/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef BLOCK_CIPHER_H
#define BLOCK_CIPHER_H

#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup BlockCipher BlockCipher
@{
**/

/*
**  "Public" Type and Method Declarations
*/

/**
 * A block cipher algorithm implementation.
 *
 * This structure contains the information and functions necessary to access a generic block cipher implementation.
 * The implementation is stricctly the cipher algorithm itself, and does not contain any mode of operation.  All
 * encryption and decryption is handled as a single block at a time (of the cipher's block size). All necessary memory
 * allocation for use of a block cipher is handled by the caller.
**/
struct BlockCipher
{

    /** The required size of the key state for this cipher algorithm. **/
    const unsigned int keyStateSize;

    /** The required block size for this cipher algorithm. */
    const unsigned int blockSize;

    /**
     * Initialize internal block cipher state for a given key.
     *
     * This call is used to establish state for use of the block cipher with the given key. The resulting state is passed to
     * the Encrypt or Decrpyt call for each block to be encrypted or decrypted. This state can serve as an optimized cache
     * of a key transform (e.g., an expanded key in AES), so that the transform does not need to be repeated for every call.
     *
     * Since only the state is passed, it is a requirement that the state include everything necessary to perform the
     * encryption or decryption (e.g., the key itself, or a transform of the key).
     *
     * The key buffer will remain valid for the entire duration of use of the key state (thus, it is safe to store a pointer
     * to the key in the key state instead of copying the key).
     *
     * This call must be thread safe, and should not depend on any non-const global data.
     *
     * This call must not be NULL.
     *
     * @param keyState [Out] The key state buffer to be initialized. This buffer must be at least @ref keyStateSize bytes
     *                 in size. This must not be NULL.
     * @param key The 32-byte initial key buffer. This must not be NULL. 
     * @return boolean  TRUE if the key state was generated
     *         succesfully.
     */
    void (* const GenerateKeyState) ( void * keyState, const void * key );

    /**
     * Encrypt a block of data.
     *
     * This call is used to encrypt a single block of plaintext (of size @ref blockSize).  All required information for the
     * encryption (e.g., key) is contained within the key state.
     *
     * This call must be thread safe, and should not depend on any non-const global data.
     *
     * This call must not be NULL. Encryption is required for all block ciphers.
     *
     * @param keyState The key state buffer. This must have been initialized by @ref GenerateKeyState. This must not be
     *                 NULL.
     * @param text The buffer containing the plaintext block to be encrypted. This buffer will contain the encrypted
     *             ciphertext after the function returnes (encryption is performed in-place in the buffer). This buffer must
     *             be at least @ref blockSize bytes in size. This must not be NULL.
     */
    void (* const Encrypt) ( const void * keyState, void * text );

    /**
     * Decrypt a block of data.
     *
     * This call is used to decrypt a single block of ciphertext (of size @ref blockSize).  All required information for the
     * decryption (e.g., key) is contained within the key state.
     *
     * This call must be thread safe, and should not depend on any non-const global data.
     *
     * This call may be NULL. With some security modes of operation, the block cipher is only used for encryption, and
     * the decryption code is never used. If it is known that decryption will never be necessary from this implementation,
     * this function may be left unimplemented. However, it is recommended to always implement decryption. Depending on how
     * the block cipher implementation is used, it may be difficult to determine that it is the lack of decryption that is
     * leading to failure.
     *
     * @param keyState The key state buffer. This must have been initialized by @ref GenerateKeyState. This must not be
     *                 NULL.
     * @param text The buffer containing the ciphertext block to be decrypted. This buffer will contain the decrypted
     *             plaintext after the function returnes (decryption is performed in-place in the buffer). This buffer must
     *             be at least @ref blockSize bytes in size. This must not be NULL.
     */
    void (* const Decrypt) ( const void * keyState, void * text );

};

/**
@}
**/


#ifdef  __cplusplus
}
#endif

#endif /* BLOCK_CIPHER_H */
