/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/blockcipher.h>


/** @defroup AES256Cipher AES256Cipher

    Implementation of a part of the FIPS PUB 197 (Rijndael) 256-bit Encryption
    and Decryption Cipher.  Only encryption is supported.

    This module is based entirely on the AES Standards Document found at
    http://csrc.nist.gov/publications/fips/fips197/fips-197.pdf.

    <b>SPECIAL NOTE:</b> The Rijndael paper defines matrices in the opposite of
    the C language. The routines affected by this are shiftRow and mixColumn
    and their inverses. The routines function identically, the row and columns
    have just been inverted. Everything else in the algorithm is not sensitive
    to this change.

@{
**/


/*
 * Note that Rijndael always uses a column size of 4, BUT it uses the opposite
 * ordering as C. This routine makes things look nicer.
 */
#define IDX(r,c)    ((unsigned char)(((r)*4) + c))

#define Nb      4
#define Nk      8
#define Nklen   (Nk*4)
#define Nr      14
#define BlkLen  (4 * Nb)
#define EKeyLen (Nb * (Nr+1) * 4)


/*
 * See FIPS PUB 197, Figure 7, page 16.
 */
static const unsigned char s_box[ 256 ] = {
    0x63, 0x7c, 0x77, 0x7b, 0xf2, 0x6b, 0x6f, 0xc5, 0x30, 0x01, 0x67, 0x2b, 0xfe, 0xd7, 0xab, 0x76,
    0xca, 0x82, 0xc9, 0x7d, 0xfa, 0x59, 0x47, 0xf0, 0xad, 0xd4, 0xa2, 0xaf, 0x9c, 0xa4, 0x72, 0xc0,
    0xb7, 0xfd, 0x93, 0x26, 0x36, 0x3f, 0xf7, 0xcc, 0x34, 0xa5, 0xe5, 0xf1, 0x71, 0xd8, 0x31, 0x15,
    0x04, 0xc7, 0x23, 0xc3, 0x18, 0x96, 0x05, 0x9a, 0x07, 0x12, 0x80, 0xe2, 0xeb, 0x27, 0xb2, 0x75,
    0x09, 0x83, 0x2c, 0x1a, 0x1b, 0x6e, 0x5a, 0xa0, 0x52, 0x3b, 0xd6, 0xb3, 0x29, 0xe3, 0x2f, 0x84,
    0x53, 0xd1, 0x00, 0xed, 0x20, 0xfc, 0xb1, 0x5b, 0x6a, 0xcb, 0xbe, 0x39, 0x4a, 0x4c, 0x58, 0xcf,
    0xd0, 0xef, 0xaa, 0xfb, 0x43, 0x4d, 0x33, 0x85, 0x45, 0xf9, 0x02, 0x7f, 0x50, 0x3c, 0x9f, 0xa8,
    0x51, 0xa3, 0x40, 0x8f, 0x92, 0x9d, 0x38, 0xf5, 0xbc, 0xb6, 0xda, 0x21, 0x10, 0xff, 0xf3, 0xd2,
    0xcd, 0x0c, 0x13, 0xec, 0x5f, 0x97, 0x44, 0x17, 0xc4, 0xa7, 0x7e, 0x3d, 0x64, 0x5d, 0x19, 0x73,
    0x60, 0x81, 0x4f, 0xdc, 0x22, 0x2a, 0x90, 0x88, 0x46, 0xee, 0xb8, 0x14, 0xde, 0x5e, 0x0b, 0xdb,
    0xe0, 0x32, 0x3a, 0x0a, 0x49, 0x06, 0x24, 0x5c, 0xc2, 0xd3, 0xac, 0x62, 0x91, 0x95, 0xe4, 0x79,
    0xe7, 0xc8, 0x37, 0x6d, 0x8d, 0xd5, 0x4e, 0xa9, 0x6c, 0x56, 0xf4, 0xea, 0x65, 0x7a, 0xae, 0x08,
    0xba, 0x78, 0x25, 0x2e, 0x1c, 0xa6, 0xb4, 0xc6, 0xe8, 0xdd, 0x74, 0x1f, 0x4b, 0xbd, 0x8b, 0x8a,
    0x70, 0x3e, 0xb5, 0x66, 0x48, 0x03, 0xf6, 0x0e, 0x61, 0x35, 0x57, 0xb9, 0x86, 0xc1, 0x1d, 0x9e,
    0xe1, 0xf8, 0x98, 0x11, 0x69, 0xd9, 0x8e, 0x94, 0x9b, 0x1e, 0x87, 0xe9, 0xce, 0x55, 0x28, 0xdf,
    0x8c, 0xa1, 0x89, 0x0d, 0xbf, 0xe6, 0x42, 0x68, 0x41, 0x99, 0x2d, 0x0f, 0xb0, 0x54, 0xbb, 0x16
};


/*
 * See FIPS PUB 197, section A.3, pages 30-32.
 * NOTE that this array only represents the high byte of the constant, and
 * is taken from the key expansion example on those pages. Since this routine
 * only supports 256-bit keys, only part of the full array (as defined in
 * the original papers on Rijndael) is presented.
 *
 * IMPORTANT NOTE: The use of this array requires a placeholder in the first
 * position, which is set to zero.
 */
static const unsigned char rco[1+((Nb*Nr)/Nk)] =
{
    0, /* This is a placeholder, not part of the definition. */
    0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40
};

/*
 * See FIPS PUB 197, section 4.2.1, page 11.
 *
 * NOTE: It is extremely important that all paths through this routine take the
 * EXACT same number of processor cycles. If this is not the case then the
 * algorithm is open to timing attacks. See section 5.1 of the original
 * Rijndael specification paper. This is not a problem here because a precomputed
 * lookup table is used.
 */
static const unsigned char xtime[256] =
{
    0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26, 28, 30, 32, 34, 36,
    38, 40, 42, 44, 46, 48, 50, 52, 54, 56, 58, 60, 62, 64, 66, 68, 70, 72, 74,
    76, 78, 80, 82, 84, 86, 88, 90, 92, 94, 96, 98, 100, 102, 104, 106, 108,
    110, 112, 114, 116, 118, 120, 122, 124, 126, 128, 130, 132, 134, 136, 138,
    140, 142, 144, 146, 148, 150, 152, 154, 156, 158, 160, 162, 164, 166, 168,
    170, 172, 174, 176, 178, 180, 182, 184, 186, 188, 190, 192, 194, 196, 198,
    200, 202, 204, 206, 208, 210, 212, 214, 216, 218, 220, 222, 224, 226, 228,
    230, 232, 234, 236, 238, 240, 242, 244, 246, 248, 250, 252, 254, 27, 25,
    31, 29, 19, 17, 23, 21, 11, 9, 15, 13, 3, 1, 7, 5, 59, 57, 63, 61, 51, 49,
    55, 53, 43, 41, 47, 45, 35, 33, 39, 37, 91, 89, 95, 93, 83, 81, 87, 85, 75,
    73, 79, 77, 67, 65, 71, 69, 123, 121, 127, 125, 115, 113, 119, 117, 107,
    105, 111, 109, 99, 97, 103, 101, 155, 153, 159, 157, 147, 145, 151, 149,
    139, 137, 143, 141, 131, 129, 135, 133, 187, 185, 191, 189, 179, 177, 183,
    181, 171, 169, 175, 173, 163, 161, 167, 165, 219, 217, 223, 221, 211, 209,
    215, 213, 203, 201, 207, 205, 195, 193, 199, 197, 251, 249, 255, 253, 243,
    241, 247, 245, 235, 233, 239, 237, 227, 225, 231, 229,
};


/*
**  "Private" Type and Method Declarations
*/
static void expandKey( void * ekey_, const void * key_ );
static void encrypt( const void * ekey_, void * data_ );
static void shiftRow( unsigned char * data );
static void mixColumn( unsigned char * data );


/*
**  Public
*/


/** External block cipher definition **/
const struct BlockCipher BlockCipher_AES_256 = {
    EKeyLen, /* keyStateSize */
    BlkLen, /* blockSize */
    expandKey, /* GenerateKeyState() */
    encrypt, /* Encrypt() */
    0 /* NULL */ /* Decrypt() */
};


/*
** Static functions
*/


/**
    Expand an encryption key for later use with @ref AES256Cipher_Encrypt. This
    expansion needs only to be done once for each key, at the expense of saving
    a longer sequence of bytes.

    @param ekey The expansion of the encryption key.
    @param key  The encryption key to expand.

    @see    FIPS 197 Section 5.2, pages 19-20;
            AES256Cipher_Encrypt
**/
static void expandKey( void * ekey_, const void * key_ )
{
    unsigned char * ekey = (unsigned char *)ekey_;
    const unsigned char * key = (const unsigned char *)key_;
    unsigned char i;
    unsigned char t0, t1, t2, t3;    /* Initialized before first use below */

    /* The first bytes are equivalent to the key. */

    for ( i = 0; i < Nklen; i++ )
        ekey[i] = key[i];


    for ( i = Nk; i < Nb * (Nr+1); i++ )
    {
        /* temp = w[i-1] */
        t0 = ekey[ (i-1)*4 ];
        t1 = ekey[ (i-1)*4 + 1 ];
        t2 = ekey[ (i-1)*4 + 2 ];
        t3 = ekey[ (i-1)*4 + 3 ];

        if ( i % Nk == 0 )
        {
            /* temp = SubWord(RotWord(temp)) xor Rcon[i/Nk] */
            unsigned char t = t0;
            t0 = s_box[ t1 ] ^ rco[ i / Nk ];
            t1 = s_box[ t2 ];
            t2 = s_box[ t3 ];
            t3 = s_box[ t ];
        }
        else if ( i % Nk == 4 )
        {
            /* temp = SubWord( temp ) */
            t0 = s_box[ t0 ];
            t1 = s_box[ t1 ];
            t2 = s_box[ t2 ];
            t3 = s_box[ t3 ];
        }

        /* w[i] = w[i-Nk] xor temp */
        ekey[ i*4 + 0 ] = ekey[ (i-Nk)*4 + 0 ] ^ t0;
        ekey[ i*4 + 1 ] = ekey[ (i-Nk)*4 + 1 ] ^ t1;
        ekey[ i*4 + 2 ] = ekey[ (i-Nk)*4 + 2 ] ^ t2;
        ekey[ i*4 + 3 ] = ekey[ (i-Nk)*4 + 3 ] ^ t3;
    }
}


/**
    Encrypt a block of plaintext.

    This routine uses the rijndael algorithm to encrypt a block of plaintext.
    The size of the block (the State) must be given and must be a multiple of
    4*Nb.

    @param ekey The expanded encryption key.
    @param data The plaintext buffer to encrypt.

    @see    FIPS 197 Section 5.1, pages 14-15
**/
static void encrypt( const void * ekey_, void * data_ )
{
    const unsigned char * ekey = (const unsigned char *)ekey_;
    unsigned char * data = (unsigned char *)data_;
    unsigned char i, j;

    /* AddRoundKey( data, w[0, Nb-1] ) */
    /* This is done for each block in the data. */
    for ( i = 0; i < BlkLen; i++ )
        data[ i ] ^= ekey[ i % (Nb*4) ];

    for ( i = 1; i < Nr; i++ )
    {
        /* SubBytes(data) */
        for ( j = 0; j < BlkLen; j++ )
            data[ j ] = s_box[ data[ j ] ];

        /* ShiftRows(data)
         * MixColumns(data) */
        /* This changes slightly because we call the routines on each block of the data. */
        for ( j = 0; j < (BlkLen / (Nb*4)); j++ )
        {
            shiftRow( data + (j * (Nb*4)) );
            mixColumn( data + (j * (Nb*4)) );
        }

        /* AddRoundKey(data, w[round*Nb, (round+1)*Nb-1]) */
        for ( j = 0; j < BlkLen; j++ )
            data[ j ] ^= ekey[ (i*(Nb*4)) + (j % (Nb*4)) ];
    }

    /* SubBytes(data) */
    for ( j = 0; j < BlkLen; j++ )
        data[ j ] = s_box[ data[ j ] ];

    /* ShiftRows(data) */
    for ( j = 0; j < (BlkLen / (Nb*4)); j++ )
        shiftRow( data + (j * (Nb*4)) );

    /* AddRoundKey(data, w[Nr*Nb, (Nr+1)*Nb-1]) */
    /* This is done for each block in the data. */
    for ( j = 0; j < BlkLen; j++ )
        data[ j ] ^= ekey[ (i*(Nb*4)) + (j % (Nb*4)) ];
}


/** @internal
    Implements the ShiftRow function as described in section 5.1.2 of the
    reference document.

    Rotates bytes in COLUMNS of the C array, which match the ROWS of the
    Rijndael document.

    @param  data    The Data block to shift.

    @see    FIPS 197 Section 5.1.2, page 17
**/
static void shiftRow( unsigned char * data )
{
    unsigned char t;     /* Initialized before first use below */

    t = data[ IDX(0,1) ];
    data[ IDX(0,1) ] = data[ IDX(1,1) ];
    data[ IDX(1,1) ] = data[ IDX(2,1) ];
    data[ IDX(2,1) ] = data[ IDX(3,1) ];
    data[ IDX(3,1) ] = t;

    t = data[ IDX(0,2) ];
    data[ IDX(0,2) ] = data[ IDX(2,2) ];
    data[ IDX(2,2) ] = t;
    t = data[ IDX(1,2) ];
    data[ IDX(1,2) ] = data[ IDX(3,2) ];
    data[ IDX(3,2) ] = t;

    t = data[ IDX(3,3) ];
    data[ IDX(3,3) ] = data[ IDX(2,3) ];
    data[ IDX(2,3) ] = data[ IDX(1,3) ];
    data[ IDX(1,3) ] = data[ IDX(0,3) ];
    data[ IDX(0,3) ] = t;
}


/** @internal
    Implements the MixColumn function as described in section 5.1.3 of the
    reference document.

    Rotates bytes in ROWS of the C array, which match the COLUMNS of the
    Rijndael document.

    @param  data    The data block to mix.

    @see    FIPS 197 Section 5.1.3, pages 17-18
**/
static void mixColumn( unsigned char * data )
{
    unsigned char tmp;       /* Initialized before first use below */

    tmp = data[IDX(0,0)] ^ data[IDX(0,1)] ^ data[IDX(0,2)] ^ data[IDX(0,3)];
    data[IDX(0,0)] ^= xtime[ (unsigned char) (data[IDX(0,0)] ^ data[IDX(0,1)] )] ^ tmp;
    data[IDX(0,1)] ^= xtime[ (unsigned char) (data[IDX(0,1)] ^ data[IDX(0,2)] )] ^ tmp;
    data[IDX(0,2)] ^= xtime[ (unsigned char) (data[IDX(0,2)] ^ data[IDX(0,3)] )] ^ tmp;
    data[IDX(0,3)] = data[IDX(0,0)] ^ data[IDX(0,1)] ^ data[IDX(0,2)] ^ tmp;

    tmp = data[IDX(1,0)] ^ data[IDX(1,1)] ^ data[IDX(1,2)] ^ data[IDX(1,3)];
    data[IDX(1,0)] ^= xtime[ (unsigned char) (data[IDX(1,0)] ^ data[IDX(1,1)] )] ^ tmp;
    data[IDX(1,1)] ^= xtime[ (unsigned char) (data[IDX(1,1)] ^ data[IDX(1,2)] )] ^ tmp;
    data[IDX(1,2)] ^= xtime[ (unsigned char) (data[IDX(1,2)] ^ data[IDX(1,3)] )] ^ tmp;
    data[IDX(1,3)] = data[IDX(1,0)] ^ data[IDX(1,1)] ^ data[IDX(1,2)] ^ tmp;

    tmp = data[IDX(2,0)] ^ data[IDX(2,1)] ^ data[IDX(2,2)] ^ data[IDX(2,3)];
    data[IDX(2,0)] ^= xtime[ (unsigned char) (data[IDX(2,0)] ^ data[IDX(2,1)] )] ^ tmp;
    data[IDX(2,1)] ^= xtime[ (unsigned char) (data[IDX(2,1)] ^ data[IDX(2,2)] )] ^ tmp;
    data[IDX(2,2)] ^= xtime[ (unsigned char) (data[IDX(2,2)] ^ data[IDX(2,3)] )] ^ tmp;
    data[IDX(2,3)] = data[IDX(2,0)] ^ data[IDX(2,1)] ^ data[IDX(2,2)] ^ tmp;

    tmp = data[IDX(3,0)] ^ data[IDX(3,1)] ^ data[IDX(3,2)] ^ data[IDX(3,3)];
    data[IDX(3,0)] ^= xtime[ (unsigned char) (data[IDX(3,0)] ^ data[IDX(3,1)] )] ^ tmp;
    data[IDX(3,1)] ^= xtime[ (unsigned char) (data[IDX(3,1)] ^ data[IDX(3,2)] )] ^ tmp;
    data[IDX(3,2)] ^= xtime[ (unsigned char) (data[IDX(3,2)] ^ data[IDX(3,3)] )] ^ tmp;
    data[IDX(3,3)] = data[IDX(3,0)] ^ data[IDX(3,1)] ^ data[IDX(3,2)] ^ tmp;
}


/**
@}
**/
