/*
**  Copyright (c) 2013-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef RECONNECTINGSTATELISTENER_H
#define RECONNECTINGSTATELISTENER_H

/** @mainpage The C-Language ReconnectingStateListener
    @section intro Introduction

        The Connection Reconnecting Listener SDK provides a simple API that monitors
        the state of connections. If a connection associated with a Reconnecting Listener
        is disconnected, the listener attempts to connect it. Using a smart back-off
        timer, the listener continues attempting to connect until it is successful.
        The listener then continues to listen, and the cycle is repeated if the connection
        becomes disconnected again.

    @section introUsage Using This Reference

        This reference defines the APIs for use by a C-based OAL application.

    @section bugs How to Report a Bug

        see the README file associated with this project.

**/

#include <dof/connection.h>

/**
 * Platform-specific public function prototype declaration specification.
 * Some platforms make use of this identifier to specially mark pulic functions (e.g., for library import/export, etc.).
 * By default, this inherits from the platform default given by @ref DOF_PCR_DECLSPEC (if defined by the PCR), or is empty.
 */
#ifndef DOF_CRL_DECLSPEC
# ifndef DOF_PCR_DECLSPEC
#  define DOF_CRL_DECLSPEC
# else
#  define DOF_CRL_DECLSPEC DOF_PCR_DECLSPEC
# endif
#endif

#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup ReconnectingStateListener ReconnectingStateListener
@{
 **/

/**
 * ReconnectingStateListener object.
 */
typedef DOFConnectionStateListener ReconnectingStateListener;

/**
    The ReconnectingStateListener callback interface.  Defined as a pointer to the @ref ReconnectingStateListenerCallbacks_t structure. Refer to the structure for more information.
**/
typedef struct ReconnectingStateListenerCallbacks_t* ReconnectingStateListenerCallbacks;

/**
 * The ReconnectingStateListener interface structure used to monitor connections.
 *
 * @ifacestruct{ReconnectingStateListenerCallbacks}
 *
 * @threadsafety
 *
 * @see ReconnectingStateListener_Create
 */
typedef struct ReconnectingStateListenerCallbacks_t
{
    /**
     * The ReconnectingStateListener interface function structure, containing a list of
     * function pointers that implement the interface.  It has the same role as
     * a vtable in C++.  This must not be NULL.
     */
    const struct ReconnectingStateListenerCallbacksFns_t
    {
        /**
         * Called to notify the user that the ReconnectingStateListener has been removed from a connection.
         *
         * This callback must not be NULL.
         *
         * @param self The callback structure given to ReconnectingStateListener_Create. This will never be NULL.
         * @param listener The ReconnectingStateListener that the callback structure was associated with. This will never be NULL.
         * @param connection The connection that the listener was removed from. This will never be NULL.
         * @param except An exception if the listener was called because of an error. This will be NULL if the ReconnectingStateListener is removed normally from the connection by the application.
         */
        void (*ConnectionRemoved)(ReconnectingStateListenerCallbacks self, ReconnectingStateListener listener, DOFConnection connection, DOFException except);

        /**
         * Called when it it safe to free the resources for this ReconnectingStateListenerCallbacks.
         *
         * This callback may be NULL to ignore notification of removal.
         *
         * @param self The callback structure given to ReconnectingStateListener_Create. This will never be NULL.
         * @param listener The ReconnectingStateListener that the callback structure was associated with. This will never be NULL.
         */
        void (*Destroyed)(ReconnectingStateListenerCallbacks self, ReconnectingStateListener listener);

    } * fns;

    /* Implementations may extend this structure. */

} ReconnectingStateListenerCallbacks_t;

/**
 * Creates a ReconnectingStateListener that can be added to connections to keep them connected if possible.
 *
 * @mustfree{ReconnectingStateListener_Destroy}
 *
 * @param callbacks The callback structure that the ReconnectingStateListener will use to send notifications. Must not be NULL.
 *
 * @return ReconnectingStateListener The ReconnectingStateListener that can be added to connections.
 *
 * @see DOFConnection_AddStateListener, DOFConnection_RemoveStateListener
 */
extern DOF_CRL_DECLSPEC ReconnectingStateListener ReconnectingStateListener_Create( ReconnectingStateListenerCallbacks callbacks );

/**
 * Sets the minimum delay in milliseconds to use between the time when a connection is initially disconnected and the initial attempt to reconnect.
 * Each time an attempt to reconnect fails, the last delay used is doubled and the new value becomes the delay before the next reconnection attempt.
 * If the calculated delay is greater than the maximum delay, the maximum delay is used.
 *
 * @param self The ReconnectingStateListener to set the minimum on. Must not be NULL.
 * @param delay The delay in ms. If this is less than 1000 then 1000 will be used.
 * @see ReconnectingStateListener_SetMaximumDelay
 */
extern DOF_CRL_DECLSPEC void ReconnectingStateListener_SetMinimumDelay(ReconnectingStateListener self, uint32 delay);

/**
 * Gets the current minimum delay from the ReconnectingStateListener.
 *
 * @param self The ReconnectingStateListener. Must not be NULL.
 *
 * @return uint32 The current minimum delay in ms.
 */
extern DOF_CRL_DECLSPEC uint32 ReconnectingStateListener_GetMinimumDelay(ReconnectingStateListener self);

/**
 * Sets the maximum delay to allow between reconnection attempts. This value should be equal to or greater than the minimum delay. If the value is less than
 * the minimum delay, the minimum delay value will be used as the maximum delay also.
 *
 * @param self The ReconnectingStateListener to set the maximum on. Must not be NULL.
 * @param delay The delay in ms. If this is less than 1000 then 1000 will be used.
 * @see ReconnectingStateListener_SetMinimumDelay
 */
extern DOF_CRL_DECLSPEC void ReconnectingStateListener_SetMaximumDelay(ReconnectingStateListener self, uint32 delay);

/**
 * Gets the current maximum delay from the ReconnectingStateListener.
 *
 * @param self The ReconnectingStateListener. Must not be NULL.
 *
 * @return uint32 The current maximum delay in ms.
 */
extern DOF_CRL_DECLSPEC uint32 ReconnectingStateListener_GetMaximumDelay(ReconnectingStateListener self);

/**
 * Destroys the ReconnectingStateListener.
 *
 * @note The ReconnectingStateListener must be removed from each connection it was added to. You must wait for the ConnectionRemoved callback to be called
 * for each connection before calling ReconnectingStateListener_Destroy to destroy the ReconnectingStateListener.
 * @note Calling @ref ReconnectingStateListener_Destroy will cause the destroyed callback for the ReconnectingStateListener to be called if one was specified.
 * The ReconnectingStateListener must not be added to new connections after calling ReconnectingStateListener_Destroy.
 *
 * @param self The ReconnectingStateListener. If NULL, no action is taken.
 *
 */
extern DOF_CRL_DECLSPEC void ReconnectingStateListener_Destroy(ReconnectingStateListener self);

/**@}**/

#ifdef  __cplusplus
}
#endif

#endif
