/*
**  Copyright (c) 2013-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/reconnectingstatelistener.h>

#include <dof/core.h>
#include <dof/error.h>
#include <dof/exception.h>
#include <dof/connection.h>
#include <dof/address.h>

#include <dof/pcr.h>

const uint8 ReconnectingStateListenerString_HexChar[] = { '0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F' };
#define ReconnectingStateListenerString_IsSpace(c)            ((c) == 0x20 || ((c) >= 0x09 && (c) <= 0x0D))
#define ReconnectingStateListenerString_IsDigit(c)            ((c) >= '0' && (c) <= '9')
#define ReconnectingStateListenerString_IsHexDigit(c)         (ReconnectingStateListenerString_IsDigit(c) || ((c) >= 'a' && (c) <= 'f') || ((c) >= 'A' && (c) <= 'F'))
#define ReconnectingStateListenerString_HexDigitToChar(c)     (ReconnectingStateListenerString_HexChar[(c)])

static uint32 ReconnectingStateListenerString_BytesToHexChar(char* buffer, uint32 dataLen, const uint8* data)
{
    uint32 offset = 0;
    uint32 i;
    for( i=1; i<=dataLen; ++i )
    {
        buffer[offset++] = ReconnectingStateListenerString_HexDigitToChar(((data[dataLen-i]) & 0xFF) >> 4);
        buffer[offset++] = ReconnectingStateListenerString_HexDigitToChar((data[dataLen-i]) & 0x0F);
    }
    return offset;
}

static void ReconnectingStateListenerString_Uint32ToHexChar(char* buffer, uint32 data)
{
    uint8 i;
    for (i = 7; i != 0xFF; i--)
    {
        buffer[i] = ReconnectingStateListenerString_HexDigitToChar( data & 0x0f );
        data >>= 4;
    }
}

const char* ReconnectingStateListenerConnectionType_StringNames[] = {"STREAM ", "DATAGRAM ", "DATAGRAM_STATELESS ", "HUB ", "POINT ", "GROUP ", "GROUP_MEMBER_ONLY "};
const uint8 ReconnectingStateListenerConnectionType_StringNamesLen[] = {7, 9, 19, 4, 6, 6, 18};

const char* ReconnectingStateListener_Log_StringHeaders[] = {"Connection Address: ", "Connection Type: ", "Exception: ", "Delay(ms): ", "Attempt Cnt: ", "ReconnectingStateListener: ", "Failed to connect. ", "Connection is connected. ", "Connection is disconnected. ", "Attempting to connect. "};
const uint8 ReconnectingStateListener_Log_StringHeadersLen[] = {20, 17, 11, 11, 13, 27, 19, 25, 28, 23};

static uint32 ReconnectingStateListener_Log_GetConnectionInfoLength(DOFConnectionState state, boolean withException)
{
    /* "Connection Address: " " Connection Type: " " Exception: " */
    return ReconnectingStateListener_Log_StringHeadersLen[0] + (DOFAddress_GetAddressSize(DOFConnectionState_GetAddress(state)) * 2)+ 1 + 
           ReconnectingStateListener_Log_StringHeadersLen[1] + ReconnectingStateListenerConnectionType_StringNamesLen[DOFConnectionState_GetConnectionType(state)] + 
           ((withException && DOFConnectionState_GetException(state) && DOFErrorException_GetMessage(DOFConnectionState_GetException(state)))? ( ReconnectingStateListener_Log_StringHeadersLen[2] + PCRString_Length(DOFErrorException_GetMessage(DOFConnectionState_GetException(state))) + 1): 0);
}


/* Buffer must be alloced to fit all the info call get length to get the length needed memory corruption will occur if lenght is too small */
static uint32 ReconnectingStateListener_Log_AddConnectionInfo(char* buffer, DOFConnectionState state, boolean withException)
{
    uint32 offset = 0;
    DOFException except;
    DOFAddress addr = DOFConnectionState_GetAddress(state);
    uint8 connType = DOFConnectionState_GetConnectionType(state);
    
    /* Add Connection Address header */
    PCRMem_Copy(buffer+offset, ReconnectingStateListener_Log_StringHeaders[0], ReconnectingStateListener_Log_StringHeadersLen[0]);
    offset += ReconnectingStateListener_Log_StringHeadersLen[0];
    /* Connection Address Data */
    offset += ReconnectingStateListenerString_BytesToHexChar(buffer+offset, DOFAddress_GetAddressSize(addr), DOFAddress_GetAddress(addr));
    buffer[offset++] = ' ';

    /* Add Connection Type Header */
    PCRMem_Copy(buffer+offset, ReconnectingStateListener_Log_StringHeaders[1], ReconnectingStateListener_Log_StringHeadersLen[1]);
    offset += ReconnectingStateListener_Log_StringHeadersLen[1];
    /* Add Connection Type String  */
    PCRMem_Copy(buffer+offset, ReconnectingStateListenerConnectionType_StringNames[connType], ReconnectingStateListenerConnectionType_StringNamesLen[connType]);
    offset += ReconnectingStateListenerConnectionType_StringNamesLen[connType];

    /* Add except */
    except = DOFConnectionState_GetException(state);
    if( withException && except )
    {
        const char* exceptString = DOFErrorException_GetMessage( except );
        if( exceptString )
        {
            /* Add except header */
            PCRMem_Copy(buffer+offset, ReconnectingStateListener_Log_StringHeaders[2], ReconnectingStateListener_Log_StringHeadersLen[2]);
            offset += ReconnectingStateListener_Log_StringHeadersLen[2];
            /* Add except String */
            PCRMem_Copy(buffer+offset, exceptString, PCRString_Length(exceptString));
            offset += PCRString_Length(exceptString)+1;
        }
    }

    return offset;
}

static uint32 ReconnectingStateListener_Log_GetConnectionAttemptInfoLen(uint32 delay, uint32 attemptCount)
{
    /* "Delay(ms): " "Attempt Cnt: " */
    return ReconnectingStateListener_Log_StringHeadersLen[3] + 9 + ReconnectingStateListener_Log_StringHeadersLen[4] + 9;
}

static uint32 ReconnectingStateListener_Log_AddConnectionAttemptInfo(char* buffer, uint32 delay, uint32 attemptCount)
{
    uint32 offset = 0;
    
    /* Add delay header */
    PCRMem_Copy(buffer+offset, ReconnectingStateListener_Log_StringHeaders[3], ReconnectingStateListener_Log_StringHeadersLen[3]);
    offset += ReconnectingStateListener_Log_StringHeadersLen[3];
    /* delay Data */
    ReconnectingStateListenerString_Uint32ToHexChar(buffer+offset, delay);
    offset += 8;
    buffer[offset++] = ' ';

    /* Add attempt header */
    PCRMem_Copy(buffer+offset, ReconnectingStateListener_Log_StringHeaders[4], ReconnectingStateListener_Log_StringHeadersLen[4]);
    offset += ReconnectingStateListener_Log_StringHeadersLen[4];
    /* attempt Data */
    ReconnectingStateListenerString_Uint32ToHexChar(buffer+offset, attemptCount);
    offset += 8;
    buffer[offset++] = ' ';

    return offset;
}

static void ReconnectingStateListener_Log_FailureToConnect(DOFConnection connection, DOFException except)
{
    /* Warn: failure to connect, address bytes, connection_type, exception string */
    DOFConnectionState state = DOFConnection_CreateState(connection);
    if( state )
    {
        uint32 len = ReconnectingStateListener_Log_StringHeadersLen[5] + ReconnectingStateListener_Log_StringHeadersLen[6] + ReconnectingStateListener_Log_GetConnectionInfoLength(state, TRUE) + 1;
        char* str = (char*)PCRMem_Alloc(len);
        if( str ) 
        {
            uint32 offset = 0;
            PCRMem_Set(str, 0, len);

            /* Add Header */
            PCRMem_Copy(str+offset, ReconnectingStateListener_Log_StringHeaders[5], ReconnectingStateListener_Log_StringHeadersLen[5] );
            offset += ReconnectingStateListener_Log_StringHeadersLen[5];

            /* Add type string. */
            PCRMem_Copy(str+offset, ReconnectingStateListener_Log_StringHeaders[6], ReconnectingStateListener_Log_StringHeadersLen[6] );
            offset += ReconnectingStateListener_Log_StringHeadersLen[6];

            /* Add connection info */
            offset += ReconnectingStateListener_Log_AddConnectionInfo(str+offset, state, TRUE);

            DOFLog_Message(DOFLOGLEVEL_WARNING, str);

            DOFConnectionState_Destroy(state);
            PCRMem_Free(str);
            return;
        }
        DOFConnectionState_Destroy(state);
    }
    DOFLog_Message(DOFLOGLEVEL_WARNING, "ReconnectingStateListener: Failed to connect. (No resouces to build info string.)");
}

static void ReconnectingStateListener_Log_ConnectAttempt(DOFConnection connection, uint32 delay, uint32 attemptCount)
{
    /* Info: Attemp to connect bytes address, connection_type, delay, # of attempts */
    DOFConnectionState state = DOFConnection_CreateState(connection);
    if( state )
    {
        uint32 len = ReconnectingStateListener_Log_StringHeadersLen[5] + ReconnectingStateListener_Log_StringHeadersLen[9] + ReconnectingStateListener_Log_GetConnectionInfoLength(state, FALSE) + ReconnectingStateListener_Log_GetConnectionAttemptInfoLen(delay, attemptCount) + 1;
        char* str = (char*)PCRMem_Alloc(len);
        if( str ) 
        {
            uint32 offset = 0;
            PCRMem_Set(str, 0, len);

            /* Add Header */
            PCRMem_Copy(str+offset, ReconnectingStateListener_Log_StringHeaders[5], ReconnectingStateListener_Log_StringHeadersLen[5] );
            offset += ReconnectingStateListener_Log_StringHeadersLen[5];

            /* Add type string. */
            PCRMem_Copy(str+offset, ReconnectingStateListener_Log_StringHeaders[9], ReconnectingStateListener_Log_StringHeadersLen[9] );
            offset += ReconnectingStateListener_Log_StringHeadersLen[9];

            /* Add connection info */
            offset += ReconnectingStateListener_Log_AddConnectionInfo(str+offset, state, FALSE);

            /* Add Attempt info */
            offset += ReconnectingStateListener_Log_AddConnectionAttemptInfo(str+offset, delay, attemptCount);

            DOFLog_Message(DOFLOGLEVEL_INFO, str);

            DOFConnectionState_Destroy(state);
            PCRMem_Free(str);
            return;
        }
        DOFConnectionState_Destroy(state);
    }
    DOFLog_Message(DOFLOGLEVEL_INFO, "ReconnectingStateListener: Attempting to connect. (No resouces to build info string.)");
}

static void ReconnectingStateListener_Log_Disconnected(DOFConnectionState state)
{
    /* Info: Disconnected address bytes, connection_type, (Exception string) */
    uint32 len = ReconnectingStateListener_Log_StringHeadersLen[5] + ReconnectingStateListener_Log_StringHeadersLen[8] + ReconnectingStateListener_Log_GetConnectionInfoLength(state, TRUE) + 1;
    char* str = (char*)PCRMem_Alloc(len);
    if( str ) 
    {
        uint32 offset = 0;
        PCRMem_Set(str, 0, len);

        /* Add Header */
        PCRMem_Copy(str+offset, ReconnectingStateListener_Log_StringHeaders[5], ReconnectingStateListener_Log_StringHeadersLen[5] );
        offset += ReconnectingStateListener_Log_StringHeadersLen[5];

        /* Add type string. */
        PCRMem_Copy(str+offset, ReconnectingStateListener_Log_StringHeaders[8], ReconnectingStateListener_Log_StringHeadersLen[8] );
        offset += ReconnectingStateListener_Log_StringHeadersLen[8];

        /* Add connection info */
        offset += ReconnectingStateListener_Log_AddConnectionInfo(str+offset, state, TRUE);

        DOFLog_Message(DOFLOGLEVEL_INFO, str);
        PCRMem_Free(str);
        return;
    }
    DOFLog_Message(DOFLOGLEVEL_INFO, "ReconnectingStateListener: Connection is disconnected. (No resouces to build info string.)");
}

static void ReconnectingStateListener_Log_Connected(DOFConnectionState state)
{
    /* Info: Connected address bytes, connection_type */
    uint32 len = ReconnectingStateListener_Log_StringHeadersLen[5] + ReconnectingStateListener_Log_StringHeadersLen[7] + ReconnectingStateListener_Log_GetConnectionInfoLength(state, FALSE) + 1;
    char* str = (char*)PCRMem_Alloc(len);
    if( str ) 
    {
        uint32 offset = 0;
        PCRMem_Set(str, 0, len);

        /* Add Header */
        PCRMem_Copy(str+offset, ReconnectingStateListener_Log_StringHeaders[5], ReconnectingStateListener_Log_StringHeadersLen[5] );
        offset += ReconnectingStateListener_Log_StringHeadersLen[5];

        /* Add type string. */
        PCRMem_Copy(str+offset, ReconnectingStateListener_Log_StringHeaders[7], ReconnectingStateListener_Log_StringHeadersLen[7] );
        offset += ReconnectingStateListener_Log_StringHeadersLen[7];

        /* Add connection info */
        offset += ReconnectingStateListener_Log_AddConnectionInfo(str+offset, state, FALSE);

        DOFLog_Message(DOFLOGLEVEL_INFO, str);
        PCRMem_Free(str);
        return;
    }
    DOFLog_Message(DOFLOGLEVEL_INFO, "ReconnectingStateListener: Connection is connected. (No resouces to build info string.)");
}

#define RECONNECT_TIMEOUT 20000

typedef struct ConnectionStateNode_t
{
    struct ConnectionStateNode_t*   next;
    DOFConnection                   connection;
    uint32                          timeToReconnect;
    uint32                          attemptCnt;
    boolean                         connected;
}ConnectionStateNode_t;

typedef struct ReconnectingStateListener_t
{
    DOFConnectionStateListener_t        base;
    ReconnectingStateListenerCallbacks  callbacks;
    PCRThread                           reconnectionThread;
    PCRMutex                            reconnectionMutex;
    ConnectionStateNode_t*              connectionListHead;
    uint32                              delayMin;
    uint32                              delayMax;
    boolean                             destroyed;
}ReconnectingStateListener_t;

static void ReconnectingStateListener_Internal_Destroy(ReconnectingStateListener_t* self);

static void ConnectionListenerRemoved( DOFConnectionStateListener self, DOFConnection connection, DOFException except );
static void ConnectionStateChanged( DOFConnectionStateListener self, DOFConnection connection, DOFConnectionState state );

const struct DOFConnectionStateListenerFns_t reconnectFns = {ConnectionListenerRemoved, ConnectionStateChanged};

static void ConnectComplete( DOFConnectionConnectCallback self, DOFConnection connection, void * context, DOFException except )
{
    /* Handle connect vs reconnect */
    if( !except )
    {
        ConnectionStateNode_t* connState = (ConnectionStateNode_t*)context;

        if( !connState->connected )
            connState->connected = TRUE;
        else
            DOFConnection_Disconnect(connection); /* Debump connection count */
    }
    else
    {
        ReconnectingStateListener_Log_FailureToConnect( connection, except ); /* LOG */
    }
}

static struct DOFConnectionConnectCallbackFns_t connectFns = {ConnectComplete};
static struct DOFConnectionConnectCallback_t connectListener = {&connectFns};

static void ReconnectComplete( DOFConnectionConnectCallback self, DOFConnection connection, void * context, DOFException except )
{
    /* Handle reconnect */
    if( except )
    {
        ReconnectingStateListener_Log_FailureToConnect( connection, except ); /* LOG */
    }
}

static struct DOFConnectionConnectCallbackFns_t reconnectCallbackFns = {ReconnectComplete};
static struct DOFConnectionConnectCallback_t reconnectListener = {&reconnectCallbackFns};

static uint32 CalulateReconnectDelay(ReconnectingStateListener_t* listener, ConnectionStateNode_t* node, boolean incAttempt)
{
    uint32 max = (listener->delayMax < listener->delayMin) ? listener->delayMin : listener->delayMax;
    uint32 delay = max;

    /* Handle rollover */
    if( incAttempt && node->attemptCnt < 4294967295 )
        node->attemptCnt += 1;

    if( node->attemptCnt < 31 )
    {
        uint32 tempDelay = listener->delayMin << node->attemptCnt;
        if( tempDelay != 0 )
            delay = tempDelay;
    }

    delay = (delay < max)? delay: max;

    /* Up to half calculated delay random backoff */
    return ( delay - (PCRRandom_Get() % (delay / 2)) );
}

static void reconnectionThreadFn(void* args)
{
    ReconnectingStateListener_t* rsl = (ReconnectingStateListener_t*)args;

    boolean destroyed = FALSE;

    do
    {
        /* Travers list of connection attempt reconnection of */
        ConnectionStateNode_t* curr = rsl->connectionListHead;

        PCRMutex_Lock(rsl->reconnectionMutex);
        while( curr )
        {
            if( curr->timeToReconnect != 0 )
            {
                curr->timeToReconnect -= 1000;

                /* Less than a second left do reconnect */
                if( curr->timeToReconnect < 1000 )
                {
                    curr->timeToReconnect = CalulateReconnectDelay(rsl, curr, TRUE);
                    ReconnectingStateListener_Log_ConnectAttempt(curr->connection, curr->timeToReconnect, curr->attemptCnt); /* LOG */
                    if( curr->connected > 0 )
                        DOFConnection_BeginReconnect(curr->connection, ((curr->timeToReconnect < RECONNECT_TIMEOUT) ? RECONNECT_TIMEOUT : curr->timeToReconnect), &reconnectListener, NULL);
                    else
                        DOFConnection_BeginConnect(curr->connection, ((curr->timeToReconnect < RECONNECT_TIMEOUT)? RECONNECT_TIMEOUT: curr->timeToReconnect), &connectListener, curr);
                }
            }
            curr = curr->next;
        }

        destroyed = rsl->destroyed;
        PCRMutex_Unlock(rsl->reconnectionMutex);

        /* lazy timer */
        PCRThread_Sleep(1000);
    } while ( !destroyed );
}

static ConnectionStateNode_t* FindConnection(ConnectionStateNode_t* head, DOFConnection connection)
{
    ConnectionStateNode_t* curr = head;
    while( curr )
    {
        if( curr->connection == connection )
            break;
        curr = curr->next;
    }
    return curr;
}

static ConnectionStateNode_t* FindAndRemove(ConnectionStateNode_t** head, DOFConnection connection)
{
    ConnectionStateNode_t* prev = *head;
    ConnectionStateNode_t* curr = (*head)->next;

    if( prev->connection == connection )
    {
        *head = curr;
        return prev;
    }

    while( curr )
    {
        if( curr->connection == connection )
        {
            /* Remove and return*/
            prev->next = curr->next;
            return curr;
        }
        prev = curr;
        curr = curr->next;
    }
    return NULL;
}

static void ConnectionStateChanged( DOFConnectionStateListener self, DOFConnection connection, DOFConnectionState state )
{
    ConnectionStateNode_t* connState;
    ReconnectingStateListener_t* listener = (ReconnectingStateListener_t*)self;
    boolean connected = DOFConnectionState_IsConnected(state);

    PCRMutex_Lock(listener->reconnectionMutex);
    connState = FindConnection(listener->connectionListHead, connection);
    if( connState )
    {
        /* known connection */
        if( connected )
        {
            /* Clear times */
            connState->attemptCnt = 0;
            connState->timeToReconnect = 0;
            ReconnectingStateListener_Log_Connected(state);
        }
        else
        {
            /* Set next attempt time */
            connState->timeToReconnect = CalulateReconnectDelay(listener, connState, FALSE);
            ReconnectingStateListener_Log_Disconnected(state);
        }
        PCRMutex_Unlock(listener->reconnectionMutex);
    }
    else
    {
        connState = (ConnectionStateNode_t*)PCRMem_Alloc(sizeof(ConnectionStateNode_t));
        if( connState )
        {
            PCRMem_Set(connState, 0, sizeof(ConnectionStateNode_t));

            connState->connection = connection;

            /* Add to list*/
            connState->next = listener->connectionListHead;
            listener->connectionListHead = connState;

            /* new connection */
            if( !connected )
            {
                connState->timeToReconnect = CalulateReconnectDelay(listener, connState, FALSE);
                DOFConnection_BeginConnect(connection, ((connState->timeToReconnect < RECONNECT_TIMEOUT)? RECONNECT_TIMEOUT: connState->timeToReconnect), &connectListener, connState);
                ReconnectingStateListener_Log_Disconnected(state);
            }
            else
            {
                DOFConnection_BeginConnect(connection, ((connState->timeToReconnect < RECONNECT_TIMEOUT)? RECONNECT_TIMEOUT: connState->timeToReconnect), &connectListener, connState); /* this is to bump count count */
                ReconnectingStateListener_Log_Connected(state);
            }
            
            PCRMutex_Unlock(listener->reconnectionMutex);
        }
        else
        {
            DOFException except;
            /* Can't save, Exception */
            PCRMutex_Unlock(listener->reconnectionMutex);
            except = DOFErrorException_Create(DOFERROR_INSUFFICIENT_RESOURCES, "Insufficient Resources");
            DOFConnection_RemoveStateListener(connection,self);
            listener->callbacks->fns->ConnectionRemoved(listener->callbacks, (ReconnectingStateListener)listener, connection, except);
            DOFException_Destroy(except);
        }
    }
}

static void ConnectionListenerRemoved( DOFConnectionStateListener self, DOFConnection connection, DOFException except )
{
    ConnectionStateNode_t* connState;
    ReconnectingStateListener_t* listener = (ReconnectingStateListener_t*)self;

    PCRMutex_Lock(listener->reconnectionMutex);
    connState = FindConnection(listener->connectionListHead, connection);
    PCRMutex_Unlock(listener->reconnectionMutex);
    if( connState )
    {
        DOFConnection_Disconnect(connState->connection);

        /* call app */
        if( listener->callbacks->fns->ConnectionRemoved )
            listener->callbacks->fns->ConnectionRemoved(listener->callbacks, (ReconnectingStateListener)listener, connection, NULL);

        PCRMutex_Lock(listener->reconnectionMutex);
        connState = FindAndRemove(&listener->connectionListHead, connection);
        PCRMutex_Unlock(listener->reconnectionMutex);

        PCRMem_Free(connState);

        ReconnectingStateListener_Internal_Destroy(listener);
    }
}

ReconnectingStateListener ReconnectingStateListener_Create( ReconnectingStateListenerCallbacks callbacks )
{
    ReconnectingStateListener_t* self = NULL;
    if( callbacks )
    {
        self = (ReconnectingStateListener_t*)PCRMem_Alloc(sizeof(ReconnectingStateListener_t));
        if(  self )
        {
            PCRMem_Set(self, 0, sizeof(ReconnectingStateListener_t));
            self->base.fns = &reconnectFns;

            self->callbacks = callbacks;
            self->delayMax = 300000;
            self->delayMin = 40000;

            self->reconnectionMutex = PCRMutex_Create();
            if( self->reconnectionMutex )
            {
                PCRMutex_Lock(self->reconnectionMutex);
                self->reconnectionThread = PCRThread_Create(reconnectionThreadFn, self);
                if( self->reconnectionThread )
                {
                    PCRMutex_Unlock(self->reconnectionMutex);
                    return (ReconnectingStateListener)self;
                }
                PCRMutex_Unlock(self->reconnectionMutex);
                PCRMutex_Destroy(self->reconnectionMutex);
            }
            PCRMem_Free(self);
            self = NULL;
        }
    }
    return (ReconnectingStateListener)self;
}

static void ReconnectingStateListener_Internal_Destroy(ReconnectingStateListener_t* self)
{
    PCRMutex_Lock(self->reconnectionMutex);
    if( self->destroyed )
    {
        if( !self->connectionListHead )
        {
            PCRMutex_Unlock(self->reconnectionMutex);
            if(self->callbacks->fns->Destroyed)
                self->callbacks->fns->Destroyed(self->callbacks, (ReconnectingStateListener)self);

            PCRMutex_Destroy(self->reconnectionMutex);
            PCRMem_Free(self);
            return;
        }
    }
    PCRMutex_Unlock(self->reconnectionMutex);
}

void ReconnectingStateListener_Destroy(ReconnectingStateListener self)
{
    if( self )
    {
        PCRMutex_Lock(((ReconnectingStateListener_t*)self)->reconnectionMutex);
        ((ReconnectingStateListener_t*)self)->destroyed = TRUE;
        PCRMutex_Unlock(((ReconnectingStateListener_t*)self)->reconnectionMutex);

        PCRThread_Join(((ReconnectingStateListener_t*)self)->reconnectionThread);
        PCRThread_Destroy(((ReconnectingStateListener_t*)self)->reconnectionThread);

        ReconnectingStateListener_Internal_Destroy(((ReconnectingStateListener_t*)self));
    }
}

void ReconnectingStateListener_SetMinimumDelay(ReconnectingStateListener self, uint32 delay)
{
    if( self )
    {
        PCRMutex_Lock(((ReconnectingStateListener_t*)self)->reconnectionMutex);
        ((ReconnectingStateListener_t*)self)->delayMin = (delay < 1000)? 2000: delay*2;
        PCRMutex_Unlock(((ReconnectingStateListener_t*)self)->reconnectionMutex);
    }
}

uint32 ReconnectingStateListener_GetMinimumDelay(ReconnectingStateListener self)
{
    if( self )
        return ((ReconnectingStateListener_t*)self)->delayMin/2;
    return 0;
}

void ReconnectingStateListener_SetMaximumDelay(ReconnectingStateListener self, uint32 delay)
{
    if( self )
    {
        PCRMutex_Lock(((ReconnectingStateListener_t*)self)->reconnectionMutex);
        ((ReconnectingStateListener_t*)self)->delayMax = (delay < 1000)? 1000: delay;
        PCRMutex_Unlock(((ReconnectingStateListener_t*)self)->reconnectionMutex);
    }
}

uint32 ReconnectingStateListener_GetMaximumDelay(ReconnectingStateListener self)
{
    if( self )
        return ((ReconnectingStateListener_t*)self)->delayMax;
    return 0;
}

