/*
**  Copyright (c) 2014-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef _DOF_INET_H
#define _DOF_INET_H

/** @mainpage The C-Language Inet Posix Transport
    @section intro Introduction

        The POSIX Internet Transport is a C-language implementation of a POSIX,
        Berkeley socket-based transport for use with the C OAL.

    @section introUsage Using This Reference

        This reference defines the APIs for use by a C-based OAL application.

    @section bugs How to Report a Bug

        see the README file associated with this project.

**/

#include <dof/pcr.h>
#include <dof/transport.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup Inet Inet Transport

    The Inet Transport is an implementation of the DOF Transport Layer (see
    @ref DTL) for TCP/IP and UDP/IP over IPv4 and IPv6.

@{
**/

/* Externally defined. */
struct sockaddr;

/** @file
    Header file for the Inet Transport interfaces, functions, defines and data types.
**/

/*
**  "Public" Type and Method Declarations
*/

/**
 * Platform-specific public function prototype declaration specification.
 * Some platforms make use of this identifier to specially mark public functions (e.g., for library import/export, etc.).
 * By default, this inherits from the platform default given by @ref DOF_PCR_DECLSPEC (if defined by the PCR), or is empty.
 */
#ifndef DOF_INET_DECLSPEC
# ifndef DOF_PCR_DECLSPEC
#  define DOF_INET_DECLSPEC
# else
#  define DOF_INET_DECLSPEC DOF_PCR_DECLSPEC
# endif
#endif

/** Default DOF port for TCP and UDP servers. **/
#define INETTRANSPORT_DEFAULT_UNICAST_PORT       (3567U)

/** Default DOF port for UDP multicast servers. **/
#define INETTRANSPORT_DEFAULT_MULTICAST_PORT     (5567U)

/** Default DOF IPv4 multicast address. **/
extern DOF_INET_DECLSPEC const uint8 INETTRANSPORT_DEFAULT_MULTICAST_ADDRESS_IPV4[4];

/** Default DOF IPv6 multicast address. **/
extern DOF_INET_DECLSPEC const uint8 INETTRANSPORT_DEFAULT_MULTICAST_ADDRESS_IPV6_NODE[16];

/** Default DOF IPv6 multicast address. **/
extern DOF_INET_DECLSPEC const uint8 INETTRANSPORT_DEFAULT_MULTICAST_ADDRESS_IPV6_LINK[16];

/** Default DOF IPv6 multicast address. **/
extern DOF_INET_DECLSPEC const uint8 INETTRANSPORT_DEFAULT_MULTICAST_ADDRESS_IPV6_SUBNET[16];

/** Default DOF IPv6 multicast address. **/
extern DOF_INET_DECLSPEC const uint8 INETTRANSPORT_DEFAULT_MULTICAST_ADDRESS_IPV6_ADMIN[16];

/** Default DOF IPv6 multicast address. **/
extern DOF_INET_DECLSPEC const uint8 INETTRANSPORT_DEFAULT_MULTICAST_ADDRESS_IPV6_SITE[16];

/** Default DOF IPv6 multicast address. **/
extern DOF_INET_DECLSPEC const uint8 INETTRANSPORT_DEFAULT_MULTICAST_ADDRESS_IPV6_ORG[16];

/** Default DOF IPv6 multicast address. **/
extern DOF_INET_DECLSPEC const uint8 INETTRANSPORT_DEFAULT_MULTICAST_ADDRESS_IPV6_GLOBAL[16];

/**
 * Initialize the Inet transport implementation for use.
 *
 * @note    This function @b must be called prior to any Inet related code
 *          (including the @ref InetTransport_CreateAddress functions).  The
 *          Inet implementation will assume this function has been called and
 *          will not perform a check.
 *
 * @warning If this function is not called prior to attempting to use the Inet
 *          transport, behavior is undefined.
 *
 * @sa      InetTransport_Shutdown
 * @sa      DOF_Init
 */
extern DOF_INET_DECLSPEC void InetTransport_Init(void);

/**
 * Release any state the Inet transport is currently maintaining, and return to an uninitialized state.
 * The Inet Transport may be reused only if it is initialized again.
 *
 * @note This function must be called after final use of any Inet related code, to release resources used by the Inet Transport.
 *
 * @warning Attempting to use the Inet transport after calling this function results in undefined behavior.
 *
 * @sa      InetTransport_Init
 * @sa      DOF_Shutdown
 */
extern DOF_INET_DECLSPEC void InetTransport_Shutdown(void);

/**
 * Creates a new address from hostname and port. 
 * @note Hostname will be resolved and reresolved when needed.
 * 
 * @param hostname The hostname this should not be NULL,
 * @param port The port number.
 *  
 * @mustfree{DOFAddress_Destroy} 
 * 
 * @return DOFAddress The address, or NULL if the address could not be created.
 */
extern DOF_INET_DECLSPEC DOFAddress InetTransport_CreateAddress( const char* hostname, uint16 port );

/**
 * Create a new address from hostname, port and an interface name e.g. eth0, wlan0. 
 *  
 * @note Hostname will be resolved and reresolved when needed. 
 * @note The interface index will only be used to force a multicast server to listen on a specific interface; it is ignored for all other address types and uses.
 *  
 * @param hostname The hostname this should not be NULL.
 * @param port The port number. 
 * @param iface The interface name over which a multicast address is valid.  0 (the default) means the platform will choose the most appropriate interface.
 * @return The new address, or NULL if one could not be created. If the SO_BINDTODEVICE socket option is not supported and the interface is greater than 0, this function will return NULL.
 *
 * @mustfree{DOFAddress_Destroy}
 */
extern DOF_INET_DECLSPEC DOFAddress InetTransport_CreateAddress_Interface( const char* hostname, uint16 port, const char* iface );

/**
 * Create a new address from an existing, standard sockaddr structure. 
 * The platform will choose the most appropriate interface to use.
 * @param addr The standard, generic sockaddr structure.  This should not be NULL.
 * @param addrSize The size of the sockaddr structure.
 * @return The new address, or NULL if one could not be created.
 *
 * @mustfree{DOFAddress_Destroy}
 */
extern DOF_INET_DECLSPEC DOFAddress InetTransport_CreateAddress_Sockaddr( const struct sockaddr * addr, uint32 addrSize );

/**
 * Create a new address from an existing, standard sockaddr structure and an interface index.
 * The interface index will only be used to force a multicast server to listen on a specific interface; it is ignored for all other address types and uses.
 * @param addr The standard, generic sockaddr structure.  This should not be NULL.
 * @param addrSize The size of the sockaddr structure.
 * @param iface The interface name over which a multicast address is valid. 
 * @return The new address, or NULL if one could not be created. If the SO_BINDTODEVICE socket option is not supported this function will return NULL.
 *
 * @mustfree{DOFAddress_Destroy}
 */
extern DOF_INET_DECLSPEC DOFAddress InetTransport_CreateAddress_Sockaddr_Interface( const struct sockaddr * addr, uint32 addrSize, const char* iface );

/**
 * Create a new IPv4 address given an address and a port.
 * @param addr The IPv4 address.  This should not be NULL.
 * @param port The port.
 * @return The new address, or NULL if one could not be created.
 *
 * @mustfree{DOFAddress_Destroy}
 */
extern DOF_INET_DECLSPEC DOFAddress InetTransport_CreateIPv4Address( const uint8 addr[4], uint16 port );

/**
 * Create a new IPv6 address given an address and a port.
 * @param addr The IPv6 address.  This should not be NULL.
 * @param port The port.
 * @return The new address, or NULL if one could not be created.
 *
 * @mustfree{DOFAddress_Destroy}
 */
extern DOF_INET_DECLSPEC DOFAddress InetTransport_CreateIPv6Address( const uint8 addr[16], uint16 port );

/**
 * Return the standard sockaddr structure represented by the @ref DOFAddress.
 *
 * @note This returns a pointer to an internal copy of the sockaddr structure
 * owned by the @ref DOFAddress.  It must not be used outside of the scope of the
 * @ref DOFAddress.
 *
 * @param addr The address.  This should not be NULL.
 * @return The standard sockaddr structure, or NULL if @a addr is NULL.
 */
extern DOF_INET_DECLSPEC const struct sockaddr * InetTransport_GetAddress( const DOFAddress addr );

/**
 * Return the size of the sockaddr structure available through @ref InetTransport_GetAddress.
 * @param addr The address.  This should not be NULL.
 * @return The size of the sockaddr structure, or 0 if @a addr is NULL.
 */
extern DOF_INET_DECLSPEC uint32 InetTransport_GetAddressSize( const DOFAddress addr );

/**
 * Return the interface index of the InetTransport address.
 * @param addr The address.  This should not be NULL.
 * @return The interface associated with the address, or 0 if no specific interface is associated.
 */
extern DOF_INET_DECLSPEC uint32 InetTransport_GetAddressInterface( const DOFAddress addr );


/**
@}
**/


#ifdef  __cplusplus
}
#endif

#endif /* _DOF_INET_H */

