/*
**  Copyright (c) 2014-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_HASHTABLE_H
#define DOF_HASHTABLE_H

#include <dof/core.h>
#include <internal/avl_multitree.h>

#ifdef __cplusplus
extern "C" {
#endif

/* @addtogroup OALHashtable
*   A growable hashtable

*   The OALHashtable implements a chained list implementation where the
*   first level of the table has n linked lists where n is some power of 2
*
@{
**/

typedef struct OALHashtable_t *OALHashtable;

/* These typedef declarations are const to ensure that no hashtable routine causes modification. They are still castable. */
typedef const void *OALHashItem_Ret_t;
typedef const void *OALHashItem_t;
typedef const void *OALHashKey_t;

typedef boolean (*OALHashtable_Traversal_fn_t)(OALHashItem_t item, void *arg);
typedef boolean (*OALHashtable_RemoveAll_fn_t)(OALHashItem_t item, void *arg);

typedef uint16  OALHashKeySize_t;

typedef enum OALHashtable_InsertionStatus
{
    OALHashtable_InsertFailure   = 0,     /* Force to be equal to FALSE. Must match OALAVLTree_InsertFailure. */
    OALHashtable_InsertSuccess   = 1,     /* Force to be equal to TRUE. Must match OALAVLTree_InsertSuccess. */
    OALHashtable_AlreadyInserted          /* Value doesn't matter, but must match OALAVLTree_AlreadyInserted. */
} OALHashtable_InsertionStatus;

/* @internal
*   Create a hashtable.

    @mustfree{OALThreadPool_Destroy}

    @param num_keys The number of keys to initialize the hashtable with.
    @return An initialized hashtable.

    @see OALHashtable_Destroy
*/
extern OALHashtable OALHashtable_Create(uint8 num_keys);

/* @internal
*   Destroy a hashtable and free up all memory associated with it.

    @param hashtable The hashtable to destroy. If the value is @c NULL, no action is taken.

    @see OALHashtable_Create
*/
extern void OALHashtable_Destroy(OALHashtable hashtable, OALHashtable_RemoveAll_fn_t remove_fn, void *arg);

/* @internal
*   Traverse all items within a hashtable.

    @param hashtable The hashtable to traverse.
    @param traverse_fn The function to be called for each node in the hashtable.
    @param arg The argument to pass to traverse_fn.
*/
extern void OALHashtable_Traverse_All(OALHashtable hashtable, OALAVLMultiTree_Traverse_All_fn_t traverse_fn, void *arg);

/* @internal
*   Traverse all items within a hashtable of a specific key.

    @param hashtable The hashtable to traverse.
    @param keys The key indicating the items to traverse.
    @param key_size The size of the key associated with the items to traverse.
    @param traverse_fn The function to be called for each node in the hashtable.
    @param arg The argument to pass to traverse_fn.
*/
extern void OALHashtable_Traverse(OALHashtable hashtable, OALHashKey_t *keys, OALHashKeySize_t *key_sizes, OALHashtable_Traversal_fn_t traverse_fn, void *arg);

/* @internal
*   Insert an item into the hashtable.

    @param hashtable The hashtable to insert an item into.
    @param item The item to insert into the hashtable.
    @param key The key to associate the item with.
    @param key_size The size of the key associated with the item.
    @return OALHashtable_InsertionStatus OALAVLTree_InsertSucess (TRUE) if succeded, OALAVLTree_InsertFailure (FALSE) is failed
*/
extern OALHashtable_InsertionStatus OALHashtable_Insert(OALHashtable hashtable, OALHashItem_t item, OALHashKey_t *keys, OALHashKeySize_t *key_sizes);

/* @internal
*   Remove the parent (root) item from the hashtable.

    @param key The key associated with the item to be removed.
    @param key_size The size of the key associated with the item to be removed.

    @return TRUE if something was deleted; otherwise FALSE.
*/
extern boolean OALHashtable_Remove_First(OALHashtable hashtable, OALHashKey_t *keys, OALHashKeySize_t *key_sizes);

/* @internal
*   Remove an item from the hashtable.

    @param hashtable The hashtable from which the item is removed.
    @param key The key associated with the item to be removed.
    @param key_size The size of the key associated with the item to be removed.

    @return TRUE if something was deleted; otherwise FALSE.
*/
extern boolean OALHashtable_Remove(OALHashtable hashtable, OALHashItem_t item, OALHashKey_t *keys, OALHashKeySize_t *key_sizes);

/* @internal
*   Retrieve an item from the hashtable.

    @param hashtable The hashtable to retrieve an item from.
    @param key The key associated with the item to be retrieved.
    @param key_size The size of the key associated with the item to be retrieved.
    @return The item associated with the given key, or NULL If no item was found.
*/
extern OALHashItem_Ret_t OALHashtable_Retrieve(OALHashtable hashtable, OALHashKey_t *keys, OALHashKeySize_t *key_sizes);

/* @internal
*   Return the number of items in a hashtable.

    @param hashtable The hashtable whose number of items is to be returned.
    @return uint32 The number of items withing the hashtable, or 0 if no items were found.
*/
extern uint32 OALHashtable_GetNumItems(OALHashtable hashtable);

#ifdef __cplusplus
}
#endif

#endif /* DOF_HASHTABLE_H */
