/*
**  Copyright (c) 2014-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef INTERNAL_QUEUEINTERNAL_H
#define INTERNAL_QUEUEINTERNAL_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif

typedef struct OALQueue_t* OALQueue;

/** Create a new instance of an OALQueue.  The queue returned will grow and shrink as needed.
 *
 *  @return     an OALQueue ready for use
 *
 *  @retval NULL    unable to create an OALQueue
 *
 *  @mustfree{OALQueue_Destroy}
 *
 *  @internal
 */
extern OALQueue OALQueue_Create(void);

/** Deallocate an OALQueue that is no longer needed.  It is YOUR responsibility to deallocate any
 *  dynamic data being stored in the OALQueue prior to calling this method.
 *
 *  @param[in]      self
 *      the OALQueue that should be deallocated
 *
 *  @internal
 */
extern void OALQueue_Destroy(OALQueue self);

/** Insert an element into an OALQueue.
 *
 *  @param[in]      self
 *      An OALQueue that will have something added.  Must not be NULL.
 *
 *  @param[in]      item
 *      Pointer to something to add to the OALQueue.  Only the address of item is stored--no copy is
 *      made.
 *  @return     TRUE if element was successfully enqueued; FALSE otherwise.
 *
 *  @internal
 */
extern boolean OALQueue_Enqueue(OALQueue self, void *item);

/** Remove the first element from an OALQueue.
 *
 *  @param[in]      self
 *      The OALQueue to retrieve from.  Must not be NULL.
 *
 *  @return     the address of some item from the OALQueue self
 *
 *  @warning    @a self must not be empty
 *
 *  @internal
 */
extern void * OALQueue_Dequeue(OALQueue self);

/** Retrieve, but not remove, the first element from an OALQueue.
 *
 *  @param[in]      self
 *      An existing OALQueue to retrieve from.  Must not be NULL.
 *
 *  @return     the front item from the OALQueue self
 *
 *  @warning    @a self must not be empty
 *
 *  @internal
 */
extern void * OALQueue_GetFront(OALQueue self);

/** Retrieve the size (number of elements) in an OALQueue.
 *
 *  @param[in]      self
 *      An existing OALQueue to retrieve the size of.  Must not be NULL.
 *
 *  @return     the number of elements in the OALQueue self
 *
 *  @internal
 */
extern uint32 OALQueue_GetSize(OALQueue self);

#ifdef  __cplusplus
}
#endif

#endif
