/*
**  Copyright (c) 2014-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include "config.h"
#include <dof/pcr.h>
#include "socket.h"
#include <dof/oal.h>
#include <dof/inet.h>

#include "inet_connection.h"
#include "inet_manager.h"
#include "inet_server.h"

static void _DOFInetConnection_Send(DTLConnection connection,
                                    DOFAddress    target,
                                    uint8 *       data,
                                    uint32        offset,
                                    uint32        length);

static void _DOFInetConnection_Close(DTLConnection connection);

static DOFAddress _DOFInetConnection_GetAddress(DTLConnection connection);

static DTLServer _DOFInetConnection_GetServer(DTLConnection connection);

static DTLTransportHandler _DOFInetConnection_GetHandler(DTLConnection connection);

static void * _DOFInetConnection_GetContext(DTLConnection connection);

static void _DOFInetConnection_SetContext(DTLConnection   connection,
                                          void          * context);

static DOFConnectionType _DOFInetConnection_GetConnectionType(DTLConnection connection);

static boolean _ReturnTrue(DTLConnection connection);

static boolean _ReturnFalse(DTLConnection connection);

static boolean _DOFInetConnection_IsPointToPoint(DTLConnection connection);

static DTLConnectionConfig _DOFInetGetConnectionConfig(DTLConnection connection);

InetConnection InetConnection_Create(DOFAddress            addr,
                                     InetSocket            socket,
                                     DTLTransportHandler   transportHandler,
                                     DTLServer             server,
                                     DOFConnectionType     type,
                                     boolean               isBound,
                                     DOFAddressType        aType,
                                     DTLConnectionConfig   config,
                                     void                * context)
{
    InetConnection ret;

    static const struct DTLConnectionFns_t connectionCallbacks =
    {
        _DOFInetConnection_Send,
        _DOFInetConnection_Close,
        _DOFInetConnection_GetAddress,
        _DOFInetConnection_GetServer,
        _DOFInetConnection_GetHandler,
        _DOFInetConnection_GetConnectionType,
        _DOFInetConnection_GetContext,
        _DOFInetConnection_SetContext,
        _ReturnTrue,
        _ReturnFalse,
        _DOFInetConnection_IsPointToPoint,
        _DOFInetGetConnectionConfig
    };

    PCR_Assert(addr);
    PCR_Assert(socket != PCRSOCKET_INVALID);
    PCR_Assert(transportHandler != NULL);

    ret = (InetConnection) PCRMem_Alloc(sizeof(InetConnection_t));
    if(ret != NULL)
    {
        ret->super.fns = &connectionCallbacks;
        ret->addr = DOFAddress_Copy(addr);
        ret->socket = socket;
        ret->transportHandler = transportHandler;
        ret->server = server;
        ret->isClosed = FALSE;
        ret->type = type;
        ret->aType = aType;
        ret->isBound = isBound;
        ret->config = config;
        if(server != NULL)
        {
            DOFInetServer_Retain((DOFInetServer) server);
        }
        ret->context = context;
    }
    return ret;
}

void InetConnection_Destroy(InetConnection connection)
{
    PCR_Assert(connection != NULL);

    if(connection->server != NULL)
    {
        DOFInetServer_Destroy((DOFInetServer) connection->server);
    }
    _DOFInetConnection_Close((DTLConnection) connection);
    DOFAddress_Destroy(connection->addr);
    InetSocket_Destroy(connection->socket);
    PCRMem_Free(connection);
}

DOFInetConnectionConfig DOFInetConnectionConfig_Create(DOFAddress addr,
                                                       InetSocket socket)
{
    DOFInetConnectionConfig ret;

    PCR_Assert(addr != NULL);
    PCR_Assert(socket != PCRSOCKET_INVALID);

    ret = (DOFInetConnectionConfig) PCRMem_Alloc(sizeof(DOFInetConnectionConfig_t));
    if(ret != NULL)
    {
        ret->addr = DOFAddress_Copy(addr);
        ret->socket = socket;
    }
    return ret;
}

void DOFInetConnectionConfig_Destroy(DOFInetConnectionConfig config)
{
    if(config != NULL)
    {
        DOFAddress_Destroy(config->addr);
        InetSocket_Destroy(config->socket);
        PCRMem_Free(config);
    }
}

static void _DOFInetConnection_Send(DTLConnection connection,
                                    DOFAddress    target,
                                    uint8 *       data,
                                    uint32        offset,
                                    uint32        length)
{
    InetConnection realConnection;

    PCR_Assert(connection != NULL);

    realConnection = (InetConnection) connection;
    if(!realConnection->isClosed)
    {
        if(realConnection->isBound)
        {
            InetManager_Send(realConnection->socket, data, offset, length);
        }
        else
        {
            if(target == NULL)
            {
                target = realConnection->addr;
            }
            InetManager_SendTo(realConnection->socket, data,
                               offset, length, target);
        }
    }
    else
        PCRMem_Free(data);
}

static void _DOFInetConnection_Close(DTLConnection connection)
{
    InetConnection conn;

    PCR_Assert(connection != NULL);

    conn = (InetConnection) connection;
    if(!conn->isClosed)
    {
        conn->isClosed = TRUE;
        PCR_Assert(conn->socket != PCRSOCKET_INVALID);
        InetManager_FindAndRemoveSocket(conn->socket, INETM_NORMAL);
    }
}

static DOFAddress _DOFInetConnection_GetAddress(DTLConnection connection)
{
    InetConnection realConnection;

    PCR_Assert(connection != NULL);
    realConnection = (InetConnection) connection;
    return realConnection->addr;
}

static DTLServer _DOFInetConnection_GetServer(DTLConnection connection)
{
    InetConnection realConnection;

    PCR_Assert(connection != NULL);
    realConnection = (InetConnection) connection;
    return realConnection->server;
}

static DTLTransportHandler _DOFInetConnection_GetHandler(DTLConnection connection)
{
    InetConnection realConnection;

    PCR_Assert(connection != NULL);
    realConnection = (InetConnection) connection;
    return realConnection->transportHandler;
}

static void * _DOFInetConnection_GetContext(DTLConnection connection)
{
    InetConnection realConnection;

    PCR_Assert(connection != NULL);
    realConnection = (InetConnection) connection;
    return realConnection->context;
}

static void _DOFInetConnection_SetContext(DTLConnection   connection,
                                          void          * context)
{
    InetConnection realConnection;

    PCR_Assert(connection != NULL);
    realConnection = (InetConnection) connection;
    realConnection->context = context;
}

static DOFConnectionType _DOFInetConnection_GetConnectionType(DTLConnection connection)
{
    PCR_Assert(connection != NULL);

    return ((InetConnection)connection)->type;
}

static boolean _ReturnTrue(DTLConnection connection)
{
    return TRUE;
}

static boolean _ReturnFalse(DTLConnection connection)
{
    return FALSE;
}

static boolean _DOFInetConnection_IsPointToPoint(DTLConnection connection)
{
    InetConnection realConnection;

    PCR_Assert(connection != NULL);
    realConnection = (InetConnection) connection;
    return ( realConnection->aType == DOFADDRESSTYPE_UNICAST );
}

static DTLConnectionConfig _DOFInetGetConnectionConfig(DTLConnection connection)
{
    InetConnection realConnection;

    PCR_Assert(connection != NULL);
    realConnection = (InetConnection) connection;
    return realConnection->config;
}
