/*
**  Copyright (c) 2014-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include "config.h"

#include <dof/pcr.h>
#include <dof/oal.h>
#include <dof/inet.h>

#include "inet_manager.h"
#include "socket.h"

#include <netdb.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>

/* Check if necessary headers are available to allow interface name resolution and if bindtodevice is available*/
#if defined(SO_BINDTODEVICE)
#include <net/if.h>
#define INET_CAN_RESTRICT_TO_INTERFACE 1
#else
#undef INET_CAN_RESTRICT_TO_INTERFACE
#endif

#ifndef RESOLVE_WAIT_PERIOD
#define RESOLVE_WAIT_PERIOD_MS 180
#endif

#ifndef PF_INET
#define PF_INET AF_INET
#endif
#define HAVE_DECL_PF_INET 1

#ifndef PF_INET6
#define PF_INET6 AF_INET6
#endif
#define HAVE_DECL_PF_INET6 1

extern const DTLTransport InetTransport;

#define DOFADDRESSTYPE_UNKNOWN 10

typedef struct InetTransportAddress_t
{
    char* hostname;

#ifdef INET_CAN_RESTRICT_TO_INTERFACE
    char* iface;
    uint32 iface_id;
#endif

    uint32 addrLen;
    union {
        struct sockaddr* base;
#if HAVE_DECL_PF_INET
        struct sockaddr_in* in;
#endif
#if HAVE_DECL_PF_INET6
        struct sockaddr_in6* in6;
#endif
    } addr;
    uint32 resolveTimestamp;
    DOFAddressType addrType;
    uint16 port;
} InetTransportAddress_t;

static uint32 InetTransport_GetSeconds(void)
{
    struct timeval tv;
    uint32 msecs;

    /* Get the system time, and convert to milliseconds */
    gettimeofday(&tv, NULL);
    msecs = ((uint32)(tv.tv_sec));
    return(msecs);
}

static uint32 InetTransport_GetAddress_Length(InetTransportAddress_t* iaddr)
{
    uint32 len = sizeof(InetTransportAddress_t);
    uint32 hlen = 0;
    uint32 ilen = 0;
    
    if( iaddr->hostname )
        hlen = PCRString_Length(iaddr->hostname)+1;

#ifdef INET_CAN_RESTRICT_TO_INTERFACE
    if( iaddr->iface )
        ilen = PCRString_Length(iaddr->iface)+1;
#endif

    return len + hlen + ilen;
}

static InetTransportAddress_t*  InetTransport_CreateAddress_Internal(const char* hostname, const char* iface, uint16 port)
{
    InetTransportAddress_t* ret = NULL;
    uint32 len = sizeof(InetTransportAddress_t);
    uint32 hlen = 0;
    uint32 ilen = 0;

#ifndef INET_CAN_RESTRICT_TO_INTERFACE
    if( iface )
        return NULL;
#endif

    if( hostname )
        hlen = PCRString_Length(hostname)+1;

    /* not wrapped because covered by check above */
    if( iface )
        ilen = PCRString_Length(iface)+1;

    ret = (InetTransportAddress_t*)PCRMem_Alloc( len + hlen + ilen );
    if ( ret )
    {
        PCRMem_Set( ret, 0, len + hlen + ilen );
        if( hostname )
        {
            ret->hostname = (char*)(((uint8*)ret)+len);
            PCRMem_Copy(ret->hostname, hostname, hlen);
        }

#ifdef INET_CAN_RESTRICT_TO_INTERFACE
        if ( iface )
        {
            ret->iface = (char*)(((uint8*)ret)+len+hlen);
            PCRMem_Copy(ret->iface, iface, ilen);
        }
#endif

        ret->port = port;
        ret->addrType = DOFADDRESSTYPE_UNKNOWN;
    }

    return ret;
}

static void InetTransport_DestroyAddress_Internal(InetTransportAddress_t* self)
{
    if( self )
    {
        if( self->addr.base )
            PCRMem_Free(self->addr.base);

        PCRMem_Free(self);
    }
}

static boolean InetTransport_SetSockaddr(InetTransportAddress_t* self, const struct sockaddr* addr, uint32 addrSize, boolean overwrite_port )
{
    boolean ret = TRUE;
    if( self )
    {
        /* alloc memory  */
        if( self->addr.base )
            PCRMem_Free(self->addr.base);
            
        self->addr.base = PCRMem_Alloc( addrSize);
        if( !self->addr.base )
        {
            /* fail */
            self->addrLen = 0;
            return FALSE;
        }

        self->addrLen = addrSize;

        /* Copy */
        PCRMem_Copy(self->addr.base, addr, addrSize);
        if( overwrite_port )
        {
            switch( addr->sa_family ) 
            {
            case AF_INET:
                self->addr.in->sin_port = PCRSwap_Net16(self->port);
                break;
            case AF_INET6:
                self->addr.in6->sin6_port = PCRSwap_Net16(self->port);
                break;
            default:
                PCRMem_Free(self->addr.base);
                self->addr.base = NULL;
                self->addrLen = 0;
                ret = FALSE;
            }   
        }
    }
    return ret;
}

boolean InetTransport_ResolveAddress( const DOFAddress addr )
{
    boolean ret = TRUE;
    if((addr != NULL) && (DOFAddress_GetTransport( addr ) == InetTransport))
    {
        InetTransportAddress_t* iaddr = (InetTransportAddress_t*)DOFAddress_GetAddress( addr );
        if( iaddr->hostname )
        {
            struct addrinfo * addrinfo;

            if (iaddr->addr.base)
                PCRMem_Free(iaddr->addr.base);

            PCRMem_Set( &iaddr->addr, 0, sizeof(iaddr->addr) );
            if ( getaddrinfo( iaddr->hostname, NULL, NULL, &addrinfo ) != 0 )
            {
                return FALSE;
            }

            if( !InetTransport_SetSockaddr(iaddr, addrinfo->ai_addr, addrinfo->ai_addrlen, TRUE) )
            {
                freeaddrinfo(addrinfo); 
                return FALSE;
            }
            freeaddrinfo(addrinfo); 
        }

#ifdef INET_CAN_RESTRICT_TO_INTERFACE
        if( iaddr->iface )
        {
            iaddr->iface_id = if_nametoindex( iaddr->iface );
            if ( !iaddr->iface_id )
            {
                ret = FALSE;
            }
        }

        if( (iaddr->hostname || iaddr->iface) && ret )
#else
        if( (iaddr->hostname) && ret )
#endif
        {
            uint32 curr = InetTransport_GetSeconds();
            iaddr->resolveTimestamp = curr + RESOLVE_WAIT_PERIOD_MS;
        }
    }
    else
    {
        ret = FALSE;
    }
    return ret; 
}

boolean InetTransport_ShouldReresolve(const DOFAddress addr)
{
    if((addr != NULL) && (DOFAddress_GetTransport( addr ) == InetTransport))
    {
        InetTransportAddress_t* iaddr = (InetTransportAddress_t*)DOFAddress_GetAddress( addr );
        if( iaddr->hostname
#ifdef INET_CAN_RESTRICT_TO_INTERFACE             
            || iaddr->iface )
#else
            )
#endif
        {
            uint32 curr = InetTransport_GetSeconds();
            /* store the initial time in a uint32 so the comparison is after rollover. */
            uint32 initialStart = iaddr->resolveTimestamp - RESOLVE_WAIT_PERIOD_MS;
            if( curr >= iaddr->resolveTimestamp || initialStart > curr )
            {
                return TRUE;
            }
        }
    }
    return FALSE;
}

void InetTransport_ClearResolve(const DOFAddress addr)
{
    if((addr != NULL) && (DOFAddress_GetTransport( addr ) == InetTransport))
    {
        InetTransportAddress_t* iaddr = (InetTransportAddress_t*)DOFAddress_GetAddress( addr );
        if( iaddr->hostname )
        {
            PCRMem_Free(iaddr->addr.base);
            iaddr->addr.base = NULL;
            iaddr->addrLen = 0;
            iaddr->resolveTimestamp = InetTransport_GetSeconds();
        }
#ifdef INET_CAN_RESTRICT_TO_INTERFACE
        iaddr->iface_id = 0;
#endif        
    }
}

DOFAddressType InetTransport_GetAddressType( DOFAddress addr )
{
    if (  addr && (DOFAddress_GetTransport( addr ) == InetTransport) )
    {
        InetTransportAddress_t* iaddr = (InetTransportAddress_t*)DOFAddress_GetAddress( addr );
        if( iaddr->addrType == DOFADDRESSTYPE_UNKNOWN )
        {
            /* Is the address a hostname and has not been resolved */
            if( iaddr->hostname && !iaddr->addr.base )
            {
                if( !InetTransport_ResolveAddress(addr) )
                {
                    /* Resolve failed return unicast */
                    return DOFADDRESSTYPE_UNICAST;
                }
                /* Resolved succeeded addr->addr.base will not be NULL */
            }

            switch( iaddr->addr.base->sa_family )
            {
#if HAVE_DECL_PF_INET6
                case PF_INET6:
                    /* Check for multicast address. */
                    if( ((uint8 *)&(iaddr->addr.in6->sin6_addr.s6_addr))[0] == 0xFF )
                    {
                        /* IPv6 does not support the idea of full "broadcast" -- it is merely another form of multicast. */
                        iaddr->addrType = DOFADDRESSTYPE_MULTICAST;
                    }
                    else
                    {
                        iaddr->addrType = DOFADDRESSTYPE_UNICAST;
                    }
                    break;
#endif

#if HAVE_DECL_PF_INET
                case PF_INET:
                    /* Check for multicast address. */
                    if( ((uint8*)&(iaddr->addr.in->sin_addr.s_addr))[0] >= 0xE0 )
                    {
                        if( ((uint32*)&(iaddr->addr.in->sin_addr.s_addr))[0] == 0xFFFFFFFF )
                            iaddr->addrType = DOFADDRESSTYPE_BROADCAST;
                        iaddr->addrType = DOFADDRESSTYPE_MULTICAST;
                    }
                    else
                    {
                        iaddr->addrType = DOFADDRESSTYPE_UNICAST;
                    }
                    break;
#endif
                default:
                    break;
            }

            if( iaddr->addrType == DOFADDRESSTYPE_UNKNOWN )
                return DOFADDRESSTYPE_UNICAST; 
        }
        return iaddr->addrType;
    }

    return DOFADDRESSTYPE_UNICAST;
}

int InetTransport_CompareAddress( uint32 size, const void * addr1, const void * addr2 )
{
    int ret = 0;
    InetTransportAddress_t* iaddr1 = (InetTransportAddress_t*)addr1;
    InetTransportAddress_t* iaddr2 = (InetTransportAddress_t*)addr2;
    if( iaddr1->hostname )
    {
        /* check hostname */
        if( iaddr2->hostname )
        {
            /* both are hostname */
            uint32 hlen = PCRString_Length(iaddr1->hostname);
            ret = hlen - PCRString_Length(iaddr2->hostname);
            if( !ret )
                ret = PCRMem_Compare(iaddr1->hostname, iaddr2->hostname, hlen);

            /* check iface if one*/
#ifdef INET_CAN_RESTRICT_TO_INTERFACE
            if( !ret )
            {
                /* iface compare */
                if( iaddr1->iface )
                {
                    if( iaddr2->iface )
                    {
                        uint32 ilen = PCRString_Length(iaddr1->iface); 
                        ret = ilen - PCRString_Length(iaddr2->iface);
                        if( !ret )
                            ret = PCRMem_Compare(iaddr1->iface, iaddr2->iface, ilen);
                    }
                    else
                    {
                        return 1;
                    }
                }
                else if( iaddr2->iface )
                {
                    return -1;
                }
            }
#endif
        }
        else
        {
            /* hostname against non-hostname */
            return 1;
        }
    }
    else if( iaddr2->hostname )
    {
        return -1;
    }
    else
    {
        const struct sockaddr *saddr1 = (const struct sockaddr *)iaddr1->addr.base;
        const struct sockaddr *saddr2 = (const struct sockaddr *)iaddr2->addr.base;

        ret = saddr1->sa_family - saddr2->sa_family;
        if( !ret )
        {
            switch( saddr1->sa_family ) 
            {
#if HAVE_DECL_PF_INET6
                case PF_INET6:
                    {
                        struct sockaddr_in6 * si61 = (struct sockaddr_in6 *)saddr1;
                        struct sockaddr_in6 * si62 = (struct sockaddr_in6 *)saddr2;
                        ret = PCRMem_Compare( &si61->sin6_addr, &si62->sin6_addr, sizeof(si61->sin6_addr) );
                        if ( !ret )
                            ret = (int)si61->sin6_port - (int)si62->sin6_port;
                    }
                    break;
#endif

#if HAVE_DECL_PF_INET
                case PF_INET:
                    {
                        struct sockaddr_in * si1 = (struct sockaddr_in *)saddr1;
                        struct sockaddr_in * si2 = (struct sockaddr_in *)saddr2;
                        ret = PCRMem_Compare( &si1->sin_addr, &si2->sin_addr, sizeof(si1->sin_addr) );
                        if ( !ret )
                            ret = (int)si1->sin_port - (int)si2->sin_port;
                    }
                    break;
#endif

                default:
                    break;
            }
        }

#ifdef INET_CAN_RESTRICT_TO_INTERFACE
        if( !ret )
        {
            /* iface compare */
            if( iaddr1->iface )
            {
                if( iaddr2->iface )
                {
                    uint32 ilen = PCRString_Length(iaddr1->iface); 
                    ret = ilen - PCRString_Length(iaddr2->iface);
                    if( !ret )
                        ret = PCRMem_Compare(iaddr1->iface, iaddr2->iface, ilen);
                }
                else
                {
                    return 1;
                }
            }
            else if( iaddr2->iface )
            {
                return -1;
            }
        }
#endif
    }
    
    return ret;
}

boolean InetTransport_AddressCopy( uint32 size, void * addr )
{
    InetTransportAddress_t* self = (InetTransportAddress_t*)addr;
    uint32 len = sizeof(InetTransportAddress_t);
    uint32 hlen = 0;
    
    if( self->hostname )
    {
        hlen = PCRString_Length(self->hostname)+1;
        self->hostname = (char*)(((uint8*)self)+len);
        self->addr.base = NULL;
    }

#ifdef INET_CAN_RESTRICT_TO_INTERFACE
    if( self->iface )
    {
        self->iface = (char*)(((uint8*)self)+len+hlen);
        self->iface_id = 0;
    }
#endif

    if( !self->hostname )
    {
        /* If no hostname this address was created with sockadder so we need to copy that */
        if( self->addrLen > 0 ) 
        {
            void *saddr = PCRMem_Alloc(self->addrLen);
            if( !saddr )
                return FALSE;

            PCRMem_Copy(saddr, self->addr.base, self->addrLen); 
            self->addr.base = (struct sockaddr*)saddr;
        }
    }
    else
    {
        /* Set time to resolve on next call resolution */
        self->resolveTimestamp = InetTransport_GetSeconds();
    }

    return TRUE;
}

void InetTransport_AddressDestroy( uint32 size, const void * addr )
{
    InetTransportAddress_t* self = (InetTransportAddress_t*)addr;
    if( self->addr.base )
    {
        PCRMem_Free(self->addr.base);
        self->addr.base = NULL;
        self->addrLen = 0;
    }
}

static DOFAddress InetTransport_CreateDOFAddress_Internal(InetTransportAddress_t* self)
{
    DOFAddress ret = NULL;
    if( self && InetTransport_IsRunning())
    {
        ret = DOFAddress_Create( InetTransport_GetAddress_Length(self), self, InetTransport );
    }
    return ret;
}

DOFAddress InetTransport_CreateAddress_Interface( const char* hostname, uint16 port, const char* iface )
{   
    DOFAddress rval = NULL;
    if( hostname )
    {
        InetTransportAddress_t *iaddr = InetTransport_CreateAddress_Internal( hostname, iface, port );
        if( iaddr )
        {
            rval = InetTransport_CreateDOFAddress_Internal( iaddr );
            InetTransport_DestroyAddress_Internal(iaddr);
        }
    }
    return rval;
}

DOFAddress InetTransport_CreateAddress( const char* hostname, uint16 port )
{
    return InetTransport_CreateAddress_Interface( hostname, port, NULL );
}

static uint16 InetTransport_Sockaddr_GetPort(const struct sockaddr * addr, uint32 addrSize)
{
    uint16 ret = 0;

    switch ( addr->sa_family )
    {
#if HAVE_DECL_PF_INET
        case AF_INET:
            ret = PCRSwap_Net16(((struct sockaddr_in*)addr)->sin_port);
            break;
#endif
#if HAVE_DECL_PF_INET6
        case AF_INET6:
            ret = PCRSwap_Net16(((struct sockaddr_in6*)addr)->sin6_port);
            break;
#endif
        default:
            break;
    }   
    return ret;
}

DOFAddress InetTransport_CreateAddress_Sockaddr_Interface( const struct sockaddr * addr, uint32 addrSize, const char* iface )
{
    DOFAddress rval = NULL; 
    if( addr && addrSize > 0 )
    {
        InetTransportAddress_t *iaddr = InetTransport_CreateAddress_Internal( NULL, iface, 0 ); 
        if( iaddr )
        {
            iaddr->port = InetTransport_Sockaddr_GetPort(addr, addrSize);
            if( !InetTransport_SetSockaddr( iaddr, addr, addrSize, FALSE ))
            {
                InetTransport_DestroyAddress_Internal(iaddr);
                return NULL;
            }

            rval = InetTransport_CreateDOFAddress_Internal( iaddr );

            /* if not resolution this will set the type to get around coal bug */
            if( !iface )
                InetTransport_GetAddressType(rval); 

            InetTransport_DestroyAddress_Internal(iaddr);
        }
    }
    return rval;
}

DOFAddress InetTransport_CreateAddress_Sockaddr( const struct sockaddr  * addr, uint32 addrSize )
{
    return InetTransport_CreateAddress_Sockaddr_Interface( addr, addrSize, NULL );
}

DOFAddress InetTransport_CreateIPv4Address(const uint8  addr[4],
                                                 uint16 port)
{
#if HAVE_DECL_PF_INET
    struct sockaddr_in in_addr;

    DOFAddress ret;

    PCRMem_Set(&in_addr, 0, sizeof(struct sockaddr_in));
    in_addr.sin_family = AF_INET;
    in_addr.sin_port = PCRSwap_Net16(port);
    PCRMem_Copy(&in_addr.sin_addr.s_addr, addr, 4);
    ret = InetTransport_CreateAddress_Sockaddr((struct sockaddr*)&in_addr, sizeof(struct sockaddr_in));
    return ret;
#else
    return NULL;
#endif
}

DOFAddress InetTransport_CreateIPv6Address(const uint8  addr[16],
                                           uint16 port)
{
#if HAVE_DECL_PF_INET6
    struct sockaddr_in6 in_addr;

    DOFAddress ret;

    PCRMem_Set(&in_addr, 0, sizeof(struct sockaddr_in6));
    in_addr.sin6_family = AF_INET6;
    in_addr.sin6_port = PCRSwap_Net16(port);
    PCRMem_Copy(&in_addr.sin6_addr.s6_addr, addr, 16);
    ret = InetTransport_CreateAddress_Sockaddr((struct sockaddr*)&in_addr, sizeof(struct sockaddr_in6));
    return ret;
#else
    return NULL;
#endif
}


const struct sockaddr * InetTransport_GetAddress( const DOFAddress addr )
{
    if((addr != NULL) && (DOFAddress_GetTransport( addr ) == InetTransport))
    {
        InetTransportAddress_t* iaddr = (InetTransportAddress_t*)DOFAddress_GetAddress(addr);
        if( !iaddr->addr.base )
        {
            /* Resolve */
            if( !InetTransport_ResolveAddress(addr) )
                return NULL; /* FAILURE */
        }
        return iaddr->addr.base; 
    }
    return NULL;
}

uint32 InetTransport_GetAddressSize( const DOFAddress addr )
{
    if((addr != NULL) && (DOFAddress_GetTransport( addr ) == InetTransport))
    {
        InetTransportAddress_t* iaddr = (InetTransportAddress_t*)DOFAddress_GetAddress(addr);
        if( !iaddr->addr.base )
        {
            /* Resolve */
            if( !InetTransport_ResolveAddress(addr) )
                return 0; /* FAILURE */
        }
        return iaddr->addrLen; 
    }
    return 0;
}

uint32 InetTransport_GetAddressInterface( const DOFAddress addr )
{
#ifdef INET_CAN_RESTRICT_TO_INTERFACE
    if((addr != NULL) && (DOFAddress_GetTransport( addr ) == InetTransport))
    {
        InetTransportAddress_t* iaddr = (InetTransportAddress_t*)DOFAddress_GetAddress(addr);
        if( iaddr->iface && iaddr->iface_id == 0 )
        {
            /* Resolve */
            if( !InetTransport_ResolveAddress(addr) )
                return 0; /* FAILURE */
        }

        return iaddr->iface_id; 
    }
#endif
    return 0;
}

