/*
**  Copyright (c) 2014-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include "config.h"
#include <dof/pcr.h>
#include "socket.h"
#include <dof/oal.h>
#include <dof/inet.h>
#include <errno.h>

#include <internal/hashtable.h>
#include <internal/timer.h>
#include <internal/threadpool.h>

#include <net/if.h>
#include <sys/socket.h>
#include <signal.h>

#include "inet_manager.h"
#include "inet_receive.h"
#include "inet_send.h"

#ifndef     INET_SELECT_TIMEOUT
#define     INET_SELECT_TIMEOUT     50
#endif

#define     LOOPBACK_DATA_SIZE              16
#define     LOOPBACK_DATA_TIMEOUT           1000

#ifndef PF_INET
#define PF_INET AF_INET
#endif
#define HAVE_DECL_PF_INET 1

#ifndef PF_INET6
#define PF_INET6 AF_INET6
#endif
#define HAVE_DECL_PF_INET6 1

const uint8 INETTRANSPORT_DEFAULT_MULTICAST_ADDRESS_IPV4[4] = { 224, 0, 23, 46 };
const uint8 INETTRANSPORT_DEFAULT_MULTICAST_ADDRESS_IPV6_NODE[16]   = { 0xFF, 0x01, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0x01, 0x2C };
const uint8 INETTRANSPORT_DEFAULT_MULTICAST_ADDRESS_IPV6_LINK[16]   = { 0xFF, 0x02, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0x01, 0x2C };
const uint8 INETTRANSPORT_DEFAULT_MULTICAST_ADDRESS_IPV6_SUBNET[16] = { 0xFF, 0x03, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0x01, 0x2C };
const uint8 INETTRANSPORT_DEFAULT_MULTICAST_ADDRESS_IPV6_ADMIN[16]  = { 0xFF, 0x04, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0x01, 0x2C };
const uint8 INETTRANSPORT_DEFAULT_MULTICAST_ADDRESS_IPV6_SITE[16]   = { 0xFF, 0x05, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0x01, 0x2C };
const uint8 INETTRANSPORT_DEFAULT_MULTICAST_ADDRESS_IPV6_ORG[16]    = { 0xFF, 0x08, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0x01, 0x2C };
const uint8 INETTRANSPORT_DEFAULT_MULTICAST_ADDRESS_IPV6_GLOBAL[16] = { 0xFF, 0x0e, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0x01, 0x2C };

PCRMutex     inetMutex;

DOFException inetMessages[6];

typedef struct LoopbackNode_t
{
    OALTimerItem ti;

    DOFAddress   addr;

    uint8        data[LOOPBACK_DATA_SIZE + 1];
} LoopbackNode_t;

typedef LoopbackNode_t   * LoopbackNode;

static PCRThread      inetThread;

static PCREvent       socketsReady;

static boolean        keepRunning = FALSE;

static uint32         cleanupNeeded;

static InetSocketNode socketListHead;
static InetSocketNode socketListTail;

static fd_set         recvSet;
static fd_set         sendSet;

#ifndef INET_DISABLE_LOOPBACK
static OALHashtable   ignoredAddresses;
static OALHashtable   localLoopbackData;
static OALTimer       loopbackTimer;
static OALThreadPool  loopbackThread;

static struct sigaction sigact;

static void _propogateLoopbackData(InetSocket socket,
                                   DOFAddress multicastAddr);

static boolean _prepareSockaddrKey(const struct sockaddr *  sockaddr,
                                          void     ** keys,
                                          uint16   *  keySizes);

static int _loopbackTimeout(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown);

#endif

static void _inetThread(void * threadData);

static void _addNode(InetSocketNode newNode);

static void _cleanupNode(InetSocketNode node);

static void _emptyList(void);

static void Inet_ConnectionComplete(InetSocketNode node);

static InetSocketNode _findNode(InetSocket sockFd);

static void _addPartialSendInfo(InetSocketNode socketNode,
                                uint8      * data,
                                uint32       offset,
                                uint32       length,
                                DOFAddress   target);

boolean InetTransport_IsRunning(void)
{
    return keepRunning;
}

static void sigpipe_handler(int signum)
{ /* DO nothing, Drop SIGPIPEs. */ }

void InetTransport_Init(void)
{
    InetSocket_Init();
    keepRunning = TRUE;
    cleanupNeeded = 0;
    socketListHead = socketListTail = NULL;
    
    /* Catch sigpipes to ignore */
    sigact.sa_handler = &sigpipe_handler;
    PCR_Assert( sigaction(SIGPIPE, &sigact, NULL) == 0) ;
    
    inetMessages[INETM_NORMAL] = NULL;
    inetMessages[INETM_ERROR] = DOFErrorException_Create( DOFERROR_TRANSPORT_FAILED, "Internal transport failure" );
    inetMessages[INETM_CONNECTION_CLOSED] = DOFErrorException_Create( DOFERROR_TERMINATED, "Connection Closed by remote host" );
    inetMessages[INETM_RECEIVE_FAILED] = DOFErrorException_Create( DOFERROR_IO_FAILED, "Receive failure" );
    inetMessages[INETM_SEND_FAILED] = DOFErrorException_Create( DOFERROR_IO_FAILED, "Send failure" );
    inetMessages[INETM_BIND_FAILED] = DOFErrorException_Create( DOFERROR_TRANSPORT_FAILED, "bind() failed" );

#ifndef INET_DISABLE_LOOPBACK
    ignoredAddresses = OALHashtable_Create(2);
    PCR_Assert(ignoredAddresses != NULL);
    localLoopbackData = OALHashtable_Create(3);
    PCR_Assert(localLoopbackData != NULL);
    loopbackThread = OALThreadPool_Create(1);
    PCR_Assert(loopbackThread != NULL);
    loopbackTimer = OALTimer_Create(loopbackThread, FALSE); /* threadpool only supports one item at a time, so just make the timer async. it'll really be sync */
    PCR_Assert(loopbackTimer != NULL);
#endif

    FD_ZERO(&recvSet);
    FD_ZERO(&sendSet);

    inetMutex = PCRMutex_Create();
    PCR_Assert(inetMutex != NULL);
    socketsReady = PCREvent_Create(FALSE);
    PCR_Assert(socketsReady != NULL);
    inetThread = PCRThread_Create(_inetThread, NULL);
    PCR_Assert(inetThread != NULL);
}

void InetTransport_Shutdown(void)
{
    PCRMutex_Lock(inetMutex);
    keepRunning = FALSE;
    PCREvent_Signal(socketsReady);
    PCRMutex_Unlock(inetMutex);

    PCRThread_Join(inetThread);
    PCRThread_Destroy(inetThread);
    PCREvent_Destroy(socketsReady);

    _emptyList();

#ifndef INET_DISABLE_LOOPBACK
    OALTimer_Destroy(loopbackTimer);
    OALThreadPool_Destroy(loopbackThread);
    OALHashtable_Destroy(localLoopbackData, NULL, NULL);
    OALHashtable_Destroy(ignoredAddresses,  NULL, NULL);
#endif

    PCRMutex_Destroy(inetMutex);

    DOFException_Destroy( inetMessages[INETM_NORMAL] );
    DOFException_Destroy( inetMessages[INETM_ERROR] );
    DOFException_Destroy( inetMessages[INETM_CONNECTION_CLOSED] );
    DOFException_Destroy( inetMessages[INETM_RECEIVE_FAILED] );
    DOFException_Destroy( inetMessages[INETM_SEND_FAILED] );
    DOFException_Destroy( inetMessages[INETM_BIND_FAILED] );

    InetSocket_Shutdown();
}

static void _inetThread(void * threadData)
{
    PCRMutex_Lock(inetMutex);
    while(keepRunning)
    {
        PCRMutex_Unlock(inetMutex);         /* Make sure we don't cause deadlock */

        PCREvent_Wait(socketsReady, PCRTIMEOUT_FOREVER);

        PCRMutex_Lock(inetMutex);
        while(keepRunning && (socketListHead != NULL))
        {
            {
                int maxSockFd = socketListTail->socketData.sockFd;

                fd_set recvCopy = recvSet;
                fd_set sendCopy = sendSet;

                PCRMutex_Unlock(inetMutex);
                {
                    int readySockets = InetSocket_Select(maxSockFd + 1, &recvCopy, &sendCopy,
                                                        NULL, INET_SELECT_TIMEOUT);

                    PCRMutex_Lock(inetMutex);
                    if(readySockets > 0)
                    {
                        InetSocketNode itr  = socketListHead;
                        InetSocketNode prev = NULL;

                        while((itr != NULL) && (readySockets > 0))
                        {
                            if(FD_ISSET(itr->socketData.sockFd, &recvCopy) && !itr->socketData.shouldDestroy)
                            {
                                PCR_Assert(itr->recvFunction != NULL);
                                itr->recvFunction(itr);
                                --readySockets;
                            }
                            if(FD_ISSET(itr->socketData.sockFd, &sendCopy) && !itr->socketData.shouldDestroy)
                            {
                                PCR_Assert(itr->sendFunction != NULL);
                                itr->sendFunction(itr);
                                --readySockets;
                            }
                            prev = itr;
                            itr = itr->next;
                        }
                    }
                }
            }
            if(cleanupNeeded > 0)
            {
                InetSocketNode itr  = socketListHead;
                InetSocketNode prev = NULL;
                InetSocketNode dtorHead = NULL;

                while(itr != NULL)
                {
                    if(itr->socketData.shouldDestroy)
                    {
                        InetSocketNode temp = itr;

                        if(socketListHead == itr)
                        {
                            socketListHead = itr->next;
                        }
                        if(socketListTail == itr)
                        {
                            socketListTail = prev;
                        }
                        FD_CLR(itr->socketData.sockFd, &recvSet);
                        FD_CLR(itr->socketData.sockFd, &sendSet);
                        itr = itr->next;
                        temp->next = dtorHead;
                        dtorHead = temp;
                        if(prev != NULL)
                        {
                            prev->next = itr;
                        }
                        --cleanupNeeded;
                    }
                    else
                    {
                        prev = itr;
                        itr = itr->next;
                    }
                }
                PCRMutex_Unlock(inetMutex);
                while(dtorHead != NULL)
                {
                    InetSocketNode backup = dtorHead->next;

                    _cleanupNode(dtorHead);
                    dtorHead = backup;
                }
            }
            else
            {
                PCRMutex_Unlock(inetMutex);
            }
            PCRThread_Yield();
            PCRMutex_Lock(inetMutex);
        }
    }
    PCRMutex_Unlock(inetMutex);
}

static void _addNode(InetSocketNode newNode)
{
    PCR_Assert(newNode != NULL);

    PCRMutex_Lock(inetMutex);
    {
        if(socketListHead == NULL)
        {
            socketListHead = socketListTail = newNode;
            newNode->next = NULL;
        }
        else if(newNode->socketData.sockFd < socketListHead->socketData.sockFd)
        {
            newNode->next = socketListHead;
            socketListHead = newNode;
        }
        else
        {
            InetSocketNode prev = socketListHead;
            InetSocketNode itr = prev->next;

            while(itr != NULL)
            {
                if(newNode->socketData.sockFd < itr->socketData.sockFd)
                {
                    newNode->next = itr;
                    prev->next = newNode;
                    break;
                }
                prev = itr;
                itr = itr->next;
            }
            if(itr == NULL)
            {
                newNode->next = NULL;
                prev->next = newNode;
                socketListTail = newNode;
            }
        }
    }
    PCRMutex_Unlock(inetMutex);
}

static void _cleanupNode(InetSocketNode node)
{
    if(node->socketData.isConnection)
    {
        node->socketData.transportHandler->fns->Closed(node->socketData.transportHandler,
                                                       (DTLConnection) node->socketData.u.connection,
                                                       node->socketData.message);
        InetConnection_Destroy(node->socketData.u.connection);
    }
    else
    {
        node->socketData.transportHandler->fns->Stopped(node->socketData.transportHandler,
                                                        (DTLServer) node->socketData.u.server,
                                                        node->socketData.message);
        DOFInetServer_Destroy(node->socketData.u.server);
        /* We didn't clear the reference in InetTransport_Start */
        DOFInetServer_Destroy(node->socketData.u.server);
    }
    if(node->socketData.sendQueue != NULL)
    {
        while(OALQueue_GetSize(node->socketData.sendQueue) != 0)
        {
            PartialSendData psd = (PartialSendData) OALQueue_Dequeue(node->socketData.sendQueue);

            PCRMem_Free(psd->data);
            PCRMem_Free(psd);
        }
        OALQueue_Destroy(node->socketData.sendQueue);
    }
    PCRMem_Free(node);
}

static void _emptyList(void)
{
    PCRMutex_Lock(inetMutex);
    {
        InetSocketNode temp = socketListHead;

        while(temp != NULL)
        {
            socketListHead = socketListHead->next;
            _cleanupNode(temp);
            temp = socketListHead;
        }
        socketListTail = NULL;
    }
    PCRMutex_Unlock(inetMutex);
}

boolean InetManager_AddServer(DOFInetServer server,
                              boolean       isMulticast)
{
    InetSocketNode newNode = PCRMem_Alloc(sizeof(InetSocketNode_t));

    PCR_Assert(server != NULL);

    if(newNode != NULL)
    {
        const struct sockaddr * sockaddr = InetTransport_GetAddress(server->address);

        newNode->socketData.sockFd = server->socket;
        newNode->socketData.transportHandler = server->handler;
        newNode->socketData.u.server = server;
        DOFInetServer_Retain(server);       /* Increment ref count for above line */
        newNode->socketData.shouldDestroy = FALSE;
        newNode->socketData.message = inetMessages[INETM_NORMAL];
        newNode->socketData.sendQueue = NULL;
        switch(sockaddr->sa_family)
        {
#if HAVE_DECL_PF_INET6
        case PF_INET6:
            newNode->socketData.isIpV6 = TRUE;
            break;
#endif
#if HAVE_DECL_PF_INET
        case PF_INET:
#endif
        default:
            newNode->socketData.isIpV6 = FALSE;
            break;
        }
        newNode->socketData.isMulticast = isMulticast;
        newNode->socketData.isConnection = (server->type == DOFSERVERTYPE_STREAM);
        newNode->recvFunction = (newNode->socketData.isConnection) ?
                                Inet_AcceptConnection   :
                                Inet_ConnectionlessReceive;
        newNode->sendFunction = (newNode->socketData.isConnection) ?
                                NULL                    :
                                Inet_ConnectionlessSend;

        _addNode(newNode);
        PCRMutex_Lock(inetMutex);
        if(newNode->socketData.isConnection)
        {
            InetSocket_Listen(server->socket);
            newNode->socketData.isConnection = FALSE;      /* not really a connection */
        }
        FD_SET(server->socket, &recvSet);
        PCREvent_Signal(socketsReady);
#ifndef INET_DISABLE_LOOPBACK
        if(isMulticast)
        {
            _propogateLoopbackData(server->socket, server->address);
            if(server->server != NULL)
            {
                _propogateLoopbackData(((DOFInetServer) server->server)->socket,
                                       server->address);
            }
        }
#endif
        PCRMutex_Unlock(inetMutex);
        return TRUE;
    }
    return FALSE;
}

#ifndef INET_DISABLE_LOOPBACK
static void _propogateLoopbackData(InetSocket socket,
                                   DOFAddress multicastAddr)
{
    LoopbackNode data = (LoopbackNode) PCRMem_Alloc(sizeof(LoopbackNode_t));

    if(data != NULL)
    {
        const struct sockaddr * sockaddr = InetTransport_GetAddress(multicastAddr);

        uint32 addrSize = InetTransport_GetAddressSize(multicastAddr);

        uint32 sentSize;

        PCRRandom_Generate(LOOPBACK_DATA_SIZE, &(data->data[1]));
        data->data[0] = TRANSPORT_PROTOCOL_IDENTIFIER;

        sentSize = InetSocket_SendTo(socket, data->data, LOOPBACK_DATA_SIZE + 1, 0, sockaddr, addrSize);

        if(sentSize == LOOPBACK_DATA_SIZE + 1)
        {
            data->addr = DOFAddress_Copy(multicastAddr);
            if(data->addr != NULL)
            {
                data->ti = OALTimer_CreateItem(loopbackTimer, _loopbackTimeout, 1, (void**)&data);
                if(data->ti) if(OALTimerItem_Insert( data->ti, LOOPBACK_DATA_TIMEOUT, FALSE, FALSE ) == DOFERROR_OK)
                {
                    void * keys[3];

                    uint16 keySizes[3];

                    if(_prepareSockaddrKey(sockaddr, keys, keySizes))
                    {
                        keys[2] = &(data->data[1]);
                        keySizes[2] = LOOPBACK_DATA_SIZE;
                        OALHashtable_Insert(localLoopbackData, data,
                                            (OALHashKey_t*) keys, keySizes);
                        return;
                    }
                }
                OALTimerItem_Destroy(data->ti, TRUE);
            }
            DOFAddress_Destroy(data->addr);
        }
        PCRMem_Free(data);
    }
}

static boolean _prepareSockaddrKey(const struct sockaddr *  sockaddr,
                                          void     ** keys,
                                          uint16   *  keySizes)
{
    PCR_Assert(sockaddr !=  NULL);
    PCR_Assert(keys != NULL);
    PCR_Assert(keySizes != NULL);

    switch (sockaddr->sa_family)
    {
#if HAVE_DECL_PF_INET6
    case PF_INET6:
    {
        struct sockaddr_in6 * s6 = (struct sockaddr_in6*) sockaddr;

        keys[0] = (void *) &(s6->sin6_addr);
        keys[1] = (void *) &(s6->sin6_port);

        keySizes[0] = sizeof(s6->sin6_addr);
        keySizes[1] = sizeof(s6->sin6_port);
        break;
    }
#endif

#if HAVE_DECL_PF_INET
    case PF_INET:
    {
        struct sockaddr_in * s = (struct sockaddr_in*) sockaddr;

        keys[0] = (void *) &(s->sin_addr);
        keys[1] = (void *) &(s->sin_port);

        keySizes[0] = sizeof(s->sin_addr);
        keySizes[1] = sizeof(s->sin_port);
        break;
    }
#endif

    default:
        return FALSE;
    }
    return TRUE;
}

static int _loopbackTimeout(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown)
{
    LoopbackNode data = (LoopbackNode) argv[0];

    PCRMutex_Lock(inetMutex);
    {
        void * keys[3];

        uint16 keySizes[3];

        if(_prepareSockaddrKey(InetTransport_GetAddress(data->addr),
                               keys, keySizes))
        {
            keys[2] = &(data->data[1]);
            keySizes[2] = LOOPBACK_DATA_SIZE;
            OALHashtable_Remove(localLoopbackData, data,
                                (OALHashKey_t*)keys, keySizes);
        }
    }
    PCRMutex_Unlock(inetMutex);
    OALTimerItem_Destroy(data->ti, FALSE);
    DOFAddress_Destroy(data->addr);
    PCRMem_Free(data);
    return 0;
}

boolean InetManager_LoopbackDetected(const struct sockaddr * fromAddr,
                                            uint8    * data,
                                            uint16     dataSize)
{
    void * keys[3];

    uint16 keySizes[3];

    if(_prepareSockaddrKey(fromAddr, keys, keySizes))
    {
        OALHashItem_t item;

        keys[2] = data;
        keySizes[2] = dataSize;
        PCRMutex_Lock(inetMutex);       /* Should already be locked,
                                           but better safe than sorry */
        item =  OALHashtable_Retrieve(localLoopbackData,
                                      (OALHashKey_t*) keys, keySizes);
        if(item != NULL)
        {
            OALHashtable_Remove(localLoopbackData, item,
                                (OALHashKey_t*) keys, keySizes);
            PCRMutex_Unlock(inetMutex);
            return TRUE;
        }
        PCRMutex_Unlock(inetMutex);
    }
    return FALSE;
}

boolean InetManager_IsLoopbackAddress(const struct sockaddr * sockaddr)
{
    void * keys[2];

    OALHashKeySize_t keySizes[2];

    if(_prepareSockaddrKey(sockaddr, keys, keySizes))
    {
        return (OALHashtable_Retrieve(ignoredAddresses, (OALHashKey_t*) keys, keySizes) != NULL);
    }
    return FALSE;
}

void InetManager_AddLoopbackAddress(const struct sockaddr * sockaddr)
{
    void * keys[2];

    OALHashKeySize_t keySizes[2];

    if(_prepareSockaddrKey(sockaddr, keys, keySizes))
    {
        PCRMutex_Lock(inetMutex);
        OALHashtable_Insert(ignoredAddresses, sockaddr, (OALHashKey_t*) keys, keySizes);
        PCRMutex_Unlock(inetMutex);
    }
}
#endif

boolean InetManager_AddConnection(InetConnection connection,
                                  boolean        isMulticast)
{
    InetSocketNode newNode = PCRMem_Alloc(sizeof(InetSocketNode_t));

    PCR_Assert(connection != NULL);

    if(newNode != NULL)
    {
        const struct sockaddr * sockaddr = InetTransport_GetAddress(connection->addr);

        newNode->socketData.sockFd = connection->socket;
        newNode->socketData.transportHandler = connection->transportHandler;
        newNode->socketData.u.connection = connection;
        newNode->socketData.shouldDestroy = FALSE;
        newNode->socketData.message = inetMessages[INETM_NORMAL];
        newNode->socketData.sendQueue = NULL;
        switch(sockaddr->sa_family)
        {
#if HAVE_DECL_PF_INET6
        case PF_INET6:
            newNode->socketData.isIpV6 = TRUE;
            break;
#endif
#if HAVE_DECL_PF_INET
        case PF_INET:
#endif
        default:
            newNode->socketData.isIpV6 = FALSE;
            break;
        }
        newNode->socketData.isMulticast = isMulticast;
        newNode->socketData.isConnection = TRUE;
        newNode->recvFunction = (isMulticast) ?
                                Inet_ConnectionlessReceive :
                                Inet_ConnectedReceive;
        newNode->sendFunction = Inet_ConnectedSend;

        _addNode(newNode);
        PCRMutex_Lock(inetMutex);
        FD_SET(connection->socket, &recvSet);
        PCREvent_Signal(socketsReady);
#ifndef INET_DISABLE_LOOPBACK
        if(isMulticast)
        {
            _propogateLoopbackData(connection->socket, connection->addr);
        }
#endif
        PCRMutex_Unlock(inetMutex);
        return TRUE;
    }
    return FALSE;
}

boolean InetManager_Connect(InetConnection connection)
{
    InetSocketNode newNode = PCRMem_Alloc(sizeof(InetSocketNode_t));

    PCR_Assert(connection != NULL);

    if(newNode != NULL)
    {
        const struct sockaddr * sockaddr = InetTransport_GetAddress(connection->addr);

        newNode->socketData.sockFd = connection->socket;
        newNode->socketData.transportHandler = connection->transportHandler;
        newNode->socketData.u.connection = connection;
        newNode->socketData.shouldDestroy = FALSE;
        newNode->socketData.message = inetMessages[INETM_NORMAL];
        newNode->socketData.sendQueue = NULL;
        switch(sockaddr->sa_family)
        {
#if HAVE_DECL_PF_INET6
        case PF_INET6:
            newNode->socketData.isIpV6 = TRUE;
            break;
#endif
#if HAVE_DECL_PF_INET
        case PF_INET:
#endif
        default:
            newNode->socketData.isIpV6 = FALSE;
            break;
        }
        newNode->socketData.isMulticast = FALSE;
        newNode->socketData.isConnection = TRUE;
        newNode->recvFunction = NULL;
        newNode->sendFunction = Inet_ConnectionComplete;

        InetSocket_SetOptionNoBlock(connection->socket);
        {
            const struct sockaddr * sockaddr = InetTransport_GetAddress(connection->addr);

            uint32 socklen = InetTransport_GetAddressSize(connection->addr);

            uint32 peeriface = InetTransport_GetAddressInterface( connection->addr );

            if ( peeriface != 0 )
            {
                char ifname[IF_NAMESIZE];
                (void) if_indextoname(peeriface, ifname);
#ifdef SO_BINDTODEVICE              
                (void) setsockopt((connection->socket), (int)SOL_SOCKET, SO_BINDTODEVICE, ifname, sizeof(ifname));
#endif
            }

            InetSocket_Connect(connection->socket, sockaddr, socklen);
        }

        _addNode(newNode);
        PCRMutex_Lock(inetMutex);
        FD_SET(connection->socket, &sendSet);
        PCREvent_Signal(socketsReady);
        PCRMutex_Unlock(inetMutex);
        return TRUE;
    }
    return FALSE;
}

void InetManager_FindAndRemoveSocket(InetSocket sockFd, InetMessages_t message)
{
    PCRMutex_Lock(inetMutex);
    {
        InetSocketNode itr = _findNode(sockFd);
        if(itr != NULL)
        {
            InetManager_RemoveNode(itr, message);
        }
    }
    PCRMutex_Unlock(inetMutex);
}

void InetManager_RemoveNode(InetSocketNode node, InetMessages_t message)
{
    /* inetMutex must already be locked. */
    PCR_Assert(node != NULL);
    if(node->socketData.message == NULL)
    {
        node->socketData.message = inetMessages[message];
    }
    node->socketData.shouldDestroy = TRUE;
    ++cleanupNeeded;
}

void InetManager_PartialSendComplete(InetSocket socket)
{
    PCRMutex_Lock(inetMutex);
    FD_CLR(socket, &sendSet);
    PCRMutex_Unlock(inetMutex);
}

void InetManager_Send(InetSocket socket,
                      uint8 *    data,
                      uint32     offset,
                      uint32     length)
{
    PCR_Assert(socket != PCRSOCKET_INVALID);
    PCR_Assert(data != NULL);
    PCR_Assert(offset < length);

    PCRMutex_Lock(inetMutex);
    {
        InetSocketNode socketNode = _findNode(socket);

        if((socketNode != NULL) && (!socketNode->socketData.shouldDestroy))
        {
            fd_set temp;

            FD_ZERO(&temp);
            FD_SET(socket, &temp);
            if(socketNode->socketData.sendQueue == NULL &&
               InetSocket_Select(socket + 1, NULL, &temp, NULL, 0) != 0)
            {
                uint32 toWrite = length - offset;
                uint32 bytesWritten = InetSocket_Send(socket, &data[offset], toWrite, MSG_DONTWAIT);

                if(bytesWritten < toWrite)
                {
                    _addPartialSendInfo(socketNode, data, bytesWritten + offset, length, NULL);
                }
                else
                {
                    PCRMem_Free(data);
                    if(bytesWritten == 0xFFFFFFFF)
                    {
                        if (errno == EAGAIN || errno == EWOULDBLOCK)
                        {
                            _addPartialSendInfo(socketNode, data, offset, length, NULL);
                        }
                        else
                        {
                            /* Error */
                            InetManager_RemoveNode(socketNode, INETM_SEND_FAILED);
                        }
                    }
                }
            }
            else
            {
                _addPartialSendInfo(socketNode, data, offset, length, NULL);
            }
        }
        else
        {
            PCRMem_Free(data);
        }
    }
    PCRMutex_Unlock(inetMutex);
}

void InetManager_SendTo(InetSocket socket,
                        uint8 *    data,
                        uint32     offset,
                        uint32     length,
                        DOFAddress target)
{
    PCR_Assert(socket != PCRSOCKET_INVALID);
    PCR_Assert(data != NULL);
    PCR_Assert(offset < length);
    PCR_Assert(target != NULL);

    PCRMutex_Lock(inetMutex);
    {
        InetSocketNode socketNode = _findNode(socket);

        if((socketNode != NULL) && (!socketNode->socketData.shouldDestroy))
        {
            fd_set temp;

            FD_ZERO(&temp);
            FD_SET(socket, &temp);
            if(socketNode->socketData.sendQueue == NULL &&
               InetSocket_Select(socket + 1, NULL, &temp, NULL, 0) != 0)
            {
                Inet_Connectionless_SendTo( socket, target, length - offset, &data[offset] );
                /* Ignore errors on directed datagram socket sends. */
                PCRMem_Free(data);
            }
            else
            {
                FD_ZERO(&temp);
                FD_SET(socket, &temp);
                if(InetSocket_Select(socket + 1, NULL, NULL, &temp, 0) == 0)
                {
                    _addPartialSendInfo(socketNode, data, offset, length, NULL);
                }
            }
        }
        else
        {
            PCRMem_Free(data);
        }
    }
    PCRMutex_Unlock(inetMutex);
}

static void Inet_ConnectionComplete(InetSocketNode node)
{
    /* inetMutex must already be locked */
    PCR_Assert(node != NULL);

    InetSocket_SetOptionNoBlock(node->socketData.sockFd);
    InetSocket_SetOptionNoDelay(node->socketData.sockFd);
    {
        node->recvFunction = Inet_ConnectedReceive;
        node->sendFunction = Inet_ConnectedSend;
    }
    FD_CLR(node->socketData.sockFd, &sendSet);
    FD_SET(node->socketData.sockFd, &recvSet);
    PCRMutex_Unlock(inetMutex);
    if(!node->socketData.transportHandler->fns->Opened(node->socketData.transportHandler,
                                                       (DTLConnection) node->socketData.u.connection))
    {
        PCRMutex_Lock(inetMutex);
        InetManager_RemoveNode(node, INETM_NORMAL);
    }
    else
    {
        PCRMutex_Lock(inetMutex);
    }
}

static InetSocketNode _findNode(InetSocket sockFd)
{
    /* inetMutex must already be locked */
    InetSocketNode ret = socketListHead;

    while((ret != NULL) && (ret->socketData.sockFd < sockFd))
    {
        ret = ret->next;
    }
    if((ret != NULL) && (ret->socketData.sockFd == sockFd))
    {
        return ret;
    }
    return NULL;
}

static void _addPartialSendInfo(InetSocketNode socketNode,
                                uint8      * data,
                                uint32       offset,
                                uint32       length,
                                DOFAddress   target)
{
    /* inetMutex must already be locked */
    PCR_Assert(socketNode != NULL);
    if(!socketNode->socketData.shouldDestroy)
    {
        if(socketNode->socketData.sendQueue == NULL)
        {
            socketNode->socketData.sendQueue = OALQueue_Create();
        }
        if(socketNode->socketData.sendQueue != NULL)
        {
            PartialSendData newNode = (PartialSendData) target == NULL        ?
                                      PCRMem_Alloc(sizeof(PartialSendData_t)) :
                                      PCRMem_Alloc(sizeof(PartialConnectionlessSendData_t));

            if(newNode != NULL)
            {
                newNode->data   = data;
                newNode->length = length;
                newNode->offset = offset;
                if(target != NULL)
                {
                    ((PartialConnectionlessSendData) newNode)->address = DOFAddress_Copy(target);
                }

                if(OALQueue_Enqueue(socketNode->socketData.sendQueue, (void *)newNode))
                {
                    FD_SET(socketNode->socketData.sockFd, &sendSet);
                }
                else
                {
                    PCRMem_Free(newNode);
                    PCRMem_Free(data);
                }
            }
        }
    }
}
