/*
**  Copyright (c) 2014-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef INET_MANAGER_H
#define INET_MANAGER_H

#include <dof/pcr.h>
#include "socket.h"
#include <dof/oal.h>
#include <dof/inet.h>

#include <internal/queue.h>

#include "inet_connection.h"
#include "inet_server.h"

#define     TRANSPORT_PROTOCOL_IDENTIFIER   127

extern PCRMutex inetMutex;

typedef struct InetSocketData_t
{
    int                   sockFd;

    DTLTransportHandler   transportHandler;

    union
    {
        InetConnection    connection;
        DOFInetServer     server;
    } u;

    OALQueue              sendQueue;

    DOFException          message;

    boolean               isIpV6;
    boolean               isMulticast;
    boolean               shouldDestroy;
    boolean               isConnection;
} InetSocketData_t;

typedef struct InetSocketNode_t * InetSocketNode;

typedef void (*InetHandlerFunction) (InetSocketNode node);

typedef struct InetSocketNode_t
{
           InetSocketData_t      socketData;

           InetHandlerFunction   recvFunction;
           InetHandlerFunction   sendFunction;

    struct InetSocketNode_t    * next;
} InetSocketNode_t;

typedef struct PartialSendData_t
{
    uint8  * data;

    uint32   length;
    uint32   offset;
} PartialSendData_t;

typedef struct PartialConnectionlessSendData_t
{
    PartialSendData_t sendData;

    DOFAddress        address;
} PartialConnectionlessSendData_t;

typedef InetSocketData_t * InetSocketData;

typedef PartialSendData_t * PartialSendData;

typedef PartialConnectionlessSendData_t * PartialConnectionlessSendData;

typedef enum InetMessages_t
{
    INETM_NORMAL,
    INETM_ERROR,
    INETM_CONNECTION_CLOSED,
    INETM_RECEIVE_FAILED,
    INETM_SEND_FAILED,
    INETM_BIND_FAILED
} InetMessages_t;

extern DOFException inetMessages[6];

boolean InetManager_AddServer(DOFInetServer server,
                              boolean       isMulticast);

boolean InetManager_AddConnection(InetConnection connection,
                                  boolean        isMulticast);

boolean InetManager_Connect(InetConnection connection);

#ifndef INET_DISABLE_LOOPBACK
boolean InetManager_LoopbackDetected(const struct sockaddr * fromAddr,
                                            uint8    * data,
                                            uint16     dataSize);

boolean InetManager_IsLoopbackAddress(const struct sockaddr * sockaddr);

void InetManager_AddLoopbackAddress(const struct sockaddr * sockaddr);
#endif

void InetManager_FindAndRemoveSocket(InetSocket sockFd, InetMessages_t message);

void InetManager_RemoveNode(InetSocketNode node, InetMessages_t message);

void InetManager_PartialSendComplete(InetSocket socket);

void InetManager_Send(InetSocket socket,
                      uint8 *    data,
                      uint32     offset,
                      uint32     length);

void InetManager_SendTo(InetSocket socket,
                        uint8 *    data,
                        uint32     offset,
                        uint32     length,
                        DOFAddress target);

boolean InetTransport_IsRunning(void);

extern boolean InetTransport_ShouldReresolve(const DOFAddress addr);

extern boolean InetTransport_ResolveAddress( const DOFAddress addr );

extern void InetTransport_ClearResolve(const DOFAddress addr);

extern DOFAddressType InetTransport_GetAddressType( DOFAddress addr );

extern int InetTransport_CompareAddress( uint32 size, const void * addr1, const void * addr2 );

extern boolean InetTransport_AddressCopy( uint32 size, void * addr );

extern void InetTransport_AddressDestroy( uint32 size, const void * addr );

#endif
