/*
**  Copyright (c) 2014-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include "config.h"
#include <dof/pcr.h>
#include "socket.h"
#include <dof/oal.h>
#include <dof/inet.h>

#include "inet_connection.h"
#include "inet_manager.h"
#include "inet_receive.h"

#ifndef PF_INET
#define PF_INET AF_INET
#endif
#define HAVE_DECL_PF_INET 1

#ifndef PF_INET6
#define PF_INET6 AF_INET6
#endif
#define HAVE_DECL_PF_INET6 1


void Inet_AcceptConnection(InetSocketNode node)
{
    /* inetMutex must already be locked */
#if HAVE_DECL_PF_INET6
    struct sockaddr_in6 sockaddr6;
    struct sockaddr * sockaddr = (struct sockaddr *) &sockaddr6;
    uint32 sockaddrSize = sizeof(sockaddr6);
#else
#if HAVE_DECL_PF_INET
    struct sockaddr_in sockaddr4;
    struct sockaddr * sockaddr = (struct sockaddr *) &sockaddr4;
    uint32 sockaddrSize = sizeof(sockaddr4);
#else
    struct sockaddr * sockaddr = NULL;
    uint32 sockaddrSize = 0;
#endif
#endif


    int newSocket;

    PCR_Assert(node != NULL);

    if(sockaddr != NULL)
    {
        newSocket = InetSocket_Accept(node->socketData.sockFd, sockaddr, &sockaddrSize);
        if(newSocket != PCRSOCKET_INVALID)
        {
            DOFAddress connAddr = InetTransport_CreateAddress_Sockaddr(sockaddr, sockaddrSize);

            if(connAddr != NULL)
            {
                InetConnection conn;

                PCRMutex_Unlock(inetMutex);
                conn = InetConnection_Create(connAddr, newSocket,
                                                            node->socketData.transportHandler,
                                                            (DTLServer) node->socketData.u.server,
                                                            node->socketData.u.server->type,
                                                            TRUE, DOFADDRESSTYPE_UNICAST, NULL, NULL);

                if(conn != NULL)
                {
                    InetSocket_SetOptionNoDelay(newSocket);
                    InetManager_AddConnection(conn, FALSE);
                    if(!node->socketData.transportHandler->fns->Opened(node->socketData.transportHandler,
                                                                       (DTLConnection) conn))
                    {
                        InetManager_FindAndRemoveSocket(newSocket, INETM_NORMAL);
                    }
                }
                DOFAddress_Destroy( connAddr );
                PCRMutex_Lock(inetMutex);
            }
        }
    }
}

void Inet_ConnectedReceive(InetSocketNode node)
{
    /* inetMutex must already be locked */
    PCR_Assert(node != NULL);
    if(!node->socketData.isMulticast)
    {
        uint32 bufferSize = InetSocket_GetRecvAvail(node->socketData.sockFd);

        uint8 * recvBuffer = (uint8 *) PCRMem_Alloc(sizeof(uint8) * bufferSize);

        if(recvBuffer != NULL)
        {
            uint32 recvSize = InetSocket_Recv(node->socketData.sockFd, recvBuffer, bufferSize, 0);

            switch(recvSize)
            {
            case 0:     /* connection closed */
            case 0xFFFFFFFF:    /* error occurred */
                if(recvSize == 0)
                {
                    InetManager_RemoveNode(node, INETM_CONNECTION_CLOSED);
                }
                else
                {
                    InetManager_RemoveNode(node, INETM_RECEIVE_FAILED);
                }
                PCRMem_Free(recvBuffer);
                break;

            default:    /* send message somewhere */
                PCRMutex_Unlock(inetMutex);
                node->socketData.transportHandler->fns->
                        Received_Connection(node->socketData.transportHandler,
                                            (DTLConnection) node->socketData.u.connection,
                                            recvSize, recvBuffer,
                                            node->socketData.u.connection->addr,
                                            node->socketData.u.connection->aType);
                PCRMutex_Lock(inetMutex);
                break;
            }
        }
    }
    else
    {
        Inet_ConnectionlessReceive(node);
    }
}

void Inet_ConnectionlessReceive(InetSocketNode node)
{
    /* inetMutex must already be locked */
    uint32 bufferSize = InetSocket_GetRecvAvail(node->socketData.sockFd);

#if HAVE_DECL_PF_INET6
    struct sockaddr_in6 sockaddr6;
    struct sockaddr * sockaddr = (struct sockaddr *) &sockaddr6;
    uint32 sockaddrSize = sizeof(sockaddr6);
#else
#if HAVE_DECL_PF_INET
    struct sockaddr_in sockaddr4;
    struct sockaddr * sockaddr = (struct sockaddr *) &sockaddr4;
    uint32 sockaddrSize = sizeof(sockaddr4);
#else
    struct sockaddr * sockaddr = NULL;
    uint32 sockaddrSize = 0;
#endif
#endif

    PCR_Assert(node != NULL);

    if(sockaddr != NULL)
    {
        uint8 * recvBuffer = (uint8 *) PCRMem_Alloc(sizeof(uint8) * bufferSize);

        if(recvBuffer != NULL)
        {
            uint32 recvSize = InetSocket_RecvFrom(node->socketData.sockFd, recvBuffer, bufferSize,
                                                  0, sockaddr, &sockaddrSize);

            if(recvSize != 0xFFFFFFFF)
            {
#ifndef INET_DISABLE_LOOPBACK
                if(node->socketData.isMulticast)
                {
                    if(InetManager_IsLoopbackAddress(sockaddr))
                    {
                        goto cleanup;
                    }
                    else if(recvBuffer[0] == TRANSPORT_PROTOCOL_IDENTIFIER)
                    {
                        const struct sockaddr * myAddr;

                        if(node->socketData.isConnection)
                        {
                            myAddr = InetTransport_GetAddress(node->socketData.u.connection->addr);
                        }
                        else
                        {
                            myAddr = InetTransport_GetAddress(node->socketData.u.server->address);
                        }
                        if(InetManager_LoopbackDetected(myAddr, &recvBuffer[1], recvSize - 1))
                        {
                            InetManager_AddLoopbackAddress(sockaddr);
                        }
                        goto cleanup;
                    }
                }
#endif

                PCRMutex_Unlock(inetMutex);
                if(node->socketData.isConnection)
                {
                    DOFAddress fromAddr = node->socketData.u.connection->addr;

                    if(node->socketData.isMulticast)
                    {
                        fromAddr = InetTransport_CreateAddress_Sockaddr(sockaddr, sockaddrSize);
                    }
                    if(fromAddr != NULL)
                    {
                        node->socketData.transportHandler->fns->
                            Received_Connection(node->socketData.transportHandler,
                                                (DTLConnection) node->socketData.u.connection,
                                                recvSize, recvBuffer,
                                                fromAddr,
                                                DOFADDRESSTYPE_UNICAST);
                        if(node->socketData.isMulticast)
                        {
                            DOFAddress_Destroy(fromAddr);
                        }
                    }
                    else
                    {
                        PCRMem_Free(recvBuffer);
                    }
                }
                else
                {
                    DOFAddress fromAddr = InetTransport_CreateAddress_Sockaddr(sockaddr, sockaddrSize);

                    if(fromAddr != NULL)
                    {
                        node->socketData.transportHandler->fns->
                                Received_Server(node->socketData.transportHandler,
                                                (DTLServer) node->socketData.u.server,
                                                recvSize, recvBuffer,
                                                fromAddr, node->socketData.u.server->aType);
                        DOFAddress_Destroy(fromAddr);
                    }
                    else
                    {
                        PCRMem_Free(recvBuffer);
                    }
                }
                PCRMutex_Lock(inetMutex);
                return;
            }
            else if ( node->socketData.isConnection )
            {
                InetManager_RemoveNode(node, INETM_RECEIVE_FAILED);
            }

#ifndef INET_DISABLE_LOOPBACK
        cleanup:
#endif
            PCRMem_Free(recvBuffer);
        }
    }
}
