/*
**  Copyright (c) 2014-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include "config.h"
#include <dof/pcr.h>
#include "socket.h"
#include <dof/oal.h>
#include <dof/inet.h>

#include "inet_server.h"
#include "inet_manager.h"

#include <sys/socket.h>

#ifndef PF_INET
#define PF_INET AF_INET
#endif
#define HAVE_DECL_PF_INET 1

#ifndef PF_INET6
#define PF_INET6 AF_INET6
#endif
#define HAVE_DECL_PF_INET6 1

static void _DOFInetServer_Send(DTLServer  server,
                                DOFAddress target,
                                uint8 *    data,
                                uint32     offset,
                                uint32     length);

static void _DOFInetServer_Stop(DTLServer server);

static uint32 _DODFInetServer_GetConnectionCount(DTLServer server);

static DOFAddress _DOFInetServer_GetAddress(DTLServer server);

static void * _DOFInetServer_GetContext(DTLServer server);

static DTLServer _DOFInetServer_GetRelatedServer(DTLServer server);

static DTLTransportHandler _DOFInetServer_GetHandler(DTLServer server);

static DOFServerType _DOFInetServer_GetServerType(DTLServer server);

static boolean _DOFInetServer_ReturnFalse(DTLServer server);

static DTLConnectionConfig _DOFInetServer_GetConnConfig(DTLServer server, DOFAddress address);

extern PCRMutex ref_count_mutex;

DOFInetServer DOFInetServer_Create(DOFAddress            addr,
                                   DTLTransportHandler   transportHandler,
                                   DOFConnectionType     type,
                                   DOFAddressType        aType,
                                   InetSocket            socket,
                                   DTLServer             relatedServer,
                                   void                * context)
{
    DOFInetServer server;

    static const struct DTLServerFns_t serverCallbacks =
    {
        _DOFInetServer_Send,
        _DOFInetServer_Stop,
        _DODFInetServer_GetConnectionCount,
        _DOFInetServer_GetAddress,
        _DOFInetServer_GetRelatedServer,
        _DOFInetServer_GetHandler,
        _DOFInetServer_GetServerType,
        _DOFInetServer_GetContext,
        _DOFInetServer_ReturnFalse,
        _DOFInetServer_ReturnFalse,
        _DOFInetServer_GetConnConfig
    };

    PCR_Assert(addr != NULL);
    PCR_Assert(transportHandler != NULL);
    PCR_Assert(socket != PCRSOCKET_INVALID);

    server = (DOFInetServer) PCRMem_Alloc(sizeof(DOFInetServer_t));
    if(server != NULL)
    {
        server->super.fns = &serverCallbacks;
        server->address = DOFAddress_Copy(addr);
        server->socket = socket;
        server->type = type;
        server->aType = aType;
        server->handler = transportHandler;
        server->server = relatedServer;
        server->context = context;
        server->connectionCount = 0;
        server->refCount = 1;
        server->isRunning = TRUE;
    }
    return server;
}

void DOFInetServer_Retain(DOFInetServer self)
{
    PCR_Assert(self != NULL);
    PCRMutex_Lock(ref_count_mutex);
    ++(self->refCount);
    PCRMutex_Unlock(ref_count_mutex);
}

void DOFInetServer_Destroy(DOFInetServer self)
{
    boolean gone;
    PCR_Assert(self != NULL);
    PCRMutex_Lock(ref_count_mutex);
    --(self->refCount);
    gone = (self->refCount == 0);
    PCRMutex_Unlock(ref_count_mutex);
    if ( gone )
    {
        _DOFInetServer_Stop((DTLServer) self);
        DOFAddress_Destroy(self->address);
        InetSocket_Destroy(self->socket);
        PCRMem_Free(self);
    }
}

DOFInetServerConfig DOFInetServerConfig_Create(DOFAddress        addr,
                                               InetSocket         socket,
                                               DOFConnectionType type)
{
    DOFInetServerConfig ret;

    PCR_Assert(addr != NULL);
    PCR_Assert(socket != PCRSOCKET_INVALID);

    ret = (DOFInetServerConfig) PCRMem_Alloc(sizeof(DOFInetServerConfig_t));
    if(ret != NULL)
    {
        ret->addr = DOFAddress_Copy(addr);
        ret->socket = socket;
        ret->type = type;
    }
    return ret;
}

void DOFInetServerConfig_Destroy(DOFInetServerConfig config)
{
    PCR_Assert(config != NULL);

    if(config != NULL)
    {
        DOFAddress_Destroy(config->addr);
        InetSocket_Destroy(config->socket);
        PCRMem_Free(config);
    }
}

static void _DOFInetServer_Send(DTLServer  server,
                                DOFAddress target,
                                uint8 *    data,
                                uint32     offset,
                                uint32     length)
{
    const struct sockaddr * saddr;

#if HAVE_DECL_PF_INET6
    static const uint8 zeroAddr6[16] = {0, 0, 0, 0,
                                        0, 0, 0, 0,
                                        0, 0, 0, 0,
                                        0, 0, 0, 0};
#endif

#if HAVE_DECL_PF_INET
    static const uint8 zeroAddr4[4] = {0, 0, 0, 0};
#endif

    PCR_Assert(server != NULL);
    PCR_Assert(target != NULL);

    saddr = InetTransport_GetAddress(target);
    switch (saddr->sa_family)
    {
#if HAVE_DECL_PF_INET6
    case PF_INET6:
        {
            const struct sockaddr_in6 * realSockaddr = (const struct sockaddr_in6*) saddr;

            if(PCRMem_Compare(&realSockaddr->sin6_addr, zeroAddr6, 16) == 0)
            {
                goto exitEarly;
            }
        }
        break;
#endif

#if HAVE_DECL_PF_INET
    case PF_INET:
        {
            const struct sockaddr_in * realSockaddr = (const struct sockaddr_in*) saddr;

            if(PCRMem_Compare(&realSockaddr->sin_addr, zeroAddr4, 4) == 0)
            {
                goto exitEarly;
            }
        }
        break;
#endif

    default:
        break;
    }

    {
        DOFInetServer realServer;

        realServer = (DOFInetServer) server;
        InetManager_SendTo(realServer->socket, data, offset, length, target);
        return;
    }

exitEarly:
    PCRMem_Free(data);
}

static void _DOFInetServer_Stop(DTLServer server)
{
    DOFInetServer serv;

    PCR_Assert(server != NULL);
    serv = (DOFInetServer) server;
    PCRMutex_Lock(inetMutex);
    if(serv->isRunning)
    {
        PCR_Assert(serv->socket != PCRSOCKET_INVALID);
        InetManager_FindAndRemoveSocket(serv->socket, INETM_NORMAL);
        serv->isRunning = FALSE;
    }
    PCRMutex_Unlock(inetMutex);
}

static uint32 _DODFInetServer_GetConnectionCount(DTLServer server)
{
    PCR_Assert(server != NULL);
    return ((DOFInetServer) server)->connectionCount;
}

static DOFAddress _DOFInetServer_GetAddress(DTLServer server)
{
    PCR_Assert(server != NULL);
    return ((DOFInetServer) server)->address;
}

static void * _DOFInetServer_GetContext(DTLServer server)
{
    PCR_Assert(server != NULL);
    return ((DOFInetServer) server)->context;
}

static DTLServer _DOFInetServer_GetRelatedServer(DTLServer server)
{
    PCR_Assert(server != NULL);
    return ((DOFInetServer) server)->server;
}

static DTLTransportHandler _DOFInetServer_GetHandler(DTLServer server)
{
    PCR_Assert(server != NULL);
    return ((DOFInetServer) server)->handler;
}

static DOFServerType _DOFInetServer_GetServerType(DTLServer server)
{
    PCR_Assert(server != NULL);
    return ((DOFInetServer) server)->type;
}

static boolean _DOFInetServer_ReturnFalse(DTLServer server)
{
    return FALSE;
}

static DTLConnectionConfig _DOFInetServer_GetConnConfig(DTLServer server, DOFAddress address)
{
    return NULL;
}
