/*
**  Copyright (c) 2014-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include "config.h"
#include <dof/pcr.h>
#include "socket.h"
#include <dof/oal.h>
#include <dof/inet.h>

#include <net/if.h>
#include <sys/socket.h>

#include "inet_connection.h"
#include "inet_manager.h"
#include "inet_server.h"
#include "inet_send.h"

#define PROTO_TCP   6
#define PROTO_UDP   17

#ifndef PF_INET
#define PF_INET AF_INET
#endif
#define HAVE_DECL_PF_INET 1

#ifndef PF_INET6
#define PF_INET6 AF_INET6
#endif
#define HAVE_DECL_PF_INET6 1

static boolean _DOFInetTransport_Open(DTLTransport          transport,
                       DTLConnectionConfig   config,
                       DOFAddress            peer,
                       DTLTransportHandler   handler,
                       DOFConnectionType     type,
                       void                * context);

static boolean _DOFInetTransport_Start(DTLTransport       transport,
                        DTLServerConfig       config,
                        DOFAddress            listenAddr,
                        DTLTransportHandler   handler,
                        DOFServerType         type,
                        DTLServer             relatedServer,
                        void                * context);

static uint32 _DOFInetTransport_GetMTU(DTLTransport      transport,
                                       DOFConnectionType type);

static DOFAddressType _DOFInetTransport_GetAddressType( DTLTransport transport, DOFAddress addr );

static int _DOFInetTransport_AddreesCompare( DTLTransport transport, uint32 size, const void * addr1, const void * addr2 );

static boolean _DOFInetTransport_AddressCopy( DTLTransport transport, uint32 size, void * addr );

static void _DOFInetTransport_AddressDestroy( DTLTransport transport, uint32 size, const void * addr );

static InetSocket _CreateSocket(const struct sockaddr   * saddr,
                                      DOFConnectionType   type);

const struct DTLTransportFns_t DOFInetTransportFuncs =
{
    _DOFInetTransport_Open, /* Open */
    _DOFInetTransport_Start, /* Start */
    _DOFInetTransport_GetMTU, /* GetMaxSize */
    NULL, /* GetMinSize */
    _DOFInetTransport_GetAddressType,
    NULL, /* IsConnectionConfigEqual */
    NULL, /* IsServerConfigEqual */
    _DOFInetTransport_AddreesCompare, /* AddressCompare */
    _DOFInetTransport_AddressCopy, /* AddressCopy */
    _DOFInetTransport_AddressDestroy  /* AddressDestroy */
}; /* Transport-specific function calls. */

const struct DTLTransport_t DOFInetTransport =
{
    &DOFInetTransportFuncs
};

const DTLTransport InetTransport = (const DTLTransport) &DOFInetTransport;

static boolean _DOFInetTransport_Open( DTLTransport          transport,
                                       DTLConnectionConfig   config,
                                       DOFAddress            peer,
                                       DTLTransportHandler   handler,
                                       DOFConnectionType     type,
                                       void                * context )
{
    if( InetTransport_ShouldReresolve(peer))
    {
        if( !InetTransport_ResolveAddress(peer) )
        {
            return FALSE;
        }
    }

    {
        const struct sockaddr *saddr = InetTransport_GetAddress(peer);
        InetSocket sock;
        uint32 peeriface = InetTransport_GetAddressInterface( peer );

        /* Setup the connected socket. */
        sock = _CreateSocket(saddr, type);
        if ( sock != PCRSOCKET_INVALID )
        {
            InetConnection conn;
            DOFAddressType aType = (type == DOFCONNECTIONTYPE_STREAM) ? DOFADDRESSTYPE_UNICAST : _DOFInetTransport_GetAddressType( transport, peer );

            if ( aType != DOFADDRESSTYPE_UNICAST )
            {
                switch (saddr->sa_family)
                {
    #if HAVE_DECL_PF_INET6
                    case PF_INET6:
                    {
                        struct sockaddr_in6 mcaddr;
                        PCRMem_Copy( &mcaddr, saddr, sizeof(mcaddr) );
                        PCRMem_Set( &mcaddr.sin6_addr.s6_addr, 0, sizeof(mcaddr.sin6_addr.s6_addr) );
                        mcaddr.sin6_port = 0;
                        InetSocket_SetOptionReuseAddr( sock );
    #ifdef INET_DISABLE_LOOPBACK
                        InetSocket_SetOptionBroadcast( sock );
    #endif
                        if ( peeriface != 0 )
                        {
                            char ifname[IF_NAMESIZE];
    #ifdef SO_BINDTODEVICE                          
                            (void) setsockopt((sock), (int)SOL_SOCKET, SO_BINDTODEVICE, if_indextoname(peeriface, ifname), sizeof(const char*));
    #endif                      
                            (void) setsockopt((sock), (int)IPPROTO_IPV6, IP_MULTICAST_IF, &peeriface, sizeof(peeriface));
                        }
                        if ( InetSocket_Bind( sock, (struct sockaddr *)&mcaddr, sizeof(mcaddr) ) != 0 )
                            goto bail;
                        break;
                    }
    #endif

    #if HAVE_DECL_PF_INET
                    case PF_INET:
                    {
                        struct sockaddr_in mcaddr;
                        PCRMem_Copy( &mcaddr, saddr, sizeof(mcaddr) );
                        PCRMem_Set( &mcaddr.sin_addr.s_addr, 0, sizeof(mcaddr.sin_addr.s_addr) );
                        mcaddr.sin_port = 0;
                        InetSocket_SetOptionReuseAddr( sock );
    #ifdef INET_DISABLE_LOOPBACK
                        InetSocket_SetOptionBroadcast( sock );
    #endif
                        if ( peeriface != 0 )
                        {
                            char ifname[IF_NAMESIZE];
    #ifdef SO_BINDTODEVICE                          
                            (void) setsockopt((sock), (int)SOL_SOCKET, SO_BINDTODEVICE, if_indextoname(peeriface, ifname), sizeof(const char*));
    #endif                      
    #ifdef HAVE_STRUCT_IP_MREQN /* mreqn has the interface id which we can use, but mreq doesn't, so we're out of luck if that's all we have. */
                            {
                                struct ip_mreqn ipmreq;
                                ipmreq.imr_ifindex = peeriface;
                                ipmreq.imr_address.s_addr = INADDR_ANY;
                                (void)PCRMem_Copy(&ipmreq.imr_multiaddr.s_addr, &((const struct sockaddr_in*)saddr)->sin_addr.s_addr, 4);
                                (void) setsockopt((sock), (int)IPPROTO_IP, IP_MULTICAST_IF, &ipmreq, sizeof(ipmreq));
                            }
    #endif
                        }
                        if ( InetSocket_Bind( sock, (struct sockaddr *)&mcaddr, sizeof(mcaddr) ) != 0 )
                            goto bail;
                        break;
                    }
    #endif

                    default:
                        break;
                }

                conn = InetConnection_Create(peer, sock, handler,
                                             NULL, type, FALSE,
                                             aType,
                                             config, context);

                if(conn != NULL)
                {
                    if(InetManager_AddConnection(conn, TRUE))
                    {
                        handler->fns->Opened(handler, (DTLConnection) conn);
                        return TRUE;
                    }
                    InetConnection_Destroy(conn);
                }
            }
            else
            {
                if ( peeriface != 0 )
                {
                    char ifname[IF_NAMESIZE];
    #ifdef SO_BINDTODEVICE  
                    (void) setsockopt((sock), (int)SOL_SOCKET, SO_BINDTODEVICE, if_indextoname(peeriface, ifname), sizeof(const char*));
    #endif
                }
                conn = InetConnection_Create(peer, sock, handler, NULL,
                                             type, TRUE, aType,
                                             config, context);

                if(conn != NULL)
                {
                    if(!InetManager_Connect(conn))
                    {
                        InetConnection_Destroy(conn);
                        goto bail;
                    }
                    return TRUE;
                }
            }
    }
    bail:
        InetSocket_Destroy( sock );
    }
    return FALSE;
}

static boolean _DOFInetTransport_Start(DTLTransport          transport,
                        DTLServerConfig       config,
                        DOFAddress            listenAddr,
                        DTLTransportHandler   handler,
                        DOFServerType         type,
                        DTLServer             relatedServer,
                        void                * context)
{
    if( InetTransport_ShouldReresolve(listenAddr))
    {
        if( !InetTransport_ResolveAddress(listenAddr) )
        {
            return FALSE;
        }
    }

    {
        const struct sockaddr *saddr = InetTransport_GetAddress(listenAddr);
        uint32 slen = InetTransport_GetAddressSize(listenAddr);
        InetSocket sock;
        uint32 peeriface = InetTransport_GetAddressInterface( listenAddr );
        DOFInetServer newServer = NULL;

        /* Setup the listening socket. */
        sock = _CreateSocket(saddr, type);

        if ( sock != PCRSOCKET_INVALID )
        {
            DOFAddressType aType = (type == DOFCONNECTIONTYPE_STREAM) ? DOFADDRESSTYPE_UNICAST : _DOFInetTransport_GetAddressType( transport, listenAddr );
            newServer = DOFInetServer_Create(listenAddr, handler, type, aType, sock, relatedServer, context);
            if(newServer != NULL)
            {
                if ( aType != DOFADDRESSTYPE_UNICAST )
                {
                    switch (saddr->sa_family)
                    {
    #if HAVE_DECL_PF_INET6
                        case PF_INET6:
                        {
                            struct sockaddr_in6 mcaddr;
                            PCRMem_Copy( &mcaddr, saddr, sizeof(mcaddr) );
                            PCRMem_Set( &mcaddr.sin6_addr.s6_addr, 0, sizeof(mcaddr.sin6_addr.s6_addr) );
                            InetSocket_SetOptionReuseAddr( sock );
    #ifdef INET_DISABLE_LOOPBACK
                            InetSocket_SetOptionBroadcast( sock );
    #endif
                            if ( peeriface != 0 )
                            {
                                char ifname[IF_NAMESIZE];
    #ifdef SO_BINDTODEVICE                              
                                (void) setsockopt((sock), (int)SOL_SOCKET, SO_BINDTODEVICE, if_indextoname(peeriface, ifname), sizeof(const char*));
    #endif                          
                            }
                            if ( InetSocket_Bind( sock, (struct sockaddr *)&mcaddr, sizeof(mcaddr) ) != 0 )
                                goto bail;
                            break;
                        }
    #endif

    #if HAVE_DECL_PF_INET
                        case PF_INET:
                        {
                            struct sockaddr_in mcaddr;
                            PCRMem_Copy( &mcaddr, saddr, sizeof(mcaddr) );
                            PCRMem_Set( &mcaddr.sin_addr.s_addr, 0, sizeof(mcaddr.sin_addr.s_addr) );
                            InetSocket_SetOptionReuseAddr( sock );
    #ifdef INET_DISABLE_LOOPBACK
                            InetSocket_SetOptionBroadcast( sock );
    #endif
                            if ( peeriface != 0 )
                            {
                                char ifname[IF_NAMESIZE];
    #ifdef SO_BINDTODEVICE
                                (void) setsockopt((sock), (int)SOL_SOCKET, SO_BINDTODEVICE, if_indextoname(peeriface, ifname), sizeof(const char*));
    #endif
                            }
                            if ( InetSocket_Bind( sock, (struct sockaddr *)&mcaddr, sizeof(mcaddr) ) != 0 )
                                goto bail;
                            break;
                        }
    #endif

                        default:
                            break;
                    }

                    InetSocket_MulticastJoin( sock, saddr, slen, InetTransport_GetAddressInterface( listenAddr ) );
                }
                else
                {
                    if ( peeriface != 0 )
                    {
                        char ifname[IF_NAMESIZE];
    #ifdef SO_BINDTODEVICE                      
                        (void) setsockopt((sock), (int)SOL_SOCKET, SO_BINDTODEVICE, if_indextoname(peeriface, ifname), sizeof(const char*));
    #endif                  
                    }
                    if ( InetSocket_Bind( sock, saddr, slen ) != 0 )
                        goto bail;
                }

                if(InetManager_AddServer(newServer, aType != DOFADDRESSTYPE_UNICAST))
                {
                    handler->fns->Started(handler, (DTLServer) newServer);
                    return TRUE;
                }
              bail:
                handler->fns->Stopped(handler, (DTLServer) newServer, inetMessages[INETM_BIND_FAILED]);
            }
            if(newServer != NULL)
            {
                DOFInetServer_Destroy(newServer);
            }
            InetSocket_Destroy( sock );
        }
        return newServer != NULL;
    }
    return FALSE;
}

static uint32 _DOFInetTransport_GetMTU(DTLTransport      transport,
                                       DOFConnectionType type)
{
    if(type == DOFCONNECTIONTYPE_STREAM)
    {
        return 0xFFFFFF;
    }
    else if(type == DOFCONNECTIONTYPE_DATAGRAM)
    {
        return 0x2000;
    }
    return 0;
}

static DOFAddressType _DOFInetTransport_GetAddressType( DTLTransport transport, DOFAddress addr )
{
    return InetTransport_GetAddressType(addr);
}

static int _DOFInetTransport_AddreesCompare( DTLTransport transport, uint32 size, const void * addr1, const void * addr2 )
{
    return InetTransport_CompareAddress( size, addr1, addr2);
}

static boolean _DOFInetTransport_AddressCopy( DTLTransport transport, uint32 size, void * addr )
{
    return InetTransport_AddressCopy(size, addr);
}

static void _DOFInetTransport_AddressDestroy( DTLTransport transport, uint32 size, const void * addr )
{
    InetTransport_AddressDestroy(size, addr);
}

static InetSocket _CreateSocket(const struct sockaddr   * saddr,
                                     DOFConnectionType   type)
{
    if(type == DOFCONNECTIONTYPE_STREAM)
    {
        return InetSocket_Create( saddr->sa_family, SOCK_STREAM, PROTO_TCP );
    }
    else if(type == DOFCONNECTIONTYPE_DATAGRAM)
    {
        return InetSocket_Create( saddr->sa_family, SOCK_DGRAM, PROTO_UDP );
    }
    PCR_Assert(FALSE);
    return PCRSOCKET_INVALID;
}

void Inet_Connectionless_SendTo( int sock, DOFAddress address, uint32 size, const uint8 * data )
{
    const struct sockaddr * sockaddr = InetTransport_GetAddress( address );
    uint32 sockaddrSize = InetTransport_GetAddressSize( address );
    uint32 iface = InetTransport_GetAddressInterface( address );

    if ( _DOFInetTransport_GetAddressType( DOFAddress_GetTransport( address ), address) != DOFADDRESSTYPE_UNICAST )
    {
        switch (sockaddr->sa_family)
        {
#if HAVE_DECL_PF_INET6
            case PF_INET6:
                (void) setsockopt((sock), (int)IPPROTO_IPV6, IP_MULTICAST_IF, &iface, sizeof(iface));
                break;
#endif

#if HAVE_DECL_PF_INET
            case PF_INET:
#ifdef HAVE_STRUCT_IP_MREQN /* mreqn has the interface id which we can use, but mreq doesn't, so we're out of luck if that's all we have. */
                {
                    struct ip_mreqn ipmreq;
                    ipmreq.imr_ifindex = iface;
                    ipmreq.imr_address.s_addr = INADDR_ANY;
                    (void)PCRMem_Copy(&ipmreq.imr_multiaddr.s_addr, &((const struct sockaddr_in*)sockaddr)->sin_addr.s_addr, 4);
                    (void) setsockopt((sock), (int)IPPROTO_IP, IP_MULTICAST_IF, &ipmreq, sizeof(ipmreq));
                }
#endif
                break;
#endif

            default:
                break;
        }
    }

    /* Ignore send errors on datagram sockets. */
    (void) InetSocket_SendTo( sock, data, size, 0, sockaddr, sockaddrSize);
}
