/*
**  Copyright (c) 2014-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include "config.h"
#include <dof/pcr.h>
#include "socket.h"

/* Standard headers */

#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <time.h>
#include <unistd.h>

#ifdef HAVE_SYS_IOCTL_H
# include <sys/ioctl.h>
#else
# include <stropts.h>
#endif


/** @addtogroup InetSocket
@{
**/

/** Maximum backlog of new connections for a listen operation. **/
#ifndef PCRSOCKET_MAX_LISTEN_BACKLOG
# ifdef SOMAXCONN
#  define PCRSOCKET_MAX_LISTEN_BACKLOG   SOMAXCONN
# else
#  define PCRSOCKET_MAX_LISTEN_BACKLOG   5
# endif
#endif


PCRStatus InetSocket_Init(void)
{
    return (PCRSTATUS_OK);
}


void InetSocket_Shutdown(void)
{
    /* No need for any shutdown logic. */
}


InetSocket InetSocket_Create(int domain, int type, int protocol)
{
    return socket((domain), (type), (protocol));
}


void InetSocket_Destroy(InetSocket s)
{
    close((s));
}


InetSocket InetSocket_Accept(InetSocket s, struct sockaddr *addr, uint32 *addrlen)
{
    return accept((s), (addr), (socklen_t*)(addrlen));
}


int InetSocket_Connect(InetSocket s, const struct sockaddr *addr, uint32 addrlen)
{
    return connect((s), (addr), (socklen_t)(addrlen));
}


int InetSocket_Bind(InetSocket s, const struct sockaddr *addr, uint32 addrlen)
{
    return bind((s), (addr), (socklen_t)(addrlen));
}


int InetSocket_Listen(InetSocket s)
{
    return listen((s), PCRSOCKET_MAX_LISTEN_BACKLOG);
}


uint32 InetSocket_Send(InetSocket s, const void *buf, uint32 len, int flags)
{
    return send((s), (buf), (len), (flags));
}


uint32 InetSocket_SendTo(InetSocket s, const void *buf, uint32 len, int flags, const struct sockaddr *addr, uint32 addrlen)
{
    return sendto((s), (buf), (len), (flags), (addr), (socklen_t)(addrlen));
}


uint32 InetSocket_Recv(InetSocket s, void *buf, uint32 len, int flags)
{
    return recv((s), (buf), (len), (flags));
}


uint32 InetSocket_RecvFrom(InetSocket s, void *buf, uint32 len, int flags, struct sockaddr *addr, uint32 *addrlen)
{
    return recvfrom((s), (buf), (len), (flags), (addr), (socklen_t*)(addrlen));
}


void InetSocket_SetOptionNoBlock  (InetSocket s)
{
    int flags = fcntl(s, F_GETFL, 0);
    fcntl(s, F_SETFL, flags|O_NONBLOCK);
}


void InetSocket_SetOptionBlock  (InetSocket s)
{
    int flags = fcntl(s, F_GETFL, 0);
    fcntl(s, F_SETFL, flags&~O_NONBLOCK);
}


void InetSocket_SetOptionNoDelay  (InetSocket s)
{
    int  opt = TRUE;
    (void) setsockopt((s), (int)IPPROTO_TCP,  TCP_NODELAY,  &opt, sizeof(opt));
}


void InetSocket_SetOptionBroadcast(InetSocket s)
{
    int  opt = TRUE;
    (void) setsockopt((s), (int)SOL_SOCKET,   SO_BROADCAST, &opt, sizeof(opt));
}


void InetSocket_SetOptionReuseAddr(InetSocket s)
{
    int  opt = TRUE;

#ifdef _DARWIN_C_SOURCE
    (void) setsockopt((s), (int)SOL_SOCKET, SO_REUSEPORT, &opt, sizeof(opt));
#else
    (void) setsockopt((s), (int)SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
#endif
}


int InetSocket_MulticastJoin(InetSocket s, const struct sockaddr *addr, uint32 addrlen, uint32 iface)
{
    switch (addr->sa_family)
    {
        case AF_INET6:
        {
            int  opt = 64;
            struct ipv6_mreq ipmreq;

            (void)setsockopt(s, (int)IPPROTO_IPV6, IPV6_MULTICAST_HOPS, (char*)&opt, sizeof(opt));
            ipmreq.ipv6mr_interface = iface;
            (void)PCRMem_Copy(&ipmreq.ipv6mr_multiaddr, &((const struct sockaddr_in6*)addr)->sin6_addr.s6_addr, 16);
            (void)setsockopt(s, (int)IPPROTO_IPV6,
#ifdef IPV6_ADD_MEMBERSHIP
                             IPV6_ADD_MEMBERSHIP,
#else
                             IPV6_JOIN_GROUP,
#endif
                             (char*)&ipmreq, sizeof(ipmreq));
            return 0;
        }

#if ( defined( HAVE_STRUCT_IP_MREQN ) || defined( HAVE_STRUCT_IP_MREQ ) )
        case AF_INET:
        {
            int  opt = 64;
#ifdef HAVE_STRUCT_IP_MREQN
            struct ip_mreqn ipmreq;
            ipmreq.imr_ifindex = iface;
            ipmreq.imr_address.s_addr = INADDR_ANY;
#else
            struct ip_mreq ipmreq;
            ipmreq.imr_interface.s_addr = INADDR_ANY;
#endif
            (void)setsockopt(s, (int)IPPROTO_IP,   IP_MULTICAST_TTL,    (char*)&opt, sizeof(opt));
            (void)PCRMem_Copy(&ipmreq.imr_multiaddr.s_addr, &((const struct sockaddr_in*)addr)->sin_addr.s_addr, 4);
            (void)setsockopt(s, (int)IPPROTO_IP, IP_ADD_MEMBERSHIP, (char*)&ipmreq, sizeof(ipmreq));
            return 0;
        }
#endif

        default:
            break;
    }
    return -1;
}


int InetSocket_MulticastLeave(InetSocket s, const struct sockaddr *addr, uint32 addrlen, uint32 iface)
{
    switch (addr->sa_family)
    {
        case AF_INET6:
        {
            int  opt = 64;
            struct ipv6_mreq ipmreq;

            (void)setsockopt(s, (int)IPPROTO_IPV6, IPV6_MULTICAST_HOPS, (char*)&opt, sizeof(opt));
            ipmreq.ipv6mr_interface = iface;
            (void)PCRMem_Copy(&ipmreq.ipv6mr_multiaddr, &((const struct sockaddr_in6*)addr)->sin6_addr.s6_addr, 16);
            (void)setsockopt(s, (int)IPPROTO_IPV6,
#ifdef IPV6_DROP_MEMBERSHIP
                             IPV6_DROP_MEMBERSHIP,
#else
                             IPV6_LEAVE_GROUP,
#endif
                             (char*)&ipmreq, sizeof(ipmreq));
            return 0;
        }

        case AF_INET:
        {
            int  opt = 64;
#ifdef HAVE_STRUCT_IP_MREQN
            struct ip_mreqn ipmreq;
            ipmreq.imr_ifindex = iface;
            ipmreq.imr_address.s_addr = INADDR_ANY;
#else
            struct ip_mreq ipmreq;
            ipmreq.imr_interface.s_addr = INADDR_ANY;
#endif
            (void)setsockopt(s, (int)IPPROTO_IP,   IP_MULTICAST_TTL,    (char*)&opt, sizeof(opt));
            (void)PCRMem_Copy(&ipmreq.imr_multiaddr.s_addr, &((const struct sockaddr_in*)addr)->sin_addr.s_addr, 4);
            (void)setsockopt(s, (int)IPPROTO_IP, IP_DROP_MEMBERSHIP, (char*)&ipmreq, sizeof(ipmreq));
            return 0;
        }

        default:
            break;
    }
    return -1;
}


int InetSocket_CheckSend(InetSocket s)
{
    struct timeval timeout;
    fd_set         fdset;

    FD_ZERO(&fdset);
    FD_SET(s, &fdset);
    timeout.tv_sec  = 0;
    timeout.tv_usec = 0;

    /* Select is used to perform a non-blocking write */
    return select((int)(s+1), 0, &fdset, 0, &timeout);
}


int InetSocket_CheckRecv(InetSocket s)
{
    struct timeval timeout;
    fd_set         fdset;

    FD_ZERO(&fdset);
    FD_SET(s, &fdset);
    timeout.tv_sec  = 0;
    timeout.tv_usec = 0;

    /* Select is used to perform a non-blocking read */
    return select((int)(s+1), &fdset, 0, 0, &timeout);
}


uint32 InetSocket_GetRecvAvail(InetSocket s)
{
    int32 dataSize;

    if ( ioctl(s, FIONREAD, (int32*)&dataSize) != 0 )
        dataSize = 0;
    return dataSize;
}


int InetSocket_Select(int nfds, fd_set *readset, fd_set *writeset, fd_set *exceptset, uint32 timeout)
{
    struct timeval tout;
    uint32 start = PCRTime_GetMS();
    uint32 elapsed = 0;
    int result;

    /* Try to ignore signal interruptions on the select call. */
    do
    {
        if (timeout < PCRTIMEOUT_FOREVER)
        {
            tout.tv_sec  = timeout / 1000;
            tout.tv_usec = (timeout % 1000) * 1000;
            result = select(nfds, readset, writeset, exceptset, &tout);

            elapsed = PCRTime_GetMS() - start;
            if ( elapsed >= timeout )
                break;
            timeout -= elapsed;
            start += elapsed;
        }
        else
        {
            result = select(nfds, readset, writeset, exceptset, NULL);
        }
    } while (result < 0 && errno == EINTR);
    return(result);
}


/**
@}
**/
