/*
**  Copyright (c) 2014-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_PCRSOCKET_H
#define DOF_PCRSOCKET_H

#include "config.h"
#include <dof/pcrtypes.h>

#include <arpa/inet.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <unistd.h>

#ifdef HAVE_SYS_IOCTL_H
# include <sys/ioctl.h>
#else
# include <stropts.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif


/** @defgroup InetSocketTypes Standard Socket Types and Definitions
    Common core basic socket types and definitions upon which the default Inet Transport Library is built.

    @ingroup PCR
@{
**/


/** Invalid socket constant. **/
#define PCRSOCKET_INVALID   ((InetSocket)-1L)


/** Representation of a socket. **/
typedef int32 InetSocket;


/**
@}
**/


/** @defgroup InetSocket InetSocket
    @ingroup PCR

    The PCR Socket layer is an optional component of the PCR.  It is not used
    directly by the OAL.  Rather, it is used by the standard Inet Transport
    Library.  If a platform does not support Berkeley-style sockets, or
    applications on the platform will not use the standard Inet Transport,
    PCR Socket support is not needed.

@{
**/


/** @file
    Header file for the Portable C Runtime socket functions.
    @ingroup PCR
**/


/**
    Initialize the socket library.

    @return Reports whether the initialization was completed successfully or not.

    @retval PCRSTATUS_OK Initialization was successful.
**/
extern PCRStatus InetSocket_Init( void );


/**
    Shutdown the socket library.
**/
extern void InetSocket_Shutdown( void );


/**
    Create a new socket.

    This is equivalent to the standard BSD socket() function.

    @param  domain      The domain of the socket being created.  This is sometimes known as the protocol family or address family.
    @param  type        The type of socket to create:
                        - SOCK_STREAM
                        - SOCK_DGRAM
                        - SOCK_RAW
    @param  protocol    The protocol number or constant to be used within the domain.  This may be 0 if there is only a single protocol of the given type in the domain.

    @return A newly created socket, or the special value @ref PCRSOCKET_INVALID on failure.
**/
extern InetSocket InetSocket_Create( int domain, int type, int protocol );


/**
    Destroy a socket.

    @param  s   The socket to destroy.
**/
extern void InetSocket_Destroy( InetSocket s );


/**
    Accept a new, incoming connection on a server socket.

    This is equivalent to the standard BSD accept() function.

    @note   This function blocks until an incoming connection is made.  Even if @ref InetSocket_Select indicates the listening socket is readable, this may still block in rare circumstances.  To prevent an unbounded block, it is recommended to make the listening socket non-blocking before accepting.

    @param  s       Listening socket with a new connection avaialable.
    @param  addr    [OUT] Source address of connecting host.
    @param  addrlen [IN/OUT] Size of memory pointed to by @a addr on input, size of actual addr structure on output.

    @return The new socket representing the new connection.
**/
extern InetSocket InetSocket_Accept( InetSocket s, struct sockaddr * addr, uint32 * addrlen );


/**
    Establish a new outbound connection.

    This is equivalent to the standard BSD connect() function.

    @note   This function blocks until the connection is made, times out, or fails.  The total time that this may take is platform- and transport-dependent, but may be significant.  To prevent an unbounded block, it is recommended to make the socket non-blocking before connecting.

    @param  s       The socket to use for the new connection.
    @param  addr    [IN] Destination address of connection.
    @param  addrlen Size of memory pointed to by @a addr.

    @return If the connection is successful, zero is returned.  Otherwise, a negative value ( < 0 ) is returned.
**/
extern int InetSocket_Connect( InetSocket s, const struct sockaddr * addr, uint32 addrlen );


/**
    Bind a socket to a local address.

    This is equivalent to the standard BSD bind() function.

    @param  s       The socket to bind to an address.
    @param  addr    [IN] Destination address of connection.
    @param  addrlen Size of memory pointed to by @a addr.

    @return If the socket is bound as requested, zero is returned.  Otherwise, a negative value ( < 0 ) is returned.
**/
extern int InetSocket_Bind( InetSocket s, const struct sockaddr * addr, uint32 addrlen );


/**
    Start listening for incoming connections using the given server socket.

    This is equivalent to the standard BSD listen() function, with the 2nd (backlog) parameter set to an appropriate value for the platform.

    @note The server socket must already be bound to a local address.

    @param  s       The server socket on which to listen.

    @return If the listen is started, zero is returned.  Otherwise, a negative value ( < 0 ) is returned.
**/
extern int InetSocket_Listen( InetSocket s );


/**
    Send data to another host.

    This is equivalent to the standard BSD send() function.

    @note Connection-based sockets must have a valid, active connection to send data.  Connectionless sockets must know the destination address.

    @param  s       The socket over which to send the data.
    @param  buf     [IN] Pointer to the start of the data to send.
    @param  len     Size of data to be sent.
    @param  flags   Flags, if any, or zero (0).

    @return On success, the actual number of bytes sent is returned (which may not always be equal to the requested size).  On failure, @c 0xFFFFFFFF is returned.
**/
extern uint32 InetSocket_Send( InetSocket s, const void * buf, uint32 len, int flags );


/**
    Send data to a non-connected host.

    This is equivalent to the standard BSD sendto() function.

    @note This is not valid for sockets that are connected via InetSocket_Connect().

    @param  s       The socket over which to send the data.
    @param  buf     [IN] Pointer to the start of the data to send.
    @param  len     Size of data to be sent.
    @param  flags   Flags, if any, or zero (0).
    @param  addr    [IN] Destination address to which the data is to be sent.
    @param  addrlen Size of memory pointed to by @a addr.

    @return On success, the actual number of bytes sent is returned (which may not always be equal to the requested size).  On failure, @c 0xFFFFFFFF is returned.
**/
extern uint32 InetSocket_SendTo( InetSocket s, const void * buf, uint32 len, int flags, const struct sockaddr * addr, uint32 addrlen );


/**
    Receive data from another host.

    This is equivalent to the standard BSD recv() function.

    @note   If there is no data to be read, this function blocks until at least one byte is availabe.  The buffer is filled with any data that is available, up to @a len.

    @note   With datagram sockets, if the buffer does not have enough space to store the entire packet, the remaining bytes are dropped.

    @param  s       The socket over which to receive the data.
    @param  buf     [OUT] Pointer to the start of the data buffer where data is to be placed.
    @param  len     Maximum size of data that can be placed in the buffer.
    @param  flags   Flags, if any, or zero (0).

    @return On success, the actual number of bytes received is returned.  If the connection has been closed, zero (0) is returned.  On failure, @c 0xFFFFFFFF is returned.
**/
extern uint32 InetSocket_Recv( InetSocket s, void * buf, uint32 len, int flags );


/**
    Receive data from a non-connected host.

    This is equivalent to the standard BSD recvfrom() function.

    @note   If there is no data to be read, this function blocks until a packet is received.  The buffer is filled with any data that is available, up to @a len.

    @param  s       The socket over which to receive the data.
    @param  buf     [OUT] Pointer to the start of the data buffer where data is to be placed.
    @param  len     Maximum size of data that can be placed in the buffer.
    @param  flags   Flags, if any, or zero (0).
    @param  addr    [IN] Destination address to which the data is to be sent.
    @param  addrlen Size of memory pointed to by @a addr.

    @return On success, the actual number of bytes received is returned.  If the connection has been closed, zero (0) is returned.  On failure, @c 0xFFFFFFFF is returned.
**/
extern uint32 InetSocket_RecvFrom( InetSocket s, void * buf, uint32 len, int flags, struct sockaddr * addr, uint32 * addrlen );


/**
    Set the non-blocking socket option for a socket.

    This is equivalent to enabling the standard non-blocking control (FIONBIO) via ioctl().

    Operations on the socket that would normally block will instead return immediately with an error.

    @param  s       The socket for which to set the option.
**/
extern void InetSocket_SetOptionNoBlock( InetSocket s );


/**
    Set the blocking socket option for a socket.

    This is equivalent to disabling the standard non-blocking control (FIONBIO) via ioctl().

    Operations on the socket may subsequently block.

    @param  s       The socket for which to set the option.
**/
extern void InetSocket_SetOptionBlock( InetSocket s );


/**
    Set the TCP no-delay socket option for a socket.

    This is equivalent to enabling the standard TCP_NODELAY socket option via setsockopt().

    This option disables traffic optimization delays in TCP in preference of lower latency.

    @param  s       The socket for which to set the option.
**/
extern void InetSocket_SetOptionNoDelay( InetSocket s );


/**
    Set the broadcast socket option for a socket.

    This is equivalent to enabling the standard SO_BROADCAST socket option via setsockopt().

    Broadcast is used to be able to send/receive broadcast packets using a datagram socket.  Normally, broadcast messages are blocked.

    @param  s       The socket for which to set the option.
**/
extern void InetSocket_SetOptionBroadcast( InetSocket s );


/**
    Set the address reuse socket option for a socket.

    This is equivalent to enabling the standard SO_REUSEADDR socket option via setsockopt().

    This allows multiple datagram sockets to be bound to the same local address.  This is useful for multicast/broadcast reception in multitasking operating systems.

    @param  s       The socket for which to set the option.
**/
extern void InetSocket_SetOptionReuseAddr( InetSocket s );


/**
    Check for the ability to send on a socket.

    @param  s   The socket to be queried.

    @return The return value is platform-dependent, but will be one of three cases:
            - Zero ( == 0 ) if a send operation would block on the socket.
            - Negative ( < 0 ) if the socket has been closed or is invalid.
            - Positive ( > 0 ) if a send operation would not block.
**/
extern int InetSocket_CheckSend( InetSocket s );

/**
    Check for the ability to receive on a socket.

    @param  s   The socket to be queried.

    @return The return value is platform-dependent, but will be one of three cases:
            - Zero ( == 0 ) if a recv operation would block on the socket.
            - Negative ( < 0 ) if the socket has been closed or is invalid.
            - Positive ( > 0 ) if a recv operation would not block.
**/
extern int InetSocket_CheckRecv( InetSocket s );

/**
    Determine the amount of data available to be read on a socket.

    @param  s   The socket to be queried.

    @return Returns the maximum number of bytes available to be read.  The
            actual value may be platform- or transport-specific, and may
            indicate a larger block of data than can be read in a single recv
            function call.  However, if there is no data to be read, this will
            return 0.  @c 0xFFFFFFFF indicates an error.
**/
extern uint32 InetSocket_GetRecvAvail( InetSocket s );

/**
    Join a multicast group.

    Multicast packets will not be received without first joining the group to which they are sent.

    @param  s       The server socket for which to join the group.
    @param  addr    [IN] Address of the multicast group.
    @param  addrlen Size of memory pointed to by @a addr.
    @param  iface   The identifier indicating on which interface to join the group.  0 indicates "any" (the platform with choose an appropriate interface).

    @return On success, zero is returned.  Otherwise, a negative value ( < 0 ) is returned.
**/
extern int InetSocket_MulticastJoin( InetSocket s, const struct sockaddr * addr, uint32 addrlen, uint32 iface );

/**
    Leave a multicast group.

    Multicast packets will no longer be received for the group.

    @param  s       The server socket for which to leave the group.
    @param  addr    [IN] Address of the multicast group.
    @param  addrlen Size of memory pointed to by @a addr.
    @param  iface   The identifier indicating on which interface to leave the group.  0 indicates "any" (the platform with choose an appropriate interface).

    @return On success, zero is returned.  Otherwise, a negative value ( < 0 ) is returned.
**/
extern int InetSocket_MulticastLeave( InetSocket s, const struct sockaddr * addr, uint32 addrlen, uint32 iface );

/**
    Determine which, if any, sets of sockets are readable, writable, or have exceptions.

    Corresponds to the standard select call.

    @param  nfds        Maximum socket identifier plus 1.
    @param  readset     [IN/OUT] Set of socket descriptors to check for readability.  On return, the sockets that are readable.
    @param  writeset    [IN/OUT] Set of socket descriptors to check for writeability.  On return, the sockets that are writable.
    @param  exceptset   [IN/OUT] Set of socket descriptors to check for exceptions.  On return the sockets that have exceptions.
    @param  timeout     Timeout value, in milliseconds, or PCRTIMEOUT_FOREVER.

    @return Returns the total number of sockets that are ready in the sets, or 0 if the timeout was reached first.  If there is an error, a negative value ( < 0 ) is returned.
**/
extern int InetSocket_Select( int nfds, fd_set * readset, fd_set * writeset, fd_set * exceptset, uint32 timeout );


/**
@}
**/


#ifdef __cplusplus
}
#endif  /* extern "C" */


/**
@}
**/

#endif /* DOF_PCRSOCKET_H */
