/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_ADDRESS_H
#define DOF_ADDRESS_H

#include <dof/core.h>
#include <dof/transport.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DOFAddress DOFAddress
    @ingroup DOF_Framework

    Represents a network address associated with a transport.
@{
**/

/*
**  "Public" Type and Method Declarations
*/

/* DOFAddressMethods */
/**
    Create a DOFAddress from a transport address and the Transport.

    Addresses are unique to a Transport, and should only be created by a
    Transport.  They are opaque to the DOFAddress, and no assumptions can or
    should be made about their type or contents, except by their associated
    Transport.

    @mustfree{DOFAddress_Destroy}

    @param  addrSize  Size of the address data.
    @param  addr      The address data.  This may be NULL if @a addrSize is 0.  This data is copied into and managed by
                      the DOFAddress.  If the address contains data requiring cleanup, the destroy function must be
                      called.
    @param  transport The transport associated with the address.  This must not be NULL.

    @return The DOFAddress, or NULL if one could not be created, or invalid parameters were used.
**/
extern DOF_OAL_DECLSPEC DOFAddress DOFAddress_Create( uint32 addrSize, const void * addr, DTLTransport transport);

/**
 *  Create a new copy of an existing @ref DOFAddress.
 *  @mustfree{DOFAddress_Destroy}
 *  @param  self The @ref DOFAddress to be copied.  This should not be NULL.
 *  @return A new copy of the @ref DOFAddress.
 *  @exception  NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DOFAddress DOFAddress_Copy( const DOFAddress self );

/**
    Compare an address against another for equality.

    Refer to @ref DOFAddress_Compare for a description how addresses are compared.

    @note If both parameters are NULL, they are considered equal. If only one is NULL, they are considered not equal.

    @param self The address to be compared.
    @param addr The address with which to compare.

    @retval TRUE The addresses are equal.
    @retval FALSE The addresses are not equal.

    @see DOFAddress_Compare
**/
extern DOF_OAL_DECLSPEC boolean DOFAddress_IsEqual( const DOFAddress self, const DOFAddress addr );

/**
    Compare an address against another.

    Address comparison is determined first by comparing the associated Transport,
    then the encapsulated data size, and finally the data itself.  The Transport
    is responsible for being able to correctly compare its own address data.

    It is generally unwise to compare addresses associated with different Transports.  Nevertheless, the behavior is deterministic and can be used for sorting.

    @note If both parameters are NULL, they are considered equal. If only one is NULL, it is considered lesser than the non-NULL address.

    @param self The address to be compared.
    @param addr The address with which to compare.

    @return Indicates how the two transport addresses compare to each other.  The return value is a standard integer, and its exact value is irrelevant (and may be platform dependent).  It is intended to be used only in comparison, as follows:
            - Zero ( == 0 ) if addr1 == addr2
            - Positive ( > 0 ) if addr1 > addr2
            - Negative ( < 0 ) if addr1 < addr2

    @see DOFAddress_IsEqual
**/
extern DOF_OAL_DECLSPEC int DOFAddress_Compare( const DOFAddress self, const DOFAddress addr );

/**
    Destroy an address.

    @param  self The address to be destroyed.  If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFAddress_Destroy( DOFAddress self );

/**
    Get the transport-specific address data stored within the
    DOFAddress.

    @note This data is owned by the DOFAddress and must be copied if it is
    needed outside the scope of the DOFAddress.  It is not valid after the
    DOFAddress is destroyed.

    @param  self The address.  This must not be NULL.

    @return The transport-specific address data stored within the DOFAddress.

    @exception  NULL is returned if @a self is NULL.
**/
extern DOF_OAL_DECLSPEC const void * DOFAddress_GetAddress( const DOFAddress self );

/**
    Get the type of address stored within the DOFAddress.

    @param  self The address.  This must not be NULL.

    @return The type of address stored within the DOFAddress.
**/
extern DOF_OAL_DECLSPEC DOFAddressType DOFAddress_GetAddressType( const DOFAddress self );

/**
    Get the size of the transport-specific address data stored
    within the DOFAddress.

    @param  self The address.  This must not be NULL.

    @return The size of the transport-specific address data stored within the DOFAddress.

    @exception  0 is returned if @a self is NULL.
**/
extern DOF_OAL_DECLSPEC uint32 DOFAddress_GetAddressSize( const DOFAddress self );

/**
    Get the transport associated with the DOFAddress.

    @param  self The address.  This must not be NULL.

    @return The transport associated with the DOFAddress.

    @exception  NULL is returned if @a self is NULL.
**/
extern DOF_OAL_DECLSPEC DTLTransport DOFAddress_GetTransport( const DOFAddress self );

/**
@}
**/

/** @defgroup DOFGroupAddress DOFGroupAddress
    @ingroup DOF_Framework

    This class represents group addresses that are used to join secure
    multicast groups.  The configuration for connections of type @ref
    DOFCONNECTIONTYPE_GROUP must use this type of address.

@{
**/

/**
 * Create a new group address from the object identifier that identifies the Secure Multicast Group.
 *
 * @mustfree{DOFAddress_Destroy}
 *
 * @param groupID The group identifier.  This should not be NULL.
 *
 * @return A new DOFGroupAddress.
 *
 * @exception NULL is returned if the address could not be allocated or @a groupOID is NULL.
 */
extern DOF_OAL_DECLSPEC DOFAddress DOFGroupAddress_Create( DOFObjectIDAuthentication groupID );

/**
 * Return the group identifier associated with the address.
 * @note The group Object ID is owned by the DOFAddress and is only valid while the DOFAddress remains valid.  It must not be freed individually.
 * @param self The DOFGroupAddress.  This must not be NULL, and should be a valid DOFGroupAddress.
 * @return The group identifier.
 */
extern DOF_OAL_DECLSPEC DOFObjectIDAuthentication DOFGroupAddress_GetGroupID( const DOFAddress self );

/**
    Destroy the given DOFGroupAddress.

    @param  self  The group address to destroy.  If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFGroupAddress_Destroy( DOFAddress self );

/**
@}
**/


#ifdef  __cplusplus
}
#endif

#endif /* DOF_ADDRESS_H */
