/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_AUTHENTICATOR_H
#define DOF_AUTHENTICATOR_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif


/** @defgroup DOFAuthenticator DOFAuthenticator
    @ingroup DOF_Framework

    This class represents security support in the DOF.  Security in the DOF is
    tied to a domain, which consists of at least a domain identifier and some
    credentials.  This class also manages callbacks represents a domains storage.

    Without at least a single domain or authenticator, a DOF cannot accept secure connections.

    Each Authenticator that is created must be destroyed using @ref DOFAuthenticator_Destroy.
@{
**/

/**
 * Destroy a DOFAuthenticator.
 *
 * @param self The DOFAuthenticator.  If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFAuthenticator_Destroy( DOFAuthenticator self );

/**
 * Create a authenticator state from the given authenticator.
 *
 * @mustfree{DOFAuthenticatorState_Destroy}
 *
 * @param self The DOFAuthenticator.  This must not be NULL. 
 *  
 * @return A copy of the state for the given DOFAuthenticator.
 */
extern DOF_OAL_DECLSPEC DOFAuthenticatorState DOFAuthenticator_CreateState( const DOFAuthenticator self );

/**
 * Generates a new, random, secure session key.
 *
 * @mustfree{PCRMem_Free}
 *
 * @return A new 32-byte session key.
 * @exception NULL is returned if there are insufficient resources to allocate a key.
 */
extern DOF_OAL_DECLSPEC uint8 * DOFAuthenticator_CreateSessionKey( void );

/**
 * Destroy a DOFAuthenticatorList
 *
 * @param self The DOFAuthenticatorList. If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFAuthenticatorList_Destroy( DOFAuthenticatorList self );


/**
 * The type of security associated with a security request.
 *
 * @see DOFAuthenticatorStorageFns_t.GetNode
 */
typedef enum DOFAuthenticatorSecurityType_t
{
    /**
    * The type of security is undefined.
    */
    DOFAUTHENTICATORSECURITYTYPE_UNDEFINED,
    /**
     * The security type is point-to-point.
     */
    DOFAUTHENTICATORSECURITYTYPE_TWO_POINT,
    /**
    * The security type is multi-point, which means multiple nodes have the same security that is limited to the security of the group.
    */
    DOFAUTHENTICATORSECURITYTYPE_MULTI_POINT,
    /**
     * The security type may be point-to-point or multi-point.
     */
    DOFAUTHENTICATORSECURITYTYPE_ANY
} DOFAuthenticatorSecurityType;


/**
@}
**/

/** @defgroup DOFAuthenticatorNode DOFAuthenticatorNode
    @ingroup DOFAuthenticator

    This class represents a authenticator node, an authenticating entity with assigned credentials and permissions.

@{
**/

/**
 * DOFAuthenticatorNodePermissionSetMapElement contains the set of permissions
 * associated with a Permission Group ID.  This structure is used with a
 * @ref DOFAuthenticatorNodePermissionSetMap.
 *
 * @ingroup DOFAuthenticator
 */
typedef struct DOFAuthenticatorNodePermissionSetMapElement_t
{
    uint32  id;                         /**< The Permission Group ID for the permissions. **/
    DOFPermissionSet permissionset;     /**< The set of permissions (grants) for the node within this Permission Group ID. This must not be NULL. **/
} DOFAuthenticatorNodePermissionSetMapElement;


typedef struct DOFAuthenticatorNodePermissionSetMap_t * DOFAuthenticatorNodePermissionSetMap; /**< Pointer to a struct DOFAuthenticatorNodePermissionSetMap_t used with DOFAuthenticatorNode **/

/**
 * DOFAuthenticatorNodePermissionSetMap_t represents the permissions for a node.
 * This may be a subset or the complete set, as needed, based on the request to the DOFAuthenticatorStorage API.
 * @note The list of elements must be ordered by the Security Scope of each element in ascending order.
 *
 * @ingroup DOFAuthenticator
 */
struct DOFAuthenticatorNodePermissionSetMap_t
{
    uint32  count;                                       /**< The count of the elements in the list below **/
    DOFAuthenticatorNodePermissionSetMapElement* list;   /**< An array of @ref DOFAuthenticatorNodePermissionSetMapElement.  Refer to the structure for more information. This may be NULL if count is 0 otherwise this must not be NULL. **/
};


/**
    The DOFAuthenticatorNode interface.  Defined as a pointer to the @ref DOFAuthenticatorNode_t structure.  Refer to the structure for more information.
**/
typedef struct DOFAuthenticatorNode_t* DOFAuthenticatorNode;

/**
 * The DOFAuthenticatorNode interface structure. Represents a node in the authentication server storage.
 *
 * This structure should be placed as the first element in a larger structure
 * representing any necessary state of the listener.  All of the functions
 * take as their first argument a pointer to this structure (a @ref
 * DOFAuthenticatorNode).  Inside each function, it can thus be type-cast as a pointer
 * to the larger structure to access the listener's state.  If no additional
 * state is required, this structure may be used directly.
 */
typedef struct DOFAuthenticatorNode_t
{

    /**
     * The assigner interface function structure, containing a list of
     * function pointers that implement the interface.  It has the same role as
     * a vtable in C++.  This must not be NULL.
     */

    const struct DOFAuthenticatorNodeFns_t
    {

        /**
         * Notifies the owner of this node that the node is no longer needed.
         * No additional calls will be made for this node.
         * Typically, this would free the node's resources.
         * The action taken is up to the storage implementation.
         *
         * @param self The DOFAuthenticatorNode object. This must not be NULL.
         */
        void (*Release)( DOFAuthenticatorNode self );

        /**
         * Gets this node's identifier.
         *
         * @param self The DOFAuthenticatorNode object. This must not be NULL.
         *
         * @return The Authentication Identity of the node.
         */
        DOFObjectIDAuthentication (*GetNodeID)( DOFAuthenticatorNode self );

        /**
         * Gets this node's unique, numeric source identifier.
         * Every node must have a unique numerical identifier, which can be used in some circumstances as a shortened alias for the node's authentication identity.
         * This value is combined with other unique information to produce a DOFObjectIDSource for use by the node, if needed.
         *
         * @param self The DOFAuthenticatorNode object. This must not be NULL.
         *
         * @return The unique, numeric source identity of the node.
         */
        uint32 (*GetSourceIdentifier)( DOFAuthenticatorNode self );

        /**
         * Gets the credential storage for this node.
         *
         * @param self The DOFAuthenticatorNode object. This must not be NULL.
         *
         * @return The private credential storage data for the node.
         */
        const uint8 * (*GetCredentialsStorage)( DOFAuthenticatorNode self );

        /**
         * Gets the byte length of the credentials store data for this node.
         *
         * @param self The DOFAuthenticatorNode object. This must not be NULL.
         *
         * @return The size of the credential storage data for the node, in bytes.
         */
        uint32 (*GetCredentialsStorageSize)( DOFAuthenticatorNode self );

        /**
         * Return whether or not this node is enabled.
         *
         * @param self The DOFAuthenticatorNode object. This must not be NULL.
         *
         * @retval Non-Zero (TRUE) The node is enabled (permitted to authenticate).
         * @retval Zero (FALSE) The node is disabled and must not be allowed to authenticate.
         */
        boolean (*IsEnabled)( DOFAuthenticatorNode self );

        /**
         * Returns whether or not a node is a member of the given group.
         *
         * @param self The DOFAuthenticatorNode object. This must not be NULL.
         * @param groupID The ID of the group in which to check.
         *
         * @retval Non-Zero (TRUE) The node is in the group.
         * @retval Zero (FALSE) The node is not in the group.
         */
        boolean (*IsMember)( DOFAuthenticatorNode self, DOFObjectIDAuthentication groupID );

        /**
         * Gets the DOFAuthenticatorNodePermissionSetMap for this node.
         *
         * @param self The DOFAuthenticatorNode object. This must not be NULL.
         *
         * @return The @ref DOFAuthenticatorNodePermissionSetMap representing the permission sets for this node and IDs that they are in.
         */
        DOFAuthenticatorNodePermissionSetMap (*GetPermissions)( DOFAuthenticatorNode self );

        /**
         * Get the unique identifier (an alias) for the remote domain.
         *
         * @note This is only called on a remote domain node.
         *
         * @param self The DOFAuthenticatorNode object. This must not be NULL.
         *
         * @return The remote domain identifier.
         */
        uint32 (*GetRemoteDomainIdentifier)( DOFAuthenticatorNode self );

        /**
         * Get the local identity of the node related to the remote domain.
         *
         * @note This is only called on a remote domain node.
         *
         * @param self The DOFAuthenticatorNode object. This must not be NULL.
         *
         * @return The local identity for the node related to the remote domain.
         */
        DOFObjectIDAuthentication (*GetLocalID)( DOFAuthenticatorNode self );

        /**
         * Get the byte length of the Security Mode for this group node.
         *
         * @note This is only called on a group node.
         *
         * @param self The DOFAuthenticatorNode object. This must not be NULL.
         *
         * @return The byte length of the group's Security Mode.
         */
        uint32 (*GetGroupSecurityModeSize)( DOFAuthenticatorNode self );

        /**
         * Retrieves the Security Mode byte array for this group node.
         *
         * @note This is only called on a group node.
         *
         * @param self The DOFAuthenticatorNode object. This must not be NULL.
         *
         * @return The group's Security Mode byte array.
         */
        const uint8 * (*GetGroupSecurityMode)( DOFAuthenticatorNode self );

        /**
         * Gets the Heartbeat timing parameter for this group node.
         *
         * @note This is only called on a group node.
         *
         * @param self The DOFAuthenticatorNode object. This must not be NULL.
         *
         * @return The group's Heartbeat period, in seconds.  Valid range is 0-255.
         */
        uint8 (*GetGroupHeartbeatPeriod)( DOFAuthenticatorNode self );

        /**
         * Gets the Minimum KEK timing parameter for this group node.
         *
         * @note This is only called on a group node.
         *
         * @param self The DOFAuthenticatorNode object. This must not be NULL.
         *
         * @return The group's Minimum KEK period, in minutes.  Valid range is 0-255.
         */
        uint8 (*GetGroupMinKEKPeriod)( DOFAuthenticatorNode self );

        /**
         * Gets the Maximum KEK timing parameter for this group node.
         *
         * @note This is only called on a group node.
         *
         * @param self The DOFAuthenticatorNode object. This must not be NULL.
         *
         * @return The group's Maximum KEK period, in minutes.  Valid range is 0-2040.  Higher numbers are capped at 2040.
         */
        uint16 (*GetGroupMaxKEKPeriod)( DOFAuthenticatorNode self );

        /**
         * Gets the current Epoch for this group node.
         *
         * @note This is only called on a group node.
         *
         * @param self The DOFAuthenticatorNode object. This must not be NULL.
         *
         * @return The group's current Epoch.
         */
        uint16 (*GetGroupEpoch)( DOFAuthenticatorNode self );

        /**
         * Gets the Security State ID (an alias) for this group node.
         *
         * @note This is only called on a group node.
         *
         * @param self The DOFAuthenticatorNode object. This must not be NULL.
         *
         * @return The group's Security State ID.
         */
        uint32 (*GetGroupStateID)( DOFAuthenticatorNode self );

    } * fns;

    /* Implementations may extend this structure. */

} DOFAuthenticatorNode_t;

/**
@}
**/


/** @defgroup DOFAuthenticatorStorage DOFAuthenticatorStorage
    @ingroup DOFAuthenticator

    This class represents the back-end storage of authenticator nodes.  This represents a read-only interface, which is
    all that is required by a @ref DOFAuthenticator to authenticate nodes.  Management of the back-end storage data is
    left to the implementation of the back-end storage.

@{
**/

/**
    The AuthenticatorStorage interface.  Defined as a pointer to the @ref DOFAuthenticatorStorage_t structure.  Refer to the structure for more information.
**/
typedef struct DOFAuthenticatorStorage_t* DOFAuthenticatorStorage;

/**
 * The AuthenticatorStorage interface structure
 *
 * This structure should be placed as the first element in a larger structure
 * representing any necessary state of the listener.  All of the functions
 * take as their first argument a pointer to this structure (a @ref
 * DOFAuthenticatorStorage).  Inside each function, it can thus be type-cast as a pointer
 * to the larger structure to access the listener's state.  If no additional
 * state is required, this structure may be used directly.
 *
 * @see DOFAuthenticatorConfigBuilder_SetAuthenticatorStorage
 */
typedef struct DOFAuthenticatorStorage_t
{
    /**
     * The assigner interface function structure, containing a list of
     * function pointers that implement the interface.  It has the same role as
     * a vtable in C++.  This must not be NULL.
     */
    const struct DOFAuthenticatorStorageFns_t
    {
        /**
         * Resolve the given credentials by updating the credential with known identity and
         * credential private storage.
         *
         * This must not be NULL.
         *
         * @param self The DOFAuthenticatorStorage object. This must not be NULL.
         * @param credentials The credentials to update. The credentials contain the information necessary for retrieval (based on credential type).
         *
         * @retval DOFERROR_OK The credentials were successfully updated.
         * @return DOFERROR_* (any other defined DOFERROR constant) may be used to indicate failure.
         */
        DOFError (*Resolve)( DOFAuthenticatorStorage self, DOFCredentials credentials );

        /**
         * Get the @ref DOFAuthenticatorNode for the specified @a nodeID.
         *
         * This must not be NULL.
         *
         * @param self The DOFAuthenticatorStorage object. This must not be NULL.
         * @param nodeID The Authentication Identity of the node to be retrieved. This must not be NULL.
         * @param type The type of credentials used in the request.
         * @param securityType The type of security associated with the get node request. May be used for determining which permissions to return with the node.
         * @param securityModeLen The Length in bytes of the securityMode.
         * @param securityMode The security mode in marshaled format, associated with the get node request. May be used for determining which permissions to
         *                     return with the node. This may Be Null.
         * @param permissions The requested permissions. The PermissionSetMap in the returned node need only contain information as it relates to the requested
         *                    permissions (more may be returned, but is not required). This may be NULL or an empty set, in which case no permission information
         *                    needs to be returned.
         * @param status [out] If no node is returned. This needs to hold a TRP error code that covers the reason for NULL node return. This will not be checked
         *                     if a node is returned. This will not be NULL. Please treat all calls as if they are for the initiator node. They will be mapped to the appropriately node.
         *
         * @return The @ref DOFAuthenticatorNode.
         * @exception NULL is returned if the node cannot be found.
         */
        DOFAuthenticatorNode (*GetNode)( DOFAuthenticatorStorage self, DOFObjectIDAuthentication nodeID, DOFCredentialsType type, DOFAuthenticatorSecurityType securityType, uint32 securityModeLen, const uint8* securityMode, DOFPermissionSet permissions, DOFError* status );

        /**
         * Get the @ref DOFAuthenticatorNode for the specified group @a nodeID.
         *
         * This must not be NULL.
         *
         * @param self The DOFAuthenticatorStorage object. This must not be NULL.
         * @param nodeID The Authentication Identity of the group node to be retrieved. This must not be NULL.
         *
         * @return The @ref DOFAuthenticatorNode.
         * @exception NULL is returned if the node cannot be found, or the node is not a group node.
         */
        DOFAuthenticatorNode (*GetGroupNode)( DOFAuthenticatorStorage self, DOFObjectIDAuthentication nodeID );

        /**
         * Get the @ref DOFAuthenticatorNode for the specified remote domain @a nodeID.
         *
         * This must not be NULL.
         *
         * @param self The DOFAuthenticatorStorage object. This must not be NULL.
         * @param domainID The Domain ID of the remote domain node to be retrieved. This must not be NULL.
         * @param initiatorID The Authentication Identity (in the remote domain)
         *                    of the initiator.  This must not be NULL.  The
         *                    resulting node may choose to map the initiator to
         *                    a local Authentication Identity.
         *
         * @return The @ref DOFAuthenticatorNode.
         * @exception NULL is returned if the node cannot be found, or the node is not a domain node.
         */
        DOFAuthenticatorNode (*GetDomainNode)( DOFAuthenticatorStorage self, DOFObjectIDDomain domainID, DOFObjectIDAuthentication initiatorID );

        /**
         * Get the domain id for this storage. This must match the authenticator configuration associated with the storage.
         *
         * This must not be NULL.
         *
         * @param self The DOFAuthenticatorStorage object. This must not be NULL.
         *
         * @return The Domain ID of the storage.  This must not be NULL.
         */
        DOFObjectIDDomain (*GetStorageDomainID)( DOFAuthenticatorStorage self );

        /**
         * Report whether or not the storage is currently available.
         *
         * This must not be NULL.
         *
         * @param self The DOFAuthenticatorStorage object. This must not be NULL.
         *
         * @retval Non-zero (TRUE) if the storage is available.
         * @retval Zero (FALSE) if the storage is not available.
         */
        boolean (*IsStorageAvailable)( DOFAuthenticatorStorage self );

    } *fns;

    /* Implementations may extend this structure. */

} DOFAuthenticatorStorage_t;

/**
@}
**/


/** @defgroup DOFAuthenticatorConfig DOFAuthenticatorConfig
    @ingroup DOFAuthenticator

    This class represents an authenticator configuration.

    Each authenticator configuration that is created must be destroyed using @ref DOFAuthenticatorConfig_Destroy.

@{
**/

/** Default maximum grant lifetime.  1440 minutes (24 hours). **/
#define DOFAUTHENTICATORCONFIG_DEFAULT_MAX_GRANT_LIFETIME 1440

/** @name Builder
@{
**/

/**
 * Create a new DOFAuthenticatorConfigBuilder.
 *
 * @mustfree{DOFAuthenticatorConfigBuilder_Destroy}
 *
 * The builder uses the following defaults:
 *  -Domain ID: Given by @a domainID
 *  -Node Identifier: Given by @a id
 *  -Storage: Given by @a storage
 *  -Max Grant Lifetime: @ref DOFAUTHENTICATORCONFIG_DEFAULT_MAX_GRANT_LIFETIME
 *  -Debug: FALSE
 *  -Debug Key: NULL (no key)
 * @param domainID The domain's identity (an Object ID).  This should not be NULL, and should not be the broadcast ID.
 * @param id The domain's node identifier.  Must be unique between each redundant node handling the same domain ID.
 * @param storage The storage for the authenticator.  This should not be NULL. If NULL the config will fail to build, if not set before build is called.
 *
 * @return A new DOFAuthenticatorConfig.
 *
 * @exception NULL is returned if an invalid argument is detected or the DOFAuthenticatorConfig could not be created.
 */
extern DOF_OAL_DECLSPEC DOFAuthenticatorConfigBuilder DOFAuthenticatorConfigBuilder_Create( DOFObjectIDDomain domainID, uint8 id, DOFAuthenticatorStorage storage );

/**
 * Creates a DOFAuthenticatorConfigBuilder from a given state.
 *
 * @mustfree{DOFAuthenticatorConfigBuilder_Destroy or @ref DOFAuthenticatorConfigBuilder_BuildAndDestroy}
 *
 * @param self The state. This must not be NULL.
 *
 * @return DOFAuthenticatorConfigBuilder The connection config builder that represents the given state. This will not have a storage set and will not build until one is set.
 * @exception NULL If the builder failed to be built.
 */
extern DOF_OAL_DECLSPEC DOFAuthenticatorConfigBuilder DOFAuthenticatorConfigBuilder_Create_State( DOFAuthenticatorState self );

/**
 * Creates a DOFAuthenticatorConfigBuilder from a given config.
 *
 * @mustfree{DOFAuthenticatorConfigBuilder_Destroy or @ref DOFAuthenticatorConfigBuilder_BuildAndDestroy}
 *
 * @param self The config. This must not be NULL.
 *
 * @return DOFAuthenticatorConfigBuilder The connection config builder that represents the given config.
 * @exception NULL If the builder failed to be built.
 */
extern DOF_OAL_DECLSPEC DOFAuthenticatorConfigBuilder DOFAuthenticatorConfigBuilder_Create_Config( DOFAuthenticatorConfig self );

/**
 * Sets the Domain ID for the domain.
 *
 * @param self The DOFAuthenticatorConfigBuilder to set the storage of.
 * @param domainID The domain's identity (an Object ID).  If NULL or the broadcast ID, the domainID will be left unchanged.
 *
 * @retval TRUE DomainID was successfully set.
 * @exception FALSE DomainID set failed.
 */
extern DOF_OAL_DECLSPEC boolean DOFAuthenticatorConfigBuilder_SetDomainID( DOFAuthenticatorConfigBuilder self, DOFObjectIDDomain domainID );

/**
 * Sets the Storage for the authenticator.
 *
 * @param self The DOFAuthenticatorConfigBuilder to set the storage of.
 * @param storage The storage for the authenticator.  If NULL there will be no change.
 *
 * @retval TRUE storage was successfully set.
 * @exception FALSE storage failed to be set.
 */
extern DOF_OAL_DECLSPEC boolean DOFAuthenticatorConfigBuilder_SetStorage( DOFAuthenticatorConfigBuilder self, DOFAuthenticatorStorage storage );

/**
 * Sets the maximum grant lifetime for the authenticator.
 * Please refer to @ref DOFAuthenticatorConfigBuilder_Create for the default value for this setting.
 *
 * @param self The DOFAuthenticatorConfigBuilder to set the max lifetime of.
 * @param max The maximum grant that the AS will pass out, in minutes.
 *
 * @retval TRUE Max grant lifetime was set.
 * @exception FALSE Max grant lifetime failed to be set.
 */
extern DOF_OAL_DECLSPEC boolean DOFAuthenticatorConfigBuilder_SetMaxGrantLifetime( DOFAuthenticatorConfigBuilder self, uint16 max );

/**
 * Sets the node identifier for the authenticator.
 * A node that supports multiple authenticators would use the same node identifier for each authenticator.
 * If storage authentication requests for a domain are handled by multiple nodes, then each node must have its own identifier.
 *
 * @param self The DOFAuthenticatorConfigBuilder to set the node identifier of.
 * @param id The node identifier.  Valid range is 0-63.  Invalid values are ignored and do not change the node identifier.
 *
 * @retval TRUE Node identifier was set.
 * @exception FALSE Node identifier was failed to be set.
 */
extern DOF_OAL_DECLSPEC boolean DOFAuthenticatorConfigBuilder_SetNodeIdentifier( DOFAuthenticatorConfigBuilder self, uint8 id );

/**
 * Sets the debug mode session key for the authenticator.
 * When debug mode for the authenticator is enabled, all distributed session keys will be the given key.
 * Even if set, this has no effect unless the debug mode flag is enabled (see @ref DOFAuthenticatorConfigBuilder_SetDebug).
 * However, it is invalid to build a authenticator with debug mode enabled, yet with a NULL debug key.
 * A debug key should @b only be used in a testing environment.
 * Please refer to @ref DOFAuthenticatorConfigBuilder_Create for the default value for this setting.
 *
 * @param self The DOFAuthenticatorConfigBuilder.  This may be NULL, in which case no action is taken.
 * @param key The debug mode session key.  This may be NULL to invalidate debug mode; otherwise, this must be a 32-byte key.
 *
 * @retval TRUE Debug key was set.
 * @exception FALSE Debug key was failed to be set.
 */
extern DOF_OAL_DECLSPEC boolean DOFAuthenticatorConfigBuilder_SetDebugKey( DOFAuthenticatorConfigBuilder self, const uint8 key[32] );

/**
 * Sets the debug mode flag for the authenticator.
 * The default debug mode flag is FALSE (disabled).  It is invalid to build a authenticator with this flag TRUE (enabled) but without a key (see @ref DOFAuthenticatorConfigBuilder_SetDebugKey).
 * A debug key should @b only be used in a testing environment.
 * Please refer to @ref DOFAuthenticatorConfigBuilder_Create for the default value for this setting.
 *
 * @param self The DOFAuthenticatorConfigBuilder.  This may be NULL, in which case no action is taken.
 * @param isDebug If the debug mode flag should be enabled (TRUE) or disabled (FALSE).
 *
 * @retval TRUE Debug was set.
 * @exception FALSE Debug was failed to be set.
 */
extern DOF_OAL_DECLSPEC boolean DOFAuthenticatorConfigBuilder_SetDebug( DOFAuthenticatorConfigBuilder self, boolean isDebug );

/**
 * Creates a DOFAuthenticatorConfig from the builder
 *
 * @param self The DOFAuthenticatorConfigBuilder to build.
 *
 * @return DOFAuthenticatorConfig The newly created DOFAuthenticatorConfig
 *
 * @exception NULL is returned if insufficient resources exist to create the config.
 */
extern DOF_OAL_DECLSPEC DOFAuthenticatorConfig DOFAuthenticatorConfigBuilder_Build( DOFAuthenticatorConfigBuilder self );

/**
 * Creates a DOFAuthenticatorConfig from the builder and destroys the builder passed in.
 *
 * @param self The DOFAuthenticatorConfigBuilder to build and destroy. This should not be NULL. 
 *
 * @return DOFAuthenticatorConfig The newly created DOFAuthenticatorConfig
 * @exception NULL is returned if an error occurs.
 */
extern DOF_OAL_DECLSPEC DOFAuthenticatorConfig DOFAuthenticatorConfigBuilder_BuildAndDestroy( DOFAuthenticatorConfigBuilder self );

/**
 * Destroy a DOFAuthenticatorConfigBuilder.
 *
 * @param self The DOFAuthenticatorConfigBuilder to destroyed. If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFAuthenticatorConfigBuilder_Destroy( DOFAuthenticatorConfigBuilder self );

/**
@}
**/


/**
 * Destroy a DOFAuthenticatorConfig.
 *
 * @param self The DOFAuthenticatorConfig.  If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFAuthenticatorConfig_Destroy( DOFAuthenticatorConfig self );

/**
 * Get the domain ID associated with the authenticator configuration.
 *
 * @param self The Authenticator Config.  This must not be NULL.
 *
 * @return The domain ID.
 */
extern DOF_OAL_DECLSPEC DOFObjectIDDomain DOFAuthenticatorConfig_GetDomainID( const DOFAuthenticatorConfig self );

/**
 * Get the node identifier associated with the authenticator configuration.
 *
 * @param self The Authenticator Config.  This must not be NULL.
 *
 * @return The node identifier.
 */
extern DOF_OAL_DECLSPEC uint8 DOFAuthenticatorConfig_GetNodeIdentifier( const DOFAuthenticatorConfig self );

/**
 * Get the storage associated with the authenticator configuration.
 *
 * @param self The Authenticator Config.  This must not be NULL.
 *
 * @return The storage.
 */
extern DOF_OAL_DECLSPEC DOFAuthenticatorStorage DOFAuthenticatorConfig_GetStorage( const DOFAuthenticatorConfig self );

/**
 * Get the maximum grant lifetime associated with the authenticator configuration.
 *
 * @param self The Authenticator Config.  This must not be NULL.
 *
 * @return The maximum grant lifetime, in minutes.
 */
extern DOF_OAL_DECLSPEC uint16 DOFAuthenticatorConfig_GetMaxGrantLifetime( const DOFAuthenticatorConfig self );

/**
 * Get the debug key associated with the authenticator configuration, if any.
 * The key is always a 256-bit key (32 bytes).
 *
 * @param self The Authenticator Config.  This must not be NULL.
 *
 * @return The debug key, or NULL if a debug key is not set.  This must be non-NULL if @ref DOFAuthenticatorConfig_IsDebug reports TRUE.
 */
extern DOF_OAL_DECLSPEC const uint8 * DOFAuthenticatorConfig_GetDebugKey( const DOFAuthenticatorConfig self );

/**
 * Get the debug mode flag associated with the authenticator configuration.
 *
 * @param self The Authenticator Config.  This must not be NULL.
 *
 * @return The debug mode flag.
 * @retval Non-zero (TRUE) if debug mode is enabled.
 * @retval Zero (FALSE) if debug mode is disabled.
 */
extern DOF_OAL_DECLSPEC boolean DOFAuthenticatorConfig_IsDebug( const DOFAuthenticatorConfig self );

/**
@}
**/

/** @defgroup DOFAuthenticatorState DOFAuthenticatorState
    @ingroup DOFAuthenticator

    Represents the instantaneous state of a DOFAuthenticator.
@{
**/

/**
 * Destroys a DOFAuthenticatorState.
 *
 * @param self The state to destroy. This may be NULL. If NULL, no action will be taken.
 */
extern DOF_OAL_DECLSPEC void DOFAuthenticatorState_Destroy( DOFAuthenticatorState self );

/**
 * Returns if the storage is available.
 *
 * @param self The state to get the value from. This must not be NULL.
 *
 * @return boolean TRUE if the authenticator storage is available.
 */
extern DOF_OAL_DECLSPEC boolean DOFAuthenticatorState_IsAvailable( const DOFAuthenticatorState self );

/**
 * Returns the domain ID.
 *
 * @param self The state to get the value from. This must not be NULL.
 *
 * @return DOFObjectIDDomain The domain ID for the authenticator.
 */
extern DOF_OAL_DECLSPEC DOFObjectIDDomain DOFAuthenticatorState_GetDomainID( const DOFAuthenticatorState self );

/**
 * Returns the max grant lifetime.
 *
 * @param self The state to get the value from. This must not be NULL.
 *
 * @return uint8 The max grant lifetime.
 */
extern DOF_OAL_DECLSPEC uint8 DOFAuthenticatorState_GetMaxGrantLifetime( const DOFAuthenticatorState self );

/**
 * Returns the node id.
 *
 * @param self The state to get the value from. This must not be NULL.
 *
 * @return uint8 The node id
 */
extern DOF_OAL_DECLSPEC uint8 DOFAuthenticatorState_GetNodeIdentifier( const DOFAuthenticatorState self );

/**
 * Returns if the authenticator is in debug mode.
 *
 * @param self The state to get the value from. This must not be NULL.
 *
 * @return boolean TRUE authenticator in debug mode, FALSE otherwise.
 */
extern DOF_OAL_DECLSPEC boolean DOFAuthenticatorState_IsDebug( const DOFAuthenticatorState self );


/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* DOF_DOMAIN_H */
