/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_BRIDGE_H
#define DOF_BRIDGE_H


#include <dof/core.h>


#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DOFOperationBridgeConfig DOFOperationBridgeConfig
    @ingroup DOFOperation

    The DOFOperationBridgeConfig class represents a security bridge configuration.

@{
**/

/**
 * @name Builder
 * @{
 */


/**
 *  Constructs a new DOFOperationBridgeConfigBuilder.  The new DOFOperationBridgeConfigBuilder will have the following configuration parameters:
 *   - credentials: NULL (unsecured)
 *   - permissions: NULL (no initial set)
 *   - permission extension: TRUE (a DOFPermissionSet can be extended as needed)
 *
 *  @mustfree{DOFOperationBridgeConfigBuilder_Destroy or @ref DOFOperationBridgeConfigBuilder_BuildAndDestroy}
 *
 *  @return a new DOFOperationBridgeConfigBuilder
 */
extern DOF_OAL_DECLSPEC DOFOperationBridgeConfigBuilder DOFOperationBridgeConfigBuilder_Create(void);

/**
 * Creates a new DOFOperationBridgeConfigBuilder from a given @ref DOFOperationBridgeConfig. The new DOFOperationBridgeConfigBuilder's initial configuration parameters will match those in the given DOFOperationBridgeConfig.
 *  
 * @mustfree{DOFOperationBridgeConfigBuilder_Destroy or @ref DOFOperationBridgeConfigBuilder_BuildAndDestroy} 
 *  
 * @param config The DOFOperationBridgeConfig whose configuration parameters to copy when creating the new builder. Should not be @c NULL.
 * 
 * @return a new DOFOperationBridgeConfigBuilder, or NULL for error.
 */
extern DOF_OAL_DECLSPEC DOFOperationBridgeConfigBuilder DOFOperationBridgeConfigBuilder_Create_Config( DOFOperationBridgeConfig config );

/**
 *  Copies a DOFOperationBridgeConfigBuilder.
 *
 *  @mustfree{DOFOperationBridgeConfigBuilder_Destroy or @ref DOFOperationBridgeConfigBuilder_BuildAndDestroy}
 *
 *  @param self The DOFOperationBridgeConfig to copy.  Should not be @c NULL.
 *
 *  @return a copy of @a self
 */
extern DOF_OAL_DECLSPEC DOFOperationBridgeConfigBuilder DOFOperationBridgeConfigBuilder_Copy(DOFOperationBridgeConfigBuilder self);

/**
 *  Destroys a DOFOperationBridgeConfigBuilder.
 *
 *  @param self The DOFOperationBridgeConfigBuilder to destroy.  If @c NULL this function will do nothing.
 */
extern DOF_OAL_DECLSPEC void DOFOperationBridgeConfigBuilder_Destroy(DOFOperationBridgeConfigBuilder self);

/**
 *  Builds a DOFOperationBridgeConfig from a DOFOperationBridgeConfigBuilder.
 *
 *  @mustfree{DOFOperationBridgeConfig_Destroy}
 *
 *  @param self The DOFOperationBridgeConfigBuilder to use.  Should not be @c NULL.
 *
 *  @return a DOFOperationBridgeConfig built from @a self
 */
extern DOF_OAL_DECLSPEC DOFOperationBridgeConfig DOFOperationBridgeConfigBuilder_Build(DOFOperationBridgeConfigBuilder self);

/**
 *  Builds a DOFOperationBridgeConfig from a DOFOperationBridgeConfigBuilder.  Regardless of success, the DOFOperationBridgeConfigBuilder will be destroyed.
 *
 *  @mustfree{DOFOperationBridgeConfig_Destroy}
 *
 *  @param self The DOFOperationBridgeConfigBuilder to use.  Should not be @c NULL.
 *
 *  @return a DOFOperationBridgeConfig built from @a self
 */
extern DOF_OAL_DECLSPEC DOFOperationBridgeConfig DOFOperationBridgeConfigBuilder_BuildAndDestroy(DOFOperationBridgeConfigBuilder self);

/**
 *  Sets the DOFCredentials to use when creating a DOFOperationBridgeConfig.
 *
 *  @param self The DOFOperationBridgeConfigBuilder.  Should not be @c NULL.
 *  @param credentials The DOFCredentials to use.  If @c NULL, @a self will construct a non-secure DOFOperationBridgeConfig
 *
 *  @retval TRUE @a credentials was successfully set for @a self
 *  @retval TRUE @a credentials was not successfully set for @a self
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationBridgeConfigBuilder_SetCredentials( DOFOperationBridgeConfigBuilder self, DOFCredentials credentials );

/**
 *  Sets the initial DOFPermissionSet to requset when creating a DOFOperationBridgeConfig.
 *
 *  @param self The DOFOperationBridgeConfigBuilder.  Should not be @c NULL.
 *  @param permissions The DOFPermissionSet to initially request.  If @c NULL, no initial permissions will be requested.
 *
 *  @retval TRUE @a permissions was successfully set as the initial set for @a self
 *  @retval TRUE @a permissions was not successfully set as the initial set for @a self
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationBridgeConfigBuilder_SetPermissions( DOFOperationBridgeConfigBuilder self, DOFPermissionSet permissions );

/**
 *  Sets whether a permission extension is permitted on a DOFOperationBridgeConfig built by some DOFOpeartionBridgeBuilder.
 *
 *  @param self The DOFOperationBridgeConfigBuilder.  Should not be @c NULL.
 *  @param extendAllowed flag to control permission extension
 *
 *  @retval TRUE permission extension option was set successfully for @a self
 *  @retval FALSE permission extension option was not set successfully for @a self
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationBridgeConfigBuilder_SetPermissionsExtendAllowed( DOFOperationBridgeConfigBuilder self, boolean extendAllowed );

/**
@}
**/

/**
 *  Copies a DOFOperationBridgeConfig.
 *
 *  @mustfree{DOFOperationBridgeConfig_Destroy}
 *
 *  @param self The DOFOperationBridgeConfig to copy.  Should not be @c NULL.
 *
 *  @return a copy of @a self
 */
extern DOF_OAL_DECLSPEC DOFOperationBridgeConfig DOFOperationBridgeConfig_Copy(DOFOperationBridgeConfig self);

/**
 *  Destroys a DOFOperationBridgeConfig.
 *
 *  @param self The DOFOperationBridgeConfig to destroy.  If @c NULL this function will do nothing.
 */
extern DOF_OAL_DECLSPEC void DOFOperationBridgeConfig_Destroy(DOFOperationBridgeConfig self);

/**
 * Gets the credentials in this DOFOperationBridgeConfig.
 *
 * @param self The DOFOperationBridgeConfig.  This must not be NULL.
 * @return The credentials held by the configuration.
 */
extern DOF_OAL_DECLSPEC DOFCredentials DOFOperationBridgeConfig_GetCredentials(DOFOperationBridgeConfig self);

/**
 * Gets the Permissions for the given DOFOperationBridgeConfig.
 *
 * @param self The DOFOperationBridgeConfig.  This must not be NULL.
 *
 * @return The permissions held by the configuration.
 */
extern DOF_OAL_DECLSPEC DOFPermissionSet DOFOperationBridgeConfig_GetPermissions(DOFOperationBridgeConfig self);

/**
 * Checks to see if the Bridge Config is set up to allow permission extensions.
 *
 * @param self The DOFOperationBridgeConfig.  This must not be NULL.
 *
 * @return TRUE if permission extensions are allowed, FALSE otherwise.
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationBridgeConfig_IsPermissionsExtendAllowed(DOFOperationBridgeConfig self);

/**
@}
**/

#ifdef  __cplusplus
}
#endif


#endif /* DOF_BRIDGE_H */
