/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/* File name is not config.h because that filename is used for build control. */

#ifndef DOF_CONFIG_H
#define DOF_CONFIG_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DOFConfig DOFConfig
    @ingroup DOF
    The DOFConfig class specifies the initial configuration to use when creating a DOF. To instantiate this class,
    you use methods of @ref DOFConfigBuilder to set configuration parameters, and call @ref DOFConfigBuilder_Build
    or DOFConfigBuilder_BuildAndDestroy to return a DOFConfig when you have finished setting the desired parameters.
    You then pass the instantiated DOFConfig to @ref DOF_Create to instantiate a DOF.

    The methods of this class are mostly getters for the parameters set using DOFConfigBuilder. Use these getters
    to get information about a DOFConfig that may or may not have been used to instantiate a DOF.
    To instead get current information about a DOF instance, use the methods of @ref DOFState.
@{
**/

/**
 * A mid-range value (128), so that the default desire to manage a secure multicast group
 * is neither strong nor weak. Use @ref DOFConfigBuilder_SetRank to set a different value.
 *
 * @see DOFConfig_GetRank
 * @see DOFState_GetRank
 **/
#define DOFCONFIG_DEFAULT_RANK 128

/**
 * Destroys a DOFConfig.
 *
 * @param self The DOFConfig to destroy. If @c NULL, no action is taken.
 * @see DOFConfigBuilder_Build
 * @see DOFConfigBuilder_BuildAndDestroy
 */
extern DOF_OAL_DECLSPEC void DOFConfig_Destroy( DOFConfig self );

/**
 * Gets the identifier that DOFs created from this @ref DOFConfig will use for operations in the unsecured domain.
 *
 * @param self The DOFConfig to get the value from. This must not be @c NULL.
 *
 * @return The DOFObjectIDSource stored in the DOFConfig.
 * @see DOFState_GetSourceID
 * @see DOFConfigBuilder_SetSourceID
 */
extern DOF_OAL_DECLSPEC DOFObjectIDSource DOFConfig_GetSourceID( DOFConfig self );

/**
 * Gets the maximum number of operations that DOFs created from this @ref DOFConfig will support.
 *
 * @deprecated The ability to set a maximum number of operations was removed. This function
 * always returns 0xFFFFFFFF.
 *
 * @param self The DOFConfig to get the value from. This must not be @c NULL.
 *
 * @return 0xFFFFFFFF
 * @see DOFConfigBuilder_SetMaxOperations
 */
  
extern DOF_OAL_DECLSPEC uint32 DOFConfig_GetMaxOperations( DOFConfig self );

/**
 * Gets the maximum number of connections that DOFs created from this @ref DOFConfig will support.
 *
 * @param self The DOFConfig to get the value from. This must not be @c NULL.
 *
 * @deprecated This functionality was replaced by @ref DOFConfig_GetConnectionLimit.
 *
 * @return The value for connection limit that is stored in the DOFConfig.
 * @see DOFState_GetMaxConnections
 * @see DOFConfigBuilder_SetMaxConnections
 */
extern DOF_OAL_DECLSPEC uint16 DOFConfig_GetMaxConnections( DOFConfig self );

/**
 * Gets the maximum number of connections that DOFs created from this @ref DOFConfig will support.
 *
 * @param self The DOFConfig to get the value from. This must not be @c NULL.
 *
 * @return The value for connection limit that is stored in the DOFConfig.
 * @see DOFState_GetConnectionLimit
 * @see DOFConfigBuilder_SetConnectionLimit
 */
extern DOF_OAL_DECLSPEC uint32 DOFConfig_GetConnectionLimit( DOFConfig self );

/**
 * Gets the maximum number of threads that DOFs created from this @ref DOFConfig will have in their
 * thread pool.
 *
 * @param self The DOFConfig to get the value from. This must not be @c NULL.
 *
 * @return The thread pool size that is stored in the DOFConfig.
 * @see DOFState_GetThreadPoolSize
 */
extern DOF_OAL_DECLSPEC uint16 DOFConfig_GetThreadPoolSize( DOFConfig self );

/**
 * Discovers whether DOFs created from this DOFConfig will have routing set to true.
 *
 * @param self The DOFConfig to get the value from. This must not be @c NULL.
 *
 * @return TRUE if the DOFConfig has routing set to true; FALSE otherwise.
 * @see DOFState_IsRouter
 * @see DOFConfigBuilder_SetRouter
 */
extern DOF_OAL_DECLSPEC boolean DOFConfig_IsRouter( DOFConfig self );

/**
 * Gets the node rank that DOFs created from this @ref DOFConfig will have.
 *
 * @param self The DOFConfig to get the value from. This must not be @c NULL.
 *
 * @return The rank stored in the DOFConfig.
 */
extern DOF_OAL_DECLSPEC uint8 DOFConfig_GetRank( DOFConfig self );

/**
 * Checks whether DOFs created from this @ref DOFConfig will validate @ref DOFValue objects to verify
 * that the associated @ref DOFType matches a required type.
 *
 * @param self The DOFConfig to get the value from. This must not be @c NULL.
 *
 * @return TRUE if the DOFConfig has validation set to true; FALSE otherwise.
 * @see DOFState_IsParameterValidation
 * @see DOFConfigBuilder_SetParameterValidation
 */
extern DOF_OAL_DECLSPEC boolean DOFConfig_IsParameterValidation( DOFConfig self );

/** @name Builder
  The DOFConfigBuilder class enables you to customize the configuration of a DOF.
  The @ref DOFConfigBuilder_Create function returns a DOFConfigBuilder with all the
  configuration parameters for a DOF set to their defaults. You can then pass the DOFConfigBuilder
  to DOFConfigBuilder functions to change the initial configuration parameters. You do not need to set any configuration parameters that you wish to leave at their default settings.

  When you have finished setting configuration parameters, you use @ref DOFConfigBuilder_Build
  or @ref DOFConfigBuilder_BuildAndDestroy to create an immutable @ref DOFConfig object that stores all the chosen settings. Pass that DOFConfig to @ref DOF_Create to create your customized DOF.

  For example, by default DOFs cannot be proxies. To create a DOF that is capable of being a proxy node, you must set routing to true, as in the following example:
  <pre>     bldr = DOFConfigBuilder_Create();
     DOFConfigBuilder_SetRouter(bldr, TRUE);
     config = DOFConfigBuilder_BuildAndDestroy(bldr);
     dof = DOF_Create(config);
     DOFConfig_Destroy(config);

     ...

     DOF_Destroy(dof);</pre>
  The following are the default configuration parameters:

  - The default maximum number of simultaneous connections, both inbound and outbound, that the DOF
    allows is 0xFFFFFFFF (effectively unlimited). Use @ref DOFConfigBuilder_SetConnectionLimit to set a different value.
  - The default number of threads the DOF can create is 20. Use @ref DOFConfigBuilder_SetThreadPoolSize to set a different value.
  - The default source identifer is a randomly generated GUID. Use @ref DOFConfigBuilder_SetSourceID to set a different value.
  - The node's default rank is 128. Use @ref DOFConfigBuilder_SetRank to set a different value.
  - By default, the DOF validates DOFValues to ensure they match required types. Use @ref DOFConfigBuilder_SetParameterValidation
    to disable validation.
  - By default, the DOF is not set to be a router. Use @ref DOFConfigBuilder_SetRouter to set routing to true.

  You can also use @ref DOFConfigBuilder_Create_Config and @ref DOFConfigBuilder_Create_State to
  create DOFConfigBuilders whose initial configuration parameters are based on another DOFConfig or on another DOF's state, rather than on the defaults. Just as with the default DOFConfigBuilder, you would then change any configuration parameters that need to be changed, and then build a new DOFConfig with the new settings.
@{
**/

/**
 * Creates a @ref DOFConfigBuilder with configuration parameters set to their defaults.
 * For information about the default values, see DOFConfigBuilder.
 *
 * @mustfree{DOFConfigBuilder_Destroy or @ref DOFConfigBuilder_BuildAndDestroy}
 *
 * @return A default DOFConfigBuilder.
 * @see DOFConfigBuilder_Create_Config
 * @see DOFConfigBuilder_Create_State
 */
extern DOF_OAL_DECLSPEC DOFConfigBuilder DOFConfigBuilder_Create( void );

/**
 * Copies configuration parameters from a @ref DOFState and creates a @ref DOFConfigBuilder with the same initial
 * configuration parameters.
 *
 * @mustfree{DOFConfigBuilder_Destroy or @ref DOFConfigBuilder_BuildAndDestroy}
 *
 * @param state The DOFState instance whose parameters to copy. This must not be @c NULL.
 *
 * @return DOFConfigBuilder A DOFConfigBuilder with parameters copied from a DOFState.
 * @see DOF_CreateState
 * @see DOFConfigBuilder_Create
 * @see DOFConfigBuilder
 */
extern DOF_OAL_DECLSPEC DOFConfigBuilder DOFConfigBuilder_Create_State( const DOFState state );

/**
 * Copies configuration parameters from a @ref DOFConfig and creates a DOFConfigBuilder with the same initial
 * configuration parameters.
 *
 * @mustfree{DOFConfigBuilder_Destroy or @ref DOFConfigBuilder_BuildAndDestroy}
 *
 * @param config The DOFConfig instance whose parameters to copy. This must not be @c NULL.
 *
 * @return A DOFConfigBuilder with parameters copied from a DOFConfig.
 * @see DOFConfigBuilder_Create
 * @see DOFConfigBuilder_Create_State
 */
extern DOF_OAL_DECLSPEC DOFConfigBuilder DOFConfigBuilder_Create_Config( const DOFConfig config );

/**
 * Sets the identifier that is used for the DOF's operations in the unsecured domain. Operations in
 * secure domains may not use this source identifier. Source identifiers in secure domains may be assigned
 * by the Authentication Server (AS) and thus, would not be under application control.
 *
 * Source identifiers must be globally unique. By default, the source identifier is a GUID that is
 * randomly generated when the DOFConfigBuilder is created. If you use this method to set a constant value
 * for the source identifier, you should not use a DOFConfig created from this DOFConfigBuilder to instantiate more
 * than one DOF. You should also ensure that you include @ref DOF_CancelAllOperations in the DOF's startup routines.
 *
 * @param self the DOFConfigBuilder to set the source identifier in. This must not be @c NULL.
 * @param sourceID The @ref DOFObjectIDSource to use as an identifier in the unsecured domain. This must not be @c NULL.
 *
 * @return TRUE if the sourceID is set in the DOFConfigBuilder.
 * @exception FALSE if the sourceID could not be set. The DOFConfigBuilder remains unchanged.
 * @see DOFObjectIDSource_Create
 * @see DOFState_GetSourceID
 * @see DOFConfig_GetSourceID
 */
extern DOF_OAL_DECLSPEC boolean DOFConfigBuilder_SetSourceID( DOFConfigBuilder self, DOFObjectIDSource sourceID );

/**
 * @deprecated The ability to set a maximum number of operations was removed.  This function no
 * longer has any effect. The maximum number of simultaneous operations that the DOF supports is 0xFFFFFFFF (effectively unlimited).
 *
 * @param self the DOFConfigBuilder to set the maximum number of operations in. This must not be @c NULL.
 * @param max Any uint32. The maximum number of operations will remain set to 0xFFFFFFFF.
 *
 * @return TRUE in cases where this function would previously have set the value, although the DOFConfigBuilder remains unchanged.
 * @exception FALSE if the arguments are invalid. The DOFConfigBuilder remains unchanged.
 */
extern DOF_OAL_DECLSPEC boolean DOFConfigBuilder_SetMaxOperations( DOFConfigBuilder self, uint32 max );

/**
 * Sets the maximum number of simultaneous connections, both inbound and outbound, that a DOF can support.
 * Inbound connections are those connected through all @ref DOFServer instances on the node, although this
 * parameter does not limit the number of servers you can create or start. Outbound connections are instances
 * of @ref DOFConnection for which you've called @ref DOFConnection_Connect or @ref DOFConnection_BeginConnect,
 * although this parameter does not limit the number of connections you can create. In addition, only calls to
 * DOFConnection_Connect or DOFConnection_BeginConnect that establish a new connection on the physical wire count
 * toward the total&mdash;not calls for connections that have already been established. The sum of all inbound
 * plus all outbound connections on the node must always be less than or equal to the value set for this parameter.
 * If the node reaches its maximum, it will not allow you to initiate or accept new connections until some
 * connections are dropped.
 *
 * Inbound DOFCONNECTIONTYPE_DATAGRAM_STATELESS connections are not limited by this parameter, although outbound
 * DATAGRAM_STATELESS connections do count toward the total.
 *
 * If you input a value for the maximum number of connections that is greater than the limit the underlying
 * operating system can handle, the number of connections you can initiate and accept will be constrained by the
 * operating system. In addition, inputting a zero means that the library will not limit the maximum number of
 * connections, but the operating system limits still apply.
 *
 * By default, this parameter is set to a large number so that the operating system will manage the maximum number
 * of connections. You would typically change this setting only for nodes with limited resources where you want to
 * limit the connections to fewer than the operating system allows. If you enforce a limit, it should be carefully
 * considered so that you do not prevent necessary connections.
 *
 * @deprecated This functionality was replaced by @ref DOFConfigBuilder_SetConnectionLimit.
 *
 * @param self the DOFConfigBuilder to set the maximum number of connections in. This must not be @c NULL.
 * @param max The maximum number of connections that the DOF can support. A value of zero means there is no limit.
 *
 * @return TRUE if the maximum value was set in the DOFConfigBuilder.
 * @exception FALSE if the maximum value could not be set. The DOFConfigBuilder remains unchanged.
 * @see DOF_SetMaxConnections
 * @see DOFState_GetMaxConnections
 * @see DOFConfig_GetMaxConnections
 */
extern DOF_OAL_DECLSPEC boolean DOFConfigBuilder_SetMaxConnections( DOFConfigBuilder self, uint16 max );

/**
 * Sets the maximum number of simultaneous connections, both inbound and outbound, that a DOF can support.
 * Inbound connections are those connected through all @ref DOFServer instances on the node, although this
 * parameter does not limit the number of servers you can create or start. Outbound connections are instances
 * of @ref DOFConnection for which you've called @ref DOFConnection_Connect or @ref DOFConnection_BeginConnect,
 * although this parameter does not limit the number of connections you can create. In addition, only calls to
 * DOFConnection_Connect or DOFConnection_BeginConnect that establish a new connection on the physical wire count
 * toward the total&mdash;not calls for connections that have already been established. The sum of all inbound
 * plus all outbound connections on the node must always be less than or equal to the value set for this parameter.
 * If the node reaches its maximum, it will not allow you to initiate or accept new connections until some
 * connections are dropped.
 *
 * Inbound DOFCONNECTIONTYPE_DATAGRAM_STATELESS connections are not limited by this parameter, although outbound
 * DATAGRAM_STATELESS connections do count toward the total.
 *
 * If you input a value for the maximum number of connections that is greater than the limit the underlying
 * operating system can handle, the number of connections you can initiate and accept will be constrained by the
 * operating system.
 *
 * By default, this parameter is set to a large number so that the operating system will manage the maximum number
 * of connections. You would typically change this setting only for nodes with limited resources where you want to
 * limit the connections to fewer than the operating system allows. If you enforce a limit, it should be carefully
 * considered so that you do not prevent necessary connections.
 *
 * @param self the DOFConfigBuilder to set the maximum number of connections in. This must not be @c NULL.
 * @param max The maximum number of connections that the DOF can support. A value of zero will prevent all connections.
 *
 * @return TRUE if the maximum value was set in the DOFConfigBuilder.
 * @exception FALSE if the maximum value could not be set. The DOFConfigBuilder remains unchanged.
 * @see DOF_SetConnectionLimit
 * @see DOFState_GetConnectionLimit
 * @see DOFConfig_GetConnectionLimit
 */
extern DOF_OAL_DECLSPEC boolean DOFConfigBuilder_SetConnectionLimit( DOFConfigBuilder self, uint32 max );

/**
 * Sets the maximum number of threads that a DOF can create to process operations.
 *
 * If you input a value for thread pool size that is greater than the limit the underlying operating
 * system can handle, the number of threads you can create will be constrained by the operating system.
 * Setting this parameter too low or too high may cause performance
 * problems, based on the application, operating system, and hardware.
 *
 * @param self the DOFConfigBuilder to set the thread pool size in. This must not be @c NULL.
 * @param count The number of threads that the DOF can create. The value must be greater than zero.
 *
 * @return TRUE if the size was set in the DOFConfigBuilder.
 * @exception FALSE if the size could not be set. The DOFConfigBuilder remains unchanged.
 * @see DOFState_GetThreadPoolSize
 * @see DOFConfig_GetThreadPoolSize
 */
extern DOF_OAL_DECLSPEC boolean DOFConfigBuilder_SetThreadPoolSize( DOFConfigBuilder self, uint16 count );

/**
 * Sets whether a DOF will route operations other than those it directly requests or provides. A DOF that
 * is a router becomes a centralized node in the network topology (a proxy in DOF architecture), capable of routing
 * operations between other requestors and providers. DOFs that are not routers act as network endpoints and
 * route only operations they initiate or that are addressed to them.
 *
 * The default DOFConfigBuilder has routing set to @c FALSE.
 *
 * @param self the DOFConfigBuilder to set the routing value in. This must not be @c NULL.
 * @param isRouter TRUE sets the node to be a centralized router. FALSE sets the node to be a network endpoint.
 *
 * @return TRUE if routing was set in the DOFConfigBuilder.
 * @exception FALSE if routing could not be set. The DOFConfigBuilder remains unchanged.
 * @see DOFState_IsRouter
 * @see DOFConfig_IsRouter
 */
extern DOF_OAL_DECLSPEC boolean DOFConfigBuilder_SetRouter( DOFConfigBuilder self, boolean isRouter );

/**
 * Sets the node's desire to manage a secure multicast group. The rank is a number from zero to 255.
 * The higher the number, the stronger the desire to be a manager. Nodes with plentiful resources
 * typically are better candidates for group manager and should have a higher rank. Nodes with
 * limited resources should have a lower rank.
 *
 * The default node rank is 128.
 *
 * @param self The DOFConfigBuilder to set the rank in. This must not be @c NULL.
 * @param rank The DOF's rank.
 *
 * @return TRUE if the rank was set in the DOFConfigBuilder.
 * @exception FALSE if the the rank could not be set. The DOFConfigBuilder remains unchanged.
 * @see DOFCONNECTIONTYPE_GROUP
 * @see DOFState_GetRank
 * @see DOFConfig_GetRank
 */
extern DOF_OAL_DECLSPEC boolean DOFConfigBuilder_SetRank( DOFConfigBuilder self, uint8 rank );

/**
 * Sets whether the DOF validates @ref DOFValue objects to verify that their associated @ref DOFType
 * matches the type required for an interface item. The default is @c TRUE. If you set validation to @c FALSE,
 * the library will bypass its normal type checking, so you must ensure that the application is using valid
 * DOFValue types.
 *
 * @param self The DOFConfigBuilder to set validation in.
 * @param validation TRUE to perform validation. FALSE to turn validation off.
 *
 * @return TRUE if validation was set in the DOFConfigBuilder.
 * @exception FALSE if validation could not be set. The DOFConfigBuilder remains unchanged.
 * @see DOFState_IsParameterValidation
 * @see DOFConfig_IsParameterValidation
 */
extern DOF_OAL_DECLSPEC boolean DOFConfigBuilder_SetParameterValidation( DOFConfigBuilder self, boolean validation );

/**
 * Builds a @ref DOFConfig based on the parameters set in this DOFConfigBuilder.
 * @mustfree{DOFConfig_Destroy}
 *
 * @param self The DOFConfigBuilder to build a DOFConfig from. This must not be @c NULL.
 *
 * @return DOFConfig The configuration for a DOF.
 * @exception NULL if the configuration could not be built.
 * @see DOFConfigBuilder_BuildAndDestroy
 * @see DOF_Create
 */
extern DOF_OAL_DECLSPEC DOFConfig DOFConfigBuilder_Build( DOFConfigBuilder self );

/**
 * Builds a @ref DOFConfig based on the parameters set in this DOFConfigBuilder and destroys the DOFConfigBuilder.
 * @mustfree{DOFConfig_Destroy}
 *
 * @param self The DOFConfigBuilder to build a DOFConfig from. This must not be @c NULL.
 *
 * @return DOFConfig The configuration for a DOF.
 * @exception NULL if the configuration could not be built. The DOFConfigBuilder is still destroyed.
 * @see DOFConfigBuilder_Build
 * @see DOF_Create
 */
extern DOF_OAL_DECLSPEC DOFConfig DOFConfigBuilder_BuildAndDestroy( DOFConfigBuilder self );

/**
 * Destroys this DOFConfigBuilder.
 *
 * @param self The DOFConfigBuilder to destroy. If @c NULL, no action is taken.
 * @see DOFConfigBuilder_Create
 * @see DOFConfigBuilder_Create_State
 * @see DOFConfigBuilder_Create_Config
 */
extern DOF_OAL_DECLSPEC void DOFConfigBuilder_Destroy( DOFConfigBuilder self );

/**
@}
**/

/**
@}
**/

/** @defgroup DOFState DOFState
    @ingroup DOF

    A DOFState instance represents a snapshot of the DOF's configuration information as it exists at the moment when you call
    @ref DOF_CreateState. After calling DOF_CreateState, use the DOFState functions to retrieve specific information about the DOF's configuration. Because the DOF's configuration information may change, you should call @ref DOF_CreateState again each time you want updated information.
@{
**/

/**
 * Destroys the DOFState.
 *
 * @param self The state to destroy. If @c NULL, no action is taken.
 * @see DOF_CreateState
 */
extern DOF_OAL_DECLSPEC void DOFState_Destroy( DOFState self );

/**
 * Gets the identifier that is used for the DOF's operations in the unsecured domain. This value may change
 * after a DOF is rebooted if source identifiers are randomly generated.
 *
 * @param self The state to get the value from. This must not be @c NULL.
 *
 * @return The DOF's DOFObjectIDSource.
 * @see DOFConfigBuilder_SetSourceID
 */
extern DOF_OAL_DECLSPEC DOFObjectIDSource DOFState_GetSourceID( DOFState self );

/**
 * Gets the maximum number of operations, including both local and remote operations, that a DOF supports.
 *
 * @deprecated The ability to set a maximum number of operations was removed. This function
 * always returns 0xFFFFFFFF.
 *
 * @param self The state to get the value from. This must not be @c NULL.
 *
 * @return 0xFFFFFFFF.
 * @see DOFConfigBuilder_SetMaxOperations
 */
extern DOF_OAL_DECLSPEC uint32 DOFState_GetMaxOperations( DOFState self );

/**
 * Gets the maximum number of connections, both inbound and outbound, that a DOF currently supports. This value may
 * change through use of @ref DOF_SetMaxConnections.
 *
 * @deprecated This functionality was replaced by @ref DOFState_GetConnectionLimit.
 *
 * @param self The state to get the value from. This must not be @c NULL.
 *
 * @return The maximum number of connections, both inbound and outbound, that a DOF currently supports.
 * @see DOFConfigBuilder_SetMaxConnections
 */
extern DOF_OAL_DECLSPEC uint16 DOFState_GetMaxConnections( DOFState self );

/**
 * Gets the maximum number of connections, both inbound and outbound, that a DOF currently supports. This value may
 * change through use of @ref DOF_SetConnectionLimit.
 *
 * @param self The state to get the value from. This must not be @c NULL.
 *
 * @return The maximum number of connections, both inbound and outbound, that a DOF currently supports.
 * @see DOFConfigBuilder_SetConnectionLimit
 */
extern DOF_OAL_DECLSPEC uint32 DOFState_GetConnectionLimit( DOFState self );


/**
 * Gets the maximum number of threads that the DOF can create to process requests.
 *
 * @param self The state to get the value from. This must not be @c NULL.
 *
 * @return The maximum number of threads that a DOF supports.
 * @see DOFConfigBuilder_SetThreadPoolSize
 */
extern DOF_OAL_DECLSPEC uint16 DOFState_GetThreadPoolSize( DOFState self );

/**
 * Discovers whether a DOF is configured as a router.
 *
 * @param self The state to get the value from. This must not be @c NULL.
 *
 * @return TRUE if the DOF is a router; FALSE otherwise.
 * @see DOFConfigBuilder_SetIsRouter
 */
extern DOF_OAL_DECLSPEC boolean DOFState_IsRouter( DOFState self );

/**
 * Gets the node's rank, which represents its desire to manage a secure multicast group.
 *
 * @param self The state to get the value from. This must not be @c NULL.
 *
 * @return The node's rank.
 * @see DOFConfigBuilder_SetRank
 */
extern DOF_OAL_DECLSPEC uint8 DOFState_GetRank( DOFState self );

/**
 * Checks whether the DOF validates @ref DOFValue objects to verify
 * that the associated @ref DOFType matches a required type.
 *
 * @param self The state to get the value from. This must not be @c NULL.
 *
 * @return TRUE if the DOF performs validation; FALSE otherwise.
 * @see DOFConfigBuilder_SetParameterValidation
 */
extern DOF_OAL_DECLSPEC boolean DOFState_IsParameterValidation( DOFState self );

/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* DOF_CONFIG_H */
