/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_CIPHERREGISTRY_H
#define DOF_CIPHERREGISTRY_H

#include <dof/blockcipher.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DOFCipherRegistry DOFCipherRegistry
    @ingroup DOF_Framework

    The block cipher registry for the C OAL.

    Required block cipher algorithms are usually compiled and pre-registered with the C OAL where available.
    For automatic pre-registration of a cipher, the cipher implementation file must contain a variable definition of the
    format:

        const struct BlockCipher BlockCipher_&lt;algorithm&gt;_&lt;strength_id&gt;

    Where &lt;algorithm&gt; is one of the @ref DOFCipherAlgorithm identifiers without the DOFCIPHERALGORITHM_ prefix,
    and &lt;strength_id&gt; is one of the @ref DOFCipherStrength identifiers without the DOFCIPHERSTRENGTH_ prefix. For
    example, to pre-register @ref DOFCIPHERALGORITHM_AES with @ref DOFCIPHERSTRENGTH_256, the following definition
    should exist:

        const struct BlockCipher BlockCipher_AES_256 = { ... }

@{
**/

/*
**  "Public" Type and Method Declarations
*/

/**
    Register a block cipher with the library.

    Block cipher implementations may already be compiled and pre-registered with the library by default. However, if
    your library was not built with the cipher your application requires, this method allows an external block cipher to
    be registered and used.  For security reasons, once a block cipher is registered, it cannot be replaced.

    @note Only the AES block cipher with 256-bit key strength is supported by the C OAL at this time. All uses of this
          cipher only require encryption (decryption support is not required).

    @param algorithm The recognized algorithm to be registered; refer to @ref DOFCipherAlgorithm.
    @param strength The security strength as defined by the OpenDof Project; refer to @ref DOFCipherStrength.
    @param  cipher The block cipher implementation; refer to @ref DOFCipher.

    @retval Non-zero (TRUE) if the cipher was successfully added to the registry.
    @exception Zero (FALSE) if the cipher could not be added. This may occur for a number of reasons, including:
     - the algorithm is not recognized by the library; or
     - the strength is not recognized by the library; or
     - a cipher was already added for the given algorithm and strength; or
     - the cipher implementation fails to meet the library's requirements for the given algorithm and strength
       (e.g., incorrect block size, excessive key state size, or missing required functionality).
**/
extern DOF_OAL_DECLSPEC boolean DOFCipherRegistry_Register( DOFCipherAlgorithm algorithm, DOFCipherStrength strength, DOFCipher cipher );


/**
    Retrieve a block cipher implementation from the library's registry.

    @param algorithm The recognized algorithm to be registered; refer to @ref DOFCipherAlgorithm.
    @param strength The security strength as defined by the OpenDOF Project; refer to @ref DOFCipherStrength.

    @return The block cipher implementation.
    @exception NULL is returned if the registry does not contain an implementation of the given algorithm and strength.
**/
extern DOF_OAL_DECLSPEC DOFCipher DOFCipherRegistry_Get( DOFCipherAlgorithm algorithm, DOFCipherStrength strength );


/**
@}
**/


#ifdef  __cplusplus
}
#endif

#endif /* DOF_CIPHERREGISTRY_H */
