/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_CONNECTION_H
#define DOF_CONNECTION_H


#include <dof/core.h>
#include <dof/transport.h>


#ifdef  __cplusplus
extern "C" {
#endif


/** @defgroup DOFConnectionConfig DOFConnectionConfig
    @ingroup DOFConnection

    The DOFConnectionConfig class represents most of the information necessary
    for a DOF to create a connection.
@{
**/

/**
 * @name Builder
 * @{
 */

/**
 * Create a connection configuration builder, given the type and target address.
 * @mustfree{DOFConnectionConfigBuilder_Destroy or @ref DOFConnectionConfigBuilder_BuildAndDestroy}
 * The builder uses the following defaults:
 *  -Connection Type: Given by @a type
 *  -Address: Given by @a address
 *  -Security Desire: @ref DOFSECURITYDESIRE_ANY
 *  -Credentials: NULL
 *  -Domain Credentials: none
 *  -Trusted Domains: none
 *  -Permissions: NULL
 *  -Transport Config: NULL
 *  -Protocol Negotiator: @ref DOF_DEFAULT_PROTOCOL_NEGOTIATOR
 *  -Bridge: NULL
 *  -Receive Filter: NULL
 *  -Send Filter: NULL
 *  -Stream Listener: NULL
 *  -Permission Extend Allowed: TRUE
 *  -Max Send Silence: @ref DOF_TIMEOUT_NEVER
 *  -Max Receive Silence: 120,000ms (2 minutes)
 * @param type The type of Connection.  This must be one of the @ref DOFConnectionType constants.
 * @param target The target @ref DOFAddress of the Connection.  This must not be NULL.
 * @return A new @ref DOFConnectionConfigBuilder.
 * @exception NULL is returned if a connection configuration builder could not be created.
 */
extern DOF_OAL_DECLSPEC DOFConnectionConfigBuilder DOFConnectionConfigBuilder_Create( DOFConnectionType type, DOFAddress target );

/**
 * Create a connection configuration builder, given an existing connection configuration.
 * @mustfree{DOFConnectionConfigBuilder_Destroy or @ref DOFConnectionConfigBuilder_BuildAndDestroy}
 * @param config The existing Connection Configuration.  This must not be NULL.
 * @return A new @ref DOFConnectionConfigBuilder.
 * @exception NULL is returned if a connection configuration builder could not be created.
 */
extern DOF_OAL_DECLSPEC DOFConnectionConfigBuilder DOFConnectionConfigBuilder_Create_Config( DOFConnectionConfig config );

/**
 * Create a connection configuration builder, given an existing connection state.
 *
 * @note This call will not use the credentials, bridge, domains, or filters from the connection from which
 * the connection state originated.
 *
 * @mustfree{DOFConnectionConfigBuilder_Destroy or @ref DOFConnectionConfigBuilder_BuildAndDestroy}
 * @param state The State of an existing Connection.  This must not be NULL.
 * @return A new @ref DOFConnectionConfigBuilder.
 * @exception NULL is returned if a connection configuration builder could not be created.
 */
extern DOF_OAL_DECLSPEC DOFConnectionConfigBuilder DOFConnectionConfigBuilder_Create_State( DOFConnectionState state );

/**
 * Create a new copy of an existing @ref DOFConnectionConfigBuilder.
 * @mustfree{DOFConnectionConfigBuilder_Destroy or @ref DOFConnectionConfigBuilder_BuildAndDestroy}
 * @param self The @ref DOFConnectionConfigBuilder to be copied.  This should not be NULL.
 * @return A new copy of the @ref DOFConnectionConfigBuilder.
 * @exception NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DOFConnectionConfigBuilder DOFConnectionConfigBuilder_Copy( DOFConnectionConfigBuilder self );

/**
   Set the Connection Type in the given DOFConnectionConfigBuilder.

   @param self The Connection Config Builder.  If @a self is NULL, no action will be taken.
   @param type The type of Connection.  This must be one of the @ref DOFConnectionType constants.

   @retval TRUE Connection type was set.
   @exception FALSE Connection type set failed.
**/
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfigBuilder_SetConnectionType( DOFConnectionConfigBuilder self, DOFConnectionType type );

/**
   Set the target Address in the given DOFConnectionConfigBuilder.

   @param self The Connection Config Builder.  If @a self is NULL, no action will be taken.
   @param target The target Address of the connection.  This must not be NULL.

   @retval TRUE The address was set.
   @exception FALSE Failed to set address.
**/
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfigBuilder_SetAddress( DOFConnectionConfigBuilder self, DOFAddress target );

/**
   Set the Security Desire in the given DOFConnectionConfigBuilder.
   Please refer to @ref DOFConnectionConfigBuilder_Create for the default value for this setting.

   @param self The Connection Config Builder.  If @a self is NULL, no action will be taken.
   @param desire The Security Desire of the connection.  This must be one of the @ref DOFSecurityDesire constants.

   @retval TRUE Security desire was set.
   @exception FALSE Security desire set failed.
**/
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfigBuilder_SetSecurityDesire( DOFConnectionConfigBuilder self, DOFSecurityDesire desire );

/**
   Set the Credentials in the given DOFConnectionConfigBuilder.
   Please refer to @ref DOFConnectionConfigBuilder_Create for the default value for this setting.

   @param self        The Connection Config Builder.  If @a self is NULL, no action will be taken.
   @param credentials The authentication Credentials that will be used on the connection.  If @a credentials is NULL, no credentials
                      are used, and the security mode and permissions are ignored.

   @retval TRUE The Credentials was set.
   @exception FALSE Failed to set Credentials.
**/
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfigBuilder_SetCredentials( DOFConnectionConfigBuilder self, DOFCredentials credentials );

/**
 * Add credentials to validate that the resulting connection is to an authorized AS (one able to process authentication requests) for a domain.
 * If the validation for any credentials fails, the connection will fail to open.
 *
 * @param self DOFConnectionConfigBuilder.  If @a self is NULL, no action will be taken.
 * @param domain The domain config to add.  This should not be NULL.
 *
 * @retval TRUE if the credentials were added successfully.
 * @exception FALSE is returned if an invalid parameter was detected or insufficient resources were available to add the credentials.
 */
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfigBuilder_AddDomain( DOFConnectionConfigBuilder self, DOFDomainConfig domain );

/**
* Sets the domain wildcard credentials for the connection.
*
* @param self The DOFConnectionConfigBuilder to add the wildcard domain credentials to. If @a self is NULL, no action will be taken.
* @param credentials The DOFCredentials to use as wildcard domain credentials. If NULL remove previously set credentials
*
* @return boolean TRUE The DOFCredentials were added to the config.
* @exception FALSE If the DOFCredentials failed to be added.
*/
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfigBuilder_SetDomainDiscoveryCredentials( DOFConnectionConfigBuilder self, DOFCredentials credentials );

/**
 * Indicate that the resulting connection is a trusted route to an authorized AS (one able to process authentication requests) for a domain.
 * The domain is not validated by any credentials.
 *
 * @param self DOFConnectionConfigBuilder.  If @a self is NULL, no action will be taken.
 * @param domainID The ID of the domain to be trusted on the resulting connection.  This should not be NULL.  This may be @ref DOFOBJECTIDDOMAIN_BROADCAST to indicate the connection is trusted for @b any domain.
 *
 * @retval TRUE if the domain route was added successfully.
 * @exception FALSE is returned if an invalid parameter was detected or insufficient resources were available to add the domain route.
 */
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfigBuilder_AddTrustedDomain( DOFConnectionConfigBuilder self, DOFObjectIDDomain domainID );

/**
   Set the Permissions in the given DOFConnectionConfigBuilder.
   Please refer to @ref DOFConnectionConfigBuilder_Create for the default value for this setting.

   @param self The Connection Config Builder.  If @a self is NULL, no action will be taken.
   @param perms The initial (and required) Permissions for the connection.  If @a perms is NULL, no initial permissions are used.

   @retval TRUE Permissions were set.
   @exception FALSE Failed to set the permissions.
**/
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfigBuilder_SetPermissions( DOFConnectionConfigBuilder self, DOFPermissionSet perms );

/**
   Set whether Permission extensions will be allowed or not.
   Please refer to @ref DOFConnectionConfigBuilder_Create for the default value for this setting.

   @param self The Connection Config Builder.  If @a self is NULL, no action will be taken.
   @param extendAllowed TRUE if permission extensions are allowed, FALSE otherwise.

   @retval TRUE Permission extend allowed was set to @a extendAllowed.
   @exception FALSE Permission extend allowed was not set to @a extendAllowed.
**/
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfigBuilder_SetPermissionsExtendAllowed( DOFConnectionConfigBuilder self, boolean extendAllowed );

/**
   Set the transport-specific configuration for the given DOFConnectionConfigBuilder
   Please refer to @ref DOFConnectionConfigBuilder_Create for the default value for this setting.

   @param self The Connection Config Builder.  If @a self is NULL, no action will be taken.
   @param config The transport-specific configuration.  If @a config is NULL, the default transport-specific configuration is used.

   @retval TRUE Transport config was set.
   @exception FALSE Transport config set failed.
**/
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfigBuilder_SetTransportConfig( DOFConnectionConfigBuilder self, DTLConnectionConfig config );

/**
    Set the @ref DOFProtocolNegotiator for the given DOFConnectionConfigBuilder.
    Please refer to @ref DOFConnectionConfigBuilder_Create for the default value for this setting.

    @param self The Connection Config Builder.  If @a self is NULL, no action will be taken.
    @param negotiator The @ref DOFProtocolNegotiator.  This must not be NULL.

    @retval TRUE Protocol negotiator was set.
    @exception FALSE Protocol negotiator set failed.
**/
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfigBuilder_SetProtocolNegotiator( DOFConnectionConfigBuilder self, DOFProtocolNegotiator negotiator );

/**
    Set the maximum time allowed to be silent before a packet must be sent for the given DOFConnectionConfigBuilder.
    Please refer to @ref DOFConnectionConfigBuilder_Create for the default value for this setting.

    @param self The Connection Config Builder.  If @a self is NULL, no action will be taken.
    @param maxSilence  The maximum silence time, in milliseconds.  This must be greater than 0.

    @retval TRUE Max send silence was set.
    @exception FALSE Failed to set max send silence.
**/
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfigBuilder_SetMaxSendSilence( DOFConnectionConfigBuilder self, uint32 maxSilence );

/**
    Set the maximum time allowed to be silent before a packet must be received for the given DOFConnectionConfigBuilder.
    Please refer to @ref DOFConnectionConfigBuilder_Create for the default value for this setting.

    @param self The Connection Config Builder.  If @a self is NULL, no action will be taken.
    @param maxSilence  The maximum silence time, in milliseconds.  This must be greater than 0.

    @retval TRUE Max receive silence was set.
    @exception FALSE Failed to set max receive silence.
**/
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfigBuilder_SetMaxReceiveSilence( DOFConnectionConfigBuilder self, uint32 maxSilence );

/**
    Set the Bridge in the DOFConnectionConfigBuilder.
    Please refer to @ref DOFConnectionConfigBuilder_Create for the default value for this setting.

    @param self The Connection Config Builder.  If @a self is NULL, no action will be taken.
    @param bridge The bridge.  If @a bridge is NULL, the default bridge is used.

    @retval TRUE The Bridge was set.
    @exception FALSE Failed to set the Bridge.
**/
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfigBuilder_SetBridge( DOFConnectionConfigBuilder self, DOFOperationBridgeConfig bridge );

/**
    Set the Receive Filter in the DOFConnectionConfigBuilder.
    Please refer to @ref DOFConnectionConfigBuilder_Create for the default value for this setting.

    @param self The Connection Config Builder.  If @a self is NULL, no action will be taken.
    @param filter The Filter.  If @a filter is NULL, the default filter is used.

    @retval TRUE The Filter was set.
    @exception FALSE Failed to set the Filter.
**/
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfigBuilder_SetReceiveFilter( DOFConnectionConfigBuilder self, DOFOperationFilter filter );

/**
    Set the Send Filter in the DOFConnectionConfigBuilder.
    Please refer to @ref DOFConnectionConfigBuilder_Create for the default value for this setting.

    @param self The Connection Config Builder.  If @a self is NULL, no action will be taken.
    @param filter The Filter.  If @a filter is NULL, the default filter is used.

    @retval TRUE The Filter was set.
    @exception FALSE Failed to set the Filter.
**/
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfigBuilder_SetSendFilter( DOFConnectionConfigBuilder self, DOFOperationFilter filter );

/**
    Set the Stream Request Listener in the DOFConnectionConfigBuilder.
    Please refer to @ref DOFConnectionConfigBuilder_Create for the default value for this setting.

    @param self The Connection Config Builder.  If @a self is NULL, no action will be taken.
    @param streamListener The DOFConnectionStreamListener.  if @a streamListener is NULL, the default is used.

    @return DOF_OAL_DECLSPEC boolean
**/
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfigBuilder_SetStreamRequestListener( DOFConnectionConfigBuilder self, DOFConnectionStreamRequestListener streamListener);

/**
   Destroy a DOFConnectionConfigBuilder instance.

   @param self The Connection Config Builder to destroy.  If @a self is NULL, no action will be taken.
**/
extern DOF_OAL_DECLSPEC void DOFConnectionConfigBuilder_Destroy( DOFConnectionConfigBuilder self );

/**
 * Build an immutable @ref DOFConnectionConfig from this DOFConnectionConfigBuilder instance.
 * @mustfree{DOFConnectionConfig_Destroy}
 *
 * @param self The Connection Config Builder.  This should not be NULL.
 * @return An immutable @ref DOFConnectionConfig.
 * @exception NULL is returned if the @ref DOFConnectionConfig cannot be created, or an invalid parameter was detected.
 */
extern DOF_OAL_DECLSPEC DOFConnectionConfig DOFConnectionConfigBuilder_Build( DOFConnectionConfigBuilder self );

/**
 * Build an immutable @ref DOFConnectionConfig from this DOFConnectionConfigBuilder instance, then destroy this DOFConnectionConfigBuilder instance.
 * This is equivalent to calling @ref DOFConnectionConfigBuilder_Build(@a self) immediately followed by @ref DOFConnectionConfigBuilder_Destroy(@a self).
 * @mustfree{DOFConnectionConfig_Destroy}
 *
 * @param self The Connection Config Builder.  This should not be NULL.
 * @return An immutable @ref DOFConnectionConfig.
 * @exception NULL is returned if the @ref DOFConnectionConfig cannot be created, or an invalid parameter was detected.  Regardless, the DOFConnectionConfigBuilder, if not NULL, is destroyed.
 */
extern DOF_OAL_DECLSPEC DOFConnectionConfig DOFConnectionConfigBuilder_BuildAndDestroy( DOFConnectionConfigBuilder self );

/** @} **/


/* These get functions return the reference to the internal copy.  They are NOT freed by the application. */

/**
   Get the type of Connection Type for the given DOFConnectionConfig.

   @param self The DOFConnectionConfig.  This must not be NULL.

   @return the Connection Type for the DOFConnectionConfig.
 */
extern DOF_OAL_DECLSPEC DOFConnectionType DOFConnectionConfig_GetConnectionType( const DOFConnectionConfig self );

/**
   Get the target Address for the given DOFConnectionConfig.

   @param self The DOFConnectionConfig.  This must not be NULL.

   @return The target Address for the DOFConnectionConfig.
 */
extern DOF_OAL_DECLSPEC DOFAddress DOFConnectionConfig_GetAddress( const DOFConnectionConfig self );

/**
   Get the SecurityDesire for the given DOFConnectionConfig.

   @param self The DOFConnectionConfig.  This must not be NULL.

   @return the Security desire for the DOFConnectionConfig.
 */
extern DOF_OAL_DECLSPEC DOFSecurityDesire DOFConnectionConfig_GetSecurityDesire( const DOFConnectionConfig self );

/**
   Get the Permissions for the given DOFConnectionConfig.

   @param self The DOFConnectionConfig.  This must not be NULL.

   @return The Permissions for the DOFConnectionConfig.
 */
extern DOF_OAL_DECLSPEC DOFPermissionSet DOFConnectionConfig_GetPermissions( const DOFConnectionConfig self );

/**
   Get the transport-specific configuration for the DOFConnectionConfig.

   @param self The DOFConnectionConfig.  This must not be NULL.

   @return The transport-specific configuration.
 */
extern DOF_OAL_DECLSPEC DTLConnectionConfig DOFConnectionConfig_GetTransportConfig( const DOFConnectionConfig self );

/**
   Get the current DOFProtocolNegotiator for the DOFConnectionConfig.

   @param self The DOFConnectionConfig.  This must not be NULL.
   @return The current negotiator being used.
**/
extern DOF_OAL_DECLSPEC DOFProtocolNegotiator DOFConnectionConfig_GetProtocolNegotiator( const DOFConnectionConfig self );

/**
   Get the current maximum send silence time in this DOFConnectionConfig.

   @param self The DOFConnectionConfig.  This must not be NULL.
   @return The current maximum send silence time, in milliseconds.
**/
extern DOF_OAL_DECLSPEC uint32 DOFConnectionConfig_GetMaxSendSilence( const DOFConnectionConfig self );

/**
   Get the current maximum receive silence time in this DOFConnectionConfig.

   @param self The DOFConnectionConfig.  This must not be NULL.
   @return The current maximum receive silence time, in milliseconds.
**/
extern DOF_OAL_DECLSPEC uint32 DOFConnectionConfig_GetMaxReceiveSilence( const DOFConnectionConfig self );

/**
   Get the credentials in this DOFConnectionConfig.

   @param self The DOFConnectionConfig.  This must not be NULL.
   @return The credentials held by the configuration.
**/
extern DOF_OAL_DECLSPEC DOFCredentials DOFConnectionConfig_GetCredentials( const DOFConnectionConfig self );

/**
 * Gets the domain wildcard credentials set in the DOFConnectionConfig.
 *
 * @param self The Connection Config.  This must not be NULL.
 *
 * @return The wildcard DOFCredentials for the given Connection Config.
 */
extern DOF_OAL_DECLSPEC DOFCredentials DOFConnectionConfig_GetDomainDiscoveryCredentials( const DOFConnectionConfig self );

/**
 * Gets the Operation Bridge set in the DOFConnectionConfig.
 *
 * @param self The Connection Config.  This must not be NULL.
 *
 * @return The DOFOperationBridgeConfig.
 */
extern DOF_OAL_DECLSPEC DOFOperationBridgeConfig DOFConnectionConfig_GetBridge( DOFConnectionConfig const self );

/**
 * Gets the Receive Filter set in the DOFConnectionConfig.
 *
 * @param self The Connection Config.  This must not be NULL.
 *
 * @return The DOFOperationFilter.
 */
extern DOF_OAL_DECLSPEC DOFOperationFilter DOFConnectionConfig_GetReceiveFilter( DOFConnectionConfig const self );

/**
 * Gets the Send Filter set in the DOFConnectionConfig.
 *
 * @param self The Connection Config.  This must not be NULL.
 *
 * @return The DOFOperationFilter.
 */
extern DOF_OAL_DECLSPEC DOFOperationFilter DOFConnectionConfig_GetSendFilter( DOFConnectionConfig const self );

/**
 * Gets the Stream Request Listner set in the DOFConnectionConfig.
 *
 * @param self The Connection Config.  This must not be NULL.
 *
 * @return the DOFConnectionStreamRequestListener.
 */
extern DOF_OAL_DECLSPEC DOFConnectionStreamRequestListener DOFConnectionConfig_GetStreamRequestListener( const DOFConnectionConfig self );

/**
 * Checks to see if the Connection Config is set up to allow permission extensions.
 *
 * @param self The Connection Config.  This must not be NULL.
 *
 * @return TRUE if permission extensions are allowed, FALSE otherwise.
 */
extern DOF_OAL_DECLSPEC boolean DOFConnectionConfig_IsPermissionsExtendAllowed( DOFConnectionConfig const self );

/**
   Destroy a DOFConnectionConfig instance.

   @param self The DOFConnectionConfig to destroy.  If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFConnectionConfig_Destroy( DOFConnectionConfig self );

/**
@}
**/


/** @defgroup DOFConnection DOFConnection
    @ingroup DOF_Framework

@{
**/

/**
   Destroy the DOFConnection and free all memory that was allocated for use by the connection.

   @param self  The connection to destroy.  If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFConnection_Destroy( DOFConnection self );


/**
 * Cause the Connection to become Connected, if possible.  If the Connection is already connected, then this call is satisfied immediately.
 * It is the responsibility of the caller to eventually call @ref DOFConnection_Disconnect when the connection is no longer desired.
 *
 * @note This call is synchronous, and waits for the Connection to be established (or fail) before returning control to
 *       the caller.
 *
 * @param self The Connection.  This must not be NULL.
 * @param timeout The maximum time to wait for the connection to
 *                be established (in milliseconds).
 * @exceptionparam
 *
 * @retval TRUE If the connection connected.
 * @exception FALSE If the connection failed to connect. If an exception was used it will give more detail of failure.
 */
extern DOF_OAL_DECLSPEC boolean DOFConnection_Connect( DOFConnection self, uint32 timeout, DOFException * pException );

/**
 * Cause the Connection to start connecting, if possible.  If the Connection is already connected, then this call
 * is satisfied immediately. It is the responsibility of the caller to eventually call @ref DOFConnection_Disconnect
 * when the connection is no longer desired.
 *
 * The callback is called to indicate that the connection either connected  successfully or failed.  Any state listeners
 * added to the connection will be notified if this causes the state to change, but are otherwise unrelated to the
 * callback specified here.
 *
 * @note This call is asynchronous, and does not wait for the Connection to be established before returning control to
 * the caller.
 * @note This connection request cannot be cancelled. It is resolved by eventual success, explicit failure, or timeout.
 * @note If the callback is notified of a successful connection, @ref DOFConnection_Disconnect can be called to
 * terminate the connection.
 *
 * @param self The Connection.  This must not be NULL.
 * @param timeout The maximum time to wait for the Connection to be established.
 * @param callback The Connection completion callback to be called when asynchronous operation finishes.  If NULL, no callback will be called, regardless of whether the connection succeeds or fails.
 * @param context The context, if any, associated with this connection request.
 *
 * @see DOFConnection_AddStateListener
 */
extern DOF_OAL_DECLSPEC void DOFConnection_BeginConnect( DOFConnection self, uint32 timeout, DOFConnectionConnectCallback callback, void * context );

/**
 * Cause the Connection to become Connected again, if possible.  If the Connection is already connected, then this call is satisfied immediately.
 * This call should only be used when a successfully established connection was terminated but the connection is still desired (@ref DOFConnection_Disconnect has not yet been called).
 *
 * @note This call is synchronous, and waits for the Connection to be established (or fail) before returning control to
 *       the caller.
 *
 * @param self The Connection.  This must not be NULL.
 * @param timeout The maximum time to wait for the connection to
 *                be established (in milliseconds).
 * @exceptionparam
 *
 * @retval TRUE If the connection reconnected.
 * @exception FALSE If the connection failed to reconnect. If an exception was used it will give more detail of failure.
 */
extern DOF_OAL_DECLSPEC boolean DOFConnection_Reconnect( DOFConnection self, uint32 timeout, DOFException * pException );

/**
 * Cause the Connection to start connecting again, if possible.  If the Connection is already connected, then this call is satisfied immediately.
 * This call should only be used when a successfully established connection was terminated but the connection is still desired (@ref DOFConnection_Disconnect has not yet been called).
 *
 * The callback is called to indicate that the connection either connected
 * successfully or failed.  Any state listeners added to the connection will be
 * notified if this causes the state to change, but are otherwise unrelated to
 * the callback specified here.
 *
 * @note This call is asynchronous, and does not wait for the Connection to be established before returning control to
 * the caller.
 *
 * @param self The Connection.  This must not be NULL.
 * @param timeout The maximum time to wait for the Connection to be established.
 * @param callback The Connection completion callback to be called when asynchronous operation finishes.  If NULL, no callback will be called, regardless of whether the connection succeeds or fails.
 * @param context The context, if any, associated with this connection request.
 *
 * @see DOFConnection_AddStateListener
 */
extern DOF_OAL_DECLSPEC void DOFConnection_BeginReconnect( DOFConnection self, uint32 timeout, DOFConnectionConnectCallback callback, void * context );

/**
   If connected then decrement the reference count associated with this DOFConnection, and if the result is zero then
   allow the Connection to close.

   @note Other DOFConnections may refer to the same underlying Connection and may have non-zero reference counts, in
   which case the Connection may not actually close.

   @param self The Connection.  This must not be NULL.
 */
extern DOF_OAL_DECLSPEC void DOFConnection_Disconnect( DOFConnection self );

/**
   Add a connection state listener to this DOFConnection. It will be called immediately with the current state of the
   connection.

   A listener must be unique among this connection's set of state listeners to be added successfully.

   @param self The DOFConnection.  This must not be NULL.
   @param listener The listener to be called whenever the connection state changes.  This must not be NULL.

   @retval TRUE if listener was successfully added to the connection.
   @exception FALSE if the listener could not be added to the connection, either because of an illegal argument or
                    because the listener is already added to the connection.

   @see DOFConnection_RemoveStateListener
 */
extern DOF_OAL_DECLSPEC boolean DOFConnection_AddStateListener( DOFConnection self, DOFConnectionStateListener listener );

/**
   Remove a connection state listener from this DOFConnection. It will no longer be called.

   @param self The DOFConnection.  This must not be NULL.
   @param listener The listener that was previously added.  If it is not found, no listeners are removed.  This should not be NULL.

   @see DOFConnection_AddStateListener
 */
extern DOF_OAL_DECLSPEC void DOFConnection_RemoveStateListener( DOFConnection self, DOFConnectionStateListener listener );

/**
   Add a connection peer listener to this DOFConnection.

   A listener must be unique among this connection's set of peer listeners to be added successfully. Only connections of
   type @ref DOFCONNECTIONTYPE_HUB can have a peer listener; attempts to add a peer listener to other types of
   connections will fail.

   @param self The Connection.  This must not be NULL.
   @param listener The listener to be called whenever the connection state changes.  This must not be NULL.

   @retval TRUE if listener was successfully added to the connection.
   @exception FALSE if the listener could not be added to the connection, either because of an illegal argument, because
                    the listener is already added to the connection, or because the connection is of the wrong type.

   @see DOFConnection_RemovePeerListener
 */
extern DOF_OAL_DECLSPEC boolean DOFConnection_AddPeerListener( DOFConnection self, DOFConnectionPeerListener listener );

/**
   Remove a connection peer listener from this DOFConnection. It will no longer be called.

   @param self The Connection.  This must not be NULL.
   @param listener The listener that was previously added.  If it is not found, no listeners are removed.  This should not be NULL.

   @see DOFConnection_AddPeerListener
 */
extern DOF_OAL_DECLSPEC void DOFConnection_RemovePeerListener( DOFConnection self, DOFConnectionPeerListener listener );

/**
   Return a snapshot of the current state of the Connection.

   @mustfree{DOFConnectionState_Destroy}

   @note The state of the Connection is dynamic and may change at any moment, due to a number of reasons.  It is usually
         best to track the state of a Connection through a state listener.

   @param self The Connection.  This must not be NULL.

   @return The snapshot of the current state of the Connection.

   @see DOFConnection_AddStateListener
 */
extern DOF_OAL_DECLSPEC DOFConnectionState DOFConnection_CreateState( const DOFConnection self );

/**
   Return the Server associated with the Connection.

   @note Not all Connections are related to Servers, so this may return NULL.

   @param self The Connection.  This must not be NULL.

   @return The associated @ref DOFServer, or NULL if there is no Server associated with the connection.
 */
extern DOF_OAL_DECLSPEC DOFServer DOFConnection_GetRelatedServer( const DOFConnection self );

/**
   Return the Connection associated with the Connection.

   @note Not all Connections are related to another Connection, so this may return NULL.

   @param self The Connection.  This must not be NULL.

   @return The associated @ref DOFConnection, or NULL if there is no Connection associated with the connection.
 */
extern DOF_OAL_DECLSPEC DOFConnection DOFConnection_GetRelatedConnection( const DOFConnection self );

/**
 * Create a connection associated with a connection, including the ability to create
 * a point connection (to a hub).  Point connections require an associated DOFConnection, so this
 * is the only method allowed to create those types of connections.  Other
 * types of connections rarely benefit by having an associated DOFConnection.
 *
 * @mustfree{DOFConnection_Destroy}
 *
 * @param self The Connection.  This must not be NULL.
 * @param config The configuration options of the connection.  This should not be NULL.
 *
 * @return The new DOFConnection.
 * @exception NULL is returned if a connection could not be created.
 */
extern DOF_OAL_DECLSPEC DOFConnection DOFConnection_CreateConnection( DOFConnection self, DOFConnectionConfig config );

/**
   Return whether the status of the Connection is connected or not.

   @param self The Connection.  This must not be NULL.
   @retval FALSE if the Connection is not connected.
   @retval TRUE if the Connection is connected.

   @see DOFConnectionState_IsConnected
 */
extern DOF_OAL_DECLSPEC boolean DOFConnection_IsConnected( const DOFConnection self );

/**
 * Adds Domain Credentials to a connection. Remove these credentials with @ref DOFConnection_RemoveDomain.
 * @note This may cause an AS route to be added for the given credentials.
 *
 * @param self The Connection to add the domain credentials to. Must not be NULL.
 * @param domain The DOFDomainConfig for the domain you wish to add. Must not be NULL.
 * @param timeout The amount of time to try and add the domain credentials.
 * @param exception [OUT] will hold an exception if one happens. May be NULL, if NULL no exception can be reported.
 *
 * @return boolean TRUE, if it was added.
 *                 FALSE, if it failed to be added.
 */
extern DOF_OAL_DECLSPEC boolean DOFConnection_AddDomain( DOFConnection self, DOFDomainConfig domain, uint32 timeout, DOFException* exception );

/**
 * Removes Domain Credentials from a connection. Removes only credential added with @ref DOFConnection_AddDomain.
 * @note This may cause an AS route to be removed.
 *
 * @param self The Connection to remove the domain credentials from. Must not be NULL.
 * @param domain The DOFDomainConfig for the domain you wish to remove. Must not be NULL.
 */
extern DOF_OAL_DECLSPEC void DOFConnection_RemoveDomain( DOFConnection self, DOFDomainConfig domain );

/**
 * Adds a Trusted Domain. This will cause the given connection to be a trusted route for the given domain.
 *
 * @param self The Connection that represents the route to the AS.
 * @param domain The domain for the trusted AS route.
 *
 * @return boolean TRUE, if it was added.
 *                 FALSE, if it failed to be added.
 */
extern DOF_OAL_DECLSPEC boolean DOFConnection_AddTrustedDomain( DOFConnection self, DOFObjectIDDomain domain );

/**
 * Removed a Trusted Domain that was added with @ref DOFConnection_AddTrustedDomain.
 *
 * @param self The connection that represents the route to the AS.
 * @param domain The domain for the trusted AS route.
 */
extern DOF_OAL_DECLSPEC void DOFConnection_RemoveTrustedDomain( DOFConnection self, DOFObjectIDDomain domain );


/** @defgroup DOFConnectionState DOFConnectionState

    Represents the instantaneous state of a DOFConnection.

@{
**/

/**
   Return whether the status of the Connection is connected or not.

   @param self The connection State.  This must not be NULL.
   @retval FALSE if the Connection is not connected.
   @retval TRUE if the Connection is connected.

   @see DOFConnection_CreateState
 */
extern DOF_OAL_DECLSPEC boolean DOFConnectionState_IsConnected( const DOFConnectionState self );

/**
   Return the active @ref DOFSecurityMode of the Connection.  This is only valid while the Connection is
   connected.  The Security Mode in effect may be different from that specified by the
   Connection configuration due to protocol negotiation at the time the Connection is established.

   @param self The Connection State.  This must not be NULL.

   @return The currently active Security Mode of the Connection.
 */
extern DOF_OAL_DECLSPEC DOFSecurityMode DOFConnectionState_GetSecurityMode( const DOFConnectionState self );

/**
   Return the direction of the Connection.

   @param self The Connection State.  This must not be NULL.

   @return The direction of the Connection.
 */
extern DOF_OAL_DECLSPEC DOFConnectionDirection DOFConnectionState_GetDirection( const DOFConnectionState self );

/**
   Return the DOFException associated with the connection state, if any.

   If an exception is returned, it indicates that the connection was closed
   unexpectedly, and typically describes why (e.g., closed from the other
   side, by the transport, or because of a communication error).  It is
   possible that multiple exceptions may occur, causing the connection to
   close.  In such cases, only the first exception is saved in the state.

   @param self The ConnectionState.  This must not be NULL.

   @return The DOFException, if any.

   @retval NULL is returned if no exception is associated with the state (e.g., the connection is still connected, or was closed normally by the application as a result of calling @ref DOFConnection_Disconnect or @ref DOFConnection_Destroy).
 */
extern DOF_OAL_DECLSPEC DOFException DOFConnectionState_GetException( const DOFConnectionState self );

/**
 * Return the DOFCredentials used by the peer node for authentication of the connection.
 * This represents the identity of the peer.
 *
 * This is only valid while the Connection is connected.
 *
 * @param self The Connection State.  This must not be NULL.
 *
 * @return The DOFCredentials used to authenticate the peer node on the connection.
 */
extern DOF_OAL_DECLSPEC DOFCredentials DOFConnectionState_GetPeerCredentials( const DOFConnectionState self );

/**
 * Return the DOFCredentials used by this node for authentication of the connection.
 * This represents the identity of the local node.
 * This is usually the same as the DOFCredentials in the DOFConnectionConfig,
 * but can be different if the connection were opened in a remote domain.
 *
 * This is only valid while the Connection is connected.
 *
 * @param self The Connection State.  This must not be NULL.
 *
 * @return The DOFCredentials used to authenticate the local node on the connection.
 */
extern DOF_OAL_DECLSPEC DOFCredentials DOFConnectionState_GetCredentials( const DOFConnectionState self );

/**
 * Create a connection based on a connection state.
 * This will create a new DOFConnection that is identical to one described by the state,
 * including any related server or connection and using the same security scope.
 *
 * @mustfree{DOFConnection_Destroy}
 *
 * @param self The Connection State.  This must not be NULL.
 *
 * @return The new DOFConnection, or NULL if one could not be created.
 */
extern DOF_OAL_DECLSPEC DOFConnection DOFConnectionState_CreateConnection( const DOFConnectionState self );

/**
 * Returns the number of bytes received on the associated connection at the point
 * the connection state was created.
 *
 * @param self The Connection State. This must not be NULL.
 *
 * @return uint64 The number of bytes received on the associated connection.
 */
extern DOF_OAL_DECLSPEC uint64 DOFConnectionState_GetReceiveByteCount( const DOFConnectionState self );

/**
 * Returns the number of bytes sent from the associated connection at the point
 * the connection state was created.
 *
 * @param self The Connection State. This must not be NULL.
 *
 * @return uint64 The number of bytes sent from associated connection.
 */
extern DOF_OAL_DECLSPEC uint64 DOFConnectionState_GetSendByteCount( const DOFConnectionState self );

/**
 * Returns the number of packets received on the associated connection at the point
 * the connection state was created.
 *
 * @param self The Connection State. This must not be NULL.
 *
 * @return uint64 The number of packets received on associated connection.
 */
extern DOF_OAL_DECLSPEC uint64 DOFConnectionState_GetReceivePacketCount( const DOFConnectionState self );

/**
 * Returns the number of packets sent from the associated connection at the point
 * the connection state was created.
 *
 * @param self The Connection State. This must not be NULL.
 *
 * @return uint64 The number of packets sent from associated connection.
 */
extern DOF_OAL_DECLSPEC uint64 DOFConnectionState_GetSendPacketCount( const DOFConnectionState self );

/**
   Get the Connection Type of the Connection.

   @param self The Connection State.  This must not be NULL.

   @return the Connection Type of the Connection.
 */
extern DOF_OAL_DECLSPEC DOFConnectionType DOFConnectionState_GetConnectionType( const DOFConnectionState self );

/**
   Get the target Address of the Connection.

   @param self The Connection State.  This must not be NULL.

   @return The Target Address for the Connection.
 */
extern DOF_OAL_DECLSPEC DOFAddress DOFConnectionState_GetAddress( const DOFConnectionState self );

/**
   Get the Security Desire of the Connection.

   @param self The Connection State.  This must not be NULL.

   @return The Security Desire for the Connection.
 */
extern DOF_OAL_DECLSPEC DOFSecurityDesire DOFConnectionState_GetSecurityDesire( const DOFConnectionState self );

/**
   Get the Permissions of the Connection.

   @param self The Connection State.  This must not be NULL.

   @return The Permissions for the Connection.
 */
extern DOF_OAL_DECLSPEC DOFPermissionSet DOFConnectionState_GetPermissions( const DOFConnectionState self );

/**
 * Checks to see if the Connection is set up to allow permission extensions.
 *
 * @param self The Connection State.  This must not be NULL.
 *
 * @return TRUE if permission extensions are allowed, FALSE otherwise.
 */
extern DOF_OAL_DECLSPEC boolean DOFConnectionState_IsPermissionsExtendAllowed( const DOFConnectionState self );

/**
   Get the transport-specific configuration of the Connection.

   @param self The Connection State.  This must not be NULL.

   @return The transport-specific configuration for the Connection.
 */
extern DOF_OAL_DECLSPEC DTLConnectionConfig DOFConnectionState_GetTransportConfig( const DOFConnectionState self );

/**
   Get the Protocol Negotiator of the Connection.

   @param self The Connection State.  This must not be NULL.

   @return the Protocol Negotiator for the Connection.
 */
extern DOF_OAL_DECLSPEC DOFProtocolNegotiator DOFConnectionState_GetProtocolNegotiator( const DOFConnectionState self );

/**
   Get the maximum send silence time of the Connection.

   @param self The Connection State.  This must not be NULL.

   @return The maximum send silence time for the Connection.
 */
extern DOF_OAL_DECLSPEC int DOFConnectionState_GetMaxSendSilence( const DOFConnectionState self );

/**
   Get the Maximum receive silence time of the Connection.

   @param self The Connection State.  This must not be NULL.

   @return The maximum receive silence time of the Connection.
 */
extern DOF_OAL_DECLSPEC int DOFConnectionState_GetMaxReceiveSilence( const DOFConnectionState self );

/**
 * Destroy the Connection State.
 *
 * @param self The Connection State.  If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFConnectionState_Destroy( DOFConnectionState self );

/**
@}
**/


/** @defgroup DOFConnectionStreamRequest DOFConnectionStreamRequest

@{
**/

/**
   Retrieve the @ref DOFConnection on which the request was made.

   @note The resulting DOFConnection is owned by the given DOFConnectionStreamRequest and valid only while the request
         is valid.

   @param self  The request. This must not be NULL.

   @return The @ref DOFConnection on which the request was made.
**/
extern DOF_OAL_DECLSPEC DOFConnection DOFConnectionStreamRequest_GetConnection( DOFConnectionStreamRequest self );

/**
   Retrieve the @ref DOFAddress to which the stream connection is requested.

   @note The resulting DOFAddress is owned by the given DOFConnectionStreamRequest and valid only while the request
         is valid.

   @param self  The request. This must not be NULL.

   @return The @ref DOFAddress to which the stream connection is requested.
**/
extern DOF_OAL_DECLSPEC DOFAddress DOFConnectionStreamRequest_GetTargetAddress( DOFConnectionStreamRequest self );

/**
@}
**/


/**
@}
**/


/** @defgroup DOFConnectionList DOFConnectionList
    @ingroup DOFConnection

@{
**/

/**
 * Release resources previously allocated while forming a DOFConnectionList.
 * @param  connectionList  The DOFConnectionList.  If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFConnectionList_Destroy(DOFConnectionList connectionList);

/**
@}
**/


#ifdef  __cplusplus
}
#endif


#endif /* DOF_CONNECTION_H */
