/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_OPERATIONCONTROL_H
#define DOF_OPERATIONCONTROL_H

#include <dof/core.h>
#include <dof/error.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DOFOperationControl DOFOperationControl
    @ingroup DOFOperation

    Specifies the control parameters of a Request Operation.

    This object is passed to certain request methods to change the default
    behavior of the request, such as the retry frequency, verbosity, and list
    of excluded Object IDs. A deep copy of the control is placed into each operation.

@{
**/

#define DOFOPERATIONCONTROL_DEFAULT_RETRY_PERIOD 3600000UL   /** Default retry period for an operation (1 hour). */
#define DOFOPERATIONCONTROL_RESPONDERS_ALL       0xFFFFU     /** Maximum allowed responders for an operation (all). **/
#define DOFOPERATIONCONTROL_HEURISTIC_ALL        0x7FFFU     /** Maximum heuristic value for an operation (all). **/

/*
**  "Public" Type and Method Declarations
*/

/* DOFOperationControl Methods */


/**
 * Create a new Operation Control for use with an Operation.
 *
 * @mustfree{DOFOperationControl_Destroy}
 *
 * @return      A new DOFOperationControl or NULL if an error occurs (usually only
 *              when out of memory).
 */
extern DOF_OAL_DECLSPEC DOFOperationControl DOFOperationControl_Create( void );

/**
 *  Create a new copy of an existing @ref DOFOperationControl.
 *  @mustfree{DOFOperationControl_Destroy}
 *  @param  self The @ref DOFOperationControl to be copied.  This should not be NULL.
 *  @return A new copy of the @ref DOFOperationControl.
 *  @exception  NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DOFOperationControl DOFOperationControl_Copy( const DOFOperationControl self );

/**
 * Destroy a DOFOperationControl.
 *
 * @note        After making this call, it should be assumed that @a self is
 *              invalid.
 *
 * @param self  A DOFOperationControl to destroy.  If NULL, this function will do
 *              nothing.
 */
extern DOF_OAL_DECLSPEC void DOFOperationControl_Destroy( DOFOperationControl self );

/**
 * Add a Provider to the list of Excluded Providers.
 *
 * Excluded Providers are not expected to respond to subsequent retries of the
 * operation.  Normally, this is done after a response is received from the
 * Provider to prevent that Provider from responding again on a retry (to
 * reduce traffic).  However, it may be used to prevent the Provider from
 * responding even to the initial request if the provider is already added to
 * the exclusion list prior to making the request.  There is no guarantee that
 * the Provider will not respond; the application must be prepared to receive
 * responses from nodes previously added to the exclusion list.  It is
 * permissible to add the same provider multiple times to the exclusion list.
 *
 * @param self     The operationControl for the operation you would like exclude
 *                 @a providerID from.  This must not be NULL.
 * @param providerID The ObjectID of some Provider that you do not want a
 *                   response from.  This should not be NULL.
 *
 * @retval Non-zero (@c TRUE) The Provider was successfully added to the exclusion list.
 * @exception Zero (@c FALSE) The Provider could not be added to the exclusion list.
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationControl_AddProviderExclusion( DOFOperationControl self, DOFObjectID providerID );

/** Remove all provider exclusions from a DOFOperationControl.
 *
 * @param self      The DOFOperationControl to clear.  This must not be NULL.
 */
extern DOF_OAL_DECLSPEC void DOFOperationControl_ClearProviderExclusions(DOFOperationControl self);

/**
 * Set whether or not a multicast or broadcast operation should automatically
 * handle exclusion of providers.  When enabled, responders to the request are
 * automatically told to no longer respond to the operation, in order to cut
 * down on traffic.  There is no guarantee that a responder will not respond
 * again; it is simply a hint to allow traffic reduction on the network.  By
 * default, auto-exclusion is enabled.
 *
 * Auto-exclusion does not apply to unicast operations, nor to any Subscribe,
 * Register, or Session operation.
 *
 * @param self          The operationControl.  This must not be NULL.
 * @param isAutoExclude Either non-zero (@c TRUE) to set automatic exclusion,
 *                      or zero (@c FALSE) to disable automatic exclusion.
 *
 * @return TRUE if auto exclude was set, FALSE otherwise.
 *
 * @sa  DOFOperationControl_IsAutoExclude
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationControl_SetAutoExclude( DOFOperationControl self, boolean isAutoExclude );

/**
 * Return whether or not the Operation Control is set to automatically handle exclusion of providers.
 *
 * @param self          The operationControl.  This must not be NULL.
 *
 * @retval Non-zero (@c TRUE) if automatic exclusion is set (on).
 * @retval Zero (@c FALSE) if automatic exclusion is not set (off).
 *
 * @sa  DOFOperationControl_SetAutoExclude
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationControl_IsAutoExclude( DOFOperationControl self );

/**
 * Set whether or a unicast request can be flooded when there are no known providers.  If there are no known providers
 * and this option is set to FALSE, the request will not be sent.  Multicast and broadcast requests are not affected by
 * this setting.  By default, this flag is TRUE.
 *
 * @param self      The operationControl.  This must not be NULL.
 * @param isFloodAllowed Either non-zero (@c TRUE) to allow the request to be flooded, or zero (@c FALSE) to prevent
 *                       flooding.
 *
 * @return TRUE if the flag was set as indicated, FALSE otherwise.
 *
 * @sa  DOFOperationControl_IsFloodAllowed
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationControl_SetFloodAllowed( DOFOperationControl self, boolean isFloodAllowed );

/**
 * Return whether or not the Operation Control allows a unicast request to be flooded when there are no known
 * providers.
 *
 * @param self  The operationControl.  This must not be NULL.
 *
 * @retval Non-zero (TRUE) if flooding is allowed.
 * @retval Zero (FALSE) if flooding is not allowed.
 *
 * @sa  DOFOperationControl_SetFloodAllowed
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationControl_IsFloodAllowed( DOFOperationControl self );

/**
 * Set a periodic retry period for a request operation.  The operation will be
 * retried at the specified period until the operation times out or completes.
 *
 * @note Either a retry period or a retry callback may be set, but not both.
 * The most recent call to either @ref DOFOperationControl_SetRetryPeriod or
 * @ref DOFOperationControl_SetRetryCallback will apply.
 *
 * @param self      A operationControl.  This must not be NULL.
 * @param period    The number of milliseconds between retries.
 *
 * @return TRUE if the retry period was set, FALSE otherwise.
 *
 * @sa  DOFOperationControl_SetRetryCallback
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationControl_SetRetryPeriod( DOFOperationControl self, uint32 period );

/**
 * Set a retry callback to dynamically determine each retry period.  Each time
 * an operation is sent, the specified callback is used to determine the time
 * until the next retry.  This allows the application dynamic control over
 * the retry behavior of the operation.
 *
 * @note Either a retry period or a retry callback may be set, but not both.
 * The most recent call to either @ref DOFOperationControl_SetRetryPeriod or
 * @ref DOFOperationControl_SetRetryCallback will apply.
 *
 * @param self     A operationControl.  This must not be NULL.
 * @param callback A user-defined callback that will determine if a retry
 *                 should take place.  This must not be NULL.
 *
 * @return TRUE if the retry callback was set, FALSE otherwise.
 *
 * @sa  DOFOperationControl_SetRetryPeriod
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationControl_SetRetryCallback( DOFOperationControl self, DOFOperationRetryCallback callback );

/**
 * Set the level of verbosity for responses desired in a DOFOperationControl.
 *
 * Refer to @ref DOFOperationResponseLevel for descriptions of the allowed levels.
 *
 * @param self      The operationControl.  This must not be NULL.
 * @param level     The desired level of response from providers.  This value
 *                  @b must be one of the possible levels listed for @ref
 *                  DOFOperationResponseLevel or undefined behavior will occur.
 *
 * @return TRUE if the response level was set, FALSE otherwise.
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationControl_SetResponseLevel( DOFOperationControl self, DOFOperationResponseLevel level );

/**
 * Get the level of verbosity for responses desired in a DOFOperationControl.
 *
 * @param self The operationControl.  This must not be NULL.
 *
 * @return The @ref DOFOperationResponseLevel.
 *
 * @exception DOFOPERATIONRESPONSELEVEL_NONE is returned if @a self is NULL.
 */
extern DOF_OAL_DECLSPEC DOFOperationResponseLevel DOFOperationControl_GetResponseLevel( DOFOperationControl self );

/**
 * Set the target connection to use for the operation.  The operation will only
 * go to the specified connection (all other connections will be ignored).  If
 * a request is sourced in a security scope that is incompatible with the security scope
 * of the connection, the request will not be sent.  Set the target connection
 * this DOFOperationControl will use.
 *
 * @param self          The operationControl.  This must not be NULL.
 * @param connection    The target connection to use.  If NULL, this operation
 *                      will use the normal routing logic for the operation
 *                      (no target connection will apply).
 *
 * @return TRUE if the target connection was set, FALSE otherwise.
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationControl_SetTargetConnection( DOFOperationControl self, DOFConnection connection );

/**
 * Get the target connection to use for the operation.
 *
 * @param self The operationControl.  This must not be NULL.
 *
 * @return The target @ref DOFConnection.  This may be NULL if there is no target connection (default routing applies).
 */
extern DOF_OAL_DECLSPEC DOFConnection DOFOperationControl_GetTargetConnection( DOFOperationControl self );

/**
 * Set whether or not a DOFOperation using this DOFOperationControl will
 * cause responses to be randomly distributed across the operation's lifetime.
 *
 * @param self          The operationControl.  This must not be NULL.
 * @param isDelay       Boolean to enable (if TRUE) or disable (if FALSE) the
 *                      delay.
 *
 * @return TRUE if the delay was set, FALSE otherwise.
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationControl_SetDelay( DOFOperationControl self, boolean isDelay );


/**
 * Set the maximum acceptable age of a cached value.  By default, no cached
 * values are allowed.
 *
 * @param self          The operationControl.  This must not be NULL.
 * @param cacheAge      The maximum acceptable age of a cached value (in
 *                      milliseconds).  A value of 0 means no cached values
 *                      will be accepted.  This value will be rounded up to the
 *                      nearest second.
 *
 * @return TRUE if the max cache age was set, FALSE otherwise.
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationControl_SetMaxCacheAge( DOFOperationControl self, uint32 cacheAge );

/**
 * Get the maximum acceptable age allowed by the DOFOperationControl.
 *
 * @param self          The operationControl.  This must not be NULL.
 *
 * @return The maximum acceptable cache age in milliseconds.
 *         A value of 0 means no cached values will be accepted.
 */
extern DOF_OAL_DECLSPEC uint32 DOFOperationControl_GetMaxCacheAge( DOFOperationControl self );

/**
 * Determine if a DOFOperationControl has delay behavior enabled or disabled.
 *
 * @param self          The operationControl.  This must not be NULL.
 *
 * @return              A boolean describing whether this operationControl
 *                      has delay behavior enabled or not.
 * @retval              TRUE @a self is valid and has delay behavior enabled.
 * @retval              FALSE @a self is valid and has delay behavior disabled.
 *
 * @exception FALSE is returned if @a self is NULL.
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationControl_IsDelay( DOFOperationControl self );

/**
 * Sets the maximum number of desired responders.  The library will limit the
 * number of responses to this count; even if more responses arrive, they will
 * be ignored and thus not passed to the application.
 *
 * This logic does not have any effect on the actual number of responses sent or on
 * the number of potential responders that take action based on the request.
 *
 * @param self          The operationControl.  This must not be NULL.
 * @param count         The total number of responses that are desired.  Valid
 *                      values are from 0 (no responses) to 65534, or 65535 to
 *                      leave responses unlimited.
 *
 * @return TRUE if the max response count was set, FALSE otherwise.
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationControl_SetMaxResponderCount( DOFOperationControl self, uint16 count );

/**
 * Get the maximum number of desired responders.
 *
 * @param self The operationControl.  This must not be NULL.
 *
 * @return The maximum number of desired responders (0-65534), or 65535 if there is no limit.
 */
extern DOF_OAL_DECLSPEC uint16 DOFOperationControl_GetMaxResponderCount( DOFOperationControl self );

/**
 * Enable randomized acceptance of the operation.  Rather than being
 * exact, this parameter causes a heuristic algorithm on the responder where each picks
 * a random number and compares it to the percentage. If their number is less than or equal
 * to the percentage then they will take action and respond.
 *
 * Since each responder determines whether to take action based solely on the percentage
 * and the random number, this method does not cause any network synchronization of
 * responders. It has an effect on both the total number of responders
 * that take action and on the total response count.
 *
 * @param self          The operationControl.  This must not be NULL.
 * @param heuristic     A heuristic, representing a percentage of receivers that should take action based on the request.
 *                      The valid range is from 0, representing 0% (or none), to 32767, representing 100% (or all).
 *                      The default is 32767 (all).
 *
 * @return TRUE if the heuristic was set, FALSE otherwise.
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationControl_SetHeuristic( DOFOperationControl self, uint16 heuristic );

/**
 * Get the heuristic.
 *
 * @param self  The operationControl.  This must not be NULL.
 *
 * @return The heuristic, in the range of 0 (representing 0% or none) to 32767 (representing 100% or all).
 */
extern DOF_OAL_DECLSPEC uint16 DOFOperationControl_GetHeuristic( DOFOperationControl self );

/**
 * Enable an acknowledgement timeout for the operation.  An operation is acknowledged
 * when a response is received to that operation (whether a normal response or an
 * exception).  For many operations, there is little difference between waiting for
 * the operation to complete vs. be acknowledged (and this timeout is therefore not
 * necessary).  However, some operations maintain state over time (such as sessions
 * or subscriptions).  These operations usually have large timeouts, but often need
 * to fail much sooner if they cannot be established.  The acknowledgement timeout
 * limits the time between when the request is made and when a response may be
 * accepted.  If a response is not received within this timeout period, the operation
 * is completed. The ack timeout is independant of other operation settings- operation
 * retry or changes to operation timeout do not override it.
 *
 * @param self    The operationControl.  This must not be NULL.
 * @param timeout The acknowledgement timeout, in milliseconds.
 *
 * @return TRUE if the ack timeout was set, FALSE otherwise.
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationControl_SetAckTimeout( DOFOperationControl self, uint32 timeout );

/**
 * Get the acknowledgement timeout.
 *
 * @param self  The operationControl.  This must not be NULL.
 *
 * @return The acknowledgement timeout, in milliseconds.
 */
extern DOF_OAL_DECLSPEC uint32 DOFOperationControl_GetAckTimeout( DOFOperationControl self );

/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* DOF_OPERATIONCONTROL_H */
