/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_CORE_H
#define DOF_CORE_H


/* System Dependent Library API */
#include <dof/pcr.h>

/* DOF Definitions */
#include <dof/dofdefs.h>

/* OAL Types */
#include <dof/oaltypes.h>

/* DOF */
#include <dof/dof.h>

/* Logging */
#include <dof/log.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DOFMem DOFMem
    @ingroup DOF

    The DOF Memory Allocator is responsible for allocating memory used
    internally by the C OAL.  It is made up of 3 main functions: @ref
    DOFMem_Alloc, @ref DOFMem_Free, and @ref DOFMem_Realloc.

    These functions can be replaced by a user by linking with user
    implementations of the functions, overriding the library versions.

@{
**/
#ifndef ENABLE_DOFALLOC
/* Use the PCR memory functions directly instead of allowing a separate DOF allocator. */
#define DOFMem_Alloc   PCRMem_Alloc
#define DOFMem_Free    PCRMem_Free
#define DOFMem_Realloc PCRMem_Realloc
#endif
#ifndef DOFMem_Alloc
/**
 * Allocate a block of memory.  The block must later by freed by @ref
 * DOFMem_Free. @mustfree{DOFMem_Free}
 *
 * @param size The size of the block of memory requested.  This must be greater than 0.
 *
 * @return A pointer to the new block of memory of the specified @a size, or
 * NULL if the request could not be satisfied.
 */
extern DOF_OAL_DECLSPEC void * DOFMem_Alloc( uint32 size );
#endif
#ifndef DOFMem_Realloc
/**
 * Reallocate a block of memory.  The block must have been allocated previously
 * by @ref DOFMem_Alloc.  When successful, the block of memory passed in must
 * no longer be referenced or freed; instead, the return value represents a new
 * block of the new size.  It is possible, but not guaranteed in any
 * circumstance, that the returned block may have the same pointer value as the
 * original block.  If the call is unsuccessful, NULL is returned and the
 * original block is left untouched (and must still be freed).
 *
 * The data contents of the original block are copied/moved to the new block,
 * up to either the given @a size or the size of the original block, whichever
 * is smaller.
 *
 * It is assumed that reallocating to the same or smaller size will always
 * succeed.
 *
 * @param mem The block of memory.  This must not be NULL.
 * @param size The new size of the allocated memory block.  It may be larger or smaller than the original allocated size.  This must be greater than 0.
 *
 * @return A pointer to the new block of memory of the specified @a size, or
 * NULL if the request could not be satisfied.
 */
extern DOF_OAL_DECLSPEC void * DOFMem_Realloc( void * mem, uint32 size );
#endif
#ifndef DOFMem_Free
/**
 * Free a block of memory.  The block must have been allocated previously by
 * @ref DOFMem_Alloc. If NULL, then no action is taken.
 *
 * @param mem The block of memory.
 */
extern DOF_OAL_DECLSPEC void DOFMem_Free( void * mem );
#endif
/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* DOF_CORE_H */
