/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_CREDENTIALS_H
#define DOF_CREDENTIALS_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif


/*
**  "Public" Type and Method Declarations
*/

/** @defgroup DOFCredentials DOFCredentials
    @ingroup DOF_Framework
@{
**/

/**
 *  Creates new credentials that use a key as the secret.
 *
 *  @param  domainID  The domain in which the @a identity exists.  If unknown, use DOFOBJECTIDDOMAIN_BROADCAST.  This should not be NULL.
 *  @param  identity  The Object ID representing the authentication identity.  This should not be NULL.
 *  @param  key       The secret to be used with the identity for authentication.  This should not be NULL.
 *
 *  @return A newly created @ref DOFCredentials.
 *  @exception NULL is returned if an invalid parameter was detected, or the credentials could not be created.
 *  @mustfree{DOFCredentials_Destroy}
 */
extern DOF_OAL_DECLSPEC DOFCredentials DOFCredentialsKey_Create( const DOFObjectIDDomain domainID, const DOFObjectIDAuthentication identity, const uint8 key[32] );

/**
 * Creates new credentials that use a password as the secret.
 *
 * @param domainID  The domain in which the @a identity exists. If unknown, use DOFOBJECTIDDOMAIN_BROADCAST.  This should not be NULL.
 * @param identity  The object ID representing the authentication identity. This should not be NULL.
 * @param password  A NULL-terminated password string to be used as the identity's authentication secret. This should not be NULL.
 *
 * @return A newly created @ref DOFCredentials
 * @exception NULL is returned if an invalid parameter was detected, or the credentials could not be created.
 * @mustfree{DOFCredentials_Destroy}
 */
extern DOF_OAL_DECLSPEC DOFCredentials DOFCredentialsPassword_Create( const DOFObjectIDDomain domainID, const DOFObjectIDAuthentication identity, const char * password );


/**
 * Creates a new credentials that will use the given credentials for resolution in the given domain.
 *
 * @param creds The base credentials
 * @param domainID The domain.  If unknown, use DOFOBJECTIDDOMAIN_BROADCAST.  This should not be NULL.
 *
 * @return DOFCredentials A newly created @ref DOFCredentials
 * @exception NULL is returned if an invalid parameter was detected, or the credentials could not be created.
 * @mustfree{DOFCredentials_Destroy}
 */
extern DOF_OAL_DECLSPEC DOFCredentials DOFCredentials_Create( DOFCredentials creds, DOFObjectIDDomain domainID );

/**
 *  Creates a new copy of an existing @ref DOFCredentials.
 *  @mustfree{DOFCredentials_Destroy}
 *  @param  self The @ref DOFCredentials to be copied. This should not be NULL.
 *  @return A new copy of the @ref DOFCredentials.
 *  @exception  NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DOFCredentials DOFCredentials_Copy( const DOFCredentials self );

/**
 *  Destroys a @ref DOFCredentials object.
 *
 *  @param  self  The @ref DOFCredentials to be destroyed. If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFCredentials_Destroy( DOFCredentials self );

/**
 *  Retrieves the type of credentials represented by the given @ref DOFCredentials.
 *
 *  @param  self  The @ref DOFCredentials to be queried. This must not be NULL.
 *
 *  @return The type of credentials.
 *
 *  @see DOFCredentialsType
 */
extern DOF_OAL_DECLSPEC DOFCredentialsType DOFCredentials_GetType( const DOFCredentials self );

/**
 *  Returns the domain in which the credentials are valid.  This may return the broadcast ID
 *  if the credentials were created without a specific domain id and have not yet been resolved.
 *  After the credentials are resolved (@ref DOFCredentials_IsResolved has returned true), this
 *  will always return a specific domain ID.
 *
 *  @param  self  The @ref DOFCredentials.  This must not be NULL.
 *
 *  @return The domain of the credentials, which may be BROADCAST if the user did not specify a domain and the credentials are not yet resolved.
 *
 *  @see DOF_ResolveCredentials
 */
extern DOF_OAL_DECLSPEC DOFObjectIDDomain DOFCredentials_GetDomainID( const DOFCredentials self );

/**
 * Returns the identity of the authorized user.  This call is only guaranteed
 * to provide a valid result after the credentials are resolved (@ref DOFCredentials_IsResolved returns
 * true), and it is not permitted that the result change after that
 * point.  However, it may return a valid result earlier.
 *
 * @param self The credentials.  This must not be NULL.
 *
 * @return The authenticated user identity, or possibly null if called before the credentials are resolved.
 *
 * @see DOF_ResolveCredentials
 */
extern DOF_OAL_DECLSPEC DOFObjectIDAuthentication DOFCredentials_GetIdentity( const DOFCredentials self );

/**
 * Returns the storage key used by the credentials for lookup.  The storage key is the identifying data known before the
 * credentials are resolved, typically used by the AS to lookup the credential for resolution.  In many cases, this is
 * simply the authentication ID (identity) of the credentials, but some forms of credentials may not know the identity
 * until after resolution completes (e.g., a fingerprint).
 *
 * @note The data returned by this call is valid only for the lifetime of the credentials. No access to this data is
 *       permitted after the credentials have been destroyed.
 *
 * @param self The credentials.  This must not be NULL.
 *
 * @return The credentials' storage key.
 */
extern DOF_OAL_DECLSPEC const uint8 * DOFCredentials_GetStorageKey( const DOFCredentials self );

/**
 * Returns the size of the storage key used by the credentials.
 *
 * @param self The credentials.  This must not be NULL.
 *
 * @return The credentials' storage key.
 *
 * @sa DOFCredentials_GetStorageKey
 */
extern DOF_OAL_DECLSPEC uint32 DOFCredentials_GetStorageKeySize( const DOFCredentials self );

/**
 * Reports whether or not the credentials are resolved, meaning that they contain a valid (not Broadcast) domain ID, valid identity, and a 256-bit shared secret.
 *
 * @param self The credentials.  This must not be NULL.
 *
 * @retval Zero (FALSE) if the credentials are resolved.
 * @retval Non-Zero (TRUE) if the credentials are not yet resvoled.
 */
extern DOF_OAL_DECLSPEC boolean DOFCredentials_IsResolved( const DOFCredentials self );

/**
 * Returns the byte array that can be stored into an authentication server datastore and used
 * to verify the credentials during authentication.  This can only be called on a DOFCredentials
 * that has been constructed.
 *
 * <b>This data must be kept private, as it contains secret data associated with the credentials. It must be
 * protected just as the secrets themselves. It must never be passed in the clear or made public.</b>
 *
 * @param self The credentials.  This must not be NULL.
 *
 * @return The private storage byte array.
 * @exception NULL is returned if the credentials do not contain the private storage information.
 *
 * @see DOFCredentials_GetPrivateStorageSize
 */
extern DOF_OAL_DECLSPEC const uint8 * DOFCredentials_GetPrivateStorage( const DOFCredentials self );

/**
 * Returns size of the private storage byte array.  This can only be called on a DOFCredentials
 * that has been constructed.
 *
 * @param self The credentials.  This must not be NULL.
 *
 * @return The size of the private storage byte array.
 * @exception 0 is returned if the credentials do not contain the private storage information.
 *
 * @see DOFCredentials_GetPrivateStorage
 */
extern DOF_OAL_DECLSPEC uint32 DOFCredentials_GetPrivateStorageSize( const DOFCredentials self );

/**
 * Sets the private storage from the AS storage
 *
 * @param self The credentials.  This must not be NULL.
 * @param dataSize The size of the data buffer passed.
 * @param data The data buffer.  This must not be NULL if @a dataSize > 0.
 *
 * @return TRUE if the private storage was set, FALSE otherwise.
 */
extern DOF_OAL_DECLSPEC boolean DOFCredentials_SetPrivateStorage( const DOFCredentials self, uint32 dataSize, const uint8 * data );

/**
 * Gets the length of a byte array for the given DOFCredentials. This represents the way the DOFCredentials would be stored if passed 
 * into the @ref DOFCredentials_CreateBytes function.
 * 
 * @param self The credentials. Should not be NULL.
 * 
 * @return uint32 The byte length of the credentials.
 */
extern DOF_OAL_DECLSPEC uint32 DOFCredentials_GetByteSize(DOFCredentials self);

/**
 * Creates a byte array for the given DOFCredentials that can be used to recreate it at a later time.
 * 
 * @note This data must be kept private, as it contains secret data associated with the credential. It must be protected 
 * just as the secrets themselves. It must never be passed in the clear or made public.
 * 
 * @param self The credentials to create the bytes from. This should not be NULL.
 * 
 * @return uint8* The byte array for the credentials. 
 * @exception NULL will be returned if the bytes could not be created. 
 * @mustfree{PCRMem_Free}  
 */
extern DOF_OAL_DECLSPEC uint8* DOFCredentials_CreateBytes( DOFCredentials self );

/**
 * Creates a DOFCredentials object from a properly formatted byte array.
 * 
 * @param size The byte length of the buffer. 
 * @param buffer The buffer containing the credentials. Should not be NULL.
 * 
 * @return DOFCredentials The credentials represented by the bytes in the buffer.
 * @exception NULL will be returned if the credentials could not be created, or the size doesn't match the exact credentials length. 
 * @mustfree{DOFCredentials_Destroy} 
 * @see DOFCredentials_CreateBytes
 */
extern DOF_OAL_DECLSPEC DOFCredentials DOFCredentials_Create_Bytes(uint32 size, uint8 * buffer);

/**
@}
**/


#ifdef  __cplusplus
}
#endif

#endif /* DOF_CREDENTIALS_H */
