/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_DOF_H
#define DOF_DOF_H


#include <dof/core.h>
#include <dof/cfg.h>


#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DOF DOF
    @ingroup DOF_Framework

    The DOF class represents a DOF networking node. Every program that uses
    DOF requires a DOF instance. The DOF controls routing between
    the components that form the basis of DOF connectivity: systems (@ref DOFSystem),
    servers (@ref DOFServer), and connections (@ref DOFConnection). While systems
    initiate and process operations, it is the DOF that determines where to route
    those operations (which other systems, connections, or servers associated with
    that DOF will receive the traffic). The DOF uses operation state,
    routing tables, security constraints, and similar information to make this determination.

    Because each DOF represents a separate DOF networking node, even
    when two DOFs reside in the same application, each is a
    separate logical node in the DOF network. So for example, while security-compatible
    systems on the same DOF can communicate directly, systems on
    separate DOFs require connections and servers to
    communicate&mdash;even if both DOFs are in the same application.

    The DOF is created using @ref DOF_Create and is initialized during creation.
    Some configuration parameters can be changed after creation to modify
    the behavior of an existing DOF and any objects created by it (for example,
    see @ref DOF_SetMaxConnections). However, changing other configuration parameters associated with
    the DOF requires taking the node down and building a new DOF with a new configuration.
    For information about all configuration parameters, see the functions of @ref DOFConfigBuilder.

    The DOF is used to create:
    - @ref DOFSystem with @ref DOF_CreateSystem
    - @ref DOFConnection with @ref DOF_CreateConnection
    - @ref DOFServer with @ref DOF_CreateServer
    - @ref DOFDomain with @ref DOF_CreateDomain
    - @ref DOFAuthenticator with @ref DOF_CreateAuthenticator

    Each DOF uses several threads internally. These are started when the DOF is created.

    The library itself must be initialized before creating a DOF instance,
    using @ref DOF_Init.
@{
**/

/*
**  "Public" Type and Method Declarations
*/


/** @name Initialization and Shutdown

    The Initialization and Shutdown module provides functions to initialize or shut down the library as a whole. These functions are included only with the C OAL. Other OALs do not require them.

@{
**/

/**
    Initializes the C Object Access Library.
    The DOF_Init function must be called before any other portion of the library can be used.

    @see DOF_Shutdown
**/
extern DOF_OAL_DECLSPEC void DOF_Init( void );

/**
    Terminates use of the C Object Access Library.
    The DOF_Shutdown function must be called when the library will no longer be used.

    @see DOF_Init
**/
extern DOF_OAL_DECLSPEC void DOF_Shutdown( void );


/**
@}
**/

/**
 * Creates a DOF instance that uses the specified configuration.
 *
 * @limitations The C OAL currently supports only a single DOF per process.
 * The DOF_Create function returns @c NULL if a DOF instance has already been
 * created.
 *
 * @mustfree{DOF_Destroy}
 *
 * @param config An instance of @ref DOFConfig, which represents a source of configuration parameters for the DOF. Use
 *               @ref DOFConfigBuilder to create an instance of DOFConfig. You may pass a @c NULL argument to use the
 *               default configuration.
 *
 * @return An instance of @ref DOF.
 * @exception NULL is returned if a DOF instance could not be created for any reason, including that invalid arguments were used in the configuration.
 *
 * @see DOF_Destroy
 * @see DOF_Init
 */
extern DOF_OAL_DECLSPEC DOF DOF_Create( DOFConfig config );

/**
 * Frees resources allocated during the creation of a DOF. This call must be
 * made for each created DOF when the DOF is no longer required&mdash;usually
 * during program termination.
 *
 * In addition to deallocating memory, calls to _Destroy functions in the API stop threads and
 * attempt to clean up state associated with objects. When destroying a DOF, you should follow these steps:
 *
    -# Call DOF_SetNodeDown.
    -# Destroy all components other than the DOF that have not been destroyed previously. This may include multiple calls to any or all of the following:
        - @ref DOFConnection_Destroy
        - @ref DOFServer_Destroy
        - @ref DOFSystem_Destroy
        - @ref DOFDomain_Destroy
        - @ref DOFAuthenticator_Destroy
    -# Finally, call DOF_Destroy.
 *
 * The DOF_Destroy function runs asynchronously. There is no guarantee that the DOF has finished being
 * destroyed when this function returns. Any attempt to use a destroyed DOF will result in undefined behavior.
 *
 * @param self The DOF instance to destroy. If the value is @c NULL, no action is taken.
 *
 * @see DOF_Create
 * @see DOF_Shutdown
 */
extern DOF_OAL_DECLSPEC void DOF_Destroy( DOF self );

/**
 * Creates a @ref DOFState, which represents a snapshot of the DOF's configuration information as it
 * exists at the moment when you call this function. After using this function to get the snapshot,
 * you can use the getter methods of DOFState to retrieve specific information about the DOF's
 * configuration, as well as some runtime information. Because the DOF's configuration information and
 * runtime statistics may change, you should call this function again each time you want updated information.
 *
 * @mustfree{DOFState_Destroy}
 *
 * @param self The DOF instance you want configuration information for. This must not be @c NULL.
 * @return A DOFState that represents the current state of the DOF.
 * @exception NULL is returned if a DOFState could not be returned.
 */
extern DOF_OAL_DECLSPEC DOFState DOF_CreateState( const DOF self );

/**
 * Creates a @ref DOFSystem that uses the specified configuration. You must create DOFSystem instances in order to create @ref
 * DOFObject instances, which are needed to perform the majority of operations DOF functionality provides. Each system created must be
 * destroyed when it is no longer needed.
 *
 * You can create a DOFSystem that uses a default configuration by passing <code>( self, NULL, 0, NULL )</code> in this function.
 * For a full list of DOFSystem defaults, see @ref DOFSystemConfigBuilder_Create.
 *
 * @mustfree{DOFSystem_Destroy}
 *
 * @param self The DOF instance to associate the system with. This must not be @c NULL.
 * @param config An instance of @ref DOFSystemConfig that represents a source of configuration parameters for the DOFSystem.
 *               Use @ref DOFSystemConfigBuilder to create an instance of DOFSystemConfig. You may pass a @c NULL argument to
 *               use the default configuration.
 * @param timeout Controls how long (in milliseconds) to wait for authentication when credentials have been set in the configuration.
 *                If the configuration does not include credentials, the timeout may be zero.
 * @exceptionparam
 * @return The newly created DOFSystem, associated with the DOF and using
 *         the specified configuration.
 * @exception NULL is returned and @a pException is set if the DOF cannot accept the new DOFSystem, credentials are set in the
 *            configuration that cannot be authenticated, or the DOFSystem could not be created for other reasons.
 */
extern DOF_OAL_DECLSPEC DOFSystem DOF_CreateSystem( DOF self, DOFSystemConfig config, uint32 timeout, DOFException * pException );

/**
 * Creates a @ref DOFConnection that uses the specified configuration. The connection is associated with the DOF
 * instance that creates it, and only that DOF can route operations through it (see @ref DOF).
 *
 * This function may be used to create the following types of connections:
 * - @ref DOFCONNECTIONTYPE_STREAM
 * - @ref DOFCONNECTIONTYPE_DATAGRAM
 * - @ref DOFCONNECTIONTYPE_DATAGRAM_STATELESS
 *
 * This function may not be used to create the following types of connections:
 * - @ref DOFCONNECTIONTYPE_GROUP. Use @ref DOFServer_CreateConnection.
 * - @ref DOFCONNECTIONTYPE_HUB. Use @ref DOFServer_CreateConnection.
 * - @ref DOFCONNECTIONTYPE_POINT. Use @ref DOFConnection_CreateConnection
 *
 * For more information about these connection types, see @ref DOFConnectionType.
 *
 * Connections do not connect automatically when created. After creating the connection, call @ref DOFConnection_Connect
 * or @ref DOFConnection_BeginConnect. To disconnect temporarily, call @ref DOFConnection_Disconnect. Disconnected
 * connections can be reconnected using DOFConnection_Connect or DOFConnection_BeginConnect. When the DOFConnection
 * is no longer needed, it must be destroyed. See @ref DOFConnection_Destroy. Destroyed connections cannot be reconnected.
 *
 * The internal library tracks references to duplicate connections, so this function can safely be called multiple times with the
 * same configuration without consuming extra resources. Such duplicate connections are frequently needed, especially by nodes
 * that create connections dynamically in response to other function calls.
 *
 * @mustfree{DOFConnection_Destroy}
 *
 * @param self The DOF instance to use to create the DOFConnection. This must not be @c NULL.
 * @param config An instance of @ref DOFConnectionConfig that represents a source of configuration parameters for the DOFConnection.
 *               Use @ref DOFConnectionConfigBuilder to create an instance of DOFConnectionConfig. This should not be @c NULL.
 * @return A new instance of DOFConnection.
 * @exception NULL is returned if the configuration is invalid, the DOF cannot accept the new DOFConnection,
 *            or the DOFConnection could not be created.
 * @see DOF_Close
 * @see DOFOperationControl_SetTargetConnection
 */
extern DOF_OAL_DECLSPEC DOFConnection DOF_CreateConnection( DOF self, DOFConnectionConfig config );

/**
 * Creates a @ref DOFServer that uses the specified configuration. The new server is associated with the DOF
 * instance that creates it, and only the DOF can accept connections or route operations through it (see @ref DOF).
 *
 * Servers do not start automatically when created. After creating the server, call @ref DOFServer_Start or
 * @ref DOFServer_BeginStart. To stop the server temporarily, call @ref DOFServer_Stop. Stopped servers can be restarted
 * using DOFServer_Start or DOFServer_BeginStart. Each DOFServer instance must be destroyed when it is no longer needed.
 * See @ref DOFServer_Destroy. Destroyed servers cannot be restarted.
 *
 * The internal library tracks references to duplicate servers, so this method can safely be called multiple times with the
 * same configuration without consuming extra resources. Such duplicate servers are frequently needed, especially by nodes
 * that create servers dynamically in response to other method calls.
 *
 * This function may be used to create unicast servers, but not to create multicast servers. Use @ref DOFServer_CreateServer
 * to create multicast servers.
 *
 * @mustfree{DOFServer_Destroy}
 *
 * @param self The DOF instance to use to create the DOFServer. This must not be @c NULL.
 * @param config An instance of @ref DOFServerConfig that represents a source of configuration parameters for the DOFServer.
 *               Use @ref DOFServerConfigBuilder to create an instance of DOFServerConfig. This should not be @c NULL and must
 *               not have a multicast address.
 * @return A new instance of DOFServer.
 * @exception NULL is returned if the configuration is invalid, the DOF cannot accept the new DOFServer,
 *            or the DOFServer could not be created.
 * @see DOF_Stop
 */
extern DOF_OAL_DECLSPEC DOFServer DOF_CreateServer( DOF self, DOFServerConfig config );

/**
 * Notifies the network that a DOF is about to be destroyed. It is beneficial for
 * network efficiency to notify other nodes before destroying a DOF, because it prevents state associated with the
 * DOF's operations on other nodes from entering a source-lost state, thereby preventing network communication related to
 * attempts to recover the lost resources. This function should be called only before destroying components and using the
 * @ref DOF_Destroy function. Continuing to send DOF operations to the rest of the network after sending a
 * DOF_SetNodeDown call will cause your DOF's state to be out of synch with other nodes and may result
 * in unpredictable behavior.
 *
 * You should always call DOF_SetNodeDown at the beginning of application cleanup for a DOF. You should @a not call
 * DOF_SetNodeDown before destroying any other components unless you intend to also destroy the DOF.
 * DOF_SetNodeDown does not replace DOF_Destroy. You must still call DOF_Destroy after calling DOF_SetNodeDown.
 * DOF_SetNodeDown does not operate as a way to stop the DOF temporarily. The DOF cannot be restarted
 * after making this call.
 * DOF_SetNodeDown will set the connection limit for the DOF to zero so that no new inbound or outbound connections
 * will be allowed to connect.
 *
 * @param self The DOF instance that is about to be destroyed. This must not be @c NULL.
 *
 * @retval TRUE if the node was set down.
 * @exception FALSE is returned if the function fails to set the node down.
 */
extern DOF_OAL_DECLSPEC boolean DOF_SetNodeDown( DOF self );

/**
 * @deprecated The ability to set a maximum number of operations was removed.  This function no
 * longer has any effect. The maximum number of simultaneous operations that the DOF supports is 0xFFFFFFFF (effectively unlimited).
 *
 * @param self the DOF to set the maximum number of operations in. This must not be @c NULL.
 * @param max Any uint32. The maximum number of operations will remain set to 0xFFFFFFFF.
 *
 * @retval TRUE in cases where this function would previously have set the value, although the DOF remains unchanged.
 * @exception FALSE if the arguments are invalid. The DOF remains unchanged.
 */ 
extern DOF_OAL_DECLSPEC boolean DOF_SetMaxOperations( DOF self, uint32 max );

/**
 * Changes the maximum number of simultaneous connections the DOF supports. The value is
 * originally set by using the default configuration or the @ref DOFConfigBuilder_SetMaxConnections function.
 * Use this function to change the original value. 
 * 
 * Because this is a runtime change, the number of currently established connections may already exceed 
 * the value you set in this function. If this occurs, it will not cause the node to drop any existing connections; 
 * however, the node will be unable to initiate or accept any new connections until the total number of connections 
 * drops below the new maximum.
 * 
 * See the DOFConfigBuilder function for more information.
 *
 * @deprecated This functionality was replaced by @ref DOF_SetConnectionLimit.
 *
 * @param self The DOF whose maximum number of connections will be changed. This must not be @c NULL.
 * @param max The maximum number of connections. A value of zero means there is no limit.
 *
 * @retval TRUE if the maximum number of connections was set.
 * @exception FALSE is returned if the function fails to set the maximum number of connections.
 * @see DOFConfigBuilder_Create
 */
extern DOF_OAL_DECLSPEC boolean DOF_SetMaxConnections( DOF self, uint16 max );

/**
 * Changes the maximum number of simultaneous connections the DOF supports. The value is
 * originally set by using the default configuration or the @ref DOFConfigBuilder_SetConnectionLimit function.
 * Use this function to change the original value.
 *
 * Because this is a runtime change, the number of currently established connections may already exceed
 * the value you set in this function. If this occurs, it will not cause the node to drop any existing connections;
 * however, the node will be unable to initiate or accept any new connections until the total number of connections
 * drops below the new maximum.
 *
 * See the DOFConfigBuilder function for more information.
 *
 * @param self The DOF whose maximum number of connections will be changed. This must not be @c NULL.
 * @param max The maximum number of connections. A value of zero will prevent all connections.
 *
 * @retval TRUE if the maximum number of connections was set.
 * @exception FALSE is returned if the function fails to set the maximum number of connections.
 * @see DOFConfigBuilder_Create
 * @see DOFConfigBuilder_SetConnectionLimit
 */
extern DOF_OAL_DECLSPEC boolean DOF_SetConnectionLimit( DOF self, uint32 max );

/**
 * Gets this DOF's runtime, which contains data about the internal runtime state of the DOF. It provides information
 * about current connections, servers, and traffic statistics.
 *
 * After getting the @ref DOFRuntime, use its functions to retrieve specific information. For example,
 * the following lines of code would retrieve a list of all connections created by myDof:
 *
 * <pre>DOFRuntime runtime = DOF_GetRuntime (myDof);
 * DOFConnectionList myConnections = DOFRuntime_GetConnections (runtime);
 * ...
 * DOFConnectionList_Destroy(myConnections);</pre>
 *
 * See the functions of @ref DOFRuntime for information about all data that can be retrieved.
 *
 * @param self The DOF whose runtime information will be retrieved. This must not be @c NULL.
 *
 * @return A DOFRuntime.
 */
extern DOF_OAL_DECLSPEC DOFRuntime DOF_GetRuntime( DOF self );

/**
 * Attempts to resolve the given credentials.
 *
 * @param self The DOF to perform the resolution under. This must not be @c NULL.
 * @param credentials The DOFCredentials to resolve. This must not be @c NULL.
 * @param timeout The amount of time to wait for the credentials to be resolved.
 * @exceptionparam
 *
 * @retval TRUE if the credentials were resolved in the time given.
 * @exception FALSE is returned if the credentials failed to be resolved in the time given.
 *            If @em pException is not @c NULL, the returned exception will give more details of the failure.
 */
extern DOF_OAL_DECLSPEC boolean DOF_ResolveCredentials( DOF self, DOFCredentials credentials, int timeout, DOFException* pException );

/**
 * Returns a DOFAuthenticator known to the DOF.
 *
 * @param self The DOF. This must not be @c NULL.
 * @param domainID The domain ID. This should not be @c NULL.
 *
 * @return The DOFAuthenticator known to the DOF.
 * @exception NULL is returned if the authenticator is not known to the DOF, or an invalid parameter was detected.
 */
extern DOF_OAL_DECLSPEC DOFAuthenticator DOF_GetAuthenticator( const DOF self, DOFObjectIDDomain domainID );

/**
 * Gets a list of all systems associated with the DOF.
 *
 * @mustfree{DOFSystemList_Destroy}
 *
 * @param self The DOF. Must not be NULL.
 *
 * @return A list of all systems.
 * @see DOFSystemList
 */
extern DOF_OAL_DECLSPEC DOFSystemList DOF_GetSystems( const DOF self );

/**
 * Cancels all outstanding operations sourced by a DOF. This function causes all connected nodes
 * to remove any existing operations that use the DOF's source identifier. Any DOF that sets a static source identifier in its
 * configuration using @ref DOFConfigBuilder_SetSourceID should include this call in its startup routine to prevent
 * sourcing new operations that have the same identifier as operations sourced by a previous instance of the DOF. This helps a DOF
 * resynchronize its state with other nodes if it crashes and is brought back online.
 *
 * This function should be called after establishing communication with other nodes but before sending any other operations.
 * It is recommended that after calling this function, you wait before sending other operations since delivery order is
 * not guaranteed.
 *
 * This function does not cancel any existing local operations.
 *
 * @param self The DOF whose operations will be canceled. This must not be @c NULL.
 */
extern DOF_OAL_DECLSPEC void DOF_CancelAllOperations( DOF self );

/**
 * Forces a connection to close. This function can be used to forcibly disconnect a connection, even if multiple references to
 * the connection exist. In contrast, @ref DOFConnection_Disconnect disconnects the connection only when the number of references
 * to disconnect calls is equal to the number of references to @ref DOFConnection_Connect and @ref DOFConnection_BeginConnect calls.
 *
 * @param self The DOF associated with the connection. This must not be @c NULL.
 * @param connection The connection to close. If @c NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOF_Close( DOF self, DOFConnection connection );

/**
 * Forces a server to stop. This method can be used to forcibly stop a server, even if multiple references to
 * the server exist. In contrast, @ref DOFServer_Stop stops the server only when the number of references
 * to stop calls is equal to the number of references to @ref DOFServer_Start and @ref DOFServer_BeginStart calls.
 *
 * @param self The DOF associated with the server. This must not be @c NULL.
 * @param server The server to stop. If @c NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOF_Stop( DOF self, DOFServer server );

/**
 * Creates a @ref DOFAuthenticator that uses the specified configuration. The DOFAuthenticator class must be implemented on
 * nodes that need to perform as Authentication Servers. All secure domains must have at least one Authentication Server.
 *
 * @mustfree{DOFAuthenticator_Destroy}
 *
 * @param self The DOF instance on which to create an authenticator. This must not be @c NULL.
 * @param config An instance of @ref DOFAuthenticatorConfig that represents a source of configuration parameters for the
 *               DOFAuthenticator. Use @ref DOFAuthenticatorConfigBuilder to create an instance of DOFAuthenticatorConfig.
 *               This should not be @c NULL.
 * @return A new instance of DOFAuthenticator.
 * @exception Returns @c NULL if the configuration is invalid, the DOF cannot accept the new DOFAuthenticator,
 *           or the DOFAuthenticator could not be created.
 */
extern DOF_OAL_DECLSPEC DOFAuthenticator DOF_CreateAuthenticator( DOF self, DOFAuthenticatorConfig config );

/**
 * Gets a DOFAuthenticatorList of all DOFAuthenticators known to the given DOF.
 *
 * @mustfree{DOFAuthenticatorList_Destroy}
 *
 * @param self The DOF to get a list of DOFAuthenticators from. This must not be @c NULL.
 * @return A count and list of all DOFAuthenticators.
 * @exception NULL is returned if the list cannot be allocated.
 * @see DOFAuthenticatorList
 */
extern DOF_OAL_DECLSPEC DOFAuthenticatorList DOF_GetAuthenticators( const DOF self );

/**
 * Creates a @ref DOFDomain that uses the specified configuration. The DOFDomain can be used to monitor connectivity to a
 * secure domain's Authentication Server.
 *
 * The internal library tracks references to duplicate domains, so this function can safely be called multiple times with the
 * same configuration without consuming extra resources. Multiple references to a DOFDomain can easily occur, because the library
 * automatically creates DOFDomain instances when you pass a DOFDomainConfig in the following functions:
 *
 * - @ref DOFConnectionConfigBuilder_AddDomain
 * - @ref DOFConnection_AddDomain
 * - @ref DOFServerConfigBuilder_AddDomain
 *
 * @mustfree{DOFDomain_Destroy}
 * @param self The DOF to create the domain on. This must not be @c NULL.
 * @param config An instance of @ref DOFDomainConfig that represents a source of configuration parameters for the DOFDomain.
 *               Use @ref DOFDomainConfigBuilder to create an instance of DOFDomainConfig. This should not be @c NULL.
 *
 * @return A new instance of DOFDomain.
 * @exception NULL is returned if the configuration is invalid or the DOFDomain could not be created.
 */
extern DOF_OAL_DECLSPEC DOFDomain DOF_CreateDomain( DOF self, DOFDomainConfig config );

/**
 * Gets a DOFDomainList of all the domains associated with the DOF.
 * @mustfree{DOFDomainList_Destroy}
 *
 * @param self The DOF to get the domains from. This must not be @c NULL.
 *
 * @return A count and a list of all the DOFDomains associated with the given DOF.
 * @exception NULL is returned if the list could not be allocated.
 * @see DOFDomainList
 */
extern DOF_OAL_DECLSPEC DOFDomainList DOF_GetDomains( DOF self );

/**
 * Adds a @ref DOFDomainListener to the given DOF, which is used to monitor the DOF's known security domains.
 *
 * @note Listeners will only be added once.
 *
 * @param self The DOF to add the listener to. This must not be @c NULL.
 * @param listener The listener to add. This should not be @c NULL.
 *
 * @retval TRUE if the listener was added or was found in the list of listeners.
 * @exception FALSE is returned if the listener could not be added, or an invalid parameter is detected.
 * @see DOFDomain
 * @see DOF_RemoveDomainListener
 */
extern DOF_OAL_DECLSPEC boolean DOF_AddDomainListener( DOF self, DOFDomainListener listener );

/**
 * Removes a @ref DOFDomainListener from the given DOF. When the DOFDomainListener is removed, this function causes
 * the DOFDomainListener's removed callback to be called, if it defines one.
 *
 * @param self The dof to remove the listener from. This must not be @c NULL.
 * @param listener The listener to remove. If @c NULL, no action is taken.
 * @see DOF_AddDomainListener
 */
extern DOF_OAL_DECLSPEC void DOF_RemoveDomainListener( DOF self, DOFDomainListener listener );

/**
 * Adds a @ref DOFInterestListener to the DOF. Interest listeners are used by nodes (usually providers or proxies) that want to
 * monitor all interest requests.
 *
 * @note Listeners will only be added once.
 *
 * @param self The DOF to add the listener to. This must not be @c NULL.
 * @param listener The listener to add. This should not be @c NULL.
 *
 * @retval TRUE if the listener was added.
 * @exception FALSE is returned if the listener could not be added.
 */
extern DOF_OAL_DECLSPEC boolean DOF_AddInterestListener( DOF self, DOFInterestListener listener );

/**
 * Removes a @ref DOFInterestListener from the DOF. When the DOFInterestListener is removed, this function causes
 * the DOFInterestListener's removed callback to be called, if it defines one.
 *
 * @param self The dof. This must not be @c NULL.
 * @param listener The listener to remove. If @c NULL, no action is taken.
 * @see DOF_AddInterestListener
 */
extern DOF_OAL_DECLSPEC void DOF_RemoveInterestListener( DOF self, DOFInterestListener listener );

/**
 * The standard Default Protocol Negotiator. For DOFConnections, this default negotiator
 * attempts to establish a connection using any standard protocol. For DOFServers, this default negotiator
 * allows connections using any standard protocol. This is the default for any DOFServer or DOFConnection.
 * @note If security is desired, see @ref DOFConnectionConfigBuilder_SetSecurityDesire and
 * @ref DOFServerConfigBuilder_SetSecurityDesire
 */
extern DOF_OAL_DECLSPEC DOFProtocolNegotiator DOF_DEFAULT_PROTOCOL_NEGOTIATOR;

/**
 * The standard Authentication Server (AS) Protocol Negotiator, for AS-only
 * connections. For DOFConnections, this negotiator attempts to establish a
 * non-secure connection to an AS and allowing only traffic using the Ticket
 * Request Protocol. For DOFServers, this negotiator only allows the connecting
 * node to negotiate a non-secure connection allowing only traffic using the
 * Ticket Request Protocol.
 */
extern DOF_OAL_DECLSPEC DOFProtocolNegotiator DOF_DEFAULT_AS_PROTOCOL_NEGOTIATOR;

/** @} **/

#ifdef  __cplusplus
}
#endif

/**
@}
**/

#endif /* DOF_DOF_H */
