/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_DOMAIN_H
#define DOF_DOMAIN_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif


/*
**  "Public" Type and Method Declarations
*/

/** @defgroup DOFDomain DOFDomain
 *  @ingroup DOF_Framework
 *
 *  This class represents security support in the DOF.  Security in the DOF is
 *  tied to a domain, which consists of at least a domain identifier and some
 *  credentials.
 *
 *   Without at least a single domain or authenticator, a DOF cannot accept secure connections.
 *
 *   Each domain that is created must be destroyed using @ref DOFDomain_Destroy.
 *
 * @{
 */

/**
    The DOFDomain state Listener interface.  Defined as a pointer to the @ref DOFDomainStateListener_t structure.  Refer to the structure for more information.
**/
typedef struct DOFDomainStateListener_t* DOFDomainStateListener;

/**
 * The listener interface structure used to obtain information on domain state.
 *
 * This structure should be placed as the first element in a larger structure
 * representing any necessary state of the listener.  All of the functions
 * take as their first argument a pointer to this structure (a @ref
 * DOFDomainListener).  Inside each function, it can thus be type-cast as a pointer
 * to the larger structure to access the listener's state.  If no additional
 * state is required, this structure may be used directly.
 *
 * @see DOFDomain_AddStateListener
 */
typedef struct DOFDomainStateListener_t
{

    /**
     * The listener interface function structure, containing a list of
     * function pointers that implement the interface.  It has the same role as
     * a vtable in C++.  This must not be NULL.
     */

    const struct DOFDomainStateListenerFns_t
    {

        /**
         * This callback is called to notify that the listener was removed.  This call is to notify the
         * program that this listener will not be used anymore and all state associated with it can be
         * freed.
         *
         * @note This callback may be NULL to ignore notification of removal.
         *
         * @param self This listener. This will not be NULL.
         * @param domain The domain that this listener was added to. This will not be NULL.
         * @param exception This will hold the reason why this listener was removed if the removal wasn't application triggered.
         */
        void (*Removed)( DOFDomainStateListener self, DOFDomain domain, DOFException exception );

        /**
         * This callback is called to notify the listener that the state of a DOFDomain has changed.
         *
         * This callback will be called immediately when the listener is added with the current state
         * and then will be called whenever the state changes.
         *
         * @param self This listener. This will not be NULL.
         * @param domain The domain that state has changed for. This will not be NULL.
         * @param state The state of the DOFDomain at the point this call was triggered. This will not be NULL.
         */
        void (*StateChanged)( DOFDomainStateListener self, DOFDomain domain, DOFDomainState state );

    } * fns;

    /* Implementations may extend this structure. */

} DOFDomainStateListener_t;

/**
 * Destroys a DOFDomain.
 *
 * @param self The domain to destroy. This may be NULL.
 */
extern DOF_OAL_DECLSPEC void DOFDomain_Destroy( DOFDomain self );

/**
 * Creates a DOFDomainState for the given DOFDomain.
 * @mustfree{DOFDomainState}
 *
 * @param self The domain to create the state for. This must not be NULL.
 *
 * @return DOFDomainState state that represents that DOFDomain at the point in time where it was created.
 * @exception NULL is returned for a failure.
 */
extern DOF_OAL_DECLSPEC DOFDomainState DOFDomain_CreateState( const DOFDomain self );

/**
 * Adds a state listener to the given domain.
 *
 * @param self The domain to add the listener too. This must not be NULL.
 * @param listener The listener to add to the domain. If this has already been added to the domain
 *                 it will not be added a second time. This must not be NULL.
 *
 * @return boolean TRUE if the listener was added.
 * @exception FALSE is returned if the listener was unable to be added.
 */
extern DOF_OAL_DECLSPEC boolean DOFDomain_AddStateListener( DOFDomain self, DOFDomainStateListener listener );

/**
 * Removes a listener previously added to a domain.
 *
 * @param self The domain to remove the listener from. This must not be NULL.
 * @param listener The listener to be removed. If the listener was found the Removed callback will be
 *                 called if available. This must not be NULL.
 */
extern DOF_OAL_DECLSPEC void DOFDomain_RemoveStateListener( DOFDomain self, DOFDomainStateListener listener );

/**
@}
**/

/** @defgroup DOFDomainConfig DOFDomainConfig
    @ingroup DOFDomain

    Represents the configuration of a DOFDomain.
@{
**/

/** @name Builder
@{
**/

/**
 * Creates a DOFDomainConfigBuilder with the given credentials and the default silence timeout.
 * @mustfree{DOFDomainConfigBuilder_Destroy}
 *
 * @param creds The credentials to create the builder with. This may be NULL. If NULL the builder will fail to be built
 *              if credentials are not set before the call to build.
 *
 * @return DOFDomainConfigBuilder The builder with the given credentials.
 * @exception NULL If the builder could not be created.
 */
extern DOF_OAL_DECLSPEC DOFDomainConfigBuilder DOFDomainConfigBuilder_Create( DOFCredentials creds );

/**
 * Creates a DOFDomainConfigBuilder with the given state.
 * @mustfree{DOFDomainConfigBuilder_Destroy}
 *
 * @param state The state. Must not be NULL.
 *
 * @return DOFDomainConfigBuilder The builder.  This will need credentials to be set before this can be built.
 * @exception NULL If the builder failed to be created.
 */
extern DOF_OAL_DECLSPEC DOFDomainConfigBuilder DOFDomainConfigBuilder_Create_State( const DOFDomainState state );

/**
 * Creates a DOFDomainConfigBuilder with the given config.
 * @mustfree{DOFDomainConfigBuilder_Destroy}
 *
 * @param config The config. Must not be NULL.
 *
 * @return DOFDomainConfigBuilder The builder.
 * @exception NULL If the builder failed to be created.
 */
extern DOF_OAL_DECLSPEC DOFDomainConfigBuilder DOFDomainConfigBuilder_Create_Config( const DOFDomainConfig config );

/**
 * Destroys the given builder.
 *
 * @param self The builder to destroy. This may be NULL. If NULL this will just return.
 */
extern DOF_OAL_DECLSPEC void DOFDomainConfigBuilder_Destroy( DOFDomainConfigBuilder self );

/**
 * Set new credentials in the builder.
 *
 * @param self The builder to set the credentials in. This must not be NULL.
 * @param creds The credentials to set in the builder. This must not be NULL.
 *
 * @return boolean TRUE If the credentials were able to be set.
 * @exception FALSE If the credentials were not able to be set.
 */
extern DOF_OAL_DECLSPEC boolean DOFDomainConfigBuilder_SetCredentials( DOFDomainConfigBuilder self, DOFCredentials creds );

/**
 * Sets the max silence time of the builder.
 *
 * @note Default silence time is an hour.
 *
 * @param self The builder to set the silence time in. This must not be NULL.
 * @param maxSilence The maximum length of time to wait between receiving a response from this domain in milliseconds. This must be greater than 0.
 *
 * @return boolean TRUE If the time was set in the builder.
 * @exception FALSE if there was a failure.
 */
extern DOF_OAL_DECLSPEC boolean DOFDomainConfigBuilder_SetMaxSilence( DOFDomainConfigBuilder self, uint32 maxSilence );

/**
 * Builds a config from the given builder.
 * @mustfree{DOFDomainConfig}
 *
 * @param self The builder to build a config from. This must not be NULL.
 *
 * @return DOFDomainConfig A config that was created from the builder.
 * @exception NULL If the config failed to be built.
 */
extern DOF_OAL_DECLSPEC DOFDomainConfig DOFDomainConfigBuilder_Build( const DOFDomainConfigBuilder self );

/**
 * Builds a config from the given builder and destroys the builder.
 * @mustfree{DOFDomainConfig}
 *
 * @note The builder given will be destroyed whether the function is successful or not.
 *
 * @param self The builder to build a config from. This may be NULL.
 *
 * @return DOFDomainConfig A config that was created from the builder.
 * @exception NULL if the config failed to be built.
 */
extern DOF_OAL_DECLSPEC DOFDomainConfig DOFDomainConfigBuilder_BuildAndDestroy( DOFDomainConfigBuilder self );

/**
@}
**/

/**
 * Destroys a DOFDomainConfig.
 *
 * @param self The DOFDomainConfig to destroy. This may be NULL. If NULL no action will be taken.
 */
extern DOF_OAL_DECLSPEC void DOFDomainConfig_Destroy( DOFDomainConfig self );

/**
 * Gets the config from the give config.
 *
 * @param self The config to return the credentials from. This may not be NULL.
 *
 * @return DOFCredentials The credentials from the config.
 */
extern DOF_OAL_DECLSPEC DOFCredentials DOFDomainConfig_GetCredentials( const DOFDomainConfig self );

/**
 * Gets the silence time from the config.
 *
 * @param self The config to return the maxSilence from. This may not be NULL.
 *
 * @return uint32 The max silence time from the config.
 */
extern DOF_OAL_DECLSPEC uint32 DOFDomainConfig_GetMaxSilence( const DOFDomainConfig self );

/**
@}
**/

/** @defgroup DOFDomainState DOFDomainState
    @ingroup DOFDomain

    Represents the instantaneous state of a DOFDomain.
@{
**/

/**
 * Destroys a DOFDomainState.
 *
 * @param self The state to destroy. This may be NULL. If NULL no action will be taken.
 */
extern DOF_OAL_DECLSPEC void DOFDomainState_Destroy( DOFDomainState self );

/**
 * Returns if the state says the DOFDomainState is connected or not.
 *
 * @param self The state to get whether the domain is connected or not. This must not be NULL.
 *
 * @return boolean TRUE If the domain is connected,
 *                 FALSE If the domain isn't connected.
 */
extern DOF_OAL_DECLSPEC boolean DOFDomainState_IsConnected( const DOFDomainState self );

/**
 * Get credentials from the state. The credentials returned will not contain private data.
 *
 * @param self The state to get the credentials from. This must not be NULL.
 *
 * @return DOFCredentials The credentials from the state.
 */
extern DOF_OAL_DECLSPEC DOFCredentials DOFDomainState_GetCredentials( const DOFDomainState self );

/**
 * Gets the max silence time from the state.
 *
 * @param self The state. This must not be NULL.
 *
 * @return uint32 The max silence time from the state.
 */
extern DOF_OAL_DECLSPEC uint32 DOFDomainState_GetMaxSilence( const DOFDomainState self );

/**
 * Frees the given DOFDomainList.  The list will not be usable after this call.
 *
 * @param self The list to Destroy. This may be NULL. If NULL no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFDomainList_Destroy( DOFDomainList self );

/**
@}
*/

#ifdef  __cplusplus
}
#endif

#endif /* DOF_DOMAIN_H */
