/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_ERROR_H
#define DOF_ERROR_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DOFError DOFError
    @ingroup DOF_Framework

@{
**/

/*
**  "Public" Type and Method Declarations
*/

/* Standard OAL status/error codes.  These codes are used with the type DOFError, typically in conjunction with DOFException. */

/* NOTE: The following error codes map to the Object Access Protocol Specification Exception PDU. Numeric values match those of the specification. */
#define DOFERROR_OK                             ((DOFError)0x0000)  /**< Success or OK (no error). **/
#define DOFERROR_NOT_SUPPORTED                  ((DOFError)0x0002)  /**< The request is not supported by the library or the OAP Interface (for example, an Invoke on an Event). @see DOFObject_Get @see DOFObject_Set @see DOFObject_Invoke @see DOFObject_BeginRegister @see DOFObject_BeginSubscribe @see DOFObject_BeginSession **/
#define DOFERROR_TYPE_MISMATCH                  ((DOFError)0x0003)  /**< The data values passed do not match the data signature of the defined Interface, therefore a value or value list could not be written to or read from the wire.  The values may mismatch by type and/or count.  @see DOFObject_Get @see DOFObject_Set @see DOFObject_Invoke @see DOFObject_BeginSubscribe @see DOFProviderException_Create **/
#define DOFERROR_INSUFFICIENT_RESOURCES         ((DOFError)0x0004)  /**< There are not enough resources available to accept or process the command. **/
#define DOFERROR_TERMINATED                     ((DOFError)0x0005)  /**< The operation, connection, or request has been terminated.  @see DOFObject_BeginRegister @see DOFObject_BeginSubscribe **/
#define DOFERROR_REJECTED                       ((DOFError)0x0006)  /**< An operation has not been accepted (for example, the provider does not allow subscriptions on the requested interface and item).  @see DOFObject_BeginSubscribe **/
#define DOFERROR_NOT_FOUND                      ((DOFError)0x0007)  /**< The requested information cannot be found (for example, a request was received that should match an existing binding, but no matching binding can be found).  @see DOFObject_Get @see DOFObject_Set @see DOFObject_Invoke @see DOFObject_BeginRegister @see DOFObject_BeginSubscribe @see DOFObject_BeginSession **/
#define DOFERROR_DUPLICATE                      ((DOFError)0x0008)  /**< An operation or entry already exists and cannot be entered again (for example, the requested Session ID is already in use).  Retry with unique information.  @see DOFObject_BeginSession **/
#define DOFERROR_TOO_MANY                       ((DOFError)0x0009)  /**< The maximum number of entries has been reached (for example, the provider has reached a maximum supported session count on the requested binding).  The request may succeed later.  @see DOFObject_BeginSession **/
#define DOFERROR_APPLICATION_ERROR              ((DOFError)0x000a)  /**< An application callback has failed (for example, the provider implementation). **/

/* NOTE: The following error codes are for use by the API.  They are not part of the protocol specification, but are part of the OAL specification.  Numeric values are library-specific. */
#define DOFERROR_FAILED                         ((DOFError)0x8000)  /**< Failed without any other, specific error. **/
#define DOFERROR_TIMEOUT                        ((DOFError)0x8001)  /**< An operation timed out. @see DOFOperation_WaitComplete @see DOFOperation_WaitAcknowledge **/
#define DOFERROR_NOT_READABLE                   ((DOFError)0x8002)  /**< The property is not readable. **/
#define DOFERROR_NOT_WRITABLE                   ((DOFError)0x8003)  /**< The property is not writable. **/
#define DOFERROR_RESULT_LOST                    ((DOFError)0x8004)  /**< One or more results for the operation have been lost (not received).  For exmaple, an Event was lost for a Register operation. **/
#define DOFERROR_PROTOCOL_NEGOTIATION_FAILED    ((DOFError)0x8005)  /**< Protocol negotiation failed when attempting to establish a connection. **/
#define DOFERROR_TRANSPORT_FAILED               ((DOFError)0x8006)  /**< Transport-level error or failure. **/
#define DOFERROR_IO_FAILED                      ((DOFError)0x8007)  /**< Transport-level Input/Output (Receive/Send) error or failure. **/
#define DOFERROR_AUTHENTICATION_FAILED          ((DOFError)0x8008)  /**< Authentication error or failure. **/
#define DOFERROR_ACK_TIMEOUT                    ((DOFError)0x8009)  /**< Ack timeout error. **/
#define DOFERROR_STALE_CREDENTIAL               ((DOFError)0x800a)  /**< The credential used was previously resolved, but failed to authenticate due to a stale cache of the resolution data. The cache has been invalidated and will resolve again on next use. **/

/* NOTE: The following error codes map directly to the Ticket Request Protocol Specification Reject PDU.  Numeric values of the lowest byte match those of the specification. */
#define DOFERROR_ACCESS_DENIED                  ((DOFError)0x9002)  /**< Access is denied.  The request could not be authorized due to a lack of one or more required permissions for the operation. **/
#define DOFERROR_UNKNOWN_INITIATOR              ((DOFError)0x9003)  /**< The initiator is not known. **/
#define DOFERROR_UNKNOWN_RESPONDER              ((DOFError)0x9004)  /**< The responder is not known. **/
#define DOFERROR_UNKNOWN_DOMAIN                 ((DOFError)0x9005)  /**< The domain in which authentication was requested is not known, cannot be found (e.g., no route), or cannot be validated. **/
#define DOFERROR_SERVICE_UNAVAILABLE            ((DOFError)0x9006)  /**< The service is not currently available.  The same request may succeed later. **/
#define DOFERROR_BAD_MODE                       ((DOFError)0x9007)  /**< The requested security mode is not acceptable. **/
#define DOFERROR_NOT_IN_GROUP                   ((DOFError)0x9009)  /**< The initiator is not a member of the requested group. **/
#define DOFERROR_UNSUPPORTED_INITIATOR          ((DOFError)0x900a)  /**< The identification mode requested by the initiator is not known.  Identification is not possible. **/
#define DOFERROR_UNSUPPORTED_RESPONDER          ((DOFError)0x900b)  /**< The identification mode requested by the responder is not known.  Identification is not possible. **/
#define DOFERROR_INVALID_INITIATOR              ((DOFError)0x900c)  /**< The identification mode requested by the initiator is known, but there is no associated private data. **/
#define DOFERROR_INVALID_RESPONDER              ((DOFError)0x900d)  /**< The identification mode requested by the responder is known, but there is no associated private data. **/
#define DOFERROR_RESOLUTION_FAILED              ((DOFError)0x900e)  /**< An error occurred in resolution.  This may indicate that the authentication information was invalid; for example, a bad password was used. **/
#define DOFERROR_VALIDATION_FAILED              ((DOFError)0x900f)  /**< An error occurred in validation.  This may indicate that the authentication information was invalid; for example, an incorrect shared key. **/

/* NOTE: The following error codes are for use by the API.  They are library-specific, and not part of the protocol or OAL specifications. */
#define DOFERROR_ILLEGAL_ARGUMENT               ((DOFError)0x10000) /**< An invalid or illegal argument was detected (for example, a NULL was passed where it should not have been). **/

/* NOTE: RESERVED are all error codes having bit 31 set, i.e. from 0x80000000 through 0xFFFFFFFF. */

/**
 * Convert the given status code to an ASCII string.
 * @param status The status code.
 * @return An ASCII string version of the status code.
 */
extern DOF_OAL_DECLSPEC const char * DOFError_GetString ( const DOFError status );


/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* DOF_ERROR_H */
