/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_EXCEPTION_H
#define DOF_EXCEPTION_H


#include <dof/core.h>


#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DOFException DOFException
    @ingroup DOF_Framework

    DOFException encapsulates DOFError codes and provider item exceptions.

@{
**/

/*
**  "Public" Type and Method Declarations
*/

/**
 * Exception type.
 */
typedef enum DOFExceptionType_t
{
    DOFEXCEPTIONTYPE_ERROR  = 1,    /**< Error code exception, with optional string. **/
    DOFEXCEPTIONTYPE_PROVIDER = 2   /**< Provider exception, based on an Interface definition and a parameter list. **/
} DOFExceptionType;

/**
 *  Create a new copy of an existing @ref DOFException.
 *  @mustfree{DOFException_Destroy}
 *  @param  self The @ref DOFException to be copied.  This should not be NULL.
 *  @return A new copy of the @ref DOFException.
 *  @exception  NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DOFException DOFException_Copy( const DOFException self );

/**
 * Construct a DOFErrorException from a DOFError code and a string.
 * @mustfree{DOFException_Destroy}
 * @param error The error code for the exception.  This should be one of the DOFError codes.
 * @param message An associated string for the exception.  If NULL, the exception is created with the default string for the error code (see @ref DOFError_GetString).
 * @return A new DOFErrorException.  If the desired exception cannot be created, a DOFErrorException representing a @ref DOFERROR_INSUFFICIENT_RESOURCES exception is returned in its place.
 * @exception NULL is returned if @a error does not represent a valid DOFError code.
 */
extern DOF_OAL_DECLSPEC DOFException DOFErrorException_Create( DOFError error, const char * message );

/**
 * Construct a DOFException using an Interface Exception (an Item within an Interface).
 * @mustfree{DOFException_Destroy}
 * @param iException The DOFInterfaceException, indicating the item identifier of the exception as defined by its related interface definition.  This should not be NULL.
 * @param parameterCount The number of parameters for the exception. This must be less than 256.
 * @param parameters The parameters for the exception.  This must not be NULL if @a parameterCount > 0.
 * @return A new DOFException.  If the desired exception cannot be allocated, a DOFException representing a @ref DOFERROR_INSUFFICIENT_RESOURCES exception is returned in its place.
 * @exception NULL is returned if the parameters do not match the Interface's defined parameter type(s).
 */
extern DOF_OAL_DECLSPEC DOFException DOFProviderException_Create( DOFInterfaceException iException, uint16 parameterCount, const DOFValue * parameters );

/**
 * Destroy a DOFException.
 * @param self The DOFException.  If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFException_Destroy( DOFException self );

/**
 * Get the type of the DOFException.
 * @param self The DOFException.  This must not be NULL.
 * @return The type of the Exception.
 */
extern DOF_OAL_DECLSPEC DOFExceptionType DOFException_GetType( const DOFException self );

/**
 * Get the error code for the DOFException.
 * @param self The DOFException.  This must not be NULL, and must be a valid DOFErrorException.
 * @return The error code for the Exception.
 * @exception @ref DOFERROR_FAILED is returned if an invalid parameter is detected.
 */
extern DOF_OAL_DECLSPEC DOFError DOFErrorException_GetError( const DOFException self );

/**
 * Get the error message string associated with the DOFException.
 * @param self The DOFException.  This must not be NULL, and should be a valid DOFErrorException.
 * @return The error message string associated with the Exception.
 * @exception NULL is returned if an invalid parameter is detected.
 */
extern DOF_OAL_DECLSPEC const char * DOFErrorException_GetMessage( const DOFException self );

/**
 * Get the extra data associated with the DOFException, if any.
 * The type of the data available depends on the specific exception.
 * @param self The DOFException.  This must not be NULL, and should be a valid DOFErrorException.
 * @return The error string associated with the Exception.
 * @exception NULL is returned if an invalid parameter is detected.
 */
extern DOF_OAL_DECLSPEC const void * DOFErrorException_GetData( const DOFException self );

/**
 * Get the Interface Exception associated with the DOFException.
 * @param self The DOFException.  This must not be NULL, and must be a valid DOFProviderException.
 * @return The Interface Exception associated with the DOFException.
 * @exception NULL is returned if an invalid parameter is detected.
 */
extern DOF_OAL_DECLSPEC DOFInterfaceException DOFProviderException_GetInterfaceException( const DOFException self );

/**
 * Get the parameters of the DOFException.
 * @param self The DOFException.  This must not be NULL, and must be a valid DOFProviderException.
 * @return The parameters of the DOFException, as an array of @ref DOFValue.
 * @exception NULL is returned if an invalid parameter is detected.
 */
extern DOF_OAL_DECLSPEC DOFValue * DOFProviderException_GetParameters( const DOFException self );

/**
 * Get the count of parameters of the DOFException.
 * @param self The DOFException.  This must not be NULL, and must be a valid DOFProviderException.
 * @return The count of parameters of the DOFException.
 * @exception 0 is returned if an invalid parameter is detected.
 */
extern DOF_OAL_DECLSPEC uint8 DOFProviderException_GetParametersCount( const DOFException self );

/**
 * Get the @ref DOFProviderInfo of the DOFException, if any.
 * @param self The DOFException.  This must not be NULL, and should be a valid DOFErrorException or DOFProviderException.
 * @return The @ref DOFProviderInfo of the DOFException, or NULL if the exception did not originate from a Provider (it originated from the library itself).
 * @exception NULL is returned if an invalid parameter is detected.
 */
extern DOF_OAL_DECLSPEC DOFProviderInfo DOFException_GetProviderInfo( const DOFException self );

/**
@}
**/


#ifdef  __cplusplus
}
#endif


#endif /* DOF_TIME_H */
