/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_FILTER_H
#define DOF_FILTER_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DOFOperationFilter DOFOperationFilter
    @ingroup DOFOperation

    The DOFOperationFilter class represents an operation filter.

@{
**/

/**
 * @name Builder
 * @{
 */


/**
 *  Construct a DOFOperationFilterBuilder.
 *
 *  @mustfree{DOFOperationFilterBuilder_Destroy or @ref DOFOperationFilterBuilder_BuildAndDestroy}
 *
 *  @return a DOFOperationFilterBuilder that can be later configured
 */
extern DOF_OAL_DECLSPEC DOFOperationFilterBuilder DOFOperationFilterBuilder_Create( void );

/**
 *  Construct a DOFOperationFilterBuilder using an existing DOFOperationFilter as a template.
 *
 *  @mustfree{DOFOperationFilterBuilder_Destroy or @ref DOFOperationFilterBuilder_BuildAndDestroy}
 *
 *  @param self A DOFOperationFilter to copy.  Should not be @c NULL.
 *
 *  @return a DOFOperationFilterBuilder based on @a self
 */
extern DOF_OAL_DECLSPEC DOFOperationFilterBuilder DOFOperationFilterBuilder_Create_Filter( DOFOperationFilter self );

/**
 *  Augment the DOFPermissionSet in a DOFOperationFilterBuilder.  If a DOFOperation requires a permission that
 *  is not present in the fullly augmented DOFPermissionSet, it will be blocked.
 *
 *  @param self The DOFOperationFilterBuilder to modify.  Should not be @c NULL.
 *  @param filter The DOFPermissionSet to filter on.  Should not be @c NULL.
 *
 *  @retval TRUE @a filter was applied
 *  @retval FALSE @a filter was not successfully applied
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationFilterBuilder_AllowByPermissions( DOFOperationFilterBuilder self, DOFPermissionSet filter );

/**
 *  Construct a DOFOperationFilter from a DOFOperationFilterBuilder.  Whether succesfuly or not, @a self will be destroyed.
 *
 *  @mustfree{DOFOperationFilter_Destroy}
 *
 *  @param self The DOFOperationFilterBuilder.  Should not be @c NULL.
 *
 *  @return a DOFOperationFilter constructed from @a self
 *
 *  @retval NULL    there was an issue building a DOFOperationFilter from @a self
 */
extern DOF_OAL_DECLSPEC DOFOperationFilter DOFOperationFilterBuilder_BuildAndDestroy( DOFOperationFilterBuilder self );

/**
 *  Construct a DOFOperationFilter from a DOFOperationFilterBuilder.
 *
 *  @mustfree{DOFOperationFilter_Destroy}
 *
 *  @param self The DOFOperationFilterBuilder to use.  Should not be @c NULL.
 *
 *  @return a DOFOperationFilter constructed from @a self
 */
extern DOF_OAL_DECLSPEC DOFOperationFilter DOFOperationFilterBuilder_Build( DOFOperationFilterBuilder self );

/**
 *  Copy a DOFOperationFilterBuilder.
 *
 *  @param self The DOFOperationFilterBuilder to copy.  Should not be @c NULL.
 *
 *  @return a copy of @a self
 */
extern DOF_OAL_DECLSPEC DOFOperationFilterBuilder DOFOperationFilterBuilder_Copy( DOFOperationFilterBuilder self );

/**
 *  Destroy a DOFOperationFilterBuilder.
 *
 *  @param self The DOFOperationFilterBuilder to destroy.  If @c NULL, this function will do nothing.
 */
extern DOF_OAL_DECLSPEC void DOFOperationFilterBuilder_Destroy( DOFOperationFilterBuilder self );

/**
@}
**/

/**
 *  Perform a copy of a DOFOperationFilter.
 *
 *  @param self The DOFOperationFilter to copy.  Should not be @c NULL.
 *
 *  @return a copy of @a self
 */
extern DOF_OAL_DECLSPEC DOFOperationFilter DOFOperationFilter_Copy( DOFOperationFilter self );

/**
 *  Destroy a DOFOperationFilter.
 *
 *  @param self The DOFOperationFilter to destory.  If @c NULL, this function will do nothing.
 */
extern DOF_OAL_DECLSPEC void DOFOperationFilter_Destroy( DOFOperationFilter self );

/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif
