/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_INTERFACE_H
#define DOF_INTERFACE_H


#include <dof/core.h>


#ifdef  __cplusplus
extern "C" {
#endif


/** @defgroup DOFInterface DOFInterface
    @ingroup DOF_Framework

    DOF Interface Definition

    A DOFInterface defines a set of possible interactions with a DOFObject, which
    is typically remote (across the network). Conceptually it is similar to
    interfaces in object-oriented programming languages such as Java or C#. It is
    composed of a set of Properties, Methods, Events and Exceptions with their
    parameter and type information.

    A DOFObject that provides (or implements) a DOFInterface is called a
    Provider, while a DOFObject that makes requests (such as invoking a method or
    setting a property value) is called a Requestor.

    @see DOFObject
@{
**/

/**
 * Construct a DOFInterface from its serialized byte array form.
 * @mustfree{DOFInterface_Destroy}
 * @param iid The Interface ID.  This should not be NULL.
 * @param dataSize The size of the serialized data.
 * @param data The serialized data.  This should not be NULL.
 * @return A new DOFInterface.
 * @exception NULL is returned if the data does not correspond to a valid Interface definition, or the DOFInterface cannot be created.
 */
extern DOF_OAL_DECLSPEC DOFInterface DOFInterface_Create( const DOFInterfaceID iid, uint32 dataSize, const uint8 * data );

/**
 * Destroy a DOFInterface.
 * @param self The Interface.  If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFInterface_Destroy( DOFInterface self );

/**
    Get an Event from the Interface, given its identifier (a @ref DOFItemID).
    @note The returned value is owned by the Interface and is only valid as long as the Interface remains valid.  It must not be freed individually.

    @param  self    The Interface.  This must not be NULL.
    @param  itemID  The identifier of the Event item within the Interface.

    @return The Event.

    @exception NULL is returned if the item could not be found in the Interface.
**/
extern DOF_OAL_DECLSPEC DOFInterfaceEvent DOFInterface_GetEvent( const DOFInterface self, DOFItemID itemID );

/**
    Get an Exception from the Interface, given its identifier (a @ref DOFItemID).
    @note The returned value is owned by the Interface and is only valid as long as the Interface remains valid.  It must not be freed individually.

    @param  self    The Interface.  This must not be NULL.
    @param  itemID  The identifier of the Exception item within the Interface.

    @return The Exception.

    @exception NULL is returned if the item could not be found in the Interface.
**/
extern DOF_OAL_DECLSPEC DOFInterfaceException DOFInterface_GetException( const DOFInterface self, DOFItemID itemID );

/**
    Get a Method from the Interface, given its identifier (a @ref DOFItemID).
    @note The returned value is owned by the Interface and is only valid as long as the Interface remains valid.  It must not be freed individually.

    @param  self    The Interface.  This must not be NULL.
    @param  itemID  The identifier of the Method item within the Interface.

    @return The Method.

    @exception NULL is returned if the item could not be found in the Interface.
**/
extern DOF_OAL_DECLSPEC DOFInterfaceMethod DOFInterface_GetMethod( const DOFInterface self, DOFItemID itemID );

/**
    Get a Property from the Interface, given its identifier (a @ref DOFItemID).
    @note The returned value is owned by the Interface and is only valid as long as the Interface remains valid.  It must not be freed individually.

    @param  self    The Interface.  This must not be NULL.
    @param  itemID  The identifier of the Property item within the Interface.

    @return The Property.

    @exception NULL is returned if the item could not be found in the Interface.
**/
extern DOF_OAL_DECLSPEC DOFInterfaceProperty DOFInterface_GetProperty( const DOFInterface self, DOFItemID itemID );

/**
    Get the serialized form of the Interface definition.
    @note The returned value is owned by the Interface and is only valid as long as the Interface remains valid.  It must not be freed individually.

    @param self  The Interface.  This must not be NULL.
    @return The serialized form of the Interface definition.  This will never be NULL.
**/
extern DOF_OAL_DECLSPEC const uint8 * DOFInterface_GetBytes( const DOFInterface self );

/**
    Get the size of the serialized form of the Interface definition.
    @param self  The Interface.  This must not be NULL.
    @return The size of the serialized form of the Interface definition.  This will always be > 0.
**/
extern DOF_OAL_DECLSPEC uint32 DOFInterface_GetBytesSize( const DOFInterface self );

/**
    Get the Interface ID associated with the Interface.
    @note The returned value is owned by the Interface and is only valid as long as the Interface remains valid.  It must not be freed individually.
    @param self  The Interface.  This must not be NULL.
    @return The Interface ID associated with the Interface.  This will never be NULL.
**/
extern DOF_OAL_DECLSPEC DOFInterfaceID DOFInterface_GetInterfaceID( const DOFInterface self );


/**
@}
**/


#ifdef  __cplusplus
}
#endif


#endif /* DOF_INTERFACE_H */
