/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_INTERFACEID_H
#define DOF_INTERFACEID_H


#include <dof/core.h>


#ifdef  __cplusplus
extern "C" {
#endif


/** @defgroup DOFInterfaceID DOFInterfaceID
    @ingroup DOF_Framework

    DOF Interface Identifier

    This class implements a DOF Interface Identifier. It is self
    descriptive, allowing its registry number, ID length, and ID data
    field to be identified and extracted.

    DOFInterfaceIDs are immutable, and should be treated as such by the application.
    With the calls of this API, DOFInterfaceIDs are created by the system. It is up
    to the application to later destroy any DOFInterfaceID that it created.

@{
**/

#define DOFINTERFACEID_MAX_SIZE            5    /**< Maximum size of the wire format representation of an Interface Identifier. **/

#define DOFINTERFACEID_TEST_REGISTRY      63    /**< Standard testing Interface Registry. **/



/** The C OAL provides and maintains this broadcast Interface ID, so applications need not create/destroy their own. **/
extern DOF_OAL_DECLSPEC DOFInterfaceID DOFINTERFACEID_WILDCARD;


/* DOFInterfaceID Methods */


/**
    Create an Interface ID from given registry and identifier.
    @mustfree{DOFInterfaceID_Destroy}
    @note Refer to the DOF Specification for more information on Interface ID Format.
    @param registry Registry in which this Interface ID is recognized.  Only registries recognized by the OpenDOF Project
           must be used.  Valid range is 0 <= @a registry <= 63.
    @param identifier The numeric identifier for the IID.  All values are valid; however, interfaces must be registered with the
                      OpenDOF Project prior to use in any product.
    @return    The newly constructed Interface ID.
    @exception NULL is returned if an invalid parameter is detected or insufficient resources exist to create the @ref DOFInterfaceID.
**/
extern DOF_OAL_DECLSPEC DOFInterfaceID DOFInterfaceID_Create( uint8 registry, uint32 identifier );

/**
 *  Create a new copy of an existing @ref DOFInterfaceID.
 *  @mustfree{DOFInterfaceID_Destroy}
 *  @param  self The @ref DOFInterfaceID to be copied.  This should not be NULL.
 *  @return A new copy of the @ref DOFInterfaceID.
 *  @exception  NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DOFInterfaceID DOFInterfaceID_Copy( const DOFInterfaceID self );

/**
    Create an Interface ID from a given Interface ID buffer.
    @mustfree{DOFInterfaceID_Destroy}
    @param dataSize   The size, in bytes, of the Interface ID buffer at @a data.
                      This should not exceed the actual size of the Interface
                      ID.
    @param data       Pointer to the given buffer. This should not be NULL, and
                      should contain the bytes representing a legal Interface
                      ID.
    @return    The newly created DOFInterfaceID.
    @exception NULL is returned if an invalid parameter is detected, or the
                    required resources cannot be allocated.
**/
extern DOF_OAL_DECLSPEC DOFInterfaceID DOFInterfaceID_Create_Bytes( uint32 dataSize, const uint8 * data );

/**
    Get pointer to the wire-representation bytes of an Interface ID.
    @param self The @ref DOFInterfaceID. This must not be NULL.
    @return Pointer to the wire-representation bytes of @a self.
**/
extern DOF_OAL_DECLSPEC const uint8 * DOFInterfaceID_GetBytes( const DOFInterfaceID self );

/**
    Get number of bytes of the wire-representation bytes of an Interface ID.
    @param self The @ref DOFInterfaceID. This must not be NULL.
    @return    Size, in byte count, of the wire-representation bytes of @a self.
**/
extern DOF_OAL_DECLSPEC uint8 DOFInterfaceID_GetBytesSize( const DOFInterfaceID self );

/**
    Get registry value of an Interface ID.
    @param self The @ref DOFInterfaceID. This must not be NULL.
    @return    Registry value of @a self.
**/
extern DOF_OAL_DECLSPEC uint8 DOFInterfaceID_GetRegistry( const DOFInterfaceID self );

/**
    Get the numeric identifier of an Interface ID.
    @param self The @ref DOFInterfaceID. This must not be NULL.
    @return    Pointer to the data portion of @a self.
**/
extern DOF_OAL_DECLSPEC uint32 DOFInterfaceID_GetIdentifier( const DOFInterfaceID self );

/**
    Destroy an Interface ID.
    @param  self The @ref DOFInterfaceID to be destroyed.  If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFInterfaceID_Destroy( DOFInterfaceID self );

/**
    Convert an Interface ID into its standard string representation.
    @mustfree{PCRMem_Free}
    The standardized Interface ID string format is described in the DOF Common Types document.
    @note The output is a null-terminated, readable string suitable for output in debug or log messages.
    @param iid The @ref DOFInterfaceID to convert. This should not be NULL.
    @return    Pointer to the newly generated string buffer.
    @exception NULL is returned if
        - the Interface ID is NULL, or
        - the required storage buffer cannot be allocated.
**/
extern DOF_OAL_DECLSPEC char * DOFInterfaceID_CreateStandardString( const DOFInterfaceID iid );

/**
    Compare two Interface IDs.
    @note This can also be used as a comparator function for sorting.
    @note If both are NULL, they are considered equal. If only one is NULL, it is considered less than the other.
    @param iid1 The first @ref DOFInterfaceID to compare.
    @param iid2 The second @ref DOFInterfaceID to compare.
    @return Zero           if exact match.
    @return Positive value if @a iid1 > @a iid2.
    @return Negative value if @a iid1 < @a iid2.
**/
extern DOF_OAL_DECLSPEC int32 DOFInterfaceID_Compare( const DOFInterfaceID iid1, const DOFInterfaceID iid2 );

/**
    Compare two Interface IDs for equality.
    @note If both parameters are NULL, they are considered equal. If only one is NULL, they are considered not equal.
    @param  iid1 The first @ref DOFInterfaceID to compare.
    @param  iid2 The second @ref DOFInterfaceID to compare.
    @return Non-zero (TRUE) if the two IDs match exactly.
    @return Zero (FALSE) if the two IDs do not match.
**/
#if DOXYGEN_MACRO_FUNCTIONS
extern DOF_OAL_DECLSPEC boolean DOFInterfaceID_IsEqual( const DOFInterfaceID iid1, const DOFInterfaceID iid2 ); /* Implemented by macro below.*/
#else
#define DOFInterfaceID_IsEqual( iid1, iid2 ) ((DOFInterfaceID_Compare( iid1, iid2 ) == 0))
#endif


/**
    Release resources previously allocated while forming an Interface ID list.
    @param  iidList The Interface ID list. If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFInterfaceIDList_Destroy( DOFInterfaceIDList iidList );


/**
@}
**/


#ifdef  __cplusplus
}
#endif


#endif /* DOF_INTERFACEID_H */
