/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_OBJECT_H
#define DOF_OBJECT_H


#include <dof/core.h>
#include <dof/control.h>


#ifdef  __cplusplus
extern "C" {
#endif


/*
**  "Public" Type and Method Declarations
*/


/** @defgroup DOFObject DOFObject
    @ingroup DOF_Framework

    The primary "class" for creating and interacting with local and remote DOF Objects.
    DOFObject implements two main interfaces:
        -# DOFProvider - used for providing information to other nodes in the network.
        -# DOFRequestor - used for requesting information from other nodes in the network.
    For each instance of DOFObject acting in its provider role, the C OAL maintains an
    internal list of active event registrations and property subscriptions.
@{
**/


/* DOFObject Methods */

/**
    Destroy the given DOFObject.

    This has no effect on any outstanding operations started from the
    DOFObject.  They will continue to be processed normally.  While the
    application cannot reference the DOFObject directly any longer, the
    DOFObject available to an operation's application callbacks will remain
    valid.

    @param  object  The @ref DOFObject.  If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFObject_Destroy( DOFObject object );


/**
    Return the Object ID associated with the given Object.
    @param  object  The @ref DOFObject. This must not be NULL.
    @return    The associated Object ID.
    @exception NULL is returned if @a obj is NULL.
**/
extern DOF_OAL_DECLSPEC DOFObjectID DOFObject_GetObjectID( const DOFObject object );


/**
    Return the associated System for a given Object.
    @param  obj     The @ref DOFObject. This must not be NULL.
    @return    The associated @ref DOFSystem.
    @exception NULL is returned if @a obj is NULL.
**/
extern DOF_OAL_DECLSPEC DOFSystem DOFObject_GetSystem( const DOFObject obj );


/**
 *  Create a new copy of an existing @ref DOFObject.
 *  @mustfree{DOFObject_Destroy}
 *  @param  self The @ref DOFObject to be copied.  This should not be NULL.
 *  @return A new copy of the @ref DOFObject.
 *  @exception  NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DOFObject DOFObject_Copy( DOFObject self );


/** @name DOF Object Provider **/
/*@{*/


/**
 * Send an Advertisement for an Object and Interface ID.  This is only a
 * notification to the network that the node may be able to provide the
 * specified binding, but without actually providing the binding.  This can be
 * useful when an application first starts up or when a new connection is
 * established, to announce its ability to provide bindings in the network.
 * This triggers matching Interest to be directed toward the node.  It is not
 * necessary to call BeginProvide for the binding prior to this call.
 *
 * @param self  The Object.  This must not be NULL.
 * @param interfaceID  The Interface ID to be advertised on the Object.  This must not be NULL.
 * @param config  The Operation Configuration, or NULL for defaults.
 */
extern DOF_OAL_DECLSPEC void DOFObject_Advertise( const DOFObject self, DOFInterfaceID interfaceID, DOFOperationControl config );


/**
    Signal that an Event occurred.  The signal is sent to all Requestors in
    compatible domains that have registered an Event handler for the Event, on
    the given Object.
    @param  obj     The @ref DOFObject on which the Event occurred.  This should not be NULL.
    @param  evt     The @ref DOFInterfaceEvent that occurred.  This should not be NULL.
    @param  parameterCount  The count of @ref DOFValue elements in @a parameters.
    @param  parameters  The Event parameters, a list of @ref DOFValue elements.  This may be NULL
                    if there are no parameters for the Event. Otherwise, the values should match the parameters defined by the Event.
    @note Though we are not notified, no event signal will be sent if
        - @a evt is NULL, or
        - @a parameters is NULL, or
        - @a evt not a DOFInterfaceEvent, or
        - A value type within @a parameters is not compatible with a type within evt's input parameter type list.
        - DOFResult could not be created, or
        - @a evt not registered for obj.
    @return TRUE for success.
    @exception FALSE is returned if
        - @a obj is NULL or if its @ref DOFObjectID is NULL, or
        - @a evt is NULL or is not an @ref DOFInterfaceEvent, or
        - @a parameters is NULL or its parameter list is not compatible with the parameters of @a evt, or
        - memory allocation failed.
**/
extern DOF_OAL_DECLSPEC boolean DOFObject_Signal( DOFObject obj, const DOFInterfaceEvent evt, uint16 parameterCount, const DOFValue parameters[] );


/**
    Ask whether any Requestors have registered an Event handler for the given object of this provider.  This allows
    an application to decide whether or not to create the event parameters and call @ref DOFObject_Signal.  When this
    function returns FALSE, a @ref DOFObject_Signal call will not do anything (so it can safely be skipped).
    @param  obj     The @ref DOFObject on which to operate. This should not be NULL.
    @param  evt     The @ref DOFInterfaceEvent in question. This should not be NULL.
    @return         TRUE    if @a evt is registered for this object.
    @return         FALSE   if @a evt is not registered for this object.
    @exception      FALSE is returned if
                    - @a obj is NULL, or
                    - @a evt is NULL or is not an @ref DOFInterfaceEvent.
**/
extern DOF_OAL_DECLSPEC boolean DOFObject_IsSignalRequired( const DOFObject obj, const DOFInterfaceEvent evt );


/**
    Report that a Property value changed. This report is sent to all Requestors that have subscribed to the given
    property on the given DOFObject.

    @param obj The object on which the property changed.  This should not be NULL.
    @param property The Property that changed.  This should not be NULL.
    @return TRUE. If @a prp is subscribed for @a obj, the change report is sent.
    @exception FALSE is returned (and the change report is not sent) if
         - @a property is NULL, or
         - @a property not a DOFInterfaceProperty, or
         - @a property not readable.
**/
extern DOF_OAL_DECLSPEC boolean DOFObject_Changed( DOFObject obj, const DOFInterfaceProperty property );


/**
    Register a set of callbacks for Providing (or implementing) an Interface on
    an Object.  This notifies the system that the given Interface is now
    available on the given Object.  For the duration of the @ref DOFOperation,
    the Provider callbacks will be used to handle requests to the Object for
    the given Interface.  As with any @ref DOFOperation, this "operation" can
    be extended or cancelled.

    @mustfree{DOFOperation_Destroy}

    @param  obj       The @ref DOFObject on which to Provide. This should not be
                      NULL and should represent a single object (neither the
                      broadcast nor a multicast object).
    @param  iface     The @ref DOFInterface to Provide. This should not be NULL.
    @param  timeout   The duration for the Provide "operation", in milliseconds.  The timeout begins after the operation has been processed and sent to appropriate connections.
    @param  callback  The callbacks to use while providing this Interface on this Object.  This may be NULL
                      (or any of its functions) if they do not apply or do not need to be implemented for the provider.
    @param  context   Application context data associated with the "operation".  This may be NULL.
    @return A reference to the new provide @ref DOFOperation.
    @exception NULL is returned if , or
            - an invalid parameter is detected; or
            - insufficient resources exist to create the operation; or
            - this node is already providing for @a iface on @a obj.
**/
extern DOF_OAL_DECLSPEC DOFOperation DOFObject_BeginProvide( DOFObject obj, DOFInterface iface, uint32 timeout, const DOFObjectProvider callback, void * context );


/*@}*/


/** @name DOF Object Requestor **/
/*@{*/


/**
    Request that an Interface definition be located in the network and retrieved.
    There is no guarantee that a definition can be found.

    @mustfree{DOFInterface_Destroy}

    @param  self          The Object on which to make the request.  This must not be NULL.
    @param  interfaceID   The identifier for the Interface to be defined.  This should not be NULL, and should not be the wildcard IID.
    @param  config        Configuration options for the request operation (NULL for default configuration).
    @param  timeout       The maximum amount of time to wait for the definition, in milliseconds.  The timeout begins after the operation has been processed and sent to appropriate connections.
    @exceptionparam

    @return The Interface retrieved.

    @exception NULL is returned and @a pException is set with one of the following errors:
                 - @ref DOFERROR_ILLEGAL_ARGUMENT if an invalid argument is detected.
                 - @ref DOFERROR_NOT_FOUND if the Interface Definition could not be found in the network.
                 - @ref DOFERROR_INSUFFICIENT_RESOURCES if the Interface could not be allocated.
**/
extern DOF_OAL_DECLSPEC DOFInterface DOFObject_Define( const DOFObject self, const DOFInterfaceID interfaceID, DOFOperationControl config, uint32 timeout, DOFException * pException );


/**
    Request that an Interface definition be located in the network and retrieved.
    There is no guarantee that a definition can be found.

    @mustfree{DOFOperation_Destroy}

    @param  self          The Object on which to make the request.  This must not be NULL.
    @param  interfaceID   The identifier for the Interface to be defined.  This should not be NULL, and should not be the wildcard IID.
    @param  config        Configuration options for the request operation (NULL for default configuration).
    @param  timeout       The maximum amount of time to wait for the definition, in milliseconds.  The timeout begins after the operation has been processed and sent to appropriate connections.
    @param  callback      The callback through which responses and the
                          completion notification will be received.  This may be null to ignore notification
                          of the received interface.  The definition can be retrieved by calling @ref DOFOperationDefine_WaitInterface.
    @param  context       Application context data associated with the asynchronous operation.

    @return A representation of the define operation.
    @exception NULL is returned if
            - an invalid parameter is detected; or
            - insufficient resources exist to create the operation.
**/
extern DOF_OAL_DECLSPEC DOFOperation DOFObject_BeginDefine( const DOFObject self, const DOFInterfaceID interfaceID, DOFOperationControl config, uint32 timeout, const DOFObjectDefineCallback callback, void * context );


/**
    Get a Property's value. The specified timeout is a maximum duration for the
    operation, and the operation fails if the timeout is reached before a
    response is received.  Otherwise, the first response received is returned.
    Broadcast operations will be cancelled immediately after the first response,
    and therefore may not reach or be processed by all available providers.

    @note Beware that a timeout of @ref DOF_TIMEOUT_NEVER may NEVER return if a response is not received!

    @mustfree{DOFResult_Destroy}

    @param obj             The object from which to get the Property's value. This should not be NULL.
    @param property        The Property to get. This should not be NULL.
    @param config          Configuration options for the request operation (NULL for default configuration).
    @param timeout         The time we wait for results, in milliseconds. The timeout begins after the operation has been processed and sent to appropriate connections.
    @exceptionparam
    @return The result of the Get operation, indicating success and containing the returned property value.
    @exception NULL is returned and @a pException is filled (if not null) with an appropriate exception in the following conditions:
            - The provider returned an exception
            - An invalid parameter was detected
            - Insufficient resources available to perform the operation
**/
extern DOF_OAL_DECLSPEC DOFResult DOFObject_Get( DOFObject obj, const DOFInterfaceProperty property, DOFOperationControl config, uint32 timeout, DOFException * pException );


/**
    Set a Property's value. The specified timeout is a maximum duration for the
    operation, and the operation fails if the timeout is reached before an
    acknowledgement is received.  Otherwise, the first acknowledgement received
    is returned.  Broadcast operations will be cancelled immediately after the
    first acknowledgement, and therefore may not reach or be processed by all
    available providers.

    @note Beware that a timeout of @ref DOF_TIMEOUT_NEVER may NEVER return if an acknowledgement is not received!

    @mustfree{DOFResult_Destroy}

    @param obj             The object on which to set the Property's value. This should not be NULL.
    @param property        The Property to set. This should not be NULL.
    @param value           The new value for the Property. The data type must match the property's data type.  NULL is permitted for a nullable type.
    @param config          Configuration options for the request operation (NULL for default configuration).
    @param timeout         The time we wait for results, in milliseconds. The timeout begins after the operation has been processed and sent to appropriate connections.
    @exceptionparam
    @return The result of the Set operation, indicating success.
    @exception NULL is returned and @a pException is filled (if not null) with an appropriate exception in the following conditions:
            - The provider returned an exception
            - An invalid parameter was detected (including an incompatible @a value passed for the @a property)
            - Insufficient resources available to perform the operation
**/
extern DOF_OAL_DECLSPEC DOFResult DOFObject_Set( DOFObject obj, const DOFInterfaceProperty property, DOFValue value, DOFOperationControl config, uint32 timeout, DOFException * pException );


/**
    Invoke a Method. The specified timeout is a maximum duration for the
    operation, and the operation fails if the timeout is reached before a
    response is received.  Otherwise, the first response received is returned.
    Broadcast operations will be cancelled immediately after the first response,
    and therefore may not reach or be processed by all available providers.

    @note Beware that a timeout of @ref DOF_TIMEOUT_NEVER may NEVER return if a response is not received!

    @param obj             The object on which to make the invocation.  This should not be NULL.
    @param method          The Method to invoke. This should not be NULL.
    @param parameterCount  The count of @ref DOFValue elements in @a parameters.
    @param parameters      The list of input parameters to the Method.  This may be NULL
                           if there are no parameters for the Method. Otherwise, the values should match the input parameters defined by the Method.
    @param config Configuration options for the request operation (NULL for default configuration).
    @param timeout         The time we wait for results, in milliseconds.  The timeout begins after the operation has been processed and sent to appropriate connections.
    @exceptionparam
    @return The result of the Invoke operation, indicating success and containing the returned method values.
    @exception NULL is returned and @a pException is filled (if not null) with an appropriate exception in the following conditions:
            - The provider returned an exception
            - An invalid parameter was detected (including incompatible @a paramaters or @a parameterCount passed for the @a method)
            - Insufficient resources available to perform the operation
**/
extern DOF_OAL_DECLSPEC DOFResult DOFObject_Invoke( DOFObject obj, const DOFInterfaceMethod method, uint16 parameterCount, const DOFValue parameters[], DOFOperationControl config, uint32 timeout, DOFException * pException );


/**
    Start an asynchronous get of a Property's value.  The specified timeout is a
    maximum duration for the operation, and the operation fails if the timeout
    is reached before a response is received.  The operation may complete
    sooner if a response is received, but multicast operations always wait for
    the full timeout.  Responses and notification of operation completion are
    received through the specified callback.  As with any @ref DOFOperation,
    this operation can be extended or cancelled.

    @mustfree{DOFOperation_Destroy}

    @param  obj           The object from which to get the Property's value. This should not be NULL.
    @param  property      The Property to get. This should not be NULL.
    @param  config        Configuration options for the request operation (NULL for default configuration).
    @param  timeout       The duration for the asynchronous operation, in milliseconds.  The timeout begins after the operation has been processed and sent to appropriate connections.
    @param  callback      The callback through which responses and the completion notification will be received.  This may be null to ignore notification
                          of the received value.  The value (result) can be retrieved by calling @ref DOFOperationGet_WaitResult.
    @param  context       Application context data associated with the asynchronous operation.

    @return A representation of the operation, or NULL if
            - an invalid parameter is detected; or
            - insufficient resources exist to create the operation.
**/
extern DOF_OAL_DECLSPEC DOFOperation DOFObject_BeginGet( DOFObject obj, const DOFInterfaceProperty property, DOFOperationControl config, uint32 timeout, const DOFObjectGetCallback callback, void * context );


/**
    Start an asynchronous set of a Property's value.  The specified timeout is a
    maximum duration for the operation, and the operation fails if the timeout
    is reached before an acknowledgement is received.  The operation may
    complete sooner if acknowledgement is received, but multicast operations
    always wait for the full timeout.  Notification of operation completion is
    received through the specified callback.  As with any @ref DOFOperation,
    this operation can be extended or cancelled.

    @mustfree{DOFOperation_Destroy}

    @param  obj           The object on which to set the Property's value. This should not be NULL.
    @param  property      The Property to set. This should not be NULL.
    @param  value         The new value for the Property. The data type must match the property's data type.  NULL is permitted for a nullable type.
    @param  config        Configuration options for the request operation (NULL for default configuration).
    @param  timeout       The duration for the asynchronous operation, in milliseconds.  The timeout begins after the operation has been processed and sent to appropriate connections.
    @param  callback      The callback through which the completion notification will be received.  This may be null to ignore notification
                          of the set completion.  The result can be retrieved by calling @ref DOFOperationSet_WaitResult.
    @param  context       Application context data associated with the asynchronous operation.

    @return A representation of the operation, or NULL if
            - an invalid parameter is detected (including incompatible @a value
              passed for the @a property); or
            - insufficient resources exist to create the operation.
**/
extern DOF_OAL_DECLSPEC DOFOperation DOFObject_BeginSet( DOFObject obj, const DOFInterfaceProperty property, DOFValue value, DOFOperationControl config, uint32 timeout, const DOFObjectSetCallback callback, void * context );


/**
    Start an asynchronous Method invocation.  The specified timeout is a
    maximum duration for the operation, and the operation fails if the timeout
    is reached before a response is received.  The operation may complete
    sooner if a response is received, but multicast operations always wait for
    the full timeout.  Responses and notification of operation completion are
    received through the specified callback.  As with any @ref DOFOperation,
    this operation can be extended or cancelled.

    @mustfree{DOFOperation_Destroy}

    @param  obj           The object on which to make the invocation. This should not be NULL.
    @param  method        The Method to invoke. This should not be NULL.
    @param  parameterCount The count of @ref DOFValue elements in @a parameters.
    @param  parameters    The list of input parameters to the Method.  This may be NULL
                          if there are no parameters for the Method. Otherwise, the values should match the input parameters defined by the Method.
    @param  config        Configuration options for the request operation (NULL for default configuration).
    @param  timeout       The duration for the asynchronous operation, in milliseconds.  The timeout begins after the operation has been processed and sent to appropriate connections.
    @param  callback      The callback through which responses and the completion notification will be received.  This may be null to ignore notification
                          of the received values.  The values (result) can be retrieved by calling @ref DOFOperationInvoke_WaitResult.
    @param  context       Application context data associated with the asynchronous operation.

    @return A representation of the operation, or NULL if
            - an invalid parameter is detected (including incompatible @a
              paramaters or @a parameterCount passed for the @a method); or
            - insufficient resources exist to create the operation.
**/
extern DOF_OAL_DECLSPEC DOFOperation DOFObject_BeginInvoke( DOFObject obj, const DOFInterfaceMethod method, uint16 parameterCount, const DOFValue parameters[], DOFOperationControl config, uint32 timeout, const DOFObjectInvokeCallback callback, void * context );


/**
    Get the list of Interfaces currently being provided on the Object.

    This does not block, as it is a simple query against the current state of
    the system.  There is no request to the network to identify additional
    Interfaces that may be available.  The Interface IDs returned may be
    limited by the Interest operations that exist.

    @mustfree{DOFInterfaceIDList_Destroy}

    @param  object      The Object for which to make the query. This should not be NULL.

    @return A new list of Interface IDs. This list can have 0 members.

    @exception NULL is returned if the list could not be allocated.
**/
extern DOF_OAL_DECLSPEC DOFInterfaceIDList DOFObject_GetInterfaces( const DOFObject object );


/**
    Register a requestor callback for watching the value of a property over time, and ask the provider of the submitted
    DOFObject and submitted DOFInterfaceProperty to notify us of such Property value changes. For the duration of the @ref
    DOFOperation, changes detected in the Property will be communicated to us through the callback, based on the given
    timing parameters. As with any @ref DOFOperation, this operation can be extended or cancelled.

    @mustfree{DOFOperation_Destroy}

    @note The Property's current value is always returned when the subscription is started.
    @note Intermediate changes that occur more frequently than minPeriod may be dropped to save bandwidth,
    but the most current value of the Property is always returned.

    @param  obj           The object on which to watch the Property. This should not be NULL.
    @param  property      The Property to watch. This should not be NULL.
    @param  minPeriod     The preferred minimum time period between change
                          notifications, in milliseconds.  This is used only as
                          a guideline; updates may occur more (or less)
                          frequently than indicated here.
    @param  config        Configuration options for the request operation (NULL for default configuration).
    @param  timeout       The duration for the asynchronous operation, in milliseconds (specifies the duration of the subscription).  The timeout begins after the operation has been processed and sent to appropriate connections.
    @param  callback      The callback through which responses and the completion notification will be received.  This may be null to ignore notification
                          of the received values during the subscription.  However, results are only reported to the callback.
    @param  context       Application context data associated with the asynchronous operation.

    @return A representation of the subscribe operation.
    @exception NULL is returned if
            - an invalid parameter is detected; or
            - insufficient resources exist to create the operation.
**/
extern DOF_OAL_DECLSPEC DOFOperation DOFObject_BeginSubscribe( DOFObject obj, const DOFInterfaceProperty property, uint32 minPeriod, DOFOperationControl config, uint32 timeout, const DOFObjectSubscribeCallback callback, void * context );


/**
    Register a requestor callback for notifying us when the submitted provider event is signaled, and ask the provider
    of that event and the submitted DOFObject to signal the event. For the duration of the @ref DOFOperation, the
    callback will be called each time an Event is signaled, along with associated data sent by the provider.  As with
    any @ref DOFOperation, this operation can be extended or cancelled.

    @mustfree{DOFOperation_Destroy}

    @param  obj           The object on which to watch for the Event. This should not be NULL.
    @param  evt           The Event to watch. This should not be NULL.
    @param  config        Configuration options for the request operation (NULL for default configuration).
    @param  timeout       The duration for the asynchronous operation, in milliseconds (specifies the duration of the registration).  The timeout begins after the operation has been processed and sent to appropriate connections.
    @param  callback      The callback through which responses and the completion notification will be received.  This may be null to ignore notification
                          of the received events.  Events are only reported to the callback.
    @param  context       Application context data associated with the asynchronous operation.

    @return A representation of the operation, or NULL if
            - an invalid parameter is detected; or
            - insufficient resources exist to create the operation.
**/
extern DOF_OAL_DECLSPEC DOFOperation DOFObject_BeginRegister( DOFObject obj, const DOFInterfaceEvent evt, DOFOperationControl config, uint32 timeout, const DOFObjectEventCallback callback, void * context );


/**
    Start an asynchronous request to open  a Session on an object.  The system will allow up to
    the specified timeout to establish a new Session for the given Interface on
    the given Object.  Operation completion or failure is reported through the
    specified callback.  As with any @ref DOFOperation, this operation can be
    extended or cancelled.

    @mustfree{DOFOperation_Destroy}

    @note Attempting to open a session upon an existing session (where the given
          Object already represents a session) is not supported.

    @param  obj           The object on which to request the Session. This should not be NULL.
    @param  interface     The Interface on which to request the Session. This should not be NULL.
    @param  sessionType   The Session primary Interface ID (a @ref DOFInterfaceID).  This interface should be provided on the session (possibly in addition to others), and defines the purpose and contract for the session.  This should not be NULL.
    @param  config        Configuration options for the request operation (NULL for default configuration).
    @param  timeout       The maximum duration for the asynchronous operation, in milliseconds (specifies the duration of the session).  The timeout begins after the operation has been processed and sent to appropriate connections.
    @param  callback      The callback through which the completion or failure notification will be received.  This may be null to ignore notification
                          of the received session.  The session can be retrieved by calling @ref DOFOperationSession_WaitSession.
    @param  context       Application context data associated with the asynchronous operation.

    @return A representation of the operation, or NULL if
            - an invalid parameter is detected; or
            - insufficient resources exist to create the operation; or
            - @a obj already represents a session.
**/
extern DOF_OAL_DECLSPEC DOFOperation DOFObject_BeginSession( DOFObject obj, const DOFInterface interface, const DOFInterfaceID sessionType, DOFOperationControl config, uint32 timeout, const DOFObjectSessionCallback callback, void * context );


/**
    Determine if there is the specified level of interest currently expressed for a given Interface on the Object.
    If @a interfaceID is the Wildcard IID, then check for @b all interfaces on the Object.
    If @a interfaceID is NULL, then check for @b any interfaces on the given Object.

    @param self        The Object.  This must not be NULL.
    @param interfaceID The Interface ID.  This may be NULL, to indicate any interface on the specified object.
    @param level       The level of interest for which to check.  Current interest at or above the specified level will give an affirmative result.

    @retval Zero (FALSE) if the specified level of interest is not present for the given Interface on the Object.
    @retval Non-Zero (TRUE) if the specified level of interest is present for the given Interface on the Object.
**/
extern DOF_OAL_DECLSPEC boolean DOFObject_IsInterestRequested( const DOFObject self, const DOFInterfaceID interfaceID, DOFInterestLevel level );


/*@}*/


/**
@}
**/


#ifdef  __cplusplus
}
#endif


#endif /* DOF_OBJECT_H */
