/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_OBJECTID_H
#define DOF_OBJECTID_H


#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif

/**
@}
**/


/** @defgroup DOFObjectID DOFObjectID
    @ingroup DOF_Framework

    DOF Object Identifier

    This class implements a DOF Object Identifier.  A @ref DOFObjectID may
    be a single ID unit of a specific class or it may be extended with one or
    more ID units called Attributes.  It is self descriptive, allowing its ID
    class to be identified and its attribute extensions to be extracted.

    DOFObjectIDs are immutable, and should be treated as such by the application.
    With the calls of this API, DOFObjectIDs are created by the system. It is up
    to the application to later destroy any DOFObjectID that it created.

@{
**/


/** The C OAL provides and maintains this broadcast Object ID, so applications need not create/destroy their own. **/
extern DOF_OAL_DECLSPEC DOFObjectID DOFOBJECTID_BROADCAST;

/** The C OAL provides and maintains this broadcast domain Object ID, so applications need not create/destroy their own. **/
extern DOF_OAL_DECLSPEC DOFObjectIDDomain DOFOBJECTIDDOMAIN_BROADCAST;


/* DOFObjectID Methods */

/**
    Create a new @ref DOFObjectID from a given class and data.
    This constructs a new @ref DOFObjectID without any Attributes.
    @mustfree{DOFObjectID_Destroy}
    @note It is the responsibility of the caller to ensure the given class is registered and that the given data
          satisfies the requirements for that class.  Refer to the documentation of the specific @ref DOFObjectIDClass,
          or the DOF Common Types Specification for more information on object ID classes, formats, and
          requirements.
    @param idClass  The @ref DOFObjectIDClass for this identifier. Use of unknown/unregistered classes is permitted
                    (e.g., classes defined after this library was built); however, classes must be registered with the
                    OpenDOF Project prior to use.  Valid range is 0 <= @a idClass < 2^30.
    @param dataSize The size, in bytes, of the given data field at @a data. Valid range is 0 <= @a dataSize <= 63.  This
                    should be 0 if @a idClass is @ref DOFOBJECTIDCLASS_BROADCAST.
    @param data     Pointer to the given data. This is ignored (and may be NULL) if @a dataSize is 0.  Otherwise, this
                    must be a valid pointer to at least @a dataSize bytes of data, and appropriate for the class of Object
                    ID being constructed.
    @return    A new DOFObjectID.
    @exception NULL is returned if
        - an invalid parameter is detected; or
        - insufficient resources exist to create the @ref DOFObjectID.
**/
extern DOF_OAL_DECLSPEC DOFObjectID DOFObjectID_Create( DOFObjectIDClass idClass, uint32 dataSize, const uint8 * data );

/**
    Create a new @ref DOFObjectID from a given class and data string.
    This constructs a @ref DOFObjectID without any Attributes.
    @mustfree{DOFObjectID_Destroy}
    @note It is the responsibility of the caller to ensure the given class is registered and that the given data
          satisfies the requirements for that class.  Refer to the documentation of the specific @ref DOFObjectIDClass,
          or the DOF Common Types Specification for more information on object ID classes, formats, and
          requirements.
    @param idClass  The @ref DOFObjectIDClass for this identifier. Use of unknown/unregistered classes is permitted
                    (e.g., classes defined after this library was built); however, classes must be registered with the
                    OpenDOF Project prior to use.  Valid range is 0 <= @a idClass < 2^30.
    @param data     The data, as a null-terminated C string. The data size is inferred by the string length.  This
                    should not be NULL, and should be a C string of 63 characters or less, and appropriate for the class
                    of Object ID being constructed. No interpretation, transformation, or parsing of the string is made.
    @return    A new DOFObjectID.
    @exception NULL is returned if
        - an invalid parameter is detected; or
        - insufficient resources exist to create the @ref DOFObjectID.
**/
extern DOF_OAL_DECLSPEC DOFObjectID DOFObjectID_Create_String( DOFObjectIDClass idClass, const char * data );

/**
 *  Create a new copy of an existing @ref DOFObjectID.
 *  @mustfree{DOFObjectID_Destroy}
 *  @param  self The @ref DOFObjectID to be copied.  This should not be NULL.
 *  @return A new copy of the @ref DOFObjectID.
 *  @exception  NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DOFObjectID DOFObjectID_Copy( const DOFObjectID self );

/**
    Create a @ref DOFObjectID from a given @ref DOFObjectID buffer.
    @mustfree{DOFObjectID_Destroy}
    @param bufferSize The size, in bytes, of the given data buffer at @a buffer.
    @param buffer     Pointer to the given buffer. This should not be NULL, and should represent a valid @ref
                      DOFObjectID in wire format of the given size.
    @return    A newly created DOFObjectID.
    @exception NULL is returned if
        - an invalid parameter is detected; or
        - insufficient resources exist to create the @ref DOFObjectID.
**/
extern DOF_OAL_DECLSPEC DOFObjectID DOFObjectID_Create_Bytes( uint32 bufferSize, const uint8 * buffer );

/**
    Get pointer to the standardized wire-format bytes of the @ref DOFObjectID.
    The value returned includes all header, class, length, attribute, and data
    fields present.
    @param self The @ref DOFObjectID.  This must not be NULL.
    @return Pointer to the wire-format of @a self.
**/
extern DOF_OAL_DECLSPEC const uint8 * DOFObjectID_GetBytes( const DOFObjectID self );

/**
    Return the standardized wire-format length of the @ref DOFObjectID.
    The value returned includes all header, class, length, attribute, and data
    fields present.
    @param self The @ref DOFObjectID.  This must not be NULL.
    @return    Size, in byte count, of the wire-format of @a self.
**/
extern DOF_OAL_DECLSPEC DOFObjectIDSize DOFObjectID_GetBytesSize( const DOFObjectID self );


/**
    Create a new @ref DOFObjectID by augmenting an existing @ref DOFObjectID with specified Attribute type and a value
    buffer.
    @mustfree{DOFObjectID_Destroy}
    @note This function stores Attributes in increasing numerical order by type, as required by the protocol specification.
    @param self              The original @ref DOFObjectID, which remains unchanged. This should not be NULL.
    @param type              The ID of the new Attribute.  See @ref DOFObjectIDAttributeType.  Valid numeric range is
                             0 <= @a type <= 127.
    @param attributeDataSize Size, in byte count, of the supplied new Attribute's data. Valid range is 0 <= @a
                             attributeDataSize <= 255.
    @param attributeData     The new Attribute's data.  This should not be NULL if @a valueSize > 0.  The data must be
                             appropriate for the given attribute type.
    @return    The new @ref DOFObjectID.
    @exception NULL is returned if
        - an invalid parameter is detected; or
        - the existing @a objectID already contains an Attribute with the same @a type; or
        - insufficient resources exist to create the @ref DOFObjectID.
**/
extern DOF_OAL_DECLSPEC DOFObjectID DOFObjectID_Create_AttributeBuffer( const DOFObjectID self, const DOFObjectIDAttributeType type, uint32 attributeDataSize, const uint8 * attributeData );

/**
    Create a new @ref DOFObjectID by augmenting an existing @ref DOFObjectID with specified Attribute type and a string.
    @mustfree{DOFObjectID_Destroy}
    @note This function stores Attributes in increasing numerical order by type, as required by the protocol specification.
    @param self      The original @ref DOFObjectID, which remains unchanged. This should not be NULL.
    @param type      The ID of the new Attribute.  See @ref DOFObjectIDAttributeType.  Valid numeric range is 0 <= @a
                     type <= 127.
    @param attributeString The new Attribute's data as a C string. This should not be NULL, and its C string length
                     should be 255 characters or less.  The data must be appropriate for the given attribute type.
    @return    The new @ref DOFObjectID.
    @exception NULL is returned if
        - an invalid parameter is detected; or
        - the existing @a objectID already contains an Attribute with the same @a type; or
        - insufficient resources exist to create the @ref DOFObjectID.
**/
extern DOF_OAL_DECLSPEC DOFObjectID DOFObjectID_Create_AttributeString( const DOFObjectID self, const DOFObjectIDAttributeType type, const char * attributeString );

/**
    Create a new @ref DOFObjectID by augmenting an existing @ref DOFObjectID with specified Attribute type and a @ref
    DOFObjectID.
    @mustfree{DOFObjectID_Destroy}
    @note This function stores Attributes in increasing numerical order by type, as required by the protocol specification.
    @param self      The original @ref DOFObjectID, which remains unchanged. This should not be NULL.
    @param type      The ID of the new Attribute.  See @ref DOFObjectIDAttributeType.  Valid numeric range is 0 <= @a
                     type <= 127.
    @param objectID  The new Attribute's data as a @ref DOFObjectID. This should not be NULL, and the total length of
                     the object ID (in byte form, see @ref DOFObjectID_GetBytesSize) should be 255 bytes or less. The
                     data must be appropriate for the given attribute type.
    @return    The new @ref DOFObjectID.
    @exception NULL is returned if
        - an invalid parameter is detected; or
        - the existing @a objectID already contains an Attribute with the same @a type; or
        - insufficient resources exist to create the @ref DOFObjectID.
**/
extern DOF_OAL_DECLSPEC DOFObjectID DOFObjectID_Create_AttributeObjectID( const DOFObjectID self, const DOFObjectIDAttributeType type, const DOFObjectID objectID );

/**
    Destroy a @ref DOFObjectID.
    @param  self The @ref DOFObjectID to be destroyed.  If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFObjectID_Destroy( DOFObjectID self );

/**
    Create a new @ref DOFObjectID using its standard string representation.
    @mustfree{DOFObjectID_Destroy}
    @param data     The data, as a null-terminated C string. The data size is inferred by the string length.  This
                    should not be NULL, and should be a C string representation of the DOFObjectID in standard string
                    format.
    @return     The new @ref DOFObjectID.
    @exception NULL is returned if
        - an invalid parameter is detected
        - insufficient resources exist to create the @ref DOFObjectID.

**/
extern DOF_OAL_DECLSPEC DOFObjectID DOFObjectID_Create_StandardString( const char * data );


/**
    Convert a @ref DOFObjectID into its standard string representation.
    @mustfree{PCRMem_Free}
    The standardized Object ID string format is described in the DOF Common Types document.
    @note The output is a null-terminated, readable string suitable for output in debug or log messages.
    @param self The @ref DOFObjectID to convert. This should not be NULL.
    @return    Pointer to the newly generated string buffer.
    @exception NULL is returned if
        - an invalid parameter is detected; or
        - insufficient resources exist to create the string.
**/
extern DOF_OAL_DECLSPEC char * DOFObjectID_CreateStandardString( const DOFObjectID self );

/**
    Compare two @ref DOFObjectID's. This can also be used as a comparator function for sorting.
    @note If both are NULL, they are considered equal. If only one is NULL, it is considered less than the other.
    @param  oid1 The first @ref DOFObjectID to compare.
    @param  oid2 The second @ref DOFObjectID to compare.
    @return Zero           if exact match.
    @return Positive value if @a oid1 > @a oid2.
    @return Negative value if @a oid1 < @a oid2.
**/
extern DOF_OAL_DECLSPEC int32 DOFObjectID_Compare( const DOFObjectID oid1, const DOFObjectID oid2 );

/**
    Compare two @ref DOFObjectID's for equality, including Attribute extensions.
    @note If both parameters are NULL, they are considered equal. If only one is NULL, they are considered not equal.
    @param  self The @ref DOFObjectID to be compared.
    @param  objectID The @ref DOFObjectID to compare against.

    @retval Non-Zero (TRUE) if the two IDs match exactly.
    @retval Zero (FALSE) if the two IDs do not match.
**/
extern DOF_OAL_DECLSPEC boolean DOFObjectID_IsEqual( const DOFObjectID self, const DOFObjectID objectID );

/**
    Compare two DOFObjectOIDs, excluding attribute extensions.
    @note If one parameter is NULL, it is considered less than the other. If both are NULL, they are considered equal.
    @param  oid1 The first @ref DOFObjectID to compare.
    @param  oid2 The second @ref DOFObjectID to compare.
    @return Zero           if exact match.
    @return Positive value if @a oid1 > @a oid2.
    @return Negative value if @a oid1 < @a oid2.
**/
extern DOF_OAL_DECLSPEC int32 DOFObjectID_CompareIgnoreAttributes( const DOFObjectID oid1, const DOFObjectID oid2 );

/**
    Compare two DOFObjectOIDs, excluding Attribute extensions.
    @note If both parameters are NULL, they are considered equal. If only one is NULL, they are considered not equal.
    @param  self The @ref DOFObjectID to be compared.
    @param  objectID The @ref DOFObjectID to compare against.
    @retval Non-Zero (TRUE) if the two IDs match exactly, excluding Attribute extensions.
    @retval Zero (FALSE) if the two IDs do not match.
**/
extern DOF_OAL_DECLSPEC boolean DOFObjectID_IsEqualIgnoreAttributes( const DOFObjectID self, const DOFObjectID objectID );

/**
    Compare the attribute extensions of two DOFObjectOIDs, excluding the base ID.
    @note If one parameter is NULL, it is considered less than the other. If both are NULL, they are considered equal.
    @param  oid1 The first @ref DOFObjectID to compare.
    @param  oid2 The second @ref DOFObjectID to compare.
    @return Zero           if exact match.
    @return Positive value if @a oid1 > @a oid2.
    @return Negative value if @a oid1 < @a oid2.
**/
extern DOF_OAL_DECLSPEC int32 DOFObjectID_CompareIgnoreBase( const DOFObjectID oid1, const DOFObjectID oid2 );

/**
    Compare two @ref DOFObjectID's for sufficient match, including Attribute extensions. Sufficient match means both
    base @ref DOFObjectID's are identical and the @a self @ref DOFObjectID has exact matches for every attribute
    extension held by the @a objectID @ref DOFObjectID. Note that @a self may contain additional attributes as well.

    @note If both parameters are NULL, they are considered equal and sufficiently matched.
          If only a single parameter is NULL, they do NOT match sufficiently.
    @param  self     The @ref DOFObjectID to be compared.
    @param  objectID The @ref DOFObjectID to compare against.

    @retval Non-Zero (TRUE) if the two IDs are sufficiently matched.
    @retval Zero (FALSE) if the two IDs are not sufficiently matched.
**/
extern DOF_OAL_DECLSPEC boolean DOFObjectID_MatchesAtLeast( const DOFObjectID self, const DOFObjectID objectID );

/**
    Get the class of a @ref DOFObjectID.
    @param self The @ref DOFObjectID. This must not be NULL.
    @return The class of this DOFObjectID, one of the DOFOBJECTCLASS constants.
**/
extern DOF_OAL_DECLSPEC DOFObjectIDClass DOFObjectID_GetIDClass( const DOFObjectID self );


/**
    Get pointer to the data field of the base @ref DOFObjectID. This does not contain any attribute data information.
    @param self The @ref DOFObjectID.  This must not be NULL.
    @return    Pointer to the actual data field of the base @ref DOFObjectID.
    @exception NULL is returned if the data field of @a self is of length 0.
**/
extern DOF_OAL_DECLSPEC const uint8 * DOFObjectID_GetData( const DOFObjectID self );

/**
    Get the number of bytes comprising the data field of a @ref DOFObjectID.
    @param self The @ref DOFObjectID. This must not be NULL.
    @return    Size, in byte count, of the data field portion of the @ref DOFObjectID.
**/
extern DOF_OAL_DECLSPEC DOFObjectIDDataSize DOFObjectID_GetDataSize( const DOFObjectID self );

/**
    Ask if a @ref DOFObjectID includes Attribute extensions.
    @param self The @ref DOFObjectID. This should not be NULL.
    @return    Non-zero (TRUE) if Attribute extensions are included in the DOFObjectID, FALSE otherwise.
    @exception Zero (FALSE) is returned if the @ref DOFObjectID is NULL.
**/
extern DOF_OAL_DECLSPEC boolean DOFObjectID_HasAttributes( const DOFObjectID self );

/**
    Ask if a @ref DOFObjectID includes an Attribute extension having a specific ID.
    @param self The @ref DOFObjectID. This must not be NULL.
    @param type The ID of the Attribute to find within all available Attribute extensions.  See @ref
                DOFObjectIDAttributeType. Valid numeric range is 0 <= @a type <= 127.
    @return    Non-zero (TRUE) if an Attribute extension with matching Attribute id is found, FALSE otherwise.
    @exception Zero (FALSE) is returned if the @ref DOFObjectID is NULL.
**/
extern DOF_OAL_DECLSPEC boolean DOFObjectID_HasAttribute( const DOFObjectID self, const DOFObjectIDAttributeType type );

/**
    From a @ref DOFObjectID, get an Attribute extension having a specific ID.
    @param self The @ref DOFObjectID. This should not be NULL.
    @param type The ID of the Attribute to find within all available Attribute extensions. The @ref DOFObjectID should
                contain the requested attribute.  See @ref DOFObjectIDAttributeType.  Valid numeric range is 0 <= @a
                type <= 127.
    @return @ref DOFObjectIDAttribute for the found Attribute.
    @exception A @ref DOFObjectIDAttribute is returned with an ID of @ref DOFOBJECTIDATTRIBUTE_INVALID if the @ref
                 DOFObjectID is NULL, or if no matching attribute is found.
**/
extern DOF_OAL_DECLSPEC DOFObjectIDAttribute DOFObjectID_GetAttribute( const DOFObjectID self, const DOFObjectIDAttributeType type );

/**
    Get the N'th Attribute extension of a @ref DOFObjectID.
    @param self  The @ref DOFObjectID. This should not be NULL.
    @param index Specifies the N'th Attribute to find within all available Attribute ID extensions. This should be less
                 than the total count of attributes on the @ref DOFObjectID.
    @return    @ref DOFObjectIDAttribute for found N'th Attribute.
    @exception A @ref DOFObjectIDAttribute is returned with an ID of @ref DOFOBJECTIDATTRIBUTE_INVALID if the @ref
                 DOFObjectID is NULL, of if no N'th Attribute exists.
**/
extern DOF_OAL_DECLSPEC DOFObjectIDAttribute DOFObjectID_GetAttributeAtIndex( const DOFObjectID self, uint8 index );

/**
    Get the count of attributes associated with a @ref DOFObjectID.
    @param self The @ref DOFObjectID. This should not be NULL.
    @return    The count of attributes associated with the @ref DOFObjectID.
    @exception 0 is returned if the @ref DOFObjectID is NULL.
**/
extern DOF_OAL_DECLSPEC uint8 DOFObjectID_GetAttributeCount( const DOFObjectID self );

/**
    Determine if the @ref DOFObjectID is of class @ref DOFOBJECTIDCLASS_BROADCAST and has no attributes.

    @param self The @ref DOFObjectID.  This must not be NULL.
    @retval Non-zero (TRUE) if the ID is a broadcast @ref DOFObjectID, as described.
    @retval Zero (FALSE) if the ID is not a broadcast @ref DOFObjectID.
 */
extern DOF_OAL_DECLSPEC boolean DOFObjectID_IsBroadcast( const DOFObjectID self );

/**
  Determine if the @ref DOFObjectID is of class @ref DOFOBJECTIDCLASS_BROADCAST and has at least one attribute.

  @param self The @ref DOFObjectID.  This must not be NULL.
  @retval Non-zero (TRUE) if the ID is a multicast @ref DOFObjectID, as described.
  @retval Zero (FALSE) if the ID is not a multicast @ref DOFObjectID.
 */
extern DOF_OAL_DECLSPEC boolean DOFObjectID_IsMulticast( const DOFObjectID self );

/**
 * Determine if the DOFObjectID is a unicast @ref DOFObjectID (neither a broadcast nor a multicast @ref DOFObjectID).
 * @param self The @ref DOFObjectID.  This must not be NULL.
 * @retval Non-zero (TRUE) if the ID is a unicast @ref DOFObjectID, as described.
 * @retval Zero (FALSE) if the ID is not a unicast @ref DOFObjectID.
 */
extern DOF_OAL_DECLSPEC boolean DOFObjectID_IsUnicast( const DOFObjectID self );

/**
    Release resources previously allocated while forming a @ref DOFObjectID list.
    @param  oidList The DOFObjectIDList.  If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFObjectIDList_Destroy( DOFObjectIDList oidList );


/** @defgroup DOFObjectIDAttribute DOFObjectIDAttribute

    The Attribute class is used to retrieve attributes from a DOFObjectID. Attributes
    are used to add additional information to the object identifier, either to
    qualify its use or modify the way that the system uses it.

    There are a set of attribute types defined by this class, although this is
    not guaranteed to be exhaustive. Further types may be defined, and this
    library is flexible enough to allow access to them. However, these kinds
    of extensions must be registered with the OpenDOF Project and will be added to
    the library once approved.

@{
**/

/**
 * Determine if the attribute is a valid attribute.
 * This check can be made on returned attributes to determine if the value is valid or not.
 * @param attribute The @ref DOFObjectIDAttribute. This must not be NULL.
 * @retval Non-zero (@c TRUE) if the attribute is valid.
 * @retval Zero (@c FALSE) if the attribute is not valid.
**/
#ifdef DOXYGEN_MACRO_FUNCTIONS
extern DOF_OAL_DECLSPEC boolean DOFObjectIDAttribute_IsValid( const DOFObjectIDAttribute attribute );
#else
#define DOFObjectIDAttribute_IsValid( attribute ) ((attribute).id < DOFOBJECTIDATTRIBUTE_INVALID)
#endif

/**
 * Get the ID of the attribute.
 * @param attribute The @ref DOFObjectIDAttribute. This must not be NULL.
 * @return The ID of the attribute.
 */
#ifdef DOXYGEN_MACRO_FUNCTIONS
extern DOF_OAL_DECLSPEC DOFObjectIDAttributeType DOFObjectIDAttribute_GetType( const DOFObjectIDAttribute attribute );
#else
#define DOFObjectIDAttribute_GetType( attribute ) ((DOFObjectIDAttributeType)(attribute).id)
#endif

/**
 * Get the size of the data associated with the attribute.
 * @param attribute The @ref DOFObjectIDAttribute. This must not be NULL.
 * @return The size of the data associated with the attribute.
 */
#ifdef DOXYGEN_MACRO_FUNCTIONS
extern DOF_OAL_DECLSPEC DOFObjectIDAttributeDataSize DOFObjectIDAttribute_GetValueSize( const DOFObjectIDAttribute attribute );
#else
#define DOFObjectIDAttribute_GetValueSize( attribute ) ((attribute).dataSize)
#endif

/**
 * Get the data associated with the attribute.
 * @param attribute The @ref DOFObjectIDAttribute. This must not be NULL.
 * @return The data associated with the attribute.
 */
#ifdef DOXYGEN_MACRO_FUNCTIONS
extern DOF_OAL_DECLSPEC const uint8 * DOFObjectIDAttribute_GetValue( const DOFObjectIDAttribute attribute );
#else
#define DOFObjectIDAttribute_GetValue( attribute ) ((attribute).data)
#endif

/**
 * Get the data associated with the attribute as a string.
 * @param attribute The @ref DOFObjectIDAttribute. This must not be NULL.
 * @return The data associated with the attribute as a string.
 */
#ifdef DOXYGEN_MACRO_FUNCTIONS
extern DOF_OAL_DECLSPEC const char * DOFObjectIDAttribute_GetValueString( const DOFObjectIDAttribute attribute );
#else
#define DOFObjectIDAttribute_GetValueString( attribute ) ((const char *)(attribute).data)
#endif

/**
 * Create a new @ref DOFObjectID from the data associated with the attribute.
 * @mustfree{DOFObjectID_Destroy}
 * @param attribute The @ref DOFObjectIDAttribute. This must not be NULL.
 * @return The new @ref DOFObjectID.
 * @exception NULL is returned if a new @ref DOFObjectID cannot be created.
 */
#ifdef DOXYGEN_MACRO_FUNCTIONS
extern DOF_OAL_DECLSPEC DOFObjectID DOFObjectIDAttribute_CreateValueObjectID( const DOFObjectIDAttribute attribute );
#else
#define DOFObjectIDAttribute_CreateValueObjectID( attribute ) (DOFObjectID_Create_Bytes(DOFObjectIDAttribute_GetValueSize((attribute)), DOFObjectIDAttribute_GetValue((attribute))))
#endif

/**
@}
**/


/** @defgroup DOFObjectIDAuthentication DOFObjectIDAuthentication

    The DOFObjectIDAuthentication class represents a @ref DOFObjectID used for authentication. It may identify a user,
    application, device, or other entity with private data shared only with the domain's authentication server (AS).
    These are typically used by @ref DOFCredentials.

@{
**/

/**
   Create a new @ref DOFObjectIDAuthentication from the specified @ref DOFObjectID.

   @mustfree{DOFObjectIDAuthentication_Destroy}

   @param id The @ref DOFObjectID from which to create the authentication ID. This should not be NULL, should not have
             any attributes, and should not be Broadcast.

   @return The new @ref DOFObjectIDAuthentication.

   @exception NULL is returned if an invalid parameter is detected or the authentication ID could not be created.
**/
extern DOF_OAL_DECLSPEC DOFObjectIDAuthentication DOFObjectIDAuthentication_Create( const DOFObjectID id );

/**
 *  Create a new copy of an existing @ref DOFObjectIDAuthentication.
 *  @mustfree{DOFObjectIDAuthentication_Destroy}
 *  @param  self The @ref DOFObjectIDAuthentication to be copied.  This should not be NULL.
 *  @return A new copy of the @ref DOFObjectIDAuthentication.
 *  @exception  NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DOFObjectIDAuthentication DOFObjectIDAuthentication_Copy( const DOFObjectIDAuthentication self );

/**
   Get the @ref DOFObjectID from the specified @ref DOFObjectIDAuthentication.

   @param self The @ref DOFObjectIDAuthentication.  This must not be NULL.

   @return The @ref DOFObjectID.
**/
extern DOF_OAL_DECLSPEC DOFObjectID DOFObjectIDAuthentication_GetObjectID( const DOFObjectIDAuthentication self );

/**
   Destroy a @ref DOFObjectIDAuthentication.
   @param  self The @ref DOFObjectIDAuthentication to be destroyed.  If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFObjectIDAuthentication_Destroy( DOFObjectIDAuthentication self );

/**
@}
**/


/** @defgroup DOFObjectIDDomain DOFObjectIDDomain

    The DOFObjectIDDomain class represents a @ref DOFObjectID used to identify a domain. A domain is the context within
    which all DOF security is defined, maintained, and enforced.

@{
**/

/**
   Create a new @ref DOFObjectIDDomain from the specified @ref DOFObjectID.

   @mustfree{DOFObjectIDDomain_Destroy}

   @param id The @ref DOFObjectID from which to create the domain ID.  This should not be NULL, and should not have any
             attributes.

   @return The new @ref DOFObjectIDDomain.

   @exception NULL is returned if an invalid parameter is detected or the domain ID could not be created.
**/
extern DOF_OAL_DECLSPEC DOFObjectIDDomain DOFObjectIDDomain_Create( const DOFObjectID id );

/**
 *  Create a new copy of an existing @ref DOFObjectIDDomain.
 *  @mustfree{DOFObjectIDDomain_Destroy}
 *  @param  self The @ref DOFObjectIDDomain to be copied.  This should not be NULL.
 *  @return A new copy of the @ref DOFObjectIDDomain.
 *  @exception  NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DOFObjectIDDomain DOFObjectIDDomain_Copy( const DOFObjectIDDomain self );

/**
   Get the @ref DOFObjectID from the specified @ref DOFObjectIDDomain.

   @param self The @ref DOFObjectIDDomain.  This must not be NULL.

   @return The @ref DOFObjectID.
**/
extern DOF_OAL_DECLSPEC DOFObjectID DOFObjectIDDomain_GetObjectID( const DOFObjectIDDomain self );

/**
   Destroy a @ref DOFObjectIDDomain.
   @param  self The @ref DOFObjectIDDomain to be destroyed.  If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFObjectIDDomain_Destroy( DOFObjectIDDomain self );

/**
@}
**/


/** @defgroup DOFObjectIDSource DOFObjectIDSource

    The DOFObjectIDSource class represents a @ref DOFObjectID used to identify a source of network operations.  Source
    IDs must always be unique to ensure operations are managed correctly throughout the system. Typically they are
    randomly generated, but tend to be large. Applications can specify their own Source IDs for non-secure traffic, but
    must ensure that they do not reuse a previous Source ID that may still be in use within the network (e.g., by
    operations sent by the node before it rebooted).  Secure Source IDs are assigned by the AS and are not under
    application control.
@{
**/

/**
   Create a new @ref DOFObjectIDSource from the specified @ref DOFObjectID.

   @mustfree{DOFObjectIDSource_Destroy}

   @param id The @ref DOFObjectID from which to create the source ID. This should not be NULL, should not have any
             attributes, should not be Broadcast, and should contain data of 20 bytes or less.

   @return The new @ref DOFObjectIDSource.

   @exception NULL is returned if an invalid parameter is detected or the source ID could not be created.
**/
extern DOF_OAL_DECLSPEC DOFObjectIDSource DOFObjectIDSource_Create( const DOFObjectID id );

/**
 *  Create a new copy of an existing @ref DOFObjectIDSource.
 *  @mustfree{DOFObjectIDSource_Destroy}
 *  @param  self The @ref DOFObjectIDSource to be copied.  This should not be NULL.
 *  @return A new copy of the @ref DOFObjectIDSource.
 *  @exception  NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DOFObjectIDSource DOFObjectIDSource_Copy( const DOFObjectIDSource self );

/**
   Get the @ref DOFObjectID from the specified @ref DOFObjectIDSource.

   @param self The @ref DOFObjectIDSource.  This must not be NULL.

   @return The @ref DOFObjectID.
**/
extern DOF_OAL_DECLSPEC DOFObjectID DOFObjectIDSource_GetObjectID( const DOFObjectIDSource self );

/**
   Destroy a @ref DOFObjectIDSource.
   @param  self The @ref DOFObjectIDSource to be destroyed.  If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFObjectIDSource_Destroy( DOFObjectIDSource self );

/**
@}
**/


/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* DOF_OBJECTID_H */
