/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_OPERATION_H
#define DOF_OPERATION_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DOFOperation DOFOperation
    @ingroup DOF_Framework

    Represents an outstanding, asynchronous operation.

    The following model is used for all asynchronous operations (that continue
    in a separate thread while the initiator can do other things):

    <code> DOFOperation &lt;type&gt;_Begin&lt;op&gt;( &lt;parameters&gt;, @ref uint32 timeout, &lt;callback(s)&gt;, void * context )</code>

    The timeout parameter represents the amount of time that the operation is
    permitted to continue.  It is represented in milliseconds.  In general, a
    timeout of @c 0 indicates that the operation is carried out but immediately
    times out, prior to any response or other normal completion.  Conversely, a
    timeout of @ref DOF_TIMEOUT_NEVER indicates that the operation never times
    out.  The timeout can be changed via @ref DOFOperation_SetTimeout any time
    before it has completed.  Once the operation is complete, attempting to
    change the timeout will have no effect.

    The callback parameter represents an object containing a set of function
    pointers (depending on the call) that will be called as the operation
    either completes or reports its status or result data.  The object may also
    contain state or other data associated with the functions, but it should
    not be specific to the operation.

    The context parameter represents data, associated with the specific
    operation, that is passed from the call point to the callback.  This allows
    state specific to the operation to be available inside the callbacks.
    Thus, the same callback may be used for multiple, different operations, and
    the callback can determine the appropriate behavior based on the
    operation's context.

    The DOFOperation represents the operation itself, and is the application's
    handle to control the operation's lifecycle and access the context.  In
    addition to changing their timeout (as already indicated), Operations can
    also be cancelled via @ref DOFOperation_Cancel, in which case the operation
    is immediately marked as complete.  If an Operation is not yet complete
    when it is passed to @ref DOFOperation_Destroy, it will first be cancelled,
    prior to being deallocated.

    Most Operations complete with the response of requested data.  Some
    operations, such as a Provide or Query operation, only complete due to a
    timeout or a cancellation.  Multicast operations using the Wildcard Object
    ID may solicit multiple responses, and thus will persist until they time
    out or are cancelled.

    In all circumstances, the Operation must eventually be complete, triggering
    the invocation of the complete callback.  In some cases, the callback may
    extend the operation's timeout (if the operation only timed out), thereby
    causing the operation to continue and the complete callback to be called
    once again (after the new timeout).  If the complete callback returns
    without extending the operation's timeout, then the Operation is flagged
    as complete (@ref DOFOperation_IsComplete returns @c TRUE).

    It is up to the application to eventually free the resources associated
    with the Operation.  It is generally not recommended to destroy the
    DOFOperation from within the callback, as this can cause memory corruption
    if the operation is still being used by the application (e.g., in another
    thread).  However, if the application "passes" ownership to the callback to
    manage the lifecycle, then the operation can safely be destroyed inside the
    callback.

    @note @ref DOFOperation_WaitComplete must @em not be called from within a
    callback for the same Operation.  Doing so results in undefined behavior.

@{
**/

/*
**  "Public" Type and Method Declarations
*/

/* DOFOperation Methods */

/**
    Cancel an Operation.

    This has no effect on an Operation that has already been cancelled.  The
    timeout of a cancelled Operation cannot be extended.

    It is safe to use this call inside the operation's callback(s).

    @param  op  The Operation to cancel.  If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFOperation_Cancel( DOFOperation op );

/**
    Destroy an Operation, freeing the Operation's resources.

    If the operation is not yet complete, this also cancels the operation,
    causing the callback to be notified of the cancellation.  Afterwards, the
    operation's resources are released.

    It is safe to use this call inside the operation's callback(s), provided
    @b all use of the operation is contained within the callback(s).  In this
    case, use of the operation outside of the callback(s) cannot be guaranteed
    (e.g., @ref DOFOperation_WaitComplete), as the callback may cause it to be
    destroyed (and thus become invalid) at any point.

    @param  op  The Operation to destroy.  If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFOperation_Destroy( DOFOperation op );

/**
    Return the application's context information held within the Operation.

    @param  op  The Operation.  This must not be NULL.

    @return The application's context.
**/
extern DOF_OAL_DECLSPEC void * DOFOperation_GetContext( const DOFOperation op );

/**
    Report the Operation's remaining timeout, in milliseconds.

    If the Operation has timed out, the time will be @c 0.

    @param  op  The Operation.  This must not be NULL.

    @return The remaining time left for the Operation, in milliseconds.
**/
extern DOF_OAL_DECLSPEC uint32 DOFOperation_GetTimeRemaining( const DOFOperation op );

/**
    Set or change an Operation's timeout.

    An Operation's timeout can be adjusted (extended or shortened) provided it
    has not been cancelled or fully completed (where no more results are
    possible). It is possible to extend the timeout of an Operation that has
    only timed out, even during the operation's completion callback.  If this
    is done, the operation state is immediately changed back to "not complete",
    and the callback will be called later to indicate completion.  If the
    timeout of a timed out operation isn't changed in the callback, the
    operation cannot be extended.

    It is safe to use this call inside the operation's callback(s).

    @param  op   The Operation.  This must not be NULL.
    @param  timeout The new timeout for the Operation, in milliseconds.

    @return TRUE if the timeout was set, FALSE otherwise.
**/
extern DOF_OAL_DECLSPEC boolean DOFOperation_SetTimeout( DOFOperation op, uint32 timeout );

/**
    Report whether or not the Operation is cancelled.

    @param  op  The Operation.  This must not be NULL.

    @retval TRUE if the operation is cancelled.
    @retval FALSE if the operation is not cancelled.
**/
extern DOF_OAL_DECLSPEC boolean DOFOperation_IsCancelled( const DOFOperation op );

/**
    Report whether or not the Operation is complete.

    An Operation is considered complete if it has either successfully completed
    or has timed out.  An Operation that has timed out can be extended with
    @ref DOFOperation_SetTimeout, in which case it is no longer considered
    complete.

    @note If this call returns TRUE, it is guaranteed that all callbacks
    associated with the operation have been called and have returned.

    @param  op  The Operation.  This must not be NULL.

    @retval TRUE if the operation is complete.
    @retval FALSE if the operation is not yet complete.

    @see DOFOperation_WaitComplete
**/
extern DOF_OAL_DECLSPEC boolean DOFOperation_IsComplete( const DOFOperation op );

/**
    Report whether or not the Operation is acknowledged.

    An Operation is considered acknowledged when a response (either
    positive or negative) has been received.  BeginQuery, BeginInterest, and
    BeginProvide Operations do not receive a response, and therefore are never acknowledged.

    @param  op  The Operation.  This must not be NULL.

    @retval TRUE if the operation is acknowledged.
    @retval FALSE if the operation is not yet acknowledged.

    @see DOFOperation_WaitAcknowledged
**/
extern DOF_OAL_DECLSPEC boolean DOFOperation_IsAcknowledged( const DOFOperation op );

/**
    Wait for the Operation to complete.

    This blocks the caller until the operation completes (including by timeout or
    cancel), or the timeout of this wait expires, whichever occurs first.  During
    this wait, information continues to be passed to the callbacks as necessary.
    If the Operation completes or times out during this call, the callback is
    called for the last time prior to this call returning, and the return value
    will be TRUE.  If this call's timeout expires first, then the return value
    will be FALSE.

    The time specified in this call does not affect the Operation itself; it
    only specifies the maximum time this call will pend waiting for the
    operation to be complete.  Avoid using the same timeout
    as the operation, since this can introduce a race condition (the operation
    may or may not actually complete before this wait does).  Instead, use @c
    DOF_TIMEOUT_NEVER.  The maximum time this call can wait is the minimum of
    the operation's time remaining and @a timeout.

    It is @b not safe to use this call if the operation's complete callback can
    result in a call to @ref DOFOperation_Destroy (refer to that function's
    documentation for more information).

    @note If this call returns TRUE, it is guaranteed that all callbacks
    associated with the operation have been called and have returned.

    @param  op      The Operation.  This must not be NULL.
    @param  timeout The maximum time to wait, in milliseconds.
    @exceptionparam

    @retval TRUE if the operation has completed.  If @a pException is not NULL,
            it will contain the same exception (if any) given to the operation's
            Complete callback.
    @retval FALSE if the operation has not completed within the specified time,
            or an error occurred.  If @a pException is not NULL, it will be set
            as follows:
        - A DOFErrorException containing @ref DOFERROR_TIMEOUT if the operation is still in progress (the wait timed out).
        - A DOFErrorException containing @ref DOFERROR_INSUFFICIENT_RESOURCES if internal resources could not be created to process this request.

    @see DOFOperation_IsComplete
**/
extern DOF_OAL_DECLSPEC boolean DOFOperation_WaitComplete( const DOFOperation op, uint32 timeout, DOFException * pException );

/**
    Wait for the Operation to be acknowledged.

    An Operation is considered acknowledged when a response (either
    positive or negative) has been received.  BeginQuery, BeginInterest, and
    BeginProvide Operations are do not receive a response, and therefore
    are never acknowledged.

    This blocks the caller until the operation is acknowledged, the operation
    completes (including by timeout or cancel), or the timeout of this wait
    expires, whichever occurs first.  During this wait, information continues
    to be passed to the callbacks as necessary.

    The time specified in this call does not affect the Operation itself; it
    only specifies the maximum time this call will pend waiting for the
    operation to be acknowledged or complete.  Avoid using the same timeout
    as the operation, since the operation may or may not actually complete
    before this wait does, causing a nondeterministic result.  Instead, use @c
    DOF_TIMEOUT_NEVER.  The maximum time this call can wait is the minimum of
    the operation's time remaining and @a timeout.

    It is @b not safe to use this call if the operation's complete callback can
    result in a call to @ref DOFOperation_Destroy (refer to that function's
    documentation for more information).

    @param  op      The Operation. Must not be NULL.
    @param  timeout The maximum time to wait, in milliseconds.
    @exceptionparam

    @retval TRUE if the operation is acknowledged.  If @a
            pException is not NULL, it will contain the same exception (if any)
            given to the operation's callback indicating acknowledgement.
    @retval FALSE if the operation has not been acknowledged within the
            specified time, the operation completed before being acknowledged
            (e.g., timed out), or an error occurred.  If @a pException is not
            NULL, it will be set as follows:
        - @c NULL if the operation was cancelled before being acknowledged.
        - A DOFErrorException containing @ref DOFERROR_TIMEOUT if the operation is still in progress (the wait timed out), or the operation timed out and was never acknowledged.
        - The exception reported to the Complete callback, if the operation is completed abnormally.
        - A DOFErrorException containing @ref DOFERROR_INSUFFICIENT_RESOURCES if internal resources could not be created to process this request.

    @see DOFOperation_IsAcknowledged
**/
extern DOF_OAL_DECLSPEC boolean DOFOperation_WaitAcknowledged( const DOFOperation op, uint32 timeout, DOFException * pException );


/**
    Get the exception associated with a completed operation, if any.
    This reports the same exception that is normally delivered to the operation's Complete callback, if any.
    If the operation is not yet complete, this will return NULL.

    @note The exception returned by this call is owned by the DOFOperation and is not valid beyond the
    lifetime of its DOFOperation.  If the exception is needed for longer, use @ref DOFException_Copy
    to obtain a separate copy of the exception.

    @param  op      The Operation. Must not be NULL.

    @return The exception associated with the operation, if any.  NULL is returned if there is no exception.

    @see DOFOperation_IsComplete
**/
extern DOF_OAL_DECLSPEC DOFException DOFOperation_GetException( const DOFOperation op );


/** @name DOFOperation.Get
 *  These calls are only for Get Operations (resulting from a @ref DOFObject_BeginGet call).
 *  @{
 */

/**
    Return the @ref DOFObject that is associated with the Get Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Get Operation.

    @return The associated @ref DOFObject.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFObject DOFOperationGet_GetObject( const DOFOperation op );

/**
    Return the @ref DOFInterfaceProperty that is associated with the Get Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Get Operation.

    @return The associated @ref DOFInterfaceProperty.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFInterfaceProperty DOFOperationGet_GetProperty( const DOFOperation op );

/**
    Return the @ref DOFOperationControl that is associated with the Get Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Get Operation.

    @return The associated @ref DOFOperationControl.  This will not be NULL.  If the Operation was started with a NULL
            control, the default control for the operation is returned here.
**/
extern DOF_OAL_DECLSPEC DOFOperationControl DOFOperationGet_GetControl( const DOFOperation op );

/**
    Wait for and return the result of the Get Operation.  If the operation already has received a result, this returns immediately with the result.

    @mustfree{DOFResult_Destroy}

    @note For broadcast or multicast operations, this returns the first result only.  Multiple results are returned only via the operation's callback.

    This blocks the caller until the operation has a result or exception (or completes
    normally), the operation times out or is cancelled, or the timeout of this
    wait expires, whichever occurs first.

    The time specified in this call has no relationship to the Operation
    itself; it only specifies how long this call will block the caller.

    It is @b not safe to use this call if the operation's complete callback can
    result in a call to @ref DOFOperation_Destroy (refer to that function's
    documentation for more information).

    @param  op  The Operation.  This must not be NULL and must represent a valid Get Operation.
    @param  timeout  The maximum time to wait for the result, in  milliseconds.  A value of 0 returns immediately whether or not a result is available.
    @exceptionparam

    @return The @ref DOFResult of the Operation, or NULL if a result cannot be returned.  There are two cases in which a result cannot be returned:
              - If the operation's result was an exception, the exception is reported via @a pException, which will be non-NULL.
              - If the operation does not have a result within the specified @a timeout, or completes without a result within the specified @a timeout, @a pException will represent a @ref DOFERROR_TIMEOUT.
**/
extern DOF_OAL_DECLSPEC DOFResult DOFOperationGet_WaitResult( const DOFOperation op, uint32 timeout, DOFException * pException );

/**
    Return the result of the Get Operation, if any.
    This does not block the caller.

    @note For broadcast or multicast operations, this returns the first result only.  Multiple results are returned only via the operation's callback.

    @param  op  The Operation.  This must not be NULL and must represent a valid Get Operation.
    @exceptionparam

    @return The @ref DOFResult of the Operation, or NULL if a result cannot be returned.  There are two cases in which a result cannot be returned:
              - If the operation's result was an exception, the exception is reported via @a pException, which will be non-NULL.
              - If the operation does not yet have a result, @a pException will represent a @ref DOFERROR_FAILED.
**/
extern DOF_OAL_DECLSPEC DOFResult DOFOperationGet_GetResult( const DOFOperation op, DOFException * pException );

/** @} **/


/** @name DOFOperation.Set
 *  These calls are only for Set Operations (resulting from a @ref DOFObject_BeginSet call).
 *  @{
 */

/**
    Return the @ref DOFObject that is associated with the Set Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Set Operation.

    @return The associated @ref DOFObject.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFObject DOFOperationSet_GetObject( const DOFOperation op );

/**
    Return the @ref DOFInterfaceProperty that is associated with the Set Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Set Operation.

    @return The associated @ref DOFInterfaceProperty.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFInterfaceProperty DOFOperationSet_GetProperty( const DOFOperation op );

/**
    Return the @ref DOFOperationControl that is associated with the Set Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Set Operation.

    @return The associated @ref DOFOperationControl.  This will not be NULL.  If the Operation was started with a NULL
            control, the default control for the operation is returned here.
**/
extern DOF_OAL_DECLSPEC DOFOperationControl DOFOperationSet_GetControl( const DOFOperation op );

/**
    Wait for and return the result of the Set Operation.  If the operation already has received a result, this returns immediately with the result.

    @mustfree{DOFResult_Destroy}

    @note For broadcast or multicast operations, this returns the first result only.  Multiple results are returned only via the operation's callback.

    This blocks the caller until the operation has a result or exception (or completes
    normally), the operation times out or is cancelled, or the timeout of this
    wait expires, whichever occurs first.

    The time specified in this call has no relationship to the Operation
    itself; it only specifies how long this call will block the caller.

    It is @b not safe to use this call if the operation's complete callback can
    result in a call to @ref DOFOperation_Destroy (refer to that function's
    documentation for more information).

    @param  op  The Operation.  This must not be NULL and must represent a valid Set Operation.
    @param  timeout  The maximum time to wait for the result, in  milliseconds.  A value of 0 returns immediately whether or not a result is available.
    @exceptionparam

    @return The @ref DOFResult of the Operation, or NULL if a result cannot be returned.  There are two cases in which a result cannot be returned:
              - If the operation's result was an exception, the exception is reported via @a pException, which will be non-NULL.
              - If the operation does not have a result within the specified @a timeout, or completes without a result within the specified @a timeout, @a pException will represent a @ref DOFERROR_TIMEOUT.
**/
extern DOF_OAL_DECLSPEC DOFResult DOFOperationSet_WaitResult( const DOFOperation op, uint32 timeout, DOFException * pException );

/**
    Return the result of the Set Operation, if any.
    This does not block the caller.

    @note For broadcast or multicast operations, this returns the first result only.  Multiple results are returned only via the operation's callback.

    @param  op  The Operation.  This must not be NULL and must represent a valid Set Operation.
    @exceptionparam

    @return The @ref DOFResult of the Operation, or NULL if a result cannot be returned.  There are two cases in which a result cannot be returned:
              - If the operation's result was an exception, the exception is reported via @a pException, which will be non-NULL.
              - If the operation does not yet have a result, @a pException will represent a @ref DOFERROR_FAILED.
**/
extern DOF_OAL_DECLSPEC DOFResult DOFOperationSet_GetResult( const DOFOperation op, DOFException * pException );

/** @} **/


/** @name DOFOperation.Invoke
 *  These calls are only for Invoke Operations (resulting from a @ref DOFObject_BeginInvoke call).
 *  @{
 */

/**
    Return the @ref DOFObject that is associated with the Invoke Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Invoke Operation.

    @return The associated @ref DOFObject.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFObject DOFOperationInvoke_GetObject( const DOFOperation op );

/**
    Return the @ref DOFInterfaceMethod that is associated with the Invoke Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Invoke Operation.

    @return The associated @ref DOFInterfaceMethod.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFInterfaceMethod DOFOperationInvoke_GetMethod( const DOFOperation op );

/**
    Return the @ref DOFOperationControl that is associated with the Invoke Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Invoke Operation.

    @return The associated @ref DOFOperationControl.  This will not be NULL.  If the Operation was started with a NULL
            control, the default control for the operation is returned here.
**/
extern DOF_OAL_DECLSPEC DOFOperationControl DOFOperationInvoke_GetControl( const DOFOperation op );

/**
    Wait for and return the result of the Invoke Operation.  If the operation already has received a result, this returns immediately with the result.

    @mustfree{DOFResult_Destroy}

    @note For broadcast or multicast operations, this returns the first result only.  Multiple results are returned only via the operation's callback.

    This blocks the caller until the operation has a result or exception (or completes
    normally), the operation times out or is cancelled, or the timeout of this
    wait expires, whichever occurs first.

    The time specified in this call has no relationship to the Operation
    itself; it only specifies how long this call will block the caller.

    It is @b not safe to use this call if the operation's complete callback can
    result in a call to @ref DOFOperation_Destroy (refer to that function's
    documentation for more information).

    @param  op  The Operation.  This must not be NULL and must represent a valid Invoke Operation.
    @param  timeout  The maximum time to wait for the result, in  milliseconds.  A value of 0 returns immediately whether or not a result is available.
    @exceptionparam

    @return The @ref DOFResult of the Operation, or NULL if a result cannot be returned.  There are two cases in which a result cannot be returned:
              - If the operation's result was an exception, the exception is reported via @a pException, which will be non-NULL.
              - If the operation does not have a result within the specified @a timeout, or completes without a result within the specified @a timeout, @a pException will represent a @ref DOFERROR_TIMEOUT.
**/
extern DOF_OAL_DECLSPEC DOFResult DOFOperationInvoke_WaitResult( const DOFOperation op, uint32 timeout, DOFException * pException );

/**
    Return the result of the Invoke Operation, if any.
    This does not block the caller.

    @note For broadcast or multicast operations, this returns the first result only.  Multiple results are returned only via the operation's callback.

    @param  op  The Operation.  This must not be NULL and must represent a valid Invoke Operation.
    @exceptionparam

    @return The @ref DOFResult of the Operation, or NULL if a result cannot be returned.  There are two cases in which a result cannot be returned:
              - If the operation's result was an exception, the exception is reported via @a pException, which will be non-NULL.
              - If the operation does not yet have a result, @a pException will represent a @ref DOFERROR_FAILED.
**/
extern DOF_OAL_DECLSPEC DOFResult DOFOperationInvoke_GetResult( const DOFOperation op, DOFException * pException );

/** @} **/


/** @name DOFOperation.Subscribe
 *  These calls are only for Subscribe Operations (resulting from a @ref DOFObject_BeginSubscribe call).
 *  @{
 */

/**
    Return the @ref DOFObject that is associated with the Subscribe Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Subscribe Operation.

    @return The associated @ref DOFObject.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFObject DOFOperationSubscribe_GetObject( const DOFOperation op );

/**
    Return the @ref DOFInterfaceProperty that is associated with the Subscribe Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Subscribe Operation.

    @return The associated @ref DOFInterfaceProperty.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFInterfaceProperty DOFOperationSubscribe_GetProperty( const DOFOperation op );

/**
    Return the @ref DOFOperationControl that is associated with the Subscribe Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Subscribe Operation.

    @return The associated @ref DOFOperationControl.  This will not be NULL.  If the Operation was started with a NULL
            control, the default control for the operation is returned here.
**/
extern DOF_OAL_DECLSPEC DOFOperationControl DOFOperationSubscribe_GetControl( const DOFOperation op );

/**
    Return the minimum period that was specified with the Subscribe Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Subscribe Operation.

    @return The associated minimum period.
**/
extern DOF_OAL_DECLSPEC uint32 DOFOperationSubscribe_GetMinPeriod( const DOFOperation op );

/** @} **/


/** @name DOFOperation.Register
 *  These calls are only for Register Operations (resulting from a @ref DOFObject_BeginRegister call).
 *  @{
 */

/**
    Return the @ref DOFObject that is associated with the Register Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Register Operation.

    @return The associated @ref DOFObject.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFObject DOFOperationRegister_GetObject( const DOFOperation op );


/**
    Return the @ref DOFInterfaceEvent that is associated with the Register Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Register Operation.

    @return The associated @ref DOFInterfaceEvent.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFInterfaceEvent DOFOperationRegister_GetEvent( const DOFOperation op );

/**
    Return the @ref DOFOperationControl that is associated with the Register Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Register Operation.

    @return The associated @ref DOFOperationControl.  This will not be NULL.  If the Operation was started with a NULL
            control, the default control for the operation is returned here.
**/
extern DOF_OAL_DECLSPEC DOFOperationControl DOFOperationRegister_GetControl( const DOFOperation op );

/** @} **/


/** @name DOFOperation.Session
 *  These calls are only for Session Operations (resulting from a @ref DOFObject_BeginSession call).
 *  @{
 */

/**
    Wait for a Session Operation to be accepted, and return a DOFObject for the
    session. If the operation has already been accepted or is completed, this
    returns immediately as defined below.

    @mustfree{DOFObject_Destroy}

    This blocks the caller until the operation is accepted or an exception is
    returned, the operation times out or is cancelled, or the timeout of this
    wait expires, whichever occurs first.

    The time specified in this call has no relationship to the Operation
    itself; it only specifies how long this call will block the caller.

    It is @b not safe to use this call if the operation's complete callback can
    result in a call to @ref DOFOperation_Destroy (refer to that function's
    documentation for more information).

    @param  op  The Operation.  This must not be NULL and must represent a valid Session Operation.
    @param  timeout  The maximum time to wait for a result, in  milliseconds. A
                     value of 0 returns immediately.
    @exceptionparam

    @return The @ref DOFObject, representing the new Session created by the Session Operation, or NULL if a session object cannot be returned.  There are two cases in which a session object cannot be returned:
              - If the operation failed with an exception, the exception is
                reported via @a pException, which will be non-NULL.
              - If the operation has not been accepted within the specified
                @a timeout, or completes without a session object within the
                specified @a timeout, @a pException will represent a
                @ref DOFERROR_TIMEOUT.
**/
extern DOF_OAL_DECLSPEC DOFObject DOFOperationSession_WaitSession( const DOFOperation op, uint32 timeout, DOFException * pException );

/**
    Return the @ref DOFObject that is associated with the Session Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Session Operation.

    @return The associated @ref DOFObject.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFObject DOFOperationSession_GetObject( const DOFOperation op );


/**
    Return the @ref DOFInterface that is associated with the Session Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Session Operation.

    @return The associated @ref DOFInterface.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFInterface DOFOperationSession_GetInterface( const DOFOperation op );

/**
    Return the Session's type (a @ref DOFInterfaceID) that is associated with the Session Operation.
    The type is always provided on the session, and indicates the
    purpose of and contract for the session.

    @param  op  The Operation.  This must not be NULL and must represent a valid Session Operation.

    @return The associated session primary @ref DOFInterfaceID.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFInterfaceID DOFOperationSession_GetSessionType( const DOFOperation op );

/**
    Return the @ref DOFOperationControl that is associated with the Session Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Session Operation.

    @return The associated @ref DOFOperationControl.  This will not be NULL.  If the Operation was started with a NULL
            control, the default control for the operation is returned here.
**/
extern DOF_OAL_DECLSPEC DOFOperationControl DOFOperationSession_GetControl( const DOFOperation op );

/**
 *  Set the DOFOperationSessionDataTransform to use on this session.
 *
 *  The data transform can be called at most once per session (subsequent calls will have no effect).
 *
 *  @param sessionOperation The DOFOperation for the session.  Should not be @c NULL.
 *  @param transform    The transform for the session.  A value of @c NULL will have no effect.
 *
 *  @retval TRUE    @a transform was set successfully
 *  @retval FALSE   there was a failure to set @a transform
 */
extern DOF_OAL_DECLSPEC boolean DOFOperationSession_SetDataTransform(DOFOperation sessionOperation, DOFOperationSessionDataTransform transform);

/** @} **/


/** @name DOFOperation.Provide
 *  These calls are only for Provide Operations (resulting from a @ref DOFObject_BeginProvide call).
 *  @{
 */

/**
    Return the @ref DOFObject that is associated with the Provide Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Provide Operation.

    @return The associated @ref DOFObject.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFObject DOFOperationProvide_GetObject( const DOFOperation op );

/**
    Return the @ref DOFInterface that is associated with the Provide Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Provide Operation.

    @return The associated @ref DOFInterface.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFInterface DOFOperationProvide_GetInterface( const DOFOperation op );

/** @} **/


/** @name DOFOperation.Interest
 *  These calls are only for Interest Operations (resulting from a @ref DOFSystem_BeginInterest call).
 *  @{
 */

/**
    Return the @ref DOFSystem that is associated with the Interest Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Interest Operation.

    @return The associated @ref DOFSystem.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFSystem DOFOperationInterest_GetSystem( const DOFOperation op );

/**
    Return the @ref DOFObjectID that is associated with the Interest Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Interest Operation.

    @return The associated @ref DOFObjectID.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFObjectID DOFOperationInterest_GetObjectID( const DOFOperation op );

/**
    Return the @ref DOFInterfaceID that is associated with the Interest Operation.

    @mustfree{DOFInterfaceID_Destroy}

    @param  op  The Operation.  This must not be NULL and must represent a valid Interest Operation.

    @return The associated @ref DOFInterfaceID.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFInterfaceID DOFOperationInterest_GetInterfaceID( DOFOperation op );

/**
    Return the @ref DOFInterestLevel that is associated with the Interest Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Interest Operation.

    @return The associated @ref DOFInterestLevel.
**/
extern DOF_OAL_DECLSPEC DOFInterestLevel DOFOperationInterest_GetInterestLevel( const DOFOperation op );

/**
    Return the @ref DOFOperationControl that is associated with the Interest Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Interest Operation.

    @return The associated @ref DOFOperationControl.  This will not be NULL.  If the Operation was started with a NULL
            control, the default control for the operation is returned here.
**/
extern DOF_OAL_DECLSPEC DOFOperationControl DOFOperationInterest_GetControl( const DOFOperation op );

/** @} **/


/** @name DOFOperation.Define
 *  These calls are only for Define Operations (resulting from a @ref DOFObject_BeginDefine call).
 *  @{
 */

/**
    Wait for and return the interface definition of the Define Operation.  If
    the operation already has received a definition or an exception, this
    returns immediately with the same.

    @mustfree{DOFInterface_Destroy}

    This blocks the caller until the operation has an interface definition or
    exception, the operation times out or is cancelled, or the timeout of this
    wait expires, whichever occurs first.

    The time specified in this call has no relationship to the Operation
    itself; it only specifies how long this call will block the caller.

    It is @b not safe to use this call if the operation's complete callback can
    result in a call to @ref DOFOperation_Destroy (refer to that function's
    documentation for more information).

    @param  op  The Operation.  This must not be NULL and must represent a valid Define Operation.
    @param  timeout  The maximum time to wait for a result, in  milliseconds. A
                     value of 0 returns immediately.
    @exceptionparam

    @return The @ref DOFInterface retreived by the Define Operation, or NULL if an interface cannot be returned.  There are two cases in which an interface cannot be returned:
              - If the operation resulted in an exception, the exception is
                reported via @a pException, which will be non-NULL.
              - If the operation does not receive an interface within the
                specified @a timeout, or the operation completes and does not
                have an interface within the specified @a timeout, @a pException
                will represent a @a DOFERROR_TIMEOUT.
**/
extern DOF_OAL_DECLSPEC DOFInterface DOFOperationDefine_WaitInterface( const DOFOperation op, uint32 timeout, DOFException * pException );

/**
    Return the @ref DOFObject that is associated with the Define Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Define Operation.

    @return The associated @ref DOFObject.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFObject DOFOperationDefine_GetObject( const DOFOperation op );

/**
    Return the @ref DOFInterfaceID that is associated with the Define Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Define Operation.

    @return The associated @ref DOFInterfaceID.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFInterfaceID DOFOperationDefine_GetInterfaceID( const DOFOperation op );

/**
    Return the @ref DOFOperationControl that is associated with the Define Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Define Operation.

    @return The associated @ref DOFOperationControl.  This will not be NULL.  If the Operation was started with a NULL
            control, the default control for the operation is returned here.
**/
extern DOF_OAL_DECLSPEC DOFOperationControl DOFOperationDefine_GetControl( const DOFOperation op );

/** @} **/


/** @name DOFOperation.Query
 *  These calls are only for Query Operations (resulting from a @ref DOFSystem_BeginQuery call).
 *  @{
 */

/**
    Return the @ref DOFQuery that is associated with a Query Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Query Operation.

    @return The associated @ref DOFQuery.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFQuery DOFOperationQuery_GetQuery( const DOFOperation op );

/**
    Return the @ref DOFSystem that is associated with the Query Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Query Operation.

    @return The associated @ref DOFSystem.  This will not be NULL.
**/
extern DOF_OAL_DECLSPEC DOFSystem DOFOperationQuery_GetSystem( const DOFOperation op );

/** @} **/


/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* DOF_OPERATION_H */
