/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_PERMISSION_H
#define DOF_PERMISSION_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif
/** @defgroup DOFPermission DOFPermission
    @ingroup DOF_Framework

@{
**/

/** Tunnel domain identifier that represents all domains.  This is typically used with the Tunnel Domain permission. **/
#define DOFPERMISSIONTUNNELDOMAIN_ALL 0x3FFFFFFF

/**
   The Action flags for a Binding Permission.

   These constants can be ORed together to express the actions for a binding permission.
**/
typedef enum DOFPermissionBindingAction_t
{
    DOFPERMISSIONBINDINGACTION_EXECUTE =    0x01, /**< Permission to Invoke a Method. **/
    DOFPERMISSIONBINDINGACTION_WRITE =      0x02, /**< Permission to Set a Property. **/
    DOFPERMISSIONBINDINGACTION_READ =       0x04, /**< Permission to Get or Subscribe to a Property, or Register for an Event. **/
    DOFPERMISSIONBINDINGACTION_PROVIDE =    0x08, /**< Permission to Provide an Interface. **/
    DOFPERMISSIONBINDINGACTION_SESSION =    0x10, /**< Permission to Open a Session. **/
    DOFPERMISSIONBINDINGACTION_ALL =        0x1F  /**< All permissions. Equivalent to ( INVOKE | WRITE | READ | PROVIDE | SESSION). **/
} DOFPermissionBindingAction;

/**
   The type of the permission.

   These constants identify the type of permission.
**/
typedef enum DOFPermissionType_t
{
    DOFPERMISSIONTYPE_BINDING =       0x0001,
    DOFPERMISSIONTYPE_I_AM =          0x0003,
    DOFPERMISSIONTYPE_ACT_AS =        0x0005,
    DOFPERMISSIONTYPE_REQUESTOR =     0x0080,
    DOFPERMISSIONTYPE_PROVIDER =      0x0082,
    DOFPERMISSIONTYPE_DEFINE =        0x0083,
    DOFPERMISSIONTYPE_TUNNEL_DOMAIN = 0x0085
} DOFPermissionType;

/**
 * Create a simple Binding type Permission from a single Object and Interface.
 *
 * @param actions The actions requested for the permission.
 * @param objectID The exact Object ID this permission is for, including any attributes.  This may be NULL to indicate
 *                 @b all object IDs (wildcard, including all attributes). This may be or include the Broadcast ID to
 *                 indicate specifically (exactly) a broadcast or multicast ID (e.g., for broadcast or multicast
 *                 operations).
 * @param interfaceID The Interface ID this permission is for.  This may be NULL or the Wildcard ID to indicate @b all interface IDs (wildcard).
 *
 * @return A new permission, as indicated.
 * @exception NULL is returned if the permission cannot be created.
 */
extern DOF_OAL_DECLSPEC DOFPermission DOFPermissionBinding_Create( DOFPermissionBindingAction actions, const DOFObjectID objectID, const DOFInterfaceID interfaceID );

/**
 * Create an Requestor type Permission.
 *
 * @mustfree{DOFPermission_Destroy}
 *
 * @return A new permission, as indicated.
 * @exception NULL is returned if the permission cannot be created.
 */
extern DOF_OAL_DECLSPEC DOFPermission DOFPermissionRequestor_Create( void );

/**
 * Create a Provider type Permission.
 *
 * @mustfree{DOFPermission_Destroy}
 *
 * @return A new permission, as indicated.
 * @exception NULL is returned if the permission cannot be created.
 */
extern DOF_OAL_DECLSPEC DOFPermission DOFPermissionProvider_Create( void );

/**
 * Create a Define type Permission.
 *
 * @mustfree{DOFPermission_Destroy}
 *
 * @param interfaceID The interface to define.  The Wildcard IID indicates permission to define all interfaces.  This should not be NULL.
 *
 * @return A new permission, as indicated.
 * @exception NULL is returned if the permission cannot be created.
 */
extern DOF_OAL_DECLSPEC DOFPermission DOFPermissionDefine_Create( const DOFInterfaceID interfaceID );

/**
 * Create a I Am type Permission.
 * This permission, if granted, allows the node to source operations with the given identifier.
 *
 * @mustfree{DOFPermission_Destroy}
 *
 * @param sourceID The source identifier.  This should not be NULL.
 *
 * @return A new permission, as indicated.
 * @exception NULL is returned if the permission cannot be created.
 */
extern DOF_OAL_DECLSPEC DOFPermission DOFPermissionIAm_Create( DOFObjectIDSource sourceID );

/**
 * Create a Act As type Permission.
 * This permission, if granted, allows the node to proxy as another.
 *
 * @mustfree{DOFPermission_Destroy}
 *
 * @param sourceID The source identifier.  This should not be NULL.  Source identifiers cannot use the Broadcast ID.
 *
 * @return A new permission, as indicated.
 * @exception NULL is returned if the permission cannot be created.
 * @see DOFPermissionActAsAny_Create
 */
extern DOF_OAL_DECLSPEC DOFPermission DOFPermissionActAs_Create( DOFObjectIDSource sourceID );

/**
 * Create a Act As Any type Permission.
 * This permission, if granted, allows the node to proxy as @b any another node (useful for routing nodes).
 *
 * @mustfree{DOFPermission_Destroy}
 *
 * @return A new permission, as indicated.
 * @exception NULL is returned if the permission cannot be created.
 * @see DOFPermissionActAs_Create
 */
extern DOF_OAL_DECLSPEC DOFPermission DOFPermissionActAsAny_Create( void );

/**
 * Create a Tunnel Domain type Permission.
 * This permission, if granted, allows the node to tunnel traffic for the specified related domain.
 *
 * @mustfree{DOFPermission_Destroy}
 *
 * @param rdid The tunnel domain identifier for the related domain.  This may be the wildcard @ref
 *             DOFPERMISSIONTUNNELDOMAIN_ALL to indicate all domains.
 *
 * @return A new permission, as indicated.
 * @exception NULL is returned if the permission cannot be created.
 */
extern DOF_OAL_DECLSPEC DOFPermission DOFPermissionTunnelDomain_Create( uint32 rdid );

/**
 * Destroys a permission created from one of the create functions.
 *
 * @param self the permission to destroy.  If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFPermission_Destroy( DOFPermission self );

/**
 * Report the number of bytes required in a buffer to marshal a @ref DOFPermission.
 *
 * @param self The @ref DOFPermission. This must not be NULL.
 *
 * @return The marshaled size of the @ref DOFPermission.
 */
extern DOF_OAL_DECLSPEC uint32 DOFPermission_GetMarshalSize(DOFPermission self);

/**
 * Marshal a @ref DOFPermission into an existing buffer.
 *
 * @param self The DOFPermission to be marshaled. This must not be NULL.
 * @param size The size of the buffer in which to marshal the DOFPermission.
 * @param buffer The buffer in which to marshal the DOFPermission. This must not be NULL, and must contain at least
 *               @a size bytes.
 *
 * @return The marshaled size of the @ref DOFPermission.
 *
 * @exception 0 will be returned if the marshaled data does not fit in the buffer, or an invalid parameter is detected.
 */
extern DOF_OAL_DECLSPEC uint32 DOFPermission_Marshal(DOFPermission self, uint32 size, uint8 * buffer);

/**
 * Unmarshals a @ref DOFPermission from a buffer.
 *
 * @param size The size of the buffer, in bytes, from which to unmarshal the @ref DOFPermission.
 * @param buffer The buffer from which to unmarshal the @ref DOFPermission. This must not be NULL, and must contain at
 *               least @a size bytes.
 *
 * @return DOFPermission The DOFPermission that is stored in the buffer.
 * @exception NULL is returned if the buffer contents of @a size could not be unmarshalled to a valid @ref
 *            DOFPermission, or there are insufficient resources to allocate the @ref DOFPermission.
 */
extern DOF_OAL_DECLSPEC DOFPermission DOFPermission_Create_Unmarshal(uint32 size, const uint8 * buffer);

/**
 * Returns the type of the given permission.
 *
 * @param self The permission.  This must not be NULL.
 *
 * @return The @ref DOFPermissionType of the specified permission.
 */
extern DOF_OAL_DECLSPEC DOFPermissionType DOFPermission_GetPermissionType( DOFPermission self );

/** @name Binding Permission Builder
@{
**/

/**
 * Creates a Permission builder for binding permission.
 *
 * @mustfree{DOFPermissionBindingBuilder_Destroy or @ref DOFPermissionBindingBuilder_BuildAndDestroy}
 *
 * The builder uses the following defaults:
 *  -Object IDs: none (no specific object means wildcard, or all objects)
 *  -Interface IDs: none (no specific interfaces means wildcard, or all interfaces)
 *  -Attributes: none (no specific attributes means wildcard, or all attributes)
 *  -Actions: The actions passed in must be valid combination of @ref DOFPermissionBindingAction.
 *
 * @param actions The actions you want to be set in this builder.
 *
 * @return A new builder for a binding type permission.
 * @exception NULL is returned if the permission builder cannot be created.
 */
extern DOF_OAL_DECLSPEC DOFPermissionBindingBuilder DOFPermissionBindingBuilder_Create( DOFPermissionBindingAction actions );

/**
 * Add an ObjectID to a binding permission builder.
 *
 * @param self     The binding permission builder.  If NULL, no action is taken.
 * @param objectID The ObjectID to add to the permission.  This may be NULL to indicate @b all object IDs (wildcard).
 *                 This may be the Broadcast ID to indicate specifically the Broadcast ID (e.g., for broadcast
 *                 operations).  The Object ID should not have any attributes.  This call does not change the set of
 *                 acceptable attributes.
 *
 * @return boolean TRUE, if the oid could be added;
 *                 FALSE, if it failed to be added.
 */
extern DOF_OAL_DECLSPEC boolean DOFPermissionBindingBuilder_AddObjectID( DOFPermissionBindingBuilder self, DOFObjectID objectID );

/**
 * Add an InterfaceID to a binding permission builder.
 *
 * @param self The binding permission builder.  If NULL, no action is taken.
 * @param interfaceID The InterfaceID to add to the permission.  This may be NULL or the Wildcard ID to indicate @b all interface IDs (wildcard).
 *
 * @return boolean TRUE, if the interface ID could be added;
 *                 FALSE, if it failed to be added.
 */
extern DOF_OAL_DECLSPEC boolean DOFPermissionBindingBuilder_AddInterfaceID( DOFPermissionBindingBuilder self, DOFInterfaceID interfaceID );

/**
 * Sets the builder to allow or not allow all attributes (any type, any data) to a binding permission builder.
 *
 * If attributes were already added to the builder this call will fail.
 *
 * @param self The binding permission builder.  If NULL, no action is taken.
 * @param allowed TRUE Sets the permission to allow all attributes.
 *                FALSE Set the permission to only allow attributes added.
 *
 * @return boolean TRUE, if the value could be set;
 *                 FALSE, if it failed to be set.
 */
extern DOF_OAL_DECLSPEC boolean DOFPermissionBindingBuilder_SetAllAttributesAllowed( DOFPermissionBindingBuilder self, boolean allowed );

/**
 * Add a wildcard attribute (any data) of a specific type to a binding permission builder.
 *
 * @param self The binding permission builder.  If NULL, no action is taken.
 * @param type The ID of the Attribute.  See @ref DOFObjectIDAttributeType.  Valid numeric range is 0 <= @a type <=
 *             127.
 *
 * @return boolean TRUE, if the attribute could be added;
 *                 FALSE, if it failed to be added.
 */
extern DOF_OAL_DECLSPEC boolean DOFPermissionBindingBuilder_AddRequiredWildcardAttribute( DOFPermissionBindingBuilder self, const DOFObjectIDAttributeType type );

/**
 * Add an attribute to a binding permission builder, based on a raw buffer.
 *
 * @param self The binding permission builder.  If NULL, no action is taken.
 * @param type The ID of the Attribute.  See @ref DOFObjectIDAttributeType.  Valid numeric range is 0 <= @a type <=
 *             127.
 * @param valueSize Size, in byte count, of the supplied Attribute's data.
 * @param value The Attribute's data.  This should not be NULL if @a valueSize > 0.
 *
 * @return boolean TRUE, if the attribute could be added;
 *                 FALSE, if it failed to be added.
 */
extern DOF_OAL_DECLSPEC boolean DOFPermissionBindingBuilder_AddRequiredAttribute_Buffer( DOFPermissionBindingBuilder self, const DOFObjectIDAttributeType type, uint32 valueSize, const uint8 * value );

/**
 * Add an attribute to a binding permission builder, based on a C string.
 *
 * @param self The binding permission builder.  If NULL, no action is taken.
 * @param type The ID of the Attribute.  See @ref DOFObjectIDAttributeType.  Valid numeric range is 0 <= @a type <=
 *             127.
 * @param value The Attribute's data (a C string).  This should not be NULL.
 *
 * @return boolean TRUE, if the attribute could be added;
 *                 FALSE, if it failed to be added.
 */
extern DOF_OAL_DECLSPEC boolean DOFPermissionBindingBuilder_AddRequiredAttribute_String( DOFPermissionBindingBuilder self, const DOFObjectIDAttributeType type, const char * value );

/**
 * Add an attribute to a binding permission builder, based on a @ref DOFObjectID.
 *
 * @param self The binding permission builder.  If NULL, no action is taken.
 * @param type The ID of the Attribute.  See @ref DOFObjectIDAttributeType.  Valid numeric range is 0 <= @a type <=
 *             127.
 * @param value The Attribute's data (a @ref DOFObjectID).  This should not be NULL.
 *
 * @return boolean TRUE, if the attribute could be added;
 *                 FALSE, if it failed to be added.
 */
extern DOF_OAL_DECLSPEC boolean DOFPermissionBindingBuilder_AddRequiredAttribute_ObjectID( DOFPermissionBindingBuilder self, const DOFObjectIDAttributeType type, const DOFObjectID value );

/**
 * Add one or more actions to a binding permission builder.
 *
 * @param self The binding permission builder.  If NULL, no action is taken.
 * @param actions The action to be added.  Must be a valid @ref DOFPermissionBindingAction.
 */
extern DOF_OAL_DECLSPEC void DOFPermissionBindingBuilder_AddActions( DOFPermissionBindingBuilder self, DOFPermissionBindingAction actions );

/**
 * Sets the binding permission actions to the given actions.
 * Please refer to @ref DOFPermissionBindingBuilder_Create for the default value for this setting.
 * @param self The binding permission builder.  If NULL, no action is taken.
 * @param actions the action to set the binding permission builder to. @ref DOFPermissionBindingAction.
 *
 * @retval TRUE Actions were set.
 * @exception FALSE Actions failed to be set.
 */
extern DOF_OAL_DECLSPEC boolean DOFPermissionBindingBuilder_SetActions( DOFPermissionBindingBuilder self, DOFPermissionBindingAction actions );

/**
 * Build a binding permission from the given binding permission builder.
 *
 * @mustfree{DOFPermission_Destroy}
 *
 * @param self The binding permission builder from which to build the permission.  This should not be NULL.
 *
 * @return A new permission, as described by the builder.
 * @exception NULL is returned if the permission cannot be created.
 */
extern DOF_OAL_DECLSPEC DOFPermission DOFPermissionBindingBuilder_Build( DOFPermissionBindingBuilder self );

/**
 * Build a binding permission from the given binding permission builder, and destroy the given builder.
 *
 * @param self The binding permission builder from which to build the permission.  This should not be NULL.
 *
 * @return A new permission, as described by the builder.
 * @exception NULL is returned if the permission cannot be created.
 */
extern DOF_OAL_DECLSPEC DOFPermission DOFPermissionBindingBuilder_BuildAndDestroy( DOFPermissionBindingBuilder self );

/**
 * Destroys a binding permission builder.
 *
 * @param self The binding permission builder to destroy.  If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFPermissionBindingBuilder_Destroy( DOFPermissionBindingBuilder self );

/**
@}
**/

/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* DOF_PERMISSION_H */
