/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_PERMISSIONSET_H
#define DOF_PERMISSIONSET_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DOFPermissionSet DOFPermissionSet
    @ingroup DOF_Framework

@{
**/

/**
 * Copy a given permission set.
 *
 * @param self The permission set to copy. Should not be NULL.
 *
 * @return DOFPermissionSet A copy of the given permission set.
 * @exception NULL if the permission set could not be copied.
 */
extern DOF_OAL_DECLSPEC DOFPermissionSet DOFPermissionSet_Copy( DOFPermissionSet self );

/**
    Destroy this DOFPermissionSet instance.

    @param self The Permission Set.  If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFPermissionSet_Destroy( DOFPermissionSet self );

/** @name Builder
@{
**/

/**
 * Create a Permission Set Builder.
 *
 * @mustfree{DOFPermissionSetBuilder_Destroy or @ref DOFPermissionSetBuilder_BuildAndDestroy}
 *
 * The builder uses the following defaults:
 *  -Permissions: none
 *
 * @return The @ref DOFPermissionSetBuilder.
 *
 * @exception NULL is returned if the builder cannot be created;
 */
extern DOF_OAL_DECLSPEC DOFPermissionSetBuilder DOFPermissionSetBuilder_Create( void );

/**
 * Create a Permission Set Builder from an existing Permission Set
 *
 * @mustfree{DOFPermissionSetBuilder_Destroy or @ref DOFPermissionSetBuilder_BuildAndDestroy}
 *
 * @param permissions The Permission Set from which to create a new builder.  This should not be NULL.
 *
 * @return The @ref DOFPermissionSetBuilder.
 * @exception NULL is returned if the builder cannot be created.
 */
extern DOF_OAL_DECLSPEC DOFPermissionSetBuilder DOFPermissionSetBuilder_Create_PermissionSet( DOFPermissionSet permissions );

/**
   Destroy this DOFPermissionSetBuilder instance.

   @param self The Permission Set Builder.  If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFPermissionSetBuilder_Destroy( DOFPermissionSetBuilder self );

/**
   Add a permission to the builder. Multiple permissions can be added.

   @param self The Permission Set Builder.  This should not be NULL.
   @param permission The Permission to add to Builder.  This should not be NULL.

   @return TRUE is returned to indicate the permission was added.
   @exception FALSE is returned to indicate the permission could not be added.
 */
extern DOF_OAL_DECLSPEC boolean DOFPermissionSetBuilder_AddPermission( DOFPermissionSetBuilder self, DOFPermission permission );

/**
 * Build an immutable @ref DOFPermissionSet from this DOFPermissionSetBuilder instance.
 * @mustfree{DOFPermissionSet_Destroy}
 * @param self The Permission Set Builder.  This should not be NULL.
 * @return An immutable @ref DOFPermissionSet.
 * @exception NULL is returned if a permission set cannot be constructed.
 */
extern DOF_OAL_DECLSPEC DOFPermissionSet DOFPermissionSetBuilder_Build( DOFPermissionSetBuilder self );

/**
 * Build an immutable @ref DOFPermissionSet from this DOFPermissionSetBuilder instance, then destroy this DOFPermissionSetBuilder instance.
 * This is equivalent to calling @ref DOFPermissionSetBuilder_Build(@a self) immediately followed by @ref DOFPermissionSetBuilder_Destroy(@a self).
 * @param self The Permission Set Builder.  This should not be NULL.
 * @return An immutable @ref DOFPermissionSet.
 * @exception NULL is returned if the @ref DOFPermissionSet cannot be created, or an invalid parameter was detected.  Regardless, the DOFPermissionSetBuilder, if not NULL, is destroyed.
 */
extern DOF_OAL_DECLSPEC DOFPermissionSet DOFPermissionSetBuilder_BuildAndDestroy( DOFPermissionSetBuilder self );

/**
 * Report the number of bytes required in a buffer to marshal a @ref DOFPermissionSet.
 *
 * @param self The @ref DOFPermissionSet. This must not be NULL.
 *
 * @return The marshaled size of the @ref DOFPermissionSet.
 */
extern DOF_OAL_DECLSPEC uint32 DOFPermissionSet_GetMarshalSize( DOFPermissionSet self );

/**
 * Marshal a @ref DOFPermissionSet into an existing buffer.
 *
 * @param self The DOFPermissionSet to be marshaled. This must not be NULL.
 * @param size The size of the buffer in which to marshal the DOFPermissionSet.
 * @param buffer The buffer in which to marshal the DOFPermissionSet. This must not be NULL, and must contain at least
 *               @a size bytes.
 *
 * @return The marshaled size of the @ref DOFPermissionSet.
 *
 * @exception 0 will be returned if the marshaled data does not fit in the buffer, or an invalid parameter is detected.
 */
extern DOF_OAL_DECLSPEC uint32 DOFPermissionSet_Marshal( DOFPermissionSet self, uint32 size, uint8 * buffer );

/**
 * Unmarshals a @ref DOFPermissionSet from a buffer.
 *
 * @param size The size of the buffer, in bytes, from which to unmarshal the @ref DOFPermissionSet.
 * @param buffer The buffer from which to unmarshal the @ref DOFPermissionSet. This must not be NULL, and must contain at
 *               least @a size bytes.
 *
 * @return DOFPermissionSet The DOFPermissionSet that is stored in the buffer.
 * @exception NULL is returned if the buffer contents of @a size could not be unmarshalled to a valid @ref
 *            DOFPermissionSet, or there are insufficient resources to allocate the @ref DOFPermissionSet.
 */
extern DOF_OAL_DECLSPEC DOFPermissionSet DOFPermissionSet_Create_Unmarshal( uint32 size, const uint8 * buffer );

/**
@}
**/

/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* DOF_PERMISSIONSET_H */
