/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_PROVIDERINFO_H
#define DOF_PROVIDERINFO_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DOFProviderInfo DOFProviderInfo
    @ingroup DOF_Framework

    Represents the source (Provider) of a result or exception returned by an operation.

    Operations that return results (either synchronously or asynchronously) are
    often able to provide additional information about the results beyond just
    the response values themselves (such as where they originated).  The values
    are contained in a DOFResult, while the additional information is contained
    in this class.

@{
**/


/*
**  "Public" Type and Method Declarations
*/

/* DOFProviderInfo Methods */

/**
 * Get the DOFConnectionState from identified by this DOFProviderInfo.
 *
 * @param self The Provider Information.  This must not be NULL.
 *
 * @return The @ref DOFConnectionState.
 */
extern DOF_OAL_DECLSPEC DOFConnectionState DOFProviderInfo_GetConnectionState( DOFProviderInfo self );

/**
 * Get the provider Object ID identified by this DOFProviderInfo.
 *
 * @note The return value is owned by the DOFProviderInfo, and must not be destroyed individually.  If it is needed beyond the scope of this DOFProviderInfo, it must be copied by the application.
 *
 * @param  self The Provider Information.  This should not be NULL.
 *
 * @return The provider Object ID.
 *
 * @exception NULL is returned if an invalid argument is detected.
 */
extern DOF_OAL_DECLSPEC DOFObjectID DOFProviderInfo_GetProviderID( const DOFProviderInfo self );

/**
 * Get the provider's address identified by this DOFProviderInfo.
 *
 * @note The return value is owned by the DOFProviderInfo, and must not be destroyed individually.  If it is needed beyond the scope of this DOFProviderInfo, it must be copied by the application.
 *
 * @param  self The Provider Information.  This should not be NULL.
 *
 * @return The provider's address.
 *
 * @exception NULL is returned if an invalid argument is detected.
 */
extern DOF_OAL_DECLSPEC DOFAddress DOFProviderInfo_GetProviderAddress( const DOFProviderInfo self );

/**
 *  Create a new copy of an existing @ref DOFProviderInfo.
 *  @mustfree{DOFProviderInfo_Destroy}
 *  @param  self The @ref DOFProviderInfo to be copied.  This should not be NULL.
 *  @return A new copy of the @ref DOFProviderInfo.
 *  @exception  NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DOFProviderInfo DOFProviderInfo_Copy( DOFProviderInfo self );

/**
 * Destroy a DOFProviderInfo.
 * Any further references to data within the DOFProviderInfo are no longer valid and must not be used.
 *
 * @param  self The Provider Information.  If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFProviderInfo_Destroy( DOFProviderInfo self );


/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* DOF_PROVIDERINFO_H */
