/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_QUERY_H
#define DOF_QUERY_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DOFQuery  DOFQuery
    @ingroup DOF_Framework

    A DOFQuery represents a way to receive notification of the availability
    of a set of Objects and/or Interfaces in the DOF network.

    A DOFQuery is used by the application to query the @ref DOFSystem to
    discover available Objects that can be accessed.  Queries are always
    against the current known state of the network, and can only report
    about Objects and Interfaces that match current Interest operations.

    A DOFQueryBuilder is used to construct a DOFQuery.  Filter blocks can
    be added to the builder, indicating the Object IDs and Interface IDs
    that are desired.  A query match style can be used to indicate how
    Object IDs are matched (see @ref DOFQueryBuilder_SetMatchStyle), and
    a restriction list can be set so that only the interfaces in the list
    are passed to the application callback (normally all interfaces
    available on a matching Object are passed to the callback).

@{
**/

/*
**  "Public" Type and Method Declarations
*/

/* DOFQuery  Methods */

/**
    Create a new, empty query specification builder.

    By default, there are no filters or restrictions in the builder, and if
    built, the resulting query will not match any objects.  At least one filter
    must be added to the query specification to be of use.  The default match
    style is set to DOFQUERYMATCHSTYLE_AT_LEAST.

    @mustfree{DOFQueryBuilder_Destroy or @ref DOFQueryBuilder_BuildAndDestroy}

    The builder uses the following defaults:
     -Filters: none
     -Restrictions: none
     -Match Style: DOFQUERYMATCHSTYLE_AT_LEAST

    @return The newly created query builder.
    @exception NULL is returned if a query builder could not be created.

    @see    DOFQueryBuilder_Destroy
**/
extern DOF_OAL_DECLSPEC DOFQueryBuilder  DOFQueryBuilder_Create( void );

/**
 Create a query specification builder from an existing query.

 @mustfree{DOFQueryBuilder_Destroy or @ref DOFQueryBuilder_BuildAndDestroy}

 @param query A query to base the call on. This should not be NULL.

 @return The newly created query builder.
 @exception NULL is returned if a query builder could not be created or @a query is NULL.

 @see    DOFQueryBuilder_Destroy
**/
extern DOF_OAL_DECLSPEC DOFQueryBuilder DOFQueryBuilder_Create_Query( DOFQuery query );

/**
    Destroy a query specification builder.

    This frees the resources allocated by @ref DOFQueryBuilder_Create.

    @param  self  The query builder.  If NULL, no action is taken.

    @see    DOFQueryBuilder_Create
**/
extern DOF_OAL_DECLSPEC void DOFQueryBuilder_Destroy( DOFQueryBuilder self );

/**
    Destroy a query specification.

    This frees the resources allocated by @ref DOFQueryBuilder_Build or @ref DOFQueryBuilder_BuildAndDestroy.

    @param  self  The Query.  If NULL, no action is taken.

    @see    DOFQueryBuilder_Build
    @see    DOFQueryBuilder_BuildAndDestroy
**/
extern DOF_OAL_DECLSPEC void DOFQuery_Destroy( DOFQuery self );

/**
    Add a filter into the query specification builder.

    Both Object IDs and Interface IDs can be added as a group into the query.
    If Object IDs are specified, the match is restricted to only that list of
    Objects.  If Interface IDs are specified, the match is restricted to
    Objects that provide ALL of the specified Interfaces (Objects that have
    only a subset, or none, of the specified Interfaces will not match).  An
    empty list (either Objects or Interfaces) is like a wildcard, where
    everything matches.  This gives four options:

    -# No Objects, No Interfaces: Match all available Objects and Interfaces.
    This is the simplest filter, but may cause a query operation to show many
    unwanted Objects and/or Interfaces.
    -# n Objects, No Interfaces: Match any of the specified Objects, no matter
    what Interfaces they provide.  If the set of Objects that are desired are
    already known, this is a quick way to learn of the available Objects in the
    system.
    -# No Objects, m Interfaces: Match any Objects that that have ALL of the
    specified Interfaces.  This is an easy way to find particular types of
    Objects (ones that provide a known set of Interfaces), but whose Object IDs
    are not known.
    -# n Objects, m Interfaces: Match any of the specified Objects, but only if
    they have ALL of the specified Interfaces.  This is the most restrictive
    filter.

    To add disjoint sets of filters (for example, "All Objects that provide I1
    and I2 AND all Objects that provide I3 and I4", these must be added to the
    query as two separate filters (using two calls to this function).

    @note If an Object matches the query, then ALL Interfaces for that Object
    are disclosed, even if they are not explicitly listed in the query.  This
    can significantly increase the number of unwanted Interfaces that are
    disclosed.  Consider using the @ref DOFQueryBuilder_AddRestriction method to
    restrict the system to only notify of Interfaces that the application
    desires.

    @param  self        The QueryBuilder.  If NULL, no action is taken.
    @param  objectIDCount Number of Object IDs.
    @param  objectIDs   List of Object IDs.  This must not be NULL if
                        @a objectIDCount > 0, and should not contain NULL or the
                        broadcast or multicast IDs.
    @param  interfaceIDCount Number of Interface IDs.
    @param  interfaceIDs List of Interface IDs.  This must not be NULL if
                         @a interfaceIDCount > 0, and should not contain NULL
                         or the wildcard ID.

    @retval Non-Zero (TRUE) The filter was added.
    @exception Zero (FALSE) The filter could not be added.
**/
extern DOF_OAL_DECLSPEC boolean DOFQueryBuilder_AddFilter( DOFQueryBuilder self, uint32 objectIDCount, const DOFObjectID objectIDs[], uint32 interfaceIDCount, const DOFInterfaceID interfaceIDs[] );

/**
    Add interfaces to the restriction list.

    The restriction list is the exclusive list of all interfaces that shall be
    considered to match the query.  An empty list means that all interfaces
    shall be considered.  The restriction list is empty by default.

    @param  self         The QueryBuilder.  If NULL, no action is taken.
    @param  interfaceIDCount  Number of interfaces.
    @param  interfaceIDs List of Interface IDs to add to the restriction list.  This must not be NULL if @a interfaceIDCount > 0.
    
    @return TRUE if restriction was added, FALSE otherwise.
**/
extern DOF_OAL_DECLSPEC boolean DOFQueryBuilder_AddRestriction( DOFQueryBuilder self, uint32 interfaceIDCount, const DOFInterfaceID interfaceIDs[] );

/**
 * Set filter matching style.
 *
 * - For DOFQUERYMATCHSTYLE_AT_LEAST, match base OID plus at least a set of attributes (additional attributes are permitted).  This is the default.
 * - For DOFQUERYMATCHSTYLE_BASE, match base OIDs only (no attributes).
 * - For DOFQUERYMATCHSTYLE_EXACT, match base OIDs and attributes exactly.
 *
 * Please refer to @ref DOFQueryBuilder_Create for the default value for this setting.
 *
 * @param self The Query.  If NULL, no action is taken.
 * @param matchStyle One of DOFQueryMatchStyle_t constants. 
 *  
 * @retval TRUE Match style was set. 
 * @exception FALSE Failed to set. 
 */
extern DOF_OAL_DECLSPEC boolean DOFQueryBuilder_SetMatchStyle( DOFQueryBuilder self, DOFQueryMatchStyle_t matchStyle );

/**
    Build a DOFQuery.

    @param self The QueryBuilder to build the query from.  This should not be NULL.

    @return The newly built query.
    @exception NULL is returned if an invalid argument is detected or the query cannot be built.
**/
extern DOF_OAL_DECLSPEC DOFQuery DOFQueryBuilder_Build( DOFQueryBuilder self );

/**
     Build a DOFQuery and destroy the builder.

     @param self The QueryBuilder to build the query from.  This should not be NULL.  This will be invalid (destroyed) after this call is made.

     @return The newly built query.
     @exception NULL is returned if an invalid argument is detected or the query cannot be built.  Regardless, the @a self builder is destroyed.
**/
extern DOF_OAL_DECLSPEC DOFQuery DOFQueryBuilder_BuildAndDestroy( DOFQueryBuilder self );

/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* DOF_QUERY_H */
