/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_REQUEST_H
#define DOF_REQUEST_H


#include <dof/core.h>


#ifdef  __cplusplus
extern "C" {
#endif


/** @defgroup DOFRequest DOFRequest
    @ingroup DOF_Framework

    Represents a Request made to a Provider.

    Also contains an "exception" status result the provider may wish
    to pass back to the Requestor.

@{
**/


/*
**  "Public" Type and Method Declarations
*/


/* DOFRequest Methods */

/**
 * Return a @ref DOFValue list result to a requestor.  This completes the
 * Invoke request, and no further references to the request are allowed.
 * The @a request must represent a method Invoke request.
 *
 * @param request The Request.  This must not be NULL.
 * @param valuesCount The count of values in the @a values array.
 * @param values  The value list to return.  This may be NULL if the method
 *                does not have any output parameters; otherise, this should
 *                not be NULL.
 */
extern DOF_OAL_DECLSPEC void DOFRequestInvoke_Return( DOFRequest request, uint16 valuesCount, const DOFValue values[] );


/**
 * Return a @ref DOFValue result to a requestor.  This completes the Get
 * request, and no further references to the request are allowed.
 * The @a request must represent a property Get request.
 *
 * @param request The Request.  This must not be NULL, and must represent a Get request.
 * @param value   The value to return.  This should not be NULL.
 */
extern DOF_OAL_DECLSPEC void DOFRequestGet_Return( DOFRequest request, DOFValue value );


/**
 * Return whether or not this request is the result of subscription activity.
 * Subscription activity may either be an initial response to a Subscribe
 * request, or a property update resulting from a call to @ref
 * DOFObject_Changed.
 * The @a request must represent a property Get request.
 *
 * @param request The Request.  This must not be NULL, and must represent a Get request.
 *
 * @retval Zero (FALSE) if the request is a normal Get request.
 * @retval Non-Zero (TRUE) if the request is the result of subscription
 *         activity.
 */
extern DOF_OAL_DECLSPEC boolean DOFRequestGet_IsSubscribe( DOFRequest request );


/**
 * Return a successful result (without any values) to a requestor.  This
 * completes the Set request, and no further references to the request are
 * allowed.  The @a request must represent a property Set request.
 *
 * @param request The Request.  This must not be NULL, and must represent a Set request.
 */
extern DOF_OAL_DECLSPEC void DOFRequestSet_Return( DOFRequest request );


/**
 * Accepts the subscription request.
 * The request remains valid until the provider SubscribeComplete callback is called.
 * The application may later choose to terminate the subscription (as long as the request remains valid) by calling @ref DOFRequest_Throw.
 * The @a request must represent a Subscribe request.
 *
 * @param request The Request.  This must not be NULL, and must represent a Subscribe request.
 */
extern DOF_OAL_DECLSPEC void DOFRequestSubscribe_Return( DOFRequest request );


/**
 * Accepts the registration request.
 * The request remains valid until the provider RegisterComplete callback is called.
 * The application may later choose to terminate the registration (as long as the request remains valid) by calling @ref DOFRequest_Throw.
 * The @a request must represent a Register request.
 *
 * @param request The Request.  This must not be NULL, and must represent a Register request.
 */
extern DOF_OAL_DECLSPEC void DOFRequestRegister_Return( DOFRequest request );


/**
 * Accepts the session request.
 * The request remains valid until the provider SessionComplete callback is called.
 * The application may later choose to terminate the session (as long as the request remains valid) by calling @ref DOFRequest_Throw.
 * The @a request must represent a Session request.
 *
 * @param request The Request.  This must not be NULL, and must represent a Session request.
 */
extern DOF_OAL_DECLSPEC void DOFRequestSession_Return( DOFRequest request );

/**
 *  Set the DOFOperationSessionDataTransform to use on this session.
 *
 *  The data transform can be called at most once per session (subsequent calls will have no effect).
 *
 *  @param request The DOFRequest for the session.  Should not be @c NULL.
 *  @param transform    The transform for the session.  A value of @c NULL will have no effect.
 *
 *  @retval TRUE    @a transform was set successfully
 *  @retval FALSE   there was a failure to set @a transform
 */
extern DOF_OAL_DECLSPEC boolean DOFRequestSession_SetDataTransform( DOFRequest request, DOFOperationSessionDataTransform transform );


/**
    Get the connection type code of the request.

    @param  request The Request.  This must not be NULL.

    @return The request connection type code, which is one of the @ref DOFRequestType constants.
**/
extern DOF_OAL_DECLSPEC DOFRequestType DOFRequest_GetRequestType( const DOFRequest request );


/**
    Get the requestor Object ID for the request.

    @param  request The Request.  This must not be NULL.

    @return The requestor Object ID.
**/
extern DOF_OAL_DECLSPEC DOFObjectIDSource DOFRequest_GetRequestorID( const DOFRequest request );


/**
    Get the requestor's address.

    @param  request The Request.  This must not be NULL.

    @return The requestor's address.
**/
extern DOF_OAL_DECLSPEC DOFAddress DOFRequest_GetRequestorAddress( const DOFRequest request );


/**
   Retrieve the associated @ref DOFConnectionState for the given DOFRequest.

   @param request TheRequest.  This must not be NULL.

   @return The @ref DOFConnectionState associated with the DOFRequest.
 */
extern DOF_OAL_DECLSPEC DOFConnectionState DOFRequest_GetConnectionState( const DOFRequest request );

/**
   Retrieve the associated @ref DOFSecurityScope for the given DOFRequest.

   @param request TheRequest.  This must not be NULL.

   @return The @ref DOFSecurityScope associated with the DOFRequest.
 */
extern DOF_OAL_DECLSPEC DOFSecurityScope DOFRequest_GetSecurityScope( const DOFRequest request );


/**
 * Throw a @ref DOFException to the Requestor and complete the request.
 * For most requests, this completes the request, and no further references to the request are allowed.
 * Subscribe, Register, and Session requests remain valid until their respective complete provider
 * callback is called; however, they can no longer be used to accept or throw another exception.
 *
 * @param request The Request.  This must not be NULL.
 * @param except  The Exception.  This should not be NULL (if NULL, a
 *                generic exception is returned to the requestor).
 */
extern DOF_OAL_DECLSPEC void DOFRequest_Throw( DOFRequest request, DOFException except );

/**
 * Returns the time remaining of the request, in milliseconds.
 *
 * If the Operation has timed out, the time will be @c 0.
 *
 * @param request The Request.  This must not be NULL.
 *
 * @return The remaining time left for the request, in milliseconds.
 */
extern DOF_OAL_DECLSPEC uint32 DOFRequest_GetTimeRemaining( DOFRequest request );
/**
@}
**/


#ifdef  __cplusplus
}
#endif


#endif /* DOF_REQUEST_H */
