/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_RESULT_H
#define DOF_RESULT_H


#include <dof/core.h>


#ifdef  __cplusplus
extern "C" {
#endif


/** @defgroup DOFResult DOFResult
    @ingroup DOF_Framework

    Represents a result of an operation, or an "exception" if the operation failed.

    Operations that return results (either synchronously or asynchronously)
    often require more information than just the response value(s). To handle
    this additional information the DOFResult type is used.

    The DOFResult can include information about the responder, as well as
    completion or exception status.

@{
**/


/*
**  "Public" Type and Method Declarations
*/


/* DOFResult Methods */


/**
    Get the provider information, if any, associated with the DOFResult.

    @note The return value is owned by the Result, and must not be destroyed individually. If it is needed beyond the
          scope of the Result, it must be copied (and later freed) by the application.

    @param  result  The Result.  This must not be NULL.

    @return The @ref DOFProviderInfo associated with the DOFResult, or NULL if
            there is no provider information associated with the DOFResult.
**/
extern DOF_OAL_DECLSPEC DOFProviderInfo DOFResult_GetProviderInfo( const DOFResult result );


/**
    Get the DOFValue held by the Result.

    @note The return value is owned by the Result, and must not be destroyed individually. If it is needed beyond the
          scope of the Result, it must be copied (and later freed) by the application.

    @param  result The Result.  This must not be NULL.

    @return The Value.

    @exception NULL is returned if the Result does not contain a Value.
**/
extern DOF_OAL_DECLSPEC DOFValue DOFResult_GetValue( const DOFResult result );


/**
    Get the list of DOFValues held by the Result.

    @note The return value is owned by the Result, and must not be destroyed individually. If it is needed beyond the
          scope of the Result, it must be copied (and later freed) by the application.

    @param  result The Result.  This must not be NULL.

    @return An array of DOFValues.  The list may be empty.

    @exception NULL is returned if the Result does not contain a list of DOFValues.

    @see DOFResult_GetValueListCount
*/
extern DOF_OAL_DECLSPEC const DOFValue * DOFResult_GetValueList( const DOFResult result );


/**
 * Get the count of DOFValues held by the Result.
 *
 * @param  result The Result.  This must not be NULL.
 *
 * @return The count of DOFValues held by the Result (may be 0).
 */
extern DOF_OAL_DECLSPEC uint16 DOFResult_GetValueListCount( const DOFResult result );


/**
    Destroy a DOFResult.
    This also destroys any value, valuelist or resultset that may be held by the DOFResult. Any further references to
    data within the DOFResult are no longer valid and must not be used.

    @param  result The Result.  If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFResult_Destroy( DOFResult result );


/**
@}
**/


#ifdef  __cplusplus
}
#endif


#endif /* DOF_RESULT_H */
