/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_ROUTEINFO_H
#define DOF_ROUTEINFO_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DOFRouteInfo DOFRouteInfo
    @ingroup DOF_Framework

@{
**/


/*
**  "Public" Method Declarations
*/

/**
 * Get the distance to a Provider over streaming (lossless) connections.
 * This does not include the distance over any datagram (lossy) connections.
 * @param self The Provider Route Information.  This must not be NULL.
 * @return The distance, in hops, to the Provider over streaming (lossless) connections.
 * @see DOFRouteInfo_GetDatagramDistance
 */
extern DOF_OAL_DECLSPEC uint16 DOFRouteInfo_GetStreamDistance( DOFRouteInfo self );

/**
 * Get the distance to a Provider over datagram (lossy) connections.
 * This does not include the distance over any stream (lossless) connections.
 * @param self The Provider Route Information.  This must not be NULL.
 * @return The distance, in hops, to the Provider over datagram (lossy) connections.
 * @see DOFRouteInfo_GetStreamDistance
 */
extern DOF_OAL_DECLSPEC uint16 DOFRouteInfo_GetDatagramDistance( DOFRouteInfo self );

/**
 * Get the Provider's Object ID (see @ref DOFObjectID).
 * @note The return value is owned by the DOFRouteInfo, and must not be destroyed individually.  If it is needed beyond the scope of this DOFRouteInfo, it must be copied by the application.
 * @param self The Provider Route Information.  This must not be NULL.
 * @return The Provider's Object ID.
 */
extern DOF_OAL_DECLSPEC DOFObjectID DOFRouteInfo_GetProviderID( DOFRouteInfo self );

/**
 * Get the Provider's address.
 * @note The return value is owned by the DOFRouteInfo, and must not be destroyed individually.  If it is needed beyond the scope of this DOFRouteInfo, it must be copied by the application.
 * @param  self The Provider Route Information.  This must not be NULL.
 * @return The Provider's address.
 */
extern DOF_OAL_DECLSPEC DOFAddress DOFRouteInfo_GetProviderAddress( const DOFRouteInfo self );

/**
 * Get the Provider's Interface ID (see @ref DOFInterfaceID).
 * @note The return value is owned by the DOFRouteInfo, and must not be destroyed individually.  If it is needed beyond the scope of this DOFRouteInfo, it must be copied by the application.
 * @param self The Provider Route Information.  This must not be NULL.
 * @return The Provider's Interface ID.
 */
extern DOF_OAL_DECLSPEC DOFInterfaceID DOFRouteInfo_GetInterfaceID( DOFRouteInfo self );

/**
 * Get the @ref DOFConnectionState representing the connection over which the Provider is available.
 * @note The return value is owned by the DOFRouteInfo, and must not be destroyed individually.  If it is needed beyond the scope of this DOFRouteInfo, it must be copied by the application.
 * @param self The Provider Route Information.  This must not be NULL.
 * @return The @ref DOFConnectionState representing the connection to the Provider.
 */
extern DOF_OAL_DECLSPEC DOFConnectionState DOFRouteInfo_GetConnectionState( DOFRouteInfo self );

/**
 * Get the @ref DOFObjectIDSource representing the source of the provider.
 * This is not necessarily the actual source ID of the provider; it is
 * merely the source from the limited perspective of the local node within
 * the network.  This ID, in combination with the operation count (see
 * @ref DOFRouteInfo_GetOperationCount), uniquely identify a single provider.
 * @note The return value is owned by the DOFRouteInfo, and must not be destroyed individually.  If it is needed beyond the scope of this DOFRouteInfo, it must be copied by the application.
 * @param self The Provider Route Information.  This must not be NULL.
 * @return The @ref DOFObjectIDSource representing the source of the provider.
 */
extern DOF_OAL_DECLSPEC DOFObjectIDSource DOFRouteInfo_GetSourceID( DOFRouteInfo self );

/**
 * Get the operation count from the source of the provider.
 * This count, in combination with the source ID (see
 * @ref DOFRouteInfo_GetSourceID), uniquely identify a single provider.
 * @param self The Provider Route Information.  This must not be NULL.
 * @return The operation count from the source of the provider.
 */
extern DOF_OAL_DECLSPEC uint32 DOFRouteInfo_GetOperationCount( DOFRouteInfo self );

/**
 * Destroy the Provider Route Information.
 * @param self The Provider Route Information.  If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFRouteInfo_Destroy( DOFRouteInfo self );


/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* DOF_ROUTEINFO_H */
