/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_RUNTIME_H
#define DOF_RUNTIME_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DOFRuntime DOFRuntime
    @ingroup DOF_Framework

    The Runtime class enables access to runtime information about the associated DOF, including traffic statistics 
    and information about active connections and servers.
    
    @see DOF_GetRuntime

@{
**/

/*
**  "Public" Type and Method Declarations
*/

/**
 * Adds a @ref DOFRuntimeConnectionStateListener that is notified when connections are established or go down.
 *
 * @param self The DOFRuntime. This must not be @c NULL.
 * @param listener The listener to add. This should not be @c NULL.
 * @return TRUE if the listener was added.
 * @exception FALSE if the listener could not be added, illegal arguments were detected, or the listener was already added.
 * @see DOFRuntime_RemoveConnectionListener
 */
extern DOF_OAL_DECLSPEC boolean DOFRuntime_AddConnectionListener( DOFRuntime self, DOFRuntimeConnectionStateListener listener );

/**
 * Removes a previously added @ref DOFRuntimeConnectionStateListener. When the listener is removed, this function causes
 * the listener's removed callback to be called, if it defines one.
 *
 * @param self The DOFRuntime. This must not be @c NULL.
 * @param listener The listener to remove.  This should not be @c NULL. If the listener is not found, no action is taken.
 * @see DOFRuntime_AddConnectionListener
 */
extern DOF_OAL_DECLSPEC void DOFRuntime_RemoveConnectionListener( DOFRuntime self, DOFRuntimeConnectionStateListener listener );

/**
 * Adds a @ref DOFRuntimeServerStateListener that is notified when servers are started or stopped.
 *
 * @param self The DOFRuntime. This must not be @c NULL.
 * @param listener The listener to add. This should not be @c NULL.
 * @return TRUE if the listener was added.
 * @exception FALSE if the listener could not be added, illegal arguments were detected, or the listener was already added.
 * @see DOFRuntime_RemoveServerListener
 */
extern DOF_OAL_DECLSPEC boolean DOFRuntime_AddServerListener( DOFRuntime self, DOFRuntimeServerStateListener listener );

/**
 * Removes a previously added @ref DOFRuntimeServerStateListener. When the listener is removed, this function causes
 * the listener's removed callback to be called, if it defines one.
 *
 * @param self The DOFRuntime. This must not be @c NULL.
 * @param listener The listener to remove. This should not be @c NULL. If the listener is not found, no action is taken.
 * @see DOFRuntime_AddServerListener
 */
extern DOF_OAL_DECLSPEC void DOFRuntime_RemoveServerListener( DOFRuntime self, DOFRuntimeServerStateListener listener );

/**
 * Gets a list of connections (@ref DOFConnection), both inbound and outbound, over which a provide operation
 * has been sent or received for a specified @ref DOFInterface on a specified @ref DOFObject. In addition to 
 * specifying exact bindings, you can do the following:
 * 
 * - Pass DOFOBJECTID_BROADCAST to get connections that are providing a specific interface on any object.
 * - Pass DOFINTERFACEID_WILDCARD to get connections that are providing any interface on a specific object.
 * - Pass both to get all connections that are providing any binding.
 * 
 * Only remote provide operations that have been received over a connection will cause the DOFConnection 
 * to appear in the collection. Local provide operations sent over a connection do not.
 *
 * @mustfree{DOFConnectionList_Destroy}
 *
 * @param self The DOFRuntime. This must not be @c NULL.
 * @param objectID The object identifier. This should not be @c NULL.
 * @param interfaceID The interface identifier. This should not be @c NULL.
 *
 * @return A list of connections that are currently carrying provide operations.
 * @exception NULL is returned if an invalid parameter is detected or the list could not be allocated.
 * @see DOFObject_BeginProvide
 */
extern DOF_OAL_DECLSPEC DOFConnectionList DOFRuntime_GetProvidingConnections( const DOFRuntime self, const DOFObjectID objectID, const DOFInterfaceID interfaceID );

/**
 * Gets a list of all connections, both inbound and outbound, that are currently open.
 *
 * @mustfree{DOFConnectionList_Destroy}
 *
 * @param self The DOFRuntime. This must not be @c NULL.
 *
 * @return A list of all connections currently open.
 * @exception NULL is returned if an error occurs.
 * @see DOFRuntime_GetConnectionCount
 */
extern DOF_OAL_DECLSPEC DOFConnectionList DOFRuntime_GetConnections( const DOFRuntime self );

/**
 * Gets a list of all servers that are currently started.
 *
 * @mustfree{DOFServerList_Destroy}
 *
 * @param self The DOFRuntime. This must not be @c NULL.
 *
 * @return A list of all started servers.
 * @exception NULL is returned if an error occurs.
 * @see DOFRuntime_GetServerCount
 */
extern DOF_OAL_DECLSPEC DOFServerList DOFRuntime_GetServers( const DOFRuntime self );

/**
 * Gets the total number of connections, both inbound and outbound, that are currently open.
 *
 * @param self The DOFRuntime. This must not be @c NULL.
 *
 * @return The total number of open connections.
 * @see DOFRuntime_GetConnections
 */
extern DOF_OAL_DECLSPEC uint16 DOFRuntime_GetConnectionCount( const DOFRuntime self );

/**
 * Gets the total number of servers that are currently started.
 *
 * @param self The DOFRuntime. This must not be @c NULL.
 *
 * @return The total number of started servers.
 * @see DOFRuntime_GetServers
 */
extern DOF_OAL_DECLSPEC uint16 DOFRuntime_GetServerCount( const DOFRuntime self );

/**
 * Gets the total number of provides currently active. The count includes both local provide operations and provide 
 * operations received from remote nodes.
 *
 * @param self The DOFRuntime. This must not be @c NULL.
 *
 * @return The number of provides currently active.
 */
extern DOF_OAL_DECLSPEC uint32 DOFRuntime_GetProvideCount( const DOFRuntime self );

/**
 * Gets a count of the number of data bytes received on all connections and servers.
 *
 * @param self The DOFRuntime. This must not be @c NULL.
 *
 * @return The current count of bytes received on all connections and servers.
 */
extern DOF_OAL_DECLSPEC uint64 DOFRuntime_GetReceiveByteCount( const DOFRuntime self );

/**
 * Gets a count of the number of data bytes sent over all connections and servers.
 *
 * @param self The DOFRuntime. This must not be @c NULL.
 *
 * @return The current count of bytes sent over all connections and servers.
 */
extern DOF_OAL_DECLSPEC uint64 DOFRuntime_GetSendByteCount( const DOFRuntime self );

/**
 * Gets a count of the packets received on all connections and servers.
 *
 * @param self The DOFRuntime. This must not be @c NULL.
 *
 * @return The current count of packets received on all connections and servers.
 */
extern DOF_OAL_DECLSPEC uint64 DOFRuntime_GetReceivePacketCount( const DOFRuntime self );

/**
 * Gets a count of the packets sent over all connections and servers.
 *
 * @param self The DOFRuntime. This must not be @c NULL.
 *
 * @return The current count of packets sent over all connections and servers.
 */
extern DOF_OAL_DECLSPEC uint64 DOFRuntime_GetSendPacketCount( const DOFRuntime self );

#ifdef  __cplusplus
}
#endif

/**
@}
**/

#endif /* DOF_RUNTIME_H */
