/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_SECURITYMODE_H
#define DOF_SECURITYMODE_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif


/*
**  "Public" Type and Method Declarations
*/

/** @defgroup DOFSecurityMode DOFSecurityMode
    @ingroup DOF_Framework

@{
**/

/**
 * The Security Mode Type represents the different security modes available, as defined by DOF Security.
 */
typedef enum DOFSecurityModeType_t
{
    DOFSECURITYMODETYPE_CCM         /**< DOF Security mode using Counter with CBC-MAC (CCM). **/
} DOFSecurityModeType;


/**
    Destroy the given @ref DOFSecurityMode.

    @param  self  The @ref DOFSecurityMode to destroy.  If NULL, no action is taken.
    @see DOFSecurityModeCCM_Create
**/
extern DOF_OAL_DECLSPEC void DOFSecurityMode_Destroy( DOFSecurityMode self );

/**
    Get the security mode type.
    @param  self  The @ref DOFSecurityMode.  This must not be NULL.
    @return The security mode type.
**/
extern DOF_OAL_DECLSPEC DOFSecurityModeType DOFSecurityMode_GetModeType( DOFSecurityMode self );

/**
 *  Create a new copy of an existing @ref DOFSecurityMode.
 *  @mustfree{DOFSecurityMode_Destroy}
 *  @param  self The @ref DOFSecurityMode to be copied.  This should not be NULL.
 *  @return A new copy of the @ref DOFSecurityMode.
 *  @exception  NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DOFSecurityMode DOFSecurityMode_Copy( const DOFSecurityMode self );

/**
 * Report the number of bytes required in a buffer to marshal a @ref DOFSecurityMode.
 *
 * @param self The @ref DOFSecurityMode. This must not be NULL.
 *
 * @return The marshaled size of the @ref DOFSecurityMode.
 */
extern DOF_OAL_DECLSPEC uint32 DOFSecurityMode_GetMarshalSize( const DOFSecurityMode self );

/**
 * Marshals the given @ref DOFSecurityMode to the given buffer.
 *
 * @param self The @ref DOFSecurityMode to marshal. This must not be NULL.
 * @param size The byte length of the given buffer
 * @param buffer The buffer to marshal into. This must not be NULL, and must contain at least
 *               @a size bytes.
 *
 * @return The marshaled size of the @ref DOFSecurityMode.
 *
 * @exception 0 will be returned if the marshaled data does not fit in the buffer, or an invalid parameter is detected.
 */
extern DOF_OAL_DECLSPEC uint32 DOFSecurityMode_Marshal( const DOFSecurityMode self, uint32 size, uint8 * buffer );


/** @defgroup DOFSecurityModeCCM DOFSecurityModeCCM
@{
**/


/**
    Create a new CCM Security Mode configuration.

    @mustfree{DOFSecurityMode_Destroy}

    @param algorithm The recognized cipher algorithm; refer to @ref DOFCipherAlgorithm.
    @param strength  The cipher security strength as defined by the OpenDOF Project; refer to @ref DOFCipherStrength.
    @param macSize   The size of the MAC to use for the CCM mode.  This should be an even integer between 4 and 16, inclusive.
    @param isEncrypt Flag indicating that the CCM mode should be encrypted (TRUE) or only authenticated (FALSE).

    @return A new @ref DOFSecurityModeCCM, created with the provided information.
    @exception NULL is returned if insufficient resources are availabe to create the mode, or if an invalid cipher or macSize is specified.
    @see DOFSecurityMode_Destroy
**/
extern DOF_OAL_DECLSPEC DOFSecurityMode DOFSecurityModeCCM_Create( DOFCipherAlgorithm algorithm, DOFCipherStrength strength, uint8 macSize, boolean isEncrypt );

/**
    Retrieve the Cipher specified for the @ref DOFSecurityModeCCM.

    @param  self  The @ref DOFSecurityModeCCM.  This must not be NULL and must represent a DOFSecurityModeCCM.

    @return The cipher algorithm.
**/
extern DOF_OAL_DECLSPEC DOFCipherAlgorithm DOFSecurityModeCCM_GetAlgorithm( const DOFSecurityMode self );

/**
    Retrieve the Cipher strength for the @ref DOFSecurityModeCCM.

    @param  self  The @ref DOFSecurityModeCCM.  This must not be NULL and must represent a DOFSecurityModeCCM.

    @return The cipher strength.
**/
extern DOF_OAL_DECLSPEC DOFCipherStrength DOFSecurityModeCCM_GetStrength( const DOFSecurityMode self );

/**
    Retrieve the MAC size specified for the @ref DOFSecurityModeCCM.

    @param  self  The @ref DOFSecurityModeCCM to be queried.  This must not be NULL and must represent a DOFSecurityModeCCM.

    @return The CCM MAC size.
**/
extern DOF_OAL_DECLSPEC uint8 DOFSecurityModeCCM_GetMACSize( const DOFSecurityMode self );

/**
    Report whether the given @ref DOFSecurityModeCCM specifies that data should be encrypted or not.  Data is always authenticated, regardless of this setting.

    @param self  The @ref DOFSecurityModeCCM to be queried.  This must not be NULL and must represent a DOFSecurityModeCCM.

    @retval Non-zero (@c TRUE) if the data is to be encrypted.
    @retval Zero (@c FALSE) if data is to only be authenticated.
**/
extern DOF_OAL_DECLSPEC boolean DOFSecurityModeCCM_IsEncrypt( const DOFSecurityMode self );


/**
@}
**/


/**
@}
**/


#ifdef  __cplusplus
}
#endif

#endif /* DOF_SECURITYMODE_H */
