/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_SERVER_H
#define DOF_SERVER_H

#include <dof/core.h>
#include <dof/transport.h>


#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DOFServerConfig DOFServerConfig
    @ingroup DOFServer

    The DOFServerConfig class represents most of the information necessary
    for a DOF to create a server.
@{
**/

/**
 * @name Builder
 * @{
 */

/**
 * Creates a server configuration builder, given the @ref DOFServerType and the listening address.
 * @mustfree{DOFServerConfigBuilder_Destroy or @ref DOFServerConfigBuilder_BuildAndDestroy}
 * 
 * The builder uses the following defaults:
 *  -Security Desire: @ref DOFSECURITYDESIRE_ANY
 *  -Credentials: none
 *  -Domain Credentials: none
 *  -Trusted Domains: none
 *  -Transport Config: NULL
 *  -Connection Limit: The maximum value of uint32 (effectively unlimited)
 *  -Protocol Negotiator: @ref DOF_DEFAULT_PROTOCOL_NEGOTIATOR
 *  -Maximum Send Silence: @ref DOF_TIMEOUT_NEVER
 *  -Maximum Receive Silence: 120,000 ms (2 minutes)
 * 
 * @param type The type of Server.  This must be one of the @ref DOFServerType constants.
 * @param listenAddress The listening @ref DOFAddress of the Server.  This should not be NULL.
 * @return A new @ref DOFServerConfigBuilder.
 * @exception NULL is returned if a server configuration builder could not be created.
 */
extern DOF_OAL_DECLSPEC DOFServerConfigBuilder DOFServerConfigBuilder_Create( DOFServerType type, DOFAddress listenAddress );

/**
 * Creates a server configuration builder, given an existing server configuration.
 * @mustfree{DOFServerConfigBuilder_Destroy or @ref DOFServerConfigBuilder_BuildAndDestroy}
 * @param config The existing Server Configuration.  This must not be NULL.
 * @return A new @ref DOFServerConfigBuilder.
 * @exception NULL is returned if a server configuration builder could not be created.
 */
extern DOF_OAL_DECLSPEC DOFServerConfigBuilder DOFServerConfigBuilder_Create_Config( DOFServerConfig config );

/**
 * Creates a server configuration builder, given an existing server state.
 * @mustfree{DOFServerConfigBuilder_Destroy or @ref DOFServerConfigBuilder_BuildAndDestroy}
 * @param config The existing Server state.  This must not be NULL.
 * @return A new @ref DOFServerConfigBuilder. This builder will not have any private data,
 *         and if the private data is needed it must be set before building this builder.
 * @exception NULL is returned if a server configuration builder could not be created.
 */
extern DOF_OAL_DECLSPEC DOFServerConfigBuilder DOFServerConfigBuilder_Create_State( DOFServerState config );

/**
 * Creates a new copy of an existing @ref DOFServerConfigBuilder.
 * @mustfree{DOFServerConfigBuilder_Destroy or @ref DOFServerConfigBuilder_BuildAndDestroy}
 * @param self The @ref DOFServerConfigBuilder to be copied.  This should not be NULL.
 * @return A new copy of the @ref DOFServerConfigBuilder.
 * @exception NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DOFServerConfigBuilder DOFServerConfigBuilder_Copy( DOFServerConfigBuilder self );

/**
   Sets the server type in the given DOFServerConfigBuilder.

   @param self The Server Config Builder.  If @a self is NULL, no action will be taken.
   @param type The type of Server.  This must be one of the @ref DOFServerType constants.

   @retval TRUE Succesfully set.
   @exception FALSE Failed to set.
 */
extern DOF_OAL_DECLSPEC boolean DOFServerConfigBuilder_SetServerType( DOFServerConfigBuilder self, DOFServerType type );

/**
   Sets the @ref DOFAddress in the given DOFServerConfigBuilder.

   @param self The Server Config Builder.  If @a self is NULL, no action will be taken.
   @param listenAddress The listening Address of the Server.  If @a listenAddress is NULL, then no change is made.

   @retval TRUE Succesfully set.
   @exception FALSE Failed to set.
 */
extern DOF_OAL_DECLSPEC boolean DOFServerConfigBuilder_SetAddress( DOFServerConfigBuilder self, DOFAddress listenAddress );

/**
   Sets the DOFSecurityDesire in the given DOFServerConfigBuilder.
   Please refer to @ref DOFServerConfigBuilder_Create for the default value for this setting.

   @param self The Server Config Builder.  If @a self is NULL, no action will be taken.
   @param desire The Security Desire of the Server.

   @retval TRUE Succesfully set.
   @exception FALSE Failed to set.
 */
extern DOF_OAL_DECLSPEC boolean DOFServerConfigBuilder_SetSecurityDesire( DOFServerConfigBuilder self, DOFSecurityDesire desire );

/**
   Sets the @ref DTLServerConfig in the given DOFServerConfigBuilder.
   Please refer to @ref DOFServerConfigBuilder_Create for the default value for this setting.

   @param self The Server Config Builder.  If @a self is NULL, no action will be taken.
   @param config The transport-specific configuration.  If @a config is NULL, the default transport-specific configuration is used.

   @retval TRUE Succesfully set.
   @exception FALSE Failed to set.
 */
extern DOF_OAL_DECLSPEC boolean DOFServerConfigBuilder_SetTransportConfig( DOFServerConfigBuilder self, DTLServerConfig config );

/**
   Sets the maximum number of inbound connections in the given DOFServerConfigBuilder.
   This prevents the server from accepting inbound connections beyond the specified maximum.
   Please refer to @ref DOFServerConfigBuilder_Create for the default value for this setting.

   @deprecated This functionality was replaced by @ref DOFServerConfigBuilder_SetConnectionLimit.

   @param self The Server Config Builder.  If @a self is NULL, no action will be taken.
   @param maxConnections The maximum number of inbound connections.  0 indicates no maximum (unlimited).

   @retval TRUE Succesfully set.
   @exception FALSE Failed to set.
 */
extern DOF_OAL_DECLSPEC boolean DOFServerConfigBuilder_SetMaxConnections( DOFServerConfigBuilder self, uint16 maxConnections );

/**
   Sets the maximum number of inbound connections in the given DOFServerConfigBuilder.
   This prevents servers created with this configuration from accepting inbound connections 
   beyond the specified maximum.
   Please refer to @ref DOFServerConfigBuilder_Create for the default value for this setting.

   @param self The Server Config Builder.  If @a self is NULL, no action will be taken.
   @param maxConnections The maximum number of inbound connections. Setting this value to zero
                         creates a server that does not accept inbound connections.

   @retval TRUE Succesfully set.
   @exception FALSE Failed to set.
 */
extern DOF_OAL_DECLSPEC boolean DOFServerConfigBuilder_SetConnectionLimit( DOFServerConfigBuilder self, uint32 maxConnections );

/**
 * Adds Credentials to the ServerConfigBuiler for the given domain. If the domainID is NULL the domain will be pulled from
 * the given Credentials.
 *
 * @param self The ServerConfigBuiler to add the credentials to.  If @a self is NULL, no action will be taken.
 * @param credentials The Credentials to add.  If @a credentials is NULL, no credentials will be used.
 *
 * @retval Non-Zero (TRUE) if the credentials were added successfully.
 * @exception Zero (FALSE) is returned if an invalid parameter was detected or insufficient resources were available to add the credentials.
 */
extern DOF_OAL_DECLSPEC boolean DOFServerConfigBuilder_AddCredentials( DOFServerConfigBuilder self, DOFCredentials credentials );

/**
* Sets the wildcard credentials for the server. These DOFCredentials will be used to attempt to get DOFCredentials for any domain that comes in that the
* DOFServerConfig does not have credentials to already handle. The given DOFCredentials will be used to create new DOFCredentials in the new domain.
*
* @param self The DOFServerConfigBuilder to add the wildcard credentials to.  If @a self is NULL, no action will be taken.
* @param credentials The DOFCredentials to use as wildcard credentials. If NULL the previously set credentials will be removed.
*
* @return boolean TRUE The DOFCredentials were added to the config.
* @exception FALSE If the DOFCredentials failed to be added.
*/
extern DOF_OAL_DECLSPEC boolean DOFServerConfigBuilder_SetWildcardCredentials( const DOFServerConfigBuilder self, DOFCredentials credentials );

/**
 * Adds credentials to validate that an inbound connection established by the server may be to an authorized AS (one able to process authentication requests) for a domain.
 * If the validation for any credentials fails, the connection will still be allowed to open, but will not be considered a valid route for the domain associated with the failed credentials.
 *
 * @param self DOFServerConfigBuilder.  If @a self is NULL, no action will be taken.
 * @param domain The domain config to add.  This should not be NULL.
 *
 * @retval Non-Zero (TRUE) if the credentials were added successfully.
 * @exception Zero (FALSE) is returned if an invalid parameter was detected or insufficient resources were available to add the credentials.
 */
extern DOF_OAL_DECLSPEC boolean DOFServerConfigBuilder_AddDomain( const DOFServerConfigBuilder self, DOFDomainConfig domain );

/**
* Sets the domain wildcard credentials for the server. These will be passed on to incoming connections.
*
* @param self The DOFServerConfigBuilder to add the wildcard domain credentials to. If @a self is NULL, no action will be taken.
* @param credentials The DOFCredentials to use as wildcard domain credentials. If NULL the previously set credentials will be removed.
*
* @return boolean TRUE The DOFCredentials were added to the config.
* @exception FALSE If the DOFCredentials failed to be added.
*/
extern DOF_OAL_DECLSPEC boolean DOFServerConfigBuilder_SetDomainDiscoveryCredentials( const DOFServerConfigBuilder self, DOFCredentials credentials );

/**
 * Indicates that resulting inbound connections are a trusted route to an authorized AS (one able to process authentication requests) for a domain.
 * The domain is not validated by any credentials.
 * The server should be properly protected (e.g., by security and/or firewalls) to ensure that arbitrary inbound connections are not used as trusted domain routes.
 *
 * @param self DOFServerConfigBuilder.  If @a self is NULL, no action will be taken.
 * @param domainID The ID of the domain to be trusted on resulting inbound connections.  This should not be NULL.  This may be @ref DOFOBJECTIDDOMAIN_BROADCAST to indicate the connection is trusted for @b any domain.
 *
 * @retval Non-Zero (TRUE) if the domain route was added successfully.
 * @exception Zero (FALSE) is returned if an invalid parameter was detected or insufficient resources were available to add the domain route.
 */
extern DOF_OAL_DECLSPEC boolean DOFServerConfigBuilder_AddTrustedDomain( const DOFServerConfigBuilder self, DOFObjectIDDomain domainID );

/**
    Sets the @ref DOFProtocolNegotiator for the given DOFServerConfigBuilder.
    Please refer to @ref DOFServerConfigBuilder_Create for the default value for this setting.

    @param self The Server Config Builder.  If @a self is NULL, no action will be taken.
    @param negotiator The @ref DOFProtocolNegotiator.  This should not be NULL.

    @retval TRUE Succesfully set.
    @exception FALSE Failed to set.
**/
extern DOF_OAL_DECLSPEC boolean DOFServerConfigBuilder_SetProtocolNegotiator( DOFServerConfigBuilder self, DOFProtocolNegotiator negotiator );

/**
    Sets the maximum time allowed to be silent before a packet must be sent for the given DOFServerConfigBuilder.
    Please refer to @ref DOFServerConfigBuilder_Create for the default value for this setting.

    @param self The Server Config Builder.  If @a self is NULL, no action will be taken.
    @param maxSilence  The maximum silence time, in milliseconds.

    @retval TRUE Succesfully set.
    @exception FALSE Failed to set.
**/
extern DOF_OAL_DECLSPEC boolean DOFServerConfigBuilder_SetMaxSendSilence( DOFServerConfigBuilder self, uint32 maxSilence );

/**
    Sets the maximum time allowed to be silent before a packet must be received for the given DOFServerConfigBuilder.
    Please refer to @ref DOFServerConfigBuilder_Create for the default value for this setting.

    @param self The Server Config Builder.  If @a self is NULL, no action will be taken.
    @param maxSilence  The maximum silence time, in milliseconds.

    @retval TRUE Succesfully set.
    @exception FALSE Failed to set.
**/
extern DOF_OAL_DECLSPEC boolean DOFServerConfigBuilder_SetMaxReceiveSilence( DOFServerConfigBuilder self, uint32 maxSilence );

/**
 *  Sets the DOFOperationFilter this DOFServerConfigBuilder will use for inbound DOFOperaitons.
 *
 *  @param self The DOFServerConfigBuilder.  If @a self is NULL this function will do nothing.
 *  @param filter   the DOFOperationFilter to apply to received DOFOperations
 *
 *  @retval TRUE    the receive DOFOperationFilter was successfully set
 *  @retval FALSE   the receive DOFOperationFilter was not successfully set
 */
extern DOF_OAL_DECLSPEC boolean DOFServerConfigBuilder_SetReceiveFilter( DOFServerConfigBuilder self, DOFOperationFilter filter );

/**
 *  Sets the DOFOperationFilter this DOFServerConfigBuilder will use for outbound DOFOperaitons.
 *
 *  @param self The DOFServerConfigBuilder.  If @a self is NULL this function will do nothing.
 *  @param filter   the DOFOperationFilter to apply to sent DOFOperations
 *
 *  @retval TRUE    the send DOFOperationFilter was successfully set
 *  @retval FALSE   the send DOFOperationFilter was not successfully set
 */
extern DOF_OAL_DECLSPEC boolean DOFServerConfigBuilder_SetSendFilter( DOFServerConfigBuilder self, DOFOperationFilter filter );

/**
 *  Sets the DOFOperationBridgeConfig this DOFServerConfigBuilder will use for security mapping.
 *
 *  @param self The DOFServerConfigBuilder.  If @a self is NULL this function will do nothing.
 *  @param bridge   the DOFOperationBridgeConfig to use
 *
 *  @retval TRUE    the DOFOperationBridgeConfig was successfully set
 *  @retval FALSE   the DOFOperationBridgeConfig was not successfully set
 */
extern DOF_OAL_DECLSPEC boolean DOFServerConfigBuilder_SetBridge( DOFServerConfigBuilder self, DOFOperationBridgeConfig bridge );

/**
   Destroys the DOFServerConfigBuilder instance.

   @param self The Server Config Builder.  If @a self is NULL, no action will be taken.
 */
extern DOF_OAL_DECLSPEC void DOFServerConfigBuilder_Destroy( DOFServerConfigBuilder self );

/**
 * Builds an immutable @ref DOFServerConfig from this DOFServerConfigBuilder instance.
 * @mustfree{DOFServerConfig_Destroy}
 *
 * @param self The Server Config Builder.  This should not be NULL.
 * @return An immutable @ref DOFServerConfig.
 * @exception NULL is returned if the @ref DOFServerConfig cannot be created, or an invalid parameter was detected.
 */
extern DOF_OAL_DECLSPEC DOFServerConfig DOFServerConfigBuilder_Build( DOFServerConfigBuilder self );

/**
 * Builds an immutable @ref DOFServerConfig from this DOFServerConfigBuilder instance, then destroy this DOFServerConfigBuilder instance.
 * This is equivalent to calling @ref DOFServerConfigBuilder_Build(@a self) immediately followed by @ref DOFServerConfigBuilder_Destroy(@a self).
 * @mustfree{DOFServerConfig_Destroy}
 *
 * @param self The Server Config Builder.  This should not be NULL.
 * @return An immutable @ref DOFServerConfig.
 * @exception NULL is returned if the @ref DOFServerConfig cannot be created, or an invalid parameter was detected.  Regardless, the DOFServerConfigBuilder, if not NULL, is destroyed.
 */
extern DOF_OAL_DECLSPEC DOFServerConfig DOFServerConfigBuilder_BuildAndDestroy( DOFServerConfigBuilder self );

/** @} **/


/* These get functions return the reference to the internal copy.  They are NOT freed by the application. */

/**
   Gets the Server type for the given DOFServerConfig.

   @param self The Server Config.  This must not be NULL.

   @return The Server Type for the given Server Config.  This will be one of the @ref DOFServerType
           constants.
 */
extern DOF_OAL_DECLSPEC DOFServerType DOFServerConfig_GetServerType(const DOFServerConfig self );

/**
   Gets the @ref DOFAddress for the given DOFServerConfig.

   @note This function returns the reference to the internal copy and should not be freed by the application.

   @param self The Server Config.  This must not be NULL.

   @return The Address for the given Server Config.
 */
extern DOF_OAL_DECLSPEC DOFAddress DOFServerConfig_GetAddress( const DOFServerConfig self );

/**
   Gets the DOFSecurityDesire from the given DOFServerConfig.

   @param self The Server Config.  This must not be NULL.

   @return The Security Desire for the given Server Config.
 */
extern DOF_OAL_DECLSPEC DOFSecurityDesire DOFServerConfig_GetSecurityDesire( const DOFServerConfig self );

/**
   Gets the @ref DTLServerConfig for the given DOFServerConfig.

   @note This function returns the reference to the internal copy and should not be freed by the application.

   @param self The Server Config.  This must not be NULL.

   @return The DTLServerConfig for the given Server Config.
 */
extern DOF_OAL_DECLSPEC DTLServerConfig DOFServerConfig_GetTransportConfig( const DOFServerConfig self );

/**
 * Gets the wildcard credentials set in the DOFServerConfig.
 *
 * @param self The server config. This must not be NULL.
 *
 * @return DOFCredentials The wildcard DOFCredentials for the given Server Config.
 */
extern DOF_OAL_DECLSPEC DOFCredentials DOFServerConfig_GetWildcardCredentials( const DOFServerConfig self );

/**
 * Gets the domain wildcard credentials set in the DOFServerConfig.
 *
 * @param self The server config. This must not be NULL.
 *
 * @return DOFCredentials The domain wildcard DOFCredentials for the given Server Config.
 */
extern DOF_OAL_DECLSPEC DOFCredentials DOFServerConfig_GetDomainDiscoveryCredentials( const DOFServerConfig self );

/**
   Gets the current DOFProtocolNegotiator for the DOFServerConfig.

   @param self The DOFServerConfig.  This must not be NULL.
   @return The current negotiator being used.
**/
extern DOF_OAL_DECLSPEC DOFProtocolNegotiator DOFServerConfig_GetProtocolNegotiator( const DOFServerConfig self );

/**
   Gets the current maximum send silence time in this DOFServerConfig.

   @param self The DOFServerConfig.  This must not be NULL.
   @return The current maximum send silence time, in milliseconds.
**/
extern DOF_OAL_DECLSPEC uint32 DOFServerConfig_GetMaxSendSilence( const DOFServerConfig self );

/**
   Gets the current maximum receive silence time in this DOFServerConfig.

   @param self The DOFServerConfig.  This must not be NULL.
   @return The current maximum receive silence time, in milliseconds.
**/
extern DOF_OAL_DECLSPEC uint32 DOFServerConfig_GetMaxReceiveSilence( const DOFServerConfig self );

/**
   Gets the maximum number of inbound connections in the DOFServerConfig.

   @deprecated This functionality was replaced by @ref DOFServerConfig_GetConnectionLimit.

   @param self The DOFServerConfig.  This must not be NULL.
   @return The maximum number of inbound connections allowed by the server.  0 indicates no maximum (unlimited).
 */
extern DOF_OAL_DECLSPEC uint16 DOFServerConfig_GetMaxConnections( const DOFServerConfig self );

/**
   Gets the maximum number of inbound connections in the DOFServerConfig.

   @param self The DOFServerConfig.  This must not be NULL.
   @return The maximum number of inbound connections allowed by the server.
 */
extern DOF_OAL_DECLSPEC uint32 DOFServerConfig_GetConnectionLimit( const DOFServerConfig self );

/**
 *  Retrieves the DOFOperationFilter applied to inbound DOFOperations.
 *
 *  @param self The DOFServerConfig to check.  This should not be @c NULL.
 *
 *  @return the DOFOperationFilter applied to inbound DOFOperations
 */
extern DOF_OAL_DECLSPEC DOFOperationFilter DOFServerConfig_GetReceiveFilter( DOFServerConfig const self );

/**
 *  Retrieves the DOFOperationFilter applied to outbound DOFOperations.
 *
 *  @param self The DOFServerConfig to check.  This should not be @c NULL.
 *
 *  @return the DOFOperationFilter applied to outbound DOFOperations
 */
extern DOF_OAL_DECLSPEC DOFOperationFilter DOFServerConfig_GetSendFilter( DOFServerConfig const self );

/**
 *  Retrieves the DOFOperationBridgeConfig used by a DOFServerConfig.
 *
 *  @param self The DOFServerConfig to check.  This should not be @c NULL.
 *
 *  @return the DOFOperationBridgeConfig
 */
extern DOF_OAL_DECLSPEC DOFOperationBridgeConfig DOFServerConfig_GetBridge( DOFServerConfig const self );

/**
   Destroys a DOFServerConfig instance.

   @param self The Server Config.  If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFServerConfig_Destroy( DOFServerConfig self );


/**
@}
**/


/** @defgroup DOFServer DOFServer
    @ingroup DOF_Framework

    Represents a DOF node server.

@{
**/


/*
**  "Public" Type and Method Declarations
*/

/* DOFServer Methods */

/**
 * Creates a server based on server passed in
 *
 * @param self the DOFServer to use to create server from. Must not be NULL!
 * @param config the DOFServerConfig for the server to create. Must not be NULL;
 *
 * @return DOFServer the Newly created DOFServer, or NULL if server could not be created
 */
extern DOF_OAL_DECLSPEC DOFServer DOFServer_CreateServer( DOFServer self, DOFServerConfig config );

/**
 *  Stops and destroys a DOFServer. No further inbound connections can be made through the server, and 
 *  existing inbound connections will be disconnected. The reference count for the server is set to zero.
 * 
 *  @note If another DOFServer object has been allocated with an identical @ref DOFServerConfig, and it 
 *  is still running, existing connections will be unaffected and inbound connections can still be 
 *  established through the identical server. Inbound connections will be affected only if this is 
 *  the only currently running server with an identical @ref DOFServerConfig.
 *
 *  @param  self The server to destroy.  If NULL, no action is taken.
 *
 *  @see    DOF_CreateServer
 *  @see    DOFServer_CreateServer
 *  @see    DOFServer_Stop
 *  @see    DOF_Stop
 */
extern DOF_OAL_DECLSPEC void DOFServer_Destroy( DOFServer self );

/**
 * Causes the Server to become Started.
 *
 * @note This call is synchronous, and waits for the Server to be established (or fail) before returning control to
 *       the caller.
 *
 * @param self The Server.  This must not be NULL.
 * @param timeout The maximum time to wait for the connection to
 *                be established (in milliseconds).
 * @exceptionparam
 *
 * @retval TRUE If the server started.
 * @exception FALSE If the server failed to start. If an exception was used it will give more detail of failure.
 *
 * @see DOF_CreateServer
 * @see DOFServer_CreateServer
 */
extern DOF_OAL_DECLSPEC boolean DOFServer_Start( DOFServer self, uint32 timeout, DOFException * pException );

/**
 * Causes the Server to become Started.
 *
 * The callback is called to indicate that the server either started
 * successfully or failed.  Any listeners added to the server will be
 * notified if this causes the state to change, but are otherwise unrelated to
 * the callback specified here.
 *
 * @note This call is asynchronous, and does not wait for the Server to be established before returning control to
 * the caller.
 *
 * @param self The Server.  This must not be NULL.
 * @param timeout The maximum time to wait for the Server to be established.
 * @param callback The Server completion callback to be called when asynchronous operation finishes.
 * @param context The context, if any, associated with this connection request.
 *
 * @see DOFServer_AddStateListener
 */
extern DOF_OAL_DECLSPEC void DOFServer_BeginStart( DOFServer self, uint32 timeout, DOFServerStartCallback callback, void * context );

/**
 * Causes the Server to become Started again, if possible.  If the Server is already started, then this call is satisfied immediately.
 * This call should only be used when a successfully started server was terminated but the server is still desired (@ref DOFServer_Stop has not yet been called).
 *
 * @note This call is synchronous, and waits for the Server to be established (or fail) before returning control to
 *       the caller.
 *
 * @param self The Server.  This must not be NULL.
 * @param timeout The maximum time to wait for the connection to
 *                be established (in milliseconds).
 * @exceptionparam
 *
 * @retval TRUE If the server restarted.
 * @exception FALSE If the server failed to restart. If an exception was used it will give more detail of failure.
 */
extern DOF_OAL_DECLSPEC boolean DOFServer_Restart( DOFServer self, uint32 timeout, DOFException * pException );

/**
 * Causes the Server to start again, if possible.  If the Server is already started, then this call is satisfied immediately.
 * This call should only be used when a successfully started server was terminated but the server is still desired (@ref DOFServer_Stop has not yet been called).
 *
 * The callback is called to indicate that the server either started
 * successfully or failed.  Any listeners added to the server will be
 * notified if this causes the state to change, but are otherwise unrelated to
 * the callback specified here.
 *
 * @note This call is asynchronous, and does not wait for the Server to be established before returning control to
 * the caller.
 *
 * @param self The Server.  This must not be NULL.
 * @param timeout The maximum time to wait for the Server to be established.
 * @param callback The Server completion callback to be called when asynchronous operation finishes.
 * @param context The context, if any, associated with this connection request.
 *
 * @see DOFServer_AddStateListener
 */
extern DOF_OAL_DECLSPEC void DOFServer_BeginRestart( DOFServer self, uint32 timeout, DOFServerStartCallback callback, void * context );

/**
 *  Stops a server and disconnects existing inbound connections to the server. If multiple calls were made to 
 *  @ref DOFServer_Start and @ref DOFServer_BeginStart for this server, this call merely decrements the reference count 
 *  without stopping the server until the reference count reaches zero. If the server is not currently started, 
 *  this call has no effect.
 *  
 *  @note If another DOFServer object has been allocated with an identical @ref DOFServerConfig, and it 
 *  is still running, the server will not actually stop and inbound connections are unaffected, even if the reference 
 *  count has reached zero. The server actually stops only when the reference reaches zero and this is the only 
 *  currently running server with an identical @ref DOFServerConfig.
 *
 *  @note If any outbound DOFConnections use the DOFServer as a related server, the server may not stop until the
 *  connections using the server are disconnected.
 *
 *  @param  self  The server to stop. This must not be NULL.
 *
 *  @see DOFServer_Destroy
 *  @see DOF_Stop
 */
extern DOF_OAL_DECLSPEC void DOFServer_Stop( DOFServer self );

/**
 * Returns the given DOFServer's DOFServerConfig. 
 * The returned DOFServerConfig is valid only while the server is available. 
 *  
 * @deprecated This function was deprecated in version 6.1 and should not be used. Its functionality was 
 * replaced by @ref DOFServer_CreateState, which should be used instead.
 *  
 * @param self The server. This must not be NULL.
 * @return The DOFServerConfig for the given server, or NULL if not available.
 */
extern DOF_OAL_DECLSPEC DOFServerConfig DOFServer_GetConfig( const DOFServer self );

/**
 * Creates a connection associated with a server, including the ability to join
 * multicast groups.  Multicast groups require an associated DOFServer, so this
 * is the only method allowed to create those types of connections.  Other
 * types of connections may benefit by having an associated DOFServer, as they
 * may share the server's resources to send/receive data.
 *
 * @mustfree{DOFConnection_Destroy}
 *
 * @param self The server.  This parameter must not be NULL.
 * @param config The configuration options of the connection.  This should not be NULL.
 *
 * @return The new DOFConnection, or NULL if one could not be created.
 */
extern DOF_OAL_DECLSPEC DOFConnection DOFServer_CreateConnection( DOFServer self, DOFConnectionConfig config );

/**
 * Adds a listener to DOFServer that notify user of state change of a server.
 *
 * @note a listenter will only be added to a server one time. Duplicate listener will fail to be added.
 *
 * @param self the server to add the listener to.  This must not be NULL.
 * @param listener the users listener to call with state change.  This should not be NULL.
 *
 * @return TRUE if Listener was added, FALSE otherwise.
 *
 * @see DOFServer_RemoveStateListener
 */
extern DOF_OAL_DECLSPEC boolean DOFServer_AddStateListener( DOFServer self, DOFServerStateListener listener );

/**
 * Removes listener from a DOFServer.
 *
 * @param self the server to remove a listener from.  This must not be NULL.
 * @param listener the users listener to remove.  This should not be NULL.  If not found, no action is taken.
 *
 * @see DOFServer_AddStateListener
 */
extern DOF_OAL_DECLSPEC void DOFServer_RemoveStateListener( DOFServer self, DOFServerStateListener listener);

/**
   Returns a snapshot of the current state of the Server.

   @mustfree{DOFServerState_Destroy}

   @note The state of the Server is dynamic and may change at any moment, due to a number of reasons.  It is usually
         best to track the state of a Server through a listener.

   @param self The Server.  This must not be NULL.

   @return The snapshow of the current state of the Server.

   @see DOFServer_AddStateListener
 */
extern DOF_OAL_DECLSPEC DOFServerState DOFServer_CreateState( const DOFServer self );

/**
   Returns whether the status of the Server is connected or not.

   @param self The Server.  This must not be NULL.
   @retval Zero (@c FALSE) if the Server is not connected.
   @retval Non-zero (@c TRUE) if the Server is connected.

   @see DOFServerState_IsStarted
 */
extern DOF_OAL_DECLSPEC boolean DOFServer_IsStarted( const DOFServer self );

/** @defgroup DOFServerState DOFServerState

    Represents the instantaneous state of a DOFServer.

@{
**/

/**
   Returns whether the status of the Server is connected or not.

   @param self The Server State.  This must not be NULL.
   @retval Zero (@c FALSE) if the Server is not connected.
   @retval Non-zero (@c TRUE) if the Server is connected.

   @see DOFServer_CreateState
 */
extern DOF_OAL_DECLSPEC boolean DOFServerState_IsStarted( const DOFServerState self );

/**
 * Returns an Exception associated with state. if one is available.
 *
 * @param self The Server State. This must not be NULL.
 *
 * @return DOFException The exeption if one is available, NULL otherwise.
 */
extern DOF_OAL_DECLSPEC DOFException DOFServerState_GetException( const DOFServerState self );

/**
 * Returns the number of bytes received on the associated server at the point
 * the server state was created.
 *
 * @param self The Server State. This must not be NULL.
 *
 * @return uint64 The number of bytes received on the associated server.
 */
extern DOF_OAL_DECLSPEC uint64 DOFServerState_GetReceiveByteCount( DOFServerState self );

/**
 * Returns the number of bytes sent from the associated server at the point
 * the server state was created.
 *
 * @param self The Server State. This must not be NULL.
 *
 * @return uint64 The number of bytes sent from associated server.
 */
extern DOF_OAL_DECLSPEC uint64 DOFServerState_GetSendByteCount( DOFServerState self );

/**
 * Returns the number of packets received on the associated server at the point
 * the server state was created.
 *
 * @param self The Server State. This must not be NULL.
 *
 * @return uint64 The number packet received on the associated server.
 */
extern DOF_OAL_DECLSPEC uint64 DOFServerState_GetReceivePacketCount( DOFServerState self );

/**
 * Returns the number of packets sent from the associated server at the point
 * the server state was created.
 *
 * @param self The Server State. This must not be NULL.
 *
 * @return uint64 The number of packets sent from associated server.
 */
extern DOF_OAL_DECLSPEC uint64 DOFServerState_GetSendPacketCount( DOFServerState self );

/**
 * Retruns the server type.
 *
 * @param self The Server State. This must not be NULL.
 *
 * @return DOFServerType The server type. @ref DOFServerType
 */
extern DOF_OAL_DECLSPEC DOFServerType DOFServerState_GetServerType( DOFServerState self );

/**
 * Returns the address.
 *
 * @param self The Server State. This must not be NULL.
 *
 * @return DOFAddress The address.
 */
extern DOF_OAL_DECLSPEC DOFAddress DOFServerState_GetAddress( DOFServerState self );

/**
 * Returns the security desire.
 *
 * @param self The Server State. This must not be NULL.
 *
 * @return DOFSecurityDesire The security desire.
 */
extern DOF_OAL_DECLSPEC DOFSecurityDesire DOFServerState_GetSecurityDesire( DOFServerState self );

/**
 * Returns the transport config.
 *
 * @param self The Server State. This must not be NULL.
 *
 * @return DTLServerConfig The Transport config.
 */
extern DOF_OAL_DECLSPEC DTLServerConfig DOFServerState_GetTransportConfig( DOFServerState self );

/**
 * Returns the protocol negotiator.
 *
 * @param self The Server State. This must not be NULL.
 *
 * @return DOFProtocolNegotiator Then protocol negotiator.
 */
extern DOF_OAL_DECLSPEC DOFProtocolNegotiator DOFServerState_GetProtocolNegotiator( DOFServerState self );

/**
 * Returns the max send silence.
 *
 * @param self The Server State. This must not be NULL.
 *
 * @return uint32 The max send silence.
 */
extern DOF_OAL_DECLSPEC uint32 DOFServerState_GetMaxSendSilence( DOFServerState self );

/**
 * Returns the max receive silence.
 *
 * @param self The Server State. This must not be NULL.
 *
 * @return uint32 The max receive silence.
 */
extern DOF_OAL_DECLSPEC uint32 DOFServerState_GetMaxReceiveSilence( DOFServerState self );

/**
 * Returns the maximum number of inbound connections the DOFServer allows.
 *
 * @deprecated This functionality was replaced by @ref DOFServerState_GetConnectionLimit.
 *
 * @param self The Server State. This must not be NULL.
 *
 * @return uint16 The max connections
 */
extern DOF_OAL_DECLSPEC uint16 DOFServerState_GetMaxConnections( DOFServerState self );

/**
 * Returns the maximum number of inbound connections the DOFServer allows.
 *
 * @param self The Server State. This must not be NULL.
 *
 * @return The maximum number of inbound connections allowed by the server.
 */
extern DOF_OAL_DECLSPEC uint32 DOFServerState_GetConnectionLimit( DOFServerState self );


/**
 * Deallocates the object representing the server's state.
 *
 * @param self The Server State.  If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFServerState_Destroy( DOFServerState self );

/**
@}
**/


/**
@}
**/


/** @defgroup DOFServerList DOFServerList
    @ingroup DOFServer

@{
**/

/**
 * Releases resources previously allocated while forming a DOFServerList.
 * @param  serverList  The DOFServerList.  If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFServerList_Destroy( DOFServerList serverList );

/**
@}
**/


#ifdef  __cplusplus
}
#endif

#endif /* DOF_SERVER_H */
