/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_SYSTEM_H
#define DOF_SYSTEM_H


#include <dof/core.h>
#include <dof/control.h>


#ifdef  __cplusplus
extern "C" {
#endif


/** @defgroup DOFSystem DOFSystem
    @ingroup DOF_Framework

    Represents a view into the DOF network.

@{
**/

/*
**  "Public" Method Declarations
*/

/**
    Express interest for an Object ID and an Interface IDs. This is a request
    only; the callback is called only when the request is complete (timed out or
    cancelled) and does not indicate a response.

    @mustfree{DOFOperation_Destroy}

    @param  sys         The system in which the Interest request is to be made.
                        This must not be NULL.
    @param  objectID    The object ID on which to express interest.  If @a level is @ref DOFINTERESTLEVEL_WATCH, then
                        this may be @ref DOFOBJECTID_BROADCAST to express interest on all object IDs (other levels do
                        not support broadcast). This should not be NULL.
    @param interfaceID  The interface ID on which to express interest.  If @a level is @ref DOFINTERESTLEVEL_WATCH
                        or @ref DOFINTERESTLEVEL_ACTIVATE, then this may be @ref DOFINTERFACEID_WILDCARD to express
                        interest on all interface IDs (other levels do not support wildcard). This should not be NULL.
    @param  level       The level of interest to request.  This should be a
                        level from @ref DOFInterestLevel.
    @param  config      Configuration options for the request operation (NULL for default configuration).
    @param  timeout     The duration for the asynchronous operation, in milliseconds. The timeout begins after the
                        operation has been processed and sent to appropriate connections.
    @param  callback    Callback when operation completes.  This may be NULL to ignore completion notification.
    @param  context     Application context data associated with the asynchronous operation.

    @return A representation of the operation, or NULL if
            - an invalid parameter is detected; or
            - insufficient resources exist to create the operation.
**/
extern DOF_OAL_DECLSPEC DOFOperation DOFSystem_BeginInterest( const DOFSystem sys, const DOFObjectID objectID, const DOFInterfaceID interfaceID, DOFInterestLevel level, DOFOperationControl config, uint32 timeout, const DOFSystemInterestCallback callback, void * context );

/**
    Query the system for a list of objects available within the system.
    The set of objects returned is scoped based on the given query filter.

    This does not block, as it is a simple query against the current state of
    the system.  There is no request to the network to identify additional
    objects that may be available.  The Object IDs returned may be limited by
    the Interest operations that exist (see @ref DOFSystem_BeginInterest).

    @mustfree{DOFObjectIDList_Destroy}

    @param  sys         The System in which to make the query.  This must not be NULL.
    @param  query       The set of Objects and/or Interfaces for the query (or @c NULL to query for everything).

    @return A new list of Object IDs. This list can have 0 members.

    @see    DOFSystem_BeginInterest
**/
extern DOF_OAL_DECLSPEC DOFObjectIDList DOFSystem_Query( const DOFSystem sys, const DOFQuery query );

/**
    Register a set of callbacks for being notified of changes in availability
    of objects within the system.  The set of objects reported is scoped
    based on the given query filter.

    For the duration of the @ref DOFOperation, the Query callbacks will be
    used to notify the application as objects matching the query are added or
    removed from the system.  As with any @ref DOFOperation, this Operation
    can be extended or cancelled.

    There is no request to the network to identify additional objects that may
    be available.  The Object IDs returned may be limited by the Interest
    operations that exist (see @ref DOFSystem_BeginInterest).

    @mustfree{DOFOperation_Destroy}

    @param  sys         The System on which to establish the query.  This must not be NULL.
    @param  query       The set of Objects and/or Interfaces to watch (or @c NULL to watch everything).
    @param  timeout     The duration for the Query operation, in milliseconds.
    @param  callback    The callback to use while the Query is in effect.  This
                        may be NULL to ignore all notification of matching
                        providers, but the operation would thus have no effect.
    @param  context     Application context data associated with the operation.

    @return A representation of the operation, or NULL if
            - an invalid parameter is detected; or
            - insufficient resources exist to create the operation.

    @see    DOFSystem_BeginInterest
**/
extern DOF_OAL_DECLSPEC DOFOperation DOFSystem_BeginQuery( const DOFSystem sys, const DOFQuery query, uint32 timeout, const DOFSystemQueryCallback callback, void * context );

/**
    Create a new @ref DOFObject in the system for use by the application.

    @mustfree{DOFObject_Destroy}

    @param  sys         The DOFSystem in which to create the new @ref DOFObject.  This must not be NULL.
    @param  objectID    The Object ID for the Object to be created. This should not be NULL.
    @return A new @ref DOFObject.
    @exception NULL is returned if an invalid argument is detected, or the Object could not be created.
**/
extern DOF_OAL_DECLSPEC DOFObject DOFSystem_CreateObject( DOFSystem sys, const DOFObjectID objectID );

/**
    Query the system for the list of Interfaces available on the given Object.

    This does not block, as it is a simple query against the current state of
    the system.  There is no request to the network to identify additional
    Interfaces that may be available. The Interface IDs returned may be limited
    to those discovered using current interest operations. (see @ref DOFSystem_BeginInterest).

    @mustfree{DOFInterfaceIDList_Destroy}

    @param  sys         The System in which to make the query.  This must not be NULL.
    @param  objectID    The Object ID for the query.  This must not be NULL.

    @return A new list of Interface IDs. This list can have 0 members.

    @exception NULL is returned if the list could not be allocated.
**/
extern DOF_OAL_DECLSPEC DOFInterfaceIDList DOFSystem_GetProviderInterfaces( const DOFSystem sys, const DOFObjectID objectID );

/**
    Adds a @ref DOFSystemActivateInterestListener to the given DOFSystem, which is used by provider applications to begin
    providing in response to @ref DOFINTERESTLEVEL_ACTIVATE interest requests.

    @param  self        The System on which to set the activation listener.  This must not be NULL.
    @param  listener    The listener to use for notification of activation requests.  This should not be NULL.

    @return TRUE if the listener was successfully added.
    @exception FALSE if the listener could not be added.
**/
extern DOF_OAL_DECLSPEC boolean DOFSystem_AddActivateInterestListener( DOFSystem self, const DOFSystemActivateInterestListener listener );

/**
    Remove a listener for when interest changes in the network.  Changes in
    interest of Objects and Interfaces will no longer be reported through the
    given listener.  If the given @a listener cannot be found, this call will
    silently ignore the request.

    @param  self        The System on which to clear the activation listener.  This must not be NULL.
    @param  listener    The listener to be removed from the list. This should not be NULL. If the listener cannot be
                        found, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFSystem_RemoveActivateInterestListener( DOFSystem self, const DOFSystemActivateInterestListener listener );

/**
    Determine if there is interest currently expressed for a given Object and Interface.
    If @a interfaceID is the Wildcard IID, then check for @b all interfaces on the given Object.
    If @a interfaceID is NULL, then check for @b any interfaces on the given Object.

    @param self        The System. This must not be NULL.
    @param objectID    The Object ID. This should not be NULL. The Broadcast ID only matches interest where the
                       Broadcast ID was used.
    @param interfaceID The Interface ID. This may be NULL to indicate any interface ID. The Wildcard ID only matches
                       interest where the Wildcard ID was used.
    @param level       The level of interest for which to check. Current interest at or above the specified level will
                       return an affirmative result.

    @return FALSE if the specified level of interest is not present for the given Object and Interface.
            TRUE if the specified level of interest is present for the given Object and Interface.
**/
extern DOF_OAL_DECLSPEC boolean DOFSystem_IsInterestRequested( const DOFSystem self, const DOFObjectID objectID, const DOFInterfaceID interfaceID, DOFInterestLevel level );

/**
    Wait for a provider providing at least the given set of interfaces to become available in the system, and return a
    new, valid @ref DOFObject for it.  The Broadcast ID may be used as a wildcard (to return the first object found).

    The matching logic uses the same logic as DOFObjectID_MatchesAtLeast, so additional attributes may be present on the
    resulting Object (e.g., an activate service may add a Provider attribute). If there are multiple matches, the first
    one found will be returned.  This is a simplified form of
    @ref DOFSystem_Query or
    @ref DOFSystem_BeginQuery, followed by @ref DOFSystem_CreateObject on the
    first match.

    There is no request to the network to identify additional objects that may
    be available.  The @ref DOFObject returned may be limited by the Interest
    operations that exist (see @ref DOFSystem_BeginInterest).

    If the Object is not available within the timeout period, the function will
    return @c NULL.

    @mustfree{DOFObject_Destroy}

    @param  sys          The System.  This must not be NULL.
    @param  objectID     The Object ID on which to wait.  This should not be NULL, and should not represent a multicast
                         Object ID. It may be the Broadcast ID to return the first object found.
    @param  interfaceIDCount  The number of Interfaces to wait for on the Object.  This should be greater than 0.
    @param  interfaceIDs The list of Interfaces to wait for on the Object.  This should not be NULL, and should not
                         contain a NULL or the Wildcard ID.
    @param  timeout      The duration to wait for the Object and Interfaces to be available, in milliseconds.

    @return A new @ref DOFObject representing the desired Object.
    @exception NULL is returned if an invalid parameter was detected, or the Object was not available within the given
                    timeout.

    @see    DOFSystem_BeginInterest
**/
extern DOF_OAL_DECLSPEC DOFObject DOFSystem_WaitProvider( const DOFSystem sys, const DOFObjectID objectID, uint32 interfaceIDCount, const DOFInterfaceID interfaceIDs[], uint32 timeout );

/**
 * Return the current route information for the best, compatible provider (if any) given an Object ID and Interface ID.
 *
 * @mustfree{DOFRouteInfo_Destroy}
 *
 * @param  sys          The System.  This must not be NULL.
 * @param  objectID     The Object ID of the desired provider. This should not be NULL, and should not be the Broadcast
 *                      or a multicast Object ID.
 * @param  interfaceID  The Interfaces ID of the desired provider. This should not be NULL, and should not be the
 *                      Wildcard Interface ID.
 *
 * @return The @ref DOFRouteInfo for the best compatible provider, or NULL if a match could not be found.
 * @exception NULL is returned if an invalid parameter is detected, or insufficient resources are available to create
 *                 the route information.
 */
extern DOF_OAL_DECLSPEC DOFRouteInfo DOFSystem_CreateProviderRoute( const DOFSystem sys, const DOFObjectID objectID, const DOFInterfaceID interfaceID );

/**
 * Add a listener to a system for when a route to a Provider available to the system is added, removed, or changed.
 * A listener should only be added once.
 *
 * @param self The DOFSystem.  This must not be NULL.
 * @param listener The listener.  This should not be NULL.
 *
 * @return TRUE if the listener was added.
 * @exception FALSE if the listener could not be added, illegal arguments were detected, or the listener has already
 *                  been addded.
 */
extern DOF_OAL_DECLSPEC boolean DOFSystem_AddRouteListener( DOFSystem self, DOFSystemRouteListener listener );

/**
 * Remove the listener on a system added by @ref DOFSystem_AddRouteListener.
 * If the listener was not added (or has already been removed), the request is ignored.
 *
 * @param self The DOFSystem.  This must not be NULL.
 * @param listener The listener.  This should not be NULL.  If not found, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFSystem_RemoveRouteListener( DOFSystem self, DOFSystemRouteListener listener );


/**
    Destroy a DOFSystem.

    This has no effect on any outstanding operations started from the
    DOFSystem, or any DOFObjects that may still exist on the DOFSystem.
    The operations will continue to be processed normally, and the
    DOFObjects will remain valid.  While the application cannot reference
    the DOFSystem directly any longer, the DOFSystem available to an
    operation's application callbacks and to DOFObjects created from the
    DOFSystem will remain valid.

    @param sys The DOFSystem to destroy.  If NULL, no action is taken.

    @see DOF_CreateSystem
**/
extern DOF_OAL_DECLSPEC void DOFSystem_Destroy( DOFSystem sys );

/**
 *  Construct a DOFSystemState.
 *
 *  @mustfree{DOFSystemState_Destroy}
 *
 *  @param self The DOFSystem who's state should be constructed.  Should not be @c NULL.
 *
 *  @return the state of @a self
 */
extern DOF_OAL_DECLSPEC DOFSystemState DOFSystem_CreateState( DOFSystem self );

/**
 * Release resources of a DOFSystemList.
 * @param  list  The DOFSystemList.  If NULL, no action is taken.
 */
extern DOF_OAL_DECLSPEC void DOFSystemList_Destroy( DOFSystemList list );

/**
@}
**/

/** @defgroup DOFSystemConfig DOFSystemConfig
    @ingroup DOFSystem

    Represents the static configuration of a DOFSystem.
@{
**/

/** @name Builder
@{
**/

/**
 *  Construct a DOFSystemConfigBuilder.  The DOFSystemConfigBuilder is created with the following defaults:
 *    - credentails: NULL
 *    - identity: NULL
 *    - receive filter: NULL
 *    - send filter: NULL
 *    - default permissions: NULL
 *    - permission extension: TRUE
 *    - threadpool size: 0
 *
 *  @mustfree{DOFSystemConfigBuilder_Destroy or @ref DOFSystemConfigBuilder_BuildAndDestroy}
 *
 *  @return a DOFSystemConfigBuilder
 *
 *  @retval NULL    Unable to aquire the necessary resources to create the DOFSystemConfigBuilder
 */
extern DOF_OAL_DECLSPEC DOFSystemConfigBuilder DOFSystemConfigBuilder_Create( void );

/**
 *  Create a DOFSystemConfigBuilder using a DOFSystemConfig as a template.
 *
 *  @mustfree{DOFSystemConfigBuilder_Destroy or @ref DOFSystemConfigBuilder_BuildAndDestroy}
 *
 *  @param self The DOFSystemConfig to use as a template.  Should not be @c NULL.
 *
 *  @return a DOFSystemConfigBuilder capable of creating a DOFSystemConfig equilivlent to @a self
 *
 *  @retval NULL    @a self is @c NULL or unable to aquire the necessary resources to create the DOFSystemConfigBuilder
 */
extern DOF_OAL_DECLSPEC DOFSystemConfigBuilder DOFSystemConfigBuilder_Create_Config( DOFSystemConfig self );

/**
 *  Create a DOFSystemConfigBuilder using a DOFSystemState as a template.
 *
 *  @mustfree{DOFSystemConfigBuilder_Destroy or @ref DOFSystemConfigBuilder_BuildAndDestroy}
 *
 *  @param self The DOFSystemState to use as a template.  Should not be @c NULL.
 *
 *  @return a DOFSystemConfigBuilder capable of creating a DOFSystemConfig equilivlent to @a self
 *
 *  @retval NULL    @a self is @c NULL or unable to aquire the necessary resources to create the DOFSystemConfigBuilder
 */
extern DOF_OAL_DECLSPEC DOFSystemConfigBuilder DOFSystemConfigBuilder_Create_State( DOFSystemState self );

/**
 *  Set the DOFCredentials that will be used for a DOFSystemConfig.
 *
 *  @param self The DOFSystemConfigBuilder.  Should not be @c NULL.
 *  @param credentials  The DOFCredentials to use.  If @c NULL the DOFSystemConfig will be non-secure.
 *
 *  @retval TRUE    the credentials were set
 *  @retval FALSE   the credentials were not set
 */
extern DOF_OAL_DECLSPEC boolean DOFSystemConfigBuilder_SetCredentials( DOFSystemConfigBuilder self, DOFCredentials credentials );

/**
 *  Set the DOFFilter that will be applied to all inbound DOFOperations.
 *
 *  @param self The DOFSystemConfigBuilder.  Should not be @c NULL.
 *  @param filter   The DOFFilter to use.  If @c NULL no inbound filtering will take place.
 *
 *  @retval TRUE    the receive filter was set
 *  @retval FALSE   the receive filter was not set
 */
extern DOF_OAL_DECLSPEC boolean DOFSystemConfigBuilder_SetReceiveFilter( DOFSystemConfigBuilder self, DOFOperationFilter filter );

/**
 *  Set the DOFFilter that will be applied to all outbound DOFOperations.
 *
 *  @param self The DOFSystemConfigBuilder.  Should not be @c NULL.
 *  @param filter   The DOFFilter to use.  If @c NULL no outbound filtering will take place.
 *
 *  @retval TRUE    the send filter was set
 *  @retval FALSE   the send filter was not set
 */
extern DOF_OAL_DECLSPEC boolean DOFSystemConfigBuilder_SetSendFilter( DOFSystemConfigBuilder self, DOFOperationFilter filter );

/**
 *  Set the DOFPermissionSet that will be initially requested.
 *
 *  @param self The DOFSystemConfigBuilder.  Should not be @c NULL.
 *  @param permissions  The DOFPermissionSet to request.  If @c NULL the initial DOFPermissionSet will be empty.
 *
 *  @retval TRUE    the initial permissions were set
 *  @retval FALSE   the initial permissions were not set
 */
extern DOF_OAL_DECLSPEC boolean DOFSystemConfigBuilder_SetPermissions( DOFSystemConfigBuilder self, DOFPermissionSet permissions );

/**
 *  Control whether a DOFSystem will be able to extend permissions if desired.
 *
 *  @param self The DOFSystemConfigBuilder.  Should not be @c NULL.
 *  @param extendAllowed    control whether permissions can be extended
 *
 *  @retval TRUE    the extend option was successfully set to @a extendAllowed
 *  @retval FALSE   the extend option was not successfully set to @a extendAllowed
 */
extern DOF_OAL_DECLSPEC boolean DOFSystemConfigBuilder_SetPermissionsExtendAllowed( DOFSystemConfigBuilder self, boolean extendAllowed );

/**
 *  Set the maximum number of threads used by a DOFSystem (although a DOFSystem is free to use fewer if it chooses to).  A value of 0 will use the default thread pool.
 *
 *  @param self The DOFSystemConfigBuilder.  Should not be @c NULL.
 *  @param maxSize  the maximum number of threads to use
 *
 *  @retval TRUE    the maximum thread count was set
 *  @retval FALSE   the maximum thread count was not set
 */
extern DOF_OAL_DECLSPEC boolean DOFSystemConfigBuilder_SetThreadpoolSize( DOFSystemConfigBuilder self, uint16 maxSize );

/**
 *  Construct a DOFSystemConfig from a DOFSystemConfigBuilder and destroy the DOFSystemConfigBuilder.
 *
 *  @mustfree{DOFSystemConfig_Destroy}
 *
 *  @param self The DOFSystemConfiBuilder.  Should not be @c NULL.
 *
 *  @return a DOFSystemConfig as configured by @a self
 */
extern DOF_OAL_DECLSPEC DOFSystemConfig DOFSystemConfigBuilder_BuildAndDestroy( DOFSystemConfigBuilder self );

/**
 *  Construct a DOFSystemConfig from a DOFSystemConfigBuilder.
 *
 *  @mustfree{DOFSystemConfig_Destroy}
 *
 *  @param self The DOFSystemConfiBuilder.  Should not be @c NULL.
 *
 *  @return a DOFSystemConfig as configured by @a self
 */
extern DOF_OAL_DECLSPEC DOFSystemConfig DOFSystemConfigBuilder_Build( DOFSystemConfigBuilder self );

/**
 *  Perform an exact copy of a DOFSystemConfigBuilder.  Any errors in the original DOFSystemConfigBuilder will also be present in the returned DOFSystemConfigBuilder.
 *
 *  @mustfree{DOFSystemConfigBuilder_Destroy or @ref DOFSystemConfigBuilder_BuildAndDestroy}
 *
 *  @param self The DOFSystemConfigBuilder to copy.  Should not be @c NULL.
 *
 *  @return a copy of @a self
 */
extern DOF_OAL_DECLSPEC DOFSystemConfigBuilder DOFSystemConfigBuilder_Copy( DOFSystemConfigBuilder self );

/**
 *  Destroy a DOFSystemConfigBuilder.
 *
 *  @param self The DOFSystemConfigBuilder to destroy.  If @c NULL this function will do nothing.
 */
extern DOF_OAL_DECLSPEC void DOFSystemConfigBuilder_Destroy( DOFSystemConfigBuilder self );

/**
@}
**/

/**
 *  Destroy a DOFSystemConfig.
 *
 *  @param self The DOFSystemConfig to destroy.  If @c NULL this function will do nothing.
 */
extern DOF_OAL_DECLSPEC void DOFSystemConfig_Destroy( DOFSystemConfig self );

/**
 *  Retrieve the DOFCredentials used in a DOFSystemConfig.
 *
 *  @param self The DOFSystemConfig.  Should not be @c NULL.
 *
 *  @return the DOFCredentials used by @a self
 */
extern DOF_OAL_DECLSPEC DOFCredentials DOFSystemConfig_GetCredentials( DOFSystemConfig self );

/**
 *  Retrieve the DOFOperationFilter for inbound traffic used in a DOFSystemConfig.
 *
 *  @param self The DOFSystemConfig.  Should not be @c NULL.
 *
 *  @return the DOFOperationFilter used for traffic received by @a self
 */
extern DOF_OAL_DECLSPEC DOFOperationFilter DOFSystemConfig_GetReceiveFilter( DOFSystemConfig self );

/**
 *  Retrieve the DOFOperationFilter for outbound traffic used in a DOFSystemConfig.
 *
 *  @param self The DOFSystemConfig.  Should not be @c NULL.
 *
 *  @return the DOFOperationFilter used for traffic sent by @a self
 */
extern DOF_OAL_DECLSPEC DOFOperationFilter DOFSystemConfig_GetSendFilter( DOFSystemConfig self );

/**
 *  Retrieve the initial DOFPermissionSet used in a DOFSystemConfig.
 *
 *  @param self The DOFSystemConfig.  Should not be @c NULL.
 *
 *  @return the initial DOFPermissionSet used in @a self
 */
extern DOF_OAL_DECLSPEC DOFPermissionSet DOFSystemConfig_GetPermissions( DOFSystemConfig self );

/**
 *  Determine if this DOFSystemConfig allows permission extension.
 *
 *  @param self The DOFSystemConfig.  Should not be @c NULL.
 *
 *  @retval TRUE    @a self allows permission extension
 *  @retval FALSE   @a self does not allow permission extension or @a self is NULL
 */
extern DOF_OAL_DECLSPEC boolean DOFSystemConfig_IsPermissionsExtendAllowed( DOFSystemConfig self );

/**
 *  Retrieve the maximum number of threads used in a DOFSystemConfig.
 *
 *  @param self The DOFSystemConfig.  Should not be @c NULL.
 *
 *  @return the maximum threads to use for the DOFSystem thread pool
 */
extern DOF_OAL_DECLSPEC uint16 DOFSystemConfig_GetThreadpoolSize( DOFSystemConfig self );

/**
@}
**/

/** @defgroup DOFSystemState DOFSystemState
    @ingroup DOFSystem

    Represents the instantaneous state of a DOFSystem.
@{
**/

/**
 *  Destroy a DOFSystemState.
 *
 *  @param self The DOFSystemState to destroy.  If @c NULL this function will do nothing.
 */
extern DOF_OAL_DECLSPEC void DOFSystemState_Destroy( DOFSystemState self );

/**
 *  Retrieve the DOFCredentials used in a DOFSystem.
 *
 *  @param self The DOFSystemState of interest.  Should not be @c NULL.
 *
 *  @return the DOFCredentials used for the relevent DOFSystem
 */
extern DOF_OAL_DECLSPEC DOFCredentials DOFSystemState_GetCredentials( DOFSystemState self );

/**
 *  Retrieve the permissions currently permitted by a DOFSystem.  If @a self is from a non-secure DOFSystem that
 *  allowed permission extension, this function will always return @c NULL.
 *
 *  @param self The DOFSystemState of itnerest.  Should not be @c NULL.
 *
 *  @return the DOFPermissionSet containing allowed permissions
 */
extern DOF_OAL_DECLSPEC DOFPermissionSet DOFSystemState_GetPermissions( DOFSystemState self );

/**
 *  Determine if permission extension is permitted by a DOFSystem.
 *
 *  @param self The DOFSystemState of interest.  Should not be @c NULL.
 *
 *  @retval TRUE permission extension is allowed
 *  @retval FALSE permission extension is not allowed
 */
extern DOF_OAL_DECLSPEC boolean DOFSystemState_IsPermissionsExtendAllowed( DOFSystemState self );

/**
 *  Retrieve the number of threads in the DOFSystem threadpool.
 *
 *  @param self The DOFSystemState of interest.  This should not be @c NULL.
 *
 *  @return the number of threads in the DOFSystem's threadpool
 *  @retval 0 the DOFSystem uses the default threadpool instead of its own
 */
extern DOF_OAL_DECLSPEC uint16 DOFSystemState_GetThreadpoolSize( DOFSystemState self );

/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* DOF_SYSTEM_H */
