/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_TRANSPORT_H
#define DOF_TRANSPORT_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @defgroup DTL DOF Transport Layer (DTL)

    The DOF Transport Layer is the OAL's abstraction of a transport.  Any
    number of transports may be implemented and used with the OAL.  Transport
    implementations are distributed separately from the OAL.

    Each transport layer works directly with a Transport Handler (another
    interface), which is identified when the transport is asked to perform a
    task.  The Transport Handler is solely responsible for calling into a
    transport as well as accepting calls from the transport.

@{
**/

/** @file
    Header file for the DOF Transport Layer interfaces, functions, defines and data types.
**/

/*
**  "Public" Type and Method Declarations
*/


/** @name Interfaces
@{
**/

/**
    The Transport Handler interface.  Defined as a pointer to the @ref DTLTransportHandler_t structure.  Refer to the structure for more information.
**/
typedef struct DTLTransportHandler_t * DTLTransportHandler;

/**
    The Transport-specific Server interface.  Defined as a pointer to the @ref DTLServer_t structure.  Refer to the structure for more information.
**/
typedef struct DTLServer_t *           DTLServer;

/**
    The Transport-specific Connection interface.  Defined as a pointer to the @ref DTLConnection_t structure.  Refer to the structure for more information.
**/
typedef struct DTLConnection_t *       DTLConnection;

/**
    The Transport-specific core interface.  Defined as a pointer to the @ref DTLTransport_t structure.  Refer to the structure for more information.
**/
typedef struct DTLTransport_t *        DTLTransport;

/**
@}
**/


/** @name DTLConnectionConfig
@{
**/

/**
 * The transport's Connection Configuration, allowing transport-specific
 * configuration of a connection.  The connection configuration is specific to
 * each transport, and is created directly from the transport.  The
 * DTLConnectionConfig is a convenient wrapper of any, arbitrary configuration
 * data used by the transport.  It is assumed the data is immutable.
 */
typedef struct DTLConnectionConfig_t * DTLConnectionConfig;

/**
 * Destroyer function for connection configuration.  This callback is used to
 * notify the Transport that the connection configuration is being destroyed.
 * This allows the Transport to clean up the configuration, if necessary.  The
 * data was originally copied into the DTLConnectionConfig and will be freed
 * after this call completes.  However, since the copy was only a shallow copy
 * of the data, the transport may need to perform a deeper free/release of its
 * held resources.
 *
 * @param config The transport-specific configuration data.
 */
typedef void (*DTLConnectionConfig_Destroyer)( const void * config );

/**
    Create a connection configuration from a transport configuration and the
    transport.

    Connection configuration is unique to a transport, and is normally created
    by the transport.  It is opaque to the DTLConnectionConfig, and no
    assumptions can or should be made about its type or contents, except by its
    associated transport.

    @mustfree{DTLConnectionConfig_Destroy}

    @param  configSize Size of the configuration data.
    @param  config     The configuration data.  This may be NULL if @a configSize is 0.
    @param  transport  The transport associated with the address.  This must not be NULL.
    @param  destroyer  The function to call when the configuration is destroyed.  This may be NULL if there is no need for this notification.

    @return The DTLConnectionConfig, or NULL if one could not be created, or
    invalid parameters were used.
**/
extern DOF_OAL_DECLSPEC DTLConnectionConfig DTLConnectionConfig_Create( uint32 configSize, const void * config, DTLTransport transport, DTLConnectionConfig_Destroyer destroyer );

/**
 *  Create a new copy of an existing @ref DTLConnectionConfig.
 *  @mustfree{DTLConnectionConfig_Destroy}
 *  @param  self The @ref DTLConnectionConfig to be copied.  This should not be NULL.
 *  @return A new copy of the @ref DTLConnectionConfig.
 *  @exception  NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DTLConnectionConfig DTLConnectionConfig_Copy( const DTLConnectionConfig self );

/**
    Compare a connection configuration against another for equality.

    Connection configuration equality is determined first by matching the
    associated transport.  If both connection configurations are associated
    with the same transport, the transport is then used to determine the
    equality of the connection configurations.  Connection configurations
    associated with different transports are never equal.

    @param self   The connection configuration to be compared.  This paramater
                  should not be NULL.
    @param config The connection configuration with which to compare.  This
                  parameter should not be NULL.

    @retval TRUE  The connection configurations are equal.
    @retval FALSE The connection configurations are not equal.

    @sa     DTLTransport_IsConnectionConfigEqual
**/
extern DOF_OAL_DECLSPEC boolean DTLConnectionConfig_IsEqual( const DTLConnectionConfig self, const DTLConnectionConfig config );

/**
    Destroy a connection configuration.

    @param  self The connection configuration to be destroyed.  If NULL, no
                 action is taken.
**/
extern DOF_OAL_DECLSPEC void DTLConnectionConfig_Destroy( DTLConnectionConfig self );

/**
    Return the transport-specific connection configuration data stored within
    the DTLConnectionConfig.

    @note This data is owned by the DTLConnectionConfig and must be copied if
    it is needed outside the scope of the DTLConnectionConfig.  It is not valid
    after the DTLConnectionConfig is destroyed.

    @param  self The connection configuration.  This parameter should not be NULL.

    @return The transport-specific connection configuration data stored within
    the DTLConnectionConfig.

    @exception  NULL is returned if @a self is NULL.
**/
extern DOF_OAL_DECLSPEC const void * DTLConnectionConfig_GetConfig( const DTLConnectionConfig self );

/**
    Return the size of the Transport-specific connection configuration data
    stored within the DTLConnectionConfig.

    @param  self The connection configuration.  This parameter should not be
                 NULL.

    @return The size of the transport-specific connection configuration data
    stored within the DTLConnectionConfig.

    @exception  NULL is returned if @a self is NULL.
**/
extern DOF_OAL_DECLSPEC uint32 DTLConnectionConfig_GetConfigSize( const DTLConnectionConfig self );

/**
    Return the transport associated with the DTLConnectionConfig.

    @param  self The connection configuration.  This parameter should not be
                 NULL.

    @return The transport associated with the DTLConnectionConfig.

    @exception  NULL is returned if @a self is NULL.
**/
extern DOF_OAL_DECLSPEC DTLTransport DTLConnectionConfig_GetTransport( const DTLConnectionConfig self );

/**
@}
**/


/** @name DTLConnection
@{
**/


/**
 * The transport's Connection interface structure, representing a
 * transport-level connection.  This interface encapsulates the transport-level
 * implementation of a connection, and is implemented by a transport that is
 * intended to work with the C OAL.  These connections can be inbound or
 * outbound, streaming (e.g., TCP) or datagram (e.g., UDP), and point-to-point
 * or point-to-multipoint (e.g., a datagram socket capable of sending multicast
 * or broadcast).
 *
 * Connections are expected to be mostly immutable.  Their properties are not
 * expected to change, with the exception of the context (it may be both set
 * and retrieved).  The context is opaque to the connection.
 *
 * This interface is called by a Transport Handler in the OAL.
 *
 * This structure should be placed as the first element in a larger structure
 * representing any necessary state of the Connection.  All of the functions
 * take as their first argument a pointer to this structure (a @ref
 * DTLConnection).  Inside each function, it can thus be type-cast as a pointer
 * to the larger structure to access the connection's state.  If no additional
 * state is required, this structure may be used directly.
 */
typedef struct DTLConnection_t
{

    /**
     * The DTLConnection interface function structure, containing a list of
     * function pointers that implement the interface.  It has the same role as
     * a vtable in C++.
     */
    const struct DTLConnectionFns_t
    {

        /**
         * Send data to the specified address.  This method should not block; if it
         * does, the call may delay the caller longer than intended, and may result in
         * the violation of certain expected timing constraints as well as poor
         * performance.  This call cannot fail.  The transport may choose to queue the
         * data for later delivery (for streaming connections), drop the data (for
         * datagram connections), or close the connection.  The actual data sent is @a
         * data[@a offset] through @a data[@a length].
         *
         * @note Ownership of the data buffer is passed into this call.  The transport
         * is required to free the @a data buffer with PCRMem_Free().
         *
         * This callback must not be NULL.
         *
         * @param connection  The connection.  This parameter must not be NULL.
         * @param target      The target address.  If this parameter is NULL, or the
         *                    connection is point-to-point, this parameter is ignored
         *                    and the peer address of the connection is assumed.
         * @param data        The data buffer.  @a data must not be NULL.
         * @param offset      The offset into the data buffer where the data to send
         *                    begins.  This parameter must be less than @a length.
         * @param length      The length of the data to send.
         */
        void (*Send)( DTLConnection connection, DOFAddress target, uint8 * data, uint32 offset, uint32 length );

        /**
         * Close the connection.
         *
         * This callback must not be NULL.
         *
         * @param connection  The connection.  This parameter must not be NULL.
         */
        void (*Close)( DTLConnection connection );

        /**
         * Get the peer Address of the connection.
         *
         * @note Ownership of the return value is @b not passed to the caller.  It is
         *       valid only while the connection is valid.
         *
         * This callback must not be NULL.
         *
         * @param connection  The connection.  This parameter must not be NULL.
         *
         * @return The peer address of the connection.
         */
        DOFAddress (*GetAddress)( DTLConnection connection );

        /**
         * Get the server associated with the connection, if any.
         *
         * @note Ownership of the return value is @b not passed to the caller.  It is
         *       valid only while the connection is valid.
         *
         * This callback must not be NULL.
         *
         * @param connection  The connection.  This parameter must not be NULL.
         *
         * @return The server associated with the connection, or NULL if there is no server associated with the connection.
         */
        DTLServer (*GetServer)( DTLConnection connection );

        /**
         * Get the Transport Handler associated with the connection.
         *
         * @note Ownership of the return value is @b not passed to the caller.  It is
         *       valid only while the connection is valid.
         *
         * This callback must not be NULL.
         *
         * @param connection  The connection.  This parameter must not be NULL.
         *
         * @return The Transport Handler associated with the connection.
         */
        DTLTransportHandler (*GetHandler)( DTLConnection connection );

        /**
         * Get the type of connection.
         *
         * This must be either @ref DOFCONNECTIONTYPE_STREAM or @ref DOFCONNECTIONTYPE_DATAGRAM.
         *
         * This callback must not be NULL.
         *
         * @param connection  The connection.  This parameter must not be NULL.
         *
         * @return The type of the connection.
         */
        DOFConnectionType (*GetConnectionType)( DTLConnection connection );

        /**
         * Get the context associated with the connection.
         * The context is opaque to the connection.  The context is normally set for
         * outbound connections when the connection is created, at the request of the
         * Transport Handler, and can be retrieved with this call.  The context is
         * normally set via @ref SetContext by the Transport Handler for
         * inbound connections created from a server, and can be retrieved with this
         * call.  If the context has not yet been set (e.g., for an inbound
         * connection), this must return NULL.
         * @param connection  The connection.  This parameter must not be NULL.
         * @return The context associated with the connection.
         */
        void * (*GetContext)( DTLConnection connection );

        /**
         * Set the context associated with the connection.
         *
         * The context is opaque to the connection.  The context is normally set for
         * outbound connections when the connection is created, at the request of the
         * Transport Handler.  This call is normally used by the Transport Handler to
         * set the context for inbound connections created from a server (the
         * connections in this case are created with a NULL context).
         *
         * This callback must not be NULL.
         *
         * @param connection  The connection.  This parameter must not be NULL.
         * @param context     The context to associate with the connection.
         */
        void (*SetContext)( DTLConnection connection, void * context );

        /**
         * Report whether or not the connection needs to be handled asynchronously.
         *
         * If the connection requires asynchronous handling (this function returns
         * TRUE), the data received must be queued for later processing and the
         * Transport Handler's receive call will return immediately.  If the connection
         * can be handled synchronously (this function returns FALSE), then data
         * received from the connection is allowed to be processed immediately,
         * blocking the caller.
         *
         * @note This is only an optimization hint: depending on the implementation of
         * the Transport Handler, data may always be handled asynchronously (queued for
         * later processing), regardless of this function's return value.
         *
         * This callback must not be NULL.
         *
         * @param connection  The connection.  This parameter must not be NULL.
         * @retval TRUE The connection is asynchronous; received packets must be queued.
         * @retval FALSE The connection is synchronous; received packets may be processed immediately (synchronously).
         */
        boolean (*IsHandleAsync)( DTLConnection connection );

        /**
         * Report whether or not the connection is inherently secure.
         *
         * Data sent to and received from a secure connection is not required to be
         * encrypted nor authenticated by the OAL to maintain its security (it is
         * assumed the transport is encrypting and validating the data).
         *
         * This callback must not be NULL.
         *
         * @param connection  The connection.  This parameter must not be NULL.
         * @retval TRUE The connection is guaranteed by the transport to be secure.
         * @retval FALSE The connection is not or may not be secure.
         */
        boolean (*IsSecure)( DTLConnection connection );

        /**
         * Report whether or not the connection is point-to-point.
         *
         * Point-to-point connections have a peer address that represents a single node
         * in the network.  A connection that has a peer address representing a group
         * of nodes (e.g., multicast or broadcast) is not point-to-point (it may be
         * referred to as point-to-multipoint).
         *
         * This callback must not be NULL.
         *
         * @param connection  The connection.  This parameter must not be NULL.
         * @retval TRUE The connection is point-to-point.
         * @retval FALSE The connection is not point-to-point.
         */
        boolean (*IsPointToPoint)( DTLConnection connection );

        /**
         * Report the connection configuration associated with the connection.
         *
         * This callback must not be NULL.
         *
         * @param connection  The connection.  This parameter must not be NULL.
         * @retval The connection configuration associated with the connection, or NULL if there is no such configuration.
         */
        DTLConnectionConfig (*GetConnectionConfig)( DTLConnection connection );

    } * fns;

    /* Implementations may extend this structure. */

} DTLConnection_t;

/**
@}
**/


/** @name DTLServerConfig
@{
**/

/**
 * The transport's Server Configuration, allowing transport-specific
 * configuration of a server.  The server configuration is specific to each
 * transport, and is created directly from the transport.  The DTLServerConfig
 * is a convenient wrapper of any, arbitrary configuration data used by the
 * transport.  It is assumed the data is immutable.
 */
typedef struct DTLServerConfig_t * DTLServerConfig;

/**
 * Destroyer function for server configuration.  This callback is used to
 * notify the transport that the server configuration is being destroyed.
 * This allows the transport to clean up the configuration, if necessary.  The
 * data was originally copied into the DTLServerConfig and will be freed
 * after this call completes.  However, since the copy was only a shallow copy
 * of the data, the transport may need to perform a deeper free/release of its
 * held resources.
 *
 * @param config The transport-specific configuration data.
 */
typedef void (*DTLServerConfig_Destroyer)( const void * config );

/**
    Create a server configuration from a transport configuration and the
    transport.

    Server configuration is unique to a transport, and is normally created by
    the transport.  It is opaque to the DTLServerConfig, and no assumptions can
    or should be made about its type or contents, except by its associated
    transport.

    @mustfree{DTLServerConfig_Destroy}

    @param  configSize Size of the configuration data.
    @param  config     The configuration data.  This may be NULL if @a configSize is 0.
    @param  transport  The transport associated with the address.  This must not be NULL.
    @param  destroyer  The function to call when the configuration is destroyed.  This may be NULL if there is no need for this notification.

    @return The DTLServerConfig, or NULL if one could not be created, or
    invalid parameters were used.
**/
extern DOF_OAL_DECLSPEC DTLServerConfig DTLServerConfig_Create( uint32 configSize, const void * config, DTLTransport transport, DTLServerConfig_Destroyer destroyer );

/**
 *  Create a new copy of an existing @ref DTLServerConfig.
 *  @mustfree{DTLServerConfig_Destroy}
 *  @param  self The @ref DTLServerConfig to be copied.  This should not be NULL.
 *  @return A new copy of the @ref DTLServerConfig.
 *  @exception  NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DTLServerConfig DTLServerConfig_Copy( const DTLServerConfig self );

/**
    Compare a server configuration against another for equality.

    Server configuration equality is determined first by matching the
    associated transport.  If both server configurations are associated with
    the same transport, the transport is then used to determine the equality of
    the server configurations.  Server configurations associated with different
    transports are never equal.

    @param self   The server configuration to be compared.  This should not be
                  NULL.
    @param config The server configuration with which to compare.  This should
                  not be NULL.

    @retval TRUE  The server configurations are equal.
    @retval FALSE The server configurations are not equal.

    @sa     DTLTransport_IsServerConfigEqual
**/
extern DOF_OAL_DECLSPEC boolean DTLServerConfig_IsEqual( const DTLServerConfig self, const DTLServerConfig config );

/**
    Destroy a server configuration.

    @param  self The server configuration to be destroyed.  If NULL, this
                 function will do nothing
**/
extern DOF_OAL_DECLSPEC void DTLServerConfig_Destroy( DTLServerConfig self );

/**
    Return the transport-specific server configuration data stored within the
    DTLServerConfig.

    @note This data is owned by the DTLServerConfig and must be copied if it is
    needed outside the scope of the DTLServerConfig.  It is not valid after the
    DTLServerConfig is destroyed.

    @param  self The server configuration.  This should not be NULL.

    @return The transport-specific server configuration data stored within the
    DTLServerConfig.

    @exception   NULL is returned if @a self is NULL.
**/
extern DOF_OAL_DECLSPEC const void * DTLServerConfig_GetConfig( const DTLServerConfig self );

/**
    Return the size of the transport-specific server configuration data stored
    within the DTLServerConfig.

    @param  self The server configuration.  This should not be NULL.

    @return The size of the transport-specific server configuration data stored
    within the DTLServerConfig.

    @exception   NULL is returned if @a self is NULL.
**/
extern DOF_OAL_DECLSPEC uint32 DTLServerConfig_GetConfigSize( const DTLServerConfig self );

/**
    Return the transport associated with the DTLServerConfig.

    @param  self The server configuration.  This should not be NULL.

    @return The transport associated with the DTLServerConfig.

    @exception   NULL is returned if @a self is NULL.
**/
extern DOF_OAL_DECLSPEC DTLTransport DTLServerConfig_GetTransport( const DTLServerConfig self );

/**
@}
**/


/** @name DTLServer
@{
**/

/**
 * The transport's Server interface structure, representing a transport-level
 * server.  This interface encapsulates the transport-level implementation of a
 * server, and is implemented by a transport that is intended to work with the
 * C OAL.  These servers can be streaming (e.g., TCP) or datagram (e.g., UDP).
 * Datagram servers may further be able to receive data via multicast or
 * broadcast.
 *
 * Servers are expected to be immutable.  Their properties are not expected to
 * change.  The context is opaque to the server.
 *
 * This interface is called by a Transport Handler in the OAL.
 *
 * This structure should be placed as the first element in a larger structure
 * representing any necessary state of the Server.  All of the functions
 * take as their first argument a pointer to this structure (a @ref
 * DTLServer).  Inside each function, it can thus be type-cast as a pointer
 * to the larger structure to access the server's state.  If no additional
 * state is required, this structure may be used directly.
 */
typedef struct DTLServer_t
{

    /**
     * The DTLServer interface function structure, containing a list of
     * function pointers that implement the interface.  It has the same role as
     * a vtable in C++.
     */
    const struct DTLServerFns_t
    {

        /**
         * Send a response to the specified address using a server.  This call is only
         * valid for datagram servers.  This method should not block; if it does, the
         * call may delay the caller longer than intended, and may result in the
         * violation of certain expected timing constraints as well as poor
         * performance.  This call cannot fail.  The transport may choose to drop the
         * data if it cannot be sent.
         *
         * @note Ownership of the data buffer is passed into this call.  The transport
         * is required to free the @a data buffer with PCRMem_Free().
         *
         * This callback must not be NULL.
         *
         * @param server      The server.  This parameter must not be NULL.
         * @param target      The target address.  This parameter must not be NULL.
         * @param data        The data buffer.  If NULL, this function does nothing.
         * @param offset      The offset into the data buffer where the data to send begins.
         *                    Must be less than @a length.
         * @param length      The length of the data to send.  Must be greater than @a offset.
         */
        void (*Send)( DTLServer server, DOFAddress target, uint8 * data, uint32 offset, uint32 length );

        /**
         * Stop the server.
         *
         * This callback must not be NULL.
         *
         * @param server      The DTLServer to stop.  If NULL, this function will do nothing.
         */
        void (*Stop)( DTLServer server );

        /**
         * Get the number of server connections.
         *
         * This callback must not be NULL.
         *
         * @param server      The server.  This parameter should not be NULL.
         *
         * @return            The number of remote connections to @a server.
         *
         * @exception         0 is returned if @a server is NULL.
         */
        uint32 (*GetConnectionCount)( DTLServer server );

        /**
         * Get the listening address of the server.
         *
         * This callback must not be NULL.
         *
         * @param server      The server.  This parameter should not be NULL.
         *
         * @return            The Transport specific DOFAddress @a server is listening
         *                    on.
         *
         * @exception         NULL is returned if @a server is NULL.
         */
        DOFAddress (*GetAddress)( DTLServer server );

        /**
         * Get the related server associated with the server, if any.
         *
         * This callback must not be NULL.
         *
         * @param server      The server.  This parameter should not be NULL.
         *
         * @return            @a server's related server (if one exists).
         *
         * @exception         NULL is returned if @a server is NULL.
         */
        DTLServer (*GetRelatedServer)( DTLServer server );

        /**
         * Get the Transport Handler associated with the server.
         *
         * This callback must not be NULL.
         *
         * @param server      The server.  This parameter should not be NULL.
         *
         * @return            The TransportHandler associated with @a server.
         *
         * @exception         NULL is returned if @a server is NULL.
         */
        DTLTransportHandler (*GetHandler)( DTLServer server );

        /**
         * Get the type of server.
         *
         * This callback must not be NULL.
         *
         * @param server      The server.  This parameter must not be NULL.
         *
         * @return            The type of Connection @a server supports.
         */
        DOFServerType (*GetServerType)( DTLServer server );

        /**
         * Get the context associated with the server.
         *
         * @note              The context is reserved for use by the
         *                    TransportHandler.
         *
         * This callback must not be NULL.
         *
         * @param server      The server.  This parameter should not be NULL.
         *
         * @return            The context of this @a server.
         *
         * @exception         NULL is returned if @a server is NULL.
         */
        void * (*GetContext)( DTLServer server );

        /**
         * Report whether or not the server needs to be handled asynchronously.
         *
         * This callback must not be NULL.
         *
         * @param server      The server.  This parameter should not be NULL.
         *
         * @return            TRUE if @a server is asynchronous, FALSE if it is not.
         *
         * @exception         FALSE is returned if @a server is NULL.
         */
        boolean (*IsHandleAsync)( DTLServer server );

        /**
         * Report whether or not the server is inherently secure.
         *
         * This callback must not be NULL.
         *
         * @param server      The server.  This parameter should not be NULL.
         *
         * @return            TRUE if @a server inheritly handles security.
         *
         * @exception         FALSE is returned if @a server is NULL.
         */
        boolean (*IsSecure)( DTLServer server );

        /**
         * Report the connection configuration associated with a connection to the given address.
         *
         * When a connection state is needed for an inbound packet from a
         * datagram server (as if it were a connection), this returns the
         * necessary configuration for the connection to that host.  This
         * will only be called for datagram servers, and always using the
         * DOFAddress passed to
         * @ref DTLTransportHandler_t::DTLTransportHandlerFns_t::Received_Server "Received_Server"
         * as the source of the datagram packet.
         *
         * This callback must not be NULL.
         *
         * @param server      The server.  This parameter should not be NULL.
         * @param address     The address of the host.  This parameter must not be NULL.
         * @retval The connection configuration associated with the connection to the given address, or NULL if there is no such configuration.
         */
        DTLConnectionConfig (*GetConnectionConfig)( DTLServer server, DOFAddress address );

    } * fns;    /* Server-specific function calls. */

    /* Implementations may extend this structure. */

} DTLServer_t;

/**
@}
**/


/** @name DTLTransport
@{
**/

/**
 * The core Transport interface structure, allowing connections and servers to
 * be created at the transport layer.  This interface is implemented by a
 * transport that is intended to work with the C OAL.
 *
 * This interface is called by a Transport Handler in the OAL to open outbound
 * connections and start servers.
 *
 * The connections and servers are separate interfaces, also implemented by the
 * transport, and are also utilized by the Transport Handler.
 *
 * This structure should be placed as the first element in a larger structure
 * representing any necessary state of the Transport.  All of the functions
 * take as their first argument a pointer to this structure (a @ref
 * DTLServer).  Inside each function, it can thus be type-cast as a pointer
 * to the larger structure to access the transport's state.  If no additional
 * state is required, this structure may be used directly.
 */
typedef struct DTLTransport_t
{

    /**
     * The DTLTransport interface function structure, containing a list of
     * function pointers that implement the interface.  It has the same role as
     * a vtable in C++.
     */
    const struct DTLTransportFns_t
    {

        /**
         * Open a connection.
         *
         * The transport associated with the @a peer address must match the transport
         * being called.  This call should not block; if it does, the call may delay
         * the caller longer than intended, and may result in the violation of certain
         * expected timing constraints.
         *
         * The parameters to this function are owned by the caller.  However, they are
         * maintained by the caller for the duration of the resulting connection.
         * Therefore, there is no need for the transport to copy these parameters, so
         * long as they are never used past the lifetime of the resulting connection.
         * The connection's lifetime terminates when the Transport Handler is notified
         * via its @ref DTLTransportHandler_t::DTLTransportHandlerFns_t::Closed "Closed"
         * method.
         *
         * This callback may be NULL if the transport does not support connections.
         *
         * @param transport The Transport.  This parameter must not be NULL.
         * @param config  Transport-specific custom parameters for the connection.
         *                There are no standard parameters defined; the implementation
         *                is free to choose what options are valid, if any, and their
         *                appropriate behavior.  If NULL, default parameters will be
         *                used.
         * @param peer    The peer address representing the remote end of the connection.
         *                @a peer must not be NULL.
         * @param handler The associated transport handler for the connection.  Once
         *                the connection is opened successfully, the
         *                @ref DTLTransportHandler_t::DTLTransportHandlerFns_t::Opened
         *                "Opened" method of @a handler must be called with the
         *                connection object that represents the connection created by
         *                this call.  If the connection cannot be opened successfully,
         *                @ref DTLTransportHandler_t::DTLTransportHandlerFns_t::Closed
         *                "Closed" must be called with the appropriate exception.  Even
         *                if this call may determine synchronously that the connection
         *                is successful or failed, a valid @ref DTLConnection must be
         *                given to one of the handler methods to indicate the success
         *                or failure, and the same DTLServer (even if already freed, in
         *                the case of a failure) must be returned.
         *                @a handler must not be NULL.
         * @param type    The type of connection to open, either @ref
         *                DOFCONNECTIONTYPE_STREAM or @ref DOFCONNECTIONTYPE_DATAGRAM.
         * @param context The context for this connection.  This context must be
         *                associated with the connection object that represents this
         *                connection, when passed to the @a handler.  The context is
         *                opaque to the connection.
         *
         * @return TRUE is returned. Success or failure is determined asynchronously via the handler callbacks as described
         *         above. It is possible that the callback is called prior to, concurrent with, or following the return of this
         *         function.
         * @exception FALSE is returned to indicate an allocation failure of the DTLConnection itself.  All other errors must
         *            result in TRUE returned and the handler's Closed method called with the appropriate exception (as
         *            described above).
         */
        boolean (*Open)( DTLTransport transport, DTLConnectionConfig config, DOFAddress peer,
                               DTLTransportHandler handler, DOFConnectionType type, void * context );

        /**
         * Start a server.
         *
         * The transport associated with the @a listenAddr address must match the
         * transport being called.  This call should not block; if it does, the call
         * may delay the caller longer than intended, and may result in the violation
         * of certain expected timing constraints.
         *
         * The parameters to this function are owned by the caller.  However, they are
         * maintained by the caller for the duration of the resulting server.
         * Therefore, there is no need for the transport to copy these parameters, so
         * long as they are never used past the lifetime of the resulting server.  The
         * server's lifetime terminates when the Transport Handler is notified via its
         * @ref DTLTransportHandler_t::DTLTransportHandlerFns_t::Stopped "Stopped"
         * method.
         *
         * This callback may be NULL if the transport does not support servers.
         *
         * @param transport  The Transport.  This parameter must not be NULL.
         * @param config     Transport defined configuration options for @a relatedServer.
         *                   If NULL, default options will be used.
         * @param listenAddress The address the server should bind to, which may be an
         *                   address that indicates all known addresses for the transport
         *                   should be bound.  It may alternatively represent a multicast
         *                   or broadcast address, in which case the server must prepare
         *                   to accept such traffic (e.g., join a multicast group).
         *                   This must not be NULL.
         * @param handler The associated transport handler for the server.  Once
         *                the server is started successfully, the @ref
         *                DTLTransportHandler_t::DTLTransportHandlerFns_t::Started
         *                "Started" method of @a handler must be called with the
         *                server object that represents the server created by this call.
         *                If the server cannot be opened successfully,
         *                @ref DTLTransportHandler_t::DTLTransportHandlerFns_t::Stopped
         *                "Stopped" must be called with the appropriate exception.  Even
         *                if this call may determine synchronously that the server
         *                is successful or failed, a valid @ref DTLServer must be given
         *                to one of the handler methods to indicate the success or failure,
         *                and the same DTLServer (even if already freed, in the case of a
         *                failure) must be returned.
         *                @a handler must not be NULL.
         * @param type    The type of server to start, either @ref
         *                DOFSERVERTYPE_STREAM or @ref DOFSERVERTYPE_DATAGRAM.
         * @param relatedServer An existing, related server, if any.  Some datagram
         *                      servers may want to utilize or reference the
         *                      configuration or implementation of an existing server.
         *                      For example, a UDP multicast server may simply join the
         *                      multicast group on the existing server's socket,
         *                      instead of creating a new, separate socket.  If a
         *                      server is shared, however, it is the transport's
         *                      responsibility to correctly identify to which server a
         *                      packet belongs.
         * @param context The context for this server.  This context must be associated
         *                with the server object that represents this server, when
         *                passed to the @a handler.
         *
         * @return TRUE is returned. Success or failure is determined asynchronously via the handler callbacks as described
         *         above. It is possible that the callback is called prior to, concurrent with, or following the return of this
         *         function.
         * @exception FALSE is returned to indicate an allocation failure of the DTLServer itself.  All other errors must result
         *            in TRUE returned and the handler's Stopped method called with the appropriate exception (as described
         *            above).
         */
         boolean (*Start)( DTLTransport transport, DTLServerConfig config, DOFAddress listenAddress,
                             DTLTransportHandler handler, DOFServerType type, DTLServer relatedServer,
                             void * context );

        /**
         * Get the Maximum Transmission Unit (MTU) size for the specified type of transmission.
         *
         * This callback may be NULL if the maximum size for all connection (and server)
         * types is effectively unlimited.
         *
         * @param transport The Transport.  This parameter must not be NULL.
         * @param type    The transmission type of interest, (e.g., @ref
         *                DOFCONNECTIONTYPE_STREAM or @ref DOFCONNECTIONTYPE_DATAGRAM).
         *
         * @return The Maximum Transmission Unit (MTU) size for the specified type of transmission, in bytes.
         *         0 indicates the specified transmission type is not supported.
         */
        uint32 (*GetMaxSize)( DTLTransport transport, DOFConnectionType type );

        /**
         * Get the Minimum Transmission size for the specified type of transmission.
         *
         * This callback may be NULL if the minimum size for all connection (and server)
         * types is zero (0).
         *
         * @param transport The Transport.  This parameter must not be NULL.
         * @param type    The transmission type of interest, (e.g., @ref
         *                DOFCONNECTIONTYPE_STREAM or @ref DOFCONNECTIONTYPE_DATAGRAM).
         *
         * @return The Minimum Transmission size for the specified type of transmission, in bytes.
         */
        uint32 (*GetMinSize)( DTLTransport transport, DOFConnectionType type );

        /**
         * Get the type of the specified address.
         *
         * This callback may be NULL if all transport addresses are unicast.
         *
         * @param transport The Transport.  This parameter must not be NULL.
         * @param address The DOFAddress.  This parameter must not be NULL.
         *
         * @return The type of the address.
         */
        DOFAddressType (*GetAddressType)( DTLTransport transport, DOFAddress address );

        /**
         * Report if two Transport DTLConnectionConfigs are equal.
         *
         * This callback may be NULL if PCRMem_Compare is sufficient to determine equality.
         *
         * @param transport The Transport.  This parameter must not be NULL.
         * @param size      The size in bytes of the configs.  This parameter must not be NULL.
         * @param config1   The first DTLConnectionConfig to compare.  Must not
         *                  be NULL and must be associated with @a transport.
         * @param config2   The second DTLConnectionConfig to compare.  Must not
         *                  be NULL and must be associated with @a transport.
         *
         * @retval TRUE  The two ConnectionConfigs are equivalent.
         * @retval FALSE The two ConnectionConfigs are different, or either
         *               ConnectionConfig has an associated transport different than
         *               @a transport.
         */
        boolean (*IsConnectionConfigEqual)( DTLTransport transport, uint32 size, const void * config1, const void * config2 );

        /**
         * Report if two Transport DTLServerConfigs are equal.
         *
         * This callback may be NULL if PCRMem_Compare is sufficient to determine equality.
         *
         * @param transport The Transport.  This parameter must not be NULL.
         * @param size      The size in bytes of the configs.  This parameter must not be NULL.
         * @param config1   The first DTLServerConfig to compare.  Must not be NULL
         *                  and must be associated with @a transport.
         * @param config2   The second DTLServerConfig to compare.  Must not be NULL
         *                  and must be associated with @a transport.
         *
         * @retval TRUE The two DTLServerConfigs are equivalent.
         * @retval FALSE The two DTLServerConfigs are different, or either ServerConfig
         *               has an associated transport different than @a transport.
         */
        boolean (*IsServerConfigEqual)( DTLTransport transport, uint32 size, const void * config1, const void * config2 );

        /**
         * Compares two transport addresses.
         *
         * This callback may be NULL if PCRMem_Compare is sufficient to determine equality.
         *
         * @param transport The Transport.  This parameter must not be NULL.
         * @param size      The size in bytes of the transport addresses.
         * @param addr1     The first address to compare.  The address must be
         *                  associated with the @a transport.  This parameter will never
         *                  be NULL.
         * @param addr2     The second address to compare.  The address must be
         *                  associated with the @a transport.  This parameter will never
         *                  be NULL.
         *
         * @return Indicates how the two transport addresses compare to each other.  The return value is a standard integer, and its exact value is irrelevant (and may be platform dependent).  It is intended to be used only in comparison, as follows:
         *          - Zero ( == 0 ) if addr1 == addr2
         *          - Positive ( > 0 ) if addr1 > addr2
         *          - Negative ( < 0 ) if addr1 < addr2
         */
        int (*AddressCompare)( DTLTransport transport, uint32 size, const void * addr1, const void * addr2 );

        /**
         * This function is called after the library makes a shallow copy of a DOFAddress and its data.  If the transport
         * requires a deeper copy of any data held by this new DOFAddress (e.g., pointers to data that are tied to the lifecycle
         * of the original DOFAddress), then the transport must make the necessary in-place adjustments to the given data. Once
         * this function returns, the copied DOFAddress (and anything to which it refers) must have its own lifecycle.
         *
         * This callback may be NULL if a shallow PCRMem_Copy is sufficient to copy the address data (a deeper copy is not
         * needed).
         *
         * @param transport The Transport.  This parameter will never be NULL.
         * @param size  The size of the address data.
         * @param addr  The shallow copy of the address data. Modifications must be made in-place. This parameter will never be
         *              NULL.
         *
         * @retval TRUE The address was successfully copied.
         * @retval FALSE The address failed to be copied.
         */
        boolean (*AddressCopy)( DTLTransport transport, uint32 size, void * addr );

        /**
         * Destroyer function for an address.  This callback is used to notify the
         * transport (or address creator) that the address is being destroyed.  This
         * allows the transport to clean up the address, if necessary.  The data was
         * originally copied into the DOFAddress and will be freed after this call
         * completes.  However, since the copy was only a shallow copy of the data, the
         * transport may need to perform a deeper free/release of its held resources.
         *
         * This callback may be NULL if no action is needed to free the address data.
         *
         * @param transport The Transport.  This parameter will never be NULL.
         * @param size  The size of the addr information.
         * @param addr  The address data.  This parameter will never be NULL.
         */
        void (*AddressDestroy)( DTLTransport transport, uint32 size, const void * addr );

    } * fns; /* Transport-specific function calls.*/

    /* Implementations may extend this structure. */

} DTLTransport_t;

/**
@}
**/


/** @name DTLTransportHandler
@{
**/

/**
 * The Transport Handler interface structure, allowing a transport to
 * communicate with the OAL.  This interface is implemented by the C OAL.
 *
 * This interface is called by a transport implementation to notify the OAL of
 * opened connections, started servers, and received packets.
 */
typedef struct DTLTransportHandler_t
{

    /**
     * The DTLTransportHandler interface function structure, containing a list of
     * function pointers that implement the interface.  It has the same role as
     * a vtable in C++.
     */
    const struct DTLTransportHandlerFns_t
    {

        /**
         * Report a connection that is now opened.
         * Called by a Transport when a inbound or outbound connection was sucessfully established.
         *
         * This callback must not be NULL.
         *
         * @param handler    The TransportHandler receiving @a connection.  This must not be NULL.
         * @param connection A Transport reprentation of the Connection.  This must not be NULL.
         *
         * @return           TRUE if @a connection was accepted by @a handler, FALSE otherwise.
         *                   There is no need to call Closed if this returns FALSE.
         */
        boolean (*Opened)( DTLTransportHandler handler, DTLConnection connection );

        /**
         * Report a connection that is now closed.
         * Called by a Transport when an inbound or outbound connection was closed.
         * The Transport is permitted to reclaim resources used by the DTLConnection only after this callback returns.
         *
         * This callback must not be NULL.
         *
         * @param handler    The TransportHandler managing @a connection.  This must not be NULL.
         * @param connection A Transport reprentation of the connection.  This must not be NULL.
         * @param except     An exception for the closure/failure of the connection, if any.  NULL indicates it was a normal closure due to Close() being called on the DTLConnection.  Otherwise, it must be a valid DOFErrorException describing the failure.
         */
        void (*Closed)( DTLTransportHandler handler, DTLConnection connection, DOFException except );

        /**
         * Report a Server that is now started.
         * Called by a Transport when a Server is started.
         *
         * This callback must not be NULL.
         *
         * @param handler The TransportHandler managing @a server.  This must not be NULL.
         * @param server  The transport representation of the Server.  This must not be NULL.
         *
         * @return        TRUE if @a server was accepted by @a handler, FALSE otherwise.
         *                There is no need to call Stopped if this returns FALSE.
         */
        boolean (*Started)( DTLTransportHandler handler, DTLServer server );

        /**
         * Report a Server that is now stopped.
         * Called by a Transport when a Server is stopped.
         * The Transport is permitted to reclaim resources used by the DTLConnection only after this callback returns.
         *
         * This callback must not be NULL.
         *
         * @param handler The TransportHandler managing @a server.  This must not be NULL.
         * @param server  The transport representation of the Server.  This must not be NULL.
         * @param except  An exception for the closure/failure of the server, if any.  NULL indicates it was a normal closure due to Stop() being called on the DTLServer.  Otherwise, it must be a valid DOFErrorException describing the failure.
         */
        void (*Stopped)( DTLTransportHandler handler, DTLServer server, DOFException except );

        /**
         * Report a new packet received on a Connection.
         * Called by a Transport when a full packet has been received by a Connection.
         * If the specified connection is asynchronous, then this method will not block.
         *
         * @note <em> Ownership of the packet is transfered to the Transport Handler.
         * </em> The data pointer must be the result of a call to PCRMem_Alloc, and
         * will be freed by the Transport Handler when it is finished with the packet
         * using PCRMem_Free.
         *
         * This callback must not be NULL.
         *
         * @param handler    The TransportHandler associated with Connection.  This must not be NULL.
         * @param connection The connection that received the packet.  This must not be NULL.
         * @param size       The number of bytes in @a packet.  This should be greater than 0.
         * @param packet     The received packet.  This should not be NULL (unless @a size is 0).
         * @param address    The address where the packet came from.  This must not be NULL.
         * @param type       The type of packet received on the connection (a @ref DOFAddressType).
         *
         * @mustfree{PCRMem_Free} packet
         *
         * @return           TRUE if @a handler accepted @a packet, FALSE otherwise.
         */
        boolean (*Received_Connection)( DTLTransportHandler handler, DTLConnection connection, uint32 size, uint8 * packet, DOFAddress address, DOFAddressType type );

        /**
         * Report a new packet received by a server.
         * Called by transport when a full packet has been received by a server.
         * If the specified server is asynchronous, then this method will not block.
         *
         * @note <em> Ownership of the packet is transfered to the Transport Handler.
         * </em> The data pointer must be the result of a call to PCRMem_Alloc, and
         * will be freed by the Transport Handler when it is finished with the packet
         * using PCRMem_Free.
         *
         * This callback must not be NULL.
         *
         * @param handler    The TransportHandler associated with Connection.  This must not be NULL.
         * @param server     The server that received the packet.  This must not be NULL.
         * @param size       The number of bytes in @a packet.  This should be greater than 0.
         * @param packet     The received packet.  This should not be NULL (unless @a size is 0).
         * @param address    The address where the packet came from.  This must not be NULL.
         * @param type       The type of packet received on the server (a @ref DOFAddressType).
         *
         * @mustfree{PCRMem_Free} packet
         *
         * @return           TRUE if @a handler accepted @a packet, FALSE otherwise.
         */
        boolean (*Received_Server)( DTLTransportHandler handler, DTLServer server, uint32 size, uint8 * packet, DOFAddress address, DOFAddressType type );

        /**
         * Determine the total packet length based on received data in a buffer.  If
         * the length cannot be determined with the data provided, then report the
         * number of additional bytes that must be read before a determination may be
         * made.
         *
         * This callback must not be NULL.
         *
         * @param handler The TransportHandler that is asking about the size.  This must not
         *                be NULL.
         * @param buffer  The data that comprises the (possibly partial) packet.
         *                This may be NULL if @a length is zero.
         * @param offset  The offset to the first byte of the PDU. This value is ignored if @a length is zero.
         * @param length  The amount of data past the offset, which may be zero.
         *
         * @return        A negative number to indicate that more data must be read
         *                (the absolute value number of additional bytes should be read), or a
         *                positive number to indicate the actual (total) length of the PDU, or
         *                zero to indicate a catastrophic, unrecoverable error.
         */
        int32 (*GetReceiveSize)( DTLTransportHandler handler, const uint8 * buffer, uint32 offset, uint32 length );

        /**
         * Get maximum initial receive size.  That is, the maximum number of bytes that
         * ever may need to be received before the packet length can be determined.
         *
         * This callback must not be NULL.
         *
         * @param handler   The TransportHandler that will determine the initial read
         *                  size.  This must not be NULL.
         *
         * @return          The size of an initial, priming read that can be used to
         *                  determine the total number of bytes in a full packet.
         */
        uint32 (*GetMaxInitialReceiveSize)( DTLTransportHandler handler );

    } * fns;

    /* Expand here */

} DTLTransportHandler_t;

/**
@}
**/

/**
@}
**/


#ifdef  __cplusplus
}
#endif

#endif /* DOF_TRANSPORT_H */
