/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_TYPE_H
#define DOF_TYPE_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif


/*
**  "Public" Type and Method Declarations
*/

/** @defgroup DOFType DOFType
    @ingroup DOF_Framework

    A DOF data type representation.

    A DOFType can be created using a variety of constructors, or can be
    obtained from an Interface Definition.  There are three main classes of
    DOFTypes: Primitives, Arrays, and Structures.  Primitive DOFTypes often map
    directly to primitive types in a language, such as integers and floating
    point numbers.  Other Primitive DOFTypes in DOF include GUIDs, Object
    IDs, Interface IDs, and BLOBs.  Arrays and Structures are different than
    Primitives in that they contain other DOFTypes.  Array DOFTypes contain a
    single sub-type, the element type.  Structure DOFTypes contain a list of
    sub-types, each of which representing a field of the structure.  In both
    cases, an element may be of any DOFType, including other Arrays and
    Structures.

@{
**/

/* Standard OAL type codes.  These codes are used mostly by DOFType. */
#define DOFTYPE_UINT8          0x00    /**< 8-bit unsigned integer type. **/
#define DOFTYPE_UINT16         0x02    /**< 16-bit unsigned integer type. **/
#define DOFTYPE_UINT32         0x04    /**< 32-bit unsigned integer type. **/
#define DOFTYPE_UINT64         0x06    /**< 64-bit unsigned integer type. **/

#define DOFTYPE_INT8           0x10    /**< 8-bit signed integer type. **/
#define DOFTYPE_INT16          0x12    /**< 16-bit signed integer type. **/
#define DOFTYPE_INT32          0x14    /**< 32-bit signed integer type. **/
#define DOFTYPE_INT64          0x16    /**< 64-bit signed integer type. **/

#define DOFTYPE_FLOAT32        0x20    /**< 32-bit IEEE floating point type. **/
#define DOFTYPE_FLOAT64        0x22    /**< 64-bit IEEE floating point type. **/

#define DOFTYPE_DATETIME       0x40    /**< 64-bit Date/Time type. **/

#define DOFTYPE_BOOLEAN        0x42    /**< Boolean type. **/

#define DOFTYPE_INTERFACEID    0x44    /**< Interface Identifier type. **/

#define DOFTYPE_OBJECTID       0x46    /**< Object Identifier type. **/

#define DOFTYPE_GUID           0x48    /**< Globally Unique Identifier (GUID) type. **/

#define DOFTYPE_STRING         0x31    /**< String type. **/

#define DOFTYPE_BLOB           0x49    /**< Blob (large, opaque, binary data) type. **/

#define DOFTYPE_ARRAY          0x4D    /**< Fixed-size Array type. **/

#define DOFTYPE_REFERENCE      0x51    /**< Reference type (to another, defined type in an interface). **/

#define DOFTYPE_NULLABLE       0x53    /**< Nullable type. **/

#define DOFTYPE_STRUCTURE      0x7801  /**< Structure type. **/

#define DOFTYPE_VOID           0xFFFF  /**< Unknown or invalid type. **/


/* DOFType Methods */

/**
    Create a new Typedef of another type.

    @mustfree{DOFType_Destroy}

    @param type  The actual type to which this reference refers.  If not NULL, this should be a defined type and should not be nullable.  If NULL, a forward reference is created.  Forward references must later be defined via @ref DOFTypeTypedef_DefineAs.

    @return The new DOFType, or NULL if it could not be created.
    @exception NULL is returned if an invalid parameter is detected or insufficient resources exist to create the type.
**/
extern DOF_OAL_DECLSPEC DOFType DOFTypeTypedef_Create( const DOFType type );

/**
    Create a new Nullable Type.
    @mustfree{DOFType_Destroy}

    @param type  The effective type when a value of this type is not the "null" value.  This must not be NULL.

    @return The new DOFType, or NULL if it could not be created.
    @exception NULL is returned if insufficient resources exist to create the type.
**/
extern DOF_OAL_DECLSPEC DOFType DOFTypeNullable_Create( const DOFType type );

/**
 *  Create a new copy of an existing @ref DOFType.
 *  @mustfree{DOFType_Destroy}
 *  @param  self The @ref DOFType to be copied.  This should not be NULL.
 *  @return A new copy of the @ref DOFType.
 *  @exception  NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DOFType DOFType_Copy( DOFType self );

/**
    Destroy the given type.

    @param  type  The DOFType to destroy.  If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFType_Destroy( DOFType type );

/**
    Retrieve the type represented within the @ref DOFType.

    @param  type  The DOFType.  This must not be NULL.

    @return The type represented within the @ref DOFType.
**/
extern DOF_OAL_DECLSPEC DOFTypeID DOFType_GetTypeID( const DOFType type );

/**
    Define a previously created reference type as its real type.
    This can only be done once, and only on a reference type that was created without an actual type (as a forward reference).

    @param  self  The reference DOFType.  This must not be NULL.
    @param  type  The real DOFType to which the reference applies.  This should not be NULL, should be a defined type, and should not be nullable.

    @return TRUE if the type was resolved.
    @exception FALSE is returned if an invalid parameter is detected.
**/
extern DOF_OAL_DECLSPEC boolean DOFTypeTypedef_DefineAs( DOFType self, const DOFType type );

/**
    Report whether or not a type is fully defined.
    Normally, all types are fully defined.  The only exception is a typedef
    type, created without an actual type (as a forward reference), on which
    @ref DOFTypeTypedef_DefineAs has not yet been called.

    @param  type    The DOFType.  This must not be NULL.

    @return TRUE if the type is defined.
    @return FALSE if the type is not defined.
**/
extern DOF_OAL_DECLSPEC boolean DOFType_IsDefined( const DOFType type );

/**
    Report the maximum length of the given type.

    @param  type  The DOFType.  This must not be NULL.

    @return The maximum length of the given type. 0 if type is NULL.
    @exception 0 is returned if the type does not have a length.
**/
extern DOF_OAL_DECLSPEC uint32 DOFType_GetMaxLength( const DOFType type );

/**
    Report the number of subtypes contained within a type.

    @param  type  The DOFType.  This must not be NULL.

    @return The number of subtypes in the type. 0 if type is NULL, or does not have any subtypes.
    @exception 0 is returned if the type does not have any subtypes.
**/
extern DOF_OAL_DECLSPEC uint32 DOFType_GetSubtypesCount( const DOFType type );

/**
    Report the minimum length of the given type.

    @param  type  The DOFType.  This must not be NULL.

    @return The minimum length of the given type. 0 if type is NULL.
    @exception 0 is returned if the type does not have a length.
**/
extern DOF_OAL_DECLSPEC uint32 DOFType_GetMinLength( const DOFType type );

/**
    Report the character encoding defined for a String type.

    @param  type  The DOFType.  This must not be NULL.

    @return The MIBenum value, a unique value identifying a specific coded character set used by the type.
    @exception DOFVALUESTRING_UNKNOWN is returned if the type does not have an encoding.
**/
extern DOF_OAL_DECLSPEC uint16 DOFType_GetEncoding( const DOFType type );

/**
    Retrieve a subtype from a type.

    @param  type  The DOFType.  This must not be NULL.
    @param  index The subtype index to retrieve (0-based).

    @return The specified subtype of the given type, or NULL if the type does not have subtypes or the specified index is out of range.
    @exception NULL is returned if the type does not have the requested subtype.
**/
extern DOF_OAL_DECLSPEC DOFType DOFType_GetSubtype( const DOFType type, uint32 index );

/**
    Retrieve the subtypes from a type.

    @param  type  The DOFType.  This must not be NULL.

    @return The array of subtypes of the given type, or NULL if the type does not have subtypes.  The count can be determined via @ref DOFType_GetSubtypesCount.
    @exception NULL is returned if the type does not have subtypes.
**/
extern DOF_OAL_DECLSPEC const DOFType * DOFType_GetSubtypes( const DOFType type );

/**
@}
**/


#ifdef  __cplusplus
}
#endif

#endif /* DOF_TYPE_H */
