/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_VALUE_H
#define DOF_VALUE_H

#include <dof/core.h>
#include <dof/value_gen.h>

#ifdef  __cplusplus
extern "C" {
#endif


/*
**  "Public" Type and Method Declarations
*/

/** @defgroup DOFValue DOFValue
    @ingroup DOF_Framework

    A generic container for any DOF Data Value.  This is the base class
    for any constructed value, including DOFValueArray and DOFValueStructure, and is
    used as the value for a properties and as parameters for methods,
    events, and exceptions.

    Each individual datatype is effectively broken into its own class, with a
    constructor and accessor methods specific to that class.  Because the C
    language does not support type inheritance, the constructors and accessor
    methods use only the generic DOFValue type.  Care should be taken when
    utilizing these methods so that only methods that apply to the specific
    type are used for that type (for example, an application should not
    attempt to get the string length of a UInt8).

    There are three main groups of values: primitive types, arrays, and
    structures.  Primitive types contain a single value, and generally
    match native and/or common C language types, such as integers and
    strings.  Arrays contain 0 or more elements of a single type.
    Structures contain 0 or more fields, each of a specific type.  The types
    contained by arrays and structures can be any type (including other
    arrays and structures).

@{
**/


/**
    @name Primitive Types

    Primitive types typically map to primitive types in a language, such as
    integers and floating point numbers.  Other Primitive types in DOF
    include GUIDs, Object IDs, Interface IDs, and BLOBs.

    Primitive DOFValues are created with a specific value, and remain
    immutable (they cannot change) for their lifetime.

    Each primitive class also defines a DOFType that describes the type.
    All primitive types are able to construct a DOFType.  Most primitive types
    also have a predefined, static DOFType that can be used; the remainder
    require additional information, and do not easily lend themselves to a
    predefined type.

@{
**/

#define DOFVALUESTRING_UNKNOWN     2  /**< Unknown encoding. **/
#define DOFVALUESTRING_US_ASCII    3  /**< US-ASCII encoding. **/
#define DOFVALUESTRING_UTF_8     106  /**< UTF-8 encoding. **/

extern DOF_OAL_DECLSPEC DOFType DOFVALUEINT8_TYPE;          /**< Pre-defined 8-bit signed Integer @ref DOFType. **/
extern DOF_OAL_DECLSPEC DOFType DOFVALUEINT16_TYPE;         /**< Pre-defined 16-bit signed Integer @ref DOFType. **/
extern DOF_OAL_DECLSPEC DOFType DOFVALUEINT32_TYPE;         /**< Pre-defined 32-bit signed Integer @ref DOFType. **/
extern DOF_OAL_DECLSPEC DOFType DOFVALUEINT64_TYPE;         /**< Pre-defined 64-bit signed Integer @ref DOFType. **/
extern DOF_OAL_DECLSPEC DOFType DOFVALUEUINT8_TYPE;         /**< Pre-defined 8-bit unsigned Integer @ref DOFType. **/
extern DOF_OAL_DECLSPEC DOFType DOFVALUEUINT16_TYPE;        /**< Pre-defined 16-bit unsigned Integer @ref DOFType. **/
extern DOF_OAL_DECLSPEC DOFType DOFVALUEUINT32_TYPE;        /**< Pre-defined 32-bit unsigned Integer @ref DOFType. **/
extern DOF_OAL_DECLSPEC DOFType DOFVALUEUINT64_TYPE;        /**< Pre-defined 64-bit unsigned Integer @ref DOFType. **/
extern DOF_OAL_DECLSPEC DOFType DOFVALUEFLOAT32_TYPE;       /**< Pre-defined 32-bit Floating Point @ref DOFType. **/
extern DOF_OAL_DECLSPEC DOFType DOFVALUEFLOAT64_TYPE;       /**< Pre-defined 64-bit Floating Point @ref DOFType. **/
extern DOF_OAL_DECLSPEC DOFType DOFVALUEDATETIME_TYPE;      /**< Pre-defined Date-Time @ref DOFType. **/
extern DOF_OAL_DECLSPEC DOFType DOFVALUEBOOLEAN_TYPE;       /**< Pre-defined Boolean @ref DOFType. **/
extern DOF_OAL_DECLSPEC DOFType DOFVALUEINTERFACEID_TYPE;   /**< Pre-defined @ref DOFInterfaceID @ref DOFType. **/
extern DOF_OAL_DECLSPEC DOFType DOFVALUEOBJECTID_TYPE;      /**< Pre-defined @ref DOFObjectID @ref DOFType. **/
extern DOF_OAL_DECLSPEC DOFType DOFVALUEGUID_TYPE;          /**< Pre-defined GUID @ref DOFType. **/

/* DOFValue Primitive Methods */

/**
    Create a new 8-bit Signed Integer Type.

    @mustfree{DOFType_Destroy}

    @return The new DOFType, or NULL if it could not be created.

**/
extern DOF_OAL_DECLSPEC DOFType DOFValueInt8Type_Create( void );

/**
    Create a new 16-bit Signed Integer Type.

    @mustfree{DOFType_Destroy}

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueInt16Type_Create( void );

/**
    Create a new 32-bit Signed Integer Type.

    @mustfree{DOFType_Destroy}

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueInt32Type_Create( void );

/**
    Create a new 64-bit Signed Integer Type.

    @mustfree{DOFType_Destroy}

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueInt64Type_Create( void );

/**
    Create a new 8-bit Unsigned Integer Type.

    @mustfree{DOFType_Destroy}

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueUInt8Type_Create( void );

/**
    Create a new 16-bit Unsigned Integer Type.

    @mustfree{DOFType_Destroy}

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueUInt16Type_Create( void );

/**
    Create a new 32-bit Unsigned Integer Type.

    @mustfree{DOFType_Destroy}

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueUInt32Type_Create( void );

/**
    Create a new 64-bit Unsigned Integer Type.

    @mustfree{DOFType_Destroy}

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueUInt64Type_Create( void );

/**
    Create a new 32-bit Floating-Point Type.

    @mustfree{DOFType_Destroy}

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueFloat32Type_Create( void );

/**
    Create a new 64-bit Floating-Point Type.

    @mustfree{DOFType_Destroy}

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueFloat64Type_Create( void );

/**
    Create a new Date-Time Type.

    @mustfree{DOFType_Destroy}

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueDateTimeType_Create( void );

/**
    Create a new Boolean Type.

    @mustfree{DOFType_Destroy}

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueBooleanType_Create( void );

/**
    Create a new GUID Type.

    @mustfree{DOFType_Destroy}

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueGUIDType_Create( void );

/**
    Create a new Interface ID Type.

    @mustfree{DOFType_Destroy}

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueInterfaceIDType_Create( void );

/**
    Create a new Object ID Type.

    @mustfree{DOFType_Destroy}

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueObjectIDType_Create( void );

/**
    Create a new String Type.
    All string types must include an encoding.

    @mustfree{DOFType_Destroy}

    @param  encoding  The MIBenum value, a unique value identifying a specific coded character set used by all strings in @a val.  This should be > 2.  Example: @ref DOFVALUESTRING_US_ASCII.
    @param  length    Maximum string length, in characters (not necessarily byte size).

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueStringType_Create( uint16 encoding, uint32 length );

/**
    Create a new BLOB Type.
    @mustfree{DOFType_Destroy}

    If the @a minSize and @a maxSize are the same, a fixed-size BLOB type is created.  Otherwise, a variable-size BLOB is created.

    @param minLength  The minimum size of this type, in bytes.  This should be less than @a maxLength.
    @param maxLength  The maximum size of this type, in bytes.  This should be greater than @a minLength.

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueBLOBType_Create( uint32 minLength, uint32 maxLength );

/**
@}
**/


/** @defgroup DOFValueArray DOFValueArray

    A DOF Array Data Value.

    A DOFValueArray contains an array of DOFValues, and handles mappings to/from
    arrays of native language types.  Any valid DOFValue can be contained in
    the array, including a @ref DOFValueStructure or another @ref DOFValueArray.  Direct
    language mappings are not provided for these nested types, but creation and
    access is availabe through the generic DOFValueArray functions (by determining
    the array length and obtaining the DOFValue for each element individually).

    @see DOFValue

@{
**/

/* DOFValueArray Methods */

/**
    Create a new Array Type.
    @mustfree{DOFType_Destroy}

    If the @a minLength and @a maxLength are the same, a fixed-length Array type is created.  Otherwise, a variable-sized Array type is created.

    @param subtype    The type of the Array elements.  This should not be NULL.
    @param minLength  The minimum number of elements in the Array.  This should be less than @a maxLength.
    @param maxLength  The maximum number of elements in the Array.  This should be greater than @a minLength.

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueArrayType_Create( const DOFType subtype, uint32 minLength, uint32 maxLength );

/**
    Create a new @ref DOFValueArray from a list of @ref DOFValue elements.
    The element type of the resulting array will be inferred based on the given elements.
    If the elments are not compatible with each other, the array will not be constructed.

    @mustfree{DOFValue_Destroy}

    @param elementCount  The number of elements in the array.  This may be 0, and should be less than 32768.
    @param elements      The array of @ref DOFValue elements to put in the @ref DOFValueArray.  This must not be NULL if @a elementCount > 0.

    @return A new DOFValueArray containing the given elements.

    @exception NULL is returned if insufficient resources are available to
    create the DOFValueArray, if any of the elements are not compatible with the
    specified @a elementType, of an illegal argument is detected.
**/
extern DOF_OAL_DECLSPEC DOFValue DOFValueArray_Create( uint16 elementCount, const DOFValue elements[] );

/**
    Create a new @ref DOFValueArray from a known array type and a list of @ref DOFValue elements.
    The number of elements is checked against the allowable length of the given array @a type.
    The elements are checked against the element type of the given array @a type.
    If either of these checks fail, the array will not be constructed.

    @mustfree{DOFValue_Destroy}

    @param type          The type of the resulting array.  This must not be NULL.
    @param elementCount  The number of elements in the array.  This may be 0, and should be less than 32768.
    @param elements      The array of @ref DOFValue elements to put in the @ref DOFValueArray.  This must not be NULL if @a elementCount > 0.

    @return A new DOFValueArray containing the given elements.

    @exception NULL is returned if insufficient resources are available to
    create the DOFValueArray, if any of the elements are not compatible with the
    specified @a elementType, of an illegal argument is detected.
**/
extern DOF_OAL_DECLSPEC DOFValue DOFValueArray_Create_Type( const DOFType type, uint16 elementCount, const DOFValue elements[] );

/**
    Report the number of elements in the DOFValueArray.

    @param  self  The DOFValueArray (a DOFValue).  This must not be NULL.

    @return The number of elements in the DOFValueArray.  This may be 0 if the array contains no elements.

    @exception 0 is returned if @a self is NULL or it is not a @ref DOFValueArray.
**/
extern DOF_OAL_DECLSPEC uint16 DOFValueArray_GetElementsCount( const DOFValue self );

/**
    Get a @ref DOFValue element from the @ref DOFValueArray.

    @note The returned data is owned by the encapsulating @ref DOFValueArray, and is
    only valid in the context of that DOFValueArray.  It must not be freed by the
    caller.  Once the DOFValueArray is freed, the DOFValue is no longer valid.  If
    the DOFValue is needed for longer than the lifetime of the DOFValueArray, it
    must be copied manually.

    @param  self   The @ref DOFValueArray in which to retrieve an element.  This must not be NULL.
    @param  index  The element index to be retrieved from @a self.

    @return The DOFValue element of the DOFValueArray.

    @exception NULL is returned if @a index specifies an element outside the
    bounds of the array, or if @a self is NULL or its type is not a @ref
    DOFValueArray.
**/
extern DOF_OAL_DECLSPEC DOFValue DOFValueArray_GetElement( const DOFValue self, uint16 index );

/**
    Get a reference to the encapsulated DOFValue array.

    This method provides low-level access to the actual array of DOFValues
    contained in the DOFValueArray.

    @note The returned data is owned by the encapsulating @ref DOFValueArray, and is
    only valid in the context of that DOFValueArray.  It must not be freed by the
    caller.  Once the DOFValueArray is freed, the DOFValue array and all the
    contained DOFValues are no longer valid.  If the array itself and/or
    DOFValues are needed for longer than the lifetime of the DOFValueArray, they
    must be copied manually.

    @param  self  The @ref DOFValueArray.  This must not be NULL.

    @return The array of DOFValues contained within the @ref DOFValueArray.

    @exception NULL is returned if @a self is NULL or its type is not a @ref DOFValueArray.
**/
extern DOF_OAL_DECLSPEC const DOFValue * DOFValueArray_GetElements( const DOFValue self );

/**
@}
**/

/** @defgroup DOFValueStructure DOFValueStructure

    A DOF Structure Data Value.

    A DOFValueStructure contains an array of DOFValues, representing fields of the
    structure.  Any valid DOFValue can be contained, including a @ref
    DOFValueArray or another @ref DOFValueStructure.

    @see DOFValue

@{
**/

/* DOFValueStructure Methods */

/**
    Create a new Structure Type.
    @mustfree{DOFType_Destroy}

    @param fieldCount  The number of fields in the Structure.
    @param fields  An array of DOFTypes, representing the field types for the Structure.  This must not be NULL if @a fieldCount > 0.

    @return The new DOFType, or NULL if it could not be created.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValueStructureType_Create( uint16 fieldCount, const DOFType fields[] );

/**
    Create a new @ref DOFValueStructure from a list of @ref DOFValue fields.

    @mustfree{DOFValue_Destroy}

    @param fieldCount  The number of fields in the structure.  This may be 0, and should be less than 32768.
    @param fields      The array of @ref DOFValue fields to put in the @ref DOFValueStructure.  This must not be NULL if @a fieldCount > 0.

    @return A new DOFValueStructure containing the given fields.

    @exception NULL is returned if insufficient resources are available to
    create the DOFValueStructure, or @a fields is NULL.
**/
extern DOF_OAL_DECLSPEC DOFValue DOFValueStructure_Create( uint16 fieldCount, const DOFValue fields[] );

/**
    Create a new @ref DOFValueStructure from a list of @ref DOFValue fields.
    The number of fields is determined by the given structure @a type.

    @mustfree{DOFValue_Destroy}

    @param type        The type of the structure to be created.  This should not be NULL.
    @param fields      The array of @ref DOFValue fields to put in the @ref DOFValueStructure (the number of fields expected is determined by the @a type).  This must not be NULL if the structure contains any fields.

    @return A new DOFValueStructure containing the given fields.

    @exception NULL is returned if insufficient resources are available to
    create the DOFValueStructure, if any of the fields are not compatible with the
    specified @a type, or if @a type or @a fields is NULL.
**/
extern DOF_OAL_DECLSPEC DOFValue DOFValueStructure_Create_Type( const DOFType type, const DOFValue fields[] );

/**
    Report the number of fields in the DOFValueStructure.

    @param  self  The DOFValueStructure (a DOFValue).  This must not be NULL.

    @return The number of fields in the DOFValueStructure.

    @exception 0 is returned if @a self is NULL or it is not a @ref DOFValueStructure.
**/
extern DOF_OAL_DECLSPEC uint16 DOFValueStructure_GetFieldsCount( const DOFValue self );

/**
    Get a @ref DOFValue field from the @ref DOFValueStructure.

    @note The returned data is owned by the encapsulating @ref DOFValueStructure,
    and is only valid in the context of that DOFValueStructure.  It must not be
    freed by the caller.  Once the DOFValueStructure is freed, the DOFValue is no
    longer valid.  If the DOFValue is needed for longer than the lifetime of
    the DOFValueStructure, it must be copied manually.

    @param  self   The @ref DOFValueStructure in which to retrieve a field.  This must not be NULL.
    @param  index  The field index to be retrieved from @a self.

    @return The DOFValue field of the DOFValueStructure.

    @exception NULL is returned if @a index does not specify a valid field of
    the structure, or if @a self is NULL or its type is not a @ref DOFValueStructure.
**/
extern DOF_OAL_DECLSPEC DOFValue DOFValueStructure_GetField( const DOFValue self, uint16 index );

/**
    Get a reference to the encapsulated DOFValue array.

    This method provides low-level access to the actual array of DOFValues
    contained in the DOFValueStructure.

    @note The returned data is owned by the encapsulating @ref DOFValueStructure, and is
    only valid in the context of that DOFValueStructure.  It must not be freed by the
    caller.  Once the DOFValueStructure is freed, the DOFValue array and all the
    contained DOFValues are no longer valid.  If the array itself and/or
    DOFValues are needed for longer than the lifetime of the DOFValueStructure, they
    must be copied manually.

    @param  self  The @ref DOFValueStructure.  This must not be NULL.

    @return The array of DOFValues contained within the @ref DOFValueStructure.

    @exception NULL is returned if @a self is NULL or its type is not a @ref DOFValueStructure.
**/
extern DOF_OAL_DECLSPEC const DOFValue * DOFValueStructure_GetFields( const DOFValue self );

/**
@}
**/

/* DOFValue Methods */

/**
    Destroy the given value.

    @param  self   The Value to destroy. If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFValue_Destroy( DOFValue self );

/**
    Destroy the given array of values.

    @param  valuesCount The count of values in the @a values array to destroy.  If 0, no action is taken.
    @param  values   The array of values to destroy.  If NULL, no action is taken.
**/
extern DOF_OAL_DECLSPEC void DOFValueList_Destroy( uint16 valuesCount, DOFValue values[] );

/**
 *  Create a new copy of an existing @ref DOFValue.
 *  @mustfree{DOFValue_Destroy}
 *  @param  self The @ref DOFValue to be copied.  This should not be NULL.
 *  @return A new copy of the @ref DOFValue.
 *  @exception  NULL is returned if @a self is NULL, or there are insufficient resources to create a copy.
 */
extern DOF_OAL_DECLSPEC DOFValue DOFValue_Copy( const DOFValue self );

/**
    Retrieve the @ref DOFType represented within the @ref DOFValue.

    @param  self   The DOFValue.  This must not be NULL.

    @return The @ref DOFType of the @ref DOFValue.

    @exception NULL is returned if @a self is NULL.
**/
extern DOF_OAL_DECLSPEC DOFType DOFValue_GetDOFType( const DOFValue self );


/**
    Determine if a value's type is compatible with another type.

    Compatibility is not commutative, but is directional.  In other words, Type
    A may be compatible with Type B, but not vice versa.  An example of this is
    comparing a 20-element Fixed-Length Array Type (Type A) with a 50-element
    Variable Length Array Type (Type B).  In this example, Type A is compatible
    with Type B because a 20-element array is within the parameters of a
    50-element Variable-Length Array.  However, Type B is not compatible with
    Type A, because the maximum (and minimum) length of Type B exceeds the
    specification of Type A.

    Array element types or Structure Field types (including order) must also be
    compatible for their parent Types to be compatible.

    If both the type and value are NULL, they are considered compatible.

    @param  value The value to be compared (containing Type A, possibly a more specific type).  This may be NULL if there is no value (compatible with some types, such as NULLABLE).
    @param  type The type to compare against (Type B, possibly a more generic type).  This should not be NULL.

    @return TRUE if the types are compatible; FALSE otherwise.
**/
extern DOF_OAL_DECLSPEC boolean DOFValue_IsCompatibleWith( DOFValue value, DOFType type );


/**
@}
**/


#ifdef  __cplusplus
}
#endif

#endif /* DOF_VALUE_H */
