/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef BN_H_
#define BN_H_

#include <dof/core.h>
#include <internal/bignum_class.h>

/* 
 * This library was originally downloaded from libtom.org 2/15/2013. libtommath was originally licensed under WTFPL.
 * The library was modified, renamed and relicensed.
 */

#ifndef MIN
   #define MIN(x,y) ((x)<(y)?(x):(y))
#endif

#ifndef MAX
   #define MAX(x,y) ((x)>(y)?(x):(y))
#endif

#ifdef __cplusplus
extern "C" {

/* C++ compilers don't like assigning void * to mp_digit * */
#define  OPT_CAST(x)  (x *)

#else

/* C on the other hand doesn't care */
#define  OPT_CAST(x)

#endif


/* detect 64-bit mode if possible */
/* Since this mode requires a gcc compiler, a check for a gcc compiler should be added if still desired.*/
#if 0 
#if defined(__x86_64__) 
   #if !(defined(MP_64BIT) && defined(MP_16BIT) && defined(MP_8BIT))
      #define MP_64BIT 
   #endif
#endif
#endif

/* some default configurations.
 *
 * A "mp_digit" must be able to hold DIGIT_BIT + 1 bits
 * A "mp_word" must be able to hold 2*DIGIT_BIT + 1 bits
 *
 * At the very least a mp_digit must be able to hold 7 bits
 * [any size beyond that is ok provided it doesn't overflow the data type]
 */
#ifdef MP_8BIT
   typedef uint8      mp_digit;
   typedef uint16     mp_word;
#elif defined(MP_16BIT)
   typedef uint16      mp_digit;
   typedef uint32      mp_word;
#elif defined(MP_64BIT)
   /* GCC specific code */
   typedef uint64  ulong64;
   typedef int64   long64;

   typedef uint64           mp_digit;
   typedef unsigned long    mp_word __attribute__ ((mode(TI))); /* Tetra Integer Mode represents a sixteen-byte integer. */

   #define DIGIT_BIT          60
#else
   /* this is the default case, 28-bit digits */
   typedef uint64  ulong64;
   typedef int64   long64;

   typedef uint32   mp_digit;
   typedef ulong64  mp_word;

#ifdef MP_31BIT   
   /* this is an extension that uses 31-bit digits */
   #define DIGIT_BIT          31
#else
   /* default case is 28-bit digits, defines MP_28BIT as a handy macro to test */
   #define DIGIT_BIT          28
   #define MP_28BIT
#endif
   
#endif

/* define heap macros */
#ifndef XMALLOC 
   #define XMALLOC  PCRMem_Alloc
   #define XFREE    PCRMem_Free
   #define XREALLOC PCRMem_Realloc
   /*#define XCALLOC  calloc*/ /* This is not being used and we do not have it in the PCR */
#endif

/* Review: This is not available in our system currently. Can be found in limits.h it could be added with configure. */
#ifndef CHAR_BIT
    #define CHAR_BIT      8
#endif

/* otherwise the bits per digit is calculated automatically from the size of a mp_digit */
#ifndef DIGIT_BIT
   #define DIGIT_BIT     ((int32)((CHAR_BIT * sizeof(mp_digit) - 1)))  /* bits per digit */
#endif

#define MP_DIGIT_BIT     DIGIT_BIT
#define MP_MASK          ((((mp_digit)1)<<((mp_digit)DIGIT_BIT))-((mp_digit)1))
#define MP_DIGIT_MAX     MP_MASK

/* equalities */
#define MP_LT        -1   /* less than */
#define MP_EQ         0   /* equal to */
#define MP_GT         1   /* greater than */

#define MP_ZPOS       0   /* positive integer */
#define MP_NEG        1   /* negative */

#define MP_OKAY       0   /* ok result */
#define MP_MEM        -2  /* out of mem */
#define MP_VAL        -3  /* invalid input */
#define MP_RANGE      MP_VAL

#define MP_YES        1   /* yes response */
#define MP_NO         0   /* no response */

/* Primality generation flags */
#define LTM_PRIME_BBS      0x0001 /* BBS style prime */
#define LTM_PRIME_SAFE     0x0002 /* Safe prime (p-1)/2 == prime */
#define LTM_PRIME_2MSB_ON  0x0008 /* force 2nd MSB to 1 */

typedef int32           mp_err;

/* you'll have to tune these... */
extern int32 KARATSUBA_MUL_CUTOFF,
             KARATSUBA_SQR_CUTOFF,
             TOOM_MUL_CUTOFF,
             TOOM_SQR_CUTOFF;

/* define this to use lower memory usage routines (exptmods mostly) */
/* #define MP_LOW_MEM */

/* default precision */
#ifndef MP_PREC
   #ifndef MP_LOW_MEM
      #define MP_PREC                 32     /* default digits of precision */
   #else
      #define MP_PREC                 8      /* default digits of precision */
   #endif   
#endif

/* size of comba arrays, should be at least 2 * 2**(BITS_PER_WORD - BITS_PER_DIGIT*2) */
#define MP_WARRAY               (1 << (sizeof(mp_word) * CHAR_BIT - 2 * DIGIT_BIT + 1))

/* the infamous mp_int structure */
typedef struct  {
    int32 used, alloc, sign;
    mp_digit *dp;
} mp_int;

/* callback for mp_prime_random, should fill dst with random bytes and return how many read [upto len] */
typedef int32 ltm_prime_callback(unsigned char *dst, int32 len, void *dat);


#define USED(m)    ((m)->used)
#define DIGIT(m,k) ((m)->dp[(k)])
#define SIGN(m)    ((m)->sign)

/* error code to char* string */
char *mp_error_to_string(int32 code);

/* ---> init and deinit bignum functions <--- */
/* init a bignum */
int32 mp_init(mp_int *a);

/* free a bignum */
void mp_clear(mp_int *a);

#ifdef VARARGS_ALLOWED
/* init a null terminated series of arguments */
int32 mp_init_multi(mp_int *mp, ...);

/* clear a null terminated series of arguments */
void mp_clear_multi(mp_int *mp, ...);
#endif

/* exchange two ints */
void mp_exch(mp_int *a, mp_int *b);

/* shrink ram required for a bignum */
int32 mp_shrink(mp_int *a);

/* grow an int32 to a given size */
int32 mp_grow(mp_int *a, int32 size);

/* init to a given number of digits */
int32 mp_init_size(mp_int *a, int32 size);

/* ---> Basic Manipulations <--- */
#define mp_iszero(a) (((a)->used == 0) ? MP_YES : MP_NO)
#define mp_iseven(a) (((a)->used > 0 && (((a)->dp[0] & 1) == 0)) ? MP_YES : MP_NO)
#define mp_isodd(a)  (((a)->used > 0 && (((a)->dp[0] & 1) == 1)) ? MP_YES : MP_NO)

/* set to zero */
void mp_zero(mp_int *a);

/* set to a digit */
void mp_set(mp_int *a, mp_digit b);

/* set a 32-bit const */
int32 mp_set_int(mp_int *a, unsigned long b);

/* get a 32-bit value */
unsigned long mp_get_int(mp_int * a);

/* initialize and set a digit */
int32 mp_init_set (mp_int * a, mp_digit b);

/* initialize and set 32-bit value */
int32 mp_init_set_int (mp_int * a, unsigned long b);

/* copy, b = a */
int32 mp_copy(mp_int *a, mp_int *b);

/* inits and copies, a = b */
int32 mp_init_copy(mp_int *a, mp_int *b);

/* trim unused digits */
void mp_clamp(mp_int *a);

/* ---> digit manipulation <--- */

/* right shift by "b" digits */
void mp_rshd(mp_int *a, int32 b);

/* left shift by "b" digits */
int32 mp_lshd(mp_int *a, int32 b);

/* c = a / 2**b */
int32 mp_div_2d(mp_int *a, int32 b, mp_int *c, mp_int *d);

/* b = a/2 */
int32 mp_div_2(mp_int *a, mp_int *b);

/* c = a * 2**b */
int32 mp_mul_2d(mp_int *a, int32 b, mp_int *c);

/* b = a*2 */
int32 mp_mul_2(mp_int *a, mp_int *b);

/* c = a mod 2**d */
int32 mp_mod_2d(mp_int *a, int32 b, mp_int *c);

/* computes a = 2**b */
int32 mp_2expt(mp_int *a, int32 b);

/* Counts the number of lsbs which are zero before the first zero bit */
int32 mp_cnt_lsb(mp_int *a);

/* I Love Earth! */

/* makes a pseudo-random int32 of a given size */
int32 mp_rand(mp_int *a, int32 digits);

/* ---> binary operations <--- */
/* c = a XOR b  */
int32 mp_xor(mp_int *a, mp_int *b, mp_int *c);

/* c = a OR b */
int32 mp_or(mp_int *a, mp_int *b, mp_int *c);

/* c = a AND b */
int32 mp_and(mp_int *a, mp_int *b, mp_int *c);

/* ---> Basic arithmetic <--- */

/* b = -a */
int32 mp_neg(mp_int *a, mp_int *b);

/* b = |a| */
int32 mp_abs(mp_int *a, mp_int *b);

/* compare a to b */
int32 mp_cmp(mp_int *a, mp_int *b);

/* compare |a| to |b| */
int32 mp_cmp_mag(mp_int *a, mp_int *b);

/* c = a + b */
int32 mp_add(mp_int *a, mp_int *b, mp_int *c);

/* c = a - b */
int32 mp_sub(mp_int *a, mp_int *b, mp_int *c);

/* c = a * b */
int32 mp_mul(mp_int *a, mp_int *b, mp_int *c);

/* b = a*a  */
int32 mp_sqr(mp_int *a, mp_int *b);

/* a/b => cb + d == a */
int32 mp_div(mp_int *a, mp_int *b, mp_int *c, mp_int *d);

/* c = a mod b, 0 <= c < b  */
int32 mp_mod(mp_int *a, mp_int *b, mp_int *c);

/* ---> single digit functions <--- */

/* compare against a single digit */
int32 mp_cmp_d(mp_int *a, mp_digit b);

/* c = a + b */
int32 mp_add_d(mp_int *a, mp_digit b, mp_int *c);

/* c = a - b */
int32 mp_sub_d(mp_int *a, mp_digit b, mp_int *c);

/* c = a * b */
int32 mp_mul_d(mp_int *a, mp_digit b, mp_int *c);

/* a/b => cb + d == a */
int32 mp_div_d(mp_int *a, mp_digit b, mp_int *c, mp_digit *d);

/* a/3 => 3c + d == a */
int32 mp_div_3(mp_int *a, mp_int *c, mp_digit *d);

/* c = a**b */
int32 mp_expt_d(mp_int *a, mp_digit b, mp_int *c);

/* c = a mod b, 0 <= c < b  */
int32 mp_mod_d(mp_int *a, mp_digit b, mp_digit *c);

/* ---> number theory <--- */

/* d = a + b (mod c) */
int32 mp_addmod(mp_int *a, mp_int *b, mp_int *c, mp_int *d);

/* d = a - b (mod c) */
int32 mp_submod(mp_int *a, mp_int *b, mp_int *c, mp_int *d);

/* d = a * b (mod c) */
int32 mp_mulmod(mp_int *a, mp_int *b, mp_int *c, mp_int *d);

/* c = a * a (mod b) */
int32 mp_sqrmod(mp_int *a, mp_int *b, mp_int *c);

/* c = 1/a (mod b) */
int32 mp_invmod(mp_int *a, mp_int *b, mp_int *c);

/* c = (a, b) */
int32 mp_gcd(mp_int *a, mp_int *b, mp_int *c);

/* produces value such that U1*a + U2*b = U3 */
int32 mp_exteuclid(mp_int *a, mp_int *b, mp_int *U1, mp_int *U2, mp_int *U3);

/* c = [a, b] or (a*b)/(a, b) */
int32 mp_lcm(mp_int *a, mp_int *b, mp_int *c);

/* finds one of the b'th root of a, such that |c|**b <= |a|
 *
 * returns error if a < 0 and b is even
 */
int32 mp_n_root(mp_int *a, mp_digit b, mp_int *c);

/* special sqrt algo */
int32 mp_sqrt(mp_int *arg, mp_int *ret);

/* is number a square? */
int32 mp_is_square(mp_int *arg, int32 *ret);

/* computes the jacobi c = (a | n) (or Legendre if b is prime)  */
int32 mp_jacobi(mp_int *a, mp_int *n, int32 *c);

/* used to setup the Barrett reduction for a given modulus b */
int32 mp_reduce_setup(mp_int *a, mp_int *b);

/* Barrett Reduction, computes a (mod b) with a precomputed value c
 *
 * Assumes that 0 < a <= b*b, note if 0 > a > -(b*b) then you can merely
 * compute the reduction as -1 * mp_reduce(mp_abs(a)) [pseudo code].
 */
int32 mp_reduce(mp_int *a, mp_int *b, mp_int *c);

/* setups the montgomery reduction */
int32 mp_montgomery_setup(mp_int *a, mp_digit *mp);

/* computes a = B**n mod b without division or multiplication useful for
 * normalizing numbers in a Montgomery system.
 */
int32 mp_montgomery_calc_normalization(mp_int *a, mp_int *b);

/* computes x/R == x (mod N) via Montgomery Reduction */
int32 mp_montgomery_reduce(mp_int *a, mp_int *m, mp_digit mp);

/* returns 1 if a is a valid DR modulus */
int32 mp_dr_is_modulus(mp_int *a);

/* sets the value of "d" required for mp_dr_reduce */
void mp_dr_setup(mp_int *a, mp_digit *d);

/* reduces a modulo b using the Diminished Radix method */
int32 mp_dr_reduce(mp_int *a, mp_int *b, mp_digit mp);

/* returns true if a can be reduced with mp_reduce_2k */
int32 mp_reduce_is_2k(mp_int *a);

/* determines k value for 2k reduction */
int32 mp_reduce_2k_setup(mp_int *a, mp_digit *d);

/* reduces a modulo b where b is of the form 2**p - k [0 <= a] */
int32 mp_reduce_2k(mp_int *a, mp_int *n, mp_digit d);

/* returns true if a can be reduced with mp_reduce_2k_l */
int32 mp_reduce_is_2k_l(mp_int *a);

/* determines k value for 2k reduction */
int32 mp_reduce_2k_setup_l(mp_int *a, mp_int *d);

/* reduces a modulo b where b is of the form 2**p - k [0 <= a] */
int32 mp_reduce_2k_l(mp_int *a, mp_int *n, mp_int *d);

/* d = a**b (mod c) */
int32 mp_exptmod(mp_int *a, mp_int *b, mp_int *c, mp_int *d);

/* ---> Primes <--- */

/* number of primes */
#ifdef MP_8BIT
   #define PRIME_SIZE      31
#else
   #define PRIME_SIZE      256
#endif

/* table of first PRIME_SIZE primes */
extern const mp_digit ltm_prime_tab[];

/* result=1 if a is divisible by one of the first PRIME_SIZE primes */
int32 mp_prime_is_divisible(mp_int *a, int32 *result);

/* performs one Fermat test of "a" using base "b".
 * Sets result to 0 if composite or 1 if probable prime
 */
int32 mp_prime_fermat(mp_int *a, mp_int *b, int32 *result);

/* performs one Miller-Rabin test of "a" using base "b".
 * Sets result to 0 if composite or 1 if probable prime
 */
int32 mp_prime_miller_rabin(mp_int *a, mp_int *b, int32 *result);

/* This gives [for a given bit size] the number of trials required
 * such that Miller-Rabin gives a prob of failure lower than 2^-96 
 */
int32 mp_prime_rabin_miller_trials(int32 size);

/* performs t rounds of Miller-Rabin on "a" using the first
 * t prime bases.  Also performs an initial sieve of trial
 * division.  Determines if "a" is prime with probability
 * of error no more than (1/4)**t.
 *
 * Sets result to 1 if probably prime, 0 otherwise
 */
int32 mp_prime_is_prime(mp_int *a, int32 t, int32 *result);

/* finds the next prime after the number "a" using "t" trials
 * of Miller-Rabin.
 *
 * bbs_style = 1 means the prime must be congruent to 3 mod 4
 */
int32 mp_prime_next_prime(mp_int *a, int32 t, int32 bbs_style);

/* makes a truly random prime of a given size (bytes),
 * call with bbs = 1 if you want it to be congruent to 3 mod 4 
 *
 * You have to supply a callback which fills in a buffer with random bytes.  "dat" is a parameter you can
 * have passed to the callback (e.g. a state or something).  This function doesn't use "dat" itself
 * so it can be NULL
 *
 * The prime generated will be larger than 2^(8*size).
 */
#define mp_prime_random(a, t, size, bbs, cb, dat) mp_prime_random_ex(a, t, ((size) * 8) + 1, (bbs==1)?LTM_PRIME_BBS:0, cb, dat)

/* makes a truly random prime of a given size (bits),
 *
 * Flags are as follows:
 * 
 *   LTM_PRIME_BBS      - make prime congruent to 3 mod 4
 *   LTM_PRIME_SAFE     - make sure (p-1)/2 is prime as well (implies LTM_PRIME_BBS)
 *   LTM_PRIME_2MSB_OFF - make the 2nd highest bit zero
 *   LTM_PRIME_2MSB_ON  - make the 2nd highest bit one
 *
 * You have to supply a callback which fills in a buffer with random bytes.  "dat" is a parameter you can
 * have passed to the callback (e.g. a state or something).  This function doesn't use "dat" itself
 * so it can be NULL
 *
 */
int32 mp_prime_random_ex(mp_int *a, int32 t, int32 size, int32 flags, ltm_prime_callback cb, void *dat);

/* ---> radix conversion <--- */
int32 mp_count_bits(mp_int *a);

int32 mp_unsigned_bin_size(mp_int *a);
int32 mp_read_unsigned_bin(mp_int *a, const unsigned char *b, int32 c);
int32 mp_to_unsigned_bin(mp_int *a, unsigned char *b);
int32 mp_to_unsigned_bin_n (mp_int * a, unsigned char *b, unsigned long *outlen);
int32 mp_to_unsigned_bin_n_pad (mp_int * a, unsigned char *b, unsigned long len);

int32 mp_signed_bin_size(mp_int *a);
int32 mp_read_signed_bin(mp_int *a, const unsigned char *b, int32 c);
int32 mp_to_signed_bin(mp_int *a,  unsigned char *b);
int32 mp_to_signed_bin_n (mp_int * a, unsigned char *b, unsigned long *outlen);

#ifdef TOUPPER_ALLOWED
int32 mp_read_radix(mp_int *a, const char *str, int32 radix);
#endif
int32 mp_toradix(mp_int *a, char *str, int32 radix);
int32 mp_toradix_n(mp_int * a, char *str, int32 radix, int32 maxlen);
int32 mp_radix_size(mp_int *a, int32 radix, int32 *size);

#define mp_read_raw(mp, str, len) mp_read_signed_bin((mp), (str), (len))
#define mp_raw_size(mp)           mp_signed_bin_size(mp)
#define mp_toraw(mp, str)         mp_to_signed_bin((mp), (str))
#define mp_read_mag(mp, str, len) mp_read_unsigned_bin((mp), (str), (len))
#define mp_mag_size(mp)           mp_unsigned_bin_size(mp)
#define mp_tomag(mp, str)         mp_to_unsigned_bin((mp), (str))

#define mp_tobinary(M, S)  mp_toradix((M), (S), 2)
#define mp_tooctal(M, S)   mp_toradix((M), (S), 8)
#define mp_todecimal(M, S) mp_toradix((M), (S), 10)
#define mp_tohex(M, S)     mp_toradix((M), (S), 16)

/* lowlevel functions, do not call! */
int32 s_mp_add(mp_int *a, mp_int *b, mp_int *c);
int32 s_mp_sub(mp_int *a, mp_int *b, mp_int *c);
#define s_mp_mul(a, b, c) s_mp_mul_digs(a, b, c, (a)->used + (b)->used + 1)
int32 fast_s_mp_mul_digs(mp_int *a, mp_int *b, mp_int *c, int32 digs);
int32 s_mp_mul_digs(mp_int *a, mp_int *b, mp_int *c, int32 digs);
int32 fast_s_mp_mul_high_digs(mp_int *a, mp_int *b, mp_int *c, int32 digs);
int32 s_mp_mul_high_digs(mp_int *a, mp_int *b, mp_int *c, int32 digs);
int32 fast_s_mp_sqr(mp_int *a, mp_int *b);
int32 s_mp_sqr(mp_int *a, mp_int *b);
int32 mp_karatsuba_mul(mp_int *a, mp_int *b, mp_int *c);
int32 mp_toom_mul(mp_int *a, mp_int *b, mp_int *c);
int32 mp_karatsuba_sqr(mp_int *a, mp_int *b);
int32 mp_toom_sqr(mp_int *a, mp_int *b);
int32 fast_mp_invmod(mp_int *a, mp_int *b, mp_int *c);
int32 mp_invmod_slow (mp_int * a, mp_int * b, mp_int * c);
int32 fast_mp_montgomery_reduce(mp_int *a, mp_int *m, mp_digit mp);
int32 mp_exptmod_fast(mp_int *G, mp_int *X, mp_int *P, mp_int *Y, int32 mode);
int32 s_mp_exptmod (mp_int * G, mp_int * X, mp_int * P, mp_int * Y, int32 mode);
void bn_reverse(unsigned char *s, int32 len);

extern const char *mp_s_rmap;

#ifdef __cplusplus
   }
#endif

#endif

