/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_CHANNELINTERNAL_H
#define DOF_CHANNELINTERNAL_H

#include <dof/core.h>
#include <dof/credentials.h>
#include <dof/transport.h>

#include <internal/oaltypesinternal.h>
#include <internal/observable.h>
#include <internal/dofinternal.h>
#include <internal/messaging.h>
#include <internal/remotechanneldata.h>
#include <internal/securityscope.h>
#include <internal/slist.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @internal @addtogroup OALChannel
@{
**/


/*
**  "Public" Type and Method Declarations
*/

#define INVALID_SECURITYSTATEID     0xFFFFFFFF
#define INVALID_REMOTE_DOMAIN_ID    0xFFFFFFFF

extern OALChannel OALChannelMgr_CreateChannel_Connection( OALConnection OALConn, DOFException * exception );
extern DOFError OALChannelMgr_CreateChannel_Server( DOFServer server );
extern boolean OALChannelMgr_CreateChannel_StreamServerConnect( DTL dtl, OALServer dofserver, DOFConnectionType connType, DOFAddress address, DTLConnectionConfig config );
extern OALChannel OALChannelMgr_CreateChannel_ConnectionlessInbound(OALChannel serverChannel, DOFConnectionType connType,
                                                                    DOFAddress address, DOFCredentials creds, DOFPermissionSet perms);
extern boolean channel_mgr_tree_Remove_Channel(OALChannel channelStored, OALChannel channelIn);
extern void OALChannelMgr_CloseChannel( OALChannel channel );
extern void OALChannel_StopClose( OALChannel channel );

typedef void (*ExtendCompleteCB) (DOFError   status,
                                  void     * context);

typedef void (*OALChannel_RouteToProvider_fn_t)( OALChannel self, DOFOperation req_op, DOFOperation prov_op,
                                                    DOFAddress addr );
typedef void (*OALChannel_RouteToRequestor_fn_t)( OALChannel self, DOFOperation respOp, DOFResult result );

typedef void (*OALChannel_MessageHandler_fn_t)( OALChannel self, OALMessage message );
typedef void (*OALChannel_Destroy_fn_t)(OALChannel channel);

typedef void (*OALChannel_ExtendPermissions)(OALChannel         channel,
                                             DOFAddress         address,
                                             DOFPermissionSet   extendPerms,
                                             ExtendCompleteCB   callback,
                                             void             * context);

typedef struct OALChannel_Connection_Node_t
{
    OALSListNode                        next;
    DOFConnection                       connection;
}OALChannel_Connection_Node_t;

typedef struct OALChannel_Functions_t
{
    OALChannel_RouteToProvider_fn_t     routeToProvider;
    OALChannel_RouteToRequestor_fn_t    routeToRequestor;
    OALChannel_MessageHandler_fn_t      messageHandler;
    OALChannel_Destroy_fn_t             destroy;
    OALChannel_ExtendPermissions        extendPermission;
}OALChannel_Functions_t;

typedef struct OALChannel_t
{
    uint32              refCount;   /* reference count */
    OALObservable       chObservable; /* Those who need to know when the channel goes down will use this for notification of such. */

    /* properties */
    DOFSecurityScope    securityScope;      /* This is probably not what you want!  Unless you are TEP, SGMP,
                                               or the PassThroughSecurityConnector, you probably want to use
                                               OALChannel_GetSecurityScope instead. */
    OALPermissionManager permissionManager;
    OALPermissionManager inboundMembranePermissionManager;
    OALPermissionManager outboundMembranePermissionManager;
    DOFSecurityMembrane securityMembrane;
    RemoteChannelData   remoteChannelData;  /* The should be NULL only for DOFSystem's. Router now depends on this. If you make this a union to hold some data for system, there may be a problem. */

    DOFConnection       sysConnection;      /* This DOFConnection is used for systems.  To get DOFConnections for remote channels
                                               use the OALConnection in remote channel data.  The OALConnection keeps a list of
                                               all of the DOFConnections that are using the OALConnection. */

    /* virtual methods */
    OALChannel_RouteToProvider_fn_t     routeToProvider;
    OALChannel_RouteToRequestor_fn_t    routeToRequestor;
    OALChannel_MessageHandler_fn_t      msg_handler_func;
    OALChannel_Destroy_fn_t             destroy;
    OALChannel_ExtendPermissions        extendPermissions;
} OALChannel_t;

extern void OALChannel_Init(void);

extern void OALChannel_Shutdown(void);

extern boolean OALChannel_Ctor(OALChannel_t* channel, boolean isRemote, DOFSecurityMembrane connector,
                               OALChannel_RouteToProvider_fn_t    routeToProviderFunc,
                               OALChannel_RouteToRequestor_fn_t   routeToRequestorFunc,
                               OALChannel_MessageHandler_fn_t     msg_handler_func,
                               OALChannel_Destroy_fn_t            destroyFunc,
                               OALChannel_ExtendPermissions       extendPermissions);

extern void OALChannel_Dtor(OALChannel_t* channel);

/* Note: This function requires that the caller have zeroed full contents of the RemoteChannelData_t structure at
   channel->remoteChannelData. */
extern boolean OALChannel_RemoteChannel_Ctor(OALChannel_t* channel,
                                             void* protocolStack,
                                             DTL transport,
                                             DOFAddress connectedAddr,
                                             DOFConnectionConfig connConfig,
                                             DOFServerConfig serverConfig);

extern void OALChannel_RemoteChannel_Dtor( OALChannel_t* channel );

extern boolean OALChannel_IsSecureGroup( OALChannel self );

#define OALChannel_IsEquivalent( self, channel ) ( OALChannel_GetPrimarySessionChannel( self ) == OALChannel_GetPrimarySessionChannel( channel ) )

extern int OALChannel_Compare( OALChannel self, OALChannel channel );

/* Note: this function must only be called from within a channel message_enqueue worker thread. */
extern void OALRemoteChannel_DeregisterSecureChannel_Immediate(OALChannel self);

#ifndef OALCHANNEL_DEBUG_REFCOUNT
extern OALChannel OALChannel_Retain(OALChannel self);
extern void OALChannel_Release(OALChannel self);
#else
#define OALChannel_Retain(ch) OALChannel_Retain_Debug((ch), __func__, __FILE__, __LINE__)
#define OALChannel_Release(ch) OALChannel_Release_Debug((ch), __func__, __FILE__, __LINE__)
extern OALChannel OALChannel_Retain_Debug(OALChannel self, const char *fn_name, const char *filename, unsigned int line_number);
extern void OALChannel_Release_Debug(OALChannel self, const char *fn_name, const char *filename, unsigned int line_number);
#endif

extern void OALChannel_Lock(OALChannel self);
extern void OALChannel_Unlock(OALChannel self);

extern DOFSecurityScope OALChannel_GetSecurityScope( OALChannel chan, boolean inbound );

extern OALChannel OALChannel_GetPrimarySessionChannel( OALChannel self );

extern boolean OALChannel_SetSecureSendBlocked( OALChannel channel );

extern DOFConnectionState OALChannel_CreateConnectionState( OALChannel channel, DOFAddress address );

extern void OALChannel_ExtendOutboundPermissions(OALChannel         channel,
                                                 DOFAddress         address,
                                                 DOFPermission      perm,
                                                 ExtendCompleteCB   callback,
                                                 void             * context);

extern void OALChannel_ExtendOutboundMembranePermissions(OALChannel         channel,
                                                         DOFPermission      perm,
                                                         ExtendCompleteCB   callback,
                                                         void             * context);

extern void OALChannel_ExtendInboundMembranePermissions(OALChannel         channel,
                                                        DOFPermission      perm,
                                                        ExtendCompleteCB   callback,
                                                        void             * context);

extern DOFSecurityMembrane OALChannel_CreateSecurityMembrane(DOF                dof,
                                                             DOFOperationFilter receiveFilter,
                                                             DOFOperationFilter sendFilter,
                                                             OALOperationBridge bridge);

extern boolean DOFSecurityMembrane_IsChannelMembrane(DOFSecurityMembrane self);

extern OALOperationBridge OALChannelMembrane_GetBridge(DOFSecurityMembrane self);

extern boolean OALChannelMembrane_SetBridge(DOFSecurityMembrane self,
                                            OALOperationBridge  newBridge);

/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* DOF_CHANNELINTERNAL_H */
