/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_CIPHER_INTERNAL_H
#define DOF_CIPHER_INTERNAL_H


#include <dof/core.h>
#include <dof/blockcipher.h>
#include <internal/oaltypesinternal.h>
#include <internal/staticassertinternal.h>
#include <internal/stream.h>


#ifdef  __cplusplus
extern "C" {
#endif


/** @internal @addtogroup OALCrypt
@{
**/


/* Not in cipherprivate.h so ccmMode_v1.c can use it. */
#define TOP_NODEID                  (0x1FFFFFFF)


/* These are not in cipherprivate.h so secureGroup and TRP and TEP can use them. */
typedef uint8  OALCryptKey[32];
typedef uint8  OALCryptExpandedKey[240];
extern DOFCipher DOFCipherRegistry_AES256Cipher;


/* These are not in cipherprivate.h so secureGroup can use a couple of them. */
#define CUR                         (0)
#define SLOT                        (1)
#define KEKKEY                      (2)
#define UNTAGGED_KEY                (KEKKEY + 1)
typedef struct DOFGroupKey_t
{
    OALCryptKey key;
    OALStream   streamCcmModeInitialState;
    boolean     receivedMulticast;
    boolean     filled;
} DOFGroupKey_t;


/* Not in cipherprivate.h so epp_v2.c can use it. */
typedef enum DecipherKeyType_t
{
    DECIPHERED_BY_CURKEY = 0,
    DECIPHERED_BY_FUTKEY,
    DECIPHERED_BY_OLDKEY
} DecipherKeyType_t;
typedef struct PeriodStore_t * OALCipherKeyState;

/* Not in cipherprivate.h so securitymode.c can use it. */
extern uint8 OALCipher_ComputeMacAuthFlags( uint8 macSize );


/**
@}
**/


/* This little section not in cipherprivate.h so srp can use it. */
/** @internal @addtogroup OALSecureHash
@{
**/


STATIC_ASSERT( sizeof( OALCryptKey ) == 32 );
typedef uint8 OALSecureHashDigest[32];


typedef struct OALSecureHashContext
{
    OALSecureHashDigest working;
    uint16              stored;
    uint16              length;
    uint32              T[ 16 ];
} OALSecureHashContext;


extern void OALSecureHash_Start ( OALSecureHashContext *ctx );
extern void OALSecureHash_Digest( OALSecureHashContext *ctx, uint32 dlen, const uint8 *data );
extern void OALSecureHash_Finish( OALSecureHashContext *ctx, OALSecureHashDigest digest );


/**
@}
**/


/* This little section not in cipherprivate.h so a few other classes can use it. */
/** @internal @addtogroup OALSecureHMAC
@{
**/


typedef OALSecureHashDigest OALSecureHMACDigest;
typedef struct OALSecureHMACContext
{
    OALCryptKey          key;
    OALSecureHashContext context;
} OALSecureHMACContext;


extern void OALSecureHMAC_Start ( OALSecureHMACContext *ctx, const OALCryptKey key );
extern void OALSecureHMAC_Digest( OALSecureHMACContext *ctx, uint32 dlen, const uint8 *data );
extern void OALSecureHMAC_Finish( OALSecureHMACContext *ctx, OALSecureHMACDigest digest );


/**
@}
**/


/** @internal @addtogroup OALCipher
@{
**/


/* Bit masks for defining cipher configuration and/or parsing the mode field of incoming secure packets. */
#define TRANSMIT_COUNT                                      0x01
#define TRANSMIT_BASE_OR_NEXT_NODEID                        0x02
#define TRANSMIT_TARGET_NODEID                              0x08
#define MANAGER_TRANSMITTED_PDU                             0x80


#define TRANSMIT_SLOT                                       0x02
#define TRANSMIT_NODEID                                     0x04
#define CIPHER_IS_HUB                                       0x10
#define CIPHER_CONFIG_USE_GIVEN_NODEID                      0x20
#define CIPHER_CONFIG_SECUREGROUP_DEFAULT       (TRANSMIT_NODEID | TRANSMIT_SLOT | TRANSMIT_COUNT)      /* multicast. */
#define CIPHER_CONFIG_POINT_TO_POINT_DEFAULT    TRANSMIT_COUNT                                 /* UDP point to point. */
#define CIPHER_CONFIG_CONNECTED_DEFAULT         0                                              /* TCP point to point. */


extern void      OALCipher_Lock                           (       OALCipher self );
extern void      OALCipher_Unlock                         (       OALCipher self );
extern OALCipher OALCipher_Create_TEP                     (       DOFSecurityMode securityMode, boolean connectionless,
                                                                  boolean tepResponder, DOFException * pException );
extern OALCipher OALCipher_Create                         (       DOFSecurityMode securityMode, uint8 configFlags,
                                                                  uint32 nodeID, OALChannel pduChannel,
                                                                  DOFException * pException );
extern boolean   OALCipher_SetPruneWatermarks             (       OALCipher self, uint32 lowStateWatermark,
                                                                  uint32 highStateWatermark );
extern void      OALCipher_SetStandardPruneOrKeyDisableDelay(     OALCipher self, uint16 secsStandardPruneOrKeyDisable);
extern boolean   OALCipher_SetCurKey                      (       OALCipher self, DOFAddress * pManagerAddress,
                                                                  OALStream streamCcmModeInitialState,
                                                                  const OALCryptKey newKey,
                                                                  boolean stateReceivedByIncomingMulticastPDU );
extern boolean   OALCipher_SetFutureKey                   (       OALCipher self, OALStream streamCcmModeInitialState,
                                                                  OALCryptKey newKey, DOFPermissionSet localPerms, DOFPermissionSet remotePerms );
extern void      OALCipher_PromoteFutureKey               (       OALCipher self,
                                                                  uint32 maxSecsToPreviousKeyReceiveDisable );
extern boolean   OALCipher_FormNewAndSetSlotKey           (       OALCipher self, uint16 secsPromoteDelay,
                                                                  OALStream streamCcmModeInitialState,
                                                                  DOFSecureGroup secureGroup );
extern boolean   OALCipher_FormNewAndSetSlotOrKEKKey      (       OALCipher self, uint16 secsPromoteDelay,
                                                                  uint8 keyType, uint8 epoch,
                                                                  DOFSecurityMode kekKeySecurityMode,
                                                                  DOFPermissionSet localPerms, DOFPermissionSet remotePerms,
                                                                  DOFSecureGroup secureGroup );
extern boolean   OALCipher_SetSlotKey                     (       OALCipher self, DOFAddress * pManagerAddress,
                                                                  OALStream streamCcmModeInitialState,
                                                            const OALCryptKey newKey, uint16 secsPromoteDelay,
                                                                  DOFSecureGroup secureGroup,
                                                                  boolean onTieOverwriteCurKeyPeriod,
                                                                  boolean stateReceivedByIncomingMulticastPDU );
extern boolean   OALCipher_SetKEKKey                      (       OALCipher self, DOFAddress * pManagerAddress,
                                                                  OALStream streamCcmModeInitialState,
                                                                  OALCryptKey newKey, uint16 secsPromoteDelay,
                                                                  uint16 kekEpoch, DOFSecurityMode kekKeySecurityMode,
                                                                  DOFPermissionSet localPerms, DOFPermissionSet remotePerms,
                                                                  DOFSecureGroup secureGroup,
                                                                  boolean stateReceivedByIncomingMulticastPDU );
extern void      OALCipher_DisableOldKeys                 (       OALCipher self );
extern void      OALCipher_SetManager                     (       OALCipher self, boolean manager );
extern OALChannel OALCipher_GetConnectionChannel          ( const OALCipher self );
extern boolean   OALCipher_IsSecureGroup                  ( const OALCipher self );
extern boolean   OALCipher_IsHub                          ( const OALCipher self );
extern boolean   OALCipher_GetKey                         ( const OALCipher self, uint8 keyType, OALCryptKey key,
                                                                  uint8 * pPeriod, uint16* pSlot );
extern boolean   OALCipher_IsKeySet                       ( const OALCipher self );
extern boolean   OALCipher_QueryReceiveHandler            ( const OALCipher self, DOFAddress address );
extern boolean   OALCipher_RequestReceiveHandler          ( const OALCipher self, DOFOperation requestOp );
extern boolean   OALCipher_RequestResponseReceiveHandler  ( const OALCipher self, uint32 nodeID );
extern boolean   OALCipher_IsEncrypt                      ( const OALCipher self );
extern boolean   OALCipher_IsAuthenticate                 ( const OALCipher self );
extern boolean   OALCipher_SecureGroupEncipherReady       (       OALCipher self, DOFAddress unicastTargetAddress,
                                                                  uint32 secureMarkingID,
                                                                boolean denyPacketReplayProtectionIfSecureUnicastSend );
extern boolean   OALCipher_IsValid_AdvertisedNNI          ( const OALCipher self );
extern DOFError  OALCipher_Encipher                       ( const OALCipher self, const OALStream * pEPPHeaderStream,
                                                                  OALStream * pPayloadStream,
                                                                  uint8 payloadEncryptOffset,
                                                                  DOFAddress unicastTargetAddress,
                                                                  boolean cipherModeTransmission,
                                                                  uint32 secureMarkingID,
                                                                  boolean ifProperPlaceNextBaseNIDOnWire );
extern DOFError  OALCipher_Decipher                       ( const OALCipher self, OALStream * pFullDecipherStream,
                                                                  uint32 offsetToBeginEPP, uint8 payloadEncryptOffset,
                                                                  DOFAddress * pSenderAddress,
                                                                  boolean * pUnicastBehaviorMarkedBySender,
                                                                  DecipherKeyType_t * pDecipherKeyType,
                                                                  OALCipherKeyState * pCipherKeyState );
extern void      OALCipher_RemoveState                    (       OALCipher self );
extern boolean   OALCipher_GetCurKeySlot                  ( const OALCipher self, uint16 * pNowSlot );
extern boolean   OALCipher_NeedRekey                      ( const OALCipher self );
extern boolean   OALCipher_GetSGMGroupKey                 ( const OALCipher self, DOFGroupKey_t * pFillGroupKey,
                                                                  DOFAddress requestorAddr );
extern void      OALCipher_ClearSecureGroupNotification   (       OALCipher self, DOFSecureGroup secureGroup );
extern void      OALCipher_Destroy                        (       OALCipher self );
extern boolean   OALCipher_IsNodeIDValid                  ( const OALCipher self );
extern OALStream OALCipher_GetSecureGroupCommonState      (       OALCipher self, OALStream streamCcmModeInitialState );
extern OALStream OALCipher_CreateInitialStateStream       (       OALCipher self, uint8 period, uint16 initialSlot,
                                                                  uint32 initialPacketNumber, boolean includeTargetNID,
                                                                  uint32 targetNodeID, uint32 baseNodeID );
extern OALStream OALCipher_CreateInitialStateStream_TEP   (       OALCipher self );
extern void      OALCipher_DisableAllPendingRekeys        (       OALCipher self );
extern boolean   OALCipher_SecureGroupGetNewNodeID        (       OALChannel secureGroupChannel );
extern void      OALCipher_AdjustKeys                     ( OALCipher cipher );
extern boolean   OALCipher_AddressIP_Port_gt_fn           ( const DOFAddress * pAddress,
                                                            const DOFAddress * pTestAddress, void * context );
extern boolean   OALCipher_AddressIP_Port_eq_fn           ( const DOFAddress * pAddress,
                                                            const DOFAddress * pTestAddress, void * context );
extern void      OALCipher_CheckAndSetChannelSecureSendBlocked(   OALCipher self );
extern void      OALCipher_ClearChannelSecureSendBlocked  (       OALCipher self );

extern DOFPermissionSet OALCipher_GetLocalPerms( OALCipher self, OALCipherKeyState keyState );
extern boolean OALCipher_SetLocalPerms( OALCipher self, OALCipherKeyState keyState, DOFPermissionSet permSet );
extern DOFPermissionSet OALCipher_GetRemotePerms( OALCipher self, OALCipherKeyState keyState );
extern boolean OALCipher_SetRemotePerms( OALCipher self, OALCipherKeyState keyState, DOFPermissionSet permSet );
extern OALCipherKeyState OALCipher_GetCurrentKeyState( OALCipher self );

/**
@}
**/


#ifdef  __cplusplus
}
#endif


#endif /* DOF_CIPHER_INTERNAL_H */
