/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_CONNECTION_INTERNAL_H
#define DOF_CONNECTION_INTERNAL_H

#include <dof/core.h>

#include <dof/connection.h>
#include <internal/oaltypesinternal.h>
#include <internal/securityscope.h>
#include <internal/slist.h>
#include <internal/timer.h>
#include <internal/transportinternal.h>
#include <internal/credentialsinternal.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @addtogroup DOFConnection
@{
**/

/**
 * The current state of a @ref DOFConnection.
 */
typedef enum DOFConnectionStatus_t
{
    DOFCONNECTIONSTATUS_DISCONNECTED = 0, /**< The connection is not currently established. **/
    DOFCONNECTIONSTATUS_CONNECTED    = 1  /**< The connection is currently established. **/
} DOFConnectionStatus;

typedef void * DOFConnectionKey;

typedef struct DOFConnectionSRL_t * DOFConnectionSRL;

typedef struct DOFConnectionConfig_t
{
    uint32                              refCount;
    DOFConnectionType                   type;
    DOFSecurityDesire                   securityDesire;
    DOFAddress                          targetAddress;
    DOFPermissionSet                    permissions;
    DTLConnectionConfig                 transportConfig;
    DOFProtocolNegotiator               negotiator;
    DOFCredentials                      credentials;
    DOFCredentials                      neuteredCredentials;
    OALDomainCredentialsTree            domainCreds; /* Tree of domain credentials keyed on domainID */
    DOFOperationBridgeConfig            bridge;
    DOFOperationFilter                  receiveFilter;
    DOFOperationFilter                  sendFilter;
    DOFConnectionSRL                    streamListener;
    uint32                              maxSendSilence;
    uint32                              maxReceiveSilence;
    uint32                              errors;
    struct
    {
        unsigned int setAddressFailed           : 1;
        unsigned int setCredentialsFailed       : 1;
        unsigned int setPermissionsFailed       : 1;
        unsigned int setTransportConfigFailed   : 1;
        unsigned int isExtendAllowed            : 1;
    } flags;
}DOFConnectionConfig_t;

typedef struct OALConnection_findContext_t
{
    OALConnection       OALConn;
    DOFConnectionConfig connConfig;
    DOFSecurityMembrane connector;
    void *              baseOAL;        /* Note: this can be an OALServer or an OALConnection.  Only used for pointer comparison. */
    boolean             baseIsOALConnection;
}OALConnection_findContext_t;

typedef struct NotifySystem_ListenerRemoved_t
{
    OALEnqueueItem_t            super;

    DOFConnection               connection;
    DOFConnectionStateListener  listener;
} NotifySystem_ListenerRemoved_t;

typedef struct NotifySystem_PeerListenerRemoved_t
{
    OALEnqueueItem_t            super;

    DOFConnection               connection;
    DOFConnectionPeerListener  listener;
} NotifySystem_PeerListenerRemoved_t;

typedef struct connLstnNode_t
{
    OALSListNode                    next;
    NotifySystem_ListenerRemoved_t  *removeNode;
    DOFConnectionStateListener      listener;
}connLstnNode_t;

typedef struct connPeerLstnNode_t
{
    OALSListNode                        next;
    NotifySystem_PeerListenerRemoved_t  *removeNode;
    DOFConnectionPeerListener           listener;
}connPeerLstnNode_t;

typedef struct CallConnectComplete_t
{
    OALEnqueueItem_t super;
    DOFConnectionConnectCallback    callback;
    DOFConnection                   connection;
    void                            *context;
    DOFException                    except;
} CallConnectComplete_t;

typedef struct pendingConnection_t
{
    OALSListNode            next;
    CallConnectComplete_t   *cbNode;
    OALTimerItem            timerItem;
    DOFConnection           pendConnection;
    uint32                  refCount;
}pendingConnection_t;

typedef struct RelatedConnectionListener_t
{
    DOFConnectionStateListener_t    connectionListener;
    DOFConnection                   connection;     /* The connection that has a related connection. */
} RelatedConnectionListener_t;

typedef struct RelatedServerListener_t
{
    DOFServerStateListener_t        serverListener;
    DOFConnection                   connection;     /* The connection that has a related server. */
}RelatedServerListener_t;

typedef struct OALConnection_t
{
    DOF                     dof;
    OALSListNode            dofConnectionList;
    DOFConnection           internalConnection;
    DOFConnectionConfig     connConfig;
    OALChannel              channel;
    DOFSecurityMembrane     connSecurityMembrane;
    connLstnNode_t *        listeners;
    DOFCredentials          peerCredentials;
    union
    {
        OALConnection       OALConn;
        OALServer           OALSrv;
    }base;
    OALAVLTree              trpRoutesTree;
    OALAVLTree              streamRequestTree;
    uint64                  byteInputCount;
    uint64                  packetInputCount;
    uint64                  byteOutputCount;
    uint64                  packetOutputCount;
    uint32                  refCount;
    uint32                  connectedRefCount;          /* Count of "connected" DOFConnections. */
    struct
    {
        unsigned int        baseIsOALConnection     : 1;    /* Set if base is OALConnection, otherwise OALServer. */
        unsigned int        isConnected             : 1;    /* A channel without isConnected set is "connecting". */
        unsigned int        isOutbound              : 1;
        unsigned int        doNotDestroy            : 1;
        unsigned int        shutdownByDOF           : 1;
    } flags;
}OALConnection_t;

typedef void (*_Notification_fn_t)( DOFConnection self, OALChannel channel );

typedef struct DOFConnection_t
{
    OALSListNode            next;                   /* Note: this must be first. */
    uint32                  refcnt;                 /* reference count for DOFConnection */
    uint32                  connectedRefCount;      /* Count of connected calls - disconnected calls */
    OALSListNode            listenerList;
    OALSListNode            peerListenerList;
    DOFConnection           relatedConnection;
    DOFServer               relatedServer;
    OALConnection           OALConn;
    OALSListNode            pendingConnections;
    OALAVLTree              trpDomainRoutes;
    OALAVLTree              trpTrustedRoutes;
    union
    {
        RelatedConnectionListener_t *   connection;
        RelatedServerListener_t *       server;
    }listener;
    boolean                 shutdownByDOF;
    boolean                 destroyInProgress;
    boolean                 stopOrDisconnectRelated;
} DOFConnection_t;

typedef struct OALConnection_TRPRoute_t
{
   union
   {
       DOFDomain           domain;
       DOFObjectIDDomain   domainID;
   }type;
   uint32                  count;
   uint32                  timeout;
   boolean                 isDomain;
}OALConnection_TRPRoute_t;

#define CONNECTION_MAX_REF_COUNT 0xFFFFFFFF         /* used for the structure ref count and connected ref count. */

/** @internal
   Test two ConnectionConfig structures to determine if they are equivalent.

   @param self The DOFConnectionConfig.
   @param config The DOFConnectionConfig to be compared against.

   @return boolean Retrun true if the two are equivalent.
 */
extern boolean DOFConnectionConfig_IsEqual(DOFConnectionConfig self, DOFConnectionConfig config);

extern void DOFConnection_RemovePendingConnections_fn(OALSListNode item, void *context);

/*  Remove both the listeners for the related connection or server, and all other listeners.
    Note: This function must be called with the connection mutex locked, and before the callback enqueue has been destroyed.
*/
extern void DOFConnection_ClearAllListeners(DOFConnection self);
extern boolean DOFConnection_Init(void);
extern void DOFConnection_Shutdown(void);

extern DOFConnection DOFConnection_Copy( const DOFConnection self );

extern void DOFConnection_ValidateComplete(const DOFConnection self, DOFException exception);

extern DOFConnection DOFConnection_CreateInternalConnection( DOF dof, DOFSecurityMembrane securityConnector, DOFConnectionConfig connConfig,
                                                             DOFServer relatedServer, DOFConnection relatedConnection, boolean shareOALConn );

extern boolean DOFConnection_CreateConnectedInboundConnection( DOF dof, DOFSecurityMembrane securityConnector, DOFConnectionConfig connConfig,
                                                               DOFServer relatedServer, OALChannel channel, boolean addListener );

extern void DOFConnection_Terminate(DOFConnection self, DOFException exception);

extern void DOFConnection_NotifyChannelUp  (DOFConnection self ,OALChannel channel);
extern void DOFConnection_NotifyChannelDown(DOFConnection self , OALChannel channel, const char * reason);

void DOFConnection_RemovePendingConnections_fn(OALSListNode item, void *arg);
extern void DOFConnection_ConnectionEvent(DOFConnection conn, boolean status, DOFConnectionState connState, DOFException exception);
extern boolean DOFConnection_AddInternalServerListener(DOFConnection conn);
extern OALChannel DOFConnection_GetChannel( DOFConnection self );

extern DOFConnectionState DOFConnectionState_Create( DOF dof, DOFConnectionConfig connConfig, DOFSecurityMembrane securityConnector,
                                              boolean isConnected, OALChannel channel, DOFConnectionDirection direction,
                                              DOFCredentials peerCredentials, DOFConnection relatedDofConnection,
                                              DOFServer relatedDofServer, uint64 inputByteCount, uint64 outputByteCount,
                                              uint64 inputPacketCount, uint64 outputPacketCount, DOFException exception);
extern DOFConnectionConfig DOFConnectionState_GetConfig( const DOFConnectionState self );
extern DOFConnectionState DOFConnectionState_Copy( DOFConnectionState self );

/**
   Create a copy of a DOFConnectionConfig, where modifications to the copy will not affect the original in any way, and
   changes to the original are isolated from the copy.

   @mustfree{DOFConnectionConfig_Destroy}

   @param self The DOFConnectionConfig.

   @return The copied, and now isolated, DOFConnectionConfig, or NULL if @a self is NULL or it could not be created.
 */
extern DOFConnectionConfig DOFConnectionConfig_Copy( const DOFConnectionConfig self );

/** @internal

   @param connConfig The DOFConnectionConfig to test.

   @return TRUE for connConfig of transport with configured GetAddressType function that says address type is Unicast.
**/
extern boolean DOFConnectionConfig_RequestTypeIsKnownAndIsUnicast( DOFConnectionConfig connConfig );
extern DOFCredentials DOFConnectionConfig_GetNeuteredCredentials( const DOFConnectionConfig self );

extern void OALConnection_ClearChannel(OALConnection OALConn);
extern void OALConnection_RemoveDOFConnection(OALConnection OALConn, DOFConnection conn);

/*  This function will traverse the list of DOFConnections for the OALConnection and remove the listeners
    that may have been added to the DOFConnections.
    Note: This function must be called with the connection mutex locked. */
extern void OALConnection_ShutdownDOFConnections(OALConnection OALConn);

extern OALConnection OALConnection_CreateConnection(DOFConnection conn, DOFConnectionConfig connConfig, DOFSecurityMembrane securityConnector, DOF dof, boolean shareOALConn);
extern void OALConnection_Connect(OALConnection OALConn, DOFException * except);
extern boolean OALConnection_ConnectInbound(OALConnection OALConn, DOFConnection conn, OALChannel channel);
extern void OALConnection_DOFConnectionConnected(OALConnection OALConn);
extern boolean OALConnection_Retain(OALConnection OALConn);
extern void OALConnection_Release(OALConnection OALConn);
extern void OALConnection_DOFConnectionDisconnected(OALConnection OALConn, DOFException exception);
extern DOFConnection OALConnection_CreateDOFConnection(OALConnection OALConn);
extern DOFConnection OALConnection_GetDOFConnection(OALConnection OALConn);
extern DOFConnection OALConnection_GetRelatedDOFConnection(OALConnection OALConn);
extern DOFServer OALConnection_GetRelatedDOFServer(OALConnection OALConn);
extern DOFServer OALConnection_CreateRealtedDOFServer(OALConnection OALConn, DOFException * pException);
extern void OALConnection_ConnectionEvent(OALConnection OALConn, OALChannel channel, boolean connected, DOFException exception);
extern void OALConnection_PeerListenerEvent(OALConnection OALConn, DOFAddress address, DOFSecurityScope scope, boolean added);
extern void OALConnection_SetPeerCredentials(OALConnection OALConn, DOFCredentials creds);
extern void OALConnection_ConnectionStreamRequestEvent(OALConnection OALConn, DOFAddress addr, boolean complete);
extern void OALConnection_UpdateInputByteCount( OALConnection conn, uint32 count );
extern void OALConnection_UpdateOutputByteCount( OALConnection conn, uint32 count );

extern boolean OALConnection_AddDynamicTRPRoute( OALConnection self, DOFDomain domain, DOFObjectIDDomain domainID, uint32 timeout, DOFException* exception );
extern void OALConnection_RemoveDynamicTRPRoute( OALConnection self, DOFDomain domain, DOFObjectIDDomain domainID );

extern boolean DOFConnection_AddDynamicTRPRoute( DOFConnection self, DOFDomain domain, DOFObjectIDDomain domainID, uint32 timeout, DOFException* exception );
extern void DOFConnection_RemoveDynamicTRPRoute( DOFConnection self, DOFDomain domain, DOFObjectIDDomain domainID );

extern DOFSecurityMembrane DOFConnection_CreateSecurityMembrane(DOF                 dof,
                                                                DOFConnectionConfig config);

/**
 * Return the DOF from which the Connection was created.
 *
 * @param self The Connection.  This must not be NULL.
 *
 * @return The DOF associated with the Connection.
 */
extern DOF_OAL_DECLSPEC DOF DOFConnection_GetDOF( const DOFConnection self );
/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /*DOF_CONNECTION_INTERNAL_H */
