/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef _DOF_CREDENTIALS_INTERNAL_H
#define _DOF_CREDENTIALS_INTERNAL_H

#include <dof/core.h>
#include <internal/stream.h>
#include <internal/avl_tree.h>
#include <internal/slist.h>
#include <internal/oaltypesinternal.h>
#include <internal/observable.h>


#ifdef __cplusplus
extern "C" {
#endif

typedef enum CredentialsUnmarshalReturn_t
{
    CREDENTIALS_FAILED_UNMARSHAL,
    CREDENTIALS_SUCCESSFUL_UNMARSHAL,
    CREDENTIALS_WRONGSTAGE_UNMARSHAL
}CredentialsUnmarshalReturn;

extern void DOFCredentials_Init(void);
extern void DOFCredentials_Shutdown(void);

extern boolean DOFCredentials_InitDOFLevel( DOF dof );
extern void DOFCredentials_ShutdownDOFLevel(void);

/**
   Compare two DOFCredentials and return true if they are equal.
   This doesn't work for Neutered Credentials

   @param self  the first DOFcredentials to use in the compare.
   @param creds The second DOFCredentials to use in the compare.

   @return boolean TRUE if the two credentials are equal.
 */
extern boolean             DOFCredentials_IsEqual(const DOFCredentials self, const DOFCredentials creds);
extern const uint8 *       DOFCredentials_GetKey( const DOFCredentials self );

extern boolean             DOFCredentials_ContainsPrivateData( const DOFCredentials self );


/* this function is for use with TRP and is probably not what you want */
extern const uint8 *       DOFCredentials_GetKeyFromStorage( const DOFCredentials self, const uint8* privateStorage );
/* this function is for use with TRP and is probably not what you want */
extern DOFCredentials      DOFCredentials_CreateCopy( const DOFCredentials self );
/* this function is for use with TRP and is probably not what you want */
extern DOFCredentials      DOFCredentials_CreateCopyWithDomain( const DOFCredentials self, DOFObjectIDDomain domain );

extern DOFCredentials      DOFCredentials_CreateNeutered( const DOFCredentials self );

extern boolean             DOFCredentials_IsCached( const DOFCredentials self );
extern void                DOFCredentials_Reset( const DOFCredentials self );
extern OALObserver         DOFCredentials_AddChangeObserver( const DOFCredentials self, OALObserver_fn_t fn, void* arg );
extern void                DOFCredentials_RemoveChangeObserver( const DOFCredentials self, OALObserver observer );

extern uint32              DOFCredentials_MarshalSize(const DOFCredentials self, boolean isRequest );
extern boolean             DOFCredentials_Internal_Marshal(const DOFCredentials self, OALStream stream, boolean isRequest );
extern CredentialsUnmarshalReturn DOFCredentials_Unmarshal( DOFCredentials* self, DOFObjectIDDomain domain, OALStream stream, boolean isRequest );

/* This returns true if the Domain can be trusted to not change */
extern boolean DOFCredentials_HasFixedDomain( const DOFCredentials self );

extern boolean DOFCredentials_Resolve( DOFCredentials self, DOF dof, int timeout, DOFException* exception );

extern boolean DOFCredentials_Internal_Resolve( DOFCredentials self, DOF dof, int timeout, DOFException* exception, boolean explicitRoute, OALChannel channel, DOFAddress address );

typedef void (*DOFCredentials_BeginResolveCallback)( DOFCredentials self, DOFException exception, void* context );
/* explicitRoute Means force it over the channel and address */
extern void DOFCredentials_Internal_BeginResolve( DOFCredentials self, DOF dof, int timeout, boolean explicitRoute, OALChannel channel, DOFAddress address, DOFCredentials_BeginResolveCallback func, void* context );
/* explicitRoute Means force it over the channel and address */
extern void DOFCredentials_Internal_BeginValidate( DOFCredentials self, DOF dof, int timeout, boolean explicitRoute, OALChannel channel, DOFAddress address, DOFCredentials_BeginResolveCallback func, void* context );

extern boolean DOFCredentials_HasAnotherResolutionStage( DOFCredentials self );

typedef struct OALDomainCredentialsTree_t
{
    DOFCredentials  wildcardCreds;
    OALAVLTree      credentials; /* Known end domains */
    OALSListNode    kownDomainCreds; /* where unknown goes when its resolved?*/
    OALSListNode    unkownDomainCreds;
    OALAVLTree      trustedDomains; /* Trusted domains */
    uint32          unknownDomainCount;
    uint32          knownDomainCount;
    uint32          refCount;
    uint32          maxSilence;
}OALDomainCredentialsTree_t;
typedef struct OALDomainCredentialsTree_t* OALDomainCredentialsTree;

extern OALDomainCredentialsTree OALDomainCredentialsTree_Create(void);
extern OALDomainCredentialsTree OALDomainCredentialsTree_Copy(OALDomainCredentialsTree self);
extern OALDomainCredentialsTree OALDomainCredentialsTree_DeepCopy( OALDomainCredentialsTree srcTree );
extern void OALDomainCredentialsTree_Destroy( OALDomainCredentialsTree self );

extern boolean OALDomainCredentialsTree_SetDomainDiscoveryCredentials( OALDomainCredentialsTree credTree, DOFCredentials creds );
DOFCredentials OALDomainCredentialsTree_GetDomainDiscoveryCredentials( OALDomainCredentialsTree self );
extern boolean OALDomainCredentialsTree_Insert( OALDomainCredentialsTree credTree, DOFDomain domain );
extern boolean OALDomainCredentialsTree_InsertTrustedDomain( OALDomainCredentialsTree self, DOFObjectIDDomain domain );

extern boolean OALDomainCredentialsTree_IsEqual( OALDomainCredentialsTree self, OALDomainCredentialsTree other );

extern boolean OALDomainCredentialsTree_HasCredentials( OALDomainCredentialsTree self );

/* NULL exception = all credentials validated,
   Exception = 1 or more failed validation requests */
typedef void (*OALDomainCredentialsTree_ValidateDomainCredentials_Callback_fn)( DOFException exception, void* context );
/* Resolution/validation forced over channel and address */
extern void OALDomainCredentialsTree_ValidateDomainCredentials( OALDomainCredentialsTree self, DOFConnection conn, OALDomainCredentialsTree_ValidateDomainCredentials_Callback_fn fn, void* context );

/* Traverse list add trusted domain routes called when connection is up completely */
extern boolean OALDomainCredentialsTree_AddTrustedDomainsRoutes( OALDomainCredentialsTree self, OALChannel channel );


typedef struct OALCredentialsTree_t
{
    DOFCredentials  wildcardCreds;
    OALAVLTree      credentials;
    OALSListNode    kownDomainCreds;
    OALSListNode    unkownDomainCreds;
    uint32          unknownDomainCount;
    uint32          knownDomainCount;
    uint32          refCount;
}OALCredentialsTree_t;
typedef struct OALCredentialsTree_t* OALCredentialsTree;

extern OALCredentialsTree OALCredentialsTree_Create(void);
extern OALCredentialsTree OALCredentialsTree_Copy(OALCredentialsTree self);
extern OALCredentialsTree OALCredentialsTree_DeepCopy( OALCredentialsTree srcTree );
extern void OALCredentialsTree_Destroy( OALCredentialsTree self );

extern boolean OALCredentialsTree_SetWildcardCredentials( OALCredentialsTree credTree, DOFCredentials creds );
DOFCredentials OALCredentialsTree_GetWildcardCredentials( OALCredentialsTree self );
extern boolean OALCredentialsTree_Insert( OALCredentialsTree credTree, DOFCredentials creds );

extern boolean OALCredentialsTree_IsEqual( OALCredentialsTree self, OALCredentialsTree other );

extern boolean OALCredentialsTree_HasCredentials( OALCredentialsTree self );

typedef void (*OALCredentialsTree_GetCredentials_Callback_Fn)( DOFCredentials self, void* context );
extern void OALCredentialsTree_GetCredentials( OALCredentialsTree self, DOFObjectIDDomain domainID, uint32 timeout, OALCredentialsTree_GetCredentials_Callback_Fn fn, void* context );

extern DOFObjectIDSource DOFCredentials_GetSID( const DOFCredentials self );
extern void DOFCredentials_UpdateSID( DOFCredentials self, DOFObjectIDSource sid );

#ifdef __cplusplus
}
#endif

#endif /* _DOF_CREDENTIALS_INTERNAL_H */

