/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef __DList_H
#define __DList_H

#include <dof/core.h>
#include <internal/slist.h>

#ifdef  __cplusplus
extern "C" {
#endif

/**
a simple doubly-linked list implementation. fun!
 inserts in the front for efficiency

 this is NOT THREAD SAFE but it probably doesn't matter since worst-case a
 thread misses a member during pointer swap
 **/

/** @internal @addtogroup OALDList
*   A double linked list

*   This is a very fast list implementation using special structures that
*   have a pointer to their same type as the first member in the structure.
*   Items are always inserted at insertion O(1).
*   Removing an item from a list has O(1) complexity. It is very important
*   not to try and add one item to two lists at the same time.
*
*   Lists are not created. Any item can start a list as long as the next
*   pointer is NULL. Normally it's custom to have a single pointer called head.
*   
*
@{
@internal
**/
 
/* IMPORTANT: the beginning of the struct you put in DList should look like this! */
typedef struct OALDListNode_t {
    struct OALDListNode_t *next; /* pointer to the next item in the list */
    struct OALDListNode_t *prev; /* pointer to the previous item in the list */
} OALDListNode_t;

typedef struct OALDListNode_t *OALDListNode;

typedef boolean (*OALDListTraversal_fn_t)(OALDListNode item, void *arg);
typedef void (*OALDListRemoveAll_fn_t)(OALDListNode item, void *arg);

/** @internal
    Insert an item into a DList.
 
    @note Items inserted into the DList always appear at the front.
 
    @param head Pointer to a DList head (first item in the list).  Must not be NULL.
    @param item The item to insert into the list.  Must not be NULL.
**/ 
extern void OALDList_Insert(OALDListNode *head, OALDListNode item);

/** @internal
    Remove an item from the DList.
 
    @param head Pointer to a list head (first item in the list).  Must not be NULL.
    @param item The item to remove.  Must not be NULL.
**/
extern void OALDList_Remove(OALDListNode *head, OALDListNode item);

/** @internal
    Remove all items in a DList and call a callback function for each item removed.

    @param head Pointer to a DList to remove all items from.  Must not be NULL.
    @param fn The callback function called for each item removed.
    @param arg Context for the function @a fn.
*/
extern void OALDList_RemoveAll(OALDListNode *head, OALDListRemoveAll_fn_t fn, void *arg);

/** @internal
    Traverse a list and call a callback function for each item in the list.

    @param head The DList to traverse.
    @param fn The function to call for each item in the DList.
    @param arg Context for the function @a fn.

    @retval TRUE    if the traversal was told to exit early
    @retval FALSE   the traversal ran to completion
*/
extern boolean OALDList_Traverse(OALDListNode head, OALDListTraversal_fn_t fn, void *arg);

/**  @internal
    Traverse a DList and return a pointer to the node where the callback function
    returns TRUE.

    @param head The DList to traverse.
    @param fn The function to call for each item in the DList.
    @param arg Context for the function @a fn.

    @return The pointer to the node in the DList where the callback function returned TRUE.
    
    @retval NULL    @a fn never returns TRUE
*/
extern OALDListNode OALDList_Retrieve(OALDListNode head, OALDListTraversal_fn_t fn, void *arg);

/**  @internal
    Traverse a DList and return a pointer to the node where the callback function
    returns TRUE.  Also, remove the node from the linked list.

    @param head The DList to traverse.  Must not be NULL.
    @param fn The function to call for each item in the DList.  Must not be NULL.
    @param arg Context for the function @a fn.

    @return The pointer to the node in the list where the callback function returned TRUE.
    
    @retval NULL    @a fn never returned TRUE and nothing was removed from the DList.
*/
extern OALDListNode OALDList_FindAndRemove(OALDListNode *head, OALDListTraversal_fn_t fn, void *arg);


#ifdef  __cplusplus
}
#endif

#endif
