/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_DOFINTERNAL_H
#define DOF_DOFINTERNAL_H

#include <dof/core.h>
#include <internal/bridgeinternal.h>
#include <internal/enqueue.h>
#include <internal/resource_pool.h>
#include <internal/slist.h>
#include <internal/avl_tree.h>
#include <internal/oaltypesinternal.h>
#include <internal/security_membrane.h>
#include <internal/securityscope.h>

#ifdef  __cplusplus
extern "C" {
#endif

typedef struct DOFSystemNode_t
{
    OALSListNode    next;
    DOFSystem       system;
}DOFSystemNode_t;

typedef struct DOFServerNode_t
{
    OALSListNode    next;
    DOFServer       server;
}DOFServerNode_t;

typedef struct DOFAuthenticatorListItem_t
{
    OALSListNode        next;
    DOFAuthenticator    authenticator;
}DOFAuthenticatorListItem_t;

extern PCRMutex ref_count_mutex;

typedef struct DOF_t
{
    DOFConfig                       config;
    DOFSecurityScope                defaultSecurityScope;
    DOFSecurityScope                emptySecurityScope;
    OALOperationBridge              defaultBridge;
    DOFOperationFilter              defaultFilter;
    OALSListNode                    authenticators; /* OALSList<DOFAuthenticatorListItem_t> */
    OALSListNode                    domainsHead; /* DOFDomains keyed by DOFCredentials */
    OALSListNode                    systems;
    OALSListNode                    connection_head;
    OALSListNode                    server_head;
    OALAVLTree                      domainListeners;
    OALAVLTree                      interestListeners;
    PCREvent                        systemsGoneEvent;
    OALSListNode                    servers;
    DOFServerManager                serverManager;
    OALThreadPool                   threadPool;
    OALEnqueue                      enqueue;
    PCRMutex                        runtime_mutex;
    uint64                          inputByteCount;
    uint64                          outputByteCount;
    uint64                          inputPacketCount;
    uint64                          outputPacketCount;
    uint32                          maxOperations;
    uint32                          authenticatorsCnt;
    uint32                          domainsCnt;
    uint32                          remainingSystemCleanupCount;
    uint32                          maxConn;

    struct
    {
        unsigned int IsNodeDown     : 1;
    } flags;
}DOF_t;

typedef struct DOFEnqueueContext_t
{
    OALEnqueueItem_t super;
    OALEnqueue enqueue;
} DOFEnqueueContext_t;

extern void DOF_DestroyEnqueue(OALThreadPoolItem node, boolean shutdown);


/**
 * @internal
 * This is used to get the current only dof
 *
 * @note ONLY TEMP, will be removed when multiple DOFs are handled
 *
 * @return DOF our dof
 */
extern DOF DOF_GetDof( void );

extern DOFConfig DOF_GetConfig( const DOF dof );

/* DOF and DOFDomain */
extern void DOF_RemoveDomain( DOF self, DOFDomain domain );

extern void DOF_NotifyNewDomain( DOF self, DOFCredentials creds, DOFConnectionState state );

/**
 * @internal
 * Return the configured default partition for the DOF. This always represents
 * a non-secure partition and domain.  This can be used for creating other
 * non-secured partitions.  In order to create a secure partition, a DOFAuthenticator
 * must exist and be used.
 *
 * In most cases the default partition is the root of a domain, but this is not
 * required.  The default partition can be configured to any valid partition
 * (root or non-root). Do not destroy the default partition. It's valid for the
 * life of the DOF.
 *
 * @param dof The DOF instance.  This must not be NULL.
 * @return The configured default partition of the DOF.
 *
 * @see DOFAuthenticator_GetRootPartition
 */
extern DOFSecurityScope DOF_GetDefaultSecurityScope( DOF dof );

extern DOFSecurityScope DOF_GetEmptySecurityScope( DOF dof );

extern OALOperationBridge DOF_GetDefaultOperationBridge( DOF dof );

extern DOFOperationFilter DOF_GetDefaultOperationFilter( DOF dof );

/**
 * @internal
 * Adds a newly created system to the dof list of systems
 *
 * @param dof the DOF
 * @param system the system to add
 *
 * @return DOFSystem The system if worked
 *                   NULL otherwise
 */
extern boolean DOF_AddSystem( DOF dof, DOFSystem system );

/**
 * @internal
 * Removes a system from the dof's list of systems
 *
 * @param dof The DOF
 * @param system the system to remove
 */
extern void DOF_RemoveSystem( DOF dof, DOFSystem system );

extern void DOF_CleanupSystems( DOF dof );

extern boolean DOF_AddServer( DOF dof, DOFServer srv );

/**
 * @internal
 * Removes a server from the dof's list of server
 *
 * @param dof The DOF
 * @param server the server to remove
 */
extern void DOF_RemoveServer( DOF dof, DOFServer server );

/**
 * @internal
 * Dispatches interest Operations to the systems that fit its DOFPartition
 *
 * @param dof the dof to use
 * @param op the operation to send to systems
 */
extern void DOF_HandleInterestRequest( DOF dof, DOFOperation op);

/**
 * @internal
 * Removes a partition from the system
 *
 * @param dof The DOF
 * @param partition The Partition to Remove
 */
extern void DOFPartitionTree_RemovePartition( DOF dof, DOFPartition partition );

/**
 * @internal
 * Creates a DOFConfig from given dof
 *
 * @param dof the DOF to create a static config too.
 *
 * @return DOFConfig a static config to return from DOF_CreateConfig
 */
extern DOFConfig DOFConfig_CreateInternal( DOF dof );

extern void DOFLogListener_Init( void );
extern void DOFLogListener_Shutdown( void );
extern void DOFLog_RemoveAllLogListeners( void );

extern void DOF_Inc_InputByteCount( DOF self, uint32 count );
extern void DOF_Inc_OutputByteCount( DOF self, uint32 count );

extern uint32 DOF_GetMaxOperations( DOF dof );
extern uint16 DOF_GetMaxConnections( DOF dof );

extern void DOF_Lock(void);
extern void DOF_Unlock(void);

extern GUID DOF_GenerateGUID(void);

extern void DOF_LogMessage( DOF self, DOFLogLevel priority, const char * message );

extern boolean DOF_InterestListeners_Init( DOF self );
extern void DOF_InterestListeners_Shutdown( DOF self );
extern void DOF_ProcessInterest( DOF self, DOFOperation reqOp );

extern boolean DOF_IsNodeDown( DOF dof );

extern DOFConfig DOFConfig_Copy( DOFConfig self );
extern DOFState DOFState_Create( DOF dof );

/**
 * Report whether or not parameter validation is currently enforced.
 * @param self  The DOF.  This must not be NULL.
 * @return TRUE if parameter validation is currently enabled.
 *         FALSE if parameter validation is currently disabled.
 */
extern boolean DOF_IsParameterValidation( const DOF self );

extern OALOperationBridge DOF_CreateBridge(DOF                  dof,
                                           DOFOperationBridgeConfig   bridgeConfig,
                                           OALChannel           channel,
                                           DOFAddress           address,
                                           uint32               timeout,
                                           DOFException       * pException);

extern boolean DOF_RemoveAuthenticator( DOF dof, DOFAuthenticator auth );

#ifdef  __cplusplus
}
#endif

#endif

