/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_ITEMINTERNAL_H
#define DOF_ITEMINTERNAL_H


#include <dof/core.h>
#include <internal/stream.h>
#include <internal/interfaceinternal.h>


#ifdef  __cplusplus
extern "C" {
#endif


/** @internal @addtogroup DOFInterface
@{
**/


#define DOFITEM_PROPERTY    0  /**< @internal This type is stored in property item descriptor. @ingroup DOFInterfaceProperty **/
#define DOFITEM_EVENT       1  /**< @internal This type is stored in event item descriptor. @ingroup DOFInterfaceEvent **/
#define DOFITEM_METHOD      2  /**< @internal This type is stored in method item descriptor. @ingroup DOFInterfaceMethod **/
#define DOFITEM_EXCEPTION   3  /**< @internal This type is stored in exception item descriptor. @ingroup DOFInterfaceException **/

#define DOFITEM_ACCESS_READ  0x01  /**< @internal The property is readable. **/
#define DOFITEM_ACCESS_WRITE 0x02  /**< @internal The property is writable. **/

#define DOFItem_GetID( itm )                                (((DOFItem)itm)->itemID)
#define DOFItem_GetType( itm )                              (((DOFItem)itm)->itemType)

typedef const uint8 * DOFItemField;

/** @internal
    Describes an item within an interface defintion.  Allows matching iid to be determined from item pointer.
    @ingroup DOFItem
**/
struct DOFItem_t
{
    DOFInterface iface;             /**< The interface to which this item is associated. **/
    DOFType *    types[2];          /**< Points to the type(s) for the item. Only methods use element[1], for output parameters.  Properties and typedefs use element[0] as a DOFType instead of DOFType *. **/
    DOFItemID    itemID;            /**< The Item ID. **/
    uint8        typeCount[2];      /**< Count of parameter types.  Only methods use element 1, for output parameters. **/
    uint8        itemType;          /**< Property, event, method, exception, or typedef. **/
    uint8        access;            /**< Property access control. **/
};

typedef struct DOFItem_t DOFInterfaceProperty_t;     /**< @internal Describes a Property within an interface defintion.  @see DOFItem_t  @ingroup DOFInterfaceProperty **/
typedef struct DOFItem_t DOFInterfaceMethod_t;       /**< @internal Describes a Method within an interface defintion.  @see DOFItem_t  @ingroup DOFInterfaceMethod **/
typedef struct DOFItem_t DOFInterfaceEvent_t;        /**< @internal Describes an Event within an interface defintion.  @see DOFItem_t  @ingroup DOFInterfaceEvent **/
typedef struct DOFItem_t DOFInterfaceException_t;    /**< @internal Describes an Exception within an interface defintion.  @see DOFItem_t  @ingroup DOFInterfaceException **/

typedef const struct DOFItem_t * DOFItem;

/** @internal
 *  Returns if property is readable
 *
 * @param prp property to check if its readable. This must not be NULL.
 *
 * @return boolean TRUE if Readable FALSE otherwise
 */
extern boolean       DOFInterfaceProperty_IsReadable                        ( const DOFInterfaceProperty prp );

/** @internal
 * Returns if property is Writable
 *
 * @param prp property to check if its Writable. This must not be NULL.
 *
 * @return boolean TRUE if Writable FALSE otherwise
 */
extern boolean       DOFInterfaceProperty_IsWritable                        ( const DOFInterfaceProperty prp );

/** @internal
   Get event parameter count from a given DOFInterfaceEvent description.

   @param evt       The event to get the parameter count for. This must not be NULL.

   @return uint8    The parameter count for the event.
**/
extern uint8         DOFInterfaceEvent_GetParameterCount                    ( const DOFInterfaceEvent evt );

/** @internal
   Get method input parameter count from a given DOFInterfaceMethod description.

   @param mtd       The method to get the input parameter count for. This must not be NULL.

   @return uint8    The input parameter count.
**/
extern uint8         DOFInterfaceMethod_GetInputParameterCount              ( const DOFInterfaceMethod mtd );

/** @internal
   Get method output parameter count from a given DOFInterfaceMethod description.

   @param mtd           This must not be NULL.

   @return uint8
**/
extern uint8         DOFInterfaceMethod_GetOutputParameterCount             ( const DOFInterfaceMethod mtd );

/** @internal
   Get exception parameter count from a given DOFInterfaceException description.

   @param xcp       The exception to get the parameter count for.  This must not be NULL.

   @return uint8    The Parameter count for the exception.
**/
extern uint8         DOFInterfaceException_GetParameterCount                ( const DOFInterfaceException xcp );

extern DOFType       DOFInterfaceProperty_GetType                           ( const DOFInterfaceProperty prp );
extern DOFType *     DOFInterfaceEvent_GetParameterTypeList                 ( const DOFInterfaceEvent evt );

/** @internal
    Allocate and create an input parameter list of DOFValue elements from incoming packet stream, based on given DOFInterfaceMethod.

    @param mtd    DOFInterfaceMethod that specifies data content format of the incoming stream. This must not be NULL.
    @return Array of @ref DOFValue elements.
**/
extern DOFType *     DOFInterfaceMethod_GetInputParameterTypeList           ( const DOFInterfaceMethod mtd );

/** @internal
    Allocate and create an output parameter list of DOFValue elements from incoming packet stream, based on given DOFInterfaceMethod.

    @param mtd    DOFInterfaceMethod that specifies data content format of the incoming stream.  This must not be NULL.
    @return Array of @ref DOFValue elements.
**/
extern DOFType *     DOFInterfaceMethod_GetOutputParameterTypeList          ( const DOFInterfaceMethod mtd );
extern DOFType *     DOFInterfaceException_GetParameterTypeList             ( const DOFInterfaceException xcp );

/** @internal
    Allocate and create DOFValue from incoming packet stream, based on given DOFInterfaceProperty.

    @param prp    DOFInterfaceProperty that specifies data content format of the incoming stream. This must not be NULL.
    @param stream Incoming data used to fill the DOFValue.
    @param failed [OUT] Set to zero (FALSE) when value was successfully unmarshaled (including a NULL value), or non-zero (TRUE) if the unmarshal failed.

    @return DOFValue
**/
extern DOFValue      DOFInterfaceProperty_CreateValueFromStream             ( const DOFInterfaceProperty prp, OALStream stream, boolean * failed );

/** @internal
    Allocate and create list of DOFValue elements from incoming packet stream, based on given DOFInterfaceEvent.

    @param evt    DOFInterfaceEvent that specifies data content format of the incoming stream. This must not be NULL.
    @param stream Incoming data used to fill the list of DOFValues.

    @return Array of @ref DOFValue elements.
**/
extern DOFValue *    DOFInterfaceEvent_CreateValueListFromStream            ( const DOFInterfaceEvent evt, OALStream stream );
extern DOFValue *    DOFInterfaceMethod_CreateInputParamValueListFromStream ( const DOFInterfaceMethod mtd, OALStream stream );
extern DOFValue *    DOFInterfaceMethod_CreateOutputParamValueListFromStream( const DOFInterfaceMethod mtd, OALStream stream );

/** @internal
    Allocate and create a list of DOFValue elements from incoming packet stream, based on given DOFInterfaceException.

    @param xcp    DOFInterfaceException that specifies data content format of the incoming stream. This must not be NULL
    @param stream Incoming data used to fill the list of DOFValues.

    @return Array of @ref DOFValue elements.
**/
extern DOFValue *    DOFInterfaceException_CreateValueListFromStream        ( const DOFInterfaceException exp, OALStream stream );

extern boolean DOFInterfaceItem_IsInputParameterListCompatible( const DOFItem item, uint8 parameterCount, const DOFValue * parameters );
extern boolean DOFInterfaceItem_IsOutputParameterListCompatible( const DOFItem item, uint8 parameterCount, const DOFValue * parameters );

/**
    Get Interface Descriptor of an Item.

    @note Passing in a value that is not a DOFItem derivative results in undefined behavior.

    @param  item The Item from which to obtain information. THis should not be NULL.

    @return The Interface ID; or NULL if
            - its stored item descriptor is not valid.

    @see    DOFItem_GetItemID
**/
extern DOFInterface     DOFItem_GetInterface( const DOFItem item );

/**
    Get InterfaceID of an Item.

    @param  item The Item from which to obtain information. This must not be NULL.
    @return The Interface ID; or NULL if
            - its stored item descriptor is not valid.
    @see    DOFItem_GetItemID
**/
extern DOFInterfaceID   DOFItem_GetInterfaceID( const DOFItem item );
extern boolean          DOFItem_RetainDesc( const DOFItem item );
extern void             DOFItem_ReleaseDesc( const DOFItem item );

extern void             DOFItem_Destroy( struct DOFItem_t * item );


extern DOFValue * DOFInterfaceItem_CreateValueListFromStream( const DOFItem item, OALStream stream, boolean evtNotXcp );


/**
@}
**/


#ifdef  __cplusplus
}
#endif


#endif /* DOF_ITEMINTERNAL_H */
