/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef _DOF_MARSHAL_H
#define _DOF_MARSHAL_H

#include <dof/core.h>
#include <internal/stream.h>

#ifdef  __cplusplus
extern "C" {
#endif

/* @internal @addtogroup OALMarshal
@{
**/


#define OALMARSHAL_INVALID_VALUE    0xFFFFFFFF


/** @internal.
    Adjust a given pointer of a compressed short data entity to point to the first byte after that entity.
    @param pBuf Pointer to the first byte of the compressed short data entity. This must not be NULL.
    @return Pointer to the first byte after that entity.
**/
#ifdef DOXYGEN_MACRO_FUNCTIONS
extern const uint8 * OALMarshal_StepOverCompressedShort( const uint8 * pBuf );/*Only 1 user. For now use define, not function*/
#else
#define OALMarshal_StepOverCompressedShort( pBuf ) ((const uint8 *)((const uint8 *)pBuf + ((*(const uint8 *)pBuf & 0x80) ? 2 : 1)))
#endif


/** @internal.
    Adjust a given pointer of a compressed long data entity to point to the first byte after that entity.
    @param pBuf Pointer to the first byte of the compressed long data entity. This must not be NULL.
    @return Pointer to the first byte after that entity.
**/
extern const uint8 * OALMarshal_StepOverCompressedLong( const uint8 * pBuf );

/*
**  "Public" Type and Method Declarations (though not exposed in the API)
*/

/** @internal
    Determine the compressed size of a 2-, 3-, or 4-byte value.

    @param  maxSize  The maximum size on the wire of the value data.
    @param  value    The value to be marshaled.

    @return The marshaled (compressed) size.
**/
extern uint8  OALMarshal_GetCompressedValueSize( uint8 maxSize, uint32 value );

/** @internal
    Marshal a compressed 2-, 3-, or 4-byte value.

    @param  maxSize    The maximum size on the wire of the value data.  Must be greater than or equal to 2 and less than or equal to 4.
    @param  value      The value to be marshaled.
    @param  bufLength  The length of the buffer in which to marshal @a value.
    @param  buffer     The buffer in which to marshal @a value.  Must not be NULL.

    @return The marshaled size, or 0 if it does not fit in the buffer.
**/
extern uint32 OALMarshal_CompressValue( uint8 maxSize, uint32 value, uint32 bufLength, uint8 * buffer );

/**
    Unmarshal a compressed 2-, 3-, or 4-byte value.

    @param  maxSize    The maximum size on the wire of the value data.  Must be greater than or equal to 2 and less than or equal to 4.
    @param  bufLength  [IN] The length of the buffer in which the value data is marshaled; [OUT] the marshaled size.  Must not be NULL.
    @param  buffer     The buffer in which the value data is marshaled.  Must not be NULL.

    @return The value data that was unmarshaled, or 0xFFFFFFFF (OALMARSHAL_INVALID_VALUE) if there was insufficient
            space in the buffer to read a valid value.
**/
extern uint32 OALMarshal_UncompressValue( uint8 maxSize, uint32 * bufLength, const uint8 * buffer );

/* If the OALMarshal_StreamS* functions do not return DOFERROR_OK, then an error occurred. */
extern DOFError OALMarshal_StreamCompress2(OALStream stream, uint16 val);
extern DOFError OALMarshal_StreamCompress3(OALStream stream, uint32 val);
extern DOFError OALMarshal_StreamCompress4(OALStream stream, uint32 val);

/* If the OALMarshal_StreamU* functions return 0xffffffff, then an error occurred. */
extern DOFError OALMarshal_StreamUncompress2(OALStream stream, uint16 *val);
extern DOFError OALMarshal_StreamUncompress3(OALStream stream, uint32 *val);
extern DOFError OALMarshal_StreamUncompress4(OALStream stream, uint32 *val);

extern boolean OALMarshal_MarshalValue(DOFValue    value,
                                       DOFType     type,
                                       uint8    ** pBuffer,
                                       uint32   *  pSize);

extern boolean OALMarshal_MarshalValues(DOFValue *  values,
                                        DOFType  *  types,
                                        uint16      count,
                                        uint8    ** pBuffer,
                                        uint32   *  pSize);

/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* _DOF_MARSHAL_H */

