/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_INTERNAL_OAP_V1_H
#define DOF_INTERNAL_OAP_V1_H

/** @internal @addtogroup Protocols
@{
**/

/** @internal @addtogroup grp_oap_v1 OAP
@{
**/

#ifdef  __cplusplus
extern "C" {
#endif

typedef struct OAP_V1_t * OAP_V1;

#include <dof/core.h>
#include <internal/blackboarddata.h>
#include <internal/channel.h>
#include <internal/operationinternal.h>
#include <internal/packet.h>
#include <internal/protocollayerresult.h>
#include <internal/stream.h>
#include <internal/timer.h>

    /*Protocol - uint16 values */
#define OAP_V1_APPID 1              /**< Protocol ID of Object Access Protocol (OAP) **/
#define OAP_ALLOWED_OAP 0

    /*OAP Options - bit values */
#define OAP_OPT_ALL_DEFINES 1       /**< An OAP ESP option **/
#define OAP_OPT_ALL_SECURITY 2      /**< An OAP ESP option **/

    /*Allowed - array index */
#define OAP_ALLOWED_ALL_DEFINES 1
#define OAP_ALLOWED_ALL_SECURITY 2


typedef struct OAP_V1_t
{
    BlackboardData blackboardData;

    boolean isConnectionless;
} OAP_V1_t;

typedef enum OAP_V1_RequestType
{
    OAP_V1_UNKNOWN,

    OAP_V1_UNICAST,
    OAP_V1_BROADCAST,
    OAP_V1_MULTICAST
} OAP_V1_RequestType;

/** Create a new OAP layer for some protocol stack.
 *
 *  @param[in]      blackboardData
 *      common data structure shared by each layer of a protocol stack
 *
 *  @param[in]      isConnectionless
 *      flag to show whether this protocol layer is connected or connectionless
 *
 *  @return         NULL if unable to create and initialize the instance, a pointer to the new
 *                  OAP_V1 otherwise
 */
extern OAP_V1 OAP_V1_Create(BlackboardData blackboardData,
                            boolean        isConnectionless);

/** Destroy and clean up an OAP layer of some protocol stack.
 *
 *  @param[in,out]  self
 *      An OAP layer to clean up and destroy.  If this is NULL, nothing will be done.
 */
extern void OAP_V1_Destroy(OAP_V1 oap);

extern boolean OAP_V1_Init(void);

extern void OAP_V1_Shutdown(void);

/** Parse and process the OAP portion of some data that came over a transport.
 *
 *  @param[in]      self
 *      instance of an OAP to parse the stream
 *
 *  @retval PLRESULT_OKDONOTHING    success
 *  @retval PLRESULT_ERROR          failure
 *  @retval PLRESULT_OKSENDDOWN     send data back to source
 */
extern ProtocolLayerResult OAP_V1_Receive(OAP_V1 self);

/** Write the OAP portion of communication between two nodes.
 *
 *  @param[in]      self
 *      the instance OAP for the protocol stack currently being used
 *
 *  @param[in]      operation
 *      the operation that should be transmitted to another node
 *
 *  @param[in]      isProviderOperation
 *      TRUE if we are sending traffic to a provider, FALSE if traffic is going to a requestor
 *
 *  @param[in]      result
 *      the result coming from a provider and going back to the requestor
 *
 *  @retval PLRESULT_OKDONOTHING    there were no problems but nothing should be sent
 *  @retval PLRESULT_OKSENDDOWN     there were no problems and something does need to be sent
 *  @retval PLRESULT_ERROR          an error occurred
 */
extern ProtocolLayerResult OAP_V1_Send(OAP_V1       self,
                                       DOFOperation operation,
                                       boolean      isProviderOperation,
                                       DOFResult    result);

extern OALTimer OAP_V1_GetTimer(void);

extern DOFOperation OAP_V1_GetExceptionOperation(void);

extern OAP_V1_RequestType OAP_V1_DetermineRequestType(DOFObjectID objectId);

#ifdef  __cplusplus
}
#endif

/**
@}
**/

/**
@}
**/

#endif /* _DOF_OAP_H */
