/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef _DOF_OPERATIONINTERNAL_H
#define _DOF_OPERATIONINTERNAL_H

#include <dof/core.h>
#include <dof/permission.h>
#include <dof/operation.h>
#include <dof/control.h>
#include <internal/avl_tree.h>
#include <internal/channel.h>
#include <internal/iteminternal.h>
#include <internal/objectidinternal.h>
#include <internal/observable.h>
#include <internal/slist.h>
#include <internal/blackboarddata.h>
#include <internal/stream.h>
#include <internal/threadpool.h>
#include <internal/source.h>
#ifdef  __cplusplus
extern "C" {
#endif

/** @addtogroup DOFOperation
@{
**/

/* OALOperationManager stuff */
boolean OALOperationManager_Init(void);
boolean OALOperationManager_Init2(void);
void OALOperationManager_Shutdown(void);
void OALOperationManager_Shutdown2(void);
void OALOperationManager_CancelAll(void);

void OALOperationManager_Traverse_All(OALAVLMultiTree_Traverse_All_fn_t traverseFn, void * context);

DOFError OALOperationManager_AddOperation(DOFOperation op, const DOFOperationID opid, boolean need_lock);
void OALOperationManager_RemoveOperationAndUnlock( DOFOperation op );
DOFOperation OALOperationManager_CheckExistingOperationWithSize(const DOFOperationID opid, uint16 opid_size);
DOFOperation OALOperationManager_CheckExistingOperation(DOFObjectIDDomain domain, DOFOperationID opid);
DOFError _operation_extend( DOFOperation op, uint32 newTimeout, boolean notOnTimerItemFiring );
DOFError DOFOperationManager_SetRetry(DOFOperation op, OALChannel channel);
DOFError OALOperationManager_RenameOperation(DOFOperation op, DOFObjectIDDomain domain, DOFOperationID new_id);
    /* Do not ever call this function!  Only for use by DOFOperation_Internal_Cancel. */
DOFError OALOperationManager_EnqueueOperationCancel(DOFOperation op);
DOFError OALOperationManager_EnqueueOperation(OALEnqueueItem node);

extern boolean OALOperationManager_SetupForwardingOperation(DOFOperation op);
/*
**  Type declarations
*/

/** Identifies the type of asynchronous operation represented by a DOFOperation. **/
typedef enum DOFOperationType {
    /*
       !!!! WARNING: !!!!
         Several code modules depend on the order of these, so don't change the order or
         add new ones without checking with other people and/or fixing *all* related code!
         Some code also depends on the relative ordering of the interest operation types
         (watch, activate, connect, and full connect)!
     */
    DOFOPERATIONTYPE_GET,           /**< Property Get operation. **/
    DOFOPERATIONTYPE_SET,           /**< Property Set operation. **/
    DOFOPERATIONTYPE_INVOKE,        /**< Method Invoke operation. **/
    DOFOPERATIONTYPE_REGISTER,      /**< Event Registration operation. **/
    DOFOPERATIONTYPE_SUBSCRIBE,     /**< Property Subscription operation. **/
    DOFOPERATIONTYPE_FINDINTERFACEDEFINITION, /**< @internal Find Interface Definition operation. **/
    DOFOPERATIONTYPE_WATCH,         /**< Watch-level Interest operation. **/
    DOFOPERATIONTYPE_ACTIVATE,      /**< Activate-level Interest operation. **/

    DOFOPERATIONTYPE_EXCEPTION,     /**< @internal Exception. **/

    DOFOPERATIONTYPE_BEGINPROVIDE,  /**< @internal Local Provide operation. **/
    DOFOPERATIONTYPE_PROVIDE,       /**< Provide operation. **/
    DOFOPERATIONTYPE_SIGNAL,        /**< @internal Signal. **/
    DOFOPERATIONTYPE_CHANGED,       /**< @internal Changed. **/

    DOFOPERATIONTYPE_CANCEL,        /**< @internal Cancel. **/
    DOFOPERATIONTYPE_CONNECT,       /**< Connect-level Interest operation. **/
    DOFOPERATIONTYPE_SESSION,       /**< Open Session operation. **/
    DOFOPERATIONTYPE_ADVERTISE,     /**< @internal Advertise operation. **/
    DOFOPERATIONTYPE_QUERY,         /**< Query operation. **/
    DOFOPERATIONTYPE_FULLCONNECT,   /**< Full connect interest request **/

    DOFOPERATIONTYPE_NOOP,          /**< &internal no-op operation. **/

    DOFOPERATIONTYPE_INVALID,       /**< Invalid or unknown operation type. **/

    /*
     *  ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** **
     *  ALL FOLLOWING TYPES ARE INTERNAL ONLY!  DO NOT USE IN APPLICATION CODE!
     *  ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** **
     */

    /* These are ONLY used by TEP */
    DOFOPERATIONTYPE_TEP_BEGIN,
    DOFOPERATIONTYPE_TEP_AUTHORIZE, /**< TEP-BeginSession Response. **/
    DOFOPERATIONTYPE_TEP_CONFIRM,
    DOFOPERATIONTYPE_TEP_CONFIRMRESPONSE,
    DOFOPERATIONTYPE_TEP_END,
    DOFOPERATIONTYPE_TEP_EXTEND,
    DOFOPERATIONTYPE_TEP_EXTENDAUTH, /**< TEP-ExtendSession Response. **/
    DOFOPERATIONTYPE_TEP_REJECT,
    DOFOPERATIONTYPE_TEP_SESSIONENDING,
    DOFOPERATIONTYPE_TEP_NOOP,
    DOFOPERATIONTYPE_TEP_NEGOTIATION_NOOP,
    DOFOPERATIONTYPE_TEP_LAST,   /* invalid */

    /* TRP only Operations */
    DOFOPERATIONTYPE_TRP_V1_PASS,
    DOFOPERATIONTYPE_TRP_V1_AS,
    DOFOPERATIONTYPE_TRP_V1_KEK,
    DOFOPERATIONTYPE_TRP_V1_RANDOM,
    DOFOPERATIONTYPE_TRP_V1_SESSION,
    DOFOPERATIONTYPE_TRP_V1_SECURITY_SCOPE,
    DOFOPERATIONTYPE_TRP_V1_CREDENTIAL,
    DOFOPERATIONTYPE_TRP_V1_LOCAL_DOMAIN,
    DOFOPERATIONTYPE_TRP_V1_REMOTE_DOMAIN,
    DOFOPERATIONTYPE_TRP_V1_VALIDATE_CREDENTIALS,
    DOFOPERATIONTYPE_TRP_LAST, /* invalid */

    /* These are for SGMP ONLY */
    /* Order of these opcode must match order found in sgmp_v1.h. */
    DOFOPERATIONTYPE_SGMP_V1_HEARTBEAT,         /* This must be left as the first SGMP type */
    DOFOPERATIONTYPE_SGMP_V1_EPOCHCHANGED,
    DOFOPERATIONTYPE_SGMP_V1_REKEY,
    DOFOPERATIONTYPE_SGMP_V1_REQUESTGROUP,
    DOFOPERATIONTYPE_SGMP_V1_REKEYEPOCH,
    DOFOPERATIONTYPE_SGMP_V1_REKEYMERGE,
    DOFOPERATIONTYPE_SGMP_LAST,                /* invalid */

    /* These are for CCMMODE ONLY */
    DOFOPERATIONTYPE_CCMMODE_V1_FIRST,         /* invalid, was DOFOPERATIONTYPE_CCMMODE_V1_PROBE. */
    DOFOPERATIONTYPE_CCMMODE_V1_QUERY,         /* opcode 1. */
    DOFOPERATIONTYPE_CCMMODE_V1_REQUEST,       /* opcode 2. */
    DOFOPERATIONTYPE_CCMMODE_V1_LAST,          /* invalid */

    /* for epp neg one only */
    DOFOPERATIONTYPE_EPP_NEGONE_SOURCELOST,
    DOFOPERATIONTYPE_EPP_NEGONE_NODEDOWN,
    DOFOPERATIONTYPE_EPP_NEGONE_CANCELALL,
    DOFOPERATIONTYPE_EPP_NEGONE_HEARTBEAT,
    DOFOPERATIONTYPE_EPP_NEGONE_PING,
    DOFOPERATIONTYPE_EPP_NEGONE_QUERY,
    DOFOPERATIONTYPE_EPP_NEGONE_SOURCEFOUND,
    DOFOPERATIONTYPE_EPP_NEGONE_LAST,

    DOFOPERATIONTYPE_ESP_V2_QUERY,
    DOFOPERATIONTYPE_ESP_V2_TERMINATE,
    DOFOPERATIONTYPE_ESP_V2_OPEN,
    DOFOPERATIONTYPE_ESP_LAST,

    DOFOPERATIONTYPE_ROUTEINFO
} DOFOperationType;

typedef struct DOFOperationList_t
{
    uint32          count;
    DOFOperation *  list;
}DOFOperationList;

typedef void (*DOFOperation_DestroyContext)(void * operationContext);

typedef struct CancelOperationContext_t
{
    OALEnqueueItem_t super;
    DOFOperation operation;
} CancelOperationContext_t;

typedef struct DOFOperationSource_t
{
    struct OALSListNode_t   *next; /* pointer to the next item in the list */
    OALChannel              channel;
    DOFAddress              address;
    OALPermissionScope      permissionGroups; /* scope ids for this source */
    DOFRequest              request;     /* router fills this when routing the 4 interest type ops to provider*/
    uint32                  oal_timeout;
    uint32                  timeoutGrace;
    OALAlias                alias;
    uint16                  distLossless;
    uint16                  distLossy;
    uint8                   retry_count;
    uint8                   sequence;
}DOFOperationSource_t;
typedef struct DOFOperationSource_t* DOFOperationSource;

typedef struct DOFOperation_t {
    OALObservable                   operation_will_close;
    void *                          context;

    DOFOperationSource              sources; /* This is head of an OALSList */

    DOFOperationType                operation_type;

    DOFOperation                    next_sub_operation;

    /* needed by DOF */
    DOFSecurityScope                securityScope;

    /* needed by all (most) */
    union {
        DOFObjectProvider           pcb;     /* 10 function pointers. */
        DOFObjectGetCallback        rcb_get; /* 2 function pointers. */
        DOFObjectSetCallback        rcb_set; /* 2 function pointers. */
        DOFObjectInvokeCallback     rcb_inv; /* 2 function pointers. */
        DOFObjectSubscribeCallback  rcb_sub; /* 2 function pointers. */
        DOFObjectEventCallback      rcb_reg; /* 2 function pointers. */
        DOFObjectDefineCallback     rcb_def; /* 2 function pointers. */
        DOFSystemQueryCallback      qcb;     /* 4 function pointers. */
        DOFObjectSessionCallback    scb;     /* 2 function pointers. */
        DOFSystemInterestCallback   icb;     /* 1 function pointer. */
        EPPNegOneQueryCallback      negOne;  /* 1 function pointer. */
    } callbacks;

    DOFInterfaceID                  interface_id;
    uint32                          proxySourceLostTime;

    /* needed by provider side */
    DOFObject                       provider_object;    /* don't set this if you aren't DOFSystem! */
    DOFInterface                    provider_interface; /* don't set this if you aren't DOFSystem! */
    DOFItem                         item;
    DOFItemID                       itemId;
    OALAVLTree                      requestTree; /* DOFSystem only user, filled only for some OAP request operations. */
    uint32                          minPeriod;   /* Property subscription requests specify this. */

    /* needed by requestor side */
    union {
        DOFType                     type;
        DOFType *                   types;
    } valtype;
    union {
        DOFValue                    value;
        DOFValue *                  values;
    } val;

    OALStream                       valueStream;
    /* needed by DOFSystem */
    DOFOperation                    myProviderOperation;
    OALObserver                     reqObserver;
    DOFQuery                        filter;
    DOFObjectIDList                 matchingObjIDs;
    PCRCondition                    notifyOpCompleteCond; /* Relies on NULL init then user must always return to NULL.*/
    uint32                          completeWaitInstanceCount;
    PCRCondition                    notifyOpAcknowledgedCond; /* Relies on NULL init. User must always return to NULL.*/
    uint32                          acceptWaitInstanceCount;
    DOFResult                       firstResult; /* Caches the first result received by the operation. */
    DOFException                    completeCallbackException;

    DOFPermission                   commandPerm;
    DOFPermission                   responsePerm;

    PCRMutex                        mutex;

    CancelOperationContext_t      * cancelContext;

    /* needed by router */
    union objectid_u
    {
        DOFObjectID                 object_id;
        DOFObjectID                 object_id_ref;
    } object_id;

    OALTimerItem                    timer_item; /* entry in the operation timeout table */
    OALTimerItem                    retry_timer_item; /* entry in router retry timer */
    OALTimerItem                    ackTimerItem;   /* used by OAP to check on long-term operation retries */

    uint32                          ref_count;

    GUID                            sessionID;   /* Session GUID.  Used only by open operation.  Need to try and optimize away */
    DOFObjectID                     sessionOID;  /* Augmented session ObjectID (base + Session attribute).  Used only by open operation.  Need to try and optimize away */
    DOFInterfaceID                  sessionIID;  /* Session type.  Used only by open operation.  Need to try and optimize away */
    DOFOperationControl             operationConfig;  /* Used by most operations */
    OALTimerItem                    opDelayTimerItem;
    OALAVLTree                      ackTree;

    DOFOperationID                  opID;
    DOFOperationID                  forwardingId;

    OALAVLTree                      providers;

    DOFOperationSessionDataTransform sessionTransformRequestor;
    DOFOperationSessionDataTransform sessionTransformProvider;

    void                          * sidKey;     /* This is only a key used only for epp neg one, the real sid is stored in the opID. */

    void                          * pbte;

    void                          * tepContext;

    DOFOperation_DestroyContext     tepContextDestroyFunction;

    DOFOperation_DestroyContext     contextDestroyFunction;

    union {
        uint32 flags;
        struct {
            unsigned int responseResultReceived                        : 1;
            unsigned int blockFollowOnIncomingSignalPDUs               : 1;
            unsigned int denyPacketReplayProtectionIfSecureUnicastSend : 1;
            unsigned int provider_oid_is_wildcard                      : 1;
            unsigned int provider_iid_is_wildcard                      : 1;
            unsigned int operation_removed_from_timer_table            : 1;
            unsigned int internally_created                            : 1;
            unsigned int timedOut                                      : 1;
            unsigned int cancelled                                     : 1;
            unsigned int completed                                     : 1;
            unsigned int completedAck                                  : 1;
            unsigned int denyTimeoutExtend                             : 1; /*Set two ways. Once set nothing clears it*/
            unsigned int value_type_is_not_list                        : 1; /* 1 is a value, 0 is a value list */
            unsigned int added_to_operation_table                      : 1;
            unsigned int dont_add_to_operation_table                   : 1;
            unsigned int activation_closing                            : 1;
            unsigned int new_retry_count                               : 1;
            unsigned int acknowledged                                  : 1;
            unsigned int fake_timeout                                  : 1;
            unsigned int timeout_hit                                   : 1;
            unsigned int cancel_enqueued                               : 1;
            unsigned int is_connectionless                             : 1;
            unsigned int oap_cancel                                    : 1;
            unsigned int no_execute                                    : 1;
            unsigned int lost                                          : 1;
            unsigned int securityRequiredSecureHeaderUnsecure          : 1;
            unsigned int already_routed                                : 1;
            unsigned int isFlooded                                     : 1;
            unsigned int commandCompletesOnAnyFinal                    : 1;
            unsigned int cancelOpAfterEppSend                          : 1;
            unsigned int subscribeOk                                   : 1;
            unsigned int observers_notified                            : 1;
        } bits;
    } flags;
} DOFOperation_t;

/*
**  "Public" Internal Type and Method Declarations
*/

DOFOperation DOFOperation_Get_Create(OALChannel requestor_channel, DOFOperationControl operationConfig,
                                     uint32 timeout, void * context, DOFObject obj,
                                     const DOFObjectGetCallback rcb, DOFInterfaceProperty prop);

DOFOperation DOFOperation_Set_Create(OALChannel requestor_channel, DOFOperationControl operationConfig,
                                     uint32 timeout, void * context, DOFObject obj,
                                     const DOFObjectSetCallback rcb, DOFInterfaceProperty prop);

DOFOperation DOFOperation_Invoke_Create(OALChannel requestor_channel, DOFOperationControl operationConfig,
                                        uint32 timeout, void * context, DOFObject obj, const DOFObjectInvokeCallback rcb,
                                        DOFInterfaceMethod method);

DOFOperation DOFOperation_Register_Create(OALChannel requestor_channel, DOFOperationControl operationConfig,
                                          uint32 timeout, void * context, DOFObject obj,
                                          const DOFObjectEventCallback rcb, DOFInterfaceEvent event);

DOFOperation DOFOperation_Subscribe_Create(OALChannel requestor_channel, DOFOperationControl operationConfig,
                                           uint32 timeout, void * context, DOFObject obj,
                                           const DOFObjectSubscribeCallback rcb, DOFInterfaceProperty prop, uint32 minPeriod);

DOFOperation DOFOperation_BeginProvide_Create(OALChannel provider_channel, uint32 timeout, void * context,
                                              DOFObject obj, const DOFObjectProvider pcb, DOFInterface iface);

DOFOperation DOFOperation_Query_Create(OALChannel query_channel, uint32 timeout, void * context,
                                       DOFQuery  filter, const DOFSystemQueryCallback qcb);

DOFOperation DOFOperation_FindInterfaceDefinition_Create(OALChannel requestor_channel, DOFOperationControl operationConfig,
                                                         uint32 timeout, void * context, DOFObject obj,
                                                         const DOFObjectDefineCallback rcb, DOFInterfaceID iid);

DOFOperation DOFOperation_Interest_Create(DOFOperationControl opConfig, OALChannel requestor_channel,
                                          DOFObjectID objectID, const DOFInterfaceID interfaceID,
                                          const DOFSystemInterestCallback icb, uint32 timeout,
                                          DOFInterestLevel level, void * context);

DOFOperation DOFOperation_Session_Create(OALChannel requestor_channel, DOFObject obj,
                                         const DOFInterface interface,
                                         DOFInterfaceID sessionIID,  DOFOperationControl operationConfig,
                                         uint32 timeout, const DOFObjectSessionCallback scb, void * ctx);

DOFOperation DOFOperation_Activate_CreateInternal(      OALChannel       requestor_channel,
                                                  const DOFInterfaceID   interfaceID,
                                                  const DOFObjectID      objectID,
                                                  const OALAlias         alias,
                                                        uint32           timeout,
                                                        uint32           timeoutGrace);

DOFOperation DOFOperation_Change_CreateInternal(      OALChannel       requestor_channel,
                                                const DOFInterfaceID   interfaceID,
                                                const DOFObjectID      objectID,
                                                const OALAlias         alias,
                                                      uint32           timeout,
                                                      uint32           timeoutGrace,
                                                      OALStream        valueStream,
                                                      DOFItemID        itemId);

DOFOperation DOFOperation_Connect_CreateInternal(      OALChannel       requestor_channel,
                                                 const DOFInterfaceID   interfaceID,
                                                 const DOFObjectID      objectID,
                                                 const OALAlias         alias,
                                                       uint32           timeout,
                                                       uint32           timeoutGrace);

DOFOperation DOFOperation_FullConnect_CreateInternal(      OALChannel       requestor_channel,
                                                     const DOFInterfaceID   interfaceID,
                                                     const DOFObjectID      objectID,
                                                     const OALAlias         alias,
                                                           uint32           timeout,
                                                           uint32           timeoutGrace);

DOFOperation DOFOperation_Define_CreateInternal(      OALChannel     requestor_channel,
                                                const DOFInterfaceID interfaceID,
                                                const DOFObjectID    objectID,
                                                      uint32         timeout,
                                                      uint32         timeoutGrace);

DOFOperation DOFOperation_Get_CreateInternal(      OALChannel       requestor_channel,
                                             const DOFInterfaceID   interfaceID,
                                             const DOFObjectID      objectID,
                                             const OALAlias         alias,
                                                   uint32           timeout,
                                                   uint32           timeoutGrace,
                                                   DOFItemID        itemId);

DOFOperation DOFOperation_Invoke_CreateInternal(      OALChannel       requestor_channel,
                                                const DOFInterfaceID   interfaceID,
                                                const DOFObjectID      objectID,
                                                const OALAlias         alias,
                                                      uint32           timeout,
                                                      uint32           timeoutGrace,
                                                      OALStream        valuesStream,
                                                      DOFItemID        itemId);

DOFOperation DOFOperation_Open_CreateInternal(      OALChannel       requestor_channel,
                                              const DOFInterfaceID   interfaceID,
                                              const DOFObjectID      objectID,
                                              const OALAlias         alias,
                                                    uint32           timeout,
                                                    uint32           timeoutGrace,
                                              const GUID *           sessionID,
                                              const DOFInterfaceID   sessionIID);

DOFOperation DOFOperation_Provide_CreateInternal(      OALChannel       requestor_channel,
                                                 const DOFInterfaceID   interfaceID,
                                                 const DOFObjectID      objectID,
                                                 const OALAlias         alias,
                                                       uint32           timeout,
                                                       uint32           timeoutGrace,
                                                       uint16           losslessHops,
                                                       uint16           lossyHops);

DOFOperation DOFOperation_Advertise_CreateInternal(      OALChannel      requestor_channel,
                                                   const DOFInterfaceID  interfaceID,
                                                   const DOFObjectID     objectID,
                                                         uint32          timeout,
                                                         uint32          timeoutGrace,
                                                         DOFAddress      address);

DOFOperation DOFOperation_Set_CreateInternal(      OALChannel       requestor_channel,
                                             const DOFInterfaceID   interfaceID,
                                             const DOFObjectID      objectID,
                                             const OALAlias         alias,
                                                   uint32           timeout,
                                                   uint32           timeoutGrace,
                                                   OALStream        valueOrStream,
                                                   DOFItemID        itemId);

DOFOperation DOFOperation_Signal_CreateInternal(      OALChannel       requestor_channel,
                                                const DOFInterfaceID   interfaceID,
                                                const DOFObjectID      objectID,
                                                const OALAlias         alias,
                                                      uint32           timeout,
                                                      uint32           timeoutGrace,
                                                      OALStream        valuesStream,
                                                      DOFItemID        itemId);

DOFOperation DOFOperation_Subscribe_CreateInternal(      OALChannel       requestor_channel,
                                                   const DOFInterfaceID   interfaceID,
                                                   const DOFObjectID      objectID,
                                                   const OALAlias         alias,
                                                         uint32           timeout,
                                                         uint32           timeoutGrace,
                                                         DOFItemID        itemId,
                                                         uint16           minimumDelta,
                                                         boolean          not_value_list);

DOFOperation DOFOperation_Watch_CreateInternal(      OALChannel       requestor_channel,
                                               const DOFInterfaceID   interfaceID,
                                               const DOFObjectID      objectID,
                                               const OALAlias         alias,
                                                     uint32           timeout,
                                                     uint32           timeoutGrace);

extern boolean DOFOperationSession_SetDataTransformProvider(DOFOperation                     sessionOperation,
                                                            DOFOperationSessionDataTransform transform);

DOFOperation DOFOperation_SourceLost_Create(void);
DOFOperation DOFOperation_SourceFound_Create(void);
DOFOperation DOFOperation_NodeDown_Create(void);
DOFOperation DOFOperation_CancelAll_Create(void);
DOFOperation DOFOperation_EppHeartbeat_Create(DOFAddress address);
DOFOperation DOFOperation_EppPing_Create(DOFAddress address);
DOFOperation DOFOperation_EppQuery_Create(OALChannel channel, DOFAddress address);

DOFOperation DOFOperation_ESP_V2_Query_Create(DOFAddress address, uint32 retry);
DOFOperation DOFOperation_ESP_V2_Terminate_Create(DOFAddress address);
DOFOperation DOFOperation_ESP_V2_Open_Create(DOFAddress address);
DOFOperation DOFOperation_ESP_V2_Open_CreateInternal(DOFOperationID opID, DOFAddress address, uint32 timeout);

CancelOperationContext_t * DOFOperationManager_CreateCancelContext(DOFOperation op);

extern void  DOFOperation_Lock(DOFOperation op);
extern void DOFOperation_Unlock(DOFOperation op);
extern void _DOFOperation_Cancel(DOFOperation op);

#ifndef DOFOPERATION_DEBUG_REFCOUNT

boolean DOFOperation_Retain(DOFOperation op);
void DOFOperation_Internal_Destroy(DOFOperation op);
#else
#define DOFOperation_Retain(op) _DOFOperation_Retain(op, __func__, __FILE__, __LINE__)
#define DOFOperation_Internal_Destroy(op) _DOFOperation_Internal_Destroy(op, __func__, __FILE__, __LINE__)
boolean _DOFOperation_Retain(DOFOperation op, const char *fn_name, const char *filename, unsigned int line_number);
void _DOFOperation_Internal_Destroy(DOFOperation op, const char *fn_name, const char *filename, unsigned int line_number);
#endif

boolean DOFOperation_SendOAPNoOp( OALChannel channel );

void DOFOperation_Internal_Cancel(DOFOperation op);

/** @internal
    Return the Operation's ID.

    This value is kept by the system to distinguish outstanding operations.
    There is usually no need for this information by the application.

    @param  op  The Operation.

    @return The Operation's ID.
**/
DOFOperationID DOFOperation_GetOperationID( const DOFOperation op );

void DOFOperation_SetLocalOID(DOFObjectIDSource newOid);

DOFError DOFOperationID_Marshal(OALStream stream, DOFOperationID id);
DOFOperationID DOFOperationID_Create(DOFOperationCount cnt, sid_t *sid, boolean implicit, boolean get_ref);
DOFOperationID DOFOperationID_CreateLocal(void);
boolean DOFOperationID_IsImplicit(DOFOperationID id);
boolean DOFOperationID_CanBeExplicit(DOFOperationID id);
void DOFOperationID_UpdateSid(DOFOperationID id, sid_t * newSid);
boolean DOFOperationID_IsSidEqual(DOFOperationID id, sid_t * sid);
DOFObjectID DOFOperationID_GetLocalOID(void);
sid_t *DOFOperationID_GetLocalSID(void);
sid_t *DOFOperationID_GetSID(DOFOperationID opid);

void *DOFOperationID_GetOPIDPtr(DOFOperationID opid);
DOFOperationCount DOFOperationID_GetCount( const DOFOperationID id );

/**
    Return the security scope that is associated with the Operation.

    @param  op  The Operation.  This must not be NULL.

    @return The associated security scope of the operation.  This will not be NULL.
**/
DOFSecurityScope DOFOperation_GetSecurityScope( const DOFOperation op );

boolean DOFOperation_Init(void);
void DOFOperation_Shutdown(void);
boolean DOFOperation_DOFInit(void);
void DOFOperation_DOFShutdown(void);

#ifndef DOFOPERATION_DEBUG_REFCOUNT
extern OALObserver DOFOperation_AddOpWillCloseObserver(DOFOperation op, OALObserver_fn_t fn);
#else
#define DOFOperation_AddOpWillCloseObserver(op, fn) _DOFOperation_AddOpWillCloseObserver(op, fn, __func__, __FILE__, __LINE__)
extern OALObserver _DOFOperation_AddOpWillCloseObserver(DOFOperation op, OALObserver_fn_t fn, const char *fn_name, const char *filename, unsigned int line_number);
#endif
extern OALObserver DOFOperation_AddOpWillCloseObserverOtherArg(DOFOperation op, OALObserver_fn_t fn, void *arg);

extern void DOFOperation_RemoveOpWillCloseObserver(DOFOperation op, OALObserver obs);

extern void DOFOperation_IncrementRetry(DOFOperation op);

/*  This is an internal function to set the oal_timeout field so that the delay sent in the EPP header is correct.
    Note: This must be called after DOFOperation_SetTimeout and OALOperationManager_AddOperation. */
void DOFOperation_SetDelay(DOFOperation op, uint32 timeout);

OALChannel DOFOperation_GetSourceChannel(DOFOperation op);

boolean DOFOperation_IsBetterSource(DOFOperation op,
                                    uint16       distLossless,
                                    uint16       distLossy);

uint16 DOFOperation_GetValueListCount( DOFOperation op );

/**
    Return the Object (see @ref DOFObject) that is associated with the Operation.

    Not valid for DOFSystem_BeginQuery Operations.

    @param  op  The Operation.

    @return The associated Object, or NULL if @a op is NULL or is a Query Operation.
**/
DOFObject DOFOperation_GetObject( const DOFOperation op );

/**
    Return the @ref DOFQuery that is associated with a Query Operation.

    Only valid for DOFSystem_BeginQuery Operations.

    @param  op  The Operation.

    @return The associated Filter, or NULL if @a op is NULL or is not a Query Operation.
**/
DOFQuery DOFOperation_GetQuery( const DOFOperation op );

/**
    Return the Method (see @ref DOFInterfaceMethod) that is associated with an Invoke Operation.

    Only valid for Invoke Operations.

    @param  op  The Operation.

    @return The associated Method, or NULL if @a op is NULL or is not an Invoke Operation.
**/
DOFInterfaceMethod DOFOperation_GetMethod( const DOFOperation op );

/**
    Return the Property (see @ref DOFInterfaceProperty) that is associated with a Get, Set, or Subscribe Operation.

    Only valid for Get, Set, or Subscribe Operations.

    @param  op  The Operation.

    @return The associated Property, or NULL if @a op is NULL or is not a Get, Set, or Subscribe Operation.
**/
DOFInterfaceProperty DOFOperation_GetProperty( const DOFOperation op );

/**
    Return the Event (see @ref DOFInterfaceEvent) that is associated with a Register Operation.

    Only valid for Register Operations.

    @param  op  The Operation.

    @return The associated Event, or NULL if @a op is NULL or is not a Register Operation.
**/
DOFInterfaceEvent DOFOperation_GetEvent( const DOFOperation op );

/**
    Return a the System (see @ref DOFSystem) associated with the Operation.

    @param  op  The Operation.

    @return The associated System, or NULL if @a op is NULL.
**/
DOFSystem DOFOperation_GetSystem( const DOFOperation op );

/**
    Return the Operation type (see @ref DOFOperationType).

    This can be used to determine the type of the operation.  In the context of
    most application callbacks, the operation type can be assumed.  However,
    there may be reasons for determining the type of an operation in other
    areas of the code.

    @param  op  The Operation.

    @return The Operation's type, or DOFOPERATIONTYPE_INVALID if @a op is NULL.
**/
DOFOperationType DOFOperation_GetType( const DOFOperation op );

DOFOperationControl DOFOperation_GetControl(DOFOperation op);

/**
    Get the Object ID portion of an Operation ID.

    @param id The Operation ID.  This should not be NULL.

    @return The Object ID portion of the Operation ID, or NULL if @a id was NULL.

    @exception  If @a id is NULL, NULL is returned.
**/
const uint8 * DOFOperationID_GetOIDBytes( const DOFOperationID id );

/**
 * Create an ObjectID from the ObjectID bytes stored in an OperationID.
 *
 * @mustfree{DOFObjectID_Destroy}
 *
 * @param id the Operation ID.  This must not be NULL.
 *
 * @return A @ref DOFObjectID.
 */
DOFObjectID DOFOperationID_GetObjectID( const DOFOperationID id );

/**
    Return the standardized wire-format length of an Operation ID.

    @param id The Operation ID.  This should not be NULL.

    @return The size, in byte count, of this Operation ID, or @c 0 if @a id is NULL.
**/
uint16 DOFOperationID_GetSize( const DOFOperationID id );
uint16 DOFOperationID_GetSizeMem(void);

/**
    Return a new copy of an existing Operation ID.

    @mustfree{DOFOperationID_Destroy}

    @param id The Operation ID.

    @return The new Operation ID copy, or NULL if @a id was NULL or a new ID could not be created.
**/
DOFOperationID DOFOperationID_Copy( DOFOperationID id );

/**
    Destroy an Operation ID.

    @param id The Operation ID. If NULL, this function does nothing.
**/
void DOFOperationID_Destroy(DOFOperationID id);

OALAVLTree DOFOperation_CreateObjectIDAVLTree( void );

DOFPermission _createNeededPermission(DOFOperationType opType,
                                      boolean          isResponse,
                                      DOFObjectID      oid,
                                      DOFInterfaceID   iid);

/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* _DOF_OPERATIONINTERNAL_H */
