/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_PACKET_H
#define DOF_PACKET_H

#include <dof/core.h>
#include <internal/stream.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @internal @addtogroup OALPacket
@{
**/

/*
**  "Public" Type and Method Declarations
*/
typedef struct OALPacket_t * OALPacket;

/* "Static" methods */

/*
*   Create a new OALPacket. Streams stored in this packet will be released when this packet is released.

    @mustfree{OALPacket_Release}

    @return The new @ref OALPacket.
    @retval NULL if failure to create the OALPacket.
    
    @internal
**/
extern OALPacket   OALPacket_Create            (void);

/* "Object" methods */
extern void        OALPacket_Retain            (OALPacket self);
extern void        OALPacket_Release           (OALPacket self);

/*
    This method prepends a stream to the OALPacket.

    @param  self    "this" OALPacket.  Must not be NULL.
    @param  stream  The stream to prepend.  Must not be NULL.
    
    @retval true    @a stream was successfully prepended to @a self    
    @retval false   Unable to prepend @a stream to @a self
    
    @internal
**/
extern boolean     OALPacket_PrependStream     (OALPacket self, OALStream stream);

/*
    This method appends a stream to the OALPacket.

    @param  self    "this" OALPacket.  Must not be NULL.
    @param  stream  The stream to append.  Must not be NULL.
    
    @retval true    @a stream was successfully appended to @a self    
    @retval false   Unable to append @a stream to @a self
    
    @internal
**/
extern boolean     OALPacket_AppendStream      (OALPacket self, OALStream stream);

/*
    Adds up the sizes of all OALStreams in this OALPacket.

    @param  self    "this" OALPacket.  Must not be NULL.

    @return Total length of the streams.
    
    @internal
**/
extern uint32      OALPacket_GetTotalStreamSize(OALPacket self);

/*
*   Copies the internal data of the packet into the provided buffer.

    @param  self            "this" OALPacket.  Must not be NULL.
    @param  destBuf         The buffer to store the copied data.  Must not be NULL.
    @param  pktDataOffset   The offset into the packet to start copying from.
    @param  length          The number of bytes to copy.
    
    @warning    @a pktDataOffset + @a length must not be greater than the value
                returned by OALPacket_GetTotalStreamSize

    @return The number of bytes actually copied.
    
    @internal
**/
extern uint32      OALPacket_CopyTo            (OALPacket self, uint8* destBuf, uint32 pktDataOffset, uint32 length);
/* OALStream   OALPacket_GetFlatStream     (OALPacket self); */

/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* DOF_PACKET_H */
