/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef _DOF_POINTERLIST_H
#define _DOF_POINTERLIST_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @internal @addtogroup OALPtrList
@{
**/

/*
**  "Public" Type and Method Declarations
*/
typedef struct OALPtrList_t *  OALPtrList;

/* "Static" methods */

/*
*   Create a new OALPtrList.

    @mustfree{OALPtrList_Destroy}

    @param  initialCapacity  Specifies how many elements to initialy allocate memory for.

    @return The new @ref OALPtrList.
    @retval NULL if failure to create the object.
    
    @internal
**/
extern OALPtrList  OALPtrList_Create(uint32 initialCapacity);

/* "Object" methods */

/*
    Destroy this OALPtrList.

    @param  self   The OALPtrList to destroy.  Must not be NULL.
    
    @internal
**/
extern void        OALPtrList_Destroy  (OALPtrList self);

/*
    Retrieve the maximum number of elements this OALPtrList can contain @em without
    allocating addiditional memory.

    @param self "this" OALPtrList.  Must not be NULL.

    @return the maximum number of elements this OALPtrList can contain @em without
            allocating addiditional memory.
            
    @internal
*/
extern uint32      OALPtrList_GetCapacity  (OALPtrList self);

/*
    Retrieve the number of elements contained in this list.

    @param self "this" OALPtrList.  Must not be NULL.

    @return the number of elements contained in this list.
    
    @internal
*/
extern uint32      OALPtrList_GetCount     (OALPtrList self);

/*
    Insert an element into the list at a given location.

    @param self     "this" OALPtrList.  Must not be NULL.
    @param item     The element to add to this list.
    @param index    The 0-based index location specifying where to add the element.
                    An index value of 0 will put the element at the front of the list.
                    An index value equal to the current element count will put the value
                    at the end of the list. For any other value (n), elements from n through
                    current element count will be shifted down this OALPtrList.

    @return         TRUE if successful, FALSE otherwise.
    
    @internal
*/
extern boolean     OALPtrList_InsertAt     (OALPtrList self, void * element, uint32 index);
extern void        OALPtrList_OverwriteAt  (OALPtrList self, void * element, uint32 index);

/*
    Removes and returns an element at a given location from the list.

    @param self     "this" OALPtrList.  Must not be NULL.
    @param index    The 0-based index location specifying where to remove the element from.
                    Must be less than the element count.

    @return         The element removed from this list.
    
    @internal
*/
extern void *      OALPtrList_RemoveAt     (OALPtrList self, uint32 index);

/*
    Returns (without removing) an element at a given location from the list.

    @param self     "this" OALPtrList.  Must not be NULL.
    @param index    The 0-based index location specifying where to retrieve the element from.
                    Must not less th an the element count.

    @return         The specified item from this list.
    
    @internal
*/
extern void *      OALPtrList_PeekAt       (OALPtrList self, uint32 index);

/*
    Returns the first index of a specified element in the list.

    @param self     "this" OALPtrList.  Must not be NULL.
    @param element  The element to find in the list.

    @return         The first index of the location of element in the list. If
                    element is not in the list then a value equal to count is returned.
                    
    @internal
*/
extern uint32      OALPtrList_FirstIndexOf (OALPtrList self, void * element);


/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* _DOF_POINTERLIST_H */
