/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_RESULTINTERNAL_H
#define DOF_RESULTINTERNAL_H


#include <dof/address.h>
#include <dof/core.h>
#include <dof/result.h>
#include <internal/channel.h>


#ifdef  __cplusplus
extern "C" {
#endif


/* @internal @addtogroup resultinternal
@{
**/


/*
**  "Public" Type and Method Declarations
*/


/**
 * Result type codes.  These help identify the data within a @ref DOFResult.
 * @ingroup DOFResult
 */
typedef enum DOFResultType_t
{
    DOFRESULTTYPE_COMPLETE    =     0,  /**< Identifies a @ref DOFResult as containing no data (acknowledgement only), as would be returned from a Set operation.  There are no values contained in the result. */
    DOFRESULTTYPE_NO_VALUE    =     1,  /**< Identifies a @ref DOFResult as containing no data (acknowledgement only), as would be returned from an Invoke operation with no outputs.  There is no list of @ref DOFValue elements contained in the result. **/
    DOFRESULTTYPE_VALUE       =     2,  /**< Identifies a @ref DOFResult as containing a @ref DOFValue, as would be returned from a Get or Subscribe operation. **/
    DOFRESULTTYPE_VALUE_LIST  =     3,  /**< Identifies a @ref DOFResult as containing a list of @ref DOFValue elements, as would be returned from an Invoke operation or reported by the signal of an Event. **/
    DOFRESULTTYPE_EXCEPTION   =     4,  /**< Identifies a @ref DOFResult as containing exception information.  This consists of a @ref DOFException, which may contain either an error code, or a @ref DOFInterfaceException and a list of @ref DOFValue parameters, defined in the Interface used for the request.  **/
    DOFRESULTTYPE_VALUE_STREAM =    7,  /**< @internal Identifies a @ref DOFResult as containing a stream containing a value. @see DOFResultType **/
    DOFRESULTTYPE_VALUELIST_STREAM =8   /**< @internal Identifies a @ref DOFResult as containing a stream containing a value. @see DOFResultType **/
} DOFResultType;


/* increments ref count of channel and address! */
extern OALChannel      DOFResult_GetProviderChannel     ( const DOFResult result );
extern DOFResult       DOFResult_Value_Create           (       OALChannel channel, DOFAddress address, DOFError status,
                                                                const DOFObjectID providerID, const DOFType type,
                                                                const DOFValue value );
extern DOFResult       DOFResult_ValueList_Create       (       OALChannel channel, DOFAddress address, DOFError status,
                                                                const DOFObjectID providerID, uint16 valueCount,
                                                                const DOFType typeList[], const DOFValue valuelist[] );
extern DOFResult       DOFResult_NoValue_Create         (       OALChannel channel, DOFAddress address, DOFError status,
                                                                const DOFObjectID providerID );
extern DOFResult       DOFResult_Exception_Create       (       OALChannel channel, DOFAddress address,
                                                                DOFException except, const DOFObjectID providerID );
extern DOFResult       DOFResult_Stream_Create          (       OALChannel channel, DOFAddress address, DOFError status,
                                                                const DOFObjectID providerID, OALStream stream,
                                                                DOFResultType type );
extern DOFResult       DOFResult_ExceptionStream_Create (       OALChannel channel, DOFAddress address,
                                                                const DOFObjectID providerID, OALStream stream,
                                                                DOFException exception );
extern DOFResult       DOFResult_Copy                   (       DOFResult srcRes );
extern DOFResult       DOFResult_DeepCopy               ( const DOFResult self );
extern DOFResult       DOFResult_DefineComplete_Create  (       OALChannel channel, DOFInterface iface,
                                                                const DOFObjectID providerID );
extern boolean         DOFResult_IsInternal             ( const DOFResult result );
extern boolean         DOFResult_IsSkippedSequence      ( const DOFResult result );
extern boolean         DOFResult_IsNewSequence          ( const DOFResult result );
extern boolean         DOFResult_IsLocal                ( const DOFResult result );
extern OALStream       DOFResult_GetStream              ( const DOFResult result );
extern void            DOFResult_SetPermission          (       DOFResult result, DOFPermission permission );
extern DOFPermission   DOFResult_GetPermission          ( const DOFResult result );
extern DOFObjectID     DOFResult_GetProviderID          ( const DOFResult result );
extern DOFType         DOFResult_GetValueType           ( const DOFResult result );
extern const DOFType * DOFResult_GetValueListTypes      ( const DOFResult result );
extern DOFAddress      DOFResult_GetAddress             ( const DOFResult result );
extern uint16          DOFResult_GetSequenceNumber      ( const DOFResult result );
extern void            DOFResult_SetSequenceNumber      (       DOFResult result, uint16 sequenceNumber,
                                                                boolean skipped_sequence, boolean new_sequence );
extern void            DOFResult_SetStream              (       DOFResult result, OALStream stream );
extern void            DOFResult_SetLocal               (       DOFResult result,  boolean local );
extern boolean         DOFResult_TransformStreamResult  (       DOFResult result, DOFOperation respOp );


/** @internal
    Get the Exception held by the Result.

    @note The return value is owned by the Result, and must not be destroyed individually. If it is needed beyond the
          scope of the Result, it must be copied (and later freed) by the application.

    @param  result The Result. This can be NULL.

    @return The Exception.

    @exception NULL is returned if the Result is NULL or does not contain an Exception.
**/
extern DOFException DOFResult_GetException( const DOFResult result );


/**
@}
**/


#ifdef  __cplusplus
}
#endif


#endif /* DOF_RESULTINTERNAL_H */
