/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef __router_internal_H
#define __router_internal_H

#include <dof/core.h>
#include <internal/channel.h>
#include <dof/interfaceid.h>
#include <dof/item.h>

#ifdef  __cplusplus
extern "C" {
#endif

typedef enum DOFRouterLocation_t
{
    DOFROUTERLOCATION_LOCAL = 1,
    DOFROUTERLOCATION_REMOTE,
    DOFROUTERLOCATION_BOTH
} DOFRouterLocation_t;

typedef enum DOFRouterStatusCode_t
{
    DOFROUTERSTATUS_SUCCESS = 0,
    DOFROUTERSTATUS_ERROR,
    DOFROUTERSTATUS_INVALID_BINDING,
    DOFROUTERSTATUS_INVALID_ALIAS,
    DOFROUTERSTATUS_INVALID_OPERATION_TYPE,
    DOFROUTERSTATUS_DUPLICATE,
    DOFROUTERSTATUS_RETRY_NOT_ALLOWED_FOR_THIS_OPERATION,
    DOFROUTERSTATUS_NO_REQUEST_CONFIG,
    DOFROUTERSTATUS_OPERATION_DOES_NOT_EXIST /* used for the rename function */
} DOFRouterStatusCode_t;

extern boolean DOFRouter_Init(void); /* call to set things up */
extern void DOFRouter_Cleanup(void); /* call when system shuts down */

extern void DOFRouter_Lock(void);
extern void DOFRouter_Unlock(void);

extern DOFRouterStatusCode_t DOFRouter_RouteToProvider(DOFOperation op);
extern DOFRouterStatusCode_t DOFRouter_RouteToRequestor(DOFOperation op, DOFResult result);

typedef struct interested_requester_t
{
    OALChannel          channel;
    DOFAddress          address;
    DOFSecurityScope    scope;
    OALAVLTree          opTree;
}interested_requester_t;

typedef struct provider_binding_table_entry_t
{
    OALAlias        alias;              /* an alias to the binding */
    DOFAddress      address;
    OALChannel      provider_channel;
    DOFObject       provider_object;    /* a begin provide for local, and 0 for remote */
    DOFInterfaceID  interface_id;
    DOFObjectID     provider_object_id; /* can't count on this being in object, because it may be 0 */
    DOFOperation    provider_op;
    OALObserver     obs;
    OALAVLTree      interested;
    uint32          ref_count; /* no need to mutex protect this, because the ref counts happen within the router mutex */
    uint32          timestamp;
    union
    {
        struct
        {
            unsigned int primary : 1; /* if 1, then consider this the primary route */
        } bits;
    } flags;
} provider_binding_table_entry_t;

extern provider_binding_table_entry_t *DOFRouter_Alias_to_pbte(DOFSecurityScope securityScope, DOFAddress address, OALAlias alias);

extern DOFOperation DOFRouter_LookupSession(DOFObjectID objectId);

/* Retrieve and increment the reference count for the primary pbte for the given binding in the given security scope. */
extern provider_binding_table_entry_t *DOFRouter_GetBindingEntry(DOFSecurityScope securityScope, DOFInterfaceID iid, DOFObjectID oid);
extern provider_binding_table_entry_t *DOFRouter_GetBindingEntry_Internal(DOFSecurityScope securityScope, DOFInterfaceID iid, DOFObjectID oid);
extern OALAlias DOFRouter_GetBindingAlias(DOFOperation op);

extern void provider_binding_table_entry_retain(provider_binding_table_entry_t *pbte);
extern void provider_binding_table_entry_release(provider_binding_table_entry_t *pbte);

extern void DOFRouter_ChannelUp(OALChannel channel);

typedef struct DOFRouter_Traverse_arg_t
{
    DOFOperation op;
    DOFObjectID oid;
    DOFInterfaceID iid;
    DOFQuery  filter;
    DOFSystem system;
    DOFSecurityScope securityScope;
    void *context;
    uint8 binding_added; /* 1 indicates binding added, and 0 indicates it is removed */
} DOFRouter_Traverse_arg_t;

typedef void (*DOFRouter_Traverse_fn_t)(DOFRouter_Traverse_arg_t *arg);
extern void DOFRouter_TraverseAllObjects(DOFRouter_Traverse_fn_t callback, DOFRouter_Traverse_arg_t *arg, DOFRouterLocation_t where);
extern void DOFRouter_TraverseAllInterfaces(DOFRouter_Traverse_fn_t callback, DOFRouter_Traverse_arg_t *arg);

extern boolean DOFRouter_CheckInterfaces(DOFObjectID oid, DOFSecurityScope securityScope, DOFInterfaceID *interfaces, uint32 count);

extern boolean DOFRouter_IsEventRegistered(DOFObjectID oid, DOFInterfaceEvent evt, DOFSecurityScope securityScope);
extern boolean DOFRouter_SignalEvent(DOFObjectID oid, DOFInterfaceEvent evt, DOFSecurityScope securityScope, DOFResult result );
extern boolean DOFRouter_PropertyChanged(DOFObjectID oid, DOFInterfaceProperty prp, DOFSecurityScope securityScope);
extern void DOFRouter_RemoveProviderSubscriptions(provider_binding_table_entry_t *pbte);

extern boolean DOFRouter_CheckInterest(DOFSecurityScope securityScope, DOFObjectID oid, DOFInterfaceID iid, DOFInterestLevel type);
extern boolean DOFRouter_EstablishPrimaryProvider( provider_binding_table_entry_t * pbte, boolean notify, boolean preferred );


extern void subscription_handle_event_ack(DOFOperation op, uint16 sequence);

extern void router_notify_reparent(DOFOperation op, OALChannel old_channel, DOFAddress old_address);

extern void interest_notify_new_provider(DOFOperation op);

extern DOFRouterStatusCode_t provider_route_normal_intercept_request_config(DOFOperation op, OALChannel *rc);

extern boolean _get_all_providing_connections(DOFConnectionList connectionList, const DOFObjectID objectID, const DOFInterfaceID interfaceID);

extern uint32 OALRouter_GetProviderCount(void);

extern boolean OALRouter_AddInterested_Operation(DOFOperation prov_op, DOFOperation interest_op);
void OALRouter_RemoveInterested_Operation(DOFOperation prov_op, DOFOperation interest_op);

extern void OALRouter_SendInterestedCancels( provider_binding_table_entry_t *pbte );

extern void OALRouter_SendInterestedUpdate( provider_binding_table_entry_t *pbte );

#ifdef  __cplusplus
}
#endif

#endif
