/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef INTERNAL_SECURITYSCOPE_H
#define INTERNAL_SECURITYSCOPE_H

#include <dof/core.h>
#include <dof/securityscope.h>

#include <internal/permissionscope.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @addtogroup DOFSecurityScope
@{
**/


#define DOFSECURITYSCOPEFLAG_ALL_SCOPES  0x3FFFFFFF  /** Security scope ID used as a wildcard to indicate @b all applicable scopes. **/
#define DOFSECURITYSCOPEFLAG_NOT_MASKING 0x3FFFFFFE  /** Security scope ID used as an indicator that the related permission does not mask any other, more specific permissions with a larger (additional) scope. **/


/*
**  "Public" Type and Method Declarations
*/

/**
 * Creates a new DOFSecurityScope.  This is needed by implementations of
 * DOFAuthenticatorStorage, where the scope for permissions are returned to the
 * library for each authenticating identity.
 *
 * @mustfree{DOFSecurityScope_Destroy}
 *
 * @param domainID The DOFObjectIDDomain of the domain in which the scope exists.  This should not be NULL.
 * @param idCount The count of Permission Group ID elements in @a ids.
 * @param ids The list of Permission Group IDs.  This may be NULL if @a idCount is 0.  Otherwise, the list must contain at least @a idCount elements.
 * @param flagCount The count of flag elements in @a flags.
 * @param flags The list of flags.  This may be NULL if @a flagCount is 0.  Otherwise, the list must contain at least @a flagCount elements.
 *
 * @return The newly created DOFSecurityScope or NULL in the case of an error.
 */
extern DOFSecurityScope DOFSecurityScope_Create( DOFObjectIDDomain domainID, uint16 idCount, const uint32 * ids, uint16 flagCount, const uint32 * flags );

/**
 * Create a copy of an existing security scope.
 * @mustfree{DOFSecurityScope_Destroy}
 * @param self The security scope.  This should not be NULL.
 * @return A new copy of the security scope.
 * @exception NULL is returned if the new security scope cannot be created or an invalid parameter was detected.
 */
extern DOFSecurityScope DOFSecurityScope_Copy( const DOFSecurityScope self );


/**
 * Return the domain ID associated with the security scope. This is always a @ref DOFObjectIDDomain.
 *
 * @param self The security scope.  This must not be NULL.
 *
 * @return The DOFObjectIDDomain that represents the domain of the security scope.  NULL indicates the non-secure
 *         domain.
 */
extern DOFObjectIDDomain DOFSecurityScope_GetDomainID( const DOFSecurityScope self );


/**
 * Return whether or not two security scopes are compatible.
 * @param self The security scope.  This Must not be NULL.
 * @param scope The security scope to be compared.  This must not be NULL.
 */
extern boolean DOFSecurityScope_IsCompatible( const DOFSecurityScope self, const DOFSecurityScope scope );

/**
 * Compares SecurityScopes  
 *  
 * @param self scope to compare against;
 * @param scope scope to compare;
 *  
 * @note NULLs will compare equal. 
 *  
 * @return uint32 0 for equal.
 *                > 0 for greater than.
 *                < 0 for less than.
 */
extern int32 DOFSecurityScope_Compare( const DOFSecurityScope self, const DOFSecurityScope scope );

/**
 * Returns the count of Permission Group IDs in the DOFSecurityScope.
 *
 * @param self The PermissionScope from which to retrieve the ID count.  This must not be NULL.
 *
 * @return The count of IDs.
 */
extern uint32 DOFSecurityScope_GetIDsCount( DOFSecurityScope self );


/**
 * Returns a pointer to the array of Permission Group IDs in the DOFSecurityScope.
 *
 * @param self The PermissionScope from which to retrieve the IDs.  This must not be NULL.
 *
 * @return A pointer to an array of IDs, NULL in the case of an error.
 */
extern const uint32 * DOFSecurityScope_GetIDs( DOFSecurityScope self );


/**
 * Returns the count of flags in the DOFSecurityScope.
 *
 * @param self The PermissionScope from which to retrieve the flag count.  This must not be NULL.
 *
 * @return The count of flags.
 */
extern uint32 DOFSecurityScope_GetFlagsCount( DOFSecurityScope self );


/**
 * Returns a pointer to the array of flags in the DOFSecurityScope.
 *
 * @param self The PermissionScope from which to retrieve the flags.  This must not be NULL.
 *
 * @return A pointer to an array of flags, NULL in the case of an error.
 */
extern const uint32 * DOFSecurityScope_GetFlags( DOFSecurityScope self );

/* DOFAuthenticator * DOFSecurityScope_GetDOFAuthenticatorPointerToPointer( const DOFSecurityScope self ); */

extern DOFSecurityScope DOFSecurityScope_Unmarshal( OALStream         stream,
                                                    DOFObjectIDDomain domainID );

extern boolean DOFSecurityScope_Marshal(DOFSecurityScope securityScope,
                                        OALStream        stream);

extern boolean DOFSecurityScope_Intersect( DOFSecurityScope   self,
                                           OALPermissionScope permsScope );

extern uint32 DOFSecurityScope_GetMarshalSize( DOFSecurityScope self );

extern int32 DOFSecurityScope_FindBestIndex( DOFSecurityScope source,
                                             DOFSecurityScope dest );

extern OALPermissionScope DOFSecurityScope_GetOALPermissionScope( DOFSecurityScope scope );

/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* _INTERNAL_SECURITYSCOPE_H */
