/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef _DOF_SERVERINTERNAL_H
#define _DOF_SERVERINTERNAL_H

#include <dof/core.h>
#include <internal/slist.h>
#include <internal/avl_tree.h>
#include <internal/oaltypesinternal.h>
#include <internal/security_membrane.h>
#include <internal/securityscope.h>
#include <internal/credentialsinternal.h>
#include <internal/enqueue.h>

#define OALSERVER_PRINT_REF_COUNT_DEBUG 0

#ifdef  __cplusplus
extern "C" {
#endif

typedef enum DOFServerStatus_t
{
    DOFSERVERSTATUS_STOPPED = 0, /* The server is not currently established. **/
    DOFSERVERSTATUS_STARTED = 1  /* The server is currently established. **/
} DOFServerStatus;

typedef struct DOFServerCallbackNodeContext_t
{
    OALEnqueueItem_t super;
    struct DOFServerCallbackNode_t * item;
} DOFServerCallbackNodeContext_t;

typedef struct DOFServerCallbackNode_t
{
    OALSListNode                    next;
    DOFServerStateListener          callback;
    DOFServer                       server;
    DOFServerCallbackNodeContext_t* destroyListener;
} DOFServerCallbackNode_t;

typedef struct DOFServerConfig_t
{
    DOFAddress              listenAddr;
    DTLServerConfig         serverConfig;
    OALCredentialsTree      creds; /* Tree of credentials keyed on domainID */
    OALDomainCredentialsTree dcreds; /* Tree of domain credentials keyed on domainID */
    DOFSecurityDesire       securityDesire;
    DOFServerType           type;
    DOFProtocolNegotiator   negotiator;
    DOFOperationFilter      receiveFilter;
    DOFOperationFilter      sendFilter;
    DOFOperationBridgeConfig      bridge;
    uint32                  refCount;
    uint32                  MaxReceiveSilence;
    uint32                  MaxSendSilence;
    uint32                  errors;
    uint32                  MaxConnections;
}DOFServerConfig_t;

typedef struct DOFServerReleaseContext_t
{
    OALEnqueueItem_t super;
    union
    {
        DOFServer dofserver;
        OALServer oalserver;
    } organizer;
} DOFServerReleaseContext_t;

typedef struct OALServer_t
{
    DOFServerManager            serverManager;
    DOFSecurityMembrane         securityMembrane;
    DOFServerConfig             config;
    OALSListNode                serverContexts;
    OALSListNode                stateCallback;
    DOFServer                   thisDOFServer;
    DOFServer                   relatedServer;
    DOFServerStateListener      relSrvList;             /* listener added to related server */
    DOFServerStatus             serverState;
    DOFServerReleaseContext_t*  cleanupnode;

    union
    {
        OALChannel          channel;                /* Default associated channel. this channel owns the DTL. Datagram case **/
        DTL                 dtl;                    /* Server owns DTL in Stream server case */
    }transport;

    uint64                  byteInputCount;
    uint64                  packetInputCount;
    uint64                  byteOutputCount;
    uint64                  packetOutputCount;
    uint32                  refCount;
    uint32                  StartCount;             /* how many times has this server been started */
    uint16                  connectionsCount;
}OALServer_t;

typedef struct DOFServer_t
{
    uint32                      refCount;
    uint32                      StartCount; /* how many times has this server been started */
    OALServer                   server;
}DOFServer_t;

typedef struct internalDOFServerStateListener_t
{
    DOFServerStateListener_t    base;
    DOFServer                   server;
} internalDOFServerStateListener_t;

typedef struct internalConnectionListener_t
{
    DOFConnectionStateListener_t base;
    OALServer                    server;
} internalConnectionListener_t;

extern DOFServerManager DOFServerManager_Create( DOF dof );
extern void DOFServerManager_Destroy( DOFServerManager self );

extern DOFServer DOFServer_Create( DOF dof, DOFServerConfig config, DOFSecurityMembrane securityConnector, DOFServer relatedServer );
extern boolean DOFServerConfig_IsEqual(DOFServerConfig self, DOFServerConfig other);
extern DOFServer DOFServer_Retain( DOFServer srv );
extern void DOFServer_Release( DOFServer srv );
extern DOFAddress DOFServer_GetListenAddr( DOFServer server );
extern void DOFServer_Kill( DOFServer server, DOFException except );
extern void OALServer_NotifyServerStarted( OALServer srv );
extern void OALServer_NotifyServerStopped( OALServer srv, DOFException except );
extern void OALServer_NotifyServerTransportDown( OALServer srv, DOFException exception );
extern void OALServer_NotifyStateChange( OALServer self, DOFException exception );
extern void OALServer_RemoveAllListener( OALServer self );

extern void OALServer_UpdateInputByteCount( OALServer server, uint32 count );
extern void OALServer_UpdateOutputByteCount( OALServer server, uint32 count );

extern DOFServerState DOFServerState_Copy(DOFServerState self);

extern uint32 DOFServerManager_GetServerCount( DOF dof );
extern DOFServer OALServer_GetDOFServer( OALServer server );
extern DOFServer OALServer_CreateDOFServer( OALServer server );

/**
 * Create a copy of a DOFServerConfig, where modifications to the copy will not affect the
 * original in any way, and changes to the original are isolated from the copy.
 * @mustfree{DOFServerConfig_Destroy}
 * @param self The DOFServerConfig.  This should not be NULL.
 * @return The copied, and now isolated, DOFServerConfig, or NULL if it could not be created, or @a self was NULL.
 */
extern DOFServerConfig DOFServerConfig_Copy( const DOFServerConfig self );

typedef void (*DOFServerConfig_GetDomainCredential_Callback)( DOFCredentials creds, void* context );

extern void DOFServerConfig_GetDomainCredential( DOFServerConfig self, DOFObjectIDDomain domainID, uint32 timeout, DOFServerConfig_GetDomainCredential_Callback fn, void * context );

#if OALSERVER_PRINT_REF_COUNT_DEBUG
#   define OALServer_Retain( srv ) OALServer_Retain_old((srv), __func__, __FILE__, __LINE__)
    extern OALServer OALServer_Retain_old( OALServer srv, const char* function, const char* filename, int line);
#else
    extern OALServer OALServer_Retain( OALServer srv );
#endif

#if OALSERVER_PRINT_REF_COUNT_DEBUG
#   define OALServer_Destroy( srv ) OALServer_Destroy_old((srv), __func__, __FILE__, __LINE__)
    extern void OALServer_Destroy_old( OALServer srv , const char* function, const char* filename, int line);
#else
    extern void OALServer_Destroy( OALServer srv );
#endif

extern DOFAddress OALServer_GetListenAddr(OALServer server);
extern DTL OALServer_GetDTL( OALServer srv );
extern OALChannel OALServer_GetSendServerChannel( OALServer server );
extern DOFServerState OALServer_CreateState( OALServer self, DOFException exception );
extern boolean OALServer_IsDatagram( OALServer self );
extern boolean OALServer_IsStarted( OALServer self );

extern void DOFServerCount_Lock(void);
extern void DOFServerCount_Unlock(void);

extern void DOFServer_Lock(void);
extern void DOFServer_Unlock(void);

typedef boolean (*server_traverse_fn_t)(OALServer server, void *context);
extern void OALServer_TraverseAllServers(server_traverse_fn_t fn, void *context);

extern boolean OALServer_AddConnectionListener(OALServer server, DOFConnection conn);
extern boolean DOFServer_IsConnectionAllowed(DOFServer server);

extern DOFSecurityMembrane DOFServer_CreateSecurityMembrane(DOF dof, DOFServerConfig config );
extern boolean OALServer_HaveStartedStreamServer( const DOFAddress addr );

/**
 * Return the DOF from which the Server was created.
 *
 * @param self The Server.  This must not be NULL.
 *
 * @return The DOF associated with the Server.
 */
extern DOF DOFServer_GetDOF( const DOFServer self );

#ifdef  __cplusplus
}
#endif

#endif /* _DOF_SERVERINTERNAL_H */
