/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef __slist_H
#define __slist_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif

/**
a simple singly-linked list implementation. fun!
 inserts in the front for efficiency

 this is NOT THREAD SAFE but it probably doesn't matter since worst-case a
 thread misses a member during pointer swap
 **/
 
/** @internal @addtogroup OALSList
*   A singly linked list

*   This is a very fast list implementation using special structures that
*   have a pointer to their same type as the first member in the structure.
*   Items are always inserted at the front of a list, insertion O(1).
*   Removing an item from a list has O(n) complexity. It is very important
*   not to try and add one item to two lists at the same time.
*
*   Lists are not created. Any item can start a list as long as the next
*   pointer is 0. Normally it's custom to have a single pointer called head.
*
@{
@internal
**/

/* IMPORTANT: the beginning of the struct you put in slist should look like this! */
typedef struct OALSListNode_t {
    struct OALSListNode_t *next; /* pointer to the next item in the list */
} OALSListNode_t;

typedef struct OALSListNode_t *OALSListNode;

typedef boolean (*OALSListTraversal_fn_t)(OALSListNode item, void *arg);
typedef void (*OALSListRemoveAll_fn_t)(OALSListNode item, void *arg);

/** @internal
    Insert an item into an SList.
 
    @note Items inserted into an SList always appear at the front.
 
    @param head Pointer to an SList head (first item in the list).  Must not be NULL.
    @param item The item to insert into the SList.  Must not be NULL.
**/ 
extern void OALSList_Insert(OALSListNode *head, OALSListNode item);

/** @internal
    Remove an item from an SList.
 
    @param head Pointer to an SList head (first item in the list).  Must not be NULL.
    @param item The item to remove.
    
    @return boolean TRUE if item was removed FALSE otherwise
**/
extern boolean OALSList_Remove(OALSListNode *head, OALSListNode item);

/** @internal
    Remove all items in a list and call a callback function for each item removed.

    @param head Pointer to the SList to remove all items from.  Must not be NULL.
    @param fn The callback function called for each item removed.
    @param arg Context for the function @a fn.
*/
extern void OALSList_RemoveAll(OALSListNode *head, OALSListRemoveAll_fn_t fn, void *arg);

/**  @internal
     Traverse an SList and return a pointer to the node where the callback
     function returns true.

     @param head The SList to traverse.
     @param fn The function to call for each item in the SList.  Must not be NULL.
     @param arg Context for the function @a fn.

     @return The pointer to the node in the list where the callback function
             returned true.  If the entire list is traversed, and the callback
             function never returns true, NULL is returned.
*/
extern boolean OALSList_Traverse(OALSListNode head, OALSListTraversal_fn_t fn, void *arg);

/**  @internal
     Traverse a list and return a pointer to the node where the callback
     function returns true.

     @param head The list to traverse.
     @param fn The function to call for each item in the list.
     @param arg Context for the function @a fn.

     @return The pointer to the node in the list where the callback function
             returned true. If the entire list is traversed, and the callback
             function never returns true, NULL is returned.
*/
extern OALSListNode OALSList_Retrieve(OALSListNode head, OALSListTraversal_fn_t fn, void *arg);

/** @internal 
    Traveses an slist for nodes to remove. If a node is removed the function returns the node where you can continue your find and remove traversal.
    
    @param head Head of the SList. Must not be NULL.
    @param fn function to find nodes to remove
    @param arg User Context
    @param start [in]Which node to start with. if NULL traverse will start at head.
                 [OUT]Node before one removed. NULL means end of list.
   
    @return OALSListNode Node to free or NULL meaning finished no node found.
*/
extern OALSListNode OALSList_FindAndRemove(OALSListNode* head, OALSListNode* start, OALSListTraversal_fn_t fn, void* arg);

/* allows slists not at the top of a struct to be useful by converting a node address into that struct */
#define node2struct(ptr, type, element) (type*)(((char*)ptr) - offsetof(type, element))

/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif
