/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_INTERNAL_STREAM_H
#define DOF_INTERNAL_STREAM_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @internal @addtogroup OALStream
@{
**/

#define OALSTREAMTYPE_READONLY  0x01
#define OALSTREAMTYPE_READWRITE 0x02
#define OALSTREAMTYPE_WRAPPED   0x03


/*
**  "Public" Type and Method Declarations
*/
typedef struct OALStream_t *    OALStream;
typedef uint8                   OALStreamType;

/* "Static" methods */

/**
    Create a new Read/Write OALStream. Data stored in this stream will be destroyed when this stream is destroyed.

    @mustfree{OALStream_Release}

    @param  size  Specifies how many bytes to allocate memory for.

    @return The new @ref OALStream (Read/Write).
    @retval NULL if failure to create the object.

    @internal
**/
extern OALStream   OALStream_CreateRW(uint32 size);

/** 
 * @internal 
 * Creates a OALStream that takes ownership of the buffer passed in. 
 *  
 * @note Data must be alloced with PCRMem_Alloc. Data is owned by stream and will be release with the stream 
 * @note Write offset will be set to the end of the buffer.  
 * 
 * @param data Data buffer to use for stream.
 * @param size Byte length of data buffer.
 * 
 * @return OALStream Stream containing data passed in or NULL representing a failure
 */
extern OALStream   OALStream_CreateRWWithData( uint8 * data, uint32 size );

/** 
 * @internal 
 * Creates a stream that wraps a buffer. Allows use of a buffer as a OALStream
 *  
 * @note Data is not owned by OALStream and will not be freed. 
 * @note Write offset will be set to the end of the buffer. 
 *  
 * @param data Data buffer to wrap.
 * @param size Byte length of the buffer.
 * 
 * @return OALStream Stream that wraps the buffer passed in or NULL representing a failure.
 */
extern OALStream OALStream_CreateRWWrapped( uint8 * data, uint32 size );

/**
    Create a new Read-Only OALStream wrapped around existing data. Data wrapped by this stream
    will not be destroyed when this stream is destroyed.

    @mustfree{OALStream_Release}

    @param  data    Pointer to the data to be wrapped.  Should not be NULL.
    @param  length  The length of the data to be wrapped.

    @return The new @ref OALStream (Read-Only).
    @retval NULL if failure to create the object.

    @internal
**/
extern OALStream   OALStream_CreateRO(const uint8* data, uint32 length);

/**
    Create a deep copy of an existing OALStream to a new Read/Write OALStream.  The newly created
    OALStream will perform (and own) a deep copy of the data in @a source.  The  deep copy will
    include the data that has been read and the data that has not been written to, or in other words,
    the new OALStream will not be minimized.

    @mustfree{OALStream_Release}

    @param  source  The OALStream from which to make the copy. May be either Read/Write or Read-Only.
                    Must not be NULL.

    @return The new @ref OALStream (Read/Write).
    @retval NULL if failure to create the object.

    @internal
**/
extern OALStream   OALStream_CloneToRW(OALStream source);

extern OALStream   OALStream_Create_Concat(OALStream first, OALStream second, uint32 freeSpace);

/* "Object" methods */
extern void            OALStream_Retain            (OALStream self);

/**
    Release (destroy) an OALStream.

    @param self         The OALStream to release.
    @param mustLock     Must be TRUE unless ref_count_mutex is already locked.

    @internal
**/
extern void            OALStream_Release           (OALStream self,
                                                    boolean   mustLock);

/**
    Return the type of this OALStream.

    @param  self   "this" OALStream.  Must not be NULL.

    @return The type of OALStream that this is.
    @retval OALSTREAMTYPE_READONLY      if this OALStream is Read-Only.
    @retval OALSTREAMTYPE_READWRITE     if this OALStream is Read/Write.

    @internal
**/
extern OALStreamType   OALStream_GetType           (OALStream self);

/**
    Returns the size of the unread data in this stream.

    @param  self   "this" OALStream.  Must not be NULL.

    @return The unread data size (in bytes).

    @internal
**/
extern uint32          OALStream_GetDataSize       (OALStream self);

/**
    Returns the remaining free space in a writable OALStream.

    @param  self   "this" OALStream.  Must not be NULL.

    @return The remaining free space (in bytes).
    @retval 0 any Read-Only OALStream.

    @internal
**/
extern uint32          OALStream_GetFreeSize       (OALStream self);

/**
    Increases the memory storage of a writable stream.

    @param  self            "this" OALStream.  Must not be NULL.
    @param  sizeIncrease    the number of bytes to add as new free space.

    @return TRUE if successful, FALSE otherwise.

    @internal
**/
extern boolean         OALStream_IncreaseFreeSize  (OALStream self, uint32 sizeIncrease);

/**
    Returns the index representing the beginning of free space in an OALStream.

    @param  self   "this" OALStream.  Must not be NULL.

    @return The offset, in bytes, to the beginning of the free space

    @internal
**/
extern uint32          OALStream_GetWriteOffset    (OALStream self);

/**
    Returns the current offset of unread data in an OALStream.

    @param  self   "this" OALStream.  Must not be NULL

    @return The offset, in bytes, of the beginning of unread data.

    @internal
**/
extern uint32          OALStream_GetReadOffset     (OALStream self);

/**
    Sets the current offset of unwritten data in an OALStream, relative
    to the beginning of the OALStream data.  If the read offset is past
    the new write offset, the read offset is adjusted to be at the
    new write offset.

    @param  self    "this" OALStream.  Must not be NULL.
    @param  offset  the new offset, in bytes, of the beginning of unwritten data.

    @internal
**/
extern void            OALStream_SetWriteOffset    (OALStream self, uint32 offset);

/**
    Sets the current offset of unread data in an OALStream, relative
    to the beginning of the OALStream data.

    @param  self    "this" OALStream.  Must not be NULL.
    @param  offset  the new offset, in bytes, of the beginning of unread data.

    @internal
**/
extern void            OALStream_SetReadOffset     (OALStream self, uint32 offset);

/**
    Adjusts the current offset of unread data in an OALStream, relative
    to the current read offset.

    @param  self         "this" OALStream.  Must not be NULL.
    @param  offsetChange the adjustment, in bytes, from the current read offset.

    @internal
**/
extern void            OALStream_AdjustReadOffset  (OALStream self, int32 offsetChange);

/**
    Adjusts the current offset of unwritten data in an OALStream, relative
    to the current write offset.

    @param  self         "this" OALStream.  Must not be NULL.
    @param  offsetChange the adjustment, in bytes, from the current write offset.

    @internal
**/
extern void            OALStream_AdjustWriteOffset (OALStream self, int32 offsetChange);

/**
    Returns a pointer to the beginning of the unread data in an OALStream.

    @param  self    "this" OALStream.  Must not be NULL

    @return A pointer to the beginning of unread data.

    @internal
**/
extern const uint8*    OALStream_GetReadPtr        (OALStream self);

/**
    Returns a pointer to the beginning of the free space in an OALStream.

    @param  self    "this" OALStream.  Must not be NULL

    @return A pointer to the beginning of free space.
    @retval NULL If this stream is Read-Only

    @internal
**/
extern uint8*          OALStream_GetWritePtr       (OALStream self);

/**
    Reads data from an OALStream and adjust the read offset/pointer.

    @param  self        "this" OALStream.  Must not be NULL
    @param  destBuf     the buffer in which to place the read data.  Must not be NULL
    @param  size        the amount of data to read, in bytes.

    @retval TRUE    if successful.
    @retval FALSE   if the operation failed.

    @internal
**/
extern boolean         OALStream_Read              (OALStream self, uint32 size,       uint8* destBuf);

/**
    Reads data from one OALStream into another OALStream.

    @param self  "this" OALStream.  Must not be NULL.
    @param size  the amount of data to read, in bytes.
    @param dest  the stream in which to place the read data.  Must not be NULL.

    @retval TRUE    @a size was read from @a self and written to @a dest
    @retval FALSE   some error occurred

    @internal
**/
extern boolean         OALStream_ReadToStream      (OALStream self, uint32 size, OALStream dest);

/**
    Writes data to a Read/Write OALStream.  Data is written at the begginning of the free space.

    @param  self        "this" OALStream.  Must not be NULL.
    @param  srcBuf      the buffer from which to read in the data.  Must not be NULL.
    @param  size        the amount of data to written, in bytes.

    @retval TRUE    if successful.
    @retval FALSE   if the operation failed.

    @internal
**/
extern boolean         OALStream_Write             (OALStream self, uint32 size, const uint8* srcBuf);


extern boolean         OALStream_Write8            (OALStream self, uint8 byte);
extern boolean         OALStream_Write16           (OALStream self, uint16 word);
extern boolean         OALStream_Write32           (OALStream self, uint32 dword);
extern boolean         OALStream_Write64           (OALStream self, uint64 ddword);

extern boolean         OALStream_Read8             (OALStream self, uint8* pByte);
extern boolean         OALStream_Read16            (OALStream self, uint16* pWord);
extern boolean         OALStream_Read32            (OALStream self, uint32* pDWord);
extern boolean         OALStream_Read64            (OALStream self, uint64* pDDWord);

/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* _DOF_STREAM_H */
