/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_SYSTEMINTERNAL_H
#define DOF_SYSTEMINTERNAL_H

#include <dof/core.h>
#include <dof/interfaceid.h>
#include <internal/channel.h>
#include <internal/enqueue.h>
#include <internal/interfaceinternal.h>
#include <internal/observable.h>
#include <internal/router.h>
#include <internal/slist.h>
#include <internal/staticassertinternal.h>

#ifdef  __cplusplus
extern "C" {
#endif

/** @addtogroup DOFSystem
@{
**/


#define QUERY_CALLBACK_INTERFACE_REMOVED    0
#define QUERY_CALLBACK_INTERFACE_ADDED      1
#define QUERY_CALLBACK_OBJECT_REMOVED       2
#define QUERY_CALLBACK_COMPLETE             3


typedef struct ExtendAcceptRequestContext_t
{
    DOFSystem sys;

    DOFOperation reqOp;

    union
    {
        DOFOperation provOp;
        DOFResult result;
    } arg2;

    boolean isOp;
} * ExtendAcceptRequestContext;

typedef enum validationStatus
{
    VALIDATED,
    VALIDATION_PENDING,
    INVALIDATED,
    VALIDATION_ERROR
} validationStatus;


typedef struct SystemContext_t
{
    OALEnqueueItem_t super;
    DOFSystem system;
    DOFOperation reqOp;
    union
    {
        DOFOperation provOp;
        DOFResult result;
    } organizer;
} SystemContext_t;


/*
    This structure is used to queue up changes for interfaces that are added or removed, or when an object
    is no longer providing.
*/
typedef struct queueDescriptor_t
{
    DOFOperation op;
    union u_t
    {
        DOFResult    result;
        DOFOperation provOp;
    }u;
} queueDescriptor_t;

typedef struct systemThread_t
{
    PCRThread         thread;
    uint16            tail;                             /* Offset to most recently added entry of queue. */
    uint16            head;                             /* Offset to oldest entry. */
    PCRCondition      cond;
    boolean           shutdown;                         /* Init to FALSE. To request thread shutdown, set to TRUE. */
    boolean           holdForExistingAcceptCompletion;  /* Init to FALSE. */
    DOFSystem         sys; /* for mutex */
} systemThread_t;


/*
    This structure is used to create a list of objects that have been added to the system
    Each time an object is added to the system a node will be added to the list so that when the
    system is shut down the objects can be destroyed.
*/
typedef struct sysObjectNode_t
{
    struct sysObjectNode_t  *next;
    DOFObjectID             objID;
    DOFObject               obj;
} sysObjectNode_t;


typedef struct sysObjectNode_t *    sysObjectNode;

typedef struct DOFSystemDestroyThreadpoolContext_t
{
    OALEnqueueItem_t super;
    OALThreadPool pool;
} DOFSystemDestroyThreadpoolContext_t;

typedef struct DOFSystemState_t
{
    DOFCredentials credentials;
    DOFSystemConfig config;
    DOFSecurityMembrane systemStateMembrane;
    uint16 poolSize;
} DOFSystemState_t;

/** @internal
    Describes a DOFSystem

    @note By the special arrangement that channel is the first parameter, these references deliver the same value:
        DOFSystem                    (ptr to DOFSystem_t struct)
        OALChannel                   (ptr to OALChannel_t struct that has the same address as its parent DOFSystem_t)
        &sys->channel                (ptr to sys's first member)
        DOFSystem_GetChannel( sys )  (ptr to sys's member known as channel)
**/
typedef struct DOFSystem_t
{
    /*channel MUST be 1st parameter!!!!! Never change it to be otherwise!!!!! This is checked by STATIC_ASSERT() below*/
    OALChannel_t                channel;
    DOFSystemState_t            state;
    DOF                         dof;
    DOFObjectID                 identity;
    sysObjectNode               objectList;
    OALThreadPool               threadPool;
    OALEnqueue                  requestQueue;
    OALEnqueue                  responseQueue;
    OALEnqueue                  query_queue;
    DOFEnqueueContext_t       * requestItem;
    DOFEnqueueContext_t       * responseItem;
    DOFEnqueueContext_t       * queryItem;
    DOFSystemDestroyThreadpoolContext_t* destroyPoolContext;
    OALAVLTree                  interestListeners; /* list of ActivateInterestListenerNode_t* */
} DOFSystem_t;
/* STATIC_ASSERT( offsetof( DOFSystem_t, channel ) == 0 ); */     /* To avoid a warning this static assert was added to DOFSystem_Create. */

typedef struct SessionIDTreeContext_t
{
    DOFObjectID sessionID;
    OALChannel  requestorChannel;
    DOFAddress  requestorAddress;
} SessionIDTreeContext_t;

extern OALTimer DOFSystem_GetTimer( void );

extern boolean DOFSystem_Init( DOF dof );
extern void DOFSystem_Shutdown( DOF dof );
extern void DOFSystem_Cleanup( DOFSystem sys );

extern DOFSystem DOFSystem_Create(DOF                 dof,
                                  DOFSystemConfig     config,
                                  DOFSecurityMembrane membrane);

extern void DOFSystem_Dtor( OALChannel chan );

extern void       Requestor_ProcessResponse( DOFOperation respOp, DOFResult respResult );
extern DOFResult  Provider_ProcessRequest  ( DOFOperation reqOp, DOFOperation provOp, boolean reqStored );

/** @internal
    Accept a change response from DOFRouter and queue it for delivery to the requestor callback

    @note This call does not and shall not block.

    @param sys              The system the query is for.
    @param op               The query operation.
    @param oid              The DOFObjectID
    @param iid              The interface id for the binding that is receiving a change notice
    @param binding_added    1 if added, 0 if removed

**/
extern void DOFSystem_AcceptQueryResponse  ( DOFSystem sys, DOFOperation op, DOFObjectID oid,
                                      DOFInterfaceID iid, uint32 binding_added );
extern void DOFSystem_QueryCleanup(void);
extern void DOFSystem_CompleteGetRequest    ( DOFSystem sys, DOFRequest request, DOFValue value );
extern void DOFSystem_CompleteSetRequest    ( DOFSystem sys, DOFRequest request );
extern void DOFSystem_CompleteInvokeRequest ( DOFSystem sys, DOFRequest request, uint16 valuesCount,
                                              const DOFValue values[] );
extern void DOFSystem_CompleteSessionRequest( DOFSystem sys, DOFRequest request );
extern void DOFSystem_CompleteSubscribeRequest( DOFSystem sys, DOFRequest request );
extern void DOFSystem_CompleteRegisterRequest ( DOFSystem sys, DOFRequest request );


extern void    DOFSystem_AcceptRequest_Wrapper( OALChannel chan, DOFOperation reqOp, const DOFOperation provOp,
                                                DOFAddress addr );
extern boolean DOFSystem_AcceptRequest      ( OALChannel chan, DOFOperation reqOp, const DOFOperation provOp );
extern void    DOFSystem_AcceptResponse_Wrapper( OALChannel chan, DOFOperation op, DOFResult respResult );
extern boolean DOFSystem_AcceptResponse     ( OALChannel chan, DOFOperation op, DOFResult respResult );
extern void    DOFSystem_OpEarlyCancelCallback(                DOFOperation op );
extern void    DOFSystem_OpTimedOutCallback(                   DOFOperation op );

/** @internal
    This function is called each time a binding becomes available, or when it is no longer available.

    @param securityScope    The securityScope.
    @param oid              The object ID for the binding that changed.
    @param iid              The interface ID for the binding that changed.
    @param binding_added    1 if added, 0 if removed.
**/
extern void DOFSystem_BindingChanged       ( DOFSecurityScope securityScope, DOFObjectID oid, DOFInterfaceID iid,
                                             uint32 binding_added);

extern void query_response_fn(OALThreadPoolItem node, boolean shutdown); /* made non-static for a unit test */

extern void DOFSystem_ReqOpClosingObserver( OALObserver self, OALObservable o, void * reqOp );

extern DOFAddress DOFSystemAddress_Create( DOFSystem sys );
extern DOFSystem DOFSystemAddress_GetSystem( const DOFAddress self );

extern void DOFSystem_DeliverResponseBackToRequestor( DOFOperation provSideReqOp, DOFResult result );

extern void DOFSystem_CheckPermissionsAndRouteResponse(DOFSystem    system,
                                                       DOFOperation operation,
                                                       DOFResult    result);

extern validationStatus DOFSystem_ValidateOperationInboundPermissions( DOFSystem sys, DOFSecurityMembrane securityMembrane,
                                                                       DOFPermission explicitPerm, DOFPermission logicalPerm,
                                                                       DOFOperation reqOp, DOFOperation provOp, DOFResult result,
                                                                       boolean provOpNotResult );

/* DO NOT HOLD ANY LOCKS WHEN THIS IS CALLED!! AND DO NOT CALL THIS IF op is or will be passed through DOFRouter.*/
extern void DOFSystem_CallCompleteWithException( DOFOperation op, DOFException except );

extern void DOFSystem_CheckPermissionsRetryAndRoute(DOFOperation op);

/* Special defines to increase readability. A system is a channel, of course. */
#define DOFSystem_GetChannel( self ) ((OALChannel)(&self->channel))
#define DOFSystem_Retain( self )     ((DOFSystem)OALChannel_Retain( &self->channel ))

/* This should be calle internally never use DOFSystem_Destroy */
#define DOFSystem_Release( self )    (OALChannel_Release( &self->channel ))


extern int DOFSystem_Handle_AckTimeout( OALTimerItem item, uint32 argc, void* argv[], boolean shutdown );

/**
 * Return the DOF from which the System was created.
 *
 * @param self The System.  This must not be NULL.
 *
 * @return The DOF associated with the System.
 */
extern DOF DOFSystem_GetDOF( const DOFSystem self );

extern DOFSystemConfig DOFSystemConfig_Copy(DOFSystemConfig self);

/**
@}
**/


#ifdef  __cplusplus
}
#endif

#endif /* DOF_SYSTEMINTERNAL_H */
