/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_INTERNAL_TEP_V1_H
#define DOF_INTERNAL_TEP_V1_H

#include <dof/address.h>
#include <dof/permissionset.h>

#include <internal/blackboarddata.h>
#include <internal/channel.h>
#include <internal/packet.h>
#include <internal/security_common.h>
#include <internal/securitymode.h>
#include <internal/stream.h>

#define TEP_V1_APPID    128

#define TEP_V1_TIMEOUT          30000       /* 30 seconds (per specification EPS.23) */

/* TEP Error Codes */
#define TEP_V1_ERROR_PARSEERROR                 0x81
#define TEP_V1_ERROR_UNACCEPTABLEMODE           0x82
#define TEP_V1_ERROR_NORESPONSE                 0x83
#define TEP_V1_ERROR_UNKNOWNDOMAIN              0x84
#define TEP_V1_ERROR_AUTHENTICATAION_FAILED     0x85
#define TEP_V1_ERROR_INSUFFICIENT_RESOURCES     0x86

#ifdef  __cplusplus
extern "C" {
#endif

typedef enum TEP_SessionEvent_t
{
    TEP_V1_SessionEvent_NEWSESSION_SUCCESS_RECVKEY,
    TEP_V1_SessionEvent_NEWSESSION_SUCCESS_SENDKEY,
    TEP_V1_SessionEvent_NEWSESSION_FAILURE,
    TEP_V1_SessionEvent_NEWSESSION_FINALIZE,

    TEP_V1_SessionEvent_REKEY_RECVKEY,
    TEP_V1_SessionEvent_REKEY_SENDKEY,
    TEP_V1_SessionEvent_REKEY_SENDKEYANDDISABLE,
    TEP_V1_SessionEvent_REKEY_DISABLEOLD,
    TEP_V1_SessionEvent_REKEY_FAILURE,

    TEP_V1_SessionEvent_ENDING
} TEP_SessionEvent_t;

typedef void* (*TEP_V1_SessionEventCallback)(void               * context,
                                             uint8              * key,
                                             TEP_SessionEvent_t   eventType,
                                             DOFSecurityMode      securityMode,
                                             OALChannel           channel,
                                             DOFSecurityScope     secScope,
                                             DOFPermissionSet     localPerms,
                                             DOFPermissionSet     otherPerms,
                                             DOFException         exception,
                                             uint32               stateID,
                                             DOFAddress           remoteAddr,
                                             OALStream          * initialState,  /* Only used on the provider side to get initial cipher state */
                                             DOFCredentials       creds);


typedef struct TEP_V1_t
{
    BlackboardData                blackboardData;

    TEP_V1_SessionEventCallback   sessionEventCallback;

    boolean                       isConnectionLess;
} TEP_V1_t;

typedef struct TEP_V1_t * TEP_V1;

extern boolean TEP_V1_Init(void);

extern void TEP_V1_Shutdown1(void);

extern void TEP_V1_Shutdown2(void);

extern TEP_V1 TEP_V1_Create(BlackboardData              blackboardData,
                     boolean                     isConnectionLess,
                     TEP_V1_SessionEventCallback sessionEventCallback);

extern void TEP_V1_Authenticate(TEP_V1            tep,
                                boolean           isInitiator,
                                void            * sessionEventCallbackContext,
                                OALChannel        associatedChannel,
                                DOFAddress        remoteAddress,
                                DOFPermissionSet  extraPerms);

extern void TEP_V1_ExtendPermissions(OALChannel         channel,
                                     DOFAddress         address,
                                     DOFPermissionSet   extendPerms,
                                     ExtendCompleteCB   callback,
                                     void             * context);

extern void TEP_V1_EndSession(OALChannel channel);

extern boolean TEP_V1_SendNoOp(OALChannel channel);

/* *********************************************************************************************/

extern void TEP_V1_Destroy(TEP_V1 self);

extern ProtocolLayerResult TEP_V1_Receive(TEP_V1 self);

extern ProtocolLayerResult TEP_V1_Send(TEP_V1       self,
                                       DOFOperation operation,
                                       boolean      isProviderOperation,
                                       DOFResult    result);

#ifdef  __cplusplus
}
#endif

#endif
