/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_INTERNAL_THREADPOOL_H
#define DOF_INTERNAL_THREADPOOL_H

#include <dof/core.h>

#ifdef  __cplusplus
extern "C" {
#endif


/** @addtogroup OALThreadPool

    The OALThreadPool is a combination of a queue and a fixed-size thread pool.
    All of the worker threads are allocated and started when the threadpool is
    created, thus there is deterministic performance throughout the lifetime of
    the pool.

    As long as free threads exist in the pool, work given to the pool will be
    executed immediately, and concurrently with existing work handled by the
    queue. It is important to remember that synchronization of the work is the
    responsibility of the user of the thread pool. It is also important to
    consider the effect of locking and blocking the worker threads, as this
    affects the performance of other work waiting in the queue. It is
    generally a bad idea to have dependencies between work in the same pool
    (e.g., a thread waiting on other work to be done by the same pool), as this
    can lead to a deadlock (if there is only one thread in the pool and its
    work is waiting on work in the queue, the queue work doesn't have a thread
    and can't be serviced).

    The simplest case of the threadpool is using a single thread. In this
    case, all work is serviced sequentially by the single thread. When the
    thread is busy, additional work can still be queued and will be processed
    in the order in which it was enqueued.

@{
**/

typedef struct OALThreadPool_t *OALThreadPool;

typedef struct OALThreadPoolItem_t * OALThreadPoolItem;

/**
    Thread Pool worker function.

    This is the standard prototype for a worker function to be used in a
    OALThreadPool work item.

    @param  argc  The number of arguments passed in @a argv. Must be less than 0x40000000
    @param  argv  An array of @a argc arguments for the function.
**/
typedef void (*OALThreadPoolFunc)(OALThreadPoolItem item,
                                  boolean           shutdown);

typedef struct OALThreadPoolItem_t
{
    OALThreadPoolItem next;

    OALThreadPoolFunc callback;
} OALThreadPoolNode_t;

extern void OALThreadPoolItem_Init(OALThreadPoolItem item,
                                   OALThreadPoolFunc fn);

/**
    Initializes the Threadpool Library.
    The OALThreadPool_Init function must be called before any other portion of the library can be used.

    @see OALThreadPool_Shutdown
**/
extern void          OALThreadPool_Init(void);

/**
    Terminates use of the Threadpool Library.
    The OALThreadPool_Shutdown function must be called when the library will no longer be used.

    @see DOF_Init
**/
extern void          OALThreadPool_Shutdown(void);

/**
    Create a new thread pool.

    @mustfree{OALThreadPool_Destroy}

    @param  nThreads  The number of threads in the pool. Must be >= 1.

    @return The newly created thread pool, or NULL if one could not be created.
**/
extern OALThreadPool OALThreadPool_Create ( uint16 nThreads );

/**
    Destroy a thread pool.

    Any work that is currently being processed by the pool (already allocated a
    worker thread) will be completed.

    If no @a destroyer function is provided, then any remaining, unprocessed
    work in the queue will also be completed normally by the worker threads
    before the pool is destroyed.

    If a @a destroyer function is provided, then any remaining, unprocessed
    work in the queue will be given to that function (using the current thread)
    before the pool is destroyed.

    @param  tpool      The thread pool to destroy. If NULL, no action is taken.
**/
extern void          OALThreadPool_Destroy( OALThreadPool tpool);

/**
    Add work to be processed by the thread pool.
    Work for a thread pool consists of a function and a list of arguments.

    If a worker thread is available, processing will start almost immediately.
    If all of the worker threads are currently busy, the work is queued for
    later processing when a worker thread becomes available.

    @param  tpool   The thread pool to which work is to be added. Must not be NULL.
    @param  item    An initialized threadpool item. Must not be NULL.

    @return TRUE if the work items was enqueued, or FALSE if the work item
    could not be enqueued.
**/
extern boolean       OALThreadPool_Run    ( OALThreadPool tpool, OALThreadPoolItem item );

/**
@}
**/

#ifdef  __cplusplus
}
#endif

#endif /* _DOF_THREADPOOL_H */