/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_TRP_V1_H
#define DOF_TRP_V1_H

#include <dof/core.h>
#include <internal/stream.h>
#include <internal/packet.h>
#include <internal/channel.h>
#include <internal/blackboarddata.h>
#include <internal/security_common.h>
#include <internal/connection_internal.h>

#define TRP_V1_OP_REJECT                0x00
#define TRP_V1_OP_KEK                   0x01
#define TRP_V1_OP_RANDOM                0x02
#define TRP_V1_OP_SESSION               0x03
#define TRP_V1_OP_SECURITY_SCOPE        0x04
#define TRP_V1_OP_DOMAINCREDENTIAL      0x05
#define TRP_V1_OP_CREDENTIAL            0x06
#define TRP_V1_OP_LOCAL_DOMAIN          0x07
#define TRP_V1_OP_REMOTE_DOMAIN         0x08
#define TRP_V1_OP_VALIDATE_CREDS        0x09
#define TRP_V1_OP_DISCOVER_REMOTE       0x0A
#define TRP_V1_OP_SECURITY_SCOPE_EXT    0x0B

#define TRP_V1_APPID    129

#define TRP_V1_SUCCESS                                0L
#define TRP_V1_PARSE_ERROR                            1L
#define TRP_V1_ACCESS_DENIED                          2L
#define TRP_V1_UNKNOWN_INITIATOR                      3L
#define TRP_V1_UNKNOWN_RESPONDER                      4L
#define TRP_V1_UNKNOWN_DOMAIN                         5L
#define TRP_V1_SERVICE_UNAVAILABLE                    6L
#define TRP_V1_BAD_MODE                               7L
#define TRP_V1_INITIATOR_NOT_GROUP_MEMBER             9L
#define TRP_V1_UNSUPPORTED_IDENTIFICATION_INITIATOR   10L
#define TRP_V1_UNSUPPORTED_IDENTIFICATION_RESPONDER   11L
#define TRP_V1_INVALID_IDENTIFICATION_INITIATOR       12L
#define TRP_V1_INVALID_IDENTIFICATION_RESPONDER       13L
#define TRP_V1_RESOLUTION_FAILED                      14L
#define TRP_V1_VALIDATION_FAILED                      15L
#define TRP_V1_INTERNAL_ERROR                         127L

#ifdef  __cplusplus
extern "C" {
#endif

typedef struct TRP_V1_t
{
    BlackboardData  blackboardData;
    boolean         isConnectionless;
} TRP_V1_t;

typedef struct TRP_V1_t* TRP_V1;

/* If AS returned an error code status will be DOSSTATUS_OK and errorCode will hold the AS returned error code */
typedef struct KEK_return_t
{
    uint8                   thb;
    uint8                   tmin;
    uint8                   tmax;
    uint16                  epoch;
    DOFSecurityScope        secScope;
    DOFSecurityMode         securityMode;
    uint32                  stateID;
    DOFObjectIDSource       sid;
    DOFPermissionSet        groupsPermissions;
    DOFPermissionSet        initPermissions;
    uint8*                  KEK;
    DOFError                status;
    uint8                   errorCode; 
    uint8*                  errorData;
}KEK_return_t;

typedef struct Random_return_t
{
    uint8*              Random;
    DOFError            status;
    uint8               errorCode;  
    uint8*              errorData;
}Random_return_t;

typedef struct Session_return_t
{
    OALStream           A;
    uint8*              sessionKey;
    uint8               grantedDuration;
    DOFSecurityScope    connSecurityScope;
    DOFPermissionSet    initPermValidation;
    DOFPermissionSet    respPermValidation;
    OALKeyGrant         initGrant;
    DOFError            status;
    uint8               errorCode;  
    uint8*              errorData;
}Session_return_t;

typedef struct Scope_return_t
{
    uint8                   grantedDuration;
    DOFSecurityScope        securityScope;
    DOFPermissionSet        permValidation;
    DOFError                status;
    uint8                   errorCode;
    uint8*                  errorData;
}Scope_return_t;

typedef struct Resolve_return_t
{
    DOFCredentials      creds;
    DOFError            status;
    uint8               errorCode;  
    uint8*              errorData;
}Resolve_return_t;

typedef struct Validate_return_t
{
    DOFCredentials      creds;
    DOFError            status;
    uint8               errorCode;  
    uint8*              errorData;
    DOFObjectIDSource   sid;
}Validate_return_t;

typedef struct RemoteDomain_return_t
{
    DOFCredentials      creds;
    uint32              remoteDomainIdentifier;
    DOFError            status;
    uint8               errorCode;  
    uint8*              errorData;
}RemoteDomain_return_t;

typedef struct KEK_return_t* KEK_return;
typedef struct Random_return_t* Random_return;
typedef struct Session_return_t* Session_return;
typedef struct Scope_return_t* Scope_return;
typedef struct Resolve_return_t* Resolve_return;
typedef struct Validate_return_t* Validate_return;
typedef struct RemoteDomain_return_t* RemoteDomain_return;

typedef union trp_v1_context
{
    KEK_return_t* KEKReturn;
    Random_return_t* RandomReturn;
    Session_return_t* SessionReturn;
    Scope_return_t* ScopeReturn;
    Resolve_return_t* ResolveReturn;
    Validate_return_t* ValidateReturn;
    RemoteDomain_return_t* RemoteDomainReturn;
}trp_v1_context;

typedef union trp_v1_context* TRP_V1_Context;

typedef void (*trp_callback_fn)(TRP_V1_Context trp_context, void* context);

extern TRP_V1 TRP_V1_Create(BlackboardData blackboardData, boolean isConnectionless);
extern void TRP_V1_Destroy(TRP_V1 self);

extern boolean TRPRouter_MutexInit(void);
extern void TRPRouter_MutexShutdown(void);

extern boolean TRP_V1_IsRunning(void);

extern boolean TRP_V1_Init(void);
extern void TRP_V1_Shutdown(void);
extern boolean TRP_V1_SM_Create(void);
extern void TRP_V1_Shutdown1(void);
extern void TRP_V1_SM_Destroy(void);

extern ProtocolLayerResult TRP_V1_Receive(TRP_V1 self);
extern ProtocolLayerResult TRP_V1_Send(TRP_V1 self, DOFOperation operation, boolean isProviderOperation, DOFResult result);

extern void TRP_V1_SM_ReceiveLocal( uint32 argc, DOFOperation* argv, boolean shutdown );
extern void TRP_V1_SM_SendLocal( uint32 argc, DOFOperation* argv, boolean shutdown );

extern void TRP_V1_RequestKEK( DOFKeyRequest keyReq, DOFObjectIDAuthentication groupID, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout);
extern void TRP_V1_RequestRandom( DOFKeyRequest keyReq, trp_callback_fn trp_fn, void* context, uint32 timeout );
extern void TRP_V1_RequestSession( OALStream rBlock, DOFKeyRequest respKeyReq, OALStream iBlock, OALSessionInitiatorBlock sessInitBlock, OALChannel channel, DOFAddress address, trp_callback_fn trp_fn, void* context, uint32 timeout );
extern void TRP_V1_RequestSecurityScope( uint8 duration, DOFKeyRequest initKeyReq, DOFObjectIDAuthentication secNodeID, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout);
extern void TRP_V1_RequestSecurityScopeAndMode(uint8 duration, DOFKeyRequest initKeyReq, DOFObjectIDAuthentication secNodeID, DOFAuthenticatorSecurityType secType, uint32 modeLen, uint8* mode, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout);
extern void TRP_V1_RequestResolveCredentials( DOFCredentials creds, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout );
extern void TRP_V1_RequestValidateCredentials(DOFCredentials creds, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout);
extern void TRP_V1_RequestRemoteDomain( DOFObjectIDDomain domain, DOFKeyRequest keyReq, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout);

/* These are exposed internally because DOFCredentials needed them. Please verify before using it anywhere else. You most likely want DOFCredentials Resolve and Validate */
extern void TRP_V1_Internal_RequestResolveCredentials( DOFObjectIDDomain domain, DOFOperation parent, DOFCredentials creds, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout, boolean force );
extern void TRP_V1_Internal_RequestValidateCredentials(DOFCredentials creds, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout, boolean force);
extern void TRP_V1_RequestRemoteDomainForcable( DOFObjectIDDomain domain, DOFKeyRequest keyReq, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout, boolean force );
/* End of DOFCredentials exposed functions */

extern boolean TRPRouter_HasAnAS(void);
extern boolean TRPRouter_KnowsAnAS(void);
extern boolean TRPRouter_AddAuthenticator( DOFAuthenticator domain );
extern void TRPRouter_RemoveAuthenticator( DOFAuthenticator domain );
extern boolean TRPRouter_AddCredentials( DOFCredentials creds, DOFConnection connection );
extern void TRPRouter_RemoveCredentials( DOFCredentials creds, DOFConnection connection );
extern boolean TRPRouter_AddTrustedDomain( DOFObjectIDDomain domain, OALChannel channel );
extern void TRPRouter_RemoveTrustedDomain( DOFObjectIDDomain domain, DOFConnection connection);
extern boolean TRPRouter_RouteToAS(DOFOperation op, OALChannel channel, DOFAddress address, boolean force);

/* Added For TEPs and TRPs use only */
extern void TRPRouter_AddDynamicRoute(DOFCredentials creds, OALChannel channel, DOFAddress address, boolean addAlways);

extern DOFError TRP_V1_GetDOFErrorFromErrorCode( uint8 errorCode );

/**
 * Triggers the given credentials to go back to unresolved status (if possible) if the TRP error code signals that 
 * something may have changed in the storage. 
 *  
 * @note The return should get passed on so that if this was triggered by the application the appropriate 
 *       DOFException will be returned to the application. 
 * 
 * @param creds Our Credentials used for the request.
 * @param initiator TRUE If we are the initiator, FALSE if not.
 * @param TRPErrorCode The error code returned by the trp request.
 * 
 * @return boolean TRUE a cache reset was triggered.
 *                 FALSE otherwise.
 */
extern boolean TRP_V1_TriggerPossibleCacheReset( DOFCredentials creds, boolean initiator, uint8 TRPErrorCode );

#ifdef  __cplusplus
}
#endif

#endif
