/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/address.h>
#include <internal/addressinternal.h>
#include <internal/enp_v1.h>


/** @addtogroup DOFAddress
@{
**/

#define DOFADDRESS_HEADER_SIZE      offsetof(DOFAddress_t, addr)

/*
**  "Private" Type and Method Declarations
*/


extern PCRMutex ref_count_mutex;
extern PCRMutex addr_mutex;

/*
**  "Public" Type and Method Declarations
*/

/*
**  "Public" Methods
*/

/* @internal */
void DOFAddress_ClearObservable( DOFAddress self )
{
    PCR_Assert( self );
    PCRMutex_Lock(addr_mutex);
    OALObservable_Destroy( self->addrDownObservable );
    self->addrDownObservable = NULL;
    PCRMutex_Unlock(addr_mutex);
}


/* @internal */
OALObserver DOFAddress_AddObserver( DOFAddress         self,
                                    OALObserver_fn_t   fn,
                                    void             * arg )
{
    OALObserver ret = NULL;

    PCR_Assert( self );
    PCRMutex_Lock(addr_mutex);
    if(self->addrDownObservable == NULL)
    {
        self->addrDownObservable = OALObservable_Create();
    }
    if(self->addrDownObservable != NULL)
    {
        ret = OALObservable_AddObserver(self->addrDownObservable, fn, arg);
    }
    PCRMutex_Unlock(addr_mutex);

    return ret;
}

void DOFAddress_RemoveObserver( DOFAddress self,
                                OALObserver observer )
{
    PCR_Assert(self != NULL);
    PCRMutex_Lock(addr_mutex);
    if(self->addrDownObservable != NULL)
    {
        OALObservable_RemoveObserver(self->addrDownObservable, observer);
    }
    PCRMutex_Unlock(addr_mutex);
}

DOFAddress DOFAddress_Create( uint32 addrSize, const void *addr, DTLTransport transport )
{
    if(transport && ((addrSize == 0 && !addr) || (addrSize > 0 && addr)))
    {
        DOFAddress newAddress;

        /* Allocate memory for the header and the address. */
        newAddress = (DOFAddress) DOFMem_Alloc( DOFADDRESS_HEADER_SIZE + addrSize );
        if (newAddress)
        {
            if ((newAddress->addrDownObservable = OALObservable_Create()) != NULL)
            {
                newAddress->len = addrSize;
                newAddress->refcnt = 1;
                newAddress->transport = transport;
                newAddress->enpPort = ENP_INVALID_PORT;
                if(addrSize != 0)
                {
                    PCRMem_Copy( ((uint8*)newAddress)+DOFADDRESS_HEADER_SIZE, addr, addrSize );
                    if( transport->fns->AddressCopy )
                    {
                        if( transport->fns->AddressCopy(transport, addrSize, ((uint8*)newAddress)+DOFADDRESS_HEADER_SIZE) )
                        {
                            return newAddress;
                        }
                    }
                    else
                    {
                        return newAddress;
                    }
                }
                else
                    return newAddress;

                OALObservable_Destroy(newAddress->addrDownObservable);
            }
            DOFMem_Free( newAddress );
        }
        DOFLog_Message( DOFLOGLEVEL_ERROR, "DOFAddress: Failed to create address (out of memory).");
    }
    else
    {
        DOFLog_Message( DOFLOGLEVEL_ERROR, "DOFAddress: Failed to create address (invalid parameters).");
    }
    return NULL;
}

DOFAddress DOFAddress_Create_AddENPPort(DOFAddress self, uint32 enpPort)
{
    if (self && ((enpPort <= ENP_MAX_PORT) || (enpPort == ENP_INVALID_PORT)))
    {
        DOFAddress newAddress;
        uint32 newAddrSize;

        /* Allocate memory for the header and the address. */
        newAddrSize = DOFADDRESS_HEADER_SIZE + DOFAddress_GetAddressSize(self);
        newAddress = (DOFAddress) DOFMem_Alloc( newAddrSize );
        if (newAddress)
        {
            if ((newAddress->addrDownObservable = OALObservable_Create()) != NULL)
            {
                boolean fail = FALSE;
                PCRMem_Copy(((uint8*)newAddress)+DOFADDRESS_HEADER_SIZE, (void *)&self->addr, self->len);
                newAddress->refcnt = 1;
                newAddress->len = self->len;
                newAddress->transport = self->transport;
                newAddress->enpPort = enpPort;

                PCR_Assert( self->transport );
                PCR_Assert( self->transport->fns );
                if ( self->transport->fns->AddressCopy )
                {
                    if( !newAddress->transport->fns->AddressCopy(newAddress->transport, newAddress->len, ((uint8*)newAddress)+DOFADDRESS_HEADER_SIZE) )
                        fail = TRUE;
                }

                if(!fail)
                    return newAddress;

                OALObservable_Destroy( newAddress->addrDownObservable );
            }
            DOFMem_Free(newAddress);
        }
    }

    DOFLog_Message( DOFLOGLEVEL_ERROR, "DOFAddress: Failed to add enp port to address.");
    return NULL;
}

DOFAddress DOFAddress_Copy( const DOFAddress self )
{
    if(self != NULL)
    {
        PCRMutex_Lock(ref_count_mutex);
        self->refcnt++;
        PCRMutex_Unlock(ref_count_mutex);
    }
    return self;
}

boolean DOFAddress_IsEqual( const DOFAddress self, const DOFAddress addr )
{
    return (DOFAddress_Compare( self, addr ) == 0);
}

int DOFAddress_CompareAddressInfo( const DOFAddress self, const DOFAddress addr)
{
    if( !self && !addr )
        return 0;

    if( !self )
        return -1;

    if( !addr )
        return 1;

    if( self->transport > addr->transport )
        return 1;
    if( self->transport < addr->transport )
        return -1;
    if (self->len > addr->len )
        return 1;
    if (self->len < addr->len )
        return 1;

    PCR_Assert( self->transport );
    PCR_Assert( self->transport->fns );
    if ( self->transport->fns->AddressCompare )
        return ( self->transport->fns->AddressCompare( self->transport, self->len, (void*)&self->addr, (void*)&addr->addr ) );

    return PCRMem_Compare( DOFAddress_GetAddress(self), DOFAddress_GetAddress(addr),  self->len );
}

/* Same as DOFAddress_IsEqual however the ENP Port is not used in the comparison. */
boolean DOFAddress_IsAddressInfoEqual( const DOFAddress self, const DOFAddress addr )
{
    return (DOFAddress_CompareAddressInfo( self, addr ) == 0);
}

int DOFAddress_Compare( const DOFAddress self, const DOFAddress addr )
{
    int rval = DOFAddress_CompareAddressInfo( self, addr );

    if ( rval == 0 && self && addr)
    {
        if( self->enpPort > addr->enpPort )
            return 1;
        if( self->enpPort < addr->enpPort )
            return -1;
    }

    return rval;
}

void DOFAddress_Destroy( DOFAddress self )
{
    boolean gone;
    if (self)
    {
        PCRMutex_Lock(ref_count_mutex);
        PCR_Assert(self->refcnt);
        --self->refcnt;
        gone = (!self->refcnt);
        PCRMutex_Unlock(ref_count_mutex);
        if ( gone )
        {
            PCRMutex_Lock(addr_mutex);
            if ( self->addrDownObservable )
            {
                OALObservable_SetChangedAndNotifyObservers( self->addrDownObservable );
                OALObservable_Destroy( self->addrDownObservable );
            }
            PCRMutex_Unlock(addr_mutex);

            PCR_Assert( self->transport );
            PCR_Assert( self->transport->fns );
            if ( self->transport->fns->AddressDestroy )
            {
                self->transport->fns->AddressDestroy( DOFAddress_GetTransport( self ),
                                                      DOFAddress_GetAddressSize( self ),
                                                      DOFAddress_GetAddress( self ) );
            }
            DOFMem_Free( self );
        }
    }
}

const void * DOFAddress_GetAddress( const DOFAddress self )
{
    if( self && self->len )
        return ((void*)&self->addr);
    return NULL;
}

DOFAddressType DOFAddress_GetAddressType( const DOFAddress self )
{
    if ( self )
    {
        PCR_Assert( self->transport );
        PCR_Assert( self->transport->fns );
        if ( self->transport->fns->GetAddressType )
        {
            return self->transport->fns->GetAddressType( DOFAddress_GetTransport( self ), self );
        }
    }
    return DOFADDRESSTYPE_UNICAST;
}

uint32 DOFAddress_GetAddressSize( const DOFAddress self )
{
    if( self )
        return self->len;
    return 0;
}

DTLTransport DOFAddress_GetTransport( const DOFAddress self )
{
    if( self )
        return self->transport;
    return NULL;
}

uint32 DOFAddress_GetENPPort( const DOFAddress self )
{
    if (self)
    {
        PCR_Assert((self->enpPort == ENP_INVALID_PORT) || (self->enpPort < ENP_MAX_PORT));
        return self->enpPort;
    }
    return ENP_INVALID_PORT;
}

/**
@}
**/
