/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/avl_tree.h>
#include <internal/avl_multitree.h>
#include <internal/dofinternal.h>

/** @internal @addtogroup OALAVLMultiTree

The OALAVLMultiTree is an AVL tree that gracefully handles items with duplicate
keys. It won't, however, gracefully handle items with duplicate keys and
duplicate data. So don't do that! :)

@{
    **/
extern void *OALAVLTree_get_root_item(OALAVLTree tree);

/** @internal
Internal avl multitree node data
**/
typedef struct multi_node {
    union {
        OALAVLTree tree;
        OALAVLMultiTreeItem_t item;
    } tree_or_item;
    uint8 item;
} multi_node;

/** @internal
Internal avl multitree instance data
**/
typedef struct OALAVLMultiTree_t {
    uint32 num_items;

    OALAVLTree parent;
    multi_node *next_tree;

    OALAVLTree_CompareGreaterThan_fn_t key_gt_fn;
    OALAVLTree_CompareEqual_fn_t key_eq_fn;

    OALAVLTree_CompareGreaterThan_fn_t item_gt_fn;
    OALAVLTree_CompareEqual_fn_t item_eq_fn;

    void *context;
} OALAVLMultiTree_t;

#define BALANCED(n) (n->opaque<0)

#define NEITHER -1
#define LEFT 0
#define RIGHT 1

typedef int8 direction;

/** @internal
Perform a basic greater-than comparison on a key and an item in the tree.

@param d1 The key in question.
@param d2 The item in the table to compare to the key.
@param context Unused.

@return True if the greater-than comparison is valid.
**/
static boolean __multi_gt(OALAVLMultiTreeKey_t d1, OALAVLMultiTreeKey_t d2, void *context)
{
    return d1 > d2;
}

/** @internal
Perform a basic is-equal comparison on a key and an item in the tree.

@param d1 The key in question.
@param d2 The item in the table to compare to the key.
@param context Unused.

@return True if the is-equal comparison is valid.
**/
static boolean __multi_eq(OALAVLMultiTreeKey_t d1, OALAVLMultiTreeKey_t d2, void *context)
{
    return d1 == d2;
}

/** @internal
Call the user item greater-than compare function.

@param key The key in question.
@param data The item in the table to compare to the key.
@param context Unused.

@return True if the greater-than comparison is valid.
**/
static boolean _multi_gt(OALAVLMultiTreeKey_t key, OALAVLMultiTreeItem_t data, OALAVLMultiTree context)
{
    return context->item_gt_fn(key, data, context->context);
}

/** @internal
Call the user item is-equal compare function.

@param key The key in question.
@param data The item in the table to compare to the key.
@param context Unused.

@return True if the is-equal comparison is valid.
**/
static boolean _multi_eq(OALAVLMultiTreeKey_t key, OALAVLMultiTreeItem_t data, OALAVLMultiTree context)
{
    return context->item_eq_fn(key, data, context->context);
}

/** @internal
Call the user item greater-than compare function.

@param key The key in question.
@param n The subtree that holds the root key.
@param context Unused.

@return True if the greater-than comparison is valid.
**/
static boolean multi_gt(OALAVLMultiTreeKey_t key, multi_node *n, OALAVLMultiTree context)
{
    return context->key_gt_fn( key, n->item ? n->tree_or_item.item : OALAVLTree_get_root_item(n->tree_or_item.tree),
                               context->context );
}

/** @internal
Call the user key is-equal compare function.

@param key The key in question.
@param n The subtree that holds the root key.
@param context Unused.

@return True if the is-equal comparison is valid.
**/
static boolean multi_eq(OALAVLMultiTreeKey_t key, multi_node *n, OALAVLMultiTree context)
{
    return context->key_eq_fn( key, n->item ? n->tree_or_item.item : OALAVLTree_get_root_item(n->tree_or_item.tree), context->context );
}

/** @internal
Constructs a multi_node for the parent tree.

@param tree The parent tree.

@return The new multi_node.
**/
static multi_node *make_multi_node(OALAVLMultiTree tree)
{
    multi_node *n = DOFMem_Alloc(sizeof(multi_node));

    if (!n)
    {
        return NULL;
    }

    n->item = 1;

    return n;
}

/** @internal
Destroy a multi_node.

@param n The multi_node to destroy.
**/
static void destroy_multi_node(multi_node *n)
{
    if(!n->item) OALAVLTree_Destroy(n->tree_or_item.tree, 0, 0);
    DOFMem_Free(n);
}

/** @internal
Create a new OALAVLMultiTree.

@param key_gt_fn The key greater-than compare function.
@param key_eq_fn The key is-equal compare function.
@param item_gt_fn The item greater-than compare function.
@param item_eq_fn The item is-equal compare function.
@param context The context to pass to the compare functions for this tree.

@return The new tree or NULL if an error occurred.
**/
OALAVLMultiTree OALAVLMultiTree_Create( OALAVLMultiTree_CompareGreaterThan_fn_t key_gt_fn,
                                        OALAVLMultiTree_CompareEqual_fn_t key_eq_fn,
                                        OALAVLMultiTree_CompareGreaterThan_fn_t item_gt_fn,
                                        OALAVLMultiTree_CompareEqual_fn_t item_eq_fn, void *context )
{
    OALAVLMultiTree_t *tree = DOFMem_Alloc(sizeof(OALAVLMultiTree_t));
    OALAVLTree parent;

    PCR_Assert(key_gt_fn&&key_eq_fn);

    if(!tree) return NULL;

    parent = OALAVLTree_Create( (OALAVLTree_CompareGreaterThan_fn_t)multi_gt, (OALAVLTree_CompareEqual_fn_t)multi_eq,
                                tree );

    if (!parent)
    {
        DOFMem_Free(tree);
        return NULL;
    }

    tree->parent = parent;
    tree->key_gt_fn = key_gt_fn;
    tree->key_eq_fn = key_eq_fn;
    tree->item_gt_fn = item_gt_fn ? item_gt_fn : (OALAVLMultiTree_CompareEqual_fn_t)__multi_gt;
    tree->item_eq_fn = item_eq_fn ? item_eq_fn : (OALAVLMultiTree_CompareEqual_fn_t)__multi_eq;
    tree->context = context;
    tree->num_items = 0;

    tree->next_tree = make_multi_node(tree);

    if(!tree->next_tree)
    {
        OALAVLTree_Destroy(parent, 0, 0);
        DOFMem_Free(tree);

        return NULL;
    }

    return tree;
}

typedef struct delete_t {
    OALAVLMultiTree_RemoveAll_fn_t fn;
    void *arg;
} delete_t;

static void _delete(multi_node *n, delete_t *arg)
{
    if(!n->item) {
        OALAVLTree_Destroy(n->tree_or_item.tree, (OALAVLTree_RemoveAll_fn_t)arg->fn, arg->arg);
    }
    else {
        arg->fn(n->tree_or_item.item, arg->arg);
    }
    DOFMem_Free(n);
}

/**
Destroys an OALAVLMultiTree.

 @param tree The tree to destroy.
 @param remove_fn The function to call for each node removed.
 @param arg The argument to pass to the remove_fn.
 **/
void OALAVLMultiTree_Destroy(OALAVLMultiTree tree, OALAVLMultiTree_RemoveAll_fn_t remove_fn, void *arg)
{
    delete_t d;

    d.fn = remove_fn;
    d.arg = arg;

    destroy_multi_node(tree->next_tree);
    OALAVLTree_Destroy(tree->parent, (OALAVLTree_RemoveAll_fn_t)_delete, &d);

    DOFMem_Free(tree);
}

/**
Retrieve an item (the first one!) with a given key.

 @param tree The tree to retrieve an item from.
 @param key The key to use for lookup.

 @return The found item or NULL if no such item exists.
 **/
OALAVLMultiTreeItem_t OALAVLMultiTree_Retrieve(OALAVLMultiTree tree, OALAVLMultiTreeItem_t key)
{
    const void  *retval = 0;

    PCR_Assert(tree);

    {
        const multi_node *n = OALAVLTree_Retrieve(tree->parent, key);

        if (!n)
            retval = 0; /* didn't find it */
        else
        {
            if (n->item)
                retval = n->tree_or_item.item;
            else
                retval = OALAVLTree_get_root_item(n->tree_or_item.tree);
        }
    }

    return retval;
}

/**
Retrieve the item associated with the given key if it's in the tree.

 @param tree The tree to retrieve an item from.
 @param key The key to use for lookup.
 @param item The item to find.

 @return The found item or NULL if no such item exists.
 **/
OALAVLMultiTreeItem_t OALAVLMultiTree_Retrieve_Item( OALAVLMultiTree tree, OALAVLMultiTreeKey_t key,
                                                     OALAVLMultiTreeItem_t item )
{
    const void *retval = 0;

    PCR_Assert(tree);

    {
        const multi_node *n = OALAVLTree_Retrieve(tree->parent, key);

        if (!n)
            retval = 0; /* didn't find it */
        else
        {
            if (n->item)
                retval = n->tree_or_item.item;
            else
                retval = OALAVLTree_Retrieve(n->tree_or_item.tree, item);
        }
    }

    return retval;
}

/**
Get the number of items in the tree.

 @param tree The tree to get the number of items from.

 @return The number of items in the tree.
 **/
uint32 OALAVLMultiTree_GetNumItems(OALAVLMultiTree tree)
{
    PCR_Assert(tree);
    return tree->num_items;
}

/******************************************************
* INSERTION                                           *
*****************************************************/

/**
Insert an item into the multi tree.

@param tree The tree to insert the item into.
@param key The key that will be used to retrieve the item.
@param item The item itself.

**/
OALAVLTree_InsertionStatus OALAVLMultiTree_Insert( OALAVLMultiTree tree, OALAVLMultiTreeKey_t key,
                                                   OALAVLMultiTreeItem_t item )
{
    PCR_Assert( tree );
    PCR_Assert( tree->parent );
    PCR_Assert( tree->next_tree );
    {
        OALAVLTree_InsertionStatus status;
        multi_node *               n;

        tree->next_tree->tree_or_item.item = item;
        switch ( status = OALAVLTree_Insert(tree->parent, key, tree->next_tree, (OALAVLTreeItem_t*)&n) )
        {
            /* status can be    OALAVLTree_InsertSuccess   (and tree->parent newly holds key/tree->next_tree combo) OR
                                OALAVLTree_AlreadyInserted (and tree->parent not changed but known to hold key). */

            default:
                return status; /* an error occurred */
            case OALAVLTree_InsertSuccess:
                PCR_Assert( n==tree->next_tree );
                /* Put the NEW pre-allocated tree in right here, for next use. */
                tree->next_tree = make_multi_node(tree);
                if(!tree->next_tree)
                {
                    OALAVLTree_Delete(tree->parent, key);
                    return OALAVLTree_InsertFailure; /* an error occurred */
                }
                break;
            case OALAVLTree_AlreadyInserted:
                if (n->item)
                {
                    /* Item held here, so create a subtree to hold it and the new item. */

                    OALAVLMultiTreeItem_t old_item = n->tree_or_item.item;

                    if ( old_item == item )
                        return OALAVLTree_AlreadyInserted;

                    /* Create a new tree to hold multiple items. */
                    n->tree_or_item.tree = OALAVLTree_CreateSubTree( tree->parent,
                                                                     (OALAVLTree_CompareGreaterThan_fn_t)_multi_gt,
                                                                     (OALAVLTree_CompareEqual_fn_t)_multi_eq, tree );
                    if (!n->tree_or_item.tree)
                    {
                        /* Restore old item. */
                        n->tree_or_item.item = old_item;
                        return OALAVLTree_InsertFailure;
                    }

                    /* Insert the old item so it is still in the tree. */
                    status = OALAVLTree_Insert(n->tree_or_item.tree, old_item, old_item, 0);
                    /* Since it is a new tree, it is impossible that this could already be inserted. */
                    if(status != OALAVLTree_InsertSuccess)
                    {
                        /* Restore old item. */
                        OALAVLTree_Destroy( n->tree_or_item.tree, 0, 0 );
                        n->tree_or_item.item = old_item;
                        return OALAVLTree_InsertFailure;
                    }
                    /* At this point, the old item is back in the tree (we just have a new subtree). */
                    n->item = 0;
                }

                /* Either there was already a tree, or we just created one. Now insert the new item. */
                status = OALAVLTree_Insert(n->tree_or_item.tree, item, item, 0);
                break;
        }
        if(status == OALAVLTree_InsertSuccess)
        {
            /* We only added something on success. */
            ++tree->num_items;
        }

        return status;
    }
}

/******************************************************
* DELETION                                           *
*****************************************************/

/**
Delete an item from the tree.

 @param tree The tree to remove an item from.
 @param key  The key that is used to find the item.
 @param item The item to remove.

 @return TRUE if something deleted; otherwise FALSE.
 **/
boolean OALAVLMultiTree_Delete(OALAVLMultiTree tree, OALAVLMultiTreeKey_t key, OALAVLMultiTreeItem_t item)
{
    PCR_Assert(tree);
    {
        const multi_node *n = OALAVLTree_Retrieve(tree->parent, key);

        if (!n)
        {
            return FALSE;
        }

        if (n->item || OALAVLTree_GetNumItems(n->tree_or_item.tree)==1)
        {
            /* we're going to delete the sub tree here */
            if(!OALAVLTree_Delete(tree->parent, key))
            {
                PCR_Assert(0); /* internal error since we know it exists */
            }
            destroy_multi_node((multi_node*)n);
        }
        else if(!OALAVLTree_Delete(n->tree_or_item.tree, item))
        {
            return FALSE; /* we have the key in the tree, but the given item is not */
        }

        --tree->num_items;
    }

    return TRUE;
}

/******************************************************
* TRAVERSE                                           *
*****************************************************/

/**
Traverse all items with a given key.

 @param tree The tree to traverse.
 @param key The key indicating the items to traverse.
 @param traverse_fn The function to be called for each item found.
 @param arg The argument to pass to traverse_fn.
 **/
void OALAVLMultiTree_Traverse( OALAVLMultiTree tree, OALAVLMultiTreeKey_t key,
                               OALAVLMultiTree_Traversal_fn_t traverse_fn, void *arg )
{
    PCR_Assert(tree);
    PCR_Assert(traverse_fn);

    {
        const multi_node *n = OALAVLTree_Retrieve(tree->parent, key);

        if (n)
        {
            if(n->item)
                traverse_fn(n->tree_or_item.item, arg);
            else
                OALAVLTree_Traverse(n->tree_or_item.tree, (OALAVLTree_Traversal_fn_t)traverse_fn, arg);
        }
    }
}

typedef struct traverse_all_t
{
    OALAVLMultiTree_Traverse_All_fn_t fn;
    void *arg;
    OALAVLMultiTree_Traverse_All_Return_Type_fn_t ret;
    boolean new_keyset;
} traverse_all_t;

static boolean __traverse_all(void *data, traverse_all_t *t)
{
    static const boolean ret_map[] = { FALSE,  /* OALAVLMULTITREE_NORMAL */
                                       TRUE,   /*                _TERMINATE_SET */
                                       TRUE }; /*                _TERMINATE_TRAVERSE. */

    t->ret = t->fn(data, t->arg, t->new_keyset);
    t->new_keyset = FALSE;
    PCR_Assert(t->ret<OALAVLMULTITREE_INVALID_RETURN);

    return ret_map[t->ret];
}

static boolean _traverse_all( multi_node *n, traverse_all_t *t )
{
    static const boolean ret_map[] = { FALSE,  /* OALAVLMULTITREE_NORMAL */
                                       FALSE,  /*                _TERMINATE_SET */
                                       TRUE }; /*                _TERMINATE_TRAVERSE. */

    t->ret = OALAVLMULTITREE_NORMAL;
    t->new_keyset = TRUE;

    if(n->item)
        __traverse_all((void*)n->tree_or_item.item, t);
    else
        OALAVLTree_Traverse(n->tree_or_item.tree, (OALAVLTree_Traversal_fn_t)__traverse_all, t);

    return ret_map[t->ret];
}

/**
Traverse all items in the mutli tree.

 @param tree The tree to traverse.
 @param traverse_fn The function to be called for each node.
 @param arg The argument to pass to traverse_fn.
 **/
void OALAVLMultiTree_Traverse_All(OALAVLMultiTree tree, OALAVLMultiTree_Traverse_All_fn_t traverse_fn, void *arg)
{
    PCR_Assert(tree);
    PCR_Assert(traverse_fn);

    {
        traverse_all_t t;

        t.fn = traverse_fn;
        t.arg = arg;
        /* other fields initialized in _traverse_all() */

        OALAVLTree_Traverse(tree->parent, (OALAVLTree_Traversal_fn_t)_traverse_all, &t);
    }
}

/**
@}
 **/
