/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/avl_tree.h>
#include <internal/mutex_pool.h>
#include <internal/dofinternal.h>

/** @internal @addtogroup OALAVLTree

 The OALAVLTree is a self-balancing tree implementation. This class
 provides the notion of parent and child trees. Also can be a "sub-tree"
 for overriding basic tree functionality in child classes.

 @{
 **/

/*
 **  "Private" Type and Method Declarations
 */

/* lame prototypes to satisfy warnings */
OALAVLTreeItem_t OALAVLTree_get_root_item(OALAVLTree tree);

/** @internal
 Internal avl tree node data
 **/

typedef struct _OALAVLTreeNode_t {
    OALAVLTreeItem_t data;

    /* don't touch these */
    struct _OALAVLTreeNode_t *nodes[2];
    int8 opaque;
} _OALAVLTreeNode_t, *_OALAVLTreeNode;

/** @internal
 Internal avl tree instance data
 **/

typedef struct OALAVLTree_t {
    _OALAVLTreeNode root;
    uint32 num_items;

    OALAVLTree_CompareGreaterThan_fn_t gt_fn;
    OALAVLTree_CompareEqual_fn_t eq_fn;

    OALAVLTree parent;
    void *context;
} OALAVLTree_t;

#define BALANCED(n) (n->opaque<0)

#define NEITHER -1
#define LEFT 0
#define RIGHT 1

typedef int8 direction;

/** @internal
 Perform a basic greater-than comparison on a key and an item in the tree.

 @param key The key in question.
 @param item The item in the table to compare to the key.
 @param context Unused.

 @return True if the greater-than comparison is valid.
 **/
static boolean basic_gt(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    return key>item;
}

/** @internal
 Perform a basic is-equal comparison on a key and an item in the tree.

 @param key The key in question.
 @param item The item in the table to compare to the key.
 @param context Unused.

 @return True if the is-equal comparison is valid.
 **/
static boolean basic_eq(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    return key==item;
}

/** @internal
 Common creation and initialization code for avl trees and sub-trees.

 @param gt_fn The greater-than comparison function.
 @param eq_fn The is-equal comparison function.
 @param context The context to be used with comparison functions.

 @return An allocated and initialized tree, or NULL if a memory error occurred.
 **/
static OALAVLTree _create(OALAVLTree_CompareGreaterThan_fn_t gt_fn, OALAVLTree_CompareEqual_fn_t eq_fn, void *context)
{
    OALAVLTree tree = (OALAVLTree)DOFMem_Alloc(sizeof(OALAVLTree_t));

    if(tree) {
        tree->root = 0;
        tree->num_items = 0;
        tree->context = context;
        tree->parent = 0;

        tree->gt_fn = gt_fn ? gt_fn : basic_gt;
        tree->eq_fn = eq_fn ? eq_fn : basic_eq;
    }

    return tree;
}

/**
 Create an AVL tree.

 @param gt_fn The greater-than comparison function.
 @param eq_fn The is-equal comparison function.
 @param context The context to be used with comparison functions.

 @return An allocated and initialized tree, or NULL if a memory error occurred.
 **/
OALAVLTree OALAVLTree_Create(OALAVLTree_CompareGreaterThan_fn_t gt_fn, OALAVLTree_CompareEqual_fn_t eq_fn, void *context)
{
    OALAVLTree tree = _create(gt_fn, eq_fn, context);

    return tree;
}

/**
 Create an AVL sub-tree.

 @param parent The parent tree.
 @param gt_fn The greater-than comparison function.
 @param eq_fn The is-equal comparison function.
 @param context The context to be used with comparison functions.

 @return An allocated and initialized tree, or NULL if a memory error occurred.
 **/
OALAVLTree OALAVLTree_CreateSubTree(OALAVLTree parent, OALAVLTree_CompareGreaterThan_fn_t gt_fn, OALAVLTree_CompareEqual_fn_t eq_fn, void *context)
{
    OALAVLTree tree = _create(gt_fn, eq_fn, context);
    if(tree) tree->parent = parent;
    return tree;
}

/** @internal

 Post-order traversal that deletes the current node in the tree.

 @param node The current node.
 @param remove_fn A function to be called when removing the data.
 @param arg The argument to be passed to the remove function.
 **/
static void _delete_traverse(_OALAVLTreeNode node, OALAVLTree_RemoveAll_fn_t remove_fn, void *arg)
{
    if(!node) return;
    _delete_traverse(node->nodes[0], remove_fn, arg);
    _delete_traverse(node->nodes[1], remove_fn, arg);
    if(remove_fn) remove_fn(node->data, arg);
    DOFMem_Free(node);
}

/**
 Destroy a tree.

 @param tree The tree to destroy.
 @param remove_fn The remove function to call when each node is destroyed.
 @param arg The argument to pass to the remove function.
 **/
void OALAVLTree_Destroy(OALAVLTree tree, OALAVLTree_RemoveAll_fn_t remove_fn, void *arg)
{
    _delete_traverse(tree->root, remove_fn, arg);

    DOFMem_Free(tree);
}

/**
 Empty a tree.

 @param tree The tree to empty.
 @param remove_fn The remove function to call when each node is destroyed.
 @param arg The argument to pass to the remove function.
 **/
void OALAVLTree_Empty(OALAVLTree tree, OALAVLTree_RemoveAll_fn_t remove_fn, void *arg)
{
    _delete_traverse(tree->root, remove_fn, arg);
    tree->root = 0;
    tree->num_items = 0;
}

/**
 Retrieve an item from the tree keyed with the given data.

 @param tree The tree to retrieve the data from.
 @param data The key to use when looking up the data.

 @return The item as found in the tree, or NULL if nothing was found.
 **/
OALAVLTreeItem_t OALAVLTree_Retrieve(OALAVLTree tree, OALAVLTreeKey_t data)
{
    _OALAVLTreeNode node;

    node = tree->root;

    while(node && !tree->eq_fn(data, node->data, tree->context)) {
        node = node->nodes[tree->gt_fn(data, node->data, tree->context)];
    }

    return node ? node->data : 0;
}

/**
 Return the number of items in a tree.

 @param tree The tree to get the number of items from.
 @return The number of items in the tree.
 **/
uint32 OALAVLTree_GetNumItems(OALAVLTree tree)
{
    PCR_Assert(tree);
    return tree->num_items;
}

/******************************************************
 * INSERTION                                           *
 *****************************************************/

static _OALAVLTreeNode avl_rotate_2(_OALAVLTreeNode *path_top, direction dir)
{
    _OALAVLTreeNode b, c, d, e;

    b = *path_top;
    d = b->nodes[dir];
    c = d->nodes[!dir];
    e = d->nodes[dir];
    *path_top = d;
    d->nodes[!dir] = b;
    b->nodes[dir] = c;
    b->opaque = NEITHER;
    d->opaque = NEITHER;

    return e;
}

static _OALAVLTreeNode avl_rotate_3(_OALAVLTreeNode *path_top, direction dir, int8 third)
{
    _OALAVLTreeNode b, f, d, c, e;

    b = *path_top;
    f = b->nodes[dir];
    d = f->nodes[!dir];
    c = d->nodes[!dir];
    e = d->nodes[dir];
    *path_top = d;
    d->nodes[!dir] = b;
    d->nodes[dir] = f;
    b->nodes[dir] = c;
    f->nodes[!dir] = e;
    d->opaque = NEITHER;

    /* assume both trees are balanced */
    b->opaque = f->opaque = NEITHER;

    if(third==NEITHER) return NULL;

    if(third==dir) {
        /* E holds the insertion so B is unbalanced */
        b->opaque = !dir;
        return e;
    }
    else {
        /* C holds the insertion so F is unbalanced */
        f->opaque = dir;
        return c;
    }
}

static void avl_rebalance_path(OALAVLTree tree, _OALAVLTreeNode path, const void *data)
{
    while(path && !(tree->eq_fn(data, path->data, tree->context))) {
        path = path->nodes[path->opaque = tree->gt_fn(data, path->data, tree->context)];
    }
}

static void avl_rebalance(OALAVLTree tree, _OALAVLTreeNode *path_top, const void *data)
{
    _OALAVLTreeNode path = *path_top;
    int8 first, second, third; /* directions where 0 is left and 1 is right */

    if(BALANCED(path)) {
        avl_rebalance_path(tree, path, data);
        return;
    }

    first = tree->gt_fn(data, path->data, tree->context);
    if(path->opaque != first) {
        /* took shorter path */
        path->opaque = NEITHER;
        avl_rebalance_path(tree, path->nodes[first], data);
        return;
    }

    /* took the longer path, so we need to rotate */
    second = tree->gt_fn(data, path->nodes[first]->data, tree->context);
    if(first==second) {
        /* two point rotate */
        path = avl_rotate_2(path_top, first);
        avl_rebalance_path(tree, path, data);
        return;
    }

    path = path->nodes[first]->nodes[second];
    if(tree->eq_fn(data, path->data, tree->context)) {
        third=NEITHER;
    }
    else {
        third = tree->gt_fn(data, path->data, tree->context);
    }

    path = avl_rotate_3(path_top, first, third);
    avl_rebalance_path(tree, path, data);
}

/**
 Insert an item into a tree.

 @param tree The tree to insert the item into.
 @param key The key for the item to insert.
 @param data The data to insert.
 @param ret Output parameter only. If non-null, holds the existing item if one already exists for the given key, or else the inserted item.

 @return Status.
 **/
OALAVLTree_InsertionStatus OALAVLTree_Insert(OALAVLTree tree, OALAVLTreeKey_t key, OALAVLTreeItem_t data, OALAVLTreeItem_t *ret)
{
    _OALAVLTreeNode *treep;
    _OALAVLTreeNode ttree;
    _OALAVLTreeNode *path_top;

    PCR_Assert(tree);

    treep = &tree->root;
    ttree = tree->root;
    path_top = &tree->root;

    /* find an appropriate place to insert data into the tree */
    while(ttree && !(tree->eq_fn(key, ttree->data, tree->context))) {
        if(!BALANCED(ttree)) path_top = treep;
        treep = &ttree->nodes[tree->gt_fn(key, ttree->data, tree->context)];
        ttree = *treep;
    }

    if(ttree) {
        if(ret) *ret = ttree->data; /* no duplicates allowed */
        return OALAVLTree_AlreadyInserted; /* no failure, but there was a dup */
    }

    ttree = DOFMem_Alloc(sizeof(_OALAVLTreeNode_t));
    if(!ttree) {
        return OALAVLTree_InsertFailure; /* error occurred */
    }

    ttree->data = data;
    ttree->nodes[0] = ttree->nodes[1] = 0;
    ttree->opaque = NEITHER;
    *treep = ttree;

    avl_rebalance(tree, path_top, key);

    ++tree->num_items;

    if(ret) *ret = ttree->data;
    return OALAVLTree_InsertSuccess; /* success! */
}

/******************************************************
 * DELETION                                           *
 *****************************************************/

static void avl_rotate_2_shrink(_OALAVLTreeNode *path_top, direction dir)
{
    _OALAVLTreeNode d, b, c;

    d = *path_top;
    b = d->nodes[!dir];
    c = b->nodes[dir];

    *path_top = b;
    b->nodes[dir] = d;
    d->nodes[!dir] = c;

    if(BALANCED(b)) {
        b->opaque = dir;
        d->opaque = !dir;
    }
    else {
        b->opaque = d->opaque = NEITHER;
    }
}

static void avl_rotate_3_shrink(_OALAVLTreeNode *path_top, direction dir)
{
    _OALAVLTreeNode b, c, d, e, f;

    f = *path_top;
    b = f->nodes[!dir];
    d = b->nodes[dir];
    c = d->nodes[!dir];
    e = d->nodes[dir];

    *path_top = d;
    d->nodes[!dir] = b;
    d->nodes[dir] = f;
    b->nodes[dir] = c;
    f->nodes[!dir] = e;

    b->opaque = f->opaque = NEITHER;

    if(d->opaque==dir) b->opaque = !dir;
    /* original code had no else here... maybe because of a possible pointer swap */
    if(d->opaque == !dir) f->opaque = dir;
    d->opaque = NEITHER;
}

static void avl_swap_del(_OALAVLTreeNode *targetp, _OALAVLTreeNode *treep, direction dir)
{
    _OALAVLTreeNode targetn = *targetp;
    _OALAVLTreeNode tree = *treep;

    *targetp = tree;
    *treep = tree->nodes[!dir];
    tree->nodes[0] = targetn->nodes[0];
    tree->nodes[1] = targetn->nodes[1];
    tree->opaque = targetn->opaque;

    DOFMem_Free(targetn);
}

static _OALAVLTreeNode *avl_rebalance_del(OALAVLTree tree, _OALAVLTreeNode *treep, const void *data, _OALAVLTreeNode *targetp)
{
    _OALAVLTreeNode targetn = *targetp;

    for(;;) {
        _OALAVLTreeNode _tree = *treep;
        direction dir = tree->gt_fn(data, _tree->data, tree->context);

        if(!_tree->nodes[dir]) break;

        if(BALANCED(_tree)) _tree->opaque = !dir;
        else if (_tree->opaque == dir) _tree->opaque = NEITHER;
        else {
            if(_tree->nodes[!dir]->opaque == dir) avl_rotate_3_shrink(treep, dir);
            else avl_rotate_2_shrink(treep, dir);
            if(_tree == targetn) targetp = &(*treep)->nodes[dir];
        }

        treep = &_tree->nodes[dir];
    }

    return targetp;
}


/**
 Remove an item from a tree.

 @param tree The tree to remove the item from.
 @param key The key for the item to remove.
 @return TRUE if something was deleted; otherwise FALSE.
 **/
boolean OALAVLTree_Delete(OALAVLTree tree, OALAVLTreeKey_t key)
{
    PCR_Assert(tree);

    {
        _OALAVLTreeNode *treep = &tree->root;
        _OALAVLTreeNode *path_top = treep;
        _OALAVLTreeNode _tree = *treep;
        _OALAVLTreeNode *targetp = NULL;

        direction dir;

        while(_tree) {
            dir = tree->gt_fn(key, _tree->data, tree->context);

            if(tree->eq_fn(key, _tree->data, tree->context)) targetp = treep;
            if(!_tree->nodes[dir]) break;
            if(BALANCED(_tree) || ((_tree->opaque == !dir) && BALANCED(_tree->nodes[!dir]))) path_top = treep;
            treep = &_tree->nodes[dir];
            _tree = *treep;
        }

        if(!targetp) {
            DOFLog_Message(  DOFLOGLEVEL_DEBUG, "OALAVLTree: Tried to delete item that isn't in tree.");
            return FALSE;
        }

        targetp = avl_rebalance_del(tree, path_top, key, targetp);

        avl_swap_del(targetp, treep, dir);
    }

    --tree->num_items;

    return TRUE;
}

typedef struct traverse_t {
    OALAVLTree_Traversal_fn_t fn;
    void *arg;
    uint8 bail;
} traverse_t;

/** @internal
 In-order traversal of the tree, calling a given function for each node.

 @param node The current node.
 @param t Pointer to a traverse_t type that contains information needed for traversal.
 **/
static void __traverse(_OALAVLTreeNode node, traverse_t *t)
{
    if(node) {
        if(!t->bail) __traverse(node->nodes[0], t);
        if(!t->bail&&t->fn(node->data, t->arg)) {
            t->bail = 1;
            return;
        }
        if(!t->bail) __traverse(node->nodes[1], t);
    }
}

/**
 Traverse a tree and call a given function for each node.

 @param tree The tree to traverse.
 @param traverse_fn The function to call for each node.
 @param arg The argument to pass to each function call.
 **/
void OALAVLTree_Traverse(OALAVLTree tree, OALAVLTree_Traversal_fn_t traverse_fn, void *arg)
{
    PCR_Assert(traverse_fn);

    {
        traverse_t t;
        t.fn = traverse_fn;
        t.arg = arg;
        t.bail = 0;

        __traverse(tree->root, &t);
    }
}

/**
 Get the root item of a tree.

 @param tree The tree from which to retrieve the root item.

 @return The root item.
 **/
OALAVLTreeItem_t OALAVLTree_get_root_item(OALAVLTree tree)
{
    return tree->root ? tree->root->data : 0;
}

/**
 @}
 **/
