/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/slist.h>
#include <internal/avl_tree.h>
#include <internal/avl_treelist.h>
#include <internal/dofinternal.h>

typedef struct OALAVLTreeList_t
{
    OALAVLTree          tree;
    uint32              refcount;
    uint32              itemCount;
    OALAVLTreeList_CompareGreaterThan_fn_t user_gt_fn;
    OALAVLTreeList_CompareEqual_fn_t       user_eq_fn;
    void*                                  user_context;
}OALAVLTreeList_t;

typedef struct treelistnode_t
{
    struct OALAVLTreeListNode_t* head;
    uint32                       count;
}treelistnode_t;

typedef struct OALAVLTreeListNode_t
{
    OALSListNode         next;
    OALAVLTreeListItem_t item;
}OALAVLTreeListNode_t;

typedef union TreeListFunctions
{
    OALAVLTreeListRemoveList_fn remove_fn;
    OALAVLTreeList_Traversal_fn_t traverse_fn;
    OALAVLTreeList_TraverseAll_fn_t traverse_all_fn;
}TreeListFunctions;

typedef struct fn_args_t
{
    TreeListFunctions fn;
    void* args;
    boolean first;
    OALAVLTreeList_Traverse_All_Return_Type_fn_t termValue;
}fn_args_t;

static boolean OALAVLTreeList_Internal_CompareGreaterThan_fn(OALAVLTreeListKey_t key, OALAVLTreeListItem_t itemx, void *contextx)
{
    treelistnode_t* item = (treelistnode_t*)itemx;
    OALAVLTreeList_t* context = (OALAVLTreeList_t*)contextx;

    if( context->user_gt_fn )
        return context->user_gt_fn(key, item->head->item, context->user_context);
    else
        return (key > (item->head->item));
}

static boolean OALAVLTreeList_Internal_CompareEqual_fn(OALAVLTreeListKey_t key, OALAVLTreeListItem_t itemx, void *contextx)
{
    treelistnode_t* item = (treelistnode_t*)itemx;
    OALAVLTreeList_t* context = (OALAVLTreeList_t*)contextx;

    if( context->user_eq_fn )
        return context->user_eq_fn(key, item->head->item, context->user_context);
    else
        return (key == item->head->item);
}

OALAVLTreeList OALAVLTreeList_Create(OALAVLTreeList_CompareGreaterThan_fn_t gt_fn, OALAVLTreeList_CompareEqual_fn_t eq_fn, void *context)
{
    OALAVLTreeList_t* retVal = (OALAVLTreeList_t*)DOFMem_Alloc(sizeof(OALAVLTreeList_t));

    if( !retVal )
        return NULL;

    retVal->tree = OALAVLTree_Create( OALAVLTreeList_Internal_CompareGreaterThan_fn, OALAVLTreeList_Internal_CompareEqual_fn, retVal);

    if( !retVal->tree )
    {
        DOFMem_Free(retVal);
        return NULL;
    }

    /* Set up fields */
    retVal->itemCount = 0;
    retVal->refcount = 1;
    retVal->user_eq_fn = eq_fn;
    retVal->user_gt_fn = gt_fn;
    retVal->user_context = context;
 
    return retVal;
}

OALAVLTreeList OALAVLTreeList_Copy(OALAVLTreeList self)
{
    PCR_Assert(self);

    if( self->refcount >=  0xFFFFFFFF )
        return NULL;

    ++(self->refcount);
    return self;
}

static void OALAVLTreeList_Internal_RemoveAllNodes(OALSListNode item, void *argx)
{
    OALAVLTreeListNode_t* listItem = (OALAVLTreeListNode_t*)item;
    fn_args_t* rm_fn = (fn_args_t*)argx;

    if( rm_fn->fn.remove_fn )
        rm_fn->fn.remove_fn(listItem->item,rm_fn->args);

    DOFMem_Free((void*)item);
}

static void OALAVLTreeList_Internal_RemoveAllLists(OALAVLTreeItem_t item, void *argx)
{
    treelistnode_t* node = (treelistnode_t*)item;
    fn_args_t* arg = (fn_args_t*)argx;

    /* traverse list and call their remove function on each item */
    arg->first = 1;
    OALSList_RemoveAll((OALSListNode*)&node->head, OALAVLTreeList_Internal_RemoveAllNodes, arg);
    DOFMem_Free((void*)item);
}

void OALAVLTreeList_Destroy(OALAVLTreeList self, OALAVLTreeListRemoveList_fn remove_fn, void *arg)
{
    if( self )
    {
        --(self->refcount);
    
        if( !self->refcount )
        {
            fn_args_t temp;
            temp.fn.remove_fn = remove_fn;
            temp.args = arg;
    
            OALAVLTree_Destroy(self->tree, OALAVLTreeList_Internal_RemoveAllLists, &temp);
            DOFMem_Free(self);
            return;
        }
    }
}

boolean OALAVLTreeList_Insert(OALAVLTreeList self, OALAVLTreeListKey_t key, OALAVLTreeListItem_t item)
{
    /* retreive item else create new list */
    treelistnode_t* itemInTree;
    OALAVLTreeListNode_t* newItem;

    itemInTree = (treelistnode_t*)OALAVLTree_Retrieve(self->tree, key);
    ++(self->itemCount);

    if( itemInTree )
    {
        newItem = (OALAVLTreeListNode_t*)DOFMem_Alloc(sizeof(OALAVLTreeListNode_t));
        if( !newItem )
            return FALSE; /* FAIL */

        newItem->item = item;
        /* add to list */
        ++(itemInTree->count);
        OALSList_Insert((OALSListNode*)&itemInTree->head, (OALSListNode)newItem);
    }
    else
    {
        itemInTree = (treelistnode_t*)DOFMem_Alloc(sizeof(treelistnode_t));
        if( itemInTree )
        {
            OALAVLTree_InsertionStatus status;

            newItem = (OALAVLTreeListNode_t*)DOFMem_Alloc(sizeof(OALAVLTreeListNode_t));
            if( !newItem )
            {
                DOFMem_Free( itemInTree );
                return FALSE; /* FAIL */
            }

            newItem->item = item;
            itemInTree->count = 1;
            itemInTree->head = NULL;
            OALSList_Insert((OALSListNode*)&itemInTree->head, (OALSListNode)newItem);
            status = OALAVLTree_Insert(self->tree,key,(OALAVLTreeItem_t)itemInTree, 0);
            if(status != OALAVLTree_InsertSuccess)
            {
                DOFMem_Free( itemInTree );
                DOFMem_Free( newItem );
                return FALSE; /* FAIL */
            }
        }
    }
    return TRUE;
}

static void OALAVLTreeList_RemoveList_Function(OALSListNode item, void *arg)
{
    OALAVLTreeListNode_t* treeListItem = (OALAVLTreeListNode_t*)item;
    fn_args_t* temp = (fn_args_t*)arg;

    /* clean up item */
    if( temp->fn.remove_fn )
        temp->fn.remove_fn(treeListItem->item, temp->args);

    DOFMem_Free((void*)item);
}


void OALAVLTreeList_DeleteList(OALAVLTreeList self, OALAVLTreeListKey_t key, OALAVLTreeListRemoveList_fn remove_fn, void *arg)
{
    treelistnode_t* itemInTree = (treelistnode_t*)OALAVLTree_Retrieve(self->tree, (OALAVLTreeKey_t) key);

    if( itemInTree )
    {
        fn_args_t temp;

        OALAVLTree_Delete(self->tree, key);

        temp.fn.remove_fn = remove_fn;
        temp.args = arg;

        self->itemCount -= itemInTree->count;

        OALSList_RemoveAll((OALSListNode*)&itemInTree->head, OALAVLTreeList_RemoveList_Function, &temp);
        DOFMem_Free(itemInTree);
    }
}

static boolean OALAVLTreeList_Internal_TreeListTraverse_fn(OALSListNode item, void *arg)
{
    OALAVLTreeListNode_t* treeListItem = (OALAVLTreeListNode_t*) item;
    fn_args_t* temp = arg;

    if( temp->fn.traverse_fn(treeListItem->item, temp->args) )
        return TRUE;
    return FALSE;
}

void OALAVLTreeList_TraverseList(OALAVLTreeList tree, OALAVLTreeListKey_t key, OALAVLTreeList_Traversal_fn_t traverse_fn, void *arg)
{
    treelistnode_t* itemInTree;

    itemInTree = (treelistnode_t*)OALAVLTree_Retrieve(tree->tree, (OALAVLTreeKey_t) key);

    if( itemInTree )
    {
        fn_args_t temp;
        temp.fn.traverse_fn = traverse_fn;
        temp.args = arg;

        /* traverse List */
        OALSList_Traverse((OALSListNode)itemInTree->head, OALAVLTreeList_Internal_TreeListTraverse_fn, &temp);
    }

    return;
}

static boolean OALAVLTreeList_Internal_AllTraverse(OALSListNode item, void *arg)
{
    OALAVLTreeListNode_t* treeListItem = (OALAVLTreeListNode_t*)item;
    fn_args_t* temp = arg;

    temp->termValue = temp->fn.traverse_all_fn(treeListItem->item, temp->args, temp->first);

    if( temp->termValue != OALAVLTREELIST_NORMAL )
        return TRUE;

    temp->first = FALSE;
    return FALSE;
}

static boolean OALAVLTreeList_Internal_TraversalAll(OALAVLTreeItem_t item, void *arg)
{
    treelistnode_t* treeListItem = (treelistnode_t*)item;
    fn_args_t* temp = arg;
    temp->first = TRUE;

    OALSList_Traverse((OALSListNode)treeListItem->head, OALAVLTreeList_Internal_AllTraverse, arg);

    if( temp->termValue == OALAVLTREELIST_TERMINATE_TRAVERSE )
        return TRUE;

    return FALSE;
}

void OALAVLTreeList_TraverseAll(OALAVLTreeList tree, OALAVLTreeList_TraverseAll_fn_t traverse_fn, void *arg)
{
    /* traverse tree and traverse lists*/
    fn_args_t temp;
    temp.fn.traverse_all_fn = traverse_fn;
    temp.args = arg;
    temp.termValue = OALAVLTREELIST_NORMAL;

    OALAVLTree_Traverse(tree->tree, OALAVLTreeList_Internal_TraversalAll, &temp);

    return;
}

uint32 OALAVLTreeList_GetNumLists(OALAVLTreeList tree)
{
    return OALAVLTree_GetNumItems(tree->tree);
}

uint32 OALAVLTreeList_GetNumItems(OALAVLTreeList self)
{
    return self->itemCount;
}

boolean OALAVLTreeList_IsKeyInTree(OALAVLTreeList self, OALAVLTreeListKey_t key)
{
    if(OALAVLTree_Retrieve(self->tree, key))
        return TRUE;

    return FALSE;
}

uint32 OALAVLTreeList_GetNumItemsInList(OALAVLTreeList self, OALAVLTreeListKey_t key )
{
    treelistnode_t* temp;

    PCR_Assert(self);
    temp = (treelistnode_t*)OALAVLTree_Retrieve(self->tree, key);

    return ((temp)? temp->count: 0);
}

static boolean OALAVLTreeList_Internal_TreeListFind_fn(OALSListNode item, void *arg)
{
    OALAVLTreeListNode_t* treeListItem = (OALAVLTreeListNode_t*)item;
    fn_args_t* temp = (fn_args_t*)arg;

    return temp->fn.traverse_fn(treeListItem->item, temp->args);
}

/**
 * removes an item a tree lists
 *
 * @note if multiple items in list need to be removed you must call this multiple times
 *
 * @param tree  your tree to remove from
 * @param key   the key to get to your list
 * @param fn    the function to call on the item returns true if it is the item to remove
 * @param arg   any context your fn needs
 *
 * @return OALAVLTreeListItem_t the item removed
 */
OALAVLTreeListItem_t OALAVLTreeList_FindAndRemove(OALAVLTreeList tree, OALAVLTreeListKey_t key, OALAVLTreeList_Traversal_fn_t fn, void *arg)
{
    treelistnode_t* itemInTree;
    OALAVLTreeListItem_t ret = NULL;
    OALAVLTreeListNode_t* node = NULL;

    if( !tree || !fn )
        return NULL;

    itemInTree = (treelistnode_t*)OALAVLTree_Retrieve(tree->tree, (OALAVLTreeKey_t) key);
    if( itemInTree )
    {
        fn_args_t temp;
        temp.fn.traverse_fn = fn;
        temp.args = arg;

        node = (OALAVLTreeListNode_t*)OALSList_Retrieve((OALSListNode)itemInTree->head, OALAVLTreeList_Internal_TreeListFind_fn, &temp);
        if( node )
        {
            ret = node->item;

            --itemInTree->count;
            --tree->itemCount;

            if( !itemInTree->count )
            {
                OALAVLTree_Delete(tree->tree, key);
                DOFMem_Free(itemInTree);
            }
            else
            {
                OALSList_Remove( (OALSListNode *)&itemInTree->head, (OALSListNode)node );
            }

            DOFMem_Free(node);
        }
    }

    return ret;
}

