/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/bignum.h>
#ifdef BN_MP_EXTEUCLID_C

/* Extended euclidean algorithm of (a, b) produces 
   a*u1 + b*u2 = u3
 */
int32 mp_exteuclid(mp_int *a, mp_int *b, mp_int *U1, mp_int *U2, mp_int *U3)
{
   mp_int u1,u2,u3,v1,v2,v3,t1,t2,t3,q,tmp;
   int32 err;

#ifdef OALVARARGS_ALLOWED
   if ((err = mp_init_multi(&u1, &u2, &u3, &v1, &v2, &v3, &t1, &t2, &t3, &q, &tmp, NULL)) != MP_OKAY) {
      return err;
   }
#else
   if ((err = mp_init(&u1)) != MP_OKAY) {
      goto _ERR_u1;
   }

   if ((err = mp_init(&u2)) != MP_OKAY) {
      goto _ERR_u2;
   }

   if ((err = mp_init(&u3)) != MP_OKAY) {
      goto _ERR_u3;
   }

   if ((err = mp_init(&v1)) != MP_OKAY) {
      goto _ERR_v1;
   }

   if ((err = mp_init(&v2)) != MP_OKAY) {
      goto _ERR_v2;
   }

   if ((err = mp_init(&v3)) != MP_OKAY) {
      goto _ERR_v3;
   }

   if ((err = mp_init(&t1)) != MP_OKAY) {
      goto _ERR_t1;
   }

   if ((err = mp_init(&t2)) != MP_OKAY) {
      goto _ERR_t2;
   }

   if ((err = mp_init(&t3)) != MP_OKAY) {
      goto _ERR_t3;
   }

   if ((err = mp_init(&q)) != MP_OKAY) {
      goto _ERR_q;
   }

   if ((err = mp_init(&tmp)) != MP_OKAY) {
      goto _ERR_tmp;
   }
#endif

   /* initialize, (u1,u2,u3) = (1,0,a) */
   mp_set(&u1, 1);
   if ((err = mp_copy(a, &u3)) != MP_OKAY)                                        { goto _ERR; }

   /* initialize, (v1,v2,v3) = (0,1,b) */
   mp_set(&v2, 1);
   if ((err = mp_copy(b, &v3)) != MP_OKAY)                                        { goto _ERR; }

   /* loop while v3 != 0 */
   while (mp_iszero(&v3) == MP_NO) {
       /* q = u3/v3 */
       if ((err = mp_div(&u3, &v3, &q, NULL)) != MP_OKAY)                         { goto _ERR; }

       /* (t1,t2,t3) = (u1,u2,u3) - (v1,v2,v3)q */
       if ((err = mp_mul(&v1, &q, &tmp)) != MP_OKAY)                              { goto _ERR; }
       if ((err = mp_sub(&u1, &tmp, &t1)) != MP_OKAY)                             { goto _ERR; }
       if ((err = mp_mul(&v2, &q, &tmp)) != MP_OKAY)                              { goto _ERR; }
       if ((err = mp_sub(&u2, &tmp, &t2)) != MP_OKAY)                             { goto _ERR; }
       if ((err = mp_mul(&v3, &q, &tmp)) != MP_OKAY)                              { goto _ERR; }
       if ((err = mp_sub(&u3, &tmp, &t3)) != MP_OKAY)                             { goto _ERR; }

       /* (u1,u2,u3) = (v1,v2,v3) */
       if ((err = mp_copy(&v1, &u1)) != MP_OKAY)                                  { goto _ERR; }
       if ((err = mp_copy(&v2, &u2)) != MP_OKAY)                                  { goto _ERR; }
       if ((err = mp_copy(&v3, &u3)) != MP_OKAY)                                  { goto _ERR; }

       /* (v1,v2,v3) = (t1,t2,t3) */
       if ((err = mp_copy(&t1, &v1)) != MP_OKAY)                                  { goto _ERR; }
       if ((err = mp_copy(&t2, &v2)) != MP_OKAY)                                  { goto _ERR; }
       if ((err = mp_copy(&t3, &v3)) != MP_OKAY)                                  { goto _ERR; }
   }

   /* make sure U3 >= 0 */
   if (u3.sign == MP_NEG) {
      mp_neg(&u1, &u1);
      mp_neg(&u2, &u2);
      mp_neg(&u3, &u3);
   }

   /* copy result out */
   if (U1 != NULL) { mp_exch(U1, &u1); }
   if (U2 != NULL) { mp_exch(U2, &u2); }
   if (U3 != NULL) { mp_exch(U3, &u3); }

   err = MP_OKAY;
_ERR: 
#ifdef OALVARARGS_ALLOWED
   mp_clear_multi(&u1, &u2, &u3, &v1, &v2, &v3, &t1, &t2, &t3, &q, &tmp, NULL);
#else
   mp_clear(&tmp);
_ERR_tmp:
   mp_clear(&q);
_ERR_q:
   mp_clear(&t3);
_ERR_t3:
   mp_clear(&t2);
_ERR_t2:
   mp_clear(&t1);
_ERR_t1:
   mp_clear(&v3);
_ERR_v3:
   mp_clear(&v2);
_ERR_v2:
   mp_clear(&v1);
_ERR_v1:
   mp_clear(&u3);
_ERR_u3:
   mp_clear(&u2);
_ERR_u2:
   mp_clear(&u1);
_ERR_u1:
#endif
   return err;
}
#endif




