/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/bridge.h>
#include <dof/credentials.h>
#include <dof/objectid.h>
#include <internal/bridgeinternal.h>
#include <internal/permissionsetinternal.h>

#define BRIDGE_SET_CREDENTIALS_ERROR 0x01
#define BRIDGE_SET_PERMISSIONS_ERROR 0x02
#define BRIDGE_SET_IDENTITY_ERROR    0x04

typedef struct DOFOperationBridgeConfigBuilder_t
{
    DOFCredentials credentials;
    DOFPermissionSet permissionSet;
    DOFObjectIDAuthentication id;
    uint32 errorFlags;
    uint32 refCount;
    boolean isExtendAllowed;
} DOFOperationBridgeConfigBuilder_t;

extern PCRMutex ref_count_mutex;

static void doErrorStuff(uint32  * errors,
                         uint32    error,
                         boolean   worked);

DOFOperationBridgeConfigBuilder DOFOperationBridgeConfigBuilder_Create(void)
{
    DOFOperationBridgeConfigBuilder builder = DOFMem_Alloc(sizeof(DOFOperationBridgeConfigBuilder_t));

    if(builder != NULL)
    {
        builder->permissionSet = NULL;
        builder->credentials = NULL;
        builder->id = NULL;
        builder->errorFlags = 0;
        builder->isExtendAllowed = TRUE;
        builder->refCount = 1;
        return builder;
    }
    return NULL;
}

DOFOperationBridgeConfigBuilder DOFOperationBridgeConfigBuilder_Create_Config( DOFOperationBridgeConfig self )
{
    if(self != NULL)
    {
        DOFOperationBridgeConfigBuilder ret = (DOFOperationBridgeConfigBuilder) DOFMem_Alloc(sizeof(DOFOperationBridgeConfigBuilder_t));

        if(ret != NULL)
        {
            DOFOperationBridgeConfigBuilder selfx = (DOFOperationBridgeConfigBuilder) self;

            ret->permissionSet = DOFPermissionSet_DeepCopy(selfx->permissionSet);
            if((ret->permissionSet != NULL) || (selfx->permissionSet == NULL))
            {
                ret->credentials = DOFCredentials_Copy(selfx->credentials);
                if((ret->credentials != NULL) || (selfx->credentials == NULL))
                {
                    ret->id = DOFObjectIDAuthentication_Copy(selfx->id);
                    if((ret->id != NULL) || (selfx->id == NULL))
                    {
                        ret->errorFlags = selfx->errorFlags;
                        ret->isExtendAllowed = selfx->isExtendAllowed;
                        ret->refCount = 1;
                        return ret;
                    }
                    DOFCredentials_Destroy(ret->credentials);
                }
                DOFPermissionSet_Destroy(ret->permissionSet);
            }
            DOFMem_Free(ret);
        }
    }
    return NULL;
}

DOFOperationBridgeConfigBuilder DOFOperationBridgeConfigBuilder_Copy(DOFOperationBridgeConfigBuilder self)
{
    if(self != NULL)
    {
        DOFOperationBridgeConfigBuilder builder = DOFMem_Alloc(sizeof(DOFOperationBridgeConfigBuilder_t));

        if(builder != NULL)
        {
            builder->credentials = DOFCredentials_Copy(self->credentials);
            if((builder->credentials != NULL) || (self->credentials == NULL))
            {
                builder->permissionSet = DOFPermissionSet_Copy(self->permissionSet);
                if((builder->permissionSet != NULL) || (self->permissionSet == NULL))
                {
                    builder->id = DOFObjectIDAuthentication_Copy(self->id);
                    if((builder->id != NULL) || (self->id == NULL))
                    {
                        builder->isExtendAllowed = self->isExtendAllowed;
                        builder->errorFlags = 0;
                        builder->refCount = 1;
                        if(!builder->isExtendAllowed)
                        {
                            DOFPermissionSet_SetGrantPerms(builder->permissionSet);
                        }
                        return builder;
                    }
                    DOFPermissionSet_Destroy(builder->permissionSet);
                }
                DOFCredentials_Destroy(builder->credentials);
            }
            DOFMem_Free(builder);
        }
    }
    return NULL;
}

void DOFOperationBridgeConfigBuilder_Destroy(DOFOperationBridgeConfigBuilder self)
{
    if(self != NULL)
    {
        boolean canFree;

        PCRMutex_Lock(ref_count_mutex);
        canFree = (--(self->refCount) == 0);
        PCRMutex_Unlock(ref_count_mutex);
        if(canFree)
        {
            DOFCredentials_Destroy(self->credentials);
            DOFPermissionSet_Destroy(self->permissionSet);
            DOFMem_Free(self);
        }
    }
}

DOFOperationBridgeConfig DOFOperationBridgeConfig_Copy(DOFOperationBridgeConfig self)
{
    if(self != NULL)
    {
        DOFOperationBridgeConfigBuilder selfx = (DOFOperationBridgeConfigBuilder) self;

        PCRMutex_Lock(ref_count_mutex);
        ++(selfx->refCount);
        PCRMutex_Unlock(ref_count_mutex);
    }
    return self;
}

void DOFOperationBridgeConfig_Destroy(DOFOperationBridgeConfig self)
{
    DOFOperationBridgeConfigBuilder_Destroy((DOFOperationBridgeConfigBuilder) self);
}

DOFOperationBridgeConfig DOFOperationBridgeConfigBuilder_Build(DOFOperationBridgeConfigBuilder self)
{
    if((self != NULL) &&
       (self->errorFlags == 0))
    {
        return (DOFOperationBridgeConfig) DOFOperationBridgeConfigBuilder_Copy(self);
    }
    return NULL;
}

DOFOperationBridgeConfig DOFOperationBridgeConfigBuilder_BuildAndDestroy(DOFOperationBridgeConfigBuilder self)
{
    if((self != NULL) &&
       (self->errorFlags == 0))
    {
        if(!self->isExtendAllowed)
        {
            DOFPermissionSet_SetGrantPerms(self->permissionSet);
        }
        return (DOFOperationBridgeConfig) self;
    }
    DOFOperationBridgeConfigBuilder_Destroy(self);
    return NULL;
}

boolean DOFOperationBridgeConfigBuilder_SetCredentials(DOFOperationBridgeConfigBuilder self,
                                                 DOFCredentials                  credentials)
{
    boolean retVal = FALSE;

    if(self != NULL)
    {
        DOFCredentials newCreds = DOFCredentials_Copy(credentials);

        retVal = ((newCreds != NULL) || (credentials == NULL));
        DOFCredentials_Destroy(self->credentials);
        self->credentials = newCreds;
        doErrorStuff(&(self->errorFlags), BRIDGE_SET_CREDENTIALS_ERROR, retVal);
    }
    return retVal;
}

boolean DOFOperationBridgeConfigBuilder_SetPermissions(DOFOperationBridgeConfigBuilder self,
                                                 DOFPermissionSet                permissions)
{
    boolean retVal = FALSE;

    if(self != NULL)
    {
        DOFPermissionSet newPerms = DOFPermissionSet_Copy(permissions);

        retVal = ((newPerms != NULL) || (permissions == NULL));
        DOFPermissionSet_Destroy(self->permissionSet);
        self->permissionSet = newPerms;
        doErrorStuff(&(self->errorFlags), BRIDGE_SET_PERMISSIONS_ERROR, retVal);
    }
    return retVal;
}

boolean DOFOperationBridgeConfigBuilder_SetPermissionsExtendAllowed(DOFOperationBridgeConfigBuilder self,
                                                             boolean                         extendAllowed)
{
    if(self != NULL)
    {
        self->isExtendAllowed = extendAllowed;
        return TRUE;
    }
    return FALSE;
}

DOFCredentials DOFOperationBridgeConfig_GetCredentials(DOFOperationBridgeConfig self)
{
    PCR_Assert(self != NULL);
    return ((DOFOperationBridgeConfigBuilder) self)->credentials;
}

DOFObjectIDAuthentication DOFOperationBridgeConfig_GetAlternateIdentity(DOFOperationBridgeConfig self)
{
    PCR_Assert(self != NULL);
    return ((DOFOperationBridgeConfigBuilder) self)->id;
}

DOFPermissionSet DOFOperationBridgeConfig_GetPermissions(DOFOperationBridgeConfig self)
{
    PCR_Assert(self != NULL);
    return ((DOFOperationBridgeConfigBuilder) self)->permissionSet;
}

boolean DOFOperationBridgeConfig_IsPermissionsExtendAllowed(DOFOperationBridgeConfig self)
{
    PCR_Assert(self != NULL);
    return ((DOFOperationBridgeConfigBuilder) self)->isExtendAllowed;
}

static void doErrorStuff(uint32  * errors,
                         uint32    error,
                         boolean   worked)
{
    if(worked)
    {
        *errors &= ~error;
    }
    else
    {
        *errors |= error;
    }
}
