/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/bridge.h>
#include <internal/bridgeinternal.h>
#include <internal/dofinternal.h>
#include <internal/permissionsetinternal.h>

#include "bridge.h"

typedef struct NonSecureBridge_t
{
    OALOperationBridge_t super;
    DOF dof;
    DOFPermissionSet permissionSet;
    DOFPermissionSet complimentSet;
} NonSecureBridge_t;

typedef NonSecureBridge_t * NonSecureBridge;

static void NonSecureBridge_Destroy(DOFSecurityMembrane self);

static DOFSecurityScope NonSecureBridge_ValidateInboundPermissions(DOFSecurityMembrane self,
                                                                   DOFPermission       explicitPermission,
                                                                   DOFPermission       logicalPermission,
                                                                   OALCipherKeyState   keyState);
static DOFSecurityScope NonSecureBridge_ValidateOutboundPermissions(DOFSecurityMembrane self,
                                                                    DOFPermission       explicitPermission,
                                                                    DOFPermission       logicalPermission);

static DOFSecurityScope NonSecureBridge_GetSecurityScope(DOFSecurityMembrane self);

static DOFSecurityMembraneExtendReturn_t NonSecureBridge_ExtendPermissions(DOFSecurityMembrane   self,
                                                                           DOFPermissionSet      extendPerms,
                                                                           ExtendCompleteCB      callback,
                                                                           void                * context);

static DOFObjectIDDomain NonSecureBridge_GetDomain(DOFSecurityMembrane self);

static DOFPermissionSet NonSecureBridge_GetInboundPerms(DOFSecurityMembrane self);

static DOFPermissionSet NonSecureBridge_GetOutboundPerms(DOFSecurityMembrane self);

static struct DOFSecurityMembrane_VTable_t vtable = {
    NonSecureBridge_Destroy,                    NonSecureBridge_GetDomain,
    NonSecureBridge_ValidateInboundPermissions, NonSecureBridge_ValidateOutboundPermissions,
    NonSecureBridge_GetSecurityScope,           NonSecureBridge_GetSecurityScope,
    NonSecureBridge_ExtendPermissions,          NonSecureBridge_ExtendPermissions,
    NonSecureBridge_GetInboundPerms,            NonSecureBridge_GetOutboundPerms
};

OALOperationBridge OALOperationBridge_CreateNonSecureBridge(DOF                      dof,
                                                            DOFOperationBridgeConfig config)
{
    NonSecureBridge bridge = (NonSecureBridge) DOFMem_Alloc(sizeof(NonSecureBridge_t));

    if(bridge != NULL)
    {
        bridge->dof = dof;
        if((config == NULL) || (DOFOperationBridgeConfig_IsPermissionsExtendAllowed(config)))
        {
            bridge->permissionSet = bridge->complimentSet = NULL;
        }
        else
        {
            DOFPermissionSet bridgePerms = DOFOperationBridgeConfig_GetPermissions(config);

            bridge->permissionSet = (bridgePerms != NULL) ? DOFPermissionSet_Copy(bridgePerms)
                                                          : DOFPermissionSet_DeepCopy(DOFPermissionSet_GetEmptyPermissionSet());
            if(bridge->permissionSet == NULL)
            {
                DOFMem_Free(bridge);
                return NULL;
            }
            else
            {
                bridge->complimentSet = DOFPermissionSet_CreateComplement(bridge->permissionSet);
                if(bridge->complimentSet == NULL)
                {
                    DOFPermissionSet_Destroy(bridge->permissionSet);
                    DOFMem_Free(bridge);
                    return NULL;
                }
            }
        }
        OALOperationBridge_Ctor(&(bridge->super), &vtable);
    }
    return (OALOperationBridge) bridge;
}

static void NonSecureBridge_Destroy(DOFSecurityMembrane self)
{
    NonSecureBridge bridge = (NonSecureBridge) self;

    DOFPermissionSet_Destroy(bridge->permissionSet);
    DOFPermissionSet_Destroy(bridge->complimentSet);
    DOFMem_Free(self);
}

static DOFSecurityScope NonSecureBridge_ValidatePermission(NonSecureBridge  bridge,
                                                           DOFPermissionSet permissionSet,
                                                           DOFPermission    permission)
{
    if(permissionSet == NULL)
    {
        return DOFSecurityScope_Copy(DOF_GetDefaultSecurityScope(bridge->dof));
    }
    else if(DOFPermissionSet_Contains(permissionSet, permission))
    {
        return DOFSecurityScope_Copy(DOF_GetDefaultSecurityScope(bridge->dof));
    }
    return DOFSecurityScope_Copy(DOF_GetEmptySecurityScope(bridge->dof));
}

static DOFSecurityScope NonSecureBridge_ValidateInboundPermissions(DOFSecurityMembrane self,
                                                                   DOFPermission       explicitPermission,
                                                                   DOFPermission       logicalPermission,
                                                                   OALCipherKeyState   keyState)
{
    NonSecureBridge bridge = (NonSecureBridge) self;

    return NonSecureBridge_ValidatePermission(bridge, bridge->permissionSet, explicitPermission);
}

static DOFSecurityScope NonSecureBridge_ValidateOutboundPermissions(DOFSecurityMembrane self,
                                                                    DOFPermission       explicitPermission,
                                                                    DOFPermission       logicalPermission)
{
    NonSecureBridge bridge = (NonSecureBridge) self;

    return NonSecureBridge_ValidatePermission(bridge, bridge->complimentSet, explicitPermission);
}

static DOFSecurityScope NonSecureBridge_GetSecurityScope(DOFSecurityMembrane self)
{
    return DOF_GetDefaultSecurityScope(((NonSecureBridge) self)->dof);
}

static DOFSecurityMembraneExtendReturn_t NonSecureBridge_ExtendPermissions(DOFSecurityMembrane   self,
                                                                           DOFPermissionSet      extendPerms,
                                                                           ExtendCompleteCB      callback,
                                                                           void                * context)
{
    PCR_Assert(FALSE);          /* we should never actually call this */
    return CANNOT_EXTEND;
}

DOFObjectIDDomain NonSecureBridge_GetDomain(DOFSecurityMembrane self)
{
    return DOFSecurityScope_GetDomainID(NonSecureBridge_GetSecurityScope(self));
}

static DOFPermissionSet NonSecureBridge_GetInboundPerms(DOFSecurityMembrane self)
{
    return ((NonSecureBridge) self)->permissionSet;
}

static DOFPermissionSet NonSecureBridge_GetOutboundPerms(DOFSecurityMembrane self)
{
    return ((NonSecureBridge) self)->complimentSet;
}
