/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/dofinternal.h>
#include <internal/cipher.h>
#include <internal/bridgeinternal.h>

#include "bridge.h"

static void NullConnector_dtor(DOFSecurityMembrane self);

static DOFSecurityScope NullConnector_validateInboundPerm(DOFSecurityMembrane self,
                                                          DOFPermission       explicitPermission,
                                                          DOFPermission       logicalPermission,
                                                          OALCipherKeyState   keyState);

static DOFSecurityScope NullConnector_validateOutboundPerm(DOFSecurityMembrane self,
                                                           DOFPermission       explicitPermission,
                                                           DOFPermission       logicalPermission);

static DOFSecurityScope NullConnector_returnScope(DOFSecurityMembrane self);

static DOFSecurityMembraneExtendReturn_t NullConnector_extendPerm(DOFSecurityMembrane   self,
                                                                  DOFPermissionSet      extendPerm,
                                                                  ExtendCompleteCB      callback,
                                                                  void                * context);

static DOFObjectIDDomain NullConnector_getDomainId(DOFSecurityMembrane self);

static DOFPermissionSet NullConnector_getInboundPerms(DOFSecurityMembrane self);

static DOFPermissionSet NullConnector_getOutboundPerms(DOFSecurityMembrane self);

typedef struct PassThroughConnector_t
{
    OALOperationBridge_t super;

    OALChannel             channel;
} PassThroughConnector_t;

typedef PassThroughConnector_t * PassThroughConnector;

static struct DOFSecurityMembrane_VTable_t nullVtable =
{
    NullConnector_dtor,                 NullConnector_getDomainId,
    NullConnector_validateInboundPerm,  NullConnector_validateOutboundPerm,
    NullConnector_returnScope,          NullConnector_returnScope,
    NullConnector_extendPerm,           NullConnector_extendPerm,
    NullConnector_getInboundPerms,      NullConnector_getOutboundPerms
};

OALOperationBridge OALOperationBridge_CreatePassThroughBridge(OALChannel channel)
{
    PassThroughConnector ret = (PassThroughConnector) DOFMem_Alloc(sizeof(PassThroughConnector_t));

    if(ret != NULL)
    {
        OALOperationBridge_Ctor(&(ret->super), &nullVtable);
        ret->channel = channel;
    }
    return (OALOperationBridge) ret;
}

static void NullConnector_dtor(DOFSecurityMembrane self)
{
    DOFMem_Free(self);
}

static DOFSecurityScope NullConnector_validateInboundPerm(DOFSecurityMembrane self,
                                                          DOFPermission       explicitPermission,
                                                          DOFPermission       logicalPermission,
                                                          OALCipherKeyState   keyState)
{
    PassThroughConnector realConnector = (PassThroughConnector) self;

    PCR_Assert(self != NULL);
    PCR_Assert( keyState != NULL );
    return DOFSecurityMembrane_ValidatePermission( DOFSecurityScope_GetDomainID( realConnector->channel->securityScope),
                                                   OALCipher_GetRemotePerms( realConnector->channel->remoteChannelData->cipher, keyState ),
                                                   logicalPermission );
}

static DOFSecurityScope NullConnector_validateOutboundPerm(DOFSecurityMembrane self,
                                                           DOFPermission       explicitPermission,
                                                           DOFPermission       logicalPermission)
{
    PassThroughConnector realConnector = (PassThroughConnector) self;

    PCR_Assert(self != NULL);
    return DOFSecurityMembrane_ValidatePermission( DOFSecurityScope_GetDomainID( realConnector->channel->securityScope),
                                                   OALCipher_GetLocalPerms( realConnector->channel->remoteChannelData->cipher,
                                                                            OALCipher_GetCurrentKeyState( realConnector->channel->remoteChannelData->cipher ) ),
                                                   explicitPermission );
}

static DOFSecurityScope NullConnector_returnScope(DOFSecurityMembrane self)
{
    PassThroughConnector realConnector = (PassThroughConnector) self;

    PCR_Assert(self != NULL);
    return realConnector->channel->securityScope;
}

static DOFSecurityMembraneExtendReturn_t NullConnector_extendPerm(DOFSecurityMembrane   self,
                                                                  DOFPermissionSet      extendPerm,
                                                                  ExtendCompleteCB      callback,
                                                                  void                * context)
{
    return EXTEND_ELSEWHERE;
}

static DOFObjectIDDomain NullConnector_getDomainId(DOFSecurityMembrane self)
{
    PassThroughConnector realConnector = (PassThroughConnector) self;

    PCR_Assert(self != NULL);
    return DOFSecurityScope_GetDomainID(realConnector->channel->securityScope);
}

static DOFPermissionSet NullConnector_getInboundPerms(DOFSecurityMembrane self)
{
    return OALCipher_GetRemotePerms(((PassThroughConnector)self)->channel->remoteChannelData->cipher, NULL);
}

static DOFPermissionSet NullConnector_getOutboundPerms(DOFSecurityMembrane self)
{
    return OALCipher_GetLocalPerms(((PassThroughConnector)self)->channel->remoteChannelData->cipher, NULL);
}
