/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/objectid.h>
#include <internal/dofinternal.h>
#include <internal/error.h>
#include <internal/mutex_pool.h>
#include <internal/permissionsetinternal.h>
#include <internal/security_common.h>
#include <internal/securityscope.h>
#include <internal/trp_v1.h>

#include "bridge.h"

#define DEFAULT_TIMEOUT             30000

typedef struct NormalSecurityConnector_t
{
    OALOperationBridge_t super;

    DOFCredentials credentials;
    DOFObjectIDAuthentication id;

    DOFPermissionSet originalInboundPermissions;
    DOFPermissionSet originalOutboundPermissions;
    DOFPermissionSet inboundPermissions;
    DOFPermissionSet outboundPermissions;

    DOFSecurityScope inboundScope;
    DOFSecurityScope outboundScope;

    OALTimerItem inboundTimerItem;
    OALTimerItem outboundTimerItem;

    uint32 inboundPermissionIndex;
    uint32 outboundPermissionIndex;

    boolean extendablePermissions;
} NormalSecurityConnector_t;

typedef NormalSecurityConnector_t * NormalSecurityConnector;

static void SecureConnector_normalDtor(DOFSecurityMembrane self);

static DOFSecurityScope SecureConnector_validateInboundPerms(DOFSecurityMembrane self,
                                                             DOFPermission       explicitPermission,
                                                             DOFPermission       logicalPermission,
                                                             OALCipherKeyState   keyState);

static DOFSecurityScope SecureConnector_validateOutboundPerms(DOFSecurityMembrane self,
                                                              DOFPermission       explicitPermission,
                                                              DOFPermission       logicalPermission);

static DOFSecurityScope SecureConnector_getInboundScope(DOFSecurityMembrane self);

static DOFSecurityScope SecureConnector_getOutboundScope(DOFSecurityMembrane self);

static DOFObjectIDDomain SecureConnector_getDomainId(DOFSecurityMembrane self);

static DOFSecurityMembraneExtendReturn_t SecureConnector_extendInboundPerms(DOFSecurityMembrane   self,
                                                                            DOFPermissionSet      extendPerm,
                                                                            ExtendCompleteCB      callback,
                                                                            void                * context);

static DOFSecurityMembraneExtendReturn_t SecureConnector_extendOutboundPerms(DOFSecurityMembrane   self,
                                                                             DOFPermissionSet      extendPerm,
                                                                             ExtendCompleteCB      callback,
                                                                             void                * context);

static DOFPermissionSet SecureConnector_getInboundPerms(DOFSecurityMembrane self);

static DOFPermissionSet SecureConnector_getOutboundPerms(DOFSecurityMembrane self);

static struct DOFSecurityMembrane_VTable_t normalVtable =
{
    SecureConnector_normalDtor,             SecureConnector_getDomainId,
    SecureConnector_validateInboundPerms,   SecureConnector_validateOutboundPerms,
    SecureConnector_getInboundScope,        SecureConnector_getOutboundScope,
    SecureConnector_extendInboundPerms,     SecureConnector_extendOutboundPerms,
    SecureConnector_getInboundPerms,        SecureConnector_getOutboundPerms
};

static OALTimer secureConnectorTimer;

boolean DOFOperationBridgeConfig_SecureInit(void)
{
    secureConnectorTimer = OALTimer_Create(DOF_GetDof()->threadPool, FALSE);

    return (secureConnectorTimer != NULL);
}

void DOFOperationBridgeConfig_SecureShutdown(void)
{
    OALTimer_Destroy(secureConnectorTimer);
}

static uint32 DOFSecureConnector_ConvertDurationToMilliseconds(uint8 duration)
{
    uint32 retVal = (duration * 8        /* minutes */
                              * 60       /* seconds */
                              * 1000);   /* milliseconds */

    if(retVal > 0)
    {
        retVal -= (1000 * 60 * 2);  /* Give ourselves a two minute grace period */
    }
    return retVal;
}

typedef struct DOFSecureConnector_RedoPermissionState_t
{
    NormalSecurityConnector secConn;

    DOFPermissionSet * permissionSet;
    DOFPermissionSet * originalPermissionSet;
    OALTimerItem * timerItem;
    DOFSecurityScope * scope;
    uint32 * permissionIndex;
} DOFSecureConnector_RedoPermissionState_t;

static void trpRedoPerms(TRP_V1_Context   trp_context,
                         void           * context)
{
    DOFSecureConnector_RedoPermissionState_t * state;

    PCR_Assert(context != NULL);
    state = (DOFSecureConnector_RedoPermissionState_t *) context;
    if(trp_context != NULL)
    {
        if(trp_context->ScopeReturn->status == PCRSTATUS_OK)
        {
            if(trp_context->ScopeReturn->errorCode == 0)
            {
                DOFPermissionSet newPermissions = DOFPermissionSet_DeepCopy(trp_context->ScopeReturn->permValidation);

                if(newPermissions != NULL)
                {
                    DOFSecurityScope newScope = DOFSecurityScope_Copy(trp_context->ScopeReturn->securityScope);

                    if(newScope != NULL)
                    {
                        if(OALTimerItem_Insert(*(state->timerItem), DOFSecureConnector_ConvertDurationToMilliseconds(trp_context->ScopeReturn->grantedDuration), FALSE, FALSE) == DOFERROR_OK)
                        {
                            OALMutexPool_Lock(&(state->secConn->super));
                            DOFPermissionSet_Destroy(*(state->permissionSet));
                            DOFSecurityScope_Destroy(*(state->scope));
                            *(state->permissionSet) = newPermissions;
                            *(state->scope) = newScope;
                            ++(*(state->permissionIndex));
                            OALMutexPool_Unlock(&(state->secConn->super));
                            return;
                        }
                        DOFSecurityScope_Destroy(newScope);
                    }
                    DOFPermissionSet_Destroy(newPermissions);
                }
            }
        }
    }
    OALOperationBridge_Destroy(&(state->secConn->super));
    DOFMem_Free(state);
}

static int DOFSecureConnector_RedoPermissions(OALTimerItem   item,
                                              uint32         argc,
                                              void         * argv[],
                                              boolean        shutdown)
{
    DOFSecureConnector_RedoPermissionState_t * state;

    PCR_Assert(argc == 1);
    PCR_Assert(argv[0] != NULL);
    state = (DOFSecureConnector_RedoPermissionState_t *) argv[0];
    if(shutdown)
    {
        OALOperationBridge_Destroy(&(state->secConn->super));
        DOFMem_Free(state);
    }
    else
    {
        DOFKeyRequest keyRequest = DOFKeyRequest_Create(state->secConn->credentials, DEFAULT_NONCE_SIZE, FALSE,
                                                        !state->secConn->extendablePermissions, NULL, *(state->originalPermissionSet));

        if(keyRequest != NULL)
        {
            TRP_V1_RequestSecurityScope(100, keyRequest, state->secConn->id,
                                        trpRedoPerms, state, NULL, NULL, 30000);
            DOFKeyRequest_Destroy(keyRequest);
        }
        else
        {
            OALOperationBridge_Destroy(&(state->secConn->super));
            DOFMem_Free(state);
        }
    }
    return 0;
}

static boolean DOFSecureConnector_SetupTimer(NormalSecurityConnector self,
                                             uint8                   inboundDuration,
                                             uint8                   outboundDuration)
{
    DOFSecureConnector_RedoPermissionState_t * state[2];

    PCR_Assert(self != NULL);
    state[0] = DOFMem_Alloc(sizeof(DOFSecureConnector_RedoPermissionState_t));
    if(state[0] != NULL)
    {
        state[0]->secConn = self;
        state[0]->permissionSet = &(self->inboundPermissions);
        state[0]->originalPermissionSet = &(self->originalInboundPermissions);
        state[0]->scope = &(self->inboundScope);
        state[0]->permissionIndex = &(self->inboundPermissionIndex);
        self->inboundTimerItem = OALTimer_CreateItem(secureConnectorTimer, DOFSecureConnector_RedoPermissions, 1, (void **) &state[0]);
        if(self->inboundTimerItem != NULL)
        {
            state[0]->timerItem = &(self->inboundTimerItem);
            state[1] = DOFMem_Alloc(sizeof(DOFSecureConnector_RedoPermissionState_t));
            if(state[1] != NULL)
            {
                state[1]->secConn = self;
                state[1]->permissionSet = &(self->outboundPermissions);
                state[1]->originalPermissionSet = &(self->originalOutboundPermissions);
                state[1]->scope = &(self->outboundScope);
                state[1]->permissionIndex = &(self->outboundPermissionIndex);
                self->outboundTimerItem = OALTimer_CreateItem(secureConnectorTimer, DOFSecureConnector_RedoPermissions, 1, (void **) &state[1]);
                if(self->outboundTimerItem != NULL)
                {
                    state[1]->timerItem = &(self->outboundTimerItem);
                    if(OALTimerItem_Insert(self->inboundTimerItem, DOFSecureConnector_ConvertDurationToMilliseconds(inboundDuration), FALSE, FALSE) == DOFERROR_OK)
                    {
                        if(OALTimerItem_Insert(self->outboundTimerItem, DOFSecureConnector_ConvertDurationToMilliseconds(outboundDuration), FALSE, FALSE) == DOFERROR_OK)
                        {
                            self->super.super.refCount += 2;
                            return TRUE;
                        }
                        OALTimerItem_Remove(self->inboundTimerItem);
                    }
                    OALTimerItem_Destroy(self->outboundTimerItem, TRUE);
                }
                DOFMem_Free(state[1]);
            }
            OALTimerItem_Destroy(self->inboundTimerItem, TRUE);
        }
        DOFMem_Free(state[0]);
    }

    return FALSE;
}

OALOperationBridge OALOperationBridge_CreateSecureBridge(DOF                       dof,
                                                         DOFCredentials            credentials,
                                                         DOFObjectIDAuthentication id,
                                                         DOFPermissionSet          inboundPermissions,
                                                         DOFSecurityScope          inboundScope,
                                                         uint8                     inboundDuration,
                                                         DOFPermissionSet          outboundPermissions,
                                                         DOFSecurityScope          outboundScope,
                                                         uint8                     outboundDuration,
                                                         boolean                   extendablePermissions)
{
    NormalSecurityConnector ret = DOFMem_Alloc(sizeof(NormalSecurityConnector_t));

    if(ret != NULL)
    {
        OALOperationBridge_Ctor(&ret->super, &normalVtable);
        ret->inboundScope = DOFSecurityScope_Copy(inboundScope);
        if(ret->inboundScope != NULL)
        {
            ret->outboundScope = DOFSecurityScope_Copy(outboundScope);
            if(ret->outboundScope != NULL)
            {
                ret->credentials = DOFCredentials_Copy(credentials);
                if((ret->credentials != NULL) || (credentials == NULL))
                {
                    ret->inboundPermissions = DOFPermissionSet_Copy(inboundPermissions);
                    if((ret->inboundPermissions != NULL) || (inboundPermissions == NULL))
                    {
                        ret->outboundPermissions = DOFPermissionSet_Copy(outboundPermissions);
                        if((ret->outboundPermissions != NULL) || (outboundPermissions == NULL))
                        {
                            ret->originalInboundPermissions = DOFPermissionSet_DeepCopy(inboundPermissions);
                            if((ret->originalInboundPermissions != NULL) || (inboundPermissions == NULL))
                            {
                                ret->originalOutboundPermissions = DOFPermissionSet_DeepCopy(outboundPermissions);
                                if((ret->originalOutboundPermissions != NULL) || (outboundPermissions == NULL))
                                {
                                    ret->id = DOFObjectIDAuthentication_Copy(id);
                                    if(ret->id != NULL)
                                    {
                                        ret->inboundPermissionIndex = 0;
                                        ret->outboundPermissionIndex = 0;
                                        if(DOFSecureConnector_SetupTimer(ret, inboundDuration, outboundDuration))
                                        {
                                            ret->extendablePermissions = extendablePermissions;
                                            return (OALOperationBridge) ret;
                                        }
                                        DOFObjectIDAuthentication_Destroy(ret->id);
                                    }
                                    DOFPermissionSet_Destroy(ret->originalOutboundPermissions);
                                }
                                DOFPermissionSet_Destroy(ret->originalInboundPermissions);
                            }
                            DOFPermissionSet_Destroy(ret->outboundPermissions);
                        }
                        DOFPermissionSet_Destroy(ret->inboundPermissions);
                    }
                    DOFCredentials_Destroy(ret->credentials);
                }
                DOFSecurityScope_Destroy(ret->outboundScope);
            }
            DOFSecurityScope_Destroy(ret->inboundScope);
        }
        DOFMem_Free(ret);
    }
    return NULL;
}

static void SecureConnector_normalDtor(DOFSecurityMembrane self)
{
    NormalSecurityConnector realSelf = (NormalSecurityConnector) self;

    PCR_Assert(self != NULL);

    DOFCredentials_Destroy(realSelf->credentials);
    DOFObjectIDAuthentication_Destroy(realSelf->id);
    DOFPermissionSet_Destroy(realSelf->inboundPermissions);
    DOFPermissionSet_Destroy(realSelf->outboundPermissions);
    DOFPermissionSet_Destroy(realSelf->originalInboundPermissions);
    DOFPermissionSet_Destroy(realSelf->originalOutboundPermissions);
    DOFSecurityScope_Destroy(realSelf->inboundScope);
    DOFSecurityScope_Destroy(realSelf->outboundScope);
    OALTimerItem_Destroy(realSelf->inboundTimerItem, FALSE);
    OALTimerItem_Destroy(realSelf->outboundTimerItem, FALSE);
    DOFMem_Free(realSelf);
}

static DOFSecurityScope SecureConnector_validateInboundPerms(DOFSecurityMembrane self,
                                                             DOFPermission       explicitPermission,
                                                             DOFPermission       logicalPermission,
                                                             OALCipherKeyState   keyState)
{
    NormalSecurityConnector realSelf = (NormalSecurityConnector) self;

    DOFSecurityScope scope;

    PCR_Assert(self != NULL);
    OALMutexPool_Lock(self);
    scope = DOFSecurityMembrane_ValidatePermission(DOFCredentials_GetDomainID(realSelf->credentials),
                                                   realSelf->inboundPermissions, explicitPermission);
    OALMutexPool_Unlock(self);
    return scope;
}

static DOFSecurityScope SecureConnector_validateOutboundPerms(DOFSecurityMembrane self,
                                                              DOFPermission       explicitPermission,
                                                              DOFPermission       logicalPermission)
{
    NormalSecurityConnector realSelf = (NormalSecurityConnector) self;

    DOFSecurityScope scope;

    PCR_Assert(self != NULL);
    OALMutexPool_Lock(self);
    scope = DOFSecurityMembrane_ValidatePermission(DOFCredentials_GetDomainID(realSelf->credentials),
                                                   realSelf->outboundPermissions, explicitPermission);
    OALMutexPool_Unlock(self);
    return scope;
}

static DOFSecurityScope SecureConnector_getInboundScope(DOFSecurityMembrane self)
{
    NormalSecurityConnector realSelf = (NormalSecurityConnector) self;
    DOFSecurityScope ret;

    PCR_Assert(self != NULL);
    OALMutexPool_Lock(&(realSelf->super));
    ret = realSelf->inboundScope;
    OALMutexPool_Unlock(&(realSelf->super));
    return ret;
}

static DOFSecurityScope SecureConnector_getOutboundScope(DOFSecurityMembrane self)
{
    NormalSecurityConnector realSelf = (NormalSecurityConnector) self;
    DOFSecurityScope ret;

    PCR_Assert(self != NULL);
    OALMutexPool_Lock(&(realSelf->super));
    ret = realSelf->outboundScope;
    OALMutexPool_Unlock(&(realSelf->super));
    return ret;
}

static DOFSecurityMembraneExtendReturn_t SecureConnector_extendPerms(DOFCredentials     credentials,
                                                                     DOFPermissionSet   originalSet,
                                                                     DOFPermissionSet   extendSet,
                                                                     void             * context);

struct ExtendPermContext_t
{
    ExtendCompleteCB originalCallback;
    void * originalContext;
    uint32 permissionIndex;
    uint32 * permissionIndexP;

    DOFPermissionSet permissionSet;
};

typedef struct ExtendPermContext_t * ExtendPermContext;

static ExtendPermContext ExtendPermContext_Create(ExtendCompleteCB   callback,
                                                  DOFPermissionSet   permissionSet,
                                                  void               * userContext);

static DOFSecurityMembraneExtendReturn_t SecureConnector_extendInboundPerms(DOFSecurityMembrane   self,
                                                                            DOFPermissionSet      extendPerm,
                                                                            ExtendCompleteCB      callback,
                                                                            void                * context)
{
    NormalSecurityConnector realSelf = (NormalSecurityConnector) self;

    PCR_Assert(self != NULL);
    PCR_Assert(extendPerm != NULL);
    if(realSelf->extendablePermissions)
    {
        ExtendPermContext fullContext = ExtendPermContext_Create(callback, realSelf->inboundPermissions, context);

        if(fullContext != NULL)
        {
            DOFSecurityMembraneExtendReturn_t retVal;

            OALMutexPool_Lock(self);
            fullContext->permissionIndex = realSelf->inboundPermissionIndex;
            fullContext->permissionIndexP = &(realSelf->inboundPermissionIndex);
            OALMutexPool_Unlock(self);
            retVal = SecureConnector_extendPerms(realSelf->credentials, realSelf->inboundPermissions,
                                                extendPerm, (void *) fullContext);
            if(retVal != EXTEND_PENDING)
            {
                DOFMem_Free(fullContext);
            }
            return retVal;
        }
    }
    return CANNOT_EXTEND;
}

static DOFSecurityMembraneExtendReturn_t SecureConnector_extendOutboundPerms(DOFSecurityMembrane   self,
                                                                             DOFPermissionSet      extendPerm,
                                                                             ExtendCompleteCB      callback,
                                                                             void                * context)
{
    NormalSecurityConnector realSelf = (NormalSecurityConnector) self;

    PCR_Assert(self != NULL);
    PCR_Assert(extendPerm != NULL);
    if(realSelf->extendablePermissions)
    {
        ExtendPermContext fullContext = ExtendPermContext_Create(callback, realSelf->outboundPermissions, context);

        if(fullContext != NULL)
        {
            DOFSecurityMembraneExtendReturn_t retVal;

            OALMutexPool_Lock(self);
            fullContext->permissionIndex = realSelf->outboundPermissionIndex;
            fullContext->permissionIndexP = &(realSelf->outboundPermissionIndex);
            OALMutexPool_Unlock(self);
            retVal = SecureConnector_extendPerms(realSelf->credentials, realSelf->outboundPermissions,
                                                extendPerm, fullContext);

            if(retVal != EXTEND_PENDING)
            {
                DOFPermissionSet_Destroy(fullContext->permissionSet);
                DOFMem_Free(fullContext);
            }
            return retVal;
        }
    }
    return CANNOT_EXTEND;
}

static void SecureConnector_extendPermCallback(TRP_V1_Context   trp_context,
                                               void           * context);

static DOFSecurityMembraneExtendReturn_t SecureConnector_extendPerms(DOFCredentials     credentials,
                                                                     DOFPermissionSet   originalSet,
                                                                     DOFPermissionSet   extendSet,
                                                                     void             * context)
{
    DOFSecurityMembraneExtendReturn_t retVal = CANNOT_EXTEND;

    DOFKeyRequest keyRequest;

    PCR_Assert(originalSet != NULL);
    PCR_Assert(extendSet != NULL);
    PCR_Assert(DOFPermissionSet_GetPermissionCount(extendSet) > 0);

    keyRequest = DOFKeyRequest_Create(credentials, DEFAULT_NONCE_SIZE, FALSE, TRUE, NULL, extendSet);
    if(keyRequest != NULL)
    {
        void * fullContext = context;

        if(fullContext != NULL)
        {
            TRP_V1_RequestSecurityScope(255, keyRequest, DOFCredentials_GetIdentity(credentials), SecureConnector_extendPermCallback,
                                        context, NULL, NULL, DEFAULT_TIMEOUT);
            retVal = EXTEND_PENDING;
        }
        DOFKeyRequest_Destroy(keyRequest);
    }
    return retVal;
}

static void SecureConnector_extendPermCallback(TRP_V1_Context   trp_context,
                                               void           * context)
{
    ExtendPermContext realContext = (ExtendPermContext) context;

    DOFError statusCode;

    PCR_Assert(context != NULL);
    if(trp_context != NULL)
    {
        if(trp_context->ScopeReturn->status == DOFERROR_OK)
        {
            if(trp_context->ScopeReturn->errorCode == 0)
            {
                /* if(realContext->permissionIndex == *(realContext->permissionIndexP)) */
                {
                    DOFPermissionSet_AddPermissionsFromSet(realContext->permissionSet, trp_context->ScopeReturn->permValidation);
                }
                statusCode = DOFERROR_OK;
            }
            else
            {
                statusCode = trp_context->ScopeReturn->errorCode;
            }
        }
        else
        {
            statusCode = trp_context->ScopeReturn->status;
        }
    }
    else
    {
        statusCode = DOFERROR_NO_AUTHENTICATION_SERVER;
    }
    if(realContext->originalCallback != NULL)
    {
        realContext->originalCallback(statusCode, realContext->originalContext);
    }
    DOFPermissionSet_Destroy(realContext->permissionSet);
    DOFMem_Free(realContext);
}

static ExtendPermContext ExtendPermContext_Create(ExtendCompleteCB   callback,
                                                  DOFPermissionSet   permissionSet,
                                                  void             * userContext)
{
    ExtendPermContext retVal = (ExtendPermContext) DOFMem_Alloc(sizeof(struct ExtendPermContext_t));

    PCR_Assert(permissionSet != NULL);
    if(retVal != NULL)
    {
        retVal->originalCallback = callback;
        retVal->originalContext = userContext;
        retVal->permissionSet = DOFPermissionSet_Copy(permissionSet);
    }
    return retVal;
}

static DOFObjectIDDomain SecureConnector_getDomainId(DOFSecurityMembrane self)
{
    NormalSecurityConnector realSelf = (NormalSecurityConnector) self;

    PCR_Assert(self != NULL);
    return DOFCredentials_GetDomainID(realSelf->credentials);
}

static DOFPermissionSet SecureConnector_getInboundPerms(DOFSecurityMembrane self)
{
    return ((NormalSecurityConnector)self)->inboundPermissions;
}

static DOFPermissionSet SecureConnector_getOutboundPerms(DOFSecurityMembrane self)
{
    return ((NormalSecurityConnector)self)->outboundPermissions;
}
