/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/address.h>
#include <dof/credentials.h>
#include <dof/connection.h>
#include <dof/exception.h>
#include <dof/server.h>
#include <dof/permissionset.h>
#include <dof/connection.h>

#include <internal/addressinternal.h>
#include <internal/bridgeinternal.h>
#include <internal/esp_v2.h>
#include <internal/channel.h>
#include <internal/cipher.h>
#include <internal/dofinternal.h>
#include <internal/enp_v1.h>
#include <internal/mutex_pool.h>
#include <internal/permission_manager.h>
#include <internal/permissioninternal.h>
#include <internal/permissionsetinternal.h>
#include <internal/remotechannel.h>
#include <internal/exception.h>
#include <internal/securitymode.h>
#include <internal/protocolstack.h>
#include <internal/serverinternal.h>
#include <internal/connection_internal.h>
#include <internal/connectionmanager.h>
#include <internal/channel_stack_common.h>
#include <internal/remotechanneldata.h>
#include <internal/remotechannelmanager.h>
#include <internal/router.h>
#include <internal/securegroupinternal.h>
#include <internal/transportinternal.h>
#include <internal/connectionless_protocolstack.h>
#include <internal/connectionless_remotechannel.h>
#include <internal/connected_protocolstack.h>

#include "channel_local.h"

#ifdef OALCHANNEL_DEBUG_REFCOUNT
#include <stdio.h>
#endif


/*
**  "Private" Type and Method Declarations
*/


extern PCRMutex ref_count_mutex;
extern PCRMutex connection_server_mutex;


/*
**  "Public" Methods
*/

boolean OALChannel_Ctor(OALChannel_t* channel, boolean isRemote, DOFSecurityMembrane securityMembrane,
                        OALChannel_RouteToProvider_fn_t    routeToProviderFunc,
                        OALChannel_RouteToRequestor_fn_t   routeToRequestorFunc,
                        OALChannel_MessageHandler_fn_t     msg_handler_func,
                        OALChannel_Destroy_fn_t            destroyFunc,
                        OALChannel_ExtendPermissions       extendPermissions)
{
    DOFSecurityScope securityScope;

    /* This function requires that the caller have zeroed full contents of the OALChannel_t structure at channel. */
    channel->refCount = 1;
    channel->chObservable = OALObservable_Create();
    if(!channel->chObservable)
    {
        goto OALChannel_Ctor_Fail;
    }

    if (isRemote)
    {
        channel->remoteChannelData = (RemoteChannelData)DOFMem_Alloc( sizeof( RemoteChannelData_t ) );

        if (!channel->remoteChannelData)
        {
            goto OALChannel_Ctor_Fail;
        }

        PCRMem_Set( channel->remoteChannelData, 0, sizeof( RemoteChannelData_t ) );
    }
    else channel->remoteChannelData = NULL;

    securityScope = DOF_GetDefaultSecurityScope(DOF_GetDof());

    /* Set this to the DOF security scope for now, it will get overwitten when security is negotiated. */
    channel->securityScope = DOFSecurityScope_Copy(securityScope);

    if (securityScope && !channel->securityScope)
    {
        goto OALChannel_Ctor_Fail;
    }
    channel->securityMembrane = DOFSecurityMembrane_Copy(securityMembrane);
    if (securityMembrane && !channel->securityMembrane)
    {
        goto OALChannel_Ctor_Fail;
    }
    channel->permissionManager = OALPermissionManager_Create();
    if(channel->permissionManager == NULL)
    {
        goto OALChannel_Ctor_Fail;
    }
    channel->outboundMembranePermissionManager = OALPermissionManager_Create();
    if(channel->outboundMembranePermissionManager == NULL)
    {
        goto OALChannel_Ctor_Fail;
    }
    channel->inboundMembranePermissionManager = OALPermissionManager_Create();
    if(channel->inboundMembranePermissionManager == NULL)
    {
        goto OALChannel_Ctor_Fail;
    }
    channel->routeToProvider = routeToProviderFunc;
    channel->routeToRequestor = routeToRequestorFunc;
    channel->msg_handler_func = msg_handler_func;
    channel->destroy = destroyFunc;
    channel->extendPermissions = extendPermissions;

    return TRUE;

    OALChannel_Ctor_Fail:

    if (channel->remoteChannelData)
    {
        DOFMem_Free(channel->remoteChannelData);
        channel->remoteChannelData = NULL;
    }
    OALObservable_Destroy(channel->chObservable);
    channel->chObservable = NULL;
    DOFSecurityScope_Destroy(channel->securityScope);
    channel->securityScope = NULL;
    DOFSecurityMembrane_Destroy(channel->securityMembrane);
    channel->securityMembrane = NULL;

    return FALSE;
}

void OALChannel_Dtor( OALChannel_t* channel )
{
    if(!channel)
        return;

    /* channel->refCount is 0 right now- because we are called here only at the time that it becomes 0. */
    DOFSecurityScope_Destroy( channel->securityScope );
    DOFConnection_Destroy(channel->sysConnection);
    OALPermissionManager_Destroy(channel->permissionManager);
    OALPermissionManager_Destroy(channel->outboundMembranePermissionManager);
    OALPermissionManager_Destroy(channel->inboundMembranePermissionManager);
    if(channel->remoteChannelData)
    {
        OALChannel_RemoteChannel_Dtor( channel );
        DOFMem_Free( channel->remoteChannelData );
    }
    DOFSecurityMembrane_Destroy(channel->securityMembrane);
    OALChannel_Lock(channel);
    OALObservable_Destroy( channel->chObservable );
    OALChannel_Unlock(channel);
}

static boolean OALChannel_RemoteChannel_CtorConfig(OALChannel_t* channel,
                                                   DOFConnectionConfig connConfig,
                                                   DOFServerConfig serverConfig)
{
    if (connConfig)
    {
        boolean retStatus = TRUE;

        channel->remoteChannelData->config.connection = DOFConnectionConfig_Copy(connConfig);
        if (channel->remoteChannelData->config.connection)
        {
            DOFConnectionType connType = DOFConnectionConfig_GetConnectionType(connConfig);

            /* Groups are configured with the server's address, which may be unicast, but the group address is multicast */
            if (connType == DOFCONNECTIONTYPE_GROUP ||
                connType == DOFCONNECTIONTYPE_GROUP_MEMBER_ONLY ||
                connType == DOFCONNECTIONTYPE_HUB)
            {
                channel->remoteChannelData->flags.defaultAddressIsMulticast = TRUE;
            }

            channel->remoteChannelData->flags.isLossy = ( connType != DOFCONNECTIONTYPE_STREAM );
            channel->remoteChannelData->flags.isConnection = TRUE;      /* This must be set for all connections including secure groups. */
            channel->remoteChannelData->negotiator = DOFConnectionConfig_GetProtocolNegotiator(connConfig);

            if( DOFConnectionConfig_GetSecurityDesire(connConfig) & DOFSECURITYDESIRE_SECURE_ANY )
            {
                channel->remoteChannelData->mode = DOFSecurityModeCCM_Create( DOFCIPHERALGORITHM_AES, DOFCIPHERSTRENGTH_256, 4, (DOFConnectionConfig_GetSecurityDesire(connConfig) == DOFSECURITYDESIRE_SECURE));
                if (!channel->remoteChannelData->mode)
                {
                    retStatus = FALSE;
                }
            }
            if (retStatus)
            {
                return TRUE;
            }
            DOFConnectionConfig_Destroy(channel->remoteChannelData->config.connection);
            channel->remoteChannelData->config.connection = NULL;
        }
    }
    else
    {
        channel->remoteChannelData->config.server = DOFServerConfig_Copy(serverConfig);
        if (channel->remoteChannelData->config.server)
        {
            channel->remoteChannelData->flags.isLossy = TRUE;
            channel->remoteChannelData->flags.isConnection = FALSE;
            channel->remoteChannelData->negotiator = DOFServerConfig_GetProtocolNegotiator( serverConfig );
            return TRUE;
        }
    }

    return FALSE;
}

/* Note: This function requires that the caller have zeroed full contents of the RemoteChannelData_t structure at
   channel->remoteChannelData. */
boolean OALChannel_RemoteChannel_Ctor(OALChannel_t* channel,
                                      void* protocolStack,
                                      DTL transport,
                                      DOFAddress connectedAddr,
                                      DOFConnectionConfig connConfig,
                                      DOFServerConfig serverConfig)
{
    PCR_Assert(connConfig || serverConfig);         /* Connection Config or server config must be passed in. */
    PCR_Assert(!connConfig || !serverConfig);      /* It is an error to pass in both a connection config and a server config. */

    channel->remoteChannelData->protocolStack = OALProtocolStack_Retain(protocolStack);

    if (channel->remoteChannelData->protocolStack || !protocolStack)
    {
        /* Transport may be NULL for cases where there is a connection with a related connection or related server. */
        channel->remoteChannelData->dtl = DTL_Retain(transport);

        if (channel->remoteChannelData->dtl || !transport)
        {
            channel->remoteChannelData->address = DOFAddress_Copy( connectedAddr );
            if (channel->remoteChannelData->address || !connectedAddr)
            {
                channel->remoteChannelData->waitingForData = FALSE;

                if (DOFAddress_GetAddressType(channel->remoteChannelData->address) != DOFADDRESSTYPE_UNICAST)
                    channel->remoteChannelData->flags.defaultAddressIsMulticast = TRUE;

                if (OALChannel_RemoteChannel_CtorConfig(channel, connConfig, serverConfig))
                {
                    /* Since 0 is valid for securityStateID and ENP_Port, set them to the defined invalid value. */
                    channel->remoteChannelData->securityStateID = INVALID_SECURITYSTATEID;
                    channel->remoteChannelData->ENP_Port = ENP_INVALID_PORT;

                    return TRUE;    /* The remote channel was successfully initialized. */
                }
                DOFAddress_Destroy(channel->remoteChannelData->address);
                channel->remoteChannelData->address = NULL;
            }
            DTL_Destroy(channel->remoteChannelData->dtl);
            channel->remoteChannelData->dtl = NULL;
        }
    }

    return FALSE;
}

void OALChannel_RemoteChannel_Dtor( OALChannel_t* channel )
{
    if(!channel)
        return;

    /* since this channel has a base channel it must be released. */
    OALChannel_Release(channel->remoteChannelData->baseChannel);
    DTL_Destroy( channel->remoteChannelData->dtl );
    OALMessage_Destroy(channel->remoteChannelData->shutdownMessage);
    OALConnection_Release(channel->remoteChannelData->failedCIShutdown);
    OALProtocolStack_Release(channel->remoteChannelData->protocolStack);

    if (channel->remoteChannelData->flags.isConnection)
    {
        DOFConnectionConfig_Destroy(channel->remoteChannelData->config.connection);
    }
    else
    {
        DOFServerConfig_Destroy(channel->remoteChannelData->config.server);
    }

    DOFAddress_Destroy( channel->remoteChannelData->address );
    OALCipher_Destroy( channel->remoteChannelData->cipher );
    DOFSecurityMode_Destroy( channel->remoteChannelData->mode );

    if (channel->remoteChannelData->registeredSSIDs_AS)
        OALAVLTreeList_Destroy(channel->remoteChannelData->registeredSSIDs_AS, NULL, NULL);

    OALHashtable_Destroy(channel->remoteChannelData->registeredSSIDs_NonAs, NULL, NULL);

    if (channel->remoteChannelData->registeredENPPorts)
        OALAVLTree_Destroy(channel->remoteChannelData->registeredENPPorts, NULL, NULL);

    DOFCredentials_Destroy(channel->remoteChannelData->secureCredentials);
    source_release(channel->remoteChannelData->secureSid);
    DOFObjectIDSource_Destroy(channel->remoteChannelData->remoteSid);
}

#ifndef OALCHANNEL_DEBUG_REFCOUNT
OALChannel OALChannel_Retain( OALChannel self )
#else
OALChannel OALChannel_Retain_Debug( OALChannel self, const char *fn_name, const char *filename, unsigned int line_number )
#endif
{
    OALChannel ret = self;

    if (ret)
    {
        PCRMutex_Lock(ref_count_mutex);
        if (ret->refCount < 0xFFFFFFFF)
        {
            ++(ret->refCount);
#ifdef OALCHANNEL_DEBUG_REFCOUNT
            fprintf(stderr, "OALChannel(%p): ++refCount = %u from %s at %s:%u\n", (void *)ret, (unsigned int) ret->refCount, fn_name, filename, line_number );
#endif
        }
        else
        {
            /* refCount is already at max so return NULL. */
            ret = NULL;
        }
        PCRMutex_Unlock(ref_count_mutex);
    }
    return ret;
}

#ifndef OALCHANNEL_DEBUG_REFCOUNT
void OALChannel_Release( OALChannel self )
#else
void OALChannel_Release_Debug( OALChannel self, const char *fn_name, const char *filename, unsigned int line_number )
#endif
{
    uint32 refCount;

    if (!self)
        return;

    PCRMutex_Lock(ref_count_mutex);
    PCR_Assert( self->refCount );
    refCount = --(self->refCount);
#ifdef OALCHANNEL_DEBUG_REFCOUNT
        fprintf(stderr, "OALChannel(%p): --refCount = %d from %s at %s:%u\n", (void *)self, self->refCount, fn_name, filename, line_number );
#endif
    PCRMutex_Unlock(ref_count_mutex);

    if ( refCount )
        return;

    PCR_Assert(self->destroy);
    self->destroy(self);
    OALMutexPool_Finalize( self );
    DOFMem_Free( self );
}

void OALChannel_Lock(OALChannel self)
{
    OALMutexPool_Lock(self);
}

void OALChannel_Unlock(OALChannel self)
{
    OALMutexPool_Unlock(self);
}

/* @internal */
boolean OALChannel_SetSecureSendBlocked( OALChannel channel )
{
    PCR_Assert( channel->remoteChannelData->cipher );
    /* channel known to be secureGroup. Do not set channel->remoteChannelData->flags.SecureSendBlocked directly, but
       because cipher may have already fixed the problem, ask cipher to do it, or not, under its own mutex protection.*/
    OALCipher_CheckAndSetChannelSecureSendBlocked( channel->remoteChannelData->cipher );

    return TRUE;
}

DOFSecurityScope OALChannel_GetSecurityScope( OALChannel chan, boolean inbound )
{
    DOFSecurityScope ret;
    if ( chan->securityMembrane )
    {
        if ( inbound )
            ret = DOFSecurityMembrane_GetInboundScope( chan->securityMembrane );
        else
            ret = DOFSecurityMembrane_GetOutboundScope( chan->securityMembrane );
    }
    else
        ret = chan->securityScope;
    return ret;
}

DOFConnectionState OALChannel_CreateConnectionState( OALChannel channel, DOFAddress address )
{
    DOFConnectionState  connectionState = NULL;

    if( channel )
    {
        DOFConnection           relatedDOFConnection = NULL;
        DOFServer               relatedDOFServer = NULL;
        DOFConnectionDirection  direction = DOFCONNECTIONDIRECTION_INBOUND;
        DOFConnectionConfig     connConfig = NULL;
        boolean                 isConnected = TRUE;

        uint64                  ibc = 0;
        uint64                  obc = 0;
        uint64                  ipc = 0;
        uint64                  opc = 0;

        PCRMutex_Lock(connection_server_mutex);

        if ( channel->remoteChannelData )
        {
            if (channel->remoteChannelData->flags.isConnection && channel->remoteChannelData->flags.isOutbound)
            {
                direction = DOFCONNECTIONDIRECTION_OUTBOUND;
            }

            if ( channel->remoteChannelData->dtl )
            {
                if ( channel->remoteChannelData->flags.isConnection )
                {
                    connConfig = DOFConnectionConfig_Copy( channel->remoteChannelData->config.connection );

                    if (channel->remoteChannelData->OALConn)
                    {
                        relatedDOFConnection = OALConnection_GetRelatedDOFConnection(channel->remoteChannelData->OALConn);
                        relatedDOFServer = OALConnection_GetRelatedDOFServer(channel->remoteChannelData->OALConn);
                    }
                    else
                    {
                        isConnected = FALSE;
                    }
                }
                else
                {
                    DOFConnectionConfigBuilder cb = DOFConnectionConfigBuilder_Create( DOFCONNECTIONTYPE_DATAGRAM_STATELESS, address );

                    if (cb)
                    {
                        DTLConnectionConfig DTLConfig;

                        DOFServerConfig sc = channel->remoteChannelData->config.server;

                        DOFConnectionConfigBuilder_SetMaxReceiveSilence( cb, DOFServerConfig_GetMaxReceiveSilence( sc ) );
                        DOFConnectionConfigBuilder_SetMaxSendSilence( cb, DOFServerConfig_GetMaxSendSilence( sc ) );
                        DOFConnectionConfigBuilder_SetProtocolNegotiator( cb, DOFServerConfig_GetProtocolNegotiator( sc ) );
                        DOFConnectionConfigBuilder_SetSecurityDesire( cb, DOFServerConfig_GetSecurityDesire( sc ) );
                        DOFConnectionConfigBuilder_SetReceiveFilter( cb, DOFServerConfig_GetReceiveFilter( sc ) );
                        DOFConnectionConfigBuilder_SetSendFilter( cb, DOFServerConfig_GetSendFilter( sc ) );
                        DOFConnectionConfigBuilder_SetBridge( cb, DOFServerConfig_GetBridge( sc ) );
                        if (DOFTransportLayer_CreateConnectionConfig( channel->remoteChannelData->dtl, address, &DTLConfig ))
                        {
                            DOFConnectionConfigBuilder_SetTransportConfig( cb, DTLConfig );
                            DTLConnectionConfig_Destroy(DTLConfig);
                        }
                        /* The permissions and Credentials in the server config are a list of permissions and credentials, so for connection state they
                           do not get copied into the connection config. */
                        connConfig = DOFConnectionConfigBuilder_BuildAndDestroy( cb );

                        isConnected = FALSE;
                    }
                }
            }
            else
            {
                isConnected = FALSE;
            }
        }
        else
        {
            connConfig = DOFConnectionConfig_Copy(channel->sysConnection->OALConn->connConfig);
        }

        if (connConfig)
        {
            DOFCredentials peerCreds = NULL;

            if(channel->remoteChannelData && channel->remoteChannelData->OALConn )
            {
                peerCreds =  channel->remoteChannelData->OALConn->peerCredentials;
            }

            connectionState = DOFConnectionState_Create( DOF_GetDof(), connConfig, channel->securityMembrane,
                                                         isConnected, channel, direction, peerCreds,
                                                         relatedDOFConnection, relatedDOFServer, ibc, obc, ipc, opc, NULL );
            /* The connectionState now holds a reference to connConfig, or connectionState was not created, in either case connConnfig
               was created or copied earlier in this function so call DOFConnectionConfig_Destroy now. */
            DOFConnectionConfig_Destroy( connConfig );
        }
        PCRMutex_Unlock(connection_server_mutex);
    }
    return connectionState;
}

void OALChannel_ExtendOutboundPermissions(OALChannel         channel,
                                          DOFAddress         address,
                                          DOFPermission      perm,
                                          ExtendCompleteCB   callback,
                                          void             * context)
{
    void * permContext;

    OALChannel_Lock(channel);
    permContext = OALPermissionManager_GetPendingContext(channel->permissionManager, perm);
    if(permContext != NULL)
    {
        boolean worked = OALPermissionManager_AddPendingContext(channel->permissionManager, permContext, callback, context);

        OALChannel_Unlock(channel);
        if(!worked)
        {
            callback(DOFERROR_INSUFFICIENT_RESOURCES, context);
        }
    }
    else
    {
        DOFPermissionSet permSet = DOFPermission_CreateExtendSet(perm);

        if(permSet != NULL)
        {
            boolean worked = OALPermissionManager_AddNewContext(channel->permissionManager, permSet,
                                                                &callback, &context);

            OALChannel_Unlock(channel);
            if(worked)
            {
                channel->extendPermissions(channel, address, permSet, callback, context);
            }
            else
            {
                callback(DOFERROR_FAILED, context);
            }
            DOFPermissionSet_Destroy(permSet);
        }
        else
        {
            OALChannel_Unlock(channel);
            callback(DOFERROR_INSUFFICIENT_RESOURCES, context);
        }
    }
}

void OALChannel_ExtendOutboundMembranePermissions(OALChannel         channel,
                                                  DOFPermission      perm,
                                                  ExtendCompleteCB   callback,
                                                  void             * context)
{
    void * permContext;

    OALChannel_Lock(channel);
    permContext = OALPermissionManager_GetPendingContext(channel->outboundMembranePermissionManager, perm);
    if(permContext != NULL)
    {
        boolean worked = OALPermissionManager_AddPendingContext(channel->outboundMembranePermissionManager, permContext, callback, context);

        OALChannel_Unlock(channel);
        if(!worked)
        {
            callback(DOFERROR_INSUFFICIENT_RESOURCES, context);
        }
    }
    else
    {
        DOFPermissionSet permSet = DOFPermission_CreateExtendSet(perm);

        if(permSet != NULL)
        {
            boolean worked = OALPermissionManager_AddNewContext(channel->outboundMembranePermissionManager, permSet,
                                                                &callback, &context);

            OALChannel_Unlock(channel);
            if(worked)
            {
                DOFSecurityMembraneExtendReturn_t status = DOFSecurityMembrane_ExtendOutboundPermissions(channel->securityMembrane, permSet,
                                                                                                         callback, context);

                switch(status)
                {
                case EXTEND_PENDING:
                    /* everything is great */
                    break;

                case EXTEND_ELSEWHERE:
                case CANNOT_EXTEND:
                default:
                    callback(DOFERROR_FAILED, context);
                    break;
                }
            }
            else
            {
                callback(DOFERROR_FAILED, context);
            }
            DOFPermissionSet_Destroy(permSet);
        }
        else
        {
            OALChannel_Unlock(channel);
            callback(DOFERROR_INSUFFICIENT_RESOURCES, context);
        }
    }
}

void OALChannel_ExtendInboundMembranePermissions(OALChannel         channel,
                                                 DOFPermission      perm,
                                                 ExtendCompleteCB   callback,
                                                 void             * context)
{
    void * permContext;

    OALChannel_Lock(channel);
    permContext = OALPermissionManager_GetPendingContext(channel->inboundMembranePermissionManager, perm);
    if(permContext != NULL)
    {
        boolean worked = OALPermissionManager_AddPendingContext(channel->inboundMembranePermissionManager, permContext, callback, context);

        OALChannel_Unlock(channel);
        if(!worked)
        {
            callback(DOFERROR_INSUFFICIENT_RESOURCES, context);
        }
    }
    else
    {
        DOFPermissionSet permSet = DOFPermission_CreateExtendSet(perm);

        if(permSet != NULL)
        {
            boolean worked = OALPermissionManager_AddNewContext(channel->inboundMembranePermissionManager, permSet,
                                                                &callback, &context);

            OALChannel_Unlock(channel);
            if(worked)
            {
                DOFSecurityMembraneExtendReturn_t status = DOFSecurityMembrane_ExtendInboundPermissions(channel->securityMembrane, permSet,
                                                                                                        callback, context);

                switch(status)
                {
                case EXTEND_PENDING:
                    /* everything is great */
                    break;

                case EXTEND_ELSEWHERE:
                case CANNOT_EXTEND:
                default:
                    callback(DOFERROR_FAILED, context);
                    break;
                }
            }
            else
            {
                callback(DOFERROR_FAILED, context);
            }
            DOFPermissionSet_Destroy(permSet);
        }
        else
        {
            OALChannel_Unlock(channel);
            callback(DOFERROR_INSUFFICIENT_RESOURCES, context);
        }
    }
}

typedef struct channel_findContext_t
{
    OALChannel ret;
    union
    {
        DOFConnectionConfig connConfig;
        DOFServerConfig serverConfig;
    }type;
    boolean isConnection;
    boolean addressFound;
}channel_findContext_t;

typedef struct channel_findContext_t* channel_findContext;

/* This function will create and return a channel to be used for a connection.  If a channel is returned
   OALChannel_Release must be called to release memory created by this function. */
OALChannel OALChannelMgr_CreateChannel_Connection( OALConnection OALConn, DOFException * exception )
{
    channel_findContext_t   retStruct;
    OALChannel              baseChannel = NULL;
    DTL                     dtl = NULL;
    DOFConnectionType       connType;
    boolean                 sharedDTL = FALSE;

    retStruct.isConnection = 1;
    retStruct.addressFound = FALSE;
    retStruct.ret = NULL;
    retStruct.type.connConfig = OALConn->connConfig;

    if (exception)
        *exception = NULL;

    if (OALConn->flags.baseIsOALConnection && OALConn->base.OALConn &&
        (DOFConnectionConfig_GetConnectionType(OALConn->connConfig) != DOFCONNECTIONTYPE_POINT))
    {
        baseChannel = OALChannel_Retain(OALConn->base.OALConn->channel);
        /* Use the DTL from the base channel. */
        sharedDTL = TRUE;
    }
    else if (!OALConn->flags.baseIsOALConnection && OALConn->base.OALSrv)
    {
        baseChannel = OALChannel_Retain(OALServer_GetSendServerChannel(OALConn->base.OALSrv));
        /* Use the DTL from the base channel. */
        sharedDTL = TRUE;
    }
    else
    {
        dtl = DTL_Create();
    }

    if(!dtl && !sharedDTL)
    {
        if (exception)
        {
            *exception = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        }
        return NULL;
    }

    if (sharedDTL && !baseChannel)
    {
        if (exception)
        {
            *exception = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        }
        return NULL;
    }

    /* set up channel */
    connType = DOFConnectionConfig_GetConnectionType(OALConn->connConfig);
    switch(connType)
    {
    case DOFCONNECTIONTYPE_STREAM:
        {
            OALConnectedProtocolStack protoStack;

            PCR_Assert(!sharedDTL);

            protoStack = OALConnectedProtocolStack_Create(DOFConnectionConfig_GetAddress(OALConn->connConfig));
            if (protoStack)
            {
                retStruct.ret = (OALChannel)OALConnectedRemoteChannel_Create( protoStack, dtl, OALConn->connSecurityMembrane, OALConn->connConfig, TRUE );
                if (retStruct.ret)
                {
                    OALConnectedProtocolStack_SetChannel( (OALConnectedProtocolStack)protoStack,(OALConnectedRemoteChannel)retStruct.ret );
                    if (OALRemoteChannelMgr_AddChannel((OALChannel)retStruct.ret))
                    {
                        OALRemoteChannelMgr_AddProtocolStack( (OALProtocolStack)protoStack );
                        retStruct.ret->remoteChannelData->flags.isUp = FALSE;
                    }
                    else
                    {
                        OALChannel_Release(retStruct.ret);
                        retStruct.ret = NULL;
                    }
                }
                OALProtocolStack_Release( (OALProtocolStack)protoStack );
            }
        }
        break;
    case DOFCONNECTIONTYPE_DATAGRAM_STATELESS:
    case DOFCONNECTIONTYPE_DATAGRAM:
        {
            retStruct.ret = (OALChannel)OALConnectionlessRemoteChannel_Create( gDefaultConnectionlessProtocolStack, dtl, OALConn->connSecurityMembrane,
                                                                               DOFConnectionConfig_GetAddress(OALConn->connConfig), OALConn->connConfig, NULL );
            if (retStruct.ret)
            {
                if (OALRemoteChannelMgr_AddChannel(retStruct.ret))
                {
                    retStruct.ret->remoteChannelData->flags.isUp = FALSE;
                }
                else
                {
                    /* The channel was not added to the remote channel manager, release the channel. */
                    OALChannel_Release(retStruct.ret);
                    retStruct.ret = NULL;
                }
            }
        }
        break;
    case DOFCONNECTIONTYPE_HUB:
    case DOFCONNECTIONTYPE_GROUP_MEMBER_ONLY:
    case DOFCONNECTIONTYPE_GROUP:
        {
            if (!OALConn->flags.baseIsOALConnection && OALConn->base.OALSrv)
            {
                DOFObjectIDAuthentication group;

                PCR_Assert( !retStruct.ret );
                if ((group = DOFGroupAddress_GetGroupID( DOFConnectionConfig_GetAddress(OALConn->connConfig) )) != NULL)
                {
                    DOFSecureGroup secureGroup;

                    if ((secureGroup = DOFSecureGroup_Create( group, gDefaultConnectionlessProtocolStack, dtl,
                                                              OALConn->connSecurityMembrane, OALConn->base.OALSrv, NULL,
                                                              OALConn->connConfig,
                                                              OALConn->connConfig->permissions )) != NULL)
                    {
                        if (OALRemoteChannelMgr_AddChannel( (OALChannel)secureGroup ))       /* This retains channel. */
                        {
                            retStruct.ret = (OALChannel)secureGroup;
                            retStruct.ret->remoteChannelData->flags.isUp = FALSE;
                        }
                        else
                        {
                            OALChannel_Release( (OALChannel)secureGroup );
                            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALChannelMgr: Failed to create secure group.");
                        }
                    }
                    else
                        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALChannelMgr: Failed to create secure group (out of memory).");
                }
                else
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALChannelMgr: Failed to create secure group (unmatched groupID).");
            }
            else
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALChannelMgr: Failed to create secure group (need base server).");
                if (exception)
                    *exception = DOFException_Copy( DOFEXCEPTION_FAILED );
                PCR_Assert( !retStruct.ret );
            }
        }
        break;
    case DOFCONNECTIONTYPE_POINT:
        {
            DOFObjectIDAuthentication group;
            DOFSecureGroup            secureGroup;
            if (OALConn->flags.baseIsOALConnection && OALConn->base.OALConn )
            {
                if ((group = DOFGroupAddress_GetGroupID( DOFConnectionConfig_GetAddress( OALConn->connConfig ) )) == NULL ||
                    (secureGroup = DOFSecureGroup_Create( group, gDefaultConnectionlessProtocolStack, dtl,
                                                          OALConn->connSecurityMembrane, NULL, OALConn->base.OALConn,
                                                          OALConn->connConfig, OALConn->connConfig->permissions )) == NULL)
                {
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALChannelMgr: Failed to create point's secure group (out of memory).");
                    /* Fall through, insufficient resources exception will be filled. */
                    PCR_Assert( !retStruct.ret );
                }
                else
                {
                    retStruct.ret = (OALChannel)secureGroup;
                    if (!OALRemoteChannelMgr_AddChannel(retStruct.ret))
                    {
                        OALChannel_Release(retStruct.ret);
                        retStruct.ret = NULL;
                    }
                    else
                        retStruct.ret->remoteChannelData->flags.isUp = FALSE;
                }
            }
            else
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALChannelMgr: Failed to create point's secure group.");
                if (exception)
                    *exception = DOFException_Copy( DOFEXCEPTION_FAILED );
                PCR_Assert( !retStruct.ret );
            }
        }
        break;
    default:
        if (exception)
            *exception = DOFException_Copy(DOFEXCEPTION_NOT_SUPPORTED);
        PCR_Assert( !retStruct.ret );
        break;
    }

    DTL_Destroy(dtl);       /* The dtl was retained by the channel so release the local copy. */

    /* If there is not a channel return NULL.  If an exception ptr was provided and it is not set, set it now. */
    if (!retStruct.ret)
    {
        OALChannel_Release(baseChannel);
        if (exception && !(*exception))
            *exception = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        return NULL;
    }
    retStruct.ret->remoteChannelData->flags.isOutbound = TRUE;

    /* retStruct.ret now a channel and needed dtl members are filled unless ->relatedServer exists and not started*/
    /* Start related server(s). They will be stopped internally when we remove connection from channel's connections list. */

    PCR_Assert(retStruct.ret->remoteChannelData->flags.isConnection);   /* This must be set for all connections including secure groups. */

    retStruct.ret->remoteChannelData->baseChannel = baseChannel;
    retStruct.ret->remoteChannelData->OALConn = OALConn;

    if (connType != DOFCONNECTIONTYPE_HUB && OALConn->connConfig->domainCreds)
    {
        /* Hub connections currently do not support domain credentials.  See bug 3801.
           This check must be removed as part of adding support for domain credentials to hub connections. */
        retStruct.ret->remoteChannelData->flags.validateDomainCreds = TRUE;
    }

    if ((DOFConnectionConfig_GetBridge(OALConn->connConfig) != NULL) && ((connType == DOFCONNECTIONTYPE_STREAM) || (connType == DOFCONNECTIONTYPE_DATAGRAM)) )
    {
        retStruct.ret->remoteChannelData->flags.createBridge = TRUE;
    }

    if ((connType == DOFCONNECTIONTYPE_STREAM) ||
        (((connType == DOFCONNECTIONTYPE_DATAGRAM_STATELESS) || (connType == DOFCONNECTIONTYPE_DATAGRAM)) && !sharedDTL))
    {
        /* This next indirectly initiates DOFTransportLayer_Opened(), which calls OALRemoteChannel_Init(). The
           latter actually starts up the connection by initiating ESP work, and if secure initiates TRP/TEP. While
           that is happening, we return to our caller, (one of the callers waits on an event). When
           the ESP/TRP/TEP work concludes, that event is signalled to finally wake up our caller. */

        if(!DTL_InitConnectionDTL(dtl, OALConn->dof, retStruct.ret, OALConn->connConfig->targetAddress->transport,
                                  OALConn->connConfig->transportConfig, OALConn->connConfig->targetAddress,
                                  (connType == DOFCONNECTIONTYPE_STREAM)?DOFCONNECTIONTYPE_STREAM:DOFCONNECTIONTYPE_DATAGRAM))
        {
            OALRemoteChannel_ConnectionInitiatedShutdown(retStruct.ret, OALConn);
            OALChannel_Release(retStruct.ret);
            if (exception)
                *exception = DOFException_Copy(DOFEXCEPTION_FAILED);
            return NULL;
        }
    }
    else if (((connType == DOFCONNECTIONTYPE_DATAGRAM_STATELESS) || (connType == DOFCONNECTIONTYPE_DATAGRAM)) && (sharedDTL))
    {
        retStruct.ret->remoteChannelData->flags.isUp = TRUE;
        if (retStruct.ret->remoteChannelData->flags.isConnection)
        {
            /*  If the channel is sharing a DTL then assign an ENP port to identify the channel.  Since the base channel cannot
                be accessed directly here, send a message to the new channel and it will register an ENP port with the base channel. */
            if (OALRemoteChannel_RegisterChannel_ENPPort(retStruct.ret))
            {
                retStruct.ret->remoteChannelData->flags.setDtlInChannelInit = TRUE;
                OALRemoteChannel_Init(retStruct.ret);
            }
            else
            {
                OALRemoteChannel_ConnectionInitiatedShutdown(retStruct.ret, OALConn);
                OALChannel_Release(retStruct.ret);
                if (exception)
                    *exception = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
                retStruct.ret = NULL;
            }
        }
        else
        {
            OALRemoteChannel_Init(retStruct.ret);
        }
    }
    else if (connType == DOFCONNECTIONTYPE_GROUP ||
             connType == DOFCONNECTIONTYPE_GROUP_MEMBER_ONLY ||
             connType == DOFCONNECTIONTYPE_HUB)
    {
        retStruct.ret->remoteChannelData->flags.setDtlInChannelInit = 1;
        OALRemoteChannel_Init( retStruct.ret );
    }
    else
    {
        PCR_Assert( connType == DOFCONNECTIONTYPE_POINT );
        PCR_Assert( dtl );

        /* This next indirectly initiates DOFTransportLayer_Opened(), which calls OALRemoteChannel_Init(). The
           latter actually starts up the connection by initiating TRP/SGMP. While that is happening, we return to
           our caller, (one of the callers waits on an event). When the TRP/SGMP work concludes (including
           KEK/groupKey), that event is signalled to finally wake up our caller. */
        if (!DTL_InitConnectionDTL( dtl, OALConn->dof, retStruct.ret, retStruct.ret->remoteChannelData->address->transport,
                                    OALConn->base.OALConn->connConfig->transportConfig,
                                    retStruct.ret->remoteChannelData->address, DOFCONNECTIONTYPE_DATAGRAM ))
        {
            OALChannel_Release(retStruct.ret);
            OALRemoteChannel_ConnectionInitiatedShutdown(retStruct.ret, OALConn);
            if (exception)
                *exception = DOFException_Copy(DOFEXCEPTION_FAILED);
            retStruct.ret = NULL;
        }
    }

    return retStruct.ret;
}

DOFError OALChannelMgr_CreateChannel_Server( DOFServer srv )
{
    OALServer server = srv->server;

    if( server && !server->transport.channel )
    {
        /* set up channel */
        switch( server->config->type )
        {
        case DOFSERVERTYPE_STREAM:
            {
                /* no channel or stack needed to be associated with stream server. */
                server->transport.dtl = DTL_Create();
                if (!server->transport.dtl)
                {
                    return DOFERROR_INSUFFICIENT_RESOURCES;
                }
                if ( DTL_InitServerDTL( OALServer_GetDTL(server), DOF_GetDof(), server, DOFAddress_GetTransport( server->config->listenAddr ), server->config->serverConfig, server->config->listenAddr, server->config->type, NULL))
                {
                    return DOFERROR_OK;
                }
                else
                {
                    return DOFERROR_INSUFFICIENT_RESOURCES;
                }
            }
            break;
        case DOFSERVERTYPE_DATAGRAM:
            {
                OALChannel      channel = NULL;
                DTL             dtl = DTL_Create();

                if(!dtl)
                    return DOFERROR_INSUFFICIENT_RESOURCES;

                channel = (OALChannel)OALConnectionlessRemoteChannel_Create(gDefaultConnectionlessProtocolStack, dtl, server->securityMembrane, OALServer_GetListenAddr(server), NULL, server->config);
                if (channel)
                {
                    OALChannel      relatedServer;

                    DTL_Destroy(dtl);   /* The dtl was retained by the channel so release the local reference. */

                    /* Get the send server before associating the server and channel.  Otherwise it will return the new channel. */
                    relatedServer = OALServer_GetSendServerChannel( server );
                    channel->remoteChannelData->baseChannel = OALChannel_Retain(relatedServer);

                    server->transport.channel = channel;
                    channel->remoteChannelData->server = server;
                    channel->remoteChannelData->flags.isUp = FALSE;

                    if (channel->remoteChannelData->baseChannel  || !relatedServer)
                    {
                        /* Add the channel to the channel manager, it will fail if we are out of memory. */
                        if(!OALRemoteChannelMgr_AddChannel(channel))
                        {
                            server->transport.channel = NULL;
                            channel->remoteChannelData->server = NULL;
                            OALChannel_Release(channel);
                            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALChannelMgr: Failed to add channel (out of memory).");
                            return DOFERROR_INSUFFICIENT_RESOURCES;
                        }

                        if( !DTL_InitServerDTL( OALServer_GetDTL(server), DOF_GetDof(), server, DOFAddress_GetTransport( server->config->listenAddr ), server->config->serverConfig, server->config->listenAddr, server->config->type, (relatedServer)?relatedServer->remoteChannelData->dtl:NULL ) )
                        {
                            channel->remoteChannelData->server = NULL;
                            server->transport.channel = NULL;
                            OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_FAILED);
                            OALChannel_Release(channel);
                            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALChannelMgr: DTL_InitServerDTL failed.");
                            return DOFERROR_INSUFFICIENT_RESOURCES;
                        }
                    }
                    else
                    {
                        server->transport.channel = NULL;
                        channel->remoteChannelData->server = NULL;
                        OALChannel_Release(channel);
                        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALChannelMgr: Failed to retain related server channel.");
                        return DOFERROR_INSUFFICIENT_RESOURCES;
                    }
                }
                else
                {
                    DTL_Destroy(dtl);
                    return DOFERROR_INSUFFICIENT_RESOURCES;
                }
            }
            break;
        default:
            return DOFERROR_FAILED;
            break;
        }
    }
    return DOFERROR_OK;
}

boolean OALChannelMgr_CreateChannel_StreamServerConnect( DTL dtl, OALServer server, DOFConnectionType connType, DOFAddress address, DTLConnectionConfig config )
{
    OALChannel                  channel = NULL;
    OALConnectedProtocolStack   protoStack = NULL;
    DOFConnectionConfig         connConfig = NULL;
    DOF                         dof       = DOF_GetDof();
    boolean                     status = FALSE;

    if(server->connectionsCount < server->config->MaxConnections) /* OK to call because server is locked already */
    {
        {
            DOFConnectionConfigBuilder ccbuilder = DOFConnectionConfigBuilder_Create( connType, address );

            if(ccbuilder == NULL)
            {
                return FALSE;
            }

/*          DOFConnectionConfigBuilder_SetCredentials( ccbuilder, DOFServerConfig_GetCredentials( server->config ) );*/
            DOFConnectionConfigBuilder_SetMaxReceiveSilence( ccbuilder, DOFServerConfig_GetMaxReceiveSilence( server->config ) );
            DOFConnectionConfigBuilder_SetMaxSendSilence( ccbuilder, DOFServerConfig_GetMaxSendSilence( server->config ) );
/*          DOFConnectionConfigBuilder_SetPermissions( ccbuilder, DOFServerConfig_GetPermissions( server->config ) );*/
            DOFConnectionConfigBuilder_SetProtocolNegotiator( ccbuilder, DOFServerConfig_GetProtocolNegotiator( server->config ) );
            DOFConnectionConfigBuilder_SetSecurityDesire( ccbuilder, DOFServerConfig_GetSecurityDesire( server->config ) );
            DOFConnectionConfigBuilder_SetReceiveFilter( ccbuilder, DOFServerConfig_GetReceiveFilter( server->config ) );
            DOFConnectionConfigBuilder_SetSendFilter( ccbuilder, DOFServerConfig_GetSendFilter( server->config ) );
            DOFConnectionConfigBuilder_SetBridge( ccbuilder, DOFServerConfig_GetBridge( server->config ) );
            DOFConnectionConfigBuilder_SetTransportConfig( ccbuilder, config );

            connConfig = DOFConnectionConfigBuilder_BuildAndDestroy( ccbuilder );
            if(server->config->dcreds)
                connConfig->domainCreds = OALDomainCredentialsTree_Copy(server->config->dcreds);
        }

        /* set up channel */
        if (connConfig)
        {
            protoStack = OALConnectedProtocolStack_Create( DOFConnectionConfig_GetAddress(connConfig) );
            if (protoStack)
            {
                channel = (OALChannel)OALConnectedRemoteChannel_Create( protoStack, dtl, server->securityMembrane,
                                                                        connConfig, FALSE );
                if (channel)
                {
                    channel->remoteChannelData->flags.createBridge = TRUE;
                    DTL_SetChannel( dtl, channel );
                    OALConnectedProtocolStack_SetChannel( (OALConnectedProtocolStack)protoStack, (OALConnectedRemoteChannel)channel );
                    if(OALRemoteChannelMgr_AddChannel(channel))
                    {
                        if (OALRemoteChannelMgr_AddProtocolStack( (OALProtocolStack)protoStack ))
                        {
                            DOFServer theServer = OALServer_CreateDOFServer(server);
                            if (theServer)
                            {
                                if (DOFConnection_CreateConnectedInboundConnection(dof, server->securityMembrane, connConfig, theServer, channel, FALSE))
                                {
                                    status = TRUE;
                                }
                                else
                                {
                                    OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_FAILED);
                                }
                                DOFServer_Destroy(theServer);
                            }
                            else
                            {
                                OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_FAILED);
                            }
                        }
                        else
                        {
                            OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_FAILED);
                        }
                    }
                }
            }
        }

        /* In the case that everything is sucessfull the remote channel manager and the connection will hold a reference to the channel.
           When the channel is shut down it will notify (and be removed from) the connection and remote channel manager. */
        DOFConnectionConfig_Destroy(connConfig);
        OALChannel_Release( channel );
        OALProtocolStack_Release( (OALProtocolStack)protoStack );
    }
    else
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ALERT, "A connection to a server was rejected because the max number of connections to the server has already been reached.");
    }

    return status;
}

OALChannel OALChannelMgr_CreateChannel_ConnectionlessInbound(OALChannel serverChannel, DOFConnectionType connType,
                                                 DOFAddress address, DOFCredentials creds, DOFPermissionSet perms)
{
    OALChannel                  channel = NULL;
    DOFConnectionConfigBuilder  cb;
    DOFConnectionConfig         connConfig = NULL;
    boolean                     status = FALSE;

    PCR_Assert(serverChannel);

    /* Build the connectionConfig from the server channel and paramaters passed in. */
    cb = DOFConnectionConfigBuilder_Create(connType, address);

    if (cb)
    {
        DTLConnectionConfig DTLConfig;
        if (DOFTransportLayer_CreateConnectionConfig(serverChannel->remoteChannelData->dtl, address, &DTLConfig))
        {
            DOFConnectionConfigBuilder_SetTransportConfig(cb, DTLConfig);
            DTLConnectionConfig_Destroy(DTLConfig);
            DOFConnectionConfigBuilder_SetSecurityDesire(cb, DOFServerConfig_GetSecurityDesire(serverChannel->remoteChannelData->config.server));
            DOFConnectionConfigBuilder_SetPermissions(cb, perms);
            DOFConnectionConfigBuilder_SetProtocolNegotiator(cb, DOFServerConfig_GetProtocolNegotiator(serverChannel->remoteChannelData->config.server));
            DOFConnectionConfigBuilder_SetMaxSendSilence(cb, DOFServerConfig_GetMaxSendSilence(serverChannel->remoteChannelData->config.server));
            DOFConnectionConfigBuilder_SetMaxReceiveSilence(cb, DOFServerConfig_GetMaxReceiveSilence(serverChannel->remoteChannelData->config.server));
            DOFConnectionConfigBuilder_SetReceiveFilter(cb, DOFServerConfig_GetReceiveFilter(serverChannel->remoteChannelData->config.server));
            DOFConnectionConfigBuilder_SetSendFilter(cb, DOFServerConfig_GetSendFilter(serverChannel->remoteChannelData->config.server));
            DOFConnectionConfigBuilder_SetBridge(cb, DOFServerConfig_GetBridge(serverChannel->remoteChannelData->config.server));
            if (creds)
            {
                DOFConnectionConfigBuilder_SetCredentials(cb, creds);
            }

            connConfig = DOFConnectionConfigBuilder_BuildAndDestroy(cb);
        }
        else
        {
            DOFConnectionConfigBuilder_Destroy(cb);
        }
    }
    if (connConfig)
    {
        DOFServer theServer = OALServer_CreateDOFServer(serverChannel->remoteChannelData->server);

        if (theServer)
        {
            if(DOFServer_IsConnectionAllowed(theServer))
            {
                channel = (OALChannel)OALConnectionlessRemoteChannel_Create(gDefaultConnectionlessProtocolStack,
                                        serverChannel->remoteChannelData->dtl, serverChannel->securityMembrane,
                                        address, connConfig, NULL);
                if (channel)
                {
                    if(OALRemoteChannelMgr_AddChannel(channel))
                    {
                        if (serverChannel->remoteChannelData->baseChannel)
                            channel->remoteChannelData->baseChannel = OALChannel_Retain(serverChannel->remoteChannelData->baseChannel);
                        else
                            channel->remoteChannelData->baseChannel = OALChannel_Retain(serverChannel);

                        if (channel->remoteChannelData->baseChannel)
                        {
                            /* Set the ENP port for the channel to identify the session. */
                            if (OALRemoteChannel_RegisterChannel_ENPPort_Immediate(channel))
                            {
                                if (OALRemoteChannelMgr_StartNegotiationTimeout(channel, ESP_V2_NEGOTIATION_TIMEOUT))
                                {
                                    if (DOFConnection_CreateConnectedInboundConnection(DOF_GetDof(), serverChannel->securityMembrane,
                                                                                       connConfig, theServer, channel, TRUE))
                                    {
                                        /* An inbound datagram connection is not up until it receives a datagram following the open. */
                                        channel->remoteChannelData->flags.waitingForInitialDatagram = TRUE;
                                        channel->remoteChannelData->flags.lossySession = TRUE;
                                        status = TRUE;
                                    }
                                }
                            }
                        }

                        if (status == FALSE)
                        {
                            OALRemoteChannel_DeregisterChannel_ENPPort_Immediate(channel);
                            OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_PROTOCOL_NEGOTIATION_FAILED);
                            OALRemoteChannelMgr_CancelNegotiationTimer(channel);
                        }
                    }
                }
            }
            else
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ALERT,"A connection to a server was rejected because the max number of connections to the server has already been reached.");
            }
            DOFServer_Destroy(theServer);
        }
        else
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALChannelMgr: Failed to create a server in OALChannelMgr_CreateChannel_ConnectionlessInbound.");
        }
    }

    DOFConnectionConfig_Destroy(connConfig);

    if (status)
    {
        return channel;
    }
    else
    {
        OALChannel_Release(channel);
        return NULL;
    }
}

boolean OALChannel_IsSecureGroup( OALChannel self )
{
    PCR_Assert( self );
    if (self->remoteChannelData && self->remoteChannelData->flags.isConnection)
    {
        DOFConnectionType connType = DOFConnectionConfig_GetConnectionType(self->remoteChannelData->config.connection);

        if (connType == DOFCONNECTIONTYPE_GROUP || connType == DOFCONNECTIONTYPE_GROUP_MEMBER_ONLY ||
            connType == DOFCONNECTIONTYPE_HUB || connType == DOFCONNECTIONTYPE_POINT )
           return TRUE;
    }

    return FALSE;
}

OALChannel OALChannel_GetPrimarySessionChannel( OALChannel self )
{
    if ( self && self->remoteChannelData && self->remoteChannelData->baseChannel )
    {
        if ( !self->remoteChannelData->flags.isConnection || self->remoteChannelData->config.connection->type == DOFCONNECTIONTYPE_DATAGRAM_STATELESS )
            return self->remoteChannelData->baseChannel;
    }
    return self;
}

int OALChannel_Compare( OALChannel self, OALChannel channel )
{
    OALChannel c1 = OALChannel_GetPrimarySessionChannel( self );
    OALChannel c2 = OALChannel_GetPrimarySessionChannel( channel );
    if ( c1 < c2 )
        return -1;
    if ( c1 > c2 )
        return 1;
    return 0;
}
