/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/bridge.h>
#include <dof/filter.h>
#include <internal/channel.h>
#include <internal/filter_internal.h>
#include <internal/security_membrane.h>

typedef struct ChannelMembrane_t
{
    DOFSecurityMembrane_t super;

    DOF dof;
    DOFOperationFilter receiveFilter;
    DOFOperationFilter sendFilter;
    OALOperationBridge bridge;
} ChannelMembrane_t;

typedef ChannelMembrane_t * ChannelMembrane;

static void ChannelMembrane_Dtor(DOFSecurityMembrane self);

static DOFSecurityScope ChannelMembrane_ValidateInboundPermissions(DOFSecurityMembrane self,
                                                                   DOFPermission       explicitPermission,
                                                                   DOFPermission       logicalPermission,
                                                                   OALCipherKeyState   keyState);

static DOFSecurityScope ChannelMembrane_ValidateOutboundPermissions(DOFSecurityMembrane self,
                                                                    DOFPermission       explicitPermission,
                                                                    DOFPermission       logicalPermission);

static DOFSecurityScope ChannelMembrane_GetInboundScope(DOFSecurityMembrane self);

static DOFSecurityScope ChannelMembrane_GetOutboundScope(DOFSecurityMembrane self);

static DOFSecurityMembraneExtendReturn_t ChannelMembrane_ExtendInboundPerms(DOFSecurityMembrane        self,
                                                                            DOFPermissionSet           extendPerms,
                                                                            ExtendMembraneCompleteCB   callback,
                                                                            void                     * context);

static DOFSecurityMembraneExtendReturn_t ChannelMembrane_ExtendOutboundPerms(DOFSecurityMembrane        self,
                                                                             DOFPermissionSet           extendPerms,
                                                                             ExtendMembraneCompleteCB   callback,
                                                                             void                     * context);

static DOFObjectIDDomain ChannelMembrane_GetDomainId(DOFSecurityMembrane self);

static DOFPermissionSet ChannelMembrane_GetInboundPerms(DOFSecurityMembrane self);

static DOFPermissionSet ChannelMembrane_GetOutboundPerms(DOFSecurityMembrane self);

static struct DOFSecurityMembrane_VTable_t vtable =
{
    ChannelMembrane_Dtor,                       ChannelMembrane_GetDomainId,
    ChannelMembrane_ValidateInboundPermissions, ChannelMembrane_ValidateOutboundPermissions,
    ChannelMembrane_GetInboundScope,            ChannelMembrane_GetOutboundScope,
    ChannelMembrane_ExtendInboundPerms,         ChannelMembrane_ExtendOutboundPerms,
    ChannelMembrane_GetInboundPerms,            ChannelMembrane_GetOutboundPerms
};

DOFSecurityMembrane OALChannel_CreateSecurityMembrane(DOF                dof,
                                                      DOFOperationFilter receiveFilter,
                                                      DOFOperationFilter sendFilter,
                                                      OALOperationBridge bridge)
{
    ChannelMembrane retVal;

    PCR_Assert(receiveFilter != NULL);
    PCR_Assert(sendFilter != NULL);
    PCR_Assert(bridge != NULL);

    retVal = (ChannelMembrane) DOFMem_Alloc(sizeof(ChannelMembrane_t));
    if(retVal != NULL)
    {
        retVal->receiveFilter = DOFOperationFilter_Copy(receiveFilter);
        if(retVal->receiveFilter != NULL)
        {
            retVal->sendFilter = DOFOperationFilter_Copy(sendFilter);
            if(retVal->sendFilter != NULL)
            {
                retVal->bridge = OALOperationBridge_Copy(bridge);
                if(retVal->bridge != NULL)
                {
                    retVal->dof = dof;
                    DOFSecurityMembrane_Ctor(&(retVal->super), &vtable);
                    return (DOFSecurityMembrane) retVal;
                }
                DOFOperationFilter_Destroy(retVal->sendFilter);
            }
            DOFOperationFilter_Destroy(retVal->receiveFilter);
        }
        DOFMem_Free(retVal);
    }
    return NULL;
}

boolean DOFSecurityMembrane_IsChannelMembrane(DOFSecurityMembrane self)
{
    PCR_Assert(self != NULL);
    return (self->vtable == &vtable);
}

OALOperationBridge OALChannelMembrane_GetBridge(DOFSecurityMembrane self)
{
    PCR_Assert(self != NULL);

    return ((ChannelMembrane) self)->bridge;
}

boolean OALChannelMembrane_SetBridge(DOFSecurityMembrane self,
                                     OALOperationBridge  newBridge)
{
    ChannelMembrane membrane = (ChannelMembrane) self;
    OALOperationBridge bridge;

    PCR_Assert(membrane != NULL);
    bridge = OALOperationBridge_Copy(newBridge);
    if(bridge != NULL)
    {
        OALOperationBridge_Destroy(membrane->bridge);
        membrane->bridge = bridge;
        return TRUE;
    }
    return FALSE;
}

static void ChannelMembrane_Dtor(DOFSecurityMembrane self)
{
    ChannelMembrane selfx = (ChannelMembrane) self;

    DOFOperationFilter_Destroy(selfx->receiveFilter);
    DOFOperationFilter_Destroy(selfx->sendFilter);
    OALOperationBridge_Destroy(selfx->bridge);
    DOFMem_Free(self);
}

static DOFSecurityScope ChannelMembrane_ValidateInboundPermissions(DOFSecurityMembrane self,
                                                                   DOFPermission       explicitPermission,
                                                                   DOFPermission       logicalPermission,
                                                                   OALCipherKeyState   keyState)
{
    ChannelMembrane selfx = (ChannelMembrane) self;

    if(DOFOperationFilter_IsAllowed(selfx->receiveFilter, logicalPermission))
    {
        return DOFSecurityMembrane_ValidateInboundPermissions((DOFSecurityMembrane)((ChannelMembrane)self)->bridge, explicitPermission, logicalPermission, keyState);
    }
    return DOFSecurityScope_Create(DOFSecurityMembrane_GetDomainID((DOFSecurityMembrane) selfx->bridge), 0, NULL, 0, NULL);
}

static DOFSecurityScope ChannelMembrane_ValidateOutboundPermissions(DOFSecurityMembrane self,
                                                                    DOFPermission       explicitPermission,
                                                                    DOFPermission       logicalPermission)
{
    ChannelMembrane selfx = (ChannelMembrane) self;

    if(DOFOperationFilter_IsAllowed(selfx->sendFilter, logicalPermission))
    {
        return DOFSecurityMembrane_ValidateOutboundPermissions((DOFSecurityMembrane)((ChannelMembrane)self)->bridge, explicitPermission, logicalPermission);
    }
    return DOFSecurityScope_Create(DOFSecurityMembrane_GetDomainID((DOFSecurityMembrane) selfx->bridge), 0, NULL, 0, NULL);
}

static DOFSecurityScope ChannelMembrane_GetInboundScope(DOFSecurityMembrane self)
{
    return DOFSecurityMembrane_GetInboundScope((DOFSecurityMembrane)((ChannelMembrane)self)->bridge);
}

static DOFSecurityScope ChannelMembrane_GetOutboundScope(DOFSecurityMembrane self)
{
    return DOFSecurityMembrane_GetOutboundScope((DOFSecurityMembrane)((ChannelMembrane)self)->bridge);
}

static DOFSecurityMembraneExtendReturn_t ChannelMembrane_ExtendInboundPerms(DOFSecurityMembrane        self,
                                                                               DOFPermissionSet           extendPerms,
                                                                               ExtendMembraneCompleteCB   callback,
                                                                               void                     * context)
{
    return DOFSecurityMembrane_ExtendInboundPermissions((DOFSecurityMembrane)((ChannelMembrane)self)->bridge,
                                                        extendPerms, callback, context);
}

static DOFSecurityMembraneExtendReturn_t ChannelMembrane_ExtendOutboundPerms(DOFSecurityMembrane        self,
                                                                                DOFPermissionSet           extendPerms,
                                                                                ExtendMembraneCompleteCB   callback,
                                                                                void                     * context)
{
    return DOFSecurityMembrane_ExtendOutboundPermissions((DOFSecurityMembrane)((ChannelMembrane)self)->bridge,
                                                         extendPerms, callback, context);
}

static DOFObjectIDDomain ChannelMembrane_GetDomainId(DOFSecurityMembrane self)
{
    return DOFSecurityMembrane_GetDomainID((DOFSecurityMembrane)((ChannelMembrane)self)->bridge);
}

static DOFPermissionSet ChannelMembrane_GetInboundPerms(DOFSecurityMembrane self)
{
    return DOFSecurityMembrane_GetInboundPermissions((DOFSecurityMembrane)((ChannelMembrane)self)->bridge);
}

static DOFPermissionSet ChannelMembrane_GetOutboundPerms(DOFSecurityMembrane self)
{
    return DOFSecurityMembrane_GetOutboundPermissions((DOFSecurityMembrane)((ChannelMembrane)self)->bridge);
}
