/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/bridge.h>
#include <dof/exception.h>
#include <internal/securityscope.h>

#include <internal/blackboarddata.h>
#include <internal/dofinternal.h>
#include <internal/channel_stack_messages.h>
#include <internal/connection_internal.h>
#include <internal/operationinternal.h>
#include <internal/permissionsetinternal.h>
#include <internal/remotechannel.h>
#include <internal/resultinternal.h>
#include <internal/serverinternal.h>

#define ROUTE_TO_PROVIDER_ARGC      4
#define ROUTE_TO_REQUESTOR_ARGC     4

/* CMT_RouteToProvider **********************************************/
static void CMT_RouteToProvider_Dtor( OALMessage msg )
{
    PCR_Assert( msg );

    DOFOperation_Internal_Destroy( (DOFOperation)msg->argv[0] );
    DOFOperation_Internal_Destroy( (DOFOperation)msg->argv[1] );
    DOFAddress_Destroy( (DOFAddress)msg->argv[2] );
}

OALMessage CMT_RouteToProvider_Create( void* recipient, DOFOperation req_op, DOFOperation prov_op, DOFAddress addr )
{
    OALMessage retVal = OALMessage_Create(ROUTE_TO_PROVIDER_ARGC, recipient,
                                          CMT_ROUTE_TO_PROVIDER,
                                          CMT_RouteToProvider_Dtor);
    /* Note: The message is created so that it is big enough for the SMT_ROUTE_TO_PROVIDER to re-use it, so
       the size and order must match. */

    if (retVal)
    {
        DOFOperation_Retain( req_op );
        retVal->argv[0] = req_op;
        if (prov_op)
            DOFOperation_Retain( prov_op );
        retVal->argv[1] = prov_op;
        retVal->argv[2] = DOFAddress_Copy( addr );
        if (!retVal->argv[2] && addr)
        {
            OALMessage_Destroy(retVal);
            retVal = NULL;
        }
    }
   return retVal;
}
/* *******************************************************************/

/* CMT_RouteToRequestor **********************************************/
static void CMT_RouteToRequestor_Dtor( OALMessage msg )
{
    PCR_Assert( msg );

    DOFOperation_Internal_Destroy( (DOFOperation)msg->argv[0] );
    DOFResult_Destroy( (DOFResult)msg->argv[1] );
    if(msg->argv[2])
        DOFAddress_Destroy( (DOFAddress)msg->argv[2] );
}

OALMessage CMT_RouteToRequestor_Create( void* recipient, DOFOperation resp_op, DOFResult result )
{
    DOFAddress queuedAddress = NULL;
    OALMessage retVal = OALMessage_Create(ROUTE_TO_REQUESTOR_ARGC, recipient,
                                          CMT_ROUTE_TO_REQUESTOR,
                                          CMT_RouteToRequestor_Dtor);
    if (retVal)
    {
        DOFOperation_Retain( resp_op );
        retVal->argv[0] = resp_op;
        retVal->argv[1] = DOFResult_Copy( result );
        if (retVal->argv[1]  || !result)
        {
            /* this is needed by operations: The address must be queued before control is returned to the caller. */
            if (resp_op)
                queuedAddress = DOFAddress_Copy( resp_op->sources->address );

            retVal->argv[2] = queuedAddress;
            if (retVal->argv[2] || !queuedAddress)
            {
                return retVal;
            }
        }
        OALMessage_Destroy(retVal);
    }

    return NULL;
}
/* *******************************************************************/

/* CMT_ReceiveFromTransport **********************************************/
static void CMT_ReceiveFromTransport_Dtor( OALMessage msg )
{
    PCR_Assert( msg );

    if (msg->argv[0])
        OALStream_Release( (OALStream)msg->argv[0], TRUE );
    if (msg->argv[1])
        DOFAddress_Destroy( (DOFAddress)msg->argv[1] );
    if (msg->argv[3])
        OALChannel_Release( (OALChannel)msg->argv[3] );
}

OALMessage CMT_ReceiveFromTransport_Create( void* recipient, OALStream data, DOFAddress rcvdFrom, boolean isMulticast )
{
    OALMessage retVal = OALMessage_Create( 4, recipient,
                                           CMT_RECEIVE_FROM_TRANSPORT,
                                           CMT_ReceiveFromTransport_Dtor);
    if (retVal)
    {
        OALStream_Retain( data );
        retVal->argv[0] = data;
        retVal->argv[1] = DOFAddress_Copy( rcvdFrom );
        if (retVal->argv[1] || !rcvdFrom)
        {
            retVal->argv[2] = (void*)(int)isMulticast;
            retVal->argv[3] = NULL; /* optimization: this arg will be used by the channel before it is passed on to the stack. */
        }
        else
        {
            OALMessage_Destroy(retVal);
            retVal = NULL;
        }
    }
    return retVal;
}
/* *******************************************************************/

/* CMT_Initialize_Create **********************************************/
OALMessage CMT_Initialize_Create(void* recipient)
{
    OALMessage retVal = OALMessage_Create( 1, recipient,
                                           CMT_INITIALIZE,
                                           NULL);
    if (retVal)
    {
        retVal->argv[0] = NULL; /* to be set by the channel if needed */
    }
    return retVal;
}
/* *******************************************************************/

/* CMT_InitializeComplete_Create **********************************************/
OALMessage CMT_InitializeComplete_Create(void* recipient)
{
    OALMessage retVal = OALMessage_Create( 1, recipient,
                                           CMT_INITIALIZE_COMPLETE,
                                           NULL);

    if (retVal)
    {
        /* The arg will not be used here, but is added to allow re-use of the message. */
        retVal->argv[0] = NULL;
    }

    return retVal;
}
/* *******************************************************************/

static void CMT_Shutdown_Dtor(OALMessage msg)
{
    DOFException_Destroy((DOFException)msg->argv[0]);
}

/* CMT_Shutdown_Create **********************************************/
OALMessage CMT_Shutdown_Create(void* recipient, DOFException exception)
{
    OALMessage retVal = OALMessage_Create( 1, recipient,
                                           CMT_SHUTDOWN,
                                           CMT_Shutdown_Dtor );
    if (retVal)
    {
        retVal->argv[0] = DOFException_Copy(exception);
    }
    return retVal;
}

void CMT_Shutdown_SetException(OALMessage msg, DOFException exception)
{
    if (msg && (msg->argv[0] == NULL))
    {
        msg->argv[0] = DOFException_Copy(exception);
    }
}
/* *******************************************************************/

static void CMT_TransportDown_Dtor(OALMessage msg)
{
    DOFException_Destroy((DOFException)msg->argv[1]);
}

/* CMT_TransportDown_Create **********************************************/
OALMessage CMT_TransportDown_Create(void* recipient, DTL transport, DOFException exception)
{
    OALMessage retVal = OALMessage_Create( 2, recipient,
                                           CMT_TRANSPORTDOWN,
                                           CMT_TransportDown_Dtor );
    if (retVal)
    {
        retVal->argv[0] = (void*)transport;
        retVal->argv[1] = DOFException_Copy(exception);
    }
    return retVal;
}
/* *******************************************************************/

/* CMT_ProcessBlockedMessage_Create *********************************/
OALMessage CMT_ProcessBlockedMessage(void * recipient, boolean clearSendBlock)
{
    OALMessage retVal = OALMessage_Create( 2, recipient,
                                           CMT_PROCESS_BLOCKED_MESSAGE,
                                           NULL);

    if (retVal)
    {
        retVal->argv[0] = (void *)(int)clearSendBlock;
    }
    return retVal;
}
/* ******************************************************************/

static void CMT_ConnectionInitiatedShutdown_Dtor( OALMessage msg )
{
    OALChannel_Release((OALChannel)msg->recipient);
    OALConnection_Release((OALConnection)msg->argv[0]);
}

/* CMT_ConnectionInitiatedShutdown_Create *********************************/
OALMessage CMT_ConnectionInitiatedShutdown_Create(void * recipient, OALConnection OALConn)
{
    OALMessage retVal = OALMessage_Create( 1, recipient,
                                           CMT_CONNECTION_INITIATED_SHUTDOWN,
                                           CMT_ConnectionInitiatedShutdown_Dtor);

    if (retVal)
    {
        /* Retain the channel so that it will not get destroyed until after this message is handled. */
        retVal->recipient = OALChannel_Retain((OALChannel)recipient);
        if (retVal->recipient)
        {
            /* Retain the connection so that it will not get destroyed until the channel is shutdown. */
            if(OALConnection_Retain(OALConn))
            {
                retVal->argv[0] = (void *)OALConn;
            }
            else
            {
                retVal->argv[0] = NULL;
                OALMessage_Destroy(retVal);
                retVal = NULL;
            }
        }
        else
        {
            OALMessage_Destroy(retVal);
            retVal = NULL;
        }
    }
    return retVal;
}

OALMessage CMT_RegisterSecureChannel_Create(void * recipient, uint32 SSID, boolean AS)
{
    if (recipient)
    {
        OALMessage retVal = OALMessage_Create(2, recipient,
                                              CMT_REGISTER_SECURE_CHANNEL,
                                              NULL);

        if (retVal)
        {
            retVal->argv[0] = (void *)SSID;
            retVal->argv[1] = (void *)(uint32)AS;
        }
        return retVal;
    }
    return NULL;
}

OALMessage CMT_APPIDQueryTimeout_Create(void * recipient)
{
    if (recipient)
    {
        return OALMessage_Create(0, recipient, CMT_APPID_QUERY_TIMEOUT, NULL);
    }
    return NULL;
}

OALMessage CMT_RegisterChannel_ENPPort_Create(void * recipient)
{
    if (recipient)
    {
        return OALMessage_Create(0, recipient, CMT_REGISTER_ENPPORT_CHANNEL, NULL);
    }
    return NULL;
}

static void CMT_RemoteDomainCallback_Dtor( OALMessage msg )
{
    PCR_Assert(msg);

    DOFCredentials_Destroy((DOFCredentials)msg->argv[0]);
}

OALMessage CMT_RemoteDomainCallback_Create(void *recipient, DOFCredentials creds)
{
    OALMessage msg = OALMessage_Create(1 ,recipient, CMT_REMOTE_DOMAIN_CALLBACK, CMT_RemoteDomainCallback_Dtor);

    if (msg)
    {
        msg->argv[0] = DOFCredentials_Copy(creds);
        if (!msg->argv[0] && creds)
        {
            OALMessage_Destroy(msg);
            msg = NULL;
        }
    }
    return msg;
}

static void CMT_ResolveCredsCallback_Dtor( OALMessage msg )
{
    PCR_Assert(msg);

    DOFCredentials_Destroy((DOFCredentials)msg->argv[0]);
}

OALMessage CMT_ResolveCredsCallback_Create(void *recipient, DOFCredentials creds)
{
    OALMessage msg = OALMessage_Create(1 ,recipient, CMT_RESOLVE_CREDS_CALLBACK, CMT_ResolveCredsCallback_Dtor);

    if (msg)
    {
        msg->argv[0] = DOFCredentials_Copy(creds);
        if (!msg->argv[0] && creds)
        {
            OALMessage_Destroy(msg);
            msg = NULL;
        }
    }
    return msg;
}

static void CMT_GetLibraryPermsCallback_Dtor( OALMessage msg )
{
    PCR_Assert(msg);

    DOFPermissionSet_Destroy((DOFPermissionSet)msg->argv[0]);
    source_release((sid_t *)msg->argv[1]);
}

OALMessage CMT_GetLibraryPermsCallback_Create(void *recipient, DOFPermissionSet perms, sid_t * secureSid)
{
    OALMessage msg = OALMessage_Create(2, recipient, CMT_GET_LIBRARY_PERMS_CALLBACK, CMT_GetLibraryPermsCallback_Dtor);

    if (msg)
    {
        msg->argv[0] = DOFPermissionSet_Copy(perms);
        if (!perms || msg->argv[0])
        {
            source_retain(secureSid);
            msg->argv[1] = secureSid;
        }
        else
        {
            msg->argv[1] = NULL;
        }
    }
    return msg;
}

static void CMT_ServerInitiatedShutdown_Dtor( OALMessage msg )
{
    PCR_Assert(msg);

    OALChannel_Release((OALChannel)msg->recipient);
    OALServer_Destroy((OALServer)msg->argv[0]);
}

OALMessage CMT_ServerInitiatedShutdown_Create(void *recipient, OALServer oalSrv)
{
    OALMessage msg = OALMessage_Create(1, recipient, CMT_SERVER_INITIATED_SHUTDOWN, CMT_ServerInitiatedShutdown_Dtor);

    PCR_Assert(oalSrv);
    if (msg)
    {
        msg->recipient = OALChannel_Retain((OALChannel)recipient);
        if (msg->recipient)
        {
            msg->argv[0] = OALServer_Retain(oalSrv);
            if (msg->argv[0])
            {
                return msg;
            }
        }
        OALMessage_Destroy(msg);
    }

    return NULL;
}

OALMessage CMT_ProcessReceiveSilence_Create(void *recipient)
{
    if (recipient)
    {
        return OALMessage_Create(0, recipient, CMT_PROCESS_RECEIVE_SILENCE, NULL);
    }
    return NULL;
}

static void CMT_SetDomainCredentials_Dtor( OALMessage msg )
{
    PCR_Assert(msg);
    OALRemoteChannel_SetCredentialsCallback_ContextDtor(msg->argv[0]);
    DOFCredentials_Destroy((DOFCredentials)(msg->argv[1]));
}

OALMessage CMT_SetDomainCredentials_Create(void *recipient, DOFCredentials creds, void *context)
{
    OALMessage msg;

    PCR_Assert(recipient);

    msg = OALMessage_Create(2, recipient, CMT_SET_DOMAIN_CREDENTIALS_CALLBACK, CMT_SetDomainCredentials_Dtor);
    if (msg)
    {
        msg->argv[0] = context;
        msg->argv[1] = DOFCredentials_Copy(creds);
        if (!msg->argv[1] && creds)
        {
            msg->argv[0] = NULL;
            OALMessage_Destroy(msg);
            msg = NULL;
        }
    }
    return msg;
}

static void CMT_CreateBridge_Dtor( OALMessage message )
{
    OALMessage_CreateBridgeMessage msg = (OALMessage_CreateBridgeMessage)message;

    PCR_Assert(msg);

    DOFOperationBridgeConfig_Destroy(msg->bridgeConfig);
}

OALMessage CMT_CreateBridge_Create(void *recipient, DOFOperationBridgeConfig bridgeConfig, DOF dof)
{
    OALMessage_CreateBridgeMessage reqMsg;

    PCR_Assert(recipient);

    reqMsg = (OALMessage_CreateBridgeMessage)DOFMem_Alloc(sizeof(OALMessage_CreateBridgeMessage_t));

    if (reqMsg)
    {
        OALMessage_Init(&reqMsg->message, recipient, CMT_CREATEBRIDGE, 0, CMT_CreateBridge_Dtor);

        reqMsg->bridgeConfig = DOFOperationBridgeConfig_Copy(bridgeConfig);
        if (reqMsg->bridgeConfig || !bridgeConfig)
        {
            reqMsg->dof = dof;
            return (OALMessage)reqMsg;
        }
        DOFMem_Free(reqMsg);
    }
    return NULL;
}

static void CMT_CreateBridgeComplete_Dtor( OALMessage msg )
{
    PCR_Assert(msg);
    OALOperationBridge_Destroy(((OALMessage_CreateBridgeCompleteMessage)msg)->bridge);
}

OALMessage CMT_CreateBridgeComplete_Create(void *recipient, OALOperationBridge bridge)
{
    OALMessage_CreateBridgeCompleteMessage msg;

    PCR_Assert(recipient);

    msg = (OALMessage_CreateBridgeCompleteMessage)DOFMem_Alloc(sizeof(OALMessage_CreateBridgeCompleteMessage_t));

    if (msg)
    {
        OALMessage_Init(&msg->message, recipient, CMT_CREATEBRIDGE_COMPLETE, 0, CMT_CreateBridgeComplete_Dtor);

        msg->bridge = OALOperationBridge_Copy(bridge);
        if (msg->bridge)
        {
            return (OALMessage)msg;
        }
        DOFMem_Free(msg);
    }
    return NULL;
}

static void CMT_PeerListenerEvent_Dtor( OALMessage msg )
{
    OLAMessage_PeerEventMessage peerMsg = (OLAMessage_PeerEventMessage)msg;

    PCR_Assert(peerMsg);
    DOFAddress_Destroy(peerMsg->address);
    DOFSecurityScope_Destroy(peerMsg->scope);
}

OALMessage CMT_PeerListenerEvent_Create( void *recipient, DOFAddress addr, DOFSecurityScope scope, boolean added)
{
    OLAMessage_PeerEventMessage peerMsg;

    PCR_Assert(recipient);

    peerMsg = (OLAMessage_PeerEventMessage)DOFMem_Alloc(sizeof(OLAMessage_PeerEventMessage_t));

    if (peerMsg)
    {
        OALMessage_Init(&peerMsg->message, recipient, CMT_PEER_LISTENER_EVENT, 0, CMT_PeerListenerEvent_Dtor);

        peerMsg->added = added;
        peerMsg->address = DOFAddress_Copy(addr);
        if (peerMsg->address)
        {
            peerMsg->scope = DOFSecurityScope_Copy(scope);
            if (!added || peerMsg->scope)
            {
                return (OALMessage)peerMsg;
            }
            DOFAddress_Destroy(peerMsg->address);
        }
        DOFMem_Free(peerMsg);
    }
    return NULL;
}

static void CMT_ConnectionStreamRequest_Dtor( OALMessage msg )
{
    OALMessage_StreamRequestMessage reqMsg = (OALMessage_StreamRequestMessage)msg;

    PCR_Assert(reqMsg);
    DOFAddress_Destroy(reqMsg->addr);
}

OALMessage CMT_ConnectionStreamRequest_Create( void *recipient, DOFAddress addr, boolean complete )
{
    OALMessage_StreamRequestMessage reqMsg;

    PCR_Assert(recipient);

    reqMsg = (OALMessage_StreamRequestMessage)DOFMem_Alloc(sizeof(OALMessage_StreamRequestMessage_t));

    if (reqMsg)
    {
        OALMessage_Init(&reqMsg->message, recipient, CMT_CONNECTION_STREAM_REQUEST, 0, CMT_ConnectionStreamRequest_Dtor);

        reqMsg->addr = DOFAddress_Copy(addr);
        if (reqMsg->addr || !addr)
        {
            reqMsg->complete = complete;
            return (OALMessage)reqMsg;
        }
        DOFMem_Free(reqMsg);
    }
    return NULL;
}

/* ******************************************************************/

void SMT_BeginNegotiationAuthorization_Dtor( OALMessage msg )
{
    PCR_Assert( msg );

    OALChannel_Release((OALChannel)msg->argv[0]);
}

/* SMT_RouteToProvider **********************************************/
static void SMT_RouteToProvider_Dtor( OALMessage msg )
{
    PCR_Assert( msg );

    DOFOperation_Internal_Destroy( (DOFOperation)msg->argv[0] );
    DOFOperation_Internal_Destroy( (DOFOperation)msg->argv[1] );

    if (msg->argv[2])
        DOFAddress_Destroy( (DOFAddress)msg->argv[2] );

    if (msg->argv[3])
        OALChannel_Release( (OALChannel)msg->argv[3] );
}

OALMessage SMT_RouteToProvider_Convert( OALMessage message, void * recipient, OALChannel associatedChannel, DOFAddress address )
{
    PCR_Assert(message->msg_type == CMT_ROUTE_TO_PROVIDER);

    if (associatedChannel)
    {
        message->recipient = recipient;
        message->msg_type = SMT_ROUTE_TO_PROVIDER;
        message->dtor = SMT_RouteToProvider_Dtor;

        if (address != message->argv[2])
        {
            DOFAddress_Destroy((DOFAddress)message->argv[2]);
            message->argv[2] = DOFAddress_Copy(address);
        }
        if (!address || (address && message->argv[2]))
        {
            message->argv[3] = OALChannel_Retain(associatedChannel);
            if (message->argv[3])
            {
                return message;
            }
        }
    }

    OALMessage_Destroy(message);
    return NULL;
}
/* *******************************************************************/

/* SMT_RouteToRequestor **********************************************/
static void SMT_RouteToRequestor_Dtor( OALMessage msg )
{
    PCR_Assert( msg );

    DOFOperation_Internal_Destroy( (DOFOperation)msg->argv[0] );
    DOFResult_Destroy( (DOFResult)msg->argv[1] );

    if (msg->argv[2])
        DOFAddress_Destroy( (DOFAddress)msg->argv[2] );

    if (msg->argv[3])
        OALChannel_Release( (OALChannel)msg->argv[3] );
}

OALMessage SMT_RouteToRequestor_Convert( OALMessage message, void * recipient, DOFAddress address, OALChannel associatedChannel )
{
    PCR_Assert(message->msg_type == CMT_ROUTE_TO_REQUESTOR);

    message->recipient = recipient;
    message->msg_type = SMT_ROUTE_TO_REQUESTOR;
    message->dtor = SMT_RouteToRequestor_Dtor;

    if (address != message->argv[2])
    {
        DOFAddress_Destroy((DOFAddress)message->argv[2]);
        message->argv[2] = DOFAddress_Copy(address);
    }
    if ((message->argv[2]) || !address)
    {
        message->argv[3] = OALChannel_Retain(associatedChannel);
        if (message->argv[3] || !associatedChannel)
        {
            return message;
        }
    }

    OALMessage_Destroy(message);
    return NULL;
}

static void SMT_Shutdown_Dtor(OALMessage msg)
{
    DOFException_Destroy((DOFException)msg->argv[0]);
}

OALMessage SMT_Shutdown_Create( void* recipient )
{
    OALMessage retVal =  OALMessage_Create( 1, recipient,
                                            SMT_SHUTDOWN,
                                            SMT_Shutdown_Dtor );

    if (retVal)
    {
        retVal->argv[0] = NULL;
    }
    return retVal;
}

void SMT_Shutdown_SetException(OALMessage msg, DOFException exception)
{
    if (msg && (msg->argv[0] == NULL))
    {
        msg->argv[0] = DOFException_Copy(exception);
    }
}

OALMessage SMT_Connected_Negotiation_Timeout_Create(void *recipient)
{
    return OALMessage_Create(0, recipient,
                             SMT_CONNECTED_NEGOTIATION_TIMEOUT,
                             NULL);
}

OALMessage SMT_Connected_Negotiation_Server_Side_Create(void *recipient)
{
    return OALMessage_Create(0, recipient,
                             SMT_NEGOTIATION_SERVER_SIDE,
                             NULL);
}

static void SMT_RemoteDomainCallback_Dtor( OALMessage msg )
{
    PCR_Assert(msg);

    DOFCredentials_Destroy((DOFCredentials)msg->argv[0]);
}

OALMessage SMT_RemoteDomainCallback_Create(void *recipient, DOFCredentials creds)
{
    OALMessage msg = OALMessage_Create(1, recipient, SMT_REMOTE_DOMAIN_CALLBACK, SMT_RemoteDomainCallback_Dtor);

    if (msg)
    {
        msg->argv[0] = DOFCredentials_Copy(creds);
        if (!msg->argv[0] && creds)
        {
            OALMessage_Destroy(msg);
            msg = NULL;
        }
    }
    return msg;
}

static void SMT_ResolveCredsCallback_Dtor( OALMessage msg )
{
    PCR_Assert(msg);

    DOFCredentials_Destroy((DOFCredentials)msg->argv[0]);
}

OALMessage SMT_ResolveCredsCallback_Create(void *recipient, DOFCredentials creds)
{
    OALMessage msg = OALMessage_Create(1, recipient, SMT_RESOLVE_CREDS_CALLBACK, SMT_ResolveCredsCallback_Dtor);

    if (msg)
    {
        msg->argv[0] = DOFCredentials_Copy(creds);
        if (!msg->argv[0] && creds)
        {
            OALMessage_Destroy(msg);
            msg = NULL;
        }
    }
    return msg;
}

static void SMT_GetLibraryPermsCallback_Dtor( OALMessage msg )
{
    PCR_Assert(msg);

    DOFPermissionSet_Destroy((DOFPermissionSet)msg->argv[0]);
    source_release((sid_t *)msg->argv[1]);
}

OALMessage SMT_GetLibraryPermsCallback_Create(void *recipient, DOFPermissionSet perms, sid_t * secureSid)
{
    OALMessage msg = OALMessage_Create(2 ,recipient, SMT_GET_LIBRARY_PERMS_CALLBACK, SMT_GetLibraryPermsCallback_Dtor);

    if (msg)
    {
        msg->argv[0] = DOFPermissionSet_Copy(perms);
        if (!perms || msg->argv[0])
        {
            source_retain(secureSid);
            msg->argv[1] = secureSid;
        }
        else
        {
            msg->argv[1] = NULL;
        }
    }
    return msg;
}

static void SMT_RouteSavedEPP_Dtor( OALMessage msg )
{
    PCR_Assert(msg);

    BlackboardData_Clear((BlackboardData)msg->argv[0]);
    DOFMem_Free((BlackboardData)msg->argv[0]);
}

/* Note: The message created here will take ownership of the BlackboardData passed in. */
OALMessage SMT_RouteSavedEPP_Create(void *recipient, BlackboardData data)
{
    OALMessage msg = OALMessage_Create(1, recipient, SMT_ROUTE_SAVED_EPP, SMT_RouteSavedEPP_Dtor);
    if (msg)
    {
        msg->argv[0] = data;
    }
    return msg;
}

