/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/connection.h>
#include <dof/error.h>
#include <internal/bridgeinternal.h>
#include <internal/channel.h>
#include <internal/channel_stack_messages.h>
#include <internal/connected_remotechannel.h>
#include <internal/connected_protocolstack.h>
#include <internal/connection_internal.h>
#include <internal/connectionmanager.h>
#include <internal/dofinternal.h>
#include <internal/epp_negative_one.h>
#include <internal/exception.h>
#include <internal/mutex_pool.h>
#include <internal/operationinternal.h>
#include <internal/remotechannel.h>
#include <internal/remotechannelmanager.h>
#include <internal/router.h>
#include <internal/tep_v1.h>
#include <internal/timer.h>
#include <internal/transportinternal.h>
#include <internal/transportmanager.h>
#include <internal/trp_v1.h>
#include <internal/serverinternal.h>

extern PCRMutex connection_server_mutex;

typedef struct OALConnectedRemoteChannel_t
{
    OALChannel_t    superclass; /* OALConnectedRemoteChannel inherits from the OALChannel abstract class. Must always be first element in this struct. */

    boolean         isClient;

} OALConnectedRemoteChannel_t;

typedef void (*OCRC_MessageHandler_fn_t)( OALConnectedRemoteChannel self, OALMessage message );

static void OALConnectedRemoteChannel_Dtor( OALChannel self );
static void OALConnectedRemoteChannel_HandleMessage( OALChannel self, OALMessage message );


OALConnectedRemoteChannel OALConnectedRemoteChannel_Create(OALConnectedProtocolStack associatedStack,
                                                           DTL associatedTransport,
                                                           DOFSecurityMembrane securityMembrane,
                                                           DOFConnectionConfig connectionConfig,
                                                           boolean isClient)
{
    OALConnectedRemoteChannel retVal = (OALConnectedRemoteChannel)DOFMem_Alloc( sizeof(OALConnectedRemoteChannel_t) );

    PCRMem_Set(retVal, 0, sizeof(OALConnectedRemoteChannel_t));

    if (!retVal)
    {
        return NULL;
    }
    PCR_Assert(securityMembrane != NULL);
    /* If this assertion is invalid we'll need to create a ChannelMembrane, but preferably at a higher level where there's more info */

    if (!OALChannel_Ctor( &(retVal->superclass),
                          TRUE, securityMembrane,
                          OALRemoteChannel_RouteToProviderLossless,
                          OALRemoteChannel_RouteToRequestorLossless,
                          OALConnectedRemoteChannel_HandleMessage,
                          OALConnectedRemoteChannel_Dtor, TEP_V1_ExtendPermissions))
    {
        DOFMem_Free(retVal);
        return NULL;
    }

    if (!OALChannel_RemoteChannel_Ctor( &(retVal->superclass),
                                        associatedStack,
                                        associatedTransport,
                                        DOFConnectionConfig_GetAddress(connectionConfig),
                                        connectionConfig,
                                        NULL))
    {
        OALChannel_Dtor(&retVal->superclass);
        DOFMem_Free(retVal);
        return NULL;
    }

    /* Create the shutdown message now so the channel can always send a shutdown. */
    retVal->superclass.remoteChannelData->shutdownMessage = OALRemoteChannel_CreateShutdownMessage(&retVal->superclass);
    if (!retVal->superclass.remoteChannelData->shutdownMessage)
    {
        OALChannel_RemoteChannel_Dtor(&retVal->superclass);
        OALChannel_Dtor(&retVal->superclass);
        DOFMem_Free(retVal);
        return NULL;
    }

    retVal->isClient = isClient;
    return retVal;
}

static void OALConnectedRemoteChannel_Dtor( OALChannel self )
{
    OALRemoteChannel_ShutdownTransport(self);

    OALTimerItem_Destroy(self->remoteChannelData->maxReceiveSilenceTimerItem, TRUE);
    OALTimerItem_Destroy(self->remoteChannelData->maxSendSilenceTimerItem, TRUE);

    OALChannel_Dtor(self);
}

static void OALConnectedRemoteChannel_Shutdown( OALConnectedRemoteChannel self, DOFException exception )
{
    if(self->superclass.remoteChannelData->flags.isInShutdownState)
        return;

    self->superclass.remoteChannelData->flags.isConnected = 0;
    self->superclass.remoteChannelData->flags.isInShutdownState = TRUE;

    OALRemoteChannelMgr_RemoveSilenceTimerItems(&self->superclass);

    OALRemoteChannel_RemoveOALConnection(&self->superclass, exception);

    OALRemoteChannelMgr_RemoveProtocolStack_Immediate(self->superclass.remoteChannelData->protocolStack);
    OALConnectedProtocolStack_ClearChannel(self->superclass.remoteChannelData->protocolStack);
    OALRemoteChannelMgr_RemoveChannel_Immediate( (OALChannel)self );
}

boolean OALConnectedRemoteChannel_GetIsClient(OALConnectedRemoteChannel self)
{
    return self->isClient;
}

static void OALConnectedRemoteChannel_HandleMessage_RouteToProvider( OALConnectedRemoteChannel self, OALMessage message )
{
    OALMessage msgToStack;
    /* Re-use the message and send it to the protocol stack for handling. */
    msgToStack = SMT_RouteToProvider_Convert(message, self->superclass.remoteChannelData->protocolStack,
                                             &self->superclass, NULL);

    if (msgToStack)
        OALRemoteChannelMgr_SendMessageToStack(msgToStack);
    else
        OALRemoteChannel_Shutdown(&self->superclass, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
}

static void OALConnectedRemoteChannel_HandleMessage_RouteToRequestor( OALConnectedRemoteChannel self, OALMessage message )
{
    OALMessage msgToStack = SMT_RouteToRequestor_Convert(message, self->superclass.remoteChannelData->protocolStack,
                                                         (DOFAddress)message->argv[2], NULL);
    /* re-use the message and send it to the protocol stack for handling. */

    if (msgToStack)
        OALRemoteChannelMgr_SendMessageToStack( msgToStack );
    else
        OALRemoteChannel_Shutdown(&self->superclass, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
}

static void OALConnectedRemoteChannel_HandleMessage_ReceiveFromTransport( OALConnectedRemoteChannel self, OALMessage message )
{
    if (message->shutdown)
    {
        OALMessage_Destroy(message);
    }
    else
    {
        /* re-use the message and send it to the protocol stack for handling. */
        message->recipient = self->superclass.remoteChannelData->protocolStack;
        message->msg_type = SMT_RECEIVE_FROM_TRANSPORT;

        message->argv[3] = OALChannel_Retain( (OALChannel)self );
        if (message->argv[3])
        {
            if(self->superclass.remoteChannelData->maxReceiveSilenceTimerItem)
                OALRemoteChannelMgr_SetRecvSilenceTimerItem( &self->superclass );

            OALRemoteChannelMgr_SendMessageToStack( message );
        }
        else
        {
            OALRemoteChannel_Shutdown(&self->superclass, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        }
    }
}

static void OALConnectedRemoteChannel_HandleMessage_Initialize( OALConnectedRemoteChannel self, OALMessage message )
{
    if (message->shutdown)
    {
        OALMessage_Destroy(message);
    }
    else
    {
        /* re-use the message and send it to the protocol stack for handling. */
        message->recipient = self->superclass.remoteChannelData->protocolStack;
        message->msg_type = SMT_BEGIN_NEGOTIATION_AUTHORIZATION;
        OALRemoteChannelMgr_SendMessageToStack( message );
    }
}

static void OALConnectedRemoteChannel_ForwardMessageToRemoteChannelManager_fn(void * context, boolean success);
static const struct EPPNegOneQueryCallbackFns_t EPPNegOneQueryCallbackFns = { OALConnectedRemoteChannel_ForwardMessageToRemoteChannelManager_fn };
static struct EPPNegOneQueryCallback_t EPPNegOneQueryCB = { &EPPNegOneQueryCallbackFns };

static void OALConnectedRemoteChannel_ForwardMessageToRemoteChannelManager_fn(void * context, boolean success)
{
    /* Connections do not wait for the response, so there is nothing to do here.*/
}

static void OALConnectedRemoteChannel_OALDomainCredentialsTree_ValidateDomainCredentials_Callback( DOFException exception, void* context )
{
    DOFConnection connection = (DOFConnection)context;
    DOFConnection_ValidateComplete( connection, exception );
    DOFConnection_Destroy(connection);
}

/* Note: This will handle a successful initialize.  Shutdown will be called for failures. */
static void OALConnectedRemoteChannel_HandleMessage_InitializeComplete( OALConnectedRemoteChannel self, OALMessage message )
{
    if (!message->shutdown)
    {
        if (self->superclass.remoteChannelData->flags.validateDomainCreds &&
            self->superclass.remoteChannelData->OALConn)
        {
            DOFConnection   conn;

            /* Validate the domain credentials*/
            self->superclass.remoteChannelData->flags.validateDomainCreds = FALSE;
            conn = OALConnection_CreateDOFConnection(self->superclass.remoteChannelData->OALConn);
            OALDomainCredentialsTree_ValidateDomainCredentials( self->superclass.remoteChannelData->OALConn->connConfig->domainCreds, conn, OALConnectedRemoteChannel_OALDomainCredentialsTree_ValidateDomainCredentials_Callback, (void*)conn );
        }
        else if (self->superclass.remoteChannelData->flags.createBridge &&
                 self->superclass.remoteChannelData->OALConn && DOFConnectionConfig_GetBridge(self->superclass.remoteChannelData->config.connection))
        {
            /* Setup the DOFOperationBridge. */
            self->superclass.remoteChannelData->flags.createBridge = FALSE;
            OALRemoteChannel_CreateBridge(&self->superclass, self->superclass.remoteChannelData->address, DOF_GetDof(), DOFConnectionConfig_GetBridge(self->superclass.remoteChannelData->config.connection));
        }
        else
        {
            self->superclass.remoteChannelData->flags.mustTerminate = TRUE;
            OALProtocolStack_CancelNegotiationTimer(self->superclass.remoteChannelData->protocolStack);
            OALRemoteChannelMgr_SetSendSilenceTimerItem( &self->superclass );
            OALRemoteChannelMgr_SetRecvSilenceTimerItem( &self->superclass );

            /* let the router know there's a new channel */
            DOFRouter_ChannelUp((OALChannel)self);

            if (self->superclass.remoteChannelData->OALConn)
            {
                self->superclass.remoteChannelData->flags.isConnected = 1;
                /* Start a sid query, but don't wait for the response. */
                EPP_NegOne_Query(&self->superclass, self->superclass.remoteChannelData->address, 8000, 3000, NULL, &EPPNegOneQueryCB);
                OALConnection_ConnectionEvent( self->superclass.remoteChannelData->OALConn, &self->superclass, TRUE, NULL );
            }
        }
    }

    OALMessage_Destroy( message );
}

static void OALConnectedRemoteChannel_Process_Shutdown( OALConnectedRemoteChannel self, DOFException exception )
{
    OALRemoteChannel_ShutdownTransport(&self->superclass);

    OALConnectedRemoteChannel_Shutdown(self, exception);
}

static void OALConnectedRemoteChannel_HandleMessage_Shutdown( OALConnectedRemoteChannel self, OALMessage message )
{
    DOFException exception = message->argv[0];

    OALConnectedRemoteChannel_Process_Shutdown(self, exception);

    OALMessage_Destroy( message );
}

static void OALConnectedRemoteChannel_HandleMessage_TransportDown( OALConnectedRemoteChannel self, OALMessage message )
{
    DTL transport = message->argv[0];
    DOFException exception = message->argv[1];

    if(transport == self->superclass.remoteChannelData->dtl)
        OALConnectedRemoteChannel_Shutdown( self, exception );

    OALMessage_Destroy( message );
}

static void OALConnectedRemoteChannel_HandleMessage_ProcessBlockedMessage(OALConnectedRemoteChannel self, OALMessage message)
{
    /*  We currently do not expect to receive this message on a connected remote channel.
        Just destroy the message and ignore it. */

    OALMessage_Destroy(message);
}

static void OALConnectedRemoteChannel_HandleMessage_ConnectionInitiatedShutdown(OALConnectedRemoteChannel self, OALMessage message)
{
    /* Clear the reference to the OALConnection.  This does not require a release for outbound connections.
       Note: Clearing the OALConn will prevent the channel from notifing the OALConn that it was shutdown.
       This prevents a race condition where the OALConn can be destroyed before the notification from
       the channel is received. */
    if (!self->superclass.remoteChannelData->flags.isOutbound)
    {
        /* For inbound connections The channel actually retains the OALConnection. */
        OALConnection_Release(self->superclass.remoteChannelData->OALConn);
    }

    PCRMutex_Lock(connection_server_mutex);
    self->superclass.remoteChannelData->OALConn = NULL;
    PCRMutex_Unlock(connection_server_mutex);

    OALConnectedRemoteChannel_Process_Shutdown(self, NULL);

    OALMessage_Destroy(message);
}

static void OALConnectedRemoteChannel_HandleMessage_RegisterSecureChannel(OALConnectedRemoteChannel self, OALMessage message)
{
    if (!message->shutdown)
    {
        OALRemoteChannel_HandleMessage_RegisterSecureChannel((OALChannel)&self->superclass, message);
    }

    OALMessage_Destroy(message);
}

static void OALConnectedRemoteChannel_HandleMessage_APPIDQueryTimeout(OALConnectedRemoteChannel self, OALMessage message)
{
    /*  We currently do not expect to receive this message on a connected remote channel.
        Just destroy the message and ignore it. */

    OALMessage_Destroy(message);
}

static void OALConnectedRemoteChannel_HandleMessage_RegisterChannel_ENPPort(OALConnectedRemoteChannel self, OALMessage message)
{
    /*  We currently do not expect to receive this message on a connected remote channel.
        Just destroy the message and ignore it. */

    OALMessage_Destroy(message);
}

static void OALConnectedRemoteChannel_HandleMessage_RemoteDomainCallback(OALConnectedRemoteChannel self, OALMessage message)
{
    /*  We currently do not expect to receive this message on a connected remote channel.
        Just destroy the message and ignore it. */

    OALMessage_Destroy(message);
}

static void OALConnectedRemoteChannel_HandleMessage_ResolveCredsCallback(OALConnectedRemoteChannel self, OALMessage message)
{
    /*  We currently do not expect to receive this message on a connected remote channel.
        Just destroy the message and ignore it. */

    OALMessage_Destroy(message);
}

static void OALConnectedRemoteChannel_HandleMessage_GetLibraryPermsCallback(OALConnectedRemoteChannel self, OALMessage message)
{
    /*  We currently do not expect to receive this message on a connected remote channel.
        Just destroy the message and ignore it. */

    OALMessage_Destroy(message);
}

static void OALConnectedRemoteChannel_HandleMessage_ServerInitiatedShutdown(OALConnectedRemoteChannel self, OALMessage message)
{
    /* Call shutdown for the channel. */
    OALConnectedRemoteChannel_Process_Shutdown(self, NULL);

    OALMessage_Destroy(message);
}

static void OALConnectedRemoteChannel_HandleMessage_ProcessReceiveSilence(OALConnectedRemoteChannel self, OALMessage message)
{
    if (!message->shutdown)
    {
        OALRemoteChannel_HandleMessage_ProcessReceiveSilence(&self->superclass, message);
    }
    OALMessage_Destroy(message);
}

static void OALConnectedRemoteChannel_HandleMessage_SetDomainCredentialsCallback(OALConnectedRemoteChannel self, OALMessage message)
{
    if (!message->shutdown)
    {
        OALRemoteChannel_HandleMessage_SetDomainCredentialsCallback(&self->superclass, message);
    }
    OALMessage_Destroy(message);
}

static void OALConnectedRemoteChannel_HandleMessage_PeerListenerEvent(OALConnectedRemoteChannel self, OALMessage message)
{
    OALRemoteChannel_HandleMessage_PeerListenerEvent((OALChannel)self, message);
    OALMessage_Destroy(message);
}

static void OALConnectedRemoteChannel_HandleMessage_CreateBridge(OALConnectedRemoteChannel self, OALMessage message)
{
    if (!message->shutdown)
    {
        OALMessage_CreateBridgeMessage msg = (OALMessage_CreateBridgeMessage)message;

        OALRemoteChannel_CreateBridge(&self->superclass, self->superclass.remoteChannelData->address, msg->dof, msg->bridgeConfig);
    }
    OALMessage_Destroy(message);
}

static void OALConnectedRemoteChannel_HandleMessage_CreateBridgeComplete(OALConnectedRemoteChannel self, OALMessage message)
{
    if (!message->shutdown)
    {
        OALMessage_CreateBridgeCompleteMessage msg = (OALMessage_CreateBridgeCompleteMessage)message;

        PCR_Assert(DOFSecurityMembrane_IsChannelMembrane(self->superclass.securityMembrane));
        OALChannelMembrane_SetBridge(self->superclass.securityMembrane, msg->bridge);

        /* Notify channel of the success. */
        OALRemoteChannel_NotifyInitComplete(&self->superclass);
    }
    OALMessage_Destroy(message);
}

static void OALConnectedRemoteChannel_HandleMessage_ConnectionStreamListener(OALConnectedRemoteChannel self, OALMessage message)
{
    OALRemoteChannel_HandleMessage_ConnectionStreamListener((OALChannel)self, message);
    OALMessage_Destroy(message);
}

static void OALConnectedRemoteChannel_HandleMessage( OALChannel self, OALMessage message )
{
    static OCRC_MessageHandler_fn_t msg_handler_funcs[] = {
        /* the order here is coupled with ChannelMsgType_t. So dont change it. */
        OALConnectedRemoteChannel_HandleMessage_RouteToProvider,
        OALConnectedRemoteChannel_HandleMessage_RouteToRequestor,
        OALConnectedRemoteChannel_HandleMessage_ReceiveFromTransport,
        OALConnectedRemoteChannel_HandleMessage_Initialize,
        OALConnectedRemoteChannel_HandleMessage_InitializeComplete,
        OALConnectedRemoteChannel_HandleMessage_Shutdown,
        OALConnectedRemoteChannel_HandleMessage_TransportDown,
        OALConnectedRemoteChannel_HandleMessage_ProcessBlockedMessage,
        OALConnectedRemoteChannel_HandleMessage_ConnectionInitiatedShutdown,
        OALConnectedRemoteChannel_HandleMessage_RegisterSecureChannel,
        OALConnectedRemoteChannel_HandleMessage_APPIDQueryTimeout,
        OALConnectedRemoteChannel_HandleMessage_RegisterChannel_ENPPort,
        OALConnectedRemoteChannel_HandleMessage_RemoteDomainCallback,
        OALConnectedRemoteChannel_HandleMessage_ResolveCredsCallback,
        OALConnectedRemoteChannel_HandleMessage_GetLibraryPermsCallback,
        OALConnectedRemoteChannel_HandleMessage_ServerInitiatedShutdown,
        OALConnectedRemoteChannel_HandleMessage_ProcessReceiveSilence,
        OALConnectedRemoteChannel_HandleMessage_SetDomainCredentialsCallback,
        OALConnectedRemoteChannel_HandleMessage_PeerListenerEvent,
        OALConnectedRemoteChannel_HandleMessage_CreateBridge,
        OALConnectedRemoteChannel_HandleMessage_CreateBridgeComplete,
        OALConnectedRemoteChannel_HandleMessage_ConnectionStreamListener
    };
    STATIC_ASSERT( (sizeof( msg_handler_funcs ) / sizeof( OCRC_MessageHandler_fn_t )) ==
                   (CMT_LAST - CMT_ROUTE_TO_PROVIDER) );   /* These are in ChannelMsgType_t. */

    PCR_Assert( self );
    PCR_Assert( message );
    PCR_Assert( message->recipient == self );
    PCR_Assert( message->msg_type < (sizeof(msg_handler_funcs) / sizeof(OCRC_MessageHandler_fn_t)) );

    msg_handler_funcs[message->msg_type]( (OALConnectedRemoteChannel)self, message );
}




