/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/addressinternal.h>
#include <internal/bridgeinternal.h>
#include <internal/channel.h>
#include <internal/channel_stack_messages.h>
#include <internal/cipher.h>
#include <internal/connectionless_remotechannel.h>
#include <internal/connectionmanager.h>
#include <internal/dofinternal.h>
#include <internal/enp_v1.h>
#include <internal/epp_negative_one.h>
#include <internal/exception.h>
#include <internal/hashtable.h>
#include <internal/messaging.h>
#include <internal/mutex_pool.h>
#include <internal/operationinternal.h>
#include <internal/operationconfiginternal.h>
#include <internal/remotechannel.h>
#include <internal/remotechannelmanager.h>
#include <internal/router.h>
#include <internal/serverinternal.h>
#include <internal/source.h>
#include <internal/tep_v1.h>
#include <internal/timer.h>
#include <internal/transportmanager.h>
#include <internal/trp_v1.h>

#include "channel_local.h"

extern PCRMutex connection_server_mutex;

typedef struct OALConnectionlessRemoteChannel_t
{
    /* OALConnectionlessRemoteChannel inherits from the OALChannel abstract class. */
    OALChannel_t                superclass; /*  This must always be first element in this struct. */
    TEP_V1_SessionEventCallback sessionEventCallback;
    uint8                       state;
} OALConnectionlessRemoteChannel_t;

#define CONNECTIONLESS_REMOTECHANNEL_STATE_INITIAL          0
#define CONNECTIONLESS_REMOTECHANNEL_STATE_ESP_NEGOTIATION  1
#define CONNECTIONLESS_REMOTECHANNEL_STATE_TEP_NEGOTIATION  2
#define CONNECTIONLESS_REMOTECHANNEL_STATE_APPID_DISCOVERY  3
#define CONNECTIONLESS_REMOTECHANNEL_STATE_STATELESS_QUERY  4
#define CONNECTIONLESS_REMOTECHANNEL_STATE_VALIDATE_CREDS   5
#define CONNECTIONLESS_REMOTECHANNEL_STATE_SIDQUERY         6
#define CONNECTIONLESS_REMOTECHANNEL_STATE_WAIT_DESIRE      7
#define CONNECTIONLESS_REMOTECHANNEL_STATE_ESTABLISHED      8

typedef void (*OCLRC_MessageHandler_fn_t)( OALConnectionlessRemoteChannel self, OALMessage message );



static void OALConnectionlessRemoteChannel_Dtor( OALChannel self );
static void OALConnectionlessRemoteChannel_HandleMessage( OALChannel self, OALMessage message );

static void* OALConnectionlessRemoteChannel_TEPSessionEventCallback(void               * context,
                                                                    uint8              * key,
                                                                    TEP_SessionEvent_t   eventType,
                                                                    DOFSecurityMode      securityMode,
                                                                    OALChannel           ch,
                                                                    DOFSecurityScope     secScope,
                                                                    DOFPermissionSet     perms,
                                                                    DOFPermissionSet     otherPerms,
                                                                    DOFException         exception,
                                                                    uint32               stateID,
                                                                    DOFAddress           remoteAddr,
                                                                    OALStream          * initialState,
                                                                    DOFCredentials       creds);

OALConnectionlessRemoteChannel OALConnectionlessRemoteChannel_Create(OALConnectionlessProtocolStack associatedStack,
                                                                     DTL associatedTransport,
                                                                     DOFSecurityMembrane securityMembrane,
                                                                     DOFAddress address,
                                                                     DOFConnectionConfig connConfig,
                                                                     DOFServerConfig serverConfig)
{
    OALConnectionlessRemoteChannel retVal = (OALConnectionlessRemoteChannel)DOFMem_Alloc( sizeof(OALConnectionlessRemoteChannel_t) );

    if (!retVal)
    {
        return NULL;
    }

    PCRMem_Set(retVal, 0, sizeof(OALConnectionlessRemoteChannel_t));
    PCR_Assert(securityMembrane != NULL);
    /* If this assertion is invalid we'll need to create a ChannelMembrane, but preferably at a higher level where there's more info */

    if (!OALChannel_Ctor(&(retVal->superclass),
                         TRUE, securityMembrane,
                         OALRemoteChannel_RouteToProviderLossy,
                         OALRemoteChannel_RouteToRequestorLossy,
                         OALConnectionlessRemoteChannel_HandleMessage,
                         OALConnectionlessRemoteChannel_Dtor,
                         TEP_V1_ExtendPermissions))
    {
        DOFMem_Free(retVal);
        return NULL;
    }

    if (!OALChannel_RemoteChannel_Ctor( &(retVal->superclass),
                                        associatedStack,
                                        associatedTransport,
                                        address,
                                        connConfig,
                                        serverConfig))
    {
        OALChannel_Dtor(&retVal->superclass);
        DOFMem_Free(retVal);
        return NULL;
    }

    /* Create the shutdown message now so the channel can always send a shutdown. */
    retVal->superclass.remoteChannelData->shutdownMessage = OALRemoteChannel_CreateShutdownMessage(&retVal->superclass);
    if (!retVal->superclass.remoteChannelData->shutdownMessage)
    {
        OALChannel_RemoteChannel_Dtor(&retVal->superclass);
        OALChannel_Dtor(&retVal->superclass);
        DOFMem_Free(retVal);
        return NULL;
    }


    /* If there is a known route to an AS set TRP_AcceptsCommands. */
    retVal->superclass.remoteChannelData->flags.TRP_AcceptsCommands = TRPRouter_KnowsAnAS();

    /* TRP_SendCommands is used along with the V0_QUERY_BIT_TRP_V1 bit to determine if the other side will accept TRP commands.
       TRP_SendCommands is set by default and V0_QUERY_BIT_TRP_V1 is cleared by default.  At the point an ESP v0 query response is
       received V0_QUERY_BIT_TRP_V1 will be set if the query response includes TRP, or TRP_SendCommands will be cleared if not.
       Outbound channels should always know if the remote node accepts TRP before TRP is used, but there is a case for inbound
       channels where TRP may be required and the channel does not know if the remote node accepts TRP or in other words
       TRP_SendCommands is set but V0_QUERY_BIT_TRP_V1 is not.  In this case the inbound channel must send an ESP v0 query
       and wait for the response.  If the remote node accepts TRP, the queued up TRP commands can be sent. otherwise they must be dropped. */
    retVal->superclass.remoteChannelData->flags.TRP_SendCommands = 1;

    retVal->state = CONNECTIONLESS_REMOTECHANNEL_STATE_INITIAL;

    if (serverConfig)
    {
        /* Currently all secure connections should be created based on a session which will already have an inbound connection. */
        retVal->sessionEventCallback = NULL;
        if (serverConfig->securityDesire & DOFSECURITYDESIRE_SECURE_ANY)
        {
            retVal->superclass.remoteChannelData->flags.secureDesire = TRUE;
        }
    }
    else
    {
        PCR_Assert(connConfig);

        if ((DOFConnectionConfig_GetConnectionType(connConfig) == DOFCONNECTIONTYPE_DATAGRAM) ||
            (connConfig->securityDesire & DOFSECURITYDESIRE_SECURE_ANY))
        {
            retVal->state = CONNECTIONLESS_REMOTECHANNEL_STATE_ESP_NEGOTIATION;
            retVal->superclass.remoteChannelData->flags.lossySession = TRUE;
            retVal->superclass.remoteChannelData->flags.waitingForInitialDatagram = TRUE;
            if (DOFConnectionConfig_GetSecurityDesire(connConfig) == DOFSECURITYDESIRE_ANY)
            {
                retVal->superclass.remoteChannelData->flags.waitingForSecurityIndicator = TRUE;
            }
        }
        else
        {
            retVal->state = CONNECTIONLESS_REMOTECHANNEL_STATE_STATELESS_QUERY;
        }
        retVal->sessionEventCallback = OALConnectionlessRemoteChannel_TEPSessionEventCallback;
        if (DOFConnectionConfig_GetSecurityDesire(connConfig) & DOFSECURITYDESIRE_SECURE_ANY)
        {
            retVal->superclass.remoteChannelData->flags.secureDesire = TRUE;
        }
    }

    return retVal;
}

static void OALConnectionlessRemoteChannel_Dtor( OALChannel self )
{
    /* nothing to free */
    if(self->remoteChannelData->dtl)
    {
        OALRemoteChannel_ShutdownTransport(self);
    }

    OALTimerItem_Destroy(self->remoteChannelData->maxReceiveSilenceTimerItem, TRUE);
    OALTimerItem_Destroy(self->remoteChannelData->maxSendSilenceTimerItem, TRUE);

    OALChannel_Dtor(self);
}

static void OALConnectionlessRemoteChannel_Shutdown( OALConnectionlessRemoteChannel self, DOFException exception )
{
    if(self->superclass.remoteChannelData->flags.isInShutdownState)
        return;

    self->superclass.remoteChannelData->flags.isConnected = 0;
    self->superclass.remoteChannelData->flags.isInShutdownState = TRUE;

    OALRemoteChannelMgr_RemoveSilenceTimerItems(&self->superclass);
    OALRemoteChannel_DeregisterChannel_ENPPort_Immediate(&self->superclass);
    OALRemoteChannelMgr_CancelNegotiationTimer(&self->superclass);
    OALRemoteChannel_DestroyBlockedMessageQueue(&self->superclass);

    /* When secure sessions using security state IDs are supported, deregister it here. */

    OALRemoteChannel_RemoveOALConnection(&(self->superclass), exception);
    OALRemoteChannel_RemoveOALServer( &(self->superclass), exception );

    if (self->superclass.remoteChannelData->baseChannel &&
        (&self->superclass == self->superclass.remoteChannelData->baseChannel->remoteChannelData->multicastChannel))
    {
        /* If this channel is listed as the multicast channel in the base channel, remove it. Note: it was not retained
            since this channel holds a reference to the base channel. */
        self->superclass.remoteChannelData->baseChannel->remoteChannelData->multicastChannel = NULL;
    }

    OALRemoteChannelMgr_RemoveChannel_Immediate( (OALChannel)self );
}

static void OALConnectionlessRemoteChannel_HandleMessage_RouteToProvider( OALConnectionlessRemoteChannel self, OALMessage message )
{
    DOFAddress addr = (DOFAddress)message->argv[2];
    /* DOFOperation req_op = message->argv[0]; */
    OALMessage msgToStack;

    if ( !addr )
    {
        /* Use the channel's default target address. */
        addr = self->superclass.remoteChannelData->address;
    }

    msgToStack = SMT_RouteToProvider_Convert(message, self->superclass.remoteChannelData->protocolStack,
                                             &self->superclass, addr);

    if (msgToStack)
    {
        OALRemoteChannelMgr_SendMessageToStack( msgToStack );
    }
    else
    {
        /* If the message could not be created this becomes a dropped packet. */
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "OALConnectionlessRemoteChannel: Route to provider: unable to send message to stack.");
    }
}

static void OALConnectionlessRemoteChannel_HandleMessage_RouteToRequestor( OALConnectionlessRemoteChannel self, OALMessage message )
{
    OALMessage msgToStack = SMT_RouteToRequestor_Convert(message, self->superclass.remoteChannelData->protocolStack,
                                                         (DOFAddress)message->argv[2], &self->superclass);

    if (msgToStack)
    {
        OALRemoteChannelMgr_SendMessageToStack( msgToStack );
    }
    else
    {
        /* If the message could not be created this becomes a dropped packet. */
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_TRACE, "OALConnectionlessRemoteChannel: Route to requestor: unable to send message to stack.");
    }
}

/* Shared with securegroup.c. */
void OALConnectionlessRemoteChannel_HandleMessage_ReceiveFromTransport( OALConnectionlessRemoteChannel self, OALMessage message )
{
    if (message->shutdown)
    {
        OALMessage_Destroy(message);
    }
    else
    {
        /* re-use the message and send it to the protocol stack for handling. */
        message->recipient = self->superclass.remoteChannelData->protocolStack;
        message->msg_type = SMT_RECEIVE_FROM_TRANSPORT;
        message->argv[3] = OALChannel_Retain( (OALChannel)self );
        if (message->argv[3])
        {
            if (self->superclass.remoteChannelData->flags.waitingForInitialDatagram)
            {
                if (!self->superclass.remoteChannelData->flags.isOutbound)
                {
                    /* Setup the silence timers for the channel. */
                    OALRemoteChannelMgr_SetSendSilenceTimerItem((OALChannel)self);
                    OALRemoteChannelMgr_SetRecvSilenceTimerItem((OALChannel)self);

                    if(self->superclass.remoteChannelData->OALConn)
                        OALConnection_ConnectionEvent(self->superclass.remoteChannelData->OALConn, (OALChannel)self, TRUE, NULL);

                    if(self->superclass.remoteChannelData->server)
                        OALServer_NotifyServerStarted(self->superclass.remoteChannelData->server);

                    /* For inbound channels, any received datagram sets this to FALSE. */
                    self->superclass.remoteChannelData->flags.waitingForInitialDatagram = FALSE;
                }

            }

            /* Note: this may just be a base channel used to interface with the DTL, the channel may be changed in
               ENP receive or EPP receive. */

            OALRemoteChannelMgr_SendMessageToStack( message );
        }
    }
}

/* The message will be re-used to start the security negotiation. */
static void OALConnectionlessRemoteChannel_Initialize_StartSecurityNegotiation(OALConnectionlessRemoteChannel self, OALMessage message)
{
    OALRemoteChannelMgr_CancelNegotiationTimer(&self->superclass);
    if (!OALRemoteChannelMgr_StartNegotiationTimeout(&self->superclass, TEP_V1_TIMEOUT))
        OALRemoteChannel_Shutdown(&self->superclass, DOFEXCEPTION_PROTOCOL_NEGOTIATION_FAILED);

    if (message->shutdown)
    {
        OALMessage_Destroy(message);
    }
    else
    {
        /*  Security negotiation must complete before this connection is established.
            If this is an inbound connection, start the security negotiation timer and wait for the client to
            start security netotiation.  If this is an outbound connection, start the security negotiation timer
            and start TEP negotiation. */
        if (self->superclass.remoteChannelData->OALConn)
        {
            if (self->superclass.remoteChannelData->flags.isOutbound)
            {
                self->superclass.remoteChannelData->flags.waitingForSecurityIndicator = FALSE;
                self->superclass.remoteChannelData->flags.secureDesire = TRUE;
                self->sessionEventCallback = OALConnectionlessRemoteChannel_TEPSessionEventCallback;

                TEP_V1_SendNoOp(&self->superclass);

                /* re-use the message and send it to the protocol stack for handling. */
                message->recipient = self->superclass.remoteChannelData->protocolStack;
                message->msg_type = SMT_BEGIN_NEGOTIATION_AUTHORIZATION;
                message->dtor = SMT_BeginNegotiationAuthorization_Dtor;
                message->argv[0] = OALChannel_Retain(&self->superclass);
                if (message->argv[0])
                {
                    OALRemoteChannelMgr_SendMessageToStack(message);
                }
                else
                {
                    OALMessage_Destroy(message);
                }
            }
            else
            {
                OALMessage_Destroy(message);
            }
        }
        else
        {
            /*  If the OALConn does not exist then the channel has already been shutdown.
                Destroy message and exit. */
            OALMessage_Destroy(message);
        }
    }
}

static void OALConnectionlessRemoteChannel_ForwardMessageToRemoteChannelManager_fn(void * context, boolean success);
static const struct EPPNegOneQueryCallbackFns_t EPPNegOneQueryCallbackFns = { OALConnectionlessRemoteChannel_ForwardMessageToRemoteChannelManager_fn };
static struct EPPNegOneQueryCallback_t EPPNegOneQueryCB = { &EPPNegOneQueryCallbackFns };

static void OALConnectionlessRemoteChannel_ForwardMessageToRemoteChannelManager_fn(void * context, boolean success)
{
    /* Connections do not wait for the response, so there is nothing to do here.*/
}

static void OALConnectionlessRemoteChannel_OALDomainCredentialsTree_ValidateDomainCredentials_Callback( DOFException exception, void* context )
{
    DOFConnection connection = (DOFConnection)context;
    DOFConnection_ValidateComplete( connection, exception );
    DOFConnection_Destroy(connection);
}

static void OALConnectionlessRemoteChannel_InitializeComplete_Notify(OALConnectionlessRemoteChannel self)
{
    OALRemoteChannelMgr_CancelNegotiationTimer(&self->superclass);

    if (self->superclass.remoteChannelData->cipher)
    {
        self->superclass.remoteChannelData->flags.sessionIsSecure = TRUE;
    }

    if (self->superclass.remoteChannelData->flags.validateDomainCreds && self->superclass.remoteChannelData->OALConn)
    {
        DOFConnection conn;

        /* Validate the domain credentials*/
        self->superclass.remoteChannelData->flags.validateDomainCreds = FALSE;
        self->state = CONNECTIONLESS_REMOTECHANNEL_STATE_VALIDATE_CREDS;
        conn = OALConnection_CreateDOFConnection(self->superclass.remoteChannelData->OALConn);
        OALDomainCredentialsTree_ValidateDomainCredentials( self->superclass.remoteChannelData->OALConn->connConfig->domainCreds, conn, OALConnectionlessRemoteChannel_OALDomainCredentialsTree_ValidateDomainCredentials_Callback, (void*)conn );
    }
    else if (self->superclass.remoteChannelData->flags.createBridge &&
             self->superclass.remoteChannelData->OALConn && DOFConnectionConfig_GetBridge(self->superclass.remoteChannelData->config.connection))
    {
        /* Setup the DOFOperationBridge. */
        self->superclass.remoteChannelData->flags.createBridge = FALSE;
        OALRemoteChannel_CreateBridge(&self->superclass, self->superclass.remoteChannelData->address, DOF_GetDof(), DOFConnectionConfig_GetBridge(self->superclass.remoteChannelData->config.connection));
    }
    else
    {
        OALRemoteChannelMgr_SetSendSilenceTimerItem(&self->superclass);
        OALRemoteChannelMgr_SetRecvSilenceTimerItem(&self->superclass);

        if (self->superclass.remoteChannelData->flags.isOutbound)
        {
            /* let the router know there's a new channel */
            DOFRouter_ChannelUp(&self->superclass);
        }

        if(self->superclass.remoteChannelData->OALConn)
        {
            self->superclass.remoteChannelData->flags.isConnected = 1;
            EPP_NegOne_Query(&self->superclass, self->superclass.remoteChannelData->address, 8000, 3000, NULL, &EPPNegOneQueryCB);

            if (!self->superclass.remoteChannelData->flags.isOutbound)
            {
                self->superclass.remoteChannelData->flags.waitingForQueryResponse = TRUE;
                /* Send an ESP query to find out what protocols the remote node will accept. */
                ESP_V2_SendQuery(&self->superclass, 8000, 3000);
            }
            OALConnection_ConnectionEvent(self->superclass.remoteChannelData->OALConn, &self->superclass, TRUE, NULL);
        }

        if(self->superclass.remoteChannelData->server)
            OALServer_NotifyServerStarted( self->superclass.remoteChannelData->server );
    }
}

/* Note: OALConnectionlessRemoteChannel_InitializeComplete will re-use or destroy message. */
static void OALConnectionlessRemoteChannel_InitializeComplete(OALConnectionlessRemoteChannel self, OALMessage message)
{
    if (message->shutdown)
    {
        OALMessage_Destroy(message);
        return;
    }

    if ((self->state == CONNECTIONLESS_REMOTECHANNEL_STATE_ESP_NEGOTIATION) ||
        (self->state == CONNECTIONLESS_REMOTECHANNEL_STATE_STATELESS_QUERY))
    {
        if (OALRemoteChannelConfig_GetSecurityDesire(&self->superclass) & DOFSECURITYDESIRE_SECURE_ANY)
        {
            self->state = CONNECTIONLESS_REMOTECHANNEL_STATE_TEP_NEGOTIATION;

            /*  OALConnectionlessRemoteChannel_Initialize_StartSecurityNegotiation will send a begin negotiation authorization message to
                the protocol stack.  It will re-use or destroy message. */
            OALConnectionlessRemoteChannel_Initialize_StartSecurityNegotiation(self, message);
        }
        else if (OALRemoteChannelConfig_GetSecurityDesire(&self->superclass) & DOFSECURITYDESIRE_ANY)
        {
            /* If the server requires security then start the security negotiation. */
            if (self->superclass.remoteChannelData->flags.isOutbound)
            {
                if (self->superclass.remoteChannelData->flags.secureDesire)
                {
                    if ((self->state == CONNECTIONLESS_REMOTECHANNEL_STATE_STATELESS_QUERY))
                    {
                        /*  If this is a stateless datagram connection the EPS session must be established first by
                            sending an ESP open command with a non-zero delay. */
                        self->state = CONNECTIONLESS_REMOTECHANNEL_STATE_ESP_NEGOTIATION;
                        self->superclass.remoteChannelData->flags.lossySession = TRUE;
                        self->superclass.remoteChannelData->flags.waitingForInitialDatagram = TRUE;
                        ESP_V2_SendOpen(&self->superclass);
                        OALMessage_Destroy(message);
                    }
                    else
                    {
                        self->state = CONNECTIONLESS_REMOTECHANNEL_STATE_TEP_NEGOTIATION;

                        /*  OALConnectionlessRemoteChannel_Initialize_StartSecurityNegotiation will send a begin negotiation authorization message to
                            the protocol stack.  It will re-use or destroy message. */
                        OALConnectionlessRemoteChannel_Initialize_StartSecurityNegotiation(self, message);
                    }
                }
                else
                {
                    if (self->state == CONNECTIONLESS_REMOTECHANNEL_STATE_ESP_NEGOTIATION)
                    {
                        ESP_V2_SendQuery(&self->superclass, 8000, 3000);
                        if (self->superclass.remoteChannelData->negotiator == DOF_DEFAULT_AS_PROTOCOL_NEGOTIATOR)
                            EPP_NegOne_Heartbeat(&self->superclass, self->superclass.remoteChannelData->address, FALSE);
                        else
                            DOFOperation_SendOAPNoOp(&self->superclass);
                    }
                    self->state = CONNECTIONLESS_REMOTECHANNEL_STATE_ESTABLISHED;
                    OALConnectionlessRemoteChannel_InitializeComplete_Notify(self);
                    OALMessage_Destroy(message);
                }
            }
            else
            {
                /*  Since the security desire is any, at this point we don't know if will become secure, or if
                    it will remain unsecure.  The first TEP, or non-TEP PDU will determine this and send
                    another initialize complete message. waitingForSecurityIndicator was set when the channel was
                    created and will get cleared when the first TEP, or non-TEP PDU is received.  Note: since a
                    message is used to trigger this function, the flag, it may already be cleared at the point
                    this message is being processed, so it cannot be set or checked by this function. */
                self->state = CONNECTIONLESS_REMOTECHANNEL_STATE_WAIT_DESIRE;
                OALMessage_Destroy(message);
            }
        }
        else
        {
            /* This is to catch the DOFSECURITYDESIRE_NOT_SEUCRE case. */
            if (self->superclass.remoteChannelData->flags.isOutbound &&
                (self->state == CONNECTIONLESS_REMOTECHANNEL_STATE_ESP_NEGOTIATION))
            {
                ESP_V2_SendQuery(&self->superclass, 8000, 3000);
                if (self->superclass.remoteChannelData->negotiator == DOF_DEFAULT_AS_PROTOCOL_NEGOTIATOR)
                    EPP_NegOne_Heartbeat(&self->superclass, self->superclass.remoteChannelData->address, FALSE);
                else
                    DOFOperation_SendOAPNoOp(&self->superclass);
            }
            self->state = CONNECTIONLESS_REMOTECHANNEL_STATE_ESTABLISHED;
            OALConnectionlessRemoteChannel_InitializeComplete_Notify(self);
            OALMessage_Destroy(message);
        }
    }
    else if (self->state == CONNECTIONLESS_REMOTECHANNEL_STATE_WAIT_DESIRE)
    {
        /* This state will only apply to inbound connections. */
        if (self->superclass.remoteChannelData->flags.secureDesire)
        {
            self->state = CONNECTIONLESS_REMOTECHANNEL_STATE_TEP_NEGOTIATION;
        }
        else
        {
            self->state = CONNECTIONLESS_REMOTECHANNEL_STATE_ESTABLISHED;
            OALConnectionlessRemoteChannel_InitializeComplete_Notify(self);
        }
        OALMessage_Destroy(message);
    }
    else if (self->state == CONNECTIONLESS_REMOTECHANNEL_STATE_TEP_NEGOTIATION)
    {
        PCR_Assert(self->superclass.remoteChannelData->flags.lossySession);

        if (self->superclass.remoteChannelData->flags.isOutbound)
        {
            /* Outbound statefull datagram connections must send a query after security negotiation.*/
            self->state = CONNECTIONLESS_REMOTECHANNEL_STATE_APPID_DISCOVERY;
            OALRemoteChannelMgr_CancelNegotiationTimer(&self->superclass);
            self->superclass.remoteChannelData->flags.sessionIsSecure = TRUE;
            self->superclass.remoteChannelData->flags.mustTerminate = TRUE;
            self->superclass.remoteChannelData->flags.waitingForQueryResponse = TRUE;
            ESP_V2_SendQuery(&self->superclass, 30000, 5000);   /* 30 second timeout with 5 second retry. */
            OALMessage_Destroy(message);
        }
        else
        {
            self->state = CONNECTIONLESS_REMOTECHANNEL_STATE_ESTABLISHED;
            OALConnectionlessRemoteChannel_InitializeComplete_Notify(self);
            OALMessage_Destroy(message);
        }
    }
    else if (self->state == CONNECTIONLESS_REMOTECHANNEL_STATE_APPID_DISCOVERY)
    {
        PCR_Assert(self->superclass.remoteChannelData->flags.isOutbound);

        self->state = CONNECTIONLESS_REMOTECHANNEL_STATE_ESTABLISHED;

        ESP_V2_SendQuery(&self->superclass, 8000, 3000);

        if (self->superclass.remoteChannelData->negotiator == DOF_DEFAULT_AS_PROTOCOL_NEGOTIATOR)
        {
            EPP_NegOne_Heartbeat(&self->superclass, self->superclass.remoteChannelData->address, FALSE);
        }
        else
        {
            /* Send an OAP no-op to establish the OAP version that will be used for the connection. */
            DOFOperation_SendOAPNoOp(&self->superclass);
        }

        OALConnectionlessRemoteChannel_InitializeComplete_Notify(self);
        OALMessage_Destroy(message);
    }
    else if (self->state == CONNECTIONLESS_REMOTECHANNEL_STATE_ESTABLISHED)
    {
        /* Nothing to do, just destroy the message and exit. */
        OALMessage_Destroy(message);
    }
    else
    {
        self->state = CONNECTIONLESS_REMOTECHANNEL_STATE_ESTABLISHED;
        OALConnectionlessRemoteChannel_InitializeComplete_Notify(self);
        OALMessage_Destroy(message);
    }
}

static void OALConnectionlessRemoteChannel_HandleMessage_Initialize( OALConnectionlessRemoteChannel self, OALMessage message )
{
    if (message->shutdown)
    {
        OALMessage_Destroy(message);
    }
    else
    {
        /*  If a channel has a base channel, set the DTL for the channel to be the same as the base channel.  If the
            base channel does not have a DTL then call shutdown for the channel. */
        if (self->superclass.remoteChannelData->flags.setDtlInChannelInit)
        {
            self->superclass.remoteChannelData->dtl = self->superclass.remoteChannelData->baseChannel->remoteChannelData->dtl;
            if (!DTL_Retain(self->superclass.remoteChannelData->dtl))
                OALRemoteChannel_Shutdown(&self->superclass, DOFEXCEPTION_FAILED);
        }

        if (self->superclass.remoteChannelData->flags.isConnection &&
            self->superclass.remoteChannelData->flags.defaultAddressIsMulticast &&
            self->superclass.remoteChannelData->baseChannel)
        {
            /* Save this channel in the base channel so that multicast traffic can be handed to this connection when it received by the server. */
            self->superclass.remoteChannelData->baseChannel->remoteChannelData->multicastChannel = &self->superclass;
        }

        if (self->superclass.remoteChannelData->flags.isConnection && !self->superclass.remoteChannelData->flags.defaultAddressIsMulticast)
        {
            if (self->superclass.remoteChannelData->flags.isOutbound)
            {
                OALRemoteChannelMgr_StartNegotiationTimeout(&self->superclass, ESP_V2_NEGOTIATION_TIMEOUT);
                ESP_V2_SendOpen(&self->superclass);
            }
            OALMessage_Destroy(message);
        }
        else
        {
            /* OALConnectionlessRemoteChannel_InitializeComplete will re-use or destroy message. */
            OALConnectionlessRemoteChannel_InitializeComplete(self, message);
        }
    }
}

/* Note: This will handle a successful initialize.  Shutdown will be called for failures. */
static void OALConnectionlessRemoteChannel_HandleMessage_InitializeComplete( OALConnectionlessRemoteChannel self, OALMessage message )
{
    /*  Use the common initialize complete function unless it is something that will only be done when the
        initialized complete message is received. */

    if (message->shutdown)
    {
        OALMessage_Destroy(message);
    }
    else
    {
        OALConnectionlessRemoteChannel_InitializeComplete(self, message);
    }
}

static void OALConnectionlessRemoteChannel_Process_Shutdown( OALConnectionlessRemoteChannel self, DOFException exception)
{
    if(self->superclass.remoteChannelData->dtl)
    {
        OALRemoteChannel_ShutdownTransport(&self->superclass);
    }

    OALConnectionlessRemoteChannel_Shutdown( self, exception );
}

static void OALConnectionlessRemoteChannel_HandleMessage_Shutdown( OALConnectionlessRemoteChannel self, OALMessage message )
{
    DOFException exception = (DOFException)message->argv[0];

    OALConnectionlessRemoteChannel_Process_Shutdown(self, exception);

    OALMessage_Destroy( message );
}

static void OALConnectionlessRemoteChannel_HandleMessage_TransportDown( OALConnectionlessRemoteChannel self, OALMessage message )
{
    DTL transport = message->argv[0];
    DOFException exception = message->argv[1];

    if(transport == self->superclass.remoteChannelData->dtl)
        OALConnectionlessRemoteChannel_Shutdown( self, exception );

    OALMessage_Destroy( message );
}

static void OALConnectionlessRemoteChannel_HandleMessage_ProcessBlockedMessage(OALConnectionlessRemoteChannel self, OALMessage message)
{
    if (message->shutdown)
    {
        OALMessage_Destroy(message);
    }
    else
    {
        OALMessage msg;

        /* Clear the argv[0] parameter for the message in case it is sent again. */
        message->argv[0] = (void *)0;
        if (self->superclass.remoteChannelData->blockedMessageQueue)
        {
            uint32 size = OALQueue_GetSize( self->superclass.remoteChannelData->blockedMessageQueue );

            if (size > 0)
            {
                /* Remove the first message from the queue and send it to the stack. */
                msg = (OALMessage)OALQueue_Dequeue( self->superclass.remoteChannelData->blockedMessageQueue );
                msg->already = message->already;
                OALRemoteChannelMgr_SendMessageToStack( msg );
            }
            /* If the queue is not empty then re-send the trigger message; destroy the message if it is empty. */
            if (size > 1)
            {
                OALRemoteChannelMgr_SendMessageToChannel_Internal( message, TRUE );
            }
            else
            {
                OALMessage_Destroy( message );
                OALQueue_Destroy(self->superclass.remoteChannelData->blockedMessageQueue);
                self->superclass.remoteChannelData->blockedMessageQueue = NULL;
            }
        }
        else
        {
            OALMessage_Destroy( message );
        }
    }
}

static void OALConnectionlessRemoteChannel_HandleMessage_ConnectionInitiatedShutdown(OALConnectionlessRemoteChannel self, OALMessage message)
{
    /* Clear the reference to the OALConnection.  This does not require a release for outbound connections.
       Note: Clearing the OALConn will prevent the channel from notifing the OALConn that it was shutdown.
       This prevents a race condition where the OALConn can be destroyed before the notification from
       the channel is received. */
    if (!self->superclass.remoteChannelData->flags.isOutbound)
    {
        /* For inbound connections The channel actually retains the OALConnection. */
        OALConnection_Release(self->superclass.remoteChannelData->OALConn);
    }
    PCRMutex_Lock(connection_server_mutex);
    self->superclass.remoteChannelData->OALConn = NULL;
    PCRMutex_Unlock(connection_server_mutex);

    OALConnectionlessRemoteChannel_Process_Shutdown(self, NULL);

    OALMessage_Destroy(message);
}

static void OALConnectionlessRemoteChannel_HandleMessage_RegisterSecureChannel(OALConnectionlessRemoteChannel self, OALMessage message)
{
    if (!message->shutdown)
    {
        OALRemoteChannel_HandleMessage_RegisterSecureChannel((OALChannel)&self->superclass, message);
    }

    OALMessage_Destroy(message);
}

static void OALConnectionlessRemoteChannel_HandleMessage_APPIDQueryTimeout(OALConnectionlessRemoteChannel self, OALMessage message)
{
    if (self->superclass.remoteChannelData->flags.waitingForInitialDatagram ||
        self->superclass.remoteChannelData->blockedMessageQueue)
    {
        /*  If the inbound statefull datagram connection is still waiting for the initial datagram,
            close the channel. */
        OALConnectionlessRemoteChannel_Process_Shutdown(self, DOFEXCEPTION_PROTOCOL_NEGOTIATION_FAILED);
    }
    OALMessage_Destroy(message);
}

static void OALConnectionlessRemoteChannel_HandleMessage_RegisterChannel_ENPPort(OALConnectionlessRemoteChannel self, OALMessage message)
{
    if (!message->shutdown)
    {
        OALRemoteChannel_HandleMessage_RegisterChannel_ENPPort((OALChannel)&self->superclass, message);
    }

    OALMessage_Destroy(message);
}

static void OALConnectionlessRemoteChannel_HandleMessage_RemoteDomainCallback(OALConnectionlessRemoteChannel self, OALMessage message)
{
    /* Todo: At one point there was a seperate check for resolving credentials and checking for a remote domain.
       Now, resolving credentials will take care of both, so at some point the RemoteDomainCallback can be removed. */
    if (message->argv[0])
    {
        OALChannel contextChannel = OALChannel_Retain(&self->superclass);

        if (contextChannel)
        {
            self->superclass.remoteChannelData->secureCredentials = (DOFCredentials)message->argv[0];
            message->argv[0] = NULL;
            OALRemoteChannel_CreateLibPermissions(OALConnectionlessRemoteChannel_GetLibraryPermsCallback, self,
                                                  self->superclass.remoteChannelData->secureCredentials,
                                                  &self->superclass, self->superclass.remoteChannelData->address);
        }
        else
        {
            OALRemoteChannel_Shutdown(&self->superclass, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        }
    }
    else
    {
        /* If credentials are not included, just shut down the channel. */
        OALRemoteChannel_Shutdown(&self->superclass, DOFEXCEPTION_AUTHENTICATION_FAILED);
    }

    OALMessage_Destroy(message);
}

static void OALConnectionlessRemoteChannel_HandleMessage_ResolveCredsCallback(OALConnectionlessRemoteChannel self, OALMessage message)
{
    if (message->argv[0])
    {
        OALChannel contextChannel = OALChannel_Retain(&self->superclass);

        if (contextChannel)
        {
            self->superclass.remoteChannelData->secureCredentials = (DOFCredentials)message->argv[0];
            message->argv[0] = NULL;
            OALRemoteChannel_CreateLibPermissions(OALConnectionlessRemoteChannel_GetLibraryPermsCallback, self,
                                                  self->superclass.remoteChannelData->secureCredentials,
                                                  &self->superclass, self->superclass.remoteChannelData->address);
        }
        else
        {
            OALRemoteChannel_Shutdown(&self->superclass, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        }
    }
    else
    {
        /* If credentials are not included, just shut down the channel. */
        OALRemoteChannel_Shutdown(&self->superclass, DOFEXCEPTION_AUTHENTICATION_FAILED);
    }

    OALMessage_Destroy(message);
}

static void OALConnectionlessRemoteChannel_HandleMessage_GetLibraryPermsCallback(OALConnectionlessRemoteChannel self, OALMessage message)
{
    if (message->argv[1])
    {
        self->superclass.remoteChannelData->secureSid = (sid_t *)message->argv[1];
        message->argv[1] = NULL;

        OALConnectionlessProtocolStack_StartTepNegotiation(self->superclass.remoteChannelData->protocolStack, &self->superclass,
                                                           (DOFPermissionSet)message->argv[0]);
    }
    else
    {
        OALRemoteChannel_Shutdown(&self->superclass, DOFEXCEPTION_AUTHENTICATION_FAILED);
    }

    OALMessage_Destroy(message);
}

static void OALConnectionlessRemoteChannel_HandleMessage_ServerInitiatedShutdown(OALConnectionlessRemoteChannel self, OALMessage message)
{
    /* The server initiated shutdown message is sent to make sure that the server is not destroyed before data
       received by the server is processed.  The message retains the server and prevents it from being destroyed
       before the reference the channel has to the server is cleared.  This is necessary since it is possible for
       the app to destroy the server before a receive from transport message is processed. */
    OALConnectionlessRemoteChannel_Process_Shutdown(self, NULL);
    self->superclass.remoteChannelData->server = NULL;

    OALMessage_Destroy(message);
}

static void OALConnectionlessRemoteChannel_HandleMessage_ProcessReceiveSilence(OALConnectionlessRemoteChannel self, OALMessage message)
{
    if (!message->shutdown)
    {
        OALRemoteChannel_HandleMessage_ProcessReceiveSilence(&self->superclass, message);
    }
    OALMessage_Destroy(message);
}

static void OALConnectionlessRemoteChannel_HandleMessage_SetDomainCredentialsCallback(OALConnectionlessRemoteChannel self, OALMessage message)
{
    if (!message->shutdown)
    {
        OALRemoteChannel_HandleMessage_SetDomainCredentialsCallback(&self->superclass, message);
    }
    OALMessage_Destroy(message);
}

static void OALConnectionlessRemoteChannel_HandleMessage_PeerListenerEvent(OALConnectionlessRemoteChannel self, OALMessage message)
{
    OALRemoteChannel_HandleMessage_PeerListenerEvent((OALChannel)self, message);
    OALMessage_Destroy(message);
}

static void OALConnectionlessRemoteChannel_HandleMessage_CreateBridge(OALConnectionlessRemoteChannel self, OALMessage message)
{
    if (!message->shutdown)
    {
        OALMessage_CreateBridgeMessage msg = (OALMessage_CreateBridgeMessage)message;

        OALRemoteChannel_CreateBridge(&self->superclass, self->superclass.remoteChannelData->address, msg->dof, msg->bridgeConfig);
    }
    OALMessage_Destroy(message);
}

static void OALConnectionlessRemoteChannel_HandleMessage_CreateBridgeComplete(OALConnectionlessRemoteChannel self, OALMessage message)
{
    if (!message->shutdown)
    {
        OALMessage_CreateBridgeCompleteMessage msg = (OALMessage_CreateBridgeCompleteMessage)message;

        PCR_Assert(DOFSecurityMembrane_IsChannelMembrane(self->superclass.securityMembrane));
        OALChannelMembrane_SetBridge(self->superclass.securityMembrane, msg->bridge);

        if (self->superclass.remoteChannelData->flags.isConnection &&
            ((DOFConnectionConfig_GetConnectionType(self->superclass.remoteChannelData->config.connection) == DOFCONNECTIONTYPE_STREAM) ||
             (DOFConnectionConfig_GetConnectionType(self->superclass.remoteChannelData->config.connection) == DOFCONNECTIONTYPE_DATAGRAM)))
        {
            /* Initialize complete was already processed for this, so just call the notify function. */
            OALConnectionlessRemoteChannel_InitializeComplete_Notify(self);
        }
        else
        {
            OALMessage newMessage;

            /* Now that the bridge is set up, send a message to start initialization. */
            newMessage = CMT_Initialize_Create(&self->superclass);
            if (newMessage)
            {
                OALRemoteChannelMgr_SendMessageToChannel(newMessage);
            }
            else
            {
                OALRemoteChannel_Shutdown(&self->superclass, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            }
        }
    }
    OALMessage_Destroy(message);
}

static void OALConnectionlessRemoteChannel_HandleMessage_ConnectionStreamListener(OALConnectionlessRemoteChannel self, OALMessage message)
{
    OALRemoteChannel_HandleMessage_ConnectionStreamListener((OALChannel)self, message);
    OALMessage_Destroy(message);
}

static void OALConnectionlessRemoteChannel_HandleMessage( OALChannel self, OALMessage message )
{
    static OCLRC_MessageHandler_fn_t msg_handler_funcs[] = {
        /* the order here is coupled with ChannelMsgType_t. So dont change it. */
        OALConnectionlessRemoteChannel_HandleMessage_RouteToProvider,
        OALConnectionlessRemoteChannel_HandleMessage_RouteToRequestor,
        OALConnectionlessRemoteChannel_HandleMessage_ReceiveFromTransport,
        OALConnectionlessRemoteChannel_HandleMessage_Initialize,
        OALConnectionlessRemoteChannel_HandleMessage_InitializeComplete,
        OALConnectionlessRemoteChannel_HandleMessage_Shutdown,
        OALConnectionlessRemoteChannel_HandleMessage_TransportDown,
        OALConnectionlessRemoteChannel_HandleMessage_ProcessBlockedMessage,
        OALConnectionlessRemoteChannel_HandleMessage_ConnectionInitiatedShutdown,
        OALConnectionlessRemoteChannel_HandleMessage_RegisterSecureChannel,
        OALConnectionlessRemoteChannel_HandleMessage_APPIDQueryTimeout,
        OALConnectionlessRemoteChannel_HandleMessage_RegisterChannel_ENPPort,
        OALConnectionlessRemoteChannel_HandleMessage_RemoteDomainCallback,
        OALConnectionlessRemoteChannel_HandleMessage_ResolveCredsCallback,
        OALConnectionlessRemoteChannel_HandleMessage_GetLibraryPermsCallback,
        OALConnectionlessRemoteChannel_HandleMessage_ServerInitiatedShutdown,
        OALConnectionlessRemoteChannel_HandleMessage_ProcessReceiveSilence,
        OALConnectionlessRemoteChannel_HandleMessage_SetDomainCredentialsCallback,
        OALConnectionlessRemoteChannel_HandleMessage_PeerListenerEvent,
        OALConnectionlessRemoteChannel_HandleMessage_CreateBridge,
        OALConnectionlessRemoteChannel_HandleMessage_CreateBridgeComplete,
        OALConnectionlessRemoteChannel_HandleMessage_ConnectionStreamListener
    };
    STATIC_ASSERT( (sizeof( msg_handler_funcs ) / sizeof( OCLRC_MessageHandler_fn_t )) ==
                   (CMT_LAST - CMT_ROUTE_TO_PROVIDER) );   /* These are in ChannelMsgType_t. */

    PCR_Assert( self );
    PCR_Assert( message );
    PCR_Assert( message->recipient == self );
    PCR_Assert( message->msg_type < (sizeof(msg_handler_funcs) / sizeof(OCLRC_MessageHandler_fn_t)) );

    msg_handler_funcs[message->msg_type]( (OALConnectionlessRemoteChannel)self, message );
}

void OALConnectionlessRemoteChannel_RemoteDomainCallback(TRP_V1_Context trp_context, void * context)
{
    OALMessage msg = NULL;
    DOFException except = DOFEXCEPTION_INSUFFICIENT_RESOURCES;

    if(trp_context != NULL)
    {
        if(trp_context->RemoteDomainReturn->status == DOFERROR_OK)
        {
            if (trp_context->RemoteDomainReturn->errorCode == TRP_V1_SUCCESS)
            {
                PCR_Assert(trp_context->RemoteDomainReturn->creds != NULL);
                /* Context holds the channel, for connectionless send this message to the channel. */
                msg = CMT_RemoteDomainCallback_Create(context, trp_context->RemoteDomainReturn->creds);
            }
            else
            {
                except = DOFErrorException_GetFromError(TRP_V1_GetDOFErrorFromErrorCode(trp_context->RemoteDomainReturn->errorCode));
            }
        }
        else
        {
            /* Map the DOFError to a DOFException. */
            except = DOFErrorException_GetFromError(trp_context->RemoteDomainReturn->status);
        }
    }

    if (msg)
    {
        OALRemoteChannelMgr_SendMessageToChannel(msg);
    }
    else
    {
        OALRemoteChannel_Shutdown((OALChannel)context, except);
    }
}

void OALConnectionlessRemoteChannel_ResolveCredsCallback( DOFCredentials self, DOFException exception, void* context )
{
    OALMessage msg = NULL;
    DOFException except = DOFEXCEPTION_INSUFFICIENT_RESOURCES;

    if( !exception )
    {
        /* Context holds the channel, for connectionless send this message to the channel. */
        msg = CMT_ResolveCredsCallback_Create( context, self );
    }
    else
    {
        except = exception;
    }

    if (msg)
    {
        OALRemoteChannelMgr_SendMessageToChannel(msg);
    }
    else
    {
        OALRemoteChannel_Shutdown((OALChannel)context, except);
    }
}

void OALConnectionlessRemoteChannel_GetLibraryPermsCallback(DOFError           status,
                                                            sid_t            * secureSid,
                                                            DOFPermissionSet   libPerms,
                                                            void             * context)
{
    OALMessage msg = NULL;
    DOFException except = DOFEXCEPTION_INSUFFICIENT_RESOURCES;

    PCR_Assert(context != NULL);
    if(status == DOFERROR_OK)
    {
        PCR_Assert(secureSid != NULL);
        msg = CMT_GetLibraryPermsCallback_Create(context, libPerms, secureSid);
    }
    else
    {
        except = DOFErrorException_GetFromError(status);
    }

    if (msg)
    {
        OALRemoteChannelMgr_SendMessageToChannel(msg);
    }
    else
    {
        OALRemoteChannel_Shutdown((OALChannel)context, except);
    }
    /* The channed was retained before calling OALRemoteChannel_CreateLibPermissions, Release it now. */
    OALChannel_Release((OALChannel)context);
}

TEP_V1_SessionEventCallback OALConnectionlessRemoteChannel_GetSessionEventCallback(OALConnectionlessRemoteChannel self)
{
    return self->sessionEventCallback;
}

static void* OALConnectionlessRemoteChannel_TEPSessionEventCallback(void               * context,
                                                                    uint8              * key,
                                                                    TEP_SessionEvent_t   eventType,
                                                                    DOFSecurityMode      securityMode,
                                                                    OALChannel           ch,
                                                                    DOFSecurityScope     secScope,
                                                                    DOFPermissionSet     perms,
                                                                    DOFPermissionSet     otherPerms,
                                                                    DOFException         exception,
                                                                    uint32               stateID,
                                                                    DOFAddress           remoteAddr,
                                                                    OALStream          * initialState,
                                                                    DOFCredentials       creds)
{
    OALConnectionlessRemoteChannel channel = (OALConnectionlessRemoteChannel)ch;

    switch (eventType)
    {
    case TEP_V1_SessionEvent_NEWSESSION_SUCCESS_RECVKEY:
        {
            boolean      tepResponder;
            OALCipher    cipher;
            DOFException except = NULL;

            PCR_Assert(channel);
            PCR_Assert(channel->superclass.remoteChannelData);
            PCR_Assert(securityMode);
            PCR_Assert(perms);
            PCR_Assert(otherPerms);
            PCR_Assert(secScope);
            tepResponder = channel->superclass.remoteChannelData->flags.isOutbound ? FALSE : TRUE;
            cipher = OALCipher_Create_TEP( securityMode, TRUE, tepResponder, &except );
            PCR_Assert(initialState);
            if (cipher)
            {
                OALCipher_SetManager( cipher, tepResponder );
                /* As TEP responder, initialState we create.
                   As TEP initiator, initialState is supplied to us by the TEP Responder. It may specify a targetNodeID.
                      If so, it is OUR CCM nodeID. The SetFutureKey() call below checks it, and if given will properly
                      overwrite our CCM nodeID, even though the _Create() call gave us one already.
                   NOTE: In our implementation the TEP responder does not assign a nodeID to the TEP initiator. */
                if (tepResponder)
                {
                    PCR_Assert( *initialState == NULL );
                    *initialState = OALCipher_CreateInitialStateStream_TEP( cipher );
                }
                if (*initialState != NULL)
                {
                    if(OALCipher_SetFutureKey(cipher, *initialState, key, perms, otherPerms))
                    {
                        if(channel->superclass.remoteChannelData->mode)
                        {
                            DOFSecurityMode_Destroy(channel->superclass.remoteChannelData->mode);
                        }
                        channel->superclass.remoteChannelData->mode = DOFSecurityMode_Copy(securityMode);
                        if (channel->superclass.securityScope)
                        {
                            DOFSecurityScope_Destroy(channel->superclass.securityScope);
                        }
                        channel->superclass.securityScope = DOFSecurityScope_Copy(secScope);

                        if (channel->superclass.remoteChannelData->mode &&
                            channel->superclass.securityScope)
                        {
                            if(channel->superclass.remoteChannelData->cipher)
                                OALCipher_Destroy(channel->superclass.remoteChannelData->cipher);
                            channel->superclass.remoteChannelData->cipher = cipher;
                        }
                        else
                        {
                            OALRemoteChannel_Shutdown(ch, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
                            return NULL;
                        }
                    }
                    else
                    {
                        OALRemoteChannel_Shutdown(ch, DOFEXCEPTION_AUTHENTICATION_FAILED);
                        return NULL;
                    }
                }
                else
                {
                    OALRemoteChannel_Shutdown(ch, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
                    return NULL;
                }
            }
            else
            {
                OALRemoteChannel_Shutdown(ch, except);
                return NULL;
            }
            DOFException_Destroy( except );
            channel->sessionEventCallback = OALConnectionlessRemoteChannel_TEPSessionEventCallback;
        }
        break;
    case TEP_V1_SessionEvent_NEWSESSION_SUCCESS_SENDKEY:
        {
            KeyChangeInfo_t* pKeyChangeInfo = OALRemoteChannel_CreateKeyChangeInfo( &channel->superclass,
                                                      channel->superclass.remoteChannelData->cipher,
                                                      stateID, NULL, NULL, NULL, NULL, FALSE, TRUE );

            if(pKeyChangeInfo != NULL)
            {
                OALProtocolStack_SetKeyChangeCallback( channel->superclass.remoteChannelData->protocolStack, OALRemoteChannel_KeyChangeCallback, pKeyChangeInfo );
            }
            else
            {
                DOFException except = DOFErrorException_Create(DOFERROR_FAILED, "Received Rekey Failure.");
                OALRemoteChannel_Shutdown(ch, except);
                DOFException_Destroy(except);
            }
        }
        break;

    case TEP_V1_SessionEvent_NEWSESSION_FINALIZE:

        PCR_Assert(DOFSecurityMembrane_IsChannelMembrane(channel->superclass.securityMembrane));
        if(!channel->superclass.remoteChannelData->flags.isOutbound)
        {
            OALOperationBridge bridge = OALOperationBridge_CreatePassThroughBridge(&(channel->superclass));

            if (bridge)
            {
                DOFSecurityMembrane_Destroy(channel->superclass.securityMembrane);
                channel->superclass.securityMembrane = DOFOperationBridgeConfig_GetMembrane(bridge);
            }
            else
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessRemoteChannel: Failed to create pass through bridge (inbound)." );
                OALRemoteChannel_Shutdown(&channel->superclass, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            }
        }
        else if (OALChannelMembrane_GetBridge(channel->superclass.securityMembrane) == DOF_GetDefaultOperationBridge(DOF_GetDof()))
        {
            OALOperationBridge bridge = OALOperationBridge_CreatePassThroughBridge(&channel->superclass);

            if (bridge)
            {
                OALChannelMembrane_SetBridge(channel->superclass.securityMembrane, bridge);
                OALOperationBridge_Destroy(bridge);
            }
            else
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessRemoteChannel: Failed to create initialState for rekey." );
                OALRemoteChannel_Shutdown(&channel->superclass, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            }
        }

        if (channel->superclass.securityMembrane)
        {
            channel->superclass.remoteChannelData->flags.sessionIsSecure = TRUE;
            OALSource_Clear(&channel->superclass, channel->superclass.remoteChannelData->address);
            OALRemoteChannel_NotifyInitComplete(&channel->superclass);
        }
        else
        {
            OALRemoteChannel_Shutdown(&channel->superclass, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        }
        break;
    case TEP_V1_SessionEvent_REKEY_RECVKEY:
        if (*initialState == NULL)
            *initialState = OALCipher_CreateInitialStateStream_TEP( channel->superclass.remoteChannelData->cipher );
        if (*initialState == NULL)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessRemoteChannel: Failed to create initialState for rekey." );
            OALRemoteChannel_Shutdown(ch, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        }
        else
        {
            if (!OALCipher_SetFutureKey(channel->superclass.remoteChannelData->cipher, *initialState, key, perms, otherPerms))
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessRemoteChannel: Failed to set future key rekey recvkey." );
                OALRemoteChannel_Shutdown(ch, DOFEXCEPTION_FAILED);
            }
        }
        break;
    case TEP_V1_SessionEvent_REKEY_SENDKEY:
        {
            KeyChangeInfo_t* pKeyChangeInfo;

            PCR_Assert(perms);
            PCR_Assert(otherPerms);

            pKeyChangeInfo = OALRemoteChannel_CreateKeyChangeInfo(&channel->superclass, channel->superclass.remoteChannelData->cipher, stateID, perms, otherPerms, NULL, NULL, FALSE, TRUE);

            if(pKeyChangeInfo != NULL)
            {
                OALProtocolStack_SetKeyChangeCallback( channel->superclass.remoteChannelData->protocolStack, OALRemoteChannel_KeyChangeCallback, pKeyChangeInfo );
            }
            else
            {
                OALRemoteChannel_Shutdown(ch, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            }
        }
        break;
    case TEP_V1_SessionEvent_REKEY_SENDKEYANDDISABLE:
        {
            KeyChangeInfo_t* pKeyChangeInfo;

            PCR_Assert(secScope);

            pKeyChangeInfo = OALRemoteChannel_CreateKeyChangeInfo(&channel->superclass, channel->superclass.remoteChannelData->cipher, stateID, NULL, NULL, NULL, secScope, TRUE, TRUE);

            if(pKeyChangeInfo != NULL)
            {
                OALProtocolStack_SetKeyChangeCallback( channel->superclass.remoteChannelData->protocolStack, OALRemoteChannel_KeyChangeCallback, pKeyChangeInfo );
            }
            else
            {
                OALRemoteChannel_Shutdown(ch, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            }
        }
        break;
    case TEP_V1_SessionEvent_REKEY_DISABLEOLD:
        {
            KeyChangeInfo_t* pKeyChangeInfo;

            PCR_Assert(secScope);

            pKeyChangeInfo = OALRemoteChannel_CreateKeyChangeInfo(&channel->superclass, channel->superclass.remoteChannelData->cipher, stateID, NULL, NULL, NULL, secScope, TRUE, FALSE);

            if(pKeyChangeInfo != NULL)
            {
                OALProtocolStack_SetKeyChangeCallback( channel->superclass.remoteChannelData->protocolStack, OALRemoteChannel_KeyChangeCallback, pKeyChangeInfo );
            }
            else
            {
                OALRemoteChannel_Shutdown(ch, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            }
        }
        break;

    case TEP_V1_SessionEvent_NEWSESSION_FAILURE:
    case TEP_V1_SessionEvent_REKEY_FAILURE:
        PCR_Assert(exception != NULL);
        OALRemoteChannel_Shutdown(ch, exception);
        break;

    case TEP_V1_SessionEvent_ENDING:
    {
        DOFException except = DOFErrorException_Create(DOFERROR_FAILED, "Secure connection ending.");
        OALRemoteChannel_Shutdown(ch, except);
        DOFException_Destroy(except);
        break;
    }


    default:
        PCR_Assert(FALSE);
    }
    return context;
}
