/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/connection.h>
#include <dof/operation.h>
#include <dof/server.h>
#include <dof/cfg.h>

#include <internal/addressinternal.h>
#include <internal/channel_stack_messages.h>
#include <internal/cipher.h>
#include <internal/connection_internal.h>
#include <internal/credentialsinternal.h>
#include <internal/dofinternal.h>
#include <internal/enp_v1.h>
#include <internal/epp_negative_one.h>
#include <internal/exception.h>
#include <internal/mutex_pool.h>
#include <internal/permissionsetinternal.h>
#include <internal/remotechannel.h>
#include <internal/remotechannelmanager.h>
#include <internal/router.h>
#include <internal/serverinternal.h>
#include <internal/tep_v1.h>
#include <internal/transportinternal.h>
#include <internal/trp_v1.h>

#include "channel_local.h"

static PCRMutex channelShutdownMutex = NULL;
extern PCRMutex connection_server_mutex;

struct traverseBySSIDContext
{
    remotechannel_ssid_traverse_fn_t fn;
    void * context;
};

boolean OALRemoteChannel_MutexInit(void)
{
    /* This is created as part of DOF_Init. */
    channelShutdownMutex = PCRMutex_Create();
    if (channelShutdownMutex)
        return TRUE;

    return FALSE;
}

void OALRemoteChannel_MutexShutdown(void)
{
    PCRMutex_Destroy(channelShutdownMutex);
    channelShutdownMutex = NULL;
}

static void OALRemoteChannel_OALDomainCredentialsTree_ValidateDomainCredentials_Callback( DOFException exception, void* context )
{
    DOFConnection connection = (DOFConnection)context;
    DOFConnection_ValidateComplete( connection, exception );
}

/* @internal
   Note: This function should only be called on success.  If Initialization fails call OALRemoteChannel_Shutdown. */
void OALRemoteChannel_HandleMessage_InitializeComplete( OALChannel self, OALMessage message )
{
    if (self->remoteChannelData->flags.validateDomainCreds && self->remoteChannelData->OALConn)
    {
        DOFConnection conn;

        /* Validate the domain credentials.  Since this will be called again after the validate is successfull clear the flag. */
        self->remoteChannelData->flags.validateDomainCreds = FALSE;
        conn = OALConnection_CreateDOFConnection(self->remoteChannelData->OALConn);
        if (conn)
        {
            /* Validate the doamin credentials.  If there is an error the channel will be shutdown (OALRemoteChannel_Shutdown).
               If the credentials are validated a new intialize complete message will be sent and this function will be called again. */
            OALDomainCredentialsTree_ValidateDomainCredentials( self->remoteChannelData->OALConn->connConfig->domainCreds, conn, OALRemoteChannel_OALDomainCredentialsTree_ValidateDomainCredentials_Callback, (void*)conn );
            DOFConnection_Destroy(conn);
        }
        else
        {
            /* The credentials cannot be validated due to the failure to create the connection. */
            DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_TRACE, "OALRemoteChannel: OALRemoteChannel_HandleMessage_InitializeComplete, failed to create connection");
            OALRemoteChannel_Shutdown(self, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        }

    }
    else
    {
        OALRemoteChannelMgr_SetSendSilenceTimerItem( self );
        OALRemoteChannelMgr_SetRecvSilenceTimerItem( self );

        if (self->remoteChannelData->OALConn)
            OALConnection_ConnectionEvent(self->remoteChannelData->OALConn, self, TRUE, NULL);
    }
}

void OALRemoteChannel_Init(OALChannel self)
{
    DOFOperationBridgeConfig bridge = NULL;
    OALMessage msg;

    if(self->remoteChannelData->flags.isConnection)
    {
        DOFConnectionType connType = DOFConnectionConfig_GetConnectionType(self->remoteChannelData->config.connection);

        if ((connType != DOFCONNECTIONTYPE_STREAM) && (connType != DOFCONNECTIONTYPE_DATAGRAM))
        {
            bridge = DOFConnectionConfig_GetBridge(self->remoteChannelData->config.connection);
        }
    }
    else
    {
        DOFSecurityDesire securityDesire;

        /* If it is possible to be unsecure, see if a bridge config is specified for the server. */
        securityDesire = DOFServerConfig_GetSecurityDesire(self->remoteChannelData->config.server);
        if ((securityDesire == DOFSECURITYDESIRE_ANY) || (securityDesire == DOFSECURITYDESIRE_NOT_SECURE))
        {
            bridge = DOFServerConfig_GetBridge(self->remoteChannelData->config.server);
        }
    }

    if(bridge != NULL)
    {
        msg = CMT_CreateBridge_Create(self, bridge, DOF_GetDof());
    }
    else
    {
        msg = CMT_Initialize_Create( self );
    }

    if ( msg )
        OALRemoteChannelMgr_SendMessageToChannel( msg );
    else
        OALRemoteChannel_Shutdown(self, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
}

/* Note: This will handle a successful initialize.  Shutdown should be called for failures. */
void OALRemoteChannel_NotifyInitComplete(OALChannel self)
{
    OALMessage msg = CMT_InitializeComplete_Create(self);

    if ( msg )
        OALRemoteChannelMgr_SendMessageToChannel( msg );
    else
        OALRemoteChannel_Shutdown(self, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
}

void OALRemoteChannel_ShutdownTransport(OALChannel self)
{
    if( !self->remoteChannelData->flags.isConnection ||(self->remoteChannelData->flags.isConnection && !self->remoteChannelData->baseChannel))
    {
        DTL_StopClose( self->remoteChannelData->dtl );
    }
}

static void sendShutdownRelatedCommands(OALChannel channel)
{
    if (channel->remoteChannelData && channel->remoteChannelData->flags.isConnection)
    {
        DOF dof;

        dof = DOF_GetDof();
        if (dof && DOF_IsNodeDown( dof ) )
            EPP_NegOne_NodeDown(channel);

        if(channel->remoteChannelData->flags.mustFinalizeTep)
        {
            TEP_V1_EndSession(channel);
        }
        if (channel->remoteChannelData->flags.mustTerminate)
        {
            ESP_V2_SendTerminate(channel);
        }
    }
}

void OALRemoteChannel_Shutdown(OALChannel self, DOFException exception)
{
    OALMessage msg = NULL;

    PCR_Assert(self->remoteChannelData->shutdownMessage);

    PCRMutex_Lock(channelShutdownMutex);

    if (!self->remoteChannelData->flags.shutdownMessageSent)
    {
        self->remoteChannelData->flags.shutdownMessageSent = TRUE;
        CMT_Shutdown_SetException(self->remoteChannelData->shutdownMessage, exception);
        sendShutdownRelatedCommands(self);

        /* If this fails it indicates that a shutdown message has already been sent so its ok to drop this one. */
        msg = OALMessage_Retain(self->remoteChannelData->shutdownMessage);
    }

    PCRMutex_Unlock(channelShutdownMutex);

    if (msg)
    {
        /*  This may fail, but if it does it is because the DOF is being shut down, in which case the channel
            will be shutdown when the remotechannel manager shuts down. msg will be destroyed if it fails. */
        OALRemoteChannelMgr_SendMessageToChannel(msg);
    }
}

OALMessage OALRemoteChannel_CreateShutdownMessage(OALChannel self)
{
    return CMT_Shutdown_Create(self, NULL);
}

void OALRemoteChannel_ConnectionInitiatedShutdown(OALChannel self, OALConnection OALConn)
{
    OALMessage msg = CMT_ConnectionInitiatedShutdown_Create(self, OALConn);
    if ( msg )
    {
        sendShutdownRelatedCommands(self);
        OALRemoteChannelMgr_SendMessageToChannel( msg );
    }
    else
    {
        if (!self->remoteChannelData->failedCIShutdown && OALConnection_Retain(OALConn))
            self->remoteChannelData->failedCIShutdown = OALConn;
        OALRemoteChannel_Shutdown(self, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
    }
}

void OALRemoteChannel_ServerInitiatedShutdown(OALChannel self, OALServer OALSrv)
{
    OALMessage msg = CMT_ServerInitiatedShutdown_Create(self, OALSrv);
    if (msg)
    {
        OALRemoteChannelMgr_SendMessageToChannel(msg);
    }
    else
    {
        if (!self->remoteChannelData->failedSIShutdown)
            self->remoteChannelData->failedSIShutdown = OALServer_Retain(OALSrv);
        OALRemoteChannel_Shutdown(self, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
    }
}

void OALRemoteChannel_ProcessRecieveSilence(OALChannel self)
{
    OALMessage msg = CMT_ProcessReceiveSilence_Create(self);
    if (msg)
    {
        OALRemoteChannelMgr_SendMessageToChannel(msg);
    }
    else
    {
        OALRemoteChannel_Shutdown(self, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
    }
}

void OALRemoteChannel_APPIDQueryTimeout(OALChannel self)
{
    OALMessage msg = CMT_APPIDQueryTimeout_Create((void *)self);

    if (msg)
    {
        OALRemoteChannelMgr_SendMessageToChannel( msg );
    }
    else
    {
        OALRemoteChannel_Shutdown(self, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
    }
}

static boolean OALChannel_SSIDTree_CompareGT_fn(OALAVLTreeListKey_t key, OALAVLTreeListItem_t item, void *context)
{
    OALChannel channel = (OALChannel)item;
    uint32 *SSIDptr = (uint32 *)key;

    return (*SSIDptr > channel->remoteChannelData->securityStateID);
}

static boolean OALConnection_tree_CompareEq_fn(OALAVLTreeListKey_t key, OALAVLTreeListItem_t item, void *context)
{
    OALChannel channel = (OALChannel)item;
    uint32 *SSIDptr = (uint32 *)key;

    return (*SSIDptr == channel->remoteChannelData->securityStateID);
}

static boolean OALRemoteChannel_FindSecureChannel(OALAVLTreeListItem_t item, void *arg)
{
    /* At this point the item is a channel whith the matching SSID so see if it is the secure channel in the context (arg). */
    return (item == arg);
}

void OALRemoteChannel_DeregisterSecureChannel_Immediate(OALChannel secureChannel)
{
    PCR_Assert(secureChannel);
    PCR_Assert(secureChannel->remoteChannelData);

    /* If the channel currently has a valid security state ID remove it from the registered tree. */
    if (secureChannel->remoteChannelData->baseChannel &&
        secureChannel->remoteChannelData->securityStateID != INVALID_SECURITYSTATEID)
    {
        OALHashKey_t        keys[1];
        OALHashKeySize_t    keySize = sizeof(secureChannel->remoteChannelData->securityStateID);

        PCR_Assert(secureChannel->remoteChannelData->baseChannel->remoteChannelData);

        if (secureChannel->remoteChannelData->flags.isSSIDAssignedFromAS)
        {
            OALAVLTreeList_FindAndRemove(secureChannel->remoteChannelData->baseChannel->remoteChannelData->registeredSSIDs_AS,
                                         &secureChannel->remoteChannelData->securityStateID,
                                         OALRemoteChannel_FindSecureChannel, secureChannel);
        }
        else
        {
            keys[0] = &secureChannel->remoteChannelData->securityStateID;
            OALHashtable_Remove(secureChannel->remoteChannelData->baseChannel->remoteChannelData->registeredSSIDs_NonAs,
                                secureChannel, keys, &keySize);
        }
        secureChannel->remoteChannelData->securityStateID = INVALID_SECURITYSTATEID;
        secureChannel->remoteChannelData->flags.isSSIDAssignedFromAS = FALSE;
    }
}

/* Note: this must be called from the message_enqueue worker thread.  The security state ID and the registered
   state ID tree are not mutex protected. */
static boolean OALRemoteChannel_RegisterSecureChannel_AS(OALChannel secureChannel, uint32 SSID)
{

    PCR_Assert(secureChannel);
    PCR_Assert(secureChannel->remoteChannelData);

    if (secureChannel->remoteChannelData->baseChannel)
    {
        OALChannel      baseChannel;

        baseChannel = secureChannel->remoteChannelData->baseChannel;

        if (!baseChannel->remoteChannelData->registeredSSIDs_AS)
        {
            baseChannel->remoteChannelData->registeredSSIDs_AS = OALAVLTreeList_Create(OALChannel_SSIDTree_CompareGT_fn,
                                                                                       OALConnection_tree_CompareEq_fn, NULL);

            /* If the registerd tree did not get created return FALSE */
            if (!baseChannel->remoteChannelData->registeredSSIDs_AS)
                return FALSE;
        }

        OALRemoteChannel_DeregisterSecureChannel_Immediate(secureChannel);

        /* Now that the tree exists, and any previous SSID for the channel is not registered, set the SSID for the channel
           and add the channel to the tree. */
        secureChannel->remoteChannelData->securityStateID = SSID;
        secureChannel->remoteChannelData->flags.isSSIDAssignedFromAS = TRUE;

        if (!OALAVLTreeList_Insert(baseChannel->remoteChannelData->registeredSSIDs_AS, &SSID, secureChannel))
        {
            secureChannel->remoteChannelData->securityStateID = INVALID_SECURITYSTATEID;
            secureChannel->remoteChannelData->flags.isSSIDAssignedFromAS = FALSE;
            return FALSE;
        }
    }
    else
    {
        secureChannel->remoteChannelData->securityStateID = SSID;
        secureChannel->remoteChannelData->flags.isSSIDAssignedFromAS = TRUE;
    }

    return TRUE;
}

static boolean OALRemoteChannel_RegisterSecureChannel_NonAS(OALChannel secureChannel, uint32 SSID)
{
    OALChannel          baseChannel;
    OALHashKey_t        keys[1];
    OALHashKeySize_t    keySize = sizeof(SSID);

    PCR_Assert(secureChannel);
    PCR_Assert(secureChannel->remoteChannelData);
    PCR_Assert(secureChannel->remoteChannelData->baseChannel);

    baseChannel = secureChannel->remoteChannelData->baseChannel;
    PCR_Assert(baseChannel->remoteChannelData);

    if (!baseChannel->remoteChannelData->registeredSSIDs_NonAs)
    {
        baseChannel->remoteChannelData->registeredSSIDs_NonAs = OALHashtable_Create(1);

        /* If the hashtable did not get created return FALSE */
        if (!baseChannel->remoteChannelData->registeredSSIDs_NonAs)
            return FALSE;
    }

    OALRemoteChannel_DeregisterSecureChannel_Immediate(secureChannel);

    /* Now that the hashtable exists and any previous SSID for the channel is not registered, set the SSID for the channel
       and add the channel to the hashtable. */

    keys[0] = &SSID;

    if (!OALHashtable_Retrieve(baseChannel->remoteChannelData->registeredSSIDs_NonAs, keys, &keySize))
    {
        secureChannel->remoteChannelData->securityStateID = SSID;
        secureChannel->remoteChannelData->flags.isSSIDAssignedFromAS = FALSE;



        if (OALHashtable_Insert(baseChannel->remoteChannelData->registeredSSIDs_NonAs, secureChannel,
                                keys, &keySize) != OALHashtable_InsertSuccess)
        {
            secureChannel->remoteChannelData->securityStateID = INVALID_SECURITYSTATEID;
            secureChannel->remoteChannelData->flags.isSSIDAssignedFromAS = FALSE;
            return FALSE;
        }
    }
    else
    {
        return FALSE;
    }

    return TRUE;
}

void OALRemoteChannel_HandleMessage_RegisterSecureChannel(OALChannel channel, OALMessage message)
{
    OALChannel  secureChannel;
    uint32      SSID;
    uint32      assignedFromAs;

    secureChannel = (OALChannel)message->recipient;
    SSID = (uint32)message->argv[0];
    assignedFromAs = (uint32)message->argv[1];


    if (SSID == INVALID_SECURITYSTATEID)
    {
        OALRemoteChannel_DeregisterSecureChannel_Immediate(secureChannel);
        secureChannel->remoteChannelData->securityStateID = INVALID_SECURITYSTATEID;
        secureChannel->remoteChannelData->flags.isSSIDAssignedFromAS = FALSE;
    }
    else
    {
        if (assignedFromAs)
            OALRemoteChannel_RegisterSecureChannel_AS(secureChannel, SSID);
        else
            OALRemoteChannel_RegisterSecureChannel_NonAS(secureChannel, SSID);
    }
}

void OALRemoteChannel_HandleMessage_ProcessReceiveSilence(OALChannel channel, OALMessage message)
{
    if (channel->remoteChannelData->waitingForData)
    {
        /* If we're waiting for data, then the ping was already sent, and no responses were received. */

        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "OALRemoteChannel: ProcessReceiveSilence, receive silence timer expired, shutting down.");
        OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_IO_FAILED);
    }
    else
    {
        DOFError error;

        channel->remoteChannelData->waitingForData = TRUE;
        EPP_NegOne_Ping( channel, NULL, EPP_NEGONE_NO_SECURE_MARKING_ID );

        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "OALRemoteChannel: ProcessReceiveSilence, receive silence timer expired, sending ping, new timeout is 10 seconds.");

        error = OALTimerItem_Insert( channel->remoteChannelData->maxReceiveSilenceTimerItem, EPP_NEGONE_MAX_PING_DELAY, FALSE, FALSE );

        if (error != DOFERROR_OK)
        {
            OALRemoteChannel_Shutdown(channel, DOFErrorException_GetFromError(error));
        }
    }
}

/* This function will send a message that will deregister and clear any previous security state ID for the channel,
   then set and register the security state ID provided. */
void OALRemoteChannel_RegisterSecureChannel(OALChannel self, uint32 SSID, boolean assignedFromAs)
{
    if (assignedFromAs)
    {
        OALMessage msg = CMT_RegisterSecureChannel_Create(self, SSID, assignedFromAs);

        if ( msg )
            OALRemoteChannelMgr_SendMessageToChannel( msg );
        else
            OALRemoteChannel_Shutdown(self, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
    }
    else
    {
        /* When this function is called for a SSID not assigned by an AS this function will already be running
           in the message_enqueue worker thread so make the call to register the SSID directly rather than
           sending a message. */
        OALRemoteChannel_RegisterSecureChannel_NonAS(self, SSID);
    }
}

/* This function will deregister and clear the security state ID stored in the channel. */
void OALRemoteChannel_DeregisterSecureChannel(OALChannel self)
{
    if (self && self->remoteChannelData && self->remoteChannelData->baseChannel)
    {
        OALMessage msg;

        /* Clear the SSID for the channel by setting it to INVALID_SECURITYSTATEID.  It will not matter if the previous
           SSID was assigned from the AS or not. */
        msg = CMT_RegisterSecureChannel_Create(self, INVALID_SECURITYSTATEID, FALSE);

        if ( msg )
            OALRemoteChannelMgr_SendMessageToChannel( msg );
        else
            OALRemoteChannel_Shutdown(self, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
    }
}

OALChannel OALRemoteChannel_GetSecureChannelByStateID_NonAs(OALChannel baseChannel, uint32 SSID)
{
    OALHashKey_t        keys[1];
    OALHashKeySize_t    keySize = sizeof(SSID);

    keys[0] = &SSID;
    return (OALChannel)OALHashtable_Retrieve(baseChannel->remoteChannelData->registeredSSIDs_NonAs, keys, &keySize);
}

static boolean OALRemoteChannel_TraverseSecureChannelsBySSID_As_fn(OALAVLTreeListItem_t item, void *arg)
{
    OALChannel secureChannel = (OALChannel)item;
    struct traverseBySSIDContext * ctx = (struct traverseBySSIDContext *)arg;

    return ctx->fn(secureChannel, ctx->context);
}

/* Note: this must only be called from the message enqueue worker thread. */
uint32 OALRemoteChannel_GetRegisteredChannelCountBySSID_As(OALChannel baseChannel, uint32 SSID)
{
    if (!baseChannel || !baseChannel->remoteChannelData || ! baseChannel->remoteChannelData->registeredSSIDs_AS)
        return 0;

    return OALAVLTreeList_GetNumItemsInList(baseChannel->remoteChannelData->registeredSSIDs_AS, &SSID);
}

void OALRemoteChannel_TraverseSecureChannelsBySSID_As(remotechannel_ssid_traverse_fn_t fn, OALChannel baseChannel,
                                                      uint32 SSID, void *context)
{
    struct traverseBySSIDContext ctx;

    ctx.fn = fn;
    ctx.context = context;

    if (baseChannel->remoteChannelData->registeredSSIDs_AS)
    {
        OALAVLTreeList_TraverseList(baseChannel->remoteChannelData->registeredSSIDs_AS, &SSID,
                                    OALRemoteChannel_TraverseSecureChannelsBySSID_As_fn, &ctx);
    }
}

void OALRemoteChannel_RouteToProviderLossy(OALChannel self, DOFOperation req_op, DOFOperation prov_op, DOFAddress addr)
{
    OALMessage msg = CMT_RouteToProvider_Create( self, req_op, prov_op, addr );

    /* For lossy protocols, if memory can not be allocated, just treat it as a dropped packet. */
    if (msg)
        OALRemoteChannelMgr_SendMessageToChannel( msg );
}

void OALRemoteChannel_RouteToProviderLossless(OALChannel self, DOFOperation req_op, DOFOperation prov_op, DOFAddress addr)
{
    OALMessage msg = CMT_RouteToProvider_Create( self, req_op, prov_op, addr );

    /* If a message can not be created for a lossless session close the session. */
    if (msg)
    {
        OALRemoteChannelMgr_SendMessageToChannel( msg );
    }
    else
    {
        OALRemoteChannel_Shutdown(self, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
    }

}

void OALRemoteChannel_RouteToRequestorLossy(OALChannel self, DOFOperation respOp, DOFResult result)
{
    OALMessage msg = CMT_RouteToRequestor_Create( self, respOp, result );

    if (msg)
        OALRemoteChannelMgr_SendMessageToChannel( msg );
}

void OALRemoteChannel_RouteToRequestorLossless(OALChannel self, DOFOperation respOp, DOFResult result)
{
    OALMessage msg = CMT_RouteToRequestor_Create( self, respOp, result );

    if (msg)
    {
        OALRemoteChannelMgr_SendMessageToChannel( msg );
    }
    else
    {
        OALRemoteChannel_Shutdown(self, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
    }
}

void OALRemoteChannel_ReceiveFromTransport( OALChannel self, OALStream data, DOFAddress rcvdFrom, boolean isMulticast, boolean isLossless )
{
    OALMessage msg = CMT_ReceiveFromTransport_Create( self, data, rcvdFrom, isMulticast );

    if ( msg )
    {
        OALRemoteChannelMgr_SendMessageToChannel( msg );
    }
    else
    {
        if (isLossless)
            OALRemoteChannel_Shutdown(self, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
    }
}

void OALRemoteChannel_NotifyTransportDown(OALChannel self, DTL transport, DOFException exception)
{
    OALMessage msg = CMT_TransportDown_Create( self, transport, exception );

    /*  The message will be placed in the message enqueue and the majority of the work will be done there. */
    if ( msg )
        OALRemoteChannelMgr_SendMessageToChannel( msg );
    else
        OALRemoteChannel_Shutdown(self, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
}

void OALRemoteChannel_DestroyBlockedMessageQueue(OALChannel channel)
{
    if (channel->remoteChannelData->blockedMessageQueue)
    {
        while (OALQueue_GetSize(channel->remoteChannelData->blockedMessageQueue))
        {
            OALMessage msg = OALQueue_Dequeue(channel->remoteChannelData->blockedMessageQueue);
            OALMessage_Destroy(msg);
        }
        OALQueue_Destroy(channel->remoteChannelData->blockedMessageQueue);
        channel->remoteChannelData->blockedMessageQueue = NULL;
    }
}

void OALRemoteChannel_ProcessBlockedMessages(OALChannel channel, boolean clearSendBlock)
{
    OALMessage msg = CMT_ProcessBlockedMessage(channel, clearSendBlock);

    if (msg)
        OALRemoteChannelMgr_SendMessageToChannel(msg);
    else
        OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
}

typedef struct CreateBridgeContext_t
{
    OALThreadPoolNode_t super;

    DOF dof;
    DOFOperationBridgeConfig bridgeConfig;
    OALChannel channel;
    DOFAddress address;
} CreateBridgeContext;

static void CreateBridge_Real(OALThreadPoolItem item,
                              boolean           shutdown)
{
    CreateBridgeContext * node = (CreateBridgeContext*) item;

    DOFException exception;
    /* DOF_CreateBridge will block, so that is why this is called using the threadpool. */
    OALOperationBridge oalBridge = DOF_CreateBridge(node->dof, node->bridgeConfig, node->channel, node->address, 30000, &exception);
    OALMessage newMessage = NULL;

    if(oalBridge != NULL)
    {
        newMessage = CMT_CreateBridgeComplete_Create(node->channel, oalBridge);
        OALOperationBridge_Destroy(oalBridge);
        if(newMessage == NULL)
        {
            /* The copy is used here only for convience so that the exception can always be destroyed at the end of this function. */
            exception = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        }
    }
    if(newMessage != NULL)
    {
        OALRemoteChannelMgr_SendMessageToChannel( newMessage );
    }
    else
    {
        OALRemoteChannel_Shutdown(node->channel, exception);
    }
    DOFException_Destroy(exception);
    DOFOperationBridgeConfig_Destroy(node->bridgeConfig);
    DOFAddress_Destroy(node->address);
    OALChannel_Release(node->channel);
    DOFMem_Free(node);
}

/* Note: address may be NULL, it is used to route TRP if no other AS route is already established. */
void OALRemoteChannel_CreateBridge(OALChannel self, DOFAddress address, DOF dof, DOFOperationBridgeConfig bridgeConfig)
{
    CreateBridgeContext * context = (CreateBridgeContext*) DOFMem_Alloc(sizeof(CreateBridgeContext));

    if (context != NULL)
    {
        OALThreadPoolItem_Init(&context->super, CreateBridge_Real);
        context->dof = dof;
        context->bridgeConfig = DOFOperationBridgeConfig_Copy(bridgeConfig);
        if (context->bridgeConfig || !bridgeConfig)
        {
            context->channel = OALChannel_Retain(self);
            if (context->channel || !self)
            {
                context->address = DOFAddress_Copy(address);
                if (context->address || !address)
                {
                    if(OALThreadPool_Run(context->dof->threadPool, &context->super))
                    {
                        /* CreateBridge_Real will run and destroy context. */
                        return;
                    }
                    DOFAddress_Destroy(context->address);
                }
                OALChannel_Release(context->channel);
            }
            DOFOperationBridgeConfig_Destroy(context->bridgeConfig);
        }
        DOFMem_Free(context);
    }
    OALRemoteChannel_Shutdown(self, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
}

static boolean channelENPPort_gt_fn(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    OALChannel channel = (OALChannel)item;
    uint32 *portPtr = (uint32 *)key;

    return *portPtr > channel->remoteChannelData->ENP_Port;
}

static boolean channelENPPort_eq_fn(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    OALChannel channel = (OALChannel)item;
    uint32 *portPtr = (uint32 *)key;

    return *portPtr == channel->remoteChannelData->ENP_Port;
}

/* Note: this function must only be called from within a channel message_enqueue worker thread. */
OALChannel OALRemoteChannel_GetRegisteredChannel_ENPPort(OALChannel baseChannel, uint32 port)
{
    if (baseChannel->remoteChannelData->registeredENPPorts)
    {
        return (OALChannel)OALAVLTree_Retrieve(baseChannel->remoteChannelData->registeredENPPorts, &port);
    }
    return NULL;
}

boolean OALRemoteChannel_RegisterChannel_ENPPort_Immediate(OALChannel channel)
{
    OALChannel          baseChannel;

    PCR_Assert(channel);
    PCR_Assert(channel->remoteChannelData);
    PCR_Assert(channel->remoteChannelData->baseChannel);
    PCR_Assert(channel->remoteChannelData->ENP_Port == ENP_INVALID_PORT);

    baseChannel = channel->remoteChannelData->baseChannel;
    PCR_Assert(baseChannel->remoteChannelData);

    if (!baseChannel->remoteChannelData->registeredENPPorts)
    {
        baseChannel->remoteChannelData->registeredENPPorts = OALAVLTree_Create(channelENPPort_gt_fn, channelENPPort_eq_fn, NULL);

        /* If the hashtable did not get created return FALSE */
        if (!baseChannel->remoteChannelData->registeredENPPorts)
            return FALSE;

        /* Since the table is created here just assign 1 to be the port number*/
        channel->remoteChannelData->ENP_Port = 1;
    }
    else
    {
        uint32 i;

        /* If the port table exists generate a port number and check to see if is already registered in the table. */
        for (i=0; i<10; i++)
        {
            uint32 rand;
            uint32 port;

            rand = PCRRandom_Get();
            port = rand & 0x7f;
            if (!OALRemoteChannel_GetRegisteredChannel_ENPPort(baseChannel, port))
            {
                channel->remoteChannelData->ENP_Port = port;
                break;
            }
            port = rand & 0x3fff;
            if (!OALRemoteChannel_GetRegisteredChannel_ENPPort(baseChannel, port))
            {
                channel->remoteChannelData->ENP_Port = port;
                break;
            }

            port = rand & 0x3fffff;
            if (!OALRemoteChannel_GetRegisteredChannel_ENPPort(baseChannel, port))
            {
                channel->remoteChannelData->ENP_Port = port;
                break;
            }
        }
    }

    /*  Add the channel to the tree.  The channel is not retained when added to this list since the channel
        already retains the base channel. */
    if ((channel->remoteChannelData->ENP_Port == ENP_INVALID_PORT) ||
        (OALAVLTree_Insert(baseChannel->remoteChannelData->registeredENPPorts,
                           &channel->remoteChannelData->ENP_Port, channel, NULL)) != OALAVLTree_InsertSuccess)
    {
        return FALSE;
    }

    return TRUE;
}

/* Note: this function must only be called from within a channel message_enqueue worker thread. */
void OALRemoteChannel_DeregisterChannel_ENPPort_Immediate(OALChannel channel)
{
    PCR_Assert(channel);
    PCR_Assert(channel->remoteChannelData);

    if (channel->remoteChannelData->ENP_Port != ENP_INVALID_PORT)
    {
        PCR_Assert(channel->remoteChannelData->baseChannel);
        PCR_Assert(channel->remoteChannelData->baseChannel->remoteChannelData);
        PCR_Assert(channel->remoteChannelData->baseChannel->remoteChannelData->registeredENPPorts);

        OALAVLTree_Delete(channel->remoteChannelData->baseChannel->remoteChannelData->registeredENPPorts,
                          &channel->remoteChannelData->ENP_Port);
    }
}

boolean OALRemoteChannel_RegisterChannel_ENPPort(OALChannel channel)
{
    OALMessage msg = CMT_RegisterChannel_ENPPort_Create( channel );

    if ( msg )
    {
        OALRemoteChannelMgr_SendMessageToChannel( msg );
        return TRUE;
    }
    else
    {
        OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        return FALSE;
    }
}

void OALRemoteChannel_HandleMessage_RegisterChannel_ENPPort(OALChannel channel, OALMessage message)
{
    if (!OALRemoteChannel_RegisterChannel_ENPPort_Immediate(channel))
        OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_FAILED);
}

typedef struct DummyCallback_t
{
    DOFObjectIDDomain domainID;
    OALChannel channel;
    OALRemoteChannel_SetCredentials_Callback fn;
    void * context;
} DummyCallback_t;

void OALRemoteChannel_SetCredentialsCallback_ContextDtor(void *context)
{
    DummyCallback_t *fullContext = (DummyCallback_t *)context;

    if (fullContext)
    {
        if (fullContext->fn)
        {
            fullContext->fn(DOFERROR_FAILED, fullContext->context);
        }
        DOFMem_Free(fullContext);
    }
}

static void OALRemoteChannel_SetDomainCredentialsCallback( DOFCredentials creds, void *context )
{
    DummyCallback_t *fullContext;
    OALMessage msg;

    fullContext = (DummyCallback_t *)context;

    msg = CMT_SetDomainCredentials_Create(fullContext->channel, creds, context);

    if (msg)
    {
        OALRemoteChannelMgr_SendMessageToChannel(msg);
    }
    else
    {
        OALRemoteChannel_Shutdown(fullContext->channel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        DOFMem_Free(context);
    }
}

static void dummyCallback(DOFCredentials   creds,
                          void           * context)
{
    DummyCallback_t * fullContext = (DummyCallback_t *) context;

    if(creds != NULL)
    {
        creds = DOFCredentials_Copy( creds );
        if ( creds )
        {
            DOFCredentials_Destroy( fullContext->channel->remoteChannelData->secureCredentials );
            fullContext->channel->remoteChannelData->secureCredentials = creds;
            fullContext->fn( DOFERROR_OK, fullContext->context );
        }
        else
            fullContext->fn( DOFERROR_INSUFFICIENT_RESOURCES, fullContext->context );
    }
    else
        fullContext->fn( DOFERROR_FAILED, fullContext->context );
}

void OALRemoteChannel_HandleMessage_SetDomainCredentialsCallback(OALChannel channel, OALMessage message)
{
    DummyCallback_t * fullContext;

    PCR_Assert(message);

    fullContext = (DummyCallback_t *)message->argv[0];

    dummyCallback((DOFCredentials)message->argv[1], fullContext);
    /* Set the callback function to NULL so that it will not get called again when the message is destroyed. */
    fullContext->fn = NULL;
}

void OALRemoteChannel_HandleMessage_PeerListenerEvent(OALChannel channel, OALMessage message)
{
    OLAMessage_PeerEventMessage peerMsg = (OLAMessage_PeerEventMessage)message;

    if (channel->remoteChannelData->OALConn)
    {
        OALConnection_PeerListenerEvent(channel->remoteChannelData->OALConn, peerMsg->address, peerMsg->scope, peerMsg->added);
    }
}

void OALRemoteChannel_HandleMessage_ConnectionStreamListener(OALChannel channel, OALMessage message)
{
    OALMessage_StreamRequestMessage reqMsg = (OALMessage_StreamRequestMessage)message;

    if (channel->remoteChannelData->OALConn && channel->remoteChannelData->flags.isOutbound)
    {
        OALConnection_ConnectionStreamRequestEvent(channel->remoteChannelData->OALConn, reqMsg->addr, reqMsg->complete);
    }
}

void OALRemoteChannel_SetCredentials(OALChannel channel, DOFObjectIDDomain domainID, uint32 timeout, OALRemoteChannel_SetCredentials_Callback fn, void* context)
{
    if (channel && channel->remoteChannelData && channel->remoteChannelData->flags.isConnection)
    {
        DOFConnection conn;

        conn = OALConnection_GetDOFConnection(channel->remoteChannelData->OALConn);

        if (conn && conn->relatedServer)
        {
            DummyCallback_t *fullContext = (DummyCallback_t *)DOFMem_Alloc(sizeof(DummyCallback_t));

            if (fullContext)
            {
                fullContext->domainID = domainID;
                fullContext->channel = channel;
                fullContext->fn = fn;
                fullContext->context = context;
                DOFServerConfig_GetDomainCredential(DOFServer_GetConfig(conn->relatedServer), domainID, timeout, OALRemoteChannel_SetDomainCredentialsCallback, fullContext);
            }
            else
            {
                fn( DOFERROR_INSUFFICIENT_RESOURCES, context );
            }
            return;
        }
    }
    fn( DOFERROR_FAILED, context );
}

DOFCredentials OALRemoteChannel_GetCredentials(RemoteChannelData self)
{
    PCR_Assert(self);
    PCR_Assert(self->flags.isConnection);
    return DOFConnectionConfig_GetCredentials( self->config.connection );
}

void OALRemoteChannel_PeerListenerEvent(OALChannel channel, DOFAddress address, DOFSecurityScope scope, boolean added)
{
    OALMessage message;

    PCR_Assert(channel);
    PCR_Assert(address);

    if (added)
    {
        PCR_Assert(scope);
    }

    message = CMT_PeerListenerEvent_Create(channel, address, scope, added);

    if (message)
    {
        OALRemoteChannelMgr_SendMessageToChannel(message);
    }
    else
    {
        OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
    }
}

void OALRemoteChannel_ConnectionStreamRequest(OALChannel channel, DOFAddress address, boolean complete)
{
    OALMessage message;

    PCR_Assert(channel);

    message = CMT_ConnectionStreamRequest_Create(channel, address, complete);

    if (message)
    {
        OALRemoteChannelMgr_SendMessageToChannel(message);
    }
    else
    {
        OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
    }
}

KeyChangeInfo_t* OALRemoteChannel_CreateKeyChangeInfo( OALChannel channel, OALCipher cipher, uint32 stateId, DOFPermissionSet localPerms, DOFPermissionSet remotePerms, DOFSecurityMode securityMode, DOFSecurityScope securityScope, boolean disableOld, boolean promoteNew )
{
    KeyChangeInfo_t* pKeyChangeInfo = (KeyChangeInfo_t*)DOFMem_Alloc( sizeof(KeyChangeInfo_t));

    if (pKeyChangeInfo)
    {
        pKeyChangeInfo->localPerms = DOFPermissionSet_Copy( localPerms );
        if ( !localPerms || pKeyChangeInfo->localPerms )
        {
            pKeyChangeInfo->remotePerms = DOFPermissionSet_Copy( remotePerms );
            if ( !remotePerms || pKeyChangeInfo->remotePerms )
            {
                pKeyChangeInfo->securityMode = DOFSecurityMode_Copy( securityMode );
                if ( !securityMode || pKeyChangeInfo->securityMode )
                {
                    pKeyChangeInfo->securityScope = DOFSecurityScope_Copy( securityScope );
                    if ( !securityScope || pKeyChangeInfo->securityScope )
                    {
                        pKeyChangeInfo->channel = channel;
                        pKeyChangeInfo->cipher = cipher;
                        pKeyChangeInfo->replaceCipher = TRUE;
                        pKeyChangeInfo->stateId = stateId;
                        pKeyChangeInfo->disableOld = disableOld;
                        pKeyChangeInfo->promoteNew = promoteNew;
                        return pKeyChangeInfo;
                    }
                }
                DOFPermissionSet_Destroy( pKeyChangeInfo->localPerms );
            }
            DOFPermissionSet_Destroy( pKeyChangeInfo->localPerms );
        }
        DOFMem_Free( pKeyChangeInfo );
    }
    return NULL;
}

boolean OALRemoteChannel_KeyChangeCallback( void* ctx )
{
    KeyChangeInfo_t* pKeyChangeInfo = (KeyChangeInfo_t*)ctx;
    OALCipherKeyState keyState = NULL;
    boolean rval = TRUE;
    /* First promote the key. */
    if(pKeyChangeInfo->promoteNew)
    {
        OALCipher_PromoteFutureKey(pKeyChangeInfo->channel->remoteChannelData->cipher, TEP_V1_TIMEOUT);
    }
    /* Remove old key(s). */
    if(pKeyChangeInfo->disableOld)
    {
        OALCipher_DisableOldKeys(pKeyChangeInfo->channel->remoteChannelData->cipher);
    }
    /* Update the rest of the key state. */
    OALChannel_Lock(pKeyChangeInfo->channel);
    if ( pKeyChangeInfo->localPerms || pKeyChangeInfo->remotePerms )
    {
        /* Always use the now-current key state (after promotion, above). */
        keyState = OALCipher_GetCurrentKeyState( pKeyChangeInfo->cipher );
    }
    pKeyChangeInfo->channel->remoteChannelData->securityStateID = pKeyChangeInfo->stateId;
    if ( pKeyChangeInfo->localPerms )
    {
        if ( !keyState || !OALCipher_SetLocalPerms( pKeyChangeInfo->cipher, keyState, pKeyChangeInfo->localPerms ) )
            rval = FALSE;
        DOFPermissionSet_Destroy( pKeyChangeInfo->localPerms );
    }
    if ( pKeyChangeInfo->remotePerms )
    {
        if ( !keyState || !OALCipher_SetRemotePerms( pKeyChangeInfo->cipher, keyState, pKeyChangeInfo->remotePerms ) )
            rval = FALSE;
        DOFPermissionSet_Destroy( pKeyChangeInfo->remotePerms );
    }
    if ( pKeyChangeInfo->securityMode )
    {
        /* Take ownership. */
        DOFSecurityMode_Destroy(pKeyChangeInfo->channel->remoteChannelData->mode);
        pKeyChangeInfo->channel->remoteChannelData->mode = pKeyChangeInfo->securityMode;
    }
    if ( pKeyChangeInfo->securityScope )
    {
        /* Take ownership. */
        DOFSecurityScope_Destroy(pKeyChangeInfo->channel->securityScope);
        pKeyChangeInfo->channel->securityScope = pKeyChangeInfo->securityScope;
    }
    OALChannel_Unlock(pKeyChangeInfo->channel);
    DOFMem_Free(pKeyChangeInfo);
    return rval;
}

/* Note: this function must only be called from within a channel message_enqueue worker thread. */
void OALRemoteChannel_RemoveOALConnection( OALChannel channel, DOFException exception )
{
    /* Notify the OALConnection that the channel is going down and clear the reference to the OALConnection so
       that the OALConnection will only be notified once.
       Note: This can be called by shutdown */

    if (channel && channel->remoteChannelData && channel->remoteChannelData->OALConn)
    {
        boolean inbound = FALSE;
        if (!channel->remoteChannelData->flags.isOutbound)
            inbound = TRUE;

        OALConnection_ConnectionEvent(channel->remoteChannelData->OALConn, channel, FALSE, exception);
        /* For inbound connections The channel actually retains the OALConnection. */
        if (inbound)
            OALConnection_Release(channel->remoteChannelData->OALConn);
        PCRMutex_Lock(connection_server_mutex);
        channel->remoteChannelData->OALConn = NULL;
        PCRMutex_Unlock(connection_server_mutex);
    }
}

/* Note: this function must only be called from within a channel message_enqueue worker thread. */
void OALRemoteChannel_RemoveOALServer( OALChannel channel, DOFException exception )
{
    if( channel && channel->remoteChannelData && channel->remoteChannelData->server )
    {
        OALServer_NotifyServerTransportDown( channel->remoteChannelData->server, exception );
        channel->remoteChannelData->server = NULL;
    }
}

DOFAddress OALRemoteChannelConfig_GetAddress(OALChannel self)
{
    if (self->remoteChannelData->flags.isConnection)
        return DOFConnectionConfig_GetAddress(self->remoteChannelData->config.connection);
    return DOFServerConfig_GetAddress(self->remoteChannelData->config.server);
}

DOFOperationBridgeConfig OALRemoteChannelConfig_GetBridgeConfig(OALChannel self)
{
    if (self->remoteChannelData->flags.isConnection)
        return DOFConnectionConfig_GetBridge(self->remoteChannelData->config.connection);
    return DOFServerConfig_GetBridge(self->remoteChannelData->config.server);
}

DOFCredentials OALRemoteChannelConfig_GetCredentials(OALChannel self)
{
    if (self->remoteChannelData->flags.isConnection)
        return DOFConnectionConfig_GetCredentials(self->remoteChannelData->config.connection);

    return NULL;
}

DOFSecurityMode OALRemoteChannelConfig_GetMode(OALChannel self)
{
    if(self->remoteChannelData && self->remoteChannelData->mode)
        return self->remoteChannelData->mode;
    return NULL;
}

DOFProtocolNegotiator OALRemoteChannelConfig_GetNegotiator(OALChannel self)
{
    if (self->remoteChannelData->flags.isConnection)
        return DOFConnectionConfig_GetProtocolNegotiator(self->remoteChannelData->config.connection);
    return DOFServerConfig_GetProtocolNegotiator(self->remoteChannelData->config.server);
}

DOFSecurityDesire OALRemoteChannelConfig_GetSecurityDesire(OALChannel self)
{
    if (self->remoteChannelData->flags.isConnection)
        return DOFConnectionConfig_GetSecurityDesire(self->remoteChannelData->config.connection);
    return DOFServerConfig_GetSecurityDesire(self->remoteChannelData->config.server);
}

DOFPermissionSet OALRemoteChannelConfig_GetPermissions(OALChannel self)
{
    if (self->remoteChannelData->flags.isConnection)
        return DOFConnectionConfig_GetPermissions(self->remoteChannelData->config.connection);

    return NULL;
}

uint32 OALRemoteChannelConfig_GetMaxReceiveSilence(OALChannel self)
{
    if (self->remoteChannelData->flags.isConnection)
        return DOFConnectionConfig_GetMaxReceiveSilence(self->remoteChannelData->config.connection);
    return DOFServerConfig_GetMaxSendSilence(self->remoteChannelData->config.server);
}

uint32 OALRemoteChannelConfig_GetMaxSendSilence(OALChannel self)
{
    if (self->remoteChannelData->flags.isConnection)
        return DOFConnectionConfig_GetMaxSendSilence(self->remoteChannelData->config.connection);
    return DOFServerConfig_GetMaxSendSilence(self->remoteChannelData->config.server);
}

DOFConnectionType OALRemoteChannelConfig_GetConnectionType(OALChannel self)
{
    PCR_Assert(self->remoteChannelData->flags.isConnection);
    return DOFConnectionConfig_GetConnectionType(self->remoteChannelData->config.connection);
}

DTLConnectionConfig OALRemoteChannelConfig_GetDTLConnectionConfig(OALChannel self)
{
    PCR_Assert(self->remoteChannelData->flags.isConnection);
    return DOFConnectionConfig_GetTransportConfig(self->remoteChannelData->config.connection);
}

DOFServerType OALRemoteChannelConfig_GetServerType(OALChannel self)
{
    PCR_Assert(!self->remoteChannelData->flags.isConnection);
    return DOFServerConfig_GetServerType(self->remoteChannelData->config.server);
}

DTLServerConfig OALRemoteChannelConfig_GetDTLServerConfig(OALChannel self)
{
    PCR_Assert(!self->remoteChannelData->flags.isConnection);
    return DOFServerConfig_GetTransportConfig(self->remoteChannelData->config.server);
}

typedef struct createLibPermissionsNode
{
    createLibPermsCallback   callback;
    void                   * context;
} createLibPermissionsNode;

static void OALRemoteChannel_CreateLibPermsCallback(TRP_V1_Context   trp_context,
                                    void           * context);

void OALRemoteChannel_CreateLibPermissions(createLibPermsCallback   callback,
                                           void                   * context,
                                           DOFCredentials           credentials,
                                           OALChannel               channel,
                                           DOFAddress               address)
{
    createLibPermissionsNode * realContext;

    PCR_Assert(callback != NULL);
    realContext = (createLibPermissionsNode*) DOFMem_Alloc(sizeof(createLibPermissionsNode));
    if(realContext != NULL)
    {
        realContext->callback = callback;
        realContext->context = context;
        TRP_V1_RequestValidateCredentials(credentials, OALRemoteChannel_CreateLibPermsCallback,
                                          (void *) realContext, channel, address, 10000);
    }
    else
    {
        (*callback)(DOFERROR_INSUFFICIENT_RESOURCES, NULL, NULL, context);
    }
}

static void OALRemoteChannel_CreateLibPermsCallback(TRP_V1_Context   trp_context,
                                                    void           * context)
{
    createLibPermissionsNode * realContext = (createLibPermissionsNode *) context;

    PCR_Assert(context != NULL);
    if(trp_context != NULL && trp_context->ValidateReturn != NULL && trp_context->ValidateReturn->sid != NULL)
    {
        if(trp_context->ValidateReturn->status == DOFERROR_OK)
        {
            if(trp_context->ValidateReturn->errorCode == TRP_V1_SUCCESS)
            {
                DOFPermissionSetBuilder builder;

                builder = DOFPermissionSetBuilder_Create();
                if(builder != NULL)
                {
                    sid_t * secureSid = lookup_explicit_sid(trp_context->ValidateReturn->sid, DOFCredentials_GetDomainID(trp_context->ValidateReturn->creds));

                    if ( !secureSid )
                        secureSid = create_sid_from_oid(trp_context->ValidateReturn->sid, DOFCredentials_GetDomainID(trp_context->ValidateReturn->creds));

                    if(secureSid != NULL)
                    {
                        if(DOFPermissionSetBuilder_AddPermission(builder, secureSid->iamPermission))
                        {
                            boolean keepGoing = TRUE;

                            if( DOFConfig_IsRouter( DOF_GetConfig( DOF_GetDof() ) ) )
                            {
                                DOFPermission actAsAny = DOFPermissionActAsAny_Create();

                                if(actAsAny != NULL)
                                {
                                    keepGoing = DOFPermissionSetBuilder_AddPermission(builder, actAsAny);
                                    DOFPermission_Destroy(actAsAny);
                                }
                                else
                                {
                                    keepGoing = FALSE;
                                }
                            }

                            if(keepGoing)
                            {
                                DOFPermissionSet tempSet = DOFPermissionSetBuilder_BuildAndDestroy(builder);

                                if(tempSet != NULL)
                                {
                                    (*realContext->callback)(DOFERROR_OK, secureSid, tempSet, realContext->context);
                                    DOFPermissionSet_Destroy(tempSet);
                                    source_release(secureSid);
                                    DOFMem_Free(context);
                                    return;
                                }
                            }
                        }
                        source_release(secureSid);
                    }
                    DOFPermissionSetBuilder_Destroy(builder);
                }
                (*realContext->callback)(DOFERROR_INSUFFICIENT_RESOURCES, NULL, NULL, realContext->context);
            }
            else
            {
                (*realContext->callback)(TRP_V1_GetDOFErrorFromErrorCode(trp_context->ValidateReturn->errorCode), NULL, NULL, realContext->context);
            }
        }
        else
        {
            (*realContext->callback)(trp_context->ValidateReturn->status, NULL, NULL, realContext->context);
        }
    }
    else
    {
        if (trp_context && trp_context->ValidateReturn && (trp_context->ValidateReturn->status == DOFERROR_OK))
        {
            (*realContext->callback)(TRP_V1_GetDOFErrorFromErrorCode(trp_context->ValidateReturn->errorCode), NULL, NULL, realContext->context);
        }
        else
        {
            (*realContext->callback)(DOFERROR_FAILED, NULL, NULL, realContext->context);
        }
    }
    DOFMem_Free(context);
}

void OALRemoteChannel_IncByteCount( OALChannel self, uint32 byteCount )
{
    if( self->remoteChannelData->flags.isConnection )
    {
        /* connection */
        OALConnection_UpdateInputByteCount( self->remoteChannelData->OALConn, byteCount );
    }
    else
    {
        /* server */
        OALServer_UpdateInputByteCount( self->remoteChannelData->server, byteCount );
    }
}
