/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/connection.h>
#include <dof/exception.h>
#include <dof/server.h>

#include <internal/channel.h>
#include <internal/dofinternal.h>
#include <internal/exception.h>
#include <internal/remotechannel.h>
#include <internal/remotechannelmanager.h>
#include <internal/avl_tree.h>
#include <internal/enqueue.h>
#include <internal/operationinternal.h>
#include <internal/hashtable.h>
#include <internal/channel_stack_messages.h>
#include <internal/epp_negative_one.h>
#include <internal/securityscope.h>

#include "channel_local.h"

#define MIN_PING_DELAY 1000

extern PCRMutex connection_server_mutex;

typedef enum RCM_MessageType
{
    RCM_ADD_CHANNEL,
    RCM_REMOVE_CHANNEL,
    RCM_ADD_PROTOCOLSTACK,
    RCM_REMOVE_PROTOCOLSTACK

} RCM_MessageType;

OALConnectionlessProtocolStack gDefaultConnectionlessProtocolStack = NULL;

static OALEnqueue  message_enqueue = NULL;
/* Note: channel_tree must only be used inside of the message_enqueue worker thread. */
static OALAVLTree  channel_tree = NULL;
static OALAVLTree  protocolstack_tree = NULL;

static OALTimer     remoteChannelTimer;

static PCRMutex protocolStackTreeMutex = NULL;
static PCRMutex messageEnqueueMutex = NULL;

boolean OALRemoteChannelMgr_Init(void)
{
    protocolStackTreeMutex = PCRMutex_Create();
    if (!protocolStackTreeMutex)
        goto EXIT_FAILURE_0;

    message_enqueue = OALEnqueue_Create(DOF_GetDof()->threadPool);
    if (!message_enqueue)
        goto EXIT_FAILURE_1;

    channel_tree = OALAVLTree_Create( NULL, NULL, NULL );
    if (!channel_tree)
        goto EXIT_FAILURE_2;

    protocolstack_tree = OALAVLTree_Create( NULL, NULL, NULL );
    if (!protocolstack_tree)
        goto EXIT_FAILURE_3;

    remoteChannelTimer = OALTimer_Create(DOF_GetDof()->threadPool, TRUE);
    if (!remoteChannelTimer)
        goto EXIT_FAILURE_6;

    gDefaultConnectionlessProtocolStack = OALConnectionlessProtocolStack_Create();
    if (!gDefaultConnectionlessProtocolStack)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALRemoteChannelMgr: create: Failed to create connectionless protocol stack." );
        goto EXIT_FAILURE_7;
    }
    if (!OALRemoteChannelMgr_AddProtocolStack( (OALProtocolStack)gDefaultConnectionlessProtocolStack ))
        goto EXIT_FAILURE_8;

    return TRUE;

    EXIT_FAILURE_8:
    OALProtocolStack_Release((OALProtocolStack)gDefaultConnectionlessProtocolStack);
    gDefaultConnectionlessProtocolStack = NULL;
    EXIT_FAILURE_7:
    OALTimer_Destroy( remoteChannelTimer );
    EXIT_FAILURE_6:
    OALAVLTree_Destroy( protocolstack_tree, NULL, NULL );
    protocolstack_tree = NULL;
    EXIT_FAILURE_3:
    OALAVLTree_Destroy( channel_tree, NULL, NULL );
    channel_tree = NULL;
    EXIT_FAILURE_2:
    OALEnqueue_Destroy( message_enqueue );
    message_enqueue = NULL;
    EXIT_FAILURE_1:
    PCRMutex_Destroy( protocolStackTreeMutex );
    protocolStackTreeMutex = NULL;
    EXIT_FAILURE_0:
    return FALSE;
}

boolean OALRemoteChannelMgr_MutexInit(void)
{
    messageEnqueueMutex = PCRMutex_Create();
    if (messageEnqueueMutex)
    {
        if (OALRemoteChannel_MutexInit())
        {
            return TRUE;
        }
        PCRMutex_Destroy(messageEnqueueMutex);
    }

    return FALSE;
}

/* This function is used to send a shutdown message to each channel, without enqueuing a message for each one. */
static void OALRemoteChannelMgr_TraverseShutdownAll_fn(OALAVLTreeItem_t item, void *notUsed)
{
    OALChannel      channel = (OALChannel)item;
    OALMessage      message;
    DOFException    exception;

    /* Create a shutdown message and call the message handler function. */
    exception = DOFErrorException_Create(DOFERROR_TERMINATED, "Shutdown by DOF_Destroy");
    message = CMT_Shutdown_Create(channel, exception);

    /* If we get to this point and there is still a server or a connection do not worry about notifying the server
       or connection since the server manager or connection manager will destroy them. */
    PCRMutex_Lock(connection_server_mutex);
    channel->remoteChannelData->server = NULL;
    channel->remoteChannelData->OALConn = NULL;
    PCRMutex_Unlock(connection_server_mutex);

    if (message)
        channel->msg_handler_func(channel, message);

    DOFException_Destroy(exception);
}

static void OALRemoteChannelMgr_ShutdownAll_fn( OALThreadPoolItem item, boolean shutdown )
{
    if (channel_tree)
    {
        OALAVLTree tempChannelTree = channel_tree;

        channel_tree = NULL;
        OALAVLTree_Destroy(tempChannelTree, OALRemoteChannelMgr_TraverseShutdownAll_fn, NULL);
    }
    if (item)
        DOFMem_Free(item);
}

static void OALRemoteChannelMgr_ShutdownAll(void)
{
    /*  If this succeeds channel_tree will be destroyed and set to NULL.  If this fails then OALRemoteChannelMgr_ShutdownAll_fn should be
        called directly after message_enqueue is destroyed. */
    OALEnqueueItem node = (OALEnqueueItem) DOFMem_Alloc(sizeof(OALEnqueueItem_t));

    if(node != NULL)
    {
        OALEnqueueItem_Init(node, OALRemoteChannelMgr_ShutdownAll_fn);
        PCRMutex_Lock(messageEnqueueMutex);
        if(message_enqueue)
        {
            OALEnqueue_Enqueue(message_enqueue, node);
        }
        else
        {
            DOFMem_Free(node);
        }

        PCRMutex_Unlock(messageEnqueueMutex);
    }
}

static void OALRemoteChannelMgr_ReleaseStacks_fn(OALAVLTreeItem_t item, void *arg)
{
    OALProtocolStack_Release( (OALProtocolStack)item );
}

void OALRemoteChannelMgr_Shutdown(void)
{
    OALEnqueue temp;
    /* The protocol stack tree will still have a reference this that will get destroyed with the tree. */
    OALProtocolStack_Release((OALProtocolStack)gDefaultConnectionlessProtocolStack);

    OALRemoteChannelMgr_ShutdownAll();
    PCRMutex_Lock(messageEnqueueMutex);
    temp = message_enqueue;
    message_enqueue = NULL;
    PCRMutex_Unlock(messageEnqueueMutex);
    /*  Call shutdown all to destroy the channel tree and send a shutdown message to each channel.  This will
        also set channel_tree to NULL. */
    OALEnqueue_Destroy( temp );

    /*  The channel_tree should have been destroyed by OALRemoteChannelMgr_ShutdownAll_fn, if not it was probably lack of resources that
        prevented it from being called. Call OALRemoteChannelMgr_ShutdownAll_fn now directly, this is safe since message_enqueue
        has been destroyed. */
    if (channel_tree)
    {
        OALRemoteChannelMgr_ShutdownAll_fn(NULL, TRUE);
    }

    OALAVLTree_Destroy( protocolstack_tree, OALRemoteChannelMgr_ReleaseStacks_fn, NULL );
    OALTimer_Destroy( remoteChannelTimer );

    PCRMutex_Destroy( protocolStackTreeMutex );
    protocolStackTreeMutex = NULL;
    protocolstack_tree = NULL;
}

void OALRemoteChannelMgr_MutexShutdown(void)
{
    PCRMutex_Destroy(messageEnqueueMutex);
    messageEnqueueMutex = NULL;
    OALRemoteChannel_MutexShutdown();
}

typedef struct ChannelNode_t
{
    OALEnqueueItem_t super;

    OALChannel       channel;
} ChannelNode_t;

/* ********************************************************************************/
static void OALRemoteChannelMgr_AddChannel_fn( OALThreadPoolItem dnode, boolean shutdown )
{
    ChannelNode_t * node;

    OALAVLTree_InsertionStatus status;

    PCR_Assert( dnode != NULL );
    node = (ChannelNode_t *) dnode;
    if ( shutdown || !channel_tree)
    {
        OALChannel_Release( node->channel );
        DOFMem_Free(node);
        return;
    }

    /* make sure that the channel is not already in the tree */
    if (channel_tree)
    {
        status = OALAVLTree_Insert( channel_tree, node->channel, node->channel, NULL );
        PCR_Assert( status != OALAVLTree_AlreadyInserted );

        if (status != OALAVLTree_InsertSuccess)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALRemoteChannelMgr: add channel: Failed to add channel (out of memory).");
        }
    }
    DOFMem_Free(node);
}

boolean OALRemoteChannelMgr_AddChannel(OALChannel channel)
{
    ChannelNode_t * node;

    PCR_Assert( channel );

    node = (ChannelNode_t *) DOFMem_Alloc(sizeof(ChannelNode_t));
    if(node != NULL)
    {
        node->channel = OALChannel_Retain( channel );
        if (node->channel)
        {
            OALEnqueueItem_Init((OALEnqueueItem) node, OALRemoteChannelMgr_AddChannel_fn);

            PCRMutex_Lock(messageEnqueueMutex);
            if ( !message_enqueue || !OALEnqueue_Enqueue( message_enqueue, (OALEnqueueItem)node ) )
            {
                PCRMutex_Unlock(messageEnqueueMutex);
                OALChannel_Release( channel );
                DOFMem_Free(node);
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALRemoteChannelMgr: Add channel: Failed to add channel to remote channel manager (out of memory).");
                return FALSE;
            }
            PCRMutex_Unlock(messageEnqueueMutex);
            return TRUE;
        }
        DOFMem_Free(node);
    }
    return FALSE;
}
/* ********************************************************************************/

/* ********************************************************************************/
/* Note: This function can only be called from within a channel message_enqueue worker thread. */
void OALRemoteChannelMgr_RemoveChannel_Immediate(OALChannel channel)
{
    if(channel)
    {
        EPP_NegOne_NotifyChannelDown(channel, channel->remoteChannelData->address);
        /* if the channel tree does not exist then the remote channel manager is probably shutting down. */
        if (channel_tree)
        {
            OALAVLTree_Delete( channel_tree, channel );
        }
        OALChannel_Lock(channel);
        OALObservable_SetChangedAndNotifyObservers(channel->chObservable);
        OALChannel_Unlock(channel);
        OALChannel_Release( channel );
    }
}

typedef struct ProtocolStackNode_t
{
    OALEnqueueItem_t super;

    OALProtocolStack protocolStack;
} ProtocolStackNode_t;

/* ********************************************************************************/
static void OALRemoteChannelMgr_AddProtocolStack_fn( OALThreadPoolItem dnode, boolean shutdown )
{
    ProtocolStackNode_t * node;

    PCR_Assert( dnode != NULL );

    node = (ProtocolStackNode_t *) dnode;
    if ( shutdown )
    {
        OALProtocolStack_Release( node->protocolStack );
        DOFMem_Free(node);
        return;
    }

    /* make sure that the protocolStack is not already in the tree */
    PCRMutex_Lock( protocolStackTreeMutex );
    {
        OALAVLTree_InsertionStatus status = OALAVLTree_Insert( protocolstack_tree, node->protocolStack, node->protocolStack, NULL );
        PCR_Assert( status != OALAVLTree_AlreadyInserted );
    }
    PCRMutex_Unlock( protocolStackTreeMutex );
    DOFMem_Free(node);
}

boolean OALRemoteChannelMgr_AddProtocolStack(OALProtocolStack protocolStack)
{
    ProtocolStackNode_t * node;

    PCR_Assert( protocolStack );

    node = (ProtocolStackNode_t *) DOFMem_Alloc(sizeof(ProtocolStackNode_t));
    if(node != NULL)
    {
        OALEnqueueItem_Init((OALEnqueueItem) node, OALRemoteChannelMgr_AddProtocolStack_fn);
        node->protocolStack = OALProtocolStack_Retain( protocolStack );
        if (node->protocolStack)
        {
            PCRMutex_Lock(messageEnqueueMutex);
            if (!message_enqueue || !OALEnqueue_Enqueue( message_enqueue, (OALEnqueueItem) node ))
            {
                PCRMutex_Unlock(messageEnqueueMutex);
                DOFMem_Free(node);
                OALProtocolStack_Release( protocolStack );
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALRemoteChannelMgr: Add protocol stack: Failed to add protocol stack to remote channel manager.");
                return FALSE;
            }
            PCRMutex_Unlock(messageEnqueueMutex);
            return TRUE;
        }
        DOFMem_Free(node);
    }
    return FALSE;
}
/* ********************************************************************************/

/* ********************************************************************************/

void OALRemoteChannelMgr_RemoveProtocolStack_Immediate( OALProtocolStack protocolStack )
{
    boolean releaseOk;

    PCRMutex_Lock( protocolStackTreeMutex );
    releaseOk = OALAVLTree_Delete(protocolstack_tree, protocolStack);
    PCRMutex_Unlock( protocolStackTreeMutex );

    if(releaseOk)
    {
        OALProtocolStack_Release(protocolStack);
    }
}

/* ********************************************************************************/

/* ********************************************************************************/
static void OALRemoteChannelMgr_SendMessageToStack_fn( OALThreadPoolItem dnode, boolean shutdown )
{
    OALProtocolStack    protocolStack;
    OALMessage          message;


    PCR_Assert( dnode != NULL);
    message = (OALMessage)dnode;
    message->shutdown = shutdown;

    PCRMutex_Lock( protocolStackTreeMutex );
    protocolStack = (OALProtocolStack)OALAVLTree_Retrieve( protocolstack_tree, message->recipient );
    PCRMutex_Unlock( protocolStackTreeMutex );

    if (!protocolStack)
    {
        OALMessage_Destroy(message);
        return;
    }

    OALProtocolStack_HandleMessage( protocolStack, message );
}

void OALRemoteChannelMgr_SendMessageToStack(OALMessage message)
{
    PCR_Assert( message );

    if(!message->already)
    {
        OALEnqueueItem_Init((OALEnqueueItem)message, OALRemoteChannelMgr_SendMessageToStack_fn);
        PCRMutex_Lock(messageEnqueueMutex);
        if (message_enqueue)
        {
            message->already = 1;
            if(!OALEnqueue_Enqueue( message_enqueue, (OALEnqueueItem)message))
            {
                OALMessage_Destroy(message);
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALRemoteChannelMgr: send message to stack: Failed to send message to stack.");
            }
        }
        else
        {
            OALMessage_Destroy(message);
        }
        PCRMutex_Unlock(messageEnqueueMutex);
    }
    else
    {
        OALEnqueueItem_Init((OALEnqueueItem)message, OALRemoteChannelMgr_SendMessageToStack_fn);     /* Let's be consistant */
        OALRemoteChannelMgr_SendMessageToStack_fn((OALThreadPoolItem)message, message->shutdown);
    }
}
/* ********************************************************************************/

/* ********************************************************************************/
static void OALRemoteChannelMgr_SendMessageToChannel_fn( OALThreadPoolItem dnode, boolean shutdown )
{
    OALMessage message;
    OALChannel channel = NULL;

    PCR_Assert( dnode != NULL);
    message = (OALMessage)dnode;
    message->shutdown = shutdown;

    if (channel_tree)
    {
        channel = (OALChannel)OALAVLTree_Retrieve( channel_tree, message->recipient );
    }

    if(!channel || channel->remoteChannelData->flags.isInShutdownState)
    {
        OALMessage_Destroy( message );
        return;
    }

    channel->msg_handler_func( channel, message );
}

void OALRemoteChannelMgr_SendMessageToChannel_Internal(OALMessage message, boolean forceEnqueue)
{
    PCR_Assert( message );

    if (!message->already || forceEnqueue)
    {
        OALEnqueueItem_Init((OALEnqueueItem)message, OALRemoteChannelMgr_SendMessageToChannel_fn);
        PCRMutex_Lock(messageEnqueueMutex);
        if(message_enqueue)
        {
            message->already = 1;
            if(!OALEnqueue_Enqueue( message_enqueue, (OALEnqueueItem)message ))
            {
                OALMessage_Destroy( message );
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALRemoteChannelMgr: internal send message to channel: Message to channel not sent (OALEnqueue_Enqueue failed).");
            }
        }
        else
        {
            OALMessage_Destroy(message);
        }
        PCRMutex_Unlock(messageEnqueueMutex);
    }
    else
    {
        OALEnqueueItem_Init((OALEnqueueItem)message, OALRemoteChannelMgr_SendMessageToChannel_fn);     /* Let's be consistant */
        OALRemoteChannelMgr_SendMessageToChannel_fn((OALThreadPoolItem)message, message->shutdown);
    }
}
/* ********************************************************************************/

void OALRemoteChannelMgr_SendShutdownToStack(OALProtocolStack protocolStack, DOFException except)
{
    OALMessage msg = NULL;

    PCRMutex_Lock(messageEnqueueMutex);
    if (protocolStack->shutdownMessage && !protocolStack->shutdownMessageSent)
    {
        protocolStack->shutdownMessageSent = TRUE;

        msg = OALMessage_Retain(protocolStack->shutdownMessage);
        SMT_Shutdown_SetException(msg, except);

    }
    PCRMutex_Unlock(messageEnqueueMutex);

    if (msg)
        OALRemoteChannelMgr_SendMessageToStack( msg );
    /* If msg is NULL then the stack was already sent the shutdown message. */
}

/* ********************************************************************************/
static boolean OALRemoteChannelMgr_TraverseSendToAll_fn(OALAVLTreeItem_t item, void *ctx)
{
    OALChannel channel = (OALChannel)item;
    DOFOperation op = (DOFOperation)ctx;
    if (op->securityScope)
    {
        if (!DOFSecurityScope_IsCompatible( OALChannel_GetSecurityScope( channel, FALSE ), op->securityScope ))
           return FALSE;
    }
    /* forwarding nodes can't route back to the source */
    if (channel != op->sources->channel && channel->remoteChannelData && channel->remoteChannelData->flags.isConnected)
    {
        channel->routeToProvider( channel, op, NULL, NULL );
    }

    return FALSE;
}

typedef struct OperationNode_t
{
    OALEnqueueItem_t super;

    DOFOperation     operation;
} OperationNode_t;

static void OALRemoteChannelMgr_SendToAll_fn( OALThreadPoolItem ditem, boolean shutdown )
{
    OperationNode_t * item = (OperationNode_t *) ditem;

    PCR_Assert(ditem != NULL);
    if (!shutdown && channel_tree)
    {
        OALAVLTree_Traverse( channel_tree, OALRemoteChannelMgr_TraverseSendToAll_fn, item->operation );
    }

    DOFOperation_Internal_Destroy( item->operation );
    DOFMem_Free(item);
}

void OALRemoteChannelMgr_SendToAll( DOFOperation operation )
{
    OperationNode_t * node;

    PCR_Assert( operation );
    node = (OperationNode_t *) DOFMem_Alloc(sizeof(OperationNode_t));
    if(node != NULL)
    {
        DOFOperation_Lock(operation);
        operation->flags.bits.isFlooded = TRUE;
        DOFOperation_Unlock(operation);
        OALEnqueueItem_Init((OALEnqueueItem) node, OALRemoteChannelMgr_SendToAll_fn);
        node->operation = operation;
        DOFOperation_Retain( operation );

        PCRMutex_Lock(messageEnqueueMutex);
        if(!message_enqueue)
        {
            DOFOperation_Internal_Destroy( operation );
            DOFMem_Free(node);
        }
        else
        {
            OALEnqueue_Enqueue( message_enqueue, (OALEnqueueItem) node );
        }
        PCRMutex_Unlock(messageEnqueueMutex);
    }
}

/* ********************************************************************************/

static boolean OALRemoteChannelMgr_TraverseSendToAllStateless_fn(OALAVLTreeItem_t item, void *ctx)
{
    OALChannel channel = (OALChannel)item;
    DOFOperation op = (DOFOperation)ctx;
    if (op->securityScope)
    {
        if (!DOFSecurityScope_IsCompatible( OALChannel_GetSecurityScope( channel, FALSE ), op->securityScope ))
           return FALSE;
    }
    /* forwarding nodes can't route back to the source */
    if (channel != op->sources->channel && channel->remoteChannelData && channel->remoteChannelData->flags.isConnected)
    {
        if((!channel->remoteChannelData->flags.secureDesire && DOFConnectionConfig_GetConnectionType(channel->remoteChannelData->config.connection) == DOFCONNECTIONTYPE_DATAGRAM_STATELESS) ||
           DOFAddress_GetAddressType(channel->remoteChannelData->address) != DOFADDRESSTYPE_UNICAST )
        {
            channel->routeToProvider( channel, op, NULL, NULL );
        }
    }

    return FALSE;
}

static void OALRemoteChannelMgr_SendToAllStateless_fn( OALThreadPoolItem ditem, boolean shutdown )
{
    OperationNode_t * item = (OperationNode_t *) ditem;

    PCR_Assert(ditem != NULL);
    if (!shutdown && channel_tree)
    {
        OALAVLTree_Traverse( channel_tree, OALRemoteChannelMgr_TraverseSendToAllStateless_fn, item->operation );
    }

    DOFOperation_Internal_Destroy( item->operation );
    DOFMem_Free(item);
}

void OALRemoteChannelMgr_SendToAllStateless( DOFOperation operation )
{
        OperationNode_t * node;

    PCR_Assert( operation );
    node = (OperationNode_t *) DOFMem_Alloc(sizeof(OperationNode_t));
    if(node != NULL)
    {
        DOFOperation_Lock(operation);
        operation->flags.bits.isFlooded = TRUE;
        DOFOperation_Unlock(operation);
        OALEnqueueItem_Init((OALEnqueueItem) node, OALRemoteChannelMgr_SendToAllStateless_fn);
        node->operation = operation;
        DOFOperation_Retain( operation );

        PCRMutex_Lock(messageEnqueueMutex);
        if(!message_enqueue)
        {
            DOFOperation_Internal_Destroy( operation );
            DOFMem_Free(node);
        }
        else
        {
            OALEnqueue_Enqueue( message_enqueue, (OALEnqueueItem) node );
        }
        PCRMutex_Unlock(messageEnqueueMutex);
    }
}
/* ********************************************************************************/

static int OALRemoteChannelMgr_ProcessRecvSilenceTimerItem_fn(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown)
{
    OALChannel channel = (OALChannel)argv[0];

    PCR_Assert(argc==1);

    if(!shutdown)   /* handle the message, or let shutdown clean everything up. */
    {
        OALRemoteChannel_ProcessRecieveSilence(channel);
    }

    return 0;
}

static int OALRemoteChannelMgr_Send_heartbeat_fn(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown)
{
    OALChannel channel = (OALChannel)argv[0];

    PCR_Assert(argc==1);
    if(!shutdown)   /* handle the message, or let shutdown clean everything up. */
    {
        EPP_NegOne_Heartbeat(channel, NULL, FALSE);
    }

    return 0;
}

boolean OALRemoteChannelMgr_SetSendSilenceTimerItem( OALChannel channel )
{
    if (channel->remoteChannelData->flags.isConnection)
    {
        uint32 maxSilence;
        maxSilence = OALRemoteChannelConfig_GetMaxSendSilence(channel);
        if (maxSilence != DOF_TIMEOUT_NEVER)
        {
            void *args[1];

            args[0] = channel;

            if(!channel->remoteChannelData->maxSendSilenceTimerItem)
                channel->remoteChannelData->maxSendSilenceTimerItem = OALTimer_CreateItem( remoteChannelTimer,
                                                                                           OALRemoteChannelMgr_Send_heartbeat_fn, 1, args );
            if(channel->remoteChannelData->maxSendSilenceTimerItem)
                return (OALTimerItem_Insert( channel->remoteChannelData->maxSendSilenceTimerItem, maxSilence, TRUE,
                                             FALSE ) == DOFERROR_OK);
        }
    }
    return TRUE;
}

boolean OALRemoteChannelMgr_SetRecvSilenceTimerItem( OALChannel channel )
{
    channel->remoteChannelData->waitingForData = FALSE;     /* We got something, who cares what */

    if (channel->remoteChannelData->flags.isConnection && !channel->remoteChannelData->flags.defaultAddressIsMulticast
        && channel->remoteChannelData->config.connection->type != DOFCONNECTIONTYPE_HUB)
    {
        uint32 maxSilence;

        maxSilence = OALRemoteChannelConfig_GetMaxReceiveSilence(channel);
        if (maxSilence != DOF_TIMEOUT_NEVER)
        {
            if ( maxSilence > EPP_NEGONE_MAX_PING_DELAY + MIN_PING_DELAY )
                maxSilence -= EPP_NEGONE_MAX_PING_DELAY;
            else
                maxSilence = MIN_PING_DELAY;

            {
                if(!channel->remoteChannelData->maxReceiveSilenceTimerItem)
                    channel->remoteChannelData->maxReceiveSilenceTimerItem = OALTimer_CreateItem(remoteChannelTimer, OALRemoteChannelMgr_ProcessRecvSilenceTimerItem_fn, 1, (void**)&channel);
                if(channel->remoteChannelData->maxReceiveSilenceTimerItem)
                    return (OALTimerItem_Insert( channel->remoteChannelData->maxReceiveSilenceTimerItem, maxSilence,
                                                TRUE, FALSE ) == DOFERROR_OK);
            }
        }
    }
    return TRUE;
}

void OALRemoteChannelMgr_RemoveSilenceTimerItems( OALChannel channel )
{
    OALTimerItem_Destroy( channel->remoteChannelData->maxSendSilenceTimerItem, TRUE );
    channel->remoteChannelData->maxSendSilenceTimerItem = NULL;
    OALTimerItem_Destroy( channel->remoteChannelData->maxReceiveSilenceTimerItem, TRUE );
    channel->remoteChannelData->maxReceiveSilenceTimerItem = NULL;
}

static int send_negotiation_timeout(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown)
{
    OALChannel channel = (OALChannel)argv[0];

    /* Note: since this is not running in the message enqueue worker thread it is not safe to use the channel. */
    PCR_Assert(argc==1);
    if(!shutdown)
    {
        /* The exception will be copied when the shutdown message is created. */
        OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_PROTOCOL_NEGOTIATION_FAILED);
    }
    OALChannel_Release(channel);

    return 0;
}

boolean OALRemoteChannelMgr_CancelNegotiationTimer(OALChannel channel)
{
    boolean status = FALSE;     /* Return TRUE if the timer was cancelled. */

    /*  Note: This function should only be called by functions handling channel and stack messages. */
    /*  Remove the timer item.  If the OALTimerItem_Remove returns TRUE release the channel.
        If OALTimerItem_Remove returns FALSE then the timer item callback function will be called. */
    if (channel->remoteChannelData->negotiationTimerItem)
    {
        if (OALTimerItem_Remove(channel->remoteChannelData->negotiationTimerItem))
        {
            /* Release the channel since the timer callback function will not be called.*/
            OALChannel_Release(channel);
            status = TRUE;
        }
        OALTimerItem_Destroy(channel->remoteChannelData->negotiationTimerItem, FALSE);
        channel->remoteChannelData->negotiationTimerItem = NULL;
    }
    return status;
}

boolean OALRemoteChannelMgr_StartNegotiationTimeout(OALChannel ch, uint32 timeout)
{
    OALChannel channel;

    /* Start the ESP negotiation timer. */
    channel = OALChannel_Retain(ch);
    if (channel)
    {
        channel->remoteChannelData->negotiationTimerItem = OALTimer_CreateItem(remoteChannelTimer, send_negotiation_timeout, 1, (void**)&channel);
        if (!channel->remoteChannelData->negotiationTimerItem ||
            OALTimerItem_Insert(channel->remoteChannelData->negotiationTimerItem, timeout, FALSE, FALSE ) != DOFERROR_OK)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALRemoteChannelMgr: Failed to start negotiation timer." );
            if (channel->remoteChannelData->negotiationTimerItem)
            {
                /* The timer item was created, so it must have failed to insert into the timer. */
                OALTimerItem_Destroy(channel->remoteChannelData->negotiationTimerItem, TRUE);
                channel->remoteChannelData->negotiationTimerItem = NULL;
            }
            OALChannel_Release(channel);
            return FALSE;
        }
        return TRUE;
    }
    return FALSE;
}
