/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/error.h>
#include <internal/addressinternal.h>
#include <internal/channel_stack_messages.h>
#include <internal/cipher.h>
#include <internal/epp_negative_one.h>
#include <internal/error.h>
#include <internal/exception.h>
#include <internal/marshal.h>
#include <internal/permissionsetinternal.h>
#include <internal/remotechannel.h>
#include <internal/remotechannelmanager.h>
#include <internal/timeinternal.h>
#include <internal/securitymode.h>
#include "cipherprivate.h"


/* The strictly CCM part of this file starts halfway down at static GenerateMac(). Above that are things that may or
   may not be strictly CCM, but need not be exposed to outside files. */


/** @internal @addtogroup OALCipher
@{
**/


/*
**  "Private" Type and Method Declarations
*/


#define CCM_BLOCK_LENGTH                    16
#define MACSIZE_16                          16


#define MAX_QUICK_NODE_CHECK_TIME           (60 * 60 * 1000) /* 1 hour, in milliseconds. */


#define NODE_ID_INVALID_OFFSET_RANGE              0x04000000         /* This is spec limit (2**29 / 8). */
#define NODE_ID_WARNING_ADDITIONAL_OFFSET_RANGE   0x04000000         /* This is spec limit (2**29 / 8). */


static uint8 GetKeyType( const OALCipher cipher, uint8 periodOffset )
{
    if (cipher->keyDesc[CUR].tagged && cipher->keyDesc[CUR].periodOffset == periodOffset)
        return CUR;
    if (cipher->keyDesc[SLOT].tagged && cipher->keyDesc[SLOT].periodOffset == periodOffset)
        return SLOT;
    if (cipher->keyDesc[KEKKEY].tagged && cipher->keyDesc[KEKKEY].periodOffset == periodOffset)
        return KEKKEY;

    return UNTAGGED_KEY;      /* Because not CUR, SLOT, or KEKKEY, it is an old key represented by none of the above. */
}


/* @internal
   Compute nodeID offset value to place on the wire.

   @param nodeID     MUST be less than or equal to TOP_NODEID.
   @param baseNodeID MUST be less than or equal to TOP_NODEID.
*/
static uint32 OALCipher_GetOffsetFromNodeIDAndBase( uint32 nodeID, uint32 baseNodeID )
{
    boolean negativeOffset = (nodeID < baseNodeID);
    uint32  retOffset      = (negativeOffset ? (baseNodeID - nodeID) : (nodeID - baseNodeID)) << 1;

    PCR_Assert( nodeID <= TOP_NODEID && baseNodeID <= TOP_NODEID );
    PCR_Assert( retOffset <= 0x3FFFFFFE );                 /* Thus, is compressable on the wire. */

    return (retOffset | (uint32)negativeOffset);
}


/* @internal
   Compute actual nodeID from offset value found on the wire.

   @param wireNodeID MUST be less than or equal to 0x3FFFFFFF.
   @param baseNodeID MUST be less than or equal to TOP_NODEID.
*/
static uint32 OALCipher_GetNodeIDFromOffsetAndBase( uint32 wireNodeID, uint32 baseNodeID )
{
    boolean negativeOffset = (boolean)(wireNodeID & 0x00000001);
    uint32  retNodeID      = wireNodeID >> 1;

    retNodeID = (negativeOffset ? (baseNodeID - retNodeID) : (retNodeID - baseNodeID)) & TOP_NODEID;

    PCR_Assert( wireNodeID <= 0x3FFFFFFF && baseNodeID <= TOP_NODEID );
    PCR_Assert( retNodeID <= TOP_NODEID );

    return retNodeID;
}


static void OALCipher_ResolveManagerSlot( OALCipher cipher, DOFAddress senderAddress, uint16 slot )
{
    if (cipher->flags.manager && OALCipher_IsSecureGroup( cipher ))          /* Slots vary only for secureGroup case. */
    {
        /*The CCM protocol spec discusses having manager adjust its slot to other members, emphasizing manager catchup*/

        /* We choose to ignore having even a WAY fast manager slow down its slot. Discussion:
           SecureGroup members (CCM non-managers) catch up their slot at every secure PDU received from SGM. If SGM
           sends none, the problem is not that the manager is WAY fast, but that the members are not synchronizing their
           slot by following whatever manager packets come along, which is a much smaller problem. But even that kind of
           non-talkative SGM will eventually send a rekey PDU- and being fast it is 1st to sense the need for slotRekey
           and distributes that rekey PDU more promptly than others might expect it to come. So, in summary, members
           either catch up their slot all along the way or as late as when the rekey PDU is received; there is no
           huge problem for either case. The same fast clock for the SGMP module means the SGM sends heartbeats
           sooner than expected and group members are less prone to step up as SGM. We choose to allow this behavior. */
        /* So we implement CCM slot catchup here. It catches up CCM slot but does not alert the SGMP module SGM
           that his rekey distribution algorithm can be slow. Here is a discussion, though belonging in DOFSecureGroup,
           is here for convenience. When CCM manager is way slow, its SGMP SGM might sense a need for slotRekey later
           than it could have. And so it might be that it distributes the new key as much as a few seconds later than
           desired- if its clock is five times as slow, it changes slot every 5 seconds instead of every 1 second. This
           whole time, members are tracking their leader's slot and, for the case given are never more than 1020 seconds
           ahead (4-1/4 minutes between max hb times times (5-1)). But rather than use the above mentioned algorithm,
           we choose instead to rekey at slot 31232 rather than anywhere near to max slot 32767. The SGM therefore
           rekeys earlier than required by 1535 slots- very sufficient to avoid any slot problem, and trivial in its
           impact: we rekey every 8 hours 40 minutes instead of 9 hours 6 minutes. Now of course the extremely behind
           SGM in this example will distribute heartbeats 5 times slower than normal, but no member actually takes over
           as SGM because we keep them down by answering their required RequestGroup PDU. */
        /* Do not allow some wild node (like 10x faster than everybody else) to influence my manager slot very much. This
           implementation: If my slot is demonstrably behind the group (but ignoring that wild node), increment slot by
           a small amount to catch up. To avoid using trees or lists and other overhead, we assume exactly one node out
           there is "wild." If none are really wild, no big deal. More than one wildly fast node is presumed to be even
           more rare than 1 wildly fast node- we do not notch out its effect here, though its effect is not dominant.
           Aside from the wild node exclusion, nodes that transmit more packets to the manager have dominant effect. */
        uint16 nowSlot;

        PCR_Assert( cipher->quickNodeStore );
        if (OALCipher_GetCurKeySlot( cipher, &nowSlot ))
        {
            /* First, gather info. NOTE: Our work is harmless if we now suddenly become NOT manager. */
            uint32 totalAheadCount;
            uint16 slotAheadCount = slot - nowSlot;

            slotAheadCount = (slotAheadCount & 0x8000) ? 0 : slotAheadCount;                  /* No behind counts. */
            if (!cipher->quickNodeStore->quickestNode.addr ||
                !DOFAddress_IsEqual( cipher->quickNodeStore->quickestNode.addr, senderAddress ))
            {
                if (!cipher->quickNodeStore->quickestNode.addr ||                             /* First quickest node. */
                    cipher->quickNodeStore->quickestNode.slotAheadCount < slotAheadCount)     /* New   quickest node. */
                {
                    /* Ignore senderAddress's report for computations; replace quickestNode for future comparisons. */
                    cipher->quickNodeStore->quickestNode.slotAheadCount = slotAheadCount;     /* Replace. */
                    DOFAddress_Destroy( cipher->quickNodeStore->quickestNode.addr );
                    cipher->quickNodeStore->quickestNode.addr = DOFAddress_Copy( senderAddress );
                    /* This is not catastrophic if the copy fails; we'll just fill it in next time around. */
                }
                else if (slotAheadCount >= 3)
                    cipher->quickNodeStore->aheadByThreeOrMoreCount++;
                else if (slotAheadCount >= 2)
                    cipher->quickNodeStore->aheadByTwoCount++;
                else
                    cipher->quickNodeStore->notAheadCount++;
            }
            else
            {
                /* Update already held ->quickestNode->addr's ahead count. */
                cipher->quickNodeStore->quickestNode.slotAheadCount = slotAheadCount;
            }
            /* Last, if appropriate, adjust curKey slot forward. */
            totalAheadCount = cipher->quickNodeStore->aheadByThreeOrMoreCount + cipher->quickNodeStore->aheadByTwoCount;
            if ((totalAheadCount * 2) > (totalAheadCount + cipher->quickNodeStore->notAheadCount))
            {
                uint16 slotForwardCount = (cipher->quickNodeStore->aheadByThreeOrMoreCount >
                                           cipher->quickNodeStore->aheadByTwoCount) ? 3 : 2;

                cipher->periodStore[PERIOD_OFFSET( cipher, CUR )].dofKey.txSlotAdjust += slotForwardCount;
                OALCipher_ResetQuickNodeStore( cipher );
                if(DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "_ResolveManagerSlot() advances manager curKey slot by 0x", (uint32)slotForwardCount );
            }
            else if (DOFTime_Compare( PCRTime_GetMS(), (cipher->quickNodeStore->beginQuickNodeStoreTime +
                                                        MAX_QUICK_NODE_CHECK_TIME) ) >= 0)
                OALCipher_ResetQuickNodeStore( cipher );
        }
    }
}

static void ScheduleNewNodeID( OALCipher cipher, uint8 quadrant )
{
    uint8 factor = 1;

    PCR_Assert( OALCipher_IsSecureGroup( cipher ) );
    /* React based on which quadrant of the warning range are we in. */
    switch (quadrant)
    {
    case 0:                                                       /* Sometime in next 1.16 hours. */
        factor *= 2;
    case 1:                                                       /* Sometime in next 2.33 hours. */
        factor *= 2;
    case 2:                                                       /* Sometime in next 4.66 hours. */
        {
            /* Random delay from 0 to (2**24 - 1) msec, or 0 to 16777215 msec; divided by factor. */
            uint32 millisecondsDelayToNewNodeIDRequest = (PCRRandom_Get() & 0x00ffffff ) / factor;

            cipher->flags.newNodeIDScheduled = 1;
            /* This next line first removes a previously inserted timer item. */
            OALTimerItem_Insert( cipher->getNodeIDItem, millisecondsDelayToNewNodeIDRequest, FALSE, FALSE );
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "Will send CCM.Request PDU in seconds 0x",
                                                         millisecondsDelayToNewNodeIDRequest / 1000 );
            break;
        }
    default:
        break;                                          /* Do nothing yet. */
    }
}


/* @internal

   @note pQuadrant filled only on NID_WARNING return.
   @return NID_OK, NID_INVALID, or NID_WARNING.
*/
static uint8 CheckNodeIDRangeViolation( const OALCipher cipher, uint8 * pQuadrant )
{
    PCR_Assert( OALCipher_IsSecureGroup( cipher ) );
    /* Only relevant for non-manager, and only if the advNNI is new and valid. */
    if (!cipher->flags.manager && OALCipher_IsValid_AdvertisedNNI( cipher ) && cipher->previousAdvNNI != cipher->advNNI)
    {
        uint32 nextNodeID = ( cipher->advNNI - cipher->nodeID ) & TOP_NODEID;               /* Normalize to nodeID 0. */

        cipher->previousAdvNNI = cipher->advNNI;                                      /* A new one, so act on it now. */

        if (nextNodeID > (TOP_NODEID - NODE_ID_INVALID_OFFSET_RANGE - NODE_ID_WARNING_ADDITIONAL_OFFSET_RANGE))
        {
            uint32 deltaWarningNodeID;
            if (nextNodeID > (TOP_NODEID - NODE_ID_INVALID_OFFSET_RANGE))
                return NID_INVALID;
            deltaWarningNodeID = TOP_NODEID - NODE_ID_INVALID_OFFSET_RANGE - nextNodeID;
            *pQuadrant = deltaWarningNodeID / MAXIMUM_NEW_ADVERTISED_NEXT_NODEID_OFFSET;
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "deltaWarningNodeID: 0x", deltaWarningNodeID );
            PCR_Assert( cipher->advNNI <= TOP_NODEID );
            return NID_WARNING;
        }
    }
    return NID_OK;
}


static void InvalidateNodeIDAndGetNewNodeID( OALCipher cipher )
{
    PCR_Assert( cipher );
    cipher->flags.nodeIDValid = 0;
    OALCipher_GetNewNodeID( cipher );
}


/**
    Create a MAC using CBC-MAC of the entire packet, with the specified IV as per CCM specification.

    @note The cipher instance mutex needs to be locked by caller of this function.
    @param ekey         The expanded key to use in computation. This must not be NULL.
    @param macAuthFlags The MAC authorization flags to use in computation.
    @param pNonce       Pointer to the nonce used to decrypt the text at data. This must not be NULL.
    @param a_dlen       Length of a_data buffer. This MUST be greater than 0.
    @param a_data       The a_data buffer. This must not be NULL.
    @param dlen         Length of text data buffer. 0 is odd but legal.
    @param data         The text data buffer. This must not be NULL.
    @param mlen         Length of MAC buffer. This must be less than or equal to 16.
    @param generatedMAC [OUT] The supplied MAC buffer. This must not be NULL.
**/
static void GenerateMac( const OALCryptExpandedKey ekey, uint8 macAuthFlags, const AESCCM_Nonce_t * pNonce,
                         uint32 a_dlen, const uint8 * a_data, uint32 dlen, const uint8* data, uint8 mlen,
                         uint8 * generatedMAC )
{
    uint32 i;
    uint8  fullMAC[MACSIZE_16];
    uint8  fullMACOffset;
    uint8  adder = (a_dlen < 0x0000FF00) ? 2 : 6;

    fullMAC[0] = macAuthFlags; /* Fill fullMAC with the IV. All 16 bytes are filled*/
    OALCipher_MarshalNonce(&fullMAC[1], pNonce); /* q is 4, n is 11: means nonce bytes to fill is 11.*/
    fullMAC[12] = (uint8)( dlen >> 24 );
    fullMAC[13] = (uint8)( dlen >> 16 );
    fullMAC[14] = (uint8)( dlen >> 8 );
    fullMAC[15] = (uint8)( dlen );                                /* B0 for use across all data (A data plus P data). */
    PCR_Assert( a_dlen );
    /* Process submitted a_data portion, modifying fullMAC along the way. */
    for (i = 0; (i < (a_dlen + adder)); i++)                      /* We will insert 2 or 6 encoded length bytes. */
    {
        fullMACOffset = i % sizeof( fullMAC );

        /* Each time we wrap to beginning of an AES-CCM block (16 bytes), encrypt fullMAC. */
        if (fullMACOffset == 0)
        {
            /* We already validated the cipher exists when the expanded key was created. */
            DOFCipherRegistry_AES256Cipher->Encrypt( ekey, fullMAC ); /* Form Y0, Y1, . . . */
        }
        if (i < adder)
        {
            switch (i)
            {
            case 0:
                fullMAC[fullMACOffset] ^= ((adder == 2) ? (uint8)(a_dlen >> 8) : 0xFF);
            break;
            case 1:
                fullMAC[fullMACOffset] ^= ((adder == 2) ? (uint8)(a_dlen & 0x00FF) : 0xFE);
            break;
            case 2:
                fullMAC[fullMACOffset] ^= (uint8)(a_dlen >> 24);
            break;
            case 3:
                fullMAC[fullMACOffset] ^= (uint8)((a_dlen >> 16) & 0x000000FF);
            break;
            case 4:
                fullMAC[fullMACOffset] ^= (uint8)((a_dlen >> 8) & 0x000000FF);
            break;
            case 5:
                fullMAC[fullMACOffset] ^= (uint8)(a_dlen & 0x000000FF);
            break;
            default:
                PCR_Assert( FALSE );
            break;
            }
        }
        else
            fullMAC[fullMACOffset] ^= a_data[i - adder];
    } /* end for. */
    DOFCipherRegistry_AES256Cipher->Encrypt( ekey, fullMAC ); /* Form last Y from Adata, will not be Y0: that was formed above*/
    /* fullMAC processing of A data now complete, do it some more for P data. */
    if (dlen > 0)
    {
        for (i = 0; i < dlen; i++)
        {
            uint8 fullMACOffset = i % sizeof( fullMAC );

            if (i && (fullMACOffset == 0))
                DOFCipherRegistry_AES256Cipher->Encrypt( ekey, fullMAC );     /* Form next Y. */
            fullMAC[fullMACOffset] ^= data[i];
        }
        DOFCipherRegistry_AES256Cipher->Encrypt( ekey, fullMAC );             /* Form last Y from P data. */
    }
    PCR_Assert( mlen <= sizeof( fullMAC ) );
    PCRMem_Copy( generatedMAC, fullMAC, mlen );                               /* Copy to out buffer. */
}


static void EncryptPData( const uint8 * pCtr0, const OALCryptExpandedKey ekey, uint8 * pData, uint32 dlen )
{
    uint32 i;
    uint32 block = 1;                    /* Generate the encrypted part of CTR stream and xor it into submitted data. */

    /* 0 dlen is odd but legal. */
    for (i = 0; i < dlen; i += CCM_BLOCK_LENGTH)
    {
        uint8 j;
        uint8 encryptedCtr[CCM_BLOCK_LENGTH];

        PCRMem_Copy( encryptedCtr, pCtr0, sizeof( encryptedCtr ) - 4 );
        encryptedCtr[12] = (uint8)(block >> 24);               /* Augment CTR with next block number. */
        encryptedCtr[13] = (uint8)(block >> 16);
        encryptedCtr[14] = (uint8)(block >> 8);
        encryptedCtr[15] = (uint8)block;
        block++;
        /* We already validated the cipher exists when the expanded key was created. */
        DOFCipherRegistry_AES256Cipher->Encrypt( ekey, encryptedCtr );
        /* xor in the ctr stream bytes.  (CTR stream only operates on the bytes of data, not the MAC) */
        for (j = 0; j < CCM_BLOCK_LENGTH; j++)
        {
            uint32 index = i + j;

            if (index < dlen)
                pData[index] ^= encryptedCtr[j];
            else
                break;
        }
    } /* end for. */
}


/*
**  "Public" Type and Method Declarations
*/


/** @internal
    Encrypt a given block of data.

    @note This call always uses the "current" key to encipher.
    @note The cipher instance mutex needs to be locked by the caller of this function.
    @param strength           The given cipher strength.
    @param eKey               The expanded key structure to use in computation. This must not be NULL.
    @param macAuthFlags       The MAC authorization flags to use in computation.
    @param pNonce             Pointer to the encipher nonce to use in computation. This must not be NULL.
    @param a_dlen             Length of the a_data block. This must be greater than 0.
    @param a_data             The a_data block. This must not be NULL.
    @param dlen               Length of the data block to encrypt (does not include maclength). This may be 0.
    @param data               The data block to encrypt. This must not be NULL if @a dlen > 0.
    @param mlen               Length of the output mac buffer. This must be less than or equal to 16.
    @param mac                The supplied output mac buffer. This must not be NULL.
    @param encrypt            TRUE for normal encryption/authentication. FALSE for authentication only.
    @return DOFERROR_OK
**/
DOFError AESCCM_Encipher( const DOFCipherStrength strength, const OALCryptExpandedKey ekey,
                          uint8 macAuthFlags, const AESCCM_Nonce_t * pNonce, uint32 a_dlen, const uint8 * a_data,
                          uint32 dlen, uint8 * data, uint8 mlen, uint8 * mac, boolean encrypt )
{
    uint8  ctr0[CCM_BLOCK_LENGTH];                          /* Only the unit test requires this not be static. */
    uint32 i;

    PCR_Assert( strength == DOFCIPHERSTRENGTH_256 );
    GenerateMac( ekey, macAuthFlags, pNonce, a_dlen, a_data, dlen, data, mlen, mac );
    /* Setup ctr0 (trailing block number is 0). */
    ctr0[0] = CTR_FLAGS_BYTE;
    OALCipher_MarshalNonce(&ctr0[1], pNonce);
    PCRMem_Set( &ctr0[12], 0, 4 );
    if (encrypt)
        EncryptPData( (const uint8 *)&ctr0[0], ekey, data, dlen );
    /* P Data encrypt or not is complete, so now fill mac. */
    /* We already validated the cipher exists when the expanded key was created. */
    DOFCipherRegistry_AES256Cipher->Encrypt( ekey, ctr0 );  /* Form S0 from ctr0. */
    for (i = 0; i < mlen; i++)
        mac[i] ^= ctr0[i];

    return DOFERROR_OK;
}


/** @internal
    Decrypt a given block of data.

    @note The cipher instance mutex needs to be locked by the caller of this function.
    @param strength           The given cipher strength.
    @param eKey               The key structure to use in computation. This must not be NULL.
    @param macAuthFlags       The MAC authorization flags to use in computation.
    @param pNonce             Pointer to the decipher nonce to use in computation. This must not be NULL.
    @param a_dlen             Length of the a_data block. This MUST be greater than 0.
    @param a_data             The a_data block. This must not be NULL.
    @param dlen               Length of the data block to decrypt (does not include maclength). This may be 0.
    @param data               The data block to decrypt. This must not be NULL if @a dlen > 0.
    @param mlen               Length of the supplied mac. This must be less than or equal to 16.
    @param mac                The supplied mac buffer. This must not be NULL.
    @param encrypt            TRUE for normal decryption/authentication. FALSE for authentication only.
    @return DOFERROR_OK for successful decryption, or DOFERROR_DECRYPTION_FAILED if final check failed.
*/
DOFError AESCCM_Decipher( const DOFCipherStrength strength, const OALCryptExpandedKey ekey,
                          uint8 macAuthFlags, const AESCCM_Nonce_t * pNonce, uint32 a_dlen, const uint8 * a_data,
                          uint32 dlen, uint8 * data, uint8 mlen, uint8 * mac, boolean encrypt )
{
    uint8    ctr0[CCM_BLOCK_LENGTH];                        /* Only the unit test requires this not be static. */
    uint8    computedMAC[MACSIZE_16];
    uint8    i;
    DOFError retStatus = DOFERROR_OK;

    PCR_Assert( strength == DOFCIPHERSTRENGTH_256 );
    ctr0[0] = CTR_FLAGS_BYTE;
    OALCipher_MarshalNonce(&ctr0[1], pNonce); /* This is ctr0. */
    PCRMem_Set( &ctr0[12], 0, 4 );
    PCR_Assert( a_dlen );                                      /* Every OAL secure packet has A data. */
    if (encrypt)
        EncryptPData( (const uint8 *)&ctr0[0], ekey, data, dlen );
    /* Compare received mac buffer with computedMAC derived from decrypted data. */
    PCR_Assert( mlen <= sizeof( computedMAC ) );
    GenerateMac( ekey, macAuthFlags, pNonce, a_dlen, a_data, dlen, data, mlen, computedMAC );
    /* We already validated the cipher exists when the expanded key was created. */
    DOFCipherRegistry_AES256Cipher->Encrypt( ekey, ctr0 );     /* Form S0 from ctr0. */
    for (i = 0; i < mlen; i++)
        computedMAC[i] ^= ctr0[i];
    if (PCRMem_Compare( mac, computedMAC, mlen ) != 0)
    {
        retStatus = DOFERROR_DECRYPTION_FAILED;
        PCRMem_Set( data, 0, dlen );                           /* Clear out data to avoid revealing it. */
    }
    PCRMem_Set( mac, 0, mlen );                                /* Clear out input mac to avoid revealing it. */

    return retStatus;
}


/** @internal
    Encrypt given stream with given cipher instance, and write CCM security mode field of the stream security header.
    @note This call always uses the "current" key to encipher.
    @note After successful encipher, the encipherPacketNumber is incremented for next use.
    @note The last 4 params influence PDU only if this cipher is of a secureGroup connection.
    @param self                   The given cipher instance. Among other things, this establishes whether to encrypt.
                                  This must not be NULL.
    @param pEPPHeaderStream       Ptr to EPP header that precedes payload. Header not encrypted, but authenticated. This
                                  must not be NULL, and its contents must not be NULL.
    @param pPayloadStream         Ptr to the given payload stream- all except payloadEncryptOffset of this is encrypted.
                                  This must not be NULL, and its contents must not be NULL.
    @param payloadEncryptOffset   Offset to 1st byte of payload that shall be encrypted (0 to encrypt all payload).
    @param unicastTargetAddress   The targeted address for this PDU. Cipher enforces that targeted PDUs are only
                                  processed by their target.  Other nodes that see the PDU are prohibited from
                                  processing the PDU.  This may be NULL if any/every node is permitted to process the
                                  PDU (multicast).
    @param denyPacketReplayProtectionIfSecureUnicastSend TRUE when the normal targetNodeID restrictions (given by @a
                                                         unicastTargetAddress) must @b not be enforced. Otherwise, this
                                                         should be FALSE.
    @param secureMarkingID        The target Node ID, overriding anything designated by @a unicastTargetAddress.
                                  Otherwise, EPP_NEGONE_NO_SECURE_MARKING_ID.
    @param ifProperPlaceNextBaseNIDOnWire TRUE for management/control PDUs, on which additional cipher state information
                                          may be included for the benefit of synchronizing with the other node(s) in the
                                          session. This is generally not necessary for most PDUs, but PDUs that are
                                          already controlling or managing session state are good candidates for
                                          distributing this additional cipher knowledge.
    @return DOFERROR_OK for encryption, or
                - DOFERROR_ILLEGAL_ARGUMENT if payloadEncryptOffset is beyond supplied encrypt data, or
                - DOFERROR_KEY_NOT_SET if cipher has no curKey- no encryption takes place, or
                - DOFERROR_REJECT_OUTGOING_PACKET_NUMBER if packet number has reached 0x40000000, or
                - DOFERROR_REQUIRED_TARGETNID_NOT_KNOWN if targetNID required but not found in cipher's addr table, or
                - DOFERROR_NODEID_NOT_VALIDATED if CIPHER_CONFIG_SECUREGROUP_DEFAULT and cipher's nodeID not valid, or
                - DOFERROR_INSUFFICIENT_RESOURCES if unable to create a new stream or write to one, or
                - DOFERROR_ENCIPHER_UNMARSHAL_ERROR if unable to read a given stream, or
                - DOFERROR_INTERNAL_ERROR on class member compression error.
**/
DOFError OALCipher_Encipher( const OALCipher self, const OALStream * pEPPHeaderStream, OALStream * pPayloadStream,
                             uint8 payloadEncryptOffset,        /* Anticipated as 0, but here for future flexibility. */
                             DOFAddress unicastTargetAddress, boolean denyPacketReplayProtectionIfSecureUnicastSend,
                             uint32 secureMarkingID, boolean ifProperPlaceNextBaseNIDOnWire )
{
    OALStream             outStream;
    int16 *               pCurKeyTxLastSlot;
    DOFCipherStrength     cipherStrength;
    const uint8 *         a_data;
    uint32                a_dlen;
    uint32                dlen;
    uint32                maxOutputLength;
    uint32                targetNID;
    DOFError              status           = DOFERROR_INTERNAL_ERROR; /* Not overwritten only in case of extreme error*/
    uint16                txSlot           = 0;
    boolean               isSecureGroup;
    boolean               includeTargetNID = FALSE;
    uint8                 mlen;
    uint8                 flags;

    /* Note: pEPPHeaderStream is NOT encrypted. At successful exit its contents and read/write offsets are as submitted.
             If encrypting, every byte of the submitted pPayloadStream is encrypted
                (except the first payloadEncryptOffset bytes, which is expected to be 0 bytes).
             Regardless, pPayloadStream returns with @a pEPPHeaderStream, followed by the new cipher header prepended, and MAC appended at the end. */
    PCR_Assert( self );
    PCR_Assert( self->addressToNodeIDTree );
    PCR_Assert( pEPPHeaderStream );
    PCR_Assert( pPayloadStream );
    a_dlen = OALStream_GetDataSize( *pEPPHeaderStream );
    PCR_Assert( a_dlen );                                            /* Every OAL secure packet has EPP header bytes. */
    dlen = OALStream_GetDataSize( *pPayloadStream );                 /* Legal (though odd) if dlen is 0. */
    mlen = DOFSecurityModeCCM_GetMACSize( self->normalSecurityMode );
    PCR_Assert( mlen <= 16 );
    maxOutputLength = a_dlen +                                       /* length of prepended EPPHeaderStream. */
                      1 +                                            /* length of flags byte. */
                      4 +                                            /* max length of nodeID (if supplied). */
                      2 +                                            /* max length of slot (if supplied). */
                      4 +                                            /* max length of packetNumber (if supplied). */
                      4 +                                            /* max length of targetNID (if supplied). */
                      4 +                                            /* max length of advNNI (if supplied). */
                      dlen +                                         /* actual payload length. */
                      mlen;                                          /* actual mac length. */
    if ((outStream = OALStream_CreateRW( maxOutputLength )) == NULL) /* Form outStream. We encrypt in place later. */
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALCipher: ERROR: Encipher failed to create rw stream.");
        return DOFERROR_INSUFFICIENT_RESOURCES;
    }
    if (OALStream_Write( outStream, a_dlen, OALStream_GetReadPtr( *pEPPHeaderStream ) ) == FALSE) /* Start outStream w/ EPPHeaderStream*/
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALCipher: ERROR: Encipher failed to read given input stream.");
        status = DOFERROR_ENCIPHER_UNMARSHAL_ERROR;
        goto errorOut;
    }
    a_data = OALStream_GetReadPtr( outStream );
    OALStream_AdjustReadOffset( outStream, a_dlen );               /* Have a_data- now ignore held EPPHeaderStream. */
    /* Done with this use of a_dlen. It will be refilled and reused below. */
    OALCipher_AdjustKeys( self );
    OALCipher_Lock( self );                                        /* Lock: working with keys, ->addressToNodeIDTree. */
    if (!OALCipher_TaggedKeyEnabled( self, CUR ))
    {
        /*
         * We should never get here if OALCipher_IsSecureGroup( self ) is true because
         * previous call to _SecureGroupEncipherReady() passed, or PDU was put in a queue.
         */
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALCipher: ERROR: Encipher has no curKey, discarding PDU." );
        status = DOFERROR_KEY_NOT_SET;
        goto errorOutUnlock;
    }
    if (!self->flags.nodeIDValid)
    {
        PCR_Assert( OALCipher_IsSecureGroup( self ) );
        /* We actually never get here because previous call to _SecureGroupEncipherReady() passed, or it went to get our
           nodeID and PDU was put in a queue until the nodeID was obtained. */
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALCipher: ERROR: Encipher has invalid nodeID, discarding PDU." );
        status = DOFERROR_NODEID_NOT_VALIDATED;
        goto errorOutUnlock;
    }
    if ((isSecureGroup = OALCipher_IsSecureGroup( self )) != FALSE)/* To a node that shares its key with others. */
    {
#if DEEPER_TREE_LOG
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "Encipher() has addressToNodeIDTree item count of 0x",
                                                     OALAVLTree_GetNumItems( self->addressToNodeIDTree ) );
#endif
        /* secureMarkingID is EPP_NEGONE_NO_SECURE_MARKING_ID EXCEPT for special multicast cases. */
        targetNID = secureMarkingID;
        if (targetNID == EPP_NEGONE_NO_SECURE_MARKING_ID &&
            unicastTargetAddress &&                                /* Caller telling us this is unicast transmission. */
            DOFAddress_GetAddressType( unicastTargetAddress ) == DOFADDRESSTYPE_UNICAST &&
            !denyPacketReplayProtectionIfSecureUnicastSend)        /* Do include targetNID. */
        {
            /* Lookup targetNID from unicastTargetAddress in table kept by this cipher instance.
               Note: self->addressToNodeIDTree is filled by OALCipher_Decipher(). */
            if (OALCipher_GetStoredTargetNodeID( self, unicastTargetAddress, &targetNID ) == FALSE)
            {
                /* Normal use of the COAL would not give us this error, for two reasons.
                        1st: We think DOF practice is, before trying to send secure unicast to a node, confirm who
                             they are. For example, before OAP requesting, express interest and get a provide back. We
                             then have their nodeID.
                        2nd: Cooperation with EPP has us being called only if his call to OALCipher_EncipherReady()
                             first verifies that we have the targetNodeID (or don't need it). */
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALCipher: ERROR: Encipher called without needed targetNodeID, discarding PDU." );
                status = DOFERROR_REQUIRED_TARGETNID_NOT_KNOWN;
                goto errorOutUnlock;
            }
        }
        includeTargetNID = ((targetNID & 0xc0000000) == 0);       /* Proceed with targetNID filled in 1 of 2 ways. */
    }
    flags = (self->flags.manager << 7) |                             /* Fill flags byte and write it to outStream. */
            (self->periodStore[PERIOD_OFFSET( self, CUR )].period << 4);        /* Place curKey's period in flags. */
    if (includeTargetNID)
    {
        PCR_Assert( isSecureGroup );
        flags |= TRANSMIT_TARGET_NODEID;
    }
    if (self->flags.manager && isSecureGroup && !self->flags.hub && ifProperPlaceNextBaseNIDOnWire &&
        OALCipher_IsValid_AdvertisedNNI( self))             /* A newly rising manager may not have computed this yet. */
       flags |= TRANSMIT_BASE_OR_NEXT_NODEID;
    if (!self->flags.implicitCount)
        flags |= TRANSMIT_COUNT;
    if (!OALStream_Write8( outStream, flags ))
    {
        status = DOFERROR_INSUFFICIENT_RESOURCES;
        goto errorOutStreamProblemUnlock;
    }
    if (!self->flags.implicitNodeID)        /* Write 5 flags-stipulated fields to outStream (starting with ->nodeID). */
    {
        uint32 wireNodeID = OALCipher_GetOffsetFromNodeIDAndBase(self->nodeID, self->periodStore[PERIOD_OFFSET( self, CUR )].baseNodeID );

        if (OALMarshal_StreamCompress4( outStream, wireNodeID ) != DOFERROR_OK)
            goto errorOutStreamProblemUnlock;                     /* PCR_Assert( status == DOFERROR_INTERNAL_ERROR ); */
    }
    pCurKeyTxLastSlot = &self->periodStore[PERIOD_OFFSET( self, CUR )].dofKey.txLastSlot;  /* To r/w DOFKey_t struct. */
    PCR_Assert( *pCurKeyTxLastSlot >= 0);                                                  /*This int16 never negative*/
    if (!self->flags.ignoreSlot)
    {
        /* Note: slot on wire is 2-byte compressed, so presentSlot limited to 32767 (~9 hours). */
        txSlot = OALCipher_GetCurKeyTxSlot( self );
        PCR_Assert( txSlot <= 0x7FFF );
        if (txSlot < *pCurKeyTxLastSlot)
            txSlot = *pCurKeyTxLastSlot;                  /* Regardless of why it happened, bump up txSlot to match. */
        else if (txSlot > *pCurKeyTxLastSlot)
            self->encipherPacketNumber = 0;               /* curKey now in new slot, so clear ->encipherPacketNumber. */
        if (*pCurKeyTxLastSlot < (int16)txSlot)
            *pCurKeyTxLastSlot = (int16)txSlot;           /* Here is ONLY place that modifies .txLastSlot*/
        PCR_Assert( *(uint16 *)pCurKeyTxLastSlot <= 0x7FFF );
        if (OALMarshal_StreamCompress2( outStream, *(uint16 *)pCurKeyTxLastSlot ) != DOFERROR_OK)
            goto errorOutStreamProblemUnlock;
    }
    else
    {
        /* Ignore slot for this session/connection. In this mode, our implementation gives no ability to set .txLastSlot
           to anything, and so .txLastSlot remains at its initialized 0, so all tx nonce use slot of 0. */
        PCR_Assert( *(uint16 *)pCurKeyTxLastSlot == 0 );
    }
    PCR_Assert( txSlot == *(uint16 *)pCurKeyTxLastSlot );
    if (self->encipherPacketNumber >= 0xC0000000)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALCipher: ERROR: Encipher asked to use illegal packet number.");
        status = DOFERROR_REJECT_OUTGOING_PACKET_NUMBER;
        goto errorOutUnlock;
    }
    if (flags & TRANSMIT_COUNT)
        if (OALMarshal_StreamCompress4( outStream, self->encipherPacketNumber ) != DOFERROR_OK)
            goto errorOutStreamProblemUnlock;                     /* PCR_Assert( status == DOFERROR_INTERNAL_ERROR ); */
    if (flags & TRANSMIT_TARGET_NODEID)
    {
        uint32 wireNodeID = OALCipher_GetOffsetFromNodeIDAndBase( targetNID,
                                                             self->periodStore[PERIOD_OFFSET( self, CUR )].baseNodeID );

        if (OALMarshal_StreamCompress4( outStream, wireNodeID ) != DOFERROR_OK)
            goto errorOutStreamProblemUnlock;                     /* PCR_Assert( status == DOFERROR_INTERNAL_ERROR ); */
    }
    if (flags & TRANSMIT_BASE_OR_NEXT_NODEID)
    {
        PCR_Assert( OALCipher_IsValid_AdvertisedNNI( self ) );
        if (OALMarshal_StreamCompress4( outStream, self->advNNI ) != DOFERROR_OK)
            goto errorOutStreamProblemUnlock;                     /* PCR_Assert( status == DOFERROR_INTERNAL_ERROR ); */
    }
    a_dlen = OALStream_GetWriteOffset( outStream );                    /*New a_dlen reflects CCM header bytes we added*/
    /* outStream's (unencrypted) header is filled. Now encipher. */
    PCR_Assert( self->normalSecurityMode );
    cipherStrength = DOFSecurityModeCCM_GetStrength( self->normalSecurityMode );
    PCR_Assert( cipherStrength == DOFCIPHERSTRENGTH_256 );
    {
        uint8          mac[MACSIZE_16];
        NonceState_t   nonceState;
        AESCCM_Nonce_t nonce;
        boolean        encrypt = DOFSecurityModeCCM_IsEncrypt( self->normalSecurityMode );
        uint8 *        data    = OALStream_GetWritePtr( outStream ); /*.. _GetReadPtr() would return const pointer*/

        PCR_Assert( a_dlen );                            /* Every OAL secure packet has A data. */
        PCR_Assert( a_dlen == (uint32)(data - a_data) ); /* a_data/data are contiguous (incoming streams are not).*/
        /* Copy payload to encrypt to outStream. */
        if (!OALStream_Write( outStream, dlen, OALStream_GetReadPtr( *pPayloadStream ) ))
        {
            status = DOFERROR_INSUFFICIENT_RESOURCES;
            goto errorOutStreamProblemUnlock;
        }
        /* We are done with *pPayloadStream now. If AESCCM_Encipher() succeeds we will release *pPayloadStream. */
        nonceState.nodeID = self->nodeID;
        nonceState.slot = txSlot;
        nonceState.packetNumber = self->encipherPacketNumber;
        OALCipher_FillNonce( nonceState, &nonce );             /* Depends on nonceState passing in by copy. */

        if (dlen)
        {
            if (dlen < payloadEncryptOffset)
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG,
                                    "OALCipher: ERROR: Encipher wrongly asked to encrypt beyond dlen." );
                status = DOFERROR_ILLEGAL_ARGUMENT;
                goto errorOutUnlock;
            }
            dlen -= payloadEncryptOffset;
            data += payloadEncryptOffset;
            a_dlen += payloadEncryptOffset;
        }
        if (!encrypt)
        {
            a_dlen += dlen;                                        /* All is ADATA. We have no PDATA. */
            dlen = 0;                                              /* pointer data is now don't care. */
        }
        /* Note: Differing mlen values do not influence the encrypted payload output at "data".
                 Differing mlen values do radically change computed mac, because macAuthFlags is different for each. */
        /* We keep mutex locked until _Encipher is complete- we do not want self->current changing during encipher*/
        status = AESCCM_Encipher( cipherStrength, self->periodStore[PERIOD_OFFSET( self, CUR )].dofKey.ekey,
                                  DOFSecurityModeCCM_GetMacAuthFlags( self->normalSecurityMode ),
                                  &nonce, a_dlen, a_data, dlen, data, mlen, mac, encrypt);
        PCR_Assert( status == DOFERROR_OK );
        {
            if (!OALStream_Write( outStream, mlen, mac ))
            {
                status = DOFERROR_INSUFFICIENT_RESOURCES;
                goto errorOutStreamProblemUnlock;
            }
            OALStream_Release( *pPayloadStream, TRUE );            /* Replace *pPayloadStream with our outStream. */
            *pPayloadStream = outStream;
            self->encipherPacketNumber++;      /* Encryption succeeded, so post-increment ->encipherPacketNumber. */
            self->lastSlotWithTX = nonceState.slot;
        }
        OALCipher_Unlock( self );
        return DOFERROR_OK;
    }
errorOutStreamProblemUnlock:
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG,
                        "OALCipher: ERROR: Encipher failed to compress a value or failed to write to a stream." );
errorOutUnlock:
    OALCipher_Unlock( self );
errorOut:
    OALStream_Release( outStream, TRUE );

    return status;
}


/* @internal
    Decrypt text block of an EPP PDU stream with a given cipher instance.

    @note After successful decipher, the succeeding decipher nonce is stored to state.
    @note The stream at pFullDecipherStream holds full data of the EPP PDU to be decrypted, but is handed to us with its
          read offset at the first byte of CCM security mode header. offsetBackToBeginEPP takes from that point in the
          stream back to the EPP PDU header's first byte. From that point forward the MAC is computed.
    @param self                   The given cipher instance. This must not be NULL.
    @param pFullDecipherStream    Pointer to the given full decipher stream. This must not be NULL, and its contents must not be NULL.
    @param offsetBackToBeginEPP   Read offset at the beginning of the EPP headers within @a pFullDecipherStream.
    @param payloadEncryptOffset   Offset to 1st payload byte that is encrypted (0 for all of text block encrypted).
    @param pSenderAddress         Pointer to sender DOFAddress. Its contents are used to record senderNodeID in our Address table. This must not be NULL, and its contents must not be NULL.
    @param pUnicastBehaviorMarkedBySender Ptr to boolean we fill as 1/0 for targetNodeID found/not. This must not be NULL.
    @param pDecipherKeyType       Pointer to DecipherKeyType_t we fill on successful decipher ONLY. This must not be NULL.
    @return DOFERROR_OK or DOFERROR_OK_NODEID_COLLISION for successful decipher, or
              - DOFERROR_TARGETNID_FOUND_BUT_NOT_SECUREGROUP if packet supplies targetNodeID when not secureGroup, or
              - DOFERROR_NON_SGM_SENT_NEXTNODEID if packet not sent by manager but supplied an advNNI, or
              - DOFERROR_SPECIFIED_DECRYPT_KEY_NOT_IN_HAND if we do not have the stream-identified key, or
              - DOFERROR_PARSE_ERROR if unable to read the given stream, or
              - DOFERROR_REJECT_SLOT if cipher has pruned state, or is blocking unknown nodes, for the incoming slot, or
              - DOFERROR_REJECT_INCOMING_PACKET_NUMBER if stream's packet number is not above state packet number, or
              - DOFERROR_RECEIVED_NODE_IDENTIFIER_NOT_US if stream's specified node ID is not us, or
              - DOFERROR_ILLEGAL_ARGUMENT if payloadEncryptOffset is too large for supplied input stream, or
              - DOFERROR_DECRYPTION_FAILED if all other conditions were fine but the decryption simply failed.
*/
DOFError OALCipher_Decipher( const OALCipher self,
                             OALStream * pFullDecipherStream, /* stream of some early beginning all the way to MAC end*/
                             uint32 offsetBackToBeginEPP,     /* Takes us back to the unencrypted EPPv2 header byte. */
                             uint8 payloadEncryptOffset, DOFAddress * pSenderAddress,
                             boolean * pUnicastBehaviorMarkedBySender, DecipherKeyType_t * pDecipherKeyType,
                             OALCipherKeyState * pCipherKeyState )
{
    uint32                hisNodeID;
    uint32                advNNI;
    uint32                timeMarkForSync = PCRTime_GetMS();
    DOFError              status          = DOFERROR_DECIPHER_UNMARSHAL_ERROR;/*Not overwritten only for extreme error*/
    DOFSecurityMode       securityMode;
    DOFCipherStrength     cipherStrength;
    NonceState_t          nonceState;
    uint16                hisSlot         = 0;
    uint8                 flags;
    uint8                 period;
    uint8                 periodOffset;
    boolean               targetNIDOnWire;
    boolean               advNNIOnWire;
    boolean               implicitCount;
    boolean               nonceStateFilledFromStorage;
    boolean               cipherIsSecureGroup;
    boolean               managerPDU;

    PCR_Assert( self );
    if (!OALStream_Read8( *pFullDecipherStream, &flags ))
        goto out;                                       /* PCR_Assert( status == DOFERROR_DECIPHER_UNMARSHAL_ERROR ); */
    targetNIDOnWire = (flags & TRANSMIT_TARGET_NODEID) != 0;
    advNNIOnWire = (flags & TRANSMIT_BASE_OR_NEXT_NODEID) != 0;
    implicitCount = (flags & TRANSMIT_COUNT) == 0;
    cipherIsSecureGroup = OALCipher_IsSecureGroup( self );
    managerPDU = (flags & MANAGER_TRANSMITTED_PDU) != 0;
    if (targetNIDOnWire && !cipherIsSecureGroup)
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_TRACE, "OALCipher: ERROR: _Decipher reject, not secureGroup, targetNID found.");
        return DOFERROR_TARGETNID_FOUND_BUT_NOT_SECUREGROUP;
    }
    if (advNNIOnWire && ( !managerPDU || !cipherIsSecureGroup ) )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "OALCipher: ERROR: _Decipher reject, non SGM sent us advNNI." );
        return DOFERROR_NON_SGM_SENT_NEXTNODEID;
    }
    PCR_Assert( pUnicastBehaviorMarkedBySender );
    /* Note: This next code line doubles up use of the return boolean pointer pUnicastBehaviorMarkedBySender:
                 FALSE means PDU received as being marked "multicast" (we return FALSE for secureGroup connection ONLY).
                 TRUE means PDU received as being marked "unicast" on secureGroup connection, or
                      connection is connectionless unicast, or
                      connection is streaming. */
    *pUnicastBehaviorMarkedBySender = (cipherIsSecureGroup ? targetNIDOnWire : TRUE);
    period = (flags >> 4) & 0x07;
    OALCipher_AdjustKeys( self );
    OALCipher_Lock( self );                                          /* Lock now that we do work with keys. */
    periodOffset = OALCipher_GetPeriodOffset( self, period );
    if (periodOffset >= PERIODCOUNT || !self->periodStore[periodOffset].enabled)
    {
        PCR_Assert( periodOffset == PERIOD_DISABLED ||
                    (periodOffset < PERIODCOUNT && !self->periodStore[periodOffset].enabled) );
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "_Decipher() reject, incoming period not available. Period: 0x",
                                                     period );
        status = DOFERROR_SPECIFIED_DECRYPT_KEY_NOT_IN_HAND;
        goto unlockMutexAndOut;
    }
    if (!self->flags.implicitNodeID)                                 /* Determine hisNodeID. */
    {
        if (OALMarshal_StreamUncompress4( *pFullDecipherStream, &hisNodeID ) != DOFERROR_OK)
            goto unlockMutexAndOut;                     /* PCR_Assert( status == DOFERROR_DECIPHER_UNMARSHAL_ERROR ); */
        hisNodeID = OALCipher_GetNodeIDFromOffsetAndBase( hisNodeID, self->periodStore[periodOffset].baseNodeID );
    }
    else
        hisNodeID = self->otherNode_nodeID;   /* If we are manager, we gave this to him; else, he told us his nodeID. */
    if (!self->flags.ignoreSlot)                                     /* Determine hisSlot. */
    {
        if (OALMarshal_StreamUncompress2( *pFullDecipherStream, &hisSlot ) != DOFERROR_OK)
            goto unlockMutexAndOut;                     /* PCR_Assert( status == DOFERROR_DECIPHER_UNMARSHAL_ERROR ); */
    }
    /* Deny accepting replayed incoming packets within hisSlot. We choose to allow a nodeID to have independant state
       entries for various incoming slot values, so accepting slot 2 is OK even though slot 3 state is already held.
       This still denies accepting replayed packets. */
    if (hisSlot < self->periodStore[periodOffset].lowAcceptedSlot)       /* Has hisSlot been pruned? */
    {
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "_Decipher() reject, incoming slot previously pruned. hisSlot 0x",
                                                     (uint32)hisSlot );
        status = DOFERROR_REJECT_SLOT;
        goto unlockMutexAndOut;
    }
    /* Lookup last received nonce state for hisNodeID/hisSlot. */
    nonceStateFilledFromStorage = OALCipher_FillNonceStateFromStorage( self, periodOffset, hisSlot, hisNodeID,
                                                                       &nonceState );
    if (!nonceStateFilledFromStorage)
    {
        PCRMem_Set( (uint8 *)(&nonceState), 0, sizeof( NonceState_t ) ); /* Build up nonceState from scratch. */
        nonceState.nodeID = hisNodeID;                                   /* Any hisSlot/hisSentPacketNumber acceptable*/
        if (implicitCount && IsStreamingConnection( self ))
            nonceState.packetNumber = 1;                                 /* 1st rcvd PDU must be pn 1. */
        nonceState.slot = hisSlot;
    }
    PCR_Assert( nonceState.nodeID == hisNodeID );
    PCR_Assert( nonceState.nodeID <= TOP_NODEID );
    PCR_Assert( !self->flags.ignoreSlot || ( hisSlot == 0 ) );
    PCR_Assert( nonceState.slot == hisSlot );
    PCR_Assert( nonceState.slot < 0x8000 );
    if (!implicitCount)
    {
        uint32 hisSentPacketNumber;                                 /* Only the success path guarantees to fill this. */

        if (OALMarshal_StreamUncompress4( *pFullDecipherStream, &hisSentPacketNumber ) != DOFERROR_OK)
            goto unlockMutexAndOut;                     /* PCR_Assert( status == DOFERROR_DECIPHER_UNMARSHAL_ERROR ); */
        PCR_Assert( hisSentPacketNumber < 0x40000000 );
        if (nonceStateFilledFromStorage && nonceState.packetNumber >= hisSentPacketNumber)
        {
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            {
                OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "_Decipher() reject, incoming pn violation. His send PN 0x",
                                                         (uint32)hisSentPacketNumber );
                OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                                      "_Decipher() reject, incoming pn violation. His associated send slot 0x",
                                      (uint32)hisSlot );
                OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                                      "_Decipher() reject, incoming pn violation. His associated send nodeID 0x",
                                      (uint32)hisNodeID );
            }
            status = DOFERROR_REJECT_INCOMING_PACKET_NUMBER;         /* hisSentPacketNumber may be a replay. */
            goto unlockMutexAndOut;
        }
        nonceState.packetNumber = hisSentPacketNumber;
    }
    else if (nonceStateFilledFromStorage)
        nonceState.packetNumber++;                                   /* For decrypting this next packet in line. */
    if (targetNIDOnWire)
    {
        uint32 targetNID;
        if (OALMarshal_StreamUncompress4( *pFullDecipherStream, &targetNID ) != DOFERROR_OK)
            goto unlockMutexAndOut;                     /* PCR_Assert( status == DOFERROR_DECIPHER_UNMARSHAL_ERROR ); */
        targetNID = OALCipher_GetNodeIDFromOffsetAndBase( targetNID, self->periodStore[periodOffset].baseNodeID );
        if (targetNID != self->nodeID)
        {
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                                      "_Decipher() reject, received targetNodeID not us. The targetNodeID: 0x",
                                      (uint32)targetNID );
            status = DOFERROR_RECEIVED_NODE_IDENTIFIER_NOT_US;
            goto unlockMutexAndOut;
        }
    }
    if (advNNIOnWire && OALMarshal_StreamUncompress4( *pFullDecipherStream, &advNNI ) != DOFERROR_OK)
        goto unlockMutexAndOut;                     /* PCR_Assert( status == DOFERROR_DECIPHER_UNMARSHAL_ERROR ); */
    /* advNNI is filled if it was on the wire, and will be accepted down below if decryption succeeds. */
    /* nonceState is filled and *pFullDecipherStream is parsed to beginning of encrypted text, so decrypt now. */
    securityMode = OALCipher_TaggedKey_PeriodOffsetMatches( self, KEKKEY, periodOffset ) ?
                      self->kekKeySecurityMode : self->normalSecurityMode; /* Normal unless incoming uses our kekKey. */
    PCR_Assert( securityMode );
    cipherStrength = DOFSecurityModeCCM_GetStrength( securityMode );
    PCR_Assert( cipherStrength == DOFCIPHERSTRENGTH_256 );
    {
        AESCCM_Nonce_t nonce;
        uint8 *        data;
        uint8 *        a_data;
        uint32         a_dlen;
        uint8          mac[MACSIZE_16];
        uint32         dlen_hold;
        boolean        decrypt    = DOFSecurityModeCCM_IsEncrypt( securityMode );
        uint8          mlen       = DOFSecurityModeCCM_GetMACSize( securityMode );
        uint32         dlen       = OALStream_GetDataSize( (OALStream)(*pFullDecipherStream) ); /*text data + mac data*/
        uint32         readOffset = OALStream_GetReadOffset( (OALStream)(*pFullDecipherStream) );

        PCR_Assert( readOffset > offsetBackToBeginEPP );
        a_dlen = readOffset - offsetBackToBeginEPP;
        PCR_Assert( dlen >= mlen );
        OALStream_AdjustReadOffset( *pFullDecipherStream, dlen - mlen );                          /* To start of mac. */
        if (!OALStream_Read( *pFullDecipherStream, mlen, mac ))
            goto unlockMutexAndOut;                     /* PCR_Assert( status == DOFERROR_DECIPHER_UNMARSHAL_ERROR ); */
        dlen -= mlen;                                   /* mac is filled and dlen is text only FROM NOW ON. */
        /* This manipulation lets us decrypt in place, bypassing _GetReadPtr(), which returns a const pointer. */
        OALStream_SetReadOffset( *pFullDecipherStream, readOffset );   /* Since readOffset must be <= writeOffset. */
        OALStream_SetWriteOffset( *pFullDecipherStream, readOffset );
        data = OALStream_GetWritePtr( *pFullDecipherStream );          /* Here we will decrypt in place. */
        a_data = data - readOffset + offsetBackToBeginEPP;

        if (dlen)
        {
            if (dlen < payloadEncryptOffset)
                goto unlockMutexAndOut;                 /* PCR_Assert( status == DOFERROR_DECIPHER_UNMARSHAL_ERROR ); */
            dlen -= payloadEncryptOffset;
            data += payloadEncryptOffset;
            a_dlen += payloadEncryptOffset;
        }
        dlen_hold = dlen;
        if (!decrypt)
        {
            a_dlen += dlen;                                            /* All is ADATA. We have no PDATA. */
            dlen = 0;                                                  /* pointer data is now don't care. */
        }

        OALCipher_FillNonce( nonceState, &nonce );                     /* Depends on nonceState passing in by copy. */
        PCR_Assert( self->periodStore[periodOffset].enabled );
        status = AESCCM_Decipher( cipherStrength, self->periodStore[periodOffset].dofKey.ekey,
                                  DOFSecurityModeCCM_GetMacAuthFlags( securityMode ),
                                  (const AESCCM_Nonce_t *)(&nonce), a_dlen, a_data, dlen, data, mlen, mac, decrypt );
        PCR_Assert( status == DOFERROR_OK || status == DOFERROR_DECRYPTION_FAILED );
        if (status == DOFERROR_OK)
        {
            /* Accept advNNI if we are not a manager (otherwise ignore it). */
            if (advNNIOnWire && !self->flags.manager)
            {
                self->advNNI = (advNNI <= TOP_NODEID) ? advNNI : UNKNOWN_ADVERTISED_NEXT_NODEID;
            }
            /* Back to decipher work. */
            OALStream_AdjustWriteOffset( *pFullDecipherStream, dlen_hold );  /* To mac begin: caller ignores mac. */
            PCR_Assert( hisNodeID == nonceState.nodeID );
            PCR_Assert( self->periodStore[periodOffset].enabled );
            PCR_Assert( pCipherKeyState );
            *pCipherKeyState = &self->periodStore[periodOffset];
            if (OALCipher_TaggedKey_PeriodOffsetMatches( self, CUR, periodOffset ))
                *pDecipherKeyType = DECIPHERED_BY_CURKEY;
            else if (OALCipher_TaggedKey_PeriodOffsetMatches( self, SLOT, periodOffset ) ||
                     OALCipher_TaggedKey_PeriodOffsetMatches( self, KEKKEY, periodOffset ))
                *pDecipherKeyType = DECIPHERED_BY_FUTKEY;
            else
            {
                *pDecipherKeyType = DECIPHERED_BY_OLDKEY;
                PCR_Assert( OALCipher_KeyIsOldKey( self, periodOffset ) );
            }
            if (!self->flags.ignoreSlot)
            {
                /* This means we are cipher to a secureGroup channel. */
                if (!self->flags.manager)           /* Decide whether and how to adjust my slot to incoming slot. */
                {
                    /* As regular member (not manager), only adjust slot to incoming PDUs from the manager that affect the current key. */
                    if (managerPDU && PERIOD_OFFSET_ENABLED( self, CUR ) &&
                        PERIOD_OFFSET( self, CUR ) == periodOffset)
                    {
                        /* Sync our curKey slot to manager slot. If we're behind, this makes our slot match his.
                                                                 If we're ahead, deny our slot increase until we match*/
                        self->periodStore[PERIOD_OFFSET( self, CUR )].dofKey.txSlotAdjust =
                            nonceState.slot -     /* A slot cannot have its high bit set, but .txSlotAdjust can. */
                            ((timeMarkForSync - self->periodStore[PERIOD_OFFSET( self, CUR )].dofKey.promoteTime) / 1000);
                    }
                }
                else
                {
                    /* As manager, adjust slot to consensus of members (but don't listen to another manager). */
                    if (*pSenderAddress && !managerPDU)
                        OALCipher_ResolveManagerSlot( self, *pSenderAddress, nonceState.slot );
                }
            }
            while (!OALCipher_UpdateStoredNonceState( self, periodOffset, &nonceState ))
            {
                /* Intentionally do not call PruneToWatermarks()- was called above in OALCipher_AdjustKeys(). Since
                   we still have cipher locked, there is no possibility that something else added slot storage. As
                   watermark pruning has not prevented us from having insufficient memory, use this strategy. */
                if (!OALCipher_PruneOneSlotStrategy( self, GetKeyType( self, periodOffset ), nonceState.slot ))
                {
                    /* If here, we could NOT store state, though we pruned to the greatest extent of our strategy.
                       We accept the packet that we decrypted, anyway. The important thing is that we not accept
                       any more packets in the slot of this packet. It is also possible that the key is disabled,
                       in which case no more packets using that key will be accepted. */
                    /* This assert demonstrates the comment above. */
                    PCR_Assert( !self->periodStore[periodOffset].enabled ||
                                self->periodStore[periodOffset].lowAcceptedSlot > nonceState.slot );
                    /* Nothing else to do. */
                    break;
                }
                /* Go try again to update stored nonce state. Maybe this time it will work. */
            } /* end while(). */
            if (cipherIsSecureGroup)
            {
                /* All multipoint datagram sessions must include an address (guaranteed by the protocol stack). */
                PCR_Assert(*pSenderAddress != NULL);
                {
                    /* Update existing or add new entry to addressToNodeIDTree. The stakes are high: Until we add
                       senderAddress/nodeID, C OAL cannot transmit secure unicast PDUs to that group member. */
                    PCR_Assert( nonceState.nodeID <= TOP_NODEID );
                    if (OALCipher_StoreAddressNodeIDPair(self, pSenderAddress, nonceState.nodeID, FALSE) <= TOP_NODEID)
                    {
                        if (self->connectionChannel &&                               /* Unit test does not have this. */
                            self->connectionChannel->remoteChannelData->flags.SecureSendBlocked &&
                            self->addrWaitingForQueryResponse &&
                            DOFAddress_Compare( self->addrWaitingForQueryResponse, *pSenderAddress ) == 0)
                        {
                            /* ->addrWaitingForQueryResponse set: we know the blockedMessageQueue is waiting for us. */
                            OALMessage msgToChannel = CMT_ProcessBlockedMessage( self->connectionChannel, TRUE );

                            if (msgToChannel)
                                OALRemoteChannelMgr_SendMessageToChannel( msgToChannel );
                            else
                            {
                                OALRemoteChannel_Shutdown(self->connectionChannel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
                                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG,
                                             "OALCipher: Cannot unblock secureGroup message queue from _Decipher()." );
                            }
                            DOFAddress_Destroy( self->addrWaitingForQueryResponse );
                            self->addrWaitingForQueryResponse = NULL;
                        }
                    }
                    else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                        OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "Decipher FAILED to store new DOFAddress of nodeID 0x",
                                                                 (uint32)nonceState.nodeID );
                }

                if (!self->flags.manager && managerPDU)
                {
                    DOFAddress addr = DOFAddress_Copy( *pSenderAddress );
                    if (addr)
                    {
                        DOFAddress_Destroy( self->managerAddress );
                        self->managerAddress = addr;
                        /* Now we know the manager, we can get a NodeID from him (if we don't already have one). */
                        if (self->flags.sendRequestPDUToNextFoundManager)
                            OALCipher_GetNewNodeID( self );
                    }
                }
                if (self->flags.nodeIDValid)
                {
                    uint8 quadrant;
                    uint8 nodeIDRangeViolation;

                    if (self->nodeID == hisNodeID || (nodeIDRangeViolation =
                                                           CheckNodeIDRangeViolation( self, &quadrant )) == NID_INVALID)
                    {
                        /* React to collision or invalid range; quadrant may NOT be filled in this path. */
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                            OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                               "Otherwise valid _Decipher() discovers nodeID collision or range violation on nodeID 0x",
                                                  (uint32)self->nodeID );
                        InvalidateNodeIDAndGetNewNodeID( self );
                        PCR_Assert( status == DOFERROR_OK );
                        /* EPP will still accept the PDU, but block future sends while we resolve this. */
                        status = DOFERROR_OK_NODEID_COLLISION;
                    }
                    else if (nodeIDRangeViolation == NID_WARNING && !self->flags.newNodeIDScheduled)
                    {
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                            OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                                "Otherwise valid _Decipher() discovers range warning and schedules getting new nodeID to replace nodeID 0x",
                                                 (uint32)self->nodeID );
                        ScheduleNewNodeID( self, quadrant );
                    }
                }
            }
        } /* end status OK. */
    }
unlockMutexAndOut:
    OALCipher_Unlock( self );
out:
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
    {
        switch (status)
        {
        case DOFERROR_DECIPHER_UNMARSHAL_ERROR:
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "OALCipher: _Decipher() reject, parse error." );
        break;

        case DOFERROR_DECRYPTION_FAILED:
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "OALCipher: _Decipher() reject by failed MAC." );
        break;

        default:
        {
            PCR_Assert( FALSE );
        }
        case DOFERROR_FAILED:                                                    /* A log message was already output. */
        case DOFERROR_OK_NODEID_COLLISION:                                       /* A log message was already output. */
        case DOFERROR_OK:                                                        /* A log message was not     output. */
        break;
        }
    }
    if (status == DOFERROR_DECIPHER_UNMARSHAL_ERROR)
        status = DOFERROR_PARSE_ERROR;

    return status;
}


DOFPermissionSet OALCipher_GetLocalPerms( OALCipher self, OALCipherKeyState keyState )
{
    DOFPermissionSet pset = NULL;
    PCR_Assert( self );
    if( keyState )
    {
        OALCipher_Lock( self );
        pset = keyState->localPerms;
        OALCipher_Unlock( self );
    }
    return pset;
}


boolean OALCipher_SetLocalPerms( OALCipher self, OALCipherKeyState keyState, DOFPermissionSet permSet )
{
    boolean rval = FALSE;
    PCR_Assert( self );
    PCR_Assert( keyState );
    OALCipher_Lock( self );
    if ( keyState->enabled )
    {
        DOFPermissionSet pset = DOFPermissionSet_Copy( permSet );
        if ( pset )
        {
            DOFPermissionSet_Destroy( keyState->localPerms );
            keyState->localPerms = pset;
            rval = TRUE;
        }
    }
    OALCipher_Unlock( self );
    return rval;
}


DOFPermissionSet OALCipher_GetRemotePerms( OALCipher self, OALCipherKeyState keyState )
{
    DOFPermissionSet pset = NULL;
    PCR_Assert( self );
    if( keyState )
    {
        OALCipher_Lock( self );
        pset = keyState->remotePerms;
        OALCipher_Unlock( self );
    }
    return pset;
}


boolean OALCipher_SetRemotePerms( OALCipher self, OALCipherKeyState keyState, DOFPermissionSet permSet )
{
    boolean rval = FALSE;
    PCR_Assert( self );
    PCR_Assert( keyState );
    OALCipher_Lock( self );
    if ( keyState->enabled )
    {
        DOFPermissionSet pset = DOFPermissionSet_Copy( permSet );
        if ( pset )
        {
            DOFPermissionSet_Destroy( keyState->remotePerms );
            keyState->remotePerms = pset;
            rval = TRUE;
        }
    }
    OALCipher_Unlock( self );
    return rval;
}


OALCipherKeyState OALCipher_GetCurrentKeyState( OALCipher self )
{
    OALCipherKeyState keyState;
    PCR_Assert( self );
    OALCipher_AdjustKeys( self );
    OALCipher_Lock( self );
    keyState = &self->periodStore[PERIOD_OFFSET( self, CUR )];
    OALCipher_Unlock( self );
    return keyState;
}

/**
@}
**/
