/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/address.h>
#include <internal/securityscope.h>
#include <internal/ccmmode_v1.h>
#include <internal/channel.h>
#include <internal/cipher.h>
#include <internal/enqueue.h>
#include <internal/marshal.h>
#include <internal/objectinternal.h>
#include <internal/operationinternal.h>
#include <internal/remotechannelmanager.h>
#include <internal/securegroupinternal.h>
#include <internal/router.h>
#include <internal/error.h>
#include "cipherccmmode_v1.h"


/* Operation functions. */

static DOFOperation Create( DOFOperationType opType, OALChannel channel, DOFAddress address, uint32 timeout,
                            boolean providerOperation )
{
    DOFOperation ret;

    PCR_Assert( opType >= DOFOPERATIONTYPE_CCMMODE_V1_FIRST );
    PCR_Assert( opType < DOFOPERATIONTYPE_CCMMODE_V1_LAST );
    ret = (DOFOperation)DOFMem_Alloc(sizeof(DOFOperation_t));
    if (ret == NULL)
        goto error_out;
    PCRMem_Set( ret, 0, sizeof(DOFOperation_t) );
    ret->ref_count = 1;
    ret->sources = (DOFOperationSource)DOFMem_Alloc( sizeof( DOFOperationSource_t ) );
    if (!ret->sources)
    {
        DOFMem_Free( ret );
        goto error_out;
    }
    PCRMem_Set( ret->sources, 0, sizeof( DOFOperationSource_t ) );
    ret->operation_will_close = OALObservable_Create();
    if (ret->operation_will_close == NULL)
        goto cleanup;
    ret->opID = DOFOperationID_CreateLocal();
    if (ret->opID == NULL)
        goto cleanup;
    ret->sources->channel = OALChannel_Retain( channel );
    if (channel && !ret->sources->channel )
        goto cleanup;
    ret->sources->address = DOFAddress_Copy( address );
    if (address && ret->sources->address == NULL)
        goto cleanup;
    ret->cancelContext = DOFOperationManager_CreateCancelContext( ret );
    if (ret->cancelContext == NULL)
        goto cleanup;
    ret->sources->oal_timeout = timeout;
    ret->operation_type = opType;
    ret->flags.bits.internally_created = (providerOperation!=0);
    return ret;

cleanup:
    DOFAddress_Destroy( ret->sources->address );
    OALChannel_Release( ret->sources->channel );
    DOFMem_Free( ret->sources );
    OALObservable_Destroy( ret->operation_will_close );
    DOFOperationID_Destroy( ret->opID );
    DOFMem_Free( ret );

error_out:
    return NULL;
}

/* internal  create the context for CCMMode that will be stored with the operation. */
static CCMMODE_V1_OperationContext CCMMODE_V1_OperationContext_Create(void)
{
    CCMMODE_V1_OperationContext ret = DOFMem_Alloc(sizeof(CCMMODE_V1_OperationContext_t));
    if (ret)
    {
        PCRMem_Set(ret, 0, sizeof(CCMMODE_V1_OperationContext_t));
    }

    return ret;
}

/* internal  destroy the context from the operation was created. */
static void CCMMODE_V1_OperationContext_Destroy(CCMMODE_V1_OperationContext self)
{
    DOFMem_Free( self );
}


DOFOperation DOFOperation_Create_CCMMODE_Query(OALChannel channel, DOFAddress address,
                                              uint32 timeout, boolean providerOperation)
{
    DOFOperation ret = Create(DOFOPERATIONTYPE_CCMMODE_V1_QUERY, channel, address, timeout, providerOperation );
    if (ret)
    {
        CCMMODE_V1_OperationContext context = CCMMODE_V1_OperationContext_Create();
        DOFSecurityScope            scope   = DOFSecurityScope_Copy( DOF_GetDefaultSecurityScope(DOF_GetDof()) );

        /* Scope should never be null for a secure group channel. */
        if (context && scope)
        {
            ret->securityScope = scope;
            ret->context = context;
        }
        else
        {
            CCMMODE_V1_OperationContext_Destroy( context );
            DOFSecurityScope_Destroy( scope );
            DOFOperation_Destroy(ret);
            ret = NULL;
        }
    }
    return ret;
}

DOFOperation DOFOperation_Create_CCMMODE_Request(OALChannel channel, DOFAddress address, uint32 timeout,
                                                 boolean providerOperation)
{
    DOFOperation ret = Create( DOFOPERATIONTYPE_CCMMODE_V1_REQUEST, channel, address, timeout, providerOperation );
    if (ret)
    {
        CCMMODE_V1_OperationContext context = CCMMODE_V1_OperationContext_Create();
        DOFSecurityScope            scope   = DOFSecurityScope_Copy( DOF_GetDefaultSecurityScope(DOF_GetDof()) );

        /* Scope should never be null for a secure group channel. */
        if (context && scope)
        {
            ret->securityScope = scope;
            ret->context = context;
        }
        else
        {
            CCMMODE_V1_OperationContext_Destroy( context );
            DOFSecurityScope_Destroy( scope );
            DOFOperation_Destroy(ret);
            ret = NULL;
        }
    }

    return ret;
}

/* CCM mode state machine functions */
#define DEFAULT_CCM_OP_TIMEOUT 16000 /* milliseconds */

static OALEnqueue CCMMODE_V1_queue;

typedef void (*CCMMODE_SM_Handler)(DOFOperation op);

boolean CCMMODE_V1_SM_Init(void)
{
    CCMMODE_V1_queue = OALEnqueue_Create(DOF_GetDof()->threadPool);
    if (CCMMODE_V1_queue)
    {
        return TRUE;
    }
    return FALSE;
}

void CCMMODE_V1_SM_Shutdown(void)
{
    OALEnqueue_Destroy(CCMMODE_V1_queue);
    CCMMODE_V1_queue = NULL;
}

CCMMODE_V1 CCMMODE_V1_Create(BlackboardData blackboardData, boolean isConnectionLess)
{
    CCMMODE_V1 ret;

    PCR_Assert(blackboardData != NULL);

    ret = (CCMMODE_V1) DOFMem_Alloc(sizeof(CCMMODE_V1_t));

    if(ret != NULL)
    {
        ret->blackboardData       = blackboardData;
        ret->isConnectionLess     = isConnectionLess;
    }
    return ret;
}

void CCMMODE_V1_Destroy(CCMMODE_V1 self)
{
    DOFMem_Free(self);
}

static void CCMMODE_V1_OpClosing(OALObserver self, OALObservable observable, void* ctx)
{
    DOFOperation op = (DOFOperation)ctx;
    /* Called with operation mutex locked. */
    if (op)
    {
        CCMMODE_V1_OperationContext ccmModeCtx = (CCMMODE_V1_OperationContext)op->context;
        if (ccmModeCtx)
        {
            CCMMODE_V1_OperationContext_Destroy(ccmModeCtx);
            op->context = NULL;
        }
        DOFOperation_Internal_Destroy(op);
    }
}

static void CCMMODE_V1_HandleIllegal(DOFOperation op)
{
}

static void CCMMODE_V1_HandleQuery_Initiator(DOFOperation op)
{
    /* Any secure packet suffices as a "response" to the query, since it will contain the information we need (NodeID). */
    /* Nothing more to do. */
}

static void CCMMODE_V1_HandleQuery_Responder(DOFOperation op)
{
    /*
     * If a PDU is received but channel does not have a cipher yet, it is not qualified to process ccmMode packets,
     * since the CCM.Query PDU is required to be sent secure.*
     */
    if ( op->sources->channel->remoteChannelData->cipher )    /* op holds info about an incoming CCM PDU. */
        OALCipher_QueryReceiveHandler(op->sources->channel->remoteChannelData->cipher, op->sources->address);
}

static void CCMMODE_V1_HandleRequest_Initiator( DOFOperation op )
{
    /* We have an incoming CCM.Request response. */
    PCR_Assert( op );
    if (op->sources->channel->remoteChannelData->cipher)
    {
        (void) OALCipher_RequestResponseReceiveHandler( op->sources->channel->remoteChannelData->cipher,
                                                       ((CCMMODE_V1_OperationContext)op->context)->nodeID );
        return;
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "CCMMODE V1: CCM.RequestResponse PDU received without cipher state." );
}

static void CCMMODE_V1_HandleRequest_Responder( DOFOperation op )
{
    /* We have an incoming CCM.Request command. */
    PCR_Assert( op );
    if (op->sources->channel->remoteChannelData->cipher) /*Cannot call OALCipher_RequestReceiveHandler w/ NULL cipher*/
    {
        (void) OALCipher_RequestReceiveHandler( op->sources->channel->remoteChannelData->cipher, op );
    }
}

typedef struct OperationNode_t
{
    OALEnqueueItem_t super;

    DOFOperation     op;
} OperationNode_t;

static void CCMMODE_V1_queueFn (OALThreadPoolItem dnode, boolean shutdown)
{
    PCR_Assert(dnode != NULL);
    {
        OperationNode_t * node = (OperationNode_t*) dnode;

        static const CCMMODE_SM_Handler functions[CCMMODE_V1_HIGHEST_OPCODE + 1][2] =
        {
            /* Note: This list is in operation type order. */
            {CCMMODE_V1_HandleIllegal,                          CCMMODE_V1_HandleIllegal},
            {CCMMODE_V1_HandleQuery_Initiator,                  CCMMODE_V1_HandleQuery_Responder},
            {CCMMODE_V1_HandleRequest_Initiator,                CCMMODE_V1_HandleRequest_Responder},
        };

        PCR_Assert(node->op);
        DOFOperation_Lock(node->op);
        PCR_Assert( node->op->operation_type >= DOFOPERATIONTYPE_CCMMODE_V1_FIRST );
        PCR_Assert( node->op->operation_type < DOFOPERATIONTYPE_CCMMODE_V1_LAST );

        functions[node->op->operation_type - DOFOPERATIONTYPE_CCMMODE_V1_FIRST]
                    [node->op->flags.bits.internally_created](node->op);
        DOFOperation_Unlock(node->op);
        DOFOperation_Internal_Destroy(node->op);
    }
    DOFMem_Free(dnode);
}

static void CCMMODE_V1_SM_ProcessToResponder(DOFOperation op)
{
    OperationNode_t * node;

    PCR_Assert(op);
    node = (OperationNode_t*) DOFMem_Alloc(sizeof(OperationNode_t));
    if(node != NULL)
    {
        OALEnqueueItem_Init((OALEnqueueItem) node, CCMMODE_V1_queueFn);
        node->op = op;
        DOFOperation_Retain( op );
        PCR_Assert( CCMMODE_V1_queue );
        if(OALEnqueue_Enqueue(CCMMODE_V1_queue, (OALEnqueueItem) node))
            return;
        DOFOperation_Internal_Destroy( op );
        DOFMem_Free(node);
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "CCMMODE V1: Insufficient resources to enqueue processing for CCM PDU." );
}

/* State machine Send functions:  The following functions are used to send data to a remote node. */

static DOFError CCMMODE_V1_RouteToProvider(DOFOperation op, OALChannel channel)
{
    /* This takes ownership of the operation (for convenience)! */
    if ( op )
    {
        /* DOFRouterStatusCode_t routerStatus; */

        /* add an observer */
        if (!op->operation_will_close)
            op->operation_will_close = OALObservable_Create();

        DOFOperation_AddOpWillCloseObserver(op, CCMMODE_V1_OpClosing);
        /* The proper error checks should be added here. */
        /* routerStatus = */ OALOperationManager_AddOperation(op, op->opID, FALSE); /* don't need op lock because nobody else knows about this op */

        channel->routeToProvider(channel, op, NULL, op->sources->address);
        DOFOperation_Internal_Destroy(op); /* Destroy the operation (we no longer need it once it is sent). */

        return DOFERROR_OK;
    }
    return DOFERROR_INTERNAL_ERROR;
}


DOFError CCMMODE_V1_SendQuery(OALChannel channel, DOFAddress address)
{
    DOFOperation op;

    op = DOFOperation_Create_CCMMODE_Query(channel, address, DEFAULT_CCM_OP_TIMEOUT, FALSE);
    op->flags.bits.denyPacketReplayProtectionIfSecureUnicastSend = 1; /* CCM.Query always sends unicast unmarked. */
    return CCMMODE_V1_RouteToProvider(op, channel);
}

/* internal

    This routine is used to send a Request PDU multicast and over the default remote connectionless channel.
*/
DOFError CCMMODE_V1_SendRequest( OALChannel channel, DOFAddress address )
{
    DOFOperation op = DOFOperation_Create_CCMMODE_Request( channel, address, DEFAULT_CCM_OP_TIMEOUT, FALSE ); /* FALSE: not a response.*/
    /* Note: Is intended that this be sent to an SGM that we know at address- the PDU goes out unicast marked.
             If address is NULL, we don't know the SGM's DOFAddress, the PDU goes out multicast unmarked. */
    return CCMMODE_V1_RouteToProvider( op, channel );
}

/* Always sends unicast as a response from the CCM.Request request. The latter is intended to come in unicast. */
DOFError CCMMODE_V1_SendRequestResponse( DOFOperation requestOp )
{
    PCR_Assert( requestOp );
    DOFOperation_Lock( requestOp );
    PCR_Assert( requestOp->context );
    requestOp->sources->channel->routeToRequestor( requestOp->sources->channel, requestOp, NULL );
    DOFOperation_Unlock( requestOp );
    return DOFERROR_OK;
}


/* Send functions */
static void CCMMODE_V1_Send_Init(CCMMODE_V1_SendData data, CCMMODE_V1 ccmMode, DOFOperation operation)
{
    PCRMem_Set(data, 0, sizeof(CCMMODE_V1_SendData_t));
    data->ccmMode = ccmMode;
    data->operation = operation;
    data->packet = ccmMode->blackboardData->sendPacket;
}

static boolean CCMMODE_V1_Send_WriteFirstByte( CCMMODE_V1_SendData data )
{
    /* We write CCM's two byte APPID 0x6001, then the one byte opcode. The latter is the "FirstByte." */
    OALStream stream = OALStream_CreateRW(3);

    if(stream)
    {
        uint32 size = OALMarshal_CompressValue(2, CCMMODE_V1_APPID, 2, OALStream_GetWritePtr(stream));
        if ((size > 0) && (size < 3))   /* 1 or 2 bytes */
        {
            OALStream_AdjustWriteOffset(stream, size);
        }
        else
        {
            data->status = CCMMODE_V1_SendStatus_MEMORYERROR;
            OALStream_Release( stream, TRUE );
            return FALSE;
        }

        if(OALStream_Write8(stream, data->opCode))
        {
            if(!OALPacket_AppendStream(data->packet, stream))
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "CCMMODE V1: send, write first byte: Failed to append to stream.");
                data->status = CCMMODE_V1_SendStatus_MEMORYERROR;
                OALStream_Release( stream, TRUE );
                return FALSE;
            }
            return TRUE;
        }
        data->status = CCMMODE_V1_SendStatus_MEMORYERROR;
        OALStream_Release( stream, TRUE );
        return FALSE;
    }
    data->status = CCMMODE_V1_SendStatus_STREAMERROR;
    return FALSE;
}


static boolean CCMMODE_V1_Send_WriteNodeID( CCMMODE_V1_SendData data )
{
    /* Append the size of the initial state as a compressed short and then append the initial state. */
    uint16    nodeIDSize;
    OALStream stream;
    boolean   success = FALSE;

    PCR_Assert( data->context->nodeID <= TOP_NODEID );
    nodeIDSize = OALMarshal_GetCompressedValueSize( 4, data->context->nodeID );
    PCR_Assert( nodeIDSize <= 4 );
    if ((stream = OALStream_CreateRW( nodeIDSize )) != NULL)
    {
        nodeIDSize = OALMarshal_CompressValue( 4, data->context->nodeID, nodeIDSize, OALStream_GetWritePtr( stream ) );
        PCR_Assert( nodeIDSize <= 4 );
        OALStream_AdjustWriteOffset( stream, nodeIDSize );
        if (OALPacket_AppendStream( data->packet, stream ))
            success = TRUE;                             /* stream gets released when data->packet gets destroyed. */
        else
        {
            data->status = CCMMODE_V1_SendStatus_STREAMERROR;
            OALStream_Release( stream, TRUE );
        }
    }
    else
        data->status = CCMMODE_V1_SendStatus_MEMORYERROR;

    return success;
}


static boolean CCMMODE_V1_Send_IterateFunctions(CCMMODE_V1_SendData data)
{
    uint8 i = 0;

    while(data->functions[i] != NULL)
    {
        if(!data->functions[i++](data))
        {
            return FALSE;
        }
    }
    return TRUE;
}


static void CCMMODE_V1_Query_Send(CCMMODE_V1_SendData data, BlackboardData blackboardData)
{
    static const CCMMODE_V1_SendUtils functions[] = {CCMMODE_V1_Send_WriteFirstByte, NULL};

    data->functions = functions;
    data->opCode = CCMMODE_V1_QUERY_OPCODE;
    blackboardData->flags.securityRequiredEncryptOrAuthenticate = TRUE;
    PCR_Assert( !blackboardData->flags.securityRequiredUnsecure);
    blackboardData->flags.streamingProtocolNotAllowed = TRUE;

    CCMMODE_V1_Send_IterateFunctions(data);
    /* REVIEW: Spec says multicast/broadcast not allowed. Do we do enough to deny that? */
}

static void CCMMODE_V1_Request_Send(CCMMODE_V1_SendData data, BlackboardData blackboardData)
{
    static const CCMMODE_V1_SendUtils functions[] = {CCMMODE_V1_Send_WriteFirstByte, NULL};

    data->functions = functions;
    data->opCode = CCMMODE_V1_REQUEST_OPCODE;
    blackboardData->flags.securityRequiredUnsecure = TRUE;
    PCR_Assert( !blackboardData->flags.securityRequiredEncryptOrAuthenticate );
    blackboardData->flags.streamingProtocolNotAllowed = TRUE;
    blackboardData->flags.securityRequiredSecureHeaderUnsecure = TRUE;
    /* The above gets the EPP security header on the wire. To use spec language, it "identifies the session." */
    CCMMODE_V1_Send_IterateFunctions(data);
    /* REVIEW: Spec says multicast/broadcast not allowed. Do we do enough to deny that?
               Right now the CCM way to send a CCM.Request is to send it by multicast. */
}


static void CCMMODE_V1_Request_Response_Send( CCMMODE_V1_SendData data, BlackboardData blackboardData )
{
    static const CCMMODE_V1_SendUtils functions[] = {CCMMODE_V1_Send_WriteFirstByte,
                                                     CCMMODE_V1_Send_WriteNodeID, NULL};

    data->functions = functions;
    data->opCode = CCMMODE_V1_REQUEST_OPCODE;
    blackboardData->flags.streamingProtocolNotAllowed = TRUE;
    blackboardData->flags.securityRequiredUnsecure = FALSE;             /* Override this from original request. */
    blackboardData->flags.securityRequiredEncryptOrAuthenticate = TRUE;
    CCMMODE_V1_Send_IterateFunctions( data );
}


static void CCMMODE_V1_InvalidOpcode_Send(CCMMODE_V1_SendData data, BlackboardData blackboardData)
{
    data->status = CCMMODE_V1_SendStatus_INVALID_OPCODE;
}

ProtocolLayerResult CCMMODE_V1_Send(CCMMODE_V1 self, DOFOperation operation, boolean isProviderOperation, DOFResult result)
{
    CCMMODE_V1_SendData_t sendData;

    /* Note: this table is in operation type order. */
    static const CCMMODE_V1_SendFunction functions[DOFOPERATIONTYPE_CCMMODE_V1_LAST -
                                                   DOFOPERATIONTYPE_CCMMODE_V1_FIRST][2] =
                                                    {
                                                        {CCMMODE_V1_InvalidOpcode_Send, CCMMODE_V1_InvalidOpcode_Send},
                                                        {CCMMODE_V1_InvalidOpcode_Send, CCMMODE_V1_Query_Send},
                                                        {CCMMODE_V1_Request_Response_Send, CCMMODE_V1_Request_Send},
                                                    };

    PCR_Assert(self);
    PCR_Assert(operation);
    PCR_Assert(operation->operation_type >= DOFOPERATIONTYPE_CCMMODE_V1_FIRST);
    PCR_Assert(operation->operation_type <  DOFOPERATIONTYPE_CCMMODE_V1_LAST);

    CCMMODE_V1_Send_Init(&sendData, self, operation);

    if (sendData.operation != NULL)
    {
        DOFOperation_Lock(sendData.operation);
        sendData.context = (CCMMODE_V1_OperationContext)sendData.operation->context;
        if (sendData.context)
        {
            /* Call the send function for this PDU. */
            functions[operation->operation_type - DOFOPERATIONTYPE_CCMMODE_V1_FIRST][isProviderOperation]
                                                                                      (&sendData, self->blackboardData);
            sendData.operation->flags.bits.securityRequiredSecureHeaderUnsecure =
                                                       self->blackboardData->flags.securityRequiredSecureHeaderUnsecure;
        }
        else
            sendData.status = CCMMODE_V1_SendStatus_TIMEOUTERROR;

        DOFOperation_Unlock(sendData.operation);
    }
    else
    {
        /* The operation did not exist, set the send status to report the error. */
        sendData.status = CCMMODE_V1_SendStatus_MEMORYERROR;
    }

    self->blackboardData->flags.isCommandNotResponse = (isProviderOperation!=0);
    self->blackboardData->associatedChannel->remoteChannelData->flags.isSSIDAssignedFromAS = TRUE;
    self->blackboardData->flags.ifProperPlaceNextBaseNIDOnWire = TRUE;
    switch(sendData.status)
    {
        case CCMMODE_V1_SendStatus_OK:
            self->blackboardData->opID = DOFOperationID_Copy(sendData.operation->opID);
            if ( self->blackboardData->opID )
            {
                self->blackboardData->delay = sendData.operation->sources->oal_timeout;
                return PLRESULT_OKSENDDOWN;
            }
            /* else fall through */
        default:
            break;
    }
    return PLRESULT_ERROR;
}

static boolean CCMMODE_V1_Receive_VerifyOpID(CCMMODE_V1_ReceiveData data)
{
    if(data->ccmMode->blackboardData->opID == NULL)
    {
        data->status = CCMMODE_V1_ReceiveStatus_STREAMERROR;
        return FALSE;
    }
    return TRUE;
}


static boolean CCMMODE_V1_Receive_RetrieveOperationFromRouter( CCMMODE_V1_ReceiveData data )
{
    data->operation = OALOperationManager_CheckExistingOperation(DOFSecurityScope_GetDomainID(DOF_GetDefaultSecurityScope(DOF_GetDof())), data->ccmMode->blackboardData->opID );
    if(data->operation == NULL)
    {
        data->status = CCMMODE_V1_ReceiveStatus_STREAMERROR;
        return FALSE;
    }
    return TRUE;
}


static boolean CCMMODE_V1_Receive_VerifyNewOpID(CCMMODE_V1_ReceiveData data)
{
    data->operation = OALOperationManager_CheckExistingOperation(DOFSecurityScope_GetDomainID(DOF_GetDefaultSecurityScope(DOF_GetDof())),data->ccmMode->blackboardData->opID);
    if(data->operation == NULL)
    {
        /* Return TRUE if we do not find an existing operation. */
        return TRUE;
    }
    DOFOperation_Internal_Destroy(data->operation);
    data->status = CCMMODE_V1_ReceiveStatus_OK_DONOTHING;
    return FALSE;
}

static void CCMMODE_V1_Receive_InitData(CCMMODE_V1_ReceiveData data, OALStream stream, CCMMODE_V1 ccmMode)
{
    PCRMem_Set(data, 0, sizeof(CCMMODE_V1_ReceiveData_t));

    data->stream = stream;
    data->ccmMode = ccmMode;
}

static void CCMMODE_V1_ReceiveOpClosing(OALObserver self, OALObservable observable, void* ctx)
{
    DOFOperation op = (DOFOperation)ctx;
    /* Called with operation mutex locked. */
    if ((CCMMODE_V1_OperationContext)op->context)
    {
        CCMMODE_V1_OperationContext_Destroy((CCMMODE_V1_OperationContext)op->context);
        op->context = NULL;
    }
    DOFOperation_Internal_Destroy(op);
}

static boolean CCMMODE_V1_Receive_IterateFunctions(CCMMODE_V1_ReceiveData data)
{
    uint8 i = 0;

    while(data->functions[i] != NULL)
    {
        if(!data->functions[i++](data))
            return FALSE;
    }
    return TRUE;
}

static void CCMMODE_V1_SM_PreProcessToResponder(CCMMODE_V1_ReceiveData receiveData)
{
    /* This takes ownership of the operation (for convenience)! */
    if (receiveData->operation)
    {
        DOFOperationID tempID = DOFOperationID_Copy(receiveData->ccmMode->blackboardData->opID);
        if ( tempID )
        {
            DOFOperationID_Destroy(receiveData->operation->opID);
            receiveData->operation->opID = tempID;
            if (!receiveData->operation->operation_will_close)
                receiveData->operation->operation_will_close = OALObservable_Create();
            /* Add a check for errors. */
            DOFOperation_AddOpWillCloseObserver(receiveData->operation, CCMMODE_V1_ReceiveOpClosing);
            OALOperationManager_AddOperation(receiveData->operation, receiveData->operation->opID, FALSE); /* don't need lock because nobody else knows about this op */
            CCMMODE_V1_SM_ProcessToResponder(receiveData->operation);
        }
        else
        {
            receiveData->status = CCMMODE_V1_ReceiveStatus_MEMORYERROR;
        }
        DOFOperation_Internal_Destroy(receiveData->operation); /* Destroy the operation (we no longer need it once it is processed). */
    }
    else
        receiveData->status = CCMMODE_V1_ReceiveStatus_MEMORYERROR;
}


static void CCMMODE_V1_ReceiveQuery(CCMMODE_V1_ReceiveData data)
{
    static const CCMMODE_V1_ReceiveUtils functions[] = {CCMMODE_V1_Receive_VerifyOpID,
                                                        CCMMODE_V1_Receive_VerifyNewOpID, NULL};

    if (data->ccmMode->blackboardData->flags.receivedAuthenticated || data->ccmMode->blackboardData->flags.receivedEncrypted)
    {
        data->functions = functions;
        if (CCMMODE_V1_Receive_IterateFunctions(data))
        {
            data->operation = DOFOperation_Create_CCMMODE_Query(data->ccmMode->blackboardData->associatedChannel,
                                                               data->ccmMode->blackboardData->transportAddr,
                                                               data->ccmMode->blackboardData->delay, TRUE);
            CCMMODE_V1_SM_PreProcessToResponder(data);
        }
    }
    else
    {
        /* If a query is received unsecure ignore it. */
        data->status = CCMMODE_V1_ReceiveStatus_OK_DONOTHING;
    }
}

static OALAVLTreeList_Traverse_All_Return_Type_fn_t CCMMODE_V1_SubmitCCMRequestToEachSecureGroup( OALAVLTreeListItem_t item, void *context, boolean newKey )
{
    OALChannel             secureGroupChannel = (OALChannel)item;
    CCMMODE_V1_ReceiveData data               = (CCMMODE_V1_ReceiveData)context;

    PCR_Assert( secureGroupChannel );
    PCR_Assert( data );
    data->operation = DOFOperation_Create_CCMMODE_Request( secureGroupChannel,
                                                           data->ccmMode->blackboardData->transportAddr,
                                                           data->ccmMode->blackboardData->delay, TRUE ); /*->operation represents incoming CCM.Request*/
    CCMMODE_V1_SM_PreProcessToResponder( data );
    return OALAVLTREELIST_NORMAL;                                                        /* OALAVLTREELIST_NORMAL: traverse full tree. */
}


static void CCMMODE_V1_ReceiveRequest( CCMMODE_V1_ReceiveData data )
{
    static const CCMMODE_V1_ReceiveUtils functions[] = {CCMMODE_V1_Receive_VerifyOpID,           /* opid not null. */
                                                        CCMMODE_V1_Receive_VerifyNewOpID, NULL}; /* opid not in table.*/

    if (data->ccmMode->blackboardData->flags.receivedUnsecure)
    {
        data->functions = functions;
        if (CCMMODE_V1_Receive_IterateFunctions( data ))
        {
            OALAVLTreeList registeredGroups; /* CCM.Request PDU sent only from secureGroup members to secureGroup SGM.*/
            /* Before the CCM.Request operation is created collect a set of secureGroup's. */
            if (data->ccmMode->blackboardData->associatedChannel->remoteChannelData->baseChannel)
                registeredGroups = data->ccmMode->blackboardData->associatedChannel->remoteChannelData->baseChannel->remoteChannelData->registeredSSIDs_AS;
            else
                registeredGroups = data->ccmMode->blackboardData->associatedChannel->remoteChannelData->registeredSSIDs_AS;
            if (!registeredGroups)
            {
                /*  This is either a point connection or there aren't any groups associated with this channel.
                    In either case the CCM.Request PDU can be dropped since there's nobody to ask. */
                data->status = CCMMODE_V1_ReceiveStatus_OK_DONOTHING;
                return;
            }
            OALAVLTreeList_TraverseAll( registeredGroups,
                                        CCMMODE_V1_SubmitCCMRequestToEachSecureGroup,
                                        data );
        }
    }
    else
        data->status = CCMMODE_V1_ReceiveStatus_OK_DONOTHING;   /* If a CCM.Request PDU is received secure ignore it. */
}


static boolean CCMMODE_V1_Receive_ReadNodeID( CCMMODE_V1_ReceiveData data )
{
    uint32 size  = OALStream_GetDataSize( data->stream );
    uint32 value = OALMarshal_UncompressValue( 4, &size, OALStream_GetReadPtr( data->stream ) );

    if ((value == 0xFFFFFFFF) || (size > 4))
    {
        data->status = CCMMODE_V1_ReceiveStatus_STREAMERROR;
        return FALSE;
    }
    OALStream_AdjustReadOffset( data->stream, size );
    data->nodeID = value;

    return TRUE;
}


static void CCMMODE_V1_ReceiveRequestResponse( CCMMODE_V1_ReceiveData data )
{
    static const CCMMODE_V1_ReceiveUtils functions[] = {CCMMODE_V1_Receive_VerifyOpID,                 /*opid not null*/
                                                        CCMMODE_V1_Receive_RetrieveOperationFromRouter,/*opid in table*/
                                                        CCMMODE_V1_Receive_ReadNodeID, NULL};

    if (!data->ccmMode->blackboardData->flags.receivedUnsecure)
    {
        data->functions = functions;
        if (CCMMODE_V1_Receive_IterateFunctions(data))
        {
            if (data->operation && data->operation->context)
            {
                DOFAddress tempAddr = DOFAddress_Copy( data->ccmMode->blackboardData->transportAddr );
                if ( tempAddr )
                {
                    CCMMODE_V1_OperationContext context = (CCMMODE_V1_OperationContext)data->operation->context;
                    context->nodeID = data->nodeID;
                    DOFAddress_Destroy(data->operation->sources->address);
                    data->operation->sources->address = tempAddr;
                    CCMMODE_V1_SM_ProcessToResponder( data->operation );
                    DOFOperation_Internal_Destroy( data->operation );
                    data->operation = NULL;
                    return;
                }
            }
        }
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "CCMMODE V1: Failed to process CCM request response." );
        data->status = CCMMODE_V1_ReceiveStatus_MEMORYERROR;
    }
    else
        data->status = CCMMODE_V1_ReceiveStatus_OK_DONOTHING; /* If a request response is received unsecure ignore it.*/
}


static void CCMMODE_V1_ReceivedReserved(CCMMODE_V1_ReceiveData data)
{
    /* This should not happen, but just exit if it does and it will get dropped */
}

ProtocolLayerResult CCMMODE_V1_Receive(CCMMODE_V1 self)
{
    /* Note: This list is in opcode order. */
    static const CCMMODE_V1_ReceiveFunction functions[CCMMODE_V1_HIGHEST_OPCODE + 1][2] = {
            /* 0 */ {CCMMODE_V1_ReceivedReserved, CCMMODE_V1_ReceivedReserved},
                    {CCMMODE_V1_ReceivedReserved, CCMMODE_V1_ReceiveQuery},
                    {CCMMODE_V1_ReceiveRequestResponse, CCMMODE_V1_ReceiveRequest},
    };

    if (self->blackboardData->receiveStream)
    {
        uint32 size = OALStream_GetDataSize( self->blackboardData->receiveStream );
        uint32 value = OALMarshal_UncompressValue(2, &size, OALStream_GetReadPtr(self->blackboardData->receiveStream));
        CCMMODE_V1_ReceiveData_t receiveData;
        uint8   firstByte;
        if ((value == 0xFFFFFFFF) || (size > 2))
            return PLRESULT_ERROR;
        if (value != CCMMODE_V1_APPID)
            return PLRESULT_ERROR;
        OALStream_AdjustReadOffset( self->blackboardData->receiveStream, size );

        if(!OALStream_Read8(self->blackboardData->receiveStream, &firstByte))
            return PLRESULT_ERROR;
        if(firstByte > CCMMODE_V1_HIGHEST_OPCODE)
            return PLRESULT_ERROR;

        CCMMODE_V1_Receive_InitData(&receiveData, self->blackboardData->receiveStream, self);
        functions[firstByte][self->blackboardData->flags.isCommandNotResponse](&receiveData);
        switch(receiveData.status)
        {
            case CCMMODE_V1_ReceiveStatus_OK:
            case CCMMODE_V1_ReceiveStatus_OK_DONOTHING:
                return PLRESULT_OKDONOTHING;
            default:
                break;
        }
    }
    return PLRESULT_ERROR;
}

