/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/addressinternal.h>
#include <internal/ccmmode_v1.h>
#include <internal/channel_stack_messages.h>
#include <internal/cipher.h>
#include <internal/epp_negative_one.h>
#include <internal/exception.h>
#include <internal/marshal.h>
#include <internal/mutex_pool.h>
#include <internal/oalstring.h>
#include <internal/observable.h>
#include <internal/remotechannel.h>
#include <internal/remotechannelmanager.h>
#include <internal/securegroupinternal.h>
#include <internal/timeinternal.h>
#include "cipherprivate.h"


/** @internal @addtogroup OALCipher
@{
**/


/*
**  "Private" Type and Method Declarations
*/


#define DEEPER_TREE_LOG                     (0)


#define MANAGER_BASE_NODEID_ADVANCE_VALUE   (0x00080000)  /* Implementation choice: 524,288. */
STATIC_ASSERT( MAXIMUM_NEW_ADVERTISED_NEXT_NODEID_OFFSET > MANAGER_BASE_NODEID_ADVANCE_VALUE );


#define GET_NEXT_PERIOD                     (0xFF)
STATIC_ASSERT( GET_NEXT_PERIOD > PERIODCOUNT );


#define NO_ENDING_HELD_STATE_COUNT_MAXIMUM  (0xFFFFFFFF)
#define UNKNOWN_BASE_NODEID                 (0xFFFFFFFF)
#define NEED_NEW_BASE_NODEID_TRIGGER_COUNT  (1000)

#define TOP_LEGAL_SLOT                      (0x7FFF)
#define ORDINARY_PRUNE_CHECK_DELAY          (1000) /* 1 second. */


#define EXCESSIVE_SLOT_TRIGGER              (0x7A00) /* 8 hours 40 minutes (26 minutes short of slot overrun). */


#define PERIOD_SET_OFFSET( cipher, keyType, newOffset )       ((cipher)->keyDesc[(keyType)].periodOffset = (newOffset))
#define KEY_EFFSLOT( cipher, keyType )                        ((cipher)->keyDesc[(keyType)].effSlot)
#define KEY_SET_EFFSLOT( cipher, keyType, newEffSlot )        ((cipher)->keyDesc[(keyType)].effSlot = (newEffSlot))
#define KEY_EPOCH( cipher, keyType )                          ((cipher)->keyDesc[(keyType)].epoch)
#define KEY_SET_EPOCH( cipher, keyType, newEpoch )            ((cipher)->keyDesc[(keyType)].epoch = (newEpoch))
#define KEY_SECUREGROUP( cipher, keyType )                    ((cipher)->keyDesc[(keyType)].secureGroup)
#define KEY_SET_SECUREGROUP( cipher, keyType, newSecureGroup )((cipher)->keyDesc[(keyType)].secureGroup = (newSecureGroup))
#define PERIOD_PROMOTETIME( cipher, keyType ) ((cipher)->periodStore[PERIOD_OFFSET( (cipher), (keyType) )].dofKey.promoteTime)
#define PERIOD_SET_PROMOTETIME( cipher, keyType, newPromoteTime ) \
                             ((cipher)->periodStore[PERIOD_OFFSET( (cipher), (keyType) )].dofKey.promoteTime = (newPromoteTime))
/* These next two masquerade birthTime by using the promoteTime member of the dofKey structure. */
#define PERIOD_BIRTHTIME( cipher, keyType )                   (PERIOD_PROMOTETIME( (cipher), (keyType) ))
#define PERIOD_SET_BIRTHTIME( cipher, keyType, newBirthTime ) (PERIOD_SET_PROMOTETIME( (cipher), (keyType), (newBirthTime)))


typedef struct AddressObserverParams_t                    /* AddrDown() observer callback uses this information. */
{
    OALAVLTree addressToNodeIDTree;
    DOFAddress address;
} AddressObserverParams_t;


typedef struct AddressToNodeID_t
{
    DOFAddress      address;
    AddressObserverParams_t * pAddressObserverParams;     /* For memory cleanup. */
    OALObserver     observer;                             /* To remove ourself as observer of DOFAddress down. */
    uint32          nodeID;
} AddressToNodeID_t;


/* Part of Nonce State. */
typedef struct SlotState_t
{
    OALAVLTree nodeIDTree;
    uint16     slot;
} SlotState_t;


/* Part of Nonce State. */
typedef struct NodeIDState_t
{
    uint32 nodeID;
    uint32 packetNumber;
} NodeIDState_t;


typedef struct PruneTraversalContext_t
{
    SlotState_t * pFoundSlotState;
    uint16        lowKeptSlot;
} PruneTraversalContext_t;


extern PCRMutex addr_mutex;


static const char cipherLogString[] = "OALCipher: ";

static void DestroyStoredKeyStateByPeriodOffset( OALCipher cipher, uint8 periodOffset );


/* @internal */
static void PrepLogOutString( char * pOutString, uint32 sizeOutString )
{
    PCRMem_Set( pOutString, 0, sizeOutString );
    PCRString_Copy( pOutString, cipherLogString, PCRString_Length( cipherLogString ) );
}


/*  @internal

    Purposely does not clear cipher->flags.nodeIDValid
*/
static void InitNodeIDMachine( OALCipher cipher )
{
    PCR_Assert( cipher );
    cipher->flags.noFirstNodeIDAssignment = 1;
    /* If we become SGM CCM manager: When we assign our own nodeID and then set our first key, we clear this flag
                                     and compute baseNodeID at 0 and advertisedNextNodeID above that.
       not SGM CCM manager: When we receive our nodeID with its associated baseNodeID, we clear this
                            flag. Following that we accept whatever the manager sends us as advertisedNextNodeID. */
}


/*  @internal

    @note The resulting cipher->nextTracked period is guaranteed to be of legal value.
    @param presentPeriod The known present period. If presented as illegal value, ->nextTrackedPeriod becomes 0.
*/
static void SetNextTrackedPeriod( OALCipher cipher, uint8 presentPeriod )
{
    cipher->nextTrackedPeriod = ++presentPeriod;
    if (cipher->nextTrackedPeriod > KEYCOUNT)
        cipher->nextTrackedPeriod = 0;
}


static void KeyTagAndPeriodOffsetEnable( OALCipher cipher, uint8 keyType, uint8 periodOffset )
{
    PCR_Assert( keyType == CUR || keyType == SLOT || keyType == KEKKEY );
    PERIOD_SET_OFFSET( cipher, keyType, periodOffset );
    cipher->keyDesc[keyType].tagged = TRUE;
    cipher->periodStore[PERIOD_OFFSET( cipher, keyType )].enabled = 1;
}


static void KeyUntag( OALCipher cipher, uint8 keyType )
{
    PCR_Assert( keyType == CUR || keyType == SLOT || keyType == KEKKEY );
    cipher->keyDesc[keyType].tagged = FALSE;
}


static void TaggedKeyDisable( OALCipher cipher, uint8 keyType )
{
    /* Nobody calls this with CUR. */
    PCR_Assert( keyType == CUR || keyType == SLOT || keyType == KEKKEY );
    if (cipher->keyDesc[keyType].tagged)
    {
        DestroyStoredKeyStateByPeriodOffset( cipher, PERIOD_OFFSET( cipher, keyType ) );
        cipher->keyDesc[keyType].tagged = FALSE;
    }
}


/* @internal */
static boolean slot_gt_fn( const uint16 * pSlot, const SlotState_t * pStoredSlotState, void * context )
{
    return *pSlot > pStoredSlotState->slot;
}


/* @internal */
static boolean slot_eq_fn( const uint16 * pSlot, const SlotState_t * pStoredSlotState, void * context )
{
    return *pSlot == pStoredSlotState->slot;
}


/* @internal */
static boolean nodeID_gt_fn( const uint32 * pNodeID, const NodeIDState_t * pStoredNodeIDState, void * context )
{
    return *pNodeID > pStoredNodeIDState->nodeID;
}


/* @internal */
static boolean nodeID_eq_fn( const uint32 * pNodeID, const NodeIDState_t * pStoredNodeIDState, void * context )
{
    return *pNodeID == pStoredNodeIDState->nodeID;
}


/** @internal
    Cleanup the given structure and its embedded memory allocations.

    @param pStoredAddress Pointer to the given structure.
    @param arg            Not used.
**/
static void CleanupStoredAddressEntries( OALAVLTreeItem_t pStoredAddress, void * arg )
{
    /* Called in OALCipher_Destroy() by OALAVLTree_Destroy() or by OALCipher_RemoveState(),
       for as many items as are held in addressToNodeIDTree.
       At each call, pStoredAddress is filled with the next item in the tree. Our function proto matches
       typedef boolean (*OALAVLTree_RemoveAll_fn_t)(OALAVLTreeItem_t item, void *arg); */
    /*For all callers, addr_mutex is locked right now. This protects cipher address tree manipulation from Addr_Down()*/
    PCR_Assert( ((AddressToNodeID_t*)pStoredAddress)->observer );              /* Remove ourself as observer of the DOFAddress observable. */
    DOFAddress_RemoveObserver( ((AddressToNodeID_t*)pStoredAddress)->address, ((AddressToNodeID_t*)pStoredAddress)->observer );
    DOFMem_Free( ((AddressToNodeID_t*)pStoredAddress)->pAddressObserverParams );         /* Removing observer prevents AddrDown() call so do this here*/
    DOFMem_Free( ((AddressToNodeID_t*)pStoredAddress) );                       /* pStoredAddress->address->nodeID not important here. */
}


/** @internal
    Observer callback called when a DOFAddress about to be freed. If DOFAddress is found within cipher tree, remove it.

    @param observable   OALObservable structure of the DOFAddress.
    @param pAddressObserverParams Pointer to structure
**/
static void AddrDown( OALObserver self, OALObservable observable, void* ctx )
{
    AddressObserverParams_t* pAddressObserverParams = (AddressObserverParams_t*)ctx;
    /* We are here ONLY for a DOFAddress structure that was placed into pAddressObserverParams->addressToNodeIDTree. */
    AddressToNodeID_t * pStoredAddress;

    PCR_Assert( observable );
    PCR_Assert( pAddressObserverParams );
    PCR_Assert( pAddressObserverParams->addressToNodeIDTree );
    PCR_Assert( pAddressObserverParams->address );
    PCRMutex_Lock( addr_mutex );
    pStoredAddress = (AddressToNodeID_t *)OALAVLTree_Retrieve( pAddressObserverParams->addressToNodeIDTree,
                                                               (OALAVLTreeKey_t)&pAddressObserverParams->address );
    /* If not in tree, do not attempt to remove it. How this can happen: A DOFAddress goes down before cipher
       is notified to remove it from tree. So the same DOFAddress is reused and therefore not placed in the tree
       (because it is already there). Then the second version is removed from the tree. Then the notification by
       the first version comes and there it is now not in the tree. So: do not try to remove it. */
    if (pStoredAddress)
    {
#if DEEPER_TREE_LOG
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "AddrDown() removing entry of nodeID 0x", pStoredAddress->nodeID );
#endif
        PCR_Assert( pStoredAddress->observer );                          /* Remove address from tree. */
        OALAVLTree_Delete( pAddressObserverParams->addressToNodeIDTree, (OALAVLTreeKey_t)&pAddressObserverParams->address );
#if DEEPER_TREE_LOG
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "AddrDown() removed an entry and now has entry num 0x",
                                  OALAVLTree_GetNumItems( pAddressObserverParams->addressToNodeIDTree ) );
#endif
        DOFMem_Free( pStoredAddress );
    }
    PCRMutex_Unlock( addr_mutex );
    /* The DOFAddress is gone and cannot call this callback again. So we do not need to remove us as its observer. */
    DOFMem_Free( pAddressObserverParams );
}


/** @internal
    Lookup given targetAddress in given OALAVLTree. If found, fill given memory with corresponding nodeID.

    @param addressToNodeIDTree Tree to find targetAddress in.
    @param targetAddress       DOFAddress to find in addressToNodeIDTree.
    @param pNodeID             Pointer to memory that will be filled with the found nodeIdentier. MUST not be NULL.
    @return TRUE for targetAddress found in addressToNodeIDTree and *pNodeID filled, FALSE otherwise.
**/
static boolean LookupAddressAndNodeID( OALAVLTree addressToNodeIDTree, DOFAddress targetAddress, uint32 * pNodeID )
{
    const AddressToNodeID_t * pStoredAddress;

    PCR_Assert( pNodeID );
    PCRMutex_Lock( addr_mutex );
    pStoredAddress = (const AddressToNodeID_t *)OALAVLTree_Retrieve( addressToNodeIDTree,
                                                                     (OALAVLTreeKey_t)&targetAddress );
    PCRMutex_Unlock( addr_mutex );
    if (pStoredAddress)
    {
        *pNodeID = pStoredAddress->nodeID;
        return TRUE;
    }
    return FALSE;
}


/* Called in FreeStoredSlotState() by OALAVLTree_Destroy() for as many NodeIDState_t * items as are held in
   pStoredSlotState->nodeIDTree. At each call, pStoredNodeIDState is filled with the next item in nodeIDTree.
   Our function proto matches typedef boolean (*OALAVLTree_RemoveAll_fn_t)(OALAVLTreeItem_t item, void *arg); */
static void FreeStoredNodeIDState(OALAVLTreeItem_t pStoredNodeIDState, void* pCountHeldState )
{
    PCR_Assert( *((uint32*)pCountHeldState) > 0 );
    if(pStoredNodeIDState)
    {
        (*((uint32*)pCountHeldState))--;            /* The decrement of held state count that our asserts watch for. */
        DOFMem_Free( ((NodeIDState_t*)pStoredNodeIDState) );
    }
    /* else add-state code had problems. It either could not create pStoredNodeIDState or could not add it to the tree. */
}


/* Called in DestroyStoredKeyStateByPeriodOffset() by OALAVLTree_Destroy() for each SlotState_t * item held in
   ->periodStore[periodOffset].slotTree. At each call, pStoredSlotState is filled with the next item in slotTree.
   Our function proto matches typedef boolean (*OALAVLTree_RemoveAll_fn_t)(OALAVLTreeItem_t item, void *arg); */
static void FreeStoredSlotState( OALAVLTreeItem_t pStoredSlotState, void* pCountHeldState )
{
    if (((SlotState_t*)pStoredSlotState)->nodeIDTree && *((uint32*)pCountHeldState) > 0)
        OALAVLTree_Destroy( ((SlotState_t*)pStoredSlotState)->nodeIDTree, FreeStoredNodeIDState,
                            (uint32 *)pCountHeldState );
    DOFMem_Free( ((SlotState_t*)pStoredSlotState) );
}


/* @internal */
static void DestroyStoredKeyStateByPeriodOffset( OALCipher cipher, uint8 periodOffset )
{
    PCR_Assert( periodOffset < KEYCOUNT );
    PCR_Assert( cipher->periodStore[periodOffset].enabled );
    if (cipher->periodStore[periodOffset].slotTree && cipher->periodStore[periodOffset].countHeldState)/*Only if state*/
    {
        uint32 removedCountHeldState = cipher->periodStore[periodOffset].countHeldState;

        OALAVLTree_Destroy( cipher->periodStore[periodOffset].slotTree, FreeStoredSlotState,
                            (uint32 *)&cipher->periodStore[periodOffset].countHeldState ); /* Decrements each removal.*/
        cipher->periodStore[periodOffset].slotTree = NULL;
        PCR_Assert( cipher->periodStore[periodOffset].countHeldState == 0 );               /* Prove full removal. */
        PCR_Assert( cipher->countHeldState >= removedCountHeldState );
        cipher->countHeldState -= removedCountHeldState;
    }
    DOFPermissionSet_Destroy( cipher->periodStore[periodOffset].localPerms );
    cipher->periodStore[periodOffset].localPerms = NULL;
    DOFPermissionSet_Destroy( cipher->periodStore[periodOffset].remotePerms );
    cipher->periodStore[periodOffset].remotePerms = NULL;
    cipher->periodStore[periodOffset].enabled = 0;
}


/* @internal
    @periodOffset If not legal, does nothing.
*/
static void EliminateKeyUseByPeriodOffset( OALCipher cipher, uint8 periodOffset )
{
    if (periodOffset < KEYCOUNT )
    {
        if (cipher->periodStore[periodOffset].enabled )
        {
            if (OALCipher_TaggedKeyEnabled( cipher, CUR ) && PERIOD_OFFSET( cipher, CUR ) == periodOffset)
                KeyUntag( cipher, CUR );
            if (OALCipher_TaggedKeyEnabled( cipher, SLOT ) && PERIOD_OFFSET( cipher, SLOT ) == periodOffset)
                KeyUntag( cipher, SLOT );
            if (OALCipher_TaggedKeyEnabled( cipher, KEKKEY ) && PERIOD_OFFSET( cipher, KEKKEY ) == periodOffset)
                KeyUntag( cipher, KEKKEY );
            if (OALCipher_KeyIsOldKey( cipher, periodOffset ))
                cipher->flags.oldKeyTimeoutEnabled = 0;
            DestroyStoredKeyStateByPeriodOffset( cipher, periodOffset );
        }
        PCR_Assert( !cipher->periodStore[periodOffset].enabled );
    }
}


/* @internal */
static void EliminateKeyUseByPeriod( OALCipher cipher, uint8 period )
{
    EliminateKeyUseByPeriodOffset( cipher, OALCipher_GetPeriodOffset( cipher, period ) );
}


/* Called in PruneStoredNonceStateUpToLowKeptSlot() by OALAVLTree_Traverse() on ->periodStore[periodOffset].slotTree.
   pStoredSlotState is filled with the lowest slot item in the tree, and we always stop traversing, since it either
   is a slot we can remove, or no other slots can be removed.
   Our function proto matches typedef boolean (*OALAVLTree_Traversal_fn_t)(OALAVLTreeItem_t item, void * arg); */
static boolean FindKillSlot_TraversalFunction( OALAVLTreeItem_t pStoredSlotState, void* pTraversalContext)
{
    if (((SlotState_t*)pStoredSlotState)->slot < ((PruneTraversalContext_t*)pTraversalContext)->lowKeptSlot)
    {
        ((PruneTraversalContext_t*)pTraversalContext)->pFoundSlotState = ((SlotState_t*)pStoredSlotState);
        return TRUE;                                          /* Found one that qualifies so stop traversing. */
    }
    PCR_Assert( !((PruneTraversalContext_t*)pTraversalContext)->pFoundSlotState );
    return TRUE; /* Stop traversing, since the traversal starts from the lowest entry and all remaining slots must be greater. */
}


/*  @internal
    Prune one slot at a time in increasing order from this key's lowAcceptedSlot.
    Do not prune @a lowKeptSlot or above, unless @a endingHeldStateCountMaximum is below this key's countHeldState.
    @param quitAtFirstPrunedSlot TRUE: Quit after successfully pruning one slot. FALSE: Prune all qualified found slots.
    @return TRUE/FALSE for pruning/not happened.
*/
static boolean PruneStoredNonceStateUpToLowKeptSlot( OALCipher cipher, uint8 periodOffset, uint16 lowKeptSlot,
                                                     uint32 endingHeldStateCountMaximum, boolean quitAtFirstPrunedSlot )
{
    /* This function is called with the cipher mutex locked. */
    boolean retBoolean = FALSE;

    PCR_Assert( cipher->periodStore[periodOffset].enabled );
    PCR_Assert( cipher->countHeldState >= cipher->periodStore[periodOffset].countHeldState );
    if (!cipher->periodStore[periodOffset].slotTree)
        return FALSE;
    while (cipher->periodStore[periodOffset].countHeldState &&                  /* No pruning to do with no heldState.*/
           (cipher->periodStore[periodOffset].lowAcceptedSlot < lowKeptSlot ||
            cipher->periodStore[periodOffset].countHeldState > endingHeldStateCountMaximum))
    {
        PruneTraversalContext_t traversalContext;
        boolean slotRemoved = FALSE;

        traversalContext.lowKeptSlot = lowKeptSlot;
        traversalContext.pFoundSlotState = NULL;                          /* Filled if matching slotState found.*/
        OALAVLTree_Traverse( cipher->periodStore[periodOffset].slotTree,
                             FindKillSlot_TraversalFunction, &traversalContext );
        if (traversalContext.pFoundSlotState)
        {
            if (traversalContext.pFoundSlotState->nodeIDTree)  /* Remove pFoundSlotState's nodeID tree members. */
            {
                uint32 originalCountHeldState;
                uint32 removedCountHeldState;

                originalCountHeldState = cipher->periodStore[periodOffset].countHeldState;
                OALAVLTree_Destroy( traversalContext.pFoundSlotState->nodeIDTree,   /* Remove state for 1 slot. */
                                    FreeStoredNodeIDState, /* Decr. n .countHeldState. */
                                    &cipher->periodStore[periodOffset].countHeldState );
                /* cipher->periodStore[periodOffset].countHeldState reduced only by that held in eliminated slot. */
                removedCountHeldState = originalCountHeldState - cipher->periodStore[periodOffset].countHeldState;
                cipher->countHeldState -= removedCountHeldState;
                PCR_Assert( cipher->countHeldState >= cipher->periodStore[periodOffset].countHeldState );
            }
            OALAVLTree_Delete( cipher->periodStore[periodOffset].slotTree,  /* Remove pStoredSlotState's .. */
                               &traversalContext.pFoundSlotState->slot ); /* .. entry in .slotTree .. */
            DOFMem_Free( traversalContext.pFoundSlotState );              /* .. and free it. */
            slotRemoved = TRUE;
            retBoolean = TRUE;
        }
        cipher->periodStore[periodOffset].lowAcceptedSlot++;                /* Whether or not actually pruned. */
        if (quitAtFirstPrunedSlot && slotRemoved)
            break;
    } /* end while. */
    if (!cipher->periodStore[periodOffset].countHeldState)
    {
        PCR_Assert( cipher->periodStore[periodOffset].slotTree );
        PCR_Assert( OALAVLTree_GetNumItems( cipher->periodStore[periodOffset].slotTree ) == 0 );
        OALAVLTree_Destroy( cipher->periodStore[periodOffset].slotTree, NULL, NULL );
        cipher->periodStore[periodOffset].slotTree = NULL;
    }

    return retBoolean;
}


/* @internal */
static void SetNodeIDValid( OALCipher cipher )
{
    cipher->flags.nodeIDValid = 1;
    cipher->flags.nodeIDInWarning = 0;
    cipher->flags.noFirstNodeIDAssignment = 0;
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ) && cipher->flags.sendRequestPDUToNextFoundManager)
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_TRACE, "OALCipher: Have new nodeID, cancelling sendRequestPDUToNextFoundManager.");
    cipher->flags.sendRequestPDUToNextFoundManager = 0;
}


static void CopyNewKeyAndExpandNewKey( OALCipher cipher, uint8 keyType, const OALCryptKey newKey )
{
    /* Note: We expand every newKey that is placed, and do not wait until the key is promoted to be curKey, because
             every key placed, before being used to send packets, can be used to decrypt incoming packets. */
    PCRMem_Copy( cipher->periodStore[cipher->keyDesc[keyType].periodOffset].dofKey.key, newKey, sizeof( OALCryptKey ) );
    PCR_Validate( DOFCipherRegistry_AES256Cipher );
    DOFCipherRegistry_AES256Cipher->GenerateKeyState( cipher->periodStore[cipher->keyDesc[keyType].periodOffset].dofKey.ekey, newKey );
}


/* @internal */
static void CurKeyBasicSetup( OALCipher cipher )
{
    cipher->periodStore[PERIOD_OFFSET( cipher, CUR )].dofKey.txSlotAdjust = 0;
    cipher->periodStore[PERIOD_OFFSET( cipher, CUR )].dofKey.txLastSlot = 0;
    cipher->encipherPacketNumber = IsStreamingConnection( cipher ) ? 1 : 0;
    OALCipher_ResetQuickNodeStore( cipher );
}


/*  @internal
    Find the next period store we can safely use.  Only used by a manager.
    @note This may eliminate a previously valid key.
*/
static uint8 GetCleanNextPeriod( OALCipher cipher )
{
    /*
     * Previous managers and lost packets may cause us to consume (no longer be able to use) excess periods.
     * Here, we look for the next period we can use that is not the one used by the current key.
     * Other keys are not important, since we are (now) the manager.
     */
    if (++cipher->nextTrackedPeriod >= PERIODCOUNT)
        cipher->nextTrackedPeriod = 0;
    if (OALCipher_TaggedKeyEnabled( cipher, CUR ) &&
        cipher->periodStore[PERIOD_OFFSET( cipher, CUR )].period == cipher->nextTrackedPeriod)
    {
        PCR_Assert( cipher->periodStore[PERIOD_OFFSET( cipher, CUR )].enabled );
        if (++cipher->nextTrackedPeriod > PERIODCOUNT)                  /* Do not overwrite enabled curKey. */
            cipher->nextTrackedPeriod = 0;
    }
    EliminateKeyUseByPeriod( cipher, cipher->nextTrackedPeriod );       /* Existing key of retPeriod is GONE (should be oldest one and least likely to be used). */

    return cipher->nextTrackedPeriod;
}


/* @internal
   From KEYCOUNT PeriodStore_t entries, get an open entry (KEYCOUNT is at least 4).

   1st: If any key using given period, eliminate its entry and return its periodOffset
                              (therefore we will never have 2 entries of same period).
   If still here, 2nd: Find an open entry and return its periodOffset.
   If still here, 3rd: Find oldest oldKey, eliminate its entry, and return its periodOffset.

   @note If all periodOffsets are filled at entry, we must have at least one oldKey that can be used.
*/
static uint8 GetOpenPeriodOffset( OALCipher cipher, uint8 period )
{
    uint8  periodOffset;
    uint32 oldestOldKey_PromoteTime;
    uint8  oldestOldKey_PeriodOffset = KEYCOUNT;                /* If used this illegal value is 1st overwritten. */

    for (periodOffset = 0; periodOffset < KEYCOUNT; periodOffset++)
        if (cipher->periodStore[periodOffset].enabled && cipher->periodStore[periodOffset].period == period)
        {
            EliminateKeyUseByPeriodOffset( cipher, periodOffset );
            return periodOffset;
        }
    oldestOldKey_PromoteTime = PCRTime_GetMS();                 /*Ensure 1st found oldKey is marked- is older than now*/
    for (periodOffset = 0; periodOffset < KEYCOUNT; periodOffset++)
    {
        if (!cipher->periodStore[periodOffset].enabled)
            return periodOffset;                                /* No key eliminated. */
        if (OALCipher_KeyIsOldKey( cipher, periodOffset ) &&
            DOFTime_Compare( oldestOldKey_PromoteTime, cipher->periodStore[periodOffset].dofKey.promoteTime ) >= 0)
        {
            PCR_Assert( cipher->periodStore[periodOffset].dofKey.promoted );
            oldestOldKey_PromoteTime = cipher->periodStore[periodOffset].dofKey.promoteTime;
            oldestOldKey_PeriodOffset = periodOffset;
        }
    }
    /* Here if all periodOffsets are enabled and none are of our period. One or more is/are oldKey. The assert proves
       that oldestOldKey_PromoteTime and oldestOldKey_PeriodOffset were written in above loop. */
    PCR_Assert( oldestOldKey_PeriodOffset < KEYCOUNT);
    EliminateKeyUseByPeriodOffset( cipher, oldestOldKey_PeriodOffset ); /* Eliminate existing old key not having our period. */
    return oldestOldKey_PeriodOffset;
}


/* @internal
   @note This function may replace a previously valid key. It may convert a previous curKey to an old key                                               .
   @param cipher      Cipher for which curKey will be replaced.
   @param newKey      Bytes of the new Key.
   @param period      Period of the supplied newKey. period must be valid (less than PERIODCOUNT).
   @param promoteTime Time that given key promoted.
 */
static void ReplaceCurKey( OALCipher cipher, const OALCryptKey newKey, uint8 period, uint32 promoteTime )
{
    /* Called with cipher mutex locked. */
    uint8 periodOffset;
    DOFPermissionSet localPerms = NULL;
    DOFPermissionSet remotePerms = NULL;

    PCR_Assert( period < PERIODCOUNT );
    periodOffset = PERIOD_OFFSET( cipher, CUR );
    if ( cipher->periodStore[periodOffset].enabled )
    {
        /* The CUR key is being replaced, but we want to keep its permissions, if any. Steal them here, and assign them to the new CUR key. */
        localPerms = cipher->periodStore[periodOffset].localPerms;
        cipher->periodStore[periodOffset].localPerms = NULL;
        remotePerms = cipher->periodStore[periodOffset].remotePerms;
        cipher->periodStore[periodOffset].remotePerms = NULL;
    }
    periodOffset = GetOpenPeriodOffset( cipher, period );     /* This may eliminate a previously valid key. */
    PCR_Assert( periodOffset < KEYCOUNT );
    KeyTagAndPeriodOffsetEnable( cipher, CUR, periodOffset ); /* By tagging as curKey, existing curKey becomes old key*/
    CurKeyBasicSetup( cipher );
    cipher->periodStore[periodOffset].dofKey.promoteTime = promoteTime;
    cipher->periodStore[periodOffset].dofKey.promoted = TRUE;
    CopyNewKeyAndExpandNewKey( cipher, CUR, newKey );
    cipher->periodStore[periodOffset].period = period;
    PCR_Assert( cipher->periodStore[periodOffset].slotTree == NULL );
    PCR_Assert( cipher->periodStore[periodOffset].localPerms == NULL );
    cipher->periodStore[periodOffset].localPerms = localPerms;
    PCR_Assert( cipher->periodStore[periodOffset].remotePerms == NULL );
    cipher->periodStore[periodOffset].remotePerms = remotePerms;
    cipher->oldKeyDisableTime = PCRTime_GetMS() + (1000UL * cipher->secsStandardPruneOrKeyDisable);
    cipher->flags.oldKeyTimeoutEnabled = 1;
    PCR_Assert( PERIOD_OFFSET_ENABLED( cipher, CUR ) );
}


/*  @internal
    @return TRUE if any pruning happened, FALSE if not.
*/
static boolean PruneAllAgedKeys( OALCipher cipher, uint16 secsPruneAge )
{
    /* Called with cipher mutex locked. */
    /* This does all-key pruning to secsPruneAge, which is most often 2*Thb. */
    uint8   periodOffset;
    boolean retBoolean = FALSE;

    /* If _Encipher() and _Decipher() are regularly called,  they call this at least once per second, via Adjust_Keys().
       If the former are not called, it does not matter that we do not prune aged keys regularly.
       Note: slotKey and kekKey are not pruned, because they are not promoted and have no age. */
    /* Only secure groups track and care about slots. */
    if (OALCipher_IsSecureGroup( cipher ))
    {
        for (periodOffset = 0; periodOffset < KEYCOUNT; periodOffset++)
        {
            if (cipher->periodStore[periodOffset].enabled && cipher->periodStore[periodOffset].dofKey.promoted &&
                cipher->periodStore[periodOffset].slotTree)
            {
                uint16 keySlot     = (PCRTime_GetMS() - cipher->periodStore[periodOffset].dofKey.promoteTime) / 1000;
                uint16 lowKeptSlot = (keySlot > secsPruneAge) ? (keySlot - secsPruneAge) : 0;

                if (PruneStoredNonceStateUpToLowKeptSlot( cipher, periodOffset, lowKeptSlot,
                                                          NO_ENDING_HELD_STATE_COUNT_MAXIMUM, FALSE ) != FALSE)
                    retBoolean = TRUE;
            }
        }
    }
    return retBoolean;
}


static void CheckBlockedMessageQueue( OALCipher cipher )
{
    PCR_Assert( cipher );
    if (OALCipher_IsSecureGroup( cipher ) && cipher->connectionChannel)
    {
        OALChannel secureGroupChannel = cipher->connectionChannel;

        OALCipher_Lock( cipher );
        if (secureGroupChannel->remoteChannelData->flags.SecureSendBlocked)
        {
            /* Create a message for the channel to unblock the send queue. */
            OALMessage msgToChannel = CMT_ProcessBlockedMessage( cipher->connectionChannel, TRUE );

            if (msgToChannel)
                OALRemoteChannelMgr_SendMessageToChannel( msgToChannel );
            else
            {
                OALRemoteChannel_Shutdown( cipher->connectionChannel, DOFEXCEPTION_INSUFFICIENT_RESOURCES );
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG,
                                    "OALCipher: Connection closed because of failure to unblock secureGroup message queue." );
            }
        }
        OALCipher_Unlock( cipher );
    }
}


static boolean StoreManagerSuppliedNonceState( OALCipher cipher, const ccmModeState_t * pOutCcmModeState,
                                               uint8 periodOffset )
{
    /* Only called for a new key. */
    boolean success = TRUE;

    cipher->periodStore[periodOffset].lowAcceptedSlot = pOutCcmModeState->slot;/*Manager told us group is at this slot*/
    /*
     * Manager tells us his *next* packet number, which we want to accept.
     * If it is 0, there is nothing to store.  Otherwise, we pretend to have received his *last* packet.
     * Regardless, we will not accept a packet (from anybody) at any slot below cipher->periodStore[periodOffset].lowAcceptedSlot.
     */
    if (pOutCcmModeState->packetNumber > 0)
    {
        NonceState_t sgmNonceState; /* To prevent reacting to PDU replays of manager's PDUs, store his nonceState now.*/

        sgmNonceState.nodeID = pOutCcmModeState->nodeID;
        sgmNonceState.slot = pOutCcmModeState->slot;
        sgmNonceState.packetNumber = pOutCcmModeState->packetNumber - 1;  /* Allow outCcmModeState.packetNumber recv. */
        success = OALCipher_UpdateStoredNonceState( cipher, periodOffset, &sgmNonceState );
    }

    return success;
}


static boolean ProcessManagerSuppliedInitialState( OALCipher cipher, uint8 periodOffset,
                                                   const ccmModeState_t * pOutCcmModeState,
                                                   DOFAddress * pManagerAddress )
{
    /* Called with cipher mutex locked. */
    boolean retBoolean;

    if ((pOutCcmModeState->flags & (MANAGER_TRANSMITTED_PDU | TRANSMIT_COUNT)) !=
                                   (MANAGER_TRANSMITTED_PDU | TRANSMIT_COUNT))
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG,
                            "OALCipher: ERROR: from manager supplied initial state, required state not found." );
        return FALSE;
    }
    retBoolean = StoreManagerSuppliedNonceState( cipher, pOutCcmModeState, periodOffset );
#if DEEPER_TREE_LOG /* This not tree related thing, but this define is useful to turn on/off this set of log messages.*/
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ) && retBoolean)
    {
        OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "Evaluating manager supplied key initial state, I am manager: 0x",
                                                 (uint32)cipher->flags.manager );
        OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                              "From manager supplied key initial state, found manager's period as 0x",
                              (uint32)pOutCcmModeState->period );
        OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                              "From manager supplied key initial state, found manager's nodeID as 0x",
                              (uint32)pOutCcmModeState->nodeID );
        OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                              "From manager supplied key initial state, found manager's slot as 0x",
                              (uint32)pOutCcmModeState->slot );
        OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                              "From manager supplied key initial state, found manager's packet number as 0x",
                              (uint32)pOutCcmModeState->packetNumber );
    }
#endif
    /* Only care about initial state if we are not (also?) a manager. */
    if (retBoolean && !cipher->flags.manager)
    {
        /* Use manager's supplied target nodeID info. */
        if (pOutCcmModeState->flags & TRANSMIT_TARGET_NODEID)
        {
            cipher->nodeID = pOutCcmModeState->targetNodeID;
            SetNodeIDValid( cipher );
            CheckBlockedMessageQueue( cipher );
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                                      "From manager supplied key initial state, updated my nodeID with targetNodeID 0x",
                                      (uint32)cipher->nodeID );
        }
        if (pManagerAddress && *pManagerAddress)
            OALCipher_StoreAddressNodeIDPair( cipher, pManagerAddress, pOutCcmModeState->nodeID, FALSE );
        /* Special: In case I become SGM manager, I use this info only as a backup, when I don't know the previous
                    SGM's ->advNNI. That is our only use of storing this generically in our class. */
        cipher->baseNodeID = pOutCcmModeState->baseNodeID;                 /* SGM's CCM baseNodeID */
#if DEEPER_TREE_LOG /* This not tree related thing, but this define is useful to turn on/off this set of log messages.*/
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ) && OALCipher_IsSecureGroup( cipher ))
            OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                                  "From manager supplied key initial state, storing other manager's baseNodeID 0x",
                                  cipher->baseNodeID );
#endif
    }
    return retBoolean;
}


/*  @internal

    @note  nodeID, slot, and packetNumber are always placed. nodeID is that of this cipher instance.
    @note  This is called only by CCM manager.
    @note  Buffer at pBuf absolutely must be at least @a *pTotalSize bytes long.
    @param self             The cipher. If NULL, @a period must be below PERIODCOUNT, and managerNodeID is assigned as 0
    @param period           Supplied period if below PERIODCOUNT. Otherwise MUST be GET_NEXT_PERIOD, and this
                            call supplies new period.
    @param slot             slot to place in initial state buffer. This must be <= 0x7FFF.
    @param packetNumber     packet number to place in initial state buffer. This must be <= 0x3FFFFFFF.
    @param includeTargetNID TRUE: include @a targetNodeID in initial state buffer.
    @param targetNodeID     targetNodeID to place in initial state buffer, if includeTargetNID true. Must be valid.
    @param baseNodeID       baseNodeID to place in initial state buffer. Must be valid, and must be 0 for TEP.
    @param pBuf             Pointer to output buffer. MUST not be NULL.
    @param pTotalSize       Pointer to totalSize. [IN] available size of pBuf. [OUT] actual filled size of pBuf.
    @param pNIDPlacedNotVal pointer to status: filled TRUE when nodeID placed in initial state buffer is not valid.
    @return TRUE for success, or
            FALSE if @a *pTotalSize is not sufficient to hold initial state, or @a period is supplied and is illegal, or
                     @a targetNodeID is supplied and is illegal, or @a baseNodeID is supplied and is illegal, or
                     for self's nodeID valid, @a targetNodeID is supplied and is same value as self's own nodeID, or
                     for self's nodeID not valid, @a targetNodeID is supplied and is 0.
*/
static boolean FillInitialStateBuf( const OALCipher self, uint8 period, uint16 slot, uint32 packetNumber,
                                    boolean includeTargetNID, uint32 targetNodeID, uint32 baseNodeID, uint8 * pBuf,
                                    uint8 * pTotalSize, boolean * pNIDPlacedNotVal )
{
    PCR_Assert( pBuf );
    PCR_Assert( period == GET_NEXT_PERIOD || period < PERIODCOUNT );
    PCR_Assert( self || period < PERIODCOUNT );
    if (pTotalSize && (!includeTargetNID || targetNodeID <= TOP_NODEID) && baseNodeID <= TOP_NODEID)
    {
        uint8  availableSize;
        uint32 managerNodeID;
        uint8  marshaledSize = 0;

        managerNodeID = self ? self->nodeID : 0;
        PCR_Assert( managerNodeID <= TOP_NODEID );
        if (includeTargetNID && managerNodeID == targetNodeID)
            return FALSE;
        *pNIDPlacedNotVal = self == NULL || !OALCipher_IsNodeIDValid( self );
        availableSize = *pTotalSize - 1;
        *pTotalSize = 1;                                      /* Note: pBuf[0] is filled at very end at success only. */
        /*If availableSize becomes too small for our need, one of the OALMarshal_CompressValue() calls below will fail*/
        if ((marshaledSize = OALMarshal_CompressValue( 4, managerNodeID, availableSize, &pBuf[*pTotalSize] )) != 0)
        {
            *pTotalSize += marshaledSize;
            availableSize -= marshaledSize;
            PCR_Assert( slot <= 0x7FFF );
            if ((marshaledSize = OALMarshal_CompressValue( 2, slot, availableSize, &pBuf[*pTotalSize] )) != 0)
            {
                *pTotalSize += marshaledSize;
                availableSize -= marshaledSize;
                PCR_Assert( packetNumber <= 0x3FFFFFFF );
                if ((marshaledSize = OALMarshal_CompressValue(4, packetNumber, availableSize,&pBuf[*pTotalSize])) != 0)
                {
                    *pTotalSize += marshaledSize;
                    availableSize -= marshaledSize;
                    marshaledSize = 0;
                    PCR_Assert( !includeTargetNID || targetNodeID <= TOP_NODEID );
                    if (!includeTargetNID || (marshaledSize = OALMarshal_CompressValue( 4, targetNodeID, availableSize,
                                                                                        &pBuf[*pTotalSize] )) != 0)
                    {
                        *pTotalSize += marshaledSize;
                        availableSize -= marshaledSize;
                        marshaledSize = 0;
                        PCR_Assert( baseNodeID <= TOP_NODEID );
                        if (baseNodeID == 0 || (marshaledSize = OALMarshal_CompressValue( 4, baseNodeID, availableSize,
                                                                                          &pBuf[*pTotalSize] )) != 0)
                        {
                            *pTotalSize += marshaledSize;
                            if (period >= PERIODCOUNT)
                            {
                                /* Not given period, so we use next known available period. */
                                PCR_Assert( period == GET_NEXT_PERIOD );
                                period = GetCleanNextPeriod( self );    /* This may eliminate a previously valid key. */
                            }
                            else
                            {
                                PCR_Assert( OALCipher_IsSecureGroup( self ) );                /* Use period given us. */
                            }
                            pBuf[0] = MANAGER_TRANSMITTED_PDU | (period << 4) | ((uint8)includeTargetNID << 3) |
                                      ((uint8)(baseNodeID != 0) << 1) | TRANSMIT_COUNT;
                            return TRUE;                                                               /* Success. */
                        }
                    }
                }
            }
        }
    }

    return FALSE;                                                                                      /* Fail. */
}


/*  @internal

    @note  nodeID, slot, and packetNumber are always placed. nodeID is that of this cipher instance.
    @note  This is called only by CCM manager.
    @param self             The cipher. If NULL, @a period must be below PERIODCOUNT, and managerNodeID is assigned as 0
    @param period           Supplied period if below PERIODCOUNT. Otherwise MUST be GET_NEXT_PERIOD, and this
                            call supplies new period.
    @param initialSlot      slot of to place in initial state buffer. This must be <= 0x7FFF.
    @param initialPacketNum packet number to place in initial state buffer. This must be <= 0x3FFFFFFF.
    @param includeTargetNID TRUE: include @a targetNodeID in initial state buffer.
    @param targetNodeID     targetNodeID to place in initial state buffer, if includeTargetNID true. Must be valid.
    @param baseNodeID       baseNodeID to place in initial state buffer. Must be valid, and must be 0 for TEP.
    @param outStream        Stream to write
    @return TRUE for success, or FALSE on fail.
*/
static boolean WriteInitialStateStream( OALCipher self, uint8 period, uint16 initialSlot, uint32 initialPacketNum,
                                        boolean includeTargetNID, uint32 targetNodeID, uint32 baseNodeID,
                                        OALStream outStream )
{
    /* @param self The cipher corresponding to the supplied initial state. This may be NULL if cipher not yet created.*/
    /* @param outStream MUST not be NULL. */
    boolean nodeIDPlacedNotValidated;
    uint8   ccmModeCompressedBuf[MAX_CCMMODE_INITIAL_STATE_SIZE];
    uint8   ccmModeCompressedBufFinalSize = MAX_CCMMODE_INITIAL_STATE_SIZE;

    PCR_Assert( outStream );
    PCR_Assert( !OALCipher_IsSecureGroup( self ) || period < PERIODCOUNT );
    PCR_Assert(  OALCipher_IsSecureGroup( self ) || period == GET_NEXT_PERIOD );
    PCR_Assert(  OALCipher_IsSecureGroup( self ) || baseNodeID == 0 );
    if ((FillInitialStateBuf( self, period, initialSlot, initialPacketNum, includeTargetNID, targetNodeID, baseNodeID,
                              ccmModeCompressedBuf, &ccmModeCompressedBufFinalSize, &nodeIDPlacedNotValidated )
                                                                                                            != FALSE) &&
        !nodeIDPlacedNotValidated)
    {
        if (OALStream_Write( outStream, ccmModeCompressedBufFinalSize, ccmModeCompressedBuf ))
            return TRUE;
    }

    return FALSE;
}


/* @internal
    @note This call first clears the buffer at @a pOutCcmModeState, then fills those fields required by @a pParseBuf.
    @param sizeParseBuf      Data size within @a pParseBuf. 0/not: lossless transport initialization / not.
    @param pParseBuf         Pointer to compressed ccm mode initial state.
    @param pOutCcmModesState Pointer to ccmModeState_t struct, holding uncompressed results. This must not be NULL.
    @return TRUE for pCcmModeState filled and representing either lossless transport initialization or a properly
            formed CCM.9 PDU with all legal values, FALSE otherwise.
*/
static boolean UncompressCcmModeState( uint8 sizeParseBuf, const uint8 * pParseBuf, ccmModeState_t * pOutCcmModeState,
                                       boolean stateReceivedByIncomingMulticastPDU, boolean cipherIsSecureGroup )
{
    uint32 slot;                                    /* uint32 for OALMarshal_UncompressValue(). Actual use is uint16. */
    uint32 lenMarshaledData;

    PCR_Assert( pOutCcmModeState );
    PCRMem_Set( pOutCcmModeState, 0, sizeof( ccmModeState_t ) );
    if (sizeParseBuf == 0)
        return TRUE;                                   /* Lossless transport initialization, valid but members all 0. */
    /* We have CCM.9 PDU. If TRUE is returned, every pOutCcmModeState member is filled, except that
       pOutCcmModeState->targetNodeID is default value 0 if T bit is clear, and
       pOutCcmModeState->baseNodeID is default value 0 if N bit is clear. */
    if (!pParseBuf ||
        !(*pParseBuf & TRANSMIT_COUNT) ||              /* Spec says these flags bits always set when not streaming. */
        !(*pParseBuf & MANAGER_TRANSMITTED_PDU))
        return FALSE;
    pOutCcmModeState->flags = *pParseBuf++;
    sizeParseBuf--;
    /* Note: any of the following _UncompressValue() calls fail if sizeParseBuf becomes too small. */
    pOutCcmModeState->manager = (boolean)(pOutCcmModeState->flags >> 7);
    pOutCcmModeState->period = (pOutCcmModeState->flags >> 4) & 0x07;
    /* Get nodeID. */
    lenMarshaledData = sizeParseBuf;
    pOutCcmModeState->nodeID = OALMarshal_UncompressValue( 4, &lenMarshaledData, pParseBuf );
    if (pOutCcmModeState->nodeID > TOP_NODEID)
        return FALSE;
    pParseBuf += lenMarshaledData;
    PCR_Assert( sizeParseBuf >= lenMarshaledData );
    sizeParseBuf -= lenMarshaledData;
    /* Get slot. */
    lenMarshaledData = sizeParseBuf;
    slot = OALMarshal_UncompressValue( 2, &lenMarshaledData, pParseBuf );
    if (slot >= 0x8000)
        return FALSE;
    pOutCcmModeState->slot = (uint16)slot;
    pParseBuf += lenMarshaledData;
    PCR_Assert( sizeParseBuf >= lenMarshaledData);
    sizeParseBuf -= lenMarshaledData;
    /* Get packet number. */
    lenMarshaledData = sizeParseBuf;
    pOutCcmModeState->packetNumber = OALMarshal_UncompressValue( 4, &lenMarshaledData, pParseBuf );
    if (pOutCcmModeState->packetNumber > 0x3FFFFFFF)
    {
        return FALSE;
    }
    PCR_Assert( sizeParseBuf >= lenMarshaledData);
    if (pOutCcmModeState->flags & TRANSMIT_TARGET_NODEID)
    {
        /* Get targetNodeID. */
        pParseBuf += lenMarshaledData;
        sizeParseBuf -= lenMarshaledData;
        lenMarshaledData = sizeParseBuf;
        pOutCcmModeState->targetNodeID = OALMarshal_UncompressValue( 4, &lenMarshaledData, pParseBuf );
        if (pOutCcmModeState->targetNodeID > TOP_NODEID ||
            pOutCcmModeState->nodeID == pOutCcmModeState->targetNodeID)              /* Spec CCM.20. */
            return FALSE;
        PCR_Assert( sizeParseBuf >= lenMarshaledData);
        if (cipherIsSecureGroup && stateReceivedByIncomingMulticastPDU)
            return FALSE;              /* Enforce CCM.21 spec. */
    } /* else already initialized to 0 above. Should not be used because flags indicate it is not there. */
    if (pOutCcmModeState->flags & TRANSMIT_BASE_OR_NEXT_NODEID)
    {
        /* Get baseNodeID. */
        pParseBuf += lenMarshaledData;
        sizeParseBuf -= lenMarshaledData;
        lenMarshaledData = sizeParseBuf;
        pOutCcmModeState->baseNodeID = OALMarshal_UncompressValue( 4, &lenMarshaledData, pParseBuf );
        if (pOutCcmModeState->baseNodeID > TOP_NODEID)
            return FALSE;
        PCR_Assert( sizeParseBuf >= lenMarshaledData);
    } /* else already initialized to 0 above. Spec says default is 0, though only secureGroup uses this. */

    return TRUE;
}


static void SendRequestPDU( const OALCipher self )
{
    PCR_Assert( self );
    if (self->connectionChannel)                                             /* Unit test does not have this channel. */
    {
        if (self->managerAddress)
        {
            self->flags.sendRequestPDUToNextFoundManager = 0;
            self->flags.nodeIDInWarning = 1;
            self->flags.newNodeIDScheduled = 0;
            CCMMODE_V1_SendRequest( self->connectionChannel, self->managerAddress );
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "OALCipher: CCM Request PDU sent." );
        }
        else
        {
            self->flags.sendRequestPDUToNextFoundManager = 1;
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE,
                                "OALCipher: CCM Request PDU will be sent when manager is found." );
        }
    }
}


/* @internal */
static int GetNodeIDTimeout_fn( OALTimerItem item, uint32 argc, void *argv[], boolean shutdown )
{
    OALCipher cipher = (OALCipher)argv[0];

    PCR_Assert( cipher );
    PCR_Assert( argc == 1 );
    OALCipher_GetNewNodeID( cipher );

    return 0;
}


static uint32 GetNextNodeIDToIssue( OALCipher cipher )
{
    uint32 nextNodeIDToIssue = cipher->nextNodeIDToIssue++;

    if ( ( cipher->nextNodeIDToIssue - cipher->baseNodeID ) > MANAGER_BASE_NODEID_ADVANCE_VALUE - NEED_NEW_BASE_NODEID_TRIGGER_COUNT )
        cipher->flags.needNewBaseNodeID = 1;
    if (cipher->nextNodeIDToIssue > TOP_NODEID)
        cipher->nextNodeIDToIssue = 0;

    return nextNodeIDToIssue;
}


/*
**  "Internal Public"  Methods.
*/


/* @internal */
/* pAddress: 1 of n DOFAddress held in the tree.  */
boolean OALCipher_AddressIP_Port_gt_fn( const DOFAddress * pAddress, const DOFAddress * pTestAddress, void * context )
{
    /* DOFAddress_Compare() can return +, 0 or -, but avltree function requires that we return boolean. */
    int retVal = DOFAddress_Compare( *pAddress, *pTestAddress );                       /* ENPPort used. */

    return (retVal > 0);               /* DOFAddress_Compare() returns + for *pAddress > *pTestAddress. */
}


/* @internal */
boolean OALCipher_AddressIP_Port_eq_fn( const DOFAddress * pAddress, const DOFAddress * pTestAddress, void * context )
{
    return DOFAddress_IsEqual( *pAddress, *pTestAddress );                               /* ENPPort used. */
}


void OALCipher_GetNewNodeID( const OALCipher cipher )
{
    PCR_Assert( cipher );
    if (cipher->flags.manager)
    {
        cipher->nodeID = GetNextNodeIDToIssue( cipher );                         /* As SGM, select my own new nodeID. */
        PCR_Assert( cipher->nodeID <= TOP_NODEID );
        SetNodeIDValid( cipher );
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            OALCipher_LogMessage(DOFLOGLEVEL_TRACE, "_GetNewNodeID() selects new self nodeID: 0x", cipher->nodeID);
        cipher->flags.sendRequestPDUToNextFoundManager = 0;
        CheckBlockedMessageQueue( cipher );
        cipher->flags.newNodeIDScheduled = 0;
    }
    else
        SendRequestPDU( cipher );                           /* Clears cipher->flags.sendRequestPDUToNextFoundManager. */
}


/* @internal */
boolean OALCipher_TaggedKeyEnabled( const OALCipher cipher, uint8 keyType )
{
    PCR_Assert( keyType == CUR || keyType == SLOT || keyType == KEKKEY );
    if (cipher->keyDesc[keyType].tagged)
    {
        PCR_Assert( PERIOD_OFFSET_ENABLED( cipher, keyType ) );
        return TRUE;
    }

    return FALSE;
}


/* @internal */
boolean OALCipher_TaggedKey_PeriodOffsetMatches( const OALCipher cipher, uint8 keyType, uint8 periodOffset )
{
    PCR_Assert( keyType == CUR || keyType == SLOT || keyType == KEKKEY );

    return (OALCipher_TaggedKeyEnabled( cipher, keyType ) && PERIOD_OFFSET( cipher, keyType ) == periodOffset);
}


/* @internal */
boolean OALCipher_KeyIsOldKey( const OALCipher cipher, uint8 periodOffset )
{
    return (cipher->periodStore[periodOffset].enabled &&
            !OALCipher_TaggedKey_PeriodOffsetMatches( cipher, CUR, periodOffset ) &&
            !OALCipher_TaggedKey_PeriodOffsetMatches( cipher, SLOT, periodOffset ) &&
            !OALCipher_TaggedKey_PeriodOffsetMatches( cipher, KEKKEY, periodOffset ));
}


/* @internal */
uint8 OALCipher_GetPeriodOffset( const OALCipher cipher, uint8 period )
{
    uint8 i;       /* Can convert to an efficient sorted list, but with only four stored keys it may not be worth it. */

    PCR_Assert( period < PERIODCOUNT );
    for (i = 0; i < KEYCOUNT; i++)
        if (cipher->periodStore[i].enabled && cipher->periodStore[i].period == period)
            return i;

    return PERIOD_DISABLED;
}


/* @internal */
boolean OALCipher_LogMessage( DOFLogLevel logLevel, const char * message, uint32 trailingValue )
{
    char outString[200];

    if ((sizeof( cipherLogString ) + PCRString_Length( message ) + 8 ) <= sizeof( outString ))
    {
        PrepLogOutString( (char *)&outString, sizeof( outString ) );
        PCRString_Concat( (outString + PCRString_Length( outString )), message, PCRString_Length( message ) );
        OALString_Uint32ToHexString( (outString + PCRString_Length( outString )), trailingValue );
        DOF_LogMessage( DOF_GetDof(), logLevel, outString );
        return TRUE;
    }

    return FALSE;
}


/* This is from the CCM spec. */
uint8 OALCipher_ComputeMacAuthFlags( uint8 macSize )
{
    return (A_DATA_PRESENT_FLAG | (((macSize - 2) / 2) << 3) | CTR_FLAGS_BYTE);
}


/** @internal
    Inform cipher that he is manager or not.

    @param self    The cipher instance. This must not be NULL.
    @param manager TRUE/FALSE for manager/not.
**/
void OALCipher_SetManager( OALCipher self, boolean manager )
{
    PCR_Assert( self );
    if (manager && OALCipher_IsSecureGroup( self ))
    {
        OALCipher_ResetQuickNodeStore( self );                                       /* Useful only for manager. */
        if (manager && !self->flags.manager)
        {
            self->flags.needNewBaseNodeID = 1;                                       /* New SGM CCM managers do this. */
            /* All the actual nodeID related work is accomplished in OALCipher_FormNewAndSetSlotOrKEKKey(). Until that
               is called, this node may operate with old ->advNNI, and similar. */
        }
        if (self->flags.noFirstNodeIDAssignment)
        {
            /* I am first member and first SGM in a new group. */
            self->advNNI = 0;               /* Anything <= TOP_NODEID, OALCipher_FormNewAndSetSlotOrKEKKey uses this. */
            CheckBlockedMessageQueue( self );
        }
    }
    self->flags.manager = manager;
}


/** @internal
    Test that given cipher is/not for a secureGroup.

    @param self The cipher. This must not be NULL.
    @return TRUE/FALSE for cipher secureGroup/not.
**/
boolean OALCipher_IsSecureGroup( const OALCipher self )
{
    PCR_Assert( self ); /* Spec says "secure group" refers to both secure multicast groups and secure unicast groups. */

    return (!self->flags.implicitNodeID && !self->flags.ignoreSlot && !self->flags.implicitCount);
}


/** @internal
    Test that given cipher is/not configured as hub.

    @param self The cipher. This must not be NULL.
    @return TRUE/FALSE for cipher configured/not as hub.
**/
boolean OALCipher_IsHub( const OALCipher self )
{
    PCR_Assert( self );

    return self->flags.hub;
}


OALChannel OALCipher_GetConnectionChannel( const OALCipher self )
{
    return self->connectionChannel;
}


/** @internal
    From given cipher and key, copy stored nonce state of given hisNodeID/hisSlot to given structure.

    @param cipher          The given cipher. This must not be NULL.
    @param periodOffset    Offset to the key held in cipher. This must be < KEYCOUNT.
    @param hisSlot         The given slot.
    @param hisNodeID       The given nodeID.
    @param pFillNonceState Pointer to the given structure. This must not be NULL.
    @return TRUE if stored nonce state found (and the given structure is filled), or FALSE otherwise.
**/
boolean OALCipher_FillNonceStateFromStorage( OALCipher cipher, uint8 periodOffset, uint16 hisSlot, uint32 hisNodeID,
                                             NonceState_t * pFillNonceState )
{
    /* Called with cipher mutex locked. */
    const SlotState_t * pSlotState;

    PCR_Assert( periodOffset < KEYCOUNT );
    PCR_Assert( cipher->periodStore[periodOffset].period < PERIODCOUNT );
    PCR_Assert( cipher->periodStore[periodOffset].enabled );
    if (cipher->periodStore[periodOffset].slotTree &&
        (pSlotState = (const SlotState_t *)OALAVLTree_Retrieve( cipher->periodStore[periodOffset].slotTree,
                                                                (OALAVLTreeKey_t)&hisSlot )) != NULL)
    {
        const NodeIDState_t * pNodeIDState;

        PCR_Assert( pSlotState->slot == hisSlot );
        pFillNonceState->slot = hisSlot;
        if (pSlotState->nodeIDTree && (pNodeIDState = (const NodeIDState_t *)
                                    OALAVLTree_Retrieve( pSlotState->nodeIDTree, (OALAVLTreeKey_t)&hisNodeID )) != NULL)
        {
            PCR_Assert( pNodeIDState->nodeID == hisNodeID );
            pFillNonceState->nodeID = hisNodeID;
            pFillNonceState->packetNumber = pNodeIDState->packetNumber;
            return TRUE;
        }
    }

    return FALSE;
}


/** @internal
    Store given nonce state into the given period storage of the given cipher instance.

    @param cipher       The cipher to accept nonceState. This must not be NULL.
    @param periodOffset Offset of period storage that is associated with nonceState. This must be < KEYCOUNT.
    @param pNonceState  Pointer to the given nonce state. This must not be NULL.
    @return TRUE for state stored, FALSE for not stored.
**/
boolean OALCipher_UpdateStoredNonceState( OALCipher cipher, uint8 periodOffset, const NonceState_t * pNonceState )
{
    /* Called with cipher mutex locked. */
    SlotState_t *   pStoredSlotState;
    NodeIDState_t * pStoredNodeIDState;

    PCR_Assert( periodOffset < KEYCOUNT );
    PCR_Assert( cipher->periodStore[periodOffset].period < PERIODCOUNT );
    PCR_Assert( cipher->periodStore[periodOffset].enabled );
    if (!cipher->periodStore[periodOffset].slotTree)
    {
        PCR_Assert( cipher->periodStore[periodOffset].countHeldState == 0 );
        if ((cipher->periodStore[periodOffset].slotTree =
             OALAVLTree_Create( (OALAVLTree_CompareGreaterThan_fn_t)slot_gt_fn,
                                (OALAVLTree_CompareEqual_fn_t)slot_eq_fn, NULL )) == NULL)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG,
                                "OALCipher: ERROR: update stored nonce state: Failed to create slot avl tree." );
            return FALSE;
        }
    }
    if ((pStoredSlotState = (SlotState_t *)OALAVLTree_Retrieve( cipher->periodStore[periodOffset].slotTree,
                                                                (OALAVLTreeKey_t)&pNonceState->slot )) == NULL)
    {
        /* Insert a now newly created SlotState_t structure into cipher->periodStore[periodOffset].slotTree. */
        if ((pStoredSlotState = DOFMem_Alloc( sizeof( SlotState_t ) )) == NULL)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG,
                                "OALCipher: ERROR: update stored nonce state: Failed to create slot state." );
            return FALSE; /*We leave safely. Whatever we create above will be destroyed at prune. Same each case below*/
        }
        /* Initialize all elements of key before inserting. */
        PCRMem_Set( pStoredSlotState, 0, sizeof( SlotState_t ) );
        pStoredSlotState->slot = pNonceState->slot;
        if ((OALAVLTree_Insert( cipher->periodStore[periodOffset].slotTree, (OALAVLTreeKey_t)&pNonceState->slot,
                                (OALAVLTreeItem_t)pStoredSlotState, NULL )) != OALAVLTree_InsertSuccess)
        {
            DOFMem_Free( (SlotState_t *)pStoredSlotState );
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG,
                                "OALCipher: ERROR: update stored nonce state: Failed to insert stored slot state." );
            return FALSE;
        }
    }
    else
    { /* Already stored in tree. */
    }
    PCR_Assert( pStoredSlotState->slot == pNonceState->slot );
    if (!pStoredSlotState->nodeIDTree &&
        (pStoredSlotState->nodeIDTree = OALAVLTree_Create( (OALAVLTree_CompareGreaterThan_fn_t)nodeID_gt_fn,
                                                           (OALAVLTree_CompareEqual_fn_t)nodeID_eq_fn, NULL )) == NULL)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG,
                            "OALCipher: ERROR: update stored nonce state: failed to create nodeID avl tree." );
        return FALSE;
    }
    if ((pStoredNodeIDState = (NodeIDState_t *)OALAVLTree_Retrieve( pStoredSlotState->nodeIDTree,
                                                                    (OALAVLTreeKey_t)&pNonceState->nodeID )) == NULL)
    {
        /* Insert a now newly created NodeIDState_t structure into pStoredSlotState->nodeIDTree. */
        if ((pStoredNodeIDState = DOFMem_Alloc( sizeof( NodeIDState_t ) )) == NULL)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG,
                                "OALCipher: ERROR: update stored nonce state: Failed to create nodeID state." );
            return FALSE;
        }
        /* Initialize all elements of key before inserting. */
        PCRMem_Set( pStoredNodeIDState, 0, sizeof( NodeIDState_t ) );
        pStoredNodeIDState->nodeID = pNonceState->nodeID;         /*New nodeID stored in a slot so incr countHeldState*/
        if ((OALAVLTree_Insert( pStoredSlotState->nodeIDTree, (OALAVLTreeKey_t)&pNonceState->nodeID,
                                (OALAVLTreeItem_t)pStoredNodeIDState, NULL )) != OALAVLTree_InsertSuccess)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG,
                                "OALCipher: ERROR: update stored nonce state: Failed to insert nodeID state." );
            DOFMem_Free( (NodeIDState_t *)pStoredNodeIDState );
            return FALSE;
        }
        cipher->periodStore[periodOffset].countHeldState++;
        cipher->countHeldState++;
        /* Can 32 bit countHeldState overflow and be 0 now? The assert below would fail, and more bad things. */
    } /* otherwise already stored in tree. */
    PCR_Assert( pStoredNodeIDState->nodeID == pNonceState->nodeID );
    PCR_Assert( cipher->periodStore[periodOffset].countHeldState > 0 );
    PCR_Assert( cipher->countHeldState > 0 );
    PCR_Assert( cipher->countHeldState >= cipher->periodStore[periodOffset].countHeldState );
    pStoredNodeIDState->packetNumber = pNonceState->packetNumber; /* The only entry updated at every successful call. */

    return TRUE;
}


/* @internal
    To all held keys, promote/demote/disable whatever is due.
*/
void OALCipher_AdjustKeys( OALCipher cipher )
{
    /* Do NOT enter with cipher mutex locked, helgrind showed that it leads to rare out of order mutex locking issues.
       This judiciously locks and unlocks cipher mutex. */
    uint32  nowTime;
    uint16  promoteEpoch                = 0;
    boolean kekKeyPromotedNeedsToNotify = FALSE;

    OALCipher_Lock( cipher );
    nowTime = PCRTime_GetMS();
    /* This next call is old key disable at 2*Thb after a curKey becomes an oldKey. */
    if (cipher->flags.oldKeyTimeoutEnabled && DOFTime_Compare( nowTime, cipher->oldKeyDisableTime ) >= 0)
        OALCipher_DisableOldKeys( cipher );                        /* Within: cipher->flags.oldKeyTimeoutEnabled = 0; */
    if (DOFTime_Compare( nowTime, cipher->nextOrdinaryPruneTime ) >= 0)/*Not prune every call- expensive at hi thruput*/
    {
        PruneAllAgedKeys( cipher, cipher->secsStandardPruneOrKeyDisable );                /* All-key 2*Thb pruning .. */
        cipher->nextOrdinaryPruneTime = nowTime + ORDINARY_PRUNE_CHECK_DELAY;             /* .. once per second. */
    }
    OALCipher_PruneToWatermarks( cipher );
                                /*Returns quickly when work not required, so no problem calling this each time through*/
    if (OALCipher_TaggedKeyEnabled( cipher, KEKKEY ))
    {
        uint16 kekKeyPromoteSlot = KEY_EFFSLOT( cipher, KEKKEY );
        uint16 kekKeyNowSlot     = (nowTime - PERIOD_BIRTHTIME( cipher, KEKKEY )) / 1000;

        /* This assert is true because we never set a kekKey with empty curKey. */
        PCR_Assert( OALCipher_TaggedKeyEnabled( cipher, CUR ) );
        if (DOFTime_Compare_15Bit( kekKeyNowSlot, kekKeyPromoteSlot ) >= 0)
        {
            cipher->lastSlotWithTX = 0;                                /* Do promote kekKey to curKey. */
            promoteEpoch = KEY_EPOCH( cipher, KEKKEY );
            if (KEY_SECUREGROUP( cipher, KEKKEY ))
                kekKeyPromotedNeedsToNotify = TRUE;
            KeyTagAndPeriodOffsetEnable( cipher, CUR, PERIOD_OFFSET( cipher, KEKKEY ) );
            KeyUntag( cipher, KEKKEY );
            CurKeyBasicSetup( cipher );
            cipher->periodStore[PERIOD_OFFSET( cipher, CUR )].dofKey.promoted = TRUE;
            PERIOD_SET_PROMOTETIME( cipher, CUR, nowTime );
            TaggedKeyDisable( cipher, SLOT );                 /* Spec: promoting kekKey cancels out slotKey. */
            PCR_Assert( PERIOD_OFFSET_ENABLED( cipher, CUR ) );
            if (cipher->kekKeySecurityMode)
            {
                DOFSecurityMode_Destroy( cipher->normalSecurityMode );   /* Promote kekKeySecurityMode. */
                cipher->normalSecurityMode = cipher->kekKeySecurityMode; /*We already retained our knowledge of it*/
                cipher->kekKeySecurityMode = NULL;                       /* Remove our knowledge of it in this var*/
            }
            else if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
            {
                /* There should always be a kekKeySecurityMode available at the moment that a kekKey promotes. */
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG,
                                "OALCipher: ERROR: kekKeySecurityMode is NULL at moment kekKey promoted." );
            }
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "kekKey promoted of epoch 0x", promoteEpoch );
            /* Nothing more to do, since slot key was disabled. */
        }
    }
    else if (OALCipher_TaggedKeyEnabled( cipher, SLOT ))
    {
        uint16 slotKeyPromoteSlot = KEY_EFFSLOT( cipher, SLOT );

        if (!(slotKeyPromoteSlot & 0x8000)) /* TEP flag indicates auto promotion is disabled. */
        {
            uint16 slotKeyNowSlot     = (nowTime - PERIOD_BIRTHTIME( cipher, SLOT )) / 1000;

            /* This assert is true because we never set an auto-promoting slotKey with empty curKey. */
            PCR_Assert( OALCipher_TaggedKeyEnabled( cipher, CUR ) );
            if (DOFTime_Compare_15Bit( slotKeyNowSlot, slotKeyPromoteSlot ) >= 0)
            {
                cipher->lastSlotWithTX = 0;                                         /* Promote slotKey to curKey. */
                KeyTagAndPeriodOffsetEnable( cipher, CUR, PERIOD_OFFSET( cipher, SLOT ) );
                KeyUntag( cipher, SLOT );
                CurKeyBasicSetup( cipher );
                cipher->periodStore[PERIOD_OFFSET( cipher, CUR )].dofKey.promoted = TRUE;
                PERIOD_SET_PROMOTETIME( cipher, CUR, nowTime );
                PCR_Assert( PERIOD_OFFSET_ENABLED( cipher, CUR ) );
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "slotKey promoted at promoteSlot 0x",
                                          slotKeyPromoteSlot );
            }
        }
    }
    OALCipher_Unlock( cipher );
    if (kekKeyPromotedNeedsToNotify)                       /* helgrind showed cipher mutex cannot be locked for this. */
        DOFSecureGroup_KEKKeyPromoted( KEY_SECUREGROUP( cipher, KEKKEY ), promoteEpoch );
}


/** @internal
    Remove one slot of nonce state storage held by the given cipher, based on attributes of a packet in hand.

    @param cipher              The given cipher. This must not be NULL.
    @param keyTypePacketInHand Key type of packet in hand (if not CUR, SLOT, or KEKKEY, the key is an oldKey).
    @param slotPacketInHand    Slot of packet in hand.
    @return TRUE/FALSE for pruning did/not happen.
**/
boolean OALCipher_PruneOneSlotStrategy( OALCipher cipher, uint8 keyTypePacketInHand, uint16 slotPacketInHand )
{
    /* Called with cipher mutex locked. */
    /* Called by _Decipher(), and only if nonceState cannot be stored. */
    uint8  periodOffset;                                       /* Only the unit test requires this not be static. */
    uint16 curKeySlot;

    for (periodOffset = 0; periodOffset < KEYCOUNT; periodOffset++)     /* Eliminate one oldKey and exit. */
    {
        if (OALCipher_KeyIsOldKey( cipher, periodOffset ))
        {
            EliminateKeyUseByPeriodOffset( cipher, periodOffset );
            return TRUE;
        }
    }
    if (OALCipher_GetCurKeySlot( cipher, &curKeySlot ))                 /* Prune one slot of curKey. If pruned, exit. */
    {
        uint16 lowKeptSlot = 1 + ((keyTypePacketInHand != CUR ||
                                   curKeySlot > slotPacketInHand) ? curKeySlot : /* Do not prune curKey's future..*/
                                  slotPacketInHand); /*..or this packet's future*/

        if (PruneStoredNonceStateUpToLowKeptSlot( cipher, PERIOD_OFFSET( cipher, CUR ), lowKeptSlot,
                                                  NO_ENDING_HELD_STATE_COUNT_MAXIMUM, TRUE )) /* TRUE: quit at 1st*/
            return TRUE;
    }
    /* Neither oldKey nor curKey was pruned. If packetInHand is SLOT/KEKKEY, prune its oldest slot then exit w/ status*/
    if (keyTypePacketInHand == SLOT || keyTypePacketInHand == KEKKEY)
        if (PruneStoredNonceStateUpToLowKeptSlot( cipher, PERIOD_OFFSET( cipher, keyTypePacketInHand ),
                                                  1 + slotPacketInHand,    /* Do not prune into this packet's future. */
                                                  NO_ENDING_HELD_STATE_COUNT_MAXIMUM, TRUE ))/*TRUE:quit at 1st pruned*/
            return TRUE;

    return FALSE;
}


/** @internal
    Remove nonce state storage entries held by the given cipher, according to watermark strategy.

    @note The strategy: When the cipher's actual held state reaches cipher's highStateWatermark, prune among all held
                        keys one slot at a time (by oldest first) until held state reaches cipher's lowStateWatermark.

    @param cipher The given cipher.
    @return TRUE/FALSE for pruning did/not happen.
**/
boolean OALCipher_PruneToWatermarks( OALCipher cipher )
{
    /* Only the unit test requires this not be static. */
    /* Called with cipher mutex locked. */
    /* Called by OALCipher_AdjustKeys(), which is called at the beginning of every _Encipher() and _Decipher() call. */
    if (cipher->countHeldState >= cipher->highStateWatermark)                        /* Prune by watermarks and exit. */
    {
        boolean pruneHappened = FALSE;

        while (cipher->countHeldState > cipher->lowStateWatermark)
        {
            uint8   periodOffset; /* Find oldest acceptable slot of all enabled keys and delete one slot from that key. */
            uint8   lowAcceptedSlotPeriodOffset = 0;
            uint32  lowAcceptedSlotTime         = 0;
            /* NOTE! This next is set TRUE when lowAcceptedSlotPeriodOffset and lowAcceptedSlotTime are FILLED. And code
                     shows only when they are FILLED are they read for any reason. */
            boolean foundEnabledPromotedKey     = FALSE;
            boolean pruned                      = FALSE;

            for (periodOffset = 0; periodOffset < KEYCOUNT; periodOffset++)
                if (cipher->periodStore[periodOffset].enabled &&
                    cipher->periodStore[periodOffset].dofKey.promoted &&
                    cipher->periodStore[periodOffset].countHeldState)
                {
                    if (foundEnabledPromotedKey)
                    {
                        uint32 newLowAcceptedSlotTime = cipher->periodStore[periodOffset].dofKey.promoteTime +
                                                        (cipher->periodStore[periodOffset].lowAcceptedSlot * 1000);

                        if (DOFTime_Compare( lowAcceptedSlotTime, newLowAcceptedSlotTime ) > 0)
                        {
                            lowAcceptedSlotPeriodOffset = periodOffset; /*periodOffset is new lowAcceptedSlotTime*/
                            lowAcceptedSlotTime = newLowAcceptedSlotTime;
                        }
                    }
                    else
                    {
                        lowAcceptedSlotPeriodOffset = periodOffset;
                        lowAcceptedSlotTime = cipher->periodStore[periodOffset].dofKey.promoteTime +
                                              (cipher->periodStore[periodOffset].lowAcceptedSlot * 1000);
                        foundEnabledPromotedKey = TRUE;
                    }
                }
            if (foundEnabledPromotedKey)
            {
                pruned = PruneStoredNonceStateUpToLowKeptSlot( cipher, lowAcceptedSlotPeriodOffset,
                                                               TOP_LEGAL_SLOT + 1,        /* TRUE: quit at 1st prune. */
                                                               NO_ENDING_HELD_STATE_COUNT_MAXIMUM, TRUE );
                PCR_Assert( pruned );
            }
            else
            {
                /* No promoted key has held state. We MUST prune, and so prune something out of our unpromoted keys. */
                if ( ( OALCipher_TaggedKeyEnabled( cipher, SLOT ) &&
                       cipher->periodStore[periodOffset = PERIOD_OFFSET( cipher, SLOT )].countHeldState ) ||
                     ( OALCipher_TaggedKeyEnabled( cipher, KEKKEY ) &&
                       cipher->periodStore[periodOffset = PERIOD_OFFSET( cipher, KEKKEY )].countHeldState ) )
                {
                    pruned = PruneStoredNonceStateUpToLowKeptSlot( cipher, periodOffset, TOP_LEGAL_SLOT + 1,
                                                                   NO_ENDING_HELD_STATE_COUNT_MAXIMUM,
                                                                   TRUE );    /* TRUE: quit at 1st prune. */
                    PCR_Assert( pruned );
                }
            }
            pruneHappened |= pruned;
        } /* end while(). */
        return pruneHappened;
    }

    return FALSE;
}


/** @internal
    From the given cipher, get slot value of its curKey.

    @param self     The cipher holding a curKey.
    @param pNowSlot Pointer to the memory location that slot is written to.
    @return TRUE for slot value available and written to the specified memory location, FALSE otherwise.
**/
boolean OALCipher_GetCurKeySlot( const OALCipher self, uint16 * pNowSlot )
{
    boolean retBoolean = FALSE;

    PCR_Assert( self );
    OALCipher_AdjustKeys( self );
    OALCipher_Lock( self );
    if (OALCipher_TaggedKeyEnabled( self, CUR ))
    {
        PCR_Assert( PERIOD_OFFSET_ENABLED( self, CUR ) );
        *pNowSlot = OALCipher_GetCurKeyTxSlot( self );
        retBoolean = TRUE;
    }
    OALCipher_Unlock( self );

    return retBoolean;
}


/** @internal
    Ask the given cipher if it needs a new key.

    @note One reason the cipher would need a new key is that its slot is approaching a limit.
    @param self The cipher.
    @return TRUE/FALSE for the cipher needs/not a new key.
**/
boolean OALCipher_NeedRekey( const OALCipher self )
{
    uint16 nowSlot;
    return (OALCipher_GetCurKeySlot( self, &nowSlot ) != FALSE && nowSlot >= EXCESSIVE_SLOT_TRIGGER);
}


/* @internal */
void OALCipher_ResetQuickNodeStore( OALCipher cipher )
{
    if (cipher->quickNodeStore)
    {
        cipher->quickNodeStore->aheadByThreeOrMoreCount = 0;
        cipher->quickNodeStore->aheadByTwoCount = 0;
        cipher->quickNodeStore->notAheadCount = 0;
        cipher->quickNodeStore->beginQuickNodeStoreTime = PCRTime_GetMS();
    }
}


/* @internal */
/* Returns Node ID. */
uint32 OALCipher_StoreAddressNodeIDPair( OALCipher cipher, DOFAddress * pSenderAddress, uint32 nodeID,
                                         boolean denyActualStoreButIssueMemberNodeIDAnyway /* A future optimization.*/ )
{
    AddressToNodeID_t * pStoredAddress;
    boolean             success = FALSE;

    /* If matching *pSenderAddress held in cipher->addressToNodeIDTree, keep/update it and exit. */
    PCRMutex_Lock( addr_mutex );
    pStoredAddress = (AddressToNodeID_t *)OALAVLTree_Retrieve( cipher->addressToNodeIDTree,
                                                               (OALAVLTreeKey_t)&(*pSenderAddress) );
    if (pStoredAddress)
    {
        success = TRUE; /* Found entry holds *pSenderAddress with some nodeID, we will update nodeID if different. */
#if DEEPER_TREE_LOG
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        {
            OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "_StoreAddressNodeIDPair already has DOFAddress at 0x",
                        (uint32)&(*pSenderAddress) );
        }
#endif
        if (nodeID <= TOP_NODEID)
        {
            /* OALCipher_Decipher() or ProcessManagerSuppliedInitialState() called us: consolidate DOFAddresses and
               pass back to EPP caller our instance of *pSenderAddress. WHY:
                1. Consider 2 incoming watch operations from the same node.
                   We consolidate into 1 structure our knowledge of that node's DOFAddress. Without this consolidation,
                   closing out either watch operation (and thus free'ing our storage of that node;s DOFAddress) will
                   call our AddrDown() observable and cause us to remove that DOFAddress from our tree. That is wrong
                   because the 2nd watch operation is still in effect and yet we have lost our tree's DOFAddress entry
                   for that node. So by doing our consolidation work here, that single DOFAddress does not go to ref
                   count 0 until both watch operations close out, and therefore our tree is able to hold the DOFAddress
                   until both ops are gone.
                2. Simple memory savings: We consolidate our allocation of DOFAddress for the same node out there. The
                                          effect is we hold 1 DOFAddress per node out there,
                                                rather than 1 DOFAddress per incoming operation (multiple can exist). */
            DOFAddress tempAddr = DOFAddress_Copy( pStoredAddress->address ); /* Original/consolidated instance (our only instance). */
            DOFAddress_Destroy( *pSenderAddress );      /* Caller's current instance. */
            *pSenderAddress = tempAddr; /* Have caller pass up the stack the consolidated instance. */
        }
        else
        {
            /* OALCipher_GetSGMGroupKey() or OALCipher_RequestReceiveHandler() called us: it is harmful for us to
               do DOFAddress consolidation- they do nothing with the DOFAddress we pass back to them. */
        }
    }
    else
    {
        /* Store brand new *pSenderAddress in tree. */
        if ((pStoredAddress = DOFMem_Alloc( sizeof( AddressToNodeID_t ) )) != NULL)
        {
            AddressObserverParams_t * pAddressObserverParams  = DOFMem_Alloc( sizeof( AddressObserverParams_t ) );

            if (pAddressObserverParams)
            {
                PCRMem_Set( pStoredAddress, 0, sizeof( AddressToNodeID_t ) );
                PCRMem_Set( pAddressObserverParams, 0, sizeof( AddressObserverParams_t ) );
                pStoredAddress->observer = DOFAddress_AddObserver( *pSenderAddress,
                                                                   AddrDown,
                                                                   pAddressObserverParams );
                /* Note that when sender address goes away, AddrDown() is called, which removes it from our
                   ->addressToNodeID list. For hub in particular, when one of its points is dropped by the hub due to
                   timeout, it destroys the point's DOFAddress and this observable mechanism comes into play- so
                   we also remove that point from our list. */
                if (pStoredAddress->observer != NULL)
                {
#if DEEPER_TREE_LOG
                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    {
                        OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                                              "_StoreAddressNodeIDPair before insert has entry num 0x",
                                              OALAVLTree_GetNumItems( cipher->addressToNodeIDTree ) );
                    }
#endif
                    /* Initialize all elements of key before inserting. */
                    pStoredAddress->address = *pSenderAddress;                 /* Fill pStoredAddress for tree use. */
                    pAddressObserverParams->addressToNodeIDTree = cipher->addressToNodeIDTree;
                    pAddressObserverParams->address = *pSenderAddress;                   /* pAddressObserverParams for observer callback. */
                    pStoredAddress->pAddressObserverParams = pAddressObserverParams;     /* Only for memory cleanup. */
                    if (OALAVLTree_Insert( cipher->addressToNodeIDTree, (OALAVLTreeKey_t)&(*pSenderAddress),
                                           (OALAVLTreeItem_t)pStoredAddress, NULL ) == OALAVLTree_InsertSuccess) /* NULL ret. */
                    {
                        success = TRUE;
                        /* I have no reason to retain the sender's DOFAddress. It stays in my tables only as long as
                           anybody else, for their own purposes, needs it to exist. Examples of such purposes: Incoming
                           OAP or other operations will keep the sender's DOFAddress alive until the operation destroys.
                           A point or group member will likely retain their hub/SGM DOFAddress for their ongoing use. */
#if DEEPER_TREE_LOG
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                        {
                            OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                                                  "_StoreAddressNodeIDPair inserted and now has entry num 0x",
                                                  OALAVLTree_GetNumItems( cipher->addressToNodeIDTree ) );
                        }
#endif
                    }
                    else
                        DOFAddress_RemoveObserver( *pSenderAddress, pStoredAddress->observer );
                }
            }
            if (!success)
            {
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                                          "_StoreAddressNodeIDPair() could not create new entry for nodeID 0x",
                                          (uint32)nodeID );
                DOFMem_Free( pStoredAddress );
                DOFMem_Free( pAddressObserverParams );
            }
        }
    }
    if (success)
    {
        if (nodeID > TOP_NODEID)
        {
            /* We are hub or SGM answering ReqGroup PDU. Give member a brand new nodeID, even if he had one before.
               To give him a brand new nodeID even if he had one before is an implementation choice. */
            PCR_Assert( nodeID == 0xFFFFFFFF );
            do
            {
                nodeID = GetNextNodeIDToIssue( cipher );
                PCR_Assert( nodeID <= TOP_NODEID );
            } while (nodeID == cipher->nodeID);
        } /* else Decipher() called us, so fall through to do nothing but update our knowledge of the PDU sender's nodeID. */
        PCR_Assert( pStoredAddress );
        pStoredAddress->nodeID = nodeID;
        /* Note: pStoredAddress->address has a ref count right now, but it might be fully destroyed very soon. When
           that happens (be it soon or late), AddrDown() is called and we remove that DOFAddress from our tree.
               Further note: We might keep a refCount for any particular DOFAddress, as when it is of our SGM, or if
               that node has an interest operation with us. That has a pleasant effect, for AddrDown() is called only
               when that status changes. Without that benefit, we continually add some DOFAddress to pStoredAddress,
               only to remove it soon after when its incoming PDU is fully processed and the DOFAddress is destroyed. */
    }
    PCRMutex_Unlock( addr_mutex );
    PCR_Assert( !success || pStoredAddress->nodeID <= TOP_NODEID );

    return success ? pStoredAddress->nodeID : 0xFFFFFFFF;
}


/*  @internal
    @note On success (TRUE return), every member of pFillGroupKey is filled. Else, none are filled.
*/
boolean OALCipher_GetSGMGroupKey( const OALCipher self, DOFGroupKey_t * pFillGroupKey, DOFAddress requestorAddr )
{
    /* REVIEW: Caller needs to track whether KEKKEY or SLOTKEY are about to promote, and then to ask for that.
               It would need a parameter to specify which key it wants. */
    PCR_Assert( self );
    PCR_Assert( pFillGroupKey );
    OALCipher_AdjustKeys( self );
    OALCipher_Lock( self );
    pFillGroupKey->filled = FALSE;
    if (OALCipher_TaggedKeyEnabled( self, CUR ))
    {
        OALCryptKey key;
        uint8       period;
        uint16      slot;
        uint32      initialStateSGMPacketNumber = 0;
        uint8       keyType                     = CUR;

        if (OALCipher_GetKey( self, keyType, key, &period, &slot ))
        {
            int32  computedTxSlot;
            uint32 targetNodeID = 0;

            if (keyType == CUR)                                        /* Adjust computedTxSlot. */
            {
                computedTxSlot = OALCipher_GetCurKeyTxSlot( self );
                PCR_Assert( self->lastSlotWithTX <= computedTxSlot );
                if (self->lastSlotWithTX == computedTxSlot)     /* OK if, for no tx at all, ->lastSlotWithTX is 0, .. */
                    initialStateSGMPacketNumber = self->encipherPacketNumber; /* because ->encipherPacketNumber is 0. */
            }
            else
                computedTxSlot = slot;
            PCR_Assert( computedTxSlot <= 0x7FFF );
            /* param FFFFFFFF tells this next to do nodeID work for us- look up already stored DOFAddress, or store
               a brand new one. In either case, return a BRAND NEW targetNodeID to us- we use it to fill key
               initialState. We choose to return the BRAND NEW targetNodeID in both cases as an implementation choice.*/
            targetNodeID = OALCipher_StoreAddressNodeIDPair( self, &requestorAddr, 0xFFFFFFFF,
                                                             (self->flags.hub ? FALSE : TRUE) );
            if (targetNodeID <= TOP_NODEID)
            {
                if ((pFillGroupKey->streamCcmModeInitialState =
                                OALCipher_CreateInitialStateStream( self, period,         /* period was filled above. */
                                                                    (uint16)computedTxSlot, initialStateSGMPacketNumber,
                                                                    TRUE, targetNodeID,
                                                                    UNKNOWN_BASE_NODEID /* retrieve baseNodeID from period. */)) != NULL)
                {
                    uint8 periodOffset = PERIOD_OFFSET( self, keyType );

                    pFillGroupKey->filled = TRUE;                              /* Success. */
                    PCRMem_Copy(&pFillGroupKey->key, &self->periodStore[periodOffset].dofKey.key, sizeof(OALCryptKey) );
                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                        OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "Manager issues group member nodeID 0x",
                                              (int)targetNodeID );
                }
                else
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG,
                                    "OALCipher: get SGM group key ERROR: Failed to create initial state stream." );
            }
            else
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG,
                                "OALCipher: ERROR: Manager FAILS to issue group member nodeID." );
        }
    }
    OALCipher_Unlock( self );

    return pFillGroupKey->filled;
}


/* @internal

   streamCcmModeInitialState Must not be NULL.
   streamCcmModeInitialState Must not be of 0 length.
*/
OALStream OALCipher_GetSecureGroupCommonState( OALCipher self, OALStream streamCcmModeInitialState )
{
    ccmModeState_t outCcmModeState;
    OALStream      retStreamCommonCcmModeInitialState = NULL;

    PCR_Assert( self );
    PCR_Assert( streamCcmModeInitialState );
    PCR_Assert( OALStream_GetDataSize( streamCcmModeInitialState ) );
    if ((UncompressCcmModeState( OALStream_GetDataSize( streamCcmModeInitialState ),
                                 OALStream_GetReadPtr( streamCcmModeInitialState ), &outCcmModeState,
                                 FALSE,              /* So we don't reject due to targetNodeID that we now strip out. */
                                 TRUE ) != FALSE))
    {
        PCR_Assert( outCcmModeState.baseNodeID != UNKNOWN_BASE_NODEID );
        outCcmModeState.flags &= ~TRANSMIT_TARGET_NODEID;                      /* Strip out state unique to one node. */
        retStreamCommonCcmModeInitialState = OALCipher_CreateInitialStateStream( self, outCcmModeState.period,
                                                                                 outCcmModeState.slot,
                                                                                 outCcmModeState.packetNumber,
                                                                                 FALSE, 0, outCcmModeState.baseNodeID );
        if (!retStreamCommonCcmModeInitialState)
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE,
                                        "OALCipher: _GetSecureGroupCommonState() create common state failure." );
    }
    else
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE,
                                      "OALCipher: _GetSecureGroupCommonState() is given bad input." );

    return retStreamCommonCcmModeInitialState;
}


OALStream OALCipher_CreateInitialStateStream_TEP( OALCipher self )
{
    return OALCipher_CreateInitialStateStream( self, GET_NEXT_PERIOD, 0, 0, FALSE, 0, 0 );
}


/* @internal

    @mustfree( OALStream_Release)
    @note  nodeID, slot, and packetNumber are always placed. nodeID is that of this cipher instance.
    @note  This is called only by CCM manager.
    @param self             Cipher corresponding to supplied initial state. This may be NULL if cipher not yet created.
    @param period           Supplied period for secureGroup, and must be valid (below PERIODCOUNT). Also must be valid
                            for @self NULL. Otherwise MUST be GET_NEXT_PERIOD, and this call supplies new period.
    @param initialSlot      slot to place in initial state buffer. Can be 0 for TEP.
    @param initialPacketNum Packet number to place in initial state buffer. Can be 0 for TEP.
    @param targetNodeID targetNodeID to place in initial state buffer, if includeTargetNID true. Must be valid.
    @param baseNodeID baseNodeID to place in initial state buffer. Must be valid, and must be 0 for TEP.
    @return OALStream or NULL on failure. @note Will always return NULL if cipher's nodeID is not validated, which can
                            happen only with secureGroup cipher use.
*/
OALStream OALCipher_CreateInitialStateStream( OALCipher self, uint8 period, uint16 initialSlot,
                                              uint32 initialPacketNum, boolean includeTargetNID, uint32 targetNodeID,
                                              uint32 paramBaseNodeID )
{
    OALStream retStreamCcmModeInitialState;

    /* secureGroup always passes in the period. TEP never does. */
    PCR_Assert( !self || !OALCipher_IsSecureGroup( self ) || period < PERIODCOUNT );
    PCR_Assert( !self ||  OALCipher_IsSecureGroup( self ) || period == GET_NEXT_PERIOD );
    PCR_Assert( !self ||  OALCipher_IsSecureGroup( self ) || paramBaseNodeID == 0 );
/* This could give smaller stream size, but having a little bigger stream size is worth it to avoid extra code/time.
    streamCcmModeInitialState = OALStream_CreateRW( OALCipher_GetInitialStateStreamSize( self, getNextPeriod,
                                                                                         pPeriod, initialSlot,
                                                                                         initialPacketNum,
                                                                                         includeTargetNID,
                                                                                         targetNodeID, baseNodeID) );
*/
    if ((retStreamCcmModeInitialState = OALStream_CreateRW( MAX_CCMMODE_INITIAL_STATE_SIZE )) != NULL)
    {
        if (self && IsStreamingConnection( self ))
        {
            PCR_Assert( OALStream_GetDataSize( retStreamCcmModeInitialState ) == 0 );
            /* Return valid but empty stream. Users cannot extract any initial state values from it. */
        }
        else
        {
            uint32 baseNodeID;

            /* secureGroup or TEP datagram. */
            PCR_Assert( paramBaseNodeID == UNKNOWN_BASE_NODEID ||            /* Not specified. */
                        paramBaseNodeID <= 0x3FFFFFFF );                     /* Specified and is compressed 32 bit. */
            if (period < PERIODCOUNT)
            {
                if (self)
                {
                    if (paramBaseNodeID == UNKNOWN_BASE_NODEID)              /* secureGroup. */
                    {
                        uint8 periodOffset = OALCipher_GetPeriodOffset( self, period );

                        baseNodeID = self->periodStore[periodOffset].baseNodeID;/*Get from periodStore, the usual case*/
                    }
                    else
                        baseNodeID = paramBaseNodeID;                        /* Use param. */
                }
                else
                    baseNodeID = 0;
            }
            else
            {
                PCR_Assert( paramBaseNodeID == 0 );                          /* TEP. */
                baseNodeID = 0;
            }
            if (WriteInitialStateStream( self, period, initialSlot, initialPacketNum, includeTargetNID, targetNodeID,
                                         baseNodeID, retStreamCcmModeInitialState ))
            {
                if (includeTargetNID)
                    self->otherNode_nodeID = targetNodeID;           /* Used by _Decipher() in non-secureGroup ONLY. */
            }
            else
            {
                OALStream_Release( retStreamCcmModeInitialState, TRUE );
                retStreamCcmModeInitialState = NULL;
            }
        }
    }
    if (retStreamCcmModeInitialState == NULL)
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG,
                        "OALCipher: ERROR: _CreateInitialStateStream failed to create stream." );

    return retStreamCcmModeInitialState;
}


/* @internal.
   Removes all non key-specific state held by self- my own nodeID, knowledge of other node's nodeIDs and of my manager.
   Key specific state remains- tx slot, tx packet number, and acceptable incoming PDU slot/packetNumber per nodeID.
*/
void OALCipher_RemoveState( OALCipher self )
{
    OALCipher_Lock( self );
    PCRMutex_Lock( addr_mutex );
    if (self->addressToNodeIDTree)
    {
        OALAVLTree_Destroy( self->addressToNodeIDTree, CleanupStoredAddressEntries, NULL );
        /* Gone. So make us a new, empty tree. */
        self->addressToNodeIDTree = OALAVLTree_Create(
                                                     (OALAVLTree_CompareGreaterThan_fn_t)OALCipher_AddressIP_Port_gt_fn,
                                                       (OALAVLTree_CompareEqual_fn_t)OALCipher_AddressIP_Port_eq_fn,
                                                       NULL );
        if (!self->addressToNodeIDTree)
        {
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALCipher: OALCipher_RemoveState() failed to allocate memory." );
        }
    }
    PCRMutex_Unlock( addr_mutex );
    self->flags.nodeIDValid = 0;
    self->flags.nodeIDInWarning = 0;
    DOFAddress_Destroy( self->managerAddress );
    self->managerAddress = NULL;
    OALCipher_Unlock( self );
}


/*  @internal.
    @param self The cipher. This can be NULL.
    @return TRUE if cipher is not NULL and cipher's nodeID is validated, FALSE otherwise.
*/
boolean OALCipher_IsNodeIDValid( const OALCipher self )
{
    return self ? (self->flags.nodeIDValid) : FALSE;
}


/* @internal */
boolean OALCipher_SecureGroupGetNewNodeID( OALChannel connectionChannel )
{
    OALCipher cipher = connectionChannel->remoteChannelData->cipher;

    PCR_Assert( cipher );
    PCR_Assert( OALCipher_IsSecureGroup( cipher ) );
    OALCipher_GetNewNodeID( cipher );

    return TRUE;
}


/** @internal

    @param self    Cipher
    @param address Sender's DOFAddress.
    @return TRUE/FALSE for did/NOT process the delivered information.
**/
boolean OALCipher_QueryReceiveHandler( const OALCipher self, DOFAddress address )
{
    PCR_Assert( self );
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                     "CCM Query received- sending back secure unicast EPP heartbeat PDU to tell sender of my nodeID 0x",
                              self->nodeID );
    EPP_NegOne_Heartbeat( self->connectionChannel,
                          address,       /* We are always given DOFAddress of Query PDU sender. */
                          FALSE );       /* Query came thru _Decipher() so we do know targetNodeID. */

    return TRUE;
}


/** @internal
    .

    @param self      Cipher
    @param requestOp The CCM.Request PDU operation
    @return TRUE/FALSE for did/NOT answer this PDU.
**/
boolean OALCipher_RequestReceiveHandler( const OALCipher self, DOFOperation requestOp )
{
    PCR_Assert( self );
    PCR_Assert( requestOp );
    if (self->flags.manager && self->connectionChannel && OALCipher_TaggedKeyEnabled( self, CUR ))
    {
        /* param FFFFFFFF tells this next to do nodeID work for us- look up already stored DOFAddress, or store
           a brand new one. In either case, return a new nodeID to us- we use it to fill key initialState. */
        uint32 targetNodeID = OALCipher_StoreAddressNodeIDPair( self, &requestOp->sources->address, 0xFFFFFFFF, FALSE );
                                                                                        /* ->address is of requestor. */

        if (targetNodeID <= TOP_NODEID)
        {
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "Manager issues group member nodeID 0x", (int)targetNodeID );
            ((CCMMODE_V1_OperationContext)(requestOp->context))->nodeID = targetNodeID;
            if (CCMMODE_V1_SendRequestResponse( requestOp ) == DOFERROR_OK)
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE,
                                    "OALCipher: CCM.Request PDU received and answered with CCM.RequestResponse PDU." );
                return TRUE;
            }
        }
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALCipher: CCM.Request PDU received but dropped." );

    return FALSE;
}


/** @internal
    .

    @param self      Cipher
    @param nodeID The nodeID delivered us from the SGM.
    @return TRUE/FALSE for did/NOT accept this PDU.
**/
boolean OALCipher_RequestResponseReceiveHandler( const OALCipher self, uint32 nodeID )
{
    if (!self->flags.manager && nodeID <= TOP_NODEID)
    {
        self->nodeID = nodeID;
        SetNodeIDValid( self );
        CheckBlockedMessageQueue( self );
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                                  "CCM.RequestResponse PDU accepted- updated my nodeID with manager supplied nodeID 0x",
                                  (uint32)self->nodeID );
        return TRUE;
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALCipher: CCM.RequestResponse PDU received but rejected." );

    return FALSE;
}


/* @internal. */
void OALCipher_SetStandardPruneOrKeyDisableDelay( OALCipher self, uint16 secsStandardPruneOrKeyDisable )
{
    self->secsStandardPruneOrKeyDisable = secsStandardPruneOrKeyDisable;
}


/** @internal
    Get the normal security mode of the given OALCipher instance.

    @param self  The cipher instance.
    @return  The normal security mode.
**/
DOFSecurityMode OALCipher_GetNormalSecurityMode( const OALCipher self )
{
    return self->normalSecurityMode;                                             /* Only the unit test requires this. */
}


/** @internal
    Get key currently in use of the given OALCipher instance.

    @param self  The cipher instance.
    @return  The DOFKey currently in use.
**/
DOFKey OALCipher_GetCurrentDOFKey( const OALCipher self )
{
    return &self->periodStore[PERIOD_OFFSET( self, CUR )].dofKey;                /* Only the unit test requires this. */
}


/** @internal
    Get encipher packet number of the given OALCipher instance.

    @param self  The cipher instance.
    @return  The encipher packet number.
**/
uint32 OALCipher_GetEncipherPacketNumber( const OALCipher self )
{
    return self->encipherPacketNumber;                                           /* Only the unit test requires this. */
}


/** @internal
    Lock this cipher instance.

    @param self  The cipher instance.
**/
void OALCipher_Lock( OALCipher self )
{
    PCR_Assert( self );                                       /* Only the unit test requires that this not be static. */
    OALMutexPool_Lock( self );
}


/** @internal
    Unlock this cipher instance.

    @param self  The cipher instance.
**/
void OALCipher_Unlock( OALCipher self )
{
    PCR_Assert( self );                                       /* Only the unit test requires that this not be static. */
    OALMutexPool_Unlock( self );
}


OALCipher OALCipher_Create_TEP( DOFSecurityMode securityMode, boolean connectionless, boolean tepResponder,
                                DOFException * pException )
{
    return OALCipher_Create( securityMode,
                             (CIPHER_CONFIG_USE_GIVEN_NODEID | (connectionless ? CIPHER_CONFIG_POINT_TO_POINT_DEFAULT :
                                                                                 CIPHER_CONFIG_CONNECTED_DEFAULT)),
                             tepResponder ? 1 : 0, NULL, pException );
}


/** @internal
    Create a cipher instance.

    @param  securityMode Security mode configuration to be used for the created OALCipher instance. Should not be NULL.
    @param  configFlags  Set of flags used to configure this cipher instance.
    @param  nodeID       Identifies this node. Must be unique within the set of communicators across a channel.
    @param  connectionChannel NULL except for secureGroup use. For that, used to send query PDU
    @exceptionparam
    @return The cipher instance, or NULL if securityMode is NULL, or securityMode not implemented, or memory allocation
            failed.
**/
OALCipher OALCipher_Create( DOFSecurityMode securityMode, uint8 configFlags, uint32 nodeID,
                            OALChannel connectionChannel, DOFException * pException )
{
    /* Note: If cipher instance is for lossless (streaming), supplied nodeID must be 1 or 0.
                0x00000000 if this node is client (the TEP initiator)
                0x00000001 if this node is server (the TEP responder) */
    /* Lossy unicast does not have this restriction. */
    if (securityMode && DOFSecurityMode_GetModeType( securityMode ) == DOFSECURITYMODETYPE_CCM)
    {
        OALCipher retCipher;

        if ((retCipher = DOFMem_Alloc( sizeof( OALCipher_t ) )) != NULL)
        {
            PCRMem_Set( retCipher, 0, sizeof( OALCipher_t ) );                              /* All members init to 0. */
            retCipher->addressToNodeIDTree = OALAVLTree_Create(
                                                     (OALAVLTree_CompareGreaterThan_fn_t)OALCipher_AddressIP_Port_gt_fn,
                                                     (OALAVLTree_CompareEqual_fn_t)OALCipher_AddressIP_Port_eq_fn,
                                                                NULL );
            retCipher->normalSecurityMode = DOFSecurityMode_Copy( securityMode );
            PCR_Assert( DOFSecurityModeCCM_GetMACSize( retCipher->normalSecurityMode ) >= MINIMUM_MAC_LENGTH );
            PCR_Assert( DOFSecurityModeCCM_GetMACSize( retCipher->normalSecurityMode ) <= MAXIMUM_MAC_LENGTH );
            if (retCipher->addressToNodeIDTree && retCipher->normalSecurityMode)
            {
                boolean success = TRUE;

                /* Do not keep a reference to the channel, since it has (or will have) a reference to us. */
                retCipher->connectionChannel = connectionChannel;
                retCipher->nextTrackedPeriod = PERIODCOUNT - 1;       /* So first used period will be 0. */
                retCipher->nextOrdinaryPruneTime = PCRTime_GetMS() + ORDINARY_PRUNE_CHECK_DELAY;
                retCipher->highStateWatermark = 0xFFFFFFFF;           /* Default behavior: never prune by watermark. */
                retCipher->lowStateWatermark = 0xFFFFFFFF;
                retCipher->flags.implicitNodeID = (boolean)((configFlags & TRANSMIT_NODEID) == 0);
                retCipher->flags.ignoreSlot = (boolean)((configFlags & TRANSMIT_SLOT) == 0);
                retCipher->flags.implicitCount = (boolean)((configFlags & TRANSMIT_COUNT) == 0);
                retCipher->flags.hub = (boolean)((configFlags & CIPHER_IS_HUB) != 0);
                if (configFlags & CIPHER_CONFIG_USE_GIVEN_NODEID)
                {
                    retCipher->nodeID = nodeID;
                    SetNodeIDValid( retCipher );
                }
                if (OALCipher_IsSecureGroup( retCipher ))
                {
                    if (((retCipher->timer = OALTimer_Create(DOF_GetDof()->threadPool, TRUE)) != NULL) &&
                        ((retCipher->getNodeIDItem = OALTimer_CreateItem( retCipher->timer, GetNodeIDTimeout_fn, 1,
                                                                          (void**)&retCipher )) != NULL) &&
                        ((retCipher->quickNodeStore = DOFMem_Alloc( sizeof( QuickNodeStore_t ) )) != NULL))
                    {
                        retCipher->secsStandardPruneOrKeyDisable = 2 * 255;                        /* Max Thb is 255. */
                        InitNodeIDMachine( retCipher );
                        PCRMem_Set( retCipher->quickNodeStore, 0, sizeof( QuickNodeStore_t ) );
                        retCipher->baseNodeID = UNKNOWN_BASE_NODEID;
                        retCipher->advNNI = UNKNOWN_ADVERTISED_NEXT_NODEID;
                        retCipher->previousAdvNNI = UNKNOWN_ADVERTISED_NEXT_NODEID;
                    }
                    else
                        success = FALSE;
                }
                if (success)
                {
                    /* All keys are disabled right now. */
                    return retCipher;
                }
            }
            OALCipher_Destroy( retCipher );                                                              /* Fail out. */
        }
        if (pException)
            *pException = DOFException_Copy( DOFEXCEPTION_INSUFFICIENT_RESOURCES );
    }
    else if (pException)
        *pException = DOFException_Copy( DOFEXCEPTION_BAD_MODE );

    return NULL;
}


boolean OALCipher_SetPruneWatermarks( OALCipher self, uint32 lowStateWatermark, uint32 highStateWatermark )
{
    if (highStateWatermark > lowStateWatermark)
    {
        self->highStateWatermark = highStateWatermark;
        self->lowStateWatermark = lowStateWatermark;
        return TRUE;
    }

    return FALSE;
}

/** @internal
    Destroy a cipher instance.

    @param  self The cipher instance.
**/
void OALCipher_Destroy( OALCipher self )
{
    uint8 periodOffset;

    if (!self)
        return;
    OALCipher_Lock( self );
    PCRMutex_Lock( addr_mutex );
    if (self->addressToNodeIDTree)
        OALAVLTree_Destroy( self->addressToNodeIDTree, CleanupStoredAddressEntries, NULL );
    PCRMutex_Unlock( addr_mutex );
    for (periodOffset = 0; periodOffset < KEYCOUNT; periodOffset++)
    {
        if (self->periodStore[periodOffset].enabled)
        {
            DestroyStoredKeyStateByPeriodOffset( self, periodOffset );/*Call EliminateKeyUseByPeriodOffset()? No need*/
        }
    }
    DOFSecurityMode_Destroy( self->normalSecurityMode );
    DOFSecurityMode_Destroy( self->kekKeySecurityMode );
    OALCipher_Unlock( self );
    if (self->timer)
    {
        OALTimerItem_Destroy( self->getNodeIDItem, TRUE );
        OALTimer_Destroy( self->timer );
    }
    if (self->quickNodeStore)
        DOFAddress_Destroy( self->quickNodeStore->quickestNode.addr );
    DOFMem_Free( self->quickNodeStore );
    DOFAddress_Destroy( self->managerAddress );
    DOFAddress_Destroy( self->addrWaitingForQueryResponse );
    OALMutexPool_Finalize( self );
    DOFMem_Free( self );
}


/* @internal

   @return computed curKey tx slot
*/
uint16 OALCipher_GetCurKeyTxSlot( const OALCipher self )
{
    uint8 curKeyOffset;                                    /* Note: We rely on callers to lock cipher before calling. */
    int32 presentSlot;
    int32 computedTxSlot;

    PCR_Assert( self );
    curKeyOffset = PERIOD_OFFSET( self, CUR );
    PCR_Assert( self->periodStore[curKeyOffset].enabled );
    PCR_Assert( self->periodStore[curKeyOffset].dofKey.promoted );
    presentSlot = (PCRTime_GetMS() - PERIOD_PROMOTETIME( self, CUR )) / 1000;
    /* presentSlot cannot be negative, but .txSlotAdjust can. */
    computedTxSlot = presentSlot + self->periodStore[curKeyOffset].dofKey.txSlotAdjust;
    if (computedTxSlot > 0x00007FFF)
        computedTxSlot = 0x00007FFF;                       /* Implementation choice: deny slot wrap from 0x7FFF to 0. */
    else if (computedTxSlot < 0)
        computedTxSlot = 0;                                /* We cannot adjust slot backwards past 0. */

    return (uint16)computedTxSlot;
}


/** @internal
    Fill the given nonce with the corresponding fields from the given nonceState structure.

    @param nonceState structure holding nodeID, slot, and packetNumber to be placed in nonce.
    @param pFillNonce Pointer to the given nonce.
**/
void OALCipher_FillNonce( NonceState_t nonceState, AESCCM_Nonce_t * pFillNonce )
{
    PCR_Assert( nonceState.slot <= 0x7FFF );
    PCR_Assert( nonceState.packetNumber <= 0x3FFFFFFF );
    nonceState.nodeID = PCRSwap_Net32( nonceState.nodeID );             /* Note: If nonceState is ever passed in by */
    nonceState.packetNumber = PCRSwap_Net32( nonceState.packetNumber ); /* reference, must not swap within nonceState.*/
    nonceState.slot = PCRSwap_Net16( nonceState.slot );
    PCRMem_Set( (uint8 *)pFillNonce, 0, sizeof( AESCCM_Nonce_t ) );
    PCRMem_Copy( pFillNonce->nodeID, (uint8 *)&nonceState.nodeID, sizeof( pFillNonce->nodeID ) );
    PCRMem_Copy( pFillNonce->packetNumber, (uint8 *)&nonceState.packetNumber, sizeof( pFillNonce->packetNumber ) );
    PCRMem_Copy( pFillNonce->slot, (uint8 *)&nonceState.slot, sizeof( pFillNonce->slot ) );
}

void OALCipher_MarshalNonce(uint8* buffer,  const AESCCM_Nonce_t * pFillNonce)
{
    PCRMem_Copy(buffer, pFillNonce->nodeID, 4);
    PCRMem_Copy(buffer+4, pFillNonce->slot, 2);
    buffer[6] = pFillNonce->nonceByteReservedAndUnusedAndMustAlwaysBeZero;
    PCRMem_Copy(buffer+7, pFillNonce->packetNumber, 4);
}

/* @internal */
boolean OALCipher_IsEncrypt( const OALCipher self )
{
    PCR_Assert( self );
    if (self->normalSecurityMode)
        return DOFSecurityModeCCM_IsEncrypt( self->normalSecurityMode );

    return FALSE;
}


/* @internal */
boolean OALCipher_IsAuthenticate( const OALCipher self )
{
    PCR_Assert( self );

    return TRUE;
}


boolean OALCipher_GetStoredTargetNodeID( const OALCipher cipher, DOFAddress unicastTargetAddress,
                                         uint32 * pTargetNodeID )
{
    /* Lookup targetNID from unicastTargetAddress in table kept by this cipher instance.
       Note: self->addressToNodeIDTree is filled by OALCipher_Decipher(). */
    /* This next call appropriately locks and unlocks addr_mutex. */

    return LookupAddressAndNodeID( cipher->addressToNodeIDTree, unicastTargetAddress, pTargetNodeID );
}


boolean OALCipher_IsValid_AdvertisedNNI( OALCipher self )
{
    if (self->advNNI <= TOP_NODEID )
        return TRUE;
    PCR_Assert( self->advNNI == UNKNOWN_ADVERTISED_NEXT_NODEID );

    return FALSE;
}


/*  @internal
    Check whether or not we are ready to be sending a packet.  Related to the blocked send queue (as a way to avoid it).
    @param self The given cipher instance.
    @param unicastTargetAddress   NULL/not: multicast/unicast transport used for transmitting this PDU.
    @param denyPacketReplayProtectionIfSecureUnicastSend FALSE when targetNodeID must be filled on unicast transport TX.
    @return TRUE/FALSE for can/cannot encipher packet to the given target address.
*/
boolean OALCipher_SecureGroupEncipherReady( OALCipher self, DOFAddress unicastTargetAddress, uint32 secureMarkingID,
                                            boolean denyPacketReplayProtectionIfSecureUnicastSend )
{
    PCR_Assert( self );
    PCR_Assert( OALCipher_IsSecureGroup( self ) );
    PCR_Assert( self->addressToNodeIDTree );
    /* secureMarkingID is EPP_NEGONE_NO_SECURE_MARKING_ID EXCEPT for special multicast cases. */
    PCR_Assert( secureMarkingID == EPP_NEGONE_NO_SECURE_MARKING_ID || !unicastTargetAddress ||
                DOFAddress_GetAddressType( unicastTargetAddress ) != DOFADDRESSTYPE_UNICAST );
    if (!self->flags.nodeIDValid)
    {
        /* If I am manager, this instantly fixes my problem. If not, it sends off a CCM.Request PDU. */
        OALCipher_GetNewNodeID( self );
    }
    if (self->flags.nodeIDValid && OALCipher_TaggedKeyEnabled( self, CUR ))
    {
        boolean success = TRUE;

        if (secureMarkingID == EPP_NEGONE_NO_SECURE_MARKING_ID &&         /* Caller not specifying targetNodeID. */
            unicastTargetAddress &&                                       /* This is unicast transmission and ... */
            /* This next line is a new understanding that we agree to in early November 2009. */
            DOFAddress_GetAddressType( unicastTargetAddress ) == DOFADDRESSTYPE_UNICAST &&
            !denyPacketReplayProtectionIfSecureUnicastSend)               /* ... we will place targetNID on the wire. */
        {
            uint32 targetNID;
            /* From unicastTargetAddress, see if targetNodeID is held in table kept by this cipher instance.
               Note: self->addressToNodeIDTree is filled by OALCipher_Decipher(). */
            success = OALCipher_GetStoredTargetNodeID( self, unicastTargetAddress, &targetNID );
            if (!success)
            {
                /* We should always have self->connectionChannel, but the unit test doesn't provide one. */
                if (self->connectionChannel)
                {
                    /* Send CCM.Query PDU so we can gather the needed targetNID. */
                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                        OALCipher_LogMessage( DOFLOGLEVEL_DEBUG,
                                    "_EncipherReady() is missing targetNID, sending unicast Query PDU, my nodeID is 0x",
                                              (uint32)self->nodeID );
                    CCMMODE_V1_SendQuery( self->connectionChannel, unicastTargetAddress );
                    self->addrWaitingForQueryResponse = DOFAddress_Copy( unicastTargetAddress );
                    /* NOTE: We have ONLY one unanswered CCM.Query PDU at a time. Reason: Along with sending out a
                             CCM.Query PDU here, our FALSE return also enables the blockedMessageQueue, which is not
                             emptied until we get our answer back. Only on emptying it are we able to find out that we
                             have another missing targetNodeID (for some other unicastTargetAddress than this one). */
                }
            }
        }
        return success;
    }

    return FALSE;
}


/** @internal **/
void OALCipher_CheckAndSetChannelSecureSendBlocked( OALCipher self )
{
    PCR_Assert( self );
    PCR_Assert( self->connectionChannel );
    OALCipher_Lock( self );
    if (!self->flags.nodeIDValid)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE,
                            "OALCipher: CheckAndSetChannelSecureSendBlocked() sets secureSendBlocked flag." );
        self->connectionChannel->remoteChannelData->flags.SecureSendBlocked = TRUE;
        /* Future work will clear this flag. */
    }
    else
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE,
                     "OALCipher: CheckAndSetChannelSecureSendBlocked() finds NO NEED to set secureSendBlocked flag." );
    OALCipher_Unlock( self );
}


/** @internal **/
void OALCipher_ClearChannelSecureSendBlocked( OALCipher self )
{
    PCR_Assert( self );
    PCR_Assert( self->connectionChannel );
    OALCipher_Lock( self );
    self->connectionChannel->remoteChannelData->flags.SecureSendBlocked = FALSE;
    OALCipher_Unlock( self );
}


/** @internal
    If future key exists, demote curKey to be an old key, then promote future key to be curKey.

    @param self The cipher.
    @param maxSecsToPreviousKeyReceiveDisable  When this delay is satisfied, the newly formed oldKey is disabled.
**/
void OALCipher_PromoteFutureKey( OALCipher self, uint32 maxSecsToPreviousKeyReceiveDisable )
{
    PCR_Assert( self );
    OALCipher_Lock( self );
    if (OALCipher_TaggedKeyEnabled( self, SLOT ))
    {
        /* By tagging a different key as curKey, existing curKey becomes an old key. */
        KeyTagAndPeriodOffsetEnable( self, CUR, PERIOD_OFFSET( self, SLOT ) ); /* Tag slotKey also as curKey, then  ..*/
        KeyUntag( self, SLOT );                        /* .. untag slotKey. This does not disable what is now curKey. */
        CurKeyBasicSetup( self );
        /* This next line here for form. Whatever value is in promoteTime is ok for a future key- they do not use slot*/
        PERIOD_SET_PROMOTETIME( self, CUR, PCRTime_GetMS() );                                     /* slot 0. */
        /* Set promoted bit now, for fast rekeys we cannot rely on above promote time to get it set in time. */
        self->periodStore[PERIOD_OFFSET( self, CUR )].dofKey.promoted = TRUE;
        PCR_Assert( PERIOD_OFFSET_ENABLED( self, CUR ) );
        /* We now have NO futKey, a curKey, and (if we previously had a curKey) an oldKey. */
        if (maxSecsToPreviousKeyReceiveDisable)       /* When shall we disable our oldKey? */
        {
            self->flags.oldKeyTimeoutEnabled = 1;
            self->oldKeyDisableTime = PCRTime_GetMS() + maxSecsToPreviousKeyReceiveDisable * 1000;
        }
        else
            OALCipher_DisableOldKeys( self );
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "OALCipher: FutureKey promoted." );
    }
    OALCipher_Unlock( self );
}


/* @internal.
    Set newKey as future key. It promotes to curKey ONLY (yes: ONLY) when _PromoteFutureKeyNow() is called.

    @param streamCcmModeInitialState Stream of initial state. This must not be NULL. Data size 0 means lossless conn.
    @return TRUE for success, or FALSE if streamCcmModeInitialState is rejected.
*/
boolean OALCipher_SetFutureKey( OALCipher self, OALStream streamCcmModeInitialState, OALCryptKey newKey, DOFPermissionSet localPerms, DOFPermissionSet remotePerms )
{
    /* Note: One way we reject is if streamCcmModeInitialState for lossless has any data held in the stream. */
    ccmModeState_t outCcmModeState;                    /* UncompressCcmModeState() initializes all fields. */
    boolean        losslessConnection;
    uint32         streamDataSize;
    uint8          period;
    uint8          periodOffset;
    boolean        retBoolean = TRUE;

    /* Note: This call does not release or retain streamCcmModeInitialState. */
    /* Note: Future key happens to use slot key for implementation, but does not behave in the same way. This works fine
             because TEP sets a future key and not a slot key, and secure groups set a slot key and not a future key. */
    PCR_Assert( self );
    PCR_Assert( streamCcmModeInitialState );
    PCR_Assert( !OALCipher_IsSecureGroup( self ) );
    losslessConnection = IsStreamingConnection( self );                  /* TRUE means no info in outCcmModeState. */
    streamDataSize = OALStream_GetDataSize( streamCcmModeInitialState );
    if (losslessConnection && streamDataSize > 0)
        return FALSE;
    if (UncompressCcmModeState( streamDataSize, OALStream_GetReadPtr( streamCcmModeInitialState ), &outCcmModeState,
                                FALSE/*TEP has no incoming multicast PDUs*/, FALSE ) == FALSE)
        return FALSE;
    OALCipher_AdjustKeys( self );                    /* Before using newKey promote/demote/disable whatever is due. */
    OALCipher_Lock( self );
    if (OALCipher_TaggedKeyEnabled( self, SLOT ))                          /* This usually does not happen. */
        EliminateKeyUseByPeriodOffset( self, PERIOD_OFFSET( self, SLOT ) );
    period = losslessConnection ? GetCleanNextPeriod( self ) :             /* Streaming conn, track period ourself. */
                                  outCcmModeState.period;
    PCR_Assert( period < PERIODCOUNT );
    periodOffset = GetOpenPeriodOffset( self, period );                    /* May eliminate a previously valid key*/
    PCR_Assert( periodOffset < KEYCOUNT );
    self->periodStore[periodOffset].period = period;
    self->periodStore[periodOffset].slotTree = NULL;   /* Previous slotTree was destroyed when periodOffset disabled. */
    if (!losslessConnection)
        retBoolean = ProcessManagerSuppliedInitialState( self, periodOffset, &outCcmModeState, NULL );
    if (retBoolean)
    {
        if (!self->flags.manager)
        {
            /* In our TEP use, _Decipher() uses this as implicitNodeID, since it not on the wire. */
            self->otherNode_nodeID = losslessConnection ? 1:                         /* from streaming TEP responder. */
                                                          outCcmModeState.nodeID;    /* connectionless TEP responder. */
            if (!self->flags.nodeIDValid)
            {
                /* TEP Responder did not supply MY nodeID. Why not? It may or not be required by spec, but reasonable
                   implementations ought supply it. Regardless of how we come to this place, our choice is to close the
                   connection or to send over a CCM.Request PDU, which, being non-secure, I can send without having a
                   nodeID. This implementation returns FALSE to indicate failure, which ought to close the connection.*/
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE,
                       "OALCipher: _SetFutureKey() fails by unusual reason: TEP responder did not supply my nodeID." );
                retBoolean = FALSE;
                goto failOut;
            }
        }
        if ( !localPerms || ( self->periodStore[periodOffset].localPerms = DOFPermissionSet_Copy( localPerms ) ) != NULL )
        {
            if ( !remotePerms || ( self->periodStore[periodOffset].remotePerms = DOFPermissionSet_Copy( remotePerms ) ) != NULL )
            {
                KeyTagAndPeriodOffsetEnable( self, SLOT, periodOffset );           /* Tag self->keys[newPeriod] as SLOT. */
                KEY_SET_EFFSLOT( self, SLOT, 0xFFFF );                             /* 0x8000+: deny automatic promote. */
                CopyNewKeyAndExpandNewKey( self, SLOT, newKey );
                PCR_Assert( PERIOD_OFFSET_ENABLED( self, SLOT ) );
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                {
                    OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "SetFutureKey accepted, period 0x", period );
                    /*OALCipher_LogMessage( DOFLOGLEVEL_TRACE, "New futKey 1st 4 bytes: 0x", *((unsigned int *)&newKey[0]) );*/
                }
            }
            else
            {
                retBoolean = FALSE;
                DOFPermissionSet_Destroy( self->periodStore[periodOffset].localPerms );
                self->periodStore[periodOffset].localPerms = NULL;
            }
        }
        else
        {
            retBoolean = FALSE;
        }
    }
failOut:
    OALCipher_Unlock( self );

    return retBoolean;
}


/* @internal.
    @note This gets new period and makes pktNum 0 and slot 0 because we know this is brand new key for future use.
    @param
    @param
    @param
    @param epoch              Used only if keyType is KEKKEY.
    @param kekKeySecurityMode Used only if keyType is KEKKEY.
    @param
*/
boolean OALCipher_FormNewAndSetSlotOrKEKKey( OALCipher self, uint16 secsPromoteDelay, uint8 keyType, uint8 epoch,
                                             DOFSecurityMode kekKeySecurityMode, DOFPermissionSet localPerms,
                                             DOFPermissionSet remotePerms, DOFSecureGroup secureGroup )
{
    /* Only called by SGM or hub. */
    OALStream streamCcmModeInitialState;
    /* These next 2 give "may be uninitialized" warning on some compilers, but if actually used they really are filled*/
    uint32    originalAdvertisedNextBaseNodeID = 0;                                     /* Used only for secureGroup. */
    uint32    originalBaseNodeID = 0;                                                   /* Used only for secureGroup. */
    uint8     ccmModeCompressedBuf[MAX_CCMMODE_INITIAL_STATE_SIZE];
    uint8     ccmModeCompressedBufFinalSize = MAX_CCMMODE_INITIAL_STATE_SIZE;
    boolean   retBoolean                    = FALSE;
    boolean   nodeIDPlacedNotValidated;

    PCR_Assert( self );
    if (self->flags.needNewBaseNodeID)
    {
        originalAdvertisedNextBaseNodeID = self->advNNI;
        originalBaseNodeID = self->baseNodeID;
        /* Change self->baseNodeID and self->advNNI. This is only place these should
           change, because only here are we secureGroup and do also create a new period. */
        if (OALCipher_IsValid_AdvertisedNNI( self ))
        {
            /* We have it by one of these means: We are SGM and simply need a new baseNodeID, or
                                                 We are SGM of a brand new secureGroup and selected 0 for it, or
                                                 We just stepped up as SGM and have it from the previous SGM. */
            self->baseNodeID = self->advNNI;     /* Used below when we set our new key. */
        }
        else
        {
            /* We just stepped up as SGM and do NOT have self->advNNI from previous SGM. */
            PCR_Assert( self->advNNI == UNKNOWN_ADVERTISED_NEXT_NODEID );
            PCR_Assert( self->baseNodeID <= TOP_NODEID ); /* We got it from an SGM before in a ReqG resp.*/
            /* We must assume that the group is still somewhere near to what that SGM told me sometime in the past. This
               is a special case, and the only place where use ->baseNodeID that we stored from another SGM. From that,
               compute our new ->baseNodeID, then ->advNNI. */
            self->baseNodeID += MANAGER_BASE_NODEID_ADVANCE_VALUE;
#if (MANAGER_BASE_NODEID_ADVANCE_VALUE <= (MAXIMUM_NEW_ADVERTISED_NEXT_NODEID_OFFSET / 2))
            self->baseNodeID += MANAGER_BASE_NODEID_ADVANCE_VALUE;           /* Again: maybe stored value not latest. */
#endif
            self->baseNodeID &= TOP_NODEID;
        }
        self->advNNI = self->baseNodeID + MANAGER_BASE_NODEID_ADVANCE_VALUE;      /* As SGM, update our self->advNNI. */
        self->advNNI &= TOP_NODEID;
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        {
            OALCipher_LogMessage(DOFLOGLEVEL_TRACE, "_FormNewAndSetSlotOrKEKKey() computes new self->baseNodeID 0x",
                                                    (uint32)self->baseNodeID);
            OALCipher_LogMessage(DOFLOGLEVEL_TRACE, "_FormNewAndSetSlotOrKEKKey() computes new self->advNNI 0x",
                                                    (uint32)self->advNNI);
        }
        self->nextNodeIDToIssue = self->baseNodeID;
        if (!self->flags.nodeIDValid)
            OALCipher_GetNewNodeID( self );
    }
    PCR_Assert( self->baseNodeID <= TOP_NODEID);
    PCR_Assert( OALCipher_IsValid_AdvertisedNNI( self) );
    /* Conclude by setting the key. */
    if (FillInitialStateBuf(self, GET_NEXT_PERIOD, 0, 0, FALSE, 0, self->baseNodeID, /* 0,0: slot, next pktNum of SGM.*/
                            ccmModeCompressedBuf, &ccmModeCompressedBufFinalSize, &nodeIDPlacedNotValidated) != FALSE &&
        (streamCcmModeInitialState = OALStream_CreateRW( ccmModeCompressedBufFinalSize )) != NULL)
    {
        if (OALStream_Write( streamCcmModeInitialState, ccmModeCompressedBufFinalSize, ccmModeCompressedBuf ) != FALSE)
        {
            OALCryptKey newKey;

            PCRSecureRandom_Generate( sizeof( OALCryptKey ), newKey );
            /* On success, either of these set key calls record self->baseNodeID into period information.*/
            if (keyType == SLOT)
                retBoolean = OALCipher_SetSlotKey( self, NULL, streamCcmModeInitialState, newKey, secsPromoteDelay,
                                                   secureGroup, FALSE, FALSE );
            else
            {
                PCR_Assert( keyType == KEKKEY );
                retBoolean = OALCipher_SetKEKKey( self, NULL, streamCcmModeInitialState, newKey, secsPromoteDelay,
                                                  epoch, kekKeySecurityMode, localPerms, remotePerms, secureGroup, FALSE );
            }
        }
        OALStream_Release( streamCcmModeInitialState, TRUE );
    }
    if (self->flags.needNewBaseNodeID)
    {
        if (retBoolean)
        {
            self->flags.needNewBaseNodeID = 0;
        }
        else
        {
            /* Fail out, but reverse what we did above to ->baseNodeID. */
            self->advNNI = originalAdvertisedNextBaseNodeID;
            self->baseNodeID = originalBaseNodeID;
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                                      "_FormNewAndSetSlotOrKEKKey() reverts, self->baseNodeID again back to 0x",
                                      (uint32)self->baseNodeID );
        }
    }

    return retBoolean;
}


/* @internal.
    Set slotKey to promote to curKey at secsPromoteDelay slots in the future.

    @note  secsPromoteDelay above 0x7FFF, or period of newKey same as existing key's period will deny newKey acceptance.
    @param secsPromoteDelay          When this delay is satisfied, slotKey (newKey) promotes to curKey.
    @param streamCcmModeInitialState Initial key state stream. Must not be NULL.
    @param onTieOverwriteCurKeyPeriod Set only when called from ReqG response.
    @return TRUE/FALSE for newKey accepted/not.
*/
boolean OALCipher_SetSlotKey( OALCipher self, DOFAddress * pManagerAddress, OALStream streamCcmModeInitialState,
                              const OALCryptKey newKey, uint16 secsPromoteDelay, DOFSecureGroup secureGroup,
                              boolean onTieOverwriteCurKeyPeriod, boolean stateReceivedByIncomingMulticastPDU )
{
    /* Note: It is convenient to pass in streamCcmModeInitialState rather than a structure of the same information,
             as we are sometimes called when an incoming slotRekey PDU has already formed that stream. */
    uint8          keyDescriptorReplaceType;
    ccmModeState_t outCcmModeState;
    uint32         streamDataSize;
    boolean        accepted = FALSE;

    /* Note: This call does not release or retain streamCcmModeInitialState. */
    PCR_Assert( self );
    PCR_Assert( streamCcmModeInitialState );
    streamDataSize = OALStream_GetDataSize( streamCcmModeInitialState ); /*0 means we do not have lossy transport info*/
    if (!streamDataSize || UncompressCcmModeState( streamDataSize, OALStream_GetReadPtr( streamCcmModeInitialState ),
                                                   &outCcmModeState, stateReceivedByIncomingMulticastPDU,
                                                   OALCipher_IsSecureGroup( self ) ) == FALSE)
        return FALSE;
    OALCipher_AdjustKeys( self );                      /* Before using newKey promote/demote/disable whatever is due. */
    OALCipher_Lock( self );
    keyDescriptorReplaceType = secsPromoteDelay ? SLOT : CUR;
    if (secsPromoteDelay <= 0x7FFF &&
        (!OALCipher_TaggedKeyEnabled( self, keyDescriptorReplaceType ) ||
         self->periodStore[PERIOD_OFFSET( self, keyDescriptorReplaceType )].period != outCcmModeState.period ||
         onTieOverwriteCurKeyPeriod))   /* onTie saves us when we miss heartbeat but SGM gives us key of same period. */
    {
        /* Special case: If period of chosen key type is identical to period of newKey, we accept it only if it came in
           from ReqG response, because in that case we asked for the key and ought to accept it from the SGM. All other
           callers are by incoming rekey. */
        uint8  periodOffset;
        uint32 nowTime = PCRTime_GetMS();

        SetNextTrackedPeriod( self, outCcmModeState.period );/*Adjust ->nextTrackedPeriod, taking this call as command*/
        if (OALCipher_TaggedKeyEnabled( self, CUR ) && secsPromoteDelay > 0)
        {
            /* Do our work on slotKey. Special case when we have both a curKey and a slot key: Overwrite slot key if its
               promote time is farther out than newKey's promote time. Nothing here makes decisions based on whether
               existing slotKey and newKey key values are same or different. */
            uint16 newKeyPromoteSlot = secsPromoteDelay + 0;    /* I believe slotRekey cannot send slot other than 0. */

            PCR_Assert( keyDescriptorReplaceType == SLOT );
            newKeyPromoteSlot = (newKeyPromoteSlot <= 0x7FFF) ? newKeyPromoteSlot : 0x7FFF;
            if (!OALCipher_TaggedKeyEnabled( self, SLOT ) ||
                DOFTime_Compare_15Bit( KEY_EFFSLOT( self, SLOT ), newKeyPromoteSlot ) > 0)
            {
                DOFPermissionSet localPerms = self->periodStore[PERIOD_OFFSET( self, CUR )].localPerms;
                if ( !localPerms || (localPerms = DOFPermissionSet_Copy( localPerms )) != NULL )
                {
                    DOFPermissionSet remotePerms = self->periodStore[PERIOD_OFFSET( self, CUR )].remotePerms;
                    if ( !remotePerms || (remotePerms = DOFPermissionSet_Copy( remotePerms )) != NULL )
                    {
                        /* If enabled, slotKey's promoteSlot is after newKey's. Spec says replace slotKey. */
                        TaggedKeyDisable( self, SLOT );                           /* Does nothing if there is no slotKey. */
                        periodOffset = GetOpenPeriodOffset( self, outCcmModeState.period );
                        PCR_Assert( periodOffset < KEYCOUNT );
                        PCR_Assert( self->periodStore[periodOffset].localPerms == NULL );
                        self->periodStore[periodOffset].localPerms = localPerms;
                        PCR_Assert( self->periodStore[periodOffset].remotePerms == NULL );
                        self->periodStore[periodOffset].remotePerms = remotePerms;
                        KeyTagAndPeriodOffsetEnable( self, SLOT, periodOffset );
                        CopyNewKeyAndExpandNewKey( self, SLOT, newKey );
                        self->periodStore[PERIOD_OFFSET( self, SLOT )].dofKey.promoted = FALSE;
                        PERIOD_SET_BIRTHTIME( self, SLOT, nowTime );           /* Into promoteTime, masquerades as birthTime. */
                        KEY_SET_EFFSLOT( self, SLOT, newKeyPromoteSlot );
                        KEY_SET_SECUREGROUP( self, SLOT, secureGroup );
                        self->periodStore[periodOffset].period = outCcmModeState.period;
                        self->periodStore[periodOffset].slotTree = NULL;
                        /* secureGroup CCM manager always includes baseNodeID in key state, is 0 as default when no flag. */
                        self->periodStore[periodOffset].baseNodeID = outCcmModeState.baseNodeID;
                        PCR_Assert( PERIOD_OFFSET_ENABLED( self, SLOT ) );
                        PCR_Assert( OALCipher_TaggedKeyEnabled( self, SLOT ) );
                        accepted = TRUE;
                    }
                    else
                        DOFPermissionSet_Destroy( localPerms );
                }
            }
            else
            {
                /* newKey's promoteSlot is same slot or after our existing slotKey's promoteSlot, so discard newKey. */
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE,
                                    "OALCipher: slotKey's promoteSlot same or after our existing slotKey." );
            }
        }
        else
        {
            /* Do our work on curKey. */
            ReplaceCurKey(self, newKey, outCcmModeState.period, nowTime); /*May replace valid key or change cur to old*/
            periodOffset = PERIOD_OFFSET( self, CUR );
            /* secureGroup CCM manager always includes baseNodeID in key state, is 0 as default when no flag. */
            self->periodStore[periodOffset].baseNodeID = outCcmModeState.baseNodeID;
            if (OALCipher_TaggedKeyEnabled( self, SLOT ))
            {
                EliminateKeyUseByPeriodOffset( self, PERIOD_OFFSET( self, SLOT ) );   /* Earlier kills later slotKey. */
                PCR_Assert( !OALCipher_TaggedKeyEnabled( self, SLOT ) );
            }
            accepted = TRUE;
        }
        if (accepted)
        {
            accepted = ProcessManagerSuppliedInitialState( self, periodOffset, &outCcmModeState, pManagerAddress );
            if (accepted && !self->flags.manager)
            {
                if (!self->flags.nodeIDValid)
                {
                    /* My nodeID was invalid before this call and streamCcmModeInitialState gave me no new nodeID. */
                    if ( secureGroup && !stateReceivedByIncomingMulticastPDU ) /* Rekey state is not supposed contain a nodeID for me. */
                    {
                        /* We are here by ReqG response and state contains no nodeID for me. Why not? It may or not
                           be required by spec, but reasonable implementations ought supply it. Regardless of how we
                           come to this place, our choice is to return FALSE or to send over a CCM.Request PDU,
                           which, being non-secure, I can send without nodeID. This implementation pro-actively
                           does the latter.*/
                        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE,
                                        "OALCipher: _SetSlotKey() gets ReqG response with no nodeID, pro-actively asking for one");
                        OALCipher_SecureGroupGetNewNodeID( DOFSecureGroup_GetChannel( secureGroup ) );
                    }
                }
            }

        }
    } /* else new key period no different than existing key period, or invalid secsPromoteDelay. */
    OALCipher_Unlock( self );
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, accepted ? "OALCipher: SetSlotKey accepted." :
                                                                    "OALCipher: SetSlotKey rejected." );

    return accepted;
}


/* @internal.
    Set kekKey to promote to curKey at secsPromoteDelay slots in the future.

    @note  secsPromoteDelay above 0x7FFF, or period of newKey same as existing key's period will deny newKey acceptance.
    @param streamCcmModeInitialState Initial key state stream. Must not be NULL.
    @param originatePeriod           T/F: period next in locally tracked sequence, or from @a streamCcmModeInitialState.
    @param secsPromoteDelay          When this delay is satisfied, kekKey (newKey) promotes to curKey.
    @param kekKeySecurityMode        DOFSecurityMode from AS for epoch that kekKey "newKey" is part of. Must not be NULL.
    @return TRUE/FALSE for newKey accepted/not.
*/
boolean OALCipher_SetKEKKey( OALCipher self, DOFAddress * pManagerAddress, OALStream streamCcmModeInitialState,
                             OALCryptKey newKey, uint16 secsPromoteDelay, uint16 kekEpoch,
                             DOFSecurityMode kekKeySecurityMode, DOFPermissionSet localPerms,
                             DOFPermissionSet remotePerms, DOFSecureGroup secureGroup,
                             boolean stateReceivedByIncomingMulticastPDU )
{
    /* Note: It is convenient to pass in streamCcmModeInitialState rather than a structure of the same information,
             as we are sometimes called when an incoming epochRekey PDU has already formed that stream. */
    ccmModeState_t outCcmModeState;
    uint32         streamDataSize;
    uint8          keyDescriptorReplaceType;
    boolean        accepted    = FALSE;
    boolean        doSetKEKKey = TRUE;

    /* Note: This call does not release or retain streamCcmModeInitialState. */
    PCR_Assert( self );
    PCR_Assert( streamCcmModeInitialState );
    PCR_Assert( kekKeySecurityMode );
    streamDataSize = OALStream_GetDataSize( streamCcmModeInitialState ); /*0 means we do not have lossy transport info*/
    if (!streamDataSize || UncompressCcmModeState( streamDataSize, OALStream_GetReadPtr( streamCcmModeInitialState ),
                                                   &outCcmModeState, stateReceivedByIncomingMulticastPDU,
                                                   OALCipher_IsSecureGroup( self ) ) == FALSE)
        return FALSE;
    OALCipher_AdjustKeys( self );                    /* Before using newKey promote/demote/disable whatever is due. */
    OALCipher_Lock( self );
    keyDescriptorReplaceType = secsPromoteDelay ? KEKKEY : CUR;
    if (secsPromoteDelay > 0x7FFF)
        doSetKEKKey = FALSE;
    else if (keyDescriptorReplaceType == KEKKEY && (OALCipher_TaggedKeyEnabled( self, KEKKEY )))
    {
        /* secureGroup NULL only when we are called by unit test. */
        if (!secureGroup || !DOFSecureGroup_LastTwoKEKRekeysFromDifferentSGMs( secureGroup ))
        {
            /* Same SGM delivering last two epochRekey's cannot override same epoch. */
            doSetKEKKey = (boolean)(DOFTime_Compare_16Bit( kekEpoch, KEY_EPOCH(self, KEKKEY )) > 0);
        }
        else
            doSetKEKKey = (boolean)(DOFTime_Compare_16Bit( kekEpoch, KEY_EPOCH(self, KEKKEY )) >= 0);
    }
    if (doSetKEKKey)
    {
        if ( !localPerms || (localPerms = DOFPermissionSet_DeepCopy( localPerms )) != NULL )
        {
            if ( !remotePerms || (remotePerms = DOFPermissionSet_DeepCopy( remotePerms )) != NULL )
            {
                /* secsPromoteDelay is OK, AND (curKey is to be replaced or kekKey replacement is acceptable). */
                uint8  periodOffset;                      /* If enabled, new epoch is after kekKey's epoch. Spec: replace it. */
                uint32 nowTime = PCRTime_GetMS();

                SetNextTrackedPeriod( self, outCcmModeState.period ); /*Adjust ->nextTrackedPeriod taking this call as command*/
                /* Do not EliminateKeyUseByPeriod() until certain that we will actually set a key. */
                if (OALCipher_TaggedKeyEnabled( self, CUR ) && secsPromoteDelay)
                {
                    uint16 newKeyPromoteSlot = secsPromoteDelay + 0;/*I believe epochRekey cannot send first slot other than 0*/

                    PCR_Assert( keyDescriptorReplaceType == KEKKEY );           /* Do our work on kekKey. */
                    newKeyPromoteSlot = (newKeyPromoteSlot <= 0x7FFF) ? newKeyPromoteSlot : 0x7FFF;
                    TaggedKeyDisable( self, KEKKEY );                           /* Does nothing if there is no kekKey. */
                    periodOffset = GetOpenPeriodOffset( self, outCcmModeState.period );
                    PCR_Assert( periodOffset < KEYCOUNT );
                    PCR_Assert( self->periodStore[periodOffset].localPerms == NULL );
                    self->periodStore[periodOffset].localPerms = localPerms;
                    PCR_Assert( self->periodStore[periodOffset].remotePerms == NULL );
                    self->periodStore[periodOffset].remotePerms = remotePerms;
                    KeyTagAndPeriodOffsetEnable( self, KEKKEY, periodOffset );
                    CopyNewKeyAndExpandNewKey( self, KEKKEY, newKey );
                    self->periodStore[PERIOD_OFFSET( self, KEKKEY )].dofKey.promoted = FALSE;
                    PERIOD_SET_BIRTHTIME( self, KEKKEY, nowTime );
                    KEY_SET_EFFSLOT( self, KEKKEY, newKeyPromoteSlot );                   /* Slot at which this key promotes. */
                    KEY_SET_EPOCH( self, KEKKEY, kekEpoch );
                    KEY_SET_SECUREGROUP( self, KEKKEY, secureGroup );
                    self->periodStore[periodOffset].period = outCcmModeState.period;
                    self->periodStore[periodOffset].slotTree = NULL;
                    /* secureGroup CCM manager always includes baseNodeID in key state, is 0 as default when no flag. */
                    self->periodStore[periodOffset].baseNodeID = outCcmModeState.baseNodeID;
                    PCR_Assert( PERIOD_OFFSET_ENABLED( self, KEKKEY ) );
                    PCR_Assert( OALCipher_TaggedKeyEnabled( self, KEKKEY ) );
                    DOFSecurityMode_Destroy( self->kekKeySecurityMode );
                    self->kekKeySecurityMode = DOFSecurityMode_Copy( kekKeySecurityMode );     /* Retain our knowledge of it. */
                    accepted = TRUE;
                }
                else if (!OALCipher_TaggedKeyEnabled( self, CUR ) || self->periodStore[PERIOD_OFFSET( self, CUR )].period != outCcmModeState.period)
                {
                    ReplaceCurKey(self, newKey, outCcmModeState.period, nowTime); /*May replace valid key or convert cur to old*/
                    periodOffset = PERIOD_OFFSET( self, CUR );
                    /* secureGroup CCM manager always includes baseNodeID in key state, is 0 as default when no flag. */
                    self->periodStore[periodOffset].baseNodeID = outCcmModeState.baseNodeID;

                    /* ReplaceCurKey may have saved the permission sets so we need to destroy them before replacing */
                    if( self->periodStore[periodOffset].localPerms )
                        DOFPermissionSet_Destroy(self->periodStore[periodOffset].localPerms);
                    self->periodStore[periodOffset].localPerms = localPerms; 

                    if( self->periodStore[periodOffset].remotePerms )
                        DOFPermissionSet_Destroy(self->periodStore[periodOffset].remotePerms);
                    self->periodStore[periodOffset].remotePerms = remotePerms;

                    if (KEY_SECUREGROUP( self, KEKKEY ))
                        DOFSecureGroup_KEKKeyPromoted( KEY_SECUREGROUP( self, KEKKEY ), kekEpoch );
                    if (OALCipher_TaggedKeyEnabled( self, KEKKEY ))
                    {
                        EliminateKeyUseByPeriodOffset( self, PERIOD_OFFSET( self, KEKKEY ) );
                        PCR_Assert( !OALCipher_TaggedKeyEnabled( self, KEKKEY ) );
                    }
                    if (OALCipher_TaggedKeyEnabled( self, SLOT))
                    {
                        EliminateKeyUseByPeriodOffset( self, PERIOD_OFFSET( self, SLOT ) );    /* Promote kekK, cancel slotK. */
                        PCR_Assert( !OALCipher_TaggedKeyEnabled( self, SLOT ) );
                    }
                    PCR_Assert( self->normalSecurityMode );
                    DOFSecurityMode_Destroy( self->normalSecurityMode );
                    self->normalSecurityMode = DOFSecurityMode_Copy( kekKeySecurityMode );     /* Retain our knowledge of it. */
                    accepted = TRUE;
                }
                else
                {
                    /* New period is same as existing curKey's period, so discard newKey. */
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE,
                                        "OALCipher: New period is equal to held curKey period." );
                    DOFPermissionSet_Destroy( remotePerms );
                    DOFPermissionSet_Destroy( localPerms );
                }
                if (accepted)
                    accepted = ProcessManagerSuppliedInitialState( self, periodOffset, &outCcmModeState, pManagerAddress );
                /*
                 * NOTE: The SGM did not deliver me my nodeID. We are here because the SGM is rekeying me, so
                 * we were NOT supposed to be given a new nodeID. I will not be here in response to a ReqG PDU,
                 * where I should have been given a new nodeID. If something has caused me to lose my nodeID
                 * (if (accepted && !self->flags.manager && !self->flags.nodeIDValid)), it should also have
                 * initiated a ReqG or a CCM.Request PDU.
                 */
            }
            else
                DOFPermissionSet_Destroy( localPerms );
        }
    }
    else
    {
        /* New kekKey epoch is disqualified to replace held kekKey epoch or invalid secsPromoteDelay, so discard newKey. */
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE,
                            "OALCipher: New key period equal to held key period, or invalid secsPromoteDelay." );
    }
    OALCipher_Unlock( self );

    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, accepted ? "OALCipher: SetKEKKey accepted." :
                                                                    "OALCipher: SetKEKKey rejected." );
    }

    return accepted;
}


/*  @internal

    @param streamCcmModeInitialState Initial key state stream. Must not be NULL.
*/
boolean OALCipher_SetCurKey( OALCipher self, DOFAddress *pManagerAddress, OALStream streamCcmModeInitialState,
                             const OALCryptKey newKey, boolean stateReceivedByIncomingMulticastPDU )
{
    /* Called only by DOFSecureGroup_RekeyMergeReceiveHandler() and DOFSecureGroup_InitialJoin(). */
    ccmModeState_t outCcmModeState;
    uint32         nowTime = PCRTime_GetMS();
    uint32         streamDataSize;
    uint8          periodOffset;
    boolean        accepted;

    PCR_Assert( self );                      /* Note: This call does not release or retain streamCcmModeInitialState. */
    PCR_Assert( streamCcmModeInitialState );
    streamDataSize = OALStream_GetDataSize( streamCcmModeInitialState ); /*0 means we do not have lossy transport info*/
    if (!streamDataSize || UncompressCcmModeState( streamDataSize, OALStream_GetReadPtr( streamCcmModeInitialState ),
                                                   &outCcmModeState, stateReceivedByIncomingMulticastPDU,
                                                   OALCipher_IsSecureGroup( self ) ) == FALSE)
        return FALSE;
    OALCipher_AdjustKeys( self );            /* Before using newKey promote/demote/disable whatever is due. */
    OALCipher_Lock( self );
    SetNextTrackedPeriod( self, outCcmModeState.period ); /* Adjust ->nextTrackedPeriod, taking this call as command. */
    /* This may replace a valid key or convert curKey to old key. */
    ReplaceCurKey( self, newKey, outCcmModeState.period, nowTime - (outCcmModeState.slot * 1000) );
    periodOffset = PERIOD_OFFSET( self, CUR);
    /* secureGroup CCM manager always includes baseNodeID in key state, is 0 as default when no flag. */
    self->periodStore[periodOffset].baseNodeID = outCcmModeState.baseNodeID;
    accepted = ProcessManagerSuppliedInitialState( self, periodOffset, &outCcmModeState, pManagerAddress );
    PCR_Assert( PERIOD_OFFSET_ENABLED( self, CUR ) );
    if (accepted)
    {
        if (OALCipher_TaggedKeyEnabled( self, SLOT))
        {
            EliminateKeyUseByPeriodOffset( self, PERIOD_OFFSET( self, SLOT ) );
            PCR_Assert( !OALCipher_TaggedKeyEnabled( self, SLOT ) );
        }
        if (!self->flags.manager && !self->flags.nodeIDValid)
        { /* NOTE: The SGM did not deliver me my nodeID, and I need one. We are here because the SGM is rekeying me, so
                   we were not supposed to be given a new nodeID. We are NOT here in response to a ReqG PDU. Presumably
                   something else has already caused me to seek a new nodeID from my SGM. The complexity here is that my
                   SGM just stepped down, and might not answer me in that way. Must I send CCM.Request to the new SGM?
                   I do not know his DOFAddr yet, so if I do that it must be sent multicast. Another choice: send ReqG*/
        }
    }
    else
        DestroyStoredKeyStateByPeriodOffset( self, PERIOD_OFFSET( self, CUR ) );

    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, accepted ? "OALCipher: SetCurKey accepted." :
                                                                    "OALCipher: SetCurKey rejected." );
        /*OALCipher_LogMessage( DOFLOGLEVEL_TRACE,
                                "_SetCurKey() stored 1st 4 bytes 0x", (uint32)(*(const uint32 newKey) );*/
    }
    OALCipher_Unlock( self );

    return accepted;
}


/* @internal
*/
void OALCipher_DisableAllPendingRekeys( OALCipher self )
{
    OALCipher_Lock( self );
    if (OALCipher_TaggedKeyEnabled( self, SLOT))
    {
        EliminateKeyUseByPeriodOffset( self, PERIOD_OFFSET( self, SLOT ) );
        PCR_Assert( !OALCipher_TaggedKeyEnabled( self, SLOT ) );
    }
    if (OALCipher_TaggedKeyEnabled( self, KEKKEY))
    {
        EliminateKeyUseByPeriodOffset( self, PERIOD_OFFSET( self, KEKKEY ) );
        PCR_Assert( !OALCipher_TaggedKeyEnabled( self, KEKKEY ) );
    }
    OALCipher_Unlock( self );
}


/** @internal
    Disable all of the cipher's old keys.

    @note If present, the old keys may have been available to decipher incoming packets.
    @param self The cipher.
**/
void OALCipher_DisableOldKeys( OALCipher self )
{
    uint8 periodOffset;                            /* Note: an old key is any key not tagged as CUR, SLOT, or KEKKEY. */

    PCR_Assert( self );
    OALCipher_Lock( self );
    self->flags.oldKeyTimeoutEnabled = 0;
    for (periodOffset = 0; periodOffset < KEYCOUNT; periodOffset++)
        if (OALCipher_KeyIsOldKey( self, periodOffset ))
            EliminateKeyUseByPeriodOffset( self, periodOffset );
    OALCipher_Unlock( self );
}


/** @internal
    Get specified key and period of this cipher instance.

    @note          The current key is used to fill the buffer at key, and not the expanded ekey.
    @param self    The cipher instance.
    @param keyType The cipher instance.
    @param key     Pointer to OALCryptKey buffer that is filled. This must not be NULL.
    @param pPeriod Pointer to uint8 buffer that is filled with period of key. This must not be NULL.
    @param pSlot   Pointer to uint16 buffer that is filled with slot of key. This must not be NULL.
    @return        TRUE if specified key is valid, FALSE if not.
**/
boolean OALCipher_GetKey( const OALCipher self, uint8 keyType, OALCryptKey key, uint8 * pPeriod, uint16* pSlot )
{
    boolean retBoolean;

    PCR_Assert( self );
    PCR_Assert( keyType == CUR || keyType == SLOT || keyType == KEKKEY );
    OALCipher_Lock( self );
    retBoolean = OALCipher_TaggedKeyEnabled( self, keyType );
    if (retBoolean)
    {
        PCR_Assert( key && pPeriod && pSlot );
        PCRMem_Copy( key, self->periodStore[PERIOD_OFFSET( self, keyType )].dofKey.key, sizeof( OALCryptKey ) );
        *pPeriod = self->periodStore[PERIOD_OFFSET( self, keyType )].period;
        if (self->periodStore[PERIOD_OFFSET( self, keyType )].dofKey.promoted)
            *pSlot = (PCRTime_GetMS() - PERIOD_PROMOTETIME( self, keyType )) / 1000;
        else
            *pSlot = 0;
    }
    OALCipher_Unlock( self );

    return retBoolean;
}


/** @internal
    Ask if this cipher instance has its curKey set.

    @param self The cipher instance.
    @return TRUE if curKey is set for this cipher instance, FALSE otherwise.
**/
boolean OALCipher_IsKeySet( const OALCipher self )
{
    PCR_Assert( self );

    return self->periodStore[PERIOD_OFFSET( self, CUR )].enabled;
}


/* @internal */
void OALCipher_ClearSecureGroupNotification( OALCipher self, DOFSecureGroup secureGroup )
{
    PCR_Assert( self );
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE,
                        "OALCipher: Cipher called to clear secureGroup notification." );
    if (KEY_SECUREGROUP( self, KEKKEY ) == secureGroup)
        KEY_SET_SECUREGROUP( self, KEKKEY, NULL );
}


/**
@}
**/
