/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_CIPHERPRIVATE_H
#define DOF_CIPHERPRIVATE_H


#include <dof/core.h>
#include <internal/avl_tree.h>
#include <internal/cipher.h>
#include <internal/timer.h>


#ifdef  __cplusplus
extern "C" {
#endif


#define nlen           11                    /* The C OAL is defined by spec that nlen is 11. */
#define q              (15 - nlen )          /* From NIST special pub 800-38C: n+q = 15.    (always) */
/* nlen is nonce length (for users max nonce length). q is length of some field that can describe payload length.
   For q of 4 then, payload is 0 thru 0xFFFFFFFF bytes long. */
#define CTR_FLAGS_BYTE (q - 1 )              /* From same NIST special pub. */
#define A_DATA_PRESENT_FLAG 0x40 /* 0x40 or 0x00. We use 0x40 because cipher (CCM) authenticates both A and P data. */


#define MINIMUM_MAC_LENGTH                  4
#define MAXIMUM_MAC_LENGTH                  16


#define MAX_CCMMODE_INITIAL_STATE_SIZE      19


#define PERIOD_OFFSET_ENABLED( cipher, keyType )(cipher->periodStore[PERIOD_OFFSET( cipher, keyType )].enabled)
#define PERIOD_OFFSET( cipher, keyType )        (cipher->keyDesc[keyType].periodOffset)


#define PERIODCOUNT                         8
#define TAGGEDKEY_COUNT                     3    /* CUR, SLOT, and KEKKEY. */
#define KEYCOUNT                            4    /* Must be above TAGGEDKEY_COUNT, less than or equal to PERIODCOUNT. */
                                                 /* Because we prune after 2 Thb, having all 4 enabled is less likely.*/

#define PERIOD_DISABLED                     255


#define UNKNOWN_ADVERTISED_NEXT_NODEID      0xFFFFFFFF


#define MAXIMUM_NEW_ADVERTISED_NEXT_NODEID_OFFSET 0x01000000             /* Related to spec limit ((2**29 / 32) - 1). */


#define NID_OK      0
#define NID_WARNING 1
#define NID_INVALID 2


#define IsStreamingConnection( cipher )                                                                                \
           ((boolean)(cipher->flags.implicitNodeID) != (boolean)(CIPHER_CONFIG_CONNECTED_DEFAULT & TRANSMIT_NODEID) && \
            (boolean)(cipher->flags.ignoreSlot)     != (boolean)(CIPHER_CONFIG_CONNECTED_DEFAULT & TRANSMIT_SLOT)   && \
            (boolean)(cipher->flags.implicitCount)  != (boolean)(CIPHER_CONFIG_CONNECTED_DEFAULT & TRANSMIT_COUNT))


typedef struct AESCCM_Nonce_t
{
    uint8 nodeID[4];
    uint8 slot[2];  /* Loosely this is secs from key promotion, but individual nodes adjust it for skew compensation. */
    uint8 nonceByteReservedAndUnusedAndMustAlwaysBeZero;
    uint8 packetNumber[4];
} AESCCM_Nonce_t;

typedef struct ccmModeState_t
{
    uint32  nodeID;
    uint32  packetNumber;
    uint32  targetNodeID;
    uint32  baseNodeID;
    uint16  slot;
    uint8   flags;                                 /* Exact copy of original first byte parsed. */
    uint8   period;
    boolean manager;                               /* The rest are not exact copies, but are unmarshaled derivations. */
} ccmModeState_t;


typedef struct DOFKey_t * DOFKey;
typedef struct DOFKey_t
{
    /* We rely on fact that all these members init to 0 at OALCipher_Create(), after which some are filled. */
    OALCryptKey         key;                /* uint8 array- Ks (or Kg ) received from AS. */
    OALCryptExpandedKey ekey;               /* A much longer uint8 array, holding our expanded version of key. */
    uint32              promoteTime;        /* .. but here we store key promote time in msec, then */
    int16               txSlotAdjust;       /* .. we use this to adjust computed tx slot forward/back, but */
    int16               txLastSlot;         /* .. we use this to disallow actual tx slot from ever going backwards. */
    boolean             promoted;           /* FALSE for slotKey and kekKey, promoteTime is birth time. */
    /* Note: promoteTime has 2 roles. promoted false: is birthTime, to calculate when to promote.
                                      promoted true: is promoteTime, to calculate slot, over passing time. */
    /* Note: We do not store slot in this struct (would be uint16 and is 2-byte compressed on the wire), because slot is
       computed as needed. When we rcve a secure packet, slot is typically obtained from the wire. When we send a secure
       packet, we place slot on the wire as secs since our curKey was promoted, modified over real time for skew. */
} DOFKey_t;


typedef struct PeriodStore_t
{
    DOFKey_t   dofKey;
    OALAVLTree slotTree;
    DOFPermissionSet localPerms;
    DOFPermissionSet remotePerms;
    uint32     baseNodeID;
    uint32     countHeldState;                            /* Of nonce state held in slotTree. 0: lowHeldSlot invalid. */
    uint16     lowAcceptedSlot;                           /* Spec calls this oldest Sg that our period "has state for"*/
#if 0 /* Benefit not worth the effort. */
    uint16     lowHeldSlot;                               /* For which we actually have state held and stored. */
#endif
    uint8      period;
    boolean    enabled;                                   /* FALSE: structure never filled or to mark key for non-use.*/
} PeriodStore_t;


typedef struct KeyDesc_t
{
    DOFSecureGroup secureGroup;
    uint16         epoch;                                 /* We need not record epoch for every key we hold. */
    uint16         effSlot;
    uint8          periodOffset;
    boolean        tagged;
} KeyDesc_t;


typedef struct QuickestNode
{
    DOFAddress addr;
    uint16     slotAheadCount;
} QuickestNode;


typedef struct QuickNodeStore_t * QuickNodeStore;
typedef struct QuickNodeStore_t
{
    uint32       beginQuickNodeStoreTime;
    QuickestNode quickestNode;
    uint32       notAheadCount;
    uint32       aheadByTwoCount;
    uint32       aheadByThreeOrMoreCount;
} QuickNodeStore_t;


typedef struct NonceState_t
{
    uint32 nodeID;
    uint32 packetNumber;
    uint16 slot;                                          /* Slot on wire is 2-byte compressed, so max slot is 32767. */
} NonceState_t;


typedef struct OALCipher_t
{
    /* We rely on fact that all these members init to 0 at OALCipher_Create(), after which some are filled. */
    PeriodStore_t   periodStore[KEYCOUNT];
    KeyDesc_t       keyDesc[TAGGEDKEY_COUNT];             /* In order from [0] through [2]: CUR, SLOT, KEKKEY. */
    OALAVLTree      addressToNodeIDTree;
    OALChannel      connectionChannel;
    OALTimer        timer;
    OALTimerItem    getNodeIDItem;
    DOFSecurityMode normalSecurityMode;
    DOFSecurityMode kekKeySecurityMode;
    QuickNodeStore  quickNodeStore;
    DOFAddress      managerAddress;
    DOFAddress      addrWaitingForQueryResponse;
    uint32          baseNodeID;
    uint32          advNNI;
    uint32          previousAdvNNI;
    uint32          nodeID;                               /* My unique ID placed in nonce and often in CCM header. */
    uint32          otherNode_nodeID;                     /* Used non-secureGroup only. */
    uint32          nextNodeIDToIssue;                    /* Used by SGM only. */
    uint32          encipherPacketNumber;
    uint32          oldKeyDisableTime;
    uint32          nextOrdinaryPruneTime;
    uint32          countHeldState;
    uint32          highStateWatermark;
    uint32          lowStateWatermark;
    struct
    {
        unsigned int oldKeyTimeoutEnabled                  : 1;
        unsigned int manager                               : 1;
        unsigned int implicitNodeID                        : 1;
        unsigned int ignoreSlot                            : 1;
        unsigned int implicitCount                         : 1;
        unsigned int hub                                   : 1;
        unsigned int nodeIDValid                           : 1;
        unsigned int needNewBaseNodeID                     : 1;
        unsigned int noFirstNodeIDAssignment               : 1;
        unsigned int sendRequestPDUToNextFoundManager      : 1;
        unsigned int nodeIDInWarning                       : 1;
        unsigned int newNodeIDScheduled                    : 1;
    } flags;
    uint16           lastSlotWithTX;
    uint16           secsStandardPruneOrKeyDisable;
    uint8            nextTrackedPeriod;
} OALCipher_t;


extern DOFError        AESCCM_Encipher                        ( const DOFCipherStrength strength,
                                                                const OALCryptExpandedKey ekey,
                                                                      uint8 macAuthFlags, const AESCCM_Nonce_t * pNonce,
                                                                      uint32 a_dlen, const uint8 * a_data, uint32 dlen,
                                                                      uint8 * data, uint8 mlen, uint8 * mac,
                                                                      boolean encrypt);
extern DOFError        AESCCM_Decipher                        ( const DOFCipherStrength strength,
                                                                const OALCryptExpandedKey ekey, uint8 macAuthFlags,
                                                                const AESCCM_Nonce_t * pNonce, uint32 a_dlen,
                                                                const uint8 * a_data, uint32 dlen, uint8 * data,
                                                                      uint8 mlen, uint8 * mac, boolean encrypt );
extern boolean         OALCipher_TaggedKeyEnabled             ( const OALCipher cipher, uint8 keyType );
extern boolean         OALCipher_GetStoredTargetNodeID        ( const OALCipher cipher, DOFAddress unicastTargetAddress,
                                                                      uint32 * pTargetNodeID );
extern uint16          OALCipher_GetCurKeyTxSlot              ( const OALCipher self );
extern void            OALCipher_FillNonce                    (       NonceState_t nonceState,
                                                                      AESCCM_Nonce_t * pFillNonce );
extern void            OALCipher_MarshalNonce                 (       uint8* buffer,  
                                                                const AESCCM_Nonce_t * pFillNonce);
extern boolean         OALCipher_FillNonceStateFromStorage    (       OALCipher cipher, uint8 periodOffset,
                                                                      uint16 hisSlot, uint32 hisNodeID,
                                                                      NonceState_t * pFillNonceState );
extern uint8           OALCipher_GetPeriodOffset              ( const OALCipher cipher, uint8 period );

extern boolean         OALCipher_TaggedKey_PeriodOffsetMatches( const OALCipher cipher, uint8 keyType,
                                                                uint8 periodOffset );
extern boolean         OALCipher_KeyIsOldKey                  ( const OALCipher cipher, uint8 periodOffset );
extern void            OALCipher_ResetQuickNodeStore          (       OALCipher cipher );
extern boolean         OALCipher_UpdateStoredNonceState       (       OALCipher cipher, uint8 periodOffset,
                                                                      const NonceState_t * pNonceState );
extern uint32          OALCipher_StoreAddressNodeIDPair       (       OALCipher cipher, DOFAddress * pSenderAddress,
                                                                      uint32 nodeID,
                                                                     boolean denyActualStoreButIssueMemberNodeIDAnyway);
extern void            OALCipher_GetNewNodeID                 ( const OALCipher cipher );

extern boolean         OALCipher_LogMessage                   (       DOFLogLevel logLevel, const char * message,
                                                                      uint32 trailingValue );
extern DOFKey          OALCipher_GetCurrentDOFKey             ( const OALCipher self );
extern DOFSecurityMode OALCipher_GetNormalSecurityMode        ( const OALCipher self );
extern uint32          OALCipher_GetEncipherPacketNumber      ( const OALCipher self );
extern boolean         OALCipher_PruneOneSlotStrategy         (       OALCipher cipher, uint8 keyTypePacketInHand,
                                                                      uint16 slotPacketInHand );
extern boolean         OALCipher_PruneToWatermarks                      (       OALCipher cipher );


#ifdef  __cplusplus
}
#endif


#endif /* DOF_CIPHERPRIVATE_H */
