/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/** @internal @addtogroup OALSecureHMAC
@{
**/


#include <dof/core.h>
#include <dof/error.h>
#include <dof/objectid.h>
#include <internal/cipherhmacutils.h>
#include <internal/error.h>
#include <internal/key_request.h>
#include <internal/marshal.h>
#include <internal/securityscope.h>
#include <internal/stream.h>


/*
**
**  "Private" Type and Method Declarations
*/


/*
**  "Internal Public"  Methods.
*/


/** @internal
    This routine is used to create the mac for the Request Group response PDU.

    @param keyRequest   The key request that contains information sent to us in the ReqG PDU.
    @param scopeStream  Stream holding the scope wire order bytes that came in on the ReqG PDU. This MUST not be NULL.
    @param KEK          The Ki key; for secureGroup this is the KEK. This should not be NULL.
    @param groupKey     The group key as provided by the secure group manager. This should not be NULL.
    @param digest       OUT: The digest where the mac is stored. This should not be NULL.
    @param initialState The Cipher initial state. This MUST not be NULL.
    @param epoch        The Epoch for the KEK.
    @param appid        The highest appid known as sent in the Request Group response.
    @param rank         The ability or desire to become the group manager.
    @param domainID     The domain OID this secure group is authorized to operate in. This should not be NULL.
**/
DOFError OALSecureHmac_CreateRequestGroupResponseMac( DOFKeyRequest keyRequest, OALStream scopeStream, OALCryptKey KEK,
                                                      OALCryptKey groupKey, OALSecureHMACDigest digest,
                                                      OALStream initialState, uint16 epoch, uint16 appid, uint8 rank,
                                                      DOFObjectID domainID )
{
    PCR_Assert( scopeStream );
    if (keyRequest && groupKey && KEK && digest && domainID)
    {
        OALStream keyRequestStream;
        DOFError  retError = DOFERROR_INSUFFICIENT_RESOURCES;

        if ((keyRequestStream = OALStream_CreateRW( DOFKeyRequest_MarshallByteLength( keyRequest, TRUE ) )) != NULL)
        {
            OALSecureHMACContext ctx;

            OALSecureHMAC_Start( &ctx, KEK );                                   /* MAC[KEKgm], fully initializes ctx. */
            if (DOFKeyRequest_Marshal( keyRequestStream, keyRequest, TRUE ))    /* From ReqG PDU Request Block. */
            {
                /* Note: Though the creds struct within keyRequest includes a domain member, that is not marshalled. */

                /* Process D block. */
                OALSecureHMAC_Digest( &ctx, DOFObjectID_GetBytesSize( domainID ), DOFObjectID_GetBytes( domainID ) );
                /* Process I block. */
                OALSecureHMAC_Digest( &ctx, 2, (uint8 *)(&epoch) );
                OALSecureHMAC_Digest( &ctx, OALStream_GetDataSize( keyRequestStream ),
                                      OALStream_GetReadPtr( keyRequestStream ));
                OALSecureHMAC_Digest( &ctx, OALStream_GetDataSize( scopeStream ), OALStream_GetReadPtr( scopeStream ));
                /* Process A and Kd block. */
                retError = OALSecureHmac_CompleteRequestGroupResponseMac( &ctx, groupKey, digest, initialState, appid, rank );
            }
            OALStream_Release( keyRequestStream, TRUE );
        }
        return retError;
    }

    return DOFERROR_ILLEGAL_ARGUMENT;
}


DOFError OALSecureHmac_CompleteRequestGroupResponseMac( OALSecureHMACContext * ctx, OALCryptKey groupKey,
                                                        OALSecureHMACDigest digest, OALStream initialState,
                                                        uint16 appid, uint8 rank )
{
    PCR_Assert( ctx );
    PCR_Assert( initialState );
    PCR_Assert( groupKey );
    PCR_Assert( digest );
    {
        uint8  marshalledData[4];
        uint32 size;
        uint32 initialStateSize;

        /* Process A block. */
        initialStateSize = OALStream_GetDataSize( initialState ); /* Initial State size field, is on wire.*/
        PCR_Assert( initialStateSize <= 0x7FFF ); /* This was unmarshaled from a 2-byte compressed, so this must be true. */
        size = OALMarshal_CompressValue( 2, initialStateSize, 2, marshalledData );
        OALSecureHMAC_Digest( ctx, size, marshalledData );
        OALSecureHMAC_Digest( ctx, OALStream_GetDataSize( initialState ), /* Initial State w/o size field.*/
                              OALStream_GetReadPtr( initialState ) );
        PCR_Assert( appid <= 0x7FFF ); /* This was unmarshaled from a 2-byte compressed, so this must be true. */
        size = OALMarshal_CompressValue( 2, appid, 2, marshalledData );   /* Appid. */
        OALSecureHMAC_Digest( ctx, size, marshalledData );
        OALSecureHMAC_Digest( ctx, 1, &rank );                            /* Rank (Desire). */
        /* Process Kd block. */
        OALSecureHMAC_Digest( ctx, sizeof( OALCryptKey ), groupKey );
        OALSecureHMAC_Finish( ctx, digest );                              /*MAC[KEKgm](D || I || A || Kd).*/
        return DOFERROR_OK;
    }
}

DOFError OALSecureHmac_CreateRequestGroupResponseMac_IStream( OALCryptKey KEK, OALCryptKey groupKey, OALSecureHMACDigest digest, uint32 I_len, const uint8* I, OALStream initialState, uint16 appid, uint8 rank, DOFObjectID domainID )
{
    if( groupKey && KEK && digest && domainID && I )
    {
        OALSecureHMACContext ctx;
        DOFError  retError = DOFERROR_INSUFFICIENT_RESOURCES;

        OALSecureHMAC_Start( &ctx, KEK );   /* MAC[KEKgm], fully initializes ctx. */

        /* Note: Though the creds struct within keyRequest includes a domain member, that is not marshalled. */
        /* Process D block. */
        OALSecureHMAC_Digest( &ctx, DOFObjectID_GetBytesSize( domainID ), DOFObjectID_GetBytes( domainID ) );

        /* Process I block. */
        OALSecureHMAC_Digest( &ctx, I_len, I );

        /* Process A and Kd block. */
        retError = OALSecureHmac_CompleteRequestGroupResponseMac( &ctx, groupKey, digest, initialState, appid, rank );
        return retError;
    }

    return DOFERROR_ILLEGAL_ARGUMENT;
}


/**
@}
**/

