/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/cipher.h>


/** @internal @addtogroup OALSecureHash

    Implementation of a part of the FIPS PUB 180-2 Secure Hash Signature
    Standard (SHS).

    This implementation only supports SHA-256.  It is based on code found in
    the public domain modified by Wei Dai, which was in turn written by Steve
    Reid.  Both implementations are in the public domain and part of the
    CryptoToolkit.

    @see    FIPS 180-2 Section 5.1.1, page 12;
            Section 5.2.1, page 13;
            Section 5.3.2, pages 13-14;
            Section 6.2, pages 18-20
@{
**/


/*
** Defines
*/

#define rotrFixed( a, b )    ((a >> b) | (a << (32-b)))
#define blk0(i) (W[i] = data[i])
#define blk2(i) (W[i&15]+=s1(W[(i-2)&15])+W[(i-7)&15]+s0(W[(i-15)&15]))

#define Ch(x,y,z) (z^(x&(y^z)))
#define Maj(x,y,z) ((x&y)|(z&(x|y)))

#define a(i) T[(0-i)&7]
#define b(i) T[(1-i)&7]
#define c(i) T[(2-i)&7]
#define d(i) T[(3-i)&7]
#define e(i) T[(4-i)&7]
#define f(i) T[(5-i)&7]
#define g(i) T[(6-i)&7]
#define h(i) T[(7-i)&7]

#define R(i) h(i)+=S1(e(i))+Ch(e(i),f(i),g(i))+K[i+j]+(j?blk2(i):blk0(i));\
    d(i)+=h(i);h(i)+=S0(a(i))+Maj(a(i),b(i),c(i))

#define S0(x) (rotrFixed(x,2)^rotrFixed(x,13)^rotrFixed(x,22))
#define S1(x) (rotrFixed(x,6)^rotrFixed(x,11)^rotrFixed(x,25))
#define s0(x) (rotrFixed(x,7)^rotrFixed(x,18)^(x>>3))
#define s1(x) (rotrFixed(x,17)^rotrFixed(x,19)^(x>>10))


/* See FIPS PUB 180-2, Section 4.2.2 on page 10 */
static const uint32 K[64] = {
    0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5,
    0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
    0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3,
    0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
    0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc,
    0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
    0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7,
    0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
    0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13,
    0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
    0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3,
    0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
    0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5,
    0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
    0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208,
    0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
};

/* See FIPS 180-2, Section 5.3.2, pages 13-14 */
static const uint32 H0[8] = {
    0x6a09e667,
    0xbb67ae85,
    0x3c6ef372,
    0xa54ff53a,
    0x510e527f,
    0x9b05688c,
    0x1f83d9ab,
    0x5be0cd19
};




/*
**  "Private" Type and Method Declarations
*/
static void SHA256Transform( uint32 *state, const uint32 *data );



/*
**  "Internal Public"  Methods.
*/


/**
    Start a new hash, initializing the hash context.

    @param  ctx     The hash context.
**/
void OALSecureHash_Start( OALSecureHashContext *ctx )
{
    /* Set the initial digest value. See Section 5.3.2. */
    PCRMem_Copy( ctx->working, H0, sizeof(H0) );

    /* Initialize values for 'state', or how much data is in the temp buffer. */
    ctx->stored = 0;
    ctx->length = 0;
}


/**
    Add to the message digest (hash) the message in @a data with length @a
    dlen.  This routine handles all necessary padding.

    @param  ctx     The hash context.
    @param  dlen    The length of the data to hash, in bytes.
    @param  data    The message/data to hash (stored as a byte array).
**/
void OALSecureHash_Digest( OALSecureHashContext *ctx, uint32 dlen, const uint8 *data )
{
    /*
     * Parse the message into 512-bit (64 byte) blocks, and handle the padding
     * as required by section 5.1.1. Note that there may be a partial block
     * already in the context, and that the data added in this call may cause 0
     * or more new blocks to be appended.
     */

    while ( dlen > 0 )
    {
        uint8* pT = (uint8*) ctx->T;
        uint8 i;

        if ( ctx->stored + dlen < 64 )
        {
            PCRMem_Copy( pT + ctx->stored, data, dlen );
            ctx->length += dlen;
            ctx->stored += dlen;
            data += dlen;
            dlen = 0;
        }
        else
        {
            PCRMem_Copy( pT + ctx->stored, data, 64 - ctx->stored );
            data += 64 - ctx->stored;
            dlen -= 64 - ctx->stored;
            ctx->length += 64 - ctx->stored;
            ctx->stored = 0;

            for ( i = 0; i < 16; i++ )
                ctx->T[i] = PCRSwap_Net32( ctx->T[i] );

            SHA256Transform( (uint32*) ctx->working, ctx->T );
        }
    }
}


/**
    Compute and return the final message digest (hash) of the entire message.

    @param  ctx     The hash context.
    @param  digest  [OUT] The resulting hash digest (stored as a byte array).
**/
void OALSecureHash_Finish( OALSecureHashContext *ctx, OALSecureHashDigest digest )
{
    uint32* wptr;
    uint8 remaining;
    uint8 i;

    /* Handle the last block. The required padding either fits or it doesn't. */
    remaining = 64 - ctx->stored;

    if ( remaining >= 9 )
    {
        uint8* ptr = (uint8*) ctx->T;

        /* The padding fits. */
        PCRMem_Set( ptr + ctx->stored, 0, remaining );
        ptr[ ctx->stored ] = 0x80;

        /* Note that we swap this back below. */
        ctx->T[ 15 ] = PCRSwap_Net32( ctx->length * 8 );

        for ( i = 0; i < 16; i++ )
            ctx->T[i] = PCRSwap_Net32( ctx->T[i] );

        SHA256Transform( (uint32*) ctx->working, ctx->T );

        wptr = (uint32*) ctx->working;
        for ( i = 0; i < 8; i++ )
            wptr[i] = PCRSwap_Net32( wptr[i] );

        ptr = ctx->working;
        for ( i = 0; i < sizeof(OALSecureHashDigest); i++ )
            digest[i] = ptr[i];
    }
    else
    {
        uint8* ptr = (uint8*) ctx->T;

        /* The padding doesn't fit - we need an additional block.
         * Note that this assumes that there is at least 1 free byte
         * in T, which there must be because otherwise it would have
         * been wrapped already and we would not be in this case.
         */
        PCR_Assert( remaining >= 1 );
        PCRMem_Set( ptr + ctx->stored, 0, remaining );
        ptr[ ctx->stored ] = 0x80;

        for ( i = 0; i < 16; i++ )
            ctx->T[i] = PCRSwap_Net32( ctx->T[i] );

        SHA256Transform( (uint32*) ctx->working, ctx->T );

        /* The last block is all zeroes except the last word. */
        PCRMem_Set( ctx->T, 0, 64 );

        /* We know the contents, so we avoid a "double swap" like
         * exists above
         */
        ctx->T[ 15 ] = ctx->length * 8;

        SHA256Transform( (uint32*) ctx->working, ctx->T );

        wptr = (uint32*) ctx->working;
        for ( i = 0; i < 8; i++ )
            wptr[i] = PCRSwap_Net32( wptr[i] );

        ptr = ctx->working;
        for ( i = 0; i < sizeof(OALSecureHashDigest); i++ )
            digest[i] = ptr[i];
    }
}


static void SHA256Transform( uint32 *state, const uint32 *data )
{
    uint32 W[16];
    uint32 T[8];
    uint8 j;

    /* Copy context->state[] to working vars */
    PCRMem_Copy(T, state, sizeof(T));
    PCRMem_Set(W, 0, sizeof(W));
    /* 64 operations, partially loop unrolled */
    for (  j=0; j < 64; j += 16 )
    {
        R( 0);
        R( 1);
        R( 2);
        R( 3);
        R( 4);
        R( 5);
        R( 6);
        R( 7);
        R( 8);
        R( 9);
        R(10);
        R(11);
        R(12);
        R(13);
        R(14);
        R(15);
    }
    /* Add the working vars back into context.state[] */
    state[0] += a(0);
    state[1] += b(0);
    state[2] += c(0);
    state[3] += d(0);
    state[4] += e(0);
    state[5] += f(0);
    state[6] += g(0);
    state[7] += h(0);
    /* Wipe variables */
    PCRMem_Set(W, 0, sizeof(W));
    PCRMem_Set(T, 0, sizeof(T));
}


/**
@}
**/
