/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/cipher.h>


/** @internal @addtogroup OALSecureHMAC

    Implementation of a part of the FIPS PUB 198 The Keyed-Hash
    Message Authentication Code (HMAC).
@{
**/


/*
**  "Internal Public"  Methods.
*/


/**
    Start a new HMAC with the given key, initializing the HMAC context.

    @param ctx The HMAC context.
    @param key The key to use with the HMAC.
**/
void OALSecureHMAC_Start( OALSecureHMACContext * ctx, const OALCryptKey key )
{
    uint8 temp[sizeof( OALCryptKey )];
    uint8 i;

    PCRMem_Copy( ctx->key, key, sizeof( OALCryptKey ) );                             /* Fill 1st ctx member. */
    OALSecureHash_Start( &ctx->context );
    PCRMem_Copy( temp, key, sizeof( temp ) );
    for ( i = 0; i < sizeof( temp ); i++ )
        temp[i] ^= 0x36;
    OALSecureHash_Digest( &ctx->context, sizeof( temp ), temp );                     /* Fill 2nd and last ctx member. */
    /* Next follows a block of 0x36. */
    PCRMem_Set( temp, 0x36, sizeof( temp ) );
    OALSecureHash_Digest( &ctx->context, sizeof( temp ), temp );
}


/**
    Add to the HMAC digest the message in @a data with length @a dlen.  This
    routine handles all necessary padding.

    @param ctx  The HMAC context.
    @param dlen The length of the data to hash, in bytes.
    @param data The message/data to hash (stored as a byte array).
**/
void OALSecureHMAC_Digest( OALSecureHMACContext *ctx, uint32 dlen, const uint8 * data )
{
    OALSecureHash_Digest( &ctx->context, dlen, data );
}


/**
    Compute and return the final HMAC digest of the entire message.

    @param  ctx     The HMAC context.
    @param  digest  [OUT] The resulting HMAC digest (stored as a byte array).
**/
void OALSecureHMAC_Finish( OALSecureHMACContext *ctx, OALSecureHMACDigest digest )
{
    uint8 temp[ sizeof(OALCryptKey) ];
    OALSecureHashDigest inner;
    uint8 i;

    OALSecureHash_Finish( &ctx->context, inner );

    /*
     * Now proceed to the 'outer' digest, which is the same secret xor with
     * 0x5C, followed by a block of 0x5C, followed by the inner digest.
     */

    OALSecureHash_Start( &ctx->context );

    PCRMem_Copy( temp, ctx->key, sizeof(temp) );
    for ( i = 0; i < sizeof(temp); i++ )
        temp[i] ^= 0x5C;

    OALSecureHash_Digest( &ctx->context, sizeof(temp), temp );


    /* Next follows a block of 0x5C. */
    PCRMem_Set( temp, 0x5C, sizeof(temp) );
    OALSecureHash_Digest( &ctx->context, sizeof(temp), temp );


    /* Finally the inner digest. */
    OALSecureHash_Digest( &ctx->context, sizeof(inner), inner );

    OALSecureHash_Finish( &ctx->context, digest );
}


/**
@}
**/
