/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/oaltypes.h>
#include <dof/pcr.h>
#include <internal/dofinternal.h>
#include <dof/connection.h>
#include <dof/address.h>
#include <dof/exception.h>
#include <dof/securitymode.h>
#include <dof/error.h>
#include <dof/transport.h>
#include <dof/permissionset.h>
#include <internal/bridgeinternal.h>
#include <internal/transportmanager.h>
#include <internal/channel.h>
#include <internal/connection_internal.h>
#include <internal/connectionless_remotechannel.h>
#include <internal/connectionless_protocolstack.h>
#include <internal/connectionmanager.h>
#include <internal/exception.h>
#include <internal/remotechannel.h>
#include <internal/remotechannelmanager.h>
#include <internal/connected_remotechannel.h>
#include <internal/connection_internal.h>
#include <internal/connected_protocolstack.h>
#include <internal/mutex_pool.h>
#include <internal/permissionsetinternal.h>
#include <internal/securegroupinternal.h>
#include <internal/securityscope.h>
#include <internal/serverinternal.h>
#include <internal/slist.h>
#include <internal/router.h>
#include <internal/dofinternal.h>

static OALTimer connectionTimer;
extern PCRMutex ref_count_mutex;
extern PCRMutex connection_server_mutex;

static void syncConnectCallback( DOFConnectionConnectCallback self, const DOFConnection connection, void * context, const DOFException except );
static const struct DOFConnectionConnectCallbackFns_t syncConnectCBFns = { syncConnectCallback };
static DOFConnectionConnectCallback_t synchronusConnectCB = { &syncConnectCBFns};

static void setupRelatedListeners( DOFConnectionConnectCallback self, const DOFConnection conn, void * context, const DOFException except );
static const struct DOFConnectionConnectCallbackFns_t setupRelatedListenersCBFns = { setupRelatedListeners };
static DOFConnectionConnectCallback_t setupRelatedListenersCB = { &setupRelatedListenersCBFns };

static void relatedConnectionConnected( DOFConnectionConnectCallback self, const DOFConnection conn, void * context, const DOFException except );
static const struct DOFConnectionConnectCallbackFns_t relatedConnCBFns = { relatedConnectionConnected };
static DOFConnectionConnectCallback_t relatedConnConnectedCB = { &relatedConnCBFns };

static void relatedServerStarted( DOFServerStartCallback self, const DOFServer server, void* context, const DOFException except );
static const struct DOFServerStartCallbackFns_t relatedServerCBFns = { relatedServerStarted };
static DOFServerStartCallback_t relatedServerStartedCB = { &relatedServerCBFns };

static void relatedConnectionListener_Fn(DOFConnectionStateListener self, DOFConnection connection, DOFConnectionState state);
static void relatedConnectionListenerRemoved_Fn(DOFConnectionStateListener self, DOFConnection connection, DOFException except );
const struct DOFConnectionStateListenerFns_t relatedConnectionListenerFns = { relatedConnectionListenerRemoved_Fn, relatedConnectionListener_Fn };
struct DOFConnectionStateListener_t relatedConnectionListener = { &relatedConnectionListenerFns };

static void relatedServerListener_Fn(DOFServerStateListener self, DOFServer server, DOFServerState state);
static void relatedServerListenerRemoved_Fn(DOFServerStateListener self, DOFServer server, DOFException except  );
const struct DOFServerStateListenerFns_t relatedServerListenerFns = { relatedServerListenerRemoved_Fn, relatedServerListener_Fn };
struct DOFServerStateListener_t relatedServerListener = { &relatedServerListenerFns };


struct DOFConnectionState_t
{
    uint32                  refCount;
    DOF                     dof;
    DOFCredentials          credentials;
    DOFCredentials          peerCredentials;
    DOFSecurityMode         mode;
    DOFConnectionStatus     status;
    DOFConnectionDirection  direction;
    DOFConnectionConfig     config;
    DOFSecurityMembrane     securityMembrane;
    DOFConnection           relatedDofConnection;
    DOFServer               relatedDofServer;
    DOFException            exception;
    uint64                  byteInputCount;
    uint64                  packetInputCount;
    uint64                  byteOutputCount;
    uint64                  packetOutputCount;

};

typedef struct connectContext_t
{
    PCREvent        evt;
    DOFException    exception;
}connectContext_t;

typedef struct connectRelatedContext_t
{
    pendingConnection_t             *pendConnNode;
    uint32                          timeout;
    uint32                          beginTime;
}connectRelatedContext_t;

typedef struct setupRelatedListenersContext_t
{
    DOFConnectionConnectCallback    callback;
    void                            *context;
}setupRelatedListenersContext_t;

/* Note: this function will not remove the node from the list.  It must be removed from the list before it is destroyed. */
static void DOFConnection_DestroyListenerNode(connLstnNode_t * listenNode)
{
    if (listenNode)
    {
        if (listenNode->removeNode)
        {
            /* If the remove node is still here it is probably because of an error adding the listener.
               Just FREE the remove here without calling the remove callback. */
            DOFMem_Free(listenNode->removeNode);
        }
        DOFMem_Free(listenNode);
    }
}

static connLstnNode_t * DOFConnection_CreateListenerNode(DOFConnectionStateListener listener)
{
    connLstnNode_t *listenerNode;

    listenerNode = (connLstnNode_t *)DOFMem_Alloc(sizeof(connLstnNode_t));
    if (listenerNode)
    {
        PCRMem_Set(listenerNode, 0, sizeof(connLstnNode_t));
        listenerNode->listener = listener;

        if (listener->fns->Removed)
        {
            listenerNode->removeNode = (NotifySystem_ListenerRemoved_t *)DOFMem_Alloc(sizeof(NotifySystem_ListenerRemoved_t));
            if (!listenerNode->removeNode)
            {
                DOFConnection_DestroyListenerNode(listenerNode);
                listenerNode = NULL;
            }
        }
    }
    return listenerNode;
}

/* Note: this function will not remove the node from the list.  It must be removed from the list before it is destroyed. */
static void DOFConnection_DestroyPeerListenerNode(connPeerLstnNode_t * listenNode)
{
    if (listenNode)
    {
        if (listenNode->removeNode)
        {
            /* If the remove node is still here it is probably because of an error adding the listener.
               Just FREE the remove here without calling the remove callback. */
            DOFMem_Free(listenNode->removeNode);
        }
        DOFMem_Free(listenNode);
    }
}

static connPeerLstnNode_t * DOFConnection_CreatePeerListenerNode(DOFConnectionPeerListener listener)
{
    connPeerLstnNode_t *listenerNode;

    listenerNode = (connPeerLstnNode_t *)DOFMem_Alloc(sizeof(connPeerLstnNode_t));
    if (listenerNode)
    {
        PCRMem_Set(listenerNode, 0, sizeof(connLstnNode_t));
        listenerNode->listener = listener;

        if (listener->fns->Removed)
        {
            listenerNode->removeNode = (NotifySystem_PeerListenerRemoved_t *)DOFMem_Alloc(sizeof(NotifySystem_PeerListenerRemoved_t));
            if (!listenerNode->removeNode)
            {
                DOFConnection_DestroyPeerListenerNode(listenerNode);
                listenerNode = NULL;
            }
        }
    }
    return listenerNode;
}

static void DOFConnection_PendingConnectionRetain(pendingConnection_t * pendConn)
{
    PCRMutex_Lock(ref_count_mutex);
    /* If this is about to roll over there will be major problems since each DOFConnection must also retain an OALConnection. */
    PCR_Assert(pendConn->refCount < 0xFFFFFFFF);
    pendConn->refCount++;
    PCRMutex_Unlock(ref_count_mutex);
}

static void DOFConnection_PendingConnectionRelease(pendingConnection_t * pendConn)
{
    uint32 refCount;
    PCRMutex_Lock(ref_count_mutex);
    pendConn->refCount--;
    refCount = pendConn->refCount;
    PCRMutex_Unlock(ref_count_mutex);

    if (!refCount)
    {
        /* Destroy the pending connection node. */
        OALTimerItem_Destroy(pendConn->timerItem, FALSE);
        DOFMem_Free(pendConn->cbNode);
        DOFConnection_Destroy(pendConn->pendConnection);
        DOFMem_Free(pendConn);
    }
}

/* This function is used by OALSList_RemoveAll to enqueue a call back for each pending connection node.
   To enqueue a single complete callback use OALConnectionMgr_enqueueConnectComplete.
   Note: This function must be called with the connection mutex locked. */
void DOFConnection_RemovePendingConnections_fn(OALSListNode item, void *context)
{
    DOFException except = (DOFException)context;
    pendingConnection_t * pendConn = (pendingConnection_t *)item;

    OALConnectionMgr_enqueueConnectComplete(pendConn, except);

    /* Destroy the pending connection node. */
    if (pendConn->timerItem)
    {
        if (OALTimerItem_Remove(pendConn->timerItem))
        {
            /* If the timer item is removed here release the pending connection for the timer context. */
            DOFConnection_PendingConnectionRelease(pendConn);
        }
    }
    /* Release the copy of the pending connection that was stored in the pending connection list. */
    DOFConnection_PendingConnectionRelease(pendConn);
}

static void DOFConnection_ClearRelatedListeners(DOFConnection self)
{
    if (self->relatedConnection && self->listener.connection)
    {
        DOFConnection_RemoveStateListener(self->relatedConnection, &(self->listener.connection->connectionListener));
        self->listener.connection = NULL;
    }
    else if (self->relatedServer && self->listener.server)
    {
        DOFServer_RemoveStateListener(self->relatedServer, &(self->listener.server->serverListener));
        self->listener.server = NULL;
    }
}

static void DOFConnection_removeListeners(OALSListNode item, void *context)
{
    DOFConnection conn = (DOFConnection)context;
    connLstnNode_t *listenNode = (connLstnNode_t *)item;

    OALConnectionMgr_NotifySystem_ListenerRemoved(conn, listenNode);
    DOFConnection_DestroyListenerNode(listenNode);
}

static void DOFConnection_RemovePeerListeners(OALSListNode item, void *context)
{
    DOFConnection conn = (DOFConnection)context;
    connPeerLstnNode_t *listenNode = (connPeerLstnNode_t *)item;

    OALConnectionMgr_NotifySystem_PeerListenerRemoved(conn, listenNode);
    DOFConnection_DestroyPeerListenerNode(listenNode);
}

/*  Remove both the listeners for the related connection or server, and all other listeners. */
void DOFConnection_ClearAllListeners(DOFConnection self)
{
    DOFConnection_ClearRelatedListeners(self);
    OALSList_RemoveAll(&self->listenerList, DOFConnection_removeListeners, self);
    OALSList_RemoveAll(&self->peerListenerList, DOFConnection_RemovePeerListeners, self);
}

/* The connection mutex must be locked when this function is called. */
boolean DOFConnection_AddInternalServerListener(DOFConnection conn)
{
    conn->listener.server = DOFMem_Alloc(sizeof(RelatedServerListener_t));
    if (conn->listener.server)
    {
        PCRMem_Set(conn->listener.server, 0, sizeof(RelatedServerListener_t));
        conn->listener.server->connection = DOFConnection_Copy(conn);
        conn->listener.server->serverListener = relatedServerListener;

        if (!conn->listener.server->connection ||
            !DOFServer_AddStateListener(conn->relatedServer, &(conn->listener.server->serverListener)))
        {
            DOFConnection_Destroy(conn->listener.server->connection);
            DOFMem_Free(conn->listener.server);
            conn->listener.server = NULL;
        }
    }

    if (conn->listener.server)
        return TRUE;

    return FALSE;
}

static void DOFConnection_DestroyRelatedContext(connectRelatedContext_t * ctx)
{
    DOFConnection_PendingConnectionRelease(ctx->pendConnNode);
    DOFMem_Free(ctx);
}

static connectRelatedContext_t * DOFConnection_CreateRelatedContext(DOFConnection conn,
                                                                    pendingConnection_t * pendConn,
                                                                    uint32 timeout)
{
    connectRelatedContext_t *ctx = (connectRelatedContext_t *)DOFMem_Alloc(sizeof(connectRelatedContext_t));

    PCR_Assert(conn);
    PCR_Assert(pendConn);

    if (ctx)
    {
        ctx->timeout = timeout;
        ctx->beginTime = PCRTime_GetMS();
        DOFConnection_PendingConnectionRetain(pendConn);
        ctx->pendConnNode = pendConn;
        return ctx;
    }
    return NULL;
}

static setupRelatedListenersContext_t * DOFConnection_CreateRelatedListenerContext(DOFConnectionConnectCallback callback, void *context)
{
    setupRelatedListenersContext_t *ctx = (setupRelatedListenersContext_t *)DOFMem_Alloc(sizeof(setupRelatedListenersContext_t));

    if (ctx)
    {
        ctx->callback = callback;
        ctx->context = context;
        return ctx;
    }
    return NULL;
}

static void DOFConnection_DestroyRelatedListenerContext(setupRelatedListenersContext_t *ctx)
{
    DOFMem_Free(ctx);
}

static void trpRoute_RemoveCredsAll_fn_t(OALAVLTreeItem_t item, void *arg)
{
    DOFConnection self = (DOFConnection)arg;
    OALConnection_RemoveDynamicTRPRoute( self->OALConn, (DOFDomain)item, NULL );
}

static void trpRoute_RemoveDomainsAll_fn_t(OALAVLTreeItem_t item, void *arg)
{
    DOFConnection self = (DOFConnection)arg;
    OALConnection_RemoveDynamicTRPRoute( self->OALConn, NULL, (DOFObjectIDDomain)item );
}

void DOFConnection_Destroy( DOFConnection self )
{
    DOFException    except;
    uint32          refCount;
    boolean         completeDestroy = FALSE;

    if (!self)
        return;

    PCRMutex_Lock(connection_server_mutex);

    if (!self->destroyInProgress)
    {
        /*  Note: calling DOFConnection_Destroy must not cause disconnect to be called unless the ref count
            for the connection drops to zero. */
        PCRMutex_Lock(ref_count_mutex);
        PCR_Assert( self->refcnt );
        refCount = --(self->refcnt);
        PCRMutex_Unlock(ref_count_mutex);

        if (!refCount)
        {
            OALSListNode tempPendList = self->pendingConnections;
            /*  If there are listener callbacks they will get enqueued and the DOFConnection must not be
                destroyed until all callbacks are called.  Set this so that any future calls to destroy will be ignored.
                The memory for the connection will be released after the last listener callback is called.
                See OALConnectionMgr_CompleteConnectionDestroy. */
            self->destroyInProgress = TRUE;
            except = DOFErrorException_Create(DOFERROR_FAILED, "Connection destroyed.");
            self->pendingConnections = NULL;
            OALSList_RemoveAll(&tempPendList, DOFConnection_RemovePendingConnections_fn, except);
            DOFException_Destroy(except);

            if (self->trpDomainRoutes)
            {
                OALAVLTree_Destroy( self->trpDomainRoutes, trpRoute_RemoveCredsAll_fn_t, self );
                self->trpDomainRoutes = NULL;
            }
            if (self->trpTrustedRoutes)
            {
                OALAVLTree_Destroy( self->trpTrustedRoutes, trpRoute_RemoveDomainsAll_fn_t, self );
                self->trpTrustedRoutes = NULL;
            }

            /* Remove listeners */
            DOFConnection_ClearAllListeners(self);

            if (self->connectedRefCount)
                OALConnection_DOFConnectionDisconnected(self->OALConn, NULL);

            OALConnection_RemoveDOFConnection(self->OALConn, self);

            /*  We will not free the connection here.  The OALConnectionMgr_NotifySystem_Destroyed function will
                complete the connection destroy process after all of the listener callbacks have been called. */
            completeDestroy = TRUE;
        }
    }

    PCRMutex_Unlock(connection_server_mutex);

    if (completeDestroy)
    {
        /* Call this outside of the connection mutex lock.  If the callback enqueue has been destroyed the
            final destroy may happen when this function is called. */
        OALConnectionMgr_NotifySystem_Destroyed(self);
    }

}

boolean DOFConnection_Init(void)
{
    connectionTimer = OALTimer_Create(DOF_GetDof()->threadPool, TRUE);

    if (connectionTimer)
        return TRUE;

    return FALSE;
}

void DOFConnection_Shutdown(void)
{
    /* Free the resources being used by DOFConnection. */
    /* This timer is used for multiple connections. */
    OALTimer_Destroy(connectionTimer);
    connectionTimer = NULL;
}

DOF DOFConnection_GetDOF( const DOFConnection self )
{
    PCR_Assert( self );
    return( self->OALConn->dof );
}

boolean DOFConnection_IsConnected( const DOFConnection self )
{
    boolean ret;
    PCR_Assert( self );
    PCRMutex_Lock(connection_server_mutex);
    ret = self->OALConn->flags.isConnected;
    PCRMutex_Unlock(connection_server_mutex);
    return( !!ret );    /* use !! to make sure either TRUE or FALSE is returned. */
}

DOFConnection DOFConnection_Copy( const DOFConnection src )
{
    DOFConnection retConn = NULL;

    if (src)
    {
        PCRMutex_Lock(ref_count_mutex);

        if (src->refcnt < 0xFFFFFFFF)
        {
            src->refcnt++;
            retConn = src;
        }
        PCRMutex_Unlock(ref_count_mutex);
    }

    return retConn;
}

void DOFConnection_ValidateComplete(const DOFConnection self, DOFException exception)
{
    /* TRP router finished validating the credentials, or it failed to validate one. */
    PCRMutex_Lock(connection_server_mutex);
    if (self->OALConn && self->OALConn->channel)
    {
        if (exception)
        {
            OALRemoteChannel_Shutdown(self->OALConn->channel, exception);
        }
        else
        {
            OALRemoteChannel_NotifyInitComplete(self->OALConn->channel);
        }
    }
    PCRMutex_Unlock(connection_server_mutex);
}

DOFConnectionState DOFConnection_CreateState(const DOFConnection self)
{
    DOFConnectionState ret = NULL;
    if ( self )
    {
        PCRMutex_Lock(connection_server_mutex);
        ret = DOFConnectionState_Create( self->OALConn->dof, self->OALConn->connConfig, self->OALConn->connSecurityMembrane,
                                         self->OALConn->flags.isConnected, self->OALConn->channel, self->OALConn->flags.isOutbound,
                                         self->OALConn->peerCredentials, OALConnection_GetRelatedDOFConnection(self->OALConn),
                                         OALConnection_GetRelatedDOFServer(self->OALConn), self->OALConn->byteInputCount,
                                         self->OALConn->byteOutputCount, self->OALConn->packetInputCount, self->OALConn->packetOutputCount, NULL);
        PCRMutex_Unlock(connection_server_mutex);
    }
    return ret;
}

DOFConnectionState DOFConnectionState_Create( DOF dof, DOFConnectionConfig connConfig, DOFSecurityMembrane securityMembrane,
                                              boolean isConnected, OALChannel channel, DOFConnectionDirection direction,
                                              DOFCredentials peerCredentials, DOFConnection relatedDofConnection,
                                              DOFServer relatedDofServer, uint64 inputByteCount, uint64 outputByteCount,
                                              uint64 inputPacketCount, uint64 outputPacketCount, DOFException exception)
{
    DOFConnectionState ret = DOFMem_Alloc( sizeof(struct DOFConnectionState_t) );
    if ( ret )
    {
        DOFError error = DOFERROR_OK;

        ret->refCount = 1;
        ret->dof = dof;
        ret->relatedDofConnection = DOFConnection_Copy(relatedDofConnection);
        ret->relatedDofServer = DOFServer_Retain(relatedDofServer);
        ret->config = DOFConnectionConfig_Copy( connConfig );
        ret->securityMembrane = DOFSecurityMembrane_Copy(securityMembrane);
        ret->exception = DOFException_Copy(exception);

        ret->byteInputCount = inputByteCount;
        ret->byteOutputCount = outputByteCount;
        ret->packetInputCount = inputPacketCount;
        ret->packetOutputCount = outputPacketCount;

        /* check for errors. */
        if ((relatedDofConnection && !ret->relatedDofConnection) ||
            (relatedDofServer && !ret->relatedDofServer) ||
            (connConfig && !ret->config) ||
            (securityMembrane && !ret->securityMembrane) ||
            (exception && !ret->exception))
        {
            error = DOFERROR_FAILED;
        }

        if ( isConnected && channel )
        {
            ret->status = DOFCONNECTIONSTATUS_CONNECTED;
            if ( channel->remoteChannelData )
            {
                ret->mode = DOFSecurityMode_Copy( channel->remoteChannelData->mode );
                if ( channel->remoteChannelData->mode && !ret->mode )
                {
                    error = DOFERROR_FAILED;
                }
                ret->credentials = DOFCredentials_Copy(channel->remoteChannelData->secureCredentials);
                if ( channel->remoteChannelData->secureCredentials && !ret->credentials )
                {
                    error = DOFERROR_FAILED;
                }
            }
            else
            {
                ret->mode = NULL;
                ret->credentials = NULL;
            }

            ret->direction = direction;
            ret->peerCredentials = DOFCredentials_CreateNeutered( peerCredentials );
            if ( peerCredentials && !ret->peerCredentials )
            {
                DOFConnectionState_Destroy(ret);
                ret = NULL;
            }
        }
        else
        {
            /* If there is no channel, we cannot be connected, and these are all undefined. */
            ret->status = DOFCONNECTIONSTATUS_DISCONNECTED;
            ret->mode = NULL;
            ret->direction = DOFCONNECTIONDIRECTION_OUTBOUND;
            ret->peerCredentials = NULL;
            ret->credentials = DOFCredentials_Copy(DOFConnectionConfig_GetNeuteredCredentials(connConfig));
            if ( DOFConnectionConfig_GetNeuteredCredentials(connConfig) && !ret->credentials )
            {
                error = DOFERROR_FAILED;
            }
        }

        if ( error != DOFERROR_OK )
        {
            DOFConnectionState_Destroy( ret );
            ret = NULL;
        }
    }

    return ret;
}

boolean DOFConnectionState_IsConnected( const DOFConnectionState self )
{
    PCR_Assert( self );
    return ( self->status == DOFCONNECTIONSTATUS_CONNECTED );
}

DOFSecurityMode DOFConnectionState_GetSecurityMode( const DOFConnectionState self )
{
    PCR_Assert( self );
    return self->mode;
}

DOFConnectionDirection DOFConnectionState_GetDirection( const DOFConnectionState self )
{
    PCR_Assert( self );
    return self->direction;
}

DOFConnectionConfig DOFConnectionState_GetConfig( const DOFConnectionState self )
{
    PCR_Assert( self );
    return self->config;
}

DOFException DOFConnectionState_GetException( const DOFConnectionState self )
{
    PCR_Assert( self );
    return self->exception;
}

DOFCredentials DOFConnectionState_GetCredentials( const DOFConnectionState self )
{
    return self->credentials;
}

DOFCredentials DOFConnectionState_GetPeerCredentials( const DOFConnectionState self )
{
    return self->peerCredentials;
}

DOFConnectionState DOFConnectionState_Copy( DOFConnectionState self )
{
    PCR_Assert( self );
    PCRMutex_Lock(ref_count_mutex);
    PCR_Assert( self->refCount < 0xFFFFFFFF );
    ++self->refCount;
    PCRMutex_Unlock(ref_count_mutex);
    return self;
}

DOFConnection DOFConnectionState_CreateConnection( const DOFConnectionState self )
{
    DOFConnection   ret = NULL;
    DOFConnection   relatedConn = NULL;
    DOFServer       relatedSrv = NULL;

    PCR_Assert(self);

    /* Since the application must not be allowed to use the internal DOFServer or DOFConnection, create
       a new related server or related connection if they are present. */
    if (self->relatedDofConnection)
        relatedConn = OALConnection_CreateDOFConnection(self->relatedDofConnection->OALConn);
    if (self->relatedDofServer)
        relatedSrv = OALServer_CreateDOFServer(self->relatedDofServer->server);

    if ((relatedConn || !self->relatedDofConnection) && (relatedSrv || !self->relatedDofServer))
    {
        ret = DOFConnection_CreateInternalConnection(self->dof, self->securityMembrane, self->config,
                                                     relatedSrv, relatedConn, TRUE);
    }
    DOFConnection_Destroy(relatedConn);
    DOFServer_Release(relatedSrv);

    return ret;
}

void DOFConnectionState_Destroy( DOFConnectionState self )
{
    if ( self )
    {
        uint32 refCount;

        PCRMutex_Lock(ref_count_mutex);
        PCR_Assert( self->refCount );
        refCount = --(self->refCount);
        PCRMutex_Unlock(ref_count_mutex);

        if ( !refCount )
        {
            DOFCredentials_Destroy(self->credentials);
            DOFCredentials_Destroy( self->peerCredentials );
            DOFSecurityMode_Destroy( self->mode );
            DOFSecurityMembrane_Destroy(self->securityMembrane);
            DOFConnectionConfig_Destroy( self->config );
            DOFConnection_Destroy(self->relatedDofConnection);
            DOFServer_Destroy(self->relatedDofServer);
            DOFException_Destroy(self->exception);
            DOFMem_Free( self );
        }
    }
}

uint64 DOFConnectionState_GetReceiveByteCount( const DOFConnectionState self )
{
    PCR_Assert(self);
    return self->byteInputCount;
}

uint64 DOFConnectionState_GetSendByteCount( const DOFConnectionState self )
{
    PCR_Assert(self);
    return self->byteOutputCount;
}

uint64 DOFConnectionState_GetReceivePacketCount( const DOFConnectionState self )
{
    PCR_Assert(self);
    return self->packetInputCount;
}

uint64 DOFConnectionState_GetSendPacketCount( const DOFConnectionState self )
{
    PCR_Assert(self);
    return self->packetOutputCount;
}

DOFConnectionType DOFConnectionState_GetConnectionType( const DOFConnectionState self )
{
    return DOFConnectionConfig_GetConnectionType(self->config);
}

DOFAddress DOFConnectionState_GetAddress( const DOFConnectionState self )
{
    return DOFConnectionConfig_GetAddress(self->config);
}

DOFSecurityDesire DOFConnectionState_GetSecurityDesire( const DOFConnectionState self )
{
    return DOFConnectionConfig_GetSecurityDesire(self->config);
}

DOFPermissionSet DOFConnectionState_GetPermissions( const DOFConnectionState self )
{
    return DOFConnectionConfig_GetPermissions(self->config);
}

boolean DOFConnectionState_IsPermissionsExtendAllowed( const DOFConnectionState self )
{
    return DOFConnectionConfig_IsPermissionsExtendAllowed(self->config);
}

DTLConnectionConfig DOFConnectionState_GetTransportConfig( const DOFConnectionState self )
{
    return DOFConnectionConfig_GetTransportConfig(self->config);
}

DOFProtocolNegotiator DOFConnectionState_GetProtocolNegotiator( const DOFConnectionState self )
{
    return DOFConnectionConfig_GetProtocolNegotiator(self->config);
}

int DOFConnectionState_GetMaxSendSilence( const DOFConnectionState self )
{
    return DOFConnectionConfig_GetMaxSendSilence(self->config);
}

int DOFConnectionState_GetMaxReceiveSilence( const DOFConnectionState self )
{
    return DOFConnectionConfig_GetMaxReceiveSilence(self->config);
}

DOFServer DOFConnection_GetRelatedServer( const DOFConnection self )
{
    if (self)
        return self->relatedServer;

    return NULL;
}

DOFConnection DOFConnection_GetRelatedConnection( const DOFConnection self )
{
    if (self)
        return self->relatedConnection;

    return NULL;
}


DOFConnection DOFConnection_CreateConnection( DOFConnection self, DOFConnectionConfig config )
{
    DOFConnection   ret = NULL;
    DOFSecurityMembrane membrane;
    PCR_Assert( self );

    membrane = DOFConnection_CreateSecurityMembrane(self->OALConn->dof, config);
    if (membrane != NULL)
    {
        ret = DOFConnection_CreateInternalConnection( self->OALConn->dof, membrane, config, NULL, self, TRUE );
        DOFSecurityMembrane_Destroy(membrane);
    }

    return ret;
}


DOFConnection DOFConnection_CreateInternalConnection( DOF dof, DOFSecurityMembrane securityMembrane, DOFConnectionConfig connConfig,
                                                      DOFServer relatedServer, DOFConnection relatedConnection, boolean shareOALConn )
{
    DOFConnectionType connectionType;
    DOFConnection     retVal      = NULL;

    /* connnConfig and dof must not be NULL. */
    if (!connConfig || !dof)
        return NULL;

    /* Validate parameters and connection config. */
    connectionType = DOFConnectionConfig_GetConnectionType( connConfig );
    if ((connectionType == DOFCONNECTIONTYPE_GROUP) ||
        (connectionType == DOFCONNECTIONTYPE_GROUP_MEMBER_ONLY) ||
        (connectionType == DOFCONNECTIONTYPE_HUB))
    {
        if (!relatedServer || DOFConnectionConfig_GetCredentials( connConfig ) == NULL)
            return NULL;                   /* A hub or secure group member must have a relatedServer and credentials. */
    }
    else if (connectionType == DOFCONNECTIONTYPE_POINT)
    {
        if (!relatedConnection || !connConfig->credentials)
            return NULL;                          /* A point member must have a relatedConnection and credentials. */
    }
    else if ((connectionType == DOFCONNECTIONTYPE_DATAGRAM_STATELESS) &&
             (DOFConnectionConfig_GetSecurityDesire(connConfig) & DOFSECURITYDESIRE_SECURE_ANY))
    {
        /* A datagram stateless connection cannot be secure. */
        return NULL;
    }

    retVal = (DOFConnection)DOFMem_Alloc(sizeof(DOFConnection_t));
    if (retVal)
    {
        PCRMem_Set( retVal, 0, sizeof( DOFConnection_t ) );

        retVal->trpTrustedRoutes = OALAVLTree_Create( NULL, NULL, NULL );
        if( retVal->trpTrustedRoutes )
        {
            retVal->trpDomainRoutes = OALAVLTree_Create( NULL, NULL, NULL );
            if( retVal->trpDomainRoutes )
            {
                retVal->refcnt = 1;
                retVal->relatedConnection = DOFConnection_Copy(relatedConnection);
                if (!relatedConnection || (relatedConnection && retVal->relatedConnection))
                {
                    retVal->relatedServer = relatedServer;
                    retVal->OALConn = OALConnection_CreateConnection( retVal, connConfig, securityMembrane, dof,
                                                                      shareOALConn );
                    if (retVal->OALConn)
                    {
                        DOFServer_Retain(relatedServer);
                        return retVal;
                    }
                    DOFConnection_Destroy(retVal->relatedConnection);
                }
                OALAVLTree_Destroy( retVal->trpDomainRoutes, NULL, NULL );
            }
            OALAVLTree_Destroy( retVal->trpTrustedRoutes, NULL, NULL );
        }
        DOFMem_Free(retVal);
    }

    return NULL;
}

/* Note: This function must be called from channel's message_enqueue worker thread. */
boolean DOFConnection_CreateConnectedInboundConnection( DOF dof, DOFSecurityMembrane securityMembrane, DOFConnectionConfig connConfig,
                                                        DOFServer relatedServer, OALChannel channel, boolean addListener )
{
    DOFConnection   conn;

    PCR_Assert(channel);

    PCRMutex_Lock(connection_server_mutex);
    conn = DOFConnection_CreateInternalConnection(dof, securityMembrane, connConfig, relatedServer, NULL, FALSE);

    if (conn)
    {
        if (OALConnection_Retain(conn->OALConn))
        {
            channel->remoteChannelData->OALConn = conn->OALConn;
            if (OALConnection_ConnectInbound(conn->OALConn, conn, channel))
            {
                if (addListener)
                {
                    if (!OALServer_AddConnectionListener(relatedServer->server, conn))
                    {
                        OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_FAILED);
                        DOFConnection_Destroy(conn);
                        conn = NULL;
                    }
                }
            }
            else
            {
                OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_FAILED);
                DOFConnection_Destroy(conn);
                conn = NULL;
            }
        }
        else
        {
            OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_FAILED);
            DOFConnection_Destroy(conn);
            conn = NULL;
        }
    }

    PCRMutex_Unlock(connection_server_mutex);

    if (conn)
    {
        /* If this is successfull the OALConnection will own conn that is created here, call destroy for conn and return TRUE. */
        DOFConnection_Destroy(conn);
        return TRUE;
    }
    return FALSE;
}


static boolean DOFConnection_removePendConnFromConnList(OALSListNode pendConn, void *arg)
{
    if (pendConn == arg)
        return TRUE;

    return FALSE;
}

static int DOFConnection_pendingConnectionTimeout_fn(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown)
{
    pendingConnection_t *   pendingConnection;

    pendingConnection_t * pendConn = (pendingConnection_t *)argv[0];

    PCR_Assert(argc==1);

    /* The connect timed out.  If the pending connection is still in the pendingConnections list then
       enqueue a connect complete callback. */
    PCRMutex_Lock(connection_server_mutex);

    pendingConnection = (pendingConnection_t *)OALSList_FindAndRemove((OALSListNode *)&(pendConn->pendConnection->pendingConnections), NULL,
                                                                      DOFConnection_removePendConnFromConnList, pendConn);

    if (pendingConnection == pendConn)
    {
        /* The item was still in the list so the complete callback must be called. */
        /* This can be called with shutdown set, but just report timeout. */
        OALConnectionMgr_enqueueConnectComplete(pendConn, DOFEXCEPTION_TIMEOUT);
        /* pendConn was retained when inserted in the list, release it now that it was removed. */
        DOFConnection_PendingConnectionRelease(pendConn);
    }

    /* The pending connection was used as a context in the timer item, release it. */
    DOFConnection_PendingConnectionRelease(pendConn);

    PCRMutex_Unlock(connection_server_mutex);

    return 0;
}

static void relatedConnectionListener_Fn(DOFConnectionStateListener self, DOFConnection connection, DOFConnectionState state)
{
    RelatedConnectionListener_t *   listener;
    boolean                         isConnected;

    listener = (RelatedConnectionListener_t *)self;
    isConnected = DOFConnectionState_IsConnected(state);

    PCRMutex_Lock(connection_server_mutex);
    if (!isConnected)
    {
        /* Notify the OALConnection that the related connection is down. */
        OALConnection_ConnectionEvent(listener->connection->OALConn, listener->connection->OALConn->channel, FALSE, NULL);
        DOFConnection_RemoveStateListener(connection, &listener->connectionListener);
        listener->connection->listener.connection = NULL;
        DOFConnection_Disconnect(connection);
    }
    PCRMutex_Unlock(connection_server_mutex);
}

static void relatedConnectionListenerRemoved_Fn(DOFConnectionStateListener self, DOFConnection connection, DOFException except )
{
    RelatedConnectionListener_t * listener;

    listener = (RelatedConnectionListener_t *)self;

    PCRMutex_Lock(connection_server_mutex);
    DOFConnection_Destroy(listener->connection);
    PCRMutex_Unlock(connection_server_mutex);
    DOFMem_Free(listener);
}

static void relatedServerListener_Fn(DOFServerStateListener self, DOFServer server, DOFServerState state)
{
    RelatedServerListener_t *   listener;
    boolean isStarted;


    listener = (RelatedServerListener_t *)self;
    isStarted = DOFServerState_IsStarted(state);

    PCRMutex_Lock(connection_server_mutex);
    if(!isStarted)
    {
        if (listener->connection->OALConn->channel)
        {
            /* Shut down the channel now that the server is down. */
            OALRemoteChannel_Shutdown(listener->connection->OALConn->channel, NULL);
        }
        DOFServer_RemoveStateListener(server, &listener->serverListener);
        listener->connection->listener.server = NULL;
        if (listener->connection->OALConn->flags.isOutbound)
        {
            DOFServer_Stop(server);
        }
    }
    PCRMutex_Unlock(connection_server_mutex);
}

static void relatedServerListenerRemoved_Fn(DOFServerStateListener self, DOFServer server, DOFException except )
{
    RelatedServerListener_t * listener;

    listener = (RelatedServerListener_t *)self;

    PCRMutex_Lock(connection_server_mutex);
    DOFConnection_Destroy(listener->connection);
    PCRMutex_Unlock(connection_server_mutex);
    DOFMem_Free(listener);
}

static void syncConnectCallback( DOFConnectionConnectCallback self, const DOFConnection conn, void * context, const DOFException except )
{
    /* Copy the exception into the context and signal the event. */
    ((connectContext_t *)context)->exception = DOFException_Copy(except);
    PCREvent_Signal(((connectContext_t *)context)->evt);
}

static CallConnectComplete_t * DOFConnection_CreateConnectCompleteNode(DOFConnection conn, DOFConnectionConnectCallback callback, void *context)
{
    CallConnectComplete_t *node = (CallConnectComplete_t *)DOFMem_Alloc(sizeof(CallConnectComplete_t));

    if (node)
    {
        PCRMem_Set(node, 0, sizeof(CallConnectComplete_t));
        node->callback = callback;
        node->connection = conn;
        node->context = context;
    }
    return node;
}

/* Note: The connection_server_mutex must be locked when this function is called. */
static pendingConnection_t * DOFConnection_CreatePendingConnectionNode(DOFConnection conn,
                                                                       DOFConnectionConnectCallback callback,
                                                                       void *context)
{
    PCR_Assert(conn);
    {
        pendingConnection_t * pendingConnection = (pendingConnection_t *)DOFMem_Alloc(sizeof(pendingConnection_t));

        if (pendingConnection)
        {
            pendingConnection->refCount = 1;
            pendingConnection->cbNode = DOFConnection_CreateConnectCompleteNode(conn, callback, context);
            if (pendingConnection->cbNode)
            {
                pendingConnection->pendConnection = DOFConnection_Copy(conn);
                if ( pendingConnection->pendConnection )
                {
                    pendingConnection->timerItem = NULL;
                    pendingConnection->next = NULL;

                    return pendingConnection;
                }
                DOFMem_Free(pendingConnection->cbNode);
            }
            DOFMem_Free(pendingConnection);
        }
    }
    return NULL;
}

/* Note: The connection_server_mutex must be locked when this function is called. */
static void DOFConnection_ConnectPendingConnection(pendingConnection_t *pendingConnection,
                                                   uint32 timeout,
                                                   DOFException *except)
{
    /* An exception is required.  This is OK since this is not a public call. */
    PCR_Assert(except);
    PCR_Assert(pendingConnection);
    PCR_Assert(pendingConnection->pendConnection);
    PCR_Assert(pendingConnection->pendConnection->OALConn);

    *except = NULL;

    if (pendingConnection->pendConnection->OALConn->flags.isConnected)
    {
        /* the OALConnection is already connected so enqueue a callback. */
        OALConnectionMgr_enqueueConnectComplete(pendingConnection, NULL);
    }
    else
    {
        if (pendingConnection->pendConnection->destroyInProgress || pendingConnection->pendConnection->shutdownByDOF)
        {
            *except = DOFException_Copy(DOFEXCEPTION_FAILED);
        }
        else
        {
            if (timeout == 0)
            {
                *except = DOFException_Copy(DOFEXCEPTION_TIMEOUT);
            }
            else
            {
                if (timeout != DOF_TIMEOUT_NEVER)
                {
                    pendingConnection->timerItem = OALTimer_CreateItem(connectionTimer, DOFConnection_pendingConnectionTimeout_fn, 1, (void**)&pendingConnection);
                    if (pendingConnection->timerItem)
                    {
                        if (OALTimerItem_Insert( pendingConnection->timerItem, timeout, FALSE, FALSE ) == DOFERROR_OK)
                        {
                            /* The pending connection node was added as a context to the timer so retain the pending connection. */
                            DOFConnection_PendingConnectionRetain(pendingConnection);
                        }
                        else
                        {
                            *except = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
                        }
                    }
                    else
                    {
                        *except = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
                    }
                }
                else
                {
                    /* A NULL timerItem serves as a flag for DOF_TIMEOUT_NEVER */
                    pendingConnection->timerItem = NULL;
                }
            }
        }

        if (!(*except))
        {
            OALConnection_Connect(pendingConnection->pendConnection->OALConn, except);
        }
    }

    if (*except)
    {
        /* If there was an exception and a timer item, remove the timerItem. */
        if (pendingConnection->timerItem)
        {
            if (OALTimerItem_Remove(pendingConnection->timerItem))
            {
                /* Only release the pending connection if the timer item was removed.
                   If the timer expires the pendingConnection will be released by the timer callback function. */
                DOFConnection_PendingConnectionRelease(pendingConnection);
            }
        }
    }
}

/* Note: The connection_server_mutex must be locked when this function is called. */
static void DOFConnection_CreateAndConnectPendingConnection(DOFConnection self,
                                                            DOFConnectionConnectCallback callback,
                                                            void *context,
                                                            DOFException *callbackException,
                                                            uint32 timeout)
{
    pendingConnection_t *pendConn;

    PCR_Assert(self);
    PCR_Assert(callbackException);

    /* There is not a related connection or server, or this is a point connection, so just connect. */
    pendConn = DOFConnection_CreatePendingConnectionNode(self, callback, context);
    if (pendConn)
    {
        /* Add the pending connection to the list.  The list will take ownership of it. */
        OALSList_Insert(&self->pendingConnections, (OALSListNode )pendConn);

        /* This will check to see if the OALConnection is connected or if it must be. */
        DOFConnection_ConnectPendingConnection(pendConn, timeout, callbackException);

        if (*callbackException)
        {
            OALSList_FindAndRemove(&self->pendingConnections, NULL, DOFConnection_removePendConnFromConnList, pendConn);
            DOFConnection_PendingConnectionRelease(pendConn);
        }
    }
    else
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFConnection: Failed to create pending connection node." );
        *callbackException = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
    }
}

/* Note: The connection_server_mutex must be locked when this function is called. */
static void DOFConnection_ConnectUsingRelated(DOFConnection self,
                                              DOFConnectionConnectCallback callback,
                                              void *context,
                                              DOFException *callbackException,
                                              uint32 timeout)
{
    setupRelatedListenersContext_t *listenerSetupContext;

    PCR_Assert(self);
    PCR_Assert(callbackException);

    /* If the connection has a base OALConnection or base OALServer then connect the related connection
       or start the related server.  If the related DOFConnection is connecting, or the related DOFServer
       is starting then just override the context so that a listener can be added if it is required. */

    listenerSetupContext = DOFConnection_CreateRelatedListenerContext(callback, context);
    if (listenerSetupContext)
    {
        if (self->stopOrDisconnectRelated)
        {
            /* If the related connection or server is already connected or started then just connect this pending connection node. */
            DOFConnection_CreateAndConnectPendingConnection(self, &setupRelatedListenersCB, listenerSetupContext, callbackException, timeout);
        }
        else
        {
            pendingConnection_t *pendConn;

            pendConn = DOFConnection_CreatePendingConnectionNode(self, &setupRelatedListenersCB, listenerSetupContext);
            if (pendConn)
            {
                connectRelatedContext_t *relatedContext;

                relatedContext = DOFConnection_CreateRelatedContext(self, pendConn, timeout);

                if (relatedContext)
                {
                    /* Add the pending connection to the list.  The list will take ownership of it. */
                    OALSList_Insert(&self->pendingConnections, (OALSListNode)pendConn);

                    if (self->relatedConnection)
                    {
                        DOFConnection_BeginConnect(self->relatedConnection, timeout, &relatedConnConnectedCB, relatedContext);
                    }
                    else
                    {
                        PCR_Assert(self->relatedServer);

                        DOFServer_BeginStart(self->relatedServer, timeout, &relatedServerStartedCB, relatedContext);
                    }
                }
                else
                {
                    DOFConnection_DestroyRelatedListenerContext(listenerSetupContext);
                    DOFConnection_PendingConnectionRelease(pendConn);
                    DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "DOFConnection: Failed to create internal related connection/server context.");
                    *callbackException = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
                }
            }
            else
            {
                DOFConnection_DestroyRelatedListenerContext(listenerSetupContext);
                DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "DOFConnection: Failed to create internal pending connection context.");
                *callbackException = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            }
        }
    }
    else
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFConnection: Failed to create internal listener context." );
        *callbackException = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
    }
}

static void setupRelatedListeners( DOFConnectionConnectCallback self, const DOFConnection conn, void * context, const DOFException except )
{
    DOFException                    exception = except;
    setupRelatedListenersContext_t  *ctx;

    PCR_Assert(conn);
    PCR_Assert(context);
    ctx = (setupRelatedListenersContext_t *)context;

    PCRMutex_Lock(connection_server_mutex);
    if (conn->destroyInProgress || conn->shutdownByDOF)
    {
        exception = DOFEXCEPTION_FAILED;
    }

    if (!exception && conn->relatedConnection && !conn->listener.connection)
    {
        /* Add a listener so that if the related connection goes down this connection is notified. */
        conn->listener.connection = DOFMem_Alloc(sizeof(RelatedConnectionListener_t));
        if (conn->listener.connection)
        {
            PCRMem_Set(conn->listener.connection, 0, sizeof(RelatedConnectionListener_t));
            conn->listener.connection->connection = DOFConnection_Copy(conn);
            conn->listener.connection->connectionListener = relatedConnectionListener;

            if (!conn->listener.connection->connection ||
                !DOFConnection_AddStateListener(conn->relatedConnection, &(conn->listener.connection->connectionListener)))
            {
                DOFMem_Free(conn->listener.connection);
                conn->listener.connection = NULL;
            }
        }

        if (!conn->listener.connection)
        {
            /* There was an error, disconnect. */
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFConnection: setup related listeners: Failed to create related connection listener." );
            exception = DOFEXCEPTION_INSUFFICIENT_RESOURCES;
            DOFConnection_Disconnect(conn);
        }
    }
    else if (!exception && conn->relatedServer && !conn->listener.server)
    {
        /* Add a listener to the server so that if it goes down this connection is notified. */
        if (!DOFConnection_AddInternalServerListener(conn))
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFConnection: setup related listeners: Failed to cretae related server listener." );
            exception = DOFEXCEPTION_INSUFFICIENT_RESOURCES;
            DOFConnection_Disconnect(conn);
        }
    }

    PCRMutex_Unlock(connection_server_mutex);

    /* Call the real connect complete callback now. */
    if (ctx && ctx->callback && ctx->callback->fns->Complete)
    {
        ctx->callback->fns->Complete(ctx->callback, conn, ctx->context, exception);
    }

    DOFMem_Free(ctx);
}

/* This is a common function that is used for the complete callback for starting rleated connections and servers. */
static void startConnWithRelatedConnSrv (connectRelatedContext_t * ctx, const DOFException except)
{
    DOFException                exception = NULL;

    if (!except)
    {
        uint32 timeout;

        /* The related connection/server is now connected/started. */

        PCRMutex_Lock(connection_server_mutex);

        if (!ctx->pendConnNode->cbNode)
        {
            /* The related connection/server connected/started, but something happened to stop the connection progress
               for the parent connection while the related connection/server was connection/starting.
               Stop the related server or disconnect the related connection and allow the context to be destroyed,
               but there is nothing else to do since the complete callback for the connection was already triggered. */
            if (ctx->pendConnNode->pendConnection->relatedConnection)
            {
                DOFConnection_Disconnect(ctx->pendConnNode->pendConnection->relatedConnection);
            }
            if (ctx->pendConnNode->pendConnection->relatedServer)
            {
                DOFServer_Stop(ctx->pendConnNode->pendConnection->relatedServer);
            }
        }
        else
        {
            if (ctx->pendConnNode->pendConnection->stopOrDisconnectRelated)
            {
                /* BeginConnect or BeginStart may have been called multiple times for the related DOFConnection or
                   related DOFServer.  If stopOrDisconnectRelated is already set then the connection should call
                   Disconnect or Stop so that only a single disconnect or stop is required at the point the
                   connectedRefCount goes to zero. */
                if (ctx->pendConnNode->pendConnection->relatedConnection)
                {
                    DOFConnection_Disconnect(ctx->pendConnNode->pendConnection->relatedConnection);
                }
                if (ctx->pendConnNode->pendConnection->relatedServer)
                {
                    DOFServer_Stop(ctx->pendConnNode->pendConnection->relatedServer);
                }
            }
            else
            {
                /* The Related Server or connection was successfully started or connected. If there is a failure later that
                   prevents the connection from connecting the related server or connection must be stopped or disconnected. */
                ctx->pendConnNode->pendConnection->stopOrDisconnectRelated = TRUE;
            }


            timeout = ctx->timeout - (PCRTime_GetMS() - ctx->beginTime);
            if (timeout > ctx->timeout)
            {
                /* This must have timed out since the related connection or server was connected.  Use 0 ms.
                   If the connection is already connected it will complete and call the callback as connected,
                   otherwise it will call the callback with a timeout exception. */
                timeout = 0;
            }

            DOFConnection_ConnectPendingConnection(ctx->pendConnNode, timeout, &exception);
        }

        PCRMutex_Unlock(connection_server_mutex);
    }
    else
    {
        exception = except;
    }

    /* If there is an exception call the callback now. */
    if (exception)
    {
        OALConnectionMgr_enqueueConnectComplete(ctx->pendConnNode, exception);

        if (!except)
        {
            /* If the exception was a result of calling DOFConnection_ConnectPendingConnection, destroy the exception. */
            DOFException_Destroy(exception);
        }
    }

    DOFConnection_DestroyRelatedContext(ctx);
}

static void relatedConnectionConnected( DOFConnectionConnectCallback self, const DOFConnection conn, void * context, const DOFException except )
{
    PCR_Assert(conn);
    PCR_Assert(context);

    startConnWithRelatedConnSrv((connectRelatedContext_t *)context, except);
}

static void relatedServerStarted( DOFServerStartCallback self, const DOFServer server, void* context, const DOFException except )
{
    PCR_Assert(context);

    startConnWithRelatedConnSrv((connectRelatedContext_t *)context, except);
}

boolean DOFConnection_Connect( DOFConnection self, uint32 timeout, DOFException * pException )
{
    connectContext_t    ctx;

    PCRMutex_Lock(connection_server_mutex);

    ctx.evt = PCREvent_Create(FALSE);
    ctx.exception = NULL;

    DOFConnection_BeginConnect(self, timeout, &synchronusConnectCB, &ctx);

    PCRMutex_Unlock(connection_server_mutex);

    PCREvent_Wait(ctx.evt, DOF_TIMEOUT_NEVER);

    PCRMutex_Lock(connection_server_mutex);

    PCREvent_Destroy(ctx.evt);

    /*  Pass back the exception from the context.  Since it was already copied when it was saved it in the
        context we do not copy it here. */
    if (pException)
        *pException = ctx.exception;
    else
        DOFException_Destroy(ctx.exception);        /* Destroy this exception if it will not be returned. */

    PCRMutex_Unlock(connection_server_mutex);

    return (!ctx.exception);
}

void DOFConnection_BeginConnect( DOFConnection self, uint32 timeout, DOFConnectionConnectCallback callback, void * context )
{
    DOFException callbackException = NULL;

    PCR_Assert(self);
    PCR_Assert( !callback || callback->fns->Complete );

    PCRMutex_Lock(connection_server_mutex);

    if (self->connectedRefCount == 0)
    {
        self->stopOrDisconnectRelated = FALSE;
    }

    if (self->connectedRefCount < CONNECTION_MAX_REF_COUNT)
    {
        ++self->connectedRefCount;

        if (self->connectedRefCount == 1)
        {
            OALConnection_DOFConnectionConnected(self->OALConn);
        }

        /* OALConn->base is a union so it could also be OALSrv, but in this case we only want to know if one or the other exists.
           If there is a base OALConnection or OALSERVER then use the related DOFConnection or DOFServer, otherwise just connect. */
        if (self->OALConn->base.OALConn && (DOFConnectionConfig_GetConnectionType(self->OALConn->connConfig) != DOFCONNECTIONTYPE_POINT))
        {
            DOFConnection_ConnectUsingRelated(self, callback, context, &callbackException, timeout);
        }
        else
        {
            /* Connect a connection that is not using a related connection or server. */
            DOFConnection_CreateAndConnectPendingConnection(self, callback, context, &callbackException, timeout);
        }

        /* If there was an exception here the connectedRefCount must be decremented. */
        if (callbackException)
        {
            --self->connectedRefCount;
            if (self->connectedRefCount == 0)
            {
                OALConnection_DOFConnectionDisconnected(self->OALConn, NULL);
            }
        }
    }
    else
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "DOFConnection: internal connected count is at maximum.");
        callbackException = DOFException_Copy(DOFEXCEPTION_TOO_MANY);
    }

    PCRMutex_Unlock(connection_server_mutex);

    /* There was an error so call the app callback now.*/
    if (callbackException)
    {
        if (callback)
        {
            callback->fns->Complete(callback, self, context, callbackException);
        }
        DOFException_Destroy(callbackException);
    }
}


OALChannel DOFConnection_GetChannel( DOFConnection self )
{
    return self->OALConn->channel;
}


/* Terminate will force the connection back to a disconnected state The connected ref count must remain the same.
   This will cause the OALConnection back to a disconnected state as well.  So, if multiple DOFConnections are
   using the same OALConnection it will cause all DOFConnections using the OALConnection to be terminated.*/
void DOFConnection_Terminate(DOFConnection self, DOFException exception)
{
    PCRMutex_Lock(connection_server_mutex);
    if (self->OALConn && self->OALConn->channel)
    {
        if (exception)
            OALRemoteChannel_Shutdown(self->OALConn->channel, exception);
        else
            OALRemoteChannel_Shutdown(self->OALConn->channel, DOFEXCEPTION_TERMINATED);
    }
    PCRMutex_Unlock(connection_server_mutex);
}

void DOFConnection_Disconnect(DOFConnection self)
{
    PCR_Assert( self );
    PCRMutex_Lock(connection_server_mutex);

    if ((self->connectedRefCount > 0) && ((--self->connectedRefCount) == 0))
    {
        /* Notify the OALConnection that the DOFConnection is no longer connected. */
        OALConnection_DOFConnectionDisconnected(self->OALConn, NULL);

        DOFConnection_ClearRelatedListeners(self);

        if (self->stopOrDisconnectRelated)
        {
            if (self->relatedConnection)
            {
                DOFConnection_Disconnect(self->relatedConnection);
            }
            if (self->relatedServer)
            {
                DOFServer_Stop(self->relatedServer);
            }
            /* Reset this flag so it is ready if connect is called again. */
            self->stopOrDisconnectRelated = FALSE;
        }
    }

    PCRMutex_Unlock(connection_server_mutex);
}

static boolean findConnectionCallback(OALSListNode node, void *context)
{
    DOFConnectionStateListener listener = (DOFConnectionStateListener)context;

    return (((connLstnNode_t *)node)->listener == listener);
}

boolean DOFConnection_AddStateListener( DOFConnection self, DOFConnectionStateListener listener )
{
    boolean retStatus = FALSE;

    if (self && listener)
    {
        connLstnNode_t * listenerNode;

        PCRMutex_Lock(connection_server_mutex);
        /* Verify that the connection manager has not been shut down by checking the shutdownByDOF flag. */
        if (!self->shutdownByDOF)
        {
            /* verify that listener is not currently in the list. */
            if (OALSList_Traverse(self->listenerList, findConnectionCallback, listener))
            {
                /* The listener was already added so return FALSE. */
                DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_TRACE, "DOFConnection, add listener: Listener was already added.");
            }
            else
            {
                listenerNode = DOFConnection_CreateListenerNode(listener);

                if (listenerNode)
                {
                    DOFConnectionState  connState;

                    connState = DOFConnection_CreateState(self);

                    if (connState)
                    {
                        /*  Insert the node into the listener list.  Since the listener callback will be called with the inital status, this
                            must be added to the list and then the callback triggered all while the connection mutex is locked. */
                        OALSList_Insert(&self->listenerList, (OALSListNode)listenerNode);
                        /* trigger the initial listener callback now. */
                        OALConnectionMgr_NotifySystem_ConnectionEvent(self, connState, listener);

                        /* The listener was added to the connection.  Any error from here out must cause the removed callback to be called. */
                        retStatus = TRUE;

                        DOFConnectionState_Destroy(connState);
                    }
                    else
                    {
                        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_WARNING, "DOFConnection: Failed to create connection state.");
                        DOFConnection_DestroyListenerNode(listenerNode);
                    }
                }
                else
                {
                    DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_WARNING, "DOFConnection: Failed to create connection listener node.");
                }
            }
        }
        PCRMutex_Unlock(connection_server_mutex);
    }

    return retStatus;
}

void DOFConnection_RemoveStateListener( DOFConnection self, DOFConnectionStateListener listener )
{
    if (self && listener)
    {
        connLstnNode_t * listenerNode;

        PCRMutex_Lock(connection_server_mutex);
        listenerNode = (connLstnNode_t *)OALSList_FindAndRemove(&self->listenerList, NULL,
                                                               findConnectionCallback, listener);
        if (listenerNode)
        {
            OALConnectionMgr_NotifySystem_ListenerRemoved(self, listenerNode);
            DOFConnection_DestroyListenerNode(listenerNode);
        }
        PCRMutex_Unlock(connection_server_mutex);
    }
}

static boolean findPeerListener_Traverse_fn(OALSListNode node, void *context)
{
    DOFConnectionPeerListener listener = (DOFConnectionPeerListener)context;

    return (((connPeerLstnNode_t *)node)->listener == listener);
}

boolean DOFConnection_AddPeerListener( DOFConnection self, DOFConnectionPeerListener listener )
{
    boolean retStatus = FALSE;

    if (self && listener && (DOFConnectionConfig_GetConnectionType(self->OALConn->connConfig) == DOFCONNECTIONTYPE_HUB))
    {
        connPeerLstnNode_t * listenerNode;

        PCRMutex_Lock(connection_server_mutex);

        /* Verify that the connection manager has not been shut down by checking the shutdownByDOF flag. */
        if (!self->shutdownByDOF)
        {
            /* verify that listener is not currently in the list. */
            if (OALSList_Traverse(self->peerListenerList, findPeerListener_Traverse_fn, listener))
            {
                /* The listener was already added so return FALSE. */
                DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_TRACE, "DOFConnection, add peer listener: Listener was already added.");
            }
            else
            {
                listenerNode = DOFConnection_CreatePeerListenerNode(listener);

                if (listenerNode)
                {
                    /*  Insert the node into the listener list. */
                    OALSList_Insert(&self->peerListenerList, (OALSListNode)listenerNode);

                    /* The listener was added to the connection.  Any error from here out must cause the removed callback to be called. */
                    retStatus = TRUE;
                }
                else
                {
                    DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_WARNING, "DOFConnection: Failed to create connection peer listener node.");
                }
            }
        }

        PCRMutex_Unlock(connection_server_mutex);
    }
    return retStatus;
}

void DOFConnection_RemovePeerListener( DOFConnection self, DOFConnectionPeerListener listener )
{
    if (self && listener)
    {
        connPeerLstnNode_t * listenerNode;

        PCRMutex_Lock(connection_server_mutex);
        listenerNode = (connPeerLstnNode_t *)OALSList_FindAndRemove(&self->peerListenerList, NULL,
                                                               findPeerListener_Traverse_fn, listener);
        if (listenerNode)
        {
            OALConnectionMgr_NotifySystem_PeerListenerRemoved(self, listenerNode);
            DOFConnection_DestroyPeerListenerNode(listenerNode);
        }
        PCRMutex_Unlock(connection_server_mutex);
    }
}

static void RemoveAll_DOFDomains( OALAVLTreeItem_t item, void *arg )
{
    DOFDomain domain = (DOFDomain)item;
    DOFConnection self = (DOFConnection)arg;

    OALConnection_RemoveDynamicTRPRoute( self->OALConn, domain, NULL );
}

static void RemoveAll_Trusted( OALAVLTreeItem_t item, void *arg )
{
    DOFObjectIDDomain domainID = (DOFObjectIDDomain)item;
    DOFConnection self = (DOFConnection)arg;

    OALConnection_RemoveDynamicTRPRoute( self->OALConn, NULL, domainID );
}

/*  The OALConnection became connected, terminated, or disconnected.  Call the complete callback for each of the
    pending connection nodes in the list and notify all connection state listeners for the connection.
    Note: This function will always be called with the connection mutex locked. */
void DOFConnection_ConnectionEvent(DOFConnection conn, boolean status, DOFConnectionState connState, DOFException exception)
{
    boolean isConnected;

    /* A connection that was attempting to connect and fails will not have a connState.  The complete callbacks must
       be called for each pending connection, but the listeners must not be called. */
    if (connState)
        isConnected = DOFConnectionState_IsConnected(connState);
    else
    {
        isConnected = FALSE;
        /* At least there is *some* exception. */
        if( !exception )
            exception = DOFEXCEPTION_FAILED;
    }


    if (isConnected)
    {
        OALSListNode tempPendList = conn->pendingConnections;

        conn->pendingConnections = NULL;
        /* remove all nodes from the pending connection list and enqueue a complete callback for each node in the list. */
        OALSList_RemoveAll(&tempPendList, DOFConnection_RemovePendingConnections_fn, NULL);
        /* Notify the Connection Manager so the connection state listener functions will be called. */
        OALConnectionMgr_NotifySystem_ConnectionEvent(conn, connState, NULL);
    }
    else
    {
        OALSListNode tempPendList = conn->pendingConnections;

        if (conn->trpDomainRoutes)
        {
            OALAVLTree_Empty( conn->trpDomainRoutes, RemoveAll_DOFDomains, conn );
        }
        if (conn->trpTrustedRoutes)
        {
            OALAVLTree_Empty( conn->trpTrustedRoutes, RemoveAll_Trusted, conn );
        }

        conn->pendingConnections = NULL;
        /* remove all nodes from the pending connection list and enqueue a complete callback for each node in the list.
           The exception is required to indicate the pending connection failed to connect. */
        OALSList_RemoveAll(&tempPendList, DOFConnection_RemovePendingConnections_fn, exception);

        /* If there is connection state then the connection was previously connected.  Notify all connections so they
           can in turn notify all listeners. */
        if (connState)
            OALConnectionMgr_NotifySystem_ConnectionEvent(conn, connState, NULL);
    }
}

boolean DOFConnection_Reconnect( DOFConnection self, uint32 timeout, DOFException * pException )
{
    if (pException)
    {
        DOFConnection_Connect(self, timeout, pException);
        if(! *pException)
        {
            DOFConnection_Disconnect(self);
            return TRUE;
        }
    }
    else
    {
        DOFException    except;

        DOFConnection_Connect(self, timeout, &except);
        if( !except )
        {
            DOFConnection_Disconnect(self);
            return TRUE;
        }
        else
        {
            DOFException_Destroy(except);
        }
    }
    return FALSE;
}

static void begin_reconnect_complete(DOFConnectionConnectCallback self, const DOFConnection connection, void * context, const DOFException except)
{
    if(!except) DOFConnection_Disconnect(connection);
}

static const struct DOFConnectionConnectCallbackFns_t reconnect_fn = {begin_reconnect_complete};
static DOFConnectionConnectCallback_t reconnect_callback = {&reconnect_fn};

void DOFConnection_BeginReconnect( DOFConnection self, uint32 timeout, DOFConnectionConnectCallback callback, void * context )
{
    DOFConnection_BeginConnect(self, timeout, &reconnect_callback, context);
}

boolean DOFConnection_AddDomain( DOFConnection self, DOFDomainConfig domainCfg, uint32 timeout, DOFException* exception )
{
    if( self && domainCfg )
    {
        DOFDomain domain = DOF_CreateDomain( DOF_GetDof(), domainCfg );
        if( domain)
        {
            if( DOFConnection_AddDynamicTRPRoute( self, domain, NULL, timeout, exception ) )
            {
                DOFDomain_Destroy( domain );
                return TRUE;
            }
            DOFDomain_Destroy( domain );
        }
        if( exception )
            (*exception) =  DOFException_Copy( DOFEXCEPTION_INSUFFICIENT_RESOURCES );
    }
    return FALSE;
}

void DOFConnection_RemoveDomain( DOFConnection self, DOFDomainConfig domainCfg )
{
    if( self && domainCfg )
    {
        DOFDomain domain = DOF_CreateDomain( DOF_GetDof(), domainCfg );
        if( domain)
        {
            DOFConnection_RemoveDynamicTRPRoute( self, domain, NULL );
            DOFDomain_Destroy( domain );
        }
    }
}

boolean DOFConnection_AddTrustedDomain( DOFConnection self, DOFObjectIDDomain domainID )
{
    if( self && domainID )
    {
        return DOFConnection_AddDynamicTRPRoute( self, NULL, domainID, DOF_TIMEOUT_NEVER, NULL );
    }
    return FALSE;
}

void DOFConnection_RemoveTrustedDomain( DOFConnection self, DOFObjectIDDomain domainID )
{
    if( self && domainID )
    {
        DOFConnection_RemoveDynamicTRPRoute( self, NULL, domainID );
    }
}

boolean DOFConnection_AddDynamicTRPRoute( DOFConnection self, DOFDomain domain, DOFObjectIDDomain domainID, uint32 timeout, DOFException* exception )
{
    OALAVLTree_InsertionStatus status;

    PCRMutex_Lock(connection_server_mutex);
    if( DOFConnection_IsConnected( self ) )
    {
        if( domain )
            status = OALAVLTree_Insert( self->trpDomainRoutes, domain, domain, NULL );
        else
            status = OALAVLTree_Insert( self->trpTrustedRoutes, domainID, domainID, NULL );

        if( status != OALAVLTree_InsertSuccess )
        {
            goto FAIL;
        }

        if( OALConnection_AddDynamicTRPRoute( self->OALConn, domain, domainID, timeout, exception ) )
        {
            PCRMutex_Unlock(connection_server_mutex);
            return TRUE;
        }

        if( domain )
            OALAVLTree_Delete( self->trpDomainRoutes, domain );
        else
            OALAVLTree_Delete( self->trpTrustedRoutes, domainID );
    }

FAIL:
    PCRMutex_Unlock(connection_server_mutex);

    if( exception )
        (*exception) =  DOFException_Copy( DOFEXCEPTION_FAILED );

    return FALSE;
}

void DOFConnection_RemoveDynamicTRPRoute( DOFConnection self, DOFDomain domain, DOFObjectIDDomain domainID )
{
    boolean status;

    PCRMutex_Lock(connection_server_mutex);

    if( domain )
        status = OALAVLTree_Delete( self->trpDomainRoutes, domain );
    else
        status = OALAVLTree_Delete( self->trpTrustedRoutes, domainID );

    if( status )
    {
        OALConnection_RemoveDynamicTRPRoute( self->OALConn, domain, domainID );
    }
    PCRMutex_Unlock(connection_server_mutex);
}

void DOFConnectionList_Destroy(DOFConnectionList self)
{
    if (self)
    {
        while(self->connectionCount--)
        {
            DOFConnection_Destroy(self->connectionArray[self->connectionCount]);
        }

        DOFMem_Free(self->connectionArray);
        DOFMem_Free(self);
    }
}
