/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/address.h>
#include <dof/bridge.h>
#include <dof/connection.h>
#include <dof/credentials.h>
#include <dof/domain.h>
#include <dof/filter.h>
#include <dof/permissionset.h>
#include <dof/transport.h>
#include <dof/objectid.h>
#include <dof/dof.h>

#include <internal/addressinternal.h>
#include <internal/credentialsinternal.h>
#include <internal/connection_internal.h>
#include <internal/permissionsetinternal.h>
#include <internal/securitymode.h>

#include "connection_local.h"

#define DEFAULT_MAX_RECEIVE_SILENCE 120000

#define ERROR_RECEIVEFILTER     0x01
#define ERROR_SENDFILTER        0x02
#define ERROR_OPERATIONBRIDGE   0x04

/* Return true if the two connection configs are equivalent. */
boolean DOFConnectionConfig_IsEqual(DOFConnectionConfig self, DOFConnectionConfig config)
{
    PCR_Assert(self);
    PCR_Assert(config);

    if (self->type != config->type
        || self->negotiator != config->negotiator
        || self->securityDesire != config->securityDesire
        || self->streamListener != config->streamListener)
    {
        return FALSE;
    }

    if (self->maxReceiveSilence != config->maxReceiveSilence)
        return FALSE;

    if (self->maxSendSilence != config->maxSendSilence)
        return FALSE;

    if (!DOFAddress_IsEqual(self->targetAddress, config->targetAddress))
        return FALSE;

    if (!DOFCredentials_IsEqual(self->credentials, config->credentials))
        return FALSE;

    if (!DTLConnectionConfig_IsEqual(self->transportConfig, config->transportConfig))
        return FALSE;

    if( !OALDomainCredentialsTree_IsEqual( self->domainCreds, config->domainCreds ) )
        return FALSE;

    /*  The permissions are not compared here.  If the credentials and everything else is equal then
        they both have the potential of negotiating the same permissions. */

    return TRUE;
}

/*
    create a DOFConnectionConfig structure.

    The return value is null if target is NULL or the alloc fails.
 */
DOFConnectionConfigBuilder DOFConnectionConfigBuilder_Create( DOFConnectionType type,
                                                              DOFAddress target )
{
    DOFConnectionConfig ret = NULL;

    if ( target )
    {
        ret = (DOFConnectionConfig) DOFMem_Alloc( sizeof(DOFConnectionConfig_t) );

        if (ret)
        {
            PCRMem_Set( ret, 0, sizeof(DOFConnectionConfig_t) );
            ret->targetAddress = DOFAddress_Copy( target );
            if (ret->targetAddress)
            {
                ret->refCount = 1;
                ret->bridge = NULL;
                ret->receiveFilter = NULL;
                ret->sendFilter = NULL;
                ret->securityDesire = DOFSECURITYDESIRE_ANY;
                ret->type = type;
                ret->negotiator = DOF_DEFAULT_PROTOCOL_NEGOTIATOR;
                ret->maxSendSilence = DOF_TIMEOUT_NEVER;
                ret->maxReceiveSilence = DEFAULT_MAX_RECEIVE_SILENCE;
                ret->flags.isExtendAllowed = TRUE;
                ret->errors = 0;

                if (ret->type == DOFCONNECTIONTYPE_DATAGRAM &&
                    DOFAddress_GetAddressType(ret->targetAddress) != DOFADDRESSTYPE_UNICAST)
                {
                    ret->type = DOFCONNECTIONTYPE_DATAGRAM_STATELESS;
                }
            }
            else
            {
                DOFMem_Free(ret);
                ret = NULL;
            }
        }
    }
    return (DOFConnectionConfigBuilder) ret;
}

DOFConnectionConfigBuilder DOFConnectionConfigBuilder_Create_Config( DOFConnectionConfig config )
{
    return (DOFConnectionConfigBuilder) DOFConnectionConfigBuilder_Build( (DOFConnectionConfigBuilder) config );
}

DOFConnectionConfigBuilder DOFConnectionConfigBuilder_Create_State( DOFConnectionState state )
{
    DOFConnectionConfigBuilder ccbuilder;

    /* Copy the non-credential related stuff from the state's config into the new builder.
       The credentials, domains, filters, and bridge are not copied. */

    ccbuilder = DOFConnectionConfigBuilder_Create(DOFConnectionState_GetConnectionType(state), DOFConnectionState_GetAddress(state));
    if (ccbuilder)
    {
        if (DOFConnectionConfigBuilder_SetSecurityDesire(ccbuilder, DOFConnectionState_GetSecurityDesire(state)))
        {
            if (DOFConnectionConfigBuilder_SetPermissions(ccbuilder, DOFConnectionState_GetPermissions(state)))
            {
                if (DOFConnectionConfigBuilder_SetTransportConfig(ccbuilder, DOFConnectionState_GetTransportConfig(state)))
                {
                    if (DOFConnectionConfigBuilder_SetProtocolNegotiator(ccbuilder, DOFConnectionState_GetProtocolNegotiator(state)))
                    {
                        if (DOFConnectionConfigBuilder_SetMaxSendSilence(ccbuilder, DOFConnectionState_GetMaxSendSilence(state)))
                        {
                            if (DOFConnectionConfigBuilder_SetMaxReceiveSilence(ccbuilder, DOFConnectionState_GetMaxReceiveSilence(state)))
                            {
                                return ccbuilder;
                            }
                        }
                    }
                }
            }
        }
        DOFConnectionConfigBuilder_Destroy(ccbuilder);
    }

    return NULL;
}

DOFConnectionConfig DOFConnectionConfig_Copy( const DOFConnectionConfig self )
{
    if ( self )
    {
        PCRMutex_Lock(ref_count_mutex);
        self->refCount++;
        PCRMutex_Unlock(ref_count_mutex);
    }
    return self;
}


void DOFConnectionConfigBuilder_Destroy( DOFConnectionConfigBuilder self )
{
    DOFConnectionConfig_Destroy( (DOFConnectionConfig) self );
}

DOFConnectionConfig DOFConnectionConfigBuilder_BuildAndDestroy( DOFConnectionConfigBuilder selfx )
{
    DOFConnectionConfig ret = (DOFConnectionConfig)selfx;

    PCR_Assert( ret );
    PCR_Assert( ret->targetAddress );

    if (ret->credentials != NULL && !ret->permissions)
    {
        /* A secure connection must have a permission set in its connection config, so give it an empty one. */
        if ((ret->permissions = DOFPermissionSet_Create()) == NULL)
            ret->flags.setPermissionsFailed = TRUE;
    }
    if (ret->domainCreds)
    {
        if (ret->type == DOFCONNECTIONTYPE_HUB)
        {
            /* Hub connections currently do not support domain credentials.  See bug 3801.
               This check must be removed as part of adding support for domain credentials to hub connections. */
            OALDomainCredentialsTree_Destroy(ret->domainCreds);
            ret->domainCreds = NULL;
        }
    }
    if ((ret->type == DOFCONNECTIONTYPE_DATAGRAM) && (DOFAddress_GetAddressType(ret->targetAddress) != DOFADDRESSTYPE_UNICAST))
    {
        /* Multicast is always stateless. */
        ret->type = DOFCONNECTIONTYPE_DATAGRAM_STATELESS;
    }
    if ((ret->type == DOFCONNECTIONTYPE_DATAGRAM_STATELESS) && (ret->securityDesire == DOFSECURITYDESIRE_ANY))
    {
        /* Stateless cannot be secure so change DOFSECURITYDESIRE_ANY to DOFSECURITYDESIRE_NOT_SECURE. */
        ret->securityDesire = DOFSECURITYDESIRE_NOT_SECURE;
    }

    if (ret->flags.setAddressFailed ||
        ret->flags.setCredentialsFailed ||
        ret->flags.setPermissionsFailed ||
        ret->flags.setTransportConfigFailed ||
        (ret->errors != 0))
    {
        DOFConnectionConfig_Destroy(ret);
        ret = NULL;
        DOFLog_Message( DOFLOGLEVEL_WARNING, "DOFConnectionConfigBuilder: One or more of the connection config builder set functions failed.");
    }

    if( ret && !ret->flags.isExtendAllowed)
    {
        DOFPermissionSet_SetGrantPerms(ret->permissions);
    }

    return ret;
}

DOFConnectionConfig DOFConnectionConfigBuilder_Build( DOFConnectionConfigBuilder selfx )
{
    /* Deep copy to make the result immutable. */
    DOFConnectionConfig self = (DOFConnectionConfig)selfx;
    DOFConnectionConfig ret = NULL;

    PCR_Assert( self );
    PCR_Assert( self->targetAddress );

    if (!self->flags.setAddressFailed &&
        !self->flags.setCredentialsFailed &&
        !self->flags.setPermissionsFailed &&
        !self->flags.setTransportConfigFailed &&
        (self->errors == 0))
    {
        ret = (DOFConnectionConfig) DOFConnectionConfigBuilder_Create( self->type, self->targetAddress );
    }

    if ( ret )
    {
        ret->transportConfig = DTLConnectionConfig_Copy( self->transportConfig );
        ret->bridge = self->bridge;
        if (ret->transportConfig || !self->transportConfig)
        {
            ret->streamListener = DOFConnectionSRL_Copy(self->streamListener);
            if (ret->streamListener || !self->streamListener)
            {
                if( self->domainCreds && (self->type != DOFCONNECTIONTYPE_HUB) )
                {
                    /* Hub connections currently do not support domain credentials.  See bug 3801.
                       This check and the one in the following check must be removed as part of
                       adding support for domain credentials to hub connections. */
                    ret->domainCreds = OALDomainCredentialsTree_DeepCopy( self->domainCreds );
                }
                if (ret->domainCreds || !self->domainCreds || (self->type == DOFCONNECTIONTYPE_HUB))
                {
                    ret->credentials = DOFCredentials_Copy( self->credentials );
                    if (ret->credentials || !self->credentials)
                    {
                        ret->neuteredCredentials = DOFCredentials_CreateNeutered( ret->credentials );
                        if( ret->neuteredCredentials || !ret->credentials )
                        {
                            ret->permissions  = DOFPermissionSet_Copy(self->permissions);
                            if( ret->permissions || !self->permissions )
                            {
                                if (ret->credentials && !ret->permissions)
                                {
                                    /* A secure connection must have a permission set in its connection config, so give it an empty one. */
                                    ret->permissions = DOFPermissionSet_Create();
                                    if (ret->permissions == NULL)
                                    {
                                        ret->flags.setPermissionsFailed = TRUE;
                                    }
                                }

                                if (!ret->flags.setPermissionsFailed)
                                {
                                    ret->securityDesire = self->securityDesire;
                                    ret->negotiator = self->negotiator;
                                    ret->maxSendSilence = self->maxSendSilence;
                                    ret->maxReceiveSilence = self->maxReceiveSilence;

                                    if ((ret->type == DOFCONNECTIONTYPE_DATAGRAM) && (DOFAddress_GetAddressType(ret->targetAddress) != DOFADDRESSTYPE_UNICAST))
                                    {
                                        /* Multicast is always stateless. */
                                        ret->type = DOFCONNECTIONTYPE_DATAGRAM_STATELESS;
                                    }
                                    if ((ret->type == DOFCONNECTIONTYPE_DATAGRAM_STATELESS) && (ret->securityDesire == DOFSECURITYDESIRE_ANY))
                                    {
                                        /* Stateless cannot be secure so change DOFSECURITYDESIRE_ANY to DOFSECURITYDESIRE_NOT_SECURE. */
                                        ret->securityDesire = DOFSECURITYDESIRE_NOT_SECURE;
                                    }
                                    ret->flags.isExtendAllowed = self->flags.isExtendAllowed;
                                    if(!ret->flags.isExtendAllowed)
                                    {
                                        DOFPermissionSet_SetGrantPerms(ret->permissions);
                                    }
                                    return ret;
                                }
                            }
                        }
                    }
                }
            }
        }
        /* There was an error filling the connection config so just destroy it. */
        DOFConnectionConfig_Destroy(ret);
    }
    return NULL;
}

DOFConnectionConfigBuilder DOFConnectionConfigBuilder_Copy( DOFConnectionConfigBuilder builder )
{
    if ( builder )
    {
        return (DOFConnectionConfigBuilder) DOFConnectionConfigBuilder_Build( builder );
    }
    return NULL;
}

void DOFConnectionConfig_Destroy( DOFConnectionConfig self )
{
    if (self)
    {
        boolean gone;
        PCRMutex_Lock(ref_count_mutex);
        PCR_Assert(self->refCount);
        --(self->refCount);
        gone = !self->refCount;
        PCRMutex_Unlock(ref_count_mutex);
        if (gone)
        {
            DOFOperationBridgeConfig_Destroy(self->bridge);
            DOFOperationFilter_Destroy(self->receiveFilter);
            DOFOperationFilter_Destroy(self->sendFilter);
            DOFConnectionSRL_Destroy(self->streamListener);
            DOFAddress_Destroy( self->targetAddress);
            DOFCredentials_Destroy( self->credentials );
            DOFCredentials_Destroy( self->neuteredCredentials );
            DOFPermissionSet_Destroy( self->permissions );
            DTLConnectionConfig_Destroy(self->transportConfig);
            OALDomainCredentialsTree_Destroy( self->domainCreds );
            DOFMem_Free( self );
        }
    }
}

/* Change the connection type. */
boolean DOFConnectionConfigBuilder_SetConnectionType( const DOFConnectionConfigBuilder selfx, DOFConnectionType type )
{
    DOFConnectionConfig self = (DOFConnectionConfig)selfx;
    if (self)
    {
        self->type = type;
        return TRUE;
    }
    return FALSE;
}

/* Set functions. */
/* Change the target address*/
boolean DOFConnectionConfigBuilder_SetAddress( const DOFConnectionConfigBuilder selfx, DOFAddress addr )
{
    DOFConnectionConfig self = (DOFConnectionConfig)selfx;
    /* Return without changes if addr is null. */
    if (self && addr)
    {
        self->flags.setAddressFailed = FALSE;
        DOFAddress_Destroy(self->targetAddress);
        self->targetAddress = DOFAddress_Copy(addr);
        if (self->targetAddress)
        {
            self->flags.setAddressFailed = FALSE;
            return TRUE;
        }
        else
        {
            self->flags.setAddressFailed = TRUE;
            DOFLog_Message( DOFLOGLEVEL_WARNING, "DOFConnectionConfigBuilder: Failed to copy address.");
        }
    }
    return FALSE;
}

/* Change the securityMode */
boolean DOFConnectionConfigBuilder_SetSecurityDesire( const DOFConnectionConfigBuilder selfx, DOFSecurityDesire desire )
{
    DOFConnectionConfig self = (DOFConnectionConfig)selfx;
    if (self)
    {
        self->securityDesire = desire;
        return TRUE;
    }
    return FALSE;
}

/* Change the credentials */
boolean DOFConnectionConfigBuilder_SetCredentials( DOFConnectionConfigBuilder selfx, DOFCredentials auth )
{
    DOFConnectionConfig self = (DOFConnectionConfig)selfx;
    if ( self )
    {
        self->flags.setCredentialsFailed = FALSE;
        DOFCredentials_Destroy( self->credentials );

        if ( auth )
        {
            self->credentials = DOFCredentials_Copy( auth );
            if (!self->credentials || !DOFCredentials_ContainsPrivateData( self->credentials ))
            {
                self->flags.setCredentialsFailed = TRUE;
                DOFLog_Message( DOFLOGLEVEL_WARNING, "DOFConnectionConfigBuilder: Failed to copy credentials.");
            }
            else
            {
                return TRUE;
            }
        }
        else
        {
            /* NULL credentials are allowed. */
            self->credentials = NULL;
            return TRUE;
        }
    }
    return FALSE;
}

boolean DOFConnectionConfigBuilder_AddDomain( DOFConnectionConfigBuilder selfx, DOFDomainConfig auth )
{
    DOFConnectionConfig self = (DOFConnectionConfig)selfx;
    if( self && auth )
    {
        if( !self->domainCreds )
            self->domainCreds = OALDomainCredentialsTree_Create();
        if( self->domainCreds )
        {
            DOFDomain domain = DOF_CreateDomain( DOF_GetDof(), auth );
            if( domain )
            {
                boolean ret = OALDomainCredentialsTree_Insert( self->domainCreds, domain );
                DOFDomain_Destroy( domain );
                return ret;
            }
        }
    }
    return FALSE;
}

boolean DOFConnectionConfigBuilder_AddTrustedDomain( DOFConnectionConfigBuilder selfx, DOFObjectIDDomain domainID )
{
    DOFConnectionConfig self = (DOFConnectionConfig)selfx;
    if( self && domainID )
    {
        if( !self->domainCreds )
            self->domainCreds = OALDomainCredentialsTree_Create();
        if( self->domainCreds )
            return OALDomainCredentialsTree_InsertTrustedDomain( self->domainCreds, domainID );
    }
    return FALSE;
}

boolean DOFConnectionConfigBuilder_SetDomainDiscoveryCredentials( DOFConnectionConfigBuilder selfx, DOFCredentials credentials )
{
    DOFConnectionConfig self = (DOFConnectionConfig)selfx;
    if( self )
    {
        if( !self->domainCreds )
            self->domainCreds = OALDomainCredentialsTree_Create();
        if( self->domainCreds )
            return OALDomainCredentialsTree_SetDomainDiscoveryCredentials( self->domainCreds, credentials );
    }
    return FALSE;
}

/* Change the permissions */
boolean DOFConnectionConfigBuilder_SetPermissions( const DOFConnectionConfigBuilder selfx, DOFPermissionSet perms )
{
    DOFConnectionConfig self = (DOFConnectionConfig)selfx;
    if (self)
    {
        self->flags.setPermissionsFailed = FALSE;
        DOFPermissionSet_Destroy(self->permissions);
        self->permissions = DOFPermissionSet_Copy(perms);
        /* Null permissions are allowed. */
        if ( perms && !self->permissions )
        {
            self->flags.setPermissionsFailed = TRUE;
            DOFLog_Message(  DOFLOGLEVEL_WARNING, "DOFConnectionConfigBuilder: Failed to copy permissions.");
        }
        else
        {
            return TRUE;
        }
    }
    return FALSE;
}

boolean DOFConnectionConfigBuilder_SetPermissionsExtendAllowed(DOFConnectionConfigBuilder self,
                                                               boolean                    extendAllowed)
{
    if(self != NULL)
    {
        DOFConnectionConfig selfx = (DOFConnectionConfig) self;

        selfx->flags.isExtendAllowed = extendAllowed;
        return TRUE;
    }
    return FALSE;
}

boolean DOFConnectionConfig_IsPermissionsExtendAllowed(DOFConnectionConfig const self)
{
    if(self != NULL)
    {
        return self->flags.isExtendAllowed;
    }
    return FALSE;
}

/* Change the transport config */
boolean DOFConnectionConfigBuilder_SetTransportConfig( DOFConnectionConfigBuilder selfx, DTLConnectionConfig config )
{
    DOFConnectionConfig self = (DOFConnectionConfig)selfx;
    if (self)
    {
        self->flags.setTransportConfigFailed = FALSE;
        DTLConnectionConfig_Destroy(self->transportConfig);
        self->transportConfig = DTLConnectionConfig_Copy(config);
        if ( config && !self->transportConfig )
        {
            self->flags.setTransportConfigFailed = TRUE;
            DOFLog_Message(  DOFLOGLEVEL_WARNING, "DOFConnectionConfigBuilder: Failed to copy DTLConnection config.");
        }
        else
        {
            return TRUE;
        }
    }
    return FALSE;
}

boolean DOFConnectionConfigBuilder_SetMaxSendSilence( DOFConnectionConfigBuilder selfx, uint32 maxSilence )
{
    DOFConnectionConfig self = (DOFConnectionConfig)selfx;
    if (self && (maxSilence > 0))
    {
        self->maxSendSilence = maxSilence;
        return TRUE;
    }
    return FALSE;
}

boolean DOFConnectionConfigBuilder_SetMaxReceiveSilence( DOFConnectionConfigBuilder selfx, uint32 maxSilence )
{
    DOFConnectionConfig self = (DOFConnectionConfig)selfx;
    if (self && (maxSilence > 0))
    {
        self->maxReceiveSilence = maxSilence;
        return TRUE;
    }
    return FALSE;
}

static void setError(uint32  * errorFlags,
                     uint32    errorCode,
                     boolean   worked)
{
    if(worked)
    {
        *errorFlags &= ~errorCode;
    }
    else
    {
        *errorFlags |= errorCode;
    }
}

boolean DOFConnectionConfigBuilder_SetBridge( DOFConnectionConfigBuilder self, DOFOperationBridgeConfig bridge )
{
    DOFConnectionConfig selfx = (DOFConnectionConfig)self;

    boolean retVal = FALSE;

    if(self != NULL)
    {
        DOFOperationBridgeConfig newBridge = DOFOperationBridgeConfig_Copy(bridge);

        retVal = ((newBridge != NULL) || (bridge == NULL));
        DOFOperationBridgeConfig_Destroy(selfx->bridge);
        selfx->bridge = newBridge;
        setError(&(selfx->errors), ERROR_OPERATIONBRIDGE, retVal);
    }
    return retVal;
}

boolean DOFConnectionConfigBuilder_SetProtocolNegotiator(DOFConnectionConfigBuilder selfx, DOFProtocolNegotiator negotiator)
{
    DOFConnectionConfig self = (DOFConnectionConfig)selfx;
    if(self)
    {
        self->negotiator = negotiator;
        return TRUE;
    }
    return FALSE;
}

boolean DOFConnectionConfigBuilder_SetReceiveFilter( DOFConnectionConfigBuilder self, DOFOperationFilter filter )
{
    DOFConnectionConfig selfx = (DOFConnectionConfig)self;

    boolean retVal = FALSE;

    if(self != NULL)
    {
        DOFOperationFilter newFilter = DOFOperationFilter_Copy(filter);

        retVal = ((newFilter != NULL) || (filter == NULL));
        DOFOperationFilter_Destroy(selfx->receiveFilter);
        selfx->receiveFilter = newFilter;
        setError(&(selfx->errors), ERROR_RECEIVEFILTER, retVal);
    }
    return retVal;
}

boolean DOFConnectionConfigBuilder_SetSendFilter( DOFConnectionConfigBuilder self, DOFOperationFilter filter )
{
    DOFConnectionConfig selfx = (DOFConnectionConfig)self;

    boolean retVal = FALSE;

    if(self != NULL)
    {
        DOFOperationFilter newFilter = DOFOperationFilter_Copy(filter);

        retVal = ((newFilter != NULL) || (filter == NULL));
        DOFOperationFilter_Destroy(selfx->sendFilter);
        selfx->sendFilter = newFilter;
        setError(&(selfx->errors), ERROR_SENDFILTER, retVal);
    }
    return retVal;
}

boolean DOFConnectionConfigBuilder_SetStreamRequestListener( DOFConnectionConfigBuilder self, DOFConnectionStreamRequestListener streamListener)
{
    DOFConnectionConfig selfx = (DOFConnectionConfig)self;

    if (selfx)
    {
        if (selfx->streamListener)
        {
            DOFConnectionSRL_Destroy(selfx->streamListener);
        }
        selfx->streamListener = NULL;

        if (streamListener)
        {
            selfx->streamListener = DOFConnectionSRL_Create(streamListener);
            if (selfx->streamListener)
            {
                return TRUE;
            }
        }
        else
        {
            return TRUE;
        }
    }

    return FALSE;
}

/* Get functions. */

DOFConnectionType DOFConnectionConfig_GetConnectionType(const DOFConnectionConfig self )
{
    if (self)
        return self->type;

    return 0;
}

DOFAddress DOFConnectionConfig_GetAddress( const DOFConnectionConfig self )
{
    if (self)
        return self->targetAddress;

    return NULL;
}

DOFSecurityDesire DOFConnectionConfig_GetSecurityDesire( const DOFConnectionConfig self )
{
    if (self)
        return self->securityDesire;
    return 0;
}

DOFCredentials DOFConnectionConfig_GetCredentials( const DOFConnectionConfig self )
{
    if (self)
        return self->credentials;
    return NULL;
}

DOFCredentials DOFConnectionConfig_GetNeuteredCredentials( const DOFConnectionConfig self )
{
    if (self)
        return self->neuteredCredentials;
    return NULL;
}

DOFCredentials DOFConnectionConfig_GetDomainDiscoveryCredentials( const DOFConnectionConfig self )
{
    if (self)
        return OALDomainCredentialsTree_GetDomainDiscoveryCredentials( self->domainCreds );
    return NULL;
}

DOFOperationBridgeConfig DOFConnectionConfig_GetBridge( DOFConnectionConfig const self )
{
    if(self != NULL)
    {
        return self->bridge;
    }
    return NULL;
}

DOFOperationFilter DOFConnectionConfig_GetReceiveFilter( DOFConnectionConfig const self )
{
    if(self != NULL)
    {
        return self->receiveFilter;
    }
    return NULL;
}

DOFOperationFilter DOFConnectionConfig_GetSendFilter( DOFConnectionConfig const self )
{
    if(self != NULL)
    {
        return self->sendFilter;
    }
    return NULL;
}

DOFPermissionSet DOFConnectionConfig_GetPermissions( const DOFConnectionConfig self )
{
    if (self)
        return self->permissions;
    return NULL;
}

DTLConnectionConfig DOFConnectionConfig_GetTransportConfig( const DOFConnectionConfig self )
{
    if (self)
        return self->transportConfig;
    return NULL;
}

DOFProtocolNegotiator DOFConnectionConfig_GetProtocolNegotiator(DOFConnectionConfig self)
{
    return self ? self->negotiator : NULL;
}

uint32 DOFConnectionConfig_GetMaxSendSilence( const DOFConnectionConfig self )
{
    if (self)
        return self->maxSendSilence;
    return 0;
}

uint32 DOFConnectionConfig_GetMaxReceiveSilence( const DOFConnectionConfig self )
{
    if (self)
        return self->maxReceiveSilence;

    return 0;
}

DOFConnectionStreamRequestListener DOFConnectionConfig_GetStreamRequestListener( const DOFConnectionConfig self )
{
    if (self)
        return DOFConnectionSRL_GetStreamRequestListener(self->streamListener);

    return NULL;
}



