/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/exception.h>
#include <dof/error.h>
#include <internal/securityscope.h>

#include <internal/dofinternal.h>
#include <internal/connectionmanager.h>
#include <internal/enqueue.h>
#include <internal/avl_multitree.h>
#include <internal/avl_tree.h>
#include <internal/avl_treelist.h>
#include <internal/connection_internal.h>
#include <internal/pointerlist.h>
#include <internal/mutex_pool.h>
#include <internal/serverinternal.h>
#include <internal/slist.h>
#include <internal/remotechannel.h>
#include <internal/runtimeinternal.h>

#include "connection_local.h"

typedef struct cbkInfo_t
{
    DOFConnection       connection;
    DOFConnectionState  state;
}cbkInfo_t;

extern PCRMutex connection_server_mutex;

static OALEnqueue       callback_enqueue = NULL;
static OALAVLTreeList   OALConnection_tree = NULL;

static DOFConnection    delayedDestroyList;     /* Used for building an slist of DOFConnections. */
static boolean          delayDestroy;

static boolean OALConnection_tree_CompareGreaterThan_fn(OALAVLTreeListKey_t key, OALAVLTreeListItem_t item, void *context)
{
    DOFAddress addr = (DOFAddress)key;
    OALConnection OALConn = (OALConnection)item;

    return (DOFAddress_Compare( addr, DOFConnectionConfig_GetAddress(OALConn->connConfig)) > 0);
}

static boolean OALConnection_tree_CompareEqual_fn(OALAVLTreeListKey_t key, OALAVLTreeListItem_t item, void *context)
{
    DOFAddress addr = (DOFAddress)key;
    OALConnection OALConn = (OALConnection)item;

    return DOFAddress_IsEqual( addr, DOFConnectionConfig_GetAddress(OALConn->connConfig) );
}

boolean OALConnectionMgr_Init(void)
{
    callback_enqueue = OALEnqueue_Create(DOF_GetDof()->threadPool);
    if (callback_enqueue)
    {
        OALConnection_tree = OALAVLTreeList_Create(OALConnection_tree_CompareGreaterThan_fn,
                                                   OALConnection_tree_CompareEqual_fn, NULL);
        if (OALConnection_tree)
        {
            return TRUE;
        }
        OALEnqueue_Destroy(callback_enqueue);
    }

    return FALSE;
}

/* The connection mutex must be locked when this is called. */
static void OALConnectionMgr_StreamRequestTree_RemoveAll(OALAVLTreeItem_t item, void *arg)
{
    DOFConnectionStreamRequest streamRequest = (DOFConnectionStreamRequest)item;

    PCR_Assert(streamRequest);

    if (streamRequest->requestCount > 0)
    {
        OALConnectionMgr_NotifySystem_StreamListenerEvent(streamRequest, TRUE);
    }
}

/* The connection mutex must be locked when this is called. */
/* OALConn must not be NULL. */
void OALConnection_DestroyStreamRequestTree(OALConnection OALConn)
{
    if (OALConn->streamRequestTree)
    {
        OALAVLTree_Destroy(OALConn->streamRequestTree, OALConnectionMgr_StreamRequestTree_RemoveAll, NULL);
        OALConn->streamRequestTree = NULL;
    }
}

/* The connection mutex must be locked when this is called. */
/* OALConn must not be NULL. */

void OALConnection_EmptyStreamRequestTree(OALConnection OALConn)
{
    if (OALConn->streamRequestTree)
    {
        OALAVLTree_Empty(OALConn->streamRequestTree, OALConnectionMgr_StreamRequestTree_RemoveAll, NULL);
    }
}

/* The connection mutex must be locked when this is called. */
static void OALConnectionMgr_OALConnection_tree_Destroy(OALAVLTreeListItem_t item, void *arg)
{
    OALConnection OALConn = (OALConnection)item;

    /* If this is an inbound connection send a shutdown message to the channel. */
    if (!OALConn->flags.isOutbound && OALConn->channel)
    {
        OALRemoteChannel_ConnectionInitiatedShutdown(OALConn->channel, OALConn);
        OALConnection_ShutdownDOFConnections(OALConn);
        OALConnection_ClearChannel(OALConn);
    }

    if (OALConn->flags.isOutbound)
    {
        /*  Take care of everyting that must happen before the callback enqueue is destroyed. */
        OALConnection_ShutdownDOFConnections(OALConn);
        OALConnection_DestroyStreamRequestTree(OALConn);
    }
}

void OALConnectionMgr_CloseInboundConnections(void)
{
    PCRMutex_Lock(connection_server_mutex);
    OALAVLTreeList_Destroy( OALConnection_tree, OALConnectionMgr_OALConnection_tree_Destroy, NULL);

    OALConnection_tree = NULL;
    PCRMutex_Unlock(connection_server_mutex);
}

static void OALConnectionMgr_CompleteConnectionDestroy_Immediate(DOFConnection connection)
{
    PCR_Assert(connection);

    /* These must not be destroyed until after all listener callbacks have been called. */
    DOFServer_Release( connection->relatedServer );
    DOFConnection_Destroy(connection->relatedConnection);

    OALConnection_Release(connection->OALConn);

    DOFMem_Free(connection);
}

static void OALConnectionMgr_delayedDestroyList_Destroy_fn(OALSListNode item, void * notUsed)
{
    OALConnectionMgr_CompleteConnectionDestroy_Immediate((DOFConnection)item);
}

void OALConnectionMgr_Shutdown(void)
{
    OALEnqueue tempEnqueue;

    PCRMutex_Lock(connection_server_mutex);
    tempEnqueue = callback_enqueue;
    callback_enqueue = NULL;
    OALConnection_tree = NULL;
    /* Until the callback_enqueue is fully destroyed, add complete connection destroy calls to the delayed destroy list. */
    delayDestroy = TRUE;
    PCRMutex_Unlock(connection_server_mutex);

    OALEnqueue_Destroy(tempEnqueue);

    /*  Clear delayDestroy flag with the mutex held so that we don't clear the flag while something is being inserted.
        Connections that are destroyed after callback_enqueue is destroyed and the delayDestroy flag is cleared will
        be destroyed immediately by the call to OALConnectionMgr_NotifySystem_Destroyed. */
    PCRMutex_Lock(connection_server_mutex);
    delayDestroy = FALSE;
    PCRMutex_Unlock(connection_server_mutex);
    /* Destroy the delayedDestroyList. */
    OALSList_RemoveAll((OALSListNode *)&delayedDestroyList, OALConnectionMgr_delayedDestroyList_Destroy_fn, NULL);
}

/* This function us used as a OALAVLTreeList_Traversal_fn_t to locate an OALConnection in the tree
   based on the DOFConnectionConfig */
static boolean OALConnection_tree_findConnConfig( OALAVLTreeListItem_t theItem, void *context )
{
    OALConnection item = (OALConnection)theItem;
    OALConnection_findContext_t *arg = (OALConnection_findContext_t *)context;

    /* Since the OALConnection_tree is keyed on address and OALAVLTreeList_TraverseList was used this function
       will only be called for OALConnections with the same DOFAddress. */

    if( DOFConnectionConfig_IsEqual(arg->connConfig, item->connConfig) )
    {
        if ((item->base.OALConn == arg->baseOAL) && (item->connSecurityMembrane == arg->connector) && (item->flags.isOutbound))
        {
            arg->OALConn = item;
            return TRUE;
        }
    }

    return FALSE;
}

/*
    This Function is used to find an OALConnection that uses the same DOFConnectionConfig.
    NULL is returned if a matching OALConnection is not found.
    Note: The connectionMutex must be locked when this function is called.
*/
void OALConnectionMgr_FindOALConnection(OALConnection_findContext_t *findCtx)
{
    if (OALConnection_tree)
    {
        OALAVLTreeList_TraverseList(OALConnection_tree, (void*)DOFConnectionConfig_GetAddress(findCtx->connConfig),
                                    OALConnection_tree_findConnConfig, (void*)findCtx);
    }
}

/* Note: The connectionMutex must be locked when this function is called. */
boolean OALConnectionMgr_AddOALConnection(OALConnection OALConn)
{
    boolean ret = FALSE;

    if (OALConnection_tree)
        ret = OALAVLTreeList_Insert(OALConnection_tree, DOFConnectionConfig_GetAddress(OALConn->connConfig), OALConn);

    return ret;
}

static boolean OALConnection_tree_findOALConnection(OALAVLTreeListItem_t item, void *arg)
{
    if (item == arg)
        return TRUE;

    return FALSE;
}

/* Note: The connectionMutex must be locked when this function is called. */
void OALConnectionMgr_RemoveOALConnection(OALConnection OALConn)
{
    if (OALConnection_tree)
    {
        /* Remove the OALConnection from the OALConnection_tree.  Since it was not retained when it was
           inserted into the tree there is nothing to release or free. */
        OALAVLTreeList_FindAndRemove(OALConnection_tree, DOFConnectionConfig_GetAddress(OALConn->connConfig),
                                     OALConnection_tree_findOALConnection, OALConn);
    }
}

/* Note: The connectionMutex must be locked when this function is called. */
uint32 OALConnectionMgr_GetConnectionCount(void)
{
    if (OALConnection_tree)
        return OALAVLTreeList_GetNumItems(OALConnection_tree);

    return 0;
}

static void OALConnectionMgr_callConnectComplete(OALThreadPoolItem dnode, boolean shutdown)
{
    /* Call the connect complete callback. */
    /* Note: No special action is required for the shutdown case. */
    CallConnectComplete_t * node;

    PCR_Assert(dnode != NULL);
    node = (CallConnectComplete_t *) dnode;
    PCR_Assert(node->callback);
    PCR_Assert(node->callback->fns->Complete);

    node->callback->fns->Complete(node->callback, node->connection, node->context, node->except);

    DOFException_Destroy(node->except);
    DOFMem_Free(node);
}

/* Enqueue one connect complete callback.
   Note: this function will take ownership of the cbNode member of the pendingConnection. */
void OALConnectionMgr_enqueueConnectComplete(pendingConnection_t *pendingConnection, DOFException except)
{
    if(pendingConnection != NULL)
    {
        PCRMutex_Lock(connection_server_mutex);

        /* cbNode is pre-allocated so that the complete callback can be enqueued here without failing.  However the
           enqueue will fail if it is being shut down, but that should not happen since the connections
           are shut down before the callback_enqueue. */
        if (pendingConnection->cbNode)
        {
            if (pendingConnection->cbNode->callback)
            {
                if (callback_enqueue)
                {
                    OALEnqueueItem_Init((OALEnqueueItem)pendingConnection->cbNode, OALConnectionMgr_callConnectComplete);
                    pendingConnection->cbNode->except = DOFException_Copy(except);

                    if (!OALEnqueue_Enqueue(callback_enqueue, (OALEnqueueItem)pendingConnection->cbNode))
                    {
                        DOFException_Destroy(except);
                        DOFMem_Free(pendingConnection->cbNode);
                        pendingConnection->cbNode = NULL;
                        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "DOFConnection: Failed to enqueue connect complete callback." );
                    }
                }
                else
                {
                    DOFMem_Free(pendingConnection->cbNode);
                    pendingConnection->cbNode = NULL;
                    /* Since this is shutting down, the channel will be shutdown by remote channel manager so just ignore this. */
                    DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_TRACE, "DOFConnection: Callback enqueue was already destroyed.");
                }

                if (except)
                {
                    DOFConnection_Disconnect(pendingConnection->pendConnection);
                }
            }
            else
            {
                DOFException_Destroy(pendingConnection->cbNode->except);
                DOFMem_Free(pendingConnection->cbNode);
            }

            /* This function will take ownership of cbNode */
            pendingConnection->cbNode = NULL;
        }

        PCRMutex_Unlock(connection_server_mutex);
    }
}

typedef struct CompleteConnectionDestroy_t
{
    OALEnqueueItem_t super;

    DOFConnection connection;
} CompleteConnectionDestroy_t;

static void OALConnectionMgr_CompleteConnectionDestroy( OALThreadPoolItem dnode, boolean shutdown )
{
    CompleteConnectionDestroy_t * node;

    PCR_Assert(dnode != NULL);
    node = (CompleteConnectionDestroy_t *)dnode;

    OALConnectionMgr_CompleteConnectionDestroy_Immediate(node->connection);
    DOFMem_Free(node);
}

static void OALConnectionMgr_ListenerRemoved_enqueue_fn( OALThreadPoolItem dnode, boolean shutdown )
{
    NotifySystem_ListenerRemoved_t * node;

    PCR_Assert(dnode != NULL);
    node = (NotifySystem_ListenerRemoved_t *) dnode;

    PCR_Assert(node->listener);
    PCR_Assert(node->listener->fns->Removed);

    node->listener->fns->Removed(node->listener, node->connection, NULL);

    DOFMem_Free(node);
}

DOFError OALConnectionMgr_NotifySystem_ListenerRemoved(DOFConnection connection, connLstnNode_t *listenNode)
{
    DOFError retStatus = DOFERROR_OK;

    PCR_Assert( connection );
    PCR_Assert(listenNode);
    PCR_Assert(listenNode->listener);

    PCRMutex_Lock(connection_server_mutex);
    if (listenNode->removeNode)
    {
        if (callback_enqueue)
        {
            OALEnqueueItem_Init((OALEnqueueItem)listenNode->removeNode, OALConnectionMgr_ListenerRemoved_enqueue_fn);
            listenNode->removeNode->listener = listenNode->listener;
            listenNode->removeNode->connection = connection;
            if (!OALEnqueue_Enqueue( callback_enqueue, (OALEnqueueItem)listenNode->removeNode))
            {
                DOFMem_Free(listenNode->removeNode);
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_INFO, "DOFConnection: Failed to enqueue listener removed message." );
                retStatus = DOFERROR_FAILED;
            }
        }
        else
        {
            /* The DOF is being shut down so just destroy the remove node and continue. */
            DOFMem_Free(listenNode->removeNode);
        }

        /* This function takes ownership of the remove node. */
        listenNode->removeNode = NULL;
    }
    PCRMutex_Unlock(connection_server_mutex);

    return retStatus;
}

static void OALConnectionMgr_PeerListenerRemoved_enqueue_fn( OALThreadPoolItem dnode, boolean shutdown )
{
    NotifySystem_PeerListenerRemoved_t * node;

    PCR_Assert(dnode != NULL);
    node = (NotifySystem_PeerListenerRemoved_t *) dnode;

    PCR_Assert(node->listener);
    PCR_Assert(node->listener->fns->Removed);

    node->listener->fns->Removed(node->listener, node->connection, NULL);

    DOFMem_Free(node);
}

DOFError OALConnectionMgr_NotifySystem_PeerListenerRemoved(DOFConnection connection, connPeerLstnNode_t *listenNode)
{
    DOFError retStatus = DOFERROR_OK;

    PCR_Assert( connection );
    PCR_Assert(listenNode);
    PCR_Assert(listenNode->listener);

    PCRMutex_Lock(connection_server_mutex);
    if (listenNode->removeNode)
    {
        if (callback_enqueue)
        {
            OALEnqueueItem_Init((OALEnqueueItem)listenNode->removeNode, OALConnectionMgr_PeerListenerRemoved_enqueue_fn);
            listenNode->removeNode->listener = listenNode->listener;
            listenNode->removeNode->connection = connection;
            if (!OALEnqueue_Enqueue(callback_enqueue, (OALEnqueueItem)listenNode->removeNode))
            {
                DOFMem_Free(listenNode->removeNode);
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_INFO, "DOFConnection: Failed to enqueue peer listener removed message." );
                retStatus = DOFERROR_FAILED;
            }
        }
        else
        {
            /* The DOF is being shut down so just destroy the remove node and continue. */
            DOFMem_Free(listenNode->removeNode);
        }

        /* This function takes ownership of the remove node. */
        listenNode->removeNode = NULL;
    }
    PCRMutex_Unlock(connection_server_mutex);

    return retStatus;
}

/* Note: This function must NOT be called with the connection mutex held. */
DOFError OALConnectionMgr_NotifySystem_Destroyed(DOFConnection connection)
{
    DOFError        retStatus = DOFERROR_OK;

    PCR_Assert( connection );

    PCRMutex_Lock(connection_server_mutex);
    if (callback_enqueue)
    {
        CompleteConnectionDestroy_t * node;

        node = (CompleteConnectionDestroy_t *) DOFMem_Alloc(sizeof(CompleteConnectionDestroy_t));
        if (node)
        {
            OALEnqueueItem_Init((OALEnqueueItem) node, OALConnectionMgr_CompleteConnectionDestroy);
            node->connection = connection;
        }

        if (!node || !OALEnqueue_Enqueue( callback_enqueue, (OALEnqueueItem) node))
        {
            DOFMem_Free(node);
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "Connection manager, notify system destroyed: Failed to enqueue." );
            /* Since the final destroy was not enqueued, add it to the delayed destroy list. */
            OALSList_Insert((OALSListNode *)&delayedDestroyList, (OALSListNode)connection);
            retStatus = DOFERROR_INSUFFICIENT_RESOURCES;
        }
        else
        {
            /* If the enqueue is successful and something is in the delayed destroy list, try to empty the list. */
            if (delayedDestroyList)
            {
                /*  Note: Since we only want to remove the item from the list if it is successfully enqueued, this
                    is a custom traversal of slist. */
                while (delayedDestroyList)
                {
                    node = (CompleteConnectionDestroy_t *) DOFMem_Alloc(sizeof(CompleteConnectionDestroy_t));
                    if(node != NULL)
                    {
                        DOFConnection next = (DOFConnection)delayedDestroyList->next;

                        OALEnqueueItem_Init((OALEnqueueItem) node, OALConnectionMgr_CompleteConnectionDestroy);
                        node->connection = delayedDestroyList;
                        if (OALEnqueue_Enqueue(callback_enqueue, (OALEnqueueItem) node))
                        {
                            /*  The connection was enqueued successfully, so remove it from the list.  It was not retained when
                                it was inserted into the list, so there is nothing to destroy, however, the connection can not be
                                accessed after it was successfully enqueued. so use the next that was saved off previously. */
                            delayedDestroyList = next;
                        }
                        else
                        {
                            DOFMem_Free(node);
                            /* Since the enqueue failed, leave the connection in the list and try to remove it later. */
                            break;
                        }
                    }
                }
            }
        }
        PCRMutex_Unlock(connection_server_mutex);
    }
    else
    {
        if (delayDestroy)
        {
            OALSList_Insert((OALSListNode *)&delayedDestroyList, (OALSListNode)connection);
            PCRMutex_Unlock(connection_server_mutex);
        }
        else
        {
            PCRMutex_Unlock(connection_server_mutex);
            /* since the callback enqueue is destroyed and delayDestroy is not set, call the final destroy immediately. */
            OALConnectionMgr_CompleteConnectionDestroy_Immediate(connection);
        }
    }

    return retStatus;
}

typedef struct NotifySystem_ConnectionEvent_t
{
    OALEnqueueItem_t super;

    DOFConnection               connection;
    DOFConnectionState          state;
    DOFConnectionStateListener  listener;
} NotifySystem_ConnectionEvent_t;

static void OALConnectionMgr_ConnectionEvent_fn( OALThreadPoolItem dnode, boolean shutdown )
{
    NotifySystem_ConnectionEvent_t * node;

    PCR_Assert( dnode != NULL );
    node = (NotifySystem_ConnectionEvent_t *) dnode;
    PCR_Assert(node->listener);
    PCR_Assert(node->listener->fns);
    PCR_Assert(node->listener->fns->StateChanged);

    /*  Call the DOFConnection callback now. */
    node->listener->fns->StateChanged(node->listener, node->connection, node->state);
    DOFConnectionState_Destroy(node->state);
    DOFMem_Free(node);
}

/* Note: the connection mutex must be locked when this function is called. */
static void OALConnectionMgr_EnqueueListenerCallback(DOFConnection conn, DOFConnectionStateListener listener, DOFConnectionState connState)
{
    PCR_Assert(conn);
    PCR_Assert(listener);

    if (callback_enqueue && connState)
    {
        NotifySystem_ConnectionEvent_t * node;

        node = (NotifySystem_ConnectionEvent_t *)DOFMem_Alloc(sizeof(NotifySystem_ConnectionEvent_t));
        if(node != NULL)
        {
            node->state = DOFConnectionState_Copy(connState);
            if (node->state)
            {
                node->connection = conn;
                node->listener = listener;
                OALEnqueueItem_Init((OALEnqueueItem) node, OALConnectionMgr_ConnectionEvent_fn);
                if (!OALEnqueue_Enqueue(callback_enqueue, (OALEnqueueItem) node))
                {
                    DOFConnection_RemoveStateListener(node->connection, node->listener);
                    DOFConnectionState_Destroy(node->state);
                    DOFMem_Free(node);
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFConnection: Failed to enqueue." );
                }
            }
            else
            {
                /* Connection state is required for the StateChanged callback.  So whether it was NULL or failed to copy
                   the listener must be removed. */
                DOFConnection_RemoveStateListener(conn, listener);
                DOFMem_Free(node);
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFConnection: Failed to copy connection state." );
            }
        }
        else
        {
            DOFConnection_RemoveStateListener(conn, listener);
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFConnection: Failed to allocate memory." );
        }
    }
    else
    {
        DOFConnection_RemoveStateListener(conn, listener);
        if (!connState)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFConnection: Connection state was NULL." );
        }
    }
}

/* Note: the connection mutex must be locked when this function is called. */
static boolean OALConnectionMgr_notifySystem_fn(OALSListNode item, void *arg)
{
    cbkInfo_t *cbkArgs = (cbkInfo_t *)arg;

    PCR_Assert(cbkArgs);
    PCR_Assert(item);

    OALConnectionMgr_EnqueueListenerCallback(cbkArgs->connection, ((connLstnNode)item)->listener, cbkArgs->state);

    /* Return FALSE so that the traverse will continue. */
    return FALSE;
}

/*  Note: this function must be called with the connection mutex already locked.
 If newListener is NULL the callback will be called for all callbacks in the list. */
void OALConnectionMgr_NotifySystem_ConnectionEvent(DOFConnection connection, DOFConnectionState state,
                                                   DOFConnectionStateListener newListener)
{
    PCR_Assert( connection );

    if (newListener)
    {
        OALConnectionMgr_EnqueueListenerCallback(connection, newListener, state);
    }
    else
    {
        cbkInfo_t cbkArgs;

        cbkArgs.connection = connection;
        cbkArgs.state = state;

        OALSList_Traverse(connection->listenerList, OALConnectionMgr_notifySystem_fn, &cbkArgs);
    }
}

typedef struct NotifySystem_PeerEvent_t
{
    OALEnqueueItem_t            super;
    DOFConnection               connection;
    DOFAddress                  address;
    DOFSecurityScope            scope;
    boolean                     added;
    DOFConnectionPeerListener   listener;
} NotifySystem_PeerEvent_t;

typedef struct notifyPeerEventCtx_t
{
    DOFConnection       conn;
    DOFAddress          address;
    DOFSecurityScope    scope;
    boolean             added;
}notifyPeerEventCtx_t;

static void OALConnectionMgr_PeerConnectionEvent_fn( OALThreadPoolItem dnode, boolean shutdown )
{
    NotifySystem_PeerEvent_t * node;

    PCR_Assert( dnode != NULL );
    node = (NotifySystem_PeerEvent_t *)dnode;
    PCR_Assert(node->listener);
    PCR_Assert(node->connection);
    PCR_Assert(node->address);

    if (node->added)
    {
        PCR_Assert(node->scope);
        if (node->listener->fns->PeerAdded)
            node->listener->fns->PeerAdded(node->listener, node->connection, node->address, node->scope);
    }
    else
    {
        if (node->listener->fns->PeerRemoved)
            node->listener->fns->PeerRemoved(node->listener, node->connection, node->address);
    }

    DOFAddress_Destroy(node->address);
    DOFSecurityScope_Destroy(node->scope);
    DOFMem_Free(node);
}

static boolean OALConnectionMgr_peerListener_tfn(OALSListNode item, void *arg)
{
    notifyPeerEventCtx_t *ctx = (notifyPeerEventCtx_t *)arg;
    DOFConnectionPeerListener listener = ((connPeerLstnNode)item)->listener;

    if (callback_enqueue)
    {
        NotifySystem_PeerEvent_t * node;

        node = (NotifySystem_PeerEvent_t *)DOFMem_Alloc(sizeof(NotifySystem_PeerEvent_t));
        if(node != NULL)
        {
            node->address = DOFAddress_Copy(ctx->address);
            if (node->address)
            {
                node->scope = DOFSecurityScope_Copy(ctx->scope);
                if (!ctx->added || node->scope)
                {
                    node->connection = ctx->conn;
                    node->listener = listener;
                    node->added = ctx->added;
                    OALEnqueueItem_Init((OALEnqueueItem)node, OALConnectionMgr_PeerConnectionEvent_fn);
                    if (!OALEnqueue_Enqueue(callback_enqueue, (OALEnqueueItem)node))
                    {
                        DOFConnection_RemovePeerListener(ctx->conn, listener);
                        DOFAddress_Destroy(node->address);
                        DOFSecurityScope_Destroy(node->scope);
                        DOFMem_Free(node);
                        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFConnection peerListener_tfn: Failed to enqueue." );
                    }
                }
                else
                {
                    DOFConnection_RemovePeerListener(ctx->conn, listener);
                    DOFAddress_Destroy(node->address);
                    DOFMem_Free(node);
                }
            }
            else
            {
                /* Connection state is required for the StateChanged callback.  So whether it was NULL or failed to copy
                   the listener must be removed. */
                DOFConnection_RemovePeerListener(ctx->conn, listener);
                DOFMem_Free(node);
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFConnection peerListener_tfn: Failed to copy DOFAddress." );
            }
        }
        else
        {
            DOFConnection_RemovePeerListener(ctx->conn, listener);
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFConnection peerListener_tfn: Failed to allocate memory." );
        }
    }
    else
    {
        DOFConnection_RemovePeerListener(ctx->conn, listener);
    }

    /* Return FALSE to iterate through all listeners. */
    return FALSE;
}

void OALConnectionMgr_NotifySystem_PeerListenerEvent(DOFConnection connection, DOFAddress address, DOFSecurityScope scope, boolean added)
{
    notifyPeerEventCtx_t ctx;

    ctx.conn = connection;
    ctx.address = address;
    ctx.scope = scope;
    ctx.added = added;

    OALSList_Traverse(connection->peerListenerList, OALConnectionMgr_peerListener_tfn, &ctx);
}

static void OALConnectionMgr_StreamRequestEvent_fn(OALThreadPoolItem dnode, boolean shutdown)
{
    DOFConnectionStreamRequest streamRequest = (DOFConnectionStreamRequest)dnode;
    DOFConnectionConfig config = NULL;
    boolean complete = FALSE;
    boolean doNothing = FALSE;

    PCR_Assert(streamRequest);

    /* lock mutex, check conditions, copy into local variables, clear enqueued flag, unlock mutex, call callback. */

    PCRMutex_Lock(connection_server_mutex);

    if (streamRequest->flags.doNothing)
    {
        doNothing = TRUE;
    }
    else
    {
        /* Config is used as a convience to get the stream request listener and as a flag to call the callback. */
        config = streamRequest->connection->OALConn->connConfig;
        complete = streamRequest->flags.complete;

    }

    streamRequest->flags.enqueued = FALSE;     /* Clear the enqueued flag so the StreamRequest can be reused as an enqueue item. */

    PCRMutex_Unlock(connection_server_mutex);

    if (doNothing)
    {
        DOFConnectionStreamRequest_Destroy(streamRequest);
    }

    if (config)
    {
        if (complete)
        {
            config->streamListener->streamListener->fns->StreamRequestComplete(config->streamListener->streamListener, config, streamRequest);
            DOFConnectionStreamRequest_Destroy(streamRequest);
        }
        else
        {
            config->streamListener->streamListener->fns->StreamRequest(config->streamListener->streamListener, config, streamRequest);
        }
    }
}

/* Note: the connection mutex must be locked when this function is called. */
boolean OALConnectionMgr_NotifySystem_StreamListenerEvent(DOFConnectionStreamRequest request, boolean complete)
{
    PCR_Assert(request);

    if (callback_enqueue && !request->connection->OALConn->connConfig->streamListener->removed)
    {
        if (request->flags.enqueued)
        {
            /* The request is already enqueued and the callback has not been called.
               The only time this can happen is if a request was enqueued and now this is the complete. The stream request is removed from the
               OALConnection's stream request tree for a complete callback so a request following a complete will cause a new stream request to be created. */
            request->flags.doNothing = TRUE;
        }
        else
        {
            if (complete)
            {
                if (request->connection->OALConn->connConfig->streamListener->streamListener->fns->StreamRequestComplete)
                {
                    OALEnqueueItem_Init((OALEnqueueItem)request, OALConnectionMgr_StreamRequestEvent_fn);
                    request->flags.enqueued = TRUE;
                    request->flags.complete = complete;
                    request->flags.doNothing = FALSE;
                    if (!OALEnqueue_Enqueue(callback_enqueue, (OALEnqueueItem)request))
                    {
                        request->flags.enqueued = FALSE;
                        /* This should never happen since enqueue should only fail if it is being shut down.  Before the enqueue was destroyed the
                           stream request tree for each OALConnection was emptied. */
                        return FALSE;
                    }
                }
                else
                {
                    DOFConnectionStreamRequest_Destroy(request);
                }
            }
            else
            {
                if (request->connection->OALConn->connConfig->streamListener->streamListener->fns->StreamRequest)
                {
                    OALEnqueueItem_Init((OALEnqueueItem)request, OALConnectionMgr_StreamRequestEvent_fn);
                    request->flags.enqueued = TRUE;
                    request->flags.complete = complete;
                    request->flags.doNothing = FALSE;
                    if (!OALEnqueue_Enqueue(callback_enqueue, (OALEnqueueItem)request))
                    {
                        request->flags.enqueued = FALSE;
                        /* This should never happen since enqueue should only fail if it is being shut down.  Before the enqueue was destroyed the
                           stream request tree for each OALConnection was emptied. */
                        return FALSE;
                    }
                }
            }
        }
    }

    return TRUE;
}

static void OALConnectionMgr_StreamRequestListenerRemoved(OALThreadPoolItem dnode, boolean shutdown)
{
    DOFConnectionSRLremoveNode removeNode = (DOFConnectionSRLremoveNode)dnode;

    PCR_Assert(removeNode);

    removeNode->streamListener->fns->Removed(removeNode->streamListener, removeNode->except);
    DOFException_Destroy(removeNode->except);
    DOFMem_Free(removeNode);
}

boolean OALConnectionMgr_NotifySystem_StreamListenerRemoved(DOFConnectionSRL srl, DOFException except)
{
    boolean retStatus = FALSE;

    PCR_Assert(srl);

    PCRMutex_Lock(connection_server_mutex);

    srl->removed = TRUE;

    if (srl->removeNode)
    {
        /* Initialize the remove node. */
        srl->removeNode->streamListener = srl->streamListener;
        srl->removeNode->except = DOFException_Copy(except);

        if (callback_enqueue)
        {
            OALEnqueueItem_Init((OALEnqueueItem)srl->removeNode, OALConnectionMgr_StreamRequestListenerRemoved);
            if (OALEnqueue_Enqueue(callback_enqueue, (OALEnqueueItem)srl->removeNode))
            {
                srl->removeNode = NULL;
                retStatus = TRUE;
            }
        }
    }
    else
    {
        retStatus = TRUE;
    }

    PCRMutex_Unlock(connection_server_mutex);

    return retStatus;
}

struct all_connections_wrapper {
    connection_manager_traverse_fn_t fn;
    void *context;
};

static OALAVLTreeList_Traverse_All_Return_Type_fn_t connections_traverse(OALAVLTreeListItem_t item, void *context, boolean newKey)
{
    OALConnection connection = (OALConnection)item;
    struct all_connections_wrapper *wrapper = (struct all_connections_wrapper *)context;

    if(wrapper->fn(connection, wrapper->context)) return OALAVLTREELIST_TERMINATE_TRAVERSE;

    return OALAVLTREELIST_NORMAL;
}

/* Note: The connectionMutex must be locked when this function is called. */
void OALConnectionMgr_TraverseAllConnections(connection_manager_traverse_fn_t fn, void *context)
{
    struct all_connections_wrapper wrapper;

    wrapper.fn = fn;
    wrapper.context = context;

    if (OALConnection_tree)
        OALAVLTreeList_TraverseAll(OALConnection_tree, connections_traverse, &wrapper);
}


