/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/exception.h>
#include <dof/error.h>
#include <dof/server.h>

#include <internal/addressinternal.h>
#include <internal/dofinternal.h>
#include <internal/exception.h>
#include <internal/avl_multitree.h>
#include <internal/avl_tree.h>
#include <internal/bridgeinternal.h>
#include <internal/connection_internal.h>
#include <internal/connectionmanager.h>
#include <internal/domaininternal.h>
#include <internal/enqueue.h>
#include <internal/mutex_pool.h>
#include <internal/pointerlist.h>
#include <internal/remotechannel.h>
#include <internal/runtimeinternal.h>
#include <internal/securegroupinternal.h>
#include <internal/serverinternal.h>
#include <internal/slist.h>
#include <internal/trp_v1.h>

#include "connection_local.h"

typedef struct connectionListNode_t
{
    OALSListNode    next;
    DOFConnection   conn;
}connectionListNode_t;

typedef struct notifyContext_t
{
    DOFConnectionState  connectionState;
    DOFException        exception;
    boolean             status;
}notifyContext_t;

static uint32 connectedCount = 0;

extern PCRMutex connection_server_mutex;
extern PCRMutex connection_server_count_mutex;

/* This function will insert the DOFConnection into the OALConnections list.
   This function must be called with the connection mutex already locked.
   FALSE will be returned if the system is out of memory or NULL is passed in for one of the paramaters. */
static boolean OALConnection_AddDOFConnection(OALConnection OALConn, DOFConnection conn)
{
    if (OALConn && conn)
    {
        connectionListNode_t * node;

        node = DOFMem_Alloc(sizeof(connectionListNode_t));
        if (node)
        {
            node->conn = conn;
            OALSList_Insert(&OALConn->dofConnectionList, (OALSListNode)node);
            return TRUE;
        }
    }
    return FALSE;
}

/* This function is formatted as a OALSlistTraversal_fn_t used to locate a DOFConnection in a slist.
   TRUE will be returned if the DOFConnection conn is stored in item. */
static boolean OALConnection_findDOFConnectionInList_tfn(OALSListNode item, void *arg)
{
    DOFConnection conn = (DOFConnection)arg;

    if (((connectionListNode_t *)item)->conn == conn)
        return TRUE;

    return FALSE;
}

void OALConnection_ClearChannel(OALConnection OALConn)
{
    if (OALConn && OALConn->channel)
    {
        if (OALConn->flags.isOutbound)
        {
            OALChannelMembrane_SetBridge(OALConn->connSecurityMembrane, DOF_GetDefaultOperationBridge(DOF_GetDof()));
        }
        OALChannel_Release(OALConn->channel);
        OALConn->channel = NULL;
        DOFCredentials_Destroy(OALConn->peerCredentials);
        OALConn->peerCredentials = NULL;
        OALConn->flags.isConnected = FALSE;
        PCR_Assert(connectedCount);
        connectedCount--;
    }
}

/* This function will remove the DOFConnection from the OALConnection's list of connections. */
void OALConnection_RemoveDOFConnection(OALConnection OALConn, DOFConnection conn)
{
    if (OALConn && conn)
    {
        connectionListNode_t * node;

        PCRMutex_Lock(connection_server_mutex);

        node = (connectionListNode_t *)OALSList_FindAndRemove(&OALConn->dofConnectionList, NULL,
                                                              OALConnection_findDOFConnectionInList_tfn, conn);

        PCRMutex_Unlock(connection_server_mutex);

        if (node)
            DOFMem_Free(node);
    }
}

/* Note: This function must be called with the connection mutex locked. */
static void OALConnection_shutdownDOFConnection_fn(OALSListNode item, void *notUsed)
{
    connectionListNode_t * node = (connectionListNode_t *)item;
    OALSListNode pendingConnectionsList = node->conn->pendingConnections;

    /* Set the list in the connection to NULL and use the temporary so that a second list destroy
       will not be called if the connection is destroyed by traversing the list. */
    node->conn->pendingConnections = NULL;
    node->conn->shutdownByDOF = TRUE;

    /* Clear the listeners and related listener functions. */
    DOFConnection_ClearAllListeners(node->conn);

    /* the connectionListNode does not retian the connection, so after destroying the pending connection list
       do not try to use the connection. */
    OALSList_RemoveAll(&pendingConnectionsList, DOFConnection_RemovePendingConnections_fn, NULL);

    DOFMem_Free(node);
}

/*  This function will traverse the list of DOFConnections for the OALConnection and remove the listeners
    that may have been added to the DOFConnections.
    Note: This function must be called with the connection mutex locked. */
void OALConnection_ShutdownDOFConnections(OALConnection OALConn)
{
    OALSListNode dofConnList = OALConn->dofConnectionList;

    OALConn->dofConnectionList = NULL;
    OALSList_RemoveAll(&dofConnList, OALConnection_shutdownDOFConnection_fn, NULL);
    OALConn->flags.shutdownByDOF = TRUE;
}

static boolean trpRoutesTree_CompareGreaterThan_fn_t(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    return (key > ((void*)((OALConnection_TRPRoute_t*)item)->type.domain));
}

static boolean trpRoutesTree_CompareEqual_fn_t(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    return (key == ((void*)((OALConnection_TRPRoute_t*)item)->type.domain));
}

static boolean OALConnection_StreamRequestCompare_gt(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    return DOFAddress_CompareAddressInfo((DOFAddress)key, ((DOFConnectionStreamRequest)item)->address) > 0;
}

static boolean OALConnection_StreamRequestCompare_eq(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    return DOFAddress_IsAddressInfoEqual((DOFAddress)key, ((DOFConnectionStreamRequest)item)->address);
}

/* This function will return an OALConnection that was created with the matching DOFConnectionConfig.
   If an OALConnection is returned it must be released using OALConnection_Release. */
OALConnection OALConnection_CreateConnection(DOFConnection conn, DOFConnectionConfig connConfig, DOFSecurityMembrane securityMembrane, DOF dof, boolean shareOALConn)
{
    /* Lookup connection and return existing OALConnection if one is found.  If an existing OALConnection is not found
       create one and return it. If an existing OALConnection is found, it may be connected.  If a new OALConnection
       is created this function will not cause it to connect. */
    OALConnection               ret = NULL;
    OALConnection_findContext_t findCtx;

    PCRMutex_Lock(connection_server_mutex);

    findCtx.connConfig = connConfig;
    findCtx.OALConn = NULL;
    findCtx.baseOAL = NULL;
    findCtx.connector = securityMembrane;
    findCtx.baseIsOALConnection = FALSE;

    /* Set what should be the base OALConnection or OALServer. */

    if (conn->relatedConnection)
    {
        if (DOFConnectionConfig_GetConnectionType(connConfig) == DOFCONNECTIONTYPE_POINT)
        {
            if (conn->relatedConnection->OALConn && (DOFAddress_GetAddressType(DOFConnectionConfig_GetAddress(conn->relatedConnection->OALConn->connConfig)) == DOFADDRESSTYPE_UNICAST))
            {
                findCtx.baseOAL = conn->relatedConnection->OALConn;
                findCtx.baseIsOALConnection = TRUE;
            }
        }
        else if ((DOFConnectionConfig_GetConnectionType(connConfig) != DOFCONNECTIONTYPE_STREAM) &&
            (DOFConnectionConfig_GetConnectionType(conn->relatedConnection->OALConn->connConfig) != DOFCONNECTIONTYPE_STREAM))
        {
            /* If the DOFConnection has a related DOFConnection get the base OAL. */
            if (conn->relatedConnection->OALConn->flags.baseIsOALConnection && conn->relatedConnection->OALConn->base.OALConn)
            {
                if (DOFAddress_IsAddressInfoEqual(DOFConnectionConfig_GetAddress(connConfig), DOFConnectionConfig_GetAddress(conn->relatedConnection->OALConn->base.OALConn->connConfig)))
                {
                    /* Note: the baseOAL is not retained since the related connection is. */
                    findCtx.baseOAL = conn->relatedConnection->OALConn->base.OALConn;
                    findCtx.baseIsOALConnection = TRUE;
                }
            }
            else if (conn->relatedConnection->OALConn->base.OALSrv)
            {
                /* If the related connection has a base OAL server then the server is datagram. */
                findCtx.baseOAL = conn->relatedConnection->OALConn->base.OALSrv;
                findCtx.baseIsOALConnection = FALSE;
            }
            else
            {
                if (DOFAddress_IsAddressInfoEqual(DOFConnectionConfig_GetAddress(connConfig), DOFConnectionConfig_GetAddress(conn->relatedConnection->OALConn->connConfig)))
                {
                    /* Use the related connection's OALConnection as this connection's base. */
                    findCtx.baseOAL = conn->relatedConnection->OALConn;
                    findCtx.baseIsOALConnection = TRUE;
                }
            }
        }
    }
    else if (conn->relatedServer)
    {
        /* Not expected to be here for connType DOFCONNECTIONTYPE_POINT, since those have conn->relatedConnection. */
        if (DOFConnectionConfig_GetConnectionType( connConfig ) != DOFCONNECTIONTYPE_STREAM &&
            DOFServerConfig_GetServerType( conn->relatedServer->server->config ) != DOFSERVERTYPE_STREAM)
        {
            /*  If the DOFConnection has a related DOFServer the servers's OALServer will be the base.
                Note: Servers do not have related connections.
                Note: the baseOAL is not retained since the related server is. */
            findCtx.baseOAL = conn->relatedServer->server;
            findCtx.baseIsOALConnection = FALSE;
        }
    }

    if (shareOALConn)
    {
        OALConnectionMgr_FindOALConnection(&findCtx);
    }
    ret = findCtx.OALConn;
    if (ret)
    {
        if (OALConnection_AddDOFConnection(ret, conn))
        {
            if (!OALConnection_Retain(ret))
            {
                OALConnection_RemoveDOFConnection(ret, conn);
                ret = NULL;
            }
        }
        else
        {
            /* Since the DOFConnection could not be added to the OALConnection's list, just return NULL. */
            ret = NULL;
        }
    }
    else
    {
        /* If a matching OALConnection is not found, create a new OALConnection. */
        if (conn && connConfig)
        {
            ret = (OALConnection)DOFMem_Alloc(sizeof(OALConnection_t));
            if (ret)
            {
                PCRMem_Set(ret, 0, sizeof(OALConnection_t));
                ret->trpRoutesTree = OALAVLTree_Create( trpRoutesTree_CompareGreaterThan_fn_t, trpRoutesTree_CompareEqual_fn_t, NULL );
                if( ret->trpRoutesTree )
                {
                    ret->refCount = 1;
                    /* If a new OALConnection is created it will be inbound if shareOALConn is FALSE. */
                    if (shareOALConn)
                        ret->flags.isOutbound = TRUE;
                    ret->connConfig = DOFConnectionConfig_Copy(connConfig);
                    ret->connSecurityMembrane = DOFSecurityMembrane_Copy(securityMembrane);
                    ret->dof = dof;
                    if (ret->connConfig && (ret->connSecurityMembrane || !securityMembrane))
                    {
                        if (ret->connConfig->streamListener)
                        {
                            ret->streamRequestTree = OALAVLTree_Create(OALConnection_StreamRequestCompare_gt, OALConnection_StreamRequestCompare_eq, NULL);
                        }
                        if (ret->streamRequestTree || !ret->connConfig->streamListener)
                        {
                            /* If the DOFConnection has a related connection or server store the base OALConnection or OALServer */
                            if (findCtx.baseOAL)
                            {
                                if (findCtx.baseIsOALConnection)
                                {
                                    /* Note: the baseOAL is not retained since the related connection is. */
                                    ret->flags.baseIsOALConnection = TRUE;
                                    ret->base.OALConn = findCtx.baseOAL;
                                }
                                else if (!findCtx.baseIsOALConnection)
                                {
                                    /* Note: the baseOAL is not retained since the related server is. */
                                    ret->flags.baseIsOALConnection = FALSE;
                                    ret->base.OALSrv = findCtx.baseOAL;
                                }
                            }

                            if (!findCtx.baseOAL || (findCtx.baseOAL && ret->base.OALConn))
                            {
                                if (OALConnectionMgr_AddOALConnection(ret))
                                {
                                    /* Add the DOFConnection to the OALConnection's list. */
                                    if (!OALConnection_AddDOFConnection(ret, conn))
                                    {
                                        OALConnection_Release(ret);
                                        ret = NULL;
                                    }
                                }
                                else
                                {
                                    OALConnection_Release(ret);
                                    ret = NULL;
                                }
                            }
                            else
                            {
                                OALConnection_Release(ret);
                                ret = NULL;
                            }
                        }
                        else
                        {
                            OALConnection_Release(ret);
                            ret = NULL;
                        }
                    }
                    else
                    {
                        OALAVLTree_Destroy( ret->trpRoutesTree, NULL, NULL );
                        DOFSecurityMembrane_Destroy(ret->connSecurityMembrane);
                        DOFConnectionConfig_Destroy(ret->connConfig);
                        DOFMem_Free(ret);
                        ret = NULL;
                    }
                }
            }
        }
    }

    PCRMutex_Unlock(connection_server_mutex);

    return ret;
}

boolean OALConnection_Retain(OALConnection OALConn)
{
    boolean retStatus = FALSE;

    if (OALConn)
    {
        PCRMutex_Lock(ref_count_mutex);
        if (OALConn->refCount < 0xFFFFFFFF)
        {
            OALConn->refCount++;
            retStatus = TRUE;
        }
        PCRMutex_Unlock(ref_count_mutex);
    }

    return retStatus;
}

void OALConnection_Release(OALConnection OALConn)
{
    if (OALConn)
    {
        uint32 refCnt;

        /* The connection_server_mutex must be locked before the refCount is tested. */
        PCRMutex_Lock(connection_server_mutex);

        PCRMutex_Lock(ref_count_mutex);
        refCnt = --OALConn->refCount;
        PCRMutex_Unlock(ref_count_mutex);

        if (refCnt == 1 && OALConn->internalConnection)
        {
            DOFConnection tempConn;

            PCR_Assert(OALConn == OALConn->internalConnection->OALConn);
            tempConn = OALConn->internalConnection;
            OALConn->internalConnection = NULL;
            DOFConnection_ClearAllListeners(tempConn);
            DOFConnection_Destroy(tempConn);
            /* Note: do not try to use OALConn after calling DOFConnection_Destroy here, it was freed. */
        }
        else if (refCnt == 0)
        {
            boolean completeDestroy = TRUE;

            OALAVLTree_Destroy( OALConn->trpRoutesTree, NULL, NULL ); /* should be empty */

            /* Note: base.OALConn and base.OALSrv are not retained since the related connection or server is. */
            if (OALConn->channel && !OALConn->flags.doNotDestroy)
            {
                /* Set the doNotDestroy flag so that the OALConnection will not be destroyed if the OALConnection is
                   retained and released (both in this thread) by calling OALRemoteChannel_ConnectionInitiatedShutdown. */
                OALConn->flags.doNotDestroy = TRUE;

                if (OALConn->flags.isOutbound)
                    OALRemoteChannel_ConnectionInitiatedShutdown(OALConn->channel, OALConn);

                OALConnection_EmptyStreamRequestTree(OALConn);
                OALConnection_ClearChannel(OALConn);

                /* Clear the doNotDestroy flag now that OALRemoteChannel_ConnectionInitiatedShutdown is done. */
                OALConn->flags.doNotDestroy = FALSE;

                /* If OALRemoteChannel_ConnectionInitiatedShutdown enqueued a message that retained the
                   OALConnection, the refCount for the OALConn will now be greater than zero.  The
                   connection_server_mutex will prevent the OALRemoteChannel_ConnectionShutdown message from
                   being processed before this is done (since it will be processed in a seperate thread). */
                PCRMutex_Lock(ref_count_mutex);
                if (OALConn->refCount > 0)
                {
                    completeDestroy = FALSE;
                }
                PCRMutex_Unlock(ref_count_mutex);
            }

            if (completeDestroy && !OALConn->flags.doNotDestroy)
            {
                OALConnectionMgr_RemoveOALConnection(OALConn);
                DOFConnection_Destroy(OALConn->internalConnection);

                DOFConnectionConfig_Destroy(OALConn->connConfig);
                DOFSecurityMembrane_Destroy(OALConn->connSecurityMembrane);
                if (OALConn->streamRequestTree)
                {
                    OALAVLTree_Destroy(OALConn->streamRequestTree, NULL, NULL);
                }
                DOFMem_Free(OALConn);
            }
        }
        else
        {
            /* No action necessary. */
        }
        PCRMutex_Unlock(connection_server_mutex);
    }
}

/* Note: The connection mutex must be locked when calling this function. */
void OALConnection_Connect(OALConnection OALConn, DOFException * except)
{
    PCR_Assert(except);
    *except = NULL;

    PCR_Assert( OALConn );
    PCR_Assert( OALConn->connConfig );

    /* If the OALConnection already has a channel, but is not connected then it is connecting.  When it is connected
       all DOFConnections will be notified. */
    if (!OALConn->channel)
    {
        if (connectedCount < DOF_GetMaxConnections(OALConn->dof))
        {
            OALConn->flags.isConnected = FALSE;
            OALConn->channel = OALChannelMgr_CreateChannel_Connection(OALConn, except);
            if (OALConn->channel)
            {
                connectedCount++;

                if (OALConn->connConfig->type == DOFCONNECTIONTYPE_HUB)
                {
                    DOFSecureGroup_SetHubConfiguredPointTimeout( (DOFSecureGroup)OALConn->channel,
                                                      DOFConnectionConfig_GetMaxReceiveSilence( OALConn->connConfig ) );
                }
            }
        }
        else
        {
            *except = DOFException_Copy(DOFEXCEPTION_TOO_MANY);
        }
    }
}

boolean OALConnection_ConnectInbound(OALConnection OALConn, DOFConnection conn, OALChannel channel)
{
    boolean             status = FALSE;

    if (OALConn && conn && channel)
    {
        PCRMutex_Lock(connection_server_mutex);

        if (connectedCount < DOF_GetMaxConnections(OALConn->dof))
        {
            if (!OALConn->channel)
            {
                OALConn->channel = OALChannel_Retain(channel);
                if (OALConn->channel)
                {
                    if (!channel->remoteChannelData->flags.isOutbound &&
                        ((DOFConnectionConfig_GetConnectionType(channel->remoteChannelData->config.connection) == DOFCONNECTIONTYPE_DATAGRAM) ||
                         (DOFConnectionConfig_GetConnectionType(channel->remoteChannelData->config.connection) == DOFCONNECTIONTYPE_STREAM)))
                    {
                        /* Start the server so that the inbound connection will not go down if the app stops the server. */
                        if (DOFServer_IsStarted(conn->relatedServer))
                        {
                            OALConn->internalConnection = DOFConnection_Copy(conn);
                            if (DOFConnection_AddInternalServerListener(conn))
                            {
                                connectedCount++;
                                status = TRUE;
                            }
                            else
                            {
                                OALChannel_Release(OALConn->channel);
                                OALConn->channel = NULL;
                                DOFConnection_Destroy(conn);
                                OALConn->internalConnection = NULL;
                            }
                        }
                        else
                        {
                            OALChannel_Release(OALConn->channel);
                        }
                    }
                    else
                    {
                        OALConn->internalConnection = DOFConnection_Copy(conn);
                        connectedCount++;
                        status = TRUE;
                    }
                }
            }
        }
        PCRMutex_Unlock(connection_server_mutex);
    }
    return status;
}

/* Note: the connection mutex must be locked when calling this function. */
void OALConnection_DOFConnectionConnected(OALConnection OALConn)
{
    if (OALConn)
    {
        OALConn->connectedRefCount++;
    }
}

/* Note: this function is called with the connection mutex already locked. */
static boolean OALConnection_ConnectComplete_tfn(OALSListNode item,  void *arg)
{
    notifyContext_t *ctx = (notifyContext_t *)arg;

    DOFConnection_ConnectionEvent(((connectionListNode_t *)item)->conn, ctx->status, ctx->connectionState, ctx->exception);

    return FALSE;   /* This function must be called for each node in the list so return FALSE. */
}

/* Note: the connection mutex must be locked when calling this function. */
void OALConnection_DOFConnectionDisconnected(OALConnection OALConn, DOFException exception)
{
    if (OALConn && OALConn->connectedRefCount > 0)
    {
        if (!exception)
            OALConn->connectedRefCount--;

        if ((exception || OALConn->connectedRefCount == 0) && OALConn->channel)
        {
            uint64 ibc;
            uint64 obc;
            uint64 ipc;
            uint64 opc;

            notifyContext_t ctx;

            PCRMutex_Lock(connection_server_count_mutex);
            ibc = OALConn->byteInputCount;
            obc = OALConn->byteOutputCount;
            ipc = OALConn->packetInputCount;
            opc = OALConn->packetOutputCount;
            PCRMutex_Unlock(connection_server_count_mutex);

            ctx.exception = exception;
            ctx.status = FALSE;

            if (!OALConn->flags.shutdownByDOF)
            {
                OALRemoteChannel_ConnectionInitiatedShutdown(OALConn->channel, OALConn);
            }
            OALConnection_EmptyStreamRequestTree(OALConn);
            OALConnection_ClearChannel(OALConn);

            /* Note: this create can fail, however the check for the failure is done later where the connection state is used. */
            ctx.connectionState = DOFConnectionState_Create(OALConn->dof, OALConn->connConfig, OALConn->connSecurityMembrane,
                                                            OALConn->flags.isConnected, NULL, OALConn->flags.isOutbound,
                                                            OALConn->peerCredentials, OALConnection_GetRelatedDOFConnection(OALConn),
                                                            OALConnection_GetRelatedDOFServer(OALConn), ibc,
                                                            obc, ipc, opc,
                                                            exception);

            OALSList_Traverse(OALConn->dofConnectionList, OALConnection_ConnectComplete_tfn, &ctx);

            if (ctx.connectionState && !OALConn->flags.shutdownByDOF)
            {
                DOFRuntime_ConnectionStateChanged(OALConn->internalConnection, ctx.connectionState);
            }

            DOFConnectionState_Destroy(ctx.connectionState);
        }
    }
}

DOFConnection OALConnection_CreateDOFConnection(OALConnection OALConn)
{
    DOFConnection   conn;
    DOFConnection   relatedConnection = NULL;
    DOFServer       relatedServer = NULL;

    PCRMutex_Lock(connection_server_mutex);

    /* If there is a related OALConnection or OALServer call OALConnection_GetDOFConnection or OALServer_GetDOFServer.*/
    if (OALConn->flags.baseIsOALConnection && OALConn->base.OALConn)
    {
        relatedConnection = OALConnection_CreateDOFConnection(OALConn->base.OALConn);
    }
    else if (OALConn->base.OALSrv)
    {
        relatedServer = OALServer_CreateDOFServer(OALConn->base.OALSrv);
    }
    conn = DOFConnection_CreateInternalConnection(OALConn->dof, OALConn->connSecurityMembrane, OALConn->connConfig,
                                                                         relatedServer, relatedConnection, TRUE);

    DOFConnection_Destroy(relatedConnection);
    DOFServer_Release(relatedServer);
    PCRMutex_Unlock(connection_server_mutex);
    return conn;
}

/* This function will return the internal DOFConnection.  Other than DOFRuntime the application or library must not
   be allowd to access this connection to call connect or disconnect. */
DOFConnection OALConnection_GetDOFConnection(OALConnection OALConn)
{
    if (OALConn)
    {
        PCRMutex_Lock(connection_server_mutex);
        if (!OALConn->internalConnection)
        {
            DOFConnection   relatedConnection = NULL;
            DOFServer       relatedServer = NULL;

            /* If there is a related OALConnection or OALServer call OALConnection_GetDOFConnection or OALServer_GetDOFServer.*/
            if (OALConn->flags.baseIsOALConnection && OALConn->base.OALConn)
            {
                relatedConnection = OALConnection_GetDOFConnection(OALConn->base.OALConn);
            }
            else if (OALConn->base.OALSrv)
            {
                relatedServer = OALServer_GetDOFServer(OALConn->base.OALSrv);
            }
            OALConn->internalConnection = DOFConnection_CreateInternalConnection(OALConn->dof, OALConn->connSecurityMembrane, OALConn->connConfig,
                                                                                 relatedServer, relatedConnection, TRUE);
        }
        PCRMutex_Unlock(connection_server_mutex);
        return OALConn->internalConnection;
    }

    return NULL;
}

DOFConnection OALConnection_GetRelatedDOFConnection(OALConnection OALConn)
{
    if (OALConn)
    {
        DOFConnection conn = OALConnection_GetDOFConnection(OALConn);

        if (conn)
            return conn->relatedConnection;
    }

    return NULL;
}

DOFServer OALConnection_GetRelatedDOFServer( OALConnection OALConn )
{
    if (OALConn)
    {
        DOFConnection conn = OALConnection_GetDOFConnection( OALConn );

        if (conn)
            return conn->relatedServer;
    }

    return NULL;
}

DOFServer OALConnection_CreateRealtedDOFServer(OALConnection OALConn, DOFException * pException)
{
    DOFServer srv = NULL;

    if (pException)
        *pException = NULL;

    PCRMutex_Lock(connection_server_mutex);
    if (!OALConn->flags.baseIsOALConnection && OALConn->base.OALSrv)
    {
        srv = OALServer_CreateDOFServer(OALConn->base.OALSrv);

        if (!srv && pException)
        {
            /* Since this exception cannot be destroyed, copy is not called here. */
            *pException = DOFEXCEPTION_INSUFFICIENT_RESOURCES;
        }
    }
    PCRMutex_Unlock(connection_server_mutex);

    return srv;
}

/* Channel will call this function when a channel is connected or has failed to connect. */
void OALConnection_ConnectionEvent(OALConnection OALConn,  OALChannel channel, boolean connected, DOFException exception)
{
    notifyContext_t ctx;

    ctx.exception  = exception;
    ctx.status = connected;
    ctx.connectionState = NULL;

    PCRMutex_Lock(connection_server_mutex);

    /* If the OALConnection is not connected or OALConnection is connected or connecting and receives a
       late message that a previous channel is shutting down then do nothing. */
    if ((OALConn->channel) && (OALConn->channel == channel))
    {
        if (OALConn->flags.isConnected)
        {
            if (!connected)
            {
                uint64 ibc;
                uint64 obc;
                uint64 ipc;
                uint64 opc;

                PCRMutex_Lock(connection_server_count_mutex);
                ibc = OALConn->byteInputCount;
                obc = OALConn->byteOutputCount;
                ipc = OALConn->packetInputCount;
                opc = OALConn->packetOutputCount;
                PCRMutex_Unlock(connection_server_count_mutex);

                OALConnection_EmptyStreamRequestTree(OALConn);
                /* The channel just notified us that it shutdown, clear the reference to the channel. */
                OALConnection_ClearChannel(OALConn);
                /* Note: this create can fail, however the check for the failure is done later where the connection state is used. */
                ctx.connectionState = DOFConnectionState_Create(OALConn->dof, OALConn->connConfig, OALConn->connSecurityMembrane,
                                                                OALConn->flags.isConnected, NULL, OALConn->flags.isOutbound,
                                                                OALConn->peerCredentials, OALConnection_GetRelatedDOFConnection(OALConn),
                                                                OALConnection_GetRelatedDOFServer(OALConn), ibc, obc, ipc, opc,
                                                                exception);
            }
        }
        else
        {
            /* When a channel exists and the isConnected flag is not set the OALConnection is in process of connecting. */
            if (!connected)
            {
                /* The Connection failed to connect, Do not set connectionState so that the listeners are not notified. */
                OALConnection_ClearChannel(OALConn);
            }
            else
            {
                boolean trp_success = TRUE;
                if (OALConn->connConfig->domainCreds)
                {
                    /* Notify the TRP router that the OALConnection is now connected. */
                    if( !OALDomainCredentialsTree_AddTrustedDomainsRoutes( OALConn->connConfig->domainCreds, OALConn->channel ) )
                    {
                        trp_success = FALSE;
                    }

                }

                if( trp_success )
                {
                    uint64 ibc;
                    uint64 obc;
                    uint64 ipc;
                    uint64 opc;

                    PCRMutex_Lock(connection_server_count_mutex);
                    ibc = OALConn->byteInputCount;
                    obc = OALConn->byteOutputCount;
                    ipc = OALConn->packetInputCount;
                    opc = OALConn->packetOutputCount;
                    PCRMutex_Unlock(connection_server_count_mutex);

                    /* The OALConnection is now connected. */
                    OALConn->flags.isConnected = TRUE;
                    /* Note: this create can fail, however the check for the failure is done later where the connection state is used. */
                    ctx.connectionState = DOFConnectionState_Create(OALConn->dof, OALConn->connConfig, OALConn->connSecurityMembrane,
                                                                    OALConn->flags.isConnected, OALConn->channel, OALConn->flags.isOutbound,
                                                                    OALConn->peerCredentials, OALConnection_GetRelatedDOFConnection(OALConn),
                                                                    OALConnection_GetRelatedDOFServer(OALConn), ibc, obc, ipc, opc,
                                                                    NULL);
                }
                else
                {
                    OALRemoteChannel_ConnectionInitiatedShutdown( OALConn->channel, OALConn );
                    OALConnection_EmptyStreamRequestTree(OALConn);
                    OALConnection_ClearChannel(OALConn);
                    ctx.exception = DOFEXCEPTION_INSUFFICIENT_RESOURCES;
                    ctx.status = FALSE;
                }
            }
        }

        OALSList_Traverse(OALConn->dofConnectionList, OALConnection_ConnectComplete_tfn, &ctx);

        if (ctx.connectionState)
        {
            /* If there is connection state notify DOFRuntime of state change.  If connectionState is null it is
               just a pending connection that failed to connect so listeners must not be notified. */
            DOFRuntime_ConnectionStateChanged(OALConn->internalConnection, ctx.connectionState);
        }
        PCRMutex_Unlock(connection_server_mutex);
        DOFConnectionState_Destroy(ctx.connectionState);
    }
    else
    {
        PCRMutex_Unlock(connection_server_mutex);
    }
}

typedef struct notifyPeerEventContext_t
{
    DOFAddress address;
    DOFSecurityScope scope;
    boolean added;
}notifyPeerEventContext_t;

static boolean OALConnection_PeerListenerEvent_tfn(OALSListNode item,  void *arg)
{
    notifyPeerEventContext_t * ctx = (notifyPeerEventContext_t *)arg;

    OALConnectionMgr_NotifySystem_PeerListenerEvent(((connectionListNode_t *)item)->conn, ctx->address, ctx->scope, ctx->added);

    /* Return FALSE so that all connections are notified. */
    return FALSE;
}

void OALConnection_PeerListenerEvent(OALConnection OALConn, DOFAddress address, DOFSecurityScope scope, boolean added)
{
    notifyPeerEventContext_t ctx;

    ctx.address = address;
    ctx.scope = scope;
    ctx.added = added;

    PCRMutex_Lock(connection_server_mutex);
    OALSList_Traverse(OALConn->dofConnectionList, OALConnection_PeerListenerEvent_tfn, &ctx);
    PCRMutex_Unlock(connection_server_mutex);
}

void OALConnection_SetPeerCredentials(OALConnection OALConn, DOFCredentials creds)
{
    if( OALConn )
    {
        PCRMutex_Lock(connection_server_mutex);
        DOFCredentials_Destroy(OALConn->peerCredentials);
        OALConn->peerCredentials = DOFCredentials_Copy(creds);
        PCRMutex_Unlock(connection_server_mutex);
    }
}

void OALConnection_ConnectionStreamRequestEvent(OALConnection OALConn, DOFAddress addr, boolean complete)
{
    DOFConnectionStreamRequest streamRequest = NULL;
    boolean notifyListener = FALSE;

    PCR_Assert(OALConn);

    PCRMutex_Lock(connection_server_mutex);

    /* Look up the streamRequest by address.
       If a stream request was not found and complete is FALSE, create a stream request and call the callback. */

    if (addr && OALConn->streamRequestTree)
    {
        streamRequest = (DOFConnectionStreamRequest)OALAVLTree_Retrieve(OALConn->streamRequestTree, addr);

        if (complete)
        {
            if (streamRequest && (streamRequest->requestCount > 0))
            {
                /* If there is a stream request decrement the count and see if the complete callback should be called.
                   Connection manager will destroy the stream request after the callback is called. */
                --streamRequest->requestCount;
                if (streamRequest->requestCount == 0)
                {
                    OALAVLTree_Delete(OALConn->streamRequestTree, addr);
                    notifyListener = TRUE;
                }
            }
        }
        else
        {
            if (streamRequest)
            {
                if (streamRequest->requestCount < 0xFFFFFFFF)
                {
                    ++streamRequest->requestCount;
                    if (streamRequest->requestCount == 1)
                    {
                        notifyListener = TRUE;
                    }
                }
            }
            else
            {
                DOFConnection requestConn;
                DOFConnection   relatedConnection = NULL;
                DOFServer       relatedServer = NULL;

                /* If there is a related OALConnection or OALServer call OALConnection_GetDOFConnection or OALServer_GetDOFServer.*/
                if (OALConn->flags.baseIsOALConnection && OALConn->base.OALConn)
                {
                    relatedConnection = OALConnection_GetDOFConnection(OALConn->base.OALConn);
                }
                else if (OALConn->base.OALSrv)
                {
                    relatedServer = OALServer_GetDOFServer(OALConn->base.OALSrv);
                }

                requestConn = DOFConnection_CreateInternalConnection(OALConn->dof, OALConn->connSecurityMembrane, OALConn->connConfig,
                                                                     relatedServer, relatedConnection, TRUE);


                if (requestConn)
                {
                    streamRequest = DOFConnectionStreamRequest_Create(requestConn, addr);
                    if (streamRequest)
                    {
                        /* Insert the stream request into the tree, which will take ownership of it. */
                        if (OALAVLTree_Insert(OALConn->streamRequestTree, addr, streamRequest, NULL) == OALAVLTree_InsertSuccess)
                        {
                            ++streamRequest->requestCount;
                            notifyListener = TRUE;
                        }
                        else
                        {
                            DOFConnectionStreamRequest_Destroy(streamRequest);
                            streamRequest = NULL;
                        }

                    }
                    /* This was created here so destroy it now.  The stream request request retained the connection in the stream request create call. */
                    DOFConnection_Destroy(requestConn);
                }

                if (!streamRequest)
                {
                    /* The stream request could not be created.  Log a message and continue.
                       If this is called again for the same OALConnection and address it will try again to
                       create the stream request.  The wierd thing will be that the count may not agree with router
                       so the complete callback will get called early, but at least the stream request callback
                       and stream request complete callbacks will get called. */
                    DOFLog_Message(DOFLOGLEVEL_WARNING, "Unable to call the stream request listener stream request callback.");
                }
            }
        }

        if (notifyListener)
        {
            if (OALConnectionMgr_NotifySystem_StreamListenerEvent(streamRequest, complete) == FALSE)
            {
                /* The connection manager failed to trigger the callback. */
                if (complete)
                {
                    /* If the complete callback cannot be called this */
                }
            }
        }
    }

    PCRMutex_Unlock(connection_server_mutex);
}

void OALConnection_UpdateInputByteCount( OALConnection conn, uint32 count )
{
    PCRMutex_Lock(connection_server_count_mutex);
    conn->byteInputCount += count;
    conn->packetInputCount++;
    PCRMutex_Unlock(connection_server_count_mutex);
}

void OALConnection_UpdateOutputByteCount( OALConnection conn, uint32 count )
{
    PCRMutex_Lock(connection_server_count_mutex);
    conn->byteOutputCount += count;
    conn->packetOutputCount++;
    PCRMutex_Unlock(connection_server_count_mutex);
}

struct OALConnectionValidateContext_t
{
    PCREvent        event;
    DOFException*   exception;
    boolean         isValidate;
    boolean         isSuccess;
};

static void OALConnection_ValidateDomainCallback( DOFCredentials self, DOFException exception, void* context )
{
    struct OALConnectionValidateContext_t* ctx = (struct OALConnectionValidateContext_t*)context;

    if( !exception )
        ctx->isSuccess = TRUE;

    if( ctx->exception )
        *(ctx->exception) = DOFException_Copy( exception );

    PCREvent_Signal( ctx->event );
}

static boolean OALConnection_ValidateDomain( OALConnection self, DOFDomain domain, uint32 timeout, DOFException* exception )
{
    boolean ret = FALSE;
    struct OALConnectionValidateContext_t ctx;

    ctx.event = PCREvent_Create(TRUE);

    if( ctx.event )
    {
        OALChannel channel = self->channel;
        DOFAddress address = self->channel->remoteChannelData->address;

        ctx.exception = exception;
        ctx.isValidate = DOFCredentials_IsResolved( DOFDomain_GetCredentials( domain ) );
        ctx.isSuccess = FALSE;

        if( ctx.isValidate )
            DOFCredentials_Internal_BeginValidate( DOFDomain_GetCredentials( domain ), DOF_GetDof(), timeout, TRUE, channel, address, OALConnection_ValidateDomainCallback, &ctx );
        else
            DOFCredentials_Internal_BeginResolve( DOFDomain_GetCredentials( domain ), DOF_GetDof(), timeout, TRUE, channel, address, OALConnection_ValidateDomainCallback, &ctx );

        if( PCREvent_Wait( ctx.event, DOF_TIMEOUT_NEVER ) == PCRSTATUS_OK )
        {
            if( ctx.isSuccess )
            {
                ret = TRUE;
            }
        }
        PCREvent_Destroy(ctx.event);
    }
    return ret;
}

/* call with connection locked */
boolean OALConnection_AddDynamicTRPRoute( OALConnection self, DOFDomain domain, DOFObjectIDDomain domainID, uint32 timeout, DOFException* exception )
{
    boolean isDomain = (domain != NULL);
    OALConnection_TRPRoute_t* item;
    boolean ret = FALSE;

    if( isDomain )
        item = (OALConnection_TRPRoute_t*)OALAVLTree_Retrieve( self->trpRoutesTree, domain );
    else
        item = (OALConnection_TRPRoute_t*)OALAVLTree_Retrieve( self->trpRoutesTree, domainID );

    if( item )
    {
        ++item->count;
        ret = TRUE;
        goto EXIT;

    }

    item = (OALConnection_TRPRoute_t*)DOFMem_Alloc(sizeof(OALConnection_TRPRoute_t));
    if( item )
    {
        OALAVLTree_InsertionStatus status = OALAVLTree_InsertFailure;
        item->isDomain = isDomain;
        item->count = 1;
        item->timeout = timeout;

        if(isDomain)
        {
            item->type.domain = DOFDomain_Copy( domain );
            if( item->type.domain )
            {
                status = OALAVLTree_Insert( self->trpRoutesTree, domain, item, NULL );
            }
        }
        else
        {
            item->type.domainID = DOFObjectIDDomain_Copy( domainID );
            if(item->type.domainID)
            {
                status = OALAVLTree_Insert( self->trpRoutesTree, domainID, item, NULL );
            }
        }

        if( status != OALAVLTree_InsertSuccess )
        {
            if( status == OALAVLTree_AlreadyInserted )
                ret = TRUE;
            DOFMem_Free( item );
            goto EXIT;
        }


        if( isDomain )
        {
            PCRMutex_Unlock(connection_server_mutex);
            if( OALConnection_ValidateDomain( self, domain, timeout, exception ) )
            {

                TRPRouter_AddDynamicRoute( DOFDomain_GetCredentials( domain ), self->channel, self->channel->remoteChannelData->address, TRUE );
                PCRMutex_Lock(connection_server_mutex);
                return TRUE;
            }
            PCRMutex_Lock(connection_server_mutex);

           OALAVLTree_Delete( self->trpRoutesTree, domain );
           DOFMem_Free( item );
        }
        else
        {
            if( TRPRouter_AddTrustedDomain( domainID, self->channel ) )
                return TRUE;

            OALAVLTree_Delete( self->trpRoutesTree, domainID );
            DOFMem_Free( item );
        }
    }

EXIT:
    return ret;
}

/* call with connection locked */
void OALConnection_RemoveDynamicTRPRoute( OALConnection self, DOFDomain domain, DOFObjectIDDomain domainID )
{
    boolean isDomain = (domain != NULL);
    OALConnection_TRPRoute_t* item;

    if( isDomain )
        item = (OALConnection_TRPRoute_t*)OALAVLTree_Retrieve( self->trpRoutesTree, domain );
    else
        item = (OALConnection_TRPRoute_t*)OALAVLTree_Retrieve( self->trpRoutesTree, domainID );

    if( item )
    {
        --item->count;

        if(item->count == 0)
        {
            if( isDomain )
            {
                OALAVLTree_Delete( self->trpRoutesTree, domain );
                TRPRouter_RemoveCredentials( DOFDomain_GetCredentials( domain ), self->internalConnection );
                DOFDomain_Destroy( item->type.domain );
            }
            else
            {
                OALAVLTree_Delete( self->trpRoutesTree, domainID );
                TRPRouter_RemoveTrustedDomain( domainID, self->internalConnection );
                DOFObjectIDDomain_Destroy( item->type.domainID );
            }
            DOFMem_Free( item );
        }
    }
}
