/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/address.h>
#include <dof/oaltypes.h>
#include <dof/connection.h>

#include <internal/addressinternal.h>
#include <internal/connection_internal.h>
#include <internal/connectionmanager.h>
#include <internal/enp_v1.h>

#include "connection_local.h"

extern PCRMutex ref_count_mutex;


/* DOFConnectionSRL is an internal object used to wrap the DOFConnectionStreamRequestListener. */

DOFConnectionSRL DOFConnectionSRL_Create(DOFConnectionStreamRequestListener streamListener)
{
    DOFConnectionSRL ret = (DOFConnectionSRL)DOFMem_Alloc(sizeof(DOFConnectionSRL_t));

    if (ret)
    {
        if (streamListener->fns->Removed)
        {
            ret->removeNode = (DOFConnectionSRLremoveNode)DOFMem_Alloc(sizeof(DOFConnectionSRLremoveNode_t));
        }
        if (ret->removeNode || !streamListener->fns->Removed)
        {
            ret->removed = FALSE;
            ret->refCount = 1;
            ret->streamListener = streamListener;
            if( ret->removeNode )
                PCRMem_Set(ret->removeNode, 0, sizeof(DOFConnectionSRLremoveNode_t));
        }
        else
        {
            DOFMem_Free(ret);
            ret = NULL;
        }
    }
    return ret;
}

DOFConnectionSRL DOFConnectionSRL_Copy(DOFConnectionSRL srl)
{
    DOFConnectionSRL retSrl = NULL;

    if (srl)
    {
        PCRMutex_Lock(ref_count_mutex);

        if (srl->refCount < 0xFFFFFFFF)
        {
            ++srl->refCount;
            retSrl = srl;
        }

        PCRMutex_Unlock(ref_count_mutex);
    }

    return retSrl;
}

void DOFConnectionSRL_Destroy(DOFConnectionSRL srl)
{
    if (srl)
    {
        uint32 refCnt;

        PCRMutex_Lock(ref_count_mutex);

        --srl->refCount;
        refCnt = srl->refCount;

        PCRMutex_Unlock(ref_count_mutex);

        if (refCnt == 0)
        {
            if (OALConnectionMgr_NotifySystem_StreamListenerRemoved(srl, NULL) == FALSE)
            {
                if (srl->removeNode)
                {
                    /* The removed callback was not removed, probably because the connection manager is shut down.
                       Free the remove node and call the removed callback from here directly. */
                    DOFMem_Free(srl->removeNode);

                    if (srl->streamListener->fns->Removed)
                    {
                        srl->streamListener->fns->Removed(srl->streamListener, NULL);
                    }
                }
            }
            DOFMem_Free(srl);
        }
    }
}


DOFConnectionStreamRequestListener DOFConnectionSRL_GetStreamRequestListener(DOFConnectionSRL srl)
{
    return srl->streamListener;
}

DOFConnectionStreamRequest DOFConnectionStreamRequest_Create(DOFConnection conn, DOFAddress addr)
{
    DOFConnectionStreamRequest ret = (DOFConnectionStreamRequest)DOFMem_Alloc(sizeof(DOFConnectionStreamRequest_t));

    PCR_Assert(conn);
    PCR_Assert(addr);

    if (ret)
    {
        PCRMem_Set(ret, 0, sizeof(DOFConnectionStreamRequest_t));

        ret->connection = DOFConnection_Copy(conn);
        if (ret->connection)
        {
            ret->address = DOFAddress_Create_AddENPPort(addr, ENP_INVALID_PORT);
            if (ret->address)
            {
                return ret;
            }
            DOFConnection_Destroy(ret->connection);
        }
        DOFMem_Free(ret);
    }
    return NULL;

}

void DOFConnectionStreamRequest_Destroy(DOFConnectionStreamRequest request)
{
    if (request)
    {
        DOFConnection_Destroy(request->connection);
        DOFAddress_Destroy(request->address);
        DOFMem_Free(request);
    }
}

DOFConnection DOFConnectionStreamRequest_GetConnection( DOFConnectionStreamRequest self )
{
    return self->connection;
}


DOFAddress DOFConnectionStreamRequest_GetTargetAddress( DOFConnectionStreamRequest self )
{
    return self->address;
}
