/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/error.h>
#include <dof/credentials.h>
#include <dof/objectid.h>
#include <internal/srp.h>
#include <internal/stream.h>
#include <internal/cipher.h>
#include <internal/marshal.h>
#include <internal/srp.h>
#include <internal/trp_v1.h>
#include <internal/resource_pool.h>
#include <internal/objectidinternal.h>
#include <internal/credentialsinternal.h>
#include <internal/permissionsetinternal.h>
#include <internal/exception.h>

#include "credentials_common.h"
#include "password.h"
#include "key.h"
#include "remotecredentials.h"

/* @addtogroup DOFCredentials
@{
**/

extern PCRMutex ref_count_mutex;

static PCRMutex credentials_mtx = NULL;
static PCRMutex credentials_tree_mtx = NULL;

OALEnqueue credentials_enqueue = NULL;

/**
    @internal
    Credentials structure.
**/

void Credentials_Lock(void)
{
    PCRMutex_Lock(credentials_mtx);
}

void Credentials_Unlock(void)
{
    PCRMutex_Unlock(credentials_mtx);
}

void CredentialsTree_Lock(void)
{
    PCRMutex_Lock(credentials_tree_mtx);
}

void CredentialsTree_Unlock(void)
{
    PCRMutex_Unlock(credentials_tree_mtx);
}

void DOFCredentials_Init(void)
{
    credentials_mtx = PCRMutex_Create();
    credentials_tree_mtx = PCRMutex_Create();
}

void DOFCredentials_Shutdown(void)
{
    PCRMutex_Destroy(credentials_mtx);
    credentials_mtx = NULL;
    PCRMutex_Destroy(credentials_tree_mtx);
    credentials_tree_mtx = NULL;
}

boolean DOFCredentials_InitDOFLevel( DOF dof )
{
    credentials_enqueue = OALEnqueue_Create( dof->threadPool );
    if( credentials_enqueue )
        return TRUE;
    return FALSE;
}

void DOFCredentials_ShutdownDOFLevel(void)
{
    OALEnqueue_Destroy( credentials_enqueue );
    credentials_enqueue = NULL;
}

DOFCredentials DOFCredentials_Copy( const DOFCredentials self )
{
    DOFCredentials rval = self;

    if (self)
    {
        PCRMutex_Lock(ref_count_mutex);
        if( self->base.refCount < 0xFFFFFFFF )
            self->base.refCount++;
        else
            rval = NULL;
        PCRMutex_Unlock(ref_count_mutex);
    }
    return( rval );
}

DOFCredentials DOFCredentials_CreateNeutered( const DOFCredentials self )
{
    DOFCredentials rval;

    if( self )
    {
        rval = (DOFCredentials)DOFMem_Alloc( sizeof( DOFCredentials_t ) );
        if( rval )
        {
            PCRMem_Set( rval, 0, sizeof( DOFCredentials_t ) );

            rval->base.domain = DOFObjectIDDomain_Copy(self->base.domain);
            if( rval->base.domain )
            {
                rval->base.identity = DOFObjectIDAuthentication_Copy( self->base.identity );
                if( rval->base.identity )
                {
                    rval->base.fns = NULL;
                    rval->base.isDomRes = self->base.isDomRes;
                    rval->base.stage = self->base.stage;
                    rval->base.type = self->base.type;
                    rval->base.refCount = 1;
                    return rval;
                }
                DOFObjectIDDomain_Destroy( rval->base.domain );
            }
            DOFMem_Free( rval );
        }
    }

    return NULL;
}

void DOFCredentials_Destroy( DOFCredentials self )
{
    boolean destroy = FALSE;

    if ( self )
    {
        PCRMutex_Lock(ref_count_mutex);
        PCR_Assert( self->base.refCount > 0 );
        if ( (--self->base.refCount) == 0 )
            destroy = TRUE;
        PCRMutex_Unlock(ref_count_mutex);
        if ( destroy )
        {
            if( self->base.fns )
            {
                self->base.fns->OALCredentials_Dtor( (OALCredentials)self );
                OALObservable_Destroy( self->base.changeObserver );
            }

            DOFObjectIDDomain_Destroy( self->base.domain );
            DOFObjectIDAuthentication_Destroy( self->base.identity );
            DOFObjectIDSource_Destroy( self->base.sid );
            DOFMem_Free( self );
        }
    }
}

DOFCredentials DOFCredentials_CreateCopy( const DOFCredentials self )
{
    DOFCredentials ret = DOFCredentials_CreateCopyWithDomain( self, DOFCredentials_GetDomainID(self) );
    if( ret )
    {
        ret->base.stage = self->base.stage;
    }
    return ret;
}

DOFCredentials DOFCredentials_CreateCopyWithDomain( const DOFCredentials self, DOFObjectIDDomain domain )
{
    DOFCredentials ret = NULL;

    switch( self->base.type )
    {
    case DOFCREDENTIALSTYPE_KEY:
        {
            ret = DOFCredentialsKey_Internal_Create( domain, self->base.identity, DOFCredentials_GetKey(self));
            if( ret )
            {
                PCRMem_Copy( ((KEYCredentials_t*)ret)->validA, ((KEYCredentials_t*)self)->validA, 40 );
                ret->base.stage = 1;
                return ret;
            }
        }
        break;
    case DOFCREDENTIALSTYPE_PASSWORD:
        {
            SRPCredentials_t* org;
            SRPCredentials_t* cpy;

            ret = SRPCredentials_InternalCreate( domain );
            org = ((SRPCredentials_t*)self);
            cpy = ((SRPCredentials_t*)ret);

            if( ret )
            {
                ret->base.identity = DOFObjectIDAuthentication_Copy( DOFCredentials_GetIdentity( self ) );
                if( ret->base.identity )
                {
                    ret->base.isDomRes = DOFObjectID_IsBroadcast( DOFObjectIDDomain_GetObjectID( ret->base.domain ) );
                    ret->base.stage = 2;

                    PCRMem_Copy( cpy->partial_x, org->partial_x, 32 );
                    PCRSecureRandom_Generate( 32, cpy->data->a );
                    OALSRP_Calc_PublicA( cpy->data->a, (uint8*)cpy->data->pubA );
                    return ret;
                }
            }
        }
        break;
    case DOFCREDENTIALSTYPE_REMOTEKEY:
        {
            ret = DOFCredentialsKey_Internal_Create( domain, self->base.identity, DOFCredentials_GetKey(self));
            if( ret )
            {
                PCRMem_Copy( ((KEYCredentials_t*)ret)->validA, ((RemoteCredentials_t*)self)->validA, 40 );
                ret->base.stage = 1;
                return ret;
            }
        }
        break;
    default:
        PCR_Assert( FALSE );
        break;
    }

    return ret;
}

DOFCredentialsType DOFCredentials_GetType( const DOFCredentials self )
{
    PCR_Assert( self );
    switch( self->base.type )
    {
        case DOFCREDENTIALSTYPE_REMOTEKEY:
            return DOFCREDENTIALSTYPE_KEY;
        default:
            return self->base.type;
    }
}

DOFObjectIDDomain DOFCredentials_GetDomainID( const DOFCredentials self )
{
    PCR_Assert( self );
    return self->base.domain;
}

DOFObjectIDAuthentication DOFCredentials_GetIdentity( const DOFCredentials self )
{
    PCR_Assert( self );
    return self->base.identity;
}

const uint8 * DOFCredentials_GetStorageKey( const DOFCredentials self )
{
    PCR_Assert( self );
    return self->base.fns->OALCredentials_GetStorageKey( (OALCredentials)self );
}

uint32 DOFCredentials_GetStorageKeySize( const DOFCredentials self )
{
    PCR_Assert( self );
    return self->base.fns->OALCredentials_GetStorageKeySize( (OALCredentials)self );
}

boolean DOFCredentials_IsResolved( const DOFCredentials self )
{
    PCR_Assert( self );
    if( self->base.fns )
        return (self->base.stage == 0 );
    return FALSE;
}

const uint8 * DOFCredentials_GetPrivateStorage( const DOFCredentials self )
{
    PCR_Assert( self );
    if( self->base.fns )
        return self->base.fns->OALCredentials_GetPrivateStorage( (OALCredentials)self );
    return NULL;
}

uint32 DOFCredentials_GetPrivateStorageSize( const DOFCredentials self )
{
    PCR_Assert( self );
    if( self->base.fns )
        return self->base.fns->OALCredentials_GetPrivateStorageSize( (OALCredentials)self );
    return 0;
}

boolean DOFCredentials_SetPrivateStorage( const DOFCredentials self, uint32 len, const uint8 * data )
{
    PCR_Assert( self );
    if( self->base.fns )
        return self->base.fns->OALCredentials_SetPrivateStorage( (OALCredentials)self, len, data );
    return FALSE;
}

boolean DOFCredentials_IsCached( const DOFCredentials self )
{
    PCR_Assert( self );
    if( self->base.fns )
        return self->base.fns->OALCredentials_IsCached( (OALCredentials)self );
    return FALSE;
}

void DOFCredentials_Reset( const DOFCredentials self )
{
    PCR_Assert( self );
    if( DOFCredentials_IsCached( self ) )
    {
        self->base.fns->OALCredentials_ResetCache( (OALCredentials)self );
        if( self->base.changeObserver )
            OALObservable_SetChangedAndNotifyObservers( self->base.changeObserver );
    }
}

OALObserver DOFCredentials_AddChangeObserver( const DOFCredentials self, OALObserver_fn_t fn, void* arg )
{
    OALObserver ret = NULL;
    if(self)
    {
        if( !self->base.changeObserver )
            self->base.changeObserver = OALObservable_Create();

        if( self->base.changeObserver )
            ret = OALObservable_AddObserver( self->base.changeObserver, fn, arg );
    }
    return ret;
}

void DOFCredentials_RemoveChangeObserver( const DOFCredentials self, OALObserver observer )
{
    if( self && observer && self->base.changeObserver )
    {
        OALObservable_RemoveObserver( self->base.changeObserver, observer );
    }
}

typedef struct notifyObservers_t
{
    OALEnqueueItem_t super;
    DOFCredentials creds;
}notifyObservers_t;

static void DOFCredentialsNotifyObservers( OALThreadPoolItem item, boolean shutdown )
{
    notifyObservers_t* node = (notifyObservers_t*)item;

    if( node->creds->base.changeObserver )
        OALObservable_SetChangedAndNotifyObservers( node->creds->base.changeObserver );

    DOFCredentials_Destroy( node->creds );
    DOFMem_Free( node );
}

boolean DOFCredentials_NotifyObservers( DOFCredentials self )
{
    if( self->base.changeObserver )
    {
        notifyObservers_t* node = (notifyObservers_t*)DOFMem_Alloc( sizeof( notifyObservers_t ) );
        if( node )
        {
            OALEnqueueItem_Init( (OALEnqueueItem)node, DOFCredentialsNotifyObservers );
            node->creds = DOFCredentials_Copy( self );

            if( OALEnqueue_Enqueue( credentials_enqueue, (OALEnqueueItem)node ) )
                return TRUE;

            DOFCredentials_Destroy( node->creds );
            DOFMem_Free( node );
        }
    }
    else
    {
        return TRUE;
    }
    return FALSE;
}

boolean DOFCredentials_ContainsPrivateData( const DOFCredentials self )
{
    PCR_Assert( self );
    return ( self->base.fns != NULL );
}

/* internal */
boolean DOFCredentials_Update( DOFCredentials self, DOFCredentials from )
{
    PCR_Assert( self && from );
    if(  self != from )
        return self->base.fns->OALCredentials_Update( (OALCredentials)self, (OALCredentials)from );
    return TRUE; /* if this was called on the same credentials no change is needed */
}

boolean DOFCredentials_HasAnotherResolutionStage( DOFCredentials self )
{
    if( self )
        return self->base.fns->OALCredentials_HasAnotherResolutionStage( (OALCredentials)self );
    return FALSE;
}

const uint8* DOFCredentials_GetKey( const DOFCredentials self )
{
    PCR_Assert( self && self->base.fns );
    return self->base.fns->OALCredentials_GetKey((OALCredentials)self);
}

const uint8* DOFCredentials_GetKeyFromStorage(const DOFCredentials self, const uint8* privateStorage)
{
    PCR_Assert( self && self->base.fns);
    return self->base.fns->OALCredentials_GetKeyFromStorage( (OALCredentials)self, privateStorage );
}

boolean DOFCredentials_HasFixedDomain( const DOFCredentials self )
{
    return (!self->base.isDomRes);
}

boolean DOFCredentials_IsEqual(const DOFCredentials self, const DOFCredentials creds)
{
    if ( self == creds )
        return TRUE;

    if ( !self || !creds )
        return FALSE;

    if ( self->base.type != creds->base.type )
        return FALSE;

    if ( !( DOFObjectID_IsEqual(DOFObjectIDDomain_GetObjectID( self->base.domain ), DOFObjectIDDomain_GetObjectID( creds->base.domain ) ) ||
          /* both could be resolved to the same but are not at this point in time need more info to make the decision */
          ( ( self->base.isDomRes && creds->base.isDomRes ) &&
          ( DOFObjectID_IsBroadcast( DOFObjectIDDomain_GetObjectID( self->base.domain ) ) || DOFObjectID_IsBroadcast( DOFObjectIDDomain_GetObjectID( creds->base.domain ) ) ) ) ) )
        return FALSE;

    if( !self->base.fns )
        return TRUE;

    /* identity had to move into the isEqual function for the specific credentials type */
    return self->base.fns->OALCredentials_IsEqual((OALCredentials)self, (OALCredentials)creds);
}

uint32 DOFCredentials_MarshalSize(const DOFCredentials self, boolean isRequest )
{
    PCR_Assert( self && self->base.fns );
    return self->base.fns->OALCredentials_MarshalSize( (OALCredentials)self, isRequest );
}

boolean DOFCredentials_Internal_Marshal(const DOFCredentials self, OALStream stream, boolean isRequest )
{
    PCR_Assert( self && stream && self->base.fns );
    return self->base.fns->OALCredentials_Marshal((OALCredentials)self, stream, isRequest);
}

CredentialsUnmarshalReturn DOFCredentials_Unmarshal( DOFCredentials* self, DOFObjectIDDomain domain, OALStream stream, boolean isRequest )
{
    uint8 type;

    PCR_Assert( !(*self) || ((*self) && (*self)->base.fns) );

    if( OALStream_Read8( stream, &type ) )
    {
        switch( type )
        {
        case DOFCREDENTIALSTYPE_KEY:
            {
                if( !(*self) )
                {
                    (*self) = KEYCredentials_InternalCreate( domain );
                    if((*self) && isRequest)
                        (*self)->base.stage = 1;
                }
                else if( DOFObjectID_IsBroadcast( DOFObjectIDDomain_GetObjectID((*self)->base.domain) ) && (domain && !DOFObjectID_IsBroadcast( DOFObjectIDDomain_GetObjectID(domain) )))
                {
                    DOFObjectIDDomain temp = (*self)->base.domain;
                    (*self)->base.domain = DOFObjectIDDomain_Copy(domain);
                    if(!(*self)->base.domain)
                    {
                        (*self)->base.domain = temp;
                        return CREDENTIALS_FAILED_UNMARSHAL;
                    }
                    DOFObjectIDDomain_Destroy(temp);
                }

                if( (*self) )
                    return (*self)->base.fns->OALCredentials_Unmarshal( (OALCredentials)(*self), stream, isRequest );
            }
            break;
        case DOFCREDENTIALSTYPE_PASSWORD:
            {
                if( !(*self) )
                {
                    (*self) = SRPCredentials_InternalCreate( domain );
                }
                else if( DOFObjectID_IsBroadcast( DOFObjectIDDomain_GetObjectID((*self)->base.domain) ) && (domain && !DOFObjectID_IsBroadcast( DOFObjectIDDomain_GetObjectID(domain) )))
                {
                    DOFObjectIDDomain temp = (*self)->base.domain;
                    (*self)->base.domain = DOFObjectIDDomain_Copy(domain);
                    if(!(*self)->base.domain)
                    {
                        (*self)->base.domain = temp;
                        return CREDENTIALS_FAILED_UNMARSHAL;
                    }
                    DOFObjectIDDomain_Destroy(temp);
                }

                if( (*self) )
                    return (*self)->base.fns->OALCredentials_Unmarshal( (OALCredentials)(*self), stream, isRequest );
            }
            break;
        default:
            break;
        }
    }
    return CREDENTIALS_FAILED_UNMARSHAL;
}

DOFObjectIDSource DOFCredentials_GetSID( const DOFCredentials self )
{
    DOFObjectIDSource sid;
    Credentials_Lock();
    sid = self->base.sid;
    Credentials_Unlock();
    return sid;
}

void DOFCredentials_UpdateSID( DOFCredentials self, DOFObjectIDSource sid )
{
    Credentials_Lock();
    if ( !self->base.sid || ( sid && !DOFObjectID_IsASAssignedEqual( DOFObjectIDSource_GetObjectID( self->base.sid ), DOFObjectIDSource_GetObjectID( sid ) ) ) )
    {
        DOFObjectIDSource sidcopy = DOFObjectIDSource_Copy( sid );
        if ( sidcopy )
        {
            DOFObjectIDSource_Destroy( self->base.sid );
            self->base.sid = sidcopy;
        }
    }
    Credentials_Unlock();
}


uint32 DOFCredentials_GetByteSize(DOFCredentials self)
{
    if( self && self->base.fns)
    {
        /* this works for current types but may need to be changed when new credentials types are added */
        return (1 + 1 + DOFObjectID_GetBytesSize(DOFObjectIDDomain_GetObjectID(self->base.domain)) + DOFObjectID_GetBytesSize(DOFObjectIDAuthentication_GetObjectID(self->base.identity)) + 32);
    }
    return 0;
}

uint8* DOFCredentials_CreateBytes(DOFCredentials self)
{
    if( self && self->base.fns )
    {
        uint32 len = DOFCredentials_GetByteSize(self);
        if( len > 0 )
        {
            uint8* buffer = (uint8*)PCRMem_Alloc(len);
            if( buffer )
            {
                if( self->base.fns->OALCredentials_MarshalStore((OALCredentials)self, len, buffer) != 0 )
                {
                    return buffer;
                }
                PCRMem_Free( buffer );
            }
        }
    }
    return NULL;
}

DOFCredentials DOFCredentials_Create_Bytes(uint32 size, uint8 * buffer)
{
    uint16 type;
    DOFCredentials ret = NULL;

    if( buffer )
    {
        OALStream stream = OALStream_CreateRWWrapped( buffer, size);
        if( stream )
        {
            /* Unmarshal type*/
            if( OALMarshal_StreamUncompress2( stream, &type) == DOFERROR_OK )
            {
                uint8 stage;
                /* Unmarshal stage 255 (storage) */
                if( OALStream_Read8(stream, &stage) && stage == 255 )
                {
                    /* Unmarshal Domain */
                    DOFObjectID domainoid = DOFObjectID_Unmarshal(stream);
                    if( domainoid )
                    {
                        DOFObjectIDDomain domain = DOFObjectIDDomain_Create(domainoid);
                        if( domain )
                        {
                            /* Unamrshal Identity */
                            DOFObjectID oid = DOFObjectID_Unmarshal(stream);
                            if( oid )
                            {
                                DOFObjectIDAuthentication identity = DOFObjectIDAuthentication_Create( oid );
                                if( identity )
                                {
                                    switch( type )
                                    {
                                    case DOFCREDENTIALSTYPE_KEY:
                                        {
                                            uint8 key[32];

                                            /* Unmarshal type specific value */
                                            if( OALStream_Read(stream, 32, key) )
                                            {
                                                if( OALStream_GetDataSize( stream ) == 0 )
                                                {
                                                    /* Create credentials */
                                                    ret = DOFCredentialsKey_Internal_Create( domain, identity, key );
                                                }
                                            }
                                        }
                                        break;

                                    case DOFCREDENTIALSTYPE_PASSWORD:
                                        {
                                            uint8 partial_x[32];

                                            /* Unmarshal type specific value */
                                            if( OALStream_Read(stream, 32, partial_x) )
                                            {
                                                if( OALStream_GetDataSize( stream ) == 0 )
                                                {
                                                    /* Create credentials */
                                                    ret = DOFCredentialsPassword_Internal_Create( domain, identity, partial_x );
                                                }
                                            }
                                        }
                                        break;

                                    default:
                                        /* No other types can be stored. */
                                        break;
                                    }
                                    DOFObjectIDAuthentication_Destroy(identity);
                                }
                                DOFObjectID_Destroy(oid);
                            }
                            DOFObjectIDDomain_Destroy(domain);
                        }
                        DOFObjectID_Destroy(domainoid);
                    }
                }
            }
            OALStream_Release(stream, FALSE);
        }
    }
    return ret;
}

/**
@}
**/
