/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/credentials.h>
#include <internal/credentialsinternal.h>
#include <internal/exception.h>
#include <internal/trp_v1.h>

#include "credentials_common.h"
#include "password.h"
#include "key.h"
#include "remotecredentials.h"

typedef struct DOFCredentials_ResolveCtx_t
{
    DOFCredentials                      creds;
    DOFException                        exception;
    OALChannel                          channel;
    DOFAddress                          address;
    DOFCredentials_BeginResolveCallback callback;
    void*                               context;
    boolean                             force;
}DOFCredentials_ResolveCtx_t;

static void DOFCredentials_ResolveCallback(TRP_V1_Context trp_context, void* context)
{
    DOFCredentials_ResolveCtx_t* ctx = (DOFCredentials_ResolveCtx_t*)context;

    if( trp_context )
    {
        if( trp_context->ResolveReturn->status == DOFERROR_OK )
        {
            if( trp_context->ResolveReturn->errorCode == 0 )
            {
                DOFCredentials_Update( ctx->creds, trp_context->RemoteDomainReturn->creds );
                ctx->callback( ctx->creds, NULL, ctx->context );
            }
            else
            {
                ctx->exception = DOFErrorException_GetFromError( TRP_V1_GetDOFErrorFromErrorCode( trp_context->ResolveReturn->errorCode ) );
            }
        }
        else
        {
            ctx->exception =  DOFErrorException_GetFromError( trp_context->ResolveReturn->status );
        }
    }
    else
    {
        ctx->exception = DOFEXCEPTION_FAILED;
    }

    if( ctx->exception )
        ctx->callback( ctx->creds, ctx->exception, ctx->context );

    DOFCredentials_Destroy( ctx->creds );
    DOFAddress_Destroy( ctx->address );
    OALChannel_Release( ctx->channel );
    DOFMem_Free( ctx );
}

static void DOFCredentials_RemoteCallback(TRP_V1_Context trp_context, void* context)
{
    DOFCredentials_ResolveCtx_t* ctx = (DOFCredentials_ResolveCtx_t*)context;

    if( trp_context )
    {
        if( trp_context->ResolveReturn->status == DOFERROR_OK )
        {
            if( trp_context->ResolveReturn->errorCode == 0 )
            {
                DOFCredentials_Update( ctx->creds, trp_context->RemoteDomainReturn->creds );
                ctx->callback( ctx->creds, NULL, ctx->context );
            }
            else
            {
                ctx->exception = DOFErrorException_GetFromError( TRP_V1_GetDOFErrorFromErrorCode( trp_context->ResolveReturn->errorCode ) );
            }
        }
        else
        {
            ctx->exception =  DOFErrorException_GetFromError( trp_context->ResolveReturn->status );
        }
    }
    else
    {
        ctx->exception = DOFEXCEPTION_FAILED;
    }

    if( ctx->exception )
        ctx->callback( ctx->creds, ctx->exception, ctx->context );

    DOFCredentials_Destroy( ctx->creds );
    DOFAddress_Destroy( ctx->address );
    OALChannel_Release( ctx->channel );
    DOFMem_Free( ctx );
}

static void DOFCredentials_ResolveFirst( DOFCredentials self, DOFException exception, void* context )
{
    DOFCredentials_ResolveCtx_t* ctx = (DOFCredentials_ResolveCtx_t*)context;
    RemoteCredentials_t* ourCreds = (RemoteCredentials_t*)ctx->creds;

    if( !exception )
    {
        DOFKeyRequest kr = DOFKeyRequest_Create( ourCreds->creds, DEFAULT_NONCE_SIZE, TRUE, FALSE, NULL, NULL );
        if( kr )
        {
            TRP_V1_RequestRemoteDomainForcable( ourCreds->base.domain, kr, DOFCredentials_RemoteCallback, ctx, ctx->channel, ctx->address, 15000, ctx->force );
            DOFKeyRequest_Destroy( kr );
            return;
        }
        else
        {
            ctx->exception = DOFEXCEPTION_INSUFFICIENT_RESOURCES;
        }
    }
    else
    {
        ctx->exception = exception;
    }

    ctx->callback( self, ctx->exception, ctx->context );

    DOFCredentials_Destroy( ctx->creds );
    DOFAddress_Destroy( ctx->address );
    OALChannel_Release( ctx->channel );
    DOFMem_Free( ctx );
}

void DOFCredentials_Internal_BeginResolve( DOFCredentials self, DOF dof, int timeout, boolean explicitRoute, OALChannel channel, DOFAddress address, DOFCredentials_BeginResolveCallback func, void* context )
{
    DOFCredentials_ResolveCtx_t* ctx;

    if( !DOFCredentials_IsResolved( self ) )
    {
        if( !explicitRoute || (explicitRoute && channel) )/* must atleast have a channel to force it anywhere */
        {
            ctx = (DOFCredentials_ResolveCtx_t*)DOFMem_Alloc( sizeof( DOFCredentials_ResolveCtx_t ) );
            if( ctx )
            {
                ctx->channel = OALChannel_Retain( channel );
                if( ctx->channel || !channel )
                {
                    ctx->address = DOFAddress_Copy( address );
                    if( ctx->address || !address )
                    {
                        ctx->creds = DOFCredentials_Copy( self );
                        if( ctx->creds )
                        {
                            ctx->exception = NULL;
                            ctx->force = explicitRoute;
                            ctx->callback = func;
                            ctx->context = context;

                            if( self->base.type != DOFCREDENTIALSTYPE_REMOTEKEY )
                            {
                                /* standard type do standard resolve through trp */
                                TRP_V1_Internal_RequestResolveCredentials( self->base.domain, NULL, self, DOFCredentials_ResolveCallback, ctx, channel, address, timeout, explicitRoute );
                            }
                            else
                            {
                                RemoteCredentials_t* selfx = (RemoteCredentials_t*)self;

                                if( DOFCredentials_IsResolved( selfx->creds ) )
                                {
                                    /* do remote domain */
                                    DOFKeyRequest kr = DOFKeyRequest_Create( selfx->creds, DEFAULT_NONCE_SIZE, TRUE, FALSE, NULL, NULL);
                                    if( kr )
                                    {
                                        TRP_V1_RequestRemoteDomainForcable( selfx->base.domain, kr, DOFCredentials_RemoteCallback, ctx, channel, address, timeout, explicitRoute );
                                        DOFKeyRequest_Destroy( kr );
                                    }
                                    else
                                    {
                                        DOFCredentials_Destroy( ctx->creds );
                                        DOFAddress_Destroy( ctx->address );
                                        OALChannel_Release( ctx->channel );
                                        DOFMem_Free( ctx );
                                        func( self, DOFEXCEPTION_INSUFFICIENT_RESOURCES, context );
                                    }
                                }
                                else
                                {
                                    DOFCredentials_Internal_BeginResolve( selfx->creds, dof, timeout, FALSE, channel, address, DOFCredentials_ResolveFirst, ctx );
                                }
                            }
                            return;
                        }
                        DOFAddress_Destroy( ctx->address );
                    }
                    OALChannel_Release( ctx->channel );
                }
                DOFMem_Free( ctx );
            }
        }
        func( self, DOFEXCEPTION_INSUFFICIENT_RESOURCES, context );
        return;
    }
    func( self, NULL, context );
}

struct DOFCredentials_ResolveWithChannelCtx_t
{
    PCREvent        event;
    DOFException    exception;
}DOFCredentials_ResolveWithChannelCtx_t;

static void DOFCredentials_ResolveWithChannelCallback( DOFCredentials creds, DOFException exception, void* context )
{
    struct DOFCredentials_ResolveWithChannelCtx_t* ctx = (struct DOFCredentials_ResolveWithChannelCtx_t*)context;
    ctx->exception = exception;
    PCREvent_Signal( ctx->event );
}

boolean DOFCredentials_Internal_Resolve( DOFCredentials self, DOF dof, int timeout, DOFException* exception, boolean explicitRoute, OALChannel channel, DOFAddress address )
{
    if( !DOFCredentials_IsResolved( self ) )
    {
        boolean ret = FALSE;
        struct DOFCredentials_ResolveWithChannelCtx_t ctx;
        ctx.event = PCREvent_Create( TRUE );

        if( ctx.event )
        {
            PCRStatus status;

            DOFCredentials_Internal_BeginResolve( self, dof, timeout, explicitRoute, channel, address, DOFCredentials_ResolveWithChannelCallback, &ctx );

            status = PCREvent_Wait( ctx.event, DOF_TIMEOUT_NEVER );
            if( status != PCRSTATUS_OK )
            {
                ctx.exception = DOFEXCEPTION_FAILED;
            }

            if( ctx.exception )
            {
                if( exception )
                    (*exception) = DOFException_Copy( ctx.exception );
            }
            else
            {
                PCR_Assert( DOFCredentials_IsResolved( self ) );
                ret = TRUE;
            }
            PCREvent_Destroy( ctx.event );
        }
        return ret;
    }
    return TRUE;
}

boolean DOFCredentials_Resolve( DOFCredentials self, DOF dof, int timeout, DOFException* exception )
{
    if( DOFCredentials_ContainsPrivateData( self ) )
        return DOFCredentials_Internal_Resolve( self, dof, timeout, exception, FALSE, NULL, NULL );

    if( *exception )
        (*exception) = DOFEXCEPTION_FAILED;
    return FALSE;
}

static void ValidateTRPCallback( TRP_V1_Context trp_context, void* context )
{
    DOFCredentials_ResolveCtx_t* ctx = (DOFCredentials_ResolveCtx_t*)context;

    if( trp_context )
    {
        if( trp_context->ValidateReturn->status == DOFERROR_OK )
        {
            if( trp_context->ValidateReturn->errorCode == 0 )
            {
                /* success */
                ctx->callback( trp_context->ValidateReturn->creds, NULL, ctx->context );
            }
            else
            {
                DOFCredentials_Reset( ctx->creds );
                ctx->exception = DOFErrorException_GetFromError( TRP_V1_GetDOFErrorFromErrorCode( trp_context->ValidateReturn->errorCode ) );
            }
        }
        else
        {
            ctx->exception =  DOFErrorException_GetFromError( trp_context->ValidateReturn->status );
        }
    }
    else
    {
        ctx->exception = DOFEXCEPTION_FAILED;
    }

    if(ctx->exception)
        ctx->callback( ctx->creds, ctx->exception, ctx->context );

    OALChannel_Release( ctx->channel );
    DOFAddress_Destroy( ctx->address );
    DOFMem_Free( ctx );
}

void DOFCredentials_Internal_BeginValidate( DOFCredentials self, DOF dof, int timeout, boolean explicitRoute, OALChannel channel, DOFAddress address, DOFCredentials_BeginResolveCallback func, void* context )
{
    DOFCredentials_ResolveCtx_t* ctx;

    if( DOFCredentials_IsResolved( self ) )
    {
        if( !explicitRoute || (explicitRoute && channel) )/* must atleast have a channel to force it anywhere */
        {
            ctx = (DOFCredentials_ResolveCtx_t*)DOFMem_Alloc( sizeof( DOFCredentials_ResolveCtx_t ) );
            if( ctx )
            {
                ctx->channel = OALChannel_Retain( channel );
                if( ctx->channel || !channel )
                {
                    ctx->address = DOFAddress_Copy( address );
                    if( ctx->address || !address )
                    {
                        ctx->creds = self;
                        ctx->exception = NULL;
                        ctx->force = explicitRoute;
                        ctx->callback = func;
                        ctx->context = context;

                        TRP_V1_Internal_RequestValidateCredentials( ctx->creds, ValidateTRPCallback, ctx, channel, address, 15000, explicitRoute );

                        return;
                    }
                    OALChannel_Release( ctx->channel );
                }
                DOFMem_Free( ctx );
            }
        }
    }
    func( self, DOFEXCEPTION_FAILED, context );
}
