/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/error.h>
#include <dof/credentials.h>
#include <dof/objectid.h>
#include <internal/trp_v1.h>
#include <internal/exception.h>
#include <internal/objectidinternal.h>
#include <internal/credentialsinternal.h>

#include "credentials_common.h"

#define OALCREDENTIALSTREE_CONNECTION_VALIDATION_TIMEOUT 15000

extern PCRMutex ref_count_mutex;

extern OALEnqueue credentials_enqueue;

typedef struct OALCredentialsListItem_t
{
    struct OALCredentialsListItem_t*    next;
    DOFCredentials                      creds;
    OALCredentialsTree                  myTree;
    OALObserver                         myObserver;
}OALCredentialsListItem_t;

static void ResolvedToNot( OALObserver self, OALObservable observable, void *arg );

static boolean OALCredentialsTree_CompareGreaterThan_fn_t(DOFObjectIDDomain domain, DOFCredentials item, void *context)
{
    return ( DOFObjectID_Compare( DOFObjectIDDomain_GetObjectID( domain ), DOFObjectIDDomain_GetObjectID( DOFCredentials_GetDomainID( item ) ) ) > 0 );
}

static boolean OALCredentialsTree_CompareEqual_fn_t(DOFObjectIDDomain domain, DOFCredentials item, void *context)
{
    return ( DOFObjectID_IsEqual( DOFObjectIDDomain_GetObjectID( domain ), DOFObjectIDDomain_GetObjectID( DOFCredentials_GetDomainID( item ) ) ) );
}

OALCredentialsTree OALCredentialsTree_Create(void)
{
    OALCredentialsTree ret = (OALCredentialsTree)DOFMem_Alloc(sizeof(OALCredentialsTree_t));

    if( ret )
    {
        ret->refCount = 1;
        ret->credentials = OALAVLTree_Create( (OALAVLTree_CompareGreaterThan_fn_t)OALCredentialsTree_CompareGreaterThan_fn_t, (OALAVLTree_CompareEqual_fn_t)OALCredentialsTree_CompareEqual_fn_t, NULL );
        if(ret->credentials)
        {
            ret->kownDomainCreds = NULL;
            ret->unkownDomainCreds =  NULL;
            ret->unknownDomainCount = 0;
            ret->knownDomainCount = 0;
            ret->wildcardCreds = NULL;
            return ret;
        }
        DOFMem_Free(ret);
    }
    return NULL;
}

OALCredentialsTree OALCredentialsTree_Copy(OALCredentialsTree self)
{
    OALCredentialsTree rval = self;

    if (self)
    {
        PCRMutex_Lock(ref_count_mutex);
        if( self->refCount < 0xFFFFFFFF )
            self->refCount++;
        else
            rval = NULL;
        PCRMutex_Unlock(ref_count_mutex);
    }
    return( rval );
}

static void OALCredentialsTree_RemoveAll_fn_t(OALAVLTreeItem_t itemx, void *arg)
{
    DOFCredentials item = (DOFCredentials)itemx;
    DOFCredentials_Destroy( item );
}

static void OALCredentialsTree_ClearList( OALSListNode itemx, void *arg )
{
    OALCredentialsListItem_t* item = (OALCredentialsListItem_t*)itemx;
    DOFCredentials_RemoveChangeObserver( item->creds, item->myObserver );
    DOFCredentials_Destroy(item->creds);
    DOFMem_Free(item);
}

void OALCredentialsTree_Destroy( OALCredentialsTree self )
{
    boolean destroy = FALSE;
    if ( self )
    {
        PCRMutex_Lock(ref_count_mutex);
        PCR_Assert( self->refCount > 0 );
        if ( (--self->refCount) == 0 )
            destroy = TRUE;
        PCRMutex_Unlock(ref_count_mutex);
        if ( destroy )
        {
            DOFCredentials_Destroy( self->wildcardCreds );
            OALAVLTree_Destroy(self->credentials, OALCredentialsTree_RemoveAll_fn_t, NULL );
            if( self->unkownDomainCreds )
                OALSList_RemoveAll(&self->unkownDomainCreds, OALCredentialsTree_ClearList,NULL);
            if( self->kownDomainCreds )
                OALSList_RemoveAll(&self->kownDomainCreds, OALCredentialsTree_ClearList,NULL);
            DOFMem_Free(self);
        }
    }
}

struct deepCopyCnt_t
{
    OALCredentialsTree dstTree;
    boolean error;
};

static boolean OALCredentialsTree_Traversal_Copy(OALAVLTreeItem_t itemx, void* arg )
{
    DOFCredentials item = (DOFCredentials)itemx;
    struct deepCopyCnt_t* ctx = (struct deepCopyCnt_t*)arg;

    if( !OALCredentialsTree_Insert( ctx->dstTree, item ) )
    {
        ctx->error = TRUE;
        return TRUE;
    }
    return FALSE;
}

static boolean OALCredentialsTree_Traversal_List_Copy(OALSListNode itemx, void* arg)
{
    OALCredentialsListItem_t* item = (OALCredentialsListItem_t*)itemx;
    struct deepCopyCnt_t* ctx = (struct deepCopyCnt_t*)arg;

    if( !OALCredentialsTree_Insert( ctx->dstTree, item->creds ))
    {
        ctx->error = TRUE;
        return TRUE;
    }
    return FALSE;
}

static boolean OALCredentialsTree_CopyKnownCredential( OALCredentialsTree self, DOFCredentials creds)
{
    OALCredentialsListItem_t* item = DOFMem_Alloc( sizeof(OALCredentialsListItem_t) );
    if( item )
    {
        item->creds =  DOFCredentials_Copy( creds );
        if( item->creds )
        {
            item->myObserver = DOFCredentials_AddChangeObserver( item->creds, ResolvedToNot, item );
            if( item->myObserver )
            {
                item->myTree = self;
                OALSList_Insert((&self->kownDomainCreds), (OALSListNode)item);
                CredentialsTree_Unlock();
                return TRUE;
            }
            DOFCredentials_Destroy( creds );
        }
        DOFMem_Free( item );
    }
    return FALSE;
}

static boolean OALCredentialsTree_Traversal_Known_List_Copy(OALSListNode itemx, void* arg)
{
    OALCredentialsListItem_t* item = (OALCredentialsListItem_t*)itemx;
    struct deepCopyCnt_t* ctx = (struct deepCopyCnt_t*)arg;

    if( !OALCredentialsTree_CopyKnownCredential( ctx->dstTree, item->creds ) )
    {
        ctx->error = TRUE;
        return TRUE;
    }
    return FALSE;
}

OALCredentialsTree OALCredentialsTree_DeepCopy( OALCredentialsTree srcTree )
{
    if( srcTree )
    {
        struct deepCopyCnt_t ctx;
        OALCredentialsTree ret = OALCredentialsTree_Create();

        ctx.dstTree = ret;
        ctx.error = FALSE;

        if( ret )
        {
            CredentialsTree_Lock();
            ret->wildcardCreds = DOFCredentials_Copy( srcTree->wildcardCreds );
            if( ret->wildcardCreds || !srcTree->wildcardCreds )
            {
                OALAVLTree_Traverse(srcTree->credentials, OALCredentialsTree_Traversal_Copy, &ctx);
                if( !ctx.error )
                {
                    OALSList_Traverse(srcTree->unkownDomainCreds, OALCredentialsTree_Traversal_List_Copy, &ctx);
                    if( !ctx.error )
                    {
                        OALSList_Traverse(srcTree->kownDomainCreds, OALCredentialsTree_Traversal_Known_List_Copy, &ctx);
                        if( !ctx.error )
                        {
                            CredentialsTree_Unlock();
                            return ret;
                        }
                    }
                }
            }
            OALCredentialsTree_Destroy( ret );
        }
        CredentialsTree_Unlock();
    }
    return NULL;
}

boolean OALCredentialsTree_SetWildcardCredentials( OALCredentialsTree credTree, DOFCredentials creds )
{
    PCR_Assert( credTree );
    CredentialsTree_Lock();
    if( creds )
    {
        DOFCredentials tempOld = credTree->wildcardCreds;
        credTree->wildcardCreds = DOFCredentials_Copy( creds );
        if( credTree->wildcardCreds )
        {
            DOFCredentials_Destroy( tempOld );
            CredentialsTree_Unlock();
            return TRUE;
        }
        credTree->wildcardCreds = tempOld;
    }
    else
    { 
        credTree->wildcardCreds = NULL;
        CredentialsTree_Unlock();
        return TRUE;
    }
    CredentialsTree_Unlock();
    return FALSE;
}

DOFCredentials OALCredentialsTree_GetWildcardCredentials( OALCredentialsTree self )
{
    if( self )
        return self->wildcardCreds;
    return NULL;
}

static boolean FindCredentialsListItem(OALSListNode item, void *arg)
{
    if( item == arg )
        return TRUE;
    return FALSE;
}

static void ResolvedToNot( OALObserver self, OALObservable observable, void *arg )
{
    OALCredentialsListItem_t* item = (OALCredentialsListItem_t*)arg;

    CredentialsTree_Lock();
    if( DOFCredentials_IsResolved( item->creds ) )
    {
        OALCredentialsListItem_t* found = (OALCredentialsListItem_t*)OALSList_FindAndRemove( &(item->myTree->unkownDomainCreds), NULL, FindCredentialsListItem, (void*)item );
        if( found )
        {
            OALSList_Insert( &(item->myTree->kownDomainCreds), (OALSListNode)item );
            --(item->myTree->unknownDomainCount);
            ++(item->myTree->knownDomainCount);
        }
    }
    else
    {
        OALCredentialsListItem_t* found = (OALCredentialsListItem_t*)OALSList_FindAndRemove( &(item->myTree->kownDomainCreds), NULL, FindCredentialsListItem, (void*)item );
        if( found )
        {
            OALSList_Insert( &(item->myTree->unkownDomainCreds), (OALSListNode)item );
            ++(item->myTree->unknownDomainCount);
            --(item->myTree->knownDomainCount);
        }
    }
    CredentialsTree_Unlock();
}

boolean OALCredentialsTree_Insert( OALCredentialsTree credTree, DOFCredentials creds )
{
    PCR_Assert( creds && credTree );

    if( DOFCredentials_ContainsPrivateData( creds ) )
    {      
        CredentialsTree_Lock();
        /* if a credential that doesn't have a fixed domain our avl tree will most likely be corrupted*/
        if( DOFCredentials_HasFixedDomain( creds ) )
        {
            DOFCredentials item = DOFCredentials_Copy( creds );
            if( item )
            {
                switch( OALAVLTree_Insert(credTree->credentials, DOFCredentials_GetDomainID( item ), item, NULL) )
                {
                case OALAVLTree_AlreadyInserted:
                    DOFCredentials_Destroy( item );
                    /* Keep fall through here */

                case OALAVLTree_InsertSuccess:
                    CredentialsTree_Unlock();
                    return TRUE;
                    break;

                default:
                case OALAVLTree_InsertFailure:
                    CredentialsTree_Unlock();
                    break;
                }
                DOFCredentials_Destroy( item );
            }
        }
        else
        {
            OALCredentialsListItem_t* item = DOFMem_Alloc( sizeof(OALCredentialsListItem_t) );
            if( item )
            {
                item->creds =  DOFCredentials_Copy( creds );
                if( item->creds )
                {
                    item->myObserver = DOFCredentials_AddChangeObserver( item->creds, ResolvedToNot, item );
                    if( item->myObserver )
                    {
                        item->myTree = credTree;
                        if( !DOFCredentials_IsResolved( item->creds ) )
                        {
                            OALSList_Insert((&credTree->unkownDomainCreds), (OALSListNode)item);
                            ++(credTree->unknownDomainCount);
                        }
                        else
                        {
                            OALSList_Insert((&credTree->kownDomainCreds), (OALSListNode)item);
                            ++(credTree->knownDomainCount);
                        }
                        CredentialsTree_Unlock();
                        return TRUE;
                    }
                    DOFCredentials_Destroy( creds );
                }
                DOFMem_Free( item );
            }
        }
        CredentialsTree_Unlock();
    }
    return FALSE;
}

struct tempTreeCmp_t
{
    boolean         cmp;
    OALCredentialsTree other;
}tempTreeCmp_t;

static boolean OALCredentialsTree_Traversal_Cmp( OALAVLTreeItem_t itemx, void* args )
{
    DOFCredentials item = (DOFCredentials)itemx;
    struct tempTreeCmp_t* arg = (struct tempTreeCmp_t*)args;
    DOFCredentials otherItem = (DOFCredentials)OALAVLTree_Retrieve( arg->other->credentials, DOFCredentials_GetDomainID( item ) );

    if( !otherItem || !DOFCredentials_IsEqual( item, otherItem ) )
    {
        arg->cmp = FALSE;
        return TRUE;
    }
    return FALSE;
}

static boolean OALCredentialsTree_Cmp(OALSListNode itemx, void* argx)
{
    OALCredentialsListItem_t* item = (OALCredentialsListItem_t*)itemx;
    DOFCredentials arg = (DOFCredentials)argx;
    if( DOFCredentials_IsEqual( item->creds, arg ) )
        return TRUE;
    return FALSE;
}

static boolean OALCredentialsTree_Traversal_List_Cmp(OALSListNode itemx, void* args)
{
    OALCredentialsListItem_t* item = (OALCredentialsListItem_t*)itemx;
    struct tempTreeCmp_t* arg = (struct tempTreeCmp_t*)args;
    arg->cmp = OALSList_Traverse( arg->other->unkownDomainCreds, OALCredentialsTree_Cmp, item->creds);
    return !arg->cmp;
}

static boolean OALCredentialsTree_Traversal_Known_List_Cmp(OALSListNode itemx, void* args)
{
    OALCredentialsListItem_t* item = (OALCredentialsListItem_t*)itemx;
    struct tempTreeCmp_t* arg = (struct tempTreeCmp_t*)args;
    arg->cmp = OALSList_Traverse( arg->other->kownDomainCreds, OALCredentialsTree_Cmp, item->creds);
    return !arg->cmp;
}

boolean OALCredentialsTree_IsEqual( OALCredentialsTree self, OALCredentialsTree other )
{
    struct tempTreeCmp_t arg;
    arg.other = other;
    arg.cmp = TRUE;

    if( self == other )
        return TRUE;

    if( !self || !other )
        return FALSE;

    CredentialsTree_Lock();

    if( !DOFCredentials_IsEqual( self->wildcardCreds, other->wildcardCreds ) )
    {
        CredentialsTree_Unlock();
        return FALSE;
    }

    if( ( OALAVLTree_GetNumItems( self->credentials ) != OALAVLTree_GetNumItems( other->credentials ) ) || 
        ( self->unknownDomainCount != other->unknownDomainCount ) || 
        ( self->knownDomainCount != other->knownDomainCount ) )
    {
        arg.cmp = FALSE;
    }
    else
    {
        OALAVLTree_Traverse( self->credentials, OALCredentialsTree_Traversal_Cmp, &arg );
        OALSList_Traverse( self->unkownDomainCreds, OALCredentialsTree_Traversal_List_Cmp, &arg);
        OALSList_Traverse( self->kownDomainCreds, OALCredentialsTree_Traversal_Known_List_Cmp, &arg);
    }
    CredentialsTree_Unlock();
    return arg.cmp;
}

boolean OALCredentialsTree_HasCredentials( OALCredentialsTree self )
{
    CredentialsTree_Lock();
    if( self && (self->unkownDomainCreds || self->kownDomainCreds || (OALAVLTree_GetNumItems(self->credentials) > 0)) )
    {
        CredentialsTree_Unlock();
        return TRUE;
    }
    CredentialsTree_Unlock();
    return FALSE;
}

typedef struct udct_t
{
    OALCredentialsTree tree;
    DOFObjectIDDomain domainID;
    OALCredentialsTree_GetCredentials_Callback_Fn fn;
    void* context;
    uint32 timeout;
    uint32 sentCount;
    uint32 responseCount;
    boolean notifiedCallback;
}udct_t;

static void OALCredentialsTree_DOFCredentials_BeginResolveCallback( DOFCredentials creds, DOFException exception, void* context )
{
    udct_t* ctx = (udct_t*)context;
    boolean goingToNotify = FALSE;
    boolean cleanup = FALSE;
    
    CredentialsTree_Lock();

    ++(ctx->responseCount);

    if( !exception )
    {
        if( DOFObjectID_IsEqual( DOFObjectIDDomain_GetObjectID( ctx->domainID ), DOFObjectIDDomain_GetObjectID( DOFCredentials_GetDomainID( creds ) ) ) )
        {
            /* Right domain */
            if( !ctx->notifiedCallback )
            {
                goingToNotify = TRUE;
                ctx->notifiedCallback = TRUE;
            }
        }        
    }

    /* received all our responses */
    if(  ctx->sentCount == ctx->responseCount )
    {
        if( !ctx->notifiedCallback )
            goingToNotify = TRUE;

        cleanup = TRUE;
    }

    CredentialsTree_Unlock();

    if( goingToNotify )
    {
        if( ctx->notifiedCallback )
            ctx->fn( creds, ctx->context );
        else
            ctx->fn( NULL, ctx->context );
    }

    if( cleanup )
    {
        DOFObjectIDDomain_Destroy( ctx->domainID );
        OALCredentialsTree_Destroy( ctx->tree );
        DOFMem_Free( ctx );
    }
}

static boolean OALCredentialsTree_UnknownDomain_Traversal(OALSListNode item, void *arg)
{
    DOFCredentials creds = (DOFCredentials)item;
    udct_t* ctx = (udct_t*)arg;

    DOFCredentials_Internal_BeginResolve( creds, DOF_GetDof(), ctx->timeout, FALSE, NULL, NULL, OALCredentialsTree_DOFCredentials_BeginResolveCallback, ctx );

    return FALSE;
}

typedef struct knownTraversal_t
{
    DOFCredentials creds;
    DOFObjectIDDomain domain;
}knownTraversal_t;

static boolean OALCredentialsTree_KnownDomain_Traversal(OALSListNode item, void *arg)
{
    DOFCredentials creds = (DOFCredentials)item;
    knownTraversal_t* ctx = (knownTraversal_t*)arg;

    if( DOFObjectID_IsEqual( DOFObjectIDDomain_GetObjectID( ctx->domain ), DOFObjectIDDomain_GetObjectID( DOFCredentials_GetDomainID( creds ) ) ) )
    {
        ctx->creds = creds;
        return TRUE;
    }
    return FALSE;
}

/* this may call the callback before leaving this function */
/* this only makes sense in the case that server config uses this structure */
void OALCredentialsTree_GetCredentials( OALCredentialsTree self, DOFObjectIDDomain domainID, uint32 timeout, OALCredentialsTree_GetCredentials_Callback_Fn fn, void* context )
{
    DOFCredentials temp;
    CredentialsTree_Lock();
    if( self )
    {
        if( self->credentials )
        {
            temp = (DOFCredentials)OALAVLTree_Retrieve( self->credentials, domainID );
            if( temp )
            {
                CredentialsTree_Unlock();
                fn( temp, context );
                return;
            }
        }

        if( self->kownDomainCreds )
        {
            knownTraversal_t ctx;
            ctx.creds = NULL;
            ctx.domain = domainID;
            OALSList_Traverse( self->kownDomainCreds, OALCredentialsTree_KnownDomain_Traversal, &ctx );
            if( ctx.creds )
            {
                CredentialsTree_Unlock();
                fn( ctx.creds, context );
                return;
            }
        }

        if( self->unknownDomainCount > 0 || 
            self->wildcardCreds != NULL )
        {
            udct_t* ctx = (udct_t*)DOFMem_Alloc( sizeof( udct_t ) );
            if( ctx )
            {
                ctx->tree = OALCredentialsTree_Copy( self );
                if( ctx->tree )
                {
                    ctx->domainID = DOFObjectIDDomain_Copy( domainID );
                    if( ctx->domainID )
                    {
                        ctx->context = context;
                        ctx->fn = fn;
                        ctx->timeout = timeout;
                        ctx->sentCount = ( self->unknownDomainCount + ((self->wildcardCreds)?1:0) );
                        ctx->responseCount = 0;
                        ctx->notifiedCallback = FALSE;
                        
                        if( self->unkownDomainCreds )
                        {
                            OALSList_Traverse( self->unkownDomainCreds, OALCredentialsTree_UnknownDomain_Traversal, ctx );
                        }

                        if( self->wildcardCreds )
                        {
                            /* wrap in domain and try to resolve */
                            DOFCredentials newCreds = DOFCredentials_Create( self->wildcardCreds, domainID );
                            DOFCredentials_Internal_BeginResolve( newCreds, DOF_GetDof(), timeout, FALSE, NULL, NULL, OALCredentialsTree_DOFCredentials_BeginResolveCallback, ctx );
                        }

                        CredentialsTree_Unlock();
                        return;
                    }
                    OALCredentialsTree_Destroy( ctx->tree );
                }
                DOFMem_Free( ctx );
            }
        }
    }

    CredentialsTree_Unlock();
    fn( NULL, context );
}

