/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/error.h>
#include <dof/credentials.h>
#include <dof/objectid.h>
#include <internal/trp_v1.h>
#include <internal/exception.h>
#include <internal/objectidinternal.h>
#include <internal/domaininternal.h>
#include <internal/credentialsinternal.h>

#include "credentials_common.h"

#define OALCREDENTIALSTREE_CONNECTION_VALIDATION_TIMEOUT 15000

extern PCRMutex ref_count_mutex;

extern OALEnqueue credentials_enqueue;

typedef struct OALDomainCredentialsListItem_t
{
    struct OALDomainCredentialsListItem_t*  next;
    DOFDomain                               domain;
    OALDomainCredentialsTree                myTree;
    OALObserver                             myObserver;
}OALDomainCredentialsListItem_t;

static void ResolvedToNot( OALObserver self, OALObservable observable, void *arg );

static boolean OALDomainCredentialsTree_CompareGreaterThan_fn_t(DOFObjectIDDomain domain, DOFDomain item, void *context)
{
    return ( DOFObjectID_Compare( DOFObjectIDDomain_GetObjectID( domain ), DOFObjectIDDomain_GetObjectID( DOFCredentials_GetDomainID( DOFDomain_GetCredentials( item ) ) ) ) > 0 );
}

static boolean OALDomainCredentialsTree_CompareEqual_fn_t(DOFObjectIDDomain domain, DOFDomain item, void *context)
{
    return ( DOFObjectID_IsEqual( DOFObjectIDDomain_GetObjectID( domain ), DOFObjectIDDomain_GetObjectID( DOFCredentials_GetDomainID( DOFDomain_GetCredentials( item ) ) ) ) );
}

static boolean OALDomainCredentialsTree_CompareGreaterThan_TD(DOFObjectIDDomain domain, DOFObjectIDDomain domain2, void *context)
{
    return ( DOFObjectID_Compare( DOFObjectIDDomain_GetObjectID( domain ), DOFObjectIDDomain_GetObjectID( domain2 ) ) > 0 );
}

static boolean OALDomainCredentialsTree_CompareEqual_TD(DOFObjectIDDomain domain, DOFObjectIDDomain domain2, void *context)
{
    return ( DOFObjectID_IsEqual( DOFObjectIDDomain_GetObjectID( domain ), DOFObjectIDDomain_GetObjectID( domain2 ) ) );
}

OALDomainCredentialsTree OALDomainCredentialsTree_Create(void)
{
    OALDomainCredentialsTree ret = (OALDomainCredentialsTree)DOFMem_Alloc(sizeof(OALDomainCredentialsTree_t));

    if( ret )
    {
        ret->refCount = 1;
        ret->credentials = OALAVLTree_Create( (OALAVLTree_CompareGreaterThan_fn_t)OALDomainCredentialsTree_CompareGreaterThan_fn_t, (OALAVLTree_CompareEqual_fn_t)OALDomainCredentialsTree_CompareEqual_fn_t, NULL );
        if(ret->credentials)
        {
            ret->trustedDomains = OALAVLTree_Create( (OALAVLTree_CompareGreaterThan_fn_t)OALDomainCredentialsTree_CompareGreaterThan_TD, (OALAVLTree_CompareEqual_fn_t)OALDomainCredentialsTree_CompareEqual_TD, NULL );
            if(ret->trustedDomains)
            {
                ret->kownDomainCreds = NULL;
                ret->unkownDomainCreds =  NULL;
                ret->unknownDomainCount = 0;
                ret->knownDomainCount = 0;
                ret->maxSilence = DOF_TIMEOUT_NEVER;
                ret->wildcardCreds = NULL;
                return ret;
            }
            OALAVLTree_Destroy(ret->credentials, NULL, NULL);
        }
        DOFMem_Free(ret);
    }
    return NULL;
}

OALDomainCredentialsTree OALDomainCredentialsTree_Copy(OALDomainCredentialsTree self)
{
    OALDomainCredentialsTree rval = self;

    if (self)
    {
        PCRMutex_Lock(ref_count_mutex);
        if( self->refCount < 0xFFFFFFFF )
            self->refCount++;
        else
            rval = NULL;
        PCRMutex_Unlock(ref_count_mutex);
    }
    return( rval );
}

static void OALDomainCredentialsTree_RemoveAll_fn_t(OALAVLTreeItem_t itemx, void *arg)
{
    DOFDomain item = (DOFDomain)itemx;
    DOFDomain_Destroy( item );
}

static void OALDomainCredentialsTree_RemoveAllTrustedDomains( OALAVLTreeItem_t item, void *arg )
{
    DOFObjectIDDomain domain = (DOFObjectIDDomain)item;
    DOFObjectIDDomain_Destroy(domain);
}

static void OALDomainCredentialsTree_ClearList( OALSListNode itemx, void *arg )
{
    OALDomainCredentialsListItem_t* item = (OALDomainCredentialsListItem_t*)itemx;
    DOFCredentials_RemoveChangeObserver( item->domain->creds, item->myObserver );
    DOFDomain_Destroy( item->domain );
    DOFMem_Free( item );
}

void OALDomainCredentialsTree_Destroy( OALDomainCredentialsTree self )
{
    boolean destroy = FALSE;
    if ( self )
    {
        PCRMutex_Lock(ref_count_mutex);
        PCR_Assert( self->refCount > 0 );
        if ( (--self->refCount) == 0 )
            destroy = TRUE;
        PCRMutex_Unlock(ref_count_mutex);
        if ( destroy )
        {
            DOFCredentials_Destroy( self->wildcardCreds );
            OALAVLTree_Destroy(self->credentials, OALDomainCredentialsTree_RemoveAll_fn_t, NULL );
            OALAVLTree_Destroy(self->trustedDomains, OALDomainCredentialsTree_RemoveAllTrustedDomains, NULL );
            if( self->unkownDomainCreds )
                OALSList_RemoveAll(&self->unkownDomainCreds, OALDomainCredentialsTree_ClearList,NULL);
            if( self->kownDomainCreds )
                OALSList_RemoveAll(&self->kownDomainCreds, OALDomainCredentialsTree_ClearList,NULL);
            DOFMem_Free(self);
        }
    }
}

struct deepCopyCnt_t
{
    OALDomainCredentialsTree dstTree;
    boolean error;
};

static boolean OALDomainCredentialsTree_Traversal_Copy(OALAVLTreeItem_t itemx, void* arg )
{
    DOFDomain item = (DOFDomain)itemx;
    struct deepCopyCnt_t* ctx = (struct deepCopyCnt_t*)arg;

    if( !OALDomainCredentialsTree_Insert( ctx->dstTree, item ) )
    {
        ctx->error = TRUE;
        return TRUE;
    }
    return FALSE;
}

static boolean OALDomainCredentialsTree_Traversal_CopyTD(OALAVLTreeItem_t itemx, void* arg )
{
    DOFObjectIDDomain domain = (DOFObjectIDDomain)itemx;
    struct deepCopyCnt_t* ctx = (struct deepCopyCnt_t*)arg;

    if( !OALDomainCredentialsTree_InsertTrustedDomain( ctx->dstTree, domain ) )
    {
        ctx->error = TRUE;
        return TRUE;
    }
    return FALSE;
}

static boolean OALDomainCredentialsTree_Traversal_List_Copy(OALSListNode itemx, void* arg)
{
    OALDomainCredentialsListItem_t* item = (OALDomainCredentialsListItem_t*)itemx;
    struct deepCopyCnt_t* ctx = (struct deepCopyCnt_t*)arg;

    if( !OALDomainCredentialsTree_Insert( ctx->dstTree, item->domain ))
    {
        ctx->error = TRUE;
        return TRUE;
    }
    return FALSE;
}

static boolean OALDomainCredentialsTree_CopyKnownCredential( OALDomainCredentialsTree self, DOFDomain domain)
{
    OALDomainCredentialsListItem_t* item = DOFMem_Alloc( sizeof(OALDomainCredentialsListItem_t) );
    if( item )
    {
        item->domain =  DOFDomain_Copy( domain );
        if( item->domain )
        {
            item->myObserver = DOFCredentials_AddChangeObserver( item->domain->creds, ResolvedToNot, item );
            if( item->myObserver )
            {
                item->myTree = self;
                OALSList_Insert((&self->kownDomainCreds), (OALSListNode)item);
                CredentialsTree_Unlock();
                return TRUE;
            }
            DOFDomain_Destroy( domain );
        }
        DOFMem_Free( item );
    }
    return FALSE;
}

static boolean OALDomainCredentialsTree_Traversal_Known_List_Copy(OALSListNode itemx, void* arg)
{
    OALDomainCredentialsListItem_t* item = (OALDomainCredentialsListItem_t*)itemx;
    struct deepCopyCnt_t* ctx = (struct deepCopyCnt_t*)arg;

    if( !OALDomainCredentialsTree_CopyKnownCredential( ctx->dstTree, item->domain ) )
    {
        ctx->error = TRUE;
        return TRUE;
    }
    return FALSE;
}

OALDomainCredentialsTree OALDomainCredentialsTree_DeepCopy( OALDomainCredentialsTree srcTree )
{
    if( srcTree )
    {
        struct deepCopyCnt_t ctx;
        OALDomainCredentialsTree ret = OALDomainCredentialsTree_Create();

        ctx.dstTree = ret;
        ctx.error = FALSE;

        if( ret )
        {
            CredentialsTree_Lock();
            ret->wildcardCreds = DOFCredentials_Copy( srcTree->wildcardCreds );
            if( ret->wildcardCreds || !srcTree->wildcardCreds )
            {
                OALAVLTree_Traverse(srcTree->credentials, OALDomainCredentialsTree_Traversal_Copy, &ctx);
                if( !ctx.error )
                {
                    OALAVLTree_Traverse(srcTree->trustedDomains, OALDomainCredentialsTree_Traversal_CopyTD, &ctx);
                    if( !ctx.error )
                    {
                        OALSList_Traverse(srcTree->unkownDomainCreds, OALDomainCredentialsTree_Traversal_List_Copy, &ctx);
                        if( !ctx.error )
                        {
                            OALSList_Traverse(srcTree->kownDomainCreds, OALDomainCredentialsTree_Traversal_Known_List_Copy, &ctx);
                            if( !ctx.error )
                            {
                                CredentialsTree_Unlock();
                                return ret;
                            }
                        }
                    }
                }
            }
            OALDomainCredentialsTree_Destroy( ret );
        }
        CredentialsTree_Unlock();
    }
    return NULL;
}

boolean OALDomainCredentialsTree_SetDomainDiscoveryCredentials( OALDomainCredentialsTree credTree, DOFCredentials creds )
{
    PCR_Assert( credTree );
    CredentialsTree_Lock();
    if( creds )
    {
        DOFCredentials tempOld = credTree->wildcardCreds;
        credTree->wildcardCreds = DOFCredentials_Copy( creds );
        if( credTree->wildcardCreds )
        {
            DOFCredentials_Destroy( tempOld );
            CredentialsTree_Unlock();
            return TRUE;
        }
        credTree->wildcardCreds = tempOld;
    }
    else
    { 
        credTree->wildcardCreds = NULL;
        CredentialsTree_Unlock();
        return TRUE;
    }
    CredentialsTree_Unlock();
    return FALSE;
}

DOFCredentials OALDomainCredentialsTree_GetDomainDiscoveryCredentials( OALDomainCredentialsTree self )
{
    if( self )
        return self->wildcardCreds;
    return NULL;
}

static boolean FindCredentialsListItem(OALSListNode item, void *arg)
{
    if( item == arg )
        return TRUE;
    return FALSE;
}

static void ResolvedToNot(OALObserver self, OALObservable observable, void *arg )
{
    OALDomainCredentialsListItem_t* item = (OALDomainCredentialsListItem_t*)arg;

    CredentialsTree_Lock();
    if( DOFCredentials_IsResolved( item->domain->creds ) )
    {
        OALDomainCredentialsListItem_t* found = (OALDomainCredentialsListItem_t*)OALSList_FindAndRemove( &(item->myTree->unkownDomainCreds), NULL, FindCredentialsListItem, (void*)item );
        if( found )
        {
            OALSList_Insert( &(item->myTree->kownDomainCreds), (OALSListNode)item );
            --(item->myTree->unknownDomainCount);
            ++(item->myTree->knownDomainCount);
        }
    }
    else
    {
        OALDomainCredentialsListItem_t* found = (OALDomainCredentialsListItem_t*)OALSList_FindAndRemove( &(item->myTree->kownDomainCreds), NULL, FindCredentialsListItem, (void*)item );
        if( found )
        {
            OALSList_Insert( &(item->myTree->unkownDomainCreds), (OALSListNode)item );
            ++(item->myTree->unknownDomainCount);
            --(item->myTree->knownDomainCount);
        }
    }
    CredentialsTree_Unlock();
}

boolean OALDomainCredentialsTree_Insert( OALDomainCredentialsTree credTree, DOFDomain domain )
{
    PCR_Assert( domain && credTree );
    CredentialsTree_Lock();
    /* if a credential that doesn't have a fixed domain our avl tree will most likely be corrupted*/
    if( DOFCredentials_HasFixedDomain( DOFDomain_GetCredentials( domain ) ) )
    {
        DOFDomain item = DOFDomain_Copy( domain );
        if( item )
        {
            switch( OALAVLTree_Insert(credTree->credentials, DOFCredentials_GetDomainID( DOFDomain_GetCredentials( domain ) ), item, NULL) )
            {
            case OALAVLTree_AlreadyInserted:
                DOFDomain_Destroy( item );
                /* Keep fall through here */

            case OALAVLTree_InsertSuccess:
                CredentialsTree_Unlock();
                return TRUE;
                break;

            default:
            case OALAVLTree_InsertFailure:
                CredentialsTree_Unlock();
                DOFDomain_Destroy( item );
                break;
            }            
        }
    }
    else
    {
        OALDomainCredentialsListItem_t* item = DOFMem_Alloc( sizeof(OALDomainCredentialsListItem_t) );
        if( item )
        {
            item->domain =  DOFDomain_Copy( domain );
            if( item->domain )
            {
                item->myObserver = DOFCredentials_AddChangeObserver( DOFDomain_GetCredentials( item->domain ), ResolvedToNot, item );
                if( item->myObserver )
                {
                    item->myTree = credTree;
                    if( !DOFCredentials_IsResolved( item->domain->creds ) )
                    {
                        OALSList_Insert((&credTree->unkownDomainCreds), (OALSListNode)item);
                        ++(credTree->unknownDomainCount);
                    }
                    else
                    {
                        OALSList_Insert((&credTree->kownDomainCreds), (OALSListNode)item);
                        ++(credTree->knownDomainCount);
                    }
                    CredentialsTree_Unlock();
                    return TRUE;
                }
                DOFDomain_Destroy( item->domain );
            }
            DOFMem_Free( item );
        }
    }
    CredentialsTree_Unlock();
    return FALSE;
}

boolean OALDomainCredentialsTree_InsertTrustedDomain( OALDomainCredentialsTree self, DOFObjectIDDomain domain )
{
    OALAVLTree_InsertionStatus ret;
    DOFObjectIDDomain d_cpy = DOFObjectIDDomain_Copy(domain);

    PCR_Assert( self );

    if( !d_cpy )
        return FALSE;

    CredentialsTree_Lock();

    ret = OALAVLTree_Insert(self->trustedDomains, d_cpy, d_cpy, NULL);
    if( ret != OALAVLTree_InsertFailure )
    {
        if( ret == OALAVLTree_AlreadyInserted )
        {
            DOFObjectIDDomain_Destroy(d_cpy);
        }

        CredentialsTree_Unlock();
        return TRUE;
    }

    DOFObjectIDDomain_Destroy(d_cpy);
    CredentialsTree_Unlock();
    return FALSE;
}

struct tempTreeCmp_t
{
    boolean         cmp;
    OALDomainCredentialsTree other;
}tempTreeCmp_t;

static boolean OALDomainCredentialsTree_Traversal_Cmp( OALAVLTreeItem_t itemx, void* args )
{
    DOFDomain item = (DOFDomain)itemx;
    struct tempTreeCmp_t* arg = (struct tempTreeCmp_t*)args;
    DOFDomain otherItem = (DOFDomain)OALAVLTree_Retrieve( arg->other->credentials, DOFCredentials_GetDomainID( DOFDomain_GetCredentials( item ) ) );

    if( !otherItem || !DOFCredentials_IsEqual( DOFDomain_GetCredentials( item ), DOFDomain_GetCredentials( otherItem ) ) )
    {
        arg->cmp = FALSE;
        return TRUE;
    }
    return FALSE;
}

static boolean OALDomainCredentialsTree_Traversal_Cmp_TD( OALAVLTreeItem_t itemx, void* args )
{
    DOFObjectIDDomain item = (DOFObjectIDDomain)itemx;
    struct tempTreeCmp_t* arg = (struct tempTreeCmp_t*)args;
    DOFObjectIDDomain otherItem = (DOFObjectIDDomain)OALAVLTree_Retrieve( arg->other->trustedDomains, item );

    if( !otherItem )
    {
        arg->cmp = FALSE;
        return TRUE;
    }
    return FALSE;
}

static boolean OALDomainCredentialsTree_Cmp(OALSListNode itemx, void* argx)
{
    OALDomainCredentialsListItem_t* item = (OALDomainCredentialsListItem_t*)itemx;
    DOFDomain arg = (DOFDomain)argx;
    if( DOFDomain_IsEqual( item->domain, arg ) )
        return TRUE;
    return FALSE;
}

static boolean OALDomainCredentialsTree_Traversal_List_Cmp(OALSListNode itemx, void* args)
{
    OALDomainCredentialsListItem_t* item = (OALDomainCredentialsListItem_t*)itemx;
    struct tempTreeCmp_t* arg = (struct tempTreeCmp_t*)args;
    arg->cmp = OALSList_Traverse( arg->other->unkownDomainCreds, OALDomainCredentialsTree_Cmp, item->domain );
    return !arg->cmp;
}

static boolean OALDomainCredentialsTree_Traversal_Known_List_Cmp(OALSListNode itemx, void* args)
{
    OALDomainCredentialsListItem_t* item = (OALDomainCredentialsListItem_t*)itemx;
    struct tempTreeCmp_t* arg = (struct tempTreeCmp_t*)args;
    arg->cmp = OALSList_Traverse( arg->other->kownDomainCreds, OALDomainCredentialsTree_Cmp, item->domain );
    return !arg->cmp;
}

boolean OALDomainCredentialsTree_IsEqual( OALDomainCredentialsTree self, OALDomainCredentialsTree other )
{
    struct tempTreeCmp_t arg;
    arg.other = other;
    arg.cmp = TRUE;

    if( self == other )
        return TRUE;

    if( !self || !other )
        return FALSE;

    CredentialsTree_Lock();

    if( !DOFCredentials_IsEqual( self->wildcardCreds, other->wildcardCreds ) )
    {
        CredentialsTree_Unlock();
        return FALSE;
    }

    if( ( OALAVLTree_GetNumItems( self->credentials ) != OALAVLTree_GetNumItems( other->credentials ) ) || 
        ( self->unknownDomainCount != other->unknownDomainCount ) || 
        ( self->knownDomainCount != other->knownDomainCount ) )
    {
        arg.cmp = FALSE;
    }
    else
    {
        OALAVLTree_Traverse( self->credentials, OALDomainCredentialsTree_Traversal_Cmp, &arg );
        OALAVLTree_Traverse( self->trustedDomains, OALDomainCredentialsTree_Traversal_Cmp_TD, &arg );
        OALSList_Traverse( self->unkownDomainCreds, OALDomainCredentialsTree_Traversal_List_Cmp, &arg);
        OALSList_Traverse( self->kownDomainCreds, OALDomainCredentialsTree_Traversal_Known_List_Cmp, &arg);
    }
    CredentialsTree_Unlock();
    return arg.cmp;
}

boolean OALDomainCredentialsTree_HasCredentials( OALDomainCredentialsTree self )
{
    CredentialsTree_Lock();
    if( self && (self->unkownDomainCreds || self->kownDomainCreds || (OALAVLTree_GetNumItems(self->credentials) > 0)) )
    {
        CredentialsTree_Unlock();
        return TRUE;
    }
    CredentialsTree_Unlock();
    return FALSE;
}

struct tdt_t
{
    OALChannel      channel;
    boolean         failed;
};

static boolean OALDomainCredentialsTree_TrustedDomainTraverse(OALAVLTreeItem_t item, void *arg)
{
    DOFObjectIDDomain domain = (DOFObjectIDDomain)item;
    struct tdt_t* ctx = (struct tdt_t*)arg;

    if( !TRPRouter_AddTrustedDomain( domain, ctx->channel ) )
    {
        ctx->failed = TRUE;
        return TRUE; /* to exit traverse early */
    }
    return FALSE;
}

extern boolean OALDomainCredentialsTree_AddTrustedDomainsRoutes( OALDomainCredentialsTree self, OALChannel channel )
{
    struct tdt_t ctx;

    ctx.failed = FALSE;
    ctx.channel = channel;

    OALAVLTree_Traverse( self->trustedDomains, OALDomainCredentialsTree_TrustedDomainTraverse, &ctx);

    return !ctx.failed;
}

typedef struct cvdc_t
{
    OALDomainCredentialsTree tree;
    DOFConnection conn;
    OALDomainCredentialsTree_ValidateDomainCredentials_Callback_fn fn;
    void* context;
    uint32 sentCount;
    uint32 responseCount;
    boolean notifiedCallback;
}cvdc_t;

static void OALDomainCredentialsTree_DOFCredentials_BeginValidateCallback( DOFCredentials self, DOFException exception, void* context )
{
    cvdc_t* ctx = (cvdc_t*)context;
    boolean goingToNotify = FALSE;
    boolean cleanup = FALSE;
    
    CredentialsTree_Lock();

    ++(ctx->responseCount);

    if( exception )
    {
        if( !ctx->notifiedCallback )
        {
            ctx->notifiedCallback = TRUE;
            goingToNotify = TRUE;
        }
    }
    else
    {
        TRPRouter_AddCredentials( self, ctx->conn );
    }

    if( ctx->responseCount == ctx->sentCount )
    {
        if( !ctx->notifiedCallback )
            goingToNotify = TRUE;

        cleanup = TRUE;
    }

    CredentialsTree_Unlock();

    if( goingToNotify )
    {
        if( ctx->notifiedCallback )
            ctx->fn( exception, ctx->context );
        else
            ctx->fn( NULL, ctx->context );
    }

    if( cleanup )
    {
        OALDomainCredentialsTree_Destroy( ctx->tree );
        DOFConnection_Destroy( ctx->conn );
        DOFMem_Free( ctx );
    }
}

static boolean ValidateDomainCredentials( OALAVLTreeItem_t item, void *arg )
{
    cvdc_t* ctx = (cvdc_t*)arg;
    DOFDomain domain = (DOFDomain)item;

    OALChannel channel = DOFConnection_GetChannel( ctx->conn );
    DOFAddress address = ( channel ) ? channel->remoteChannelData->address : NULL;

    if( !DOFCredentials_IsResolved( DOFDomain_GetCredentials( domain ) ) )
    {
        DOFCredentials_Internal_BeginResolve( DOFDomain_GetCredentials( domain ), DOF_GetDof(), OALCREDENTIALSTREE_CONNECTION_VALIDATION_TIMEOUT, TRUE, channel, address, OALDomainCredentialsTree_DOFCredentials_BeginValidateCallback, ctx );
    }
    else
    {
        DOFCredentials_Internal_BeginValidate( DOFDomain_GetCredentials( domain ), DOF_GetDof(), OALCREDENTIALSTREE_CONNECTION_VALIDATION_TIMEOUT, TRUE, channel, address, OALDomainCredentialsTree_DOFCredentials_BeginValidateCallback, ctx );
    }

    return FALSE;
}

static boolean ValidateNonFixedDomainCredentials( OALSListNode item, void *arg )
{
    cvdc_t* ctx = (cvdc_t*)arg;
    OALDomainCredentialsListItem_t* credl = (OALDomainCredentialsListItem_t*)item;

    OALChannel channel = DOFConnection_GetChannel( ctx->conn );
    DOFAddress address = ( channel ) ? channel->remoteChannelData->address : NULL;

    if( !DOFCredentials_IsResolved( credl->domain->creds ) )
    {
        DOFCredentials_Internal_BeginResolve( credl->domain->creds, DOF_GetDof(), OALCREDENTIALSTREE_CONNECTION_VALIDATION_TIMEOUT, TRUE, channel, address, OALDomainCredentialsTree_DOFCredentials_BeginValidateCallback, ctx );
    }
    else
    {
        DOFCredentials_Internal_BeginValidate( credl->domain->creds, DOF_GetDof(), OALCREDENTIALSTREE_CONNECTION_VALIDATION_TIMEOUT, TRUE, channel, address, OALDomainCredentialsTree_DOFCredentials_BeginValidateCallback, ctx );
    }

    return FALSE;
}

/* for connections only */
extern void OALDomainCredentialsTree_ValidateDomainCredentials( OALDomainCredentialsTree self, DOFConnection conn, OALDomainCredentialsTree_ValidateDomainCredentials_Callback_fn fn, void* context )
{
    uint32 sentCount;
    CredentialsTree_Lock();
    PCR_Assert( self && conn && fn );

    sentCount = OALAVLTree_GetNumItems( self->credentials ) + self->unknownDomainCount + self->knownDomainCount;

    /* no credentials to validate */
    if( sentCount == 0 )
    {
        CredentialsTree_Unlock();
        fn( NULL, context );
        return;
    }

    {
        cvdc_t* ctx = (cvdc_t*)DOFMem_Alloc( sizeof( cvdc_t ) );
        if( ctx )
        {
            PCRMem_Set( ctx, 0, sizeof( cvdc_t ) );
            ctx->conn = DOFConnection_Copy( conn );
            if( ctx->conn )
            {
                ctx->tree = OALDomainCredentialsTree_Copy( self );
                if( ctx->tree )
                {
                    ctx->sentCount = sentCount;
                    ctx->responseCount = 0;
                    ctx->fn = fn;
                    ctx->context = context;
                    ctx->notifiedCallback = FALSE;

                    /* Traverse domain credentials. */
                    OALAVLTree_Traverse( self->credentials, ValidateDomainCredentials, ctx );

                    /* Traverse unknown domain credentials. */
                    OALSList_Traverse( self->unkownDomainCreds, ValidateNonFixedDomainCredentials, ctx );

                    /* Traverse the credentials that at one time may have not known their domain */
                    OALSList_Traverse( self->kownDomainCreds, ValidateNonFixedDomainCredentials, ctx );

                    CredentialsTree_Unlock();
                    return;
                }
                DOFConnection_Destroy( ctx->conn );
            }
            DOFMem_Free( ctx );
        }
    }

    CredentialsTree_Unlock();
    fn( DOFEXCEPTION_INSUFFICIENT_RESOURCES, context );
}

