/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/error.h>
#include <dof/credentials.h>
#include <dof/objectid.h>
#include <internal/srp.h>
#include <internal/stream.h>
#include <internal/cipher.h>
#include <internal/marshal.h>
#include <internal/objectidinternal.h>
#include <internal/credentialsinternal.h>

#include "credentials_common.h"
#include "key.h"

/* KEY Credential functions */
static void KEYCredentials_Dtor( OALCredentials self );
static const uint8* KEYCredentials_GetPrivateStorage( OALCredentials self );
static uint32 KEYCredentials_GetPrivateStorageSize( OALCredentials self );
static boolean KEYCredentials_SetPrivateStorage( OALCredentials self, uint32 len, const uint8* data );
static boolean KEYCredentials_Marshal( OALCredentials self, OALStream stream, boolean isRequest );
static uint32 KEYCredentials_MarshalSize( OALCredentials self, boolean isRequest );
static CredentialsUnmarshalReturn KEYCredentials_Unmarshal( OALCredentials self, OALStream stream, boolean isRequest );
static boolean KEYCredentials_IsEqual( OALCredentials self, OALCredentials other);
static const uint8* KEYCredentials_GetKeyFromStorage(OALCredentials self, const uint8* privateStorage);
static const uint8* KEYCredentials_GetKey( OALCredentials self );
static const uint8* KEYCredentials_GetStorageKey( OALCredentials self );
static uint32 KEYCredentials_GetStorageKeySize( OALCredentials self );
static boolean KEYCredentials_IsCached( OALCredentials self );
static void KEYCredentials_ResetCache( OALCredentials self );
static boolean KEYCredentials_Update( OALCredentials self, OALCredentials from );
static boolean KEYCredentials_HasAnotherResolutionStage( OALCredentials self );
static uint32 KEYCredentials_MarshalStore( OALCredentials self, uint32 size, uint8 * buffer );

static const OALCredentials_fns_t KEYCredentialsFNS = { KEYCredentials_Dtor,
                                                        KEYCredentials_GetPrivateStorage,
                                                        KEYCredentials_GetPrivateStorageSize,
                                                        KEYCredentials_SetPrivateStorage,
                                                        KEYCredentials_Marshal,
                                                        KEYCredentials_MarshalSize,
                                                        KEYCredentials_Unmarshal,
                                                        KEYCredentials_IsEqual,
                                                        KEYCredentials_GetKeyFromStorage,
                                                        KEYCredentials_GetKey,
                                                        KEYCredentials_GetStorageKey,
                                                        KEYCredentials_GetStorageKeySize,
                                                        KEYCredentials_IsCached,
                                                        KEYCredentials_ResetCache,
                                                        KEYCredentials_Update,
                                                        KEYCredentials_HasAnotherResolutionStage,
                                                        KEYCredentials_MarshalStore };

DOFCredentials KEYCredentials_InternalCreate( DOFObjectIDDomain domain )
{
    KEYCredentials_t* ret = DOFMem_Alloc( sizeof( KEYCredentials_t ) );
    if( ret )
    {
        PCRMem_Set( ret, 0, sizeof( KEYCredentials_t ) );
        ret->data = (KEYCredentialsData_t*)DOFMem_Alloc( sizeof( KEYCredentialsData_t ) );
        if( ret->data )
        {
            PCRMem_Set( ret->data, 0, sizeof( KEYCredentialsData_t ) );
            ret->base.domain = DOFObjectIDDomain_Copy( (domain) ? domain : DOFOBJECTIDDOMAIN_BROADCAST );
            if(ret->base.domain)
            {
                ret->base.refCount = 1;
                ret->base.type = DOFCREDENTIALSTYPE_KEY;
                ret->base.fns = &KEYCredentialsFNS;
                ret->base.isDomRes = !domain || DOFObjectID_IsWildcard( DOFObjectIDDomain_GetObjectID( ret->base.domain ) );
                ret->base.stage = (ret->base.isDomRes)?1:0;

                return (DOFCredentials)ret;
            }
            DOFMem_Free( ret->data );
        }
        DOFMem_Free( ret );
    }
    return NULL;
}

DOFCredentials DOFCredentialsKey_Internal_Create( const DOFObjectIDDomain domain, const DOFObjectIDAuthentication identity, const uint8 key[32] )
{
    if( domain && identity && key )
    {
        KEYCredentials_t* ret = (KEYCredentials_t*)KEYCredentials_InternalCreate( domain );
        if( ret )
        {
            ret->base.identity = DOFObjectIDAuthentication_Copy( identity );
            if(ret->base.identity)
            {
                PCRMem_Copy( ((KEYCredentials_t*)ret)->key, key, 32 );
                return (DOFCredentials)ret;
            }
            DOFCredentials_Destroy((DOFCredentials)ret);
        }
    }
    return NULL;
}

DOFCredentials DOFCredentialsKey_Create( const DOFObjectIDDomain domain, const DOFObjectIDAuthentication identity, const uint8 key[32] )
{
    KEYCredentials_t* ret = (KEYCredentials_t*)DOFCredentialsKey_Internal_Create(domain,identity,key);

    if( ret )
    {
        OALSecureHMACContext ctx;
        /* Calculate ValidA */
        PCRSecureRandom_Generate( 8, ((uint8*)((KEYCredentials_t*)ret)->validA) );
        OALSecureHMAC_Start( &ctx, ((KEYCredentials_t*)ret)->key );
        OALSecureHMAC_Digest( &ctx, 8, ((uint8*)((KEYCredentials_t*)ret)->validA) );
        OALSecureHMAC_Finish( &ctx, (((KEYCredentials_t*)ret)->validA+8) );
    }
    return (DOFCredentials)ret;
}

static void KEYCredentialsData_Destroy( KEYCredentialsData_t* data )
{
    DOFMem_Free( data );
}

static void KEYCredentials_Dtor( OALCredentials selfx )
{
    KEYCredentials_t* self = (KEYCredentials_t*)selfx;

    if( self->data )
        KEYCredentialsData_Destroy( self->data );
}

static const uint8* KEYCredentials_GetPrivateStorage( OALCredentials self )
{
    if( self->stage == 0 )
        return ((KEYCredentials_t*)self)->key;
    return NULL;
}

static uint32 KEYCredentials_GetPrivateStorageSize( OALCredentials selfx )
{
    return 32; /* spec */
}

static boolean KEYCredentials_SetPrivateStorage( OALCredentials selfx, uint32 len, const uint8* data )
{
    KEYCredentials_t* self = (KEYCredentials_t*)selfx;
    OALSecureHMACContext ctx;
    DOFObjectID oid = DOFObjectIDDomain_GetObjectID( self->base.domain );

    PCRMem_Copy( self->key, data, 32 );
    /* calculate ValidB */
    PCRSecureRandom_Generate( 8, ((uint8*)self->data->validB) );
    OALSecureHMAC_Start( &ctx, self->key );
    OALSecureHMAC_Digest( &ctx, 8, ((uint8*)self->validA) );
    OALSecureHMAC_Digest( &ctx, 8, ((uint8*)self->data->validB) );
    OALSecureHMAC_Digest( &ctx, (oid ? DOFObjectID_GetBytesSize( oid ) : 0), DOFObjectID_GetBytes( oid ) );
    OALSecureHMAC_Finish( &ctx, (self->data->validB+8) );
    return TRUE;
}

static boolean KEYCredentials_Marshal( OALCredentials selfx, OALStream stream, boolean isRequest )
{
    KEYCredentials_t* self = (KEYCredentials_t*)selfx;
    switch( self->base.stage )
    {
    case 0:
        {
            if( OALStream_Write8( stream, DOFCREDENTIALSTYPE_KEY ) &&
                OALStream_Write8( stream, 0 ) &&
                DOFObjectID_Marshal( DOFObjectIDAuthentication_GetObjectID( self->base.identity ), stream ))
                return TRUE;
        }
        break;
    case 1:
        {
            if( OALStream_Write8( stream, DOFCREDENTIALSTYPE_KEY ) &&
                OALStream_Write8( stream, 1 ) )
            {
                if( isRequest )
                {
                    uint32      o_len;
                    DOFObjectID oid   = DOFObjectIDAuthentication_GetObjectID( self->base.identity );

                    o_len = (oid ? DOFObjectID_GetBytesSize( oid ): 0);
                    if( (OALMarshal_StreamCompress3( stream, (o_len + 40) ) == DOFERROR_OK) &&
                         DOFObjectID_Marshal( oid, stream ) &&
                         OALStream_Write( stream, 40, self->validA ) )
                    {
                        return TRUE;
                    }
                }
                else
                {
                    if( (OALMarshal_StreamCompress3( stream, 40 ) == DOFERROR_OK) &&
                         OALStream_Write( stream, 40, self->data->validB ) )
                    {
                        return TRUE;
                    }
                }
            }
        }
        break;
    default:
        break;
    }
    return FALSE;
}

static uint32 KEYCredentials_MarshalSize( OALCredentials selfx, boolean isRequest )
{
    uint32 ret = 0;
    KEYCredentials_t* self = (KEYCredentials_t*)selfx;

    switch( self->base.stage )
    {
    case 0:
        ret = 2 +  DOFObjectID_GetBytesSize( DOFObjectIDAuthentication_GetObjectID( self->base.identity ) );
        break;
    case 1:
        if( isRequest )
        {
            uint32 o_len = DOFObjectID_GetBytesSize( DOFObjectIDAuthentication_GetObjectID( self->base.identity ) );
            ret = 2 + OALMarshal_GetCompressedValueSize( 3, (o_len + 40) ) + o_len + 40;
        }
        else
        {
            ret = 43;
        }
        break;
    default:
        break;
    }

    return ret;
}

static boolean KEYVerifyValidB( KEYCredentials_t* self )
{
    OALSecureHMACContext ctx;
    uint8 validB_temp[32];
    DOFObjectID oid = DOFObjectIDDomain_GetObjectID( self->base.domain );

    OALSecureHMAC_Start( &ctx, self->key );
    OALSecureHMAC_Digest( &ctx, 8, ((uint8*)self->validA) );
    OALSecureHMAC_Digest( &ctx, 8, ((uint8*)self->data->validB) );
    OALSecureHMAC_Digest( &ctx, (oid ? DOFObjectID_GetBytesSize( oid ) : 0), DOFObjectID_GetBytes( oid ));
    OALSecureHMAC_Finish( &ctx, validB_temp );
    if(PCRMem_Compare( self->data->validB+8, validB_temp, 32 ) == 0 )
        return TRUE;
    return FALSE;
}

static CredentialsUnmarshalReturn KEYCredentials_Unmarshal( OALCredentials selfx, OALStream stream, boolean isRequest )
{
    uint8 stage = 0;
    KEYCredentials_t* self = (KEYCredentials_t*)selfx;

    if( OALStream_Read8( stream, &stage ) )
    {
        if( (stage == self->base.stage) || (stage == 0 && self->base.stage == 1 && isRequest) )
        {
            switch( stage )
            {
            case 0:
                self->base.stage = stage;
                if((self->base.identity = DOFObjectIDAuthentication_Unmarshal(stream)) != NULL)
                {
                    return CREDENTIALS_SUCCESSFUL_UNMARSHAL;
                }
                break;
            case 1:
                {
                    uint32 len = 0;
                    if( isRequest )
                    {

                        if( (DOFERROR_OK == OALMarshal_StreamUncompress3( stream, &len )) &&
                            ((self->base.identity = DOFObjectIDAuthentication_Unmarshal(stream)) != NULL) &&
                            OALStream_Read( stream, 40, self->validA ) )
                            return CREDENTIALS_SUCCESSFUL_UNMARSHAL;
                    }
                    else
                    {
                        if( (DOFERROR_OK == OALMarshal_StreamUncompress3( stream, &len )) && (len == 40) &&
                            OALStream_Read( stream, 40, self->data->validB ) &&
                            KEYVerifyValidB( self ) )
                        {
                            --self->base.stage;
                            return CREDENTIALS_SUCCESSFUL_UNMARSHAL;
                        }
                    }
                }
                break;
            default:
                break;
            }
        }
        else
        {
            return CREDENTIALS_WRONGSTAGE_UNMARSHAL;
        }
    }
    return CREDENTIALS_FAILED_UNMARSHAL;
}

static boolean KEYCredentials_IsEqual( OALCredentials selfx, OALCredentials otherx)
{
    KEYCredentials_t* self = (KEYCredentials_t*)selfx;
    KEYCredentials_t* other = (KEYCredentials_t*)otherx;

    if( DOFObjectID_IsEqual(DOFObjectIDAuthentication_GetObjectID( self->base.identity ), DOFObjectIDAuthentication_GetObjectID( other->base.identity ) ) )
    {
        switch( self->base.stage )
        {
        case 0:
            return ( PCRMem_Compare( self->key, other->key, 32 ) == 0 );
            break;
        case 1:
            return ( PCRMem_Compare( self->validA, other->validA, 32 ) == 0 );
            break;
        default:
            break;
        }
    }
    return FALSE;
}

static const uint8* KEYCredentials_GetKeyFromStorage(OALCredentials selfx, const uint8* privateStorage)
{
    return privateStorage;
}

static const uint8* KEYCredentials_GetKey( OALCredentials self )
{
    return ((const uint8*)((KEYCredentials_t*)self)->key);
}

static const uint8* KEYCredentials_GetStorageKey( OALCredentials self )
{
    return DOFObjectID_GetBytes( DOFObjectIDAuthentication_GetObjectID( ((KEYCredentials_t*)self)->base.identity ) );
}

static uint32 KEYCredentials_GetStorageKeySize( OALCredentials self )
{
    return DOFObjectID_GetBytesSize( DOFObjectIDAuthentication_GetObjectID( ((KEYCredentials_t*)self)->base.identity ) );
}

static boolean KEYCredentials_IsCached( OALCredentials self )
{
    /* only cached if it is a domain discovery key credentials that are resolved */
    if( self->stage == 0 && self->isDomRes )
        return TRUE;
    return FALSE;
}

static void KEYCredentials_ResetCache( OALCredentials selfx )
{
    KEYCredentials_t* self = (KEYCredentials_t*)selfx;
    
    if( selfx->stage == 0 && selfx->isDomRes )
    {
        DOFObjectIDDomain tempd = NULL;
        if( self->base.isDomRes )
        {
            tempd = DOFObjectIDDomain_Create( DOFOBJECTID_BROADCAST );
            if( !tempd )
                return;

            DOFObjectIDDomain_Destroy( self->base.domain );
            self->base.domain = tempd;
        }       
        selfx->stage = 1; /* we are not resolved anymore */
        PCRMem_Set( ((void*)self->key), 0, 32 ); /* May not be needed */
    }
}

static boolean KEYCredentials_Update( OALCredentials selfx, OALCredentials otherx )
{
    KEYCredentials_t* self = (KEYCredentials_t*)selfx;
    KEYCredentials_t* other = (KEYCredentials_t*)otherx;

    if( selfx->type == otherx->type &&
        otherx->stage == 0 )
    {
        if( selfx->stage == 0 )
            return TRUE;

        if( !DOFObjectID_IsEqual( DOFObjectIDDomain_GetObjectID( self->base.domain ), DOFObjectIDDomain_GetObjectID( other->base.domain ) ) )
        {
            DOFObjectIDDomain tempd = DOFObjectIDDomain_Copy( other->base.domain );
            if( !tempd )
                return FALSE;

            DOFObjectIDDomain_Destroy( self->base.domain );
            self->base.domain = tempd;
        }
        
        PCRMem_Copy( ((void*)self->key), ((uint8*)other->key), 32 );
        
        self->base.stage = 0;

        DOFCredentials_NotifyObservers( (DOFCredentials)selfx );

        return TRUE;
    }

    return FALSE;
}

static boolean KEYCredentials_HasAnotherResolutionStage( OALCredentials self )
{
    return FALSE;
}

static uint32 KEYCredentials_MarshalStore( OALCredentials self, uint32 size, uint8 * buffer )
{
    uint32 ret = 0;
    OALStream stream = OALStream_CreateRWWrapped( buffer, size );
    if( stream )
    {
        OALStream_SetWriteOffset( stream, 0 );
        /* marshal type */
        if( OALStream_Write8( stream, DOFCREDENTIALSTYPE_KEY ) )
        {
            /* marshal stage 255 */
            if( OALStream_Write8( stream, 255 ) )
            {
                /* marshal domain */
                if( DOFObjectID_Marshal( DOFObjectIDDomain_GetObjectID(self->domain), stream ) )
                {
                    /* marshal identity */
                    if( DOFObjectID_Marshal( DOFObjectIDAuthentication_GetObjectID(self->identity), stream ) )
                    {
                        /* marshal key */
                        if( OALStream_Write( stream,32,((KEYCredentials_t*)self)->key ) )
                        {
                            ret = OALStream_GetWriteOffset(stream);
                        }
                    }
                }
            }
        }
        OALStream_Release( stream, FALSE );
    }
    return ret;
}

