/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/error.h>
#include <dof/credentials.h>
#include <dof/objectid.h>
#include <internal/srp.h>
#include <internal/stream.h>
#include <internal/marshal.h>
#include <internal/objectidinternal.h>
#include <internal/credentialsinternal.h>

#include "credentials_common.h"
#include "password.h"

/* SRP Credential functions */
static void SRPCredentials_Dtor( OALCredentials self );
static const uint8* SRPCredentials_GetPrivateStorage( OALCredentials self );
static uint32 SRPCredentials_GetPrivateStorageSize( OALCredentials self );
static boolean SRPCredentials_SetPrivateStorage( OALCredentials self, uint32 len, const uint8* data );
static boolean SRPCredentials_Marshal( OALCredentials self, OALStream stream, boolean isRequest );
static uint32 SRPCredentials_MarshalSize( OALCredentials self, boolean isRequest );
static CredentialsUnmarshalReturn SRPCredentials_Unmarshal( OALCredentials self, OALStream stream, boolean isRequest );
static boolean SRPCredentials_IsEqual( OALCredentials self, OALCredentials other);
static const uint8* SRPCredentials_GetKeyFromStorage(OALCredentials self, const uint8* privateStorage);
static const uint8* SRPCredentials_GetKey( OALCredentials self );
static const uint8* SRPCredentials_GetStorageKey( OALCredentials self );
static uint32 SRPCredentials_GetStorageKeySize( OALCredentials self );
static boolean SRPCredentials_IsCached( OALCredentials self );
static void SRPCredentials_ResetCache( OALCredentials self );
static boolean SRPCredentials_Update( OALCredentials self, OALCredentials from );
static boolean SRPCredentials_HasAnotherResolutionStage( OALCredentials self );
static uint32 SRPCredentials_MarshalStore( OALCredentials self, uint32 size, uint8 * buffer );

static const OALCredentials_fns_t SRPCredentialsFNS = { SRPCredentials_Dtor,
                                                        SRPCredentials_GetPrivateStorage,
                                                        SRPCredentials_GetPrivateStorageSize,
                                                        SRPCredentials_SetPrivateStorage,
                                                        SRPCredentials_Marshal,
                                                        SRPCredentials_MarshalSize,
                                                        SRPCredentials_Unmarshal,
                                                        SRPCredentials_IsEqual,
                                                        SRPCredentials_GetKeyFromStorage,
                                                        SRPCredentials_GetKey,
                                                        SRPCredentials_GetStorageKey,
                                                        SRPCredentials_GetStorageKeySize,
                                                        SRPCredentials_IsCached,
                                                        SRPCredentials_ResetCache,
                                                        SRPCredentials_Update,
                                                        SRPCredentials_HasAnotherResolutionStage,
                                                        SRPCredentials_MarshalStore };

DOFCredentials DOFCredentialsPassword_Create( const DOFObjectIDDomain domain, const DOFObjectIDAuthentication identity, const char * password )
{
    SRPCredentials_t* ret = NULL;

    if( identity && password )
    {
        ret = (SRPCredentials_t*)SRPCredentials_InternalCreate( domain );
        if( ret )
        {
            ret->base.isDomRes = DOFObjectID_IsBroadcast( DOFObjectIDDomain_GetObjectID( ret->base.domain ) );
            ret->base.identity = DOFObjectIDAuthentication_Copy( identity );
            ret->base.stage = 2;

            /* init SRP data for this credential */
            OALSRP_Calc_Partial_x( DOFObjectIDAuthentication_GetObjectID( identity ), password, ret->partial_x );
            PCRSecureRandom_Generate( 32, ret->data->a );
            OALSRP_Calc_PublicA( ret->data->a, (uint8*)ret->data->pubA );
        }
    }
    return (DOFCredentials)ret;
}

DOFCredentials DOFCredentialsPassword_Internal_Create( const DOFObjectIDDomain domain, const DOFObjectIDAuthentication identity, const uint8* partial_x )
{
    SRPCredentials_t* ret = NULL;

    if( identity && partial_x )
    {
        ret = (SRPCredentials_t*)SRPCredentials_InternalCreate( domain );
        if( ret )
        {
            ret->base.isDomRes = DOFObjectID_IsBroadcast( DOFObjectIDDomain_GetObjectID( ret->base.domain ) );
            ret->base.identity = DOFObjectIDAuthentication_Copy( identity );
            ret->base.stage = 2;

            /* init SRP data for this credential */
            PCRMem_Copy(ret->partial_x, partial_x, 32);
            PCRSecureRandom_Generate( 32, ret->data->a );
            OALSRP_Calc_PublicA( ret->data->a, (uint8*)ret->data->pubA );
        }
    }
    return (DOFCredentials)ret;
}

DOFCredentials SRPCredentials_InternalCreate( DOFObjectIDDomain domain )
{
    SRPCredentials_t* ret = DOFMem_Alloc( sizeof( SRPCredentials_t ) );
    if( ret )
    {
        PCRMem_Set( ret, 0, sizeof( SRPCredentials_t ) );
        ret->data = (SRPCredentialsData_t*)DOFMem_Alloc( sizeof( SRPCredentialsData_t ) );
        if( ret->data )
        {
            PCRMem_Set( ret->data, 0, sizeof( SRPCredentialsData_t ) );
            ret->base.domain = DOFObjectIDDomain_Copy( (domain) ? domain : DOFOBJECTIDDOMAIN_BROADCAST );
            if(ret->base.domain)
            {
                ret->base.type = DOFCREDENTIALSTYPE_PASSWORD;
                ret->base.refCount = 1;
                ret->base.isDomRes = !domain || DOFObjectID_IsWildcard( DOFObjectIDDomain_GetObjectID( ret->base.domain ) );
                ret->base.fns = &SRPCredentialsFNS;
                ret->base.stage = 2;
                return (DOFCredentials)ret;
            }
            DOFMem_Free(ret->data);
        }
        DOFMem_Free(ret);
    }

    return NULL;
}

static void SRPCredentialsData_Destroy( SRPCredentialsData_t* data )
{
    if( data->storage )
        DOFMem_Free( data->storage );
    DOFMem_Free( data );
}

static void SRPCredentials_Dtor( OALCredentials selfx )
{
    SRPCredentials_t* self = (SRPCredentials_t*)selfx;
    if( self->data )
        SRPCredentialsData_Destroy( self->data );
}

static const uint8* SRPCredentials_GetPrivateStorage( OALCredentials selfx )
{
    SRPCredentials_t* self = (SRPCredentials_t*)selfx;
    if( self->data )
    {
        if( !self->data->storage )
        {
            self->data->storage = DOFMem_Alloc( 192 );

            if( self->data->storage )
            {
                PCRSecureRandom_Generate( 32, self->data->storage );
                OALSRP_Calc_v( self->partial_x, self->data->storage, (self->data->storage+32));
                PCRSecureRandom_Generate( 32, (self->data->storage+160));
                PCRMem_Copy( self->key,(self->data->storage+160), 32 );
            }
        }
        return self->data->storage;
    }
    return NULL;
}

static uint32 SRPCredentials_GetPrivateStorageSize( OALCredentials selfx )
{
    return 192; /* as per spec */
}

static boolean SRPCredentials_SetPrivateStorage( OALCredentials selfx, uint32 len, const uint8* data )
{
    SRPCredentials_t* self = (SRPCredentials_t*)selfx;

    PCRMem_Copy( self->data->srps, data, 32 );
    PCRMem_Copy( self->data->v, (data+32), 128 );
    PCRMem_Copy( self->key, (data+160), 32 );

    /* calc Pub B, S and Ksrp */
    OALSRP_CalculateAS( self->data->v, (uint8*)self->data->pubA, (uint8*)self->data->pubB, (uint8*)self->data->srpS, (uint8*)self->data->Ksrp );
    return TRUE;
}

static boolean SRPCredentials_Marshal( OALCredentials selfx, OALStream stream, boolean isRequest )
{
    SRPCredentials_t* self = (SRPCredentials_t*)selfx;
    switch( self->base.stage )
    {
    case 0:
        {
            if( OALStream_Write8( stream, DOFCREDENTIALSTYPE_PASSWORD ) &&
                OALStream_Write8( stream, 0 ) &&
                DOFObjectID_Marshal( DOFObjectIDAuthentication_GetObjectID( self->base.identity ), stream ))
            {
                return TRUE;
            }
        }
        break;
    case 1:
        {
            if( OALStream_Write8( stream, DOFCREDENTIALSTYPE_PASSWORD ) &&
                OALStream_Write8( stream, 1 ) )
            {
                if( isRequest )
                {
                    if( ( OALMarshal_StreamCompress3( stream, 32 ) == DOFERROR_OK ) &&
                        OALStream_Write( stream, 32, (uint8*)self->validA ) )
                    {
                        return TRUE;
                    }
                }
                else
                {
                    if( ( OALMarshal_StreamCompress3( stream, 64 ) == DOFERROR_OK ) &&
                        OALStream_Write( stream, 32, (uint8*)self->data->validB ) &&
                        OALStream_Write( stream, 32, (uint8*)self->data->KEY ) )
                    {
                        --self->base.stage;
                        return TRUE;
                    }
                }
            }
        }
        break;
    case 2:
        {
            if( OALStream_Write8( stream, DOFCREDENTIALSTYPE_PASSWORD ) &&
                OALStream_Write8( stream, 2 ) )
            {
                if( isRequest )
                {
                    if( ( OALMarshal_StreamCompress3( stream, (DOFObjectID_GetBytesSize( DOFObjectIDAuthentication_GetObjectID( self->base.identity ) ) + 128) ) == DOFERROR_OK ) &&
                         DOFObjectID_Marshal( DOFObjectIDAuthentication_GetObjectID( self->base.identity ), stream ) &&
                         OALStream_Write( stream, 128, (uint8*)self->data->pubA ) )
                    {
                        return TRUE;
                    }
                }
                else
                {
                    if( self->base.isDomRes )
                    {
                        if( OALStream_Write8( stream, 0) )
                        {
                            --self->base.stage;
                            return TRUE;
                        }
                    }
                    else
                    {
                        if( ( OALMarshal_StreamCompress3( stream, 160 ) == DOFERROR_OK ) &&
                             OALStream_Write( stream, 128, (uint8*)self->data->pubB ) &&
                             OALStream_Write( stream, 32, self->data->srps ) )
                        {
                            --self->base.stage;
                            return TRUE;
                        }
                    }
                }
            }
        }
        break;
    default:
        break;
    }
    return FALSE;
}

static uint32 SRPCredentials_MarshalSize( OALCredentials selfx, boolean isRequest )
{
    uint32 ret = 0;
    SRPCredentials_t* self = (SRPCredentials_t*)selfx;

    switch( self->base.stage )
    {
    case 0:
        {
            ret = 2 +  DOFObjectID_GetBytesSize( DOFObjectIDAuthentication_GetObjectID( self->base.identity ) );
        }
        break;
    case 1:
        {
            if( isRequest )
                ret = 34 + OALMarshal_GetCompressedValueSize( 3, 32 );
            else
                ret = 64 + 2 + OALMarshal_GetCompressedValueSize( 3, 64 );
        }
        break;
    case 2:
        {
            uint32 oidlen = DOFObjectID_GetBytesSize( DOFObjectIDAuthentication_GetObjectID( self->base.identity ) );
            if( isRequest )
                ret = 130 + oidlen + OALMarshal_GetCompressedValueSize( 3, (oidlen + 128) );
            else
                ret = 2 + ((!self->base.isDomRes) ? (OALMarshal_GetCompressedValueSize( 3, 160 ) + 160) : OALMarshal_GetCompressedValueSize( 3,0 ) );
        }
        break;
    default:
        return FALSE;
        break;
    }
    return ret;
}

static boolean SRP_UnmarshalCalc( SRPCredentials_t* self, boolean isRequest )
{
    switch( self->base.stage )
    {
    case 1:
        {
            if( isRequest )
            {
                OALSRP_CalculateValidBAndKey( (uint8*)self->data->pubA, (uint8*)self->validA, (uint8*)self->data->srpS, (uint8*)self->key, DOFObjectID_GetBytesSize(DOFObjectIDDomain_GetObjectID( self->base.domain )), (uint8*)DOFObjectID_GetBytes(DOFObjectIDDomain_GetObjectID( self->base.domain )), self->data->Ksrp, (uint8*)self->data->validB, (uint8*)self->data->KEY );
            }
            else
            {
                if(!OALSRP_CalculateKShared( (uint8*)self->data->pubA, (uint8*)self->validA, (uint8*)self->data->srpS, (uint8*)self->data->validB, (uint8*)self->data->KEY, (uint8*)self->data->Ksrp, DOFObjectID_GetBytesSize(DOFObjectIDDomain_GetObjectID( self->base.domain )), (uint8*)DOFObjectID_GetBytes(DOFObjectIDDomain_GetObjectID( self->base.domain )), (uint8*)self->key ))
                {
                    return FALSE;
                }
                --self->base.stage;
            }
        }
        break;
    case 2:
        {
            if( !isRequest )
            {
                OALSRP_CalculateNode( (uint8*)self->data->a, (uint8*)self->data->pubA, (uint8*)self->data->pubB, (uint8*)self->partial_x, self->data->srps, (uint8*)self->data->srpS, (uint8*)self->data->Ksrp, (uint8*)self->validA );
                --self->base.stage;
            }
        }
        break;
    default:
        return FALSE;
        break;
    }
    return TRUE;
}

static CredentialsUnmarshalReturn SRPCredentials_Unmarshal( OALCredentials selfx, OALStream stream, boolean isRequest )
{
    uint8 stage = 0;
    SRPCredentials_t* self = (SRPCredentials_t*)selfx;

    if( OALStream_Read8( stream, &stage ) )
    {
        if( (stage == self->base.stage) || ( stage == (self->base.stage-1) && isRequest) || stage == 0 )
        {
            switch( stage )
            {
            case 0:
                {
                    self->base.stage = stage;
                    if ((self->base.identity = DOFObjectIDAuthentication_Unmarshal(stream)) != NULL)
                        return CREDENTIALS_SUCCESSFUL_UNMARSHAL;
                }
                break;
            case 1:
                {
                    if( isRequest )
                    {
                        uint32 len = 0;
                        self->base.stage = stage;
                        if( (DOFERROR_OK == OALMarshal_StreamUncompress3( stream, &len )) && (len == 32) &&
                             OALStream_Read( stream, 32, (uint8*)self->validA ) &&
                             SRP_UnmarshalCalc(self, isRequest) )
                            return CREDENTIALS_SUCCESSFUL_UNMARSHAL;
                    }
                    else
                    {
                        uint32 len = 0;
                        if( (DOFERROR_OK == OALMarshal_StreamUncompress3( stream, &len )) && (len == 64) &&
                             OALStream_Read( stream, 32, (uint8*)self->data->validB ) &&
                             OALStream_Read( stream, 32, (uint8*)self->data->KEY ) &&
                             SRP_UnmarshalCalc(self, isRequest) )
                            return CREDENTIALS_SUCCESSFUL_UNMARSHAL;

                    }
                }
                break;
            case 2:
                {
                    if( isRequest )
                    {
                        uint32 len ;
                        if( (DOFERROR_OK == OALMarshal_StreamUncompress3( stream, &len )) &&
                            ((self->base.identity = DOFObjectIDAuthentication_Unmarshal(stream)) != NULL) &&
                            OALStream_Read( stream, 128, (uint8*)self->data->pubA ) &&
                            SRP_UnmarshalCalc(self, isRequest) )
                            return CREDENTIALS_SUCCESSFUL_UNMARSHAL;
                    }
                    else
                    {
                        uint32 len = 0;
                        if( DOFERROR_OK == OALMarshal_StreamUncompress3( stream, &len ) )
                        {
                             if( len == 0 )
                                 return CREDENTIALS_SUCCESSFUL_UNMARSHAL;

                             if( OALStream_Read( stream, 128, (uint8*)self->data->pubB ) &&
                                 OALStream_Read( stream, 32, self->data->srps ) &&
                                 SRP_UnmarshalCalc(self, isRequest) )
                                 return CREDENTIALS_SUCCESSFUL_UNMARSHAL;
                        }

                    }
                }
                break;
            default:
                break;
            }
        }
        else
        {
            return CREDENTIALS_WRONGSTAGE_UNMARSHAL;
        }
    }
    return CREDENTIALS_FAILED_UNMARSHAL;
}

static boolean SRPCredentials_IsEqual( OALCredentials selfx, OALCredentials otherx)
{
    SRPCredentials_t* self = (SRPCredentials_t*)selfx;
    SRPCredentials_t* other = (SRPCredentials_t*)otherx;

    if( DOFObjectID_IsEqual(DOFObjectIDAuthentication_GetObjectID( self->base.identity ), DOFObjectIDAuthentication_GetObjectID( other->base.identity ) ) )
    {
        switch( self->base.stage )
        {
        case 0:
            return ( PCRMem_Compare( self->key, other->key, 32 ) == 0 );
            break;
        case 1:
            return ( PCRMem_Compare( self->validA, other->validA, 32 ) == 0 );
            break;
        case 2:
            return ( PCRMem_Compare( self->partial_x, other->partial_x, 32 ) == 0 );
            break;
        default:
            break;
        }
    }
    return FALSE;
}

static const uint8* SRPCredentials_GetKeyFromStorage( OALCredentials selfx, const uint8* privateStorage )
{
    return (privateStorage + 160); /* spec */
}

static const uint8* SRPCredentials_GetKey( OALCredentials self )
{
    return ((const uint8*)((SRPCredentials_t*)self)->key);
}

static const uint8* SRPCredentials_GetStorageKey( OALCredentials self )
{
    return DOFObjectID_GetBytes( DOFObjectIDAuthentication_GetObjectID( ((SRPCredentials_t*)self)->base.identity ) );
}

static uint32 SRPCredentials_GetStorageKeySize( OALCredentials self )
{
    return DOFObjectID_GetBytesSize( DOFObjectIDAuthentication_GetObjectID( ((SRPCredentials_t*)self)->base.identity ) );
}

static boolean SRPCredentials_IsCached( OALCredentials self )
{
    if( self->stage == 0 || self->stage == 1 || ( self->isDomRes && !DOFObjectID_IsBroadcast( DOFObjectIDDomain_GetObjectID( self->domain ) ) ) )
        return TRUE; 
    return FALSE;
}

static void SRPCredentials_ResetCache( OALCredentials self )
{
    SRPCredentials_t* selfx = (SRPCredentials_t*)self;

    if( selfx->base.stage == 0 || selfx->base.stage == 1 || ( self->isDomRes && !DOFObjectID_IsBroadcast( DOFObjectIDDomain_GetObjectID( self->domain ) ) ) )
    {
        /* Free storage if created */
        if( selfx->data->storage )
            PCRMem_Free( selfx->data->storage );

        /* memset all data */
        PCRMem_Set( selfx->data, 0, sizeof(SRPCredentialsData_t) );
        PCRMem_Set( selfx->key, 0, 32 );
        PCRMem_Set( selfx->validA, 0, 40 );
        
        /* reset to broadcast domain if discovery */
        if( selfx->base.isDomRes )
        {
            DOFObjectIDDomain_Destroy( selfx->base.domain );
            selfx->base.domain = DOFObjectIDDomain_Create( DOFOBJECTID_BROADCAST );
        }   

        /* new random a */
        PCRSecureRandom_Generate( 32, selfx->data->a );
        /* calculate new Public A */
        OALSRP_Calc_PublicA( selfx->data->a, (uint8*)selfx->data->pubA );

        /* reset to stage 2 */
        selfx->base.stage = 2;
    }
    
}

static boolean SRPCredentials_Update( OALCredentials self, OALCredentials from )
{
    return FALSE;
}

static boolean SRPCredentials_HasAnotherResolutionStage( OALCredentials self )
{
    if( self->stage >= 1 )
        return TRUE;
    return FALSE;
}

static uint32 SRPCredentials_MarshalStore( OALCredentials self, uint32 size, uint8 * buffer )
{
    uint32 ret = 0;
    OALStream stream = OALStream_CreateRWWrapped( buffer, size );
    if( stream )
    {
        OALStream_SetWriteOffset( stream, 0 );
        /* marshal type */
        if( OALStream_Write8( stream, DOFCREDENTIALSTYPE_PASSWORD ) )
        {
            /* marshal stage 255 */
            if( OALStream_Write8( stream, 255 ) )
            {
                /* marshal domain */
                if( DOFObjectID_Marshal( DOFObjectIDDomain_GetObjectID( self->domain ), stream ) )
                {
                    /* marshal identity */
                    if( DOFObjectID_Marshal( DOFObjectIDAuthentication_GetObjectID( self->identity ), stream ) )
                    {
                        /* marshal key */
                        if( OALStream_Write( stream, 32, ((SRPCredentials_t*)self)->partial_x ) )
                        {
                            ret = OALStream_GetWriteOffset(stream);
                        }
                    }
                }
            }
        }
        OALStream_Release( stream, FALSE );
    }
    return ret;
}

