/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/objectid.h>
#include <dof/credentials.h>

#include <internal/stream.h>
#include <internal/objectidinternal.h>
#include <internal/credentialsinternal.h>

#include "credentials_common.h"
#include "remotecredentials.h"
#include "key.h"

static void RemoteCredentials_Dtor( OALCredentials self );
static const uint8* RemoteCredentials_GetPrivateStorage( OALCredentials self );
static uint32 RemoteCredentials_GetPrivateStorageSize( OALCredentials self );
static boolean RemoteCredentials_SetPrivateStorage( OALCredentials self, uint32 len, const uint8* data );
static boolean RemoteCredentials_Marshal( OALCredentials self, OALStream stream, boolean isRequest );
static uint32 RemoteCredentials_MarshalSize( OALCredentials self, boolean isRequest );
static CredentialsUnmarshalReturn RemoteCredentials_Unmarshal( OALCredentials self, OALStream stream, boolean isRequest );
static boolean RemoteCredentials_IsEqual( OALCredentials self, OALCredentials other);
static const uint8* RemoteCredentials_GetKeyFromStorage(OALCredentials self, const uint8* privateStorage);
static const uint8* RemoteCredentials_GetKey( OALCredentials self );
static const uint8* RemoteCredentials_GetStorageKey( OALCredentials self );
static uint32 RemoteCredentials_GetStorageKeySize( OALCredentials self );
static boolean RemoteCredentials_IsCached( OALCredentials self );
static void RemoteCredentials_ResetCache( OALCredentials self );
static boolean RemoteCredentials_Update( OALCredentials self, OALCredentials from );
static boolean RemoteCredentials_HasAnotherResolutionStage( OALCredentials self );
static uint32 RemoteCredentials_MarshalStore( OALCredentials self, uint32 size, uint8 * buffer );

static const OALCredentials_fns_t RemoteCredentialsFNS = { RemoteCredentials_Dtor,
                                                           RemoteCredentials_GetPrivateStorage,
                                                           RemoteCredentials_GetPrivateStorageSize,
                                                           RemoteCredentials_SetPrivateStorage,
                                                           RemoteCredentials_Marshal,
                                                           RemoteCredentials_MarshalSize,
                                                           RemoteCredentials_Unmarshal,
                                                           RemoteCredentials_IsEqual,
                                                           RemoteCredentials_GetKeyFromStorage,
                                                           RemoteCredentials_GetKey,
                                                           RemoteCredentials_GetStorageKey,
                                                           RemoteCredentials_GetStorageKeySize,
                                                           RemoteCredentials_IsCached,
                                                           RemoteCredentials_ResetCache,
                                                           RemoteCredentials_Update,
                                                           RemoteCredentials_HasAnotherResolutionStage,
                                                           RemoteCredentials_MarshalStore };



DOFCredentials DOFCredentials_Create( DOFCredentials creds, DOFObjectIDDomain domainID )
{
    DOFObjectIDDomain tempd = DOFObjectIDDomain_Copy( domainID );
    if( tempd )
    {
        DOFObjectIDAuthentication tempi = DOFObjectIDAuthentication_Copy( DOFCredentials_GetIdentity( creds ) );
        if( tempi )
        {
            DOFCredentials tempc = DOFCredentials_Copy( creds );
            if( tempc )
            {
                RemoteCredentials_t* ret = (RemoteCredentials_t*)DOFMem_Alloc(sizeof(RemoteCredentials_t));
                if( ret )
                {
                    PCRMem_Set( ret, 0, sizeof( RemoteCredentials_t ) );
                    ret->base.fns = &RemoteCredentialsFNS;
                    ret->base.refCount = 1;
                    ret->base.type = DOFCREDENTIALSTYPE_REMOTEKEY;
                    ret->base.domain = tempd;
                    ret->base.identity = tempi;
                    ret->creds = tempc;
                    ret->base.stage = 1; /* unresolved */
                    ret->base.isDomRes = DOFObjectID_IsBroadcast( DOFObjectIDDomain_GetObjectID( tempd ) );

                    return (DOFCredentials)ret;
                }
                DOFCredentials_Destroy( tempc );
            }
            DOFObjectIDAuthentication_Destroy( tempi );
        }
        DOFObjectIDDomain_Destroy( tempd );
    }
    return NULL;
}

static void RemoteCredentials_Dtor( OALCredentials self )
{
    DOFCredentials_Destroy( ((RemoteCredentials_t*)self)->creds );
}

static const uint8* RemoteCredentials_GetPrivateStorage( OALCredentials self )
{
    return NULL;
}

static uint32 RemoteCredentials_GetPrivateStorageSize( OALCredentials self )
{
    return 0;
}

static boolean RemoteCredentials_SetPrivateStorage( OALCredentials self, uint32 len, const uint8* data )
{
    return FALSE;
}

static boolean RemoteCredentials_Marshal( OALCredentials self, OALStream stream, boolean isRequest )
{
    PCR_Assert( self->stage == 0 );
    if( OALStream_Write8( stream, DOFCREDENTIALSTYPE_KEY ) &&
        OALStream_Write8( stream, 0 ) &&
        DOFObjectID_Marshal( DOFObjectIDAuthentication_GetObjectID( self->identity ), stream ))
        return TRUE;
    return FALSE;
}

static uint32 RemoteCredentials_MarshalSize( OALCredentials self, boolean isRequest )
{
    PCR_Assert( self->stage == 0 );
    return (2 +  DOFObjectID_GetBytesSize( DOFObjectIDAuthentication_GetObjectID( self->identity ) ));
}

static CredentialsUnmarshalReturn RemoteCredentials_Unmarshal( OALCredentials self, OALStream stream, boolean isRequest )
{
    return CREDENTIALS_FAILED_UNMARSHAL;
}

static boolean RemoteCredentials_IsEqual( OALCredentials selfx, OALCredentials otherx)
{
    RemoteCredentials_t* self = (RemoteCredentials_t*)selfx;
    RemoteCredentials_t* other = (RemoteCredentials_t*)otherx;

    if( self->base.stage == 0 && other->base.stage == 0 ) /* resolved key should match */
    {
        if( PCRMem_Compare( self->key, other->key, 32 ) != 0 )
            return FALSE;
    }
    /* Identity is based on the base credential so this works for the identity comparison */
    return DOFCredentials_IsEqual( self->creds, other->creds );
}

static const uint8* RemoteCredentials_GetKeyFromStorage(OALCredentials self, const uint8* privateStorage)
{
    return privateStorage;
}

static const uint8* RemoteCredentials_GetKey( OALCredentials self )
{
    if( self->stage == 0 )
        return ((const uint8*)((RemoteCredentials_t*)self)->key);
    return NULL;
}

static const uint8* RemoteCredentials_GetStorageKey( OALCredentials self )
{
    return NULL;
}

static uint32 RemoteCredentials_GetStorageKeySize( OALCredentials self )
{
    return 0;
}

static boolean RemoteCredentials_IsCached( OALCredentials selfx )
{
    if( selfx->stage == 0 )
        return TRUE; /* these are cached if they are resolved */
    return FALSE;
}

static void RemoteCredentials_ResetCache( OALCredentials selfx )
{
    RemoteCredentials_t* self = (RemoteCredentials_t*)selfx;

    if( selfx->stage == 0 ) /* Unresolved nothing to reset */
    {
        DOFObjectIDDomain origD = NULL;
        DOFObjectIDAuthentication origI = NULL;

        if( self->base.isDomRes )
        {
            origD = self->base.domain;
            self->base.domain = DOFObjectIDDomain_Create( DOFOBJECTID_BROADCAST );
            if( !self->base.domain )
            {
                self->base.domain = origD;
                return;
            }
        }

        if( !DOFObjectID_IsEqual( DOFObjectIDAuthentication_GetObjectID( self->base.identity ), DOFObjectIDAuthentication_GetObjectID( DOFCredentials_GetIdentity( self->creds ) ) ) )
        {
            origI = self->base.identity;
            self->base.identity = DOFObjectIDAuthentication_Copy( DOFCredentials_GetIdentity( self->creds ) );
            if( !self->base.identity )
            {
                self->base.identity = origI;
                if( origD )
                {
                    DOFObjectIDDomain_Destroy( self->base.domain );
                    self->base.domain = origD;
                }
                return;
            }            
        }

        PCRMem_Set( ((void*)self->key), 0, 32 ); /* May not be needed */

        selfx->stage = 1; /* we are not resolved anymore */

        DOFCredentials_NotifyObservers( (DOFCredentials)selfx ); /* notify observer of changes */

        DOFObjectIDDomain_Destroy( origD );
        DOFObjectIDAuthentication_Destroy( origI );
    }
}

static boolean RemoteCredentials_Update( OALCredentials selfx, OALCredentials otherx )
{
    RemoteCredentials_t* self = (RemoteCredentials_t*)selfx;

    PCR_Assert( otherx->type == DOFCREDENTIALSTYPE_KEY );

    if( otherx->stage == 0 )
    {
        if( selfx->stage != 0 )
        {
            DOFObjectIDDomain origDomain = NULL;

            if( !DOFObjectID_IsEqual( DOFObjectIDDomain_GetObjectID( self->base.domain ), DOFObjectIDDomain_GetObjectID( DOFCredentials_GetDomainID( (DOFCredentials)otherx ) ) ) )
            {
                DOFObjectIDDomain tempd = DOFObjectIDDomain_Copy( DOFCredentials_GetDomainID( (DOFCredentials)otherx ) );
                if( !tempd )
                    return FALSE;

                origDomain = self->base.domain;
                self->base.domain = tempd;
            }

            if( !DOFObjectID_IsEqual( DOFObjectIDAuthentication_GetObjectID( self->base.identity ), DOFObjectIDAuthentication_GetObjectID( DOFCredentials_GetIdentity( (DOFCredentials)otherx ) ) ) )
            {
                DOFObjectIDAuthentication tempA = DOFObjectIDAuthentication_Copy( DOFCredentials_GetIdentity( (DOFCredentials)otherx ) );
                if( !tempA )
                {
                    DOFObjectIDDomain_Destroy( self->base.domain );
                    self->base.domain = origDomain;
                    return FALSE;
                }

                DOFObjectIDAuthentication_Destroy( self->base.identity);
                self->base.identity = tempA;
            }
            
            PCRMem_Copy( ((uint8*)self->key), DOFCredentials_GetKey( (DOFCredentials)otherx ), 32 );
            PCRMem_Copy( self->validA, ( (otherx->type == DOFCREDENTIALSTYPE_REMOTEKEY)?((RemoteCredentials_t*)otherx)->validA : ((KEYCredentials_t*)otherx)->validA ), 40 );
            self->base.stage = 0;

            DOFCredentials_NotifyObservers( (DOFCredentials)selfx ); /* notify observer of changes */
            DOFObjectIDDomain_Destroy( origDomain );
            return TRUE;
        }
        else
        {
            /* no need to update */
            return TRUE;
        }
    }
    return FALSE;
}

static boolean RemoteCredentials_HasAnotherResolutionStage( OALCredentials self )
{
    return FALSE;
}

static uint32 RemoteCredentials_MarshalStore( OALCredentials self, uint32 size, uint8 * buffer )
{
    return 0;
}

