/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/** @addtogroup DOF
@{
**/

/* Include external and internal oal headers. */
#include <dof/cfg.h>
#include <dof/dof.h>
#include <dof/domain.h>
#include <dof/interface.h>
#include <dof/oaltypes.h>
#include <dof/objectid.h>
#include <dof/server.h>
#include <dof/system.h>
#include <internal/bridgeinternal.h>
#include <internal/dofinternal.h>
#include <internal/enqueue.h>
#include <internal/exception.h>
#include <internal/connectionless_protocolstack.h>
#include <internal/connectionmanager.h>
#include <internal/filter_internal.h>
#include <internal/oap_v1.h>
#include <internal/credentialsinternal.h>
#include <internal/interfaceidinternal.h>
#include <internal/securegroupinternal.h>
#include <internal/srp.h>
#include <internal/connection_internal.h>
#include <internal/dofinternal.h>
#include <internal/objectinternal.h>
#include <internal/operationinternal.h>
#include <internal/operationconfiginternal.h>
#include <internal/remotechannelmanager.h>
#include <internal/threadpool.h>
#include <internal/transportinternal.h>
#include <internal/transportmanager.h>
#include <internal/router.h>
#include <internal/securityscope.h>
#include <internal/systeminternal.h>
#include <internal/tep_v1.h>
#include <internal/trp_v1.h>
#include <internal/sgmp_v1.h>
#include <internal/ccmmode_v1.h>
#include <internal/resource_pool.h>
#include <internal/epp_negative_one.h>
#include <internal/serverinternal.h>
#include <internal/domaininternal.h>
#include <internal/transportinternal.h>
#include <internal/runtimeinternal.h>

#ifndef DOFRESOURCE_NUM_MUTEXES
#  define DOFRESOURCE_NUM_MUTEXES 0 /* Dynamic */
#endif

#ifndef DOF_SYSTEM_CLEANUP_GRACE_PERIOD
#  define DOF_SYSTEM_CLEANUP_GRACE_PERIOD 5000
#endif

OALResourcePool         mutex_pool;
PCRMutex                ref_count_mutex;
PCRMutex                addr_mutex;
PCRMutex                connection_server_mutex;
PCRMutex                connection_server_count_mutex;

static PCRMutex oal_mutex = NULL;

static PCRMutex dof_mutex = NULL;
static boolean  dof_init  = FALSE;
static DOF      gDOF      = NULL;


static DOFAuthenticator DOF_CreateAuthenticatorInternal( DOF dof, DOFAuthenticatorConfig domainConfig);


void DOF_Lock(void)
{
    PCRMutex_Lock(dof_mutex);
}

void DOF_Unlock(void)
{
    PCRMutex_Unlock(dof_mutex);
}

/* DOFMem */
#ifndef DOFMem_Free
void DOFMem_Free( void * mem ) { PCRMem_Free( mem ); }
#endif
#ifndef DOFMem_Alloc
void * DOFMem_Alloc  ( uint32 size ) { return PCRMem_Alloc( size ); }
#endif
#ifndef DOFMem_Realloc
void * DOFMem_Realloc( void * mem, uint32 size ) { return PCRMem_Realloc( mem, size ); }
#endif

void init_mutex_pool(void);
void destroy_mutex_pool(void);
extern boolean DOFSystem_QueryInit(void);
static boolean DOF_CreateInternal( DOF dof, DOFConfig config );
static boolean DOF_DestroyAuthenticatorsInternal( DOF dof );

void DOF_Init( void )
{
    if( !dof_init )
    {
        boolean isFail = FALSE;
        /*  Most of these functions have no error checking/logging, but failure to succeed at
         *  DOF_Init is a catastrophic failure and should be treated as such. Right now
         *  most of these calls silently fail.
         */
        PCR_Init();

        oal_mutex  = PCRMutex_Create(); /* Protect library init/creates */
        PCR_Assert(oal_mutex);

        init_mutex_pool();
        DOFLogListener_Init( );
        isFail |= !DOFOperation_Init();
        isFail |= !OALOperationManager_Init2();
        isFail |= !DOFObjectID_Init();
        isFail |= !DOFInterfaceID_Init();
        isFail |= !DOFDomain_Init1();
        DOFException_Init();
        DOFCredentials_Init();
        isFail |= !TRPRouter_MutexInit();
        isFail |= !DOFType_Init();
        isFail |= !OALSRP_Init();
        isFail |= !TRP_V1_Init();
        PCR_Assert(!isFail);
        dof_init = TRUE;
    }
}

typedef struct DOFDomainListenerRemoveNode_t* DOFDomainListenerRemoveNode;
typedef struct DOFDomainListenerNode_t* DOFDomainListenerNode;

typedef struct DOFDomainListenerRemoveNode_t
{
    OALEnqueueItem_t        base;
    DOFDomainListener       listener;
    DOFException            exception;
    DOFDomainListenerNode   node;
}DOFDomainListenerRemoveNode_t;

typedef struct DOFDomainListenerNode_t
{
    OALSListNode                    next;
    DOFDomainListenerRemoveNode_t   self;
}DOFDomainListenerNode_t;

static boolean CompareGreaterThan_DOFDomainListeners( OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context )
{
    return ( ((DOFDomainListener)key) > ((DOFDomainListenerNode)item)->self.listener );
}

static boolean CompareEqual_DOFDomainListeners( OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context )
{
    return ( ((DOFDomainListener)key) == ((DOFDomainListenerNode)item)->self.listener );
}

DOF DOF_Create( DOFConfig config )
{
    PCRMutex_Lock(oal_mutex);
    if( !gDOF )
    {
        gDOF = DOFMem_Alloc(sizeof(DOF_t));
        if(gDOF)
        {
            PCRMem_Set(gDOF,0,sizeof(DOF_t));
            dof_mutex = PCRMutex_Create();
            if(dof_mutex)
            {
                gDOF->runtime_mutex = PCRMutex_Create();
                if(gDOF->runtime_mutex)
                {
                    if( config )
                    {
                        gDOF->config = DOFConfig_Copy( config );
                    }
                    else
                    {
                        gDOF->config = DOFConfigBuilder_BuildAndDestroy( DOFConfigBuilder_Create() );
                    }

                    if( gDOF->config )
                    {
                        /* handle dof create */
                        if( DOF_CreateInternal( gDOF, gDOF->config ))
                        {
                            gDOF->threadPool = OALThreadPool_Create( DOFConfig_GetThreadPoolSize( gDOF->config ) );
                            if ( gDOF->threadPool )
                            {
                                if( DOFCredentials_InitDOFLevel( gDOF ) )
                                {
                                    gDOF->domainListeners = OALAVLTree_Create( CompareGreaterThan_DOFDomainListeners, CompareEqual_DOFDomainListeners, NULL );
                                    if( gDOF->domainListeners )
                                    {
                                        if( DOF_InterestListeners_Init( gDOF ) )
                                        {
                                            if( DOFDomain_Init2() )
                                            {
                                                if(DOFOperationBridgeConfig_Init())
                                                {
                                                    gDOF->defaultBridge = OALOperationBridge_CreateNonSecureBridge(gDOF, NULL);
                                                    if(gDOF->defaultBridge != NULL)
                                                    {
                                                        gDOF->enqueue = OALEnqueue_Create(gDOF->threadPool);
                                                        if ( gDOF->enqueue )
                                                        {
                                                            gDOF->systemsGoneEvent = PCREvent_Create(TRUE);
                                                            if ( gDOF->systemsGoneEvent )
                                                            {
                                                                PCREvent_Signal( gDOF->systemsGoneEvent );
                                                                if(DOFSecureGroups_Init())
                                                                {
                                                                    if (OAP_V1_Init())
                                                                    {
                                                                        if((gDOF->serverManager = DOFServerManager_Create( gDOF )))
                                                                        {
                                                                            if(EPP_NegOne_Init())
                                                                            {
                                                                                if(DOFOperation_DOFInit())
                                                                                {
                                                                                    if(DOFRuntime_Init( gDOF ))
                                                                                    {
                                                                                        if(DOFSystem_QueryInit())
                                                                                        {
                                                                                            if(DOFSystem_Init( gDOF ))
                                                                                            {
                                                                                                if(OALRemoteChannelMgr_Init())
                                                                                                {
                                                                                                    if(OALConnectionlessProtocolStack_Init())
                                                                                                    {
                                                                                                        if(DOFConnection_Init())
                                                                                                        {
                                                                                                            if(OALConnectionMgr_Init())
                                                                                                            {
                                                                                                                if(OALOperationManager_Init())
                                                                                                                {
                                                                                                                    if(DOFRouter_Init())
                                                                                                                    {
                                                                                                                        if(TEP_V1_Init())
                                                                                                                        {
                                                                                                                            if(TRP_V1_SM_Create())
                                                                                                                            {
                                                                                                                                if(SGMP_V1_SM_Init())
                                                                                                                                {
                                                                                                                                    if(CCMMODE_V1_SM_Init())
                                                                                                                                    {
                                                                                                                                        if(DOFTransportLayer_Init())
                                                                                                                                        {
                                                                                                                                            /* old DOF_Init end */
                                                                                                                                            DOFOperation_SetLocalOID( DOFConfig_GetSourceID( gDOF->config ) );

                                                                                                                                            PCRMutex_Unlock(oal_mutex);
                                                                                                                                            return gDOF;
                                                                                                                                        }
                                                                                                                                        CCMMODE_V1_SM_Shutdown();
                                                                                                                                    }
                                                                                                                                    SGMP_V1_SM_Shutdown();
                                                                                                                                }
                                                                                                                                TRP_V1_Shutdown1();
                                                                                                                                TRP_V1_SM_Destroy();
                                                                                                                            }
                                                                                                                            TEP_V1_Shutdown1();
                                                                                                                            TEP_V1_Shutdown2();
                                                                                                                        }
                                                                                                                        DOFRouter_Cleanup();
                                                                                                                    }
                                                                                                                    OALOperationManager_CancelAll();
                                                                                                                    OALOperationManager_Shutdown();
                                                                                                                }
                                                                                                                OALConnectionMgr_CloseInboundConnections();
                                                                                                                OALConnectionMgr_Shutdown();
                                                                                                            }
                                                                                                            DOFConnection_Shutdown();
                                                                                                        }
                                                                                                        OALConnectionlessProtocolStack_Shutdown();
                                                                                                    }
                                                                                                    OALRemoteChannelMgr_Shutdown();
                                                                                                }
                                                                                                DOFSystem_Shutdown(gDOF);
                                                                                            }
                                                                                            DOFSystem_QueryCleanup();
                                                                                        }
                                                                                        DOFRuntime_Shutdown();
                                                                                    }
                                                                                    DOFOperation_DOFShutdown();
                                                                                }
                                                                                EPP_NegOne_Shutdown();
                                                                            }
                                                                            DOFServerManager_Destroy(gDOF->serverManager);
                                                                        }
                                                                        OAP_V1_Shutdown();
                                                                    }
                                                                }
                                                                PCREvent_Destroy( gDOF->systemsGoneEvent );
                                                            }
                                                            OALEnqueue_Destroy( gDOF->enqueue );
                                                        }
                                                        OALOperationBridge_Destroy(gDOF->defaultBridge);
                                                    }
                                                    DOFOperationBridgeConfig_Shutdown();
                                                }
                                                DOFDomain_Shutdown2();
                                            }
                                            DOF_InterestListeners_Shutdown( gDOF );
                                        }
                                        OALAVLTree_Destroy( gDOF->domainListeners, NULL, NULL );
                                    }
                                    DOFCredentials_ShutdownDOFLevel();
                                }
                                OALThreadPool_Destroy( gDOF->threadPool );
                            }
                            DOF_DestroyAuthenticatorsInternal( gDOF );
                            DOFOperationFilter_Destroy( gDOF->defaultFilter );
                            OALOperationBridge_Destroy( gDOF->defaultBridge );
                            DOFSecurityScope_Destroy( gDOF->defaultSecurityScope );
                            DOFSecurityScope_Destroy( gDOF->emptySecurityScope );
                        }
                        DOFConfig_Destroy(gDOF->config);
                    }
                    PCRMutex_Destroy(gDOF->runtime_mutex);
                }
                PCRMutex_Destroy( dof_mutex );
                dof_mutex = NULL;
            }
            DOFMem_Free(gDOF);
            gDOF = NULL;
        }
    }
    PCRMutex_Unlock(oal_mutex);
    return NULL;
}

static void DOF_RemoveServersNodes( DOF dof );
static void DOF_RemoveAllDomainListeners( DOF self );

/*  When a structure that contains an enqueue is destroyed it must not be destroyed from the
    worker thread that belongs to the enqueue.  To prevent the error case use this function to
    guarantee that a thread other than the worker thread will call OALEnqueue_Destroy.  */
void DOF_DestroyEnqueue(OALThreadPoolItem dnode, boolean shutdown)
{
    DOFEnqueueContext_t * node = (DOFEnqueueContext_t*) dnode;

    PCR_Assert(node != NULL);

    OALEnqueue_Destroy(node->enqueue);
    DOFMem_Free(dnode);
}

void DOF_Destroy( DOF dof )
{
    if( !dof || dof != gDOF )
        return;

    TRP_V1_Shutdown1(); /* Stop trp from adding new operations */

    DOF_RemoveAllDomainListeners( dof );

    OALConnectionMgr_CloseInboundConnections();
    TEP_V1_Shutdown1();             /* TEP keeps state that Operation Manager won't
                                       know about.  Thus, it must be shut down before
                                       Operation Manager.  This will empty the data
                                       structures, but keep the mutex around just in
                                       case an operation is being fed through cancel. */
    OALOperationManager_CancelAll();
    DOF_InterestListeners_Shutdown( gDOF );
    DOFSystem_Shutdown( dof );
    OALEnqueue_Empty( dof->enqueue ); /* Finish system cleanup before continuing. */
    OALRemoteChannelMgr_Shutdown();
    OALConnectionlessProtocolStack_Shutdown();
    DOFConnection_Shutdown();
    OALConnectionMgr_Shutdown();
    DOFServerManager_Destroy( dof->serverManager );
    DOF_RemoveServersNodes(dof); /* frees the DOFServer nodes from the DOF. DOF doesn't care about them any more */
    DOFTransportLayer_Shutdown();
    OAP_V1_Shutdown();
    SGMP_V1_SM_Shutdown();
    CCMMODE_V1_SM_Shutdown();
    DOFRouter_Cleanup();
    EPP_NegOne_Shutdown();
    DOFOperation_DOFShutdown();
    TEP_V1_Shutdown2();             /* Clean up last bit of state */
    DOFOperationBridgeConfig_Shutdown();
    TRP_V1_SM_Destroy();
    DOF_DestroyAuthenticatorsInternal( dof );
    DOFCredentials_ShutdownDOFLevel();
    OALEnqueue_Destroy(dof->enqueue);
    DOFDomain_Shutdown2();
    OALOperationManager_Shutdown();
    DOFRuntime_Shutdown(); /* this needs to be after the enqueue destroy */
    PCRMutex_Destroy(dof->runtime_mutex);
    OALThreadPool_Destroy(dof->threadPool);
    PCREvent_Destroy( dof->systemsGoneEvent );
    DOFConfig_Destroy( dof->config );

    /* These cannot be free'd until systems and operations have been shut down */
    DOFOperationFilter_Destroy(dof->defaultFilter);
    OALOperationBridge_Destroy(dof->defaultBridge );
    DOFSecurityScope_Destroy( dof->defaultSecurityScope );
    DOFSecurityScope_Destroy(dof->emptySecurityScope);

    DOFMem_Free( dof );
    gDOF = NULL;
    PCRMutex_Destroy( dof_mutex );
    dof_mutex = NULL;
}

void DOF_Shutdown( void )
{
    if( dof_init )
    {
        DOFDomain_Shutdown1();
        TRPRouter_MutexShutdown();
        DOFCredentials_Shutdown();
        TRP_V1_Shutdown();
        OALSRP_Shutdown();
        DOFException_Shutdown();
        DOFInterfaceID_Shutdown();
        DOFObjectID_Shutdown();
        OALOperationManager_Shutdown2();
        DOFOperation_Shutdown();
        DOFLogListener_Shutdown();
        DOFType_Shutdown();
        destroy_mutex_pool();
        PCRMutex_Destroy(oal_mutex);
        oal_mutex = NULL;
        PCR_Shutdown();
        dof_init = FALSE;
    }
}

/**
 * @internal
 *
 * used to get the one dof until we change to handle multiple dofs
 * @note This call is to be remove after every class has an ptr
 *       to a dof if needed.
 *
 * @return DOF The Systems DOF
 */
DOF DOF_GetDof()
{
    return gDOF;
}

DOFConfig DOF_GetConfig( const DOF dof )
{
    return dof->config;
}

DOFState DOF_CreateState( const DOF dof )
{
    return DOFState_Create( dof );
}

static boolean findAuthenticatorByID(OALSListNode itemx, void *arg)
{
    DOFAuthenticatorListItem_t* item = (DOFAuthenticatorListItem_t*)itemx;
    DOFObjectID domainID = (DOFObjectID)arg;

    return (DOFObjectID_Compare( DOFObjectIDDomain_GetObjectID( DOFAuthenticator_GetDomainID(item->authenticator )), domainID ) == 0 );
}

DOFAuthenticator DOF_GetAuthenticator( const DOF self, DOFObjectIDDomain domainID )
{
    DOFAuthenticatorListItem_t* ret;

    if( self )
    {
        DOF_Lock();
        ret = (DOFAuthenticatorListItem_t*)OALSList_Retrieve( self->authenticators, findAuthenticatorByID, (void*)DOFObjectIDDomain_GetObjectID( domainID ) );
        DOF_Unlock();
        if( ret )
            return ret->authenticator;
    }
    return NULL;
}

typedef struct partitionContext_t
{
    PCREvent        event;
    DOFPartition    partition;
}partitionContext_t;

/* caller must retain DOFAuthenticator if wanted */
static DOFAuthenticator DOF_CreateAuthenticatorInternal( DOF dof, DOFAuthenticatorConfig authenticatorConfig)
{
    DOFAuthenticatorListItem_t* dml = (DOFAuthenticatorListItem_t*)DOFMem_Alloc( sizeof(DOFAuthenticatorListItem_t) );
    if ( dml )
    {
        dml->authenticator = DOFAuthenticator_CreateInternal(dof, authenticatorConfig);
        if ( dml->authenticator )
        {
            if( TRPRouter_AddAuthenticator( dml->authenticator ) ) /* this will limit us to just one domain entry per domainID*/
            {
                dml->next = NULL;
                DOF_Lock();
                OALSList_Insert( &dof->authenticators, (OALSListNode)dml );
                ++dof->authenticatorsCnt;
                DOF_Unlock();
                return dml->authenticator;
            }
            DOFAuthenticator_Destroy( dml->authenticator );
        }
        DOFMem_Free( dml );
    }
    return NULL;
}

static void removeAuthenticators( OALSListNode itemx, void *arg )
{
    DOFAuthenticatorListItem_t* item = (DOFAuthenticatorListItem_t*)itemx;

    --(DOF_GetDof()->authenticatorsCnt);
    DOFAuthenticator_DOFTriggeredDestroy( item->authenticator );
    DOFMem_Free( item );
}

static boolean DOF_DestroyAuthenticatorsInternal( DOF dof )
{
    DOF_Lock();
    OALSList_RemoveAll( &dof->authenticators, removeAuthenticators, NULL );
    DOF_Unlock();
    return TRUE;
}

static boolean removeAuthenticator( OALSListNode itemx, void *arg )
{
    DOFAuthenticatorListItem_t* item = (DOFAuthenticatorListItem_t*)itemx;

    if( item->authenticator == arg )
    {
        return TRUE;
    }
    return FALSE;
}

boolean DOF_RemoveAuthenticator( DOF dof, DOFAuthenticator auth )
{
    DOFAuthenticatorListItem_t* item = NULL;
    DOF_Lock();
    item = (DOFAuthenticatorListItem_t*)OALSList_FindAndRemove( &dof->authenticators, NULL, removeAuthenticator, auth );
    if( item )
    {
        --(dof->authenticatorsCnt);
        DOFMem_Free( item );
    }
    DOF_Unlock();
    return TRUE;
}

GUID DOF_GenerateGUID(void)
{
    GUID guid;

    PCRRandom_Generate( sizeof(guid), guid.bytes );
    guid.bytes[8] &= 0x3F; /* guid specification dictates that 2 bits of byte 8 and 4 bits of byte 6 be fixed. */
    guid.bytes[8] |= 0x80;
    guid.bytes[6] &= 0x0F;
    guid.bytes[6] |= 0x40;
    return guid;
}

static boolean DOF_CreateInternal( DOF dof, DOFConfig config )
{
    {
        dof->maxConn = DOFConfig_GetConnectionLimit( dof->config );
        dof->maxOperations = DOFConfig_GetMaxOperations( dof->config );
    }

    {
        uint32 scopeid = 0; /* All one scope. */
        DOFObjectIDDomain domain = DOFObjectIDDomain_Create( DOFObjectIDSource_GetObjectID( DOFConfig_GetSourceID( dof->config) ) );
        dof->defaultSecurityScope = DOFSecurityScope_Create( domain, 1, &scopeid, 0, NULL );
        dof->emptySecurityScope = DOFSecurityScope_Create(domain, 0, NULL, 0, NULL);
        DOFObjectIDDomain_Destroy( domain );
    }

    if( !dof->defaultSecurityScope ||
        !dof->emptySecurityScope )
        goto FAIL;

    dof->defaultFilter = DOFOperationFilter_CreateNullFilter();
    if(dof->defaultFilter == NULL)
    {
        goto FAIL;
    }

    return TRUE;

FAIL:
    DOFSecurityScope_Destroy(dof->defaultSecurityScope);
    DOFSecurityScope_Destroy(dof->emptySecurityScope);
    return FALSE;
}

DOFSecurityScope DOF_GetDefaultSecurityScope( DOF dof )
{
    if( dof )
        return dof->defaultSecurityScope;
    return NULL;
}

DOFSecurityScope DOF_GetEmptySecurityScope( DOF dof )
{
    if( dof )
        return dof->emptySecurityScope;
    return NULL;
}

OALOperationBridge DOF_GetDefaultOperationBridge( DOF dof )
{
    if( dof )
        return dof->defaultBridge;
    return NULL;
}

DOFOperationFilter DOF_GetDefaultOperationFilter(DOF dof)
{
    PCR_Assert(dof != NULL);
    return dof->defaultFilter;
}

DOFConnection DOF_CreateConnection( DOF dof, DOFConnectionConfig config )
{
    if (dof && config)
    {
        DOFSecurityMembrane membrane = DOFConnection_CreateSecurityMembrane(dof, config);

        if(membrane != NULL)
        {
            DOFConnection ret = DOFConnection_CreateInternalConnection( dof, membrane, config, NULL, NULL, TRUE );

            DOFSecurityMembrane_Destroy(membrane);
            return ret;
        }
    }
    return NULL;
}

static boolean findServer( OALSListNode itemx, void *argx )
{
    DOFServerNode_t* item = (DOFServerNode_t*)itemx;
    DOFServerConfig arg = (DOFServerConfig)argx;

    return DOFServerConfig_IsEqual(item->server->server->config, arg);
}

boolean DOF_AddServer( DOF dof, DOFServer srv )
{
    DOFServerNode_t* node = NULL;
    if( dof && srv )
    {
        node = DOFMem_Alloc(sizeof( DOFServerNode_t ));
        if( node )
        {
            node->next = NULL;
            node->server = srv;
            if( node->server )
            {
                DOF_Lock();
                OALSList_Insert( &dof->servers, (void*)node);
                DOF_Unlock();
                return TRUE;
            }
            else
                DOFMem_Free(node);
        }
    }
    return FALSE;
}

DOFServer DOF_CreateServer( DOF dof, DOFServerConfig config )
{
    if( dof && config && DOFAddress_GetAddressType( DOFServerConfig_GetAddress( config ) ) == DOFADDRESSTYPE_UNICAST )
    {
        DOFServerNode_t* node;


        node = (DOFServerNode_t*)OALSList_Retrieve( dof->servers, findServer, config);
        if( node )
        {
            return DOFServer_Retain( node->server );
        }
        else
        {
            node = DOFMem_Alloc(sizeof( DOFServerNode_t ));
            if( node )
            {
                DOFSecurityMembrane membrane = DOFServer_CreateSecurityMembrane(dof, config);
                if(membrane != NULL)
                {
                    node->next = NULL;
                    node->server = DOFServer_Create( dof, config, membrane, NULL );
                    DOFSecurityMembrane_Destroy(membrane);
                    if( node->server )
                    {
                        DOF_Lock();
                        OALSList_Insert( &dof->servers, (void*)node);
                        DOF_Unlock();
                        return node->server;
                    }
                }
                DOFMem_Free(node);
            }
        }
    }
    return NULL;
}

boolean DOF_SetNodeDown( DOF dof )
{
    PCR_Assert( dof );
    DOF_Lock();
    dof->flags.IsNodeDown = TRUE;
    DOF_SetConnectionLimit(dof, 0);
    DOF_Unlock();
    return TRUE;
}

boolean DOF_IsNodeDown( DOF dof )
{
    PCR_Assert( dof );
    return dof->flags.IsNodeDown;
}

void init_mutex_pool(void)
{
    /* mutex pool stuff */
    PCR_Assert(!mutex_pool);

    ref_count_mutex = PCRMutex_Create();
    addr_mutex = PCRMutex_Create();
    connection_server_mutex = PCRMutex_Create();
    connection_server_count_mutex = PCRMutex_Create();

    mutex_pool = OALResourcePool_Create(DOFRESOURCE_NUM_MUTEXES, (OALResourcePool_Create_Resource_fn_t)PCRMutex_Create, (OALResourcePool_Destroy_Resource_fn_t)PCRMutex_Destroy); /* this number can be tuned */
    PCR_Assert(mutex_pool);
    OALThreadPool_Init();
    OALTimer_Init();
    OALRemoteChannelMgr_MutexInit();
    source_mutex_init();
    DOFOperationControl_Init();
    OALPermissionScope_Init();
    DOFPermissionSet_Init();
    OALObservable_Init();
    EPP_NegOne_Mutex_Init();
    OALProtocolStack_Init();
}

void destroy_mutex_pool(void)
{
    DOFRuntime_RemoveAllRoutingCallback(); /* Destroys the RoutingInfoChanged Callback OALAVLTree if it exist*/

    PCR_Assert(mutex_pool);

    OALProtocolStack_Cleanup();
    EPP_NegOne_Mutex_Shutdown();
    OALObservable_Shutdown();
    DOFPermissionSet_Shutdown();
    OALPermissionScope_Shutdown();
    DOFOperationControl_Shutdown();
    source_mutex_shutdown();
    OALTimer_Shutdown();
    OALRemoteChannelMgr_MutexShutdown();
    OALThreadPool_Shutdown();

    OALResourcePool_Destroy(mutex_pool);
    mutex_pool = 0;

    PCRMutex_Destroy(ref_count_mutex);
    PCRMutex_Destroy(addr_mutex);
    PCRMutex_Destroy(connection_server_mutex);
    PCRMutex_Destroy(connection_server_count_mutex);
    ref_count_mutex = 0;
    addr_mutex = 0;
    connection_server_mutex = NULL;
}

struct SendInterestToSystem_t
{
    OALEnqueueItem_t parent;
    DOFSystem        system;
    DOFOperation     op;
}SendInterestToSystem_t;

/* Only caller is DOF_HandleInterestRequest(). */
static void DOF_SendInterestToSystem(OALThreadPoolItem item, boolean shutdown)
{
    struct SendInterestToSystem_t* node = (struct SendInterestToSystem_t*)item;
    if( !shutdown )
    {
        if( DOFSecurityScope_IsCompatible( node->op->securityScope, OALChannel_GetSecurityScope( DOFSystem_GetChannel( node->system ), FALSE ) ))
        {
            /* 3rd param NULL means is interest op, one of _ACTIVATE, _WATCH, _CONNECT, or _FULLCONNECT. */
            if( !DOFSystem_AcceptRequest( DOFSystem_GetChannel( node->system ), node->op, NULL ) )
            {
                DOFLog_Message( DOFLOGLEVEL_DEBUG, "DOF: Interest operation was not delivered to system." );
            }
        }
    }

    DOFSystem_Release( node->system );
    DOFOperation_Internal_Destroy( node->op );
    DOFMem_Free( node );
}

static boolean DOF_EnqueueInterestToSystem( OALSListNode itemx, void *arg )
{
    DOFSystemNode_t* item = (DOFSystemNode_t*)itemx;
    DOFOperation op = (DOFOperation)arg;
    struct SendInterestToSystem_t* node = (struct SendInterestToSystem_t*)DOFMem_Alloc( sizeof( struct SendInterestToSystem_t ) );
    if( node )
    {
        OALEnqueueItem_Init( (OALEnqueueItem)node, DOF_SendInterestToSystem );

        node->system = DOFSystem_Retain( item->system );
        if( node->system )
        {
            DOFOperation_Retain( op );
            node->op = op;

            if ( OALEnqueue_Enqueue( DOF_GetDof()->enqueue, (OALEnqueueItem)node ) )
                return FALSE; /* continue */
            DOFOperation_Internal_Destroy( op );
            DOFSystem_Release( node->system );
        }

        DOFMem_Free( node );
    }
    DOFLog_Message( DOFLOGLEVEL_WARNING, "DOF: Interest operation could not be enqueued for system." );
    return FALSE;
}

/* Called only in interest.c, where op's type is one of _ACTIVATE, _WATCH, _CONNECT, or _FULLCONNECT. */
void DOF_HandleInterestRequest( DOF dof, DOFOperation op)
{
    DOF_Lock();
    DOF_ProcessInterest( dof, op ); /* notify dof's listeners */
    OALSList_Traverse( dof->systems, DOF_EnqueueInterestToSystem, op); /* notify compatible systems */
    DOF_Unlock();
}

static boolean DOF_CleanupSystem( OALSListNode itemx, void *arg )
{
    DOFSystemNode_t* item = (DOFSystemNode_t*)itemx;

    DOFSystem_Cleanup( item->system );
    return FALSE;
}

void DOF_CleanupSystems( DOF dof )
{
    uint32 remainingSystemCleanupCount;

    DOF_Lock();
    OALSList_Traverse( dof->systems, DOF_CleanupSystem, NULL );
    remainingSystemCleanupCount = dof->remainingSystemCleanupCount;
    DOF_Unlock();
    while (remainingSystemCleanupCount)
    {
        uint32 systemsLeft;

        PCREvent_Wait(dof->systemsGoneEvent, DOF_SYSTEM_CLEANUP_GRACE_PERIOD);

        DOF_Lock();
        systemsLeft = dof->remainingSystemCleanupCount;
        DOF_Unlock();

        /* Shutting_down prevents new entries, so this should *always* be less, except in the case of system leak (runtime error). */
        PCR_Validate( systemsLeft < remainingSystemCleanupCount );
        remainingSystemCleanupCount = systemsLeft;
    }
}

boolean DOF_AddSystem( DOF dof, DOFSystem system )
{
    if (dof && system)
    {
        DOFSystemNode_t * newNode = (DOFSystemNode_t*)DOFMem_Alloc(sizeof(DOFSystemNode_t));
        if( newNode )
        {
            newNode->next = NULL;
            newNode->system = system;
            DOF_Lock();
            OALSList_Insert( &dof->systems, (OALSListNode)newNode);
            PCREvent_Reset(dof->systemsGoneEvent);
            DOF_Unlock();
            return TRUE;
        }
        DOFMem_Free(newNode);
    }

    return FALSE;
}

static boolean DOF_FindAndRemoveSystem( OALSListNode itemx, void *arg )
{
    DOFSystemNode_t* item = (DOFSystemNode_t*)itemx;
    DOFSystem system = (DOFSystem)arg;
    return ( item->system == system );
}

void DOF_RemoveSystem( DOF dof, DOFSystem system )
{
    DOFSystemNode_t* node;

    PCR_Assert(dof == DOFSystem_GetDOF(system));

    if ( !dof || !system )
        return;

    DOF_Lock();
    node = (DOFSystemNode_t*)OALSList_FindAndRemove( &dof->systems, NULL, DOF_FindAndRemoveSystem, (void*)system);
    DOF_Unlock();
    /* It should always be in the list. */
    DOFMem_Free(node);
    /* Note: If we just removed last DOFSystem from dof->systems, then dof->systems is now NULL. */
}

static boolean DOF_FindAndRemoveServer(OALSListNode itemx, void *arg)
{
    DOFServerNode_t* item = (DOFServerNode_t*)itemx;
    DOFServer server = (DOFServer)arg;

    return ( item->server == server );
}

void DOF_RemoveServer( DOF dof, DOFServer server )
{
    DOFServerNode_t* node = NULL;

    if( !dof || !server )
        return;

    DOF_Lock();
    node = (DOFServerNode_t*)OALSList_FindAndRemove( &dof->servers, NULL, DOF_FindAndRemoveServer, (void*)server);
    DOFMem_Free(node);
    DOF_Unlock();
}

static void DOF_RemoveAllServerNodes(OALSListNode item, void *arg)
{
    DOFMem_Free( item );
}

static void DOF_RemoveServersNodes( DOF dof )
{
    if( !dof )
        return;

    DOF_Lock();
    OALSList_RemoveAll( &dof->servers, DOF_RemoveAllServerNodes, NULL);
    DOF_Unlock();
}

boolean DOF_IsParameterValidation( const DOF self )
{
    PCR_Assert( self );

    return DOFConfig_IsParameterValidation( self->config );
}

static boolean count_systems( OALSListNode item, void *arg )
{
    uint32 *count = (uint32*)arg;

    ++(*count);
    return FALSE; /* don't stop early */
}

struct sys_list_wrapper
{
    DOFSystemList sys_list;
    uint32 desired_count;
};

static boolean add_systems( OALSListNode item, void *arg )
{
    DOFSystemNode_t * node = (DOFSystemNode_t *)item;
    DOFSystem system;
    struct sys_list_wrapper* wrapper = (struct sys_list_wrapper*)arg;

    PCR_Assert(wrapper->sys_list->systemCount < wrapper->desired_count);

    system = DOFSystem_Retain( node->system );
    if(!system)
        return TRUE;

    wrapper->sys_list->systemArray[wrapper->sys_list->systemCount++] = system;
    return FALSE;
}

static void remove_free_systems( struct sys_list_wrapper *wrapper )
{
    uint32 i;
    for( i=0; i < wrapper->sys_list->systemCount; ++i )
    {
        DOFSystem_Release(wrapper->sys_list->systemArray[i]);
    }
}

DOFSystemList DOF_GetSystems(const DOF self)
{
    /* since locking a mutex to keep the system count is a horrible pain, let's just iterate
     over the list of systems twice. it isn't like it's going to take a long time */
    uint32 count = 0;
    DOFSystemList sys_list = NULL;

    DOF_Lock();
    OALSList_Traverse((OALSListNode)self->systems, count_systems, &count);
    sys_list = DOFMem_Alloc(sizeof(struct DOFSystemList_t));
    if(sys_list)
    {
        /* set count in traverse */
        sys_list->systemCount = 0;
        sys_list->systemArray = DOFMem_Alloc(sizeof(DOFSystem)*count);
        if(sys_list->systemArray)
        {
            struct sys_list_wrapper wrapper;
            wrapper.desired_count = count;
            wrapper.sys_list = sys_list;
            if( OALSList_Traverse((OALSListNode)self->systems, add_systems, &wrapper) )
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "DOF: Failed to copy a system.");
                remove_free_systems(&wrapper);
                DOFMem_Free(sys_list);
                DOF_Unlock();
                return NULL;
            }
            PCR_Assert(count==sys_list->systemCount);
        }
        else
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "DOF: Failed to allocate space for system list array.");
            DOFMem_Free(sys_list);
            sys_list = NULL;
        }
    }
    else
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "DOF: Failed to allocate space for system list.");
    }
    DOF_Unlock();

    return sys_list;
}

typedef struct CompatibleSystemContext_t
{
    DOFSecurityScope scope;
    DOFSystemList list;
    uint32 count;
} CompatibleSystemContext_t;

void DOF_CancelAllOperations( DOF self )
{
    EPP_NegOne_CancelAll();
}

void DOF_Close( DOF self, DOFConnection conn )
{
    if( self && conn )
    {
        DOFException exception = DOFErrorException_Create(DOFERROR_TERMINATED, "Terminated by DOF.");
        DOFConnection_Terminate(conn, exception);
        DOFException_Destroy(exception);
    }
}

void DOF_Stop( DOF self, DOFServer srv )
{
    if( self && srv )
    {
        DOFException exception = DOFErrorException_Create(DOFERROR_TERMINATED, "Terminated by DOF.");
        DOFServer_Kill( srv, exception );
        DOFException_Destroy(exception);
    }
}

DOFAuthenticator DOF_CreateAuthenticator( DOF self, DOFAuthenticatorConfig config )
{
    if( self && config )
        return DOF_CreateAuthenticatorInternal( self, config );
    return NULL;
}

static boolean buildAuthenticatorList(OALSListNode itemx, void *arg)
{
    DOFAuthenticatorListItem_t* item = (DOFAuthenticatorListItem_t*)itemx;
    DOFAuthenticatorList ret = (DOFAuthenticatorList)arg;

    ret->list[ret->count] = DOFAuthenticator_Copy(item->authenticator);
    ++ret->count;
    return FALSE;
}

DOFAuthenticatorList DOF_GetAuthenticators( const DOF self )
{
    if( self )
    {
        DOFAuthenticatorList ret;
        DOF_Lock();
        ret = DOFMem_Alloc( sizeof(struct DOFAuthenticatorList_t) + sizeof(DOFAuthenticator) * self->authenticatorsCnt);
        if( ret )
        {
            ret->list = ((DOFAuthenticator*)((uint8*)ret)+sizeof(struct DOFAuthenticatorList_t));
            OALSList_Traverse((OALSListNode)self->authenticators, buildAuthenticatorList, (void*)ret);
            DOF_Unlock();
            return ret;
        }
        DOF_Unlock();
    }
    return NULL;
}

boolean DOF_ResolveCredentials( DOF self, DOFCredentials credentials, int timeout, DOFException* exception )
{
    return DOFCredentials_Resolve( credentials, self, timeout, exception );
}

DOFRuntime DOF_GetRuntime( DOF dof )
{
    return (DOFRuntime)dof;
}

void DOF_Inc_InputByteCount( DOF self, uint32 count )
{
    if( self )
    {
        PCRMutex_Lock( self->runtime_mutex );
        self->inputByteCount += count;
        ++self->inputPacketCount;
        PCRMutex_Unlock( self->runtime_mutex );
    }
}

void DOF_Inc_OutputByteCount( DOF self, uint32 count )
{
    if( self )
    {
        PCRMutex_Lock( self->runtime_mutex );
        self->outputByteCount += count;
        ++self->outputPacketCount;
        PCRMutex_Unlock( self->runtime_mutex );
    }
}

boolean DOF_SetMaxConnections( DOF dof, uint16 max )
{
    if( dof )
    {
        if ( max == 0 )
        {
            /* For DOF_SetMaxConnections, 0 means max. */
            return DOF_SetConnectionLimit( dof, DOF_UNLIMITED );
        }
        return DOF_SetConnectionLimit( dof, max );
    }
    return FALSE;
}

boolean DOF_SetConnectionLimit( DOF dof, uint32 max )
{
    if( dof )
    {
        DOF_Lock();
            dof->maxConn = max;
        DOF_Unlock();
        return TRUE;
    }
    return FALSE;
}

uint16 DOF_GetMaxConnections( DOF dof )
{
    uint16 ret = 0;
    if( dof )
    {
        DOF_Lock();
            ret = dof->maxConn;
        DOF_Unlock();
    }
    return ret;
}

boolean DOF_SetMaxOperations( DOF dof, uint32 max )
{
    if( dof )
        return TRUE;
    return FALSE;
}

uint32 DOF_GetMaxOperations( DOF dof )
{
    return DOF_UNLIMITED;
}

/* New DOFDomain stuff */
typedef struct DOFDomainNode_t
{
    OALSListNode    next;
    DOFDomain       domain;
} DOFDomainNode_t;

static boolean DOF_FindDOFDomain( OALSListNode item, void *arg )
{
    DOFDomainConfig cfg = (DOFDomainConfig)arg;
    DOFDomainNode_t* cfgNode = (DOFDomainNode_t*)item;

    return DOFCredentials_IsEqual( cfg->creds, cfgNode->domain->creds );
}

DOFDomain DOF_CreateDomain( DOF self, DOFDomainConfig config )
{
    DOFDomainNode_t* node;

    DOF_Lock();
    node = (DOFDomainNode_t*)OALSList_Retrieve( self->domainsHead, DOF_FindDOFDomain, (void*)config );
    if( node )
    {
        /* alredy have the domain */
        DOF_Unlock();
        DOFDomain_UpdateSilenceTimer( node->domain, config->maxSilence );
        return DOFDomain_Copy( node->domain );
    }

    node = (DOFDomainNode_t*)DOFMem_Alloc( sizeof( DOFDomainNode_t ) );
    if( node )
    {
        node->domain = DOFDomain_CreateInternal( config );
        if( node->domain )
        {
            node->domain->dof = self;
            node->next = NULL;
            OALSList_Insert( &self->domainsHead, (OALSListNode)node );
            ++(self->domainsCnt);
            DOF_Unlock();
            return node->domain;
        }
        DOFMem_Free( node );
    }
    DOF_Unlock();
    return NULL;
}

static boolean RemoveDomain(OALSListNode item, void *arg)
{
    DOFDomainNode_t* node = (DOFDomainNode_t*)item;
    return (node->domain == arg);
}

void DOF_RemoveDomain( DOF self, DOFDomain domain )
{
    DOFDomainNode_t* node;

    if( self )
    {
        DOF_Lock();
        node = (DOFDomainNode_t*)OALSList_FindAndRemove( &self->domainsHead, NULL, RemoveDomain, domain );
        if( node )
        {
            --(self->domainsCnt);
            DOFMem_Free( node );
        }
        DOF_Unlock();
    }
}

struct cdl_t
{
    DOFDomainList   list;
    uint32          addedCount;
};

static boolean DOF_AddDOFDomainToList( OALSListNode item, void *arg )
{
    struct cdl_t* ctx = (struct cdl_t*)arg;
    DOFDomainNode_t* cfgNode = (DOFDomainNode_t*)item;

    ctx->list->list[ctx->addedCount] = DOFDomain_Copy( cfgNode->domain );

    if( ctx->list->list[ctx->addedCount++] ) /* check and inc added count */
        return FALSE;
    return TRUE; /* early exit error */
}

DOFDomainList DOF_GetDomains( DOF self )
{
    DOFDomainList ret;

    DOF_Lock();
    ret = (DOFDomainList)DOFMem_Alloc( sizeof( struct DOFDomainList_t ) + ( self->domainsCnt * sizeof( DOFDomain ) ) );

    if( ret )
    {
        struct cdl_t ctx;
        ctx.addedCount = 0;
        ctx.list = ret;

        PCRMem_Set( ret, 0, sizeof( struct DOFDomainList_t ) + ( self->domainsCnt * sizeof( DOFDomain ) ) );

        ret->count = self->domainsCnt;
        ret->list = (DOFDomain*)(((uint8*)ret) + sizeof( struct DOFDomainList_t ));

        if( !OALSList_Traverse( self->domainsHead, DOF_AddDOFDomainToList, (void*)&ctx ) )
        {
            /* success */
            DOF_Unlock();
            return ret;
        }
        DOFDomainList_Destroy( ret ); /* destroy whole list even though we may have failed early */
    }
    DOF_Unlock();
    return NULL;
}

void DOFDomainList_Destroy( DOFDomainList self )
{
    if( self )
    {
        uint32 i = 0;
        for( ; i < self->count; ++i )
        {
            DOFDomain_Destroy( self->list[i] );
        }
        DOFMem_Free( self );
    }
}

static void DomainListenerNodeRemoved(OALThreadPoolItem item, boolean shutdown)
{
    DOFDomainListenerRemoveNode node = (DOFDomainListenerRemoveNode)item;
    if( node->listener->fns->Removed )
        node->listener->fns->Removed( node->listener, gDOF, node->exception);
    DOFMem_Free( node->node ); /* we are done with it. */
}

static void DOFDomainListernerNode_EnqueueRemoved( DOFDomainListenerRemoveNode node )
{
    OALEnqueueItem_Init( (OALEnqueueItem)node, DomainListenerNodeRemoved );
    OALEnqueue_Enqueue( gDOF->enqueue, (OALEnqueueItem)node );
}

typedef struct DOFDomainListenerNodeNotify_t
{
    OALEnqueueItem_t    base;
    DOFDomainListener   listener;
    DOFCredentials      creds;
    DOFConnectionState  state;
    DOFDomainConfig     config;
}DOFDomainListenerNodeNotify_t;

static void DomainListenerNodeNotify(OALThreadPoolItem item, boolean shutdown)
{
    DOFDomainListenerNodeNotify_t* enqueueItem = (DOFDomainListenerNodeNotify_t*)item;

    enqueueItem->listener->fns->DomainAdded( enqueueItem->listener, gDOF, enqueueItem->config, enqueueItem->creds, enqueueItem->state );

    DOFDomainConfig_Destroy( enqueueItem->config );
    DOFConnectionState_Destroy( enqueueItem->state );
    DOFCredentials_Destroy( enqueueItem->creds );
    DOFMem_Free( enqueueItem );
}

struct dlnntc_t
{
    DOFCredentials      creds;
    DOFConnectionState  state;
    DOFDomainConfig     config;
    OALSListNode        head;
};

static boolean DomainListenerNodeNotifyTraverse( OALAVLTreeItem_t item, void *arg )
{
    struct dlnntc_t* ctx = (struct dlnntc_t*)arg;
    DOFDomainListenerNode node = (DOFDomainListenerNode)item;
    DOFDomainListenerNodeNotify_t* enqueueItem = (DOFDomainListenerNodeNotify_t*)DOFMem_Alloc( sizeof(DOFDomainListenerNodeNotify_t) );
    if( enqueueItem )
    {
        OALEnqueueItem_Init( (OALEnqueueItem)enqueueItem, DomainListenerNodeNotify );
        enqueueItem->creds = DOFCredentials_Copy( ctx->creds );
        if( enqueueItem->creds )
        {
            enqueueItem->state = DOFConnectionState_Copy( ctx->state );
            if( enqueueItem->state )
            {
                enqueueItem->config = DOFDomainConfig_Copy( ctx->config );
                if( enqueueItem->config )
                {
                    enqueueItem->listener = node->self.listener;
                    OALEnqueue_Enqueue( gDOF->enqueue, (OALEnqueueItem)enqueueItem );
                    return FALSE;
                }
                DOFConnectionState_Destroy( enqueueItem->state );
            }
            DOFCredentials_Destroy( enqueueItem->creds );
        }
        DOFMem_Free( enqueueItem );
    }
    OALSList_Insert( &(ctx->head), (OALSListNode)node ); /* so we can remove and call removed for this listener because it wasn't notified */
    return FALSE; /* continue traverse */
}

static void RemoveAll_DomainListeners(OALAVLTreeItem_t item, void *arg)
{
    ((DOFDomainListenerNode)item)->self.exception = arg; /* if an arg was passed it is a DOFException */
    DOFDomainListernerNode_EnqueueRemoved( &(((DOFDomainListenerNode)item)->self) );
}

static void DOF_Internal_RemoveAllDomainListeners( DOF self, boolean destroy, DOFException exception )
{
    DOF_Lock();
    if( destroy )
    {
        OALAVLTree_Destroy( self->domainListeners, RemoveAll_DomainListeners, exception );
        self->domainListeners = NULL;
    }
    else
    {
        OALAVLTree_Empty( self->domainListeners, RemoveAll_DomainListeners, exception );
    }
    DOF_Unlock();
}

static void DOF_RemoveAllDomainListeners( DOF self )
{
    DOF_Internal_RemoveAllDomainListeners( self, TRUE, NULL );
}

static void FailedToNotify_RemovedCall( OALSListNode item, void *arg )
{
    DOFDomainListenerNode node = (DOFDomainListenerNode)item;
    DOF dof = (DOF)arg;
    if( OALAVLTree_Delete( dof->domainListeners, node->self.listener ) )
        DOFDomainListernerNode_EnqueueRemoved( &(node->self) );
}

void DOF_NotifyNewDomain( DOF self, DOFCredentials creds, DOFConnectionState state )
{
    DOF_Lock();
    if( OALAVLTree_GetNumItems( self->domainListeners ) > 0 )
    {
        struct dlnntc_t ctx;

        ctx.creds = creds;
        ctx.state = state;
        ctx.head = NULL;
        ctx.config = DOFDomainConfigBuilder_BuildAndDestroy( DOFDomainConfigBuilder_Create( creds ) );

        if( ctx.config )
        {
            OALAVLTree_Traverse( self->domainListeners, DomainListenerNodeNotifyTraverse, &ctx );
            DOFDomainConfig_Destroy( ctx.config );
        }
        else
        {
            DOF_Internal_RemoveAllDomainListeners( self, FALSE, DOFEXCEPTION_INSUFFICIENT_RESOURCES ); /* Can't notify any, do to alloc failure, remove all */
        }

        if( ctx.head != NULL )
        {
            OALSList_RemoveAll( &ctx.head, FailedToNotify_RemovedCall, self );
        }
    }
    DOF_Unlock();
}

boolean DOF_AddDomainListener( DOF self, DOFDomainListener listener )
{
    if( self && listener )
    {
        DOFDomainListenerNode node = (DOFDomainListenerNode)DOFMem_Alloc( sizeof( DOFDomainListenerNode_t ) );
        if( node )
        {
            OALAVLTree_InsertionStatus status;
            PCRMem_Set( node, 0, sizeof( DOFDomainListenerNode_t ) );

            node->self.node = node;
            node->self.listener = listener;

            DOF_Lock();
            status = OALAVLTree_Insert( self->domainListeners, listener, node, NULL );
            DOF_Unlock();

            switch( status )
            {
            case OALAVLTree_AlreadyInserted:
                DOFMem_Free( node );
                /* Fall through */
            case OALAVLTree_InsertSuccess:
                return TRUE;
                break;
            case OALAVLTree_InsertFailure:
            default:
                DOFMem_Free( node );
                break;
            }
        }
    }
    return FALSE;
}

void DOF_RemoveDomainListener( DOF self, DOFDomainListener listener )
{
    if( self && listener )
    {
        DOFDomainListenerNode node;

        DOF_Lock();
        node = (DOFDomainListenerNode)OALAVLTree_Retrieve( self->domainListeners, listener );
        if( node )
        {
            if( OALAVLTree_Delete( self->domainListeners, listener ) )
            {
                DOF_Unlock();
                DOFDomainListernerNode_EnqueueRemoved( &(node->self) );
                return;
            }
            else
                PCR_Assert(FALSE); /* We just found it! Why wasn't it removed? */
        }
        DOF_Unlock();
    }
}

