/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/dof.h>
#include <dof/address.h>
#include <dof/connection.h>
#include <dof/filter.h>
#include <dof/interfaceid.h>
#include <dof/objectid.h>
#include <dof/permission.h>
#include <dof/permissionset.h>
#include <internal/securityscope.h>
#include <dof/system.h>
#include <internal/bridgeinternal.h>
#include <internal/exception.h>
#include <internal/systeminternal.h>
#include <internal/trp_v1.h>
#include <stdlib.h>

/** @addtogroup DOF DOF
@{
**/

static OALOperationBridge CreateBridge(DOF               dof,
                                       DOFSystemConfig   config,
                                       DOFException    * pException,
                                       uint32            timeout)
{
    DOFOperationBridgeConfigBuilder builder = DOFOperationBridgeConfigBuilder_Create();

    if(builder != NULL)
    {
        DOFOperationBridgeConfig bridgeConfig;

        DOFOperationBridgeConfigBuilder_SetCredentials(builder, DOFSystemConfig_GetCredentials(config));
        DOFOperationBridgeConfigBuilder_SetPermissions(builder, DOFSystemConfig_GetPermissions(config));
        DOFOperationBridgeConfigBuilder_SetPermissionsExtendAllowed(builder, DOFSystemConfig_IsPermissionsExtendAllowed(config));
        bridgeConfig = DOFOperationBridgeConfigBuilder_BuildAndDestroy(builder);
        if(bridgeConfig != NULL)
        {
            OALOperationBridge ret = DOF_CreateBridge(dof, bridgeConfig, NULL, NULL, timeout, pException);

            DOFOperationBridgeConfig_Destroy(bridgeConfig);
            return ret;
        }
    }
    return NULL;
}

DOFSystem DOF_CreateSystem(DOF               dof,
                           DOFSystemConfig   systemConfig,
                           uint32            timeout,
                           DOFException    * pException)
{
    DOFSecurityMembrane membrane = NULL;

    PCR_Assert( dof );
    if(pException != NULL)
    {
        *pException = NULL;
    }
    if(systemConfig != NULL)
    {
        DOFOperationFilter receiveFilter = DOFSystemConfig_GetReceiveFilter(systemConfig);
        DOFOperationFilter sendFilter = DOFSystemConfig_GetSendFilter(systemConfig);
        OALOperationBridge bridge = CreateBridge(dof, systemConfig, pException, timeout);

        if(receiveFilter == NULL)
        {
            receiveFilter = DOF_GetDefaultOperationFilter(dof);
        }
        if(sendFilter == NULL)
        {
            sendFilter = DOF_GetDefaultOperationFilter(dof);
        }
        if((receiveFilter != NULL) &&
           (sendFilter != NULL) &&
           (bridge != NULL))
        {
            membrane = OALChannel_CreateSecurityMembrane(dof, receiveFilter, sendFilter, bridge);
        }
        OALOperationBridge_Destroy(bridge);
    }
    else
    {
        membrane = OALChannel_CreateSecurityMembrane(dof, DOF_GetDefaultOperationFilter(dof),
                                                     DOF_GetDefaultOperationFilter(dof),
                                                     DOF_GetDefaultOperationBridge(dof));
    }
    if(membrane != NULL)
    {
        DOFSystem system = DOFSystem_Create(dof, systemConfig, membrane);

        DOFSecurityMembrane_Destroy(membrane);
        if(system == NULL)
        {
            if(pException != NULL)
            {
                *pException = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            }
        }
        return system;
    }
    else if((pException != NULL) && (*pException == NULL))
    {
        *pException = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
    }

    return NULL;                                /* Error out. */
}


/**
@}
**/
