/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/dof.h>
#include <internal/bridgeinternal.h>
#include <internal/exception.h>
#include <internal/permissionsetinternal.h>
#include <internal/trp_v1.h>

typedef struct CreateSecureBridgeContext_t
{
    PCRSemaphore semaphore;
    DOFCredentials resolvedCredentials;
    DOFException exception;
    DOF dof;
    OALChannel channel;
    DOFAddress address;
    boolean fixedPermissions;
} CreateSecureBridgeContext_t;

typedef CreateSecureBridgeContext_t * CreateSecureBridgeContext;

typedef struct CreateSecureBridgeScopeContext_t
{
    CreateSecureBridgeContext fullContext;
    DOFPermissionSet permissionSet;
    DOFSecurityScope securityScope;
    DOFException exception;
    OALChannel channel;
    DOFAddress address;
    uint8 duration;
} CreateSecureBridgeScopeContext_t;

typedef CreateSecureBridgeScopeContext_t * CreateSecureBridgeScopeContext;

static boolean CreateBridge_ResolveCredentials(CreateSecureBridgeContext context,
                                               uint32                    timeout);

static boolean CreateBridge_ResolveScope(CreateSecureBridgeScopeContext context,
                                         DOFObjectIDAuthentication      id,
                                         uint32                         timeout);

OALOperationBridge DOF_CreateBridge(DOF                  dof,
                                    DOFOperationBridgeConfig   bridgeConfig,
                                    OALChannel           channel,
                                    DOFAddress           address,
                                    uint32               timeout,
                                    DOFException       * pException)
{
    OALOperationBridge bridge = NULL;

    if(dof != NULL)
    {
        if(pException != NULL)
        {
            *pException = NULL;
        }
        if((bridgeConfig == NULL) ||
           (DOFOperationBridgeConfig_GetCredentials(bridgeConfig) == NULL))
        {
            bridge = OALOperationBridge_CreateNonSecureBridge(dof, bridgeConfig);
        }
        else
        {
            CreateSecureBridgeContext_t context;

            context.semaphore = PCRSemaphore_Create(0);
            if(context.semaphore != NULL)
            {
                boolean goingToBeOk = TRUE;

                uint32 currentTime = PCRTime_GetMS();
                uint32 absoluteTimeout = currentTime + timeout;

                context.exception = NULL;
                context.dof = dof;
                context.channel = channel;
                context.address = address;
                context.resolvedCredentials = DOFOperationBridgeConfig_GetCredentials(bridgeConfig);
                if(!DOFCredentials_IsResolved(context.resolvedCredentials))
                {
                    goingToBeOk = CreateBridge_ResolveCredentials(&context, timeout);
                }
                else
                {
                    goingToBeOk = TRUE;
                }
                if(goingToBeOk)
                {
                    CreateSecureBridgeScopeContext_t scopeContexts[2];

                    DOFPermissionSet permSet = DOFOperationBridgeConfig_GetPermissions(bridgeConfig);

                    DOFObjectIDAuthentication id = DOFOperationBridgeConfig_GetAlternateIdentity(bridgeConfig);

                    if(id == NULL)
                    {
                        id = DOFCredentials_GetIdentity(context.resolvedCredentials);
                    }
                    context.fixedPermissions = !DOFOperationBridgeConfig_IsPermissionsExtendAllowed(bridgeConfig);
                    scopeContexts[0].permissionSet = (permSet != NULL) ? DOFPermissionSet_Copy(permSet)
                                                                       : DOFPermissionSet_DeepCopy(DOFPermissionSet_GetEmptyPermissionSet());
                    if(scopeContexts[0].permissionSet != NULL)
                    {
                        scopeContexts[1].permissionSet = DOFPermissionSet_CreateComplement(scopeContexts[0].permissionSet);
                        if(scopeContexts[1].permissionSet != NULL)
                        {
                            scopeContexts[0].duration = scopeContexts[1].duration = 1;
                            scopeContexts[0].securityScope = scopeContexts[1].securityScope = NULL;
                            scopeContexts[0].exception = scopeContexts[1].exception = NULL;
                            scopeContexts[0].fullContext = scopeContexts[1].fullContext = &context;
                            scopeContexts[0].channel = scopeContexts[1].channel = channel;
                            scopeContexts[0].address = scopeContexts[1].address = address;
                            currentTime = PCRTime_GetMS();
                            if (absoluteTimeout - currentTime <= timeout)
                                timeout = absoluteTimeout - currentTime;
                            else
                                timeout = 0;
                            if (CreateBridge_ResolveScope(&(scopeContexts[0]), id, timeout))
                            {
                                currentTime = PCRTime_GetMS();
                                if (absoluteTimeout - currentTime <= timeout)
                                    timeout = absoluteTimeout - currentTime;
                                else
                                    timeout = 0;
                                if(CreateBridge_ResolveScope(&(scopeContexts[1]), id, timeout))
                                {
                                    PCRSemaphore_Wait(context.semaphore);
                                    PCRSemaphore_Wait(context.semaphore);
                                    if((scopeContexts[0].exception == NULL) &&
                                       (scopeContexts[1].exception == NULL))
                                    {
                                        bridge = OALOperationBridge_CreateSecureBridge(dof, context.resolvedCredentials, id, scopeContexts[1].permissionSet,
                                                                                       scopeContexts[1].securityScope, scopeContexts[1].duration,
                                                                                       scopeContexts[0].permissionSet, scopeContexts[0].securityScope,
                                                                                       scopeContexts[0].duration, !context.fixedPermissions);
                                    }
                                }
                                else
                                {
                                    PCRSemaphore_Wait(context.semaphore);
                                }
                            }
                            if (pException != NULL)
                            {
                                if (scopeContexts[0].exception != NULL)
                                {
                                    *pException = scopeContexts[0].exception;
                                    DOFException_Destroy(scopeContexts[1].exception);
                                }
                                else if(scopeContexts[1].exception != NULL)
                                {
                                    *pException = scopeContexts[1].exception;
                                }
                            }
                            else
                            {
                                DOFException_Destroy(scopeContexts[0].exception);
                                DOFException_Destroy(scopeContexts[1].exception);
                            }
                            DOFSecurityScope_Destroy(scopeContexts[0].securityScope);
                            DOFSecurityScope_Destroy(scopeContexts[1].securityScope);
                            DOFPermissionSet_Destroy(scopeContexts[1].permissionSet);
                        }
                        else
                        {
                            context.exception = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
                        }
                        DOFPermissionSet_Destroy(scopeContexts[0].permissionSet);
                    }
                    else
                    {
                        context.exception = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
                    }
                }
            }
            else
            {
                context.exception = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            }

            if((context.exception != NULL) && (pException != NULL))
            {
                if (pException != NULL)
                {
                    *pException = context.exception;
                }
                else
                {
                    DOFException_Destroy(context.exception);
                }
            }
            PCRSemaphore_Destroy(context.semaphore);
        }
    }
    return bridge;
}

static void CreateBridge_ResolveCredsCallback(DOFCredentials   self,
                                              DOFException     exception,
                                              void           * context)
{
    CreateSecureBridgeContext fullContext = (CreateSecureBridgeContext) context;

    if(exception != NULL)
    {
        fullContext->resolvedCredentials = DOFCredentials_Copy(self);
        if(fullContext->resolvedCredentials != NULL)
        {
            fullContext->exception = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        }
    }
    else
    {
        fullContext->exception = DOFException_Copy(exception);
    }
    PCRSemaphore_Signal(fullContext->semaphore);
}

static boolean CreateBridge_ResolveCredentials(CreateSecureBridgeContext context,
                                               uint32                    timeout)
{
    DOFCredentials_Internal_BeginResolve(context->resolvedCredentials, context->dof, timeout,
                                         FALSE, context->channel, context->address, CreateBridge_ResolveCredsCallback, (void *) context);
    PCRSemaphore_Wait(context->semaphore);
    return (context->exception == NULL);
}

static void CreateBridge_ResolveScopeCallback(TRP_V1_Context   trp_context,
                                              void           * context)
{
    CreateSecureBridgeScopeContext fullContext = (CreateSecureBridgeScopeContext) context;

    if(trp_context != NULL)
    {
        if((trp_context->ScopeReturn->status == DOFERROR_OK) &&
           (trp_context->ScopeReturn->errorCode == 0))
        {
            fullContext->duration = trp_context->ScopeReturn->grantedDuration;
            fullContext->securityScope = DOFSecurityScope_Copy(trp_context->ScopeReturn->securityScope);
            if(fullContext->securityScope != NULL)
            {
                DOFPermissionSet newPerms = DOFPermissionSet_Copy(trp_context->ScopeReturn->permValidation);

                DOFPermissionSet_Destroy(fullContext->permissionSet);
                fullContext->permissionSet = newPerms;
            }
            if((fullContext->securityScope == NULL) || (fullContext->permissionSet == NULL))
            {
                fullContext->exception = DOFException_Copy(DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            }
        }
        else
        {
            fullContext->exception = DOFErrorException_Create((trp_context->ScopeReturn->status == DOFERROR_OK) ? trp_context->ScopeReturn->errorCode
                                                                                                                : trp_context->ScopeReturn->status, NULL);
        }
    }
    else
    {
        fullContext->exception = DOFException_Copy(DOFEXCEPTION_RESOLUTION_FAILED);
    }
    PCRSemaphore_Signal(fullContext->fullContext->semaphore);
}

static boolean CreateBridge_ResolveScope(CreateSecureBridgeScopeContext context,
                                         DOFObjectIDAuthentication      id,
                                         uint32                         timeout)
{
    DOFKeyRequest keyRequest;

    keyRequest = DOFKeyRequest_Create(context->fullContext->resolvedCredentials, DEFAULT_NONCE_SIZE, FALSE,
                                      context->fullContext->fixedPermissions, NULL, context->permissionSet);
    if(keyRequest != NULL)
    {
        TRP_V1_RequestSecurityScope(context->duration, keyRequest, id, CreateBridge_ResolveScopeCallback,
                                    context, context->channel, context->address, timeout);
        DOFKeyRequest_Destroy(keyRequest);
    }
    return (keyRequest != NULL);
}
