/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/cfg.h>
#include <dof/dof.h>
#include <dof/objectid.h>
#include <internal/dofinternal.h>
#include <internal/operationinternal.h>

extern PCRMutex ref_count_mutex;

typedef struct DOFConfig_t
{
    DOFObjectIDSource   sourceID;
    uint32              refCount;
    uint32              maxOperations;
    uint32              maxConnections;
    uint16              poolSize;
    uint8               nodeRank;
    struct
    {
        unsigned int isRouter       : 1;
        unsigned int doValidation   : 1;
        unsigned int error          : 1;

    } flags;
} DOFConfig_t;

typedef struct DOFState_t
{
    struct DOFConfig_t parent;
}DOFState_t;

typedef struct DOFConfigBuilder_t
{
    struct DOFConfig_t parent;
}DOFConfigBuilder_t;

static DOFConfigBuilder DOFConfigBuilder_Internal_Create( DOFObjectIDSource sourceID, uint32 maxOperations, uint32 maxConnections, uint16 poolSize, boolean isRouter, uint8 nodeRank, boolean doValidation )
{
    if( sourceID )
    {
        DOFConfigBuilder ret = (DOFConfigBuilder)DOFMem_Alloc( sizeof( DOFConfig_t ) );
        if( ret )
        {
            PCRMem_Set( ret, 0, sizeof( DOFConfig_t ) );
            ret->parent.sourceID = DOFObjectIDSource_Copy( sourceID );
            if( ret->parent.sourceID )
            {
                ret->parent.maxConnections = maxConnections;
                ret->parent.maxOperations = maxOperations;
                ret->parent.nodeRank = nodeRank;
                ret->parent.poolSize = poolSize;
                ret->parent.flags.isRouter = !!isRouter;
                ret->parent.flags.doValidation = !!doValidation;
                ret->parent.refCount = 1;
                return ret;
            }
            DOFMem_Free( ret );
        }
    }
    return NULL;
}

DOFConfigBuilder DOFConfigBuilder_Create( void )
{
    /* Create Default GUID source ID */
    GUID guid = DOF_GenerateGUID();
    DOFObjectID oid = DOFObjectID_Create( DOFOBJECTIDCLASS_GUID, sizeof(guid), guid.bytes );
    DOFObjectIDSource sourceID = DOFObjectIDSource_Create( oid );
    DOFObjectID_Destroy( oid );
    PCR_Assert( !DOFObjectID_IsWildcard( (DOFObjectID)sourceID ) );

    if( sourceID )
    {
        /* Default config values */
        DOFConfigBuilder ret = DOFConfigBuilder_Internal_Create( sourceID, DOF_UNLIMITED, DOF_UNLIMITED, 20, FALSE, DOFCONFIG_DEFAULT_RANK, TRUE );
        DOFObjectIDSource_Destroy( sourceID );
        return ret;
    }
    return NULL;
}

DOFConfigBuilder DOFConfigBuilder_Create_State( const DOFState state )
{
    if( state )
        return (DOFConfigBuilder)DOFConfigBuilder_Build( (DOFConfigBuilder)state );
    return NULL;
}

DOFConfigBuilder DOFConfigBuilder_Create_Config( const DOFConfig config )
{
    if( config )
        return (DOFConfigBuilder)DOFConfigBuilder_Build( (DOFConfigBuilder)config );
    return NULL;
}

boolean DOFConfigBuilder_SetSourceID( DOFConfigBuilder self, DOFObjectIDSource sourceID )
{
    if( self && sourceID )
    {
        DOFObjectIDSource_Destroy( self->parent.sourceID );
        self->parent.sourceID = DOFObjectIDSource_Copy( sourceID );
        if( self->parent.sourceID )
        {
            self->parent.flags.error = 0;
            return TRUE;
        }
        self->parent.flags.error = 1;
    }
    return FALSE;
}

boolean DOFConfigBuilder_SetMaxOperations( DOFConfigBuilder self, uint32 count )
{
    if( self )
    {
        self->parent.maxOperations = DOF_UNLIMITED;
        return TRUE;
    }
    return FALSE;
}

boolean DOFConfigBuilder_SetMaxConnections( DOFConfigBuilder self, uint16 count )
{
    if( self )
    {
        if ( count == 0 )
        {
            /* For SetMaxConnections, zero means unlimited. */
            return DOFConfigBuilder_SetConnectionLimit( self, DOF_UNLIMITED );
        }
        return DOFConfigBuilder_SetConnectionLimit( self, count );
    }
    return FALSE;
}

boolean DOFConfigBuilder_SetConnectionLimit( DOFConfigBuilder self, uint32 count )
{
    if( self )
    {
        self->parent.maxConnections = count;
        return TRUE;
    }
    return FALSE;
}

boolean DOFConfigBuilder_SetThreadPoolSize( DOFConfigBuilder self, uint16 count )
{
    if( self && count > 0 )
    {
        self->parent.poolSize = count;
        return TRUE;
    }
    return FALSE;
}

boolean DOFConfigBuilder_SetRouter( DOFConfigBuilder self, boolean isRouter )
{
    if( self )
    {
        self->parent.flags.isRouter = !!isRouter;
        return TRUE;
    }
    return FALSE;
}

boolean DOFConfigBuilder_SetRank( DOFConfigBuilder self, uint8 rank )
{
    if( self )
    {
        self->parent.nodeRank = rank;
        return TRUE;
    }
    return FALSE;
}

boolean DOFConfigBuilder_SetParameterValidation( DOFConfigBuilder self, boolean validation )
{
    if( self )
    {
        self->parent.flags.doValidation = !!validation;
        return TRUE;
    }
    return FALSE;
}

DOFConfig DOFConfigBuilder_Build( DOFConfigBuilder self )
{
    if( self && !self->parent.flags.error )
        return DOFConfigBuilder_BuildAndDestroy( DOFConfigBuilder_Internal_Create( self->parent.sourceID, self->parent.maxOperations, self->parent.maxConnections, self->parent.poolSize, self->parent.flags.isRouter, self->parent.nodeRank, self->parent.flags.doValidation ) );
    return NULL;
}

DOFConfig DOFConfigBuilder_BuildAndDestroy( DOFConfigBuilder self )
{
    DOFConfig ret = (DOFConfig)self;

    if( ret && ret->flags.error )
    {
        DOFConfigBuilder_Destroy( self );
        ret = NULL;
    }
    return ret;
}

void DOFConfigBuilder_Destroy( DOFConfigBuilder self )
{
    if( self )
    {
        boolean free_ok = FALSE;

        PCRMutex_Lock(ref_count_mutex);
        if( !( --(self->parent.refCount) ) )
            free_ok = TRUE;
        PCRMutex_Unlock(ref_count_mutex);

        if( free_ok )
        {
            DOFObjectIDSource_Destroy( self->parent.sourceID );
            DOFMem_Free( self );
        }
    }
}

DOFConfig DOFConfig_Copy( DOFConfig self )
{
    if( self )
    {
        PCRMutex_Lock(ref_count_mutex);
        if( self->refCount < 0xFFFFFFFF )
        {
            ++self->refCount;
            PCRMutex_Unlock(ref_count_mutex);
            return self;
        }
        PCRMutex_Unlock(ref_count_mutex);
    }
    return NULL;
}

void DOFConfig_Destroy( DOFConfig self )
{
    DOFConfigBuilder_Destroy( (DOFConfigBuilder)self );
}

DOFObjectIDSource DOFConfig_GetSourceID( DOFConfig self )
{
    if( self )
        return self->sourceID;
    return NULL;
}

uint32 DOFConfig_GetMaxOperations( DOFConfig self )
{
    /* This function was deprecated and will always return 0. */
    return 0;
}

uint16 DOFConfig_GetMaxConnections( DOFConfig self )
{
    if( self )
    {
        uint32 max;

        max = DOFConfig_GetConnectionLimit( self );
        if ( max == 0 )
        {
            /* With the Set/Get Connection Limit functions zero internally means zero, and this
               function defines 0 as unlimited.  Return 1 since that is the closest to zero. */
            return 1;
        }
        if ( max > 0xFFFF )
        {
            return 0xFFFF;
        }
        return max;
    }
    return 0;
}

uint32 DOFConfig_GetConnectionLimit( DOFConfig self )
{
    if( self )
        return self->maxConnections;
    return 0;
}

uint16 DOFConfig_GetThreadPoolSize( DOFConfig self )
{
    if( self )
        return self->poolSize;
    return 0;
}

boolean DOFConfig_IsRouter( DOFConfig self )
{
    if( self )
        return self->flags.isRouter;
    return FALSE;
}

uint8 DOFConfig_GetRank( DOFConfig self )
{
    if( self )
        return self->nodeRank;
    return 0;
}

boolean DOFConfig_IsParameterValidation( DOFConfig self )
{
    if( self )
        return self->flags.doValidation;
    return TRUE; /* Safest, enforces validation */
}

DOFState DOFState_Create( DOF dof )
{
    DOFConfigBuilder cb;

    DOF_Lock();

    cb = (DOFConfigBuilder)DOF_GetConfig(dof);
    DOFConfigBuilder_SetConnectionLimit(cb, dof->maxConn);

    DOF_Unlock();

    return (DOFState)DOFConfigBuilder_Build( cb );
}

void DOFState_Destroy( DOFState self )
{
    DOFConfig_Destroy( (DOFConfig)self );
}

DOFObjectIDSource DOFState_GetSourceID( DOFState self )
{
    if( self )
        return self->parent.sourceID;
    return NULL;
}

uint32 DOFState_GetMaxOperations( DOFState self )
{
    /* This function was deprecated and will always return DOF_UNLIMITED. */
    return DOF_UNLIMITED;
}

uint16 DOFState_GetMaxConnections( DOFState self )
{
    if( self )
    {
        uint32 max;

        max = DOFState_GetConnectionLimit(self);
        if ( max == 0 )
        {
            /* With the Set/Get Connection Limit functions zero internally means zero, and this
               function defines 0 as unlimited.  Return 1 since that is the closest to zero. */
            return 1;
        }
        if ( max > 0xFFFF )
        {
            return max;
        }
        return max;
    }
    return 0;
}

uint32 DOFState_GetConnectionLimit( DOFState self )
{
    if( self )
    {
        return self->parent.maxConnections;
    }
    return 0;
}

uint16 DOFState_GetThreadPoolSize( DOFState self )
{
    if( self )
        return self->parent.poolSize;
    return 0;
}

boolean DOFState_IsRouter( DOFState self )
{
    if( self )
        return self->parent.flags.isRouter;
    return FALSE;
}

uint8 DOFState_GetRank( DOFState self )
{
    if( self )
        return self->parent.nodeRank;
    return 0;
}

boolean DOFState_IsParameterValidation( DOFState self )
{
    if( self )
        return self->parent.flags.doValidation;
    return TRUE; /* Safest, enforces validation */
}
