/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/dof.h>
#include <internal/dofinternal.h>
#include <internal/operationinternal.h>
#include <internal/slist.h>
#include <internal/enqueue.h>

static DOF dof = NULL;

typedef struct DOFInterestListenerNode_t* DOFInterestListenerNode;
typedef struct DOFInterestListenerRemoveNode_t* DOFInterestListenerRemoveNode;

typedef struct DOFInterestListenerRemoveNode_t
{
    OALEnqueueItem_t            base;
    DOFInterestListener         listener;
    DOFException                exception;
    DOFInterestListenerNode     node;
}DOFInterestListenerRemoveNode_t;

typedef struct DOFInterestListenerNode_t
{
    OALSListNode                    next;
    DOFInterestListenerRemoveNode_t self;
}DOFInterestListenerNode_t;

static boolean DOFInterestListenerNodeCompareGreaterThan(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    return ( ((DOFInterestListener)key) > ((DOFInterestListenerNode)item)->self.listener );
}

static boolean DOFInterestListenerNodeCompareEqual(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    return ( ((DOFInterestListener)key) == ((DOFInterestListenerNode)item)->self.listener );
}

boolean DOF_InterestListeners_Init( DOF self )
{
    dof = self;
    self->interestListeners = OALAVLTree_Create( DOFInterestListenerNodeCompareGreaterThan, DOFInterestListenerNodeCompareEqual, NULL );
    if( self->interestListeners )
    {
        return TRUE;
    }
    return FALSE;
}

static void DOFInterestListernerNode_EnqueueRemoved( DOFInterestListenerRemoveNode node );

static void InterestListenerRemoveAll(OALAVLTreeItem_t item, void *arg)
{
    DOFInterestListenerNode node = (DOFInterestListenerNode)item;
    node->self.exception = arg;
    DOFInterestListernerNode_EnqueueRemoved( &(node->self) );
}

void DOF_InterestListeners_Shutdown( DOF self )
{
    OALAVLTree_Destroy( self->interestListeners, InterestListenerRemoveAll, NULL );
}

typedef struct InterestContext_t
{
    DOF                 self;
    DOFOperation        reqOp;
    DOFInterestLevel    level;
    OALSListNode        head;
    boolean             added;
}InterestContext_t;

typedef struct EnqueuedInterestListenerCall_t
{
    OALEnqueueItem_t        base;
    DOF                     self;
    DOFOperation            reqOp;
    DOFInterestLevel        level;
    DOFInterestListenerNode node;
    boolean                 added;
}EnqueuedInterestListenerCall_t;

static void callInterestListener(OALThreadPoolItem item, boolean shutdown)
{
    EnqueuedInterestListenerCall_t* node = (EnqueuedInterestListenerCall_t*)item;

    if( node->added )
        node->node->self.listener->fns->InterestAdded( node->node->self.listener, node->self, node->reqOp->sources->request, node->reqOp->object_id.object_id, node->reqOp->interface_id, node->level );
    else
        node->node->self.listener->fns->InterestRemoved( node->node->self.listener, node->self, node->reqOp->sources->request, node->reqOp->object_id.object_id, node->reqOp->interface_id, node->level );

    DOFOperation_Internal_Destroy( node->reqOp );
    DOFMem_Free( node );
}

static boolean NotifyDOFInterestListeners( OALAVLTreeItem_t item, void *arg )
{
    DOFInterestListenerNode node = (DOFInterestListenerNode)item;
    InterestContext_t* ctx = (InterestContext_t*)arg;

    if( ( ctx->added && node->self.listener->fns->InterestAdded ) ||
        ( !ctx->added && node->self.listener->fns->InterestRemoved ) )
    {
        EnqueuedInterestListenerCall_t* enqueueCallCtx = (EnqueuedInterestListenerCall_t*)DOFMem_Alloc( sizeof( EnqueuedInterestListenerCall_t ) );
        if( enqueueCallCtx )
        {
            DOFOperation_Retain( ctx->reqOp );
            enqueueCallCtx->added = ctx->added;
            enqueueCallCtx->level = ctx->level;
            enqueueCallCtx->reqOp = ctx->reqOp;
            enqueueCallCtx->self = ctx->self;
            enqueueCallCtx->node = node;

            OALEnqueueItem_Init( (OALEnqueueItem)enqueueCallCtx, callInterestListener );
            OALEnqueue_Enqueue( ctx->self->enqueue, (OALEnqueueItem)enqueueCallCtx );
            return FALSE;
        }
    }
    OALSList_Insert( &(ctx->head), (OALSListNode)node );
    return FALSE;
}

static void RemoveAllFailedToNotify( OALSListNode item, void *arg )
{
    DOFInterestListenerNode node = (DOFInterestListenerNode)item;
    OALAVLTree_Delete( dof->interestListeners, node->self.listener );
    DOFInterestListernerNode_EnqueueRemoved( &(node->self) );
}

void DOF_ProcessInterest( DOF self, DOFOperation reqOp )
{
    if( self && reqOp ) 
    {
        InterestContext_t ctx;

        DOFOperation_Retain( reqOp );
        ctx.added = (reqOp->flags.bits.activation_closing == 0); /* Added or cancelled */
        ctx.reqOp = reqOp;
        ctx.head = NULL;
        ctx.self = self;

        switch( reqOp->operation_type )
        {
        case DOFOPERATIONTYPE_WATCH:
            ctx.level = DOFINTERESTLEVEL_WATCH;
            break;
        case DOFOPERATIONTYPE_ACTIVATE:
            ctx.level = DOFINTERESTLEVEL_ACTIVATE;
            break;
        case DOFOPERATIONTYPE_CONNECT:
            /* Fall Through */
        case DOFOPERATIONTYPE_FULLCONNECT:
            ctx.level = DOFINTERESTLEVEL_CONNECT;
            break;
        default:
            return;
            break;
        }

        DOF_Lock();
        OALAVLTree_Traverse( self->interestListeners, NotifyDOFInterestListeners, &ctx );

        if( ctx.head )
            OALSList_RemoveAll( &(ctx.head), RemoveAllFailedToNotify, NULL );
        DOF_Unlock();

        DOFOperation_Internal_Destroy( reqOp );
    }
}

boolean DOF_AddInterestListener( DOF self, DOFInterestListener listener )
{
    if( self && listener )
    {
        DOFInterestListenerNode node = (DOFInterestListenerNode)DOFMem_Alloc( sizeof( DOFInterestListenerNode_t ) );
        if( node )
        {
            OALAVLTree_InsertionStatus status;

            PCRMem_Set( node, 0, sizeof( DOFInterestListenerNode_t ) );
            node->self.node = node;
            node->self.listener = listener;

            DOF_Lock();
            status = OALAVLTree_Insert( self->interestListeners, listener, node, NULL );
            DOF_Unlock();

            switch( status )
            {
            case OALAVLTree_AlreadyInserted:
                DOFMem_Free( node );
                /* Fall through */
            case OALAVLTree_InsertSuccess:
                return TRUE;
                break;
            case OALAVLTree_InsertFailure:
            default:
                DOFMem_Free( node );
                break;
            }
        }
    }
    return FALSE;
}

static void InterestListenerNodeRemoved(OALThreadPoolItem item, boolean shutdown)
{
    DOFInterestListenerRemoveNode node = (DOFInterestListenerRemoveNode)item;
    if( node->listener->fns->Removed )
        node->listener->fns->Removed( node->listener, dof, node->exception);
    DOFMem_Free( node->node ); 
}

static void DOFInterestListernerNode_EnqueueRemoved( DOFInterestListenerRemoveNode node )
{
    OALEnqueueItem_Init( (OALEnqueueItem)node, InterestListenerNodeRemoved );
    OALEnqueue_Enqueue( dof->enqueue, (OALEnqueueItem)node );
}

void DOF_RemoveInterestListener( DOF self, DOFInterestListener listener )
{
    if( self && listener )
    {
        DOFInterestListenerNode node;

        DOF_Lock();
        node = (DOFInterestListenerNode)OALAVLTree_Retrieve( self->interestListeners, listener );
        if( node )
        {
            if( OALAVLTree_Delete( self->interestListeners, listener ) )
            {
                DOF_Unlock();
                DOFInterestListernerNode_EnqueueRemoved( &(node->self) );
                return;
            }
            else
                PCR_Assert(FALSE); /* We just found it! Why wasn't it removed? */
        }
        DOF_Unlock();
    }
}
