/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/dof.h>
#include <dof/log.h>
#include <dof/oaltypes.h>
#include <dof/type.h>
#include <dof/value.h>
#include <dof/value_gen.h>
#include <internal/dofinternal.h>
#include <internal/slist.h>

#define LOGGER_THREADS 1


typedef struct DOFLogListenerNode_t
{
    OALSListNode    next;
    DOFLogLevel     priority;
    DOFLogListener  callback;
    OALEnqueue      enqueue;
}DOFLogListenerNode_t;


static PCRMutex mutex = NULL;
static OALThreadPool threadpool = NULL;
static DOFLogLevel activeLevels = 0;
static uint32 nlisteners= 0;
static OALSListNode loglisteners = NULL;


void DOFLogListener_Init( void )
{
    mutex = PCRMutex_Create();
    PCR_Assert( mutex );
}

typedef struct loglistener_t
{
    uint32      refCount;
    uint32      timestamp;
    DOFLogLevel priority;
    char*       message;
}loglistener_t;
typedef struct loglistener_t* loglistener;

typedef struct LogContext_t
{
    OALEnqueueItem_t super;
    DOFLogListener log;
    loglistener aLog;
    boolean isLog;
} LogContext_t;

static loglistener loglistener_Create( DOFLogLevel priority, const char * prefix, const char * message )
{
    uint32 plen = (prefix)? PCRString_Length( prefix ):0;
    uint32 mlen = PCRString_Length( message );
    uint32 messageLength = plen + mlen + ((plen)?3:1);
    loglistener ret = DOFMem_Alloc( sizeof( loglistener_t ) + messageLength );

    if( ret )
    {
        PCRMem_Set(ret,0, sizeof( loglistener_t ) + messageLength );
        ret->priority = priority;
        ret->timestamp = PCRTime_GetMS();
        ret->refCount = 1;
        ret->message = (char*)( ((uint8*)ret) + sizeof( loglistener_t ) );

        if( plen )
        {
            PCRMem_Copy( ret->message, prefix, plen );
            ret->message[ plen ] = 58; /* colon */
            ret->message[ plen+1 ] = 32; /* space */
        }
        PCRMem_Copy( (ret->message + ((plen)?plen+2:0)), message, mlen );
        ret->message[ messageLength-1 ] = 0;
    }
    return ret;
}

static loglistener loglistener_Copy( loglistener self )
{
    if(self)
    {
        loglistener ret = self;
        PCRMutex_Lock(ref_count_mutex);
        if( self->refCount < 0xFFFFFFFF )
            ++(self->refCount);
        else
            ret = NULL;
        PCRMutex_Unlock(ref_count_mutex);
        return ret;
    }
    return NULL;
}

static void loglistener_Destroy( loglistener self )
{
    if(self)
    {
        boolean freeable = FALSE;
        PCRMutex_Lock(ref_count_mutex);
        PCR_Assert( self->refCount > 0 );
        (self->refCount)--;
        if( !(self->refCount))
            freeable = TRUE;
        PCRMutex_Unlock(ref_count_mutex);

        if( freeable )
        {
            DOFMem_Free( self );
        }
    }
}

static void callLogListener( OALThreadPoolItem dnode, boolean shutdown )
{
    LogContext_t * node = (LogContext_t*)dnode;

    PCR_Assert(dnode != NULL);
    if ( !node->isLog )
    {
        if ( node->log->fns->Removed )
            node->log->fns->Removed( node->log, NULL );
    }
    else
    {
        if ( node->log->fns->LogMessage )
            node->log->fns->LogMessage( node->log, node->aLog->timestamp, node->aLog->priority, node->aLog->message );
        loglistener_Destroy( node->aLog );
    }
    DOFMem_Free(node);
}

static void internalRemoveLogListener( DOFLogListenerNode_t* listener )
{
    /* Called with mutex held. */
    LogContext_t * node = (LogContext_t*) DOFMem_Alloc(sizeof(LogContext_t));

    if(node != NULL)
    {
        OALEnqueueItem_Init((OALEnqueueItem) node, callLogListener);
        node->log = listener->callback;
        node->isLog = FALSE;
        PCR_Assert( listener->enqueue );
        if(!OALEnqueue_Enqueue( listener->enqueue, (OALEnqueueItem) node ))
        {
            DOFMem_Free(node);
        }
        OALEnqueue_Destroy( listener->enqueue );
        DOFMem_Free( listener );
        --nlisteners;
    }
}

static void internalRemoveLogListenerSList(OALSListNode node, void* arg)
{
    internalRemoveLogListener((DOFLogListenerNode_t*)node);
}

void DOFLogListener_Shutdown( void )
{
    OALThreadPool pool;
    PCRMutex_Lock( mutex );
    OALSList_RemoveAll( &loglisteners, internalRemoveLogListenerSList, NULL );
    pool = threadpool;
    threadpool = NULL;
    activeLevels = 0;
    PCR_Assert(  nlisteners == 0 );
    PCR_Assert( loglisteners == NULL );
    PCRMutex_Unlock( mutex );
    if ( pool )
        OALThreadPool_Destroy( pool );
    PCRMutex_Destroy( mutex );
    mutex = NULL;
}

boolean DOFLog_AddListener( DOFLogLevel priority, DOFLogListener callback )
{
    if ( callback )
    {
        OALThreadPool pool = NULL;

        PCRMutex_Lock( mutex );
        PCR_Assert( ( nlisteners > 0 ) ^ ( threadpool == NULL ) );
        if ( threadpool || ( threadpool = OALThreadPool_Create( LOGGER_THREADS ) ) != NULL )
        {
            DOFLogListenerNode_t* logListener = (DOFLogListenerNode_t *) DOFMem_Alloc( sizeof( DOFLogListenerNode_t ) );
            if ( logListener )
            {
                logListener->enqueue = OALEnqueue_Create( threadpool );
                if ( logListener->enqueue )
                {
                    logListener->callback = callback;
                    logListener->priority = priority;
                    logListener->next = NULL;

                    OALSList_Insert( &loglisteners, (OALSListNode)logListener);
                    ++nlisteners;
                    activeLevels |= priority;
                    PCRMutex_Unlock( mutex );
                    return TRUE;
                }
                DOFMem_Free( logListener );
            }
            if ( !nlisteners )
            {
                pool = threadpool;
                threadpool = NULL;
            }
        }
        PCRMutex_Unlock( mutex );
        if ( pool )
            OALThreadPool_Destroy( pool );
    }
    return FALSE;
}

static boolean findLogListener( OALSListNode itemx, void *argx )
{
    DOFLogListenerNode_t* item = (DOFLogListenerNode_t*)itemx;
    DOFLogListener arg = (DOFLogListener)argx;
    /* Called with mutex held. */
    return ( item->callback == arg );
}

static boolean logLevelTraverse( OALSListNode itemx, void *argx )
{
    DOFLogListenerNode_t* item = (DOFLogListenerNode_t*)itemx;
    DOFLogLevel* arg = (DOFLogLevel*)argx;
    /* Called with mutex held. */
    *arg |= item->priority;
    return FALSE;
}

void DOFLog_RemoveListener( DOFLogListener callback )
{
    if ( callback )
    {
        DOFLogListenerNode_t* logListener;
        OALThreadPool pool = NULL;

        PCRMutex_Lock( mutex );
        logListener = (DOFLogListenerNode_t*) OALSList_FindAndRemove( &loglisteners, NULL, findLogListener, callback );
        if ( logListener )
            internalRemoveLogListener( logListener );
        activeLevels = 0;
        if ( !nlisteners )
        {
            pool = threadpool;
            threadpool = NULL;
        }
        else
        {
            OALSList_Traverse( loglisteners, logLevelTraverse, (void*)&activeLevels);
        }
        PCRMutex_Unlock( mutex );
        if ( pool )
            OALThreadPool_Destroy( pool );
    }
}

boolean DOFLog_IsLogLevel( DOFLogLevel priority )
{
    boolean ret;
    PCRMutex_Lock( mutex );
    ret = ( activeLevels & priority ) != 0;
    PCRMutex_Unlock( mutex );
    return ret;
}

static boolean logListenerTraverse( OALSListNode itemx, void* argx )
{
    DOFLogListenerNode_t* item = (DOFLogListenerNode_t*)itemx;
    struct loglistener_t* arg = (struct loglistener_t*)argx;

    /* Called with mutex held. */
    if ( item->priority & arg->priority )
    {
        LogContext_t * node = (LogContext_t*) DOFMem_Alloc(sizeof(LogContext_t));

        if(node != NULL)
        {
            OALEnqueueItem_Init((OALEnqueueItem) node, callLogListener);

            node->aLog = loglistener_Copy(arg);
            if(node->aLog)
            {
                node->log = item->callback;
                node->isLog = TRUE;
                if(OALEnqueue_Enqueue( item->enqueue, (OALEnqueueItem) node))
                {
                    return FALSE; /* Continue the SList traverse */
                }
                loglistener_Destroy(node->aLog);
            }
            DOFMem_Free(node);
        }
    }
    return FALSE; /* Continue the SList traverse */
}

static void DOFLog_Internal_Message( DOFLogLevel priority, const char * prefix, const char * message )
{
    PCRMutex_Lock( mutex );
    if ( activeLevels & priority )
    {
        loglistener logl = loglistener_Create( priority, prefix, message );

        if( logl )
        {
            OALSList_Traverse( loglisteners, logListenerTraverse, logl );
        }
        loglistener_Destroy( logl );
    }
    PCRMutex_Unlock( mutex );
}

void DOFLog_Message( DOFLogLevel priority, const char * message )
{
    DOFLog_Internal_Message( priority, NULL, message );
}

void DOF_LogMessage( DOF self, DOFLogLevel priority, const char * message )
{
    PCR_Assert(self);
    DOFLog_Internal_Message( priority, "DOF", message );
}

