/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/domain.h>
#include <dof/authenticator.h>
#include <internal/domaininternal.h>
#include <dof/oaltypes.h>
#include <dof/dof.h>
#include <dof/credentials.h>
#include <dof/objectid.h>
#include <internal/dofinternal.h>
#include <internal/trp_v1.h>
#include <internal/exception.h>

#define DOFAUTHENTICATOR_DEFAULT_MAX_LIFETIME 0xB4 /* 24 hours */

extern PCRMutex ref_count_mutex;

#define DOFAUTHENTICATORCONFIGBUILDER_ERROR_DOMAINID    0x1
#define DOFAUTHENTICATORCONFIGBUILDER_ERROR_DEBUGKEY    0x2

typedef struct DOFAuthenticatorConfig_t
{
    DOFObjectIDDomain       domainID;
    DOFAuthenticatorStorage storage;
    uint8*                  debugKey;
    uint8                   nodeID;
    uint8                   maxLifetime;
    uint8                   flags;
    boolean                 isDebug;
}DOFAuthenticatorConfig_t;

typedef struct DOFAuthenticatorConfigBuilder_t
{
    DOFAuthenticatorConfig_t   config;
}DOFAuthenticatorConfigBuilder_t;

typedef struct DOFAuthenticator_t
{
    DOFAuthenticatorConfig_t   config;
    DOF                        dof;
    uint32                     refCount;
} DOFAuthenticator_t;

#define DOFAUTHENTICATORSTATE_FLAG_ISDEBUG      0x1
#define DOFAUTHENTICATORSTATE_FLAG_ISAVALILABLE 0x2

typedef struct DOFAuthenticatorState_t
{
    DOFObjectIDDomain       domainID;
    uint8                   nodeID;
    uint8                   maxLifetime;
    uint8                   flags;
}DOFAuthenticatorState_t;

static boolean DOFAuthenticatorConfig_SetDebugKey(DOFAuthenticatorConfig self, uint8* debugKey )
{
    PCR_Assert( self );

    if( debugKey )
    {
        PCR_Assert( self->debugKey == NULL );
        self->debugKey = DOFMem_Alloc( DOF_SECURITY_KEY_LENGTH );
        if( !self->debugKey )
            return FALSE;

        PCRMem_Copy( self->debugKey, debugKey, DOF_SECURITY_KEY_LENGTH );
    }
    return TRUE;
}

static boolean DOFAuthenticatorConfig_CopyInit(DOFAuthenticatorConfig self, DOFAuthenticatorConfig from)
{
    PCR_Assert(from && self);

    self->domainID = DOFObjectIDDomain_Copy( from->domainID );
    if(self->domainID)
    {
        self->nodeID = from->nodeID;
        self->storage = from->storage;
        self->maxLifetime = from->maxLifetime;
        self->isDebug = from->isDebug;
        if( DOFAuthenticatorConfig_SetDebugKey( self, from->debugKey ) )
        {
            return TRUE;
        }
        DOFObjectIDDomain_Destroy(self->domainID);
        self->domainID = NULL;
    }
    return FALSE;
}

DOFAuthenticator DOFAuthenticator_CreateInternal( DOF dof, DOFAuthenticatorConfig config )
{
    DOFAuthenticator ret = (DOFAuthenticator_t*)DOFMem_Alloc(sizeof(DOFAuthenticator_t));
    
    if( ret )
    {
        PCRMem_Set( ret, 0, sizeof(DOFAuthenticator_t) );
        if( DOFAuthenticatorConfig_CopyInit( (&ret->config), config) )
        {
            ret->refCount = 1;
            ret->dof = dof;
            return ret;
        }
        DOFMem_Free(ret);
    }
    return NULL;
}

void DOFAuthenticator_DOFTriggeredDestroy( DOFAuthenticator self )
{
    self->config.storage = NULL;
}

void DOFAuthenticator_Destroy( DOFAuthenticator self )
{
    uint32  ref;
    if( !self )
        return;

    PCRMutex_Lock( ref_count_mutex );
    --(self->refCount);
    ref = self->refCount;
    PCRMutex_Unlock( ref_count_mutex );

    if( !ref )
    {
        DOF_RemoveAuthenticator( DOF_GetDof(), self );
        TRPRouter_RemoveAuthenticator( self );
        DOFObjectIDDomain_Destroy( self->config.domainID );
        DOFMem_Free(self->config.debugKey);
        DOFMem_Free( self );
    }
}

void DOFAuthenticatorList_Destroy( DOFAuthenticatorList self )
{
    if( self )
    {
        uint32 i = 0;

        for( ; i < self->count; ++i )
        {
            DOFAuthenticator_Destroy(self->list[i]);
        }
        DOFMem_Free(self);
    }
}

void DOFAuthenticatorConfig_Destroy( DOFAuthenticatorConfig self )
{
    if( self )
    {
        DOFObjectIDDomain_Destroy( self->domainID );
        DOFMem_Free(self->debugKey);
        DOFMem_Free( self );
    }
}

DOFAuthenticator DOFAuthenticator_Copy( const DOFAuthenticator self )
{
    DOFAuthenticator ret = NULL;
    if(!self)
        return NULL;

    PCRMutex_Lock(ref_count_mutex);
    if( self->refCount < 0xFFFFFFFF )
    {
        self->refCount++;
        ret = self;
    }
    PCRMutex_Unlock(ref_count_mutex);

    return ret;
}

DOFAuthenticatorStorage DOFAuthenticator_GetStorage(const DOFAuthenticator self)
{
    if( self )
        return self->config.storage;
    return NULL;
}

uint8 DOFAuthenticator_GetID(const DOFAuthenticator self)
{
    if( self )
        return self->config.nodeID;
    return 0;
}

uint8 DOFAuthenticator_GetMaxLifetime( const DOFAuthenticator self )
{
    PCR_Assert(self);
    return self->config.maxLifetime;
}

DOFObjectIDDomain DOFAuthenticator_GetDomainID( const DOFAuthenticator self )
{
    if( self )
        return self->config.domainID;
    return NULL;
}

boolean DOFAuthenticator_IsDebug( DOFAuthenticator self )
{
    if( self )
        return self->config.isDebug;
    return FALSE;
}

uint8* DOFAuthenticator_GetDebugKey( DOFAuthenticator self )
{
    if( self )
        return self->config.debugKey;
    return NULL;
}

uint8* DOFAuthenticator_CreateSessionKey( void )
{
    uint8* ret = PCRMem_Alloc(32);

    if( ret )
        PCRSecureRandom_Generate( 32, ret );

    return ret;
}

DOFAuthenticatorConfigBuilder DOFAuthenticatorConfigBuilder_Create( DOFObjectIDDomain domain, uint8 id, DOFAuthenticatorStorage storage )
{
    DOFAuthenticatorConfigBuilder ret = NULL;
    if( domain && !DOFObjectID_IsBroadcast( DOFObjectIDDomain_GetObjectID( domain ) ) )
    {
        ret = (DOFAuthenticatorConfigBuilder_t*)DOFMem_Alloc( sizeof(DOFAuthenticatorConfigBuilder_t) );
        if( ret )
        {
            PCRMem_Set( ret, 0, sizeof( DOFAuthenticatorConfigBuilder_t ) );
            ret->config.domainID = DOFObjectIDDomain_Copy( domain );
            if( ret->config.domainID )
            {
                ret->config.storage = storage;
                ret->config.nodeID = id;
                ret->config.maxLifetime = DOFAUTHENTICATOR_DEFAULT_MAX_LIFETIME;
                return ret;
            }
            DOFMem_Free(ret);
        }
    }
    return NULL;
}

boolean DOFAuthenticatorConfigBuilder_SetStorage( DOFAuthenticatorConfigBuilder self, DOFAuthenticatorStorage storage )
{
    if( self && storage )
    {
        self->config.storage = storage;
        return TRUE;
    }
    return FALSE;
}

boolean DOFAuthenticatorConfigBuilder_SetMaxGrantLifetime( DOFAuthenticatorConfigBuilder self, uint16 lifetime )
{
    if( self )
    {
        self->config.maxLifetime = (lifetime<=8)?1:((lifetime+7)/8); /* insure liftime is greater than 0 and round up;*/
        return TRUE;
    }
    return FALSE;
}

DOFAuthenticatorConfig DOFAuthenticatorConfigBuilder_Build( DOFAuthenticatorConfigBuilder self )
{
    DOFAuthenticatorConfig ret;

    if( !self || self->config.flags || !self->config.storage )
        return NULL;

    ret = (DOFAuthenticatorConfig_t*)DOFMem_Alloc( sizeof(DOFAuthenticatorConfig_t) );

    if( ret )
    {
        PCRMem_Set( ret, 0, sizeof(DOFAuthenticatorConfig_t) );
        if( DOFAuthenticatorConfig_CopyInit( ret, (&self->config)) )
            return ret;
        DOFMem_Free( ret );
    }

    return NULL;
}

DOFAuthenticatorConfig DOFAuthenticatorConfigBuilder_BuildAndDestroy( DOFAuthenticatorConfigBuilder self )
{
    DOFAuthenticatorConfig ret = (DOFAuthenticatorConfig)self;
    if( ret && (ret->flags || !ret->storage) )
    {
        DOFAuthenticatorConfig_Destroy( ret );
        ret = NULL;
    }
    return ret;
}

void DOFAuthenticatorConfigBuilder_Destroy( DOFAuthenticatorConfigBuilder self )
{
    if( self )
    {
        DOFObjectIDDomain_Destroy( self->config.domainID );
        DOFMem_Free(self->config.debugKey);
        DOFMem_Free( self );
    }
}

uint16 DOFAuthenticatorConfig_GetMaxGrantLifetime( const DOFAuthenticatorConfig self )
{
    PCR_Assert(self);
    return ((self->maxLifetime)*8);
}

uint8 DOFAuthenticatorConfig_GetNodeIdentifier( const DOFAuthenticatorConfig self )
{
    PCR_Assert(self);
    return self->nodeID;
}

DOFObjectIDDomain DOFAuthenticatorConfig_GetDomainID( const DOFAuthenticatorConfig self )
{
    PCR_Assert(self);
    return self->domainID;
}

boolean DOFAuthenticatorConfig_IsDebug( const DOFAuthenticatorConfig self )
{
    PCR_Assert(self);
    return self->isDebug;
}

const uint8 * DOFAuthenticatorConfig_GetDebugKey( const DOFAuthenticatorConfig self )
{
    PCR_Assert(self);
    return self->debugKey;
}

boolean DOFAuthenticatorConfigBuilder_SetDebug( DOFAuthenticatorConfigBuilder self, boolean isDebug )
{
    if( self )
    {
        self->config.isDebug = isDebug;
        return TRUE;
    }
    return FALSE;
}

boolean DOFAuthenticatorConfigBuilder_SetDebugKey( DOFAuthenticatorConfigBuilder self, const uint8 key[DOF_SECURITY_KEY_LENGTH] )
{
    if( self )
    {
        if( self->config.debugKey == NULL )
        {
            self->config.debugKey = DOFMem_Alloc(DOF_SECURITY_KEY_LENGTH);
            if( self->config.debugKey == NULL )
            {
                self->config.flags |= DOFAUTHENTICATORCONFIGBUILDER_ERROR_DEBUGKEY;
                return FALSE;
            }
        }
        PCRMem_Copy(self->config.debugKey, key, DOF_SECURITY_KEY_LENGTH);
        self->config.flags &= ~DOFAUTHENTICATORCONFIGBUILDER_ERROR_DEBUGKEY;
        return TRUE;
    }
    return FALSE;
}

boolean DOFAuthenticatorConfigBuilder_SetNodeIdentifier( DOFAuthenticatorConfigBuilder self, uint8 id )
{
    if( self && ( id < 64 ) )
    {
        self->config.nodeID = id;
        return TRUE;
    }
    return FALSE;
}

boolean DOFAuthenticatorConfigBuilder_SetDomainID( DOFAuthenticatorConfigBuilder self, DOFObjectIDDomain domainID )
{
    if( self && domainID )
    {
        DOFObjectIDDomain_Destroy( self->config.domainID );
        self->config.domainID = DOFObjectIDDomain_Copy(domainID);
        if( self->config.domainID )
        {
            self->config.flags &= ~DOFAUTHENTICATORCONFIGBUILDER_ERROR_DOMAINID;
            return TRUE;
        }
        self->config.flags |= DOFAUTHENTICATORCONFIGBUILDER_ERROR_DOMAINID;
    }
    return FALSE;
}

DOFAuthenticatorState DOFAuthenticator_CreateState( DOFAuthenticator auth )
{
    if( auth )
    {
        DOFAuthenticatorState ret = (DOFAuthenticatorState)DOFMem_Alloc( sizeof( DOFAuthenticatorState_t ) );
        if( ret )
        {
            PCRMem_Set( ret, 0, sizeof( DOFAuthenticatorState_t ) );

            ret->domainID = DOFObjectIDDomain_Copy( auth->config.domainID );
            if( ret->domainID )
            {
                ret->nodeID = auth->config.nodeID;
                ret->maxLifetime = auth->config.maxLifetime;

                if( auth->config.isDebug )
                    ret->flags |= DOFAUTHENTICATORSTATE_FLAG_ISDEBUG;

                if( auth->config.storage && auth->config.storage->fns->IsStorageAvailable( auth->config.storage ) )
                    ret->flags |= DOFAUTHENTICATORSTATE_FLAG_ISAVALILABLE;

                return ret;
            }
            DOFMem_Free( ret );
        }
    }
    return NULL;
}

DOFAuthenticatorConfigBuilder DOFAuthenticatorConfigBuilder_Create_State( DOFAuthenticatorState state )
{
    if( state )
    {
        DOFAuthenticatorConfigBuilder ret = DOFAuthenticatorConfigBuilder_Create( state->domainID, state->nodeID, NULL );
        if( ret )
        {
            if( DOFAuthenticatorConfigBuilder_SetMaxGrantLifetime( ret, state->maxLifetime ) )
            {
                if( DOFAuthenticatorConfigBuilder_SetDebug( ret, DOFAuthenticatorState_IsDebug( state ) ) )
                    return ret;
            }
            DOFAuthenticatorConfigBuilder_Destroy( ret );
        }
    }
    return NULL;
}

DOFAuthenticatorConfigBuilder DOFAuthenticatorConfigBuilder_Create_Config( DOFAuthenticatorConfig self )
{
    if( self )
    {
        DOFAuthenticatorConfigBuilder ret = (DOFAuthenticatorConfigBuilder_t*)DOFMem_Alloc( sizeof(DOFAuthenticatorConfigBuilder_t) );
        if( ret )
        {
            PCRMem_Set( ret, 0, sizeof( DOFAuthenticatorConfigBuilder_t ) );
            if( DOFAuthenticatorConfig_CopyInit( &(ret->config), self ) )
                return ret;
            DOFMem_Free(ret);
        }
    }
    return NULL;
}

void DOFAuthenticatorState_Destroy( DOFAuthenticatorState self )
{
    if( self )
    {
        DOFObjectIDDomain_Destroy( self->domainID );
        DOFMem_Free( self );
    }
}

boolean DOFAuthenticatorState_IsAvailable( const DOFAuthenticatorState self )
{
    PCR_Assert(self);
    return (self->flags & DOFAUTHENTICATORSTATE_FLAG_ISAVALILABLE);
}

DOFObjectIDDomain DOFAuthenticatorState_GetDomainID( const DOFAuthenticatorState self )
{
    PCR_Assert(self);
    return self->domainID;
}

uint8 DOFAuthenticatorState_GetMaxGrantLifetime( const DOFAuthenticatorState self )
{
    PCR_Assert(self);
    return self->maxLifetime;
}

uint8 DOFAuthenticatorState_GetNodeIdentifier( const DOFAuthenticatorState self )
{
    PCR_Assert(self);
    return self->nodeID;
}

boolean DOFAuthenticatorState_IsDebug( const DOFAuthenticatorState self )
{
    PCR_Assert(self);
    return (self->flags & DOFAUTHENTICATORSTATE_FLAG_ISDEBUG);
}


/* New Domain stuff */
static OALEnqueue domainEnqueue = NULL;
static OALTimer   domainTimer = NULL;
static PCRMutex   domainMtx = NULL;

void DOFDomain_Lock(void)
{
    PCRMutex_Lock( domainMtx );
}

void DOFDomain_Unlock(void)
{
    PCRMutex_Unlock( domainMtx );
}

#define DOFDOMAIN_SILENCE_BACKOFF 15000
#define DOFDOMAIN_SILENCE_RETRY_TIME 5000

typedef struct DOFDomainStateListenerNode_t* DOFDomainStateListenerNode;
typedef struct DOFDomainStateListenerSListNode_t* DOFDomainStateListenerSListNode;

typedef struct DOFDomainStateListenerSListNode_t
{
    OALSListNode                next;
    DOFDomainStateListenerNode  self;
}DOFDomainStateListenerSListNode_t;

typedef struct DOFDomainStateListenerNode_t
{
    OALEnqueueItem_t        base;
    DOFDomainStateListener  listener;
    DOFDomain               domain;
    DOFException            exception;
    DOFDomainStateListenerSListNode node;
}DOFDomainStateListenerNode_t;

static void DOFDomain_NotifyStateRemoved( DOFDomainStateListenerNode node );
static boolean DOFDomainStateListenerNode_NotifyStateChanged( DOFDomainStateListenerNode node, DOFDomainState state );


boolean DOFDomain_Init1( void )
{
    domainMtx = PCRMutex_Create();
    if( domainMtx )
        return TRUE;
    return FALSE;
}

boolean DOFDomain_Init2( void )
{
    domainEnqueue = OALEnqueue_Create( DOF_GetDof()->threadPool );
    if( domainEnqueue )
    {
        domainTimer = OALTimer_Create( DOF_GetDof()->threadPool, TRUE );
        if( domainTimer )
            return TRUE;

        OALEnqueue_Destroy( domainEnqueue );
        domainEnqueue = NULL;
    }
    return FALSE;
}

void DOFDomain_Shutdown2( void )
{
    OALEnqueue_Destroy( domainEnqueue );
    domainEnqueue = NULL;
    OALTimer_Destroy( domainTimer );
    domainTimer = NULL;
}

void DOFDomain_Shutdown1(void)
{
    PCRMutex_Destroy( domainMtx );
    domainMtx = NULL;
}

static void DOFDomain_NotifyRemoved( DOFDomain self, boolean destroy, DOFException exception );

struct ddntc_t
{
    DOFDomainState  state;
    OALSListNode    head;
};

static boolean DOFDomainNotifyTraverse(OALAVLTreeItem_t item, void *arg)
{
    DOFDomainStateListenerNode node = (DOFDomainStateListenerNode)item;
    struct ddntc_t* ctx = (struct ddntc_t*)arg;

    if( !DOFDomainStateListenerNode_NotifyStateChanged( node, ctx->state ) )
    {
        OALSList_Insert( &ctx->head, (OALSListNode)node->node );
    }
    return FALSE;
}

static void FailedToNotify_RemovedCall( OALSListNode item, void *arg )
{
    DOFDomainStateListenerSListNode node = (DOFDomainStateListenerSListNode)item;
    DOFDomain domain = (DOFDomain)arg;
    if( OALAVLTree_Delete( domain->listeners, node->self->listener ) )
    {
        node->self->exception = DOFEXCEPTION_INSUFFICIENT_RESOURCES;
        DOFDomain_NotifyStateRemoved( node->self );
    }
}

static void DOFDomain_NotifyStateChanged( DOFDomain self )
{
    if( self )
    {
        if( self->listeners )
        {
            struct ddntc_t ctx;

            ctx.head = NULL;

            DOFDomain_Lock();
            ctx.state = DOFDomain_CreateState(self);
            if( ctx.state )
            {
                OALAVLTree_Traverse( self->listeners, DOFDomainNotifyTraverse, &ctx );
                DOFDomainState_Destroy( ctx.state );
                ctx.state = NULL;
            }
            else
            {
                /* Couldn't notify any because of alloc failure so we remove all listeners */
                DOFDomain_NotifyRemoved( self, FALSE, DOFEXCEPTION_INSUFFICIENT_RESOURCES );
            }

            if( ctx.head != NULL )
            {
                /* remove just the listeners that failed */
                OALSList_RemoveAll( &ctx.head, FailedToNotify_RemovedCall, self );
            }
            DOFDomain_Unlock();
        }
    }
}

static void DOFDomainValidate( DOFCredentials self, DOFException exception, void* context )
{
    DOFDomain domain = (DOFDomain)context;

    if( !exception )
    {
        boolean changed;
        DOFDomain_Lock();
        changed = !domain->flags.isConnected;
        domain->flags.validating1 = FALSE;
        domain->flags.validating2 = FALSE;
        domain->flags.validating3 = FALSE;
        domain->flags.isConnected = TRUE;
        DOFDomain_SetUpSilenceTimer( domain );
        if( changed )
            DOFDomain_NotifyStateChanged( domain );
        DOFDomain_Unlock();
    }

    DOFDomain_Destroy( domain );
}

typedef struct DOFDomainValidateNode_t
{
    OALEnqueueItem_t    base;
    DOFDomain           domain;
    uint8               stage;
}DOFDomainValidateNode_t;

static void EnqueuedDomainTRPCall(OALThreadPoolItem item, boolean shutdown)
{
    DOFDomainValidateNode_t* node = (DOFDomainValidateNode_t*)item;
    if(!shutdown)
    {
        if( !DOFCredentials_IsResolved(  node->domain->creds ) )
            DOFCredentials_Internal_BeginResolve(  node->domain->creds, DOF_GetDof(), 15000-(DOFDOMAIN_SILENCE_RETRY_TIME*node->stage), FALSE, NULL, NULL, DOFDomainValidate,  node->domain );
        else
            DOFCredentials_Internal_BeginValidate(  node->domain->creds, DOF_GetDof(), 15000-(DOFDOMAIN_SILENCE_RETRY_TIME*node->stage), FALSE, NULL, NULL, DOFDomainValidate,  node->domain );
    }
    else
    {
        DOFDomain_Destroy( node->domain );
    }
    DOFMem_Free( item );
}

static void DOFDomain_SendDomainValidate( DOFDomain self, uint8 stage )
{
    if( domainEnqueue )
    {
        DOFDomain domain = DOFDomain_Copy(self); 
        if( domain )
        {
            DOFDomainValidateNode_t* node = DOFMem_Alloc(sizeof(DOFDomainValidateNode_t));
            if( node )
            {
                OALEnqueueItem_Init( (OALEnqueueItem)node, EnqueuedDomainTRPCall );
                node->domain = domain;
                node->stage = stage;

                if( OALEnqueue_Enqueue(domainEnqueue, (OALEnqueueItem)node) )
                    return;

                DOFMem_Free( node );
            }
            DOFDomain_Destroy( domain );
        }
    }
}

static int DOFDomain_HandleMaxSilence( OALTimerItem item, uint32 argc, void *argv[], boolean shutdown )
{
    if( argc == 1 )
    {
        DOFDomain self = (DOFDomain)argv[0];
        if( !shutdown )
        {
            DOFDomain_Lock();
            if( self->timerItem )
            {
                if( self->flags.validating3 )
                {
                    if( self->flags.isConnected )
                    {
                        self->flags.isConnected = FALSE;
                        DOFDomain_NotifyStateChanged( self );
                    }
                }
                else if( self->flags.validating2 )
                {
                    self->flags.validating3 = TRUE;
                    OALTimerItem_Insert( self->timerItem, DOFDOMAIN_SILENCE_RETRY_TIME, FALSE, FALSE );
                    DOFDomain_SendDomainValidate( self, 2 );
                }
                else if( self->flags.validating1 )
                {
                    self->flags.validating2 = TRUE;
                    OALTimerItem_Insert( self->timerItem, DOFDOMAIN_SILENCE_RETRY_TIME, FALSE, FALSE );
                    DOFDomain_SendDomainValidate( self, 1 );
                }
                else
                {
                    self->flags.validating1 = TRUE;
                    OALTimerItem_Insert( self->timerItem, DOFDOMAIN_SILENCE_RETRY_TIME, FALSE, FALSE );
                    DOFDomain_SendDomainValidate( self, 0 );
                }
            }
            DOFDomain_Unlock();
        }
        else
        {
            DOFDomain_Destroy(self);
        }
    }
    return 0;   
}

void DOFDomain_SetUpSilenceTimer( DOFDomain self )
{
    uint32 reltime = (self->maxSilence>DOFDOMAIN_SILENCE_BACKOFF)?self->maxSilence-DOFDOMAIN_SILENCE_BACKOFF:self->maxSilence;
    OALTimerItem_Insert( self->timerItem, reltime, FALSE, TRUE );
}

void DOFDomain_UpdateSilenceTimer( DOFDomain self, uint32 maxSilence )
{
    /* Only update if it is a shorter time*/
    DOFDomain_Lock();
    if( self->maxSilence > maxSilence )
    {
        uint32 oldSilence = self->maxSilence; 
        self->maxSilence = maxSilence;

        /* don't do this if we are already validating */
        if( !self->flags.validating1 )
        {
            uint32 reltime = (oldSilence - (OALTimerItem_TimeRemaining( self->timerItem ) + ((oldSilence>DOFDOMAIN_SILENCE_BACKOFF)?DOFDOMAIN_SILENCE_BACKOFF:0)));
            int32 theNewTime = (self->maxSilence - (((self->maxSilence>DOFDOMAIN_SILENCE_BACKOFF)?DOFDOMAIN_SILENCE_BACKOFF:0) + reltime));

            if( theNewTime > 0 )
            {
                OALTimerItem_Insert( self->timerItem, theNewTime, FALSE, TRUE );
            }
            else
            {
                /* Trigger it now */
                OALTimerItem_Insert( self->timerItem, 0, FALSE, TRUE );
            }
        }
    }
    DOFDomain_Unlock();
}

static boolean DOFDomainStateListenerNode_gt(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    return ( key > (void*)((DOFDomainStateListenerNode)item)->listener );
}

static boolean DOFDomainStateListenerNode_eq(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    return ( key == (void*)((DOFDomainStateListenerNode)item)->listener );
}

DOFDomain DOFDomain_CreateInternal( DOFDomainConfig config )
{
    DOFDomain ret = (DOFDomain)DOFMem_Alloc( sizeof( DOFDomain_t ) );
    if( ret )
    {
        PCRMem_Set( ret, 0, sizeof( DOFDomain_t ) );

        ret->refCount = 1;
        ret->maxSilence = config->maxSilence;

        ret->creds = DOFCredentials_Copy( config->creds );
        if( ret->creds )
        {
            ret->listeners = OALAVLTree_Create( DOFDomainStateListenerNode_gt, DOFDomainStateListenerNode_eq, NULL );
            if( ret->listeners )
            {
                ret->timerItem = OALTimer_CreateItem( domainTimer, DOFDomain_HandleMaxSilence, 1, (void**)&ret );
                if( ret->timerItem )
                {
                    OALTimerItem_Insert( ret->timerItem, 0, FALSE, TRUE ); /* trigger validation now */
                    return ret;
                }
                OALAVLTree_Destroy( ret->listeners, NULL, NULL );
            }
            DOFCredentials_Destroy( ret->creds );
        }
        DOFMem_Free( ret );
    }
    return NULL;
}

DOFDomain DOFDomain_Copy( DOFDomain self )
{
    DOFDomain ret = NULL;

    DOFDomain_Lock();
    if( self )
    {
        PCR_Assert(self->refCount != 0);
        if( self->refCount < 0xFFFFFFFF )
        {
            self->refCount++;
            ret = self;
        }
    }
    DOFDomain_Unlock();
    return ret;
}

DOFCredentials DOFDomain_GetCredentials( DOFDomain self )
{
    if( self )
        return self->creds;
    return NULL;
}

static void DOFDomainListeners_RemoveAll_fn( OALAVLTreeItem_t item, void *arg )
{
    ((DOFDomainStateListenerNode)item)->exception = (DOFException)arg;
    DOFDomain_NotifyStateRemoved((DOFDomainStateListenerNode)item);
}

void DOFDomain_NotifyRemovedDOF( DOFDomain self )
{
    if( self->dof )
    {
        OALTimerItem temp = self->timerItem;
        DOFDomain_NotifyRemoved( self, TRUE, NULL );
        self->dof = NULL; /* not associated with this DOF after next call */
        self->timerItem = NULL;

        DOFDomain_Unlock();
        /* can't hold domain lock if this is firing at the same time */
        OALTimerItem_Destroy( temp, TRUE );
        DOFDomain_Lock();
    }
}

static void DOFDomain_NotifyRemoved( DOFDomain self, boolean destroy, DOFException exception )
{
    DOFDomain_Lock();
    if( self->listeners )
    {
        if( destroy ) 
        {
            OALAVLTree_Destroy( self->listeners, DOFDomainListeners_RemoveAll_fn, exception );
            self->listeners = NULL;
        }
        else
        {
            OALAVLTree_Empty( self->listeners, DOFDomainListeners_RemoveAll_fn, exception );
        }
    }
    DOFDomain_Unlock();
}

/* 
   called before domain is destroy when the dof is going away
   This needs to do everything destroy would but free the memory
*/
void DOFDomain_RemoveDOF( DOFDomain self )
{
    DOFDomain_Lock();
    if( self && self->dof )
    {
        DOFDomain_NotifyRemovedDOF( self );
    }
    DOFDomain_Unlock();
}

void DOFDomain_Destroy( DOFDomain self )
{
    boolean destroy = FALSE;

    DOFDomain_Lock();
    if( self )
    {
        PCR_Assert( self->refCount > 0x0 );
        self->refCount--;
        if( self->refCount == 0 )
            destroy = TRUE;   
        
        if( destroy )
        {
            if( self->dof )
            {
                DOF_RemoveDomain( self->dof, self );
                DOFDomain_NotifyRemovedDOF( self );
            }

            DOFCredentials_Destroy( self->creds );
            DOFMem_Free( self );
        }
    }
    DOFDomain_Unlock();
}

boolean DOFDomain_IsEqual( const DOFDomain self, const DOFDomain other )
{
    if( ( self == other ) ||
        DOFCredentials_IsEqual( self->creds, other->creds ) )
        return TRUE;

    return FALSE;
}

DOFDomainState DOFDomain_CreateState( const DOFDomain self )
{
    if( self )
    {
        DOFDomainState ret = (DOFDomainState)DOFMem_Alloc( sizeof( DOFDomainState_t ) );
        if( ret )
        {
            ret->creds = DOFCredentials_CreateNeutered( self->creds );
            if( ret->creds )
            {
                ret->refCount = 1;
                ret->isConnected = self->flags.isConnected;
                ret->maxSilence = self->maxSilence;
                return ret;
            }
            DOFMem_Free( ret );
        }
    }
    return NULL;
}

static void DOFDomainStateListenerNode_Removed(OALThreadPoolItem item, boolean shutdown)
{
    DOFDomainStateListenerNode node = (DOFDomainStateListenerNode)item;
    if( node->listener->fns->Removed )
        node->listener->fns->Removed( node->listener, node->domain, node->exception );
    DOFMem_Free( node );
}

static void DOFDomain_NotifyStateRemoved( DOFDomainStateListenerNode node )
{
    OALEnqueueItem_Init( (OALEnqueueItem)node, DOFDomainStateListenerNode_Removed );

    if( !OALEnqueue_Enqueue( domainEnqueue, (OALEnqueueItem)node ) )
    {
        DOFLog_Message(DOFLOGLEVEL_WARNING, "DOFDomain failed to notify a listener that it was removed.");
        DOFMem_Free( node );
    }
}

typedef struct DOFDomainStateListenerChangedNode_t
{
    OALEnqueueItem_t        base;
    DOFDomainStateListener  listener;
    DOFDomainState          state;
    DOFDomain               domain;
}DOFDomainStateListenerChangedNode_t;

static void DOFDomainStateListenerNode_StateChanged(OALThreadPoolItem item, boolean shutdown)
{
    DOFDomainStateListenerChangedNode_t* node = (DOFDomainStateListenerChangedNode_t*)item;
    node->listener->fns->StateChanged( node->listener, node->domain, node->state );
    DOFDomainState_Destroy( node->state );
    DOFDomain_Destroy( node->domain );
    DOFMem_Free( node );
}

static boolean DOFDomainStateListenerNode_NotifyStateChanged( DOFDomainStateListenerNode node, DOFDomainState state )
{
    DOFDomainStateListenerChangedNode_t* item = (DOFDomainStateListenerChangedNode_t*)DOFMem_Alloc( sizeof( DOFDomainStateListenerChangedNode_t ) );
    if( item )
    {
        item->state = DOFDomainState_Copy(state);
        if( item->state )
        {
            item->domain = DOFDomain_Copy( node->domain );
            if( item->domain )
            {
                OALEnqueueItem_Init( (OALEnqueueItem)item, DOFDomainStateListenerNode_StateChanged );
                item->listener = node->listener;
                OALEnqueue_Enqueue( domainEnqueue, (OALEnqueueItem)item );
                return TRUE;
            }
            DOFDomainState_Destroy( item->state );
        }
        DOFMem_Free( item );
    }
    return FALSE;
}

boolean DOFDomain_AddStateListener( DOFDomain self, DOFDomainStateListener listener )
{
    OALAVLTree_InsertionStatus status = OALAVLTree_InsertFailure;
    if( self && self->dof )
    {
        DOFDomainStateListenerNode node = (DOFDomainStateListenerNode)DOFMem_Alloc( sizeof( DOFDomainStateListenerNode_t ) + sizeof( DOFDomainStateListenerSListNode_t ) );
        if( node )
        {
            DOFDomainState state = DOFDomain_CreateState(self);
            if( state )
            {
                PCRMem_Set( node, 0, (sizeof( DOFDomainStateListenerNode_t ) + sizeof( DOFDomainStateListenerSListNode_t )) );

                node->node = (DOFDomainStateListenerSListNode)(((uint8*)node) + sizeof( DOFDomainStateListenerNode_t ));
                node->node->self = node;

                node->listener = listener;
                node->domain = self;
                
                DOFDomain_Lock();
                status = OALAVLTree_Insert( self->listeners, listener, node, NULL );
                if( status == OALAVLTree_InsertSuccess )
                {
                    if( !DOFDomainStateListenerNode_NotifyStateChanged( node, state ) ) 
                    {
                        /* failed to notify the initial change so we just remove it and return failure */
                        if( !OALAVLTree_Delete( self->listeners, listener ) )
                        {
                            PCR_Assert( FALSE ); /* we just inserted it why didn't we find it? */
                        }
                        DOFDomain_Unlock();
                        DOFDomainState_Destroy( state );
                        DOFMem_Free( node );
                        return FALSE;
                    }
                    DOFDomain_Unlock();
                    DOFDomainState_Destroy( state );
                    return TRUE;
                }
                DOFDomain_Unlock();
                DOFDomainState_Destroy( state );
            }
            DOFMem_Free( node );
        }
    }
    return (status != OALAVLTree_InsertFailure);
}

void DOFDomain_RemoveStateListener( DOFDomain self, DOFDomainStateListener listener )
{
    if( self )
    {
        DOFDomainStateListenerNode node;
        DOFDomain_Lock();
        node = (DOFDomainStateListenerNode)OALAVLTree_Retrieve( self->listeners, listener );
        if( node )
        {
            if( OALAVLTree_Delete( self->listeners, listener ) )
            {
                DOFDomain_NotifyStateRemoved( node );
                DOFDomain_Unlock();
                return;
            }
        }
        DOFDomain_Unlock();
    }
}

DOFDomainConfigBuilder DOFDomainConfigBuilder_Create( DOFCredentials creds )
{
    DOFDomainConfigBuilder ret = (DOFDomainConfigBuilder)DOFMem_Alloc( sizeof( DOFDomainConfigBuilder_t ) );
    if( ret )
    {
        PCRMem_Set( ret, 0, sizeof( DOFDomainConfigBuilder_t ) );
        ret->base.creds = DOFCredentials_Copy( creds );
        if( ret->base.creds || !creds )
        {
            ret->base.refCount = 1;
            ret->base.maxSilence = DOFDOMAIN_HOURSILENCE;
            return ret;
        }
        DOFMem_Free( ret );
    }
    return NULL;
}

DOFDomainConfigBuilder DOFDomainConfigBuilder_Create_State( const DOFDomainState self )
{
    if( self )
    {
        DOFDomainConfigBuilder ret = DOFDomainConfigBuilder_Create( NULL );
        if( ret )
        {
            if( DOFDomainConfigBuilder_SetMaxSilence( ret, self->maxSilence ) )
                return ret;
            DOFDomainConfigBuilder_Destroy( ret );
        }
    }
    return NULL;
}

DOFDomainConfigBuilder DOFDomainConfigBuilder_Create_Config( const DOFDomainConfig config )
{
    if( config )
    {
        DOFDomainConfigBuilder ret = DOFDomainConfigBuilder_Create( config->creds );
        if( ret )
        {
            if( DOFDomainConfigBuilder_SetMaxSilence( ret, config->maxSilence ) )
                return ret;
            DOFDomainConfigBuilder_Destroy( ret );
        }
    }
    return NULL;
}

void DOFDomainConfigBuilder_Destroy( DOFDomainConfigBuilder self )
{
    if( self )
    {
        DOFCredentials_Destroy( self->base.creds );
        DOFMem_Free( self );
    }
}

boolean DOFDomainConfigBuilder_SetCredentials( DOFDomainConfigBuilder self, DOFCredentials creds )
{
    if( self && creds )
    {
        DOFCredentials temp = self->base.creds;
        self->base.creds = DOFCredentials_Copy( creds );
        if( self->base.creds )
        {
            DOFCredentials_Destroy( temp );
            self->error = FALSE; /* only thing that can fail currently */
            return TRUE;
        }
        self->base.creds = temp;
        self->error = TRUE;
    }
    return FALSE;
}

boolean DOFDomainConfigBuilder_SetMaxSilence( DOFDomainConfigBuilder self, uint32 maxSilence )
{
    if( self && maxSilence > 0)
    {
        self->base.maxSilence = maxSilence;
        return TRUE;
    }
    return FALSE;
}

DOFDomainConfig DOFDomainConfigBuilder_Build( const DOFDomainConfigBuilder self )
{
    if( !self->error ) /* can't build if error happened on a set */
    {
        DOFDomainConfig ret = (DOFDomainConfig)DOFMem_Alloc( sizeof( DOFDomainConfig_t ) );
        if( ret )
        {
            ret->creds = DOFCredentials_Copy( self->base.creds );
            if( ret->creds )
            {
                ret->maxSilence = self->base.maxSilence;
                ret->refCount = 1;
                return ret;
            }
            DOFMem_Free( ret );
        }
    }
    return NULL;
}

DOFDomainConfig DOFDomainConfigBuilder_BuildAndDestroy( DOFDomainConfigBuilder self )
{
    DOFDomainConfig ret = (DOFDomainConfig)self;

    if( self && self->error )
    {
        DOFDomainConfig_Destroy( ret );
        ret = NULL;
    }
    return ret;
}

DOFDomainConfig DOFDomainConfig_Copy( DOFDomainConfig self )
{
    DOFDomainConfig ret = NULL;
    if( self )
    {
        PCRMutex_Lock(ref_count_mutex);
        if( self->refCount < 0xFFFFFFFF )
        {
            ++self->refCount;
            ret = self;
        }
        PCRMutex_Unlock(ref_count_mutex);
    }
    return ret;
}

void DOFDomainConfig_Destroy( DOFDomainConfig self )
{
    if( self )
    {
        boolean destroy = FALSE;

        PCRMutex_Lock(ref_count_mutex);
        PCR_Assert( self->refCount > 0 );
        --self->refCount;
        if( self->refCount == 0 )
            destroy = TRUE;   
        PCRMutex_Unlock(ref_count_mutex);

        if( destroy )
        {
            DOFCredentials_Destroy( self->creds );
            DOFMem_Free( self );
        }
    }
}

DOFCredentials DOFDomainConfig_GetCredentials( const DOFDomainConfig self )
{
    if( self )
        return self->creds;
    return NULL;
}

uint32 DOFDomainConfig_GetMaxSilence( const DOFDomainConfig self )
{
    if( self )
        return self->maxSilence;
    return 0;
}

DOFDomainState DOFDomainState_Create( DOFDomain domain )
{
    return DOFDomain_CreateState( domain );
}

DOFDomainState DOFDomainState_Copy( DOFDomainState self )
{
    DOFDomainState ret = NULL;
    if( self )
    {
        PCRMutex_Lock(ref_count_mutex);
        if( self->refCount <= 0xFFFFFFFF )
        {
            ++self->refCount;
            ret = self;
        }
        PCRMutex_Unlock(ref_count_mutex);
    }
    return ret;
}

void DOFDomainState_Destroy( DOFDomainState self )
{
    if( self )
    {
        boolean destroy = FALSE;

        PCRMutex_Lock(ref_count_mutex);
        PCR_Assert( self->refCount > 0 );
        if ( (--self->refCount) == 0 )
            destroy = TRUE;  
        PCRMutex_Unlock(ref_count_mutex);

        if( destroy )
        {
            DOFCredentials_Destroy( self->creds );
            DOFMem_Free( self );
        }
    }
}

boolean DOFDomainState_IsConnected( const DOFDomainState self )
{
    if( self )
        return self->isConnected;
    return FALSE;
}

DOFCredentials DOFDomainState_GetCredentials( const DOFDomainState self )
{
    if( self )
        return self->creds;
    return NULL;
}

uint32 DOFDomainState_GetMaxSilence( const DOFDomainState self )
{
    if( self )
        return self->maxSilence;
    return 0;
}
