/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/addressinternal.h>
#include <internal/dofinternal.h>
#include <internal/enp_v1.h>
#include <internal/marshal.h>
#include <internal/remotechannel.h>
#include <internal/remotechannelmanager.h>
#include <internal/router.h>
#include <internal/serverinternal.h>

/*
**  "Private" Type and Method Declarations
*/

/* Masks for the first ENP byte, flags and version. */
#define ENP_VERSION_MASK    0x7F
#define ENP_FLAGS_PRESENT_MASK  0x80

/* Masks for the second ENP byte assuming flags are present. */
#define ENP_V1_LENGTH_MASK                  0x03
#define ENP_V1_SOURCE_PORT_PRESENT          0x04
#define ENP_V1_DEST_PORT_PRESENT            0x08
#define ENP_V1_BYTE_TWO_RESERVED_BITS       0xF0

#define ENP_V1_PORT_MAX_BYTES               3

#define ENP_MAX_HEADER_SIZE                 11


typedef struct ENP_V1_t
{
    BlackboardData      blackboardData;
    boolean             lenInferencePermitted;
    boolean             cpyStreamOnReceive;

    uint8   numLengthBytes; /* The number of length byte fields to expect in the packet. Defaults to zero on creation. */
} ENP_V1_t;



/*
**  "Public" Methods
*/

ENP_V1  ENP_V1_Create(BlackboardData blackboardData, boolean lengthInferencePermitted, boolean copyStreamOnReceive)
{
    ENP_V1 retVal = (ENP_V1)DOFMem_Alloc( sizeof(ENP_V1_t) );

    if (!retVal)
        return NULL;

    retVal->blackboardData = blackboardData;
    retVal->lenInferencePermitted = lengthInferencePermitted;
    retVal->cpyStreamOnReceive = copyStreamOnReceive;
    retVal->numLengthBytes = 0;

    return retVal;
}

/*
 *
 */
void ENP_V1_Destroy(ENP_V1 enp)
{
    DOFMem_Free(enp);
}

ProtocolLayerResult ENP_V1_Receive(ENP_V1 self)
{
    BlackboardData  blackboardData = self->blackboardData;
    const uint32    dataLen = OALStream_GetDataSize(blackboardData->receiveStream);
    const uint8*    pPacketStart = OALStream_GetReadPtr(blackboardData->receiveStream);
    const uint8*    pData = pPacketStart;
    uint32          encapsulatedPacketSize = 0;
    boolean         sourcePortPresent = FALSE;
    boolean         destPortPresent = FALSE;
    boolean         flagsPresent;
    uint8           i;

    /* Verify ENP Version header. */
    if ( (*pData & ENP_VERSION_MASK) != ENP_V1_VERSION )
        return PLRESULT_ERROR;

    if (blackboardData->associatedChannel->remoteChannelData->flags.ENP_V0_QueryInProgress &&
        blackboardData->flags.isMulticast)
    {
        blackboardData->associatedChannel->remoteChannelData->ENP_V0_QueryVersions |= V0_QUERY_BIT_ENP_V1;
    }


    /* Check for flags */
    flagsPresent = *pData++ & ENP_FLAGS_PRESENT_MASK;

    if (flagsPresent)
    {
        self->numLengthBytes = *pData & ENP_V1_LENGTH_MASK;         /* read length byte count from packet. */
        sourcePortPresent = *pData & ENP_V1_SOURCE_PORT_PRESENT;    /* see if a source port was included. */
        destPortPresent = *pData & ENP_V1_DEST_PORT_PRESENT;        /* see if a dest port was included */
        pData++;
        /* Reserved bits are not checked. */
    }
    else
    {
        self->numLengthBytes = 0; /* flags are not present so set to the default value of zero. */
    }


    /* Check that we can read all the length bytes necessary. */
    if (self->numLengthBytes > (dataLen - 0x02))
        return PLRESULT_NEEDMOREDATA;

    /* If there are no length bytes to be read and the length of the packet cannot
       be infered then it is an error. If it can be infered then the size of the
       packet is infered from the size of the received stream. The typical case for using
       inference is when the transport is packet-based. */
    if ( !(self->numLengthBytes))
    {
        if ( !(self->lenInferencePermitted) )
            return PLRESULT_ERROR;
        else
            encapsulatedPacketSize = dataLen - (pData - pPacketStart);
    }
    else
        for (i = 0; i < self->numLengthBytes; ++i, ++pData)
            encapsulatedPacketSize = (encapsulatedPacketSize << 8) + *pData;

    /* This pointer arithmetic depends on OALStream using one continuous block for buffer */
    if ( encapsulatedPacketSize > (dataLen - (pData - pPacketStart)) )
        return PLRESULT_NEEDMOREDATA;

    if (self->lenInferencePermitted)
    {
        int32           toAdjust;

        /* If a packet is received with padding at the end of the packet adjust the write offset so that the extra bytes are not passed on. */

        /* one byte for the header, one byte if we had flags */
        toAdjust = encapsulatedPacketSize + 1 - dataLen + self->numLengthBytes;
        if(flagsPresent)
        {
            ++toAdjust;
        }
        OALStream_AdjustWriteOffset(blackboardData->receiveStream, toAdjust);
    }

    if (sourcePortPresent)
    {
        DOFAddress  newAddr;
        uint32      size;
        uint32      port;

        size = encapsulatedPacketSize;
        port = OALMarshal_UncompressValue(ENP_V1_PORT_MAX_BYTES, &size, pData);
        if ( port == OALMARSHAL_INVALID_VALUE )
            return PLRESULT_ERROR;

        pData += size;      /* OALMarshal_UncompressValue will set the actual size of the port. */
        encapsulatedPacketSize -= size;

        newAddr = DOFAddress_Create_AddENPPort(self->blackboardData->transportAddr, port);
        if (newAddr)
        {
            DOFAddress_Destroy(self->blackboardData->transportAddr);
            self->blackboardData->transportAddr = newAddr;
        }
        else {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "ENP V1: Failed to create address with enp port.");
            return PLRESULT_ERROR;
        }
    }

    if (destPortPresent)
    {
        OALChannel  newChannel;
        uint32      size;
        uint32      port;

        size = encapsulatedPacketSize;
        port = OALMarshal_UncompressValue(ENP_V1_PORT_MAX_BYTES, &size, pData);
        if ( port == OALMARSHAL_INVALID_VALUE )
            return PLRESULT_ERROR;

        pData += size;      /* OALMarshal_UncompressValue will set the actual size of the port. */
        encapsulatedPacketSize -= size;

        /* Lookup the channel registered with the ENP port. */
        newChannel = OALRemoteChannel_GetRegisteredChannel_ENPPort(self->blackboardData->associatedChannel, port);
        if (newChannel)
        {
            OALChannel_Release(self->blackboardData->associatedChannel);
            self->blackboardData->associatedChannel = OALChannel_Retain(newChannel);

            if (self->blackboardData->associatedChannel)
            {
                if (!self->blackboardData->associatedChannel->remoteChannelData->flags.isOutbound &&
                    self->blackboardData->associatedChannel->remoteChannelData->flags.waitingForInitialDatagram)
                {
                    newChannel->remoteChannelData->flags.waitingForInitialDatagram = FALSE;
                    newChannel->remoteChannelData->flags.mustTerminate = TRUE;
                    OALRemoteChannel_NotifyInitComplete(newChannel);
                }
            }
            else
            {
                DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ENP V1: Failed to retain new channel after lookup by ENP port number");
                /* Note: we cannot return PLRESULT_OKDONOTHING here.  It will cause an infinite loop in connected protocol stack.*/
                return PLRESULT_ERROR;
            }
        }
        else
        {
            DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ENP V1: channel/ENP Port Lookup did not find a channel.");
            return PLRESULT_ERROR;
        }
    }

    if (self->blackboardData->associatedChannel->remoteChannelData->flags.lossySession &&
        !self->blackboardData->associatedChannel->remoteChannelData->flags.waitingForInitialDatagram)
    {
        /*  Lossy sessions must drop the packet if it is not from the address configured in the channel. */
        if (!DOFAddress_IsEqual(self->blackboardData->transportAddr, self->blackboardData->associatedChannel->remoteChannelData->address))
        {
            DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ENP V1: Received from an address other than the channel's address.");
            return PLRESULT_ERROR;
        }
    }

    OALStream_AdjustReadOffset(blackboardData->receiveStream, pData - pPacketStart);

    if (self->cpyStreamOnReceive)
    {
        /* Create a new stream and copy over the data.
           The typical case for this is when the transport
           is stream-based. */
        OALStream tmpStream = OALStream_CreateRW(encapsulatedPacketSize);

        if (tmpStream)
        {
            OALStream_ReadToStream(blackboardData->receiveStream, encapsulatedPacketSize, tmpStream);
            OALStream_Release( blackboardData->receiveStream, TRUE );
            blackboardData->receiveStream = tmpStream;
        }
        else
        {
            DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ENP V1: Failed to create OALStream.") ;
            return PLRESULT_ERROR;
        }
    }

    return PLRESULT_OKSENDUP;
}


ProtocolLayerResult ENP_V1_Send(ENP_V1 self)
{
    OALStream enpHeader;
    OALPacket packet = self->blackboardData->sendPacket;
    uint32    minSize = 0;
    uint32    length = OALPacket_GetTotalStreamSize(packet);
    uint32    port;
    uint8     senderPort[ENP_V1_PORT_MAX_BYTES];
    uint8     receiverPort[ENP_V1_PORT_MAX_BYTES];
    uint8     headerSize = 1;
    uint8     version = ENP_V1_VERSION;
    uint8     flags = 0;
    uint8     lengthSize = 0;
    uint8     senderPortSize = 0;
    uint8     receiverPortSize = 0;

    if ( (length > 0xffffff) || !length )
        return PLRESULT_ERROR;

    if ( !self->blackboardData->transportAddr )
    {
        PCR_Assert( self->blackboardData->associatedChannel->remoteChannelData->address );
        self->blackboardData->transportAddr = DOFAddress_Copy( self->blackboardData->associatedChannel->remoteChannelData->address );
        if (!self->blackboardData->transportAddr)
        {
            DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_WARNING, "ENP V1: Failed to copy address.");
            return PLRESULT_ERROR;
        }
    }
    else
    {
        /*  If this is for a session, do not allow it to send to an address other than one configured in the channel. */
        if (self->blackboardData->associatedChannel->remoteChannelData->flags.lossySession)
        {
            if (!DOFAddress_IsEqual(self->blackboardData->transportAddr, self->blackboardData->associatedChannel->remoteChannelData->address))
            {
                DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_INFO, "ENP V1: Target address did not match.");
                return PLRESULT_ERROR;
            }
        }
    }
    if ( self->blackboardData->transportAddr->transport->fns->GetMinSize )
        minSize = self->blackboardData->transportAddr->transport->fns->GetMinSize(self->blackboardData->transportAddr->transport, (self->lenInferencePermitted)? DOFCONNECTIONTYPE_DATAGRAM : DOFCONNECTIONTYPE_STREAM);

    port = self->blackboardData->associatedChannel->remoteChannelData->ENP_Port;
    if ( port != ENP_INVALID_PORT )
    {
        flags |= ENP_V1_SOURCE_PORT_PRESENT;
        senderPortSize = OALMarshal_CompressValue(ENP_V1_PORT_MAX_BYTES, port, sizeof(senderPort), &senderPort[0]);
        PCR_Assert((senderPortSize > 0) && (senderPortSize <= ENP_V1_PORT_MAX_BYTES));
        length += senderPortSize;
    }

    port = DOFAddress_GetENPPort(self->blackboardData->transportAddr);
    if ( port != ENP_INVALID_PORT )
    {
        flags |= ENP_V1_DEST_PORT_PRESENT;
        receiverPortSize = OALMarshal_CompressValue(ENP_V1_PORT_MAX_BYTES, port, sizeof(receiverPort), &receiverPort[0]);
        PCR_Assert((receiverPortSize > 0) && (receiverPortSize <= ENP_V1_PORT_MAX_BYTES));
        length += receiverPortSize;
    }

    if ( !self->lenInferencePermitted || ( ( length + headerSize ) < minSize ) )
    {
        lengthSize = (length > 0xffff) ? 3 : ((length > 0xff) ? 2 : 1);
        flags |= lengthSize;
        headerSize += lengthSize;
    }

    if ( flags )
    {
        version |= ENP_FLAGS_PRESENT_MASK;
        ++headerSize;
    }

    if ( self->blackboardData->transportAddr->transport->fns->GetMaxSize &&
         ( length + headerSize ) > self->blackboardData->transportAddr->transport->fns->GetMaxSize(self->blackboardData->transportAddr->transport, (self->lenInferencePermitted)? DOFCONNECTIONTYPE_DATAGRAM : DOFCONNECTIONTYPE_STREAM) )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "ENP V1: Packet too large for transport." );
        return PLRESULT_ERROR;
    }

    length = PCRSwap_Net32( length );
    enpHeader = OALStream_CreateRW(ENP_MAX_HEADER_SIZE);
    if ( !enpHeader ||
         !OALStream_Write(enpHeader, 1, &version) ||
         ( flags && !OALStream_Write(enpHeader, 1, &flags) ) ||
         !OALStream_Write(enpHeader, lengthSize, ((uint8*)&length) + (4-lengthSize)) ||
         !OALStream_Write(enpHeader, senderPortSize, senderPort) ||
         !OALStream_Write(enpHeader, receiverPortSize, receiverPort) ||
         !OALPacket_PrependStream( packet, enpHeader ) )
    {
        OALStream_Release( enpHeader, FALSE );
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "ENP V1: Failed to prepend stream." );
        return PLRESULT_ERROR;
    }

    return PLRESULT_OKSENDDOWN;
}
