/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/address.h>
#include <dof/connection.h>
#include <dof/objectid.h>
#include <dof/permission.h>

#include <internal/dofinternal.h>
#include <internal/enqueue.h>
#include <internal/epp_negative_one.h>
#include <internal/exception.h>
#include <internal/marshal.h>
#include <internal/objectidinternal.h>
#include <internal/operationinternal.h>
#include <internal/stream.h>
#include <internal/avl_multitree.h>
#include <internal/marshal.h>
#include <internal/observable.h>
#include <internal/channel.h>
#include <internal/remotechannelmanager.h>
#include <internal/router.h>
#include <internal/source.h>
#include <internal/securityscope.h>
#include <internal/systeminternal.h>

typedef struct EPP_NegOne_t
{
    BlackboardData      blackboardData;
    boolean             isConnectionless;
} EPP_NegOne_t;

typedef struct SidQueryContext_t
{
    void *      context;
    OALChannel  channel;
    DOFAddress  address;
}SidQueryContext_t;

typedef ProtocolLayerResult (*EPP_NegOne_Receive_fn_t)(EPP_NegOne);
typedef ProtocolLayerResult (*EPP_NegOne_Send_fn_t)(EPP_NegOne, DOFOperation op);

#define EPP_NEG1_OPCODE_NODEDOWN        0x00
#define EPP_NEG1_OPCODE_SOURCELOST      0x01
#define EPP_NEG1_OPCODE_RENAME          0x02
#define EPP_NEG1_OPCODE_PING            0x03
#define EPP_NEG1_OPCODE_CANCELALL       0x04
#define EPP_NEG1_OPCODE_HEARTBEAT       0x05
#define EPP_NEG1_OPCODE_QUERY           0x06
#define EPP_NEG1_OPCODE_SOURCEFOUND     0x08

#define EPP_NEG1_APPID                  0x7fff

/* i'm going to use a multitree instead of a treelist or hashtable here. The reason is that we're going to key on channel,
 which is a pointer, and then have the value be an operation. I won't often need to traverse over all operations belonging
 to a tree but I will often need to remove an operation associated with a channel. Since there can be thousands of operations
 on a channel, and since what comes in must eventually be removed, log(n) delete is better than O(1) insert and O(n) delete
 with tree list. Since channel is only a pointer, and this guarantees correctness, there's no reason for a hashtable -- sds */
static OALAVLMultiTree channel_operation_tree;
static OALAVLMultiTree sid_operation_tree;

static OALEnqueue  sid_operation_enqueue = NULL;

static void channel_operation_observer(OALObserver self, OALObservable observable, void* ctx);

static OALStream EPP_NegOne_Send_Lost_Found(EPP_NegOne self, DOFOperationID id, DOFOperation op, uint8 opcode);

static PCRMutex mutex;

static boolean tree_gt(void * key, DOFOperation item, void *context)
{
    return key > item->sidKey;
}

static boolean tree_eq(void * key, DOFOperation item, void *context)
{
    return key == item->sidKey;
}

/* Struct and compare functions for channel_operation_tree */
typedef struct channelTreeItem_t
{
    OALChannel channel;
    DOFAddress address;
    DOFOperation op;
}channelTreeItem_t;

static boolean channelTree_gt(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    return (OALChannel)key > ((channelTreeItem_t *)item)->channel;
}

static boolean channelTree_eq(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    return (OALChannel)key == ((channelTreeItem_t *)item)->channel;
}

static boolean channelItem_gt_fn(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    boolean ret = DOFAddress_Compare(((channelTreeItem_t *)key)->address, ((channelTreeItem_t *)item)->address);

    if (ret)
    {
        return ret > 0;
    }
    return (((channelTreeItem_t *)key)->op > ((channelTreeItem_t *)item)->op);

}

static boolean channelItem_eq_fn(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    return ((DOFAddress_Compare(((channelTreeItem_t *)key)->address, ((channelTreeItem_t *)item)->address) == 0) && (((channelTreeItem_t *)key)->op == ((channelTreeItem_t *)item)->op));
}

void EPP_NegOne_Mutex_Init(void)
{
    mutex = PCRMutex_Create();
    PCR_Assert(mutex);
}

static void EPP_NegOne_DummyRemoveFn(OALAVLMultiTreeItem_t item, void *arg)
{
    /* This will prevent a crash in case all of the items are not removed from this tree. */
}

boolean EPP_NegOne_Init(void)
{
    PCR_Assert(!channel_operation_tree); /* don't call init without calling shutdown */


    sid_operation_tree = OALAVLMultiTree_Create((OALAVLMultiTree_CompareGreaterThan_fn_t)tree_gt, (OALAVLMultiTree_CompareEqual_fn_t)tree_eq, 0, 0, 0); /* create a tree with default compare fns and no context */
    if (sid_operation_tree)
    {
        channel_operation_tree = OALAVLMultiTree_Create((OALAVLMultiTree_CompareGreaterThan_fn_t)channelTree_gt, (OALAVLMultiTree_CompareEqual_fn_t)channelTree_eq,
                                                        (OALAVLMultiTree_CompareGreaterThan_fn_t)channelItem_gt_fn, (OALAVLMultiTree_CompareEqual_fn_t)channelItem_eq_fn, NULL);
        if (channel_operation_tree)
        {
            sid_operation_enqueue = OALEnqueue_Create(DOF_GetDof()->threadPool);
            if (sid_operation_tree)
            {
                return TRUE;
            }
            OALAVLMultiTree_Destroy(channel_operation_tree, EPP_NegOne_DummyRemoveFn, NULL);
            channel_operation_tree = NULL;
        }
        OALAVLMultiTree_Destroy(sid_operation_tree, EPP_NegOne_DummyRemoveFn, NULL);
        sid_operation_tree = NULL;
    }

    return FALSE;
}

void EPP_NegOne_Mutex_Shutdown(void)
{
    PCRMutex_Destroy(mutex);
    mutex = NULL;
}

void EPP_NegOne_Shutdown(void)
{
    OALEnqueue_Destroy(sid_operation_enqueue);
    PCR_Assert( !OALAVLMultiTree_GetNumItems( channel_operation_tree ) );
    OALAVLMultiTree_Destroy(channel_operation_tree, EPP_NegOne_DummyRemoveFn, NULL);
    OALAVLMultiTree_Destroy(sid_operation_tree, EPP_NegOne_DummyRemoveFn, NULL);
    channel_operation_tree = NULL;
    sid_operation_tree = NULL;
    sid_operation_enqueue = NULL;
}

typedef struct OperationContext_t
{
    OALEnqueueItem_t super;
    DOFOperation operation;
    boolean doFree;
} OperationContext_t;

static void EPP_NegOne_Channel_Operation_Observer(OALThreadPoolItem dnode, boolean shutdown)
{
    OperationContext_t * node;

    PCR_Assert(dnode != NULL);
    node = (OperationContext_t *) dnode;
    channel_operation_observer(NULL, NULL, node->operation);

    DOFOperation_Internal_Destroy(node->operation);
    if(node->doFree)
    {
        DOFMem_Free(node);
    }
}

typedef struct sidEnqueueNode_t
{
    OALEnqueueItem_t super;
    DOFOperation op;
    void * sidKey;
} sidEnqueueNode_t;

static void EPP_NegOne_UpdateOperationSid_Internal(OALThreadPoolItem dnode, boolean shutdown)
{
    sidEnqueueNode_t * node;

    PCR_Assert(dnode);
    node = (sidEnqueueNode_t *)dnode;

    PCRMutex_Lock(mutex);

    if (node->op->sidKey)
    {
        OALAVLMultiTree_Delete(sid_operation_tree, node->op->sidKey, node->op);
        node->op->sidKey = node->sidKey;
        if (!OALAVLMultiTree_Insert(sid_operation_tree, node->op->sidKey, node->op))
        {
            /* The operation will not be able to be looked up by sid for node down or similar, but should otherwise be ok*/
            DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "EPP negative one, update operation sid: Failed to instert into sid operation tree\n");
        }
    }
    /* If the the op's sidKey is not set then it was never added to the tree, so don't add it now. */

    PCRMutex_Unlock(mutex);

    DOFOperation_Internal_Destroy(node->op);
    DOFMem_Free(node);
}

boolean EPP_NegOne_UpdateOperationSid(DOFOperation op, sid_t * newSid)
{
    sidEnqueueNode_t * node;

    node = (sidEnqueueNode_t *)DOFMem_Alloc(sizeof(sidEnqueueNode_t));
    if (node)
    {
        OALEnqueueItem_Init((OALEnqueueItem)node, EPP_NegOne_UpdateOperationSid_Internal);
        DOFOperation_Retain(op);
        node->op = op;
        node->sidKey = newSid;
        if (!OALEnqueue_Enqueue(sid_operation_enqueue, (OALEnqueueItem)node))
        {
            DOFOperation_Internal_Destroy(node->op);
            DOFMem_Free(node);
            return FALSE;
        }
        return TRUE;
    }
    return FALSE;
}

static void EPP_NegOne_RemoveSidOperation(OALThreadPoolItem dnode, boolean shutdown)
{
    sidEnqueueNode_t * node;

    PCR_Assert(dnode != NULL);
    node = (sidEnqueueNode_t *)dnode;

    PCRMutex_Lock(mutex);

    if (node->op->sidKey)
    {
        OALAVLMultiTree_Delete(sid_operation_tree, node->op->sidKey, node->op);
        node->op->sidKey = NULL;
    }
    PCRMutex_Unlock(mutex);

    DOFOperation_Internal_Destroy(node->op);    /* It was retained before enqueue was called, so release it now. */
    DOFMem_Free(node);
}

static void channel_operation_observer(OALObserver self, OALObservable observable, void* ctx)
{
    DOFOperation op = (DOFOperation)ctx;
    PCR_Assert(op);

    /* if observable != NULL, then called with operation lock held */
    if(observable)
    {
        OperationContext_t * node = (OperationContext_t *) DOFMem_Alloc(sizeof(OperationContext_t));

        if(node != NULL)
        {
            OALEnqueueItem_Init((OALEnqueueItem) node, EPP_NegOne_Channel_Operation_Observer);
            node->operation = op;
            node->doFree = TRUE;
            if(OALOperationManager_EnqueueOperation((OALEnqueueItem) node))
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP negative one, channel operation observer: Failed to enqueue cleanup function." );

                /* no choice but to call function directly. */
                DOFOperation_Unlock(op);
                EPP_NegOne_Channel_Operation_Observer((OALThreadPoolItem) node, FALSE);
                DOFOperation_Lock(op);
            }
        }
        else
        {
            OperationContext_t node;
            OALEnqueueItem_Init((OALEnqueueItem) &node, EPP_NegOne_Channel_Operation_Observer);
            node.operation = op;
            node.doFree = FALSE;
            EPP_NegOne_Channel_Operation_Observer((OALThreadPoolItem) &node, FALSE);
        }
    }
    else
    {
        channelTreeItem_t *foundItem;
        channelTreeItem_t chtItem;
        sidEnqueueNode_t *  node;

        PCRMutex_Lock(mutex);
        DOFOperation_Lock(op);

        /* setup chtItem so the comparision functions will work for channel_operation_tree. */
        chtItem.channel = DOFOperation_GetSourceChannel(op);
        chtItem.address = op->sources->address;
        chtItem.op = op;

        /* since the channelTreeItem was allocated, find the item, so the channelTreeItem can be freed. */
        foundItem = (channelTreeItem_t *)OALAVLMultiTree_Retrieve_Item(channel_operation_tree, chtItem.channel, &chtItem);
        if (foundItem)
        {
            if (OALAVLMultiTree_Delete(channel_operation_tree, chtItem.channel, &chtItem))
            {
                OALChannel_Release(foundItem->channel);
                DOFAddress_Destroy(foundItem->address);
                DOFMem_Free(foundItem);
            }
        }

        node = (sidEnqueueNode_t *)DOFMem_Alloc(sizeof(sidEnqueueNode_t));

        if (node)
        {
            OALEnqueueItem_Init((OALEnqueueItem)node, EPP_NegOne_RemoveSidOperation);
            DOFOperation_Retain(op);
            node->op = op;
            node->sidKey = NULL;

            if (!OALEnqueue_Enqueue(sid_operation_enqueue, (OALEnqueueItem)node))
            {
                /* This should only fail if the enqueue has been shutdown, so just destroy node. */
                DOFOperation_Internal_Destroy(node->op);
                DOFMem_Free(node);
            }
        }
        else
        {
            /* Since memory could not be allocated for node, remove the op from the tree directly.
               The mutex protecting sid_operation_tree is already locked here and the enqueue is only used
               in this case for consistency, since everything uses the enqueue. */
            OALAVLMultiTree_Delete(sid_operation_tree, op->sidKey, op);
            op->sidKey = NULL;
        }
        DOFOperation_Unlock(op);
        PCRMutex_Unlock(mutex);
    }
}

EPP_NegOne EPP_NegOne_Create(BlackboardData blackboardData, boolean isConnectionless)
{
    EPP_NegOne retVal = (EPP_NegOne)DOFMem_Alloc( sizeof(EPP_NegOne_t) );

    if(!retVal)
        return NULL;

    retVal->blackboardData = blackboardData;
    retVal->isConnectionless = isConnectionless;

    return retVal;
}

static void EPP_NegOne_AddSidOperation(OALThreadPoolItem dnode, boolean shutdown)
{
    sidEnqueueNode_t * node;

    PCR_Assert(dnode != NULL);
    node = (sidEnqueueNode_t *)dnode;

    PCRMutex_Lock(mutex);

    node->op->sidKey = node->sidKey;
    OALAVLMultiTree_Insert(sid_operation_tree, node->sidKey, node->op);

    PCRMutex_Unlock(mutex);

    DOFOperation_Internal_Destroy(node->op);    /* It was retained before enqueue was called, so release it now. */
    DOFMem_Free(node);
}

static int32 EPP_NegOne_AddCommandOperation_Internal(DOFOperation op, OALChannel channel, boolean add_observer)
{
    channelTreeItem_t  *chItem;
    int32               rval = -1; /* failure */

    chItem = (channelTreeItem_t *)DOFMem_Alloc(sizeof(channelTreeItem_t));
    if (chItem)
    {
        PCRMutex_Lock(mutex);
        DOFOperation_Lock(op);

        /* setup chItem with the channel, address, and operation. The operation does not need to be retained since the
           observer that is added for the operation will keep the operation around. */
        chItem->channel = OALChannel_Retain(DOFOperation_GetSourceChannel(op));
        chItem->address = DOFAddress_Copy(op->sources->address);
        chItem->op = op;

        /* try to add our data to the tree */

        if (chItem->channel && chItem->address)
        {
            switch ( OALAVLMultiTree_Insert(channel_operation_tree, chItem->channel, chItem) )
            {
                case OALAVLTree_AlreadyInserted:
                    OALChannel_Release(chItem->channel);
                    DOFAddress_Destroy(chItem->address);
                    DOFMem_Free(chItem);
                    chItem = NULL;
                    rval = 0;
                    break;

                case OALAVLTree_InsertSuccess:
                    /* add our observer */
                    if ( !add_observer || DOFOperation_AddOpWillCloseObserver(op, channel_operation_observer) )
                    {
                        sidEnqueueNode_t *  node;

                        node = (sidEnqueueNode_t *)DOFMem_Alloc(sizeof(sidEnqueueNode_t));
                        if (node)
                        {
                            OALEnqueueItem_Init((OALEnqueueItem)node, EPP_NegOne_AddSidOperation);
                            DOFOperation_Retain(op);
                            node->op = op;
                            PCR_Assert(op->opID);
                            node->sidKey = DOFOperationID_GetSID(op->opID);
                            PCR_Assert(node->sidKey);

                            if (OALEnqueue_Enqueue(sid_operation_enqueue, (OALEnqueueItem)node))
                            {
                                rval = 0;
                            }
                            else
                            {
                                DOFOperation_Internal_Destroy(node->op);
                                DOFMem_Free(node);
                                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP negative one, Failed to enqueue operation add for sid_operation_tree");
                                DOFOperation_Cancel(op);    /* Just let the observer remove the op from the channel_operation_tree. */
                            }
                        }
                        else
                        {
                            DOFOperation_Cancel(op);    /* Just let the observer remove the op from the channel_operation_tree. */
                        }
                    }
                    else
                    {
                        OALAVLMultiTree_Delete(channel_operation_tree, chItem->channel, chItem);
                        OALChannel_Release(chItem->channel);
                        DOFAddress_Destroy(chItem->address);
                        DOFMem_Free(chItem);
                        DOFOperation_Cancel(op);
                    }
                    break;

                default:
                    OALChannel_Release(chItem->channel);
                    DOFAddress_Destroy(chItem->address);
                    DOFMem_Free(chItem);
                    DOFOperation_Cancel(op);
                    break;
            }
        }
        else
        {
            OALChannel_Release(chItem->channel);
            DOFAddress_Destroy(chItem->address);
            DOFMem_Free(chItem);
        }

        DOFOperation_Unlock(op);
        PCRMutex_Unlock(mutex);
    }
    return rval;
}

int32 EPP_NegOne_AddCommandOperation(DOFOperation op, OALChannel channel)
{
    return EPP_NegOne_AddCommandOperation_Internal(op, channel, TRUE);
}

static boolean channel_node_down_traverse(OALAVLMultiTreeItem_t item, void *arg)
{
    channelTreeItem_t *chItem = (channelTreeItem_t *)item;
    boolean rval = FALSE; /* Don't exit early. */
    if (DOFAddress_IsEqual(chItem->address, (DOFAddress)arg))
    {
        DOFOperation_Lock(chItem->op);

        if (!(chItem->op->flags.bits.cancelled))
        {
            /*
             * There are two possible reasons to be called:
             *   Cancel All PDU:
             *     We are traversing by (matching) operation ID (arg != NULL): cancel all (matching) operations.
             *   Node Down PDU or channel down notification:
             *     We are traversing by channel / address (node) because it is going or has gone down (arg == NULL):
             *       If the operation is not in the operation table, then there is no need to keep it.
             *       If seq == 0, then the node *must* be the source, and we need to cancel it.
             *       If seq > 0, then send source lost.
             */
            if ( arg || !chItem->op->flags.bits.added_to_operation_table || !chItem->op->sources->sequence || !EPP_NegOne_SourceLost( chItem->op ) )
            {
                DOFOperation_Internal_Cancel( chItem->op );
            }
        }
        DOFOperation_Unlock(chItem->op);
    }
    return rval;
}

static void EPP_NegOne_NotifyChannelDown_Internal(OALChannel channel, DOFAddress address)
{
    OALAVLMultiTree_Traverse(channel_operation_tree, channel, channel_node_down_traverse, address);
}

/* This function will return TRUE if the permissions are vaildated or if it is not secure and do not require validation. */
static boolean EPP_NegOne_ValidatePermissions(EPP_NegOne self, DOFObjectID oid, DOFPermissionType permType)
{
    boolean retVal = FALSE;

    if (self && oid)
    {
        PCR_Assert((permType == DOFPERMISSIONTYPE_I_AM) || (permType == DOFPERMISSIONTYPE_ACT_AS));

        if (self->blackboardData->associatedChannel->remoteChannelData->flags.sessionIsSecure &&
            self->blackboardData->associatedChannel->remoteChannelData->cipher)
        {
            DOFObjectIDSource   sourceID;

            sourceID = DOFObjectIDSource_Create(oid);
            if (sourceID)
            {
                sid_t *             sid;

                sid = lookup_explicit_sid(sourceID, DOFSecurityMembrane_GetDomainID(self->blackboardData->associatedChannel->securityMembrane));

                if (sid)
                {
                    OALPermissionScope  scope;

                    scope = DOFPermissionSet_Validate(OALCipher_GetRemotePerms( self->blackboardData->associatedChannel->remoteChannelData->cipher, self->blackboardData->keyState ),
                                                      (permType == DOFPERMISSIONTYPE_I_AM) ? sid->iamPermission : sid->actAsPermission);

                    if((scope != NULL) && (OALPermissionScope_GetIdCount( scope ) > 0))
                    {
                        retVal = TRUE;
                    }

                    OALPermissionScope_Destroy(scope);
                    source_release(sid);
                }
                DOFObjectIDSource_Destroy(sourceID);
            }
        }
        else
        {
            /* Since it's not secure the permissions don't need to be checked. */
            retVal = TRUE;
        }
    }

    return(retVal);
}

typedef struct EPP_NegOne_ExtendPermCtx_t
{
    DOFOperation pendingOperation;
    OALChannel channel;
    DOFAddress targetAddr;
    boolean isProviderBound;
}EPP_NegOne_ExtendPermCtx_t;

static void EPP_NegoneOne_ExtendCompleteCallback(DOFError status, void *context)
{
    EPP_NegOne_ExtendPermCtx_t *fullContext = (EPP_NegOne_ExtendPermCtx_t *)context;

    if (status == DOFERROR_OK)
    {
        fullContext->channel->routeToProvider(fullContext->channel, fullContext->pendingOperation, NULL, fullContext->targetAddr);
    }
    DOFOperation_Internal_Destroy(fullContext->pendingOperation);
    OALChannel_Release(fullContext->channel);
    DOFAddress_Destroy(fullContext->targetAddr);
    DOFMem_Free(fullContext);
}

static EPP_NegOne_ExtendPermCtx_t * EPP_Negone_CreateExtendPermCtx(OALChannel channel, DOFAddress address, DOFOperation op)
{
    EPP_NegOne_ExtendPermCtx_t *ctx = (EPP_NegOne_ExtendPermCtx_t *)DOFMem_Alloc(sizeof(EPP_NegOne_ExtendPermCtx_t));

    if (ctx)
    {
        ctx->channel = OALChannel_Retain(channel);
        if (ctx->channel)
        {
            ctx->targetAddr = DOFAddress_Copy(address);
            if (ctx->targetAddr)
            {
                ctx->pendingOperation = op;
                if (ctx->pendingOperation)
                {
                    DOFOperation_Retain(op);
                    return ctx;
                }
                DOFAddress_Destroy(ctx->targetAddr);
            }
            OALChannel_Release(ctx->channel);
        }
        DOFMem_Free(ctx);
    }
    return NULL;
}

/* This function will return TRUE if the permissions are vaildated or if it is not secure and do not require validation. */
static boolean EPP_NegOne_ValidateOutboundPermissions(EPP_NegOne self, DOFObjectID oid, DOFPermissionType permType, DOFOperation op)
{
    boolean retVal = FALSE;

    if (self && oid)
    {
        PCR_Assert((permType == DOFPERMISSIONTYPE_I_AM) || (permType == DOFPERMISSIONTYPE_ACT_AS));

        if (self->blackboardData->associatedChannel->remoteChannelData->flags.sessionIsSecure &&
            self->blackboardData->associatedChannel->remoteChannelData->cipher)
        {
            DOFObjectIDSource   sourceID;

            sourceID = DOFObjectIDSource_Create(oid);
            if (sourceID)
            {
                sid_t * sid = lookup_explicit_sid(sourceID, DOFCredentials_GetDomainID(self->blackboardData->associatedChannel->remoteChannelData->secureCredentials));

                if (sid)
                {
                    OALPermissionScope  scope;

                    scope = DOFPermissionSet_Validate(OALCipher_GetLocalPerms( self->blackboardData->associatedChannel->remoteChannelData->cipher, OALCipher_GetCurrentKeyState(self->blackboardData->associatedChannel->remoteChannelData->cipher)),
                                                      (permType == DOFPERMISSIONTYPE_I_AM) ? sid->iamPermission : sid->actAsPermission);

                    if (scope == NULL)
                    {
                        if (permType == DOFPERMISSIONTYPE_ACT_AS)
                        {
                            /* Try to extend permissions. */
                            DOFPermissionSetBuilder permSetBuilder = DOFPermissionSetBuilder_Create();

                            if (permSetBuilder)
                            {
                                if (DOFPermissionSetBuilder_AddPermission(permSetBuilder, sid->actAsPermission))
                                {
                                    DOFPermissionSet permSet = DOFPermissionSetBuilder_BuildAndDestroy(permSetBuilder);

                                    if (permSet)
                                    {
                                        EPP_NegOne_ExtendPermCtx_t * context = EPP_Negone_CreateExtendPermCtx(self->blackboardData->associatedChannel,
                                                                                                              self->blackboardData->transportAddr,
                                                                                                              op);
                                        if (context)
                                        {
                                            self->blackboardData->associatedChannel->extendPermissions(self->blackboardData->associatedChannel,
                                                                                                       self->blackboardData->transportAddr,
                                                                                                       permSet,
                                                                                                       EPP_NegoneOne_ExtendCompleteCallback,
                                                                                                       context);
                                        }
                                        DOFPermissionSet_Destroy(permSet);
                                    }
                                }
                                else
                                {
                                    DOFPermissionSetBuilder_Destroy(permSetBuilder);
                                }
                            }
                        }
                    }
                    else
                    {
                        if (OALPermissionScope_GetIdCount( scope ) > 0)
                        {
                            retVal = TRUE;
                        }
                    }

                    OALPermissionScope_Destroy(scope);
                    source_release(sid);
                }
                DOFObjectIDSource_Destroy(sourceID);
            }
        }
        else
        {
            /* Since it's not secure the permissions don't need to be checked. */
            retVal = TRUE;
        }
    }

    return(retVal);
}

static void context_destroy(void * operationContext)
{
    DOFOperation_Internal_Destroy( (DOFOperation)operationContext );
}

static boolean EPP_NegOne_SourceLost_Internal( EPP_NegOne self, DOFOperation op,  boolean floodOp )
{
    DOFOperation source_lost_op;

    /* Called with operation locked. */
    /* If "self" is NULL then this is a result of a channel or node down. */
    PCR_Assert(op);

    /* Do not send/forward if there is no opID or if the sequence will be greater than 255. */
    if (!op->opID || (op->sources->sequence > 0 && self && (self->blackboardData->seq >= 255)))
    {
        return TRUE;
    }

    source_lost_op = DOFOperation_SourceLost_Create();
    if ( !source_lost_op )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP Common, source lost: Failed to create source lost op." );
        return FALSE; /* no room for more state */
    }

    source_lost_op->context = (void *)op;
    DOFOperation_Retain( op );
    source_lost_op->contextDestroyFunction = context_destroy;

    if (!self && op->sources->sequence == 0)
    {
        /* If this is a local op set the sequence to 255 so it will not be forwarded by anyone who receives it. */
        source_lost_op->sources->sequence = 255;
    }
    else
    {
        if (self)
        {
            source_lost_op->sources->sequence = self->blackboardData->seq + 1;
        }
        else
        {
            source_lost_op->sources->sequence = 1;
        }
    }

    if (floodOp)
    {
        op->flags.bits.lost = 1; /* lost state */

        op->sources->distLossless = op->sources->distLossy = 0xffff;
        if ( DOFOperation_GetTimeRemaining( op ) > 16000 )
            DOFOperation_SetTimeout(op, 16000); /* cap timeout to 16s */

        OALRemoteChannelMgr_SendToAll(source_lost_op);
    }
    else
    {
        uint32 currentTime = PCRTime_GetMS();

        /* Only send this if on has not been sent in the last 16 seconds. */
        if ((op->proxySourceLostTime == 0) || ((currentTime - op->proxySourceLostTime) < 16000))
        {
            op->sources->channel->routeToProvider(op->sources->channel, source_lost_op, NULL, op->sources->address);
            op->proxySourceLostTime = PCRTime_GetMS();
        }
    }
    DOFOperation_Internal_Destroy(source_lost_op);
    return TRUE;
}

static boolean EPP_NegOne_SourceFound_Internal( EPP_NegOne self, DOFOperation op, boolean floodOp )
{
    DOFOperation source_found_op;

    /* Called with operation locked. */
    PCR_Assert(op);

    if ((uint32)op->sources->sequence + op->flags.bits.internally_created >= 255)
    {
        return TRUE;
    }

    source_found_op = DOFOperation_SourceFound_Create();
    if ( !source_found_op )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP Common, source found: Failed to create source found op." );
        return FALSE; /* no room for more state */
    }

    source_found_op->context = (void *)op;
    DOFOperation_Retain( op );
    source_found_op->contextDestroyFunction = context_destroy;

    OALRemoteChannelMgr_SendToAll(source_found_op);

    DOFOperation_Internal_Destroy(source_found_op);
    return TRUE;
}

void EPP_NegOne_NotifyChannelDown(OALChannel channel, DOFAddress address)
{
    PCRMutex_Lock(mutex);
    EPP_NegOne_NotifyChannelDown_Internal(channel, address);
    PCRMutex_Unlock(mutex);
}

void EPP_NegOne_Destroy(EPP_NegOne epp)
{
    if(!epp)
        return;

    DOFMem_Free( epp );
}

static ProtocolLayerResult EPP_NegOne_Receive_NodeDown(EPP_NegOne self)
{
    DOFObjectID    oid = DOFObjectID_Unmarshal(self->blackboardData->receiveStream);
    ProtocolLayerResult rval = PLRESULT_OKDONOTHING;

    if (EPP_NegOne_ValidatePermissions(self, oid, DOFPERMISSIONTYPE_I_AM))
    {
        if ( !oid || DOFObjectID_HasAttributes( oid ) || DOFObjectID_GetDataSize( oid ) > 20 || !EPP_NegOne_ProcessNodeDown( oid, self->blackboardData->associatedChannel, self->blackboardData->transportAddr, TRUE ) )
            rval = PLRESULT_ERROR;
    }

    DOFObjectID_Destroy( oid );
    return rval;
}

boolean EPP_NegOne_ProcessNodeDown( DOFObjectID oid, OALChannel channel, DOFAddress addr, boolean notifyChannelDown )
{
    const  uint8 * oid_bytes;

    if ( !oid || !channel || !addr )
        return FALSE;

    oid_bytes = DOFObjectID_GetBytes( oid );
    PCR_Assert( DOFObjectID_GetSize_Bytes( oid_bytes ) );

    PCRMutex_Lock( mutex );
    /* first cancel all nodes with matching SID */
    OALAVLMultiTree_Traverse( sid_operation_tree, oid_bytes, channel_node_down_traverse, (void*)1 );
    /* for any other operations, call the channel going down function for its channel/address pair even though
     it isn't really going down (Note: optimize me -- is it really going down?) */
    if (notifyChannelDown)
        EPP_NegOne_NotifyChannelDown_Internal( channel, addr );
    PCRMutex_Unlock( mutex );

    return TRUE;
}

static DOFOperationID unmarshal_opid(EPP_NegOne self)
{
    DOFOperationCount cnt;
    DOFObjectID oid = DOFObjectID_Unmarshal(self->blackboardData->receiveStream);
    DOFObjectIDSource srcid = DOFObjectIDSource_Create( oid );

    if ( !srcid || OALMarshal_StreamUncompress4(self->blackboardData->receiveStream, &cnt) != DOFERROR_OK)
        goto error1;

    /* See if the oid is our oid, either the secure sid for the channel (if secure) or our local oid. */
    if ( ( self->blackboardData->associatedChannel->remoteChannelData->flags.sessionIsSecure &&
               DOFObjectID_IsEqual( oid, DOFObjectIDSource_GetObjectID( self->blackboardData->associatedChannel->remoteChannelData->secureSid->oid ) ) ) ||
             ( !self->blackboardData->associatedChannel->remoteChannelData->flags.sessionIsSecure &&
               DOFObjectID_IsEqual( oid, DOFOperationID_GetLocalOID() ) ) )
    {
        /* This is us, create an opid. */
        DOFObjectID_Destroy(oid);
        DOFObjectIDSource_Destroy( srcid );
        return DOFOperationID_Create(cnt, DOFOperationID_GetLocalSID(), FALSE, TRUE);
    }
    else
    {
        /* received an opid on the wire. it will need a resolved sid, however */
        sid_t *sid;

        if (self->blackboardData->associatedChannel->remoteChannelData->secureCredentials)
        {
            sid = lookup_explicit_sid(srcid, DOFSecurityMembrane_GetDomainID(self->blackboardData->associatedChannel->securityMembrane));
        }
        else
        {
            sid = lookup_explicit_sid(srcid, NULL);
        }

        if(sid)
        {
            DOFObjectID_Destroy(oid);
            DOFObjectIDSource_Destroy( srcid );
            return DOFOperationID_Create(cnt, sid, FALSE, FALSE);
        }
        /* else fall through */
    }

error1:
    DOFObjectID_Destroy(oid);
    DOFObjectIDSource_Destroy( srcid );
    return NULL;
}

static ProtocolLayerResult EPP_NegOne_Receive_SourceLost(EPP_NegOne self)
{
    ProtocolLayerResult ret = PLRESULT_OKDONOTHING;

    DOFOperationID opid = unmarshal_opid(self);
    DOFOperation op;

    if(!opid) goto out;

    op = OALOperationManager_CheckExistingOperation(DOFSecurityMembrane_GetDomainID(self->blackboardData->associatedChannel->securityMembrane), opid);

    if(op)
    {
        DOFOperation_Lock(op);
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP negative one, received a Source Lost and foud a matching operation." );
        if (OALChannel_IsEquivalent(op->sources->channel, self->blackboardData->associatedChannel) && DOFAddress_IsEqual(op->sources->address, self->blackboardData->transportAddr))
        {
            /* If my parent sent the source lost then change the operation to the source lost state, and flood the source lost. */
            if (!EPP_NegOne_SourceLost_Internal(self, op, TRUE))
            {
                ret = PLRESULT_ERROR;
            }
        }
        else if (!op->flags.bits.internally_created)
        {
            /* This is a local operation or in other words we are the source so send a source found. */
            /* Do not send self so that the function knows that this is the source. (force sequence 1) */
            if (op->operation_type == DOFOPERATIONTYPE_BEGINPROVIDE)
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP negative one, receive Source Lost: Sending retry for provide." );
                DOFOperation_IncrementRetry(op);
                DOFSystem_CheckPermissionsRetryAndRoute(op);
            }
            else if (!EPP_NegOne_SourceFound_Internal(NULL, op, TRUE))
            {
                ret = PLRESULT_ERROR;
            }
        }
        else
        {
            /* Forward the source lost to my parent and set the proxySourceLostTime. */
            if (!EPP_NegOne_SourceLost_Internal(self, op, FALSE))
            {
                ret = PLRESULT_ERROR;
            }
        }
        DOFOperation_Unlock(op);

        DOFOperation_Internal_Destroy(op);
    }

    DOFOperationID_Destroy(opid);

out:
    return ret;
}

static ProtocolLayerResult EPP_NegOne_Receive_OperationRename(EPP_NegOne self)
{
    DOFOperationID      old_opid = NULL;
    DOFObjectIDDomain   domain;
    DOFOperation        op;
    DOFError            renameErr = DOFERROR_OK;

    /* unmarshal the opid */
    old_opid = unmarshal_opid(self);

    if(!old_opid)
        return PLRESULT_ERROR;

    /* The rename requires permissions to be checked on the new opID which was in the EPP header so permissions were already checked. */

    if (DOFObjectID_Compare(DOFOperationID_GetObjectID(old_opid), DOFOperationID_GetObjectID(self->blackboardData->opID)) == 0)
    {
        domain = DOFSecurityMembrane_GetDomainID(self->blackboardData->associatedChannel->securityMembrane);
        op = OALOperationManager_CheckExistingOperation(domain, old_opid);
        if (op)
        {
            DOFOperation_Lock(op);
            renameErr = OALOperationManager_RenameOperation(op, domain, self->blackboardData->opID);

            if (renameErr == DOFERROR_OK)
            {
                /* The rename was successfull so update the operation state. */
                op->sources->sequence = self->blackboardData->seq;
                op->sources->retry_count = self->blackboardData->retry;
                _operation_extend( op, self->blackboardData->delay, TRUE );
                op->sources->oal_timeout = self->blackboardData->delayWithoutGracePeriod;
            }

            DOFOperation_Unlock(op);
            /* since CheckExistingOperation retains the operation call internal destroy now. */
            DOFOperation_Internal_Destroy(op);
        }
    }

    /* old_opid was used to look up the operation and is no longer needed. */
    DOFOperationID_Destroy(old_opid);

    if (renameErr == DOFERROR_INSUFFICIENT_RESOURCES)
    {
        return PLRESULT_ERROR;
    }
    return PLRESULT_OKDONOTHING;
}

static ProtocolLayerResult EPP_NegOne_Receive_Ping(EPP_NegOne self)
{
    /* Special info: Per the spec, when Ping is sent secure, it may be sent multicast ONLY if it is "marked as unicast"
                     in the CCM header. After CCM.Decipher() has processed this incoming PDU, it has overwritten the
                     stack's understanding of transport by clearing self->blackboardData->flags.isMulticast to 0.
                     This test therefore will never fail when the incoming PDU is secure. When not secure, this test
                     could possibly fail out, as was always intended for that situation. */
    if(!self->blackboardData->flags.isMulticast)
    {
        /* immediately respond to the ping. */
        OALStream stream = OALStream_CreateRW( 3 );

        if ( stream )
        {
            if (OALMarshal_StreamCompress2(stream, EPP_NEG1_APPID) == DOFERROR_OK)
            {
                if (OALStream_Write8( stream, EPP_NEG1_OPCODE_PING ))
                {
                    /* Since this is called from a receive function the sendPacket may not be created yet. */
                    if(!self->blackboardData->sendPacket)
                    {
                        self->blackboardData->sendPacket = OALPacket_Create();
                    }
                    if ( self->blackboardData->sendPacket )
                    {
                        if (OALPacket_PrependStream( self->blackboardData->sendPacket, stream ))
                        {
                            self->blackboardData->flags.isCommandNotResponse = FALSE;
                            return PLRESULT_OKSENDDOWN;
                        }
                    }
                }
            }
            OALStream_Release(stream, FALSE);
        }
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP negative one, receive ping: Failed to reply to ping." );
    }
    /* silently ignoring ping from multicast/broadcast, or error */
    return PLRESULT_OKDONOTHING;
}

static ProtocolLayerResult EPP_NegOne_Receive_PingResponse(EPP_NegOne self)
{
    return PLRESULT_OKDONOTHING;
}

static ProtocolLayerResult EPP_NegOne_Receive_CancelAll(EPP_NegOne self)
{
    DOFObjectID    oid = DOFObjectID_Unmarshal(self->blackboardData->receiveStream);
    ProtocolLayerResult rval = PLRESULT_OKDONOTHING;

    if (EPP_NegOne_ValidatePermissions(self, oid, DOFPERMISSIONTYPE_I_AM))
    {
        if (!oid || DOFObjectID_HasAttributes( oid ) || DOFObjectID_GetDataSize( oid ) > 20 || !EPP_NegOne_ProcessNodeDown( oid, self->blackboardData->associatedChannel, self->blackboardData->transportAddr, FALSE))
            rval = PLRESULT_ERROR;
    }

    DOFObjectID_Destroy(oid);
    return rval;
}

static ProtocolLayerResult EPP_NegOne_Receive_Heartbeat(EPP_NegOne self)
{
    return PLRESULT_OKDONOTHING;
}

static ProtocolLayerResult EPP_NegOne_Receive_Query(EPP_NegOne self)
{
    if(!self->blackboardData->flags.isMulticast)
    {
        DOFObjectID sid;

        if(self->blackboardData->associatedChannel->remoteChannelData->flags.sessionIsSecure &&
           (self->blackboardData->associatedChannel->remoteChannelData->cipher != NULL))
        {
            sid = DOFObjectIDSource_GetObjectID(self->blackboardData->associatedChannel->remoteChannelData->secureSid->oid);
        }
        else
        {
            sid = DOFOperationID_GetLocalOID();
        }

        /* immediately respond to the query. */
        if (sid)
        {
            OALStream       stream;
            DOFObjectIDSize sidSize;

            sidSize = DOFObjectID_GetBytesSize( sid );
            stream = OALStream_CreateRW( 3 + sidSize );

            if (stream)
            {
                if (OALMarshal_StreamCompress2(stream, EPP_NEG1_APPID) == DOFERROR_OK)
                {
                    if (OALStream_Write8( stream, EPP_NEG1_OPCODE_QUERY ))
                    {
                        if (DOFObjectID_Marshal(sid, stream))
                        {
                            /* Since this is called from a receive function the sendPacket may not be created yet. */
                            if (!self->blackboardData->sendPacket)
                                self->blackboardData->sendPacket = OALPacket_Create();
                            if (self->blackboardData->sendPacket)
                            {
                                if (OALPacket_PrependStream( self->blackboardData->sendPacket, stream ))
                                {
                                    self->blackboardData->flags.isCommandNotResponse = FALSE;
                                    return PLRESULT_OKSENDDOWN;
                                }
                            }
                        }
                    }
                }
                OALStream_Release(stream, FALSE);
            }
        }
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP negative one, receive query: Failed to respond to a sid query." );
        return PLRESULT_ERROR;
    }

    /* silently ignoring query from multicast/broadcast */
    return PLRESULT_OKDONOTHING;
}

static ProtocolLayerResult EPP_NegOne_Receive_QueryResponse(EPP_NegOne self)
{
    DOFObjectID oid = DOFObjectID_Unmarshal(self->blackboardData->receiveStream);
    DOFObjectIDSource srcid = DOFObjectIDSource_Create( oid );
    ProtocolLayerResult rval = PLRESULT_OKDONOTHING;

    if ( !srcid )
    {
        rval = PLRESULT_ERROR;
    }
    else
    {
        if (EPP_NegOne_ValidatePermissions(self, oid, DOFPERMISSIONTYPE_I_AM))
        {
            sid_t *sid;
            DOFOperation queryOp;

            if (!self->blackboardData->associatedChannel->remoteChannelData->remoteSid &&
                self->blackboardData->associatedChannel->remoteChannelData->flags.isConnection)
            {
                DOFConnectionType connType;

                connType = DOFConnectionConfig_GetConnectionType(self->blackboardData->associatedChannel->remoteChannelData->config.connection);
                if ((connType == DOFCONNECTIONTYPE_DATAGRAM) || (connType == DOFCONNECTIONTYPE_STREAM))
                {
                    self->blackboardData->associatedChannel->remoteChannelData->remoteSid = DOFObjectIDSource_Copy(srcid);

                    if (!self->blackboardData->associatedChannel->remoteChannelData->remoteSid)
                    {
                        rval = PLRESULT_ERROR;
                    }
                }
            }

            sid = lookup_sid(self->blackboardData->associatedChannel, self->blackboardData->transportAddr);

            if(sid)
            {
                if (!sid->oid)
                {
                    if (source_set_oid(sid, srcid) != 0)
                    {
                        rval = PLRESULT_ERROR;
                    }
                }
                source_release(sid);
            }

            queryOp = OALOperationManager_CheckExistingOperation(DOFSecurityMembrane_GetDomainID(self->blackboardData->associatedChannel->securityMembrane), self->blackboardData->opID);
            if (queryOp)
            {
                SidQueryContext_t * context;

                DOFOperation_Lock(queryOp);
                context = (SidQueryContext_t *)queryOp->context;
                queryOp->context = NULL;
                OALOperationManager_RemoveOperationAndUnlock(queryOp);

                if (context)
                {
                    if (context->address && context->channel)
                    {
                        sid_t * ctxSid;

                        ctxSid = lookup_sid(context->channel, context->address);
                        if (ctxSid)
                        {
                            if (!ctxSid->oid)
                            {
                                if (source_set_oid(ctxSid, srcid) != 0)
                                {
                                    rval = PLRESULT_ERROR;
                                }
                            }
                            source_release(ctxSid);
                        }
                    }
                    if (queryOp->callbacks.negOne)
                    {
                        if (rval == PLRESULT_OKDONOTHING)
                        {
                            queryOp->callbacks.negOne->fns->Complete(context->context, TRUE);
                        }
                        else
                        {
                            queryOp->callbacks.negOne->fns->Complete(context->context, FALSE);
                        }
                    }

                    DOFAddress_Destroy(context->address);
                    OALChannel_Release(context->channel);
                    DOFMem_Free(context);
                }

                DOFOperation_Cancel(queryOp);
                DOFOperation_Internal_Destroy(queryOp);
            }
        }
    }

    DOFObjectIDSource_Destroy( srcid );
    DOFObjectID_Destroy(oid);
    return rval;
}

ProtocolLayerResult EPP_NegOne_ReparentOperation(DOFOperation op, OALChannel new_channel, DOFAddress new_address, boolean isConnectionless)
{
    OALChannel          old_channel = NULL;
    OALChannel          channel = NULL;
    DOFAddress          address = NULL;
    ProtocolLayerResult ret = PLRESULT_OKDONOTHING;
    DOFSecurityScope    securityScope;
    DOFSecurityScope    securityScope_newChannel;

    PCRMutex_Lock(mutex);
    DOFOperation_Lock(op);

    old_channel = op->sources->channel;

    PCR_Assert(old_channel);

    securityScope = OALChannel_GetSecurityScope(old_channel, TRUE);
    securityScope_newChannel = OALChannel_GetSecurityScope(new_channel, TRUE);
    address = DOFAddress_Copy(new_address);
    channel = OALChannel_Retain(new_channel);
    if ( !DOFSecurityScope_IsCompatible(securityScope_newChannel, securityScope) || ( new_address && !address ) || (new_channel && !channel) )
    {
        DOFOperation_Internal_Cancel(op); /* no memory to handle reparenting, so destroy it */
        DOFAddress_Destroy( address );
        OALChannel_Release(channel);
        ret = PLRESULT_ERROR;
    }
    else
    {
        channel_operation_observer(NULL, NULL, op); /* clean up old entries */
        DOFAddress_Destroy( op->sources->address );
        op->sources->address = address;
        op->sources->channel = channel;

        op->flags.bits.is_connectionless = isConnectionless;
        op->flags.bits.lost = 0;
        router_notify_reparent(op, old_channel, address);

        /* add the new state back to our tables.  If this fails there is not much that can be done,
           so let the operation time out or get cancelled. */
        (void)EPP_NegOne_AddCommandOperation_Internal( op, new_channel, FALSE );                  /* If this call fails, it logs a message. */

        OALChannel_Release(old_channel);
    }

    DOFOperation_Unlock(op);
    PCRMutex_Unlock(mutex);

    return ret;
}

static ProtocolLayerResult EPP_NegOne_Receive_SourceFound(EPP_NegOne self)
{
    DOFOperationID opid;
    ProtocolLayerResult ret = PLRESULT_OKDONOTHING;

    opid = unmarshal_opid(self);

    if(opid)
    {
        DOFOperation op;

        if (EPP_NegOne_ValidatePermissions(self, DOFOperationID_GetObjectID(opid), DOFPERMISSIONTYPE_ACT_AS))
        {
            op = OALOperationManager_CheckExistingOperation(DOFSecurityMembrane_GetDomainID(self->blackboardData->associatedChannel->securityMembrane),opid);
            if(op)
            {
                DOFOperation_Lock(op);

                if (op->flags.bits.lost)
                {
                    /* Set the operation state to the source found state. */
                    op->sources->sequence = self->blackboardData->seq;
                    op->sources->retry_count = self->blackboardData->retry;
                    _operation_extend( op, self->blackboardData->delay, TRUE );
                    op->sources->oal_timeout = self->blackboardData->delayWithoutGracePeriod;
                    if(OALChannel_IsEquivalent(op->sources->channel, self->blackboardData->associatedChannel) && DOFAddress_IsEqual(op->sources->address, self->blackboardData->transportAddr))
                    {
                        op->flags.bits.lost = 0;
                        ret = PLRESULT_OKDONOTHING; /* we "re-parented" with the same source -- boring */
                    }
                    else
                    {
                        ret = EPP_NegOne_ReparentOperation(op, self->blackboardData->associatedChannel, self->blackboardData->transportAddr, self->isConnectionless);
                    }

                    /* Make sure the proxySourceLostTime is zero. */
                    op->proxySourceLostTime = 0;
                    /* Flood out the source found. */
                    EPP_NegOne_SourceFound_Internal(self, op, TRUE);
                }
                else if (op->proxySourceLostTime)
                {
                    if (OALChannel_IsEquivalent(op->sources->channel, self->blackboardData->associatedChannel) && DOFAddress_IsEqual(op->sources->address, self->blackboardData->transportAddr))
                    {
                        /* When the proxySourceLostTime is set, ignore the source found if it is not from our parent. */
                        if ((PCRTime_GetMS() - op->proxySourceLostTime) < 16000)
                        {
                            /* If a proxy source lost time was set less than 16 seconds ago and the source found was sent by our parent,
                               set the proxySourceLostTime to 0 and flood the source found. */
                            op->proxySourceLostTime = 0;
                            EPP_NegOne_SourceFound_Internal(self, op, TRUE);
                        }
                        else
                        {
                            /* Set the proxy source lost time to zero, but do not send the source found. */
                            op->proxySourceLostTime = 0;
                        }
                    }
                }
                else
                {
                    /* If the op is not in the lost state and the proxySourceLostTime is zero then do not send out the source found. */
                }
                DOFOperation_Unlock(op);

                /* Call internal destroy to offset CheckExistingOperation. */
                DOFOperation_Internal_Destroy(op);
            }
        }

        DOFOperationID_Destroy(opid);
    }

    return ret;
}

#define VERIFY_STREAM_MINSIZE( stream, size ) do{ if (OALStream_GetDataSize(stream) < size){    \
                                                   retVal = PLRESULT_ERROR;                     \
                                                   goto CLEANUPANDEXIT;}                        \
                                              }while (0)

ProtocolLayerResult EPP_NegOne_Receive(EPP_NegOne self)
{
    OALStream           stream = self->blackboardData->receiveStream;
    uint8               opcode;
    ProtocolLayerResult retVal = PLRESULT_ERROR;
    uint16              appID;

#define OPCODE_LIMIT 9

    static const EPP_NegOne_Receive_fn_t receiveFunctions[OPCODE_LIMIT][2] = {
        { 0, EPP_NegOne_Receive_NodeDown },
        { 0, EPP_NegOne_Receive_SourceLost },
        { 0, EPP_NegOne_Receive_OperationRename },
        { EPP_NegOne_Receive_PingResponse, EPP_NegOne_Receive_Ping },
        { 0, EPP_NegOne_Receive_CancelAll },
        { 0, EPP_NegOne_Receive_Heartbeat },
        { EPP_NegOne_Receive_QueryResponse, EPP_NegOne_Receive_Query },
        { 0, 0 },
        { 0, EPP_NegOne_Receive_SourceFound }
    };

    if(self->blackboardData->associatedChannel->remoteChannelData->flags.sessionIsSecure &&
       !(self->blackboardData->flags.receivedAuthenticated || self->blackboardData->flags.receivedEncrypted))
    {
        return PLRESULT_ERROR;
    }

    VERIFY_STREAM_MINSIZE( stream, 3 );

    if (OALMarshal_StreamUncompress2(stream, &appID) == DOFERROR_OK)
    {
        PCR_Assert(appID == EPP_NEG1_APPID); /* this asserts because the appid was peeked at before this stream got routed here. Therefore it is programmer error */

        if (OALStream_Read8(stream, &opcode))
        {
            if(opcode < OPCODE_LIMIT)
            {
                EPP_NegOne_Receive_fn_t fn = receiveFunctions[opcode][self->blackboardData->flags.isCommandNotResponse];
                if(fn)
                {
                    retVal = fn(self);
                }
            }
        }
    }

    CLEANUPANDEXIT:
    return retVal;
}

static OALStream EPP_NegOne_Send_Lost_Found(EPP_NegOne self, DOFOperationID id, DOFOperation op, uint8 opcode)
{
    DOFOperationID tempID = NULL;

    if(self->blackboardData->associatedChannel->remoteChannelData->flags.sessionIsSecure &&
       (self->blackboardData->associatedChannel->remoteChannelData->cipher != NULL))
    {
        sid_t * sid = DOFOperationID_GetSID(id);

        if(sid)
        {
            sid_t * localSid = DOFOperationID_GetLocalSID();

            if(DOFObjectID_Compare(DOFObjectIDSource_GetObjectID(sid->oid), DOFObjectIDSource_GetObjectID(localSid->oid)) == 0)
            {
                sid = self->blackboardData->associatedChannel->remoteChannelData->secureSid;

                if (sid)
                {
                    tempID = DOFOperationID_Create(DOFOperationID_GetCount(id), sid, FALSE, TRUE);
                }
            }
            else
            {
                tempID = DOFOperationID_Copy(id);
            }
        }
    }
    else
    {
        tempID = DOFOperationID_Copy(id);
    }

    if (tempID)
    {
        if ((opcode == EPP_NEG1_OPCODE_SOURCELOST) || EPP_NegOne_ValidateOutboundPermissions(self, DOFOperationID_GetObjectID(tempID), DOFPERMISSIONTYPE_ACT_AS, op))
        {
            OALStream stream = OALStream_CreateRW(DOFOperationID_GetSize(tempID)+1/*opcode*/+2/*protocol id*/);

            if(stream)
            {
                if (OALMarshal_StreamCompress2(stream, EPP_NEGONE_ID) == DOFERROR_OK)
                {
                    if (OALStream_Write8(stream, opcode))
                    {
                        if (DOFOperationID_Marshal(stream, tempID) == DOFERROR_OK)
                        {
                            DOFOperationID_Destroy(tempID);
                            return stream;
                        }
                    }
                }
                OALStream_Release(stream, FALSE);
            }
        }
        DOFOperationID_Destroy(tempID);
    }
    return NULL;
}

static ProtocolLayerResult send_source_lost(EPP_NegOne self, DOFOperation op)
{
    OALStream stream;

    DOFOperation_Lock(op);
    if(!op->flags.bits.cancelled && op->context)
    {
        DOFOperation lost_op = (DOFOperation)op->context;

        DOFOperation_Unlock(op);
        DOFOperation_Lock(lost_op);
        if (lost_op->flags.bits.cancelled || DOFOperation_GetTimeRemaining( lost_op ) == 0)
        {
            DOFOperation_Unlock(lost_op);
            return PLRESULT_OKDONOTHING;
        }

        self->blackboardData->flags.isCommandNotResponse = TRUE;
        self->blackboardData->seq = op->sources->sequence;
        stream = EPP_NegOne_Send_Lost_Found(self, lost_op->opID, op, EPP_NEG1_OPCODE_SOURCELOST);

        DOFOperation_Unlock(lost_op);

        if(!stream || !OALPacket_AppendStream(self->blackboardData->sendPacket, stream))
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP negative one, send source lost: Failed to append source lost data to packet." );
            return PLRESULT_ERROR;
        }

        return PLRESULT_OKSENDDOWN;
    }

    DOFOperation_Unlock(op);
    return PLRESULT_ERROR;
}

static ProtocolLayerResult send_node_down(EPP_NegOne self, DOFOperation op)
{
    uint32 stream_size = 3; /* opcode + protocol id */
    OALStream stream;
    const uint8 *oid_bytes;
    uint8 oid_bytes_size;

    if(self->blackboardData->associatedChannel->remoteChannelData->flags.sessionIsSecure &&
       (self->blackboardData->associatedChannel->remoteChannelData->cipher != NULL))
    {
        oid_bytes = DOFObjectID_GetBytes(DOFObjectIDSource_GetObjectID(self->blackboardData->associatedChannel->remoteChannelData->secureSid->oid));
    }
    else
    {
        oid_bytes = DOFObjectID_GetBytes(DOFOperationID_GetLocalOID());
    }

    if (!oid_bytes)
    {
        return PLRESULT_ERROR;
    }

    oid_bytes_size = DOFObjectID_GetSize_Bytes(oid_bytes);
    PCR_Assert(oid_bytes_size);

    stream_size += oid_bytes_size;

    stream = OALStream_CreateRW(stream_size);
    if(stream)
    {
        if (OALMarshal_StreamCompress2(stream, EPP_NEGONE_ID) == DOFERROR_OK)
        {
            if (OALStream_Write8(stream, EPP_NEG1_OPCODE_NODEDOWN))
            {
                if (OALStream_Write(stream, oid_bytes_size, oid_bytes))
                {
                    if(OALPacket_AppendStream(self->blackboardData->sendPacket, stream))
                    {
                        self->blackboardData->flags.isCommandNotResponse = TRUE;

                        return PLRESULT_OKSENDDOWN;
                    }
                }
            }
        }
        OALStream_Release(stream, FALSE);
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP negative one, send node down: Failed to send node down." );
    return PLRESULT_ERROR;
}

static ProtocolLayerResult send_cancelAll(EPP_NegOne self, DOFOperation op)
{
    uint32 stream_size = 3; /* opcode + protocol id */
    OALStream stream;
    const uint8 *oid_bytes;
    uint8 oid_bytes_size;

    if(self->blackboardData->associatedChannel->remoteChannelData->flags.sessionIsSecure &&
       (self->blackboardData->associatedChannel->remoteChannelData->cipher != NULL))
    {
        oid_bytes = DOFObjectID_GetBytes(DOFObjectIDSource_GetObjectID(self->blackboardData->associatedChannel->remoteChannelData->secureSid->oid));
    }
    else
    {
        oid_bytes = DOFObjectID_GetBytes(DOFOperationID_GetLocalOID());
    }

    if (!oid_bytes)
    {
        return PLRESULT_ERROR;
    }

    oid_bytes_size = DOFObjectID_GetSize_Bytes(oid_bytes);
    PCR_Assert(oid_bytes_size);

    stream_size += oid_bytes_size;

    stream = OALStream_CreateRW(stream_size);
    if(stream)
    {
        if (OALMarshal_StreamCompress2(stream, EPP_NEGONE_ID) == DOFERROR_OK)
        {
            if (OALStream_Write8(stream, EPP_NEG1_OPCODE_CANCELALL))
            {
                if (OALStream_Write(stream, oid_bytes_size, oid_bytes))
                {
                    if(OALPacket_AppendStream(self->blackboardData->sendPacket, stream))
                    {
                        self->blackboardData->flags.isCommandNotResponse = TRUE;

                        return PLRESULT_OKSENDDOWN;
                    }
                }
            }
        }
        OALStream_Release(stream, FALSE);
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP negative one, send node down: Failed to send cancell all." );
    return PLRESULT_ERROR;
}

static ProtocolLayerResult send_heartbeat(EPP_NegOne self, DOFOperation op)
{
    OALStream stream = OALStream_CreateRW(1/*opcode*/+2/*protocol id*/);
    if(stream)
    {
        if (OALMarshal_StreamCompress2(stream, EPP_NEGONE_ID) == DOFERROR_OK)
        {
            if (OALStream_Write8(stream, EPP_NEG1_OPCODE_HEARTBEAT))
            {
                self->blackboardData->flags.isCommandNotResponse = TRUE;
                self->blackboardData->flags.multicastProtocolNotAllowed = 1;
                /* This next is set only if we are being sent as a reaction to an incoming CCM.Probe PDU. */
                self->blackboardData->flags.denyPacketReplayProtectionIfSecureUnicastSend = op->flags.bits.denyPacketReplayProtectionIfSecureUnicastSend;
                if (op->opID)
                {
                    self->blackboardData->opID = DOFOperationID_Copy(op->opID);
                }

                if (self->blackboardData->opID || !op->opID)
                {
                    if(OALPacket_AppendStream(self->blackboardData->sendPacket, stream))
                    {
                        return PLRESULT_OKSENDDOWN;
                    }
                }
            }
        }
        OALStream_Release(stream, FALSE);
    }

    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP negative one, send heartbeat: Failed to send heartbeat." );
    return PLRESULT_ERROR;
}

static ProtocolLayerResult send_ping(EPP_NegOne self, DOFOperation op)
{
    OALStream stream = OALStream_CreateRW(1/*opcode*/+2/*protocol id*/);

    if(stream)
    {
        if (OALMarshal_StreamCompress2(stream, EPP_NEGONE_ID) == DOFERROR_OK)
        {
            if (OALStream_Write8(stream, EPP_NEG1_OPCODE_PING))
            {
                self->blackboardData->flags.multicastProtocolNotAllowed = 1;
                self->blackboardData->flags.isCommandNotResponse = TRUE;
                self->blackboardData->secureMarkingID = (uint32)op->context;
                if (op->opID)
                    self->blackboardData->opID = DOFOperationID_Copy(op->opID);

                if (self->blackboardData->opID || !op->opID)
                {
                    if(OALPacket_AppendStream(self->blackboardData->sendPacket, stream))
                    {
                        return PLRESULT_OKSENDDOWN;
                    }
                }
            }
        }
        OALStream_Release(stream, FALSE);
    }

    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP negative one, send ping: Failed to send ping." );
    return PLRESULT_ERROR;
}

static ProtocolLayerResult send_query(EPP_NegOne self, DOFOperation op)
{
    OALStream stream = OALStream_CreateRW(1/*opcode*/+2/*protocol id*/);

    if(stream)
    {
        if (OALMarshal_StreamCompress2(stream, EPP_NEGONE_ID) == DOFERROR_OK)
        {
            if (OALStream_Write8(stream, EPP_NEG1_OPCODE_QUERY))
            {
                self->blackboardData->flags.isCommandNotResponse = TRUE;
                self->blackboardData->flags.multicastProtocolNotAllowed = 1;

                if (op->opID)
                    self->blackboardData->opID = DOFOperationID_Copy(op->opID);

                if (self->blackboardData->opID || !op->opID)
                {
                    if (OALPacket_AppendStream(self->blackboardData->sendPacket, stream))
                    {
                        return PLRESULT_OKSENDDOWN;
                    }
                }
            }
        }
        OALStream_Release(stream, FALSE);
    }

    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP negative one, send query: Failed to send query." );
    return PLRESULT_ERROR;
}

static ProtocolLayerResult send_source_found(EPP_NegOne self, DOFOperation op)
{
    DOFOperation_Lock(op);
    if(!op->flags.bits.cancelled && op->context)
    {
        DOFOperation lost_op = (DOFOperation)op->context;
        OALStream stream;

        DOFOperation_Unlock(op);
        DOFOperation_Lock(lost_op);
        if ( !lost_op->flags.bits.cancelled &&
             DOFOperation_GetTimeRemaining( lost_op ) > 0 &&
             lost_op->sources->sequence + lost_op->flags.bits.internally_created < 255 )
        {
            self->blackboardData->flags.isCommandNotResponse = TRUE;
            self->blackboardData->delay = DOFOperation_GetTimeRemaining(lost_op);
            self->blackboardData->seq = lost_op->sources->sequence + lost_op->flags.bits.internally_created;
            self->blackboardData->retry = lost_op->sources->retry_count;
            stream = EPP_NegOne_Send_Lost_Found(self, lost_op->opID, op, EPP_NEG1_OPCODE_SOURCEFOUND);

            DOFOperation_Unlock(lost_op);

            if(!stream || !OALPacket_AppendStream(self->blackboardData->sendPacket, stream))
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP negative one, send source found: Failed to append source found data to packet." );
                return PLRESULT_ERROR;
            }
            return PLRESULT_OKSENDDOWN;
        }

        DOFOperation_Unlock( lost_op );
        return PLRESULT_OKDONOTHING;
    }

    DOFOperation_Unlock(op);
    return PLRESULT_ERROR;
}


ProtocolLayerResult EPP_NegOne_Send(EPP_NegOne self, DOFOperation op)
{
    ProtocolLayerResult result;

    /* Note: this list is in operation type order. */
    static const EPP_NegOne_Send_fn_t sendFunctions[] = {
        send_source_lost,
        send_node_down,
        send_cancelAll,
        send_heartbeat,
        send_ping,
        send_query,
        send_source_found
    };

    PCR_Assert(self);
    PCR_Assert(self->blackboardData);
    PCR_Assert(self->blackboardData->sendPacket);
    PCR_Assert(op->operation_type>=DOFOPERATIONTYPE_EPP_NEGONE_SOURCELOST&&op->operation_type<DOFOPERATIONTYPE_EPP_NEGONE_LAST);

    result = sendFunctions[op->operation_type-DOFOPERATIONTYPE_EPP_NEGONE_SOURCELOST](self, op);

    if (self->blackboardData->opID && !self->blackboardData->seq)
        self->blackboardData->seq = 1;

    return result;
}

void EPP_NegOne_GoingDown(void)
{
    DOFOperation node_down_op = DOFOperation_NodeDown_Create();

    if (node_down_op)
    {
        OALRemoteChannelMgr_SendToAll(node_down_op);
        DOFOperation_Internal_Destroy(node_down_op);
    }
}

void EPP_NegOne_NodeDown(OALChannel channel)
{
    DOFOperation node_down_op = DOFOperation_NodeDown_Create();

    if (node_down_op && channel)
    {
        node_down_op->sources->channel = OALChannel_Retain(channel);
        if (node_down_op->sources->channel)
        {
            node_down_op->sources->oal_timeout = 1000;

            channel->routeToProvider(channel, node_down_op, NULL, node_down_op->sources->address);
        }
    }
    DOFOperation_Internal_Destroy(node_down_op);
}

void EPP_NegOne_CancelAll(void)
{
    DOFOperation op = DOFOperation_CancelAll_Create();

    if (op)
    {
        OALRemoteChannelMgr_SendToAll(op);
        DOFOperation_Internal_Destroy(op);
    }
    else
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP negative one, send cancel all: Failed to create operation." );
    }
}

/*  Send an EPP heartbeat.  A channel may be provided to direct the heartbeat.
    If channel is NULL a heartbeat is sent to all channels.
*/
void EPP_NegOne_Heartbeat(OALChannel channel, DOFAddress address, boolean denyPacketReplayProtectionIfSecureUnicastSend)
{
    DOFOperation heartbeat_op = DOFOperation_EppHeartbeat_Create(address);

    if (heartbeat_op)
    {
        if (channel)
        {
            heartbeat_op->sources->channel = OALChannel_Retain(channel);

            if (heartbeat_op->sources->channel)
            {
                DOFOperation_SetTimeout(heartbeat_op, 1000);
                /* This next is set only when we are returning EPP heartbeat to sender of a CCM.Probe PDU. */
                heartbeat_op->flags.bits.denyPacketReplayProtectionIfSecureUnicastSend =
                                                                      denyPacketReplayProtectionIfSecureUnicastSend ? 1 : 0;

                channel->routeToProvider( channel, heartbeat_op, NULL, heartbeat_op->sources->address );
            }
        }
        else
        {
            OALRemoteChannelMgr_SendToAll(heartbeat_op);
        }
        DOFOperation_Internal_Destroy(heartbeat_op);
    }
}

void EPP_NegOne_Ping(OALChannel channel, DOFAddress address, uint32 secureMarkingID)
{
    DOFOperation ping_op = DOFOperation_EppPing_Create(address);

    if (ping_op)
    {
        if (channel)
        {
            ping_op->sources->channel = OALChannel_Retain(channel);

            if (ping_op->sources->channel)
            {
                ping_op->context = (void *)secureMarkingID;

                channel->routeToProvider( channel, ping_op, NULL, ping_op->sources->address );
            }
            else
            {
                DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "EPP negative one ping: Failed to retain channel for ping operation");
            }
        }
        DOFOperation_Internal_Destroy(ping_op);
    }
    else
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP negative one ping: Failed to create ping operation." );
}

static void EPP_NegOne_SidQuery_Observer_fn(OALObserver self, OALObservable observable, void* ctx)
{
     DOFOperation queryOp = (DOFOperation)ctx;
    if (queryOp->context)
    {
        if (queryOp->callbacks.negOne)
            queryOp->callbacks.negOne->fns->Complete(((SidQueryContext_t *)queryOp->context)->context, FALSE);
        DOFAddress_Destroy(((SidQueryContext_t *)queryOp->context)->address);
        OALChannel_Release(((SidQueryContext_t *)queryOp->context)->channel);
    }
    DOFOperation_Internal_Destroy(queryOp);
}

static boolean EPP_NegOne_SidQuery_SetupOp(DOFOperation query_op, OALChannel channel, uint32 timeout, uint32 retry)
{
    query_op->opID = DOFOperationID_CreateLocal();

    if (query_op->opID)
    {
        if (!query_op->operationConfig)
            query_op->operationConfig = DOFOperationControl_Create();
        if (query_op->operationConfig)
        {
            if (DOFOperation_AddOpWillCloseObserver(query_op, EPP_NegOne_SidQuery_Observer_fn))
            {
                if (retry)
                {
                    DOFOperationControl_SetRetryPeriod(query_op->operationConfig, retry);
                    DOFOperationManager_SetRetry(query_op, channel);
                }
                if (timeout)
                {
                    DOFOperation_SetTimeout(query_op, timeout);
                }
                if (OALOperationManager_AddOperation(query_op, query_op->opID, TRUE) == DOFERROR_OK)
                {
                    return TRUE;
                }
            }
        }
    }
    return FALSE;
}

boolean EPP_NegOne_Query(OALChannel channel, DOFAddress address, uint32 timeout, uint32 retry,
                      void * context, EPPNegOneQueryCallback callbackFn)
{
    DOFOperation query_op = DOFOperation_EppQuery_Create(channel, address);

    PCR_Assert(channel&&address);

    if(query_op)
    {
        if (context && callbackFn)
        {
            SidQueryContext_t * ctx;

            if (EPP_NegOne_SidQuery_SetupOp(query_op, channel, timeout, retry))
            {

                ctx = (SidQueryContext_t *)DOFMem_Alloc(sizeof(SidQueryContext_t));

                if (!ctx)
                {
                    DOFOperation_Destroy(query_op);
                    return FALSE;
                }
                ctx->address = DOFAddress_Copy(address);
                if (!ctx->address)
                {
                    DOFMem_Free(ctx);
                    DOFOperation_Destroy(query_op);
                    return FALSE;
                }
                ctx->channel = OALChannel_Retain(channel);
                if (!ctx->channel)
                {
                    DOFAddress_Destroy(ctx->address);
                    DOFMem_Free(ctx);
                    DOFOperation_Destroy(query_op);
                    return FALSE;
                }
                ctx->context = context;
                query_op->context = ctx;
                query_op->callbacks.negOne = callbackFn;

                channel->routeToProvider(channel, query_op, NULL, query_op->sources->address);
                DOFOperation_Internal_Destroy(query_op);
                return TRUE;
            }
            else
            {
                DOFOperation_Destroy(query_op);
                return FALSE;
            }
        }
        else
        {
            /* If a context is not provided just send without an opID. */
            channel->routeToProvider(channel, query_op, NULL, query_op->sources->address);
            DOFOperation_Internal_Destroy(query_op);
            return TRUE;
        }
    }
    else
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP negative one query: Failed to create query op." );
    }
    return FALSE;
}

boolean EPP_NegOne_SourceLost(DOFOperation op)
{
    return EPP_NegOne_SourceLost_Internal(NULL, op, TRUE);
}

