/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/interfaceid.h>
#include <dof/objectid.h>
#include <dof/operation.h>
#include <dof/server.h>
#include <internal/blackboarddata.h>
#include <internal/channel_stack_messages.h>
#include <internal/cipher.h>
#include <internal/dofinternal.h>
#include <internal/epp_v2.h>
#include <internal/epp_negative_one.h>
#include <internal/marshal.h>
#include <internal/objectidinternal.h>
#include <internal/operationinternal.h>
#include <internal/permissioninternal.h>
#include <internal/remotechannel.h>
#include <internal/remotechannelmanager.h>
#include <internal/securegroupinternal.h>
#include <internal/source.h>
#include <internal/error.h>

extern PCRMutex ref_count_mutex;

/*
**  "Private" Type and Method Declarations
*/

typedef struct EPP_V2_t
{
    BlackboardData      blackboardData;
    boolean             isConnectionLess;

    /* Flag byte. Defaults to zero on creation. */
    uint8               flagsByte;

} EPP_V2_t;

#define EPPHDR_VERSION_MASK     0x7f
#define EPPHDR_FLAGS_MASK       0x80

#define EPPV2_FLAG_SECURITY_HEADER              0x80
#define EPPV2_FLAG_OPID                         0x60
#define EPPV2_FLAG_OPID_IMPLICIT_SEND           0x20
#define EPPV2_FLAG_OPID_IMPLICIT_RECV           0x40
#define EPPV2_FLAG_OPID_EXPLICIT                0x60
#define EPPV2_FLAG_OPID_NONE                    0x00
#define EPPV2_FLAG_RESPONSE                     0x10
#define EPPV2_FLAG_COMMAND_RESERVED             0x08
#define EPPV2_FLAG_COMMAND_SEQUENCE             0x04
#define EPPV2_FLAG_COMMAND_RETRY                0x02
#define EPPV2_FLAG_COMMAND_COMPLETES            0x01

#define EPPV2_FLAG_RESPONSE_RESERVED            0x0E
#define EPPV2_FLAG_RESPONSE_INTERMEDIATE        0x01


#define EPPV2_SECURITYHEADER_FLAG_SECURE        0x80
#define EPPV2_SECURITYHEADER_FLAG_RESERVED      0x70
#define EPPV2_SECURITYHEADER_FLAG_DOMAINID      0x08
#define EPPV2_SECURITYHEADER_FLAG_SECURITYSCOPE 0x04
#define EPPV2_SECURITYHEADER_FLAG_AS            0x02
#define EPPV2_SECURITYHEADER_FLAG_STATE         0x01


#define EPPV2_FLAG_DEFAULTS     0x00

struct traverseDecipherContext
{
    BlackboardData      blackboardData;
    OALStream           stream;
    uint32              offsetToBeginEPP;
    uint32              count;
};
/*
**  "Public" Methods
*/

EPP_V2  EPP_V2_Create(BlackboardData blackboard, boolean isConnectionLess)
{
    EPP_V2 retVal = (EPP_V2)DOFMem_Alloc( sizeof(EPP_V2_t) );

    if (retVal)
    {
        retVal->blackboardData = blackboard;
        retVal->isConnectionLess = isConnectionLess;

        retVal->flagsByte = EPPV2_FLAG_DEFAULTS;
    }

    return retVal;
}

void EPP_V2_Destroy(EPP_V2 epp)
{
    DOFMem_Free(epp);
}

#define VERIFY_STREAM_MINSIZE( stream, size ) do{ if (OALStream_GetDataSize(stream) < size){    \
                                                   retVal = PLRESULT_ERROR;                     \
                                                   goto CLEANUPANDEXIT;}                        \
                                              }while (0)

static boolean EPP_V2_ValidateTunnelPermission(uint32 remoteDomainID, OALCipher cipher, OALCipherKeyState keyState)
{
    boolean retStatus = TRUE;

    if (remoteDomainID != INVALID_REMOTE_DOMAIN_ID)
    {
        DOFPermission rdidTunnelPerm;
        OALPermissionScope scope = NULL;

        rdidTunnelPerm = DOFPermissionTunnelDomain_Create(remoteDomainID);

        if (rdidTunnelPerm)
        {
            scope = DOFPermissionSet_Validate(OALCipher_GetRemotePerms( cipher, keyState ), rdidTunnelPerm);
        }

        if ((scope == NULL) || (OALPermissionScope_GetIdCount(scope) == 0))
        {
            retStatus = FALSE;
        }

        OALPermissionScope_Destroy(scope);
        DOFPermission_Destroy(rdidTunnelPerm);
    }
    return retStatus;
}

static boolean EPP_V2_Decipher(BlackboardData blackboardData, OALStream stream, uint32 offsetToBeginEPP)
{
    boolean           unicastBehaviorMarkedBySender;
    DecipherKeyType_t decipherKeyType;
    DOFError          decipherStatus;
    OALCipher         cipher = blackboardData->associatedChannel->remoteChannelData->cipher;

    if (!cipher)
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "EPP_V2_Deciper: decipher called for a channel withouth a cipher.");
        return FALSE;
    }

    PCR_Assert( blackboardData->transportAddr );
    PCR_Assert( !OALCipher_IsSecureGroup( cipher ) || /*cipher was configured with same channel as ->associatedChannel*/
                OALCipher_GetConnectionChannel( cipher ) == blackboardData->associatedChannel );
    /* Note: stream is given here with its read offset at the 1st byte of its embedded CCM security mode header.
                    That point is a few bytes before the encryption boundary.
             offsetToBeginEPP is well before that. From that place _Decipher() computes MAC.
                    That point is known as the security boundary. */
    decipherStatus = OALCipher_Decipher( cipher, &stream, offsetToBeginEPP,
                                         0,             /* payloadEncryptOffset- may be 0 forever. */
                                         &blackboardData->transportAddr, (boolean *)&unicastBehaviorMarkedBySender,
                                         &decipherKeyType, &blackboardData->keyState );
    /* Note: decipherStatus may not be a DOFStatus that is in the public API and should not be returned to the app. */
    if (decipherStatus == DOFERROR_OK_NODEID_COLLISION)
    {
        /* Only here when self->blackboardData->associatedChannel is a secureGroup. */
        PCR_Assert( OALCipher_IsSecureGroup( cipher ) );
        /* Though this next could perhaps be called within the above _Decipher() call (because cipher was configured
           with this same channel), it is an EPP activity and is best done here. */
        OALChannel_SetSecureSendBlocked( blackboardData->associatedChannel );
        /* Accept incoming PDU, anyway, by falling through. */
    }
    else if (decipherStatus != DOFERROR_OK)
    {
        return FALSE;
    }

    /* The sender of secure packets controls our understanding of how we receive them. */
    blackboardData->flags.isMulticast = unicastBehaviorMarkedBySender ? 0 : 1;

    if (OALCipher_IsHub( cipher ))
    {
        DOFHub_UpdateAddressTable( (DOFSecureGroup)blackboardData->associatedChannel,
                                   blackboardData->transportAddr, FALSE, 0 );
        /* This can return FALSE if this is not a point member for the hub or there was a failure.
           Since this only updates a timeout and decipher did not fail, continue to process this packet. */
    }

    return TRUE;
}

static boolean EPP_V2_Traverse_SSIDs(OALChannel secureChannel, void* context)
{
    struct traverseDecipherContext* ctx = (struct traverseDecipherContext*)context;
    OALStream newStream = NULL;

    PCR_Assert(ctx->blackboardData->associatedChannel == NULL);

    if (ctx->count > 1)
    {
        --(ctx->count);
        newStream = OALStream_CloneToRW(ctx->stream);

        if (!newStream)
        {
            /* There is no memory to create the new stream so skip testing for this channel and go to the next.
               If this channel was correct this failure will cause the packet to be dropped. */
            return FALSE;
        }
    }
    else
    {
        newStream = ctx->stream;
    }

    /* Since this is only for connectionless a failure to retain the channel will result in a dropped packet. */
    ctx->blackboardData->associatedChannel = OALChannel_Retain(secureChannel);

    if (ctx->blackboardData->associatedChannel)
    {
        if (EPP_V2_Decipher(ctx->blackboardData, newStream, ctx->offsetToBeginEPP))
        {
            if (newStream != ctx->stream)
            {
                /* Use the new stream. */
                OALStream_Release(ctx->stream, TRUE);
                ctx->stream = newStream;
            }
            return TRUE;
        }

        /* Right SSID wrong channel, set the channel to NULL and see if the next one works. */
        OALChannel_Release(ctx->blackboardData->associatedChannel);
        ctx->blackboardData->associatedChannel = NULL;
    }

    if (newStream != ctx->stream)
    {
        /* Since it didn't work, discard the new stream. */
        OALStream_Release(newStream, FALSE);
    }

    return FALSE;
}

ProtocolLayerResult EPP_V2_Receive( EPP_V2 self )
{
    OALStream         stream = self->blackboardData->receiveStream;
    uint32            offsetToBeginEPP;
    uint8             header;
    uint8             securityHeaderFlagsByte;
    boolean           flagsPresent;
    DOFOperationCount opCount;
    DOFObjectIDSource pktID = NULL;

    ProtocolLayerResult retVal = PLRESULT_ERROR;
    offsetToBeginEPP = OALStream_GetReadOffset( stream );

    VERIFY_STREAM_MINSIZE( stream, 1 );
    OALStream_Read8(stream, &header);

    /* Verify EPP Version header */
    if ( (header & EPPHDR_VERSION_MASK) != EPPV2_VERSION )
    {
        retVal = PLRESULT_ERROR;
        goto CLEANUPANDEXIT;
    }

    PCR_Assert( self->blackboardData->associatedChannel );
    /* If a EPP_V0 query is in progress set the bit to indicate that we received a v2 packet. */
    if (self->blackboardData->associatedChannel->remoteChannelData->flags.EPP_V0_QueryInProgress &&
        self->blackboardData->flags.isMulticast)
    {
        self->blackboardData->associatedChannel->remoteChannelData->EPP_V0_QueryVersions |= V0_QUERY_BIT_EPP_V2;
    }

    /* Check for flags */
    flagsPresent = header >> 0x07;

    if (flagsPresent)
    {
        VERIFY_STREAM_MINSIZE( stream, 1 );
        OALStream_Read8(stream, &(self->flagsByte));
    }
    else
        self->flagsByte = EPPV2_FLAG_DEFAULTS;

    if ( ( self->flagsByte & EPPV2_FLAG_OPID ) == EPPV2_FLAG_OPID_EXPLICIT )
    {
        /* Read Explicit SID. */
        DOFObjectID oid = DOFObjectID_Unmarshal(stream);

        pktID = DOFObjectIDSource_Create( oid );
        DOFObjectID_Destroy( oid );
        if ( pktID == NULL )
        {
            retVal = PLRESULT_ERROR;
            goto CLEANUPANDEXIT;
        }
        else
        { /* Here for legal pktID: not broadcast, no attributes, and its oid data length is below 21. */
        }
    }

    /* Note: ->flagsByte can have both 0x60 bits set for cmd and resp. Can be 0x20 bit for cmd. Can be 0x40 for resp.*/
    if ( self->flagsByte & EPPV2_FLAG_OPID )
    {
        /* read Operation Count */
        if ( OALMarshal_StreamUncompress4( stream, &opCount ) != DOFERROR_OK )
        {
            retVal = PLRESULT_ERROR;
            goto CLEANUPANDEXIT;
        }
    }

    /* OPID's are actually created towards the end of this function */

    if ( self->flagsByte & EPPV2_FLAG_RESPONSE )
    {
        /* It is an error for a response to indicate implicit sender. */
        if ((self->flagsByte & EPPV2_FLAG_OPID) == EPPV2_FLAG_OPID_IMPLICIT_SEND)
        {
            retVal = PLRESULT_ERROR;
            goto CLEANUPANDEXIT;
        }

        self->blackboardData->flags.isCommandNotResponse = FALSE;
        self->blackboardData->flags.responseIsIntermediate = self->flagsByte & EPPV2_FLAG_RESPONSE_INTERMEDIATE;
    }
    else
    {
        /* It is an error for a command to indicate implicit receiver. */
        if ((self->flagsByte & EPPV2_FLAG_OPID) == EPPV2_FLAG_OPID_IMPLICIT_RECV)
        {
            retVal = PLRESULT_ERROR;
            goto CLEANUPANDEXIT;
        }

        self->blackboardData->flags.isCommandNotResponse = TRUE;
        self->blackboardData->flags.commandCompletesOnAnyFinal = self->flagsByte & EPPV2_FLAG_COMMAND_COMPLETES;

        if ( self->flagsByte & EPPV2_FLAG_COMMAND_SEQUENCE )
        {
            /* read SEQ field */
            VERIFY_STREAM_MINSIZE( stream, 1 );
            OALStream_Read8(stream, &(self->blackboardData->seq));
            self->blackboardData->flags.isFlooded = ( self->blackboardData->seq > 0 );
        }

        if ( self->flagsByte & EPPV2_FLAG_COMMAND_RETRY )
        {
            /* read Retry field */
            VERIFY_STREAM_MINSIZE( stream, 1 );
            OALStream_Read8(stream, &(self->blackboardData->retry));
        }

        {
            /* read Delay field and add the grace period. */
            uint8 delayByte = 0;
            VERIFY_STREAM_MINSIZE( stream, 1 );
            OALStream_Read8(stream, &delayByte);
            if (delayByte == 0)
                self->blackboardData->delay = 0;
            else if (delayByte <= 128)
            {
                self->blackboardData->delayWithoutGracePeriod = 1000UL *delayByte;
                self->blackboardData->delay = self->blackboardData->delayWithoutGracePeriod + 4000UL;
            }
            else
            {
                self->blackboardData->delayWithoutGracePeriod = 32000UL * (delayByte-124);
                self->blackboardData->delay = self->blackboardData->delayWithoutGracePeriod + 250UL * (delayByte-112);
            }
        }
    }


    /* SECURITY UNMARSHAL **********************/

    if ( self->flagsByte & EPPV2_FLAG_SECURITY_HEADER )
    {
        uint32    securityStateID = INVALID_SECURITYSTATEID;

        VERIFY_STREAM_MINSIZE( stream, 1 );
        OALStream_Read8(stream, &securityHeaderFlagsByte);

        if ( securityHeaderFlagsByte & EPPV2_SECURITYHEADER_FLAG_STATE )
        {
            /* Get the security state ID. */
            if ( OALMarshal_StreamUncompress4( stream, &securityStateID ) != DOFERROR_OK )
            {
                retVal = PLRESULT_ERROR;
                goto CLEANUPANDEXIT;
            }
            self->blackboardData->securityStateID = securityStateID;
        }
        else
        {
            /* The default for the SSID is zero. */
            self->blackboardData->securityStateID = 0;
        }

        if (securityHeaderFlagsByte & EPPV2_SECURITYHEADER_FLAG_DOMAINID)
        {
            if (!self->isConnectionLess)
            {
                uint32 remoteDomainID;

                /* We are not set up to handle tunneling at this point, so we will just drop this packet.
                   For for streaming connections we still need to keep nonce state in sync, so we need to continue to unmarshal and decrypt this packet. */

                /* Unmarshal the Remote Domain Identifier and the Remote Operation Security Scope. */
                if (OALMarshal_StreamUncompress4(stream, &remoteDomainID) != DOFERROR_OK)
                {
                    retVal = PLRESULT_ERROR;
                    goto CLEANUPANDEXIT;
                }
                self->blackboardData->remoteDomainID = remoteDomainID;
                self->blackboardData->remoteOpSecurityScope = DOFSecurityScope_Unmarshal(stream, DOFSecurityScope_GetDomainID(self->blackboardData->associatedChannel->securityScope));
                if (self->blackboardData->remoteOpSecurityScope == NULL)
                {
                    retVal = PLRESULT_ERROR;
                    goto CLEANUPANDEXIT;
                }
            }
            else
            {
                /* We are not set up to handle tunneling at this point, so we will just drop this packet.
                   for connectionless that means we can drop it now. */
                retVal = PLRESULT_OKDONOTHING;
                goto CLEANUPANDEXIT;
            }
        }

        if ( securityHeaderFlagsByte & EPPV2_SECURITYHEADER_FLAG_SECURITYSCOPE )
        {
            /* Unmarshal the Security Scope. */
            self->blackboardData->securityScope = DOFSecurityScope_Unmarshal(stream, DOFSecurityScope_GetDomainID(self->blackboardData->associatedChannel->securityScope));
            if (self->blackboardData->securityScope == NULL)
            {
                retVal = PLRESULT_ERROR;
                goto CLEANUPANDEXIT;
            }
        }

        if ( self->isConnectionLess )
        {
            if ( securityHeaderFlagsByte & EPPV2_SECURITYHEADER_FLAG_AS )
            {
                if (securityHeaderFlagsByte & EPPV2_SECURITYHEADER_FLAG_SECURE)
                {
                    /* For point connections the channel that receives the datagram will have the SSID (the channel is the secure group) */
                    PCR_Assert( self->blackboardData->associatedChannel );
                    if (self->blackboardData->associatedChannel->remoteChannelData->securityStateID !=
                        self->blackboardData->securityStateID)
                    {
                        struct traverseDecipherContext  ctx;
                        OALChannel                      baseChannel;
                        OALChannel                      previousChannel;

                        if (self->blackboardData->associatedChannel->remoteChannelData->baseChannel)
                            baseChannel = self->blackboardData->associatedChannel->remoteChannelData->baseChannel;
                        else
                            baseChannel = self->blackboardData->associatedChannel;

                        /* Set associatedChannel to NULL.  When the packet is deciphered it will be set to a valid secure channel.
                           If it cannot be deciphered by any of the channels associated with the security state ID the
                           associatedChannel will still be set to NULL. */
                        previousChannel = self->blackboardData->associatedChannel;
                        self->blackboardData->associatedChannel = NULL;

                        ctx.blackboardData = self->blackboardData;
                        /* The context will take ownership of stream and may destroy it and replace it with a new stream. */
                        ctx.stream = stream;
                        ctx.offsetToBeginEPP = offsetToBeginEPP;
                        ctx.count = OALRemoteChannel_GetRegisteredChannelCountBySSID_As(baseChannel, self->blackboardData->securityStateID);

                        OALRemoteChannel_TraverseSecureChannelsBySSID_As(EPP_V2_Traverse_SSIDs,
                                                                         baseChannel, self->blackboardData->securityStateID, &ctx);
                        /* Note: stream may have been destroyed when the list was traversed.  do not use it until after
                           this next check and the context returns ownership to stream. */
                        if (ctx.stream != stream)
                        {
                            stream = ctx.stream;
                        }

                        /* previousChannel is not released until now since this channel allows access to the base channel.
                           The base channel is not retained above, so keeping previousChannel around until this point it guarantees
                           access to the base channel until the new channel is retained. */
                        OALChannel_Release(previousChannel);
                        if (!self->blackboardData->associatedChannel)
                        {
                            DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "EPP V2 Receive: lookup by security state ID failed to locate a registered channel.");
                            retVal = PLRESULT_ERROR;
                            goto CLEANUPANDEXIT;
                        }
                    }
                    else
                    {
                        if (!EPP_V2_Decipher(self->blackboardData, stream, offsetToBeginEPP))
                        {
                            retVal = PLRESULT_ERROR;
                            goto CLEANUPANDEXIT;
                        }
                    }
                }
                else
                {
                    /* The packet was not sent secure so the secure channel must be found by traversing the list of
                       secure channels by state ID later in the protocol stack. */
                    self->blackboardData->flags.getChannelBySSIDFromAsInProtocolStack = TRUE;
                    self->blackboardData->flags.receivedUnsecure = TRUE;
                }
            }
            else
            {
                if (!self->blackboardData->associatedChannel->remoteChannelData->flags.isConnection ||
                     (DOFConnectionConfig_GetConnectionType(self->blackboardData->associatedChannel->remoteChannelData->config.connection) == DOFCONNECTIONTYPE_DATAGRAM_STATELESS))
                {
                    OALChannel  baseChannel;

                    if (self->blackboardData->associatedChannel->remoteChannelData->baseChannel)
                        baseChannel = self->blackboardData->associatedChannel->remoteChannelData->baseChannel;
                    else
                        baseChannel = self->blackboardData->associatedChannel;

                    if (baseChannel->remoteChannelData->registeredSSIDs_NonAs)
                    {
                        OALChannel  previousChannel;
                        OALChannel  newChannel;

                        previousChannel = self->blackboardData->associatedChannel;
                        newChannel = OALRemoteChannel_GetSecureChannelByStateID_NonAs(baseChannel, securityStateID );
                        self->blackboardData->associatedChannel = OALChannel_Retain(newChannel);
                        OALChannel_Release(previousChannel);
                        if (!self->blackboardData->associatedChannel)
                        {
                            retVal = PLRESULT_ERROR;
                            goto CLEANUPANDEXIT;
                        }
                    }
                }
                if (!self->blackboardData->associatedChannel->remoteChannelData->flags.isConnection)
                {
                    retVal = PLRESULT_ERROR;
                    goto CLEANUPANDEXIT;
                }

                if (securityHeaderFlagsByte & EPPV2_SECURITYHEADER_FLAG_SECURE)
                {
                    if (!EPP_V2_Decipher(self->blackboardData, stream, offsetToBeginEPP))
                    {
                        retVal = PLRESULT_ERROR;
                        goto CLEANUPANDEXIT;
                    }
                }
                else
                    self->blackboardData->flags.receivedUnsecure = TRUE;
            }
        }
        else
        {
            if (securityHeaderFlagsByte & EPPV2_SECURITYHEADER_FLAG_SECURE)
            {
                if (!EPP_V2_Decipher(self->blackboardData, stream, offsetToBeginEPP))
                {
                    retVal = PLRESULT_ERROR;
                    goto CLEANUPANDEXIT;
                }

                if (securityHeaderFlagsByte & EPPV2_SECURITYHEADER_FLAG_DOMAINID)
                {
                    /* We are not set up to handle tunneling at this point, so we will just drop this packet.
                       Now that decipher was called the nonce state has been updated, so return PLRESULT_OKDONOTHING which will cause this packet to be dropped. */
                    retVal = PLRESULT_OKDONOTHING;
                    goto CLEANUPANDEXIT;
                }
            }
            else
            {
                self->blackboardData->flags.receivedUnsecure = TRUE;
            }
        }

        self->blackboardData->receiveStream = stream;

        if (self->blackboardData->associatedChannel->remoteChannelData->cipher)
        {
            if (self->blackboardData->flags.receivedUnsecure)
                self->blackboardData->flags.receivedUnsecureOnSecureChannel = TRUE;
            else
            {
                /* a cipher can encrypt and authenticate or just do one or the other. */
                if (OALCipher_IsEncrypt(self->blackboardData->associatedChannel->remoteChannelData->cipher))
                    self->blackboardData->flags.receivedEncrypted = TRUE;
                if (OALCipher_IsAuthenticate(self->blackboardData->associatedChannel->remoteChannelData->cipher))
                    self->blackboardData->flags.receivedAuthenticated = TRUE;
            }
        }
    }
    else
    {
        self->blackboardData->flags.receivedUnsecure = TRUE;
    }

    /************************************************/

    if ( ( self->flagsByte & EPPV2_FLAG_OPID ) == EPPV2_FLAG_OPID_EXPLICIT )
    {
        if ( ( self->blackboardData->associatedChannel->remoteChannelData->flags.sessionIsSecure &&
               DOFObjectID_IsEqual( DOFObjectIDSource_GetObjectID( pktID ), DOFObjectIDSource_GetObjectID( self->blackboardData->associatedChannel->remoteChannelData->secureSid->oid ) ) ) ||
             ( !self->blackboardData->associatedChannel->remoteChannelData->flags.sessionIsSecure &&
               DOFObjectID_IsEqual( DOFObjectIDSource_GetObjectID( pktID ), DOFOperationID_GetLocalOID() ) ) )
        {
            if (self->blackboardData->flags.isCommandNotResponse)
            {
                retVal = PLRESULT_OKDONOTHING;
                goto CLEANUPANDEXIT;
            }
            /* pktID is my own */
            self->blackboardData->opID = DOFOperationID_Create(opCount, DOFOperationID_GetLocalSID(), FALSE, TRUE);
        }
        else if ( self->blackboardData->flags.isCommandNotResponse )
        {
            /* First, lookup in explicit sourceID tree using the pktID from the packet. */
            sid_t *sid;

            if (self->blackboardData->flags.receivedUnsecure)
            {
                sid = lookup_explicit_sid(pktID, NULL);
            }
            else
            {
                sid = lookup_explicit_sid(pktID, DOFSecurityMembrane_GetDomainID(self->blackboardData->associatedChannel->securityMembrane));
            }

            if ( !sid )
            {
                /* Not found via explicit pktID. */
                if ( self->blackboardData->seq == 0 )
                {
                    /* We have a pktID and a sequence 0, so look up by channel/address. */
                    sid = lookup_sid(self->blackboardData->associatedChannel, self->blackboardData->transportAddr);
                    if ( !sid )
                    {
                        /* Not found (we don't have it yet), and we have everything needed to completely populate a SID, so create it. */
                        goto CREATENEWSID;
                    }
                    else
                    {
                        /* Found in ->associatedChannel lookup. */
                        int rval = source_set_oid( sid, pktID );

                        if ( rval < 0 )
                        {
                            /* Error. */
                            source_release(sid);
                            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "EPP V2 Receive: Failed to set existing sid oid." );
                            retVal = PLRESULT_OKDONOTHING;
                            goto CLEANUPANDEXIT;
                        }
                        else if ( rval > 0 )
                        {
                            /* sid already has an oid. Treat the old SID as NodeDown, and create a new SID. */
                            EPP_NegOne_ProcessNodeDown( DOFObjectIDSource_GetObjectID( sid->oid ),
                                                        self->blackboardData->associatedChannel,
                                                        self->blackboardData->transportAddr, TRUE );
                            source_set_channel_address( sid, NULL, NULL );
                            source_release( sid );
                            goto CREATENEWSID;
                        }
                        else
                        { /* Just use sid already looked up. */ }
                    }
                }
                else
                {
                    /* Not found (we don't have it yet), and we have everything needed to completely populate a SID, so create it. */
                  CREATENEWSID:
                    if (self->blackboardData->flags.receivedUnsecure)
                    {
                        sid = create_sid_from_oid(pktID, NULL);
                    }
                    else
                    {
                        sid = create_sid_from_oid(pktID, DOFSecurityMembrane_GetDomainID(self->blackboardData->associatedChannel->securityMembrane));
                    }

                    if ( !sid )
                    {
                        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "EPP V2: Failed to create sid from source ID." );
                        retVal = PLRESULT_OKDONOTHING;
                        goto CLEANUPANDEXIT;
                    }
                }
            }
            PCR_Assert(sid != NULL);        /* We should always have a sid by this point */
            if(self->blackboardData->associatedChannel->remoteChannelData->flags.sessionIsSecure &&
               (self->blackboardData->associatedChannel->remoteChannelData->cipher != NULL))    /* We're secure, so let's check the identity permission */
            {
                OALPermissionScope scope;

                scope = DOFPermissionSet_Validate(OALCipher_GetRemotePerms( self->blackboardData->associatedChannel->remoteChannelData->cipher, self->blackboardData->keyState ),
                                                  (self->blackboardData->seq == 0) ? sid->iamPermission
                                                                                   : sid->actAsPermission);
                if((scope == NULL) || (OALPermissionScope_GetIdCount(scope) == 0))    /* Permission unknown (is this wrong?) or denied */
                {
                    source_release(sid);
                    OALPermissionScope_Destroy(scope);
                    retVal = PLRESULT_OKDONOTHING;
                    goto CLEANUPANDEXIT;
                }
                /* If we get here, everything is ok */
                OALPermissionScope_Destroy(scope);

                if (!EPP_V2_ValidateTunnelPermission(self->blackboardData->remoteDomainID, self->blackboardData->associatedChannel->remoteChannelData->cipher, self->blackboardData->keyState))
                {
                    source_release(sid);
                    retVal = PLRESULT_OKDONOTHING;
                    goto CLEANUPANDEXIT;
                }
            }
            if ( self->blackboardData->seq == 0 )
            {
                /* This next returns 0 for success or -1 for failure. */
                if (source_set_channel_address( sid, self->blackboardData->associatedChannel, self->blackboardData->transportAddr ) != 0)
                {
                    /* Error. */
                    source_release(sid); /* can't do anything with it */
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "EPP V2: Failed to set existing sid channel and address." );
                    retVal = PLRESULT_OKDONOTHING;
                    goto CLEANUPANDEXIT;
                }
            }

            /* at this point we can create opid */
            self->blackboardData->opID = DOFOperationID_Create(opCount, sid, FALSE, FALSE);
            if ( !self->blackboardData->opID )
                source_release( sid );
            else
            { /* sid ownership transferred to new ->opID. */
            }
        }
        else
        {
            /* Responses always have the original pktID their command was sent with. */
            sid_t *sid;

            if (self->blackboardData->flags.receivedUnsecure)
            {
                sid = lookup_explicit_sid(pktID, NULL);
            }
            else
            {
                sid = lookup_explicit_sid(pktID, DOFSecurityMembrane_GetDomainID(self->blackboardData->associatedChannel->securityMembrane)); 
            }

            if (!sid)
            {
                /*No record of sid at all. This means the source is unknown so we cannot accept this incoming response*/
                retVal = PLRESULT_OKDONOTHING;
                goto CLEANUPANDEXIT;
            }
            self->blackboardData->opID = DOFOperationID_Create(opCount, sid, FALSE, FALSE);
            if ( !self->blackboardData->opID )
                source_release( sid );
            /* sid ownership transferred to new ->opID. */
        }

        if ( !self->blackboardData->opID )
        {
            /* OPID could not be created, so drop the packet. */
            retVal = PLRESULT_OKDONOTHING;
            goto CLEANUPANDEXIT;
        }
    }
    else if ( ( self->flagsByte & EPPV2_FLAG_OPID ) == EPPV2_FLAG_OPID_IMPLICIT_SEND )
    {
        /* Implicit Sender pktID: Receive pkt is a command, opCount was filled from the wire. */
        sid_t *sid = lookup_sid(self->blackboardData->associatedChannel, self->blackboardData->transportAddr);
        /* sid info is looked for in source.c's source_tree. */

        if(!sid)
        {
            if (self->blackboardData->associatedChannel->remoteChannelData->secureCredentials)
            {
                sid = create_sid_from_channel_address(self->blackboardData->associatedChannel,
                                                      self->blackboardData->transportAddr,
                                                      DOFCredentials_GetDomainID(self->blackboardData->associatedChannel->remoteChannelData->secureCredentials));
            }
            else
            {
                sid = create_sid_from_channel_address(self->blackboardData->associatedChannel, self->blackboardData->transportAddr, NULL);
            }
        }
        if(sid)
        {
            self->blackboardData->opID = DOFOperationID_Create(opCount, sid, TRUE, FALSE);
            if ( !self->blackboardData->opID )
                source_release( sid );      /* Note: sid not removed from source.c's source_tree for this error case. */
            else
            { /* sid ownership transferred to new ->opID. Note: sid may or may not be in the explicit sid tree. */
            }
        }
        else
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "EPP V2: Failed to create sid from channel and address." );
            retVal = PLRESULT_OKDONOTHING;
            goto CLEANUPANDEXIT;
        }

        if ( !self->blackboardData->opID )
        {
            /* OPID could not be created, so drop the packet. */
            retVal = PLRESULT_OKDONOTHING;
            goto CLEANUPANDEXIT;
        }
    }
    else if ( ( self->flagsByte & EPPV2_FLAG_OPID ) == EPPV2_FLAG_OPID_IMPLICIT_RECV )
    {
        /* Implicit Receiver pktID: Receive pkt is a response, opCount was filled from the wire. */
        /* Our own SID is the implicitly sent SID. */
        self->blackboardData->opID = DOFOperationID_Create(opCount, DOFOperationID_GetLocalSID(), TRUE, TRUE);

        if ( !self->blackboardData->opID )
        {
            /* OPID could not be created, so drop the packet. */
            retVal = PLRESULT_OKDONOTHING;
            goto CLEANUPANDEXIT;
        }
    }

    retVal = PLRESULT_OKSENDUP;

    CLEANUPANDEXIT:
    if (pktID)
        DOFObjectIDSource_Destroy(pktID);

    return retVal;
}

typedef struct EPP_ExtendPermContext_t
{
    OALProtocolStack protocolStack;
    BlackboardData savedBlackboard;
} EPP_ExtendPermContext_t;

static void EPP_V2_ExtendPermissionsCallback(DOFError status, void *context)
{
    EPP_ExtendPermContext_t *fullContext;
    OALMessage msg;

    PCR_Assert(context);
    fullContext = (EPP_ExtendPermContext_t *)context;

    if (status == DOFERROR_OK)
    {
        /* Send a message to the protocol stack since the callback is not called in the message enqueue worker thread. */
        msg = SMT_RouteSavedEPP_Create(fullContext->protocolStack, fullContext->savedBlackboard);
        if (msg)
        {
            OALRemoteChannelMgr_SendMessageToStack(msg);
        }
        else
        {
            DOFMem_Free(fullContext->savedBlackboard);
        }
    }
    else
    {
        DOFMem_Free(fullContext->savedBlackboard);
    }
    DOFMem_Free(fullContext);
}

ProtocolLayerResult EPP_V2_Send(EPP_V2 self)
{
    OALStream   header_and_flags_stream;
    uint32      headerAndFlagsSize = 0;
    uint8       flagByte = 0x00;                /* reserved bits must be clear, so initialize flag byte here. */
    uint8       securityHeaderFlagsByte = 0x00;
    uint8       noHeaderSizeError = 1;
    const uint8 *     opID_objID_bytes = NULL;
    boolean     includeSecurityHeader;

    /* If security is not allowed for this PDU, check for contradictions and clear includeSecurityHeader. */
    if (self->blackboardData->flags.securityRequiredUnsecure && !self->blackboardData->flags.securityRequiredSecureHeaderUnsecure)
    {
        /* The encrypt or authenticate flags should not be set if the unsecure flag was set. */
        if (self->blackboardData->flags.securityRequiredAuthenticate || self->blackboardData->flags.securityRequiredEncrypt)
            return PLRESULT_ERROR;

        includeSecurityHeader = FALSE;
    }
    else
    {
        includeSecurityHeader = (self->blackboardData->associatedChannel->remoteChannelData->flags.sessionIsSecure &&
                                 !self->blackboardData->flags.forceUnsecure);
        /* Check to see if authentication or encryption is required and the securityRequiredSecureHeaderUnsecure bit is set.*/
        if (includeSecurityHeader)
        {
            if ((self->blackboardData->remoteDomainID != INVALID_REMOTE_DOMAIN_ID) && (self->blackboardData->remoteOpSecurityScope == NULL))
            {
                return PLRESULT_ERROR;
            }
            if (self->blackboardData->flags.securityRequiredSecureHeaderUnsecure &&
                (self->blackboardData->flags.securityRequiredAuthenticate || self->blackboardData->flags.securityRequiredEncrypt))
            {
                return PLRESULT_ERROR;
            }
        }
        if (self->blackboardData->flags.securityRequiredEncrypt)
        {
            if (!self->blackboardData->associatedChannel->remoteChannelData->cipher ||
                !OALCipher_IsEncrypt(self->blackboardData->associatedChannel->remoteChannelData->cipher))
            {
                return PLRESULT_ERROR;
            }
        }
        if (self->blackboardData->flags.securityRequiredAuthenticate)
        {
            if (!self->blackboardData->associatedChannel->remoteChannelData->cipher ||
                !OALCipher_IsAuthenticate(self->blackboardData->associatedChannel->remoteChannelData->cipher))
            {
                return PLRESULT_ERROR;
            }
        }
        if (self->blackboardData->flags.securityRequiredEncryptOrAuthenticate)
        {
            if (!self->blackboardData->associatedChannel->remoteChannelData->cipher ||
                (!OALCipher_IsAuthenticate(self->blackboardData->associatedChannel->remoteChannelData->cipher) &&
                 !OALCipher_IsEncrypt(self->blackboardData->associatedChannel->remoteChannelData->cipher)))
            {
                return PLRESULT_ERROR;
            }
        }
    }

    if(self->blackboardData->flags.isCommandNotResponse &&
       self->blackboardData->associatedChannel->remoteChannelData->flags.sessionIsSecure &&
       (self->blackboardData->opID != NULL) &&
       (self->blackboardData->associatedChannel->remoteChannelData->cipher != NULL))
    {
        sid_t * sid = DOFOperationID_GetSID(self->blackboardData->opID);

        if(sid != NULL && !self->blackboardData->flags.forceUnsecure)
        {
            sid_t * localSid;
            OALPermissionScope scope;

            /* Validate Tunnel permission. */
            if (!EPP_V2_ValidateTunnelPermission(self->blackboardData->remoteDomainID, self->blackboardData->associatedChannel->remoteChannelData->cipher, OALCipher_GetCurrentKeyState(self->blackboardData->associatedChannel->remoteChannelData->cipher)))
            {
                return PLRESULT_OKDONOTHING;
            }

            localSid = DOFOperationID_GetLocalSID();

            if(DOFObjectID_Compare(DOFObjectIDSource_GetObjectID(sid->oid),
                                   DOFObjectIDSource_GetObjectID(localSid->oid)) == 0)
            {
                DOFOperationID tempID;

                sid = self->blackboardData->associatedChannel->remoteChannelData->secureSid;
                if (sid)
                {
                    /* If there is a secure sid use it for the opID. */
                    tempID = DOFOperationID_Create(DOFOperationID_GetCount(self->blackboardData->opID), sid, FALSE, TRUE);
                    if(tempID != NULL)
                    {
                        DOFOperationID_Destroy(self->blackboardData->opID);
                        self->blackboardData->opID = tempID;
                    }
                    else
                    {
                        return PLRESULT_ERROR;
                    }
                }
                else
                    return PLRESULT_ERROR;
            }

            scope = DOFPermissionSet_Validate(OALCipher_GetLocalPerms( self->blackboardData->associatedChannel->remoteChannelData->cipher, OALCipher_GetCurrentKeyState( self->blackboardData->associatedChannel->remoteChannelData->cipher ) ),
                                              (self->blackboardData->seq == 0) ? sid->iamPermission
                                                                               : sid->actAsPermission);
            if(scope == NULL)
            {
                boolean extending = FALSE;
                if (self->blackboardData->seq != 0)
                {
                    boolean isExtendAllowed = (self->blackboardData->associatedChannel->remoteChannelData->flags.isConnection)
                            ? DOFConnectionConfig_IsPermissionsExtendAllowed(self->blackboardData->associatedChannel->remoteChannelData->config.connection)
                            : TRUE;
                    if ((self->blackboardData->associatedChannel->extendPermissions != NULL) &&
                        (self->blackboardData->associatedChannel->remoteChannelData->cipher != NULL) &&
                        isExtendAllowed)
                    {
                        EPP_ExtendPermContext_t *extendContext = (EPP_ExtendPermContext_t *)DOFMem_Alloc(sizeof(EPP_ExtendPermContext_t));
                        if (extendContext)
                        {
                            extendContext->savedBlackboard = BlackboardData_Save(self->blackboardData);
                            if (extendContext->savedBlackboard)
                            {
                                extendContext->protocolStack = self->blackboardData->associatedChannel->remoteChannelData->protocolStack;
                                if (extendContext->protocolStack)
                                {
                                    OALChannel_ExtendOutboundPermissions(self->blackboardData->associatedChannel, self->blackboardData->transportAddr, sid->actAsPermission, EPP_V2_ExtendPermissionsCallback, extendContext);

                                    extending = TRUE;

                                    /* Call Init for blackboard data so that the copy will hold the reference to all of the objects. */
                                    BlackboardData_Init(self->blackboardData);

                                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "Permission to use sid is unknown, trying to extend permissions.");
                                }
                                else
                                {
                                    DOFMem_Free(extendContext->savedBlackboard);
                                    DOFMem_Free(extendContext);
                                }
                            }
                            else
                            {
                                DOFMem_Free(extendContext);
                            }
                        }
                    }
                }

                if (!extending)
                {
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "Permission to use sid is denied, cannot send");
                }
                return PLRESULT_OKDONOTHING;    /* We don't want to close the connection for a permission failure */
            }
            else if(OALPermissionScope_GetIdCount(scope) == 0)        /* No permission to use this sid */
            {
                OALPermissionScope_Destroy(scope);
                DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_WARNING, "EPP V2: Permission to use sid is denied, therefore cannot send");
                return PLRESULT_OKDONOTHING;    /* We don't want to close the connection for a permission failure */
            }
            OALPermissionScope_Destroy(scope);
        }
    }

    headerAndFlagsSize = 3;     /* version, flags, and delay. */

    if (self->blackboardData->opID)
    {
        headerAndFlagsSize += 4;/* max op count size */
        if (DOFOperationID_IsImplicit(self->blackboardData->opID) && !self->blackboardData->flags.isCommandNotResponse )
        {
            /* Special internal only C OAL use of a near-OID. A true OID of type _BROADCAST must have length field of 0
               and no data after. We have a near-OID with _BROADCAST class but non-standard data thereafter. We will
               NOT place self->blackboardData->opID on the wire- hence the name implicit. */
            flagByte |= EPPV2_FLAG_OPID_IMPLICIT_RECV;
        }
        else
        {
            uint16 opid_size = DOFOperationID_GetSize(self->blackboardData->opID);
            if(!opid_size)
            {
                DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_ERROR, "EPP V2: Dropping packet, operation ID size is zero");
                return PLRESULT_ERROR;
            }

            headerAndFlagsSize += opid_size;
            flagByte |= EPPV2_FLAG_OPID_EXPLICIT; /* explicit RID */
        }
    }

    if ( self->blackboardData->flags.isCommandNotResponse )
    {
        if (self->blackboardData->flags.commandCompletesOnAnyFinal)
        {
            flagByte |= EPPV2_FLAG_COMMAND_COMPLETES;
        }

        if ( self->blackboardData->flags.isFlooded && !self->blackboardData->seq )
            self->blackboardData->seq++;
        if ( self->blackboardData->seq )
        {
            headerAndFlagsSize++;
            flagByte |= EPPV2_FLAG_COMMAND_SEQUENCE;
        }

        if ( self->blackboardData->retry )
        {
            headerAndFlagsSize++;
            flagByte |= EPPV2_FLAG_COMMAND_RETRY;
        }
    }
    else
    {
        flagByte |= EPPV2_FLAG_RESPONSE;
        if (self->blackboardData->flags.responseIsIntermediate)
        {
            flagByte |= EPPV2_FLAG_RESPONSE_INTERMEDIATE;
        }
    }


    if (includeSecurityHeader)
    {
        flagByte |= EPPV2_FLAG_SECURITY_HEADER;

        headerAndFlagsSize++;

        if (!self->blackboardData->flags.securityRequiredSecureHeaderUnsecure)
            securityHeaderFlagsByte |= EPPV2_SECURITYHEADER_FLAG_SECURE;

        if (self->isConnectionLess)
        {
            /* The default value for the security state ID is zero. */
            if (self->blackboardData->associatedChannel->remoteChannelData->securityStateID != 0 &&
                self->blackboardData->associatedChannel->remoteChannelData->securityStateID != INVALID_SECURITYSTATEID)
            {
                securityHeaderFlagsByte |= EPPV2_SECURITYHEADER_FLAG_STATE;
                headerAndFlagsSize += OALMarshal_GetCompressedValueSize(4, self->blackboardData->associatedChannel->remoteChannelData->securityStateID);
            }
            if (self->blackboardData->associatedChannel->remoteChannelData->flags.isSSIDAssignedFromAS)
                securityHeaderFlagsByte |= EPPV2_SECURITYHEADER_FLAG_AS;
        }

        if (self->blackboardData->remoteDomainID != INVALID_REMOTE_DOMAIN_ID)
        {
            securityHeaderFlagsByte |= EPPV2_SECURITYHEADER_FLAG_DOMAINID;
            headerAndFlagsSize += OALMarshal_GetCompressedValueSize(4, self->blackboardData->remoteDomainID);
            headerAndFlagsSize += DOFSecurityScope_GetMarshalSize(self->blackboardData->remoteOpSecurityScope);
        }

        if (self->blackboardData->securityScope)
        {
            securityHeaderFlagsByte |= EPPV2_SECURITYHEADER_FLAG_SECURITYSCOPE;
            headerAndFlagsSize += DOFSecurityScope_GetMarshalSize(self->blackboardData->securityScope);
        }
    }

    header_and_flags_stream = OALStream_CreateRW(headerAndFlagsSize);
    /* when header_and_flags_stream is created, the size of the stream is known so if a write fails it is probably
       because the size was not calculated correctly. */
    if(header_and_flags_stream != NULL)
    {
        noHeaderSizeError &= OALStream_Write8(header_and_flags_stream, EPPHDR_FLAGS_MASK | EPPV2_VERSION ); /* Header: version 2 + flags bit set. */
        noHeaderSizeError &= OALStream_Write8(header_and_flags_stream, flagByte);

        if ( ( flagByte & EPPV2_FLAG_OPID ) == EPPV2_FLAG_OPID_EXPLICIT )
        {
            opID_objID_bytes = DOFOperationID_GetOIDBytes(self->blackboardData->opID);
            /* write Explicit SID. */
            noHeaderSizeError &= OALStream_Write( header_and_flags_stream, DOFObjectID_GetSize_Bytes(opID_objID_bytes),
                                                  opID_objID_bytes );
        }

        if ( flagByte & EPPV2_FLAG_OPID )
        {
            DOFOperationCount opcount = DOFOperationID_GetCount(self->blackboardData->opID);

            if (OALMarshal_StreamCompress4( header_and_flags_stream, opcount ) != DOFERROR_OK)
                noHeaderSizeError = 0;
        }

        if ( flagByte & EPPV2_FLAG_COMMAND_SEQUENCE )
        {
            /* write SEQ field */
            noHeaderSizeError &= OALStream_Write8(header_and_flags_stream, (self->blackboardData->seq));
        }

        if ( flagByte & EPPV2_FLAG_COMMAND_RETRY )
        {
            /* write Retry field */
            noHeaderSizeError &= OALStream_Write8(header_and_flags_stream, (self->blackboardData->retry));
        }

        /* write Delay field */
        if (!(flagByte & EPPV2_FLAG_RESPONSE))
        {
            uint8 delayByte;
            uint32 seconds;

            if ( self->blackboardData->delay > 1000UL * EPPV2_DELAY_MAXSECONDS )
                seconds = EPPV2_DELAY_MAXSECONDS;
            else
                seconds = ( self->blackboardData->delay + 999 ) / 1000;  /* Round up */

            if (seconds <= 128)
                delayByte = seconds;
            else
                delayByte = ((seconds+31)/32) + 124;
                /*  adding 31 will cause the delay byte to round up, 124 comes from the grace period function where
                    seconds = 128 + (delayByte - 128) * 32 */

            noHeaderSizeError &= OALStream_Write8(header_and_flags_stream, delayByte);
        }

        /* Write the security header. */
        if (includeSecurityHeader)
        {
            noHeaderSizeError &= OALStream_Write8(header_and_flags_stream, securityHeaderFlagsByte);

            if (securityHeaderFlagsByte & EPPV2_SECURITYHEADER_FLAG_STATE)
                if ( OALMarshal_StreamCompress4( header_and_flags_stream, self->blackboardData->associatedChannel->remoteChannelData->securityStateID ) != DOFERROR_OK)
                    noHeaderSizeError = 0;

            if (securityHeaderFlagsByte & EPPV2_SECURITYHEADER_FLAG_DOMAINID)
            {
                /* The domainID flag controls both the remote domain ID and the remote operation security scope. */
                if (OALMarshal_StreamCompress4(header_and_flags_stream, self->blackboardData->remoteDomainID) != DOFERROR_OK)
                    noHeaderSizeError = 0;
                noHeaderSizeError &= DOFSecurityScope_Marshal(self->blackboardData->remoteOpSecurityScope, header_and_flags_stream);
            }

            if (securityHeaderFlagsByte & EPPV2_SECURITYHEADER_FLAG_SECURITYSCOPE)
            {
                noHeaderSizeError &= DOFSecurityScope_Marshal(self->blackboardData->securityScope, header_and_flags_stream);
            }
        }
    }
    else
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "EPP V2: Failed to create epp v2 send stream." );
        return PLRESULT_ERROR;
    }
    /* Verify that everything was added to header_and_flags_stream. */
    if (!noHeaderSizeError)
    {
        PCR_Assert(FALSE);
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "EPP V2: Failed to calculate correct stream size." );
        OALStream_Release(header_and_flags_stream, TRUE);
        return PLRESULT_ERROR;
    }

    if (securityHeaderFlagsByte & EPPV2_SECURITYHEADER_FLAG_SECURE)
    {
        uint32    sendPacketSize;
        OALStream stream;
        uint8*    pData;
        DOFError  status;
        OALCipher cipher = self->blackboardData->associatedChannel->remoteChannelData->cipher;

        if (!self->blackboardData->associatedChannel->remoteChannelData->cipher)
        {
            /* When secureGroup connection terminates locally, a nodeDown PDU goes out multicast secure to the group. */
            OALStream_Release(header_and_flags_stream, TRUE);               /* This can happen when nodeDown PDU is sent .. */
            return PLRESULT_ERROR;                                    /* .. but we do not have cipher setup yet. */
            /* There is a bigger issue. The spec will soon say that a going-down connection cannot lose its transport
               related capability until after the node down PDU is sent. Is that guaranteed in the C-OAL? */
        }
        sendPacketSize = OALPacket_GetTotalStreamSize(self->blackboardData->sendPacket);
        stream = OALStream_CreateRW( sendPacketSize );
        if(stream == NULL)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "EPP V2: Failed to create epp v2 encryption stream." );
            OALStream_Release(header_and_flags_stream, FALSE);      /* Nobody else knows about this */
            return PLRESULT_ERROR;
        }
        pData = OALStream_GetWritePtr( stream );
        OALPacket_CopyTo( self->blackboardData->sendPacket, pData, 0, sendPacketSize );
        OALStream_AdjustWriteOffset( stream, sendPacketSize );
        /* Note: self->blackboardData->flags.isMulticast has meaning for receive only. It is not filled on send. */
        /* Note: header_and_flags_stream is NOT encrypted by _Encipher() and remains intact.
                       Its beginning is known as the security boundary.
                 If encrypt is enabled, every byte of "stream" is encrypted (unless param payloadEncryptOffset not 0).
                       Its beginning is known as the encryption boundary.
                 Whether encrypting, or not, "stream" returns with a cipher header prepended and a MAC appended. */
        if ((status = OALCipher_Encipher( cipher,
                                &header_and_flags_stream,                       /* Not modified by _Encipher(). */
                                &stream,                                        /*     Modified by _Encipher(). */
                                0,   /* payloadEncryptOffset- this may be 0 forever but is available for flexibility. */
                                self->blackboardData->transportAddr,            /* Is NULL for sending multicast. */
                                self->blackboardData->flags.denyPacketReplayProtectionIfSecureUnicastSend,
                                self->blackboardData->secureMarkingID,
                                self->blackboardData->flags.ifProperPlaceNextBaseNIDOnWire )) != DOFERROR_OK)
        {
            PCR_Assert( status == DOFERROR_FAILED || status == DOFERROR_INSUFFICIENT_RESOURCES ||      /* public API. */
                        status == DOFERROR_ILLEGAL_ARGUMENT );        /* not in public API, not for returning to app. */
            PCRMutex_Lock(ref_count_mutex);
            OALStream_Release( stream, FALSE );
            OALStream_Release(header_and_flags_stream, FALSE);
            PCRMutex_Unlock(ref_count_mutex);
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "EPP V2: Failed to encipher packet." );
            return PLRESULT_ERROR;
        }
        OALPacket_Release( self->blackboardData->sendPacket );
        self->blackboardData->sendPacket = OALPacket_Create();
        if(!OALPacket_PrependStream( self->blackboardData->sendPacket, stream ))
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP V2: Failed to prepend enciphered stream to packet." );
            return PLRESULT_ERROR;
        }
    }

    if(!OALPacket_PrependStream( self->blackboardData->sendPacket, header_and_flags_stream ))
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "EPP V2: Failed to prepend header_and_flags_stream to packet." );
        return PLRESULT_ERROR;
    }

    return PLRESULT_OKSENDDOWN;
}


