/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/connection.h>

#include <internal/addressinternal.h>
#include <internal/blackboarddata.h>
#include <internal/ccmmode_v1.h>
#include <internal/esp_v2.h>
#include <internal/channel.h>
#include <internal/channel_stack_messages.h>
#include <internal/enp_v1.h>
#include <internal/epp_negative_one.h>
#include <internal/epp_v2.h>
#include <internal/exception.h>
#include <internal/marshal.h>
#include <internal/oap_v1.h>
#include <internal/packet.h>
#include <internal/protocollayerresult.h>
#include <internal/protocolstack.h>
#include <internal/remotechannel.h>
#include <internal/remotechanneldata.h>
#include <internal/remotechannelmanager.h>
#include <internal/router.h>
#include <internal/sgmp_v1.h>
#include <internal/stream.h>
#include <internal/tep_v1.h>
#include <internal/trp_v1.h>

/*
**  "Private" Type and Method Declarations
*/

#define ESP_V2_OPCODE_QUERY         0x00
#define ESP_V2_OPCODE_TERMINATE     0x05
#define ESP_V2_OPCODE_OPEN          0x06
#define ESP_V2_OPCODE_OPENSECURE    0x07

typedef ProtocolLayerResult (*LCP_StateHandlerFunc) (ESP_V2 self, uint8 lcpOpcode);

typedef ProtocolLayerResult (*ESP_V2_Receive_fn_t)(BlackboardData);
typedef ProtocolLayerResult (*ESP_V2_Send_fn_t)(BlackboardData, DOFOperation);

typedef enum ccp_states
{
    LCP_REQSENT,
    LCP_ACKRCVD,
    LCP_ACKSENT
} ccp_states;

/* LCP State Handler Method Prototypes */
static ProtocolLayerResult  LCP_ReqSent_StateHandler(ESP_V2 self, uint8 lcpOpcode);
static ProtocolLayerResult  LCP_AckRcvd_StateHandler(ESP_V2 self, uint8 lcpOpcode);
static ProtocolLayerResult  LCP_AckSent_StateHandler(ESP_V2 self, uint8 lcpOpcode);

static ProtocolLayerResult (*state_fns[])(ESP_V2 self, uint8 lcpOpcode) = { LCP_ReqSent_StateHandler, LCP_AckRcvd_StateHandler, LCP_AckSent_StateHandler };

typedef struct ESP_V2_t
{
    BlackboardData          blackboardData;

    void*                   protocolStack;
    ConfigRequestHandler_t  reqHandler;
    ConfigNegAckHandler_t   negAckHandler;
    ConfigRejectHandler_t   rejectHandler;

    uint8                   currentState;
} ESP_V2_t;

/* ***** ESP Packet Builders ***** */
static boolean buildESPConfigRequest(OALPacket packet, ConfigOptionList configOpts);
static boolean buildESPConfigAck(OALPacket packet);
static boolean buildESPConfigNegAck(OALPacket packet, ConfigOptionList configOpts);
static boolean buildESPConfigReject(OALPacket packet, ConfigOptionList configOpts);

/* ***** Parsing Helpers ***** */
static uint8 AttributeData_GetLength(const uint8 * buf);
static boolean AttributeData_StreamRead(OALStream stream, uint16 * pAttributeData);

/*
**  "Public" Type and Method Declarations
*/

ESP_V2 ESP_V2_Create(BlackboardData blackboardData, void* protocolStack, ConfigRequestHandler_t reqHandler, ConfigNegAckHandler_t negAckHandler, ConfigRejectHandler_t rejectHandler)
{
    ESP_V2 retVal = (ESP_V2)DOFMem_Alloc( sizeof(ESP_V2_t) );

    if(!retVal)
    {
        return NULL;
    }

    retVal->blackboardData = blackboardData;
    retVal->protocolStack = protocolStack;
    retVal->reqHandler = reqHandler;
    retVal->negAckHandler = negAckHandler;
    retVal->rejectHandler = rejectHandler;

    retVal->currentState = LCP_REQSENT;

    return retVal;
}

void ESP_V2_Destroy(ESP_V2 ccp)
{
    DOFMem_Free(ccp);
}

ProtocolLayerResult ESP_V2_SendInitialNegotiation(ESP_V2 self, ConfigOptionList desiredOpts)
{
    /* write request to packet. */
    if( buildESPConfigRequest( self->blackboardData->sendPacket, desiredOpts ) )
    {
        self->blackboardData->flags.isCommandNotResponse = TRUE;
        return PLRESULT_OKSENDDOWN;
    }
    else
    {
        return PLRESULT_ERROR;
    }
}

/* This fuctionality should probably be moved into ESP_V2_Receive */
ProtocolLayerResult ESP_V2_TCP_Receive(ESP_V2 self)
{
    uint16  attributeData;
    uint8   lcpOpCode;

    OALStream stream = self->blackboardData->receiveStream;

    /* check the length */
    if((OALStream_GetDataSize(stream) == 0) || OALStream_GetDataSize(stream) < (uint32)((uint32)AttributeData_GetLength(OALStream_GetReadPtr(stream)) + 1) )
        return PLRESULT_ERROR;

    /* get the attributeData/APPID */
    if(!AttributeData_StreamRead(stream, &attributeData))
        return PLRESULT_ERROR;

    /* check the attributeData/APPID */
    if(attributeData != ESP_PROTOCOLID)
        return PLRESULT_ERROR;

    /* get the LCP Opcode */
    if(!OALStream_Read8(stream, &lcpOpCode))
        return PLRESULT_ERROR;

    return state_fns[self->currentState](self, lcpOpCode);
}

/* LCP State Handler Methods */

static ProtocolLayerResult LCP_ReqSent_StateHandler(ESP_V2 self, uint8 lcpOpcode)
{
    OALStream rcvStream = self->blackboardData->receiveStream;
    OALPacket sndPacket = self->blackboardData->sendPacket;

    switch(lcpOpcode)
    {
    case LCP_CONFIGURE_REQUEST:
        {
            ESP_V2_CfgHandlerResponse_t cfgResponse;
            ConfigOptionList optList = NULL;

            if( !self->blackboardData->flags.isCommandNotResponse )
                return PLRESULT_CLOSE_CONNECTION;

            optList = ESPV2_ConfigOption_Unmarshall_List( rcvStream );
            if(!optList)
                return PLRESULT_CLOSE_CONNECTION;

            cfgResponse = self->reqHandler(self->protocolStack, optList);
            ESPV2_ConfigOptionList_Destroy( optList );

            switch(cfgResponse.responseCode)
            {
            case ESP_V2_CFGHANDLER_RESPONSE_ACK:
                {
                    OALMessage message = SMT_Connected_Negotiation_Server_Side_Create(self->protocolStack);
                    buildESPConfigAck( sndPacket );
                    self->blackboardData->flags.isCommandNotResponse = FALSE;
                    self->currentState = LCP_ACKSENT; /* Change to Ack-Sent state. */
                    if( message )
                    {
                        OALRemoteChannelMgr_SendMessageToStack(message);
                        return PLRESULT_OKSENDDOWN;
                    }
                    return PLRESULT_ERROR;
                }

            case ESP_V2_CFGHANDLER_RESPONSE_NEGACK:
                if ( buildESPConfigNegAck( sndPacket, cfgResponse.optList ) )
                {
                    ESPV2_ConfigOptionList_Destroy( cfgResponse.optList );
                    self->blackboardData->flags.isCommandNotResponse = FALSE;
                    return PLRESULT_OKSENDDOWN;
                }
                else
                    return PLRESULT_ERROR;

            case ESP_V2_CFGHANDLER_RESPONSE_REJECT:
                if ( buildESPConfigReject ( sndPacket, cfgResponse.optList ) )
                {
                    ESPV2_ConfigOptionList_Destroy( cfgResponse.optList );
                    self->blackboardData->flags.isCommandNotResponse = FALSE;
                    return PLRESULT_OKSENDDOWN;
                }
                else
                    return PLRESULT_ERROR;

            case ESP_V2_CFGHANDLER_RESPONSE_REQUEST:
                if ( buildESPConfigRequest ( sndPacket, cfgResponse.optList ) )
                {
                    ESPV2_ConfigOptionList_Destroy( cfgResponse.optList );
                    self->blackboardData->flags.isCommandNotResponse = TRUE;
                    return PLRESULT_OKSENDDOWN;
                }
                else
                    return PLRESULT_ERROR;

            case ESP_V2_CFGHANDLER_RESPONSE_CLOSE:
                {
                    return PLRESULT_CLOSE_CONNECTION;
                }

            default:
                {
                    return PLRESULT_CLOSE_CONNECTION;
                }
            }
        }

    case LCP_CONFIGURE_ACK:
        {
            if( self->blackboardData->flags.isCommandNotResponse )
                return PLRESULT_CLOSE_CONNECTION;
            self->currentState = LCP_ACKRCVD; /* Change to Ack-Rcvd state. */
            return PLRESULT_OKDONOTHING;
        }
    case LCP_CONFIGURE_NAK:
    case LCP_CONFIGURE_REJECT:
        {
            ESP_V2_CfgHandlerResponse_t cfgResponse;
            RejectList optList = NULL;

            if( self->blackboardData->flags.isCommandNotResponse )
                return PLRESULT_CLOSE_CONNECTION;

            if ( lcpOpcode == LCP_CONFIGURE_NAK )
                optList = ESPV2_ConfigOption_Unmarshall_List( rcvStream );
            else
                optList = ESPV2_Reject_Unmarshall_List( rcvStream );

            if(!optList)
                return PLRESULT_CLOSE_CONNECTION;

            cfgResponse = (lcpOpcode == LCP_CONFIGURE_NAK) ?
                          self->negAckHandler(self->protocolStack, optList) :
                          self->rejectHandler(self->protocolStack, optList);
            ESPV2_RejectList_Destroy( optList );
            switch(cfgResponse.responseCode)
            {
            case ESP_V2_CFGHANDLER_RESPONSE_REQUEST:
                if ( buildESPConfigRequest( sndPacket, cfgResponse.optList ) )
                {
                    ESPV2_ConfigOptionList_Destroy( cfgResponse.optList );
                    self->blackboardData->flags.isCommandNotResponse = TRUE;
                    return PLRESULT_OKSENDDOWN;
                }
                else
                    return PLRESULT_CLOSE_CONNECTION;

            case ESP_V2_CFGHANDLER_RESPONSE_CLOSE:
                return PLRESULT_CLOSE_CONNECTION;
            default:
                {
                    return PLRESULT_ERROR;
                }
            }
        }

    default:
        return PLRESULT_CLOSE_CONNECTION;
    }
}

static ProtocolLayerResult  LCP_AckRcvd_StateHandler(ESP_V2 self, uint8 lcpOpcode)
{
    OALStream rcvStream = self->blackboardData->receiveStream;
    OALPacket sndPacket = self->blackboardData->sendPacket;

    switch (lcpOpcode)
    {
    case LCP_CONFIGURE_REQUEST:
        {
            ESP_V2_CfgHandlerResponse_t cfgResponse;
            ConfigOptionList optList = NULL;

            if( !self->blackboardData->flags.isCommandNotResponse )
                return PLRESULT_CLOSE_CONNECTION;

            optList = ESPV2_ConfigOption_Unmarshall_List( rcvStream );
            if (!optList)
                return PLRESULT_CLOSE_CONNECTION;

            cfgResponse = self->reqHandler(self->protocolStack, optList);
            ESPV2_ConfigOptionList_Destroy( optList );

            switch (cfgResponse.responseCode)
            {
            case ESP_V2_CFGHANDLER_RESPONSE_ACK:
                {
                    buildESPConfigAck( sndPacket );
                    self->blackboardData->flags.isCommandNotResponse = FALSE;
                    return PLRESULT_SENDDOWN_THENNEGOTIATIONCOMPLETE;
                }
            case ESP_V2_CFGHANDLER_RESPONSE_NEGACK:
                if ( buildESPConfigNegAck( sndPacket, cfgResponse.optList ) )
                {
                    ESPV2_ConfigOptionList_Destroy( cfgResponse.optList );
                    self->blackboardData->flags.isCommandNotResponse = FALSE;
                    return PLRESULT_OKSENDDOWN;
                }
                else
                    return PLRESULT_ERROR;

            case ESP_V2_CFGHANDLER_RESPONSE_REJECT:
                if ( buildESPConfigReject ( sndPacket, cfgResponse.optList ) )
                {
                    ESPV2_ConfigOptionList_Destroy( cfgResponse.optList );
                    self->blackboardData->flags.isCommandNotResponse = FALSE;
                    return PLRESULT_OKSENDDOWN;
                }
                else
                    return PLRESULT_CLOSE_CONNECTION;

            case ESP_V2_CFGHANDLER_RESPONSE_CLOSE:
                return PLRESULT_CLOSE_CONNECTION;

            default:
                {
                    return PLRESULT_ERROR;
                }
            }
        }
        break;

    default:
        return PLRESULT_CLOSE_CONNECTION;
    }
}

static ProtocolLayerResult  LCP_AckSent_StateHandler(ESP_V2 self, uint8 lcpOpcode)
{
    OALStream rcvStream = self->blackboardData->receiveStream;
    OALPacket sndPacket = self->blackboardData->sendPacket;

    switch (lcpOpcode)
    {
    case LCP_CONFIGURE_REQUEST:
        {
            ESP_V2_CfgHandlerResponse_t cfgResponse;
            ConfigOptionList optList = NULL;

            if( !self->blackboardData->flags.isCommandNotResponse )
                return PLRESULT_CLOSE_CONNECTION;

            optList = ESPV2_ConfigOption_Unmarshall_List( rcvStream );

            if(!optList)
                return PLRESULT_CLOSE_CONNECTION;

            cfgResponse = self->reqHandler(self->protocolStack, optList);
            ESPV2_ConfigOptionList_Destroy( optList );

            switch(cfgResponse.responseCode)
            {
            case ESP_V2_CFGHANDLER_RESPONSE_ACK:
                {
                    OALMessage message = SMT_Connected_Negotiation_Server_Side_Create(self->protocolStack);
                    buildESPConfigAck( sndPacket );
                    self->blackboardData->flags.isCommandNotResponse = FALSE;
                    self->currentState = LCP_ACKSENT; /* Change to Ack-Sent state. */
                    if( message )
                    {
                        OALRemoteChannelMgr_SendMessageToStack(message);
                        return PLRESULT_OKSENDDOWN;
                    }
                    return PLRESULT_ERROR;
                }

            case ESP_V2_CFGHANDLER_RESPONSE_NEGACK:
                if ( buildESPConfigNegAck( sndPacket, cfgResponse.optList ) )
                {
                    ESPV2_ConfigOptionList_Destroy( cfgResponse.optList );
                    self->blackboardData->flags.isCommandNotResponse = FALSE;
                    return PLRESULT_OKSENDDOWN;
                }
                else
                    return PLRESULT_ERROR;

            case ESP_V2_CFGHANDLER_RESPONSE_REJECT:
                if ( buildESPConfigReject ( sndPacket, cfgResponse.optList ) )
                {
                    ESPV2_ConfigOptionList_Destroy( cfgResponse.optList );
                    self->blackboardData->flags.isCommandNotResponse = FALSE;
                    return PLRESULT_OKSENDDOWN;
                }
                else
                    return PLRESULT_ERROR;

            case ESP_V2_CFGHANDLER_RESPONSE_REQUEST:
                if ( buildESPConfigRequest ( sndPacket, cfgResponse.optList ) )
                {
                    ESPV2_ConfigOptionList_Destroy( cfgResponse.optList );
                    self->blackboardData->flags.isCommandNotResponse = TRUE;
                    return PLRESULT_OKSENDDOWN;
                }
                else
                    return PLRESULT_CLOSE_CONNECTION;

            case ESP_V2_CFGHANDLER_RESPONSE_CLOSE:
                return PLRESULT_CLOSE_CONNECTION;

            default:
                {
                    return PLRESULT_ERROR;
                }
            }
        }

    case LCP_CONFIGURE_ACK:
        return PLRESULT_NEGOTIATIONCOMPLETE;

    case LCP_CONFIGURE_NAK:
    case LCP_CONFIGURE_REJECT:
        {
            ESP_V2_CfgHandlerResponse_t cfgResponse;
            ConfigOptionList optList = NULL;

            if( self->blackboardData->flags.isCommandNotResponse )
                return PLRESULT_CLOSE_CONNECTION;

            if ( lcpOpcode == LCP_CONFIGURE_NAK )
                optList = ESPV2_ConfigOption_Unmarshall_List( rcvStream );
            else
                optList = ESPV2_Reject_Unmarshall_List( rcvStream );
            if (!optList)
                return PLRESULT_CLOSE_CONNECTION;

            cfgResponse = (lcpOpcode == LCP_CONFIGURE_NAK) ?
                          self->negAckHandler(self->protocolStack, optList) :
                          self->rejectHandler(self->protocolStack, optList);
            ESPV2_RejectList_Destroy( optList );
            switch (cfgResponse.responseCode)
            {
            case ESP_V2_CFGHANDLER_RESPONSE_REQUEST:
                if ( buildESPConfigRequest( sndPacket, cfgResponse.optList ) )
                {
                    ESPV2_ConfigOptionList_Destroy( cfgResponse.optList );
                    self->blackboardData->flags.isCommandNotResponse = TRUE;
                    return PLRESULT_OKSENDDOWN;
                }
                else
                    return PLRESULT_CLOSE_CONNECTION;

            case ESP_V2_CFGHANDLER_RESPONSE_CLOSE:
                return PLRESULT_CLOSE_CONNECTION;

            default:
                {
                    return PLRESULT_ERROR;
                }
            }
        }

    default:
        return PLRESULT_CLOSE_CONNECTION;
    }
}

/* ESP Packet Builders */

#define ESP_CONFIGREQUEST_HEADER_SIZE   2
static const uint8 ccpConfigRequestHeader[] = {0x00, LCP_CONFIGURE_REQUEST};

#define ESP_CONFIGACK_HEADER_SIZE   2
static const uint8 ccpConfigAckHeader[] = {0x00, LCP_CONFIGURE_ACK};

#define ESP_CONFIGNEGACK_HEADER_SIZE   2
static const uint8 ccpConfigNegAckHeader[] = {0x00, LCP_CONFIGURE_NAK};

#define ESP_CONFIGREJECT_HEADER_SIZE   2
static const uint8 ccpConfigRejectHeader[] = {0x00, LCP_CONFIGURE_REJECT};

/** @internal
 * Build ESP Request.
 *
 * @param packet        The packet to write the config request to.
 * @param configOpts    The configuration options to request.
 *
 * @return boolean
 * @retval Non-zero (TRUE)  if successful
 * @retval Zero     (FALSE) if failed
 */
static boolean buildESPConfigRequest(OALPacket packet, ConfigOptionList configOpts)
{
    OALStream headerStream = OALStream_CreateRO( ccpConfigRequestHeader, ESP_CONFIGREQUEST_HEADER_SIZE );
    if( headerStream )
    {
        if( OALPacket_AppendStream( packet, headerStream ) )
        {
            OALStream configOptsStream = ESPV2_ConfigOption_Marshall_List( configOpts );
            if( configOptsStream )
            {
                if( OALPacket_AppendStream( packet, configOptsStream ) )
                    return TRUE;

                OALStream_Release( configOptsStream, FALSE );
            }
        }
        else
        {
            OALStream_Release( headerStream, FALSE );
        }
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "ESP V2: build config request: Failed to append ccp config request to stream.");
    return FALSE;
}

static boolean buildESPConfigAck(OALPacket packet)
{
    OALStream headerStream = OALStream_CreateRO(ccpConfigAckHeader, ESP_CONFIGACK_HEADER_SIZE);
    if(headerStream)
    {
        if(OALPacket_AppendStream(packet, headerStream))
            return TRUE;

        OALStream_Release(headerStream, FALSE);
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "ESP V2: build config ack: Failed to append ccp config ack to stream.");
    return FALSE;
}

/** @internal
 * Build ESP Negative Acknowledge.
 *
 * @param packet        The packet to write the NAK to.
 * @param configOpts    The configuration options to NAK.
 *
 * @return boolean
 * @retval Non-zero (TRUE)  if successful
 * @retval Zero     (FALSE) if failed
 */
static boolean buildESPConfigNegAck(OALPacket packet, ConfigOptionList configOpts)
{
    OALStream headerStream = OALStream_CreateRO(ccpConfigNegAckHeader, ESP_CONFIGNEGACK_HEADER_SIZE);
    if( headerStream )
    {
        if( OALPacket_AppendStream(packet, headerStream) )
        {
            OALStream configOptsStream = ESPV2_ConfigOption_Marshall_List(configOpts);
            if( configOptsStream )
            {
                if( OALPacket_AppendStream(packet, configOptsStream) )
                    return TRUE;

                OALStream_Release( configOptsStream, FALSE );
            }
        }
        else
        {
            OALStream_Release(headerStream, FALSE);
        }
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "ESP V2: build config neg ack: Failed to append ccp config neg ack to stream.");
    return FALSE;
}

/** @internal
 * Build ESP Reject.
 *
 * @param packet        The packet to write the reject to.
 * @param configOpts    The configuration options to reject.
 *
 * @return boolean
 * @retval Non-zero (TRUE)  if successful
 * @retval Zero     (FALSE) if failed
 */
static boolean buildESPConfigReject(OALPacket packet, ConfigOptionList configOpts)
{
    OALStream headerStream = OALStream_CreateRO(ccpConfigRejectHeader, ESP_CONFIGREJECT_HEADER_SIZE);
    if( headerStream )
    {
        if( OALPacket_AppendStream(packet, headerStream) )
        {
            OALStream configOptsStream = ESPV2_Reject_Marshall_List(configOpts);
            if( configOptsStream )
            {
                if( OALPacket_AppendStream( packet, configOptsStream ) )
                    return TRUE;

                OALStream_Release( configOptsStream, FALSE );
            }
        }
        else
        {
            OALStream_Release( headerStream, FALSE );
        }
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "ESP V2: build config reject: Failed to append ccp config reject to stream.");
    return FALSE;
}


/*
    Gets the length of a attributeData based on the byte first byte of
    it, which is passed in as a pointer to the buffer containing the
    ESP packet.

    @param  buf  The buffer containing the protocol as the first and
    possibly second byte.

    @returns The protocol ID length.
    @retval  1 if the ID is one byte.
    @retval  2 if the ID is extended to another byte.
*/
static uint8 AttributeData_GetLength(const uint8 * buf)
{
    uint8 cont = (*buf) & ESP_PROTOCOLID_CONT;
    return (cont != 0) ? sizeof(uint16) : sizeof(uint8);
}


/*
    Reads a attributeData from a stream.  The value in the stream is either 1 or
    2 bytes so it is converted to a 2-byte value and returned in @ref
    pAttributeData.

    @param  stream  The stream containing the attributeData.
    @param  pAttributeData  The converted attributeData is placed here.

    @returns The success code of the read.
    @retval  TRUE if the stream read was successful.
    @retval  FALSE if it was not.
*/
static boolean AttributeData_StreamRead(OALStream stream, uint16 * pAttributeData)
{
    boolean rval;
    uint8 cont = (*OALStream_GetReadPtr(stream)) & ESP_PROTOCOLID_CONT;
    if (cont != 0)
    {
        uint16 attributeData16;
        rval = OALStream_Read16(stream, &attributeData16);
        *pAttributeData = attributeData16;
    }
    else
    {
        uint8 attributeData8;
        rval = OALStream_Read8(stream, &attributeData8);
        *pAttributeData = attributeData8;
    }
    return rval;
}

static ProtocolLayerResult send_ESP_V2_Command(BlackboardData blackboardData, DOFOperation op)
{
    /* This is a common function for commands that only include an opcode. */
    OALStream   CommandStream = OALStream_CreateRW(3);
    boolean     opcodeWriteStatus;

    if (!CommandStream)
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ESP V2: send ESP command: failed to create command stream.");
        return PLRESULT_ERROR;
    }

    if (OALMarshal_StreamCompress2(CommandStream, PROTOCOLID_ESP_V2) != DOFERROR_OK)   /* Write the ESP_V2 appID. */
    {
        OALStream_Release( CommandStream, FALSE );
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ESP V2: send ESP command: failed to write to command stream.");
        return PLRESULT_ERROR;
    }
    switch (op->operation_type)
    {
    case DOFOPERATIONTYPE_ESP_V2_QUERY:
        opcodeWriteStatus = OALStream_Write8(CommandStream, ESP_V2_OPCODE_QUERY);
        break;
    case DOFOPERATIONTYPE_ESP_V2_TERMINATE:
        opcodeWriteStatus = OALStream_Write8(CommandStream, ESP_V2_OPCODE_TERMINATE);
        break;
    case DOFOPERATIONTYPE_ESP_V2_OPEN:
        opcodeWriteStatus = OALStream_Write8(CommandStream, ESP_V2_OPCODE_OPEN);
        break;
    default:
        opcodeWriteStatus = FALSE;
        break;
    }
    if (!opcodeWriteStatus)
    {
        OALStream_Release( CommandStream, FALSE );
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ESP V2: send ESP command: failed to write to command stream.");
        return PLRESULT_ERROR;
    }

    blackboardData->flags.isCommandNotResponse = TRUE;
    if (op->opID)
    {
        if (blackboardData->opID)
        {
            DOFOperationID_Destroy(blackboardData->opID);
        }
        blackboardData->opID = DOFOperationID_Copy(op->opID);
        if (!blackboardData->opID)
        {
            return PLRESULT_ERROR;
        }
    }
    blackboardData->delay = op->sources->oal_timeout;
    blackboardData->retry = op->sources->retry_count;

    if (!blackboardData->sendPacket)
    {
        blackboardData->sendPacket = OALPacket_Create();
        if (!blackboardData->sendPacket)
        {
            OALStream_Release( CommandStream, FALSE );
            return PLRESULT_ERROR;
        }
    }

    if (!OALPacket_AppendStream( blackboardData->sendPacket, CommandStream))
    {
        OALStream_Release( CommandStream, FALSE );
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ESP V2: send APP V0 query: failed to append stream to send packet.");
        return PLRESULT_ERROR;
    }

    return PLRESULT_OKSENDDOWN;

}

static uint32 ESP_V2_StreamUnmarshalAPPIDList(OALStream stream)
{
    uint16 appid;
    uint32 bitFieldList = 0;

    while (OALMarshal_StreamUncompress2(stream, &appid) == DOFERROR_OK)
    {
        switch (appid)
        {
        case OAP_V1_APPID:
            bitFieldList |= V0_QUERY_BIT_OAP_V1;
            break;
        case TEP_V1_APPID:
            bitFieldList |= V0_QUERY_BIT_TEP_V1;
            break;
        case TRP_V1_APPID:
            bitFieldList |= V0_QUERY_BIT_TRP_V1;
            break;
        case SGMP_V1_APPID:
            bitFieldList |= V0_QUERY_BIT_SGMP_V1;
            break;
        case CCMMODE_V1_APPID:
            bitFieldList |= V0_QUERY_BIT_CCMMODE_V1;
            break;
        /* As new protocols are added they must be included in this switch statement. */
        default:
            break;
        }
    }
    return bitFieldList;
}

static boolean ESP_V2_StreamCreateAPPIDList(OALStream stream, uint32 APPIDsToAdvertise)
{
    /* Write each APPID.  This list must be in increasing order. New APPIDs must be added to this list. */
    if (APPIDsToAdvertise & V0_QUERY_BIT_OAP_V1)
    {
        if (OALMarshal_StreamCompress2(stream, OAP_V1_APPID) != DOFERROR_OK)
            return FALSE;
    }
    if (APPIDsToAdvertise & V0_QUERY_BIT_TEP_V1)
    {
        if (OALMarshal_StreamCompress2(stream, TEP_V1_APPID) != DOFERROR_OK)
            return FALSE;
    }
    if (APPIDsToAdvertise & V0_QUERY_BIT_TRP_V1)
    {
        if (OALMarshal_StreamCompress2(stream, TRP_V1_APPID) != DOFERROR_OK)
            return FALSE;
    }
    if (APPIDsToAdvertise & V0_QUERY_BIT_SGMP_V1)
    {
        if (OALMarshal_StreamCompress2(stream, SGMP_V1_APPID) != DOFERROR_OK)
            return FALSE;
    }
    if (APPIDsToAdvertise & V0_QUERY_BIT_CCMMODE_V1)
    {
        if (OALMarshal_StreamCompress2(stream, CCMMODE_V1_APPID) != DOFERROR_OK)
            return FALSE;
    }
    return TRUE;
}

/* Note: this function must be updated as new protocols are added. */
static boolean verifyRequiredProtocols(OALChannel channel, uint32 APPIDList)
{
    DOFSecurityDesire       desire;
    DOFProtocolNegotiator   negotiator;
    boolean                 requiredProtocolsAvailable = TRUE;

    PCR_Assert(channel);
    PCR_Assert(channel->remoteChannelData);
    PCR_Assert(channel->remoteChannelData->flags.isConnection);

    desire = DOFConnectionConfig_GetSecurityDesire(channel->remoteChannelData->config.connection);
    negotiator = DOFConnectionConfig_GetProtocolNegotiator(channel->remoteChannelData->config.connection);

    if (negotiator == DOF_DEFAULT_AS_PROTOCOL_NEGOTIATOR)
    {
        /* if this is a connection to an AS make sure that TRP is supported. */
        if (!(APPIDList & V0_QUERY_BIT_TRP_V1))
        {
            requiredProtocolsAvailable = FALSE;
        }
    }
    else if (negotiator == DOF_DEFAULT_PROTOCOL_NEGOTIATOR)
    {
        if (desire == DOFSECURITYDESIRE_NOT_SECURE)
        {
            /*  If this is the default negotiator and it is unsecure, make sure that OAP is supported. */
            if (!(APPIDList & V0_QUERY_BIT_OAP_V1))
            {
                requiredProtocolsAvailable = FALSE;
            }
        }
        else if (desire & DOFSECURITYDESIRE_SECURE_ANY)
        {
            /* If this is the default negotiator and it is secure, make sure that TEP is supported. */
            if (!(APPIDList & V0_QUERY_BIT_TEP_V1))
            {
                requiredProtocolsAvailable = FALSE;
            }
        }
        else if (desire == DOFSECURITYDESIRE_ANY)
        {
            if (channel->remoteChannelData->flags.sessionIsSecure)
            {
                if (!(APPIDList & V0_QUERY_BIT_OAP_V1))
                {
                    requiredProtocolsAvailable = FALSE;
                }
            }
            else if (channel->remoteChannelData->flags.secureDesire)
            {
                /* if the "secureDesire" flag is set and the channel is not yet secure, make sure that TEP is supported. */
                if (!(APPIDList & V0_QUERY_BIT_TEP_V1))
                {
                    requiredProtocolsAvailable = FALSE;
                }
            }
            else if (APPIDList & V0_QUERY_BIT_TEP_V1 && !(APPIDList & V0_QUERY_BIT_OAP_V1))
            {
                /* If TEP is supported, but OAP is not then it must be secure.  set secure desire and leave return value as TRUE. */
                channel->remoteChannelData->flags.secureDesire = TRUE;
            }
            else if (!(APPIDList & V0_QUERY_BIT_TEP_V1) && !(APPIDList & V0_QUERY_BIT_OAP_V1))
            {
                /* If the session is not secure verify that OAP or TEP is supported. */
                requiredProtocolsAvailable = FALSE;
            }
        }
        else
        {
            requiredProtocolsAvailable = FALSE;
        }
    }
    else
    {
        requiredProtocolsAvailable = FALSE;
    }

    /*  Todo: if TRP was not advertised by the remote host and the channel has
        domain credentials or a trusted domain, return FALSE. */
    return requiredProtocolsAvailable;
}

static uint32 buildAdvertisedAPPIDList(OALChannel channel, boolean inSession)
{
    DOFSecurityDesire   desire;
    uint32              APPIDsToAdvertise = 0;

    PCR_Assert(channel);
    PCR_Assert(channel->remoteChannelData);

    if (channel->remoteChannelData->flags.isConnection)
        desire = DOFConnectionConfig_GetSecurityDesire(channel->remoteChannelData->config.connection);
    else
        desire = DOFServerConfig_GetSecurityDesire(channel->remoteChannelData->config.server);

    if (inSession)
    {
        if (desire == DOFSECURITYDESIRE_ANY)
        {
            APPIDsToAdvertise |= V0_QUERY_BIT_OAP_V1;
            APPIDsToAdvertise |= V0_QUERY_BIT_TEP_V1;
            APPIDsToAdvertise |= V0_QUERY_BIT_TRP_V1;
            APPIDsToAdvertise |= V0_QUERY_BIT_CCMMODE_V1;
            if (channel->remoteChannelData->flags.secureDesire && !channel->remoteChannelData->flags.sessionIsSecure)
            {
                /* If the session will be secure, but it is not yet don't advertise OAP. */
                APPIDsToAdvertise &= ~V0_QUERY_BIT_OAP_V1;
            }
        }
        else if (desire & DOFSECURITYDESIRE_SECURE_ANY)
        {
            if (channel->remoteChannelData->flags.sessionIsSecure)
            {
                APPIDsToAdvertise |= V0_QUERY_BIT_OAP_V1;
                APPIDsToAdvertise |= V0_QUERY_BIT_TEP_V1;
                APPIDsToAdvertise |= V0_QUERY_BIT_TRP_V1;
                APPIDsToAdvertise |= V0_QUERY_BIT_CCMMODE_V1;
            }
            else
            {
                APPIDsToAdvertise |= V0_QUERY_BIT_TEP_V1;
                APPIDsToAdvertise |= V0_QUERY_BIT_TRP_V1;
                APPIDsToAdvertise |= V0_QUERY_BIT_CCMMODE_V1;
            }
        }
        else if (desire == DOFSECURITYDESIRE_NOT_SECURE)
        {
            APPIDsToAdvertise = V0_QUERY_UNSECURE_APPIDS;
        }
    }
    else
    {

        if (desire == DOFSECURITYDESIRE_NOT_SECURE)
        {
            APPIDsToAdvertise = V0_QUERY_UNSECURE_APPIDS;
        }
        else
        {
            APPIDsToAdvertise = ESP_V0_QUERY_BITMASK;
        }
    }

    /* If there isn't a route to an AS don't advertise TRP. */
    if (!channel->remoteChannelData->flags.TRP_AcceptsCommands)
    {
        APPIDsToAdvertise &= ~V0_QUERY_BIT_TRP_V1;
    }

    if (channel->remoteChannelData->negotiator == DOF_DEFAULT_AS_PROTOCOL_NEGOTIATOR)
    {
        /* Don't advertise SGMP or OAP if the AS negotiator is used. */
        APPIDsToAdvertise &= ~V0_QUERY_BIT_SGMP_V1;
        APPIDsToAdvertise &= ~V0_QUERY_BIT_OAP_V1;
    }

    return APPIDsToAdvertise;
}

static void ESP_OpenOp_Server_Observer_fn(OALObserver self, OALObservable observable, void* ctx)
{
    DOFOperation_Internal_Destroy((DOFOperation)ctx);
}

static ProtocolLayerResult ESP_V2_ReceiveQuery(BlackboardData blackboardData)
{
    ProtocolLayerResult plRes;
    uint32 APPIDsToAdvertise = ESP_V0_QUERY_BITMASK;
    /* create a stream for the APPID list, two bytes for each new APPID plus one for the ESP APPID used for the header. */
    OALStream queryResponse = OALStream_CreateRW(ESP_V0_APPID_COUNT * 2 + 1);

    if (!queryResponse)
        return PLRESULT_ERROR;

    /* Write ESP appID, and the query response opcode. */
    if (OALMarshal_StreamCompress2(queryResponse, PROTOCOLID_ESP_V2) != DOFERROR_OK)   /* Write the ESP_V2 appID. */
    {
        OALStream_Release(queryResponse, FALSE);
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ESP V2: send ESP query response: failed to write to command stream.");
        return PLRESULT_ERROR;
    }
    if (!OALStream_Write8(queryResponse, ESP_V2_OPCODE_QUERY))
    {
        OALStream_Release(queryResponse, FALSE);
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ESP V2: send ESP query response: failed to write to command stream.");
        return PLRESULT_ERROR;
    }

    blackboardData->flags.isCommandNotResponse = FALSE;

    if (blackboardData->associatedChannel->remoteChannelData->flags.isConnection &&
        ((DOFConnectionConfig_GetConnectionType(blackboardData->associatedChannel->remoteChannelData->config.connection) == DOFCONNECTIONTYPE_DATAGRAM) ||
         blackboardData->associatedChannel->remoteChannelData->flags.secureDesire))
    {
        APPIDsToAdvertise = buildAdvertisedAPPIDList(blackboardData->associatedChannel, TRUE);
    }
    else
    {
        /* If the query is not received on a session advertise all protocols. */
        APPIDsToAdvertise = buildAdvertisedAPPIDList(blackboardData->associatedChannel, FALSE);
    }

    if (!ESP_V2_StreamCreateAPPIDList(queryResponse, APPIDsToAdvertise))
    {
        OALStream_Release(queryResponse, FALSE);
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ESP V2: send ESP query response: failed to write to APPID list to command stream.");
        return PLRESULT_ERROR;
    }
    /* Write the list to the send packet. */
    if (!blackboardData->sendPacket)
        blackboardData->sendPacket = OALPacket_Create();

    if (!blackboardData->sendPacket)
    {
        OALStream_Release(queryResponse, FALSE);
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "ESP V2: handle ESP V0 query: Failed to create packet for APP V0 query." );
        return PLRESULT_ERROR;
    }

    if (!OALPacket_AppendStream( blackboardData->sendPacket, queryResponse ))
        return PLRESULT_ERROR;

    plRes = ((OALProtocolStack)blackboardData->associatedChannel->remoteChannelData->protocolStack)->ENP_EPP_send_func(blackboardData->associatedChannel->remoteChannelData->protocolStack, blackboardData);
    if (plRes  == PLRESULT_OKSENDDOWN )
    {
        if (blackboardData->flags.isMulticast)
        {
            if (!connectionlessV0Query_AddDelayedResponse(blackboardData))
                return PLRESULT_ERROR;
        }
        else
        {
            /* If this is a unicast query respond immediately. */
            if( DOFTransportLayer_SendTo(blackboardData->associatedChannel->remoteChannelData->dtl, blackboardData->sendPacket, blackboardData->transportAddr) )
            {
                /* We just sent a packet, so if there is a maxSendSilenceTimerItem reset it. */
                if(blackboardData->associatedChannel->remoteChannelData->maxSendSilenceTimerItem)
                {
                    OALRemoteChannelMgr_SetSendSilenceTimerItem(blackboardData->associatedChannel);
                }
            }
            else
            {
                return PLRESULT_ERROR;
            }
        }
    }
    return PLRESULT_OKDONOTHING;
}

static ProtocolLayerResult ESP_V2_ReceiveQueryResponse(BlackboardData blackboardData)
{
    DOFOperation queryOp;
    /*  Set the flag for the appids received that we use.  We may not be required to send a response if
        all known appids are used before we are required to send our response. */

    queryOp = OALOperationManager_CheckExistingOperation(DOFSecurityScope_GetDomainID(DOF_GetDefaultSecurityScope(DOF_GetDof())), blackboardData->opID);

    if (queryOp)
    {
        uint32 oldList = blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids;

        blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids = ESP_V2_StreamUnmarshalAPPIDList(blackboardData->receiveStream);

        if (blackboardData->associatedChannel->remoteChannelData->flags.isOutbound)
        {
            if ( blackboardData->associatedChannel->remoteChannelData->negotiator == DOF_DEFAULT_AS_PROTOCOL_NEGOTIATOR )
            {
                /* Regardless of what was received, clear the OAP bit if the DOF_DEFAULT_AS_PROTOCOL_NEGOTIATOR was used. */
                blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids &= ~V0_QUERY_BIT_OAP_V1;
            }

            if (blackboardData->associatedChannel->remoteChannelData->flags.waitingForQueryResponse)
            {
                blackboardData->associatedChannel->remoteChannelData->flags.waitingForQueryResponse = FALSE;

                if (verifyRequiredProtocols(blackboardData->associatedChannel, blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids))
                {
                    if(!(blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids & V0_QUERY_BIT_TRP_V1))
                    {
                        /* Clear TRP send commands since TRP was not included in the query response. */
                        blackboardData->associatedChannel->remoteChannelData->flags.TRP_SendCommands = FALSE;
                    }
                    /* If the required protocols are supported then notify the channel. */
                    OALRemoteChannel_NotifyInitComplete(blackboardData->associatedChannel);
                }
                else
                {
                    /* If the required protocols are not supported, send a message to shut down the channel. */
                    DOFException except = DOFException_Copy(DOFEXCEPTION_PROTOCOL_NEGOTIATION_FAILED);
                    OALRemoteChannel_Shutdown(blackboardData->associatedChannel, except);
                    DOFException_Destroy(except);
                }
            }
        }
        else
        {
            /* This is written to assume TRP V1.  It will require changes when multiple TRP versions are supported. */
            if (blackboardData->associatedChannel->remoteChannelData->flags.TRP_SendCommands &&
                !(blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids & V0_QUERY_BIT_TRP_V1))
            {
                /* Clear TRP send commands, and discard all messages from the blocked message queue. */
                blackboardData->associatedChannel->remoteChannelData->flags.TRP_SendCommands = FALSE;
                OALRemoteChannel_DestroyBlockedMessageQueue(blackboardData->associatedChannel);
            }
            if ((blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids & V0_QUERY_BIT_TRP_V1) &&
                blackboardData->associatedChannel->remoteChannelData->blockedMessageQueue)
            {
                OALRemoteChannel_ProcessBlockedMessages(blackboardData->associatedChannel, TRUE);
            }
            if (!(oldList & V0_QUERY_BIT_OAP_V1) && (blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids & V0_QUERY_BIT_OAP_V1))
            {
                DOFRouter_ChannelUp(blackboardData->associatedChannel);
            }
        }

        /* Call internal destroy to offset the call to check existing operation. */
        DOFOperation_Internal_Destroy(queryOp);
        /* The response has been processed so cancel the operation. */
        DOFOperation_Cancel(queryOp);
    }
    return PLRESULT_OKDONOTHING;
}

static ProtocolLayerResult ESP_V2_ReceiveOpen(BlackboardData blackboardData)
{
    uint32      APPIDsToAdvertise = ESP_V0_QUERY_BITMASK;    /* Start with all APPIDs*/
    OALStream   openResponseStream;

    if (blackboardData->flags.isMulticast)
    {
        /* Multicast should not send an open. */
        return PLRESULT_ERROR;
    }

    /* create a stream for the APPID list, two bytes for each APPID plus three for the ESP APPID and opcode. */
    openResponseStream = OALStream_CreateRW(ESP_V0_APPID_COUNT * 2 + 3);

    if (!openResponseStream)
        return PLRESULT_ERROR;

    blackboardData->flags.isCommandNotResponse = FALSE;

    /* Write ESP appID, and the query response opcode. */
    if (OALMarshal_StreamCompress2(openResponseStream, PROTOCOLID_ESP_V2) != DOFERROR_OK)   /* Write the ESP_V2 appID. */
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ESP V2: send ESP query response: failed to write to command stream.");
        OALStream_Release(openResponseStream, FALSE);
        return PLRESULT_ERROR;
    }
    if (OALRemoteChannelConfig_GetSecurityDesire(blackboardData->associatedChannel) & DOFSECURITYDESIRE_SECURE_ANY)
    {
        if (!OALStream_Write8(openResponseStream, ESP_V2_OPCODE_OPENSECURE))
        {
            DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ESP V2: send ESP query response: failed to write to command stream.");
            OALStream_Release(openResponseStream, FALSE);
            return PLRESULT_ERROR;
        }
    }
    else
    {
        if (!OALStream_Write8(openResponseStream, ESP_V2_OPCODE_OPEN))
        {
            DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ESP V2: send ESP query response: failed to write to command stream.");
            OALStream_Release(openResponseStream, FALSE);
            return PLRESULT_ERROR;
        }
    }

    if (blackboardData->opID && !blackboardData->associatedChannel->remoteChannelData->flags.isConnection)
    {
        DOFOperation    openOp;

        /*  If there is an opID and delay the remote node is trying to start a statefull datagram connection.
            Create an inbound connection and channel and respond using the new channel.*/

        /* See if the operation exists, it might be a retry. */
        openOp = OALOperationManager_CheckExistingOperation(DOFSecurityScope_GetDomainID(DOF_GetDefaultSecurityScope(DOF_GetDof())),blackboardData->opID);

        if (!openOp && blackboardData->delay > 0)
        {
            /*  Create an operation, if the operation times out before the handshake is complete it will close
                the inbound connection.  Cap the timeout at 30 seconds. */
            openOp = DOFOperation_ESP_V2_Open_CreateInternal(blackboardData->opID, blackboardData->transportAddr,
                                                               (blackboardData->delay>30000)?30000:blackboardData->delay);

            if (openOp)
            {
                DOFOperation_Lock(openOp);
                PCR_Assert(openOp->operation_will_close);
                if (DOFOperation_AddOpWillCloseObserver(openOp, ESP_OpenOp_Server_Observer_fn))
                {
                    if (!OALOperationManager_AddOperation(openOp, DOFOperation_GetOperationID(openOp), FALSE) == DOFERROR_OK)
                    {
                        DOFOperation_Internal_Destroy(openOp);
                        DOFOperation_Unlock(openOp);
                        openOp = NULL;
                    }
                }
                else
                {
                    DOFOperation_Internal_Destroy(openOp);
                    DOFOperation_Unlock(openOp);
                    openOp = NULL;
                }
                if (openOp)
                    DOFOperation_Unlock(openOp);
            }

            if (!openOp)
            {
                /* If the operation was not created for some reason, send back and empty list. */
                APPIDsToAdvertise = 0;
            }
        }

        if (openOp && blackboardData->delay > 0)
        {
            if (openOp->sources->channel == NULL)
            {
                openOp->sources->channel = OALChannelMgr_CreateChannel_ConnectionlessInbound(blackboardData->associatedChannel,
                                                                                              DOFCONNECTIONTYPE_DATAGRAM,
                                                                                              blackboardData->transportAddr, NULL, NULL);
            }

            if (openOp->sources->channel)
            {
                if (blackboardData->associatedChannel != openOp->sources->channel)
                {
                    OALChannel_Release(blackboardData->associatedChannel);
                    blackboardData->associatedChannel = OALChannel_Retain(openOp->sources->channel);

                    if(blackboardData->associatedChannel)
                    {
                        APPIDsToAdvertise = buildAdvertisedAPPIDList(openOp->sources->channel, TRUE);
                    }
                    else
                    {
                        /* Since the connection cannot be established, respond with an empty list. */
                        APPIDsToAdvertise = 0;
                    }
                }
                else
                {
                    APPIDsToAdvertise = buildAdvertisedAPPIDList(openOp->sources->channel, TRUE);
                }
            }
            else
            {
                /* Since the connection cannot be established, respond with an empty list. */
                APPIDsToAdvertise = 0;
            }
            DOFOperation_Internal_Destroy(openOp);
        }
        else if (openOp && blackboardData->delay == 0)
        {
            /*  This is a cancel for the operation. Call internal destroy to offset the call to
                CheckExistingOperation, and then Destroy to cancel and finsh the destroy. */
            if (openOp->sources->channel && openOp->sources->channel->remoteChannelData)
            {
                openOp->sources->channel->remoteChannelData->flags.mustTerminate = FALSE;
            }
            DOFOperation_Internal_Destroy(openOp);
            DOFOperation_Cancel(openOp);
            OALStream_Release(openResponseStream, FALSE);
            return PLRESULT_ERROR;
        }
        else if (!openOp && blackboardData->delay == 0)
        {
            /*  This is like receiving an in session query out of session.  Return the list of APPIDs that are
                acceptable at the time, and do not use an ENP port. */
            APPIDsToAdvertise = buildAdvertisedAPPIDList(blackboardData->associatedChannel, TRUE);
        }
        else
        {
            /* If the operation can not be created, return an empty list of APPIDs. */
            APPIDsToAdvertise = 0;
        }
    }
    else
    {
        /* It is a protocol error if the the open does not have an opID so just ignore it. */
        OALStream_Release(openResponseStream, FALSE);
        return PLRESULT_ERROR;
    }

    if (!ESP_V2_StreamCreateAPPIDList(openResponseStream, APPIDsToAdvertise))
    {
        OALStream_Release(openResponseStream, FALSE);
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ESP V2: send ESP open response: failed to write to APPID list to command stream.");
        return PLRESULT_ERROR;
    }

    /* Write the lst to the send packet. */
    if (!blackboardData->sendPacket)
        blackboardData->sendPacket = OALPacket_Create();
    if (!blackboardData->sendPacket)
    {
        OALStream_Release(openResponseStream, FALSE);
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "ESP V2: handle ESP V0 query: Failed to create packet for APP V0 query." );
        return PLRESULT_ERROR;
    }

    /* OALPacket_AppendStream will take ownership of the stream, but it does not retain it. */
    if (!OALPacket_AppendStream(blackboardData->sendPacket, openResponseStream))
    {
        OALStream_Release(openResponseStream, FALSE);
        return PLRESULT_ERROR;
    }

    return PLRESULT_OKSENDDOWN;
}

static ProtocolLayerResult ESP_V2_HandleOpenResponse(BlackboardData blackboardData, boolean securityRequired)
{
    ProtocolLayerResult retStatus = PLRESULT_OKDONOTHING;

    if (blackboardData->opID && blackboardData->transportAddr)
    {
        DOFOperation openOp;

        openOp = OALOperationManager_CheckExistingOperation(DOFSecurityScope_GetDomainID(DOF_GetDefaultSecurityScope(DOF_GetDof())), blackboardData->opID);
        if (openOp)
        {
            DOFOperation_Lock(openOp);
            if (openOp->sources->channel)
            {
                openOp->sources->channel->remoteChannelData->ESP_V0_QueryAppids = ESP_V2_StreamUnmarshalAPPIDList(blackboardData->receiveStream);
                if ( openOp->sources->channel->remoteChannelData->negotiator == DOF_DEFAULT_AS_PROTOCOL_NEGOTIATOR )
                    openOp->sources->channel->remoteChannelData->ESP_V0_QueryAppids &= ~V0_QUERY_BIT_OAP_V1;
                if (blackboardData->transportAddr)
                {
                    if (!DOFAddress_IsAddressInfoEqual(blackboardData->transportAddr, openOp->sources->address))
                    {
                        DOFOperation_Unlock(openOp);
                        DOFOperation_Cancel(openOp);
                        /* Call internal destroy to offset the call to check existing operation. */
                        DOFOperation_Internal_Destroy(openOp);
                        /* The open response came from a different address, so drop the response. */
                        return PLRESULT_ERROR;
                    }

                    /* Stateless datagram connections can't use security, but if the required protocols are OK, we can still open stateless. */
                    if (securityRequired && OALRemoteChannelConfig_GetConnectionType( blackboardData->associatedChannel ) != DOFCONNECTIONTYPE_DATAGRAM_STATELESS)
                    {
                        if (OALRemoteChannelConfig_GetSecurityDesire(blackboardData->associatedChannel) == DOFSECURITYDESIRE_NOT_SECURE)
                        {
                            DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "Connectionless remote channel, handle ESP V2 open response: server requires security, client requires no security.");
                            OALRemoteChannel_Shutdown(openOp->sources->channel, DOFEXCEPTION_PROTOCOL_NEGOTIATION_FAILED);
                            DOFOperation_Unlock(openOp);
                            DOFOperation_Cancel(openOp);
                            /* Call internal destroy to offset the call to check existing operation. */
                            DOFOperation_Internal_Destroy(openOp);
                            return PLRESULT_OKDONOTHING;
                        }
                        blackboardData->associatedChannel->remoteChannelData->flags.secureDesire = TRUE;
                    }

                    /* Verify the required protocols are supported by the remote node. */
                    if (verifyRequiredProtocols(openOp->sources->channel, openOp->sources->channel->remoteChannelData->ESP_V0_QueryAppids))
                    {
                        if (openOp->sources->channel->remoteChannelData->flags.lossySession)
                        {
                            /* If the connection type is datagram or it is known that it is secure then it must be starting a session. */

                            /* Verify that the ENP Port is valid. */
                            if (DOFAddress_GetENPPort(blackboardData->transportAddr) == ENP_INVALID_PORT)
                            {
                                DOFException except;

                                DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ESP V2: handle ESP V2 open response: received and invalid ENP port in and open response.");
                                except = DOFException_Copy(DOFEXCEPTION_PROTOCOL_NEGOTIATION_FAILED);
                                OALRemoteChannel_Shutdown(openOp->sources->channel, except);
                                DOFException_Destroy(except);
                                DOFOperation_Unlock(openOp);
                                DOFOperation_Cancel(openOp);
                                /* Call internal destroy to offset the call to check existing operation. */
                                DOFOperation_Internal_Destroy(openOp);
                                return PLRESULT_ERROR;
                            }

                            /* Copy the new address into the channel since it now has the ENP port added to it. */
                            DOFAddress_Destroy(openOp->sources->channel->remoteChannelData->address);
                            openOp->sources->channel->remoteChannelData->address = DOFAddress_Copy(blackboardData->transportAddr);
                            if (!openOp->sources->channel->remoteChannelData->address)
                            {
                                DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ESP V2: handle ESP V2 open response: failed to copy address into channel.");
                                OALRemoteChannel_Shutdown(openOp->sources->channel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
                                DOFOperation_Unlock(openOp);
                                DOFOperation_Cancel(openOp);
                                /* Call internal destroy to offset the call to check existing operation. */
                                DOFOperation_Internal_Destroy(openOp);
                                return PLRESULT_ERROR;
                            }

                            /* Success. */
                            /* The EPS session is up now, so if it goes down send a terminate. */
                            openOp->sources->channel->remoteChannelData->flags.mustTerminate = TRUE;
                            openOp->sources->channel->remoteChannelData->flags.waitingForInitialDatagram = FALSE;
                        }
                        OALRemoteChannel_NotifyInitComplete(openOp->sources->channel);
                    }
                    else
                    {
                        DOFException except;

                        /* If the required protocols are not available, send a cancel for the ESP open operation. */
                        /*  Set the address in blackboard data back to the address in the operation, so it sends on the
                            "none" session, or in other words without the ENP port. */
                        DOFAddress_Destroy(blackboardData->transportAddr);
                        blackboardData->transportAddr = DOFAddress_Copy(openOp->sources->address);
                        if (blackboardData->transportAddr)
                        {
                            /* Set the timeout to zero so that it will cancel the operation. */
                            DOFOperation_SetTimeout(openOp, 0);
                            send_ESP_V2_Command(blackboardData, openOp);
                            /* Set the status to ok send down so that the response will be sent. */
                            retStatus = PLRESULT_OKSENDDOWN;
                        }
                        openOp->sources->channel->remoteChannelData->flags.mustTerminate = FALSE;
                        except = DOFException_Copy(DOFEXCEPTION_PROTOCOL_NEGOTIATION_FAILED);
                        OALRemoteChannel_Shutdown(openOp->sources->channel, except);
                        DOFException_Destroy(except);
                    }
                }
            }
            DOFOperation_Unlock(openOp);
            /* The response has been processed so cancel the operation. */
            DOFOperation_Cancel(openOp);
            /* Call internal destroy to offset the call to check existing operation. */
            DOFOperation_Internal_Destroy(openOp);
        }
    }
    return retStatus;
}

static ProtocolLayerResult ESP_V2_ReceiveOpenResponse(BlackboardData blackboardData)
{
    return ESP_V2_HandleOpenResponse(blackboardData, FALSE);
}

static ProtocolLayerResult ESP_V2_ReceiveOpenResponseSecure(BlackboardData blackboardData)
{
    return ESP_V2_HandleOpenResponse(blackboardData, TRUE);
}

static ProtocolLayerResult ESP_V2_ReceiveTerminate(BlackboardData blackboardData)
{
    OALStream       terminateResponseStream;
    DOFException    except;

    if (blackboardData->flags.isMulticast)
    {
        /* Multicast should not send an open. */
        return PLRESULT_ERROR;
    }

    if (!blackboardData->associatedChannel->remoteChannelData->flags.isConnection)
    {
        /* If this is not a connection then just exit. */
        return PLRESULT_ERROR;
    }

    /* Since this channel was terminated clear the mustTerminate flag so that this channel will not send a terminate command. */
    blackboardData->associatedChannel->remoteChannelData->flags.mustTerminate = FALSE;
    except = DOFErrorException_Create(DOFERROR_TERMINATED, "Connection Closed by remote host");
    OALRemoteChannel_Shutdown(blackboardData->associatedChannel, except);
    DOFException_Destroy(except);

    /* create a stream for the APPID list, two bytes for each APPID plus three for the ESP APPID and opcode. */
    terminateResponseStream = OALStream_CreateRW(3);

    if (!terminateResponseStream)
        return PLRESULT_ERROR;

    blackboardData->flags.isCommandNotResponse = FALSE;

    /* Write ESP appID, and the query response opcode. */
    if (OALMarshal_StreamCompress2(terminateResponseStream, PROTOCOLID_ESP_V2) != DOFERROR_OK)   /* Write the ESP_V2 appID. */
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ESP V2: send ESP query response: failed to write to command stream.");
        OALStream_Release(terminateResponseStream, FALSE);
        return PLRESULT_ERROR;
    }
    if (!OALStream_Write8(terminateResponseStream, ESP_V2_OPCODE_TERMINATE))
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "ESP V2: send ESP query response: failed to write to command stream.");
        OALStream_Release(terminateResponseStream, FALSE);
        return PLRESULT_ERROR;
    }

    /* Write the lst to the send packet. */
    if (!blackboardData->sendPacket)
        blackboardData->sendPacket = OALPacket_Create();
    if (!blackboardData->sendPacket)
    {
        OALStream_Release(terminateResponseStream, FALSE);
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "ESP V2: handle ESP V0 query: Failed to create packet for APP V0 query." );
        return PLRESULT_ERROR;
    }
    /* OALPacket_AppendStream will take ownership of the stream, but it does not retain it. */
    if (!OALPacket_AppendStream(blackboardData->sendPacket, terminateResponseStream))
    {
        OALStream_Release(terminateResponseStream, FALSE);
        return PLRESULT_ERROR;
    }

    return PLRESULT_OKSENDDOWN;
}

static ProtocolLayerResult ESP_V2_ReceiveTerminateResponse(BlackboardData blackboardData)
{
    return PLRESULT_OKDONOTHING;
}

ProtocolLayerResult ESP_V2_Receive(BlackboardData blackboardData)
{
    OALStream           stream = blackboardData->receiveStream;
    uint8               opcode;
    ProtocolLayerResult retVal = PLRESULT_ERROR;
    uint16              appID;

#define OPCODE_LIMIT 8
    /* Note: this list is in opcode order. */
    static const ESP_V2_Receive_fn_t receiveFunctions[OPCODE_LIMIT][2] = {
        {ESP_V2_ReceiveQueryResponse, ESP_V2_ReceiveQuery},             /* 0  Query */
        {NULL, NULL},   /* 1  Configuration request */
        {NULL, NULL},   /* 2  Configuration acknowledge */
        {NULL, NULL},   /* 3  Configuration negative acknowledge */
        {NULL, NULL},   /* 4  Configuration reject */
        {ESP_V2_ReceiveTerminateResponse, ESP_V2_ReceiveTerminate},     /* 5  Terminate */
        {ESP_V2_ReceiveOpenResponse, ESP_V2_ReceiveOpen},               /* 6  Open */
        {ESP_V2_ReceiveOpenResponseSecure, NULL}                        /* 7  Open (secure) */
    };

    /* After a channel becomes secure, verify that all subsequent packets are secure. */
    if (blackboardData->associatedChannel->remoteChannelData->flags.sessionIsSecure &&
        !(blackboardData->flags.receivedAuthenticated || blackboardData->flags.receivedEncrypted))
    {
        return PLRESULT_ERROR;
    }

    /* Read the appID, since it is 0 it should only read 1 byte. */
    if ((OALStream_GetDataSize(stream) == 0) || (OALMarshal_StreamUncompress2(stream, &appID) != DOFERROR_OK))
    {
        return PLRESULT_ERROR;
    }

    if (OALStream_GetDataSize(stream) == 0)
    {
        return PLRESULT_OKDONOTHING;
    }
    else
    {
        /* Read one byte for the opcode. */
        if(!OALStream_Read8(stream, &opcode))
        {
            return PLRESULT_ERROR;
        }
    }

    if (opcode < OPCODE_LIMIT)
    {
        ESP_V2_Receive_fn_t fn = receiveFunctions[opcode][blackboardData->flags.isCommandNotResponse];
        if (fn)
        {
            retVal = fn(blackboardData);
        }
    }

    return retVal;
}

static ProtocolLayerResult send_query(BlackboardData blackboardData, DOFOperation op)
{
    ProtocolLayerResult status;

    status = send_ESP_V2_Command(blackboardData, op);
    return status;
}

static ProtocolLayerResult send_terminate(BlackboardData blackboardData, DOFOperation op)
{
    ProtocolLayerResult status;

    if (blackboardData->associatedChannel->remoteChannelData->flags.isConnection &&
        !blackboardData->associatedChannel->remoteChannelData->flags.defaultAddressIsMulticast)
    {
        status = send_ESP_V2_Command(blackboardData, op);
    }
    else
    {
        status = PLRESULT_OKDONOTHING;
    }
    return status;
}

static ProtocolLayerResult send_open(BlackboardData blackboardData, DOFOperation op)
{
    return send_ESP_V2_Command(blackboardData, op);
}

ProtocolLayerResult ESP_V2_Send(BlackboardData blackboardData, DOFOperation op)
{
    ProtocolLayerResult result;
    /* Note: this list is in operation type order. and is for DOFOPERATIONTYPE_ESP_V2_QUERY - DOFOPERATIONTYPE_ESP_LAST*/
    static const ESP_V2_Send_fn_t sendFunctions[] = {
        send_query,             /* DOFOPERATIONTYPE_ESP_V2_QUERY */
        send_terminate,         /* DOFOPERATIONTYPE_ESP_V2_TERMINATE */
        send_open               /* DOFOPERATIONTYPE_ESP_V2_OPEN */
    };

    PCR_Assert(op->operation_type >= DOFOPERATIONTYPE_ESP_V2_QUERY && op->operation_type < DOFOPERATIONTYPE_ESP_LAST);

    result = sendFunctions[op->operation_type - DOFOPERATIONTYPE_ESP_V2_QUERY](blackboardData, op);

    if (blackboardData->opID)
        blackboardData->seq = 0;

    return result;
}

static void ESPV2Operation_Observer_fn(OALObserver self, OALObservable observable, void* ctx)
{
    DOFOperation_Internal_Destroy((DOFOperation)ctx);
}

static void ESP_QueryOp_InboundConnection_Observer_fn(OALObserver self, OALObservable observable, void* ctx)
{
    DOFOperation queryOp = (DOFOperation)ctx;
    DOFOperation_Lock(queryOp);
    if(queryOp->sources->channel)
        OALRemoteChannel_APPIDQueryTimeout(queryOp->sources->channel);
    DOFOperation_Unlock(queryOp);
    DOFOperation_Internal_Destroy(queryOp);
}

boolean ESP_V2_SendQuery(OALChannel channel, uint32 timeout, uint32 retry)
{
    boolean status = FALSE;
    OALObserver_fn_t observerFn;

    if (channel && channel->remoteChannelData && channel->remoteChannelData->flags.isConnection)
    {
        DOFOperation appQueryOp = DOFOperation_ESP_V2_Query_Create(channel->remoteChannelData->address, retry);
        if (appQueryOp)
        {
            appQueryOp->sources->channel = OALChannel_Retain(channel);
            if(appQueryOp->sources->channel)
            {
                DOFOperation_SetTimeout(appQueryOp, timeout);
                if (DOFOperationManager_SetRetry(appQueryOp, appQueryOp->sources->channel) == DOFERROR_OK)
                {
                    PCR_Assert(appQueryOp->operation_will_close);

                    if (channel->remoteChannelData->flags.isOutbound)
                        observerFn = ESPV2Operation_Observer_fn;
                    else
                        observerFn = ESP_QueryOp_InboundConnection_Observer_fn;

                    if (DOFOperation_AddOpWillCloseObserver(appQueryOp, observerFn))
                    {
                        if (OALOperationManager_AddOperation(appQueryOp, DOFOperation_GetOperationID(appQueryOp), TRUE) == DOFERROR_OK)
                        {
                            channel->routeToProvider(channel, appQueryOp, NULL, appQueryOp->sources->address);
                            status = TRUE;
                        }
                    }
                }
            }
            DOFOperation_Internal_Destroy(appQueryOp);
        }
    }
    return status;
}

boolean ESP_V2_SendTerminate(OALChannel channel)
{
    boolean status = FALSE;

    if (channel && channel->remoteChannelData && channel->remoteChannelData->flags.isConnection)
    {
        DOFOperation terminateOp = DOFOperation_ESP_V2_Terminate_Create(channel->remoteChannelData->address);
        if (terminateOp)
        {
            terminateOp->sources->channel = OALChannel_Retain(channel);
            if(terminateOp->sources->channel)
            {
                channel->routeToProvider(channel, terminateOp, NULL, terminateOp->sources->address);
                status = TRUE;
            }
            DOFOperation_Internal_Destroy(terminateOp);
        }
    }
    return status;
}

boolean ESP_V2_SendOpen(OALChannel channel)
{
    boolean status = FALSE;

    if (channel && channel->remoteChannelData && channel->remoteChannelData->flags.isConnection)
    {
        DOFOperation openOp = DOFOperation_ESP_V2_Open_Create(channel->remoteChannelData->address);
        if (openOp)
        {
            openOp->sources->channel = OALChannel_Retain(channel);
            if(openOp->sources->channel)
            {
                PCR_Assert(openOp->operation_will_close);
                if (DOFOperation_AddOpWillCloseObserver(openOp, ESPV2Operation_Observer_fn))
                {
                    DOFOperation_SetTimeout(openOp, 30000);     /* 30 second timeout to match TCP connections. */

                    if (OALOperationManager_AddOperation(openOp, DOFOperation_GetOperationID(openOp), FALSE) == DOFERROR_OK)
                    {
                        /*  The operation manager will take ownership of the operation.  The operation will either
                            time out or be destroyed at the point when an open response is received. */

                        if (DOFOperationManager_SetRetry(openOp, openOp->sources->channel) == DOFERROR_OK)
                        {

                            if (channel->remoteChannelData->flags.lossySession)
                            {
                                channel->remoteChannelData->flags.waitingForInitialDatagram = TRUE;
                            }
                            else
                            {
                                /*  If this is not for a lossy session set the delay to zero to get the list of currently acceptable APPIDs.
                                    This must be done after DOFOperation_Set Timeout and OALOperationManager_AddOperation. */
                                DOFOperation_SetDelay(openOp, 0);
                            }

                            channel->routeToProvider(channel, openOp, NULL, openOp->sources->address);
                            status = TRUE;
                        }
                    }
                }
            }
            DOFOperation_Internal_Destroy(openOp);
        }
    }
    return status;
}




