/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/dofinternal.h>
#include <internal/esp_v2_options.h>
#include <internal/securitymode.h>
#include <dof/securitymode.h>
#include <internal/esp_v2.h>
#include <internal/trp_v1.h>
#include <internal/protocolstack.h>

/*
**  "Private" Type and Method Declarations
*/


/*
**  "Public" Methods
*/

ESPV2_ConfigOption  ESPV2_ConfigOption_Create(uint8 attributeCode, uint16 attributeData, uint8 valueLength)
{
    ESPV2_ConfigOption retVal = (ESPV2_ConfigOption)DOFMem_Alloc( offsetof(ESPV2_ConfigOption_t, valueData) + valueLength );

    if(retVal)
    {
        retVal->refCount = 0x01;
        retVal->attributeCode = attributeCode;
        retVal->attributeData = attributeData;
        retVal->valueLength = valueLength;
    }

    return retVal;
}

ESPV2_ConfigOption ESPV2_ConfigOption_CreateFromMode(DOFSecurityMode mode)
{
    ESPV2_ConfigOption retVal = ESPV2_ConfigOption_Create(ESP_ATTR_TYPE_CCM, PROTOCOLID_CCMMODE_V1, 3);

    if( retVal )
    {
        if( !DOFSecurityModeCCM_MarshalData(mode, retVal->valueData) )
        {
            ESPV2_ConfigOption_Destroy(retVal);
            retVal = NULL;
        }
    }
    return retVal;
}

ESPV2_ConfigOption ESPV2_ConfigOption_CreateSec(DOFSecurityDesire desire)
{
    /* add new security modes here!!! for ESP */
    ESPV2_ConfigOption retVal = ESPV2_ConfigOption_Create(ESP_ATTR_TYPE_CCM, PROTOCOLID_CCMMODE_V1, 3);

    if( retVal )
    {
        retVal->valueData[0] = 1;
        retVal->valueData[1] = DOFCIPHERSTRENGTH_256;
        retVal->valueData[2] = 0x39; /* Tmax = 7(16) Tmin = 1(4)*/

        if( desire ==  DOFSECURITYDESIRE_SECURE_ANY)
        {
            retVal->valueData[2] |= 0x40;
        }
        else if( desire == DOFSECURITYDESIRE_SECURE )
        {
            retVal->valueData[2] |= 0xC0;
        }
    }

    return retVal;
}

DOFSecurityMode ESPV2_ConfigOption_SecModeFromConfigOption(ESPV2_ConfigOption self, DOFSecurityDesire desire, boolean* nak)
{
    if( self )
    {
        uint8 pos = 1;
        uint8 i = 0;
        uint8 cCnt = self->valueData[0];

        if( self->valueLength > (1+cCnt) )
        {
            for(; i < cCnt; ++i, ++pos ) 
            {
                if(self->valueData[pos] == DOFCIPHERSTRENGTH_256 )
                {
                    uint8 Tmin;
                    uint8 Tmax;
                    uint8 sec;

                    pos = 1+cCnt; /* calculate the position of the tmin tmax... */

                    Tmin = ((self->valueData[pos] & 0x07)*2)+2;
                    Tmax = (((self->valueData[pos] & 0x38)>>3)*2)+2;
                    sec = (self->valueData[pos] & 0xC0)>>6;

                    if( Tmin < 4 || Tmax > 16 || Tmax < Tmin )
                    {
                        return NULL;
                    }

                    switch( desire )
                    {
                    case DOFSECURITYDESIRE_SECURE_AUTHENTICATE_ONLY:
                        if( !(sec == 0x0) )
                            *nak = TRUE;
                        break;
                    case DOFSECURITYDESIRE_SECURE:
                        if( !(sec == 0x3) )
                            *nak = TRUE;
                        break;
                    default:
                        break;
                    }

                    return DOFSecurityModeCCM_Create( DOFCIPHERALGORITHM_AES, DOFCIPHERSTRENGTH_256, Tmin, (sec & 0x02));
                }
            }
        }
    }

    return NULL;
}

ESPV2_ConfigOption ESPV2_ConfigOption_Unmarshall(OALStream stream)
{
    uint8   attributeCode;
    uint8   valueLength;
    uint16  attributeData;
    ESPV2_ConfigOption cfgOpt = NULL;

    if( OALStream_GetDataSize(stream) >= 4 &&
        OALStream_Read8( stream, &attributeCode ) &&
        OALStream_Read16( stream, &attributeData ) &&
        OALStream_Read8( stream, &valueLength ) &&
        OALStream_GetDataSize(stream) >= valueLength )
    {
        cfgOpt = ESPV2_ConfigOption_Create( attributeCode, attributeData, valueLength );
        if( cfgOpt )
        {
            if( !OALStream_Read( stream, valueLength, cfgOpt->valueData ))
            {
                /* Error */
                ESPV2_ConfigOption_Destroy(cfgOpt);
                cfgOpt = NULL;
            }
        }
    }

    return cfgOpt;
}

ConfigOptionList ESPV2_ConfigOption_Unmarshall_List(OALStream stream)
{
    ConfigOptionList retVal = OALPtrList_Create( 4 );/* initial size is just a guess. */

    if(retVal != NULL)
    {
        uint8 i = 0;
        while( OALStream_GetDataSize(stream) > 0 )
        {
            ESPV2_ConfigOption configOption = ESPV2_ConfigOption_Unmarshall(stream);
            if(!configOption || !OALPtrList_InsertAt( retVal, configOption, i++ ) )
            {
                goto HANDLE_ERROR_CLEANUP;
            }
        }
        return retVal;

        HANDLE_ERROR_CLEANUP:
        {
            while( OALPtrList_GetCount(retVal) )
            {
                ESPV2_ConfigOption configOption = OALPtrList_RemoveAt( retVal, 0 );
                ESPV2_ConfigOption_Destroy( configOption );
            }
            OALPtrList_Destroy( retVal );
            return NULL;
        }
    }
    return NULL;
}

boolean ESPV2_ConfigOption_Marshall(ESPV2_ConfigOption cfgOpt, OALStream stream)
{
    if( OALStream_Write8( stream, cfgOpt->attributeCode ) &&
        OALStream_Write8( stream, (uint8)(cfgOpt->attributeData >> 8 )) &&
        OALStream_Write8( stream, (uint8)(cfgOpt->attributeData) ) &&
        OALStream_Write8( stream, cfgOpt->valueLength ) &&
        OALStream_Write( stream, cfgOpt->valueLength, cfgOpt->valueData ) )
        return TRUE;
    return FALSE;
}

OALStream ESPV2_ConfigOption_Marshall_List(ConfigOptionList cfgOptList)
{
    uint32 cfgOptsCount;
    uint32 totalSize = 0;
    uint32 i;
    OALStream retVal;

    PCR_Assert( cfgOptList != NULL );
    cfgOptsCount = OALPtrList_GetCount( cfgOptList );
    for(i = 0; i < cfgOptsCount; ++i)
    {
        totalSize += ((ESPV2_ConfigOption)OALPtrList_PeekAt( cfgOptList, i ))->valueLength + 4;
    }

    retVal = OALStream_CreateRW(totalSize);
    if( retVal )
    {
        for(i = 0; i < cfgOptsCount; ++i)
        {
            if( !ESPV2_ConfigOption_Marshall( (ESPV2_ConfigOption)OALPtrList_PeekAt( cfgOptList, i ), retVal) )
            {
                OALStream_Release(retVal, FALSE);
                return NULL;
            }
        }
    }

    return retVal;
}

void ESPV2_ConfigOption_Destroy(ESPV2_ConfigOption configOpt)
{
    if( configOpt )
    {
        /* NOTE: This doesn't need to be protected because it only gets accessed in one thread. */
        --(configOpt->refCount);
        if( !(configOpt->refCount) )
        {
            DOFMem_Free(configOpt);
        }
    }
}

ESPV2_ConfigOption ESPV2_ConfigOption_Retain(ESPV2_ConfigOption self)
{
    if( self )
        ++(self->refCount);
    return self;
}

void ESPV2_ConfigOptionList_Destroy(ConfigOptionList optList)
{
    if(optList != NULL)
    {
        const uint32 listSize = OALPtrList_GetCount(optList);
        uint32 i = 0;

        for(; i < listSize; ++i)
        {
            ESPV2_ConfigOption_Destroy( OALPtrList_PeekAt(optList, i) );
        }
        OALPtrList_Destroy(optList);
    }
}

ConfigOptionList ESPV2_ConfigOptionList_Copy(ConfigOptionList optList)
{
    if ( optList )
    {
        const uint32 listSize = OALPtrList_GetCount(optList);
        ConfigOptionList retVal = OALPtrList_Create( listSize );

        if( retVal != NULL )
        {
            uint32 i;

            for(i = 0; i < listSize; ++i)
            {
                ESPV2_ConfigOption opt = OALPtrList_PeekAt(optList, i);
                ESPV2_ConfigOption_Retain( opt );
                OALPtrList_InsertAt( retVal, opt, i );
            }
            return retVal;
        }
    }
    return NULL;
}

static ESPV2_Reject ESPV2_Reject_Create_Internal( uint8 attributeCode, uint8 count )
{
    ESPV2_Reject retVal = (ESPV2_Reject)DOFMem_Alloc( offsetof(ESPV2_Reject_t, attributeData) + (sizeof(uint16) * count) );

    if(retVal)
    {
        retVal->refCount = 0x01;
        retVal->attributeCode = attributeCode;
        retVal->attributeDataCnt = count;
    }

    return retVal;
}

ESPV2_Reject ESPV2_Reject_Create( uint8 attributeCode )
{
    return ESPV2_Reject_Create_Internal( attributeCode, 0 );
}

ESPV2_Reject ESPV2_TEPReject_Create( void )
{
    ESPV2_Reject retVal = ESPV2_Reject_Create_Internal( ESP_ATTR_TYPE_TEP, 1 );

    if( retVal )
    {
        retVal->attributeData[0] = PROTOCOLID_TEP_V1;
    }

    return retVal;
}

ESPV2_Reject ESPV2_OAPReject_Create( void )
{
    ESPV2_Reject retVal = ESPV2_Reject_Create_Internal( ESP_ATTR_TYPE_OAP, 1 );

    if( retVal )
    {
        retVal->attributeData[0] = PROTOCOLID_OAP_V1;
    }

    return retVal;
}

ESPV2_Reject ESPV2_CCMReject_Create( void )
{
    ESPV2_Reject retVal = ESPV2_Reject_Create_Internal( ESP_ATTR_TYPE_CCM, 1 );

    if( retVal )
    {
        retVal->attributeData[0] = PROTOCOLID_CCMMODE_V1;
    }

    return retVal;
}

ESPV2_Reject ESPV2_TRPReject_Create( void )
{
    ESPV2_Reject retVal = NULL;

    retVal = ESPV2_Reject_Create_Internal( ESP_ATTR_TYPE_TRP, 1 );

    if( retVal )
    {
        retVal->attributeData[0] = PROTOCOLID_TRP_V1;
    }

    return retVal;
}

ESPV2_Reject ESPV2_GenReject_Create( void )
{
    ESPV2_Reject retVal = ESPV2_Reject_Create_Internal( ESP_ATTR_TYPE_Gen, 0 );

    if( retVal )
    {
        /* Nothing Yet */
    }

    return retVal;
}

ESPV2_Reject ESPV2_Reject_Unmarshall(OALStream stream)
{
    uint16  i;
    uint8   attributeCode;
    uint16  dataCount;
    ESPV2_Reject cfgOpt;

    if( OALStream_GetDataSize(stream) < 3 || !OALStream_Read8( stream, &attributeCode ) || !OALStream_Read16( stream, &dataCount ) || OALStream_GetDataSize(stream) < (dataCount * 2U))
        return NULL;

    cfgOpt = ESPV2_Reject_Create_Internal( attributeCode, dataCount );

    if(cfgOpt)
    {
        for( i = 0; i < dataCount; ++i )
        {
            OALStream_Read16( stream, &cfgOpt->attributeData[i] );
        }
    }
    return cfgOpt;
}

RejectList ESPV2_Reject_Unmarshall_List(OALStream stream)
{
    RejectList retVal = OALPtrList_Create( 2 );/* initial size is just a guess. */

    if(retVal != NULL)
    {
        uint8 i = 0;
        while( OALStream_GetDataSize(stream) > 0 )
        {
            ESPV2_Reject configReject = ESPV2_Reject_Unmarshall(stream);
            if(!configReject || !OALPtrList_InsertAt( retVal, configReject, i++ ))
                goto HANDLE_ERROR_CLEANUP;
        }
        return retVal;

        HANDLE_ERROR_CLEANUP:
        {
            while( OALPtrList_GetCount(retVal) )
            {
                ESPV2_Reject configReject = OALPtrList_RemoveAt( retVal, 0 );
                ESPV2_Reject_Destroy( configReject );
            }
            OALPtrList_Destroy( retVal );
            return NULL;
        }
    }
    return NULL;
}

boolean ESPV2_Reject_Marshall(ESPV2_Reject reject, OALStream stream)
{
    uint16 i = 0;
    if( OALStream_Write8( stream, reject->attributeCode ) &&
        OALStream_Write8( stream, (uint8)(reject->attributeDataCnt >> 8 )) &&
        OALStream_Write8( stream, (uint8)(reject->attributeDataCnt) ))
    {
        for( ; i < reject->attributeDataCnt; ++i )
        {
            if( !OALStream_Write8( stream, (uint8)(reject->attributeData[i] >> 8 )) ||
                !OALStream_Write8( stream, (uint8)(reject->attributeData[i]) ) )
            {
                return FALSE;
            }
        }
        return TRUE;
    }
    return FALSE;
}

OALStream ESPV2_Reject_Marshall_List(RejectList rejectList)
{
    uint32 cfgRejectCount;
    uint32 totalSize = 0;
    uint32 i;
    OALStream retVal;

    PCR_Assert( rejectList != NULL );
    cfgRejectCount = OALPtrList_GetCount( rejectList );
    for(i = 0; i < cfgRejectCount; ++i)
    {
        totalSize += (((ESPV2_Reject)OALPtrList_PeekAt( rejectList, i ))->attributeDataCnt * 2) + 3;
    }

    retVal = OALStream_CreateRW(totalSize);
    if(retVal)
    {
        for(i = 0; i < cfgRejectCount; ++i)
        {
            if( !ESPV2_Reject_Marshall( (ESPV2_Reject)OALPtrList_PeekAt( rejectList, i ), retVal) )
            {
                OALStream_Release( retVal, FALSE );
                return NULL;
            }
        }
    }
    return retVal;
}

void ESPV2_Reject_Destroy( ESPV2_Reject self )
{
    DOFMem_Free( self );
}

void ESPV2_RejectList_Destroy(RejectList optList)
{
    uint32 listSize;
    uint32 i = 0;

    PCR_Assert( optList != NULL );
    listSize = OALPtrList_GetCount( optList );
    for(; i < listSize; ++i)
    {
        ESPV2_Reject_Destroy( OALPtrList_PeekAt(optList, i) );
    }

    OALPtrList_Destroy(optList);
}


/* FOR TEP */
/* False return is a reject */
/* secMode = Null in nak case means that the one they wanted doesn't work with our desire */
/* reject secMode will always be null */
/* *streamError is TRUE if unmarshalling issues occurred */
boolean ESPV2_CheckRejectOrNakIfNakReturnASecurityMode(OALStream stream, DOFSecurityDesire desire, DOFSecurityMode* secMode, boolean * streamError)
{
    uint8 opcode = 0;

    PCR_Assert(stream);
    PCR_Assert(streamError != NULL);

    if(!OALStream_Read8(stream,&opcode))
    {
        *streamError = TRUE;
        return FALSE;
    }
    *streamError = FALSE;
    if( opcode == LCP_CONFIGURE_NAK )
    {
        if( secMode )
        {
            boolean nak = FALSE;
            ESPV2_ConfigOption cn = ESPV2_ConfigOption_Unmarshall(stream);
            if( cn )
            {
                DOFSecurityMode sec = ESPV2_ConfigOption_SecModeFromConfigOption(cn,desire,&nak);
                if( !nak )
                    (*secMode) = sec;
                else
                    DOFSecurityMode_Destroy(sec);
                ESPV2_ConfigOption_Destroy(cn);
                return TRUE;
            }
        }
    }

    return FALSE;
}

/* FOR TEP */
boolean ESPV2_MarshalCCMNak(OALStream stream, DOFSecurityDesire desire)
{
    boolean ret = FALSE;
    ESPV2_ConfigOption cfg = ESPV2_ConfigOption_CreateSec(desire);
    PCR_Assert(stream);

    if(cfg &&
       OALStream_Write8(stream, LCP_CONFIGURE_NAK) &&
       ESPV2_ConfigOption_Marshall(cfg,stream))
    {
        ret = TRUE;
    }
    ESPV2_ConfigOption_Destroy(cfg);
    return ret;
}

/* FOR TEP */
boolean ESPV2_MarshalCCMReject(OALStream stream)
{
    boolean ret = FALSE;
    ESPV2_Reject cfg = ESPV2_CCMReject_Create();
    PCR_Assert(stream);

    if(cfg &&
       OALStream_Write8(stream, LCP_CONFIGURE_REJECT) &&
       ESPV2_Reject_Marshall(cfg,stream))
    {
        ret = TRUE;
    }
    ESPV2_Reject_Destroy(cfg);
    return ret;
}
