/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/exception.h>
#include <dof/error.h>
#include <dof/value.h>
#include <dof/providerinfo.h>
#include <internal/exception.h>
#include <internal/iteminternal.h>
#include <internal/dofinternal.h>


extern PCRMutex ref_count_mutex;

#define CreateException( name )     DOFEXCEPTION_##name = DOFErrorException_Create( DOFERROR_##name, DOFError_GetString( DOFERROR_##name ) );


/** @addtogroup DOFException
@{
**/

/*
**  "Private" Type and Method Declarations
*/

typedef struct DOFException_t
{
    uint32           refCount;
    DOFExceptionType type;
    union {
        struct {
            const char * errorString;
            const void * data;
            DOFError code;
        } error;
        struct {
            DOFInterfaceException except;
            DOFValue * parameters;
        } provider;
        struct {
            OALStream stream;
            DOFItemID itemId;
        } stream;
    } u;
    DOFProviderInfo info;
    boolean isStream;
    boolean valuesAreSeperateAlloc;
} DOFException_t;

static DOFException_t outOfMemory;

DOFException DOFEXCEPTION_NOT_SUPPORTED = NULL;
DOFException DOFEXCEPTION_TYPE_MISMATCH = NULL;
DOFException DOFEXCEPTION_INSUFFICIENT_RESOURCES = NULL;
DOFException DOFEXCEPTION_TERMINATED = NULL;
DOFException DOFEXCEPTION_REJECTED = NULL;
DOFException DOFEXCEPTION_NOT_FOUND = NULL;
DOFException DOFEXCEPTION_DUPLICATE = NULL;
DOFException DOFEXCEPTION_TOO_MANY = NULL;
DOFException DOFEXCEPTION_APPLICATION_ERROR = NULL;

DOFException DOFEXCEPTION_FAILED = NULL;
DOFException DOFEXCEPTION_TIMEOUT = NULL;
DOFException DOFEXCEPTION_NOT_READABLE = NULL;
DOFException DOFEXCEPTION_NOT_WRITABLE = NULL;
DOFException DOFEXCEPTION_RESULT_LOST = NULL;
DOFException DOFEXCEPTION_PROTOCOL_NEGOTIATION_FAILED = NULL;
DOFException DOFEXCEPTION_TRANSPORT_FAILED = NULL;
DOFException DOFEXCEPTION_IO_FAILED = NULL;
DOFException DOFEXCEPTION_AUTHENTICATION_FAILED = NULL;
DOFException DOFEXCEPTION_ACK_TIMEOUT = NULL;
DOFException DOFEXCEPTION_STALE_CREDENTIAL = NULL;

DOFException DOFEXCEPTION_ACCESS_DENIED = NULL;
DOFException DOFEXCEPTION_UNKNOWN_INITIATOR = NULL;
DOFException DOFEXCEPTION_UNKNOWN_RESPONDER = NULL;
DOFException DOFEXCEPTION_UNKNOWN_DOMAIN = NULL;
DOFException DOFEXCEPTION_SERVICE_UNAVAILABLE = NULL;
DOFException DOFEXCEPTION_BAD_MODE = NULL;
DOFException DOFEXCEPTION_NOT_IN_GROUP = NULL;
DOFException DOFEXCEPTION_UNSUPPORTED_INITIATOR = NULL;
DOFException DOFEXCEPTION_UNSUPPORTED_RESPONDER = NULL;
DOFException DOFEXCEPTION_INVALID_INITIATOR = NULL;
DOFException DOFEXCEPTION_INVALID_RESPONDER = NULL;
DOFException DOFEXCEPTION_RESOLUTION_FAILED = NULL;
DOFException DOFEXCEPTION_VALIDATION_FAILED = NULL;

DOFException DOFEXCEPTION_ILLEGAL_ARGUMENT = NULL;

static DOFException exceptionMap_OAP[11];
static DOFException exceptionMap_API[11];
static DOFException exceptionMap_TRP[16];

static DOFException _create( DOFError error, const char * errorString, DOFProviderInfo info, const void * data )
{
    if ( error > DOFERROR_OK )
    {
        uint32       slen = ( errorString ) ? PCRString_Length( errorString ) + 1 : 0;
        DOFException e    = DOFMem_Alloc( sizeof( DOFException_t ) + slen );
        if ( e )
        {
            e->refCount = 1;
            e->type = DOFEXCEPTIONTYPE_ERROR;
            if ( errorString )
            {
                PCRMem_Copy( &e[1], errorString, slen );
                e->u.error.errorString = (const char *) &e[1];
            }
            else
            {
                e->u.error.errorString = DOFError_GetString( error );
            }
            e->info = DOFProviderInfo_Copy( info );
            e->u.error.code = error;
            e->u.error.data = data;
            e->isStream = FALSE;
            e->valuesAreSeperateAlloc = FALSE;
            return e;
        }
        return &outOfMemory;
    }
    return NULL;
}

void DOFException_Init( void )
{
    outOfMemory.type = DOFEXCEPTIONTYPE_ERROR;
    outOfMemory.u.error.code = DOFERROR_INSUFFICIENT_RESOURCES;
    outOfMemory.u.error.errorString = DOFError_GetString( DOFERROR_INSUFFICIENT_RESOURCES );

    CreateException( NOT_SUPPORTED );
    CreateException( TYPE_MISMATCH );
    CreateException( INSUFFICIENT_RESOURCES );
    CreateException( TERMINATED );
    CreateException( REJECTED );
    CreateException( NOT_FOUND );
    CreateException( DUPLICATE );
    CreateException( TOO_MANY );
    CreateException( APPLICATION_ERROR );

    CreateException( FAILED );
    CreateException( TIMEOUT );
    CreateException( NOT_READABLE );
    CreateException( NOT_WRITABLE );
    CreateException( RESULT_LOST );
    CreateException( PROTOCOL_NEGOTIATION_FAILED );
    CreateException( TRANSPORT_FAILED );
    CreateException( IO_FAILED );
    CreateException( AUTHENTICATION_FAILED );
    CreateException( ACK_TIMEOUT );
    CreateException( STALE_CREDENTIAL );

    CreateException( ACCESS_DENIED );
    CreateException( UNKNOWN_INITIATOR );
    CreateException( UNKNOWN_RESPONDER );
    CreateException( UNKNOWN_DOMAIN );
    CreateException( SERVICE_UNAVAILABLE );
    CreateException( BAD_MODE );
    CreateException( NOT_IN_GROUP );
    CreateException( UNSUPPORTED_INITIATOR );
    CreateException( UNSUPPORTED_RESPONDER );
    CreateException( INVALID_INITIATOR );
    CreateException( INVALID_RESPONDER );
    CreateException( RESOLUTION_FAILED );
    CreateException( VALIDATION_FAILED );

    CreateException( ILLEGAL_ARGUMENT );

    exceptionMap_OAP[DOFERROR_OK]                       = NULL;
    exceptionMap_OAP[1]                                 = DOFEXCEPTION_FAILED;
    exceptionMap_OAP[DOFERROR_NOT_SUPPORTED]            = DOFEXCEPTION_NOT_SUPPORTED;           /*  DOFERROR_NOT_SUPPORTED          ((DOFError)0x0002) */
    exceptionMap_OAP[DOFERROR_TYPE_MISMATCH]            = DOFEXCEPTION_TYPE_MISMATCH;           /*  DOFERROR_TYPE_MISMATCH          ((DOFError)0x0003) */
    exceptionMap_OAP[DOFERROR_INSUFFICIENT_RESOURCES]   = DOFEXCEPTION_INSUFFICIENT_RESOURCES;  /*  DOFERROR_INSUFFICIENT_RESOURCES ((DOFError)0x0004) */
    exceptionMap_OAP[DOFERROR_TERMINATED]               = DOFEXCEPTION_TERMINATED;              /*  DOFERROR_TERMINATED             ((DOFError)0x0005) */
    exceptionMap_OAP[DOFERROR_REJECTED]                 = DOFEXCEPTION_REJECTED;                /*  DOFERROR_REJECTED               ((DOFError)0x0006) */
    exceptionMap_OAP[DOFERROR_NOT_FOUND]                = DOFEXCEPTION_NOT_FOUND;               /*  DOFERROR_NOT_FOUND              ((DOFError)0x0007) */
    exceptionMap_OAP[DOFERROR_DUPLICATE]                = DOFEXCEPTION_DUPLICATE;               /*  DOFERROR_DUPLICATE              ((DOFError)0x0008) */
    exceptionMap_OAP[DOFERROR_TOO_MANY]                 = DOFEXCEPTION_TOO_MANY;                /*  DOFERROR_TOO_MANY               ((DOFError)0x0009) */
    exceptionMap_OAP[DOFERROR_APPLICATION_ERROR]        = DOFEXCEPTION_APPLICATION_ERROR;       /*  DOFERROR_APPLICATION_ERROR      ((DOFError)0x000a) */

    exceptionMap_API[DOFERROR_FAILED & 0xFF]                       = DOFEXCEPTION_FAILED;                      /*  DOFERROR_FAILED                         ((DOFError)0x8000) */
    exceptionMap_API[DOFERROR_TIMEOUT & 0xFF]                      = DOFEXCEPTION_TIMEOUT;                     /*  DOFERROR_TIMEOUT                        ((DOFError)0x8001) */
    exceptionMap_API[DOFERROR_NOT_READABLE & 0xFF]                 = DOFEXCEPTION_NOT_READABLE;                /*  DOFERROR_NOT_READABLE                   ((DOFError)0x8002) */
    exceptionMap_API[DOFERROR_NOT_WRITABLE & 0xFF]                 = DOFEXCEPTION_NOT_WRITABLE;                /*  DOFERROR_NOT_WRITABLE                   ((DOFError)0x8003) */
    exceptionMap_API[DOFERROR_RESULT_LOST & 0xFF]                  = DOFEXCEPTION_RESULT_LOST;                 /*  DOFERROR_RESULT_LOST                    ((DOFError)0x8004) */
    exceptionMap_API[DOFERROR_PROTOCOL_NEGOTIATION_FAILED & 0xFF]  = DOFEXCEPTION_PROTOCOL_NEGOTIATION_FAILED; /*  DOFERROR_PROTOCOL_NEGOTIATION_FAILED    ((DOFError)0x8005) */
    exceptionMap_API[DOFERROR_TRANSPORT_FAILED & 0xFF]             = DOFEXCEPTION_TRANSPORT_FAILED;            /*  DOFERROR_TRANSPORT_FAILED               ((DOFError)0x8006) */
    exceptionMap_API[DOFERROR_IO_FAILED & 0xFF]                    = DOFEXCEPTION_IO_FAILED;                   /*  DOFERROR_IO_FAILED                      ((DOFError)0x8007) */
    exceptionMap_API[DOFERROR_AUTHENTICATION_FAILED & 0xFF]        = DOFEXCEPTION_AUTHENTICATION_FAILED;       /*  DOFERROR_AUTHENTICATION_FAILED          ((DOFError)0x8008) */
    exceptionMap_API[DOFERROR_ACK_TIMEOUT & 0xFF]                  = DOFEXCEPTION_ACK_TIMEOUT;                 /*  DOFERROR_ACK_TIMEOUT                    ((DOFError)0x8009) */
    exceptionMap_API[DOFERROR_STALE_CREDENTIAL & 0xFF]             = DOFEXCEPTION_STALE_CREDENTIAL;            /*  DOFERROR_STALE_CREDENTIAL               ((DOFError)0x800a) */

    exceptionMap_TRP[0]                                         = DOFEXCEPTION_FAILED;
    exceptionMap_TRP[1]                                         = DOFEXCEPTION_FAILED;
    exceptionMap_TRP[DOFERROR_ACCESS_DENIED & 0xFF]             = DOFEXCEPTION_ACCESS_DENIED;           /*  DOFERROR_ACCESS_DENIED                  ((DOFError)0x9002) */
    exceptionMap_TRP[DOFERROR_UNKNOWN_INITIATOR & 0xFF]         = DOFEXCEPTION_UNKNOWN_INITIATOR;       /*  DOFERROR_UNKNOWN_INITIATOR              ((DOFError)0x9003) */
    exceptionMap_TRP[DOFERROR_UNKNOWN_RESPONDER & 0xFF]         = DOFEXCEPTION_UNKNOWN_RESPONDER;       /*  DOFERROR_UNKNOWN_RESPONDER              ((DOFError)0x9004) */
    exceptionMap_TRP[DOFERROR_UNKNOWN_DOMAIN & 0xFF]            = DOFEXCEPTION_UNKNOWN_DOMAIN;          /*  DOFERROR_UNKNOWN_DOMAIN                 ((DOFError)0x9005) */
    exceptionMap_TRP[DOFERROR_SERVICE_UNAVAILABLE & 0xFF]       = DOFEXCEPTION_SERVICE_UNAVAILABLE;     /*  DOFERROR_SERVICE_UNAVAILABLE            ((DOFError)0x9006) */
    exceptionMap_TRP[DOFERROR_BAD_MODE & 0xFF]                  = DOFEXCEPTION_BAD_MODE;                /*  DOFERROR_BAD_MODE                       ((DOFError)0x9007) */
    exceptionMap_TRP[8]                                         = DOFEXCEPTION_FAILED;
    exceptionMap_TRP[DOFERROR_NOT_IN_GROUP & 0xFF]              = DOFEXCEPTION_NOT_IN_GROUP;            /*  DOFERROR_NOT_IN_GROUP                   ((DOFError)0x9009) */
    exceptionMap_TRP[DOFERROR_UNSUPPORTED_INITIATOR & 0xFF]     = DOFEXCEPTION_UNSUPPORTED_INITIATOR;   /*  DOFERROR_UNSUPPORTED_INITIATOR          ((DOFError)0x900a) */
    exceptionMap_TRP[DOFERROR_UNSUPPORTED_RESPONDER & 0xFF]     = DOFEXCEPTION_UNSUPPORTED_RESPONDER;   /*  DOFERROR_UNSUPPORTED_RESPONDER          ((DOFError)0x900b) */
    exceptionMap_TRP[DOFERROR_INVALID_INITIATOR & 0xFF]         = DOFEXCEPTION_INVALID_INITIATOR;       /*  DOFERROR_INVALID_INITIATOR              ((DOFError)0x900c) */
    exceptionMap_TRP[DOFERROR_INVALID_RESPONDER & 0xFF]         = DOFEXCEPTION_INVALID_RESPONDER;       /*  DOFERROR_INVALID_RESPONDER              ((DOFError)0x900d) */
    exceptionMap_TRP[DOFERROR_RESOLUTION_FAILED & 0xFF]         = DOFEXCEPTION_RESOLUTION_FAILED;       /*  DOFERROR_RESOLUTION_FAILED              ((DOFError)0x900e) */
    exceptionMap_TRP[DOFERROR_VALIDATION_FAILED & 0xFF]         = DOFEXCEPTION_VALIDATION_FAILED;       /*  DOFERROR_VALIDATION_FAILED              ((DOFError)0x900f) */

    /* DOFEXCEPTION_ILLEGAL_ARGUMENT */
}

void DOFException_Shutdown( void )
{
    DOFException_Destroy( DOFEXCEPTION_NOT_SUPPORTED );
    DOFException_Destroy( DOFEXCEPTION_TYPE_MISMATCH );
    DOFException_Destroy( DOFEXCEPTION_INSUFFICIENT_RESOURCES );
    DOFException_Destroy( DOFEXCEPTION_TERMINATED );
    DOFException_Destroy( DOFEXCEPTION_REJECTED );
    DOFException_Destroy( DOFEXCEPTION_NOT_FOUND );
    DOFException_Destroy( DOFEXCEPTION_DUPLICATE );
    DOFException_Destroy( DOFEXCEPTION_TOO_MANY );
    DOFException_Destroy( DOFEXCEPTION_APPLICATION_ERROR );

    DOFException_Destroy( DOFEXCEPTION_FAILED );
    DOFException_Destroy( DOFEXCEPTION_TIMEOUT );
    DOFException_Destroy( DOFEXCEPTION_NOT_READABLE );
    DOFException_Destroy( DOFEXCEPTION_NOT_WRITABLE );
    DOFException_Destroy( DOFEXCEPTION_RESULT_LOST );
    DOFException_Destroy( DOFEXCEPTION_PROTOCOL_NEGOTIATION_FAILED );
    DOFException_Destroy( DOFEXCEPTION_TRANSPORT_FAILED );
    DOFException_Destroy( DOFEXCEPTION_IO_FAILED );
    DOFException_Destroy( DOFEXCEPTION_AUTHENTICATION_FAILED );
    DOFException_Destroy( DOFEXCEPTION_ACK_TIMEOUT );
    DOFException_Destroy( DOFEXCEPTION_STALE_CREDENTIAL );

    DOFException_Destroy( DOFEXCEPTION_ACCESS_DENIED );
    DOFException_Destroy( DOFEXCEPTION_UNKNOWN_INITIATOR );
    DOFException_Destroy( DOFEXCEPTION_UNKNOWN_RESPONDER );
    DOFException_Destroy( DOFEXCEPTION_UNKNOWN_DOMAIN );
    DOFException_Destroy( DOFEXCEPTION_SERVICE_UNAVAILABLE );
    DOFException_Destroy( DOFEXCEPTION_BAD_MODE );
    DOFException_Destroy( DOFEXCEPTION_NOT_IN_GROUP );
    DOFException_Destroy( DOFEXCEPTION_UNSUPPORTED_INITIATOR );
    DOFException_Destroy( DOFEXCEPTION_UNSUPPORTED_RESPONDER );
    DOFException_Destroy( DOFEXCEPTION_INVALID_INITIATOR );
    DOFException_Destroy( DOFEXCEPTION_INVALID_RESPONDER );
    DOFException_Destroy( DOFEXCEPTION_RESOLUTION_FAILED );
    DOFException_Destroy( DOFEXCEPTION_VALIDATION_FAILED );

    DOFException_Destroy( DOFEXCEPTION_ILLEGAL_ARGUMENT );
}

/* Map a DOFError to a DOFException.  The exception returned is not a retained reference.*/
DOFException DOFErrorException_GetFromError(DOFError error)
{
    DOFException exception = DOFEXCEPTION_FAILED;

    if ( (error <= DOFERROR_APPLICATION_ERROR) )
    {
        exception = exceptionMap_OAP[error];
    }
    else if ( (error >= DOFERROR_FAILED) && (error <= DOFERROR_STALE_CREDENTIAL) )
    {
        exception = exceptionMap_API[error & 0xFF];
    }
    else if ( (error >= DOFERROR_ACCESS_DENIED) && (error <= DOFERROR_VALIDATION_FAILED) )
    {
        exception = exceptionMap_TRP[error & 0xFF];
    }
    else if ( error == DOFERROR_ILLEGAL_ARGUMENT )
    {
        exception = DOFEXCEPTION_ILLEGAL_ARGUMENT;
    }
    else
    {
        exception = DOFEXCEPTION_FAILED;
    }

    return exception;
}

/*
**  "Public" Methods
*/


DOFException DOFErrorException_Create( DOFError error, const char * errorString )
{
    return _create( error, errorString, NULL, NULL );
}


DOFException DOFErrorException_Create_ProviderInfo( DOFError error, const char * errorString, DOFProviderInfo info )
{
    return _create( error, errorString, info, NULL );
}

DOFException DOFErrorException_Create_Data( DOFError error, const char * errorString, const void * data )
{
    return _create( error, errorString, NULL, data );
}

DOFException DOFProviderException_Create( DOFInterfaceException except, uint16 parameterCount, const DOFValue * parameters )
{
    if ( parameterCount < 256 )
        return DOFProviderException_Create_ProviderInfo( except, (uint8)parameterCount, parameters, NULL );
    return NULL;
}


DOFException DOFProviderException_Create_ProviderInfo( DOFInterfaceException except, uint8 parameterCount, const DOFValue * parameters, DOFProviderInfo info )
{
    if ( except && ( !parameterCount || parameters ) )
    {
        if ( DOFInterfaceItem_IsInputParameterListCompatible( (DOFItem)except, parameterCount, parameters ) )
        {
            DOFException e = DOFMem_Alloc( sizeof( DOFException_t ) + sizeof( DOFValue ) * parameterCount );
            if ( e )
            {
                uint8 i;
                PCRMem_Set(e,0,(sizeof( DOFException_t ) + sizeof( DOFValue ) * parameterCount));
                e->refCount = 1;
                e->type = DOFEXCEPTIONTYPE_PROVIDER;
                e->isStream = FALSE;
                e->valuesAreSeperateAlloc = FALSE;
                e->u.provider.except = except;
                e->info = DOFProviderInfo_Copy( info );
                e->u.provider.parameters = (DOFValue *) &e[1];
                for ( i = 0; i < parameterCount; i++ )
                {
                    if( parameters[i] )
                    {
                        e->u.provider.parameters[i] = DOFValue_Copy( parameters[i] );
                        if ( !e->u.provider.parameters[i] )
                        {
                            DOFException_Destroy( e );
                            e = NULL;
                            break;
                        }
                    }
                }
            }
            else
            {
                DOFLog_Message(  DOFLOGLEVEL_ERROR, "DOFProviderException: Failed to create provider exception." );
                return &outOfMemory;
            }

            return e;
        }
    }
    return NULL;
}

DOFException DOFProviderException_CreateStream_ProviderInfo(OALStream       valueStream,
                                                      DOFProviderInfo info,
                                                      DOFItemID       itemId)
{
    DOFException e;

    PCR_Assert(valueStream != NULL);
    e = DOFMem_Alloc(sizeof(DOFException_t));
    if(e != NULL)
    {
        e->refCount = 1;
        e->type = DOFEXCEPTIONTYPE_PROVIDER;
        e->u.stream.stream = valueStream;
        e->info = DOFProviderInfo_Copy(info);
        e->u.stream.itemId = itemId;
        OALStream_Retain(valueStream);
        e->isStream = TRUE;
        e->valuesAreSeperateAlloc = FALSE;
    }
    return e;
}

boolean DOFException_IsStream(DOFException e)
{
    PCR_Assert(e != NULL);
    return e->isStream;
}

OALStream DOFException_GetStream(DOFException e)
{
    PCR_Assert(e != NULL);
    PCR_Assert(e->isStream);

    return e->u.stream.stream;
}

DOFItemID DOFStreamException_GetItemId(DOFException e)
{
    PCR_Assert(e != NULL);
    PCR_Assert(e->isStream);

    return e->u.stream.itemId;
}

DOFException DOFException_Copy( const DOFException self )
{
    if (self != NULL && self != &outOfMemory )
    {
        PCRMutex_Lock(ref_count_mutex);
        if ( self->refCount >= 0xFFFFFFFF )
        {
            PCRMutex_Unlock(ref_count_mutex);
            return &outOfMemory;
        }
        self->refCount++;
        PCRMutex_Unlock(ref_count_mutex);
    }
    return self;
}

void DOFException_Destroy( DOFException self )
{
    if ( self && self != &outOfMemory )
    {
        boolean gone = FALSE;
        PCRMutex_Lock(ref_count_mutex);
        self->refCount--;
        gone = (self->refCount == 0);
        PCRMutex_Unlock(ref_count_mutex);
        if (gone)
        {
            if ( self->type == DOFEXCEPTIONTYPE_PROVIDER )
            {
                if(self->isStream)
                {
                    OALStream_Release(self->u.stream.stream, TRUE);
                }
                else
                {
                    uint8 i;
                    uint8 count = DOFInterfaceException_GetParameterCount( self->u.provider.except );

                    for ( i = 0; i < count; i++ )
                        DOFValue_Destroy( self->u.provider.parameters[i] );
                }
            }
            DOFProviderInfo_Destroy( self->info );
            if(self->valuesAreSeperateAlloc)
            {
                DOFMem_Free(self->u.provider.parameters);
            }
            DOFMem_Free( self );
        }
    }
}


DOFExceptionType DOFException_GetType( const DOFException self )
{
    if ( self )
        return self->type;
    return DOFEXCEPTIONTYPE_ERROR;
}


DOFError DOFErrorException_GetError( const DOFException self )
{
    if ( self && self->type == DOFEXCEPTIONTYPE_ERROR )
        return self->u.error.code;
    return DOFERROR_FAILED;
}


const char * DOFErrorException_GetMessage( const DOFException self )
{
    if ( self && self->type != DOFEXCEPTIONTYPE_PROVIDER )
        return self->u.error.errorString;
    return NULL;
}


const void * DOFErrorException_GetData( const DOFException self )
{
    if ( self && self->type != DOFEXCEPTIONTYPE_PROVIDER )
        return self->u.error.data;
    return NULL;
}


DOFInterfaceException DOFProviderException_GetInterfaceException( const DOFException self )
{
    if ( self && self->type == DOFEXCEPTIONTYPE_PROVIDER )
        return self->u.provider.except;
    return NULL;
}

DOFValue * DOFProviderException_GetParameters( const DOFException self )
{
    if ( self && self->type == DOFEXCEPTIONTYPE_PROVIDER )
        return self->u.provider.parameters;
    return NULL;
}


uint8 DOFProviderException_GetParametersCount( const DOFException self )
{
    if ( self && self->type == DOFEXCEPTIONTYPE_PROVIDER )
        return DOFInterfaceException_GetParameterCount( self->u.provider.except );
    return 0;
}


DOFProviderInfo DOFException_GetProviderInfo( const DOFException self )
{
    if ( self )
        return self->info;
    return NULL;
}


DOFException DOFException_Copy_ProviderInfo( DOFException self, DOFProviderInfo info )
{
    DOFException ret = NULL;
    if ( self )
    {
        if(DOFException_GetType(self) == DOFEXCEPTIONTYPE_ERROR)
        {
            ret = DOFErrorException_Create_ProviderInfo( DOFErrorException_GetError( self ), DOFErrorException_GetMessage( self ), info );
        }
        else
        {
            PCR_Assert( DOFException_GetType(self) == DOFEXCEPTIONTYPE_PROVIDER );
            ret = DOFProviderException_Create_ProviderInfo( DOFProviderException_GetInterfaceException( self ), DOFProviderException_GetParametersCount( self ), DOFProviderException_GetParameters( self ), info);
        }
    }
    return ret;
}


/**
@}
**/
