/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/filter.h>
#include <internal/filter_internal.h>
#include <internal/permissionsetinternal.h>

#define FILTER_PERMISSIONS_ERROR    0x01

typedef struct DOFOperationFilterBuilder_t
{
    DOFPermissionSet allowedByPerms;
    uint32 refCount;
    boolean permsWorked;
} DOFOperationFilterBuilder_t;

extern PCRMutex ref_count_mutex;

static DOFPermissionSet DOFOperationFilter_CreateFullPermissions(void)
{
    DOFPermissionSetBuilder builder = DOFPermissionSetBuilder_Create();

    if(builder != NULL)
    {
        DOFPermission perm = DOFPermissionBinding_Create(DOFPERMISSIONBINDINGACTION_ALL, NULL, NULL);

        if(perm != NULL)
        {
            DOFPermissionSetBuilder_AddPermission(builder, perm);
            DOFPermission_Destroy(perm);
            return DOFPermissionSetBuilder_BuildAndDestroy(builder);
        }
        DOFPermissionSetBuilder_Destroy(builder);
    }
    return NULL;
}

DOFOperationFilter DOFOperationFilter_CreateNullFilter(void)
{
    DOFOperationFilterBuilder ret = (DOFOperationFilterBuilder) DOFMem_Alloc(sizeof(DOFOperationFilterBuilder_t));

    if(ret != NULL)
    {
        ret->allowedByPerms = DOFOperationFilter_CreateFullPermissions();
        if(ret->allowedByPerms != NULL)
        {
            ret->refCount = 1;
            ret->permsWorked = TRUE;
            return (DOFOperationFilter) ret;
        }
        DOFMem_Free(ret);
    }
    return NULL;
}

DOFOperationFilterBuilder DOFOperationFilterBuilder_Create(void)
{
    DOFOperationFilterBuilder retVal = (DOFOperationFilterBuilder) DOFMem_Alloc(sizeof(DOFOperationFilterBuilder_t));

    if(retVal != NULL)
    {
        retVal->allowedByPerms = DOFPermissionSet_DeepCopy(DOFPermissionSet_GetEmptyPermissionSet());
        if(retVal->allowedByPerms != NULL)
        {
            retVal->refCount = 1;
            retVal->permsWorked = FALSE;
            return retVal;
        }
        DOFMem_Free(retVal);
    }
    return NULL;
}

DOFOperationFilterBuilder DOFOperationFilterBuilder_Create_Filter(DOFOperationFilter self)
{
    return DOFOperationFilterBuilder_Copy((DOFOperationFilterBuilder) self);
}

boolean DOFOperationFilterBuilder_AllowByPermissions(DOFOperationFilterBuilder self,
                                                     DOFPermissionSet          filter)
{
    boolean retVal = FALSE;

    if((self != NULL) && (filter != NULL))
    {
        retVal = DOFPermissionSet_AddPermissionsFromSet(self->allowedByPerms, filter);
        self->permsWorked |= retVal;
    }
    return retVal;
}

DOFOperationFilter DOFOperationFilterBuilder_BuildAndDestroy(DOFOperationFilterBuilder self)
{
    if(self != NULL)
    {
        if(self->permsWorked)
        {
            return (DOFOperationFilter) self;
        }
        DOFOperationFilterBuilder_Destroy(self);
    }
    return NULL;
}

DOFOperationFilter DOFOperationFilterBuilder_Build(DOFOperationFilterBuilder self)
{
    if(self != NULL)
    {
        if(self->permsWorked)
        {
            return (DOFOperationFilter) DOFOperationFilterBuilder_Copy(self);
        }
    }
    return NULL;
}

DOFOperationFilterBuilder DOFOperationFilterBuilder_Copy(DOFOperationFilterBuilder self)
{
    DOFOperationFilterBuilder retVal = (DOFOperationFilterBuilder) DOFMem_Alloc(sizeof(DOFOperationFilterBuilder_t));

    if(retVal != NULL)
    {
        retVal->allowedByPerms = DOFPermissionSet_DeepCopy(self->allowedByPerms);
        if(retVal->allowedByPerms != NULL)
        {
            retVal->refCount = 1;
            retVal->permsWorked = self->permsWorked;
            return retVal;
        }
        DOFMem_Free(retVal);
    }
    return NULL;
}

void DOFOperationFilterBuilder_Destroy(DOFOperationFilterBuilder self)
{
    if(self != NULL)
    {
        boolean freeOk;

        PCRMutex_Lock(ref_count_mutex);
        freeOk = (--(self->refCount) == 0);
        PCRMutex_Unlock(ref_count_mutex);
        if(freeOk)
        {
            DOFPermissionSet_Destroy(self->allowedByPerms);
            DOFMem_Free(self);
        }
    }
}

boolean DOFOperationFilter_IsAllowed(DOFOperationFilter self,
                                     DOFPermission      perm)
{
    PCR_Assert(perm != NULL);
    return DOFPermissionSet_Contains(((DOFOperationFilterBuilder) self)->allowedByPerms, perm);
}

DOFPermissionSet DOFOperationFilter_GetAllowByPermissions(DOFOperationFilter self)
{
    return (self != NULL) ? ((DOFOperationFilterBuilder) self)->allowedByPerms
                          : NULL;
}

DOFOperationFilter DOFOperationFilter_Copy(DOFOperationFilter self)
{
    if(self != NULL)
    {
        DOFOperationFilterBuilder selfx = (DOFOperationFilterBuilder) self;

        uint32 old;

        PCRMutex_Lock(ref_count_mutex);
        old = (selfx->refCount)++;
        if(old > (selfx->refCount))
        {
            selfx->refCount = old;
            PCRMutex_Unlock(ref_count_mutex);
            return NULL;
        }
        PCRMutex_Unlock(ref_count_mutex);
    }
    return self;
}

void DOFOperationFilter_Destroy(DOFOperationFilter self)
{
    DOFOperationFilterBuilder_Destroy((DOFOperationFilterBuilder) self);
}
