/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/hashtable.h>
#include <internal/dofinternal.h>

/*
**  "Private" type and method declarations
*/

typedef struct OALHashtable_t {
    OALAVLMultiTree parent;
    uint8 num_keys;
} OALHashtable_t;

typedef struct key_value_pair {
    void *key;
    OALHashKeySize_t key_size;
} key_value_pair;

typedef struct hash_item_t {
    OALHashItem_t           item;
    uint32                  hash; /* saves time */
    key_value_pair          pairs[1];
} hash_item_t;

typedef struct hash_search_item_t {
    OALHashKey_t *keys;
    OALHashKeySize_t *key_sizes;

    OALHashItem_t           item;
    uint32                  hash; /* saves time */
} hash_search_item_t;

int32 multi_compare(hash_search_item_t *key, hash_item_t *item, uint8 num_keys);

/* Bob Jenkins one-at-a-time hash! with a slight modification */
static uint32 joaat_hash(OALHashKey_t *keys, OALHashKeySize_t *sizes, uint8 num_keys)
{
    uint32 hash = 0;
    uint16 i, size;
    uint8 j;
    const uint8 *key;

    for(j=0; j<num_keys; j++) {
        size = sizes[j];
        key = (const uint8*)keys[j];

        for (i = 0; i < size; i++) {
            hash += key[i];
            hash += (hash << 10);
            hash ^= (hash >> 6);
        }

        hash += (hash << 3);
        hash ^= (hash >> 11);
        hash += (hash << 15);
    }

    return hash;
}

int32 multi_compare(hash_search_item_t *key, hash_item_t *item, uint8 num_keys)
{
    uint8 i = 0;

    for(i=0; i<num_keys; ++i) {
        int32 val = 0;

        if(key->key_sizes[i]!=item->pairs[i].key_size) {
            if(key->key_sizes[i]>item->pairs[i].key_size) return 1;
            else return -1;
        }

        val = PCRMem_Compare(key->keys[i], item->pairs[i].key, key->key_sizes[i]);
        if(val) return val;
    }

    return 0;
}

static boolean hashtable_item_gt(hash_item_t *key, hash_item_t *item, OALHashtable context)
{
    return key->item > item->item;
}

static boolean hashtable_item_eq(hash_item_t *key, hash_item_t *item, OALHashtable context)
{
    return key->item == item->item;
}

static boolean hashtable_key_gt(hash_search_item_t *key, hash_item_t *item, OALHashtable context)
{
    if(key->hash==item->hash) return (multi_compare(key, item, context->num_keys)>0);
    else return key->hash>item->hash;
}

static boolean hashtable_key_eq(hash_search_item_t *key, hash_item_t *item, OALHashtable context)
{
    if(key->hash==item->hash) {
        if(multi_compare(key, item, context->num_keys)) return FALSE;
#if 0
        if(key->item) {
            if(key->item==item->item) return TRUE;
            return FALSE; /* not equal */
        }
#endif
        return TRUE;
    }
    return FALSE;
}

OALHashtable OALHashtable_Create(uint8 num_keys)
{
    OALHashtable h = DOFMem_Alloc(sizeof(OALHashtable_t));

    if(!h) {
        return NULL;
    }

    PCR_Assert(num_keys);

    h->parent = OALAVLMultiTree_Create((OALAVLMultiTree_CompareGreaterThan_fn_t)hashtable_key_gt, (OALAVLMultiTree_CompareEqual_fn_t)hashtable_key_eq, (OALAVLMultiTree_CompareGreaterThan_fn_t)hashtable_item_gt, (OALAVLMultiTree_CompareEqual_fn_t)hashtable_item_eq, h);

    if(!h->parent) {
        DOFMem_Free(h);
        return NULL;
    }

    h->num_keys = num_keys;

    return h;
}


typedef struct traverse_t {
    boolean (*fn)(void *, void*);
    void *arg;
    uint8 num_keys;
} traverse_t;

static void remove_all(OALHashItem_t node, void* arg)
{
    if(((traverse_t*)arg)->fn) ((OALHashtable_RemoveAll_fn_t)((traverse_t*)arg)->fn)(((hash_item_t*)node)->item, ((traverse_t*)arg)->arg);
    DOFMem_Free(((hash_item_t*)node));
}

void OALHashtable_Destroy(OALHashtable hashtable, OALHashtable_RemoveAll_fn_t fn, void *arg)
{
    if ( hashtable )
    {
        traverse_t t;

        t.fn = (boolean(*)(void*,void*))fn;
        t.arg = arg;
        t.num_keys = hashtable->num_keys;

        OALAVLMultiTree_Destroy(hashtable->parent, remove_all, &t);
        DOFMem_Free(hashtable);
    }
}

static void hash_item_key_init(hash_search_item_t *item, OALHashKey_t *keys, OALHashKeySize_t *key_sizes, uint8 num_keys)
{
    item->keys = keys;
    item->key_sizes = key_sizes;
    item->hash = joaat_hash(keys, key_sizes, num_keys);
    item->item = 0;
}

OALHashtable_InsertionStatus OALHashtable_Insert(OALHashtable hashtable, OALHashItem_t item, OALHashKey_t *keys, OALHashKeySize_t *key_sizes)
{
    hash_search_item_t hsit;
    hash_item_t *hit;
    uint8 c = 0;
    uint16 total_size = 0;
    uint32 offset;
    OALAVLTree_InsertionStatus status;

    PCR_Assert(hashtable);
    for(c=0; c<hashtable->num_keys; ++c) {
        total_size += key_sizes[c];
    }

    PCR_Assert(total_size>0);

    offset = offsetof(hash_item_t, pairs);
    offset += (sizeof(key_value_pair)*hashtable->num_keys);
    hit = DOFMem_Alloc(offset + total_size);

    if(!hit) {
        return OALHashtable_InsertFailure;
    }

    hash_item_key_init(&hsit, keys, key_sizes, hashtable->num_keys);

    for(c=0; c<hashtable->num_keys; ++c) {
        hit->pairs[c].key = (uint8 *)hit + offset;
        offset += (hit->pairs[c].key_size = key_sizes[c]);

        PCRMem_Copy((void*)hit->pairs[c].key, keys[c], key_sizes[c]); /* copy the key */
    }

    hit->hash = joaat_hash(keys, key_sizes, hashtable->num_keys);
    hit->item = item;

    if((status=OALAVLMultiTree_Insert(hashtable->parent, &hsit, hit)) != OALAVLTree_InsertSuccess) {
        DOFMem_Free(hit);
    }

    return (OALHashtable_InsertionStatus)status; /* These two status enums match. */
}

static const hash_item_t *hashtable_find(OALHashtable hashtable, OALHashKey_t *keys, OALHashKeySize_t *key_sizes)
{
    hash_search_item_t item;
    const hash_item_t *found;

    hash_item_key_init(&item, keys, key_sizes, hashtable->num_keys);

    found = OALAVLMultiTree_Retrieve(hashtable->parent, &item);

    return found;
}

boolean OALHashtable_Remove_First(OALHashtable hashtable, OALHashKey_t *keys, OALHashKeySize_t *key_sizes)
{
    hash_search_item_t item;
    const hash_item_t *found;

    PCR_Assert(hashtable);
    hash_item_key_init(&item, keys, key_sizes, hashtable->num_keys);

    found = OALAVLMultiTree_Retrieve(hashtable->parent, &item);

    if(!found) {
        return FALSE; /* nothing to do */
    }

    {
        boolean ret = OALAVLMultiTree_Delete(hashtable->parent, &item, found);
        PCR_Assert(ret); /* at this point it's an internal error if nothing is found */
    }

    DOFMem_Free((hash_item_t*)found);

    return TRUE;
}

typedef struct find_remove_t {
    OALHashItem_t item; /* the user item we want to find in the given key set */
    hash_item_t *found; /* the found item */
} find_remove_t;

static boolean _find_and_remove(OALAVLMultiTreeItem_t item, void* fr)
{
    if(((hash_item_t*)item)->item==((find_remove_t*)fr)->item) {
        ((find_remove_t*)fr)->found = ((hash_item_t*)item);
        return TRUE;
    }
    return FALSE; /* keep searching */
}

boolean OALHashtable_Remove(OALHashtable hashtable, OALHashItem_t item, OALHashKey_t *keys, OALHashKeySize_t *key_sizes)
{
    hash_search_item_t hit;
    find_remove_t fr;

    PCR_Assert(hashtable);
    hash_item_key_init(&hit, keys, key_sizes, hashtable->num_keys);

    fr.item = item;
    fr.found = 0;

    OALAVLMultiTree_Traverse(hashtable->parent, &hit, _find_and_remove, &fr);

    if(!fr.found) return FALSE; /* nothing to do */

    {
        boolean ret = OALAVLMultiTree_Delete(hashtable->parent, &hit, fr.found);
        PCR_Assert(ret); /* internal error if not found here */
    }

    DOFMem_Free(fr.found);

    return TRUE;
}

OALHashItem_Ret_t OALHashtable_Retrieve(OALHashtable hashtable, OALHashKey_t *keys, OALHashKeySize_t *key_sizes)
{
    PCR_Assert(hashtable);

    {
        const hash_item_t *hit = (hashtable_find(hashtable, keys, key_sizes));
        return hit ? hit->item : 0;
    }
}

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t _traverse_all(OALAVLMultiTreeItem_t item, void* t, boolean new_set)
{
    return ((OALAVLMultiTree_Traverse_All_fn_t)((traverse_t*)t)->fn)(((hash_item_t*)item)->item, ((traverse_t*)t)->arg, new_set);
}

void OALHashtable_Traverse_All(OALHashtable hashtable, OALAVLMultiTree_Traverse_All_fn_t traverse_fn, void *arg)
{
    traverse_t t;

    PCR_Assert(hashtable);
    PCR_Assert(traverse_fn);

    t.arg = arg;
    t.fn = (boolean(*)(void*, void*))traverse_fn;

    OALAVLMultiTree_Traverse_All(hashtable->parent, _traverse_all, &t);
}

static boolean _traverse(OALAVLMultiTreeItem_t item, void* t)
{
    return ((OALAVLMultiTree_Traversal_fn_t)((traverse_t*)t)->fn)(((OALAVLMultiTreeItem_t)((hash_item_t*)item)->item), ((traverse_t*)t)->arg);
}

void OALHashtable_Traverse(OALHashtable hashtable, OALHashKey_t *keys, OALHashKeySize_t *key_sizes, OALHashtable_Traversal_fn_t traverse_fn, void *arg)
{
    hash_search_item_t item;
    traverse_t t;

    PCR_Assert(hashtable);
    PCR_Assert(traverse_fn);

    hash_item_key_init(&item, keys, key_sizes, hashtable->num_keys);

    t.arg = arg;
    t.fn = (boolean(*)(void*, void*))traverse_fn;

    OALAVLMultiTree_Traverse(hashtable->parent, &item, _traverse, &t);
}

uint32 OALHashtable_GetNumItems(OALHashtable hashtable)
{
    PCR_Assert(hashtable);

    return OALAVLMultiTree_GetNumItems(hashtable->parent);
}