/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/interface.h>
#include <dof/interfaceid.h>
#include <dof/type.h>
#include <internal/type.h>
#include <internal/oaltypesinternal.h>
#include <internal/dofinternal.h>
#include <internal/iteminternal.h>
#include <internal/interfaceinternal.h>
#include <internal/marshal.h>
#include "interfaceGetItem.h"


extern PCRMutex ref_count_mutex;


/** @internal @addtogroup DOFInterface
    DOF Interface
@{
**/


/*
**  "Private" Type and Method Declarations
*/


static boolean readItems( DOFInterface iface, uint8 itemType, uint32 * pSize, const uint8 ** ppParse )
{
    const uint8 * pParse;
    uint32        lenMarshaledData;
    uint32        size;
    uint16        ui16;
    uint16        count;
    int           lastID = -1;

    PCR_Assert( iface && pSize && ppParse );
    pParse = *ppParse;
    PCR_Assert( pParse );
    size = *pSize;
    /* Count of Items: 2-byte compressed */
    lenMarshaledData = size;
    count = (uint16)OALMarshal_UncompressValue( 2, &lenMarshaledData, pParse );
    if ( count == (uint16) OALMARSHAL_INVALID_VALUE )
    {
        DOFLog_Message( DOFLOGLEVEL_TRACE , "DOFInterface: Parse error." );
        return FALSE;
    }
    size -= lenMarshaledData;
    pParse += lenMarshaledData;
    for (ui16 = 0; ui16 < count; ++ui16)
    {
        int                paramListCount;
        struct DOFItem_t * item = DOFMem_Alloc( sizeof( struct DOFItem_t ) );

        if (!item)
        {
            DOFLog_Message( DOFLOGLEVEL_TRACE, "DOFInterface: Allocation failed during interface construction." );
            return FALSE;
        }
        PCRMem_Set( item, 0, sizeof( struct DOFItem_t ) );
        item->iface = iface;
        item->itemType = itemType;
        switch (itemType)
        {
            default:
                paramListCount = 0;
                break;
            case DOFITEM_EVENT:
            case DOFITEM_EXCEPTION:
                paramListCount = 1;
                break;
            case DOFITEM_METHOD:
                paramListCount = 2;
                break;
        }
        /* All Items: Item ID (2-byte compressed) */
        lenMarshaledData = size;
        item->itemID = (DOFItemID) OALMarshal_UncompressValue( 2, &lenMarshaledData, pParse );
        if (item->itemID == (DOFItemID) OALMARSHAL_INVALID_VALUE || lastID >= (int) item->itemID)
            goto error;
        lastID = item->itemID;
        size -= lenMarshaledData;
        pParse += lenMarshaledData;

        if ( paramListCount > 0 )
        {
            int   paramList;

            /* Method / Event / Exception: Parameter List(s) */
            for ( paramList = 0; paramList < paramListCount; ++paramList )
            {
                uint8 param;
                if ( size < 1 )
                    goto error;
                size -= 1;
                item->typeCount[ paramList ] = *(pParse++);
                item->types[ paramList ] = DOFTypeListAlloc_Create( item->typeCount[ paramList ] );
                if ( !item->types[ paramList ] )
                {
                    DOFLog_Message( DOFLOGLEVEL_TRACE, "DOFInterface: Allocation failed during interface construction." );
                    goto error;
                }
                for ( param = 0; param < item->typeCount[ paramList ]; ++param )
                {
                    lenMarshaledData = size;
                    item->types[ paramList ][ param ] =
                         DOFType_Copy(DOFTypeContext_UnmarshalTypeRef( iface->typeContext, &lenMarshaledData, pParse ));
                    if (!item->types[ paramList ][ param ])
                        goto error;
                    size -= lenMarshaledData;
                    pParse += lenMarshaledData;
                }
            }
        }
        else
        {
            PCR_Assert( itemType == DOFITEM_PROPERTY );
            /* Property: Access Byte */
            if (size < 1)
                goto error;
            --size;
            item->access = *(pParse++) & 0x03;
            /* Property: Single Type Field */
            lenMarshaledData = size;
            item->types[0] = (DOFType *)DOFType_Copy( DOFTypeContext_UnmarshalTypeRef( iface->typeContext,
                                                                                       &lenMarshaledData, pParse ) );
            if (!item->types[0])
                goto error;
            size -= lenMarshaledData;
            pParse += lenMarshaledData;
        }

        if (OALAVLTree_Insert( iface->itemTree, (uint8 *)0 + item->itemID, item, NULL ) != OALAVLTree_InsertSuccess)
        {
        error:
            DOFLog_Message( DOFLOGLEVEL_TRACE, "DOFInterface: Parse error or allocation failed during interface construction." );
            DOFItem_Destroy( item );
            return FALSE;
        }
    }
    *pSize = size;
    *ppParse = pParse;
    return TRUE;
}



/*
**  "Public" Methods
*/



static boolean item_gt_fn( OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context )
{
    return (DOFItemID)key > ((DOFItem)item)->itemID;
}

static boolean item_eq_fn( OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context )
{
    return (DOFItemID)key == ((DOFItem)item)->itemID;
}

DOFInterface DOFInterface_Create( const DOFInterfaceID iid, uint32 dataSize, const uint8 * data )
{
    DOFInterface_t * iface;

    if (!iid || !data || dataSize < 5)
        return NULL;

    iface = DOFMem_Alloc( offsetof( DOFInterface_t, data ) + dataSize );
    if (iface)
    {
        /* Fill the interface descriptor. */
        PCRMem_Set( iface, 0, offsetof( DOFInterface_t, data ) );
        iface->refCount = 1;
        iface->iid = DOFInterfaceID_Copy( iid );
        if (iface->iid)
        {
            iface->itemTree = OALAVLTree_Create( item_gt_fn, item_eq_fn, NULL );
            if (iface->itemTree)
            {
                iface->dataSize = dataSize;
                PCRMem_Copy( iface->data, data, dataSize );
                iface->typeContext = DOFTypeContext_Create_Unmarshal( &dataSize, data );
                if (iface->typeContext)
                {
                    data += dataSize;
                    dataSize = iface->dataSize - dataSize;

                    /* Read in the items. */
                    if (readItems( iface, DOFITEM_PROPERTY, &dataSize, &data ) &&
                        readItems( iface, DOFITEM_EVENT, &dataSize, &data ) &&
                        readItems( iface, DOFITEM_METHOD, &dataSize, &data ) &&
                        readItems( iface, DOFITEM_EXCEPTION, &dataSize, &data ) &&
                        dataSize == 0)
                    {
                        /* No need to keep the context any longer. */
                        DOFTypeContext_Destroy( iface->typeContext );
                        iface->typeContext = NULL;
                        return iface;
                    }
                    else
                        DOFLog_Message(  DOFLOGLEVEL_ERROR, "DOFInterface_Create: Failed to parse interface definition." );
                }
                else
                    DOFLog_Message(  DOFLOGLEVEL_ERROR, "DOFInterface_Create: Failed to parse interface definition." );
            }
            else
                DOFLog_Message(  DOFLOGLEVEL_ERROR, "DOFInterface_Create: Allocation failure." );
        }
        else
            DOFLog_Message(  DOFLOGLEVEL_ERROR, "DOFInterface_Create: Allocation failure." );
        DOFInterface_Destroy( iface );
    }
    else
        DOFLog_Message(  DOFLOGLEVEL_ERROR, "DOFInterface_Create: Allocation failure." );

    return( NULL );
}

const uint8 * DOFInterface_GetBytes( const DOFInterface self )
{
    return self->data;
}

uint32 DOFInterface_GetBytesSize( const DOFInterface self )
{
    return self->dataSize;
}


/** @internal
    Retain a reference to a previously created DOFInterface table, or interface descriptor.

    @param  theInterface The interface descriptor to retain. This must not be NULL.
    @return TRUE for success, FALSE when @theInterface is NULL or resources are insufficient to retain the reference.
**/
boolean DOFInterface_Retain( DOFInterface_t * theInterface )
{
    boolean rval = TRUE;

    PCR_Assert( theInterface );
    PCRMutex_Lock( ref_count_mutex );
    if ((++theInterface->refCount) == 0)
    {
        --theInterface->refCount;
        rval = FALSE;
    }
    PCRMutex_Unlock( ref_count_mutex );

    return rval;
}

static void treeItemDestroy(OALAVLTreeItem_t item, void* ctx)
{
    DOFItem_Destroy((struct DOFItem_t*)item);
}

/** @internal
    De-allocate resources of a previously created DOFInterface table, or interface descriptor.

    @note Where the given @a destroyInterface is NULL, this call does nothing and safely returns.

    @param  destroyInterface The interface descriptor to de-allocate.
**/
void DOFInterface_Destroy( DOFInterface_t * destroyInterface )
{
    if (destroyInterface)
    {
        uint32 testCount;

        PCRMutex_Lock( ref_count_mutex );

        PCR_Assert( destroyInterface->refCount );
        testCount = --destroyInterface->refCount;
        PCRMutex_Unlock( ref_count_mutex );
        if (!testCount)
        {
            if (destroyInterface->itemTree)
                OALAVLTree_Destroy( destroyInterface->itemTree, treeItemDestroy, NULL );
            DOFTypeContext_Destroy( destroyInterface->typeContext );
            DOFInterfaceID_Destroy( destroyInterface->iid );
            DOFMem_Free( destroyInterface );
        }
    }
}


/** @internal
    Retrieve the Interface ID for the given interface descriptor.

    @param  iface The interface descriptor to query.
**/
DOFInterfaceID DOFInterface_GetInterfaceID( DOFInterface iface )
{
    if (iface)
        return iface->iid;

    return NULL;
}


/**
@}
**/
