/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/interfaceid.h>
#include <internal/interfaceidinternal.h>
#include <internal/marshal.h>
#include "interfaceidprivate.h"
#include <internal/dofinternal.h>


/** @addtogroup DOFInterfaceID
    DOF Interface Identifier
@{
**/


extern PCRMutex ref_count_mutex;


static boolean InterfaceID_BytesSizeValid( uint8 iidBytesSize )
{
    return (iidBytesSize == 2 || iidBytesSize == 3 || iidBytesSize == 5) ? TRUE : FALSE;
}


/* Return the size of the InterfaceID based on the length bytes. */
static uint8 InterfaceIDBytes_GetSize( const uint8 * pIIDData )
{
    static const uint8 sizes[] = { 0 /* 0 length invalid */, 2, 3, 5 };

    return pIIDData ? sizes[*pIIDData & 0x03] : 0;
}


/** @internal
    Determine if IID representation is in its most compact form.

    @param iidDataSize Length in bytes of the IID wire represenation held at @a pIIDData. Must be 2, 4, or 5.
    @param pIIDData Pointer to IID wire representation.
    @return TRUE for IID representation is in most compact form, FALSE otherwise
*/
static boolean DataFieldMostCompact( uint8 iidDataSize, const uint8 * pIIDData )
{
    /* Illegal zero front padding in pIIDData,     for iidDataSize:
                                     --------      ---------------
                                     none               1
                                     0x00               2
                                     0x0000             4 */
    switch (iidDataSize)
    {
        case 4:
            if (pIIDData[1])
                return TRUE;
            /* 2nd byte 0, so fall through to test leading byte. */
        case 2:
            if (pIIDData[0] == 0)
                return FALSE;
            break;
        case 1:
            break;
        default:
            PCR_Assert( FALSE );
            break;
    }

    return TRUE;
}


/*
**  "Public" Methods
*/


DOFInterfaceID DOFInterfaceID_Create( uint8 registry, uint32 identifier )
{
    if (registry <= DOFINTERFACEID_MAX_REGISTRY_VALUE)
    {
        DOFInterfaceID newInterfaceID = (DOFInterfaceID)DOFMem_Alloc( sizeof( DOFInterfaceID_t ) );

        if (newInterfaceID)
        {
            newInterfaceID->refCount = 1;
            newInterfaceID->len = 1;
            newInterfaceID->identifier = identifier;
            if (identifier <= 0xFF)
            {
                newInterfaceID->len = 2;
                newInterfaceID->iid[0] = (registry << 2) | 0x01;
            }
            else if (identifier <= 0xFFFF)
            {
                newInterfaceID->len = 3;
                newInterfaceID->iid[0] = (registry << 2) | 0x02;
            }
            else
            {
                newInterfaceID->len = 5;
                newInterfaceID->iid[0] = (registry << 2) | 0x03;
            }
            identifier = PCRSwap_Net32( identifier );
            PCRMem_Copy( &(newInterfaceID->iid[1]), ((uint8*)&identifier) + (5 - newInterfaceID->len),
                         newInterfaceID->len - 1 );
            return newInterfaceID;
        }
    }

    return NULL;                                                                 /* Error. */
}


DOFInterfaceID DOFInterfaceID_Create_Bytes( uint32 bufferSize, const uint8 * pIIDBuffer )
{
    if (pIIDBuffer && InterfaceID_BytesSizeValid( bufferSize ) &&
        (bufferSize == InterfaceIDBytes_GetSize( pIIDBuffer )) &&
        DataFieldMostCompact( (bufferSize - 1), (pIIDBuffer + 1) ))
    {
        DOFInterfaceID newInterfaceID = (DOFInterfaceID)DOFMem_Alloc( sizeof( DOFInterfaceID_t ) );

        if (newInterfaceID)
        {
            uint32 identifier = 0;
            newInterfaceID->refCount = 1;
            newInterfaceID->len = (uint8)bufferSize;
            PCRMem_Copy( &(newInterfaceID->iid[0]), pIIDBuffer, bufferSize );
            PCRMem_Copy( ((uint8*)&identifier) + ( 5 - bufferSize ), pIIDBuffer + 1, bufferSize - 1 );
            newInterfaceID->identifier = PCRSwap_Net32( identifier );
            return newInterfaceID;
        }
    }

    return NULL;
}


const uint8 * DOFInterfaceID_GetBytes( const DOFInterfaceID self )
{
    return self ? self->iid : NULL;
}


uint8 DOFInterfaceID_GetBytesSize( const DOFInterfaceID self )
{
    return self ? self->len : 0;

}


uint32 DOFInterfaceID_GetIdentifier( const DOFInterfaceID self )
{
    return self->identifier;
}


uint8 DOFInterfaceID_GetRegistry( const DOFInterfaceID self )
{
    return self ? (self->iid[0] >> 2) : 0;
}


void DOFInterfaceID_Destroy( DOFInterfaceID self )
{
    if (self)
    {
        boolean freeIt = FALSE;

        PCRMutex_Lock( ref_count_mutex );
        PCR_Assert( self->refCount );
        if (!--self->refCount)
            freeIt = TRUE;
        PCRMutex_Unlock( ref_count_mutex );
        if (freeIt)
            DOFMem_Free( self );
    }
}


/*
**  "Private" Methods
*/


/** @internal
    Unmarshal an Interface ID from a buffer.
    @mustfree{DOFInterfaceID_Destroy}
    @param  length  [IN] The total length of the buffer; [OUT] the length of the data unmarshaled.
    @param  buffer  The buffer in which the @ref DOFInterfaceID is marshaled. This should not be NULL.
    @return The new @ref DOFInterfaceID created.
    @exception NULL is returned if
        - @a buffer is NULL, or
        - the unmarshaled Interface ID would not be legal, or
        - @length is too small to correctly unmarshal the Interface ID, or
        - there were insufficient resources to create the Interface ID.
**/
DOFInterfaceID DOFInterfaceID_Create_Unmarshal( uint32 * length, const uint8 * buffer )
{
    DOFInterfaceID retID = NULL;

    if (*length > 1)
    {
        /* We must have at least two bytes to have a valid interfaceID. */
        uint8 len = InterfaceIDBytes_GetSize( buffer );

        if (*length < len)
            return NULL;

        retID = DOFInterfaceID_Create_Bytes( len, buffer );
        if (retID)
            *length = len;
        else
            *length = 0; /* Here if buffer NULL or IID not legal or allocate failure. */
    }

    return retID;
}


/** @internal
    Unmarshal an Interface ID from a stream.
    @mustfree{DOFInterfaceID_Destroy}
    @param  stream The stream in which the @ref DOFInterfaceID is marshaled. This should not be NULL and should describe
                   a legal @ref DOFInterfaceID.
    @return The new @ref DOFInterfaceID created.
    @exception NULL is returned if @a stream is NULL or it describes an illegal @ref DOFInterfaceID.
**/
DOFInterfaceID DOFInterfaceID_Unmarshal( OALStream stream )
{
    if (stream)
    {
        DOFInterfaceID retID   = NULL;
        uint32         len     = OALStream_GetDataSize( stream );
        const uint8 *  readPtr = OALStream_GetReadPtr( stream );

        retID = DOFInterfaceID_Create_Unmarshal( &len, readPtr );
        OALStream_AdjustReadOffset( stream, len );
        return retID;
    }

    return FALSE;
}


/** @internal
    Marshal an Interface ID into a buffer.
    @param  self    The @ref DOFInterfaceID to be marshaled.  Must not be NULL.  Must have a valid length.
    @param  length  The length of the buffer in which to marshal the Interface ID.
    @param  buffer  The buffer in which to marshal the Interface ID.
    @return The length of the marshaled Interface ID.
    @exception 0 is returned if the marshalled Interface ID cannot fit in the buffer.
**/
uint8 DOFInterfaceID_Marshal( const DOFInterfaceID self, uint32 length, uint8 * buffer )
{
    uint8 marshalLength = DOFInterfaceID_GetBytesSize( self );

    PCR_Assert( InterfaceID_BytesSizeValid( marshalLength ) );
    if (marshalLength <= length)
    {
        PCRMem_Copy( buffer, DOFInterfaceID_GetBytes( self ), marshalLength );
        return marshalLength;
    }

    return 0;
}


/** @internal
    Marshal an Interface ID into a stream.
    @param  self    The @ref DOFInterfaceID to be marshaled.  Must not be NULL.  Must have a valid length.
    @param  stream  The stream in which to marshal the Interface ID.
    @return TRUE for success.
    @exception FALSE is returned if the marshalled Interface ID cannot be placed in the stream.
**/
boolean DOFInterfaceID_MarshalToStream( const DOFInterfaceID self, OALStream stream )
{
    uint32 writeLen = OALStream_GetFreeSize( stream );
    uint8* writePtr = OALStream_GetWritePtr( stream );

    writeLen = DOFInterfaceID_Marshal( self, writeLen, writePtr );

    if (writeLen != 0)
    {
        OALStream_AdjustWriteOffset( stream, writeLen );
        return TRUE;
    }

    return FALSE;
}


/**
@}
**/

