/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/interface.h>
#include <dof/interfaceid.h>
#include <dof/item.h>
#include <dof/value.h>
#include <internal/dofinternal.h>
#include <internal/interfaceinternal.h>
#include <internal/iteminternal.h>
#include <internal/stream.h>
#include <internal/type.h>
#include <internal/value.h>


/** @internal @addtogroup DOFItem
@{
**/


/*
**  "Private" Type and Method Declarations
*/


/*
**  "Public" Methods, internal
*/


DOFInterface DOFItem_GetInterface( const DOFItem item )
{
    if (item)
        return item->iface;

    return NULL;
}


DOFInterfaceID DOFItem_GetInterfaceID( const DOFItem item )
{
    DOFInterface pInterfaceDescriptor;

    PCR_Assert( item );
    pInterfaceDescriptor = DOFItem_GetInterface( item );
    if (pInterfaceDescriptor)
        return DOFInterface_GetInterfaceID( pInterfaceDescriptor );

    return NULL;
}


boolean DOFItem_RetainDesc( const DOFItem item )
{
    DOFInterface_t * pInterfaceDescriptor = DOFItem_GetInterface( item );

    return pInterfaceDescriptor ? DOFInterface_Retain( pInterfaceDescriptor ) : FALSE;
}


void DOFItem_ReleaseDesc( const DOFItem item )
{
    DOFInterface interface = DOFItem_GetInterface( item );
    DOFInterface_Destroy( interface );
}


void DOFItem_Destroy( struct DOFItem_t * item )
{
    if (item)
    {
        /* For rules, see paramListCount in readItems() of interface.c. */
        if (item->itemType == DOFITEM_PROPERTY)
            DOFType_Destroy( (DOFType)item->types[0] );
        /* The original naked code that follows here triggered code review. It is kept here, and may or may not be more
           efficient than more direct approaches. Comments and assert are added to demonstrate code is proper. */
        else
            DOFTypeListAlloc_Destroy( item->typeCount[0], item->types[0] ); /* Event, method, exception all have this.*/
        PCR_Assert( item->itemType == DOFITEM_METHOD || item->types[1] == NULL ); /* == NULL: next call does nothing. */
        DOFTypeListAlloc_Destroy( item->typeCount[1], item->types[1] );   /* Even for _METHOD, can sometime == NULL . */
        /* [More directly, can test type before uselessly calling last _Destroy(). Tradeoff: More taste, less filling]*/
        DOFMem_Free( item );
    }
}


boolean DOFInterfaceItem_IsInputParameterListCompatible( const DOFItem item, uint8 parameterCount,
                                                         const DOFValue * parameters )
{
    PCR_Assert( item != NULL );

    /* Check the parameter count of the item. */
    if (item->typeCount[0] != parameterCount)
        return FALSE;

    /* Check each value's type against the item's parameter types. */
    return DOFTypeList_IsValueCompatibleWith( parameterCount, item->types[0], parameters );
}


boolean DOFInterfaceItem_IsOutputParameterListCompatible( const DOFItem item, uint8 parameterCount,
                                                          const DOFValue * parameters )
{
    PCR_Assert( item != NULL );

    /* Check the parameter count of the item. */
    if (item->typeCount[1] != parameterCount)
        return FALSE;

    /* Check each value's type against the item's parameter types. */
    return DOFTypeList_IsValueCompatibleWith( parameterCount, item->types[1], parameters );
}


DOFValue * DOFInterfaceItem_CreateValueListFromStream( const DOFItem item, OALStream stream, boolean evtNotXcp )
{
    /* Called for event or exception, and nothing else. */
    uint8      elementCount;
    DOFValue * retValues;

    PCR_Assert( item );
    if (evtNotXcp)
    {
        PCR_Assert( item->itemType == DOFITEM_EVENT );
        elementCount = DOFInterfaceEvent_GetParameterCount( (DOFInterfaceEvent)item );
    }
    else
    {
        PCR_Assert( item->itemType == DOFITEM_EXCEPTION );
        elementCount = DOFInterfaceException_GetParameterCount( (DOFInterfaceException)item );
    }
    retValues = DOFValueListAlloc_Create( elementCount );
    if (retValues)
    {
        uint8     i;
        DOFType * types;

        if (evtNotXcp)
            types = DOFInterfaceEvent_GetParameterTypeList( (DOFInterfaceEvent)item );
        else
            types = DOFInterfaceException_GetParameterTypeList( (DOFInterfaceException)item );
        for (i = 0; i < elementCount; i++)
        {
            boolean failed;

            retValues[i] = DOFType_CreateValue_UnmarshalFromStream( *(types++), stream, &failed );
            if (failed)
            {
                DOFValueListAlloc_Destroy( elementCount, retValues );
                return NULL;
            }
        }
    }

    return retValues;
}


/**
@}
**/
