/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/error.h>
#include <dof/value.h>
#include <internal/marshal.h>
#include <internal/value.h>


/** @internal @addtogroup OALMarshal
    Common marshal/unmarshal routines.
@{
**/


const uint8 * OALMarshal_StepOverCompressedLong( const uint8 * pBuf )                           /* docs in marshal.h. */
{
    return (const uint8 *)(pBuf + ((*pBuf & 0x80) ? ((*pBuf & 0x40) ? 4 : 2) : 1));
}


/*
**  "Private" Type and Method Declarations
*/


uint8 OALMarshal_GetCompressedValueSize( uint8 maxSize, uint32 value )
{
    uint8 lenbytes = ( 1 + ( value > 0x7F ) + ( value > 0x3FFF ) );
    if ( lenbytes > 2 )
        return( maxSize );
    return( lenbytes );
}

uint32 OALMarshal_CompressValue( uint8 maxSize, uint32 value, uint32 bufLength, uint8 * buffer )
{
    uint8 lenSize = OALMarshal_GetCompressedValueSize(maxSize,value);

    PCR_Assert(buffer);
    PCR_Assert( maxSize >= 2 );
    PCR_Assert( maxSize <= 4 );
    PCR_Assert( lenSize <= maxSize );
    if ( bufLength < lenSize )
        return 0;
    switch ( lenSize )
    {
        case 4:
            PCR_Assert( value <= 0x3FFFFFFF );
            *(buffer++) = (uint8)(( value >> 24 ) & 0x3F ) | 0xC0;
            *(buffer++) = (uint8)(( value >> 16 ) & 0xFF );
            *(buffer++) = (uint8)(( value >>  8 ) & 0xFF );
            *(buffer++) = (uint8)(  value         & 0xFF );
            break;
        case 3:
            PCR_Assert( value <= 0x3FFFFF );
            *(buffer++) = (uint8)(( value >> 16 ) & 0x3F ) | 0xC0;
            *(buffer++) = (uint8)(( value >>  8 ) & 0xFF );
            *(buffer++) = (uint8)(  value         & 0xFF );
            break;
        case 2:
            if ( maxSize == 2 )
            {
                PCR_Assert( value <= 0x7FFF );
                *(buffer++) = (uint8)(( value >>  8 ) & 0x7F ) | 0x80;
            }
            else
            {
                PCR_Assert( value <= 0x3FFF );
                *(buffer++) = (uint8)(( value >>  8 ) & 0x3F ) | 0x80;
            }
            *(buffer++) = (uint8)(  value         & 0xFF );
            break;
        case 1:
            PCR_Assert( value <= 0x7F );
            *(buffer++) = (uint8)(  value         & 0x7F );
            break;
        default:
            /* Invalid computed size! */
            PCR_Assert( 0 );
            break;
    }
    return( lenSize );
}

uint32 OALMarshal_UncompressValue( uint8 maxSize, uint32 * bufLength, const uint8 * buffer )
{
    uint32 value = 0;
    uint8 used = 0;
    uint8 size = maxSize;
    uint8 mask;

    PCR_Assert(bufLength&&buffer);
    PCR_Assert( maxSize >= 2 );
    PCR_Assert( maxSize <= 4 );

    if ( *bufLength < 1 )
        return OALMARSHAL_INVALID_VALUE;

    switch ( buffer[0] >> 6 )
    {
        case 0x02:
            /* Two Bytes */
            if ( maxSize > 2 )
                mask = 0x3F;
            else
                mask = 0x7F;
            size = 2;
            break;
        case 0x03:
            /* Three/Four Bytes */
            if ( maxSize > 2 )
                mask = 0x3F;
            else
                mask = 0x7F;
            break;
        default:
            /* One Byte */
            size = 1;
            mask = 0x7F;
            break;
    }

    if ( *bufLength < size )
        return OALMARSHAL_INVALID_VALUE;
    value = buffer[used++] & mask;
    while ( used < size )
        value = ( value << 8 ) | buffer[used++];
    if ( OALMarshal_GetCompressedValueSize( maxSize, value ) != size )
        return OALMARSHAL_INVALID_VALUE;
    *bufLength = used;
    return( value );
}

static DOFError compress_internal(OALStream stream, uint32 val, uint32 size)
{
    uint32 ret = OALMarshal_CompressValue(size, val, OALStream_GetFreeSize(stream), OALStream_GetWritePtr(stream));

    if(ret)
    {
        OALStream_AdjustWriteOffset(stream, (int32)ret);
        return DOFERROR_OK;
    }

    return DOFERROR_FAILED;
}

static DOFError uncompress_internal(OALStream stream, uint32 *val, uint32 size)
{
    uint32 buflen;
    uint32 ret;

    PCR_Assert(stream&&val);

    buflen = OALStream_GetDataSize(stream);
    ret = OALMarshal_UncompressValue(size, &buflen, OALStream_GetReadPtr(stream));

    if(ret!=OALMARSHAL_INVALID_VALUE)
    {
        OALStream_AdjustReadOffset(stream, (int32)buflen);
        *val = ret;
        return DOFERROR_OK;
    }

    return DOFERROR_FAILED;
}

DOFError OALMarshal_StreamCompress2(OALStream stream, uint16 val)
{
    return compress_internal(stream, val, 2);
}

DOFError OALMarshal_StreamCompress3(OALStream stream, uint32 val)
{
    return compress_internal(stream, val, 3);
}

DOFError OALMarshal_StreamCompress4(OALStream stream, uint32 val)
{
    return compress_internal(stream, val, 4);
}

DOFError OALMarshal_StreamUncompress2(OALStream stream, uint16 *val)
{
    /* assertion on val done in OALMarshal_UncompressValue */
    uint32 _val;
    DOFError err = uncompress_internal(stream, &_val, 2);

    if(err!=DOFERROR_OK)
        return err;

    *val = (uint16)_val;
    return DOFERROR_OK;
}

DOFError OALMarshal_StreamUncompress3(OALStream stream, uint32 *val)
{
    /* assertion on val done in OALMarshal_UncompressValue */
    return uncompress_internal(stream, val, 3);
}

DOFError OALMarshal_StreamUncompress4(OALStream stream, uint32 *val)
{
    /* assertion on val done in OALMarshal_UncompressValue */
    return uncompress_internal(stream, val, 4);
}

boolean OALMarshal_MarshalValue(DOFValue    value,
                                DOFType     type,
                                uint8    ** pBuffer,
                                uint32   *  pSize)
{
    return OALMarshal_MarshalValues(&value, &type, 1, pBuffer, pSize);
}

boolean OALMarshal_MarshalValues(DOFValue *  values,
                                 DOFType  *  types,
                                 uint16      count,
                                 uint8    ** pBuffer,
                                 uint32   *  pSize)
{
    uint16 i;

    (*pSize) = 0;
    for(i = 0; i < count; ++i)
    {
        (*pSize) += DOFValue_GetMarshalSize(values[i], types[i]);
    }
    (*pBuffer) = PCRMem_Alloc(*pSize);
    if((*pBuffer) != NULL)
    {
        uint32 writeSize = 0;

        for(i = 0; i < count; ++i)
        {
            uint32 temp = DOFValue_Marshal(values[i], types[i],
                                           (*pSize) - writeSize, &((*pBuffer)[writeSize]));

            if((temp > 0) || (values[i] == NULL))
            {
                writeSize += temp;
            }
            else
            {
                PCRMem_Free(*pBuffer);
                return FALSE;
            }
        }
        return TRUE;
    }
    return FALSE;
}

/**
@}
**/
