/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/connection.h>
#include <dof/system.h>
#include <dof/interface.h>
#include <internal/securityscope.h>
#include <dof/permissionset.h>
#include <dof/providerinfo.h>
#include <dof/result.h>
#include <dof/error.h>
#include <dof/exception.h>
#include <dof/value.h>
#include <internal/bridgeinternal.h>
#include <internal/epp_negative_one.h>
#include <internal/exception.h>
#include <internal/interfaceinternal.h>
#include <internal/marshal.h>
#include <internal/oap_v1.h>
#include <internal/operationconfiginternal.h>
#include <internal/permissionsetinternal.h>
#include <internal/providerinfo.h>
#include <internal/resultinternal.h>
#include <internal/requestinternal.h>
#include <internal/router.h>
#include <internal/value.h>

#include "oap_v1_receiveutil.h"

static void OAP_V1_ReceiveInvalidOpCode(OAP_V1_ReceiveData_t * data);

static boolean OAP_V1_Receive_ReadInterestPDU(OAP_V1_ReceiveData data);

static void OAP_V1_Receive_FinishOperation(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveAdvertise(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveActivate(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveChange(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveConnect(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveDefine(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveDefineComplete(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveException(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveFullConnect(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveGet(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveGetComplete(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveInvoke(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveInvokeComplete(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveOpen(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveOpenAccepted(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveProvide(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveRegister(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveRegisterAccepted(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveSignal(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveSet(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveSetComplete(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveSubscribe(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveSubscribeAccepted(OAP_V1_ReceiveData data);

static void OAP_V1_ReceiveWatch(OAP_V1_ReceiveData data);

static void OAP_V1_Receive_SendOperationToProvider(OAP_V1_ReceiveData data);

static void OAP_V1_Receive_SendOperationToRequestor(OAP_V1_ReceiveData data);

static uint32 OAP_V1_CalculateGrace(BlackboardData bbData);

ProtocolLayerResult OAP_V1_Receive(OAP_V1 self)
{
    OAP_V1_ReceiveData_t receiveData;

    uint16 appId;

    DOFOperation op = NULL;

    static const OAP_V1_ReceiveFunction receiveFunctions[32][2] = {
                    /* Cancel */
        /*  0 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveInvalidOpCode},
        /*  1 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveInvalidOpCode},
                    /* Change */
        /*  2 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveChange},
                    /* Full Connect */
        /*  3 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveFullConnect},
                    /* Connect */
        /*  4 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveConnect},
                    /* Advertise */
        /*  5 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveAdvertise},
                    /* Define */
        /*  6 */   {OAP_V1_ReceiveDefineComplete,    OAP_V1_ReceiveDefine},
        /*  7 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveInvalidOpCode},
        /*  8 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveInvalidOpCode},
                    /* Exception */
        /*  9 */   {OAP_V1_ReceiveException,         OAP_V1_ReceiveInvalidOpCode},
                    /* Get */
        /* 10 */   {OAP_V1_ReceiveGetComplete,       OAP_V1_ReceiveGet},
        /* 11 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveInvalidOpCode},
                    /* Invoke */
        /* 12 */   {OAP_V1_ReceiveInvokeComplete,    OAP_V1_ReceiveInvoke},
        /* 13 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveInvalidOpCode},
                    /* Open */
        /* 14 */   {OAP_V1_ReceiveOpenAccepted,      OAP_V1_ReceiveOpen},
        /* 15 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveInvalidOpCode},
                    /* Provide */
        /* 16 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveProvide},
        /* 17 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveInvalidOpCode},
        /* 18 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveInvalidOpCode},
        /* 19 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveInvalidOpCode},
                    /* Set */
        /* 20 */   {OAP_V1_ReceiveSetComplete,       OAP_V1_ReceiveSet},
        /* 21 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveInvalidOpCode},
                    /* Signal */
        /* 22 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveSignal},
        /* 23 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveInvalidOpCode},
                    /* Subscribe */
        /* 24 */   {OAP_V1_ReceiveSubscribeAccepted, OAP_V1_ReceiveSubscribe},
                    /* Register */
        /* 25 */   {OAP_V1_ReceiveRegisterAccepted,  OAP_V1_ReceiveRegister},
        /* 26 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveInvalidOpCode},
        /* 27 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveInvalidOpCode},
                    /* Activate */
        /* 28 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveActivate},
        /* 29 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveInvalidOpCode},
                    /* Watch */
        /* 30 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveWatch},
        /* 31 */   {OAP_V1_ReceiveInvalidOpCode,     OAP_V1_ReceiveInvalidOpCode}
    };

    if(self->blackboardData->associatedChannel->remoteChannelData->flags.secureDesire &&
       !(self->blackboardData->flags.receivedAuthenticated || self->blackboardData->flags.receivedEncrypted))
    {
        return PLRESULT_ERROR;
    }
    if((OALMarshal_StreamUncompress2(self->blackboardData->receiveStream, &appId) != DOFERROR_OK) || (appId != OAP_V1_APPID))
    {
        /* Did we fail or unmarshal the wrong thing? */
        return PLRESULT_ERROR;
    }
    else if(self->blackboardData->opID != NULL)
    {
        op = OALOperationManager_CheckExistingOperation(DOFSecurityMembrane_GetDomainID(self->blackboardData->associatedChannel->securityMembrane), self->blackboardData->opID);
        if ( op &&
             !self->blackboardData->delay &&
             self->blackboardData->flags.isCommandNotResponse &&
             OALChannel_IsEquivalent( op->sources->channel, self->blackboardData->associatedChannel ) &&
             DOFAddress_IsEqual( op->sources->address, self->blackboardData->transportAddr ) &&
             ( (int8)(self->blackboardData->retry - op->sources->retry_count) >= 0 ) )          /* if the retry count is lower than last seen do not cancel the op. */
        {
            /* EPP Cancel */
            DOFOperation_Cancel(op);
            DOFOperation_Internal_Destroy(op);
            return PLRESULT_OKDONOTHING;
        }
    }

    if ( op )
    {
        DOFOperation_Lock(op);
        /*  Is this a loopback of our own request? */
        if(!op->flags.bits.internally_created && self->blackboardData->flags.isCommandNotResponse)
        {
            DOFOperation_Internal_Destroy(op);
            DOFOperation_Unlock(op);
            return PLRESULT_OKDONOTHING;
        }
        /*  Has this operation already been cancelled? */
        else if(op->flags.bits.cancelled)
        {
            DOFOperation_Internal_Destroy(op);
            DOFOperation_Unlock(op);
            return PLRESULT_OKDONOTHING;
        }
        /*  Have we already received the maximum number of valid responses? */
        else if(!op->flags.bits.internally_created &&
                !self->blackboardData->flags.isCommandNotResponse &&
                op->operationConfig != NULL &&
                DOFOperationControl_GetMaxResponderCount(op->operationConfig) == 0)
        {
            DOFOperation_Internal_Destroy(op);
            DOFOperation_Unlock(op);
            return PLRESULT_OKDONOTHING;
        }
        else
        {
            DOFOperation_Unlock(op);
        }
        /* Reparenting must be done after unmarshalling and validating the operation */
    }

    OAP_V1_ReceiveData_Init(&receiveData, self, op);
    if(OALStream_Read8(self->blackboardData->receiveStream, &receiveData.firstByte))
    {
        receiveFunctions[receiveData.firstByte & 0x1F][self->blackboardData->flags.isCommandNotResponse](&receiveData);
    }
    else
    {
        /* No-op. */
        receiveData.status = OAP_V1_ReceiveStatus_DONOTHING;
    }
    switch(receiveData.status)
    {
    case OAP_V1_ReceiveStatus_MEMORY_ERROR:
        break;

    case OAP_V1_ReceiveStatus_OK:
        if(op != NULL)
        {
            DOFOperation_Lock(op);
            if ( self->blackboardData->flags.isCommandNotResponse )
            {
                boolean same_source = ( OALChannel_IsEquivalent( op->sources->channel, self->blackboardData->associatedChannel ) && DOFAddress_IsEqual( op->sources->address, self->blackboardData->transportAddr ) );
                if ( op->flags.bits.lost ||
                     self->blackboardData->seq < op->sources->sequence ||
                     ( self->blackboardData->seq == op->sources->sequence && !same_source && op->sources->channel->remoteChannelData->flags.isLossy &&
                       ( !self->blackboardData->associatedChannel->remoteChannelData->flags.isLossy ||
                         ( self->blackboardData->associatedChannel->remoteChannelData->flags.isConnection && !op->sources->channel->remoteChannelData->flags.isConnection ) ||
                         ( !self->blackboardData->associatedChannel->remoteChannelData->flags.isConnection && !op->sources->channel->remoteChannelData->flags.isConnection && ( (int8)(self->blackboardData->retry - op->sources->retry_count) >= 0 ) ) ) ) )
                {
                    /* Update the source and operation state (sequence & address). */
                    if ( same_source || EPP_NegOne_ReparentOperation( op, self->blackboardData->associatedChannel, self->blackboardData->transportAddr, self->blackboardData->associatedChannel->remoteChannelData->flags.isLossy ) == PLRESULT_OKDONOTHING)
                    {
                        op->sources->sequence = self->blackboardData->seq;
                        goto process_as_retry;
                    }
                }
                else if ( same_source && self->blackboardData->seq > op->sources->sequence )
                {
                    /* Lost State - Send Source Lost. */
                    EPP_NegOne_SourceLost( op );
                    /* No further processing. */
                }
                else if ( same_source && ( (int8)(self->blackboardData->retry - op->sources->retry_count) > 0 ) )
                {
                    process_as_retry:
                    /* Update the operation state (retry and delay). */
                    if ( self->blackboardData->delay )
                    {
                        op->sources->retry_count = self->blackboardData->retry;
                        op->sources->oal_timeout = self->blackboardData->delay;
                        op->flags.bits.new_retry_count = 1;
                        OAP_V1_Receive_SendOperationToProvider(&receiveData);
                    }
                    else
                        DOFOperation_Cancel( op );
                }
                else
                {
                    /* A duplicate, old, or more distant operation; ignore. */
                }
            }
            else
            {
                if(!op->flags.bits.internally_created &&
                    op->operationConfig != NULL &&
                    DOFOperationControl_GetMaxResponderCount(op->operationConfig) != DOFOPERATIONCONTROL_RESPONDERS_ALL)
                {
                    uint16 newResponderCount = DOFOperationControl_GetMaxResponderCount(op->operationConfig) - 1;

                    DOFOperationControl_SetMaxResponderCount(op->operationConfig, newResponderCount);
                }
            }
            DOFOperation_Unlock(op);
        }

    case OAP_V1_ReceiveStatus_INVALID_BINDING:
    case OAP_V1_ReceiveStatus_DONOTHING:
    case OAP_V1_ReceiveStatus_NOOPERATION:
        OAP_V1_ReceiveData_Clear(&receiveData);
        return PLRESULT_OKDONOTHING;

    case OAP_V1_ReceiveStatus_NOT_SUPPORTED:
        {
            DOFResult r = DOFResult_Exception_Create( self->blackboardData->associatedChannel, self->blackboardData->transportAddr, DOFEXCEPTION_NOT_SUPPORTED, DOFOBJECTID_BROADCAST );
            if ( r && (OAP_V1_Send( self, OAP_V1_GetExceptionOperation(), FALSE, r ) == PLRESULT_OKSENDDOWN ))
            {
                DOFResult_Destroy( r );
                OAP_V1_ReceiveData_Clear(&receiveData);
                return PLRESULT_OKSENDDOWN;
            }
        }
        OAP_V1_ReceiveData_Clear(&receiveData);
        return PLRESULT_OKDONOTHING;

    default:
    case OAP_V1_ReceiveStatus_INVALID_OPCODE:
    case OAP_V1_ReceiveStatus_STREAM_ERROR:
        OAP_V1_ReceiveData_Clear(&receiveData);
        return PLRESULT_ERROR;
    }
    OAP_V1_ReceiveData_Clear(&receiveData);
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "OAP V1: Receive failed. (out of memory error)" );
    return PLRESULT_OKDONOTHING;
}

static void OAP_V1_Receive_SendOperationToProvider_Route(DOFOperation op)
{
    switch(op->operation_type)
    {
    case DOFOPERATIONTYPE_PROVIDE:
        DOFRouter_RouteToRequestor(op, NULL);

    case DOFOPERATIONTYPE_CHANGED:
    case DOFOPERATIONTYPE_SIGNAL:
        /* Do NOTHING in these cases! Already routed in the regular OAP processing. */
        break;

    default:
        DOFRouter_RouteToProvider(op);
        break;
    }
}

static void OAP_V1_ReceiveInvalidOpCode(OAP_V1_ReceiveData data)
{
    data->status = OAP_V1_ReceiveStatus_INVALID_OPCODE;
}

static void OAP_V1_ReceiveAdvertise(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_ReadFullBinding,
                                                     OAP_V1_Receive_SetOperationConfig, NULL};

    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data) && !data->operation)
    {
        data->operation =
            DOFOperation_Advertise_CreateInternal(data->oap->blackboardData->associatedChannel, data->interfaceID,
                                                  data->objectID, data->oap->blackboardData->delay,
                                                  OAP_V1_CalculateGrace(data->oap->blackboardData), data->oap->blackboardData->transportAddr);

        if(data->operation != NULL)
        {
            OAP_V1_Receive_FinishOperation(data);
            OAP_V1_Receive_SendOperationToProvider(data);
        }
        else
        {
            data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
        }
    }
}

/** Process an Activate request.
 *
 *  @param[in,out]  data
 *      all data related to the incoming request
 */
static void OAP_V1_ReceiveActivate(OAP_V1_ReceiveData data)
{
    if(OAP_V1_Receive_ReadInterestPDU(data) && !data->operation)
    {
        data->operation =
            DOFOperation_Activate_CreateInternal(data->oap->blackboardData->associatedChannel, data->interfaceID,
                                                 data->objectID, data->alias, data->oap->blackboardData->delay,
                                                 OAP_V1_CalculateGrace(data->oap->blackboardData));
        if(data->operation != NULL)
        {
            OAP_V1_Receive_FinishOperation(data);
            OAP_V1_Receive_SendOperationToProvider(data);
        }
        else
        {
            data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
        }
    }
}

static void OAP_V1_ReceiveChange(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_ReadItemID,
                                                     OAP_V1_Receive_ReadBinding,
                                                     OAP_V1_Receive_ReadSequenceNumber, NULL};

    data->functions = functions;
    data->aliasAddress = data->oap->blackboardData->transportAddr;
    if(OAP_V1_Receive_IterateFunctions(data))
    {
        if(data->operation == NULL)
        {
            data->operation =
                DOFOperation_Change_CreateInternal(data->oap->blackboardData->associatedChannel, data->interfaceID, data->objectID,
                                                   data->alias, data->oap->blackboardData->delay,
                                                   OAP_V1_CalculateGrace(data->oap->blackboardData), data->stream, data->itemID);
        }
        if(data->operation != NULL)
        {
            if(OAP_V1_Receive_CreateValueResult(data))
            {
                DOFResult_SetSequenceNumber(data->result, data->sequenceNumber, FALSE, FALSE);
                OAP_V1_Receive_FinishOperation(data);
                OAP_V1_Receive_SendOperationToRequestor(data);
            }
        }
    }
}

static void OAP_V1_ReceiveConnect(OAP_V1_ReceiveData data)
{
    if(OAP_V1_Receive_ReadInterestPDU(data))
    {
        if(DOFObjectID_IsUnicast(data->objectID) &&
           !DOFInterfaceID_IsEqual(data->interfaceID, DOFINTERFACEID_WILDCARD))
        {
            if(!data->operation)
            {
                data->operation =
                    DOFOperation_Connect_CreateInternal(data->oap->blackboardData->associatedChannel, data->interfaceID,
                                                        data->objectID, data->alias, data->oap->blackboardData->delay,
                                                        OAP_V1_CalculateGrace(data->oap->blackboardData));
                if(data->operation != NULL)
                {
                    OAP_V1_Receive_FinishOperation(data);
                    OAP_V1_Receive_SendOperationToProvider(data);
                }
                else
                {
                    data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
                }
            }
        }
        else
        {
            data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
        }
    }
}

static void OAP_V1_ReceiveDefine(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_ReadFullBinding, NULL};

    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data) && !data->operation)
    {
        data->operation =
            DOFOperation_Define_CreateInternal(data->oap->blackboardData->associatedChannel,
                                               data->interfaceID, data->objectID,
                                               data->oap->blackboardData->delay,
                                               OAP_V1_CalculateGrace(data->oap->blackboardData));
        if(data->operation != NULL)
        {
            OAP_V1_Receive_FinishOperation(data);
            OAP_V1_Receive_SendOperationToProvider(data);
        }
        else
        {
            data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
        }
    }
}

static void OAP_V1_ReceiveDefineComplete(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_VerifyOperation,
                                                     OAP_V1_Receive_CreateValueResult, NULL};

    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data))
    {
        OAP_V1_Receive_SendOperationToRequestor(data);
    }
}

static void OAP_V1_ReceiveFullConnect(OAP_V1_ReceiveData data)
{
    if(OAP_V1_Receive_ReadInterestPDU(data) && !data->operation)
    {
        if(DOFObjectID_IsUnicast(data->objectID) &&
           !DOFInterfaceID_IsEqual(data->interfaceID, DOFINTERFACEID_WILDCARD))
        {
            if(!data->operation)
            {
                data->operation =
                    DOFOperation_FullConnect_CreateInternal(data->oap->blackboardData->associatedChannel, data->interfaceID,
                                                            data->objectID,  data->alias, data->oap->blackboardData->delay,
                                                            OAP_V1_CalculateGrace(data->oap->blackboardData));
                if(data->operation != NULL)
                {
                    OAP_V1_Receive_FinishOperation(data);
                    OAP_V1_Receive_SendOperationToProvider(data);
                }
                else
                {
                    data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
                }
            }
        }
        else
        {
            data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
        }
    }
}

static void OAP_V1_ReceiveGet(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_SetOperationConfig,
                                                     OAP_V1_Receive_ReadResponseControl,
                                                     OAP_V1_Receive_ReadItemID,
                                                     OAP_V1_Receive_ReadBinding, NULL};

    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data))
    {
        if ( !data->operation )
        {
            data->operation =
                DOFOperation_Get_CreateInternal(data->oap->blackboardData->associatedChannel,
                                                data->interfaceID, data->objectID, data->alias,
                                                data->oap->blackboardData->delay,
                                                OAP_V1_CalculateGrace(data->oap->blackboardData), data->itemID);
            if(data->operation != NULL)
            {
                OAP_V1_Receive_FinishOperation(data);
                OAP_V1_Receive_SendOperationToProvider(data);
            }
            else
            {
                data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
            }
        }
    }
}

static void OAP_V1_ReceiveGetComplete(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_VerifyOperation,
                                                     OAP_V1_Receive_ReadProviderID,
                                                     OAP_V1_Receive_CreateValueResult, NULL};

    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data))
    {
        OAP_V1_Receive_SendOperationToRequestor(data);
    }
}

static void OAP_V1_ReceiveException(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_VerifyOperation,
                                                     OAP_V1_Receive_ReadProviderID,
                                                     OAP_V1_Receive_ReadItemID, NULL};

    if(data->oap->blackboardData->flags.isMulticast)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_INFO, "OAP V1: Rejecting multicast exception.");
        return;
    }
    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data))
    {
        DOFException except = NULL;

        DOFProviderInfo provInfo = NULL;

        DOFConnectionState connState;

        if(data->operation->ackTimerItem != NULL)
        {
            OALTimerItem_Remove(data->operation->ackTimerItem);
        }
        connState = OALChannel_CreateConnectionState(data->oap->blackboardData->associatedChannel,
                                                     data->oap->blackboardData->transportAddr);
        if(connState != NULL)
        {
            provInfo = DOFProviderInfo_Create(connState, data->providerID,
                                              data->oap->blackboardData->transportAddr);
            DOFConnectionState_Destroy(connState);
        }
        if(!(data->firstByte & 0x80))
        {
            if(provInfo != NULL)
            {
                except = DOFProviderException_CreateStream_ProviderInfo(data->stream, provInfo, data->itemID);
            }
        }
        else if(provInfo != NULL)
        {
            except = DOFErrorException_Create_ProviderInfo(data->itemID, NULL, provInfo);
        }
        DOFProviderInfo_Destroy(provInfo);
        if(except == NULL)
        {
            data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
            return;
        }

        data->result = DOFResult_Exception_Create(data->oap->blackboardData->associatedChannel,
                                                  data->oap->blackboardData->transportAddr,
                                                  except, data->providerID);
        DOFException_Destroy(except);
        if(data->result ==  NULL)
        {
            data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
            return;
        }
        DOFResult_SetStream(data->result, data->stream);
        OAP_V1_Receive_SendOperationToRequestor(data);
    }
}

static void OAP_V1_ReceiveInvoke(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_SetOperationConfig,
                                                     OAP_V1_Receive_ReadResponseControl,
                                                     OAP_V1_Receive_ReadItemID,
                                                     OAP_V1_Receive_ReadBinding, NULL};

    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data))
    {
        if ( !data->operation )
        {
            data->operation =
                DOFOperation_Invoke_CreateInternal(data->oap->blackboardData->associatedChannel,
                                                   data->interfaceID, data->objectID,  data->alias,
                                                   data->oap->blackboardData->delay,
                                                   OAP_V1_CalculateGrace(data->oap->blackboardData),
                                                   data->stream, data->itemID);
            if(data->operation != NULL)
            {
                if ( !data->operation->opID )
                    OAP_V1_Receive_FinishOperation(data);
                OAP_V1_Receive_SendOperationToProvider(data);
            }
            else
            {
                data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
            }
        }
        else
        {
            DOFOperation_Lock(data->operation);
            if(data->operation->valueStream != NULL)
            {
                OALStream_Release(data->operation->valueStream, TRUE);
                data->operation->valueStream = data->stream;
                OALStream_Retain(data->stream);
            }
            DOFOperation_Unlock(data->operation);
        }
    }
}

static void OAP_V1_ReceiveInvokeComplete(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_VerifyOperation,
                                                     OAP_V1_Receive_ReadProviderID,
                                                     OAP_V1_Receive_CreateValueListResult, NULL};

    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data))
    {
        OAP_V1_Receive_SendOperationToRequestor(data);
    }
}

static void OAP_V1_ReceiveOpen(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_SetOperationConfig,
                                                     OAP_V1_Receive_ReadBinding,
                                                     OAP_V1_Receive_ReadSessionIID,
                                                     OAP_V1_Receive_ReadSessionGUID, NULL};

    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data))
    {
        if ( !data->operation )
        {
            data->operation =
                DOFOperation_Open_CreateInternal(data->oap->blackboardData->associatedChannel, data->interfaceID,
                                                 data->objectID, data->alias, data->oap->blackboardData->delay,
                                                 OAP_V1_CalculateGrace(data->oap->blackboardData), &data->sessionGUID, data->sessionIID);
            if(data->operation != NULL)
            {
                OAP_V1_Receive_FinishOperation(data);
                OAP_V1_Receive_SendOperationToProvider(data);
            }
            else
            {
                data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
            }
        }
    }
}

static void OAP_V1_ReceiveOpenAccepted(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_VerifyOperation, NULL};

    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data))
    {
        if(data->operation->ackTimerItem != NULL)
        {
            OALTimerItem_Remove(data->operation->ackTimerItem);
        }
        if(data->operation != NULL)
        {
            data->result = DOFResult_NoValue_Create(data->oap->blackboardData->associatedChannel,
                                                    data->oap->blackboardData->transportAddr,
                                                    DOFERROR_OK, data->operation->object_id.object_id);
            if(data->result == NULL)
            {
                data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
                return;
            }
            OAP_V1_Receive_SendOperationToRequestor(data);
            return;
        }
        data->status = OAP_V1_ReceiveStatus_NOOPERATION;
    }
}

static void OAP_V1_ReceiveProvide(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_ReadAlias,
                                                     OAP_V1_Receive_ReadFullBinding,
                                                     OAP_V1_Receive_ReadLosslessCount,
                                                     OAP_V1_Receive_ReadLossyCount, NULL};

    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data))
    {
        if ( data->oap->blackboardData->associatedChannel->remoteChannelData->flags.isLossy )
        {
            ++(data->lossyCount);
        }
        else
        {
            ++(data->losslessCount);
        }
        if ( !data->operation )
        {
            data->operation =
                DOFOperation_Provide_CreateInternal(data->oap->blackboardData->associatedChannel, data->interfaceID,
                                                    data->objectID, data->alias, data->oap->blackboardData->delay,
                                                    OAP_V1_CalculateGrace(data->oap->blackboardData), data->losslessCount, data->lossyCount);
            if(data->operation != NULL)
            {
                OAP_V1_Receive_FinishOperation(data);
                data->result = DOFResult_NoValue_Create(data->oap->blackboardData->associatedChannel,
                                                        data->oap->blackboardData->transportAddr,
                                                        DOFERROR_OK, data->objectID);
                if(data->result == NULL)
                {
                    data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
                    return;
                }
                /* These next two calls output a log message if they return anything but success. */
                EPP_NegOne_AddCommandOperation(data->operation, data->oap->blackboardData->associatedChannel);
                OALOperationManager_AddOperation(data->operation, data->oap->blackboardData->opID, FALSE); /* don't need lock because nobody knows about this op */
                OAP_V1_Receive_SendOperationToRequestor(data);
            }
            else
            {
                data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
            }
        }
        else
        {
            DOFOperation_Lock(data->operation);
            data->operation->sources->distLossless = data->losslessCount;
            data->operation->sources->distLossy = data->lossyCount;
            DOFOperation_Unlock(data->operation);
        }
    }
}

static void OAP_V1_ReceiveRegister(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_SetOperationConfig,
                                                     OAP_V1_Receive_ReadResponseControl,
                                                     OAP_V1_Receive_ReadItemID,
                                                     OAP_V1_Receive_ReadBinding, NULL};

    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data))
    {
        if ( !data->operation )
        {
            data->operation =
                DOFOperation_Subscribe_CreateInternal(data->oap->blackboardData->associatedChannel,
                                                      data->interfaceID, data->objectID,  data->alias,
                                                      data->oap->blackboardData->delay,
                                                      OAP_V1_CalculateGrace(data->oap->blackboardData),
                                                      data->itemID, data->minimumDelta, FALSE);
            if(data->operation != NULL)
            {
                OAP_V1_Receive_FinishOperation(data);
                OAP_V1_Receive_SendOperationToProvider(data);
            }
            else
            {
                data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
            }
        }
    }
}

static void OAP_V1_ReceiveRegisterAccepted(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_VerifyOperation,
                                                     OAP_V1_Receive_ReadProviderID,
                                                     OAP_V1_Receive_ReadSequenceNumber,
                                                     OAP_V1_Receive_CreateEmptyResult, NULL};

    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data))
    {
        if(data->operation->ackTimerItem != NULL)
        {
            OALTimerItem_Remove(data->operation->ackTimerItem);
        }
        DOFResult_SetSequenceNumber(data->result, data->sequenceNumber, FALSE, FALSE);
        OAP_V1_Receive_SendOperationToRequestor(data);
    }
}

static void OAP_V1_ReceiveSignal(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_ReadItemID,
                                                     OAP_V1_Receive_ReadBinding,
                                                     OAP_V1_Receive_ReadSequenceNumber, NULL};

    data->aliasAddress = data->oap->blackboardData->transportAddr;
    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data))
    {
        if(data->operation == NULL)
        {
            data->operation =
                DOFOperation_Signal_CreateInternal(data->oap->blackboardData->associatedChannel,
                                                   data->interfaceID, data->objectID, data->alias, data->oap->blackboardData->delay,
                                                   OAP_V1_CalculateGrace(data->oap->blackboardData), data->stream, data->itemID);
        }
        if(data->operation != NULL)
        {
            if(OAP_V1_Receive_CreateValueListResult(data))
            {
                DOFResult_SetSequenceNumber(data->result, data->sequenceNumber, FALSE, FALSE);
                OAP_V1_Receive_FinishOperation(data);
                OAP_V1_Receive_SendOperationToRequestor(data);
            }
        }
    }
}

static void OAP_V1_ReceiveSet(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_SetOperationConfig,
                                                     OAP_V1_Receive_ReadResponseControl,
                                                     OAP_V1_Receive_ReadItemID,
                                                     OAP_V1_Receive_ReadBinding, NULL};

    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data))
    {
        if ( !data->operation )
        {
            data->operation =
                DOFOperation_Set_CreateInternal(data->oap->blackboardData->associatedChannel, data->interfaceID,
                                                data->objectID,  data->alias, data->oap->blackboardData->delay,
                                                OAP_V1_CalculateGrace(data->oap->blackboardData), data->stream, data->itemID);
            if(data->operation != NULL)
            {
                if ( !data->operation->opID )
                    OAP_V1_Receive_FinishOperation(data);
                OAP_V1_Receive_SendOperationToProvider(data);
            }
            else
            {
                data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
            }
        }
        else
        {
            DOFOperation_Lock(data->operation);
            if(data->operation->valueStream != NULL)
            {
                OALStream_Release(data->operation->valueStream, TRUE);
                data->operation->valueStream = data->stream;
                OALStream_Retain(data->stream);
            }
            DOFOperation_Unlock(data->operation);
        }
    }
}

static void OAP_V1_ReceiveSetComplete(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_VerifyOperation,
                                                     OAP_V1_Receive_ReadProviderID,
                                                     OAP_V1_Receive_CreateEmptyResult,
                                                     NULL};

    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data))
    {
        OAP_V1_Receive_SendOperationToRequestor(data);
    }
}

static void OAP_V1_ReceiveSubscribe(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_SetOperationConfig,
                                                     OAP_V1_Receive_ReadResponseControl,
                                                     OAP_V1_Receive_ReadItemID,
                                                     OAP_V1_Receive_ReadBinding,
                                                     OAP_V1_Receive_ReadMinimumDelta, NULL};

    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data))
    {
        if ( !data->operation )
        {
            data->operation =
                DOFOperation_Subscribe_CreateInternal(data->oap->blackboardData->associatedChannel,
                                                      data->interfaceID, data->objectID,  data->alias,
                                                      data->oap->blackboardData->delay,
                                                      OAP_V1_CalculateGrace(data->oap->blackboardData), data->itemID,
                                                      data->minimumDelta, TRUE);
            if(data->operation != NULL)
            {
                OAP_V1_Receive_FinishOperation(data);
                OAP_V1_Receive_SendOperationToProvider(data);
            }
            else
            {
                data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
            }
        }
    }
}

static void OAP_V1_ReceiveSubscribeAccepted(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_VerifyOperation,
                                                     OAP_V1_Receive_ReadProviderID,
                                                     OAP_V1_Receive_ReadSequenceNumber,
                                                     OAP_V1_Receive_CreateValueResult, NULL};

    data->functions = functions;
    if(OAP_V1_Receive_IterateFunctions(data))
    {
        if(data->operation->ackTimerItem != NULL)
        {
            OALTimerItem_Remove(data->operation->ackTimerItem);
        }
        DOFResult_SetSequenceNumber(data->result, data->sequenceNumber, FALSE, FALSE);
        OAP_V1_Receive_SendOperationToRequestor(data);
    }
}

static void OAP_V1_ReceiveWatch(OAP_V1_ReceiveData data)
{
    if(OAP_V1_Receive_ReadInterestPDU(data) && !data->operation)
    {
        data->operation =
            DOFOperation_Watch_CreateInternal(data->oap->blackboardData->associatedChannel, data->interfaceID,
                                              data->objectID, data->alias, data->oap->blackboardData->delay,
                                              OAP_V1_CalculateGrace(data->oap->blackboardData));
        if(data->operation != NULL)
        {
            OAP_V1_Receive_FinishOperation(data);
            OAP_V1_Receive_SendOperationToProvider(data);
        }
        else
        {
            data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
        }
    }
}

static boolean OAP_V1_Receive_ReadInterestPDU(OAP_V1_ReceiveData data)
{
    static const OAP_V1_ReceiveHelper functions[] = {OAP_V1_Receive_VerifyOpID,
                                                     OAP_V1_Receive_SetOperationConfig,
                                                     OAP_V1_Receive_ReadResponseControl,
                                                     OAP_V1_Receive_ReadBinding, NULL};

    data->functions = functions;
    return OAP_V1_Receive_IterateFunctions(data);
}

static void OAP_V1_Receive_FinishOperation(OAP_V1_ReceiveData data)
{
    /* op mutex protection not needed, as this is called only just after the operation is created. */
    data->operation->flags.bits.is_connectionless = data->oap->isConnectionless;
    data->operation->sources->retry_count = data->oap->blackboardData->retry;
    data->operation->sources->sequence = data->oap->blackboardData->seq;
    /* We have an OAP op received from wire- mark it with EPP header flag that was on the wire. */
    data->operation->flags.bits.commandCompletesOnAnyFinal = data->oap->blackboardData->flags.commandCompletesOnAnyFinal;
    PCR_Assert(data->operation->opID == NULL);
    PCR_Assert(data->operation->sources->address == NULL);
    data->operation->ackTree = data->ackTree;
    data->operation->operationConfig = data->opConfig;
    /* We can fail and the caller will not know. */
    data->operation->opID = DOFOperationID_Copy(data->oap->blackboardData->opID);
    data->operation->sources->address = DOFAddress_Copy(data->oap->blackboardData->transportAddr);
    data->operation->sources->request = DOFRequest_Create(NULL, data->operation, NULL);
    if ( !data->operation->sources->request )
        data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
}

static void OAP_V1_Receive_SendOperationToProvider(OAP_V1_ReceiveData data)
{
    PCR_Assert(data != NULL);
    PCR_Assert(data->operation != NULL);
    /* Update flooded status */
    if ( data->oap->blackboardData->flags.isFlooded || data->oap->blackboardData->flags.isMulticast )
        data->operation->flags.bits.isFlooded = TRUE;
    /* These next two calls output a log message if they return anything but success. */
    if ( !data->operation->flags.bits.added_to_operation_table )
    {
        EPP_NegOne_AddCommandOperation(data->operation, data->oap->blackboardData->associatedChannel);
        OALOperationManager_AddOperation(data->operation, data->oap->blackboardData->opID, FALSE); /* don't need lock because nobody else knows about it */
    }
    else
    {
        /* Already added, but at least update the timer. */
        _operation_extend( data->operation, data->operation->sources->oal_timeout, TRUE );
    }
    data->operation->sources->oal_timeout = data->oap->blackboardData->delayWithoutGracePeriod; /* Reset timeout to what was on the wire. */
    if ( !OAP_V1_Receive_VerifyPermissions(data) )
    {
        data->status = OAP_V1_ReceiveStatus_DONOTHING;
    }
    else
        OAP_V1_Receive_SendOperationToProvider_Route(data->operation);
}

static void OAP_V1_Receive_SendOperationToRequestor(OAP_V1_ReceiveData data)
{
    PCR_Assert(data->result);
    if ( OAP_V1_Receive_VerifyPermissions(data) )
    {
        DOFRouter_RouteToRequestor(data->operation, data->result);
    }
    else
    {
        PCR_Assert(data->status == OAP_V1_ReceiveStatus_DONOTHING);
    }
}

static uint32 OAP_V1_CalculateGrace(BlackboardData bbData)
{
    return bbData->delay - bbData->delayWithoutGracePeriod;
}
