/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/interface.h>
#include <dof/interfaceid.h>
#include <dof/item.h>
#include <dof/objectid.h>
#include <dof/permissionset.h>
#include <dof/result.h>
#include <dof/error.h>
#include <dof/system.h>
#include <dof/value.h>
#include <internal/bridgeinternal.h>
#include <internal/interfaceidinternal.h>
#include <internal/marshal.h>
#include <internal/objectidinternal.h>
#include <internal/operationinternal.h>
#include <internal/operationconfiginternal.h>
#include <internal/permissioninternal.h>
#include <internal/permissionsetinternal.h>
#include <internal/remotechannel.h>
#include <internal/resultinternal.h>
#include <internal/router.h>
#include <internal/securityscope.h>
#include <internal/stream.h>
#include <internal/value.h>

#include "oap_v1_receiveutil.h"

typedef struct ExtendPermContext_t
{
    DOFOperation op;

    DOFResult result;

    OALPermissionScope scope;

    OALChannel channel;

    boolean isCommand;
    boolean isDelayed;
} * ExtendPermContext;

static void OAP_V1_Receive_extendPermCallback(DOFError   status,
                                              void     * context);


static uint8 OAP_V1_ReceiveData_DetermineAliasSize(OAP_V1_ReceiveData_t * self);

void OAP_V1_ReceiveData_Init(OAP_V1_ReceiveData self,
                             OAP_V1             oap,
                             DOFOperation       op)
{
    PCR_Assert(oap);
    PCR_Assert(oap->blackboardData->receiveStream);

    PCRMem_Set(self, 0, sizeof(OAP_V1_ReceiveData_t));
    self->stream = oap->blackboardData->receiveStream;
    self->oap = oap;
    self->operation = op;

    self->ackTree = NULL;
}

void OAP_V1_ReceiveData_Clear(OAP_V1_ReceiveData self)
{
    PCR_Assert(self);

    if ( self->ackTree != NULL && ( !self->operation || self->ackTree != self->operation->ackTree ) )
    {
        OALAVLTree_Destroy(self->ackTree, NULL, NULL);
    }
    DOFInterfaceID_Destroy( self->sessionIID );
    DOFObjectID_Destroy( self->providerID );
    DOFObjectID_Destroy( self->objectID );
    DOFInterfaceID_Destroy( self->interfaceID );
    if ( !self->operation || self->opConfig != self->operation->operationConfig )
    {
        DOFOperationControl_Destroy( self->opConfig );
    }
    DOFOperation_Internal_Destroy( self->operation );
    DOFResult_Destroy( self->result );
    DOFSecurityScope_Destroy( self->secScope );
}

boolean OAP_V1_Receive_SetOperationConfig(OAP_V1_ReceiveData data)
{
    PCR_Assert(data != NULL);
    if ( data->operation )
    {
        data->opConfig = data->operation->operationConfig;
    }
    else
    {
        data->opConfig = DOFOperationControl_Create();
        if ( !data->opConfig )
        {
            data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
            return FALSE;
        }
        DOFOperationControl_SetRetryPeriod( data->opConfig, DOF_TIMEOUT_NEVER );
    }
    PCR_Assert( data->opConfig );
    return TRUE;
}

boolean OAP_V1_Receive_VerifyPermissions(OAP_V1_ReceiveData data)
{
    OALPermissionScope scope;
    DOFPermission permToCheck;

    PCR_Assert(data != NULL);

    permToCheck = DOFPermission_Copy(data->oap->blackboardData->flags.isCommandNotResponse ?
                                                              data->operation->commandPerm :
                                                              data->operation->responsePerm);
    if(permToCheck == NULL)     /* Response to a broadcast command */
    {
        PCR_Assert(!data->oap->blackboardData->flags.isCommandNotResponse);
        permToCheck = _createNeededPermission(data->operation->operation_type, TRUE,
                                              data->providerID, data->operation->interface_id);
        if(permToCheck == NULL)
        {
            data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
            return FALSE;
        }
    }
    if(data->result != NULL)
    {
        DOFResult_SetPermission(data->result, permToCheck);
    }
    if ( data->oap->blackboardData->associatedChannel->remoteChannelData->cipher != NULL )
    {
        scope = DOFPermissionSet_Validate(OALCipher_GetRemotePerms( data->oap->blackboardData->associatedChannel->remoteChannelData->cipher, data->oap->blackboardData->keyState ),
                                          permToCheck);
    }
    else
    {
        scope = OALPermissionScope_Copy(DOFSecurityScope_GetOALPermissionScope(DOF_GetDefaultSecurityScope(DOF_GetDof())));
    }

    if ( scope != NULL )
    {
        boolean havePerms;
        boolean isSameDomain = ( DOFObjectID_Compare( DOFObjectIDDomain_GetObjectID(DOFSecurityScope_GetDomainID(OALChannel_GetSecurityScope(data->oap->blackboardData->associatedChannel, TRUE))),
                                                      DOFObjectIDDomain_GetObjectID(DOFSecurityScope_GetDomainID(data->oap->blackboardData->associatedChannel->securityScope)) ) == 0 );

        data->secScope = DOFSecurityMembrane_ValidateInboundPermissions(data->oap->blackboardData->associatedChannel->securityMembrane,
                                                                        data->operation->commandPerm, permToCheck, data->oap->blackboardData->keyState);
        if(data->secScope != NULL)
        {
            if ( isSameDomain )
            {
                havePerms = DOFSecurityScope_Intersect(data->secScope, scope);
            }
            else
            {
                havePerms = ((DOFSecurityScope_GetIDsCount(data->secScope) > 0) && (OALPermissionScope_GetIdCount(scope) > 0));
            }
        }
        else
        {
            ExtendPermContext context = (ExtendPermContext) DOFMem_Alloc(sizeof(struct ExtendPermContext_t));

            if(context != NULL)
            {
                context->result = DOFResult_Copy(data->result);
                if((data->result == NULL) || (context->result != NULL))
                {
                    context->scope = OALPermissionScope_Copy(scope);
                    if(context->scope != NULL)
                    {
                        context->op = (DOFOperation_Retain(data->operation), data->operation);
                        context->isCommand = data->oap->blackboardData->flags.isCommandNotResponse;
                        context->isDelayed = data->opDelay;
                        context->channel = OALChannel_Retain(data->oap->blackboardData->associatedChannel);
                        OALChannel_ExtendInboundMembranePermissions(data->oap->blackboardData->associatedChannel, permToCheck, OAP_V1_Receive_extendPermCallback, context);
                    }
                    else
                    {
                        DOFResult_Destroy(context->result);
                        DOFMem_Free(context);
                    }
                }
                else
                {
                    DOFMem_Free(context);
                }
            }
            OALPermissionScope_Destroy(scope);
            data->status = OAP_V1_ReceiveStatus_DONOTHING;
            DOFPermission_Destroy(permToCheck);
            return FALSE;
        }
        if(!havePerms)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_INFO, "OAP V1: Rejecting incoming OAP request (permission denied)");
            data->status = OAP_V1_ReceiveStatus_DONOTHING;
        }
        else
        {
            DOFSecurityScope_Destroy(data->operation->securityScope);
            data->operation->securityScope = DOFSecurityScope_Copy(data->secScope);
        }
        OALPermissionScope_Destroy(scope);
        DOFPermission_Destroy(permToCheck);
        return havePerms;
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_INFO, "OAP V1: Rejecting incoming OAP request (permission denied)");
    data->status = OAP_V1_ReceiveStatus_DONOTHING;
    DOFPermission_Destroy(permToCheck);
    return FALSE;
}

boolean OAP_V1_Receive_IterateFunctions(OAP_V1_ReceiveData data)
{
    uint8 i = 0;

    while(data->functions[i] != NULL)
    {
        if(! data->functions[i++](data))
        {
            PCR_Assert(data->status != OAP_V1_ReceiveStatus_OK);
            return FALSE;
        }
        PCR_Assert(data->status == OAP_V1_ReceiveStatus_OK);
    }
    return TRUE;
}

boolean OAP_V1_Receive_VerifyOpID(OAP_V1_ReceiveData data)
{
    if(data->oap->blackboardData->opID != NULL)
    {
        return TRUE;
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, verify opID: Received null opid?" );
    data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
    return FALSE;
}

boolean OAP_V1_Receive_VerifyOperation(OAP_V1_ReceiveData data)
{
    if(data->operation != NULL)
    {
        if(data->operation->operation_type < DOFOPERATIONTYPE_INVALID)
        {
            return TRUE;
        }
        else
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, verify operation: received a non-OAP operation." );
            data->status = OAP_V1_ReceiveStatus_NOOPERATION;    /* Should this be a fatal error? */
        }
    }
    else
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, verify operation: expected an operation but didn't find it." );
        data->status = OAP_V1_ReceiveStatus_NOOPERATION;
    }
    return FALSE;
}

boolean OAP_V1_Receive_ReadBinding(OAP_V1_ReceiveData data)
{
    if(! OAP_V1_Receive_ReadAlias(data))
    {
        return FALSE;
    }
    if(data->alias != 0)
    {
        if(data->oap->blackboardData->flags.isMulticast)
        {
            data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
            return FALSE;
        }
        else
        {
            provider_binding_table_entry_t *pbte = DOFRouter_Alias_to_pbte(OALChannel_GetSecurityScope(data->oap->blackboardData->associatedChannel, TRUE),
                                                                           data->aliasAddress, data->alias);
            if(pbte != NULL)
            {
                data->objectID = DOFObjectID_Copy(pbte->provider_object_id);
                if(data->objectID != NULL)
                {
                    data->interfaceID = DOFInterfaceID_Copy(pbte->interface_id);
                    if(data->interfaceID != NULL)
                    {
                        data->requestType = OAP_V1_UNICAST;
                        provider_binding_table_entry_release(pbte);
                        return TRUE;
                    }
                }
                provider_binding_table_entry_release(pbte);
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "OAP V1 receive, read binding: Failed to copy the oid and iid from the pbte (out of memory)." );
                data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
                return FALSE;
            }
            data->status = OAP_V1_ReceiveStatus_INVALID_BINDING;
        }
    }
    else
    {
        return OAP_V1_Receive_ReadFullBinding(data);
    }
    return FALSE;
}

static uint8 OAP_V1_ReceiveData_DetermineAliasSize(OAP_V1_ReceiveData self)
{
    switch(self->firstByte >> 6)
    {
    case 0:
    case 1:
    case 2:
        return (self->firstByte >> 6);
        break;

    case 3:
        return 4;
        break;

    default:
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive data, determine alias size: invalid size." );
        self->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
        return 0xFF;
    }
}

boolean OAP_V1_Receive_ReadAlias(OAP_V1_ReceiveData data)
{
    uint8 aliasSize = OAP_V1_ReceiveData_DetermineAliasSize(data);

    switch(aliasSize)
    {
    case 1:
        {
            uint8 ret;
            if(! OALStream_Read8(data->stream, &ret))
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read alias: Failed to read alias8.");
                data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
                return FALSE;
            }
            data->alias = ret;
        }
        break;

    case 2:
        {
            uint16 ret;
            if(! OALStream_Read16(data->stream, &ret))
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read alias: Failed to read alias16.");
                data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
                return FALSE;
            }
            data->alias = ret;
        }
        break;

    case 4:
        if(! OALStream_Read32(data->stream, &(data->alias)))
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read alias: Failed to read alias32.");
            data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
            return FALSE;
        }
        break;

    case 0xFF:
        data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
        return FALSE;
        break;

    default:
    case 0:
        data->alias = 0;
        break;
    }
    return TRUE;
}

boolean OAP_V1_Receive_ReadFullBinding(OAP_V1_ReceiveData data)
{
    uint32 size = OALStream_GetDataSize(data->stream);

    data->interfaceID = DOFInterfaceID_Create_Unmarshal(&size, OALStream_GetReadPtr(data->stream));
    if ( data->interfaceID )
    {
        OALStream_AdjustReadOffset(data->stream, size);
        data->objectID = DOFObjectID_Unmarshal(data->stream);
        if(data->objectID)
        {
            data->requestType = OAP_V1_DetermineRequestType(data->objectID);
            return TRUE;
        }
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read full binding: Failed to read." );
    data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
    return FALSE;
}

boolean OAP_V1_Receive_ReadProviderID(OAP_V1_ReceiveData data)
{
    if(data->firstByte & 0x20)
    {
        data->providerID = DOFObjectID_Unmarshal(data->stream);
        if ( !data->providerID )
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read providerID: Failed to read." );
            data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
            return FALSE;
        }
    }
    else if ( data->operation->flags.bits.provider_oid_is_wildcard )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read providerID: Broadcast operation response does not contain Provider ID." );
        data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
        return FALSE;
    }
    else
    {
        data->providerID = DOFObjectID_Copy( data->operation->object_id.object_id );
        if ( !data->providerID )
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read providerID: Failed to copy." );
            data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
            return FALSE;
        }
    }
    return TRUE;
}

boolean OAP_V1_Receive_ReadInterfaceID(OAP_V1_ReceiveData data)
{
    uint32 size = OALStream_GetDataSize(data->stream);
    data->interfaceID = DOFInterfaceID_Create_Unmarshal(&size, OALStream_GetReadPtr(data->stream));
    if (data->interfaceID)
    {
        OALStream_AdjustReadOffset(data->stream, size);
        return TRUE;
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read interfaceID: Failed to read." );
    data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
    return FALSE;
}

boolean OAP_V1_Receive_ReadResponseControl(OAP_V1_ReceiveData data)
{
    if(data->firstByte & 0x20)
    {
        uint8 buffer;

        PCR_Assert( data->opConfig );
        if(!OALStream_Read8(data->stream, &buffer))
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read response control: Failed to read response control flags." );
            data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
            return FALSE;
        }
        DOFOperationControl_SetResponseLevel( data->opConfig, (buffer & 0x30) >> 4 );
        /* Hueristic */
        if(buffer & 0x01)
        {
            uint16 val;

            if(OALMarshal_StreamUncompress2(data->stream, &val) != DOFERROR_OK)
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read response control: Failed to unmarshal heuristic." );
                data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
                return FALSE;
            }
            DOFOperationControl_SetHeuristic( data->opConfig, (uint16) val );
        }
        /* No Execute */
        else if ( buffer & 0x08 )
        {
            DOFOperationControl_SetHeuristic( data->opConfig, 0 );
        }
        else
        {
            DOFOperationControl_SetHeuristic( data->opConfig, DOFOPERATIONCONTROL_HEURISTIC_ALL );
        }
        /* Delay */
        DOFOperationControl_SetDelay( data->opConfig, (buffer & 0x02) );
        /* Ack List */
        if(buffer & 0x04)
        {
            uint8 i;
            if(!OALStream_Read8(data->stream, &buffer))
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read response control: Failed to unmarshal ack size." );
                data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
                return FALSE;
            }
            if(buffer > 0)
            {
                OALAVLTree tree;
                if(data->operation == NULL)
                {
                    tree = data->ackTree = DOFOperation_CreateObjectIDAVLTree();
                }
                else
                {
                    DOFOperation_Lock( data->operation );
                    if(data->operation->ackTree == NULL)
                        data->operation->ackTree = DOFOperation_CreateObjectIDAVLTree();
                    tree = data->ackTree = data->operation->ackTree;
                }
                if ( !tree )
                {
                    data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
                    return FALSE;
                }
                for(i = 0; i < buffer; ++i)
                {
                    DOFObjectID objectID = DOFObjectID_Unmarshal(data->stream);

                    if ( objectID )
                    {
                        if(DOFOperationControl_AddProviderExclusion( data->opConfig, objectID ))
                        {
                            if ( OALAVLTree_Insert(tree, objectID, objectID, NULL) != OALAVLTree_InsertSuccess )
                            {
                                DOFObjectID_Destroy( objectID );
                            }
                        }
                        else
                        {
                            DOFObjectID_Destroy(objectID);
                        }
                        data->opDelay = FALSE;
                    }
                    else
                    {
                        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read response control: Failed to unmarshal ack oid." );
                        data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
                        return FALSE;
                    }
                }
                if ( data->operation )
                    DOFOperation_Unlock( data->operation );
            }
        }
        /* Cache Delay */
        if(buffer & 0x40)
        {
            uint16 val;
            if(OALMarshal_StreamUncompress2(data->stream, &val) != DOFERROR_OK)
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read response control: Failed to unmarshal time limit for cached values." );
                data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
                return FALSE;
            }
            DOFOperationControl_SetMaxCacheAge( data->opConfig, (uint32)1000U * val );
        }
    }
    return TRUE;
}

boolean OAP_V1_Receive_ReadItemID(OAP_V1_ReceiveData data)
{
    uint16 fullItemId;

    if(OALMarshal_StreamUncompress2(data->stream, &fullItemId) != DOFERROR_OK)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read itemID: Failed to read." );
        data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
        return FALSE;
    }
    data->itemID = (uint16) fullItemId;
    return TRUE;
}

boolean OAP_V1_Receive_ReadMinimumDelta(OAP_V1_ReceiveData data)
{
    if( OALMarshal_StreamUncompress2(data->stream, &(data->minimumDelta)) != DOFERROR_OK )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read minimum delta: Failed to read." );
        data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
        return FALSE;
    }
    return TRUE;
}

boolean OAP_V1_Receive_ReadSequenceNumber(OAP_V1_ReceiveData data)
{
    uint16 sequenceNumber;

    if(OALStream_Read16(data->stream, &sequenceNumber))
    {
        data->sequenceNumber = sequenceNumber;
        return TRUE;
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read sequence number: Failed to read." );
    data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
    return FALSE;
}

boolean OAP_V1_Receive_ReadLosslessCount(OAP_V1_ReceiveData data)
{
    if(OALStream_GetDataSize(data->stream) > 0)
    {
        DOFError error = OALMarshal_StreamUncompress2(data->stream, &(data->losslessCount));

        if(error != DOFERROR_OK)
        {
            data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
            return FALSE;
        }
    }
    return TRUE;
}

boolean OAP_V1_Receive_ReadLossyCount(OAP_V1_ReceiveData data)
{
    if(OALStream_GetDataSize(data->stream) > 0)
    {
        DOFError error = OALMarshal_StreamUncompress2(data->stream, &(data->lossyCount));

        if(error != DOFERROR_OK)
        {
            data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
            return FALSE;
        }
    }
    return TRUE;
}

boolean OAP_V1_Receive_ReadSessionGUID(OAP_V1_ReceiveData data)
{
    uint8 attribute, length;

    if ( OALStream_Read8( data->stream, &attribute ) && OALStream_Read8( data->stream, &length ) )
    {
        if ( attribute == DOFOBJECTIDATTRIBUTE_SESSION && length == sizeof(GUID) )
        {
            if ( OALStream_Read( data->stream, sizeof(GUID), data->sessionGUID.bytes ) )
            {
                return TRUE;
            }
        }
        else
        {
            data->status = OAP_V1_ReceiveStatus_NOT_SUPPORTED;
            return FALSE;
        }
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read sessionGUID: Failed to read." );
    data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
    return FALSE;
}

boolean OAP_V1_Receive_ReadSessionIID(OAP_V1_ReceiveData data)
{
    DOFInterfaceID sessionIID = DOFInterfaceID_Unmarshal(data->stream);
    if (sessionIID)
    {
        data->sessionIID = sessionIID;
        return TRUE;
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OAP V1 receive, read sessionIID: Failed to read." );
    data->status = OAP_V1_ReceiveStatus_STREAM_ERROR;
    return FALSE;
}

boolean OAP_V1_Receive_CreateEmptyResult(OAP_V1_ReceiveData data)
{
    DOFObjectID pid = (data->providerID != NULL) ? data->providerID
                                                 : data->operation->object_id.object_id;

    data->result = DOFResult_NoValue_Create(data->oap->blackboardData->associatedChannel,
                                            data->oap->blackboardData->transportAddr,
                                            DOFERROR_OK, pid);
    if(data->result != NULL)
    {
        return TRUE;
    }
    data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
    return FALSE;
}

boolean OAP_V1_Receive_CreateValueResult(OAP_V1_ReceiveData data)
{
    DOFObjectID pid = (data->providerID != NULL) ? data->providerID :
                                                   data->operation->object_id.object_id;

    data->result = DOFResult_Stream_Create(data->oap->blackboardData->associatedChannel,
                                           data->oap->blackboardData->transportAddr,
                                           DOFERROR_OK, pid, data->stream,
                                           DOFRESULTTYPE_VALUE_STREAM);
    if(data->result != NULL)
    {
        return TRUE;
    }
    data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
    return FALSE;
}

boolean OAP_V1_Receive_CreateValueListResult(OAP_V1_ReceiveData data)
{
    DOFObjectID pid = (data->providerID != NULL) ?
                      data->providerID :
                      data->operation->object_id.object_id;

    data->result = DOFResult_Stream_Create(data->oap->blackboardData->associatedChannel,
                                           data->oap->blackboardData->transportAddr,
                                           DOFERROR_OK, pid, data->stream,
                                           DOFRESULTTYPE_VALUELIST_STREAM);
    if(data->result != NULL)
    {
        return TRUE;
    }
    data->status = OAP_V1_ReceiveStatus_MEMORY_ERROR;
    return FALSE;
}

static void OAP_V1_Receive_extendPermCallback(DOFError   status,
                                              void     * context)
{
    ExtendPermContext realContext = (ExtendPermContext) context;

    PCR_Assert(context != NULL);
    PCR_Assert(realContext->op != NULL);
    PCR_Assert(realContext->scope != NULL);
    PCR_Assert(realContext->channel != NULL);
    if(status == DOFERROR_OK)
    {
        /* The keyState only matters if we have a cipher. */
        OALCipherKeyState keyState = (realContext->channel->remoteChannelData->cipher) ? OALCipher_GetCurrentKeyState( realContext->channel->remoteChannelData->cipher ) : NULL;
        DOFSecurityScope secScope =
            DOFSecurityMembrane_ValidateInboundPermissions(realContext->channel->securityMembrane,
                                                            (realContext->result != NULL) ? DOFResult_GetPermission(realContext->result)
                                                                                          : realContext->op->commandPerm, realContext->op->commandPerm,
                                                            keyState);

        if(secScope != NULL)
        {
            boolean hasPerms = TRUE;
            boolean isSameDomain =
                DOFObjectID_Compare( DOFObjectIDDomain_GetObjectID(DOFSecurityScope_GetDomainID(realContext->channel->securityScope)),
                                     DOFObjectIDDomain_GetObjectID(DOFSecurityScope_GetDomainID(secScope)) ) == 0;

            if ( isSameDomain )
            {
                hasPerms = DOFSecurityScope_Intersect(secScope, realContext->scope);
            }
            else
            {
                hasPerms = (OALPermissionScope_GetIdCount((DOFSecurityScope_GetOALPermissionScope(secScope))) > 0);
            }
            if( hasPerms )
            {
                DOFSecurityScope newScope = DOFSecurityScope_Copy(secScope);

                DOFOperation_Lock(realContext->op);
                DOFSecurityScope_Destroy(realContext->op->securityScope);
                realContext->op->securityScope = newScope;
                DOFOperation_Unlock(realContext->op);
                if(realContext->result != NULL)
                {
                    DOFRouter_RouteToRequestor(realContext->op, realContext->result);
                }
                else
                {
                    DOFRouter_RouteToProvider( realContext->op );
                }
            }
            else
            {
                /* Permission denied */
            }
            DOFSecurityScope_Destroy(secScope);
        }
        else
        {
            /* Memory failure? */
        }
    }
    else
    {
        /* Report error? */
    }
    OALChannel_Release(realContext->channel);
    DOFOperation_Internal_Destroy(realContext->op);
    DOFResult_Destroy(realContext->result);
    OALPermissionScope_Destroy(realContext->scope);
    DOFMem_Free(context);
}
