/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#if OAP_DEBUG_OUTPUT
#include <stdio.h>
#endif

#include <dof/exception.h>
#include <internal/securityscope.h>
#include <dof/permissionset.h>
#include <dof/value.h>
#include <internal/error.h>
#include <internal/exception.h>
#include <internal/interfaceidinternal.h>
#include <internal/marshal.h>
#include <internal/oap_v1.h>
#include <internal/operationconfiginternal.h>
#include <internal/operationinternal.h>
#include <internal/resultinternal.h>
#include <internal/serverinternal.h>
#include <internal/stream.h>
#include <internal/systeminternal.h>

#include "oap_v1_sendutil.h"

#define OAP_V1_ADVERTISE_OPCODE      5
#define OAP_V1_ACTIVATE_OPCODE      28
#define OAP_V1_CHANGE_OPCODE         2
#define OAP_V1_CONNECT_OPCODE        4
#define OAP_V1_DEFINE_OPCODE         6
#define OAP_V1_EXCEPTION_OPCODE      9
#define OAP_V1_FULLCONNECT_OPCODE    3
#define OAP_V1_GET_OPCODE           10
#define OAP_V1_INVOKE_OPCODE        12
#define OAP_V1_SESSION_OPCODE       14
#define OAP_V1_PROVIDE_OPCODE       16
#define OAP_V1_REGISTER_OPCODE      25
#define OAP_V1_SET_OPCODE           20
#define OAP_V1_SIGNAL_OPCODE        22
#define OAP_V1_SUBSCRIBE_OPCODE     24
#define OAP_V1_WATCH_OPCODE         30

static void OAP_V1_SendInvalidOpType(OAP_V1_SendData data);

static void OAP_V1_Send_sendInterestPDU(OAP_V1_SendData data);

static void OAP_V1_SendAdvertise(OAP_V1_SendData data);

static void OAP_V1_SendActivate(OAP_V1_SendData data);

static void OAP_V1_SendCancel(OAP_V1_SendData data);

static void OAP_V1_SendChange(OAP_V1_SendData data);

static void OAP_V1_SendConnect(OAP_V1_SendData data);

static void OAP_V1_SendDefine(OAP_V1_SendData data);

static void OAP_V1_SendDefineComplete(OAP_V1_SendData data);

static void OAP_V1_SendException(OAP_V1_SendData data);

static void OAP_V1_SendGet(OAP_V1_SendData data);

static void OAP_V1_SendGetComplete(OAP_V1_SendData data);

static void OAP_V1_SendInvoke(OAP_V1_SendData data);

static void OAP_V1_SendInvokeComplete(OAP_V1_SendData data);

static void OAP_V1_SendNoOp(OAP_V1_SendData data);

static void OAP_V1_SendOpen(OAP_V1_SendData data);

static void OAP_V1_SendOpenAccepted(OAP_V1_SendData data);

static void OAP_V1_SendProvide(OAP_V1_SendData data);

static void OAP_V1_SendRegister(OAP_V1_SendData data);

static void OAP_V1_SendRegisterAccepted(OAP_V1_SendData data);

static void OAP_V1_SendSet(OAP_V1_SendData data);

static void OAP_V1_SendSetComplete(OAP_V1_SendData data);

static void OAP_V1_SendSignal(OAP_V1_SendData data);

static void OAP_V1_SendSubscribe(OAP_V1_SendData data);

static void OAP_V1_SendSubscribeAccepted(OAP_V1_SendData data);

static void OAP_V1_SendWatch(OAP_V1_SendData data);

static boolean OAP_V1_Send_OkToForwardCancel(DOFOperation op);

ProtocolLayerResult OAP_V1_Send(OAP_V1       self,
                                DOFOperation operation,
                                boolean      isProviderOperation,       /* TRUE if going to the provider */
                                DOFResult    result)
{
    DOFOperationType opType;

    OAP_V1_SendData_t sendData;

    OAP_V1_RequestType requestType = OAP_V1_UNICAST;

    /* These functions are ordered by operation type.  See operationinternal.h */
    static const OAP_V1_SendFunction sendFunctions[DOFOPERATIONTYPE_INVALID][2] = {
        /* DOFOPERATIONTYPE_GET */
            {OAP_V1_SendGetComplete,        OAP_V1_SendGet},
        /* DOFOPERATIONTYPE_SET */
            {OAP_V1_SendSetComplete,        OAP_V1_SendSet},
        /* DOFOPERATIONTYPE_INVOKE */
            {OAP_V1_SendInvokeComplete,     OAP_V1_SendInvoke},
        /* DOFOPERATIONTYPE_REGISTER */
            {OAP_V1_SendRegisterAccepted,   OAP_V1_SendRegister},
        /* DOFOPERATIONTYPE_SUBSCRIBE */
            {OAP_V1_SendSubscribeAccepted,  OAP_V1_SendSubscribe},
        /* DOFOPERATIONTYPE_FINDINTERFACEDEFINITION */
            {OAP_V1_SendDefineComplete,     OAP_V1_SendDefine},
        /* DOFOPERATIONTYPE_WATCH */
            {OAP_V1_SendInvalidOpType,      OAP_V1_SendWatch},
        /* DOFOPERATIONTYPE_ACTIVATE */
            {OAP_V1_SendInvalidOpType,      OAP_V1_SendActivate},
        /* DOFOPERATIONTYPE_EXCEPTION */
            {OAP_V1_SendException,          OAP_V1_SendInvalidOpType},
        /* DOFOPERATIONTYPE_BEGINPROVIDE */
            {OAP_V1_SendProvide,            OAP_V1_SendProvide},
        /* DOFOPERATIONTYPE_PROVIDE */
            {OAP_V1_SendProvide,            OAP_V1_SendProvide},
        /* DOFOPERATIONTYPE_SIGNAL */
            {OAP_V1_SendSignal,             OAP_V1_SendInvalidOpType},
        /* DOFOPERATIONTYPE_CHANGED */
            {OAP_V1_SendChange,             OAP_V1_SendInvalidOpType},
        /* DOFOPERATIONTYPE_CANCEL */
            {OAP_V1_SendCancel,             OAP_V1_SendCancel},
        /* DOFOPERATIONTYPE_CONNECT */
            {OAP_V1_SendInvalidOpType,      OAP_V1_SendConnect},
        /* DOFOPERATIONTYPE_SESSION */
            {OAP_V1_SendOpenAccepted,       OAP_V1_SendOpen},
        /* DOFOPERATIONTYPE_ADVERTISE */
            {OAP_V1_SendInvalidOpType,      OAP_V1_SendAdvertise},
        /* DOFOPERATIONTYPE_QUERY */
            {OAP_V1_SendInvalidOpType,      OAP_V1_SendInvalidOpType},
        /* DOFOPERATIONTYPE_FULLCONNECT */
            {OAP_V1_SendInvalidOpType,      OAP_V1_SendConnect},
        /* DOFOPERATIONTYPE_NOOP */
            {OAP_V1_SendNoOp,               OAP_V1_SendNoOp},
    };

    PCR_Assert(self);
    PCR_Assert(operation != NULL);
    PCR_Assert(operation->operation_type != DOFOPERATIONTYPE_QUERY);

    DOFOperation_Lock(operation);
    if(operation->flags.bits.isFlooded && !result) /* Only commands can be flooded. */
    {
        self->blackboardData->flags.isFlooded = TRUE;
    }
    if(operation->flags.bits.oap_cancel && !result) /* Only commands can be cancelled. */
    {
        if(operation->flags.bits.internally_created &&
           (!OAP_V1_Send_OkToForwardCancel(operation)))
        {
            DOFOperation_Unlock(operation);
            return PLRESULT_OKDONOTHING;
        }
        opType = DOFOPERATIONTYPE_CANCEL;
    }
    else if(operation->operation_type == DOFOPERATIONTYPE_NOOP)
    {
        opType = DOFOPERATIONTYPE_NOOP;
    }
    else
    {
        if(operation->operation_type == DOFOPERATIONTYPE_EXCEPTION)
        {
            requestType = OAP_V1_DetermineRequestType(operation->object_id.object_id);
        }
        if(result && (DOFResult_GetException(result)))
        {
            opType = DOFOPERATIONTYPE_EXCEPTION;
        }
        else
        {
            opType = operation->operation_type;
        }
    }
    DOFOperation_Unlock(operation);
    /* Among other things, this next fills sendData->result with result. */
    OAP_V1_SendData_Init(&sendData, self, operation, result, requestType);
    /* Mark blackboardData so outgoing EPP header in the wire will include this flag or not, then send sendData. */
    self->blackboardData->flags.commandCompletesOnAnyFinal = operation->flags.bits.commandCompletesOnAnyFinal;
    sendFunctions[opType][isProviderOperation](&sendData);
    switch(sendData.status)
    {
    case OAP_V1_SendStatus_OK:
        PCR_Assert(self->blackboardData->flags.isCommandNotResponse || !(self->blackboardData->flags.isMulticast));
        /* The operation ID and sequence should be based on the forwarded operation, but everything else should be the real operation. */
        if ( self->blackboardData->flags.isCommandNotResponse )
        {
            if ( ( (uint16) operation->sources->sequence + operation->flags.bits.internally_created ) >= 256 )
            {
                /* Cannot forward packets with a sequence > 255. */
                DOFSecurityScope_Destroy(sendData.secScope);
                return PLRESULT_OKDONOTHING;
            }
            self->blackboardData->seq = operation->sources->sequence + operation->flags.bits.internally_created;
            self->blackboardData->retry = operation->sources->retry_count;
            if(operation->forwardingId != NULL)
            {
                self->blackboardData->opID = DOFOperationID_Copy( operation->forwardingId );
                if ( !operation->flags.bits.isFlooded && !operation->sources->sequence )
                    self->blackboardData->seq = 0;
            }
            else
            {
                self->blackboardData->opID = DOFOperationID_Copy( operation->opID );
            }
        }
        else
        {
            self->blackboardData->opID = DOFOperationID_Copy( operation->opID );
        }
        self->blackboardData->flags.isFlooded = operation->flags.bits.isFlooded;
        /* Fall through */

    case OAP_V1_SendStatus_OKNOOPID:
        if(self->blackboardData->flags.isCommandNotResponse && opType != DOFOPERATIONTYPE_CANCEL && opType != DOFOPERATIONTYPE_NOOP)
        {
            if ( operation->flags.bits.internally_created )
            {
                self->blackboardData->delay = operation->sources->oal_timeout;
            }
            else
            {
                self->blackboardData->delay = DOFOperation_GetTimeRemaining( operation );
            }
        }
        DOFSecurityScope_Destroy(sendData.secScope);
        return PLRESULT_OKSENDDOWN;
        break;

    case OAP_V1_SendStatus_DONOTSEND:
        DOFSecurityScope_Destroy(sendData.secScope);
        return PLRESULT_OKDONOTHING;
        break;

    default:
    case OAP_V1_SendStatus_NOTENOUGHVALUES:
    case OAP_V1_SendStatus_INVALID_OPCODE:
        return PLRESULT_ERROR;
        break;
    }

    return PLRESULT_OKDONOTHING;
}

static void OAP_V1_SendInvalidOpType(OAP_V1_SendData data)
{
    data->status = OAP_V1_SendStatus_INVALID_OPCODE;
}

static void OAP_V1_SendAdvertise(OAP_V1_SendData data)
{
    static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                  OAP_V1_Send_WriteFirstByte,
                                                  OAP_V1_Send_WriteFullBinding, NULL};

    data->opCode = OAP_V1_ADVERTISE_OPCODE;
    data->functions = functions;
    data->oap->blackboardData->flags.isCommandNotResponse = TRUE;
    OAP_V1_Send_IterateFunctions(data);
}

static void OAP_V1_SendActivate(OAP_V1_SendData data)
{
    data->opCode = OAP_V1_ACTIVATE_OPCODE;
    OAP_V1_Send_sendInterestPDU(data);
}

static void OAP_V1_SendCancel(OAP_V1_SendData data)
{
    OALStream stream = OALStream_CreateRW(OALMarshal_GetCompressedValueSize(2, OAP_V1_APPID));

    if(stream != NULL)
    {
        if(OALMarshal_StreamCompress2(stream, OAP_V1_APPID) == DOFERROR_OK)
        {
            if(OALPacket_AppendStream(data->packet, stream))
            {
                data->oap->blackboardData->flags.isCommandNotResponse = TRUE;
            }
            else
            {
                OALStream_Release(stream, FALSE);
                data->status = OAP_V1_SendStatus_BADALLOC;
            }
        }
        else
        {
            OALStream_Release(stream, FALSE);
            data->status = OAP_V1_SendStatus_STREAMERROR;
        }
    }
    else
    {
        data->status = OAP_V1_SendStatus_BADALLOC;
    }
}

static void OAP_V1_SendChange(OAP_V1_SendData data)
{
    static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                  OAP_V1_Send_WriteFirstByte,
                                                  OAP_V1_Send_WriteItemID,
                                                  OAP_V1_Send_WriteBinding,
                                                  OAP_V1_Send_WriteSequenceNumber, NULL};

    data->opCode = OAP_V1_CHANGE_OPCODE;
    data->functions = functions;
    data->oap->blackboardData->flags.isCommandNotResponse = TRUE;
    PCR_Assert(data->valueStream != NULL);
    if(OAP_V1_Send_IterateFunctions(data))
    {
        OALStream_Retain(data->valueStream);
        if(!OALPacket_AppendStream(data->packet, data->valueStream))
        {
            OALStream_Release(data->valueStream, TRUE);
            data->status = OAP_V1_SendStatus_BADALLOC;
        }
        else
        {
            data->status = OAP_V1_SendStatus_OKNOOPID;
        }
    }
    else
    {
        data->status = OAP_V1_SendStatus_DONOTSEND;
    }
}

static void OAP_V1_SendConnect(OAP_V1_SendData data)
{
    PCR_Assert(DOFObjectID_IsUnicast(data->operation->object_id.object_id));
    PCR_Assert(!DOFInterfaceID_IsEqual(data->operation->interface_id, DOFINTERFACEID_WILDCARD));
    data->opCode = OAP_V1_CONNECT_OPCODE;
    if (!data->oap->blackboardData->associatedChannel->remoteChannelData->flags.isOutbound)
    {
        OALChannel baseChannel = data->oap->blackboardData->associatedChannel->remoteChannelData->baseChannel;
        if (baseChannel && !baseChannel->remoteChannelData->flags.isConnection &&
            OALServer_HaveStartedStreamServer(baseChannel->remoteChannelData->address))
        {
            data->opCode = OAP_V1_FULLCONNECT_OPCODE;
        }
    }
    OAP_V1_Send_sendInterestPDU(data);
}

static void OAP_V1_SendDefine(OAP_V1_SendData data)
{
    if(data->oap->blackboardData->flags.isMulticast)
    {
#if OAP_DEBUG_OUTPUT
        fprintf(stderr, "OAP Rejecting Define - Cannot be send on a multicast channel\n");
#endif
        data->status = OAP_V1_SendStatus_DONOTSEND;
    }
    else
    {
        static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                      OAP_V1_Send_WriteFirstByte,
                                                      OAP_V1_Send_WriteFullBinding, NULL};

        data->opCode = OAP_V1_DEFINE_OPCODE;
        data->functions = functions;
        data->oap->blackboardData->flags.isCommandNotResponse = TRUE;
        OAP_V1_Send_IterateFunctions(data);
    }
}

static void OAP_V1_SendDefineComplete(OAP_V1_SendData data)
{
    if(!data->oap->blackboardData->flags.isMulticast)
    {
        static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_WriteFirstByte, NULL};

        data->opCode = OAP_V1_DEFINE_OPCODE;
        data->functions = functions;
        data->oap->blackboardData->flags.isCommandNotResponse = FALSE;
        if(OAP_V1_Send_IterateFunctions(data))
        {
            OALStream_Retain(data->valueStream);
            if(!OALPacket_AppendStream(data->packet, data->valueStream))
            {
                OALStream_Release(data->valueStream, TRUE);
                data->status = OAP_V1_SendStatus_BADALLOC;
            }
        }
    }
    else
    {
        data->status = OAP_V1_SendStatus_DONOTSEND;
    }
}

static void OAP_V1_SendException(OAP_V1_SendData data)
{
    PCR_Assert(DOFResult_GetException(data->result) != NULL);

    if(data->oap->blackboardData->flags.isMulticast)
    {
#if OAP_DEBUG_OUTPUT
        fprintf(stderr, "OAP Rejecting Exception - Cannot be send on a multicast channel\n");
#endif
        data->status = OAP_V1_SendStatus_DONOTSEND;
    }
    else
    {
        static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                      OAP_V1_Send_WriteFirstByte,
                                                      OAP_V1_Send_WritePID,
                                                      OAP_V1_Send_WriteExceptionError, NULL};

        if(data->verbosity < 2)
        {
            goto earlyExit;
        }
        data->oap->blackboardData->flags.multicastProtocolNotAllowed = TRUE;
        data->opCode = OAP_V1_EXCEPTION_OPCODE;
        data->firstByteFlags |= OAP_V1_SENDDATA_INTERNAL | OAP_V1_SENDDATA_FINAL | OAP_V1_SENDDATA_PID;
        data->functions = functions;
        data->oap->blackboardData->flags.isCommandNotResponse = FALSE;
        if(OAP_V1_Send_IterateFunctions(data))
        {
            if (data->valueStream != NULL)
            {
                OALStream_Retain(data->valueStream);
                if(!OALPacket_AppendStream(data->packet, data->valueStream))
                {
                    OALStream_Release(data->valueStream, TRUE);
                    data->status = OAP_V1_SendStatus_BADALLOC;
                    return;
                }
            }
        }
        return;

    earlyExit:
        data->status = OAP_V1_SendStatus_DONOTSEND;
    }
}

static void OAP_V1_SendGet(OAP_V1_SendData data)
{
    static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                  OAP_V1_Send_WriteFirstByte,
                                                  OAP_V1_Send_WriteResponseControl,
                                                  OAP_V1_Send_WriteItemID,
                                                  OAP_V1_Send_WriteBinding, NULL};

    data->opCode = OAP_V1_GET_OPCODE;
    data->firstByteFlags |= OAP_V1_SENDDATA_ALIAS | OAP_V1_SENDDATA_FLAGS;
    data->commandControlFlags |= OAP_V1_COMMANDCONTROL_NOEXECUTE | OAP_V1_COMMANDCONTROL_VERBOSITY |
                                 OAP_V1_COMMANDCONTROL_HEURISTIC | OAP_V1_COMMANDCONTROL_DELAY |
                                 OAP_V1_COMMANDCONTROL_CACHE | OAP_V1_COMMANDCONTROL_ACKLIST;
    data->functions = functions;
    data->oap->blackboardData->flags.isCommandNotResponse = TRUE;
    OAP_V1_Send_IterateFunctions(data);
}

static void OAP_V1_SendGetComplete(OAP_V1_SendData data)
{
    if(!data->oap->blackboardData->flags.isMulticast && (data->verbosity >= 1))
    {
        static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                      OAP_V1_Send_WriteFirstByte,
                                                      OAP_V1_Send_WritePID, NULL};

        data->opCode = OAP_V1_GET_OPCODE;
        data->firstByteFlags = OAP_V1_SENDDATA_PID;
        data->functions = functions;
        data->oap->blackboardData->flags.isCommandNotResponse = FALSE;
        PCR_Assert(data->valueStream != NULL);
        if(OAP_V1_Send_IterateFunctions(data))
        {
            OALStream_Retain(data->valueStream);
            if(!OALPacket_AppendStream(data->packet, data->valueStream))
            {
                OALStream_Release(data->valueStream, TRUE);
                data->status = OAP_V1_SendStatus_BADALLOC;
            }
        }
    }
    else
    {
        data->status = OAP_V1_SendStatus_DONOTSEND;
    }
}

static void OAP_V1_SendInvoke(OAP_V1_SendData data)
{
    static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                  OAP_V1_Send_WriteFirstByte,
                                                  OAP_V1_Send_WriteResponseControl,
                                                  OAP_V1_Send_WriteItemID, OAP_V1_Send_WriteBinding, NULL};

    data->opCode = OAP_V1_INVOKE_OPCODE;
    data->firstByteFlags |= OAP_V1_SENDDATA_ALIAS | OAP_V1_SENDDATA_FLAGS;
    data->commandControlFlags |= OAP_V1_COMMANDCONTROL_NOEXECUTE | OAP_V1_COMMANDCONTROL_VERBOSITY |
                                 OAP_V1_COMMANDCONTROL_HEURISTIC | OAP_V1_COMMANDCONTROL_DELAY |
                                 OAP_V1_COMMANDCONTROL_ACKLIST;
    data->functions = functions;
    data->oap->blackboardData->flags.isCommandNotResponse = TRUE;
    PCR_Assert(data->valueStream != NULL);
    if(OAP_V1_Send_IterateFunctions(data))
    {
        OALStream_Retain(data->valueStream);
        if(!OALPacket_AppendStream(data->packet, data->valueStream))
        {
            OALStream_Release(data->valueStream, TRUE);
            data->status = OAP_V1_SendStatus_BADALLOC;
        }
    }
}

static void OAP_V1_SendInvokeComplete(OAP_V1_SendData data)
{
    if(!data->oap->blackboardData->flags.isMulticast && (data->verbosity >= 1))
    {
        static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                      OAP_V1_Send_WriteFirstByte,
                                                      OAP_V1_Send_WritePID, NULL};

        data->opCode = OAP_V1_INVOKE_OPCODE;
        data->firstByteFlags = OAP_V1_SENDDATA_PID;
        data->functions = functions;
        data->oap->blackboardData->flags.isCommandNotResponse = FALSE;
        PCR_Assert(data->valueStream != NULL);
        if(OAP_V1_Send_IterateFunctions(data))
        {
            OALStream_Retain(data->valueStream);
            if(!OALPacket_AppendStream(data->packet, data->valueStream))
            {
                OALStream_Release(data->valueStream, TRUE);
                data->status = OAP_V1_SendStatus_BADALLOC;
                return;
            }
        }
    }
    else
    {
        data->status = OAP_V1_SendStatus_DONOTSEND;
    }
}

static void OAP_V1_SendNoOp(OAP_V1_SendData data)
{
    OALStream stream = OALStream_CreateRW(OALMarshal_GetCompressedValueSize(2, OAP_V1_APPID));

    if(stream != NULL)
    {
        if(OALMarshal_StreamCompress2(stream, OAP_V1_APPID) == DOFERROR_OK)
        {
            if(OALPacket_AppendStream(data->packet, stream))
            {
                data->status = OAP_V1_SendStatus_OKNOOPID;
                data->oap->blackboardData->flags.isCommandNotResponse = TRUE;
            }
            else
            {
                OALStream_Release(stream, FALSE);
                data->status = OAP_V1_SendStatus_BADALLOC;
            }
        }
        else
        {
            OALStream_Release(stream, FALSE);
            data->status = OAP_V1_SendStatus_STREAMERROR;
        }
    }
    else
    {
        data->status = OAP_V1_SendStatus_BADALLOC;
    }
}

static void OAP_V1_SendOpen(OAP_V1_SendData data)
{
    if(OAP_V1_DetermineRequestType(data->operation->object_id.object_id) == OAP_V1_UNICAST)
    {
        static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                      OAP_V1_Send_WriteFirstByte,
                                                      OAP_V1_Send_WriteBinding,
                                                      OAP_V1_Send_SessionIID,
                                                      OAP_V1_Send_SessionGUID, NULL};

        data->opCode = OAP_V1_SESSION_OPCODE;
        data->firstByteFlags |= OAP_V1_SENDDATA_ALIAS;
        data->functions = functions;
        data->oap->blackboardData->flags.isCommandNotResponse = TRUE;
        if(OAP_V1_Send_IterateFunctions(data))
        {
            if(!data->operation->flags.bits.internally_created)
            {
                uint32 ackTimeout = DOFOperationControl_GetAckTimeout(data->operation->operationConfig);

                if (ackTimeout < DOF_TIMEOUT_NEVER)
                {
                    if(data->operation->ackTimerItem == NULL)
                    {
                        data->operation->ackTimerItem = OALTimer_CreateItem( OAP_V1_GetTimer(),
                                                                             DOFSystem_Handle_AckTimeout, 1,
                                                                             (void **)&(data->operation) );
                    }
                    if ((data->operation->ackTimerItem != NULL) &&
                        !OALTimerItem_IsInserted( data->operation->ackTimerItem ))
                    {
                        OALTimerItem_Insert( data->operation->ackTimerItem, ackTimeout, FALSE, FALSE );
                    }
                    else
                    { /* ack timeout already set, leave it alone. */
                    }
                }
            }
        }
    }
    else
    {
#if OAP_DEBUG_OUTPUT
        fprintf(stderr, "OAP Rejecting Open - Binding must be unicast\n");
#endif
        data->status = OAP_V1_SendStatus_DONOTSEND;
    }
}

static void OAP_V1_SendOpenAccepted(OAP_V1_SendData data)
{
    if(!data->oap->blackboardData->flags.isMulticast)
    {
        static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                      OAP_V1_Send_WriteFirstByte, NULL};

        data->opCode = OAP_V1_SESSION_OPCODE;
        data->functions = functions;
        data->oap->blackboardData->flags.isCommandNotResponse = FALSE;
        OAP_V1_Send_IterateFunctions(data);
    }
    else
    {
        data->status = OAP_V1_SendStatus_DONOTSEND;
    }
}

static void OAP_V1_SendProvide(OAP_V1_SendData data)
{
    PCR_Assert( data->operation->opID );                                   /* Provide PDU is invalid without an opID. */
    PCR_Assert(data->operation->object_id.object_id != NULL);
    PCR_Assert(data->operation->interface_id != NULL);
    if (DOFObjectID_IsWildcard(data->operation->object_id.object_id) ||
        DOFInterfaceID_IsWildcard(data->operation->interface_id))
    {
#if OAP_DEBUG_OUTPUT
        fprintf(stderr, "OAP Rejecting Provide - Binding must be non-wildcard\n");
#endif
        data->status = OAP_V1_SendStatus_DONOTSEND;
    }
    else
    {
        static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                      OAP_V1_Send_WriteFirstByte,
                                                      OAP_V1_Send_WriteAlias,
                                                      OAP_V1_Send_WriteFullBinding,
                                                      OAP_V1_Send_WriteHopCount, NULL};

        data->opCode = OAP_V1_PROVIDE_OPCODE;
        data->firstByteFlags |= OAP_V1_SENDDATA_ALIAS;
        data->functions = functions;
        data->oap->blackboardData->flags.isCommandNotResponse = TRUE;
        OAP_V1_Send_IterateFunctions(data);
    }
}

static void OAP_V1_SendRegister(OAP_V1_SendData data)
{
    static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                  OAP_V1_Send_WriteFirstByte,
                                                  OAP_V1_Send_WriteResponseControl,
                                                  OAP_V1_Send_WriteItemID,
                                                  OAP_V1_Send_WriteBinding, NULL};

    data->opCode = OAP_V1_REGISTER_OPCODE;
    data->firstByteFlags |= /* OAP_V1_SENDDATA_ALIAS | */ OAP_V1_SENDDATA_FLAGS;
    data->commandControlFlags |= OAP_V1_COMMANDCONTROL_NOEXECUTE | OAP_V1_COMMANDCONTROL_VERBOSITY |
                                 OAP_V1_COMMANDCONTROL_HEURISTIC | OAP_V1_COMMANDCONTROL_DELAY |
                                 OAP_V1_COMMANDCONTROL_ACKLIST;
    data->functions = functions;
    data->oap->blackboardData->flags.isCommandNotResponse = TRUE;
    if(OAP_V1_Send_IterateFunctions(data))
    {
        if(!data->operation->flags.bits.internally_created)
        {
            uint32 ackTimeout = DOFOperationControl_GetAckTimeout(data->operation->operationConfig);

            if (ackTimeout < DOF_TIMEOUT_NEVER)
            {
                if(data->operation->ackTimerItem == NULL)
                {
                    data->operation->ackTimerItem =
                         OALTimer_CreateItem(OAP_V1_GetTimer(), DOFSystem_Handle_AckTimeout, 1, (void **)&(data->operation));
                }
                if ((data->operation->ackTimerItem != NULL) && !OALTimerItem_IsInserted(data->operation->ackTimerItem))
                    OALTimerItem_Insert( data->operation->ackTimerItem, ackTimeout, FALSE, FALSE );
                else
                { /* ack timeout already set, leave it alone. */
                }
            }
        }
    }
}

static void OAP_V1_SendRegisterAccepted(OAP_V1_SendData data)
{
    if(!data->oap->blackboardData->flags.isMulticast && (data->verbosity >= 1))
    {
        static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                      OAP_V1_Send_WriteFirstByte,
                                                      OAP_V1_Send_WritePID,
                                                      OAP_V1_Send_WriteSequenceNumber, NULL};

        data->opCode = OAP_V1_REGISTER_OPCODE;
        data->firstByteFlags = OAP_V1_SENDDATA_PID;
        data->functions = functions;
        data->oap->blackboardData->flags.isCommandNotResponse = FALSE;
        OAP_V1_Send_IterateFunctions(data);
    }
    else
    {
        data->status = OAP_V1_SendStatus_DONOTSEND;
    }
}

static void OAP_V1_SendSet(OAP_V1_SendData data)
{
    static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                  OAP_V1_Send_WriteFirstByte,
                                                  OAP_V1_Send_WriteResponseControl,
                                                  OAP_V1_Send_WriteItemID,
                                                  OAP_V1_Send_WriteBinding, NULL};

    data->opCode = OAP_V1_SET_OPCODE;
    data->firstByteFlags |= OAP_V1_SENDDATA_ALIAS | OAP_V1_SENDDATA_FLAGS;
    data->commandControlFlags |= OAP_V1_COMMANDCONTROL_NOEXECUTE | OAP_V1_COMMANDCONTROL_VERBOSITY |
                                 OAP_V1_COMMANDCONTROL_HEURISTIC | OAP_V1_COMMANDCONTROL_DELAY |
                                 OAP_V1_COMMANDCONTROL_ACKLIST;
    data->functions = functions;
    data->oap->blackboardData->flags.isCommandNotResponse = TRUE;
    if(OAP_V1_Send_IterateFunctions(data))
    {
        OALStream_Retain(data->valueStream);
        if(!OALPacket_AppendStream(data->packet, data->valueStream))
        {
            OALStream_Release(data->valueStream, TRUE);
            data->status = OAP_V1_SendStatus_BADALLOC;
        }
    }
}

static void OAP_V1_SendSetComplete(OAP_V1_SendData data)
{
    if(!data->oap->blackboardData->flags.isMulticast && (data->verbosity >= 1))
    {
        static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                      OAP_V1_Send_WriteFirstByte,
                                                      OAP_V1_Send_WritePID, NULL};

        data->opCode = OAP_V1_SET_OPCODE;
        data->firstByteFlags = OAP_V1_SENDDATA_PID;
        data->functions = functions;
        data->oap->blackboardData->flags.isCommandNotResponse = FALSE;
        OAP_V1_Send_IterateFunctions(data);
    }
    else
    {
        data->status = OAP_V1_SendStatus_DONOTSEND;
    }
}

static void OAP_V1_SendSignal(OAP_V1_SendData data)
{
    static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                  OAP_V1_Send_WriteFirstByte,
                                                  OAP_V1_Send_WriteItemID,
                                                  OAP_V1_Send_WriteBinding,
                                                  OAP_V1_Send_WriteSequenceNumber, NULL};

    data->opCode = OAP_V1_SIGNAL_OPCODE;
    data->firstByteFlags |= /* OAP_V1_SENDDATA_ALIAS | */ 0;
    data->functions = functions;
    data->oap->blackboardData->flags.isCommandNotResponse = TRUE;
    if(OAP_V1_Send_IterateFunctions(data))
    {
        if(data->valueStream != NULL)
        {
            OALStream_Retain(data->valueStream);
            if(!OALPacket_AppendStream(data->packet, data->valueStream))
            {
                OALStream_Release(data->valueStream, TRUE);
                data->status = OAP_V1_SendStatus_BADALLOC;
                return;
            }
        }
        if(! data->oap->isConnectionless)
        {
            data->status = OAP_V1_SendStatus_OKNOOPID;
        }
    }
}

static void OAP_V1_SendSubscribe(OAP_V1_SendData data)
{
    static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                  OAP_V1_Send_WriteFirstByte,
                                                  OAP_V1_Send_WriteResponseControl,
                                                  OAP_V1_Send_WriteItemID,
                                                  OAP_V1_Send_WriteBinding,
                                                  OAP_V1_Send_WriteMinimumDelta, NULL};

    data->opCode = OAP_V1_SUBSCRIBE_OPCODE;
    data->firstByteFlags |= /* OAP_V1_SENDDATA_ALIAS | */ OAP_V1_SENDDATA_FLAGS;
    data->commandControlFlags |= OAP_V1_COMMANDCONTROL_NOEXECUTE | OAP_V1_COMMANDCONTROL_VERBOSITY |
                                 OAP_V1_COMMANDCONTROL_HEURISTIC | OAP_V1_COMMANDCONTROL_DELAY |
                                 OAP_V1_COMMANDCONTROL_ACKLIST;
    data->functions = functions;
    data->oap->blackboardData->flags.isCommandNotResponse = TRUE;
    if(OAP_V1_Send_IterateFunctions(data))
    {
        if(!data->operation->flags.bits.internally_created)
        {
            uint32 ackTimeout = DOFOperationControl_GetAckTimeout(data->operation->operationConfig);

            if (ackTimeout < DOF_TIMEOUT_NEVER)
            {
                if(data->operation->ackTimerItem == NULL)
                {
                    data->operation->ackTimerItem =
                         OALTimer_CreateItem(OAP_V1_GetTimer(), DOFSystem_Handle_AckTimeout, 1, (void **)&(data->operation));
                }
                if ((data->operation->ackTimerItem != NULL) && !OALTimerItem_IsInserted(data->operation->ackTimerItem))
                    OALTimerItem_Insert( data->operation->ackTimerItem, ackTimeout, FALSE, FALSE );
                else
                { /* ack timeout already set, leave it alone. */
                }
            }
        }
    }
}

static void OAP_V1_SendSubscribeAccepted(OAP_V1_SendData data)
{
    if(!data->oap->blackboardData->flags.isMulticast && (data->verbosity >= 1))
    {
        static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                      OAP_V1_Send_WriteFirstByte,
                                                      OAP_V1_Send_WritePID,
                                                      OAP_V1_Send_WriteSequenceNumber, NULL};

        data->opCode = OAP_V1_SUBSCRIBE_OPCODE;
        data->firstByteFlags = OAP_V1_SENDDATA_PID;
        data->functions = functions;
        data->oap->blackboardData->flags.isCommandNotResponse = FALSE;
        PCR_Assert(data->valueStream != NULL);
        if(OAP_V1_Send_IterateFunctions(data))
        {
            OALStream_Retain(data->valueStream);
            if(!OALPacket_AppendStream(data->packet, data->valueStream))
            {
                OALStream_Release(data->valueStream, TRUE);
                data->status = OAP_V1_SendStatus_BADALLOC;
                return;
            }
        }
    }
    else
    {
        data->status = OAP_V1_SendStatus_DONOTSEND;
    }
}

static void OAP_V1_SendWatch(OAP_V1_SendData data)
{
    data->opCode = OAP_V1_WATCH_OPCODE;
    OAP_V1_Send_sendInterestPDU(data);
}

static void OAP_V1_Send_sendInterestPDU(OAP_V1_SendData data)
{
    static const OAP_V1_SendHelper functions[] = {OAP_V1_Send_VerifyPermissions,
                                                  OAP_V1_Send_WriteFirstByte,
                                                  OAP_V1_Send_WriteResponseControl,
                                                  OAP_V1_Send_WriteBinding, NULL};

    data->firstByteFlags |= OAP_V1_SENDDATA_ALIAS | OAP_V1_SENDDATA_FLAGS;
    data->commandControlFlags |= OAP_V1_COMMANDCONTROL_HEURISTIC | OAP_V1_COMMANDCONTROL_DELAY;
    data->functions = functions;
    data->oap->blackboardData->flags.isCommandNotResponse = TRUE;
    OAP_V1_Send_IterateFunctions(data);
}

static boolean OAP_V1_Send_OkToForwardCancel(DOFOperation op)
{
    /* Called with operation mutex locked. */
    switch ( op->operation_type )
    {
        case DOFOPERATIONTYPE_CANCEL:
        case DOFOPERATIONTYPE_CHANGED:
        case DOFOPERATIONTYPE_SIGNAL:
        case DOFOPERATIONTYPE_EXCEPTION:
        case DOFOPERATIONTYPE_NOOP:
        case DOFOPERATIONTYPE_QUERY:
            return FALSE;
        default:
            break;
    }
    return (!op->flags.bits.cancelOpAfterEppSend);
}
