/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/connection.h>
#include <dof/interfaceid.h>
#include <dof/objectid.h>
#include <dof/permissionset.h>
#include <dof/providerinfo.h>
#include <dof/value.h>
#include <dof/exception.h>
#include <internal/securityscope.h>
#include <dof/server.h>
#include <internal/bridgeinternal.h>
#include <internal/exception.h>
#include <internal/interfaceinternal.h>
#include <internal/iteminternal.h>
#include <internal/marshal.h>
#include <internal/permissioninternal.h>
#include <internal/permissionsetinternal.h>
#include <internal/operationconfiginternal.h>
#include <internal/remotechannel.h>
#include <internal/resultinternal.h>
#include <internal/router.h>
#include <internal/value.h>
#include <internal/type.h>

#include "oap_v1_sendutil.h"

#define LAST_TWO_BYTE_ALIAS     0xFFFF
#define LAST_ONE_BYTE_ALIAS     0xFF
#define MAXIMUM_MINIMUM_DELTA   32767U

typedef struct ExtendPermContext_t
{
    DOFOperation op;
    DOFResult result;
    OALChannel channel;
    DOFAddress address;

    boolean isProviderBound;
} * ExtendPermContext;

static void OAP_V1_Send_extendPermCallback(DOFError   status,
                                            void     * context);

static boolean OAP_V1_Send_WriteAlias_Internal(OAP_V1_SendData data);

static boolean OAP_V1_Send_WriteInterfaceID_Internal(OAP_V1_SendData data);

static boolean OAP_V1_Send_WriteItemID_Internal(OAP_V1_SendData data);

static boolean OAP_V1_Send_WriteObjectID_Internal(OAP_V1_SendData data);

static boolean OAP_V1_Send_WriteHopCount_Internal(OAP_V1_SendData data);

static boolean OAP_V1_Send_WriteProviderID_Internal(OAP_V1_SendData data);

static boolean OAP_V1_Send_WriteResponseControl_Internal(OAP_V1_SendData data);

static boolean OAP_V1_Send_WriteMinimumDelta_Internal(OAP_V1_SendData data);

static boolean OAP_V1_Send_WriteSequenceNumber_Internal(OAP_V1_SendData data);

static boolean OAP_V1_Send_SessionGUID_Internal(OAP_V1_SendData data);

static boolean OAP_V1_Send_SessionIID_Internal(OAP_V1_SendData data);

static uint16 OAP_V1_Send_mapError(DOFError         error,
                                   DOFOperationType opType);

void OAP_V1_SendData_Init(OAP_V1_SendData    self,
                          OAP_V1             oap,
                          DOFOperation       operation,
                          DOFResult          result,
                          OAP_V1_RequestType requestType)
{
    PCR_Assert(oap);

    PCRMem_Set(self, 0, sizeof(OAP_V1_SendData_t));
    self->packet = oap->blackboardData->sendPacket;
    self->oap = oap;
    self->operation = operation;
    self->result = result;
    self->verbosity = (requestType == OAP_V1_UNICAST) ? 2 : 1;
    self->heuristic = DOFOPERATIONCONTROL_HEURISTIC_ALL;
    if(result)
    {
        DOFOperationResponseLevel verbosity;
        OALStream                 stream = DOFResult_GetStream( result );

        /*Note: A DOFResult is of types. If result is a stream type, stream is filled. Otherwise result is some other
                type. Further, a DOFResult can be transformed from stream type to non-stream type, but not the reverse*/
        if (stream)
        {
            self->valueStream = stream;
        }
        if (operation && operation->operationConfig &&
            (verbosity = DOFOperationControl_GetVerbosity(operation->operationConfig)) <= DOFOPERATIONRESPONSELEVEL_ALL)
           self->verbosity = (uint8)verbosity;
    }
    else if ( operation )
    {
        DOFOperation_Lock( operation );
        self->valueStream = operation->valueStream;
        DOFOperation_Unlock( operation );
    }
}

typedef struct ExtendPermissionContext_t
{
    DOFOperation pendingOperation;
    OALChannel channel;
    DOFAddress target;
    union {
        DOFOperation providerOperation;
        DOFResult result;
    } extraParam;
    boolean isProviderBound;
} ExtendPermissionContext_t;

static ExtendPermissionContext_t * createExtendContext(OALChannel   channel,
                                                       DOFAddress   target,
                                                       DOFOperation pendingOp,
                                                       DOFOperation providerOp,
                                                       DOFResult    result)
{
    ExtendPermissionContext_t * ret = DOFMem_Alloc(sizeof(ExtendPermissionContext_t));

    if(ret != NULL)
    {
        ret->channel = OALChannel_Retain(channel);
        if(ret->channel != NULL)
        {
            ret->pendingOperation = (DOFOperation_Retain(pendingOp), pendingOp);
            if(ret->pendingOperation != NULL)
            {
                ret->target = DOFAddress_Copy(target);
                if((ret->target != NULL) || (target == NULL))
                {
                    if(result != NULL)
                    {
                        ret->extraParam.result = DOFResult_Copy(result);
                        if(ret->extraParam.result != NULL)
                        {
                            ret->isProviderBound = FALSE;
                            return ret;
                        }
                    }
                    else
                    {
                        if(providerOp != NULL)
                        {
                            ret->extraParam.providerOperation = (DOFOperation_Retain(providerOp), providerOp);
                        }
                        else
                        {
                            ret->extraParam.providerOperation = NULL;
                        }
                        if((ret->extraParam.providerOperation != NULL) || (providerOp == NULL))
                        {
                            ret->isProviderBound = TRUE;
                            return ret;
                        }
                    }
                }
                DOFOperation_Internal_Destroy(ret->pendingOperation);
            }
            OALChannel_Release(ret->channel);
        }
        DOFMem_Free(ret);
    }
    return NULL;
}

static void extendCompleteCallback(DOFError   status,
                                   void     * context)
{
    ExtendPermissionContext_t * fullContext = (ExtendPermissionContext_t *) context;

    if(status == DOFERROR_OK)
    {
        if(fullContext->isProviderBound)
        {
            fullContext->channel->routeToProvider(fullContext->channel, fullContext->pendingOperation,
                                                  fullContext->extraParam.providerOperation, fullContext->target);
        }
        else
        {
            fullContext->channel->routeToRequestor(fullContext->channel, fullContext->pendingOperation,
                                                  fullContext->extraParam.result);
        }
    }
    DOFOperation_Internal_Destroy(fullContext->pendingOperation);
    OALChannel_Release(fullContext->channel);
    DOFAddress_Destroy(fullContext->target);
    if(fullContext->isProviderBound)
    {
        DOFOperation_Internal_Destroy(fullContext->extraParam.providerOperation);
    }
    else
    {
        DOFResult_Destroy(fullContext->extraParam.result);
    }
    DOFMem_Free(fullContext);
}

boolean OAP_V1_Send_VerifyPermissions(OAP_V1_SendData data,
                                      boolean         marshalling)
{
    if(!marshalling)
    {
        DOFPermission permToCheck = NULL;

        PCR_Assert(data != NULL);
        PCR_Assert(data->oap->blackboardData->associatedChannel->securityMembrane != NULL);

        data->secScope = OALChannel_GetSecurityScope(data->oap->blackboardData->associatedChannel, FALSE);
        if(data->oap->blackboardData->flags.isCommandNotResponse)
        {
            permToCheck = DOFPermission_Copy(data->operation->commandPerm);
        }
        else
        {
            PCR_Assert(data->result != NULL);
            permToCheck = DOFPermission_Copy(DOFResult_GetPermission(data->result));
            PCR_Assert(permToCheck != NULL);
        }
        if ( data->secScope == data->oap->blackboardData->associatedChannel->securityScope )
        {
            /* Pass-through (null) connector: Nothing to check. */
            data->secScope = DOFSecurityScope_Copy( DOF_GetDefaultSecurityScope(DOF_GetDof()) );
        }
        else
        {
            data->secScope =
                DOFSecurityMembrane_ValidateOutboundPermissions(data->oap->blackboardData->associatedChannel->securityMembrane,
                                                                 permToCheck, permToCheck);
        }
        if(data->secScope != NULL)
        {
            OALPermissionScope scope;

            if ( data->oap->blackboardData->associatedChannel->remoteChannelData->cipher != NULL )
            {
                scope =
                    DOFPermissionSet_Validate(OALCipher_GetLocalPerms( data->oap->blackboardData->associatedChannel->remoteChannelData->cipher,
                                                                       OALCipher_GetCurrentKeyState( data->oap->blackboardData->associatedChannel->remoteChannelData->cipher ) ), permToCheck);
            }
            else
            {
                scope = OALPermissionScope_Copy( DOFSecurityScope_GetOALPermissionScope( DOF_GetDefaultSecurityScope( DOF_GetDof() ) ) );
            }

            if(scope != NULL)
            {
                boolean isSameDomain = ( DOFObjectID_Compare( DOFObjectIDDomain_GetObjectID(DOFSecurityScope_GetDomainID(data->secScope)),
                                                              DOFObjectIDDomain_GetObjectID(DOFSecurityScope_GetDomainID(data->oap->blackboardData->associatedChannel->securityScope)) ) == 0 );
                boolean havePerms;
                if ( isSameDomain )
                {
                    havePerms = DOFSecurityScope_Intersect(data->secScope, scope);
                }
                else
                {
                    havePerms = ((DOFSecurityScope_GetIDsCount(data->secScope) > 0) && (OALPermissionScope_GetIdCount(scope) > 0));
                }
                if ( !havePerms )
                {
                    /* Permission denied */
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_INFO, "OAP V1: Send permission denied." );
                    data->status = OAP_V1_SendStatus_DONOTSEND;
                }
                OALPermissionScope_Destroy(scope);
                DOFPermission_Destroy(permToCheck);
                return havePerms;
            }
            else if ( (data->oap->blackboardData->associatedChannel->extendPermissions != NULL) &&
                      (data->oap->blackboardData->associatedChannel->remoteChannelData->cipher != NULL) &&
                      (data->oap->blackboardData->associatedChannel->remoteChannelData->flags.isConnection)
                            ? DOFConnectionConfig_IsPermissionsExtendAllowed(data->oap->blackboardData->associatedChannel->remoteChannelData->config.connection)
                            : TRUE)
            {
                void * context = createExtendContext(data->oap->blackboardData->associatedChannel,
                                                     data->oap->blackboardData->transportAddr,
                                                     data->operation, data->oap->blackboardData->providerOp,
                                                     data->result);

                if(context != NULL)
                {
                    OALChannel_ExtendOutboundPermissions(data->oap->blackboardData->associatedChannel, data->oap->blackboardData->transportAddr,
                                                         permToCheck, extendCompleteCallback, context);
                }
            }
            data->status = OAP_V1_SendStatus_DONOTSEND;
            DOFPermission_Destroy(permToCheck);
            return FALSE;
        }
        else
        {
            ExtendPermContext context = (ExtendPermContext) DOFMem_Alloc(sizeof(struct ExtendPermContext_t));

            if(context != NULL)
            {
                context->op = (DOFOperation_Retain(data->operation), data->operation);
                context->channel = OALChannel_Retain(data->oap->blackboardData->associatedChannel);
                context->result = DOFResult_Copy(data->result);
                if((data->result == NULL) || (context->result != NULL))
                {
                    context->isProviderBound = data->oap->blackboardData->flags.isCommandNotResponse;
                    context->address = DOFAddress_Copy( data->oap->blackboardData->transportAddr );
                    OALChannel_ExtendOutboundMembranePermissions(data->oap->blackboardData->associatedChannel, permToCheck, OAP_V1_Send_extendPermCallback, context);
                }
                else
                {
                    DOFMem_Free(context);
                }
            }
            data->status = OAP_V1_SendStatus_DONOTSEND;
            DOFPermission_Destroy(permToCheck);
            return FALSE;
        }
    }
    return TRUE;
}

boolean OAP_V1_Send_IterateFunctions(OAP_V1_SendData data)
{
    uint8 i = 0;

    while((data->functions[i] != NULL) && !data->exitEarly)
    {
        if(! data->functions[i++](data, FALSE))
        {
            PCR_Assert(data->status != OAP_V1_SendStatus_OK);
            return FALSE;
        }
        PCR_Assert(data->status == OAP_V1_SendStatus_OK);
    }
    data->stream = OALStream_CreateRW(OALMarshal_GetCompressedValueSize(2, OAP_V1_APPID) +
                                      1 + data->aliasSize + data->flagsSize + data->hopSize +
                                      data->itemIDSize + data->iIDSize + data->oIDSize + data->pIDSize +
                                      data->minimumDeltaSize + data->sequenceNumberSize +
                                      data->sessionGUIDSize + data->sessionIIDSize + data->exclusionListSize);
    if(data->stream == NULL)
    {
        data->status = OAP_V1_SendStatus_BADALLOC;
        return FALSE;
    }
    i = 0;
    while((data->functions[i] != NULL) && !data->exitEarly)
    {
        if(! data->functions[i++](data, TRUE))
        {
            OALStream_Release(data->stream, FALSE);
            return FALSE;
        }
    }
    if(!OALPacket_AppendStream(data->packet, data->stream))
    {
        data->status = OAP_V1_SendStatus_BADALLOC;
        OALStream_Release(data->stream, TRUE);
        return FALSE;
    }
    return TRUE;
}

boolean OAP_V1_Send_WriteFirstByte(OAP_V1_SendData data,
                                   boolean         marshalling)
{
    if(!marshalling)
    {
        if(data->firstByteFlags & OAP_V1_SENDDATA_ALIAS)
        {
            if(!data->oap->blackboardData->flags.isMulticast)
            {
                if(data->operation->operation_type==DOFOPERATIONTYPE_BEGINPROVIDE ||
                   data->operation->operation_type==DOFOPERATIONTYPE_PROVIDE)
                {
                    data->alias = DOFRouter_GetBindingAlias( data->operation );
                }
                else if(data->oap->blackboardData->providerOp)
                {
                    data->alias = data->oap->blackboardData->providerOp->sources->alias;
                }

                if(data->alias > LAST_TWO_BYTE_ALIAS)
                {
                    data->aliasSize = 4;
                }
                else if(data->alias > LAST_ONE_BYTE_ALIAS)
                {
                    data->aliasSize = 2;
                }
                else if(data->alias > 0x00)
                {
                    data->aliasSize = 1;
                }
                else
                {
                    data->aliasSize = 0;
                }
            }
            else
            {
                data->aliasSize = 0;
            }
            if ( data->aliasSize == 4 )
                data->firstByte |= 0xC0;
            else
                data->firstByte |= (data->aliasSize << 6);
        }
        if(data->firstByteFlags & OAP_V1_SENDDATA_INTERNAL)
        {
            if(DOFResult_IsInternal(data->result))
            {
                data->firstByte |= 0x80;
            }
        }
        if(data->firstByteFlags & OAP_V1_SENDDATA_FINAL)
        {
            if(/* isFinal */ TRUE)
            {
                data->firstByte |= (1 << 6);
            }
        }
        if(data->firstByteFlags & OAP_V1_SENDDATA_FLAGS)
        {
            data->firstByte |= 0x20;
        }
        if(data->operation && (data->firstByteFlags & OAP_V1_SENDDATA_PID))
        {
            DOFProviderInfo pinfo = DOFResult_GetProviderInfo( data->result );
            DOFOperation_Lock( data->operation );
            if(data->operation->flags.bits.provider_oid_is_wildcard &&
               pinfo &&
               ((data->providerID = DOFProviderInfo_GetProviderID(pinfo)) != NULL))
            {
                data->firstByte |= (data->operation->flags.bits.provider_oid_is_wildcard << 5);
            }
            DOFOperation_Unlock( data->operation );
        }
        data->firstByte |= data->opCode;
        return TRUE;
    }
    else
    {
        return ((OALMarshal_StreamCompress2(data->stream, OAP_V1_APPID) == DOFERROR_OK) &&
                OALStream_Write8(data->stream, data->firstByte));
    }
}

boolean OAP_V1_Send_WritePID(OAP_V1_SendData data,
                             boolean         marshalling)
{
    if(!marshalling)
    {
        if(data->providerID)
        {
            data->pIDSize = DOFObjectID_GetBytesSize(data->providerID);
        }
        return TRUE;
    }
    else
    {
        return OAP_V1_Send_WriteProviderID_Internal(data);
    }
}

boolean OAP_V1_Send_WriteBinding(OAP_V1_SendData data,
                                 boolean         marshalling)
{
    if(data->aliasSize > 0)
    {
        if(!marshalling)
        {
            data->oIDSize = 0;
            data->iIDSize = 0;

            data->interfaceID = NULL;
            data->objectID = NULL;

            return TRUE;
        }
        else
        {
            return OAP_V1_Send_WriteAlias_Internal(data);
        }
    }
    return OAP_V1_Send_WriteFullBinding(data, marshalling);
}

boolean OAP_V1_Send_WriteAlias(OAP_V1_SendData data,
                               boolean         marshalling)
{
    if(marshalling)
    {
        return OAP_V1_Send_WriteAlias_Internal(data);
    }
    return TRUE;
}

boolean OAP_V1_Send_WriteFullBinding(OAP_V1_SendData data,
                                     boolean         marshalling)
{
    if(!marshalling)
    {
        data->oIDSize = DOFObjectID_GetBytesSize(data->operation->object_id.object_id);
        data->iIDSize = DOFInterfaceID_GetBytesSize(data->operation->interface_id);
        data->interfaceID = data->operation->interface_id;
        data->objectID = data->operation->object_id.object_id_ref;

        return TRUE;
    }
    else
    {
        if(OAP_V1_Send_WriteInterfaceID_Internal(data))
        {
            return OAP_V1_Send_WriteObjectID_Internal(data);
        }
        return FALSE;
    }
}

boolean OAP_V1_Send_WriteResponseControl(OAP_V1_SendData data,
                                         boolean         marshalling)
{
    if(!marshalling)
    {
        uint16  heuristic = DOFOPERATIONCONTROL_HEURISTIC_ALL;
        uint16  cacheDelay = 0;
        uint8   verbosity = DEFAULT_VERBOSITY;
        boolean delayed = FALSE;

        if (data->operation->operationConfig)
        {
            if(data->commandControlFlags & OAP_V1_COMMANDCONTROL_ACKLIST)
            {
                for(data->ackListCount = 0; data->ackListCount < MAX_ACK; ++(data->ackListCount))
                {
                    data->acks[data->ackListCount] = DOFOperationControl_PopExclusionQueueHead(data->operation->operationConfig);
                    if(data->acks[data->ackListCount] == NULL)
                        break;
                    data->exclusionListSize += DOFObjectID_GetBytesSize(data->acks[data->ackListCount]);
                }
            }

            verbosity = DOFOperationControl_GetVerbosity(data->operation->operationConfig);
            heuristic = DOFOperationControl_GetHeuristic(data->operation->operationConfig);
            cacheDelay = DOFOperationControl_GetMaxCacheAge_Sec(data->operation->operationConfig);
            delayed = DOFOperationControl_IsDelay(data->operation->operationConfig);
        }

        if(data->exclusionListSize != 0)
        {
            if(data->ackListCount == MAX_ACK)
            {
                data->noExecute = TRUE;
            }
            data->flagsSize = 2;
        }
        else
        {
            if((verbosity != DEFAULT_VERBOSITY) && (verbosity != data->verbosity))
            {
                data->verbosity = verbosity;
                data->flagsSize = 1;
            }

            if(data->commandControlFlags & OAP_V1_COMMANDCONTROL_HEURISTIC)
            {
                if(heuristic != DOFOPERATIONCONTROL_HEURISTIC_ALL)
                {
                    if(data->flagsSize == 0)
                    {
                        data->flagsSize = 1;
                    }
                    if ( heuristic )
                    {
                        data->flagsSize += OALMarshal_GetCompressedValueSize(2, heuristic);
                        data->heuristic = heuristic;
                    }
                    else
                    {
                        data->noExecute = TRUE;
                    }
                }
            }
            if(data->commandControlFlags & OAP_V1_COMMANDCONTROL_CACHE)
            {
                if(cacheDelay != 0)
                {
                    if(data->flagsSize == 0)
                    {
                        data->flagsSize = 1;
                    }
                    data->flagsSize += OALMarshal_GetCompressedValueSize(2, cacheDelay);
                    data->cacheDelay = cacheDelay;
                }
            }
            if(data->commandControlFlags & OAP_V1_COMMANDCONTROL_DELAY)
            {
                if(delayed)
                {
                    if(data->flagsSize == 0)
                    {
                        data->flagsSize = 1;
                    }
                    data->isDelayed = TRUE;
                }
            }
        }

        /* Flags bit is on by default, so we'll clear it if we don't need it */
        if(data->flagsSize == 0)
        {
            data->firstByte &= 0xDF;
        }
        return TRUE;
    }
    else
    {
        return OAP_V1_Send_WriteResponseControl_Internal(data);
    }
}

boolean OAP_V1_Send_WriteItemID(OAP_V1_SendData data,
                                boolean         marshalling)
{
    if(!marshalling)
    {
        if(data->operation->item != NULL) {
            data->itemID = DOFItem_GetID(data->operation->item);
        }
        else data->itemID = data->operation->itemId;

        data->itemIDSize = OALMarshal_GetCompressedValueSize( 2, data->itemID );
        return TRUE;
    }
    else
    {
        return OAP_V1_Send_WriteItemID_Internal(data);
    }
}

boolean OAP_V1_Send_WriteMinimumDelta(OAP_V1_SendData data,
                                      boolean         marshalling)
{
    if(!marshalling)
    {
        uint32 minPeriod = data->operation->minPeriod;
        if ( minPeriod > MAXIMUM_MINIMUM_DELTA )
            minPeriod = MAXIMUM_MINIMUM_DELTA;
        data->minimumDeltaSize = OALMarshal_GetCompressedValueSize( 2, minPeriod );
        return TRUE;
    }
    else
    {
        return OAP_V1_Send_WriteMinimumDelta_Internal(data);
    }
}

boolean OAP_V1_Send_WriteSequenceNumber(OAP_V1_SendData data,
                                        boolean         marshalling)
{
    if(!marshalling)
    {
        data->sequenceNumberSize = 2;
        return TRUE;
    }
    else
    {
        return OAP_V1_Send_WriteSequenceNumber_Internal(data);
    }
}

boolean OAP_V1_Send_WriteHopCount(OAP_V1_SendData data,
                                  boolean         marshalling)
{
    if(!marshalling)
    {
        if(data->operation->sources->distLossy != 0)
        {
            data->hopSize = OALMarshal_GetCompressedValueSize(2, data->operation->sources->distLossy) +
                            OALMarshal_GetCompressedValueSize(2, data->operation->sources->distLossless);
        }
        else if(data->operation->sources->distLossless != 0)
        {
            data->hopSize = OALMarshal_GetCompressedValueSize(2, data->operation->sources->distLossless);
        }
        return TRUE;
    }
    else
    {
        return OAP_V1_Send_WriteHopCount_Internal(data);
    }
}

boolean OAP_V1_Send_SessionGUID(OAP_V1_SendData data,
                                boolean         marshalling)
{
    if(!marshalling)
    {
        data->sessionGUIDSize = sizeof(GUID) + 2;
        data->sessionGUID = data->operation->sessionID;

        return TRUE;
    }
    else
    {
        return OAP_V1_Send_SessionGUID_Internal(data);
    }
}

boolean OAP_V1_Send_SessionIID(OAP_V1_SendData data,
                               boolean         marshalling)
{
    if(!marshalling)
    {
        data->sessionIIDSize = DOFInterfaceID_GetBytesSize(data->operation->sessionIID);
        data->sessionIID = data->operation->sessionIID;

        return TRUE;
    }
    else
    {
        return OAP_V1_Send_SessionIID_Internal(data);
    }
}

static uint16 OAP_V1_Send_mapError(DOFError         error,
                                   DOFOperationType opType)
{
    static DOFError const LAST_ERROR = 0x0A;
    if(error & 0x8000)
    {
        switch(error)
        {
        case DOFERROR_NOT_READABLE:
        case DOFERROR_NOT_WRITABLE:
            return DOFERROR_NOT_SUPPORTED;

        case DOFERROR_RESULT_LOST:
            return DOFERROR_INSUFFICIENT_RESOURCES;

        default:        /* Application did something bad */
        case DOFERROR_FAILED:
        case DOFERROR_TIMEOUT:
        case DOFERROR_PROTOCOL_NEGOTIATION_FAILED:
        case DOFERROR_TRANSPORT_FAILED:
        case DOFERROR_IO_FAILED:
        case DOFERROR_AUTHENTICATION_FAILED:
            return DOFERROR_APPLICATION_ERROR;
        }
    }
    else if(error <= LAST_ERROR)
    {
        return error;
    }
    else
    {
        return DOFERROR_NOT_SUPPORTED;      /* generic error */
    }
}

boolean OAP_V1_Send_WriteExceptionError(OAP_V1_SendData data,
                                        boolean         marshalling)
{
    if(!marshalling)
    {
        DOFException e = DOFResult_GetException(data->result);

        if(e != NULL)
        {
            switch (DOFException_GetType(e))
            {
            case DOFEXCEPTIONTYPE_ERROR:
                if ( data->operation )
                {
                    data->itemID = OAP_V1_Send_mapError(DOFErrorException_GetError(e), data->operation->operation_type);
                }
                else
                {
                    data->itemID = DOFErrorException_GetError(e);
                }
                break;
            case DOFEXCEPTIONTYPE_PROVIDER:
                if(DOFException_IsStream(e))
                {
                    data->itemID = DOFStreamException_GetItemId(e);
                }
                else
                {
                    data->itemID = DOFItem_GetID(DOFProviderException_GetInterfaceException(e));
                }
                break;
            default:
                data->itemID = DOFERROR_FAILED;
                break;
            }
            data->itemIDSize = OALMarshal_GetCompressedValueSize(2, data->itemID);
            return TRUE;
        }
        data->status = OAP_V1_SendStatus_DONOTSEND;     /* better status? */
        return FALSE;
    }
    else
    {
        return OAP_V1_Send_WriteItemID_Internal(data);
    }
}

static boolean OAP_V1_Send_WriteInterfaceID_Internal(OAP_V1_SendData data)
{
    if(data->interfaceID != NULL)
    {
        if(!OALStream_Write(data->stream, data->iIDSize, DOFInterfaceID_GetBytes(data->interfaceID)))
        {
            data->status = OAP_V1_SendStatus_STREAMERROR;
            return FALSE;
        }
    }
    return TRUE;
}

static boolean OAP_V1_Send_WriteObjectID_Internal(OAP_V1_SendData data)
{
    if(data->objectID != NULL)
    {
        if(!OALStream_Write(data->stream, data->oIDSize, DOFObjectID_GetBytes(data->objectID)))
        {
            data->status = OAP_V1_SendStatus_STREAMERROR;
            return FALSE;
        }
    }
    return TRUE;
}

static boolean OAP_V1_Send_WriteHopCount_Internal(OAP_V1_SendData data)
{
    if(data->hopSize > 0)
    {
        if(OALMarshal_StreamCompress2(data->stream, data->operation->sources->distLossless) != DOFERROR_OK)
        {
            data->status = OAP_V1_SendStatus_STREAMERROR;
            return FALSE;
        }
        if(data->operation->sources->distLossy > 0)
        {
            if(OALMarshal_StreamCompress2(data->stream, data->operation->sources->distLossy) != DOFERROR_OK)
            {
                data->status = OAP_V1_SendStatus_STREAMERROR;
                return FALSE;
            }
        }
    }
    return TRUE;
}

static boolean OAP_V1_Send_WriteProviderID_Internal(OAP_V1_SendData data)
{
    if(data->providerID != NULL)
    {
        if(!OALStream_Write(data->stream, data->pIDSize, DOFObjectID_GetBytes(data->providerID)))
        {
            data->status = OAP_V1_SendStatus_STREAMERROR;
            return FALSE;
        }
    }
    return TRUE;
}

static boolean OAP_V1_Send_WriteResponseControl_Internal(OAP_V1_SendData data)
{
    if(data->flagsSize != 0)
    {
        uint8 flagsByte = ((data->verbosity & 0x03) << 4);

        if((data->commandControlFlags & OAP_V1_COMMANDCONTROL_HEURISTIC) &&
           data->heuristic != DOFOPERATIONCONTROL_HEURISTIC_ALL)
        {
            flagsByte |= 0x01;
        }
        if((data->commandControlFlags & OAP_V1_COMMANDCONTROL_DELAY) &&
           data->isDelayed)
        {
            flagsByte |= 0x02;
        }
        if(data->exclusionListSize != 0)
        {
            flagsByte |= 0x04;
        }
        if((data->commandControlFlags & OAP_V1_COMMANDCONTROL_NOEXECUTE)
           && data->noExecute)
        {
            flagsByte |= 0x08;
        }
        if((data->commandControlFlags & OAP_V1_COMMANDCONTROL_CACHE) && (data->cacheDelay != 0))
        {
            flagsByte |= 0x40;
        }

        if(!OALStream_Write8(data->stream, flagsByte))
        {
            data->status = OAP_V1_SendStatus_STREAMERROR;
            return FALSE;
        }
        if(flagsByte & 0x01)
        {
            if(OALMarshal_StreamCompress2(data->stream, data->heuristic) != DOFERROR_OK)
            {
                data->status = OAP_V1_SendStatus_STREAMERROR;
                return FALSE;
            }
        }
        if(data->exclusionListSize != 0)
        {
            uint8 i;

            if(!OALStream_Write8(data->stream, data->ackListCount))
            {
                data->status = OAP_V1_SendStatus_STREAMERROR;
                return FALSE;
            }
            for(i = 0; i < data->ackListCount; ++i)
            {
                if(!DOFObjectID_Marshal(data->acks[i], data->stream))
                {
                    data->status = OAP_V1_SendStatus_STREAMERROR;
                    return FALSE;
                }
                DOFObjectID_Destroy(data->acks[i]);
                data->acks[i] = NULL;
            }
        }
        if(flagsByte & 0x40)
        {
            DOFError err = OALMarshal_StreamCompress2(data->stream, data->cacheDelay);

            if(err != DOFERROR_OK)
            {
                data->status = OAP_V1_SendStatus_STREAMERROR;
                return FALSE;
            }
        }
    }
    return TRUE;
}

static boolean OAP_V1_Send_WriteItemID_Internal(OAP_V1_SendData data)
{
    if(data->itemIDSize > 0)
    {
        if(OALMarshal_StreamCompress2(data->stream, data->itemID) != DOFERROR_OK)
        {
            data->status = OAP_V1_SendStatus_STREAMERROR;
            return FALSE;
        }
        /* Fall through */
    }
    return TRUE;
}

static boolean OAP_V1_Send_WriteAlias_Internal(OAP_V1_SendData data)
{
    switch(data->aliasSize)
    {
    case 4:
        if(OALStream_Write32(data->stream, data->alias))
        {
            return TRUE;
        }
        break;

    case 2:
        {
            uint16 alias = (uint16) data->alias;

            if(OALStream_Write16(data->stream, alias))
            {
                return TRUE;
            }
        }

    case 1:
        {
            uint8 alias = (uint16) data->alias;

            if(OALStream_Write8(data->stream, alias))
            {
                return TRUE;
            }
        }

    default:
        return TRUE;
    }
    data->status = OAP_V1_SendStatus_STREAMERROR;
    return FALSE;
}

static boolean OAP_V1_Send_WriteMinimumDelta_Internal(OAP_V1_SendData data)
{
    if(data->minimumDeltaSize != 0)
    {
        uint16 minPeriod;

        if ( data->operation->minPeriod > MAXIMUM_MINIMUM_DELTA )
            minPeriod = MAXIMUM_MINIMUM_DELTA;
        else
            minPeriod = (uint16)data->operation->minPeriod;
        if(OALMarshal_StreamCompress2(data->stream, minPeriod)!=DOFERROR_OK)
        {
            data->status = OAP_V1_SendStatus_STREAMERROR;
            return FALSE;
        }
    }
    return TRUE;
}

static boolean OAP_V1_Send_WriteSequenceNumber_Internal(OAP_V1_SendData data)
{
    if(data->sequenceNumberSize != 0)
    {
        uint16 sequenceNumber;

        sequenceNumber = DOFResult_GetSequenceNumber(data->result);
        if(!OALStream_Write16(data->stream, sequenceNumber))
        {
            data->status = OAP_V1_SendStatus_STREAMERROR;
            return FALSE;
        }
    }
    return TRUE;
}

static boolean OAP_V1_Send_SessionGUID_Internal(OAP_V1_SendData data)
{
    if(data->sessionGUIDSize)
    {
        static const uint8 header[] = { DOFOBJECTIDATTRIBUTE_SESSION, sizeof(GUID) };
        if ( !OALStream_Write(data->stream, 2, header) ||
             !OALStream_Write(data->stream, sizeof(GUID), data->sessionGUID.bytes) )
        {
            data->status = OAP_V1_SendStatus_STREAMERROR;
            return FALSE;
        }
    }
    return TRUE;
}

static boolean OAP_V1_Send_SessionIID_Internal(OAP_V1_SendData data)
{
    if(data->sessionIID != NULL)
    {
        if(!OALStream_Write(data->stream, data->sessionIIDSize, DOFInterfaceID_GetBytes(data->sessionIID)))
        {
            data->status = OAP_V1_SendStatus_STREAMERROR;
            return FALSE;
        }
    }
    return TRUE;
}

static void OAP_V1_Send_extendPermCallback(DOFError   status,
                                           void     * context)
{
    ExtendPermContext realContext = (ExtendPermContext) context;

    PCR_Assert(context != NULL);
    PCR_Assert(realContext->op != NULL);
    PCR_Assert(realContext->channel != NULL);
    if(status == DOFERROR_OK)
    {
        if(realContext->isProviderBound)
        {
            realContext->channel->routeToProvider(realContext->channel,
                                                  realContext->op,
                                                  NULL,
                                                  realContext->address);
        }
        else
        {
            realContext->channel->routeToRequestor(realContext->channel,
                                                   realContext->op,
                                                   realContext->result);
        }
    }
    else if((realContext->op->callbacks.pcb != NULL) && (realContext->op->callbacks.pcb->fns != NULL) && (realContext->op->callbacks.pcb->fns->Complete != NULL))
    {
        /* Report error? */
    }
    DOFOperation_Internal_Destroy(realContext->op);
    DOFAddress_Destroy(realContext->address);
    DOFResult_Destroy(realContext->result);
    OALChannel_Release(realContext->channel);
    DOFMem_Free(context);
}
