/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/bridge.h>
#include <dof/core.h>
#include <dof/item.h>
#include <dof/interface.h>
#include <dof/interfaceid.h>
#include <dof/object.h>
#include <dof/objectid.h>
#include <dof/operation.h>
#include <dof/permissionset.h>
#include <dof/result.h>
#include <dof/exception.h>
#include <dof/error.h>
#include <dof/type.h>
#include <internal/securityscope.h>
#include <dof/value.h>
#include <dof/system.h>
#include <internal/bridgeinternal.h>
#include <internal/exception.h>
#include <internal/interfaceinternal.h>
#include <internal/interfaceidinternal.h>
#include <internal/marshal.h>
#include <internal/objectinternal.h>
#include <internal/objectidinternal.h>
#include <internal/observable.h>
#include <internal/operationinternal.h>
#include <internal/permissioninternal.h>
#include <internal/requestinternal.h>
#include <internal/resultinternal.h>
#include <internal/router.h>
#include <internal/systeminternal.h>
#include <internal/value.h>
#include <internal/slist.h>
#include <internal/channel.h>
#include <internal/operationconfiginternal.h>
#include <internal/mutex_pool.h>

/** @addtogroup DOFObject
@{
**/

#define ADVERTISE_TIMEOUT 5000 /* 5 seconds */

extern PCRMutex ref_count_mutex;

/*
**  "Private" Type and Method Declarations
*/


/** @internal
   Observer callback to shutdown and/or disable things when a BeginProvide operation closes.

   @param o      BeginProvide observable that BeginProvideOp is observing
                 (originally stored at provOp->operation_will_close).
   @param provOp The specific provider operation that has is in the process of shutting down.
 */
static void BeginProvideOpClosing( OALObserver self, OALObservable o, void * provOp )
{
    PCR_Assert( o );
    PCR_Assert( provOp );
    PCR_Assert( ((DOFOperation)provOp)->flags.bits.cancelled ); /* Cancel already happened or we would not be called. */
    /* Note: No effort is made to figure out when the request operation has no more BeginProvide's.
             This decision made 5/23/2007, and further that it is the DOFOperation class that will figure this out
             and close out the request operation when it has no more BeginProvide operations attached. */
    /* We don't remove ourselves as observer of o (which is provOp->operation_will_close) because DOFOperation_Cancel()
       logic guarantees that provOp is not cancelled twice, and it takes a _Cancel() to notify its observers. */
    if (!((DOFOperation)provOp)->flags.bits.completedAck)
        if (!DOFSystem_AcceptRequest( DOFSystem_GetChannel( ((DOFOperation)provOp)->provider_object->sys ), NULL,
                                      (DOFOperation)provOp ))        /* NULL reqOp: This is cancel or timeout. */
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFObject: BeginProvideOpClosing() failed not notify DOFSystem." );
        }
    DOFOperation_Internal_Destroy( (DOFOperation)provOp ); /* Unbump ref count bump from setting up observer callback.*/
}


/** @internal
    This function looks at the system's object list and to see if an object with the same objectID has
    already been added.  If the object has already been added to the system then the reference count for
    the object is incremented

    @param node The node from the system list of objects.
    @param objID The object ID to search for.

    @return True if the object ID is found, and False if it is not found.
**/
static boolean DOFObject_InSysList(OALSListNode node, void* arg)
{
    sysObjectNode searchNode = (sysObjectNode)arg;
    return (DOFObjectID_IsEqual(((sysObjectNode)node)->objID, searchNode->objID) /* &&
            DOFPartition_IsEqual(((sysObjectNode)node)->partition, searchNode->partition) */);
}


/** @internal
    This function removes an object from its system's object list.

    @param obj The object that should be removed from the system list of objects
**/
static void DOFObject_RemoveFromSysList(DOFObject obj)
{
    /* Must be called with obj->sys mutex locked. */
    sysObjectNode_t searchNode;
    OALSListNode    node;

    PCR_Assert( obj );
    PCR_Assert( obj->sys );
    PCR_Assert( obj->sys->objectList );
    searchNode.objID = obj->oid;
    node = OALSList_FindAndRemove( (OALSListNode *)&obj->sys->objectList, NULL, DOFObject_InSysList, &searchNode );
    if (node)
        DOFMem_Free(node);
    else
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFObject: remove from SysList: Called with object not in list? weird" );
}


/*
**  "Internal Public"  Methods
*/


/*
**  "Public" Methods
*/


void DOFObject_Destroy( DOFObject obj )
{
    if (obj)
    {
        DOFSystem sys = obj->sys;
        PCRMutex sysMutex = OALMutexPool_Get( sys );

        PCRMutex_Lock( sysMutex );
        PCR_Assert( obj->refCount );
        obj->refCount--;
        if(!obj->refCount)
        {
            DOFObject_RemoveFromSysList(obj);   /* This must be called with sysMutex locked. */

            PCRMutex_Unlock( sysMutex );
            OALMutexPool_Release( sys );

            DOFSystem_Release( sys );
            DOFObjectID_Destroy(obj->oid);
            DOFMem_Free( obj );
        }
        else
        {
            PCRMutex_Unlock( sysMutex );
            OALMutexPool_Release( sys );
        }
    }
}


DOFObjectID DOFObject_GetObjectID( const DOFObject obj )
{
    if( !obj )
        return NULL;

    return (const DOFObjectID)obj->oid;
}


DOFSystem DOFObject_GetSystem( const DOFObject obj )
{
    if( !obj )
        return NULL;
    return obj->sys;
}


DOFObject DOFObject_Copy( DOFObject self )
{
    if (self)
    {
        boolean overflow = FALSE;
        PCRMutex sysMutex = OALMutexPool_Get( self->sys );

        PCRMutex_Lock( sysMutex );
        self->refCount++;
        if (!self->refCount)
        {
            overflow = TRUE;
            --self->refCount;   /* Fail: restore back to the max refcount value. */
        }
        PCRMutex_Unlock( sysMutex );
        OALMutexPool_Release( self->sys );

        if (!overflow)
            return self;    /* Success. */
    }

    return NULL;
}


/*
**  "Public" Methods, Provider side.
*/


boolean DOFObject_IsSignalRequired( const DOFObject obj, const DOFInterfaceEvent evt )
{
    if (obj && obj->oid)
    {
        if (evt && DOFItem_GetType(evt) == DOFITEM_EVENT )
        {
            return DOFRouter_IsEventRegistered( obj->oid, evt,
                                        OALChannel_GetSecurityScope( DOFSystem_GetChannel( DOFObject_GetSystem( obj ) ),
                                                                     TRUE ) );        /* inbound because towards dof. */
        }
    }

    return FALSE;          /* Error out: evt is illegal or its item type not an event. We know evt is not registered. */
}


boolean DOFObject_Changed( DOFObject obj, const DOFInterfaceProperty prp )
{
    boolean success = FALSE;

    if (obj && obj->oid)
    {
        if (prp && DOFItem_GetType(prp) == DOFITEM_PROPERTY && DOFInterfaceProperty_IsReadable( prp) )
        {
            if (DOFRouter_PropertyChanged( obj->oid, prp,
                                       OALChannel_GetSecurityScope( DOFSystem_GetChannel( DOFObject_GetSystem( obj ) ),
                                                                    TRUE ) ))        /* inbound because towards dof. */
                success = TRUE;
        }
        else
        { /* Error out: prp illegal, item type not property, or prp not readable, so we know prp not subscribed. */
        }
    }
    if (!success)
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFObject: DOFObject_Changed() did not send its changed report" );

    return success;
}

typedef struct PendingSignalMarshallContext_t
{
    OALEnqueueItem_t super;
    DOFInterfaceEvent event;
    DOFObject object;
    uint8 * buffer;
    uint32 bufferSize;
} * PendingSignalMarshallContext;

static void FinishPendingSignalMarshall(OALThreadPoolItem ditem,
                                        boolean           shutdown);

static PendingSignalMarshallContext PendingSignalMarshallContext_Create(DOFInterfaceEvent   event,
                                                                        DOFObject           object,
                                                                        uint8             * buffer,
                                                                        uint32              bufferSize)
{
    PendingSignalMarshallContext context = (PendingSignalMarshallContext) DOFMem_Alloc(sizeof(struct PendingSignalMarshallContext_t));

    if(context != NULL)
    {
        if(DOFInterface_Retain(DOFInterfaceEvent_GetInterface(event)))
        {
            context->event = event;
            context->object = DOFObject_Copy(object);
            if(context->object != NULL)
            {
                context->buffer = buffer;
                context->bufferSize = bufferSize;
                OALEnqueueItem_Init((OALEnqueueItem) context, FinishPendingSignalMarshall);
                return context;
            }
            DOFInterface_Destroy(DOFInterfaceEvent_GetInterface(event));
        }
        DOFMem_Free(context);
    }
    return NULL;
}

static void PendingSignalMarshallContext_Destroy(PendingSignalMarshallContext self)
{
    DOFInterface_Destroy(DOFInterfaceEvent_GetInterface(self->event));
    DOFObject_Destroy(self->object);
    PCRMem_Free(self->buffer);
    DOFMem_Free(self);
}

static void FinishPendingSignalMarshall(OALThreadPoolItem ditem,
                                        boolean           shutdown)
{
    PendingSignalMarshallContext item = (PendingSignalMarshallContext) ditem;

    /*if(item != NULL)*/ /* This is unneeded */
    {
        DOFOperation sessionOp = DOFRouter_LookupSession(DOFObject_GetObjectID(item->object));

        if(sessionOp != NULL)
        {
            uint8 * newBuffer = sessionOp->sessionTransformProvider->fns->TransformSendData(sessionOp->sessionTransformProvider,
                                                                                            DOFInterfaceEvent_GetInterfaceID(item->event),
                                                                                            &item->bufferSize, item->buffer);

            if(newBuffer != item->buffer)
            {
                PCRMem_Free(item->buffer);
                item->buffer = newBuffer;
            }
        }
        if(item->buffer != NULL)
        {
            OALStream stream = OALStream_CreateRWWithData(item->buffer, item->bufferSize);
            if(stream != NULL)
            {
                DOFResult result = DOFResult_Stream_Create(DOFSystem_GetChannel( DOFObject_GetSystem( item->object ) ), 0,
                                                            DOFERROR_OK, item->object->oid, stream, DOFRESULTTYPE_VALUELIST_STREAM );

                item->buffer = NULL;
                OALStream_Release(stream, TRUE);
                if (result)
                {
                    DOFPermission perm = _createNeededPermission( DOFOPERATIONTYPE_SIGNAL, FALSE,
                                                                DOFObject_GetObjectID( item->object ),
                                                                DOFInterfaceEvent_GetInterfaceID( item->event ) );

                    if (perm != NULL)
                    {
                        DOFResult_SetPermission( result, perm );                            /* Bumps perm's refCount. */
                        DOFPermission_Destroy( perm );
                        if(!DOFRouter_SignalEvent( item->object->oid, item->event,
                                        OALChannel_GetSecurityScope( DOFSystem_GetChannel( DOFObject_GetSystem( item->object ) ),
                                                                    TRUE ), result )) /* inbound because towards dof.*/
                        {
                            /* report failure? */
                        }
                    }
                    DOFResult_Destroy( result );
                }
            }
        }
        PendingSignalMarshallContext_Destroy(item);
    }
}

boolean DOFObject_Signal( DOFObject obj, const DOFInterfaceEvent evt, uint16 valueCount, const DOFValue value[] )
{
    boolean success = FALSE;

    if (obj && obj->oid)
    {
        if (evt && valueCount == DOFInterfaceEvent_GetParameterCount( evt ) && DOFItem_GetType(evt) == DOFITEM_EVENT)
        {
            if (!DOF_IsParameterValidation(obj->sys->dof) ||
                DOFTypeList_IsValueCompatibleWith( valueCount, DOFInterfaceEvent_GetParameterTypeList( evt ), value ))
            {
                uint8 * buffer;
                uint32 marshalSize;

                if(OALMarshal_MarshalValues((DOFValue*)value, DOFInterfaceEvent_GetParameterTypeList( evt ), valueCount, &buffer, &marshalSize))
                {
                    PendingSignalMarshallContext item = PendingSignalMarshallContext_Create(evt, obj, buffer, marshalSize);

                    if(item != NULL)
                    {
                        success = OALEnqueue_Enqueue(obj->sys->responseQueue, (OALEnqueueItem)item);
                        if(!success)
                        {
                            PendingSignalMarshallContext_Destroy(item);
                        }
                    }
                }
            }
        }
    }
    if (!success)
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFObject: Failed to signal." );

    return success;
}


typedef struct ExtendPermContext_t
{
    DOFOperation op;
    DOFSecurityMembrane secConn;
} * ExtendPermContext;


static DOFOperation CheckScopeThenRouteProviderOrComplete( DOFOperation op, OALObserver_fn_t fn )
{
    PCR_Assert( op );
    if (OALPermissionScope_GetIdCount(DOFSecurityScope_GetOALPermissionScope( op->securityScope )) != 0)
    {
        if (DOFOperation_AddOpWillCloseObserver( op, fn ) == NULL ||
            DOFRouter_RouteToProvider( op ) != DOFROUTERSTATUS_SUCCESS)
        {
            DOFOperation_Internal_Destroy( op );
            op = NULL;
        }
        else
        { /* Fall through to success out; op is not NULL. */
        }
    }
    else
        DOFSystem_CallCompleteWithException( op, DOFEXCEPTION_ACCESS_DENIED );     /* Fail out, but with non-NULL op. */

    return op;
}


static DOFOperation ExtendOutboundPermsOrDestroyOp( DOFOperation op, DOFObject obj, ExtendCompleteCB fn )
{
    ExtendPermContext context = (ExtendPermContext)DOFMem_Alloc( sizeof( struct ExtendPermContext_t ) );

    if (context != NULL)
    {
        context->secConn = DOFSecurityMembrane_Copy(
                                    (DOFSystem_GetChannel( DOFObject_GetSystem( obj ) ))->securityMembrane );
        context->op = (DOFOperation_Retain( op ), op);
        OALChannel_ExtendOutboundMembranePermissions(op->sources->channel, op->commandPerm, fn, context);
        return op;        /* fn eventually calls DOFRouter_RouteToProvider( op ) or calls op complete CB. */
    }
    return NULL;
}


static void RequestCallback( DOFError status, void * context )
{
    boolean           accessDenied = TRUE;
    ExtendPermContext realContext  = (ExtendPermContext)context;

    PCR_Assert( context != NULL );
    PCR_Assert( realContext->op != NULL );
    PCR_Assert( realContext->secConn != NULL );
    PCR_Assert( realContext->op->securityScope == NULL );
    if (status == DOFERROR_OK)
    {
        DOFSecurityScope secScope = DOFSecurityMembrane_ValidateOutboundPermissions(realContext->secConn, realContext->op->commandPerm, realContext->op->commandPerm);
        if (secScope != NULL)
        {
            if (OALPermissionScope_GetIdCount(DOFSecurityScope_GetOALPermissionScope( secScope )) != 0)
            {
                OALObserver successObserver;

                accessDenied = FALSE;
                DOFOperation_Lock( realContext->op );
                DOFSecurityScope_Destroy( realContext->op->securityScope );
                realContext->op->securityScope = secScope;
                successObserver = DOFOperation_AddOpWillCloseObserver( realContext->op,
                                                    (realContext->op->operation_type == DOFOPERATIONTYPE_BEGINPROVIDE) ?
                                                               BeginProvideOpClosing : DOFSystem_ReqOpClosingObserver );
                DOFOperation_Unlock( realContext->op );
                if (!successObserver || DOFRouter_RouteToProvider( realContext->op ) != DOFROUTERSTATUS_SUCCESS)
                    DOFSystem_CallCompleteWithException( realContext->op, DOFEXCEPTION_INSUFFICIENT_RESOURCES );
                else
                { /* Fall through to success out. */
                }
            }
            else
            {
                DOFSecurityScope_Destroy(secScope);
            }
        }
    }
    if (accessDenied)
        DOFSystem_CallCompleteWithException( realContext->op, DOFEXCEPTION_ACCESS_DENIED );
    DOFOperation_Internal_Destroy( realContext->op );
    DOFSecurityMembrane_Destroy( realContext->secConn );
    DOFMem_Free( context );
}


static void GetNeededPermCB( DOFError status, void* context )
{
    RequestCallback( status, context );
}


DOFOperation DOFObject_BeginProvide( DOFObject obj, DOFInterface iface, uint32 timeout, const DOFObjectProvider pcb,
                                     void * ctx )
{
    DOFOperation provOp = NULL;

    if (obj && DOFObjectID_IsUnicast( DOFObject_GetObjectID( obj ) ) && iface &&
        (provOp = DOFOperation_BeginProvide_Create( DOFSystem_GetChannel( DOFObject_GetSystem( obj ) ), timeout, ctx,
                                                    obj, pcb, iface )) != NULL)
    {
        /* obj passed here contains obj and sys info that DOFRouter will reference to direct incoming requests to us.
           Specifically: incoming request is matched to provOp. DOFRouter passes it to the DOFSys stored in provOp->obj.
           DOFSystem calls the proper provider callback (one of the pcb choices) and passes op to it. If the callback
           happens to be generic to work on all possible objects, it uses provOp->obj to compute which object. */
        if (provOp->securityScope != NULL)
            provOp = CheckScopeThenRouteProviderOrComplete( provOp, BeginProvideOpClosing );  /* Returns NULL or not. */
        else
        {
            /* We want the op, but cannot conclude its disposition until we go to AS to extend permission. */
            provOp = ExtendOutboundPermsOrDestroyOp( provOp, obj, GetNeededPermCB );           /* Returns NULL or not. */
        }
    }

    return provOp;
}

static void AdvertisePermCallback(DOFError status, void * context)
{
    boolean success = FALSE;
    ExtendPermContext realContext = (ExtendPermContext) context;

    PCR_Assert(context != NULL);
    PCR_Assert(realContext->op != NULL);
    PCR_Assert(realContext->secConn != NULL);
    PCR_Assert(realContext->op->securityScope == NULL);
    if(status == DOFERROR_OK)
    {
        DOFSecurityScope secScope = DOFSecurityMembrane_ValidateOutboundPermissions(realContext->secConn, realContext->op->commandPerm, realContext->op->commandPerm);
        if (secScope != NULL)
        {
            if (OALPermissionScope_GetIdCount(DOFSecurityScope_GetOALPermissionScope(secScope)) > 0)
            {
                DOFOperation_Lock( realContext->op );
                DOFSecurityScope_Destroy( realContext->op->securityScope );
                realContext->op->securityScope = secScope;
                success = (DOFRouter_RouteToProvider( realContext->op ) == DOFROUTERSTATUS_SUCCESS);
                DOFOperation_Unlock( realContext->op );
            }
            else
            {
                DOFSecurityScope_Destroy(secScope);
            }
        }
    }
    DOFOperation_Internal_Destroy(realContext->op);
    DOFSecurityMembrane_Destroy(realContext->secConn);
    DOFMem_Free(context);
    if (!success)
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFObject: Failed to deliver advertise operation." );
}

void DOFObject_Advertise( DOFObject obj, DOFInterfaceID interfaceID, DOFOperationControl operationConfig )
{
    /* There seems to be an assumption that if this fails it isn't a big deal, since the fixed public API return value
       is void. Failure simply means that the application does not get the benefit of advertising its capability to
       other DOF nodes. So we lump every error in one log message. */
    if (obj != NULL && interfaceID != NULL)
    {
        boolean      success     = FALSE;
        DOFOperation advertiseOp = DOFOperation_Advertise_CreateInternal(DOFSystem_GetChannel(DOFObject_GetSystem(obj)),
                                                                         interfaceID, DOFObject_GetObjectID(obj),
                                                                         ADVERTISE_TIMEOUT, 0, NULL);

        if (advertiseOp != NULL)
        {
            if ((advertiseOp->opID = DOFOperationID_CreateLocal()) != NULL)
            {
                if (operationConfig)
                    advertiseOp->operationConfig = DOFOperationControl_Copy( operationConfig );
                if (!operationConfig || advertiseOp->operationConfig)
                {
                    advertiseOp->flags.bits.internally_created = 0;
                    DOFSecurityScope_Destroy(advertiseOp->securityScope);
                    advertiseOp->securityScope = DOFSecurityMembrane_ValidateOutboundPermissions(
                                           advertiseOp->sources->channel->securityMembrane, advertiseOp->commandPerm, advertiseOp->commandPerm);
                    if (advertiseOp->securityScope == NULL)
                    {
                        ExtendPermContext context =
                                    (ExtendPermContext)DOFMem_Alloc( sizeof( struct ExtendPermContext_t ) );

                        if (context != NULL)
                        {
                            context->secConn = DOFSecurityMembrane_Copy(
                                    (DOFSystem_GetChannel( DOFObject_GetSystem( obj ) ))->securityMembrane );
                            if (context->secConn)
                            {
                                context->op = (DOFOperation_Retain(advertiseOp), advertiseOp);
                                OALChannel_ExtendInboundMembranePermissions(advertiseOp->sources->channel, advertiseOp->commandPerm, AdvertisePermCallback, context);
                            }
                            else
                                DOFMem_Free( context );
                        }
                    }
                    else if (OALPermissionScope_GetIdCount(DOFSecurityScope_GetOALPermissionScope(advertiseOp->securityScope)) > 0)
                        success = (DOFRouter_RouteToProvider( advertiseOp ) == DOFROUTERSTATUS_SUCCESS);
                    /* All cases: we don't return the operation; we let it timeout naturally. */
                }
            }
            DOFOperation_Internal_Destroy( advertiseOp );
        }
        if (!success)
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFObject: Failed to deliver advertise operation." );
    }
}


/*
**  "Public" Methods, Requestor side.
*/


DOFInterface DOFObject_Define( const DOFObject obj, const DOFInterfaceID interfaceID, DOFOperationControl config,
                               uint32 timeout, DOFException * pException )
{
    DOFInterface iface = NULL;

    if ( obj && interfaceID && !DOFInterfaceID_IsWildcard( interfaceID ) )
    {
        DOFOperation reqOp;

        reqOp = DOFObject_BeginDefine( obj, interfaceID, config, timeout, NULL, NULL );

        if (reqOp)
        {
            DOFException except;
            /* This next call fills *pException if a reason is found to fill it. */
            iface = DOFOperationDefine_WaitInterface( reqOp, timeout, &except );
            DOFOperation_Destroy( reqOp );
            /* reqOp now complete, with the first valid response accepted, and all other responses discarded. */
            if ( DOFErrorException_GetError( except ) == DOFERROR_TIMEOUT )
            {
                /* No valid response from reqOp by given by timeout. */
                PCR_Assert( !iface );
                if ( pException )
                    *pException = DOFException_Copy( DOFEXCEPTION_NOT_FOUND );
            }
            else if ( !iface )
            {
                /* Everything worked up to here, except the Interface couldn't actually be created. */
                if ( pException )
                    *pException = DOFException_Copy( DOFEXCEPTION_FAILED );
            }
            else
            {
                /* Success. */
                if ( pException )
                    *pException = NULL;
            }
            DOFException_Destroy( except );
            PCR_Assert( !pException || iface || *pException );
        }
        else if ( pException )
            *pException = DOFException_Copy( DOFEXCEPTION_INSUFFICIENT_RESOURCES );
    }
    else if ( pException )
        *pException = DOFException_Copy( DOFEXCEPTION_ILLEGAL_ARGUMENT );

    return iface;
}


DOFOperation DOFObject_BeginDefine( const DOFObject obj, const DOFInterfaceID interfaceID, DOFOperationControl config,
                                    uint32 timeout, const DOFObjectDefineCallback rcb, void * ctx )
{
    DOFOperation reqOp = NULL;

    if (obj && interfaceID &&
        (reqOp = DOFOperation_FindInterfaceDefinition_Create( DOFSystem_GetChannel( DOFObject_GetSystem( obj ) ),
                                                              config, timeout, ctx, obj, rcb,
                                                              interfaceID )) != NULL)
    {
        if (reqOp->securityScope != NULL)
            reqOp = CheckScopeThenRouteProviderOrComplete( reqOp, DOFSystem_ReqOpClosingObserver ); /*Ret NULL/not*/
        else
        {
            /* We want the op, but cannot conclude its disposition until we go to AS to extend permission. */
            reqOp = ExtendOutboundPermsOrDestroyOp( reqOp, obj, RequestCallback );         /* Returns NULL or not. */
        }
    }

    return reqOp;
}


DOFResult DOFObject_Get( DOFObject obj, const DOFInterfaceProperty prp, DOFOperationControl operationConfig,
                         uint32 timeout, DOFException * pException )
{
    DOFResult retResult = NULL;

    if (!obj || !obj->oid || !prp)
    {
        if (pException)
            *pException = DOFException_Copy( DOFEXCEPTION_ILLEGAL_ARGUMENT );
    }
    else
    {
        DOFOperation op = DOFObject_BeginGet( obj, prp, operationConfig, timeout, NULL, NULL );

        if (op)
        {
            /* This next call fills *pException if a reason is found to fill it. */
            retResult = DOFOperationGet_WaitResult( op, timeout, pException );
            DOFOperation_Destroy( op );
            PCR_Assert( !pException || retResult || *pException );
        }
        else if (pException)
            *pException = DOFException_Copy( DOFEXCEPTION_INSUFFICIENT_RESOURCES );
    }

    return retResult;
}


DOFResult DOFObject_Set( DOFObject obj, const DOFInterfaceProperty prp, DOFValue value,
                         DOFOperationControl operationConfig, uint32 timeout, DOFException * pException )
{
    DOFResult retResult = NULL;

    if( !obj || !obj->oid || !prp )
    {
        if ( pException )
            *pException = DOFException_Copy( DOFEXCEPTION_ILLEGAL_ARGUMENT );
    }
    else if ( DOF_IsParameterValidation(obj->sys->dof) && !DOFValue_CheckCompatible( value, DOFInterfaceProperty_GetType( prp ) ) )
    {
        if ( pException )
            *pException = DOFException_Copy( DOFEXCEPTION_TYPE_MISMATCH );
    }
    else
    {
        DOFOperation op = DOFObject_BeginSet( obj, prp, value, operationConfig, timeout, NULL, NULL );
        if (op)
        {
            retResult = DOFOperationSet_WaitResult( op, timeout, pException );
            DOFOperation_Destroy( op );
            PCR_Assert( !pException || retResult || *pException );
        }
        else if (pException)
            *pException = DOFException_Copy( DOFEXCEPTION_INSUFFICIENT_RESOURCES );
    }

    return retResult;
}


DOFResult DOFObject_Invoke( DOFObject obj, const DOFInterfaceMethod mtd, uint16 valuesCount, const DOFValue values[],
                            DOFOperationControl operationConfig, uint32 timeout, DOFException * pException )
{
    DOFResult retResult = NULL;

    if (!obj || !obj->oid || !mtd || (valuesCount && !values))
    {
        if (pException)
            *pException = DOFException_Copy( DOFEXCEPTION_ILLEGAL_ARGUMENT );
    }
    /* values can legally be NULL. This next call tests properly whether values is NULL or not. */
    else if (valuesCount != DOFInterfaceMethod_GetInputParameterCount( mtd ) ||
             (DOF_IsParameterValidation(obj->sys->dof) &&
              !DOFTypeList_IsValueCompatibleWith( valuesCount,
                                                  DOFInterfaceMethod_GetInputParameterTypeList( mtd ), values )))
    {
        if (pException)
            *pException = DOFException_Copy( DOFEXCEPTION_TYPE_MISMATCH );
    }
    else
    {
        DOFOperation op = DOFObject_BeginInvoke( obj, mtd, valuesCount, values, operationConfig, timeout, NULL, NULL );

        if (op)
        {
            retResult = DOFOperationInvoke_WaitResult( op, timeout, pException );
            DOFOperation_Destroy( op );
            PCR_Assert( !pException || retResult || *pException );
        }
        else if (pException)
            *pException = DOFException_Copy( DOFEXCEPTION_INSUFFICIENT_RESOURCES );
    }

    return retResult;
}


DOFOperation DOFObject_BeginGet( DOFObject obj, const DOFInterfaceProperty prp, DOFOperationControl operationConfig,
                                 uint32 timeout, const DOFObjectGetCallback rcb, void * ctx )
{
    DOFOperation reqOp = NULL;

    if (obj && prp && (reqOp = DOFOperation_Get_Create( DOFSystem_GetChannel( DOFObject_GetSystem( obj ) ),
                                                        operationConfig, timeout, ctx, obj, rcb, prp )) != NULL)
    {
        if (!DOFInterfaceProperty_IsReadable(prp))
            DOFSystem_CallCompleteWithException( reqOp, DOFEXCEPTION_NOT_READABLE );
        else
        {
            /* Send OAP pkt to provider to request the DOFValue corresponding to reqOp's property prp. When that
               provider sends us the expected OAP response, call requestor callback rcb. */
            PCR_Assert( reqOp->provider_object == obj);
            PCR_Assert( reqOp->item == (DOFItem)prp );
            PCR_Assert( reqOp->object_id.object_id == DOFObject_GetObjectID( obj ) );
            if (reqOp->securityScope != NULL)
                reqOp = CheckScopeThenRouteProviderOrComplete( reqOp, DOFSystem_ReqOpClosingObserver ); /*Ret NULL/not*/
            else
            {
                /* We want the op, but cannot conclude its disposition until we go to AS to extend permission. */
                reqOp = ExtendOutboundPermsOrDestroyOp( reqOp, obj, RequestCallback );         /* Returns NULL or not. */
            }
        }
    }

    return reqOp;
}

typedef struct PendingRequestMarshallContext_t
{
    OALEnqueueItem_t super;
    DOFOperation operation;
    uint8 * buffer;
    uint32 bufferSize;
} * PendingRequestMarshallContext;

static void FinishPendingRequestMarshall(OALThreadPoolItem ditem,
                                         boolean           shutdown);

static PendingRequestMarshallContext PendingRequestMarshallContext_Create(DOFOperation   operation,
                                                                          uint8        * buffer,
                                                                          uint32         bufferSize)
{
    PendingRequestMarshallContext context = (PendingRequestMarshallContext) DOFMem_Alloc(sizeof(struct PendingRequestMarshallContext_t));

    if(context != NULL)
    {
        context->operation = (DOFOperation_Retain(operation), operation);
        context->buffer = buffer;
        context->bufferSize = bufferSize;
        OALEnqueueItem_Init((OALEnqueueItem) context, FinishPendingRequestMarshall);
        return context;
    }
    return NULL;
}

static void PendingRequestMarshallContext_Destroy(PendingRequestMarshallContext self)
{
    DOFOperation_Internal_Destroy(self->operation);
    PCRMem_Free(self->buffer);
    DOFMem_Free(self);
}

static void FinishPendingRequestMarshall(OALThreadPoolItem ditem,
                                         boolean           shutdown)
{
    PendingRequestMarshallContext item = (PendingRequestMarshallContext) ditem;

    if(!shutdown)
    {
        DOFOperation sessionOp = DOFRouter_LookupSession(item->operation->object_id.object_id);

        if(sessionOp != NULL)
        {
            uint8 * newBuffer = sessionOp->sessionTransformRequestor->fns->TransformSendData(sessionOp->sessionTransformRequestor,
                                                                                             item->operation->interface_id, &item->bufferSize,
                                                                                             item->buffer);

            if(newBuffer != item->buffer)
            {
                PCRMem_Free(item->buffer);
                item->buffer = newBuffer;
            }
        }
        if(item->buffer != NULL)
        {
            item->operation->valueStream = OALStream_CreateRWWithData(item->buffer, item->bufferSize);
            if(item->operation->valueStream != NULL)
            {
                item->buffer = NULL;
                /* Send OAP pkt to provider to request that it set reqOp's property prp to reqOp's DOFValue value. When that
                provider sends us the expected OAP response, call reqOp's requestor callback rcb. */
                if (item->operation->securityScope != NULL)
                {
                    /* bump ref count because the call to CheckScopeThenRouteProviderOrComplete may release it in the failure case
                       and the call to DOFSystem_CallCompleteWithException to report the failure needs it */
                    if( DOFOperation_Retain(item->operation) )
                    {
                        if( CheckScopeThenRouteProviderOrComplete(item->operation, DOFSystem_ReqOpClosingObserver) == NULL )
                        {
                            DOFSystem_CallCompleteWithException(item->operation, DOFEXCEPTION_FAILED);
                            DOFOperation_Internal_Destroy(item->operation);
                            item->operation = NULL;
                        }
                        else
                        {
                            DOFOperation_Internal_Destroy(item->operation);
                        }
                    }
                    else
                    {
                        DOFSystem_CallCompleteWithException(item->operation, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
                    }
                }
                else
                {
                    /* We want the op, but cannot conclude its disposition until we go to AS to extend permission. */
                    if(ExtendOutboundPermsOrDestroyOp( item->operation, item->operation->provider_object, RequestCallback ) == NULL)
                    {
                        DOFSystem_CallCompleteWithException(item->operation, DOFEXCEPTION_FAILED);
                    }
                }
            }
        }
    }
    PendingRequestMarshallContext_Destroy(item);
}

DOFOperation DOFObject_BeginSet( DOFObject obj, const DOFInterfaceProperty prp, DOFValue value,
                                 DOFOperationControl operationConfig, uint32 timeout, const DOFObjectSetCallback rcb,
                                 void * ctx )
{
    DOFOperation reqOp = NULL;

    if (obj && prp &&
        (!DOF_IsParameterValidation(obj->sys->dof) || DOFValue_CheckCompatible( value, DOFInterfaceProperty_GetType( prp ) )) &&
        (reqOp = DOFOperation_Set_Create( DOFSystem_GetChannel( DOFObject_GetSystem( obj ) ), operationConfig,
                                          timeout, ctx, obj, rcb, prp )) != NULL)
    {
        if (!DOFInterfaceProperty_IsWritable(prp))
        {
            DOFSystem_CallCompleteWithException( reqOp, DOFEXCEPTION_NOT_WRITABLE );
            return reqOp;
        }
        else
        {
            uint8 * buffer;
            uint32 bufferSize;

            if(OALMarshal_MarshalValue(value, DOFInterfaceProperty_GetType( prp ), &buffer, &bufferSize))
            {
                PendingRequestMarshallContext item;

                PCR_Assert( reqOp->provider_object == obj);
                PCR_Assert( reqOp->item == (DOFItem)prp );
                PCR_Assert( reqOp->object_id.object_id == DOFObject_GetObjectID( obj ) );
                item = PendingRequestMarshallContext_Create(reqOp, buffer, bufferSize);
                if(item != NULL)
                {
                    if(!OALEnqueue_Enqueue(obj->sys->requestQueue, (OALEnqueueItem) item))
                    {
                        PendingRequestMarshallContext_Destroy(item);
                        DOFSystem_CallCompleteWithException(reqOp, DOFEXCEPTION_FAILED);
                    }
                    return reqOp;
                }
                PCRMem_Free(buffer);
            }
        }
        DOFOperation_Internal_Destroy(reqOp);
    }

    return NULL;
}


DOFOperation DOFObject_BeginInvoke( DOFObject obj, const DOFInterfaceMethod mtd, uint16 valuesCount,
                                    const DOFValue values[], DOFOperationControl operationConfig, uint32 timeout,
                                    const DOFObjectInvokeCallback rcb, void * ctx )
{
    DOFOperation reqOp = NULL;

    /* values can legally be NULL. This next line tests properly whether values can be NULL or not. */
    if (obj && mtd && valuesCount == DOFInterfaceMethod_GetInputParameterCount( mtd ) &&
        (!DOF_IsParameterValidation(obj->sys->dof) ||
         DOFTypeList_IsValueCompatibleWith( valuesCount, DOFInterfaceMethod_GetInputParameterTypeList( mtd ), values ))
        && (reqOp = DOFOperation_Invoke_Create( DOFSystem_GetChannel( DOFObject_GetSystem( obj ) ), operationConfig,
                                                timeout, ctx, obj, rcb, mtd)) != NULL)
    {
        uint8 * buffer;
        uint32 marshalSize;

        if(OALMarshal_MarshalValues((DOFValue*) values, DOFInterfaceMethod_GetInputParameterTypeList( mtd ), valuesCount, &buffer, &marshalSize))
        {
            PendingRequestMarshallContext item;

            PCR_Assert( reqOp->provider_object == obj);
            PCR_Assert( reqOp->item == (DOFItem)mtd );
            PCR_Assert( reqOp->object_id.object_id == DOFObject_GetObjectID( obj ) );
            item = PendingRequestMarshallContext_Create(reqOp, buffer, marshalSize);
            if(item != NULL)
            {
                if(!OALEnqueue_Enqueue(obj->sys->requestQueue, (OALEnqueueItem) item))
                {
                    PendingRequestMarshallContext_Destroy(item);
                    DOFSystem_CallCompleteWithException(reqOp, DOFEXCEPTION_FAILED);
                }
                return reqOp;
            }
            PCRMem_Free(buffer);
        }
        DOFOperation_Internal_Destroy(reqOp);
    }

    return NULL;
}


DOFOperation DOFObject_BeginSubscribe( DOFObject obj, const DOFInterfaceProperty prp, uint32 minPeriod,
                                       DOFOperationControl operationConfig, uint32 timeout,
                                       const DOFObjectSubscribeCallback rcb, void * ctx )
{
    DOFOperation reqOp = NULL;

    if (obj && prp &&
        (reqOp = DOFOperation_Subscribe_Create( DOFSystem_GetChannel( DOFObject_GetSystem( obj ) ), operationConfig,
                                                                      timeout, ctx, obj, rcb, prp, minPeriod )) != NULL)
    {
        if (!DOFInterfaceProperty_IsReadable( prp ))
            DOFSystem_CallCompleteWithException( reqOp, DOFEXCEPTION_NOT_READABLE );
        else
        {
            /* Send OAP pkt to provider to tell him to send subscription reports, and to have requestor callback rcb be
               called when that provider sends us an OAP response that is that property subscription report. */
            if (reqOp->securityScope != NULL)
                reqOp = CheckScopeThenRouteProviderOrComplete( reqOp, DOFSystem_ReqOpClosingObserver ); /*Ret NULL/not*/
            else
            {
                /* We want the op, but cannot conclude its disposition until we go to AS to extend permission. */
                reqOp = ExtendOutboundPermsOrDestroyOp( reqOp, obj, RequestCallback );         /* Returns NULL or not. */
            }
        }
    }

    return reqOp;
}


DOFOperation DOFObject_BeginRegister( DOFObject obj, const DOFInterfaceEvent evt, DOFOperationControl operationConfig,
                                      uint32 timeout, const DOFObjectEventCallback rcb, void * ctx )
{
    DOFOperation reqOp = NULL;

    if (obj && evt &&
        (reqOp = DOFOperation_Register_Create( DOFSystem_GetChannel( DOFObject_GetSystem( obj ) ), operationConfig,
                                               timeout, ctx, obj, rcb, evt )) != NULL)
    {
        /* Send OAP pkt to provider to tell him to send event notifications, and to have requestor callback rcb be
           called when that provider sends us an OAP response that is that event notificiation. */
        if (reqOp->securityScope != NULL)
            reqOp = CheckScopeThenRouteProviderOrComplete( reqOp, DOFSystem_ReqOpClosingObserver ); /* Ret NULL or not*/
        else
        {
            /* We want the op, but cannot conclude its disposition until we go to AS to extend permission. */
            reqOp = ExtendOutboundPermsOrDestroyOp( reqOp, obj, RequestCallback );             /* Returns NULL or not. */
        }
    }

    return reqOp;
}


DOFOperation DOFObject_BeginSession( DOFObject obj, const DOFInterface interface,
                                     const DOFInterfaceID sessionIID, DOFOperationControl operationConfig,
                                     uint32 timeout, const DOFObjectSessionCallback scb, void * ctx )
{
    DOFOperation reqOp = NULL;

    if (obj && interface &&
        (reqOp = DOFOperation_Session_Create( DOFSystem_GetChannel(  DOFObject_GetSystem( obj ) ), obj, interface,
                                              sessionIID, operationConfig, timeout, scb, ctx )) != NULL)
    {
        if (reqOp->securityScope != NULL)
            reqOp = CheckScopeThenRouteProviderOrComplete( reqOp, DOFSystem_ReqOpClosingObserver ); /* Ret NULL or not*/
        else
        {
            /* We want the op, but cannot conclude its disposition until we go to AS to extend permission. */
            reqOp = ExtendOutboundPermsOrDestroyOp( reqOp, obj, RequestCallback );             /* Returns NULL or not. */
        }
    }

    return reqOp;
}


DOFInterfaceIDList DOFObject_GetInterfaces( const DOFObject obj )
{
    return( DOFSystem_GetProviderInterfaces( DOFObject_GetSystem( obj ), DOFObject_GetObjectID( obj ) ) );
}


boolean DOFObject_IsInterestRequested( const DOFObject self, const DOFInterfaceID interfaceID, DOFInterestLevel level )
{   
    DOFSystem sys;
    PCR_Assert(self); /* self must not be NULL */
    sys = DOFObject_GetSystem(self); /* since self is not NULL this will return a system */
    return DOFSystem_IsInterestRequested(sys, DOFObject_GetObjectID(self), interfaceID, level); 
}


/**
@}
**/
