/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/objectid.h>
#include <internal/objectidinternal.h>
#include "objectidprivate.h"


/** @addtogroup DOFObjectID
@{
**/

DOFObjectID DOFOBJECTID_BROADCAST;
DOFObjectIDDomain DOFOBJECTIDDOMAIN_BROADCAST;


/*
**  "Public" Methods
*/


DOFObjectID DOFObjectID_Create_AttributeBuffer( const DOFObjectID sourceOID, const DOFObjectIDAttributeType attributeID,
                                                uint32 attributeDataSize,
                                                const uint8 * attributeData )
{
    if (sourceOID && (attributeData || attributeDataSize == 0) && attributeDataSize < 256 && attributeID < DOFOBJECTIDATTRIBUTE_INVALID)
    {
        /* Special legal case is allowed: NULL attributeData and attributeDataSize 0. The returned OID is tagged to have
           a new attribute, but the added attribute is appropriately marked to have 0 attribute data size. */
        DOFObjectIDSize oidSize    = DOFObjectID_GetBytesSize( sourceOID );
        uint32          structSize = DOFOBJECTID_HEADER_SIZE + oidSize + attributeDataSize + 2;
                                                   /* 2: new attribute marshalling overhead. */
        DOFObjectID     retOID     = (DOFObjectID)DOFMem_Alloc( structSize );

        if (retOID)
        {
            uint8 * pInsert    = (uint8 *)(retOID->oid + DOFObjectID_GetBaseSize( sourceOID ));
            boolean avpFollows = FALSE;

            DOFObjectID_InitStruct( retOID, structSize - DOFOBJECTID_HEADER_SIZE );
            PCRMem_Copy( retOID->oid, sourceOID->oid, oidSize );
            if (DOFObjectID_HasAttributes( sourceOID ))
            {
                uint8 * pNextAttribute = pInsert;

                while (1) /* Parse thru existing Attributes. Insert new Attribute so attributeIDs in increasing order.*/
                {
                    uint16 offsetToNextAttribute;

                    if ((*pNextAttribute & 0x7F) == (uint8)attributeID) /* 0x7F: 7 bits of presently held attributeID.*/
                    {
                        DOFObjectID_Destroy( retOID );
                        DOFLog_Message(  DOFLOGLEVEL_DEBUG, "DOFObjectID: Create found a Duplicate attributeID." );
                        return NULL;        /* Error- duplicate attributeID. */
                    }
                    offsetToNextAttribute = 2 +                         /* 2: present attribute marshalling overhead. */
                                            *(uint8 *)(pNextAttribute + 1); /* +1: location of attribute length field.*/
                    if ((*pNextAttribute & 0x7F) > (uint8)attributeID)
                    {
                        /* Insert new Attribute before this one. pInsert and pNextAttribute now at present Attribute. */
                        DOFObjectIDSize remainingBytes = oidSize - (DOFObjectIDSize)(pInsert - retOID->oid);

                        pNextAttribute += offsetToNextAttribute;                        /* Just after this Attribute. */
                        PCRMem_Move( pNextAttribute, pInsert, remainingBytes );
                        avpFollows = TRUE;
                        break;
                    }
                    else
                    {
                        /* Our new Attribute will not be placed before this one. */
                        pInsert += offsetToNextAttribute;                               /* Just after this Attribute. */
                        if (!(*pNextAttribute & 0x80))                  /* 0x80: "another attribute is present" flag. */
                        {
                            *pNextAttribute |= 0x80; /* At end of existing Attributes, so place ours immediately after*/
                            break;
                        }

                        pNextAttribute = pInsert;
                    }
                } /* end while(1) */
            }
            else
            {
                /* sourceOID does not, but retOID has Attribute. */
                *(uint8 *)(retOID->oid + DOFObjectID_GetClassSize( retOID )) |= 0x80;   /* 0x80: attribute(s) present.*/
            }
            /* Perform actual insertion. */
            *pInsert++ = (attributeID | ((uint8)avpFollows << 7));                      /* Or in 0x80 or 0x00. */
            *pInsert++ = attributeDataSize;                                             /* attribute data length byte.*/
            PCRMem_Copy( pInsert, attributeData, attributeDataSize );
            return retOID;                                                              /* Success. */
        }
    }

    return NULL;                                                                        /* Error. */
}


DOFObjectID DOFObjectID_Create_AttributeString( const DOFObjectID sourceOID, const DOFObjectIDAttributeType type,
                                                const char * value )
{
    uint32 len = value ? PCRString_Length( value ) : 0;

    /* len greater than 255 is an illegal parameter, because such an attribute length cannot be expressed in a
           DOFObjectID. Reason: the OID attribute length field is only 1 byte.
           The next call cannot reject len greater than 255 because it accepts len as a uint8 only. */
    return (len <= 255) ? DOFObjectID_Create_AttributeBuffer( sourceOID, type, len, (const uint8 *)value ) : NULL;
}


DOFObjectID DOFObjectID_Create_AttributeObjectID(const DOFObjectID sourceOID,
                                                 const DOFObjectIDAttributeType attributeID, const DOFObjectID objectID)
{
    DOFObjectID oid = NULL;

    if (objectID)
    {
        uint32 len = DOFObjectID_GetBytesSize( objectID );

        /* len greater than 255 is an illegal parameter, because such an attribute length cannot be expressed in a
           DOFObjectID. Reason: the OID attribute length field is only 1 byte.
           The next call cannot reject len greater than 255 because it accepts len as a uint8 only. */
        if (len <= 255)
            oid = DOFObjectID_Create_AttributeBuffer( sourceOID, attributeID, len, objectID->oid );
    }

    return oid;
}


/*
**  "Private" Methods
*/


/* @internal */
boolean DOFObjectID_Init( void )
{
    return ((DOFOBJECTID_BROADCAST = DOFObjectID_Create( DOFOBJECTIDCLASS_BROADCAST, 0, NULL )) != NULL &&
            (DOFOBJECTIDDOMAIN_BROADCAST = DOFObjectIDDomain_Create( DOFOBJECTID_BROADCAST )) != NULL);
}


/* @internal */
void DOFObjectID_Shutdown( void )
{
    DOFObjectID_Destroy( DOFOBJECTID_BROADCAST );
    DOFOBJECTID_BROADCAST = NULL;
    DOFObjectIDDomain_Destroy( DOFOBJECTIDDOMAIN_BROADCAST );
    DOFOBJECTIDDOMAIN_BROADCAST = NULL;
}


/**
@}
**/
