/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/objectid.h>
#include <internal/dofinternal.h>
#include "objectidprivate.h"


/** @addtogroup DOFObjectID
    DOF Object Identifier
@{
**/


static boolean matchesAtLeast( const DOFObjectID self, const DOFObjectID objectID, boolean matchMulticast )
{
    uint8 n;

    if (self == objectID)
        return TRUE;
    if (!self || !objectID)
        return FALSE;

    /* Base OID's are NOT compared if matchMulticast is TRUE and objectID is multicast. Only for that special case do
       we allow for success when base OIDs are not identical. This is reasonable because objectID, being multicast, is
       allowed to match other OIDs. */
    if ((!matchMulticast || !DOFObjectID_IsWildcard( objectID )) &&
        (DOFObjectID_IsEqualIgnoreAttributes( self, objectID ) == 0))
        return FALSE;

    /* Base OIDs match or the multicast case applies. Does every attribute in objectID match attributes held in self? */
    if (!DOFObjectID_HasAttributes( objectID ))
        return TRUE;
    /* Walk objectID's attributes list. Only objectID's attributes must be in self. self can have extras. */
    for (n = 0; n < DOFObjectID_GetAttributeCount( objectID ); n++)
    {
        DOFObjectIDAttribute attributeOID = DOFObjectID_GetAttributeAtIndex( objectID, n );

        if (DOFObjectID_HasAttribute( self, attributeOID.id ))
        {
            DOFObjectIDAttribute attributeSelf = DOFObjectID_GetAttribute( self, attributeOID.id );

            PCR_Assert( attributeSelf.id == attributeOID.id );
            if ((attributeSelf.dataSize == attributeOID.dataSize) &&
                (PCRMem_Compare( attributeSelf.data, attributeOID.data, attributeSelf.dataSize ) == 0))
               continue;               /* This attribute matches. */
        }
        return FALSE;
    }

    return TRUE;                       /* Success: every attribute in objectID matches attributes found in self. */
}


/*
**  "Public" Methods
*/


int32 DOFObjectID_CompareIgnoreAttributes( const DOFObjectID oid1, const DOFObjectID oid2 )
{
/* Compare baseOIDs, but ignore the single attribute extension bit. This is our DOFObjectID_CompareBase() call. */
    DOFObjectIDSize baseSize1;
    DOFObjectIDSize baseSize2;
    uint32          stripOffset1;
    uint32          stripOffset2;
    uint32          i;

    if (oid1 == oid2)
        return 0;
    if (!oid1)
        return -1;
    if (!oid2)
        return 1;
    /* Preserve the concept of comparing earliest bytes to latest, where earliest byte differences ARE the answer.
       Prep for loop byte by byte compare, with on the fly bit stripping, then do it. */
    baseSize1 = DOFObjectID_GetBaseSize( oid1 );
    baseSize2 = DOFObjectID_GetBaseSize( oid2 );
    if (baseSize1 > baseSize2)
        baseSize1 = baseSize2;       /* If baseSizes differ, we will bail early with our answer. */
    stripOffset1 = DOFObjectID_GetClassSize( oid1 );
    stripOffset2 = DOFObjectID_GetClassSize( oid2 );
    if (stripOffset1 > stripOffset2)
        stripOffset1 = stripOffset2; /* If classes differ, we will bail with our answer before even using stripOffset.*/
    for (i = 0; i < baseSize1; i++)
    {
        int32 retInt32;
        uint8 val1 = *(uint8 *)(oid1->oid + i);
        uint8 val2 = *(uint8 *)(oid2->oid + i);

        if (i == stripOffset1)
        {
            PCR_Assert( stripOffset1 == DOFObjectID_GetClassSize( oid2 ) );
            val1 &= 0x7F;
            val2 &= 0x7F;
        }
        retInt32 = val1 - val2;
        if (retInt32 != 0)
            return retInt32;
        PCR_Assert( i != stripOffset1 ||                            /* At this moment of passing stripOffset test . . */
                    baseSize1 == DOFObjectID_GetBaseSize( oid2 ) ); /* . .  we also proved baseSizes are equal. */
    }
    return 0; /* Base OIDs are identical except perhaps their single attribute extension bits. */
}


boolean DOFObjectID_IsEqualIgnoreAttributes( const DOFObjectID oid1, const DOFObjectID oid2 )
{
    return (boolean)(DOFObjectID_CompareIgnoreAttributes( oid1, oid2 ) == 0);
}


/** @internal
    Compare two @ref DOFObjectID's for sufficient match, including Attribute extensions. Sufficient match means both
    base @ref DOFObjectID's are identical, unless @a objectID is wildcard (composed of @ref DOFObjectIDClass @ref
    DOFOBJECTIDCLASS_BROADCAST and has attribute(s)), and the @a self @ref DOFObjectID has exact matches for every
    attribute extension held by the @a objectID @ref DOFObjectID. Note that @a self may contain additional attributes as
    well.

    @note If @a objectID is not wildcard, this function behaves the same as DOFObjectID_MatchesAtLeast().
    @note If both parameters are NULL, they are considered equal and sufficiently matched.
          If only a single parameter is NULL, they do NOT match sufficiently.
    @param  self     The @ref DOFObjectID to be compared.
    @param  objectID The @ref DOFObjectID to compare against.

    @retval Non-Zero (TRUE) if the two IDs are sufficiently matched.
    @retval Zero (FALSE) if the two IDs are not sufficiently matched.
**/
boolean DOFObjectID_MatchesAtLeast_Mcast( const DOFObjectID self, const DOFObjectID objectID )
{
    return matchesAtLeast( self, objectID, TRUE );
}


boolean DOFObjectID_MatchesAtLeast( const DOFObjectID self, const DOFObjectID objectID )
{
    return matchesAtLeast( self, objectID, FALSE );
}


/**
@}
**/
