/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/objectid.h>
#include <internal/objectidinternal.h>
#include "objectidprivate.h"


/** @addtogroup DOFObjectID
@{
**/


/*
**  "Public" Methods
*/


int32 DOFObjectID_Compare( const DOFObjectID oid1, const DOFObjectID oid2 )
{
    uint32 safeCompareLength;

    if (oid1 == oid2)
        return 0;                   /* Even if both are NULL. */
    if (!oid1)
        return -1;
    if (!oid2)
        return 1;
    /*
     * Perform straight memcmp(). Regardless of its implementation detail, we are counting on the compare result to be
     * determined in the standard "memcmp" algorithm, where earliest bytes of each oid have priority, in turn. This
     * means that oid class, length, and attribute flag byte (in the early byte positions) differences take priority
     * over follow-on oid data.
     */
    PCR_Assert( oid1->len >= 2 );   /* All valid Object IDs have 2+ bytes of wire representation. */
    PCR_Assert( oid2->len >= 2 );
    safeCompareLength = (oid1->len <= oid2->len) ? oid1->len : oid2->len;
    /* safeCompareLength ensures that all comparisons take place on valid memory. We are conveniently able to ignore
       accounting for the trailing bytes of whichever oid is longer, because oid format forces the oid with extra
       bytes to have its earlier length byte(s) and/or attribute flag(s) to be of higher values than for the shorter
       length oid. Those higher values force the proper compare result even though its trailing bytes are not involved
       in the comparison. */

    return PCRMem_Compare( (const uint8 *)oid1->oid, (const uint8 *)oid2->oid, safeCompareLength );
    /* Note: Our return value must be int32- not int8! For cygwin implementation at least, an int8 loses info. */
}

boolean DOFObjectID_IsEqual( const DOFObjectID oid1, const DOFObjectID oid2 )
{
    return (boolean)(DOFObjectID_Compare( oid1, oid2 ) == 0);
}


/**
@}
**/
