/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/error.h>
#include <dof/objectid.h>
#include <internal/objectidinternal.h>
#include <internal/marshal.h>
#include "objectidprivate.h"


/** @addtogroup DOFObjectID
    DOF Object Identifier
@{
**/


#define MAX_CLASS_VALUE (0x3FFFFFFF) /* Above this cannot be marshalled. */


extern PCRMutex ref_count_mutex;


/*
**  "Public" Methods
*/


DOFObjectID DOFObjectID_Create( DOFObjectIDClass oidClass, uint32 dataSize, const uint8 * pData )
{
    if ((!dataSize || pData) &&                                       /* Cannot have dataSize without pData. */
        dataSize <= MAX_OID_DATA_SIZE && oidClass <= MAX_CLASS_VALUE &&
        (oidClass != DOFOBJECTIDCLASS_BROADCAST || !dataSize))        /* dataSize must be 0 if oidClass is BROADCAST. */
    {
        DOFObjectIDDataSize structSize;
        DOFObjectID         newObjID;
        uint8               oidClassSize = OALMarshal_GetCompressedValueSize( DOFOBJECTID_MAX_CLASS_SIZE, oidClass );

        /* Above call won't return 3 because DOFOBJECTID_MAX_CLASS_SIZE (4) was passed in. */
        PCR_Assert( oidClassSize == 1 || oidClassSize == 2 || oidClassSize == DOFOBJECTID_MAX_CLASS_SIZE );
        structSize = DOFOBJECTID_HEADER_SIZE + 1 + oidClassSize + dataSize;              /* 1: length of length byte. */
        newObjID = (DOFObjectID)DOFMem_Alloc( structSize + 1 );   /* +1: allows adding null-terminator, just in case. */
        if (newObjID)
        {
            DOFObjectID_InitStruct( newObjID, structSize - DOFOBJECTID_HEADER_SIZE );
            OALMarshal_CompressValue( DOFOBJECTID_MAX_CLASS_SIZE, oidClass, oidClassSize, newObjID->oid );
            newObjID->oid[oidClassSize] = dataSize;
            PCRMem_Copy( &(newObjID->oid[oidClassSize + 1]), pData, dataSize );
            newObjID->oid[oidClassSize + 1 + dataSize] = 0; /*null-terminate in case anything looks at oid as a string*/
            return newObjID;
        }
    }

    return NULL;                                                              /* Error. */
}

DOFObjectID DOFObjectID_Create_String( DOFObjectIDClass oidClass, const char * pData )
{
    if (pData)
        return DOFObjectID_Create( oidClass, PCRString_Length( pData ), (const uint8 *)pData );

    return NULL;
}

DOFObjectID DOFObjectID_Create_Bytes( uint32 bufferSize, const uint8 * pOIDBuffer )
{
    uint32      len  = bufferSize;
    DOFObjectID rval = DOFObjectID_Create_Unmarshal( &len, pOIDBuffer );

    if ( rval )
    {
        if ( len != bufferSize )
        {
            DOFObjectID_Destroy( rval );
            rval = NULL;
        }
    }
    return rval;
}


const uint8 * DOFObjectID_GetBytes( const DOFObjectID self )
{
    return self ? self->oid : NULL;
}


DOFObjectIDSize DOFObjectID_GetBytesSize( const DOFObjectID self )
{
    PCR_Assert( self );

    return self->len;
}

DOFObjectID DOFObjectID_Create_StandardString( const char * data )
{
    if (data)
    {
        DOFObjectID newObjID;
        DOFError err;
        uint32 size = 0;

        /* Call parseInternal to find out how big the buffer needs to be. */
        err = DOFObjectID_Create_ParseInternal(data, &size, NULL);

        if (err == DOFERROR_OK)
        {
            /* Create the DOFObjectID using the size that was just computed. */
            newObjID = (DOFObjectID)DOFMem_Alloc( DOFOBJECTID_HEADER_SIZE + size + 1 ); /* Adds space for null-terminator, just in case. */

            if (newObjID)
            {
                DOFObjectID_InitStruct( newObjID, size );
                /* Now that the size is computed and the DOFObjectID is created, call parseInternal again to fill the oid buffer. */
                err = DOFObjectID_Create_ParseInternal( data, &size, newObjID->oid );

                if (err == DOFERROR_OK)
                {
                    return newObjID;
                }

                DOFObjectID_Destroy(newObjID);
            }
        }
    }
    return NULL;
}

void DOFObjectID_Destroy( DOFObjectID self )
{
    if (self)
    {
        PCRMutex_Lock( ref_count_mutex );
        PCR_Assert( self->refCount );
        if (!--self->refCount)
            DOFMem_Free( self );
        PCRMutex_Unlock( ref_count_mutex );
    }
}


/*
**  "Private" Methods
*/


/* @internal */
void DOFObjectID_InitStruct( DOFObjectID newObjID, uint32 dataLen )
{
    newObjID->refCount = 1;
    newObjID->len = dataLen;
}


/**
@}
**/
