/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/objectid.h>
#include <internal/marshal.h>
#include <internal/staticassertinternal.h>
#include "objectidprivate.h"


/** @addtogroup DOFObjectID
    DOF Object Identifier
@{
**/


/*
**  "Private" Methods
*/


/*
    Determine the length of the objectID class by using the unmarshall function.  This function assumes that the
    bytes provided are a valid objectID.
*/
uint32 DOFObjectID_GetClassSize_Bytes( const uint8 * pBytes )
{
    uint32 size = 4;

    (void)OALMarshal_UncompressValue( DOFOBJECTID_MAX_CLASS_SIZE, &size, pBytes );

    return size;
}

uint32 DOFObjectID_GetClassSize( DOFObjectID self )
{
    return DOFObjectID_GetClassSize_Bytes( self->oid );
}


/** @internal
    Get the byte length of the wire representation of a @ref DOFObjectID.

    @param  pBytes Pointer to wire represention of the @ref DOFObjectID. This should not be NULL.
    @return Full byte length of the wire representation of the @ref DOFObjectID at @ pBytes.
    @exception 0 is returned for NULL @ pBytes
**/
DOFObjectIDSize DOFObjectID_GetSize_Bytes( const uint8 * pBytes )
{
    if (pBytes)
    {
        uint8  lengthByte;
        uint32 computedSize;
        uint8  classSize = DOFObjectID_GetClassSize_Bytes( pBytes );

        /* DOFObjectID_Create() and DOFObjectID_Create_Unmarshal() ensure this assert about classSize. */
        PCR_Assert( classSize == 1 || classSize == 2 || classSize == DOFOBJECTID_MAX_CLASS_SIZE );
        lengthByte = *(const uint8 *)(pBytes + classSize);
        computedSize =  classSize + 1 + (lengthByte & OID_DATA_LEN_MASK);                /* 1: length of length byte. */
        PCR_Assert( computedSize >= 2 && (computedSize <= (DOFOBJECTID_MAX_CLASS_SIZE + 1 + MAX_OID_DATA_SIZE)) );
        if ((lengthByte & 0x80) != 0)                                              /* bit 7: attrribute present flag. */
        {
            const uint8 * pNextAttribute = (const uint8 *)(pBytes + computedSize);             /* We have attributes. */

            while (1)                                  /* Add to computedSize as we parse through all the Attributes. */
            {
                computedSize += (2 + *(const uint8 *)(pNextAttribute + 1));     /* 2: attribute marshalling overhead. */
                if (!(*pNextAttribute & 0x80))
                    break;                              /* No more Attributes */
                pNextAttribute = (const uint8 *)(pBytes + computedSize);
            }
        }
        {
            return computedSize;
        }
    }

    return 0;
}


/**
@}
**/
