/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/objectid.h>
#include <dof/error.h>

#include <internal/marshal.h>

#include "objectidprivate.h"

struct parseCtx
{
    const char* oid;
    uint8* buffer;
    uint32 buffLen;
    uint32 oidLen;
    uint32 currOidPos;
    uint32 currBufferPos;
}parseCtx;

/* Operations on OID string */
#define PARSECTX_PEEK_CHAR_OID(ctx) ( (ctx)->oid[(ctx)->currOidPos] )
#define PARSECTX_PEEK_NEXT_CHAR_OID(ctx) ( (ctx)->oid[(ctx)->currOidPos+1] )
#define PARSECTX_READ_CHAR_OID(ctx) ( (ctx)->oid[(ctx)->currOidPos++] )
#define PARSECTX_GET_CURRENT_POS_OID(ctx) ( (ctx)->oid+(ctx)->currOidPos )
#define PARSECTX_STEP_OID(ctx, count)((ctx)->currOidPos+=(count))

/* Operations on DOFObjectID buffer */
#define PARSECTX_GET_CURRENT_POS_BUF(ctx)( ((ctx)->buffer)? (ctx)->buffer+(ctx)->currBufferPos: NULL )
#define PARSECTX_STEP_BUF(ctx, count)((ctx)->currBufferPos+=(count))
#define PARSECTX_WRITE_AT_POS_BUF(ctx, pos, value) do{ if((ctx)->buffer) *(pos) = (value); } while(0)
#define PARSECTX_OR_AT_POS_BUF(ctx, pos, value) do{ if((ctx)->buffer) *(pos) |= (value); } while(0)
#define PARSECTX_WRITE_BUF(ctx, value)( ((ctx)->buffer)? (ctx)->buffer[(ctx)->currBufferPos++] = (value): (ctx)->currBufferPos++ )
#define PARSECTX_CHECK_LEN(ctx, len) (((ctx)->buffer)? (((ctx)->currBufferPos+len <= (ctx)->buffLen)? DOFERROR_OK: DOFERROR_INSUFFICIENT_RESOURCES): DOFERROR_OK)

/* Operation to read from OID straight to buffer */
#define PARSECTX_WRITE_BUF_FROM_OID(ctx) (((ctx)->buffer)? (ctx)->buffer[(ctx)->currBufferPos++] = (ctx)->oid[(ctx)->currOidPos]: ((ctx)->currBufferPos++),((ctx)->currOidPos++))

#define IS_DIGIT(c) (((c) >= '0' && (c) <= '9'))
#define DIGIT2VALUE(c) (c-48)

#define HEX2VALUE(c) ( (IS_DIGIT(c))? DIGIT2VALUE(c) : ((c) >= 'A' && (c) <= 'F')? (c-55): (c-87) )
#define VALIDHEXSEP(c) ( (c) == ' ' || (c) == ':' || (c) == '-' )
#define VALIDHEX(c) ( ((c) >= '0' && (c) <= '9') || ((c) >= 'A' && (c) <= 'F') || ((c) >= 'a' && (c) <= 'f') )
#define VALIDHEXBYTE(s) ( VALIDHEX((s)[0]) && VALIDHEX((s)[1]) )
#define VALIDNUMBER(c) ((c) >= '0' && (c) <= '9')

#define VALIDASCIICHAR(c) (((uint8)c) >= 32 && ((uint8)c) <= 126 )

#define IS_ESCAPED(c) ( (c) == '(' || (c) == ')' || (c) == '[' || (c) == ']' || (c) == '{' || (c) == '}' || (c) == '\\' || (c) == '|' )

static DOFError parseFormatOID( struct parseCtx* ctx );

static DOFError parseHexField( struct parseCtx* ctx )
{
    /* Hex fields start with { and end with } can contain space, dash and colon*/
    if( PARSECTX_READ_CHAR_OID(ctx) == '{' && PARSECTX_PEEK_CHAR_OID(ctx) != '}' )
    {
        while( PARSECTX_PEEK_CHAR_OID(ctx) != '}' )
        {
            if( VALIDHEXBYTE( PARSECTX_GET_CURRENT_POS_OID(ctx) ) )
            {
                if( PARSECTX_CHECK_LEN(ctx, 1) == DOFERROR_OK )
                {
                    PARSECTX_WRITE_BUF(ctx, HEX2VALUE( PARSECTX_PEEK_CHAR_OID(ctx) ) << 4 | HEX2VALUE( PARSECTX_PEEK_NEXT_CHAR_OID(ctx) ) );
                    PARSECTX_STEP_OID(ctx,2);

                    if( VALIDHEXSEP( PARSECTX_PEEK_CHAR_OID(ctx) ) )
                    {
                        if( PARSECTX_PEEK_NEXT_CHAR_OID(ctx) == '}')
                        {
                            /* no seperator after byte block */
                            return DOFERROR_ILLEGAL_ARGUMENT;
                        }
                        PARSECTX_STEP_OID(ctx,1);
                    }
                }
                else
                {
                    return DOFERROR_INSUFFICIENT_RESOURCES;
                }
            }
            else
            {
                return DOFERROR_ILLEGAL_ARGUMENT;
            }
        }
        PARSECTX_STEP_OID(ctx,1);
        return DOFERROR_OK;
    }
    return DOFERROR_FAILED;
}

static DOFError parseStringField( struct parseCtx* ctx )
{
    /* Copy into buffer until end or */
    while( ctx->currOidPos < (ctx->oidLen-1) )
    {
        char curr = PARSECTX_PEEK_CHAR_OID(ctx);
        if( curr == ']' || curr == '(' )
        {
            break; /* End of string field */
        }
        else if( curr == '\\' )
        {
            /* Handle escaped char */
            PARSECTX_STEP_OID(ctx,1);
            if( !IS_ESCAPED(PARSECTX_PEEK_CHAR_OID(ctx)) || PARSECTX_CHECK_LEN(ctx,1) != DOFERROR_OK )
                return DOFERROR_FAILED;
            PARSECTX_WRITE_BUF_FROM_OID(ctx);
        }
        else
        {
            if( VALIDASCIICHAR( curr ) && PARSECTX_CHECK_LEN(ctx,1) == DOFERROR_OK )
                PARSECTX_WRITE_BUF_FROM_OID(ctx);
            else
                return DOFERROR_FAILED;
        }
    }
    return DOFERROR_OK;
}

static DOFError parseOIDClass( struct parseCtx* ctx )
{
    if( PARSECTX_PEEK_CHAR_OID(ctx) == '{' && PARSECTX_PEEK_NEXT_CHAR_OID(ctx) != '}' )
    {
        /* Hex */
        uint8 classSize = 0;
        uint32 oidClass = 0;
        PARSECTX_STEP_OID(ctx,1);
        while( PARSECTX_PEEK_CHAR_OID(ctx) != '}' )
        {
            if( VALIDHEXBYTE( PARSECTX_GET_CURRENT_POS_OID(ctx) ) )
            {
                oidClass <<= 8;
                oidClass += (HEX2VALUE( PARSECTX_PEEK_CHAR_OID(ctx) ) << 4 | HEX2VALUE( PARSECTX_PEEK_NEXT_CHAR_OID(ctx)));
                PARSECTX_STEP_OID(ctx,2);

                if( VALIDHEXSEP( PARSECTX_PEEK_CHAR_OID(ctx) ) )
                {
                    if( PARSECTX_PEEK_NEXT_CHAR_OID(ctx) == '}')
                    {
                        /* no seperator after byte block */
                        return DOFERROR_ILLEGAL_ARGUMENT;
                    }
                    PARSECTX_STEP_OID(ctx,1);
                }
            }
            else
            {
                return DOFERROR_ILLEGAL_ARGUMENT;
            }
        }
        PARSECTX_STEP_OID(ctx,1);

        classSize = OALMarshal_GetCompressedValueSize(4, oidClass);
        if( PARSECTX_CHECK_LEN(ctx, classSize) == DOFERROR_OK )
        {
            if(PARSECTX_GET_CURRENT_POS_BUF(ctx))
                classSize = OALMarshal_CompressValue(4, oidClass, classSize, PARSECTX_GET_CURRENT_POS_BUF(ctx));

            PARSECTX_STEP_BUF(ctx, classSize);
        }

        return DOFERROR_OK;
    }
    else
    {
        /* Number */
        uint8 classSize = 0;
        uint32 oidClass = 0;
        while( IS_DIGIT( PARSECTX_PEEK_CHAR_OID(ctx) ) )
        {
            oidClass*=10;
            oidClass+= DIGIT2VALUE(PARSECTX_PEEK_CHAR_OID(ctx));
            PARSECTX_STEP_OID(ctx,1);
        }

        classSize = OALMarshal_GetCompressedValueSize(4, oidClass);
        if( PARSECTX_CHECK_LEN(ctx, classSize) == DOFERROR_OK )
        {
            if(PARSECTX_GET_CURRENT_POS_BUF(ctx))
                classSize = OALMarshal_CompressValue(4, oidClass, classSize, PARSECTX_GET_CURRENT_POS_BUF(ctx));

            PARSECTX_STEP_BUF(ctx, classSize);
        }

        return DOFERROR_OK;
    }
    return DOFERROR_FAILED;
}

static DOFError parseAttributeID( struct parseCtx* ctx )
{
    if( PARSECTX_PEEK_CHAR_OID(ctx) == '{' )
    {
        return parseHexField(ctx);
    }
    else
    {
        uint8 avpid = 0;
        while( IS_DIGIT(PARSECTX_PEEK_CHAR_OID(ctx)) )
        {
            avpid *= 10;
            avpid += DIGIT2VALUE(PARSECTX_PEEK_CHAR_OID(ctx));
            PARSECTX_STEP_OID(ctx,1);
        }

        if( PARSECTX_CHECK_LEN(ctx,1) == DOFERROR_OK )
        {
            PARSECTX_WRITE_BUF(ctx, avpid);
            return DOFERROR_OK;
        }
    }
    return DOFERROR_FAILED;
}

static DOFError parseAttributeData( struct parseCtx* ctx )
{
    if( PARSECTX_PEEK_CHAR_OID(ctx) == '[' )
    {
        return parseFormatOID(ctx);
    }
    else if( PARSECTX_PEEK_CHAR_OID(ctx) == '{' )
    {
        return parseHexField(ctx);
    }
    else
    {
        return parseStringField(ctx);
    }
}

static DOFError parseAttribute( struct parseCtx* ctx )
{
    if( parseAttributeID(ctx) == DOFERROR_OK )
    {
        /* seperated by ':' */
        if(PARSECTX_READ_CHAR_OID(ctx) == ':' && PARSECTX_CHECK_LEN(ctx,1) == DOFERROR_OK )
        {
            uint8* length = PARSECTX_GET_CURRENT_POS_BUF(ctx);
            PARSECTX_STEP_BUF(ctx,1);

            if( parseAttributeData(ctx) == DOFERROR_OK )
            {
                PARSECTX_WRITE_AT_POS_BUF( ctx, length, (PARSECTX_GET_CURRENT_POS_BUF(ctx) - (length+1)));
                return DOFERROR_OK;
            }
        }
    }
    return DOFERROR_FAILED;
}

static DOFError parseAttributes( struct parseCtx* ctx )
{
    /* AVPs surrounded by '(' ')' but needs at least an avp */
    if( PARSECTX_READ_CHAR_OID(ctx) == '(' &&  PARSECTX_PEEK_CHAR_OID(ctx) != ')' )
    {
        while( PARSECTX_PEEK_CHAR_OID(ctx) != ')' )
        {
            uint8* avpID  = PARSECTX_GET_CURRENT_POS_BUF(ctx);

            if( parseAttribute(ctx) != DOFERROR_OK )
                return DOFERROR_FAILED;

            /* multiple seperated by '|' */
            if( PARSECTX_PEEK_CHAR_OID(ctx) == '|' && PARSECTX_PEEK_NEXT_CHAR_OID(ctx) != ')' )
            {
                PARSECTX_OR_AT_POS_BUF(ctx, avpID, 0x80); /* set that there is a next attribute */
                PARSECTX_STEP_OID(ctx,1);
            }
        }
        PARSECTX_STEP_OID(ctx,1);
        return DOFERROR_OK;
    }
    return DOFERROR_FAILED;
}

static DOFError parseFormatOID( struct parseCtx* ctx )
{
    /* oid must start with '[' */
    if( PARSECTX_PEEK_CHAR_OID(ctx) == '[' )
    {
        PARSECTX_STEP_OID(ctx,1);
        /* Get class id */
        if( parseOIDClass(ctx) == DOFERROR_OK )
        {
            /* seperated by ':' */
            if( PARSECTX_READ_CHAR_OID(ctx) == ':' && PARSECTX_CHECK_LEN(ctx,1) == DOFERROR_OK )
            {
                uint8* length = PARSECTX_GET_CURRENT_POS_BUF(ctx);
                PARSECTX_STEP_BUF(ctx,1);

                /* Get data */
                if(PARSECTX_PEEK_CHAR_OID(ctx) == '{')
                {
                    /* hex data */
                    if( parseHexField( ctx ) != DOFERROR_OK )
                        return DOFERROR_FAILED;
                }
                else
                {
                    /* string data */
                    if( parseStringField( ctx ) != DOFERROR_OK )
                        return DOFERROR_FAILED;
                }

                /* Write length */
                PARSECTX_WRITE_AT_POS_BUF(ctx, length, (PARSECTX_GET_CURRENT_POS_BUF(ctx)-(length+1)));

                /* Check if attributes exist */
                if(PARSECTX_PEEK_CHAR_OID(ctx) == '(' )
                {
                    PARSECTX_OR_AT_POS_BUF(ctx, length, 0x80); /* set that there are attributes */
                    if( parseAttributes( ctx ) != DOFERROR_OK )
                        return DOFERROR_FAILED;
                }

                /* Ends with ] */
                if(PARSECTX_READ_CHAR_OID(ctx) == ']')
                {
                    return DOFERROR_OK;
                }
            }
        }
    }
    return DOFERROR_FAILED;
}

DOFError DOFObjectID_Create_ParseInternal( const char* oid, uint32* size, uint8* buffer )
{
    struct parseCtx ctx;

    ctx.oid = oid;
    ctx.buffer = buffer;
    ctx.currOidPos = 0;
    ctx.currBufferPos = 0;

    if( oid )
    {
        if( size )
        {
            ctx.buffLen = (*size);
            ctx.oidLen = PCRString_Length(oid);
            if( PARSECTX_PEEK_CHAR_OID(&ctx) == '[' )
            {
                /* Format OID */
                if( parseFormatOID( &ctx ) == DOFERROR_OK )
                {
                    (*size) = ctx.currBufferPos;
                    return DOFERROR_OK;
                }
            }
            else if( PARSECTX_PEEK_CHAR_OID(&ctx) == '{' )
            {
                /* HEX OID */
                if( parseHexField( &ctx ) == DOFERROR_OK )
                {
                    (*size) = ctx.currBufferPos;
                    return DOFERROR_OK;
                }
            }
            (*size) = 0;
        }
    }
    return DOFERROR_FAILED;
}

