/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/objectid.h>
#include <internal/oalstring.h>
#include <internal/dofinternal.h>


/** @addtogroup DOFObjectID
    DOF Object Identifier
@{
**/
#define IS_PRINTABLE(c)         ( ((uint8)c) >= 32U && ((uint8)c) < 127U )
#define IS_ESCAPED(c)           ( (c) == '(' || (c) == ')' || (c) == '[' || (c) == ']' || (c) == '{' || (c) == '}' || (c) == '\\' || (c) == '|' )

/*
**  "Private" Type and Method Declarations
*/


/** @internal
    Determine the size of the resulting string when converting data from an Object ID into a standard string format.

    There are two formats, either a hex-string surrounded by braces, or an ASCII string with any special characters escaped.  This routine will automatically choose which format is shorter.

    @param data          The data to turn into a string.
    @param dataSize      Size of the data (in bytes).

    @return Byte count of resulting string, excluding string terminator.
**/
#define ObjectID_DataToStringLength( data, dataSize ) ObjectID_DataToString( (data), (dataSize), NULL )


/** @internal
    Convert data from an Object ID into a standard string format.

    There are two formats, either a hex-string surrounded by braces, or an
    ASCII string with any special characters escaped.  This routine will
    automatically choose which format is shorter.  There is @b no terminator
    placed on the string.

    @note   <b>This routine assumes sufficient space exists in the buffer!</b>
            Passing NULL for @a pBuf is permitted to only determine the length of the resulting string.

    @param  data          The data to turn into a string.
    @param  dataSize      Size of the data (in bytes).
    @param  pBuf          Buffer in which to place the new string.  NULL to only return the length.

    @return Filled byte count (no terminator is present).
**/
static uint32 ObjectID_DataToString( const uint8 * data, uint32 dataSize, char *pBuf )
{
    uint32 len = 0, i, nonprintable, escaped;

    PCR_Assert( data || dataSize == 0 );
    /* Determine if the data is printable... */
    for (i = 0, nonprintable = 0, escaped = 0; i < dataSize; i++)
    {
        if (!IS_PRINTABLE( data[i] ))
            nonprintable++;
        else if (IS_ESCAPED( data[i] ))
            escaped++;
    }
    if ( nonprintable == 0 )
    {
        /* Printable, so copy as a string, escaping where necessary. */
        if ( pBuf )
        {
            for (i = 0; i < dataSize; i++)
            {
                if (IS_ESCAPED( data[i] ))
                {
                    pBuf[len++] = '\\';
                    pBuf[len++] = data[i];
                }
                else
                    pBuf[len++] = data[i];
            }
        } else {
            len = dataSize + escaped; /* Count escaped characters twice. */
        }
    }
    else
    {
        /* Non-printable, so format as hex string. */
        if ( pBuf )
        {
            pBuf[len++] = '{';
            for (i = 0; i < dataSize; i++)
            {
                pBuf[len++] = OALString_HexDigitToChar( (data[i] >> 4) & 0x0F );
                pBuf[len++] = OALString_HexDigitToChar( (data[i]     ) & 0x0F );
            }
            pBuf[len++] = '}';
        } else {
            len = dataSize * 2 + 2;
        }
    }
    return len;
}


/** @internal
    Determine the size of the resulting string when converting an Object ID into its standard string format.

    There are multiple valid representations for a single Object ID in the standard Object ID string format.

    @param oid           The DOFObjectID to convert.

    @return Byte count of resulting string, excluding string terminator.
**/
static uint32 ObjectID_ToStringLength( const DOFObjectID oid )
{
    uint32 len = 0;

    /* Note: All these string functions can be exercised with objectid_test.c, which outputs the string to console. */
    PCR_Assert( oid );
    len = 7 /* [{xx}: and trailing ] */ + ObjectID_DataToStringLength( DOFObjectID_GetData( oid ),
                                                                       DOFObjectID_GetDataSize( oid ) );
    if (DOFObjectID_GetIDClass( oid ) & 0xFF000000)
        len += 6;                                           /* Six more hex digits. */
    else if (DOFObjectID_GetIDClass( oid ) & 0xFF0000)
        len += 4;                                           /* Four more hex digits. */
    else if (DOFObjectID_GetIDClass( oid ) & 0xFF00)
        len += 2;                                           /* Two more hex digits. */
    /* Handle Attributes, if any. */
    if (DOFObjectID_HasAttributes( oid ))
    {
        uint8 i;                                            /* Max attribute count is under uint8. */
        uint8 attributeCount = DOFObjectID_GetAttributeCount( oid );

        len += 2;                                           /* surrounding ( ) */
        for (i = 0; i < attributeCount; i++)
        {
            DOFObjectID embedOID;
            DOFObjectIDAttribute avpDescriptor = DOFObjectID_GetAttributeAtIndex( oid, i );

            if (!DOFObjectIDAttribute_IsValid(avpDescriptor))
                break;  /* Done with Attributes. If here, some error took place. */
            if (i)
                len++;
            len += 5;  /* {xx}: */
            /* Handle embedded Object IDs. */
            embedOID = DOFObjectID_Create_Bytes( DOFObjectIDAttribute_GetValueSize( avpDescriptor ),
                                                 DOFObjectIDAttribute_GetValue( avpDescriptor ) );
            if ( embedOID )
            {
                len += ObjectID_ToStringLength( embedOID ); /* Recurse to compute string rep length of found OID. */
                DOFObjectID_Destroy( embedOID );
            } else {
                /* Hex Data. */
                len += ObjectID_DataToStringLength( DOFObjectIDAttribute_GetValue( avpDescriptor ),
                                                    DOFObjectIDAttribute_GetValueSize( avpDescriptor ) );
            }
        } /* end for(). */
    }

    return len;
}


/** @internal
    Convert an Object ID into its standard string format.

    There are multiple valid representations for a single Object ID in the standard Object ID string format.

    @note   <b>This routine assumes sufficient space exists in the buffer!</b>
            Refer to @ref ObjectID_ToStringLength to determine the size of
            string required.

    @param oid           The DOFObjectID to convert.
    @param pBuf          Pointer to buffer to fill

    @return Filled byte count, excluding placed string terminator.
**/
static uint32 ObjectID_ToString( const DOFObjectID oid, char * pBuf  )
{
    DOFObjectIDClass oidClass;
    uint32           len = 0;

    PCR_Assert( oid );
    PCR_Assert( pBuf );
    pBuf[len++] = '[';
    pBuf[len++] = '{';
    /* Class */
    oidClass = DOFObjectID_GetIDClass( oid );
    if (oidClass & 0xFF000000)
    {
        pBuf[len++] = OALString_HexDigitToChar( (oidClass >> 28) & 0x0F );
        pBuf[len++] = OALString_HexDigitToChar( (oidClass >> 24) & 0x0F );
    }
    if (oidClass & 0xFFFF0000)
    {
        pBuf[len++] = OALString_HexDigitToChar( (oidClass >> 20) & 0x0F );
        pBuf[len++] = OALString_HexDigitToChar( (oidClass >> 16) & 0x0F );
    }
    if (oidClass & 0xFFFFFF00)
    {
        pBuf[len++] = OALString_HexDigitToChar( (oidClass >> 12) & 0x0F );
        pBuf[len++] = OALString_HexDigitToChar( (oidClass >>  8) & 0x0F );
    }
    pBuf[len++] = OALString_HexDigitToChar( (oidClass >>  4) & 0x0F );
    pBuf[len++] = OALString_HexDigitToChar( (oidClass      ) & 0x0F );
    pBuf[len++] = '}';
    pBuf[len++] = ':';
    /* Data */
    len += ObjectID_DataToString( DOFObjectID_GetData( oid ), DOFObjectID_GetDataSize( oid ), &pBuf[len] );
    /* Handle Attributes, if any. */
    if (DOFObjectID_HasAttributes( oid ))
    {
        uint8 i;
        uint8 attributeCount = DOFObjectID_GetAttributeCount( oid );

        pBuf[len++] = '(';
        for (i = 0; i < attributeCount; i++)
        {
            DOFObjectID embedOID;
            DOFObjectIDAttribute avpDescriptor = DOFObjectID_GetAttributeAtIndex( oid, i );

            if (!DOFObjectIDAttribute_IsValid(avpDescriptor))
                break;  /* Done with Attributes. If here, some error took place. */
            if (i)
                pBuf[len++] = '|';
            pBuf[len++] = '{';
            pBuf[len++] = OALString_HexDigitToChar( (DOFObjectIDAttribute_GetType( avpDescriptor ) >> 4) & 0x0F );
            pBuf[len++] = OALString_HexDigitToChar( (DOFObjectIDAttribute_GetType( avpDescriptor )     ) & 0x0F );
            pBuf[len++] = '}';
            pBuf[len++] = ':';

            /* Handle embedded Object IDs. */
            embedOID = DOFObjectID_Create_Bytes( DOFObjectIDAttribute_GetValueSize( avpDescriptor ),
                                                 DOFObjectIDAttribute_GetValue( avpDescriptor ) );
            if ( embedOID )
            {
                len += ObjectID_ToString( embedOID, &pBuf[len] ); /* Recurse to output string rep of found OID. */
                DOFObjectID_Destroy( embedOID );
            } else {
                /* Hex Data. */
                len += ObjectID_DataToString( DOFObjectIDAttribute_GetValue( avpDescriptor ),
                                              DOFObjectIDAttribute_GetValueSize( avpDescriptor ), &pBuf[len] );
            }
        } /* end for(). */
        pBuf[len++] = ')';
    }
    pBuf[len++] = ']';

    return len;
}


char * DOFObjectID_CreateStandardString( const DOFObjectID oid )
{
    char * pRetval;
    uint32 len;

    if (!oid)
        return NULL;

    len = ObjectID_ToStringLength( oid );
    /* Use PCRMem_Alloc() and not DOFMem_Alloc() because app caller will be freeing memory with PCRMem_Destroy(). */
    pRetval = (char *)PCRMem_Alloc( len + 1 );
    if (pRetval)
    {
        ObjectID_ToString( oid, pRetval );
        pRetval[len] = 0;
    }

    return pRetval;
}


/**
@}
**/
