/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/dofinternal.h>
#include <internal/marshal.h>
#include <internal/objectidinternal.h>
#include "objectidprivate.h"


/*
**  "Private" Type and Method Declarations
*/


/*
**  "Private" Methods
*/


/* @internal */
boolean DOFObjectID_Marshal(DOFObjectID self, OALStream stream)
{
    if (self && stream)
        return OALStream_Write( stream, DOFObjectID_GetBytesSize( self ), self->oid );

    return FALSE;
}


/** @internal
    Create a new @ref DOFObjectID from a given wire representation of @ref DOFObjectID.

    @mustfree{DOFObjectID_Destroy}
    @param length [IN]  Pointer to 32 bit length of @a buffer. This should point to a length above 1.
                  [OUT] Pointer to 32 bit length of bytes in @ buffer actually used to form the new @ref DOFObjectID.
    @param buffer Pointer to the wire representation of @ref DOFObjectID. This should not be NULL.
    @return @ref DOFObjectID
    @exception Return NULL for
        - NULL @a buffer, or
        - @length less than 2, or
        - invalid wire representation of the @ref DOFObjectID at @a buffer, or
        - insufficient bytes held in @a buffer to hold the wire represention of the given @ref DOFObjectID.
*/
DOFObjectID DOFObjectID_Create_Unmarshal( uint32 * length, const uint8 * buffer )
{
    uint32  len = *length;

    /* Legal OID described at buffer must have at least 2 bytes. */
    if (buffer && len >= 2)
    {
        uint32 classSize = len;
        uint32 class = OALMarshal_UncompressValue( DOFOBJECTID_MAX_CLASS_SIZE, &classSize, buffer );

        /* Legal OID described at buffer must have its class representation be correctly compressed. */
        if (classSize != OALMARSHAL_INVALID_VALUE)
        {
            uint32 computedSize;

            /* Above call won't return 3 because DOFOBJECTID_MAX_CLASS_SIZE (4) was passed in. */
            if( classSize == 1 || classSize == 2 || classSize == DOFOBJECTID_MAX_CLASS_SIZE )
            {
                computedSize = classSize + 1;                              /* 1: length of length byte. */
                /* Legal OID described at buffer must have enough bytes to describe its OID class. */
                if (len >= computedSize)
                {
                    uint8 lenByte = buffer[classSize];

                    /* Legal OID described at buffer must have its length byte bit 6 be 0. */
                    if (!(lenByte & 0x40))
                    {
                        boolean hasAttr;
                        uint8   dataLen = lenByte & OID_DATA_LEN_MASK;

                        /* Legal broadcast OID described at buffer must have no base data, though it can have attribute(s)*/
                        if ((class == DOFOBJECTIDCLASS_BROADCAST) && (dataLen > 0))
                            goto notvalid;
                        computedSize += dataLen;
                        hasAttr = lenByte & 0x80;                       /* Valid OID base; check attributes. */
                        while (hasAttr)
                        {
                            /* Legal OID described at buffer must have enough bytes to hold each new found attribute. */
                            if (len >= computedSize + 2)                /* 2: attribute marshalling overhead. */
                            {
                                hasAttr = buffer[computedSize] & 0x80;  /* bit 7: next attribute present flag. */
                                computedSize += (2 + buffer[computedSize + 1]);
                            }
                            else
                                goto notvalid;
                        }
                        /* Legal OID described at buffer must have enough buffer bytes, final check. */
                        if (len >= computedSize)
                        {
                            DOFObjectID newObjID = (DOFObjectID)DOFMem_Alloc( DOFOBJECTID_HEADER_SIZE + computedSize + 1 );
                                                                        /* Adds space for null-terminator, just in case. */

                            *length = computedSize;
                            if (newObjID)
                            {
                                DOFObjectID_InitStruct( newObjID, computedSize );
                                PCRMem_Copy( newObjID->oid, buffer, computedSize );
                                newObjID->oid[computedSize] = 0;
                                return newObjID;                    /* Success. */
                            }
                            /* buffer describes valid OID, but due to alloc failure we cannot return the newly created OID*/
                            goto allocErrorOut;
                        }
                    }
                }
            }
        }
    }
notvalid:
    /* buffer does not describe a valid OID, but do not log a message. The caller may have called us to find out if the
       buffer does or does not obey the rules of a valid OID. He learns that by our NULL return. */
allocErrorOut:
    *length = 0;

    return NULL;
}

/** @internal
    Create a new @ref DOFObjectID from a given wire representation of @ref DOFObjectID held in a @ref OALStream.

    @mustfree{DOFObjectID_Destroy}
    @param stream The @ref OALStream holding the wire representation of @ref DOFObjectID. This must not be NULL.
    @return @ref DOFObjectID
    @exception Return NULL for
        - @ stream holding wire representation of length less than 2, or
        - invalid wire representation of the @ref DOFObjectID held in @ stream, or
        - insufficient bytes held in the @a stream wire represention of the given @ref DOFObjectID.
*/
DOFObjectID DOFObjectID_Unmarshal( OALStream stream )
{
    uint32 marshalSize;

    PCR_Assert( stream );
    marshalSize = OALStream_GetDataSize( stream );
    if (marshalSize)
    {
        DOFObjectID oid = DOFObjectID_Create_Unmarshal( &marshalSize, OALStream_GetReadPtr( stream ) );

        if (oid)
        {
            /* Above call does not adjust stream parameters, so do that here. */
            PCR_Assert( marshalSize < 0x80000000 ); /*Even OIDs with max 256 attributes have max bytes size below this*/
            OALStream_AdjustReadOffset( stream, (int32)marshalSize );              /* marshalSize is positive number. */
            return oid;
        }
    }

    return NULL;
}

/* @internal
    This function is used for an opID or SID where the max length is 20 data bytes and
    there is not an AVP list.

    This function will return the number of bytes for SID (including the class and
    length byte), or zero if it is invalid. */
uint32 DOFObjectID_Unmarshal_GetSidLength( const uint8 * pBytes, uint32 max_length )
{
    uint32 sid_bytes_len;
    uint8  length_byte;
    uint8  class_len;

    if ((pBytes == NULL) || (max_length < 2))
        return 0;
    class_len = DOFObjectID_GetClassSize_Bytes( pBytes );
    sid_bytes_len = class_len + 1;                       /* 1: make sure that we have at least 1 byte for the length. */
    if (max_length < sid_bytes_len)
        return 0;
    length_byte = *(const uint8 *)(pBytes + class_len);
    /*  make sure that we have at least the number of bytes specified by the length.
        A length greater than 20 is invalid. */
    sid_bytes_len += length_byte;
    if ((length_byte > DOFOBJECTID_MAX_SIDBYTES_SIZE) || (max_length < sid_bytes_len))
        return 0;

    return sid_bytes_len;
}
