/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/objectid.h>
#include <internal/objectidinternal.h>
#include "objectidprivate.h"


/** @addtogroup DOFObjectIDSource
@{
**/


/*
**  "Public" Methods
*/


DOFObjectIDSource DOFObjectIDSource_Create( const DOFObjectID id )
{
    /* The protocol spec that describes operation identifiers, under the "Type 19" section, mandates requirements that
       can be interpreted to mean that a source ID cannot be broadcast, have attributes, or have data length above 20.*/
    /* Spec "base length max 20," refers to 6 bit "BASE LENGTH" of the basic OID spec, not DOFObjectID_GetBaseSize(). */
    if (id && !DOFObjectID_IsBroadcast( id ) && !DOFObjectID_HasAttributes( id ) &&
        (DOFObjectID_GetDataSize( id ) <= DOFOBJECTID_MAX_SIDBYTES_SIZE))
        return (DOFObjectIDSource)DOFObjectID_Copy( id );

    return NULL;
}


DOFObjectIDSource DOFObjectIDSource_Copy( DOFObjectIDSource self )
{
    return (DOFObjectIDSource)DOFObjectID_Copy( (DOFObjectID) self );
}


DOFObjectID DOFObjectIDSource_GetObjectID( const DOFObjectIDSource self )
{
    return (DOFObjectID)self;
}


void DOFObjectIDSource_Destroy( DOFObjectIDSource self )
{
    DOFObjectID_Destroy( (DOFObjectID)self );
}


/*
**  "Private" Methods
*/

boolean DOFObjectID_IsSourceID( DOFObjectID id )
{
    /* Spec "base length max 20," refers to 6 bit "BASE LENGTH" of the basic OID spec, not DOFObjectID_GetBaseSize(). */
    if (id && !DOFObjectID_IsBroadcast( id ) && !DOFObjectID_HasAttributes( id ) &&
        (DOFObjectID_GetDataSize( id ) <= DOFOBJECTID_MAX_SIDBYTES_SIZE))
        return TRUE;

    return FALSE;
}

static uint32 DOFObjectID_ASAssignedGetNodeID( DOFObjectID self )
{
    uint8        i        = 0;
    uint8        nodeLen  = DOFObjectID_GetDataSize( self ) - 6; /* This subtracts the data after the node id from the length */
    const uint8* selfData = DOFObjectID_GetData( self );
    uint32       ret      = 0;

    PCR_Assert(nodeLen > 0 && nodeLen <= 4);

    for( ; i < nodeLen; ++i )
       ((uint8*)&ret)[i] = selfData[nodeLen-(i+1)];

    return ret;
}

boolean DOFObjectID_IsASAssignedEqual( DOFObjectID self, DOFObjectID other)
{
    DOFObjectIDClass sc = DOFObjectID_GetIDClass( self );

    if (sc == DOFOBJECTIDCLASS_AS && sc == DOFObjectID_GetIDClass( other ) &&
        DOFObjectID_ASAssignedGetNodeID( self ) == DOFObjectID_ASAssignedGetNodeID( other ))
        return TRUE;

    return FALSE;
}

/**
@}
**/

