/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/dof.h>
#include <dof/core.h>
#include <internal/observable.h>
#include <internal/slist.h>
#include <internal/objectinternal.h>
#include <internal/operationinternal.h>

/* @addtogroup OALObservable
*
*   The OALObservable implements the observer/observable design pattern
*   by keeping a linked list of observers associated with an observable
*   structure. An observer is attached to the system with a callback
*   on notification. When the user adds an observer, he/she will receive
*   an OALObserver type which should be remembered if they wil want to
*   remove that observer later.
@{
**/

/*
**  "Private" Type and Method Declarations
*/

typedef struct OALObservable_t {
    struct OALObserver *observers;
    uint8 dirty; /* ack we have to waste a whole byte! :( */
} OALObservable_t;

typedef struct OALObserver_t {
    struct OALObserver   *next;
    OALObserver_fn_t    fn;
    void                *arg;
    OALObservable_t     *observable;
} OALObserver_t;

/*
**  "Public" Methods
*/

void OALObservable_Init(void)
{
}

void OALObservable_Shutdown(void)
{
}

/* "Static" methods */

/* @internal
*   Create a new OALObservable.

 @return A new @ref OALObservable.
 @retval NULL if an error occurred while creating the object.
*/

OALObservable OALObservable_Create(void)
{
    OALObservable o = DOFMem_Alloc(sizeof(OALObservable_t));

    if(o) 
    {
        o->observers = 0;
        o->dirty = 0;
    }

    return o;
}

static void remove_all(OALSListNode node, void* arg)
{
    DOFMem_Free(node);
}

/* "Object" methods */

/* @internal
*   Destroy an OALObservable object.

    @param observable The observable to destroy.

    @note This will remove all items in the slist as well as the desired object.
*/

void OALObservable_Destroy(OALObservable observable)
{
    if ( observable )
    {
        OALSList_RemoveAll((OALSListNode*)&observable->observers, remove_all, 0);
        DOFMem_Free(observable);
    }
}

/* @internal
*   Add an observer to the observable.

    @param  observable  The observable to add an observer to.
    @param  fn  The function to be called when this observable notifies its observers.
    @param  arg The argument to be passed to the callback function.

    @return An observer that the caller should keep track of if they intend to remove
            it from the observable.
*/

OALObserver OALObservable_AddObserver(OALObservable observable, OALObserver_fn_t fn, void *arg)
{
    OALObserver_t *o;

    PCR_Assert(fn);

    o = DOFMem_Alloc(sizeof(OALObserver_t));
    if(!o)
        return NULL;

    o->next = 0;
    o->fn = fn;
    o->arg = arg;
    o->observable = observable;

    OALSList_Insert((OALSListNode*)&observable->observers, (OALSListNode)o);

    return o;
}

/* @internal
*   Remove an observer to the observable.

    @param  observer  The observer to remove from the observable.
*/

void OALObservable_RemoveObserver(OALObservable observable, OALObserver observer)
{
    OALSList_Remove((OALSListNode*)&observable->observers, (OALSListNode)observer);
    DOFMem_Free(observer);
}

/* @internal
*   Mark an observable "dirty".

    @param  observable  The observable to mark as changed.
*/

void OALObservable_SetChanged(OALObservable observable)
{
    observable->dirty = 1;
}

static boolean notify(OALSListNode node, void *unused)
{
    OALObserver observer = (OALObserver)node;
    PCR_Assert(observer->fn);
    observer->fn(observer, observer->observable, observer->arg);
    return FALSE;
}

/* @internal
*   Notify the observers (call their callbacks) if an observable has changed.

    @param  observable The observable to check and notify.
*/
void OALObservable_NotifyObservers(OALObservable observable)
{
    if(!observable->dirty) return; /* nothing to do */

    OALSList_Traverse((OALSListNode)observable->observers, notify, 0);
    observable->dirty = 0;
}


/* @internal
*   Notify the observers (call their callbacks).

    @param  observable The observable to notify.
*/
void OALObservable_SetChangedAndNotifyObservers(OALObservable observable)
{
    OALObservable_SetChanged(observable);
    OALObservable_NotifyObservers(observable);
}

/**
@}
**/
