/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/exception.h>
#include <dof/query.h>
#include <dof/interface.h>
#include <dof/interfaceid.h>
#include <dof/item.h>
#include <dof/object.h>
#include <dof/objectid.h>
#include <dof/operation.h>
#include <internal/securityscope.h>
#include <dof/request.h>
#include <dof/result.h>
#include <dof/error.h>
#include <dof/system.h>
#include <dof/value.h>
#include <internal/avl_tree.h>
#include <internal/bridgeinternal.h>
#include <internal/exception.h>
#include <internal/filterinternal.h>
#include <internal/interfaceinternal.h>
#include <internal/interfaceidinternal.h>
#include <internal/objectinternal.h>
#include <internal/objectidinternal.h>
#include <internal/operationinternal.h>
#include <internal/permissioninternal.h>
#include <internal/resultinternal.h>
#include <internal/systeminternal.h>
#include <internal/marshal.h>
#include <internal/value.h>
#include <internal/operationconfiginternal.h>
#include <internal/requestinternal.h>
#include <internal/router.h>
#include <internal/error.h>
#include <internal/mutex_pool.h>
#include <dof/type.h>
#include <internal/source.h>

#ifdef DOFOPERATION_DEBUG_REFCOUNT
#include <stdio.h>
#endif

static uint32 opcnt = 0;

static void _increment_retry(DOFOperation op);

#define CNT_IMPLICIT_MASK 0x80000000UL



typedef struct MatchAndMarkContext_t
{
    DOFObjectID  sessionID;
    DOFOperation provOp;
} MatchAndMarkContext_t;


typedef struct DOFOperationID_t
{
    uint32              ref_count;
    DOFOperationCount   cnt; /* DO NOT MOVE ME */
    sid_t               *sid;
} DOFOperationID_t;

static DOFObjectIDSource local_oid  = NULL;
static sid_t *local_sid; /* not cached */

static boolean is_initialized = FALSE;

static PCRMutex local_oid_mutex;
extern PCRMutex ref_count_mutex;

static void _cleanup(DOFOperation op);

static void _cleanup_r(DOFOperation op);

void DOFOperation_Cancel_Sync(DOFOperation op);

static DOFOperation _createInterestInternal(      OALChannel       requestor_channel,
                                            const DOFInterfaceID   interfaceID,
                                            const DOFObjectID      objectID,
                                            const OALAlias         alias,
                                                  uint32           timeout,
                                                  uint32           timeoutGrace,
                                                  DOFOperationType type);

boolean DOFOperation_Init(void)
{
    local_oid_mutex = PCRMutex_Create();
    if(local_oid_mutex != NULL)
    {
        return TRUE;
    }
    return FALSE;
}

void DOFOperation_Shutdown(void)
{
    PCRMutex_Destroy(local_oid_mutex);
    local_oid_mutex = NULL;
}

boolean DOFOperation_DOFInit(void)
{
    is_initialized = TRUE;
    if(source_init())
    {
        return TRUE;
    }
    return FALSE;
}

void DOFOperation_DOFShutdown(void)
{
    is_initialized = FALSE;
    if (local_oid != NULL)
    {
        DOFObjectIDSource_Destroy(local_oid);
        local_oid = NULL;
        if ( local_sid )
        {
            source_release(local_sid);
            local_sid = NULL;
        }
    }
    source_shutdown();
}


/** @addtogroup DOFOperation
@{
**/


/*
**  "Private" Type and Method Declarations
*/


/*
**  "Public" Methods
*/

/* Internal "Static" Factory Methods */

static DOFOperation _create(void)
{
    DOFOperation ret = (DOFOperation) DOFMem_Alloc(sizeof(DOFOperation_t));

    if(ret != NULL)
    {
        OALObservable observer = OALObservable_Create();

        CancelOperationContext_t * cancelContext;

        if(observer == NULL)
        {
            DOFMem_Free(ret);
        }
        else
        {
            cancelContext = DOFOperationManager_CreateCancelContext(ret);
            if(cancelContext != NULL)
            {
                PCRMem_Set(ret, 0, sizeof(DOFOperation_t));
                ret->sources = (DOFOperationSource)DOFMem_Alloc(sizeof(DOFOperationSource_t));
                if( ret->sources != NULL )
                {
                    PCRMem_Set(ret->sources, 0, sizeof(DOFOperationSource_t));
                    ret->ref_count = 1;
                    ret->operation_will_close = observer;
                    ret->cancelContext = cancelContext;
                    return ret;
                }
                DOFMem_Free(cancelContext);
            }
            OALObservable_Destroy(observer);
            DOFMem_Free(ret);
        }
    }

    return NULL;
}

static void _cleanup(DOFOperation op)
{
    PCR_Assert(op != NULL);

    OALObservable_Destroy(op->operation_will_close);
    DOFMem_Free(op->sources);
    DOFMem_Free(op);
}

static DOFOperation _create_r(DOFOperationType type, OALChannel requestor_channel,
                              DOFOperationControl operationConfig, DOFSecurityScope securityScope,
                              uint32 timeout, void * context, DOFObject obj)
{
    DOFOperation ret;

    PCR_Assert(requestor_channel);
    /*  Some operations do not have an object.  Once we have a real solution, these asserts should
     *  be restored.
     */

    ret = _create();

    if(!ret) return NULL;

    if (obj)
    {
        ret->provider_object = DOFObject_Copy(obj);     /* use the copy method so that ref counts are incremented */
        if (!ret->provider_object)
            goto cleanup;
        ret->object_id.object_id = DOFObjectID_Copy(DOFObject_GetObjectID(obj));
        if (!ret->object_id.object_id)
            goto cleanup;
        ret->flags.bits.provider_oid_is_wildcard = DOFObjectID_IsWildcard(ret->object_id.object_id);
    }

    ret->securityScope = securityScope; /* Take ownership of security scope. */

    ret->operation_type = type;

    ret->sources->channel = OALChannel_Retain(requestor_channel);

    ret->sources->oal_timeout = timeout;
    ret->context = context;

    if (operationConfig != NULL)
    {
        ret->operationConfig = DOFOperationControl_Copy(operationConfig);
        if(ret->operationConfig == NULL)
        {
            goto cleanup;
        }
    }
    else
    {
        ret->operationConfig = DOFOperationControl_Create();
        if(ret->operationConfig == NULL)
        {
            goto cleanup;
        }
    }

    ret->opID = DOFOperationID_CreateLocal();
    if(ret->opID == NULL)
    {
        goto cleanup;
    }

    ret->ref_count = 1;

    return ret;

cleanup:
    _cleanup_r(ret);
    return NULL;
}

static void _cleanup_r(DOFOperation op)
{
    PCR_Assert(op != NULL);

    if(op->sources->channel != NULL)
    {
        OALChannel_Release(op->sources->channel);
    }
    if(op->opID != NULL)
    {
        DOFOperationID_Destroy(op->opID);
    }
    if(op->provider_object != NULL)
    {
        DOFObject_Destroy(op->provider_object);
    }
    if(op->object_id.object_id != NULL)
    {
        DOFObjectID_Destroy(op->object_id.object_id);
    }
    if(op->securityScope != NULL)
    {
        DOFSecurityScope_Destroy(op->securityScope);
    }
    if(op->operationConfig != NULL)
    {
        DOFOperationControl_Destroy(op->operationConfig);
    }
    _cleanup(op);
}

/* @internal */
DOFOperation DOFOperation_SourceLost_Create(void)
{
    DOFOperation ret = _create();

    if(ret != NULL)
    {
        ret->operation_type = DOFOPERATIONTYPE_EPP_NEGONE_SOURCELOST;
    }
    return ret;
}

/* @internal */
DOFOperation DOFOperation_SourceFound_Create(void)
{
    DOFOperation ret = _create();

    if(ret != NULL)
    {
        ret->operation_type = DOFOPERATIONTYPE_EPP_NEGONE_SOURCEFOUND;
    }
    return ret;
}

/* @internal */
DOFOperation DOFOperation_NodeDown_Create(void)
{
    DOFOperation ret = _create();

    if(ret != NULL)
    {
        ret->operation_type = DOFOPERATIONTYPE_EPP_NEGONE_NODEDOWN;
        return ret;
    }
    return NULL;
}

DOFOperation DOFOperation_CancelAll_Create(void)
{
    DOFOperation ret = _create();

    if(ret != NULL)
    {
        ret->operation_type = DOFOPERATIONTYPE_EPP_NEGONE_CANCELALL;
        ret->flags.bits.internally_created = TRUE;
        return ret;
    }
    return NULL;
}

DOFOperation DOFOperation_EppHeartbeat_Create(DOFAddress address)
{
    DOFOperation ret = _create();

    if(ret != NULL)
    {
        ret->opID = DOFOperationID_CreateLocal();
        if(ret->opID == NULL)
        {
            _cleanup(ret);
            return NULL;
        }
        ret->sources->address = DOFAddress_Copy(address);
        if(address && !ret->sources->address)
        {
            DOFOperation_Internal_Destroy(ret);
            return NULL;
        }
        ret->securityScope = DOFSecurityScope_Copy(DOF_GetDefaultSecurityScope(DOF_GetDof()));
        ret->flags.bits.internally_created = TRUE;
        ret->operation_type = DOFOPERATIONTYPE_EPP_NEGONE_HEARTBEAT;
    }
    return ret;
}

DOFOperation DOFOperation_EppPing_Create(DOFAddress address)
{
    DOFOperation ret = _create();

    if(ret != NULL)
    {
        ret->opID = DOFOperationID_CreateLocal();
        if(ret->opID == NULL)
        {
            _cleanup(ret);
            return NULL;
        }
        ret->sources->address = DOFAddress_Copy(address);
        if(address && !ret->sources->address)
        {
            DOFOperation_Internal_Destroy(ret);
            return NULL;
        }
        ret->securityScope = DOFSecurityScope_Copy(DOF_GetDefaultSecurityScope(DOF_GetDof()));
        ret->flags.bits.internally_created = TRUE;
        ret->operation_type = DOFOPERATIONTYPE_EPP_NEGONE_PING;
    }
    return ret;
}

DOFOperation DOFOperation_EppQuery_Create(OALChannel channel, DOFAddress address)
{
    DOFOperation ret = _create();

    if(ret) {
        ret->flags.bits.internally_created = FALSE;
        ret->operation_type = DOFOPERATIONTYPE_EPP_NEGONE_QUERY;

        if(address) {
            ret->sources->address = DOFAddress_Copy(address);
            if(!ret->sources->address) {
                DOFOperation_Internal_Destroy(ret);
                return NULL;
            }
        }
        ret->securityScope = DOFSecurityScope_Copy(DOF_GetDefaultSecurityScope(DOF_GetDof()));
        ret->sources->channel = OALChannel_Retain(channel);
    }

    return ret;
}

DOFOperation DOFOperation_ESP_V2_Query_Create(DOFAddress address, uint32 retry)
{
    DOFOperation ret = _create();

    if(ret != NULL)
    {
        ret->operationConfig = DOFOperationControl_Create();
        if(ret->operationConfig == NULL)
        {
            DOFOperation_Internal_Destroy(ret);
            return NULL;
        }

        DOFOperationControl_SetRetryPeriod(ret->operationConfig, retry);

        ret->opID = DOFOperationID_CreateLocal();
        if(ret->opID == NULL)
        {
            DOFOperation_Internal_Destroy(ret);
            return NULL;
        }

        ret->sources->address = DOFAddress_Copy(address);
        if(address && !ret->sources->address)
        {
            DOFOperation_Internal_Destroy(ret);
            return NULL;
        }
        ret->securityScope = DOFSecurityScope_Copy(DOF_GetDefaultSecurityScope(DOF_GetDof()));
        ret->flags.bits.internally_created = FALSE;
        ret->operation_type = DOFOPERATIONTYPE_ESP_V2_QUERY;
    }
    return ret;
}

DOFOperation DOFOperation_ESP_V2_Terminate_Create(DOFAddress address)
{
    DOFOperation ret = _create();

    if(ret != NULL)
    {
        /* The terminate is sent without an opID. */
        ret->sources->address = DOFAddress_Copy(address);
        if(address && !ret->sources->address)
        {
            DOFOperation_Internal_Destroy(ret);
            return NULL;
        }
        ret->securityScope = DOFSecurityScope_Copy(DOF_GetDefaultSecurityScope(DOF_GetDof()));
        ret->flags.bits.internally_created = FALSE;
        ret->operation_type = DOFOPERATIONTYPE_ESP_V2_TERMINATE;
    }
    return ret;
}

DOFOperation DOFOperation_ESP_V2_Open_Create(DOFAddress address)
{
    DOFOperation ret = _create();

    if(ret != NULL)
    {
        ret->operationConfig = DOFOperationControl_Create();
        if(ret->operationConfig == NULL)
        {
            DOFOperation_Internal_Destroy(ret);
            return NULL;
        }
        DOFOperationControl_SetRetryPeriod(ret->operationConfig, 5000);

        ret->opID = DOFOperationID_CreateLocal();
        if(ret->opID == NULL)
        {
            DOFOperation_Internal_Destroy(ret);
            return NULL;
        }
        ret->sources->address = DOFAddress_Copy(address);
        if(address && !ret->sources->address)
        {
            DOFOperation_Internal_Destroy(ret);
            return NULL;
        }
        ret->securityScope = DOFSecurityScope_Copy(DOF_GetDefaultSecurityScope(DOF_GetDof()));
        ret->flags.bits.internally_created = FALSE;
        ret->operation_type = DOFOPERATIONTYPE_ESP_V2_OPEN;
    }
    return ret;
}

DOFOperation DOFOperation_ESP_V2_Open_CreateInternal(DOFOperationID opID, DOFAddress address, uint32 timeout)
{
    DOFOperation ret = _create();

    if(ret != NULL)
    {
        ret->opID = DOFOperationID_Copy(opID);
        if(ret->opID == NULL)
        {
            _cleanup(ret);
            return NULL;
        }
        ret->sources->address = DOFAddress_Copy(address);
        if(address && !ret->sources->address)
        {
            DOFOperation_Internal_Destroy(ret);
            return NULL;
        }
        ret->securityScope = DOFSecurityScope_Copy(DOF_GetDefaultSecurityScope(DOF_GetDof()));
        ret->flags.bits.internally_created = TRUE;
        ret->operation_type = DOFOPERATIONTYPE_ESP_V2_OPEN;
        DOFOperation_SetTimeout(ret, timeout);
    }
    return ret;
}

/* @internal
    Factory method for DOFOperations of type DOFOPERATIONTYPE_GET.

    @param timeout  The amount of milliseconds that the operation will continue before completing.
    @param context  Data that is passed to the callback.
    @param obj      The object that this operation is associated with.
    @param rcb      The function that will be called as the operation either completes or returns additional data.
    @param prop     The property to retrieve the value of.

    @return The created DOFOperation or NULL if the operation could not be created or this method was used in error
*/
DOFOperation DOFOperation_Get_Create(OALChannel requestor_channel, DOFOperationControl operationConfig,
                                     uint32 timeout, void * context, DOFObject obj,
                                     const DOFObjectGetCallback rcb, DOFInterfaceProperty prop)
{
    DOFInterfaceID iid;

    PCR_Assert(prop);
    iid = DOFInterfaceID_Copy(DOFInterfaceProperty_GetInterfaceID(prop));

    if(iid != NULL)
    {
        DOFPermission commandPerm = _createNeededPermission(DOFOPERATIONTYPE_GET, FALSE, DOFObject_GetObjectID(obj), iid);
        DOFPermission responsePerm = NULL;

        if(DOFObjectID_IsUnicast(DOFObject_GetObjectID(obj)))
        {
            responsePerm = _createNeededPermission(DOFOPERATIONTYPE_GET, TRUE, DOFObject_GetObjectID(obj), iid);
        }
        if((commandPerm != NULL) && ((responsePerm != NULL) || (!DOFObjectID_IsUnicast(DOFObject_GetObjectID(obj)))))
        {
            DOFSecurityScope secScope =
                DOFSecurityMembrane_ValidateOutboundPermissions(requestor_channel->securityMembrane, commandPerm, commandPerm);
            DOFOperation     ret      = _create_r( DOFOPERATIONTYPE_GET, requestor_channel, operationConfig,
                                                   secScope, timeout, context, obj ); /* Takes ownership of secScope. */

            if(ret != NULL)
            {
                PCR_Assert((responsePerm == NULL) == ret->flags.bits.provider_oid_is_wildcard);
                ret->commandPerm = commandPerm;
                ret->responsePerm = responsePerm;
                ret->callbacks.rcb_get = rcb;
                ret->item = (DOFItem)prop;
                ret->itemId = DOFItem_GetID(ret->item);
                DOFItem_RetainDesc(ret->item);
                PCR_Assert( iid );
                ret->interface_id = iid;
                ret->flags.bits.value_type_is_not_list = 1;
                if ( DOFObjectID_IsUnicast( DOFObject_GetObjectID( obj ) ))
                    ret->flags.bits.commandCompletesOnAnyFinal = 1; /* Directed OAP requests put this EPP header flag on wire.*/
                else
                    DOFOperationControl_SetDefaultResponseLevel( ret->operationConfig, DOFOPERATIONRESPONSELEVEL_SUCCESS );
                if (DOFOperationControl_GetVerbosity(ret->operationConfig) == 0)
                {
                    DOFOperationControl_SetResponseLevel(ret->operationConfig, 1);
                }
                return ret;
            }
            else
                DOFSecurityScope_Destroy(secScope);
        }
        DOFPermission_Destroy(commandPerm);
        DOFPermission_Destroy(responsePerm);
        DOFInterfaceID_Destroy(iid);
    }
    return NULL;
}

/* @internal
    Factory method for DOFOperations of type DOFOPERATIONTYPE_SET.

    @param timeout  The amount of milliseconds that the operation will continue before completing.
    @param context  Data that is passed to the callback.
    @param obj      The object that this operation is associated with.
    @param rcb      The function that will be called as the operation either completes or returns additional data.
    @param prop     The property to set the value of.
    @param value    The value to set the property to.

    @return The created DOFOperation or NULL if the operation could not be created or this method was used in error
*/
DOFOperation DOFOperation_Set_Create(OALChannel requestor_channel, DOFOperationControl operationConfig,
                                     uint32 timeout, void * context, DOFObject obj,
                                     const DOFObjectSetCallback rcb, DOFInterfaceProperty prop)
{
    DOFInterfaceID iid;

    PCR_Assert(prop != NULL);

    iid = DOFInterfaceID_Copy(DOFInterfaceProperty_GetInterfaceID(prop));

    if(iid != NULL)
    {
        DOFPermission commandPerm = _createNeededPermission(DOFOPERATIONTYPE_SET, FALSE, DOFObject_GetObjectID(obj), iid);
        DOFPermission responsePerm = NULL;

        if(DOFObjectID_IsUnicast(DOFObject_GetObjectID(obj)))
        {
            responsePerm = _createNeededPermission(DOFOPERATIONTYPE_SET, TRUE, DOFObject_GetObjectID(obj), iid);
        }
        if((commandPerm != NULL) && ((responsePerm != NULL) || (!DOFObjectID_IsUnicast(DOFObject_GetObjectID(obj)))))
        {
            DOFSecurityScope secScope =
                DOFSecurityMembrane_ValidateOutboundPermissions(requestor_channel->securityMembrane, commandPerm, commandPerm);

            DOFOperation ret;

            ret = _create_r(DOFOPERATIONTYPE_SET, requestor_channel, operationConfig,
                            secScope, timeout, context, obj); /* Takes ownership of secScope. */

            if(ret != NULL)
            {
                PCR_Assert((responsePerm == NULL) == ret->flags.bits.provider_oid_is_wildcard);
                ret->callbacks.rcb_set = rcb;
                ret->flags.bits.value_type_is_not_list = 1;
                ret->commandPerm = commandPerm;
                ret->responsePerm = responsePerm;
                ret->item = (DOFItem)prop;
                ret->itemId = DOFItem_GetID(ret->item);
                DOFItem_RetainDesc(ret->item);
                PCR_Assert( iid );
                ret->interface_id = iid;
                if ( DOFObjectID_IsUnicast( DOFObject_GetObjectID( obj ) ))
                    ret->flags.bits.commandCompletesOnAnyFinal = 1; /* Directed OAP requests put this EPP header flag on wire.*/
                else
                    DOFOperationControl_SetDefaultResponseLevel( ret->operationConfig, DOFOPERATIONRESPONSELEVEL_SUCCESS );
                return ret;
            }
            else
                DOFSecurityScope_Destroy(secScope);
            DOFPermission_Destroy(commandPerm);
            DOFPermission_Destroy(responsePerm);
        }
        DOFInterfaceID_Destroy(iid);
    }

    return NULL;
}

/* @internal
    Factory method for DOFOperations of type DOFOPERATIONTYPE_INVOKE.

    @param timeout  The amount of milliseconds that the operation will continue before completing.
    @param context  Data that is passed to the callback.
    @param obj      The object that this operation is associated with.
    @param rcb      The function that will be called as the operation either completes or returns additional data.
    @param method   The method to invoke.
    @param valuesCount The count of parameters to the method.
    @param values   The parameters/arguments to the method.

    @return The created DOFOperation or NULL if the operation could not be created or this method was used in error
*/
DOFOperation DOFOperation_Invoke_Create(OALChannel requestor_channel, DOFOperationControl operationConfig,
                                        uint32 timeout, void * context, DOFObject obj, const DOFObjectInvokeCallback rcb,
                                        DOFInterfaceMethod method)
{
    DOFInterfaceID iid;

    PCR_Assert(method != NULL);
    iid = DOFInterfaceID_Copy(DOFInterfaceMethod_GetInterfaceID(method));
    if(iid != NULL)
    {
        DOFPermission commandPerm = _createNeededPermission(DOFOPERATIONTYPE_INVOKE, FALSE, DOFObject_GetObjectID(obj), iid);
        DOFPermission responsePerm = NULL;

        if(DOFObjectID_IsUnicast(DOFObject_GetObjectID(obj)))
        {
            responsePerm = _createNeededPermission(DOFOPERATIONTYPE_INVOKE, TRUE, DOFObject_GetObjectID(obj), iid);
        }
        if((commandPerm != NULL) && ((responsePerm != NULL) || (!DOFObjectID_IsUnicast(DOFObject_GetObjectID(obj)))))
        {
            DOFSecurityScope secScope =
                DOFSecurityMembrane_ValidateOutboundPermissions(requestor_channel->securityMembrane, commandPerm, commandPerm);

            DOFOperation ret;

            ret = _create_r(DOFOPERATIONTYPE_INVOKE, requestor_channel, operationConfig,
                            secScope, timeout, context, obj); /* Takes ownership of secScope. */

            if(ret != NULL)
            {
                PCR_Assert((responsePerm == NULL) == ret->flags.bits.provider_oid_is_wildcard);
                ret->callbacks.rcb_inv = rcb;
                /* invoke uses a list */
                ret->commandPerm = commandPerm;
                ret->responsePerm = responsePerm;
                ret->item = (DOFItem)method;
                ret->itemId = DOFItem_GetID(ret->item);
                DOFItem_RetainDesc(ret->item);
                PCR_Assert( iid );
                ret->interface_id = iid;
                if ( DOFObjectID_IsUnicast( DOFObject_GetObjectID( obj ) ))
                    ret->flags.bits.commandCompletesOnAnyFinal = 1; /* Directed OAP requests put this EPP header flag on wire.*/
                else
                    DOFOperationControl_SetDefaultResponseLevel( ret->operationConfig, DOFOPERATIONRESPONSELEVEL_SUCCESS );
                return ret;
            }
            else
                DOFSecurityScope_Destroy(secScope);
            DOFPermission_Destroy(commandPerm);
            DOFPermission_Destroy(responsePerm);
        }
        DOFInterfaceID_Destroy(iid);
    }
    return NULL;
}

/* @internal
    Factory method for DOFOperations of type DOFOPERATIONTYPE_REGISTER.

    @param timeout  The amount of milliseconds that the operation will continue before completing.
    @param context  Data that is passed to the callback.
    @param obj      The object that this operation is associated with.
    @param rcb      The function that will be called as the operation either completes or returns additional data.
    @param event    The event to register for.

    @return The created DOFOperation or NULL if the operation could not be created or this method was used in error
*/
DOFOperation DOFOperation_Register_Create(OALChannel requestor_channel, DOFOperationControl operationConfig,
                                          uint32 timeout, void * context, DOFObject obj,
                                          const DOFObjectEventCallback rcb, DOFInterfaceEvent event)
{
    DOFInterfaceID iid;

    PCR_Assert(event != NULL);

    iid = DOFInterfaceID_Copy(DOFInterfaceEvent_GetInterfaceID(event));
    if(iid != NULL)
    {
        DOFPermission commandPerm = _createNeededPermission(DOFOPERATIONTYPE_REGISTER, FALSE, DOFObject_GetObjectID(obj), iid);
        DOFPermission responsePerm = NULL;

        if(DOFObjectID_IsUnicast(DOFObject_GetObjectID(obj)))
        {
            responsePerm = _createNeededPermission(DOFOPERATIONTYPE_INVOKE, TRUE, DOFObject_GetObjectID(obj), iid);
        }
        if((commandPerm != NULL) && ((responsePerm != NULL) || (!DOFObjectID_IsUnicast(DOFObject_GetObjectID(obj)))))
        {
            DOFSecurityScope secScope =
                DOFSecurityMembrane_ValidateOutboundPermissions(requestor_channel->securityMembrane, commandPerm, commandPerm);

            DOFOperation ret;

            ret = _create_r(DOFOPERATIONTYPE_REGISTER, requestor_channel, operationConfig,
                            secScope, timeout, context, obj); /* Takes ownership of secScope. */

            if(ret != NULL)
            {
                PCR_Assert((responsePerm == NULL) == ret->flags.bits.provider_oid_is_wildcard);
                ret->commandPerm = commandPerm;
                ret->responsePerm = responsePerm;
                ret->callbacks.rcb_reg = rcb;
                ret->item = (DOFItem)event;
                ret->itemId = DOFItem_GetID(ret->item);
                DOFItem_RetainDesc(ret->item);
                PCR_Assert( iid );
                ret->interface_id = iid;
                ret->operation_type = DOFOPERATIONTYPE_REGISTER;
                if ( !DOFObjectID_IsUnicast( DOFObject_GetObjectID( obj ) ) )
                    DOFOperationControl_SetDefaultResponseLevel( ret->operationConfig, DOFOPERATIONRESPONSELEVEL_SUCCESS );
                return ret;
            }
            else
                DOFSecurityScope_Destroy(secScope);
        }
        DOFPermission_Destroy(commandPerm);
        DOFPermission_Destroy(responsePerm);
        DOFInterfaceID_Destroy(iid);
    }
    return NULL;
}

/* @internal
    Factory method for DOFOperations of type DOFOPERATIONTYPE_SUBSCRIBE.

    @param timeout      The amount of milliseconds that the operation will continue before completing.
    @param context      Data that is passed to the callback.
    @param obj          The object that this operation is associated with.
    @param rcb          The function that will be called as the operation either completes or returns additional data.
    @param prop         The property to subscribe to.
    @param minPeriod    The minimum period between subscription notifications.

    @return The created DOFOperation or NULL if the operation could not be created or this method was used in error
*/
DOFOperation DOFOperation_Subscribe_Create(OALChannel requestor_channel, DOFOperationControl operationConfig,
                                           uint32 timeout, void * context, DOFObject obj,
                                           const DOFObjectSubscribeCallback rcb, DOFInterfaceProperty prop, uint32 minPeriod)
{
    DOFInterfaceID iid;

    PCR_Assert(prop != NULL);

    iid = DOFInterfaceID_Copy(DOFInterfaceProperty_GetInterfaceID(prop));
    if(iid != NULL)
    {
        DOFPermission commandPerm = _createNeededPermission(DOFOPERATIONTYPE_SUBSCRIBE, FALSE, DOFObject_GetObjectID(obj), iid);
        DOFPermission responsePerm = NULL;

        if(DOFObjectID_IsUnicast(DOFObject_GetObjectID(obj)))
        {
            responsePerm = _createNeededPermission(DOFOPERATIONTYPE_INVOKE, TRUE, DOFObject_GetObjectID(obj), iid);
        }
        if((commandPerm != NULL) && ((responsePerm != NULL) || (!DOFObjectID_IsUnicast(DOFObject_GetObjectID(obj)))))
        {
            DOFSecurityScope secScope =
                DOFSecurityMembrane_ValidateOutboundPermissions(requestor_channel->securityMembrane, commandPerm, commandPerm);

            DOFOperation ret;

            ret = _create_r(DOFOPERATIONTYPE_SUBSCRIBE, requestor_channel, operationConfig,
                            secScope, timeout, context, obj); /* Takes ownership of secScope. */

            if(ret != NULL)
            {
                PCR_Assert((responsePerm == NULL) == ret->flags.bits.provider_oid_is_wildcard);
                ret->callbacks.rcb_sub = rcb;
                ret->minPeriod = minPeriod;
                ret->item = (DOFItem)prop;
                ret->itemId = DOFItem_GetID(ret->item);
                DOFItem_RetainDesc(ret->item);
                PCR_Assert( iid );
                ret->interface_id = iid;
                ret->flags.bits.value_type_is_not_list = 1;
                ret->commandPerm = commandPerm;
                ret->responsePerm = responsePerm;
                ret->operation_type = DOFOPERATIONTYPE_SUBSCRIBE;
                if ( !DOFObjectID_IsUnicast( DOFObject_GetObjectID( obj ) ) )
                    DOFOperationControl_SetDefaultResponseLevel( ret->operationConfig, DOFOPERATIONRESPONSELEVEL_SUCCESS );
                return ret;
            }
            else
                DOFSecurityScope_Destroy(secScope);
        }
        DOFPermission_Destroy(commandPerm);
        DOFPermission_Destroy(responsePerm);
        DOFInterfaceID_Destroy(iid);
    }
    return NULL;
}

/* @internal
Factory method for DOFOperations of type DOFOPERATIONTYPE_BEGINPROVIDE.

@param timeout  The amount of milliseconds that the operation will continue before completing.
@param context  Data that is passed to the callback.
@param obj      The object that this operation is associated with.
@param pcb      The functions that will be called to provide the object.
@param iface    The interface being provided. This must not be NULL.

@return The created DOFOperation or NULL if the operation could not be created or this method was used in error
*/
DOFOperation DOFOperation_BeginProvide_Create(OALChannel provider_channel, uint32 timeout, void * context,
                                              DOFObject obj, const DOFObjectProvider pcb, DOFInterface iface)
{
    DOFInterfaceID iid;
    DOFObjectID oid;
    DOFObject obj_;
    DOFSecurityScope securityScope;
    DOFPermission commandPerm;
    DOFOperationID opId;
    DOFOperation ret = NULL;

    PCR_Assert( iface );
    if (DOFInterface_Retain( iface ) == FALSE)
    {
        goto exit;
    }
    iid = DOFInterfaceID_Copy(DOFInterface_GetInterfaceID(iface));
    if(iid == NULL)
    {
        goto cleanup1;
    }
    oid = DOFObjectID_Copy(DOFObject_GetObjectID(obj));
    if(oid == NULL)
    {
        goto cleanup2;
    }
    obj_ = DOFObject_Copy(obj);
    if(obj_ == NULL)
    {
        goto cleanup3;
    }
    commandPerm = _createNeededPermission(DOFOPERATIONTYPE_BEGINPROVIDE, FALSE, DOFObject_GetObjectID(obj_), iid);
    if(commandPerm == NULL)
    {
        goto cleanup4;
    }
    securityScope = DOFSecurityMembrane_ValidateOutboundPermissions(provider_channel->securityMembrane, commandPerm, commandPerm);
    opId = DOFOperationID_CreateLocal(); 
    if(opId == NULL)
    {
        goto cleanup5;
    }
    ret = _create();
    if (ret)
    {
        ret->operation_type = DOFOPERATIONTYPE_BEGINPROVIDE;
        ret->sources->channel = OALChannel_Retain(provider_channel);
        ret->commandPerm = commandPerm;
        ret->sources->oal_timeout = timeout;
        ret->context = context;
        ret->provider_object = obj_;
        ret->securityScope = securityScope;
        ret->provider_interface = iface;
        ret->callbacks.pcb = pcb;
        PCR_Assert( iid );
        ret->interface_id = iid;
        ret->object_id.object_id = oid;
        ret->opID = opId;
        PCR_Assert( !ret->sources->address );
        return ret;
    }
    /* Fail out. */
    DOFOperationID_Destroy(opId);
cleanup5:    
    DOFSecurityScope_Destroy(securityScope);
    DOFPermission_Destroy(commandPerm);
cleanup4:
    DOFObject_Destroy(obj);
cleanup3:
    DOFObjectID_Destroy(oid);
cleanup2:
    DOFInterfaceID_Destroy(iid);
cleanup1:
    DOFInterface_Destroy(iface);
exit:
    return NULL;
}

/* @internal
Factory method for DOFOperations of type DOFOPERATIONTYPE_QUERY.

@param timeout  The amount of milliseconds that the operation will continue before completing.
@param context  Data that is passed to the callback.

@return The created DOFOperation or NULL if the operation could not be created or this method was used in error
*/
DOFOperation DOFOperation_Query_Create(OALChannel requestor_channel, uint32 timeout, void * context,
                                       DOFQuery  filter, const DOFSystemQueryCallback qcb)
{
    DOFQuery filterCopy = NULL;

    if ( filter )
        filterCopy = DOFQuery_DeepCopy(filter);
    if(filterCopy != NULL || filter == NULL)
    {
        DOFObjectIDList matchingOids = DOFMem_Alloc(sizeof(struct DOFObjectIDList_t));

        if(matchingOids != NULL)
        {
            DOFOperationID opId = DOFOperationID_CreateLocal();

            if(opId != NULL)
            {
                DOFOperation ret = _create();

                if(ret != NULL)
                {
                    ret->operation_type = DOFOPERATIONTYPE_QUERY;

                    ret->sources->oal_timeout = timeout;
                    ret->context = context;
                    ret->callbacks.qcb = qcb;
                    ret->filter = filterCopy;
                    ret->sources->channel = OALChannel_Retain(requestor_channel);

                    ret->securityScope = DOFSecurityScope_Copy(requestor_channel->securityScope);
                    ret->matchingObjIDs = matchingOids;
                    ret->matchingObjIDs->count = 0;
                    ret->matchingObjIDs->list = NULL;
                    ret->opID = opId;
                    return ret;
                }
                DOFOperationID_Destroy(opId);
            }
            DOFMem_Free(matchingOids);
        }
        DOFQuery_Destroy(filterCopy);
    }
    return NULL;
}


/* @internal
     Factory method for DOFOperations of type DOFOPERATIONTYPE_FINDINTERFACEDEFINITION.

    @param timeout  The amount of milliseconds that the operation will continue before completing.
    @param context  Data that is passed to the callback.
    @param sys      The DOFSystem that this operation is associated with.
    @param rcb      The function that will be called when the response is received or timeout occurs.
    @param iid      The Interface ID of the interface being requested.

    @return The created DOFOperation or NULL if the operation could not be created or this method was used in error
*/
DOFOperation DOFOperation_FindInterfaceDefinition_Create(OALChannel requestor_channel, DOFOperationControl operationConfig,
                                                         uint32 timeout, void * context, DOFObject obj,
                                                         const DOFObjectDefineCallback rcb, DOFInterfaceID iid)
{
    PCR_Assert(iid);
    if(!DOFInterfaceID_IsWildcard(iid))
    {
        DOFInterfaceID iidCopy = DOFInterfaceID_Copy(iid);

        if(iidCopy != NULL)
        {
            DOFPermission commandPerm = _createNeededPermission(DOFOPERATIONTYPE_FINDINTERFACEDEFINITION, FALSE, DOFObject_GetObjectID(obj), iid);
            DOFPermission responsePerm = _createNeededPermission(DOFOPERATIONTYPE_FINDINTERFACEDEFINITION, TRUE, DOFObject_GetObjectID(obj), iid);

            if((commandPerm != NULL) && (responsePerm != NULL))
            {
                DOFSecurityScope secScope = DOFSecurityMembrane_ValidateOutboundPermissions(requestor_channel->securityMembrane, commandPerm, commandPerm);

                {
                    DOFOperation ret = _create_r( DOFOPERATIONTYPE_FINDINTERFACEDEFINITION,
                                                  requestor_channel, operationConfig, secScope,
                                                  timeout, context, obj ); /* Takes ownership of secScope. */
                    if(ret != NULL)
                    {
                        ret->callbacks.rcb_def = rcb;
                        ret->interface_id = iidCopy;
                        ret->commandPerm = commandPerm;
                        ret->responsePerm = responsePerm;
                        PCR_Assert( ret->interface_id );
                        ret->flags.bits.commandCompletesOnAnyFinal = 1; /* Directed OAP requests put this EPP header flag on wire.*/
                        return ret;
                    }
                    else
                        DOFSecurityScope_Destroy(secScope);
                }
            }
            DOFPermission_Destroy(commandPerm);
            DOFPermission_Destroy(responsePerm);
            DOFInterfaceID_Destroy(iidCopy);
        }
    }
    return NULL;
}

DOFOperation DOFOperation_Interest_Create(DOFOperationControl opConfig, OALChannel requestor_channel,
                                          DOFObjectID objectID, const DOFInterfaceID interfaceID,
                                          const DOFSystemInterestCallback icb, uint32 timeout,
                                          DOFInterestLevel level, void * context)
{
    DOFOperation ret = NULL;

    DOFOperationType opType;

    DOFPermission commandPerm;

    switch(level)
    {
    case DOFINTERESTLEVEL_WATCH:
        opType = DOFOPERATIONTYPE_WATCH;
        break;

    case DOFINTERESTLEVEL_ACTIVATE:
        opType = DOFOPERATIONTYPE_ACTIVATE;
        break;

    case DOFINTERESTLEVEL_CONNECT:
        opType = DOFOPERATIONTYPE_CONNECT;
        break;

    default:
        return NULL;
    }
    commandPerm = _createNeededPermission(opType, FALSE, objectID, NULL);
    if(commandPerm != NULL)
    {
        DOFSecurityScope secScope =
                DOFSecurityMembrane_ValidateOutboundPermissions(requestor_channel->securityMembrane, commandPerm, commandPerm);

        ret = _create_r(opType, requestor_channel, opConfig, secScope,
                        timeout, context, NULL);  /* Takes ownership of secScope. */
        if(ret != NULL)
        {
            ret->callbacks.icb = icb;
            ret->object_id.object_id = DOFObjectID_Copy(objectID);
            if(ret->object_id.object_id)
            {
                ret->flags.bits.provider_oid_is_wildcard = DOFObjectID_IsWildcard(ret->object_id.object_id);
                ret->commandPerm = commandPerm;
                ret->interface_id = DOFInterfaceID_Copy(interfaceID);
                if(ret->interface_id)
                {
                    return ret;
                }
            }
            _cleanup_r(ret);
        }
        else
            DOFSecurityScope_Destroy(secScope);
        DOFPermission_Destroy(commandPerm);
    }
    return NULL;
}

static uint8* dummyTransform(DOFOperationSessionDataTransform   self,
                             DOFInterfaceID                     sessionId,
                             uint32                           * pSize,
                             uint8                            * buffer)
{
    return buffer;
}

static struct DOFOperationSessionDataTransformFns_t defautlSessionTransformFns = { dummyTransform, dummyTransform };
static DOFOperationSessionDataTransform_t defaultSessionTransform = { &defautlSessionTransformFns };

DOFOperation DOFOperation_Session_Create(OALChannel requestor_channel, DOFObject obj,
                                         const DOFInterface interface,
                                         const DOFInterfaceID sessionIID,  DOFOperationControl operationConfig,
                                         uint32 timeout, const DOFObjectSessionCallback scb, void * ctx)
{
    DOFObject objCopy;

    PCR_Assert(requestor_channel);
    PCR_Assert(obj);

    objCopy = DOFObject_Copy(obj);

    if(objCopy != NULL)
    {
        DOFObjectID objIdCopy = DOFObjectID_Copy(DOFObject_GetObjectID(obj));

        if(objIdCopy != NULL)
        {
            DOFInterfaceID interfaceIDCopy = DOFInterfaceID_Copy(DOFInterface_GetInterfaceID(interface));

            if(interfaceIDCopy != NULL)
            {
                DOFPermission commandPerm = _createNeededPermission(DOFOPERATIONTYPE_SESSION, FALSE, objIdCopy, interfaceIDCopy);
                DOFPermission responsePerm = _createNeededPermission(DOFOPERATIONTYPE_SESSION, TRUE, objIdCopy, interfaceIDCopy);

                if((commandPerm != NULL) && (responsePerm != NULL))
                {
                    DOFInterfaceID sessionIidCopy = DOFInterfaceID_Copy(sessionIID);

                    if(sessionIidCopy != NULL)
                    {
                        if(DOFInterface_Retain(interface))
                        {
                            DOFSecurityScope secScope = DOFSecurityMembrane_ValidateOutboundPermissions(requestor_channel->securityMembrane, commandPerm, commandPerm);

                            DOFOperation ret = _create();

                            if(ret != NULL)
                            {
                                /* This next must be ok to properly call DOFObjectID_Create_AttributeBuffer. */
                                STATIC_ASSERT( sizeof( GUID ) <= 256 );
                                STATIC_ASSERT( sizeof( GUID ) <= 16 );

                                ret->provider_object = objCopy;
                                ret->object_id.object_id = objIdCopy;
                                PCR_Assert( objIdCopy );
                                ret->flags.bits.provider_oid_is_wildcard = DOFObjectID_IsWildcard(objIdCopy);

                                ret->operation_type = DOFOPERATIONTYPE_SESSION;

                                ret->sources->oal_timeout = timeout;
                                ret->context = ctx;
                                ret->sources->channel = OALChannel_Retain(requestor_channel);

                                ret->provider_interface = interface;
                                ret->interface_id = interfaceIDCopy;
                                ret->sessionID = DOF_GenerateGUID();
                                ret->sessionOID = DOFObjectID_Create_AttributeBuffer( objIdCopy, DOFOBJECTIDATTRIBUTE_SESSION,
                                                                                      sizeof(GUID), ret->sessionID.bytes );
                                ret->sessionIID = sessionIidCopy;
                                ret->commandPerm = commandPerm;
                                ret->responsePerm = responsePerm;
                                ret->sessionTransformRequestor = &defaultSessionTransform;
                                ret->sessionTransformProvider = &defaultSessionTransform;

                                if (operationConfig != NULL)
                                {
                                    ret->operationConfig = DOFOperationControl_Copy(operationConfig);
                                }
                                else
                                {
                                    ret->operationConfig = DOFOperationControl_Create();
                                }

                                ret->callbacks.scb = scb;

                                ret->opID = DOFOperationID_CreateLocal();

                                ret->ref_count = 1;
                                ret->securityScope = secScope;
                                if ( ret->object_id.object_id && ret->interface_id && ret->sessionOID && ret->sessionIID && ret->operationConfig && ret->opID )
                                    return ret;
                                DOFOperation_Destroy( ret );
                                return NULL;
                            }
                            else
                            {
                                DOFSecurityScope_Destroy(secScope);
                            }
                            DOFInterface_Destroy(interface);
                        }
                        DOFInterfaceID_Destroy(sessionIidCopy);
                    }
                }
                /* Destroy perms no matter what! */
                DOFPermission_Destroy(commandPerm);
                DOFPermission_Destroy(responsePerm);
                DOFInterfaceID_Destroy(interfaceIDCopy);
            }
            DOFObjectID_Destroy(objIdCopy);
        }
        DOFObject_Destroy(objCopy);
    }
    return NULL;
}

boolean DOFOperationSession_SetDataTransform(DOFOperation                     sessionOperation,
                                             DOFOperationSessionDataTransform transform)
{
    if((sessionOperation != NULL) &&
       (sessionOperation->operation_type == DOFOPERATIONTYPE_SESSION) &&
       (transform != NULL))
    {
        if(sessionOperation->sessionTransformRequestor == &defaultSessionTransform)
        {
            sessionOperation->sessionTransformRequestor = transform;
            return TRUE;
        }
    }
    return FALSE;
}

boolean DOFOperationSession_SetDataTransformProvider(DOFOperation                     sessionOperation,
                                                     DOFOperationSessionDataTransform transform)
{
    PCR_Assert(sessionOperation != NULL);
    PCR_Assert(sessionOperation->operation_type == DOFOPERATIONTYPE_SESSION);
    if(transform != NULL)
    {
        if(sessionOperation->sessionTransformProvider == &defaultSessionTransform)
        {
            sessionOperation->sessionTransformProvider = transform;
            return TRUE;
        }
    }
    return FALSE;
}

static DOFOperation _create_internal(OALChannel requestor_channel, const DOFInterfaceID interfaceID,
                                     const DOFObjectID objectID, const OALAlias alias,
                                     uint32 timeout, uint32 timeoutGrace, DOFOperationType opType)
{
    DOFObjectID oidCopy = (objectID != NULL) ? DOFObjectID_Copy(objectID) : NULL;

    if(objectID != NULL && oidCopy == NULL)
    {
        goto exit;
    }
    else
    {
        DOFInterfaceID iidCopy = DOFInterfaceID_Copy(interfaceID);

        if(interfaceID != NULL && iidCopy == NULL)
        {
            goto cleanup1;
        }
        else
        {
            DOFOperation ret = _create();

            if(ret != NULL)
            {
                if (requestor_channel)
                {
                    ret->sources->channel = OALChannel_Retain(requestor_channel);
                }

                ret->object_id.object_id = oidCopy;
                ret->interface_id = iidCopy;
                PCR_Assert( (objectID && oidCopy) || (!objectID && !oidCopy) );
                ret->flags.bits.provider_oid_is_wildcard = ( objectID ) ? DOFObjectID_IsWildcard(oidCopy) : 0;
                ret->flags.bits.provider_iid_is_wildcard = iidCopy ? DOFInterfaceID_IsWildcard(iidCopy) : FALSE;

                ret->sources->alias = alias;

                ret->sources->oal_timeout = timeout;
                ret->sources->timeoutGrace = timeoutGrace;
                ret->operation_type = opType;

                ret->flags.bits.internally_created = 1;

                return ret;
            }
            DOFInterfaceID_Destroy(iidCopy);
        }
    }

cleanup1:
    DOFObjectID_Destroy(oidCopy);
exit:
    return NULL;
}

DOFOperation DOFOperation_Activate_CreateInternal(OALChannel requestor_channel, const DOFInterfaceID interfaceID,
                                                  const DOFObjectID objectID,  const OALAlias alias, uint32 timeout, uint32 timeoutGrace)
{
    return _createInterestInternal(requestor_channel, interfaceID, objectID,
                                   alias, timeout, timeoutGrace, DOFOPERATIONTYPE_ACTIVATE);
}

DOFOperation DOFOperation_Change_CreateInternal(      OALChannel       requestor_channel,
                                                const DOFInterfaceID   interfaceID,
                                                const DOFObjectID      objectID,
                                                const OALAlias         alias,
                                                      uint32           timeout,
                                                      uint32           timeoutGrace,
                                                      OALStream        valueStream,
                                                      DOFItemID        itemId)
{
    DOFPermission commandPerm = _createNeededPermission(DOFOPERATIONTYPE_CHANGED, FALSE, objectID, interfaceID);

    if(commandPerm != NULL)
    {

        DOFOperation ret = _create_internal(requestor_channel, interfaceID, objectID,
                                            alias, timeout, timeoutGrace, DOFOPERATIONTYPE_CHANGED);

        if(ret != NULL)
        {
            /* The _create_internal function doesn't do exactly what we need */
            ret->operation_type = DOFOPERATIONTYPE_CHANGED;
            ret->commandPerm = commandPerm;
            if(valueStream != NULL)
            {
                OALStream_Retain(valueStream);
                ret->valueStream = valueStream;
            }
            ret->itemId = itemId;
            ret->flags.bits.value_type_is_not_list   = 1;
            return ret;
        }
        DOFPermission_Destroy(commandPerm);
    }
    return NULL;
}

DOFOperation DOFOperation_Connect_CreateInternal(      OALChannel       requestor_channel,
                                                 const DOFInterfaceID   interfaceID,
                                                 const DOFObjectID      objectID,
                                                 const OALAlias         alias,
                                                       uint32           timeout,
                                                       uint32           timeoutGrace)
{
    return _createInterestInternal(requestor_channel, interfaceID, objectID,
                                   alias, timeout, timeoutGrace, DOFOPERATIONTYPE_CONNECT);
}

DOFOperation DOFOperation_FullConnect_CreateInternal(      OALChannel       requestor_channel,
                                                     const DOFInterfaceID   interfaceID,
                                                     const DOFObjectID      objectID,
                                                     const OALAlias         alias,
                                                           uint32           timeout,
                                                           uint32           timeoutGrace)
{
    return _createInterestInternal(requestor_channel, interfaceID, objectID,
                                   alias, timeout, timeoutGrace, DOFOPERATIONTYPE_FULLCONNECT);
}

DOFOperation DOFOperation_Define_CreateInternal(      OALChannel     requestor_channel,
                                                const DOFInterfaceID interfaceID,
                                                const DOFObjectID    objectID,
                                                      uint32         timeout,
                                                      uint32         timeoutGrace)
{
    DOFPermission commandPerm = _createNeededPermission(DOFOPERATIONTYPE_FINDINTERFACEDEFINITION, FALSE, NULL, interfaceID);
    DOFPermission responsePerm = _createNeededPermission(DOFOPERATIONTYPE_FINDINTERFACEDEFINITION, TRUE, NULL, interfaceID);

    if((commandPerm != NULL) && (responsePerm != NULL))
    {
        DOFOperation ret = _create_internal( requestor_channel, interfaceID, objectID,
                                             0, timeout, timeoutGrace, DOFOPERATIONTYPE_FINDINTERFACEDEFINITION );

        if(ret != NULL)
        {
            ret->operation_type = DOFOPERATIONTYPE_FINDINTERFACEDEFINITION;
            ret->commandPerm = commandPerm;
            ret->responsePerm = responsePerm;
            ret->securityScope = DOFSecurityScope_Copy(OALChannel_GetSecurityScope(requestor_channel, TRUE));
            return ret;
        }
    }
    DOFPermission_Destroy(commandPerm);
    DOFPermission_Destroy(responsePerm);
    return NULL;
}

DOFOperation DOFOperation_Get_CreateInternal(      OALChannel       requestor_channel,
                                             const DOFInterfaceID   interfaceID,
                                             const DOFObjectID      objectID,
                                             const OALAlias         alias,
                                                   uint32           timeout,
                                                   uint32           timeoutGrace,
                                                   DOFItemID        itemId)
{
    DOFPermission commandPerm = _createNeededPermission(DOFOPERATIONTYPE_GET, FALSE, objectID, interfaceID);
    DOFPermission responsePerm = NULL;

    if(DOFObjectID_IsUnicast(objectID))
    {
        responsePerm = _createNeededPermission(DOFOPERATIONTYPE_GET, TRUE, objectID, interfaceID);
    }
    if((commandPerm != NULL) && ((responsePerm != NULL) || (!DOFObjectID_IsUnicast(objectID))))
    {
        DOFOperation ret = _create_internal(requestor_channel, interfaceID, objectID,
                                            alias, timeout, timeoutGrace, DOFOPERATIONTYPE_GET);

        if(ret != NULL)
        {
            PCR_Assert((responsePerm == NULL) == ret->flags.bits.provider_oid_is_wildcard);
            ret->itemId = itemId;
            ret->commandPerm = commandPerm;
            ret->responsePerm = responsePerm;
            ret->securityScope = DOFSecurityScope_Copy(OALChannel_GetSecurityScope(requestor_channel, TRUE));
            return ret;
        }
        DOFPermission_Destroy(commandPerm);
        DOFPermission_Destroy(responsePerm);
    }
    return NULL;
}

DOFOperation DOFOperation_Invoke_CreateInternal(      OALChannel       requestor_channel,
                                                const DOFInterfaceID   interfaceID,
                                                const DOFObjectID      objectID,
                                                const OALAlias         alias,
                                                      uint32           timeout,
                                                      uint32           timeoutGrace,
                                                      OALStream        valuesStream,
                                                      DOFItemID        itemId)
{
    DOFPermission commandPerm = _createNeededPermission(DOFOPERATIONTYPE_INVOKE, FALSE, objectID, interfaceID);
    DOFPermission responsePerm = NULL;

    if(DOFObjectID_IsUnicast(objectID))
    {
        responsePerm = _createNeededPermission(DOFOPERATIONTYPE_INVOKE, TRUE, objectID, interfaceID);
    }
    if((commandPerm != NULL) && ((responsePerm != NULL) || (!DOFObjectID_IsUnicast(objectID))))
    {
        DOFOperation ret = _create_internal(requestor_channel, interfaceID, objectID,
                                        alias, timeout, timeoutGrace, DOFOPERATIONTYPE_INVOKE);

        if(ret != NULL)
        {
            PCR_Assert((responsePerm == NULL) == ret->flags.bits.provider_oid_is_wildcard);
            ret->valueStream = valuesStream;
            OALStream_Retain(valuesStream);
            ret->itemId = itemId;
            ret->commandPerm = commandPerm;
            ret->responsePerm = responsePerm;
            ret->securityScope = DOFSecurityScope_Copy(OALChannel_GetSecurityScope(requestor_channel, TRUE));
            return ret;
        }
    }
    DOFPermission_Destroy(commandPerm);
    DOFPermission_Destroy(responsePerm);
    return NULL;
}

DOFOperation DOFOperation_Open_CreateInternal(      OALChannel       requestor_channel,
                                              const DOFInterfaceID   interfaceID,
                                              const DOFObjectID      objectID,
                                              const OALAlias         alias,
                                                    uint32           timeout,
                                                    uint32           timeoutGrace,
                                              const GUID *           sessionID,
                                              const DOFInterfaceID   sessionIID)
{
    DOFInterfaceID sessionIIDCopy = DOFInterfaceID_Copy(sessionIID);

    PCR_Assert( sessionID );
    if(sessionIIDCopy != NULL)
    {
        DOFPermission commandPerm = _createNeededPermission(DOFOPERATIONTYPE_SESSION, FALSE, objectID, interfaceID);
        DOFPermission responsePerm = _createNeededPermission(DOFOPERATIONTYPE_SESSION, TRUE, objectID, interfaceID);

        if((commandPerm != NULL) && (responsePerm != NULL))
        {
            DOFObjectID sessionOID;

            /* This next must be ok to properly call DOFObjectID_Create_AttributeBuffer. */
            STATIC_ASSERT( sizeof( GUID ) <= 256 );

            sessionOID = DOFObjectID_Create_AttributeBuffer( objectID, DOFOBJECTIDATTRIBUTE_SESSION, sizeof( GUID ),
                                                             sessionID->bytes );
            if ( sessionOID )
            {
                DOFOperation ret = _create_internal(requestor_channel, interfaceID, objectID,
                                                    alias, timeout, timeoutGrace, DOFOPERATIONTYPE_SESSION);
                if(ret != NULL)
                {
                    ret->operation_type = DOFOPERATIONTYPE_SESSION;
                    ret->sessionID = *sessionID;
                    ret->sessionOID = sessionOID;
                    ret->sessionIID = sessionIIDCopy;
                    ret->commandPerm = commandPerm;
                    ret->responsePerm = responsePerm;
                    ret->securityScope = DOFSecurityScope_Copy(OALChannel_GetSecurityScope(requestor_channel, TRUE));
                    ret->sessionTransformRequestor = &defaultSessionTransform;
                    ret->sessionTransformProvider = &defaultSessionTransform;
                    return ret;
                }
                DOFObjectID_Destroy( sessionOID );
            }
        }
        DOFPermission_Destroy(commandPerm);
        DOFPermission_Destroy(responsePerm);
        DOFInterfaceID_Destroy(sessionIIDCopy);
    }
    return NULL;
}

DOFOperation DOFOperation_Provide_CreateInternal(      OALChannel       requestor_channel,
                                                 const DOFInterfaceID   interfaceID,
                                                 const DOFObjectID      objectID,
                                                 const OALAlias         alias,
                                                       uint32           timeout,
                                                       uint32           timeoutGrace,
                                                       uint16           losslessHops,
                                                       uint16           lossyHops)
{
    /* This should be the scope from the either the inbound permission set of the security connector, 
    or the peer's permission set from the channel, based on the permission required for the operation. */
    DOFPermission commandPerm = _createNeededPermission(DOFOPERATIONTYPE_PROVIDE, FALSE, objectID, interfaceID);

    if (commandPerm != NULL)
    {
        DOFOperation ret = _create_internal(requestor_channel, interfaceID, objectID,
                                            alias, timeout, timeoutGrace, DOFOPERATIONTYPE_PROVIDE);

        if(ret != NULL)
        {
            ret->operation_type = DOFOPERATIONTYPE_PROVIDE;
            ret->commandPerm = commandPerm;
            ret->sources->distLossless = losslessHops;
            ret->sources->distLossy = lossyHops;
            ret->securityScope = DOFSecurityScope_Copy(OALChannel_GetSecurityScope(requestor_channel, TRUE));
        }
        return ret;
    }
    return NULL;
}

DOFOperation DOFOperation_Advertise_CreateInternal(      OALChannel      requestor_channel,
                                                   const DOFInterfaceID  interfaceID,
                                                   const DOFObjectID     objectID,
                                                         uint32          timeout,
                                                         uint32          timeoutGrace,
                                                         DOFAddress      address)
{
    return _createInterestInternal(requestor_channel, interfaceID, objectID, 0,
                                   timeout, timeoutGrace, DOFOPERATIONTYPE_ADVERTISE);
}

DOFOperation DOFOperation_Set_CreateInternal(      OALChannel       requestor_channel,
                                             const DOFInterfaceID   interfaceID,
                                             const DOFObjectID      objectID,
                                             const OALAlias         alias,
                                                   uint32           timeout,
                                                   uint32           timeoutGrace,
                                                   OALStream        valueStream,
                                                   DOFItemID        itemId)
{
    DOFPermission commandPerm = _createNeededPermission(DOFOPERATIONTYPE_SET, FALSE, objectID, interfaceID);
    DOFPermission responsePerm = NULL;

    if(DOFObjectID_IsUnicast(objectID))
    {
        responsePerm = _createNeededPermission(DOFOPERATIONTYPE_SET, TRUE, objectID, interfaceID);
    }
    if((commandPerm != NULL) && ((responsePerm != NULL) || (!DOFObjectID_IsUnicast(objectID))))
    {
        DOFOperation ret = _create_internal(requestor_channel, interfaceID, objectID,
                                        alias, timeout, timeoutGrace, DOFOPERATIONTYPE_SET);

        if(ret != NULL)
        {
            PCR_Assert((responsePerm == NULL) == ret->flags.bits.provider_oid_is_wildcard);
            ret->valueStream = valueStream;
            OALStream_Retain(valueStream);
            ret->itemId = itemId;
            ret->commandPerm = commandPerm;
            ret->responsePerm = responsePerm;
            ret->securityScope = DOFSecurityScope_Copy(OALChannel_GetSecurityScope(requestor_channel, TRUE));
            return ret;
        }
    }
    DOFPermission_Destroy(commandPerm);
    DOFPermission_Destroy(responsePerm);
    return NULL;
}

DOFOperation DOFOperation_Signal_CreateInternal(      OALChannel       requestor_channel,
                                                const DOFInterfaceID   interfaceID,
                                                const DOFObjectID      objectID,
                                                const OALAlias         alias,
                                                      uint32           timeout,
                                                      uint32           timeoutGrace,
                                                      OALStream        valuesStream,
                                                      DOFItemID        itemId)
{
    DOFPermission commandPerm = _createNeededPermission(DOFOPERATIONTYPE_SIGNAL, FALSE, objectID, interfaceID);

    if(commandPerm != NULL)
    {
        DOFOperation ret = _create_internal(requestor_channel, interfaceID, objectID,
                                            alias, timeout, timeoutGrace, DOFOPERATIONTYPE_SIGNAL);

        if(ret != NULL)
        {
            ret->operation_type = DOFOPERATIONTYPE_SIGNAL;
            ret->commandPerm = commandPerm;
            if(valuesStream != NULL)
            {
                OALStream_Retain(valuesStream);
                ret->valueStream = valuesStream;
            }
            ret->itemId = itemId;
            ret->securityScope = DOFSecurityScope_Copy(OALChannel_GetSecurityScope(requestor_channel, TRUE));
            return ret;
        }
        DOFPermission_Destroy(commandPerm);
    }
    return NULL;
}

static boolean objid_gt(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    return DOFObjectID_Compare( (DOFObjectID) key, (DOFObjectID) item ) > 0;
}
static boolean objid_eq(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    return DOFObjectID_Compare( (DOFObjectID) key, (DOFObjectID) item ) == 0;
}

DOFOperation DOFOperation_Subscribe_CreateInternal(      OALChannel       requestor_channel,
                                                   const DOFInterfaceID   interfaceID,
                                                   const DOFObjectID      objectID,
                                                   const OALAlias         alias,
                                                         uint32           timeout,
                                                         uint32           timeoutGrace,
                                                         DOFItemID        itemId,
                                                         uint16           minimumDelta,
                                                         boolean          not_value_list)
{
    static const DOFOperationType types[] = { DOFOPERATIONTYPE_REGISTER, DOFOPERATIONTYPE_SUBSCRIBE};

    DOFPermission commandPerm = _createNeededPermission(types[not_value_list], FALSE, objectID, interfaceID);
    DOFPermission responsePerm = NULL;

    if(DOFObjectID_IsUnicast(objectID))
    {
        responsePerm = _createNeededPermission(types[not_value_list], TRUE, objectID, interfaceID);
    }
    if((commandPerm != NULL) && ((responsePerm != NULL) || (!DOFObjectID_IsUnicast(objectID))))
    {
        DOFOperation ret = _create_internal(requestor_channel, interfaceID, objectID,
                                            alias, timeout, timeoutGrace, types[not_value_list]);

        if(ret != NULL)
        {
            PCR_Assert((responsePerm == NULL) == ret->flags.bits.provider_oid_is_wildcard);
            ret->itemId = itemId;
            ret->minPeriod = minimumDelta;
            ret->operation_type = types[not_value_list];
            ret->commandPerm = commandPerm;
            ret->responsePerm = responsePerm;
            ret->flags.bits.value_type_is_not_list = not_value_list;
            ret->providers = OALAVLTree_Create(objid_gt, objid_eq, NULL);
            ret->securityScope = DOFSecurityScope_Copy(OALChannel_GetSecurityScope(requestor_channel, TRUE));
            return ret;
        }
    }
    DOFPermission_Destroy(commandPerm);
    DOFPermission_Destroy(responsePerm);
    return NULL;
}

DOFOperation DOFOperation_Watch_CreateInternal(      OALChannel       requestor_channel,
                                               const DOFInterfaceID   interfaceID,
                                               const DOFObjectID      objectID,
                                               const OALAlias         alias,
                                                     uint32           timeout,
                                                     uint32           timeoutGrace)
{
    return _createInterestInternal(requestor_channel, interfaceID, objectID,
                                   alias, timeout, timeoutGrace, DOFOPERATIONTYPE_WATCH);
}

static DOFOperation _createInterestInternal(      OALChannel       requestor_channel,
                                            const DOFInterfaceID   interfaceID,
                                            const DOFObjectID      objectID,
                                            const OALAlias         alias,
                                                  uint32           timeout,
                                                  uint32           timeoutGrace,
                                                  DOFOperationType type)
{
    DOFPermission commandPerm = _createNeededPermission(type, FALSE, objectID, interfaceID);

    if(commandPerm != NULL)
    {
        DOFOperation ret = _create_internal(requestor_channel, interfaceID, objectID,
                                            alias, timeout, timeoutGrace, type);

        if(ret != NULL)
        {
            ret->commandPerm = commandPerm;
            ret->securityScope = DOFSecurityScope_Copy(OALChannel_GetSecurityScope(requestor_channel, TRUE));
            return ret;
        }
        return NULL;
    }
    return NULL;
}

/* Object methods */

static void _internal_cancel(DOFOperation op, boolean deny, boolean oap)
{
    PCR_Assert(op);
    DOFOperation_Lock(op);
    if(deny)
        op->flags.bits.denyTimeoutExtend = 1;
    if(oap)
        op->flags.bits.oap_cancel = 1;
    if (op->flags.bits.timeout_hit)                                       /* Set in timer_observer. */
    {
        op->flags.bits.cancelled = 1;
        OALOperationManager_EnqueueOperationCancel(op);
        if (op->flags.bits.fake_timeout)
        {
            DOFOperation_Internal_Destroy( op );
            op->flags.bits.fake_timeout = 0;
        }
    }
    else if (!op->flags.bits.cancelled)
    {
        boolean timerExtendNotAvailable = FALSE;

        /* Not here for fully processed timeout (.timeout_hit TRUE)- that is caught above. We are here for a cancel. */
        op->flags.bits.cancelled = 1;
        if (op->timer_item)
        {
            DOFError error;

            /* This next atomically includes op->timer_item->flags.firing in its decision, and temporarily locks
               threadpool_mutex to do so. */
            error = _operation_extend( op, 0, TRUE );      /* Force cancel through timer, but not on timer item firing. */
            PCR_Assert( error == DOFERROR_OK || error == DOFERROR_REJECTED || error == DOFERROR_FAILED ||
                        error == DOFERROR_INSUFFICIENT_RESOURCES );
            if (error == DOFERROR_OK)
            {
                op->flags.bits.fake_timeout = 1;
                DOFOperation_Retain(op);
                /* We extended by 0 msecs, so the timer handler routine will very soon call us AGAIN, and we will
                   bump the ref count down in the path above because both .timeout_hit and .fake_timeout are set. */
            }
            else if (DOFERROR_REJECTED)
            {
                /* Here because op->timer_item->flags.firing is true. It means we are here for this race: Op cancel came
                   before our triggered timeout was processed by timer_observer(). Exit now to let timer_observer() set
                   .timeout_hit, finish its other work, then conclude by calling us again. */
            }
            else
                timerExtendNotAvailable = TRUE; /* Here for shutdown, failed timer insert, or insufficient resources. */
        }
        else
            timerExtendNotAvailable = TRUE;
        if (timerExtendNotAvailable)
            OALOperationManager_EnqueueOperationCancel(op);
    }

    DOFOperation_Unlock(op);
}

/** @internal
Cancel an outstanding Operation.

If the operation was still in progress, op's completion flag is set to
indicate the operation is cancelled (readable by DOFOperation_IsComplete()).
op's callback is then called for final cleanup, and not called again for op.

@param  op  The Operation to cancel.
**/
void DOFOperation_Internal_Cancel(DOFOperation op)
{
    _internal_cancel(op, FALSE, FALSE);
}

void DOFOperation_Cancel(DOFOperation op)
{
    if(op)
    {
        _internal_cancel(op, TRUE, FALSE);
    }
}

static void DOFRequest_CleanupRequestTreeWrapper(OALAVLTreeItem_t item, void* arg)
{
    DOFRequest_CleanupRequestTree((DOFRequest)item, NULL);
}

void _DOFOperation_Cancel(DOFOperation op)
{
    /* called with op lock held */

    /* Must not return until DOFSystem has called the callback for the last time and that callback has returned.
     That way, all context data has been deallocated and DOFOperation_Destroy() immediately follow this call. */

    /* Notify all my observers that I am gone and that they should close.
     For them to close means that they would destroy their operation.*/
    PCR_Assert( op->flags.bits.cancelled );
    if(op->operation_will_close) OALObservable_SetChangedAndNotifyObservers(op->operation_will_close);
    op->flags.bits.observers_notified = 1;

    if( op->requestTree )
    {
        /* Happens for SUBSCRIBE, REGISTER, and SESSION requests only. */
        OALAVLTree_Destroy( op->requestTree, DOFRequest_CleanupRequestTreeWrapper, NULL );
        op->requestTree = NULL;
    }

    if(op->operation_type == DOFOPERATIONTYPE_BEGINPROVIDE)
    {
        if( op->flags.bits.already_routed )
        {
            op->flags.bits.oap_cancel = 1;
            _increment_retry(op);
            interest_notify_new_provider(op);
        }
    }                                 /* internally created was commented out here so that a cancel is routed to systems
                                         so that the provider session complete callback will get called. */
    else if (!op->flags.bits.oap_cancel&& /* !op->flags.bits.internally_created&& */ op->interface_id&&op->object_id.object_id&&
             op->sources->channel&&op->flags.bits.added_to_operation_table&&/*!op->address&&*/!op->flags.bits.completed&&
             op->operation_type!=DOFOPERATIONTYPE_CHANGED&&op->operation_type!=DOFOPERATIONTYPE_SIGNAL&&
             (op->flags.bits.commandCompletesOnAnyFinal ? op->firstResult == NULL : TRUE))
    {
        op->flags.bits.oap_cancel = 1;
        if(op->operation_type != DOFOPERATIONTYPE_PROVIDE && op->operation_type != DOFOPERATIONTYPE_BEGINPROVIDE)
        {
            _increment_retry(op);
            DOFRouter_RouteToProvider(op);
        }
    }
}


void DOFOperation_Destroy( DOFOperation op )
{
    if (op)
    {
        DOFOperation_Cancel(op);
        DOFOperation_Internal_Destroy(op);
    }
}

static void internalDOFOperation_DestroySources(OALSListNode node, void *arg)
{
    DOFOperationSource source = (DOFOperationSource)node;
    DOFAddress_Destroy(source->address);
    OALChannel_Release(source->channel);
    DOFRequest_Release(source->request);
    OALPermissionScope_Destroy(source->permissionGroups);
    DOFMem_Free(source);
}

static void DestroyObjectID(OALAVLTreeItem_t item, void *arg)
{
    DOFObjectID_Destroy((DOFObjectID)item);
}

#ifndef DOFOPERATION_DEBUG_REFCOUNT
void DOFOperation_Internal_Destroy(DOFOperation op)
#else
void _DOFOperation_Internal_Destroy(DOFOperation op, const char *fn_name, const char *filename, unsigned int line_number)
#endif
{
    boolean free_ok = FALSE;

    if (op)
    {
        PCRMutex_Lock(ref_count_mutex);
        if ( op->ref_count > 0 )
        {
            if ( op->ref_count == 1 )
            {
                /*
                 * In case the operation isn't already cancelled, do so now.
                 * WARNING: This locks the operation mutex, which generally should not be locked when the ref_count is being decremented.
                 * It is safe if it is the last one, though.
                 * The cancellation may cause the ref count to increase, so we check again afterwards.
                 */
                PCRMutex_Unlock( ref_count_mutex );
                DOFOperation_Cancel( op );
                PCRMutex_Lock( ref_count_mutex );
            }
            if (!--(op->ref_count))
                free_ok = TRUE;

#ifdef DOFOPERATION_DEBUG_REFCOUNT
            fprintf(stderr,"DOFOperation(%p): --refCount = %u from %s at %s:%u\n", (void *)op, ((int)op->ref_count), fn_name, filename, line_number);
#endif
        }
        PCRMutex_Unlock(ref_count_mutex);

        if (free_ok)
        {
            DOFOperation_Lock(op);
            /* This clears out op->timer_item and op->retry_timer_item. */
            OALOperationManager_RemoveOperationAndUnlock( op ); /* Unlocks operation mutex. */

            OALTimerItem_Destroy(op->ackTimerItem, TRUE);

            DOFException_Destroy( op->completeCallbackException );
            if (op->requestTree)
                OALAVLTree_Destroy( op->requestTree, DOFRequest_CleanupRequestTreeWrapper, NULL );
            DOFObject_Destroy( op->provider_object );
            DOFInterface_Destroy( op->provider_interface );
            DOFQuery_Destroy( op->filter );

            if( op->sources )
            {
                OALSList_RemoveAll((OALSListNode*)&op->sources, internalDOFOperation_DestroySources,NULL);
            }

            DOFOperationID_Destroy(op->opID);
            DOFOperationID_Destroy(op->forwardingId);

            if (op->valueStream != NULL)
                OALStream_Release(op->valueStream, TRUE);

            if (op->flags.bits.value_type_is_not_list)
                DOFValue_Destroy( op->val.value );
            else
                DOFValueListAlloc_Destroy( DOFOperation_GetValueListCount(op), op->val.values );
            DOFPermission_Destroy(op->commandPerm);
            DOFPermission_Destroy(op->responsePerm);
            OALObservable_Destroy(op->operation_will_close);
            DOFObjectIDList_Destroy(op->matchingObjIDs);
            DOFSecurityScope_Destroy(op->securityScope);
            DOFOperationControl_Destroy(op->operationConfig);
            if (op->ackTree != NULL)
                OALAVLTree_Destroy(op->ackTree, DestroyObjectID, NULL);
            if (op->providers != NULL)
                OALAVLTree_Destroy(op->providers, DestroyObjectID, NULL);
            DOFObjectID_Destroy(op->sessionOID);
            DOFInterfaceID_Destroy(op->sessionIID);
            PCR_Assert( !op->notifyOpCompleteCond ); /* Because all signalling expected complete before !ref_count*/
            PCR_Assert( !op->notifyOpAcknowledgedCond );
            DOFResult_Destroy(op->firstResult);

            if ( op->item )
                DOFItem_ReleaseDesc(op->item);
            if (op->interface_id)
                DOFInterfaceID_Destroy(op->interface_id);

            if (op->object_id.object_id)
                DOFObjectID_Destroy(op->object_id.object_id);

            if (op->tepContextDestroyFunction != NULL)
                op->tepContextDestroyFunction(op->tepContext);

            if (op->contextDestroyFunction != NULL)
                op->contextDestroyFunction(op->context);

            OALMutexPool_Finalize( op );
            DOFMem_Free(op->cancelContext);
            DOFMem_Free(op);
        }
    }
}

boolean DOFOperation_SendOAPNoOp( OALChannel channel )
{
    if(channel)
    {
        DOFOperation op;

        op = _create();

        if(op)
        {
            op->operation_type = DOFOPERATIONTYPE_NOOP;
            op->sources->channel = OALChannel_Retain(channel);
            channel->routeToProvider(channel, op, NULL, NULL);

            DOFOperation_Internal_Destroy(op);
            return TRUE;
        }
    }
    return FALSE;
}

void * DOFOperation_GetContext( const DOFOperation op )
{
    if ( !op )
        return NULL;
    return op->context;
}

DOFObject DOFOperation_GetObject( const DOFOperation op )
{
    if (!op || (op->operation_type==DOFOPERATIONTYPE_QUERY)) return NULL;
    return op->provider_object;
}

DOFQuery DOFOperation_GetQuery( const DOFOperation op )
{
    if (!op || (op->operation_type!=DOFOPERATIONTYPE_QUERY)) return NULL;
    return op->filter;
}


DOFInterfaceMethod DOFOperation_GetMethod( const DOFOperation op )
{
    if (!op || (op->operation_type!=DOFOPERATIONTYPE_INVOKE)) return NULL;
    return((DOFInterfaceMethod)op->item);
}


DOFInterfaceProperty DOFOperation_GetProperty( const DOFOperation op )
{
    if(!op || ( (op->operation_type!=DOFOPERATIONTYPE_GET) &&
                (op->operation_type!=DOFOPERATIONTYPE_SET) &&
                (op->operation_type!=DOFOPERATIONTYPE_SUBSCRIBE) )) return NULL;
    return((DOFInterfaceProperty)op->item);
}

DOFInterfaceEvent DOFOperation_GetEvent( const DOFOperation op )
{
    if(!op || (op->operation_type!=DOFOPERATIONTYPE_REGISTER)) return NULL;
    return((DOFInterfaceEvent)op->item);
}

DOFOperationID DOFOperation_GetOperationID( const DOFOperation op )
{
    if ( !op )
        return NULL;
    return op->opID;
}


uint32 DOFOperation_GetTimeRemaining( const DOFOperation op )
{
    uint32 ret = 0;
    if ( op )
    {
        DOFOperation_Lock( op );
        if ( op->timer_item )
        {
            ret = OALTimerItem_TimeRemaining(op->timer_item);
            /* If an operation had a timeout, and was later changed to DOF_TIMEOUT_NEVER the operation,
               will still have a timer item. */
            if(ret > op->sources->timeoutGrace)
            {
                ret -= op->sources->timeoutGrace;
            }
            else
            {
                ret = 0;
            }
            if( (ret == 0) && (op->sources->oal_timeout == DOF_TIMEOUT_NEVER) && !op->flags.bits.cancelled )
            {
                ret = DOF_TIMEOUT_NEVER;
            }
        }
        else if( !op->flags.bits.cancelled )
        {
            ret = DOF_TIMEOUT_NEVER;
        }
        DOFOperation_Unlock( op );
    }
    return ret;
}


DOFSystem DOFOperation_GetSystem( const DOFOperation op )
{
    if ( !op )
        return NULL;
    return (DOFSystem)op->sources->channel;
}


DOFOperationType DOFOperation_GetType( const DOFOperation op )
{
    if ( !op )
        return DOFOPERATIONTYPE_INVALID;
    return op->operation_type;
}


boolean DOFOperation_SetTimeout( DOFOperation op, uint32 newTimeout )
{
    if ( !op )
        return FALSE;
    DOFOperation_Lock(op);
    if (!op->flags.bits.denyTimeoutExtend)
    {
        if (newTimeout)
        {
            if (!op->flags.bits.cancelled)
            {
                /* If current timeout is 0 it is very likely that we are called by the application callback that was
                   notified of timeout and wants to prevent the operation from closing. This an operation "save."
                   Whether this is a save or not, newTimeout is legit and we are not denied extending, so:
                            If this is a "save," inform DOFSystem that we extended the timeout.*/
                /* One strange way this next call will fail is if op->timer_item is firing at this moment. In that case,
                   the call does nothing and op will soon be serviced as a timeout. */
                if (DOFERROR_OK == _operation_extend( op, newTimeout, FALSE ))
                {
                    op->flags.bits.completed = 0;     /* If not a "save," this harmless because .completed already 0. */
                    if ( op->operation_type < DOFOPERATIONTYPE_INVALID && op->operation_type != DOFOPERATIONTYPE_BEGINPROVIDE && !op->flags.bits.lost )
                    {
                        DOFOperation_IncrementRetry(op);
                        DOFRouter_RouteToProvider(op);
                    }
                }
            }
            else
            { /* Previous code paths mandated that the operation is irreversably going down. */
            }
        }
        else
        {
            /* New timeout is 0, which we interpret as a command to perform an immediate timeout. We need not actually
               change timeout value to 0. Instead, call the timeout's callback here to do the same thing a true timeout
               would have done. */
            {
                DOFSystem_OpTimedOutCallback( op );
                /* If app callback X called us, it is required that this not initiate a call to that same app callback X
                   until it has returned. DOFSystem code guarantees that. */
            }
            /* Why we do nothing if either op bits "cancelled" or "completedAck" are set: In both cases, previous code
               paths mandated that the operation is irreversably going down. */
        }
    }
    else
    {
        /* op was explicitly cancelled OR DOFSystem code denied timeout extensions because it is shutting down the
           operation. In both cases, the operation is irreversably going down. */
    }
    DOFOperation_Unlock(op);
    return TRUE;
}


boolean DOFOperation_IsCancelled( const DOFOperation op )
{
    if ( !op )
        return FALSE;
    else
    {
        boolean cancelled;

        DOFOperation_Lock(op);
        cancelled = op->flags.bits.cancelled;
        DOFOperation_Unlock(op);
        return cancelled;
    }
}


boolean DOFOperation_IsComplete( const DOFOperation op )
{
    /* Note: Unit test is found in system/systemCallback_unit.c, testEight_OpWaitComplete. */
    /* In response to an operation completing by any means, DOFSystem makes a
    final call to the operation's callback. DOFSystem must set the operation's
    "complete" flag before that call, so callback code can cleanup context
    based upon its DOFOperation_IsComplete() test. */
    if ( !op )
        return FALSE;
    else
    {
        boolean retVal;

        DOFOperation_Lock(op);
        retVal = (boolean)op->flags.bits.completedAck;
        DOFOperation_Unlock(op);

        return retVal;
    }
}


boolean DOFOperation_IsAcknowledged( const DOFOperation op )
{
    /* Note: Unit test is found in system/systemCallback_unit.c, testNine_OpWaitAcknowledged. */
    if ( !op )
        return FALSE;
    else
    {
        boolean retVal;

        DOFOperation_Lock(op);
        retVal = op->flags.bits.acknowledged;
        DOFOperation_Unlock(op);

        return retVal;
    }
}


boolean DOFOperation_WaitComplete( const DOFOperation op, uint32 timeout, DOFException * pException )
{
    /* Note: Unit test is found in system/systemCallback_unit.c, testEight_OpWaitComplete. */
    boolean retStatus = TRUE;

    PCR_Assert( op );

    DOFOperation_Lock(op);
    /* Static DOFSystem calls ProvideHandler() and ResponseHandler() lock their provOp/respOp mutex also, except while
       they call their user callback. The callbacks must not manipulate op flags without setting op mutex. */
    if (!op->flags.bits.completedAck)
    {
        if (!op->notifyOpCompleteCond)
        {
            op->notifyOpCompleteCond = PCRCondition_Create();
            PCR_Assert( !op->flags.bits.completedAck );     /* Because op is locked. */
            PCR_Assert( !op->completeWaitInstanceCount );
        }
        if (op->notifyOpCompleteCond)
        {
            DOFError status;

            ++op->completeWaitInstanceCount;                /* All manipulation is right here, under DOFOp_Lock(). */
            PCR_Assert( op->completeWaitInstanceCount );
            PCR_Assert( !op->flags.bits.completedAck );     /* Because op is locked. */
            status = PCRCondition_Wait( op->notifyOpCompleteCond, op->mutex, timeout );
            PCR_Assert( status == PCRSTATUS_OK || status == PCRSTATUS_TIMEOUT );
            if (status == PCRSTATUS_TIMEOUT || !op->flags.bits.completed || !op->flags.bits.completedAck)
            {
                retStatus = FALSE;                          /* op not complete- Our _Wait timed out. */
                if ( pException )
                    *pException = DOFException_Copy( DOFEXCEPTION_TIMEOUT );
            }
            else
            {
                PCR_Assert( retStatus != FALSE );           /* op has completed. */
                PCR_Assert( op->flags.bits.completed );
                PCR_Assert( op->flags.bits.completedAck );
                if ( pException )
                    *pException = DOFException_Copy( op->completeCallbackException );
            }
            PCR_Assert( op->completeWaitInstanceCount );    /* Multiple threads can call _WaitComplete() for same op. */
            if (!--op->completeWaitInstanceCount)
            {
                PCRCondition_Destroy( op->notifyOpCompleteCond ); /* Only in this path was the condition var created. */
                op->notifyOpCompleteCond = NULL;                  /* DOFSystem chooses behavior based on NULL or not. */
            }
        }
        else
        {
            /* Error: Though we know op has not called its callback the final time, we have not and cannot timeout. */
            retStatus = FALSE;       /* op->notifyOpCompleteCond create failure (out of memory?) */
            if ( pException )
                *pException = DOFException_Copy( DOFEXCEPTION_INSUFFICIENT_RESOURCES );
        }
    }
    else
    {
        if ( pException )
            *pException = DOFException_Copy( op->completeCallbackException );
    }
    DOFOperation_Unlock(op);

    return retStatus;
}


boolean DOFOperation_WaitAcknowledged( const DOFOperation op, uint32 timeout, DOFException * pException )
{
    /* Note: Unit test is found in system/systemCallback_unit.c, testNine_OpWaitAcknowledged_CompleteTriggered,
       testTen_OpWaitAcknowledged_OpTimeoutFirst, testEleven_OpWaitAcknowledged_WaitTimeoutFirst, and
       testTwelve_OpWaitAcknowledged_SignalTriggered. */
    boolean retStatus = TRUE;

    /* Note: For request operations, op->flags.bits.acknowledged is set at first provider response.
             For provider operations, op->flags.bits.acknowledged is set right away. */
    PCR_Assert( op );
    DOFOperation_Lock( op );
    if (!op->flags.bits.acknowledged && !op->flags.bits.completedAck)
    {
        if (!op->notifyOpAcknowledgedCond)
        {
            op->notifyOpAcknowledgedCond = PCRCondition_Create();
            PCR_Assert( !op->flags.bits.completedAck ); /* Because op is locked. */
            PCR_Assert( !op->acceptWaitInstanceCount );
        }
        if (op->notifyOpAcknowledgedCond)
        {
            op->acceptWaitInstanceCount++;              /* All manipulation is right here, under DOFOp_Lock(). */
            PCR_Assert( op->acceptWaitInstanceCount );
            PCR_Assert( !op->flags.bits.completedAck ); /* Because op is locked. */
            PCRCondition_Wait( op->notifyOpAcknowledgedCond, op->mutex, timeout );
            /* We were signaled by either DOFSystem or by op timeout. We don't check for who or why. */
            if (!op->flags.bits.acknowledged)
            {
                retStatus = FALSE;      /* op is not accepted. */
                if (pException)
                {
                    if ( !op->flags.bits.completedAck )
                        *pException = DOFException_Copy( DOFEXCEPTION_TIMEOUT );
                    else
                    {
                        if(op->firstResult != NULL)
                        {
                            *pException = DOFException_Copy(DOFResult_GetException(op->firstResult));
                        }
                        else if(op->completeCallbackException != NULL)
                        {
                            *pException = DOFException_Copy( op->completeCallbackException );
                        }
                    }
                }
            }
            else if (pException)
            {
                if(op->firstResult != NULL)
                {
                    *pException = DOFException_Copy(DOFResult_GetException(op->firstResult));
                }
                else if(op->completeCallbackException != NULL)
                {
                    *pException = DOFException_Copy( op->completeCallbackException );
                }
            }
            PCR_Assert( op->acceptWaitInstanceCount );  /* Multiple threads can call _WaitAcknowledged() for same op. */
            if (!--op->acceptWaitInstanceCount)
            {
                PCRCondition_Destroy( op->notifyOpAcknowledgedCond ); /* Only in this path was condition var created. */
                op->notifyOpAcknowledgedCond = NULL;              /* DOFSystem chooses behavior based on NULL or not. */
            }
        }
        else
        {
            /* Error: Though we know op has not been accepted, we have not and cannot timeout. */
            retStatus = FALSE;      /* op->notifyOpAcknowledgedCond create failure (out of memory?) */
            if (pException)
                *pException = DOFException_Copy( DOFEXCEPTION_INSUFFICIENT_RESOURCES );
        }
    }
    else
    {
        retStatus = op->flags.bits.acknowledged;
        if (pException)
        {
            if(op->firstResult != NULL)
            {
                *pException = DOFException_Copy(DOFResult_GetException(op->firstResult));
            }
            else if(op->completeCallbackException != NULL)
            {
                *pException = DOFException_Copy( op->completeCallbackException );
            }
        }
    }
    DOFOperation_Unlock( op );

    return retStatus;
}

static DOFResult _getresult(DOFOperation op, DOFException *pException)
{
    DOFException except    = NULL;
    DOFResult    result    = NULL;

    PCR_Assert( op );
    if (op->firstResult)
    {
        if (!(except = DOFResult_GetException( op->firstResult )))
            result = op->firstResult;
    }
    else
        except = DOFEXCEPTION_FAILED;
    if ( pException )
        *pException = DOFException_Copy( except );

    return result;
}

static DOFResult _waitresult(DOFOperation op, uint32 timeout, DOFException *pException)
{
    boolean      doCleanup = FALSE;
    DOFException except    = NULL;
    DOFResult    result    = NULL;

    PCR_Assert( op );
    {
        if ( DOFOperation_WaitAcknowledged( op, timeout, &except ) )
        {
            DOFException_Destroy( except );
            if ((result = _getresult( op, &except )) == NULL)
                doCleanup = TRUE;
            else
            {
                /*On failure, this approach sets except AND also returns NULL, matching public API for calls that care*/
                if ( !(result = DOFResult_Copy( result ) ) )
                    except = DOFEXCEPTION_INSUFFICIENT_RESOURCES;
            }
        }
        else
        {
            if ( except )
                doCleanup = TRUE;
            else
                except = DOFEXCEPTION_FAILED;
        }
    }
    if ( pException )
        *pException = DOFException_Copy( except );
    if (doCleanup)
        DOFException_Destroy( except );

    return result;
}

DOFResult DOFOperationGet_WaitResult( const DOFOperation op, uint32 timeout, DOFException *pException )
{
    /* this call must only be made with get operations */
    PCR_Assert( op );
    PCR_Assert( op->operation_type==DOFOPERATIONTYPE_GET );

    return _waitresult(op, timeout, pException);
}

DOFResult DOFOperationSet_WaitResult( const DOFOperation op, uint32 timeout, DOFException *pException )
{
    /* this call must only be made with set operations */
    PCR_Assert( op );
    PCR_Assert( op->operation_type==DOFOPERATIONTYPE_SET );

    return _waitresult(op, timeout, pException);
}

DOFResult DOFOperationInvoke_WaitResult( const DOFOperation op, uint32 timeout, DOFException *pException )
{
    /* this call must only be made invoke operations */
    PCR_Assert( op );
    PCR_Assert( op->operation_type==DOFOPERATIONTYPE_INVOKE );

    return _waitresult(op, timeout, pException);
}

DOFResult DOFOperationGet_GetResult( const DOFOperation op, DOFException *pException )
{
    /* this call must only be made with get operations */
    PCR_Assert( op );
    PCR_Assert( op->operation_type==DOFOPERATIONTYPE_GET );

    return _getresult(op, pException);
}

DOFResult DOFOperationSet_GetResult( const DOFOperation op, DOFException *pException )
{
    /* this call must only be made with set operations */
    PCR_Assert( op );
    PCR_Assert( op->operation_type==DOFOPERATIONTYPE_SET );

    return _getresult(op, pException);
}

DOFResult DOFOperationInvoke_GetResult( const DOFOperation op, DOFException *pException )
{
    /* this call must only be made invoke operations */
    PCR_Assert( op );
    PCR_Assert( op->operation_type==DOFOPERATIONTYPE_INVOKE );

    return _getresult(op, pException);
}


DOFException DOFOperation_GetException( const DOFOperation op )
{
    DOFException ret = NULL;
    if ( op )
    {
        DOFOperation_Lock( op );
        ret = op->completeCallbackException;
        DOFOperation_Unlock( op );
    }
    return ret;
}

/** @internal
    Lock the Operation by blocking until the lock can be obtained.

    @param  op The Operation to lock.

    @see DOFOperation_Unlock
**/
void DOFOperation_Lock(DOFOperation op)
{
    PCRMutex mutex;

    PCR_Assert(op);
    mutex = OALMutexPool_Get(op);
    PCRMutex_Lock(mutex);
    op->mutex = mutex;
}

/** @internal
    Unlock the Operation.

    @param  op The Operation to unlock.

    @see DOFOperation_Lock
**/
void DOFOperation_Unlock(DOFOperation op)
{
    PCR_Assert(op);
    PCR_Assert(OALMutexPool_GetNoRef(op)==op->mutex);
    PCRMutex_Unlock(op->mutex);
    OALMutexPool_Release(op);
}

#ifndef DOFOPERATION_DEBUG_REFCOUNT
boolean DOFOperation_Retain(DOFOperation op)
#else
boolean _DOFOperation_Retain(DOFOperation op, const char *fn_name, const char *filename, unsigned int line_number)
#endif
{
    boolean ret = TRUE;
    PCR_Assert(op);
    PCRMutex_Lock(ref_count_mutex);
#ifdef DOFOPERATION_DEBUG_REFCOUNT
        if( op->ref_count == 0 )
            fprintf(stderr,"DOFOperation(%p): op refCount is zero, will assert. %s, at %s:%u\n", (void *)op, fn_name, filename, line_number);
#endif

    if( op->ref_count == 0 ) /* For operations, we agreed that no case should exist that bumps up from 0, but if it does let the caller decide what to do */
        ret = FALSE;

    ++op->ref_count;
#ifdef DOFOPERATION_DEBUG_REFCOUNT
    fprintf(stderr,"DOFOperation(%p): ++refCount = %u from %s at %s:%u\n", (void *)op, (unsigned int)op->ref_count, fn_name, filename, line_number);
#endif
    PCRMutex_Unlock(ref_count_mutex);
    return ret;
}

const uint8 * DOFOperationID_GetOIDBytes (const DOFOperationID id)
{
    PCR_Assert(id);
    return (id->sid->oid ? DOFObjectID_GetBytes( DOFObjectIDSource_GetObjectID( id->sid->oid ) ) : NULL);
}

DOFObjectID DOFOperationID_GetObjectID(const DOFOperationID id)
{
    PCR_Assert(id);
    return DOFObjectIDSource_GetObjectID( id->sid->oid );
}

DOFOperationCount DOFOperationID_GetCount (const DOFOperationID id)
{
    PCR_Assert(id);
    return id->cnt&0x7fffffff;
}

uint16 DOFOperationID_GetSizeMem(void)
{
    return sizeof(struct DOFOperationID_t) - offsetof(struct DOFOperationID_t, cnt);
}

uint16 DOFOperationID_GetSize(const DOFOperationID id)
{
    if (!id || !id->sid->oid)
        return 0;

    return sizeof(id->cnt) + DOFObjectID_GetBytesSize(DOFObjectIDSource_GetObjectID( id->sid->oid ));
}

static void DOFOperation_ActuallySetLocalOID(DOFObjectIDSource newOid)
{
    DOFObjectIDSource tempId;

    PCR_Assert(newOid != NULL);
    tempId = DOFObjectIDSource_Copy( newOid );
    if(tempId != NULL)
    {
        sid_t * tempSid = create_sid_from_oid(tempId, NULL);

        if(tempSid != NULL)
        {
            /* Let's clean up the old state... */
            if (local_oid != NULL)
            {
                DOFObjectIDSource_Destroy(local_oid);
            }
            if ( local_sid )
            {
                source_release(local_sid);
            }
            /* Copy the temps and we're good to go... */
            local_oid = tempId;
            local_sid = tempSid;
        }
        else
        {
            DOFObjectIDSource_Destroy(tempId);
        }
    }
}

void DOFOperation_SetLocalOID(DOFObjectIDSource newOid)
{
    if (is_initialized)
    {
        PCRMutex_Lock(local_oid_mutex);
        DOFOperation_ActuallySetLocalOID(newOid);
        PCRMutex_Unlock(local_oid_mutex);
    }
    else
    {
        DOFOperation_ActuallySetLocalOID(newOid);
    }
}

DOFError DOFOperationID_Marshal(OALStream stream, DOFOperationID id)
{
    DOFError result = DOFERROR_INTERNAL_ERROR;

    PCR_Assert(stream&&id);
    {
        const uint8 * oid_bytes;

        oid_bytes = DOFOperationID_GetOIDBytes(id);
        if (!oid_bytes || !OALStream_Write( stream, DOFObjectID_GetSize_Bytes(oid_bytes), oid_bytes ))
            return DOFERROR_INTERNAL_ERROR;

        result = OALMarshal_StreamCompress4(stream, DOFOperationID_GetCount(id));
        /* fall through */
    }

    return result;
}

DOFOperationID DOFOperationID_Create(DOFOperationCount cnt, sid_t *sid, boolean implicit, boolean get_ref)
{
    DOFOperationID opid = NULL;

    /* This is the only check to protect against creating a sid after we are shuting down */
    if(is_initialized)
    {
        opid = DOFMem_Alloc(sizeof(DOFOperationID_t));
        if(!opid) 
            return NULL;

        opid->ref_count = 1;

        if(get_ref) 
            source_retain(sid);

        opid->sid = sid;
        opid->cnt = cnt;

        if(implicit)
            opid->cnt = cnt|CNT_IMPLICIT_MASK; /* set flag indicating implicit */
    }

    return opid;
}

DOFOperationID DOFOperationID_CreateLocal(void)
{
    DOFOperationID ret;

    PCRMutex_Lock(local_oid_mutex); 
    ret = DOFOperationID_Create(opcnt++, local_sid, FALSE, TRUE); /* send explicit */
    PCRMutex_Unlock(local_oid_mutex);

    return ret;
}

DOFOperationID DOFOperationID_Copy(DOFOperationID id)
{
    if(!id) return NULL;

    PCRMutex_Lock(local_oid_mutex);
    PCR_Assert(id->ref_count != 0);
    ++(id->ref_count);
    PCRMutex_Unlock(local_oid_mutex);

    return id;
}

boolean DOFOperationID_IsImplicit(DOFOperationID id)
{
    return ((id->cnt&CNT_IMPLICIT_MASK)!=0);
}

boolean DOFOperationID_CanBeExplicit(DOFOperationID id)
{
    if (id->sid->oid || id->sid->explicitSid)
        return TRUE;

    return FALSE;
}

void DOFOperationID_UpdateSid(DOFOperationID id, sid_t * newSid)
{
    sid_t * oldSid = id->sid;
    source_retain(newSid);
    id->sid = newSid;
    source_release(oldSid);
}

boolean DOFOperationID_IsSidEqual(DOFOperationID id, sid_t * sid)
{
    if (id->sid == sid)
        return TRUE;

    return FALSE;
}

void DOFOperationID_Destroy(DOFOperationID opid)
{
    if (opid)
    {
        PCRMutex_Lock(local_oid_mutex);
        PCR_Assert( opid->ref_count );
        if (--(opid->ref_count) == 0) {
            source_release(opid->sid);
            DOFMem_Free(opid);
        }
        PCRMutex_Unlock( local_oid_mutex );
    }
}

DOFObjectID DOFOperationID_GetLocalOID(void)
{
    return DOFObjectIDSource_GetObjectID( local_oid );
}

sid_t *DOFOperationID_GetSID(DOFOperationID opid)
{
    PCR_Assert(opid);
    return opid->sid;
}

sid_t *DOFOperationID_GetLocalSID(void)
{
    return local_sid;
}

void *DOFOperationID_GetOPIDPtr(DOFOperationID opid)
{
    PCR_Assert(opid);

    return &((DOFOperationID_t*)opid)->cnt;
}

OALObserver DOFOperation_AddOpWillCloseObserverOtherArg(DOFOperation op, OALObserver_fn_t fn, void *arg)
{
    OALObserver obs = NULL;

    DOFOperation_Lock(op);
    if(!op->flags.bits.cancelled && (obs = OALObservable_AddObserver(op->operation_will_close, fn, arg))) {
        DOFOperation_Retain(op);
    }
    DOFOperation_Unlock(op);

    return obs;
}

void DOFOperation_RemoveOpWillCloseObserver(DOFOperation op, OALObserver obs)
{
    DOFOperation_Lock(op);
    if ( obs )
    {
        OALObservable_RemoveObserver( op->operation_will_close, obs );
        /* If the observer hasn't been called yet, we need to decrement the ref count from when it was added.
           Otherwise, the callback should have already taken care of that. */
        if ( !op->flags.bits.observers_notified )
            DOFOperation_Internal_Destroy( op );
    }
    DOFOperation_Unlock(op);
}

/**
    @internal
    @note When callback function fn is eventually called, it must release op, i.e.: call DOFOperation_Destroy( op ).
    @param op   DOFOperation that, when finally closing, will have its "closing callback" called with "arg param" op.
    @param fn  Callback function called when op will close.

    @return The observer.
**/
#ifndef DOFOPERATION_DEBUG_REFCOUNT
OALObserver DOFOperation_AddOpWillCloseObserver(DOFOperation op, OALObserver_fn_t fn)
#else
OALObserver _DOFOperation_AddOpWillCloseObserver(DOFOperation op, OALObserver_fn_t fn, const char *fn_name, const char *filename, unsigned int line_number)
#endif
{
    OALObserver obs = NULL;

    DOFOperation_Lock(op);
    if(!op->flags.bits.cancelled && (obs = OALObservable_AddObserver(op->operation_will_close, fn, op)))
    {
#ifndef DOFOPERATION_DEBUG_REFCOUNT
        DOFOperation_Retain(op);
#else
        _DOFOperation_Retain(op, fn_name, filename, line_number);
#endif
    }

    DOFOperation_Unlock(op);
    return obs;
}

static void _increment_retry(DOFOperation op)
{
    ++(op->sources->retry_count);
    op->flags.bits.new_retry_count = 1;
}

void DOFOperation_IncrementRetry(DOFOperation op)
{
    DOFOperation_Lock(op);
    _increment_retry(op);
    DOFOperation_Unlock(op);
}

/*  This is an internal function to set the oal_timeout field so that the delay sent in the EPP header is correct.
    Note: This must be called after DOFOperation_SetTimeout and OALOperationManager_AddOperation. */
void DOFOperation_SetDelay(DOFOperation op, uint32 timeout)
{
    PCR_Assert(op);
    op->sources->oal_timeout = timeout;
}

OALChannel DOFOperation_GetSourceChannel(DOFOperation op)
{
    PCR_Assert(op);
    return op->sources->channel;
}

DOFSecurityScope DOFOperation_GetSecurityScope( const DOFOperation op )
{
    PCR_Assert(op);
    return (op->securityScope);
}

boolean DOFOperation_IsBetterSource(DOFOperation op,
                                    uint16       distLossless,
                                    uint16       distLossy)
{
    uint32 newDist = ( (uint32)(op->sources->distLossy) << 15 ) + op->sources->distLossless;
    uint32 oldDist = ( (uint32)(distLossy) << 15 ) + distLossless;
    return ( newDist < oldDist );
}

uint16 DOFOperation_GetValueListCount( DOFOperation op )
{
    if ( op && op->item )
    {
        switch ( DOFItem_GetType(op->item) )
        {
            case DOFITEM_EVENT:
                return DOFInterfaceEvent_GetParameterCount((DOFInterfaceEvent)op->item);
            case DOFITEM_METHOD:
                return DOFInterfaceMethod_GetInputParameterCount((DOFInterfaceMethod)op->item);
            default:
                /* Nothing else should have a list... */
                break;
        }
    }
    return 0;
}

DOFOperationControl DOFOperation_GetControl(DOFOperation op)
{
    DOFOperationControl config;
    DOFOperation_Lock(op);
    config = op->operationConfig;
    DOFOperation_Unlock(op);
    /* DOFOperationControl is immutable, so cannot change. */
    return config;
}

DOFObject DOFOperationGet_GetObject(const DOFOperation op)
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_GET);

    return DOFOperation_GetObject(op);
}

DOFInterfaceProperty DOFOperationGet_GetProperty( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_GET);

    return DOFOperation_GetProperty(op);
}

DOFOperationControl DOFOperationGet_GetControl( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_GET);

    return DOFOperation_GetControl(op);
}

DOFObject DOFOperationSet_GetObject( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_SET);

    return DOFOperation_GetObject(op);
}

DOFInterfaceProperty DOFOperationSet_GetProperty( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_SET);

    return DOFOperation_GetProperty(op);
}

DOFOperationControl DOFOperationSet_GetControl( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_SET);

    return DOFOperation_GetControl(op);
}

DOFObject DOFOperationInvoke_GetObject( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_INVOKE);

    return DOFOperation_GetObject(op);
}

DOFInterfaceMethod DOFOperationInvoke_GetMethod( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_INVOKE);

    return DOFOperation_GetMethod(op);
}

DOFOperationControl DOFOperationInvoke_GetControl( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_INVOKE);

    return DOFOperation_GetControl(op);
}

DOFObject DOFOperationSubscribe_GetObject( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_SUBSCRIBE);

    return DOFOperation_GetObject(op);
}

DOFInterfaceProperty DOFOperationSubscribe_GetProperty( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_SUBSCRIBE);

    return DOFOperation_GetProperty(op);
}

DOFOperationControl DOFOperationSubscribe_GetControl( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_SUBSCRIBE);

    return DOFOperation_GetControl(op);
}

uint32 DOFOperationSubscribe_GetMinPeriod( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_SUBSCRIBE);

    return op->minPeriod;
}

DOFObject DOFOperationRegister_GetObject( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_REGISTER);

    return DOFOperation_GetObject(op);
}

DOFInterfaceEvent DOFOperationRegister_GetEvent( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_REGISTER);

    return DOFOperation_GetEvent(op);
}

DOFOperationControl DOFOperationRegister_GetControl( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_REGISTER);

    return DOFOperation_GetControl(op);
}

DOFObject DOFOperationSession_WaitSession( const DOFOperation op, uint32 timeout, DOFException * pException )
{
    DOFResult result;
    DOFObject ret = NULL;

    PCR_Assert( op );
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_SESSION);

    result = _waitresult(op, timeout, pException);

    if(result) {
        ret = DOFSystem_CreateObject( DOFObject_GetSystem(op->provider_object), op->sessionOID );
        DOFResult_Destroy(result);
    }

    return ret;
}

DOFObject DOFOperationSession_GetObject( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_SESSION);

    return DOFOperation_GetObject(op);
}

DOFInterface DOFOperationSession_GetInterface( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_SESSION);

    return op->provider_interface;
}

DOFInterfaceID DOFOperationSession_GetSessionType( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_SESSION);

    return op->sessionIID;
}

DOFOperationControl DOFOperationSession_GetControl( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_SESSION);

    return DOFOperation_GetControl(op);
}

DOFObject DOFOperationProvide_GetObject( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_BEGINPROVIDE);

    return DOFOperation_GetObject(op);
}

DOFInterface DOFOperationProvide_GetInterface( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_BEGINPROVIDE);

    return op->provider_interface;
}

DOFSystem DOFOperationInterest_GetSystem( const DOFOperation op )
{
    PCR_Assert((op->operation_type==DOFOPERATIONTYPE_WATCH)||(op->operation_type==DOFOPERATIONTYPE_ACTIVATE)||
        (op->operation_type==DOFOPERATIONTYPE_CONNECT)||(op->operation_type==DOFOPERATIONTYPE_FULLCONNECT));

    return (DOFSystem)op->sources->channel;
}

DOFObjectID DOFOperationInterest_GetObjectID( const DOFOperation op )
{
    PCR_Assert((op->operation_type==DOFOPERATIONTYPE_WATCH)||(op->operation_type==DOFOPERATIONTYPE_ACTIVATE)||
        (op->operation_type==DOFOPERATIONTYPE_CONNECT)||(op->operation_type==DOFOPERATIONTYPE_FULLCONNECT));

    return op->object_id.object_id;
}

DOFInterfaceID DOFOperationInterest_GetInterfaceID( DOFOperation op )
{
    PCR_Assert((op->operation_type==DOFOPERATIONTYPE_WATCH)||(op->operation_type==DOFOPERATIONTYPE_ACTIVATE)||
        (op->operation_type==DOFOPERATIONTYPE_CONNECT)||(op->operation_type==DOFOPERATIONTYPE_FULLCONNECT));

    return op->interface_id;
}

DOFInterestLevel DOFOperationInterest_GetInterestLevel( const DOFOperation op )
{
    switch(op->operation_type) {
        case DOFOPERATIONTYPE_WATCH:
            return DOFINTERESTLEVEL_WATCH;
        case DOFOPERATIONTYPE_ACTIVATE:
            return DOFINTERESTLEVEL_ACTIVATE;
        case DOFOPERATIONTYPE_CONNECT:
        case DOFOPERATIONTYPE_FULLCONNECT:
            return DOFINTERESTLEVEL_CONNECT;
        default:
            PCR_Assert(0); /* type not valid */
    }

    return -1; /* should assert above */
}

DOFOperationControl DOFOperationInterest_GetControl( const DOFOperation op )
{
    PCR_Assert((op->operation_type==DOFOPERATIONTYPE_WATCH)||(op->operation_type==DOFOPERATIONTYPE_ACTIVATE)||
        (op->operation_type==DOFOPERATIONTYPE_CONNECT)||(op->operation_type==DOFOPERATIONTYPE_FULLCONNECT));

    return DOFOperation_GetControl(op);
}

DOFInterface DOFOperationDefine_WaitInterface( const DOFOperation op, uint32 timeout, DOFException * pException )
{
    DOFResult result;
    DOFInterface ret = NULL;

    PCR_Assert( op );
    PCR_Assert( op->operation_type == DOFOPERATIONTYPE_FINDINTERFACEDEFINITION );

    result = _waitresult(op, timeout, pException);

    if(result) {
        DOFValue val = DOFResult_GetValue( result );
        if ( val && DOFType_GetTypeID( DOFValue_GetDOFType( val ) ) == DOFTYPE_BLOB)
        {
            ret = DOFInterface_Create( op->interface_id, DOFValueBLOB_GetSize( val ), DOFValueBLOB_Get( val ) );
        }
        DOFResult_Destroy(result);
    }

    return ret;
}

DOFObject DOFOperationDefine_GetObject( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_FINDINTERFACEDEFINITION);

    return DOFOperation_GetObject(op);
}

DOFInterfaceID DOFOperationDefine_GetInterfaceID( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_FINDINTERFACEDEFINITION);

    return op->interface_id;
}

DOFOperationControl DOFOperationDefine_GetControl( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_FINDINTERFACEDEFINITION);

    return DOFOperation_GetControl(op);
}

DOFQuery DOFOperationQuery_GetQuery( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_QUERY);

    return op->filter;
}

/**
    Return the @ref DOFSystem that is associated with the Query Operation.

    @param  op  The Operation.  This must not be NULL and must represent a valid Query Operation.

    @return The associated @ref DOFSystem.  This will not be NULL.
**/
DOFSystem DOFOperationQuery_GetSystem( const DOFOperation op )
{
    PCR_Assert(op->operation_type==DOFOPERATIONTYPE_QUERY);

    return (DOFSystem)op->sources->channel;
}

OALAVLTree DOFOperation_CreateObjectIDAVLTree( void )
{
    return OALAVLTree_Create(objid_gt, objid_eq, NULL);
}

DOFPermission _createNeededPermission(DOFOperationType opType,
                                      boolean          isResponse,
                                      DOFObjectID      oid,
                                      DOFInterfaceID   iid)
{
    static DOFPermissionBindingAction bindingPerms[DOFOPERATIONTYPE_INVALID][2] =
    {
    /* DOFOPERATIONTYPE_GET */
        { DOFPERMISSIONBINDINGACTION_READ,      DOFPERMISSIONBINDINGACTION_PROVIDE },
    /* DOFOPERATIONTYPE_SET */
        { DOFPERMISSIONBINDINGACTION_WRITE,     DOFPERMISSIONBINDINGACTION_PROVIDE },
    /* DOFOPERATIONTYPE_INVOKE */
        { DOFPERMISSIONBINDINGACTION_EXECUTE,   DOFPERMISSIONBINDINGACTION_PROVIDE },
    /* DOFOPERATIONTYPE_REGISTER */
        { DOFPERMISSIONBINDINGACTION_READ,      DOFPERMISSIONBINDINGACTION_PROVIDE },
    /* DOFOPERATIONTYPE_SUBSCRIBE */
        { DOFPERMISSIONBINDINGACTION_READ,      DOFPERMISSIONBINDINGACTION_PROVIDE },
    /* DOFOPERATIONTYPE_FINDINTERFACEDEFINITION */
        { 0,                                    0 },
    /* DOFOPERATIONTYPE_WATCH */
        { 0,                                    0 },
    /* DOFOPERATIONTYPE_ACTIVATE */
        { 0,                                    0 },
    /* DOFOPERATIONTYPE_EXCEPTION */
        { 0,                                    DOFPERMISSIONBINDINGACTION_PROVIDE },
    /* DOFOPERATIONTYPE_BEGINPROVIDE */
        { DOFPERMISSIONBINDINGACTION_PROVIDE,   0 },
    /* DOFOPERATIONTYPE_PROVIDE */
        { DOFPERMISSIONBINDINGACTION_PROVIDE,   0 },
    /* DOFOPERATIONTYPE_SIGNAL */
        { DOFPERMISSIONBINDINGACTION_PROVIDE,   0 },
    /* DOFOPERATIONTYPE_CHANGED */
        { DOFPERMISSIONBINDINGACTION_PROVIDE,   0 },
    /* DOFOPERATIONTYPE_CANCEL */
        { DOFPERMISSIONBINDINGACTION_PROVIDE,   0 },
    /* DOFOPERATIONTYPE_CONNECT */
        { 0,                                    0 },
    /* DOFOPERATIONTYPE_SESSION */
        { DOFPERMISSIONBINDINGACTION_SESSION,   DOFPERMISSIONBINDINGACTION_PROVIDE },
    /* DOFOPERATIONTYPE_ADVERTISE */
        { 0,                                    0 },
    /* DOFOPERATIONTYPE_QUERY */
        { 0,                                    0 },
    /* DOFOPERATIONTYPE_FULLCONNECT */
        { 0,                                    0 },
    /* DOFOPERATIONTYPE_NOOP */
        { 0,                                    0 }
    };

    PCR_Assert(opType < DOFOPERATIONTYPE_INVALID);
    switch(opType)
    {
    case DOFOPERATIONTYPE_EXCEPTION:        /* Is this ever hit? */
        PCR_Assert(isResponse);
        return DOFPermissionBinding_Create(bindingPerms[opType][isResponse],oid,iid);
        break;

    case DOFOPERATIONTYPE_SIGNAL:
    case DOFOPERATIONTYPE_CHANGED:
    case DOFOPERATIONTYPE_BEGINPROVIDE:
    case DOFOPERATIONTYPE_PROVIDE:
    case DOFOPERATIONTYPE_CANCEL:
        PCR_Assert(!isResponse);
        /* fall through */

    case DOFOPERATIONTYPE_GET:
    case DOFOPERATIONTYPE_SET:
    case DOFOPERATIONTYPE_INVOKE:
    case DOFOPERATIONTYPE_REGISTER:
    case DOFOPERATIONTYPE_SUBSCRIBE:
    case DOFOPERATIONTYPE_SESSION:
        return DOFPermissionBinding_Create(bindingPerms[opType][isResponse],oid,iid);
        break;

    case DOFOPERATIONTYPE_WATCH:
    case DOFOPERATIONTYPE_ACTIVATE:
    case DOFOPERATIONTYPE_CONNECT:
    case DOFOPERATIONTYPE_FULLCONNECT:
        return DOFPermissionRequestor_Create();
        break;

    case DOFOPERATIONTYPE_ADVERTISE:
        return DOFPermissionProvider_Create();
        break;

    case DOFOPERATIONTYPE_FINDINTERFACEDEFINITION:
        return (isResponse) ? DOFPermissionDefine_Create(iid) : DOFPermissionRequestor_Create();
        break;

    case DOFOPERATIONTYPE_NOOP:
        return NULL;

    default:
        PCR_Assert(FALSE);      /* What type got missed above? */
        return NULL;
    }
    return NULL;
}

/**
@}
**/
