/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/interfaceid.h>
#include <dof/pcr.h>
#include <internal/dofinternal.h>
#include <internal/hashtable.h>
#include <internal/slist.h>
#include <internal/channel.h>
#include <internal/epp_v2.h>
#include <internal/operationinternal.h>
#include <internal/operationconfiginternal.h>
#include <internal/enqueue.h>
#include <internal/systeminternal.h>

#define DOFOPERATION_CLEANUP_GRACE_PERIOD 5000

typedef struct cleanupOpNode_t
{
    OALSListNode    next;
    DOFOperation    op;
}cleanupOpNode_t;

static OALEnqueue enqueue;
static OALHashtable operation_table = NULL;
static OALTimer operation_timer = NULL, retransmit_timer = NULL;
static PCRMutex mutex = NULL;
static PCREvent operationsGoneEvent = NULL;

static boolean shutting_down = FALSE;

boolean OALOperationManager_Init(void)
{
    PCR_Assert(!operation_timer&&!retransmit_timer);
    PCR_Assert(!operation_table);
    PCR_Assert(!enqueue);

    operation_timer = OALTimer_Create(DOF_GetDof()->threadPool, TRUE);
    if(operation_timer != NULL)
    {
        retransmit_timer = OALTimer_Create(DOF_GetDof()->threadPool, FALSE);
        if(retransmit_timer)
        {
            /* By setting this to 2 it does not include the domain. */
            operation_table = OALHashtable_Create(3);
            if(operation_table != NULL)
            {
                enqueue = OALEnqueue_Create(DOF_GetDof()->threadPool);
                if(enqueue != NULL)
                {
                    operationsGoneEvent = PCREvent_Create(TRUE);
                    if(operationsGoneEvent != NULL)
                    {
                        PCREvent_Signal( operationsGoneEvent );
                        shutting_down = FALSE;
                        return TRUE;
                    }
                    OALEnqueue_Destroy(enqueue);
                    enqueue = NULL;
                }
                OALHashtable_Destroy(operation_table, NULL, NULL);
                operation_table = NULL;
            }
            OALTimer_Destroy(retransmit_timer);
            retransmit_timer = NULL;
        }
        OALTimer_Destroy(operation_timer);
        operation_timer = NULL;
    }
    return FALSE;

#if 0
    source_init();
#endif
}

boolean OALOperationManager_Init2(void)
{
    PCR_Assert(!mutex);
    mutex = PCRMutex_Create();
    if(mutex != NULL)
        return TRUE;
    return FALSE;
}

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t _cleanup(OALAVLMultiTreeItem_t op, void* listHead, boolean new_key)
{
    /* called with mutex held */
    cleanupOpNode_t *   opNodePtr;

    /*  Add the cancel to a list so the operation can be cancelled after the traverse is done.
        This will prevent the operation manager being locked before an operation is locked. */

    opNodePtr = (cleanupOpNode_t *)DOFMem_Alloc(sizeof(cleanupOpNode_t));
    if (opNodePtr)
    {
        opNodePtr->next = NULL;
        opNodePtr->op = ((DOFOperation)op);
        DOFOperation_Retain(opNodePtr->op);
        OALSList_Insert(((OALSListNode*)listHead), (OALSListNode)opNodePtr);
    }
    else
    {
        /* Todo: what should happen if the node is not allocated. */
    }

    return OALAVLMULTITREE_NORMAL;
}

void OALOperationManager_Shutdown(void)
{
    PCR_Assert(operation_timer);
    PCR_Assert(!operation_table);
    PCR_Assert(!operationsGoneEvent);
    PCR_Assert(mutex);

    OALTimer_Destroy(operation_timer);
    OALTimer_Destroy(retransmit_timer);
    retransmit_timer = operation_timer = NULL;

    OALEnqueue_Destroy(enqueue);
    enqueue = NULL;
}

void OALOperationManager_Shutdown2(void)
{
    PCR_Assert(mutex);
    PCRMutex_Destroy(mutex);
    mutex = NULL;
}

static void internalCancelAllTraverse_fn(OALSListNode item, void * unused)
{
    DOFOperation op = ((cleanupOpNode_t *)item)->op;
    DOFOperation_Cancel(op);
    DOFOperation_Internal_Destroy(op);
    DOFMem_Free(item);
}

void OALOperationManager_CancelAll(void)
{
    uint32       lastCount;
    OALSListNode opDestroyList = NULL;

    PCR_Assert(operation_timer);
    PCR_Assert(operation_table);
    PCR_Assert(mutex);
    PCR_Assert(!shutting_down);

    PCRMutex_Lock(mutex);
    /*  Build a list of all the operations so they can be cancelled.  Since the operation manager is locked
        the individual operations cannot be locked to be cancelled, so this will just build a list. */
    OALHashtable_Traverse_All(operation_table, _cleanup, &opDestroyList);
    shutting_down = TRUE;
    lastCount = OALHashtable_GetNumItems( operation_table );
    PCRMutex_Unlock(mutex);
    /* shutting_down TRUE: So even though mutex now unlocked, operation_table cannot increase for any reason. */

    /* Now that the operation manager is unlocked it is ok to cancel the operations. */
    OALSList_RemoveAll(&opDestroyList, internalCancelAllTraverse_fn, NULL);
    while ( lastCount )
    {
        uint32 operationsLeft;

        PCREvent_Wait( operationsGoneEvent, DOFOPERATION_CLEANUP_GRACE_PERIOD );

        PCRMutex_Lock( mutex );
        operationsLeft = OALHashtable_GetNumItems( operation_table );
        PCRMutex_Unlock( mutex );

        /* Shutting_down prevents new entries, so this should *always* be less, except in the case of an operation leak (runtime error). */
        PCR_Validate( operationsLeft < lastCount );
        lastCount = operationsLeft;            /* lastCount must be 0, because we can assert operationsLeft is 0. */
    }
    PCREvent_Destroy( operationsGoneEvent );
    operationsGoneEvent = NULL;

    OALHashtable_Destroy(operation_table, 0, 0);
    operation_table = NULL;

    /* Empty the queue of operations in process of being cancelled. */
    OALEnqueue_Empty( enqueue );
}

void OALOperationManager_Traverse_All(OALAVLMultiTree_Traverse_All_fn_t traverseFn, void * context)
{
    PCR_Assert(operation_table);
    PCR_Assert(mutex);

    PCRMutex_Lock(mutex);
    OALHashtable_Traverse_All(operation_table, traverseFn, context);
    PCRMutex_Unlock(mutex);
}

static int _timer_observer(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown)
{
    DOFOperation op = (DOFOperation)argv[0];
    boolean for_system;

    DOFOperation_Lock(op);
    if(op->flags.bits.timeout_hit) {
        DOFOperation_Unlock(op);
        return 0;
    }

    op->flags.bits.timeout_hit = 1;

    if(op->flags.bits.lost)
    {
        DOFOperation_Cancel(op);
        DOFOperation_Unlock(op);
        return 0;
    }

    PCRMutex_Lock(mutex);
    for_system = (!shutting_down&&!op->flags.bits.cancelled&&!op->flags.bits.fake_timeout&&!op->flags.bits.internally_created);
    PCRMutex_Unlock(mutex);

    if (for_system)
        DOFSystem_OpTimedOutCallback(op);
    else
        DOFOperation_Cancel(op);
    DOFOperation_Unlock(op);
    return 0;
}

static int _retry_observer(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown)
{
    PCR_Assert( argc == 2 );

    if ( !shutdown )
    {
        DOFOperation op      = (DOFOperation) argv[0];
        OALChannel   channel = (OALChannel) argv[1];

        DOFOperation_Lock( op );
        DOFOperation_IncrementRetry( op );
        if ( channel )
        {
            /* Directly route retry to channel.  This works for all operation types. */
            DOFOperationManager_SetRetry( op, channel );    /* Schedule the next retry. */
            channel->routeToProvider( channel, op, NULL, NULL );
        }
        else
        {
            DOFSystem_CheckPermissionsRetryAndRoute(op);
        }
        DOFOperation_Unlock( op );
    }

    return 0;
}

DOFError DOFOperationManager_SetRetry(DOFOperation op, OALChannel channel)
{
    DOFError result = DOFERROR_OK;
    DOFOperation_Lock( op );
    if ( !op->flags.bits.internally_created )
    {
        uint32 timeout = DOF_TIMEOUT_NEVER;

        /* While requestor side OAP requests will always have an opConfig, it is not required that every other op will*/
        if ( op->operationConfig )
        {
            DOFOperationRetryCallback cb = DOFOperationControl_GetRetryCallback( op->operationConfig );
            if (cb)
            {
                DOFOperation_Unlock( op );
                PCR_Assert( cb->fns );
                PCR_Assert( cb->fns->Retry );
                timeout = cb->fns->Retry ( cb, op );
                DOFOperation_Lock( op );
            }
            else
                timeout = DOFOperationControl_GetRetryPeriod( op->operationConfig );

            /* Force a retry at least as often as needed to keep the operation alive at the minimal EPP layer. */
            if (timeout > EPPV2_DELAY_MAXMILLISECONDS)
                timeout = EPPV2_DELAY_MAXMILLISECONDS;               /* We don't leave with DOF_TIMEOUT_NEVER. */
        }
        else if ( op->operation_type == DOFOPERATIONTYPE_BEGINPROVIDE )
        {
            /* Force a retry at least as often as needed to keep the operation alive at the minimal EPP layer. */
            timeout = EPPV2_DELAY_MAXMILLISECONDS;
        }

        if ( timeout != DOF_TIMEOUT_NEVER )
        {
            void * argv[2];

            argv[0] = op;
            argv[1] = channel;
            if ( !op->retry_timer_item )
                op->retry_timer_item = OALTimer_CreateItem(retransmit_timer, _retry_observer, lengthof(argv), argv);
            if (!op->retry_timer_item ||
                OALTimerItem_Insert( op->retry_timer_item, timeout, FALSE, FALSE ) != DOFERROR_OK)
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "OperationManager: failed to schedule next operation retry.");
                result = DOFERROR_INSUFFICIENT_RESOURCES;
            }
        }
        else
        { /* No opConfig and op not beginProvide- no retry. Never here for an OAP request on the requestor side. */
        }
    }
    else
    { /* No retry of operations that came in to us across the wire. */
    }
    DOFOperation_Unlock( op );
    return result;
}


/** @internal
 Adds an operation to the operation manager.

 @param op        The operation to add.
 @param opid      The operation ID which will be used for lookup in the operation table. MUST not be NULL when @a op is
                  configured for operation timeout.
 @param need_lock Ignored.

 @return DOFERROR_FAILED for NULL @a opID or if called during operation manager shutdown, OR
         DOFERROR_INSUFFICIENT_RESOURCES for insufficient memory or timer failure or op manager tables found full, OR
         DOFERROR_OK.
**/
DOFError OALOperationManager_AddOperation(DOFOperation op, const DOFOperationID opid, boolean need_lock)
{
    /* ok to do in any thread */
    if (!opid)
    {
        if (op->sources->oal_timeout != DOF_TIMEOUT_NEVER)
        {
            /* Timeout wanted, but it won't happen when opid is NULL. */
            PCR_Assert( FALSE );
        }
        return DOFERROR_FAILED;
    }

    need_lock = TRUE;
    if (need_lock)
    {
        DOFOperation_Lock(op);
    }
    PCRMutex_Lock(mutex);

    if (shutting_down)
    {
        DOFOperation_Cancel( op );
        PCRMutex_Unlock(mutex);
        if (need_lock)
        {
            DOFOperation_Unlock(op);
        }
        return DOFERROR_FAILED;
    }

    if (!op->flags.bits.added_to_operation_table && !op->flags.bits.dont_add_to_operation_table)
    {
        const void *keys[3];
        uint16 key_sizes[3];
        void * sidPtr = DOFOperationID_GetSID(opid);
        uint32 count = DOFOperationID_GetCount(opid);

        PCR_Assert(op->securityScope); /* Security Scope is now required because we key on Domain and OperationID*/
        keys[0] = &count;
        key_sizes[0] = sizeof(count);
        keys[1] = (void *)&sidPtr;
        key_sizes[1] = sizeof(sidPtr);
        keys[2] = DOFObjectID_GetBytes( DOFObjectIDDomain_GetObjectID( DOFSecurityScope_GetDomainID(op->securityScope) ) );
        key_sizes[2] = DOFObjectID_GetBytesSize( DOFObjectIDDomain_GetObjectID( DOFSecurityScope_GetDomainID(op->securityScope) ) );

        PCR_Assert(keys[0] && keys[1] && keys[2]); /* don't allow a null pointer */

        if(OALHashtable_Insert(operation_table, op, keys, key_sizes)!=OALHashtable_InsertSuccess) {
            op->flags.bits.cancelled = 1; /* don't let it cancel */
            PCRMutex_Unlock(mutex);
            if (need_lock)
            {
                DOFOperation_Unlock(op);
            }
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OperationManager: failed to insert a new entry.");
            return DOFERROR_INSUFFICIENT_RESOURCES;
        }

        /* One strange way this next call will fail is if op->timer_item is firing at this moment, but when we add an op
           here, it is typically a new operation and would typically not even have a timer_item yet. */
        if (DOFERROR_OK != _operation_extend( op, op->sources->oal_timeout, TRUE ))
        {
            OALHashtable_Remove(operation_table, op, (const void**)keys, key_sizes);
            op->flags.bits.cancelled = 1; /* don't let it cancel */
            PCRMutex_Unlock(mutex);
            if (need_lock)
            {
                DOFOperation_Unlock(op);
            }
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OperationManager: failed because _operation_extend() failed.");
            return DOFERROR_INSUFFICIENT_RESOURCES;
        }

        op->flags.bits.added_to_operation_table = 1;
        PCREvent_Reset( operationsGoneEvent );
    }

    PCRMutex_Unlock(mutex);
    if (need_lock)
    {
        DOFOperation_Unlock(op);
    }

    return DOFERROR_OK;
}

boolean OALOperationManager_SetupForwardingOperation(DOFOperation op)
{
    PCR_Assert(op != NULL);

    if(op->forwardingId != NULL)
    {
        return TRUE;
    }
    else
    {
        PCRMutex_Lock(mutex);
        op->forwardingId = DOFOperationID_CreateLocal();
        if(op->forwardingId != NULL)
        {
            void const * keys[3];
            uint16 key_sizes[3];
            void * sidPtr = DOFOperationID_GetSID(op->forwardingId);
            uint32 count = DOFOperationID_GetCount(op->forwardingId);

            keys[0] = &count;
            key_sizes[0] = sizeof(count);
            keys[1] = (void *)&sidPtr;
            key_sizes[1] = sizeof(sidPtr);
            keys[2] = DOFObjectID_GetBytes( DOFObjectIDDomain_GetObjectID( DOFSecurityScope_GetDomainID(op->securityScope) ) );
            key_sizes[2] = DOFObjectID_GetBytesSize( DOFObjectIDDomain_GetObjectID( DOFSecurityScope_GetDomainID(op->securityScope) ) );
            if(OALHashtable_Insert(operation_table, op, keys, key_sizes)==OALHashtable_InsertSuccess)
            {
                PCRMutex_Unlock(mutex);
                return TRUE;
            }
            DOFOperationID_Destroy(op->forwardingId);
            op->forwardingId = NULL;
        }
        PCRMutex_Unlock(mutex);
        return FALSE;
    }
}

void OALOperationManager_RemoveOperationAndUnlock( DOFOperation op )
{
    /* Called with operation mutex locked. */
    /* WARNING: This call UNLOCKS the operation mutex! */
    OALTimerItem timer_item = op->timer_item;
    OALTimerItem retry_timer_item = op->retry_timer_item;

    op->timer_item = NULL;
    op->retry_timer_item = NULL;
    if ( op->flags.bits.added_to_operation_table )
    {
        const void *keys[3];
        uint16 key_sizes[3];
        uint32 count = DOFOperationID_GetCount(op->opID);
        void * sidPtr = DOFOperationID_GetSID(op->opID);

        keys[0] = &count;
        key_sizes[0] = sizeof(count);
        keys[1] = (void *)&sidPtr;
        key_sizes[1] = sizeof(sidPtr);
        keys[2] = DOFObjectID_GetBytes( DOFObjectIDDomain_GetObjectID( DOFSecurityScope_GetDomainID(op->securityScope) ) );
        key_sizes[2] = DOFObjectID_GetBytesSize( DOFObjectIDDomain_GetObjectID( DOFSecurityScope_GetDomainID(op->securityScope) ) );

        PCRMutex_Lock( mutex );
        if (OALHashtable_Remove( operation_table, op, keys, key_sizes ))
        {
            op->flags.bits.added_to_operation_table = 0;
        }
        if(op->forwardingId != NULL)
        {
            sidPtr = DOFOperationID_GetSID(op->forwardingId);
            count = DOFOperationID_GetCount(op->forwardingId);
            OALHashtable_Remove(operation_table, op, keys, key_sizes);
        }
        if ( OALHashtable_GetNumItems( operation_table ) == 0 )
        {
            PCREvent_Signal( operationsGoneEvent );
        }
        PCRMutex_Unlock( mutex );
    }
    DOFOperation_Unlock( op );
    OALTimerItem_Destroy( timer_item, TRUE );
    OALTimerItem_Destroy( retry_timer_item, TRUE );
}

/**
 Check if an operation already exists.

 @param opid The operation ID to check for.
 @param opid_size The size of the operation ID.

 @return A DOFOperation with this ID; otherwise NULL if nothing is found.
 **/
DOFOperation OALOperationManager_CheckExistingOperation(DOFObjectIDDomain domain, DOFOperationID opid)
{
    DOFOperation op = NULL;
    if (opid && domain)
    {
        const void *keys[3];
        uint16 key_sizes[3];
        uint32 count = DOFOperationID_GetCount(opid);
        void * sidPtr = DOFOperationID_GetSID(opid);
        keys[0] = &count;
        key_sizes[0] = sizeof(count);
        keys[1] = (void *)&sidPtr;
        key_sizes[1] = sizeof(sidPtr);
        keys[2] = DOFObjectID_GetBytes( DOFObjectIDDomain_GetObjectID( domain ) );
        key_sizes[2] = DOFObjectID_GetBytesSize( DOFObjectIDDomain_GetObjectID( domain ) );

        PCRMutex_Lock(mutex);
        if (!shutting_down)
            op = (DOFOperation)OALHashtable_Retrieve(operation_table, keys, key_sizes);
        if (op)
        {
            /*
             * WARNING: If this locks the operation, this will cause a deadlock.
             * It turns out this is ok for now since op uses the refcount mutex to bump ref counts, and does not lock the operation.
             */
            if( !DOFOperation_Retain(op) )
                op = NULL;
        }
        PCRMutex_Unlock(mutex);
    }
    return op;
}


DOFError _operation_extend( DOFOperation op, uint32 newTimeout, boolean notOnTimerItemFiring )
{
    DOFError retError = DOFERROR_OK;

    /* Called with op mutex locked. */
    op->sources->oal_timeout = newTimeout;
    PCRMutex_Lock(mutex);
    if(shutting_down&&(op->sources->oal_timeout)) /* 0 timeout is ok during shutdown */
    {
        PCRMutex_Unlock(mutex);
        return DOFERROR_FAILED;
    }
    PCRMutex_Unlock(mutex);
    if(op->sources->oal_timeout==DOF_TIMEOUT_NEVER)
    {
        /* Cancel existing timeout, if any. */
        if ( op->timer_item )
        {
            OALTimerItem_Remove( op->timer_item );
        }

        return DOFERROR_OK;
    }
    if (!op->timer_item)
        op->timer_item = OALTimer_CreateItem(operation_timer, _timer_observer, 1, (void**)&op);
    if (op->timer_item)
        retError = OALTimerItem_Insert( op->timer_item, op->sources->oal_timeout, FALSE, notOnTimerItemFiring );
    else
        return DOFERROR_INSUFFICIENT_RESOURCES;
    if (retError == DOFERROR_OK)
    {
        op->flags.bits.fake_timeout = 0;
        op->flags.bits.timeout_hit = 0;
    }
    else
    {
        PCR_Assert( retError == DOFERROR_REJECTED || retError == DOFERROR_FAILED );
    }

    return retError;
}


/**
 Replace an operation's opid with a new one.

 @param op The operation to rename.  This should not be NULL.
 @param domain the domain to use when adding the operation to operation manager.  This should not be NULL.
 @param new_id The new oid.  This should not be NULL.

 @return DOFERROR_OK if all goes well; otherwise DOFERROR_INSUFFICIENT_RESOURCES if no resources available. DOFERROR_FAILED is returned if the old opid isn't found.
 **/
DOFError OALOperationManager_RenameOperation(DOFOperation op, DOFObjectIDDomain domain, DOFOperationID new_id)
{
    /* ok to do in any thread */
    DOFError status = DOFERROR_FAILED;

    if ( op && new_id && domain )
    {
        DOFOperation_Lock(op);
        PCRMutex_Lock(mutex);
        if ( !shutting_down )
        {
            DOFOperationID old_store, new_store;
            const void *keys[3];
            uint16 key_sizes[3];
            uint32 count = DOFOperationID_GetCount(new_id);
            void * sidPtr = DOFOperationID_GetSID(new_id);

            status = DOFERROR_INSUFFICIENT_RESOURCES;

            old_store = op->opID;
            new_store = DOFOperationID_Copy(new_id);


            /* If copy failed, then leave the operation unmodified. */
            if ( new_store )
            {
                /* Store the new opid in the operation table.  Do this before removing based on old key in case out of memory. */
                keys[0] = &count;
                key_sizes[0] = sizeof(count);
                keys[1] = (void *)&sidPtr;
                key_sizes[1] = sizeof(sidPtr);
                keys[2] = DOFObjectID_GetBytes( DOFObjectIDDomain_GetObjectID( domain ) );
                key_sizes[2] = DOFObjectID_GetBytesSize( DOFObjectIDDomain_GetObjectID( domain ) );

                if ( OALHashtable_Insert(operation_table, op, keys, key_sizes) == OALHashtable_InsertSuccess )
                {
                    /* remove the operation from the tree with old key */
                    count = DOFOperationID_GetCount(old_store);
                    sidPtr = DOFOperationID_GetSID(old_store);
                    /* domain set above */

                    OALHashtable_Remove(operation_table, op, keys, key_sizes);

                    op->opID = new_store;
                    DOFOperationID_Destroy(old_store);
                    status = DOFERROR_OK;
                }
                else
                {
                    DOFOperationID_Destroy(new_store);
                }

            }
        }
        PCRMutex_Unlock(mutex);
        DOFOperation_Unlock(op);
    }
    return status;
}

static void __DOFOperation_Cancel(OALThreadPoolItem dnode, boolean shutdown)
{
    CancelOperationContext_t * node = (CancelOperationContext_t *) dnode;

    PCR_Assert(dnode != NULL);
    DOFOperation_Lock(node->operation);
    PCR_Assert(node->operation->flags.bits.cancelled);
    /* At this moment we guarantee op (argv[0]) has our ref count, so next call cannot deallocate op and its members. */
    _DOFOperation_Cancel(node->operation);
    OALOperationManager_RemoveOperationAndUnlock(node->operation); /* call this directly now */  /* Unlocks operation mutex. */
    DOFOperation_Internal_Destroy(node->operation); /* release other side of queue */
}

CancelOperationContext_t * DOFOperationManager_CreateCancelContext(DOFOperation op)
{
    CancelOperationContext_t * node;

    node = (CancelOperationContext_t *) DOFMem_Alloc(sizeof(CancelOperationContext_t));
    if(node != NULL)
    {
        OALEnqueueItem_Init((OALEnqueueItem) node, __DOFOperation_Cancel);
        node->operation = op;
    }
    return node;
}

/** @internal
 Queues an operation for cancel. MUST BE CALLED WITH OPERATION LOCK HELD.

 @param op The operation to cancel.
 @return DOFERROR_OK if successful, DOFERROR_INSUFFICIENT_RESOURCES if no resources, or DOFERROR_FAILED if operation was already enqueued.
 **/
DOFError OALOperationManager_EnqueueOperationCancel(DOFOperation op)
{
    /* Called with operation mutex locked. */
    PCR_Assert( op->flags.bits.cancelled );
    PCR_Assert(op->cancelContext != NULL);
    if(op->flags.bits.cancel_enqueued)
    {
        return DOFERROR_FAILED; /* don't do it again please */
    }

    PCR_Assert( enqueue );
    PCR_Assert(op->cancelContext->operation == op);
    DOFOperation_Retain(op);
    if (OALEnqueue_Enqueue(enqueue, (OALEnqueueItem) op->cancelContext)) {
        op->flags.bits.cancel_enqueued = 1;
        return DOFERROR_OK;
    }
    return DOFERROR_INSUFFICIENT_RESOURCES;
}

DOFError OALOperationManager_EnqueueOperation(OALEnqueueItem node)
{
    PCR_Assert( enqueue );

    return OALEnqueue_Enqueue(enqueue, node) ? DOFERROR_OK : DOFERROR_INSUFFICIENT_RESOURCES;
}
