/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/error.h>
#include <dof/permission.h>
#include <internal/avl_tree.h>
#include <internal/timer.h>
#include <internal/dofinternal.h>
#include <internal/source.h>
#include <dof/address.h>
#include <dof/objectid.h>
#include <internal/epp_negative_one.h>
#include <internal/operationinternal.h>
#include <internal/slist.h>

#ifdef OALSOURCE_REF_COUNT_DEBUG
#include <stdio.h>
#endif

#define SOURCE_FLAGS_IMPLICIT           0x1
#define SOURCE_FLAGS_QUERIED            0x2

#define QUERY_TIMEOUT 10000
#define STATE_TIMEOUT 60000

static PCRMutex source_mutex;
static OALAVLTree explicit_source_tree;
static OALAVLTree source_tree;
static OALTimer source_timer;

static void actually_destroy(sid_t *sid, boolean remove_from_source_tree);

static boolean source_gt(sid_t *s1, sid_t *s2, void *context)
{
    int rval = OALChannel_Compare( s1->channel, s2->channel );
    if ( rval == 0 )
    {
        return ( DOFAddress_Compare(s1->address, s2->address) > 0 );
    }
    return ( rval > 0 );
}

static boolean source_eq(sid_t *s1, sid_t *s2, void *context)
{
    if (OALChannel_IsEquivalent(s1->channel, s2->channel))
    {
        return (DOFAddress_IsEqual(s1->address, s2->address));
    }

    return FALSE;
}

static boolean explicit_source_gt(sid_t *s1, sid_t *s2, void *context)
{
    int32 cmpVar;

    PCR_Assert(s1->oid&&s2->oid);

    cmpVar = DOFObjectID_Compare(DOFObjectIDDomain_GetObjectID(s1->domain), DOFObjectIDDomain_GetObjectID(s2->domain));
    if (cmpVar == 0)
    {
        cmpVar =  DOFObjectID_Compare(DOFObjectIDSource_GetObjectID( s1->oid ), DOFObjectIDSource_GetObjectID( s2->oid ));
    }

    return (cmpVar > 0);
}

static boolean explicit_source_eq(sid_t *s1, sid_t *s2, void *context)
{
    int32 cmpVar;

    PCR_Assert(s1->oid&&s2->oid);

    cmpVar = DOFObjectID_Compare(DOFObjectIDDomain_GetObjectID(s1->domain), DOFObjectIDDomain_GetObjectID(s2->domain));
    if (cmpVar == 0)
    {
        cmpVar =  DOFObjectID_Compare(DOFObjectIDSource_GetObjectID( s1->oid ), DOFObjectIDSource_GetObjectID( s2->oid ));
    }

    return (cmpVar == 0);
}

boolean source_mutex_init(void)
{
    source_mutex = PCRMutex_Create();
    if(source_mutex != NULL)
    {
        return TRUE;
    }

    return FALSE;
}

boolean source_init(void)
{
    PCR_Assert(!source_tree&&!source_timer);

    source_timer = OALTimer_Create(DOF_GetDof()->threadPool, FALSE);
    if(source_timer != NULL)
    {
        source_tree = OALAVLTree_Create((OALAVLTree_CompareGreaterThan_fn_t)source_gt, (OALAVLTree_CompareEqual_fn_t)source_eq, NULL);
        if(source_tree != NULL)
        {
            explicit_source_tree = OALAVLTree_Create((OALAVLTree_CompareGreaterThan_fn_t)explicit_source_gt, (OALAVLTree_CompareEqual_fn_t)explicit_source_eq, NULL);
            if(explicit_source_tree != NULL)
            {
                return TRUE;
            }
            OALAVLTree_Destroy(source_tree, NULL, NULL);
        }
        OALTimer_Destroy(source_timer);
    }
    return FALSE;
}

void source_mutex_shutdown(void)
{
    PCRMutex_Destroy(source_mutex);
    source_mutex = NULL;
}

void source_shutdown(void)
{
    OALAVLTree tempSourceTree;

    PCRMutex_Lock( source_mutex );
    PCR_Assert(source_tree&&source_timer);
    OALTimer_Destroy(source_timer);
    source_timer = NULL;
    tempSourceTree = source_tree;
    source_tree = NULL;  /* _source_remove() finds source_tree NULL so it doesn't access source_tree while we destroy. Isn't used anymore.*/
    OALAVLTree_Destroy(tempSourceTree, /*_source_remove*/ NULL, NULL);
    OALAVLTree_Destroy(explicit_source_tree, NULL, NULL);

    source_tree = NULL;
    source_timer = NULL;
    explicit_source_tree = NULL;
    PCRMutex_Unlock( source_mutex );
}

#ifdef OALSOURCE_REF_COUNT_DEBUG
static void source_inc_ref_count(sid_t *sid, const char * file, const char * func, uint32 line)
#else
static void source_inc_ref_count(sid_t *sid)
#endif
{
    /* should be called with lock held */
    if(!sid->state_timeout)
    {
        ++sid->ref_count;
#ifdef OALSOURCE_REF_COUNT_DEBUG
        fprintf(stderr,"DEBUG inc_ref_count sid %p, refCount %d, file %s, func %s, line %d\n", (void *)sid, sid->ref_count, file, func, line);
#endif
    }
    else
    {
#ifdef OALSOURCE_REF_COUNT_DEBUG
        fprintf(stderr,"DEBUG in inc_ref_count sid %p, refCount %d, file %s, func %s, line %d\n", (void *)sid, sid->ref_count, file, func, line);
#endif
        OALTimerItem_Destroy(sid->state_timeout, FALSE);
        sid->state_timeout = NULL;
    }
}

boolean source_isOIDSet(sid_t *sid)
{
    if (sid)
    {
        boolean set = FALSE;
        PCRMutex_Lock(source_mutex);
        if (sid->oid)
            set = TRUE;
        PCRMutex_Unlock(source_mutex);
        if (set)
            return TRUE;
    }
    return FALSE;
}

typedef struct updateOpIDSidNode_t
{
    OALSListNode    next;
    DOFOperation    op;
}updateOpIDSidNode_t;

typedef struct updateOpIDContext_t
{
    OALSListNode *  listHead;
    sid_t *         oldSid;
}updateOpIDContext_t;

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t _updateSidTraverseFn(OALAVLMultiTreeItem_t op, void* context, boolean new_key)
{
    updateOpIDSidNode_t * opNodePtr;

    /*  Add the operation to a list so it  can be updated with the correct opID after the traverse is done.  */

    if (DOFOperationID_IsSidEqual(((DOFOperation)op)->opID, ((updateOpIDContext_t*)context)->oldSid))
    {
        opNodePtr = (updateOpIDSidNode_t *)DOFMem_Alloc(sizeof(updateOpIDSidNode_t));
        if (opNodePtr)
        {
            opNodePtr->next = NULL;
            opNodePtr->op = ((DOFOperation)op);
            DOFOperation_Retain(opNodePtr->op);
            OALSList_Insert(((updateOpIDContext_t*)context)->listHead, (OALSListNode)opNodePtr);
        }
    }

    return OALAVLMULTITREE_NORMAL;
}

static void internalRemoveThenAddWithNewSid(OALSListNode item, void* arg)
{
    sid_t* newSid = (sid_t*)arg;
    updateOpIDSidNode_t *   opNodePtr = (updateOpIDSidNode_t *)item;

    /* Remove the operation from op manager. */
    DOFOperation_Lock(opNodePtr->op);

    OALOperationManager_RemoveOperationAndUnlock(opNodePtr->op);

    /* Now add the operation back with the new opID. */
    DOFOperation_Lock(opNodePtr->op);
    DOFOperationID_UpdateSid(opNodePtr->op->opID, newSid);
    EPP_NegOne_UpdateOperationSid(opNodePtr->op, newSid);

    if (OALOperationManager_AddOperation(opNodePtr->op, opNodePtr->op->opID, TRUE) != DOFERROR_OK)
    {
    }
    DOFOperation_Unlock(opNodePtr->op);

    DOFOperation_Internal_Destroy(opNodePtr->op);
    DOFMem_Free(opNodePtr);
}

static void source_merge(sid_t * oldSid, sid_t * newSid)
{
    OALSListNode        opUpdateList = NULL;
    updateOpIDContext_t context;

    context.listHead = &opUpdateList;
    context.oldSid = oldSid;

    OALOperationManager_Traverse_All(_updateSidTraverseFn, &context);

    OALSList_RemoveAll(&opUpdateList, internalRemoveThenAddWithNewSid, newSid);
}

int source_set_oid(sid_t *sid, DOFObjectIDSource oid)
{
    sid_t * explicitSid = NULL;
    sid_t * implicitSid = NULL;

    PCR_Assert(oid);

    PCRMutex_Lock( source_mutex );
    if ( sid->oid )
    {
        /* The SID already has an OID?  Then this is an explicit SID, Check if it is the same... */
        /* If the source changed its SID, it likely rebooted, so report it. */
        int rval = ( DOFObjectID_Compare( DOFObjectIDSource_GetObjectID(sid->oid), DOFObjectIDSource_GetObjectID(oid) ) != 0 );
        PCRMutex_Unlock( source_mutex );
        return rval;
    }

    /* Try to lookup the explicit SID, it may already exist but not associated with this channel and address yet. */
    explicitSid = lookup_explicit_sid(oid, sid->domain);
    if (explicitSid)
    {
        if (!sid->explicitSid)
        {
            sid->explicitSid = explicitSid;
            /*  The explicitSid is not retained.  When it is destroyed any implicit sids that
                reference it will be destroyed. */

            implicitSid = sid;
        }
    }
    else
    {
        /* An explicit sid was not found, make this one explicit. */
        sid->actAsPermission = DOFPermissionActAs_Create(oid);
        if(sid->actAsPermission == NULL)
        {
            PCRMutex_Unlock(source_mutex);
            return -1;
        }
        sid->iamPermission = DOFPermissionIAm_Create(oid);
        if(sid->iamPermission == NULL)
        {
            DOFPermission_Destroy(sid->actAsPermission);
            sid->actAsPermission = NULL;
            PCRMutex_Unlock(source_mutex);
            return -1;
        }
        sid->oid = DOFObjectIDSource_Copy(oid);
        if ( !sid->oid )
        {
            DOFPermission_Destroy(sid->iamPermission);
            DOFPermission_Destroy(sid->actAsPermission);
            sid->iamPermission = NULL;
            sid->actAsPermission = NULL;
            PCRMutex_Unlock( source_mutex );
            return -1;
        }
        if (explicit_source_tree &&
            (OALAVLTree_Insert( explicit_source_tree, sid, sid, NULL ) != OALAVLTree_InsertSuccess))
        {

#if 0
            DOFPermission_Destroy(sid->iamPermission);
            DOFPermission_Destroy(sid->actAsPermission);
            sid->iamPermission = NULL;
            sid->actAsPermission = NULL;
            DOFObjectIDSource_Destroy(sid->oid);
            sid->oid = NULL;
            PCRMutex_Unlock(source_mutex);
            return -1;
#endif
        }
    }

    PCRMutex_Unlock( source_mutex );

    if (explicitSid)
    {
        if (implicitSid)
        {
            source_merge(implicitSid, explicitSid);
        }
        source_release(explicitSid);
    }


    return 0;
}

#ifdef OALSOURCE_REF_COUNT_DEBUG
int source_set_channel_address_Debug(sid_t *sid, OALChannel channel, DOFAddress address, const char * file, const char * func, uint32 line)
#else
int source_set_channel_address(sid_t *sid, OALChannel channel, DOFAddress address)
#endif
{
    PCRMutex_Lock(source_mutex);
    if ( sid->channel )
    {
        /* The SID already has a channel / address?  Check if it is the same... */
        if ( !OALChannel_IsEquivalent( sid->channel, channel ) || DOFAddress_Compare( sid->address, address ) != 0 )
        {
            sid_t * implicitSid;

            if (channel && address)
            {
                /* There can be multiple channels/addresses associated with a single SID! */
                implicitSid = create_sid_from_channel_address_explicit(channel, address, sid->domain, sid->explicitSid);
                if (implicitSid)
                {
                    /* This release will leave the sid in the timeout state, so if something comes in before it
                       times out it will still be used. */
                    source_release(implicitSid);
                }
            }
            else
            {
                if (source_tree)
                    OALAVLTree_Delete(source_tree, sid);
                if (sid->oid)
                {
                    /* If this is an explicit sid, just clear the channel and address. */
                    OALChannel_Release(sid->channel);
                    sid->channel = NULL;
                    DOFAddress_Destroy(sid->address);
                    sid->address = NULL;
                }
                else
                {
                    /* If this is only an implicit sid go ahead and release the sid.  Note that sid->explicitSid is not retaind, so don't release it here. */
                    sid->explicitSid = NULL;
                    source_release(sid);
                }
            }
        }
        else
        {
            /* No change. */
            PCRMutex_Unlock(source_mutex);
            return 0;
        }
    }
    else
    {
        /* If a channel was provided insert the sid into source_tree. */
        if ( channel )
        {
            if (channel->remoteChannelData->secureSid == sid)
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "Source: received command with my own source");
                PCRMutex_Unlock(source_mutex);
                return -1;
            }
            sid->address = DOFAddress_Copy(address);
            if ( !sid->address )
            {
                PCRMutex_Unlock(source_mutex);
                return -1;
            }
            sid->channel = OALChannel_Retain(OALChannel_GetPrimarySessionChannel( channel ));
            if( source_tree )
            {
                sid_t* stored = NULL;
                OALAVLTree_InsertionStatus insertRet = OALAVLTree_Insert(source_tree, sid, sid, (OALAVLTreeItem_t*)&stored);
                if( insertRet != OALAVLTree_InsertSuccess )
                {
                    /* handle stored sid's explicit SourceID not matching new sid's explicit SourceID */
                    if( insertRet == OALAVLTree_AlreadyInserted )
                    {
                        sid_t* esid1 = (sid->explicitSid) ? sid->explicitSid : sid;
                        sid_t* esid2 = (stored->explicitSid)?stored->explicitSid:stored;

                        if( esid1->oid && esid2->oid && !explicit_source_eq(esid1, esid2, NULL))
                        {
                            EPP_NegOne_ProcessNodeDown( DOFObjectIDSource_GetObjectID( esid2->oid ), stored->channel, stored->address, TRUE );
                            source_set_channel_address( stored, NULL, NULL );

                            if( OALAVLTree_Insert(source_tree, sid, sid, NULL) == OALAVLTree_InsertSuccess )
                            {
                                PCRMutex_Unlock(source_mutex);
                                return 0;
                            }
                        }

                        /* Fall through is desired if insertion failed to clean up */
                    }

                    DOFAddress_Destroy(sid->address);
                    sid->address = NULL;
                    OALChannel_Release(sid->channel);
                    sid->channel = NULL;

                    PCRMutex_Unlock(source_mutex);
                    return -1;
                }
            }
        }
    }
    PCRMutex_Unlock(source_mutex);

    return 0;
}

static sid_t *OALSource_create_sid(void)
{
    sid_t *sid = DOFMem_Alloc(sizeof(sid_t));

    if(sid != NULL)
    {
        PCRMem_Set(sid, 0, sizeof(sid_t));

        sid->ref_count = 1;
    }
    return sid;
}

sid_t *create_sid_from_oid(DOFObjectIDSource oid, DOFObjectIDDomain domain)
{
    sid_t *sid = OALSource_create_sid();

    if(!sid) return NULL;

    sid->domain = DOFObjectIDDomain_Copy(domain);
    if (domain && !sid->domain)
    {
        DOFMem_Free(sid);
        return NULL;
    }

    if(source_set_oid(sid, oid))
    {
        /* error occurred */
        DOFObjectIDDomain_Destroy(sid->domain);
        DOFMem_Free(sid);
        return NULL;
    }

    return sid;
}

sid_t* create_sid_from_channel_address_explicit(OALChannel channel, DOFAddress address, DOFObjectIDDomain domain, sid_t* sid_explicit)
{
    sid_t *sid = OALSource_create_sid();

    if(!sid)
        return NULL;

    sid->domain = DOFObjectIDDomain_Copy(domain);
    if (domain && !sid->domain)
    {
        DOFMem_Free(sid);
        return NULL;
    }

    if(sid_explicit)
    {
        sid->explicitSid = sid_explicit;
    }

    if(source_set_channel_address(sid, channel, address))
    {
        /* error occurred */
        DOFObjectIDDomain_Destroy(sid->domain);
        DOFMem_Free(sid);
        return NULL;
    }

    if (channel->remoteChannelData && channel->remoteChannelData->remoteSid)
    {
        /* If this fail just return, the channel and address were set and the oid can get set later. */
        source_set_oid(sid, channel->remoteChannelData->remoteSid);
    }

    return sid;                                        /* If channel, sid is also placed into source.c's source_tree. */
}

sid_t *create_sid_from_channel_address(OALChannel channel, DOFAddress address, DOFObjectIDDomain domain)
{
    sid_t *sid = OALSource_create_sid();

    if(!sid)
        return NULL;

    sid->domain = DOFObjectIDDomain_Copy(domain);
    if (domain && !sid->domain)
    {
        DOFMem_Free(sid);
        return NULL;
    }

    if(source_set_channel_address(sid, channel, address))
    {
        /* error occurred */
        DOFObjectIDDomain_Destroy(sid->domain);
        DOFMem_Free(sid);
        return NULL;
    }

    if (channel->remoteChannelData && channel->remoteChannelData->remoteSid)
    {
        /* If this fail just return, the channel and address were set and the oid can get set later. */
        source_set_oid(sid, channel->remoteChannelData->remoteSid);
    }

    return sid;                                        /* If channel, sid is also placed into source.c's source_tree. */
}

#ifdef OALSOURCE_REF_COUNT_DEBUG
sid_t *lookup_sid_Debug(OALChannel channel, DOFAddress address, const char * file, const char * func, uint32 line)
#else
sid_t *lookup_sid(OALChannel channel, DOFAddress address)
#endif
{
    sid_t *sid = NULL;
    sid_t compare_sid;

    compare_sid.ref_count = 0;
    compare_sid.oid = NULL;
    compare_sid.channel = OALChannel_GetPrimarySessionChannel( channel );
    compare_sid.address = address;

    PCRMutex_Lock(source_mutex);
    if (source_tree)
        sid = (sid_t*)OALAVLTree_Retrieve(source_tree, &compare_sid);
    if(sid)
    {
        if (sid->explicitSid)
        {
            sid = sid->explicitSid;
        }

#ifdef OALSOURCE_REF_COUNT_DEBUG
        source_inc_ref_count(sid, file, func, line); /* while we have the lock */
#else
        source_inc_ref_count(sid); /* while we have the lock */
#endif
    }
    PCRMutex_Unlock(source_mutex);

    return sid;
}

#ifdef OALSOURCE_REF_COUNT_DEBUG
sid_t *lookup_explicit_sid_Debug(DOFObjectIDSource oid, DOFObjectIDDomain domain, const char * file, const char * func, uint32 line)
#else
sid_t *lookup_explicit_sid(DOFObjectIDSource oid, DOFObjectIDDomain domain)
#endif
{
    sid_t *sid = NULL;
    sid_t compare_sid;

    compare_sid.ref_count = 0;
    compare_sid.oid = oid;
    compare_sid.domain = domain;
    compare_sid.channel = NULL;
    compare_sid.address = NULL;

    PCRMutex_Lock(source_mutex);
    if (explicit_source_tree)
        sid = (sid_t*)OALAVLTree_Retrieve(explicit_source_tree, &compare_sid);
    if(sid)
    {
#ifdef OALSOURCE_REF_COUNT_DEBUG
        source_inc_ref_count(sid, file, func, line); /* while we have the lock */
#else
        source_inc_ref_count(sid); /* while we have the lock */
#endif
    }
    PCRMutex_Unlock(source_mutex);

    return sid;
}

#ifdef OALSOURCE_REF_COUNT_DEBUG
void source_retain_Debug(sid_t *sid, const char * file, const char * func, uint32 line)
#else
void source_retain(sid_t *sid)
#endif
{
    if (sid)
    {
        PCRMutex_Lock(source_mutex);
#ifdef OALSOURCE_REF_COUNT_DEBUG
        source_inc_ref_count(sid, file, func, line); /* bump sid ref count */
#else
        source_inc_ref_count(sid);
#endif
        PCRMutex_Unlock(source_mutex);
    }
}

typedef struct sidCleanupNode_t
{
    OALSListNode    next;
    sid_t *         sid;
}sidCleanupNode_t;

typedef struct sidCleanupCtx_t
{
    OALSListNode *  listHead;
    sid_t *         sid;

}sidCleanupCtx_t;

static void cleanupRemoveSidFn(OALSListNode sid, void * unused)
{
    source_release(((sid_t*)sid));
}

static boolean cleanupImplicitFn(OALAVLTreeItem_t item, void *arg)
{
    sid_t* implicitSid = (sid_t*)item;
    sidCleanupCtx_t* ctx = (sidCleanupCtx_t*)arg;
    if (implicitSid->explicitSid == ctx->sid)
    {
        sidCleanupNode_t * node = (sidCleanupNode_t *)DOFMem_Alloc(sizeof(sidCleanupNode_t));
        if (node)
            OALSList_Insert(ctx->listHead, (OALSListNode)node);
    }
    return FALSE;
}

static void actually_destroy(sid_t *sid, boolean actually_destroy)
{

    /* Traverse the source_tree and release any sid with a reference to sid. */
    if(sid->channel&&actually_destroy)
    {
        if (sid->oid)
        {
            OALSListNode    sidList = NULL;
            sidCleanupCtx_t ctx;

            ctx.listHead = &sidList;
            ctx.sid = sid;
            if (source_tree)
            {
                /* Since this is an explicit sid, traverse the source_tree and release any implicit sids that reference it. */
                OALAVLTree_Traverse(source_tree, cleanupImplicitFn, sid);
            }

            /* Since the sids cannot be released while traversing the tree, release all of the sids in the list. */
            OALSList_RemoveAll(ctx.listHead, cleanupRemoveSidFn, NULL);
        }
        if (source_tree)
            OALAVLTree_Delete(source_tree, sid);
    }
    if(sid->oid && explicit_source_tree)
        OALAVLTree_Delete(explicit_source_tree, sid);
    DOFAddress_Destroy(sid->address);
    OALChannel_Release(sid->channel);
    DOFObjectIDSource_Destroy(sid->oid);
    DOFObjectIDDomain_Destroy(sid->domain);
    DOFPermission_Destroy(sid->iamPermission);
    DOFPermission_Destroy(sid->actAsPermission);
    OALTimerItem_Destroy(sid->state_timeout, FALSE);
    DOFMem_Free(sid);
}

static int timeout_fn(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown)
{
    /*if ref count 1 then only reference is the one in the timer. destroy it. if shutting down, also destroy reference*/
    PCRMutex_Lock(source_mutex);
    if( ((sid_t*)argv[0])->state_timeout == item )
    {
        if((((sid_t*)argv[0])->ref_count==1)||shutdown)
        {
            actually_destroy((sid_t*)argv[0], TRUE);
        }
        else
        {
            PCR_Assert( ((sid_t*)argv[0])->ref_count != 0 ); /* If ref count 0, shouldn't have this sid and cannot use it.*/
            OALTimerItem_Destroy(((sid_t*)argv[0])->state_timeout, FALSE);
            ((sid_t*)argv[0])->state_timeout = NULL;
            /* ok to explicitly dec ref count here */
            --(((sid_t*)argv[0])->ref_count);
#ifdef OALSOURCE_REF_COUNT_DEBUG
            fprintf(stderr,"DEBUG in timeout_fn sid %p, refCount %d\n", argv[0], ((sid_t*)argv[0])->ref_count);
#endif
        }
    }
    PCRMutex_Unlock(source_mutex);
    return 0;
}

#ifdef OALSOURCE_REF_COUNT_DEBUG
void source_release_Debug(sid_t *sid, const char * file, const char * func, uint32 line)
#else
void source_release(sid_t *sid)
#endif
{
    if (sid)
    {
        PCRMutex_Lock(source_mutex);
#ifdef OALSOURCE_REF_COUNT_DEBUG
        fprintf(stderr,"DEBUG Release sid %p, refCount %d, file %s, func %s, line %d\n", (void *)sid, sid->ref_count - 1, file, func, line);
#endif
        if(!--sid->ref_count)
        {
            PCR_Assert(!sid->state_timeout); /* since the timer calls actually_destroy, this better not be set when ref is 0 */
            sid->ref_count = 1; /* keep reference for timer */
            if(source_timer == NULL)        /* Shutting down */
            {
                actually_destroy(sid, TRUE);
            }
            else
            {
                if (source_timer)
                {
                    sid->state_timeout = OALTimer_CreateItem(source_timer, timeout_fn, 1, (void**)&sid);
                    if(!sid->state_timeout)
                    {
                        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "Source: Failed to create state timeout item." );
                        actually_destroy(sid, TRUE);
                    }
                    else if (OALTimerItem_Insert( sid->state_timeout, STATE_TIMEOUT, FALSE, FALSE ) != DOFERROR_OK)
                    {
                        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "Source: Failed to insert query timer item." );
                        /* just free it */
                        actually_destroy(sid, TRUE);
                    }
                }
            }
        }
        PCRMutex_Unlock(source_mutex);
    }
}

#ifdef OALSOURCE_REF_COUNT_DEBUG
void OALSource_Clear_Debug(OALChannel channel, DOFAddress address, const char * file, const char * func, uint32 line)
#else
void OALSource_Clear(OALChannel channel, DOFAddress address)
#endif
{
    sid_t *sid;

    /* This is called at the point a session becomes secure so that something that comes in on the channel and address does not find
       the sid that is now in the unsecure domain. */

    PCRMutex_Lock(source_mutex);

    sid = lookup_sid(channel, address);

    if (sid)
    {
#ifdef OALSOURCE_REF_COUNT_DEBUG
        printf("DEGBUG in OALSource_Clear sid %p, file %s, func %s, line %d\n", (void *)sid, file, func, line);
#endif
        if (source_tree)
            OALAVLTree_Delete(source_tree, sid);
        if (sid->oid)
        {
            /* If this is an explicit sid, clear the channel and address. */
            OALChannel_Release(sid->channel);
            sid->channel = NULL;
            DOFAddress_Destroy(sid->address);
            sid->address = NULL;
        }
        else
        {
            /* Note that sid->explicitSid is not retaind, so don't release it here. */
            sid->explicitSid = NULL;
        }
        /* The sid was retained in lookup_sid so release it now. */
        source_release(sid);
    }

    PCRMutex_Unlock(source_mutex);
}
