/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/address.h>
#include <dof/connection.h>
#include <dof/control.h>
#include <dof/objectid.h>

#include <internal/connection_internal.h>
#include <internal/objectinternal.h>
#include <internal/operationconfiginternal.h>

#include "operationconfig_struct.h"

/** @addtogroup DOFOperationControl
@{
**/

/*
**  "Private" Type and Method Declarations
*/


/*
**  "Public" Type and Method Declarations
*/

static PCRMutex operationConfigMutex;

/**
 * @internal
 * sets up DOFOperationControl's Class Mutex
 */
void DOFOperationControl_Init(void)
{
    operationConfigMutex = PCRMutex_Create();
}

/**
 * @internal
 * Destroys DOFOperationControl's Class Mutex
 */
void DOFOperationControl_Shutdown(void)
{
    PCRMutex_Destroy(operationConfigMutex);
}

static void exclusionQueueEmpty( DOFOperationControl self )
{
    PCR_Assert( self );
    while ( self->exclusionQueueHead )
    {
        struct ProviderExclusionNode_t * temp = self->exclusionQueueHead;
        DOFObjectID_Destroy( self->exclusionQueueHead->oid );
        self->exclusionQueueHead = self->exclusionQueueHead->next;
        DOFMem_Free(temp);
    }
    self->exclusionQueueTail = (struct ProviderExclusionNode_t *) &self->exclusionQueueHead;
}

static boolean exclusionQueueCopyTo( DOFOperationControl self, DOFOperationControl dest )
{
    /* Called with operationConfigMutex held. */
    struct ProviderExclusionNode_t * d = (struct ProviderExclusionNode_t *) &dest->exclusionQueueHead;
    struct ProviderExclusionNode_t * s;
    PCR_Assert( self && dest && d );
    s = self->exclusionQueueHead;
    while ( s )
    {
        d->next = DOFMem_Alloc( sizeof( struct ProviderExclusionNode_t ) );
        if ( d->next )
        {
            d = d->next;
            d->next = NULL;
            d->oid = DOFObjectID_Copy( s->oid );
            if ( d->oid )
            {
                s = s->next;
                continue;
            }
        }
        exclusionQueueEmpty( dest );
        return FALSE;
    }
    dest->exclusionQueueTail = d;
    return TRUE;
}

DOFOperationControl DOFOperationControl_Create(void)
{
    DOFOperationControl ret = (DOFOperationControl) DOFMem_Alloc(sizeof(struct DOFOperationControl_t));

    if(ret != NULL)
    {
        PCRMem_Set(ret, 0, sizeof(struct DOFOperationControl_t));
        ret->refCount = 1;
        ret->respCount = DOFOPERATIONCONTROL_RESPONDERS_ALL;
        ret->percentage = DOFOPERATIONCONTROL_HEURISTIC_ALL;
        ret->timeout = DOFOPERATIONCONTROL_DEFAULT_RETRY_PERIOD;
        ret->verbosity = DEFAULT_VERBOSITY;
        ret->isAutoExclusion = TRUE;
        ret->isAllowBroadcast = TRUE;
        ret->exclusionQueueHead = NULL;
        ret->exclusionQueueTail = (struct ProviderExclusionNode_t *) &ret->exclusionQueueHead;
        ret->ackTimeout = DOF_TIMEOUT_NEVER;
        return ret;
    }

    return NULL;
}

DOFOperationControl DOFOperationControl_Copy(DOFOperationControl self)
{
    if(self) {
        DOFOperationControl ret = (DOFOperationControl) DOFMem_Alloc(sizeof(struct DOFOperationControl_t));

        if(ret)
        {
            PCRMem_Set(ret, 0, sizeof(struct DOFOperationControl_t));
            ret->refCount = 1;
            PCRMutex_Lock(operationConfigMutex);
            ret->percentage = self->percentage;
            ret->verbosity = self->verbosity;
            ret->security = self->security;
            ret->isAutoExclusion = self->isAutoExclusion;
            ret->isAllowBroadcast = self->isAllowBroadcast;
            ret->isDelay = self->isDelay;
            ret->retryCallback = self->retryCallback;
            ret->timeout = self->timeout;
            ret->respCount = self->respCount;
            ret->cacheDelay = self->cacheDelay;
            ret->ackTimeout = self->ackTimeout;
            ret->connection = DOFConnection_Copy(self->connection);

            if(self->channel)
            {
                OALChannel_Retain(self->channel);
                ret->channel = self->channel;
            }

            if( !exclusionQueueCopyTo( self, ret ) )
            {
                DOFMem_Free(ret);
                ret = NULL;
            }
            PCRMutex_Unlock(operationConfigMutex);

            return ret;
        }
    }
    return NULL;
}

void DOFOperationControl_Destroy(DOFOperationControl self)
{
    boolean gone;
    if ( self )
    {
        PCRMutex_Lock( operationConfigMutex );
        PCR_Assert( self->refCount );
        gone = --(self->refCount) == 0;
        PCRMutex_Unlock( operationConfigMutex );
        if ( gone )
        {
            DOFConnection_Destroy(self->connection);
            exclusionQueueEmpty(self);
            OALChannel_Release(self->channel);
            DOFMem_Free(self);
        }
    }
}

boolean DOFOperationControl_AddProviderExclusion(DOFOperationControl self, DOFObjectID provider)
{
    boolean ret = FALSE;

    if ( self && !DOFObjectID_IsWildcard( provider ) )
    {
        /* Queue gets a copy. */
        provider = DOFObjectID_Copy( provider );
        if ( provider )
        {
            PCRMutex_Lock(operationConfigMutex);
            self->exclusionQueueTail->next = DOFMem_Alloc( sizeof( struct ProviderExclusionNode_t ) );
            if ( self->exclusionQueueTail->next )
            {
                self->exclusionQueueTail = self->exclusionQueueTail->next;
                self->exclusionQueueTail->next = NULL;
                self->exclusionQueueTail->oid = provider;
                ret = TRUE;
            }
            else
                DOFObjectID_Destroy( provider );
            PCRMutex_Unlock(operationConfigMutex);
        }
    }
    return ret;
}

void DOFOperationControl_ClearProviderExclusions(DOFOperationControl self)
{
    if(self != NULL)
    {
        PCRMutex_Lock(operationConfigMutex);
        exclusionQueueEmpty( self );
        PCRMutex_Unlock(operationConfigMutex);
    }
}

boolean DOFOperationControl_SetRetryPeriod(DOFOperationControl self,
                                     uint32       timeout)
{
    if(self != NULL)
    {
        PCRMutex_Lock(operationConfigMutex);
        self->timeout = timeout;
        self->retryCallback = NULL;
        PCRMutex_Unlock(operationConfigMutex);
        return TRUE;
    }
    return FALSE;
}

/**
 * @internal
 * Gets the Retry Period Set buy DOFOperationControl_SetRetryPeriod()
 *
 * @param self the Request Config to retrieve from
 *
 * @return uint32 the length of time before a retry
 */
uint32 DOFOperationControl_GetRetryPeriod(DOFOperationControl self)
{
    uint32 ret;

    if(self)
    {
        PCRMutex_Lock(operationConfigMutex);
        ret = self->timeout;
        PCRMutex_Unlock(operationConfigMutex);
    }
    else
    {
        ret = DOFOPERATIONCONTROL_DEFAULT_RETRY_PERIOD;
    }
    return ret;
}

boolean DOFOperationControl_SetRetryCallback(DOFOperationControl self,
                                         DOFOperationRetryCallback cb)
{
    if(self != NULL)
    {
        PCRMutex_Lock(operationConfigMutex);
        self->timeout = DOFOPERATIONCONTROL_DEFAULT_RETRY_PERIOD;
        self->retryCallback = cb;
        PCRMutex_Unlock(operationConfigMutex);
        return TRUE;
    }
    return FALSE;
}

/**
 * @internal
 * Gets the users callback
 *
 * @param self the operations request config
 *
 * @return DOFOperationRetryCallback pointer to the callback function
 */
DOFOperationRetryCallback DOFOperationControl_GetRetryCallback(DOFOperationControl self)
{
    DOFOperationRetryCallback ret;
    PCR_Assert(self);
    PCRMutex_Lock(operationConfigMutex);
    ret = self->retryCallback;
    PCRMutex_Unlock(operationConfigMutex);
    return ret;
}

DOFOperationResponseLevel DOFOperationControl_GetResponseLevel(DOFOperationControl self)
{
    DOFOperationResponseLevel ret = DOFOPERATIONRESPONSELEVEL_NONE;
    if ( self )
    {
        PCRMutex_Lock(operationConfigMutex);
        if ( self->verbosity == DEFAULT_VERBOSITY )
            ret = DOFOPERATIONRESPONSELEVEL_EXCEPTION;
        else
            ret = self->verbosity;
        PCRMutex_Unlock(operationConfigMutex);
    }
    return ret;
}

boolean DOFOperationControl_SetResponseLevel(DOFOperationControl self,
                                       DOFOperationResponseLevel level)
{
    if(self != NULL)
    {
        PCRMutex_Lock(operationConfigMutex);
        if(level > DOFOPERATIONRESPONSELEVEL_ALL)
        {
            self->verbosity = DOFOPERATIONRESPONSELEVEL_ALL;
        }
        else
        {
            self->verbosity = level;
        }
        PCRMutex_Unlock(operationConfigMutex);
        return TRUE;
    }
    return FALSE;
}

void DOFOperationControl_SetDefaultResponseLevel(DOFOperationControl self,
                                                DOFOperationResponseLevel level)
{
    PCR_Assert(level <= DOFOPERATIONRESPONSELEVEL_ALL);
    if(self != NULL)
    {
        PCRMutex_Lock(operationConfigMutex);
        if ( self->verbosity == DEFAULT_VERBOSITY )
        {
            self->verbosity = level;
        }
        PCRMutex_Unlock(operationConfigMutex);
    }
}

DOFConnection DOFOperationControl_GetTargetConnection(DOFOperationControl self)
{
    DOFConnection ret = NULL;
    if ( self )
    {
        PCRMutex_Lock(operationConfigMutex);
        ret = self->connection;
        PCRMutex_Unlock(operationConfigMutex);
    }
    return ret;
}

boolean DOFOperationControl_SetTargetConnection(DOFOperationControl self,
                                            DOFConnection    connection)
{
    if ( self )
    {
        PCRMutex_Lock(operationConfigMutex);
        self->connection = DOFConnection_Copy(connection);
        PCRMutex_Unlock(operationConfigMutex);
        return TRUE;
    }
    return FALSE;
}

boolean DOFOperationControl_SetDelay(DOFOperationControl self,
                                 boolean            delay)
{
    if(self !=  NULL)
    {
        PCRMutex_Lock(operationConfigMutex);
        self->isDelay = !!delay;
        PCRMutex_Unlock(operationConfigMutex);
        return TRUE;
    }
    return FALSE;
}

boolean DOFOperationControl_SetMaxCacheAge(DOFOperationControl self,
                                       uint32             cacheDelay)
{
    if(self != NULL)
    {
        PCRMutex_Lock(operationConfigMutex);
        if ( cacheDelay > 32767000U )
            self->cacheDelay = 32767;
        else
            self->cacheDelay = ((cacheDelay + 999UL) / 1000UL);
        PCRMutex_Unlock(operationConfigMutex);
        return TRUE;
    }
    return FALSE;
}

uint16 DOFOperationControl_GetMaxCacheAge_Sec(DOFOperationControl self)
{
    uint16 ret;
    PCR_Assert(self != NULL);
    PCRMutex_Lock(operationConfigMutex);
    ret = self->cacheDelay;
    PCRMutex_Unlock(operationConfigMutex);
    return ret;
}

uint32 DOFOperationControl_GetMaxCacheAge(DOFOperationControl self)
{
    uint32 ret;
    PCR_Assert(self != NULL);
    PCRMutex_Lock(operationConfigMutex);
    ret = 1000UL * self->cacheDelay;
    PCRMutex_Unlock(operationConfigMutex);
    return ret;
}

boolean DOFOperationControl_IsDelay(DOFOperationControl self)
{
    boolean ret = FALSE;
    if(self != NULL)
    {
        PCRMutex_Lock(operationConfigMutex);
        ret = self->isDelay;
        PCRMutex_Unlock(operationConfigMutex);
    }
    return ret;
}

boolean DOFOperationControl_SetMaxResponderCount(DOFOperationControl self,
                                             uint16             count)
{
    if(self !=  NULL)
    {
        PCRMutex_Lock(operationConfigMutex);
        self->respCount = count;
        PCRMutex_Unlock(operationConfigMutex);
        return TRUE;
    }
    return FALSE;
}

boolean DOFOperationControl_SetHeuristic(DOFOperationControl self,
                                     uint16             percentage)
{
    if(self !=  NULL)
    {
        PCRMutex_Lock(operationConfigMutex);
        self->percentage = percentage;
        PCRMutex_Unlock(operationConfigMutex);
        return TRUE;
    }
    return FALSE;
}

/** @internal
 * Retains the Request Config
 *
 * @note the Request Config must be Destroyed by calling DOFOperationControl_Destroy;
 *
 * @param self the operation's request config
 */
void DOFOperationControl_Retain(DOFOperationControl self)
{
    if(self != NULL)
    {
        PCRMutex_Lock(operationConfigMutex);
        ++(self->refCount);
        PCRMutex_Unlock(operationConfigMutex);
    }
}

/**
 * @internal
 * Gets the Verbosity Level Requested
 *
 * @param self the operation's request config
 *
 * @return uint8 the level
 */
uint8 DOFOperationControl_GetVerbosity(DOFOperationControl self)
{
    uint8 ret;
    PCR_Assert(self);
    PCRMutex_Lock(operationConfigMutex);
    ret = self->verbosity;
    PCRMutex_Unlock(operationConfigMutex);
    return ret;
}

uint16 DOFOperationControl_GetMaxResponderCount(DOFOperationControl self)
{
    uint16 ret;
    PCR_Assert(self);
    PCRMutex_Lock(operationConfigMutex);
    ret = self->respCount;
    PCRMutex_Unlock(operationConfigMutex);
    return ret;
}

uint16 DOFOperationControl_GetHeuristic(DOFOperationControl self)
{
    uint16 ret;
    PCR_Assert(self);
    PCRMutex_Lock(operationConfigMutex);
    ret = self->percentage;
    PCRMutex_Unlock(operationConfigMutex);
    return ret;
}

boolean DOFOperationControl_SetAckTimeout(DOFOperationControl self, uint32 timeout)
{
    if(self !=  NULL)
    {
        PCRMutex_Lock(operationConfigMutex);
        self->ackTimeout = timeout;
        PCRMutex_Unlock(operationConfigMutex);
        return TRUE;
    }
    return FALSE;
}

uint32 DOFOperationControl_GetAckTimeout(DOFOperationControl self)
{
    uint32 ret;
    PCR_Assert(self);
    PCRMutex_Lock(operationConfigMutex);
    ret = self->ackTimeout;
    PCRMutex_Unlock(operationConfigMutex);
    return ret;
}

/**
 * @internal
 * Locks the Request Config
 *
 * @param self the operation's request config
 */
void DOFOperationControl_Lock  (DOFOperationControl self)
{
    PCR_Assert(self);

    PCRMutex_Lock(operationConfigMutex);
}

/**
 * @internal
 * Unlocks the Request Config
 *
 * @param self the operation's request config
 */
void DOFOperationControl_Unlock(DOFOperationControl self)
{
    PCR_Assert(self);

    PCRMutex_Unlock(operationConfigMutex);
}

boolean DOFOperationControl_SetAutoExclude( DOFOperationControl self, boolean isAutomatic )
{
    if(self != NULL)
    {
        PCRMutex_Lock(operationConfigMutex);
        self->isAutoExclusion = !!isAutomatic;
        PCRMutex_Unlock(operationConfigMutex);
        return TRUE;
    }
    return FALSE;
}

boolean DOFOperationControl_IsAutoExclude( DOFOperationControl self )
{
    boolean ret = FALSE;
    if(self != NULL)
    {
        PCRMutex_Lock(operationConfigMutex);
        ret = self->isAutoExclusion;
        PCRMutex_Unlock(operationConfigMutex);
    }
    return ret;
}

boolean DOFOperationControl_SetFloodAllowed( DOFOperationControl self, boolean isAllowed )
{
    if(self)
    {
        PCRMutex_Lock(operationConfigMutex);
        self->isAllowBroadcast = !!isAllowed;
        PCRMutex_Unlock(operationConfigMutex);
        return TRUE;
    }
    return FALSE;
}

boolean DOFOperationControl_IsFloodAllowed( DOFOperationControl self )
{
    boolean ret = FALSE;
    if(self != NULL)
    {
        PCRMutex_Lock(operationConfigMutex);
        ret = self->isAllowBroadcast;
        PCRMutex_Unlock(operationConfigMutex);
    }
    return ret;
}


DOFObjectID DOFOperationControl_PopExclusionQueueHead( DOFOperationControl self )
{
    DOFObjectID ret = NULL;
    if ( self )
    {
        struct ProviderExclusionNode_t * tmp;
        PCRMutex_Lock(operationConfigMutex);
        tmp = self->exclusionQueueHead;
        if ( tmp )
        {
            ret = tmp->oid;
            self->exclusionQueueHead = tmp->next;
            if ( !self->exclusionQueueHead )
                self->exclusionQueueTail = (struct ProviderExclusionNode_t *) &self->exclusionQueueHead;
        }
        PCRMutex_Unlock(operationConfigMutex);
        DOFMem_Free( tmp );
    }
    /* Queue's copy is returned and must be freed by caller. */
    return ret;
}


boolean DOFOperationControl_IsProviderExcluded( DOFOperationControl self, DOFObjectID provID )
{
    boolean ret = FALSE;
    if ( self )
    {
        struct ProviderExclusionNode_t * tmp;
        PCRMutex_Lock(operationConfigMutex);
        tmp = self->exclusionQueueHead;
        while ( tmp )
        {
            if ( DOFObjectID_Compare( tmp->oid, provID ) == 0 )
            {
                ret = TRUE;
                break;
            }
            tmp = tmp->next;
        }
        PCRMutex_Unlock(operationConfigMutex);
    }
    return ret;
}


static boolean addToTree( OALSListNode node_, void* arg )
{
    struct ProviderExclusionNode_t* node = (struct ProviderExclusionNode_t*)node_;
    OALAVLTree tree = (OALAVLTree)arg;
    DOFObjectID oid = DOFObjectID_Copy( node->oid );
    if((oid != NULL) && (OALAVLTree_Insert( tree, oid, oid, NULL ) != OALAVLTree_InsertSuccess))
    {
        DOFObjectID_Destroy(oid);
    }
    return FALSE;
}

void DOFOperationControl_ExclusionQueueAddToTree( DOFOperationControl self, OALAVLTree tree )
{
    if ( self )
    {
        PCRMutex_Lock( operationConfigMutex );
        OALSList_Traverse( (OALSListNode) self->exclusionQueueHead, addToTree, tree );
        PCRMutex_Unlock( operationConfigMutex );
    }
}


/**
@}
**/
