/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/dof.h>
#include <internal/dofinternal.h>
#include <internal/packet.h>
#include <internal/pointerlist.h>

extern PCRMutex ref_count_mutex;

/*
**  "Private" Type and Method Declarations
*/
typedef struct OALPacket_t
{
    uint32      refCount;   /* reference count */
    uint32      streamSize;
    OALPtrList  streamList;
}OALPacket_t;

extern PCRMutex ref_count_mutex;

/*
**  "Public" Methods
*/

/* "Static" methods */

OALPacket OALPacket_Create(void)
{
    /* allocate memory for the packet struct */
    OALPacket newPacket = (OALPacket)(DOFMem_Alloc( sizeof(OALPacket_t) ));
    if (!newPacket)
        return NULL;
    newPacket->streamList = OALPtrList_Create(0x08);
    if(newPacket->streamList != NULL)
    {
        newPacket->refCount = 1;
        newPacket->streamSize = 0;
        return newPacket;
    }
    DOFMem_Free(newPacket);
    return NULL;
}

void OALPacket_Retain(OALPacket self)
{
    PCR_Assert(self);

    PCRMutex_Lock(ref_count_mutex);
    ++(self->refCount);
    PCR_Assert( self->refCount );   /* Rollover to 0 is disastrous and is caused by coding error of some kind. */
    PCRMutex_Unlock(ref_count_mutex);
}

void OALPacket_Release(OALPacket self)
{
    uint32 listCount;
    uint32 i, j;
    uint32 refCount_temp;

    if(!self)
        return;

    PCRMutex_Lock(ref_count_mutex);
    PCR_Assert( self->refCount );
    refCount_temp = --(self->refCount);
    PCRMutex_Unlock(ref_count_mutex);

    if(refCount_temp)
        return;

    listCount = OALPtrList_GetCount(self->streamList);

    /* remove and destroy each of the streams in the list */
    PCRMutex_Lock(ref_count_mutex);     /* This lock is an optimization to avoid multiple locks and unlocks
                                           while releasing OALStreams.  See OALStream_Release for more information. */
    for (i = 0, j = listCount-1; i < listCount; ++i, --j)
    {
        /* remove from the back towards the front so that memory moving does not take place */
        OALStream stream = (OALStream)OALPtrList_RemoveAt( self->streamList, j );
        OALStream_Release(stream, FALSE);
    }
    PCRMutex_Unlock(ref_count_mutex);
    /* destroy the list of streams itself */
    OALPtrList_Destroy(self->streamList);
    /* Finally destroy "this" */
    DOFMem_Free(self);
}

boolean OALPacket_PrependStream(OALPacket self, OALStream stream)
{
    PCR_Assert(self);
    PCR_Assert(stream != NULL);

    if(OALPtrList_InsertAt(self->streamList, stream, 0))
    {
        self->streamSize += OALStream_GetDataSize(stream);
        return TRUE;
    }
    return FALSE;
}

boolean OALPacket_AppendStream(OALPacket self, OALStream stream)
{
    uint32 streamCount;

    PCR_Assert(self);
    PCR_Assert(stream != NULL);

    streamCount = OALPtrList_GetCount(self->streamList);
    if(OALPtrList_InsertAt(self->streamList, stream, streamCount))
    {
        self->streamSize += OALStream_GetDataSize(stream);
        return TRUE;
    }
    return FALSE;
}

uint32 OALPacket_GetTotalStreamSize(OALPacket self)
{
    PCR_Assert(self != NULL);

    return self->streamSize;
}

uint32 OALPacket_CopyTo(OALPacket self, uint8* destBuf, uint32 pktDataOffset, uint32 length)
{
    uint32 streamCount;
    uint32 bytesCopied = 0;
    uint32 streamIndex; /* This is the index to a stream inside this packet. */
    uint32 absoluteIndex = 0; /* This is the byte index into the entire packet. */
    boolean locatingFirstStream = TRUE;

    PCR_Assert(self);
    PCR_Assert((pktDataOffset + length) <= OALPacket_GetTotalStreamSize(self));
    PCR_Assert(destBuf != NULL);

    streamCount = OALPtrList_GetCount(self->streamList);

    for (streamIndex = 0; streamIndex < streamCount; streamIndex++)
    {
        OALStream stream = (OALStream)(OALPtrList_PeekAt(self->streamList, streamIndex));
        const uint32 streamSize = OALStream_GetDataSize(stream);

        if (locatingFirstStream)/* first locate the stream which contains the start of the data to copy */
        {
            uint32 nextStreamIndex = absoluteIndex + streamSize;
            if (nextStreamIndex > pktDataOffset)
            {
                /* The start of the data to copy is in this stream, start copying from this stream. */
                const uint32 streamOffset = pktDataOffset - absoluteIndex;
                const uint8* pData = (const uint8*)(OALStream_GetReadPtr(stream) + streamOffset);
                uint32 lengthToCopy = streamSize - streamOffset;
                if (lengthToCopy > length)
                    lengthToCopy = length;
                PCRMem_Copy(destBuf, pData, lengthToCopy);
                bytesCopied = lengthToCopy;
                destBuf += lengthToCopy;
                locatingFirstStream = FALSE;
            }
            else
            {
                /* Look for start of data in next stream */
                absoluteIndex = nextStreamIndex;
            }
        }
        else /* Copy data from remaining streams, if any */
        {
            uint32 lengthToCopy = length - bytesCopied;
            if (lengthToCopy)
            {
                const uint8* pData = OALStream_GetReadPtr(stream);
                if (lengthToCopy > streamSize)
                    lengthToCopy = streamSize;
                PCRMem_Copy(destBuf, pData, lengthToCopy);
                bytesCopied += lengthToCopy;
                destBuf += lengthToCopy;
            }
            else
                break;
        }
    }
    return bytesCopied;
}

