/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/error.h>
#include <internal/dofinternal.h>
#include <internal/securityscope.h>
#include <internal/marshal.h>
#include <internal/permissionscope.h>
#include <internal/error.h>

#define PERMISSIONSCOPEMETHOD_REFERENCE         0
#define PERMISSIONSCOPEMETHOD_RELATIVE          1
#define PERMISSIONSCOPEMETHOD_ABSOLUTE          2
#define PERMISSIONSCOPEMETHOD_CONTINUED         3

#define FULLLY_COMPATIBLE_SECURITYSCOPE DOFSECURITYSCOPEFLAG_ALL_SCOPES
#define NOT_MASKING_ID_SECURITYSCOPE DOFSECURITYSCOPEFLAG_NOT_MASKING
#define LAST_NON_RESERVED_SECURITYSCOPE 0x1FFFFFFF

#define OALPERMISSIONSONSCOPE_ALLOC_STEP 4

typedef struct OALPermissionScope_t
{
    uint32  idCount;
    uint32  marshalSize;
    uint32  ids[1];       /* must be last */
} OALPermissionScope_t;

static PCRMutex permScopeMutex = NULL;

boolean OALPermissionScope_Init(void)
{
    PCR_Assert(permScopeMutex == NULL);
    permScopeMutex = PCRMutex_Create();
    return (permScopeMutex != NULL);
}

void OALPermissionScope_Shutdown(void)
{
    PCRMutex_Destroy(permScopeMutex);
    permScopeMutex = NULL;
}

OALPermissionScope OALPermissionScope_Create(uint32 idCount)
{
    OALPermissionScope ret = DOFMem_Alloc(offsetof(struct OALPermissionScope_t, ids) +
                                          (sizeof(uint32) * idCount));

    if(ret != NULL)
    {
        PCRMem_Set( ret, 0, offsetof(struct OALPermissionScope_t, ids) + (sizeof(uint32) * idCount));
        ret->marshalSize = (OALMarshal_GetCompressedValueSize(4, 0) * idCount);
        ret->idCount = idCount;
    }
    return ret;
}

OALPermissionScope OALPermissionScope_Copy( OALPermissionScope src )
{
    if( src )
    {
        OALPermissionScope ret = DOFMem_Alloc(offsetof(struct OALPermissionScope_t, ids) + (sizeof(uint32) * src->idCount));

        if( ret )
        {
            ret->idCount = src->idCount;
            ret->marshalSize = src->marshalSize;
            PCRMem_Copy( ret->ids, src->ids, (sizeof(uint32) * src->idCount));
            return ret;
        }
    }
    return NULL;
}

void OALPermissionScope_Destroy(OALPermissionScope scope)
{
    DOFMem_Free(scope);
}

void OALPermissionScope_Lock(void)
{
    PCRMutex_Lock(permScopeMutex);
}

void OALPermissionScope_Unlock(void)
{
    PCRMutex_Unlock(permScopeMutex);
}

/* These functions are needed because currently we access the internal pointer but we need these sorted so this replaces PCRMem_set*/
/* sets the complete list of IDs like PCRMem_Set but sorts as they are inserted */
void OALPermissionScope_SetIDs(OALPermissionScope self, uint32 i_cnt, const uint32* ids, uint32 f_cnt, const uint32* flags, uint32 * realFlagsCount)
{
    uint32 i;
    uint32 cnt = 0;

    PCR_Assert(self != NULL);
    PCR_Assert((i_cnt + f_cnt) == self->idCount);

    if(realFlagsCount != NULL)
    {
        *realFlagsCount = 0;
    }
    for( i=0; i< i_cnt; ++i )
    {
        self->ids[cnt] = ids[i];
        self->marshalSize += OALMarshal_GetCompressedValueSize(4, ids[i]);
        ++cnt;
        if((realFlagsCount != NULL) && (ids[i] > LAST_NON_RESERVED_SECURITYSCOPE))
        {
            ++(*realFlagsCount);
        }
    }

    for( i=0; i< f_cnt; ++i )
    {
        self->ids[cnt] = flags[i];
        self->marshalSize += OALMarshal_GetCompressedValueSize(4, flags[i]);
        ++cnt;
        if((realFlagsCount != NULL) && (flags[i] > LAST_NON_RESERVED_SECURITYSCOPE))
        {
            ++(*realFlagsCount);
        }
    }
}

/* These functions are needed because currently we access the internal pointer but we need these sorted so this replaces ptr[offset] = id for use in a for loop */
/* This tries to set an id at specified location in id list but if the id is a flag it is inserted at the end instead
 * offset[in] is the offset of the one to set
 * offset[out] is the next offset to set. it may be the same since flags will be placed at the end.
 */
void OALPermissionScope_SetID(OALPermissionScope self, uint32* offset, uint32 id)
{
     if( self && ((*offset) < self->idCount) )
     {
         OALPermissionScope_Lock();
         self->marshalSize -= OALMarshal_GetCompressedValueSize(4, self->ids[(*offset)]);
         self->marshalSize += OALMarshal_GetCompressedValueSize(4, id);
         self->ids[(*offset)] = id;
         (*offset) += 1;
         OALPermissionScope_Unlock();
     }
}

uint32 OALPermissionScope_CalculateMarshalSize(OALPermissionScope self)
{
    uint32 ret;

    PCR_Assert(self != NULL);

    OALPermissionScope_Lock();
    ret = (self->idCount/63) + 1;
    ret += self->marshalSize;
    OALPermissionScope_Unlock();
    return ret;
}

uint32 OALPermissionScope_GetIdCount(OALPermissionScope scope)
{
    PCR_Assert(scope != NULL);

    return scope->idCount;
}

uint32 const * OALPermissionScope_GetIds(OALPermissionScope scope)
{
    PCR_Assert(scope != NULL);

    return scope->ids;
}

void OALPermissionScope_SetIdCount(OALPermissionScope scope,
                                   uint32             count)
{
    PCR_Assert(scope != NULL);

    scope->idCount = count;
}

boolean OALPermissionScope_Marshal(OALPermissionScope self,
                                   OALStream          stream)
{
    uint32 scopeOffset = 0;
    uint32 toWrite;

    PCR_Assert(self != NULL);
    PCR_Assert(stream != NULL);
    OALPermissionScope_Lock();
    do
    {
        uint8 firstByte;

        toWrite = self->idCount - scopeOffset;
        if(toWrite > 63)
        {
            toWrite = 63;
        }
        if(scopeOffset == 0)
        {
            firstByte = 0x80;
        }
        else
        {
            firstByte = 0xC0;
        }
        firstByte |= toWrite;
        if(OALStream_Write8(stream, firstByte))
        {
            uint32 i;

            for(i = 0; i < toWrite; ++i)
            {
                if(OALMarshal_StreamCompress4(stream, self->ids[scopeOffset + i]) != DOFERROR_OK)
                {
                    OALPermissionScope_Unlock();
                    return FALSE;
                }
            }
            scopeOffset += toWrite;
        }
        else
        {
            OALPermissionScope_Unlock();
            return FALSE;
        }
    } while(toWrite == 63 /* scopeOffset < scope->permissionCount */);
    OALPermissionScope_Unlock();
    return TRUE;
}

#define MAX_RELATIVE_BOUND  0x20000000

DOFError OALPermissionScope_Unmarshal(OALStream            stream,
                                      uint16               count,
                                      OALPermissionScope * permissionScopes,
                                      OALPermissionScope   connectionScope)
{
    uint32 * permBuffer = NULL;

    uint32 readPermissions = 0;

    uint16 readScopes = 0;

    boolean mustContinue = FALSE;
    boolean relativePermissions = FALSE;

    if( permissionScopes && connectionScope )
    {
        while(readScopes < count)
        {
            uint8 firstByte;
            uint8 method;
            uint8 permsToRead;

            if(!OALStream_Read8(stream, &firstByte))
            {
                return DOFERROR_PARSE_ERROR;
            }
            method = (firstByte >> 6);
            permsToRead = (firstByte & 0x3F);
            switch(method)
            {
            case PERMISSIONSCOPEMETHOD_REFERENCE:
                permissionScopes[readScopes] = OALPermissionScope_Copy(permissionScopes[readScopes - (permsToRead + 1)]);
                ++readScopes;
                break;

            case PERMISSIONSCOPEMETHOD_CONTINUED:
                if(!mustContinue)
                    return DOFERROR_PARSE_ERROR;
            case PERMISSIONSCOPEMETHOD_RELATIVE:
                if(method != PERMISSIONSCOPEMETHOD_CONTINUED)
                {
                    relativePermissions = TRUE;
                }
            case PERMISSIONSCOPEMETHOD_ABSOLUTE:
                {
                    uint32 * writeBuffer;

                    uint8 i;

                    if(permsToRead < 63)
                    {
                        permissionScopes[readScopes] = OALPermissionScope_Create(readPermissions + permsToRead);
                        if(permissionScopes[readScopes] == NULL)
                        {
                            DOFMem_Free(permBuffer);
                            return DOFERROR_INSUFFICIENT_RESOURCES;
                        }
                        if(permBuffer != NULL)
                        {
                            PCRMem_Copy(permissionScopes[readScopes]->ids, permBuffer,
                                        sizeof(uint32) * readPermissions);
                            DOFMem_Free(permBuffer);
                            permBuffer = NULL;
                        }
                        writeBuffer = &(permissionScopes[readScopes]->ids[readPermissions]);
                        mustContinue = FALSE;
                        ++readScopes;
                    }
                    else
                    {
                        uint32 * newBuffer = (readPermissions == 0) ?
                                             DOFMem_Alloc(sizeof(uint32) * permsToRead) :
                                             DOFMem_Realloc(permBuffer, sizeof(uint32) * (readPermissions + permsToRead));

                        if(newBuffer == NULL)
                        {
                            DOFMem_Free(permBuffer);
                            return DOFERROR_INSUFFICIENT_RESOURCES;
                        }
                        permBuffer = newBuffer;
                        writeBuffer = &permBuffer[readPermissions];
                        mustContinue = TRUE;
                    }
                    for(i = 0; i < permsToRead; ++i)
                    {
                        if(OALMarshal_StreamUncompress4(stream, &writeBuffer[i]) != DOFERROR_OK)
                        {
                            DOFMem_Free(permBuffer);
                            return DOFERROR_PARSE_ERROR;
                        }
                        if(relativePermissions)
                        {
                            if(writeBuffer[i] < MAX_RELATIVE_BOUND)
                            {
                                if(connectionScope->idCount < writeBuffer[i])
                                {
                                    DOFMem_Free(permBuffer);
                                    return DOFERROR_PARSE_ERROR;
                                }
                                writeBuffer[i] = connectionScope->ids[writeBuffer[i]];
                            }
                        }
                    }
                    if(mustContinue)
                    {
                        readPermissions += permsToRead;
                    }
                    else
                    {
                        readPermissions = 0;
                        relativePermissions = FALSE;
                    }
                }
                break;

            default:        /* should not hit */
                DOF_LogMessage(DOF_GetDof(), DOFLOGLEVEL_WARNING, "OALPermissionScope: OALPermissionScope_Unmarshal hit default case that should not be hit.");
                DOFMem_Free(permBuffer); 
                return DOFERROR_PARSE_ERROR;
                break;
            }
        }
    }
    else
    {
        DOF_LogMessage(DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALPermissionScope: OALPermissionScope_Unmarshal was given NULL parameters.");
        return DOFERROR_PARSE_ERROR;
    }
    return DOFERROR_OK;
}

boolean OALPermissionScope_IsCompatible(OALPermissionScope self,
                                        OALPermissionScope scope)
{
    uint8 i;
    uint8 j;

    PCR_Assert(self&&scope);

    /* traverse flags first for fully compatible first */
    OALPermissionScope_Lock();
    for( i = 0; i < self->idCount; ++i)
    {
        if(self->ids[i] == FULLLY_COMPATIBLE_SECURITYSCOPE)
        {
            OALPermissionScope_Unlock();
            return TRUE;
        }
    }

    /* traverse flags first for fully compatible first */
    for( i = 0; i < scope->idCount; ++i)
    {
        if(scope->ids[i] == FULLLY_COMPATIBLE_SECURITYSCOPE)
        {
            OALPermissionScope_Unlock();
            return TRUE;
        }
    }

    for(i = 0; i < (self->idCount); ++i)
    {
        for(j = 0; j < (scope->idCount); ++j)
        {
            if(self->ids[i] == scope->ids[j])
            {
                OALPermissionScope_Unlock();
                return TRUE;
            }
        }
    }
    OALPermissionScope_Unlock();
    return FALSE;
}

/* only for use with none terminal permissions */
boolean OALPermissionScope_IsMasking(OALPermissionScope self)
{
    if( self )
    {
        uint32 i = 0;
        OALPermissionScope_Lock();
        for( ; i < self->idCount; ++i)
        {
            if( self->ids[i] == NOT_MASKING_ID_SECURITYSCOPE ) /* if found that means not masking */
            {
                OALPermissionScope_Unlock();
                return FALSE;
            }
        }
        OALPermissionScope_Unlock();
    }
    return TRUE;
}

static OALPermissionScope get_compatible_scope(OALPermissionScope s1, OALPermissionScope s2)
{
    /* called with s1 having less permission ids to look through */
    uint32 i = 0;
    uint32 *storage;
    boolean s1All = FALSE;
    boolean s2All = FALSE;
    OALPermissionScope scope = NULL;

    OALPermissionScope_Lock();
    for(i = 0; i < s1->idCount; ++i)
    {
        if(s1->ids[i] == FULLLY_COMPATIBLE_SECURITYSCOPE)
        {
            s1All = TRUE;
            break;
        }
    }

    for(i = 0; i < s2->idCount; ++i)
    {
        if(s2->ids[i] == FULLLY_COMPATIBLE_SECURITYSCOPE)
        {
            s2All = TRUE;
            break;
        }
    }

    if( s1All && s2All)
    {
        /* return all id */
        scope = OALPermissionScope_Create(1);
        if(scope)
        {
            scope->ids[0] = FULLLY_COMPATIBLE_SECURITYSCOPE;
        }
    }
    else if(s1All || s2All)
    {
        uint32 cnt = (s1All)?s2->idCount:s1->idCount;
        uint32* ptr = (s1All)?s2->ids:s1->ids;
        scope = OALPermissionScope_Create(cnt);
        if(scope)
        {
            for(i=0 ; i<cnt; ++i)
            {
                scope->ids[i] = ptr[i];
            }
        }
    }
    else
    {
        storage = DOFMem_Alloc(s1->idCount*sizeof(uint32));
        if(storage)
        {
            unsigned int i = 0, k=0;

            for(; i<(s1->idCount); ++i)
            {
                unsigned int j = 0;

                for(; j<(s2->idCount); ++j)
                {
                    if(s1->ids[i]==s2->ids[j])
                    {
                        storage[k++] = s1->ids[i];
                        break;
                    }
                }
            }

            scope = OALPermissionScope_Create(k);
            if(scope) {
                for(i=0; i<k; ++i) scope->ids[i] = storage[i];
            }

            DOFMem_Free(storage);
        }
    }
    OALPermissionScope_Unlock();
    return scope;
}

OALPermissionScope OALPermissionScope_GetCompatibleScope(OALPermissionScope self,
                                                         OALPermissionScope scope)
{
    return (self->idCount < scope->idCount) ? get_compatible_scope(self, scope) : get_compatible_scope(scope, self);
}

int32 OALPermissionScope_Compare(OALPermissionScope self, OALPermissionScope scope)
{
    int32 ret;
    OALPermissionScope_Lock();
    ret = self->idCount - scope->idCount;
    if(ret == 0)
    {
        uint32 o,i;
        for(o = 0; o<self->idCount; ++o)
        {
            boolean found = FALSE;
            for(i = 0; i<scope->idCount; ++i)
            {
                if(self->ids[o] == scope->ids[i])
                {
                    found = TRUE;
                    break;
                }
            }

            if(!found)
            {
                OALPermissionScope_Unlock();
                return 1;
            }
        }
    }
    OALPermissionScope_Unlock();
    return ret;
}

OALPermissionScopeBuilder OALPermissionScopeBuilder_Create( void )
{
    OALPermissionScopeBuilder ret = DOFMem_Alloc(sizeof(struct OALPermissionScopeBuilder_t));

    if(ret != NULL)
    {
        ret->ids = DOFMem_Alloc((sizeof(uint32) * OALPERMISSIONSONSCOPE_ALLOC_STEP));
        if( ret->ids )
        {
            ret->alc = OALPERMISSIONSONSCOPE_ALLOC_STEP;
            ret->idCount = 0;
        }
        else
        {
            DOFMem_Free(ret);
            ret = NULL;
        }
    }
    return ret;
}

OALPermissionScopeBuilder OALPermissionScopeBuilder_CreateCopy( OALPermissionScopeBuilder self )
{
    OALPermissionScopeBuilder ret = OALPermissionScopeBuilder_Create();
    if( ret )
    {
        uint32 i;

        for( i = 0; i < self->idCount; ++i )
        {
            if( !OALPermissionScopeBuilder_AddID( ret, self->ids[i] ) )
            {
                OALPermissionScopeBuilder_Destroy( ret );
                ret = NULL;
                break;
            }
        }
    }
    return ret;
}

boolean OALPermissionScopeBuilder_AddID( OALPermissionScopeBuilder self, uint32 ID )
{
    if( self->idCount == self->alc )
    {
        void * newIds = DOFMem_Realloc( self->ids, (sizeof(uint32) * ( self->alc + OALPERMISSIONSONSCOPE_ALLOC_STEP )));

        if(newIds != NULL)
        {
            self->ids = newIds;
            self->alc += OALPERMISSIONSONSCOPE_ALLOC_STEP;
        }
        else
        {
            return FALSE;
        }
    }

    self->ids[ self->idCount ] = ID;
    ++self->idCount;
    return TRUE;
}

static void OALPermissionScopeBuilder_RemoveIDAtPosition( OALPermissionScopeBuilder self, uint32 position )
{
    uint32 i;

    --self->idCount;
    for( i = position; i < self->idCount; ++i )
    {
        self->ids[i] = self->ids[i+1];
    }
}

boolean OALPermissionScopeBuilder_DoIntersection( OALPermissionScopeBuilder self, OALPermissionScopeBuilder other )
{
    boolean masking = FALSE;
    int32 i;

    for( i = ( self->idCount - 1 ); i >= 0 ; --i )
    {
        uint32 j;
        boolean found = FALSE;
        for( j = 0; j < other->idCount; ++j )
        {
            if( self->ids[i] == other->ids[j])
            {
                found = TRUE;
                break;
            }
        }

        if( !found )
        {
            OALPermissionScopeBuilder_RemoveIDAtPosition( self, i);
            masking = TRUE;
        }
    }

    return masking;
}

OALPermissionScope OALPermissionScopeBuilder_Build( OALPermissionScopeBuilder self )
{
    OALPermissionScope ret = OALPermissionScope_Create( self->idCount );

    if( ret )
    {
        uint32 i = 0;
        uint32 next = 0;

        for( ; i < ret->idCount; ++i )
        {
            OALPermissionScope_SetID(ret, &next, self->ids[i]);
        }
    }

    return ret;
}

OALPermissionScope OALPermissionScopeBuilder_BuildAndDestroy( OALPermissionScopeBuilder self )
{
    OALPermissionScope ret = OALPermissionScopeBuilder_Build( self );
    OALPermissionScopeBuilder_Destroy( self );
    return ret;
}

void OALPermissionScopeBuilder_Destroy( OALPermissionScopeBuilder self )
{
    DOFMem_Free(self->ids);
    DOFMem_Free(self);
}
