/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/domain.h>
#include <dof/permission.h>
#include <dof/permissionset.h>
#include <dof/objectid.h>
#include <dof/interfaceid.h>
#include <dof/error.h>
#include <internal/objectidinternal.h>
#include <internal/permissioninternal.h>
#include <internal/permissionsetinternal.h>
#include <internal/interfaceidinternal.h>
#include <internal/marshal.h>
#include <internal/avl_tree.h>
#include "permissionset_structs.h"

extern PCRMutex ref_count_mutex;

#define DOFPERMISSIONBINDINGACTION_ACTIONSMASK DOFPERMISSIONBINDINGACTION_ALL
#define DOFPERMISSIONBINDINGACTION_FLAGSMASK 0xE0
#define DOFPERMISSIONBINDING_ALLAVPS 0x80
#define DOFPERMISSIONBINDING_WIID 0x40

static boolean DOFPermissionBindingBuilder_Internal_AddObjectID( DOFPermissionBindingBuilder self, DOFObjectID oid );

DOFPermission DOFPermissionRequestor_Create()
{
    DOFPermission ret = (DOFPermission)DOFMem_Alloc( sizeof( DOFPermission_t ) );
    if( ret )
    {
        ret->type = DOFPERMISSIONTYPE_REQUESTOR;
        ret->dataLength = 0;
        ret->refCount = 1;
    }
    return ret;
}

DOFPermission DOFPermissionProvider_Create()
{
    DOFPermission ret = (DOFPermission)DOFMem_Alloc( sizeof( DOFPermission_t ) );
    if( ret )
    {
        ret->type = DOFPERMISSIONTYPE_PROVIDER;
        ret->dataLength = 0;
        ret->refCount = 1;
    }
    return ret;
}

DOFPermission DOFPermissionDefine_Create( const DOFInterfaceID iid )
{
    DefinePermission_t* ret = NULL;

    if( iid )
    {
        ret = (DefinePermission_t*)DOFMem_Alloc( sizeof( DefinePermission_t ) );
        if( ret )
        {
            ret->iid = DOFInterfaceID_Copy( iid );
            if( ret->iid )
            {
                ret->base.type = DOFPERMISSIONTYPE_DEFINE;
                ret->base.dataLength = DOFInterfaceID_GetBytesSize( iid );
                ret->base.refCount = 1;
                return (DOFPermission)ret;
            }
            DOFMem_Free( ret );
        }
    }
    return NULL;
}

static DOFPermission DOFPermissionEPP_Internal_Create( DOFPermissionType type, DOFObjectIDSource sourceID )
{
    EPPPermission_t* ret = (EPPPermission_t*)DOFMem_Alloc( sizeof( EPPPermission_t ) );
    if( ret )
    {
        ret->sid = DOFObjectID_Copy( DOFObjectIDSource_GetObjectID( sourceID ) );
        if (!sourceID || ret->sid)
        {
            ret->base.type = type;
            ret->base.dataLength = ret->sid ? DOFObjectID_GetBytesSize( ret->sid ) : 0;
            ret->base.refCount = 1;
            return (DOFPermission)ret;
        }
        DOFMem_Free( ret );
    }
    return NULL;
}

DOFPermission DOFPermissionIAm_Create( DOFObjectIDSource sourceID )
{
    if ( sourceID )
        return DOFPermissionEPP_Internal_Create(DOFPERMISSIONTYPE_I_AM, sourceID);
    return NULL;
}

DOFPermission DOFPermissionActAs_Create( DOFObjectIDSource sourceID )
{
    if ( sourceID )
        return DOFPermissionEPP_Internal_Create(DOFPERMISSIONTYPE_ACT_AS, sourceID);
    return NULL;
}

DOFPermission DOFPermissionActAsAny_Create( void )
{
    return DOFPermissionEPP_Internal_Create(DOFPERMISSIONTYPE_ACT_AS, NULL);
}

DOFPermission DOFPermissionTunnelDomain_Create( uint32 rdid )
{
    TunPermission_t* ret = (TunPermission_t*)DOFMem_Alloc( sizeof( TunPermission_t ) );
    if( ret )
    {
        ret->base.type = DOFPERMISSIONTYPE_TUNNEL_DOMAIN;
        ret->base.dataLength = OALMarshal_GetCompressedValueSize(4,rdid);
        ret->base.refCount = 1;
        ret->rdi = rdid;
    }
    return (DOFPermission)ret;
}

static boolean DOFPermissionBindingBuilder_AddObjectIDStripAttribute( DOFPermissionBindingBuilder self, DOFObjectID oid )
{
    if( self )
    {
        if( DOFObjectID_HasAttributes( oid ) )
        {
            DOFObjectID baseOID = DOFObjectID_Create( DOFObjectID_GetIDClass(oid), DOFObjectID_GetDataSize(oid), DOFObjectID_GetData(oid) );
            if( baseOID )
            {
                if( DOFPermissionBindingBuilder_Internal_AddObjectID( self, baseOID ) )
                {
                    uint8 i;
                    uint8 avpCnt = DOFObjectID_GetAttributeCount( oid );
                    boolean allAdded = TRUE;

                    for( i = 0; i < avpCnt; ++i )
                    {
                       DOFObjectIDAttribute atr_o = DOFObjectID_GetAttributeAtIndex( oid, i);
                       if( !DOFPermissionBindingBuilder_AddRequiredAttribute_Buffer( self, atr_o.id, atr_o.dataSize, atr_o.data ) )
                       {
                           allAdded = FALSE;
                           break;
                       }
                    }
                    DOFObjectID_Destroy( baseOID );
                    return allAdded;
                }
                DOFObjectID_Destroy( baseOID );
            }
        }
        else
        {
            return DOFPermissionBindingBuilder_AddObjectID( self, oid );
        }
    }
    return FALSE;
}

DOFPermission DOFPermissionBinding_Create( DOFPermissionBindingAction action, const DOFObjectID oid, const DOFInterfaceID iid )
{
    boolean error = TRUE;
    DOFPermissionBindingBuilder perm = DOFPermissionBindingBuilder_Create( action ); /* This sets the all attributes flag that will be cleared if needed below */

    if( perm )
    {
        if( !iid || DOFPermissionBindingBuilder_AddInterfaceID( perm, iid ) )
        {
            if( oid )
            {
                if( iid || !DOFObjectID_HasAttributes( oid ) )
                {
                    if( DOFPermissionBindingBuilder_SetAllAttributesAllowed( perm, FALSE ) )
                    {
                        if( DOFPermissionBindingBuilder_Internal_AddObjectID( perm, oid ) )
                            error = FALSE;
                    }
                }
                else
                {
                    /* This will clear the all attributes when the attributes are added from oid */
                    if( DOFPermissionBindingBuilder_AddObjectIDStripAttribute( perm, oid ) )
                        error = FALSE;
                }
            }
            else
            {
                error = FALSE;
            }
        }

        if( error )
        {
            DOFPermissionBindingBuilder_Destroy( perm );
            perm = NULL;
        }
    }
    return DOFPermissionBindingBuilder_BuildAndDestroy( perm );
}

static void oidDestroy( OALAVLTreeItem_t itemx, void *arg )
{
    DOFObjectID item = (DOFObjectID)itemx;
    DOFPermissionBindingBuilder self = (DOFPermissionBindingBuilder)arg;

    if( self )
        self->base.dataLength -= (item ? DOFObjectID_GetBytesSize( item ): 0);
    DOFObjectID_Destroy(item);
}

static void iidDestroy( OALAVLTreeItem_t itemx, void *arg )
{
    DOFInterfaceID item = (DOFInterfaceID)itemx;
    DOFPermissionBindingBuilder self = (DOFPermissionBindingBuilder)arg;

    if( self )
        self->base.dataLength -= DOFInterfaceID_GetBytesSize(item);
    DOFInterfaceID_Destroy(item);
}

static void avpDestroy( OALAVLTreeItem_t itemx, void *arg )
{
    DOFObjectIDAttribute* item = (DOFObjectIDAttribute*)itemx;
    DOFPermissionBindingBuilder self = (DOFPermissionBindingBuilder)arg;

    if( self )
        self->base.dataLength -= (2+item->dataSize);
    DOFMem_Free(item);
}

static void BindingPermission_Dtor( DOFPermission self )
{
    DOFPermissionBuilder_t* perm = (DOFPermissionBuilder_t*)self;

    if( perm->objectIDs )
        OALAVLTree_Destroy( perm->objectIDs, oidDestroy, NULL);
    if( perm->interfaces )
        OALAVLTree_Destroy( perm->interfaces, iidDestroy, NULL );
    if( perm->avps )
        OALAVLTree_Destroy( perm->avps, avpDestroy, NULL);
}

static void DefinedPermission_Dtor( DOFPermission self )
{
    DOFInterfaceID_Destroy(((DefinePermission_t*)self)->iid);
}

static void EPPPermission_Dtor( DOFPermission self )
{
    DOFObjectID_Destroy(((EPPPermission_t*)self)->sid);
}

void DOFPermission_Destroy( DOFPermission self )
{
    boolean free_ok = FALSE;
    if( self )
    {
        PCRMutex_Lock(ref_count_mutex);
        if( !--(self->refCount) ) free_ok = TRUE;
        PCRMutex_Unlock(ref_count_mutex);

        if( free_ok )
        {
            switch( self->type )
            {
            case DOFPERMISSIONTYPE_BINDING:
            case DOFPERMISSIONTYPE_BINDING_EXPLICIT:
                BindingPermission_Dtor(self);
                break;
            case DOFPERMISSIONTYPE_I_AM:
                EPPPermission_Dtor(self);
                break;
            case DOFPERMISSIONTYPE_ACT_AS:
                EPPPermission_Dtor(self);
                break;
            case DOFPERMISSIONTYPE_REQUESTOR:
            case DOFPERMISSIONTYPE_PROVIDER:
                break;
            case DOFPERMISSIONTYPE_DEFINE:
                DefinedPermission_Dtor(self);
                break;
            case DOFPERMISSIONTYPE_TUNNEL_DOMAIN:
            default:
                break;
            }
            DOFMem_Free(self);
            return;
        }
    }
}

DOFPermissionType DOFPermission_GetPermissionType( DOFPermission self )
{
    if( self )
        switch( self->type )
        {
            case DOFPERMISSIONTYPE_BINDING_EXPLICIT:
                return DOFPERMISSIONTYPE_BINDING;
            default:
                return self->type;
        }
    return 0;
}

static boolean objectid_gt( OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context )
{
    return (DOFObjectID_Compare( (DOFObjectID) key, (DOFObjectID) item ) > 0);
}

static boolean objectid_eq( OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context )
{
    return (DOFObjectID_Compare( (DOFObjectID) key, (DOFObjectID) item ) == 0);
}

static boolean iid_gt( OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context )
{
    return (DOFInterfaceID_Compare( (DOFInterfaceID)key, (DOFInterfaceID)item ) > 0);
}

static boolean iid_eq( OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context )
{
    return (DOFInterfaceID_Compare( (DOFInterfaceID)key, (DOFInterfaceID)item ) == 0);
}

static boolean avp_gt( OALAVLTreeKey_t keyx, OALAVLTreeItem_t itemx, void *context )
{
    DOFObjectIDAttribute* key = (DOFObjectIDAttribute*)keyx;
    DOFObjectIDAttribute* item = (DOFObjectIDAttribute*)itemx;
    return ((key->id & 0x7F) > (item->id & 0x7F)); /* mask off A for the compare */
}

static boolean avp_eq( OALAVLTreeKey_t keyx, OALAVLTreeItem_t itemx, void *context )
{
    DOFObjectIDAttribute* key = (DOFObjectIDAttribute*)keyx;
    DOFObjectIDAttribute* item = (DOFObjectIDAttribute*)itemx;
    return ((key->id & 0x7F) == (item->id & 0x7F)); /* mask off A for the compare */
}

static DOFPermissionBindingBuilder DOFPermissionBindingBuilder_Internal_Create( void )
{
    DOFPermissionBindingBuilder ret = (DOFPermissionBindingBuilder)DOFMem_Alloc( sizeof( DOFPermissionBuilder_t ) );
    if( ret )
    {
        PCRMem_Set( ret, 0, sizeof( DOFPermissionBuilder_t ) );
        ret->base.type = DOFPERMISSIONTYPE_BINDING;
        ret->base.dataLength = 4;
        ret->base.refCount = 1;
        ret->flags = DOFPERMISSIONBINDING_WIID | DOFPERMISSIONBINDING_ALLAVPS;
        ret->objectIDs = OALAVLTree_Create( objectid_gt, objectid_eq, NULL );
        if(ret->objectIDs != NULL)
        {
            ret->interfaces = OALAVLTree_Create( iid_gt, iid_eq, NULL );
            if(ret->interfaces != NULL)
            {
                ret->avps = OALAVLTree_Create( avp_gt, avp_eq, NULL );
                if(ret->avps != NULL)
                {
                    return ret;
                }
                OALAVLTree_Destroy(ret->interfaces, NULL, NULL);
            }
            OALAVLTree_Destroy(ret->objectIDs, NULL, NULL);
        }
        DOFMem_Free(ret);
        ret = NULL;
    }
    return ret;
}

DOFPermissionBindingBuilder DOFPermissionBindingBuilder_Create( DOFPermissionBindingAction actions )
{
    DOFPermissionBindingBuilder ret = NULL;
    if( actions > 0 && actions <= DOFPERMISSIONBINDINGACTION_ALL )
    {
        ret = DOFPermissionBindingBuilder_Internal_Create();
        if( ret )
        {
            ret->flags |= actions;
        }
    }
    return ret;
}

static boolean DOFPermissionBindingBuilder_Internal_AddObjectID( DOFPermissionBindingBuilder self, DOFObjectID oid )
{
    if( self )
    {
        if( oid )
        {
            DOFObjectID temp = DOFObjectID_Copy(oid);
            if( temp )
            {
                OALAVLTree_InsertionStatus ret = OALAVLTree_Insert( self->objectIDs, temp, temp, NULL );

                switch( ret )
                {
                case OALAVLTree_InsertSuccess:
                    self->base.dataLength += DOFObjectID_GetBytesSize( oid );
                    return TRUE;
                case OALAVLTree_AlreadyInserted:
                    DOFObjectID_Destroy( temp );
                    return TRUE;
                default:
                    break;
                }
                DOFObjectID_Destroy( temp );
            }
        }
        else
        {
            if( OALAVLTree_GetNumItems(self->objectIDs) > 0 )
            {
                OALAVLTree_Empty(self->objectIDs, oidDestroy, self);
            }
            return TRUE;
        }
    }
    return FALSE;
}

boolean DOFPermissionBindingBuilder_AddObjectID( DOFPermissionBindingBuilder self, DOFObjectID oid )
{
    if( !DOFObjectID_HasAttributes(oid) )
        return DOFPermissionBindingBuilder_Internal_AddObjectID( self, oid );

    return FALSE;
}

boolean DOFPermissionBindingBuilder_AddInterfaceID( DOFPermissionBindingBuilder self, DOFInterfaceID iid )
{
    if( self )
    {
        if( iid && !DOFInterfaceID_IsWildcard(iid) )
        {
            DOFInterfaceID temp = DOFInterfaceID_Copy(iid);
            if( temp )
            {
                OALAVLTree_InsertionStatus ret = OALAVLTree_Insert( self->interfaces, temp, temp, NULL );
                switch( ret )
                {
                case OALAVLTree_InsertSuccess:
                    self->base.dataLength += DOFInterfaceID_GetBytesSize( iid );
                    self->flags &= ~DOFPERMISSIONBINDING_WIID;
                    return TRUE;
                case OALAVLTree_AlreadyInserted:
                    DOFInterfaceID_Destroy( temp );
                    return TRUE;
                default:
                    break;
                }
                DOFInterfaceID_Destroy( temp );
            }
        }
        else
        {
            if( OALAVLTree_GetNumItems(self->interfaces) > 0 )
            {
                OALAVLTree_Empty(self->interfaces, iidDestroy, self);
            }
            self->flags |= DOFPERMISSIONBINDING_WIID;
            return TRUE;
        }
    }
    return FALSE;
}

static boolean DOFPermissionBindingBuilder_Internal_AddRequiredAttribute( DOFPermissionBindingBuilder self, boolean wildcard, const DOFObjectIDAttributeType id, uint32 len, const uint8 * data )
{
    if ( self && len < 256 && id < DOFOBJECTIDATTRIBUTE_INVALID )
    {
        DOFObjectIDAttribute* avp;

        avp = DOFMem_Alloc(sizeof(DOFObjectIDAttribute) + len);
        if( avp )
        {
            OALAVLTree_InsertionStatus ret;

            avp->id = id;
            avp->dataSize = len;

            if( wildcard )
            {
                avp->id |= 0x80;
                avp->data = NULL;
            }
            else
            {
                avp->data = (((uint8*)avp)+sizeof(DOFObjectIDAttribute));
                PCRMem_Copy((uint8*)avp->data, data, len);
            }

            ret = OALAVLTree_Insert(self->avps, avp, avp, NULL);
            switch( ret )
            {
            case OALAVLTree_InsertSuccess:
                self->base.dataLength += (2+len);
                self->flags &= ~DOFPERMISSIONBINDING_ALLAVPS; /* clear the all avp flag */
                return TRUE;
            case OALAVLTree_AlreadyInserted:
                DOFMem_Free( avp );
                return TRUE;
            default:
                break;
            }
            DOFMem_Free(avp);
        }
    }
    return FALSE;
}

boolean DOFPermissionBindingBuilder_SetAllAttributesAllowed( DOFPermissionBindingBuilder self, boolean allowed )
{
    /* fail if the builder already has avps */
    if( self && ( OALAVLTree_GetNumItems( self->avps ) == 0 ) )
    {
        if( allowed )
            self->flags |= DOFPERMISSIONBINDING_ALLAVPS;
        else
            self->flags &= ~DOFPERMISSIONBINDING_ALLAVPS;

        return TRUE;
    }
    return FALSE;
}

boolean DOFPermissionBindingBuilder_AddRequiredWildcardAttribute( DOFPermissionBindingBuilder self, const DOFObjectIDAttributeType id )
{
    return DOFPermissionBindingBuilder_Internal_AddRequiredAttribute(self, TRUE, id, 0, NULL);
}

boolean DOFPermissionBindingBuilder_AddRequiredAttribute_Buffer( DOFPermissionBindingBuilder self, const DOFObjectIDAttributeType id, uint32 len, const uint8 * data )
{
    return DOFPermissionBindingBuilder_Internal_AddRequiredAttribute(self, FALSE, id, len, data);
}

boolean DOFPermissionBindingBuilder_AddRequiredAttribute_String( DOFPermissionBindingBuilder self, const DOFObjectIDAttributeType id, const char * data )
{
    DOFObjectIDSize len = PCRString_Length( data );
    if ( len <= 255 )
        return DOFPermissionBindingBuilder_AddRequiredAttribute_Buffer( self, id, len, (const uint8 *) data );
    return FALSE;
}

boolean DOFPermissionBindingBuilder_AddRequiredAttribute_ObjectID( DOFPermissionBindingBuilder self, const DOFObjectIDAttributeType id, const DOFObjectID data )
{
    DOFObjectIDSize len = data ? DOFObjectID_GetBytesSize( data ) : 0;

    if ( len <= 255 )
        return DOFPermissionBindingBuilder_AddRequiredAttribute_Buffer( self, id, len, DOFObjectID_GetBytes( data ) );
    return FALSE;
}

void DOFPermissionBindingBuilder_AddActions( DOFPermissionBindingBuilder self, DOFPermissionBindingAction action )
{
    if ( self && ( action > 0 && action <= DOFPERMISSIONBINDINGACTION_ALL ) )
        self->flags |= action;
}

boolean DOFPermissionBindingBuilder_SetActions( DOFPermissionBindingBuilder self, DOFPermissionBindingAction action )
{
    if ( self && ( action > 0  && action <= DOFPERMISSIONBINDINGACTION_ALL ) )
    {
        self->flags = (self->flags & DOFPERMISSIONBINDINGACTION_FLAGSMASK) + action;
        return TRUE;
    }
    return FALSE;
}

struct permissionBuilderAndError_t
{
    DOFPermissionBindingBuilder perm;
    boolean error;
};

static boolean oidCopy( OALAVLTreeItem_t item, void *arg )
{
    DOFObjectID oid = (DOFObjectID)item;
    struct permissionBuilderAndError_t* ctx = (struct permissionBuilderAndError_t*)arg;

    if( !DOFPermissionBindingBuilder_Internal_AddObjectID(ctx->perm, oid) )
    {
        ctx->error = TRUE;
        return TRUE; /* Exit Early For Error */
    }
    return FALSE;
}

static boolean iidCopy( OALAVLTreeItem_t item, void *arg )
{
    DOFInterfaceID iid = (DOFInterfaceID)item;
    struct permissionBuilderAndError_t* ctx = (struct permissionBuilderAndError_t*)arg;

    if( !DOFPermissionBindingBuilder_AddInterfaceID(ctx->perm,iid) )
    {
        ctx->error = TRUE;
        return TRUE; /* Exit Early For Error */
    }
    return FALSE;
}

static boolean avpCopy(  OALAVLTreeItem_t item, void *arg )
{
    DOFObjectIDAttribute* avp = (DOFObjectIDAttribute*)item;
    struct permissionBuilderAndError_t* ctx = (struct permissionBuilderAndError_t*)arg;

    if( !DOFPermissionBindingBuilder_Internal_AddRequiredAttribute( ctx->perm, (avp->id & 0x80), (avp->id & 0x7F), avp->dataSize, avp->data ) )
    {
        ctx->error = TRUE;
        return TRUE; /* Exit Early For Error */
    }
    return FALSE;
}

static boolean DOFPermissionBindingBuilder_IsWithoutWildcard( DOFPermissionBindingBuilder self )
{
    if(self->base.type == DOFPERMISSIONTYPE_BINDING_EXPLICIT ||
       (!(self->flags & (DOFPERMISSIONBINDING_ALLAVPS | DOFPERMISSIONBINDING_WIID)) &&
       OALAVLTree_GetNumItems(self->objectIDs) == 1 &&
       OALAVLTree_GetNumItems(self->interfaces) == 1 &&
       OALAVLTree_GetNumItems(self->avps) == 0))
        return TRUE;
    return FALSE;
}

static boolean OID_GetAVPs( OALAVLTreeItem_t item, void *arg )
{
    DOFObjectID objectID = (DOFObjectID)item;
    struct permissionBuilderAndError_t* ctx = (struct permissionBuilderAndError_t*)arg;
    uint8 i = 0;
    uint8 cnt = DOFObjectID_GetAttributeCount(objectID);
    uint16 len = ctx->perm->base.dataLength;

    for( ; i<cnt; ++i )
    {
        DOFObjectIDAttribute atrb = DOFObjectID_GetAttributeAtIndex(objectID, i);
        if( !DOFPermissionBindingBuilder_Internal_AddRequiredAttribute(ctx->perm, (atrb.id&0x80), atrb.id, atrb.dataSize, atrb.data) )
        {
            ctx->error = TRUE;
        }
    }

    ctx->perm->base.dataLength = len;

    return TRUE; /* should only be one */
}

/* converts permission to type Without wildcard */
static DOFPermission DOFPermissionBindingBuilder_BuildWithoutWildcard( DOFPermissionBindingBuilder self )
{
    if( self )
    {
        struct permissionBuilderAndError_t ctx;

        ctx.error = FALSE;
        ctx.perm = self;

        self->base.type = DOFPERMISSIONTYPE_BINDING_EXPLICIT;
        PCR_Assert( OALAVLTree_GetNumItems(self->objectIDs) == 1 );
        OALAVLTree_Traverse(self->objectIDs, OID_GetAVPs, &ctx);
        if( !ctx.error )
        {
            self->base.dataLength -= 3; /* we only have the actions byte */
            return (DOFPermission)self;
        }
        DOFPermissionBindingBuilder_Destroy( self );
        return NULL;
    }
    return (DOFPermission)self;
}

DOFPermission DOFPermissionBindingBuilder_Build( DOFPermissionBindingBuilder self )
{
    DOFPermissionBindingBuilder ret = NULL;

    if( self )
    {
        ret = DOFPermissionBindingBuilder_Internal_Create();
        if( ret )
        {
            struct permissionBuilderAndError_t ctx;

            ctx.error = FALSE;
            ctx.perm = ret;

            ret->flags = self->flags;
            ret->base.type = self->base.type;
            OALAVLTree_Traverse( self->objectIDs, oidCopy, &ctx);
            OALAVLTree_Traverse( self->interfaces, iidCopy, &ctx);
            OALAVLTree_Traverse( self->avps, avpCopy, &ctx);
            if( !ctx.error )
            {
                ret->base.dataLength = self->base.dataLength;
                return (DOFPermission)ret;
            }
            DOFPermissionBindingBuilder_Destroy( ret );
        }
    }
    return (DOFPermission)ret;
}

DOFPermission DOFPermissionBindingBuilder_BuildAndDestroy( DOFPermissionBindingBuilder self )
{
    if( self )
    {
        if( !DOFPermissionBindingBuilder_IsWithoutWildcard(self) )
        {
            return (DOFPermission)self;
        }
        else
        {
            return DOFPermissionBindingBuilder_BuildWithoutWildcard(self);
        }
    }
    return NULL;
}

void DOFPermissionBindingBuilder_Destroy( DOFPermissionBindingBuilder self )
{
    DOFPermission_Destroy((DOFPermission)self);
}

DOFPermission DOFPermission_Copy( DOFPermission self )
{
    if(self)
    {
        DOFPermission ret = self;

        PCRMutex_Lock(ref_count_mutex);
        if( self->refCount < 0xFFFFFFFF )
            ++(self->refCount);
        else
            ret = NULL;
        PCRMutex_Unlock(ref_count_mutex);
        return ret;
    }
    return NULL;
}

uint32 DOFPermission_MarshalByteLength( DOFPermission self )
{
    if( self )
        return (OALMarshal_GetCompressedValueSize( 2, (self->type)) + (((self->type & 0x0001) != 0)?(OALMarshal_GetCompressedValueSize( 2, (self->dataLength) )+ self->dataLength):0));
    return 0;
}

struct marshalContext_t
{
    OALStream   stream;
    boolean     fail;
}marshalContext_t;

static boolean oidMarshal( OALAVLTreeItem_t item, void *arg )
{
    DOFObjectID oid = (DOFObjectID)item;
    struct marshalContext_t* ctx = (struct marshalContext_t*)arg;

    if( DOFObjectID_Marshal( oid, ctx->stream ) )
        return FALSE;
    ctx->fail = TRUE;
    return TRUE; /* exit early */
}

static boolean iidMarshal( OALAVLTreeItem_t item, void *arg )
{
    DOFInterfaceID iid = (DOFInterfaceID)item;
    struct marshalContext_t* ctx = (struct marshalContext_t*)arg;

    if( DOFInterfaceID_MarshalToStream( iid, ctx->stream ) )
        return FALSE;
    ctx->fail = TRUE;
    return TRUE; /* exit early */
}

static boolean avpMarshal( OALAVLTreeItem_t item, void *arg )
{
    DOFObjectIDAttribute* avp = (DOFObjectIDAttribute*)item;
    struct marshalContext_t* ctx = (struct marshalContext_t*)arg;

    if( OALStream_Write8( ctx->stream, avp->id ) &&
        OALStream_Write8( ctx->stream, avp->dataSize ) &&
        (!avp->dataSize || ( avp->dataSize && OALStream_Write( ctx->stream, avp->dataSize, avp->data ) )))
        return FALSE;
    ctx->fail = TRUE;
    return TRUE; /* exit early */
}

static boolean marshal_binding_wo_permission( DOFPermission self, OALStream stream )
{
    DOFPermissionBindingBuilder perm = (DOFPermissionBindingBuilder)self;

    if ( OALMarshal_StreamCompress2(stream, self->dataLength) == DOFERROR_OK )
    {
        uint32 offset = OALStream_GetWriteOffset( stream );
        if ( OALStream_Write8(stream, (perm->flags&DOFPERMISSIONBINDINGACTION_ACTIONSMASK)) )
        {
            struct marshalContext_t ctx;
            ctx.fail = FALSE;
            ctx.stream = stream;

            OALAVLTree_Traverse(perm->interfaces, iidMarshal, &ctx );
            if( !ctx.fail )
            {
                OALAVLTree_Traverse(perm->objectIDs, oidMarshal,  &ctx );
                if( !ctx.fail )
                {
                    if ( OALStream_GetWriteOffset( stream ) - offset == self->dataLength )
                        return TRUE;
                }
            }
        }
    }
    return FALSE;
}

static boolean marshal_binding_permission( DOFPermission self, OALStream stream )
{
    DOFPermissionBindingBuilder perm = (DOFPermissionBindingBuilder)self;
    uint8 avpCnt = ((OALAVLTree_GetNumItems(perm->avps)>0)?OALAVLTree_GetNumItems(perm->avps):((perm->flags & DOFPERMISSIONBINDING_ALLAVPS)?0:255));

    if ( OALMarshal_StreamCompress2(stream, self->dataLength) == DOFERROR_OK )
    {
        uint32 offset = OALStream_GetWriteOffset( stream );
        if ( OALStream_Write8(stream, (perm->flags&DOFPERMISSIONBINDINGACTION_ACTIONSMASK)) &&
             OALStream_Write8(stream, OALAVLTree_GetNumItems(perm->interfaces)) &&
             OALStream_Write8(stream, OALAVLTree_GetNumItems(perm->objectIDs)) &&
             OALStream_Write8(stream, avpCnt) )
        {
            struct marshalContext_t ctx;
            ctx.fail = FALSE;
            ctx.stream = stream;

            OALAVLTree_Traverse(perm->interfaces, iidMarshal, &ctx );
            if( !ctx.fail )
            {
                OALAVLTree_Traverse(perm->objectIDs, oidMarshal, &ctx );
                if( !ctx.fail )
                {
                    OALAVLTree_Traverse(perm->avps, avpMarshal, &ctx );
                    if( !ctx.fail )
                    {
                        if ( OALStream_GetWriteOffset( stream ) - offset == self->dataLength )
                            return TRUE;
                    }
                }
            }
        }
    }
    return FALSE;
}

static boolean marshal_epp_permission( DOFPermission self, OALStream stream )
{
    EPPPermission_t* perm = (EPPPermission_t*)self;

    if(OALMarshal_StreamCompress2(stream, self->dataLength) == DOFERROR_OK)
    {
        if( self->dataLength )
        {
            if( DOFObjectID_Marshal(perm->sid, stream) )
            {
                return TRUE;
            }
        }
        else
        {
            return TRUE;
        }
    }
    return FALSE;
}

static boolean marshal_tun_permission( DOFPermission self, OALStream stream )
{
    TunPermission_t* perm = (TunPermission_t*)self;

    if( (OALMarshal_StreamCompress2(stream, self->dataLength) == DOFERROR_OK) &&
        (OALMarshal_StreamCompress4(stream,perm->rdi) == DOFERROR_OK))
        return TRUE;

    return FALSE;
}

static boolean marshal_unknown_permission( DOFPermission self, OALStream stream )
{
    if( self->type & 0x0001 )/* odd */
    {
        UnknownTypePermission_t* perm = (UnknownTypePermission_t*)self;
        if((OALMarshal_StreamCompress2(stream, self->dataLength) == DOFERROR_OK) &&
            OALStream_Write(stream, self->dataLength, perm->data))
            return TRUE;
    }
    else
    {
        return TRUE;
    }
    return FALSE;
}

boolean DOFPermission_Internal_Marshal( DOFPermission self, OALStream stream )
{
    PCR_Assert(self);
    if(OALMarshal_StreamCompress2(stream, self->type) == DOFERROR_OK)
    {

        switch(self->type)
        {
        case DOFPERMISSIONTYPE_BINDING:
            return marshal_binding_permission( self, stream );
            break;
        case DOFPERMISSIONTYPE_I_AM:
        case DOFPERMISSIONTYPE_ACT_AS:
            return marshal_epp_permission(self, stream);
            break;
        case DOFPERMISSIONTYPE_BINDING_EXPLICIT:
            return marshal_binding_wo_permission( self, stream );
            break;
        case DOFPERMISSIONTYPE_DEFINE:
            if( (OALMarshal_StreamCompress2( stream, self->dataLength ) == DOFERROR_OK) && DOFInterfaceID_MarshalToStream(((DefinePermission_t*)self)->iid, stream))
                return TRUE;
            return FALSE;
            break;
        case DOFPERMISSIONTYPE_REQUESTOR:
        case DOFPERMISSIONTYPE_PROVIDER:
            return TRUE;
            break;
        case DOFPERMISSIONTYPE_TUNNEL_DOMAIN:
            return marshal_tun_permission(self, stream);
            break;
        default:
            return marshal_unknown_permission(self, stream);
            break;
        }
    }
    return FALSE;
}

static DOFPermission unmarshal_binding_permission( OALStream stream )
{
    DOFPermissionBindingBuilder perm = DOFPermissionBindingBuilder_Internal_Create();

    if( perm )
    {
        uint8 i;
        uint8 oidCnt = 0;
        uint8 iidCnt = 0;
        uint8 avpCnt = 0;
        uint16 len = 0;

        if((OALMarshal_StreamUncompress2( stream, &len ) == DOFERROR_OK) &&
            OALStream_Read8( stream, &perm->flags ) && !(perm->flags & 0xE0) &&
            OALStream_Read8( stream, &iidCnt ) &&
            OALStream_Read8( stream, &oidCnt ) &&
            OALStream_Read8( stream, &avpCnt ) )
        {
            uint8 lastID = 0;
            DOFObjectID lastOID = NULL;
            DOFInterfaceID lastIID = NULL;
            uint32 startOffset = OALStream_GetReadOffset(stream);

            if ( iidCnt == 0 )
            {
                perm->flags |= DOFPERMISSIONBINDING_WIID;
            }
            else
            {
                for( i = 0; i < iidCnt; i++ )
                {
                    DOFInterfaceID iid = DOFInterfaceID_Unmarshal( stream );
                    if(!iid ||
                       (i>0 && (DOFInterfaceID_Compare(iid,lastIID)<=0)) ||
                       !DOFPermissionBindingBuilder_AddInterfaceID( perm, iid ))
                    {
                        DOFInterfaceID_Destroy( iid );
                        DOFPermission_Destroy( (DOFPermission)perm );
                        return NULL;
                    }
                    lastIID = iid;
                    DOFInterfaceID_Destroy( iid );
                }
            }

            for( i = 0; i < oidCnt; i++ )
            {
                DOFObjectID oid = DOFObjectID_Unmarshal( stream );
                if(!oid ||
                   (i>0 && (DOFObjectID_Compare(oid,lastOID)<=0)) ||
                   !DOFPermissionBindingBuilder_AddObjectID( perm, oid ))
                {
                    DOFObjectID_Destroy( oid );
                    DOFPermission_Destroy( (DOFPermission)perm );
                    return NULL;
                }
                lastOID = oid;
                DOFObjectID_Destroy( oid );
            }

            if( avpCnt == 0)
            {
                perm->flags |= DOFPERMISSIONBINDING_ALLAVPS;
            }
            else if( avpCnt != 255 )
            {
                for( i = 0; i < avpCnt; i++ )
                {
                    uint8 id;
                    uint8 dataLen;

                    if(OALStream_Read8(stream, &id) && OALStream_Read8(stream, &dataLen) && OALStream_GetDataSize(stream)>=dataLen)
                    {
                        if((i>0 && id<=lastID) ||
                           !DOFPermissionBindingBuilder_Internal_AddRequiredAttribute( perm, (id & 0x80), (id & 0x7F), dataLen, OALStream_GetReadPtr(stream) ))
                        {
                            DOFPermission_Destroy( (DOFPermission)perm );
                            return NULL;
                        }
                        OALStream_AdjustReadOffset(stream,dataLen);
                        lastID = id;
                    }
                    else
                    {
                        DOFPermission_Destroy( (DOFPermission)perm );
                        return NULL;
                    }
                }
            }

            if( startOffset+(len-4) != OALStream_GetReadOffset(stream) )
            {
                /* permission was longer than it said it was going to be: Fail */
                DOFPermission_Destroy( (DOFPermission)perm );
                return NULL;
            }

            perm = (DOFPermissionBindingBuilder)DOFPermissionBindingBuilder_BuildAndDestroy( perm );
            return(DOFPermission)perm;
        }
        DOFPermission_Destroy( (DOFPermission)perm );
    }
    return NULL;
}

static DOFPermission unmarshal_binding_wo_permission( OALStream stream )
{
    DOFPermissionBindingBuilder perm = DOFPermissionBindingBuilder_Internal_Create();

    if( perm )
    {
        uint16 len = 0;
        DOFInterfaceID iid = NULL;
        DOFObjectID oid = NULL;

        if(OALMarshal_StreamUncompress2( stream, &len ) == DOFERROR_OK)
        {
            uint32 startOffset = OALStream_GetReadOffset(stream);
            if( OALStream_Read8( stream, &perm->flags ) && !(perm->flags & 0xE0) &&
                ((iid = DOFInterfaceID_Unmarshal( stream )) != NULL) &&
                ((oid = DOFObjectID_Unmarshal( stream )) != NULL) &&
                DOFPermissionBindingBuilder_AddInterfaceID( perm, iid ) &&
                DOFPermissionBindingBuilder_Internal_AddObjectID( perm, oid ) &&
               ((startOffset+len)==OALStream_GetReadOffset(stream)))
            {
                DOFObjectID_Destroy(oid);
                DOFInterfaceID_Destroy(iid);
                return DOFPermissionBindingBuilder_BuildAndDestroy( perm );
            }
            DOFObjectID_Destroy(oid);
            DOFInterfaceID_Destroy(iid);
        }
        DOFPermission_Destroy( (DOFPermission)perm );
    }
    return NULL;
}

static DOFPermission unmarshal_epp_permission( OALStream stream, DOFPermissionType type )
{
    EPPPermission_t* ret = NULL;
    uint16 len = 0;
    DOFObjectID temp = NULL;

    if(DOFERROR_OK == OALMarshal_StreamUncompress2( stream, &len ))
    {
        if( len > 0 )
        {
            uint32_t start = OALStream_GetDataSize(stream);

            if(((temp = DOFObjectID_Unmarshal(stream)) == NULL))
            {
                return NULL;
            }

            if( ((start - OALStream_GetDataSize(stream)) != len) || !DOFObjectID_IsSourceID(temp) )
            {
                DOFObjectID_Destroy(temp);
                return NULL;
            }
        }
        else if( type != DOFPERMISSIONTYPE_ACT_AS )
        {
            /* ACT AS is the only epp perm allowed to not have a sid */
            return NULL;
        }

        ret = (EPPPermission_t*)DOFMem_Alloc( sizeof(EPPPermission_t));

        if( ret )
        {
            ret->base.type = type;
            ret->base.refCount = 1;
            ret->base.dataLength = len;
            ret->sid = temp;
        }
        else
        {
            DOFObjectID_Destroy(temp);
        }
    }

    return (DOFPermission)ret;
}

static DOFPermission unmarshal_tun_permission( OALStream stream )
{
    TunPermission_t* ret = NULL;
    uint16 len = 0;
    uint32 rdi;

    if( DOFERROR_OK == OALMarshal_StreamUncompress2( stream, &len ) &&
        DOFERROR_OK == OALMarshal_StreamUncompress4( stream, &rdi ) )
    {
        ret = (TunPermission_t*)DOFMem_Alloc( sizeof(TunPermission_t));
        if( ret )
        {
            ret->base.refCount = 1;
            ret->base.dataLength = len;
            ret->base.type = DOFPERMISSIONTYPE_TUNNEL_DOMAIN;
            ret->rdi = rdi;
        }
    }
    return (DOFPermission)ret;
}

static DOFPermission unmarshal_unknowntype( OALStream stream, DOFPermissionType type )
{
    uint16 len;

    if( type & 0x0001 ) /* odd */
    {
        if( DOFERROR_OK == OALMarshal_StreamUncompress2( stream, &len ) )
        {
            UnknownTypePermission_t* ret = (UnknownTypePermission_t*)DOFMem_Alloc(sizeof(UnknownTypePermission_t) + len );
            if( ret )
            {
                ret->data = (((uint8*)ret)+sizeof(UnknownTypePermission_t));
                if( OALStream_Read( stream, len, ret->data ) )
                {
                    ret->base.refCount = 1;
                    ret->base.type = type;
                    ret->base.dataLength = len;
                    return (DOFPermission)ret;
                }
                DOFMem_Free( ret );
            }
        }
    }
    else
    {
        DOFPermission ret = (DOFPermission)DOFMem_Alloc(sizeof(DOFPermission_t));

        if( ret )
        {
            ret->type = type;
            ret->dataLength = 0;
            ret->refCount = 1;
            return ret;
        }
    }

    return NULL;
}

DOFPermission DOFPermission_Unmarshal( OALStream stream, boolean unknownAllowed )
{
    uint16 type = 0;

    if( DOFERROR_OK == OALMarshal_StreamUncompress2( stream, &type ) )
    {
        switch( type )
        {
        case DOFPERMISSIONTYPE_BINDING:
            {
                return unmarshal_binding_permission( stream );
            }
            break;
        case DOFPERMISSIONTYPE_I_AM:
        case DOFPERMISSIONTYPE_ACT_AS:
            {
                return unmarshal_epp_permission(stream,type);
            }
            break;
        case DOFPERMISSIONTYPE_BINDING_EXPLICIT:
            {
                return unmarshal_binding_wo_permission( stream );
            }
            break;
        case DOFPERMISSIONTYPE_REQUESTOR:
            {
                return DOFPermissionRequestor_Create();
            }
            break;
        case DOFPERMISSIONTYPE_PROVIDER:
            {
                return DOFPermissionProvider_Create();
            }
        case DOFPERMISSIONTYPE_DEFINE:
            {
                uint16 len = 0;
                if( DOFERROR_OK == OALMarshal_StreamUncompress2( stream, &len )  )
                {
                    DOFInterfaceID iid = DOFInterfaceID_Unmarshal( stream );
                    DOFPermission ret = DOFPermissionDefine_Create( iid );
                    DOFInterfaceID_Destroy( iid );
                    return ret;
                }
            }
            break;
        case DOFPERMISSIONTYPE_TUNNEL_DOMAIN:
            {
                return unmarshal_tun_permission(stream);
            }
            break;
        default:
            if( unknownAllowed )
            {
                return unmarshal_unknowntype( stream, type );
            }
            break;
        }
    }
    return NULL;
}

static uint16 DOFPermission_GetActions( DOFPermission self )
{
    if( self->type == DOFPERMISSIONTYPE_BINDING || self->type == DOFPERMISSIONTYPE_BINDING_EXPLICIT )
        return (((DOFPermissionBuilder_t*)self)->flags & DOFPERMISSIONBINDINGACTION_ACTIONSMASK);
    return 0;
}

static DOFInterfaceID DOFPermission_GetInterfaceID( DOFPermission self )
{
    if( self->type == DOFPERMISSIONTYPE_DEFINE )
        return ((DefinePermission_t*)self)->iid;
    return NULL;
}

static boolean DOFPermission_HasInterfaceID( DOFPermission self, DOFInterfaceID iid )
{
    if( self->type == DOFPERMISSIONTYPE_BINDING || self->type == DOFPERMISSIONTYPE_BINDING_EXPLICIT )
    {
        if( OALAVLTree_GetNumItems(((DOFPermissionBuilder_t*)self)->interfaces) == 0 )
            return TRUE;
        return (OALAVLTree_Retrieve(((DOFPermissionBuilder_t*)self)->interfaces, iid)!=NULL);
    }
    return FALSE;
}

static boolean compare_AVPS(DOFObjectIDAttribute* B, DOFObjectIDAttribute* A)
{
    if((B->id & 0x7F) == (A->id & 0x7F)) /* make sure IDs Match */
    {
        if( (B->id & 0x80) || /* we are comparing against a wildcard any data matches */
            ( !( A->id & 0x80 ) && (B->dataSize == A->dataSize && (PCRMem_Compare(B->data, A->data, B->dataSize) == 0)))) /* not a wildcard so data has to be identical */
        {
            /* match */
            return TRUE;
        }
    }
    return FALSE;
}

struct _batc_t
{
    DOFPermissionBuilder_t* B;
    boolean success;
}_batc_t;
                                        
static boolean binding_IsAttributeInList(OALAVLTreeItem_t item, void *ctx)
{
    DOFObjectIDAttribute* Bavp = (DOFObjectIDAttribute*)OALAVLTree_Retrieve( ((struct _batc_t*)ctx)->B->avps, ((DOFObjectIDAttribute*)item) );

    /* item from A and Bavp is from B */
    if( !( Bavp && compare_AVPS( Bavp, ((DOFObjectIDAttribute*)item) ) ) )
    {
        ((struct _batc_t*)ctx)->success = FALSE;
        return TRUE;
    }

    return FALSE;
}

static boolean Binding_AllAttributesIn( DOFPermissionBuilder_t* B, DOFPermissionBuilder_t* A )
{
    struct _batc_t ctx;

    if( B->flags & DOFPERMISSIONBINDING_ALLAVPS )
        return TRUE;

    if( A->flags & DOFPERMISSIONBINDING_ALLAVPS )
        return FALSE;

    /* This compares no attribute A to no attribute B */
    if( (OALAVLTree_GetNumItems(A->avps) == 0) && ( OALAVLTree_GetNumItems(B->avps) == 0 ) )
        return TRUE;

    if( OALAVLTree_GetNumItems(A->avps) != OALAVLTree_GetNumItems(B->avps) )
        return FALSE;

    ctx.B = B;
    ctx.success = TRUE;
    OALAVLTree_Traverse(A->avps, binding_IsAttributeInList, (void*)&ctx);
    return ctx.success;
}

struct bndCmp
{
    boolean ret;
    DOFPermissionBuilder_t* B;
}bndCmp;

struct bpfowa_t
{
    boolean ret;
    DOFObjectID A;
    DOFPermissionBuilder_t* B;
}bpfowa_t;

static boolean findObjectID( OALAVLTreeItem_t item, void *arg )
{
    DOFObjectID B = (DOFObjectID)item;
    struct bpfowa_t* ctx = (struct bpfowa_t*)arg;

    /* need this compare because w/o wildcards still has avps but they have been added to avp list */
    if( DOFObjectID_CompareIgnoreAttributes(ctx->A, B)==0)
    {
        ctx->ret = TRUE;
        return TRUE;
    }
    return FALSE;
}

static boolean binding_IsObjectsInList( OALAVLTreeItem_t item, void *arg )
{
    DOFObjectID oid = (DOFObjectID)item;
    struct bndCmp* ctx = (struct bndCmp*)arg;
    struct bpfowa_t context;

    context.A = oid;
    context.B = ctx->B;
    context.ret = FALSE;

    OALAVLTree_Traverse(ctx->B->objectIDs, findObjectID, &context);

    if(!context.ret)
    {
        ctx->ret = FALSE;
        return TRUE; /* exit early. didn't find the oid */
    }
    return FALSE;
}

static boolean Binding_AllObjectsIn( DOFPermissionBuilder_t* B, DOFPermissionBuilder_t* A )
{
    struct bndCmp ctx;

    if(OALAVLTree_GetNumItems(B->objectIDs) == 0)
        return TRUE;

    if( OALAVLTree_GetNumItems(A->objectIDs) == 0 )
        return FALSE;

    ctx.B = B;
    ctx.ret = TRUE;

    OALAVLTree_Traverse(A->objectIDs, binding_IsObjectsInList, (void*)&ctx);

    return ctx.ret;
}

static boolean binding_IsInterfacesInList( OALAVLTreeItem_t item, void *arg )
{
    DOFInterfaceID iid = (DOFInterfaceID)item;
    struct bndCmp* ctx = (struct bndCmp*)arg;

    if(OALAVLTree_Retrieve(ctx->B->interfaces, iid) == NULL)
    {
        ctx->ret = FALSE;
        return TRUE; /* exit early error */
    }

    return FALSE;
}

static boolean Binding_AllInterfacesIn( DOFPermissionBuilder_t* B, DOFPermissionBuilder_t* A )
{
    struct bndCmp ctx;

    if(OALAVLTree_GetNumItems(B->interfaces)==0)
        return TRUE;

    if(OALAVLTree_GetNumItems(A->interfaces)==0)
        return FALSE;

    ctx.B = B;
    ctx.ret = TRUE;

    OALAVLTree_Traverse(A->interfaces, binding_IsInterfacesInList, (void*)&ctx);

    return ctx.ret;
}

static boolean Permission_IsActionAllowedBy( DOFPermissionBuilder_t* B, DOFPermissionBuilder_t* A )
{
    if( ((B->flags & A->flags) & DOFPERMISSIONBINDINGACTION_ACTIONSMASK) == (A->flags & DOFPERMISSIONBINDINGACTION_ACTIONSMASK) )
        return TRUE;
    return FALSE;
}

/* check if A is allowed by B */
static boolean Binding_IsAllowedBy( DOFPermissionBuilder_t* B, DOFPermissionBuilder_t* A, boolean alreadyNormalized )
{
    if( alreadyNormalized || A->base.type == DOFPERMISSIONTYPE_BINDING_EXPLICIT )
    {
        if( Permission_IsActionAllowedBy(B, A) && Binding_AllObjectsIn(B, A) && Binding_AllInterfacesIn(B, A) && Binding_AllAttributesIn(B, A) )
            return TRUE;
    }
    else
    {
        DOFPermissionSet normalized = DOFPermission_CreateNormalized( (DOFPermission)A );

        if( normalized )
        {
            uint16 i;
            boolean allowedBy = TRUE;

            for( i = 0; i < normalized->permsCnt; ++i )
            {
                if( !( Permission_IsActionAllowedBy( B, (DOFPermissionBuilder_t*)normalized->permission_list[i].perm ) &&
                      Binding_AllObjectsIn( B, (DOFPermissionBuilder_t*)normalized->permission_list[i].perm ) &&
                      Binding_AllInterfacesIn( B, (DOFPermissionBuilder_t*)normalized->permission_list[i].perm ) &&
                      Binding_AllAttributesIn( B, (DOFPermissionBuilder_t*)normalized->permission_list[i].perm ) ) )
                {
                    allowedBy = FALSE;
                    break;
                }
            }
            DOFPermissionSet_Destroy( normalized );
            return allowedBy;
        }
    }
    return FALSE;
}

static boolean Defined_IsAllowedBy( DOFPermission B, DefinePermission_t* A )
{
    DOFInterfaceID iid = DOFPermission_GetInterfaceID( B );
    if( ( iid && DOFInterfaceID_IsWildcard( iid )) ||
        DOFInterfaceID_IsEqual( A->iid, iid ) ||
        ((DOFPermission_GetActions( B ) & DOFPERMISSIONBINDINGACTION_PROVIDE) &&
          DOFPermission_HasInterfaceID( B, iid ) ) )
        return TRUE;
    return FALSE;
}

static boolean Requestor_IsAllowedBy( DOFPermission B, DOFPermission A )
{
    if( B->type == DOFPERMISSIONTYPE_REQUESTOR ||
        (DOFPermission_GetActions( B ) & ( DOFPERMISSIONBINDINGACTION_READ | DOFPERMISSIONBINDINGACTION_WRITE | DOFPERMISSIONBINDINGACTION_EXECUTE | DOFPERMISSIONBINDINGACTION_SESSION ) ))
        return TRUE;
    return FALSE;
}

static boolean Provider_IsAllowedBy( DOFPermission B, DOFPermission A )
{
    if( B->type == DOFPERMISSIONTYPE_PROVIDER ||
        B->type == DOFPERMISSIONTYPE_DEFINE ||
        (DOFPermission_GetActions( B ) & DOFPERMISSIONBINDINGACTION_PROVIDE) )
        return TRUE;
    return FALSE;
}

static boolean I_AM_IsAllowedBy( DOFPermission B, DOFPermission A )
{
    if( B->type == DOFPERMISSIONTYPE_I_AM )
    {
        if( (DOFObjectID_IsEqual(((EPPPermission_t*)A)->sid,((EPPPermission_t*)B)->sid)) ||
            (DOFObjectID_IsASAssignedEqual(((EPPPermission_t*)A)->sid, ((EPPPermission_t*)B)->sid)))
            return TRUE;
    }
    return FALSE;
}

static boolean ActAs_IsAllowedBy( DOFPermission B, DOFPermission A )
{
    if( B->type == DOFPERMISSIONTYPE_ACT_AS )
    {
        if( (((EPPPermission_t*)B)->sid) == NULL ||
            ( DOFObjectID_IsEqual( ((EPPPermission_t*)A)->sid,((EPPPermission_t*)B)->sid ) ) ||
              DOFObjectID_IsASAssignedEqual( ((EPPPermission_t*)A)->sid, ((EPPPermission_t*)B)->sid ) )
            return TRUE;
    }
    else if( B->type == DOFPERMISSIONTYPE_I_AM )
    {
        if((((EPPPermission_t*)A)->sid) == NULL)
            return FALSE;
        return I_AM_IsAllowedBy(B, A);
    }
    return FALSE;
}

static boolean Tun_IsAllowedBy( DOFPermission B, DOFPermission A )
{
    if( B->type == DOFPERMISSIONTYPE_TUNNEL_DOMAIN )
    {
        if( ((TunPermission_t*)B)->rdi == ((TunPermission_t*)A)->rdi ||
            ((TunPermission_t*)B)->rdi == DOFPERMISSIONTUNNELDOMAIN_ALL )
            return TRUE;
    }
    return FALSE;
}

boolean DOFPermission_IsAllowedBy( DOFPermission B, DOFPermission A, boolean compareExplicit, boolean alreadyNormalized )
{
    if( B == A )
        return TRUE;

    if( compareExplicit && DOFPermission_IsExplicit( A ) && A->type != B->type )
        return FALSE;

    switch( A->type )
    {
    case DOFPERMISSIONTYPE_BINDING:
    case DOFPERMISSIONTYPE_BINDING_EXPLICIT:
        if( B->type == DOFPERMISSIONTYPE_BINDING || B->type == DOFPERMISSIONTYPE_BINDING_EXPLICIT )
            return Binding_IsAllowedBy((DOFPermissionBuilder_t*)B, (DOFPermissionBuilder_t*)A, alreadyNormalized );
        break;
    case DOFPERMISSIONTYPE_I_AM:
        return I_AM_IsAllowedBy(B, A);
        break;
    case DOFPERMISSIONTYPE_ACT_AS:
        return ActAs_IsAllowedBy(B, A);
        break;
    case DOFPERMISSIONTYPE_REQUESTOR:
        return Requestor_IsAllowedBy(B, A);
        break;
    case DOFPERMISSIONTYPE_PROVIDER:
        return Provider_IsAllowedBy(B, A);
        break;
    case DOFPERMISSIONTYPE_DEFINE:
        return Defined_IsAllowedBy(B, (DefinePermission_t*)A);
        break;
    case DOFPERMISSIONTYPE_TUNNEL_DOMAIN:
        return Tun_IsAllowedBy(B, A);
        break;
    default:
        break;
    }
    return FALSE;
}

static uint8 DOFPermissionBinding_ActionCount( DOFPermission selfx )
{
    uint8 ret = 0;
    DOFPermissionBuilder_t* self = (DOFPermissionBuilder_t*)selfx;

    if( ( self->flags & DOFPERMISSIONBINDINGACTION_ACTIONSMASK ) == DOFPERMISSIONBINDINGACTION_EXECUTE )
        ++ret;
    if( ( self->flags & DOFPERMISSIONBINDINGACTION_ACTIONSMASK ) == DOFPERMISSIONBINDINGACTION_WRITE )
        ++ret;
    if( ( self->flags & DOFPERMISSIONBINDINGACTION_ACTIONSMASK ) == DOFPERMISSIONBINDINGACTION_READ )
        ++ret;
    if( ( self->flags & DOFPERMISSIONBINDINGACTION_ACTIONSMASK ) == DOFPERMISSIONBINDINGACTION_PROVIDE )
        ++ret;
    if( ( self->flags & DOFPERMISSIONBINDINGACTION_ACTIONSMASK ) == DOFPERMISSIONBINDINGACTION_SESSION )
        ++ret;
    return ret;
}

struct intersect_t
{
    DOFPermissionBuilder_t* self;
    DOFPermissionBuilder_t* other;
    boolean success;
}intersect_t;

static boolean IntersectAttributeTraverse( OALAVLTreeItem_t item, void *arg )
{
    struct intersect_t* ctx = (struct intersect_t*)arg;

    if( OALAVLTree_GetNumItems( ctx->other->avps ) == 0 || OALAVLTree_Retrieve( ctx->other->avps, (DOFObjectIDAttribute*)item ) != NULL )
    {
        ctx->success = TRUE;
        return TRUE;
    }
    return FALSE;
}

static boolean IntersectInterfaceIDTraverse( OALAVLTreeItem_t item, void *arg )
{
    struct intersect_t* ctx = (struct intersect_t*)arg;

    if( !item || ( OALAVLTree_GetNumItems( ctx->other->interfaces ) == 0 ) || OALAVLTree_Retrieve( ctx->other->interfaces, (DOFInterfaceID)item ) != NULL )
    {
        if( ctx->self->flags & DOFPERMISSIONBINDING_ALLAVPS || ctx->other->flags & DOFPERMISSIONBINDING_ALLAVPS )
        {
            ctx->success = TRUE;
        }
        else
        {
            if( OALAVLTree_GetNumItems( ctx->self->avps ) )
                OALAVLTree_Traverse( ctx->self->avps, IntersectAttributeTraverse, arg );
            else
                IntersectAttributeTraverse( NULL, arg );
        }
        return TRUE;
    }
    return FALSE;
}

static boolean IntersectObjectIDTraverse( OALAVLTreeItem_t item, void *arg )
{
    struct intersect_t* ctx = (struct intersect_t*)arg;

    if( !item || ( OALAVLTree_GetNumItems( ctx->other->objectIDs ) == 0 ) || OALAVLTree_Retrieve( ctx->other->objectIDs, (DOFObjectID)item ) != NULL )
    {
        if( OALAVLTree_GetNumItems( ctx->self->interfaces ) )
            OALAVLTree_Traverse( ctx->self->interfaces, IntersectInterfaceIDTraverse, arg );
        else
            IntersectInterfaceIDTraverse( NULL, arg );
        return TRUE;
    }
    return FALSE;
}

boolean DOFPermission_Intersects( DOFPermission self, DOFPermission other )
{
    if( ( self->type == DOFPERMISSIONTYPE_BINDING || self->type == DOFPERMISSIONTYPE_BINDING_EXPLICIT ) &&
        ( other->type == DOFPERMISSIONTYPE_BINDING || other->type == DOFPERMISSIONTYPE_BINDING_EXPLICIT ) )
    {
        struct intersect_t ctx;

        ctx.self = (DOFPermissionBuilder_t*)self;
        ctx.other = (DOFPermissionBuilder_t*)other;
        ctx.success = FALSE;

        if( ( ctx.self->flags & DOFPERMISSIONBINDINGACTION_ACTIONSMASK ) & ( ctx.other->flags & DOFPERMISSIONBINDINGACTION_ACTIONSMASK ) )
        {
            if( OALAVLTree_GetNumItems( ctx.self->objectIDs ) )
                OALAVLTree_Traverse( ctx.self->objectIDs, IntersectObjectIDTraverse, &ctx );
            else
                IntersectObjectIDTraverse( NULL, &ctx );

        }
        return ctx.success;
    }
    return FALSE;
}

static boolean DOFPermissionBinding_IsExpandable( DOFPermission self )
{
    uint32 count =  ( ( OALAVLTree_GetNumItems( ((DOFPermissionBuilder_t *)self)->objectIDs ) > 0 )? OALAVLTree_GetNumItems( ((DOFPermissionBuilder_t *)self)->objectIDs ) : 1 ) * ( ( OALAVLTree_GetNumItems( ((DOFPermissionBuilder_t *)self)->interfaces ) > 0 )? OALAVLTree_GetNumItems( ((DOFPermissionBuilder_t *)self)->interfaces ) : 1 ) * DOFPermissionBinding_ActionCount( self );
    if( count > 1)
        return TRUE;
    return FALSE;
}

boolean DOFPermission_IsExpandable( DOFPermission self )
{
    if( self->type == DOFPERMISSIONTYPE_BINDING &&
        DOFPermissionBinding_IsExpandable( self ) )
        return TRUE;
    return FALSE;
}

struct normCtx_t
{
    DOFPermissionBuilder_t*     orig;
    DOFPermissionSetBuilder     normalized;
    DOFPermissionBindingBuilder builder;
    boolean                     error;
} normCtx_t;

static boolean NormalizeActions( DOFPermissionBindingAction action, DOFPermissionBindingBuilder builder, DOFPermissionSetBuilder normalizedset )
{
    if( action & DOFPERMISSIONBINDINGACTION_EXECUTE )
    {
        DOFPermission perm;

        DOFPermissionBindingBuilder_SetActions( builder, DOFPERMISSIONBINDINGACTION_EXECUTE );
        perm = DOFPermissionBindingBuilder_Build( builder );

        if( !perm )
            return FALSE;

        DOFPermissionSetBuilder_AddPermission( normalizedset, perm );
        DOFPermission_Destroy( perm );
    }

    if( action & DOFPERMISSIONBINDINGACTION_WRITE )
    {
        DOFPermission perm;

        DOFPermissionBindingBuilder_SetActions( builder, DOFPERMISSIONBINDINGACTION_WRITE );
        perm = DOFPermissionBindingBuilder_Build( builder );

        if( !perm )
            return FALSE;

        DOFPermissionSetBuilder_AddPermission( normalizedset, perm );
        DOFPermission_Destroy( perm );
    }

    if( action & DOFPERMISSIONBINDINGACTION_READ )
    {
        DOFPermission perm;

        DOFPermissionBindingBuilder_SetActions( builder, DOFPERMISSIONBINDINGACTION_READ );
        perm = DOFPermissionBindingBuilder_Build( builder );

        if( !perm )
            return FALSE;

        DOFPermissionSetBuilder_AddPermission( normalizedset, perm );
        DOFPermission_Destroy( perm );
    }

    if( action & DOFPERMISSIONBINDINGACTION_PROVIDE )
    {
        DOFPermission perm;

        DOFPermissionBindingBuilder_SetActions( builder, DOFPERMISSIONBINDINGACTION_PROVIDE );
        perm = DOFPermissionBindingBuilder_Build( builder );

        if( !perm )
            return FALSE;

        DOFPermissionSetBuilder_AddPermission( normalizedset, perm );
        DOFPermission_Destroy( perm );
    }

    if( action & DOFPERMISSIONBINDINGACTION_SESSION )
    {
        DOFPermission perm;

        DOFPermissionBindingBuilder_SetActions( builder, DOFPERMISSIONBINDINGACTION_SESSION );
        perm = DOFPermissionBindingBuilder_Build( builder );

        if( !perm )
            return FALSE;

        DOFPermissionSetBuilder_AddPermission( normalizedset, perm );
        DOFPermission_Destroy( perm );
    }
    return TRUE;
}

static boolean NormalizeInterfaceIDTraverse( OALAVLTreeItem_t item, void *arg )
{
    DOFInterfaceID iid = (DOFInterfaceID)item;
    struct normCtx_t* ctx = (struct normCtx_t*)arg;

    if( iid )
    {
        if( !DOFPermissionBindingBuilder_AddInterfaceID( ctx->builder, iid ) )
        {
            ctx->error = TRUE;
            return TRUE;
        }
    }

    if( !NormalizeActions( ctx->orig->flags & DOFPERMISSIONBINDINGACTION_ACTIONSMASK, ctx->builder, ctx->normalized ) )
    {
        ctx->error = TRUE;
        return TRUE;
    }

    DOFPermissionBindingBuilder_AddInterfaceID( ctx->builder, NULL ); /* resets to just oid */

    return FALSE;
}

static boolean NormalizeObjectIDTraverse( OALAVLTreeItem_t item, void *arg )
{
    DOFObjectID oid = (DOFObjectID)item;
    struct normCtx_t* ctx = (struct normCtx_t*)arg;

    if( oid )
    {
        if( !DOFPermissionBindingBuilder_AddObjectID( ctx->builder, oid ) )
        {
            ctx->error = TRUE;
            return TRUE;
        }
    }

    if( OALAVLTree_GetNumItems( ctx->orig->interfaces ) )
        OALAVLTree_Traverse( ctx->orig->interfaces, NormalizeInterfaceIDTraverse, arg );
    else
        NormalizeInterfaceIDTraverse( NULL, arg );

    if( ctx->error )
        return TRUE;

    DOFPermissionBindingBuilder_AddObjectID( ctx->builder, NULL );

    return FALSE;
}

static boolean NormalizeAttributeTraverse( OALAVLTreeItem_t item, void *arg )
{
    DOFObjectIDAttribute* attr = (DOFObjectIDAttribute*)item;
    struct normCtx_t* ctx = (struct normCtx_t*)arg;

    if( !DOFPermissionBindingBuilder_Internal_AddRequiredAttribute( ctx->builder,(attr->id & 0x80), (attr->id & 0x7F), attr->dataSize, attr->data ) )
    {
        ctx->error = TRUE;
        return TRUE;
    }

    return FALSE;
}

static boolean DOFPermission_CreateNormalized_Helper_SetAttibutesFromPermission(struct normCtx_t* ctx)
{
    if( OALAVLTree_GetNumItems( ctx->orig->avps ) )
    {
        OALAVLTree_Traverse( ctx->orig->avps, NormalizeAttributeTraverse, ctx );
    }
    else
    {
        if( !DOFPermissionBindingBuilder_SetAllAttributesAllowed( ctx->builder, ( ctx->orig->flags & DOFPERMISSIONBINDING_ALLAVPS ) ) )
            ctx->error = TRUE;
    }

    return !ctx->error;
}

DOFPermissionSet DOFPermission_CreateNormalized( DOFPermission selfx )
{
    /* DOFPERMISSIONTYPE_BINDING is the only type that needs to be normalized */
    if( selfx->type == DOFPERMISSIONTYPE_BINDING &&
        DOFPermission_IsExpandable( selfx ) )
    {
        DOFPermissionBuilder_t* self = (DOFPermissionBuilder_t*)selfx;
        /* Figure out how many permissions this expands too. */
        uint32 count =  ( ( OALAVLTree_GetNumItems( self->objectIDs ) > 0 )? OALAVLTree_GetNumItems( self->objectIDs ) : 1 ) *
                        ( ( OALAVLTree_GetNumItems( self->interfaces ) > 0 )? OALAVLTree_GetNumItems( self->interfaces ) : 1 ) *
                        DOFPermissionBinding_ActionCount( selfx );

        DOFPermissionSetBuilder ret = (DOFPermissionSetBuilder)DOFPermissionSet_CreateInternal( count, TRUE );
        if( ret )
        {
            struct normCtx_t ctx;

            ctx.orig = self;
            ctx.normalized = ret;
            ctx.error = FALSE;

            ctx.builder = DOFPermissionBindingBuilder_Internal_Create();
            if( ctx.builder )
            {
                if( DOFPermission_CreateNormalized_Helper_SetAttibutesFromPermission( &ctx ) )
                {
                    if( !ctx.error )
                    {
                        if( OALAVLTree_GetNumItems( self->objectIDs ) )
                            OALAVLTree_Traverse( self->objectIDs, NormalizeObjectIDTraverse, &ctx );
                        else
                            NormalizeObjectIDTraverse( NULL, &ctx );
                    }
                }
                DOFPermissionBindingBuilder_Destroy( ctx.builder );
            }
            else
                ctx.error = TRUE;

            if( ctx.error )
                DOFPermissionSetBuilder_Destroy( ctx.normalized );
            else
                return DOFPermissionSetBuilder_BuildAndDestroy( ctx.normalized );
        }
    }
    else
    {
        DOFPermissionSetBuilder ret = (DOFPermissionSetBuilder)DOFPermissionSet_CreateInternal( 1, TRUE );
        /* ret is checked in these next calls so there is no need to check here */
        DOFPermissionSetBuilder_AddPermission( ret, selfx ); 
        return DOFPermissionSetBuilder_BuildAndDestroy( ret );
    }
    return NULL;
}

boolean DOFPermission_IsExplicit( DOFPermission self )
{
    return ( self->type == DOFPERMISSIONTYPE_REQUESTOR ||
             self->type == DOFPERMISSIONTYPE_PROVIDER );
}

static boolean DOFPermissionBinding_IsTerminal( DOFPermission self )
{
    if( OALAVLTree_GetNumItems(((DOFPermissionBuilder_t*)self)->objectIDs) == 1 &&
        OALAVLTree_GetNumItems(((DOFPermissionBuilder_t*)self)->interfaces) == 1 &&
       (!(((DOFPermissionBuilder_t*)self)->flags & DOFPERMISSIONBINDING_ALLAVPS)) &&
       ( (((DOFPermissionBuilder_t*)self)->flags & DOFPERMISSIONBINDINGACTION_ACTIONSMASK) == DOFPERMISSIONBINDINGACTION_EXECUTE ||
         (((DOFPermissionBuilder_t*)self)->flags & DOFPERMISSIONBINDINGACTION_ACTIONSMASK) == DOFPERMISSIONBINDINGACTION_WRITE ||
         (((DOFPermissionBuilder_t*)self)->flags & DOFPERMISSIONBINDINGACTION_ACTIONSMASK) == DOFPERMISSIONBINDINGACTION_READ ||
         (((DOFPermissionBuilder_t*)self)->flags & DOFPERMISSIONBINDINGACTION_ACTIONSMASK) == DOFPERMISSIONBINDINGACTION_PROVIDE ||
         (((DOFPermissionBuilder_t*)self)->flags & DOFPERMISSIONBINDINGACTION_ACTIONSMASK) == DOFPERMISSIONBINDINGACTION_SESSION ) )
    {
        return TRUE;
    }
    return FALSE;
}

boolean DOFPermission_IsTerminal(  DOFPermission self )
{
    switch( self->type )
    {
    case DOFPERMISSIONTYPE_BINDING:
    case DOFPERMISSIONTYPE_BINDING_EXPLICIT:
        return DOFPermissionBinding_IsTerminal( self );
        break;
    case DOFPERMISSIONTYPE_DEFINE:
        return ((DefinePermission_t *)self)->iid ? DOFInterfaceID_IsWildcard(((DefinePermission_t *)self)->iid) : FALSE;
        break;
    case DOFPERMISSIONTYPE_I_AM:
    case DOFPERMISSIONTYPE_ACT_AS:
    case DOFPERMISSIONTYPE_REQUESTOR:
    case DOFPERMISSIONTYPE_PROVIDER:
    case DOFPERMISSIONTYPE_TUNNEL_DOMAIN:
        return TRUE;
        break;
    default:
        break;
    }
    return FALSE;
}

static uint32 binding_Complement_cnt( DOFPermission selfx )
{
    DOFPermissionBuilder_t* self = (DOFPermissionBuilder_t*)selfx;
    uint32 ret = 0;

    if( self->flags & DOFPERMISSIONBINDINGACTION_PROVIDE )
    {
        ret += 4;
    }
    if( self->flags & (DOFPERMISSIONBINDINGACTION_READ | DOFPERMISSIONBINDINGACTION_WRITE | DOFPERMISSIONBINDINGACTION_SESSION | DOFPERMISSIONBINDINGACTION_EXECUTE) )
    {
        ++ret;
    }
    return ret;
}

static boolean binding_Complement( DOFPermission selfx, DOFPermissionSetBuilder set )
{
    DOFPermissionBuilder_t* self = (DOFPermissionBuilder_t*)selfx;
    DOFPermissionBuilder_t* prm;

    if( self->flags & DOFPERMISSIONBINDINGACTION_PROVIDE )
    {
        prm = (DOFPermissionBuilder_t*)DOFPermissionBindingBuilder_Build( (DOFPermissionBindingBuilder)selfx );
        if( !prm )
            goto FAIL;

        prm->flags = (prm->flags&DOFPERMISSIONBINDINGACTION_FLAGSMASK)+DOFPERMISSIONBINDINGACTION_READ;
        if( !DOFPermissionSetBuilder_AddPermission(set,(DOFPermission)prm) )
            goto FAIL_WITHPERM;
        DOFPermission_Destroy((DOFPermission )prm);

        prm = (DOFPermissionBuilder_t*)DOFPermissionBindingBuilder_Build( (DOFPermissionBindingBuilder)selfx );
        if( !prm )
            goto FAIL;
        prm->flags = (prm->flags&DOFPERMISSIONBINDINGACTION_FLAGSMASK)+DOFPERMISSIONBINDINGACTION_WRITE;
        if( !DOFPermissionSetBuilder_AddPermission(set,(DOFPermission)prm) )
            goto FAIL_WITHPERM;
        DOFPermission_Destroy((DOFPermission )prm);

        prm = (DOFPermissionBuilder_t*)DOFPermissionBindingBuilder_Build( (DOFPermissionBindingBuilder)selfx );
        if( !prm )
            goto FAIL;
        prm->flags = (prm->flags&DOFPERMISSIONBINDINGACTION_FLAGSMASK)+DOFPERMISSIONBINDINGACTION_EXECUTE;
        if( !DOFPermissionSetBuilder_AddPermission(set,(DOFPermission)prm) )
            goto FAIL_WITHPERM;
        DOFPermission_Destroy((DOFPermission )prm);

        prm = (DOFPermissionBuilder_t*)DOFPermissionBindingBuilder_Build( (DOFPermissionBindingBuilder)selfx );
        if( !prm )
            goto FAIL;
        prm->flags = (prm->flags&DOFPERMISSIONBINDINGACTION_FLAGSMASK)+DOFPERMISSIONBINDINGACTION_SESSION;
        if( !DOFPermissionSetBuilder_AddPermission(set,(DOFPermission)prm) )
            goto FAIL_WITHPERM;
        DOFPermission_Destroy((DOFPermission )prm);
    }

    if( self->flags & (DOFPERMISSIONBINDINGACTION_READ | DOFPERMISSIONBINDINGACTION_WRITE | DOFPERMISSIONBINDINGACTION_SESSION | DOFPERMISSIONBINDINGACTION_EXECUTE) )
    {
        prm = (DOFPermissionBuilder_t*)DOFPermissionBindingBuilder_Build( (DOFPermissionBindingBuilder)selfx );
        if( !prm )
            goto FAIL;
        prm->flags = (prm->flags&DOFPERMISSIONBINDINGACTION_FLAGSMASK)+DOFPERMISSIONBINDINGACTION_PROVIDE;
        if( !DOFPermissionSetBuilder_AddPermission(set,(DOFPermission)prm) )
            goto FAIL_WITHPERM;
        DOFPermission_Destroy((DOFPermission )prm);
    }
    return TRUE;

FAIL_WITHPERM:
    DOFPermission_Destroy((DOFPermission )prm);
FAIL:
    return FALSE;

}

uint32 DOFPermission_ComplementCount( DOFPermission self )
{
    uint32 ret = 0;

    if( self->type == DOFPERMISSIONTYPE_BINDING || self->type == DOFPERMISSIONTYPE_BINDING_EXPLICIT )
        ret = binding_Complement_cnt( self );
    else if( self->type == DOFPERMISSIONTYPE_REQUESTOR ||
             self->type == DOFPERMISSIONTYPE_PROVIDER ||
             self->type == DOFPERMISSIONTYPE_DEFINE ||
             self->type == DOFPERMISSIONTYPE_TUNNEL_DOMAIN)
        ++ret;

    return ret;
}

DOFPermissionSet DOFPermission_CreateComplement( DOFPermission self )
{
    uint32 cnt = DOFPermission_ComplementCount( self );
    DOFPermissionSetBuilder ret = (DOFPermissionSetBuilder)DOFPermissionSet_CreateInternal(cnt, TRUE);

    if( ret )
    {
        switch( self->type )
        {
        case DOFPERMISSIONTYPE_BINDING:
        case DOFPERMISSIONTYPE_BINDING_EXPLICIT:
            {
                if( !binding_Complement( self, ret ) )
                {
                    DOFPermissionSetBuilder_Destroy( ret );
                    return NULL;
                }
            }
            break;
        case DOFPERMISSIONTYPE_I_AM:
        case DOFPERMISSIONTYPE_ACT_AS:
            /* No Complement */
            break;
        case DOFPERMISSIONTYPE_REQUESTOR:
            {
                DOFPermission per = DOFPermissionProvider_Create();
                if(per != NULL)
                {
                    if( !DOFPermissionSetBuilder_AddPermission( ret, per ) )
                    {
                        DOFPermission_Destroy( per );
                        DOFPermissionSetBuilder_Destroy( ret );
                        return NULL;
                    }
                    DOFPermission_Destroy( per );
                }

            }
            break;
        case DOFPERMISSIONTYPE_PROVIDER:
        case DOFPERMISSIONTYPE_DEFINE:
            {
                DOFPermission per = DOFPermissionRequestor_Create();
                if(per != NULL)
                {
                    if( !DOFPermissionSetBuilder_AddPermission( ret, per ) )
                    {
                        DOFPermission_Destroy( per );
                        DOFPermissionSetBuilder_Destroy( ret );
                        return NULL;
                    }
                    DOFPermission_Destroy( per );
                }
            }
            break;
        case DOFPERMISSIONTYPE_TUNNEL_DOMAIN:
            /* Add the permission as its complement */
            if( !DOFPermissionSetBuilder_AddPermission( ret, self ) )
            {
                DOFPermissionSetBuilder_Destroy( ret );
                return NULL;
            }
            break; 
        default:
            break;
        }
        return DOFPermissionSetBuilder_BuildAndDestroy( ret );
    }
    return NULL;
}

static boolean DOFPermissionExtend_AddBindingActionPerms(DOFPermissionSetBuilder builder,
                                                         DOFObjectID             objectId,
                                                         DOFInterfaceID          interfaceId)
{
    static DOFPermissionBindingAction const requestorActions[5] =
    {
        DOFPERMISSIONBINDINGACTION_EXECUTE,     DOFPERMISSIONBINDINGACTION_WRITE,
        DOFPERMISSIONBINDINGACTION_READ,        DOFPERMISSIONBINDINGACTION_SESSION, 0
    };

    uint8 i = 0;

    while(requestorActions[i] != 0)
    {
        DOFPermission perm = DOFPermissionBinding_Create(requestorActions[i], objectId, interfaceId);

        if(perm != NULL)
        {
            boolean keepGoing = DOFPermissionSetBuilder_AddPermission(builder, perm);

            DOFPermission_Destroy(perm);
            if(!keepGoing)
            {
                return FALSE;
            }
        }
        ++i;
    }
    return TRUE;
}

static boolean DOFPermissionExtend_AddBindingActionProviderPerms(DOFPermissionSetBuilder builder,
                                                                 DOFObjectID             objectId,
                                                                 DOFInterfaceID          interfaceId)
{
    static DOFPermissionBindingAction const requestorActions[5] =
    {
        DOFPERMISSIONBINDINGACTION_PROVIDE, 0
    };

    uint8 i = 0;

    while(requestorActions[i] != 0)
    {
        DOFPermission perm = DOFPermissionBinding_Create(requestorActions[i], objectId, interfaceId);

        if(perm != NULL)
        {
            boolean keepGoing = DOFPermissionSetBuilder_AddPermission(builder, perm);

            DOFPermission_Destroy(perm);
            if(!keepGoing)
            {
                return FALSE;
            }
        }
        ++i;
    }
    return TRUE;
}

typedef boolean (*BuilderCallback)(DOFPermissionSetBuilder builer,
                                   DOFObjectID             objectId,
                                   DOFInterfaceID          interfaceId);

typedef struct ExtendSetBindingContext_t
{
    DOFPermissionSetBuilder builder;
    DOFObjectID objectId;
    DOFPermissionBuilder_t * realPerm;
    BuilderCallback builderCallback;
    boolean keepGoing;
} ExtendSetBindingContext_t;

static boolean extendBinding_interfaceIdTraversal(OALAVLTreeItem_t   item,
                                                  void             * arg)
{
    ExtendSetBindingContext_t * context = (ExtendSetBindingContext_t*) arg;

    context->keepGoing = context->builderCallback(context->builder, context->objectId, (DOFInterfaceID) item);
    return !(context->keepGoing);
}

static boolean extendBinding_objectIdTraversal(OALAVLTreeItem_t   item,
                                               void             * arg)
{
    ExtendSetBindingContext_t * extendContext = (ExtendSetBindingContext_t*) arg;

    extendContext->objectId = (DOFObjectID) item;
    OALAVLTree_Traverse(extendContext->realPerm->interfaces, extendBinding_interfaceIdTraversal, extendContext);
    return !(extendContext->keepGoing);
}

static void DOFPermissionExtend_CreatePartialBindingPerms(ExtendSetBindingContext_t * extendContext,
                                                          BuilderCallback             builderCallback)
{
    extendContext->builderCallback = builderCallback;
    OALAVLTree_Traverse(extendContext->realPerm->objectIDs, extendBinding_objectIdTraversal, extendContext);
    if(extendContext->keepGoing)
    {
        extendContext->keepGoing = extendContext->builderCallback(extendContext->builder, extendContext->objectId, NULL);
        if(extendContext->keepGoing)
        {
            extendContext->objectId = NULL;
            OALAVLTree_Traverse(extendContext->realPerm->interfaces, extendBinding_interfaceIdTraversal, extendContext);
            if(extendContext->keepGoing)
            {
                extendContext->keepGoing = extendContext->builderCallback(extendContext->builder, NULL, NULL);
            }
        }
    }
}

/*  This function should create a permission set containing the following:
 *   - if the permission contains the PROVIDE action, PROVIDE on the following bindings:
 *      1) OID:IID
 *      2) *:IID
 *      3) OID:*
 *      4) *:*
 *   - if the permision cotains ANY non-provide action, READ, WRITE, EXECUTE, and SESSION on the following bindings
 *      1) OID:IID
 *      2) *:IID
 *      3) OID:*
 *      4) *:*
 *
 *  The permissions should include EVERY possible combination OIDs and IIDs, but because the only permissions that
 *  are passed in are created internally and they have exactly one IID and one OID, we don't cover those cases (yet).
 */
static DOFPermissionSet DOFPermissionExtend_CreateBindingPerms(DOFPermission self)
{
    DOFPermissionSetBuilder builder;

    PCR_Assert(self != NULL);
    PCR_Assert((self->type == DOFPERMISSIONTYPE_BINDING) ||
               (self->type == DOFPERMISSIONTYPE_BINDING_EXPLICIT));
    builder = DOFPermissionSetBuilder_Create();
    if(builder != NULL)
    {
        DOFPermissionBuilder_t * realPerm = (DOFPermissionBuilder_t*) self;
        ExtendSetBindingContext_t extendContext;

        extendContext.builder = builder;
        extendContext.realPerm = realPerm;
        extendContext.keepGoing = FALSE;
        /*  Because every permission that'll ever be passed in is created by _createNeededPerm, we can safely assume it has exactly one action, one oid, and one iid.
         */
        if(((realPerm->flags & DOFPERMISSIONBINDINGACTION_ACTIONSMASK) & ~DOFPERMISSIONBINDINGACTION_PROVIDE) != 0)
        {
            DOFPermissionExtend_CreatePartialBindingPerms(&extendContext, DOFPermissionExtend_AddBindingActionPerms);
        }
        if(((realPerm->flags & DOFPERMISSIONBINDINGACTION_ACTIONSMASK) & DOFPERMISSIONBINDINGACTION_PROVIDE) == DOFPERMISSIONBINDINGACTION_PROVIDE)
        {
            DOFPermissionExtend_CreatePartialBindingPerms(&extendContext, DOFPermissionExtend_AddBindingActionProviderPerms);
        }
        if(extendContext.keepGoing)
        {
            return DOFPermissionSetBuilder_BuildAndDestroy(builder);
        }
        DOFPermissionSetBuilder_Destroy(builder);
    }
    return NULL;
}

DOFPermissionSet DOFPermission_CreateExtendSet(DOFPermission self)
{
    PCR_Assert(self != NULL);
    switch(self->type)
    {
    case DOFPERMISSIONTYPE_BINDING:
    case DOFPERMISSIONTYPE_BINDING_EXPLICIT:
        return DOFPermissionExtend_CreateBindingPerms(self);
        break;

    default:
        {
            DOFPermissionSetBuilder builder = DOFPermissionSetBuilder_Create();

            if(builder != NULL)
            {
                if(DOFPermissionSetBuilder_AddPermission(builder, self))
                {
                    return DOFPermissionSetBuilder_BuildAndDestroy(builder);
                }
                DOFPermissionSetBuilder_Destroy(builder);
            }
        }
    }
    return NULL;
}

struct isEqualContext_t
{
    OALAVLTree  otherTree;
    void*       toCompare;
    boolean     Is_Match;
}isEqualContext_t;

static boolean Binding_IsEqual_OID_Inner_Traversal_fn(OALAVLTreeItem_t item, void *arg)
{
    DOFObjectID oid = (DOFObjectID)item;
    struct isEqualContext_t* ctx = (struct isEqualContext_t*)arg;

    if( DOFObjectID_IsEqual(oid, ((DOFObjectID)ctx->toCompare)) )
    {
        ctx->Is_Match = TRUE;
        return TRUE; /* end traversal */
    }
    return FALSE;
}

static boolean Binding_IsEqual_OID_Outer_Traversal_fn(OALAVLTreeItem_t item, void *arg)
{
    DOFObjectID oid = (DOFObjectID)item;
    struct isEqualContext_t* ctx = (struct isEqualContext_t*)arg;

    ctx->Is_Match = FALSE;
    ctx->toCompare = oid;

    OALAVLTree_Traverse( ctx->otherTree, Binding_IsEqual_OID_Inner_Traversal_fn, ctx );

    if( !ctx->Is_Match )
        return TRUE; /* exit early because we could not find oid */
    return FALSE;
}

static boolean Binding_IsEqual_IID_Inner_Traversal_fn(OALAVLTreeItem_t item, void *arg)
{
    DOFInterfaceID iid = (DOFInterfaceID)item;
    struct isEqualContext_t* ctx = (struct isEqualContext_t*)arg;

    if( DOFInterfaceID_IsEqual(iid, ((DOFInterfaceID)ctx->toCompare)) )
    {
        ctx->Is_Match = TRUE;
        return TRUE; /* end traversal */
    }
    return FALSE;
}

static boolean Binding_IsEqual_IID_Outer_Traversal_fn(OALAVLTreeItem_t item, void *arg)
{
    DOFInterfaceID iid = (DOFInterfaceID)item;
    struct isEqualContext_t* ctx = (struct isEqualContext_t*)arg;

    ctx->Is_Match = FALSE;
    ctx->toCompare = iid;

    OALAVLTree_Traverse( ctx->otherTree, Binding_IsEqual_IID_Inner_Traversal_fn, ctx );

    if( !ctx->Is_Match )
        return TRUE; /* exit early because we could not find oid */
    return FALSE;
}

static boolean Binding_IsEqual_AVP_Inner_Traversal_fn(OALAVLTreeItem_t item, void *arg)
{
    struct isEqualContext_t* ctx = (struct isEqualContext_t*)arg;
    DOFObjectIDAttribute* avp = (DOFObjectIDAttribute*)item;
    DOFObjectIDAttribute* other = (DOFObjectIDAttribute*)ctx->toCompare;

    if( ( avp->id == other->id ) &&
        ( avp->dataSize == other->dataSize ) &&
        PCRMem_Compare(avp->data, other->data, avp->dataSize))
    {
        ctx->Is_Match = TRUE;
        return TRUE; /* end traversal */
    }
    return FALSE;
}

static boolean Binding_IsEqual_AVP_Outer_Traversal_fn(OALAVLTreeItem_t item, void *arg)
{
    DOFObjectIDAttribute* avp = (DOFObjectIDAttribute*)item;
    struct isEqualContext_t* ctx = (struct isEqualContext_t*)arg;

    ctx->Is_Match = FALSE;
    ctx->toCompare = avp;

    OALAVLTree_Traverse( ctx->otherTree, Binding_IsEqual_AVP_Inner_Traversal_fn, ctx );

    if( !ctx->Is_Match )
        return TRUE; /* exit early because we could not find oid */
    return FALSE;
}

static boolean DOFPermission_Binding_IsEqual(DOFPermission self, DOFPermission other)
{
    DOFPermissionBuilder_t* selfx = (DOFPermissionBuilder_t*)self;
    DOFPermissionBuilder_t* otherx = (DOFPermissionBuilder_t*)other;

    if( self->dataLength == other->dataLength )
    {
        if( selfx->flags == otherx->flags )
        {
            if( OALAVLTree_GetNumItems( selfx->objectIDs ) == OALAVLTree_GetNumItems( otherx->objectIDs ) )
            {
                if( OALAVLTree_GetNumItems( selfx->interfaces ) == OALAVLTree_GetNumItems( otherx->interfaces ) )
                {
                    if( OALAVLTree_GetNumItems( selfx->avps ) == OALAVLTree_GetNumItems( otherx->avps ) )
                    {
                        struct isEqualContext_t ctx;

                        ctx.Is_Match = TRUE;

                        if( OALAVLTree_GetNumItems( selfx->objectIDs ) )
                        {
                            ctx.Is_Match = FALSE;
                            ctx.otherTree = otherx->objectIDs;
                            OALAVLTree_Traverse( selfx->objectIDs, Binding_IsEqual_OID_Outer_Traversal_fn, &ctx );
                        }

                        if( ctx.Is_Match )
                        {
                            if( OALAVLTree_GetNumItems( selfx->interfaces ) )
                            {
                                ctx.Is_Match = FALSE;
                                ctx.otherTree = otherx->interfaces;
                                OALAVLTree_Traverse( selfx->interfaces, Binding_IsEqual_IID_Outer_Traversal_fn, &ctx );
                            }

                            if( ctx.Is_Match )
                            {
                                if( OALAVLTree_GetNumItems( selfx->avps ) )
                                {
                                    ctx.Is_Match = FALSE;
                                    ctx.otherTree = otherx->avps;
                                    OALAVLTree_Traverse( selfx->avps, Binding_IsEqual_AVP_Outer_Traversal_fn, &ctx );
                                }

                                if( ctx.Is_Match )
                                    return TRUE;
                            }
                        }
                    }
                }
            }
        }
    }

    return FALSE;
}

static boolean DOFPermission_EPP_IsEqual(DOFPermission self, DOFPermission other)
{
    if( DOFObjectID_IsEqual(((EPPPermission_t*)self)->sid, ((EPPPermission_t*)other)->sid ) )
        return TRUE;
    return FALSE;
}

static boolean DOFPermission_DEFINE_IsEqual(DOFPermission self, DOFPermission other)
{
    if( DOFInterfaceID_IsEqual( ((DefinePermission_t*)self)->iid, ((DefinePermission_t*)other)->iid ) )
        return TRUE;
    return FALSE;
}

static boolean DOFPermission_TUNNEL_IsEqual(DOFPermission self, DOFPermission other)
{
    if( ((TunPermission_t*)self)->rdi == ((TunPermission_t*)other)->rdi )
        return TRUE;
    return FALSE;
}

boolean DOFPermission_IsEqual( DOFPermission self, DOFPermission other )
{
    if( self->type == other->type )
    {
        switch( self->type )
        {
        case DOFPERMISSIONTYPE_BINDING:
        case DOFPERMISSIONTYPE_BINDING_EXPLICIT:
            return DOFPermission_Binding_IsEqual( self, other );
            break;
        case DOFPERMISSIONTYPE_I_AM:
        case DOFPERMISSIONTYPE_ACT_AS:
            return DOFPermission_EPP_IsEqual( self, other );
            break;
        case DOFPERMISSIONTYPE_REQUESTOR:
        case DOFPERMISSIONTYPE_PROVIDER:
            return TRUE;
            break;
        case DOFPERMISSIONTYPE_DEFINE:
            return DOFPermission_DEFINE_IsEqual( self, other );
            break;
        case DOFPERMISSIONTYPE_TUNNEL_DOMAIN:
            return DOFPermission_TUNNEL_IsEqual( self, other );
            break;
        default:
            break;
        }
    }
    return FALSE;
}

uint32 DOFPermission_GetMarshalSize(DOFPermission self)
{
    return DOFPermission_MarshalByteLength(self);
}

uint32 DOFPermission_Marshal( DOFPermission self, uint32 size, uint8* buffer )
{
    uint32 len = 0;
    OALStream stream = OALStream_CreateRWWrapped( buffer, size );

    if( stream )
    {
        OALStream_SetWriteOffset( stream, 0 );
        if(DOFPermission_Internal_Marshal( self, stream ))
        {
            len = OALStream_GetWriteOffset( stream );
        }
        OALStream_Release( stream, FALSE );
    }
    return len;
}

DOFPermission DOFPermission_Create_Unmarshal( uint32 size, const uint8 * buffer )
{
    DOFPermission perm = NULL;
    OALStream stream = OALStream_CreateRWWrapped( (uint8*)buffer, size );

    if( stream )
    {
        perm = DOFPermission_Unmarshal( stream, TRUE );
        OALStream_Release( stream, FALSE );
    }

    return perm;
}

boolean DOFPermission_IsKnownType( DOFPermission self )
{
    if( self )
    {
        switch( self->type )
        {
        case DOFPERMISSIONTYPE_BINDING:
        case DOFPERMISSIONTYPE_BINDING_EXPLICIT:
        case DOFPERMISSIONTYPE_I_AM:
        case DOFPERMISSIONTYPE_ACT_AS:
        case DOFPERMISSIONTYPE_REQUESTOR:
        case DOFPERMISSIONTYPE_PROVIDER:
        case DOFPERMISSIONTYPE_DEFINE:
        case DOFPERMISSIONTYPE_TUNNEL_DOMAIN:
            return TRUE;
            break;
        default:
            break;
        }
    }
    return FALSE;
}
