/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/channel.h>
#include <internal/oaltypesinternal.h>
#include <internal/permissioninternal.h>
#include <internal/permissionsetinternal.h>
#include <internal/permission_manager.h>
#include <internal/slist.h>

typedef struct PendingPermissionObservers_t
{
    OALSListNode_t super;

    ExtendCompleteCB callback;
    void * callbackContext;
} PendingPermissionObservers_t;

typedef struct PendingPermissions_t
{
    OALSListNode_t super;

    DOFPermissionSet extendPerms;
    PendingPermissionObservers_t * observers;
} PendingPermissions_t;

typedef struct OALPermissionManager_t
{
    PendingPermissions_t * pendingPermissions;
} OALPermissionManager_t;

OALPermissionManager OALPermissionManager_Create(void)
{
    OALPermissionManager ret = (OALPermissionManager) DOFMem_Alloc(sizeof(OALPermissionManager_t));

    if(ret != NULL)
    {
        ret->pendingPermissions = NULL;
    }
    return ret;
}

void OALPermissionManager_Destroy(OALPermissionManager self)
{
    /* the SLists should be cleaned up via callbacks */
    DOFMem_Free(self);
}

typedef struct CheckPermissionContext_t
{
    DOFPermission perm;
    PendingPermissions_t * pendingRequest;
} CheckPermissionContext_t;

static boolean checkForExistingRequest(OALSListNode   item,
                                       void         * arg)
{
    PendingPermissions_t * pendingPermissions = (PendingPermissions_t*) item;
    CheckPermissionContext_t * context = (CheckPermissionContext_t*) arg;

    if(DOFPermissionSet_Contains(pendingPermissions->extendPerms, context->perm))
    {
        context->pendingRequest = pendingPermissions;
        return TRUE;
    }
    return FALSE;
}

void * OALPermissionManager_GetPendingContext(OALPermissionManager self,
                                              DOFPermission        perm)
{
    CheckPermissionContext_t context;

    PCR_Assert(self != NULL);
    PCR_Assert(perm != NULL);

    context.perm = perm;
    context.pendingRequest = NULL;
    OALSList_Traverse((OALSListNode) self->pendingPermissions, checkForExistingRequest, &context);
    return context.pendingRequest;
}

boolean OALPermissionManager_AddPendingContext(OALPermissionManager   self,
                                               void                 * pendingContext,
                                               ExtendCompleteCB       callback,
                                               void                 * callbackContext)
{
    PendingPermissions_t * pending;
    PendingPermissionObservers_t * observer;

    PCR_Assert(self != NULL);
    PCR_Assert(pendingContext != NULL);

    pending = (PendingPermissions_t*) pendingContext;
    observer = (PendingPermissionObservers_t*) DOFMem_Alloc(sizeof(PendingPermissionObservers_t));
    if(observer != NULL)
    {
        observer->callback = callback;
        observer->callbackContext = callbackContext;
        OALSList_Insert((OALSListNode*)&pending->observers, &(observer->super));
        return TRUE;
    }
    return FALSE;
}

typedef struct ExtentContext_t
{
    OALPermissionManager manager;
    PendingPermissions_t * perms;
} ExtendContext_t;

static void notifyObservers(OALSListNode   item,
                            void         * arg)
{
    PendingPermissionObservers_t * observer = (PendingPermissionObservers_t*) item;

    observer->callback(*((DOFError*) arg), observer->callbackContext);
    DOFMem_Free(observer);
}

static void extendCompleteCB(DOFError   status,
                             void     * context)
{
    ExtendContext_t * realContext = (ExtendContext_t*) context;

    OALSList_Remove((OALSListNode*) &realContext->manager->pendingPermissions, &(realContext->perms->super));
    OALSList_RemoveAll((OALSListNode*) &realContext->perms->observers, notifyObservers, &status);
    DOFPermissionSet_Destroy(realContext->perms->extendPerms);
    DOFMem_Free(realContext->perms);
    DOFMem_Free(realContext);
}

boolean OALPermissionManager_AddNewContext(OALPermissionManager    self,
                                           DOFPermissionSet        permSet,
                                           ExtendCompleteCB      * callback,
                                           void                 ** callbackContext)
{
    PendingPermissions_t * pending;

    PCR_Assert(self);
    PCR_Assert(*callback != NULL);

    pending = (PendingPermissions_t*) DOFMem_Alloc(sizeof(PendingPermissions_t));
    if(pending != NULL)
    {
        pending->extendPerms = DOFPermissionSet_Copy(permSet);
        if(pending->extendPerms != NULL)
        {
            pending->observers = (PendingPermissionObservers_t*) DOFMem_Alloc(sizeof(PendingPermissionObservers_t));
            if(pending->observers != NULL)
            {
                ExtendContext_t * realContext = (ExtendContext_t*) DOFMem_Alloc(sizeof(ExtendContext_t));

                if(realContext != NULL)
                {
                    realContext->manager = self;
                    realContext->perms = pending;
                    OALSList_Insert((OALSListNode*) &self->pendingPermissions, (OALSListNode) pending);
                    pending->observers->callback = *callback;
                    pending->observers->callbackContext = *callbackContext;
                    pending->observers->super.next = NULL;
                    *callback = extendCompleteCB;
                    *callbackContext = realContext;
                    return TRUE;
                }
                DOFMem_Free(pending->observers);
            }
            DOFPermissionSet_Destroy(permSet);
        }
        DOFMem_Free(pending);
    }
    return FALSE;
}
