/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/dof.h>
#include <dof/domain.h>
#include <dof/interfaceid.h>
#include <dof/objectid.h>
#include <dof/permissionset.h>
#include <dof/permission.h>
#include <internal/securityscope.h>
#include <dof/system.h>
#include <internal/permissionscope.h>
#include <internal/interfaceinternal.h>
#include <internal/interfaceidinternal.h>
#include <internal/marshal.h>
#include <internal/objectinternal.h>
#include <internal/objectidinternal.h>
#include <internal/permissionsetinternal.h>
#include <internal/permissioninternal.h>
#include <internal/dofinternal.h>
#include "permissionset_structs.h"

/* REVIEW: this is the number of permissions we step alloc by everytime we realloc our struct. tune this to average permissionset use. */
#define DOFPERMISSIONSET_ALLOC_STEP 4

/* DOFPermissionSet flags */
#define DOFPERMISSIONSET_GRANT      0x08
#define DOFPERMISSIONSET_BUILDER    0x04
#define DOFPERMISSIONSET_STARTED    0x02
#define DOFPERMISSIONSET_VALIDATED  0x01
#define DOFPERMISSIONSET_NOT_VALIDATED 0xFE

extern PCRMutex ref_count_mutex;
static PCRMutex permission_set_mutex = NULL;
static DOFPermissionSet empty_set = NULL;

static OALPermissionScope ALLPermissionScope = NULL;

static boolean DOFPermissionSet_ExpandIfNeeded(DOFPermissionSet self, uint16 add_count);
static boolean DOFPermissionSet_Internal_AddPermission_Helper(DOFPermissionSet self, DOFPermission perm, OALPermissionScope scope);

#define DOFPermissionSet_Lock() PCRMutex_Lock( permission_set_mutex )
#define DOFPermissionSet_Unlock()PCRMutex_Unlock( permission_set_mutex )

void DOFPermissionSet_Init(void)
{
    permission_set_mutex = PCRMutex_Create();
    PCR_Assert(permission_set_mutex);

    empty_set = DOFPermissionSet_Create();
    PCR_Assert(empty_set);

    ALLPermissionScope = OALPermissionScope_Create(1);
    if( ALLPermissionScope )
    {
        uint32 next = 0;
        OALPermissionScope_SetID(ALLPermissionScope,&next,DOFSECURITYSCOPEFLAG_ALL_SCOPES);
    }
}

void DOFPermissionSet_Shutdown(void)
{
    DOFPermissionSet_Destroy(empty_set);
    empty_set = NULL;

    PCRMutex_Destroy(permission_set_mutex);
    permission_set_mutex = 0;

    OALPermissionScope_Destroy(ALLPermissionScope);
    ALLPermissionScope = NULL;
}

void DOFPermissionSet_Destroy(DOFPermissionSet self)
{
    boolean free_ok = FALSE;

    if( !self )
        return;

    PCRMutex_Lock(ref_count_mutex);
    if(!--(self->refCount))
        free_ok = TRUE;
    PCRMutex_Unlock(ref_count_mutex);

    if( free_ok )
    {
        uint16 i = 0;
        for( ; i < self->permsCnt; ++i )
        {
            DOFPermission_Destroy( self->permission_list[i].perm );
            OALPermissionScope_Destroy( self->permission_list[i].scope );
        }
        DOFMem_Free( self->permission_list );
        DOFMem_Free( self );
    }
}

DOFPermissionSetBuilder DOFPermissionSetBuilder_Create_PermissionSet( DOFPermissionSet self )
{
    DOFPermissionSetBuilder ret = DOFPermissionSetBuilder_Create();

    if( ret )
    {
        if( !DOFPermissionSet_AddPermissionsFromSet( ( DOFPermissionSet)ret, self ) )
        {
            DOFPermissionSetBuilder_Destroy( ret );
            ret = NULL;
        }
    }
    return ret;
}

DOFPermissionSet DOFPermissionSet_CreateInternal( uint16 cnt, boolean builder )
{
    DOFPermissionSet ret = (DOFPermissionSet)DOFMem_Alloc( sizeof( DOFPermissionSet_t ));
    if( ret )
    {
        PCRMem_Set( ret, 0, sizeof( DOFPermissionSet_t ));
        if( cnt > 0 )
        {
            ret->permission_list = (permission_node_t*)DOFMem_Alloc( sizeof( permission_node_t ) * cnt );
            if( !ret->permission_list )
            {
                DOFMem_Free(ret);
                return NULL;
            }
            PCRMem_Set( ret->permission_list, 0, (sizeof( permission_node_t ) * cnt ));
        }

        if(builder) ret->flags |= DOFPERMISSIONSET_BUILDER;
        ret->free = cnt;
        ret->permsCnt = 0;
        ret->refCount = 1;
    }
    return ret;
}

DOFPermissionSetBuilder DOFPermissionSetBuilder_Create( void )
{
    return (DOFPermissionSetBuilder)DOFPermissionSet_CreateInternal( DOFPERMISSIONSET_ALLOC_STEP, TRUE );
}

void DOFPermissionSetBuilder_Destroy( DOFPermissionSetBuilder self )
{
    DOFPermissionSet_Destroy((DOFPermissionSet)self);
}

static boolean DOFPermissionSet_Internal_Addable( DOFPermissionSet self, DOFPermission perm )
{
    uint16 i;

    for ( i = 0; i < self->permsCnt; ++i)
    {
        if( DOFPermission_IsEqual( self->permission_list[i].perm, perm ) )
        {
            return FALSE;
        }
    }

    return TRUE;
}

boolean DOFPermissionSetBuilder_AddExactPermission( DOFPermissionSetBuilder self, DOFPermission perm )
{
    DOFPermissionSet selfx = (DOFPermissionSet)self;

    /* if not extendable || builder, we can add a permission */
    if( !self || !perm || !(selfx->flags & DOFPERMISSIONSET_BUILDER) || (selfx->permsCnt == 0xFFFF ))
        return FALSE;

    DOFPermissionSet_Lock();
    if( DOFPermissionSet_ExpandIfNeeded( selfx, DOFPERMISSIONSET_ALLOC_STEP ) )
    {
        if( DOFPermissionSet_Internal_AddPermission_Helper( selfx, perm, NULL) )
        {
            DOFPermissionSet_Unlock();
            return TRUE;
        }
    }
    DOFPermissionSet_Unlock();
    return FALSE;
}

boolean DOFPermissionSetBuilder_AddPermission( DOFPermissionSetBuilder self, DOFPermission perm )
{
    DOFPermissionSet selfx = (DOFPermissionSet)self;

    /* if not extendable || builder, we can add a permission */
    if( !self || !perm || !(selfx->flags & DOFPERMISSIONSET_BUILDER) || (selfx->permsCnt == 0xFFFF ))
        return FALSE;

    DOFPermissionSet_Lock();
    if(DOFPermissionSet_Internal_Addable( selfx, perm ))
    {
        if( DOFPermissionSet_ExpandIfNeeded( selfx, DOFPERMISSIONSET_ALLOC_STEP ) )
        {
            if( DOFPermissionSet_Internal_AddPermission_Helper( selfx, perm, NULL) )
            {
                DOFPermissionSet_Unlock();
                return TRUE;
            }
        }
    }
    DOFPermissionSet_Unlock();
    return FALSE;
}

DOFPermissionSet DOFPermissionSet_DeepCopy( DOFPermissionSet self )
{
    DOFPermissionSet ret = NULL;
    if( self )
    {
        ret = (DOFPermissionSet)DOFPermissionSet_CreateInternal( self->permsCnt, FALSE );
        if( ret )
        {
            boolean fail = FALSE;
            uint16 i = 0;

            for( ; i < self->permsCnt; ++i )
            {
                ret->permission_list[i].perm = DOFPermission_Copy( self->permission_list[i].perm );
                if( !ret->permission_list[i].perm )
                {
                    fail = TRUE;
                    break;
                }

                ret->permission_list[i].scope = OALPermissionScope_Copy( self->permission_list[i].scope );
                if( !(!self->permission_list[i].scope || ret->permission_list[i].scope) )
                {
                    fail = TRUE;
                    break;
                }

            }

            if( !fail )
            {
                ret->free = 0;
                ret->refCount = 1;
                ret->flags = self->flags;
                ret->permsCnt = self->permsCnt;
                ret->vbyteLen = self->vbyteLen;
                ret->rbyteLen = self->rbyteLen;
            }
            else
            {
                DOFPermissionSet_Destroy( ret );
                ret = NULL;
            }
        }
    }
    return ret;
}

DOFPermissionSet DOFPermissionSetBuilder_Build( DOFPermissionSetBuilder self )
{
    DOFPermissionSet ret = DOFPermissionSet_DeepCopy( (DOFPermissionSet)self );
    if( ret )
        ret->flags &= ~DOFPERMISSIONSET_BUILDER;
    return ret;
}

DOFPermissionSet DOFPermissionSetBuilder_BuildAndDestroy( DOFPermissionSetBuilder self )
{
    DOFPermissionSet ret = (DOFPermissionSet)self;
    if( ret )
        ret->flags &= ~DOFPERMISSIONSET_BUILDER;
    return ret;
}

DOFPermissionSet DOFPermissionSet_Create( void )
{
    return (DOFPermissionSet)DOFPermissionSet_CreateInternal(0, FALSE); /* empty PermissionSet */
}

static boolean DOFPermissionSet_ExpandIfNeeded(DOFPermissionSet self, uint16 add_count)
{
    if(self->free < add_count)
    {
        permission_node_t* temp = NULL;

        if( self->permission_list )
            temp = (permission_node_t*)DOFMem_Realloc( self->permission_list, sizeof( permission_node_t ) * (self->permsCnt + self->free + add_count));
        else
            temp = (permission_node_t*)DOFMem_Alloc( sizeof( permission_node_t ) * (add_count));


        if(!temp)
        {
            return FALSE;
        }

        self->permission_list = temp;
        self->free += add_count;
    }
    return TRUE;
}

static boolean DOFPermissionSet_Internal_AddPermission_Helper(DOFPermissionSet self, DOFPermission perm, OALPermissionScope scope)
{
    if(self && perm)
    {
        DOFPermission temp = DOFPermission_Copy(perm);
        if( temp )
        {
            PCR_Assert(self->free > 0); /* memory should already have been expanded */
            self->permission_list[self->permsCnt].perm = temp;
            self->permission_list[self->permsCnt].scope = OALPermissionScope_Copy(scope);
            if( !scope || self->permission_list[self->permsCnt].scope )
            {
                self->rbyteLen += DOFPermission_MarshalByteLength(perm);
                if(scope)
                    self->vbyteLen += OALPermissionScope_CalculateMarshalSize(scope);

                ++self->permsCnt;
                --self->free;
                return TRUE;
            }
            self->permission_list[self->permsCnt].perm = NULL;
            DOFPermission_Destroy( temp );
        }
    }
    return FALSE;
}

boolean DOFPermissionSet_AddPermissionsFromSet( DOFPermissionSet self, const DOFPermissionSet addFrom )
{
    DOFPermissionSet_Lock();
    if( self && addFrom && self->permsCnt != 0xFFFF && ((self->flags & DOFPERMISSIONSET_VALIDATED) == (addFrom->flags & DOFPERMISSIONSET_VALIDATED)) )
    {
        if( DOFPermissionSet_ExpandIfNeeded( self, addFrom->permsCnt ) )
        {
            uint16 i;
            boolean fail = FALSE;
            for( i = 0; i < addFrom->permsCnt; ++i )
            {
                if ( addFrom->permission_list[i].scope != NULL || DOFPermissionSet_Internal_Addable( self, addFrom->permission_list[i].perm ))
                {
                    if( !DOFPermissionSet_Internal_AddPermission_Helper( self, addFrom->permission_list[i].perm, addFrom->permission_list[i].scope ) )
                    {
                        fail = TRUE;
                        break;
                    }
                }
            }
            DOFPermissionSet_Unlock();
            return !fail;
        }
    }
    DOFPermissionSet_Unlock();
    return FALSE;
}

boolean DOFPermissionSet_AddTrustedPermission( DOFPermissionSet self, const DOFPermission perm )
{
    if(self && perm)
    {
        DOFPermissionSet_Lock();
        /* this call is only allowed on a validated permission set */
        if(self->flags&DOFPERMISSIONSET_VALIDATED)
        {
            if(DOFPermissionSet_Internal_Addable( self, perm ))
            {
                /* Do we need more memory */
                if(DOFPermissionSet_ExpandIfNeeded(self,1))
                {
                    if(DOFPermissionSet_Internal_AddPermission_Helper(self,perm,ALLPermissionScope))
                    {
                        DOFPermissionSet_Unlock();
                        return TRUE;
                    }
                }
            }
        }
        DOFPermissionSet_Unlock();
    }
    return FALSE;
}

void DOFPermissionSet_SetGrantPerms( DOFPermissionSet self )
{
    if( self )
        self->flags |= DOFPERMISSIONSET_GRANT;
}

void DOFPermissionSet_ClearGrantPerms( DOFPermissionSet self )
{
    if( self )
        self->flags &= ~DOFPERMISSIONSET_GRANT;
}

struct perm_val_t
{
    DOFPermission   perm;
    uint32          cnt;
    uint32          alc;
    boolean         found;
    boolean         isGrant;
    uint32*         ids; /* alloc 8 realloc when more are needed */
}perm_val_t;

static boolean permission_validate( permission_node_t* item, struct perm_val_t* arg )
{
    uint32 i;

    if( DOFPermission_IsAllowedBy( item->perm, arg->perm, !arg->isGrant, FALSE ) )
    {
        /* we may have groups but not the full list so we don't add them and we may end up extending */
        if(DOFPermission_IsTerminal(item->perm) ||
          (!DOFPermission_IsTerminal(item->perm) && !OALPermissionScope_IsMasking(item->scope)))
        {
            uint32 const * ids = OALPermissionScope_GetIds(item->scope);

            if( ( ( arg->cnt ) + OALPermissionScope_GetIdCount(item->scope) ) > arg->alc )
            {
                uint32 alcTemp = ( ( ( ( arg->cnt + 1 ) + OALPermissionScope_GetIdCount(item->scope) ) + 8 ) );
                uint32* idsTemp = (uint32*)DOFMem_Realloc( arg->ids, ( sizeof( uint32 ) * alcTemp ) );
                if( !idsTemp )
                {
                    return FALSE; /* end early error case */
                }
                arg->ids = idsTemp;
                arg->alc = alcTemp;
            }

            arg->found = TRUE;
            for( i = 0; i < OALPermissionScope_GetIdCount(item->scope); ++i )
            {
                if( ids[i] != DOFSECURITYSCOPEFLAG_NOT_MASKING ) /* Not Masking scope should not be in the validation scopes */
                {
                    arg->ids[arg->cnt] = ids[i];
                    ++arg->cnt;
                }
            }
        }
    }
    return TRUE;
}

OALPermissionScope DOFPermissionSet_Validate( DOFPermissionSet self, DOFPermission perm )
{
    OALPermissionScope ret = NULL;

    if( self )
    {
        struct perm_val_t temp;
        boolean exitEarly = FALSE;

        DOFPermissionSet_Lock();
        if( (self->flags & DOFPERMISSIONSET_VALIDATED) )
        {
            temp.cnt = 0;
            temp.alc = 8;
            temp.found = FALSE;
            temp.perm = perm;
            temp.isGrant = (self->flags & DOFPERMISSIONSET_GRANT);

            if( temp.perm )
            {
                temp.ids = (uint32*)DOFMem_Alloc( sizeof( uint32 ) * 8 );
                if( temp.ids )
                {
                    uint16 i = 0;

                    for( ; i < self->permsCnt; ++i )
                    {
                        if(!permission_validate( &self->permission_list[i], &temp ))
                        {
                            exitEarly = TRUE;
                            break;
                        }
                    }

                    if( !exitEarly )
                    {
                        if( temp.found )
                        {
                            ret = OALPermissionScope_Create( temp.cnt );
                            if( ret )
                            {
                                uint32 i = 0;
                                uint32 next = 0;

                                for( ; i<temp.cnt; ++i )
                                {
                                    OALPermissionScope_SetID( ret, &next, temp.ids[i]);
                                }
                            }
                        }
                    }
                    DOFMem_Free( temp.ids );
                }
            }
        }
        DOFPermissionSet_Unlock();
    }
    return ret;
}

boolean DOFPermissionSet_Contains( DOFPermissionSet self, DOFPermission perm )
{
    uint16 i;

    PCR_Assert(self != NULL);
    DOFPermissionSet_Lock();
    for( i = 0; i < self->permsCnt; ++i )
    {
        if( DOFPermission_IsAllowedBy( self->permission_list[i].perm, perm, FALSE, FALSE ) )
        {
            DOFPermissionSet_Unlock();
            return TRUE;
        }
    }
    DOFPermissionSet_Unlock();
    return FALSE;
}

OALPermissionScope DOFPermissionSet_ValidateRequestorProvider( DOFPermissionSet self, boolean provider )
{
    DOFPermission perm;

    if( provider )
    {
        perm = DOFPermissionProvider_Create();
    }
    else
    {
        perm = DOFPermissionRequestor_Create();
    }
    if(perm != NULL)
    {
        OALPermissionScope ret = DOFPermissionSet_Validate(self, perm);

        DOFPermission_Destroy(perm);
        return ret;
    }
    return NULL;
}

OALPermissionScope DOFPermissionSet_ValidateDefine( DOFPermissionSet self, DOFInterfaceID iid )
{
    DOFPermission perm = DOFPermissionDefine_Create(iid);

    if(perm != NULL)
    {
        OALPermissionScope ret = DOFPermissionSet_Validate( self, perm);

        DOFPermission_Destroy(perm);
        return ret;
    }
    return NULL;
}

OALPermissionScope DOFPermissionSet_ValidateBinding( DOFPermissionSet self, DOFPermissionBindingAction access, const DOFObjectID oid, DOFInterfaceID iid )
{
    DOFPermission perm = DOFPermissionBinding_Create( access, oid, iid );

    if(perm != NULL)
    {
        OALPermissionScope ret = DOFPermissionSet_Validate( self, perm);

        DOFPermission_Destroy(perm);
        return ret;
    }
    return NULL;
}

DOFPermissionSet DOFPermissionSet_Copy( DOFPermissionSet self )
{
    DOFPermissionSet ret = NULL;

    if( self )
    {
        PCRMutex_Lock(ref_count_mutex);
        if( self->refCount < 0xFFFFFFFF )
        {
            ++(self->refCount);
            ret = self;
        }
        PCRMutex_Unlock(ref_count_mutex);
    }
    return ret;
}

uint32 DOFPermissionSet_GetPermissionCount( DOFPermissionSet self )
{
    if( self )
        return self->permsCnt;
    return 0;
}

DOFPermissionSet DOFPermissionSet_CreateComplement(DOFPermissionSet self)
{
    DOFPermissionSet ret = NULL;
    uint16 i = 0;
    uint32 cnt = 0;

    DOFPermissionSet_Lock();
    for( ; i < self->permsCnt; ++i )
    {
        cnt += DOFPermission_ComplementCount( self->permission_list[i].perm );
    }

    ret = DOFPermissionSet_CreateInternal( cnt, FALSE );
    if( ret )
    {
        for( i = 0; i < self->permsCnt; ++i )
        {
            DOFPermissionSet pr = DOFPermission_CreateComplement( self->permission_list[i].perm );
            if( !(pr && DOFPermissionSet_AddPermissionsFromSet( ret, pr )) )
            {
                DOFPermissionSet_Destroy( pr );
                DOFPermissionSet_Destroy( ret );
                ret = NULL;
                break;
            }
            DOFPermissionSet_Destroy( pr );
        }
    }
    DOFPermissionSet_Unlock();
    return ret;
}

struct pfv_t
{
    uint32              groupID;
    boolean             hadError;
    union
    {
        DOFPermissionSet    groupPerms;
        permission_node_t*  reqPerm;
    }type;
}pfv_t;

static boolean permission_start_scopes(permission_node_t* item, void* arg)
{
    item->scope = (OALPermissionScope)OALPermissionScopeBuilder_Create();
    if( item->scope )
        return TRUE;
    return FALSE;
}

void DOFPermissionSet_CleanupFailedBuildingValidation( DOFPermissionSet request )
{
    DOFPermissionSet_Lock();
    if( request )
    {
        uint16 i = 0;
        for( ; i < request->permsCnt; ++i )
        {
            OALPermissionScopeBuilder_Destroy( (OALPermissionScopeBuilder)request->permission_list[i].scope );
            request->permission_list[i].scope = NULL;
        }
    }
    DOFPermissionSet_Unlock();
    
}

boolean DOFPermissionSet_StartBuildingValidation( DOFPermissionSet request )
{
    DOFPermissionSet_Lock();
    if( request )
    {
        boolean exitEarly = FALSE;
        uint16 i = 0;
        request->vbyteLen = 0;

        for( ; i < request->permsCnt; ++i )
        {
            if(!permission_start_scopes( &request->permission_list[i], NULL ))
            {
                exitEarly = TRUE;
                break;
            }
        }

        if( !exitEarly )
        {
            request->flags |= DOFPERMISSIONSET_STARTED;
            DOFPermissionSet_Unlock();
            return TRUE;
        }
    }
    DOFPermissionSet_Unlock();
    return FALSE;
}

/* This is only used in the AS validation  */
static boolean DOFPermission_Internal_AddIDToValidation( DOFPermission request, OALPermissionScopeBuilder scope, uint32 groupID, DOFPermissionSet groupPerms, boolean* isMasking )
{
    boolean hadError = FALSE;
    uint16 k = 0;
    boolean masking = FALSE;

    DOFPermissionSet_Lock();
    for( ; k < groupPerms->permsCnt; ++k )
    {
        if( DOFPermission_IsAllowedBy( groupPerms->permission_list[k].perm, request, FALSE, TRUE ) )
        {
            masking = FALSE;
            if( !OALPermissionScopeBuilder_AddID( scope, groupID ) )
            {
                 hadError = TRUE;
            }
            break; /* we are in a single group id so we have it and it can't be masked we can exit early */
        }
        else if( DOFPermission_Intersects( request, groupPerms->permission_list[k].perm ) )
        {
            masking = TRUE;
        }
    }

    if( !hadError )
    {
        DOFPermissionSet_Unlock();
        (*isMasking) = masking;
        return TRUE;
    }

    DOFPermissionSet_Unlock();
    return FALSE;
}

/* This is only used in the AS validation  */
boolean DOFPermissionSet_AddALLScopesToValidation(DOFPermissionSet permission_set, DOFPermissionSet sids)
{
    uint16 i;
    boolean mask;
    boolean ret = TRUE;

    for( i = 0; i < permission_set->permsCnt; ++i )
    {
        if( !DOFPermission_Internal_AddIDToValidation(permission_set->permission_list[i].perm, (OALPermissionScopeBuilder)permission_set->permission_list[i].scope, DOFSECURITYSCOPEFLAG_ALL_SCOPES, sids, &mask) )
        {
            ret = FALSE;
            break;
        }
    }

    return ret;
}

static OALPermissionScopeBuilder DOFPermissionSet_ReducePermissionScopes( DOFPermissionSet self, boolean* masking )
{
    uint16 i;
    OALPermissionScopeBuilder ret = NULL;

    if( self )
    {
        ret = OALPermissionScopeBuilder_CreateCopy( (OALPermissionScopeBuilder)self->permission_list[0].scope );
        if( ret )
        {
            for( i = 1; i < self->permsCnt; ++i )
            {
                if( OALPermissionScopeBuilder_DoIntersection( ret, (OALPermissionScopeBuilder)self->permission_list[i].scope ) )
                    (*masking) = TRUE;
            }
        }
    }
    return ret;
}

/* This is only used in the AS validation  */
boolean DOFPermissionSet_AddIDToValidation( DOFPermissionSet request, DOFAuthenticatorNodePermissionSetMap perm_map )
{
    uint16 i;
    uint16 j;

    /* REVIEW: This should be optimized */
    for( i = 0; i < request->permsCnt; ++i )
    {
        if( DOFPermission_IsKnownType( request->permission_list[i].perm ) )
        {
            DOFPermissionSet normalized = DOFPermission_CreateNormalized( request->permission_list[i].perm );
            if( normalized )
            {
                if( DOFPermissionSet_GetPermissionCount(normalized) > 0 ) /* We should have at least the original permission */
                {
                    if (DOFPermissionSet_StartBuildingValidation(normalized)) 
                    {
                        boolean masking = FALSE;

                        for( j = 0; j < normalized->permsCnt; ++j )
                        {
                            uint32 k;
                            for( k = 0; k < perm_map->count; ++k )
                            {
                                boolean mask;
                                if( !DOFPermission_Internal_AddIDToValidation( normalized->permission_list[j].perm, (OALPermissionScopeBuilder)normalized->permission_list[j].scope, perm_map->list[k].id, perm_map->list[k].permissionset, &mask ) )
                                {
                                    DOFPermissionSet_Destroy( normalized );
                                    return FALSE;
                                }

                                if( mask )
                                    masking = TRUE;
                            }
                        }

                        if( request->permission_list[i].scope )
                            OALPermissionScopeBuilder_Destroy( (OALPermissionScopeBuilder)request->permission_list[i].scope );

                        request->permission_list[i].scope = (OALPermissionScope)DOFPermissionSet_ReducePermissionScopes( normalized, &masking );
                        if( !request->permission_list[i].scope )
                        {
                            DOF_LogMessage(DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFPermissionSet: Failed to create copy of normalized permission scopes.");
                            DOFPermissionSet_Destroy( normalized );
                            return FALSE;
                        }

                        DOFPermissionSet_CleanupFailedBuildingValidation( normalized );

                        if( !masking )
                            OALPermissionScopeBuilder_AddID( (OALPermissionScopeBuilder)request->permission_list[i].scope, DOFSECURITYSCOPEFLAG_NOT_MASKING );
                    }
                    else
                    {
                        DOFLog_Message(DOFLOGLEVEL_WARNING, "DOFPermission: Failed to start building normalized permission scopes.");
                        DOFPermissionSet_Destroy( normalized );
                        return FALSE;
                    }
                }
                else
                {
                    DOFLog_Message(DOFLOGLEVEL_WARNING, "DOFPermission: Normalization did not return at least one permissions.");
                    DOFPermissionSet_Destroy( normalized );
                    return FALSE;
                }
                DOFPermissionSet_Destroy( normalized );
            }
            else
                return FALSE;
        }
        else
        {
            /* This adds an empty scope since we do not know how to deal with this permission type, so it is rejected. */
            if( request->permission_list[i].scope )
                OALPermissionScopeBuilder_Destroy( (OALPermissionScopeBuilder)request->permission_list[i].scope );

            request->permission_list[i].scope = (OALPermissionScope)OALPermissionScopeBuilder_Create(); 
            if( !request->permission_list[i].scope )
                return FALSE;
        }
    }
    return TRUE;
}

boolean DOFPermissionSet_FinishBuildingValidation( DOFPermissionSet request )
{
    DOFPermissionSet_Lock();
    if( (request->flags & DOFPERMISSIONSET_STARTED) )
    {
        uint16 i = 0;
        boolean exitEarly = FALSE;
        request->flags &= ~DOFPERMISSIONSET_STARTED;

        for( ; i < request->permsCnt; ++i )
        {
            request->permission_list[i].scope = OALPermissionScopeBuilder_BuildAndDestroy( (OALPermissionScopeBuilder)request->permission_list[i].scope );
            request->vbyteLen += OALPermissionScope_CalculateMarshalSize( request->permission_list[i].scope );
            if( !request->permission_list[i].scope )
            {
                exitEarly = TRUE;
                break;
            }
        }

        if( !exitEarly )
        {
            request->flags |= DOFPERMISSIONSET_VALIDATED;
            DOFPermissionSet_Unlock();
            return TRUE;
        }
        else
        {
            for(; i > 0; --i )
            {
                OALPermissionScope_Destroy(request->permission_list[i].scope);
            }
        }
    }
    DOFPermissionSet_Unlock();
    return FALSE;
}

boolean DOFPermissionSet_UnmarshalValidation( DOFPermissionSet self, OALStream stream, OALPermissionScope connectionScope )
{
    uint16 cnt;
    boolean exitEarly = FALSE;

    DOFPermissionSet_Lock();
    if( self )
    {
        if( ( OALMarshal_StreamUncompress2( stream, &cnt ) == DOFERROR_OK ) && ( self->permsCnt == cnt ) )
        {
            OALPermissionScope* permissionScopes = (OALPermissionScope*)DOFMem_Alloc( sizeof( OALPermissionScope ) * cnt );
            if( permissionScopes )
            {
                if( DOFERROR_OK == OALPermissionScope_Unmarshal( stream, cnt, permissionScopes, connectionScope ) )
                {
                    uint16 i = 0;
                    for( ; i < cnt; ++i )
                    {
                        if( self->permission_list[i].scope )
                        {
                            self->vbyteLen -= OALPermissionScope_CalculateMarshalSize( self->permission_list[i].scope );
                            OALPermissionScope_Destroy(self->permission_list[i].scope);
                        }
                        self->permission_list[i].scope = permissionScopes[i];
                        if( self->permission_list[i].scope )
                        {
                            self->vbyteLen += OALPermissionScope_CalculateMarshalSize( self->permission_list[i].scope );
                        }
                        else
                        {
                            /* Error case */
                            exitEarly = TRUE;
                            break;
                        }
                    }
                }
                else
                {
                    /* Error case */
                    exitEarly = TRUE;
                }

                DOFMem_Free( permissionScopes );

                if( !exitEarly )
                {
                    self->flags |= DOFPERMISSIONSET_VALIDATED;
                    DOFPermissionSet_Unlock();
                    return TRUE;
                }
            }
        }
        self->flags &= DOFPERMISSIONSET_NOT_VALIDATED;
    }
    DOFPermissionSet_Unlock();
    return FALSE;
}

DOFPermissionSet DOFPermissionSet_UnmarshalRequest( OALStream stream, boolean unknownAllowed)
{
    uint16 i = 0;
    uint16 cnt;
    DOFPermissionSet self = NULL;

    DOFPermissionSet_Lock();
    if( ( OALMarshal_StreamUncompress2( stream, &cnt ) == DOFERROR_OK ) )
    {
        self = DOFPermissionSet_CreateInternal( cnt, FALSE );
        if( self )
        {
            self->flags |= DOFPERMISSIONSET_BUILDER;
            for( ; i < cnt; ++i )
            {
                DOFPermission prm = DOFPermission_Unmarshal( stream, unknownAllowed );
                if( prm && DOFPermissionSetBuilder_AddExactPermission((DOFPermissionSetBuilder)self, prm ) )
                {
                    DOFPermission_Destroy( prm );
                }
                else
                {
                    DOFPermission_Destroy( prm );
                    DOFPermissionSet_Destroy(self);
                    DOFPermissionSet_Unlock();
                    return NULL;
                }
            }

            DOFPermissionSet_Unlock();
            self->flags &= ~DOFPERMISSIONSET_BUILDER;
            return self;
        }
    }
    DOFPermissionSet_Unlock();
    return NULL;
}

uint32 DOFPermissionSet_MarshalValidationSize( DOFPermissionSet self )
{
    return (self->vbyteLen + OALMarshal_GetCompressedValueSize( 2, self->permsCnt ));
}

boolean DOFPermissionSet_MarshalValidation( DOFPermissionSet self, OALStream stream )
{
    DOFPermissionSet_Lock();
    if( self && (self->flags & DOFPERMISSIONSET_VALIDATED) && (OALMarshal_StreamCompress2( stream, self->permsCnt ) == DOFERROR_OK) )
    {
        uint16 i = 0;
        for( ; i < self->permsCnt; ++i )
        {
            if(!OALPermissionScope_Marshal( self->permission_list[i].scope, stream ))
            {
                DOFPermissionSet_Unlock();
                return FALSE;
            }
        }
        DOFPermissionSet_Unlock();
        return TRUE;
    }
    DOFPermissionSet_Unlock();
    return FALSE;
}

uint32 DOFPermissionSet_MarshalRequestSize( DOFPermissionSet self )
{
    return (self->rbyteLen + OALMarshal_GetCompressedValueSize(2,self->permsCnt));
}

boolean DOFPermissionSet_MarshalRequest( DOFPermissionSet self, OALStream stream )
{
    DOFPermissionSet_Lock();
    if( self && (OALMarshal_StreamCompress2( stream, self->permsCnt ) == DOFERROR_OK) )
    {
        uint16 i = 0;
        for( ; i < self->permsCnt; ++i )
        {
            PCR_Assert( self->permission_list[i].perm );
            if(!DOFPermission_Internal_Marshal( self->permission_list[i].perm, stream ))
            {
                DOFPermissionSet_Unlock();
                return FALSE;
            }
        }
        DOFPermissionSet_Unlock();
        return TRUE;
    }
    DOFPermissionSet_Unlock();
    return FALSE;
}

DOFPermissionSet DOFPermissionSet_GetEmptyPermissionSet( void )
{
    return empty_set;
}

uint32 DOFPermissionSet_GetMarshalSize( DOFPermissionSet self )
{
    return DOFPermissionSet_MarshalRequestSize(self);
}

uint32 DOFPermissionSet_Marshal( DOFPermissionSet self, uint32 size, uint8 * buffer )
{
    uint32 len = 0;
    OALStream stream = OALStream_CreateRWWrapped( buffer, size );

    if( stream )
    {
        OALStream_SetWriteOffset( stream, 0 );
        if( DOFPermissionSet_MarshalRequest( self, stream ) )
        {
            len = OALStream_GetWriteOffset( stream );
        }
        OALStream_Release( stream, FALSE );
    }
    return len;
}

DOFPermissionSet DOFPermissionSet_Create_Unmarshal( uint32 size, const uint8 * buffer )
{
    DOFPermissionSet perms = NULL;
    OALStream stream = OALStream_CreateRWWrapped( (uint8*)buffer, size );

    if( stream )
    {
        perms = DOFPermissionSet_UnmarshalRequest( stream, TRUE );
        OALStream_Release( stream, FALSE );
    }
    return perms;
}
