/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/pointerlist.h>

#define PTRLIST_SIZE_INCREASE   8

/** @internal @addtogroup OALPtrList
    A dynamicaly allocated list of void pointers

    The OALPtrList does not assume "ownership" of anything it is used to
    contain. Destroying the OALPtrList will not destroy its contained elements.

@{
**/

/*
**  "Private" Type and Method Declarations
*/

typedef struct OALPtrList_t
{
    uint32  elementCount;   /* The number of elements in this list */
    uint32  capacity;       /* The amount of elements which can be added
                               to ptrBuf without requiring a resize */
    void**  ptrBuf;         /* The array of stored pointers */
} OALPtrList_t;


/*
**  "Public" Methods
*/

/* "Static" methods */

OALPtrList OALPtrList_Create(uint32 initialCapacity)
{
    /* allocate memory for the list struct */
    OALPtrList newList = (OALPtrList)(DOFMem_Alloc( sizeof(OALPtrList_t) ));
    /* compute the required size of the storage buffer based on initialCapacity */
    const uint32 listBufSize = (uint32)( sizeof(void*) * initialCapacity );

    if(!newList)
    {
        return NULL;
    }
    /* allocate memory for the storage buffer */
    newList->ptrBuf = DOFMem_Alloc(listBufSize);

    if(!newList->ptrBuf)
    {
        DOFMem_Free( newList );
        return NULL;
    }

    /* initialize the rest of the struct members' values */
    newList->elementCount = 0;
    newList->capacity = initialCapacity;

    return newList;
}

/* "Object" methods */

void OALPtrList_Destroy(OALPtrList self)
{
    PCR_Assert(self);

    DOFMem_Free(self->ptrBuf);
    DOFMem_Free(self);
}

uint32 OALPtrList_GetCount(OALPtrList self)
{
    PCR_Assert(self);

    return self->elementCount;
}

uint32 OALPtrList_GetCapacity(OALPtrList self)
{
    PCR_Assert(self);

    return self->capacity;
}

boolean OALPtrList_InsertAt(OALPtrList self, void* item, uint32 index)
{
    PCR_Assert(self);
    PCR_Assert(index <= self->elementCount);
    PCR_Assert(self->elementCount <= self->capacity);

    /* check if resize is needed */
    if(self->elementCount == self->capacity)/* resize is needed */
    {
        uint32 newSize;
        void* newPtrBuf;

        /* Doing it correctly now. */
        uint32 newCapacity = self->capacity + PTRLIST_SIZE_INCREASE;
        /* check for overflow */
        if((sizeof(void*) * newCapacity) < (sizeof(void*) * self->capacity))
        {
            return FALSE; /* We are already full, we can't insert more! */
        }
        newSize = (uint32)( sizeof(void*) * newCapacity );
        newPtrBuf = DOFMem_Realloc((void*)self->ptrBuf, newSize);
        if(!newPtrBuf)
        {
            return FALSE; /* realloc failed, we can't insert item */
        }
        else
        {
            self->ptrBuf = (void**)newPtrBuf;
        }

        self->capacity = newCapacity;
        /* resize complete */
    }

    /* check to see if we must shift some elements over before inserting the item */
    if(index < self->elementCount)
    {
        const uint32 shiftSize = (uint32)( sizeof(void*) * (self->elementCount - index) );
        PCRMem_Move( self->ptrBuf+index+1, self->ptrBuf+index, shiftSize );
    }

    /* add the item in the appropriate spot */
    self->ptrBuf[index] = (void*)item;
    ++(self->elementCount);

    PCR_Assert(self->elementCount <= self->capacity);

    return TRUE;
}

void OALPtrList_OverwriteAt(OALPtrList self, void* element, uint32 index)
{
    PCR_Assert(self);
    PCR_Assert(index < self->elementCount);

    self->ptrBuf[index] = element;
}

void* OALPtrList_RemoveAt(OALPtrList self, uint32 index)
{
    void* returnValue;

    PCR_Assert(self);
    PCR_Assert(index < self->elementCount);

    returnValue = self->ptrBuf[index];

    if(index < (self->elementCount-1))/* item is not at the end of the array so a memory "shift" is needed */
    {
        const uint32 shiftSize = (uint32)( sizeof(void*) * (self->elementCount - (index+1)) );
        PCRMem_Move( self->ptrBuf+index, self->ptrBuf+index+1, shiftSize);
    }

    --(self->elementCount);

    return returnValue;
}

void* OALPtrList_PeekAt(OALPtrList self, uint32 index)
{
    PCR_Assert(self);
    PCR_Assert(index < self->elementCount);

    return self->ptrBuf[index];
}

uint32 OALPtrList_FirstIndexOf(OALPtrList self, void* element)
{
    uint32 i;

    PCR_Assert(self);

    for(i = 0; i < self->elementCount; ++i)
    {
        if(self->ptrBuf[i] == element)
            break;
    }
    return i;
}

/**
@}
**/
