/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>

#include <dof/address.h>
#include <dof/result.h>
#include <internal/bridgeinternal.h>
#include <internal/dofinternal.h>
#include <internal/exception.h>
#include <internal/protocolstack.h>
#include <internal/connected_protocolstack.h>
#include <internal/protocollayerresult.h>
#include <internal/esp_v2.h>
#include <internal/enp_v1.h>
#include <internal/epp_v2.h>
#include <internal/oap_v1.h>
#include <internal/tep_v1.h>
#include <internal/trp_v1.h>
#include <internal/epp_negative_one.h>
#include <internal/resultinternal.h>
#include <internal/securitymode.h>
#include <internal/source.h>
#include <internal/router.h>
#include <internal/marshal.h>
#include <internal/messaging.h>
#include <internal/transportmanager.h>
#include <internal/channel_stack_common.h>
#include <internal/connected_remotechannel.h>
#include <internal/remotechannel.h>
#include <internal/error.h>
#include <internal/timer.h>
#include <internal/channel_stack_messages.h>
#include <internal/remotechannelmanager.h>
#include <internal/threadpool.h>

#ifdef OALPROTOCOLSTACK_PACKET_TRACE
#include <stdio.h>
#endif

/* Order of these states is CRUCIAL. */
#define CONNECTEDSTACK_STATE_EPS_NEGOTIATION    0x00
#define CONNECTEDSTACK_STATE_APP_NEGOTIATION    0x01
#define CONNECTEDSTACK_STATE_AUTH_NEGOTIATION   0x02
#define CONNECTEDSTACK_STATE_OPEN               0x03

#define EPS_NEGOTIATION_TIMEOUT 10000   /* ten seconds */
#define ESP_NEGOTIATION_TIMEOUT 30000   /* thirty seconds */
#define TEP_NEGOTIATION_TIMEOUT 30000   /* thirty seconds */

typedef struct OALConnectedProtocolStack_t
{
    OALProtocolStack_t      superClass;         /* OALConnectedProtocolStack inherits from the OALProtocolStack abstract class. Must always be first element in this struct. */

    OALStream               fragment;               /* Storage of the fragmented data received. */
    DOFAddress              connectedAddr;
    OALChannel              associatedChannel;      /* A reference to this stack's associated channel. We do not destroy this. */

    /***** Protocol Layers ************************************/
    ENP_V1      enpV1;
    EPP_V2      eppV2;
    ESP_V2      ccpV1;
    OAP_V1      oapV1;
    EPP_NegOne  eppNegOne;
    TEP_V1      tepV1;
    TRP_V1      trpV1;
    /**********************************************************/

    ConfigOptionList    myNegotiatedOpts;       /* This is the list of configuration options that "we" have negotiated for. */
    ConfigOptionList    theirNegotiatedOpts;    /* This is the list of configuration options that "they" have negotiated for. */

    DOFSecurityDesire   secDes;
    union
    {
        struct
        {
            unsigned int amClientNotServer      : 1;
            unsigned int requestOAP             : 1;
            unsigned int rejectedOAP            : 1;
            unsigned int requestTRP             : 1;
            unsigned int rejectedTRP            : 1;
            unsigned int requestTEP             : 1;
            unsigned int requireTEP             : 1;
            unsigned int rejectedTEP            : 1;
            unsigned int requestSEC             : 1;
        } bit;
        uint16 byte;
    } flags;                                /**< Flags and state of server. **/

    uint8                   stackState;     /* The state of this stack. Must be one of the CONNECTEDSTACK_STATE_X's above. */
} OALConnectedProtocolStack_t;

typedef void (*OCPS_MessageHandler_fn_t)( OALConnectedProtocolStack self, OALMessage message );

static void OALConnectedProtocolStack_Dtor( OALProtocolStack selfx );
static void OALConnectedProtocolStack_HandleMessage( OALProtocolStack selfx, OALMessage message );
static ProtocolLayerResult OALConnectedProtocolStack_EPP_ENP_Send_fn(OALProtocolStack selfx, BlackboardData blackboardData);

static boolean OALConnectedProtocolStack_Send(OALConnectedProtocolStack self, ProtocolLayerIndicator_t protocolLayerIndicator, uint16 attributeData, uint32 argc, void* argv[]);

static void OALConnectedProtocolStack_HandleReceiveData(OALConnectedProtocolStack self, OALStream data);

static void OALConnectedProtocolStack_ResolveCredsCallback( DOFCredentials self,
                                                            DOFException exception,
                                                            void* context );

static void OALConnectedProtocolStack_GetLibraryPermsCallback(DOFError           status,
                                                              sid_t            * secureSid,
                                                              DOFPermissionSet   outPerms,
                                                              void             * context);

static void* OALConnectedProtocolStack_TEP_V1_SessionEventCallback(void               * context,
                                                                   uint8              * key,
                                                                   TEP_SessionEvent_t   eventType,
                                                                   DOFSecurityMode      securityMode,
                                                                   OALChannel           channel,
                                                                   DOFSecurityScope     secScope,
                                                                   DOFPermissionSet     localPerms,
                                                                   DOFPermissionSet     otherPerms,
                                                                   DOFException         exception,
                                                                   uint32               stateID,
                                                                   DOFAddress           remoteAddr,
                                                                   OALStream          * initialState,
                                                                   DOFCredentials       creds);

static ConfigOptionList createRequestOptionsList(OALConnectedProtocolStack self, boolean supportOldVersion);
ESP_V2_CfgHandlerResponse_t configRequestHandler(void* protocolStack, ConfigOptionList cfgRequestOpts);
ESP_V2_CfgHandlerResponse_t configNegAckHandler(void* protocolStack, ConfigOptionList cfgNegAckOpts);
ESP_V2_CfgHandlerResponse_t configRejectHandler(void* protocolStack, ConfigOptionList cfgRejectOpts);
static void OALConnectedProtocolStack_DoPostNegotiationStuff(OALConnectedProtocolStack self);
static void OALConnectedProtocolStack_HandleReceiveError_Shutdown(OALConnectedProtocolStack self);

OALConnectedProtocolStack OALConnectedProtocolStack_Create( DOFAddress connected_addr )
{
    OALConnectedProtocolStack retVal = (OALConnectedProtocolStack)DOFMem_Alloc( sizeof(OALConnectedProtocolStack_t) );

    if ( retVal )
    {
        retVal->superClass.shutdownMessageSent = FALSE;
        retVal->superClass.shutdownMessage = SMT_Shutdown_Create((void *)&retVal->superClass);
        if (retVal->superClass.shutdownMessage)
        {
            OALProtocolStack_Ctor( &(retVal->superClass),
                                   OALConnectedProtocolStack_Dtor,
                                   OALConnectedProtocolStack_HandleMessage,
                                   OALConnectedProtocolStack_EPP_ENP_Send_fn);

            retVal->fragment = NULL;
            retVal->stackState = CONNECTEDSTACK_STATE_EPS_NEGOTIATION;
            retVal->connectedAddr = DOFAddress_Copy( connected_addr );
            if (!connected_addr || retVal->connectedAddr)
            {
                retVal->enpV1 = ENP_V1_Create( &(retVal->superClass.blackboardData), FALSE, TRUE);
                retVal->eppV2 = EPP_V2_Create( &(retVal->superClass.blackboardData), FALSE );
                retVal->ccpV1 = ESP_V2_Create( &(retVal->superClass.blackboardData), (void*)retVal, configRequestHandler, configNegAckHandler, configRejectHandler);
                retVal->oapV1 = NULL; /*OAP_V1_Create( &(retVal->superClass.blackboardData), FALSE );*/
                retVal->eppNegOne = EPP_NegOne_Create( &(retVal->superClass.blackboardData), FALSE );
                retVal->tepV1 = NULL; /*TEP_V1_Create( &(retVal->superClass.blackboardData), FALSE, OALConnectedProtocolStack_TEP_V1_SessionEventCallback );*/
                retVal->trpV1 = NULL; /*TRP_V1_Create( &(retVal->superClass.blackboardData), FALSE );*/

                if (retVal->enpV1 && retVal->eppV2 && retVal->ccpV1 && retVal->eppNegOne)
                {
                    /* SEC stuff is initialized after the channel is set */
                    retVal->flags.bit.requestOAP = TRUE;
                    retVal->flags.bit.rejectedOAP = FALSE;
                    retVal->flags.bit.requestTRP = TRUE;
                    retVal->flags.bit.rejectedTRP = FALSE;
                    retVal->flags.bit.requestTEP = FALSE;
                    retVal->flags.bit.requireTEP = FALSE;
                    retVal->flags.bit.rejectedTEP = FALSE;
                    retVal->flags.bit.requestSEC = FALSE;

                    retVal->myNegotiatedOpts = NULL;
                    retVal->theirNegotiatedOpts = NULL;
                    retVal->secDes = DOFSECURITYDESIRE_ANY;

                    return retVal;
                }
                if (retVal->enpV1)
                    ENP_V1_Destroy(retVal->enpV1);
                if (retVal->eppV2)
                    EPP_V2_Destroy(retVal->eppV2);
                if (retVal->ccpV1)
                    ESP_V2_Destroy(retVal->ccpV1);
                if (retVal->eppNegOne)
                    EPP_NegOne_Destroy(retVal->eppNegOne);
            }
            DOFMem_Free(retVal->superClass.shutdownMessage);
        }
        DOFMem_Free(retVal);
    }

    return NULL;
}

static void OALConnectedProtocolStack_Dtor( OALProtocolStack selfx )
{
    OALConnectedProtocolStack self = (OALConnectedProtocolStack)selfx;
    if ( !self )
        return;

    if ( self->theirNegotiatedOpts )
        ESPV2_ConfigOptionList_Destroy( self->theirNegotiatedOpts );
    if ( self->myNegotiatedOpts )
        ESPV2_ConfigOptionList_Destroy( self->myNegotiatedOpts );

    if ( self->trpV1 )
        TRP_V1_Destroy( self->trpV1 );
    if ( self->tepV1 )
        TEP_V1_Destroy( self->tepV1 );
    if ( self->eppNegOne )
        EPP_NegOne_Destroy( self->eppNegOne );
    if ( self->oapV1 )
        OAP_V1_Destroy( self->oapV1 );
    if ( self->ccpV1 )
        ESP_V2_Destroy( self->ccpV1 );
    if ( self->eppV2 )
        EPP_V2_Destroy( self->eppV2 );
    if ( self->enpV1 )
        ENP_V1_Destroy( self->enpV1 );

    if ( self->connectedAddr )
        DOFAddress_Destroy( self->connectedAddr );
    if ( self->fragment )
        OALStream_Release( self->fragment, TRUE );

    OALProtocolStack_Dtor( (OALProtocolStack)self );

    DOFMem_Free( self );
}

void OALConnectedProtocolStack_SetChannel( OALConnectedProtocolStack self, OALConnectedRemoteChannel remoteChannel )
{
    self->associatedChannel = (OALChannel)remoteChannel;
    self->secDes = (self->associatedChannel->remoteChannelData->flags.isConnection)?(DOFConnectionConfig_GetSecurityDesire(self->associatedChannel->remoteChannelData->config.connection)):
                                                                                    (DOFServerConfig_GetSecurityDesire(self->associatedChannel->remoteChannelData->config.server));
    self->flags.bit.amClientNotServer = OALConnectedRemoteChannel_GetIsClient( remoteChannel );
    self->flags.bit.requestOAP = self->associatedChannel->remoteChannelData->negotiator != DOF_DEFAULT_AS_PROTOCOL_NEGOTIATOR;
    self->flags.bit.requireTEP = self->secDes & DOFSECURITYDESIRE_SECURE_ANY;
    self->flags.bit.requestTEP = self->flags.bit.amClientNotServer && self->flags.bit.requireTEP;
    self->flags.bit.requestSEC = self->flags.bit.requestTEP;
    self->flags.bit.requestTRP = TRUE;
}

void OALConnectedProtocolStack_ClearChannel( OALConnectedProtocolStack self )
{
    self->associatedChannel = NULL;
}

OALChannel OALConnectedProtocolStack_GetChannel( OALConnectedProtocolStack self )
{
    return self->associatedChannel;
}

static void OALConnectedProtocolStack_ProcessShutdown( OALConnectedProtocolStack self, DOFException except )
{
    if (self->associatedChannel && except)
    {
        /* Send a shutdown message to the channel just in case this was a protocol stack error.
           The shutdown will only be sent to the channel once, so if the protocol stack is shutting down
           as a reslut of the channel shutting down this will do nothing. */
        OALRemoteChannel_Shutdown(self->associatedChannel, except);
    }
    OALRemoteChannelMgr_RemoveProtocolStack_Immediate(&self->superClass);
}

static void OALConnectedProtocolStack_HandleMessage_RouteToProvider( OALConnectedProtocolStack self, OALMessage message )
{
    ProtocolLayerIndicator_t    protocolLayerIndicator;
    DOFOperation                req_op;
    DOFOperation                prov_op;

    PCR_Assert( message->argc >= 2 );

    req_op      = message->argv[0];
    prov_op     = message->argv[1];

    protocolLayerIndicator = OALProtocolStack_DetermineProtocol(req_op->operation_type);

    switch ( protocolLayerIndicator )
    {
    case PLI_OAP:
        {
            void* args[4];
            /* args to OAP_V1_Send */
            args[0] = req_op;
            args[1] = prov_op;
            args[2] = (void*)TRUE;
            args[3] = NULL;

            OALConnectedProtocolStack_Send( self, protocolLayerIndicator, PROTOCOLID_OAP_V1, 4, args );
            goto CleanupAndExit;
        }
        break;

    case PLI_TEP:
        {
            void* args[3];
            /* args to TEP_V1_Send */
            args[0] = req_op;
            args[1] = (void*)TRUE;
            args[2] = NULL;

            OALConnectedProtocolStack_Send( self, protocolLayerIndicator, PROTOCOLID_TEP_V1, 3, args );
            goto CleanupAndExit;
        }
        break;

    case PLI_TRP:
        {
            void* args[3];
            args[0] = req_op;
            args[1] = (void*)TRUE;
            args[2] = NULL;

            OALConnectedProtocolStack_Send( self, protocolLayerIndicator, PROTOCOLID_TRP_V1, 3, args );
            goto CleanupAndExit;
        }
        break;
    case PLI_EPP_NEGONE:
        {
            OALConnectedProtocolStack_Send( self, protocolLayerIndicator, PROTOCOLID_EPP_NEGONE, 1, (void**)&req_op );
            goto CleanupAndExit;
        }
    default:

        break;
    }

    CleanupAndExit:
    return;
}

static void OALConnectedProtocolStack_HandleMessage_RouteToRequestor( OALConnectedProtocolStack self, OALMessage message )
{
    ProtocolLayerIndicator_t    protocolLayerIndicator;
    DOFOperation                operation;
    DOFResult                   result;

    PCR_Assert( message->argc >= 2 );

    operation = message->argv[0];
    result = message->argv[1];

    protocolLayerIndicator = OALProtocolStack_DetermineProtocol(operation->operation_type);

    switch ( protocolLayerIndicator )
    {
    case PLI_OAP:
        {
            void* args[4];
            /* args to OAP_V1_Send */
            args[0] = operation;
            args[1] = NULL;
            args[2] = (void*)FALSE;
            args[3] = result;

            if ( OALConnectedProtocolStack_Send( self, protocolLayerIndicator, PROTOCOLID_OAP_V1, 4, args ) )
            {
                /* Now that PDU is well along the way to going on the wire, cancel its operation if previously marked. */
                if (operation->flags.bits.cancelOpAfterEppSend)
                    DOFOperation_Cancel( operation );
            }
            goto CleanupAndExit;
        }
        break;
    case PLI_TRP:
        {
            void* args[3];
            args[0] = operation;
            args[1] = (void*)FALSE;
            args[2] = result;

            OALConnectedProtocolStack_Send( self, protocolLayerIndicator, PROTOCOLID_TRP_V1, 3, args );
            goto CleanupAndExit;
        }
        break;
    default:

        break;
    }

    CleanupAndExit:
    return;
}

static void OALConnectedProtocolStack_HandleMessage_ReceiveFromTransport( OALConnectedProtocolStack self, OALMessage message )
{
    if (!message->shutdown)
    {
        OALStream data;

        PCR_Assert( message->argc > 0 );

        data = (OALStream)message->argv[0];

        if (self->associatedChannel)
        {
            OALConnectedProtocolStack_HandleReceiveData( self, data );
        }
    }
}

static void OALConnectedProtocolStack_HandleMessage_BeginNegotiationAuthorization( OALConnectedProtocolStack self, OALMessage message )
{
    if (!message->shutdown)
    {
        if (!OALProtocolStack_StartNegotiationTimeout((OALProtocolStack)self, EPS_NEGOTIATION_TIMEOUT))
        {
            OALConnectedProtocolStack_ProcessShutdown(self, DOFEXCEPTION_FAILED);
            return;
        }

        if( self->flags.bit.amClientNotServer )
        {
            /* If this is the client start ESP negotiation. */
            OALConnectedProtocolStack_Send( self, PLI_ESP, PROTOCOLID_ESP, 0, NULL );
        }
    }
}

static void OALConnectedProtocolStack_HandleMessage_Shutdown( OALConnectedProtocolStack self, OALMessage message )
{
    OALConnectedProtocolStack_ProcessShutdown(self, (DOFException)message->argv[0]);
}

static void OALConnectedProtocolStack_HandleMessage_Negotiation_Timeout(OALConnectedProtocolStack self, OALMessage message)
{
    OALConnectedProtocolStack_ProcessShutdown(self, DOFEXCEPTION_TIMEOUT);
}

static void OALConnectedProtocolStack_HandleMessage_BeginNegotiationServerSide(OALConnectedProtocolStack self, OALMessage message)
{
    if (!message->shutdown)
    {
        if( !self->flags.bit.amClientNotServer )
        {
            OALConnectedProtocolStack_Send( self, PLI_ESP, PROTOCOLID_ESP, 0, NULL );
        }
    }
}

static void OALConnectedProtocolStack_HandleMessage_RemoteDomainCallback(OALConnectedProtocolStack self, OALMessage message)
{
    /* Todo: At one point there was a seperate check for resolving credentials and checking for a remote domain.
       Now, resolving credentials will take care of both, so at some point the RemoteDomainCallback can be removed. */
    if (self->associatedChannel)
    {
        if (message->argv[0])
        {
            self->associatedChannel->remoteChannelData->secureCredentials = (DOFCredentials)message->argv[0];
            message->argv[0] = NULL;
            OALRemoteChannel_CreateLibPermissions(OALConnectedProtocolStack_GetLibraryPermsCallback, self,
                                                  self->associatedChannel->remoteChannelData->secureCredentials,
                                                  self->associatedChannel, self->associatedChannel->remoteChannelData->address);
        }
        else
        {
            /* If credentials are not included, just shut down the channel. */
            OALRemoteChannel_Shutdown(self->associatedChannel, DOFEXCEPTION_AUTHENTICATION_FAILED);
        }
    }
}

static void OALConnectedProtocolStack_HandleMessage_ResolveCredsCallback(OALConnectedProtocolStack self, OALMessage message)
{
    if (self->associatedChannel)
    {
        if (message->argv[0])
        {
            self->associatedChannel->remoteChannelData->secureCredentials = (DOFCredentials)message->argv[0];
            message->argv[0] = NULL;
            OALRemoteChannel_CreateLibPermissions(OALConnectedProtocolStack_GetLibraryPermsCallback, self,
                                                  self->associatedChannel->remoteChannelData->secureCredentials,
                                                  self->associatedChannel, self->associatedChannel->remoteChannelData->address);
        }
        else
        {
            /* If credentials are not included, just shut down the channel. */
            OALRemoteChannel_Shutdown(self->associatedChannel, DOFEXCEPTION_AUTHENTICATION_FAILED);
        }
    }
}

static void OALConnectedProtocolStack_HandleMessage_GetLibraryPermsCallback(OALConnectedProtocolStack self, OALMessage message)
{
    if (self->associatedChannel)
    {
        if (message->argv[1])
        {
            self->associatedChannel->remoteChannelData->secureSid = (sid_t *)message->argv[1];
            message->argv[1] = NULL;

            TEP_V1_Authenticate( self->tepV1, self->flags.bit.amClientNotServer, self, self->associatedChannel,
                                 self->associatedChannel->remoteChannelData->address, (DOFPermissionSet)message->argv[0] );
        }
        else
        {
            OALRemoteChannel_Shutdown(self->associatedChannel, DOFEXCEPTION_AUTHENTICATION_FAILED);
        }
    }
}

static void OALConnectedProtocolStack_HandleMessage_RouteSavedEPP(OALConnectedProtocolStack self, OALMessage message)
{
    BlackboardData blackboardData = &(self->superClass.blackboardData);

    BlackboardData_Restore(blackboardData, (BlackboardData)message->argv[0]);
    /* The blackboard data in the stack now owns the objects in blackboardData so call
       blackboardData_Init so that the objects will not be destroyed when the message is destroyed. */
    BlackboardData_Init((BlackboardData)message->argv[0]);

    /* This is hard coded for EPP_V2 and ENP_V1 and must be updated when different or multiple versions are added. */
    if (self->superClass.ENP_EPP_send_func(&self->superClass, blackboardData) == PLRESULT_OKSENDDOWN)
    {
        if (DOFTransportLayer_SendTo(blackboardData->associatedChannel->remoteChannelData->dtl, blackboardData->sendPacket, blackboardData->transportAddr))
        {
            if (blackboardData->associatedChannel->remoteChannelData->maxSendSilenceTimerItem)
            {
                OALRemoteChannelMgr_SetSendSilenceTimerItem(blackboardData->associatedChannel);
            }
        }
    }
    BlackboardData_Clear(blackboardData);
}

static void OALConnectedProtocolStack_HandleMessage( OALProtocolStack selfx, OALMessage message )
{
    OALConnectedProtocolStack self = (OALConnectedProtocolStack)selfx;
    static OCPS_MessageHandler_fn_t msg_handler_funcs[] = {
        /* the order here is coupled with StackMsgType_t. So dont change it. */
        OALConnectedProtocolStack_HandleMessage_RouteToProvider,
        OALConnectedProtocolStack_HandleMessage_RouteToRequestor,
        OALConnectedProtocolStack_HandleMessage_ReceiveFromTransport,
        OALConnectedProtocolStack_HandleMessage_BeginNegotiationAuthorization,
        OALConnectedProtocolStack_HandleMessage_Shutdown,
        OALConnectedProtocolStack_HandleMessage_Negotiation_Timeout,
        OALConnectedProtocolStack_HandleMessage_BeginNegotiationServerSide,
        OALConnectedProtocolStack_HandleMessage_RemoteDomainCallback,
        OALConnectedProtocolStack_HandleMessage_ResolveCredsCallback,
        OALConnectedProtocolStack_HandleMessage_GetLibraryPermsCallback,
        OALConnectedProtocolStack_HandleMessage_RouteSavedEPP
    };

    PCR_Assert( self );
    PCR_Assert( message );
    PCR_Assert( message->recipient == self );
    PCR_Assert( message->msg_type < (sizeof(msg_handler_funcs) / sizeof(OCPS_MessageHandler_fn_t)) );

    msg_handler_funcs[message->msg_type]( self, message );
}

static ProtocolLayerResult OALConnectedProtocolStack_EPP_ENP_Send_fn(OALProtocolStack selfx, BlackboardData blackboardData)
{
    OALConnectedProtocolStack self = (OALConnectedProtocolStack)selfx;
    ProtocolLayerResult retStatus;

    retStatus = EPP_V2_Send(self->eppV2);
    if ( retStatus != PLRESULT_OKSENDDOWN )
        return retStatus;

    retStatus = ENP_V1_Send(self->enpV1);

    return retStatus;
}

static boolean OALConnectedProtocolStack_Send(OALConnectedProtocolStack self, ProtocolLayerIndicator_t protocolLayerIndicator, uint16 attributeData, uint32 argc, void* argv[])
{
    boolean rval = FALSE;
    BlackboardData blackboardData = &(self->superClass.blackboardData);

    BlackboardData_Init( blackboardData );

    blackboardData->sendPacket = OALPacket_Create();
    if ( !blackboardData->sendPacket )
    {
        goto CleanupAndExit;
    }

    blackboardData->associatedChannel = OALChannel_Retain( self->associatedChannel );
    if (self->associatedChannel && !blackboardData->associatedChannel)
    {
        goto CleanupAndExit;
    }

    switch ( protocolLayerIndicator )
    {
    case PLI_OAP:
        PCR_Assert( argc == 4 );
        blackboardData->providerOp = (DOFOperation)argv[1];
        if (blackboardData->providerOp)
            DOFOperation_Retain(blackboardData->providerOp);
        if ( self->stackState == CONNECTEDSTACK_STATE_OPEN )
        {
            if ( !self->oapV1 || OAP_V1_Send(self->oapV1, (DOFOperation)argv[0], (boolean)((int)argv[2]), (DOFResult)argv[3]) != PLRESULT_OKSENDDOWN )
                goto CleanupAndExit;
        }
        else
            goto CleanupAndExit;

        break;

    case PLI_TEP:
        PCR_Assert( argc == 3 );
        if (self->stackState >= CONNECTEDSTACK_STATE_AUTH_NEGOTIATION)
        {
            if ( !self->tepV1 || TEP_V1_Send(self->tepV1, (DOFOperation)argv[0], (boolean)((int)argv[1]), (DOFResult)argv[2]) != PLRESULT_OKSENDDOWN )
                goto CleanupAndExit;
        }
        else
            goto CleanupAndExit;
        break;

    case PLI_TRP:
        PCR_Assert( argc == 3 );
        if (self->stackState >= CONNECTEDSTACK_STATE_AUTH_NEGOTIATION)
        {
            if ( !self->trpV1 || TRP_V1_Send(self->trpV1, (DOFOperation)argv[0], (boolean)((int)argv[1]), (DOFResult)argv[2]) != PLRESULT_OKSENDDOWN )
                goto CleanupAndExit;
        }
        else
            goto CleanupAndExit;
        break;

    case PLI_ESP:
        {
            ProtocolLayerResult result = PLRESULT_ERROR;

            if (self->stackState == CONNECTEDSTACK_STATE_EPS_NEGOTIATION)
            {
                /* The client will send the first ccp negotiation packet which ends eps negotiation. */
                self->stackState = CONNECTEDSTACK_STATE_APP_NEGOTIATION;
                if (! OALProtocolStack_CancelNegotiationTimer((OALProtocolStack)self) || !OALProtocolStack_StartNegotiationTimeout((OALProtocolStack)self, ESP_NEGOTIATION_TIMEOUT))
                {
                    OALConnectedProtocolStack_HandleReceiveError_Shutdown(self);
                    goto CleanupAndExit;
                }
            }

            {
                ConfigOptionList optsToSend = createRequestOptionsList( self, FALSE );
                if ( optsToSend )
                {
                    /* send ccp configuration request. */
                    result = ESP_V2_SendInitialNegotiation( self->ccpV1, optsToSend );
                    if ( self->myNegotiatedOpts )
                        ESPV2_ConfigOptionList_Destroy( self->myNegotiatedOpts );
                    self->myNegotiatedOpts = optsToSend;
                }
            }
            if ( result != PLRESULT_OKSENDDOWN )
                goto CleanupAndExit;
        }
        break;

    case PLI_EPP_NEGONE:
        if ( EPP_NegOne_Send(self->eppNegOne, (DOFOperation)argv[0]) != PLRESULT_OKSENDDOWN )
            goto CleanupAndExit;

        break;

    default:

        break;
    }

#ifdef OALPROTOCOLSTACK_PACKET_TRACE
    {
        uint32 i;
        uint32 theLen = OALPacket_GetTotalStreamSize(blackboardData->sendPacket);
        uint8* ptr = DOFMem_Alloc(theLen);

        OALPacket_CopyTo(blackboardData->sendPacket, ptr, 0, theLen);

        printf("Send: ");
        for(i = 0; i<theLen; ++i)
        {
            printf("0x%02X ", ptr[i]);
        }
        printf("\n");

        DOFMem_Free(ptr);
    }
#endif

    /* Verify that streaming is allowed. */
    if (blackboardData->flags.streamingProtocolNotAllowed)
        goto CleanupAndExit;

/*EPP_Send:*/
    if ( EPP_V2_Send(self->eppV2) != PLRESULT_OKSENDDOWN )
        goto CleanupAndExit;

/*ENP_Send:*/
    if ( ENP_V1_Send(self->enpV1) != PLRESULT_OKSENDDOWN )
        goto CleanupAndExit;

    /* Send to Transport Layer */
    if(DOFTransportLayer_SendTo( self->associatedChannel->remoteChannelData->dtl, blackboardData->sendPacket, NULL ))
    {
        rval = TRUE;
        /* We just sent a packet, so if there is a maxSendSilenceTimerItem reset it. */
        if(self->associatedChannel->remoteChannelData->maxSendSilenceTimerItem)
        {
            OALRemoteChannelMgr_SetSendSilenceTimerItem(self->associatedChannel);
        }
    }
    else
    {
        OALRemoteChannel_Shutdown(self->associatedChannel, DOFEXCEPTION_FAILED);
    }

    CleanupAndExit:
    if ( self->superClass.keyChangeCallback )
    {
        if ( !self->superClass.keyChangeCallback( self->superClass.keyChangeCallbackContext ) )
            OALConnectedProtocolStack_HandleReceiveError_Shutdown( self );
        self->superClass.keyChangeCallback = NULL;
        self->superClass.keyChangeCallbackContext = NULL;
    }
    BlackboardData_Clear( &(self->superClass.blackboardData) );

    return rval;
}

static void OALConnectedProtocolStack_HandleReceiveData(OALConnectedProtocolStack self, OALStream data)
{
    uint32                  appProtocolID;
    uint32                  bufLength;
    ProtocolLayerResult     result;
    BlackboardData          blackboardData = &(self->superClass.blackboardData);
    uint32                  dataLen = OALStream_GetDataSize(data);
    OALChannel              origChannel = self->associatedChannel;

    PCR_Assert(self->associatedChannel);
    PCR_Assert(self->associatedChannel->remoteChannelData);

    /* handle previously received fragment */
    if ( self->fragment )
    {
        OALStream temp = OALStream_Create_Concat(self->fragment, data, 0); /* concatenate data to the end of the fragment */
        if ( !temp )
        {
            OALConnectedProtocolStack_HandleReceiveError_Shutdown( self );
            goto CleanupAndExit;
        }
        data = temp;
        OALStream_Release( self->fragment, TRUE );
        self->fragment = NULL;
    }
    else
        OALStream_Retain( data );

    /* There may be several "packets" in this stream so loop. */
    for ( ;; )
    {
        /* We need at least 2 bytes to do anything worth while. */
        switch ( OALStream_GetDataSize( data ) )
        {
            case 1:
                /* store fragment and return. */
                self->fragment = data;
                OALStream_Retain( self->fragment );
                /* Fall through */
            case 0:
                goto CleanupAndExit;
            default:
                break;
        }

        BlackboardData_Clear( blackboardData );

        OALStream_Retain( data );
        blackboardData->receiveStream = data;
        blackboardData->flags.isMulticast = FALSE;
        blackboardData->transportAddr = DOFAddress_Copy( self->connectedAddr );
        if ( !blackboardData->transportAddr )
        {
            OALConnectedProtocolStack_HandleReceiveError_Shutdown( self );
            goto CleanupAndExit;
        }

        blackboardData->associatedChannel = OALChannel_Retain( self->associatedChannel );
        if (!blackboardData->associatedChannel)
        {
            OALConnectedProtocolStack_HandleReceiveError_Shutdown( self );
            goto CleanupAndExit;
        }

        blackboardData->sendPacket = OALPacket_Create();
        if ( !blackboardData->sendPacket )
        {
            OALConnectedProtocolStack_HandleReceiveError_Shutdown( self );
            goto CleanupAndExit;
        }

        if ( self->stackState == CONNECTEDSTACK_STATE_EPS_NEGOTIATION )
        {/* Handle EPS Negotiation. */
            const uint8* pNegotiationReq = (const uint8 *)OALStream_GetReadPtr(blackboardData->receiveStream);
            const uint16 negotiationReq = (*pNegotiationReq << 8) + *(pNegotiationReq+1);

            /* Determine if this is a negotiation request. */
            if ( !(negotiationReq & 0x8080) ) /* check for no ENP/EPP flag bits set */
            {
                /*  Due to our state we must assume this is a negotiation request.  Because we only support 1 ENP version
                    and 1 EPP version we will respond by asserting those versions by sending an EPP heartbeat.
                    Even though we assert our versions here, the EPS negotation is not complete unitl the client
                    sends the first non-negotiation packet. */
                OALStream_AdjustReadOffset(blackboardData->receiveStream, 2);
                EPP_NegOne_Heartbeat(blackboardData->associatedChannel, blackboardData->transportAddr, FALSE);
                self->stackState = CONNECTEDSTACK_STATE_APP_NEGOTIATION;
                continue;
            }
            else if ( (negotiationReq & 0x8000) == 0x8000 ) /* check that the flag bit of the ENP header is set. */
            {
                /* At this point this is either a version-asserted packet. Or a packet in error.
                   If the latter case holds then the error will be caught by the protocol layers. */
                self->stackState = CONNECTEDSTACK_STATE_APP_NEGOTIATION;
                if (! OALProtocolStack_CancelNegotiationTimer((OALProtocolStack)self) || !OALProtocolStack_StartNegotiationTimeout((OALProtocolStack)self, ESP_NEGOTIATION_TIMEOUT))
                {
                    OALConnectedProtocolStack_HandleReceiveError_Shutdown(self);
                    goto CleanupAndExit;
                }
            }
            else
            {
                /*  The flag bit of the ENP header was not set. but the EPP flag bit was.
                    Since this is not a valid version negotiation close the connection. */
                OALConnectedProtocolStack_HandleReceiveError_Shutdown( self );
                goto CleanupAndExit;
            }
        }
        else
        {
            const uint8* readPtr = OALStream_GetReadPtr(blackboardData->receiveStream);

            /*  If the eps negotiation is complete and the flag bit for ENP is not set,
                the connection must be closed. (ENP.3) */
            if (!(*readPtr & 0x80))
            {
                OALConnectedProtocolStack_HandleReceiveError_Shutdown( self );
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectedProtocolsStack: Flag bit was not set on a non-negotiation datagram.");
                goto CleanupAndExit;
            }
        }

        /* Process ENP */
        switch ( ENP_V1_Receive(self->enpV1) )
        {
        case PLRESULT_ERROR:
            OALConnectedProtocolStack_HandleReceiveError_Shutdown( self );
            goto CleanupAndExit;
        case PLRESULT_NEEDMOREDATA:
            /* store fragment and return. */
            self->fragment = blackboardData->receiveStream;
            OALStream_Retain( self->fragment );
            goto CleanupAndExit;
        case PLRESULT_OKDONOTHING:
            continue;
        case PLRESULT_OKSENDUP:
            break; /* All we have to do here to send data up to EPP is just not return. */
        default:
            PCR_Assert(FALSE); /* Should never ever get here. */
            break;
        }

        /* Process EPP */
        switch ( EPP_V2_Receive(self->eppV2) )
        {
        case PLRESULT_ERROR:
            OALConnectedProtocolStack_HandleReceiveError_Shutdown( self );
            goto CleanupAndExit;
        case PLRESULT_OKNOMOREDATA:
            goto CleanupAndContinue;
        case PLRESULT_OKSENDUP:
            break; /* All we have to do here to send data up (to an application layer) is just not return. */

        case PLRESULT_OKDONOTHING:
            continue;

        case PLRESULT_OKSENDDOWN:
            if ( ENP_V1_Send(self->enpV1) != PLRESULT_OKSENDDOWN )
            {
                goto CleanupAndContinue;
            }

            /* Send to Transport Layer */
            if(DOFTransportLayer_SendTo(self->associatedChannel->remoteChannelData->dtl, blackboardData->sendPacket, NULL))
            {
                /* We just sent a packet, so if there is a maxSendSilenceTimerItem reset it. */
                if(self->associatedChannel->remoteChannelData->maxSendSilenceTimerItem)
                {
                    OALRemoteChannelMgr_SetSendSilenceTimerItem(self->associatedChannel);
                }
            }
            else
            {
                OALConnectedProtocolStack_HandleReceiveError_Shutdown(self);
                goto CleanupAndExit;
            }
            continue;

        default:
            PCR_Assert(FALSE); /* Should never ever get here. */
            break;
        }

#ifdef OALPROTOCOLSTACK_PACKET_TRACE
        {
            uint32 i;
            uint32 theLen = OALStream_GetDataSize(blackboardData->receiveStream);
            const uint8* ptr = OALStream_GetReadPtr(blackboardData->receiveStream);

            printf("Recv: ");

            for(i = 0; i<theLen; ++i)
            {
                printf("0x%02X ", ptr[i]);
            }

            printf("\n");
        }
#endif

        /* peek at the AppID, doesnt modify the stream */
        bufLength = OALStream_GetDataSize(blackboardData->receiveStream);
        appProtocolID = OALMarshal_UncompressValue( 2 /* AppID can be 2 bytes */, &bufLength, OALStream_GetReadPtr(blackboardData->receiveStream));
        if (appProtocolID == OALMARSHAL_INVALID_VALUE)
        {
            OALConnectedProtocolStack_HandleReceiveError_Shutdown( self );
            goto CleanupAndExit;
        }

        if ( (self->stackState == CONNECTEDSTACK_STATE_APP_NEGOTIATION) && (appProtocolID != PROTOCOLID_EPP_NEGONE) )
        {
            /* Process ESP */
            switch ( ESP_V2_TCP_Receive(self->ccpV1) )
            {
            case PLRESULT_ERROR:
            case PLRESULT_CLOSE_CONNECTION:
                OALConnectedProtocolStack_HandleReceiveError_Shutdown( self );
                goto CleanupAndExit;

            case PLRESULT_OKDONOTHING:
                continue; /* Do nothing. */

            case PLRESULT_NEGOTIATIONCOMPLETE:
                if (OALProtocolStack_CancelNegotiationTimer((OALProtocolStack)self))
                    OALConnectedProtocolStack_DoPostNegotiationStuff(self);
                else
                {
                    OALConnectedProtocolStack_HandleReceiveError_Shutdown(self);
                    goto CleanupAndExit;
                }
                continue;

            case PLRESULT_SENDDOWN_THENNEGOTIATIONCOMPLETE:
                if (OALProtocolStack_CancelNegotiationTimer((OALProtocolStack)self))
                    OALConnectedProtocolStack_DoPostNegotiationStuff(self);
                else
                {
                    OALConnectedProtocolStack_HandleReceiveError_Shutdown(self);
                    goto CleanupAndExit;
                }

                /* FALL THROUGH INTENDED */
            case PLRESULT_OKSENDDOWN:
                {
                    ProtocolLayerResult plRes;

                    /* Send the response down through the stack. */
                    plRes = EPP_V2_Send(self->eppV2);
                    if ( plRes != PLRESULT_OKSENDDOWN )
                    {
                        if (plRes != PLRESULT_OKDONOTHING)
                        {
                            OALConnectedProtocolStack_HandleReceiveError_Shutdown( self );
                        }
                        goto CleanupAndExit;
                    }

                    plRes = ENP_V1_Send(self->enpV1);
                    if ( plRes != PLRESULT_OKSENDDOWN )
                    {
                        if (plRes != PLRESULT_OKDONOTHING)
                        {
                            OALConnectedProtocolStack_HandleReceiveError_Shutdown( self );
                        }
                        goto CleanupAndExit;
                    }

                    /* Send to Transport Layer */
                    if(DOFTransportLayer_SendTo(self->associatedChannel->remoteChannelData->dtl, blackboardData->sendPacket, NULL))
                    {
                        /* We just sent a packet, so if there is a maxSendSilenceTimerItem reset it. */
                        if(self->associatedChannel->remoteChannelData->maxSendSilenceTimerItem)
                        {
                            OALRemoteChannelMgr_SetSendSilenceTimerItem(self->associatedChannel);
                        }
                    }
                    else
                    {
                        OALConnectedProtocolStack_HandleReceiveError_Shutdown(self);
                        goto CleanupAndExit;
                    }
                }
                continue;

            default:
                PCR_Assert(FALSE); /* Should never ever get here. */
                break;
            }
        }

        switch ( appProtocolID )
        {
        case PROTOCOLID_OAP_V1:
            if ( self->stackState == CONNECTEDSTACK_STATE_OPEN )
                result = (self->oapV1) ? OAP_V1_Receive(self->oapV1) : PLRESULT_ERROR;
            else
                result = PLRESULT_ERROR;
            break;

        case PROTOCOLID_EPP_NEGONE:
            result = EPP_NegOne_Receive(self->eppNegOne);
            break;

        case PROTOCOLID_TEP_V1:
            if ( self->stackState >= CONNECTEDSTACK_STATE_AUTH_NEGOTIATION )
                result = (self->tepV1) ? TEP_V1_Receive(self->tepV1) : PLRESULT_ERROR;
            else
                result = PLRESULT_ERROR;
            break;

        case PROTOCOLID_TRP_V1:
            if ( self->stackState >= CONNECTEDSTACK_STATE_AUTH_NEGOTIATION )
                result = (self->trpV1) ? TRP_V1_Receive(self->trpV1) : PLRESULT_ERROR;
            else
                result = PLRESULT_ERROR;
            break;

        default:
            result = PLRESULT_ERROR;
            break;
        }

        /* Process result */
        switch ( result )
        {
        case PLRESULT_ERROR:
            OALConnectedProtocolStack_HandleReceiveError_Shutdown( self );
            goto CleanupAndExit;

        case PLRESULT_OKSENDDOWN:
            if ( EPP_V2_Send(self->eppV2) != PLRESULT_OKSENDDOWN )
                goto CleanupAndContinue;

            if ( ENP_V1_Send(self->enpV1) != PLRESULT_OKSENDDOWN )
                goto CleanupAndContinue;

            /* Send to Transport Layer */
            if(DOFTransportLayer_SendTo( blackboardData->associatedChannel->remoteChannelData->dtl, blackboardData->sendPacket, blackboardData->transportAddr))
            {
                /* We just sent a packet, so if there is a maxSendSilenceTimerItem reset it. */
                if(blackboardData->associatedChannel->remoteChannelData->maxSendSilenceTimerItem)
                {
                    OALRemoteChannelMgr_SetSendSilenceTimerItem(blackboardData->associatedChannel);
                }
            }
            else
            {
                OALConnectedProtocolStack_HandleReceiveError_Shutdown(self);
                goto CleanupAndExit;
            }
            continue;

        case PLRESULT_OKDONOTHING:
            break; /* Do nothing. */

        default:
            PCR_Assert(FALSE); /* Should never ever get here. */
            break;
        }

        CleanupAndContinue:
        if ( self->superClass.keyChangeCallback )
        {
            boolean success = self->superClass.keyChangeCallback( self->superClass.keyChangeCallbackContext );
            self->superClass.keyChangeCallback = NULL;
            self->superClass.keyChangeCallbackContext = NULL;
            if ( !success )
            {
                OALConnectedProtocolStack_HandleReceiveError_Shutdown( self );
                goto CleanupAndExit;
            }
        }

    } /* while(dataToProcess) */

    CleanupAndExit:
    /* This is only done here for connected channels since connected channels
       don't change but in the future it may. When that happens this should be
       called when proper channel is set and dataLen should be set to 0. This
       will only handle error cases then. */
    if( dataLen > 0 )
    {
        OALRemoteChannel_IncByteCount(origChannel, dataLen);
    }

    BlackboardData_Clear( blackboardData );
    OALStream_Release( data, TRUE );

    if ( self->superClass.keyChangeCallback )
    {
        if ( !self->superClass.keyChangeCallback( self->superClass.keyChangeCallbackContext ) )
            OALConnectedProtocolStack_HandleReceiveError_Shutdown( self );
        self->superClass.keyChangeCallback = NULL;
        self->superClass.keyChangeCallbackContext = NULL;
    }

    return;
}

static ConfigOptionList createRequestOptionsList(OALConnectedProtocolStack self, boolean supportOldVersion)
{
    DOFSecurityMode mode = self->associatedChannel->remoteChannelData->mode;
    ConfigOptionList    retVal = OALPtrList_Create(3);
    ESPV2_ConfigOption  opt;
    uint8 i = 0;

    if ( retVal )
    {
        boolean status = TRUE;

        if( self->flags.bit.amClientNotServer && self->flags.bit.requestOAP )
        {
            opt = ESPV2_ConfigOption_Create(ESP_ATTR_TYPE_OAP, PROTOCOLID_OAP_V1, 0);
            if (opt)
            {
                if (!OALPtrList_InsertAt(retVal, opt, i++))
                {
                    ESPV2_ConfigOption_Destroy(opt);
                    status = FALSE;
                }
            }
            else
            {
                status = FALSE;
            }
        }

        if ( status && self->flags.bit.requestTRP )
        {
            opt = ESPV2_ConfigOption_Create(ESP_ATTR_TYPE_TRP, PROTOCOLID_TRP_V1, 0);
            if (opt)
            {
                if (!OALPtrList_InsertAt(retVal, opt, i++))
                {
                    ESPV2_ConfigOption_Destroy(opt);
                    status = FALSE;
                }
            }
            else
            {
                status = FALSE;
            }
        }

        // supportOldVersion: This is a work around to allow C to be a secure server with some old versions of the other libraries that are incorrect.
        if ( status && (self->flags.bit.amClientNotServer || supportOldVersion) && self->flags.bit.requestTEP )
        {
            opt = ESPV2_ConfigOption_Create(ESP_ATTR_TYPE_TEP, PROTOCOLID_TEP_V1, 0);
            if (opt)
            {
                if (!OALPtrList_InsertAt(retVal, opt, i++))
                {
                    ESPV2_ConfigOption_Destroy(opt);
                    status = FALSE;
                }

                if( status )
                {
                    if( mode )
                        opt = ESPV2_ConfigOption_CreateFromMode(mode);
                    else
                        opt = ESPV2_ConfigOption_CreateSec(self->secDes);
                    if( opt )
                    {
                        if (!OALPtrList_InsertAt(retVal, opt, i++))
                        {
                            ESPV2_ConfigOption_Destroy(opt);
                            status = FALSE;
                        }
                    }
                    else
                    {
                        status = FALSE;
                    }
                }
            }
            else
            {
                status = FALSE;
            }
        }

        if (status == FALSE)
        {
            ESPV2_ConfigOptionList_Destroy(retVal);
            retVal = NULL;
        }
    }

    return retVal;
}

/* ****** ESP Configuration Handler Callbacks ***** */
ESP_V2_CfgHandlerResponse_t configRequestHandler(void* protocolStack, ConfigOptionList cfgRequestOpts)
{
    OALConnectedProtocolStack self = (OALConnectedProtocolStack)protocolStack;
    DOFProtocolNegotiator protocolNeg = self->associatedChannel->remoteChannelData->negotiator;
    DOFSecurityDesire secDes = self->secDes;
    uint32 optCount;

    /* Pointer list will grow as needed so start with a capacity of 3. */
    ConfigOptionList nackList = OALPtrList_Create(3);
    uint32 nackListCount = 0;
    boolean wasOapRequested = FALSE;
    boolean wasTepRequested = FALSE;

    boolean rejectTEP = FALSE;
    boolean suggestTEP = FALSE;
    boolean rejectTRP = FALSE;

    RejectList rejectList;
    uint32 rejectListCount = 0;

    ESP_V2_CfgHandlerResponse_t retVal;
    uint32 i;

    PCR_Assert(cfgRequestOpts != NULL);
    optCount = OALPtrList_GetCount(cfgRequestOpts);
    rejectList = OALPtrList_Create(optCount);
    if ( !nackList || !rejectList )
    {
        goto EXIT_CLOSE_CONNECTION;
    }
    if( optCount > 0 ) /* currently server can just ack */
    {
        for ( i = 0; i < optCount; ++i )
        {
            ESPV2_ConfigOption opt = OALPtrList_PeekAt(cfgRequestOpts, i);

            switch ( opt->attributeCode )
            {
            case ESP_ATTR_TYPE_TEP:
                if ( opt->attributeData == PROTOCOLID_TEP_V1 )
                {
                    if ( self->flags.bit.amClientNotServer || secDes == DOFSECURITYDESIRE_NOT_SECURE )
                    {
                        if( !self->flags.bit.rejectedTEP )
                        {
                            rejectTEP = TRUE;
                        }
                        else
                        {
                            goto EXIT_CLOSE_CONNECTION;
                        }
                    }
                    else
                    {
                        wasTepRequested = TRUE;
                    }
                }
                else
                {
                    if( !self->flags.bit.rejectedTEP )
                    {
                        rejectTEP = TRUE;
                        if( !self->flags.bit.amClientNotServer && secDes != DOFSECURITYDESIRE_NOT_SECURE )
                        {
                            suggestTEP = TRUE;
                        }

                    }
                }
                break;
            case ESP_ATTR_TYPE_OAP:
                if( opt->attributeData == PROTOCOLID_OAP_V1 )
                {
                    wasOapRequested = TRUE;
                }
                else
                {
                    if( !self->flags.bit.rejectedOAP )
                    {
                        ESPV2_Reject cfgR = ESPV2_OAPReject_Create();

                        if (!cfgR)
                        {
                            goto EXIT_CLOSE_CONNECTION;
                        }
                        if (!OALPtrList_InsertAt(rejectList, cfgR, rejectListCount++))
                        {
                            ESPV2_Reject_Destroy(cfgR);
                            goto EXIT_CLOSE_CONNECTION;
                        }
                    }
                    else
                    {
                        goto EXIT_CLOSE_CONNECTION;
                    }
                }
                break;
            case ESP_ATTR_TYPE_CCM:
                /* mode we understand */
                if( opt->attributeData == PROTOCOLID_CCMMODE_V1 && secDes != DOFSECURITYDESIRE_NOT_SECURE)
                {
                    /* security allowed */
                    boolean nakSec = FALSE;
                    DOFSecurityMode sec = ESPV2_ConfigOption_SecModeFromConfigOption( opt, secDes, &nakSec);
                    if( sec )
                    {
                        if( !nakSec )
                        {
                            /* Mode Acceptable We Shall Use It! */
                            DOFSecurityMode_Destroy( self->associatedChannel->remoteChannelData->mode );
                            self->associatedChannel->remoteChannelData->mode = sec;
                        }
                        else
                        {
                            ESPV2_ConfigOption cn = ESPV2_ConfigOption_CreateSec(secDes);
                            DOFSecurityMode_Destroy(sec);
                            if (!OALPtrList_InsertAt(nackList, cn,nackListCount++))
                            {
                                ESPV2_ConfigOption_Destroy(cn);
                                goto EXIT_CLOSE_CONNECTION;
                            }
                        }
                    }
                    else
                    {
                        goto EXIT_CLOSE_CONNECTION;
                    }
                }
                else
                {
                    ESPV2_Reject cfgR = ESPV2_CCMReject_Create();
                    if (!cfgR)
                    {
                        goto EXIT_CLOSE_CONNECTION;
                    }
                    if (!OALPtrList_InsertAt(rejectList, cfgR, rejectListCount++))
                    {
                        ESPV2_Reject_Destroy(cfgR);
                        goto EXIT_CLOSE_CONNECTION;
                    }
                }
                break;
            case ESP_ATTR_TYPE_TRP:
                if ( opt->attributeData != PROTOCOLID_TRP_V1 )
                {
                    if( !self->flags.bit.rejectedTRP )
                    {
                        rejectTRP = TRUE;
                    }
                    else
                    {
                        goto EXIT_CLOSE_CONNECTION;
                    }
                }
                break;
            case ESP_ATTR_TYPE_Gen:
                /* Gen is not currently supported */
            default: /* unrecognizable attribute code...goes in the reject list. */
                {
                    ESPV2_Reject cfgR = ESPV2_Reject_Create(opt->attributeCode);

                    if (!cfgR)
                    {
                        goto EXIT_CLOSE_CONNECTION;
                    }
                    if (!OALPtrList_InsertAt(rejectList, cfgR, rejectListCount++))
                    {
                        ESPV2_Reject_Destroy(cfgR);
                        goto EXIT_CLOSE_CONNECTION;
                    }
                }
            }
        }

        if( rejectTEP )
        {
            ESPV2_Reject cfgR;
            if( suggestTEP )
            {
                cfgR = ESPV2_TEPReject_Create();
            }
            else
            {
                cfgR = ESPV2_Reject_Create(ESP_ATTR_TYPE_TEP);
            }

            if (!cfgR)
            {
                goto EXIT_CLOSE_CONNECTION;
            }
            if (!OALPtrList_InsertAt(rejectList, cfgR, rejectListCount++))
            {
                ESPV2_Reject_Destroy(cfgR);
                goto EXIT_CLOSE_CONNECTION;
            }
            self->flags.bit.rejectedTEP = 1;
        }

        if( rejectTRP )
        {
            ESPV2_Reject cfgR = ESPV2_TRPReject_Create();
            if (!cfgR)
            {
                goto EXIT_CLOSE_CONNECTION;
            }
            if (!OALPtrList_InsertAt(rejectList, cfgR, rejectListCount++))
            {
                ESPV2_Reject_Destroy(cfgR);
                goto EXIT_CLOSE_CONNECTION;
            }
            self->flags.bit.rejectedTRP = 1;
        }

        /* make sure OAP was requested */
        if ( !rejectListCount && !self->flags.bit.amClientNotServer && !wasOapRequested && protocolNeg != DOF_DEFAULT_AS_PROTOCOL_NEGOTIATOR )
        {
            ESPV2_ConfigOption oapOpt = ESPV2_ConfigOption_Create(ESP_ATTR_TYPE_OAP, PROTOCOLID_OAP_V1, 0);

            if (!oapOpt)
            {
                goto EXIT_CLOSE_CONNECTION;
            }
            if (!OALPtrList_InsertAt(nackList, oapOpt, nackListCount++))
            {
                ESPV2_ConfigOption_Destroy(oapOpt);
                goto EXIT_CLOSE_CONNECTION;
            }
        }

        /* make sure TEP is requested, if needed */
        if ( !rejectListCount && !self->flags.bit.amClientNotServer && !wasTepRequested && self->flags.bit.requireTEP )
        {
            ESPV2_ConfigOption tepOpt = ESPV2_ConfigOption_Create(ESP_ATTR_TYPE_TEP, PROTOCOLID_TEP_V1, 0);

            if (!tepOpt)
            {
                goto EXIT_CLOSE_CONNECTION;
            }
            if (!OALPtrList_InsertAt(nackList, tepOpt, nackListCount++))
            {
                ESPV2_ConfigOption_Destroy(tepOpt);
                goto EXIT_CLOSE_CONNECTION;
            }
        }
    }

    if ( rejectListCount )
    {
        ESPV2_ConfigOptionList_Destroy(nackList);
        retVal.responseCode = ESP_V2_CFGHANDLER_RESPONSE_REJECT;
        retVal.optList = rejectList;
    }
    else if ( nackListCount )
    {
        ESPV2_RejectList_Destroy(rejectList);
        retVal.responseCode = ESP_V2_CFGHANDLER_RESPONSE_NEGACK;
        retVal.optList = nackList;
    }
    else
    {
        ESPV2_ConfigOptionList_Destroy(nackList);
        ESPV2_ConfigOptionList_Destroy(rejectList);
        retVal.responseCode = ESP_V2_CFGHANDLER_RESPONSE_ACK;
        retVal.optList = NULL;
        self->flags.bit.requestTRP = TRUE; /* resets it so it is requested if needed other direction */
        self->flags.bit.rejectedOAP = FALSE;
        self->flags.bit.rejectedTRP = FALSE;
        self->flags.bit.rejectedTEP = FALSE;

        if(self->theirNegotiatedOpts)
            ESPV2_ConfigOptionList_Destroy(self->theirNegotiatedOpts);
        self->theirNegotiatedOpts = ESPV2_ConfigOptionList_Copy( cfgRequestOpts );
        if(!self->theirNegotiatedOpts)
            goto EXIT_CLOSE_CONNECTION_NOFREE;
    }

    return retVal;

EXIT_CLOSE_CONNECTION:
    ESPV2_ConfigOptionList_Destroy(nackList);
    ESPV2_ConfigOptionList_Destroy(rejectList);
EXIT_CLOSE_CONNECTION_NOFREE:
    retVal.responseCode = ESP_V2_CFGHANDLER_RESPONSE_CLOSE;
    retVal.optList = NULL;
    return retVal;
}

ESP_V2_CfgHandlerResponse_t configNegAckHandler(void* protocolStack, ConfigOptionList cfgNegAckOpts)
{
    OALConnectedProtocolStack self = (OALConnectedProtocolStack)protocolStack;
    DOFSecurityDesire secDes = self->secDes;
    uint32 optCount;

    boolean canSendHack = FALSE;

    boolean resendRequest = FALSE;
    ESP_V2_CfgHandlerResponse_t retVal;

    uint32 i;

    PCR_Assert(cfgNegAckOpts != NULL);
    optCount = OALPtrList_GetCount(cfgNegAckOpts);
    for ( i = 0; i < optCount; ++i )
    {
        ESPV2_ConfigOption opt = OALPtrList_PeekAt(cfgNegAckOpts, i);

        switch ( opt->attributeCode )
        {
        case ESP_ATTR_TYPE_TEP:
            /* they want TEP if we are not Nonsecure then we can add to request */
            if ( opt->attributeData == PROTOCOLID_TEP_V1 )
            {
                if ( secDes != DOFSECURITYDESIRE_NOT_SECURE )
                {
                    canSendHack = TRUE;
                    self->flags.bit.requestTEP = TRUE;
                    resendRequest = TRUE;
                }
            }
            break;
        case ESP_ATTR_TYPE_OAP:
            if( opt->attributeData == PROTOCOLID_OAP_V1 )
            {
                self->flags.bit.requestOAP = TRUE;
                resendRequest = TRUE;
            }
            break;
        case ESP_ATTR_TYPE_TRP:
            if( opt->attributeData == PROTOCOLID_TRP_V1 )
            {
                self->flags.bit.requestTRP = TRUE;
                resendRequest = TRUE;
            }
            break;
        case ESP_ATTR_TYPE_CCM:
            /* we only support one mode and we will ask for it if we need it so a NAK does nothing */
            if( opt->attributeData == PROTOCOLID_CCMMODE_V1 )
            {
                if( secDes != DOFSECURITYDESIRE_NOT_SECURE )
                {
                    boolean nakSec = FALSE;
                    DOFSecurityMode sec = ESPV2_ConfigOption_SecModeFromConfigOption( opt, secDes, &nakSec);
                    if( sec )
                    {
                        /* Currently we will pick the mode server NAKed with */
                        DOFSecurityMode_Destroy( self->associatedChannel->remoteChannelData->mode );
                        self->associatedChannel->remoteChannelData->mode = sec;
                        self->flags.bit.requestSEC = TRUE;
                        resendRequest = TRUE;
                    }
                }
            }
            break;
        default: /* unrecognizable attribute code */
            break;
        }
    }

    if ( resendRequest )
    {
        ConfigOptionList optsToRequest = createRequestOptionsList( self,  canSendHack);

        if (optsToRequest)
        {
            retVal.responseCode = ESP_V2_CFGHANDLER_RESPONSE_REQUEST;
            retVal.optList = optsToRequest;
            if( self->myNegotiatedOpts )
                ESPV2_ConfigOptionList_Destroy( self->myNegotiatedOpts );
            self->myNegotiatedOpts = ESPV2_ConfigOptionList_Copy( optsToRequest );
            if( self->myNegotiatedOpts )
            {
                return retVal;
            }
        }
    }

    retVal.responseCode = ESP_V2_CFGHANDLER_RESPONSE_CLOSE;
    retVal.optList = NULL;
    return retVal;
}

ESP_V2_CfgHandlerResponse_t configRejectHandler(void* protocolStack, RejectList cfgRejectOpts)
{
    OALConnectedProtocolStack self = (OALConnectedProtocolStack)protocolStack;
    DOFProtocolNegotiator protocolNeg = self->associatedChannel->remoteChannelData->negotiator;
    uint32 optCount;

    boolean resendRequest = FALSE;
    ESP_V2_CfgHandlerResponse_t retVal = { ESP_V2_CFGHANDLER_RESPONSE_CLOSE, NULL };

    uint32 i;

    PCR_Assert(cfgRejectOpts != NULL);
    optCount = OALPtrList_GetCount(cfgRejectOpts);
    for (i = 0; i < optCount; ++i )
    {
        ESPV2_Reject opt = OALPtrList_PeekAt(cfgRejectOpts, i);

        /* These don't look at the APPIDs because we asked for what we know so there is nothing else to try */
        switch ( opt->attributeCode )
        {
        case ESP_ATTR_TYPE_TEP:
            if(!self->flags.bit.requireTEP)
            {
                self->flags.bit.requestTEP = FALSE;
                resendRequest = TRUE;
            }
            else
            {
                goto EXIT_CLOSE_CONNECTION;
            }
            break;
        case ESP_ATTR_TYPE_OAP:
            if( protocolNeg == DOF_DEFAULT_AS_PROTOCOL_NEGOTIATOR )
            {
                self->flags.bit.requestOAP = FALSE;
                resendRequest = TRUE;
            }
            else
            {
                goto EXIT_CLOSE_CONNECTION;
            }
            break;
        case ESP_ATTR_TYPE_CCM:
            if(!self->flags.bit.requireTEP)
            {
                resendRequest = TRUE;
            }
            else
            {
                goto EXIT_CLOSE_CONNECTION;
            }
            break;
        case ESP_ATTR_TYPE_TRP:
            if( !(self->flags.bit.amClientNotServer && protocolNeg == DOF_DEFAULT_AS_PROTOCOL_NEGOTIATOR) )
            {
                self->flags.bit.requestTRP = FALSE;
                resendRequest = TRUE;
            }
            else
            {
                goto EXIT_CLOSE_CONNECTION;
            }
            break;
        case ESP_ATTR_TYPE_Gen:
        default: /* unrecognizable attribute code */
            goto EXIT_CLOSE_CONNECTION;
            break;
        }
    }

    if ( resendRequest )
    {
        if ( self->myNegotiatedOpts )
            ESPV2_ConfigOptionList_Destroy( self->myNegotiatedOpts );
        self->myNegotiatedOpts = createRequestOptionsList( self, FALSE );
        if ( self->myNegotiatedOpts )
        {
            retVal.optList = ESPV2_ConfigOptionList_Copy( self->myNegotiatedOpts );
            retVal.responseCode = ESP_V2_CFGHANDLER_RESPONSE_REQUEST;
            return retVal;
        }
    }

    EXIT_CLOSE_CONNECTION:
    retVal.responseCode = ESP_V2_CFGHANDLER_RESPONSE_CLOSE;
    retVal.optList = NULL;
    return retVal;
}

static void OALConnectedProtocolStack_DoPostNegotiationStuff(OALConnectedProtocolStack self)
{
    ConfigOptionList optLists[2];
    uint8 i;

    /* Note: This was added to close the channel when an invalid negotiation finished.
             The whole solution will come with a rewrite of ESP */
    /* PCR_Assert(self->myNegotiatedOpts != NULL); */
    /* PCR_Assert(self->theirNegotiatedOpts != NULL); */
    if( !self->myNegotiatedOpts || !self->theirNegotiatedOpts )
    {
        OALRemoteChannel_Shutdown( self->associatedChannel, DOFEXCEPTION_PROTOCOL_NEGOTIATION_FAILED );
        return;
    }

    PCR_Assert(self->associatedChannel != NULL);
    PCR_Assert(self->associatedChannel->remoteChannelData != NULL);

    optLists[0] = self->myNegotiatedOpts;
    optLists[1] = self->theirNegotiatedOpts;

    for ( i=0; i < 2; ++i )         /* myNegotiatedOpts and theirNegotiatedOpts */
    {
        ConfigOptionList optList = optLists[i];
        uint32 optCount = OALPtrList_GetCount(optList);
        uint32 j;

        for ( j=0; j < optCount; ++j )
        {
            ESPV2_ConfigOption opt = OALPtrList_PeekAt(optList, j);

            switch ( opt->attributeCode )
            {
            case ESP_ATTR_TYPE_TEP:
                if ( (opt->attributeData == PROTOCOLID_TEP_V1) && !self->tepV1 )
                {
                    self->associatedChannel->remoteChannelData->flags.secureDesire = TRUE;

                    self->tepV1 = TEP_V1_Create( &(self->superClass.blackboardData), FALSE, OALConnectedProtocolStack_TEP_V1_SessionEventCallback );
                    if (!self->tepV1)
                    {
                        OALRemoteChannel_Shutdown(self->associatedChannel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
                    }
                }
                break;
            case ESP_ATTR_TYPE_OAP:
                if ( (opt->attributeData == PROTOCOLID_OAP_V1) && !self->oapV1 )
                {
                    self->oapV1 = OAP_V1_Create( &(self->superClass.blackboardData), FALSE );
                    if (!self->oapV1)
                    {
                        OALRemoteChannel_Shutdown(self->associatedChannel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
                    }
                }
                break;
            case ESP_ATTR_TYPE_TRP:
                if ( (opt->attributeData == PROTOCOLID_TRP_V1))
                {
                    if( i == 0 )
                        self->associatedChannel->remoteChannelData->flags.TRP_SendCommands = 1;
                    else
                        self->associatedChannel->remoteChannelData->flags.TRP_AcceptsCommands = 1;

                    if( !self->trpV1 )
                        self->trpV1 = TRP_V1_Create( &(self->superClass.blackboardData), FALSE );
                    if (!self->trpV1)
                    {
                        OALRemoteChannel_Shutdown(self->associatedChannel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
                    }
                }
                break;
            case ESP_ATTR_TYPE_CCM:
                break;
            default: /* unrecognizable attribute code */
                break;
            }
        }
    }

    if ( self->tepV1 )
    {
        if (!self->associatedChannel->remoteChannelData->mode)
        {
            self->associatedChannel->remoteChannelData->mode = DOFSecurityModeCCM_Create( DOFCIPHERALGORITHM_AES, DOFCIPHERSTRENGTH_256, 4, (self->secDes == DOFSECURITYDESIRE_SECURE));
        }
        if (self->associatedChannel->remoteChannelData->mode)
        {
            if (OALProtocolStack_StartNegotiationTimeout((OALProtocolStack)self, TEP_NEGOTIATION_TIMEOUT))
            {
                self->associatedChannel->remoteChannelData->flags.secureDesire = TRUE;
                self->stackState = CONNECTEDSTACK_STATE_AUTH_NEGOTIATION;
                if(self->flags.bit.amClientNotServer)
                {
                    DOFCredentials creds = OALRemoteChannel_GetCredentials(self->associatedChannel->remoteChannelData);

                    if (!creds)
                    {
                        OALRemoteChannel_Shutdown(self->associatedChannel, DOFEXCEPTION_AUTHENTICATION_FAILED);
                    }
                    else
                    {
                        if( DOFCredentials_IsResolved(creds) )
                        {
                            self->associatedChannel->remoteChannelData->secureCredentials = DOFCredentials_Copy(creds);
                            OALRemoteChannel_CreateLibPermissions(OALConnectedProtocolStack_GetLibraryPermsCallback, (void *) self,
                                                                  self->associatedChannel->remoteChannelData->secureCredentials,
                                                                  self->associatedChannel, self->associatedChannel->remoteChannelData->address);
                        }
                        else
                        {
                            DOFCredentials_Internal_BeginResolve( creds, DOF_GetDof(), 30000, FALSE, self->associatedChannel, self->associatedChannel->remoteChannelData->address, OALConnectedProtocolStack_ResolveCredsCallback, self );
                        }
                    }
                }
                else
                {
                    /* A secure inbound connection currently does not support a bridge. */
                    self->associatedChannel->remoteChannelData->flags.createBridge = FALSE;
                    TEP_V1_Authenticate( self->tepV1,
                                         self->flags.bit.amClientNotServer,
                                         self,  self->associatedChannel,
                                         self->associatedChannel->remoteChannelData->address,
                                         NULL );
                }
            }
            else
            {
                OALConnectedProtocolStack_HandleReceiveError_Shutdown(self);
            }
        }
        else
        {
            OALRemoteChannel_Shutdown(self->associatedChannel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        }
    }
    else
    {
        self->stackState = CONNECTEDSTACK_STATE_OPEN;
        /* Notify channel of the success. */
        OALRemoteChannel_NotifyInitComplete(self->associatedChannel);
    }
}

static void OALConnectedProtocolStack_ResolveCredsCallback( DOFCredentials self, DOFException exception, void* context )
{
    OALMessage msg = NULL;
    DOFException except = DOFEXCEPTION_INSUFFICIENT_RESOURCES;
    OALProtocolStack stk = (OALProtocolStack)context;

    if( !exception )
    {
        /* Since it's not safe to use the associatedChannel in the stack here, send a message to the stack. */
        msg = SMT_ResolveCredsCallback_Create(stk, self);
    }
    else
    {
        except = exception;
    }

    if (msg)
    {
        OALRemoteChannelMgr_SendMessageToStack(msg);
    }
    else
    {
        /* Since it's not safe to get the channel out of the protocol stack, send a stack shutdown message. */
        OALProtocolStack_Shutdown(stk, except);
    }
}

/* Note: this function is a TRP callback so be carefull not to use members of protocolstack that are not immutable */
static void OALConnectedProtocolStack_GetLibraryPermsCallback(DOFError           status,
                                                              sid_t            * secureSid,
                                                              DOFPermissionSet   libPerms,
                                                              void             * context)
{
    OALMessage msg = NULL;
    DOFException except = DOFEXCEPTION_INSUFFICIENT_RESOURCES;
    OALProtocolStack stk = (OALProtocolStack)context;
    PCR_Assert(context != NULL);
    if(status == DOFERROR_OK)
    {
        PCR_Assert(secureSid != NULL);

        msg = SMT_GetLibraryPermsCallback_Create(stk, libPerms, secureSid);
    }
    else
    {
        /* Map the DOFError to a DOFException. */
        except = DOFErrorException_GetFromError(status);
    }

    if (msg)
    {
        /* If for some reason a message could not be allocated, just drop this an let a timeout destroy the channel. */
        OALRemoteChannelMgr_SendMessageToStack(msg);
    }
    else
    {
        OALProtocolStack_Shutdown(stk, except);
    }
}

/* Note: this function must be called from the message enqueue worker thread (message processing). */
static void OALConnectedProtocolStack_HandleReceiveError_Shutdown(OALConnectedProtocolStack self)
{
    DOFException     except = NULL;

    if ( self->stackState < CONNECTEDSTACK_STATE_OPEN )
        except = DOFErrorException_Create(DOFERROR_FAILED, "Failed to connect.");
    else
        except = DOFErrorException_Create(DOFERROR_FAILED, "Protocol stack failure");

    OALRemoteChannel_Shutdown(self->associatedChannel, except);
    DOFException_Destroy(except);
}

/* **************************************************/


static void* OALConnectedProtocolStack_TEP_V1_SessionEventCallback(void               * context,
                                                                   uint8              * key,
                                                                   TEP_SessionEvent_t   eventType,
                                                                   DOFSecurityMode      securityMode,
                                                                   OALChannel           channel,
                                                                   DOFSecurityScope     secScope,
                                                                   DOFPermissionSet     localPerms,
                                                                   DOFPermissionSet     otherPerms,
                                                                   DOFException         exception,
                                                                   uint32               stateID,
                                                                   DOFAddress           remoteAddr,
                                                                   OALStream          * initialState,
                                                                   DOFCredentials       creds)
{
    OALConnectedProtocolStack self = (OALConnectedProtocolStack)context;

    /* Note: since the associated channel in protocol stack can be cleared,
       do not trust using it in this function, rather use the channel passed in from TEP. */

    switch ( eventType )
    {
    case TEP_V1_SessionEvent_NEWSESSION_SUCCESS_RECVKEY:
        {
            DOFException except = NULL;
            OALCipher    cipher;

            PCR_Assert(securityMode);
            cipher = OALCipher_Create_TEP( securityMode, FALSE,
                                           self->flags.bit.amClientNotServer ? FALSE : TRUE, &except );
            if ( cipher )
            {
                if(*initialState == NULL)
                {
                    OALCipher_SetManager(cipher, self->flags.bit.amClientNotServer ? FALSE : /* We are TEP Initiator. */
                                                                                     TRUE);  /* We are TEP Responder. */
                    *initialState = OALCipher_CreateInitialStateStream_TEP( cipher );
                }
                if (*initialState != NULL)
                {
                    if(OALCipher_SetFutureKey(cipher, *initialState, key, localPerms, otherPerms))
                    {
                        if ( channel->remoteChannelData->cipher )
                            OALCipher_Destroy(channel->remoteChannelData->cipher);
                        channel->remoteChannelData->cipher = cipher;
                    }
                    else
                    {
                        OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_AUTHENTICATION_FAILED);
                    }
                }
                else
                {
                    OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
                }
            }
            else
            {
                OALRemoteChannel_Shutdown(channel, except);
            }
            DOFException_Destroy( except );     /* Destroy the exception from the call to OALCipher_Create */
        }
        break;

    case TEP_V1_SessionEvent_NEWSESSION_FAILURE:
    case TEP_V1_SessionEvent_REKEY_FAILURE:
        PCR_Assert(exception != NULL);
        OALRemoteChannel_Shutdown(channel, exception);
        break;

    case TEP_V1_SessionEvent_NEWSESSION_SUCCESS_SENDKEY:
        {
            KeyChangeInfo_t* pKeyChangeInfo;

            PCR_Assert(securityMode);
            PCR_Assert(localPerms);
            PCR_Assert(otherPerms);
            PCR_Assert(secScope);

            pKeyChangeInfo = OALRemoteChannel_CreateKeyChangeInfo( channel, channel->remoteChannelData->cipher, stateID, localPerms, otherPerms, securityMode, secScope, FALSE, TRUE );

            if(pKeyChangeInfo != NULL)
            {
                OALProtocolStack_SetKeyChangeCallback( &self->superClass, OALRemoteChannel_KeyChangeCallback, pKeyChangeInfo );
            }
            else
            {
                DOFException except = DOFErrorException_Create(DOFERROR_INSUFFICIENT_RESOURCES, "Failed to allocate key change state.");
                OALRemoteChannel_Shutdown(channel, except);
                DOFException_Destroy(except);
            }
        }
        break;

    case TEP_V1_SessionEvent_NEWSESSION_FINALIZE:
        PCR_Assert(self->stackState != CONNECTEDSTACK_STATE_OPEN);
        self->stackState = CONNECTEDSTACK_STATE_OPEN;
        PCR_Assert(DOFSecurityMembrane_IsChannelMembrane(channel->securityMembrane));
        if(!channel->remoteChannelData->flags.isOutbound)
        {
            OALOperationBridge bridge = OALOperationBridge_CreatePassThroughBridge(channel);

            if (bridge)
            {
                DOFSecurityMembrane_Destroy(channel->securityMembrane);
                channel->securityMembrane = DOFOperationBridgeConfig_GetMembrane(bridge);
            }
            else
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectedProtocolsStack: Flag bit was not set on a non-negotiation datagram.");
                OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            }
        }
        else if(OALChannelMembrane_GetBridge(channel->securityMembrane) == DOF_GetDefaultOperationBridge(DOF_GetDof()))
        {
            OALOperationBridge bridge = OALOperationBridge_CreatePassThroughBridge(channel);

            if (bridge)
            {
                OALChannelMembrane_SetBridge(channel->securityMembrane, bridge);
                OALOperationBridge_Destroy(bridge);
            }
            else
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectedProtocolsStack: Flag bit was not set on a non-negotiation datagram.");
                OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            }
        }

        if (channel->securityMembrane)
        {
            /* Notify channel about the success. */
            OALRemoteChannel_NotifyInitComplete(channel);
            OALSource_Clear(channel, channel->remoteChannelData->address);
            channel->remoteChannelData->flags.sessionIsSecure = TRUE;
        }
        else
        {
            OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        }
        break;

    case TEP_V1_SessionEvent_REKEY_RECVKEY:
        if (*initialState == NULL)
            *initialState = OALCipher_CreateInitialStateStream_TEP( channel->remoteChannelData->cipher );
        if (*initialState != NULL)
        {
            if (!OALCipher_SetFutureKey(channel->remoteChannelData->cipher, *initialState, key, localPerms, otherPerms))
            {
                OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_AUTHENTICATION_FAILED);
            }
        }
        else
        {
            OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        }
        break;

    case TEP_V1_SessionEvent_REKEY_DISABLEOLD:
        {
            KeyChangeInfo_t* pKeyChangeInfo;

            PCR_Assert(secScope);

            pKeyChangeInfo = OALRemoteChannel_CreateKeyChangeInfo( channel, channel->remoteChannelData->cipher, stateID, NULL, NULL, NULL, secScope, TRUE, FALSE );

            if(pKeyChangeInfo != NULL)
            {
                OALProtocolStack_SetKeyChangeCallback( &self->superClass, OALRemoteChannel_KeyChangeCallback, pKeyChangeInfo );
            }
            else
            {
                OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            }
        }
        break;

    case TEP_V1_SessionEvent_REKEY_SENDKEY:
        {
            KeyChangeInfo_t* pKeyChangeInfo =
                OALRemoteChannel_CreateKeyChangeInfo( channel, channel->remoteChannelData->cipher, stateID, NULL, NULL, NULL, NULL, FALSE, TRUE );

            if(pKeyChangeInfo != NULL)
            {
                OALProtocolStack_SetKeyChangeCallback( &self->superClass, OALRemoteChannel_KeyChangeCallback, pKeyChangeInfo );
            }
            else
            {
                OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            }
        }
        break;

    case TEP_V1_SessionEvent_REKEY_SENDKEYANDDISABLE:
        {
            KeyChangeInfo_t* pKeyChangeInfo;

            PCR_Assert(localPerms);
            PCR_Assert(otherPerms);
            PCR_Assert(secScope);

            pKeyChangeInfo = OALRemoteChannel_CreateKeyChangeInfo( channel, channel->remoteChannelData->cipher, stateID, localPerms, otherPerms, NULL, secScope, TRUE, TRUE );

            if(pKeyChangeInfo != NULL)
            {
                OALProtocolStack_SetKeyChangeCallback( &self->superClass, OALRemoteChannel_KeyChangeCallback, pKeyChangeInfo );
            }
            else
            {
                OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            }
        }
        break;

    case TEP_V1_SessionEvent_ENDING:
        {
            DOFException except = DOFErrorException_Create(DOFERROR_OK, "TEP Connection ending");
            OALRemoteChannel_Shutdown(channel, except);
            DOFException_Destroy(except);
            break;
        }

    default:
        PCR_Assert( FALSE ); /* error */
        break;
    }

    return context;
}

