/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>

#include <dof/address.h>
#include <dof/result.h>

#include <internal/channel_stack_messages.h>
#include <internal/dofinternal.h>
#include <internal/protocolstack.h>
#include <internal/connectionless_protocolstack.h>
#include <internal/connectionless_remotechannel.h>
#include <internal/protocollayerresult.h>
#include <internal/enp_v1.h>
#include <internal/epp_v2.h>
#include <internal/exception.h>
#include <internal/marshal.h>
#include <internal/remotechannel.h>
#include <internal/remotechanneldata.h>
#include <internal/remotechannelmanager.h>
#include <internal/oap_v1.h>
#include <internal/epp_negative_one.h>
#include <internal/tep_v1.h>
#include <internal/trp_v1.h>
#include <internal/sgmp_v1.h>
#include <internal/ccmmode_v1.h>
#include <internal/resultinternal.h>
#include <internal/router.h>
#include <internal/messaging.h>
#include <internal/transportinternal.h>
#include <internal/transportmanager.h>

#include "connectionless_protocolstack_struct.h"
#include "connectionless_v0query.h"

#ifdef OALPROTOCOLSTACK_PACKET_TRACE
#include <stdio.h>
#endif

struct traverseSSIDContext
{
    uint32                          channelCount;
    uint32                          maxCount;
    OALChannel *                    channelList;
};

typedef void (*OCLPS_MessageHandler_fn_t)( OALConnectionlessProtocolStack self, OALMessage message );

static void OALConnectionlessProtocolStack_Dtor( OALProtocolStack selfx );
static void OALConnectionlessProtocolStack_HandleMessage( OALProtocolStack selfx, OALMessage message );
static ProtocolLayerResult OALConnectionlessProtocolStack_EPP_ENP_Send_fn(OALProtocolStack selfx, BlackboardData blackboardData);

static boolean OALConnectionlessProtocolStack_Send(OALConnectionlessProtocolStack self,
                                                ProtocolLayerIndicator_t protocolLayerIndicator,
                                                uint16 attributeData,
                                                DOFAddress destAddr,
                                                OALChannel associatedChannel,
                                                uint32 argc,
                                                void* argv[]);

static void OALConnectionlessProtocolStack_HandleReceiveData(OALConnectionlessProtocolStack self,
                                                             OALStream data,
                                                             DOFAddress srcAddr,
                                                             boolean isMulticast,
                                                             OALChannel associatedChannel);

static void* OALConnectionlessProtocolStack_TEP_V1_SessionEventCallback(void               * context,
                                                                        uint8              * key,
                                                                        TEP_SessionEvent_t   eventType,
                                                                        DOFSecurityMode      securityMode,
                                                                        OALChannel           channel,
                                                                        DOFSecurityScope     secScope,
                                                                        DOFPermissionSet     perms,
                                                                        DOFPermissionSet     otherPerms,
                                                                        DOFException         exception,
                                                                        uint32               stateID,
                                                                        DOFAddress           remoteAddr,
                                                                        OALStream          * outStream,
                                                                        DOFCredentials       creds);

static boolean OALConnectionlessProtocolStack_SGMP_KeyChangedCallback(void * ctx)
{
    PCREvent_Signal((PCREvent)ctx);
    return TRUE;
}


OALConnectionlessProtocolStack OALConnectionlessProtocolStack_Create(void)
{
    OALConnectionlessProtocolStack retVal = (OALConnectionlessProtocolStack)DOFMem_Alloc( sizeof(OALConnectionlessProtocolStack_t) );

    if (!retVal)
    {
        return NULL;
    }

    retVal->superClass.shutdownMessageSent = FALSE;
    retVal->superClass.shutdownMessage = SMT_Shutdown_Create((void *)&retVal->superClass);
    if (!retVal->superClass.shutdownMessage)
    {
        DOFMem_Free(retVal);
        return NULL;
    }

    OALProtocolStack_Ctor( &(retVal->superClass),
                           OALConnectionlessProtocolStack_Dtor,
                           OALConnectionlessProtocolStack_HandleMessage,
                           OALConnectionlessProtocolStack_EPP_ENP_Send_fn );

    retVal->enpV1 =     ENP_V1_Create( &(retVal->superClass.blackboardData), TRUE, FALSE );
    if (retVal->enpV1)
    {
        retVal->eppV2 =     EPP_V2_Create( &(retVal->superClass.blackboardData), TRUE );
        if (retVal->eppV2)
        {
            retVal->oapV1 =     OAP_V1_Create( &(retVal->superClass.blackboardData), TRUE );
            if (retVal->oapV1)
            {
                retVal->eppNegOne = EPP_NegOne_Create( &(retVal->superClass.blackboardData), TRUE );
                if (retVal->eppNegOne)
                {
                    retVal->tepV1 =     TEP_V1_Create ( &(retVal->superClass.blackboardData), TRUE, OALConnectionlessProtocolStack_TEP_V1_SessionEventCallback );
                    if (retVal->tepV1)
                    {
                        retVal->trpV1 =     TRP_V1_Create( &(retVal->superClass.blackboardData), TRUE );
                        if (retVal->trpV1)
                        {
                            retVal->sgmpV1 =    SGMP_V1_Create( &(retVal->superClass.blackboardData));
                            if (retVal->sgmpV1)
                            {
                                retVal->ccmModeV1 =   CCMMODE_V1_Create( &(retVal->superClass.blackboardData), TRUE);
                                if (retVal->ccmModeV1)
                                {
                                    return retVal;
                                }
                                SGMP_V1_Destroy(retVal->sgmpV1);
                            }
                            TRP_V1_Destroy(retVal->trpV1);
                        }
                        TEP_V1_Destroy(retVal->tepV1);
                    }
                    EPP_NegOne_Destroy(retVal->eppNegOne);
                }
                OAP_V1_Destroy(retVal->oapV1);
            }
            EPP_V2_Destroy(retVal->eppV2);
        }
        ENP_V1_Destroy(retVal->enpV1);
    }

    DOFMem_Free(retVal->superClass.shutdownMessage);
    DOFMem_Free(retVal);
    return NULL;
}

static void OALConnectionlessProtocolStack_Dtor( OALProtocolStack selfx )
{
    OALConnectionlessProtocolStack self = (OALConnectionlessProtocolStack)selfx;
    if(!self)
        return;

    CCMMODE_V1_Destroy( self->ccmModeV1 );
    SGMP_V1_Destroy( self->sgmpV1 );
    TRP_V1_Destroy( self->trpV1 );
    TEP_V1_Destroy( self->tepV1 );
    EPP_NegOne_Destroy( self->eppNegOne );
    OAP_V1_Destroy( self->oapV1 );
    EPP_V2_Destroy( self->eppV2 );
    ENP_V1_Destroy( self->enpV1 );

    OALProtocolStack_Dtor( (OALProtocolStack)self );

    DOFMem_Free( self );
}

static void OALConnectionlessProtocolStack_AddToBlockedMessageQueue(OALChannel channel, OALMessage message)
{
    boolean status = TRUE;

    if (!channel->remoteChannelData->blockedMessageQueue)
    {
        channel->remoteChannelData->blockedMessageQueue = OALQueue_Create();
    }

    if (channel->remoteChannelData->blockedMessageQueue)
    {
        OALMessage msg = OALMessage_Retain(message);
        if (msg)
        {
            status = OALQueue_Enqueue(channel->remoteChannelData->blockedMessageQueue, msg);
        }
        else
        {
            status = FALSE;
        }
    }
    else
    {
        status = FALSE;
    }

    if (status == FALSE)
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: Failed to create blocked message queue\n");
        OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
    }
}

static void OALConnectionlessProtocolStack_HandleMessage_RouteToProvider( OALConnectionlessProtocolStack self, OALMessage message )
{
    ProtocolLayerIndicator_t    protocolLayerIndicator;
    DOFOperation                req_op;
    DOFOperation                prov_op;
    DOFAddress                  destAddr;
    OALChannel                  associatedChannel;

    PCR_Assert( message->argc >= 4 );

    req_op              = message->argv[0];
    prov_op             = message->argv[1];
    destAddr            = message->argv[2];
    associatedChannel   = message->argv[3];

    protocolLayerIndicator = OALProtocolStack_DetermineProtocol(req_op->operation_type);

    switch (protocolLayerIndicator)
    {
    case PLI_OAP:
        {
            void* args[4];
            /* args to OAP_V1_Send */
            args[0] = req_op;
            args[1] = prov_op;
            args[2] = (void*)TRUE;
            args[3] = NULL;

            /* Do not send OAP from a unicast connection unless it is known that the remote host can accept it. */
            if (!(associatedChannel->remoteChannelData->ESP_V0_QueryAppids & V0_QUERY_BIT_OAP_V1) &&
                !associatedChannel->remoteChannelData->flags.defaultAddressIsMulticast)
            {
                if ( !associatedChannel->remoteChannelData->flags.isOutbound )
                    DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: Not sending OAP datagram, an initial OAP datagram has not been received");
                break;
            }

            OALConnectionlessProtocolStack_Send( self, protocolLayerIndicator, PROTOCOLID_OAP_V1, destAddr, associatedChannel, 4, args );
            goto CleanupAndExit;
        }
        break;

    case PLI_TEP:
        {
            void* args[3];
            /* args to TEP_V1_Send */
            args[0] = req_op;
            args[1] = (void*)TRUE;
            args[2] = NULL;

            OALConnectionlessProtocolStack_Send( self, protocolLayerIndicator, PROTOCOLID_TEP_V1, destAddr, associatedChannel, 3, args );
            goto CleanupAndExit;
        }
        break;

    case PLI_TRP:
        {
            boolean addrIsMulticast;
            void* args[3];
            args[0] = req_op;
            args[1] = (void*)TRUE;
            args[2] = NULL;

            if (destAddr)
                addrIsMulticast = (DOFAddress_GetAddressType(destAddr) != DOFADDRESSTYPE_UNICAST);
            else
                addrIsMulticast = associatedChannel->remoteChannelData->flags.defaultAddressIsMulticast;

            if (!addrIsMulticast && !(associatedChannel->remoteChannelData->ESP_V0_QueryAppids & V0_QUERY_BIT_TRP_V1))
            {
                if (!associatedChannel->remoteChannelData->flags.isOutbound
                    && associatedChannel->remoteChannelData->flags.TRP_SendCommands)
                {
                    /*  If it is unknown whether the remote host accepts TRP commands, send a query and queue up commands
                        until the response is received.  Note: TRP_SendCommand is set by default for inbound connectionless
                        remote channels, so if the query comes back and TRP command are not allowed this must be cleared. */
                    OALConnectionlessProtocolStack_AddToBlockedMessageQueue(associatedChannel, message);
                    ESP_V2_SendQuery(associatedChannel, 8000, 3000);        /* 8 second timeout with 3 second retry. */
                    break;
                }
                else
                {
                    /* If the remote host does not accept TRP commands, drop this command. */
                    DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_WARNING, "Unable to send TRP Command because it is not allowed.");
                    break;
                }
            }

            OALConnectionlessProtocolStack_Send( self, protocolLayerIndicator, PROTOCOLID_TRP_V1, destAddr, associatedChannel, 3, args );
            goto CleanupAndExit;
        }
        break;

    case PLI_SGMP:
        {
            void* args[3];
            args[0] = req_op;
            args[1] = (void*)TRUE;
            args[2] = NULL;

            OALConnectionlessProtocolStack_Send( self, protocolLayerIndicator, PROTOCOLID_SGMP_V1, destAddr, associatedChannel, 3, args );
            goto CleanupAndExit;
        }
    case PLI_CCMMODE:
        {
            void* args[3];
            args[0] = req_op;
            args[1] = (void*)TRUE;
            args[2] = NULL;

            OALConnectionlessProtocolStack_Send( self, protocolLayerIndicator, PROTOCOLID_CCMMODE_V1, destAddr, associatedChannel, 3, args );
            goto CleanupAndExit;
        }
    case PLI_ESP:
        {
            OALConnectionlessProtocolStack_Send( self, protocolLayerIndicator, PROTOCOLID_ESP, destAddr, associatedChannel, 1, (void**)&req_op);
            goto CleanupAndExit;
        }
    case PLI_EPP_NEGONE:
        {
            OALConnectionlessProtocolStack_Send( self, protocolLayerIndicator, PROTOCOLID_EPP_NEGONE, destAddr, associatedChannel, 1, (void**)&req_op );
            goto CleanupAndExit;
        }
    default:
        break;
    }

    CleanupAndExit:
    return;
}

static void OALConnectionlessProtocolStack_HandleMessage_RouteToRequestor( OALConnectionlessProtocolStack self, OALMessage message )
{
    ProtocolLayerIndicator_t    protocolLayerIndicator;
    DOFOperation                operation;
    DOFResult                   result;
    DOFAddress                  destAddr;
    OALChannel                  associatedChannel;

    PCR_Assert( message->argc >= 4 );

    operation           = message->argv[0];
    result              = message->argv[1];
    destAddr            = message->argv[2];
    associatedChannel   = message->argv[3];

    protocolLayerIndicator = OALProtocolStack_DetermineProtocol(operation->operation_type);

    switch (protocolLayerIndicator)
    {
    case PLI_OAP:
        {
            void* args[4];
            /* args to OAP_V1_Send */
            args[0] = operation;
            args[1] = NULL;
            args[2] = (void*)FALSE;
            args[3] = result;

            /* Do not send OAP from an inbound connection until the requestor has sent OAP. */
            if (!(associatedChannel->remoteChannelData->ESP_V0_QueryAppids & V0_QUERY_BIT_OAP_V1) &&
                !associatedChannel->remoteChannelData->flags.defaultAddressIsMulticast)
            {
                if ( !associatedChannel->remoteChannelData->flags.isOutbound )
                    DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: Not sending OAP datagram, an initial OAP datagram has not been received");
                break;
            }

            if ( OALConnectionlessProtocolStack_Send( self, protocolLayerIndicator, PROTOCOLID_OAP_V1, destAddr, associatedChannel, 4, args ) )
            {
                /* Now that PDU is well along the way to going on the wire, cancel its operation if previously marked.*/
                if (operation->flags.bits.cancelOpAfterEppSend)
                    DOFOperation_Cancel( operation );
            }
            break;
        }
    case PLI_TRP:
        {
            void* args[3];

            args[0] = operation;
            args[1] = (void*)FALSE;
            args[2] = result;
            OALConnectionlessProtocolStack_Send( self, protocolLayerIndicator, PROTOCOLID_TRP_V1, destAddr, associatedChannel, 3, args );
            break;
        }
    case PLI_SGMP:
        {
            void* args[3];

            args[0] = operation;
            args[1] = (void*)FALSE;
            args[2] = NULL;
            OALConnectionlessProtocolStack_Send( self, protocolLayerIndicator, PROTOCOLID_SGMP_V1, destAddr, associatedChannel, 3, args );
            break;
        }
    case PLI_CCMMODE:
        {
            void* args[3];

            args[0] = operation;
            args[1] = (void*)FALSE;
            args[2] = NULL;
            OALConnectionlessProtocolStack_Send( self, protocolLayerIndicator, PROTOCOLID_CCMMODE_V1, destAddr, associatedChannel, 3, args );
            break;
        }
    default:

        break;
    }

    return;
}

static void OALConnectionlessProtocolStack_HandleMessage_ReceiveFromTransport( OALConnectionlessProtocolStack self, OALMessage message )
{
    if (!message->shutdown)
    {
        OALStream data;
        DOFAddress srcAddr;
        boolean isMulticast;
        OALChannel associatedChannel;

        PCR_Assert( message->argc == 4 );

        data                = (OALStream)message->argv[0];
        srcAddr             = (DOFAddress)message->argv[1];
        isMulticast         = (boolean)((int)message->argv[2]);
        associatedChannel   = (OALChannel)message->argv[3];

        OALConnectionlessProtocolStack_HandleReceiveData( self, data, srcAddr, isMulticast, associatedChannel );
    }
}

static void OALConnectionlessProtocolStack_HandleMessage_BeginNegotiationAuthorization( OALConnectionlessProtocolStack self, OALMessage message )
{
    if (!message->shutdown)
    {
        OALChannel      associatedChannel;
        DOFCredentials  creds;

        PCR_Assert( message->argc == 1 );
        associatedChannel = (OALChannel)message->argv[0];

        creds = OALRemoteChannel_GetCredentials(associatedChannel->remoteChannelData);
        if (!creds)
        {
            OALRemoteChannel_Shutdown(associatedChannel, DOFEXCEPTION_AUTHENTICATION_FAILED);
            return;
        }

        if(DOFCredentials_IsResolved(creds))
        {
            OALChannel contextChannel = OALChannel_Retain(associatedChannel);

            /* Make sure that the channel passed in as context is retained since _CreateLibPermissions is asynchronous */
            if (contextChannel)
            {
                associatedChannel->remoteChannelData->secureCredentials = DOFCredentials_Copy(creds);

                if (associatedChannel->remoteChannelData->secureCredentials)
                {
                    OALRemoteChannel_CreateLibPermissions(OALConnectionlessRemoteChannel_GetLibraryPermsCallback, contextChannel, creds,
                                                          associatedChannel, associatedChannel->remoteChannelData->address);
                }
                else
                {
                    OALChannel_Release(contextChannel);
                    OALRemoteChannel_Shutdown(associatedChannel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
                }
            }
            else
            {
                OALRemoteChannel_Shutdown(associatedChannel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            }
        }
        else
        {
            OALChannel contextChannel = OALChannel_Retain(associatedChannel);


            DOFCredentials_Internal_BeginResolve(creds, DOF_GetDof(), 30000, FALSE, contextChannel, contextChannel->remoteChannelData->address,
                                                  OALConnectionlessRemoteChannel_ResolveCredsCallback, contextChannel);
        }
    }
}

void OALConnectionlessProtocolStack_StartTepNegotiation( OALConnectionlessProtocolStack self, OALChannel channel,
                                                         DOFPermissionSet perms )
{
    TEP_V1_Authenticate( self->tepV1, TRUE, channel, channel, channel->remoteChannelData->address, perms );
}

static void OALConnectionlessProtocolStack_HandleMessage_Shutdown( OALConnectionlessProtocolStack self, OALMessage message )
{
    OALRemoteChannelMgr_RemoveProtocolStack_Immediate(&self->superClass);
}

static void OALConnectionlessProtocolStack_HandleMessage_Timeout(OALConnectionlessProtocolStack self, OALMessage message)
{
    PCR_Assert(FALSE); /* shouldn't receive this message on connectionless stack! */
}

static void OALConnectionlessProtocolStack_HandleMessage_BeginNegotiationServerSide(OALConnectionlessProtocolStack self, OALMessage message)
{
    PCR_Assert(FALSE); /* shouldn't receive this message on connectionless stack! */
}

static void OALConnectionlessProtocolStack_HandleMessage_RemoteDomainCallback(OALConnectionlessProtocolStack self, OALMessage message)
{
    /* This should not be called for connectionless protocol stacks, but just return without doing anything,
       the message will be destroyed. */
}

static void OALConnectionlessProtocolStack_HandleMessage_ResolveCredsCallback(OALConnectionlessProtocolStack self, OALMessage message)
{
    /* This should not be called for connectionless protocol stacks, but just return without doing anything,
       the message will be destroyed. */
}

static void OALConnectionlessProtocolStack_HandleMessage_GetLibraryPermsCallback(OALConnectionlessProtocolStack self, OALMessage message)
{
    /* This should not be called for connectionless protocol stacks, but just return without doing anything,
       the message will be destroyed. */
}

static void OALConnectionlessProtocolStack_HandleMessage_RouteSavedEPP(OALConnectionlessProtocolStack self, OALMessage message)
{
    BlackboardData              blackboardData = &(self->superClass.blackboardData);

    BlackboardData_Restore(blackboardData, (BlackboardData)message->argv[0]);
    /* The blackboard data in the stack now owns the objects in blackboardData so call
       blackboardData_Init so that the objects will not be destroyed when the message is destroyed. */
    BlackboardData_Init((BlackboardData)message->argv[0]);

    /* This is hard coded for EPP_V2 and ENP_V1 and must be updated when different or multiple versions are added. */
    if (self->superClass.ENP_EPP_send_func(&self->superClass, blackboardData) == PLRESULT_OKSENDDOWN)
    {
        if (DOFTransportLayer_SendTo(blackboardData->associatedChannel->remoteChannelData->dtl, blackboardData->sendPacket, blackboardData->transportAddr))
        {
            if (blackboardData->associatedChannel->remoteChannelData->maxSendSilenceTimerItem)
            {
                OALRemoteChannelMgr_SetSendSilenceTimerItem(blackboardData->associatedChannel);
            }
        }
    }
    BlackboardData_Clear(blackboardData);
}

static void OALConnectionlessProtocolStack_HandleMessage( OALProtocolStack selfx, OALMessage message )
{
    OALConnectionlessProtocolStack self = (OALConnectionlessProtocolStack)selfx;
    static OCLPS_MessageHandler_fn_t msg_handler_funcs[] = {
        /* the order here is coupled with StackMsgType_t. So dont change it. */
        OALConnectionlessProtocolStack_HandleMessage_RouteToProvider,
        OALConnectionlessProtocolStack_HandleMessage_RouteToRequestor,
        OALConnectionlessProtocolStack_HandleMessage_ReceiveFromTransport,
        OALConnectionlessProtocolStack_HandleMessage_BeginNegotiationAuthorization,
        OALConnectionlessProtocolStack_HandleMessage_Shutdown,
        OALConnectionlessProtocolStack_HandleMessage_Timeout,
        OALConnectionlessProtocolStack_HandleMessage_BeginNegotiationServerSide,
        OALConnectionlessProtocolStack_HandleMessage_RemoteDomainCallback,
        OALConnectionlessProtocolStack_HandleMessage_ResolveCredsCallback,
        OALConnectionlessProtocolStack_HandleMessage_GetLibraryPermsCallback,
        OALConnectionlessProtocolStack_HandleMessage_RouteSavedEPP
    };

    PCR_Assert( self );
    PCR_Assert( message );
    PCR_Assert( message->recipient == self );
    PCR_Assert( message->msg_type < (sizeof(msg_handler_funcs) / sizeof(OCLPS_MessageHandler_fn_t)) );

    msg_handler_funcs[message->msg_type]( self, message );
}

static ProtocolLayerResult OALConnectionlessProtocolStack_EPP_ENP_Send_fn(OALProtocolStack selfx,
                                                                          BlackboardData blackboardData)
{
    OALConnectionlessProtocolStack self = (OALConnectionlessProtocolStack)selfx;
    ProtocolLayerResult result;

    result = EPP_V2_Send(self->eppV2);
    if ( result != PLRESULT_OKSENDDOWN )
        return result;

    result = ENP_V1_Send(self->enpV1);
    if ( result != PLRESULT_OKSENDDOWN )
        return result;

    /*  If there is an EPP V0 query in progress and we are sending a multicast packet, set the flag to indicate
        EPP_V2 was used. */
    if (blackboardData->associatedChannel->remoteChannelData->flags.EPP_V0_QueryInProgress && blackboardData->flags.isMulticast)
        blackboardData->associatedChannel->remoteChannelData->EPP_V0_QueryVersions |= V0_QUERY_BIT_EPP_V2;

    /*  If there is an ENP V0 query in progress and we are sending a multicast packet, set the flag to indicate
        ENP_V1 was used. */
    if (blackboardData->associatedChannel->remoteChannelData->flags.ENP_V0_QueryInProgress && blackboardData->flags.isMulticast)
        blackboardData->associatedChannel->remoteChannelData->EPP_V0_QueryVersions |= V0_QUERY_BIT_ENP_V1;

    return PLRESULT_OKSENDDOWN;
}

static ProtocolLayerResult OALConnectionlessProtocolStack_ReceiveAppData(OALConnectionlessProtocolStack self,
                                                                         BlackboardData blackboardData,
                                                                         uint16 appProtocolID)
{
    ProtocolLayerResult result;

    switch ( appProtocolID )
    {
    case PROTOCOLID_ESP_V2:
        result = ESP_V2_Receive(blackboardData);
        break;

    case PROTOCOLID_OAP_V1:
        result = OAP_V1_Receive(self->oapV1);
        break;

    case PROTOCOLID_EPP_NEGONE:
        result = EPP_NegOne_Receive(self->eppNegOne);
        break;

    case PROTOCOLID_TEP_V1:
        result = TEP_V1_Receive(self->tepV1);
        break;

    case PROTOCOLID_TRP_V1:
        result = TRP_V1_Receive(self->trpV1);
        break;

    case PROTOCOLID_SGMP_V1:
        result = SGMP_V1_Receive(self->sgmpV1);
        break;
    case PROTOCOLID_CCMMODE_V1:
        result = CCMMODE_V1_Receive(self->ccmModeV1);
        break;

    default:
        result = PLRESULT_ERROR;
        break;
    }

    return result;
}

static void OALConnectionlessProtocolStack_HandleReceiveData(OALConnectionlessProtocolStack self,
                                                             OALStream data,
                                                             DOFAddress srcAddr,
                                                             boolean isMulticast,
                                                             OALChannel associatedChannel)
{
    uint32                      eppLength;
    uint32                      ccpLength;
    uint32                      appProtocolID;
    uint32                      bufLength;
    ProtocolLayerResult         result;
    uint8                       headerByte;
    BlackboardData              blackboardData = &(self->superClass.blackboardData);
    uint32                      dataLen = OALStream_GetDataSize(data);
    OALChannel                  origChannel = associatedChannel;


    BlackboardData_Clear( blackboardData );

    OALStream_Retain( data );
    blackboardData->receiveStream = data;
    blackboardData->transportAddr = DOFAddress_Copy( srcAddr );
    blackboardData->flags.isMulticast = isMulticast;
    blackboardData->associatedChannel = OALChannel_Retain( associatedChannel );
    if (!blackboardData->associatedChannel || !blackboardData->transportAddr || !OALStream_GetDataSize(data))
    {
        goto CleanupAndExit;
    }

    /* Handle ENP V0 */
    headerByte = *(OALStream_GetReadPtr(data));

    if (headerByte == 0x00)
    {
        uint32  enpLength;

        /*  Determine if this is a request or a response.  If it is a request then it will not have any more data
            or the next byte will be zero.  A response will have more data and the byte following the headerByte
            will not be zero.  */
        OALStream_AdjustReadOffset(blackboardData->receiveStream, 1);
        enpLength = OALStream_GetDataSize(data);
        if ((enpLength == 0) || (*(OALStream_GetReadPtr(data))==0x00) )
        {
            /* This packet is a ENP V0 query */
            OALConnectionlessProtocolStack_Handle_ENP_V0_Query(self, blackboardData);
        }
        else
        {
            /*  This is an ENP V0 query response. */
            OALConnectionlessProtocolStack_handle_ENP_V0_QueryResponse(blackboardData);
        }
        goto CleanupAndExit;
    }

    /* Now that we handled queries, we need at least 2 bytes to do anything worth while. */
    if (OALStream_GetDataSize(data) < 2)
        goto CleanupAndExit;


    /* Process ENP */
    switch ( ENP_V1_Receive(self->enpV1) )
    {
    case PLRESULT_ERROR:
    case PLRESULT_NEEDMOREDATA:
    case PLRESULT_OKDONOTHING:
        goto CleanupAndExit;


    case PLRESULT_OKSENDUP:
        break; /* All we have to do here to send data up to EPP is just not return. */

    default:
        PCR_Assert(FALSE); /* Should never ever get here. */
        break;
    }

    /* we need at least 1 byte at this point to do anything worth while. */

    eppLength = OALStream_GetDataSize(blackboardData->receiveStream);
    if (eppLength < 1)
        goto CleanupAndExit;

    /* Handle EPP V0 */
    headerByte = *(OALStream_GetReadPtr( blackboardData->receiveStream ));
    if (headerByte == 0x00)
    {
        OALStream_AdjustReadOffset(blackboardData->receiveStream, 1);
        if (eppLength == 1)
        {
            /* This packet is an EPP V0 query */
            OALConnectionlessProtocolStack_Handle_EPP_V0_Query(self, blackboardData);
        }
        else
        {
            /*  This is an EPP V0 query response. */
            OALConnectionlessProtocolStack_handle_EPP_V0_QueryResponse(blackboardData);
        }
        goto CleanupAndExit;
    }

    /* Process EPP */
    switch ( EPP_V2_Receive(self->eppV2) )
    {
    case PLRESULT_ERROR:
    case PLRESULT_OKNOMOREDATA:
        goto CleanupAndExit;

    case PLRESULT_OKSENDUP:
        break; /* All we have to do here to send data up (to an application layer) is just not return. */

    case PLRESULT_OKSENDDOWN:
        if ( ENP_V1_Send(self->enpV1) != PLRESULT_OKSENDDOWN )
            goto CleanupAndExit;

        /* Send to transport layer. */
        if (DOFTransportLayer_SendTo( blackboardData->associatedChannel->remoteChannelData->dtl, blackboardData->sendPacket, blackboardData->transportAddr))
        {
            /* We just sent a packet, so if there is a maxSendSilenceTimerItem reset it. */
            if(blackboardData->associatedChannel->remoteChannelData->maxSendSilenceTimerItem)
            {
                OALRemoteChannelMgr_SetSendSilenceTimerItem(blackboardData->associatedChannel);
            }
        }
        goto CleanupAndExit;

    case PLRESULT_OKDONOTHING:
        goto CleanupAndExit;

    default:
        PCR_Assert(FALSE); /* Should never ever get here. */
        break;
    }

    /* we need at least 1 byte at this point to do anything worth while. */
    ccpLength = OALStream_GetDataSize(blackboardData->receiveStream);
    if (ccpLength < 1)
        goto CleanupAndExit;

#ifdef OALPROTOCOLSTACK_PACKET_TRACE
    {
        uint32 i;
        uint32 theLen = OALStream_GetDataSize(blackboardData->receiveStream);
        const uint8* ptr = OALStream_GetReadPtr(blackboardData->receiveStream);

        printf("Recv: ");

        for(i = 0; i<theLen; ++i)
        {
            printf("0x%02X ", ptr[i]);
        }

        printf("\n");
    }
#endif

    /* Application Protocol Mux */
    /* peek at the AppID, doesnt modify the stream */
    bufLength = OALStream_GetDataSize(blackboardData->receiveStream);
    appProtocolID = OALMarshal_UncompressValue( 2 /* AppID can be 2 bytes */, &bufLength, OALStream_GetReadPtr(blackboardData->receiveStream));
    if (appProtocolID == OALMARSHAL_INVALID_VALUE)
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "Connectionless remotechannel: Dropping received datagram, V0_QUERY_BIT_OAP_V1 is not set\n");
        goto CleanupAndExit;
    }

    if (appProtocolID != PROTOCOLID_SGMP_V1)
    {
        /*  If the channel is still the server's channel, and the default address is multicast, check to see if there is
            a connection that was created to handle the multicast traffic. */
        if (blackboardData->associatedChannel->remoteChannelData->flags.defaultAddressIsMulticast &&
            !blackboardData->associatedChannel->remoteChannelData->flags.isConnection &&
            blackboardData->associatedChannel->remoteChannelData->baseChannel &&
            blackboardData->associatedChannel->remoteChannelData->baseChannel->remoteChannelData->multicastChannel)
        {
            OALChannel  newChannel;
            OALChannel  previousChannel = blackboardData->associatedChannel;

            newChannel = blackboardData->associatedChannel->remoteChannelData->baseChannel->remoteChannelData->multicastChannel;
            blackboardData->associatedChannel = OALChannel_Retain(newChannel);
            OALChannel_Release(previousChannel);
            if (newChannel && !blackboardData->associatedChannel)
            {
                goto CleanupAndExit;
            }
        }
    }

    if (blackboardData->associatedChannel->remoteChannelData->flags.waitingForSecurityIndicator)
    {
        if (blackboardData->associatedChannel->remoteChannelData->flags.isOutbound)
        {
            blackboardData->associatedChannel->remoteChannelData->flags.waitingForInitialDatagram = FALSE;
        }
        else
        {
            /* See if we are still waiting for the first app PDU to determine if the session will be secure. */
            if ((appProtocolID == PROTOCOLID_TEP_V1) || (appProtocolID == PROTOCOLID_TRP_V1))
            {
                blackboardData->associatedChannel->remoteChannelData->flags.waitingForSecurityIndicator = FALSE;
                blackboardData->associatedChannel->remoteChannelData->flags.secureDesire = TRUE;
                OALRemoteChannel_NotifyInitComplete(blackboardData->associatedChannel);
            }
            else if ((appProtocolID != PROTOCOLID_EPP_NEGONE) && (appProtocolID != PROTOCOLID_ESP_V2))
            {
                blackboardData->associatedChannel->remoteChannelData->flags.waitingForSecurityIndicator = FALSE;
                /*  Since the first PDU (other than EPP neg one or ESP) was not TEP or TRP this session will be unsecure.
                    Clear the TEP bit so that TEP will be dropped. */
                blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids &= ~V0_QUERY_BIT_TEP_V1;
                blackboardData->associatedChannel->remoteChannelData->flags.noSessionSecurity = TRUE;
                OALRemoteChannel_NotifyInitComplete(blackboardData->associatedChannel);
            }
            /* Leave waitingForSecurityIndicator set if it is a EPP negone packet. */
        }
    }

    /* At this point the channel is set to the channel associated with the connection or server that is configured
       to process the remainder of the packet.  The address, or security state ID was used to look up this channel
       and the associated channel may have been changed in enp receive or epp recieve. */
    if (blackboardData->associatedChannel->remoteChannelData->maxReceiveSilenceTimerItem)
        OALRemoteChannelMgr_SetRecvSilenceTimerItem( blackboardData->associatedChannel );

    OALRemoteChannel_IncByteCount( blackboardData->associatedChannel, dataLen);
    dataLen = 0;

    result = PLRESULT_OKDONOTHING;
    if ( blackboardData->associatedChannel->remoteChannelData->flags.isOutbound &&
         !blackboardData->associatedChannel->remoteChannelData->flags.defaultAddressIsMulticast )
    {
        /* Incoming multicast SGMP PDUs do NOT take this path because .defaultAddressIsMulticast is TRUE.
           Incoming point SGMP PDUs DO take this path because .defaultAddressIsMulticast is FALSE. */
        switch ( appProtocolID )
        {
        case PROTOCOLID_OAP_V1:
            if (!(blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids & V0_QUERY_BIT_OAP_V1))
            {
                DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "Connectionless remotechannel: Dropping received datagram, V0_QUERY_BIT_OAP_V1 is not set\n");
                result = PLRESULT_ERROR;
            }
            break;

        case PROTOCOLID_EPP_NEGONE:
            break;

        case PROTOCOLID_TEP_V1:
            if (!(blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids & V0_QUERY_BIT_TEP_V1))
            {
                DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "Connectionless remotechannel: Dropping received datagram, V0_QUERY_BIT_TEP_V1 is not set\n");
                result = PLRESULT_ERROR;
            }
            break;

        case PROTOCOLID_TRP_V1:
            if (blackboardData->flags.isCommandNotResponse)
            {
                if (!(blackboardData->associatedChannel->remoteChannelData->flags.TRP_AcceptsCommands))
                {
                    DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "Connectionless remotechannel: Dropping received TRP command, TRP_AcceptsCommands is not set\n");
                    result = PLRESULT_ERROR;
                }
            }
            else
            {
                /* We should not recieve a response if the remote node does not accept commands. */
                if (!(blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids & V0_QUERY_BIT_TRP_V1))
                {
                    DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "Connectionless remotechannel: Dropping received TRP response, V0_QUERY_BIT_TRP_V1 is not set\n");
                    result = PLRESULT_ERROR;
                }
            }
            break;

        case PROTOCOLID_SGMP_V1:
            if (!(blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids & V0_QUERY_BIT_SGMP_V1))
            {
                DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "Connectionless remotechannel: Dropping received datagram, V0_QUERY_BIT_SGMP_V1 is not set\n");
                result = PLRESULT_ERROR;
            }
            break;
        case PROTOCOLID_CCMMODE_V1:
            if (!(blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids & V0_QUERY_BIT_CCMMODE_V1))
            {
                DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "Connectionless remotechannel: Dropping received datagram, V0_QUERY_BIT_CCMMODE_V1 is not set\n");
                result = PLRESULT_ERROR;
            }
            break;

        default:
            break;
        }
    }
    else if ( !blackboardData->associatedChannel->remoteChannelData->flags.isOutbound &&
              !blackboardData->associatedChannel->remoteChannelData->flags.defaultAddressIsMulticast )
    {
        /* If TEP is received on an inbound channel and the noSessionSecurity flag is set then drop this PDU. */
        if (appProtocolID == PROTOCOLID_TEP_V1 && blackboardData->associatedChannel->remoteChannelData->flags.noSessionSecurity)
        {
            DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "Connectionless remotechannel: Dropping received datagram, TEP received after noSessionSecurity was set.");
            result = PLRESULT_ERROR;
        }
    }

    /* the associated channel in blackboardData is the correct channel to process this PDU. Proceed to call the function
       for the given protocol ID. */
    if ( result != PLRESULT_ERROR )
        result = OALConnectionlessProtocolStack_ReceiveAppData(self, blackboardData, appProtocolID);

    if ((result != PLRESULT_ERROR) &&
        blackboardData->associatedChannel &&
        !blackboardData->associatedChannel->remoteChannelData->flags.isOutbound)
    {
        /* If this is an inbound channel set the query bit to indicate that protocol */
        switch ( appProtocolID )
        {
        case PROTOCOLID_OAP_V1:
            if (!(blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids & V0_QUERY_BIT_OAP_V1))
            {
                DOFRouter_ChannelUp(blackboardData->associatedChannel);
                blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids |= V0_QUERY_BIT_OAP_V1;
            }
            break;

        case PROTOCOLID_EPP_NEGONE:
            break;

        case PROTOCOLID_TEP_V1:
            blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids |= V0_QUERY_BIT_TEP_V1;
            break;

        case PROTOCOLID_TRP_V1:
            /* Set the bit if this is a multicast channel, otherwise do not set the bit.  For unicast the remote
               node may send TRP commands, but may not be able to accept TRP commands so do not set the bit. */
            if (blackboardData->associatedChannel->remoteChannelData->flags.defaultAddressIsMulticast)
            {
                blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids |= V0_QUERY_BIT_TRP_V1;
            }
            break;

        case PROTOCOLID_SGMP_V1:
            blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids |= V0_QUERY_BIT_SGMP_V1;
            break;
        case PROTOCOLID_CCMMODE_V1:
            blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids |= V0_QUERY_BIT_CCMMODE_V1;
            break;

        default:
            break;
        }
    }


    /* Process result */
    switch ( result )
    {
    case PLRESULT_ERROR:
        /* do something besides drop packet? probably not. */
        break;

    case PLRESULT_OKSENDDOWN:
        if (OALConnectionlessProtocolStack_EPP_ENP_Send_fn((OALProtocolStack)self, blackboardData) != PLRESULT_OKSENDDOWN)
            break;

        /* Send to transport layer. */
        if (DOFTransportLayer_SendTo( blackboardData->associatedChannel->remoteChannelData->dtl, blackboardData->sendPacket, blackboardData->transportAddr))
        {
            /* We just sent a packet, so if there is a maxSendSilenceTimerItem reset it. */
            if(blackboardData->associatedChannel->remoteChannelData->maxSendSilenceTimerItem)
            {
                OALRemoteChannelMgr_SetSendSilenceTimerItem(blackboardData->associatedChannel);
            }
        }
        break;

    case PLRESULT_OKDONOTHING:
        break; /* Do nothing. */
    case PLRESULT_OKNOMOREDATA:
        break; /* This should not happen on connectinless so treat it as a bad packet and drop it. */

    default:
        PCR_Assert(FALSE); /* Should never ever get here. */
        break;
    }

    CleanupAndExit:
    if( dataLen > 0 )
    {
        OALRemoteChannel_IncByteCount(origChannel, dataLen);
    }

    BlackboardData_Clear( blackboardData );

    if(self->superClass.keyChangeCallback)
    {
        if ( !self->superClass.keyChangeCallback( self->superClass.keyChangeCallbackContext ) )
            OALRemoteChannel_Shutdown( associatedChannel, DOFEXCEPTION_INSUFFICIENT_RESOURCES );
        self->superClass.keyChangeCallback = NULL;
        self->superClass.keyChangeCallbackContext = NULL;
    }
    BlackboardData_Clear( blackboardData );

    return;
}

static boolean OALConnectionlessProtocolStack_Send(OALConnectionlessProtocolStack self,
                                                ProtocolLayerIndicator_t protocolLayerIndicator,
                                                uint16 attributeData,
                                                DOFAddress destAddr,
                                                OALChannel associatedChannel,
                                                uint32 argc,
                                                void* argv[])
{
    boolean rval = FALSE;
    uint32  appidQuery = 0;
    BlackboardData blackboardData = &(self->superClass.blackboardData);

    BlackboardData_Clear( blackboardData );

    blackboardData->associatedChannel = OALChannel_Retain( associatedChannel );

    if (associatedChannel && !blackboardData->associatedChannel)
    {
        goto CleanupAndExit;
    }

    blackboardData->sendPacket = OALPacket_Create();

    if (!blackboardData->sendPacket)
        goto CleanupAndExit;

    blackboardData->transportAddr = DOFAddress_Copy(destAddr);

    if (destAddr && ! blackboardData->transportAddr)
        goto CleanupAndExit;

    switch (protocolLayerIndicator)
    {
    case PLI_OAP:
        /* When sent secure, all cases of unicast send need replay protection, so do not set
           ->flags.denyPacketReplayProtectionIfSecureUnicastSend. */
        PCR_Assert( argc == 4 );
        blackboardData->providerOp = (DOFOperation)argv[1];
        if (blackboardData->providerOp)
            DOFOperation_Retain(blackboardData->providerOp);
        if (!self->oapV1 || OAP_V1_Send(self->oapV1, (DOFOperation)argv[0], (boolean)((int)argv[2]), (DOFResult)argv[3]) != PLRESULT_OKSENDDOWN )
            goto CleanupAndExit;
        appidQuery = V0_QUERY_BIT_OAP_V1;
        break;

    case PLI_TEP:
        /* Could be sent secure in future, so do not set ->flags.denyPacketReplayProtectionIfSecureUnicastSend. */
        PCR_Assert( argc == 3 );
        if (!self->tepV1 || TEP_V1_Send(self->tepV1, (DOFOperation)argv[0], (boolean)((int)argv[1]), (DOFResult)argv[2]) != PLRESULT_OKSENDDOWN )
            goto CleanupAndExit;
        appidQuery = V0_QUERY_BIT_TEP_V1;
        break;

    case PLI_TRP:
        /* Could be sent secure in future, so do not set ->flags.denyPacketReplayProtectionIfSecureUnicastSend. */
        PCR_Assert( argc == 3 );
        if ( !self->trpV1 || TRP_V1_Send(self->trpV1, (DOFOperation)argv[0], (boolean)((int)argv[1]), (DOFResult)argv[2]) != PLRESULT_OKSENDDOWN )
            goto CleanupAndExit;
        appidQuery = V0_QUERY_BIT_TRP_V1;
        break;

    case PLI_SGMP:
        {
            /* We are supposed to know targetNodeID of our destination, so do not set
               ->flags.denyPacketReplayProtectionIfSecureUnicastSend for normal PDUs.
               REVIEW: Note, however, that a hub wanting to send multiple unicast rekey or epochChanged PDUs to
               different point members, is motivated to call OALEncipher() one time, setting
               denyPacketReplayProtectionIfSecureUnicastSend for that call. Code needs be added to get that send PDU
               thru the protocol stack and out on the wire. When this happens it might be that those PDUs do not pass
               thru this block, so we do not need to set ->flags.denyPacketReplayProtectionIfSecureUnicastSend here. */
            PCREvent evt = NULL;

            if ( !self->sgmpV1 || SGMP_V1_Send(self->sgmpV1, (DOFOperation)argv[0], (boolean)((int)argv[1]), (DOFResult)argv[2],
                              &evt) != PLRESULT_OKSENDDOWN )
                goto CleanupAndExit;
            if (evt)
                OALProtocolStack_SetKeyChangeCallback( &self->superClass, OALConnectionlessProtocolStack_SGMP_KeyChangedCallback, evt );
            appidQuery = V0_QUERY_BIT_SGMP_V1;
        }
        break;

    case PLI_CCMMODE:
        {
            /* CCM.Query PDU is sent secure/unicast and unmarked */
            blackboardData->flags.denyPacketReplayProtectionIfSecureUnicastSend = 1;
            if ( CCMMODE_V1_Send(self->ccmModeV1, (DOFOperation)argv[0], (boolean)((int)argv[1]), (DOFResult)argv[2]) != PLRESULT_OKSENDDOWN )
                goto CleanupAndExit;

            appidQuery = V0_QUERY_BIT_CCMMODE_V1;
        }
        break;

    case PLI_ESP:
        {
            if (ESP_V2_Send(blackboardData, (DOFOperation)argv[0]) != PLRESULT_OKSENDDOWN)
                goto CleanupAndExit;
        }
        break;

    case PLI_EPP_NEGONE:
        {
            /* Setting of blackboardeData->flags.denyPacketReplayProtectionIfSecureUnicastSend is done within this call. */
            if ( EPP_NegOne_Send(self->eppNegOne, (DOFOperation)argv[0]) != PLRESULT_OKSENDDOWN )
                goto CleanupAndExit;
        }
        break;
    default:
        break;
    }

#ifdef OALPROTOCOLSTACK_PACKET_TRACE
    {
        uint32 i = 0;
        uint32 theLen = OALPacket_GetTotalStreamSize(blackboardData->sendPacket);
        uint8* ptr = DOFMem_Alloc(theLen);

        OALPacket_CopyTo(blackboardData->sendPacket, ptr, 0, theLen);

        printf("Send: ");
        for(; i<theLen; ++i)
        {
            printf("0x%02X ", ptr[i]);
        }
        printf("\n");

        DOFMem_Free(ptr);
    }
#endif

    /* Verify that the packet can be sent to blackboardData->transportAddr. */
    if (blackboardData->flags.datagramProtocolP2PNotAllowed)
    {
        /* Datagram point to point is not allowed, verify that the address is multicast. */
        if ((!blackboardData->transportAddr || (blackboardData->transportAddr == blackboardData->associatedChannel->remoteChannelData->address)) &&
            !blackboardData->associatedChannel->remoteChannelData->flags.defaultAddressIsMulticast)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: PDU not allowed over datagram point to point." );
            goto CleanupAndExit;
        }
        else
        {
            /* The address is not the one in the channel.  Check to see what the address is. */
            /* a NULL address indicates multicast. */
            if (blackboardData->transportAddr && (DOFAddress_GetAddressType(blackboardData->transportAddr) == DOFADDRESSTYPE_UNICAST))
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: PDU not allowed over datagram point to point." );
                goto CleanupAndExit;
            }
        }
    }
    if (blackboardData->flags.multicastProtocolNotAllowed)
    {
        /* Verify that the address is not multicast. */
        if ((!blackboardData->transportAddr || (blackboardData->transportAddr == blackboardData->associatedChannel->remoteChannelData->address)) &&
            blackboardData->associatedChannel->remoteChannelData->flags.defaultAddressIsMulticast)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: PDU not allowed over datagram multicast." );
            goto CleanupAndExit;
        }
        else
        {
            /* The address is not the one in the channel.  Check to see what the address is. */
            if (blackboardData->transportAddr && (DOFAddress_GetAddressType(blackboardData->transportAddr) != DOFADDRESSTYPE_UNICAST))
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: PDU not allowed over datagram multicast." );
                goto CleanupAndExit;
            }
        }
    }

    /* Add the EPP and ENP data.*/
    if (OALConnectionlessProtocolStack_EPP_ENP_Send_fn((OALProtocolStack)self, blackboardData) == PLRESULT_OKSENDDOWN)
    {
        /* If there is and EPS query in progress set the bit for the APPID used. */
        if (blackboardData->associatedChannel->remoteChannelData->flags.ESP_V0_QueryInProgress &&
            blackboardData->flags.isMulticast)
           blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids |= appidQuery;

        /* Send to Transport Layer */
        /* If there is an address in blackboard data use it, otherwise get the address from the channel. */
        if (DOFTransportLayer_SendTo( associatedChannel->remoteChannelData->dtl, blackboardData->sendPacket,
                                       blackboardData->transportAddr ? blackboardData->transportAddr
                                                                     : associatedChannel->remoteChannelData->address))
        {
            rval = TRUE;
            /* We just sent a packet, so if there is a maxSendSilenceTimerItem reset it. */
            if(blackboardData->associatedChannel->remoteChannelData->maxSendSilenceTimerItem)
            {
                OALRemoteChannelMgr_SetSendSilenceTimerItem(blackboardData->associatedChannel);
            }
        }
        else
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "OALConnectionlessProtocolStack: failed to send." );
        }
    }

    CleanupAndExit:
    if(self->superClass.keyChangeCallback)
    {
        if ( !self->superClass.keyChangeCallback( self->superClass.keyChangeCallbackContext ) )
            OALRemoteChannel_Shutdown( associatedChannel, DOFEXCEPTION_INSUFFICIENT_RESOURCES );
        self->superClass.keyChangeCallback = NULL;
        self->superClass.keyChangeCallbackContext = NULL;
    }
    BlackboardData_Clear( blackboardData );

    return rval;
}



static void* OALConnectionlessProtocolStack_TEP_V1_SessionEventCallback(void               * context,
                                                                        uint8              * key,
                                                                        TEP_SessionEvent_t   eventType,
                                                                        DOFSecurityMode      securityMode,
                                                                        OALChannel           channel,
                                                                        DOFSecurityScope     secScope,
                                                                        DOFPermissionSet     perms,
                                                                        DOFPermissionSet     otherPerms,
                                                                        DOFException         exception,
                                                                        uint32               stateID,
                                                                        DOFAddress           remoteAddr,
                                                                        OALStream          * outStream,
                                                                        DOFCredentials       creds)
{
    OALConnectionlessRemoteChannel remoteChannel;
    TEP_V1_SessionEventCallback channelSessionEventCallback;

    remoteChannel = context ? (OALConnectionlessRemoteChannel)context : (OALConnectionlessRemoteChannel)channel;
    channelSessionEventCallback = OALConnectionlessRemoteChannel_GetSessionEventCallback( remoteChannel );

    return channelSessionEventCallback(context, key, eventType, securityMode, (OALChannel)remoteChannel,
                                       secScope, perms, otherPerms, exception, stateID, remoteAddr, outStream, creds);
}
