/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>

#include <internal/blackboarddata.h>
#include <internal/channel.h>
#include <internal/channel_stack_common.h>
#include <internal/connectionless_protocolstack.h>
#include <internal/connectionless_remotechannel.h>
#include <internal/enp_v1.h>
#include <internal/marshal.h>
#include <internal/operationinternal.h>
#include <internal/packet.h>
#include <internal/protocollayerresult.h>
#include <internal/remotechannel.h>
#include <internal/remotechanneldata.h>
#include <internal/remotechannelmanager.h>
#include <internal/stream.h>
#include <internal/timer.h>
#include <internal/transportinternal.h>
#include <internal/dofinternal.h>

#include "connectionless_protocolstack_struct.h"
#include "connectionless_v0query.h"


static OALTimer protocolQueryTimer = NULL;


static int protocolTimerSendResponse(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown)
{
    protocolQueryNode_t *arg = (protocolQueryNode_t*)argv[0];
    boolean sendQueryResponse = TRUE;

    PCR_Assert(argc==1);

    if(!shutdown)
    {
        switch (arg->queryType)
        {
            case QUERY_TYPE_ENP_V0:
                if ((arg->channel->remoteChannelData->ENP_V0_QueryVersions & V0_QUERY_ENP_BITMASK) == V0_QUERY_ENP_BITMASK)
                    sendQueryResponse = FALSE;
                arg->channel->remoteChannelData->flags.ENP_V0_QueryInProgress = FALSE;
                break;
            case QUERY_TYPE_EPP_V0:
                if ((arg->channel->remoteChannelData->EPP_V0_QueryVersions & V0_QUERY_EPP_BITMASK) == V0_QUERY_EPP_BITMASK)
                    sendQueryResponse = FALSE;
                arg->channel->remoteChannelData->flags.EPP_V0_QueryInProgress = FALSE;
                break;
            case QUERY_TYPE_ESP_V2:
                if ((arg->channel->remoteChannelData->ESP_V0_QueryAppids & ESP_V0_QUERY_BITMASK) == ESP_V0_QUERY_BITMASK)
                    sendQueryResponse = FALSE;
                arg->channel->remoteChannelData->flags.ESP_V0_QueryInProgress = FALSE;
                break;
            default:
                sendQueryResponse = FALSE;
        }

        if (sendQueryResponse)
        {
            if (DOFTransportLayer_SendTo(arg->channel->remoteChannelData->dtl, arg->sendPacket, NULL))
            {
                /* We just sent a packet, so if there is a maxSendSilenceTimerItem reset it. */
                if(arg->channel->remoteChannelData->maxSendSilenceTimerItem)
                {
                    OALRemoteChannelMgr_SetSendSilenceTimerItem(arg->channel);
                }
            }
        }
    }

    OALChannel_Release(arg->channel);
    OALPacket_Release(arg->sendPacket);
    DOFMem_Free(arg);

    return 0;
}

boolean connectionlessV0Query_AddDelayedResponse(BlackboardData blackboardData)
{
    protocolQueryNode_t *   queryTimerArg;
    uint32                  queryResponseDelay;

    /*  When a multicast query is received, set a random timer between zero and three minutes.
        Start tracking APPIDs as packets and query responses are received.  When the timer
        expires only send a query response if there is knowledge of APPIDs that were not recieved
        in a packet or query response. */

    /* If a query is already in progress exit now. */
    if (blackboardData->associatedChannel->remoteChannelData->flags.ESP_V0_QueryInProgress)
        return FALSE;
    blackboardData->associatedChannel->remoteChannelData->flags.ESP_V0_QueryInProgress = TRUE;
    queryResponseDelay = PCRRandom_Get();
    queryResponseDelay %= MAX_QUERY_DELAY;

    queryTimerArg = DOFMem_Alloc(sizeof(protocolQueryNode_t));
    if (!queryTimerArg)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: handle ESP V0 query: Failed to create protocol query node." );
        return FALSE;
    }

    queryTimerArg->channel = OALChannel_Retain(blackboardData->associatedChannel);
    if (!queryTimerArg->channel)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: handle ESP V0 query: Failed to create protocol query node." );
        DOFMem_Free(queryTimerArg);
        return FALSE;
    }

    OALPacket_Retain(blackboardData->sendPacket);
    queryTimerArg->sendPacket = blackboardData->sendPacket;
    queryTimerArg->queryType = QUERY_TYPE_ESP_V2;

    /* Set the bit field to all zeros and as packets and query responses are received set the bit for the version used. */
    blackboardData->associatedChannel->remoteChannelData->ESP_V0_QueryAppids = 0;
    if(!OALTimer_Insert(protocolQueryTimer, queryResponseDelay, protocolTimerSendResponse,
                        1, (void**)&queryTimerArg, FALSE))
    {
        OALChannel_Release(blackboardData->associatedChannel);
        OALPacket_Release(blackboardData->sendPacket);
        DOFMem_Free(queryTimerArg);
        return FALSE;
    }
    return TRUE;
}

boolean OALConnectionlessProtocolStack_Init(void)
{
    if ((protocolQueryTimer = OALTimer_Create(DOF_GetDof()->threadPool, TRUE)) == NULL)
        return FALSE;                   /* Error out. */

    return TRUE;                        /* Success. */
}

void OALConnectionlessProtocolStack_Shutdown(void)
{
    if (protocolQueryTimer)
        OALTimer_Destroy(protocolQueryTimer);
}

boolean OALConnectionlessProtocolStack_Send_ENP_V0_Query(OALConnectionlessProtocolStack self, OALChannel channel)
{
    uint8           enpV0Command[] = {0x00};
    OALStream       enpCommand = OALStream_CreateRW(1);
    OALPacket       sendPacket;

    if (enpCommand == NULL)
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: send ENP V0 query: failed to create ENP command stream.");
        return FALSE;
    }
    if (!OALStream_Write(enpCommand, 1, enpV0Command))
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: send ENP V0 query: failed to write to ENP command stream.");
        OALStream_Release(enpCommand, FALSE);
        return FALSE;
    }
    sendPacket = OALPacket_Create();
    if (!sendPacket)
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: send ENP V0 query: failed to create ENP command packet.");
        OALStream_Release(enpCommand, FALSE);
        return FALSE;
    }
    if (!OALPacket_AppendStream(sendPacket, enpCommand))
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: send ENP V0 query: failed to append to ENP command packet.");
        OALPacket_Release(sendPacket);
        OALStream_Release(enpCommand, FALSE);
        return FALSE;
    }

    if (DOFTransportLayer_SendTo( channel->remoteChannelData->dtl, sendPacket, channel->remoteChannelData->address))
    {
        /* We just sent a packet, so if there is a maxSendSilenceTimerItem reset it. */
        if(channel->remoteChannelData->maxSendSilenceTimerItem)
        {
            OALRemoteChannelMgr_SetSendSilenceTimerItem(channel);
        }
    }

    OALPacket_Release(sendPacket);

    return TRUE;
}

boolean OALConnectionlessProtocolStack_Handle_ENP_V0_Query(OALConnectionlessProtocolStack self,
                                                          BlackboardData blackboardData)
{
    /*  Reply with 0x00 to indicate ENP V0, followed by 1 byte for each supported ENP version.
        As new ENP versions are added they must be added to this enpV0Response. */
    uint8 enpV0Response[] = { 0x00, ENP_V1_VERSION };
    OALStream enpResponse = OALStream_CreateRW(2);

    if (enpResponse == NULL)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: handle ENP V0 query: Failed to create ENP query response stream." );
        return FALSE;
    }
    if (!OALStream_Write(enpResponse, 2, enpV0Response))
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: handle ENP V0 query: Failed to write to ENP query response stream." );
        OALStream_Release(enpResponse, FALSE);
        return FALSE;
    }

    if (!blackboardData->sendPacket)
        blackboardData->sendPacket = OALPacket_Create();
    if (!blackboardData->sendPacket)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: handle ENP V0 query: Failed to Create ENP query response packet." );
        OALStream_Release(enpResponse, FALSE);
        return FALSE;
    }
    if(!OALPacket_AppendStream( blackboardData->sendPacket, enpResponse ))
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: handle ENP V0 query: Failed to append stream data to enp query response packet." );
        OALStream_Release(enpResponse, FALSE);
        return FALSE;
    }

    /* If the query was received multicast send the response multicast. */
    if (blackboardData->flags.isMulticast)
    {
        protocolQueryNode_t *   queryTimerArg;
        uint32                  queryResponseDelay;

        /*  When a multicast query is received, set a random timer between zero and three minutes.
            Start tracking EPP versions as packets and query responses are received.  When the timer
            expires only send a query response if there is knowledge of versions of ENP that were not recieved
            in a packet or query response. */

        /* If a query is already in progress exit now. */
        if (blackboardData->associatedChannel->remoteChannelData->flags.ENP_V0_QueryInProgress)
            return FALSE;
        blackboardData->associatedChannel->remoteChannelData->flags.ENP_V0_QueryInProgress = TRUE;
        queryResponseDelay = PCRRandom_Get();
        queryResponseDelay %= MAX_QUERY_DELAY;

        queryTimerArg = DOFMem_Alloc(sizeof(protocolQueryNode_t));
        if (queryTimerArg == NULL)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: handle ENP V0 query: Failed to retain channel." );
            return FALSE;
        }

        queryTimerArg->channel = OALChannel_Retain(blackboardData->associatedChannel);
        if (!queryTimerArg->channel)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: handle ENP V0 query: Failed to allocate queryTimerArg." );
            DOFMem_Free(queryTimerArg);
            return FALSE;
        }


        OALPacket_Retain(blackboardData->sendPacket);
        queryTimerArg->sendPacket = blackboardData->sendPacket;
        queryTimerArg->queryType = QUERY_TYPE_ENP_V0;

        /* Set the bit field to all zeros and as packets and query responses are received set the bit for the version used. */
        blackboardData->associatedChannel->remoteChannelData->ENP_V0_QueryVersions = 0;
        if(!OALTimer_Insert(protocolQueryTimer, queryResponseDelay, protocolTimerSendResponse,
                            1, (void**)&queryTimerArg, FALSE))
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: handle ENP V0 query: Failed to insert query timer item." );
            OALChannel_Release(queryTimerArg->channel);
            OALPacket_Release(queryTimerArg->sendPacket);
            DOFMem_Free(queryTimerArg);
            return FALSE;
        }
    }
    else
    {
        /* If this is a unicast query respond immediately. */
        if (DOFTransportLayer_SendTo( blackboardData->associatedChannel->remoteChannelData->dtl, blackboardData->sendPacket, blackboardData->transportAddr))
        {
            /* We just sent a packet, so if there is a maxSendSilenceTimerItem reset it. */
            if(blackboardData->associatedChannel->remoteChannelData->maxSendSilenceTimerItem)
            {
                OALRemoteChannelMgr_SetSendSilenceTimerItem(blackboardData->associatedChannel);
            }
        }
    }
    return TRUE;
}

boolean OALConnectionlessProtocolStack_handle_ENP_V0_QueryResponse(BlackboardData blackboardData)
{
    /*  Set the flag for the versions received that we use.  We may not be required to send a response if
        all known versions are used before we are required to send our response. */
    uint8   enpVersion;
    while (OALStream_Read8(blackboardData->receiveStream, &enpVersion))
    {
        switch (enpVersion)
        {
        case ENP_V1_VERSION:
            blackboardData->associatedChannel->remoteChannelData->ENP_V0_QueryVersions |= V0_QUERY_BIT_ENP_V1;
            break;
        /* As new versions of ENP are added they must be included in this switch statement. */
        default:
            break;
        }
    }

    return TRUE;
}

boolean OALConnectionlessProtocolStack_Send_EPP_V0_Query(OALConnectionlessProtocolStack self, OALChannel channel)
{
    uint8           eppV0Command[] = {0x01, 0x00};
    OALStream       eppCommand = OALStream_CreateRW(2);
    OALPacket       sendPacket;

    /* Note: this is hard coded to use ENP_V1 for now.  At some point it should be changed to use ENP_V1 send or the specific ENP version used. */
    if (eppCommand == NULL)
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: send EPP V0 query: failed to create EPP command stream.");
        return FALSE;
    }
    if (!OALStream_Write(eppCommand, 2, eppV0Command))
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: send EPP V0 query: failed to write to EPP command stream.");
        OALStream_Release(eppCommand, FALSE);
        return FALSE;
    }
    sendPacket = OALPacket_Create();
    if (!sendPacket)
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: send EPP V0 query: failed to create EPP command packet.");
        OALStream_Release(eppCommand, FALSE);
        return FALSE;
    }
    if (!OALPacket_AppendStream(sendPacket, eppCommand))
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: send EPP V0 query: failed to append to EPP command packet.");
        OALStream_Release(eppCommand, FALSE);
        OALPacket_Release(sendPacket);
        return FALSE;
    }

    if (DOFTransportLayer_SendTo( channel->remoteChannelData->dtl, sendPacket, channel->remoteChannelData->address))
    {
        /* We just sent a packet, so if there is a maxSendSilenceTimerItem reset it. */
        if(channel->remoteChannelData->maxSendSilenceTimerItem)
        {
            OALRemoteChannelMgr_SetSendSilenceTimerItem(channel);
        }
    }

    OALPacket_Release(sendPacket);

    return TRUE;
}


boolean OALConnectionlessProtocolStack_Handle_EPP_V0_Query(OALConnectionlessProtocolStack self,
                                                          BlackboardData blackboardData)
{
    /*  Reply with 0x00 to indicate EPP V0, followed by 1 byte for each supported EPP version.
        As new EPP versions are added they must be included here. */
    const uint8 eppV0Response[] = { 0x00, EPPV2_VERSION };
    OALStream eppResponse = OALStream_CreateRW(2);

    if (eppResponse == NULL)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: handle EPP V0 query: Failed to create epp query response stream." );
        return FALSE;
    }
    OALStream_Write(eppResponse, 2, eppV0Response);
    if (!blackboardData->sendPacket)
        blackboardData->sendPacket = OALPacket_Create();
    if (!blackboardData->sendPacket)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: handle EPP V0 query: Failed to create packet for EPP response." );
        OALStream_Release(eppResponse, FALSE);
        return FALSE;
    }
    if(!OALPacket_AppendStream( blackboardData->sendPacket, eppResponse ))
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: handle EPP V0 query: Failed to append epp query response to stream." );
        OALStream_Release(eppResponse, FALSE);
        return FALSE;
    }

    if ( ENP_V1_Send(self->enpV1) != PLRESULT_OKSENDDOWN )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: handle EPP V0 query: Failed to send epp query response via enp." );
        return FALSE;
    }

    /* If the query was received multicast send the response multicast. */
    if (blackboardData->flags.isMulticast)
    {
        protocolQueryNode_t *   queryTimerArg;
        uint32                  queryResponseDelay;

        /*  When a multicast query is received, set a random timer between zero and three minutes.
            Start tracking EPP versions as packets and query responses are received.  When the timer
            expires only send a query response if there is knowledge of versions of EPP that were not recieved
            in a packet or query response. */

        /* If a query is already in progress exit now. */
        if (blackboardData->associatedChannel->remoteChannelData->flags.EPP_V0_QueryInProgress)
            return FALSE;
        blackboardData->associatedChannel->remoteChannelData->flags.EPP_V0_QueryInProgress = TRUE;
        queryResponseDelay = PCRRandom_Get();
        queryResponseDelay %= MAX_QUERY_DELAY;

        queryTimerArg = DOFMem_Alloc(sizeof(protocolQueryNode_t));
        if (!queryTimerArg)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: handle EPP V0 query: Failed to allocate queryTimerArg for epp query." );
            return FALSE;
        }

        queryTimerArg->channel = OALChannel_Retain(blackboardData->associatedChannel);
        if (!queryTimerArg->channel)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: handle EPP V0 query: Failed to retain channel for epp query." );
            DOFMem_Free(queryTimerArg);
            return FALSE;
        }

        OALPacket_Retain(blackboardData->sendPacket);
        queryTimerArg->sendPacket = blackboardData->sendPacket;
        queryTimerArg->queryType = QUERY_TYPE_EPP_V0;

        /* Set the bit field to all zeros and as packets and query responses are received set the bit for the version used. */
        blackboardData->associatedChannel->remoteChannelData->EPP_V0_QueryVersions = 0;
        if(!OALTimer_Insert(protocolQueryTimer, queryResponseDelay, protocolTimerSendResponse,
                            1, (void**)&queryTimerArg, FALSE))
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALConnectionlessProtocolStack: handle EPP V0 query: Failed to insert epp query response timer item." );
            OALChannel_Release(queryTimerArg->channel);
            OALPacket_Release(queryTimerArg->sendPacket);
            DOFMem_Free(queryTimerArg);
            return FALSE;
        }
    }
    else
    {
        /* If this is a unicast query respond immediately. */
        if (DOFTransportLayer_SendTo(blackboardData->associatedChannel->remoteChannelData->dtl, blackboardData->sendPacket, blackboardData->transportAddr))
        {
            /* We just sent a packet, so if there is a maxSendSilenceTimerItem reset it. */
            if(blackboardData->associatedChannel->remoteChannelData->maxSendSilenceTimerItem)
            {
                OALRemoteChannelMgr_SetSendSilenceTimerItem(blackboardData->associatedChannel);
            }
        }
    }
    return TRUE;
}

boolean OALConnectionlessProtocolStack_handle_EPP_V0_QueryResponse(BlackboardData blackboardData)
{
    /*  Set the flag for the versions received that we use.  We may not be required to send a response if
        all known versions are used before we are required to send our response. */
    uint8   eppVersion;
    while (OALStream_Read8(blackboardData->receiveStream, &eppVersion))
    {
        switch (eppVersion)
        {
        case EPPV2_VERSION:
            blackboardData->associatedChannel->remoteChannelData->EPP_V0_QueryVersions |= V0_QUERY_BIT_EPP_V2;
            break;
        /* As new versions of EPP are added they must be included in this switch statement. */
        default:
            break;
        }
    }

    return TRUE;
}

