/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/error.h>

#include <internal/channel_stack_messages.h>
#include <internal/connected_protocolstack.h>
#include <internal/exception.h>
#include <internal/protocolstack.h>
#include <internal/remotechannel.h>
#include <internal/remotechannelmanager.h>

extern PCRMutex ref_count_mutex;

static OALTimer protocolstack_timer;
static OALThreadPool pool;

void OALProtocolStack_Init(void)
{
    pool = OALThreadPool_Create(1);
    protocolstack_timer = OALTimer_Create(pool, FALSE); /* it's really synchronous, but threadpool has 1 thread */
    PCR_Assert(protocolstack_timer);
}

void OALProtocolStack_Cleanup(void)
{
    OALTimer_Destroy(protocolstack_timer);
    protocolstack_timer = NULL;
    OALThreadPool_Destroy(pool);
    pool = NULL;
}

static int send_negotiation_timeout(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown)
{
    OALConnectedProtocolStack self = (OALConnectedProtocolStack)argv[0];

    PCR_Assert(argc==1);
    if(!shutdown)
    {
        OALMessage message = SMT_Connected_Negotiation_Timeout_Create(self);
        if(message)
        {
            OALRemoteChannelMgr_SendMessageToStack(message);
        }
        else
        {
            OALChannel channel = OALConnectedProtocolStack_GetChannel(self);
            if (channel)
            {
                OALRemoteChannel_Shutdown(channel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            }
        }
    }
    OALProtocolStack_Release((OALProtocolStack)self);

    return 0;
}

boolean OALProtocolStack_CancelNegotiationTimer(OALProtocolStack self)
{
    boolean status = FALSE;     /* Return TRUE if the timer was cancelled. */

    /* Note: This function should only be called by functions handling protocol stack messages. */
    /*  Remove the timer item.  If the OALTimerItem_Remove returns TRUE release the protocol stack.
        If OALTimerItem_Remove returns FALSE then the timer item callback function will be called. */
    if (self->negotiationTimerItem)
    {
        if (OALTimerItem_Remove(self->negotiationTimerItem))
        {
            /* Release the protocol stack since the timer callback function will not be called.*/
            OALProtocolStack_Release(self);
            status = TRUE;
        }
        OALTimerItem_Destroy(self->negotiationTimerItem, FALSE);
        self->negotiationTimerItem = NULL;
    }
    return status;
}

boolean OALProtocolStack_StartNegotiationTimeout(OALProtocolStack self, uint32 timeout)
{
    OALProtocolStack stk;
    /* Start the ESP negotiation timer. */
    stk = OALProtocolStack_Retain(self);
    if (!stk)
    {
        return FALSE;
    }

    self->negotiationTimerItem = OALTimer_CreateItem(protocolstack_timer, send_negotiation_timeout, 1, (void**)&stk);
    if (!self->negotiationTimerItem ||
        OALTimerItem_Insert( self->negotiationTimerItem, timeout, FALSE, FALSE ) != DOFERROR_OK)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALProtocolStack: handle message, begin negotiation authorization: Failed to insert negotiation timeout timer." );
        if (self->negotiationTimerItem)
        {
            /* The timer item was created, so it must have failed to insert into the timer. */
            OALTimerItem_Destroy(self->negotiationTimerItem, TRUE);
            self->negotiationTimerItem = NULL;
        }
        OALProtocolStack_Release((OALProtocolStack)self);
        return FALSE;
    }
    return TRUE;
}

void OALProtocolStack_Ctor( OALProtocolStack self,
                            OALProtocolStack_Destroy_fn_t        destroy_func,
                            OALProtocolStack_MessageHandler_fn_t msg_handler_func,
                            OALProtocolStack_ENP_EPP_Send_fn_t enp_epp_send_func )
{
    BlackboardData_Init( &(self->blackboardData) );
    self->negotiationTimerItem = NULL;
    self->destroy_func = destroy_func;
    self->msg_handler_func = msg_handler_func;
    self->ENP_EPP_send_func = enp_epp_send_func;
    self->refCount = 1;
    self->keyChangeCallback = NULL;
    self->keyChangeCallbackContext = NULL;
}

void OALProtocolStack_Dtor(OALProtocolStack self)
{
    OALMessage_Destroy(self->shutdownMessage);
    OALProtocolStack_CancelNegotiationTimer(self);
    BlackboardData_Clear( &self->blackboardData );
}

void OALProtocolStack_Shutdown( OALProtocolStack self, DOFException except )
{
    OALRemoteChannelMgr_SendShutdownToStack( self, except );
}

OALProtocolStack OALProtocolStack_Retain( OALProtocolStack self )
{
    OALProtocolStack ret = self;

    if (ret)
    {
        PCRMutex_Lock(ref_count_mutex);
        if (ret->refCount < 0xFFFFFFFF)
        {
            ++ret->refCount;
        }
        else
        {
            /* refCount is already at max so return NULL. */
            ret = NULL;
        }
        PCRMutex_Unlock(ref_count_mutex);
    }

    return ret;
}

void OALProtocolStack_Release( OALProtocolStack self )
{
    if (self)
    {
        uint32 count;

        PCRMutex_Lock(ref_count_mutex);
        PCR_Assert( self->refCount );
        count = --self->refCount;
        PCRMutex_Unlock(ref_count_mutex);

        if (!count && self->destroy_func)
            self->destroy_func( self );
    }
}

void OALProtocolStack_HandleMessage( OALProtocolStack self, OALMessage message )
{
    self->msg_handler_func( self, message );

    OALMessage_Destroy(message);
}

ProtocolLayerIndicator_t OALProtocolStack_DetermineProtocol( DOFOperationType opType )
{
    ProtocolLayerIndicator_t retVal;
    PCR_Assert( opType != DOFOPERATIONTYPE_ROUTEINFO );
    if (opType > DOFOPERATIONTYPE_EPP_NEGONE_LAST)
    {
        retVal = PLI_ESP;
    }
    else if(opType > DOFOPERATIONTYPE_CCMMODE_V1_LAST)
    {
        retVal = PLI_EPP_NEGONE;
    }
    else if (opType > DOFOPERATIONTYPE_SGMP_LAST)
    {
        retVal = PLI_CCMMODE;
    }
    else if(opType > DOFOPERATIONTYPE_TRP_LAST)
    {
        retVal = PLI_SGMP;
    }
    else if(opType > DOFOPERATIONTYPE_TEP_LAST)
    {
        retVal = PLI_TRP;
    }
    else if(opType > DOFOPERATIONTYPE_INVALID)
    {
        retVal = PLI_TEP;
    }
    else
    {
        retVal = PLI_OAP;
    }

    return retVal;
}

void OALProtocolStack_SetKeyChangeCallback( OALProtocolStack self, KeyChangeCallback_fn_t cb, void* context)
{
    self->keyChangeCallback = cb;
    self->keyChangeCallbackContext = context;
}
