/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/address.h>
#include <dof/objectid.h>
#include <dof/connection.h>
#include <dof/providerinfo.h>
#include <internal/channel.h>
#include <internal/providerinfo.h>
#include <internal/connection_internal.h>


extern PCRMutex ref_count_mutex;

/** @addtogroup DOFProviderInfo
@{
**/

/*
**  "Private" Type and Method Declarations
*/

typedef struct DOFProviderInfo_t
{
    uint32             refCount;
    DOFConnectionState connectionState;/* The DOFConnection associated with this DOFProviderInfo */
    DOFObjectID        providerID;     /* The Object ID of the Provider associated with this DOFProviderInfo. */
    DOFAddress         address;        /* The address of the Provider associated with this DOFProviderInfo. */
} DOFProviderInfo_t;


DOFProviderInfo DOFProviderInfo_Create( DOFConnectionState connectionState, DOFObjectID providerID, DOFAddress providerAddress )
{
    if ( connectionState && providerID )
    {
        DOFProviderInfo rval = (DOFProviderInfo) DOFMem_Alloc( sizeof(DOFProviderInfo_t) );
        if ( rval )
        {
            rval->refCount = 1;
            rval->connectionState = DOFConnectionState_Copy( connectionState );
            if ( rval->connectionState )
            {
                rval->providerID = DOFObjectID_Copy( providerID );
                if ( rval->providerID )
                {
                    rval->address = DOFAddress_Copy( providerAddress );
                    if ( rval->address || !providerAddress )
                    {
                        return rval;
                    }
                }
            }
            DOFMem_Free( rval );
        }
    }
    return NULL;
}



/*
**  "Public" Methods
*/


DOFProviderInfo DOFProviderInfo_Copy( const DOFProviderInfo self )
{
    if (self != NULL )
    {
        PCRMutex_Lock(ref_count_mutex);
        if ( self->refCount >= 0xFFFFFFFF )
        {
            PCRMutex_Unlock(ref_count_mutex);
            return NULL;
        }
        self->refCount++;
        PCRMutex_Unlock(ref_count_mutex);
    }
    return self;
}


void DOFProviderInfo_Destroy( DOFProviderInfo self )
{
    if ( self )
    {
        boolean gone = FALSE;
        PCRMutex_Lock(ref_count_mutex);
        if(!--self->refCount) gone = TRUE;
        PCRMutex_Unlock(ref_count_mutex);
        if (gone)
        {
            DOFConnectionState_Destroy( self->connectionState );
            DOFObjectID_Destroy( self->providerID );
            DOFAddress_Destroy( self->address );
            DOFMem_Free( self );
        }
    }
}


DOFObjectID DOFProviderInfo_GetProviderID( const DOFProviderInfo self )
{
    if ( self )
        return self->providerID;
    return NULL;
}


DOFAddress DOFProviderInfo_GetProviderAddress( const DOFProviderInfo self )
{
    if ( self )
        return self->address;
    return NULL;
}


DOFConnectionState DOFProviderInfo_GetConnectionState( const DOFProviderInfo self )
{
    if ( self )
        return self->connectionState;
    return NULL;
}


/**
@}
**/
