/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/query.h>
#include <dof/interfaceid.h>
#include <dof/objectid.h>
#include <internal/filterinternal.h>
#include <internal/interfaceinternal.h>
#include <internal/interfaceidinternal.h>
#include <internal/objectidinternal.h>
#include <internal/router.h>
#include <internal/slist.h>

/** @addtogroup DOFQuery
@{
**/

/*
**  "Private" Type and Method Declarations
*/

/*
**  "Public" Methods
*/

#define QUERYSTYLEMASK ((uint8)(DOFQUERYMATCHSTYLE_AT_LEAST | DOFQUERYMATCHSTYLE_BASE | DOFQUERYMATCHSTYLE_EXACT))

static boolean DOFFilterBlock_IsAllOIDs(interestBlock_t* block);
static boolean DOFFilterBlock_IsAllIIDs(interestBlock_t* block);
static boolean DOFFilterBlock_IsObjInBlock(interestBlock_t* block, DOFObjectID obj, uint8 matchStyle);
static boolean DOFFilterBlock_IsObjAndInterfacesProvided(interestBlock_t* block, DOFObjectID obj, DOFSecurityScope securityScope);

DOFQueryBuilder  DOFQueryBuilder_Create( void )
{
    DOFQueryBuilder  retVal = (DOFQueryBuilder)DOFMem_Alloc(sizeof(DOFQueryBuilder_t));

    if( !retVal )
        return NULL;

    retVal->q.blockCnt = 0;
    retVal->nextBlock = 0;
    retVal->q.pInterestBlocks = NULL;
    retVal->q.flags = DOFQUERYMATCHSTYLE_AT_LEAST; /* default */
    retVal->q.totalBinds = 0;

    retVal->q.restrictionIIDs.count = 0;
    retVal->q.restrictionIIDs.list = NULL;
    return retVal;
}

DOFQueryBuilder DOFQueryBuilder_Create_Query(DOFQuery query)
{
    if (query)
    {
        DOFQueryBuilder retVal = DOFQueryBuilder_Create();
        DOFQuery copied;

        if(!retVal)
            return NULL;

        copied = DOFQuery_DeepCopy(query);

        if(!copied) {
            DOFQueryBuilder_Destroy(retVal);
            return NULL;
        }

        retVal->q = *copied; /* shallow copy */

        DOFMem_Free(copied); /* just free this but don't destroy anything inside */

        return retVal;
    }
    return NULL;
}

static boolean clear_interest_blocks(OALSListNode node, void *arg)
{
    unsigned int i;
    interestBlock_t *block = (interestBlock_t*)node;

    for(i=0; i<block->oidCnt; ++i) DOFObjectID_Destroy(block->pOIDs[i]);
    DOFMem_Free(block->pOIDs);

    for(i=0; i<block->iidCnt; ++i) DOFInterfaceID_Destroy(block->pIIDs[i]);
    DOFMem_Free(block->pIIDs);
    DOFMem_Free(block);

    return FALSE;
}

static void _destroy(void *self, interestBlock_t *blocks, DOFInterfaceIDList restrictions)
{
    if(blocks) {
        OALSList_Traverse((OALSListNode)blocks, clear_interest_blocks, NULL);
    }

    DOFInterfaceIDList_DestroyInterfaceIDs(restrictions);
    DOFMem_Free(self);
}

void DOFQueryBuilder_Destroy(DOFQueryBuilder self)
{
    if(self) _destroy(self, self->q.pInterestBlocks, &self->q.restrictionIIDs);
}

void DOFQuery_Destroy( DOFQuery  self )
{
    if(self) _destroy(self, self->pInterestBlocks, &self->restrictionIIDs);
}

boolean DOFQueryBuilder_AddFilter( DOFQueryBuilder filter, uint32 nObjects, const DOFObjectID objectIDs[], uint32 nInterfaces, const DOFInterfaceID interfaceIDs[] )
{
    interestBlock_t* valAdd;

    if( !filter ||
       !( valAdd  = (interestBlock_t*)DOFMem_Alloc(sizeof(interestBlock_t)) ) )
    {
        return FALSE;
    }
    PCR_Assert( objectIDs || !nObjects );
    PCR_Assert( interfaceIDs || !nInterfaces );

    valAdd->oidCnt = 0;
    valAdd->iidCnt = 0;
    valAdd->pOIDs = NULL;
    valAdd->pIIDs = NULL;

    if ( nObjects )
    {
        valAdd->pOIDs = (DOFObjectID *) DOFMem_Alloc( sizeof(DOFObjectID) * nObjects );
        if ( !valAdd->pOIDs )
            goto fail;
    }
    if ( nInterfaces )
    {
        valAdd->pIIDs = (DOFInterfaceID *) DOFMem_Alloc( sizeof(DOFInterfaceID) * nInterfaces );
        if ( !valAdd->pIIDs )
            goto fail;
    }

    if( nObjects > 0 )
    {
        /* add objectIDs if any */
        for( valAdd->oidCnt = 0; valAdd->oidCnt < nObjects; ++valAdd->oidCnt )
        {
            if ( !objectIDs[valAdd->oidCnt] || !DOFObjectID_IsUnicast( objectIDs[valAdd->oidCnt] ) || ( valAdd->pOIDs[valAdd->oidCnt] = DOFObjectID_Copy(objectIDs[valAdd->oidCnt]) ) == NULL )
                goto fail;
        }
    }

    /* add InterfaceIDs if any*/
    if( nInterfaces > 0 )
    {
        for( valAdd->iidCnt = 0; valAdd->iidCnt < nInterfaces; ++valAdd->iidCnt )
        {
            if ( !interfaceIDs[valAdd->iidCnt] || DOFInterfaceID_IsWildcard( interfaceIDs[valAdd->iidCnt] ) || ( valAdd->pIIDs[valAdd->iidCnt] = DOFInterfaceID_Copy(interfaceIDs[valAdd->iidCnt]) ) == NULL )
                goto fail;
        }
    }

    /* add to start */
    OALSList_Insert((OALSListNode*)&filter->q.pInterestBlocks, (OALSListNode)valAdd);

    /* is catch all case?*/
    if( nObjects == 0 && nInterfaces == 0 )
    {
            filter->q.flags |= FILTER_FLAGS_CATCH_ALL;
            valAdd->flags = FILTER_FLAGS_CATCH_ALL;
            filter->q.totalBinds += 1;
    }
    else if( nObjects == 0 )
    {
            valAdd->flags = FILTER_FLAGS_CATCH_ALL_OBJECTS;
            filter->q.totalBinds += nInterfaces;
    }
    else if( nInterfaces == 0 )
    {
            valAdd->flags = FILTER_FLAGS_CATCH_ALL_INTERFACES;
            filter->q.totalBinds += nObjects;
    }
    else
    {
            valAdd->flags = FILTER_FLAGS_NO_CATCH_ALL;
            filter->q.totalBinds += (nInterfaces*nObjects);
    }

    /* set values */
    ++(filter->q.blockCnt);
    ++(filter->nextBlock);
    valAdd->blockID = filter->nextBlock;

    return TRUE;/*valAdd->blockID;*/

fail:
    while ( valAdd->iidCnt-- > 0 )
        DOFInterfaceID_Destroy( valAdd->pIIDs[valAdd->iidCnt] );
    DOFMem_Free( valAdd->pIIDs );
    while ( valAdd->oidCnt-- > 0 )
        DOFObjectID_Destroy( valAdd->pOIDs[valAdd->oidCnt] );
    DOFMem_Free( valAdd->pOIDs );
    DOFMem_Free( valAdd );
    return FALSE;
}


boolean DOFQueryBuilder_AddRestriction( DOFQueryBuilder filter, uint32 nInterfaces, const DOFInterfaceID interfaceIDs[] )
{
    boolean failure = FALSE;
    uint32 i;
    uint32 cnt;
    DOFInterfaceID* temp;

    if ( !filter )
        return FALSE;

    PCR_Assert( interfaceIDs || !nInterfaces );

    i = 0;
    cnt = filter->q.restrictionIIDs.count;
    filter->q.restrictionIIDs.count += nInterfaces;
    
    if( filter->q.restrictionIIDs.count > nInterfaces )
        temp = (DOFInterfaceID*)DOFMem_Realloc(filter->q.restrictionIIDs.list ,sizeof (DOFInterfaceID) * filter->q.restrictionIIDs.count);
    else
        temp = (DOFInterfaceID*)DOFMem_Alloc(sizeof(DOFInterfaceID) * nInterfaces);

    if(!temp)
    {
        filter->q.restrictionIIDs.count -= nInterfaces;
        DOFLog_Message(  DOFLOGLEVEL_ERROR, "DOFQueryBuilder: Failed to alloc space for new restriction iids." );
        return FALSE;
    }

    filter->q.restrictionIIDs.list = temp;

    for(; cnt < filter->q.restrictionIIDs.count; ++cnt )
    {
        filter->q.restrictionIIDs.list[cnt] = DOFInterfaceID_Copy(interfaceIDs[i++]);
        if( !filter->q.restrictionIIDs.list[cnt] )
        {
            failure = TRUE;
            break;
        }
    }

    if( failure )
    {
        /* roll back */
        DOFLog_Message(  DOFLOGLEVEL_ERROR, "DOFQueryBuilder: Failed to copy restriction iids." );

        filter->q.restrictionIIDs.count -= nInterfaces;
        cnt = filter->q.restrictionIIDs.count;

        while( filter->q.restrictionIIDs.list[cnt] && cnt < ( filter->q.restrictionIIDs.count + nInterfaces ) )
            DOFInterfaceID_Destroy(filter->q.restrictionIIDs.list[cnt++]);

        if(filter->q.restrictionIIDs.count == 0)
        {
            DOFMem_Free(filter->q.restrictionIIDs.list);
            filter->q.restrictionIIDs.list = NULL;
        }
        return FALSE;
    }
    return TRUE;
}

/** @internal
 * gets the number of filter blocks
 *
 * @param filter the filter to get the block count from
 *
 * @return uint32 block count
 */
uint32 DOFQuery_GetBlockCount(DOFQuery  filter)
{
    if( !filter )
        return 0;

    return (filter->blockCnt-1);
}

/** @internal
 * does this filter contain the catch all case
 *
 * @param filter the filter to check
 *
 * @return boolean TRUE if it is catch all
 */
boolean DOFQuery_IsCatchAll(DOFQuery  filter)
{
    if( filter && ( filter->pInterestBlocks == NULL || (filter->flags & FILTER_FLAGS_CATCH_ALL) == FILTER_FLAGS_CATCH_ALL ) )
        return TRUE;
    else
        return FALSE;
}

/** @internal
 * returns the restricted iid for this filter
 *
 * @param filter the filter to get restricted iids from
 *
 * @return DOFInterfaceIDList list of iids
 */
DOFInterfaceIDList DOFQuery_GetRestrictIIDs(DOFQuery  filter)
{
    if( filter )
    {
        return &filter->restrictionIIDs;
    }

    return NULL;
}


/** @internal
 * return if an objectid and its provided iids are of interest
 *
 * @param filter filter to check
 * @param obj object to look for
 * @param partition the partition
 *
 * @return boolean TRUE if it is of interest and provides the iids
 *                 FALSE if it is not of interest
 */
boolean DOFQuery_QueryMatch(DOFQuery filter, DOFObjectID objID, DOFSecurityScope securityScope)
{
    interestBlock_t* curr = NULL;

    if( !filter )
        return TRUE;

    /* is catch all */
    if( DOFQuery_IsCatchAll(filter) )
        return TRUE;

    /* get head of list */
    curr = filter->pInterestBlocks;

    while( curr )
    {
        /* Is curr interest block set for all OIDs, else check if objID is in curr interest block. */
        if( DOFFilterBlock_IsObjInBlock(curr, objID, (filter->flags & QUERYSTYLEMASK)) )
        {
            /* Is curr interest block set for all IIDs, else check if we are providing on any of curr's IIDs. */
            if( DOFFilterBlock_IsObjAndInterfacesProvided(curr, objID, securityScope) )
            {
                return TRUE;
            }
        }
        curr = curr->next;
    }
    return FALSE;
}

/** @internal
 * checks if obj is in this filter block
 *
 * @param block block to check
 * @param obj objectid to check for
 *
 * @return boolean TRUE if obj is covered in this interestBlock
 *                 FALSE otherwise
 */
static boolean DOFFilterBlock_IsObjInBlock(interestBlock_t* block, DOFObjectID obj, uint8 matchStyle)
{
    /* DOFQueryMatchStyle_t defines order of the three functions used here. */
    static boolean (* const match_fns[])(const DOFObjectID oid1, const DOFObjectID oid2) =
                                               { DOFObjectID_MatchesAtLeast,          /* DOFQUERYMATCHSTYLE_AT_LEAST. */
                                                 DOFObjectID_IsEqualIgnoreAttributes, /* DOFQUERYMATCHSTYLE_BASE. */
                                                 DOFObjectID_IsEqual };               /* DOFQUERYMATCHSTYLE_EXACT. */
    uint32 cnt;

    /* if catch all obj */
    if( DOFFilterBlock_IsAllOIDs(block) )
        return TRUE;

    PCR_Assert( matchStyle < QUERYSTYLEMASK);
    /* loop through array */
    for(cnt=0; cnt < block->oidCnt; ++cnt )
    {
        if (match_fns[matchStyle]( obj, block->pOIDs[cnt] ))
        {
            return TRUE;
        }
    }

   return FALSE;
}

/** @internal
 * get if an objectid provides the interfaceids to fit in to the filter block
 *
 * @param block block to check against
 * @param obj object to check
 * @param partition partition to check against
 *
 * @return boolean TRUE if it objectID Provides InterfaceIDs
 *                 FALSE otherwise
 */
static boolean DOFFilterBlock_IsObjAndInterfacesProvided(interestBlock_t* block, DOFObjectID obj, DOFSecurityScope securityScope)
{

    if( DOFFilterBlock_IsAllIIDs(block) )
        return TRUE;
    else
    {
        if( DOFRouter_CheckInterfaces( obj, securityScope, block->pIIDs, block->iidCnt ) )
            return TRUE;
    }

    return FALSE;
}

/** @internal
 * gets if this block is a all objectid block
 *
 * @param block the block to check
 *
 * @return boolean TRUE if it is for all ObjectIDs
 */
static boolean DOFFilterBlock_IsAllOIDs(interestBlock_t* block)
{
    if( block->flags & FILTER_FLAGS_CATCH_ALL_OBJECTS )
        return TRUE;
    else
        return FALSE;
}

/** @internal
 * gets if this block is a all interfaceid block
 *
 * @param block the block to check
 *
 * @return boolean TRUE if it is for all InterfaceIDs
 */
static boolean DOFFilterBlock_IsAllIIDs(interestBlock_t* block)
{
    if( block->flags & FILTER_FLAGS_CATCH_ALL_INTERFACES )
        return TRUE;
    else
        return FALSE;
}

/** @internal
 * get all the bindings possible from a filter
 *
 * @param filter the filter to get the bindings of
 *
 * @return bindingList* ptr to the bindingList created Must be Destroyed
 */
bindingList* DOFQuery_GetBindings(DOFQuery  filter)
{
    interestBlock_t* block;
    bindingList*     retVal = NULL;

    if( !filter )
        return NULL;

    retVal = DOFMem_Alloc(sizeof(bindingList));

    if( !retVal )
    { /* check if alloc worked */
        return NULL;
    }

    if( DOFQuery_IsCatchAll(filter) )
    {
        /* declare catch all binding list and return */
        retVal->nBindings = 1;
        retVal->pBindings = DOFMem_Alloc(sizeof(binding_t));
        if( !retVal->pBindings )
        {
            DOFMem_Free(retVal);
            return NULL;
        }

        retVal->pBindings[0].oid = DOFOBJECTID_BROADCAST;
        retVal->pBindings[0].iid = DOFINTERFACEID_WILDCARD;
        return retVal;
    }

    /* declare size of retVal */
    retVal->nBindings = 0;
    retVal->pBindings = DOFMem_Alloc(sizeof(binding_t) * filter->totalBinds);

    if (!retVal->pBindings)
    {
        DOFMem_Free(retVal);
        return NULL;
    }

    block = filter->pInterestBlocks;

    while( block )
    {
        uint32 cnt = 0;

        if( DOFFilterBlock_IsAllOIDs(block) )
        {
            /* create bindings for all interfaces in block with wildcard oid */
            for(; cnt < block->iidCnt; ++cnt )
            {
                retVal->pBindings[retVal->nBindings].oid = DOFOBJECTID_BROADCAST;
                retVal->pBindings[retVal->nBindings].iid = block->pIIDs[cnt];
                ++(retVal->nBindings);
            }
        }
        else if( DOFFilterBlock_IsAllIIDs(block) )
        {
            /* create bindings for all objects in block with wildcard iid */
            for(; cnt < block->oidCnt; ++cnt )
            {
                retVal->pBindings[retVal->nBindings].oid = block->pOIDs[cnt];
                retVal->pBindings[retVal->nBindings].iid = DOFINTERFACEID_WILDCARD;
                ++(retVal->nBindings);
            }
        }
        else
        {
            /* create bindings for each oid with all iids from block */
            for(; cnt < block->oidCnt; ++cnt )
            {
                uint32 cnt2 = 0;
                for(; cnt2 < block->iidCnt; ++cnt2)
                {
                    retVal->pBindings[retVal->nBindings].oid = block->pOIDs[cnt];
                    retVal->pBindings[retVal->nBindings].iid = block->pIIDs[cnt2];
                    ++(retVal->nBindings);
                }
            }
        }
        block = block->next;
    }

    return retVal;
}

/** @internal
 * destroy a bindinglist
 *
 * @param bindings binding list to destroy
 */
void bindingList_Destroy(bindingList* bindings)
{
    if( !bindings )
        return;

    DOFMem_Free( bindings->pBindings );
    DOFMem_Free( bindings );
}

struct params {
    DOFQuery into;
    boolean success;
};

static boolean deep_copy_add_filter(OALSListNode node, void* arg)
{
    interestBlock_t* block = (interestBlock_t*)node;
    struct params* p = (struct params*)arg;
    struct DOFQueryBuilder_t builder;

    PCRMem_Set(&builder, 0, sizeof(builder));

    if(!DOFQueryBuilder_AddFilter(&builder, block->oidCnt, block->pOIDs, block->iidCnt, block->pIIDs))
    {
        p->success = FALSE;
        return TRUE;
    }

    OALSList_Insert((OALSListNode*)&p->into->pInterestBlocks, (OALSListNode)builder.q.pInterestBlocks);

    return FALSE;
}

/** @internal
 * deep copies the filter passed in
 *
 * @param filter The filter to copy
 *
 * @return A copy of the filter passed in
 */
DOFQuery  DOFQuery_DeepCopy(DOFQuery  filter)
{
    DOFQuery  retVal;
    uint32 cnt = 0;
    struct params p;

    if( !filter )
        return NULL;

    retVal = (DOFQuery)DOFMem_Alloc(sizeof(*retVal));
    if(!retVal)
    {
        DOFLog_Message(  DOFLOGLEVEL_ERROR, "DOFQuery: Failed to create. " );
        return NULL;
    }

    PCRMem_Set(retVal, 0, sizeof(*retVal));
    retVal->blockCnt = filter->blockCnt;

    /* copy interest blocks */
    p.into = retVal;
    p.success = TRUE;

    OALSList_Traverse((OALSListNode)filter->pInterestBlocks, deep_copy_add_filter, &p);
    if(!p.success) 
    {
        DOFLog_Message(  DOFLOGLEVEL_ERROR, "DOFQuery: Failed to add filters. " );
        DOFQuery_Destroy(retVal);
        return NULL;
    }

    /* copy restriction */
    retVal->restrictionIIDs.count = filter->restrictionIIDs.count;
    retVal->restrictionIIDs.list = DOFMem_Alloc(sizeof(DOFInterfaceID) * retVal->restrictionIIDs.count);

    if( !retVal->restrictionIIDs.list )
    {
        DOFLog_Message(  DOFLOGLEVEL_ERROR, "DOFQuery: Failed to create restriction iids." );
        DOFQuery_Destroy(retVal);
        return NULL;
    }

    PCRMem_Set(retVal->restrictionIIDs.list, 0, sizeof(DOFInterfaceID) * retVal->restrictionIIDs.count);

    for(cnt = 0; cnt < filter->restrictionIIDs.count; ++cnt )
    {
        retVal->restrictionIIDs.list[cnt] = DOFInterfaceID_Copy(filter->restrictionIIDs.list[cnt]);
        if(!retVal->restrictionIIDs.list[cnt])
        {
            DOFLog_Message(  DOFLOGLEVEL_ERROR, "DOFQuery: Failed to copy restriction iids." );
            DOFQuery_Destroy(retVal);
            return NULL;
        }
    }

    retVal->totalBinds = filter->totalBinds;
    retVal->flags = filter->flags;

    return retVal;
}

/** @internal
 * Checks if InterfaceID is in the restricted list or the list is empty or the filter is null
 *
 * @param filter  the filter that has the restricted list to check
 * @param iid  the iid to check
 *
 * @return boolean TRUE if it is in list or list is for every one or filter is null
 *                 FALSE if there is a list and this iid is not in it
 */
boolean DOFQuery_IsRestricted(DOFQuery  filter, DOFInterfaceID iid)
{
    if( !filter )
        return TRUE;

    if( filter->restrictionIIDs.count == 0 )
        return TRUE;
    else
    {
        uint32 cnt = 0;

        for( ; cnt < filter->restrictionIIDs.count; ++cnt )
        {
            if( DOFInterfaceID_IsEqual(filter->restrictionIIDs.list[cnt], iid) )
                return TRUE;
        }
    }
    return FALSE;
}

boolean DOFQueryBuilder_SetMatchStyle(DOFQueryBuilder self, DOFQueryMatchStyle_t matchStyle)
{
    if (self && matchStyle < QUERYSTYLEMASK)
    {
        self->q.flags &= ~QUERYSTYLEMASK;
        self->q.flags |= matchStyle;
        return TRUE;
    }
    return FALSE;
}

DOFQuery DOFQueryBuilder_Build(DOFQueryBuilder builder)
{
    return DOFQuery_DeepCopy(&builder->q);
}

DOFQuery DOFQueryBuilder_BuildAndDestroy(DOFQueryBuilder builder)
{
    DOFQuery ret = DOFQueryBuilder_Build(builder);

    if(ret)
        DOFQueryBuilder_Destroy(builder);
    else
        DOFLog_Message(  DOFLOGLEVEL_ERROR, "DOFQueryBuilder: Failed to create query." );
    return ret;
}

/**
@}
**/
