/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/connection.h>
#include <dof/error.h>
#include <dof/exception.h>
#include <dof/object.h>
#include <dof/request.h>
#include <dof/transport.h>
#include <dof/value.h>
#include <internal/channel.h>
#include <internal/connection_internal.h>
#include <internal/exception.h>
#include <internal/iteminternal.h>
#include <internal/mutex_pool.h>
#include <internal/objectinternal.h>
#include <internal/objectidinternal.h>
#include <internal/operationinternal.h>
#include <internal/operationconfiginternal.h>
#include <internal/remotechanneldata.h>
#include <internal/requestinternal.h>
#include <internal/securityscope.h>
#include <internal/systeminternal.h>
#include <internal/transportinternal.h>
#include <internal/value.h>


/** @addtogroup DOFRequest
@{
**/

/*
**  "Private" Type and Method Declarations
*/


/* DOFRequest code assumes these values can be expressed in one byte. Otherwise, more DOFRequest mutex locking would be
   needed to protect very small platforms. */
typedef enum RequestPendState_t
{
    APP_CANNOT_ACCEPT_OR_THROW = 0,
    NORMAL_PEND,
    SUBSCRIBE_CALLBACK_OK_NOW_NEED_GET,
    APP_CAN_THROW_ONLY
} RequestPendState_t;


typedef struct DOFRequest_t
{
    DOFSystem           system;          /* The DOFSystem associated with this request. */
    DOFConnectionState  connectionState; /* The DOFConnectionState associated with this request */
    DOFAddress          addr;            /* The requestor's address. */
    DOFObjectIDSource   requestorID;     /* The requestor objectID for this request. */
    DOFRequestType      requestType;     /* The directed type for this req -- one of the @ref DOFRequestType constants*/
    DOFOperationType    opType;          /* The operation type of the request that this is. */
    DOFException        exception;       /* Exception storage used by the set/get exception calls. */
    DOFSecurityScope    scope;
    uint32              refCount;
    /* These next three are used only for pend. */
    DOFOperation        provOp;
    DOFOperation        reqOp;
    RequestPendState_t  pendState;
} DOFRequest_t;


/** @internal
    Create a DOFRequest. This fills a request with constants meaning a locally generated request.
    It is used by DOFRouter to provide a request for a local operation.

    @param system    The DOFSystem associated with this request
    @param op        The operation representing this request.
    @param connState The connection state associated with this request.

    @return The created request.
**/
DOFRequest DOFRequest_Create(DOFSystem system, DOFOperation op, DOFConnectionState connState)
{
    boolean    success    = FALSE;
    DOFRequest newRequest = (DOFRequest)DOFMem_Alloc( sizeof( DOFRequest_t ) );

    if (newRequest)
    {
        PCRMem_Set( newRequest, 0, sizeof( DOFRequest_t ) );
        newRequest->refCount = 1;
        newRequest->pendState = APP_CANNOT_ACCEPT_OR_THROW;
        newRequest->system = system;
        newRequest->requestType = DOFREQUESTTYPE_UNICAST;
        success = TRUE;
        if (op)
        {
            /* Locking op acknowledges that op->sources (the head of an OALSList) is intended to change as additional
               sources are added to an operation, though understood that versin 5.0 use does not insert or remove
               from this op->sources slist. Even so, locking op here acknowledges that the address of an operation is
               understood able to change on the fly. */
            DOFOperation_Lock( op );
            PCR_Assert( op->sources );
            if (op->sources->address)
            {
                if ((newRequest->addr = DOFAddress_Copy( op->sources->address )) == NULL)
                    success = FALSE;
            }
            if ( op->securityScope )
            {
                if ( (newRequest->scope = DOFSecurityScope_Copy( op->securityScope )) == NULL )
                    success = FALSE;
            }
            DOFOperation_Unlock( op );
            if (success)
            {
                if (connState)
                    newRequest->connectionState = DOFConnectionState_Copy( connState );
                else
                    newRequest->connectionState = OALChannel_CreateConnectionState( op->sources->channel,
                                                                                    op->sources->address );
                if (newRequest->connectionState)
                {
                    if (op->opID)
                    {
                        newRequest->requestorID = DOFObjectIDSource_Create( DOFOperationID_GetObjectID( op->opID ) );
                        if (!newRequest->requestorID)
                        {
                            success = FALSE;
                        }
                    }
                }
                else
                    success = FALSE;
            }
            if (success)
            {
                newRequest->opType = DOFOperation_GetType( op );
                if (op->object_id.object_id)
                {
                    if (DOFObjectID_IsMulticast( op->object_id.object_id ))
                        newRequest->requestType = DOFREQUESTTYPE_MULTICAST;
                    else if (DOFObjectID_IsBroadcast( op->object_id.object_id ))
                        newRequest->requestType = DOFREQUESTTYPE_BROADCAST;
                }
            }
            else
            {
                DOFRequest_Release( newRequest );
                newRequest = NULL;
            }
        }
        else
            newRequest->opType = DOFOPERATIONTYPE_INVALID;
    }

    return newRequest;
}


void DOFRequest_Retain( DOFRequest self )
{
    if (self)
    {
        /* This retain is the kind that is deemed not possible of failure. It asserts if the refCount rolls over to 0.*/
        PCRMutex_Lock( ref_count_mutex );
        ++(self->refCount);
        PCR_Assert( self->refCount );
        PCRMutex_Unlock( ref_count_mutex );
    }
}


/** @internal
    Release a DOFRequest.

    @note For NULL request, this returns harmlessly, having done nothing.

    @param  request The DOFRequest to destroy

    @return void
**/
void DOFRequest_Release( DOFRequest request )
{
    if (request)
    {
        boolean freeOK = FALSE;

        PCRMutex_Lock( ref_count_mutex );
        if (!--(request->refCount))
            freeOK = TRUE;
        PCRMutex_Unlock( ref_count_mutex );
        if (freeOK)
        {
            DOFObjectIDSource_Destroy( request->requestorID );
            DOFException_Destroy( request->exception );
            DOFConnectionState_Destroy( request->connectionState );
            DOFAddress_Destroy( request->addr );
            DOFSecurityScope_Destroy( request->scope );
            /* These next two should already be cleaned up and set to NULL, but just in case we attempt to destroy a
               pending request, these will free the associated operation. */
            DOFOperation_Internal_Destroy( request->reqOp );
            DOFOperation_Internal_Destroy( request->provOp );
            /* request->system not released because it was purposely not retained at request create. Interactions
               between a DOFSystem and operations are carefully controlled. Our request cleans up when it is unpended,
               which happens in the ordinary interaction of a DOFSystem and operations.*/
            OALMutexPool_Finalize( request );
            DOFMem_Free( request );
        }
    }
}


/* @internal */
DOFException DOFRequest_GetException( const DOFRequest request )
{
    PCR_Assert( request );

    return request->exception;
}


/* @internal */
DOFOperation DOFRequest_GetProviderOp( const DOFRequest request )
{
    DOFOperation op;

    PCR_Assert( request );
    /* Locking request mutex guards data slicing, where unpend() can come along anytime and change ->provOp on the fly*/
    OALMutexPool_Lock( request );
    op = request->provOp;
    OALMutexPool_Unlock( request );

    return op;
}


/* @internal */
DOFObjectID DOFRequest_GetResultOID( const DOFRequest request )
{
    return DOFObject_GetObjectID( DOFOperation_GetObject( DOFRequest_GetProviderOp( request ) ) );
}


static void unpendAndUnlock( DOFRequest request )
{
    /* Called with the request lock held. */
    DOFOperation reqOp;
    DOFOperation provOp;

    PCR_Assert( request );
    request->pendState = APP_CANNOT_ACCEPT_OR_THROW;
    reqOp = request->reqOp;
    request->reqOp = NULL;
    provOp = request->provOp;
    request->provOp = NULL;
    OALMutexPool_Unlock( request );

    DOFOperation_Internal_Destroy( reqOp );
    DOFOperation_Internal_Destroy( provOp );
}


/* internal. */
void DOFRequest_Pend( DOFRequest request, DOFOperation reqOp, DOFOperation provOp )
{
    PCR_Assert( request );
    OALMutexPool_Lock( request );
    request->reqOp = reqOp;
    request->provOp = provOp;
    request->pendState = NORMAL_PEND;
    OALMutexPool_Unlock( request );
    DOFOperation_Retain( reqOp );
    /* reqOp owns its DOFRequest (us), so no need to retain reqOp while we pend. */
    DOFOperation_Retain( provOp );
}


void DOFRequest_Unpend( DOFRequest request )
{
    PCR_Assert( request );
    OALMutexPool_Lock( request );
    unpendAndUnlock( request );
}


/* internal. */
DOFOperation DOFRequest_GetReqOp( const DOFRequest request )
{
    DOFOperation ret;

    PCR_Assert( request );
    /* Locking request mutex guards data slicing, where unpend() can come along anytime and change ->reqOp on the fly.*/
    OALMutexPool_Lock( request );
    ret = request->reqOp;
    OALMutexPool_Unlock( request );

    return ret;
}


/*
**  "Public" Type and Method Declarations
*/


/*
**  "Public" Methods
*/


DOFObjectIDSource DOFRequest_GetRequestorID( const DOFRequest request )
{
    /* This call does not allocate memory for the requestor OID. Within parameter request's structure exists a
       DOFObjectID that we return. It remains valid until the DOFRequest is destroyed. */
    PCR_Assert( request );

    return request->requestorID;
}


DOFAddress DOFRequest_GetRequestorAddress( const DOFRequest request )
{
    /* This call does not allocate memory for the requestor address. Within parameter request's structure exists the
       DOFOperation, containing a DOFAddress that we return. It remains valid until the DOFRequest is destroyed. */
    PCR_Assert( request );

    return request->addr;
}



DOFRequestType DOFRequest_GetRequestType( const DOFRequest request )
{
    PCR_Assert( request );

    return request->requestType;
}


DOFSystem DOFRequest_GetSystem( const DOFRequest request )
{
    PCR_Assert( request );

    return request->system;
}


DOFConnectionState DOFRequest_GetConnectionState( const DOFRequest request )
{
    PCR_Assert( request );

    return request->connectionState;
}

DOFSecurityScope DOFRequest_GetSecurityScope( const DOFRequest request )
{
    PCR_Assert( request );
    return request->scope;
}


/*********************** functions to deal with pending and subsequently completing requests *************************/


static void requestThrow( DOFRequest request, DOFException except )
{
    boolean cleanupException = FALSE;

    /*Called with request mutex locked- temporarily unlocks it for calling the proper DOFSystem_CompleteXXX() function*/
    PCR_Assert( request );
    PCR_Assert( except );
    PCR_Assert( !request->exception );
    request->exception = DOFException_Copy( except ); /* We have no way to notify anything in COAL if this copy fails.*/
    OALMutexPool_Unlock( request );
    switch (request->opType)
    {
        case DOFOPERATIONTYPE_GET:
        case DOFOPERATIONTYPE_SUBSCRIBE:
        case DOFOPERATIONTYPE_CHANGED:
            DOFSystem_CompleteGetRequest( request->system, request, NULL );
        break;
        case DOFOPERATIONTYPE_INVOKE:
            DOFSystem_CompleteInvokeRequest( request->system, request, 0, NULL );
        break;
        case DOFOPERATIONTYPE_SET:
            DOFSystem_CompleteSetRequest( request->system, request );
        break;
        case DOFOPERATIONTYPE_SESSION:
            DOFSystem_CompleteSessionRequest( request->system, request );
        break;
        case DOFOPERATIONTYPE_REGISTER:
            DOFSystem_CompleteRegisterRequest( request->system, request );
        break;
        default:
            /* Here only because app did something wrong, not even sure how. Whatever the reason, we have no call to
               make and have nothing else we can do for the situation. We do need to cleanup the exception copy. */
            cleanupException = TRUE;
        break;
    }
    OALMutexPool_Lock( request );
    if (cleanupException)
    {
        DOFException_Destroy( request->exception );
        request->exception = NULL;
    }
}


/* Intent of the docs:
        Throw means no follow-on throw or accept.
        For _Get, _Changed, _Set, _Invoke: no throw to follow accept.
   Implementation:
        Follow-on throw: Detect and do nothing.
        _Get, _Changed, _Set, _Invoke, throw after accept: detect and do nothing, but request is free'd so crash likely.
        Throw after XXXComplete callback: detect and do nothing, but request is free'd so crash is likely. */
void DOFRequest_Throw( DOFRequest request, DOFException except )                /* Called only by app, never by COAL. */
{
    PCR_Assert( request );
    OALMutexPool_Lock( request );
    if (request->pendState != APP_CANNOT_ACCEPT_OR_THROW)
    {
        boolean requestDone;

        request->pendState = APP_CANNOT_ACCEPT_OR_THROW;
        if (!except)
            except = DOFEXCEPTION_FAILED;
        else if (DOFException_GetType( except ) == DOFEXCEPTIONTYPE_PROVIDER &&
                 !DOFInterfaceID_IsEqual(
                           DOFInterfaceException_GetInterfaceID( DOFProviderException_GetInterfaceException( except ) ),
                                          request->reqOp->interface_id ))
            except = DOFEXCEPTION_APPLICATION_ERROR;
        requestThrow( request, except );
        /* Check ->reqOp before unpend() allows ->reqOp to (probably) be deallocated. */
        requestDone = (request->reqOp->operation_type != DOFOPERATIONTYPE_SUBSCRIBE &&
                               request->reqOp->operation_type != DOFOPERATIONTYPE_REGISTER &&
                               request->reqOp->operation_type != DOFOPERATIONTYPE_SESSION) ? TRUE : FALSE;

        unpendAndUnlock( request );
        if (requestDone)
        {
            DOFRequest_Release( request );
        }
        else
        {
            /*request is placed in request->provOp's requestTree; clean up only when request is removed from the tree*/
        }
    }
    else
        OALMutexPool_Unlock( request );
}


boolean DOFRequest_IsPending( const DOFRequest request )
{
    PCR_Assert( request );
    /* We rely here on request->pendState always being expressible in one byte. Otherwise we would want to lock the
       request mutex to guard data slicing where unpend() can come along anytime and change it. */

    return (request->pendState != APP_CANNOT_ACCEPT_OR_THROW);
}


void DOFRequest_DisablePend( DOFRequest self )
{
    PCR_Assert( self );
    OALMutexPool_Lock( self );
    self->pendState = APP_CANNOT_ACCEPT_OR_THROW;
    OALMutexPool_Unlock( self );
}


/* Called in DOFSystem's CleanupAndCallProvideCompleteCallback() for one DOFRequest. Also cCalled in
   DOF_Operation_Internal_Destroy() AND _DOFOperation_Cancel() by OALAVLTree_Destroy() for as many RequestTreeContext_t
   items as are held in op->requestTree. At each call, pCtx is filled with the next item in requestTree. Our function
   proto matches typedef boolean (*OALAVLTree_RemoveAll_fn_t)(OALAVLTreeItem_t item, void arg); */
void DOFRequest_CleanupRequestTree( DOFRequest request, void * arg )
{
    PCR_Assert( request );
    DOFRequest_Unpend( request );
    DOFRequest_Release( request );
    /* Note: It happens that pend and unpend manipulate provOp's refCount, and pend/unpend always happen when
       inserting/removing requests from the tree. That could mean that separately manipulationg provOp's refCount is
       not needed. But we keep that in place (harmless), to allow pend/unpend to discontinue its manipulation of
       provOp's refCount, IF THAT WERE EVER THOUGHT TO BE BENEFICIAL. */
}

/* Intent of the docs for providing on _SUBSCRIBE, _REGISTER, and _SESSION requests:
        _SUBSCRIBE/_REGISTER happen with NULL callback, _SESSION does not. With NULL callback, app cannot accept/throw.
        For non-NULL callback:
            Behavior won't happen if supplied request is not accepted.
            For XXXComplete callback NULL, original callback must accept/throw, and throw cannot follow.
            For XXXComplete callback, accepted behavior can be rejected with throw, until XXXComplete callback called.
    Implementation:
        Presence or absence of accept/throw has nothing to do with cleanup.
        Multiple accept/throw: Detect extras, do nothing.
        Accept/Throw after XXXComplete callback: detect and do nothing, but request is free'd so crash is likely. */
static void DOFRequest_Return( DOFRequest request, DOFOperationType expected )
{
    DOFOperationType opType;

    PCR_Assert( request );
    PCR_Assert( request->reqOp->operation_type == DOFOPERATIONTYPE_SUBSCRIBE ||
                request->reqOp->operation_type == DOFOPERATIONTYPE_REGISTER ||
                request->reqOp->operation_type == DOFOPERATIONTYPE_SESSION );
    OALMutexPool_Lock( request );
    opType = request->opType;
    if (opType == expected && request->pendState == NORMAL_PEND)
    {
        request->pendState = (opType == DOFOPERATIONTYPE_SUBSCRIBE) ? SUBSCRIBE_CALLBACK_OK_NOW_NEED_GET :
                                                                      APP_CAN_THROW_ONLY;
        OALMutexPool_Unlock( request );
        switch (opType)
        {
            case DOFOPERATIONTYPE_SUBSCRIBE:
                DOFSystem_CompleteSubscribeRequest( request->system, request );
                /* We get here when the app has a Subscribe provider callback and during that call or after
                   calls DOFRequestSubscribe_Return(), WHICH MEANS "enable the subscribe." App can later call
                   DOFRequest_Throw() to terminate the subscribe, up until the subscribe op is cancelled. */
            break;
            case DOFOPERATIONTYPE_REGISTER:
                DOFSystem_CompleteRegisterRequest( request->system, request );
                /* We get here when the app has a Register provider callback and during that call or after calls
                   DOFRequestRegister_Return(), WHICH MEANS "enable the registration." App can later call
                   DOFRequest_Throw() to terminate the registration, up until the register op is cancelled. */
            break;
            case DOFOPERATIONTYPE_SESSION:
                DOFSystem_CompleteSessionRequest( request->system, request );
                /* We get here when the app has a Session provider callback and during that call or after calls
                   DOFRequestSession_Return(), WHICH MEANS "enable the session." App can later call
                   DOFRequest_Throw() to terminate the session, up until the session op is cancelled. */
            break;
            default:
                /* This can never occur with the assertion of operation type at the top. */
            break;
        }
        return;
    }
    /*App tried to return using wrong API call. We have no choice but to unpend, but we throw an exception if possible*/
    OALMutexPool_Unlock( request );
    DOFRequest_Throw( request, DOFEXCEPTION_APPLICATION_ERROR );
}


void DOFRequestSubscribe_Return( DOFRequest request )                           /* Called only by app, never by COAL. */
{
    DOFRequest_Return( request, DOFOPERATIONTYPE_SUBSCRIBE );
}


void DOFRequestRegister_Return( DOFRequest request )                            /* Called only by app, never by COAL. */
{
    DOFRequest_Return( request, DOFOPERATIONTYPE_REGISTER );
}


void DOFRequestSession_Return( DOFRequest request )                             /* Called only by app, never by COAL. */
{
    DOFRequest_Return( request, DOFOPERATIONTYPE_SESSION );
}

boolean DOFRequestSession_SetDataTransform(DOFRequest                       request,
                                           DOFOperationSessionDataTransform transform)
{
    if(request != NULL)
    {
        return DOFOperationSession_SetDataTransformProvider(request->reqOp, transform);
    }
    return FALSE;
}

void DOFRequestSubscribe_Accept( DOFRequest request )    /* Internal API call, so app never calls us, only DOFSystem. */
{
    PCR_Assert( request );
    PCR_Assert( request->opType == DOFOPERATIONTYPE_SUBSCRIBE );
    PCR_Assert( request->reqOp->operation_type == DOFOPERATIONTYPE_SUBSCRIBE );
    PCR_Assert( request->pendState == NORMAL_PEND );
    OALMutexPool_Lock( request );
    request->pendState = SUBSCRIBE_CALLBACK_OK_NOW_NEED_GET;
    OALMutexPool_Unlock( request );
}


/* Intent of the docs for providing on _GET, _SET, and _INVOKE requests:
        It doesn't happen with NULL callback. When callback is called, its request must be accepted or thrown, and not
        more than once between the two choices.
   Implementation:
        If said request never accepted/thrown, original request op and the provide op do not clean up.
        Multiple accept/throw: Detect extras, do nothing. Request is free'd at 1st instance, so crash is likely. */
static void DOFRequest_Return_NoCompleteCallback( DOFRequest request, DOFOperationType expected, DOFValue value,
                                                  uint16 valuesCount, const DOFValue values[], boolean existingLock )
{
    boolean requestIsComplete = FALSE;

    PCR_Assert( request );
    PCR_Assert( expected == DOFOPERATIONTYPE_GET || expected == DOFOPERATIONTYPE_SET ||
                expected == DOFOPERATIONTYPE_INVOKE );
    PCR_Assert( request->reqOp->operation_type == DOFOPERATIONTYPE_GET ||
                request->reqOp->operation_type == DOFOPERATIONTYPE_CHANGED ||
                request->reqOp->operation_type == DOFOPERATIONTYPE_SET ||
                request->reqOp->operation_type == DOFOPERATIONTYPE_INVOKE );
    if (!existingLock)
        OALMutexPool_Lock( request );
    if (request->pendState != APP_CANNOT_ACCEPT_OR_THROW && request->pendState != APP_CAN_THROW_ONLY)
    {
        request->pendState = APP_CANNOT_ACCEPT_OR_THROW;
        if (expected == request->opType ||
            (expected == DOFOPERATIONTYPE_GET && request->opType == DOFOPERATIONTYPE_CHANGED))
        {
            OALMutexPool_Unlock( request );
            switch (request->opType)
            {
                /* These 3 calls use request only to obtain operation, exception, and/or OID. We can keep request lock*/
                case DOFOPERATIONTYPE_GET:
                case DOFOPERATIONTYPE_CHANGED:
                    DOFSystem_CompleteGetRequest( request->system, request, value );
                    break;
                case DOFOPERATIONTYPE_SET:
                    DOFSystem_CompleteSetRequest( request->system, request );
                    break;
                case DOFOPERATIONTYPE_INVOKE:
                    DOFSystem_CompleteInvokeRequest( request->system, request, valuesCount, values );
                    break;
                default:
                    {
                        PCR_Assert( FALSE );
                    }
                    break;
            }
            OALMutexPool_Lock( request );
        }
        else
        {
            requestThrow( request, DOFEXCEPTION_APPLICATION_ERROR );
            /* Though not sure what app wants, assume we can cleanup by falling through to unpend and destroy request.*/
        }
        unpendAndUnlock( request );
        if (existingLock)
            OALMutexPool_Lock( request );
        requestIsComplete = TRUE;
    }
    else
    { /* App wrongly called DOFRequestXXX_Return() more than once. To mimimize damage do NOTHING. */
        if (!existingLock)
            OALMutexPool_Unlock( request );
    }
    if (requestIsComplete)
        DOFRequest_Release( request ); /* GetComplete, SetComplete, InvokeComplete callback never called, so req done.*/
}


boolean DOFRequestGet_IsSubscribe( DOFRequest request )
{
    PCR_Assert( request );
    return ( request->opType != DOFOPERATIONTYPE_GET );
}


void DOFRequestGet_Return( DOFRequest request, DOFValue value )                 /* Called only by app, never by COAL. */
{
    PCR_Assert( request );
    OALMutexPool_Lock( request );
    if (request->pendState != APP_CANNOT_ACCEPT_OR_THROW && request->pendState != APP_CAN_THROW_ONLY)
    {
        if (request->opType == DOFOPERATIONTYPE_SUBSCRIBE)
        {
            /* We were called by app in one of these ways:
                    Subscribe had callback, it caused DOFRequestSubscribe_Return() to be called, and it called us.
                                                        (request->pendState is SUBSCRIBE_CALLBACK_OK_NOW_NEED_GET)
                    Subscribe had NO callback, so ->Get callback was called instead, it caused us to be called.
                                                                            (request->pendState is NORMAL_PEND) */
            request->pendState = (request->pendState == SUBSCRIBE_CALLBACK_OK_NOW_NEED_GET) ?
                                                                  APP_CAN_THROW_ONLY :            /* op not complete. */
                                                                  APP_CANNOT_ACCEPT_OR_THROW;     /* op  is complete. */
            OALMutexPool_Unlock( request );
            DOFSystem_CompleteGetRequest( request->system, request, value );
            /* If DOFRequestSubscribe_Return() was used to get us here, then APP_CAN_THROW_ONLY is in effect, and the
               original subscribe op shall stay alive until the app throws a request exception or the op is cancelled.
               We do not here unpend the request nor release our refCount on it. But this next block is the other case*/
            if (request->pendState == APP_CANNOT_ACCEPT_OR_THROW)
            {
                /* The subscribe op had no callback, so we were instead called by the ->Get callback. All use of the
                   request is now complete so unpend and release it. */
                DOFRequest_Unpend( request );
                DOFRequest_Release( request );
            }
            return;
        }
        else
        {
            PCR_Assert( request->reqOp->operation_type == DOFOPERATIONTYPE_GET ||
                        request->reqOp->operation_type == DOFOPERATIONTYPE_CHANGED );
            DOFRequest_Return_NoCompleteCallback( request,
                                                  DOFOPERATIONTYPE_GET, /* Call accepts _GET or _CHANGED. */
                                                  value, 0, NULL, TRUE );
        }
    }
    OALMutexPool_Unlock( request );
}


void DOFRequestSet_Return( DOFRequest request )                                 /* Called only by app, never by COAL. */
{
    PCR_Assert( request );
    DOFRequest_Return_NoCompleteCallback( request, DOFOPERATIONTYPE_SET, NULL, 0, NULL, FALSE );
}


void DOFRequestInvoke_Return( DOFRequest request, uint16 valuesCount,
                              const DOFValue values[] )                         /* Called only by app, never by COAL. */
{
    PCR_Assert( request );
    DOFRequest_Return_NoCompleteCallback( request, DOFOPERATIONTYPE_INVOKE, NULL, valuesCount, values, FALSE );
}

uint32 DOFRequest_GetTimeRemaining( DOFRequest request )
{
    PCR_Assert( request );
    return DOFOperation_GetTimeRemaining( request->reqOp );
}

/**
@}
**/
