/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/dof.h>
#include <internal/resource_pool.h>
#include <internal/dofinternal.h>

#ifdef OALRESOURCEPOOL_TRACK_USED
#include <stdio.h>
#endif

#ifndef OALRESOURCEPOOL_ALLOC_BLOCK_SIZE
# define OALRESOURCEPOOL_ALLOC_BLOCK_SIZE 16
#endif

#define OALRESOURCEPOOL_ALLOC_FAILURE_RETRY_PERIOD 500  /* Time, in milliseconds, to wait if (re)allocation fails before trying again. */


/* the pool uses a binary search method with insertion sort. we can't use tree
 to keep track of it because tree is dependent on this. plus, the array is very
 tiny (usually!) */
struct OALResourcePool_Resource_t {
    void *resource;
    OALResourcePoolKey_t key;
    uint32 ref_count;
};

struct OALResourcePool_t {
    PCRMutex mutex;
    PCRCondition cond;
    OALResourcePool_Create_Resource_fn_t create_fn;
    OALResourcePool_Destroy_Resource_fn_t destroy_fn;
    uint16 initial;
    uint16 total;
    uint16 used;
#ifdef OALRESOURCEPOOL_TRACK_USED
    uint16 max_used;
#endif
    struct OALResourcePool_Resource_t * resources;
};

OALResourcePool OALResourcePool_Create(uint16 num_resources, OALResourcePool_Create_Resource_fn_t create_fn, OALResourcePool_Destroy_Resource_fn_t destroy_fn)
{
    uint32 size;
    OALResourcePool p = (OALResourcePool)DOFMem_Alloc ( sizeof(struct OALResourcePool_t) );

    PCR_Assert( create_fn );
    PCR_Assert( destroy_fn );

    if ( p != NULL )
    {
        PCRMem_Set( p, 0, sizeof(struct OALResourcePool_t) );
        p->create_fn = create_fn;           /* functions need to be set early in case of error */
        p->destroy_fn = destroy_fn;
        p->initial = num_resources;
        if ( !num_resources )
            num_resources = OALRESOURCEPOOL_ALLOC_BLOCK_SIZE * 2;
        p->total = num_resources;
        size = sizeof(struct OALResourcePool_Resource_t) * num_resources;

        p->resources = (struct OALResourcePool_Resource_t *) DOFMem_Alloc( size );
        if ( p->resources != NULL)
        {
            PCRMem_Set( p->resources, 0, size );
            p->mutex = PCRMutex_Create();
            if ( p->mutex != NULL)
            {
                p->cond = PCRCondition_Create();
                if ( p->cond != NULL)
                {
                    boolean error = FALSE;

                    do {
                        --num_resources;
                        p->resources[num_resources].resource = create_fn();
                        if ( !p->resources[num_resources].resource )
                        {
                            error = TRUE;
                        }
                    } while ( num_resources && !error);
                    if(!error)
                    {
                        return p;
                    }
                    PCRCondition_Destroy(p->cond);
                }
                PCRMutex_Destroy(p->mutex);
            }
            DOFMem_Free(p->resources);
        }
        DOFMem_Free(p);
    }
    return NULL;
}

void OALResourcePool_Destroy(OALResourcePool pool)
{
    PCR_Assert( pool );
    PCR_Assert( pool->destroy_fn );
#ifdef OALRESOURCEPOOL_TRACK_USED
    printf( "Pool %p destroyed with %u of %u in use.\n", pool, (unsigned int) pool->used, pool->total );
#endif
#ifndef OALRESOURCEPOOL_KEEP_AFTER_USE
    PCR_Assert( !pool->used ); /* don't call this when resources are in use */
#endif

    PCRMutex_Lock( pool->mutex );
    do {
        --(pool->total);
        pool->destroy_fn( pool->resources[pool->total].resource );
    } while ( pool->total );

    PCRMutex_Unlock( pool->mutex );
    PCRMutex_Destroy( pool->mutex );
    PCRCondition_Destroy( pool->cond );
    DOFMem_Free( pool->resources );
    DOFMem_Free( pool );
}

static uint16 find_resource(OALResourcePool pool, OALResourcePoolKey_t key)
{
    /* use simple binary search to see if the resource is in the table */
    uint16 start = 0;
    uint16 end = pool->used;
    uint16 i = end >> 1;

    for(;start<end;) {
        if(pool->resources[i].key<key) {
            start = i+1; /* no need to go past here */
        }
        else {
            end = i; /* don't go past here */
        }
        i = ((start+end)>>1);
    }

    return i; /* didn't find it */
}

static void *get_resource(OALResourcePool pool, OALResourcePoolKey_t key, boolean refs)
{
    uint16 index;
    struct OALResourcePool_Resource_t *r;

    PCR_Assert(key); /* no null keys allowed */
    PCR_Assert(pool);
    PCRMutex_Lock(pool->mutex);

  retry:
    r = NULL;
    if(pool->used) {
        /* must see if we are already using this key */
        index = find_resource(pool, key);
        if(index<pool->used)
            r = &pool->resources[index];
    } else {
        index = 0;
    }

    if(r&&(r->key==key)) {
        void *resource;

#ifndef OALRESOURCEPOOL_KEEP_AFTER_USE
        PCR_Assert(r->ref_count);
#endif

        /* found one in use */
        if(refs) ++r->ref_count;
        resource = r->resource;
        PCRMutex_Unlock(pool->mutex);
        return resource;
    }
    else {
        struct OALResourcePool_Resource_t resource;

        if ( pool->used == pool->total )
        {
            if ( !pool->initial && ( (uint32)pool->total + OALRESOURCEPOOL_ALLOC_BLOCK_SIZE <= 0xFFFFU ) )
            {
                /* no resources left so grow the pool size */
                uint32 oldSize = sizeof(struct OALResourcePool_Resource_t) * pool->total;
                uint32 addSize = sizeof(struct OALResourcePool_Resource_t) * OALRESOURCEPOOL_ALLOC_BLOCK_SIZE;
                struct OALResourcePool_Resource_t * resources = (struct OALResourcePool_Resource_t *) DOFMem_Realloc( pool->resources, oldSize + addSize );

                if ( resources )
                {
                    /* Successfully allocated more space. */
                    uint16 rsrc = pool->total;
                    pool->resources = resources;
                    PCRMem_Set( &resources[pool->total], 0, addSize);
                    while ( rsrc < ( pool->total + OALRESOURCEPOOL_ALLOC_BLOCK_SIZE ) )
                    {
                        if ( ( pool->resources[rsrc].resource = pool->create_fn() ) == NULL )
                            break;
                        else
                            ++rsrc;
                    }
                    if ( rsrc > pool->total )
                    {
                        pool->total = rsrc;
                        /* Signal in case anyone is waiting. */
                        PCRCondition_Signal( pool->cond );
#ifdef OALRESOURCEPOOL_TRACK_USED
                        printf( "Pool %p resized up to %u.\n", pool, (unsigned int) pool->total );
#endif
                        goto retry;
                    }
                }
                /* No resources left (and couldn't alloc more), so wait for a resource (or more memory) to free up. */
                PCRCondition_Wait( pool->cond, pool->mutex, OALRESOURCEPOOL_ALLOC_FAILURE_RETRY_PERIOD );
            }
            else
            {
                /* No resources left (and not allowed to alloc more), so wait (forever) for a resource to free up. */
                PCRCondition_Wait( pool->cond, pool->mutex, PCRTIMEOUT_FOREVER );
            }
            goto retry;
        }

        PCR_Assert(pool->used < pool->total); /* should not be here without resources */

        /* we'll take this one */
        resource = pool->resources[pool->used];

        PCR_Assert(refs);
        resource.ref_count = 1;
        resource.key = key;

        if(pool->used) {
            /* if there is at least one resource being used, then make room for this new one */
            PCRMem_Move(&pool->resources[index+1], &pool->resources[index], sizeof(struct OALResourcePool_Resource_t)*(pool->used-index));
        }

        pool->resources[index] = resource;
        ++pool->used; /* tick up the resource usage count */
#ifdef OALRESOURCEPOOL_TRACK_USED
        if ( pool->used > pool->max_used )
        {
            pool->max_used = pool->used;
            printf( "Pool %p use up to %u.\n", pool, (unsigned int) pool->used );
        }
#endif

        PCRMutex_Unlock(pool->mutex);

        return resource.resource;
    }
}

void *OALResourcePool_GetResourceNoRef(OALResourcePool pool, OALResourcePoolKey_t key)
{
    return get_resource(pool, key, FALSE);
}

void *OALResourcePool_GetResource(OALResourcePool pool, OALResourcePoolKey_t key)
{
    return get_resource(pool, key, TRUE);
}

#ifdef OALRESOURCEPOOL_KEEP_AFTER_USE
void OALRESOURCEPOOL_KEEP_AFTER_USE_OALResourcePool_ReleaseResource(OALResourcePool pool, OALResourcePoolKey_t key, boolean destroy)
#else
void OALResourcePool_ReleaseResource(OALResourcePool pool, OALResourcePoolKey_t key)
#endif
{
    uint16 index=0;
    struct OALResourcePool_Resource_t *r;

    PCR_Assert(pool);
    PCR_Assert(key);
    PCRMutex_Lock(pool->mutex);

    index = find_resource(pool, key);
    r = &pool->resources[index];

#ifdef OALRESOURCEPOOL_KEEP_AFTER_USE
    if ( !destroy )
        --r->ref_count;
    else if ( r->key == key )
#else
    PCR_Assert(r->key==key);
    if ( !--r->ref_count )
#endif
    {
        /* we need to clear this one out of the list */
        PCR_Assert( r->ref_count == 0 );
        r->key = 0;
#ifdef OALRESOURCEPOOL_KEEP_AFTER_USE
        {
            void * rsrc = pool->create_fn();
            if ( rsrc )
            {
                pool->destroy_fn( r->resource );
                r->resource = rsrc;
            }
        }
#endif
        if ( pool->used > 1 )
        {
            struct OALResourcePool_Resource_t resource = *r; /* back this up for a second */
            PCRMem_Move(r, r+1, sizeof(struct OALResourcePool_Resource_t)*(pool->used-1-index));
            pool->resources[pool->used-1] = resource;
        }

        --pool->used;
        if ( !pool->initial && pool->used < ( pool->total - OALRESOURCEPOOL_ALLOC_BLOCK_SIZE * 2 ) )
        {
            /* too many resources left so shrink the pool size */
            uint16 newtotal = pool->total - OALRESOURCEPOOL_ALLOC_BLOCK_SIZE;
            if ( newtotal < OALRESOURCEPOOL_ALLOC_BLOCK_SIZE * 2 )
                newtotal = OALRESOURCEPOOL_ALLOC_BLOCK_SIZE * 2;
            while ( pool->total > newtotal )
            {
                pool->destroy_fn( pool->resources[--pool->total].resource );
            }
            pool->resources = (struct OALResourcePool_Resource_t *) DOFMem_Realloc( pool->resources, sizeof(struct OALResourcePool_Resource_t) * ( pool->total ) );
            PCR_Assert( pool->resources ); /* Smaller size must always work! */
#ifdef OALRESOURCEPOOL_TRACK_USED
            printf( "Pool %p resized down to %u.\n", pool, (unsigned int) pool->total );
#endif
        }
        PCRCondition_Signal( pool->cond );
    }

    PCRMutex_Unlock(pool->mutex);
}
