/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/address.h>
#include <dof/core.h>
#include <dof/dof.h>
#include <dof/error.h>
#include <dof/exception.h>
#include <dof/interface.h>
#include <dof/objectid.h>
#include <dof/permission.h>
#include <dof/providerinfo.h>
#include <dof/result.h>
#include <dof/type.h>
#include <dof/value.h>
#include <internal/channel.h>
#include <internal/connection_internal.h>
#include <internal/interfaceinternal.h>
#include <internal/iteminternal.h>
#include <internal/mutex_pool.h>
#include <internal/objectinternal.h>
#include <internal/operationinternal.h>
#include <internal/permissioninternal.h>
#include <internal/providerinfo.h>
#include <internal/resultinternal.h>
#include <internal/router.h>
#include <internal/type.h>
#include <internal/value.h>
#include <internal/exception.h>


/*
**  "Private" Type and Method Declarations
*/


/* for exceptions */
#define ISNEWSEQUENCE 0x01
#define ISSKIPPEDSEQUENCE 0x02
#define ISLOCAL 0x04

typedef struct StoredValue_t
{
    struct
    {
        DOFType  type;
        DOFValue value;
    } value;
    struct
    {
        DOFType  * types;
        DOFValue * values;
        uint16     count;
    } valueList;
    OALStream    stream;
    DOFException except;
} StoredValue_t;


typedef struct DOFResult_t
{
    DOFError        status;      /* The status of this result, must be one of the DOFERROR constants. */
    DOFProviderInfo provInfo;    /* The DOFProviderInfo associated with this result. */
    DOFResultType   resultType;  /* The type of this result, must be one of the DOFResultType constants. */
    DOFObjectID     providerID;  /* The object ID of the provider which returned this result. */
    DOFAddress      address;     /* The address of the responder, in case the application wants to connect. */
    OALChannel      channel;
    uint32          refCount;
    StoredValue_t   storedValue; /*The returned value stored in result: either a DOFValue or list of DOFValue elements*/
    uint16          sequenceNumber;
    uint8           flags;
    DOFPermission   permission;
} DOFResult_t;


/** @addtogroup DOFResult
@{
**/


/** @internal
    Sets up common DOFResult fields

    @param  channel     The channel from which the result was received.
    @param  address     The address from which the result was received.
    @param  status      The status of this DOFResult. Must be one of the valid DOFError codes in status.h
    @param  resultType  The type of DOFResult that this DOFResult is. Must be a valid DOFResultType from result.h
    @param  providerID  The OID of the provider object returning this result.

    @return The created DOFResult or NULL if the result could not be created or this method was used in error.
**/
static DOFResult create( OALChannel channel, DOFAddress address, DOFError status, DOFResultType resultType,
                         const DOFObjectID providerID )
{
    DOFResult retResult = (DOFResult)DOFMem_Alloc( sizeof( DOFResult_t ) );

    if (retResult)
    {
        PCRMem_Set( retResult, 0, sizeof( DOFResult_t ) );
        if (!providerID || (retResult->providerID = DOFObjectID_Copy( providerID )) != NULL)
        {
            if (!address || (retResult->address = DOFAddress_Copy( address )) != NULL)
            {
                boolean success = TRUE;

                retResult->refCount = 1;
                retResult->status = status;
                retResult->resultType = resultType;

                if (channel)
                {
                    success = FALSE;

                    retResult->channel = OALChannel_Retain( channel );
                    if (retResult->channel)
                    {
                        DOFConnectionState connState = OALChannel_CreateConnectionState( retResult->channel,
                                                                                         retResult->address );

                        if (connState)
                        {
                            retResult->provInfo = DOFProviderInfo_Create( connState, retResult->providerID,
                                                                          retResult->address );
                            DOFConnectionState_Destroy( connState );
                            if (retResult->provInfo)
                                success = TRUE;
                        }
                    }
                }
                if (success)
                    return retResult;   /* Success. */
                OALChannel_Release( retResult->channel );
                DOFAddress_Destroy( retResult->address );
            }
            DOFObjectID_Destroy( retResult->providerID );
        }
        DOFMem_Free( retResult );
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "DOFResult: Failed to allocate result." );

    return NULL;
}


static boolean DOFResult_SetValuesInStreamResult( DOFResult result, uint16 count, const DOFType types[],
                                                  DOFValue values[] )
{
    /* Used on requestor side only. OAP creates result from what comes in on the wire, and does not fill
       ->storedValue.valueList.values or ->storedValue.valueList.count. */
    OALStream holdStream;
    DOFType * tempTypes;

    PCR_Assert( result != NULL );
    PCR_Assert( result->resultType == DOFRESULTTYPE_VALUELIST_STREAM );
    PCR_Assert( result->storedValue.valueList.values == NULL );
    holdStream = result->storedValue.stream;      /* result->storedValue is a union, so .stream overlies .value.type. */
    /* Note: Passing in NULL types does NOT return NULL as is our customary expectation, because it is NOT an error.
             The exception to this, as expected, is that allocation failure still returns NULL; that is a true error. */
    tempTypes = DOFTypeListAlloc_Copy( count, types );
    if (tempTypes)
    {
        /* Transform result from DOFRESULTTYPE_VALUELIST_STREAM to DOFRESULTTYPE_VALUE_LIST. */
        result->storedValue.valueList.types = tempTypes;
        result->storedValue.valueList.values = values; /* Transfer ownership of values from caller. */
        result->storedValue.valueList.count = count;
        /* Special note: The next 2 code lines must be done as a pair. DOFResult_Destroy() won't cleanup stream
           because we changed result->resultType. */
        result->resultType = DOFRESULTTYPE_VALUE_LIST;
        OALStream_Release( holdStream, TRUE ); /*stream was prepared for us to extract info. We did, so cleanup stream*/
        return TRUE;            /* Success. */
    }
    else
    {
        /* Because ->resultType is unchanged, then DOFResult_Destroy() will cleanup stream. */
        return FALSE;           /* Fail. */
    }
}

static boolean DOFResult_SetExceptionInExceptionResult( DOFResult result, DOFInterfaceException interfaceExcept, uint8 count, DOFValue values[] )
{
    DOFException oldException = DOFResult_GetException( result );

    OALMutexPool_Lock( result );

    PCR_Assert( result != NULL );
    PCR_Assert( result->resultType == DOFRESULTTYPE_EXCEPTION );

    result->storedValue.except = DOFProviderException_Create_ProviderInfo( interfaceExcept, count, values, DOFException_GetProviderInfo( oldException ) );
    if( result->storedValue.except )
    {
        DOFException_Destroy( oldException );
        OALMutexPool_Unlock( result );
        return TRUE;
    }
    result->storedValue.except = oldException;
    OALMutexPool_Unlock( result );
    return FALSE;
}

static boolean DOFResult_SetValueInStreamResult( DOFResult result, const DOFType type, const DOFValue value )
{
    /* Used on requestor side only. OAP creates result from what comes in on the wire, and does not fill
       ->storedValue.value.value. */
    OALStream holdStream;
    DOFType   tempType;

    PCR_Assert( result != NULL );
    PCR_Assert( result->resultType == DOFRESULTTYPE_VALUE_STREAM );
    PCR_Assert( result->storedValue.value.value == NULL );
    holdStream = result->storedValue.stream;      /* result->storedValue is a union, so .stream overlies .value.type. */
    tempType = DOFType_Copy( type );
    if (tempType != NULL || type == NULL)
    {
        /* Transform result from DOFRESULTTYPE_VALUE_STREAM to DOFRESULTTYPE_VALUE. */
        result->storedValue.value.type = tempType;
        result->storedValue.value.value = value;
        /* Special note: The next 2 code lines must be done as a pair. DOFResult_Destroy() won't cleanup stream
           because we changed result->resultType. */
        result->resultType = DOFRESULTTYPE_VALUE;
        OALStream_Release( holdStream, TRUE ); /*stream was prepared for us to extract info. We did, so cleanup stream*/
        return TRUE;            /* Success. */
    }
    else
    {
        /* Because ->resultType is unchanged, then DOFResult_Destroy() will cleanup stream. */
        return FALSE;           /* Fail. */
    }
}


/** @internal
    Factory method for creating a DOFResult which contains a DOFValue.

    @param  channel     The channel from which the result was received.
    @param  address     The address from which the result was received.
    @param  status      The status of this DOFResult. Must be one of the valid DOFError codes in status.h
    @param  providerID  The OID of the provider object returning this result.
    @param  type        The type of the stored value returned from the provider object.
    @param  value       The stored value returned from the provider object.

    @return The created DOFResult or NULL if the result could not be created or this method was used in error.
**/
DOFResult DOFResult_Value_Create( OALChannel channel, DOFAddress address, DOFError status,
                                  const DOFObjectID providerID, const DOFType type, const DOFValue value )
{
    DOFResult retResult = NULL;

    /* The intention is that value and type both exist. It would be nice to know if we can assert both. */
    if ((retResult = create( channel, address, status, DOFRESULTTYPE_VALUE, providerID )) != NULL)
    {
        retResult->storedValue.value.type = DOFType_Copy( type );
        retResult->storedValue.value.value = DOFValue_Copy( value );
    }

    return retResult;
}


/** @internal
    Factory method for creating a DOFResult which contains a list of DOFValue elements.

    @param  channel     The channel from which the result was received.
    @param  address     The address from which the result was received.
    @param  status      The status of this DOFResult. Must be one of the valid DOFError codes in status.h
    @param  providerID  The OID of the provider object returning this result.
    @param  typeList    The type of the stored valuelist returned from the provider object.
    @param  valueList   The stored valuelist returned from the provider object.

    @return The created DOFResult or NULL if the result could not be created or this method was used in error
**/
DOFResult DOFResult_ValueList_Create( OALChannel channel, DOFAddress address, DOFError status,
                                      const DOFObjectID providerID, uint16 valueCount, const DOFType typeList[],
                                      const DOFValue valueList[] )
{
    DOFResult retResult = NULL;

    /* The intention is that valueList and typeList both exist and that valueCount is non-zero. Can we assert? */
    if ((retResult = create( channel, address, status, DOFRESULTTYPE_VALUE_LIST, providerID )) != NULL)
    {
        retResult->storedValue.valueList.count = valueCount;
        retResult->storedValue.valueList.types = DOFTypeListAlloc_Copy( valueCount, typeList );
        retResult->storedValue.valueList.values = DOFValueListAlloc_Copy( valueCount, valueList );
    }

    return retResult;
}


/** @internal
    Factory method for creating a DOFResult which contains neither a DOFValue nor a list of DOFValue elements.

    @param  channel     The channel from which the result was received.
    @param  address     The address from which the result was received.
    @param  status      The status of this DOFResult. Must be one of the valid DOFError codes in status.h
    @param  providerID  The OID of the provider object returning this result.

    @return The created DOFResult or NULL if the result could not be created or this method was used in error
**/
DOFResult DOFResult_NoValue_Create( OALChannel channel, DOFAddress address, DOFError status,
                                    const DOFObjectID providerID )
{
    DOFResult retResult = NULL;

    if ((retResult = create( channel, address, status, DOFRESULTTYPE_NO_VALUE, providerID )) != NULL)
    {
        PCR_Assert( retResult->storedValue.value.value == NULL ); /* create() memset to NULL then never changed these.*/
        PCR_Assert( retResult->storedValue.value.type == NULL );
    }

    return retResult;
}


DOFResult DOFResult_Stream_Create( OALChannel channel, DOFAddress address, DOFError status,
                                   const DOFObjectID providerID, OALStream stream, DOFResultType type )
{
    DOFResult retResult;

    PCR_Assert( providerID );
    PCR_Assert( stream );
    PCR_Assert( type == DOFRESULTTYPE_VALUE_STREAM || type == DOFRESULTTYPE_VALUELIST_STREAM );
    retResult = create( channel, address, status, type, providerID );
    if (retResult)
        retResult->storedValue.stream = (OALStream_Retain( stream ), stream);

    return retResult;
}

DOFResult DOFResult_ExceptionStream_Create(      OALChannel   channel,
                                                 DOFAddress   address,
                                           const DOFObjectID  providerID,
                                                 OALStream    stream,
                                                 DOFException exception)
{
    DOFResult ret;

    ret = create(channel, address, DOFERROR_OK, DOFRESULTTYPE_EXCEPTION, providerID);
    if(ret != NULL)
    {
        ret->storedValue.stream = (OALStream_Retain(stream), stream);
        ret->storedValue.except = DOFProviderException_CreateStream_ProviderInfo(stream, ret->provInfo, DOFInterfaceException_GetItemID(DOFProviderException_GetInterfaceException(exception)));
        if(ret->storedValue.except != NULL)
        {
            return ret;
        }
        DOFResult_Destroy(ret);
    }
    return NULL;
}


OALStream DOFResult_GetStream( DOFResult result )
{
    OALStream retStream;

    PCR_Assert( result );
    OALMutexPool_Lock( result );
    retStream = result->storedValue.stream;
    OALMutexPool_Unlock( result );

    return retStream;
}


/**
 * @internal
 * Creates an Exception Result
 *
 * @param channel         The channel from which the result was received.
 * @param address         The address from which the result was received.
 * @param except          The DOFException. This must not be NULL.
 * @param providerID      The providerID.
 *
 * @return DOFResult      The created result that must be destroyed at some point
 */
DOFResult DOFResult_Exception_Create( OALChannel channel, DOFAddress address, DOFException except,
                                      const DOFObjectID providerID )
{
    DOFResult retResult;

    PCR_Assert( except );
    if ((retResult = create( channel, address, DOFERROR_OK, DOFRESULTTYPE_EXCEPTION, providerID )) != NULL)
    {
        DOFException temp = except;

        if(DOFException_GetProviderInfo(except) == NULL)
            temp = DOFException_Copy_ProviderInfo(except, DOFResult_GetProviderInfo(retResult));
        else
            temp = DOFException_Copy( temp );

        if ((retResult->storedValue.except = temp ) != NULL)
            return retResult;
        DOFResult_Destroy( retResult );
    }

    return NULL;
}

/** @internal
    Create a shallow copy of a DOFResult.

    @note Must be freed with DOFResult_Destroy().
    @param srcRes

    @return The copied DOFResult (which can be NULL), or NULL if
            - memory for the copied result structure could not be allocated.
**/
DOFResult DOFResult_Copy( DOFResult srcRes )
{
    if (srcRes)
    {
        PCRMutex_Lock( ref_count_mutex );
        if ((++srcRes->refCount) == 0)
        {
            --(srcRes->refCount);
            srcRes = NULL;        /* Fail. */
        }
        PCRMutex_Unlock( ref_count_mutex );

    }

    return srcRes;
}

/**
 * Deep Copy a result
 * 
 * @param self The result to copy.
 * 
 * @return DOFResult The result if successful, NULL otherwise.
 */
DOFResult DOFResult_DeepCopy( const DOFResult self )
{
    DOFResult ret = NULL;

    if( self )
    {
        OALMutexPool_Lock( self );
        ret = create( self->channel, self->address, self->status, self->resultType, self->providerID );
        if( ret )
        {
            ret->flags = self->flags;
            ret->sequenceNumber = self->sequenceNumber;

            ret->permission = DOFPermission_Copy( self->permission );
            if( !self->permission || ret->permission )
            {
                boolean fail = FALSE;

                switch( self->resultType )
                {
                case DOFRESULTTYPE_VALUE:
                    ret->storedValue.value.value = DOFValue_Copy( self->storedValue.value.value );
                    if( ret->storedValue.value.value )
                    {
                        ret->storedValue.value.type = DOFType_Copy( self->storedValue.value.type );
                        if( ret->storedValue.value.type )
                        {
                            break;
                        }
                    }
                    fail = TRUE;
                    break;
                case DOFRESULTTYPE_VALUE_LIST:
                    ret->storedValue.valueList.values = DOFValueListAlloc_Copy( self->storedValue.valueList.count, self->storedValue.valueList.values );
                    if( ret->storedValue.valueList.values )
                    {
                        ret->storedValue.valueList.types = DOFTypeListAlloc_Copy( self->storedValue.valueList.count, self->storedValue.valueList.types );
                        if( ret->storedValue.valueList.types )
                        {
                            ret->storedValue.valueList.count = self->storedValue.valueList.count;
                            ret->storedValue.valueList.count = self->storedValue.valueList.count;
                            break;
                        }
                    }
                    fail = TRUE;
                    break;
                case DOFRESULTTYPE_EXCEPTION:
                    ret->storedValue.except = DOFException_Copy( self->storedValue.except );
                    if( !self->storedValue.except || ret->storedValue.except )
                    {
                        if( self->storedValue.stream )
                        {
                            ret->storedValue.stream = OALStream_CloneToRW( self->storedValue.stream );
                            if( ret->storedValue.stream )
                            {
                                break;
                            }
                        }
                        else
                        {
                            /* no stream to copy */
                            break;
                        }
                    }
                    fail = TRUE;
                    break;
                case DOFRESULTTYPE_VALUE_STREAM:
                case DOFRESULTTYPE_VALUELIST_STREAM:
                    if( self->storedValue.stream )
                    {
                        ret->storedValue.stream = OALStream_CloneToRW( self->storedValue.stream );
                        if( ret->storedValue.stream )
                        {
                            break;
                        }
                    }
                    else
                    {
                        /* no stream to copy */
                        break;
                    }
                    fail = TRUE;
                    break;
                default:
                    break;
                }

                /* Failed */
                if( fail )
                {
                    DOFResult_Destroy( ret );
                    ret = NULL;
                }
            }
        }
        OALMutexPool_Unlock( self );
    }

    return ret;
}

/** @internal
    Form a DOFResult that holds interface definition of the given interface or an error status.

    @param channel     The channel from which the result was received.
    @param address     The address from which the result was received.
    @param iface       The Interface we are asked to define. This should not be NULL.
    @param providerID  provider OID of the incoming request that need our response.

    @return DOFResult that holds a DOFValue with the interface definition, or NULL if
        - @a iface is NULL, or
        - memory allocations failed.
**/
DOFResult DOFResult_DefineComplete_Create( OALChannel channel, DOFInterface iface, const DOFObjectID providerID )
{
    if (iface)
    {
        uint32 size = DOFInterface_GetBytesSize(iface);
        OALStream stream = OALStream_CreateRW(size);

        if(stream != NULL)
        {
            if(OALStream_Write(stream, size, DOFInterface_GetBytes(iface)))
            {
                DOFResult ret = DOFResult_Stream_Create(channel, NULL, DOFERROR_OK, providerID, stream, DOFRESULTTYPE_VALUE_STREAM);

                OALStream_Release(stream, FALSE);
                return ret;
            }
            OALStream_Release(stream, FALSE);
        }
    }
    return NULL;
}


/*
**  "Public" Methods
*/


void DOFResult_Destroy( DOFResult result )
{
    if (result)
    {
        boolean free_ok = FALSE;

        PCRMutex_Lock( ref_count_mutex );
        if ((--result->refCount) == 0)
            free_ok = TRUE;
        PCRMutex_Unlock( ref_count_mutex );
        if (free_ok)
        {
            /* DOFValues and list of DOFValues are reference counted so always call destroy */
            switch (result->resultType)
            {
            case DOFRESULTTYPE_VALUE:
                DOFValue_Destroy( result->storedValue.value.value );
                DOFType_Destroy( result->storedValue.value.type );
                break;
            case DOFRESULTTYPE_VALUE_LIST:
                DOFValueListAlloc_Destroy( result->storedValue.valueList.count, result->storedValue.valueList.values );
                DOFTypeListAlloc_Destroy( result->storedValue.valueList.count, result->storedValue.valueList.types );
                break;
            case DOFRESULTTYPE_EXCEPTION:
                DOFException_Destroy( result->storedValue.except );
                OALStream_Release(result->storedValue.stream, TRUE);
                break;
            case DOFRESULTTYPE_VALUE_STREAM:
            case DOFRESULTTYPE_VALUELIST_STREAM:
                /* DOFSystem_AcceptResponse() changes ->resultType away from this case so we don't get here. But if it
                   fails then ->resultType doesn't change and we must cleanup stream. */
                OALStream_Release( result->storedValue.stream, TRUE );
                break;
            default:
                break;
            }
            OALMutexPool_Finalize( result );
            DOFAddress_Destroy( result->address );
            OALChannel_Release( result->channel );
            DOFObjectID_Destroy( result->providerID );
            DOFProviderInfo_Destroy( result->provInfo );
            DOFPermission_Destroy( result->permission );
            DOFMem_Free( result );
        }
    }
}


DOFObjectID DOFResult_GetProviderID( const DOFResult result )
{
    PCR_Assert( result );

    return result->providerID;
}


DOFProviderInfo DOFResult_GetProviderInfo( const DOFResult result )
{
    PCR_Assert( result );

    return result->provInfo;
}


DOFValue DOFResult_GetValue( const DOFResult result )
{
    DOFValue retValue = NULL;

    if (result)
    {
        OALMutexPool_Lock( result );
        if (result->resultType == DOFRESULTTYPE_VALUE)
            retValue = result->storedValue.value.value;
        OALMutexPool_Unlock( result );
    }

    return retValue;
}


DOFType DOFResult_GetValueType( const DOFResult result )
{
    DOFType retType = NULL;

    if (result)
    {
        OALMutexPool_Lock( result );
        if (result->resultType == DOFRESULTTYPE_VALUE)
            retType = result->storedValue.value.type;
        OALMutexPool_Unlock( result );
    }

    return retType;
}


const DOFValue * DOFResult_GetValueList( const DOFResult result )
{
    DOFValue * pRetValues = NULL;

    if (result)
    {
        OALMutexPool_Lock( result );
        if (result->resultType == DOFRESULTTYPE_VALUE_LIST)
            pRetValues = result->storedValue.valueList.values;
        OALMutexPool_Unlock( result );
    }

    return pRetValues;
}


const DOFType * DOFResult_GetValueListTypes( const DOFResult result )
{
    DOFType * pRetTypes = NULL;

    if (result)
    {
        OALMutexPool_Lock( result );
        if (result->resultType == DOFRESULTTYPE_VALUE_LIST)
            pRetTypes = result->storedValue.valueList.types;
        OALMutexPool_Unlock( result );
    }

    return pRetTypes;
}


uint16 DOFResult_GetValueListCount( const DOFResult result )
{
    uint16 ret = 0;

    if (result)
    {
        OALMutexPool_Lock( result );
        if (result->resultType == DOFRESULTTYPE_VALUE_LIST)
            ret = result->storedValue.valueList.count;
        OALMutexPool_Unlock( result );
    }

    return ret;
}


DOFException DOFResult_GetException ( const DOFResult result )
{
    DOFException retException = NULL;

    if (result)
    {
        OALMutexPool_Lock( result );
        if (result->resultType == DOFRESULTTYPE_EXCEPTION)
            retException = result->storedValue.except;
        OALMutexPool_Unlock( result );
    }

    return retException;
}


boolean DOFResult_IsInternal( DOFResult result )
{
    boolean retBoolean;

    PCR_Assert(result != NULL);
    OALMutexPool_Lock( result );
    retBoolean = result->resultType == DOFRESULTTYPE_EXCEPTION &&
                 DOFException_GetType( result->storedValue.except ) == DOFEXCEPTIONTYPE_ERROR;
    OALMutexPool_Unlock( result );

    return retBoolean;
}


boolean DOFResult_IsSkippedSequence( DOFResult result )
{
    PCR_Assert( result );

    return !!(result->flags & ISSKIPPEDSEQUENCE);
}


boolean DOFResult_IsNewSequence( DOFResult result )
{
    PCR_Assert( result );

    return !!(result->flags & ISNEWSEQUENCE);
}


boolean DOFResult_IsLocal( DOFResult result )
{
    PCR_Assert( result );

    return !!(result->flags & ISLOCAL);
}


uint16 DOFResult_GetSequenceNumber( DOFResult result )
{
    PCR_Assert( result );

    return result->sequenceNumber;
}


void DOFResult_SetSequenceNumber( DOFResult result, uint16 sequenceNumber, boolean skipped_sequence, boolean new_sequence )
{
    PCR_Assert( result );
    OALMutexPool_Lock( result );
    result->sequenceNumber = sequenceNumber;
    if (skipped_sequence)
        result->flags |= ISSKIPPEDSEQUENCE;
    else
        result->flags &= ~ISSKIPPEDSEQUENCE;
    if (new_sequence)
        result->flags |= ISNEWSEQUENCE;
    else
        result->flags &= ~ISNEWSEQUENCE;
    OALMutexPool_Unlock( result );
}


void DOFResult_SetLocal( DOFResult result, boolean local )
{
    PCR_Assert( result );
    OALMutexPool_Lock( result );
    if (local)
        result->flags |= ISLOCAL;
    else
        result->flags &= ~ISLOCAL;
    OALMutexPool_Unlock( result );
}

void DOFResult_SetStream(DOFResult result,
                         OALStream stream)
{
    PCR_Assert(result->storedValue.stream == NULL);
    OALMutexPool_Lock( result );
    result->storedValue.stream = (OALStream_Retain(stream), stream);
    OALMutexPool_Unlock( result );
}

DOFAddress DOFResult_GetAddress( DOFResult result )
{
    PCR_Assert( result != NULL );

    return result->address;
}


OALChannel DOFResult_GetProviderChannel( DOFResult result )
{
    PCR_Assert(result != NULL);

    return result->channel;
}


void DOFResult_SetPermission( DOFResult result, DOFPermission permission )
{
    PCR_Assert( result != NULL );
    PCR_Assert( permission != NULL );
    PCR_Assert( result->permission == NULL );
    result->permission = DOFPermission_Copy( permission );
}


DOFPermission DOFResult_GetPermission( DOFResult result )
{
    PCR_Assert( result != NULL );

    return result->permission;
}


boolean DOFResult_TransformStreamResult( DOFResult respResult, DOFOperation respOp )
{
    OALStream resultStream;
    boolean   success = FALSE;
    boolean   killStream = FALSE;

    PCR_Assert( respOp );
    PCR_Assert( respResult );
    OALMutexPool_Lock( respResult );
    resultStream = DOFResult_GetStream( respResult );
    PCR_Assert( resultStream );
    if ( respOp->operation_type != DOFOPERATIONTYPE_FINDINTERFACEDEFINITION )
    {
        DOFOperation sessionOp = DOFRouter_LookupSession(respOp->object_id.object_id);
        if(sessionOp != NULL)
        {
            uint32 bufferSize = OALStream_GetDataSize(resultStream);
            uint8 * buffer = (uint8*) OALStream_GetReadPtr(resultStream);
            uint8 * newBuffer = sessionOp->sessionTransformRequestor->fns->TransformReceiveData(sessionOp->sessionTransformRequestor,
                                                                                                respOp->interface_id, &bufferSize, buffer);

            if(newBuffer != buffer)
            {
                resultStream = OALStream_CreateRWWithData(newBuffer, bufferSize);
                if(resultStream == NULL)
                {
                    PCRMem_Free(newBuffer);
                    OALMutexPool_Unlock( respResult );
                    return FALSE;
                }
                killStream = TRUE;
            }
        }
    }

    if (respResult->resultType == DOFRESULTTYPE_VALUE_STREAM)
    {
        DOFValue value;

        if(respOp->operation_type == DOFOPERATIONTYPE_FINDINTERFACEDEFINITION)
        {
            value = DOFValueBLOB_Create(OALStream_GetDataSize(resultStream), OALStream_GetReadPtr(resultStream));
            if(value != NULL)
            {
                success = DOFResult_SetValueInStreamResult(respResult, NULL, value);
            }
        }
        else
        {
            boolean  resultTransformFailure;
            PCR_Assert(DOFItem_GetType( respOp->item ) == DOFITEM_PROPERTY);
            value = DOFInterfaceProperty_CreateValueFromStream( (DOFInterfaceProperty)respOp->item, resultStream,
                                                                &resultTransformFailure );
            if (!resultTransformFailure)
            success = DOFResult_SetValueInStreamResult( respResult,/*This transforms respResult to DOFRESULTTYPE_VALUE*/
                                                        DOFInterfaceProperty_GetType( (DOFInterfaceProperty)respOp->item ),
                                                        value );
        }
    }
    else if( respResult->resultType == DOFRESULTTYPE_VALUELIST_STREAM)
    {
        DOFValue * values;

        switch(DOFItem_GetType(respOp->item))
        {
            case DOFITEM_EVENT:
                values = DOFInterfaceEvent_CreateValueListFromStream( (DOFInterfaceEvent)respOp->item, resultStream );
                if (values)
                    success = DOFResult_SetValuesInStreamResult(respResult,/*Transforms respResult to DOFRESULTTYPE_VALUE_LIST*/
                                                                DOFInterfaceEvent_GetParameterCount((DOFInterfaceEvent)respOp->item),
                                                                DOFInterfaceEvent_GetParameterTypeList((DOFInterfaceEvent)respOp->item),
                                                                values);
                break;

            case DOFITEM_METHOD:
                values = DOFInterfaceMethod_CreateOutputParamValueListFromStream((DOFInterfaceMethod)respOp->item, resultStream);
                if(values)
                {
                    success = DOFResult_SetValuesInStreamResult(respResult,/*Transforms respResult to DOFRESULTTYPE_VALUE_LIST*/
                                                                DOFInterfaceMethod_GetOutputParameterCount((DOFInterfaceMethod)respOp->item),
                                                                DOFInterfaceMethod_GetOutputParameterTypeList((DOFInterfaceMethod)respOp->item),
                                                                values);
                }
                break;

            default:
                break;
        }
    }
    else if( respResult->resultType == DOFRESULTTYPE_EXCEPTION )
    {
        DOFException except = DOFResult_GetException( respResult );
        if ( DOFException_GetType( except ) == DOFEXCEPTIONTYPE_PROVIDER )
        {
            DOFInterfaceException interfaceExcept = DOFInterface_GetException( (respOp->item)?respOp->item->iface:respOp->provider_interface, DOFStreamException_GetItemId( except ) );
            if ( interfaceExcept )
            {
                DOFValue * values = DOFInterfaceException_CreateValueListFromStream( interfaceExcept, resultStream );
                if( values )
                {
                    uint8 count = DOFInterfaceException_GetParameterCount( interfaceExcept );
                    success = DOFResult_SetExceptionInExceptionResult( respResult, interfaceExcept, count, values );
                    /* The ownership could be passed to the exception as an optimization */
                    DOFValueListAlloc_Destroy( count, values );
                }
            }
        }
        else
        {
            /* Nothing to transform. */
            success = TRUE;
        }
    }
    else
    {
        /* the same result is routed to multiple local requestors */
    }

    if(killStream)
    {
        OALStream_Release(resultStream, FALSE);
    }

    OALMutexPool_Unlock( respResult );
    return success;
}


/**
@}
**/
