/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/interfaceid.h>
#include <dof/pcr.h>
#include <internal/channel.h>
#include <internal/dofinternal.h>
#include <internal/enqueue.h>
#include <internal/epp_negative_one.h>
#include <internal/interfaceidinternal.h>
#include <internal/operationinternal.h>
#include <internal/remotechannel.h>
#include <internal/remotechannelmanager.h>
#include <internal/requestinternal.h>
#include <internal/router.h>
#include <internal/securityscope.h>
#include <internal/slist.h>
#include <internal/systeminternal.h>
#include <internal/threadpool.h>
#include "r.h"
#include "interest.h"
#include "advertise.h"
#include "router_fns.h"


/** @internal
    Handle interest stuff for router.
@{
**/

#define SIDQUERY_TIMEOUT 8000
#define SIDQUERY_RETRY_PERIOD 3000

static OALHashtable interest_table;

/* 0 full binding, 1 is *I, 2 is O*, and 3 is ** */
static OALHashtable interest_tables[4];

static OALSListNode interestObserverList = NULL;

/** @internal
    Initialize the watch system.
**/
boolean interest_init(void)
{
    interest_tables[BINDING_TABLE] = OALHashtable_Create(3);
    interest_tables[BINDING_TABLE_INTERFACE] = OALHashtable_Create(2);
    interest_tables[BINDING_TABLE_OBJECT] = OALHashtable_Create(2);
    interest_tables[BINDING_TABLE_STARSTAR] = OALHashtable_Create(1);

    interest_table = OALHashtable_Create(3);

    if(interest_tables[0] && interest_tables[1] && interest_tables[2] && interest_tables[3] && interest_table)
    {
        return TRUE;
    }
    OALHashtable_Destroy(interest_tables[BINDING_TABLE], NULL, NULL);
    OALHashtable_Destroy(interest_tables[BINDING_TABLE_INTERFACE], NULL, NULL);
    OALHashtable_Destroy(interest_tables[BINDING_TABLE_OBJECT], NULL, NULL);
    OALHashtable_Destroy(interest_tables[BINDING_TABLE_STARSTAR], NULL, NULL);
    return FALSE;
}

static boolean _cleanup(OALHashItem_t op, void *unused)
{
    DOFOperation_Internal_Destroy(((DOFOperation)op));

    return FALSE;
}

static void query_observer( OALObserver self, OALObservable observable, void* ctx );

/**
    Clean up the interest system.
**/
void interest_cleanup(void)
{
    OALHashtable_Destroy(interest_tables[BINDING_TABLE], _cleanup, 0);
    OALHashtable_Destroy(interest_tables[BINDING_TABLE_INTERFACE], _cleanup, 0);
    OALHashtable_Destroy(interest_tables[BINDING_TABLE_OBJECT], _cleanup, 0);
    OALHashtable_Destroy(interest_tables[BINDING_TABLE_STARSTAR], _cleanup, 0);

    OALHashtable_Destroy(interest_table, _cleanup, 0);
}

typedef struct figure_t
{
    OALHashtable ht;
    OALHashtable ht2;
    uint8 type;
} figure_t;

static void figure_table_from_op(DOFOperation op, figure_t *t)
{
    /* Operation's mutex and router_mutex must already be locked. */
    uint8 index = 0;
    static OALHashtable *ltables[] = { &local_binding_tables[BINDING_TABLE], &local_binding_tables[BINDING_TABLE_INTERFACE],
        &local_binding_tables[BINDING_TABLE_OBJECT], &local_binding_tables[BINDING_TABLE] };
    static OALHashtable *rtables[] = { &remote_binding_tables[BINDING_TABLE], &remote_binding_tables[BINDING_TABLE_INTERFACE],
        &remote_binding_tables[BINDING_TABLE_OBJECT], &remote_binding_tables[BINDING_TABLE] };

    index |= (op->flags.bits.provider_iid_is_wildcard<<1);
    index |= (op->flags.bits.provider_oid_is_wildcard);

    t->type = index; /* corresponds to the order as defined in r.h */
    t->ht = *ltables[index];
    t->ht2 = *rtables[index];
}

typedef struct interestObservers_t
{
    OALSListNode    next;
    OALObserver     provOpObserver;
    OALObserver     reqOpObserver;
    OALObserver     channelObserver;
    DOFOperation    provOp;
    DOFOperation    reqOp;
    OALChannel      channel;
    DOFAddress      address;
}interestObservers_t;

typedef struct removeInterestObserver_t
{
    OALThreadPoolNode_t super;

    DOFOperation    provOp;
    OALObserver     provOpObserver;
    DOFOperation    reqOp;
    OALObserver     reqOpObserver;
}removeInterestObserver_t;

static void removeOpObservers_fn(OALThreadPoolItem item, boolean shutdown)
{
    removeInterestObserver_t * removeNode = (removeInterestObserver_t *)item;

    if (removeNode->provOp && removeNode->provOpObserver)
    {
        DOFOperation_RemoveOpWillCloseObserver(removeNode->provOp, removeNode->provOpObserver);
    }

    if (removeNode->reqOp && removeNode->reqOpObserver)
    {
        DOFOperation_RemoveOpWillCloseObserver(removeNode->reqOp, removeNode->reqOpObserver);
    }

    DOFOperation_Internal_Destroy(removeNode->reqOp);
    DOFOperation_Internal_Destroy(removeNode->provOp);
    DOFMem_Free(removeNode);
}


static boolean interestObserverList_traverseReqOp(OALSListNode node, void *context)
{
    interestObservers_t * inNode = (interestObservers_t *)node;

    if (inNode->reqOpObserver)
    {
        if (inNode->reqOpObserver == context)
        {
            return TRUE;
        }
    }

    return FALSE;
}

static void interestReqOpObserver_fn(OALObserver observer, OALObservable observable, void *arg)
{
    interestObservers_t *node;

    PCRMutex_Lock(router_mutex);

    node = (interestObservers_t *)OALSList_FindAndRemove(&interestObserverList, NULL, interestObserverList_traverseReqOp, observer);

    PCRMutex_Unlock(router_mutex);

    if (node)
    {
        removeInterestObserver_t * removeProvOpObserver;

        OALRemoteChannel_ConnectionStreamRequest( node->channel, node->address, TRUE );

        OALChannel_Lock(node->channel);
        OALObservable_RemoveObserver(node->channel->chObservable, node->channelObserver);
        OALChannel_Unlock(node->channel);

        /* Since the request op is already locked this cannot lock the provide op.  Use a seperate thread to remove it. */
        removeProvOpObserver = (removeInterestObserver_t *)DOFMem_Alloc(sizeof(removeInterestObserver_t));

        if (removeProvOpObserver)
        {
            /* transfer ownership of the provOp to the removeProvOpObserver. */
            removeProvOpObserver->provOp = node->provOp;
            node->provOp = NULL;
            removeProvOpObserver->provOpObserver = node->provOpObserver;
            removeProvOpObserver->reqOp = NULL;
            removeProvOpObserver->reqOpObserver = NULL;

            OALThreadPoolItem_Init((OALThreadPoolItem)removeProvOpObserver, removeOpObservers_fn);
            if (!OALThreadPool_Run((DOF_GetDof())->threadPool, (OALThreadPoolItem)removeProvOpObserver))
            {
                /* If this fails just leave the observer until the observable is destroyed. */
                DOFOperation_Internal_Destroy(removeProvOpObserver->provOp);
                DOFMem_Free(removeProvOpObserver);
            }
        }

        OALChannel_Release(node->channel);
        DOFAddress_Destroy(node->address);
        DOFOperation_Internal_Destroy(node->provOp);
        DOFOperation_Internal_Destroy(node->reqOp);
        DOFMem_Free(node);
    }

    /* This internal destroy is required by the operation observer functions. */
    DOFOperation_Internal_Destroy((DOFOperation)arg);
}

static boolean interestObserverList_traverseProvOp(OALSListNode node, void *context)
{
    interestObservers_t * inNode = (interestObservers_t *)node;

    if (inNode->provOpObserver)
    {
        if (inNode->provOpObserver == context)
        {
            return TRUE;
        }
    }

    return FALSE;
}

static void interestProvOpObserver_fn(OALObserver observer, OALObservable observable, void *arg)
{
    interestObservers_t *node;

    PCRMutex_Lock(router_mutex);

    node = (interestObservers_t *)OALSList_FindAndRemove(&interestObserverList, NULL, interestObserverList_traverseProvOp, observer);

    PCRMutex_Unlock(router_mutex);

    if (node)
    {
        OALRemoteChannel_ConnectionStreamRequest( node->channel, node->address, TRUE );

        /* Since this is the provide op it is safe to lock the request op to remove it's observer*/
        DOFOperation_RemoveOpWillCloseObserver(node->reqOp, node->reqOpObserver);

        OALChannel_Lock(node->channel);
        OALObservable_RemoveObserver(node->channel->chObservable, node->channelObserver);
        OALChannel_Unlock(node->channel);

        OALChannel_Release(node->channel);
        DOFAddress_Destroy(node->address);
        DOFOperation_Internal_Destroy(node->provOp);
        DOFOperation_Internal_Destroy(node->reqOp);
        DOFMem_Free(node);
    }

    /* This internal destroy is required by the operation observer functions. */
    DOFOperation_Internal_Destroy((DOFOperation)arg);
}

static boolean interestObserverList_traverseChannel(OALSListNode node, void *context)
{
    interestObservers_t * inNode = (interestObservers_t *)node;

    if (inNode->channelObserver)
    {
        if (inNode->channelObserver == context)
        {
            return TRUE;
        }
    }

    return FALSE;
}

static void interestChannelObserver_fn(OALObserver observer, OALObservable observable, void *arg)
{
    interestObservers_t *node;

    PCRMutex_Lock(router_mutex);

    node = (interestObservers_t *)OALSList_FindAndRemove(&interestObserverList, NULL, interestObserverList_traverseChannel, observer);

    PCRMutex_Unlock(router_mutex);

    if (node)
    {
        removeInterestObserver_t * removeBothObservers;

        OALRemoteChannel_ConnectionStreamRequest( node->channel, node->address, TRUE );

        /* The channel is locked here so it is not safe to lock the operations.  Use a seperate thread to remove the operation observers. */
        removeBothObservers = (removeInterestObserver_t *)DOFMem_Alloc(sizeof(removeInterestObserver_t));

        if (removeBothObservers)
        {
            /* transfer ownership of the provOp  and reqOp to the removeProvOpObserver. */
            removeBothObservers->provOp = node->provOp;
            node->provOp = NULL;
            removeBothObservers->provOpObserver = node->provOpObserver;
            removeBothObservers->reqOp = node->reqOp;
            node->reqOp = NULL;
            removeBothObservers->reqOpObserver = node->reqOpObserver;

            OALThreadPoolItem_Init((OALThreadPoolItem)removeBothObservers, removeOpObservers_fn);
            if (!OALThreadPool_Run((DOF_GetDof())->threadPool, (OALThreadPoolItem)removeBothObservers))
            {
                /* If this fails just leave the observers until the observable is destroyed. */
                DOFOperation_Internal_Destroy(removeBothObservers->provOp);
                DOFOperation_Internal_Destroy(removeBothObservers->reqOp);
                DOFMem_Free(removeBothObservers);
            }
        }

        DOFOperation_Internal_Destroy(node->provOp);
        DOFOperation_Internal_Destroy(node->reqOp);
        DOFAddress_Destroy(node->address);
        OALChannel_Release(node->channel);
        DOFMem_Free(node);
    }
}

static boolean addInterestObservers(DOFOperation reqOp, DOFOperation provOp, OALChannel channel, DOFAddress address)
{
    interestObservers_t * newNode = (interestObservers_t *)DOFMem_Alloc(sizeof(interestObservers_t));
    if (newNode)
    {
        newNode->reqOpObserver = DOFOperation_AddOpWillCloseObserver(reqOp, interestReqOpObserver_fn);
        if (newNode->reqOpObserver)
        {
            newNode->provOpObserver = DOFOperation_AddOpWillCloseObserver(provOp, interestProvOpObserver_fn);
            if (newNode->provOpObserver)
            {
                newNode->channel = OALChannel_Retain(channel);
                if (newNode->channel)
                {
                    OALChannel_Lock(newNode->channel);
                    newNode->channelObserver = OALObservable_AddObserver(newNode->channel->chObservable, interestChannelObserver_fn, newNode->channel);
                    OALChannel_Unlock(newNode->channel);
                    if (newNode->channelObserver)
                    {
                        newNode->address = DOFAddress_Copy(address);
                        if (newNode->address)
                        {
                            if (DOFOperation_Retain(reqOp))
                            {
                                newNode->reqOp = reqOp;
                                if (DOFOperation_Retain(provOp))
                                {
                                    newNode->provOp = provOp;
                                    newNode->next = NULL;
                                    PCRMutex_Lock(router_mutex);
                                    OALSList_Insert(&interestObserverList, (OALSListNode)newNode);
                                    PCRMutex_Unlock(router_mutex);
                                    return TRUE;
                                }
                                DOFOperation_Internal_Destroy(newNode->reqOp);
                            }
                            DOFAddress_Destroy(newNode->address);
                        }
                        OALChannel_Lock(newNode->channel);
                        OALObservable_RemoveObserver(newNode->channel->chObservable, newNode->channelObserver);
                        OALChannel_Unlock(newNode->channel);
                    }
                    OALChannel_Release(newNode->channel);
                }
                DOFOperation_RemoveOpWillCloseObserver(provOp, newNode->provOpObserver);
            }
            DOFOperation_RemoveOpWillCloseObserver(reqOp, newNode->reqOpObserver);
        }
        DOFMem_Free(newNode);
    }
    return FALSE;
}

static void processConnectInterest( DOFOperation reqOp, DOFOperation provOp )
{
    /* Called with provOp lock held. */
    /*DOFOperation_Lock( provOp );*/
    DOFOperation_Lock( reqOp );
    /*
     * Conditions:
     *   1. There exists at least one Connect level Interest Operation (streaming) or Full-Connect level (datagram)
     *      that has the connection as its source *AND* there exists at least one matching, security-compatible
     *      Provide Operation accessible by that connection (e.g., provided locally or provided remotely and we are
     *      configured as a router); OR
     *   2. There exists at least one Provide Operation that has the connection as its source *AND* there exists at
     *      least one matching, security-compatible Interest Operation of level Connect or Full Connect that is (or
     *      shortly will be) sent on that connection (e.g., local interest, or remote interest and we are configured
     *      as a router).
     * At this point in the code, we have already matched up the provide and interest, checked and verified security
     * compatibility, and we are actually ready to send a provide (in reaction to Interest, or beacuse of a new
     * Provide), so the only thing left to check is the interest level, connection type, which connection to notify,
     * and whether this is opening or closing.
     */
    if ( ( reqOp->operation_type == DOFOPERATIONTYPE_CONNECT ) || ( reqOp->operation_type == DOFOPERATIONTYPE_FULLCONNECT ) )
    {
        if ( !reqOp->flags.bits.activation_closing && !provOp->flags.bits.activation_closing )
        {
            /* Notify Requestor Connection */
            if ( reqOp->sources->channel->remoteChannelData && reqOp->sources->channel->remoteChannelData->flags.isConnection &&
                 ( ( reqOp->sources->channel->remoteChannelData->flags.isLossy && ( reqOp->operation_type == DOFOPERATIONTYPE_FULLCONNECT ) ) ||
                   !reqOp->sources->channel->remoteChannelData->flags.isLossy ) )
            {
                DOFAddress addr = reqOp->sources->channel->remoteChannelData->address;
                if ( !addr )
                    addr = reqOp->sources->channel->remoteChannelData->config.connection->targetAddress;

                if (addInterestObservers(reqOp, provOp, reqOp->sources->channel, addr))
                    OALRemoteChannel_ConnectionStreamRequest( reqOp->sources->channel, addr, FALSE );
            }
            /* Notify Provider Connection */
            if ( provOp->sources->channel->remoteChannelData && provOp->sources->channel->remoteChannelData->flags.isConnection)
            {
                DOFAddress addr = provOp->sources->channel->remoteChannelData->address;
                if ( !addr )
                    addr = provOp->sources->channel->remoteChannelData->config.connection->targetAddress;

                if (addInterestObservers(reqOp, provOp, provOp->sources->channel, addr))
                    OALRemoteChannel_ConnectionStreamRequest( provOp->sources->channel, addr, FALSE );
            }
        }
    }
    DOFOperation_Unlock( reqOp );
    /*DOFOperation_Unlock( provOp );*/
}

typedef struct __notify_new_t
{
    OALEnqueueItem_t super;

    DOFOperation provOp;
    DOFAddress address;
    OALChannel channel;
    DOFOperation reqOp; /* this is used for DOFConnectInterestListeners */
} __notify_new_t;

static void __notify_new(OALThreadPoolItem dnode, boolean shutdown)
{
    /* Security scope already checked by caller. */
    __notify_new_t * node;

    PCR_Assert(dnode != NULL);
    node = (__notify_new_t *) dnode;
    PCR_Assert( node->reqOp && node->provOp );
    /* we are going to send a provide pdu */
    if (!shutdown)
    {
        DOFAddress oldAddress = NULL;

        DOFOperation_Lock( node->provOp );             /* We manipulate op's address param, need protection. */
        PCR_Assert( node->provOp->operation_type == DOFOPERATIONTYPE_PROVIDE || node->provOp->operation_type == DOFOPERATIONTYPE_BEGINPROVIDE );
        if ( router_forwards() ||
             !( node->provOp->flags.bits.internally_created && node->channel->remoteChannelData ) )
        {
            /* Don't actually forward any provides if the interest is being cancelled (it still needs to be processed for connect interest, though). */
            if ( !node->reqOp->flags.bits.cancelled )
            {
                OALRouter_AddInterested_Operation(node->provOp, node->reqOp);

                /* Note: provideOp's ->address is temporarily overwritten with the requestor's DOFAddress. */
                if (node->address)
                {
                    oldAddress = node->provOp->sources->address;
                    node->provOp->sources->address = node->address;
                }
                /* Note: OALRemoteChannel_RouteToRequestor() is called here, which calls
                   CMT_RouteToRequestor_Create(). The latter takes providerOp's address and stores it for its actual use. Any
                   other call than OALRemoteChannel_RouteToRequestor() needs to do the same thing. */
                node->channel->routeToRequestor(node->channel, node->provOp, 0);
                if (node->address)
                    node->provOp->sources->address = oldAddress;
                /* Note: provideOp now has its original ->address member restored. */
                if ( !node->provOp->flags.bits.cancelled )
                    processConnectInterest( node->reqOp, node->provOp );
            }
            else
            {
                OALRouter_RemoveInterested_Operation(node->provOp, node->reqOp);
            }
        }
        DOFOperation_Unlock( node->provOp );
    }

    DOFOperation_Internal_Destroy(node->provOp);
    DOFOperation_Internal_Destroy(node->reqOp);
    OALChannel_Release(node->channel);
    DOFAddress_Destroy(node->address);
    DOFMem_Free(node);
}

static void __notify_existing(provider_binding_table_entry_t *pbte, DOFOperation op)
{
    /* Security scope already checked by caller. */
    if ( router_IsProvideCompatibleWithInterest( pbte->provider_op, op ) )
    {
        __notify_new_t * node;

        DOFOperation pop = pbte->provider_op;

        node = (__notify_new_t *) DOFMem_Alloc(sizeof(__notify_new_t));
        if(node != NULL)
        {
            OALEnqueueItem_Init((OALEnqueueItem) node, __notify_new);
            node->provOp = (DOFOperation_Retain(pop), pop);
            node->channel = OALChannel_Retain(op->sources->channel);
            node->address = DOFAddress_Copy(op->sources->address);
            node->reqOp = (DOFOperation_Retain(op), op);
            if(DOFERROR_OK != OALOperationManager_EnqueueOperation((OALEnqueueItem) node))
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OAPRouter: Unable to enqueue notification of provide.");
                DOFOperation_Internal_Destroy(pop);
                DOFOperation_Internal_Destroy(op);
                OALChannel_Release(op->sources->channel);
                DOFAddress_Destroy(node->address);
                DOFMem_Free( node );
            }
        }
    }
}

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t _notify_existing(OALAVLMultiTreeItem_t pbte, void* op, boolean new_keyset )
{
    if (DOFSecurityScope_IsCompatible( ((provider_binding_table_entry_t*)pbte)->provider_op->securityScope, ((DOFOperation)op)->securityScope ) &&
        (DOFObjectID_IsBroadcast( ((DOFOperation)op)->object_id.object_id ) ||
         DOFObjectID_MatchesAtLeast_Mcast( ((provider_binding_table_entry_t*)pbte)->provider_op->object_id.object_id, ((DOFOperation)op)->object_id.object_id )))
    {
        if ((((DOFOperation)op)->interface_id ? DOFInterfaceID_IsWildcard( ((DOFOperation)op)->interface_id ) : FALSE) ||
             DOFInterfaceID_IsEqual( ((DOFOperation)op)->interface_id, ((provider_binding_table_entry_t*)pbte)->provider_op->interface_id ))
          __notify_existing(((provider_binding_table_entry_t*)pbte), ((DOFOperation)op));
    }

    return OALAVLMULTITREE_NORMAL; /* need all bindings! */
}

static void _interest_keys(DOFSecurityScope securityScope, DOFOperation op, const void *keys[3], uint16 key_sizes[3], uint8 type)
{
    _make_keys( securityScope, op,
                op->sources->channel,                                                           /* channel not used. */
                keys, key_sizes );

    if ( type == BINDING_TABLE_OBJECT )
    {
        keys[1] = keys[2];
        key_sizes[1] = key_sizes[2];
    }
}

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t _notify_existing_starstar(OALAVLMultiTreeItem_t pbte, void* op, boolean new_keyset)
{
    if ( DOFSecurityScope_IsCompatible(((provider_binding_table_entry_t *)pbte)->provider_op->securityScope, ((DOFOperation)op)->securityScope) )
    {
        __notify_existing(((provider_binding_table_entry_t *)pbte), ((DOFOperation)op));
    }

    return OALAVLMULTITREE_NORMAL; /* need all bindings! */
}

static void interest_notify(DOFOperation op, OALHashtable mt, uint8 type)
{
    if ( type == BINDING_TABLE_STARSTAR )
    {
        OALHashtable_Traverse_All(mt, _notify_existing_starstar, op);
    }
    else
    {
        OALHashtable_Traverse_All(mt, _notify_existing, op);
    }

    /* No use notifying anyone if it's cancelled (provides are only sent in reaction to active interest). */
    if ( !op->flags.bits.cancelled )
    {
        DOFRouter_TraverseAdvertiseToRoutingInterests(op);
    }
}

/** @internal
    Add a new query operation.

    @param op The operation.
    @param mt The hashtable to insert the query operation into.
    @param type The type of operation. (wildcard, etc.)

    @return TRUE for success; otherwise FALSE.
**/
static boolean interest_add(DOFOperation op, OALHashtable mt, uint8 type)
{
    const void *keys[3];
    uint16 key_sizes[3];

    _interest_keys(op->securityScope, op, keys, key_sizes, type);

    switch(OALHashtable_Insert(interest_tables[type], op, keys, key_sizes))
    {
        default:
            break;
        case OALHashtable_InsertSuccess:
            DOFOperation_Retain(op);
            if(!DOFOperation_AddOpWillCloseObserver(op, query_observer))
            {
                (void)OALHashtable_Remove(interest_tables[type], op, keys, key_sizes);
                DOFOperation_Internal_Destroy(op);
                break;
            }
            /* Fall through */
        case OALHashtable_AlreadyInserted:
            return TRUE;
    }

    return FALSE;
}

struct DelayedCanelNode_t
{
    OALEnqueueItem_t super;

    DOFOperation operation;
};

/** @internal
    Remove an operation from the interest tables.

    @param op The operation.
    @param type The type of operation. (wildcard, etc.)
**/
static void interest_remove(DOFOperation op, uint8 type)
{
    const void *keys[3];
    uint16 key_sizes[3];

    _interest_keys(op->securityScope, op, keys, key_sizes, type);

    if(OALHashtable_Remove(interest_tables[type], op, keys, key_sizes))
    {
        DOFOperation_Internal_Destroy(op);
    }
}

static void query_observer(OALObserver self, OALObservable observable, void* ctx)
{
    DOFOperation op = (DOFOperation)ctx;
    figure_t t;

    figure_table_from_op(op, &t);

    interest_remove(op, t.type);
    DOFOperation_Internal_Destroy(op);
}

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t _notify_new_provider( OALAVLMultiTreeItem_t op, void* prov_op, boolean new_set )
{
    if (DOFSecurityScope_IsCompatible( ((DOFOperation)op)->securityScope, (((DOFOperation)prov_op)->securityScope )) &&
        (DOFObjectID_IsBroadcast( ((DOFOperation)op)->object_id.object_id ) ||
         DOFObjectID_MatchesAtLeast_Mcast( ((DOFOperation)prov_op)->object_id.object_id, ((DOFOperation)op)->object_id.object_id )) &&
        router_IsProvideCompatibleWithInterest( ((DOFOperation)prov_op), ((DOFOperation)op) ))
    {
        if ((((DOFOperation)op)->interface_id ? DOFInterfaceID_IsWildcard( ((DOFOperation)op)->interface_id ) : FALSE) ||
            DOFInterfaceID_IsEqual( ((DOFOperation)op)->interface_id, ((DOFOperation)prov_op)->interface_id ))
        {
            __notify_new_t * node;

            node = (__notify_new_t *) DOFMem_Alloc(sizeof(__notify_new_t));
            if(node != NULL)
            {
                OALEnqueueItem_Init((OALEnqueueItem) node, __notify_new);
                node->provOp = (DOFOperation_Retain(((DOFOperation)prov_op)), ((DOFOperation)prov_op));
                node->channel = OALChannel_Retain(((DOFOperation)op)->sources->channel);
                node->address = DOFAddress_Copy(((DOFOperation)op)->sources->address);
                node->reqOp = (DOFOperation_Retain(((DOFOperation)op)), ((DOFOperation)op));
                /* queue up send of provide */
                if (DOFERROR_OK == OALOperationManager_EnqueueOperation((OALEnqueueItem) node))
                    return OALAVLMULTITREE_NORMAL; /* keep going */

                OALChannel_Release(((DOFOperation)op)->sources->channel);
                DOFAddress_Destroy(((DOFOperation)op)->sources->address);
                DOFOperation_Internal_Destroy(((DOFOperation)prov_op));
                DOFOperation_Internal_Destroy(((DOFOperation)op));
                DOFMem_Free(node);
            }

            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OAPRouter: Failed to allocate state to notify interest of a new provide op.");
            return OALAVLMULTITREE_TERMINATE_TRAVERSE; /* out of memory; terminate */
        }
    }

    return OALAVLMULTITREE_NORMAL; /* keep going */
}

static void _queryCallback_newProvider_fn(void * context, boolean success);
static const struct EPPNegOneQueryCallbackFns_t EPPNegOneNewProviderFns = { _queryCallback_newProvider_fn };
static struct EPPNegOneQueryCallback_t EPPNegOneQueryNewProviderCB = { &EPPNegOneNewProviderFns };

/* Callback when SID is found so we can actually route the interest appropriately. */
static void _queryCallback_newProvider_fn(void * context, boolean success)
{
    DOFOperation op = (DOFOperation)context;
    if ( success )
    {
        DOFOperation_Lock( op );
        PCRMutex_Lock( router_mutex );
        interest_notify_new_provider( op );
        PCRMutex_Unlock( router_mutex );
        DOFOperation_Unlock( op );
    }

    DOFOperation_Internal_Destroy( op );
}

/** @internal
    Notify existing interests that a new provider is available.

    @param op The new provide operation.
**/
void interest_notify_new_provider(DOFOperation op)
{
    /* look through interest table for all interests on this binding, and send a PROVIDE */
    if (DOFOperationID_CanBeExplicit(op->opID))
    {
        OALHashtable_Traverse_All(interest_table, _notify_new_provider, op);
    }
    else
    {
        if (EPP_NegOne_Query(op->sources->channel, op->sources->address, SIDQUERY_TIMEOUT, SIDQUERY_RETRY_PERIOD, op, &EPPNegOneQueryNewProviderCB))
        {
            DOFOperation_Retain(op);
        }
    }
}

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t _interest_notify_new_channel( OALAVLMultiTreeItem_t op, void* channel, boolean new_keyset )
{
    /* only send our locally generated requests unless we forward */
    if ((!((DOFOperation)op)->flags.bits.internally_created) || (router_forwards() && !OALChannel_IsEquivalent( ((OALChannel)channel), ((DOFOperation)op)->sources->channel )))
        ((OALChannel)channel)->routeToProvider( ((OALChannel)channel), ((DOFOperation)op), NULL, NULL );

    return OALAVLMULTITREE_NORMAL; /* don't exit early */
}

/**
    Send all local interests to a new channel.

    @param channel The new channel.
**/
void interest_notify_new_channel(OALChannel channel)
{
    int i;

    /* don't send these to local system */
    if (channel->remoteChannelData)
    {
        for (i=0; i<3; ++i)
            OALHashtable_Traverse_All( interest_tables[i], _interest_notify_new_channel, channel );
    }
}

typedef struct interest_t {
    DOFObjectID oid;
    DOFInterfaceID iid;
    DOFOperationType type;
    boolean found;
} interest_t;

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t interest_exists(OALAVLMultiTreeItem_t op, void* i, boolean new_set)
{
    if ( DOFObjectID_MatchesAtLeast_Mcast( ((interest_t*)i)->oid, ((DOFOperation)op)->object_id.object_id ) )
    {
        if (DOFInterfaceID_IsWildcard(((DOFOperation)op)->interface_id) || DOFInterfaceID_IsEqual(((DOFOperation)op)->interface_id, ((interest_t*)i)->iid))
        {
            if (((DOFOperation)op)->operation_type >= ((interest_t *)i)->type)
            {
                ((interest_t*)i)->found = TRUE;
                return OALAVLMULTITREE_TERMINATE_TRAVERSE;
            }
        }
    }
    return OALAVLMULTITREE_NORMAL;
}

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t _interest_exists_any( OALAVLMultiTreeItem_t op, void* i, boolean new_set )
{
    if ( DOFObjectID_MatchesAtLeast_Mcast( ((interest_t*)i)->oid, ((DOFOperation)op)->object_id.object_id ) )
    {
        if ( ((DOFOperation)op)->operation_type >= ((interest_t*)i)->type )
        {
            ((interest_t*)i)->found = TRUE;
            return OALAVLMULTITREE_TERMINATE_TRAVERSE;
        }
    }
    return OALAVLMULTITREE_NORMAL;
}

/**
 Check to see if an interest operation exists given a security id, iid, and oid.

 @param securityScope The security scope.
 @param oid The object id.
 @param iid The interface id.
 @param type The interest level.

 @return TRUE if the activation exists; otherwise FALSE.
 **/
boolean DOFRouter_CheckInterest(DOFSecurityScope securityScope, DOFObjectID oid, DOFInterfaceID iid, DOFInterestLevel type)
{
    static const DOFOperationType map[] = { DOFOPERATIONTYPE_WATCH, DOFOPERATIONTYPE_ACTIVATE, DOFOPERATIONTYPE_CONNECT, DOFOPERATIONTYPE_FULLCONNECT };
    interest_t i;

    PCRMutex_Lock(router_mutex);
    PCR_Assert(type<=DOFINTERESTLEVEL_CONNECT_FULL);
    i.type = map[type];
    i.found = FALSE;
    i.oid = oid;
    i.iid = iid;

    if ( iid )
    {
        OALHashtable_Traverse_All( interest_table, interest_exists, &i );
    } else {
        OALHashtable_Traverse_All( interest_table, _interest_exists_any,  &i );
    }
    PCRMutex_Unlock(router_mutex);
    return i.found;
}

static void interest_observer(OALObserver self, OALObservable observable, void* ctx)
{
    DOFOperation op = (DOFOperation)ctx;
    OALHashKey_t keys[3];
    OALHashKeySize_t key_sizes[3];

    _make_keys( op->securityScope, op,
                op->sources->channel,                                                           /* channel not used. */
                keys, key_sizes );

    OALHashtable_Remove(interest_table, op, keys, key_sizes);

    op->flags.bits.activation_closing = 1;                        /* Intend that InterestRemoved callback be called. */
    DOF_HandleInterestRequest( DOF_GetDof(), op);

    DOFOperation_Internal_Destroy(op);
}

static void _queryCallback_SendToAll_fn(void * context, boolean success);
static const struct EPPNegOneQueryCallbackFns_t EPPNegOneQueryCallbackFns = { _queryCallback_SendToAll_fn };
static struct EPPNegOneQueryCallback_t EPPNegOneQueryCB = { &EPPNegOneQueryCallbackFns };

static void _queryCallback_SendToAll_fn(void * context, boolean success)
{
    if (success)
        OALRemoteChannelMgr_SendToAll((DOFOperation)context);

    DOFOperation_Internal_Destroy((DOFOperation)context);
}

DOFRouterStatusCode_t route_interest(DOFOperation op)
{
    figure_t t;
    OALHashKey_t keys[3];
    OALHashKeySize_t key_sizes[3];

    _make_keys( op->securityScope, op,
                op->sources->channel,                                                       /* channel not used. */
                keys, key_sizes );

    figure_table_from_op(op, &t);

    if ( !op->flags.bits.new_retry_count )
    {
        /* oap will reject wildcard IID and *.* activates */

        if(!interest_add(op, t.ht, t.type))
            return DOFROUTERSTATUS_ERROR;

        switch (OALHashtable_Insert(interest_table, op, keys, key_sizes))
        {
            case OALHashtable_InsertSuccess:
                if(!DOFOperation_AddOpWillCloseObserver(op, interest_observer))
                {
                    OALHashtable_Remove(interest_table, op, keys, key_sizes);
                    return DOFROUTERSTATUS_ERROR;
                }
                if(!op->sources->request)
                {
                    op->sources->request = DOFRequest_Create(NULL, op, NULL);
                }
                /* Intend that InterestAdded callback be called. */
                op->flags.bits.activation_closing = 0;
                DOF_HandleInterestRequest( DOF_GetDof(), op);
                break;
            case OALHashtable_AlreadyInserted:
                break;
            default:
                return DOFROUTERSTATUS_ERROR;
        }
    }

    interest_notify(op,t.ht,t.type);

    if ((!op->flags.bits.internally_created) || /* internally_created would mean this watch PDU came to us on the wire*/
        router_forwards() )
    {
        /* We originated this watch PDU or we are an OAS and this watch PDU came in on the wire. So send this watch
           PDU to as many places as is appropriate. */
        OALChannel rc = NULL;

        interest_notify(op,t.ht2,t.type);

        if (provider_route_normal_intercept_request_config(op, &rc) == DOFROUTERSTATUS_NO_REQUEST_CONFIG)
        {
            if (DOFOperationID_CanBeExplicit(op->opID))
            {
                OALRemoteChannelMgr_SendToAll(op);
            }
            else
            {
                if (EPP_NegOne_Query(op->sources->channel, op->sources->address, SIDQUERY_TIMEOUT, SIDQUERY_RETRY_PERIOD, op, &EPPNegOneQueryCB))
                {
                    DOFOperation_Retain(op);
                    return DOFROUTERSTATUS_SUCCESS;
                }
                else
                {
                    return DOFROUTERSTATUS_ERROR;
                }
            }
        }
        else
        { /* op was properly sent (or dropped) within above provider_route_normal_intercept_request_config() call. */
        }
    }

    return DOFROUTERSTATUS_SUCCESS;        /* Success does not mean "sent," it means properly sent or dropped. */
}

struct notify_info {
    DOFOperation adv_op;
    OALHashtable_Traversal_fn_t fnct;
    void * context;
};

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t _notify_new_advertise(OALAVLMultiTreeItem_t op, void* info, boolean new_set)
{
    if(DOFSecurityScope_IsCompatible(((DOFOperation)op)->securityScope, ((struct notify_info*)info)->adv_op->securityScope)) {
        if ( DOFObjectID_IsBroadcast( ((DOFOperation)op)->object_id.object_id ) || DOFObjectID_MatchesAtLeast_Mcast( ((struct notify_info*)info)->adv_op->object_id.object_id, ((DOFOperation)op)->object_id.object_id ) )
        {
            if ( ( ((struct notify_info*)info)->adv_op->interface_id ? DOFInterfaceID_IsWildcard( ((struct notify_info*)info)->adv_op->interface_id ) : FALSE ) ||
                 ( ((DOFOperation)op)->interface_id ? DOFInterfaceID_IsWildcard( ((DOFOperation)op)->interface_id ) : FALSE ) ||
                 DOFInterfaceID_IsEqual( ((DOFOperation)op)->interface_id, ((struct notify_info*)info)->adv_op->interface_id ) )
            {
                if ( ((struct notify_info*)info)->fnct( ((DOFOperation)op), ((struct notify_info*)info)->context ) )
                    return OALAVLMULTITREE_TERMINATE_TRAVERSE;
                return OALAVLMULTITREE_NORMAL; /* keep going */
            }
        }
    }
    return OALAVLMULTITREE_NORMAL; /* keep going */
}

void DOFRouter_TraverseInterestTables(DOFOperation                  op,
                                      OALHashtable_Traversal_fn_t   fnct,
                                      void                        * context)
{
    struct notify_info i;

    i.adv_op = op;
    i.fnct = fnct;
    i.context = context;

    OALHashtable_Traverse_All( interest_table, _notify_new_advertise, &i );
}

/**
@}
**/
