/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/dof.h>
#include <dof/interfaceid.h>
#include <internal/router.h>
#include <internal/slist.h>
#include <internal/channel.h>
#include <internal/operationinternal.h>
#include <internal/enqueue.h>
#include <internal/systeminternal.h>
#include <internal/runtimeinternal.h>
#include <internal/dofinternal.h>
#include "r.h"
#include "interest.h"
#include "advertise.h"
#include "router_fns.h"

#include <dof/core.h>
#include <dof/interfaceid.h>
#include <dof/pcr.h>
#include <dof/object.h>

/** @internal
    The router.
@{
**/

extern PCRMutex     ref_count_mutex;

PCRMutex router_mutex;

/**
    Init the router.
**/
boolean DOFRouter_Init(void)
{
    router_mutex = PCRMutex_Create();
    if(router_mutex != NULL)
    {
        if(advertise_init())
        {
            if(interest_init())
            {
                if(subscriptions_init())
                {
                    if(fns_init())
                    {
                        return TRUE;
                    }
                    subscriptions_cleanup();
                }
                interest_cleanup();
            }
            advertise_cleanup();
        }
        PCRMutex_Destroy(router_mutex);
        router_mutex = NULL;
    }
    return FALSE;
}

/**
    Clean up after the router.
**/
void DOFRouter_Cleanup(void)
{
    PCRMutex_Lock(router_mutex);
    fns_cleanup();
    PCRMutex_Unlock(router_mutex);

    interest_cleanup();
    /* No operations should exist at this point. */
    subscriptions_cleanup();
    advertise_cleanup();

    PCRMutex_Destroy(router_mutex);
    router_mutex = NULL;
}

void DOFRouter_Lock(void)
{
    if (router_mutex)
        PCRMutex_Lock(router_mutex);
}

void DOFRouter_Unlock(void)
{
    if (router_mutex)
        PCRMutex_Unlock(router_mutex);
}

/** @internal
    Fill the given keys with information suitable for looking up a binding.

    @param securityScope The security scope.
    @param iid The interface id.
    @param oid The object id.
    @param keys Holder for the three keys.
    @param key_sizes Holder for the three key sizes.
**/
void __make_keys(DOFSecurityScope securityScope, DOFInterfaceID iid, DOFObjectID oid, const void *keys[3], OALHashKeySize_t key_sizes[3])
{
    DOFObjectID did = DOFObjectIDDomain_GetObjectID(DOFSecurityScope_GetDomainID(securityScope));

    keys[0] = DOFObjectID_GetBytes(did);
    keys[1] = DOFInterfaceID_GetBytes(iid);
    keys[2] = DOFObjectID_GetBytes(oid); /* Used only by tables using 3 keys; ignored by those using only 2 keys. */

    key_sizes[0] = did ? DOFObjectID_GetBytesSize( did ) : 0;
    key_sizes[1] = iid ? DOFInterfaceID_GetBytesSize(iid) : 0;
    key_sizes[2] = oid ? DOFObjectID_GetBytesSize( oid ) : 0; /* Used only by tables using 3 keys; ignored by those using only 2 keys. */
}

/** @internal
    Make keys from a given operation suitable for binding lookup.

    @param op The operation that holds the oid, and iid.
    @param channel Holds the security ID, neither are used.
    @param keys Holder for the three keys.
    @param key_sizes Holder for the three key sizes.
**/
void _make_keys(DOFSecurityScope securityScope, DOFOperation op, OALChannel channel, const void *keys[3], OALHashKeySize_t key_sizes[3])
{
    __make_keys(securityScope, op->interface_id, op->object_id.object_id, keys, key_sizes);
}

/**
    Bump the ref count of a pbte.

    @param pbte The pbte to increase the ref count on.
**/
void provider_binding_table_entry_retain(provider_binding_table_entry_t *pbte)
{
    PCR_Assert(pbte);
    PCRMutex_Lock(router_mutex);
    ++(pbte->ref_count);
    PCRMutex_Unlock(router_mutex);
}

static void Interested_RemoveAll_fn_t(OALAVLTreeItem_t item, void *arg)
{
    /* Free node */
    interested_requester_t* interested = (interested_requester_t*)item;
    OALAVLTree_Destroy(interested->opTree, NULL, NULL);
    OALChannel_Release(interested->channel);
    DOFAddress_Destroy(interested->address);
    DOFSecurityScope_Destroy(interested->scope);
    DOFMem_Free((void*)item);
}

/**
    Decrement the ref count of a pbte.

    @param pbte The pbte to decrement the ref count of.
**/
void provider_binding_table_entry_release(provider_binding_table_entry_t *pbte)
{
    if(pbte)
    {
        boolean destroy = FALSE;

        PCRMutex_Lock(router_mutex);
        if(!--(pbte->ref_count))
        {
            destroy = TRUE;
        }
        PCRMutex_Unlock(router_mutex);

        if(destroy)
        {
            if(pbte->interested)
                OALAVLTree_Destroy( pbte->interested, Interested_RemoveAll_fn_t, NULL);
            OALChannel_Release(pbte->provider_channel);
            DOFOperation_Internal_Destroy(pbte->provider_op);
            DOFObject_Destroy(pbte->provider_object);
            DOFAddress_Destroy(pbte->address);
            DOFMem_Free(pbte);
        }
    }
}

/** @internal
Remove a binding from the tables.

@param tables The array of tables (local or remote) to remove the pbte from.
@param pbte The pbte to remove.
**/
void tables_remove_binding(OALHashtable *tables, provider_binding_table_entry_t *pbte)
{
    const void *keys[3];
    OALHashKeySize_t key_sizes[3];

    _make_keys( pbte->provider_op->securityScope, pbte->provider_op,
                pbte->provider_op->sources->channel,                                             /* channel not used. */
                keys, key_sizes );

    OALHashtable_Remove(tables[BINDING_TABLE], pbte, keys, key_sizes);
    OALHashtable_Remove(tables[BINDING_TABLE_INTERFACE], pbte, keys, key_sizes);

    keys[1] = keys[2];
    key_sizes[1] = key_sizes[2];

    OALHashtable_Remove(tables[BINDING_TABLE_OBJECT], pbte, keys, key_sizes);
}


/** @internal
    Add a binding to the tables.

    @param tables The array of tables (local or remote) to add the pbte to.
    @param pbte The pbte to add.

    @return TRUE if everything went ok; otherwise FALSE.
**/
boolean tables_add_binding(OALHashtable *tables, provider_binding_table_entry_t *pbte)
{
    const void *keys[3];
    OALHashKeySize_t key_sizes[3];
    const void *okeys[2];
    OALHashKeySize_t okey_sizes[2];

    _make_keys( pbte->provider_op->securityScope, pbte->provider_op,
                pbte->provider_op->sources->channel,                                             /* channel not used. */
                keys, key_sizes );

    if(OALHashtable_Insert(tables[BINDING_TABLE], pbte, keys, key_sizes) != OALHashtable_InsertSuccess) goto out_fail_binding;
    if(OALHashtable_Insert(tables[BINDING_TABLE_INTERFACE], pbte, keys, key_sizes) != OALHashtable_InsertSuccess) goto out_fail_interface;

    okeys[0] = keys[0];
    okeys[1] = keys[2];
    okey_sizes[0] = key_sizes[0];
    okey_sizes[1] = key_sizes[2];

    if(OALHashtable_Insert(tables[BINDING_TABLE_OBJECT], pbte, okeys, okey_sizes) != OALHashtable_InsertSuccess) goto out_fail_object;

    return TRUE;

out_fail_object:
    OALHashtable_Remove(tables[BINDING_TABLE_INTERFACE], pbte, keys, key_sizes);

out_fail_interface:
    OALHashtable_Remove(tables[BINDING_TABLE], pbte, keys, key_sizes);

out_fail_binding:
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "OAPRouter: Failed to add new binding.");
    return FALSE;
}

static boolean find_binding(OALHashItem_t pbte, void* found)
{
    *((provider_binding_table_entry_t**)found) = ((provider_binding_table_entry_t*)pbte);

    return TRUE; /* exit traversal early */
}

/** @internal
    See if a binding exists.

    @param table The table to search in.
    @param op The operation looking for this binding.

    @return The pbte; otherwise NULL if none found.
**/
provider_binding_table_entry_t *binding_exists(OALHashtable table, DOFOperation op)
{
    const void *keys[3];
    OALHashKeySize_t key_sizes[3];
    provider_binding_table_entry_t *found = NULL;

    _make_keys( op->securityScope, op,
                op->sources->channel,                                                            /* channel not used. */
                keys, key_sizes );

    OALHashtable_Traverse(table, keys, key_sizes, find_binding, &found);

    return found;
}

provider_binding_table_entry_t *DOFRouter_GetBindingEntry_Internal(DOFSecurityScope securityScope, DOFInterfaceID iid, DOFObjectID oid)
{
    /* Called with router mutex locked. */
    provider_binding_table_entry_t *pbte = NULL;

    const void *keys[3];
    uint16 key_sizes[3];

    __make_keys(securityScope, iid, oid, keys, key_sizes);

    OALHashtable_Traverse(local_binding_tables[BINDING_TABLE], keys, key_sizes, find_binding, &pbte);
    if ( !pbte )
        OALHashtable_Traverse(remote_binding_tables[BINDING_TABLE], keys, key_sizes, find_binding, &pbte);

    return pbte;
}

/** @internal
    Remove a local alias.

    @param ht The alias table.
    @param alias The alias to remove.
**/
void alias_remove(OALHashtable ht, OALAlias alias)
{
    if(alias) {
        OALHashKeySize_t alias_len = sizeof(alias);
        OALAlias *alias_ptr = &alias;

        OALHashtable_Remove_First(ht, (const void**)&alias_ptr, &alias_len);
    }
}

/**
    Retrieve a pbte based on a local alias.

    @param ht The alias table.
    @param alias The alias to retrieve.

    @return The pbte for that alias; otherwise NULL if none found.
**/
provider_binding_table_entry_t *alias_retrieve(OALHashtable ht, OALAlias alias)
{
    if(alias) {
        OALHashKeySize_t alias_len = sizeof(alias);
        OALAlias *alias_ptr = &alias;

        return (provider_binding_table_entry_t*)OALHashtable_Retrieve(ht, (const void**)&alias_ptr, &alias_len);
    }

    return 0;
}

/**
    Insert a local alias.

    @param ht The alias table.
    @param alias The alias to insert.
    @param pbte The pbte for the alias to be entered.

    @return TRUE if everything went ok; otherwise FALSE.
**/
boolean alias_insert(OALHashtable ht, OALAlias alias, provider_binding_table_entry_t *pbte)
{
    if(alias) {
        OALHashKeySize_t alias_len = sizeof(alias);
        OALAlias *alias_ptr = &alias;

        /* Only the primary provider should ever be in the alias table. */
        PCR_Assert( pbte->flags.bits.primary );
        if(OALHashtable_Insert(ht, pbte, (const void**)&alias_ptr, &alias_len) != OALHashtable_InsertSuccess) {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "OAPRouter: Failed to add primary binding alias.");
            return FALSE;
        }
    }

    return TRUE; /* no alias, so it's ok */
}

static boolean pbte_figure_primary_remote(OALHashItem_t pbte, void* best)
{
    ((provider_binding_table_entry_t*)pbte)->flags.bits.primary = FALSE;
    if ( !*((provider_binding_table_entry_t**)best) || DOFOperation_IsBetterSource( ((provider_binding_table_entry_t*)pbte)->provider_op, (*((provider_binding_table_entry_t**)best))->provider_op->sources->distLossless, (*((provider_binding_table_entry_t**)best))->provider_op->sources->distLossy ) )
        *((provider_binding_table_entry_t**)best) = ((provider_binding_table_entry_t*)pbte);

    return FALSE; /* search all */
}

static boolean pbte_figure_primary_local(OALHashItem_t pbte, void* best)
{
    /* Keep existing local primary. */
    if ( ((provider_binding_table_entry_t*)pbte)->flags.bits.primary )
    {
        *((provider_binding_table_entry_t**)best) = ((provider_binding_table_entry_t*)pbte);
        return TRUE; /* Only one primary; we're done. */
    }
    if ( !*((provider_binding_table_entry_t**)best) )
        *((provider_binding_table_entry_t**)best) = ((provider_binding_table_entry_t*)pbte);

    return FALSE; /* search all */
}

static void MergeInterestedOp_Traversal(OALAVLTreeItem_t item, void *arg)
{
    interested_requester_t* interested = (interested_requester_t*)arg;

    OALAVLTree_InsertionStatus status = OALAVLTree_Insert(interested->opTree, item, item, NULL);
    if (status == OALAVLTree_InsertFailure)
    {
        DOFLog_Message(DOFLOGLEVEL_WARNING, "OALRouter: Failed to add interested for new primary provider, insert failure.");
    }
}

static boolean MergeInterested_Traversal(OALAVLTreeItem_t item, void *arg)
{
    interested_requester_t* ainterested = (interested_requester_t*)item;
    interested_requester_t* destInterested;
    OALAVLTree interested = (OALAVLTree)arg;

    OALAVLTree_InsertionStatus status = OALAVLTree_Insert(interested, ainterested, ainterested, (OALAVLTreeItem_t*)&destInterested);
    switch(status)
    {
    case OALAVLTree_InsertSuccess:
        /* DO Nothing */
        break;

    case OALAVLTree_AlreadyInserted:
    default:
        /* destroy the opTree and add all items to the new tree. */
        OALAVLTree_Destroy(ainterested->opTree, MergeInterestedOp_Traversal, destInterested);
        DOFSecurityScope_Destroy(ainterested->scope);
        DOFAddress_Destroy(ainterested->address);
        OALChannel_Release(ainterested->channel);
        DOFMem_Free(ainterested);
        break;
    }
    return FALSE;
}

static void OALRouter_MergeInterested(OALAVLTree* dest, OALAVLTree* src)
{
    if(*dest != NULL)
    {
        DOFLog_Message(DOFLOGLEVEL_WARNING, "OALRouter: New primary provider already has an interested list.");
        if(*src != NULL)
        {
            OALAVLTree_Traverse(*dest, MergeInterested_Traversal, *src); 
            OALAVLTree_Destroy(*dest, NULL, NULL);
            /* dest is set below so no need to set it to NULL */
        }
        else
        {
            /* This should never happen but if we already have a list on the dest and non on the src, don't kill the list */
            return; 
        }
    }

    /* Move list to dest */
    *dest = *src;
    *src = NULL;
}

boolean DOFRouter_EstablishPrimaryProvider( provider_binding_table_entry_t * pbte, boolean notify, boolean preferred )
{
    /* Called with router mutex locked. */
    const void *keys[3];
    OALHashKeySize_t key_sizes[3];
    provider_binding_table_entry_t * best = ( preferred ) ? pbte : NULL;

    PCR_Assert( pbte );
    PCR_Assert( pbte->provider_op );

    /* Remove the alias, since it must always point to the primary provider, which may be changing. */
    alias_remove( local_alias_table, pbte->alias );
    /* Traverse over all matching bindings and figure out the best one to be the primary provider. */
    _make_keys( pbte->provider_op->securityScope, pbte->provider_op,
                pbte->provider_op->sources->channel,                                     /* channel not used. */
                keys, key_sizes );
    /* First, search for a local provider.  Keep existing primary provider, if there is one. */
    OALHashtable_Traverse( local_binding_tables[BINDING_TABLE], keys, key_sizes,
                           pbte_figure_primary_local, &best );
    /* Next, search for a remote provider.  It would be nice to skip this, but we need to make sure we clear ALL remote pbte primary flags. */
    OALHashtable_Traverse( remote_binding_tables[BINDING_TABLE], keys, key_sizes,
                           pbte_figure_primary_remote, &best );
    if ( best )
    {
        boolean ret;

        if( best != pbte )
        {
            if(pbte->interested)
            {
                /* This should not have to merge lists but does if needed */
                OALRouter_MergeInterested(&best->interested, &pbte->interested);
            }
            pbte->flags.bits.primary = FALSE; /* Guarantee */
        }

        best->flags.bits.primary = TRUE;
        /* Add the alias back in, so that it points to the primary provider. */
        ret = alias_insert( local_alias_table, best->alias, best );

        if ( ret )
        {
            /* Send provide PDU to matching interest. */
            interest_notify_new_provider(best->provider_op);
        }
        return ret;
    }
    return FALSE;
}

/* this is called with the operation locked and with security compatibility confirmed */
void router_notify_reparent(DOFOperation op, OALChannel old_channel, DOFAddress old_address)
{
    PCRMutex_Lock(router_mutex);

    if(op->operation_type==DOFOPERATIONTYPE_PROVIDE)
    {
        /* reparent the binding */
        provider_binding_table_entry_t *pbte = (provider_binding_table_entry_t*)op->pbte;

        if(!pbte)
        {
            DOFOperation_Internal_Cancel(op); /* what can we do with it? */
        }
        else
        {
            DOFAddress addrcopy = DOFAddress_Copy(op->sources->address);

            if ( op->sources->address && !addrcopy )
            {
                /* can't copy address, so drop it */
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "OAPRouter: Failed to copy address for reparenting operation.");
                DOFOperation_Internal_Cancel(op);
            }
            else
            {
                OALChannel ch = OALChannel_Retain(op->sources->channel); /* increment for pbte */

                OALChannel_Release(pbte->provider_channel);
                pbte->provider_channel = ch;
                DOFAddress_Destroy(pbte->address);
                pbte->address = addrcopy;

                DOFRouter_EstablishPrimaryProvider( pbte, router_forwards(), TRUE );

                /* notify system that routing info changed */
                DOFRuntime_RoutingInfoChanged(pbte, DOFROUTEINFOSTATE_CHANGED);
            }
            /* provider_binding_table_entry_release(pbte); - Removed since we never retain this pbte here */
        }
    }
    else if((op->operation_type == DOFOPERATIONTYPE_SUBSCRIBE) || (op->operation_type == DOFOPERATIONTYPE_REGISTER))
    {
        router_subscription_reparent(op,old_channel,old_address);
    }

    PCRMutex_Unlock(router_mutex);
}

boolean router_forwards(void)
{
    return DOFConfig_IsRouter( DOF_GetConfig( DOF_GetDof() ) );
}

static boolean Interested_CompareGreaterThan_fn_t(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    interested_requester_t* iKey = (interested_requester_t*)key;
    interested_requester_t* iItem = (interested_requester_t*)item;

    int32 ret = OALChannel_Compare(iKey->channel, iItem->channel);
    if(ret == 0)
    {
        ret = DOFAddress_Compare(iKey->address, iItem->address);
        if(ret == 0)
        {
            ret = DOFSecurityScope_Compare(iKey->scope, iItem->scope);
        }
    }
    return ret > 0;
}

static boolean Interested_CompareEqual_fn_t(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    interested_requester_t* iKey = (interested_requester_t*)key;
    interested_requester_t* iItem = (interested_requester_t*)item;

    int32 ret = OALChannel_Compare(iKey->channel, iItem->channel);
    if(ret == 0)
    {
        ret = DOFAddress_Compare(iKey->address, iItem->address);
        if(ret == 0)
        {
            ret = DOFSecurityScope_Compare(iKey->scope, iItem->scope);
        }
    }
    return (ret == 0);
}

static boolean OALRouter_AddInterestedOp_Internal(OALAVLTree opTree, DOFOperation op)
{
    OALAVLTree_InsertionStatus status = OALAVLTree_Insert(opTree, op, op, NULL);
    switch (status)
    {
    case OALAVLTree_InsertSuccess:
        return TRUE;
    case OALAVLTree_AlreadyInserted:
        return TRUE;
    case OALAVLTree_InsertFailure:
        DOF_LogMessage(DOF_GetDof(), DOFLOGLEVEL_WARNING, "OALRouter: Interested op failed to be added due to insert failure in data structure.");
    default:
        break;
    }
return FALSE;
}

static boolean OALRouter_AddInterested_Internal(provider_binding_table_entry_t* pbte, DOFOperation op)
{
    if( pbte )
    {
        if(!pbte->interested)
        {
            pbte->interested = OALAVLTree_Create( Interested_CompareGreaterThan_fn_t, Interested_CompareEqual_fn_t, NULL);
            if(pbte->interested)
            {
                interested_requester_t* interested = (interested_requester_t*)DOFMem_Alloc(sizeof(interested_requester_t));
                if(interested)
                {
                    interested->channel = OALChannel_Retain( op->sources->channel );
                    if(interested->channel || !op->sources->channel)
                    {
                        interested->address = DOFAddress_Copy(op->sources->address);
                        if(interested->address || !op->sources->address)
                        {
                            interested->scope = DOFSecurityScope_Copy(op->securityScope);
                            if(interested->scope || !op->securityScope)
                            {
                                interested->opTree = OALAVLTree_Create(NULL, NULL, NULL);
                                if (interested->opTree)
                                {
                                    if (OALRouter_AddInterestedOp_Internal(interested->opTree, op))
                                    {
                                        OALAVLTree_InsertionStatus inst_status = OALAVLTree_Insert(pbte->interested, interested, interested, NULL);

                                        switch(inst_status)
                                        {
                                        case OALAVLTree_InsertSuccess:
                                            return TRUE;
                                        case OALAVLTree_AlreadyInserted:
                                            DOF_LogMessage(DOF_GetDof(), DOFLOGLEVEL_WARNING, "OALRouter: Interested failed to be added due to already inserted in data structure.");
                                            break;
                                        case OALAVLTree_InsertFailure:
                                            DOF_LogMessage(DOF_GetDof(), DOFLOGLEVEL_WARNING, "OALRouter: Interested failed to be added due to insert failure in data structure.");
                                        default:
                                            break;
                                        }
                                    }
                                    OALAVLTree_Destroy(interested->opTree, NULL, NULL);
                                }
                                DOFSecurityScope_Destroy(interested->scope);
                            }
                            DOFAddress_Destroy(interested->address);
                        }
                        OALChannel_Release(interested->channel);
                    }
                    DOFMem_Free(interested);
                }
                OALAVLTree_Destroy(pbte->interested, NULL, NULL);
                pbte->interested = NULL;
            }
            DOF_LogMessage(DOF_GetDof(), DOFLOGLEVEL_WARNING, "OALRouter: Interested failed to be added due an error.");
            return FALSE;
        }
        else
        {
            interested_requester_t keys;
            interested_requester_t* item;

            keys.channel = op->sources->channel;
            keys.address = op->sources->address;
            keys.scope = op->securityScope;
            item = (interested_requester_t*)OALAVLTree_Retrieve(pbte->interested, &keys);

            if (item)
            {
                if (OALRouter_AddInterestedOp_Internal(item->opTree, op))
                {
                    return TRUE;
                }
                else
                {
                    DOF_LogMessage(DOF_GetDof(), DOFLOGLEVEL_WARNING, "OALRouter: found interested tree, but insert failed.");
                    return FALSE;
                }
            }
            else
            {
                /* Allocate item and add to tree. */
                item = (interested_requester_t*)DOFMem_Alloc(sizeof(interested_requester_t));
                if(item)
                {
                    item->channel = OALChannel_Retain( op->sources->channel );
                    if(item->channel || !op->sources->channel)
                    {
                        item->address = DOFAddress_Copy(op->sources->address);
                        if(item->address || !op->sources->address)
                        {
                            item->scope = DOFSecurityScope_Copy(op->securityScope);
                            if(item->scope || !op->securityScope)
                            {
                                item->opTree = OALAVLTree_Create(NULL, NULL, NULL);
                                if (item->opTree)
                                {
                                    if (OALRouter_AddInterestedOp_Internal(item->opTree, op))
                                    {
                                        OALAVLTree_InsertionStatus inst_status = OALAVLTree_Insert(pbte->interested, item, item, NULL);

                                        switch(inst_status)
                                        {
                                        case OALAVLTree_InsertSuccess:

                                            return TRUE;

                                        case OALAVLTree_AlreadyInserted:
                                            DOF_LogMessage(DOF_GetDof(), DOFLOGLEVEL_WARNING, "OALRouter: Interested failed to be added due to already inserted in data structure.");
                                            break;
                                        case OALAVLTree_InsertFailure:
                                            DOF_LogMessage(DOF_GetDof(), DOFLOGLEVEL_WARNING, "OALRouter: Interested failed to be added due to insert failure in data structure.");
                                        default:
                                            break;
                                        }
                                    }
                                    OALAVLTree_Destroy(item->opTree, NULL, NULL);
                                }
                                DOFSecurityScope_Destroy(item->scope);
                            }
                            DOFAddress_Destroy(item->address);
                        }
                        OALChannel_Release(item->channel);
                    }
                    DOFMem_Free(item);
                }
            }
        }
    }
    return FALSE;
}

boolean OALRouter_AddInterested_Operation(DOFOperation prov_op, DOFOperation interest_op)
{
    boolean ret = FALSE;
    provider_binding_table_entry_t* pbte;
    DOFRouter_Lock();
    pbte = DOFRouter_GetBindingEntry(prov_op->securityScope, prov_op->interface_id, prov_op->object_id.object_id);
    if(pbte)
    {
        ret = OALRouter_AddInterested_Internal(pbte, interest_op);
        provider_binding_table_entry_release(pbte);
    }
    else
        DOF_LogMessage(DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALRouter: Interested failed to be added due to the provider being removed.");
    DOFRouter_Unlock();
    return ret;
}

void OALRouter_RemoveInterested_Operation(DOFOperation prov_op, DOFOperation interest_op)
{
    provider_binding_table_entry_t* pbte;

    DOFRouter_Lock();
    pbte = DOFRouter_GetBindingEntry(prov_op->securityScope, prov_op->interface_id, prov_op->object_id.object_id);
    if(pbte)
    {
        if(pbte->interested)
        {
            interested_requester_t intreq = {interest_op->sources->channel, interest_op->sources->address, interest_op->securityScope, NULL};

            interested_requester_t* intObj = (interested_requester_t*)OALAVLTree_Retrieve(pbte->interested, &intreq);
            if( intObj )
            {
                uint32 itemCount = 0;
                if (intObj->opTree)
                {
                    OALAVLTree_Delete(intObj->opTree, interest_op);
                    itemCount = OALAVLTree_GetNumItems(intObj->opTree);
                }

                if (itemCount == 0)
                {
                    OALAVLTree_Delete(pbte->interested, intObj);
                    OALAVLTree_Destroy(intObj->opTree, NULL, NULL);
                    OALChannel_Release(intObj->channel);
                    DOFAddress_Destroy(intObj->address);
                    DOFSecurityScope_Destroy(intObj->scope);
                    DOFMem_Free(intObj);
                }
            }
        }
        provider_binding_table_entry_release(pbte);
    }
    else
        DOF_LogMessage(DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OALRouter: Interested failed to be added due to the provider being removed.");
    DOFRouter_Unlock();
}

static void Interested_Traversal_SendCancels(OALAVLTreeItem_t item, void *arg)
{
    interested_requester_t* node = (interested_requester_t*)item;
    provider_binding_table_entry_t* pbte = (provider_binding_table_entry_t*)arg;

    /* Send cancel */
    if( pbte->provider_op )
    {   
        DOFAddress oldAddress;
         
        DOFOperation_Lock(pbte->provider_op);
        
        DOFOperation_SetTimeout( pbte->provider_op, 0 );
        if ( node->address ) 
        {
            oldAddress = pbte->provider_op->sources->address;
            pbte->provider_op->sources->address = node->address;
        }              

        node->channel->routeToRequestor(node->channel, pbte->provider_op, 0);

        if(node->address)
            pbte->provider_op->sources->address = oldAddress;

        DOFOperation_Unlock(pbte->provider_op);
    }

    /* Free node */
    OALAVLTree_Destroy(node->opTree, NULL, NULL);
    OALChannel_Release(node->channel);
    DOFAddress_Destroy(node->address);
    DOFSecurityScope_Destroy(node->scope);
    DOFMem_Free((void*)item);
}

void OALRouter_SendInterestedCancels( provider_binding_table_entry_t *pbte )
{
    if(pbte->interested) 
    {
        OALAVLTree_Destroy(pbte->interested, Interested_Traversal_SendCancels, pbte);
        pbte->interested = NULL;
    }
}

static boolean Interested_Traversal_SendRetries(OALAVLTreeItem_t item, void *arg)
{
    interested_requester_t* node = (interested_requester_t*)item;
    provider_binding_table_entry_t* pbte = (provider_binding_table_entry_t*)arg;

    /* Send cancel */
    if( pbte->provider_op )
    {   
        DOFAddress oldAddress;
        
        DOFOperation_Lock(pbte->provider_op);
        if ( node->address ) 
        {
            oldAddress = pbte->provider_op->sources->address;
            pbte->provider_op->sources->address = node->address;
        }              

        node->channel->routeToRequestor(node->channel, pbte->provider_op, 0);

        if(node->address)
            pbte->provider_op->sources->address = oldAddress;

        DOFOperation_Unlock(pbte->provider_op);
    }
    return FALSE;
}

void OALRouter_SendInterestedUpdate( provider_binding_table_entry_t *pbte )
{
    if(pbte->interested) 
    {
        OALAVLTree_Traverse(pbte->interested, Interested_Traversal_SendRetries, pbte);
    }
}

/**
@}
**/
