/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/dof.h>
#include <dof/error.h>
#include <dof/exception.h>
#include <dof/interfaceid.h>
#include <dof/object.h>
#include <dof/operation.h>
#include <dof/pcr.h>
#include <internal/channel.h>
#include <internal/connection_internal.h>
#include <internal/dofinternal.h>
#include <internal/epp_negative_one.h>
#include <internal/exception.h>
#include <internal/hashtable.h>
#include <internal/interfaceidinternal.h>
#include <internal/oap_v1.h>
#include <internal/observable.h>
#include <internal/operationconfiginternal.h>
#include <internal/operationinternal.h>
#include <internal/remotechannelmanager.h>
#include <internal/requestinternal.h>
#include <internal/resultinternal.h>
#include <internal/router.h>
#include <internal/runtimeinternal.h>
#include <internal/securityscope.h>
#include <internal/slist.h>
#include <internal/systeminternal.h>

#include "r.h"
#include "router_fns.h"
#include "interest.h"
#include "advertise.h"


#define BINDING_TABLE 0
#define BINDING_TABLE_INTERFACE 1
#define BINDING_TABLE_OBJECT 2

/** @internal
 The routing functions.
 @{
 **/

OALHashtable local_binding_tables[3];

OALHashtable remote_binding_tables[3];


static OALHashtable session_table;

typedef struct session_table_entry_t
{
    DOFOperation reqOp;
    OALChannel   provChannel;
    DOFAddress   provAddress;
    OALAlias     baseAlias;
} session_table_entry_t;

OALHashtable local_alias_table;

static OALAlias alias = 0;

typedef struct traverse_op_and_result_t {
    DOFOperation op;
    boolean found;
    boolean local;
} traverse_op_and_result_t;


static session_table_entry_t * getSessionTableEntry( DOFObjectID objectID );


/** @internal
 Initialize the fns system.
 **/
boolean fns_init(void)
{
    local_binding_tables[BINDING_TABLE] = OALHashtable_Create(3);
    local_binding_tables[BINDING_TABLE_INTERFACE] = OALHashtable_Create(2);
    local_binding_tables[BINDING_TABLE_OBJECT] = OALHashtable_Create(2);

    remote_binding_tables[BINDING_TABLE] = OALHashtable_Create(3);
    remote_binding_tables[BINDING_TABLE_INTERFACE] = OALHashtable_Create(2);
    remote_binding_tables[BINDING_TABLE_OBJECT] = OALHashtable_Create(2);

    local_alias_table = OALHashtable_Create(1);
    session_table = OALHashtable_Create(1);

    if(local_binding_tables[BINDING_TABLE] && local_binding_tables[BINDING_TABLE_INTERFACE] &&
       local_binding_tables[BINDING_TABLE_OBJECT] && remote_binding_tables[BINDING_TABLE] &&
       remote_binding_tables[BINDING_TABLE_INTERFACE] && remote_binding_tables[BINDING_TABLE_OBJECT] &&
       local_alias_table && session_table)
    {
        return TRUE;
    }
    OALHashtable_Destroy(local_binding_tables[BINDING_TABLE], NULL, NULL);
    OALHashtable_Destroy(local_binding_tables[BINDING_TABLE_INTERFACE], NULL, NULL);
    OALHashtable_Destroy(local_binding_tables[BINDING_TABLE_OBJECT], NULL, NULL);
    OALHashtable_Destroy(remote_binding_tables[BINDING_TABLE], NULL, NULL);
    OALHashtable_Destroy(remote_binding_tables[BINDING_TABLE_INTERFACE], NULL, NULL);
    OALHashtable_Destroy(remote_binding_tables[BINDING_TABLE_OBJECT], NULL, NULL);
    OALHashtable_Destroy(local_alias_table, NULL, NULL);
    OALHashtable_Destroy(session_table, NULL, NULL);
    return FALSE;
}

static boolean _cleanup(OALHashItem_t pbte, void *unused)
{
    /* Called with router mutex held. */
    ((provider_binding_table_entry_t*)pbte)->provider_op->flags.bits.activation_closing = 1;
    provider_binding_table_entry_release(((provider_binding_table_entry_t*)pbte));

    return FALSE;
}

static OALAlias _get_next_alias(void)
{
    return ++alias;
}

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t removeSessionProvide(OALAVLMultiTreeItem_t pbte, void* entry, boolean new_keyset)
{
    /* Called with router mutex held. */
    if ( DOFObjectID_MatchesAtLeast( ((provider_binding_table_entry_t*)pbte)->provider_object_id, ((session_table_entry_t*)entry)->reqOp->sessionOID ) )
    {
        DOFOperation_Cancel( ((provider_binding_table_entry_t*)pbte)->provider_op );
    }
    return OALAVLMULTITREE_NORMAL;
}

static boolean _cleanup_session_table(OALHashItem_t entry, void *unused)
{
    /* Called with router mutex held. */
    OALHashtable_Traverse_All( local_binding_tables[BINDING_TABLE], removeSessionProvide, ((session_table_entry_t*)entry) );
    OALHashtable_Traverse_All( remote_binding_tables[BINDING_TABLE], removeSessionProvide, ((session_table_entry_t*)entry) );
    DOFOperation_Internal_Destroy( ((session_table_entry_t*)entry)->reqOp );
    OALChannel_Release( ((session_table_entry_t*)entry)->provChannel );
    DOFAddress_Destroy( ((session_table_entry_t*)entry)->provAddress );
    DOFMem_Free( ((session_table_entry_t*)entry) );
    return FALSE;
}

/** @internal
 Clean up the fns system.
 **/
void fns_cleanup(void)
{
    /* Called with router mutex held. */
    OALHashtable_Destroy(local_alias_table, 0, 0);

    OALHashtable_Destroy(session_table, _cleanup_session_table, 0);

    OALHashtable_Destroy(local_binding_tables[BINDING_TABLE], _cleanup, 0);
    OALHashtable_Destroy(local_binding_tables[BINDING_TABLE_INTERFACE], 0, 0);
    OALHashtable_Destroy(local_binding_tables[BINDING_TABLE_OBJECT], 0, 0);

    OALHashtable_Destroy(remote_binding_tables[BINDING_TABLE], _cleanup, 0);
    OALHashtable_Destroy(remote_binding_tables[BINDING_TABLE_INTERFACE], 0, 0);
    OALHashtable_Destroy(remote_binding_tables[BINDING_TABLE_OBJECT], 0, 0);
}

provider_binding_table_entry_t *DOFRouter_GetBindingEntry(DOFSecurityScope securityScope, DOFInterfaceID iid, DOFObjectID oid)
{
    provider_binding_table_entry_t *pbte;

    PCRMutex_Lock(router_mutex);
    pbte = DOFRouter_GetBindingEntry_Internal( securityScope, iid, oid );
    if ( pbte )
    {
        /* We found one, but make sure it's the primary. */
        if ( !pbte->flags.bits.primary )
            pbte = alias_retrieve( local_alias_table, pbte->alias );
        if ( pbte )
            provider_binding_table_entry_retain( pbte );
    }
    PCRMutex_Unlock(router_mutex);
    return pbte;
}

OALAlias DOFRouter_GetBindingAlias(DOFOperation op)
{
    provider_binding_table_entry_t *pbte;
    OALAlias alias = 0;

    PCRMutex_Lock( router_mutex );
    if ( op->pbte )
    {
        pbte = (provider_binding_table_entry_t *)op->pbte;
    }
    else
    {
        pbte = binding_exists( local_binding_tables[BINDING_TABLE], op );
        if ( !pbte )
            pbte = binding_exists( remote_binding_tables[BINDING_TABLE], op );
    }
    if ( pbte )
    {
        alias = pbte->alias;
    }
    PCRMutex_Unlock( router_mutex );
    return alias;
}

/**
 Given an alias, return a provider_binding_table_entry_t (pbte) for that alias.

 @param securityScope The security scope.
 @param address The address associated with the alias.
 @param alias The alias itself.

 @return A valid pbte for that alias or NULL if none was found.
 **/
provider_binding_table_entry_t *DOFRouter_Alias_to_pbte(DOFSecurityScope securityScope, DOFAddress address, OALAlias alias)
{
    provider_binding_table_entry_t *pbte = NULL;

    PCRMutex_Lock(router_mutex);

    if(address)
    {
        /* We do not handle reverse alias lookups for subscribe/register.  This is OK, since we never send subscribe/register with an alias. */
    }
    else
    {
        pbte = alias_retrieve(local_alias_table, alias);
    }

    if ( pbte && (!DOFSecurityScope_IsCompatible(securityScope, pbte->provider_op->securityScope)) )
        pbte = NULL;
    if ( pbte )
        provider_binding_table_entry_retain(pbte);

    PCRMutex_Unlock(router_mutex);

    return pbte;
}

static boolean _provider_route_normal_send_operation(provider_binding_table_entry_t *pbte, DOFOperation op, boolean local, boolean wildcard, boolean noProvideOp);

static int _timerDelayFunc(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown)
{
    provider_binding_table_entry_t* pbte = (provider_binding_table_entry_t *)argv[0];
    DOFOperation op = (DOFOperation)argv[1];

    PCR_Assert( argc == 2 );
    PCR_Assert( pbte );
    PCR_Assert( op );

    if ( !shutdown )
    {
        DOFOperation_Lock( op );
        PCRMutex_Lock(router_mutex);
        _provider_route_normal_send_operation( pbte, op, FALSE /* Pretend non-local to avoid recursive delay */, FALSE /* At this point, we treat it as a non-wildcard operation */, FALSE /* Normal */ );
        PCRMutex_Unlock(router_mutex);
        DOFOperation_Unlock( op );
    }
    DOFOperation_Internal_Destroy( op );
    provider_binding_table_entry_release( pbte );

    return 0;
}

typedef struct routerSidQueryContext_t
{
    DOFOperation    op;
    DOFOperation    providerOp;
    OALChannel      providerChannel;
    DOFAddress      address;
}routerSidQueryContext_t;

static void _sendOpAfterQuery_fn(void * context, boolean success);
static const struct EPPNegOneQueryCallbackFns_t EPPNegOneQueryCallbackFns_SendOp = { _sendOpAfterQuery_fn };
static struct EPPNegOneQueryCallback_t EPPNegOneQueryCB_SendOp = { &EPPNegOneQueryCallbackFns_SendOp };

static void _sendOpAfterQuery_fn(void * context, boolean success)
{
    routerSidQueryContext_t * ctx = (routerSidQueryContext_t *)context;

    PCR_Assert(ctx);
    if (success)
    {
        ctx->providerChannel->routeToProvider(ctx->providerChannel, ctx->op, ctx->providerOp, ctx->address);
    }
    DOFOperation_Internal_Destroy(ctx->op);
    DOFOperation_Internal_Destroy(ctx->providerOp);
    OALChannel_Release(ctx->providerChannel);
    DOFAddress_Destroy(ctx->address);
    DOFMem_Free(ctx);
}

/* returns TRUE if routed */
static boolean _provider_route_normal_send_operation(provider_binding_table_entry_t *pbte, DOFOperation op, boolean local, boolean wildcard, boolean noProvideOp)
{
    /* Called with op mutex held. */
    if ( local )
    {
        if (!op->sources->request)
        {
            op->sources->request = DOFRequest_Create((DOFSystem)pbte->provider_channel, op, NULL);
            if ( !op->sources->request )
                return FALSE; /* error occurred */
        }

        /* has this operation been acked? if so, don't route it */
        if ( !op->flags.bits.cancelled &&
             op->ackTree &&
             OALAVLTree_Retrieve(op->ackTree, pbte->provider_object_id) )
            return TRUE; /* Do not execute, but it was "successful". */

        if ( op->operationConfig && !op->flags.bits.cancelled )
        {
            if ( (PCRRandom_Get() & 0x7FFF) > DOFOperationControl_GetHeuristic( op->operationConfig ) )
                return TRUE; /* Do not execute, but it was "successful". */
            if ( DOFOperationControl_IsDelay( op->operationConfig ) )
            {
                uint32 delay = (PCRRandom_Get() % op->sources->oal_timeout);

                if(delay > 0)
                {
                    void *argv[2];

                    argv[0] = pbte;
                    argv[1] = op;
                    if ( OALTimer_Insert( OAP_V1_GetTimer(), delay, _timerDelayFunc, 2, argv, FALSE) )
                    {
                        provider_binding_table_entry_retain( pbte );
                        DOFOperation_Retain( op );
                        return TRUE;
                    }
                    /* If the insertion fails, just execute the operation now. */
                }
            }
        }
    }

    if ( ( op->operation_type == DOFOPERATIONTYPE_REGISTER || op->operation_type == DOFOPERATIONTYPE_SUBSCRIBE ) && !op->flags.bits.cancelled )
    {
        /* Setup subscription state information. */
        subscription_add( op, pbte );
    }

    if ( !local && wildcard )
        return FALSE; /* Forwarding, so don't process here if it's a broadcast oid. */

    /* If a channel/address is specified with a request config, we need to trap that here. */
    /* This is mainly for register/subscribe/session, since they can only be sent to a known provider anyway. */
    if(!op->flags.bits.internally_created && op->operationConfig)
    {
        DOFConnection conn = DOFOperationControl_GetTargetConnection(op->operationConfig);
        if(conn)
        {
            OALChannel channel = DOFConnection_GetChannel(conn);
            if(!OALChannel_IsEquivalent( channel, pbte->provider_channel ))
                return FALSE;
        }
    }

    if(DOFSecurityScope_IsCompatible(OALChannel_GetSecurityScope(pbte->provider_channel, FALSE), op->securityScope))
    {
        DOFOperation sidQueryOp = op;

        if(op->flags.bits.internally_created && pbte->provider_op->flags.bits.internally_created)
            /* If we have an internally created op and internally created provide operation, we're forwarding */
        {
            if(!OALOperationManager_SetupForwardingOperation(op))
            {
                return FALSE;
            }
        }
        if (!sidQueryOp->opID || DOFOperationID_CanBeExplicit(sidQueryOp->opID))
        {
            pbte->provider_channel->routeToProvider(pbte->provider_channel, op, (noProvideOp)? NULL : pbte->provider_op, pbte->provider_op->sources->address);
        }
        else
        {
            routerSidQueryContext_t * ctx;

            ctx = (routerSidQueryContext_t *)DOFMem_Alloc(sizeof(routerSidQueryContext_t));

            if (ctx)
            {
                ctx->address = DOFAddress_Copy(pbte->provider_op->sources->address);
                if (ctx->address || !pbte->provider_op->sources->address)
                {
                    DOFOperation_Retain(op);
                    ctx->op = op;
                    if (noProvideOp)
                    {
                        ctx->providerOp = NULL;
                    }
                    else
                    {
                        DOFOperation_Retain(pbte->provider_op);
                        ctx->providerOp = pbte->provider_op;
                    }
                    ctx->providerChannel = OALChannel_Retain(pbte->provider_channel);

                    if (!EPP_NegOne_Query(sidQueryOp->sources->channel, sidQueryOp->sources->address, 8000, 3000, ctx, &EPPNegOneQueryCB_SendOp))
                    {
                        DOFOperation_Internal_Destroy(ctx->op);
                        DOFOperation_Internal_Destroy(ctx->providerOp);
                        OALChannel_Release(ctx->providerChannel);
                        DOFAddress_Destroy(ctx->address);
                        DOFMem_Free(ctx);
                        return FALSE;
                    }
                }
                else
                {
                    DOFMem_Free(ctx);
                    return FALSE;
                }
            }
            else
                return FALSE;
        }
        return TRUE;
    }

    return FALSE;
}

static boolean __provider_route_normal_wildcard(OALHashItem_t pbte, void* t)
{
    /* Called with op mutex held. */
    if ( !((provider_binding_table_entry_t*)pbte)->provider_op->flags.bits.lost && DOFObjectID_MatchesAtLeast_Mcast(((provider_binding_table_entry_t*)pbte)->provider_object_id, ((traverse_op_and_result_t*)t)->op->object_id.object_id) )
    {
        if ( _provider_route_normal_send_operation(((provider_binding_table_entry_t*)pbte), ((traverse_op_and_result_t*)t)->op, ((traverse_op_and_result_t*)t)->local, TRUE, FALSE) )
            ((traverse_op_and_result_t*)t)->found = TRUE;
    }

    return FALSE;
}

static boolean __provider_route_normal(OALHashItem_t pbte, void* t)
{
    /* Called with op mutex held. */
    if(!((provider_binding_table_entry_t*)pbte)->provider_op->flags.bits.lost&&((provider_binding_table_entry_t*)pbte)->flags.bits.primary)
    {
        if(_provider_route_normal_send_operation(((provider_binding_table_entry_t*)pbte), ((traverse_op_and_result_t*)t)->op, ((traverse_op_and_result_t*)t)->local, FALSE, FALSE))
            ((traverse_op_and_result_t*)t)->found = TRUE;

        if(!((traverse_op_and_result_t*)t)->op->sources->channel)
        {
            ((traverse_op_and_result_t*)t)->op->sources->channel = OALChannel_Retain(((provider_binding_table_entry_t*)pbte)->provider_channel);
        }

        return TRUE; /* not wildcard, so take the first one */
    }

    return FALSE; /* try again */
}

DOFRouterStatusCode_t provider_route_normal_intercept_request_config(DOFOperation op, OALChannel *rc)
{
    /* before we do anything, we need to see if the user wanted to override the routing info with DOFOperationControl */
    /* This logic is bypassed for subscribe, register, and session, since it *must* be sent to the provider. */
    if(!op->flags.bits.internally_created && op->operationConfig && op->operation_type!=DOFOPERATIONTYPE_SUBSCRIBE && op->operation_type!=DOFOPERATIONTYPE_REGISTER && op->operation_type!=DOFOPERATIONTYPE_SESSION)
    {
        DOFConnection conn = DOFOperationControl_GetTargetConnection(op->operationConfig);

        if(conn)
        {
            *rc = DOFConnection_GetChannel(conn);

            /* check to make sure the requestor and provider channel are security ID compatible */
            if(*rc)
            {
                if(DOFSecurityScope_IsCompatible(op->securityScope, OALChannel_GetSecurityScope(*rc, FALSE)))
                {
                    /* is the channel a remote channel? if so, then send it */
                    if((*rc)->remoteChannelData)
                    {
                        (*rc)->routeToProvider(*rc, op, NULL, NULL);
                        return DOFROUTERSTATUS_SUCCESS;
                    }
                }
                else
                {
                    DOFResult res = DOFResult_Exception_Create(NULL, NULL, DOFEXCEPTION_ACCESS_DENIED, NULL);
                    /* If result did not create, this next call does nothing. Would be useless to make a memory exception.*/
                    DOFSystem_DeliverResponseBackToRequestor( op, res );                /* Destroys result for us. */
                }

                return DOFROUTERSTATUS_SUCCESS;
            }

            return DOFROUTERSTATUS_ERROR; /* do not send if channel is null- stop the caller from sending also. */
        }
    }

    return DOFROUTERSTATUS_NO_REQUEST_CONFIG;                                           /* caller will send op. */
}

static void _queryCallback_SendToAll_fn(void * context, boolean success);
static const struct EPPNegOneQueryCallbackFns_t EPPNegOneQueryCallbackFns = { _queryCallback_SendToAll_fn };
static struct EPPNegOneQueryCallback_t EPPNegOneQueryCB = { &EPPNegOneQueryCallbackFns };

static void _queryCallback_SendToAll_fn(void * context, boolean success)
{
    DOFOperation op = (DOFOperation) context;
    if (success)
        OALRemoteChannelMgr_SendToAll(op);

    DOFOperation_Internal_Destroy(op);
}

static DOFRouterStatusCode_t provider_route_normal(DOFOperation op)
{
    typedef boolean (*provider_route_fn)(provider_binding_table_entry_t *pbte, traverse_op_and_result_t *t);
    static const OALHashtable_Traversal_fn_t fns[] = { __provider_route_normal, __provider_route_normal_wildcard };

    provider_binding_table_entry_t *pbte;
    const void *keys[3];
    uint16 key_sizes[3];
    OALChannel rc = 0;

    traverse_op_and_result_t t;
    t.found = FALSE;
    t.op = op;

    /* Called with op mutex held. */

    /* if it's internally created, only check local bindings. if the user made it, then check local
     and remote bindings */

    if(provider_route_normal_intercept_request_config(op, &rc)!=DOFROUTERSTATUS_NO_REQUEST_CONFIG)
        return DOFROUTERSTATUS_SUCCESS; /* op was properly sent (or dropped) within above call. */

    /* request coming in with an alias -- locate the primary provider (always in the alias table) and send it. */
    if(op->sources->alias)
    {
        /* op has an alias */
        pbte = alias_retrieve(local_alias_table, op->sources->alias);
        if(pbte) {
            if(_provider_route_normal_send_operation(pbte, op, TRUE, FALSE, FALSE))
                return DOFROUTERSTATUS_SUCCESS;
        }

        return DOFROUTERSTATUS_INVALID_BINDING; /* invalid alias or failure to send to provider */
    }

    _make_keys( op->securityScope, op,
                op->sources->channel,                                                           /* channel not used. */
                keys, key_sizes );

    t.local = TRUE;

    /* check for locally provided ops */
    OALHashtable_Traverse(local_binding_tables[op->flags.bits.provider_oid_is_wildcard], keys, key_sizes,
                          fns[op->flags.bits.provider_oid_is_wildcard], &t);

    if ( op->flags.bits.provider_oid_is_wildcard )
    {
        goto send_to_all;
    }

    if ( !t.found )
    {
        /* check for remotely provided ops */
        if ( !op->flags.bits.internally_created || router_forwards() )
        {
            t.local = FALSE;
            /* request is going out, so it needs to refer to the remote binding tables */
            /* if the binding isn't found, then send it everywhere */

            /* keys already made above */
            /* _make_keys(op->securityScope, op, op->sources->channel, keys, key_sizes); */

            OALHashtable_Traverse(remote_binding_tables[op->flags.bits.provider_oid_is_wildcard], keys, key_sizes,
                                  fns[op->flags.bits.provider_oid_is_wildcard], &t);

            if ( !t.found && ( op->flags.bits.internally_created || DOFOperationControl_IsFloodAllowed( op->operationConfig ) ) )
            {
                /* binding not found, so send everywhere */
                goto send_to_all;
            }
            /* finished with all local bindings and remote bindings */
        }

        /* binding not found for request */
        if ( !t.found )
        {
            if ( !op->flags.bits.isFlooded && !op->flags.bits.provider_oid_is_wildcard )
            {
                /* if not, send an exception */
                DOFResult result = DOFResult_Exception_Create(0, 0, DOFEXCEPTION_NOT_FOUND, 0);
                /* If result did not create, this next call does nothing. Would be useless to make a memory exception.*/
                DOFSystem_DeliverResponseBackToRequestor( op, result );                  /* Destroys result for us. */
            }
            return DOFROUTERSTATUS_INVALID_BINDING;
        }
    }
    return DOFROUTERSTATUS_SUCCESS; /* local request found local binding.. return */

send_to_all:
    if ( !op->flags.bits.internally_created || router_forwards() )
    {
        /* Session-based operations cannot flood, but can be directed to known session provider. */
        if ( DOFObjectID_HasAttribute( op->object_id.object_id,  DOFOBJECTIDATTRIBUTE_SESSION ) )
        {
            session_table_entry_t * entry = getSessionTableEntry( op->object_id.object_id );
            if ( entry )
            {
                pbte = alias_retrieve( local_alias_table, entry->baseAlias );
                if ( pbte )
                {
                    if(_provider_route_normal_send_operation(pbte, op, FALSE, FALSE, TRUE))
                        return DOFROUTERSTATUS_SUCCESS;
                }
            }
            return DOFROUTERSTATUS_INVALID_BINDING;
        }
        else if ( op->operation_type == DOFOPERATIONTYPE_REGISTER || op->operation_type == DOFOPERATIONTYPE_SUBSCRIBE )
        {
            if(!op->flags.bits.cancelled)
            {
                if ( op->flags.bits.provider_oid_is_wildcard )
                {
                    /* Scan the remote tables (skipped previously) and setup the subscription information. */
                    t.local = FALSE;

                    /* keys already made above */
                    /* _make_keys(op->securityScope, op, op->sources->channel, keys, key_sizes); */

                    OALHashtable_Traverse(remote_binding_tables[op->flags.bits.provider_oid_is_wildcard], keys, key_sizes,
                                          __provider_route_normal_wildcard, &t);
                    /* The above only sets up the routing information for responses, we still need to flood everywhere (below). */
                }
                else
                {
                    /* Cannot flood a subscribe/register; if it wasn't found, then no state is setup to accept responses, thus it is useless to send it.*/
                    return DOFROUTERSTATUS_INVALID_BINDING;
                }
            }
        }
        if (DOFOperationID_CanBeExplicit(op->opID))
        {
            OALRemoteChannelMgr_SendToAll(op);
        }
        else
        {
            if (EPP_NegOne_Query(op->sources->channel, op->sources->address, 8000, 3000, op, &EPPNegOneQueryCB))
            {
                DOFOperation_Retain(op);
            }
            else
            {
                return DOFROUTERSTATUS_ERROR;
            }
        }
    }
    return DOFROUTERSTATUS_SUCCESS;
}

static DOFRouterStatusCode_t requestor_route_normal(DOFOperation op, DOFResult result)
{
    op->sources->channel->routeToRequestor(op->sources->channel, op, result);

    return DOFROUTERSTATUS_SUCCESS;
}

static DOFRouterStatusCode_t requestor_route_stateful(DOFOperation op, DOFResult result)
{
    /* For stateful, OAP unicast, remote operations, non-success result causes the operation to terminate. */
    /* Local operations must be processed first by the DOFSystem, and are then cancelled. */
    if ( !op->flags.bits.provider_oid_is_wildcard && op->flags.bits.internally_created && DOFResult_GetException(result) )
    {
        DOFOperation_Cancel( op );
    }
    return requestor_route_normal( op, result );
}

static DOFRouterStatusCode_t requestor_route_subscribe(DOFOperation op, DOFResult result)
{
    DOFRouterStatusCode_t rval = DOFROUTERSTATUS_SUCCESS;

    /* this is a subscribe accepted */
    /* Check that it came from the primary provider. */
    if ( subscription_accept( op, result ) )
    {
        /* Accept the response only from the primary provider. */
        rval = requestor_route_stateful(op, result);
    }

    return rval;
}

static void _provide_and_begin_provide_cleanup(provider_binding_table_entry_t *pbte, OALHashtable *tables)
{
    provider_binding_table_entry_t *pbte_found[2] = {NULL,NULL};

    /* this call always made with at least one extra ref count on pbte */

    PCRMutex_Lock(router_mutex);

    tables_remove_binding(tables, pbte);

    _cleanup(pbte, NULL);

    if(((pbte_found[0] = binding_exists(local_binding_tables[BINDING_TABLE], pbte->provider_op)) == NULL)
       && ((pbte_found[1] = binding_exists(remote_binding_tables[BINDING_TABLE], pbte->provider_op)) == NULL)) {
        /* no more binding, so tell dofsystem */
        DOFSystem_BindingChanged(OALChannel_GetSecurityScope(pbte->provider_channel, FALSE), pbte->provider_object_id, pbte->interface_id, 0);
    }
    if(pbte->flags.bits.primary) { /* removing primary, so make another primary */
        DOFRouter_EstablishPrimaryProvider( pbte, FALSE, FALSE );
    }

    DOFRouter_RemoveProviderSubscriptions( pbte );

    DOFRuntime_RoutingInfoChanged(pbte, DOFROUTEINFOSTATE_REMOVED);

    provider_binding_table_entry_release(pbte);

    PCRMutex_Unlock(router_mutex);
}

static void begin_provide_observer(OALObserver self, OALObservable observable, void* pbte)
{
    _provide_and_begin_provide_cleanup((provider_binding_table_entry_t*)pbte, local_binding_tables);
}

static void provide_observer(OALObserver self, OALObservable observable, void* pbte)
{
    _provide_and_begin_provide_cleanup((provider_binding_table_entry_t*)pbte, remote_binding_tables);
}

static provider_binding_table_entry_t *pbte_create(DOFOperation provider_op, OALObserver_fn_t observer_fn, OALAlias alias)
{
    provider_binding_table_entry_t *pbte = DOFMem_Alloc(sizeof(provider_binding_table_entry_t));

    if (!pbte)
        goto out_malloc;

    PCR_Assert(provider_op);

    pbte->ref_count = 2; /* one for create, and one for observer */

    pbte->alias = alias ? alias : _get_next_alias();
    if (provider_op->sources->address)
    {
        pbte->address = DOFAddress_Copy(provider_op->sources->address);
        if (!pbte->address)
            goto out_address;
    }
    else
        pbte->address = NULL;

    pbte->provider_channel = OALChannel_Retain( provider_op->sources->channel );
    pbte->provider_object = DOFObject_Copy(provider_op->provider_object);
    pbte->interface_id = provider_op->interface_id;
    pbte->provider_object_id = provider_op->object_id.object_id;
    pbte->provider_op = provider_op; /* retained in observer */
    pbte->timestamp = PCRTime_GetMS();
    pbte->provider_op->flags.bits.activation_closing = 0;
    pbte->interested = NULL;

    pbte->obs = DOFOperation_AddOpWillCloseObserverOtherArg(provider_op, observer_fn, pbte);
    if(pbte->obs == NULL)
    {
        goto out_observable;
    }

    PCRMem_Set( &pbte->flags, 0, sizeof(pbte->flags ) );
    provider_op->pbte = pbte;

    return pbte;

out_observable:
    DOFObject_Destroy(pbte->provider_object);
    OALChannel_Release(pbte->provider_channel);
    DOFAddress_Destroy(pbte->address);

out_address:
    DOFMem_Free(pbte);
out_malloc:

    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "OAPRouter: Failed to create provider entry.");
    return NULL;
}

static session_table_entry_t * getSessionTableEntry( DOFObjectID objectID )
{
    DOFObjectIDAttribute attributeSessionID = DOFObjectID_GetAttribute( objectID, DOFOBJECTIDATTRIBUTE_SESSION );
    OALHashKeySize_t size = attributeSessionID.dataSize; /* Note: .dataSize (uint8) cannot exceed OALHashKeySize_t (uint16). */
    const void * bytes = attributeSessionID.data;
    PCR_Assert( attributeSessionID.id == DOFOBJECTIDATTRIBUTE_SESSION );
    return (session_table_entry_t *)OALHashtable_Retrieve( session_table, &bytes, &size );
}

DOFOperation DOFRouter_LookupSession(DOFObjectID objectId)
{
    if(DOFObjectID_HasAttribute(objectId, DOFOBJECTIDATTRIBUTE_SESSION))
    {
        session_table_entry_t * entry = getSessionTableEntry(objectId);

        if(entry != NULL)
        {
            return entry->reqOp;
        }
    }
    return NULL;
}

boolean router_IsProvideCompatibleWithInterest( DOFOperation provOp, DOFOperation reqOp )
{
    if (DOFObjectID_HasAttribute( provOp->object_id.object_id, DOFOBJECTIDATTRIBUTE_SESSION ))
    {
        /* it belongs to a session, so find the session op */
        session_table_entry_t * entry = getSessionTableEntry( provOp->object_id.object_id );

        /* Ensure the request's channel and address match the session's source. */
        if ( entry && OALChannel_IsEquivalent( entry->reqOp->sources->channel, reqOp->sources->channel ) && DOFAddress_IsEqual( entry->reqOp->sources->address, reqOp->sources->address ) )
            return TRUE; /* Accept: session found, and channel and address match. */
        return FALSE; /* Reject: prevent this provide operation from reaching this requestor. */
    }
    return TRUE; /* Accept: no session. */
}

static boolean checkProvideCompatibleWithSession( DOFOperation provOp )
{
    /* see if provide has a session Attribute in its oid */
    if (DOFObjectID_HasAttribute( provOp->object_id.object_id, DOFOBJECTIDATTRIBUTE_SESSION ))
    {
        /* it belongs to a session, so find the session op */
        session_table_entry_t * entry = getSessionTableEntry( provOp->object_id.object_id );

        /* Ensure the provider's channel and address match the session provider's source. */
        if ( entry && OALChannel_IsEquivalent( provOp->sources->channel, entry->provChannel ) && DOFAddress_IsEqual( provOp->sources->address, entry->provAddress ) )
        {
            /* Accept: session found, and channel and address match. */
            return TRUE;
        }
        /* Reject: prevent this provide operation from being accepted. */
        if ( !provOp->flags.bits.internally_created )
            provOp->completeCallbackException = DOFException_Copy( DOFEXCEPTION_REJECTED );
        DOFOperation_Cancel( provOp );
        return FALSE;
    }
    return TRUE; /* Accept: no session. */
}

/* called by _route_to_provider(). */
static DOFRouterStatusCode_t route_begin_provide(DOFOperation op)
{
    /* I've been given a begin provide, so add it to my tables. */
    provider_binding_table_entry_t * pbte;
    OALAlias                         alias = 0;

    if ( !checkProvideCompatibleWithSession( op ) )
        return DOFROUTERSTATUS_ERROR;

    pbte = DOFRouter_GetBindingEntry_Internal( op->securityScope, op->interface_id, op->object_id.object_id );
    if ( pbte )
    {
        alias = pbte->alias;

        pbte = alias_retrieve( local_alias_table, alias );
        if ( pbte && pbte->provider_op == op )
        {
            /* Send provide PDU to matching interest for primary provide (keep it alive). */
            if( !op->flags.bits.cancelled )
            {
                if( !op->flags.bits.new_retry_count )
                {
                    interest_notify_new_provider(op);
                }
                else
                {
                    OALRouter_SendInterestedUpdate( pbte );
                }
            }
            else
                OALRouter_SendInterestedCancels(pbte);

            return DOFROUTERSTATUS_SUCCESS;
        }
        if (op->flags.bits.already_routed)
        {
            /* Nothing more to do -- tables are already updated. */
            return DOFROUTERSTATUS_SUCCESS;
        }
    }

     /* Ignore canceled provides */
    if ( op->flags.bits.cancelled )
        return DOFROUTERSTATUS_SUCCESS;

    pbte = pbte_create( op, begin_provide_observer, alias );
    if (!pbte)
        goto out_pbte;

    if (!tables_add_binding(local_binding_tables, pbte))
        goto out_tables;

    if ( !DOFRouter_EstablishPrimaryProvider( pbte, TRUE, FALSE ) )
        goto out_alias;

    if ( alias == 0 )
    {
        /* We created a new binding, so let the system know. */
        DOFSystem_BindingChanged( OALChannel_GetSecurityScope( pbte->provider_channel, FALSE ),
                                  pbte->provider_object_id, pbte->interface_id, 1 );
    }

    DOFRuntime_RoutingInfoChanged(pbte, DOFROUTEINFOSTATE_ADDED);

    op->flags.bits.already_routed = 1;

    return DOFROUTERSTATUS_SUCCESS;

out_alias:
    tables_remove_binding(local_binding_tables, pbte);

out_tables:
    DOFOperation_RemoveOpWillCloseObserver(op, pbte->obs);
    provider_binding_table_entry_release(pbte); /* sets ref count to 2 */
    provider_binding_table_entry_release(pbte);

out_pbte:
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "OAPRouter: Failed to add new local provide information to router.");
    return DOFROUTERSTATUS_ERROR;
}

/* Called only by _route_to_requestor. */
static DOFRouterStatusCode_t route_provide(DOFOperation op, DOFResult result)
{
    /* Called with op's lock held, op is a provide operation being passed back to the requestor. */
    provider_binding_table_entry_t *pbte;
    OALAlias alias = 0;

    /* Check validity of session provide, if necessary. */
    if ( !checkProvideCompatibleWithSession( op ) )
        return DOFROUTERSTATUS_ERROR;

    /* Ignore canceled provides */
    if ( op->flags.bits.cancelled )
        return DOFROUTERSTATUS_SUCCESS;

    pbte = DOFRouter_GetBindingEntry_Internal( op->securityScope, op->interface_id, op->object_id.object_id );
    if ( pbte )
    {
        alias = pbte->alias;

        pbte = alias_retrieve( local_alias_table, alias );
        if ( pbte && pbte->provider_op == op )
        {
            /* Send provide PDU to matching interest for primary provide (keep it alive). */
            interest_notify_new_provider(op);
            return DOFROUTERSTATUS_SUCCESS;
        }
        if (op->flags.bits.already_routed)
        {
            /* Nothing more to do -- tables are already updated. */
            return DOFROUTERSTATUS_SUCCESS;
        }
    }

    pbte = pbte_create( op, provide_observer, alias );
    /* provide_observer() called when op (a provide op) cancels- removes pbte from 2 binding tables and frees it. */

    if (!pbte)
        goto out_pbte;

    if(!tables_add_binding(remote_binding_tables, pbte))
        goto out_tables;
    /* pbte's provide op now added or was already present in remote_binding_tables. If forwarding node, pbte's
       provide op now added or was already present in local_binding_tables. */

    if ( !DOFRouter_EstablishPrimaryProvider( pbte, router_forwards(), TRUE ) )
        goto out_alias;

    if ( alias == 0 )
    {
        /* We added our newly created pbte to remote_binding_tables. */
        DOFSystem_BindingChanged( OALChannel_GetSecurityScope( pbte->provider_channel, FALSE ),
                                  pbte->provider_object_id, pbte->interface_id, 1 );
    }

    DOFRuntime_RoutingInfoChanged(pbte, DOFROUTEINFOSTATE_ADDED); /*Calls _traverse_routing_callbacks() EVERY TIME*/

    op->flags.bits.already_routed = 1;

    return DOFROUTERSTATUS_SUCCESS;

out_alias:
    tables_remove_binding(remote_binding_tables, pbte);

out_tables:
    DOFOperation_RemoveOpWillCloseObserver(op, pbte->obs);
    provider_binding_table_entry_release(pbte); /* sets ref count to 2 */
    provider_binding_table_entry_release(pbte);

out_pbte:
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "OAPRouter: Failed to add new remote provide information to router.");
    return DOFROUTERSTATUS_ERROR;
}

static DOFRouterStatusCode_t route_provider_death(DOFOperation op)
{
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "OAPRouter: Invalid provider operation route attempt.");
    return DOFROUTERSTATUS_SUCCESS;
}

struct _chanaddr
{
    OALChannel channel;
    DOFAddress address;
};

static boolean _advertiseTraversal(OALHashItem_t op, void* ca)
{
    ((struct _chanaddr*)ca)->channel->routeToProvider(((struct _chanaddr*)ca)->channel, ((DOFOperation)op), NULL, ((struct _chanaddr*)ca)->address);
    return FALSE;
}

static void _queryCallback_SendToAllStateless_fn(void * context, boolean success);
static const struct EPPNegOneQueryCallbackFns_t EPPNegOneQueryCallbackStatelessFns = { _queryCallback_SendToAllStateless_fn };
static struct EPPNegOneQueryCallback_t EPPNegOneQueryStatelessCB = { &EPPNegOneQueryCallbackStatelessFns };

static void _queryCallback_SendToAllStateless_fn(void * context, boolean success)
{
    if (success)
        OALRemoteChannelMgr_SendToAllStateless((DOFOperation)context);

    DOFOperation_Internal_Destroy((DOFOperation)context);
}

static DOFRouterStatusCode_t route_provider_request_config_or_flood(DOFOperation op)
{
    DOFRouter_AddAdvertise(op);
    if(!op->flags.bits.internally_created)
    {
        OALChannel rc = 0;

        if(provider_route_normal_intercept_request_config(op, &rc) != DOFROUTERSTATUS_NO_REQUEST_CONFIG)
            return DOFROUTERSTATUS_SUCCESS;   /* op was properly sent (or dropped) within above call. */
    }
    else
    {
        struct _chanaddr ca;
        ca.channel = op->sources->channel;
        ca.address = op->sources->address;
        DOFRouter_TraverseInterestTables(op, _advertiseTraversal, &ca);
        if (!router_forwards())
            return DOFROUTERSTATUS_SUCCESS;
    }

    if (DOFOperationID_CanBeExplicit(op->opID))
    {
        OALRemoteChannelMgr_SendToAllStateless(op);
    }
    else
    {
        if (EPP_NegOne_Query(op->sources->channel, op->sources->address, 8000, 3000, op, &EPPNegOneQueryStatelessCB))
        {
            DOFOperation_Retain(op);
            return DOFROUTERSTATUS_SUCCESS;
        }
        else
        {
            return DOFROUTERSTATUS_ERROR;
        }
    }

    return DOFROUTERSTATUS_SUCCESS;
}

static DOFRouterStatusCode_t route_requestor_death(DOFOperation op, DOFResult result)
{
    return DOFROUTERSTATUS_SUCCESS;
}


/** @internal
 Observer callback to cleanup session_table when a session operation closes.

 @param observable    Requestor observable for sessionOp (originally stored at sessionOp->operation_will_close).
 @param sessionOpenOp The specific requestor operation that is in the process of closing.
**/
static void CleanupSessionTableForOp( OALObserver self, OALObservable observable, void * sessionOpenOp )
{
    /* This matches typedef void (*OALObserver_fn_t)(OALObservable observable, void *arg); in internal/observable.h. */
    const void *            bytes;
    session_table_entry_t * entry;
    OALHashKeySize_t        size  = sizeof(GUID);
    DOFOperation            op = (DOFOperation)sessionOpenOp;

    PCR_Assert( observable );
    PCR_Assert( op );
    /* Note: _DOFOperation_Cancel() causes us to be called, and does so with op locked. */
    PCR_Assert( op->flags.bits.cancelled );
    bytes = op->sessionID.bytes;
    entry = (session_table_entry_t *)OALHashtable_Retrieve( session_table, &bytes, &size );
    if (entry && entry->reqOp == op && OALHashtable_Remove( session_table, entry, &bytes, &size ))
        _cleanup_session_table( entry, NULL );
    DOFOperation_Internal_Destroy( op ); /* DOFOperation_AddOpWillCloseObserver() had retained this. */
}


static DOFRouterStatusCode_t route_provider_session( DOFOperation op )
{
    /* Called with op and router mutexes held. op is an OAP Open request. */
    OALHashKeySize_t        size  = sizeof(GUID);
    const void *            bytes = op->sessionID.bytes;
    session_table_entry_t * entry = (session_table_entry_t *)OALHashtable_Retrieve( session_table, &bytes, &size );

    if (!op->flags.bits.cancelled)
    {
        /* Normal or retry: op->flags.bits.new_retry_count tells us which. Is this a new session? */
        if (!entry)
        {
            /* New session: record it, add our observer to the OAP open reqOp, then submit to DOFSystem processing. */
            OALObserver                      observer;
            provider_binding_table_entry_t * pbte = DOFRouter_GetBindingEntry( op->securityScope, op->interface_id,
                                                                               op->object_id.object_id );
                                             /* On non-NULL returned pbte, its ref count is bumped up. */

            if (!pbte)
            {
                /* Failed OAP open session request:  send DOFException back to requestor. */
                DOFResult result = DOFResult_Exception_Create( NULL, NULL, DOFEXCEPTION_NOT_FOUND, NULL );

                /* If result did not create, this next call does nothing. Would be useless to make a memory exception.*/
                DOFSystem_DeliverResponseBackToRequestor( op, result );                  /* Destroys result for us. */
                return DOFROUTERSTATUS_ERROR;
            }
            entry = DOFMem_Alloc( sizeof(session_table_entry_t) );
            if (entry)
            {
                if ((observer = DOFOperation_AddOpWillCloseObserver( op, CleanupSessionTableForOp )) == FALSE)
                {
                    goto out_of_memory;
                }
                else
                {
                    if (OALHashtable_Insert( session_table, entry, &bytes, &size ) != OALHashtable_InsertSuccess)
                        goto out_of_memory;
                    entry->provAddress = DOFAddress_Copy( pbte->address );
                    if (!pbte->address || entry->provAddress )
                    {
                        entry->provChannel = OALChannel_Retain(pbte->provider_channel);
                        entry->reqOp = op;
                        DOFOperation_Retain( op );
                        entry->baseAlias = pbte->alias;
                        op->sources->alias = pbte->alias;
                        provider_binding_table_entry_release( pbte );
                    }
                    else
                        goto out_of_memory;
                }
            }
            else
            {
                DOFResult result;

            out_of_memory:
                provider_binding_table_entry_release( pbte );
                PCRMem_Free( entry );
                /* Failed OAP open session request:  send DOFException back to requestor. */
                result = DOFResult_Exception_Create( NULL, NULL, DOFEXCEPTION_INSUFFICIENT_RESOURCES, NULL );
                /* If result dud not create, this next call does nothing. */
                DOFSystem_DeliverResponseBackToRequestor( op, result );                /* Destroys result for us. */
                return DOFROUTERSTATUS_ERROR;
            }
        }
        else
        {
            /* We already know about this session. Do not submit to DOFSystem processing. */
            if (entry->reqOp != op)
            {
                /* Duplicate OAP open session request:  send DOFException back to requestor. */
                DOFResult result = DOFResult_Exception_Create( NULL, NULL, DOFEXCEPTION_DUPLICATE, NULL );

                /* If result did not create, this next call does nothing. Would be useless to make a memory exception.*/
                DOFSystem_DeliverResponseBackToRequestor( op, result );                  /* Destroys result for us. */
                return DOFROUTERSTATUS_DUPLICATE;
            }
            /* Otherwise, same old session we know about, coming in as a retry, which is OK. */
            PCR_Assert( op->flags.bits.new_retry_count );
        }
    }
    else
    { /* op's observer will be called to cleanup. */
    }

    /* Do this next for cancel, normal new request, or retry request. */
    provider_route_normal( op );      /* Calls DOFSystem_AcceptRequest_Wrapper(). */

    return DOFROUTERSTATUS_SUCCESS;
}

typedef struct RouteContext_t
{
    OALEnqueueItem_t super;
    DOFOperation operation;
    DOFResult result;
} RouteContext_t;

static void _route_to_provider(OALThreadPoolItem dnode, boolean shutdown)
{
    static DOFRouterStatusCode_t (* const route_provider_fns[DOFOPERATIONTYPE_INVALID])(DOFOperation op) = {
        provider_route_normal, /* DOFOPERATIONTYPE_GET */
        provider_route_normal, /* DOFOPERATIONTYPE_SET */
        provider_route_normal, /* DOFOPERATIONTYPE_INVOKE */
        provider_route_normal, /* DOFOPERATIONTYPE_REGISTER */
        provider_route_normal, /* DOFOPERATIONTYPE_SUBSCRIBE */
        provider_route_normal, /* DOFOPERATIONTYPE_FINDINTERFACEDEFINITION */
        route_interest, /* DOFOPERATIONTYPE_WATCH */
        route_interest, /* DOFOPERATIONTYPE_ACTIVATE */
        provider_route_normal, /* DOFOPERATIONTYPE_EXCEPTION */
        route_begin_provide, /* DOFOPERATIONTYPE_BEGINPROVIDE */
        route_provider_death, /* DOFOPERATIONTYPE_PROVIDE */
        route_provider_death, /* DOFOPERATIONTYPE_SIGNAL */
        provider_route_normal, /* DOFOPERATIONTYPE_CHANGED */
        route_provider_death, /* DOFOPERATIONTYPE_CANCEL */
        route_interest, /* DOFOPERATIONTYPE_CONNECT */
        route_provider_session, /* DOFOPERATION_SESSION */
        route_provider_request_config_or_flood, /* DOFOPERATIONTYPE_ADVERTISE */
        route_provider_death, /* DOFOPERATIONTYPE_QUERY */
        route_interest, /* DOFOPERATIONTYPE_FULLCONNECT */
        route_provider_death /* DOFOPERATIONTYPE_NOOP */
    };

    RouteContext_t * node;

    PCR_Assert(dnode != NULL);
    node = (RouteContext_t*) dnode;

    DOFOperation_Lock(node->operation);
    PCR_Assert(node->operation->operation_type<DOFOPERATIONTYPE_INVALID);

    /* Handle ACKs */
    if ( !node->operation->flags.bits.internally_created )
    {
        if ( !node->operation->ackTree )
            node->operation->ackTree = DOFOperation_CreateObjectIDAVLTree();
        if ( node->operation->ackTree )
            DOFOperationControl_ExclusionQueueAddToTree( node->operation->operationConfig, node->operation->ackTree );
    }
    
    PCRMutex_Lock(router_mutex);

    route_provider_fns[node->operation->operation_type](node->operation);
    /* This next is modeled after a similar solution used in operation_manager.c's _timer_observer(). */
    if (node->operation->flags.bits.cancelled && !node->operation->flags.bits.internally_created)
        DOFSystem_OpEarlyCancelCallback( node->operation );                 /* Fixes bug 3195. */

    node->operation->flags.bits.new_retry_count = 0;

    PCRMutex_Unlock(router_mutex);
    DOFOperation_Unlock(node->operation);
    DOFOperation_Internal_Destroy(node->operation);
    DOFMem_Free(node);
}


static void _route_to_requestor(OALThreadPoolItem dnode, boolean shutdown)
{
    static DOFRouterStatusCode_t (* const route_requestor_fns[DOFOPERATIONTYPE_INVALID])(DOFOperation op, DOFResult result) = {
        requestor_route_normal, /* DOFOPERATIONTYPE_GET */
        requestor_route_normal, /* DOFOPERATIONTYPE_SET */
        requestor_route_normal, /* DOFOPERATIONTYPE_INVOKE */
        requestor_route_subscribe, /* DOFOPERATIONTYPE_REGISTER */
        requestor_route_subscribe, /* DOFOPERATIONTYPE_SUBSCRIBE */
        requestor_route_normal, /* DOFOPERATIONTYPE_FINDINTERFACEDEFINITION */
        route_requestor_death, /* DOFOPERATIONTYPE_WATCH */
        route_requestor_death, /* DOFOPERATIONTYPE_ACTIVATE */
        requestor_route_normal, /* DOFOPERATIONTYPE_EXCEPTION */
        route_requestor_death, /* DOFOPERATIONTYPE_BEGINPROVIDE */
        route_provide,   /* DOFOPERATIONTYPE_PROVIDE */
        subscription_requestor_changed, /* DOFOPERATIONTYPE_SIGNAL */
        subscription_requestor_changed, /* DOFOPERATIONTYPE_CHANGED */
        route_requestor_death, /* DOFOPERATIONTYPE_CANCEL */
        route_requestor_death, /* DOFOPERATIONTYPE_CONNECT */
        requestor_route_stateful, /* DOFOPERATIONTYPE_SESSION */
        route_requestor_death, /* DOFOPERATIONTYPE_ADVERTISE */
        route_requestor_death, /* DOFOPERATIONTYPE_QUERY */
        route_requestor_death, /* DOFOPERATIONTYPE_FULLCONNECT */
        route_requestor_death /* DOFOPERATIONTYPE_NOOP */
    };

    RouteContext_t * node;

    PCR_Assert(dnode != NULL);
    node = (RouteContext_t*) dnode;
    DOFOperation_Lock(node->operation);
    PCR_Assert(node->operation->operation_type<DOFOPERATIONTYPE_INVALID);
    /* Mark reqOp with cancel flag- cancel happens after EPP_V2_Send() is called, which happens ONLY on provider side.*/
    if (node->operation->flags.bits.commandCompletesOnAnyFinal)
        node->operation->flags.bits.cancelOpAfterEppSend = 1;
    /* Ignore the Not Found exception response for flooded operations. */
    if ( !( node->operation->flags.bits.isFlooded && DOFErrorException_GetError( DOFResult_GetException( node->result ) ) == DOFERROR_NOT_FOUND ) )
    {
        PCRMutex_Lock(router_mutex);
        route_requestor_fns[node->operation->operation_type](node->operation, node->result);
        PCRMutex_Unlock(router_mutex);
    }
    DOFOperation_Unlock(node->operation);
    DOFOperation_Internal_Destroy(node->operation);
    DOFResult_Destroy(node->result);
    DOFMem_Free(node);
}

static boolean _enqueue_one(OALEnqueueFunc fn, DOFOperation op, DOFResult result)
{
    RouteContext_t * node;

    node = (RouteContext_t *) DOFMem_Alloc(sizeof(RouteContext_t));
    if(node != NULL)
    {
        OALEnqueueItem_Init((OALEnqueueItem) node, fn);
        node->result = DOFResult_Copy( result );
        if((node->result != NULL) || (result == NULL))
        {
            DOFOperation_Retain(op);
            node->operation = op;
            if (OALOperationManager_EnqueueOperation((OALEnqueueItem) node) == DOFERROR_OK)
                return TRUE;
            DOFOperation_Internal_Destroy( node->operation );
            DOFResult_Destroy( node->result );
        }
        DOFMem_Free(node);
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "OAPRouter: Failed to enqueue operation for routing.");
    return FALSE;
}

/* send to a provider */
DOFRouterStatusCode_t DOFRouter_RouteToProvider(DOFOperation op)
{
    if ( DOFERROR_OK != DOFOperationManager_SetRetry(op, NULL) )
        return DOFROUTERSTATUS_ERROR;

    /* don't need a lock because operation is unknown outside */
    if ( !op->flags.bits.added_to_operation_table && op->opID && DOFERROR_OK != OALOperationManager_AddOperation( op, op->opID, FALSE ) )
        return DOFROUTERSTATUS_ERROR;

    if ( !_enqueue_one( _route_to_provider, op, 0 ) )
        return DOFROUTERSTATUS_ERROR;

    return DOFROUTERSTATUS_SUCCESS;
}

/* send from a provider to a requestor */
DOFRouterStatusCode_t DOFRouter_RouteToRequestor(DOFOperation op, DOFResult result)
{
    if (!_enqueue_one(_route_to_requestor, op, result))
        return DOFROUTERSTATUS_ERROR;

    return DOFROUTERSTATUS_SUCCESS;
}

void DOFRouter_ChannelUp(OALChannel channel)
{
    /* send out all my interests */
    PCRMutex_Lock(router_mutex);
    interest_notify_new_channel(channel);
    PCRMutex_Unlock(router_mutex);
}

typedef struct traverse_t {
    DOFRouter_Traverse_fn_t fn;
    DOFRouter_Traverse_arg_t *arg;
} traverse_t;

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t object_traverse(OALAVLMultiTreeItem_t pbte, void* arg, boolean new_keyset)
{
    if(((provider_binding_table_entry_t*)pbte)->flags.bits.primary)
    {
        if(DOFSecurityScope_IsCompatible(OALChannel_GetSecurityScope(((provider_binding_table_entry_t*)pbte)->provider_channel, FALSE), ((traverse_t*)arg)->arg->securityScope))
        {
            ((traverse_t*)arg)->arg->oid = ((provider_binding_table_entry_t*)pbte)->provider_object_id;
            ((traverse_t*)arg)->fn(((traverse_t*)arg)->arg);
        }
    }

    return OALAVLMULTITREE_NORMAL;
}

/**
 Traverse over all bindings (local and remote) in the router and call the callback.

 @param callback The function to call.
 @param arg The argument to pass to the callback function.
 **/
void DOFRouter_TraverseAllObjects(DOFRouter_Traverse_fn_t callback, DOFRouter_Traverse_arg_t *arg, DOFRouterLocation_t where)
{
    PCR_Assert(callback);

    PCRMutex_Lock(router_mutex);

    {
        traverse_t t;

        t.fn = callback;
        t.arg = arg;

        if( where & DOFROUTERLOCATION_LOCAL )
            OALHashtable_Traverse_All(local_binding_tables[BINDING_TABLE_OBJECT], object_traverse, &t);
        if( where & DOFROUTERLOCATION_REMOTE )
            OALHashtable_Traverse_All(remote_binding_tables[BINDING_TABLE_OBJECT], object_traverse, &t);
    }

    PCRMutex_Unlock(router_mutex);
}

static boolean interface_traverse(OALHashItem_t pbte, void* arg)
{
    if(((provider_binding_table_entry_t*)pbte)->flags.bits.primary && DOFSecurityScope_IsCompatible( ((provider_binding_table_entry_t*)pbte)->provider_op->securityScope, ((traverse_t*)arg)->arg->securityScope ))
    {
        ((traverse_t*)arg)->arg->iid = ((provider_binding_table_entry_t*)pbte)->interface_id;
        ((traverse_t*)arg)->fn(((traverse_t*)arg)->arg);
    }

    return OALAVLMULTITREE_NORMAL;
}

/**

 Traverse over all interfaces in the system and call the given callback for each.

 @param callback The callback to call for each interface.
 @param arg The argument to pass to the callback.
 **/
void DOFRouter_TraverseAllInterfaces(DOFRouter_Traverse_fn_t callback, DOFRouter_Traverse_arg_t *arg)
{
    PCR_Assert(callback);
    PCR_Assert(arg->oid);

    PCRMutex_Lock(router_mutex);

    {
        traverse_t t;
        const void *keys[2];
        uint16 key_sizes[2];
        DOFObjectID did = DOFObjectIDDomain_GetObjectID(DOFSecurityScope_GetDomainID(arg->securityScope));

        keys[0] = DOFObjectID_GetBytes(did);
        keys[1] = DOFObjectID_GetBytes(arg->oid);

        key_sizes[0] = did ? DOFObjectID_GetBytesSize(did) : 0;
        key_sizes[1] = DOFObjectID_GetBytesSize(arg->oid);

        t.fn = callback;
        t.arg = arg;

        OALHashtable_Traverse(local_binding_tables[BINDING_TABLE_OBJECT], keys, key_sizes, interface_traverse, &t);
        OALHashtable_Traverse(remote_binding_tables[BINDING_TABLE_OBJECT], keys, key_sizes, interface_traverse, &t);
    }

    PCRMutex_Unlock(router_mutex);
}

static boolean interface_check(OALHashItem_t pbte, void* found)
{
    *((uint32*)found) = TRUE;

    return TRUE;
}

/**
 For the given security id and object, check if all the given interfaces are provided. Written for DOFQuery .

 @param oid The oid in question.
 @param partition The partition the object resides in.
 @param interfaces The interfaces that are desired.
 @param count The number of interfaces in the array.

 @return TRUE if all exist; otherwise FALSE.
 **/
boolean DOFRouter_CheckInterfaces(DOFObjectID oid, DOFSecurityScope securityScope, DOFInterfaceID *interfaces, uint32 count)
{
    uint32 index = 0;
    uint32 found = 0;
    uint32 counted = 0;

    const void *keys[3];
    uint16 key_sizes[3];
    DOFObjectID did;

    if(!securityScope) securityScope = DOF_GetDefaultSecurityScope(DOF_GetDof());

    did = DOFObjectIDDomain_GetObjectID(DOFSecurityScope_GetDomainID(securityScope));
    keys[0] = DOFObjectID_GetBytes(did);
    key_sizes[0] = did ? DOFObjectID_GetBytesSize( did ) : 0;

    keys[2] = DOFObjectID_GetBytes(oid);
    key_sizes[2] = oid ? DOFObjectID_GetBytesSize(oid) : 0;

    PCRMutex_Lock(router_mutex);

    for ( ; index<count; ++index )
    {
        found = 0;

        keys[1] = DOFInterfaceID_GetBytes(interfaces[index]);
        key_sizes[1] = DOFInterfaceID_GetBytesSize(interfaces[index]);
        /* first check local binding */
        OALHashtable_Traverse(local_binding_tables[BINDING_TABLE], keys, key_sizes,
                              interface_check, &found);

        if (found)
        {
            ++counted;
            continue;
        }

        /* now check remote binding */
        OALHashtable_Traverse(remote_binding_tables[BINDING_TABLE], keys, key_sizes,
                              interface_check, &found);

        if (found)
            ++counted;
    }

    PCRMutex_Unlock(router_mutex);

    return (counted==count);
}

struct _gapc /* for Get All Providing Connections */
{
    uint32 currentAlloc;
    DOFConnectionList cl;
    DOFObjectID objectID;
    DOFInterfaceID interfaceID;
    boolean iidW;
    boolean oidW;
    boolean success;
}_gapc;

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t _build_providing_connection_list(OALAVLMultiTreeItem_t pbte, void* ctx, boolean new_set)
{
    boolean add = FALSE;

    if( ((provider_binding_table_entry_t*)pbte)->provider_channel->remoteChannelData->flags.isConnection )
    {
        if( ((struct _gapc*)ctx)->oidW && ((struct _gapc*)ctx)->iidW )
        {
            /* every one */
            add = TRUE;
        }
        else if( !((struct _gapc*)ctx)->oidW && !((struct _gapc*)ctx)->iidW )
        {
            /* check full binding */
            if( DOFObjectID_IsEqual(((provider_binding_table_entry_t*)pbte)->provider_object_id, ((struct _gapc*)ctx)->objectID) && DOFInterfaceID_IsEqual(((provider_binding_table_entry_t*)pbte)->interface_id, ((struct _gapc*)ctx)->interfaceID) )
            {
                add = TRUE;
            }
        }
        else if( ((struct _gapc*)ctx)->oidW )
        {
            /* all objects providing the interface */
            if( DOFInterfaceID_IsEqual(((provider_binding_table_entry_t*)pbte)->interface_id, ((struct _gapc*)ctx)->interfaceID) )
            {
                add = TRUE;
            }
        }
        else
        {
            /* all interfaces this object is providing */
            if( DOFObjectID_IsEqual(((provider_binding_table_entry_t*)pbte)->provider_object_id, ((struct _gapc*)ctx)->objectID) )
            {
                add = TRUE;
            }
        }

        /* Only add if match (above) and the connection actually still exists (this pbte/provide might be in process of cancelling). */
        if( add && ((provider_binding_table_entry_t*)pbte)->provider_channel->remoteChannelData->OALConn )
        {
            /* Do we need to realloc our list */
            if( ((struct _gapc*)ctx)->cl->connectionCount == ((struct _gapc*)ctx)->currentAlloc )
            {
                /*resize*/
                DOFConnection* ret;
                ((struct _gapc*)ctx)->currentAlloc += 8; /* this should be tuned */
                ret = DOFMem_Realloc(((struct _gapc*)ctx)->cl->connectionArray, (sizeof(DOFConnection)*(((struct _gapc*)ctx)->currentAlloc)));

                if( ret )
                {
                    ((struct _gapc*)ctx)->cl->connectionArray = ret;
                }
                else
                {
                    ((struct _gapc*)ctx)->success = FALSE;
                    return OALAVLMULTITREE_TERMINATE_TRAVERSE;
                }
            }

            /* add connection inc count */
            ((struct _gapc*)ctx)->cl->connectionArray[((struct _gapc*)ctx)->cl->connectionCount] = DOFConnection_Copy(OALConnection_GetDOFConnection(((provider_binding_table_entry_t*)pbte)->provider_channel->remoteChannelData->OALConn));
            if( ((struct _gapc*)ctx)->cl->connectionArray[((struct _gapc*)ctx)->cl->connectionCount] == NULL )
            {
                ((struct _gapc*)ctx)->success = FALSE;
                return OALAVLMULTITREE_TERMINATE_TRAVERSE;
            }
            ++((struct _gapc*)ctx)->cl->connectionCount;
        }
    }

    return OALAVLMULTITREE_NORMAL;
}

/* This is for DOFRuntime_GetProvidingConnections */
/* assumes connectionList has been created but not the array of connections in it */
boolean _get_all_providing_connections(DOFConnectionList connectionList, const DOFObjectID objectID, const DOFInterfaceID interfaceID)
{
    struct _gapc ctx;                  /* If ctx is used below, every member if filled- no need to fill ctx with 0's. */

    PCR_Assert(connectionList);

    connectionList->connectionArray = DOFMem_Alloc(sizeof(DOFConnection)*8);

    if( !connectionList->connectionArray )
        return FALSE;

    ctx.success = TRUE;
    ctx.currentAlloc = 8;
    ctx.cl = connectionList;
    ctx.objectID = objectID;
    ctx.cl->connectionCount = 0;
    ctx.interfaceID = interfaceID;
    ctx.oidW = DOFObjectID_IsWildcard(objectID);
    ctx.iidW = DOFInterfaceID_IsWildcard(interfaceID);

    PCRMutex_Lock(router_mutex);
    OALHashtable_Traverse_All(remote_binding_tables[BINDING_TABLE], _build_providing_connection_list, (void*)&ctx);
    PCRMutex_Unlock(router_mutex);

    return ctx.success;
}

/* This is for DOFRuntime_GetProvideCount */
uint32 OALRouter_GetProviderCount( void )
{
    uint32 count;
    PCRMutex_Lock( router_mutex );
    count = OALHashtable_GetNumItems( remote_binding_tables[BINDING_TABLE] );
    count += OALHashtable_GetNumItems( local_binding_tables[BINDING_TABLE] );
    PCRMutex_Unlock( router_mutex );
    return count;
}

/**
 @}
 **/
