/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/interface.h>
#include <dof/interfaceid.h>
#include <dof/system.h>
#include <dof/dof.h>
#include <dof/permission.h>
#include <dof/permissionset.h>
#include <dof/providerinfo.h>
#include <internal/bridgeinternal.h>
#include <internal/iteminternal.h>
#include <internal/router.h>
#include <internal/hashtable.h>
#include <internal/exception.h>
#include <internal/channel.h>
#include <internal/operationinternal.h>
#include <internal/permissioninternal.h>
#include <dof/result.h>
#include <internal/resultinternal.h>
#include <internal/securityscope.h>
#include <internal/systeminternal.h>
#include <dof/object.h>
#include <internal/iteminternal.h>
#include "r.h"
#include <dof/item.h>
#include <internal/timer.h>

/**  @internal
    Deals with router subscriptions.
@{
**/

static OALHashtable subscription_table;

typedef struct subscription_information /* one per interface item */
{
    OALChannel accept_channel;
    DOFAddress accept_address;
    DOFObjectID oid;
    OALAVLMultiTree op_table_by_chan_addr;
    OALHashtable op_unique_table;
    provider_binding_table_entry_t * pbte; /* This should not be used directly; only used for comparison against the primary pbte to ensure it's the same. */
    uint16 accept_sequence;
    boolean accepted;
} subscription_information;

typedef struct subscription_information_and_op
{
    DOFOperation op;
    subscription_information *si;
} subscription_information_and_op;

typedef struct subscription_information_and_op_subscribe
{
    subscription_information_and_op opsi;
    DOFResult pending_result; /* min delay handling requires we keep a reference to the latest result */
    DOFOperation dummy;
    OALTimerItem timer_item;
    uint8 not_ok_to_send;
} subscription_information_and_op_subscribe;

static OALTimer backlog_timer; /* used for guaranteed event backlog and min delay */
/*static PCRMutex mutex;*/

static DOFOperation create_dumb_op(OALChannel channel, DOFInterfaceID iid, DOFObjectID oid, DOFItem item, DOFItemID item_id, DOFOperationType type, boolean no_table)
{
    /* No need to make a copy of IID or OID since internal operations always make their own copies */
    DOFOperation op;

    if (type == DOFOPERATIONTYPE_CHANGED)
    {
        op = DOFOperation_Change_CreateInternal(channel, iid, oid, 0, DOF_TIMEOUT_NEVER, 0, 0, item_id);
    }
    else if (type == DOFOPERATIONTYPE_SIGNAL)
    {
        op = DOFOperation_Signal_CreateInternal(channel, iid, oid, 0, DOF_TIMEOUT_NEVER, 0, 0, item_id);
    }
    else
    {
        PCR_Assert(FALSE);
        return NULL;
    }

    if ( !op )
        return NULL;
    else
    {
        DOFSecurityScope_Destroy(op->securityScope);
        op->securityScope = DOFSecurityMembrane_ValidateOutboundPermissions(channel->securityMembrane, op->commandPerm, op->commandPerm);

        if ( op->securityScope && OALPermissionScope_GetIdCount(DOFSecurityScope_GetOALPermissionScope(op->securityScope)) == 0 )
        {
            /* Denied. */
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OAPRouter: Change or signal failed permission check." );
            DOFOperation_Internal_Destroy(op);
            return NULL;
        }
    }

    op->item = item;
    op->itemId = item_id;
    DOFItem_RetainDesc(item);

    op->flags.bits.dont_add_to_operation_table = no_table; /* tell router not to add it to the op table */
    op->flags.bits.internally_created = FALSE;
    op->operation_type = type;
    return op;
}

static void subscription_make_keys(DOFSecurityScope securityScope, DOFObjectID oid, DOFInterfaceID iid, DOFItemID *item_id, OALHashKey_t *keys, OALHashKeySize_t *key_sizes)
{
    DOFObjectID did = DOFObjectIDDomain_GetObjectID(DOFSecurityScope_GetDomainID(securityScope));

    keys[0] = DOFObjectID_GetBytes(did);
    /* This assert not commited, here to gather data for below use of DOFObjectID_GetBytesSize(). */
    /*PCR_Assert( did );*/
    key_sizes[0] = did ? DOFObjectID_GetBytesSize( did ) : 0;

    keys[1] = DOFObjectID_GetBytes(oid);
    /* This assert not commited, here to gather data for below use of DOFObjectID_GetBytesSize(). */
    /*PCR_Assert( oid );*/
    key_sizes[1] = oid ? DOFObjectID_GetBytesSize( oid ) : 0;

    keys[2] = DOFInterfaceID_GetBytes(iid);
    key_sizes[2] = DOFInterfaceID_GetBytesSize(iid);

    keys[3] = item_id;
    key_sizes[3] = sizeof(DOFItemID);
}

static void channel_addr_table_make_keys(OALChannel *channel, DOFAddress address, OALHashKey_t *keys, OALHashKeySize_t *key_sizes)
{
    /* key on the address of channel */
    keys[0] = channel;
    key_sizes[0] = sizeof(OALChannel);

    if ( address )
    {
        keys[1] = DOFAddress_GetAddress(address);
        key_sizes[1] = DOFAddress_GetAddressSize(address);
    }
    else
    {
        keys[1] = NULL;
        key_sizes[1] = 0;
    }
}

/**
    Initialize the subscription system.
**/
boolean subscriptions_init(void)
{
    subscription_table = OALHashtable_Create(4);
    if ( subscription_table )
    {
        backlog_timer = OALTimer_Create(DOF_GetDof()->threadPool, FALSE);
        if ( backlog_timer )
        {
            return TRUE;
        }
        OALHashtable_Destroy(subscription_table, NULL, NULL);
    }
    return FALSE;
}

/**
    Clean up the subscription system.
**/
void subscriptions_cleanup(void)
{
    /* all operations at this point should be cancelled, so there's no point in cleaning this up */
    OALHashtable_Destroy(subscription_table, 0, 0);
    OALTimer_Destroy(backlog_timer);
}

struct find_si_t {
    subscription_information *si;
    DOFSecurityScope securityScope;
};

static boolean _find_si(OALHashItem_t si, void* arg)
{
    if ( DOFSecurityScope_IsCompatible(OALChannel_GetSecurityScope(((subscription_information*)si)->accept_channel, TRUE), ((struct find_si_t*)arg)->securityScope) )
    {
        PCR_Assert( ((struct find_si_t*)arg)->si == NULL );
        ((struct find_si_t*)arg)->si = ((subscription_information*)si);
        return TRUE; /* exit early */
    }
    return FALSE;
}

static subscription_information *find_si(DOFObjectID oid, DOFInterfaceID iid, DOFItemID item_id, DOFSecurityScope securityScope)
{
    struct find_si_t find;

    OALHashKey_t keys[4] = {NULL,NULL,NULL,NULL};
    OALHashKeySize_t key_sizes[4] = {0,0,0,0};

    find.si = NULL;
    find.securityScope = securityScope;

    subscription_make_keys(securityScope, oid, iid, &item_id, keys, key_sizes);
    OALHashtable_Traverse(subscription_table, keys, key_sizes, _find_si, &find);

    return find.si;
}

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t sendTerminated(OALAVLMultiTreeItem_t op, void* pbte, boolean new_set)
{
    DOFResult terminatedResult = DOFResult_Exception_Create( ((provider_binding_table_entry_t*)pbte)->provider_channel, ((provider_binding_table_entry_t*)pbte)->address, DOFEXCEPTION_TERMINATED, ((provider_binding_table_entry_t*)pbte)->provider_object_id );
    /* If result did not create, this next call does nothing. Would be useless to make a memory exception.*/
    DOFSystem_DeliverResponseBackToRequestor( ((DOFOperation)op), terminatedResult ); /* Destroys result for us. */
    return OALAVLMULTITREE_NORMAL;
}

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t removeProviderSubscription(OALAVLMultiTreeItem_t si, void* pbte, boolean new_set)
{
    if ( ((subscription_information*)si)->pbte == ((provider_binding_table_entry_t*)pbte) )
    {
        ((subscription_information*)si)->pbte = NULL;
        ((subscription_information*)si)->accepted = FALSE;
        OALHashtable_Traverse_All( ((subscription_information*)si)->op_unique_table, sendTerminated, ((provider_binding_table_entry_t*)pbte));
    }
    return OALAVLMULTITREE_NORMAL;
}

void DOFRouter_RemoveProviderSubscriptions(provider_binding_table_entry_t *pbte)
{
    OALHashtable_Traverse_All( subscription_table, removeProviderSubscription, pbte);
}

/** @internal
    Get event registered status for a given binding, item_id, and securityID.

    @param oid    DOFObjectID of the given binding.
    @param evt    DOFInterfaceEvent, representing the Interface ID of the binding and Item ID of the event.
    @param securityScope  The given securityScope.

    @return TRUE if the event's given binding, item_id, and security_id is registered, FALSE otherwise.
**/
boolean DOFRouter_IsEventRegistered(DOFObjectID oid, DOFInterfaceEvent evt, DOFSecurityScope securityScope)
{
    boolean found = FALSE;

    PCR_Assert( oid );
    PCR_Assert( evt );

    PCRMutex_Lock(router_mutex);

    {
        subscription_information *si = find_si(oid, DOFInterfaceEvent_GetInterfaceID( evt ), DOFInterfaceEvent_GetItemID( evt ), securityScope);

        if ( si && si->accepted )
            found = TRUE;
    }

    PCRMutex_Unlock(router_mutex);

    return found;
}

typedef struct traverse_t {
    DOFResult result;
    DOFObjectID oid;
    DOFOperationType type;
    DOFOperation op;
} traverse_t;

/* This is the call used for minDelay on subscriptions for properties. */
static int subscribe_timer_fn(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown)
{
    subscription_information_and_op_subscribe *opsi = (subscription_information_and_op_subscribe*)argv[0];
    int ret = 0;

    PCR_Assert(argc==1);
    PCR_Assert( opsi );
    PCRMutex_Lock(router_mutex);
    if (opsi->timer_item)
    {
        if (opsi->dummy)
        {
            /* now it's time to send our operation */
            OALChannel channel = opsi->opsi.op->sources->channel;
            DOFOperation dummy = opsi->dummy;
            DOFResult result = opsi->pending_result;
            opsi->dummy = NULL;
            opsi->pending_result = NULL;
            PCRMutex_Unlock( router_mutex );

            channel->routeToRequestor( channel, dummy, result );

            DOFOperation_Internal_Destroy( dummy );
            DOFResult_Destroy( result );
            return 0;
        }
        else
        {
            /* there's no operation to send, so now we need to clear the not_ok_to_send flag and turn off the timer */
            opsi->not_ok_to_send = 0;
            ret = 1; /* This causes the timer to no longer be periodic. */
        }
    }
    else
    {
        /* subscription_observer(), if already called, may have NULL'ed out opsi->timer_item. */
    }
    PCRMutex_Unlock( router_mutex );

    return ret;
}

static int check_subscribe_ok_to_send(subscription_information_and_op_subscribe *opsi, DOFOperation dummy, DOFResult result)
{
    PCRMutex_Lock(router_mutex);
    if(opsi->not_ok_to_send)
    {
        /* mindelay says we can't send right now */
        DOFResult r = DOFResult_Copy( result );
        DOFOperation o = DOFOperation_Retain( dummy ) ? dummy : NULL;

        if ( r && o )
        {
            if(opsi->pending_result) /* is a result already pending? */
            {
                /* there is already a result pending, so let's free it and replace it with our new result */
                DOFResult_Destroy(opsi->pending_result); /* release our reference to the old result */
                /*  Note: helgrind reports a mutex lock out of order because this operation destroy is called with the
                    router mutex held.  However, there is really no risk of deadlock since the operation will only be
                    locked if the ref count goes to zero as a result of calling the destroy function. */
                DOFOperation_Internal_Destroy(opsi->dummy); /* release our reference to the old dummy op */
            }
            opsi->pending_result = r; /* copy the new result in and let the timer fn send it later */
            opsi->dummy = o;
        }
        else
        {
            DOFResult_Destroy( r );
            DOFOperation_Internal_Destroy( o );
        }

        PCRMutex_Unlock(router_mutex);

        return -1; /* not ok to send it */
    }

    opsi->not_ok_to_send = 1; /* prevent the next send */

    if(!opsi->timer_item)
        opsi->timer_item = OALTimer_CreateItem(backlog_timer, subscribe_timer_fn, 1, (void**)&opsi);
    if (opsi->timer_item)
        OALTimerItem_Insert( opsi->timer_item, opsi->opsi.op->minPeriod, TRUE, FALSE );

    PCRMutex_Unlock(router_mutex); /* it's ok to send, so proceed normally */

    return 0; /* it's ok to send it */
}

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t _subscription_send_signal_or_change(OALAVLMultiTreeItem_t opsi, void* t, boolean new_keyset)
{
    /* the optimizer should probably optimize these out (readable code ftw) */
    subscription_information_and_op_subscribe *subscribe_opsi = (subscription_information_and_op_subscribe*)opsi;

    if (((subscription_information_and_op*)opsi)->op->flags.bits.internally_created &&      /* we're internally created */
        !((subscription_information_and_op*)opsi)->op->flags.bits.subscribeOk &&            /* we haven't accepted a subscribe/reg on this operation */
        (!((traverse_t*)t)->result || DOFResult_GetStream( ((traverse_t*)t)->result ) == NULL))    /* we aren't forwarding from somewhere else */
    {
        return OALAVLMULTITREE_NORMAL;
    }

    if(((subscription_information_and_op*)opsi)->op->sources->address)
    {
        boolean more_to_do = FALSE;

        if ( ((traverse_t*)t)->type==DOFOPERATIONTYPE_CHANGED )
        {
            /* we're about to send a changed event, so we need to set up the timer for min delay processing
             to ensure another isn't sent too soon */
            more_to_do = ( ((subscription_information_and_op*)opsi)->op->minPeriod != 0); /* if minPeriod == 0, then there's no need to do any extra processing */
        }

        {
            DOFResult temp;
            OALChannel c = DOFResult_GetProviderChannel(((traverse_t*)t)->result);
            DOFOperation dummy_op = create_dumb_op(c, ((subscription_information_and_op*)opsi)->op->interface_id, ((traverse_t*)t)->oid, ((subscription_information_and_op*)opsi)->op->item, ((subscription_information_and_op*)opsi)->op->itemId, ((traverse_t*)t)->type, 1);

            if (!dummy_op)
            {
                goto out;
            }

            dummy_op->sources->address = DOFAddress_Copy(((subscription_information_and_op*)opsi)->op->sources->address);
            if ( !dummy_op->sources->address && ((subscription_information_and_op*)opsi)->op->sources->address )
            {
                DOFOperation_Internal_Destroy(dummy_op);
                goto out;
            }

            dummy_op->callbacks = ((subscription_information_and_op*)opsi)->op->callbacks;

            dummy_op->sources->alias = ((subscription_information_and_op*)opsi)->op->sources->alias;

            if ( more_to_do )
            {
                if ( !check_subscribe_ok_to_send(subscribe_opsi, dummy_op, ((traverse_t*)t)->result) )
                {
                    more_to_do = FALSE; /* free the dummy op after sending it */
                }
                else
                {
                    DOFOperation_Internal_Destroy(dummy_op); /* check_subscribe_ok_to_send took my reference */
                    return OALAVLMULTITREE_NORMAL; /* can't do anything else */
                }
            }

            /* route the changed or signal pdu by passing it back through the router */
            temp = DOFResult_DeepCopy( ((traverse_t*)t)->result );
            if( temp )
            {
                ((subscription_information_and_op*)opsi)->op->sources->channel->routeToRequestor(((subscription_information_and_op*)opsi)->op->sources->channel, dummy_op, temp);
                DOFResult_Destroy( temp );
            }
            else
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OAPRouter: Failed to copy result.");
            }

            if(!more_to_do)
            {
                DOFOperation_Internal_Destroy(dummy_op);
            }
        }
        /*  we only need to send to a remote address once */
        return OALAVLMULTITREE_TERMINATE_SET;
    }
    else
    {
        DOFResult temp;
        /* if address is 0, then send to all operations for this chan/addr pair */

        if ( ((traverse_t*)t)->type==DOFOPERATIONTYPE_CHANGED )
        {
            /* handle local throttling */
            if ( check_subscribe_ok_to_send(subscribe_opsi, ((subscription_information_and_op*)opsi)->op, ((traverse_t*)t)->result) )
                return OALAVLMULTITREE_NORMAL;
        }

        temp = DOFResult_DeepCopy( ((traverse_t*)t)->result );
        if( temp )
        {
            ((subscription_information_and_op*)opsi)->op->sources->channel->routeToRequestor(((subscription_information_and_op*)opsi)->op->sources->channel, ((subscription_information_and_op*)opsi)->op, temp);
            DOFResult_Destroy( temp );
        }
        else
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OAPRouter: Sending Signal/Change to all operations, failed to copy result.");
        }

        return OALAVLMULTITREE_NORMAL;
    }

out:
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OAPRouter: Failed to create operation to send signal or change.");
    return OALAVLMULTITREE_TERMINATE_SET;
}

static void subscription_send_signal_or_change(subscription_information *si, DOFOperation op, DOFObjectID oid, DOFItem item, DOFItemID itemID, DOFSecurityScope securityScope, DOFResult result, DOFOperationType type, uint16 seq)
{
    PCR_Assert(si);
    PCR_Assert(type==DOFOPERATIONTYPE_CHANGED||type==DOFOPERATIONTYPE_SIGNAL);

    PCRMutex_Lock(router_mutex);
    /* Check that the sequence is later (old/same sequences are dropped here). */
    if ( (int16)( seq - si->accept_sequence ) > 0 )
    {
        traverse_t t;

        t.result = result;
        t.oid = oid;
        t.type = type;
        t.op = op;

        DOFResult_SetSequenceNumber(result, seq, ( (int16)( seq - si->accept_sequence ) > 1 ), ( (int16)( seq - si->accept_sequence ) > 0 ));
        si->accept_sequence = seq;
        OALAVLMultiTree_Traverse_All(si->op_table_by_chan_addr, _subscription_send_signal_or_change, &t);
    }
    PCRMutex_Unlock(router_mutex);
}

/** @internal
    Deliver an event signal report for a given binding, item_id, and securityID for sending to requestor.

    @param oid    DOFObjectID of the given binding.
    @param evt    DOFInterfaceEvent, representing the Interface ID of the binding and Item ID of the event.
    @param securityScope The given securityScope.
    @param result The DOFResult containing the list of DOFValues required in an event signal report.
    @return TRUE for event found in subscription table (and signal therefore sent), FALSE otherwise
**/
boolean DOFRouter_SignalEvent(DOFObjectID oid, DOFInterfaceEvent evt, DOFSecurityScope securityScope, DOFResult result)
{
    DOFItemID                  itemid;
    subscription_information * si;
    boolean                    retBoolean = FALSE;

    PCR_Assert( oid );
    PCR_Assert( evt );
    PCR_Assert( result );
    /* Send event signal to requestor unless evt (oid/iid/item_id) is not registered for obj. If registered, call
        DOFRouter_RouteToRequestor on the operation. For all registered events, call this once for operations with
        a non-zero address. Call it for all operations with a zero address because they are local and router is
        handling the routing to DOFSystem. */

    PCRMutex_Lock( router_mutex );
    itemid = DOFInterfaceEvent_GetItemID( evt );
    si = find_si( oid, DOFInterfaceEvent_GetInterfaceID( evt ), itemid, securityScope );
    if ( si )
    {
        subscription_send_signal_or_change( si, 0, oid, (DOFItem)evt, itemid, securityScope, result,
                                            DOFOPERATIONTYPE_SIGNAL, si->accept_sequence + 1 );
        retBoolean = TRUE;
    }
    PCRMutex_Unlock( router_mutex );

    return retBoolean;
}

static void DOFRouter_ExtendPermissions_Callback(DOFError   status,
                                                 void     * context)
{
    DOFOperation op = (DOFOperation) context;

    if(status == DOFERROR_OK)
    {
        PCR_Assert(op->securityScope == NULL);
        op->securityScope = DOFSecurityMembrane_ValidateOutboundPermissions(op->sources->channel->securityMembrane, op->commandPerm, op->commandPerm);
        if(op->securityScope)
        {
            if(OALPermissionScope_GetIdCount( DOFSecurityScope_GetOALPermissionScope(op->securityScope) ) != 0)
            {
                DOFRouter_RouteToProvider(op);
            }
        }
    }
    DOFOperation_Internal_Destroy(op);
}

static boolean DOFRouter_ExtendPermissions(DOFOperation op)
{
    DOFOperation_Retain(op);
    OALChannel_ExtendOutboundMembranePermissions(op->sources->channel, op->commandPerm, DOFRouter_ExtendPermissions_Callback, op);
    return TRUE;
}

/** @internal
    Deliver a property changed report for a given binding, item_id, and securityID.

    @note No notification of property change will be sent if prp not subscribed for obj.
    @param oid          DOFObjectID of the given binding.
    @param prp          DOFInterfaceProperty, representing the Interface ID of the binding and Item ID of the Property.
    @param securityScope The given securityScope.
    @return TRUE for success, FALSE for fail.
**/
boolean DOFRouter_PropertyChanged(DOFObjectID oid, DOFInterfaceProperty prp, DOFSecurityScope securityScope)
{
    subscription_information * si;
    DOFInterfaceID             iid;
    DOFItemID                  id;
    OALChannel                 acceptChannel = NULL;

    PCR_Assert( oid );
    PCR_Assert( prp );
    PCR_Assert( DOFInterfaceProperty_IsReadable( prp ) );
    iid = DOFInterfaceProperty_GetInterfaceID( prp );
    id = DOFInterfaceProperty_GetItemID( prp );

    /* Router may know of one or several requests for this binding/item_id/security_id property subscription report. The
       several requests may have different associated minPeriod values. Mark each such matching request as "changed." As
       each matching request qualifies by minPeriod for report transmission, obtain the required current DOFValue
       (embedded within a DOFResult) by calling DOFRouter_RouteToRequestor() with a request operation of type
       DOFOPERATIONTYPE_CHANGED. When its response comes back, forward it along as a property change response. If that
       forwarding is to be transmitted across the wire, an OAP Changed PDU will be formed. If by local loopback, call
       DOFSystem_AcceptResponse_Wrapper(). */

    /* Ignore mindelay here! Ultimately this will go into a timing loop in DOFRouter. */

    /* Determine if we need to route a property changed report, or not. */
    PCRMutex_Lock( router_mutex );
    si = find_si( oid, iid, id, securityScope );
    if (si)
        acceptChannel = si->accept_channel;
    PCRMutex_Unlock( router_mutex );
    if (acceptChannel)
    {
        /* We need to route a property changed report. */
        boolean      success = FALSE;
        DOFOperation op      = create_dumb_op( acceptChannel, iid, oid, (DOFItem)prp, id, DOFOPERATIONTYPE_CHANGED, 1 );

        if (op)
        {
            if(op->securityScope != NULL)
            {
                if(DOFRouter_RouteToProvider( op ) == DOFROUTERSTATUS_SUCCESS)
                {
                    success = TRUE;
                }
            }
            else
            {
                success = DOFRouter_ExtendPermissions(op);
            }
            DOFOperation_Internal_Destroy( op );
        }
        return success;
    }
    else
        return TRUE; /* We don't need to route a property changed report, so we do nothing, and we call that a success. */
}

static void subscription_observer(OALObserver self, OALObservable observable, void* ctx)
{
    subscription_information_and_op* opsi  = (subscription_information_and_op*)ctx;
    OALHashKeySize_t op_size               = sizeof(opsi->op);
    OALHashKey_t     keys[4]               = {NULL,NULL,NULL,NULL};
    OALHashKeySize_t key_sizes[4]          = {0,0,0,0};
    OALTimerItem     subscriptionTimerItem = NULL;

    PCR_Assert( opsi );
    PCRMutex_Lock(router_mutex);
    if (opsi->op->operation_type == DOFOPERATIONTYPE_SUBSCRIBE)
    {
        subscription_information_and_op_subscribe * _opsi = (subscription_information_and_op_subscribe * )opsi;

        subscriptionTimerItem = _opsi->timer_item;
        _opsi->timer_item = NULL;            /* OALTimerItem_Destroy() called after mutex unlocked. */
        DOFResult_Destroy( _opsi->pending_result );
        DOFOperation_Internal_Destroy( _opsi->dummy );
    }
    OALAVLMultiTree_Delete(opsi->si->op_table_by_chan_addr, opsi, opsi);
    OALHashtable_Remove_First(opsi->si->op_unique_table, (const void**)&opsi->op, &op_size);
    /* clean up the SI if there are no more operations for it */
    if (!OALAVLMultiTree_GetNumItems(opsi->si->op_table_by_chan_addr))
    {
        /* nothing left, so free that subscription info */
        DOFItemID item_id = opsi->op->itemId;

        subscription_make_keys( OALChannel_GetSecurityScope( opsi->si->accept_channel, FALSE ), opsi->si->oid,
                                opsi->op->interface_id, &item_id, keys, key_sizes );
        OALHashtable_Remove(subscription_table, opsi->si, keys, key_sizes);
        OALChannel_Release(opsi->si->accept_channel);
        DOFObjectID_Destroy(opsi->si->oid);
        DOFAddress_Destroy(opsi->si->accept_address);
        OALAVLMultiTree_Destroy(opsi->si->op_table_by_chan_addr, NULL, NULL);
        OALHashtable_Destroy(opsi->si->op_unique_table, 0, 0);
        DOFMem_Free(opsi->si);
    }
    PCRMutex_Unlock(router_mutex);
    OALTimerItem_Destroy( subscriptionTimerItem, TRUE );
    /* Above won't return until timer_item fully destroyed, subscribe_timer_fn() can be called while this happens.*/
    DOFOperation_Internal_Destroy(opsi->op);
    DOFMem_Free( opsi );
}

static boolean OALChannelAndAddress_CompareGreaterThan_fn_t(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    subscription_information_and_op* opsi_k = (subscription_information_and_op*)key;
    subscription_information_and_op* opsi_i = (subscription_information_and_op*)item;
    int32_t channelcmp = OALChannel_Compare(opsi_k->op->sources->channel, opsi_i->op->sources->channel) > 0;
    if( channelcmp <= 0 ) 
    {
        int32_t addresscmp = DOFAddress_Compare(opsi_k->op->sources->address, opsi_i->op->sources->address);
        if( !(channelcmp == 0 && addresscmp > 0) )
            return FALSE;
    }
    return TRUE;
}

static boolean OALChannelAndAddress_CompareEqual_fn_t(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    subscription_information_and_op* opsi_k = (subscription_information_and_op*)key;
    subscription_information_and_op* opsi_i = (subscription_information_and_op*)item;
    return OALChannel_Compare(opsi_k->op->sources->channel, opsi_i->op->sources->channel) == 0 && DOFAddress_Compare(opsi_k->op->sources->address, opsi_i->op->sources->address) == 0;
}

static boolean OPSI_CompareGreaterThan_fn_t(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    subscription_information_and_op* opsi_k = (subscription_information_and_op*)key;
    subscription_information_and_op* opsi_i = (subscription_information_and_op*)item;
    return opsi_k->op > opsi_i->op;
}

static boolean OPSI_CompareEqual_fn_t(OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context)
{
    subscription_information_and_op* opsi_k = (subscription_information_and_op*)key;
    subscription_information_and_op* opsi_i = (subscription_information_and_op*)item;
    return opsi_k->op == opsi_i->op;
}

/** @internal
    Add a new subscription.

    @param op The operation that represents the subscription to add.
    @param pbte Holds information on the provider.

    @return TRUE if successful, or FALSE if failure occurs.
**/
boolean subscription_add(DOFOperation op, provider_binding_table_entry_t *pbte)
{
    /* there is a new subscribe, so add it to the tables. on the requesting side,
    this is going to be routed to a remote provider(s) and each route needs an
    entry into the subscribe tables. on the providing side, the same information
    must be put into the tables on a per-subscription basis */

    OALHashKeySize_t op_size = sizeof(op);
    OALObserver o = NULL;
    boolean new_si = FALSE;

    subscription_information *si;

    DOFItemID item_id = op->itemId;

    subscription_information_and_op *opsi;

    if( op->flags.bits.internally_created )
    {
        /* provider side */
        if( op->flags.bits.provider_oid_is_wildcard )
        {
            DOFObjectID oid = DOFObjectID_Copy( pbte->provider_object_id );
            if(OALAVLTree_InsertSuccess != OALAVLTree_Insert( op->providers, oid, oid, NULL ) )
                DOFObjectID_Destroy( oid );
        }
    }

    /* we only want one subscription per provider id, so let's see if one exists to use */
    si = find_si( pbte->provider_object_id, pbte->interface_id, item_id, pbte->provider_op->securityScope );

    if ( !si )
    {
        OALHashKey_t keys[4] = {NULL,NULL,NULL,NULL};
        OALHashKeySize_t key_sizes[4] = {0,0,0,0};

        /* we don't have one, so we get to make a new one */
        si = DOFMem_Alloc(sizeof(subscription_information));
        if ( !si )
            goto out_fail;

        new_si = TRUE;

        PCRMem_Set(si, 0, sizeof(subscription_information));
        si->op_table_by_chan_addr = OALAVLMultiTree_Create(OALChannelAndAddress_CompareGreaterThan_fn_t, OALChannelAndAddress_CompareEqual_fn_t, OPSI_CompareGreaterThan_fn_t, OPSI_CompareEqual_fn_t, NULL);
        if ( !si->op_table_by_chan_addr )
            goto out_free;

        si->op_unique_table = OALHashtable_Create(1);
        if ( !si->op_unique_table )
            goto out_free;

        subscription_make_keys(pbte->provider_op->securityScope, pbte->provider_object_id, pbte->interface_id, &item_id, keys, key_sizes);

        if ( OALHashtable_Insert(subscription_table, si, keys, key_sizes) != OALHashtable_InsertSuccess )
            goto out_free;

        si->oid = DOFObjectID_Copy(pbte->provider_object_id);
        si->accept_channel = OALChannel_Retain(pbte->provider_channel);
        PCR_Assert(si->accept_channel);
        si->accept_address = DOFAddress_Copy( pbte->address );
        si->pbte = pbte; /* We do not retain it, because we only use it for comparison. */
        si->accepted = FALSE;
    }
    else
    {
        if ( pbte->flags.bits.primary )
        {
            OALChannel channel = OALChannel_Retain(pbte->provider_channel);
            DOFAddress address = DOFAddress_Copy( pbte->address );
            if ( channel && address )
            {
                OALChannel_Release( si->accept_channel );
                si->accept_channel = channel;
                DOFAddress_Destroy( si->accept_address );
                si->accept_address = address;
                si->pbte = pbte; /* We do not retain it, because we only use it for comparison. */
            }
            else
            {
                OALChannel_Release( channel );
                DOFAddress_Destroy( address );
            }
        }

        /* see if this op exists in the unique table */
        if ( OALHashtable_Retrieve(si->op_unique_table, (const void**)&op, &op_size) )
            return TRUE; /* Must be a retry; nothing more to do. */
    }

    if ( op->operation_type==DOFOPERATIONTYPE_SUBSCRIBE )
    {
        opsi = (subscription_information_and_op*)DOFMem_Alloc(sizeof(subscription_information_and_op_subscribe));
        if ( !opsi )
            goto out_free;
        PCRMem_Set(opsi, 0, sizeof(subscription_information_and_op_subscribe));
        if ( op->minPeriod )
        {
            ((subscription_information_and_op_subscribe*)opsi)->not_ok_to_send = 1;
            ((subscription_information_and_op_subscribe*)opsi)->timer_item = OALTimer_CreateItem(backlog_timer, subscribe_timer_fn, 1, (void**)&opsi);
            if(((subscription_information_and_op_subscribe*)opsi)->timer_item)
                OALTimerItem_Insert( ((subscription_information_and_op_subscribe*)opsi)->timer_item, op->minPeriod,
                                     TRUE, FALSE );
        }
    }
    else
    {
        opsi = (subscription_information_and_op*)DOFMem_Alloc(sizeof(subscription_information_and_op));
        if ( !opsi )
            goto out_free;
        PCRMem_Set(opsi, 0, sizeof(subscription_information_and_op));
    }

    /* store this operation in the table */
    opsi->si = si;
    opsi->op = op;

    o = DOFOperation_AddOpWillCloseObserverOtherArg(op, subscription_observer, opsi);

    if ( !o )
        goto out_free_opsi;

    if ( OALAVLMultiTree_Insert(si->op_table_by_chan_addr, opsi, opsi) != OALAVLTree_InsertSuccess )
        goto out_remove_observer;
    if ( OALHashtable_Insert(si->op_unique_table, op, (const void**)&op, &op_size) != OALHashtable_InsertSuccess )
        goto out_remove_op_table_by_chan_addr;

    return TRUE;

out_remove_op_table_by_chan_addr:
    OALAVLMultiTree_Delete(si->op_table_by_chan_addr, opsi, opsi);

out_remove_observer:
    DOFOperation_RemoveOpWillCloseObserver(op, o);

out_free_opsi:
    if( op->operation_type==DOFOPERATIONTYPE_SUBSCRIBE && op->minPeriod && (((subscription_information_and_op_subscribe*)opsi)->timer_item))
    {
        OALTimerItem_Destroy(((subscription_information_and_op_subscribe*)opsi)->timer_item, TRUE);
    }
    DOFMem_Free(opsi);

out_free:
    if ( new_si )
    {
        OALChannel_Release( si->accept_channel );
        DOFAddress_Destroy( si->accept_address );
        DOFObjectID_Destroy( si->oid );
        OALHashtable_Destroy( si->op_unique_table, 0, 0 );
        OALAVLMultiTree_Destroy( si->op_table_by_chan_addr, NULL, NULL );
        DOFMem_Free( si );
    }

out_fail:
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OAPRouter: Failed to add subscription information to router.");
    return FALSE;
}

struct reparent
{
    DOFOperation op;
    OALChannel channel;
    DOFAddress address;
}reparent;

static boolean internal_reparent(OALAVLTreeItem_t item, void *arg_)
{
    DOFObjectID providerID = (DOFObjectID)item;
    struct reparent* arg = (struct reparent*)arg_;
    subscription_information *si = find_si( providerID, arg->op->interface_id, arg->op->itemId, arg->op->securityScope);

    if ( si )
    {
        OALChannel newchannel;
        DOFAddress newaddress;
        subscription_information_and_op *opsi = NULL;
        subscription_information_and_op opsit;
        
        newchannel = arg->op->sources->channel;
        newaddress = arg->op->sources->address;
        arg->op->sources->channel = arg->channel;
        arg->op->sources->address = arg->address;

        opsit.op = arg->op;
        opsit.si = si;

        /* Lookup using old channel and address */
        opsi = (subscription_information_and_op*)OALAVLMultiTree_Retrieve_Item(si->op_table_by_chan_addr, &opsit, &opsit);
        if( opsi )
        {
            if(OALAVLMultiTree_Delete(opsi->si->op_table_by_chan_addr, opsi, opsi))
            {
                arg->op->sources->channel = newchannel;
                arg->op->sources->address = newaddress;
                /* reinsert with new channel and address */
                if(OALAVLMultiTree_Insert(opsi->si->op_table_by_chan_addr, opsi, opsi) != OALAVLTree_InsertSuccess)
                {
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "OAPRouter: Failed to reparent a subscription or register");
                }
                return FALSE;
            }
        }
        /* Reset channel address for failure case */
        arg->op->sources->channel = newchannel;
        arg->op->sources->address = newaddress;
    }
    return FALSE;
}

/** @internal This is called when a subscribe/register reparents.

 @param op The opeation that reparented.
**/
void router_subscription_reparent(DOFOperation op, OALChannel oldChannel, DOFAddress address)
{
    struct reparent cntx;

    cntx.op = op;
    cntx.channel = oldChannel;
    cntx.address = address;

    if( !op->flags.bits.provider_oid_is_wildcard )
    {
        internal_reparent(op->object_id.object_id, &cntx);
    }
    else
    {
        OALAVLTree_Traverse( op->providers, internal_reparent,(void*)&cntx);
    }
}

/** @internal
    This is called when a subscribe/register accept comes in by the router.

    @param op The subscribe/register operation.
    @param result Holds information about the provider.

    @return TRUE if the subscription is acceptable form the provider, FALSE
            otherwise.
**/
boolean subscription_accept(DOFOperation op, DOFResult result)
{
    DOFItemID item_id = op->itemId;
    DOFObjectID oid = ( op->flags.bits.provider_oid_is_wildcard ) ? DOFResult_GetProviderID( result ) : op->object_id.object_id;
    subscription_information *si = find_si(oid, op->interface_id, item_id, op->securityScope);

    if ( si )
    {
        /*
         * Only process packet and update sequence information when from the primary provider.
         * This means that both the channel and address must match.
         * Sequence is checked in subscription_send_signal_or_change.
         */
        if ( DOFAddress_IsEqual( DOFResult_GetAddress( result ), si->accept_address ) &&
             OALChannel_IsEquivalent( DOFResult_GetProviderChannel( result ), si->accept_channel ) )
        {
            if ( !si->accepted || DOFResult_IsLocal(result))
            {
                if ( !DOFResult_IsLocal(result) )
                    si->accept_sequence = DOFResult_GetSequenceNumber(result);

                DOFResult_SetSequenceNumber( result, si->accept_sequence, FALSE, !si->accepted );
                si->accepted = TRUE;
            }
            else if ( !DOFResult_IsLocal(result) )
            {
                uint16 seq = DOFResult_GetSequenceNumber(result);
                DOFResult_SetSequenceNumber(result, seq, ( (int16)( seq - si->accept_sequence ) > 1 ), ( (int16)( seq - si->accept_sequence ) > 0 ));
                si->accept_sequence = seq;
            }
            return TRUE;
        }
    }

    return FALSE;
}

/** @internal
    This is called when a signal/changed comes in by the router.

    @param op The signal/changed operation.
    @param result Holds information about the provider.

    @return only DOFROUTERSTATUS_SUCCESS because the caller doesn't care if any subscriptions are here or not.
**/
DOFRouterStatusCode_t subscription_requestor_changed(DOFOperation op, DOFResult result)
{
    DOFItemID item_id = op->itemId;
    subscription_information *si = find_si(op->object_id.object_id, op->interface_id, item_id, op->securityScope);

    if ( si )
    {
        uint16 seq = ( DOFResult_IsLocal(result) ) ? ( si->accept_sequence + 1 ) : DOFResult_GetSequenceNumber(result);

        /*
         * Only process packet and update sequence information when notification is newer and from the known provider.
         * This means that both the channel and address must match.
         * Sequence is checked in subscription_send_signal_or_change.
         */
        if ( si->accepted &&
             DOFAddress_IsEqual( op->sources->address, si->accept_address ) &&
             OALChannel_IsEquivalent( op->sources->channel, si->accept_channel ) )
        {
            subscription_send_signal_or_change(si, op, op->object_id.object_id, op->item, item_id, op->securityScope, result, op->operation_type, seq);
        }
    }

    return DOFROUTERSTATUS_SUCCESS;
}

/**
@}
**/
