/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/runtime.h>
#include <dof/core.h>
#include <dof/connection.h>
#include <dof/objectid.h>
#include <internal/securityscope.h>
#include <dof/server.h>
#include <internal/connection_internal.h>
#include <internal/router.h>
#include <internal/dofinternal.h>
#include <internal/runtimeinternal.h>
#include <internal/operationinternal.h>
#include <internal/serverinternal.h>
#include "runtime_local.h"
#include <internal/slist.h>
#include <internal/threadpool.h>
#include <internal/connectionmanager.h>

static PCRMutex runtime_mutex;

extern PCRMutex connection_server_mutex;

typedef struct DOFRuntime_t
{
    DOF_t     dof;
}DOFRuntime_t;

struct runtime_listener_wrapper {
    struct connection_state_wrapper_t *next;
    union {
        DOFRuntimeConnectionStateListener connection;
        DOFRuntimeServerStateListener server;
    } listener;
    uint32 notify_count; /* high bit will be used to indicate pending remove */
};

boolean DOFRuntime_Init( DOF dof )
{
    runtime_mutex = PCRMutex_Create();
    if(runtime_mutex != NULL)
    {
        if(DOFRouteInfo_Init(dof))
        {
            return TRUE;
        }
        PCRMutex_Destroy(runtime_mutex);
    }
    return FALSE;
}

void DOFRuntime_Shutdown( void )
{
    DOFRouteInfo_Shutdown();
    PCRMutex_Destroy(runtime_mutex);
    runtime_mutex = NULL;
}

typedef struct RuntimeThreadPoolNode_t
{
    OALEnqueueItem_t super;

    void * argv[4];
} RuntimeThreadPoolNode_t;

boolean DOFRuntime_AddConnectionListener(DOFRuntime self, DOFRuntimeConnectionStateListener listener)
{
    struct runtime_listener_wrapper *wrapper;
    if(listener)
    {
        PCR_Assert(listener->fns->StateChanged); 

        wrapper = DOFMem_Alloc(sizeof(*wrapper));
        if(wrapper)
        {
            wrapper->listener.connection = listener;
            wrapper->notify_count = 0;
            PCRMutex_Lock(runtime_mutex);
            OALSList_Insert(&self->dof.connection_head, (OALSListNode)wrapper);
            PCRMutex_Unlock(runtime_mutex);
            return 1;
        }
    }
    return 0;
}

boolean DOFRuntime_AddServerListener(DOFRuntime self, DOFRuntimeServerStateListener listener)
{
    struct runtime_listener_wrapper *wrapper;
    if(listener)
    {
        PCR_Assert(listener->fns->StateChanged); 

        wrapper = DOFMem_Alloc(sizeof(*wrapper));
        if(wrapper)
        {
            wrapper->listener.server = listener;
            wrapper->notify_count = 0;
            PCRMutex_Lock(runtime_mutex);
            OALSList_Insert(&self->dof.server_head, (OALSListNode)wrapper);
            PCRMutex_Unlock(runtime_mutex);
            return 1;
        }
    }
    return 0;
}

static boolean remove_connection_listener(OALSListNode item, void *arg)
{
    struct runtime_listener_wrapper* wrapper = (struct runtime_listener_wrapper*)item;
    DOFRuntimeConnectionStateListener listener = (DOFRuntimeConnectionStateListener)arg;

    if(wrapper->listener.connection==listener)
        return 1;
    return 0;
}

static boolean remove_server_listener(OALSListNode item, void *arg)
{
    struct runtime_listener_wrapper* wrapper = (struct runtime_listener_wrapper*)item;
    DOFRuntimeServerStateListener listener = (DOFRuntimeServerStateListener)arg;

    if(wrapper->listener.server==listener)
        return 1;
    return 0;
}

static void notify_connection_listener_removed(OALThreadPoolItem dnode, boolean shutdown)
{
    RuntimeThreadPoolNode_t * node = (RuntimeThreadPoolNode_t *) dnode;
    if(((struct runtime_listener_wrapper*)node->argv[0])->listener.connection->fns->Removed )
        ((struct runtime_listener_wrapper*)node->argv[0])->listener.connection->fns->Removed(((struct runtime_listener_wrapper*)node->argv[0])->listener.connection, (struct DOFRuntime_t*)node->argv[1], NULL);
    DOFMem_Free((struct runtime_listener_wrapper*)node->argv[0]);
    DOFMem_Free(node);
}

static void notify_server_listener_removed(OALThreadPoolItem dnode, boolean shutdown)
{
    RuntimeThreadPoolNode_t * node = (RuntimeThreadPoolNode_t *) dnode;
    if(((struct runtime_listener_wrapper*)node->argv[0])->listener.server->fns->Removed )
        ((struct runtime_listener_wrapper*)node->argv[0])->listener.server->fns->Removed(((struct runtime_listener_wrapper*)node->argv[0])->listener.server, (struct DOFRuntime_t*)node->argv[1], NULL);
    DOFMem_Free((struct runtime_listener_wrapper*)node->argv[0]);
    DOFMem_Free(node);
}

void DOFRuntime_RemoveConnectionListener(DOFRuntime self, DOFRuntimeConnectionStateListener listener)
{
    struct runtime_listener_wrapper *wrapper;
    if(listener)
    {   
        PCRMutex_Lock(runtime_mutex);
        wrapper = (struct runtime_listener_wrapper*)OALSList_FindAndRemove(&self->dof.connection_head, ((void *)0), remove_connection_listener, listener);
        if(wrapper)
        {
            if(!wrapper->notify_count)
            {
                RuntimeThreadPoolNode_t * node = (RuntimeThreadPoolNode_t *) DOFMem_Alloc(offsetof (RuntimeThreadPoolNode_t, argv) + sizeof(void *) * 2);
                if(node != ((void *)0))
                {
                    OALEnqueueItem_Init((OALEnqueueItem) node, notify_connection_listener_removed);
                    node->argv[0] = wrapper;
                    node->argv[1] = (struct DOFRuntime_t*)DOF_GetDof();
                    if(!OALEnqueue_Enqueue(DOF_GetDof()->enqueue, (OALEnqueueItem) node))
                    {
                        DOFMem_Free(node);
                    }
                }
            }
            else
            {
                wrapper->notify_count |= 0x80000000;
            }
        }
        PCRMutex_Unlock(runtime_mutex);
    }
}

void DOFRuntime_RemoveServerListener(DOFRuntime self, DOFRuntimeServerStateListener listener)
{
    struct runtime_listener_wrapper *wrapper;
    if(listener)
    {
        PCRMutex_Lock(runtime_mutex);
        wrapper = (struct runtime_listener_wrapper*)OALSList_FindAndRemove(&self->dof.server_head, ((void *)0), remove_server_listener, listener);
        if(wrapper)
        {
            if(!wrapper->notify_count)
            {
                RuntimeThreadPoolNode_t * node = (RuntimeThreadPoolNode_t *) DOFMem_Alloc(offsetof (RuntimeThreadPoolNode_t, argv) + sizeof(void *) * 2);
                if(node != ((void *)0))
                {
                    OALEnqueueItem_Init((OALEnqueueItem) node, notify_server_listener_removed);
                    node->argv[0] = wrapper;
                    node->argv[1] = (struct DOFRuntime_t*)DOF_GetDof();
                    if( !OALEnqueue_Enqueue(DOF_GetDof()->enqueue, (OALEnqueueItem) node) )
                    {
                        DOFMem_Free(node);
                    }
                }
            }
            else
            {
                wrapper->notify_count |= 0x80000000;
            }
        }
        PCRMutex_Unlock(runtime_mutex);
    }
}

static void notify_connection_listener(OALThreadPoolItem dnode, boolean shutdown)
{
    RuntimeThreadPoolNode_t * node = (RuntimeThreadPoolNode_t *) dnode;

    ((struct runtime_listener_wrapper*)node->argv[0])->listener.connection->fns->StateChanged(((struct runtime_listener_wrapper*)node->argv[0])->listener.connection, (struct DOFRuntime_t*)node->argv[2], (DOFConnection)node->argv[3], (DOFConnectionState)node->argv[1]);
    DOFConnectionState_Destroy((DOFConnectionState)node->argv[1]);

    PCRMutex_Lock(runtime_mutex);
    --(((struct runtime_listener_wrapper*)node->argv[0])->notify_count);
    if( !(((struct runtime_listener_wrapper*)node->argv[0])->notify_count&0x7fffffff) )
    {
        if( ((struct runtime_listener_wrapper*)node->argv[0])->notify_count&0x80000000 )
            DOFMem_Free(node->argv[0]);
    }
    PCRMutex_Unlock(runtime_mutex);
    DOFMem_Free(node);
}

static void notify_server_listener(OALThreadPoolItem dnode, boolean shutdown)
{
    RuntimeThreadPoolNode_t * node = (RuntimeThreadPoolNode_t *) dnode;

    ((struct runtime_listener_wrapper*)node->argv[0])->listener.server->fns->StateChanged(((struct runtime_listener_wrapper*)node->argv[0])->listener.server, (struct DOFRuntime_t*)node->argv[2], (DOFServer)node->argv[3], (DOFServerState)node->argv[1]);

    DOFServerState_Destroy((DOFServerState)node->argv[1]);

    PCRMutex_Lock(runtime_mutex);
    --(((struct runtime_listener_wrapper*)node->argv[0])->notify_count);

    if( !(((struct runtime_listener_wrapper*)node->argv[0])->notify_count&0x7fffffff) )
    {
        if( ((struct runtime_listener_wrapper*)node->argv[0])->notify_count&0x80000000 )
            DOFMem_Free(node->argv[0]);
    }
    PCRMutex_Unlock(runtime_mutex);
    DOFMem_Free(node);
}

typedef struct DOFConnectionStateInfo_t
{
    DOFConnectionState state; DOFConnection obj;
} * DOFConnectionStateInfo;

typedef struct DOFServerStateInfo_t
{
    DOFServerState state; DOFServer obj;
} * DOFServerStateInfo;

static boolean notify_connection_traverse(OALSListNode item, void *arg)
{
    struct runtime_listener_wrapper* wrapper = (struct runtime_listener_wrapper*)item;
    DOFConnectionStateInfo stateinfo = (DOFConnectionStateInfo)arg;
    RuntimeThreadPoolNode_t * node = (RuntimeThreadPoolNode_t *) DOFMem_Alloc(offsetof (RuntimeThreadPoolNode_t, argv) + sizeof(void *) * 4);

    if( node != ((void *)0) )
    {
        OALEnqueueItem_Init((OALEnqueueItem) node, notify_connection_listener);
        ++(wrapper->notify_count);
        node->argv[0] = wrapper;
        node->argv[1] = DOFConnectionState_Copy(stateinfo->state);
        node->argv[2] = (struct DOFRuntime_t*)DOF_GetDof();
        node->argv[3] = stateinfo->obj;
        if( !OALEnqueue_Enqueue(DOF_GetDof()->enqueue, (OALEnqueueItem) node) )
        {
            DOFMem_Free(node);
        }
    }
    return 0;
}

static boolean notify_server_traverse(OALSListNode item, void *arg)
{
    struct runtime_listener_wrapper* wrapper =(struct runtime_listener_wrapper*)item;
    DOFServerStateInfo stateinfo = (DOFServerStateInfo)arg;
    RuntimeThreadPoolNode_t * node = (RuntimeThreadPoolNode_t *) DOFMem_Alloc(offsetof (RuntimeThreadPoolNode_t, argv) + sizeof(void *) * 4);

    if( node != ((void *)0) )
    {
        OALEnqueueItem_Init((OALEnqueueItem) node, notify_server_listener);
        ++(wrapper->notify_count);
        node->argv[0] = wrapper;
        node->argv[1] = DOFServerState_Copy(stateinfo->state);
        node->argv[2] = (struct DOFRuntime_t*)DOF_GetDof();
        node->argv[3] = stateinfo->obj;
        if( !OALEnqueue_Enqueue(DOF_GetDof()->enqueue, (OALEnqueueItem) node) )
        {
            DOFMem_Free(node);
        }
    }
    return 0;
}

void DOFRuntime_ConnectionStateChanged(DOFConnection obj, DOFConnectionState state)
{
    struct DOFConnectionStateInfo_t stateinfo;
    stateinfo.state = state;
    stateinfo.obj = obj;
    PCRMutex_Lock(runtime_mutex);
    OALSList_Traverse(DOF_GetDof()->connection_head, notify_connection_traverse, &stateinfo);
    PCRMutex_Unlock(runtime_mutex);
}

void DOFRuntime_ServerStateChanged(DOFServer obj, DOFServerState state)
{
    struct DOFServerStateInfo_t stateinfo;
    stateinfo.state = state;
    stateinfo.obj = obj;
    PCRMutex_Lock(runtime_mutex);
    OALSList_Traverse(DOF_GetDof()->server_head, notify_server_traverse, &stateinfo);
    PCRMutex_Unlock(runtime_mutex);
}

DOFRouteInfo DOFRuntime_CreateProviderRoute( const DOFRuntime self, const DOFObjectID oid, const DOFInterfaceID iid, const DOFSecurityScope scope )
{
    DOFRouteInfo info = NULL;

    if( self && oid && iid )
    {
        provider_binding_table_entry_t *pbte = DOFRouter_GetBindingEntry( scope, iid, oid );
        if(pbte)
        {
            info = DOFMem_Alloc(sizeof(*info));
            if(!info)
            {
                goto err_alloc;
            }

            info->address = DOFAddress_Copy( pbte->address );
            if(!info->address && pbte->address)
                goto err_address_copy;

            info->connectionState = OALChannel_CreateConnectionState(pbte->provider_channel, pbte->address);
            if(!info->connectionState)
                goto err_connection_state;

            info->securityScope = DOFSecurityScope_Copy(pbte->provider_op->securityScope);
            if(!info->securityScope)
                goto err_partition_copy;

            info->oid = DOFObjectID_Copy(oid);
            if(!info->oid)
                goto err_oid_copy;

            info->iid = DOFInterfaceID_Copy(iid);
            if(!info->iid)
                goto err_iid_copy;

            info->distLossless = pbte->provider_op->sources->distLossless;
            info->distLossy = pbte->provider_op->sources->distLossy;

            goto out_success;

        err_iid_copy:
            DOFObjectID_Destroy(info->oid);

        err_oid_copy:
            DOFSecurityScope_Destroy(info->securityScope);

        err_partition_copy:
            DOFConnectionState_Destroy(info->connectionState);

        err_connection_state:
            DOFAddress_Destroy(info->address);

        err_address_copy:
            DOFMem_Free(info);
            info = NULL;
        }

    err_alloc:
    out_success:
        provider_binding_table_entry_release(pbte);
    }

    return info;
}

DOFConnectionList DOFRuntime_GetProvidingConnections( const DOFRuntime self, const DOFObjectID objectID, const DOFInterfaceID interfaceID)
{
    PCR_Assert(self);

    if( objectID && interfaceID )
    {
        DOFConnectionList ret = DOFMem_Alloc(sizeof(struct DOFConnectionList_t));

        if( ret )
        {
            /* This next call sets all the fields in ret so no need for a memset */
            if( _get_all_providing_connections(ret,objectID,interfaceID))
            {
                return ret;
            }
            /* Error Cleanup */
            DOFConnectionList_Destroy(ret);
        }
    }
    return NULL;
}

struct connection_wrapper {
    DOFConnectionList conn_list;
    uint32 desired_count;
    boolean success;
};

static boolean add_connections(OALConnection connection, struct connection_wrapper *wrapper)
{
    DOFConnection c;

    PCR_Assert(wrapper->conn_list->connectionCount<wrapper->desired_count);

    if(wrapper->success)
    {
        c = DOFConnection_Copy(OALConnection_GetDOFConnection(connection));
        if(!c) wrapper->success = FALSE;
        wrapper->conn_list->connectionArray[wrapper->conn_list->connectionCount++] = c;
    }
    else
    {
        wrapper->conn_list->connectionArray[wrapper->conn_list->connectionCount++] = NULL;
    }

    return FALSE;
}

DOFConnectionList DOFRuntime_GetConnections( const DOFRuntime self )
{
    uint32 count;
    DOFConnectionList conn_list = DOFMem_Alloc(sizeof(struct DOFConnectionList_t));

    if(conn_list)
    {
        PCRMutex_Lock(connection_server_mutex);
        conn_list->connectionCount = 0;

        count = OALConnectionMgr_GetConnectionCount();
        conn_list->connectionArray = DOFMem_Alloc(sizeof(DOFConnection)*count);

        if(!conn_list->connectionArray)
        {
            DOFMem_Free(conn_list);
            conn_list = NULL;
        }
        else
        {
            struct connection_wrapper wrapper;

            wrapper.conn_list = conn_list;
            wrapper.desired_count = count;
            wrapper.success = TRUE;

            OALConnectionMgr_TraverseAllConnections((connection_manager_traverse_fn_t)add_connections, &wrapper);
            PCR_Assert(conn_list->connectionCount==count);
            if(!wrapper.success)
            {
                DOFConnectionList_Destroy(conn_list);
                conn_list = NULL;
            }
        }
        PCRMutex_Unlock(connection_server_mutex);
    }

    return conn_list;
}

struct server_list_wrapper
{
    DOFServerList server_list;
    uint32 desired_count;
    boolean success;
};

static boolean add_servers(OALServer server, void* context)
{
    struct server_list_wrapper* wrapper = (struct server_list_wrapper*)context;
    DOFServer s;

    PCR_Assert(wrapper->server_list->serverCount<wrapper->desired_count);

    if(wrapper->success)
    {
        s = DOFServer_Retain(OALServer_GetDOFServer(server));
        if(!s)
            wrapper->success = FALSE;
        wrapper->server_list->serverArray[wrapper->server_list->serverCount++] = s;
    }
    else
    {
        wrapper->server_list->serverArray[wrapper->server_list->serverCount++] = NULL;
    }
    return FALSE;
}

DOFServerList DOFRuntime_GetServers( const DOFRuntime self )
{
    uint32 count;
    DOFServerList server_list = DOFMem_Alloc(sizeof(struct DOFServerList_t));

    if(server_list)
    {
        DOFServer_Lock();
        count = DOFServerManager_GetServerCount(DOF_GetDof());
        server_list->serverArray = DOFMem_Alloc(sizeof(DOFServer)*count);
        server_list->serverCount = 0;
        if(!server_list->serverArray)
        {
            DOFMem_Free(server_list);
            server_list = NULL;
        }
        else
        {
            struct server_list_wrapper wrapper;

            wrapper.server_list = server_list;
            wrapper.desired_count = count;
            wrapper.success = TRUE;

            OALServer_TraverseAllServers(add_servers, &wrapper);
            PCR_Assert(server_list->serverCount == count);
            if(!wrapper.success)
            {
                DOFServerList_Destroy(server_list);
                server_list = NULL;
            }
        }
        DOFServer_Unlock();
    }

    return server_list;
}

uint16 DOFRuntime_GetConnectionCount( const DOFRuntime self )
{
    if( self )
    {
        uint32 count;
        PCRMutex_Lock( connection_server_mutex );
        count = OALConnectionMgr_GetConnectionCount();
        PCRMutex_Unlock( connection_server_mutex );
        if ( count > 0xFFFF )
            count = 0xFFFF;
        return (uint16)count;
    }
    return 0;
}

uint16 DOFRuntime_GetServerCount( const DOFRuntime self )
{
    if( self )
    {
        uint32 count;
        DOFServer_Lock();
        count = DOFServerManager_GetServerCount( (DOF)self );
        DOFServer_Unlock();
        if ( count > 0xFFFF )
            count = 0xFFFF;
        return (uint16)count;
    }
    return 0;
}

uint32 DOFRuntime_GetProvideCount( const DOFRuntime self )
{
    if( self )
        return OALRouter_GetProviderCount();
    return 0;
}

/* NOTE: These functions do not protect access of the variables because it is not critical that we get the instantaneous counts. */
uint64 DOFRuntime_GetReceiveByteCount( const DOFRuntime self )
{
    if( self )
        return ((DOF)self)->inputByteCount;
    return 0;
}

uint64 DOFRuntime_GetSendByteCount( const DOFRuntime self )
{
    if( self )
        return ((DOF)self)->outputByteCount;
    return 0;
}

uint64 DOFRuntime_GetReceivePacketCount( const DOFRuntime self )
{
    if( self )
        return ((DOF)self)->inputPacketCount;
    return 0;
}

uint64 DOFRuntime_GetSendPacketCount( const DOFRuntime self )
{
    if( self )
        return ((DOF)self)->outputPacketCount;
    return 0;
}

