/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/address.h>
#include <dof/core.h>
#include <dof/dof.h>
#include <dof/interfaceid.h>
#include <dof/objectid.h>
#include <dof/routeinfo.h>
#include <dof/system.h>
#include <dof/runtime.h>
#include <internal/channel.h>
#include <internal/runtimeinternal.h>
#include <internal/connection_internal.h>
#include <internal/interfaceidinternal.h>
#include <internal/objectinternal.h>
#include <internal/operationinternal.h>
#include <internal/router.h>
#include <internal/runtimesysteminternal.h>
#include <internal/securityscope.h>
#include <internal/slist.h>
#include <internal/systeminternal.h>
#include <internal/transportinternal.h>

#include "runtime_local.h"

/** @addtogroup DOFRouteInfo
@{
**/

static RoutingInfoChange_t * routingCallbacks = NULL;

static OALEnqueue runtime_enqueue = NULL;

boolean DOFRouteInfo_Init( DOF dof )
{
    runtime_enqueue = OALEnqueue_Create( dof->threadPool );
    return (runtime_enqueue != NULL);
}

void DOFRouteInfo_Shutdown( void )
{
    OALEnqueue_Destroy( runtime_enqueue );
    runtime_enqueue = NULL;
}

DOFConnectionState DOFRouteInfo_GetConnectionState(DOFRouteInfo self)
{
    if(!self)
        return NULL;
    return self->connectionState;
}

uint16 DOFRouteInfo_GetStreamDistance(DOFRouteInfo self)
{
    if(!self)
        return 0;
    return self->distLossless;
}

uint16 DOFRouteInfo_GetDatagramDistance(DOFRouteInfo self)
{
    if(!self)
        return 0;
    return self->distLossy;
}

DOFObjectID DOFRouteInfo_GetProviderID(DOFRouteInfo self)
{
    if(!self)
        return NULL;

    PCR_Assert( self->oid );
    return self->oid;
}

DOFAddress DOFRouteInfo_GetProviderAddress( const DOFRouteInfo self )
{
    if( !self )
        return NULL;
    return self->address;
}

DOFInterfaceID DOFRouteInfo_GetInterfaceID(DOFRouteInfo self)
{
    if(!self)
        return NULL;

    PCR_Assert( self->iid );
    return self->iid;
}

static void DOFRouteInfo_Empty(DOFRouteInfo self)
{
    if( self )
    {
        DOFSecurityScope_Destroy(self->securityScope);
        DOFConnectionState_Destroy(self->connectionState);
        DOFInterfaceID_Destroy( self->iid );
        DOFObjectID_Destroy( self->oid );
        DOFAddress_Destroy( self->address );

        PCRMem_Set(self, 0, sizeof(*self));
    }
}

void DOFRouteInfo_Destroy(DOFRouteInfo self)
{
    if( self )
    {
        DOFRouteInfo_Empty(self);
        DOFMem_Free( self );
    }
}


RoutingInfoChange_t * DOFRuntime_GetRoutingInfoCallbacks( void )
{
    return routingCallbacks;
}


/**
@}
**/

/** @addtogroup DOFSystem
@{
**/
static DOFRouteInfo DOFRuntime_FillRouteInfo(provider_binding_table_entry_t *pbte, DOFRouteInfoState change)
{
    DOFRouteInfo ri = DOFMem_Alloc(sizeof(DOFRouteInfo_t));

    if( !ri )
        return NULL;

    PCRMem_Set(ri,0,sizeof(DOFRouteInfo_t));

    if( pbte )
    {
        if( pbte->provider_channel )
            ri->connectionState = OALChannel_CreateConnectionState( pbte->provider_channel, pbte->address );
        if( ri->connectionState )
        {
            ri->address = DOFAddress_Copy( pbte->address );
            if( ri->address || !pbte->address )
            {
                PCR_Assert( pbte->interface_id );
                ri->iid = DOFInterfaceID_Copy( pbte->interface_id );
                if( ri->iid )
                {
                    PCR_Assert( pbte->provider_object_id );
                    ri->oid = DOFObjectID_Copy( pbte->provider_object_id );
                    if( ri->oid )
                    {
                        PCR_Assert( pbte->provider_channel );
                        ri->securityScope = DOFSecurityScope_Copy( OALChannel_GetSecurityScope( pbte->provider_channel,
                                                                                                FALSE ) );
                        if( ri->securityScope )
                        {
                            if( pbte->provider_op )
                            {
                                ri->distLossless = pbte->provider_op->sources->distLossless;
                                ri->distLossy = pbte->provider_op->sources->distLossy;
                            }
                            return ri;
                        }
                        DOFObjectID_Destroy( ri->oid );
                    }
                    DOFInterfaceID_Destroy( ri->iid );
                }
                DOFAddress_Destroy( ri->address );
            }
            DOFConnectionState_Destroy( ri->connectionState );
        }
    }

    DOFMem_Free(ri);
    return NULL;
}

static boolean DOFRuntime_FillRouteInfoFromRoutingInfo(DOFRouteInfo self, DOFRouteInfo ri)
{

    if( self && ri )
    {
        self->address = DOFAddress_Copy(ri->address);
        if( self->address )
        {
            self->connectionState = DOFConnectionState_Copy(ri->connectionState);
            if( self->connectionState )
            {
                self->securityScope = DOFSecurityScope_Copy(ri->securityScope);
                if( self->securityScope )
                {
                    self->oid = DOFObjectID_Copy(ri->oid);
                    if( self->oid )
                    {
                        self->iid = DOFInterfaceID_Copy(ri->iid);
                        if( self->iid )
                        {
                            self->distLossless = ri->distLossless;
                            self->distLossy = ri->distLossy;
                            return TRUE;
                        }
                        DOFObjectID_Destroy( self->oid );
                        self->oid = NULL;
                    }
                    DOFSecurityScope_Destroy( self->securityScope );
                    self->securityScope = NULL;
                }
                DOFConnectionState_Destroy( self->connectionState );
                self->connectionState = NULL;
            }
            DOFAddress_Destroy( self->address );
            self->address = NULL;
        }
    }
    return FALSE;
}

static void DOFRuntime_ClearRouteInfo( DOFRouteInfo self )
{
    DOFInterfaceID_Destroy( self->iid );
    DOFObjectID_Destroy( self->oid );
    DOFSecurityScope_Destroy( self->securityScope );
    DOFConnectionState_Destroy( self->connectionState );
    DOFAddress_Destroy( self->address );
}

static void internal_remove_all_ric(OALSListNode item, void *arg)
{
    DOFMem_Free(((RoutingInfoChange_t*)item));
}

/* 
 * removes callback to call when route info has changed
 */
void DOFRuntime_RemoveAllRoutingCallback( void )
{
    if( routingCallbacks )
    {
        OALSList_RemoveAll( ((OALSListNode*)&routingCallbacks), internal_remove_all_ric, NULL);
        routingCallbacks = NULL;
    }
}

typedef struct RoutingChange_t
{
    uint8               type;
    DOFRouteInfo        route_info;
    DOFRequest          request;
} RoutingChange_t;

static DOFOperation DOFOperation_CreateRoutingChange(RoutingChange_t* rci)
{
    DOFOperation ret = DOFMem_Alloc(sizeof(DOFOperation_t)+sizeof(RoutingChange_t)+sizeof(DOFRouteInfo_t));
    RoutingChange_t* opRci;

    if( ret )
    {
        PCRMem_Set(ret,0,sizeof(DOFOperation_t)+sizeof(RoutingChange_t)+sizeof(DOFRouteInfo_t));

        opRci = (RoutingChange_t*)(((uint8*)ret) + sizeof(DOFOperation_t));
        opRci->route_info = (DOFRouteInfo)(((uint8*)ret) + sizeof(DOFOperation_t) + +sizeof(RoutingChange_t));
        if( DOFRuntime_FillRouteInfoFromRoutingInfo( opRci->route_info, rci->route_info) )
        {
            opRci->type = rci->type;
            opRci->request = NULL;

            ret->cancelContext = DOFOperationManager_CreateCancelContext(ret);
            if ( ret->cancelContext )
            {
                ret->opID = DOFOperationID_CreateLocal();
                if( ret->opID )
                {
                    ret->operation_will_close = OALObservable_Create();
                    if( ret->operation_will_close )
                    {
                        ret->operation_type = DOFOPERATIONTYPE_ROUTEINFO;
                        ret->context = opRci;
                        ret->ref_count = 1;
                        return ret;
                    }
                    DOFOperationID_Destroy(ret->opID);
                }
                DOFMem_Free(ret->cancelContext);
            }
            DOFRuntime_ClearRouteInfo( opRci->route_info );
        }
        DOFMem_Free(ret);
    }
    return NULL;
}

static boolean internal_traverse_routing_callbacks(OALSListNode item_, void* arg_)
{
    RoutingInfoChange_t* item = (RoutingInfoChange_t*)item_;
    RoutingChange_t* arg = (RoutingChange_t*)arg_;
    PCR_Assert(item->callback);
    PCR_Assert(item->callback->fns);
    PCR_Assert(item->callback->fns->RouteStatus);

    if (DOFSecurityScope_IsCompatible( OALChannel_GetSecurityScope( DOFSystem_GetChannel( item->system ), FALSE ),
                                       arg->route_info->securityScope ))
        item->callback->fns->RouteStatus( item->callback, item->system, arg->route_info, arg->type );
    return FALSE;
}

void DOFRuntime_HandleRouteInfo( DOFOperation op )
{
    RoutingChange_t* rci;
    DOFOperation_Lock(op);
    rci = op->context;
    if( routingCallbacks )
    {
        OALSList_Traverse( ((OALSListNode)routingCallbacks), internal_traverse_routing_callbacks, (void*)rci );
    }
    DOFRouteInfo_Empty(rci->route_info);
    DOFOperation_Unlock(op);
    DOFOperation_Cancel( op );
}

typedef struct OperationContext_t
{
    OALEnqueueItem_t super;
    DOFOperation operation;
} OperationContext_t;

static void handleRouteInfoListenerOp( OALThreadPoolItem dnode, boolean shutdown )
{
    PCR_Assert(dnode != NULL);
    if( !shutdown )
    {
        OperationContext_t * node = (OperationContext_t*) dnode;
        DOFOperation rio = node->operation;
        DOFRuntime_HandleRouteInfo( rio );
        DOFOperation_Internal_Destroy(rio);
    }
}

static boolean internal_traverse_routing_callbacks_handle(OALSListNode item_, void* arg_)
{
    RoutingChange_t* rci = (RoutingChange_t*)arg_;
    DOFOperation op = DOFOperation_CreateRoutingChange( rci );
    if( op )
    {
        OperationContext_t * node = (OperationContext_t*) DOFMem_Alloc(sizeof(OperationContext_t));
        if(node != NULL)
        {
            OALEnqueueItem_Init((OALEnqueueItem) node, handleRouteInfoListenerOp);
            node->operation = op;
            if(!OALEnqueue_Enqueue( runtime_enqueue, (OALEnqueueItem) node))
            {
                DOFOperation_Internal_Destroy( op );
                DOFMem_Free(node);
            }
        }
        else
        {
            DOFOperation_Internal_Destroy( op );
        }
    }
    return FALSE;
}

void DOFRuntime_RoutingInfoChanged( provider_binding_table_entry_t *pbte, DOFRouteInfoState change )
{
    if( pbte )
    {
        if( routingCallbacks )
        {
            DOFRouteInfo ri = DOFRuntime_FillRouteInfo( pbte, change );

            if( ri )
            {
                RoutingChange_t rci;

                rci.route_info = ri;
                rci.type = (uint8)change;
                rci.request = NULL;
                OALSList_Traverse( ((OALSListNode)routingCallbacks), internal_traverse_routing_callbacks_handle, (void*)&rci);

                DOFRouteInfo_Destroy(ri);
            }
        }
    }
}

/**
@}
**/
