/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/error.h>
#include <internal/addressinternal.h>
#include <internal/mutex_pool.h>
#include <internal/remotechannel.h>
#include "securegroupprivate.h"


/** @addtogroup DOFSecureGroup
@{
**/


/*
**
**  "Private" Type and Method Declarations
*/

/* @internal

   @note typedef int (*OALTimer_fn_t)(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown);
*/
static int PointTimeout_fn( OALTimerItem item, uint32 argc, void *argv[], boolean shutdown )
{
    boolean              deleted;
    PCRMutex             mutex;
    pointAddressInfo_t * pPointAddressInfo = argv[0];

    PCR_Assert( argc == 1 );
    PCR_Assert( pPointAddressInfo );
    PCR_Assert( pPointAddressInfo->address );
    PCR_Assert( pPointAddressInfo->secureGroup );
    PCR_Assert( pPointAddressInfo->secureGroup->flags.hub );
    PCR_Assert( pPointAddressInfo->timerItem );
    mutex = OALMutexPool_Get( pPointAddressInfo->secureGroup );
    PCRMutex_Lock( mutex );
    PCR_Assert( pPointAddressInfo->secureGroup->hubPointTree );
    deleted = OALAVLTree_Delete( pPointAddressInfo->secureGroup->hubPointTree,
                                 (OALAVLTreeKey_t)&pPointAddressInfo->address );
    PCRMutex_Unlock( mutex );
    OALMutexPool_Release( pPointAddressInfo->secureGroup );
    if (deleted)
    {
        OALRemoteChannel_PeerListenerEvent((OALChannel)pPointAddressInfo->secureGroup, pPointAddressInfo->address, NULL, FALSE);

        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        {
            /* Though it would be nice to add to this message the cipher nodeID, we do not have that information. */
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                "Point timed out- hub removed stored DOFAddress from hubPointTree." );
        }
    }
    else
    {
        /* Odd that the tree entry is gone- who removed it? Whoever did presumably destroyed the DOFAddress associated
           with the tree entry. That has nothing to do with our destroy below of pPointAddressInfo->address. */
        /* Are we confident enough to assert that deleted is TRUE, then we wouldn't have this quandary to consider. */
    }
    /* Cleanup. */
    DOFAddress_Destroy( pPointAddressInfo->address );                       /* From the _Copy() of our caller. */
    /* Presumably this is the last reference to this DOFAddress- if so cipher's address observer will be notified
       and cipher will remove this point DOFAddress from his list. */
    OALTimerItem_Destroy( pPointAddressInfo->timerItem, FALSE );     /* Our timeout removed timerItem from the timer. */
    DOFMem_Free( pPointAddressInfo );

    return 0;
}


/*
**
**  "Public" Type and Method Declarations
*/


/* @internal
    Called by hub when answering a ReqGroup by sending back a ReqGroupResponse, or
           by any secure packet coming in that is delivered to the hub mux.
    If point's @a receivedAddr is already in hubPointTree, give it full timeout,
                             else if @a addNewEntry TRUE, add point's @a receivedAddr to hubPointTree with full timeout.

    @param secureGroup
    @param receivedAddr
    @param addNewEntry
    @param hisTransmittedAPPID Comes from incoming RequestGroup PDU, used only when addNewEntry is TRUE.
    @return TRUE for update happened, FALSE otherwise.
*/
boolean DOFHub_UpdateAddressTable( DOFSecureGroup secureGroup, const DOFAddress receivedAddr, boolean addNewEntry,
                                   uint16 hisTransmittedAPPID )
{
    DOFAddress copiedAddr;
    boolean    success = FALSE;

    PCR_Assert( secureGroup );
    PCR_Assert( secureGroup->flags.hub );
    PCR_Assert( receivedAddr );
    copiedAddr = DOFAddress_Copy( receivedAddr );
    if (copiedAddr)
    {
        pointAddressInfo_t * pPointAddressInfo;
        PCRMutex             mutex = OALMutexPool_Get( secureGroup );

        PCR_Assert( secureGroup->hubPointTree );
        PCRMutex_Lock( mutex );
        pPointAddressInfo = (pointAddressInfo_t *)OALAVLTree_Retrieve( secureGroup->hubPointTree,
                                                                       (OALAVLTreeKey_t)&copiedAddr );
        PCRMutex_Unlock( mutex );
        if (pPointAddressInfo)
        {
            /* Update the entry held in ->hubPointTree. Update its ->address and its ->timerItem timeout. */
            /* Detail: Eliminate ancient DOFAddr structure equivalent to receivedAddr, then put back in our fresh
               structure. Needed because cipher looks up targetNodeID by physical address of the DOFAddress structure.*/
            DOFAddress_Destroy( pPointAddressInfo->address ); /* We retained it before, so release our hold on it.*/
            pPointAddressInfo->address = copiedAddr;
            PCR_Assert( pPointAddressInfo->timerItem );
            if (OALTimerItem_Insert( pPointAddressInfo->timerItem, secureGroup->hubConfiguredPointTimeout, FALSE,
                                     FALSE ) == DOFERROR_OK)
            {
                success = TRUE;
                /*With OAP traffic in play this is too annoying place to put a log message- every OAP PDU gets us here*/
            }
            else
            {
                /* We were unable to extend the timeout, but the hub will continue to receive from the point and it
                   may get extended next time it receives something from the point. */
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG,
                                                        "DOFHub_UpdateAddressTable() failed timer item insert." );
            }
        }
        else if ((pPointAddressInfo = DOFMem_Alloc( sizeof( pointAddressInfo_t ) )) != NULL)
        {
            OALTimerItem timerItem = OALTimer_CreateItem( secureGroup->timer, PointTimeout_fn, 1,
                                                          (void**)&pPointAddressInfo );

            if (timerItem != NULL)
            {
                if (OALTimerItem_Insert(timerItem, secureGroup->hubConfiguredPointTimeout, FALSE, FALSE) == DOFERROR_OK)
                {
                    /* Initialize all elements of key before inserting. ALL filled so no need to memset the structure.*/
                    pPointAddressInfo->address = copiedAddr;
                    pPointAddressInfo->secureGroup = secureGroup;
                    pPointAddressInfo->timerItem = timerItem;
                    pPointAddressInfo->hisAPPID = hisTransmittedAPPID;
                    PCRMutex_Lock( mutex );
                    success = OALAVLTree_Insert(secureGroup->hubPointTree, (OALAVLTreeKey_t)&copiedAddr,
                                                (OALAVLTreeItem_t)pPointAddressInfo, NULL ) != OALAVLTree_InsertFailure;
                    PCRMutex_Unlock( mutex );
                    if (success)
                    {
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                                     "Hub adds new DOFAddress to hubPointTree of point timeout (msec) ",
                                                       (uint32)secureGroup->hubConfiguredPointTimeout );
                    }
                    else
                        OALTimerItem_Remove( timerItem );
                }
                if (!success)
                    OALTimerItem_Destroy( timerItem, TRUE );
            }
            if (!success)
                DOFMem_Free( pPointAddressInfo );
        }
        if (!success)
            DOFAddress_Destroy( copiedAddr );
        OALMutexPool_Release( secureGroup );
    }
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ) && !success)
        DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_TRACE, "Hub FAILED to add received DOFAddress to hubPointTree.");

    return success;
}


/**
@}
**/
