/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/bridgeinternal.h>
#include <internal/connectionless_remotechannel.h>
#include <internal/exception.h>
#include <internal/mutex_pool.h>
#include <internal/oalstring.h>
#include <internal/remotechannel.h>
#include <internal/remotechannelmanager.h>
#include <internal/security_common.h>
#include <internal/serverinternal.h>
#include <internal/staticassertinternal.h>
#include <internal/timeinternal.h>
#include "securegroupprivate.h"

/** @addtogroup DOFSecureGroup
@{
**/

/*
**
**  "Private" Type and Method Declarations
*/


#define INITIALJOIN_TOTAL_ATTEMPT_COUNT     (2)    /* Spec says multiple: we try twice. */
#define SGM_RANK_HUB                        (255)  /* Spec says applying this value should not matter to a hub. */
#define SGM_RANK_POINT                      (0)    /* Spec says applying this value should not matter to a point. */

/* Fill this list with every SGMP APPID that we have implemented.
        Every SGMP APPID less than our highest capability must be in the list (i.e. we must implement every such APPID).
        Each entry must be placed in increasing order.
        Absolute maximum entry count is 32. Otherwise some asserts can be hit. */
uint16 IMPLEMENTED_SGMP_APPID_LIST[] =
{
    SGMP_V1_APPID,                          /* SGMP_V1_APPID (130) must be in this list, and must be first entry. */
                                            /* RequestGroup PDU will ONLY be sent at the highest level in this list. */
};
/* NOTE: Number of entries in this list MUST not be 0 and MUST match IMPLEMENTED_SGMP_PROTOCOLS_COUNT. */
STATIC_ASSERT( IMPLEMENTED_SGMP_PROTOCOLS_COUNT > 0 );
STATIC_ASSERT( IMPLEMENTED_SGMP_PROTOCOLS_COUNT == lengthof( IMPLEMENTED_SGMP_APPID_LIST ) );
STATIC_ASSERT( IMPLEMENTED_SGMP_APPID_LIST_LENGTH > 0  );
STATIC_ASSERT( IMPLEMENTED_SGMP_APPID_LIST_LENGTH <= 32  );

typedef struct routeToProviderInfo_t
{
    OALChannel   channel;
    DOFOperation reqOp;
    DOFOperation provOp;
} routeToProviderInfo_t;


uint32 myHighestAPPIDBit;


static const char secureGroupLogString[] = "[DOFSecureGroup] ";


/* @internal */
static void PrepLogOutString( char * pOutString, uint32 sizeOutString )
{

    PCRMem_Set( pOutString, 0, sizeOutString );
    PCRString_Copy( pOutString, secureGroupLogString, PCRString_Length( secureGroupLogString ) );
}


/* @internal
    Disable the given secureGroup's secure group behavior, and cleanup its structure members.

    @note This does not destroy secureGroup structure.
          That is done by DOFSecureGroup_Release() and DOFSecureGroup_Dtor().
    @param secureGroup The given secureGroup.
*/
static void DisableAndCleanup( DOFSecureGroup secureGroup )
{
    PCR_Assert( secureGroup );
    PCR_Assert( ((OALChannel)secureGroup)->remoteChannelData );
    DOFSecureGroup_DenyGroupBehavior( secureGroup );
    if (((OALChannel)secureGroup)->remoteChannelData->cipher)
        OALCipher_ClearSecureGroupNotification( ((OALChannel)secureGroup)->remoteChannelData->cipher, secureGroup );
    /* Scrub KEK on purpose. This is in lieu of sending a LeaveGroup PDU (which does not exist). */
    PCRMem_Set( &secureGroup->keks[0].KEK, 0, sizeof( OALCryptKey ) );
    PCRMem_Set( &secureGroup->keks[1].KEK, 0, sizeof( OALCryptKey ) );
}


/** @internal
   Initialize secureGroup behavior for the given secureGroup.

   @param secureGroup The given secureGroup.
   @return TRUE if group behavior is already in place or for successfully initiated, or
           FALSE if resources could not be allocated.
**/
static boolean InitiateMyGroupBehavior( DOFSecureGroup secureGroup )
{
    /*We do not lock secureGroup mutex because secureGroup machine is not up yet; other threads are not making changes*/
    PCR_Assert( secureGroup );
    if (!secureGroup->timer)
    {
        PCR_Assert( secureGroup->keks[secureGroup->flags.fut ^ 1].filled );
        PCR_Assert( secureGroup->keks[secureGroup->flags.fut ^ 1].Thb );
        secureGroup->sendHeartbeatTimeout = secureGroup->keks[secureGroup->flags.fut ^ 1].Thb * 1000;
#if (SCALE_THB > 1)
        secureGroup->sendHeartbeatTimeout = DOFSecureGroup_ScaleHeartbeatTimeout( secureGroup,
                                                                                  secureGroup->sendHeartbeatTimeout );
#endif
        if ((secureGroup->timer = OALTimer_Create(DOF_GetDof()->threadPool, TRUE)) != NULL)
        {
            PCR_Assert( secureGroup->workEnqueue );
            PCR_Assert( !secureGroup->heartbeatTimerItem );
            secureGroup->heartbeatTimerItem = OALTimer_CreateItem( secureGroup->timer,
                                                                   DOFSecureGroup_HeartbeatTimeout_fn, 1,
                                                                   (void**)&secureGroup );
            PCR_Assert( !secureGroup->getKEKTimerItem );
            secureGroup->getKEKTimerItem = OALTimer_CreateItem( secureGroup->timer, DOFSecureGroup_GetKEKTimeout_fn,
                                                                1, (void**)&secureGroup );
            PCR_Assert( !secureGroup->grantTimerItem );
            secureGroup->grantTimerItem = OALTimer_CreateItem( secureGroup->timer, DOFSecureGroup_GrantTimeout_fn,
                                                               1, (void**)&secureGroup );
            PCR_Assert( !secureGroup->epochCompleteTimerItem );
            secureGroup->epochCompleteTimerItem = OALTimer_CreateItem( secureGroup->timer,
                                                                       DOFSecureGroup_EpochCompleteTimeout_fn,
                                                                       1, (void**)&secureGroup );
            PCR_Assert( !secureGroup->slotRekeyTimerItem );
            secureGroup->slotRekeyTimerItem = OALTimer_CreateItem( secureGroup->timer,
                                                                   DOFSecureGroup_SlotRekeyTimeout_fn, 1,
                                                                   (void**)&secureGroup );
            PCR_Assert( secureGroup->keks[DOFSecureGroup_GetOpKEK( secureGroup )].Tmax );
            if (secureGroup->heartbeatTimerItem && secureGroup->getKEKTimerItem && secureGroup->grantTimerItem &&
                secureGroup->epochCompleteTimerItem && secureGroup->slotRekeyTimerItem &&
                DOFSecureGroup_SetEpochTimeout( secureGroup,
               (secureGroup->keks[DOFSecureGroup_GetOpKEK(secureGroup)].Tmax*SGMP_TMAX_MSEC_MULTIPLIER/SCALE_TMAX_TMIN),
                                                FALSE ) /* FALSE because we do not have ->grantTimerItem inserted yet.*/
                                                        != FALSE &&
                DOFSecureGroup_SetSlotRekeyTimeout( secureGroup, SLOTREKEY_TIMEOUT ) != FALSE &&
                (secureGroup->flags.pointMember || secureGroup->flags.hub || /* point, hub NOT enable hb timeout. */
                 DOFSecureGroup_ExtendHeartbeatTimeout( secureGroup ) != FALSE))
                {
                    PCR_Assert( !secureGroup->flags.sgmScheduleSecondSlotRekey );
                    return TRUE;                                                /* Success. */
                }
            /* Error out. */
            OALTimerItem_Destroy( secureGroup->heartbeatTimerItem, TRUE );
            secureGroup->heartbeatTimerItem = NULL;
            OALTimerItem_Destroy( secureGroup->getKEKTimerItem, TRUE );
            secureGroup->getKEKTimerItem = NULL;
            OALTimerItem_Destroy( secureGroup->grantTimerItem, TRUE );
            secureGroup->grantTimerItem = NULL;
            OALTimerItem_Destroy( secureGroup->epochCompleteTimerItem, TRUE );
            secureGroup->epochCompleteTimerItem = NULL;
            OALTimerItem_Destroy( secureGroup->slotRekeyTimerItem, TRUE );
            secureGroup->slotRekeyTimerItem = NULL;
            OALTimer_Destroy( secureGroup->timer );
            secureGroup->timer = NULL;                                         /* All the NULLs in case we try again. */
        }
        return FALSE;                                                          /* Error out. */
    }
    PCR_Assert( secureGroup->flags.pointMember || secureGroup->heartbeatTimerItem );

    return TRUE;                           /* Pause to ask: Why initiating group behavior again if already initiated? */
}


/* @internal */
static void CleanupStoredPointAddresses(OALAVLTreeItem_t pPointAddressInfo, void * arg )
{
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                        "At SecureGroup_Dtor(), hub removes one stored DOFAddress from hubPointTree." );
    DOFAddress_Destroy( ((pointAddressInfo_t*)pPointAddressInfo)->address );
    DOFMem_Free( ((pointAddressInfo_t*)pPointAddressInfo) );
}


/* Called in DOFSecureGroup_RouteToProvider() by OALAVLTree_Traverse() for as many pointAddressInfo_t * items as are
   held in ->hubPointTree. At each call, pStoredAddressInfoItem is filled with the next item in the tree.

    @note Our function proto represents:
                                   boolean (*OALAVLTree_Traversal_fn_t)( pointAddressInfo_t * pStoredAddressInfoItem,
                                                                         routeToProviderInfo_t * pRouteToProviderInfo );
*/
static boolean RouteToProviderFromHubPointTree( OALAVLTreeItem_t item, void * arg )
{
    const pointAddressInfo_t * pStoredAddressInfoItem;
    routeToProviderInfo_t *    pRouteToProviderInfo;

    PCR_Assert( item );
    pStoredAddressInfoItem = item;
    PCR_Assert( pStoredAddressInfoItem->address );                   /* We are to send unicast. */
    PCR_Assert( pStoredAddressInfoItem->secureGroup );
    PCR_Assert( pStoredAddressInfoItem->secureGroup->flags.hub );
    PCR_Assert( arg );
    pRouteToProviderInfo = arg;
    PCR_Assert( pRouteToProviderInfo->channel );
    PCR_Assert( pRouteToProviderInfo->reqOp );
    PCR_Assert( !pRouteToProviderInfo->provOp );                     /* Why we are here in the 1st place. */
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                   "Hub re-issues multicast request as unicast request to one entry in hubPointTree." );
    OALRemoteChannel_RouteToProviderLossy( pRouteToProviderInfo->channel, pRouteToProviderInfo->reqOp,
                                           pRouteToProviderInfo->provOp, pStoredAddressInfoItem->address );
    return FALSE;
}


static void SecureGroup_Dtor( OALChannel chan )
{
    DOFSecureGroup self = (DOFSecureGroup)chan;

    PCR_Assert( self );
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "SecureGroup_Dtor() called." );
    /* These next two asserts show that secureGroup's Process_Shutdown() was already called, which is required. */
    PCR_Assert( chan->remoteChannelData->flags.isInShutdownState );
    PCR_Assert( !self->timer );
    /* ->workEnqueue was already destroyed in Process_Shutdown() or DOFSecureGroup_Create(). */
    OALRemoteChannelMgr_RemoveSilenceTimerItems( chan );
    DOFAddress_Destroy( self->lastReceivedSGMAddress );
    DOFAddress_Destroy( self->lastKEKRekeySGMAddress );
    DOFObjectIDAuthentication_Destroy( self->group );
    OALMutexPool_Finalize( self->extendPermissionsFIFO );
    OALQueue_Destroy( self->extendPermissionsFIFO );
    OALSList_RemoveAll( (OALSListNode *)&self->epochWorkList, DOFSecureGroup_DeallocateInList,
                        NULL );
    PCR_Assert( !self->flags.hub || self->hubPointTree );
    PCR_Assert( self->flags.hub || !self->hubPointTree );
    if (self->hubPointTree)
        OALAVLTree_Destroy( self->hubPointTree, CleanupStoredPointAddresses, NULL );
    PCREvent_Destroy( self->specialEvent );
    PCRMutex_Destroy( self->reqGRespMutex );
    OALMutexPool_Finalize( &self->reqKEvent );
    OALChannel_Dtor( chan );                                                /* Call parent class destructor. */
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "SecureGroup_Dtor() exits." );
}


static boolean SetupSGM( DOFSecureGroup secureGroup, OALCipher cipher )
{
    boolean success;
    uint8 opkek = DOFSecureGroup_GetOpKEK( secureGroup );

    OALCipher_SetManager( cipher, TRUE );
    success = OALCipher_FormNewAndSetSlotOrKEKKey( cipher, 0, KEKKEY, secureGroup->keks[opkek].epoch, secureGroup->keks[opkek].securityModeConfig, secureGroup->keks[opkek].myPerms, secureGroup->keks[opkek].otherPerms, secureGroup );
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, success ? "Stored slotKey in SetupSGM.":
                                                                         "_SetSlotKey() not accepted in SetupSGM." );
    if (success)
    {
        PCR_Assert( OALCipher_IsNodeIDValid( cipher ) ); /* As SGM, we self assigned. */
        DOFSecureGroup_InitiateMySGMBehavior( secureGroup, 0, FALSE );
        return TRUE;
    }

    return FALSE;
}


static boolean SetupMember( DOFSecureGroup secureGroup, OALCipher cipher, DOFGroupKey_t * pFillGroupKey )
{
    uint8 opkek = DOFSecureGroup_GetOpKEK( secureGroup );
    PCR_Assert( !secureGroup->flags.hub );
    PCR_Assert( pFillGroupKey->filled );
    PCR_Assert( pFillGroupKey->streamCcmModeInitialState );
    PCR_Assert( !OALCipher_IsNodeIDValid( cipher ) );       /* Next call fills nodeID from streamCcmModeInitialState. */
    if (OALCipher_SetKEKKey( cipher, &secureGroup->lastReceivedSGMAddress,
                             pFillGroupKey->streamCcmModeInitialState, pFillGroupKey->key,
                             0, secureGroup->keks[opkek].epoch, secureGroup->keks[opkek].securityModeConfig,
                             secureGroup->keks[opkek].myPerms, secureGroup->keks[opkek].otherPerms, secureGroup, pFillGroupKey->receivedMulticast ) == TRUE)
    {
        /* NOTE: "simple takeover" as SGM is not possible at this moment, because we do not yet have cipher's advertised
           next nodeID- it comes to cipher only when we receive a secure PDU from the SGM. Former code here that checked
           for simple takeover is now removed. On receiving each future SGMP HB we will check for simple takeover. */
        if (!OALCipher_IsNodeIDValid( cipher ))
            OALCipher_SecureGroupGetNewNodeID( DOFSecureGroup_GetChannel_Internal(secureGroup) ); /*SGM gave no nodeID*/
        if (!secureGroup->flags.pointMember)
        {
            if (DOFSecureGroup_ExtendHeartbeatTimeout( secureGroup ) == FALSE)
            {
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG, "Extend heartbeat error in SetupMember()." );
                return FALSE;                                                                           /* Error out. */
            }
        }
        return TRUE;
    }
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG, "Set cur key error in SetupMember." );

    return FALSE;                                                                    /* Error out. */
}


/*
**  "Internal Public"  Methods.
*/


/* @internal. */
OALChannel DOFSecureGroup_GetChannel( DOFSecureGroup secureGroup )
{
    return DOFSecureGroup_GetChannel_Internal( secureGroup );
}


/* @internal. */
void DOFSecureGroup_NewSecurityScopeToChannelSecurityScope( DOFSecurityScope newSecurityScope, OALChannel channel )
{
    /* We require caller to decide whether to lock channel's mutex, or not. */
    DOFSecurityScope oldScope;

    /* We don't use the channel getter call because secureGroup use does not involve the DOFOperationBridgeConfig.
       We stuff and access channel's securityScope directly. */
    PCR_Assert( channel );
    PCR_Assert( newSecurityScope );
    newSecurityScope = DOFSecurityScope_Copy( newSecurityScope );
    PCR_Assert( newSecurityScope );                /* The copy will NEVER be NULL if NULL is not supplied. */
    oldScope = channel->securityScope;
    channel->securityScope = newSecurityScope;
    DOFSecurityScope_Destroy( oldScope );
}


/* @internal. */
void DOFSecureGroup_NewSecureSIDToRemoteChannel( DOFObjectIDSource newSID, OALChannel channel )
{
    /* We require caller to decide whether to lock channel's mutex, or not. */
    sid_t * pNewSid;

    PCR_Assert( channel );
    PCR_Assert( channel->remoteChannelData );
    PCR_Assert( newSID );

    if (channel->remoteChannelData->secureCredentials)
    {
        pNewSid = create_sid_from_oid( newSID, DOFCredentials_GetDomainID(channel->remoteChannelData->secureCredentials) );
    }
    else
    {
        pNewSid = create_sid_from_oid( newSID, NULL );
    }

    if (pNewSid != NULL)
    {
        sid_t * pOldSid = channel->remoteChannelData->secureSid;

        channel->remoteChannelData->secureSid = pNewSid;
        if (pOldSid)
            source_release( pOldSid );
    }
    else
    {
        /* Is there anything that is required here? */
    }
}


/** @internal

   @param appID
**/
uint32 DOFSecureGroup_ComputeAPPIDVersionBit( uint16 appID )
{
    uint32 i;
    uint32 computedBit = 0x0001;

    for (i = 0; i < IMPLEMENTED_SGMP_APPID_LIST_LENGTH; i++)
    {
        if (appID == IMPLEMENTED_SGMP_APPID_LIST[i])
            return computedBit;
        computedBit <<= 1;
    }

    return 0;                                         /* appID not in our list. */
}


/* The DOFRouter calls this when handed locally originated requests.
       This not called when request comes in across the wire, UNLESS. */
void DOFSecureGroup_RouteToProvider( OALChannel channel, DOFOperation req_op, DOFOperation prov_op, DOFAddress addr )
{
    DOFSecureGroup secureGroup = (DOFSecureGroup)channel;

    PCR_Assert( channel );
    PCR_Assert( req_op );
    if (!secureGroup->flags.hub || prov_op)
        OALRemoteChannel_RouteToProviderLossy( channel, req_op, prov_op, addr ); /* Do the normal thing. */
    else
    {
        /* Special work because we are hub. Send this request to every point in hubPointTree. */
        PCRMutex              mutex;
        routeToProviderInfo_t routeToProviderInfo;

        routeToProviderInfo.channel = channel;                          /* No memset, all 3 members are filled. */
        routeToProviderInfo.reqOp = req_op;
        routeToProviderInfo.provOp = prov_op;
        mutex = OALMutexPool_Get( secureGroup );
        PCRMutex_Lock( mutex );
        PCR_Assert( secureGroup->hubPointTree );
        OALAVLTree_Traverse( secureGroup->hubPointTree, RouteToProviderFromHubPointTree, &routeToProviderInfo );
        PCRMutex_Unlock( mutex );
        OALMutexPool_Release( secureGroup );
    }
}



/* @internal
    Deny given secureGroup's transmit/receive capability.

    @note Does nothing for FALSE return of DOFSecureGroup_NowUpOrNeverUp().
    @note Afterward, restoring capability requires successful DOFConnection_Connect() call.
    @note This does not cleanup secureGroup struct members. That is done by DisableAndCleanup().
    @param secureGroup The given secureGroup
*/
void DOFSecureGroup_DenyGroupBehavior( DOFSecureGroup secureGroup )
{
    boolean nowUpOrNeverUp;

    /* Called by DisableAndCleanup() or ProcessShutdown(). Note: can be called early by InitialJoin() through
       DisableAndCleanup() if it fails to join and before trying to join a 2nd time. */
    PCR_Assert( secureGroup );
    PCR_Assert( ((OALChannel)secureGroup)->remoteChannelData );
    nowUpOrNeverUp = DOFSecureGroup_NowUpOrNeverUp( secureGroup );
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessage(DOFLOGLEVEL_TRACE, "DenyGroupBehavior called with nowUpOrNeverUp: ", nowUpOrNeverUp);
#endif
    if (nowUpOrNeverUp)
    {
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
        boolean eventToSignalExists;
#endif

        secureGroup->up = FALSE;
        DOFSecureGroup_ChangeSGMState( secureGroup, !BE_SGM ); /*Redundant to not be SGM if sg disabled; here for form*/

        OALMutexPool_Lock( secureGroup->reqGRespMutex );
        /* RequestGroupCallback(), AcceptHighestPriorityRequestGroupResponse() also lock this mutex. */
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                       "DenyGroupBehavior has reqG mutex lock, finds ->reqGRespEvent 0x",
                                       secureGroup->reqGRespEvent ? 1 : 0 );
        eventToSignalExists = secureGroup->reqGRespEvent ? TRUE : FALSE;
        /*->reqGRespEvent filled in IssueRequestGroupPDU(), NULL'ed at end AcceptHighestPriorityRequestGroupResponse()*/
#endif
        /* If AcceptHighestPriorityRequestGroupResponse() is waiting for ReqG response PDU, signal it. */
        if (secureGroup->reqGRespEvent)
        {
            PCR_Assert( !secureGroup->up );
            /*Tell AcceptHighestPriorityRequestGroupResponse to stop waiting for response and to cancel original reqOp*/
            PCREvent_Signal( secureGroup->reqGRespEvent );
        }
        OALMutexPool_Unlock( secureGroup->reqGRespMutex );
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, eventToSignalExists ?
              "DenyGroupBehavior() signaled AcceptHighestPriorityRequestGroupResponse() to stop waiting for response." :
              "DenyGroupBehavior() did NOT need to signal AcceptHighestPriorityRequestGroupResponse()." );
#endif

        OALMutexPool_Lock( &secureGroup->reqKEvent );           /* GetKEKCallback(), GetKEKUsingTRP() also lock this. */
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                       "DenyGroupBehavior has reqK mutex lock, finds ->reqKCallbackEvent 0x",
                                       secureGroup->reqKEvent.reqKCallbackEvent ? 1 : 0 );
        eventToSignalExists = secureGroup->reqKEvent.reqKCallbackEvent ? TRUE : FALSE;
#endif

        /* If GetKEKUsingTRP() is waiting for get KEK response PDU, signal it. */
        if (secureGroup->reqKEvent.reqKCallbackEvent)
        {
            PCR_Assert( !secureGroup->up );
            /* Tell GetKEKUsingTRP() to stop waiting for response. */
            PCREvent_Signal( secureGroup->reqKEvent.reqKCallbackEvent );
        }
        OALMutexPool_Unlock( &secureGroup->reqKEvent );
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, eventToSignalExists ?
                                         "DenyGroupBehavior() signaled GetKEKUsingTRP() to stop waiting for response." :
                                         "DenyGroupBehavior() did NOT need to signal GetKEKUsingTRP()." );
#endif
    }
}


void DOFSecureGroup_CleanupSecureGroupTimer( DOFSecureGroup secureGroup )
{
    OALTimerItem tempHbItem;
    OALTimerItem tempGkItem;
    OALTimerItem tempGItem;
    OALTimerItem tempEcItem;
    OALTimerItem tempSrItem;
    OALTimer     tempTimer;
    PCRMutex     mutex;
    boolean      act = FALSE;

    PCR_Assert( secureGroup );
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "CleanupSecureGroupTimer() called." );
#endif
    mutex = OALMutexPool_Get( secureGroup );
    PCRMutex_Lock( mutex );
    if (secureGroup->timer)
    {
        act = TRUE;
        tempHbItem = secureGroup->heartbeatTimerItem;
        tempGItem  = secureGroup->grantTimerItem;
        tempEcItem = secureGroup->epochCompleteTimerItem;
        tempSrItem = secureGroup->slotRekeyTimerItem;
        tempGkItem = secureGroup->getKEKTimerItem;
        tempTimer = secureGroup->timer;
        secureGroup->heartbeatTimerItem = NULL;
        secureGroup->grantTimerItem = NULL;
        secureGroup->epochCompleteTimerItem = NULL;
        secureGroup->slotRekeyTimerItem = NULL;
        secureGroup->getKEKTimerItem = NULL;
        secureGroup->timer = NULL;
    }
    PCRMutex_Unlock( mutex );
    OALMutexPool_Release( secureGroup );

    if (act)
    {
        /* Now the five timer callbacks are no longer callable. */
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_TRACE, "CleanupSecureGroupTimer now destroys Hb item.");
#endif
        OALTimerItem_Destroy( tempHbItem, TRUE );
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "CleanupSecureGroupTimer now destroys G item." );
#endif
        /* First eliminate grant timer as an issue, so it cannot not fire in the middle of getKEK timer cleanup. */
        OALTimerItem_Destroy( tempGItem, TRUE );
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_TRACE, "CleanupSecureGroupTimer now destroys Ec item.");
#endif
        OALTimerItem_Destroy( tempEcItem, TRUE );
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_TRACE, "CleanupSecureGroupTimer now destroys Sr item.");
#endif
        OALTimerItem_Destroy( tempSrItem, TRUE );

#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_TRACE, "CleanupSecureGroupTimer now destroys Gk item.");
#endif
        OALTimerItem_Destroy( tempGkItem, TRUE );

        /* Now all five timer callbacks are known to be complete. */
        PCR_Assert( tempTimer );
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "CleanupSecureGroupTimer now destroys timer." );
#endif
        OALTimer_Destroy( tempTimer );
    }
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "CleanupSecureGroupTimer now exits." );
#endif
}


/* @internal.

   @note This blocks until various TRP and then SGMP PDUs are answered.
   */
void DOFSecureGroup_InitialJoin( OALThreadPoolItem dnode, boolean shutdown )
{
    SecureGroupContext_t * node;
    uint8                  opKEK;
    DOFGroupKey_t          fillGroupKey;
    DOFCredentials         initialCreds;
    OALChannel             secureGroupChannel;
    boolean                joined = FALSE;
    DOFException           except = NULL;
    OALChannel             contextChannel;

    node = (SecureGroupContext_t *)dnode;
    PCR_Assert( node != NULL );
    PCR_Assert( node->secureGroup );
    PCR_Assert( ((OALChannel)node->secureGroup)->remoteChannelData );
    PCR_Assert(node->secureGroup->flags.pointMember || !((OALChannel)node->secureGroup)->remoteChannelData->flags.isUp);
    node->secureGroup->up = FALSE;
    secureGroupChannel = DOFSecureGroup_GetChannel_Internal( node->secureGroup );
    PCR_Assert( secureGroupChannel->remoteChannelData );
    PCR_Assert( secureGroupChannel->remoteChannelData->config.connection );
    initialCreds = secureGroupChannel->remoteChannelData->config.connection->credentials;
    PCR_Assert( initialCreds );

    if( !DOFCredentials_IsResolved( initialCreds ) )
    {
        DOFException except = NULL;

        /* If this function DOFSecureGroup_InitialJoin ever goes to not blocking this should be changed because DOFCredentials_Resolve Blocks */
        if( !DOFCredentials_Resolve( initialCreds, DOF_GetDof(), 25000, &except ) ||
            except )
        {
            DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_DEBUG, "InitialJoin() failed to resolve credentials.");
            goto disableCleanupAndOut;
        }
    }

    /* Keep node->secureGroup->creds the same as submitted, update remoteChannelData. */
    PCR_Assert( secureGroupChannel->remoteChannelData->secureCredentials == NULL );
    secureGroupChannel->remoteChannelData->secureCredentials = DOFCredentials_Copy( initialCreds );
    node->secureGroup->specialSuccessFlag = FALSE;                                /* Written TRUE on success. */
    contextChannel = OALChannel_Retain(secureGroupChannel);

    OALRemoteChannel_CreateLibPermissions( OALConnectionlessRemoteChannel_GetLibraryPermsCallback,
                                           contextChannel,
                                           secureGroupChannel->remoteChannelData->secureCredentials,
                                           secureGroupChannel, secureGroupChannel->remoteChannelData->address );
    /* _getLibraryPermsCallback sends channel message to our handler HandleMessage_GetLibraryPermsCallback(). */
    PCR_Assert( node->secureGroup->specialEvent );
    PCREvent_Wait( node->secureGroup->specialEvent, 25000 );
    /* 25 secs in case _CreateLibPermissions() TRP Validate PDU has some high timeout. Currently it is 10 secs*/
    if (!node->secureGroup->specialSuccessFlag)
    {
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                               "_InitialJoin: _CreateLibPermissions fails after initialCreds domainID match." );
        goto disableCleanupAndOut;
    }

    if (!node->secureGroup->flags.activeSGM)
    {
        uint8     joinAttemptCount;
        uint8     hisRank;
        uint16    hisHighestAPPID;
        OALCipher cipher;
        boolean   success;
        uint16    previousLoopObtainedEpoch = 0;

        for (joinAttemptCount = 0; joinAttemptCount < INITIALJOIN_TOTAL_ATTEMPT_COUNT; joinAttemptCount++)
        {
            DOFError status;

            opKEK = DOFSecureGroup_GetOpKEK( node->secureGroup );
            /* opKEK is what we fill and what JoinGroup() uses below. */
            if (joinAttemptCount)
                OALRemoteChannel_DeregisterSecureChannel( (OALChannel)node->secureGroup );       /* Stored last time. */
            DisableAndCleanup( node->secureGroup );               /* In case some old secureGroup is submitted to us. */
            PCR_Assert( node->secureGroup->keks[opKEK].filled == FALSE);

            /* GetKEK will use TRP to get a KEK from an authentication server.
               Note: this function blocks until it receives a KEK or a timeout.*/
            status = DOFSecureGroup_GetKEK( node->secureGroup, &node->secureGroup->keks[opKEK] );
            if (status != DOFERROR_OK)
            {
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "GetKEK() fail in InitialJoin." );
                goto disableCleanupAndOut;       /* status: _BAD_PARAM, _INT_ERR, _ACCESS_DENIED or _REQUEST_TIMEOUT. */
            }
            PCR_Assert( node->secureGroup->keks[opKEK].filled );

            node->secureGroup->beginGatherGroupStateTime = PCRTime_GetMS();
            PCR_Assert( node->secureGroup->keks[opKEK].securityScope );
            /* Note: Since our secureGroup machine is not up yet, no other changes to this channel's securityScope can
                     happen right now, and so we do not lock secureGroup mutex for this next call. */
            DOFSecureGroup_NewSecurityScopeToChannelSecurityScope( node->secureGroup->keks[opKEK].securityScope,
                                                                   (OALChannel)node->secureGroup );
            DOFSecureGroup_NewSecureSIDToRemoteChannel( node->secureGroup->keks[opKEK].sid,
                                                        (OALChannel)node->secureGroup );
            /* Use opKEK's just received stateID to setup our channel. */
            OALRemoteChannel_RegisterSecureChannel( (OALChannel)node->secureGroup,
                                                    node->secureGroup->keks[opKEK].stateID, TRUE );

            /*Returning here by going, as a fail path, to deregisterDisableCleanupAndOut shows no added valgrind probs*/

            if (joinAttemptCount && node->secureGroup->keks[opKEK].epoch == previousLoopObtainedEpoch)
                break;                                                     /* No 2nd try ReqG PDUs, epoch still same. */
            previousLoopObtainedEpoch = node->secureGroup->keks[opKEK].epoch;
            if (node->secureGroup->flags.hub)
            {
                PCR_Assert( !joined );
                break;
            }
            else
            {
                /* With channel setup we can now send a RequestGroup PDU and receive back a RequestGroupResponse PDU*/
                /* Note: This call will initialize the structure at &fillGroupKey. */
                joined = DOFSecureGroup_JoinGroup( node->secureGroup, (node->secureGroup->flags.pointMember != FALSE),
                                                   &fillGroupKey,           /* _JoinGroup() memsets this to zeroes. */
                                                   node->secureGroup->keks[opKEK].epoch, FALSE, &hisRank,
                                                   &hisHighestAPPID, REQUESTGROUP_SEND_TIMEOUT );
                /*Note: JoinGroup will take 2 seconds when we do not join, because we try twice and timeout is 1 second.
                        The earliest successful return happens, for each of two attempts, only after a full second of
                        watching for multiple possible returns and keeping only the one response with highest priority*/
                if (joined)
                {
                    PCR_Assert( fillGroupKey.filled );
                    PCR_Assert( fillGroupKey.streamCcmModeInitialState );
                    PCR_Assert( hisHighestAPPID >= SGMP_V1_APPID);                              /* SGMP checked this. */
                    break;
                }
                else
                {
                    /* opKEK valid but we will not use it's contents- cleanup whatever is left. */
                    DOFSecureGroup_CleanupKEKInfoAndDisable( &node->secureGroup->keks[opKEK] );
                    /* Also holds myPerms and otherPerms, but ownership transferred to secureGroup->remoteChannelData.*/
                }
            }
            /* If not joined, maybe I just missed getting a new KEK that the SGM did get and is using. Loop to try
               again by first getting a new KEK. */
        } /* end for(). */

        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "InitialJoin finds AS gives us encrypt 1/0: ",
                                    DOFSecurityModeCCM_IsEncrypt( node->secureGroup->keks[opKEK].securityModeConfig ) );
        if (!joined && (node->secureGroup->flags.pointMember || node->secureGroup->flags.doNotBeSGM))
        {
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                    "Not joined when configured for must join in InitialJoin." );
            goto deregisterDisableCleanupAndOut;
        }

        if (!InitiateMyGroupBehavior( node->secureGroup ))
        {
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_DEBUG, "Initiate group behavior error in InitialJoin.");
            goto deregisterDisableCleanupAndOut;
        }

        /* Extend grant from AS to Tmax. next getKEK is already scheduled above by InitiateMyGroupBehavior(). */
        PCR_Assert( node->secureGroup->keks[opKEK].Tmax );
        if (!DOFSecureGroup_SetGrantTimeout( node->secureGroup,
                                      (node->secureGroup->keks[opKEK].Tmax*SGMP_TMAX_MSEC_MULTIPLIER/SCALE_TMAX_TMIN) ))
        {
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG, "Extend grant error in InitialJoin." );
            goto deregisterDisableCleanupAndOut;
        }
        /* We ARE a group member. If we did not join an existing group above, we are not a point and are our own group*/
        if ((cipher = OALCipher_Create( node->secureGroup->keks[opKEK].securityModeConfig,
                               (CIPHER_CONFIG_SECUREGROUP_DEFAULT | (node->secureGroup->flags.hub ? CIPHER_IS_HUB : 0)),
                                        0,/*0: This nodeID param is not used. For secureGroup, nodeID assigns by rules*/
                                        (OALChannel)node->secureGroup, &except )) == NULL)
        {
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG, "Cipher create error in InitialJoin." );
            goto deregisterDisableAllCleanupAndOut;                                            /* Error out. */
        }
        PCR_Assert( !except );

        ((OALChannel)node->secureGroup)->remoteChannelData->cipher = cipher;
        /*Now that cipher is created we can take either of these next paths; both set cipher key and both fill success*/
        if (joined)
        {
            success = SetupMember( node->secureGroup, cipher, &fillGroupKey );
            OALStream_Release( fillGroupKey.streamCcmModeInitialState, TRUE );
        }
        else
        {
            /* We are SGM or hub- tell cipher he is manager, set a groupKey, and initiate other SGM behavior. */
            PCR_Assert( !node->secureGroup->flags.doNotBeSGM );
            success = SetupSGM( node->secureGroup, cipher );
            /* Wireshark cannot decode secure PDUs until it gets the key by seeing a (non-secure) ReqG response. No
               attempt to send a slot rekey PDU out here can help, because it is secure. */
            /* OK to transmit secure PDUs immediately, because key is new. */
            /* Nobody else has the key we just formed- nobody can receive our hb. So no need to send out early hb.
               And we disable sending this first hb by passing 0 for appidVersionBit. */
        }

        if (success)
        {
            node->secureGroup->up = TRUE;
            node->secureGroup->upSeen = TRUE;
            ((OALChannel)node->secureGroup)->remoteChannelData->flags.isUp = TRUE;
            OALRemoteChannel_NotifyInitComplete( (OALChannel)node->secureGroup );
            goto initialJoinSuccessOut;
        }
        else
            goto deregisterDisableAllKillCipherCleanupAndOut;
    }
    else
    {
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG, "InitialJoin called when wrongly set as SGM." );
        goto disableCleanupAndOut;
    }

deregisterDisableAllKillCipherCleanupAndOut:
    OALCipher_Destroy( ((OALChannel)node->secureGroup)->remoteChannelData->cipher );
    ((OALChannel)node->secureGroup)->remoteChannelData->cipher = NULL; /*SecureGroup_Dtor to find no cipher to destroy*/
deregisterDisableAllCleanupAndOut:
    DOFSecureGroup_CleanupSecureGroupTimer( node->secureGroup );
deregisterDisableCleanupAndOut:
    if (joined)
    {
        PCR_Assert( fillGroupKey.filled );
        OALStream_Release( fillGroupKey.streamCcmModeInitialState, TRUE );
    }
    OALRemoteChannel_DeregisterSecureChannel( (OALChannel)node->secureGroup );
    DOFSecureGroup_CleanupKEKInfoAndDisable( &node->secureGroup->keks[opKEK] );
disableCleanupAndOut:
    PCR_Assert( node->secureGroup->flags.pointMember ||
                !((OALChannel)node->secureGroup)->remoteChannelData->flags.isUp );
    DisableAndCleanup( node->secureGroup );
    PCR_Assert( node->secureGroup->up == FALSE );
    if (!except)
        except = DOFException_Copy( DOFEXCEPTION_FAILED );
    OALRemoteChannel_Shutdown( (OALChannel)node->secureGroup, except );
    /* Causes HandleMessage_Shutdown() to be called from DOFConnection_NotifyChannelDown(). The latter also releases
       OALChannelMgr_CreateChannel_Connection()'s ownership established early in DOFConnection_BeginConnect(). */
    DOFException_Destroy( except );
initialJoinSuccessOut:
    DOFMem_Free( node );
}


DOFSecurityScope DOFSecureGroup_GetSecurityScope( DOFSecureGroup self )
{
    return self ? self->keks[DOFSecureGroup_GetOpKEK( self )].securityScope : NULL;
}


#if (SCALE_THB > 1)
uint32 DOFSecureGroup_ScaleHeartbeatTimeout( DOFSecureGroup secureGroup, uint32 hbTimeout )
{
    uint32 retTimeout;

    /* We have scaling, arrange it so our minimum heartbeat separation is 2 seconds. */
    PCR_Assert( hbTimeout );
    PCR_Assert( SCALE_THB > 1 );
    retTimeout = hbTimeout / SCALE_THB;

    return (retTimeout >= 2000) ? retTimeout : 2000;
}
#endif


/* @internal */
boolean DOFSecureGroup_LogMessage( DOFLogLevel logLevel, const char * message, uint32 trailingValue )
{
    char outString[200]; /* Final output strips 1-7 leading ascii '0's from trailingValue. */

    /* 8 is character length of our uint32, when converted to hex. */
    if ((sizeof( secureGroupLogString ) + PCRString_Length( message ) + 8 ) <= sizeof( outString ))
    {
        char  trailingString[9];
        uint8 i;

        /* i marks earliest non ascii '0' of trailingString (hex string version of trailingValue). */
        OALString_Uint32ToHexString( trailingString, trailingValue );               /* Fills first 8 bytes. */
        trailingString[8] = 0;                                                      /* Terminate our C string. */
        for (i = 0; i < 7; i++)
        {
            if (trailingString[i] != '0')
                break;
        }
        PrepLogOutString( (char *)&outString, sizeof( outString ) );
        PCRString_Concat( (outString + PCRString_Length( outString )), message, PCRString_Length( message ) );
        PCRString_Concat( (outString + PCRString_Length( outString )), &trailingString[i],
                                                                       PCRString_Length( &trailingString[i] ) );
        DOF_LogMessage( DOF_GetDof(), logLevel, outString );
        return TRUE;
    }

    return FALSE;
}


/* @internal */
boolean DOFSecureGroup_LogMessageNoTrailer( DOFLogLevel logLevel, const char * message )
{
    char outString[200];

    if ((sizeof( secureGroupLogString ) + PCRString_Length( message )) <= sizeof( outString ))
    {
        PrepLogOutString( (char *)&outString, sizeof( outString ) );
        PCRString_Concat( (outString + PCRString_Length( outString )), message, PCRString_Length( message ) );
        DOF_LogMessage( DOF_GetDof(), logLevel, outString );
        return TRUE;
    }

    return FALSE;
}


/* @internal

    Called by OALConnection_Connect() for a hub.
*/
void DOFSecureGroup_SetHubConfiguredPointTimeout( DOFSecureGroup self, uint32 maxReceiveSilence )
{
    PCR_Assert( self );
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                   "Hub set to timeout points silent at milliseconds  0x", maxReceiveSilence );
    self->hubConfiguredPointTimeout = maxReceiveSilence;
}


/** @internal
    Create a new @ref DOFSecureGroup.

    @note This allocates resources that must later be freed by calling @ref DOFSecureGroup_Release().
    @note Until secure group is destroyed do not stop server passed in here. That would halt secure group communication.

    @param group        The OID of the multicast group.
    @param associatedStack
    @param associatedTransport
    @param securityConector MUST be set
    @param listenServer DOFServer through which incoming PDUs will be heard. If not @a pointMember this must not be NULL
    @param relatedConnection Holds configured hubAddress for point. If @a pointMember, this must not be NULL
    @param connConfig   The connConfig given when the DOFConnection was created.  This also contains the authentication
                        credentials used to join the multicast group.
    @param myPerms      The initial permission set of the new multicast group. Should not be NULL.
    @return A new @ref DOFSecureGroup, or NULL if
            - groupOID is NULL, or
            - connConfig is NULL, or
            - connConfig's credentials.creds is NULL, or
            - connConfig's permission set is NULL, or
            - those creds have no domainID, or
            - allocation or copy errors happen.
**/
DOFSecureGroup DOFSecureGroup_Create( const DOFObjectIDAuthentication group,
                                      OALConnectionlessProtocolStack associatedStack, DTL associatedTransport,
                                      DOFSecurityMembrane securityConnector, OALServer listenServer,
                                      OALConnection relatedConnection, DOFConnectionConfig connConfig,
                                      DOFPermissionSet myPerms )
{
    DOFConnectionType cType       = DOFConnectionConfig_GetConnectionType( connConfig );
    boolean           pointMember = (cType == DOFCONNECTIONTYPE_POINT);
    boolean           hub         = (cType == DOFCONNECTIONTYPE_HUB);
    STATIC_ASSERT( offsetof( DOFSecureGroup_t, channel ) == 0 );

    PCR_Assert( pointMember || listenServer );                                  /* A group member must have a server. */
    PCR_Assert( !pointMember|| relatedConnection );                             /* A point must have a connection. */
    PCR_Assert( securityConnector );
    if (group && connConfig && connConfig->credentials &&
        DOFCredentials_GetDomainID( connConfig->credentials ))
    {
        if (connConfig->permissions)
        {
            DOFSecureGroup retGroup = DOFMem_Alloc( sizeof( DOFSecureGroup_t ) );

            if (retGroup)
            {
                OALChannel           secureGroupChannel = (OALChannel)retGroup;
                boolean              success            = TRUE;
                boolean              constructed        = FALSE;

                PCRMem_Set( retGroup, 0, sizeof( DOFSecureGroup_t ) );
                /* We check now the security connector passed to the group _CreateConnection() call.
                   If NULL was passed into that call, it was replaced with the default security
                   before we are called. */
                PCR_Assert(DOFSecurityMembrane_IsChannelMembrane(securityConnector));
                if (OALChannelMembrane_GetBridge(securityConnector) == DOF_GetDefaultOperationBridge( DOF_GetDof() ))
                {
                    OALOperationBridge newBridge = OALOperationBridge_CreatePassThroughBridge(secureGroupChannel);

                    if (newBridge == NULL)
                    {
                        success = FALSE;
                    }
                    else
                    {
                        OALChannelMembrane_SetBridge(securityConnector, newBridge);
                        OALOperationBridge_Destroy(newBridge);
                    }
                }
                else
                {
                   /* This channel, though a secureGroup channel, is viewed according to the security connector
                      specified by the app programmer. */
                }
                /* OAP traffic sent across this channel (a secureGroup channel) is always sent secure. The distinction
                   offered by the selected channel security connector is that it allows an OAS to decide whether or not
                   to additionally bridge received secure traffic to non secure connections. */
                /* Note: SGMP traffic always sent according to the rules of SGMP (sometimes not secure). */
                /* Note: No matter what happens here we leave passed in securityConnector alone. */
                /* Note: Nothing can release chan w/o us calling this _Ctor 1st (sets required destroy function member*/
                if (success && ((constructed = OALChannel_Ctor( secureGroupChannel, TRUE, securityConnector,
                                                                DOFSecureGroup_RouteToProvider,
                                                                OALRemoteChannel_RouteToRequestorLossy,
                                                                DOFSecureGroup_HandleMessage, SecureGroup_Dtor,
                                                                DOFSecureGroup_ExtendTheirPermissions )) != FALSE))
                {
                    if( DOFConfig_IsRouter( DOF_GetConfig( DOF_GetDof() ) ) )
                    {
                        /* For router nodes only, add actAsAny permission to connConfig->permissions ("myPerms"). */
                        DOFPermission actAsAnyPerm = DOFPermissionActAsAny_Create();

                        success = FALSE;
                        if (actAsAnyPerm != NULL)
                        {
                            DOFPermissionSetBuilder builder = DOFPermissionSetBuilder_Create();

                            if (builder != NULL)
                            {
                                if (DOFPermissionSetBuilder_AddPermission( builder, actAsAnyPerm ))
                                {
                                    DOFPermissionSet actAsAnyPerms = DOFPermissionSetBuilder_Build( builder );

                                    if (actAsAnyPerms)
                                    {
                                        success = DOFPermissionSet_AddPermissionsFromSet( connConfig->permissions, actAsAnyPerms );
                                    }
                                    DOFPermissionSet_Destroy( actAsAnyPerms );
                                }
                                DOFPermissionSetBuilder_Destroy( builder );
                            }
                            DOFPermission_Destroy( actAsAnyPerm );
                        }
                    }
                    /* Only with constructed TRUE can we call the channel release function (DOFSecureGroup_Release(). */
                    if (constructed && success && ((retGroup->group = DOFObjectIDAuthentication_Copy( group )) != NULL))
                    {
                        if ((retGroup->extendPermissionsFIFO = OALQueue_Create()) != NULL)
                        {
                            if (!hub ||
                                (retGroup->hubPointTree = OALAVLTree_Create(
                                                     (OALAVLTree_CompareGreaterThan_fn_t)OALCipher_AddressIP_Port_gt_fn,
                                                     (OALAVLTree_CompareEqual_fn_t)OALCipher_AddressIP_Port_eq_fn,
                                                                             NULL )) != NULL)
                            {
                                if ((retGroup->workEnqueue = OALEnqueue_Create(DOF_GetDof()->threadPool)) != NULL)
                                {
                                    if ((retGroup->flags.pointMember = pointMember) != FALSE)
                                    {
                                        PCR_Assert( relatedConnection->connConfig );
                                        PCR_Assert( relatedConnection->connConfig->targetAddress );
                                        retGroup->lastReceivedSGMAddress =                 /* Configured hub DOFAddr. */
                                                        DOFAddress_Copy( relatedConnection->connConfig->targetAddress );
                                        if (!retGroup->lastReceivedSGMAddress)
                                            success = FALSE;
                                    }
                                    if (success)
                                    {
                                        /* Pre-allocate shutdown message now so channel can always send it. */
                                        if ((secureGroupChannel->remoteChannelData->shutdownMessage =
                                                  OALRemoteChannel_CreateShutdownMessage( secureGroupChannel )) != NULL)
                                        {
                                            /* Success: finish and out. */
                                            DOFAddress relatedAddress =
                                                                  listenServer ? OALServer_GetListenAddr( listenServer )
                                                                               : retGroup->lastReceivedSGMAddress;

                                            retGroup->specialEvent = PCREvent_Create( FALSE );
                                            if (retGroup->specialEvent)
                                            {
                                                if ((retGroup->reqGRespMutex = PCRMutex_Create()) != NULL)
                                                {
                                                    /* Note: connConfig->permissions is either filled by the app or was
                                                       filled by the connConfig builder with an empty permission set.
                                                       For the empty permission set, SGMP traffic is always fine. On the
                                                       other hand, OAP traffic is at mercy of otherPerms that the AS
                                                       gives back. This next fills remoteChData->myPerms from
                                                       connConfig and gives empty otherPerms. */
                                                    OALChannel_RemoteChannel_Ctor( secureGroupChannel, associatedStack,
                                                                                   associatedTransport, relatedAddress,
                                                                                   connConfig, NULL );
                                                    if (cType == DOFCONNECTIONTYPE_GROUP_MEMBER_ONLY)
                                                        retGroup->flags.doNotBeSGM = 1;
                                                    if (hub)
                                                        retGroup->flags.hub = 1;
                                                    ((OALChannel)retGroup)->remoteChannelData->flags.secureDesire = 1;
                                                    /*These next allow our secureGroup channel participate in TRP PDUs*/
                                                    ((OALChannel)retGroup)->remoteChannelData->flags.TRP_AcceptsCommands
                                                                                                = TRPRouter_KnowsAnAS();
                                                    ((OALChannel)retGroup)->remoteChannelData->flags.TRP_SendCommands
                                                                                                                    = 1;
                                                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                                                    {
                                                        if (SCALE_TMAX_TMIN > 1)
                                                            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                                                                       "SPECIAL TMAX-TMIN divisor: 0x",
                                                                                       SCALE_TMAX_TMIN );
                                                        if (SCALE_THB > 1)
                                                            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                                                                       "SPECIAL THB divisor: 0x",
                                                                                       SCALE_THB );
                                                    }
                                                    PCR_Assert( retGroup->up == FALSE );
                                                    return retGroup;                                  /* Success out. */
                                                }
                                                else { /* ->reqGRespMutex alloc failure. */ }
                                            }
                                            else { /* ->specialEvent alloc failure. */ }
                                        } else { /* Failed to create shutdown message */ }
                                    } else { /* ->lastReceivedSGMAddress alloc failure. */ }
                                    OALEnqueue_Destroy( retGroup->workEnqueue );/*Not cleaned up in SecureGroup_Dtor()*/
                                } else { /* ->workEnqueue alloc failure. */ }
                            } else { /* ->hubPointTree alloc failure. */ }
                        } else { /* ->extendPermissionFIFO alloc failure. */ }
                    } else { /* constructor failed, or IsRouter add permission failure, or copy problems. */ }
                }
                if (constructed)
                    DOFSecureGroup_Release( retGroup );             /* retGroup took owership of all its members. */
                else
                {
                    DOFMem_Free( retGroup );
                }
            } else { /* retGroup alloc failure. */ }
        } else { /* connConfig->permissions not present. */ }
    } else { /* Parameter problems. */ }

    return NULL;                                                                                        /* Error out. */
}


/* @internal */
uint8 DOFSecureGroup_GetOpKEK( DOFSecureGroup secureGroup )
{
    PCR_Assert( secureGroup );

    return secureGroup->flags.fut ? 0 : 1;
}


/* @internal
    Note: This does nothing for hub when beSGM param is FALSE
   */
void DOFSecureGroup_ChangeSGMState( DOFSecureGroup secureGroup, boolean beSGM )
{
    PCR_Assert( secureGroup );
    PCR_Assert( !beSGM || !secureGroup->flags.doNotBeSGM );
    PCR_Assert( !beSGM || !secureGroup->flags.pointMember );
    if (beSGM || !secureGroup->flags.hub)
    {
        boolean  wasSGM;
        PCRMutex mutex = OALMutexPool_Get( secureGroup );

        if (((OALChannel)secureGroup)->remoteChannelData->cipher)
            OALCipher_SetManager( ((OALChannel)secureGroup)->remoteChannelData->cipher, beSGM );

        PCRMutex_Lock( mutex );
        wasSGM = (boolean)secureGroup->flags.activeSGM;
        secureGroup->flags.activeSGM = beSGM;
        secureGroup->flags.becomeSGMInPlay = 0;
        secureGroup->flags.simpleTakeoverInPlay = 0;
        secureGroup->flags.noSlotRekeyReceivedSinceFirstHBTimeout = 0;
        PCRMutex_Unlock( mutex );
        OALMutexPool_Release( secureGroup );
        if (beSGM)
        {
            do
            {
                secureGroup->tieBreaker = PCRRandom_Get();          /* New tiebreaker each time we step up to be SGM. */
            } while (secureGroup->tieBreaker == 0);
        }
        else
            secureGroup->tieBreaker = 0;
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ) && wasSGM != beSGM)
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_DEBUG, "SGM state change, new state: ", (uint32)beSGM );
    }
    else
    {
        /* A hub never stops being SGM, regardless of what circumstances called us. */
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG,
                                           "DOFSecureGroup_ChangeSGMState() called to take hub down as SGM- ignored." );
    }
}


/** @internal
   Initialize DOFSecureGroups (one per DOF) to hold information commonly used by each secureGroup of a DOF.

   @return Success or fail.
**/
boolean DOFSecureGroups_Init( void )
{
    PCR_Assert( DOFSECUREGROUP_MY_LOWEST_APPID == SGMP_V1_APPID );
    PCR_Assert( DOFSECUREGROUP_MY_HIGHEST_APPID >= DOFSECUREGROUP_MY_LOWEST_APPID );
    myHighestAPPIDBit = DOFSecureGroup_ComputeAPPIDVersionBit( DOFSECUREGROUP_MY_HIGHEST_APPID );

    return TRUE;
}


/** @internal
   Fill internal flags to force SGM behavior of the given secureGroup.

   @param secureGroup      The given secureGroup.
   @param appidVersionBit  0 at ordinary call, 1 bit is set for simple takeover and is appid of SGM that we joined with.
   @param checkKEKNow
   @return TRUE.
**/
boolean DOFSecureGroup_InitiateMySGMBehavior( DOFSecureGroup secureGroup, uint32 appidVersionBit, boolean checkKEKNow )
{
    PCRMutex mutex = OALMutexPool_Get( secureGroup );

    PCR_Assert( secureGroup );
    PCR_Assert( !secureGroup->flags.pointMember );
    PCR_Assert( secureGroup->heartbeatTimerItem );
    DOFSecureGroup_ChangeSGMState( secureGroup, BE_SGM );
    secureGroup->activeMemberAPPIDTable = appidVersionBit;     /* Not altered until a member joins group. */
    secureGroup->pastActiveMemberAPPIDTable = appidVersionBit; /* Tmax: Make all appid inactive until proven active. */
    PCR_Assert( secureGroup->keks[DOFSecureGroup_GetOpKEK( secureGroup )].Tmax );
    PCR_Assert( secureGroup->keks[DOFSecureGroup_GetOpKEK( secureGroup )].Tmin );
    secureGroup->sgmNextActiveMemberAPPIDTableCheckTime = PCRTime_GetMS() +
             (secureGroup->keks[DOFSecureGroup_GetOpKEK( secureGroup )].Tmax*SGMP_TMAX_MSEC_MULTIPLIER/SCALE_TMAX_TMIN);
    PCR_Assert( secureGroup->sgmNextActiveMemberAPPIDTableCheckTime );
    PCRMutex_Lock( mutex );
    DOFSecureGroup_SetEpochTimeout( secureGroup,
                                    (checkKEKNow ? 0 :
              (secureGroup->keks[DOFSecureGroup_GetOpKEK(secureGroup)].Tmin*SGMP_TMIN_MSEC_MULTIPLIER/SCALE_TMAX_TMIN)),
                                    TRUE );
    if (!secureGroup->flags.hub)
    {
        /* Do not use DOFSecureGroup_ExtendHeartbeatTimeout(), its coding will NOT timeout in 1 msec. */
        /* Note: mutex lock prevents sg cleanup from stealing ->heartbeatTimerItem away from us. */
        if (OALTimerItem_Insert( secureGroup->heartbeatTimerItem, 1, FALSE, FALSE ) != DOFERROR_OK)/*1st hb out 1 msec*/
        {
            /* Cannot set our heartbeat timeout; we will send no heartbeats at all. */
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_DEBUG,
                                                   "DOFSecureGroup_InitiateMySGMBehavior() failed timer item insert.");
        }
    }
    PCRMutex_Unlock( mutex );
    OALMutexPool_Release( secureGroup );

    return TRUE;
}


/* @internal
   Return rank and also update group state gathered flag.
*/
uint8 DOFSecureGroup_GetRank( DOFSecureGroup secureGroup )
{
    PCR_Assert( secureGroup );
    if (secureGroup->flags.hub)
        return SGM_RANK_HUB;
    if (secureGroup->flags.pointMember)
        return SGM_RANK_POINT;
    if (!secureGroup->flags.groupStateGathered)
    {
        PCRMutex mutex;

        if (secureGroup->flags.activeSGM)
        { /* Once SGM, fall through to use normal rank, to try to stay SGM. */
        }
        else
        {
            uint32 scaledHeartbeatTimeout;
            uint8  opKEK = DOFSecureGroup_GetOpKEK( secureGroup );

            PCR_Assert( secureGroup->keks[opKEK].filled );
            PCR_Assert( secureGroup->keks[opKEK].Thb );
            scaledHeartbeatTimeout = secureGroup->keks[opKEK].Thb * 1000;
#if (SCALE_THB > 1)
            scaledHeartbeatTimeout = DOFSecureGroup_ScaleHeartbeatTimeout( secureGroup, scaledHeartbeatTimeout );
#endif
            if (DOFTime_Compare( PCRTime_GetMS(),
                                 (secureGroup->beginGatherGroupStateTime + (2 * scaledHeartbeatTimeout)) ) <= 0)
            {
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                 "_GetRank() returns 0 because 2 Thb has not passed by. 2 Thb msec: 0x", (2 * scaledHeartbeatTimeout) );
                return 0;                          /* Spec: rank is 0 until group state is gathered (2 Thb). */
            }
        }
        mutex = OALMutexPool_Get( secureGroup );
        PCRMutex_Lock( mutex );
        secureGroup->flags.groupStateGathered = 1;
        PCRMutex_Unlock( mutex );
        OALMutexPool_Release( secureGroup );
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "_GetRank() ends early node-rank suppression." );
    }

    return DOFConfig_GetRank( DOF_GetConfig( DOF_GetDof() ) );
}


/* @internal */
uint8 DOFSecureGroup_StepDownTest( DOFSecureGroup secureGroup, uint16 hisHighestAPPID, uint8 hisRank )
{
    uint8 myRank;

    PCR_Assert( secureGroup );
    if (secureGroup->flags.hub)
        return NOT_STEPDOWN;
    if (secureGroup->flags.doNotBeSGM)
        return DO_STEPDOWN;
    if (hisHighestAPPID != DOFSECUREGROUP_MY_HIGHEST_APPID)
        return (hisHighestAPPID > DOFSECUREGROUP_MY_HIGHEST_APPID) ? DO_STEPDOWN : NOT_STEPDOWN;
    myRank = DOFSecureGroup_GetRank( secureGroup );
    if (hisRank != myRank)                                                 /* APPIDs are equal. */
        return (hisRank > myRank) ? DO_STEPDOWN : NOT_STEPDOWN;

    return TIE_STEPDOWN;                                                   /* APPIDs and also ranks are equal. */
}


void DOFSecureGroup_UpdateLastReceivedSGMAddress( DOFSecureGroup secureGroup, DOFAddress sgmAddress )
{
    PCRMutex   mutex;

    PCR_Assert( secureGroup );                  /* Several hb PDUs or 1 ReqGroupResponse PDU could come at same time. */
    mutex = OALMutexPool_Get( secureGroup );
    PCRMutex_Lock( mutex );
    /*Note: sgmAddress and ->lastReceivedSGMAddress can be identical structure, for which copying early avoids trouble*/
    if (!sgmAddress || ((sgmAddress = DOFAddress_Copy( sgmAddress )) != NULL))
    {
        /* Perform the update. */
        DOFAddress_Destroy( secureGroup->lastReceivedSGMAddress );                               /* Out with the old. */
        secureGroup->lastReceivedSGMAddress = sgmAddress; /* In with the new: can be NULL, equal to old, or new.*/
    }
    else
    {
        /* copy failure: skip the update this time. */
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "_UpdateLastReceivedSGMAddress() copy failure." );
    }
    PCRMutex_Unlock( mutex );
    OALMutexPool_Release( secureGroup );
}


/* Called by GetKEK() when rejected by AS, and by HeartbeatTimeoutHandler() when we cannot be SGM and nobody else is,
   by GrantTimeout_fn(), and by HandleMessage_ConnectionInitiatedShutdown(). */
void DOFSecureGroup_LeaveGroup( DOFSecureGroup secureGroup, DOFException exception )
{
    PCR_Assert( exception );
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "_LeaveGroup called with DOFException 0x",
                                                      DOFErrorException_GetError( exception ) );

    DisableAndCleanup( secureGroup );
    OALRemoteChannel_Shutdown( (OALChannel)secureGroup, exception );         /* HandleMessage_Shutdown called by msg. */
}


/* @internal. */
void DOFSecureGroup_AddVersionToAPPIDTable( DOFSecureGroup secureGroup, uint32 appid )
{
    uint32 appidBit;

    PCR_Assert( secureGroup );
    appidBit = DOFSecureGroup_ComputeAPPIDVersionBit( appid );
    secureGroup->activeMemberAPPIDTable |= appidBit;
    secureGroup->pastActiveMemberAPPIDTable |= appidBit;
}


/* @internal.
    SGMP rekey PDUs specify when the rekey takes effect in a seconds promote delay number. That key promotion should not
    extend farther out than one SGMP heartbeat period. For a given seconds promote delay number, this returns that same
    number or the heartbeat period in seconds, whichever is lower.
*/
uint8 DOFSecureGroup_ReduceToThb( DOFSecureGroup secureGroup, uint16 secsPromoteDelay )
{
    /* Thb is a uint8 of seconds, so we return uint8 at most. */
    uint32 scaledHeartbeatTimeout;
    uint16 scaledSecsPromoteDelay;
    uint8  opKEK = DOFSecureGroup_GetOpKEK( secureGroup );

    PCR_Assert( secureGroup->keks[opKEK].filled );
    PCR_Assert( secureGroup->keks[opKEK].Thb );
    scaledHeartbeatTimeout = secureGroup->keks[opKEK].Thb * 1000;
#if (SCALE_THB > 1)
    scaledHeartbeatTimeout = DOFSecureGroup_ScaleHeartbeatTimeout( secureGroup, scaledHeartbeatTimeout );
#endif
    scaledSecsPromoteDelay = scaledHeartbeatTimeout / 1000;
    PCR_Assert( scaledSecsPromoteDelay <= 0xFF );                                                 /* 0xFF is max Thb. */
    if (secsPromoteDelay > scaledSecsPromoteDelay)
        secsPromoteDelay = scaledSecsPromoteDelay;
    PCR_Assert( secsPromoteDelay <= 0xFF );

    return secsPromoteDelay;
}


/* @internal
   Allows use at 1st try where sg known not up.
*/
boolean DOFSecureGroup_NowUpOrNeverUp( const DOFSecureGroup self )
{
    return (!self->upSeen || self->up);                       /* ->upSeen latches TRUE when self joins a secureGroup. */
}


DOFObjectIDAuthentication DOFSecureGroup_GetGroupID( DOFSecureGroup self )
{
    return self->group;
}

/**
@}
**/
