/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/cipherhmacutils.h>
#include <internal/error.h>
#include <internal/mutex_pool.h>
#include <internal/security_common.h>
#include <internal/sgmp_v1.h>
#include "securegroupprivate.h"


/** @addtogroup DOFSecureGroup
@{
**/


/*
**
**  "Private" Type and Method Declarations
*/


#define REQUEST_GROUP_NONCE_SIZE (DEFAULT_NONCE_SIZE)         /* As sender, we choose nonce length, spec allows 1-16. */


typedef struct RequestGroupContext_t * RequestGroupContext;
typedef struct RequestGroupContext_t
{
    PCREvent        reqGRespEvent;
    OALStream       streamI;
    OALKeyGrant     ticket;
    DOFGroupKey_t * keyInfo;
    DOFAddress      address;
    DOFError        callbackStatus;
    uint32          nonTimeoutCallbackNeedsContext_count;     /* Used only by instances of RequestGroupCallback(). */
    uint16          appid;
    uint8           rank;
    boolean         receivedMulticast;
    volatile boolean contextFilled;
    boolean         consumerIsGone;
    boolean         timeoutOverrodeAnAllocatingCallbackInstance;
    boolean         someFinalCallbackMustFreeContext;         /* Used only by instances of RequestGroupCallback(). */
} RequestGroupContext_t;


typedef struct AcceptedRequestGroupResponse_t
{
    DOFAddress  address;
    uint16      appid;
    uint8       rank;
    boolean     filled;
} AcceptedRequestGroupResponse_t;


/** @internal
    This function is used to extract the group key from the mac and key given in the ticket using the
    ticket and KEK.

    @param  KEK         The KEK provided by the auth server in the authorize KEK PDU.
    @param  ticket      The ticket that contains the MAC and KEY.
    @param  groupKey    The buffer where the group key will be stored.
**/
static void extractGroupKeyFromMacKekAndKey( OALCryptKey KEK, OALKeyGrant ticket, OALCryptKey groupKey )
{
    OALCryptExpandedKey ekey;
    uint8               i;
    uint8               size = sizeof(ticket->key);

    PCRMem_Copy( groupKey, ticket->mac, size );
    PCR_Validate( DOFCipherRegistry_AES256Cipher );
    DOFCipherRegistry_AES256Cipher->GenerateKeyState( ekey, KEK );
    DOFCipherRegistry_AES256Cipher->Encrypt( ekey, &groupKey[0] );
    DOFCipherRegistry_AES256Cipher->Encrypt( ekey, &groupKey[16] );

    for (i = 0; i < size; i++)
        groupKey[i] ^= ticket->key[i];
}


/** @internal

    @note  Function prototype is SGMP_V1_RequestGroupCallback_fn.
    @param address         DOFAddress of node that sent the ReqG Response PDU. If NULL, response isn't best so drop it.
    @param ticket          the ticket received in the ReqG response PDU or null for timeout.
    @param hisHighestAPPID highest APPID of the sender SGM.
    @param hisRank         rank of the sender SGM.
    @param streamCcmModeInitialState
    @param status          the status of the request group.
    @param ctx             where the event and other information is passed back from the request group.
    @param streamCcmModeInitialState
    @param receivedMulticast
**/
static void RequestGroupCallback( DOFAddress address, OALKeyGrant ticket, uint16 hisHighestAPPID, uint8 hisRank,
                                  DOFError status, void * ctx, OALStream streamCcmModeInitialState,
                                  boolean receivedMulticast, DOFOperation respOp, PCRMutex reqGRespMutex )
{
    /* Matches typedef void (*SGMP_V1_RequestGroupCallback_fn) in sgmp_v1.h. */
    boolean             snapshotSomeFinalCallbackMustFreeContext;
    uint32              snapshotNonTimeoutCallbackNeedsContext_count;
    DOFError            computedStatus;
    RequestGroupContext context = ctx;

    /* We are called only by SGMP_V1_HandleRequestGroupResponse_Initiator() or SGMP_V1_OpClosing(). */
    /* NOTE: Both callers call with respOp locked. respOp may or not be cancelled, but has a refCount and is valid. */
    PCR_Assert( reqGRespMutex );
    /* Strategy described: This callback is called once when no SGM answers our ReqG PDU- the timeout, or twice when one
       SGM delivers us a ReqG Response- once for the answer and then once later when the reqG operation times out. This
       callback can also be called more times when additional SGMs respond. Strategy: Lock/Unlock reqGRespMutex once
       ONLY in this callback, and if we signal AcceptHighestPriorityRequestGroupResponse() then do not exit here until
       it has consumed our context. For extreme callback combinations (especially when responses come in very near to
       the timeout), other strategies can allow context to be freed before a last caller is done with it. */
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                   "RequestGroupCallback about to lock mutex, called with status: 0x", status );
#endif
    OALMutexPool_Lock( reqGRespMutex );
    /* AcceptHighestPriorityRequestGroupResponse() and DOFSecureGroup_DenyGroupBehavior() also lock this mutex. */
    PCR_Assert( status == DOFERROR_OK ||                 /* SGMP_V1_HandleRequestGroupResponse_Initiator() called us. */
                status == DOFERROR_TIMEOUT );                                       /* SGMP_V1_OpClosing() called us. */
    if (status == DOFERROR_OK)
    {
        /* Exit immediately if this callback instance is useless. */
        boolean quickExit = FALSE;

        if (DOFOperation_IsCancelled(respOp))/*Can be by t/o or AcceptHighestPriorityRequestGroupResponse cancelled us*/
            quickExit = TRUE;                /* context already free'd IF we were called before with DOFERROR_TIMEOUT.*/
        else
        {
            /* context not freed yet. Reason: respOp not timed out, therefore we haven't been called to free context. */
            if (context->consumerIsGone)
            {
                /*We did not bump ->nonTimeoutCallbackNeedsContext_count, so not part of the context freeing mechanism*/
                quickExit = TRUE;
            }
        }
        if (quickExit)
        {
            OALMutexPool_Unlock( reqGRespMutex );
            return; /*AcceptHighestPriorityRequestGroupResponse() has no need of our info; context might even be freed*/
        }
    }
    else
    {
        /* We are called once and only once with DOFERROR_TIMEOUT, and that by SGMP_V1_OpClosing(). */
        PCR_Assert( DOFOperation_IsCancelled( respOp ) );
    }
    PCR_Assert( context );    /* Both callers give us context; with both mutexes locked we know context is not freed. */

    /* Note: Calls from SGMP_V1_HandleRequestGroupResponse_Initiator are in strict sequence, the 2nd call not happening
       until we have returned from the 1st. However, the call from SGMP_V1_OpClosing() comes from a different thread and
       so a call to us can happen from either thread at any time. Rarely, then, two instances of this callback may be
       executing in parallel, which are a problem if context can be destroyed too soon. Only 2 situations lead to this:
           1. Timeout got us here with SGM response following just after (which is stuck now at the mutex lock above),
              harmless because trailing SGM response exits early above on finding op cancelled, never touching context.
           2. SGM response got us here with timeout just after. This is safe because we do not enable freeing context
              until timeout gets to the the end of this callback, and it is stuck right now at the mutex lock above. */

    computedStatus = status;                        /* Will be overwritten below for special errors or circumstances. */
    if (status == DOFERROR_OK)
    {
        /* Fill all other entries of context (ctx). DOFERROR_TIMEOUT fills only context->callbackStatus. */
        if (address)
        {
            if (ticket)
            {
                if ((context->address = DOFAddress_Copy( address )) != NULL)
                {
                    if ((ticket = OALKeyGrant_Retain( ticket )) != NULL)
                    {
                        context->keyInfo->filled = TRUE;                          /* Success. */
                        context->ticket = ticket;
                        context->keyInfo->streamCcmModeInitialState = streamCcmModeInitialState;
                        OALStream_Retain( context->keyInfo->streamCcmModeInitialState );
                        context->appid = hisHighestAPPID;
                        context->rank = hisRank;
                        context->receivedMulticast = receivedMulticast;
                    }
                    else
                    {
                        computedStatus = DOFERROR_INSUFFICIENT_RESOURCES;
                        DOFAddress_Destroy( context->address );
                    }
                }
                else
                    computedStatus = DOFERROR_INSUFFICIENT_RESOURCES;
            }
            else
            {
                computedStatus = DOFERROR_MISSING_TICKET;
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                        "RequestGroupCallback unexpectedly finds no sender ticket and drops response" );
            }
        }
        else
        {
            computedStatus = DOFERROR_MISSING_SENDER_ADDRESS;
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                    "RequestGroupCallback unexpectedly finds no sender DOFAddress and drops response" );
        }
        PCR_Assert( computedStatus == DOFERROR_OK || computedStatus == DOFERROR_INSUFFICIENT_RESOURCES ||
                    computedStatus == DOFERROR_MISSING_TICKET || computedStatus == DOFERROR_MISSING_SENDER_ADDRESS );
    } /* end if (status == DOFERROR_OK) */

    context->contextFilled = TRUE;                         /* Cleared by AcceptHighestPriorityRequestGroupResponse(). */
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                         "RequestGroupCallback signals event with computedStatus: 0x", computedStatus );
#endif
    PCR_Assert( context->reqGRespEvent );
    PCREvent_Signal( context->reqGRespEvent );             /* Signal AcceptHighestPriorityRequestGroupResponse(). */
    /* Before unlocking mutex, prepare for free'ing context- done only by last callback that doesn't return above.
       Also, detect timeout callback overriding an unconsumed non-timeout callback. */
    if (status == DOFERROR_TIMEOUT)                                                  /* Original status given us. */
    {
        /* DOFERROR_TIMEOUT means SGMP_V1_OpClosing() called us; happens either by ordinary timeout or op cancel. */
        if (context->callbackStatus != DOFERROR_NOT_FOUND)
        {
            PCR_Assert( context->callbackStatus == DOFERROR_OK );
            /* Special case. We are timeout callback coming along just after a non-timeout callback whose context was
               not yet consumed by AcceptHighestPriorityRequestGroupResponse(). */
            context->timeoutOverrodeAnAllocatingCallbackInstance = TRUE;
            computedStatus = DOFERROR_TIMEOUT;        /* Override. Consumer will still cleanup other context members. */
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                       "RequestGroupCallback overrides a DOFERROR_OK instance with DOFERROR_TIMEOUT." );
        }
        PCR_Assert( !context->someFinalCallbackMustFreeContext );
        context->someFinalCallbackMustFreeContext = TRUE; /* Context freed by us or callback that still needs context.*/
        snapshotNonTimeoutCallbackNeedsContext_count = context->nonTimeoutCallbackNeedsContext_count;
    }
    else
    {
        PCR_Assert( context->callbackStatus == DOFERROR_NOT_FOUND ); /*non timeout cbck does not override timeout cbck*/
        snapshotNonTimeoutCallbackNeedsContext_count = ++context->nonTimeoutCallbackNeedsContext_count;
    }
    snapshotSomeFinalCallbackMustFreeContext = context->someFinalCallbackMustFreeContext;
    context->callbackStatus = computedStatus;                                    /* Set context->status just in time. */
    OALMutexPool_Unlock( reqGRespMutex );

    /* Even though the mutex is unlocked, exactly one instance of non-timeout callback can be here at the same time. The
       timeout callback can get here and execute in "parallel," however. context is valid for both instances, and there
       is no guarantee as to which callback will first complete and exit. After
       context->someFinalCallbackMustFreeContext gets set by op timeout callback getting this far, all new
       non-timeout callback instances harmlessly return, above, and do not get here. */
    PCR_Assert( status == DOFERROR_TIMEOUT || context->nonTimeoutCallbackNeedsContext_count > 0 ); /*From us, at least*/
    while (context->contextFilled)               /* Naked (not bit-set) boolean is readable without mutex protection. */
        PCRThread_Sleep( 4 );   /* Move on only after AcceptHighestPriorityRequestGroupResponse has consumed context. */

    PCR_Assert( status != DOFERROR_TIMEOUT || context->consumerIsGone ); /*Original status then readable naked boolean*/
    if (status != DOFERROR_TIMEOUT)
    {
        PCR_Assert( snapshotNonTimeoutCallbackNeedsContext_count > 0 );
        PCR_Assert( context->nonTimeoutCallbackNeedsContext_count > 0 );                     /* From us, at least. */
        /* Safe because no other non-timeout callback executes in parallel, so this member cannot be manipulating. */
        --context->nonTimeoutCallbackNeedsContext_count;/*Do not read context after this, can free in another instance*/
    }

    if (snapshotSomeFinalCallbackMustFreeContext)
    {
        boolean freeContext = FALSE;

        if (status == DOFERROR_TIMEOUT)
        {
            /* We are the callback instance that set ->someFinalCallbackMustFreeContext. */
            if (snapshotNonTimeoutCallbackNeedsContext_count == 0)
                freeContext = TRUE; /* The ordinary case. All non-timeout callback instances completed ahead of us. */
            else
            { /* One or more non-timeout callback instances have not completed yet and one of them will free context. */
            }
        }
        else
        {
            PCR_Assert( snapshotNonTimeoutCallbackNeedsContext_count > 0 );
            if (snapshotNonTimeoutCallbackNeedsContext_count == 1)
                freeContext = TRUE;   /* We are the final non-timeout callback instance to complete after op timeout. */
            else
            {/*1 or more other non-timeout callback instances have not completed yet and 1 of them will free context. */
            }
        }
        if (freeContext)
        {
            PCR_Assert(context->consumerIsGone); /*Happened when AcceptHighestPriorityRequestGroupResponse saw timeout*/
            DOFMem_Free( ctx );  /* We won't be called again and AcceptHighestPriorityRequestGroupResponse() is gone. */
        }
    }
}


/* @internal */
static DOFOperation IssueRequestGroupPDU( DOFSecureGroup secureGroup, OALStream streamI, DOFKEKInfo KEKInfo,
                                     DOFKeyRequest keyRequest, RequestGroupContext_t * context,
                                     DOFAddress targetAddress, uint32 timeout, OALSecureHMACContext * pHmacI )
{
    DOFCredentials    secureCreds;
    DOFObjectIDDomain domainID;
    uint32            size;
    DOFOperation      reqGroupOp = NULL;

    /* Fill streamI, send ReqG PDU, keep streamI for use in checking ReqG response. */
    PCR_Assert( secureGroup );
    PCR_Assert( secureGroup->reqGRespEvent == NULL );
    PCR_Assert( streamI );
    PCR_Assert( KEKInfo );
    PCR_Assert( keyRequest );
    PCR_Assert( KEKInfo->securityScope );
    PCR_Assert( context );
    PCR_Assert( pHmacI );
    secureCreds = ((OALChannel)DOFSecureGroup_GetChannel_Internal( secureGroup ))->remoteChannelData->secureCredentials;
    PCR_Assert( secureCreds );
    domainID = DOFCredentials_GetDomainID( secureCreds );
    PCR_Assert( domainID );
    /* Security.7 PDU spec: domainID not broadcast or multicast OID. */
    size = DOFObjectID_GetBytesSize( DOFObjectIDDomain_GetObjectID( domainID ));
    if (OALStream_Write( streamI, size, DOFObjectID_GetBytes( DOFObjectIDDomain_GetObjectID( domainID ) ) ) &&  /* D. */
        OALStream_Write16( streamI, KEKInfo->epoch ) && DOFKeyRequest_Marshal( streamI, keyRequest, TRUE ) &&   /* I. */
        DOFSecurityScope_Marshal( KEKInfo->securityScope, streamI ) &&                                     /* more I. */
        (secureGroup->reqGRespEvent = PCREvent_Create( FALSE )) != NULL)
    {
        /* Note: Though the creds struct within keyRequest includes a domain member, that is not marshalled. */
        /* Note: keyRequest's creds member holds identity, which is the groupOID of the group we are trying to join. */
        context->reqGRespEvent = secureGroup->reqGRespEvent;
        /* context->keyInfo struct pointer was filled by caller- RequestGroupCallback() will fill that structure. */
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_DEBUG, "SGMP RequestGroup PDU being sent, as SGM: ",
                                                          (uint32)secureGroup->flags.activeSGM );
        /* SGMP spec "A node should only attempt to join a group using the latest version of SGMP that it understands"*/
        switch (DOFSECUREGROUP_MY_HIGHEST_APPID)
        {
        case SGMP_V1_APPID:
            {
                PCR_Assert( secureGroup->reqGRespMutex );
                /* For 1 sec, this call will accept multiple responses, calling RequestGroupCallback()
                   for each. From one call to SGMP, our callback is called for each received response
                   until timeout. WHY: The one answer returned is the one having highest priority. */
                reqGroupOp = SGMP_V1_SendRequestGroup( DOFSecureGroup_GetChannel_Internal( secureGroup ),
                                                       targetAddress, streamI, domainID, KEKInfo->securityScope,
                                                       RequestGroupCallback, context, timeout, FALSE,
                                                       secureGroup->reqGRespMutex );
                /* Above call destroys his use of streamI. We have our own reference of streamI for use after return. */
            }
        break;
        default:
        break;
        }
        if (reqGroupOp != NULL)
        {
            OALSecureHMAC_Start( pHmacI, KEKInfo->KEK );                       /* Form hmac of D and I. */
            OALSecureHMAC_Digest( pHmacI, OALStream_GetDataSize( streamI ),
                                  OALStream_GetReadPtr( streamI ) );
        }
        else
        {
            PCREvent_Destroy( secureGroup->reqGRespEvent );                    /* Fail out. */
            secureGroup->reqGRespEvent = NULL;
        }
    }

    return reqGroupOp;
}


/* @internal */
static DOFError AcceptHighestPriorityRequestGroupResponse( DOFSecureGroup secureGroup, RequestGroupContext_t * context,
                                                           DOFKEKInfo KEKInfo, OALSecureHMACContext * pHmacI,
                                                           AcceptedRequestGroupResponse_t * pAcceptedResponse,
                                                           DOFGroupKey_t * pGroupKeyInfo, DOFOperation reqGroupOp )
{
    /* Both RequestGroupCallback() and DOFSecureGroup_DenyGroupBehavior() can signal our event. */
    PCR_Assert( secureGroup );
    PCR_Assert( KEKInfo );
    PCR_Assert( pHmacI );
    PCR_Assert( pAcceptedResponse );
    PCR_Assert( pGroupKeyInfo );
    PCR_Assert( secureGroup->reqGRespEvent );                     /* IssueRequestGroupPDU() created this event. */
    PCR_Assert( secureGroup->reqGRespMutex );
    PCR_Assert( context );                                        /* context is freed only in RequestGroupCallback(). */
    PCR_Assert( secureGroup->reqGRespEvent == context->reqGRespEvent );
    OALMutexPool_Lock( secureGroup->reqGRespMutex );  /* RequestGroupCallback, DenyGroupBehavior also lock this. */
    context->callbackStatus = DOFERROR_NOT_FOUND;                 /* Overwritten only by RequestGroupCallback(). */
    PCRMem_Set( pAcceptedResponse, 0, sizeof( AcceptedRequestGroupResponse_t ) );

    /* Note: At this moment, context->contextFilled is still FALSE from when RequestGroupAndFillKey() initialized it. */
    while (TRUE)
    {
        boolean  cleanupContextFromReqGCallback = FALSE;
        boolean  earlyPointExit                 = FALSE;
        DOFError callbackStatus                 = DOFERROR_NOT_FOUND;
        boolean  upToDown                       = TRUE;

        PCR_Assert( context->consumerIsGone == FALSE );
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                "AcceptHighestPriorityRequestGroupResponse() at loop top." );
#endif
        /* Before waiting for a ReqG response, be sure secureGroup did not transition from up to down. */
        if (DOFSecureGroup_NowUpOrNeverUp( secureGroup ))
        {
            OALMutexPool_Unlock( secureGroup->reqGRespMutex );
            /* Wait for RequestGroupCallback() to deliver us a response to or timeout of the ReqG op that was created in
               the SGMP_V1_SendRequestGroup call in IssueRequestGroupPDU(). */
            /* Note: RequestGroupCallback() sets context->contextFilled before it signals us. */
            PCREvent_Wait( secureGroup->reqGRespEvent, PCRTIMEOUT_FOREVER );
            /* Note: _DenyGroupBehavior() can also signal this event- if so then secureGroup is NOT up,
                     transitionUpToDown will remain set at TRUE, and we will exit immediately with DOFERROR_TERMINATED*/
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
        "AcceptHighestPriorityRequestGroupResponse() is signaled and will now lock the mutex, signalled with status 0x",
                                           context->callbackStatus );
#endif
            OALMutexPool_Lock( secureGroup->reqGRespMutex );
            callbackStatus = context->callbackStatus;
            PCR_Assert( callbackStatus == DOFERROR_OK || callbackStatus == DOFERROR_MISSING_SENDER_ADDRESS ||
                        callbackStatus == DOFERROR_INSUFFICIENT_RESOURCES || callbackStatus == DOFERROR_TIMEOUT ||
                        callbackStatus == DOFERROR_MISSING_TICKET ||     /* 5 from RequestGroupCallback() signaling us*/
                        callbackStatus == DOFERROR_NOT_FOUND );          /* Can be, when DenyGroupBehavior signals us.*/
            /* Before we act on gathered info from callback, catch that secureGroup transitioned from up to down. */
            if (DOFSecureGroup_NowUpOrNeverUp( secureGroup ))
            {
                upToDown = FALSE;
                if (callbackStatus == DOFERROR_OK)
                {
                    /* Structure at pGroupKeyInfo holds all new info from the SGM that answered us. */
                    OALCryptKey          extractedGroupKey;
                    OALSecureHMACContext hmacThisReqGResponse;
                    OALSecureHMACDigest  digest;                             /* To extract the group key. */

                    PCR_Assert( context->keyInfo );
                    PCR_Assert( context->keyInfo->filled );
                    cleanupContextFromReqGCallback = TRUE;
                    extractGroupKeyFromMacKekAndKey( KEKInfo->KEK, context->ticket, extractedGroupKey );
                    /* Verify group key is correct by creating MAC and comparing it to the one in ticket. */
                    PCRMem_Copy( &hmacThisReqGResponse, pHmacI, sizeof( OALSecureHMACContext ) );
                    /* This next completes hmac starting with pHmacI from above, by adding A and Kd block elements. */
                    if (OALSecureHmac_CompleteRequestGroupResponseMac( &hmacThisReqGResponse, extractedGroupKey,
                                                                       digest,          /* digest is filled. */
                                                                       context->keyInfo->streamCcmModeInitialState,
                                                                       context->appid, context->rank ) == DOFERROR_OK)
                    {
                        PCR_Assert( context->appid >= SGMP_V1_APPID );                        /* SGMP checked this. */
                        if (PCRMem_Compare( context->ticket->mac, digest, sizeof( context->ticket->mac ) ) == 0)
                        {
                            /* Successful response. Analyze and keep in pAcceptedResponse only response of
                               highest priority, based on highestAPPID, desire, and groupKey (as tiebreak). */
                            boolean acceptResponse = TRUE;

                            if (pAcceptedResponse->filled)          /* First time here, will not be filled. */
                            {
                                boolean tieAppid = (context->appid == pAcceptedResponse->appid);

                                if (tieAppid && (context->rank == pAcceptedResponse->rank))
                                {
                                    int         comp;                      /* Must be int !!!! */
                                    OALCryptKey newKeyXOR, acceptedKeyXOR; /* Tiebreak w/ XORed key compare. */

                                    PCRMem_Copy( &newKeyXOR[0], extractedGroupKey, sizeof( OALCryptKey ) );
                                    PCRMem_Copy( &acceptedKeyXOR[0],
                                                 pGroupKeyInfo->key,   /* Filled previously in our loop. */
                                                 sizeof( OALCryptKey ) );
                                    DOFSecureGroup_XORTwoKeys( &KEKInfo->KEK[0], &newKeyXOR[0], &acceptedKeyXOR[0] );
                                    comp = PCRMem_Compare( &newKeyXOR[0], &acceptedKeyXOR[0],
                                                           sizeof( OALCryptKey ) );
                                    if (comp >= 0)
                                        acceptResponse = FALSE; /* < newKey lost. == also don't use newKey. */
                                    /* Note: If 2 SGM's deliver ReqGResponse with identical appid, rank, and
                                             groupKey, we get to this line, which causes whichever SGM is
                                             accepted by member first to be the one finally accepted. */
                                }
                                else if (context->appid < pAcceptedResponse->appid ||
                                         (tieAppid && (context->rank < pAcceptedResponse->rank)))
                                    acceptResponse = FALSE;
                            }
                            if (acceptResponse)
                            {
                                /* This is 1st response or a winning response. */
                                DOFAddress newlyCopiedAddress = DOFAddress_Copy( context->address );

                                if (newlyCopiedAddress || !context->address)
                                {
                                    /* We can and will keep this response. */
                                    if (!pAcceptedResponse->filled)
                                        PCRMem_Set( pGroupKeyInfo, 0, sizeof( DOFGroupKey_t ) );
                                    /* First store things intended to return to caller in pGroupKey. */
                                    pGroupKeyInfo->receivedMulticast = context->receivedMulticast;
                                    PCRMem_Copy( pGroupKeyInfo->key, extractedGroupKey, sizeof( OALCryptKey ) );
                                    PCR_Assert( pAcceptedResponse->filled ||
                                                pGroupKeyInfo->streamCcmModeInitialState == NULL );
                                    OALStream_Release( pGroupKeyInfo->streamCcmModeInitialState, TRUE );
                                    pGroupKeyInfo->streamCcmModeInitialState =
                                                                            context->keyInfo->streamCcmModeInitialState;
                                    OALStream_Retain( pGroupKeyInfo->streamCcmModeInitialState );
                                    /* Next store things intended to return to caller in pAcceptedResponse. */
                                    pAcceptedResponse->appid = context->appid;
                                    pAcceptedResponse->rank = context->rank;
                                    PCR_Assert( pAcceptedResponse->filled || pAcceptedResponse->address == NULL );
                                    DOFAddress_Destroy( pAcceptedResponse->address );
                                    pAcceptedResponse->address = newlyCopiedAddress;
                                    pAcceptedResponse->filled = TRUE;
                                    /* Last nagging detail. */
                                    if (secureGroup->flags.pointMember)
                                        earlyPointExit = TRUE;     /* Need no consensus of best hub- there is only 1. */
                                }
                                else
                                { /*Don't update pAcceptedResponse due to copy error, but don't lose previous response*/
                                }
                            }
                        }
                        else
                        {
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
                            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                               "AcceptHighestPriorityRequestGroupResponse() finds failed validation." );
#endif
                            callbackStatus = DOFERROR_VALIDATION_FAILED;
                        }
                    }
                    else
                    {
                        /* _CompleteRequestGroupResponseMac() returned DOFERROR_ILLEGAL_ARGUMENT. */
                        callbackStatus = DOFERROR_INTERNAL_ERROR;
                    }
                    PCR_Assert( callbackStatus == DOFERROR_OK || callbackStatus == DOFERROR_VALIDATION_FAILED ||
                                callbackStatus == DOFERROR_INTERNAL_ERROR );
                }
                else
                {
                    PCR_Assert( !cleanupContextFromReqGCallback );         /* Here for callbackStatus != DOFERROR_OK. */
                    if (context->timeoutOverrodeAnAllocatingCallbackInstance)
                        cleanupContextFromReqGCallback = TRUE;
                }
            }
        }
        if (upToDown)
        {
            /* Act on terminate order, we will exit now. */
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                        "AcceptHighestPriorityRequestGroupResponse() sees transition from up to down" );
#endif
            if (callbackStatus == DOFERROR_OK)
                cleanupContextFromReqGCallback = TRUE; /* Cleanup contents of the valid ReqG response just handed us. */
            if (pAcceptedResponse->filled)
            {
                pAcceptedResponse->filled = FALSE;                    /* Don't use what we gathered, and clean it up. */
                DOFAddress_Destroy( pAcceptedResponse->address );
                OALStream_Release( pGroupKeyInfo->streamCcmModeInitialState, TRUE );
            }
            callbackStatus = DOFERROR_TERMINATED;      /* Force DOFERROR_TERMINATED return, even if previous response.*/
        }

        if (cleanupContextFromReqGCallback)
        {
            OALStream_Release( context->keyInfo->streamCcmModeInitialState, TRUE );
            OALKeyGrant_Release( context->ticket );
            DOFAddress_Destroy( context->address );
            context->timeoutOverrodeAnAllocatingCallbackInstance = FALSE;
        }

        if (callbackStatus == DOFERROR_TIMEOUT) /*Original request operation timed out (or we cancelled it down below)*/
        {
            /* Return, whether good, bad, or ugly. RequestGroupCallback() won't call us again and is waiting for us to
               exit, and when we do it will free context and exit for the last time (relating to this reqGroupOp). */
            DOFError retError = pAcceptedResponse->filled ?
                               DOFERROR_OK :     /* We have a good answer, happens ONLY w/ DOFERROR_OK from a callback*/
                               DOFERROR_TIMEOUT; /* No good answers came or secureGroup transitioned from up to down. */
            /* Note: These callbackStatus values may have been used in the course of handling ReqG responses, but are
               not returned to our caller: DOFERROR_INSUFFICIENT_RESOURCES, _MISSING_SENDER_ADDRESS, _MISSING_TICKET,
               _INTERNAL_ERROR, and _VALIDATION_FAILED. */

            PCREvent_Destroy( secureGroup->reqGRespEvent );
            secureGroup->reqGRespEvent = NULL; /* Tells _DenyGroupBehavior() we have nothing in progress. */
            context->consumerIsGone = TRUE;    /* Tells RequestGroupCallback() to ignore possible follow on responses.*/
            context->contextFilled = FALSE;                              /* Reset to let RequestGroupCallback() exit. */
            OALMutexPool_Unlock( secureGroup->reqGRespMutex );
            return retError; /*All context members are cleaned up. context itself is free'd in RequestGroupCallback().*/
        }
        else if (earlyPointExit ||                      /* The point found its hub and need not examine more responses*/
                 callbackStatus == DOFERROR_TERMINATED) /* We detected secureGroup went down. */
        {
            if (earlyPointExit)
            {
                PCR_Assert( callbackStatus == DOFERROR_OK );
                PCR_Assert( pAcceptedResponse->filled );
            }
            DOFOperation_Cancel( reqGroupOp ); /*Get RequestGroupCallback() called for last time; will signal us again*/
        }

        /*callbackStatus here can be DOFERROR_OK or 1 of the bad errors like MISSING_TICKET, but not DOFERROR_TIMEOUT*/
        context->callbackStatus = DOFERROR_NOT_FOUND;            /* Not overwritten if _DenyGroupBehavior signals us. */
        context->contextFilled = FALSE;                          /* Reset to allow RequestGroupCallback() to exit. */
        PCRThread_Yield();                                       /* Play nice for early RequestGroupCallback() exit. */
    } /* end while (TRUE). */
    PCR_Assert( FALSE );                                         /* The while loop returns only and never breaks out. */
}


/** @internal

    @note This blocks while waiting a full second to accept only one of possibly multiple responses.

    @param secureGroup      the secure Group or channel to use to send the request group.  This must not be NULL.
    @param targetAddress    NULL to send RequestGroup PDU multicast, not NULL to send it unicast.
    @param KEKInfo          The key and other information provided by the AS used to prove knowledge of the group.
    @param pGroupKeyInfo    The key and other related information returned to caller.
    @param pHisHighestAPPID Highest APPID of the SGM that returns our ReqGroupResp PDU [OUTPUT fills at this pointer].
    @param pHisRank         Rank of the SGM that returns our RequestGroupResponse PDU [OUTPUT fills at this pointer].
    @param pHisAddress      DOFAddress of the SGM that returns our RequestGroupResp PDU [OUTPUT fills at this pointer].
    @param timeout          The timeout passed to SGMP for the request on the wire.
    @return DOFERROR_OK if response from a secure group manager was received and structure at pGroupKeyInfo is filled.
**/
static DOFError RequestGroupAndFillKey( DOFSecureGroup secureGroup, DOFAddress targetAddress, DOFKEKInfo KEKInfo,
                                        DOFGroupKey_t * pGroupKeyInfo, uint16 * pHisHighestAPPID, uint8 * pHisRank,
                                        DOFAddress * pHisAddress, uint32 timeout )
{
    DOFCredentials      secureCreds;
    RequestGroupContext requestGroupContext;
    DOFError            retError = DOFERROR_INSUFFICIENT_RESOURCES;

    PCR_Assert( secureGroup );
    PCR_Assert( !secureGroup->flags.hub );
    secureCreds = ((OALChannel)DOFSecureGroup_GetChannel_Internal( secureGroup ))->remoteChannelData->secureCredentials;
    PCR_Assert( secureCreds );
    requestGroupContext = DOFMem_Alloc( sizeof( RequestGroupContext_t ) );
    if (requestGroupContext)
    {
        AcceptedRequestGroupResponse_t acceptedResponse;
        DOFCredentials                 keyRequestCredentials;
        DOFKeyRequest                  keyRequest = NULL;

        PCRMem_Set( requestGroupContext, 0, sizeof( RequestGroupContext_t ) );
        acceptedResponse.filled = FALSE;       /* Initializing the full structure happens when needed. */
        if ((keyRequestCredentials = DOFCredentialsKey_Create( DOFCredentials_GetDomainID( secureCreds ),
                                                             secureGroup->group, KEKInfo->KEK )) != NULL)
        {
            keyRequest = DOFKeyRequest_Create( keyRequestCredentials, REQUEST_GROUP_NONCE_SIZE, FALSE, FALSE, NULL, NULL );
            DOFCredentials_Destroy( keyRequestCredentials );                           /* keyRequest now owns it. */
        }
        if (keyRequest)
        {
            OALStream streamI = OALStream_CreateRW( 1000 );          /* REVIEW: How get total size w/o much work? */

            if (streamI)
            {
                /* Note: Though not necessarily obvious, scope of localGroupKeyInfo needs to be here. Do not change
                         this. The final user of this structure is the if (cleanupContextFromReqGCallback) code block of
                         AcceptHighestPriorityRequestGroupResponse(), called just below, and this structure needs still
                         be valid at that moment. */
                DOFGroupKey_t        localGroupKeyInfo;
                                    /* This local struct is filled by RequestGroupCallback() via requestGroupContext. */
                                    /* Interlock allows using it for each of several possible incoming ReqG responses.*/
                OALSecureHMACContext hmacI;
                DOFOperation         reqGroupOp;

                /* It is tempting to place pGroupKeyInfo into requestGroupContext->keyInfo. Cannot, for pGroupKeyInfo
                   holds the most recently accepted (and best) info that is selected from multiple SGM responses. To
                   use that here would have any follow-on SGM response overwrite our currently found "best" selection.*/
                requestGroupContext->keyInfo = &localGroupKeyInfo;
                reqGroupOp = IssueRequestGroupPDU( secureGroup, streamI, KEKInfo, keyRequest, requestGroupContext,
                                                   targetAddress, timeout, &hmacI );
                if (reqGroupOp)
                {
                    /* requestGroupContext is freed within this call or in RequestGroupCallback(). Cannot free it
                       outside those.*/
                    /* Note: This call will initialize the structures at pGroupKeyInfo and at &acceptedResponse */
                    retError = AcceptHighestPriorityRequestGroupResponse( secureGroup, requestGroupContext, KEKInfo,
                                                                          &hmacI, &acceptedResponse, pGroupKeyInfo,
                                                                          reqGroupOp );
                    if (acceptedResponse.filled)
                    {
                        *pHisHighestAPPID = acceptedResponse.appid;
                        *pHisRank = acceptedResponse.rank;
                        *pHisAddress = acceptedResponse.address;   /* Only an exchange of location, no copy required. */
                        /* pGroupKeyInfo holds everything else. */
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                            DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_DEBUG,
                                                               "SGMP RequestGroup response PDU received and accepted.");
                    }
                }
                else
                {
                    DOFMem_Free( requestGroupContext );          /* Here because the normal mechanism did not engage. */
                    PCR_Assert( !acceptedResponse.filled );
                }
                OALStream_Release( streamI, FALSE );
            }
            DOFKeyRequest_Destroy( keyRequest );
        }
        PCR_Assert( acceptedResponse.filled || retError != DOFERROR_OK );
        pGroupKeyInfo->filled = acceptedResponse.filled;
    }
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "RequestGroupAndFillKey() exiting with retError 0x", retError );
#endif

    return retError;
}


/*
**  "Internal Public"  Methods.
*/


/** @internal
    This routine is used to send a request group PDU to secure group managers and then wait for a response. This routine
    sends the request group PDU a few times to increase the chances for a group manager to receive the request group and
    respond with an authorize group PDU.

    @note This blocks for one second for each ReqG PDU sent out.

    @param secureGroup      Contains information related to the secure group.
    @param sendUnicast      TRUE for unicast send of RequestGroup PDU, FALSE for multicast send.
    @param pFillGroupKey    Pointer to groupKey structure to be filled by this call.
    @param kekEpoch         Epoch of the KEK to use in sending RequestGroup PDU.
    @param oneAttempt       TRUE/FALSE: Send/Not send retries of the requestGroup PDU.
    @param pHisRank         Rank of the SGM that answered us.
    @param pHisHighestAPPID Highest APPID of the SGM that answered us.
    @param timeout          The time to wait for each response to the request group.
    @return boolean         TRUE for joined, FALSE for not joined (in which case other return values not valid).
**/
boolean DOFSecureGroup_JoinGroup( DOFSecureGroup secureGroup, boolean sendUnicast, DOFGroupKey_t * pFillGroupKey,
                                  uint16 kekEpoch, boolean oneAttempt, uint8 * pHisRank, uint16 * pHisHighestAPPID,
                                  uint32 timeout )
{
    uint8 n;      /* Try to reach a secure group manager. If a secure group manager is found then fill pFillGroupKey. */

    /* NOTE: Choose timeout large enough to allow a full "send-receive-respond-receive" cycle before sending repeat
             request, because our implementation has our SGM answer every qualified RequestGroup PDU that comes to it.*/
    for (n = 0; n < (oneAttempt ? 1 : REQUESTGROUP_TOTAL_ATTEMPT_COUNT); n++)
    {
        DOFError   callbackStatus;
        DOFAddress hisAddress;                              /* Filled by RequestGroupAndFillKey(). */
        uint8      kekOffset;                               /* Request groupKey using kek we have of specified epoch. */
        uint8      futKEK = (uint8)secureGroup->flags.fut;
        uint8      opKEK  = futKEK ^ 1;

        /* At first connect, secureGroup is not up, so cannot test secureGroup->up here and then fail out. */
        if (sendUnicast && !secureGroup->lastReceivedSGMAddress)
        {
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                    "Cannot send unicast ReqG PDU with NULL lastReceivedSGMAddress." );
            return FALSE;
        }
        if (secureGroup->keks[futKEK].filled && secureGroup->keks[futKEK].epoch == kekEpoch)
            kekOffset = futKEK;
        else
        {
            PCR_Assert( secureGroup->keks[opKEK].filled );
            if (secureGroup->keks[opKEK].epoch == kekEpoch)
                kekOffset = opKEK;
            else
                return FALSE;                               /* Error out, we do not know of a KEK with epoch kekEpoch.*/
        }
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                       "Sending RequestGroup PDU from JoinGroup(), using kekOffset 0x", kekOffset );
        /* Note This call will initialize the structure at pFillGroupKey. */
        callbackStatus = RequestGroupAndFillKey( secureGroup,
                                                 (sendUnicast ? secureGroup->lastReceivedSGMAddress : NULL),
                                                 &secureGroup->keks[kekOffset], pFillGroupKey,
                                                 pHisHighestAPPID, pHisRank, &hisAddress, timeout );
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        {
            if (callbackStatus == DOFERROR_TERMINATED)
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                 "secureGroup went down before JoinGroup() answer came from ReqG PDU" );
            else
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, (callbackStatus == DOFERROR_OK) ?
                                                    "Success after ReqG." : "NO ANSWER after ReqG." );
        }
        if (callbackStatus == DOFERROR_OK)
        {
            PCR_Assert( pFillGroupKey->filled );
            PCR_Assert( pFillGroupKey->streamCcmModeInitialState );
            DOFSecureGroup_UpdateLastReceivedSGMAddress( secureGroup, hisAddress );
            DOFAddress_Destroy( hisAddress );
            return TRUE;
        }
    } /* end for(). */

    return FALSE;
}


/** @internal
    Try to join group. Either joins group and fully accepts key from an SGM, or we become SGM.

    @note This blocks 1 second while accepting possibly multiple ReqG responses.

    @param argc     Number of arguments in argv, presented always as 1.
    @param argv     The secureGroup to join group.
    @param shutdown
**/
void DOFSecureGroup_JoinGroupToDecideToBeSGM( OALThreadPoolItem dnode, boolean shutdown )
{
    SecureGroupContext_t * node = (SecureGroupContext_t *) dnode;

    PCR_Assert( dnode != NULL );       /* Note: This runs in its own thread, brought up with OALQueue_Enqueue() call. */
    PCR_Assert( node->secureGroup );
    PCR_Assert( !node->secureGroup->flags.pointMember );
    PCR_Assert( !node->secureGroup->flags.hub );
    PCR_Assert( node->secureGroup->heartbeatTimerItem );
    if (!shutdown)
    {
        if (node->secureGroup->flags.becomeSGMInPlay)
        {
            boolean doBeSGM = FALSE;

            if (node->secureGroup->up)
            {
                DOFGroupKey_t fillGroupKey;
                uint8         kekOffset;
                uint8         hisRank;
                uint16        hisHighestAPPID;
                boolean       joined;
                uint8         futKEK = node->secureGroup->flags.fut;
                uint8         opKEK  = futKEK ^ 1;

                if (node->secureGroup->keks[futKEK].filled)
                    kekOffset = futKEK;
                else
                {
                    PCR_Assert( node->secureGroup->keks[opKEK].filled );
                    kekOffset = opKEK;
                }
                PCR_Assert( node->secureGroup->heartbeatTimerItem );
                /* At success, JoinGroup fills fillGroupKey.streamCcmModeInitialState. We consume it. */
                /* Start process of becoming SGM. First see if any SGM at all has risen to answer our multicast join. */
                joined = DOFSecureGroup_JoinGroup( node->secureGroup, FALSE,
                                                   &fillGroupKey, /* _JoinGroup() memsets this to zeroes. */
                                                   node->secureGroup->keks[kekOffset].epoch, TRUE, &hisRank,
                                                   &hisHighestAPPID, REQUESTGROUP_SEND_TIMEOUT );
                /* The above "FALSE" param has us send multicast. With some logical reasons, we could have chosen to
                   send unicast (to our recenty known SGM), but we have decided that to multicast gives the benefit of
                   catching a very recently available SGM, which helps better avoid having multiple SGMs. */
                /* Note: JoinGroup will take 1 second when we do not join, because we try once and timeout is 1 second.
                         The earliest successful return happens only after a full second of watching for multiple
                         possible returns and keeping only the one response with highest priority. */
                if (node->secureGroup->up)
                {
                    OALCipher cipher = ((OALChannel)node->secureGroup)->remoteChannelData->cipher;

                    PCR_Assert( node->secureGroup->heartbeatTimerItem );
                    /* A hb could have come in while we were involved in JoinGroup(). If hb is from lower priority node,
                       then ->flags.becomeSGMInPlay is still set and we are poised to become SGM now. If from a tied or
                       higher priority node, then ->flags.becomeSGMInPlay is cleared and we do not become SGM. */
                    if (node->secureGroup->flags.becomeSGMInPlay &&
                        OALCipher_IsNodeIDValid( cipher ))         /* nodeID can become invalid at any moment. */
                    {
                        doBeSGM = TRUE;                            /* Subject to step down test. */
                        if (joined)
                        {
                            uint8 stepDown;

                            /* Use SGM-delivered key NOW. We might gently takeover, but at least we use their key. */
                            /* Use _SetSlotKey() with 0 delay to avoid state loss attendant w/ nearly same _SetCurKey().
                               If fillGroupKey is new to us, we apparently never heard the rekey PDU.
                               If fillGroupKey is the same key we know, we apparently missed the hb. */
                            if (!OALCipher_SetSlotKey( cipher, &node->secureGroup->lastReceivedSGMAddress,
                                                       fillGroupKey.streamCcmModeInitialState, fillGroupKey.key, 0,
                                                       node->secureGroup, TRUE, fillGroupKey.receivedMulticast ))
                            { /* fillGroupKey not necessarily a new random key. If here then the key is already stored*/
                                doBeSGM = FALSE;
                                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                       "_SetSlotKey() not accepted (from _JoinGroupToDecideToBeSGM)." );
                            }
                            else
                            {
                                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                 "Stored slotKey from other SGM(from from _JoinGroupToDecideToBeSGM)" );
                                stepDown = DOFSecureGroup_StepDownTest( node->secureGroup, hisHighestAPPID, hisRank );
                                if (stepDown == NOT_STEPDOWN && OALCipher_IsNodeIDValid( cipher ))
                                { /* doBeSGM is TRUE. Gently takeover. Fall thru to nuke other SGM with my blazing hb.*/
                                }
                                else
                                {
                                    doBeSGM = FALSE;
                                    if (node->secureGroup->up != FALSE)  /* We joined group content that he stay SGM. */
                                    {
                                        if (DOFSecureGroup_ExtendHeartbeatTimeout( node->secureGroup ) == FALSE)
                                        { /* How to notify? Cannot prevent hb to and will wrongly try to be SGM.*/
                                        }
                                    }
                                    else
                                    { /* This is our very 1st join effort. No heartbeat timer item yet. */
                                    }
                                }
                            }
                        }
                        if (doBeSGM)
                        {
                            uint32 highestAPPIDVersionBit;
                            PCRMutex mutex = OALMutexPool_Get( node->secureGroup );

                            PCR_Assert( myHighestAPPIDBit );
                            if (joined)
                            {
                                highestAPPIDVersionBit = DOFSecureGroup_ComputeAPPIDVersionBit( hisHighestAPPID );
                                /* Could be 0. */
                                /* We still use hisHighestAPPID's highestAPPIDVersionBit so present SGM gets our hb. */
                                if (!highestAPPIDVersionBit)/*Requires bit to have hb go out. We want that in takeover*/
                                {
                                    /* If nobody uses myHighestAPPIDBit, agreed that resultant hb's are wasted effort.*/
                                    highestAPPIDVersionBit = myHighestAPPIDBit;
                                }
                            }
                            else
                                highestAPPIDVersionBit = myHighestAPPIDBit;
                            PCRMutex_Lock( mutex );
                            node->secureGroup->flags.mustSlotRekey = 1; /*As SGM declare slotRekey to GetKEK_timeout()*/
                            PCRMutex_Unlock( mutex );
                            OALMutexPool_Release( node->secureGroup );
                            PCR_Assert( node->secureGroup->heartbeatTimerItem );
                            if (DOFSecureGroup_InitiateMySGMBehavior( node->secureGroup, highestAPPIDVersionBit, TRUE )
                                                                                                               != FALSE)
                            {
                                /* Order of events when 1st HeartbeatTimeoutHandler() is called in 1 msec:
                                    Send hb using present group key.
                                    Check KEK.
                                    If new KEK, form new kekKey and send it out in epochRekey (using present        .
                                    groupKey). Form new slotKey, send it in slotRekey (using present groupKey)      .
                                    Both newKeys eventually promote at my and all other nodes. */
                                /* This may be ordinary case where we step up as SGM or simple takeover. In either case
                                   we should be aware that old SGM may not have sent epochRekey. */

                                /* Details about EpochRekey PDU for new epoch, when we step us the new SGM:
                                    CASE 1: node->secureGroup->flags.noEpochRekeySinceNewEpochFound is SET
                                       Last SGM did not send a epochRekey PDU for this epoch, or we did not hear it.
                                       When we soon chkKEK in GetKEKTimeout_fn(), and if we so happen to find a new KEK,
                                       it may be that all the other group members already know about it and already have
                                       an epochRekey for it from the last SGM. But we do not know whether that happened
                                       or not, and irregardless of it we will form a new random kekKey and send it out.
                                       If others do have a kekKey from the 1st SGM, ours will override because we are a
                                       new SGM.
                                    CASE 2: node->secureGroup->flags.noEpochRekeySinceNewEpochFound is CLEAR
                                       Because we did get an epoch rekey, we either already have a kekKey in our cipher
                                       or we do not. If we do, with our new role as SGM we will automatically receive
                                       notification from cipher that we should send the cleanup epochRekey PDU. If there
                                       is not a kekKey in cipher it is because it promoted already and that means that
                                       the old SGM already sent the cleanup epochRekey (or if it did not get it sent
                                       then we do not know how long ago the kekKey promoted and it is therefore wrong
                                       for us to send out a "substitute" cleanup epochRekey. */
                            }
                            else
                            {
                                /* Could try to reschedule becoming SGM but why bother? We aren't ready right now, and
                                   somebody else is SGM or will step up on their own. If not we will try again later. */
                                doBeSGM = FALSE;
                                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                               "InitiateMySGMBehavior() failed in JoinGroupToDecideToBeSGM, may try to be SGM later." );
                            }
                        }
                    }
                    else
                    { /* hb apparently came in, or cipher is not ready for transmit, so do nothing here. */
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ) && joined)
                            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
               "_JoinGroupToDecideToBeSGM() decides not to use the reqG response, no longer interested in being SGM." );
                    }
                }
                else
                { /* secureGroup connection went down during JoinGroup(). */
                }
                if (joined)
                    OALStream_Release( fillGroupKey.streamCcmModeInitialState, TRUE );
            }
            if (!doBeSGM)
                DOFSecureGroup_ChangeSGMState(node->secureGroup, !BE_SGM);   /* Something prevents us from being SGM. */
        }
        else
        { /* incoming hb apparently cleared .becomeSGMInPlay, so do not be SGM. */
        }
    }
    else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                 "JoinGroupToDecideToBeSGM called with shutdown TRUE, so thread closes without work." );
    DOFMem_Free(node);
}


/* @internal
   Either steps us down and fully accepts his key, or not.

   @note This blocks for 1 second waiting for possibly multiple ReqG responses to come back.
*/
void DOFSecureGroup_JoinGroupToDecideToStepDownAsSGM( DOFSecureGroup secureGroup )
{
    PCR_Assert( !secureGroup->flags.pointMember );
    PCR_Assert( !secureGroup->flags.hub );
    if (secureGroup->flags.activeSGM &&              /* DOFSecureGroup_HeartbeatReceiveHandler may have cleared this. */
        secureGroup->up)
    {
        boolean       joined;
        uint16        hisHighestAPPID;                        /* Filled by JoinGroup(). */
        DOFGroupKey_t fillGroupKeyInfo;                       /* Same. */
        uint8         hisRank;                                /* Same. */
        uint8         kekOffset;
        uint8         futKEK = secureGroup->flags.fut;
        uint8         opKEK  = futKEK ^ 1;

        if (secureGroup->keks[futKEK].filled)
            kekOffset = futKEK;
        else
        {
            PCR_Assert( secureGroup->keks[opKEK].filled );
            kekOffset = opKEK;
        }
        /* At success, this next fills fillGroupKeyInfo.streamCcmModeInitialState. We must consume it. */
        joined = DOFSecureGroup_JoinGroup( secureGroup, FALSE,
                                           &fillGroupKeyInfo,  /* _JoinGroup() memsets this to zeroes. */
                                           secureGroup->keks[kekOffset].epoch,
                                           TRUE, &hisRank, &hisHighestAPPID, REQUESTGROUP_SEND_TIMEOUT );
        /* Note: JoinGroup will take 1 second when we do not join, because we try once and timeout is 1 second.
                 The earliest successful return happens only after a full second of watching for multiple
                 possible returns and keeping only the one response with highest priority. */
        if (joined && secureGroup->flags.activeSGM &&                     /* Incoming packet handlers can clear this. */
            secureGroup->up)
        {
            /*Other SGM has my KEK. Shall I step down? */
            uint8 stepDown = DOFSecureGroup_StepDownTest( secureGroup, hisHighestAPPID, hisRank );

            PCR_Assert( hisHighestAPPID >= SGMP_V1_APPID);
            /* Note: I am SGM so do not use successful receipt of ReqG response PDU to extend my hb timeout. */
            PCR_Assert( fillGroupKeyInfo.filled );
            PCR_Assert( fillGroupKeyInfo.streamCcmModeInitialState );
            if (stepDown == TIE_STEPDOWN)
            {
                uint8       period;                               /* Use group key to perhaps tie break. */
                uint16      slot;
                int         compare;                              /* Must be int !!!! */
                OALCryptKey myGroupKey, hisKeyXOR, myKeyXOR;

                /* Get current group key from cipher, XOR it, compare, and decide on stepping down. */
                OALCipher_GetKey( ((OALChannel)secureGroup)->remoteChannelData->cipher, CUR, myGroupKey, &period,
                                  &slot );
                PCRMem_Copy( &hisKeyXOR[0], &fillGroupKeyInfo.key[0], sizeof( OALCryptKey ) );
                PCRMem_Copy( &myKeyXOR[0], &myGroupKey[0], sizeof( OALCryptKey ) ); /* Spec says compare by XORs. */
                DOFSecureGroup_XORTwoKeys( &secureGroup->keks[opKEK].KEK[0], &hisKeyXOR[0], &myGroupKey[0]);
                compare = PCRMem_Compare( &hisKeyXOR[0], &myGroupKey[0], sizeof( OALCryptKey ) );
                if (compare)
                    stepDown = (compare > 0) ? DO_STEPDOWN : NOT_STEPDOWN;
                PCR_Assert( compare || stepDown == TIE_STEPDOWN );
            }
            if (stepDown == NOT_STEPDOWN && secureGroup->up &&
                secureGroup->flags.activeSGM)                         /* Incoming packet handlers can clear this. */
            {
                /* At the very least, he will hear our RequestGroupResponse PDU soon and will step down as SGM. Enable
                   our hb tx at hisHighestAPPID. */
                DOFSecureGroup_AddVersionToAPPIDTable( secureGroup, hisHighestAPPID );
                /* Note: If we do not know hisHighestAPPID, we should have lost to him. */
                /* In case he has same key, send hb to him so he steps down NOW, so he stops letting new members join.*/
                /* REVIEW: Send at hisHighestAPPID instead of at DOFSECUREGROUP_MY_HIGHEST_APPID. */
                PCR_Assert( DOFSecureGroup_GetSecurityScope( secureGroup ) );
                SGMP_V1_SendHeartbeat( (OALChannel)secureGroup, secureGroup->lastReceivedSGMAddress,
                                       DOFSecureGroup_GetSecurityScope( secureGroup ), DOFSECUREGROUP_MY_HIGHEST_APPID,
                                       DOFSecureGroup_GetRank( secureGroup ), secureGroup->tieBreaker );
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                {
                    uint16 txSlot;

                    OALCipher_GetCurKeySlot( ((OALChannel)secureGroup)->remoteChannelData->cipher, &txSlot );
                    DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "Unicast Heartbeat V1 sent of slot 0x",
                                               (uint32)txSlot );
                }
            }
            else if (secureGroup->up && secureGroup->flags.activeSGM) /* Incoming packet handlers can clear .activeSGM*/
            {
                /* Step down, even if still tied (which means keys are the same). If TIE_STEPDOWN, do nothing more.
                   If DO_STEPDOWN, merge my group by sending nowRekey PDU (on receipt, they will invalidate their
                   nodeID and pick a new one, whereas I invalidate my nodeID just after I send the nowRekey PDU. */
                PCR_Assert( !secureGroup->flags.becomeSGMInPlay );
                if (stepDown == DO_STEPDOWN)
                {
                    OALCipher cipher = ((OALChannel)secureGroup)->remoteChannelData->cipher;

                    /* I lost, he has different group key, I joined him. So get my group members into his group. */
                    PCREvent eventRekeyPDU = PCREvent_Create( FALSE );

                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                 "<><><><As SGM, I lost to other SGM and have different keys. I am merging my group." );
                    OALCipher_DisableAllPendingRekeys( cipher );
                    if (eventRekeyPDU)
                    {
                        OALStream streamCommonCcmModeInitialState =        /* We consume this returned stream, below. */
                              OALCipher_GetSecureGroupCommonState( cipher, fillGroupKeyInfo.streamCcmModeInitialState );

                        if (streamCommonCcmModeInitialState)
                        {
                            /* I am still SGM for a short time- I can get my nodeID even if it becomes invalid somehow*/
                            DOFError status; /* Event guarantees rekeyMerge PDU goes out before we set new key.
                                                After our SGMP RekeyMerge PDU encrypts, OALConnectedProtocolStack_Send()
                                                calls OALConnectionlessProtocolStack_SGMP_KeyChangedCallback(), which
                                                signals our event. PCREvent() has no refCount- we created it and we
                                                destroy it. */

                            PCR_Assert( DOFSecureGroup_GetSecurityScope( secureGroup ) );
                            /* This sent with my present cipher nodeID and curKey. */
                            if ((status = SGMP_V1_SendRekeyMerge( (OALChannel)secureGroup, NULL,
                                                                  DOFSecureGroup_GetSecurityScope( secureGroup ),
                                                                  eventRekeyPDU, streamCommonCcmModeInitialState,
                                                                  fillGroupKeyInfo.key )) == DOFERROR_OK )
                            {
                                PCRStatus pcrStatus = PCREvent_Wait( eventRekeyPDU, 1000 ); /* 1 sec ought encrypt PDU*/

                                if (pcrStatus == PCRSTATUS_TIMEOUT)
                                { /* Since PDU never even got to EPP, implement some kind of resend? */
                                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                                        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                                            "I failed to send mergeRekey PDU." );
                                }
                                else
                                {
                                    PCR_Assert( pcrStatus == PCRSTATUS_OK );  /* event says PDU is thru _Encipher(). */
                                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                                        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                                            "I sent mergeRekey PDU multicast." );
                                }
                            }
                            else if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                            {
                                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG,
                             "DOFSecureGroup_JoinGroupToDecideToStepDownAsSGM() failed to send SGMP rekey merge PDU." );
                            }
                            OALStream_Release( streamCommonCcmModeInitialState, FALSE );
                        }
                        PCREvent_Destroy( eventRekeyPDU );
                    }
                    else if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG,
                                                            "ERROR: Failed to send mergeRekey PDU multicast." );
                    /* Work on myself now: step down as SGM (which also tells cipher he is not manager), remove cipher
                       state, set new curKey with new nodeID, set surviving SGM state, and send to him unicast surrender
                       SGMP.Heartbeat PDU. */
                    DOFSecureGroup_ChangeSGMState( secureGroup, !BE_SGM );
                    if (DOFSecureGroup_ExtendHeartbeatTimeout( secureGroup ) == FALSE)
                    { /* Error. How to notify somebody that we cannot extend hb timeout? */
                    }
                    OALCipher_RemoveState( cipher );        /* Cipher also loses knowledge of his manager. */
                    /* This next gives me new nodeID from surviving SGM (requires cipher not be manager). */
                    if ((OALCipher_SetCurKey( cipher, &secureGroup->lastReceivedSGMAddress,
                                              fillGroupKeyInfo.streamCcmModeInitialState,
                                              fillGroupKeyInfo.key, fillGroupKeyInfo.receivedMulticast )))
                    {
#if 0 /* Done above in _SetCurKey() call. This #if 0 purposely kept as best way to document what is happening. */
                        /*fillGroupKeyInfo.streamCcmModeInitialState has SGM state. Get it into cipher tables. */
                        OALCipher_SetNodeState( cipher, &secureGroup->lastReceivedSGMAddress,
                                                fillGroupKeyInfo.streamCcmModeInitialState );
                        /* Send unicast hb to surviving SGM, so he sends his heartbeat. I MUST lose so he will send it*/
                        /* We can do this because we just acquired his DOFAddress as we received his ReqG Resp PDU. */
                        /* We want him to send his hb to everybody, particularly to my merging group members, that they
                           more quickly will know his DOFAddress, and can thus more quickly ask for their new nodeID. */
#endif
                         /*DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "Stored curKey 1st 4 bytes 0x",
                                                      (uint32)(*((unsigned int *)&fillGroupKeyInfo.key[0])) );*/
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                      "Sending unicast surrender SGMP.Heartbeat to surviving manager" );
                        PCR_Assert( DOFSecureGroup_GetSecurityScope( secureGroup ) );
                        /* REVIEW: Send at hisHighestAPPID instead of at DOFSECUREGROUP_MY_HIGHEST_APPID. */
                        SGMP_V1_SendHeartbeat( (OALChannel)secureGroup, secureGroup->lastReceivedSGMAddress,
                                               DOFSecureGroup_GetSecurityScope( secureGroup ),
                                               DOFSECUREGROUP_MY_HIGHEST_APPID,
                                               DOFSecureGroup_GetRank( secureGroup ), 0 ); /* 0: tieBreaker so I lose.*/
                        /*Even though I know my new SGM and could inform cipher, let his return mcast HB inform cipher*/
                    }
                    else if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG,
                                               "DOFSecureGroup_JoinGroupToDecideToStepDownAsSGM() failed to set key." );
                }
                DOFSecureGroup_ChangeSGMState( secureGroup, !BE_SGM );
            }
        }
        if (joined)
            OALStream_Release( fillGroupKeyInfo.streamCcmModeInitialState, TRUE );
    }
}


void DOFSecureGroup_XORTwoKeys( OALCryptKey src, OALCryptKey dest1, OALCryptKey dest2 )
{
    uint8 i;

    for (i = 0; i < sizeof( OALCryptKey ); i++)
    {
        if (dest1)
            dest1[i] ^= src[i];
        if (dest2)
            dest2[i] ^= src[i];
    }
}


/**
@}
**/
