/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/error.h>
#include <internal/exception.h>
#include <internal/mutex_pool.h>
#include <internal/operationinternal.h>
#include <internal/permissionsetinternal.h>
#include <internal/remotechannel.h>
#include <internal/resultinternal.h>
#include <internal/security_common.h>
#include <internal/timeinternal.h>
#include <internal/trp_v1.h>
#include "securegroupprivate.h"


/** @addtogroup DOFSecureGroup
@{
**/


/*
**
**  "Private" Type and Method Declarations
*/


#define ILLEGAL_SSID                        0xFFFFFFFF


typedef struct oapSendInfo_t
{
    DOFOperation pendingOp;
    DOFResult    pendingResult;
} oapSendInfo_t;


typedef struct extendPerms_t
{
    DOFOperation   pendingOp;
    DOFResult      pendingResult;
    DOFOperation   provOp;
    DOFAddress     address;
    DOFSecureGroup secureGroup;
    boolean        requestNotResponse;
} extendPerms_t;


typedef struct extendTheirPerms_t
{
    DOFSecureGroup        secureGroup;
    void                * context;
    ExtendCompleteCB      extendCompleteCallback;
} extendTheirPerms_t;


typedef struct RequestKekContext_t
{
    reqKEvent_t * pReqKMutexSyncObject;
    PCREvent      event;
    boolean       filledByCallback;
    DOFKEKInfo    kekInfo;
    DOFError      callbackStatus;
    DOFError      trpStatus;                           /* Filled and valid only when ->callbackStatus is DOFERROR_OK. */
} RequestKekContext_t;


/* @internal */
static uint32 MillisDelayForOneMoreEpochCheck( DOFSecureGroup secureGroup )
{
    uint32 bailMillis = OALTimerItem_TimeRemaining( secureGroup->grantTimerItem );

    if (bailMillis)
    {
        if (bailMillis > SAFE_GET_EPOCH_MILLISECONDS)
        {
            /* Even suffering through various thread delays, we very likely can schedule out in the future to get
               at least one more epoch check. */
            return bailMillis - SAFE_GET_EPOCH_MILLISECONDS;
        }
        else
            return 1;      /* Urgent- schedule epoch check for essentially now. */
    }
    else
        return  0;         /* As soon as its timer thread starts up, we bail! So don't encourage another epoch check. */
}


/** @internal
    Cleanup the given KEKInfo structure of the given secureGroup.

    @param pKEKInfo Pointer to the given KEKInfo structure.
**/
static void CleanupKEKInfo( DOFKEKInfo_t * pKEKInfo )
{
    if (pKEKInfo->filled)
    {
        DOFPermissionSet_Destroy( pKEKInfo->otherPerms );
        DOFPermissionSet_Destroy( pKEKInfo->myPerms );
        DOFSecurityMode_Destroy( pKEKInfo->securityModeConfig );
        DOFSecurityScope_Destroy( pKEKInfo->securityScope );
        DOFObjectIDSource_Destroy( pKEKInfo->sid );
    }
}


/* hub: NULL returns. point: secureGroup returns. member: ->baseChannel returns.*/
static OALChannel DetermineFallbackChannel( DOFSecureGroup secureGroup )
{
    OALChannel fallbackChannel = NULL;

    if (secureGroup->flags.pointMember)
        fallbackChannel = (OALChannel)secureGroup;
    else if (!secureGroup->flags.hub)
    {
        /* secureGroup's sendServer channel. sendServer is a "unicast" server for listening purposes, but when
           submitting to it a send DOFAddress of IPaddr (typically) 224.0.23.46:5567, it will send out multicast. */
        fallbackChannel = ((OALChannel)secureGroup)->remoteChannelData->baseChannel;
    }

    return fallbackChannel;
}


/* hub: NULL returns. point: ->lastReceivedSGMAddress returns. member: ->remoteChannelData->address returns. */
static DOFAddress DetermineFallbackAddress( DOFSecureGroup secureGroup )
{
    DOFAddress fallbackAddress = NULL;

    PCR_Assert( secureGroup );
    if (secureGroup->flags.pointMember)
    {
        PCR_Assert( secureGroup->lastReceivedSGMAddress );
        fallbackAddress = secureGroup->lastReceivedSGMAddress;
    }
    else if (!secureGroup->flags.hub)
    {
        /* secureGroup's listenServer DOFAddress. listenServer is a "multicast" server for listening purposes, and
           listens at its DOFAddress IPaddr of (typically) 224.0.23.46:5567. Submitting this DOFAddress to a send
           server will output multicast packet. */
        fallbackAddress = ((OALChannel)secureGroup)->remoteChannelData->address;
    }

    return fallbackAddress;
}


/*
    @note Does not update secureGroup->opKEKInfo.
    Return TRUE if pNewKEKInfo used to update secureGroup->futKEKInfo.
*/
static boolean UpdateFutureKEKInfo( DOFSecureGroup secureGroup, DOFKEKInfo_t * pNewKEKInfo )
{
    uint16  greatestKnownEpoch;
    boolean retBoolean = FALSE;

    PCR_Assert( pNewKEKInfo->filled );
    /* OALCipher_AdjustKeys() locks cipher while promoting futKEK, and so shall we while we update futKEK. */
    OALCipher_Lock( ((OALChannel)DOFSecureGroup_GetChannel_Internal( secureGroup ))->remoteChannelData->cipher );
    if (DOFSecureGroup_GetGreatestKnownEpoch( secureGroup, &greatestKnownEpoch ) == FALSE ||
        DOFTime_Compare_16Bit( pNewKEKInfo->epoch, greatestKnownEpoch ) > 0)
    {
        boolean  futKEKReplaced = FALSE;
        uint8    futKEK         = secureGroup->flags.fut;
        uint32   seed           = PCRRandom_Get();
        PCRMutex mutex          = OALMutexPool_Get( secureGroup );

        if (secureGroup->keks[futKEK].filled)
        {
            /* pNewKEKInfo's epoch greater than ours so accept into futKEK. */
            DOFSecureGroup_CleanupKEKInfoAndDisable( &secureGroup->keks[futKEK] );
            futKEKReplaced = TRUE;
        }
        PCRMutex_Lock( mutex );
        secureGroup->flags.noEpochRekeySinceNewEpochFound = 1;
        PCRMutex_Unlock( mutex );
        OALMutexPool_Release( secureGroup );
        secureGroup->lastAcceptedKEKEpoch = pNewKEKInfo->epoch;
        PCRMem_Copy( &secureGroup->keks[futKEK], pNewKEKInfo, sizeof( DOFKEKInfo_t ) );/*Copy ENTIRE kekInfo structure*/
        PCR_Assert( secureGroup->keks[futKEK].filled );
        PCR_Assert( pNewKEKInfo->securityScope );
        PCR_Assert( pNewKEKInfo->securityModeConfig );

        /* Since we accept known random KEK, seed random number generator with XOR'ed high 32 bits of that randomness.*/
        PCRRandom_Seed( seed ^ *(uint32 *)secureGroup->keks[futKEK].KEK);
        retBoolean = TRUE;
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        {
            if (futKEKReplaced)
                DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                           "Our older futKEK is replaced by an accepted futKEK of epoch 0x",
                                           secureGroup->keks[futKEK].epoch );
            else
                DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                           "Our empty futKEK is filled by an accepted futKEK of epoch 0x",
                                           secureGroup->keks[futKEK].epoch );
        }
    }
    else
    { /* Our KEK info has epoch the same or greater than that submitted to us in pNewKEKInfo. Ignore. */
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "KEK obtained but NOT new." );
        CleanupKEKInfo( pNewKEKInfo ); /* Does not disable pNewKEKInfo, just cleans it up. */
    }
    OALCipher_Unlock( ((OALChannel)DOFSecureGroup_GetChannel_Internal( secureGroup ))->remoteChannelData->cipher );

    return retBoolean;
}


/** @internal
    The callback provided to receive the KEK from the call to TRP_RequestKEK

    @note Function prototype is trp_callback_fn.
    @param  trp_context The return information from TRP.
    @param  ctx         The location to save the information provided by TRP.
**/
static void GetKEKCallback( TRP_V1_Context trp_context, void * ctx )
{
    if (ctx)
    {
        reqKEvent_t *         pReqKMutexSyncObject;
        RequestKekContext_t * pContext       = ctx;
        boolean               cleanupContext = FALSE;

        PCR_Assert( pContext );
        PCR_Assert( pContext->pReqKMutexSyncObject );
        pReqKMutexSyncObject = pContext->pReqKMutexSyncObject;
        OALMutexPool_Lock( pReqKMutexSyncObject );             /* GetKEKUsingTRP(), DenyGroupBehavior() also lock it. */
        if (pContext->event)
        {
            /* We now fill pContext and we will signal GetKEKUsingTRP(). */
            pContext->kekInfo->filled = FALSE;
            PCR_Assert( !pContext->filledByCallback );
            pContext->filledByCallback = TRUE;
            if (trp_context && trp_context->KEKReturn)
            {
                pContext->callbackStatus = trp_context->KEKReturn->status;
                pContext->trpStatus = trp_context->KEKReturn->errorCode;/*Only useful for ->callbackStatus DOFERROR_OK*/
                /* Look at the response from TRP and fill in the DOFKEKInfo struct only if we got back a valid KEK. */
                if (pContext->callbackStatus != DOFERROR_OK)
                {
                    /* TRP PDU was sent but we got no response back from AS. This path is AS timeout, and nothing else*/
                    PCR_Assert( pContext->callbackStatus == DOFERROR_TIMEOUT );
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                            "GetKEKCallback() sees TRP ->status as TIMEOUT.");
#endif
                }
                else if (trp_context->KEKReturn->errorCode == TRP_V1_SUCCESS)
                {
                    /* Add iamPermission to pContext->kekInfo->otherPerms and to ->myPerms, then return successful. */
                    DOFPermission iamPerm;
                    boolean       success = FALSE;

                    PCR_Assert( trp_context->KEKReturn->securityMode ); /* ONLY this path fills pContext (DOFKEKInfo).*/
                    PCR_Assert( trp_context->KEKReturn->secScope );
                    if ((iamPerm = DOFPermissionIAm_Create( trp_context->KEKReturn->sid )) != NULL)
                    {
                        success =
                          DOFPermissionSet_AddTrustedPermission( trp_context->KEKReturn->groupsPermissions, iamPerm ) &&
                          DOFPermissionSet_AddTrustedPermission( trp_context->KEKReturn->initPermissions, iamPerm );

                        DOFPermission_Destroy( iamPerm );
                        if (success)
                        {
                            PCR_Assert( pContext->kekInfo->myPerms == NULL );
                            PCR_Assert( pContext->kekInfo->otherPerms == NULL );
                            PCR_Assert( pContext->kekInfo->sid == NULL );
                            PCR_Assert( pContext->kekInfo->securityModeConfig == NULL );
                            PCR_Assert( pContext->kekInfo->securityScope == NULL );
                            pContext->kekInfo->myPerms =
                                                   DOFPermissionSet_Copy( trp_context->KEKReturn->initPermissions );
                            pContext->kekInfo->otherPerms =
                                                   DOFPermissionSet_Copy( trp_context->KEKReturn->groupsPermissions );
                            pContext->kekInfo->sid = DOFObjectIDSource_Copy( trp_context->KEKReturn->sid );
                            pContext->kekInfo->securityModeConfig =
                                                          DOFSecurityMode_Copy( trp_context->KEKReturn->securityMode );
                            pContext->kekInfo->securityScope = DOFSecurityScope_Copy(trp_context->KEKReturn->secScope);
                            /* Check success of the above retains/copies. */
                            if ((pContext->kekInfo->myPerms            || !trp_context->KEKReturn->initPermissions) &&
                                (pContext->kekInfo->otherPerms         || !trp_context->KEKReturn->groupsPermissions) &&
                                (pContext->kekInfo->sid                || !trp_context->KEKReturn->sid) &&
                                (pContext->kekInfo->securityModeConfig || !trp_context->KEKReturn->securityMode) &&
                                (pContext->kekInfo->securityScope      || !trp_context->KEKReturn->secScope))
                            {
                                /* pContext->callbackStatus is still DOFERROR_OK. Finish filling pContext and exit. */
                                pContext->kekInfo->epoch = trp_context->KEKReturn->epoch;
                                PCRMem_Copy( &pContext->kekInfo->KEK, trp_context->KEKReturn->KEK,
                                             sizeof( pContext->kekInfo->KEK ) );
                                pContext->kekInfo->Thb = trp_context->KEKReturn->thb;
                                pContext->kekInfo->Tmin = trp_context->KEKReturn->tmin;
                                pContext->kekInfo->Tmax = trp_context->KEKReturn->tmax;
                                pContext->kekInfo->stateID = trp_context->KEKReturn->stateID;
                                pContext->kekInfo->filled = TRUE;
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
                                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                                    DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_TRACE, "GetKEKCallback() succeeds.");
#endif
                            }
                            else
                            {
                                success = FALSE;
                                /* Cleanup what we retained/copied. 1 or more will be NULL. */
                                DOFPermissionSet_Destroy( pContext->kekInfo->myPerms );
                                DOFPermissionSet_Destroy( pContext->kekInfo->otherPerms );
                                DOFObjectIDSource_Destroy( pContext->kekInfo->sid );
                                DOFSecurityMode_Destroy( pContext->kekInfo->securityModeConfig );
                                DOFSecurityScope_Destroy( pContext->kekInfo->securityScope );
                            }
                        }
                    }
                    PCR_Assert( pContext->callbackStatus == DOFERROR_OK );
                    if (!success)
                        pContext->callbackStatus = DOFERROR_INSUFFICIENT_RESOURCES;
                }
                else
                {
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                        DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "GetKEKCallback() gets back AS errorCode 0x",
                                                   (uint32)trp_context->KEKReturn->errorCode );
#endif
                    PCR_Assert( pContext->callbackStatus == DOFERROR_OK );   /* Tainted with bad pContext->trpStatus. */
                }
            }
            else
            {
                /* Here if TRP router had no route and we gave NULL/NULL for fallback channel/address. It sent no PDU.*/
                pContext->callbackStatus = DOFERROR_SERVICE_UNAVAILABLE;/*TRP is the unavail service. AS'es may abound*/
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ) && EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING)
                    DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_TRACE, "Error: GetKEKCallback() got no return data");
            }
            /* pContext->callbackStatus is filled for every path that gets here, where DOFERROR_OK means TRP actually
               received something back from the AS. However, DOFERROR_OK can be tainted with bad pContext->trpStatus. */
            PCR_Assert( !pContext->kekInfo->filled ||
                        (pContext->callbackStatus == DOFERROR_OK && pContext->trpStatus == DOFERROR_OK) );
            PCR_Assert( pContext->kekInfo->filled ||
                        (pContext->callbackStatus != DOFERROR_OK || /* For this, pContext->trpStatus has no meaning. */
                         pContext->trpStatus != DOFERROR_OK) );     /*Otherwise, pContext->trpStatus is very important*/
            PCREvent_Signal( pContext->event );
        }
        else
        {
            /* We have not filled pContext in any way. */
            cleanupContext = TRUE;                          /* Because we do not signal code that cleans up pContext. */
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                 "GetKEKCallback finds no PCREvent, does nothing but cleanup context" );
#endif
        }
        /* As we leave, our ownership of ->securityMode, ->secScope, ->initPermissions, ->sid, and ->groupsPermissions:
                   for     DOFERROR_OK: we bumped their ref counts.
                   for not DOFERROR_OK: we did not. */
        PCR_Assert( pContext );
        PCR_Assert( pContext->pReqKMutexSyncObject );
        PCR_Assert( pReqKMutexSyncObject );
        OALMutexPool_Unlock( pReqKMutexSyncObject ); /* NOTE! If instead use context->pReqKMutexSyncObject, can crash.*/
        if (cleanupContext)
            DOFMem_Free( pContext );
    }
    else
    { /* Here only rarely. GetKEKUsingTRP() cleaned up and wants us not to cleanup and not to signal. */
    }
}


/** @internal
    Get KEK from AS using TRP request.

    @note This call blocks for good answer up to 15 secs.
          After work concludes, pObtainedKEKInfo is filled and DOFERROR_OK is returned.

    @param secureGroup
    @param group            The group that should be joined (an Object ID).
    @param permissionSet    The permission set for which we are asking authentication.
    @param creds            The credentials to use.
    @param pObtainedKEKInfo An out param ptr. If this call is successful by obtaining a new KEK, fill
                            the structure with all of the information provided by TRP.
    @param pMostDescriptiveDOFException An out param ptr. If this call not successful in obtaining a new KEK, fill with
                            the DOFException that best describes the reason for failure.
    @return DOFERROR_ILLEGAL_ARGUMENT if given @a group, @a permissionSet, @a creds, or @a pObtainedKEKInfo are NULL, or
            DOFERROR_INTERNAL_ERROR if memory could not be allocated or other resources could not be obtained, or
            DOFERROR_ACCESS_DENIED if TRP did not return the KEK from the AS, or
            DOFERROR_OK if we successfully initiated work to get KEK.
**/
static DOFError GetKEKUsingTRP( DOFSecureGroup secureGroup, DOFObjectIDAuthentication group,
                                DOFPermissionSet permissionSet, DOFCredentials creds, DOFKEKInfo_t * pObtainedKEKInfo,
                                DOFException * pMostDescriptiveDOFException )
{
    /* Our ONLY caller is GetKEK(). He can't call again until we return to him. */
    boolean  leaveGroup = FALSE;
    DOFError retError   = DOFERROR_ILLEGAL_ARGUMENT;

    PCR_Assert( secureGroup );
    PCR_Assert( pMostDescriptiveDOFException );
    *pMostDescriptiveDOFException = NULL;                           /* Filled only when we do not return DOFERROR_OK. */
    if (group && permissionSet && creds && pObtainedKEKInfo)
    {
        DOFKeyRequest keyRequest;

        PCR_Assert( DOFCredentials_IsResolved( creds ) );
        retError = DOFERROR_INSUFFICIENT_RESOURCES;
        if ((keyRequest = DOFKeyRequest_Create( creds, REQUEST_KEK_NONCE_SIZE, TRUE, FALSE, NULL, permissionSet )) != NULL)
        {
            OALChannel fallbackChannel     = DetermineFallbackChannel( secureGroup );
            DOFAddress fallbackSendAddress = DetermineFallbackAddress( secureGroup );
            uint8      secsTimeout         = 1;   /* Until success and not reject, we try at 1, 2, 4, 8 secs timeout. */

            while (TRUE)
            {
                boolean escape = TRUE;                                                       /* Only one escape path. */

                /* Before sending PDU, be sure secureGroup did not transition from up to down. */
                if (DOFSecureGroup_NowUpOrNeverUp( secureGroup ))
                {
                    PCR_Assert( secureGroup->reqKEvent.reqKCallbackEvent == NULL );
                    secureGroup->reqKEvent.reqKCallbackEvent = PCREvent_Create( FALSE );
                    if (secureGroup->reqKEvent.reqKCallbackEvent)
                    {
                        RequestKekContext_t * context = DOFMem_Alloc( sizeof( RequestKekContext_t ) );

                        if (context)
                        {
                            /* We are committed: will send TRP Request and wait for response. */
                            reqKEvent_t * pReqKMutexSyncObject;

                            /*Each time through loop, fill all members of new context- no need to clear context struct*/
                            context->kekInfo = pObtainedKEKInfo; /* Caller cleared the structure at pObtainedKEKInfo. */
                            context->filledByCallback = FALSE;
                            context->event = secureGroup->reqKEvent.reqKCallbackEvent;
                            context->pReqKMutexSyncObject = &secureGroup->reqKEvent;
                            /* Call TRP. */
                            if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG, "Sending TRP Request KEK PDU." );
                            TRP_V1_RequestKEK( keyRequest, group, GetKEKCallback, context, fallbackChannel,
                                               fallbackSendAddress, (secsTimeout * 1000) );
                            PCREvent_Wait(context->event, PCRTIMEOUT_FOREVER);/*Block til cback or shutdown signals us*/
                            PCR_Assert( context );
                            if (context->filledByCallback)
                            {
                                /* GetKEKCallback() called us. Be sure secureGroup did not transition from up to down.*/
                                if (DOFSecureGroup_NowUpOrNeverUp( secureGroup ))
                                {
                                    retError = context->callbackStatus;
                                    if (retError == DOFERROR_TIMEOUT)
                                    {
                                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                                            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                                                       "Getting KEK timed out at (secs) 0x",
                                                                       secsTimeout );
                                        PCR_Assert( escape );
                                        if (secsTimeout < 8)
                                        {
                                            /* Try again by falling through to go back up to loop top. */
                                            escape = FALSE;
                                            secsTimeout *= 2;                /* Logarithmic backoff: 1, 2, 4, 8 secs. */
                                        }
                                    }
                                    else
                                    {
                                        /* Done evaluating, have actual response. */
                                        PCR_Assert( escape );
                                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                                        {
                                            if (context->callbackStatus == DOFERROR_OK)
                                            {
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
                                                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                          "TRP Request KEK PDU caused GetKEKCallback() to signal us." );
                                                if (context->trpStatus != DOFERROR_OK)
                                                    DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                                            "                       BUT the callback saw trpStatus 0x:",
                                                                               (uint32)context->trpStatus );
#endif
                                            }
                                            else
                                                DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                       "GetKEKUsingTRP signalled by GetKEKCallback with unexpected callback status 0x:",
                                                                           (uint32)context->callbackStatus );
                                        }
                                        PCR_Assert( context->kekInfo );
                                        if (context->kekInfo->filled)
                                        {
                                            /* Success. The caller of GetKEKUsingTRP() must free not only the new
                                               DOFKEKInfo struct but any malloc'ed members that are referenced. */
                                            PCR_Assert( context->callbackStatus == DOFERROR_OK );
                                            PCR_Assert( context->trpStatus == DOFERROR_OK );
                                            /* Note: context->kekInfo->otherPerms and ->myPerms are as sent by AS but
                                               with iamPermission added. */
                                        }
                                        else if (context->callbackStatus == DOFERROR_OK)
                                        {
                                            PCR_Assert( retError == DOFERROR_OK );
                                            PCR_Assert( context->trpStatus != DOFERROR_OK );
                                            if (TRP_V1_TriggerPossibleCacheReset(creds, TRUE, context->trpStatus))
                                            {
                                                retError = DOFERROR_STALE_CREDENTIAL;
                                            }
                                            else
                                            {
                                                retError = TRP_V1_GetDOFErrorFromErrorCode( context->trpStatus );
                                            }
                                            /* This path means leave group for most possible ->trpStatus, but not all.*/
                                            if (context->trpStatus != TRP_V1_INTERNAL_ERROR &&
                                                context->trpStatus != TRP_V1_SERVICE_UNAVAILABLE)
                                               leaveGroup = TRUE;
                                        }
                                    }
                                }
                                else
                                    retError = DOFERROR_TERMINATED;
                            }
                            else
                            {
                                /* context->event was signaled from DenyGroupBehavior(), called by Process_Shutdown().*/
                                retError = DOFERROR_TERMINATED;
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
                                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                              "GetKEKUsingTRP() is signalled from Process_Shutdown()" );
#endif
                            }
                            /* Cleanup event and related synchronization. */
                            pReqKMutexSyncObject = &secureGroup->reqKEvent;
                            OALMutexPool_Lock(pReqKMutexSyncObject); /*GetKEKCallback/DenyGroupBehavior also lock this*/
                            PCR_Assert( &secureGroup->reqKEvent == context->pReqKMutexSyncObject );
                            /* context->callbackStatus filled by callback or by special code paths above. */
                            if (context->filledByCallback)
                            {
                                /*reqK op concluded by calling our callback, so cleanup. We are safe to do this in all
                                  situations, including when some sort of trailing reqK op cancel happens, like when
                                  DOF_Destroy() cleans up all operations that are still hanging around. That situation
                                  won't call our callback again- the TRP contract is to call our callback EXACTLY once*/
                                DOFMem_Free( context );                               /* escape can be TRUE or FALSE. */
                                context = NULL;
                            }
                            else
                            {
                                /* Callback to be called at DOF_Destroy() op cleanup; specify that it destroy context.*/
                                context->event = NULL;
                                PCR_Assert( escape );
                            }
                            PCREvent_Destroy( secureGroup->reqKEvent.reqKCallbackEvent );
                            secureGroup->reqKEvent.reqKCallbackEvent = NULL;
                            OALMutexPool_Unlock( pReqKMutexSyncObject );
                        }
                        else
                        {
                            PCREvent_Destroy( secureGroup->reqKEvent.reqKCallbackEvent );
                            secureGroup->reqKEvent.reqKCallbackEvent = NULL;
                        }
                    }
                }
                else
                    retError = DOFERROR_TERMINATED;
                if (escape)
                {
                    PCR_Assert( retError != DOFERROR_OK ||  pObtainedKEKInfo->securityScope );
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    {
                        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                         (retError == DOFERROR_TIMEOUT) ? "GetKEKUsingTRP() loop escape by timeout" :
                                                                          ((retError == DOFERROR_TERMINATED) ?
                                                                           "GetKEKUsingTRP() loop escape by terminate" :
                                                                           "GetKEKUsingTRP() loop escape") );
                    }
#endif
                    break;
                }
            } /* end while() */
            DOFKeyRequest_Destroy( keyRequest );
        }
    }
    PCR_Assert( *pMostDescriptiveDOFException == NULL );
    if (retError != DOFERROR_OK)  /* Technically this test not needed, but here for form at least; also saves cycles. */
    {
        *pMostDescriptiveDOFException = DOFErrorException_GetFromError( retError );
        PCR_Assert( *pMostDescriptiveDOFException != NULL );
        if (leaveGroup)
            retError = DOFERROR_ACCESS_DENIED; /* Caller will leave group but pass on the DOFEXCEPTION that cause it. */
    }

    return retError;
}


static void ExtendPermisionsFIFO_Cleanup( oapSendInfo_t * pOapSendInfo )
{
    DOFOperation_Internal_Destroy( pOapSendInfo->pendingOp );
    DOFResult_Destroy( pOapSendInfo->pendingResult );
    DOFMem_Free( pOapSendInfo );
}


/** @internal
    The callback provided to receive the KEK from the call to TRP_RequestKEK

    @note Function prototype is trp_callback_fn, representing:
                                    void (*trp_callback_fn)( TRP_V1_Context trp_context, extendPerms_t * pExtendPerms );
    @param  trp_context The return information from TRP.
    @param  ctx         The location to save the information provided by TRP.
**/
static void ExtendPermissionsCallback( TRP_V1_Context trp_context, void * ctx )
{
    extendPerms_t * pExtendPerms;

    PCR_Assert( ctx );
    pExtendPerms = ctx;
    PCR_Assert( pExtendPerms->secureGroup );
    PCR_Assert( pExtendPerms->pendingOp );
    if (trp_context)
    {
        boolean newPermissionInfo = FALSE;

        /* Look at TRP response and act only if we have valid KEK info. */
        if (trp_context->KEKReturn->status != DOFERROR_OK)
        {
            if (trp_context->KEKReturn->status == DOFERROR_ACCESS_DENIED)
            {
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                   "ExtendPermissionCallback() finds DENY KEKRequest, leaving group." );
                DOFSecureGroup_LeaveGroup( pExtendPerms->secureGroup, DOFEXCEPTION_ACCESS_DENIED );
            }
            else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                          "ExtendPermissionCallback() finds KEK Request timed out or internal error." );
        }
        else if (trp_context->KEKReturn && trp_context->KEKReturn->errorCode == 0)
        {
            OALCipher cipher;
            OALCipherKeyState keyState;

            PCR_Assert( trp_context->KEKReturn->tmax && trp_context->KEKReturn->tmin && trp_context->KEKReturn->thb );
            cipher = (DOFSecureGroup_GetChannel_Internal( pExtendPerms->secureGroup ))->remoteChannelData->cipher;
            keyState = OALCipher_GetCurrentKeyState( cipher );
            /* Add AS'es currently authorized (Pg + matching scopes) to our "theirPerms". */
            DOFPermissionSet_AddPermissionsFromSet( OALCipher_GetRemotePerms( cipher, keyState ),
                                                    trp_context->KEKReturn->groupsPermissions );
            /* Add AS'es currently authorized (Initiator security scope + matching scopes) to our "myPerms". */
            DOFPermissionSet_AddPermissionsFromSet( OALCipher_GetLocalPerms( cipher, keyState ),
                                                    trp_context->KEKReturn->initPermissions );
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                    "ExtendPermissionCallback() received new permissions." );
            /* Our positive response means pendingOp is now permissioned, so resubmit it. */
            if (pExtendPerms->requestNotResponse)
            {
                PCR_Assert( !pExtendPerms->pendingResult );
                DOFSecureGroup_RouteToProvider( DOFSecureGroup_GetChannel_Internal( pExtendPerms->secureGroup ),
                                                pExtendPerms->pendingOp, pExtendPerms->provOp, pExtendPerms->address );
            }
            else
            {
                PCR_Assert( pExtendPerms->pendingResult );
                OALRemoteChannel_RouteToRequestorLossy( DOFSecureGroup_GetChannel_Internal( pExtendPerms->secureGroup ),
                                                        pExtendPerms->pendingOp, pExtendPerms->pendingResult );
            }
            newPermissionInfo = TRUE;
        }
        else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        {
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                     "ExtendPermissionCallback() gets KEKRequestResponse, but with error indicators." );
        }
        /* Last, resend queued up OAP packets if newPermissionInfo, or at the very least cleanup the queue. */
        {
            PCRMutex mutex = OALMutexPool_Get( pExtendPerms->secureGroup->extendPermissionsFIFO );

            PCRMutex_Lock( mutex );
            if (pExtendPerms->secureGroup->flags.extendPermissionsASRequestInProgress)
            {
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    if (OALQueue_GetSize( pExtendPerms->secureGroup->extendPermissionsFIFO ) != 0)
                        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                            "ExtendPermissionsCallback() dequeues outgoing PDUs." );
                while (OALQueue_GetSize( pExtendPerms->secureGroup->extendPermissionsFIFO ) != 0)
                {
                    oapSendInfo_t * pOapSendInfo = OALQueue_Dequeue( pExtendPerms->secureGroup->extendPermissionsFIFO );

                    if (newPermissionInfo)
                    {
                        if (pOapSendInfo->pendingResult)
                            OALRemoteChannel_RouteToRequestorLossy(
                                                        DOFSecureGroup_GetChannel_Internal( pExtendPerms->secureGroup ),
                                                                   pOapSendInfo->pendingOp,
                                                                   pOapSendInfo->pendingResult );
                        else
                            DOFSecureGroup_RouteToProvider(
                                                        DOFSecureGroup_GetChannel_Internal( pExtendPerms->secureGroup ),
                                                            pOapSendInfo->pendingOp, pExtendPerms->provOp,
                                                            pExtendPerms->address );
                    }
                    /* ExtendPermissionsFIFO_Cleanup will free pOapSendInfo after releasing its members. */
                    ExtendPermisionsFIFO_Cleanup( pOapSendInfo );
                }
                pExtendPerms->secureGroup->flags.extendPermissionsASRequestInProgress = 0;
            }
            PCRMutex_Unlock( mutex );
            OALMutexPool_Release( pExtendPerms->secureGroup->extendPermissionsFIFO );
        }
    }
    else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
    {
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                            "ExtendPermissionsCallback() cannot act, has no AS response information" );
    }
    DOFOperation_Internal_Destroy( pExtendPerms->pendingOp );
    DOFAddress_Destroy( pExtendPerms->address );
    DOFResult_Destroy( pExtendPerms->pendingResult );
    DOFSecureGroup_Release( pExtendPerms->secureGroup );
    DOFMem_Free( pExtendPerms );
}


/** @internal
    The callback provided to receive the KEK from the call to TRP_RequestKEK

    @note Function prototype is trp_callback_fn, representing:
                          void (*trp_callback_fn)( TRP_V1_Context trp_context, extendTheirPerms_t * pExtendTheirPerms );
    @param  trp_context The return information from TRP.
    @param  ctx         The location to save the information provided by TRP.
**/
static void ExtendTheirPermissionsCallback( TRP_V1_Context trp_context, void * ctx )
{
    extendTheirPerms_t * pExtendTheirPerms;
    DOFError             status = DOFERROR_INTERNAL_ERROR;

    PCR_Assert( ctx );
    pExtendTheirPerms = ctx;
    PCR_Assert( pExtendTheirPerms->secureGroup );
    PCR_Assert( pExtendTheirPerms->extendCompleteCallback );
    if (trp_context && trp_context->KEKReturn)
    {
        /* Look at the response from TRP and fill in the DOFKEKInfo struct only if we get back a valid KEK. */
        if (trp_context->KEKReturn->status == DOFERROR_OK && trp_context->KEKReturn->errorCode == 0)
        {
            uint8    opKEK = DOFSecureGroup_GetOpKEK( pExtendTheirPerms->secureGroup );
            OALCipher cipher;
            OALCipherKeyState keyState;

            cipher = (DOFSecureGroup_GetChannel_Internal( pExtendTheirPerms->secureGroup ))->remoteChannelData->cipher;
            keyState = OALCipher_GetCurrentKeyState( cipher );
            if (DOFPermissionSet_AddPermissionsFromSet( pExtendTheirPerms->secureGroup->keks[opKEK].otherPerms,
                                                         trp_context->KEKReturn->groupsPermissions ) &&
                DOFPermissionSet_AddPermissionsFromSet( pExtendTheirPerms->secureGroup->keks[opKEK].myPerms,
                                                         trp_context->KEKReturn->initPermissions ) &&
                /* Inform channel of upgraded perms- they did NOT diminish, only increased. We must do this because the
                   final evaluation is against .remoteChannelData->theirPerms. */
                DOFPermissionSet_AddPermissionsFromSet( OALCipher_GetRemotePerms( cipher, keyState ),
                                                        trp_context->KEKReturn->groupsPermissions ) &&
                DOFPermissionSet_AddPermissionsFromSet( OALCipher_GetLocalPerms( cipher, keyState ),
                                                        trp_context->KEKReturn->initPermissions ))
            {
                status = DOFERROR_OK;
            }
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            {
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                    "ExtendTheirPermissionCallback gets good TRP status." );
#endif
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                    "ExtendTheirPermissionCallback received new permissions." );
            }
        }
        else
        {
            if (trp_context->KEKReturn->status == DOFERROR_ACCESS_DENIED)
            {
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                              "ExtendTheirPermissionCallback() finds DENY KEKRequest, leaving group." );
                DOFSecureGroup_LeaveGroup( pExtendTheirPerms->secureGroup, DOFEXCEPTION_ACCESS_DENIED );
            }
            else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                     "ExtendTheirPermissionCallback() finds KEK Request timed out or internal error." );
            status = DOFERROR_FAILED;
        }
    }
    else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
    {
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,       /* Why no or bad trp_context? */
                                            "ExtendTheirPermissionCallback gets empty TRP status." );
    }
    pExtendTheirPerms->extendCompleteCallback( status, pExtendTheirPerms->context );
    /* Expected ->extendCompleteCallback is _extendSecureGroupPermsCallback(). */
    DOFSecureGroup_Release( pExtendTheirPerms->secureGroup );
    DOFMem_Free( pExtendTheirPerms );
}


/*
**  "Internal Public"  Methods.
*/


boolean DOFSecureGroup_SetGrantTimeout( DOFSecureGroup secureGroup, uint32 grantMillis )
{
    PCRMutex mutex;
    boolean  retBoolean = FALSE;

    PCR_Assert( secureGroup );
    PCR_Assert( grantMillis );
    mutex = OALMutexPool_Get( secureGroup );
    PCRMutex_Lock( mutex );     /* Guard against CleanupSecureGroupTimer(), which can steal timer items away from us. */
    if (DOFSecureGroup_NowUpOrNeverUp( secureGroup ) && !secureGroup->channel.remoteChannelData->flags.isInShutdownState)
    {
        PCR_Assert( secureGroup->grantTimerItem );          /* Asserts valid because sg is up under mutex protection. */
        PCR_Assert( secureGroup->getKEKTimerItem );
        if (OALTimerItem_Insert( secureGroup->grantTimerItem, grantMillis, FALSE, TRUE ) == DOFERROR_OK)
        {
            uint32 millisOneMoreCheck;

            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                           "_SetGrantTimeout() extends AS grant to (msecs) 0x", (uint32)grantMillis );
            millisOneMoreCheck = MillisDelayForOneMoreEpochCheck( secureGroup );
            if (millisOneMoreCheck)
            {
                uint32 nextCheckKEKTimeDelay = OALTimerItem_TimeRemaining( secureGroup->getKEKTimerItem );

                if (nextCheckKEKTimeDelay > millisOneMoreCheck)
                {
                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                        DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                "To avoid exhausting grant, _SetGrantTimeout() moved next epoch check in by (msecs) 0x",
                                                   (uint32)(nextCheckKEKTimeDelay - millisOneMoreCheck) );
                    DOFSecureGroup_SetEpochTimeout( secureGroup, millisOneMoreCheck, FALSE );
                                                       /* FALSE: no duplicate computation. */
                }
            }
            else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                    "_SetGrantTimeout() finds that secureGroup grant is exhausted." );
            retBoolean = TRUE;
        }
        else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "_SetGrantTimeout() fails timer item insert" );
    }
    else
        DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_TRACE, "_SetGrantTimeout() finds secureGroup went down.");
    PCRMutex_Unlock( mutex );
    OALMutexPool_Release( secureGroup );

    return retBoolean;
}


/** @internal
    Cleanup the given KEKInfo structure of the given secureGroup, and disable that KEKInfo.

    @param pKEKInfo Pointer to the given KEKInfo structure.
**/
void DOFSecureGroup_CleanupKEKInfoAndDisable( DOFKEKInfo_t * pKEKInfo )
{
    CleanupKEKInfo( pKEKInfo );
    pKEKInfo->filled = FALSE;
}


/*  @internal

    @return TRUE/FALSE if we do/not have a KEK.
*/
boolean DOFSecureGroup_GetGreatestKnownEpoch( DOFSecureGroup secureGroup, uint16 * pGreatestKnownEpoch )
{
    boolean retBoolean = TRUE;
    uint8   futKEK     = secureGroup->flags.fut;
    uint8   opKEK      = futKEK ^ 1;

    if (secureGroup->keks[futKEK].filled)
    {
        *pGreatestKnownEpoch = secureGroup->keks[futKEK].epoch;
        PCR_Assert( secureGroup->keks[opKEK].filled );
        PCR_Assert( DOFTime_Compare_16Bit( secureGroup->keks[futKEK].epoch, secureGroup->keks[opKEK].epoch ) > 0 );
    }
    else if (secureGroup->keks[opKEK].filled)
        *pGreatestKnownEpoch = secureGroup->keks[opKEK].epoch;
    else
        retBoolean = FALSE;

    return retBoolean;
}


/* @internal

   @note This blocks until AS responds, up to 15 seconds.

   @param secureGroup
   @param pFillKEKInfo
   @return DOFERROR_OK if KEK was obtained, or
           DOFERROR_ACCESS_DENIED if AS specifically denied our new KEK (we leave the group in this case), or (anything
                                  else) if KEK was not obtained.
*/
DOFError DOFSecureGroup_GetKEK( DOFSecureGroup secureGroup, DOFKEKInfo_t * pFillKEKInfo )
{
    DOFCredentials secureCreds;
    DOFError       status;
    DOFException   mostDescriptiveDOFException;                  /* Will be initialized within GetKEKUsingTRP() call. */

    PCR_Assert( secureGroup );
    secureCreds = ((OALChannel)DOFSecureGroup_GetChannel_Internal(secureGroup))->remoteChannelData->secureCredentials;
    PCR_Assert( secureCreds );
    PCR_Assert( pFillKEKInfo );
    PCRMem_Set( pFillKEKInfo, 0, sizeof( DOFKEKInfo_t ) );

    /* At first join, secureGroup is not up, so cannot test secureGroup->up here and then use answer to fail out. */

    PCR_Assert( (DOFSecureGroup_GetChannel_Internal( secureGroup ))->remoteChannelData->flags.isConnection );
    /* REVIEW: This is called only at Tmax, and that we submit our base myPerms to AS. Our base myPerms are held in
               connConfig, filled originally by the application. Is it called only that often?
               Can optimize: For SGM we must call this every Tmin, can choose to submit accumuated myPerms until Tmax.*/
    status = GetKEKUsingTRP( secureGroup, secureGroup->group,               /* This call blocks until we have answer. */
                             DOFConnectionConfig_GetPermissions(
                            (DOFSecureGroup_GetChannel_Internal( secureGroup ))->remoteChannelData->config.connection ),
                             secureCreds, pFillKEKInfo, &mostDescriptiveDOFException );
    if (status == DOFERROR_OK)
    {
        PCR_Assert( pFillKEKInfo->filled );
        PCR_Assert( pFillKEKInfo->Tmax && pFillKEKInfo->Tmin && pFillKEKInfo->Thb );
        PCR_Assert( pFillKEKInfo->securityScope );
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        {
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "Valid KEK has Tmin of 0x: ", pFillKEKInfo->Tmin );
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "Valid KEK has Tmax of 0x: ", pFillKEKInfo->Tmax );
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "Valid KEK has Thb of 0x: ", pFillKEKInfo->Thb );
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "Valid KEK has epoch of 0x: ", pFillKEKInfo->epoch );
            /*DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "Valid KEK, 4 bytes: ",
                                                            *((unsigned int *)pFillKEKInfo->KEK ));*/
        }
        if (secureGroup->grantTimerItem)
        {
            /* Extend grant from AS to Tmax. */
            if (DOFSecureGroup_SetGrantTimeout( secureGroup,
                                          (pFillKEKInfo->Tmax*SGMP_TMAX_MSEC_MULTIPLIER/SCALE_TMAX_TMIN) ) == FALSE)
            {
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG,
                                                        "DOFSecureGroup_GetKEK() failed to extend grant." );
            }
        }
        else
        { /* InitiateGroupBehavior() has not been called yet to initialize timer and fill ->grantTimerItem. */
        }
        /* pFillKEKInfo->securityScope will eventually be used when this KEK promotes.
           If instead we make a 2nd trip through InitialJoin(), it will be cleaned up there. */
    }
    else
    {
        PCR_Assert( mostDescriptiveDOFException != NULL );
        if (status == DOFERROR_ACCESS_DENIED)
        {
            /* mostDescriptiveDOFException holds precise reason why AS answer said to leave the group. */
            DOFSecureGroup_LeaveGroup( secureGroup, mostDescriptiveDOFException );
        }
        else
        { /* Remain in group UNTIL Tmax expires without getting a new KEK- enforced by timerItem grantTimerItem. */
        }
    }

    return status;
}


/* @internal

   @note This blocks up to 15 seconds until AS responds.
*/
boolean DOFSecureGroup_GetKEKAndUpdateFutKEKInfo( DOFSecureGroup secureGroup )
{
    /* The common function for getting the KEK. secureGroup->flags.finishedGettingKEK is set only in this function. */
    PCRMutex     mutex;
    DOFKEKInfo_t fillKEKInfo;
    DOFError     status;
    uint32       nextCheckKEKTimeDelay;
    uint8        futKEK;
    uint8        kekOffset;
    boolean      futureKEKUpdated = FALSE;

    PCR_Assert( secureGroup );
    mutex = OALMutexPool_Get( secureGroup );
    PCRMutex_Lock( mutex );
    if (!secureGroup->up)
    {
        secureGroup->flags.finishedGettingKEK = 1;
        goto cleanupMutexExit;
    }
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
    if (!secureGroup->channel.remoteChannelData->flags.isInShutdownState)
    {
        if (DOFLog_IsLogLevel(DOFLOGLEVEL_TRACE) && secureGroup->grantTimerItem)
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "GetKEKAndUpdateFutKEKInfo() ENTRY finds remaining Grant msec 0x",
                                       OALTimerItem_TimeRemaining( secureGroup->grantTimerItem) );
    }
#endif
    PCRMutex_Unlock( mutex );
    /* AdjustKeys() in case traffic has been minimal. This promotes an existing futKEK before we get new KEK. */
    OALCipher_AdjustKeys( ((OALChannel)secureGroup)->remoteChannelData->cipher );
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "GetKEKAndUpdateFutKEKInfo() now calls GetKEK()." );
    /* GetKEK will block while a request is sent until it receives a response or it times out. */
    status = DOFSecureGroup_GetKEK( secureGroup, &fillKEKInfo );                           /* Does not mutex protect. */
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "GetKEKAndUpdateFutKEKInfo() returned from GetKEK()." );
    PCRMutex_Lock( mutex );
    if (secureGroup->up)              /* Repeated check here appropriate- GetKEK() required some portion of a second. */
    {
        futKEK = secureGroup->flags.fut;
        kekOffset = secureGroup->keks[futKEK].filled ? futKEK : DOFSecureGroup_GetOpKEK( secureGroup );
        if (status == DOFERROR_OK)
        {
            PCR_Assert( secureGroup->keks[kekOffset].Tmin );/*asserts proper because TRP rejects KEK where these are 0*/
            PCR_Assert( secureGroup->keks[kekOffset].Tmax );
            /* The _AdjustKeys() call is the mechanism that promotes a futKEK, and locks the cipher to do it. This next
               call also locks the cipher to prevent other threads from calling _AdjustKeys() while we are updating. */
            futureKEKUpdated = UpdateFutureKEKInfo(secureGroup, &fillKEKInfo); /*Fill futKEK or not. Clean fillKEKInfo*/
        }
        secureGroup->flags.finishedGettingKEK = 1;                     /* Must be before _SetEpochTimeout(), below. */
        /* We could assert status not DOFERROR_ACCESS_DENIED, if certain that secureGroup->up was taken down by that*/
        if (status != DOFERROR_ACCESS_DENIED)
        {
            /* Prepare to call _SetEpochTimeout(). Note: _SetGrantTimeout() was called above in GetKEK(), if there was a
               good answer from the AS. */
            PCR_Assert( secureGroup->keks[kekOffset].Tmax );
            PCR_Assert( secureGroup->keks[kekOffset].Tmin );
            nextCheckKEKTimeDelay = (secureGroup->flags.activeSGM ?
                                         (secureGroup->keks[kekOffset].Tmin*SGMP_TMIN_MSEC_MULTIPLIER) :
                                         (secureGroup->keks[kekOffset].Tmax*SGMP_TMAX_MSEC_MULTIPLIER))/SCALE_TMAX_TMIN;
            PCR_Assert( nextCheckKEKTimeDelay );
            DOFSecureGroup_SetEpochTimeout( secureGroup, nextCheckKEKTimeDelay, TRUE );
        }
        else
        { /* GetKEK() already took action to leave the group. */
        }
    }
    else
    {
        /* If KEK Grant expires, it can call _LeaveGroup(), which takes down secureGroup->up. We avoided useless work.*/
        secureGroup->flags.finishedGettingKEK = 1;
    }
cleanupMutexExit:
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
    if (secureGroup->channel.remoteChannelData->flags.isInShutdownState)
    {
        if (DOFLog_IsLogLevel(DOFLOGLEVEL_TRACE) && secureGroup->grantTimerItem)
                DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "GetKEKAndUpdateFutKEKInfo() EXIT finds remaining Grant msec 0x",
                                                              OALTimerItem_TimeRemaining( secureGroup->grantTimerItem) );
    }
#endif
    PCRMutex_Unlock( mutex );
    OALMutexPool_Release( secureGroup );


    return futureKEKUpdated;                        /* TRUE for new KEK, FALSE for same old KEK or AS did not answer. */
}


/** @internal
 * used for extending permissions for an op associated with channel
 *
 * @param chan
 * @param address
 * @param permSet
 * @param pendingOp Operation that triggered the extending of permissions that will be handled when trp returns
 * @param provOp
 * @param pendingResult
 * @param requestNotResponse
 * @param unused Supplied to match function prototype of OALChannel_ExtendPermissions
 */
void DOFSecureGroup_ExtendPermissions( OALChannel chan, DOFAddress address, DOFPermissionSet permSet,
                                       DOFOperation pendingOp, DOFOperation provOp, DOFResult pendingResult,
                                       boolean requestNotResponse, boolean unused )
{
    DOFCredentials secureCreds;
    boolean        success = FALSE;
    DOFSecureGroup sg      = (DOFSecureGroup)chan;

    PCR_Assert( sg );
    secureCreds = ((OALChannel)DOFSecureGroup_GetChannel_Internal( sg ))->remoteChannelData->secureCredentials;
    PCR_Assert( secureCreds );
    if (permSet && pendingOp && sg->group && DOFCredentials_IsResolved( secureCreds ))
    {
        /* pendingResult: pendingOp is an OAP response. NULL pendingResult: pendingOp is an OAP request. */
        PCRMutex mutex = OALMutexPool_Get( sg->extendPermissionsFIFO );

        PCRMutex_Lock( mutex );
        if (sg->flags.extendPermissionsASRequestInProgress)
        {
            oapSendInfo_t * pOapSendInfo = DOFMem_Alloc( sizeof( oapSendInfo_t ) );

            if (pOapSendInfo != NULL)
            {
                PCRMem_Set( pOapSendInfo, 0, sizeof( oapSendInfo_t ) );
                if (!pendingResult ||
                    (pOapSendInfo->pendingResult = DOFResult_Copy( pendingResult )) != NULL)
                {
                    DOFOperation_Retain( pendingOp );
                    pOapSendInfo->pendingOp = pendingOp;
                    if (OALQueue_Enqueue( sg->extendPermissionsFIFO, pOapSendInfo ) != FALSE)
                    {
                        success = TRUE;
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                             "ExtendPermissions() enqueues outgoing OAP PDU, pending answer from AS." );
                    }
                    else
                    {
                        /* Cleanup then fall through to give fail log message. */
                        DOFOperation_Internal_Destroy( pendingOp );
                        DOFResult_Destroy( pendingResult );
                    }
                }
                if (!success)
                    DOFMem_Free( pOapSendInfo );
            }
        }
        else
        {
            /* OAP calls us if he has otherwise given up. We need to:
                   a) Start up our TRP Request and return without too much time (at our return the OAP caller then
                      returns to the protocol stack caller an error code that says: drop this send PDU.
                   b) When we get our KEK Request response, either clean it up or resubmit it to the protocol stack. */
            uint32        timeout    = (pendingOp->sources->oal_timeout < 15000) ? pendingOp->sources->oal_timeout :
                                                                                   15000;
            DOFKeyRequest keyRequest = DOFKeyRequest_Create(secureCreds, REQUEST_KEK_NONCE_SIZE, FALSE, FALSE, NULL, permSet);

            if (keyRequest)
            {
                extendPerms_t * pExtendPerms = DOFMem_Alloc( sizeof( extendPerms_t ) );

                if (pExtendPerms)
                {
                    PCRMem_Set( pExtendPerms, 0, sizeof( extendPerms_t ) );
                    if (!pendingResult ||
                        (pExtendPerms->pendingResult = DOFResult_Copy( pendingResult )) != NULL)
                    {
                        if (!address ||
                            (pExtendPerms->address = DOFAddress_Copy( address )) != NULL)
                        {
                            if ((pExtendPerms->secureGroup = DOFSecureGroup_Retain( sg )) != NULL)
                            {
                                OALChannel fallbackChannel     = DetermineFallbackChannel( sg );
                                DOFAddress fallbackSendAddress = DetermineFallbackAddress( sg );

                                pExtendPerms->pendingOp = pendingOp;
                                DOFOperation_Retain( pendingOp );
                                pExtendPerms->provOp = provOp;
                                pExtendPerms->requestNotResponse = requestNotResponse;
                                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                                        "ExtendPermissions() sends TRP RequestKEK" );
                                TRP_V1_RequestKEK( keyRequest, sg->group, ExtendPermissionsCallback, pExtendPerms,
                                                   fallbackChannel, fallbackSendAddress, timeout );
                                /* pExtendPerms ownership transfers to ExtendPermissionsCallback(). */
                                sg->flags.extendPermissionsASRequestInProgress = TRUE;
                                success = TRUE;
                            }
                            else
                                DOFAddress_Destroy( address );
                        }
                        else
                            DOFResult_Destroy( pendingResult );
                    }
                    if (!success)
                        DOFMem_Free( pExtendPerms );
                }
                DOFKeyRequest_Destroy( keyRequest );
            }
        }
        PCRMutex_Unlock( mutex );
        OALMutexPool_Release( sg->extendPermissionsFIFO );
    }
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ) && !success)
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG, "ExtendPermissions() ERROR, OAP PDU dropped" );
}


/** @internal
 * used for extending permissions for an op associated with channel
 *
 * @param sgChannel   SecureGroup/Channel called on
 * @param address
 * @param extendPerms
 * @param extendCompleteCallback
 * @param context
 */
void DOFSecureGroup_ExtendTheirPermissions( OALChannel sgChannel, DOFAddress address, DOFPermissionSet extendPerms,
                                            ExtendCompleteCB extendCompleteCallback, void * context )
{
    DOFCredentials secureCreds;
    DOFError       status      = DOFERROR_ILLEGAL_ARGUMENT;
    DOFSecureGroup secureGroup = (DOFSecureGroup)sgChannel;

    PCR_Assert( extendCompleteCallback );
    PCR_Assert( secureGroup );
    secureCreds = ((OALChannel)DOFSecureGroup_GetChannel_Internal( secureGroup ))->remoteChannelData->secureCredentials;
    PCR_Assert( secureCreds );
    if (extendPerms && secureGroup->group && DOFCredentials_IsResolved( secureCreds ))
    {
        /* OAP calls us if he has otherwise given up. We need to:
               a) Start up our TRP Request and return without too much time (at our return the OAP caller then
                  returns to the protocol stack caller an error code that says: drop this incoming PDU.
               b) When we get our KEK Request response, check for success, then clean it up. */
        DOFKeyRequest keyRequest;

        status = DOFERROR_INSUFFICIENT_RESOURCES;
        if ((keyRequest = DOFKeyRequest_Create(secureCreds, REQUEST_KEK_NONCE_SIZE, FALSE, FALSE, NULL, extendPerms)) != NULL)
        {
            boolean              ok                = FALSE;
            extendTheirPerms_t * pExtendTheirPerms = DOFMem_Alloc( sizeof( extendTheirPerms_t ) );

            if (pExtendTheirPerms != NULL)
            {
                /* No memset at pExtendTheirPermissions, all 3 members are filled. */
                if ((pExtendTheirPerms->secureGroup = DOFSecureGroup_Retain( secureGroup )) != NULL)
                {
                    /* DO KEK request with extendPerms in keyRequest, representing the myPerms field of the request. */
                    OALChannel fallbackChannel     = DetermineFallbackChannel( secureGroup );
                    DOFAddress fallbackSendAddress = DetermineFallbackAddress( secureGroup );          /* multicast. */

                    pExtendTheirPerms->context = context;
                    pExtendTheirPerms->extendCompleteCallback = extendCompleteCallback;
                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                            "ExtendTheirPermissions() sends TRP RequestKEK" );
                    /* This next does not block, and pExtendTheirPerms is freed at end of the callback. */
                    TRP_V1_RequestKEK( keyRequest, secureGroup->group, ExtendTheirPermissionsCallback,
                                       pExtendTheirPerms, fallbackChannel, fallbackSendAddress, 5000 );/*5 sec timeout*/
                    ok = TRUE;
                }
                else
                    DOFMem_Free( pExtendTheirPerms );
            }
            DOFKeyRequest_Destroy( keyRequest );
            if (ok)
                return;
        }
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG,
                                     "ExtendTheirPermissions() memory allocate failure, does NOT send TRP RequestKEK" );
    }
    else if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG,
                                            "ExtendTheirPermissions() param failure, does NOT send TRP RequestKEK" );
    extendCompleteCallback( status, context );
}


/**
@}
**/
