/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>


#include <dof/credentials.h>
#include <dof/dof.h>
#include <dof/item.h>
#include <dof/objectid.h>
#include <dof/pcr.h>
#include <dof/permissionset.h>
#include <dof/request.h>
#include <dof/securitymode.h>
#include <dof/server.h>
#include <dof/system.h>
#include <dof/value.h>
#include <internal/channel_stack_messages.h>
#include <internal/connectionless_remotechannel.h>
#include <internal/epp_negative_one.h>
#include <internal/exception.h>
#include <internal/mutex_pool.h>
#include <internal/operationinternal.h>
#include <internal/remotechannel.h>
#include <internal/remotechannelmanager.h>
#include <internal/timeinternal.h>
#include "securegroupprivate.h"

extern PCRMutex connection_server_mutex;


/** @addtogroup DOFSecureGroup
@{
**/


/*
**
**  "Private" Type and Method Declarations
*/


typedef void (*secureGroup_MessageHandler_fn_t)( OALChannel, OALMessage );


static boolean IsCCM_Op( DOFOperationType opType )
{
    return opType >= DOFOPERATIONTYPE_CCMMODE_V1_FIRST && opType < DOFOPERATIONTYPE_CCMMODE_V1_LAST;
}

static void CheckBlockedMessageQueueTimeout( OALChannel secureGroupChannel )
{
    if (OALQueue_GetSize( secureGroupChannel->remoteChannelData->blockedMessageQueue ) == 1)
    {
        /* Set queue timeout. */
        secureGroupChannel->remoteChannelData->blockedMessageQueueTimeoutExpirationTime =
                                                                        PCRTime_GetMS() + BLOCKED_MESSAGE_QUEUE_TIMEOUT;
        secureGroupChannel->remoteChannelData->blockedMessageQueueTimeoutEnabled = TRUE;
    }
    else if (secureGroupChannel->remoteChannelData->blockedMessageQueueTimeoutEnabled &&
             DOFTime_Compare( PCRTime_GetMS(),
                              secureGroupChannel->remoteChannelData->blockedMessageQueueTimeoutExpirationTime ) > 0)
    {
        DOFSecureGroup_LeaveGroup( (DOFSecureGroup)secureGroupChannel, DOFEXCEPTION_FAILED );
    }
}


static void HandleMessage_RouteToProviderOrRequestor( OALChannel secureGroupChannel, OALMessage message,
                                                      boolean toProvider )
{
    OALMessage   msgToStack;
    boolean      drop = TRUE;
    DOFAddress   addr = (DOFAddress)message->argv[2];

    if (!message->shutdown)
    {
        if (toProvider)
            msgToStack = SMT_RouteToProvider_Convert( message, secureGroupChannel->remoteChannelData->protocolStack,
                                                      secureGroupChannel, addr );
        else
            msgToStack = SMT_RouteToRequestor_Convert( message, secureGroupChannel->remoteChannelData->protocolStack,
                                                       addr, secureGroupChannel );
        if (msgToStack)
        {
            DOFOperation reqOrRespOp = message->argv[0];
            /* reqOrRespOp is a request for toProvider TRUE, otherwise it is a response. */

            /*  If SecureSendBlocked is true or there is a non empty queue then just place the stack message created in
                the blockedMessageQueue. One case may be if this is a secure group and the nodeID is invalid. */
            if (!reqOrRespOp->flags.bits.securityRequiredSecureHeaderUnsecure &&
                secureGroupChannel->remoteChannelData->cipher &&
                !IsCCM_Op( reqOrRespOp->operation_type ) &&
                (secureGroupChannel->remoteChannelData->flags.SecureSendBlocked ||      /* Set by Encipher not ready. */
                 (secureGroupChannel->remoteChannelData->blockedMessageQueue &&
                  (OALQueue_GetSize( secureGroupChannel->remoteChannelData->blockedMessageQueue ) != 0)) ||
                 !OALCipher_SecureGroupEncipherReady( secureGroupChannel->remoteChannelData->cipher, addr,
                                                 EPP_NEGONE_NO_SECURE_MARKING_ID, /*Don't know where to get real value*/
                                                 reqOrRespOp->flags.bits.denyPacketReplayProtectionIfSecureUnicastSend )
                )
               )
            {
                if (!secureGroupChannel->remoteChannelData->blockedMessageQueue)
                    secureGroupChannel->remoteChannelData->blockedMessageQueue = OALQueue_Create();
                if (secureGroupChannel->remoteChannelData->blockedMessageQueue)
                {
                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                           "RouteToProviderOrRequestor message handler adds to blocked message queue" );
                    OALQueue_Enqueue( secureGroupChannel->remoteChannelData->blockedMessageQueue, msgToStack );
                    CheckBlockedMessageQueueTimeout( secureGroupChannel );
                    drop = FALSE;
                }
            }
            else
            {
                /* Immediate success. */
                /* We can be here with ->blockedMessageQueue existing and empty but also knowing that cipher is ready.
                   For that situation it may be concerning that by the time our PDU gets to the cipher encipher call
                   that cipher won't be ready. Here we rely on message handling to know that our PDU is fully processed
                   through the protocol stack before any other PDU is processed. That means that if cipher is ready now
                   it will also be ready as our PDU is fed through it. */
                if (secureGroupChannel->remoteChannelData->maxSendSilenceTimerItem)
                    OALRemoteChannelMgr_SetSendSilenceTimerItem( secureGroupChannel );
                OALRemoteChannelMgr_SendMessageToStack( msgToStack );
                drop = FALSE;
            }
        }
    }
    if (drop)
    {
        OALMessage_Destroy( message );
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG,
                                                "HandleMessage_RouteToProviderOrRequestor() drops message." );
    }
}


static void HandleMessage_RouteToProvider( OALChannel secureGroupChannel, OALMessage message )
{
    HandleMessage_RouteToProviderOrRequestor( secureGroupChannel, message, TRUE );
}


static void HandleMessage_RouteToRequestor( OALChannel secureGroupChannel, OALMessage message )
{
    HandleMessage_RouteToProviderOrRequestor( secureGroupChannel, message, FALSE );
}


static void HandleMessage_ReceiveFromTransport( OALChannel secureGroupChannel, OALMessage message )
{
    if (message->shutdown)
        OALMessage_Destroy(message);
    else
    {
        /* This called for point member only, when udp packet comes in from hub. This next handles incoming SGMP PDUs.*/
        OALConnectionlessRemoteChannel_HandleMessage_ReceiveFromTransport(
                                                          (OALConnectionlessRemoteChannel)secureGroupChannel, message );
    }
}

static void Initialize( OALChannel secureGroupChannel )
{
    DOFSecureGroup         secureGroup = (DOFSecureGroup)secureGroupChannel;
    SecureGroupContext_t * node;

    PCR_Assert( secureGroup );
    PCR_Assert( secureGroup->workEnqueue );
    /* _AddChannel() does not retain, but instead we assume ownership of the CTOR refCount. */
    PCR_Assert( secureGroup->flags.pointMember || !secureGroupChannel->remoteChannelData->flags.isUp );
    secureGroup->flags.initializeCalled = TRUE;
    if (secureGroupChannel->remoteChannelData->flags.setDtlInChannelInit)
    {
        secureGroupChannel->remoteChannelData->dtl =
                                             secureGroupChannel->remoteChannelData->baseChannel->remoteChannelData->dtl;
        if (!DTL_Retain( secureGroupChannel->remoteChannelData->dtl ))
            OALRemoteChannel_Shutdown( secureGroupChannel, DOFEXCEPTION_FAILED );
    }
    node = (SecureGroupContext_t *)DOFMem_Alloc( sizeof( SecureGroupContext_t ) );
    if (node != NULL)
    {
        OALEnqueueItem_Init( (OALEnqueueItem)node, DOFSecureGroup_InitialJoin );
        node->secureGroup = secureGroup;
        if (OALEnqueue_Enqueue( secureGroup->workEnqueue, (OALEnqueueItem)node ))
        {
            /* Get KEK, join group, _NotifyInitComplete(DOFERROR_OK), and set ->remoteChannelData->flags.up. */
            return;
        }
        else
        {
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG, "Initialize() failed to enqueue." );
            DOFMem_Free( node );
        }
    }
    OALRemoteChannel_Shutdown( secureGroupChannel, DOFEXCEPTION_FAILED );
}


#define DOFSECUREGROUP_ESP_PROTOCOL_DISCOVERY_TIMEOUT (8000)
#define DOFSECUREGROUP_ESP_PROTOCOL_DISCOVERY_RETRY   (3000)


static void HandleMessage_Initialize( OALChannel secureGroupChannel, OALMessage message )
{
    if (!message->shutdown)
    {
        PCR_Assert(secureGroupChannel->remoteChannelData);
        PCR_Assert(secureGroupChannel->remoteChannelData->flags.isConnection);
        if (DOFConnectionConfig_GetConnectionType( secureGroupChannel->remoteChannelData->config.connection ) ==
            DOFCONNECTIONTYPE_POINT)
        {
           /* This query allows each side of the connection to do protocol discovery. */
           secureGroupChannel->remoteChannelData->flags.waitingForQueryResponse = TRUE;
           ESP_V2_SendQuery( secureGroupChannel, DOFSECUREGROUP_ESP_PROTOCOL_DISCOVERY_TIMEOUT,  /* 8 second timeout. */
                                                 DOFSECUREGROUP_ESP_PROTOCOL_DISCOVERY_RETRY );  /* 3 second retry. */
        }
        else
           Initialize( secureGroupChannel );
    }
    OALMessage_Destroy( message );
}

static void HandleMessage_InitializeComplete( OALChannel secureGroupChannel, OALMessage message )
{
    if (!message->shutdown)
    {
        /* Can get here twice: once at the conclusion of a successful InitialJoin(), and then again a 2nd time if
           this next call sets up a TRP operation which succeeds. */
        if (((DOFSecureGroup)secureGroupChannel)->flags.initializeCalled)
        {
            OALRemoteChannel_HandleMessage_InitializeComplete( secureGroupChannel, message );
            secureGroupChannel->remoteChannelData->flags.sessionIsSecure = TRUE;
            /* Since secure groups are connections set the isConnected flag. */
            secureGroupChannel->remoteChannelData->flags.isConnected = TRUE;
        }
        else
            Initialize( secureGroupChannel );                 /* Sets .initializeCalled then enqueue's InitialJoin(). */
    }
    OALMessage_Destroy(message);
}


static void Process_Shutdown( OALChannel secureGroupChannel, DOFException exception )
{
    DOFSecureGroup secureGroup;
    PCRMutex mutex;

    PCR_Assert( secureGroupChannel );
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "Process_Shutdown called." );
#endif
    secureGroup = (DOFSecureGroup)secureGroupChannel;
    mutex = OALMutexPool_Get( secureGroup );
    PCRMutex_Lock( mutex );
    PCR_Assert( secureGroupChannel->remoteChannelData->flags.isInShutdownState == 0 );
    secureGroupChannel->remoteChannelData->flags.isInShutdownState = 1;     /* Flag to logger in SecureGroup_Dtor(). */
    PCRMutex_Unlock( mutex );
    OALMutexPool_Release( secureGroup );

    /* This next happens also in SecureGroup_Dtor(). Harmless if called more than once. */
    DOFSecureGroup_DenyGroupBehavior( secureGroup );                         /* Disable SGMP packet flow and more. */
    PCR_Assert( secureGroup->specialEvent );
    PCREvent_Signal( secureGroup->specialEvent );
    DOFSecureGroup_CleanupSecureGroupTimer( secureGroup ); /* Returns only after every sg timer callback has completed*/
    /* This next call gets the connection listener callback called with DOFCONNECTIONSTATUS_DISCONNECTED status.
       We believe secureGroupChannel has only one connection, but if more this calls all the listener callbacks. */
    OALRemoteChannel_RemoveOALConnection( secureGroupChannel, exception );
    DOFSecureGroup_CleanupKEKInfoAndDisable( &secureGroup->keks[0] );
    DOFSecureGroup_CleanupKEKInfoAndDisable( &secureGroup->keks[1] );
    OALRemoteChannelMgr_RemoveSilenceTimerItems( secureGroupChannel );
    /* One or the other of my keks are registered. Deregister both. */
    /* DOFDestroy() does this SSID work by different means, if it gets called before we do. Harmless to do twice. */
    OALRemoteChannel_DeregisterSecureChannel_Immediate( secureGroupChannel );
    /* REVIEW: There is a chance that the SecurityStateID can change, and in that case there may be two
       securityStateIDs that must be deregistered.  Currently channel only supports one.
       How do we want to handle two securityStateIDs associated with a single channel. */
    /* HandleMessage_RouteToProvider() or HandleMessage_RouteToRequestor() created ->blockedMessageQueue, so here in
       Process_Shutdown() is a good place to destroy it. */
    OALRemoteChannel_DestroyBlockedMessageQueue( secureGroupChannel );
    /* These next two calls wait for the enqueue's related threads to terminate. */
    OALEnqueue_Destroy( secureGroup->workEnqueue );
    OALRemoteChannelMgr_RemoveChannel_Immediate( secureGroupChannel );
    /* The above call removes the secureGroupChannel reference held by connectionManager. Keep it down here below all
       the action so _Dtor cannot be called until now. */
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "Process_Shutdown() exits." );
#endif
}


static void HandleMessage_Shutdown( OALChannel secureGroupChannel, OALMessage message )
{
    DOFException exception;

    PCR_Assert( message );
    exception = message->argv[0];                                   /* Note that argv[1] is used in _TransportDown(). */
    Process_Shutdown( secureGroupChannel, exception );
    OALMessage_Destroy( message );
}


static void HandleMessage_TransportDown( OALChannel secureGroupChannel, OALMessage message )
{
    /* Only point members have independant connection, so we expect this to be received only if we are a point. */
    DOFException exception;

    PCR_Assert( message );
    exception = message->argv[1];                                   /* Note that argv[0] is used in _Shutdown(). */
    Process_Shutdown( secureGroupChannel, exception );
    OALMessage_Destroy( message );
}


static void HandleMessage_ProcessBlockedMessage( OALChannel secureGroupChannel, OALMessage message )
{
    if (!message->shutdown)
    {
        /* message is our trigger to dequeue from ->blockedMessageQueue. */
        if (message->argv[0])
        {
            PCR_Assert( secureGroupChannel->remoteChannelData->cipher );
            OALCipher_ClearChannelSecureSendBlocked( secureGroupChannel->remoteChannelData->cipher );
        }
        if (secureGroupChannel->remoteChannelData->blockedMessageQueue)
        {
            uint32 size = OALQueue_GetSize( secureGroupChannel->remoteChannelData->blockedMessageQueue );

            if (size > 0)
            {
                /* Remove the first message from the queue and send it to the stack. */
                OALMessage msg = OALQueue_Dequeue( secureGroupChannel->remoteChannelData->blockedMessageQueue );

                PCR_Assert( msg );
                msg->already = message->already;
                OALRemoteChannelMgr_SendMessageToStack( msg );
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                               "Emptied 1 entry from blocked message queue, entries now in queue: 0x",
                                               size-1 );
                if (size > 1) /* We dequeued just above, so >1 now means "not empty." */
                {
                    /* The queue is still not empty ; re-send the trigger message so we are notified again. */
                    message->argv[0] = (void *)0;    /* Clear the argv[0] param that has extra meaning. */
                    OALRemoteChannelMgr_SendMessageToChannel_Internal( message, TRUE );
                    return;
                }
                else
                { /* queue is now empty. */
                }
            }
            else
            { /* Nothing to dequeue. */
            }
        }
        else
        { /* Nothing to dequeue. */
        }
    }
    OALMessage_Destroy(message);
}

static void HandleMessage_ConnectionInitiatedShutdown( OALChannel secureGroupChannel, OALMessage message )
{
    /* Clear the reference to the OALConnection.  This does not require a release.
       Note: Clearing the OALConn will prevent the channel from notifing the OALConn that it was shutdown.
       This prevents a race condition where the OALConn can be destroyed before the notification from
       the channel is received. */
    PCRMutex_Lock(connection_server_mutex);
    secureGroupChannel->remoteChannelData->OALConn = NULL;
    PCRMutex_Unlock(connection_server_mutex);
    Process_Shutdown( secureGroupChannel, NULL );
    OALMessage_Destroy( message );
}

static void HandleMessage_RegisterSecureChannel( OALChannel secureGroupChannel, OALMessage message )
{
    if (!message->shutdown)
        OALRemoteChannel_HandleMessage_RegisterSecureChannel( secureGroupChannel, message );

    OALMessage_Destroy( message );
}


static void HandleMessage_RemoteDomainCallback( OALChannel secureGroupChannel, OALMessage message )
{
    /*  TRP_V1_RequestRemoteDomain() call got us here. */
    if (message->argv[0])
    {
        /* We have new creds for use when asking AS for KEK. Replace secureGroup's storage of that creds. */
        DOFCredentials_Destroy( secureGroupChannel->remoteChannelData->secureCredentials );
        secureGroupChannel->remoteChannelData->secureCredentials = (DOFCredentials)message->argv[0];
        message->argv[0] = NULL;                        /* So message destroy below doesn't kill what we just placed. */
        ((DOFSecureGroup)secureGroupChannel)->specialSuccessFlag = TRUE;
    }
    else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                            "HandleMessage_RemoteDomainCallback() gets no secureCredentials." );
    PCREvent_Signal( ((DOFSecureGroup)secureGroupChannel)->specialEvent );
    OALMessage_Destroy( message );
}


static void HandleMessage_GetLibraryPermsCallback( OALChannel secureGroupChannel, OALMessage message )
{
    /*  OALRemoteChannel_CreateLibPermissions() call got us here. */
    if (message->argv[1])
    {
        secureGroupChannel->remoteChannelData->secureSid = (sid_t *)message->argv[1];
        message->argv[1] = NULL;                        /* So message destroy below doesn't kill what we just placed. */
        ((DOFSecureGroup)secureGroupChannel)->specialSuccessFlag = TRUE;
    }
    else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                            "HandleMessage_GetLibraryPermsCallback() gets no secureSid." );
    PCREvent_Signal( ((DOFSecureGroup)secureGroupChannel)->specialEvent );
    OALMessage_Destroy( message );
}


static void HandleMessage_ProcessReceiveSilence( OALChannel secureGroupChannel, OALMessage message )
{
    if (!message->shutdown)
        OALRemoteChannel_HandleMessage_ProcessReceiveSilence( secureGroupChannel, message );
    OALMessage_Destroy( message );
}


static void HandleMessage_Unexpected( OALChannel secureGroupChannel, OALMessage message )
{
    /*  We currently do not expect to receive this message on a secure group channel.
        Just destroy the message and ignore it. */
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "MessageHandler called with unexpected message type." );
    OALMessage_Destroy( message );
}

static void HandleMessage_SetDomainCredentialsCallback(OALChannel secureGroupChannel, OALMessage message)
{
    if (!message->shutdown)
    {
        OALRemoteChannel_HandleMessage_SetDomainCredentialsCallback(secureGroupChannel, message);
    }
    OALMessage_Destroy(message);
}

static void HandleMessage_PeerListenerEvent(OALChannel secureGroupChannel, OALMessage message)
{
    OALRemoteChannel_HandleMessage_PeerListenerEvent(secureGroupChannel, message);
    OALMessage_Destroy(message);
}

static void HandleMessage_CreateBridge(OALChannel secureGroupChannel, OALMessage message)
{
    if (!message->shutdown)
    {
        OALMessage_CreateBridgeMessage msg = (OALMessage_CreateBridgeMessage)message;
        DOFAddress addr = NULL;

        PCR_Assert(secureGroupChannel->remoteChannelData);
        PCR_Assert(secureGroupChannel->remoteChannelData->flags.isConnection);
        if (DOFConnectionConfig_GetConnectionType( secureGroupChannel->remoteChannelData->config.connection ) == DOFCONNECTIONTYPE_POINT)
        {
            addr = secureGroupChannel->remoteChannelData->address;
        }
        OALRemoteChannel_CreateBridge((OALChannel)secureGroupChannel, addr, msg->dof, msg->bridgeConfig);
    }
    OALMessage_Destroy(message);
}

static void HandleMessage_CreateBridgeComplete(OALChannel secureGroupChannel, OALMessage message)
{
    if (!message->shutdown)
    {
        OALMessage newMessage;

        PCR_Assert(DOFSecurityMembrane_IsChannelMembrane(secureGroupChannel->securityMembrane));
        OALChannelMembrane_SetBridge(secureGroupChannel->securityMembrane, ((OALMessage_CreateBridgeCompleteMessage)message)->bridge);

        /* Now that the bridge is set up, send a message to start initialization. */
        newMessage = CMT_Initialize_Create(secureGroupChannel);
        if (newMessage)
        {
            /* Since we are already in the thread for handling messages, copy the already flag to prevent an extra enqueue. */
            newMessage->already = message->already;

            OALRemoteChannelMgr_SendMessageToChannel(newMessage);
        }
        else
        {
            OALRemoteChannel_Shutdown(secureGroupChannel, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        }
    }
    OALMessage_Destroy(message);
}

static void HandleMessage_ConnectionStreamRequest(OALChannel secureGroupChannel, OALMessage message)
{
    OALRemoteChannel_HandleMessage_ConnectionStreamListener(secureGroupChannel, message);
    OALMessage_Destroy(message);
}

void DOFSecureGroup_HandleMessage( OALChannel secureGroupChannel, OALMessage message )
{
    static secureGroup_MessageHandler_fn_t msg_handler_funcs[] =
    {
        HandleMessage_RouteToProvider,        /* The order here is coupled with ChannelMsgType_t. So don't change it. */
        HandleMessage_RouteToRequestor,
        HandleMessage_ReceiveFromTransport,
        HandleMessage_Initialize,
        HandleMessage_InitializeComplete,
        HandleMessage_Shutdown,
        HandleMessage_TransportDown,
        HandleMessage_ProcessBlockedMessage,
        HandleMessage_ConnectionInitiatedShutdown,
        HandleMessage_RegisterSecureChannel,
        HandleMessage_Unexpected,                       /* If needed, could be HandleMessage_APPIDQueryTimeout. */
        HandleMessage_Unexpected,                       /* If needed, could be HandleMessage_RegisterChannel_ENPPort. */
        HandleMessage_RemoteDomainCallback,
        HandleMessage_Unexpected,                       /* If needed, could be HandleMessage_ResolveCredsCallback. */
        HandleMessage_GetLibraryPermsCallback,
        HandleMessage_Unexpected,                       /* If needed, could be HandleMessage_ServerInitiatedShutdown. */
        HandleMessage_ProcessReceiveSilence,
        HandleMessage_SetDomainCredentialsCallback,
        HandleMessage_PeerListenerEvent,
        HandleMessage_CreateBridge,
        HandleMessage_CreateBridgeComplete,
        HandleMessage_ConnectionStreamRequest
    };
    STATIC_ASSERT( (sizeof( msg_handler_funcs ) / sizeof( secureGroup_MessageHandler_fn_t )) ==
                   (CMT_LAST - CMT_ROUTE_TO_PROVIDER) );   /* These are in ChannelMsgType_t. */

    PCR_Assert( secureGroupChannel );
    PCR_Assert( message );
    PCR_Assert( message->recipient == secureGroupChannel );
    PCR_Assert( message->msg_type < (sizeof( msg_handler_funcs ) / sizeof( secureGroup_MessageHandler_fn_t )) );

    msg_handler_funcs[message->msg_type]( secureGroupChannel, message );
}


/**
@}
**/
