/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/addressinternal.h>
#include <internal/cipherhmacutils.h>
#include <internal/error.h>
#include <internal/mutex_pool.h>
#include <internal/operationinternal.h>
#include <internal/permissionsetinternal.h>
#include <internal/remotechannel.h>
#include <internal/sgmp_v1.h>
#include <internal/timeinternal.h>
#include "securegroupprivate.h"


/** @addtogroup DOFSecureGroup
@{
**/


/*
**
**  "Private" Type and Method Declarations
*/


typedef struct InfoContext_t
{
    OALEnqueueItem_t super;
    DOFSecureGroup   secureGroup;
    OALCryptKey *    pKey;
    DOFAddress       managerAddress;
    OALStream        streamCcmModeInitialState;
    uint16           getKEKWindowMinutes;
    uint16           secsPromoteDelay;
    uint16           epoch;
    boolean          receivedMulticast;
} InfoContext_t;


/** @internal
    If SGM for the given secureGroup, get DOFGroupKey_t and DOFKEKInfo_t structures.

    @param secureGroup         The secureGroup.
    @param pGroupKeyStruct     Pointer to the buffer to hold the group key.
    @param requestorAddr
    @return DOFERROR_OK if buffer at pGroupKey is filled, or
            DOFERROR_ILLEGAL_ARGUMENT if given group or pGroupKey are NULL, or
            DOFERROR_GROUP_NOT_AVAILABLE if given group is not available on this node, or
            DOFERROR_SGM_CANNOT_GET_GROUP_KEY if group key cannot be returned.
**/
static DOFError GetSGMGroupKey( DOFSecureGroup secureGroup, DOFGroupKey_t * pGroupKeyStruct, DOFAddress requestorAddr )
{
    PCRMutex mutex;
    DOFError retStatus = DOFERROR_OK;

    /* SGM/hub only. If we stop being SGM in the middle of this, prepare answer anyway- caller must decide what to do.*/
    PCR_Assert( !secureGroup->flags.pointMember );        /* If point calls this, need to test for it and do nothing. */
    PCR_Assert( requestorAddr );
    if (!secureGroup || !pGroupKeyStruct)
        return DOFERROR_ILLEGAL_ARGUMENT;
    mutex = OALMutexPool_Get( secureGroup );
    PCRMutex_Lock( mutex );
    if (!secureGroup->up)
        retStatus = DOFERROR_GROUP_NOT_AVAILABLE;
    else
    {
        /* REVIEW: Can we just use pGroupKeyStruct directly? */
        DOFGroupKey_t fillGroupKey;                       /* We are SGM, so fill pGroupKeyStruct. */

        PCRMem_Set( &fillGroupKey, 0, sizeof( DOFGroupKey_t ));
        if (OALCipher_GetSGMGroupKey( ((OALChannel)secureGroup)->remoteChannelData->cipher, &fillGroupKey,
                                      requestorAddr ) == FALSE)
            retStatus = DOFERROR_SGM_CANNOT_GET_GROUP_KEY;    /* Through no fault of his, he does not join our group. */
        else
        {
            /* Every member of pFillGroupKey is filled by the above call, and might be CUR, SLOT, or KEKKEY. */
            PCRMem_Copy( pGroupKeyStruct, &fillGroupKey, sizeof( DOFGroupKey_t ) );
            /* Note: Only caller of this function is DOFSecureGroup_ProcessRequestGroup(). We return the real key, even
                     if it is the KEKKey that is about to promote. If we XOR'ed it with KEK, they cannot communicate. */
        }
    }
    PCRMutex_Unlock( mutex );
    OALMutexPool_Release( secureGroup );

    return retStatus;
}


static void StoreKEKKey( DOFSecureGroup secureGroup, DOFAddress * pManagerAddress, OALStream streamCcmModeInitialState,
                         OALCryptKey newKeyXOR, uint16 secsPromoteDelay, uint16 epoch,
                         boolean rekeyPDUCameToUsMulticast )
{
    /* Note: This call does NOT consume streamCcmModeInitialState. */
    if (secureGroup->keks[secureGroup->flags.fut].filled &&
        epoch == secureGroup->keks[secureGroup->flags.fut].epoch)
    {
        DOFSecureGroup_XORTwoKeys( &secureGroup->keks[secureGroup->flags.fut].KEK[0], &newKeyXOR[0], NULL );
        PCR_Assert( secureGroup->keks[secureGroup->flags.fut].filled );
        /* this is being done before the call to this function. if that changes this should be uncommented */
/*      secsPromoteDelay = DOFSecureGroup_ReduceToThb( secureGroup, secsPromoteDelay );*/
        if (OALCipher_SetKEKKey( ((OALChannel)secureGroup)->remoteChannelData->cipher, pManagerAddress,
                                 streamCcmModeInitialState, newKeyXOR, secsPromoteDelay, epoch,
                                 secureGroup->keks[secureGroup->flags.fut].securityModeConfig,
                                 secureGroup->keks[secureGroup->flags.fut].myPerms,
                                 secureGroup->keks[secureGroup->flags.fut].otherPerms,
                                 secureGroup, rekeyPDUCameToUsMulticast ) != FALSE)
        { /* Success, we are done. */
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            {
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "Stored reconstituted kekKey" );
                /*DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "Stored reconstituted kekKey 1st 4 bytes 0x",*/
                /*                           (uint32)(*((unsigned int *)&newKeyXOR[0])) );                   */
            }
        }
        else
        { /* There are legitimate reasons that cipher would reject _SetKEKKey(). */
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            {
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "Cipher rejects reconstituted kekKey" );
                /*DOFSecureGroup_LogMessage(DOFLOGLEVEL_TRACE,"Cipher rejects reconstituted keKey 1st 4 bytes 0x",*/
                /*                                             (uint32)(*((unsigned int *)&newKeyXOR[0])) );      */
            }
        }
    }
    else
    { /* We can be here if we had previously a epochRekey PDU which was stored as a kekKey and which has now already
         promoted to curKey. It can happen because the SGM sent us a 2nd epochRekey PDU, but sent it so late that we
         did not get it until after the kekKey already promoted. */
    }
}


/* @internal. */
static void CheckAndInsertEpochThread( OALThreadPoolItem dnode, boolean shutdown )
{
    InfoContext_t     * node = (InfoContext_t *)dnode;

    PCR_Assert( dnode != NULL );
    if (shutdown)
    {
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                         "CheckAndInsertEpochInfoThread called with shutdown TRUE, so thread closes." );
    }
    else
    {
        PCRMutex            mutex;
        epochChWorkNode_t * pEpochChWorkNode;

        mutex = OALMutexPool_Get( node->secureGroup );
        PCRMutex_Lock( mutex );                                    /* Deny other thread epochWorkList access. */

        if (!node->secureGroup->channel.remoteChannelData->flags.isInShutdownState)
        {
            /* Is epoch in our list? */
            pEpochChWorkNode = (epochChWorkNode_t *)OALSList_Retrieve( (OALSListNode)node->secureGroup->epochWorkList,
                                                                        DOFSecureGroup_EpochInList,
                                                                        &node->epoch );
            if (!pEpochChWorkNode)
            {
                uint32 ourTminMS;                      /* If not previously scheduled, schedule future kek check. */
                uint32 randomTimeout;

                PCR_Assert( node->secureGroup->getKEKTimerItem );
                OALTimerItem_Remove( node->secureGroup->getKEKTimerItem );
                ourTminMS = node->getKEKWindowMinutes * 60 * 1000;
                ourTminMS /= SCALE_TMAX_TMIN;               /* Ordinarily  SCALE_TMAX_TMIN is 1, so compiler will do nothing. */
                if ((pEpochChWorkNode = DOFMem_Alloc( sizeof( epochChWorkNode_t ) )) != NULL)
                {
                    /* Put info in new epochChWorkNode_t, then insert into epochWorkList. */
                    pEpochChWorkNode->epoch = node->epoch;
                    pEpochChWorkNode->epochCompleteTimeMS = PCRTime_GetMS() + ourTminMS +
                                                            EPOCHCHANGED_REPEAT_DELAY +/*Allow for members that missed 1st PDU*/
                                                            (3 * 1000);  /* Bump rekey out 3 secs, not required to be prompt. */
                    OALSList_Insert( (OALSListNode *)&node->secureGroup->epochWorkList, (OALSListNode)pEpochChWorkNode );
                    PCR_Assert( node->secureGroup->epochWorkList ); /*OALSlist_Insert will not fail so this cannot be NULL now*/
                }
                else if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                    DOFSecureGroup_LogMessage( DOFLOGLEVEL_DEBUG,
                                   "CheckAndInsertEpochThread cannot alloc pEpochChWorkNode, dropping notification of epoch 0x",
                                               (uint32)node->epoch );
                randomTimeout = ourTminMS ? (PCRRandom_Get() % ourTminMS) : 0;
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "CheckAndInsertEpochThread random delay to getKEK (msec): 0x",
                                               randomTimeout );
                DOFSecureGroup_SetEpochTimeout( node->secureGroup, randomTimeout, TRUE );              /* Schedule kek check. */
                /* By time kek is checked, we can have multiple entries in the list (if epoch changes rapidly). */
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                               "I added to epochWorkList for epochComplete to happen in (secs) 0x",
                                               (uint32)(ourTminMS / 1000) );
            }
            else
            {
                /* Drop, this is a duplicate and the epoch is already held in our slist. */
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                 "epochWorkList already has the epoch, dropping incoming epochChanged notification of epoch 0x",
                                               (uint32)node->epoch );
            }
        }
        PCRMutex_Unlock( mutex );
        OALMutexPool_Release( node->secureGroup );
    }

    DOFMem_Free( node );
}


/* @internal.

   This can block if another thread is getting a KEK right now, until the other is done.
*/
static void CheckKEKAndStoreKEKKeyThread( OALThreadPoolItem dnode, boolean shutdown )
{
    uint8            futKEK;
    uint8            opKEK;
    PCRMutex         mutex;
    InfoContext_t  * node = (InfoContext_t *)dnode;

    PCR_Assert( dnode != NULL );
    if (shutdown)
    {
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                          "CheckKEKAndStoreKEKKeyThread called with shutdown TRUE, so thread closes." );
        goto cleanup1;
    }
    mutex = OALMutexPool_Get( node->secureGroup );
    PCRMutex_Lock( mutex );
    while (node->secureGroup->flags.gettingKEK)
    {
        PCRMutex_Unlock( mutex );
        PCRThread_Sleep( 1 );
        PCRMutex_Lock( mutex );
    }
    /* If GetKEKTimeout_fn() had been in process, it is done and we have latest KEK. */
    /* REVIEW: Need accept an unpromoted epochRekey even of same epoch if it comes from a new SGM? */
    futKEK = node->secureGroup->flags.fut;
    opKEK = futKEK ^ 1;

    if (node->secureGroup->channel.remoteChannelData->flags.isInShutdownState)
    {
        /* the secure group has started to shut down so the following is invalid and pointless. */
        goto cleanup2;
    }
    PCR_Assert(node->secureGroup->keks[opKEK].filled);     /* Otherwise we cannot receive the secure kek rekey PDU. */
    if ((DOFTime_Compare_16Bit( node->epoch, node->secureGroup->keks[opKEK].epoch ) > 0) &&     /* Newer than opKEK. */
        (!node->secureGroup->keks[futKEK].filled ||          /* We need get new KEK to take advantage of new kekKEY. */
         (DOFTime_Compare_16Bit( node->epoch, node->secureGroup->keks[futKEK].epoch ) >= 0))) /* Not older than futKEK*/
    {
        /* Incoming rekey is of epoch newer that opKEK and not less than futKEK. */
        epochChWorkNode_t * pEpochChWorkNode;

        /* See if we have epoch or if we need to get epoch. Store rcvd kekKey only after having correct epoch. */
        if (node->secureGroup->keks[futKEK].filled == FALSE || node->epoch != node->secureGroup->keks[futKEK].epoch)
        {
            /* Need epoch later than what we have. */
            node->secureGroup->flags.finishedGettingKEK = 0;                             /* Nothing else clears this. */

            /* If we are within 1 second of getting a new KEK, wait in this thread-yielding loop for it to happen. */
            PCR_Assert( node->secureGroup->getKEKTimerItem );
            while (!node->secureGroup->flags.finishedGettingKEK && /* Set in DOFSecureGroup_GetKEKAndUpdateFutKEKInfo.*/
                   OALTimerItem_TimeRemaining( node->secureGroup->getKEKTimerItem ) < 1000)
            {
                PCRMutex_Unlock( mutex );
                /* Wait for existing get-KEK work to finish. We leave loop only after timeout causes GetKEKTimeout_fn to
                   execute (for which case we leave with .finishedGettingKEK set), or before that happens that something
                   pushes the timeout out beyond 1 sec (for which case we leave with .finishedGettingKEK clear). */
                PCRThread_Sleep( 50 );
                PCRMutex_Lock( mutex );

                /* The mutex was unlocked, so verify again that the timer item exists since OALTimerItem_TimeRemaining asserts that an item was passed in. */
                if (node->secureGroup->channel.remoteChannelData->flags.isInShutdownState)
                {
                    /* The secure group is shutting down, and the timer item may not exist. */
                    goto cleanup2;
                }
            }
            if (!node->secureGroup->flags.finishedGettingKEK)
            {
                /* We didn't get new KEK but we need one, so schedule getting KEK right now. */
                DOFSecureGroup_SetEpochTimeout( node->secureGroup, 0, FALSE );
                while (!node->secureGroup->flags.finishedGettingKEK)
                {
                    PCRMutex_Unlock( mutex );
                    PCRThread_Sleep( 1 );
                    PCRMutex_Lock( mutex );
                    if (node->secureGroup->channel.remoteChannelData->flags.isInShutdownState)
                    {
                        goto cleanup2;
                    }
                }
            }
        }
        else
        { /* Our present futKEK matches epoch of the incoming KEKKey PDU. */
        }

        /* After deciding about get-KEK status, store or discard our newKEKKey. */
        if (node->secureGroup->up && node->secureGroup->keks[futKEK].filled &&
            node->epoch == node->secureGroup->keks[futKEK].epoch)
        {
            node->secsPromoteDelay = DOFSecureGroup_ReduceToThb( node->secureGroup, node->secsPromoteDelay );
            StoreKEKKey( node->secureGroup, &node->managerAddress, node->streamCcmModeInitialState, (uint8 *)node->pKey,
                         node->secsPromoteDelay, node->epoch, node->receivedMulticast );
            /* This epochRekey was sent to all group members, and so we need not epochRekey the group if we happen to
               become SGM. Except in the rarest case, we are NOT SGM right now. Even if we just became SGM (how?),
               everybody, even me, got this epochRekey. So disable the epochComplete timeout service routine, since that
               is where an SGM, old or new, creates a new kekKey for the group and sends it out. */
            OALTimerItem_Remove( node->secureGroup->epochCompleteTimerItem );
        }
        else
        {
            /* Drop incoming new kekKEY, since the AS did not give us a KEK of given epoch. We may have been
               rejected by AS, or the epoch given us may be bogus, or the epoch given us may be old- for all
               these it is very appropriate that we drop the new kekKEY. OR !!!!! it may be that we simply
               failed to contact the AS at all due to LAN problems or similar- in this case the fact that we
               discard the new kekKEY may be inconvenient for us- unless we get the 2nd sent kek rekey, we may
               stop hearing heartbeats and find ourselves stepping up as SGM (but hopefully our ReqG PDU will be
               answered by the SGM and we can rejoin the group instead). */
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,"drop new epochRekey, do not have given epoch." );
        }
        /* If we needed to get new KEK (it needs list entry), we have it now, so can remove list entry. */
        /* Remove epoch's entry in epochWorkList so if we become SGM we don't attempt to send out our own epochRekey. */
        /* mutex is locked, denying other threads access to epochWorkList. */
        pEpochChWorkNode = (epochChWorkNode_t *)OALSList_FindAndRemove((OALSListNode *)&node->secureGroup->epochWorkList, NULL,
                                                                        DOFSecureGroup_EpochInList,
                                                                        &node->epoch );
        if (pEpochChWorkNode)
            DOFMem_Free( pEpochChWorkNode );
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        {
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                "Received epochRekey PDU, cancels epochCompleteTimeout() function, incoming epoch is 0x", node->epoch );
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, pEpochChWorkNode ?
                                                "... and it removes that epoch entry from epochWorkList" :
                                                "... and it finds no epoch entry in epochWorkList" );
        }
    }
    else
    {
        /* Ignore epochRekey for ancient epochs. */
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "thread ignores epochRekey for ancient epoch." );
    }

cleanup2:
    PCRMutex_Unlock( mutex );
    OALMutexPool_Release( node->secureGroup );
cleanup1:
    DOFAddress_Destroy( node->managerAddress );
    DOFMem_Free( node->pKey );
    OALStream_Release( node->streamCcmModeInitialState, FALSE );
    DOFMem_Free( node );
}


/*
**
**  "Public" Type and Method Declarations
**
*/


/* @internal
   Handle incoming heartbeat PDU information.

   If we are not SGM, test to see if my our priority is higher than the SGM. If we have enough group information,
   takeover as SGM by the simple takeover mechanism, which skips the ordinarily required send of a ReqG PDU.

   If we are SGM, test to see which SGM shall step down. If we lose, stop acting as SGM.

   @param secureGroup     The secureGroup.
   @param hisAddress      DOFAddress of the transmitting SGM.
   @param hisHighestAPPID The highest SGMP APPID we can understand.
   @param hisRank         The ability or desire to become the group manager.
   @param tieBreaker      Random number supplied by the sender. This can be any 4-byte vaue.
*/
void DOFSecureGroup_HeartbeatReceiveHandler( DOFSecureGroup secureGroup, DOFAddress hisAddress, uint16 hisHighestAPPID,
                                             uint8 hisRank, uint32 tieBreaker )
{
    PCRMutex mutex;
    boolean  acceptHeartbeat = FALSE;
    boolean  simpleTakeoverHappened = FALSE;

    if (!secureGroup->up || secureGroup->flags.pointMember || secureGroup->flags.hub)
        return;
    mutex = OALMutexPool_Get( secureGroup );
    DOFSecureGroup_AddVersionToAPPIDTable( secureGroup, hisHighestAPPID );
    /* In case we become SGM, we want to know everyody's appid. */
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
        DOFSecureGroup_LogMessage( DOFLOGLEVEL_DEBUG, "SGMP heartbeat PDU received, as SGM: ",
                                   (uint32)secureGroup->flags.activeSGM );
    /* Process state machine reactions to incoming hb. */
    if (!secureGroup->flags.activeSGM)
    {
        /*Note: We cannot use simple takeover mechanism until we have sending SGM's advertisedNextNodeID- he sends it
                occasionally in his secure PDUs. We might have received it earlier or just now in his SGMP.Heartbeat. */
        if (DOFSecureGroup_StepDownTest( secureGroup, hisHighestAPPID, hisRank ) != NOT_STEPDOWN)
            acceptHeartbeat = TRUE;                              /* For do stepdown or tie, don't do simple takeover. */
        else if (OALCipher_IsValid_AdvertisedNNI( ((OALChannel)secureGroup)->remoteChannelData->cipher ))
        {
            /* Ready for simple takeover, meaning we skip the step of sending ReqG PDU before becoming SGM. */
            PCRMutex_Lock( mutex );
            secureGroup->flags.becomeSGMInPlay = 1; /* All 3 needed in _HeartbeatTimeoutHandler() for simple takeover.*/
            secureGroup->flags.simpleTakeoverInPlay = 1;
            secureGroup->flags.hbCameInSinceLastHeartbeatTimeout = 0;
            PCRMutex_Unlock( mutex );
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                    "HB received as non-SGM but ignored- will do simple takeover." );
            DOFSecureGroup_HeartbeatTimeoutHandler( secureGroup );              /* Simple takeover happens within. */
            /*We become SGM. ->flags.mustSlotRekey will be 1. A new HBTO is also scheduled to send a quick vanquish hb*/
            simpleTakeoverHappened = TRUE;
        }
        else
        {
            acceptHeartbeat = TRUE;                       /* Cannot simple takeover without his advertisedNextNodeID. */
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                   "HB received as non-SGM, wanted to do simple takeover but cannot- advertisedNextNodeID not known." );
        }
    }
    if (acceptHeartbeat)
    {
        /* Not here for simple takeover. */
        PCRMutex_Lock( mutex );
        secureGroup->flags.activeSGM = 0;                                  /* For form, we won't be here if I am SGM. */
        secureGroup->flags.becomeSGMInPlay = 0;                            /* hb came in: prevent us from stepping up.*/
        secureGroup->flags.primedToLeaveGroupAt3Thb = 0;
        secureGroup->flags.hbCameInSinceLastHeartbeatTimeout = 1;
        PCRMutex_Unlock( mutex );
    }
    OALMutexPool_Release( secureGroup );
    if (secureGroup->flags.activeSGM && !simpleTakeoverHappened)
    {
        uint8 testResult;

        /* Some other SGM sent me a heartbeat w/ my groupKey. Shall I step down as SGM? If so, step down quietly. */
        /* Note: Though his sent hb may not be at his highest version, test is against his highest understood version.*/
        testResult = DOFSecureGroup_StepDownTest( secureGroup, hisHighestAPPID, hisRank );
        if (testResult == TIE_STEPDOWN)
        {
            /* Already have the SGM's groupKey, the normal tiebreak (proven since I received hb). Use special tiebreak*/
            testResult = (tieBreaker < secureGroup->tieBreaker) ? NOT_STEPDOWN:  /* 2**32 chance of equal tieBreaker. */
                                                                  DO_STEPDOWN;
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            {
                DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                            "<>HB received as SGM, AND for otherwise tie vote we use tieBreak. His: 0x",
                                            (uint32)tieBreaker );
                DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                           "<>HB received as SGM, AND for otherwise tie vote we use tieBreak. Mine: 0x",
                                            (uint32)secureGroup->tieBreaker );
            }
        }
        PCR_Assert( testResult != TIE_STEPDOWN );
        if (testResult == NOT_STEPDOWN)
        {
            /* He will hear our hb and he will stop being SGM. So enable our tx of his version-level. */
            DOFSecureGroup_AddVersionToAPPIDTable( secureGroup, hisHighestAPPID );
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                    "<>HB received when I am SGM- I stay SGM. Sending HB to nuke him and establish myself to others." );
            /* Send mcast hb, to have the lesser priority SGM stop letting new members join, and to have all group
               members lose their belief that that node is the real SGM. He MUST lose so this will work, but my test
               just above shows that he will. */
            /* REVIEW: Send at hisHighestAPPID instead of at DOFSECUREGROUP_MY_HIGHEST_APPID? */
            SGMP_V1_SendHeartbeat( (OALChannel)secureGroup, NULL, DOFSecureGroup_GetSecurityScope( secureGroup ),
                                   DOFSECUREGROUP_MY_HIGHEST_APPID, DOFSecureGroup_GetRank( secureGroup ),
                                   secureGroup->tieBreaker );
        }
        else
        {
            PCR_Assert( testResult == DO_STEPDOWN );
            DOFSecureGroup_ChangeSGMState( secureGroup, !BE_SGM );                      /* I lose, so stop being SGM. */
            /* We already have that SGM's group key, proven in that I received his hb. So do not merge my group, in fact
               I cannot merge my group because both group's have the same key and I would cause all to invalidate
               state when I only want my members to do that.
               REVIEW: Is it true that there can be no overlap in my group's nodeID values? There can be overlap if
                       truly to two separated SGM's somehow picked the same groupKey. */
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                    "<>HB received when I am SGM- I step down as SGM." );
        }
    }
    if (acceptHeartbeat)
    {
        /* Not here for simple takeover. */
        if (!secureGroup->flags.activeSGM)                               /* Other things may have cleared .activeSGM. */
        {
            /* We accept the incoming hb and we are no longer SGM (if we were), so push out the heartbeat timeout. */
            if (DOFSecureGroup_ExtendHeartbeatTimeout( secureGroup ) == FALSE)
            { /* How to notify somebody that we cannot prevent hb timeout and we will wrongly try to become SGM?*/
            }
            /* Note: An SGM can send same heartbeat multiple times across different APPIDs. We have no good way to weed
                     out duplicates. As a non-SGM, we just process whatever hb comes in, duplicate or not. */
        }
        /* Last of all, store new hisAddress to secureGroup. */
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        {
            if (!hisAddress)
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "!!! received hb from SGM of NULL DOFAddress" );
            else if (DOFAddress_IsEqual( hisAddress, secureGroup->lastReceivedSGMAddress ) == FALSE)
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                    "!!! received hb from a different SGM as last time" );
        }
        if (hisAddress)
            DOFSecureGroup_UpdateLastReceivedSGMAddress( secureGroup, hisAddress );
    }
}


/* @internal
   Handle incoming slotRekey PDU information. If valid, we rekey.

   @param secureGroup          The secureGroup.
   @param secsPromoteDelay     When the supplied groupKey takes effect.
   @param streamCcmModeInitialState
   @param pGroupKey            Pointer to supplied groupKey.
   @param receivedMulticast
*/
void DOFSecureGroup_RekeyReceiveHandler( DOFSecureGroup secureGroup, uint16 secsPromoteDelay,
                                         OALStream streamCcmModeInitialState, const uint8 * pGroupKey,
                                         boolean receivedMulticast )
{
    /* Note: This call does NOT consume streamCcmModeInitialState. */
    if (secureGroup->up)                                              /* Note: Rekey PDU also known as slotRekey PDU. */
    {
        if (!secureGroup->flags.activeSGM) /* Rekeys sent by SGMs, only. Ordinarily I get no rekey if SGM, so ignore. */
        {
            PCRMutex mutex = OALMutexPool_Get( secureGroup );

            PCRMutex_Lock( mutex );
            secureGroup->flags.noSlotRekeyReceivedSinceFirstHBTimeout = 0;
            PCRMutex_Unlock( mutex );
            OALMutexPool_Release( secureGroup );
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG, "SGMP slot rekey PDU received." );

            /* Cannot call DOFSecureGroup_UpdateLastReceivedSGMAddress- we are not passed DOFAddress of the sender SGM*/

            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            {
                DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_TRACE, "_RekeyReceiveHandler() received slotRekey PDU");
                /*DOFSecureGroup_LogMessage(DOFLOGLEVEL_TRACE,"I received slotRekey PDU with slotKey 1st 4 bytes: 0x",*/
                /*                          (uint32)(*(const uint32 *)pGroupKey) );                                   */
            }
            secsPromoteDelay = DOFSecureGroup_ReduceToThb( secureGroup, secsPromoteDelay );
            /* Decision: We do not accept incoming same period slot rekey PDU. Not sure why that would happen. One
               test shows it happen, where multicast loop is allowed, but we do not ever allow that. If it happens
               that this means our member loses rekeys and cannot stay with the group, so be it. It will eventually
               send a new ReqG PDU and will get back in sync then. */
            if (OALCipher_SetSlotKey( ((OALChannel)secureGroup)->remoteChannelData->cipher,
                                      NULL, /* We don't have hisAddress. */
                                      streamCcmModeInitialState, pGroupKey, secsPromoteDelay, secureGroup, FALSE,
                                      receivedMulticast ))
            {
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                {
                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "Stored slotKey" );
                    /*DOFSecureGroup_LogMessage(DOFLOGLEVEL_TRACE,"Stored slotKey 1st 4 bytes (from slotRekey PDU)0x",*/
                    /*                          (uint32)(*(const uint32 *)pGroupKey) );                               */
                }
            }
            else
            {
                /* A reason to be here; there is previous slotKey stored that will promote sooner than ours would. */
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                {
                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "Rejected slotKey" );
                    /*DOFSecureGroup_LogMessage(DOFLOGLEVEL_TRACE,"Rejected slotKey 1st 4 bytes(from slotRkey PDU)0x",*/
                    /*                          (uint32)(*(const unsigned int *)(pGroupKey)));                        */
                }
            }
        }
    }
}


/* @internal
   Handle incoming mergeRekey PDU information. If valid, we rekey.

   @param secureGroup          The secureGroup.
   @param streamCcmModeInitialState
   @param pGroupKey            Pointer to supplied groupKey.
*/
void DOFSecureGroup_RekeyMergeReceiveHandler( DOFSecureGroup secureGroup, OALStream streamCcmModeInitialState,
                                              const uint8 * pGroupKey, boolean receivedMulticast )
{
    /* Note: This call does NOT consume streamCcmModeInitialState. */
    if (secureGroup->up && !secureGroup->flags.pointMember && !secureGroup->flags.activeSGM)
    {
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG, "SGMP merge rekey PDU received." );
        if (DOFSecureGroup_ExtendHeartbeatTimeout( secureGroup ) == FALSE)        /* Since willing to rekey from him. */
        { /* Error. How to notify somebody we cannot prevent hb timeout, and we will wrongly try to be SGM? */
        }
        OALCipher_DisableAllPendingRekeys( ((OALChannel)secureGroup)->remoteChannelData->cipher );
        OALCipher_RemoveState( ((OALChannel)secureGroup)->remoteChannelData->cipher );
        /* Cipher class has no nodeID and no manager to get it from, and the _SetCurKey() call will not give us a new
           nodeID. But cipher watches incoming secure PDUs to find the new manager, and will automatically go get a new
           nodeID when he or she is found. */
        if (OALCipher_SetCurKey( ((OALChannel)secureGroup)->remoteChannelData->cipher,
                                 NULL, /* Don't have DOFAddress of the new manager, the one who sourced initial state.*/
                                 streamCcmModeInitialState,
                                 pGroupKey, receivedMulticast ))
        {
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            {
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "Stored curKey" );
                /*DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "Stored nowKey 1st 4 bytes (from mergeRekey PDU) 0x",*/
                /*                                              (uint32)(*(const unsigned int *)pGroupKey) );        */
            }
        }
        else
        { /* How to notify that we could not set curKey. */
        }
    }
    else
    { /* Ignore this PDU if I am a point member, or I am SGM. */
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "I rejected mergeRekey PDU." );
    }
}


/* @internal
   Handle incoming epochRekey PDU information. If valid, we rekey.

   @param secureGroup               The secureGroup.
   @param hisAddress
   @param secsPromoteDelay          Future slot offset from curKey slot. At this future slot, kekKey promotes to curKey.
   @param streamCcmModeInitialState Initial state distributed along with pKey by the key originator.
   @param pKey                      Pointer to supplied groupKey.
   @param epoch                     KEK's epoch to which the supplied groupKey belongs.
   @param receivedMulticast
*/
void DOFSecureGroup_RekeyEpochReceiveHandler( DOFSecureGroup secureGroup, DOFAddress hisAddress,
                                              uint16 secsPromoteDelay, OALStream streamCcmModeInitialState,
                                              uint8 * pKey, uint16 epoch, boolean receivedMulticast )
{
    /* Note: This call does not consume streamCcmModeInitialState. Any extended work here is by OALEnqueue_Enqueue(). */
    if (secureGroup->up && !secureGroup->flags.activeSGM)
    {
        PCRMutex mutex = OALMutexPool_Get( secureGroup );

        PCR_Assert( hisAddress );
        PCRMutex_Lock( mutex );
        if (secureGroup->flags.noEpochRekeySinceNewEpochFound && (secureGroup->lastAcceptedKEKEpoch == epoch))
            secureGroup->flags.noEpochRekeySinceNewEpochFound = 0;
        PCRMutex_Unlock( mutex );
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG, "SGMP receives epoch rekey PDU." );
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "Received epochRekey PDU of epoch 0x", epoch );
        /* Store info about hisAddress in ->lastReceivedSGMAddress and ->lastKEKRekeySGMAddress. */
        /* Go ahead and lock early since the following will lock it anyways. (saves lock/unlock only to lock again). */
        PCRMutex_Lock( mutex );
        if (!secureGroup->flags.pointMember)
            DOFSecureGroup_UpdateLastReceivedSGMAddress( secureGroup, hisAddress );
        secureGroup->flags.lastTwoKEKRekeysFromDifferentSGM = 0;
        if (secureGroup->lastKEKRekeySGMAddress)
        {
            secureGroup->flags.lastTwoKEKRekeysFromDifferentSGM =
                                      !((boolean)DOFAddress_IsEqual( secureGroup->lastKEKRekeySGMAddress, hisAddress ));
            DOFAddress_Destroy( secureGroup->lastKEKRekeySGMAddress );
        }
        secureGroup->lastKEKRekeySGMAddress = DOFAddress_Copy( hisAddress );
        PCRMutex_Unlock( mutex );
        if (epoch > secureGroup->keks[DOFSecureGroup_GetOpKEK( secureGroup )].epoch)
        {
            /* Do work of checking and possibly getting KEK in a new thread. Fill structure to pass to enqueue thread.*/
            InfoContext_t * nodeRekeyKEKInfo = DOFMem_Alloc( sizeof( InfoContext_t ) );

            if (nodeRekeyKEKInfo != NULL)
            {
                PCRMem_Set( nodeRekeyKEKInfo, 0, sizeof( InfoContext_t ) );
                nodeRekeyKEKInfo->pKey = (OALCryptKey *)DOFMem_Alloc( sizeof( OALCryptKey ) );
                if (nodeRekeyKEKInfo->pKey != NULL)
                {
                    nodeRekeyKEKInfo->secureGroup = secureGroup;
                    nodeRekeyKEKInfo->streamCcmModeInitialState = streamCcmModeInitialState;
                    nodeRekeyKEKInfo->receivedMulticast = receivedMulticast;
                    nodeRekeyKEKInfo->secsPromoteDelay = secsPromoteDelay;
                    nodeRekeyKEKInfo->epoch = epoch;
                    PCRMem_Copy( nodeRekeyKEKInfo->pKey, pKey, sizeof( OALCryptKey ) );
                    OALStream_Retain( streamCcmModeInitialState ); /* These next 2 release at enqueue thread exit.*/
                    nodeRekeyKEKInfo->managerAddress = DOFAddress_Copy( hisAddress );
                    /* Enqueue work. */
                    PCR_Assert( secureGroup->workEnqueue );
                    OALEnqueueItem_Init( (OALEnqueueItem)nodeRekeyKEKInfo, CheckKEKAndStoreKEKKeyThread );
                    if (!OALEnqueue_Enqueue( secureGroup->workEnqueue, (OALEnqueueItem)nodeRekeyKEKInfo))
                    {
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG,
                                                                "_RekeyEpochReceiveHandler failed to enqueue." );
                        DOFAddress_Destroy( nodeRekeyKEKInfo->managerAddress );
                        OALStream_Release( streamCcmModeInitialState, TRUE );
                        DOFMem_Free( nodeRekeyKEKInfo->pKey );
                        DOFMem_Free( nodeRekeyKEKInfo );

                    }
                }
                else
                {
                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG,
                                                        "_RekeyEpochReceiveHandler() failed to allocate pKey memory." );
                    DOFMem_Free( nodeRekeyKEKInfo );
                }
            }
            else if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG,
                                                    "_RekeyEpochReceiveHandler() failed to allocate memory." );
        }
        else
        {
            /* epoch same or less than our opKEK, already got our kekKey for that. */
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                    "epochRekey PDU of same or less epoch than our opKEK- drop it." );
        }
        OALMutexPool_Release( secureGroup );
    }
    else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "I rejected epochRekey PDU." );
}


/* @internal. */
boolean DOFSecureGroup_EpochInList( OALSListNode pEpochChWorkNode, void* pEpoch )
{
    return (((epochChWorkNode_t*)pEpochChWorkNode)->epoch == *((uint16*)pEpoch));
}


/* @internal. */
void DOFSecureGroup_DeallocateInList( OALSListNode pEpochChWorkNode, void * arg )
{
    DOFMem_Free( ((epochChWorkNode_t*)pEpochChWorkNode) );
}


/* @internal
   Handle incoming EpochChanged PDU information.

   If we are SGM and do not have the indicated KEK epoch, go get it from the AS.

   @param secureGroup The secureGroup.
   @param address
   @param epoch       Epoch number of the newly discovered KEK.
*/
void DOFSecureGroup_EpochChangedReceiveHandler( DOFSecureGroup secureGroup, DOFAddress hisAddress, uint16 epoch,
                                                uint16 getKEKWindowMinutes )
{
    PCR_Assert( secureGroup );
    if (secureGroup->up)
    {
        uint8    futKEK;
        uint8    opKEK;
        PCRMutex mutex = OALMutexPool_Get( secureGroup );

        PCRMutex_Lock( mutex );
        futKEK = secureGroup->flags.fut;
        opKEK = futKEK ^ 1;
        PCR_Assert( secureGroup->keks[opKEK].filled );
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        {
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                       "_EpochChangedReceiveHandler() received EpochChanged PDU of epoch 0x",
                                       (uint32)epoch );
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                "_EpochChangedReceiveHandler() received EpochChanged PDU with on-the-wire getKEKWindow (minutes) of 0x",
                                       (uint32)getKEKWindowMinutes );
        }
        if (DOFTime_Compare_16Bit( epoch, (secureGroup->keks[futKEK].filled ?
                                           secureGroup->keks[futKEK].epoch : secureGroup->keks[opKEK].epoch) ) > 0)
        {
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                DOFSecureGroup_LogMessage( DOFLOGLEVEL_DEBUG, "SGMP epoch changed PDU received, as SGM: ",
                                           (uint32)secureGroup->flags.activeSGM );
            /* Schedule KEK check. */
            if (secureGroup->flags.activeSGM)
            {
                /* kek check with 0 ms delay. If new KEK we send mcast EpochChanged PDU then later epochRekey PDU. */
                /* We trust secure PDU, we are SGM, so getKEK right away. */
                DOFSecureGroup_SetEpochTimeout( secureGroup, 0, FALSE );
            }
            else
            {
                boolean correctSGM = DOFAddress_IsEqual( secureGroup->lastReceivedSGMAddress, hisAddress );

                PCRMutex_Unlock( mutex );
                OALMutexPool_Release( secureGroup );
                if (correctSGM)
                {
                    /* Within new thread CheckAndInsertEpochThread, check epoch in list and possibly insert it. */
                    InfoContext_t * nodeCheckInsertEpochInfo = DOFMem_Alloc( sizeof( InfoContext_t ) );

                    if (nodeCheckInsertEpochInfo != NULL)
                    {
                        PCRMem_Set( nodeCheckInsertEpochInfo, 0, sizeof( InfoContext_t ) );
                        OALEnqueueItem_Init( (OALEnqueueItem)nodeCheckInsertEpochInfo, CheckAndInsertEpochThread );
                        nodeCheckInsertEpochInfo->secureGroup = secureGroup;
                        nodeCheckInsertEpochInfo->getKEKWindowMinutes = getKEKWindowMinutes;
                        nodeCheckInsertEpochInfo->epoch = epoch;
                        PCR_Assert( secureGroup->workEnqueue );
                        if (!OALEnqueue_Enqueue( secureGroup->workEnqueue, (OALEnqueueItem)nodeCheckInsertEpochInfo))
                        {
                            if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG,
                                                                    "_EpochChangedReceiveHandler failed to enqueue." );
                            DOFMem_Free( nodeCheckInsertEpochInfo );
                        }
                    }
                    else if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                        DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_DEBUG,
                                                           "_EpochChangedReceiveHandler() failed to allocate memory.");
                }
                else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                           "As member, rejecting EpochChanged PDU because sender address not my SGM." );
                return;
            }
        }
        else
        { /* We already know of the epoch we are told about.  It's either older or the same. */
        }
        PCRMutex_Unlock( mutex );
        OALMutexPool_Release( secureGroup );
    }
}


/** @internal
    This routine is used to receive an SGMP RequestGroup PDU and then determine if an RequestGroupResponse reply
    PDU should be sent. A reply, if sent, will only be sent if this node is the secure group manager.

    @param secureGroup    The secureGroup we are authorized to operate in.
    @param requestGroupOp The operation associated with the incoming ReqG PDU.
    @param keyRequest     The request block has information used to verify the signature (SIG).
    @param domain         The domain OID sent as part of the incoming ReqG PDU.
    @param I_Stream       The wire order bytes sent as part of the incoming ReqG PDU.
    @param I_offset       The offset into I_stream where the I block starts.
    @param Scope_offset   The offset into I_stream where the scope bytes start.
    @param epoch          The epoch sent as part of the incoming ReqG PDU.
    @param hisTransmittedAPPID APPID of the actual RequestGroup PDU transmitted to us (not sent to us as a PDU param).
**/
void DOFSecureGroup_ProcessRequestGroup( DOFSecureGroup secureGroup, DOFOperation requestGroupOp, DOFKeyRequest keyRequest, DOFObjectIDDomain domain, OALStream I_Stream, uint32 I_offset, uint32 Scope_offset, uint16 epoch, uint16 hisTransmittedAPPID )
{
    /* Requestor gave us his scope. Protocol spec says hub ought keep track of it, on a per point basis.
       FUTURE: Decide what to do with it, and whether the benefit extends to an SGM tracking normal group member scope*/
    PCR_Assert( secureGroup );

    if (secureGroup->up && secureGroup->flags.activeSGM)
    {
        DOFCredentials            secureCreds;
        OALCryptKey               hisKEK;
        uint8                     offsetKEK;
        DOFObjectIDAuthentication keyRequestIdentity = NULL;
        boolean                   reject             = TRUE;
        PCRMutex                  mutex              = OALMutexPool_Get( secureGroup );

        PCR_Assert( !secureGroup->flags.pointMember );
        PCR_Assert( requestGroupOp );
        PCR_Assert( requestGroupOp->sources->address );                        /* Hub sends future sgmp pdus to here. */
        if (keyRequest)
        {
            /* SGMP_V1_Receive_ReadKeyRequestBlock() fills keyRequest->credentials.base.domain with
               DOFOBJECTIDDOMAIN_BROADCAST. The actual placement is accomplished in KEYCredentials_InternalCreate(). */
            PCR_Assert( DOFCredentials_GetDomainID( keyRequest->credentials ) != NULL );
            keyRequestIdentity = DOFCredentials_GetIdentity( keyRequest->credentials );
            /* Note: incoming keyRequest's creds member holds identity of the requestor, which is the groupOID of the
               group that the requestor is trying to join. */
        }
        PCRMutex_Lock( mutex );

        if (!secureGroup->channel.remoteChannelData->flags.isInShutdownState)
        {
            secureCreds = (DOFSecureGroup_GetChannel_Internal(secureGroup))->remoteChannelData->secureCredentials;
            /* Test incoming keyRequest and function parameters; See that we have the things a ReqG PDU should have. */
            if (keyRequest && domain && keyRequestIdentity &&
                DOFPermissionSet_GetPermissionCount( keyRequest->permission_set ) == 0)
            {
                PCR_Assert( I_Stream );                      /* SGMP receive guarantees scopeStream is a valid stream. */
                if (OALStream_GetDataSize( I_Stream ) > 0 && /* Spec: Even if one byte of 0, we must have scope. */
                    DOFObjectID_IsEqual( DOFObjectIDDomain_GetObjectID( DOFCredentials_GetDomainID( secureCreds ) ),
                                         DOFObjectIDDomain_GetObjectID( domain ) ) &&
                    DOFObjectIDAuthentication_IsEqual( secureGroup->group, keyRequestIdentity ))
                {
                    reject = FALSE;
                    if (secureGroup->keks[0].filled && secureGroup->keks[0].epoch == epoch)
                        offsetKEK = 0;
                    else if (secureGroup->keks[1].filled && secureGroup->keks[1].epoch == epoch)
                        offsetKEK = 1;
                    else
                    {
                        /* We reject incoming requestGroup PDU. But shall we move in our time to get a new KEK? */
                        uint16  greatestKnownEpoch;
                        boolean greatestKnownEpochIsValid = DOFSecureGroup_GetGreatestKnownEpoch( secureGroup,
                                                                                                  &greatestKnownEpoch );

                        reject = TRUE;
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                     "Received but rejected incoming RequestGroup PDU, of unknown epoch." );
                        if (greatestKnownEpochIsValid && DOFTime_Compare_16Bit( epoch, greatestKnownEpoch ) > 0)
                        {
                            /* He claims to have an epoch later than we know about. We don't trust him so don't go get a new
                               KEK right away. But spec says to at least move in our scheduled time to go get new KEK. */
                            uint32 nextCheckKEKDelayMS = OALTimerItem_TimeRemaining( secureGroup->getKEKTimerItem );

                            if (DOFTime_Compare( nextCheckKEKDelayMS, 5) >= 0)                                 /* 5 msec. */
                            {
                                /* Move in by half. */
                                DOFSecureGroup_SetEpochTimeout( secureGroup, (nextCheckKEKDelayMS / 2), FALSE );
                            }
                            else
                            { /* (nextCheck) is before (now plus a little), so KEK check is past due or about to happen. */
                            }
                        }
                    }
                }
                else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
          "Received but rejected incoming RequestGroup PDU, not of our domain or not of our group, or scopeStream empty." );
            }
            else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                              "Received but rejected incoming RequestGroup PDU with improper parameters." );
        }

        if (!reject)
            PCRMem_Copy( hisKEK, secureGroup->keks[offsetKEK].KEK, sizeof( OALCryptKey ) );

        PCRMutex_Unlock( mutex );
        OALMutexPool_Release( secureGroup );
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_DEBUG, "SGM receives SGMP RequestGroup PDU, accepted: ",
                                       (uint32)( reject == FALSE) );
        if (!reject)
        {
            DOFError      status;
            DOFGroupKey_t groupKeyInfo;           /* Send unicast RequestGroupResp PDU to he that sent us a ReqG PDU. */
            boolean       success = FALSE;

            PCR_Assert( requestGroupOp->sources->address );                    /* Hub sends future sgmp pdus to here. */
            PCRMem_Set( &groupKeyInfo, 0, sizeof( DOFGroupKey_t ));
            DOFOperation_Lock( requestGroupOp );
            status = GetSGMGroupKey(secureGroup, &groupKeyInfo, requestGroupOp->sources->address); /* Gets key/state. */
            if (status == DOFERROR_OK)
            {

                OALSecureHMACDigest digest;
                uint8               myRank = DOFSecureGroup_GetRank( secureGroup );
                const uint8*        I;
                uint32              I_Len;

                OALStream_SetReadOffset(I_Stream, I_offset);
                I = OALStream_GetReadPtr(I_Stream);
                I_Len = OALStream_GetDataSize(I_Stream);

                /* Set it back to where the scope is because of the unmarshal*/
                OALStream_SetReadOffset(I_Stream, Scope_offset);

                /* REVIEW: Send at hisHighestAPPID instead of at DOFSECUREGROUP_MY_HIGHEST_APPID. */
                if( OALSecureHmac_CreateRequestGroupResponseMac_IStream( hisKEK, groupKeyInfo.key, digest,  
                                                                         I_Len, I, groupKeyInfo.streamCcmModeInitialState, 
                                                                         DOFSECUREGROUP_MY_HIGHEST_APPID, myRank, 
                                                                         DOFObjectIDDomain_GetObjectID( DOFCredentials_GetDomainID( secureCreds ) ) ) == DOFERROR_OK ) /* mac is in digest. */
                {
                    OALKeyGrant      ticket;
                    OALCryptKey      generatedKey;
                    DOFSecurityScope scope = DOFSecurityScope_Unmarshal( I_Stream, domain );

                    /* Success, so send unicast ReqG response PDU to node that sent us the ReqG PDU. */
                    DOFSecureGroup_GenerateKeyFromMacKekAndGroupKey( hisKEK, generatedKey, digest, groupKeyInfo.key );
                    ticket = OALKeyGrant_Create( digest, generatedKey );
                    /* Send ReqGroupResponse. REVIEW: Send on whichever APPID that ReqGroup PDU came in on. */
                    if (ticket)
                    {
                        /* REVIEW: Send at hisHighestAPPID instead of at DOFSECUREGROUP_MY_HIGHEST_APPID? */
                        if (SGMP_V1_SendRequestGroupResponse( requestGroupOp->sources->channel, ticket,
                                                              requestGroupOp, DOFSECUREGROUP_MY_HIGHEST_APPID,
                                                              /* Above param is OUR highest understood SGMP APPID, not
                                                                 necessarily the APPID that we are sending across. */
                                                              myRank,
                                                              groupKeyInfo.streamCcmModeInitialState ) == DOFERROR_OK)
                        {
                            success = TRUE;
                            if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG,
                                                                    "SGM sends RequestGroup response PDU." );
                            /* GetSGMGroupKey() above filled cipher's address table with reqestGroupOp->address. */
                            if (secureGroup->flags.hub)
                            {
                                if (DOFHub_UpdateAddressTable( secureGroup, requestGroupOp->sources->address, TRUE,
                                                               hisTransmittedAPPID ))
                                {
                                    /* The call retained point's requestGroupOp->address so cipher can keep it in its
                                     tables. If the point times out, ->address is destroyed, and cipher loses it. */

                                    /* Notify the peer listeners a peer was added. */
                                    if (requestGroupOp->sources->address && scope)
                                    {
                                        OALRemoteChannel_PeerListenerEvent((OALChannel)secureGroup, requestGroupOp->sources->address, scope, TRUE);
                                    }
                                }
                                else
                                {
                                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                                        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                           "Hub FAILED to add point to hubPointTree, and dropped it." );
                                    /* We answered the point, but here lost knowledge of him. */
                                }
                            }
                            else
                            { /* As SGM, I have no reason to retain the sender's DOFAddress. It stays in cipher's
                                 tables only as long as anybody else, for their own purposes, needs it to exist. */
                            }
                        }
                        else
                            DOFSecureGroup_ChangeSGMState(secureGroup, !BE_SGM);/*Failed to send ReqGResp so step down*/
                        /* Add his version to our list of tx hb versions. If his version is higher than we know and we
                           are not sending heartbeats at any version, then enable hb tx at our highest version. The new
                           node that just joined will get our hb at either our highest version or at whatever version
                           we have been sending all along. */
                        DOFSecureGroup_AddVersionToAPPIDTable( secureGroup, hisTransmittedAPPID );
                        /* Note: Nothing added when we do not know hisTransmittedAPPID. It is possible that he can join
                                 us with such an APPID. Presumably he will detect that he wins in our ReqGroupResponse
                                 and then do a simple takeover. */
                        if (!secureGroup->activeMemberAPPIDTable)
                        {
                            /* Not only do we not know his APPID, we also are not transmitting any hb at all. */
                            DOFSecureGroup_AddVersionToAPPIDTable( secureGroup, DOFSECUREGROUP_MY_HIGHEST_APPID );
                                                             /* So he sees our hb to takeover. */
                            PCR_Assert( secureGroup->activeMemberAPPIDTable );
                        }
                        else
                        { /* Whether we know his version or not, at least we are sending heartbeat(s). */
                        }
                        OALKeyGrant_Release(ticket);
                    }
                    else
                    { /* Could not create ticket, so step down as SGM. */
                    }

                    DOFSecurityScope_Destroy(scope);
                }
                else
                { /* Cannot form a RequestGroupResponse PDU, so step down as SGM. */
                }
                OALStream_Release( groupKeyInfo.streamCcmModeInitialState, TRUE );
            }
            else
            {
                PCR_Assert( groupKeyInfo.filled == FALSE );             /* We have no group key, so step down as SGM. */
                PCR_Assert( groupKeyInfo.streamCcmModeInitialState == NULL );
            }
            DOFOperation_Unlock( requestGroupOp );
            if (!success)
            {
                DOFSecureGroup_ChangeSGMState( secureGroup, !BE_SGM );             /* Our fault, so step down as SGM. */
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                        "As SGM, accepted RequestGroup PDU not answered." );
            }
        }
    }
    else if (FALSE && DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
    {
        DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                   "DOFSecureGroup_ProcessRequestGroup() rejects incoming ReqG PDU, secureGroup up: 0x",
                                   (uint32)secureGroup->up );
        DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                   "DOFSecureGroup_ProcessRequestGroup() rejects incoming ReqG PDU, I am SGM: 0x",
                                   secureGroup->flags.activeSGM );
    }
}


/**
@}
**/
