/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/bridgeinternal.h>
#include <internal/cipherhmacutils.h>
#include <internal/connectionless_remotechannel.h>
#include <internal/error.h>
#include <internal/exception.h>
#include <internal/mutex_pool.h>
#include <internal/oalstring.h>
#include <internal/remotechannel.h>
#include <internal/remotechannelmanager.h>
#include <internal/security_common.h>
#include <internal/serverinternal.h>
#include <internal/sgmp_v1.h>
#include <internal/timeinternal.h>
#include "securegroupprivate.h"


/** @addtogroup DOFSecureGroup
@{
**/


/*
**
**  "Private" Type and Method Declarations
*/


typedef struct rekeyInfo_t
{
    OALCryptKey    key;
    DOFSecureGroup secureGroup;
    OALStream      streamCcmModeInitialState;
    uint16         epoch;
    uint8          secsPromoteDelay;
    boolean        epochRekeyInProgress;
} rekeyInfo_t;


/* Called in SendKEKRekeyIfSGM() or DOFSecureGroup_SendSlotRekeyIfSGM() by OALAVLTree_Traverse() for as many
   pointAddressInfo_t * items as are held in ->hubPointTree. At each call, pStoredAddressInfoItem is filled with the
   next item in the tree. Our function proto represents:
      boolean (*OALAVLTree_Traversal_fn_t)(pointAddressInfo_t * pStoredAddressInfoItem, rekeyInfo_t * pRekeyInfo); */
static boolean SendKEKOrSlotRekeyFromHubPointTree( OALAVLTreeItem_t item, void * arg )
{
    boolean                    fail                   = TRUE;
    const pointAddressInfo_t * pStoredAddressInfoItem = item;
    rekeyInfo_t *              pRekeyInfo             = arg;

    /* FUTURE: Optimize send by having one PDU be encrypted, without targetNodeID, and have it sent multiple times. In
       doing that, OALEncipher() would need be called once somehow, and in making that call its parameter
       denyPacketReplayProtectionIfSecureUnicastSend would be set. */
    PCR_Assert( pStoredAddressInfoItem );
    PCR_Assert( pStoredAddressInfoItem->address );                                      /* We send unicast. */
    PCR_Assert( pStoredAddressInfoItem->secureGroup );
    PCR_Assert( pStoredAddressInfoItem->secureGroup->flags.hub );
    PCR_Assert( DOFSecureGroup_GetSecurityScope( pStoredAddressInfoItem->secureGroup ) );
    /* When multiple SGMP versions come along, use pStoredAddressInfoItem->hisAPPID to select which version of
       _SendRekeyEpoch() or _SendRekey() to call. */
    if (pRekeyInfo->epochRekeyInProgress)
    {
        if (SGMP_V1_SendRekeyEpoch( (OALChannel)pStoredAddressInfoItem->secureGroup, pStoredAddressInfoItem->address,
                                    DOFSecureGroup_GetSecurityScope( pStoredAddressInfoItem->secureGroup ),
                                    pRekeyInfo->secsPromoteDelay, pRekeyInfo->epoch,
                                    pRekeyInfo->streamCcmModeInitialState, pRekeyInfo->key ) == DOFERROR_OK)
            fail = FALSE;
    }
    else
    {
        if (SGMP_V1_SendRekey( (OALChannel)pStoredAddressInfoItem->secureGroup, pStoredAddressInfoItem->address,
                               DOFSecureGroup_GetSecurityScope( pStoredAddressInfoItem->secureGroup ),
                               pRekeyInfo->secsPromoteDelay,
                               pRekeyInfo->streamCcmModeInitialState, pRekeyInfo->key ) == DOFERROR_OK)
            fail = FALSE;
    }
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ) && fail)
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG,
                                            "SendKEKOrSlotRekeyFromHubPointTree() failed to send Rekey PDU." );

    return FALSE; /* Keep going. */
}


static void SendRekeyToPointMembers( DOFSecureGroup secureGroup, OALCryptKey key, uint16 secsPromoteDelay,
                                     OALStream streamCcmModeInitialState, boolean epochRekeyInProgress )
{
    PCRMutex    mutex;
    rekeyInfo_t rekeyInfo;

    PCRMem_Set( &rekeyInfo, 0, sizeof( rekeyInfo_t ) );
    PCRMem_Copy( rekeyInfo.key, key, sizeof( OALCryptKey ) );
    rekeyInfo.epochRekeyInProgress = epochRekeyInProgress;
    rekeyInfo.secureGroup = secureGroup;
    rekeyInfo.secsPromoteDelay = (uint8)secsPromoteDelay;
    rekeyInfo.streamCcmModeInitialState = streamCcmModeInitialState;
    mutex = OALMutexPool_Get( secureGroup );
    PCRMutex_Lock( mutex );
    rekeyInfo.epoch = secureGroup->keks[secureGroup->flags.fut].epoch;
    OALAVLTree_Traverse( secureGroup->hubPointTree, SendKEKOrSlotRekeyFromHubPointTree, &rekeyInfo );
    PCRMutex_Unlock( mutex );
    OALMutexPool_Release( secureGroup );
}


/*
**  "Internal Public"  Methods.
*/


/* @internal */
void DOFSecureGroup_SendKEKRekeyIfSGM( DOFSecureGroup secureGroup, uint16 secsPromoteDelay )
{
    boolean success = FALSE;

    PCR_Assert( secureGroup );
    PCR_Assert( !secureGroup->flags.pointMember );
    if (secureGroup->up && secureGroup->flags.activeSGM)
    {
        OALCryptKey kekKey;
        uint16      slot;
        uint8       period;

        if (OALCipher_GetKey( ((OALChannel)secureGroup)->remoteChannelData->cipher, KEKKEY, kekKey, &period, &slot ))
        {
            OALStream streamCcmModeInitialState;                     /* Send stored kekKey and its initialState. */
            OALCipher cipher = ((OALChannel)secureGroup)->remoteChannelData->cipher;

            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ) && slot != 0)
                DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "Retrieved kekKey wrongly has non-zero slot: 0x", slot );
            if ((streamCcmModeInitialState = OALCipher_CreateInitialStateStream( cipher, period, slot, 0, FALSE, 0,
                                                                                 -1 ))/*-1: get baseNodeID from period*/
                                                                                       != NULL)
            {
                /* Note: At the moment kekKey promotes to curKey, futKEK promotes to opKEK.  */
                DOFSecureGroup_XORTwoKeys( &secureGroup->keks[secureGroup->flags.fut].KEK[0], &kekKey[0], NULL );
                                                                               /* kekKEY always transmits XOR. */
                /* Adjust secsPromoteDelay if it exceeds Thb, then send. */
                secsPromoteDelay = DOFSecureGroup_ReduceToThb( secureGroup, secsPromoteDelay );
                PCR_Assert( secsPromoteDelay <= 255 );
                PCR_Assert( DOFSecureGroup_GetSecurityScope( secureGroup ) );
                if (secureGroup->up) /* Tested again in case we stepped down just now. */
                {
                    success = TRUE;
                    if (secureGroup->flags.hub)
                        SendRekeyToPointMembers(secureGroup, kekKey, secsPromoteDelay, streamCcmModeInitialState, TRUE);
                    else if (SGMP_V1_SendRekeyEpoch( (OALChannel)secureGroup, NULL,
                                                     DOFSecureGroup_GetSecurityScope( secureGroup ),
                                                     (uint8)secsPromoteDelay,
                                                     secureGroup->keks[secureGroup->flags.fut].epoch,
                                                     streamCcmModeInitialState, kekKey ) == DOFERROR_OK)
                    {
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                                       "I sent epochRekey PDU multicast of secsPromoteDelay 0x",
                                                       (uint32)secsPromoteDelay );
                    }
                    else
                        success = FALSE;
                }
                OALStream_Release( streamCcmModeInitialState, TRUE );
            }
        }
    }
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ) && !success)
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "_SendKEKRekeyIfSGM() failed to send KEK rekey." );
}


/* @internal */
void DOFSecureGroup_SendSlotRekeyIfSGM( DOFSecureGroup secureGroup, uint16 secsPromoteDelay )
{
    boolean success = FALSE;

    PCR_Assert( secureGroup );
    PCR_Assert( !secureGroup->flags.pointMember );
    if (secureGroup->up && secureGroup->flags.activeSGM)
    {
        OALCryptKey slotKey;
        uint16      slot;
        uint8       slotKeyPeriod;

        if (OALCipher_GetKey( ((OALChannel)secureGroup)->remoteChannelData->cipher, SLOT, slotKey, &slotKeyPeriod,
                              &slot ))
        {
            OALStream streamCcmModeInitialState;                 /* Send initialState of cipher's stored slotKey. */
            OALCipher cipher = ((OALChannel)secureGroup)->remoteChannelData->cipher;

            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ) && slot != 0)
                DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "Retrieved slotKey wrongly has non-zero slot: 0x",
                                           (uint32)slot );
            if ((streamCcmModeInitialState = OALCipher_CreateInitialStateStream( cipher, slotKeyPeriod,
                                                                                 slot, 0, FALSE, 0,
                                                                                 -1 ))/*-1: get baseNodeID from period*/
                                                                                       != NULL)
            {
                success = TRUE;
                secsPromoteDelay = DOFSecureGroup_ReduceToThb( secureGroup, secsPromoteDelay );
                PCR_Assert( DOFSecureGroup_GetSecurityScope( secureGroup ) );
                if (secureGroup->flags.hub)
                    SendRekeyToPointMembers( secureGroup, slotKey, secsPromoteDelay, streamCcmModeInitialState,
                                             FALSE );
                else if (SGMP_V1_SendRekey((OALChannel)secureGroup, NULL,
                                           DOFSecureGroup_GetSecurityScope( secureGroup ),
                                           (uint8)secsPromoteDelay, streamCcmModeInitialState,
                                           slotKey) == DOFERROR_OK)
                {
                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                        DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                                   "I sent slotRekey PDU multicast of secsPromoteDelay 0x",
                                                   (uint32)secsPromoteDelay );
                }
                else
                    success = FALSE;
                OALStream_Release( streamCcmModeInitialState, TRUE );
            }
        }
    }
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ) && !success)
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "_SendSlotRekeyIfSGM() failed to send slot rekey." );
}


/* @internal */
void DOFSecureGroup_KEKKeyPromoted( DOFSecureGroup secureGroup, uint16 epoch )
{
    /* Called only from cipher.c. */
    uint8    futKEK;
    PCRMutex mutex = OALMutexPool_Get( secureGroup );

    PCR_Assert( secureGroup );
    PCRMutex_Lock( mutex );
    futKEK = secureGroup->flags.fut;
    if (secureGroup->keks[futKEK].filled && secureGroup->keks[futKEK].epoch == epoch)
    {
        uint8 opKEK;                   /* Promote futKEK to be opKEK and actuate new STATEas in STATEgm and STATEmem. */
        uint8 oldOpKEK;

        secureGroup->flags.fut ^= 1;  /* Promote fut to op. This single step actuates new epoch, KEK, Tmax, and Tmin. */
        oldOpKEK = secureGroup->flags.fut;
        PCR_Assert( oldOpKEK == 0 || oldOpKEK == 1 );
        PCR_Assert( secureGroup->keks[oldOpKEK].filled );
        opKEK = oldOpKEK ^ 1;
        PCR_Assert( secureGroup->keks[opKEK].filled );
        PCR_Assert( secureGroup->keks[opKEK].Thb );
        secureGroup->sendHeartbeatTimeout = secureGroup->keks[opKEK].Thb * 1000;                  /* Actuate New Thb. */
#if (SCALE_THB > 1)
        secureGroup->sendHeartbeatTimeout = DOFSecureGroup_ScaleHeartbeatTimeout( secureGroup,
                                                                                  secureGroup->sendHeartbeatTimeout );
#endif
        /* If new opKEK holds a securityScope and replacement is enabled, replace same within secureGroup's channel. */
        PCR_Assert( secureGroup->keks[opKEK].securityScope );
        /*Note: secureGroup mutex is locked; other thread secureGroup machinations are held off, including destruction*/
        DOFSecureGroup_NewSecurityScopeToChannelSecurityScope( secureGroup->keks[opKEK].securityScope,
                                                               (OALChannel)secureGroup );
        DOFSecureGroup_NewSecureSIDToRemoteChannel( secureGroup->keks[opKEK].sid, (OALChannel)secureGroup );
        /* Set SSID in remoteChannelData and register it with RemoteChannelMgr. */
        OALRemoteChannel_RegisterSecureChannel( (OALChannel)secureGroup, secureGroup->keks[opKEK].stateID, TRUE );
        DOFSecureGroup_CleanupKEKInfoAndDisable( &secureGroup->keks[oldOpKEK] ); /* Cleanup old opKEK and disable it. */
        PCR_Assert( secureGroup->keks[oldOpKEK].filled == FALSE );
        PCR_Assert( secureGroup->keks[opKEK].filled );
        PCR_Assert( secureGroup->keks[opKEK].Tmax );
        secureGroup->sgmNextActiveMemberAPPIDTableCheckTime = PCRTime_GetMS() +
                                              (secureGroup->keks[opKEK].Tmax*SGMP_TMAX_MSEC_MULTIPLIER/SCALE_TMAX_TMIN);
        PCR_Assert( secureGroup->sgmNextActiveMemberAPPIDTableCheckTime );
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "futKEK promoted to opKEK. futKEK now unfilled." );
    }
    PCRMutex_Unlock( mutex );
    OALMutexPool_Release( secureGroup );
}


boolean DOFSecureGroup_LastTwoKEKRekeysFromDifferentSGMs( DOFSecureGroup self )
{
    return !!self->flags.lastTwoKEKRekeysFromDifferentSGM;
}


/**
@}
**/
