/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/exception.h>
#include <internal/mutex_pool.h>
#include <internal/sgmp_v1.h>
#include <internal/timeinternal.h>
#include "securegroupprivate.h"


/** @addtogroup DOFSecureGroup
@{
**/


#define SGMP_V1_NEED_SEND_HB( secureGroup )    ((secureGroup->activeMemberAPPIDTable & SGMP_V1_APPID_130_BIT) != 0 || \
                                                SendHBEvenWithoutGroupMember( secureGroup ))
#define SLOT_REKEYPRECEDESPROMOTESECONDS       (2)
#define KEK_REKEYPRECEDESPROMOTESECONDS        (SLOT_REKEYPRECEDESPROMOTESECONDS)
#define SLOT_SECONDREKEYPRECEDESPROMOTESECONDS (2)
#define KEK_SECONDREKEYPRECEDESPROMOTESECONDS  (SLOT_SECONDREKEYPRECEDESPROMOTESECONDS)


/*
**
**  "Private" Type and Method Declarations
*/


typedef struct epochChangedInfo_t
{
    uint16         epoch;
    uint8          Tmin;
} epochChangedInfo_t;


/* Called in GetKEKTimeout_fn() by OALAVLTree_Traverse() for as many pointAddressInfo_t * items as are held
   in ->hubPointTree. At each call, pStoredAddressInfoItem is filled with the next item in the tree.

   @note Function prototype is OALAVLTree_Traversal_fn_t, representing:
                                        void (*OALAVLTree_Traversal_fn_t)( pointAddressInfo_t * pStoredAddressInfoItem,
                                                                           epochChangedInfo_t * pEpochChangedInfo ); */
static boolean SendEpochChangedFromHubPointTree( OALAVLTreeItem_t item, void * arg )
{
    const pointAddressInfo_t * pStoredAddressInfoItem;
    epochChangedInfo_t * pEpochChangedInfo;

    /* REVIEW: Optimize send by having one PDU be encrypted, without targetNodeID, and have it sent multiple times.
       In doing that, OALEncipher() would need be called once somehow, and in making that call its parameter
       denyPacketReplayProtectionIfSecureUnicastSend would be set. */
    PCR_Assert( item );
    pStoredAddressInfoItem = item;
    PCR_Assert( pStoredAddressInfoItem->address );                                      /* We send unicast. */
    PCR_Assert( pStoredAddressInfoItem->secureGroup );
    PCR_Assert( pStoredAddressInfoItem->secureGroup->flags.hub );
    PCR_Assert( DOFSecureGroup_GetSecurityScope( pStoredAddressInfoItem->secureGroup ) );
    PCR_Assert( arg );
    pEpochChangedInfo = arg;
    /* REVIEW: When multiple SGMP versions come along, use pStoredAddressInfoItem->hisAPPID to select which version
               of _SendEpochChanged() to call. */
    /* Note: Decided to send no retry on EpochChanged PDU. If point misses it, he can still go to AS on RekeyEpoch PDU.
             This allows better scaling when hub has a huge number of point members. */
    if (SGMP_V1_SendEpochChanged( (OALChannel)pStoredAddressInfoItem->secureGroup, pStoredAddressInfoItem->address,
                                  DOFSecureGroup_GetSecurityScope( pStoredAddressInfoItem->secureGroup ),
                                  pEpochChangedInfo->epoch, pEpochChangedInfo->Tmin, 0, 0 ) == DOFERROR_OK)
    {
    }
    else if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
        DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_DEBUG,
                                           "SendEpochChangedFromHubPointTree() failed to send SGMP epoch changed PDU.");

    return FALSE;
}


static boolean SendHBEvenWithoutGroupMember( DOFSecureGroup secureGroup )
{
    return FALSE;
}


/*
**
**  "Public" Type and Method Declarations
*/


/* @internal

   @note This can block while cipher nodeID not valid.
   @note typedef int (*OALTimer_fn_t)(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown);
*/
int DOFSecureGroup_SlotRekeyTimeout_fn( OALTimerItem item, uint32 argc, void *argv[], boolean shutdown )
{
    DOFSecureGroup secureGroup = argv[0];

    PCR_Assert( argc == 1 );
    PCR_Assert( secureGroup );
    if (shutdown)
    {
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))/*We think that a timer callback is never called with shutdown TRUE*/
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                          "SlotRekeyTimeout_fn() called with shutdown TRUE, so timer handler closes." );
    }
    else if (DOFSecureGroup_NowUpOrNeverUp( secureGroup ))
    {
        uint32   timeout = SLOTREKEY_TIMEOUT;
        PCRMutex mutex   = OALMutexPool_Get( secureGroup );

        if (secureGroup->flags.activeSGM)
        {
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "_SlotRekeyTimeout_fn() called as SGM." );
#endif
            if (!secureGroup->flags.sgmScheduleSecondSlotRekey)
            {
                if (secureGroup->flags.sgmScheduleSlotKeyPromote)
                {
                    /* If secure traffic hasn't promoted slot key already, do it now. */
                    OALCipher_AdjustKeys( ((OALChannel)secureGroup)->remoteChannelData->cipher );
                    PCRMutex_Lock( mutex );
                    secureGroup->flags.sgmScheduleSlotKeyPromote = 0;
                    PCRMutex_Unlock( mutex );
                    /*If here, we just sent 2 slotRekey PDUs in last few seconds, so don't check rekey need right now.*/
                }
                else if (secureGroup->flags.mustSlotRekey ||
                         OALCipher_NeedRekey( ((OALChannel)secureGroup)->remoteChannelData->cipher ) != FALSE)
                {
                    uint16 secsPromoteDelay = DOFSecureGroup_ReduceToThb( secureGroup, SLOT_SECSPROMOTEDELAY );

                    PCRMutex_Lock( mutex );
                    secureGroup->flags.mustSlotRekey = 0;
                    PCRMutex_Unlock( mutex );
                    /* Form new slotKey, set our cipher with it, and send it in slotRekey PDU. */
                    if (secsPromoteDelay < SLOT_REKEYPRECEDESPROMOTESECONDS)
                        secsPromoteDelay = SLOT_REKEYPRECEDESPROMOTESECONDS;                            /* 2 seconds. */
                    /* NOTE!!!! Be sure SLOT_SECSPROMOTEDELAY is at least 2. If less, it is possible that our outgoing
                                _SendRekey() PDU could go out too slowly and wrongly encipher with our brand new key. */
                    if (OALCipher_FormNewAndSetSlotOrKEKKey( ((OALChannel)secureGroup)->remoteChannelData->cipher,
                                                             secsPromoteDelay, SLOT, 0,    /* 0: ignored for slotKey. */
                                                             NULL, NULL, NULL, secureGroup ) != FALSE)
                    {
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                            DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_TRACE, "Stored slotKey, sending rekey PDU.");
                        DOFSecureGroup_SendSlotRekeyIfSGM( secureGroup, secsPromoteDelay );       /* REVIEW: enqueue. */
                        PCRMutex_Lock( mutex );
                        secureGroup->flags.sgmScheduleSecondSlotRekey = 1;
                        PCRMutex_Unlock( mutex );
                        /* We will send a 2nd slotRekey by scheduling us to be called again. */
                        if (secsPromoteDelay ==  SLOT_SECONDREKEYPRECEDESPROMOTESECONDS)
                        {
                            timeout = 1000;                                /* Lets have at least 1 second to 2nd PDU. */
                            /* This adjustment means the 2nd PDU sent will have secsPromoteDelay 1 second beyond my (I
                               am SGM) actual promote time, but that is OK. It is better in this case that we have PDU
                               separation in order to achieve on the wire redundancy. */
                        }
                        else
                            timeout = (secsPromoteDelay  > SLOT_SECONDREKEYPRECEDESPROMOTESECONDS) ?
                                      ((secsPromoteDelay - SLOT_SECONDREKEYPRECEDESPROMOTESECONDS) * 1000) : 0;
                        /* 2nd rekey PDU usually sent 2 seconds before key promotion. */
                    }
                    else
                    {
                        /* We did not set a new slot key. One possible reason is that secsPromoteDelay is later than a
                           previous slotKey that is about to promote. So we won't send a PDU. What to do in this case?*/
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                        "_SetSlotKey() not accepted, so slotRekey PDU not sent. secsPromoteDelay is 0x",
                                                       secsPromoteDelay );
                    }
                }
            }
            else
            {
                /* Send 2nd slotRekey PDU- this call won't send it if our slot key has already promoted. */
                DOFSecureGroup_SendSlotRekeyIfSGM( secureGroup, SLOT_SECONDREKEYPRECEDESPROMOTESECONDS );/*REVIEW: enq*/
                PCRMutex_Lock( mutex );
                secureGroup->flags.sgmScheduleSecondSlotRekey = 0;
                secureGroup->flags.sgmScheduleSlotKeyPromote = 1;
                PCRMutex_Unlock( mutex );
                timeout = (SLOT_SECONDREKEYPRECEDESPROMOTESECONDS * 1000) + 300; /*Secure traffic before this promotes*/
            }
        }
        else
        {
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                    "_SlotRekeyTimeout_fn() called as group member." );
            PCRMutex_Lock( mutex );
            secureGroup->flags.sgmScheduleSecondSlotRekey = 0;
            secureGroup->flags.sgmScheduleSlotKeyPromote = 0;
            PCRMutex_Unlock( mutex );
        }
        if (DOFSecureGroup_SetSlotRekeyTimeout( secureGroup, timeout ) == FALSE)
        {
            /* We won't check slot key timeout anymore, and if we are or become SGM, we will not rekey anymore. */
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                 DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG,
                                                     "DOFSecureGroup_SlotRekeyTimeout_fn() failed to extend timeout." );
        }
        OALMutexPool_Release( secureGroup );
    }

    return 0;
}


/* @internal */
/* checkAgainstGrant CANNOT be TRUE if secureGroup->grantTimerItem is not yet present in timer. */
boolean DOFSecureGroup_SetEpochTimeout( DOFSecureGroup secureGroup, uint32 newCheckKEKTimeDelay,
                                        boolean checkAgainstGrant )
{
    /*Caller MUST lock secureGroup mutex- guards against CleanupSecureGroupTimer() which can steal timer items from us*/
    boolean  retBoolean = FALSE;

    PCR_Assert( secureGroup );
    if (DOFSecureGroup_NowUpOrNeverUp( secureGroup ) && !secureGroup->channel.remoteChannelData->flags.isInShutdownState)
    {
        if (checkAgainstGrant)
        {
            uint32 millisToGrantExpire         = OALTimerItem_TimeRemaining( secureGroup->grantTimerItem );
            uint32 maxAcceptableNextCheckDelay = (millisToGrantExpire > SAFE_GET_EPOCH_MILLISECONDS) ?
                                                 (millisToGrantExpire - SAFE_GET_EPOCH_MILLISECONDS) : 0;

            if (maxAcceptableNextCheckDelay > 0)
            {
                /* Schedule a new KEK request. How soon? */
                if (newCheckKEKTimeDelay <= maxAcceptableNextCheckDelay)
                { /* Leave newCheckKEKTimeDelay alone. */
                }
                else
                {
                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                        DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                 "To avoid exhausting grant, _SetEpochTimeout() moved specified next epoch check in by (milliseconds) ",
                                                         (uint32)(newCheckKEKTimeDelay - maxAcceptableNextCheckDelay) );
                    newCheckKEKTimeDelay = maxAcceptableNextCheckDelay;/*SAFE_GET_EPOCH_MILLISECONDS before expiration*/
                }
            }
            else
            {
                /* Do not schedule a new KEK request. */
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_TRACE,
                                                       "_SetEpochTimeout() finds that we lost AS grant or soon will.");
                return FALSE;
            }
        }

        PCR_Assert(secureGroup->getKEKTimerItem);
        /* REVIEW: For newCheckKEKTimeDelay 0, do we get immediate timer service? */
        retBoolean = (OALTimerItem_Insert( secureGroup->getKEKTimerItem, newCheckKEKTimeDelay,
                                           FALSE, FALSE ) == DOFERROR_OK);
    }
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
    {
        if (retBoolean)
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "_SetEpochTimeout() placed new timeout (secs) 0x",
                                       (uint32)(newCheckKEKTimeDelay / 1000) );
        else
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "_SetEpochTimeout() did not place new timeout." );
    }

    return retBoolean;
}


/* @internal */
boolean DOFSecureGroup_ExtendHeartbeatTimeout( DOFSecureGroup secureGroup )
{
    boolean retBoolean = FALSE;
    PCRMutex mutex     = OALMutexPool_Get( secureGroup );

    PCR_Assert( !secureGroup->flags.pointMember );
    PCRMutex_Lock( mutex ); /* Guard against CleanupSecureGroupTimer(), which can steal timer items away from us. */
    if (DOFSecureGroup_NowUpOrNeverUp( secureGroup ))
    {
        PCR_Assert( secureGroup->heartbeatTimerItem );       /* Assert valid because sg is up under mutex protection. */
        retBoolean = OALTimerItem_Insert( secureGroup->heartbeatTimerItem,
                                          (secureGroup->flags.activeSGM ? secureGroup->sendHeartbeatTimeout :
                                                                          secureGroup->sendHeartbeatTimeout * 3 / 2),
                                          FALSE, FALSE ) == DOFERROR_OK;
    }
    PCRMutex_Unlock( mutex );
    OALMutexPool_Release( secureGroup );

    return retBoolean;
}


/* @internal */
static boolean SetEpochCompleteTimeout( DOFSecureGroup secureGroup, uint32 newTimeout )
{
    boolean  retBoolean = FALSE;
    PCRMutex mutex      = OALMutexPool_Get( secureGroup );

    PCRMutex_Lock( mutex ); /* Guard against CleanupSecureGroupTimer(), which can steal timer items away from us. */
    if (DOFSecureGroup_NowUpOrNeverUp( secureGroup ))
    {
        PCR_Assert( secureGroup->epochCompleteTimerItem );   /* Assert valid because sg is up under mutex protection. */
        retBoolean = OALTimerItem_Insert( secureGroup->epochCompleteTimerItem, (newTimeout ? newTimeout : 1),
                                          FALSE, FALSE ) == DOFERROR_OK;
    }
    PCRMutex_Unlock( mutex );
    OALMutexPool_Release( secureGroup );

    return retBoolean;
}


/* @internal. */
static uint32 Enqueue_JoinGroupToDecideToBeSGM( DOFSecureGroup secureGroup, uint32 newTimeout )
{
    SecureGroupContext_t * node;

    PCR_Assert( newTimeout != 0xFFFFFFFF );
    if ((node = (SecureGroupContext_t *)DOFMem_Alloc( sizeof( SecureGroupContext_t ) )) != NULL)
    {
        /* Run JoinGroupToDecideToBeSGM in thread. If join .becomeSGMInPlay goes to 0. Else step us as SGM. */
        OALEnqueueItem_Init((OALEnqueueItem)node, DOFSecureGroup_JoinGroupToDecideToBeSGM);
        node->secureGroup = secureGroup;
        PCR_Assert( secureGroup->workEnqueue );
        if (OALEnqueue_Enqueue( secureGroup->workEnqueue, (OALEnqueueItem)node))
        {
            uint32 minTimeout = (REQUESTGROUP_TOTAL_ATTEMPT_COUNT * REQUESTGROUP_SEND_TIMEOUT) + 500;

            /* Note: We still step down if we hear hb from superior SGM. */
            /* newTimeout still set to 3/2 ->sendHeartbeatTimeout. If we become SGM it will change*/
            if (newTimeout < minTimeout)
                newTimeout = minTimeout; /* Since could take up to minTimeout for us to become SGM*/
            return newTimeout;
        }
        else
            DOFMem_Free(node);
    }
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG, "HeartbeatTimeoutHandler() failed to enqueue." );

    return 0xFFFFFFFF;
}


/* @internal */
boolean DOFSecureGroup_SetSlotRekeyTimeout( DOFSecureGroup secureGroup, uint32 newTimeout )
{
    boolean retBoolean = FALSE;

    if (DOFSecureGroup_NowUpOrNeverUp( secureGroup ))
    {
        PCRMutex mutex = OALMutexPool_Get( secureGroup );

        PCRMutex_Lock( mutex ); /* Guard against CleanupSecureGroupTimer(), which can steal timer items away from us. */
        if (secureGroup->slotRekeyTimerItem)
            retBoolean = OALTimerItem_Insert( secureGroup->slotRekeyTimerItem, (newTimeout ? newTimeout : 1),
                                              FALSE, FALSE ) == DOFERROR_OK;
        PCRMutex_Unlock( mutex );
        OALMutexPool_Release( secureGroup );
    }
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
    {
        if (retBoolean)
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "SetSlotRekeyTimeout() placed new timeout (msecs) 0x",
                                       (uint32)newTimeout );
        else
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "SetSlotRekeyTimeout() did not place new timeout" );
    }

    return retBoolean;
}


/** @internal
   Handle heartbeat timeout. If SGM, send heartbeat and check group things, otherwise setup to perhaps be SGM in future.

   When SGM, this is called every Thb.
   When not SGM, this is NEVER called until SGM disappears, and then at (3 * Thb / 2) after last heartbeat from SGM.

   @note This can block while cipher nodeID not valid
   @note If not able to be an SGM, it is required to leave the group after 2 * Thb after last heartbeat from SGM. This
   case is not coded here because this implemenation is able and willing to be an SGM.
   @param secureGroup The secureGroup that this potential SGM is a member of.
**/
void DOFSecureGroup_HeartbeatTimeoutHandler( DOFSecureGroup secureGroup )
{
    PCR_Assert( secureGroup );
    PCR_Assert( !secureGroup->flags.pointMember );
    PCR_Assert( !secureGroup->flags.hub );
    if (DOFSecureGroup_NowUpOrNeverUp( secureGroup ))
    {
        /* Delays getting here can be vary. Before locking mutex, snapshot time to make SGM's output hb stable. */
        uint32   newTimeout;
        uint32   entryTime                         = PCRTime_GetMS();
        PCRMutex mutex                             = OALMutexPool_Get( secureGroup );
        boolean  hbCameInSinceLastHeartbeatTimeout = secureGroup->flags.hbCameInSinceLastHeartbeatTimeout;

        PCR_Assert( secureGroup->heartbeatTimerItem );
        PCR_Assert( !secureGroup->flags.becomeSGMInPlay || !secureGroup->flags.activeSGM );  /* Not at same time. */
        PCRMutex_Lock( mutex );
        secureGroup->flags.hbCameInSinceLastHeartbeatTimeout = 0;           /* Clear this since we just sent it. */
        PCRMutex_Unlock( mutex );
        newTimeout = (secureGroup->sendHeartbeatTimeout * 3) / 2;                           /* Possible overwrite.*/
        PCR_Assert( newTimeout != 0xFFFFFFFF );      /* If for no other reason, because we divided by 2. */
        if (secureGroup->flags.activeSGM)            /* DOFSecureGroup_HeartbeatReceiveHandler() may have cleared.*/
        {
            boolean needSendHeartbeat = SGMP_V1_NEED_SEND_HB( secureGroup );

            PCR_Assert( !secureGroup->flags.doNotBeSGM );
            if (needSendHeartbeat)
            {
                /* Send out hb at SGMP version numbers for members that have joined our group. Might not send any. */
                /*REVIEW: Send from our list of implemented SGMP APPID instead of only DOFSECUREGROUP_MY_HIGHEST_APPID*/
                PCR_Assert( DOFSecureGroup_GetSecurityScope( secureGroup ) );
                if (SGMP_V1_SendHeartbeat( (OALChannel)secureGroup, NULL,
                                            DOFSecureGroup_GetSecurityScope( secureGroup ),
                                            DOFSECUREGROUP_MY_HIGHEST_APPID, DOFSecureGroup_GetRank( secureGroup ),
                                            secureGroup->tieBreaker ) != DOFERROR_OK)
                {
                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                            "Stepping down as SGM at failure to send hb." );
                    DOFSecureGroup_ChangeSGMState( secureGroup, !BE_SGM );
                    goto sgmSteppedDown;             /* And leave newTimeout at 3/2 sendHeartbeatTimeout. */
                }
            }
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            {
                uint16 txSlot;  /* We may or not have sent hb, but we did not FAIL to send it if it was required. */

                OALCipher_GetCurKeySlot( ((OALChannel)secureGroup)->remoteChannelData->cipher, &txSlot );
                DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, (needSendHeartbeat ?
                                                              "Heartbeat V1 sent of slot 0x" :
                                                              "Heartbeat V1 held of slot 0x"), (uint32)txSlot );
            }
            /* Right here: Send SGMP_V2_APPID hb- for now, there is none and no function to call. */

            /* Cipher slots come from promote time, which does not move around. Overhead coming in and thread
               swapping varies our time in getting here. So, compensate by using entryTime. The alternative is
               to use timer as periodic, but that has disadvantages as we use hb timer flexibly. */
            newTimeout = secureGroup->sendHeartbeatTimeout + entryTime - PCRTime_GetMS(); /* Subtract delta time. */
            if (newTimeout > secureGroup->sendHeartbeatTimeout)
            {
                PCR_Assert( newTimeout >= 0x80000000 );   /* We underflowed and have a huge 32 bit number . . */
                newTimeout = 1;                           /* . . so instead send new hb in 1 msec. */
            }
        }
        else if (secureGroup->flags.becomeSGMInPlay)
        {
            PCR_Assert( !secureGroup->flags.doNotBeSGM );
            if (!hbCameInSinceLastHeartbeatTimeout)
            {
                PCR_Assert( secureGroup->heartbeatTimerItem );
                if (!secureGroup->flags.simpleTakeoverInPlay)
                {
                    uint32 newTimeoutCandidate;                   /* Step up by normal means. */

                    PCR_Assert( newTimeout != 0xFFFFFFFF );
                    newTimeoutCandidate = Enqueue_JoinGroupToDecideToBeSGM( secureGroup, newTimeout );

                    if (newTimeoutCandidate != 0xFFFFFFFF)
                        newTimeout = newTimeoutCandidate; /* newTimeout smaller if it changes. Usually no change. */
                    else
                    { /* Failed call: did not enqueue so take no action to be SGM; continue waiting for heartbeat.*/
                    }
                }
                else
                {
                    PCRMutex_Lock( mutex );
                    secureGroup->flags.mustSlotRekey = 1;
                    PCRMutex_Unlock( mutex );
                    if (DOFSecureGroup_InitiateMySGMBehavior( secureGroup,
                                                              secureGroup->activeMemberAPPIDTable, TRUE ))
                    {

                        /* secureGroup->flags.simpleTakeoverInPlay was cleared in InitiateMySGMBehavior().*/
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                         "Stepped up as SGM by simple takeover, HB goes out now." );
                        /* Order of events when 1st HeartbeatTimeoutHandler() is called in 1 msec:
                            Send vanquish hb using present group key                                       .
                            Check KEK                                                                      .
                            Form new slotKey if indicated and send it in slotRekey (using present groupkey).
                            If new KEK, form new kekKey and send it in epochRekey (using present groupKey)*/
                        /* This is the simple takeover case. Be aware that old SGM may not have sent epochRekey. */
                        if (secureGroup->flags.noEpochRekeySinceNewEpochFound)
                        {
                            /* Last SGM did not send a epochRekey PDU for this epoch, or we did not hear it.
                               When we soon chkKEK in GetKEKTimeout_fn(), and if we so happen to find a new
                               KEK, it may be that all the other group members already know about it and
                               already have a epochRekey for it from the last SGM. But we do not know
                               whether that happened or not, and irregardless of it we will form a new
                               random kekKey and send it out. If others do have a kekKey from the 1st SGM,
                               ours will override because we are a new SGM. */
                        }
                        else
                        {
                             /* Because we did get an epoch rekey, we either already have a kekKey in our
                             cipher or we do not. If we do, with our new role as SGM we will automatically
                             receive notification from cipher that we should send the cleanup epochRekey
                             PDU. If there is not a kekKey in cipher it is because it promoted already and
                             that means that the old SGM already sent the cleanup epochRekey (or if it did
                             not get it sent then we do not know how long ago the kekKey promoted and it
                             is therefore wrong for us to send out a "substitute" cleanup epochRekey. */
                        }
                        goto mutexRelease;   /* We already set hb timeout, need not fall through to set it again. */
                    }
                    else
                    { /* Abort simple takeover as SGM. Something is wrong, but at least there is an SGM out there.*/
                    }
                }
            }
            else
            {
                DOFSecureGroup_ChangeSGMState( secureGroup, !BE_SGM );
                /* And leave newTimeout at 3/2 sendHeartbeatTimeout. */
            }
        }
        else
        {
            /* Not SGM and not ->flags.becomeSGMInPlay, just an ordinary group member. */
            if (secureGroup->flags.doNotBeSGM)
            {
                if (secureGroup->flags.primedToLeaveGroupAt3Thb)
                {
                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                         "3 Thb without HB and we are doNotBeSGM, so leave group.");
                    DOFSecureGroup_LeaveGroup( secureGroup, DOFEXCEPTION_TERMINATED );
                    goto mutexRelease;
                }
                else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ) && !secureGroup->flags.primedToLeaveGroupAt3Thb)
                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                        "1-1/2 Thb happened without HB and we are doNotBeSGM." );
                PCRMutex_Lock( mutex );
                secureGroup->flags.primedToLeaveGroupAt3Thb = 1;                /* We consumed 1-1/2 Thb already. */
                PCRMutex_Unlock( mutex );
            }
            else
            {
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                {
                    if (secureGroup->flags.simpleTakeoverInPlay)
                        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                            "simpleTakeover detected, now becomeSGMInPlay." );
                    else
                        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                            "HB receive timeout. becomeSGMInPlay." );
                }
                PCRMutex_Lock( mutex );
                secureGroup->flags.becomeSGMInPlay = 1;      /* We have been 3/2 Thb without an incoming hb. */
                secureGroup->flags.noSlotRekeyReceivedSinceFirstHBTimeout = 1;
                PCRMutex_Unlock( mutex );
                newTimeout = PCRRandom_Get() % (secureGroup->sendHeartbeatTimeout / 2);  /* half ->sendHBTimeout. */
                PCR_Assert( newTimeout <= (secureGroup->sendHeartbeatTimeout / 2));
                ++newTimeout;                                                            /*Not 0, ignore msec diff*/
            }
            PCRMutex_Lock( mutex );
            secureGroup->flags.hbCameInSinceLastHeartbeatTimeout = 0;
            PCRMutex_Unlock( mutex );
        }
    sgmSteppedDown:
        PCRMutex_Lock( mutex );/*Guard against CleanupSecureGroupTimer(), which can steal timer items away from us*/
        if (DOFSecureGroup_NowUpOrNeverUp( secureGroup ))
        {
            PCR_Assert( secureGroup->heartbeatTimerItem ); /* Assert valid because sg is up under mutex protection*/
            if (OALTimerItem_Insert( secureGroup->heartbeatTimerItem, newTimeout, FALSE, FALSE) != DOFERROR_OK)
            {
                /* Error- cannot fill timer or we could not send our heartbeat. We aren't coming back. How to
                   notify somebody that we cannot become SGM and that we are now a lame group node that will watch
                   its heartbeat timeout no more? And if an SGM we send heartbeats no more. */
            }
        }
        PCRMutex_Unlock( mutex );
    mutexRelease:
        OALMutexPool_Release( secureGroup );
    }
}


/** @internal
   Queue heartbeat timeout of a given secureGroup to be handled in thread separate from the timer thread.

   @note typedef int (*OALTimer_fn_t)(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown);
   @param secureGroup The secureGroup.
**/
int DOFSecureGroup_HeartbeatTimeout_fn( OALTimerItem item, uint32 argc, void *argv[], boolean shutdown )
{
    DOFSecureGroup secureGroup = (DOFSecureGroup)argv[0];

    PCR_Assert( secureGroup );
    PCR_Assert( argc == 1 );
    PCR_Assert( !secureGroup->flags.pointMember );
    if (!shutdown)
        DOFSecureGroup_HeartbeatTimeoutHandler( secureGroup );
    else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                         "_HeartbeatTimeout_fn() called with shutdown TRUE, so timer handler closes." );

    return 0;
}


/* @internal

    @note typedef int (*OALTimer_fn_t)(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown);
    @note This can block while waiting for ReqG response to come back.
*/
int DOFSecureGroup_GetKEKTimeout_fn( OALTimerItem item, uint32 argc, void *argv[], boolean shutdown )
{
    DOFSecureGroup secureGroup = argv[0];

    PCR_Assert( argc == 1 );
    PCR_Assert( secureGroup );
    if (!shutdown)
    {
        PCRMutex mutex;
        uint8    futKEK;
        uint16   foundFutEpoch;
        uint32   epochCompleteDelayMS   = 0xFFFFFFFF;
        boolean  futureKEKUpdated       = FALSE;
        boolean  memberSendEpochChanged = FALSE;
        boolean  sgmSendEpochChanged    = FALSE;
        uint32   nowTime                = PCRTime_GetMS();

        if (!DOFSecureGroup_NowUpOrNeverUp( secureGroup ))
            goto finishedGettingKEKAndOutNoMutex;                     /* secureGroup trying to go down. Not go to AS. */
        mutex = OALMutexPool_Get( secureGroup );
        PCRMutex_Lock( mutex );
        secureGroup->flags.gettingKEK = 1;
        futKEK = secureGroup->flags.fut;
        PCR_Assert( secureGroup->keks[futKEK ^ 1].filled );                         /* opKEK is filled. */
        PCRMutex_Unlock( mutex );
        futureKEKUpdated = DOFSecureGroup_GetKEKAndUpdateFutKEKInfo( secureGroup ); /*Also sets up next getKEK timeout*/
        /* secureGroup->flags.finishedGettingKEK was set by above call, and is not set anywhere else. */

        if (futureKEKUpdated)
        {
            epochChWorkNode_t * pEpochChWorkNode;

            /* React to new KEK, both as SGM or member, but react differently. Is epoch in our list? */
            OALTimerItem_Remove( secureGroup->epochCompleteTimerItem ); /*Deny it, will start it up again if indicated*/
            PCRMutex_Lock( mutex );                                     /* Deny other thread access to epochWorkList. */
            secureGroup->flags.gettingKEK = 0;
            futKEK = secureGroup->flags.fut;   /* Redo since futKEK may have promoted in GetKEKAndUpdateFutKEKInfo(). */
            foundFutEpoch = secureGroup->keks[futKEK].epoch;
            pEpochChWorkNode = (epochChWorkNode_t *)OALSList_FindAndRemove( (OALSListNode *)&secureGroup->epochWorkList, NULL,
                                                                             DOFSecureGroup_EpochInList,
                                                                             &foundFutEpoch );
            if (pEpochChWorkNode)
            {
                /* We now have new epoch matching previous epochChanged PDU that we received as a member.
                   We are SGM only if we stepped up since then. In either case, schedule epochComplete timeout at time
                   specified in that epochChanged PDU. */
                epochCompleteDelayMS = pEpochChWorkNode->epochCompleteTimeMS - nowTime;
                if (epochCompleteDelayMS > 0x7FFFFFFF)
                    epochCompleteDelayMS = 0;                           /* Behind, so schedule epochComplete for now. */
            }
            else
            {
                /* We now have new epoch without having a previously received epochChanged PDU. */
                PCR_Assert( epochCompleteDelayMS == 0xFFFFFFFF ); /* No epochComplete 'til EpochChanged PDU from SGM. */
                if (secureGroup->flags.activeSGM)
                {
                    PCR_Assert( secureGroup->keks[futKEK].Tmin );
                    sgmSendEpochChanged = TRUE;
                    /* As SGM, we just learned of new epoch. Send multicast EpochChanged PDU to group. */
                    /* Schedule epochComplete timeout from KEK's Tmin, at which time we will send KEKRekey PDU twice. */
                    epochCompleteDelayMS = secureGroup->keks[futKEK].Tmin * SGMP_TMIN_MSEC_MULTIPLIER / SCALE_TMAX_TMIN;
                    PCR_Assert( epochCompleteDelayMS );
                }
                else
                    memberSendEpochChanged = TRUE;                           /* Send unicast EpochChanged PDU to SGM. */
            }
            /* Remove all epoch information from list- frees memory. But pEpochChWorkNode is NOT in list so not freed.*/
            OALSList_RemoveAll( (OALSListNode *)&secureGroup->epochWorkList,
                                 DOFSecureGroup_DeallocateInList, NULL );
            PCR_Assert( !secureGroup->epochWorkList );
            if (pEpochChWorkNode)
            {
                /* Re-insert our found node. */
                OALSList_Insert( (OALSListNode *)&secureGroup->epochWorkList, (OALSListNode)pEpochChWorkNode );
                PCR_Assert( secureGroup->epochWorkList );                            /* NULL means no entry. */
            }
            PCRMutex_Unlock( mutex );
        }
        /* For all paths above, we are finished getting KEK. */
        PCRMutex_Lock( mutex );
        secureGroup->flags.gettingKEK = 0;
        PCRMutex_Unlock( mutex );
        if (!DOFSecureGroup_NowUpOrNeverUp( secureGroup ))
        {
            /* Can happen when AS returned explicit rejection- we just left the group, or at app shutdown. */
            PCRMutex_Lock( mutex );                                   /* Deny other thread access to epochWorkList. */
            OALSList_RemoveAll( (OALSListNode *)&secureGroup->epochWorkList,
                                DOFSecureGroup_DeallocateInList, NULL );
            PCRMutex_Unlock( mutex );
            goto finishedGettingKEKAndOut;
        }
        /* Act on all info gathered. */
        if (epochCompleteDelayMS < 0x8000000)
        {
            PCR_Assert( futureKEKUpdated ); /* Since epochCompleteDelayMS only gets updated if we received a new kek. */
            PCRMutex_Lock( mutex );/*Guard against CleanupSecureGroupTimer(), which can steal timer items away from us*/
            if (secureGroup->epochCompleteTimerItem)                              /* Schedule epochCompleteTimerItem. */
            {
                OALTimerItem_Insert( secureGroup->epochCompleteTimerItem, epochCompleteDelayMS, FALSE, FALSE );
                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                    DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                               "At new found KEK, scheduled epochComplete service out (secs) 0x",
                                               (uint32)(epochCompleteDelayMS / 1000) );
            }
            PCRMutex_Unlock( mutex );
        }
        if (DOFSecureGroup_NowUpOrNeverUp( secureGroup ))
        {
            if (memberSendEpochChanged)
            {
                PCR_Assert( futureKEKUpdated );                  /* Demonstrates that foundFutEpoch was filled above. */
                PCR_Assert( secureGroup->lastReceivedSGMAddress );
                PCR_Assert( DOFSecureGroup_GetSecurityScope( secureGroup ) );
                if (SGMP_V1_SendEpochChanged( (OALChannel)secureGroup, secureGroup->lastReceivedSGMAddress, /* unicast*/
                                              DOFSecureGroup_GetSecurityScope( secureGroup ), foundFutEpoch,
                                              0,                                         /* 0 because we are not SGM. */
                                              0, 0 ) == DOFERROR_OK)
                {
                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                        DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                                   "As member, sent EpochChanged unicast to SGM, of epoch 0x",
                                                   (uint32)foundFutEpoch );
                }
                else
                { /* unicast PDU not sent to SGM, not a big deal. */
                }
                /* Continue normally, expecting a epochRekey PDU from SGM. If never heard, do nothing special. */
            }
            else if (secureGroup->flags.activeSGM)   /* DOFSecureGroup_HeartbeatReceiveHandler may have cleared this. */
            {
                boolean mustSlotRekey;

                PCRMutex_Lock( mutex );
                mustSlotRekey = secureGroup->flags.mustSlotRekey;
                PCRMutex_Unlock( mutex );
                if (mustSlotRekey)
                {
                    /* We are new SGM that just stepped up or we are existing SGM that simply needs to send a slot
                       rekey, so schedule slot rekey in 1 msec (includes follow on send.) We don't need to check for
                       step down this time through. */
                    if (DOFSecureGroup_SetSlotRekeyTimeout( secureGroup, 1 ) == FALSE)
                    {
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
                            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG,
                                                   "DOFSecureGroup_GetKEKTimeout_fn() failed to initiate slot rekey." );
                    }
                }
                else if (!secureGroup->flags.hub)
                {
                    /* We are SGM. Step down if another SGM has our latest KEK and has higher priority. */
                    DOFSecureGroup_JoinGroupToDecideToStepDownAsSGM( secureGroup );
                    /* Either step down & fully accept his key, or not. */

                    /* The above call blocks for one second (to gather answers from other SGM(s) out there) and
                       therefore causes our epoch changed PDU to be sent out only after a delay of one second. This
                       delay is harmless, and gives us the chance to step down as SGM if needed. Thus only if we are
                       meant to continue as SGM do we send out epoch changed PDU with the extra work to ourself of
                       setting up for EPOCH REKEY SCHEDULUING. */
                }
                if (secureGroup->flags.activeSGM &&
                    DOFSecureGroup_NowUpOrNeverUp( secureGroup )) /*Check again- we locked mutex for a while above*/
                {
                    /* We are still SGM, and after checking KEK, perhaps scheduling slot rekey, and checking for other
                       SGMs out there, see about sending EpochChanged. */
                    if (sgmSendEpochChanged)
                    {
                        /* Decisions above intend us to send 2 EpochChanged PDUs; multicast (hub unicast transport). */
                        PCR_Assert( futureKEKUpdated );    /* This demonstrates that foundFutEpoch was filled above. */
                        PCR_Assert( DOFSecureGroup_GetSecurityScope( secureGroup ) );
                        if (secureGroup->flags.hub)
                        {
                            uint32             numItems;
                            epochChangedInfo_t epochChangedInfo;

                            /* No memset because both members are now filled. */
                            epochChangedInfo.epoch = foundFutEpoch;
                            PCRMutex_Lock( mutex );
                            epochChangedInfo.Tmin = secureGroup->keks[futKEK].Tmin;
                            OALAVLTree_Traverse( secureGroup->hubPointTree, SendEpochChangedFromHubPointTree,
                                                 &epochChangedInfo );
                            numItems = OALAVLTree_GetNumItems( secureGroup->hubPointTree );
                            PCRMutex_Unlock( mutex );
                            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                                DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                                           "At EpochChanged hub send, hubPointTree entry count now 0x",
                                                           numItems );
                        }
                        else if (SGMP_V1_SendEpochChanged( (OALChannel)secureGroup, NULL,
                                                           DOFSecureGroup_GetSecurityScope( secureGroup ),
                                                           foundFutEpoch, secureGroup->keks[futKEK].Tmin, 1,
                                                           EPOCHCHANGED_REPEAT_DELAY ) == DOFERROR_OK)
                        {
                            /* Note: The retry is handled within the function call. */
                            /* Note: Retry PDU sends the same "Tmin" value. That Tmin is anywhere from 1 min. to 4
                                     hours. If 2nd send is the only PDU received by a group member, it will believe
                                     the end of the go-to-AS window is 3 secs later than everyone else- harmless. */
                            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                                DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_TRACE,
                                          "As SGM, sent EpochChanged PDU multicast, with 1 automatic retry scheduled.");
                        }
                        else
                        {
                            /* epoch changed PDU send failed. Since SGM and not hub, step down. */
                            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                                DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                                        "SGM Failed to send epoch changed PDU" );
                            if (secureGroup->flags.activeSGM)
                            {
                                DOFSecureGroup_ChangeSGMState( secureGroup, !BE_SGM );
                                if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                                    DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                                        "Stepping down as SGM due to epoch changed PDU send failure." );
                            }
                        }
                    }
                    else
                    {
                        PCR_Assert( !futureKEKUpdated );
                        /* AS did not answer, or returned one of its error reports, or it returned the same KEK we
                           already had. For all of these, do nothing and check again at normal time. */
                    }
                }
                else
                {
                    /* We stepped down as SGM. */
                    PCR_Assert( !DOFSecureGroup_NowUpOrNeverUp( secureGroup ) || !secureGroup->flags.hub );
                }
            }
            else
            { /* No need to send epochChanged PDU. Either there was never a need, or SGM was going to but stepped down*/
            }
        }
        else
        { /* secure is going down. */
        }
    finishedGettingKEKAndOut:
        OALMutexPool_Release( secureGroup );
    }
finishedGettingKEKAndOutNoMutex:
#if EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "GetKEKTimeout_fn() exits." );
#endif

    return 0;
}

/* @internal
    @note typedef int (*OALTimer_fn_t)(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown);
*/
int DOFSecureGroup_GrantTimeout_fn( OALTimerItem item, uint32 argc, void *argv[], boolean shutdown )
{
    DOFSecureGroup secureGroup = argv[0];

    PCR_Assert( argc == 1 );
    PCR_Assert( secureGroup );
    if (shutdown)
    {
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer(DOFLOGLEVEL_TRACE,
                                               "GrantTimeout_fn() called with shutdown TRUE, so timer handler closes.");
    }
    else if (!((OALChannel)secureGroup)->remoteChannelData->flags.isInShutdownState)
    {
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                        "GrantTimeout_fn() called with channel not in shutdown state, so LeaveGroup." );
        DOFSecureGroup_LeaveGroup( secureGroup, DOFEXCEPTION_AUTHENTICATION_FAILED );
    }
    else if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                            "GrantTimeout_fn() called with channel in shutdown state, so do nothing." );

    return 0;
}


/* @internal

    @note typedef int (*OALTimer_fn_t)(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown);
    @note This can block while cipher nodeID not valid.
*/
int DOFSecureGroup_EpochCompleteTimeout_fn( OALTimerItem item, uint32 argc, void *argv[], boolean shutdown )
{
    /* Here, an old or new SGM creaes a new kekKey for the group and sends it out. */
    DOFSecureGroup secureGroup = argv[0];

    PCR_Assert( argc == 1 );
    PCR_Assert( secureGroup );
    if (shutdown)
    {
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                      "EpochCompleteTimeout_fn() called with shutdown TRUE, so timer handler closes." );
    }
    else if (DOFSecureGroup_NowUpOrNeverUp( secureGroup ))
    {
        PCRMutex mutex = OALMutexPool_Get( secureGroup );

        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE, "epochComplete service, am SGM: 0x",
                                                          secureGroup->flags.activeSGM );
        if (secureGroup->flags.activeSGM)
        {
            uint32 timeout = 0xFFFFFFFF;

            if (!secureGroup->flags.sgmScheduleSecondKEKRekey )
            {
                /* Here to send 1st epoch rekey, or, we have already sent both and will now promote cipher's KEKKey. */
                if (secureGroup->flags.sgmScheduleKEKKeyPromote)
                {
                    /* We just sent two epochRekey PDUs in the last few seconds. Now promote cipher's KEKkey, the act
                       of which will cause our futKEK to promote. */
                    /* If secure traffic hasn't promoted epoch key already, do it now. */
                    OALCipher_AdjustKeys( ((OALChannel)secureGroup)->remoteChannelData->cipher );
                    PCRMutex_Lock( mutex );
                    secureGroup->flags.sgmScheduleKEKKeyPromote = 0;
                    PCRMutex_Unlock( mutex );
                }
                else
                {
                    /* Send our 1st epochRekey PDU, after 1st forming new kekKey and setting cipher with it. */
                    uint16 secsPromoteDelay = DOFSecureGroup_ReduceToThb( secureGroup, KEK_SECSPROMOTEDELAY );
                    uint8  futKEK           = secureGroup->flags.fut;

                    PCR_Assert( secureGroup->keks[futKEK].filled );
                    if (secsPromoteDelay < KEK_REKEYPRECEDESPROMOTESECONDS)
                        secsPromoteDelay = KEK_REKEYPRECEDESPROMOTESECONDS;                             /* 2 seconds. */
                    /* NOTE!!!! Be sure KEK_SECSPROMOTEDELAY is at least 2. If less, it is possible that our outgoing
                                _SendRekey() PDU could go out too slowly and wrongly encipher with our brand new key. */
                    if (OALCipher_FormNewAndSetSlotOrKEKKey( ((OALChannel)secureGroup)->remoteChannelData->cipher,
                                                             secsPromoteDelay, KEKKEY, secureGroup->keks[futKEK].epoch,
                                                             secureGroup->keks[futKEK].securityModeConfig,
                                                             secureGroup->keks[futKEK].myPerms,
                                                             secureGroup->keks[futKEK].otherPerms,
                                                             secureGroup ) != FALSE)
                    {
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE, "Stored KEKKey." );
                        DOFSecureGroup_SendKEKRekeyIfSGM( secureGroup, secsPromoteDelay );        /* REVIEW: enqueue. */
                        PCRMutex_Lock( mutex );
                        secureGroup->flags.sgmScheduleSecondKEKRekey = 1;
                        secureGroup->flags.noEpochRekeySinceNewEpochFound = 0;
                        PCRMutex_Unlock( mutex );
                        /* We will send a 2nd epochRekey by scheduling us to be called again. */
                        if (secsPromoteDelay ==  KEK_SECONDREKEYPRECEDESPROMOTESECONDS)
                        {
                            timeout = 1000;                                /* Lets have at least 1 second to 2nd PDU. */
                            /* This adjustment means the 2nd PDU sent will have secsPromoteDelay 1 second beyond my (I
                               am SGM) actual promote time, but that is OK. It is better in this case that we have PDU
                               separation in order to achieve on the wire redundancy. */
                        }
                        else
                            timeout = (secsPromoteDelay  > KEK_SECONDREKEYPRECEDESPROMOTESECONDS) ?
                                      ((secsPromoteDelay - KEK_SECONDREKEYPRECEDESPROMOTESECONDS) * 1000) : 0;
                        /* 2nd rekey PDU usually sent 2 seconds before key promotion. */
                    }
                    else
                    {
                        /* We did not set a new kek key. One possible reason is that secsPromoteDelay is later than a
                           previous slotKey that is about to promote. So we won't send a PDU. What to do in this case?*/
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                            DOFSecureGroup_LogMessage( DOFLOGLEVEL_TRACE,
                                        "_SetKEKKey() not accepted, so epochRekey PDU not sent. secsPromoteDelay is 0x",
                                                       secsPromoteDelay );
                    }
                }
            }
            else
            {
                /* Send 2nd KEKRekey PDU- this call won't send it if our KEK key has already promoted. */
                DOFSecureGroup_SendKEKRekeyIfSGM( secureGroup, KEK_SECONDREKEYPRECEDESPROMOTESECONDS ); /* REVIEW: enq*/
                PCRMutex_Lock( mutex );
                secureGroup->flags.sgmScheduleSecondKEKRekey = 0;
                secureGroup->flags.sgmScheduleKEKKeyPromote = 1;
                PCRMutex_Unlock( mutex );
                timeout = (KEK_SECONDREKEYPRECEDESPROMOTESECONDS * 1000) + 300; /* Secure traffic before this promotes*/
            }
            if (timeout != 0xFFFFFFFF)
            {
                if (SetEpochCompleteTimeout( secureGroup, timeout ))
                {
                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                        DOFSecureGroup_LogMessage(DOFLOGLEVEL_TRACE,
                          "EpochCompleteTimeout_fn() schedules another 'epochComplete service' out (msec) 0x", timeout);
                }
                else
                {
                    if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE ))
                        DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_TRACE,
                                       "EpochCompleteTimeout_fn() could not schedule another 'epochComplete service'" );
                }
            }
            else
            { /* Decided not to extend epoch complete timeout. */
            }
        }
        else
        {
            PCRMutex_Lock( mutex );                                                                       /* not SGM. */
            secureGroup->flags.sgmScheduleSecondKEKRekey = 0;
            secureGroup->flags.sgmScheduleKEKKeyPromote = 0;
            PCRMutex_Unlock( mutex );
        }
        OALMutexPool_Release( secureGroup );
    }

    return 0;
}


/**
@}
**/
