/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef _DOF_SECUREGROUPPRIVATE_H
#define _DOF_SECUREGROUPPRIVATE_H


#include <dof/core.h>
#include <internal/cipher.h>
#include <internal/channel.h>
#include <internal/securegroupinternal.h>
#include <internal/sgmp_v1.h>
#include <internal/slist.h>


#ifdef  __cplusplus
extern "C" {
#endif


/** @addtogroup DOFSecureGroup
@{
**/


#define EXTRA_TRP_AND_SGMP_CALLBACK_LOGGING (0)


#define SCALE_TMAX_TMIN                     (1)
#define SCALE_THB                           (1)

STATIC_ASSERT( SCALE_TMAX_TMIN > 0 );
STATIC_ASSERT( SCALE_THB > 0 );


#define NOT_STEPDOWN 0
#define TIE_STEPDOWN 1
#define DO_STEPDOWN  2


#define IMPLEMENTED_SGMP_PROTOCOLS_COUNT    (1)


extern uint16 IMPLEMENTED_SGMP_APPID_LIST[IMPLEMENTED_SGMP_PROTOCOLS_COUNT];         /* List filled in securegroup.c. */


#define IMPLEMENTED_SGMP_APPID_LIST_LENGTH (lengthof( IMPLEMENTED_SGMP_APPID_LIST ))
#define DOFSECUREGROUP_MY_HIGHEST_APPID    (IMPLEMENTED_SGMP_APPID_LIST[IMPLEMENTED_SGMP_APPID_LIST_LENGTH - 1])
#define DOFSECUREGROUP_MY_LOWEST_APPID     (IMPLEMENTED_SGMP_APPID_LIST[0])


/* Special defines to increase readability. A secureGroup is a channel, of course. */
#define DOFSecureGroup_GetChannel_Internal( self ) ((OALChannel)(&self->channel))
#define DOFSecureGroup_Retain( self )              ((DOFSecureGroup)OALChannel_Retain( &self->channel ))
#define DOFSecureGroup_Release( self )             (OALChannel_Release( &self->channel ))


#define SGMP_V1_APPID_130_BIT               0x0001
#define BE_SGM                              TRUE
#define REQUESTGROUP_TOTAL_ATTEMPT_COUNT    2    /* Spec says "the node should retry the RequestGroup PDU." */
#define REQUESTGROUP_SEND_TIMEOUT           1000 /*Enough msec to allow receipt of response before we send 2nd request*/
#define SLOT_SECSPROMOTEDELAY               6
#define KEK_SECSPROMOTEDELAY                SLOT_SECSPROMOTEDELAY
#define EPOCHCHANGED_REPEAT_DELAY           (3000)             /* 3 seconds. */
#define REQUEST_KEK_NONCE_SIZE              DEFAULT_NONCE_SIZE /* As sender, we choose nonce length, spec allows 1-16.*/
#define SGMP_TMIN_MSEC_MULTIPLIER           (1000 * 60 * 1)
#define SGMP_TMAX_MSEC_MULTIPLIER           (1000 * 60 * 8)
#define SAFE_GET_EPOCH_MILLISECONDS         (1501)             /* We can generally get a KEK in 1.5 seconds. */

#define SLOTREKEY_TIMEOUT         (1000 * 60 * 7) /* 7 minutes. */


typedef struct epochChWorkNode_t
{
    OALSListNode_t slistNode;
    uint32         epochCompleteTimeMS;
    uint16         epoch;
} epochChWorkNode_t;


typedef struct reqKEvent_t
{
    PCREvent reqKCallbackEvent;
} reqKEvent_t;


typedef struct DOFKEKInfo_t * DOFKEKInfo;
typedef struct DOFKEKInfo_t
{
    /* These first three are not part of AS state, not placed below OALCryptKey to emphasize separation of "AS state."*/
    DOFPermissionSet      myPerms;    /*perms from AS in _RequestKEKResponse. They are my qualified perms. */
    DOFPermissionSet      otherPerms; /*perms from AS in _RequestKEKResponse. They are of OTHERs that interact with me*/
    DOFSecurityScope      securityScope;

    /* These next seven are AS state. AS guarantees any changes to these are accompanied by a new epoch. */
    OALCryptKey           KEK;
    DOFObjectIDSource     sid;
    uint32                stateID;
    DOFSecurityMode       securityModeConfig;
    uint8                 Tmin;                 /* 1 minute resolution. Max of 4 hours 15 minutes. */
    uint8                 Tmax;                 /* 8 minute resolution. Max of 34 hours. */
    uint8                 Thb;                  /* 1 second resolution. Max of 4 minutes 15 seconds. */

    uint16                epoch;                /* Not placed above the uint8's to emphasize AS state is "per epoch." */

    boolean filled;
} DOFKEKInfo_t;


typedef struct secureGroupFlags
{
    unsigned int sgmScheduleKEKKeyPromote               : 1;
    unsigned int sgmScheduleSecondKEKRekey              : 1;
    unsigned int sgmScheduleSlotKeyPromote              : 1;
    unsigned int sgmScheduleSecondSlotRekey             : 1;
    unsigned int noEpochRekeySinceNewEpochFound         : 1; /* REVIEW: Is this automatic now? */
    unsigned int noSlotRekeyReceivedSinceFirstHBTimeout : 1;
    unsigned int simpleTakeoverInPlay                   : 1;
    unsigned int primedToLeaveGroupAt3Thb               : 1;
    unsigned int doNotBeSGM                             : 1; /*Example: ap requestor only wants out of group if no SGM*/
    unsigned int lastTwoKEKRekeysFromDifferentSGM       : 1;
    unsigned int groupStateGathered                     : 1; /*beginGatherGroupStateTime meaningful only if this FALSE*/
    unsigned int finishedGettingKEK                     : 1;
    unsigned int gettingKEK                             : 1;
    unsigned int mustSlotRekey                          : 1;
    unsigned int fut                                    : 1;
    unsigned int hbCameInSinceLastHeartbeatTimeout      : 1;
    unsigned int activeSGM                              : 1;
    unsigned int becomeSGMInPlay                        : 1;
    unsigned int pointMember                            : 1;
    unsigned int hub                                    : 1;
    unsigned int extendPermissionsASRequestInProgress   : 1;
    unsigned int initializeCalled                       : 1;
} secureGroupFlags;


/** @internal
    Describes a DOFSecureGroup

    @note By the special arrangement that channel is the first parameter, these references deliver the same value:
        DOFSecureGroup                            (ptr to DOFSecureGroup_t struct)
        OALChannel                 (ptr to OALChannel_t struct that has the same address as its parent DOFSecureGroup_t)
        &secureGroup->channel                     (ptr to secureGroup's first member)
        DOFSecureGroup_GetChannel( secureGroup )  (ptr to secureGroup's member known as channel)
        DOFSecureGroup_GetChannel_Internal( sg )  (ptr to secureGroup's member known as channel)
**/
typedef struct DOFSecureGroup_t
{
    /*channel MUST be 1st parameter!!!!! Never change it to be otherwise!!!!! This is checked by STATIC_ASSERT() below*/
    OALChannel_t              channel;
    /* Note: channel has member RemoteChannelData remoteChannelData. It holds our OALCipher. */
    DOFKEKInfo_t              keks[2];
    PCREvent                  reqGRespEvent;
    reqKEvent_t               reqKEvent;
    DOFObjectIDAuthentication group;                        /* A globally unique OID. */
    OALAVLTree                hubPointTree;                 /* Hub list of active points. */
    OALEnqueue                workEnqueue;
    OALQueue                  extendPermissionsFIFO;
    DOFAddress                lastReceivedSGMAddress;       /* Of the SGM giving us last ReqGResponse or last hb PDU. */
    DOFAddress                lastKEKRekeySGMAddress;
    PCRMutex                  reqGRespMutex;
    PCREvent                  specialEvent;
    OALTimer                  timer;
    OALTimerItem              getKEKTimerItem;
    OALTimerItem              grantTimerItem;
    OALTimerItem              heartbeatTimerItem;
    OALTimerItem              slotRekeyTimerItem;           /* SGM only. */
    OALTimerItem              epochCompleteTimerItem;
    epochChWorkNode_t *       epochWorkList;                /* Relies on being set NULL by SG_Create(). */
    uint32                    activeMemberAPPIDTable; /*Bit 0: 0'th entry of SGMP_IMPLEMENTED_APPID_LIST. Only for SGM*/
    uint32                    pastActiveMemberAPPIDTable;   /* Bits line up with activeMemberAPPIDTable. Only for SGM.*/
    uint32                    hubConfiguredPointTimeout;
    /* REVIEW: This check time is not used yet to invalidate entries in the active version table, when known that
               no joins on particular versions have come in since our last check. So right now, heartbeats are sent
               forever across any version that is enabled. */
    uint32                    sgmNextActiveMemberAPPIDTableCheckTime;
    uint32                    sendHeartbeatTimeout;         /* Thb * 1000. Largest is 255000, so uint16 not sufficient*/
    uint32                    beginGatherGroupStateTime;    /* Useful until ->flags.groupStateGathered is set. */
    uint32                    tieBreaker;
    secureGroupFlags          flags;
    uint16                    lastAcceptedKEKEpoch;
    uint8                     epochWorkSemaphore;           /* Relies on being set to 0 by SG_Create(). */
    boolean                   specialSuccessFlag;
    boolean                   upSeen;
    boolean                   up;
} DOFSecureGroup_t;
/* STATIC_ASSERT( offsetof( DOFSecureGroup_t, channel ) == 0 ); */  /* To avoid a warning ths static assert was added to DOFSceureGroup_Create. */


typedef struct pointAddressInfo_t
{
    DOFAddress     address;
    DOFSecureGroup secureGroup;
    OALTimerItem   timerItem;
    uint16         hisAPPID;
} pointAddressInfo_t;


typedef struct SecureGroupContext_t
{
    OALEnqueueItem_t super;
    DOFSecureGroup secureGroup;
} SecureGroupContext_t;


typedef struct secureGroupServerStateListener_t
{
    DOFServerStateListener_t dofServerStateListener;                                                /* Must be first. */
    DOFSecureGroup           secureGroup;
} secureGroupServerStateListener_t;


extern uint32 myHighestAPPIDBit;


extern void             DOFSecureGroup_HeartbeatTimeoutHandler( DOFSecureGroup secureGroup );
extern int              DOFSecureGroup_HeartbeatTimeout_fn( OALTimerItem item, uint32 argc, void *argv[],
                                                            boolean shutdown);
extern int              DOFSecureGroup_GetKEKTimeout_fn(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown);
extern int              DOFSecureGroup_GrantTimeout_fn(OALTimerItem item, uint32 argc, void *argv[], boolean shutdown);
extern uint8            DOFSecureGroup_GetRank( DOFSecureGroup secureGroup );
extern void             DOFSecureGroup_ChangeSGMState( DOFSecureGroup secureGroup, boolean beSGM );
extern boolean          DOFSecureGroup_InitiateMySGMBehavior( DOFSecureGroup secureGroup, uint32 appidVersionBit,
                                                              boolean checkKEKNow );
extern void             DOFSecureGroup_JoinGroupToDecideToBeSGM( OALThreadPoolItem item, boolean shutdown );
extern uint8            DOFSecureGroup_StepDownTest(DOFSecureGroup secureGroup, uint16 hisHighestAPPID, uint8 hisRank);
extern void             DOFSecureGroup_AddVersionToAPPIDTable( DOFSecureGroup secureGroup, uint32 appid );
extern boolean          DOFSecureGroup_ExtendHeartbeatTimeout( DOFSecureGroup secureGroup );
extern boolean          DOFSecureGroup_SetGrantTimeout        ( DOFSecureGroup secureGroup, uint32 grantMillis );
extern boolean          DOFSecureGroup_SetSlotRekeyTimeout( DOFSecureGroup secureGroup, uint32 newTimeout );
extern boolean          DOFSecureGroup_SetEpochTimeout( DOFSecureGroup secureGroup, uint32 newGetKEKTimeout,
                                                        boolean checkAgainstGrant );
extern uint8            DOFSecureGroup_GetKEKRekeyStateIndex( DOFSecureGroup secureGroup, uint8 epoch );
extern uint8            DOFSecureGroup_GetOpenKEKRekeyStateIndex( DOFSecureGroup secureGroup );
extern void             DOFSecureGroup_UpdateLastReceivedSGMAddress(DOFSecureGroup secureGroup, DOFAddress sgmAddress);
extern uint8            DOFSecureGroup_GetOpKEK( DOFSecureGroup secureGroup );
extern boolean          DOFSecureGroup_GetKEKAndUpdateFutKEKInfo( DOFSecureGroup secureGroup );
extern uint8            DOFSecureGroup_ReduceToThb( DOFSecureGroup secureGroup, uint16 secsPromoteDelay );
extern void             DOFSecureGroup_SendKEKRekeyIfSGM( DOFSecureGroup secureGroup, uint16 secsPromoteDelay );
extern void             DOFSecureGroup_SendSlotRekeyIfSGM( DOFSecureGroup secureGroup, uint16 secsPromoteDelay );
extern void             DOFSecureGroup_JoinGroupToDecideToStepDownAsSGM( DOFSecureGroup secureGroup );
extern int              DOFSecureGroup_EpochCompleteTimeout_fn( OALTimerItem item, uint32 argc, void *argv[],
                                                                boolean shutdown );
extern int              DOFSecureGroup_SlotRekeyTimeout_fn( OALTimerItem item, uint32 argc, void *argv[],
                                                            boolean shutdown );
extern boolean          DOFSecureGroup_EpochInList( OALSListNode pEpochChWorkNode, void* pEpoch );
extern void             DOFSecureGroup_DeallocateInList( OALSListNode pEpochChWorkNode, void * arg );
extern void             DOFSecureGroup_LeaveGroup               ( DOFSecureGroup secureGroup, DOFException exception );
extern boolean          DOFSecureGroup_NowUpOrNeverUp           ( const DOFSecureGroup self );
extern void             DOFSecureGroup_HandleMessage            ( OALChannel secureGroupChannel, OALMessage message );
extern void             DOFSecureGroup_InitialJoin              ( OALThreadPoolItem dnode, boolean shutdown );
extern void             DOFSecureGroup_DenyGroupBehavior        ( DOFSecureGroup secureGroup );
extern void             DOFSecureGroup_CleanupKEKInfoAndDisable ( DOFKEKInfo_t * pKEKInfo );
extern void             DOFSecureGroup_CleanupSecureGroupTimer  ( DOFSecureGroup secureGroup );
extern boolean          DOFSecureGroup_GetGreatestKnownEpoch    ( DOFSecureGroup secureGroup,
                                                                  uint16 * pGreatestKnownEpoch );
extern DOFError         DOFSecureGroup_GetKEK                   ( DOFSecureGroup secureGroup,
                                                                  DOFKEKInfo_t * pFillKEKInfo );
extern void             DOFSecureGroup_ExtendPermissions        ( OALChannel chan, DOFAddress address,
                                                                  DOFPermissionSet permSet, DOFOperation pendingOp,
                                                                  DOFOperation provOp, DOFResult pendingResult,
                                                                  boolean requestNotResponse, boolean unused );
extern void             DOFSecureGroup_ExtendTheirPermissions   ( OALChannel sgChannel, DOFAddress address,
                                                                  DOFPermissionSet extendPerms,
                                                                  ExtendCompleteCB extendCompleteCallback,
                                                                  void * context );
extern void             DOFSecureGroup_RouteToProvider          ( OALChannel channel, DOFOperation req_op,
                                                                  DOFOperation prov_op, DOFAddress addr );
extern boolean          DOFSecureGroup_JoinGroup                ( DOFSecureGroup secureGroup, boolean sendUnicast,
                                                                  DOFGroupKey_t * pFillGroupKey, uint16 kekEpoch,
                                                                  boolean oneAttempt, uint8 * pHisRank,
                                                                  uint16 * pHisHighestAPPID, uint32 timeout );
extern uint32           DOFSecureGroup_ComputeAPPIDVersionBit   ( uint16 appID );
extern void             DOFSecureGroup_NewSecureSIDToRemoteChannel( DOFObjectIDSource newSID, OALChannel channel );

extern void             DOFSecureGroup_NewSecurityScopeToChannelSecurityScope( DOFSecurityScope newSecurityScope,
                                                                               OALChannel channel );
extern void             DOFSecureGroup_GenerateKeyFromMacKekAndGroupKey(OALCryptKey KEK, OALCryptKey key,
                                                                        OALSecureHMACDigest mac, OALCryptKey groupKey);
extern void             DOFSecureGroup_XORTwoKeys( OALCryptKey src, OALCryptKey dest1, OALCryptKey dest2 );
#if (SCALE_THB > 1)
extern uint32           DOFSecureGroup_ScaleHeartbeatTimeout    ( DOFSecureGroup secureGroup, uint32 hbTimeout );
#endif


/**
@}
**/


#ifdef  __cplusplus
}
#endif


#endif /* _DOF_SECUREGROUPPRIVATE_H */
