/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/security_common.h>
#include <internal/stream.h>
#include <internal/cipher.h>

extern PCRMutex ref_count_mutex;

OALKeyGrant OALKeyGrant_Create( uint8* mac, uint8* key )
{
    OALKeyGrant ret = (OALKeyGrant)DOFMem_Alloc( sizeof( OALKeyGrant_t ) );

    if( ret )
    {
        ret->refCount = 1;
        PCRMem_Copy( ret->mac, mac, 32 );
        PCRMem_Copy( ret->key, key, 32 );
    }
    return ret;
}

OALKeyGrant OALKeyGrant_Retain( OALKeyGrant grant )
{
    PCR_Assert(grant != NULL);

    PCRMutex_Lock(ref_count_mutex);
    ++(grant->refCount);
    PCRMutex_Unlock(ref_count_mutex);
    return grant;
}

void OALKeyGrant_Release( OALKeyGrant grant )
{
    if(grant)
    {
        boolean doFree;

        PCRMutex_Lock(ref_count_mutex);
        doFree = (--(grant->refCount) == 0);
        PCRMutex_Unlock(ref_count_mutex);
        if( doFree )
        {
            DOFMem_Free( grant );
        }
    }
}

OALKeyGrant OALKeyGrant_Unmarshal( OALStream stream )
{
    OALKeyGrant ret = (OALKeyGrant)DOFMem_Alloc( sizeof( OALKeyGrant_t ) );

    if( ret )
    {
        ret->refCount = 1;
        if( OALStream_GetDataSize( stream ) >= 64 )
        {
            if( OALStream_Read( stream, 32, ret->mac ) &&
                OALStream_Read( stream, 32, ret->key ) )
                return ret;
        }
        DOFMem_Free( ret );
    }
    return NULL;
}

boolean OALKeyGrant_Marshal( OALKeyGrant grant, OALStream stream )
{
    if( grant && stream && OALStream_GetFreeSize( stream ) >= 64)
    {
        if( OALStream_Write( stream, 32, grant->mac ) &&
            OALStream_Write( stream, 32, grant->key ) )
            return TRUE;
    }
    return FALSE;
}
uint32 OALKeyGrant_MarshalSize(OALKeyGrant self)
{
    return 64;
}

void OALKeyGrant_ExtractK(OALKeyGrant grant, const uint8 *shared_secret, uint8 *place_to_put_it)
{
    OALCryptExpandedKey ekey;
    int32 i = 0;
    PCR_Assert(grant);
    PCR_Assert(shared_secret);
    PCR_Assert(place_to_put_it);

    PCRMem_Copy( place_to_put_it, grant->mac, 32 );
    PCR_Validate( DOFCipherRegistry_AES256Cipher );
    DOFCipherRegistry_AES256Cipher->GenerateKeyState( ekey, shared_secret );
    DOFCipherRegistry_AES256Cipher->Encrypt( ekey, &place_to_put_it[0] );
    DOFCipherRegistry_AES256Cipher->Encrypt( ekey, &place_to_put_it[16] );

    for( ; i < 32 ; ++i )
    {
        place_to_put_it[i] = grant->key[i] ^ place_to_put_it[i];
    }
}
