/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/credentialsinternal.h>
#include <internal/permissionsetinternal.h>
#include <internal/security_common.h>

extern PCRMutex ref_count_mutex;

DOFKeyRequest DOFKeyRequest_Create( DOFCredentials credentials, uint8 nonce_size, boolean new_login, boolean fixedPermissions,
                                    const uint8 *nonce, DOFPermissionSet permission_set )
{
    DOFKeyRequest ret;

    PCR_Assert( nonce_size <= MAX_KEYREQUEST_NONCE_SIZE );
    ret = (DOFKeyRequest)DOFMem_Alloc( sizeof( DOFKeyRequest_t ) );
    if( ret )
    {
        /* Clearing ret struct not needed because every member is filled. */
        ret->credentials = DOFCredentials_Copy( credentials );
        if((ret->credentials != NULL) || (credentials == NULL))
        {
            ret->permission_set = (permission_set != NULL) ? DOFPermissionSet_DeepCopy(permission_set)
                                                           : DOFPermissionSet_Create();
            if(ret->permission_set != NULL)
            {
                ret->refCount = 1;
                ret->nonceLength = nonce_size;
                ret->newLogin = new_login;
                ret->fixedPermissions = fixedPermissions;
                if (nonce)
                {
                    PCRMem_Copy( ret->nonce, nonce, ret->nonceLength );
                }
                else
                {
                    PCRRandom_Generate( ret->nonceLength, ret->nonce );
                }
                return ret;
            }
            DOFCredentials_Destroy(ret->credentials);
        }
        DOFMem_Free(ret);
    }
    return NULL;
}

DOFKeyRequest DOFKeyRequest_Retain(DOFKeyRequest self)
{
    PCR_Assert(self != NULL);

    PCRMutex_Lock(ref_count_mutex);
    ++(self->refCount);
    PCRMutex_Unlock(ref_count_mutex);
    return self;
}

void DOFKeyRequest_Destroy(DOFKeyRequest self)
{
    if(self)
    {
        boolean doFree;

        PCRMutex_Lock(ref_count_mutex);
        doFree = (--(self->refCount) == 0);
        PCRMutex_Unlock(ref_count_mutex);
        if( doFree )
        {
            DOFKeyRequest_Cleanup(self);
            DOFMem_Free( self );
        }
    }
}

void DOFKeyRequest_Cleanup(DOFKeyRequest self)
{
    PCR_Assert(self != NULL);

    DOFPermissionSet_Destroy( self->permission_set );
    DOFCredentials_Destroy( self->credentials );
}

boolean DOFKeyRequest_Marshal(OALStream     stream,
                              DOFKeyRequest self,
                              boolean       isRequest)
{
    uint8 firstByte;

    PCR_Assert(stream != NULL);
    PCR_Assert(self != NULL);
    firstByte = (self->nonceLength - 1);
    if(self->newLogin)
    {
        firstByte |= 0x80;
    }
    if(self->fixedPermissions)
    {
        firstByte |= 0x40;
    }
    if(!OALStream_Write8(stream, firstByte))
    {
        return FALSE;
    }
    if(!DOFCredentials_Internal_Marshal(self->credentials, stream, isRequest))
    {
        return FALSE;
    }
    if(!OALStream_Write(stream, self->nonceLength, self->nonce))
    {
        return FALSE;
    }
    return DOFPermissionSet_MarshalRequest(self->permission_set, stream);
}

boolean DOFKeyRequest_Unmarshal(OALStream           stream,
                                DOFKeyRequest       self,
                                DOFObjectIDDomain   domain,
                                boolean             isRequest,
                                boolean             unknownAllowed)
{
    uint8 firstByte;

    PCR_Assert(self != NULL);
    if(OALStream_Read8(stream, &firstByte))
    {
        self->nonceLength = (firstByte & 0x0F) + 1;
        self->newLogin = (firstByte >> 7);
        self->fixedPermissions = ((firstByte & 0x40) >> 6);
        self->credentials = NULL;
        self->refCount = 1;
        if(CREDENTIALS_SUCCESSFUL_UNMARSHAL == DOFCredentials_Unmarshal(&(self->credentials), domain, stream, isRequest))
        {
            if(OALStream_Read(stream, self->nonceLength, self->nonce))
            {
                DOFPermissionSet_Destroy(self->permission_set);
                self->permission_set = DOFPermissionSet_UnmarshalRequest(stream, unknownAllowed);
                if(self->permission_set)
                {
                }
                return (self->permission_set != NULL);
            }
        }
    }
    return FALSE;
}

uint32 DOFKeyRequest_MarshallByteLength(DOFKeyRequest self,
                                        boolean       isRequest)
{
    PCR_Assert(self != NULL);

    return 1 + DOFCredentials_MarshalSize(self->credentials, isRequest) +
           self->nonceLength + DOFPermissionSet_MarshalRequestSize(self->permission_set);
}
