/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/error.h>
#include <internal/marshal.h>
#include <internal/security_common.h>

#define DEFAULT_ATTRIBUTE_CODE      2

extern PCRMutex ref_count_mutex;

OALModeOfOperation OALModeOfOperation_Create(uint16 appId,
                                             uint8  length)
{
    OALModeOfOperation ret = DOFMem_Alloc( sizeof(OALModeOfOperation_t) + (length * sizeof(uint8)));

    if(ret != NULL)
    {
        ret->refCount = 1;
        ret->appId = appId;
        ret->attributeCode = DEFAULT_ATTRIBUTE_CODE;
        ret->valueLength = length;
        ret->options = ((uint8*)ret + sizeof(OALModeOfOperation_t));
    }
    return ret;
}

OALModeOfOperation OALModeOfOperation_Retain(OALModeOfOperation modeOfOperation)
{
    PCR_Assert(modeOfOperation != NULL);

    PCRMutex_Lock(ref_count_mutex);
    ++(modeOfOperation->refCount);
    PCRMutex_Unlock(ref_count_mutex);
    return modeOfOperation;
}

void OALModeOfOperation_Release(OALModeOfOperation modeOfOperation)
{
    boolean doFree;

    PCR_Assert(modeOfOperation != NULL);

    PCRMutex_Lock(ref_count_mutex);
    doFree = (--(modeOfOperation->refCount) == 0);
    PCRMutex_Unlock(ref_count_mutex);
    if(doFree)
    {
        DOFMem_Free(modeOfOperation);
    }
}

uint32 OALModeOfOperation_MarshalSize(OALModeOfOperation modeOfOperation)
{
    PCR_Assert(modeOfOperation != NULL);

    return 4 + modeOfOperation->valueLength;
}

boolean OALModeOfOperation_Marshal(OALModeOfOperation self,
                                   OALStream          stream)
{
    PCR_Assert(self != NULL);
    PCR_Assert(stream != NULL);

    if(OALStream_Write8(stream, self->attributeCode))
    {
        if(OALStream_Write16(stream, self->appId))
        {
            if(OALStream_Write8(stream, self->valueLength))
            {
                return OALStream_Write(stream, self->valueLength, self->options);
            }
        }
    }
    return FALSE;
}

OALModeOfOperation OALModeOfOperation_Unmarshal(OALStream stream)
{
    uint8 attributeCode;

    if(OALStream_Read8(stream, &attributeCode))
    {
        uint16 appId;

        PCR_Assert(attributeCode == 2); 
        if(OALStream_Read16(stream, &appId))
        {
            uint8 length;

            if(OALStream_Read8(stream, &length))
            {
                OALModeOfOperation ret = OALModeOfOperation_Create(appId, length);

                if(ret != NULL)
                {
                    if(length > 0)
                    {
                        if(OALStream_Read(stream, length, ret->options))
                        {
                            ret->attributeCode = attributeCode;
                            return ret;
                        }
                    }
                    else
                    {
                        return ret;
                    }
                }
                DOFMem_Free(ret);
            }
        }
    }
    return NULL;
}
