/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/security_common.h>

extern PCRMutex ref_count_mutex;

OALSessionInitiatorBlock OALSessionInitiatorBlock_Create(uint8               duration,
                                                         DOFSecurityMode     modeOfOperation,
                                                         DOFKeyRequest       keyRequest)
{
    OALSessionInitiatorBlock ret = DOFMem_Alloc(sizeof(OALSessionInitiatorBlock_t));

    PCR_Assert(modeOfOperation);
    PCR_Assert(keyRequest);

    if(ret != NULL)
    {
        ret->keyRequest = DOFKeyRequest_Retain(keyRequest);
        if(ret->keyRequest != NULL)
        {
            ret->modeOfOperation = DOFSecurityMode_Copy(modeOfOperation);
            if(ret->modeOfOperation != NULL)
            {
                ret->refCount = 1;
                ret->duration = duration;
                return ret;
            }
            DOFKeyRequest_Destroy(ret->keyRequest);
        }
        DOFMem_Free(ret);
    }
    return NULL;
}

OALSessionInitiatorBlock OALSessionInitiatorBlock_Retain(OALSessionInitiatorBlock initBlock)
{
    PCR_Assert(initBlock != NULL);

    PCRMutex_Lock(ref_count_mutex);
    ++(initBlock->refCount);
    PCRMutex_Unlock(ref_count_mutex);
    return initBlock;
}

void OALSessionInitiatorBlock_Release(OALSessionInitiatorBlock initBlock)
{
    if (initBlock != NULL)
    {
        boolean doFree;
        PCRMutex_Lock(ref_count_mutex);
        doFree = (--(initBlock->refCount) == 0);
        PCRMutex_Unlock(ref_count_mutex);
        if(doFree)
        {
            DOFSecurityMode_Destroy(initBlock->modeOfOperation);
            DOFKeyRequest_Destroy(initBlock->keyRequest);
            DOFMem_Free(initBlock);
        }
    }
}

OALSessionInitiatorBlock OALSessionInitiatorBlock_Unmarshal(OALStream   stream,
                                                            DOFObjectIDDomain domain,
                                                            boolean     isRequest,
                                                            boolean     unknownAllowed)
{
    uint8 duration;

    if(OALStream_Read8(stream, &duration))
    {
        DOFSecurityMode modeOfOperation = DOFSecurityMode_Create_UnmarshalFromStream(stream);
        if(modeOfOperation != NULL)
        {
            DOFKeyRequest keyRequest = DOFKeyRequest_Create( NULL, 0, FALSE, FALSE, NULL, NULL );

            if( keyRequest )
            {
                DOFPermissionSet_Destroy(keyRequest->permission_set);
                PCRMem_Set(keyRequest, 0, sizeof(DOFKeyRequest_t));
                keyRequest->refCount = 1; /* sets it so if unmarshal fails before it sets it the destroy will work correctly */
                if(DOFKeyRequest_Unmarshal(stream, keyRequest, domain, isRequest, unknownAllowed))
                {
                    OALSessionInitiatorBlock ret = OALSessionInitiatorBlock_Create(duration, modeOfOperation, keyRequest);
                    DOFSecurityMode_Destroy( modeOfOperation );
                    DOFKeyRequest_Destroy( keyRequest );
                    return ret;
                }
                DOFKeyRequest_Destroy(keyRequest);
            }
            DOFSecurityMode_Destroy( modeOfOperation );
        }
    }
    return NULL;
}

uint32 OALSessionInitiatorBlock_MarshalSize(OALSessionInitiatorBlock self,
                                            boolean                  isRequest)
{
    PCR_Assert(self != NULL);

    return 1 + DOFSecurityMode_GetMarshalSize( self->modeOfOperation ) +
           DOFKeyRequest_MarshallByteLength(self->keyRequest, isRequest);
}

boolean OALSessionInitiatorBlock_Marshal(OALSessionInitiatorBlock self,
                                         OALStream                stream,
                                         boolean                  isRequest)
{
    PCR_Assert(self != NULL);
    PCR_Assert(stream != NULL);

    if(OALStream_Write8(stream, self->duration))
    {
        if(DOFSecurityMode_MarshalToStream(self->modeOfOperation, stream))
        {
            return DOFKeyRequest_Marshal(stream, self->keyRequest, isRequest);
        }
    }
    return FALSE;
}
