/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/bridge.h>
#include <internal/permissionsetinternal.h>
#include <internal/security_membrane.h>
#include <internal/securityscope.h>

extern PCRMutex ref_count_mutex;

void DOFSecurityMembrane_Ctor(DOFSecurityMembrane        self,
                              DOFSecurityMembrane_VTable vtable)
{
    PCR_Assert(self != NULL);
    self->refCount = 1;
    self->vtable = vtable;
}

DOFSecurityMembrane DOFSecurityMembrane_Copy(DOFSecurityMembrane self)
{
    if(self != NULL)
    {
        PCRMutex_Lock(ref_count_mutex);
        ++(self->refCount);
        PCRMutex_Unlock(ref_count_mutex);
    }
    return self;
}

void DOFSecurityMembrane_Destroy(DOFSecurityMembrane self)
{
    if(self != NULL)
    {
        boolean gone;

        PCRMutex_Lock(ref_count_mutex);
        --(self->refCount);
        gone = (self->refCount == 0);
        PCRMutex_Unlock(ref_count_mutex);
        if ( gone )
        {
            PCR_Assert(self->vtable != NULL);
            PCR_Assert(self->vtable->dtor != NULL);

            self->vtable->dtor(self);
        }
    }
}

DOFSecurityScope DOFSecurityMembrane_ValidatePermission(DOFObjectIDDomain domainId,
                                                        DOFPermissionSet  permSet,
                                                        DOFPermission     perm)
{
    OALPermissionScope permRet;

    if ( permSet )
    {
        permRet = DOFPermissionSet_Validate(permSet, perm);
    }
    else
    {
        /*
         * A NULL permission set only occurs in the case of an outbound permission set of a secure connector.
         * Effectively, this turns off permission checking in the connector, so just return all scopes.
         */
        static const uint32 flags[] = { DOFSECURITYSCOPEFLAG_ALL_SCOPES };

        return DOFSecurityScope_Create(domainId, 0, NULL, 1, flags);
    }

    if(permRet != NULL)
    {
        DOFSecurityScope ret = DOFSecurityScope_Create(domainId, OALPermissionScope_GetIdCount(permRet),
                                                       OALPermissionScope_GetIds(permRet), 0, NULL);

        OALPermissionScope_Destroy(permRet);
        return ret;
    }
    return NULL;
}

DOFSecurityScope DOFSecurityMembrane_ValidateInboundPermissions(DOFSecurityMembrane  self,
                                                                DOFPermission        explicitPermission,
                                                                DOFPermission        logicalPermission,
                                                                OALCipherKeyState    keyState)
{
    PCR_Assert(self != NULL);
    PCR_Assert(self->vtable != NULL);
    PCR_Assert(self->vtable->validateInboundPerms != NULL);
    PCR_Assert(explicitPermission != NULL);
    PCR_Assert(logicalPermission != NULL);

    return self->vtable->validateInboundPerms(self, explicitPermission, logicalPermission, keyState);
}

DOFSecurityScope DOFSecurityMembrane_ValidateOutboundPermissions(DOFSecurityMembrane self,
                                                                 DOFPermission       explicitPermission,
                                                                 DOFPermission       logicalPermission)
{
    PCR_Assert(self != NULL);
    PCR_Assert(self->vtable != NULL);
    PCR_Assert(self->vtable->validateOutboundPerms != NULL);
    PCR_Assert(explicitPermission != NULL);
    PCR_Assert(logicalPermission != NULL);

    return self->vtable->validateOutboundPerms(self, explicitPermission, logicalPermission);
}

DOFSecurityScope DOFSecurityMembrane_GetInboundScope(DOFSecurityMembrane self)
{
    PCR_Assert(self != NULL);
    PCR_Assert(self->vtable != NULL);
    PCR_Assert(self->vtable->getInboundScope != NULL);

    return self->vtable->getInboundScope(self);
}

DOFSecurityScope DOFSecurityMembrane_GetOutboundScope(DOFSecurityMembrane self)
{
    PCR_Assert(self != NULL);
    PCR_Assert(self->vtable != NULL);
    PCR_Assert(self->vtable->getOutboundScope != NULL);

    return self->vtable->getOutboundScope(self);
}

DOFObjectIDDomain DOFSecurityMembrane_GetDomainID(DOFSecurityMembrane self)
{
    PCR_Assert(self != NULL);
    PCR_Assert(self->vtable != NULL);
    PCR_Assert(self->vtable->getDomainId != NULL);

    return self->vtable->getDomainId(self);
}

DOFSecurityMembraneExtendReturn_t DOFSecurityMembrane_ExtendInboundPermissions(DOFSecurityMembrane        self,
                                                                               DOFPermissionSet           extendPerms,
                                                                               ExtendMembraneCompleteCB   callback,
                                                                               void                     * context)
{
    PCR_Assert(self != NULL);
    PCR_Assert(self->vtable != NULL);
    PCR_Assert(self->vtable->extendInboundPerms != NULL);

    return self->vtable->extendInboundPerms(self, extendPerms, callback, context);
}

DOFSecurityMembraneExtendReturn_t DOFSecurityMembrane_ExtendOutboundPermissions(DOFSecurityMembrane        self,
                                                                                DOFPermissionSet           extendPerms,
                                                                                ExtendMembraneCompleteCB   callback,
                                                                                void                     * context)
{
    PCR_Assert(self != NULL);
    PCR_Assert(self->vtable != NULL);
    PCR_Assert(self->vtable->extendOutboundPerms != NULL);

    return self->vtable->extendOutboundPerms(self, extendPerms, callback, context);
}

DOFPermissionSet DOFSecurityMembrane_GetInboundPermissions(DOFSecurityMembrane self)
{
    PCR_Assert(self != NULL);
    PCR_Assert(self->vtable != NULL);
    PCR_Assert(self->vtable->getInboundPerms != NULL);

    return self->vtable->getInboundPerms(self);
}

DOFPermissionSet DOFSecurityMembrane_GetOutboundPermissions(DOFSecurityMembrane self)
{
    PCR_Assert(self != NULL);
    PCR_Assert(self->vtable != NULL);
    PCR_Assert(self->vtable->getOutboundPerms != NULL);

    return self->vtable->getOutboundPerms(self);
}
