/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/securitymode.h>
#include <internal/ccmmode_v1.h>
#include <internal/cipher.h>
#include <internal/marshal.h>
#include <internal/securitymode.h>
#include <internal/stream.h>


/* @addtogroup DOFSecurityModeCCM
@{
**/

#define DOF_CCM_MIN_MAC_SIZE    4
#define DOF_CCM_MAX_MAC_SIZE    16

typedef struct DOFSecurityMode_t
{
    uint32                refCount;
    DOFCipherAlgorithm    algorithm;
    DOFCipherStrength     strength;
    uint8                 flags;
    uint8                 macAuthFlags;
    boolean               encrypt;
} DOFSecurityMode_t;

static boolean setMACSize( const DOFSecurityMode self, uint8 macSize );


/*
**  "Public" Type and Method Declarations (though not exposed in the API)
*/

extern PCRMutex ref_count_mutex;


DOFSecurityMode DOFSecurityModeCCM_Create( DOFCipherAlgorithm algorithm, DOFCipherStrength strength, uint8 macSize, boolean isEncrypt )
{
    DOFSecurityMode ret = (DOFSecurityMode)DOFMem_Alloc( sizeof( DOFSecurityMode_t ) );

    if (ret)
    {
        switch ( algorithm )
        {
            case DOFCIPHERALGORITHM_AES:
                if ( ( ( macSize & 1 ) == 0 ) && 
                     (strength == DOFCIPHERSTRENGTH_128 || strength == DOFCIPHERSTRENGTH_192 || strength == DOFCIPHERSTRENGTH_256) )
                {
                    ret->algorithm = algorithm;
                    ret->strength = strength;
                    ret->encrypt = !!isEncrypt;
                    if ( setMACSize( ret, macSize ) )
                    {
                        ret->refCount = 1;
                        return ret;
                    }
                }
                break;
            default:
                break;
        }
        DOFLog_Message( DOFLOGLEVEL_ERROR, "DOFSecurityModeCCM: Failed to create security mode.");
        DOFMem_Free(ret);
    }
    else
        DOFLog_Message( DOFLOGLEVEL_ERROR, "DOFSecurityModeCCM: Failed to allocate security mode.");

    return NULL;
}


/** @internal
    Create a new @ref DOFSecurityMode from the marshaled contents of a buffer.

    @param  length  [IN] The total length of the buffer; [OUT] the length of the data unmarshaled.
    @param  buffer  The buffer in which the DOFSecurityMode is marshaled.

    @return The new @ref DOFSecurityMode created, or NULL if insufficient
    resources are available to create it or there were insufficient data bytes
    in the buffer to correctly unmarshal the specified @ref DOFType.
**/
DOFSecurityMode DOFSecurityMode_Create_Unmarshal( uint32 * length, const uint8 *buffer )
{
    DOFSecurityMode ret = (DOFSecurityMode)DOFMem_Alloc( sizeof( DOFSecurityMode_t ) );

    if (ret)
    {
        uint32  used     = 0;
        uint32  inLength = *length;

        PCRMem_Set(ret, 0, sizeof( DOFSecurityMode_t ));
        if ( inLength < 7 )
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFSecurityMode: Failed to unmarshal security mode size.");
            used = 1;                    /* Transport-provided security has no mode descriptor. inLength should be 1. */
        }
        else
        {
            if (inLength >= 7 &&         /* 6 bytes of descriptor. */
                buffer[used++] == 2 &&   /* Attribute code 2 is security mode of operation. */
                buffer[used++] == (CCMMODE_V1_APPID >> 8) &&
                buffer[used++] == (CCMMODE_V1_APPID & 0xFF) &&
                buffer[used++] >= 3)
            {
                uint8 i = 0;
                uint8 cCnt = buffer[used++];
                ret->flags = buffer[used+cCnt];

                if ( used+cCnt < inLength )/* We have a valid CCM mode of operation descriptor. */
                {
                    *length = used+cCnt+1;

                    for( ; i<cCnt; ++i, ++used )
                    {
                        /* C OAL only supports 256-bit strength AES. */
                        if(buffer[used] == DOFCIPHERSTRENGTH_256 )
                        {
                            ret->encrypt = ((ret->flags & 0x80) != 0);
                            ret->algorithm = DOFCIPHERALGORITHM_AES;
                            ret->strength = DOFCIPHERSTRENGTH_256;
                            ret->refCount = 1;
                            if ( setMACSize( ret, ((ret->flags & 0x07)*2)+2 ) )
                            {
                                return ret;
                            }
                        }
                    }
                }
            }
            /* Here for bad descriptor or some new security mode of operation descriptor to be defined in the future*/
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFSecurityMode: Unsupported security mode descriptor.");
        }
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "DOFSecurityMode: Failed to create security mode.");
        DOFMem_Free( ret );
    }
    else
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "DOFSecurityMode: Failed to create security mode.");
    }

    return NULL;
}



/** @internal
    Unmarshal a @ref DOFSecurityMode from an @ref OALStream.

    @param  stream  The OALStream in which the DOFSecurityMode is marshaled.

    @return The new @ref DOFSecurityMode created, or NULL if insufficient
    resources are available to create it or there were insufficient data bytes
    in the stream to correctly unmarshal the specified @ref DOFType.
**/
DOFSecurityMode DOFSecurityMode_Create_UnmarshalFromStream( OALStream stream )
{
    uint32          marshalSize = OALStream_GetDataSize( stream );
    DOFSecurityMode smc         = DOFSecurityMode_Create_Unmarshal( &marshalSize, OALStream_GetReadPtr( stream ) );

    if (smc)
        OALStream_AdjustReadOffset( stream, (int32) marshalSize & 0x7FFFFFFFUL );
    else
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFSecurityMode: Failed to unmarshal security mode.");

    return smc;
}


void DOFSecurityMode_Destroy( DOFSecurityMode self )
{
    if (self)
    {
        PCRMutex_Lock( ref_count_mutex );
        --(self->refCount);
        PCRMutex_Unlock( ref_count_mutex );
        if (!self->refCount)
            DOFMem_Free( self );
    }
}

DOFSecurityModeType DOFSecurityMode_GetModeType(DOFSecurityMode self)
{
    PCR_Assert(self);

    return DOFSECURITYMODETYPE_CCM;
}


DOFSecurityMode DOFSecurityMode_Copy( const DOFSecurityMode self )
{
    if (self)
    {
        PCRMutex_Lock( ref_count_mutex );
        if (self->refCount == 0xFFFFFFFF)
        {
            PCRMutex_Unlock( ref_count_mutex );
            return NULL;
        }
        ++(self->refCount);
        PCRMutex_Unlock( ref_count_mutex );
    }

    return self;
}


uint32 DOFSecurityMode_GetMarshalSize( const DOFSecurityMode self )
{
    PCR_Assert( self );
    return 7;    /* We only have CCM security modes to describe, so far. They all marshal as 7 bytes. */
}

uint32 DOFSecurityMode_Marshal( const DOFSecurityMode self, uint32 length, uint8 *buffer )
{
    PCR_Assert( self );
    {
        uint32 marshalSize = DOFSecurityMode_GetMarshalSize( self );

        if (!buffer || length < marshalSize)
            return 0;
        PCR_Assert(self->flags > 0);
        {
            *buffer++ = 2;
            *buffer++ = CCMMODE_V1_APPID >> 8;
            *buffer++ = CCMMODE_V1_APPID & 0xFF;
            *buffer++ = 3;
            *buffer++ = 1;
            *buffer++ = (uint8)self->strength;
            *buffer = self->flags;
        }

        return marshalSize;
    }
}

uint8 DOFSecurityModeCCM_MarshalData(const DOFSecurityMode self, uint8 * buffer)
{
    *buffer++ = 1;
    *buffer++ = (uint8)self->strength;
    *buffer = self->flags;
    return 3;
}


/** @internal
    Marshal a @ref DOFSecurityMode into an @ref OALStream.

    @param  self    The DOFSecurityMode to be marshaled.
    @param  stream  The OALStream in which to marshal the DOFSecurityMode.

    @return TRUE if the marshal was successful; FALSE otherwise.
**/
boolean DOFSecurityMode_MarshalToStream( const DOFSecurityMode self, OALStream stream )
{
    uint32 marshalSize = DOFSecurityMode_Marshal( self, OALStream_GetFreeSize( stream ),
                                                  OALStream_GetWritePtr( stream ) );

    if ( marshalSize )
        OALStream_AdjustWriteOffset( stream, (int32) marshalSize & 0x7FFFFFFFUL );

    return( marshalSize > 0 );
}

/** @internal


   @param self   The first DOFSecurityMode to compare.
   @param config The second DOFSecurityMode to compare.

   @return boolean  Return True if the two specified DOFSecurityMode are equal.  If both are NULL
    return TRUE, if only one of the two are NULL return FALSE.
**/
boolean DOFSecurityMode_IsEqual( const DOFSecurityMode self, const DOFSecurityMode config )
{

    /* If both are NULL then return TRUE*/
    if (!self && !config)
        return TRUE;

    /* If one or the other is NULL then return FALSE */
    if (!self || !config)
        return FALSE;

    return ( ( self->algorithm == config->algorithm ) &&
             ( self->strength == config->strength ) &&
             ( self->encrypt == config->encrypt ) &&
             ( self->macAuthFlags == config->macAuthFlags ) );
}


DOFCipherAlgorithm DOFSecurityModeCCM_GetAlgorithm( const DOFSecurityMode self )
{
    PCR_Assert( self );

    return self->algorithm;
}


DOFCipherStrength DOFSecurityModeCCM_GetStrength( const DOFSecurityMode self )
{
    PCR_Assert( self );

    return self->strength;
}


uint8 DOFSecurityModeCCM_GetMACSize( const DOFSecurityMode self )
{
    uint8 macLengthDescriptor;

    PCR_Assert( self );
    macLengthDescriptor = (self->flags & 0x07);
    PCR_Assert( macLengthDescriptor && macLengthDescriptor <= 7 );
    return (macLengthDescriptor * 2) + 2;
}

boolean setMACSize( const DOFSecurityMode self, uint8 macSize )
{
    PCR_Assert( self );
    if ( (macSize >= DOF_CCM_MIN_MAC_SIZE) && (macSize <= DOF_CCM_MAX_MAC_SIZE) )
    {
        self->macAuthFlags = OALCipher_ComputeMacAuthFlags( macSize );
        self->flags = ((macSize-2)/2) | (((macSize-2)/2) <<3) | ((self->encrypt)?0xC0:0x00);
        return TRUE;
    }
    return FALSE;
}


uint8 DOFSecurityModeCCM_GetMacAuthFlags( const DOFSecurityMode self )
{
    PCR_Assert( self );

    return self->macAuthFlags;
}


boolean DOFSecurityModeCCM_IsEncrypt( const DOFSecurityMode self )
{
    PCR_Assert( self );

    return self->encrypt;
}


/**
@}
**/
