/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/objectid.h>
#include <dof/permissionset.h>
#include <dof/securityscope.h>
#include <dof/error.h>

#include <internal/marshal.h>
#include <internal/objectidinternal.h>
#include <internal/permissionsetinternal.h>
#include <internal/permissionscope.h>
#include <internal/securityscope.h>

#include "securityscope_structs.h"

/* @addtogroup DOFSecurityScope
@{
**/

/*
**  "Public" Type and Method Declarations (though not exposed in the API)
*/

extern PCRMutex ref_count_mutex;

DOFSecurityScope DOFSecurityScope_Create(DOFObjectIDDomain domain, uint16 idCount, const uint32 * ids, uint16 flagCount, const uint32 * flags)
{
    DOFSecurityScope ret = (DOFSecurityScope) DOFMem_Alloc(sizeof(DOFSecurityScope_t));

    if(ret != NULL)
    {
        ret->domain = DOFObjectIDDomain_Copy(domain);
        if(ret->domain != NULL)
        {
            ret->permissionScope = OALPermissionScope_Create( idCount + flagCount );
            if (ret->permissionScope)
            {
                ret->refCount = 1;

                if( ids || flags)
                {
                    OALPermissionScope_SetIDs(ret->permissionScope, idCount, ids, flagCount, flags, &(ret->flagCount));
                }

                return ret;
            }
            DOFObjectIDDomain_Destroy( ret->domain );
        }
        DOFMem_Free(ret);
        ret = NULL;
    }
    return ret;
}

DOFSecurityScope DOFSecurityScope_Copy( const DOFSecurityScope self )
{
    if(self != NULL)
    {
        PCRMutex_Lock(ref_count_mutex);
        ++(self->refCount);
        PCRMutex_Unlock(ref_count_mutex);
    }
    return self;
}

void DOFSecurityScope_Destroy( DOFSecurityScope self )
{
    if(self != NULL)
    {
        boolean okfree = FALSE;
        PCRMutex_Lock(ref_count_mutex);
        if ( --self->refCount == 0 )
            okfree = TRUE;
        PCRMutex_Unlock(ref_count_mutex);
        if ( okfree )
        {
            DOFObjectIDDomain_Destroy(self->domain);
            OALPermissionScope_Destroy(self->permissionScope);
            DOFMem_Free(self);
        }
    }
}

DOFObjectIDDomain DOFSecurityScope_GetDomainID( const DOFSecurityScope self )
{
    PCR_Assert(self != NULL);

    return self->domain;
}

int32 DOFSecurityScope_FindBestIndex(DOFSecurityScope source,
                                     DOFSecurityScope dest)
{
    uint32 const * sourceIds;
    uint32 const * destIds;

    uint8 i, j;

    PCR_Assert(source&&dest);

    PCR_Assert(source->domain == dest->domain);

    if(source==dest) return 0;

    OALPermissionScope_Lock();
    sourceIds = OALPermissionScope_GetIds(source->permissionScope);
    destIds = OALPermissionScope_GetIds(dest->permissionScope);
    for(i=0; i < OALPermissionScope_GetIdCount(source->permissionScope); ++i) {
        for(j=0; j < OALPermissionScope_GetIdCount(dest->permissionScope); ++j) {
            if(sourceIds[i] == destIds[j])
            {
                OALPermissionScope_Unlock();
                return i;
            }
        }
    }
    OALPermissionScope_Unlock();

    return -1;
}

boolean DOFSecurityScope_IsCompatible( const DOFSecurityScope self, const DOFSecurityScope scope )
{
    PCR_Assert(self != NULL);
    PCR_Assert(scope != NULL);

    if(DOFObjectID_Compare(DOFObjectIDDomain_GetObjectID( self->domain ), DOFObjectIDDomain_GetObjectID( scope->domain )) == 0)
    {
        boolean retVal;

        OALPermissionScope_Lock();
        retVal = OALPermissionScope_IsCompatible(self->permissionScope, scope->permissionScope);
        OALPermissionScope_Unlock();
        return retVal;
    }
    return FALSE;
}

int32 DOFSecurityScope_Compare( const DOFSecurityScope self, const DOFSecurityScope scope )
{
    /* NULL or same is equal */
    if(self == scope)
        return 0;
        
    if(self && scope)
    {
        int32 ret = DOFObjectID_Compare(DOFObjectIDDomain_GetObjectID( self->domain ), DOFObjectIDDomain_GetObjectID( scope->domain ));
        if(ret == 0)
        {
            OALPermissionScope_Lock();
            ret = OALPermissionScope_Compare(self->permissionScope, scope->permissionScope);
            OALPermissionScope_Unlock();
        }
        return ret;
    }
    else if( scope )
        return 1;
    else
        return -1;
    return 0;
}

DOFSecurityScope DOFSecurityScope_Unmarshal(OALStream   stream,
                                            DOFObjectIDDomain domain)
{
    uint16 count;

    PCR_Assert(stream != NULL);
    PCR_Assert(domain != NULL);
    if(OALMarshal_StreamUncompress2(stream, &count) == DOFERROR_OK)
    {
        DOFSecurityScope ret = DOFSecurityScope_Create( domain, count, NULL, 0, NULL );
        if(ret!= NULL)
        {
            uint32 i;
            uint32 next = 0;

            for(i = 0; i < count; ++i)
            {
                uint32 temp;

                if(OALMarshal_StreamUncompress4(stream, &temp) != DOFERROR_OK)
                {
                    DOFSecurityScope_Destroy(ret);
                    return NULL;
                }

                OALPermissionScope_SetID(ret->permissionScope, &next, temp);
            }
            return ret;
        }
    }
    return NULL;
}

boolean DOFSecurityScope_Marshal(DOFSecurityScope securityScope,
                                 OALStream        stream)
{
    PCR_Assert(securityScope != NULL);
    PCR_Assert(stream != NULL);

    OALPermissionScope_Lock();
    if(OALMarshal_StreamCompress2(stream, OALPermissionScope_GetIdCount(securityScope->permissionScope)) == DOFERROR_OK)
    {
        uint32 const * ids = OALPermissionScope_GetIds(securityScope->permissionScope);

        uint16 i;

        for(i = 0; i < OALPermissionScope_GetIdCount(securityScope->permissionScope); ++i)
        {
            if(OALMarshal_StreamCompress4(stream, ids[i]) != DOFERROR_OK)
            {
                OALPermissionScope_Unlock();
                return FALSE;
            }
        }
        OALPermissionScope_Unlock();
        return TRUE;
    }
    OALPermissionScope_Unlock();
    return FALSE;
}

uint32 DOFSecurityScope_GetMarshalSize(DOFSecurityScope self)
{
    uint32 retVal;

    uint32 const * ids;

    uint16 i;

    PCR_Assert(self != NULL);
    OALPermissionScope_Lock();
    ids = OALPermissionScope_GetIds(self->permissionScope);
    retVal = OALMarshal_GetCompressedValueSize(2, OALPermissionScope_GetIdCount(self->permissionScope));
    for(i = 0; i < OALPermissionScope_GetIdCount(self->permissionScope); ++i)
    {
        retVal += OALMarshal_GetCompressedValueSize(4, ids[i]);
    }
    OALPermissionScope_Unlock();
    return retVal;
}

OALPermissionScope DOFSecurityScope_GetOALPermissionScope( DOFSecurityScope scope )
{
    if( scope )
        return scope->permissionScope;
    return NULL;
}

boolean DOFSecurityScope_Intersect(DOFSecurityScope   self,
                                   OALPermissionScope permsScope)
{
    uint32 unionedSize = 0;

    uint32 const * ids;
    uint32 const * readableIds;

    uint8 i;
    uint8 j;

    PCR_Assert(self != NULL);
    PCR_Assert(permsScope != NULL);

    OALPermissionScope_Lock();

    ids = OALPermissionScope_GetIds(self->permissionScope);
    /* loop self for Allscopes copy permSopes if found */
    for( i = 0; i < OALPermissionScope_GetIdCount(self->permissionScope); ++i )
    {
        if( ids[i] == DOFSECURITYSCOPEFLAG_ALL_SCOPES )
        {
            OALPermissionScope temp = self->permissionScope;
            self->permissionScope = OALPermissionScope_Copy(permsScope);
            if( self->permissionScope )
            {
                OALPermissionScope_Destroy( temp );
                OALPermissionScope_Unlock();
                return ( OALPermissionScope_GetIdCount( self->permissionScope ) > 0 );
            }
            self->permissionScope = temp;
            OALPermissionScope_Unlock();
            return FALSE;
        }
    }
    readableIds = OALPermissionScope_GetIds(permsScope);
    /* loop permsScope for Allscopes */
    for ( i = 0; i < OALPermissionScope_GetIdCount(permsScope); ++i )
    {
        if ( readableIds[i] == DOFSECURITYSCOPEFLAG_ALL_SCOPES )
        {
            OALPermissionScope_Unlock();
            return ( OALPermissionScope_GetIdCount( self->permissionScope ) > 0 );
        }
    }

    /* loop all self and compare with permScope */
    for(i = 0; i < OALPermissionScope_GetIdCount(self->permissionScope); ++i)
    {
        for(j = 0; j < OALPermissionScope_GetIdCount(permsScope); ++j)
        {
            if( ids[i] == readableIds[j] )
            {
                uint32 offset = unionedSize++;

                PCR_Assert(offset < OALPermissionScope_GetIdCount(self->permissionScope));
                OALPermissionScope_SetID(self->permissionScope, &offset, ids[i]);
                j = OALPermissionScope_GetIdCount(permsScope);                    /* done with inner loop */
            }
        }
        if(OALPermissionScope_GetIdCount(self->permissionScope) == unionedSize)
        {
            i = OALPermissionScope_GetIdCount(self->permissionScope);             /* done with outer loop */
        }
    }
    OALPermissionScope_SetIdCount(self->permissionScope, unionedSize);
    OALPermissionScope_Unlock();
    return (unionedSize != 0);
}


uint32 DOFSecurityScope_GetIDsCount( DOFSecurityScope self )
{
    if( self )
        return OALPermissionScope_GetIdCount(self->permissionScope);
    return 0;
}

const uint32* DOFSecurityScope_GetIDs( DOFSecurityScope self )
{
    if( self )
        return OALPermissionScope_GetIds(self->permissionScope);
    return NULL;
}

uint32 DOFSecurityScope_GetFlagsCount( DOFSecurityScope self )
{
    if( self )
        return self->flagCount;
    return 0;
}

const uint32* DOFSecurityScope_GetFlags( DOFSecurityScope self )
{
    if(self != NULL)
    {
        return &(OALPermissionScope_GetIds(self->permissionScope)[OALPermissionScope_GetIdCount(self->permissionScope)]);
    }
    return NULL;
}

DOFSecurityScope DOFSecurityScope_CreateCompatibleScope( DOFSecurityScope self, DOFSecurityScope other )
{
    DOFSecurityScope ret = (DOFSecurityScope) DOFMem_Alloc(sizeof(DOFSecurityScope_t));

    if(ret != NULL)
    {
        ret->domain = DOFObjectIDDomain_Copy(self->domain);
        if(ret->domain != NULL)
        {
            ret->permissionScope = OALPermissionScope_GetCompatibleScope(self->permissionScope, other->permissionScope);;
            if(ret->permissionScope)
            {
                ret->refCount = 1;
                return ret;
            }
            DOFObjectIDDomain_Destroy( ret->domain );
        }
        DOFMem_Free(ret);
        ret = NULL;
    }
    return ret;
}

/**
@}
**/
