/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/oaltypes.h>
#include <dof/server.h>
#include <dof/address.h>
#include <dof/transport.h>
#include <dof/exception.h>
#include <dof/dof.h>
#include <internal/dofinternal.h>
#include <internal/bridgeinternal.h>
#include <internal/epp_negative_one.h>
#include <internal/protocolstack.h>
#include <internal/exception.h>
#include <internal/connectionmanager.h>
#include <internal/serverinternal.h>
#include <internal/runtimeinternal.h>
#include <internal/connectionless_remotechannel.h>
#include <internal/connectionless_protocolstack.h>
#include <internal/remotechannelmanager.h>
#include <internal/transportmanager.h>
#include <internal/connection_internal.h>
#include <internal/remotechannel.h>
#include <internal/permissionsetinternal.h>

#include "server_common.h"

#if OALSERVER_PRINT_REF_COUNT_DEBUG
#include <stdio.h>
#endif


extern PCRMutex connection_server_mutex;
extern PCRMutex connection_server_count_mutex;

typedef struct OALServerNode_t
{
    OALSListNode    next;
    OALServer       server;
}OALServerNode_t;

typedef struct OALServerNodeCmp_t
{
    DOF dof;
    DOFServerConfig config;
    OALOperationBridge securityConnector;
    DOFServer relatedServer;
    OALServer retServer;
}OALServerNodeCmp_t;

struct serverRestart_t
{
    DOFServerStartCallback callback;
    void *context;
}serverRestart_t;

typedef struct StartCallbackCall_t
{
    OALEnqueueItem_t super;
    DOFServerStartCallback callback;
    void * context;
    DOFServer srv;
    DOFException except;
} StartCallbackCall_t;

static void DOFServer_RemoveAllListener( DOFServer self );

static void serverStartComplete( DOFServerStartCallback self, const DOFServer server, void * context, const DOFException except );
static const struct DOFServerStartCallbackFns_t serverStartFns = { serverStartComplete };
static DOFServerStartCallback_t serverCallback = { &serverStartFns };

static serverContext_t* DOFServerContext_Create( DOFServer srv, DOFServerStartCallback callback, void* context);

void DOFServer_Lock(void)
{
    PCRMutex_Lock( connection_server_mutex );
}

void DOFServer_Unlock(void)
{
    PCRMutex_Unlock( connection_server_mutex );
}

void DOFServerCount_Lock(void)
{
    PCRMutex_Lock( connection_server_count_mutex );
}

void DOFServerCount_Unlock(void)
{
    PCRMutex_Unlock( connection_server_count_mutex );
}

static boolean DOFServer_IsDOFAvailable(DOFServer srv)
{
    if( srv->server && srv->server->serverManager )
        return TRUE;
    return FALSE;
}

DOFServerManager DOFServerManager_Create( DOF dof )
{
    DOFServerManager ret = (DOFServerManager)DOFMem_Alloc( sizeof( DOFServerManager_t ) );

    if( ret )
    {
        ret->oalServers = NULL;
        ret->count = 0;
        ret->dof = dof;
        ret->server_enqueue = OALEnqueue_Create( dof->threadPool );
        if( ret->server_enqueue )
        {
            ret->server_mutex = connection_server_mutex;
            if( ret->server_mutex )
            {
                ret->server_timer = OALTimer_Create( dof->threadPool, TRUE );
                if( ret->server_timer )
                {
                    return ret;
                }
                ret->server_mutex = NULL;
            }
            OALEnqueue_Destroy(ret->server_enqueue);
        }
        DOFMem_Free(ret);
    }
    return NULL;
}

static void StopTheServersFunc(OALSListNode itemx, void *arg)
{
    OALServerNode_t* item = (OALServerNode_t*)itemx;
    OALServer_RealStop(item->server, FALSE, NULL);
    OALServer_RemoveAllListener(item->server);
    OALServer_RemoveManager( item->server );
    DOFMem_Free(itemx);
}

void DOFServerManager_Destroy( DOFServerManager self )
{
    if( self )
    {
        DOFServer_Lock();
        if( self->oalServers )
            OALSList_RemoveAll(&self->oalServers, StopTheServersFunc, NULL);
        DOFServer_Unlock();
        OALEnqueue_Destroy( self->server_enqueue );
        OALTimer_Destroy( self->server_timer );
        DOFMem_Free( self );
    }
}

DOFServer DOFServer_CreateServer(DOFServer self, DOFServerConfig config)
{
    if(self)
    {
        DOFSecurityMembrane membrane;

        DOFServer_Lock();
        if( !DOFServer_IsDOFAvailable( self ) )
        {
            DOFServer_Unlock();
            return NULL;
        }
        membrane = DOFServer_CreateSecurityMembrane(self->server->serverManager->dof, config);
        if(membrane != NULL)
        {
            if( self->server->thisDOFServer != self )
            {
                DOFServer ret = DOFServer_Create(  DOFServer_GetDOF( self ), config, membrane, self );
                DOFServer_Unlock();
                DOFSecurityMembrane_Destroy(membrane);
                return ret;
            }
            else
            {
                DOFServer ret = NULL;
                DOFServer relserver = DOFMem_Alloc(sizeof(DOFServer_t));

                if( relserver )
                {
                    relserver->refCount = 1;
                    relserver->StartCount = 0;
                    relserver->server = OALServer_Retain( self->server );
                    if( relserver->server )
                    {
                        ret = DOFServer_Create(DOFServer_GetDOF(relserver), config, membrane, relserver);
                        if( ret )
                        {
                            DOFServer_Destroy(relserver);
                            DOFServer_Unlock();
                            DOFSecurityMembrane_Destroy(membrane);
                            return ret;
                        }
                        OALServer_Destroy(relserver->server);
                    }
                    DOFMem_Free(relserver);
                }
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFServer: Failed to Create a DOFServer.");
            }
            DOFSecurityMembrane_Destroy(membrane);
        }
        DOFServer_Unlock();
    }
    return NULL;
}

DOFServer DOFServer_Create( DOF dof, DOFServerConfig config, DOFSecurityMembrane securityConnector, DOFServer relatedServer )
{
    if( dof && config )
    {
        DOFServer ret = DOFMem_Alloc(sizeof(DOFServer_t));
        if( ret )
        {
            ret->refCount = 1;
            ret->StartCount = 0;
            ret->server = OALServer_Create( dof, config, securityConnector, relatedServer );
            if( ret->server )
            {
                return ret;
            }
            DOFMem_Free(ret);
        }
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFServer: Failed to allocate memory for a DOFServer.");
    }
    return NULL;
}

static void enqueuedStartCallbackCall( OALThreadPoolItem dnode, boolean shutdown )
{
    StartCallbackCall_t * node = (StartCallbackCall_t*) dnode;

    PCR_Assert( dnode != NULL );
    node->callback->fns->Complete( node->callback, node->srv, node->context, node->except );

    DOFServer_Release( node->srv );
    DOFMem_Free( dnode );
}

static boolean enqueueFailedStart( StartCallbackCall_t* node, DOFServerStartCallback callback, void* context, DOFServer srv, DOFException except )
{
    if(node != NULL)
    {
        DOFServer temp = DOFServer_Retain( srv );
        if( temp )
        {
            OALEnqueueItem_Init((OALEnqueueItem)node, enqueuedStartCallbackCall);
            node->callback = callback;
            node->context = context;
            node->srv = temp;
            node->except = except;

            if(OALEnqueue_Enqueue( srv->server->serverManager->server_enqueue, (OALEnqueueItem)node))
            {
                return TRUE;
            }
            DOFServer_Destroy( temp );
        }
    }
    return FALSE;
}

void removeAllContexts(OALSListNode item, void *user_arg)
{
    serverContext_t* arg = (serverContext_t*)item;
    DOFException except = (DOFException)user_arg;

    OALTimerItem_Remove( arg->timer_item );
    OALTimerItem_Destroy( arg->timer_item, TRUE );

    if( enqueueFailedStart(arg->enqComplete, arg->callback, arg->context, arg->server, except ) )
    {
        return;
    }
    PCR_Assert(FALSE); /* Should not hit here */
}

/* called when a server is being destroyed to remove all the start calls caused by it*/
static void DOFServer_CompleteStop( DOFServer srv )
{
    if( !srv || !srv->StartCount  )
        return;

    if(srv->StartCount > 0)
    {
        srv->server->StartCount -= srv->StartCount;
        srv->StartCount = 0;

        if( !srv->server->StartCount )
            OALServer_RealStop( srv->server, FALSE, NULL );
    }
}

/*
 *  This is to check if this is the DOFServer created in the OALServer
 */
static boolean DOFServer_Internal_CheckIfOALServersDOFServer( DOFServer srv )
{
    if(srv->server && (srv == srv->server->thisDOFServer))
        return TRUE;
    return FALSE;
}

static void DOFServer_Internal_HandleOALServerOwnedDOFServer( DOFServer srv )
{
    DOFServer_RemoveAllListener(srv);
    DOFServer_CompleteStop(srv);

    if( srv->refCount == 1 ) /* Check if last calls bumped refcount if not we can still free */
    {
        uint32 _refCount = srv->server->refCount;
        if ( _refCount == 0 )
        {
            ++srv->server->refCount;
            OALServer_Destroy( srv->server );
            srv->server = NULL;
        }
    }
}

static void DOFServer_Internal_HandleRefCountHitZero( DOFServer srv )
{
    DOFServer_CompleteStop(srv);
    DOFServer_RemoveAllListener( srv );

    if( srv->refCount == 0 ) /* Check if last calls bumped refcount if not we can still free */
    {
        DOF_RemoveServer( DOF_GetDof(), srv );
        OALServer_Destroy( srv->server );
        srv->server = NULL;
        DOFMem_Free( srv );
    }
}

void DOFServer_Release( DOFServer srv )
{
    uint32 _refCount;

    if(srv)
    {
        DOFServer_Lock();
        PCR_Assert( srv->refCount );
        _refCount = --(srv->refCount);

        if( _refCount == 1 &&
            DOFServer_Internal_CheckIfOALServersDOFServer( srv ) )
        {
            DOFServer_Internal_HandleOALServerOwnedDOFServer( srv );
        }
        else if( _refCount == 0 )
        {
            DOFServer_Internal_HandleRefCountHitZero( srv );
        }
        DOFServer_Unlock();
    }
}

DOFServer DOFServer_Retain( DOFServer srv )
{
    if( srv )
    {
        DOFServer_Lock();
        ++(srv->refCount);
        PCR_Assert( srv->refCount );   /* Rollover to 0 is disastrous and is most likely caused by coding error of some kind. */
        DOFServer_Unlock();
        return srv;
    }
    return NULL;
}

void DOFServer_Destroy( DOFServer server )
{
    DOFServer_Release( server );
}

DOF DOFServer_GetDOF( const DOFServer self )
{
    PCR_Assert( self );
    DOFServer_Lock();
    if( DOFServer_IsDOFAvailable(self) )
    {
        DOFServer_Unlock();
        return self->server->serverManager->dof;
    }
    DOFServer_Unlock();
    return NULL;
}

static serverContext_t* DOFServerContext_Create( DOFServer srv, DOFServerStartCallback callback, void* context)
{
    serverContext_t* retVal = (serverContext_t*)DOFMem_Alloc( sizeof( serverContext_t ) );
    if( retVal )
    {
        PCRMem_Set( retVal, 0, sizeof( serverContext_t ) );
        retVal->enqComplete = (StartCallbackCall_t*)DOFMem_Alloc( sizeof( StartCallbackCall_t ) );
        if( retVal->enqComplete )
        {
            retVal->enqNode = (ServerContext_Wrapper_t*)DOFMem_Alloc( sizeof( ServerContext_Wrapper_t ) );
            if( retVal->enqNode )
            {
                retVal->enqNode->item = retVal;
                retVal->callback = callback;
                retVal->context = context;
                retVal->server = srv;
                return retVal;
            }
            DOFMem_Free( retVal->enqComplete );
        }
        DOFMem_Free( retVal );
    }

    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFServer: Failed to create server context");
    return NULL;
}

void DOFServerContext_Destroy( serverContext_t* srvCtx )
{
    if( srvCtx->enqNode )
        DOFMem_Free( srvCtx->enqNode );
    DOFMem_Free(srvCtx->enqComplete);
    DOFMem_Free( srvCtx );
}

static int server_Timeout(OALTimerItem _item, uint32 argc, void *argv[], boolean shutdown)
{
    serverContext_t *item = (serverContext_t*)argv[0];
    if( !shutdown && argc == 1 )
    {
        DOFServer_Lock();
        if( item->server->server->serverContexts )
            OALSList_Remove(&item->server->server->serverContexts, (OALSListNode)item);
        DOFServer_Unlock();

        /* Not enqueued because this is called in timers thread */
        item->callback->fns->Complete( item->callback, item->server, item->context, DOFEXCEPTION_TIMEOUT );
    }

    OALTimerItem_Destroy( _item, FALSE );
    DOFServerContext_Destroy( item );
    return 0;
}

static boolean DOFServer_IsRelatedUseable( DOFServer srv )
{
    if( srv->server->relSrvList ) /* if we created our own Server Listener we have a usable related server */
        return TRUE;
    return FALSE;
}

typedef struct relatedServer_t
{
    DOFServer               srv;
    uint32                  start;
    uint32                  timeout;
    DOFServerStartCallback  callback;
    void*                   context;
    StartCallbackCall_t*    enqComplete;
}relatedServer_t;

static relatedServer_t* DOFServer_CreateRelatedContext( DOFServer srv, DOFServerStartCallback callback, void* context, uint32 timeout )
{
    relatedServer_t* ret = (relatedServer_t*)DOFMem_Alloc( sizeof( relatedServer_t ) );
    if( ret )
    {
        ret->enqComplete = (StartCallbackCall_t*)DOFMem_Alloc( sizeof( StartCallbackCall_t ) );
        if( ret->enqComplete )
        {
            ret->srv = srv;
            ret->callback = callback;
            ret->context = context;
            ret->timeout = timeout;
            ret->start = PCRTime_GetMS();
            return ret;
        }
        DOFMem_Free(ret);
    }

    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFServer: Failed to allocate related server context");
    return NULL;
}

static void DOFServer_Internal_BeginStart( DOFServer srv, uint32 timeout, DOFServerStartCallback callback, void* context, boolean returning )
{
    serverContext_t* ctx;
    PCR_Assert(srv);

    DOFServer_Lock();

    if(!DOFServer_IsDOFAvailable(srv))
    {
        DOFServer_Unlock();
        callback->fns->Complete(callback, srv, context, DOFEXCEPTION_FAILED);
        return;
    }

    if( srv->server->serverState == DOFSERVERSTATUS_STARTED  )
    {
        StartCallbackCall_t * node = (StartCallbackCall_t*) DOFMem_Alloc(sizeof(StartCallbackCall_t));

        if(node != NULL)
        {
            if( !enqueueFailedStart( node, callback, context, srv, NULL ) )
            {
                DOFMem_Free(node);
            }
            ++srv->StartCount;
            ++srv->server->StartCount;
            DOFServer_Unlock();
        }
        else
        {
            DOFServer_Unlock();
            callback->fns->Complete(callback, srv, context, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        }
        return;
    }

    if( !returning && DOFServer_IsRelatedUseable( srv ) )
    {
        relatedServer_t* con = DOFServer_CreateRelatedContext( srv, callback, context, timeout );
        if(!con)
        {
            DOFServer_Unlock();
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFServer: Failed to create related server context in begin start.");
            callback->fns->Complete(callback, srv, context, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
            return;
        }
        DOFServer_Unlock();
        DOFServer_BeginStart( srv->server->relatedServer, timeout, &serverCallback, con );
        return;
    }

    ctx = DOFServerContext_Create( srv, callback, context );

    if( ctx )
    {
        OALSList_Insert( &srv->server->serverContexts, (void*)ctx );
        if((ctx->timer_item = OALTimer_CreateItem(srv->server->serverManager->server_timer, server_Timeout, 1, (void**)&ctx)) != NULL )
        {
            if(OALTimerItem_Insert( ctx->timer_item, timeout, FALSE, FALSE ) == DOFERROR_OK)
            {
                OALChannelMgr_CreateChannel_Server( srv );
                DOFServer_Unlock();
                return;
            }
            OALTimerItem_Destroy( ctx->timer_item, TRUE );
            ctx->timer_item = NULL;
        }
        OALSList_Remove( &srv->server->serverContexts, (void*)ctx );
        DOFServerContext_Destroy(ctx);
    }
    DOFServer_Unlock();
    callback->fns->Complete(callback, srv, context, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
    return;
}

static void serverStartComplete( DOFServerStartCallback self, const DOFServer server, void * context, const DOFException except )
{
    relatedServer_t* rSrv = (relatedServer_t*)context;

    DOFServer_Lock();
    if( !except )
    {
        /* we started this server call start on next one */
        uint32 newTime = PCRTime_GetMS();

        if( rSrv->callback == self ) /* if its ours we have another related server to start */
            ((relatedServer_t*)rSrv->context)->start = newTime;

        newTime -= rSrv->start;
        /* at this point newTime holds how long it has already taken. */
        if (newTime >= rSrv->timeout)
        {
            /* This timed out, enqueue the failure. */
            DOFServer_Stop(server);
            enqueueFailedStart(rSrv->enqComplete, rSrv->callback, rSrv->context, rSrv->srv, DOFEXCEPTION_TIMEOUT);
            rSrv->enqComplete = NULL;/* will be freed in enqueue */
        }
        else
        {
            newTime = rSrv->timeout - newTime;

            if( !DOFServer_AddStateListener( server, rSrv->srv->server->relSrvList ) )
            {
                DOFServer_Stop(server);
                enqueueFailedStart( rSrv->enqComplete, rSrv->callback, rSrv->context, rSrv->srv, DOFEXCEPTION_FAILED );
                rSrv->enqComplete = NULL;/* will be freed in enqueue */
            }
            else
            {
                DOFServer_Internal_BeginStart( rSrv->srv, newTime, rSrv->callback, rSrv->context, TRUE );
            }
        }
    }
    else
    {
        if( server->server->relatedServer && server->server->relatedServer->server->serverState == DOFSERVERSTATUS_STARTED )
        {
            DOFServer_RemoveStateListener( server->server->relatedServer, server->server->relSrvList );
            DOFServer_Stop( server->server->relatedServer ); /* this should stop the server we just started */
        }

        enqueueFailedStart( rSrv->enqComplete, rSrv->callback, rSrv->context, rSrv->srv, except );
        rSrv->enqComplete = NULL; /* will be freed in enqueue */
    }
    DOFServer_Unlock();
    DOFMem_Free( rSrv->enqComplete );
    DOFMem_Free( context );
}

void DOFServer_BeginStart( DOFServer srv, uint32 timeout, DOFServerStartCallback callback, void* context )
{
    DOFServer_Internal_BeginStart( srv, timeout, callback, context, FALSE );
}

typedef struct startContext
{
    PCREvent        event;
    DOFException    except;
}startContext;

static void CompleteStart ( DOFServerStartCallback self, const DOFServer server, void* context, const DOFException except )
{
    startContext* ctx = (startContext*)context;
    PCR_Assert(!ctx->except);
    ctx->except = DOFException_Copy(except);
    PCREvent_Signal(ctx->event);
}

static const struct DOFServerStartCallbackFns_t dsscbsfns = {CompleteStart};
static DOFServerStartCallback_t dsscbs = {&dsscbsfns};

boolean DOFServer_Start( DOFServer srv, uint32 timeout, DOFException *pException )
{
    PCRStatus status;
    startContext sc;
    PCR_Assert(srv);

    DOFServer_Lock();
    if(!DOFServer_IsDOFAvailable(srv))
    {
        if(pException)
            *pException = DOFException_Copy( DOFEXCEPTION_FAILED );
        DOFServer_Unlock();
        return FALSE;
    }

    if( srv->server->serverState == DOFSERVERSTATUS_STARTED )
    {

        ++srv->StartCount;
        ++srv->server->StartCount;
        DOFServer_Unlock();
        if( pException )
            *pException = NULL;
        return TRUE;
    }
    DOFServer_Unlock();

    sc.except = NULL;
    sc.event = PCREvent_Create( FALSE );
    if( !sc.event )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFServer: Failed to create pcr event in server start");
        if( pException )
            *pException = DOFException_Copy(DOFEXCEPTION_FAILED);
        return FALSE;
    }

    DOFServer_BeginStart( srv, timeout, &dsscbs, &sc );
    status = PCREvent_Wait( sc.event, DOF_TIMEOUT_NEVER );

    if( pException )
    {
        if( status != PCRSTATUS_OK )
            *pException = DOFException_Copy( DOFEXCEPTION_FAILED );
        else
            *pException = sc.except;
    }

    PCREvent_Destroy( sc.event );

    return ((status == PCRSTATUS_OK) && !sc.except);
}

/* currently only for DOF_Stop so verify contract before use*/
void DOFServer_Kill( DOFServer server, DOFException except )
{
    PCR_Assert(server);
    DOFServer_Lock();
    if( server->server )
        OALServer_RealStop( server->server, FALSE, except );
    DOFServer_Unlock();
}

void DOFServer_StopInternal( DOFServer srv, DOFException exception )
{
    if( !srv || !srv->StartCount  )
        return;

    if(srv->StartCount > 0)
    {
        --srv->StartCount;
        --srv->server->StartCount;
        if( !srv->server->StartCount )
            OALServer_RealStop( srv->server, FALSE, exception );
    }
}

void DOFServer_Stop( DOFServer srv )
{
    PCR_Assert(srv);
    DOFServer_Lock();
    DOFServer_StopInternal( srv, NULL );
    DOFServer_Unlock();
}

boolean DOFServer_IsStarted( const DOFServer self )
{
    PCR_Assert( self );
    return OALServer_IsStarted( self->server );
}

DOFServerState DOFServer_CreateState( const DOFServer self )
{
    PCR_Assert( self );
    return OALServer_CreateState( self->server, NULL );
}

DOFConnection DOFServer_CreateConnection( DOFServer self, DOFConnectionConfig config )
{
    DOFConnection ret = NULL;
    DOFSecurityMembrane membrane;

    PCR_Assert(self);
    if( config )
    {
        membrane = DOFConnection_CreateSecurityMembrane(self->server->serverManager->dof, config);
        if(membrane != NULL)
        {
            ret = DOFConnection_CreateInternalConnection( DOF_GetDof(), membrane, config, self, NULL, TRUE );
            DOFSecurityMembrane_Destroy(membrane);
        }
    }

    return ret;
}

static boolean findStateCallback(OALSListNode itemx, void *arg)
{
    DOFServerCallbackNode_t* item = (DOFServerCallbackNode_t*)itemx;
    void** args = (void**)arg;
    DOFServer srv = (DOFServer)args[0];
    DOFServerStateListener callback = (DOFServerStateListener)args[1];

    if( srv == item->server && item->callback == callback )
        return TRUE;
    return FALSE;
}

typedef struct StateCallbackCall_t
{
    OALEnqueueItem_t super;
    DOFServer srv;
    DOFServerStateListener callback;
    DOFServerState state;
} StateCallbackCall_t;

static void DOFServer_StateCallbackCall( OALThreadPoolItem dnode, boolean shutdown )
{
    StateCallbackCall_t * node = (StateCallbackCall_t*) dnode;

    if( node->state )
    {
        node->callback->fns->StateChanged( node->callback, node->srv, node->state ); /* normal state listener callback */
        DOFServerState_Destroy( node->state );
    }
    else
    {
        node->callback->fns->Removed( node->callback, node->srv, NULL ); /* removed listener call */
    }
    DOFMem_Free(node);
}

void DOFServer_EnqueueListenerCall(DOFServer self, DOFServerStateListener callback, DOFServerState serverState )
{
    if( self->server->serverManager && self->server->serverManager->server_enqueue )
    {
        StateCallbackCall_t * node = (StateCallbackCall_t*) DOFMem_Alloc(sizeof(StateCallbackCall_t));

        if(node != NULL)
        {
            OALEnqueueItem_Init((OALEnqueueItem) node, DOFServer_StateCallbackCall);
            node->srv = self;
            node->callback = callback;
            node->state = (serverState)?DOFServerState_Copy( serverState ) : NULL;

            if( OALEnqueue_Enqueue( self->server->serverManager->server_enqueue, (OALEnqueueItem) node) )
            {
                return;
            }
        }
        else
        {
            DOFServer_RemoveStateListenerInternal( self, callback );
        }
    }
}

boolean DOFServer_AddStateListener( DOFServer self, DOFServerStateListener callback)
{
    void* args[2];
    args[0] = self;
    args[1] = (DOFServerStateListener)callback;

    if(self && callback)
    {
        DOFServer_Lock();

        if(!DOFServer_IsDOFAvailable(self))
        {
            DOFServer_Unlock();
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFServer: DOF is Destroyed");
            return FALSE;
        }

        if( OALSList_Retrieve( self->server->stateCallback, findStateCallback, args ) == NULL)
        {
            DOFServerCallbackNode_t* scn = DOFMem_Alloc( sizeof( DOFServerCallbackNode_t ) );
            if( scn )
            {
                scn->destroyListener = DOFMem_Alloc( sizeof( DOFServerCallbackNodeContext_t ) );
                if( scn->destroyListener )
                {
                    DOFServerState serverState = DOFServer_CreateState(self);
                    if( serverState )
                    {
                        scn->callback = callback;
                        scn->server = self;
                        scn->next = NULL;

                        OALSList_Insert( &self->server->stateCallback, (OALSListNode)scn );

                        DOFServer_EnqueueListenerCall( self, callback, serverState );
                        DOFServerState_Destroy( serverState );

                        DOFServer_Unlock();
                        return TRUE;
                    }
                    DOFMem_Free( scn->destroyListener );
                }
                DOFMem_Free( scn );
            }
        }
        DOFServer_Unlock();
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFServer: Failed to add server listener");
    }
    return FALSE;
}

static void notifyDestroyListener( OALThreadPoolItem dnode, boolean shutdown )
{
    DOFServerCallbackNodeContext_t * node = (DOFServerCallbackNodeContext_t*) dnode;
    DOFServerStateListener sl;

    PCR_Assert(dnode != NULL);
    sl = (DOFServerStateListener)node->item->callback;
    if( sl && sl->fns->Removed )
        sl->fns->Removed( sl, node->item->server, NULL );
    DOFServer_Destroy(node->item->server);
    DOFMem_Free( node->item );
    DOFMem_Free( node );
}

void removeAllStateCallback(OALSListNode itemx, void* argx)
{
    DOFServerCallbackNode_t* item = (DOFServerCallbackNode_t*)itemx;
    DOFServerCallbackNodeContext_t * node = item->destroyListener;

    if(DOFServer_IsDOFAvailable( item->server ))
    {
        DOFServer temp = DOFServer_Retain(item->server);
        if(temp)
        {
            OALEnqueueItem_Init((OALEnqueueItem) node, notifyDestroyListener);
            node->item = item;

            if( OALEnqueue_Enqueue( ((DOF)DOF_GetDof())->serverManager->server_enqueue, (OALEnqueueItem) node) )
            {
                return;
            }
            PCR_Assert( FALSE ); /* should not hit */
        }
    }
    DOFMem_Free( node );
}

static boolean removeAllMyStateCallback(OALSListNode itemx, void *arg)
{
    DOFServerCallbackNode_t* item = (DOFServerCallbackNode_t*)itemx;
    DOFServer self = (DOFServer)arg;
    if( item->server == self )
        return TRUE;
    return FALSE;
}

void DOFServer_RemoveAllListener( DOFServer self )
{
    DOFServer_Lock();
    if( self->server->stateCallback )
    {
        DOFServerCallbackNode_t* node = NULL;
        do
        {
            node = (DOFServerCallbackNode_t*)OALSList_FindAndRemove( &self->server->stateCallback, NULL, removeAllMyStateCallback, self );
            if( node )
            {
                DOFServerCallbackNodeContext_t * enode = node->destroyListener;
                OALEnqueueItem_Init((OALEnqueueItem) enode, notifyDestroyListener);
                enode->item = node;
                DOFServer_Retain(node->server); /* bump server ref count */

                OALEnqueue_Enqueue( self->server->serverManager->server_enqueue, (OALEnqueueItem) enode);
            }
        }while(node);
    }
    DOFServer_Unlock();
}

void DOFServer_RemoveStateListenerInternal( DOFServer self, DOFServerStateListener callback)
{
    void* args[2];
    DOFServerCallbackNode_t* rem;
    args[0] = self;
    args[1] = (DOFServerStateListener)callback;

    DOFServer_Lock();
    if( DOFServer_IsDOFAvailable(self) )
    {
        rem = (DOFServerCallbackNode_t*)OALSList_FindAndRemove( &self->server->stateCallback, NULL, findStateCallback, args );
        if( rem )
        {
            DOFServerCallbackNodeContext_t * node = rem->destroyListener;
            OALEnqueueItem_Init((OALEnqueueItem) node, notifyDestroyListener);
            node->item = rem;
            DOFServer_Retain(rem->server);
            OALEnqueue_Enqueue( self->server->serverManager->server_enqueue, (OALEnqueueItem) node);
        }
    }
    DOFServer_Unlock();
}

void DOFServer_RemoveStateListener( DOFServer self, DOFServerStateListener callback)
{
    if( self && callback )
    {
        DOFServer_Lock();
        DOFServer_RemoveStateListenerInternal( self, callback );
        DOFServer_Unlock();
    }
}

DOFAddress DOFServer_GetListenAddr(DOFServer server)
{
    return OALServer_GetListenAddr( server->server );
}

DOFServerConfig DOFServer_GetConfig( const DOFServer self )
{
    if( self && self->server )
        return self->server->config;
    return NULL;
}

boolean DOFServer_Restart( DOFServer self, uint32 timeout, DOFException * pException )
{
    boolean ret = DOFServer_Start(self, timeout, pException);
    if( ret )
        DOFServer_Stop(self);
    return ret;
}

static void CompleteRestart ( DOFServerStartCallback self, const DOFServer server, void * contextx, const DOFException except )
{
    struct serverRestart_t* context = (struct serverRestart_t*)contextx;
    if(!except)
        DOFServer_Stop(server);
    context->callback->fns->Complete(context->callback, server, context->context, except);
    DOFMem_Free(context);
}

static const struct DOFServerStartCallbackFns_t restart_fn = {CompleteRestart};
static DOFServerStartCallback_t restart_callback = {&restart_fn};

void DOFServer_BeginRestart( DOFServer self, uint32 timeout, DOFServerStartCallback callback, void *context )
{
    struct serverRestart_t* ctx = DOFMem_Alloc(sizeof(struct serverRestart_t));

    if( ctx )
    {
        ctx->callback = callback;
        ctx->context = context;
        DOFServer_BeginStart( self, timeout, &restart_callback, (void*)ctx );
    }
    else
    {
        callback->fns->Complete(callback, self, context, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
    }
}



uint32 DOFServerManager_GetServerCount( DOF dof )
{
    if( dof )
        return dof->serverManager->count;
    return 0;
}


void DOFServerList_Destroy(DOFServerList self)
{
    if( self )
    {
        while(self->serverCount--)
        {
            DOFServer_Release(self->serverArray[self->serverCount]);
        }

        DOFMem_Free(self->serverArray);
        DOFMem_Free(self);
    }
}

boolean DOFServer_IsConnectionAllowed(DOFServer server)
{
    boolean ret = FALSE;
    if( server )
    {
        DOFServer_Lock();
        ret = (server->server->connectionsCount < server->server->config->MaxConnections);
        DOFServer_Unlock();
    }
    return ret;
}
