/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/oaltypes.h>
#include <dof/server.h>
#include <dof/address.h>
#include <dof/transport.h>
#include <dof/exception.h>
#include <dof/dof.h>
#include <internal/bridgeinternal.h>
#include <internal/dofinternal.h>
#include <internal/epp_negative_one.h>
#include <internal/protocolstack.h>
#include <internal/exception.h>
#include <internal/connectionmanager.h>
#include <internal/serverinternal.h>
#include <internal/runtimeinternal.h>
#include <internal/connectionless_remotechannel.h>
#include <internal/connectionless_protocolstack.h>
#include <internal/remotechannelmanager.h>
#include <internal/transportmanager.h>
#include <internal/connection_internal.h>
#include <internal/remotechannel.h>
#include <internal/permissionsetinternal.h>

#include "server_common.h"

#if OALSERVER_PRINT_REF_COUNT_DEBUG
#include <stdio.h>
#endif

extern PCRMutex connection_server_mutex;

typedef struct OALServerNode_t
{
    OALSListNode    next;
    OALServer       server;
}OALServerNode_t;

typedef struct OALServerNodeCmp_t
{
    DOF dof;
    DOFServerConfig config;
    DOFSecurityMembrane securityMembrane;
    DOFServer relatedServer;
    OALServer retServer;
}OALServerNodeCmp_t;

typedef struct ServerStateListener_t
{
    DOFServerStateListener_t    base;
    OALServer                   server;
}ServerStateListener_t;

static void serverStateChanged( DOFServerStateListener self, const DOFServer server, DOFServerState state );
static void serverRemoved( DOFServerStateListener self, DOFServer server, DOFException except );

static const struct DOFServerStateListenerFns_t serverStateFns = { serverRemoved, serverStateChanged };
static DOFServerStateListener_t serverStateListener = { &serverStateFns };

static  boolean OALServers_TraversalAndFind(OALSListNode itemx, void *argx)
{
    OALServerNode_t* item = (OALServerNode_t*)itemx;
    OALServerNodeCmp_t* arg = (OALServerNodeCmp_t*)argx;

    if( item->server->securityMembrane == arg->securityMembrane )
    {
        if( item->server->relatedServer == arg->relatedServer )
        {
            if( DOFServerConfig_IsEqual( item->server->config, arg->config ) )
            {
                arg->retServer = OALServer_Retain( item->server );
                if( arg->retServer )
                {
                    return TRUE;
                }
            }
        }
    }
    return FALSE;
}

#if OALSERVER_PRINT_REF_COUNT_DEBUG
    OALServer OALServer_Create_old( DOF dof, DOFServerConfig config, DOFSecurityMembrane securityMembrane, DOFServer relatedServer, const char* function, const char* filename, int line)
#else
    OALServer OALServer_Create( DOF dof, DOFServerConfig config, DOFSecurityMembrane securityMembrane, DOFServer relatedServer )
#endif
{
    boolean found;
    OALServerNodeCmp_t cmp;
    cmp.dof = dof;
    cmp.securityMembrane = securityMembrane;
    cmp.relatedServer = relatedServer;
    cmp.config = config;
    cmp.retServer = NULL;

    DOFServer_Lock();
    found = OALSList_Traverse( dof->serverManager->oalServers, OALServers_TraversalAndFind, &cmp);
    DOFServer_Unlock();
    if( found )
    {
        return cmp.retServer;
    }
    else
    {
        OALServer ret = NULL;
        OALServerNode_t* node = NULL;
        if( !( relatedServer && (OALServer_IsDatagram( relatedServer->server ) && (config->type == DOFSERVERTYPE_DATAGRAM))) )
        {
            ret = DOFMem_Alloc(sizeof(OALServer_t));
            if( !ret )
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFServer: Failed to create an Server.");
                return NULL;
            }
            PCRMem_Set(ret,0,sizeof(OALServer_t));
        }
        else
        {
            /* our listener for our related server is stored with us */
            ServerStateListener_t* ssl;
            ret = DOFMem_Alloc( sizeof( OALServer_t ) + sizeof( ServerStateListener_t ) );
            if( !ret )
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFServer: Failed to create an Server.");
                return NULL;
            }
            PCRMem_Set(ret,0, sizeof( OALServer_t ) + sizeof( ServerStateListener_t ) );
            ret->relSrvList = (DOFServerStateListener)(((uint8*)ret) + sizeof(OALServer_t));
            ssl = (ServerStateListener_t*)ret->relSrvList;
            PCRMem_Copy(&ssl->base, &serverStateListener, sizeof( DOFServerStateListener_t ));
            ssl->server = ret;
        }

        ret->cleanupnode = DOFMem_Alloc(sizeof(DOFServerReleaseContext_t));
        if( ret->cleanupnode )
        {
            ret->relatedServer = DOFServer_Retain( relatedServer );
            if(!relatedServer || (relatedServer && ret->relatedServer))
            {
                ret->serverManager = dof->serverManager;
                ret->serverState = DOFSERVERSTATUS_STOPPED;
                ret->config = DOFServerConfig_Copy( config );
                if( ret->config )
                {
                    ret->securityMembrane = DOFSecurityMembrane_Copy( securityMembrane );
                    if( ret->securityMembrane )
                    {
                        ret->refCount = 1;
                        ret->thisDOFServer = DOFMem_Alloc(sizeof(DOFServer_t));
                        if( ret->thisDOFServer )
                        {
                            ret->thisDOFServer->refCount = 1;
                            ret->thisDOFServer->server = ret;
                            ret->thisDOFServer->StartCount = 0;

                            node = DOFMem_Alloc( sizeof(OALServerNode_t) );
                            if( node )
                            {
                                node->server = ret;
                                node->next = NULL;

                                DOFServer_Lock();
                                OALSList_Insert( &dof->serverManager->oalServers, (OALSListNode)node );
                                ++dof->serverManager->count;
                                DOFServer_Unlock();
#if OALSERVER_PRINT_REF_COUNT_DEBUG
                                fprintf( stderr,"OALServer: %p +%d %s:%d (%s)\n", ret, ret->refCount, filename, line, function );
#endif
                                return ret;
                            }
                            DOFMem_Free(ret->thisDOFServer);
                        }
                        DOFSecurityMembrane_Destroy(ret->securityMembrane);
                    }
                    DOFServerConfig_Destroy(ret->config);
                }
                DOFServer_Destroy(ret->relatedServer);
            }
            DOFMem_Free(ret->cleanupnode);
        }
        DOFMem_Free(ret);
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFServer: Failed to create a Server.");
    return NULL;
}

static void destroyOALServer( OALThreadPoolItem dnode, boolean shutdown )
{
    DOFServerReleaseContext_t * node = (DOFServerReleaseContext_t *) dnode;

    DOFServer_Lock();

    PCR_Assert(dnode != NULL);
#if OALSERVER_PRINT_REF_COUNT_DEBUG
    fprintf( stderr, "OALServer Free: %p\n", node->organizer.oalserver );
#endif
    DOFMem_Free( node->organizer.oalserver->thisDOFServer ); /* DOFServer Free */
    /* complete free */
    DOFServerConfig_Destroy( node->organizer.oalserver->config );
    DOFSecurityMembrane_Destroy( node->organizer.oalserver->securityMembrane );

    DOFMem_Free( node->organizer.oalserver );
    DOFMem_Free (node );

    DOFServer_Unlock();
}

OALChannel OALServer_GetSendServerChannel( OALServer server )
{
    OALChannel ret = NULL;
    OALServer relSrv = server;

    DOFServer_Lock();

    while( relSrv )
    {
        if( relSrv->relatedServer && OALServer_IsDatagram( relSrv ) && OALServer_IsDatagram( relSrv->relatedServer->server ) )
        {
            relSrv = relSrv->relatedServer->server;
        }
        else
        {
            ret = relSrv->transport.channel;
            relSrv = NULL;
        }
    }

    DOFServer_Unlock();
    return ret;
}

DTL OALServer_GetDTL( OALServer srv )
{
    if( srv && srv->transport.channel)
    {
        if( srv->config->type != DOFSERVERTYPE_STREAM )
            return srv->transport.channel->remoteChannelData->dtl;
        else
            return srv->transport.dtl;
    }
    return NULL;
}

static void notifyServerComplete( OALThreadPoolItem dnode, boolean shutdown )
{
    ServerContext_Wrapper_t * node = (ServerContext_Wrapper_t*) dnode;

    node->item->callback->fns->Complete( node->item->callback, node->item->server, node->item->context, node->item->except);
    DOFException_Destroy( node->item->except );
    DOFServerContext_Destroy( node->item );
}

static void notifyServerContexts(OALSListNode itemx, void* argx)
{
    serverContext_t* item = (serverContext_t*)itemx;
    DOFException arg = (DOFException)argx;

    DOFServer_Lock();
    if(OALTimerItem_Remove(item->timer_item))
    {
        OALTimerItem_Destroy(item->timer_item, TRUE);
        item->timer_item = NULL;
        if( !arg )
        {
            ++item->server->StartCount;
            ++item->server->server->StartCount;
        }
        item->except = DOFException_Copy( arg );
        if( item->server->server->serverManager->server_enqueue )
        {
            ServerContext_Wrapper_t * node = item->enqNode;

            OALEnqueueItem_Init((OALEnqueueItem)node, notifyServerComplete);
            if(OALEnqueue_Enqueue( item->server->server->serverManager->server_enqueue, (OALEnqueueItem)node ))
            {
                DOFServer_Unlock();
                return;
            }
            /* only case is shutting down */
        }
        /* failure path */
        item->callback->fns->Complete( item->callback, item->server, item->context, DOFEXCEPTION_INSUFFICIENT_RESOURCES);
        if( !arg )
        {
            --item->server->StartCount;
            --item->server->server->StartCount;
        }
        DOFServerContext_Destroy( item );
        DOFException_Destroy( arg );
    }
    DOFServer_Unlock();
}

void OALServer_NotifyServerStarted( OALServer srv )
{
    DOFServer_Lock();
    srv->serverState = DOFSERVERSTATUS_STARTED;
    OALSList_RemoveAll( &srv->serverContexts, notifyServerContexts, NULL );
    OALServer_NotifyStateChange( srv, NULL );
    DOFServer_Unlock();
}

void OALServer_NotifyServerStopped( OALServer srv, DOFException except )
{
    boolean change = FALSE;
    DOFServer_Lock();
    if( srv->serverState == DOFSERVERSTATUS_STARTED )
        change = TRUE;
    srv->serverState = DOFSERVERSTATUS_STOPPED;
    OALSList_RemoveAll( &srv->serverContexts, notifyServerContexts, except);
    if( change )
        OALServer_NotifyStateChange( srv, except );
    DOFServer_Unlock();
}

void OALServer_NotifyServerTransportDown( OALServer srv, DOFException exception )
{
    DOFServer_Lock();
    if( srv->serverState == DOFSERVERSTATUS_STARTED )
    {
        OALServer_NotifyStateChange( srv, exception );
    }
    if( srv->transport.channel )
    {
        OALChannel_Release(srv->transport.channel);
        srv->transport.channel = NULL;
    }
    DOFServer_Unlock();
}

void OALServer_RemoveManager( OALServer srv )
{
    srv->serverManager = NULL;
}

/* set server back to stopped, Call with DOFServer_Lock */
void OALServer_RealStop( OALServer srv, boolean fromRelated, DOFException exception )
{
    if( srv->transport.channel )
    {
        if( srv->config->type == DOFSERVERTYPE_DATAGRAM )
        {
            OALChannel chnl = srv->transport.channel;
            srv->transport.channel = NULL;
            DTL_StopClose( chnl->remoteChannelData->dtl ); /* close so we don't receive any more packets */
            if( !fromRelated && !exception )
            {
                OALRemoteChannel_ServerInitiatedShutdown(chnl, srv);
            }
            else
            {
                OALRemoteChannel_Shutdown( chnl, exception );
            }
            OALChannel_Release( chnl );

            if( srv->relatedServer )
            {
                if( srv->relSrvList )
                    DOFServer_RemoveStateListenerInternal( srv->relatedServer, srv->relSrvList );
                if( !fromRelated )
                    DOFServer_StopInternal( srv->relatedServer, exception );
            }
        }
        else
        {
            DTL dtl = srv->transport.dtl;
            srv->transport.dtl = NULL;
            DTL_StopClose( dtl );
            DTL_Destroy( dtl );
        }

        if( srv->serverState == DOFSERVERSTATUS_STARTED )
        {
            srv->serverState = DOFSERVERSTATUS_STOPPED;
            OALServer_NotifyStateChange( srv, exception );
        }
    }
}

DOFServerState OALServer_CreateState( OALServer self, DOFException exception )
{
    DOFServerState ret = NULL;
    if ( self )
    {
        ret = DOFMem_Alloc( sizeof( DOFServerState_t ) );
        if ( ret )
        {
            PCRMem_Set( ret, 0, sizeof( DOFServerState_t ) );
            DOFServer_Lock();
            ret->address = DOFAddress_Copy( self->config->listenAddr );
            if( ret->address )
            {
                ret->transportConfig = DTLServerConfig_Copy( self->config->serverConfig );
                if( ret->transportConfig || !self->config->serverConfig )
                {
                    ret->exception = DOFException_Copy( exception );
                    ret->refCount = 1;

                    DOFServerCount_Lock();
                    ret->byteInputCount = self->byteInputCount;
                    ret->byteOutputCount = self->byteOutputCount;
                    ret->packetInputCount = self->packetInputCount;
                    ret->packetOutputCount = self->packetOutputCount;
                    DOFServerCount_Unlock();

                    ret->status = self->serverState;
                    ret->protocolNeg = self->config->negotiator;
                    ret->maxConnections = self->config->MaxConnections;
                    ret->receiveSilence = self->config->MaxReceiveSilence;
                    ret->sendSilence = self->config->MaxSendSilence;
                    ret->securityDesire = self->config->securityDesire;
                    ret->type = self->config->type;
                    DOFServer_Unlock();
                    return ret;

                }
                DOFAddress_Destroy( ret->address );
            }
            DOFServer_Unlock();
            DOFMem_Free( ret );
        }
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFServer: Failed to create server state.");
    }
    return NULL;
}

boolean OALServer_IsStarted( OALServer self )
{
    boolean ret = FALSE;
    if ( self )
    {
        DOFServer_Lock();
        ret = self->serverState == DOFSERVERSTATUS_STARTED;
        DOFServer_Unlock();
    }
    return ret;
}

boolean OALServer_IsDatagram( OALServer self )
{
    PCR_Assert( self );
    return ( self->config->type == DOFSERVERTYPE_DATAGRAM );
}

static boolean notifyStateChange(OALSListNode item, void *arg)
{
    DOFServerCallbackNode_t* srvNde = (DOFServerCallbackNode_t*)item;
    DOFServerState state = (DOFServerState)arg;

    DOFServer_EnqueueListenerCall( srvNde->server, srvNde->callback, state );
    return FALSE;
}

void OALServer_RemoveAllListener( OALServer self )
{
    DOFServer_Lock();
    if( self->stateCallback )
        OALSList_RemoveAll( &self->stateCallback, removeAllStateCallback, self );
    DOFServer_Unlock();
}

void OALServer_NotifyStateChange( OALServer self, DOFException exception )
{
    DOFServerState args = OALServer_CreateState( self, exception );

    if( args )
    {
        DOFServer_Lock();
        OALSList_Traverse( self->stateCallback, notifyStateChange, args);
        DOFRuntime_ServerStateChanged( self->thisDOFServer, args );
        DOFServer_Unlock();
        DOFServerState_Destroy( (DOFServerState)args );
    }
    else
    {
        OALServer_RemoveAllListener( self ); /* Remove all because state couldn't be created */
    }
}

DOFAddress OALServer_GetListenAddr(OALServer server)
{
    if( server )
        return server->config->listenAddr;
    return NULL;
}

static void handleRelatedGoingDown( OALServer srv )
{
    DOFServer_Lock();
    if( srv->transport.channel )
    {
        DOFException exception = DOFErrorException_Create(DOFERROR_TERMINATED, "Related Server Stopped.");
        OALServer_RealStop( srv, TRUE, exception );
        DOFException_Destroy( exception );
    }
    DOFServer_Unlock();
}

static void serverStateChanged( DOFServerStateListener self, const DOFServer server, DOFServerState state )
{
    ServerStateListener_t* realSelf = (ServerStateListener_t*)self;

    switch( state->status )
    {
    case DOFSERVERSTATUS_STOPPED:
        handleRelatedGoingDown( realSelf->server );
        break;
    default:
        break;
    }
}

static void serverRemoved( DOFServerStateListener self, DOFServer server, DOFException except )
{
    ServerStateListener_t* realSelf = (ServerStateListener_t*)self;
    handleRelatedGoingDown( realSelf->server );
}

#if OALSERVER_PRINT_REF_COUNT_DEBUG
OALServer OALServer_Retain_old( OALServer srv, const char* function, const char* filename, int line)
#else
OALServer OALServer_Retain( OALServer srv )
#endif
{
    OALServer ret = NULL;
    if( srv )
    {
        DOFServer_Lock();
        PCR_Assert( srv->refCount < 0xFFFFFFFF );
        if( ++(srv->refCount) )
        {
#if OALSERVER_PRINT_REF_COUNT_DEBUG
            fprintf( stderr,"OALServer: %p +%d %s:%d (%s)\n", srv, srv->refCount, filename, line, function );
#endif
            ret = srv;
        }
        DOFServer_Unlock();
    }
    return ret;
}

static boolean findOALServerAndRemove( OALSListNode itemx, void* arg )
{
    OALServerNode_t* item = (OALServerNode_t*)itemx;
    OALServer srv = (OALServer)arg;
    if( item->server == srv )
        return TRUE;
    return FALSE;
}

#if OALSERVER_PRINT_REF_COUNT_DEBUG
void OALServer_Destroy_old( OALServer srv , const char* function, const char* filename, int line)
#else
void OALServer_Destroy( OALServer srv )
#endif
{
    OALServerNode_t* node = NULL;
    uint32 _refCount;

    if (!srv)
        return;

    DOFServer_Lock();
    PCR_Assert( srv->refCount );
    _refCount = --(srv->refCount);
#if OALSERVER_PRINT_REF_COUNT_DEBUG
    fprintf( stderr,"OALServer: %p -%d DOF %d %s:%d (%s)\n", srv, srv->refCount, srv->thisDOFServer->refCount, filename, line, function );
#endif

    if ( _refCount || srv->thisDOFServer->refCount > 1 ) /* our dof server has been retained externally */
    {
        DOFServer_Unlock();
        return;
    }


    /* remove from manager. will not be used by any other starts */
    if ( srv->serverManager && (node = (OALServerNode_t*)OALSList_FindAndRemove( &srv->serverManager->oalServers, NULL, findOALServerAndRemove, srv)) != NULL )
    {
        DOFMem_Free(node);
        --srv->serverManager->count;
    }

    if ( srv->serverContexts )
    {
        DOFException except = DOFErrorException_Create(DOFERROR_FAILED, "Server destroyed.");
        OALSList_RemoveAll( &srv->serverContexts, removeAllContexts, except );
        DOFException_Destroy( except );
    }

    OALServer_RemoveAllListener( srv );

    OALServer_RealStop( srv, FALSE, NULL );

    if( srv->refCount != 0 )
    {
        DOFServer_Unlock();
        return;
    }

    if ( srv->relatedServer )
    {
        DOFServer_Release( srv->relatedServer );
        srv->relatedServer = NULL;
    }

    if( srv->cleanupnode )
    {
        OALEnqueueItem_Init((OALEnqueueItem)srv->cleanupnode, destroyOALServer);
        srv->cleanupnode->organizer.oalserver = srv;
        if (srv->serverManager)
        {
            if ( !OALEnqueue_Enqueue( srv->serverManager->server_enqueue, (OALEnqueueItem)srv->cleanupnode) )
            {
                DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_WARNING, "DOFServer: Failed to cleanup.");
            }
            else
                srv->cleanupnode = NULL;
        }
        else
        {
#if OALSERVER_PRINT_REF_COUNT_DEBUG
            fprintf( stderr, "OALServer Free: %p\n", srv );
#endif
            DOFMem_Free( srv->thisDOFServer ); /* DOFServer Free */
            DOFServerConfig_Destroy( srv->config );
            DOFSecurityMembrane_Destroy( srv->securityMembrane );
            DOFMem_Free( srv->cleanupnode );
            DOFMem_Free( srv );
        }
    }
    DOFServer_Unlock();
}

void OALServer_UpdateInputByteCount( OALServer server, uint32 count )
{
    DOFServerCount_Lock();
    server->byteInputCount += count;
    server->packetInputCount++;
    DOFServerCount_Unlock();
}

void OALServer_UpdateOutputByteCount( OALServer server, uint32 count )
{
    DOFServerCount_Lock();
    server->byteOutputCount += count;
    server->packetOutputCount++;
    DOFServerCount_Unlock();
}

/* THIS SERVER CAN ONLY BE USED BY DOFRUNTIME */
DOFServer OALServer_GetDOFServer( OALServer server )
{
    if( server )
        return server->thisDOFServer;
    return NULL;
}

DOFServer OALServer_CreateDOFServer( OALServer server )
{
    if( server )
    {
        DOFServer relserver = DOFMem_Alloc(sizeof(DOFServer_t));
        if( relserver )
        {
            relserver->server = OALServer_Retain( server );
            if( relserver->server )
            {
                relserver->refCount = 1;
                relserver->StartCount = 0;
                return relserver;
            }
            DOFMem_Free( relserver );
        }
    }
    return NULL;
}

struct all_servers_wrapper
{
    server_traverse_fn_t fn;
    void *context;
};

static boolean servers_traverse(OALSListNode itemx, void *arg)
{
    OALServerNode_t* item = (OALServerNode_t*)itemx;
    struct all_servers_wrapper* wrapper = (struct all_servers_wrapper*) arg;

    if(wrapper->fn(item->server, wrapper->context))
        return TRUE;
    return FALSE;
}

/* Call with server lock */
void OALServer_TraverseAllServers(server_traverse_fn_t fn, void *context)
{
    struct all_servers_wrapper ctx;
    PCR_Assert(fn);
    ctx.fn = fn;
    ctx.context = context;
    OALSList_Traverse(((DOF)DOF_GetDof())->serverManager->oalServers, servers_traverse, &ctx);
}


static void internalConnectionListener_Removed( DOFConnectionStateListener self, DOFConnection connection, DOFException except )
{
    internalConnectionListener_t* selfx = (internalConnectionListener_t*)self;
    DOFServer_Lock();
    --selfx->server->connectionsCount;
    DOFServer_Unlock();
    OALServer_Destroy(selfx->server);
    DOFMem_Free(self);
}

static void internalConnectionListener_StateChanged( DOFConnectionStateListener self, DOFConnection connection, DOFConnectionState state )
{
    /* Blank on purpose */
}

static const struct DOFConnectionStateListenerFns_t serverConnectionListener_fns= {internalConnectionListener_Removed, internalConnectionListener_StateChanged};

boolean OALServer_AddConnectionListener(OALServer server, DOFConnection conn)
{
    internalConnectionListener_t* listener = (internalConnectionListener_t*)DOFMem_Alloc(sizeof(internalConnectionListener_t));
    if( listener )
    {
        listener->base.fns = &serverConnectionListener_fns;
        listener->server = OALServer_Retain(server);
        if( listener->server )
        {
            if( DOFConnection_AddStateListener( conn, (DOFConnectionStateListener)listener ) )
            {
                DOFServer_Lock();
                ++server->connectionsCount;
                DOFServer_Unlock();
                return TRUE;
            }
            OALServer_Destroy( listener->server );
        }
        DOFMem_Free( listener );
    }

    return FALSE;
}

struct server_addr_wrapper
{
    DOFAddress address;
    boolean success;
};

static boolean checkServerAddr(OALServer server, void* ctx)
{
    struct server_addr_wrapper* wrapper = (struct server_addr_wrapper*)ctx;
    wrapper->success = ( server->serverState == DOFSERVERSTATUS_STARTED &&
                         server->config->type == DOFSERVERTYPE_STREAM &&
                         DOFAddress_IsEqual( server->config->listenAddr,  wrapper->address ) );
    return wrapper->success; /* If successful, we're done with the search. */
}

boolean OALServer_HaveStartedStreamServer( const DOFAddress addr )
{
    struct server_addr_wrapper wrapper;

    wrapper.address = addr;
    wrapper.success = FALSE;

    DOFServer_Lock();
    OALServer_TraverseAllServers(checkServerAddr, &wrapper);
    DOFServer_Unlock();

    return wrapper.success;
}
