/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/bridge.h>
#include <dof/server.h>
#include <dof/address.h>
#include <dof/transport.h>
#include <dof/dof.h>
#include <internal/dofinternal.h>
#include <internal/protocolstack.h>
#include <internal/credentialsinternal.h>
#include <internal/connectionmanager.h>
#include <internal/serverinternal.h>
#include <internal/connectionless_remotechannel.h>
#include <internal/connectionless_protocolstack.h>
#include <internal/remotechannelmanager.h>
#include <internal/transportmanager.h>
#include <internal/connection_internal.h>
#include <internal/remotechannel.h>
#include <internal/permissionsetinternal.h>

#include "server_common.h"

#define DOFSERVERCONFIGFAILUREBITS_ADDRESS          0x01
#define DOFSERVERCONFIGFAILUREBITS_TRANSPORTCONFIG  0x02
#define DOFSERVERCONFIGFAILUREBITS_RECEIVEFILTER    0x04
#define DOFSERVERCONFIGFAILUREBITS_SENDFILTER       0x08
#define DOFSERVERCONFIGFAILUREBITS_OPERATIONBRIDGE  0x10

static void setError( uint32* errorFlags, uint32 errorCode, boolean worked )
{
    if(worked)
        *errorFlags &= ~errorCode;
    else
        *errorFlags |= errorCode;
}

DOFServerConfigBuilder DOFServerConfigBuilder_Create( DOFServerType type, DOFAddress listenAddr )
{
    DOFServerConfig ret = NULL;

    if ( listenAddr )
    {
        ret = DOFMem_Alloc( sizeof(DOFServerConfig_t) );
        if( ret )
        {
            PCRMem_Set( ret, 0, sizeof(DOFServerConfig_t) );
            ret->refCount = 1;
            ret->securityDesire = DOFSECURITYDESIRE_ANY;
            ret->type = type;
            ret->MaxReceiveSilence = 120000;
            ret->MaxConnections = DOF_UNLIMITED;
            ret->MaxSendSilence = DOF_TIMEOUT_NEVER;
            ret->negotiator = DOF_DEFAULT_PROTOCOL_NEGOTIATOR;
            ret->listenAddr = DOFAddress_Copy( listenAddr );
            ret->receiveFilter = NULL;
            ret->sendFilter = NULL;
            ret->bridge = NULL;
            if ( !ret->listenAddr )
            {
                DOFLog_Message(  DOFLOGLEVEL_ERROR, "DOFServerConfigBuilder: Failed to copy address into server config builder.");
                DOFServerConfig_Destroy( ret );
                ret = NULL;
            }
        }
    }
    else
        DOFLog_Message(  DOFLOGLEVEL_ERROR, "DOFServerConfigBuilder: Failed to create server config builder.");

    return (DOFServerConfigBuilder) ret;
}

DOFServerConfigBuilder DOFServerConfigBuilder_Create_Config( DOFServerConfig config )
{
    return (DOFServerConfigBuilder) DOFServerConfigBuilder_Build( (DOFServerConfigBuilder) config );
}

DOFServerConfigBuilder DOFServerConfigBuilder_Create_State( DOFServerState state )
{
    if( state )
    {
        DOFServerConfigBuilder ret = DOFServerConfigBuilder_Create( DOFServerState_GetServerType( state ), DOFServerState_GetAddress( state ) );
        if( ret )
        {
            if( DOFServerConfigBuilder_SetConnectionLimit( ret, DOFServerState_GetConnectionLimit( state ) ) )
            {
                if( DOFServerConfigBuilder_SetProtocolNegotiator( ret, DOFServerState_GetProtocolNegotiator( state ) ) )
                {
                    if( DOFServerConfigBuilder_SetMaxReceiveSilence( ret, DOFServerState_GetMaxReceiveSilence( state ) ) )
                    {
                        if( DOFServerConfigBuilder_SetMaxSendSilence( ret, DOFServerState_GetMaxSendSilence( state ) ) )
                        {
                            if( DOFServerConfigBuilder_SetSecurityDesire( ret, DOFServerState_GetSecurityDesire( state ) ) )
                            {
                                if( DOFServerConfigBuilder_SetTransportConfig( ret, DOFServerState_GetTransportConfig( state ) ) )
                                    return ret;
                            }
                        }
                    }
                }
            }
            DOFServerConfigBuilder_Destroy( ret );
        }
    }
    return NULL;
}

DOFServerConfigBuilder DOFServerConfigBuilder_Copy( DOFServerConfigBuilder builder )
{
    return (DOFServerConfigBuilder) DOFServerConfigBuilder_Build( builder );
}

DOFServerConfig DOFServerConfig_Copy( const DOFServerConfig self )
{
    if ( self )
    {
        PCRMutex_Lock(ref_count_mutex);
        self->refCount++;
        PCRMutex_Unlock(ref_count_mutex);
    }
    return self;
}

void DOFServerConfigBuilder_Destroy( DOFServerConfigBuilder self )
{
    DOFServerConfig_Destroy( (DOFServerConfig) self );
}

DOFServerConfig DOFServerConfigBuilder_BuildAndDestroy( DOFServerConfigBuilder selfx )
{
    DOFServerConfig self = (DOFServerConfig)selfx;
    if( self && self->errors == 0 )
    {
        return self;
    }
    else
    {
        DOFServerConfig_Destroy(self);
    }
    return NULL;
}

DOFServerConfig DOFServerConfigBuilder_Build( DOFServerConfigBuilder selfx )
{
    /* Deep copy to make the result immutable. */
    DOFServerConfig ret = NULL;
    DOFServerConfig self = (DOFServerConfig)selfx;

    PCR_Assert( self->listenAddr );

    if(self && self->listenAddr && self->errors == 0)
    {
        ret = DOFMem_Alloc( sizeof(DOFServerConfig_t) );
        if( ret )
        {
            PCRMem_Set( ret, 0, sizeof(DOFServerConfig_t) );
            ret->refCount = 1;
            ret->securityDesire = self->securityDesire;
            ret->MaxReceiveSilence = self->MaxReceiveSilence;
            ret->MaxSendSilence = self->MaxSendSilence;
            ret->MaxConnections = self->MaxConnections;
            ret->negotiator = self->negotiator;
            ret->type = self->type;
            ret->serverConfig = DTLServerConfig_Copy( self->serverConfig );
            if( self->creds )
            {
                ret->creds = OALCredentialsTree_DeepCopy( self->creds );
                if(!ret->creds)
                {
                    goto FAIL;
                }
            }
            if( self->dcreds )
            {
                ret->dcreds = OALDomainCredentialsTree_DeepCopy( self->dcreds );
                if(!ret->dcreds)
                {
                    goto FAIL;
                }
            }
            ret->listenAddr = DOFAddress_Copy( self->listenAddr );
            if ( !ret->listenAddr )
            {
                goto FAIL;
            }
        }
    }
    return ret;

FAIL:
    DOFLog_Message(  DOFLOGLEVEL_WARNING, "DOFServerConfig: Failed to build DOFServerConfig.");
    DOFServerConfig_Destroy( ret );
    return NULL;
}

void DOFServerConfig_Destroy( DOFServerConfig self )
{
    if (self)
    {
        boolean gone;
        PCRMutex_Lock(ref_count_mutex);
        PCR_Assert( self->refCount );
        gone = (!--self->refCount);
        PCRMutex_Unlock(ref_count_mutex);
        if ( gone )
        {
            DOFAddress_Destroy( self->listenAddr);
            DTLServerConfig_Destroy( self->serverConfig );
            OALCredentialsTree_Destroy( self->creds );
            OALDomainCredentialsTree_Destroy( self->dcreds );
            DOFOperationBridgeConfig_Destroy( self->bridge );
            DOFMem_Free( self );
        }
    }
}

typedef struct dscgcc_t
{
    DOFServerConfig_GetDomainCredential_Callback fn;
    void* context;
}dscgcc_t;

static void DOFServerConfig_Internal_GetDomainCredential( DOFCredentials self, void* context )
{
    dscgcc_t* ctx = (dscgcc_t*)context;

    ctx->fn( self, ctx->context );

    DOFMem_Free( ctx );
}

/* May call directly back into the callback function  before leaving this function. */
void DOFServerConfig_GetDomainCredential( DOFServerConfig self, DOFObjectIDDomain domainID, uint32 timeout, DOFServerConfig_GetDomainCredential_Callback fn, void* context )
{
    dscgcc_t* ctx = (dscgcc_t*)DOFMem_Alloc( sizeof( dscgcc_t ) );
    if( ctx )
    {
        ctx->fn = fn;
        ctx->context = context;

        OALCredentialsTree_GetCredentials( self->creds, domainID, timeout, DOFServerConfig_Internal_GetDomainCredential, ctx );
    }
    else
        fn( NULL, context );
}

boolean DOFServerConfig_IsEqual( DOFServerConfig self, DOFServerConfig other )
{
    PCR_Assert(self);
    PCR_Assert(other);

    if( self->type != other->type || self->MaxReceiveSilence != other->MaxReceiveSilence || self->MaxConnections != other->MaxConnections || self->MaxSendSilence != other->MaxSendSilence || self->securityDesire != other->securityDesire )
        return FALSE;

    if( !OALCredentialsTree_IsEqual( self->creds, other->creds ) )
        return FALSE;

    if( !OALDomainCredentialsTree_IsEqual( self->dcreds, other->dcreds ) )
        return FALSE;

    if( !DOFAddress_IsEqual( self->listenAddr, other->listenAddr ) )
        return FALSE;

    if( !DTLServerConfig_IsEqual( self->serverConfig, other->serverConfig) )
        return FALSE;

    return TRUE;
}

boolean DOFServerConfigBuilder_SetServerType( const DOFServerConfigBuilder selfx, DOFServerType type )
{
    DOFServerConfig self = (DOFServerConfig)selfx;
    if(selfx)
    {
        self->type = type;
        return TRUE;
    }
    return FALSE;
}

boolean DOFServerConfigBuilder_SetAddress( const DOFServerConfigBuilder selfx, DOFAddress addr )
{
    DOFServerConfig self = (DOFServerConfig)selfx;
    if ( self && addr )
    {
        DOFAddress temp = self->listenAddr;
        self->listenAddr = DOFAddress_Copy( addr );
        if( self->listenAddr )
        {
            setError( &(self->errors), DOFSERVERCONFIGFAILUREBITS_ADDRESS, TRUE );
            DOFAddress_Destroy( temp );
            return TRUE;
        }
        else
        {
            self->listenAddr = temp;
            setError( &(self->errors), DOFSERVERCONFIGFAILUREBITS_ADDRESS, FALSE );
            DOFLog_Message(  DOFLOGLEVEL_WARNING, "DOFServerConfigBuilder: Failed to set DOFAddress!" );
        }
    }
    return FALSE;
}

boolean DOFServerConfigBuilder_SetSecurityDesire( const DOFServerConfigBuilder selfx, DOFSecurityDesire desire )
{
    DOFServerConfig self = (DOFServerConfig)selfx;
    if( self )
    {
        self->securityDesire = desire;
        return TRUE;
    }
    return FALSE;
}

boolean DOFServerConfigBuilder_SetTransportConfig( DOFServerConfigBuilder selfx, DTLServerConfig config )
{
    DOFServerConfig self = (DOFServerConfig)selfx;
    if( config && self )
    {
        DTLServerConfig temp = self->serverConfig;
        self->serverConfig = DTLServerConfig_Copy( config );
        if( self->serverConfig )
        {
            setError( &(self->errors), DOFSERVERCONFIGFAILUREBITS_TRANSPORTCONFIG, TRUE );
            DTLServerConfig_Destroy( temp );
            return TRUE;
        }
        else
        {
            self->serverConfig = temp;
            setError( &(self->errors), DOFSERVERCONFIGFAILUREBITS_TRANSPORTCONFIG, FALSE );
            DOFLog_Message(  DOFLOGLEVEL_WARNING, "DOFServerConfigBuilder: Failed to set DTLServerConfig." );
        }
    }
    return FALSE;
}

boolean DOFServerConfigBuilder_SetMaxConnections( DOFServerConfigBuilder selfx, uint16 maxConnections )
{
    DOFServerConfig self = (DOFServerConfig)selfx;
    if( self )
    {
        if( maxConnections != 0 )
        {
            ((DOFServerConfig)self)->MaxConnections = maxConnections;
        }
        else
        {
            ((DOFServerConfig)self)->MaxConnections = DOF_UNLIMITED;
        }
        return TRUE;
    }
    return FALSE;
}

boolean DOFServerConfigBuilder_SetConnectionLimit( DOFServerConfigBuilder selfx, uint32 maxConnections )
{
    DOFServerConfig self = (DOFServerConfig)selfx;
    if( self )
    {
        self->MaxConnections = maxConnections;
        return TRUE;
    }
    return FALSE;
}

boolean DOFServerConfigBuilder_AddCredentials( DOFServerConfigBuilder selfx, DOFCredentials auth )
{
    DOFServerConfig self = (DOFServerConfig)selfx;
    if( self && auth )
    {
        if( !self->creds )
        {
            self->creds = OALCredentialsTree_Create();
            if( !self->creds )
            {
                return FALSE;
            }
        }
        return OALCredentialsTree_Insert( self->creds, auth );
    }
    return FALSE;
}

boolean DOFServerConfigBuilder_SetWildcardCredentials( const DOFServerConfigBuilder selfx, DOFCredentials credentials )
{
    DOFServerConfig self = (DOFServerConfig)selfx;
    if( self )
    {
        if( !self->creds )
        {
            self->creds = OALCredentialsTree_Create();
            if( !self->creds )
            {
                return FALSE;
            }
        }
        return OALCredentialsTree_SetWildcardCredentials( self->creds, credentials );
    }
    return FALSE;
}

DOFCredentials DOFServerConfig_GetWildcardCredentials( const DOFServerConfig self )
{
    if( self && self->creds )
        return OALCredentialsTree_GetWildcardCredentials( self->creds );
    return NULL;
}

boolean DOFServerConfigBuilder_AddDomain( const DOFServerConfigBuilder selfx, DOFDomainConfig auth )
{
    DOFServerConfig self = (DOFServerConfig)selfx;
    if( self && auth )
    {
        DOFDomain domain;
        if( !self->dcreds )
        {
            self->dcreds = OALDomainCredentialsTree_Create();
            if( !self->dcreds )
            {
                return FALSE;
            }
        }

        domain = DOF_CreateDomain( DOF_GetDof(), auth );
        if( domain )
        {
            boolean ret = OALDomainCredentialsTree_Insert( self->dcreds, domain );
            DOFDomain_Destroy( domain );
            return ret;
        }
    }
    return FALSE;
}

boolean DOFServerConfigBuilder_AddTrustedDomain( const DOFServerConfigBuilder selfx, DOFObjectIDDomain domainID )
{
    DOFServerConfig self = (DOFServerConfig)selfx;
    if( self && domainID )
    {
        if( !self->dcreds )
        {
            self->dcreds = OALDomainCredentialsTree_Create();
            if( !self->dcreds )
            {
                return FALSE;
            }
        }
        if( OALDomainCredentialsTree_InsertTrustedDomain( self->dcreds, domainID ) )
            return TRUE;
    }
    return FALSE;
}

boolean DOFServerConfigBuilder_SetDomainDiscoveryCredentials( const DOFServerConfigBuilder selfx, DOFCredentials credentials )
{
    DOFServerConfig self = (DOFServerConfig)selfx;
    if( self )
    {
        if( !self->dcreds )
        {
            self->dcreds = OALDomainCredentialsTree_Create();
            if( !self->dcreds )
            {
                return FALSE;
            }
        }
        return OALDomainCredentialsTree_SetDomainDiscoveryCredentials( self->dcreds, credentials );
    }
    return FALSE;
}

DOFCredentials DOFServerConfig_GetDomainDiscoveryCredentials( const DOFServerConfig self )
{
    if( self && self->dcreds )
        return OALDomainCredentialsTree_GetDomainDiscoveryCredentials( self->dcreds );
    return NULL;
}

DOFServerType DOFServerConfig_GetServerType(const DOFServerConfig self )
{
    if( self )
        return self->type;
    return 0;
}

DOFAddress DOFServerConfig_GetAddress( const DOFServerConfig self )
{
    if( self )
        return self->listenAddr;
    return NULL;
}

DOFSecurityDesire DOFServerConfig_GetSecurityDesire( const DOFServerConfig self )
{
    if( self )
        return self->securityDesire;
    return 0;
}

DTLServerConfig DOFServerConfig_GetTransportConfig( const DOFServerConfig self )
{
    if( self )
        return self->serverConfig;
    return NULL;
}

DOFProtocolNegotiator DOFServerConfig_GetProtocolNegotiator( const DOFServerConfig self )
{
    return (self ? self->negotiator : NULL);
}

boolean DOFServerConfigBuilder_SetProtocolNegotiator( DOFServerConfigBuilder selfx, DOFProtocolNegotiator negotiator)
{
    DOFServerConfig self = (DOFServerConfig)selfx;
    if(self)
    {
        self->negotiator = negotiator;
        return TRUE;
    }
    return FALSE;
}

uint32 DOFServerConfig_GetMaxReceiveSilence( const DOFServerConfig self )
{
    if( self )
        return self->MaxReceiveSilence;
    return 0;
}

uint32 DOFServerConfig_GetMaxSendSilence( const DOFServerConfig self )
{
    if( self )
        return self->MaxSendSilence;
    return 0;
}

uint16 DOFServerConfig_GetMaxConnections( const DOFServerConfig self )
{
    if( self )
    {
        if( self->MaxConnections == 0 )
        {
            return 1;
        }
        else if( self->MaxConnections < 0xFFFF )
        {
            return self->MaxConnections;
        }

        /* FALL through intended */
    }
    return 0;
}

uint32 DOFServerConfig_GetConnectionLimit( const DOFServerConfig self )
{
    if( self )
    {
        return self->MaxConnections;
    }
    return 0;
}

DOFOperationFilter DOFServerConfig_GetReceiveFilter(DOFServerConfig const self)
{
    if(self != NULL)
    {
        return self->receiveFilter;
    }
    return NULL;
}

DOFOperationFilter DOFServerConfig_GetSendFilter(DOFServerConfig const self)
{
    if(self != NULL)
    {
        return self->sendFilter;
    }
    return NULL;
}

DOFOperationBridgeConfig DOFServerConfig_GetBridge(DOFServerConfig const self)
{
    if(self != NULL)
    {
        return self->bridge;
    }
    return NULL;
}

boolean DOFServerConfigBuilder_SetMaxReceiveSilence( DOFServerConfigBuilder selfx, uint32 maxSilence )
{
    DOFServerConfig self = (DOFServerConfig)selfx;
    if (self)
    {
        self->MaxReceiveSilence = maxSilence;
        return TRUE;
    }
    return FALSE;
}

boolean DOFServerConfigBuilder_SetReceiveFilter( DOFServerConfigBuilder self, DOFOperationFilter filter )
{
    DOFServerConfig selfx = (DOFServerConfig)self;

    boolean retVal = FALSE;

    if(self != NULL)
    {
        DOFOperationFilter newFilter = DOFOperationFilter_Copy(filter);

        retVal = ((newFilter != NULL) || (filter == NULL));
        DOFOperationFilter_Destroy(selfx->receiveFilter);
        selfx->receiveFilter = newFilter;
        setError(&(selfx->errors), DOFSERVERCONFIGFAILUREBITS_RECEIVEFILTER, retVal);
    }
    return retVal;
}

boolean DOFServerConfigBuilder_SetSendFilter( DOFServerConfigBuilder self, DOFOperationFilter filter )
{
    DOFServerConfig selfx = (DOFServerConfig)self;

    boolean retVal = FALSE;

    if(self != NULL)
    {
        DOFOperationFilter newFilter = DOFOperationFilter_Copy(filter);

        retVal = ((newFilter != NULL) || (filter == NULL));
        DOFOperationFilter_Destroy(selfx->sendFilter);
        selfx->sendFilter = newFilter;
        setError(&(selfx->errors), DOFSERVERCONFIGFAILUREBITS_SENDFILTER, retVal);
    }
    return retVal;
}

boolean DOFServerConfigBuilder_SetBridge( DOFServerConfigBuilder self, DOFOperationBridgeConfig bridge )
{
    DOFServerConfig selfx = (DOFServerConfig)self;

    boolean retVal = FALSE;

    if(self != NULL)
    {
        DOFOperationBridgeConfig newBridge = DOFOperationBridgeConfig_Copy(bridge);

        retVal = ((newBridge != NULL) || (bridge == NULL));
        DOFOperationBridgeConfig_Destroy(selfx->bridge);
        selfx->bridge = newBridge;
        setError(&(selfx->errors), DOFSERVERCONFIGFAILUREBITS_OPERATIONBRIDGE, retVal);
    }
    return retVal;
}

boolean DOFServerConfigBuilder_SetMaxSendSilence(DOFServerConfigBuilder selfx, uint32 maxSilence)
{
    DOFServerConfig self = (DOFServerConfig)selfx;
    if(self)
    {
        self->MaxSendSilence = maxSilence;
        return TRUE;
    }
    return FALSE;
}

boolean DOFServerState_IsStarted( const DOFServerState self )
{
    PCR_Assert( self );
    return ( self->status == DOFSERVERSTATUS_STARTED );
}

DOFException DOFServerState_GetException( const DOFServerState self )
{
    PCR_Assert( self );
    return self->exception;
}

DOFServerState DOFServerState_Copy( DOFServerState self )
{
    PCR_Assert( self );
    DOFServer_Lock();
    PCR_Assert( self->refCount < 0xFFFFFFFF );
    ++self->refCount;
    DOFServer_Unlock();
    return self;
}

uint64 DOFServerState_GetReceiveByteCount( DOFServerState self )
{
    PCR_Assert( self );
    return self->byteInputCount;
}

uint64 DOFServerState_GetSendByteCount( DOFServerState self )
{
    PCR_Assert( self );
    return self->byteOutputCount;
}

uint64 DOFServerState_GetReceivePacketCount( DOFServerState self )
{
    PCR_Assert( self );
    return self->packetInputCount;
}

uint64 DOFServerState_GetSendPacketCount( DOFServerState self )
{
    PCR_Assert( self );
    return self->packetOutputCount;
}

DOFServerType DOFServerState_GetServerType( DOFServerState self )
{
    PCR_Assert( self );
    return self->type;
}

DOFAddress DOFServerState_GetAddress( DOFServerState self )
{
    PCR_Assert( self );
    return self->address;
}

DOFSecurityDesire DOFServerState_GetSecurityDesire( DOFServerState self )
{
    PCR_Assert( self );
    return self->securityDesire;
}

DTLServerConfig DOFServerState_GetTransportConfig( DOFServerState self )
{
    PCR_Assert( self );
    return self->transportConfig;
}

DOFProtocolNegotiator DOFServerState_GetProtocolNegotiator( DOFServerState self )
{
    PCR_Assert( self );
    return self->protocolNeg;
}

uint32 DOFServerState_GetMaxSendSilence( DOFServerState self )
{
    PCR_Assert( self );
    return self->sendSilence;
}

uint32 DOFServerState_GetMaxReceiveSilence( DOFServerState self )
{
    PCR_Assert( self );
    return self->receiveSilence;
}

uint16 DOFServerState_GetMaxConnections( DOFServerState self )
{
    PCR_Assert( self );
    if( self->maxConnections == 0 )
    {
        return 1;
    }
    else if( self->maxConnections < 0xFFFF )
    {
        return self->maxConnections;
    }
    return 0;
}

uint32 DOFServerState_GetConnectionLimit( DOFServerState self )
{
    PCR_Assert( self );
    return self->maxConnections;
}

void DOFServerState_Destroy( DOFServerState self )
{
    if ( self )
    {
        uint32 _refCount;

        DOFServer_Lock();
        PCR_Assert( self->refCount );
        _refCount = --(self->refCount);
        DOFServer_Unlock();

        if(!_refCount)
        {
            DOFException_Destroy(self->exception);
            DOFAddress_Destroy( self->address );
            DTLServerConfig_Destroy( self->transportConfig );
            DOFMem_Free( self );
        }
    }
}
