/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/address.h>
#include <internal/channel.h>
#include <internal/key_request.h>
#include <internal/objectinternal.h>
#include <internal/operationinternal.h>
#include <internal/sgmp_v1.h>
#include "sgmp_v1_operation.h"


/* the callers of the callers of this will log the error */
static DOFOperation _create( DOFOperationType opType, OALChannel channel, DOFAddress address,
                             uint32 timeout, boolean providerOperation )
{
    DOFOperation ret;

    PCR_Assert( opType >= DOFOPERATIONTYPE_SGMP_V1_HEARTBEAT );
    PCR_Assert( opType < DOFOPERATIONTYPE_SGMP_LAST );
    ret = (DOFOperation)DOFMem_Alloc( sizeof( DOFOperation_t ) );
    if ( ret )
    {
        PCRMem_Set( ret, 0, sizeof( DOFOperation_t ) );
        ret->sources = (DOFOperationSource)DOFMem_Alloc( sizeof( DOFOperationSource_t ) );
        if ( ret->sources )
        {
            PCRMem_Set( ret->sources, 0, sizeof( DOFOperationSource_t ) );
            ret->operation_will_close = OALObservable_Create();
            if ( ret->operation_will_close )
            {
                ret->sources->channel = OALChannel_Retain( channel );
                if ( ret->sources->channel )
                {
                    ret->sources->address = DOFAddress_Copy( address );
                    if ( !address || ret->sources->address )
                    {
                        ret->cancelContext = DOFOperationManager_CreateCancelContext( ret );
                        if ( ret->cancelContext )
                        {
                            ret->ref_count = 1;
                            ret->operation_type = opType;
                            if (providerOperation)
                                ret->flags.bits.internally_created = 1;
                            else
                                ret->flags.bits.internally_created = 0;
                            ret->sources->oal_timeout = timeout;
                            return ret;
                        }
                        DOFAddress_Destroy( ret->sources->address );
                    }
                    OALChannel_Release( ret->sources->channel );
                }
                OALObservable_Destroy( ret->operation_will_close );
            }
            DOFMem_Free( ret->sources );
        }
        DOFMem_Free( ret );
    }
    return NULL;
}


SGMP_V1_OperationContext SGMP_V1_OperationContext_Create( void )
{
    SGMP_V1_OperationContext ret = DOFMem_Alloc( sizeof( SGMP_V1_OperationContext_t ) );

    if (ret)
        PCRMem_Set( ret, 0, sizeof( SGMP_V1_OperationContext_t ) );

    return ret;
}


void SGMP_V1_OperationContext_Destroy(void* context)
{
    SGMP_V1_OperationContext self = (SGMP_V1_OperationContext)context;
    if (self)
    {
        DOFObjectID_Destroy( self->group );
        DOFKeyRequest_Destroy( self->keyRequest );
        DOFObjectIDDomain_Destroy( self->sgmpdomain );
        OALStream_Release( self->I_Rec_Stream, FALSE );
        OALKeyGrant_Release( self->ticket );
        OALStream_Release( self->initialState, TRUE );
        OALChannel_Release( self->secureGroupChannel );
        DOFMem_Free( self->key );
        OALStream_Release( self->sgmpStreamI, TRUE );
        /* self->requestGroupContext not released here because it is involved with responses from an original SGMP ReqG
           multicast request. It is OK to release it at time and in the path of need. */
        /*self->keyChangedEvent not released here because our use did not increment its refCount, because a PCREvent has
          no refCount. Release happens with at creator/caller, which is SGMP_V1_SendRekeyMerge(). */
        DOFMem_Free( self );
    }
}


DOFOperation DOFOperation_Create_Heartbeat( OALChannel channel, DOFAddress address, DOFSecurityScope securityScope,
                                            uint32 tieBreaker, uint16 appid, uint8 desire, uint32 timeout,
                                            boolean providerOperation )
{
    DOFOperation ret;

    PCR_Assert( securityScope );
    if ((ret = _create( DOFOPERATIONTYPE_SGMP_V1_HEARTBEAT, channel, address, timeout, providerOperation )) != NULL)
    {
        if ((ret->securityScope = DOFSecurityScope_Copy( DOF_GetDefaultSecurityScope( DOF_GetDof() ) )) != NULL)
        {
            SGMP_V1_OperationContext context = SGMP_V1_OperationContext_Create();

            if (context)
            {
                /* Success. */
                context->tieBreaker = tieBreaker;
                context->appid = appid;
                context->desire = desire;
                ret->context = context;
                ret->contextDestroyFunction = SGMP_V1_OperationContext_Destroy;
                return ret;
            }
        }
        DOFOperation_Internal_Destroy( ret );                       /* ret->securityScope cleans up here. */
    }

    return NULL;
}


DOFOperation DOFOperation_Create_EpochChanged( OALChannel channel, DOFAddress address, DOFSecurityScope securityScope,
                                               uint16 epoch, uint16 Tmin, uint32 timeout, boolean providerOperation,
                                               boolean doSendRetry, uint32 retryPeriod )
{
    DOFOperation ret;

    PCR_Assert( securityScope );
    if ((ret = _create( DOFOPERATIONTYPE_SGMP_V1_EPOCHCHANGED, channel, address, timeout, providerOperation )) != NULL)
    {
        if ((ret->securityScope = DOFSecurityScope_Copy( DOF_GetDefaultSecurityScope( DOF_GetDof() ) )) != NULL)
        {
            ret->opID = DOFOperationID_CreateLocal();   /* Without, AddOperation won't work and timeout won't happen. */
            if (ret->opID)
            {
                SGMP_V1_OperationContext context = SGMP_V1_OperationContext_Create();

                if (context)
                {
                    if (doSendRetry)
                    {
                        if ((ret->operationConfig = DOFOperationControl_Create()) != NULL)
                        {
                            DOFOperationControl_SetRetryPeriod( ret->operationConfig, retryPeriod );
                            DOFOperationManager_SetRetry( ret, channel );
                        }
                        else
                        {
                            SGMP_V1_OperationContext_Destroy( context );
                            goto failOut;
                        }
                    }
                    /* Success. */
                    context->epoch = epoch;
                    context->Tmin = Tmin;
                    ret->context = context;
                    ret->contextDestroyFunction = SGMP_V1_OperationContext_Destroy;
                    return ret;
                }
            }
        }
    failOut:
        DOFOperation_Internal_Destroy( ret );   /* ret->securityScope and ->opID clean up here. ret->context is NULL. */
    }

    return NULL;
}


DOFOperation DOFOperation_Create_Rekey( OALChannel channel, DOFAddress address, DOFSecurityScope securityScope,
                                        OALStream initialState, uint8* key, uint8 delay, uint32 timeout,
                                        boolean providerOperation )
{
    DOFOperation ret;

    PCR_Assert( securityScope );
    if ((ret = _create( DOFOPERATIONTYPE_SGMP_V1_REKEY, channel, address, timeout, providerOperation )) != NULL)
    {
        if ((ret->securityScope = DOFSecurityScope_Copy( DOF_GetDefaultSecurityScope( DOF_GetDof() ) )) != NULL)
        {
            SGMP_V1_OperationContext context = SGMP_V1_OperationContext_Create();

            if (context)
            {
                if ((context->key = DOFMem_Alloc( SGMP_V1_KEY_SIZE )) != NULL)
                {
                    /* Success. */
                    PCRMem_Copy( context->key, key, SGMP_V1_KEY_SIZE );
                    context->delay = delay;
                    context->initialState = initialState;
                    OALStream_Retain( initialState );
                    ret->context = context;
                    ret->contextDestroyFunction = SGMP_V1_OperationContext_Destroy;
                    return ret;
                }
                SGMP_V1_OperationContext_Destroy( context );
            }
        }
        DOFOperation_Internal_Destroy( ret );     /* ret->securityScope cleans up here. ret->context is NULL. */
     }

    return NULL;
}


DOFOperation DOFOperation_Create_RekeyEpoch( OALChannel channel, DOFAddress address, DOFSecurityScope securityScope,
                                             uint8* key, OALStream initialState, uint8 delay, uint16 epoch,
                                             uint32 timeout, boolean providerOperation )
{
    DOFOperation ret;

    PCR_Assert( securityScope );
    if ((ret = _create( DOFOPERATIONTYPE_SGMP_V1_REKEYEPOCH, channel, address, timeout, providerOperation )) != NULL)
    {
        if ((ret->securityScope = DOFSecurityScope_Copy( DOF_GetDefaultSecurityScope( DOF_GetDof() ) )) != NULL)
        {
            SGMP_V1_OperationContext context = SGMP_V1_OperationContext_Create();

            if (context)
            {
                if ((context->key = DOFMem_Alloc( SGMP_V1_KEY_SIZE )) != NULL)
                {
                    /* Success. */
                    PCRMem_Copy( context->key, key, SGMP_V1_KEY_SIZE );
                    context->delay = delay;
                    context->epoch = epoch;
                    context->initialState = initialState;
                    OALStream_Retain( initialState );
                    ret->context = context;
                    ret->contextDestroyFunction = SGMP_V1_OperationContext_Destroy;
                    return ret;
                }
                SGMP_V1_OperationContext_Destroy( context );
            }
        }
        DOFOperation_Internal_Destroy( ret );     /* ret->securityScope cleans up here. ret->context is NULL. */
    }

    return NULL;
}


DOFOperation DOFOperation_Create_RekeyMerge( OALChannel channel, DOFAddress address, DOFSecurityScope securityScope,
                                             PCREvent event,  uint8* key, OALStream initialState, uint32 timeout,
                                             boolean providerOperation )
{
    DOFOperation ret;

    PCR_Assert( securityScope );
    PCR_Assert( providerOperation || event );
    if ((ret = _create( DOFOPERATIONTYPE_SGMP_V1_REKEYMERGE, channel, address, timeout, providerOperation )) != NULL)
    {
        if ((ret->securityScope = DOFSecurityScope_Copy( DOF_GetDefaultSecurityScope( DOF_GetDof() ) )) != NULL)
        {
            SGMP_V1_OperationContext context = SGMP_V1_OperationContext_Create();

            if (context)
            {
                if ((context->key = DOFMem_Alloc( SGMP_V1_KEY_SIZE )) != NULL)
                {
                    /* Success. */
                    PCRMem_Copy( context->key, key, SGMP_V1_KEY_SIZE );
                    context->keyChangedEvent = event;                                 /* No refCount to bump. */
                    context->initialState = initialState;
                    OALStream_Retain( initialState );
                    ret->context = context;
                    ret->contextDestroyFunction = SGMP_V1_OperationContext_Destroy;
                    return ret;
                }
                SGMP_V1_OperationContext_Destroy( context );
            }
        }
        DOFOperation_Internal_Destroy( ret );     /* ret->securityScope cleans up here. ret->context is NULL. */
    }

    return NULL;
}


DOFOperation DOFOperation_Create_RequestGroup( OALChannel channel, DOFAddress address, DOFSecurityScope securityScope,
                                               OALStream streamI, DOFKeyRequest keyRequest, DOFObjectIDDomain domain,
                                               OALStream I_Rec_Stream, uint32 IOffset, uint32 scopeOffset, SGMP_V1_RequestGroupCallback_fn callbackFn,
                                               void * rgSendCtx, uint16 epoch, uint32 timeout,
                                               boolean providerOperation, PCRMutex reqGRespMutex )
{
    DOFOperation ret;

    PCR_Assert( securityScope );
    if ((ret = _create( DOFOPERATIONTYPE_SGMP_V1_REQUESTGROUP, channel, address, timeout, providerOperation )) != NULL)
    {
        ret->opID = DOFOperationID_CreateLocal();
        if (ret->opID)
        {
            SGMP_V1_OperationContext context = SGMP_V1_OperationContext_Create();

            if (context)
            {
                ret->securityScope = DOFSecurityScope_Copy( DOF_GetDefaultSecurityScope( DOF_GetDof() ) );
                if (ret->securityScope &&
                    (rgSendCtx ||          /* Send does not use keyRequest parameter, which can be NULL. */
                     (keyRequest && ((context->keyRequest = DOFKeyRequest_Retain( keyRequest )) != NULL))))
                {
                    if (providerOperation)
                    {
                        PCR_Assert( domain );
                        context->sgmpdomain = DOFObjectIDDomain_Copy( domain );
                    }
                    if (!providerOperation || context->sgmpdomain != NULL)
                    {
                        /* Success. */
                        if (streamI)
                        {
                            context->sgmpStreamI = streamI;
                            OALStream_Retain( context->sgmpStreamI );
                        }
                        if (I_Rec_Stream)
                        {
                            PCR_Assert( providerOperation );
                            context->I_Rec_Stream = I_Rec_Stream;                   /* Here when we receive a ReqG PDU. */
                            OALStream_Retain( context->I_Rec_Stream );
                            context->I_Offset = IOffset;
                            context->Scope_Offset = scopeOffset;
                        }
                        context->requestGroupContext = rgSendCtx;
                        context->requestGroupResponseMutex = reqGRespMutex;
                        context->callbackFn = callbackFn;
                        context->epoch = epoch;
                        ret->context = context;
                        ret->contextDestroyFunction = SGMP_V1_OperationContext_Destroy;
                        return ret;
                    }
                }
                SGMP_V1_OperationContext_Destroy( context );
            }
        }
        DOFOperation_Internal_Destroy( ret );    /* ret->opID && ->securityScope clean up here. ret->context is NULL. */
    }

    return NULL;
}
