/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/connection_internal.h>
#include <internal/marshal.h>
#include <internal/operationinternal.h>
#include <internal/router.h>
#include <internal/securegroupinternal.h>
#include <internal/sgmp_v1.h>
#include "sgmp_v1_operation.h"
#include "sgmp_v1_receive.h"
#include "sgmp_v1_receiveutil.h"
#include "sgmp_v1_sm.h"


#define SGMP_V1_RECEIVE_OP_TIMEOUT 1000


static OALAVLTreeList_Traverse_All_Return_Type_fn_t SGMP_V1_findChannelForRequestGroup(OALAVLTreeListItem_t item, void *context, boolean newKey)
{
    PCR_Assert( item );
    /* See if this is channel where the domainID and groupID both match, then see if the channel is a secureGroup. */
    if (DOFObjectID_Compare(DOFObjectIDDomain_GetObjectID(DOFCredentials_GetDomainID(((OALChannel)item)->remoteChannelData->config.connection->credentials)),
                            DOFObjectIDDomain_GetObjectID(((SGMP_V1_ReceiveData)context)->domain)) == 0)
    {
        if (DOFObjectIDAuthentication_IsEqual( DOFSecureGroup_GetGroupID( (DOFSecureGroup)item ),
                                               DOFCredentials_GetIdentity(
                                                            ((SGMP_V1_ReceiveData)context)->keyRequest->credentials ) ))
        {
            if (OALChannel_IsSecureGroup( (OALChannel)item ))
            {
                ((SGMP_V1_ReceiveData)context)->secureGroupChannel = OALChannel_Retain( (OALChannel)item );
                return (((SGMP_V1_ReceiveData)context)->secureGroupChannel!=NULL)? OALAVLTREELIST_TERMINATE_TRAVERSE : OALAVLTREELIST_NORMAL;
            }
        }
    }

    return OALAVLTREELIST_NORMAL;
}


static void SGMP_V1_ReceiveReserved(SGMP_V1_ReceiveData data)
{
    /* This should not happen, but just exit if it does and it will get dropped */
}


static void SGMP_V1_ReceiveOpClosing(OALObserver self, OALObservable observable, void* ctx)
{
    DOFOperation op = (DOFOperation)ctx;
    DOFOperation_Lock(op);
    if ((SGMP_V1_OperationContext)op->context)
    {
        SGMP_V1_OperationContext_Destroy((SGMP_V1_OperationContext)op->context);
        op->context = NULL;
    }
    DOFOperation_Unlock(op);
    DOFOperation_Internal_Destroy(op);
}


ProtocolLayerResult SGMP_V1_Receive(SGMP_V1 self)
{
    /* Note: this list is in opcode order.  */
    static const SGMP_V1_ReceiveFunction functions[SGMP_V1_HIGHEST_OPCODE + 1][2] = {
            /* 0 */ {SGMP_V1_ReceiveReserved,                   SGMP_V1_ReceiveHeartbeat},
            /* 1 */ {SGMP_V1_ReceiveReserved,                   SGMP_V1_ReceiveEpochChanged},
            /* 2 */ {SGMP_V1_ReceiveReserved,                   SGMP_V1_ReceiveRekey},
            /* 3 */ {SGMP_V1_ReceiveRequestGroupResponse,       SGMP_V1_ReceiveRequestGroup},
            /* 4 */ {SGMP_V1_ReceiveReserved,                   SGMP_V1_ReceiveReserved},
            /* 5 */ {SGMP_V1_ReceiveReserved,                   SGMP_V1_ReceiveRekeyEpoch},
            /* 6 */ {SGMP_V1_ReceiveReserved,                   SGMP_V1_ReceiveReserved},
            /* 7 */ {SGMP_V1_ReceiveReserved,                   SGMP_V1_ReceiveRekeyMerge}
    };

    /* Note: It is legal that SGMP PDUs are sent without opid. It is also legal that they come in with same opid as
       previously received (like when send operations retry and simply bump the retry count). We do not check pid's. */
    if (self->blackboardData->receiveStream)
    {
        SGMP_V1_ReceiveData_t receiveData;
        uint8                 firstByte;
        uint32                size  = OALStream_GetDataSize( self->blackboardData->receiveStream );
        uint32                value = OALMarshal_UncompressValue( 2, &size,
                                                            OALStream_GetReadPtr(self->blackboardData->receiveStream ));

        OALStream_AdjustReadOffset( self->blackboardData->receiveStream, size );
        if ((value == 0xFFFFFFFF) || (size > 2))
            return PLRESULT_ERROR;
        if (value != SGMP_V1_APPID)
            return PLRESULT_ERROR;

        if (!OALStream_Read8(self->blackboardData->receiveStream, &firstByte))
            return PLRESULT_ERROR;
        if (firstByte > SGMP_V1_HIGHEST_OPCODE)
            return PLRESULT_ERROR;

        PCR_Assert( self->blackboardData->associatedChannel );            /* If none, why did protocol stack call us? */
        if (firstByte != SGMP_V1_REQUESTGROUP_OPCODE)
        {
            /* This is some SGMP PDU other than ReqG or ReqG response, and all these others all secure. */
            /* self->blackboardData->associatedChannel was transformed in EPP_V2_Receive to be intended channel rather
               than the channel our PDU came in on, such as server channel. For SGMP secure PDUs, we get here only by
               passing through OALCipher_Decipher(), so it is highly unlikely that we will fail this next little test.*/
            DOFConnectionType connType = DOFConnectionConfig_GetConnectionType(
                                        self->blackboardData->associatedChannel->remoteChannelData->config.connection );

            if (connType != DOFCONNECTIONTYPE_GROUP && connType != DOFCONNECTIONTYPE_GROUP_MEMBER_ONLY &&
                connType != DOFCONNECTIONTYPE_HUB && connType != DOFCONNECTIONTYPE_POINT)
               return PLRESULT_ERROR;
            /* This check might be considered optional, but it allows the individual handlers to assert on it. */
        }
        SGMP_V1_Receive_InitData(&receiveData, self->blackboardData->receiveStream, self);
        PCR_Assert( receiveData.status == SGMP_V1_ReceiveStatus_OK );
        functions[firstByte][self->blackboardData->flags.isCommandNotResponse](&receiveData);
        SGMP_V1_Receive_CleanupData(&receiveData);
        switch(receiveData.status)
        {
        case SGMP_V1_ReceiveStatus_OK:
        case SGMP_V1_ReceiveStatus_OK_DONOTHING:
            return PLRESULT_OKDONOTHING;
        break;

        case SGMP_V1_ReceiveStatus_STREAMERROR:
        case SGMP_V1_ReceiveStatus_MEMORYERROR:
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "SGMP V1: _V1_Receive() had a memory or stream error");
        default:
            return PLRESULT_ERROR;
        break;
        }
    }
    return PLRESULT_ERROR;
}


void SGMP_V1_ReceiveHeartbeat(SGMP_V1_ReceiveData data)
{
    static const SGMP_V1_ReceiveUtils functions[] = {SGMP_V1_Receive_ReadAppID,
                                                     SGMP_V1_Receive_ReadDesire,
                                                     SGMP_V1_Receive_ReadTieBreaker, NULL};

    if (data->sgmp->blackboardData->flags.receivedAuthenticated)
    {
        PCR_Assert( data->sgmp->blackboardData->associatedChannel );   /* The secureGroup channel to accept this PDU. */
        data->functions = functions;
        if (SGMP_V1_Receive_IterateFunctions( data ))
        {
            DOFSecurityScope scope =
                       DOFSecureGroup_GetSecurityScope( (DOFSecureGroup)data->sgmp->blackboardData->associatedChannel );

            if (scope)
            {
                DOFConnectionType connType = DOFConnectionConfig_GetConnectionType(
                                  data->sgmp->blackboardData->associatedChannel->remoteChannelData->config.connection );

                PCR_Assert( connType == DOFCONNECTIONTYPE_GROUP || connType == DOFCONNECTIONTYPE_GROUP_MEMBER_ONLY ||
                            connType == DOFCONNECTIONTYPE_HUB || connType == DOFCONNECTIONTYPE_POINT );
                data->operation = DOFOperation_Create_Heartbeat( data->sgmp->blackboardData->associatedChannel,
                                                                 data->sgmp->blackboardData->transportAddr, scope,
                                                                 data->tieBreaker, data->appid, data->desire,
                                                                 SGMP_V1_RECEIVE_OP_TIMEOUT, TRUE );
                if (data->operation)
                {
                    SGMP_V1_SM_ProcessToResponder(data->operation);
                    DOFOperation_Internal_Destroy(data->operation);
                    data->operation = NULL;
                }
                else
                {
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "SGMP V1: failed to create sgmp heartbeat op");
                    data->status = SGMP_V1_ReceiveStatus_MEMORYERROR;
                }
            }
            else
                data->status = SGMP_V1_ReceiveStatus_OK_DONOTHING;
        }
        else
            data->status = SGMP_V1_ReceiveStatus_OK_DONOTHING;                               /* Basic PDU tests fail. */
    }
    else
        data->status = SGMP_V1_ReceiveStatus_OK_DONOTHING;                     /* Ignore heartbeat received unsecure. */
}


void SGMP_V1_ReceiveEpochChanged(SGMP_V1_ReceiveData data)
{
    static const SGMP_V1_ReceiveUtils functions[] = {SGMP_V1_Receive_ReadTmin,
                                                     SGMP_V1_Receive_ReadEpoch, NULL};

    if (data->sgmp->blackboardData->flags.receivedAuthenticated)
    {
        data->functions = functions;
        if (SGMP_V1_Receive_IterateFunctions( data ))
        {
            DOFSecurityScope scope =
                       DOFSecureGroup_GetSecurityScope( (DOFSecureGroup)data->sgmp->blackboardData->associatedChannel );

            if (scope)
            {
                DOFConnectionType connType = DOFConnectionConfig_GetConnectionType(
                                  data->sgmp->blackboardData->associatedChannel->remoteChannelData->config.connection );

                PCR_Assert( connType == DOFCONNECTIONTYPE_GROUP || connType == DOFCONNECTIONTYPE_GROUP_MEMBER_ONLY ||
                            connType == DOFCONNECTIONTYPE_HUB || connType == DOFCONNECTIONTYPE_POINT );
                data->operation = DOFOperation_Create_EpochChanged(data->sgmp->blackboardData->associatedChannel,
                                                                   data->sgmp->blackboardData->transportAddr, scope,
                                                                   data->epoch, data->Tmin, SGMP_V1_RECEIVE_OP_TIMEOUT,
                                                                   TRUE, 0, 0);
                if (data->operation)
                {
                    SGMP_V1_SM_ProcessToResponder(data->operation);
                    DOFOperation_Internal_Destroy(data->operation);
                    data->operation = NULL;
                }
                else
                {
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "SGMP V1: failed to create kek changed op");
                    data->status = SGMP_V1_ReceiveStatus_MEMORYERROR;
                }
            }
        }
    }
    else
    {
        /* If a epoch changed is received unsecure ignore it. */
        data->status = SGMP_V1_ReceiveStatus_OK_DONOTHING;
    }
}


void SGMP_V1_ReceiveRekey(SGMP_V1_ReceiveData data)
{
    static const SGMP_V1_ReceiveUtils functions[] = {SGMP_V1_Receive_ReadDelay,
                                                     SGMP_V1_Receive_ReadInitialState,
                                                     SGMP_V1_Receive_ReadKey, NULL};

    if (data->sgmp->blackboardData->flags.receivedAuthenticated && data->sgmp->blackboardData->flags.receivedEncrypted)
    {
        data->functions = functions;
        if (SGMP_V1_Receive_IterateFunctions( data ))
        {
            DOFSecurityScope scope =
                       DOFSecureGroup_GetSecurityScope( (DOFSecureGroup)data->sgmp->blackboardData->associatedChannel );

            if (scope)
            {
                DOFConnectionType connType = DOFConnectionConfig_GetConnectionType(
                                  data->sgmp->blackboardData->associatedChannel->remoteChannelData->config.connection );

                PCR_Assert( connType == DOFCONNECTIONTYPE_GROUP || connType == DOFCONNECTIONTYPE_GROUP_MEMBER_ONLY ||
                            connType == DOFCONNECTIONTYPE_HUB || connType == DOFCONNECTIONTYPE_POINT );
                data->operation = DOFOperation_Create_Rekey(data->sgmp->blackboardData->associatedChannel,
                                                            data->sgmp->blackboardData->transportAddr, scope,
                                                            data->initialState, data->key,
                                                            data->delay, SGMP_V1_RECEIVE_OP_TIMEOUT, TRUE);

                if (data->operation)
                {
                    ((SGMP_V1_OperationContext)data->operation->context)->receivedMulticast =
                                                                          data->sgmp->blackboardData->flags.isMulticast;
                    SGMP_V1_SM_ProcessToResponder(data->operation);
                    DOFOperation_Internal_Destroy(data->operation);
                    data->operation = NULL;
                }
                else
                {
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "SGMP V1: failed to create rekeay op");
                    data->status = SGMP_V1_ReceiveStatus_MEMORYERROR;
                }
            }
        }
    }
    else
    {
        /* If a rekey is received unsecure ignore it. */
        data->status = SGMP_V1_ReceiveStatus_OK_DONOTHING;
    }
}


void SGMP_V1_ReceiveRekeyEpoch(SGMP_V1_ReceiveData data)
{
    static const SGMP_V1_ReceiveUtils functions[] = {SGMP_V1_Receive_ReadDelay,
                                                     SGMP_V1_Receive_ReadInitialState,
                                                     SGMP_V1_Receive_ReadEpoch,
                                                     SGMP_V1_Receive_ReadKey, NULL};

    if (data->sgmp->blackboardData->flags.receivedAuthenticated && data->sgmp->blackboardData->flags.receivedEncrypted)
    {
        data->functions = functions;
        if (SGMP_V1_Receive_IterateFunctions( data ))
        {
            DOFSecurityScope scope =
                       DOFSecureGroup_GetSecurityScope( (DOFSecureGroup)data->sgmp->blackboardData->associatedChannel );

            if (scope)
            {
                DOFConnectionType connType = DOFConnectionConfig_GetConnectionType(
                                  data->sgmp->blackboardData->associatedChannel->remoteChannelData->config.connection );

                PCR_Assert( connType == DOFCONNECTIONTYPE_GROUP || connType == DOFCONNECTIONTYPE_GROUP_MEMBER_ONLY ||
                            connType == DOFCONNECTIONTYPE_HUB || connType == DOFCONNECTIONTYPE_POINT );
                data->operation = DOFOperation_Create_RekeyEpoch( data->sgmp->blackboardData->associatedChannel,
                                                                  data->sgmp->blackboardData->transportAddr, scope,
                                                                  data->key, data->initialState, data->delay,
                                                                  data->epoch, SGMP_V1_RECEIVE_OP_TIMEOUT, TRUE );
                if (data->operation)
                {
                    ((SGMP_V1_OperationContext)data->operation->context)->receivedMulticast =
                                                                          data->sgmp->blackboardData->flags.isMulticast;
                    SGMP_V1_SM_ProcessToResponder(data->operation);
                    DOFOperation_Internal_Destroy(data->operation);
                    data->operation = NULL;
                }
                else
                {
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "SGMP V1: failed to create rekey kek op");
                    data->status = SGMP_V1_ReceiveStatus_MEMORYERROR;
                }
            }
        }
    }
    else
    {
        /* If a rekey epoch is received unsecure ignore it. */
        data->status = SGMP_V1_ReceiveStatus_OK_DONOTHING;
    }
}


void SGMP_V1_ReceiveRekeyMerge(SGMP_V1_ReceiveData data)
{
    static const SGMP_V1_ReceiveUtils functions[] = {SGMP_V1_Receive_ReadInitialState,
                                                     SGMP_V1_Receive_ReadKey, NULL};

    if (data->sgmp->blackboardData->flags.receivedAuthenticated && data->sgmp->blackboardData->flags.receivedEncrypted)
    {
        data->functions = functions;
        if (SGMP_V1_Receive_IterateFunctions( data ))
        {
            DOFSecurityScope scope =
                       DOFSecureGroup_GetSecurityScope( (DOFSecureGroup)data->sgmp->blackboardData->associatedChannel );

            if (scope)
            {
                DOFConnectionType connType = DOFConnectionConfig_GetConnectionType(
                                  data->sgmp->blackboardData->associatedChannel->remoteChannelData->config.connection );

                PCR_Assert( connType == DOFCONNECTIONTYPE_GROUP || connType == DOFCONNECTIONTYPE_GROUP_MEMBER_ONLY ||
                            connType == DOFCONNECTIONTYPE_HUB || connType == DOFCONNECTIONTYPE_POINT );
                data->operation = DOFOperation_Create_RekeyMerge( data->sgmp->blackboardData->associatedChannel,
                                                                  data->sgmp->blackboardData->transportAddr, scope,
                                                                  NULL, data->key, data->initialState,
                                                                  SGMP_V1_RECEIVE_OP_TIMEOUT, TRUE );

                if (data->operation)
                {
                    ((SGMP_V1_OperationContext)data->operation->context)->receivedMulticast =
                                                                          data->sgmp->blackboardData->flags.isMulticast;
                    SGMP_V1_SM_ProcessToResponder(data->operation);
                    DOFOperation_Internal_Destroy(data->operation);
                    data->operation = NULL;
                }
                else
                {
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "SGMP V1: failed to create rekey merge op");
                    data->status = SGMP_V1_ReceiveStatus_MEMORYERROR;
                }
            }
        }
    }
    else
    {
        /* If a rekey merge is received unsecure ignore it. */
        data->status = SGMP_V1_ReceiveStatus_OK_DONOTHING;
    }
}


void SGMP_V1_ReceiveRequestGroup( SGMP_V1_ReceiveData data )
{
    static const SGMP_V1_ReceiveUtils functions[] = {SGMP_V1_Receive_VerifyOpID,
                                                     SGMP_V1_Receive_VerifyNewOpID,
                                                     SGMP_V1_Receive_ReadDomain,
                                                     SGMP_V1_Receive_ReadEpoch,
                                                     SGMP_V1_Receive_ReadKeyRequestBlock,
                                                     NULL};

    PCR_Assert( data );
    PCR_Assert( data->sgmp );
    PCR_Assert( data->sgmp->blackboardData );
    PCR_Assert( data->sgmp->blackboardData->associatedChannel );
    PCR_Assert( data->sgmp->blackboardData->associatedChannel->remoteChannelData);
    PCR_Assert( data->status == SGMP_V1_ReceiveStatus_OK );
    if (data->sgmp->blackboardData->flags.receivedUnsecure)
    {
        data->functions = functions;
        /* Have no length byte for our incoming I block, so we iterate our way to fill data-> members. */
        if (SGMP_V1_Receive_IterateFunctions(data))
        {
            OALAVLTreeList registeredGroups;

            PCR_Assert( DOFCredentials_GetDomainID( data->keyRequest->credentials ) != NULL ); /*DOFOBJECTID_BROADCAST*/
            /* Before the request group operation is created see if the intended group exists, in two steps. */
            if (data->sgmp->blackboardData->associatedChannel->remoteChannelData->baseChannel)
                registeredGroups = data->sgmp->blackboardData->associatedChannel->remoteChannelData->baseChannel->remoteChannelData->registeredSSIDs_AS;
            else
                registeredGroups = data->sgmp->blackboardData->associatedChannel->remoteChannelData->registeredSSIDs_AS;

            if (registeredGroups)
            {
                /* Traverse entire tree of registered channels to see if there is a secureGroup channel to send PDU to*/
                OALAVLTreeList_TraverseAll( registeredGroups,
                                            SGMP_V1_findChannelForRequestGroup, data );
                if (data->secureGroupChannel)
                {
                    DOFSecurityScope scope = DOFSecureGroup_GetSecurityScope((DOFSecureGroup)data->secureGroupChannel);

                    if (scope)
                    {

                        /* Send ReqG PDU to the found secureGroup channel. */
                        DOFOperation reqGOp = DOFOperation_Create_RequestGroup(
                                                                          data->sgmp->blackboardData->associatedChannel,
                                                                          data->sgmp->blackboardData->transportAddr,
                                                                          scope, NULL, data->keyRequest, data->domain,
                                                                          data->stream, data->I_BlockOffset, OALStream_GetReadOffset(data->stream),
                                                                          NULL, NULL, data->epoch, 2000, TRUE, NULL );

                        if (reqGOp)
                        {
                            boolean success = FALSE;

                            PCR_Assert( reqGOp->context );
                            PCR_Assert( reqGOp->operation_will_close );
                            /* Use opID held in blackboardData. */
                            DOFOperationID_Destroy( reqGOp->opID );
                            reqGOp->opID = DOFOperationID_Copy( data->sgmp->blackboardData->opID );
                            if (!data->sgmp->blackboardData->opID     /* Is it legal to be NULL? We allow it for now. */
                                || reqGOp->opID)
                            {
                                if (DOFOperation_AddOpWillCloseObserver( reqGOp,                  /* bumps ref count. */
                                                                         SGMP_V1_ReceiveOpClosing ) != NULL)
                                {
                                    /* nobody else knows about operation so don't need lock */
                                    if (!reqGOp->opID ||               /* When no opID, _AddOperation() does nothing. */
                                        OALOperationManager_AddOperation( reqGOp, reqGOp->opID,
                                                                          FALSE ) == DOFERROR_OK)
                                    {
                                        /* Act on incoming ReqG PDU. */
                                        SGMP_V1_OperationContext ctx = (SGMP_V1_OperationContext)reqGOp->context;

                                        PCR_Assert( ctx );
                                        ctx->secureGroupChannel = data->secureGroupChannel;
                                        data->secureGroupChannel = NULL;    /* Allow operation to take ownership. */
                                        SGMP_V1_SM_ProcessToResponder( reqGOp );
                                        PCR_Assert( data->status == SGMP_V1_ReceiveStatus_OK );
                                        success = TRUE;
                                    }
                                    else
                                        DOFOperation_Cancel( reqGOp );    /* Cause observer to be called for cleanup. */
                                }
                            }
                            DOFOperation_Internal_Destroy( reqGOp );
                            if (success)
                                return;
                        }
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE))
                            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "SGMP V1: incoming reqGroup PDU, failed to create op" );
                        data->status = SGMP_V1_ReceiveStatus_MEMORYERROR;
                        return;
                    }
                    else
                    {
                        /* No scope in secureGroup's opKEK. */
                        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE))
                            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "SGMP V1: reject incoming reqGroup PDU, no scope held" );
                    }
                }
                else
                { /* No matching secureGroup channel to send to. Silently drop the PDU. */
                }
            }
            else
            { /* No secureGroup channels registered with this channel. Silently drop the PDU. */
            }
        }
        else
        {
            if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE))
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "SGMP V1: reject incoming request group PDU at parse" );
            data->status = SGMP_V1_ReceiveStatus_STREAMERROR;        /* Or _MEMORYERROR- either may be at fault here. */
            return;
        }
    }
    else
    {
        /* If a request group is received secure ignore it. */
        /* TODO: It is now possible to receive a secure request group, but what should happen when we do? */
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE))
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "SGMP V1: drop incoming request group PDU because it was sent secure" );
    }
    data->status = SGMP_V1_ReceiveStatus_OK_DONOTHING;
}


void SGMP_V1_ReceiveRequestGroupResponse(SGMP_V1_ReceiveData data)
{
    static const SGMP_V1_ReceiveUtils functions[] = {SGMP_V1_Receive_VerifyOpID,
                                                     SGMP_V1_Receive_RetrieveOperationFromRouter,
                                                     SGMP_V1_Receive_ReadInitialState,
                                                     SGMP_V1_Receive_ReadAppID,
                                                     SGMP_V1_Receive_ReadDesire,
                                                     SGMP_V1_Receive_ReadTicket, NULL};

    PCR_Assert( data->status == SGMP_V1_ReceiveStatus_OK );
    if (data->sgmp->blackboardData->flags.receivedUnsecure)
    {
        data->functions = functions;
        if (SGMP_V1_Receive_IterateFunctions( data ) && data->operation && data->operation->context)
        {
            SGMP_V1_OperationContext context = (SGMP_V1_OperationContext)data->operation->context;

            PCR_Assert( context );
            /* Since we can receive multiple request group responses clear out anything from the previous response*/
            OALKeyGrant_Release( context->ticket );
            OALStream_Release( context->initialState, TRUE );

            /* Use address held in blackboardData. */
            DOFAddress_Destroy(data->operation->sources->address);
            data->operation->sources->address = DOFAddress_Copy( data->sgmp->blackboardData->transportAddr );
            if (data->operation->sources->address || !data->sgmp->blackboardData->transportAddr)
            {
                context->ticket = data->ticket;
                OALKeyGrant_Retain( context->ticket );
                context->initialState = data->initialState;
                OALStream_Retain(context->initialState);
                context->receivedMulticast = data->sgmp->blackboardData->flags.isMulticast;
                context->appid = data->appid;
                context->desire = data->desire;
                SGMP_V1_SM_ProcessToResponder( data->operation );
                PCR_Assert( data->status == SGMP_V1_ReceiveStatus_OK );
            }
            else
                data->status = SGMP_V1_ReceiveStatus_MEMORYERROR;
            DOFOperation_Internal_Destroy( data->operation );
            data->operation = NULL;
        }
        else
            data->status = SGMP_V1_ReceiveStatus_STREAMERROR;        /* Or _MEMORYERROR- either may be at fault here. */
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE) && data->status != SGMP_V1_ReceiveStatus_OK)
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "SGMP V1: reject incoming reqG response PDU at parse" );
    }
    else
    { /* If a request group response is received secure ignore it. TODO: still true? */
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_TRACE))
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "SGMP V1: drop incoming reqG response PDU because it was sent secure" );
        data->status = SGMP_V1_ReceiveStatus_OK_DONOTHING;
    }
}
