/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/key_request.h>
#include <internal/marshal.h>
#include <internal/operationinternal.h>
#include <internal/router.h>
#include "sgmp_v1_receiveutil.h"


void SGMP_V1_Receive_InitData( SGMP_V1_ReceiveData data, OALStream stream, SGMP_V1 sgmp )
{
    PCRMem_Set( data, 0, sizeof(SGMP_V1_ReceiveData_t) ); /* Note: sets data->status to 0, SGMP_V1_ReceiveStatus_OK,. */
    data->stream = stream;
    data->sgmp = sgmp;
}


void SGMP_V1_Receive_CleanupData( SGMP_V1_ReceiveData data )
{
    DOFObjectIDDomain_Destroy( data->domain );
    DOFMem_Free( data->key );
    OALKeyGrant_Release( data->ticket );
    OALStream_Release( data->initialState, TRUE );
    DOFKeyRequest_Destroy( data->keyRequest );
    OALChannel_Release( data->secureGroupChannel );
}


boolean SGMP_V1_Receive_IterateFunctions( SGMP_V1_ReceiveData data )
{
    uint8 i = 0;

    while(data->functions[i] != NULL)
    {
        if(!data->functions[i++](data))
            return FALSE;
    }

    return TRUE;
}


boolean SGMP_V1_Receive_ReadDelay( SGMP_V1_ReceiveData data )
{
    if (!OALStream_Read8( data->stream, &(data->delay) ))
    {
        data->status = SGMP_V1_ReceiveStatus_STREAMERROR;
        return FALSE;
    }

    return TRUE;
}

boolean SGMP_V1_Receive_ReadAppID( SGMP_V1_ReceiveData data )
{
    uint32 size;
    uint32 value;

    size = OALStream_GetDataSize( data->stream );
    value = OALMarshal_UncompressValue( 2, &size, OALStream_GetReadPtr( data->stream ) );
    if ((value == 0xFFFFFFFF) || (size > 2) || /* Pure uncompress problem. */
        value < SGMP_V1_APPID)                 /* SGMP spec says sender's highest appid has this minimum. */
    {
        data->status = SGMP_V1_ReceiveStatus_STREAMERROR;
        return FALSE;
    }
    OALStream_AdjustReadOffset( data->stream, size );
    data->appid = (uint16)value;

    return TRUE;
}


boolean SGMP_V1_Receive_ReadEpoch( SGMP_V1_ReceiveData data )
{
    data->I_BlockOffset = OALStream_GetReadOffset(data->stream);
    if (!OALStream_Read16( data->stream, &(data->epoch) ))
    {
        data->status = SGMP_V1_ReceiveStatus_STREAMERROR;
        return FALSE;
    }
    
    return TRUE;
}


boolean SGMP_V1_Receive_ReadTieBreaker(SGMP_V1_ReceiveData data)
{
    if (!OALStream_Read32(data->stream, &(data->tieBreaker)))
    {
        data->status = SGMP_V1_ReceiveStatus_STREAMERROR;
        return FALSE;
    }
    return TRUE;
}


boolean SGMP_V1_Receive_ReadDesire( SGMP_V1_ReceiveData data )
{
    if (!OALStream_Read8( data->stream, &(data->desire) ))
    {
        data->status = SGMP_V1_ReceiveStatus_STREAMERROR;
        return FALSE;
    }

    return TRUE;
}


boolean SGMP_V1_Receive_ReadTmin( SGMP_V1_ReceiveData data )
{
    if (!OALStream_Read16( data->stream, &(data->Tmin) ))
    {
        data->status = SGMP_V1_ReceiveStatus_STREAMERROR;
        return FALSE;
    }

    return TRUE;
}


boolean SGMP_V1_Receive_ReadKey( SGMP_V1_ReceiveData data )
{
    data->key = DOFMem_Alloc( SGMP_V1_KEY_SIZE );

    return data->key ? OALStream_Read( data->stream, SGMP_V1_KEY_SIZE, data->key ) : FALSE;
}


boolean SGMP_V1_Receive_ReadDomain( SGMP_V1_ReceiveData data )
{
    DOFObjectID oid;
    uint32      length = OALStream_GetDataSize( data->stream );

    if ((oid = DOFObjectID_Create_Unmarshal( &length, OALStream_GetReadPtr( data->stream ) )) != NULL)
    {
        PCR_Assert( length >= 2 );                               /* Or the above call returns NULL. */
        OALStream_AdjustReadOffset( data->stream, length );
        data->domain = DOFObjectIDDomain_Create( oid );
        DOFObjectID_Destroy( oid );                              /* Cleanup 1st, then check success. */
        if (data->domain)
            return TRUE;
    }

    return FALSE;
}


boolean SGMP_V1_Receive_ReadKeyRequestBlock( SGMP_V1_ReceiveData data )
{
    data->keyRequest = DOFKeyRequest_Create( NULL, 0, FALSE, FALSE, NULL, NULL );

    if (data->keyRequest != NULL)
    {
        PCR_Assert( data->keyRequest->credentials == NULL );
        /* Instead of NULL could place actual domain, but DOFSecureGroup_ProcessRequestGroup() ignores what we place. */
        if (DOFKeyRequest_Unmarshal( data->stream, data->keyRequest, NULL, TRUE, FALSE ))
        {
            if (data->keyRequest->credentials && DOFCredentials_GetDomainID( data->keyRequest->credentials ) != NULL)
               return TRUE;                                                           /* Is DOFOBJECTID_BROADCAST. */
        }
        data->status = SGMP_V1_ReceiveStatus_STREAMERROR;
    }
    else
        data->status = SGMP_V1_ReceiveStatus_MEMORYERROR;

    return FALSE;
}


boolean SGMP_V1_Receive_ReadTicket( SGMP_V1_ReceiveData data )
{

    data->ticket = OALKeyGrant_Unmarshal( data->stream );
    if (!data->ticket)
    {
        data->status = SGMP_V1_ReceiveStatus_STREAMERROR;
        return FALSE;
    }

    return TRUE;
}


boolean SGMP_V1_Receive_ReadInitialState( SGMP_V1_ReceiveData data )
{
    uint32 initialStateSize;
    uint32 size;

    size = OALStream_GetDataSize( data->stream );
    initialStateSize = OALMarshal_UncompressValue( 2, &size, OALStream_GetReadPtr( data->stream ) );
    if ((initialStateSize == 0xFFFFFFFF) || (size > 2))
    {
        data->status = SGMP_V1_ReceiveStatus_STREAMERROR;
        return FALSE;
    }
    OALStream_AdjustReadOffset( data->stream, size );
    PCR_Assert( data->initialState == NULL );                              /* As setup by SGMP_V1_Receive_InitData(). */
    data->initialState = OALStream_CreateRW( initialStateSize );
    if (data->initialState)
    {
        if (OALStream_Write( data->initialState, initialStateSize, OALStream_GetReadPtr( data->stream ) ))
        {
            OALStream_AdjustReadOffset( data->stream, initialStateSize );
            return TRUE;
        }
    }

    return FALSE;
}


boolean SGMP_V1_Receive_VerifyOpID( SGMP_V1_ReceiveData data )
{
    if (data->sgmp->blackboardData->opID == NULL)
    {
        data->status = SGMP_V1_ReceiveStatus_STREAMERROR;
        return FALSE;
    }

    return TRUE;
}


boolean SGMP_V1_Receive_VerifyNewOpID( SGMP_V1_ReceiveData data )
{
    data->operation = OALOperationManager_CheckExistingOperation( data->domain, data->sgmp->blackboardData->opID );
    if (data->operation == NULL)
        return TRUE;                        /* Return TRUE if we do not find an existing operation: opid is NEW. */
    DOFOperation_Internal_Destroy(data->operation);
    data->status = SGMP_V1_ReceiveStatus_OK_DONOTHING;

    return FALSE;
}


boolean SGMP_V1_Receive_RetrieveOperationFromRouter( SGMP_V1_ReceiveData data )
{
    data->operation = OALOperationManager_CheckExistingOperation(
                                            DOFSecurityScope_GetDomainID( DOF_GetDefaultSecurityScope( DOF_GetDof() ) ),
                                                                  data->sgmp->blackboardData->opID );
    if (data->operation == NULL)
    {
        data->status = SGMP_V1_ReceiveStatus_MEMORYERROR;
        return FALSE;
    }

    return TRUE;
}
