/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/blackboarddata.h>
#include "sgmp_v1_send.h"


static void SGMP_V1_InvalidOpcode_Send(SGMP_V1_SendData data, BlackboardData blackboardData)
{
    data->status = SGMP_V1_SendStatus_INVALID_OPCODE;
}


static void SGMP_V1_RequestGroup_Send(SGMP_V1_SendData data, BlackboardData blackboardData)
{
    static const SGMP_V1_SendUtils functions[] = {SGMP_V1_Send_WriteFirstByte,
                                                  SGMP_V1_Send_I,
                                                  NULL};

    PCR_Assert( data );
    if (data->operation)
    {
        boolean success = TRUE;

        PCR_Assert( data->status == SGMP_V1_SendStatus_OK );
        PCR_Assert( data->operation->context );                      /* Filled in DOFOperation_Create_RequestGroup(). */
        if (!((SGMP_V1_OperationContext)(data->operation->context))->secure)
        {
            if (!blackboardData->transportAddr)
            {
                blackboardData->transportAddr =
                                   DOFAddress_Copy( blackboardData->associatedChannel->remoteChannelData->address );
                if (!blackboardData->transportAddr && blackboardData->associatedChannel->remoteChannelData->address)
                    success = FALSE;
            }
            if (blackboardData->associatedChannel->remoteChannelData->baseChannel)
            {
                OALChannel temp = blackboardData->associatedChannel;

                blackboardData->associatedChannel = OALChannel_Retain( temp->remoteChannelData->baseChannel );
                if (!blackboardData->associatedChannel && temp->remoteChannelData->baseChannel)
                    success = FALSE;
                OALChannel_Release( temp );
            }
        }

        if (success)
        {
            /* data->operation->context was filled with SGMP_V1_OperationContext. */
            /* data->context is known to compiler as SGMP_V1_OperationContext. */
            /* cast is because compiler only knows data->operation->context as void * */
            data->context = (SGMP_V1_OperationContext)data->operation->context;
            data->functions = functions;
            data->opCode = SGMP_V1_REQUESTGROUP_OPCODE;
            blackboardData->flags.streamingProtocolNotAllowed = TRUE;
            DOFOperation_Lock( data->operation );

            SGMP_V1_Send_IterateFunctions( data );
        }
        else
            data->status = SGMP_V1_SendStatus_MEMORYERROR;
        DOFOperation_Unlock( data->operation );
    }
    else
        data->status = SGMP_V1_SendStatus_TIMEOUTERROR;
}


ProtocolLayerResult SGMP_V1_Send(SGMP_V1 self, DOFOperation operation, boolean isProviderOperation, DOFResult result,
                                 PCREvent * keyChangedEvent)
{
    SGMP_V1_SendData_t sendData;

    /* Note: this is in operation type order.  */
    static const SGMP_V1_SendFunction functions[DOFOPERATIONTYPE_SGMP_LAST - DOFOPERATIONTYPE_SGMP_V1_HEARTBEAT][2] = {
    /* 0 */ {SGMP_V1_InvalidOpcode_Send,                SGMP_V1_Heartbeat_Send},
    /* 1 */ {SGMP_V1_InvalidOpcode_Send,                SGMP_V1_EpochChanged_Send},
    /* 2 */ {SGMP_V1_InvalidOpcode_Send,                SGMP_V1_Rekey_Send},
    /* 3 */ {SGMP_V1_RequestGroupResponse_Send,         SGMP_V1_RequestGroup_Send},
    /* 4 */ {SGMP_V1_InvalidOpcode_Send,                SGMP_V1_RekeyEpoch_Send},
    /* 5 */ {SGMP_V1_InvalidOpcode_Send,                SGMP_V1_RekeyMerge_Send}
    };

    PCR_Assert(self);
    PCR_Assert(operation);

    PCR_Assert(operation->operation_type >= DOFOPERATIONTYPE_SGMP_V1_HEARTBEAT);
    PCR_Assert(operation->operation_type <  DOFOPERATIONTYPE_SGMP_LAST);

    SGMP_V1_Send_Init( &sendData, self, operation) ;         /* Fills .status with 0 (SGMP_V1_SendStatus_OK). */
    PCR_Assert( sendData.status == SGMP_V1_SendStatus_OK );  /* We rely on this- nothing sets it this way after this. */

    functions[operation->operation_type - DOFOPERATIONTYPE_SGMP_V1_HEARTBEAT][isProviderOperation]
                                                                                      (&sendData, self->blackboardData);

    if (operation->operation_type == DOFOPERATIONTYPE_SGMP_V1_REQUESTGROUP && isProviderOperation &&
        self->blackboardData->seq == 0)
    {
        /* Avoid ->seq of 0- that means to a recipient that we are the source of this PDU. This can cause the recipient
           to drop our incoming PDU if it has knowledge stored that a previous PDU came in from us, with certain
           attributes. If we rebooted since that time, our new PDU is unlikely to match in every detail, and the
           recipient must drop us. To avoid this, use non-zero ->seq. This matches what we have already done in
           TRP, TEP, ESP, EPP_NegOne, and allows the recipient know we are not claiming to match those previously stored
           PDU attributes, and EPP_V2_Receive() will not be forced to drop our incoming PDU. */
        self->blackboardData->seq = 1;
    }

    if (isProviderOperation)
        self->blackboardData->flags.isCommandNotResponse = TRUE;
    else
        self->blackboardData->flags.isCommandNotResponse = FALSE;
    if (sendData.context)
        *keyChangedEvent = ((SGMP_V1_OperationContext)(sendData.context))->keyChangedEvent;
    self->blackboardData->associatedChannel->remoteChannelData->flags.isSSIDAssignedFromAS = 1;
    self->blackboardData->flags.ifProperPlaceNextBaseNIDOnWire = 1;
    switch(sendData.status)
    {
    case SGMP_V1_SendStatus_OK:
        self->blackboardData->opID = DOFOperationID_Copy(sendData.operation->opID);
    case SGMP_V1_SendStatus_OKNOOPID:
        self->blackboardData->delay = sendData.operation->sources->oal_timeout;
        return PLRESULT_OKSENDDOWN;
    break;

    case SGMP_V1_SendStatus_MEMORYERROR:
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "SGMP V1: Out of memory while trying to send.");
        return PLRESULT_ERROR;
    case SGMP_V1_SendStatus_STREAMERROR:
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "SGMP V1: Had an issue writing to a stream while trying to send.");
        return PLRESULT_ERROR;
    case SGMP_V1_SendStatus_TIMEOUTERROR:
    case SGMP_V1_SendStatus_INVALID_OPCODE:
    default:
        return PLRESULT_ERROR;
    break;
    }
}


void SGMP_V1_Heartbeat_Send(SGMP_V1_SendData data, BlackboardData blackboardData)
{
    static const SGMP_V1_SendUtils functions[] = {SGMP_V1_Send_WriteFirstByte,
                                                  SGMP_V1_Send_WriteAppID,
                                                  SGMP_V1_Send_WriteDesire,
                                                  SGMP_V1_Send_WriteTieBreaker, NULL};

    if (data->operation)
    {
        DOFOperation_Lock(data->operation);
        if (data->operation->context)
        {
            data->context = (SGMP_V1_OperationContext) data->operation->context;
            data->functions = functions;
            data->opCode = SGMP_V1_HEARTBEAT_OPCODE;
            blackboardData->flags.securityRequiredAuthenticate = TRUE;
            blackboardData->flags.streamingProtocolNotAllowed = TRUE;

            SGMP_V1_Send_IterateFunctions(data);
        }
        else
            data->status = SGMP_V1_SendStatus_TIMEOUTERROR;

        DOFOperation_Unlock(data->operation);
    }
    else
        data->status = SGMP_V1_SendStatus_TIMEOUTERROR;
}


void SGMP_V1_EpochChanged_Send(SGMP_V1_SendData data, BlackboardData blackboardData)
{
    static const SGMP_V1_SendUtils functions[] = {SGMP_V1_Send_WriteFirstByte,
                                                  SGMP_V1_Send_WriteTmin,
                                                  SGMP_V1_Send_WriteEpoch, NULL};

    if (data->operation)
    {
        DOFOperation_Lock(data->operation);
        if (data->operation->context)
        {
            data->context = (SGMP_V1_OperationContext) data->operation->context;
            data->functions = functions;
            data->opCode = SGMP_V1_EPOCHCHANGED_OPCODE;
            blackboardData->flags.securityRequiredAuthenticate = TRUE;
            blackboardData->flags.streamingProtocolNotAllowed = TRUE;

            SGMP_V1_Send_IterateFunctions(data);
        }
        else
            data->status = SGMP_V1_SendStatus_TIMEOUTERROR;

        DOFOperation_Unlock(data->operation);
    }
    else
        data->status = SGMP_V1_SendStatus_TIMEOUTERROR;
}


void SGMP_V1_Rekey_Send(SGMP_V1_SendData data, BlackboardData blackboardData)
{
    static const SGMP_V1_SendUtils functions[] = {SGMP_V1_Send_WriteFirstByte,
                                                  SGMP_V1_Send_WriteDelay,
                                                  SGMP_V1_Send_WriteInitialState,
                                                  SGMP_V1_Send_WriteKey, NULL};

    if (data->operation)
    {
        DOFOperation_Lock(data->operation);
        if (data->operation->context)
        {
            data->context = (SGMP_V1_OperationContext) data->operation->context;
            data->functions = functions;
            data->opCode = SGMP_V1_REKEY_OPCODE;
            blackboardData->flags.securityRequiredAuthenticate = TRUE;
            blackboardData->flags.securityRequiredEncrypt = TRUE;
            blackboardData->flags.streamingProtocolNotAllowed = TRUE;

            SGMP_V1_Send_IterateFunctions(data);
        }
        else
            data->status = SGMP_V1_SendStatus_TIMEOUTERROR;

        DOFOperation_Unlock(data->operation);
    }
    else
        data->status = SGMP_V1_SendStatus_TIMEOUTERROR;
}


void SGMP_V1_RekeyEpoch_Send(SGMP_V1_SendData data, BlackboardData blackboardData)
{
    static const SGMP_V1_SendUtils functions[] = {SGMP_V1_Send_WriteFirstByte,
                                                  SGMP_V1_Send_WriteDelay,
                                                  SGMP_V1_Send_WriteInitialState,
                                                  SGMP_V1_Send_WriteEpoch,
                                                  SGMP_V1_Send_WriteKey, NULL};

    if (data->operation)
    {
        DOFOperation_Lock(data->operation);
        if (data->operation->context)
        {
            data->context = (SGMP_V1_OperationContext) data->operation->context;
            data->functions = functions;
            data->opCode = SGMP_V1_REKEYEPOCH_OPCODE;
            blackboardData->flags.securityRequiredAuthenticate = TRUE;
            blackboardData->flags.securityRequiredEncrypt = TRUE;
            blackboardData->flags.streamingProtocolNotAllowed = TRUE;

            SGMP_V1_Send_IterateFunctions(data);
        }
        else
            data->status = SGMP_V1_SendStatus_TIMEOUTERROR;

        DOFOperation_Unlock(data->operation);
    }
    else
        data->status = SGMP_V1_SendStatus_TIMEOUTERROR;
}


void SGMP_V1_RekeyMerge_Send(SGMP_V1_SendData data, BlackboardData blackboardData)
{
    static const SGMP_V1_SendUtils functions[] = {SGMP_V1_Send_WriteFirstByte,
                                                  SGMP_V1_Send_WriteInitialState,
                                                  SGMP_V1_Send_WriteKey, NULL};

    if (data->operation)
    {
        DOFOperation_Lock(data->operation);
        if (data->operation->context)
        {
            data->context = (SGMP_V1_OperationContext) data->operation->context;
            data->functions = functions;
            data->opCode = SGMP_V1_REKEYMERGE_OPCODE;
            blackboardData->flags.securityRequiredAuthenticate = TRUE;
            blackboardData->flags.securityRequiredEncrypt = TRUE;
            blackboardData->flags.streamingProtocolNotAllowed = TRUE;

            SGMP_V1_Send_IterateFunctions(data);
        }
        else
            data->status = SGMP_V1_SendStatus_TIMEOUTERROR;

        DOFOperation_Unlock(data->operation);
    }
    else
        data->status = SGMP_V1_SendStatus_TIMEOUTERROR;
}


void SGMP_V1_RequestGroupResponse_Send(SGMP_V1_SendData data, BlackboardData blackboardData)
{
    static const SGMP_V1_SendUtils functions[] = {SGMP_V1_Send_WriteFirstByte,
                                                  SGMP_V1_Send_WriteInitialState,
                                                  SGMP_V1_Send_WriteAppID,
                                                  SGMP_V1_Send_WriteDesire,
                                                  SGMP_V1_Send_WriteTicket, NULL};

    if (data->operation)
    {
        DOFOperation_Lock(data->operation);
        if (data->operation->context)
        {
            data->context = (SGMP_V1_OperationContext) data->operation->context;
            data->functions = functions;
            data->opCode = SGMP_V1_REQUESTGROUP_OPCODE;
            blackboardData->flags.streamingProtocolNotAllowed = TRUE;

            SGMP_V1_Send_IterateFunctions(data);
        }
        else
            data->status = SGMP_V1_SendStatus_TIMEOUTERROR;

        DOFOperation_Unlock(data->operation);
    }
    else
        data->status = SGMP_V1_SendStatus_TIMEOUTERROR;
}
