/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/channel.h>
#include <internal/key_request.h>
#include <internal/marshal.h>
#include <internal/sgmp_v1.h>
#include "sgmp_v1_sendutil.h"


void SGMP_V1_Send_Init( SGMP_V1_SendData data, SGMP_V1 sgmp, DOFOperation operation )
{
    PCRMem_Set( data, 0, sizeof( SGMP_V1_SendData_t ) );
    data->sgmp = sgmp;
    data->operation = operation;
    data->packet = sgmp->blackboardData->sendPacket;
}


boolean SGMP_V1_Send_IterateFunctions( SGMP_V1_SendData data )
{
    uint8 i = 0;

    while (data->functions[i] != NULL)
    {
        if (!data->functions[i++](data))
            return FALSE;
    }

    return TRUE;
}


boolean SGMP_V1_Send_WriteFirstByte( SGMP_V1_SendData data )
{
    uint32    size   = OALMarshal_GetCompressedValueSize( 2, SGMP_V1_APPID );
    OALStream stream = OALStream_CreateRW( size + 1 );                             /* 1-2 bytes appid, 1 byte opcode. */

    PCR_Assert( data->status == SGMP_V1_SendStatus_OK );
    if (stream)
    {
        size = OALMarshal_CompressValue( 2, SGMP_V1_APPID, size, OALStream_GetWritePtr( stream ) );
        if (size > 0 && size < 3)   /* 1 or 2 bytes */
        {
            OALStream_AdjustWriteOffset( stream, size );
            if (OALStream_Write8( stream, data->opCode ) && OALPacket_AppendStream( data->packet, stream ))
                return TRUE;                                                                 /* Success. */
            data->status = SGMP_V1_SendStatus_STREAMERROR;
        }
        OALStream_Release( stream, FALSE );
    }
    if (data->status != SGMP_V1_SendStatus_STREAMERROR)
        data->status  = SGMP_V1_SendStatus_MEMORYERROR;

    return FALSE;
}


boolean SGMP_V1_Send_I( SGMP_V1_SendData data )
{
    boolean retBoolean;

    PCR_Assert( data );
    PCR_Assert( data->context );
    PCR_Assert( data->context->sgmpStreamI );
    OALStream_Retain(data->context->sgmpStreamI); /*->sgmpStreamI destroys at data->packet destroy, could be immediate*/
    if ((retBoolean = OALPacket_AppendStream( data->packet, data->context->sgmpStreamI )) == FALSE)
        OALStream_Release( data->context->sgmpStreamI, TRUE); /* Fail. */

    return retBoolean;                                        /* Ref count still intact for context destroy function. */
}


boolean SGMP_V1_Send_WriteDelay( SGMP_V1_SendData data )
{
    OALStream stream = OALStream_CreateRW( 1 ); /* Allow 1 byte */

    PCR_Assert( data->status == SGMP_V1_SendStatus_OK );
    if (stream)
    {
        if (OALStream_Write8( stream, data->context->delay ) && OALPacket_AppendStream( data->packet, stream ))
            return TRUE;
        data->status = SGMP_V1_SendStatus_STREAMERROR;
        OALStream_Release( stream, FALSE );
    }
    else
        data->status = SGMP_V1_SendStatus_MEMORYERROR;

    return FALSE;
}


boolean SGMP_V1_Send_WriteAppID(SGMP_V1_SendData data)
{
    uint32    size   = OALMarshal_GetCompressedValueSize( 2, data->context->appid );
    OALStream stream = OALStream_CreateRW( size );

    PCR_Assert( data->status == SGMP_V1_SendStatus_OK );
    if (stream)
    {
        PCR_Assert( data->context->appid <= 0x7FFF );
        size = OALMarshal_CompressValue( 2, data->context->appid, size, OALStream_GetWritePtr( stream ) );
        if ((size > 0) && (size < 3))                                          /* appid is 1 or 2 bytes */
        {
            OALStream_AdjustWriteOffset( stream, size );
            if (OALPacket_AppendStream( data->packet, stream ))
                return TRUE;
            data->status = SGMP_V1_SendStatus_STREAMERROR;
        }
        OALStream_Release( stream, FALSE );
    }
    if (data->status != SGMP_V1_SendStatus_STREAMERROR)
        data->status  = SGMP_V1_SendStatus_MEMORYERROR;

    return FALSE;
}


boolean SGMP_V1_Send_WriteTieBreaker(SGMP_V1_SendData data)
{
    OALStream stream = OALStream_CreateRW( 4 );

    PCR_Assert( data->status == SGMP_V1_SendStatus_OK );
    if (stream)
    {
        if (OALStream_Write32( stream, data->context->tieBreaker ) && OALPacket_AppendStream( data->packet, stream ))
            return TRUE;
        data->status = SGMP_V1_SendStatus_STREAMERROR;
        OALStream_Release( stream, FALSE );
    }
    else
        data->status  = SGMP_V1_SendStatus_MEMORYERROR;

    return FALSE;
}


boolean SGMP_V1_Send_WriteDesire( SGMP_V1_SendData data )
{
    OALStream stream = OALStream_CreateRW( 1 );

    PCR_Assert( data->status == SGMP_V1_SendStatus_OK );
    if (stream)
    {
        if (OALStream_Write8( stream, data->context->desire ) && OALPacket_AppendStream( data->packet, stream ))
            return TRUE;
        data->status = SGMP_V1_SendStatus_STREAMERROR;
        OALStream_Release( stream, FALSE );
    }
    else
        data->status = SGMP_V1_SendStatus_MEMORYERROR;

    return FALSE;
}


boolean SGMP_V1_Send_WriteTmin(SGMP_V1_SendData data)
{
    OALStream stream = OALStream_CreateRW( 2 );

    PCR_Assert( data->status == SGMP_V1_SendStatus_OK );
    if (stream)
    {
        if (OALStream_Write16( stream, data->context->Tmin ) && OALPacket_AppendStream( data->packet, stream ))
            return TRUE;
        data->status = SGMP_V1_SendStatus_STREAMERROR;
        OALStream_Release( stream, FALSE );
    }
    else
        data->status = SGMP_V1_SendStatus_MEMORYERROR;

    return FALSE;
}


boolean SGMP_V1_Send_WriteKey(SGMP_V1_SendData data)
{
    OALStream stream = OALStream_CreateRW( SGMP_V1_KEY_SIZE );

    PCR_Assert( data->status == SGMP_V1_SendStatus_OK );
    if (stream)
    {
        if (OALStream_Write( stream, SGMP_V1_KEY_SIZE, data->context->key ) &&
            OALPacket_AppendStream( data->packet, stream ))
           return TRUE;
        data->status = SGMP_V1_SendStatus_STREAMERROR;
        OALStream_Release( stream, FALSE );
    }
    else
        data->status = SGMP_V1_SendStatus_MEMORYERROR;

    return FALSE;
}


boolean SGMP_V1_Send_WriteEpoch( SGMP_V1_SendData data )
{
    OALStream stream = OALStream_CreateRW( 2 );

    PCR_Assert( data->status == SGMP_V1_SendStatus_OK );
    if (stream)
    {
        if (OALStream_Write16( stream, data->context->epoch ) && OALPacket_AppendStream( data->packet, stream ))
            return TRUE;
        data->status = SGMP_V1_SendStatus_STREAMERROR;
        OALStream_Release( stream, FALSE );
    }
    else
        data->status = SGMP_V1_SendStatus_MEMORYERROR;

    return FALSE;
}


boolean SGMP_V1_Send_WriteTicket(SGMP_V1_SendData data)
{
    OALStream stream = OALStream_CreateRW( OALKeyGrant_MarshalSize( data->context->ticket ) );

    PCR_Assert( data->status == SGMP_V1_SendStatus_OK );
    if (stream)
    {
        if (OALKeyGrant_Marshal( data->context->ticket, stream ) && OALPacket_AppendStream( data->packet, stream ))
            return TRUE;
        data->status = SGMP_V1_SendStatus_STREAMERROR;
        OALStream_Release( stream, FALSE );
    }
    else
        data->status = SGMP_V1_SendStatus_MEMORYERROR;

    return FALSE;
}


boolean SGMP_V1_Send_WriteInitialState( SGMP_V1_SendData data )
{
    PCR_Assert( data );
    PCR_Assert( data->context );
    PCR_Assert( data->context->initialState );
    {
        /* append the size of the initial state as a compressed short and then append the initial state. */
        uint16    initialStateSize = OALStream_GetDataSize( data->context->initialState );
        uint32    size             = OALMarshal_GetCompressedValueSize( 2, initialStateSize );
        OALStream stream           = OALStream_CreateRW( size );

        PCR_Assert( data->status == SGMP_V1_SendStatus_OK );
        PCR_Assert( initialStateSize <= 0x7FFF );
        if (stream)
        {
            size = OALMarshal_CompressValue( 2, initialStateSize, size, OALStream_GetWritePtr( stream ) );
            if (size > 0 && size <= 2)
            {
                OALStream_AdjustWriteOffset( stream, size );
                if ( OALPacket_AppendStream( data->packet, stream ) )
                {
                    if ( OALPacket_AppendStream( data->packet, data->context->initialState ) )
                    {
                        OALStream_Retain(data->context->initialState);
                        return TRUE;
                    }
                }
                else
                {
                    OALStream_Release(stream, FALSE);
                }
                data->status = SGMP_V1_SendStatus_STREAMERROR;
            }
            else
            {
                OALStream_Release(stream, FALSE);
                data->status = SGMP_V1_SendStatus_STREAMERROR;
            }
        }
        if (data->status != SGMP_V1_SendStatus_STREAMERROR)
            data->status  = SGMP_V1_SendStatus_MEMORYERROR;
    }

    return FALSE;
}
