/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/operation.h>
#include <internal/cipher.h>
#include <internal/enqueue.h>
#include <internal/error.h>
#include <internal/operationinternal.h>
#include <internal/router.h>
#include <internal/securegroupinternal.h>
#include <internal/sgmp_v1.h>
#include "sgmp_v1_sm.h"
#include "sgmp_v1_operation.h"


typedef struct SGMP_V1_ResponseContext_t
{
    DOFAddress                      address;
    OALKeyGrant                     ticket;
    OALStream                       initialState;
    void *                          requestGroupResponseContext;
    PCRMutex                        requestGroupResponseMutex;
    SGMP_V1_RequestGroupCallback_fn callbackFn;
    uint16                          appid;
    uint8                           desire;
    boolean                         receivedMulticast;
} SGMP_V1_ResponseContext_t;


typedef struct SGMPContext_t
{
    OALEnqueueItem_t super;
    DOFOperation op;
    SGMP_V1_ResponseContext_t * responseContext;
} SGMPContext_t;


static OALEnqueue SGMP_V1_queue;


typedef void (*SGMP_SM_Handler)(DOFOperation op, SGMP_V1_ResponseContext_t * responseContext);


static void SGMP_V1_OpClosing( OALObserver self, OALObservable observable, void* ctx )
{
    DOFOperation op = (DOFOperation)ctx;
    SGMP_V1_OperationContext sgmpCtx;

    DOFOperation_Lock( op );
    sgmpCtx = (SGMP_V1_OperationContext)op->context;
    if (sgmpCtx)
    {
        if (sgmpCtx->callbackFn)
        {
            PCR_Assert( sgmpCtx->requestGroupContext );
            PCR_Assert( sgmpCtx->requestGroupResponseMutex );

            sgmpCtx->callbackFn( NULL/* address */, NULL/* ticket. */, 0/* appid */, 0/* desire */,
                                 DOFERROR_TIMEOUT, sgmpCtx->requestGroupContext, NULL/* streamCcmModeInitialState */,
                                 FALSE/* receivedMulticast */, op, sgmpCtx->requestGroupResponseMutex );

            sgmpCtx->callbackFn = NULL;
            /* Removes callback so we will not call it again, not even in SGMP_V1_HandleRequestGroupResponse_Initiator*/
        }
        SGMP_V1_OperationContext_Destroy( sgmpCtx );
        op->context = NULL;
    }
    DOFOperation_Unlock( op );
    if (sgmpCtx)
        DOFOperation_Internal_Destroy( op ); /* op had been retained. */
}


static void SGMP_V1_HandleHeartbeat_Initiator(DOFOperation op, SGMP_V1_ResponseContext_t * responseContext)
{
    /* This should never happen, but if it does return safely, (no harm done) */
}


static void SGMP_V1_HandleHeartbeat_Responder(DOFOperation op, SGMP_V1_ResponseContext_t * responseContext)
{
    if (op)
    {
        PCR_Assert( op->sources->channel );
        PCR_Assert( OALChannel_IsSecureGroup( op->sources->channel ) );
        DOFOperation_Lock(op);
        /* Cast the channel to a secure group since the securegroup is a channel. */
        if ( op->context )
        {
            DOFSecureGroup_HeartbeatReceiveHandler((DOFSecureGroup)op->sources->channel, op->sources->address,
                                                   ((SGMP_V1_OperationContext)op->context)->appid,
                                                   ((SGMP_V1_OperationContext)op->context)->desire,
                                                   ((SGMP_V1_OperationContext)op->context)->tieBreaker);
            SGMP_V1_OperationContext_Destroy((SGMP_V1_OperationContext)op->context);
            op->context = NULL;
        }
        DOFOperation_Unlock(op);
        DOFOperation_Cancel(op);
    }
}


static void SGMP_V1_HandleEpochChanged_Initiator(DOFOperation op, SGMP_V1_ResponseContext_t * responseContext)
{
    /* This should never happen, but if it does return safely, (no harm done) */
}


static void SGMP_V1_HandleEpochChanged_Responder(DOFOperation op, SGMP_V1_ResponseContext_t * responseContext)
{
    if (op)
    {
        PCR_Assert( op->sources->channel );
        PCR_Assert( OALChannel_IsSecureGroup( op->sources->channel ) );
        DOFOperation_Lock(op);
        if ( op->context )
        {
            DOFSecureGroup_EpochChangedReceiveHandler((DOFSecureGroup)op->sources->channel, op->sources->address,
                                                      ((SGMP_V1_OperationContext)op->context)->epoch,
                                                      ((SGMP_V1_OperationContext)op->context)->Tmin);
            SGMP_V1_OperationContext_Destroy((SGMP_V1_OperationContext)op->context);
            op->context = NULL;
        }
        DOFOperation_Unlock(op);
        DOFOperation_Cancel(op);
    }
}


static void SGMP_V1_HandleRekey_Initiator(DOFOperation op, SGMP_V1_ResponseContext_t * responseContext)
{
    /* This should never happen, but if it does return safely, (no harm done) */
}


static void SGMP_V1_HandleRekey_Responder(DOFOperation op, SGMP_V1_ResponseContext_t * responseContext)
{
    if (op)
    {
        PCR_Assert( op->sources->channel );
        PCR_Assert( OALChannel_IsSecureGroup( op->sources->channel ) );
        DOFOperation_Lock(op);
        /* Note: This next call does not consume streamCcmModeInitialState. */
        if ( op->context )
        {
            DOFSecureGroup_RekeyReceiveHandler( (DOFSecureGroup)op->sources->channel,
                                                ((SGMP_V1_OperationContext)op->context)->delay,
                                                ((SGMP_V1_OperationContext)op->context)->initialState,
                                                ((SGMP_V1_OperationContext)op->context)->key,
                                                ((SGMP_V1_OperationContext)op->context)->receivedMulticast );
            SGMP_V1_OperationContext_Destroy((SGMP_V1_OperationContext)op->context);
            op->context = NULL;
        }
        DOFOperation_Unlock(op);
        DOFOperation_Cancel(op);
    }
}


static void SGMP_V1_HandleRekeyEpoch_Initiator(DOFOperation op, SGMP_V1_ResponseContext_t * responseContext)
{
    /* This should never happen, but if it does return safely, (no harm done) */
}


static void SGMP_V1_HandleRekeyEpoch_Responder(DOFOperation op, SGMP_V1_ResponseContext_t * responseContext)
{
    if (op)
    {
        PCR_Assert( op->sources->channel );
        PCR_Assert( OALChannel_IsSecureGroup( op->sources->channel ) );
        DOFOperation_Lock(op);
        /* Note: This next call does not consume streamCcmModeInitialState. */
        if ( op->context )
        {
            DOFSecureGroup_RekeyEpochReceiveHandler( (DOFSecureGroup)op->sources->channel, op->sources->address,
                                                   ((SGMP_V1_OperationContext)op->context)->delay,
                                                   ((SGMP_V1_OperationContext)op->context)->initialState,
                                                   ((SGMP_V1_OperationContext)op->context)->key,
                                                   ((SGMP_V1_OperationContext)op->context)->epoch,
                                                   ((SGMP_V1_OperationContext)op->context)->receivedMulticast );
            SGMP_V1_OperationContext_Destroy((SGMP_V1_OperationContext)op->context);
            op->context = NULL;
        }
        DOFOperation_Unlock(op);
        DOFOperation_Cancel(op);
    }
}


static void SGMP_V1_HandleRekeyMerge_Initiator( DOFOperation op, SGMP_V1_ResponseContext_t * responseContext )
{
    /* This should never happen, but if it does return safely, (no harm done) */
}


static void SGMP_V1_HandleRekeyMerge_Responder( DOFOperation op, SGMP_V1_ResponseContext_t * responseContext )
{
    if (op)
    {
        PCR_Assert( op->sources->channel );
        PCR_Assert( OALChannel_IsSecureGroup( op->sources->channel ) );
        DOFOperation_Lock(op);
        /* Note: This next call does not consume streamCcmModeInitialState. */
        if ( op->context )
        {
            DOFSecureGroup_RekeyMergeReceiveHandler( (DOFSecureGroup)op->sources->channel,
                                                     ((SGMP_V1_OperationContext)op->context)->initialState,
                                                     ((SGMP_V1_OperationContext)op->context)->key,
                                                     ((SGMP_V1_OperationContext)op->context)->receivedMulticast );
            SGMP_V1_OperationContext_Destroy((SGMP_V1_OperationContext)op->context);
            op->context = NULL;
        }
        DOFOperation_Unlock(op);
        DOFOperation_Cancel(op);
    }
}


static void SGMP_V1_HandleRequestGroup_Responder( DOFOperation op, SGMP_V1_ResponseContext_t * responseContext )
{
    PCR_Assert( op );
    /* op->context holds info about an incoming RequestGroup PDU. Lock to protect from op cancel that might happen
       before we are done with the context. */
    DOFOperation_Lock( op );
    if ( op->context ) /* Filled in DOFOperation_Create_RequestGroup(), but possibly freed by an early cancel. */
    {
        PCR_Assert( ((SGMP_V1_OperationContext)op->context)->secureGroupChannel );
        PCR_Assert( OALChannel_IsSecureGroup( ((SGMP_V1_OperationContext)op->context)->secureGroupChannel ) );
        DOFSecureGroup_ProcessRequestGroup( (DOFSecureGroup)((SGMP_V1_OperationContext)op->context)->secureGroupChannel,
                                            op, ((SGMP_V1_OperationContext)op->context)->keyRequest,
                                            ((SGMP_V1_OperationContext)op->context)->sgmpdomain,
                                            ((SGMP_V1_OperationContext)op->context)->I_Rec_Stream,
                                            ((SGMP_V1_OperationContext)op->context)->I_Offset,
                                            ((SGMP_V1_OperationContext)op->context)->Scope_Offset,
                                            ((SGMP_V1_OperationContext)op->context)->epoch,
                                            SGMP_V1_APPID ); /* Next note- how we know _V1_APPID is sender's highest. */
        /* Note: We know the sender sent to us using SGMP_V1_APPID simply because we are SGMP_V1 receive handler code. We
           know SGMP_V1_APPID is the sender's highest understood SGMP appid version because the protocol spec, ReqG PDU
           section, says "A node should only attempt to join a group using the latest version of SGMP that it understands"*/
    }
    DOFOperation_Unlock(op);
    /* Do not cancel the operation here since secureGroup code may send a request group response using it. */
}


static void SGMP_V1_HandleRequestGroupResponse_Initiator( DOFOperation op, SGMP_V1_ResponseContext_t * responseContext )
{
    /* We handle a ReqG response here. Since we can receive multiple request group responses we must NOT use op here and
        should especially not cancel op. Instead, use responseContext, which was purposely filled with all we need. */
    boolean ok = FALSE;

    if (responseContext)         /* This can probably be asserted, but check in case of a path we have not found yet. */
    {
        if (responseContext->callbackFn) /* SGMP_V1_OpClosing() NULLs this; means op cancelled so deny other callbacks*/
        {
            /* Call the function specified in the original request group send operation: RequestGroupCallback(). */
            PCR_Assert( responseContext->requestGroupResponseContext );
            PCR_Assert( responseContext->requestGroupResponseMutex );
            ok = TRUE;

            responseContext->callbackFn( responseContext->address, responseContext->ticket, responseContext->appid,
                                         responseContext->desire, DOFERROR_OK,
                                         responseContext->requestGroupResponseContext, responseContext->initialState,
                                         responseContext->receivedMulticast, op,
                                         responseContext->requestGroupResponseMutex );


            /* Release things we retained in responseContext. */
            DOFAddress_Destroy( responseContext->address );
            OALKeyGrant_Release( responseContext->ticket );
            OALStream_Release( responseContext->initialState, TRUE );
            DOFMem_Free( responseContext );
        }
    }
    if (!ok)
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "[SGMP] Ignoring a ReqG response- it has no specified callback");
}


static void SGMP_V1_queueFn (OALThreadPoolItem dnode, boolean shutdown)
{
    boolean         isSecureGroup;
    SGMPContext_t * node;
    /* In this former approach, 8 entries are reserved, but only the first 6 are populated.
    static const    SGMP_SM_Handler functions[SGMP_V1_HIGHEST_OPCODE - SGMP_V1_HEARTBEAT_OPCODE + 1][2] =*/
    static const    SGMP_SM_Handler functions[DOFOPERATIONTYPE_SGMP_LAST - DOFOPERATIONTYPE_SGMP_V1_HEARTBEAT][2] =
    {
        /* Note: this list is in operation type order. */
        /* Heartbeat */
        {SGMP_V1_HandleHeartbeat_Initiator,             SGMP_V1_HandleHeartbeat_Responder},
        /* KEK Changed */
        {SGMP_V1_HandleEpochChanged_Initiator,          SGMP_V1_HandleEpochChanged_Responder},
        /* Rekey */
        {SGMP_V1_HandleRekey_Initiator,                 SGMP_V1_HandleRekey_Responder},
        /* Request Group */
        {SGMP_V1_HandleRequestGroupResponse_Initiator,  SGMP_V1_HandleRequestGroup_Responder},
        /* Rekey KEK */
        {SGMP_V1_HandleRekeyEpoch_Initiator,            SGMP_V1_HandleRekeyEpoch_Responder},
        /* Rekey Immediate */
        {SGMP_V1_HandleRekeyMerge_Initiator,            SGMP_V1_HandleRekeyMerge_Responder}
    };

    PCR_Assert( dnode != NULL );
    node = (SGMPContext_t *)dnode;
    PCR_Assert( node->op );
    PCR_Assert( node->op->operation_type >= DOFOPERATIONTYPE_SGMP_V1_HEARTBEAT );
    PCR_Assert( node->op->operation_type < DOFOPERATIONTYPE_SGMP_LAST );
    if (node->op->operation_type != DOFOPERATIONTYPE_SGMP_V1_REQUESTGROUP ||
        !node->op->flags.bits.internally_created)                    /* ReqG response was not internally created. */
    {
        /* Note: On incoming ReqG response, DOFOperation_Create_RequestGroup() is called with provider operation
           FALSE, which clears op->flags.bits.internally_created, and lets us be here. */

        /* This is any incoming PDU except SGMP ReqG command. */
        /* Prove that our operative channel is a secureGroup channel. */
        PCR_Assert( node->op->sources->channel );
        isSecureGroup = OALChannel_IsSecureGroup( node->op->sources->channel );
    }
    else
    {
        /* This is incoming ReqG PDU (not ReqG response), op->flags.bits.internally_created is TRUE. */
        isSecureGroup = TRUE;
        /* These asserts check work accomplished in SGMP_V1_findChannelForRequestGroup(). If here, the channel that
           will be submitted to DOFSecureGroup_ProcessRequestGroup() is known to be a secureGroup channel. */
        PCR_Assert( node->op->context );
        PCR_Assert( ((SGMP_V1_OperationContext)node->op->context)->secureGroupChannel );
        PCR_Assert( OALChannel_IsSecureGroup( ((SGMP_V1_OperationContext)node->op->context)->secureGroupChannel ) );
    }
    if (isSecureGroup)
        functions[node->op->operation_type - DOFOPERATIONTYPE_SGMP_V1_HEARTBEAT]
                 [node->op->flags.bits.internally_created]( node->op, node->responseContext );
    else
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE,
                        "[SGMP] Received but rejected incoming PDU of non-secureGroup connection type." );
    DOFOperation_Internal_Destroy( node->op );
    DOFMem_Free( dnode );
}


void SGMP_V1_SM_ProcessToResponder( DOFOperation op )
{
    boolean                     success         = TRUE;
    SGMP_V1_ResponseContext_t * responseContext = NULL;

    PCR_Assert( op );
    DOFOperation_Retain( op );

    /* Note: do not use op->channel here since a request group may not have it set. */

    if (op->operation_type == DOFOPERATIONTYPE_SGMP_V1_REQUESTGROUP && !op->flags.bits.internally_created)
    {
        /* We have a ReqG response PDU: .internally_created TRUE means we just received a ReqG command PDU, FALSE means
           we just received a ReqG response PDU. */
        success = FALSE;
        /* ReqG response is useless if it holds no ticket. */
        if (((SGMP_V1_OperationContext)op->context)->ticket)
        {
            if ((responseContext = DOFMem_Alloc( sizeof( SGMP_V1_ResponseContext_t ) )) != NULL)
            {
                PCRMem_Set( responseContext, 0, sizeof( SGMP_V1_ResponseContext_t ));
                responseContext->ticket = ((SGMP_V1_OperationContext)op->context)->ticket;
                if (OALKeyGrant_Retain( responseContext->ticket ) != NULL)
                {
                    responseContext->address = DOFAddress_Copy( op->sources->address );
                    if (responseContext->address || !op->sources->address)
                    {
                        responseContext->initialState = ((SGMP_V1_OperationContext)op->context)->initialState;
                        OALStream_Retain( responseContext->initialState );
                        responseContext->receivedMulticast = ((SGMP_V1_OperationContext)op->context)->receivedMulticast;
                        responseContext->requestGroupResponseContext =
                                                           ((SGMP_V1_OperationContext)op->context)->requestGroupContext;
                        responseContext->requestGroupResponseMutex =
                                                     ((SGMP_V1_OperationContext)op->context)->requestGroupResponseMutex;
                        /* This next is RequestGroupCallback() in securegroupGetGroupKey.c. */
                        responseContext->callbackFn = ((SGMP_V1_OperationContext)op->context)->callbackFn;
                        responseContext->appid = ((SGMP_V1_OperationContext)op->context)->appid;
                        responseContext->desire = ((SGMP_V1_OperationContext)op->context)->desire;
                        success = TRUE;
                    }
                    else
                    {
                        OALKeyGrant_Release( responseContext->ticket );
                        responseContext->ticket = NULL;
                    }
                }
                if (!success)
                {
                    DOFMem_Free( responseContext );
                    responseContext = NULL;
                }
            }
        }
        if (!success)
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "[SGMP] failed to create response context");
    }
    if (success)
    {
        SGMPContext_t * node = (SGMPContext_t *)DOFMem_Alloc( sizeof( SGMPContext_t ) );

        success = FALSE;
        if (node != NULL)
        {
            OALEnqueueItem_Init( (OALEnqueueItem)node, SGMP_V1_queueFn );
            node->op = op;
            node->responseContext = responseContext;                 /* NULL except when receiving ReqG response PDU. */
            if (OALEnqueue_Enqueue( SGMP_V1_queue, (OALEnqueueItem)node ))
                success = TRUE;
            else
            {
                DOFMem_Free( node );
                if (responseContext)
                {
                    OALKeyGrant_Release( responseContext->ticket );
                    DOFAddress_Destroy( responseContext->address );
                    OALStream_Release( responseContext->initialState, TRUE );
                    DOFMem_Free(responseContext);
                }
            }
        }
        else
        {
            if (responseContext)
            {
                OALKeyGrant_Release( responseContext->ticket );
                DOFAddress_Destroy( responseContext->address );
                OALStream_Release( responseContext->initialState, TRUE );
                DOFMem_Free(responseContext);
            }
        }
    }
    if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ) && !success)
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "[SGMP] failed to enqueue process to responder");
}


boolean SGMP_V1_SM_Init( void )
{
    SGMP_V1_queue = OALEnqueue_Create( DOF_GetDof()->threadPool );

    return SGMP_V1_queue ? TRUE : FALSE;
}


void SGMP_V1_SM_Shutdown( void )
{
    OALEnqueue_Destroy( SGMP_V1_queue );
}


SGMP_V1 SGMP_V1_Create( BlackboardData blackboardData )
{
    SGMP_V1 ret;

    PCR_Assert( blackboardData != NULL );

    ret = (SGMP_V1)DOFMem_Alloc( sizeof( SGMP_V1_t ) );
    if (ret != NULL)
        ret->blackboardData = blackboardData;                                             /* Only one member to fill. */
    else
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "[SGMP] failed to create sgmp v1" );

    return ret;
}


void SGMP_V1_Destroy( SGMP_V1 self )
{
    if (self != NULL)
        DOFMem_Free( self );
}


/* Send functions:  The following functions are used by a secure group to send data to a remote node. */


/* @internal

     @param address NULL/not: send multicast/unicast.
     @param appid
     @param desire
     @param tieBreaker number supplied by hb sender to possibly be used by receiver. This can be any 4 byte value.
*/
DOFError SGMP_V1_SendHeartbeat( OALChannel channel, DOFAddress address, DOFSecurityScope securityScope, uint16 appid,
                                uint8 desire, uint32 tieBreaker )
{
    DOFOperation op;

    PCR_Assert( securityScope );
    op = DOFOperation_Create_Heartbeat( channel, address, securityScope, tieBreaker, appid, desire,
                                        SGMP_V1_SEND_OP_TIMEOUT, FALSE );

    if (op)
    {
        /* call route to provider to send the PDU */
        channel->routeToProvider(channel, op, NULL, op->sources->address);
        DOFOperation_Internal_Destroy(op);
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
            DOFSecureGroup_LogMessage( DOFLOGLEVEL_DEBUG, "[SGMP] heartbeat PDU sent multicast or unicast. Multicast: ",
                                                          (address == NULL) );

        return DOFERROR_OK;
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "[SGMP] failed to create heartbeat op");

    return DOFERROR_INTERNAL_ERROR;
}


/* address NULL/not: send multicast/unicast. Multicast send is only by SGM. */
DOFError SGMP_V1_SendEpochChanged( OALChannel channel, DOFAddress address, DOFSecurityScope securityScope,
                                   uint16 epoch, uint16 getKEKWindowMinutes, uint8 retryCount, uint32 retryPeriod )
{
    uint32   timeout;
    DOFError status;

    PCR_Assert( securityScope );
    if (retryCount < 100 && (!retryCount || retryPeriod < 20000) && (!retryCount || retryPeriod >= 10) &&
        (timeout = SGMP_V1_SEND_OP_TIMEOUT + (retryCount * retryPeriod) - (retryCount ? (retryPeriod / 2): 0)) < 15000)
    {
        DOFOperation op;

        /* We cannot directly control retryCount in the operation we now create. But we tell it that retry is in
           effect, and that op timeout is a little after the final retry has been sent. */
        status = DOFERROR_INSUFFICIENT_RESOURCES;
        if ((op = DOFOperation_Create_EpochChanged( channel, address, securityScope, epoch, getKEKWindowMinutes,
                                                    timeout, FALSE, (retryCount > 0), retryPeriod )) != NULL)
        {
            PCR_Assert( op->operation_will_close );                      /* _Create_EpochChanged() did this. */
            if (DOFOperation_AddOpWillCloseObserver( op, SGMP_V1_OpClosing ) != NULL)
            {
                /* For retryCount DOFOperationManager_SetRetry() was already called by _Create_EpochChanged(). */
                PCR_Assert( op->opID );                                   /* Filled by _Create_EpochChanged(). */
                status = OALOperationManager_AddOperation( op, op->opID, FALSE );
                if (status == DOFERROR_OK)
                    channel->routeToProvider( channel, op, NULL, op->sources->address );
                else
                    DOFOperation_Cancel( op );             /* cleanup: op was retained in _AddOpWillCloseObserver(). */
            }
            DOFOperation_Internal_Destroy( op );
        }
        if (status != DOFERROR_OK)
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "[SGMP] failed to send epochChanged op" );
    }
    else
    {
        /* Ordinary is like 1 retry happening a few seconds later. Anything wildly outside of this we just won't do. */
        status = DOFERROR_INTERNAL_ERROR;
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "[SGMP] send epoch changed retry violation" );
    }

    return status;
}


DOFError SGMP_V1_SendRekey( OALChannel channel, DOFAddress address, DOFSecurityScope securityScope,
                            uint8 secsPromoteDelay, OALStream streamCcmModeInitialState, uint8* key )
{
    DOFOperation op;

    PCR_Assert( securityScope );
    op = DOFOperation_Create_Rekey( channel, address, securityScope, streamCcmModeInitialState, key, secsPromoteDelay,
                                    SGMP_V1_SEND_OP_TIMEOUT, FALSE );
    if (op)
    {
        channel->routeToProvider(channel, op, NULL, op->sources->address);
        DOFOperation_Internal_Destroy(op);
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG, "[SGMP] slot rekey PDU sent." );
        return DOFERROR_OK;
    }
    else
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "[SGMP] failed to create rekey op");

    return DOFERROR_INTERNAL_ERROR;
}


DOFError SGMP_V1_SendRekeyEpoch( OALChannel channel, DOFAddress address, DOFSecurityScope securityScope,
                                 uint8 secsPromoteDelay, uint16 epoch, OALStream streamCcmModeInitialState, uint8* key )
{
    DOFOperation op;

    PCR_Assert( securityScope );
    op = DOFOperation_Create_RekeyEpoch( channel, address, securityScope, key, streamCcmModeInitialState,
                                         secsPromoteDelay, epoch, SGMP_V1_SEND_OP_TIMEOUT, FALSE );
    if (op)
    {
        channel->routeToProvider(channel, op, NULL, op->sources->address);
        DOFOperation_Internal_Destroy(op);
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG, "[SGMP] epoch rekey PDU sent." );
        return DOFERROR_OK;
    }
    else
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "[SGMP] failed to create rekeyEpoch");

    return DOFERROR_INTERNAL_ERROR;
}


DOFError SGMP_V1_SendRekeyMerge( OALChannel channel, DOFAddress address, DOFSecurityScope securityScope, PCREvent event,
                                 OALStream streamCcmModeInitialState, uint8* key )
{
    DOFOperation op;

    PCR_Assert( securityScope );
    op = DOFOperation_Create_RekeyMerge( channel, address, securityScope, event, key, streamCcmModeInitialState,
                                         SGMP_V1_SEND_OP_TIMEOUT, FALSE );
    if (op)
    {
        channel->routeToProvider(channel, op, NULL, op->sources->address);
        DOFOperation_Internal_Destroy(op);
        if (DOFLog_IsLogLevel( DOFLOGLEVEL_DEBUG ))
            DOFSecureGroup_LogMessageNoTrailer( DOFLOGLEVEL_DEBUG, "[SGMP] merge rekey PDU sent." );
        return DOFERROR_OK;
    }
    else
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "[SGMP] failed to create rekey merge");

    return DOFERROR_INTERNAL_ERROR;
}


/* address NULL/not: send multicast/unicast. */
DOFOperation SGMP_V1_SendRequestGroup( OALChannel channel, DOFAddress address, OALStream streamI,
                                       DOFObjectIDDomain domain, DOFSecurityScope securityScope,
                                       SGMP_V1_RequestGroupCallback_fn callbackFn, void * rgSendCtx, uint32 timeout,
                                       boolean secure, PCRMutex reqGRespMutex )
{
    DOFOperation op;

    PCR_Assert( domain );
    PCR_Assert( rgSendCtx );
    PCR_Assert( securityScope );
    PCR_Assert( callbackFn );
    /* Security.7 PDU spec: domain not broadcast or multicast OID.*/
    if ((op = DOFOperation_Create_RequestGroup( channel, address, securityScope, streamI,
                                                NULL,    /* Placeholder: keyRequest not needed in sending reqG PDU. */
                                                domain,  /* Security.7 PDU: domain not broadcast or mcast OID. */
                                                NULL,    /* Placeholder: I received stream not needed in sending reqG PDU. */
                                                0,       /* Placeholder: only needed with I received Stream. */
                                                0,       /* Placeholder: only needed with I received Stream. */
                                                callbackFn, rgSendCtx,
                                                0x7F,    /* Placeholder: epoch not needed in sending reqG PDU. */
                                                timeout, FALSE, reqGRespMutex )) != NULL)
    {
        DOFError status = DOFERROR_INSUFFICIENT_RESOURCES;

        PCR_Assert( op->operation_will_close );              /* _Create_RequestGroup() did this. */
        if (DOFOperation_AddOpWillCloseObserver( op, SGMP_V1_OpClosing ) != NULL)
        {
            PCR_Assert( op->opID );                         /* Filled by _Create_RequestGroup(). */
            status = OALOperationManager_AddOperation( op, op->opID, FALSE );
            if (status == DOFERROR_OK)
                channel->routeToProvider( channel, op, NULL, op->sources->address );                   /* Retains op. */
            else
                DOFOperation_Cancel( op );                  /* cleanup: op was retained in _AddOpWillCloseObserver(). */
        }
        DOFOperation_Internal_Destroy( op );
        if (status != DOFERROR_OK)
            op = NULL;
    }
    if (op == NULL)
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "[SGMP] failed to create requestgroup op");

    return op;
}


/* Always sends unicast as a response from the RequestGroup request. The latter may have come in multicast or unicast.*/
/* Note: the channel will be a connectionless channel if unsecure, or a secure group channel if secure. */
DOFError SGMP_V1_SendRequestGroupResponse( OALChannel channel, OALKeyGrant ticket, DOFOperation requestGroupOp,
                                           uint16 appid, uint8 desire, OALStream streamCcmModeInitialState )
{
    if (requestGroupOp)
    {
        SGMP_V1_OperationContext context;
        DOFError                 status = DOFERROR_INSUFFICIENT_RESOURCES;

        PCR_Assert( requestGroupOp->context ); /*A ReqG response w/o things held in requestGroupOp->context is useless*/
        PCR_Assert( ticket );                  /*A ReqG response w/o a ticket is useless. */
        DOFOperation_Lock( requestGroupOp );
        DOFOperation_Retain( requestGroupOp );
        context = (SGMP_V1_OperationContext)requestGroupOp->context;
        context->ticket = ticket;
        if (OALKeyGrant_Retain( context->ticket ) != NULL)
        {
            context->appid = appid;
            context->desire = desire;
            context->initialState = streamCcmModeInitialState;
            OALStream_Retain( streamCcmModeInitialState );
            channel->routeToRequestor( channel, requestGroupOp, NULL );
            status = DOFERROR_OK;
        }
        DOFOperation_Unlock( requestGroupOp );
        DOFOperation_Internal_Destroy( requestGroupOp );

        return status;
    }

    return DOFERROR_INTERNAL_ERROR;
}
