/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/objectid.h>
#include <internal/avl_tree.h>
#include <internal/cipher.h>
#include <internal/srp.h>
#include <internal/dofinternal.h>
#include <internal/bignum.h>

/** @internal @addtogroup SRP
    Secure Remote Password Functionality
@{
**/

#ifdef WORDS_BIGENDIAN
#define SRP_CONSTANT(a,b) b, a
#else
#define SRP_CONSTANT(a,b) a, b
#endif

static const uint8 N_init[] = {0xEE,0xAF,0x0A,0xB9,0xAD,0xB3,0x8D,0xD6,0x9C,0x33,0xF8,0x0A,0xFA,0x8F,0xC5,0xE8,0x60,0x72,0x61,0x87,0x75,0xFF,0x3C,0x0B,0x9E,0xA2,0x31,0x4C,0x9C,0x25,0x65,0x76,0xD6,0x74,0xDF,0x74,0x96,0xEA,0x81,0xD3,0x38,0x3B,0x48,0x13,0xD6,0x92,0xC6,0xE0,0xE0,0xD5,0xD8,0xE2,0x50,0xB9,0x8B,0xE4,0x8E,0x49,0x5C,0x1D,0x60,0x89,0xDA,0xD1,0x5D,0xC7,0xD7,0xB4,0x61,0x54,0xD6,0xB6,0xCE,0x8E,0xF4,0xAD,0x69,0xB1,0x5D,0x49,0x82,0x55,0x9B,0x29,0x7B,0xCF,0x18,0x85,0xC5,0x29,0xF5,0x66,0x66,0x0E,0x57,0xEC,0x68,0xED,0xBC,0x3C,0x05,0x72,0x6C,0xC0,0x2F,0xD4,0xCB,0xF4,0x97,0x6E,0xAA,0x9A,0xFD,0x51,0x38,0xFE,0x83,0x76,0x43,0x5B,0x9F,0xC6,0x1D,0x2F,0xC0,0xEB,0x06,0xE3};

/* constants*/
static mp_int SRP_g; /* 2 */
static mp_int SRP_three; /* 3 */
static mp_int SRP_N;


#define SRP_PUBLIC (128)
#define SRP_PRIVATE (32)

boolean OALSRP_Init()
{
    /* Init Constants */
    if( mp_init(&SRP_g) == MP_OKAY )
    {
        if( mp_init(&SRP_three) == MP_OKAY )
        {
            if( mp_init(&SRP_N) == MP_OKAY )
            {
                if( mp_read_unsigned_bin(&SRP_N, N_init, 128 ) == MP_OKAY )
                {
                    mp_set(&SRP_g, 2);
                    mp_set(&SRP_three, 3);
                    return TRUE;
                }
                mp_clear(&SRP_N);
            }
            mp_clear(&SRP_three);
        }
        mp_clear(&SRP_g);
    }
    return FALSE;
}

void OALSRP_Shutdown()
{
    mp_clear(&SRP_g);
    mp_clear(&SRP_three);
    mp_clear(&SRP_N);
}

boolean OALSRP_Calc_PublicA( uint8* privateA, uint8* pubA )
{
    mp_int privA_bn;
    mp_int a;
    long unsigned int size = 128;
    boolean ret = FALSE;

    if( mp_init(&privA_bn) == MP_OKAY )
    {
        if( mp_init(&a) == MP_OKAY )
        {
            /* Read Private A */
            if( mp_read_unsigned_bin(&privA_bn, privateA, SRP_PRIVATE ) == MP_OKAY )
            {
                /* A = g ^ a, N */
                if( mp_exptmod( &SRP_g, &privA_bn, &SRP_N, &a) == MP_OKAY )
                {
                    /* Write Public A out */
                    if( mp_to_unsigned_bin_n_pad(&a, pubA, size) == MP_OKAY )
                    {
                        ret = TRUE;
                    }
                }
            }
            mp_clear(&a);
        }
        mp_clear(&privA_bn);
    }

    return ret;
}

boolean OALSRP_Calc_Partial_x( DOFObjectID identity, const char* password, uint8* partial_x )
{
    OALSecureHashContext ctx;

    OALSecureHash_Start( &ctx );
    OALSecureHash_Digest( &ctx, DOFObjectID_GetBytesSize( identity ), DOFObjectID_GetBytes( identity ) );
    OALSecureHash_Digest( &ctx, PCRString_Length(password), (uint8*)password );
    OALSecureHash_Finish( &ctx, partial_x );

    return TRUE;
}

boolean OALSRP_Calc_v( uint8* partial_x, uint8* s, uint8* v )
{
    OALSecureHashContext ctx;
    OALSecureHashDigest x;
    mp_int x_bn;
    mp_int v_bn;
    long unsigned int size = 128;
    boolean ret = FALSE;
    
    OALSecureHash_Start( &ctx );
    OALSecureHash_Digest( &ctx, 32, s );
    OALSecureHash_Digest( &ctx, 32, partial_x );
    OALSecureHash_Finish( &ctx, x );
    
    if( mp_init(&x_bn) == MP_OKAY )
    {
        if( mp_init(&v_bn) == MP_OKAY )
        {
            /* Read x */
            if( mp_read_unsigned_bin( &x_bn, x, 32 ) == MP_OKAY )
            {
                /* v = g ^ x, N */
                if( mp_exptmod( &SRP_g, &x_bn, &SRP_N, &v_bn) == MP_OKAY )
                {
                    /* Write v */
                    if( mp_to_unsigned_bin_n_pad(&v_bn, v, size) == MP_OKAY )
                        ret = TRUE;
                }
            } 
        }
        mp_clear(&v_bn);
    }
    mp_clear(&x_bn);
    
    return ret;
}

boolean OALSRP_CalculateAS( uint8* v, uint8* pubA, uint8* pubB, uint8* srpS, uint8* Ksrp )
{
    OALSecureHashContext ctx;
    OALSecureHashDigest u;
    mp_int temp_bn, temp2_bn, temp3_bn, A_bn, B_bn, b_bn, v_bn, u_bn, S_bn;
    uint8 b[32];
    long unsigned int size = 128;
    boolean ret = FALSE;
    
    if( mp_init( &B_bn ) == MP_OKAY )
    {
        if( mp_init( &A_bn ) == MP_OKAY )
        {
            if( mp_init( &b_bn )  == MP_OKAY )
            {
                if( mp_init( &v_bn ) == MP_OKAY )
                {
                    if( mp_init( &u_bn ) == MP_OKAY )
                    {
                        if( mp_init( &S_bn ) == MP_OKAY )
                        {
                            if( mp_init( &temp_bn ) == MP_OKAY )
                            {
                                if( mp_init( &temp2_bn ) == MP_OKAY )
                                {
                                    if( mp_init( &temp3_bn ) == MP_OKAY )
                                    {
                                        PCRSecureRandom_Generate( 32, b );
                                        /* g^b,N */
                                        if( mp_read_unsigned_bin( &b_bn, b, 32 ) == MP_OKAY )
                                        {
                                            if( mp_exptmod( &SRP_g, &b_bn, &SRP_N, &temp_bn) == MP_OKAY )
                                            {
                                                /* 3 * v */
                                                if( mp_read_unsigned_bin( &v_bn, v, 128 ) == MP_OKAY )
                                                {
                                                    if( mp_mul(&SRP_three, &v_bn, &temp2_bn) == MP_OKAY )
                                                    {
                                                        /*B=(3v + g^b,N) % N  */
                                                        if( mp_add(&temp2_bn, &temp_bn, &temp3_bn) == MP_OKAY )
                                                        {
                                                            if( mp_mod(&temp3_bn, &SRP_N, &B_bn ) == MP_OKAY )
                                                            {
                                                                /* Write Public B */
                                                                if( mp_to_unsigned_bin_n_pad(&B_bn, pubB, size) == MP_OKAY )
                                                                {
                                                                    /* u = SHA( A || B ) */
                                                                    OALSecureHash_Start( &ctx );
                                                                    OALSecureHash_Digest( &ctx, 128, pubA );
                                                                    OALSecureHash_Digest( &ctx, 128, pubB );
                                                                    OALSecureHash_Finish( &ctx, u );
                                                                    
                                                                    if( mp_read_unsigned_bin( &u_bn, u, 32 ) == MP_OKAY )
                                                                    {
                                                                        if( mp_read_unsigned_bin( &A_bn, pubA, 128 ) == MP_OKAY )
                                                                        {
                                                                            /*v^u,N*/
                                                                            if( mp_exptmod( &v_bn, &u_bn, &SRP_N, &temp_bn) == MP_OKAY )
                                                                            {
                                                                                /* (A*(v^u,N)) % N */
                                                                                if( mp_mul(&A_bn,&temp_bn, &temp2_bn) == MP_OKAY )
                                                                                {
                                                                                    if( mp_mod(&temp2_bn, &SRP_N, &temp3_bn) == MP_OKAY )
                                                                                    {
                                                                                        /* S = ((A*(v^u,N)) % N) ^ b, N */
                                                                                        if( mp_exptmod( &temp3_bn, &b_bn, &SRP_N, &S_bn) == MP_OKAY )
                                                                                        {
                                                                                            /* Write S */
                                                                                            size = 128;
                                                                                            if( mp_to_unsigned_bin_n_pad(&S_bn, srpS, size) == MP_OKAY )
                                                                                            {
                                                                                                /* Ksrp = SHA(S) */
                                                                                                OALSecureHash_Start( &ctx );
                                                                                                OALSecureHash_Digest( &ctx, 128, srpS );
                                                                                                OALSecureHash_Finish( &ctx, Ksrp );

                                                                                                ret = TRUE;
                                                                                            }
                                                                                        }
                                                                                    }
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        mp_clear( &temp3_bn );
                                    }
                                    mp_clear( &temp2_bn );
                                }
                                mp_clear( &temp_bn );
                            }
                            mp_clear( &S_bn );
                        }
                        mp_clear( &u_bn );
                    }
                    mp_clear( &v_bn );
                }
                mp_clear( &b_bn );
            }
            mp_clear( &A_bn );
        }
        mp_clear( &B_bn );
    }
    return ret;
}

boolean OALSRP_CalculateNode( uint8* privateA, uint8* pubA, uint8* pubB, uint8* partial_x, uint8* srps, uint8* srpS, uint8* Ksrp, uint8* validA )
{
    mp_int t1;
    mp_int t2;
    mp_int u;
    mp_int x;
    mp_int w;
    mp_int Si;
    mp_int B_bn;
    mp_int priA;
    OALSecureHashContext shacontext;
    OALSecureHashDigest digestU;
    OALSecureHashDigest digestX;
    long unsigned int size = 128;
    boolean ret = FALSE;
    
    if( mp_init(&t1) == MP_OKAY )
    {   
        if( mp_init(&t2) == MP_OKAY )
        {
            if( mp_init(&u) == MP_OKAY )
            {
                if( mp_init(&x) == MP_OKAY )
                {
                    if( mp_init(&w) == MP_OKAY )
                    {
                        if( mp_init(&Si) == MP_OKAY )
                        {
                            if( mp_init(&B_bn) == MP_OKAY )
                            {
                                if( mp_init(&priA) == MP_OKAY )
                                {
                                    /* u = SHA( PUB_A + PUB_B) */
                                    OALSecureHash_Start(&shacontext);
                                    OALSecureHash_Digest(&shacontext, SRP_PUBLIC, pubA);
                                    OALSecureHash_Digest(&shacontext, SRP_PUBLIC, pubB);
                                    OALSecureHash_Finish(&shacontext, digestU);
                                    
                                    OALSecureHash_Start(&shacontext);
                                    OALSecureHash_Digest(&shacontext, 32, srps);
                                    OALSecureHash_Digest(&shacontext, 32, partial_x);
                                    OALSecureHash_Finish(&shacontext, digestX);

                                    /* set up bignumber */
                                    if( mp_read_unsigned_bin(&u, digestU,32) == MP_OKAY )
                                    {
                                        if( mp_read_unsigned_bin(&B_bn, pubB, SRP_PUBLIC) == MP_OKAY )
                                        {
                                            if( mp_read_unsigned_bin(&priA, privateA, SRP_PRIVATE) == MP_OKAY )
                                            {
                                                if( mp_read_unsigned_bin(&x, digestX,32) == MP_OKAY )
                                                {
                                                    /* w = SRPn - ( SRPg ^ x, SRPn ) */
                                                    if( mp_exptmod(&SRP_g, &x, &SRP_N, &t1) == MP_OKAY )
                                                    {
                                                        if( mp_sub(&SRP_N, &t1, &w) == MP_OKAY )
                                                        {
                                                            /* S = ((B + 3w)^(a + ux), N */
                                                            /* 3 * w */
                                                            if( mp_mul(&SRP_three, &w, &t1) == MP_OKAY )
                                                            {
                                                                /* PUB_B + 3w */
                                                                if( mp_add(&B_bn, &t1, &t2) == MP_OKAY )
                                                                {
                                                                    /* u * x */
                                                                    if( mp_mul(&u, &x, &Si) == MP_OKAY )
                                                                    {
                                                                        /* PRIV_A + u*x */
                                                                        if( mp_add(&priA, &Si, &t1) == MP_OKAY )
                                                                        {
                                                                            /* S = ((PUB_B + 3w)^(PRIV_A + ux), SRP_N */
                                                                            if( mp_exptmod(&t2, &t1, &SRP_N, &Si) == MP_OKAY )
                                                                            {
                                                                                /* load S into byte buffer */
                                                                                if( mp_to_unsigned_bin_n_pad(&Si, srpS, size) == MP_OKAY )
                                                                                {
                                                                                    OALSecureHash_Start(&shacontext);
                                                                                    OALSecureHash_Digest(&shacontext, 128, srpS);
                                                                                    OALSecureHash_Finish(&shacontext, Ksrp);

                                                                                    OALSecureHash_Start(&shacontext);
                                                                                    OALSecureHash_Digest(&shacontext, SRP_PUBLIC, pubA);
                                                                                    OALSecureHash_Digest(&shacontext, SRP_PUBLIC, pubB);
                                                                                    OALSecureHash_Digest(&shacontext, 128, srpS);
                                                                                    OALSecureHash_Finish(&shacontext, validA);
                                                                                    ret = TRUE;
                                                                                }
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }

                                    mp_clear(&priA);
                                }
                                mp_clear(&B_bn);
                            }
                            mp_clear(&Si);
                        }
                        mp_clear(&w);
                    }
                    mp_clear(&x);
                }
                mp_clear(&u);
            }
            mp_clear(&t2);
        }
        mp_clear(&t1);
    }
    return ret;
}

boolean OALSRP_CalculateKShared( uint8* pubA, uint8* validA, uint8* srpS, uint8* validB, uint8* KEY, uint8* Ksrp, uint32 d_len, uint8* domain, uint8* key )
{
    OALSecureHashContext ctx;
    OALCryptExpandedKey ekey;
    int32 i = 0;
    uint8 temp[32];
    boolean ret = TRUE;

    PCRMem_Copy( temp, validB, 32 );
    PCR_Validate( DOFCipherRegistry_AES256Cipher );
    DOFCipherRegistry_AES256Cipher->GenerateKeyState( ekey, Ksrp );
    DOFCipherRegistry_AES256Cipher->Encrypt( ekey, &temp[0] );
    DOFCipherRegistry_AES256Cipher->Encrypt( ekey, &temp[16] );

    for( ; i < 32 ; ++i )
    {
        temp[i] = KEY[i] ^ temp[i];
    }

    PCRMem_Copy( key, temp, 32);

    OALSecureHash_Start( &ctx);
    OALSecureHash_Digest( &ctx, 128, pubA);
    OALSecureHash_Digest( &ctx, 32, validA);
    OALSecureHash_Digest( &ctx, 128, srpS);
    OALSecureHash_Digest( &ctx, 32, key);
    OALSecureHash_Digest( &ctx, d_len, domain);
    OALSecureHash_Finish( &ctx, temp );

    for( i = 0; i < 32 ; ++i )
    {
        if( temp[i] != validB[i] )
            ret = FALSE;
    }

    if(!ret)
        PCRMem_Set( key, 0, 32);

    return ret;
}

boolean OALSRP_CalculateValidBAndKey( uint8* pubA, uint8* validA, uint8* srpS, uint8* key, uint32 d_len, uint8* domain, uint8* Ksrp, uint8* validB, uint8* KEY )
{
    OALSecureHashContext ctx;
    OALCryptExpandedKey ekey;
    int32 i = 0;

    OALSecureHash_Start( &ctx);
    OALSecureHash_Digest( &ctx, 128, pubA);
    OALSecureHash_Digest( &ctx, 32, validA);
    OALSecureHash_Digest( &ctx, 128, srpS);
    OALSecureHash_Digest( &ctx, 32, key);
    OALSecureHash_Digest( &ctx, d_len, domain);
    OALSecureHash_Finish( &ctx, validB );

    PCRMem_Copy( KEY, validB, 32 );
    PCR_Validate( DOFCipherRegistry_AES256Cipher );
    DOFCipherRegistry_AES256Cipher->GenerateKeyState( ekey, Ksrp );
    DOFCipherRegistry_AES256Cipher->Encrypt( ekey, &KEY[0] );
    DOFCipherRegistry_AES256Cipher->Encrypt( ekey, &KEY[16] );

    for( ; i < 32 ; ++i )
    {
        KEY[i] = KEY[i] ^ key[i];
    }

    return TRUE;
}

/**
@}
**/
