/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/stream.h>
#include <internal/dofinternal.h>

/** @internal @addtogroup OALStream
@{
**/

extern PCRMutex ref_count_mutex;

/*
**  "Private" Type and Method Declarations
*/
typedef union DataBufPtr_t
{
    uint8*          rw;     /* read/write pointer */
    const uint8*    ro;     /* read-only  pointer */
} DataBufPtr_t;

typedef struct OALStream_t
{
    uint32      refCount;   /* reference count */

    uint32          readOffset;
    uint32          writeOffset;
    uint32          dataBufSize;
    OALStreamType   streamType;
    DataBufPtr_t    dataBuf;
} OALStream_t;


/*
**  "Public" Methods
*/

/* "Static" methods */

static boolean static_readCheck(OALStream self,
                                uint32    size);

static boolean static_writeCheck(OALStream self,
                                 uint32    size);

OALStream OALStream_CreateRW(uint32 size)
{
    /* When this OALStream() is eventually destroyed, the buffer at data IS freed, presuming it WAS allocated.*/
    OALStream newStream = DOFMem_Alloc( sizeof(OALStream_t) );
    if (newStream) 
    {
        newStream->dataBuf.rw = (uint8*)PCRMem_Alloc(size);
        if (newStream->dataBuf.rw)
        {
            newStream->refCount = 1;
            newStream->readOffset = 0;
            newStream->writeOffset = 0;
            newStream->dataBufSize = size;
            newStream->streamType = OALSTREAMTYPE_READWRITE;
            return newStream;
        }
        DOFMem_Free(newStream);
    }
    return NULL;
}

OALStream OALStream_CreateRO(const uint8 * data, uint32 length)
{
    /* When this OALStream() is eventually destroyed, the buffer at data is NOT freed, presuming it was NOT allocated.*/
    const uint32 allocSize = (uint32)( offsetof(OALStream_t, dataBuf.ro) + sizeof(DataBufPtr_t) );
    OALStream newStream = DOFMem_Alloc(allocSize);

    if (!newStream)
        return NULL;

    newStream->refCount = 1;
    newStream->readOffset = 0;
    newStream->writeOffset = length;
    newStream->dataBufSize = length;
    newStream->streamType = OALSTREAMTYPE_READONLY;
    newStream->dataBuf.ro = data;

    return newStream;
}

OALStream OALStream_CreateRWWithData(uint8 * data, uint32 size)
{
    OALStream newStream = DOFMem_Alloc( sizeof(OALStream_t) );

    if (!newStream)
        return NULL;

    newStream->refCount = 1;
    newStream->readOffset = 0;
    newStream->writeOffset = size;
    newStream->dataBufSize = size;
    newStream->streamType = OALSTREAMTYPE_READWRITE;
    newStream->dataBuf.rw = data;

    return newStream;
}

OALStream OALStream_CreateRWWrapped(uint8 * data, uint32 size)
{
    OALStream newStream = DOFMem_Alloc( sizeof(OALStream_t) );

    if (!newStream)
        return NULL;

    newStream->refCount = 1;
    newStream->readOffset = 0;
    newStream->writeOffset = size;
    newStream->dataBufSize = size;
    newStream->streamType = OALSTREAMTYPE_WRAPPED;
    newStream->dataBuf.rw = data;

    return newStream;
}

/* Note: this must not minimize the stream, there are some cases where an exact clone is required. */
OALStream OALStream_CloneToRW(OALStream source)
{
    uint32 readOffset = OALStream_GetReadOffset(source);
    uint32 sourceDataSize = source->dataBufSize;
    OALStream newStream = OALStream_CreateRW( sourceDataSize );
    const uint8* pSourceData = OALStream_GetReadPtr(source) - readOffset;

    if (!newStream)
        return NULL;

    newStream->readOffset = readOffset;
    newStream->writeOffset = OALStream_GetWriteOffset(source);
    newStream->dataBufSize = sourceDataSize;
    newStream->streamType = OALSTREAMTYPE_READWRITE;
    PCRMem_Copy(newStream->dataBuf.rw, pSourceData, sourceDataSize);

    return newStream;
}

OALStream OALStream_Create_Concat(OALStream first, OALStream second, uint32 freeSpace)
{
    uint32 sizeFirst = OALStream_GetDataSize(first);
    uint32 sizeSecond = OALStream_GetDataSize(second);
    uint32 sizeTotal = sizeFirst + sizeSecond;
    OALStream retVal;

    /* detect overflow */
    if ( sizeTotal < sizeFirst )
        return NULL;
    sizeTotal += freeSpace;
    if ( sizeTotal < freeSpace )
        return NULL;

    retVal = OALStream_CreateRW( sizeTotal );
    if (!retVal)
        return NULL;

    PCRMem_Copy( (retVal->dataBuf.rw + retVal->writeOffset), (first->dataBuf.ro + first->readOffset), sizeFirst );
    OALStream_AdjustWriteOffset( retVal, sizeFirst );
    PCRMem_Copy( (retVal->dataBuf.rw + retVal->writeOffset), (second->dataBuf.ro + second->readOffset), sizeSecond );
    OALStream_AdjustWriteOffset( retVal, sizeSecond );

    return retVal;
}

void OALStream_Retain(OALStream self)
{
    PCR_Assert(self);

    PCRMutex_Lock(ref_count_mutex);
    ++(self->refCount);
    PCR_Assert( self->refCount );   /* Rollover to 0 is disastrous and is caused by coding error of some kind. */
    PCRMutex_Unlock(ref_count_mutex);
}

void OALStream_Release(OALStream self,
                       boolean   mustLock)
{
    uint32 oldRefCount;

    if (!self)
        return;

    if(mustLock)
    {
        PCRMutex_Lock(ref_count_mutex);
    }
    PCR_Assert( self->refCount );
    oldRefCount = --(self->refCount);
    if(mustLock)
    {
        PCRMutex_Unlock(ref_count_mutex);
    }

    if(oldRefCount)
        return;

    if ( self->streamType == OALSTREAMTYPE_READWRITE )
    {
        PCRMem_Free(self->dataBuf.rw);
    }

    DOFMem_Free(self);
}

OALStreamType OALStream_GetType(OALStream self)
{
    PCR_Assert(self);

    return self->streamType;
}

uint32 OALStream_GetDataSize(OALStream self)
{
    PCR_Assert(self);
    PCR_Assert(self->writeOffset >= self->readOffset);

    return self->writeOffset - self->readOffset;
}

static boolean OALStream_IsWriteable( OALStream self )
{
    if( ( self->streamType == OALSTREAMTYPE_READWRITE ) ||
        ( self->streamType == OALSTREAMTYPE_WRAPPED ) )
    {
        return TRUE;
    }
    return FALSE;
}

uint32 OALStream_GetFreeSize(OALStream self)
{
    PCR_Assert(self);

    return(OALStream_IsWriteable(self)) ? (self->dataBufSize - self->writeOffset) : 0x00000000;
}

boolean OALStream_IncreaseFreeSize(OALStream self, uint32 sizeIncrease)
{
    uint32 newSize;
    uint8* newMem;

    PCR_Assert(self);

    /* we cant go tweaking memory we dont own */
    if (self->streamType == OALSTREAMTYPE_READONLY)
        return FALSE;

    newSize = self->dataBufSize + sizeIncrease;
    if(newSize < self->dataBufSize)
    {
        return FALSE;
    }
    newMem = PCRMem_Realloc( self->dataBuf.rw, newSize );

    if (!newMem)
    {
        return FALSE;
    }
    else
    {
        self->dataBuf.rw = newMem;
        self->dataBufSize = newSize;
        return TRUE;
    }
}

uint32 OALStream_GetWriteOffset(OALStream self)
{
    PCR_Assert(self);

    return self->writeOffset;
}

uint32 OALStream_GetReadOffset(OALStream self)
{
    PCR_Assert(self);

    return self->readOffset;
}

void OALStream_SetWriteOffset(OALStream self, uint32 offset)
{
    PCR_Assert(self);
    /* It may be nice to have a warning here if out of bounds */

    PCR_Assert(offset <= self->dataBufSize);
    if(offset <= self->dataBufSize)
    {
        self->writeOffset = offset;
        if (self->readOffset > self->writeOffset)
            self->readOffset = self->writeOffset;
    }
}

void OALStream_SetReadOffset(OALStream self, uint32 offset)
{
    PCR_Assert(self);
    /* It may be nice to have a warning here if out of bounds */

    PCR_Assert(offset <= self->dataBufSize);
    if(offset <= self->dataBufSize)
    {
        self->readOffset = offset;
    }
}

void OALStream_AdjustReadOffset(OALStream self, int32 offsetChange)
{
    boolean isOk;

    PCR_Assert(self);
    /* assert on overflow */
    isOk = (offsetChange > 0) ? (self->readOffset <  (self->readOffset + offsetChange)) :
                                (self->readOffset >= (self->readOffset + offsetChange));
    PCR_Assert(isOk);
    if(isOk)    /* It may be nice to have a warning here if out of bounds or past writeOffset */
    {
        self->readOffset += offsetChange;
        PCR_Assert(self->readOffset <= self->writeOffset); /* Postcondition */
    }
}

void OALStream_AdjustWriteOffset(OALStream self, int32 offsetChange)
{
    boolean isOk;

    PCR_Assert(self);

    /* assert on overflow */
    isOk = (offsetChange > 0) ? (self->writeOffset <  (self->writeOffset + offsetChange)) :
                                (self->writeOffset >= (self->writeOffset + offsetChange));
    PCR_Assert(isOk);
    if(isOk)        /* It may be nice to have a warning here if out of bounds or past writeOffset */
    {
        self->writeOffset += offsetChange;
        if (self->writeOffset > self->dataBufSize)
            self->readOffset = self->dataBufSize;
        PCR_Assert(self->readOffset <= self->writeOffset); /* Postcondition */
    }
}

const uint8* OALStream_GetReadPtr(OALStream self)
{
    PCR_Assert(self);

    return self->dataBuf.ro + self->readOffset;
}

uint8* OALStream_GetWritePtr(OALStream self)
{
    PCR_Assert(self);

    return(OALStream_IsWriteable(self)) ? (self->dataBuf.rw + self->writeOffset) : NULL;
}

boolean OALStream_Read(OALStream self, uint32 size, uint8* destBuf)
{
    if(static_readCheck(self, size))
    {
        PCRMem_Copy(destBuf, OALStream_GetReadPtr(self), size);
        self->readOffset += size;
        return TRUE;
    }
    return FALSE;
}

boolean OALStream_ReadToStream(OALStream self, uint32 size, OALStream dest)
{
    uint32 maxReadLen;

    PCR_Assert(self);
    PCR_Assert(dest);

    if((dest->dataBufSize - dest->writeOffset) < size)
    {
        return FALSE;
    }
    maxReadLen = self->writeOffset - self->readOffset;
    if (size > maxReadLen)
        return FALSE;

    PCRMem_Copy(OALStream_GetWritePtr(dest), OALStream_GetReadPtr(self), size);
    self->readOffset += size;
    dest->writeOffset += size;
    return TRUE;
}

boolean OALStream_Write(OALStream self, uint32 size, const uint8* srcBuf)
{
    if(static_writeCheck(self, size))
    {
        PCRMem_Copy(OALStream_GetWritePtr(self), srcBuf, size);
        self->writeOffset += size;
        return TRUE;
    }
    return FALSE;
}




boolean OALStream_Write8(OALStream self, uint8 value)
{
    if(static_writeCheck(self, sizeof(uint8)))
    {
        *(OALStream_GetWritePtr(self)) = PCRSwap_Net8(value);
        self->writeOffset += sizeof(uint8);
        return TRUE;
    }
    return FALSE;
}

boolean OALStream_Write16(OALStream self, uint16 value)
{
    if(static_writeCheck(self, sizeof(uint16)))
    {
        value = PCRSwap_Net16( value );
        PCRMem_Copy( OALStream_GetWritePtr( self ), &value,  sizeof( value ) );
        self->writeOffset += sizeof(uint16);
        return TRUE;
    }
    return FALSE;
}

boolean OALStream_Write32(OALStream self, uint32 value)
{
    if(static_writeCheck(self, sizeof(uint32)))
    {
        value = PCRSwap_Net32( value );
        PCRMem_Copy( OALStream_GetWritePtr( self ), &value,  sizeof( value ) );
        self->writeOffset += sizeof(uint32);
        return TRUE;
    }
    return FALSE;
}

boolean OALStream_Write64(OALStream self, uint64 value)
{
    if(static_writeCheck(self, sizeof(uint64)))
    {
        value = PCRSwap_Net64( value );
        PCRMem_Copy( OALStream_GetWritePtr( self ), &value,  sizeof( value ) );
        self->writeOffset += sizeof(uint64);
        return TRUE;
    }
    return FALSE;
}

/* Returns TRUE for success. */
boolean OALStream_Read8(OALStream self, uint8* pValue)
{
    if(static_readCheck(self, sizeof(uint8)))
    {
        *pValue = *OALStream_GetReadPtr(self);
        self->readOffset += sizeof(uint8);

        return TRUE;
    }
    return FALSE;
}

boolean OALStream_Read16(OALStream self, uint16* pValue)
{
    if(static_readCheck(self, sizeof(uint16)))
    {
        uint16 tmp;
        PCRMem_Copy( &tmp, OALStream_GetReadPtr( self ), sizeof( tmp ) );
        *pValue = PCRSwap_Net16( tmp );
        self->readOffset += sizeof(uint16);

        return TRUE;
    }
    return FALSE;
}

boolean OALStream_Read32(OALStream self, uint32* pValue)
{
    if(static_readCheck(self, sizeof(uint32)))
    {
        uint32 tmp;
        PCRMem_Copy( &tmp, OALStream_GetReadPtr( self ), sizeof( tmp ) );
        *pValue = PCRSwap_Net32( tmp );
        self->readOffset += sizeof(uint32);

        return TRUE;
    }
    return FALSE;
}

boolean OALStream_Read64(OALStream self, uint64* pValue)
{
    if(static_readCheck(self, sizeof(uint64)))
    {
        uint64 tmp;
        PCRMem_Copy( &tmp, OALStream_GetReadPtr( self ), sizeof( tmp ) );
        *pValue = PCRSwap_Net64( tmp );
        self->readOffset += sizeof(uint64);

        return TRUE;
    }
    return FALSE;
}

static boolean static_readCheck(OALStream self,
                                uint32    size)
{
    PCR_Assert(self);
    PCR_Assert(self->writeOffset >= self->readOffset);

    return (size <= (self->writeOffset - self->readOffset));
}

static boolean static_writeCheck(OALStream self,
                                 uint32    size)
{
    PCR_Assert(self);
    PCR_Assert(OALStream_IsWriteable(self));
    PCR_Assert(self->writeOffset >= self->readOffset);

    return (size <= (self->dataBufSize - self->writeOffset));
}

/**
@}
**/

