/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/connection.h>
#include <dof/core.h>
#include <dof/error.h>
#include <dof/exception.h>
#include <dof/object.h>
#include <dof/permissionset.h>
#include <internal/securityscope.h>
#include <internal/bridgeinternal.h>
#include <internal/credentialsinternal.h>
#include <internal/mutex_pool.h>
#include <internal/objectinternal.h>
#include <internal/operationinternal.h>
#include <internal/permission_manager.h>
#include <internal/permissioninternal.h>
#include <internal/resultinternal.h>
#include <internal/runtimesysteminternal.h>
#include <internal/systeminternal.h>
#include "system_interest.h"


/** @addtogroup DOFSystem
@{
**/


/*
**  "Private" Type and Method Declarations
*/


static boolean remove_ricb( OALSListNode item, void* arg )
{
    if (((RoutingInfoChange_t*)item)->callback == ((RoutingInfoChange_t*)arg)->callback)
        return TRUE;

    return FALSE;
}


static void extendCB_Accept( DOFError status, void * context )
{
    ExtendAcceptRequestContext realContext = (ExtendAcceptRequestContext)context;

    PCR_Assert( context != NULL );
    if (status == DOFERROR_OK)
    {
        /* REVIEW: These next two calls can be checked for boolean return and log message on fail. */
        if (realContext->isOp)
            DOFSystem_AcceptRequest( DOFSystem_GetChannel( realContext->sys ), realContext->reqOp,
                                     realContext->arg2.provOp );
        else
            DOFSystem_AcceptResponse( DOFSystem_GetChannel( realContext->sys ), realContext->reqOp,
                                      realContext->arg2.result );
    }
    if (realContext->isOp)
        DOFOperation_Internal_Destroy( realContext->arg2.provOp );
    else
        DOFResult_Destroy( realContext->arg2.result );
    DOFSystem_Release( realContext->sys );
    DOFOperation_Internal_Destroy( realContext->reqOp );
    DOFMem_Free( context );
}


/** @internal
    This function looks at the system's object list to see if an object with the same objectID has already been added.

    @param node       The node from the system list of objects.
    @param searchNode The node holding the object ID to search for.

    @return TRUE if the object ID is found, and FALSE if it is not found.
**/
static boolean InSysObjList( OALSListNode node, void* arg )
{
    const sysObjectNode searchNode = (sysObjectNode)arg;
    /* Called with system mutex locked. */
    return DOFObjectID_IsEqual( ((sysObjectNode)node)->objID, searchNode->objID );
}


static boolean OpIsBeginProvide( DOFOperation op )
{
    return (op->operation_type == DOFOPERATIONTYPE_BEGINPROVIDE);
}


/*
**  "Internal Public"  Methods.
*/


void DOFSystem_CallCompleteWithException( DOFOperation op, DOFException except )
{
    /* DO NOT HOLD ANY LOCKS WHEN THIS IS CALLED!! AND DO NOT CALL THIS IF op is or will be passed through DOFRouter.*/
    if (op)
    {
        DOFOperation_Retain( op );                 /* MUST! Because ->Complete() callback below might destroy the op. */
        DOFOperation_Lock( op );
        /* NOTE: Simulate what DOFSystem's Requestor_ProcessResponse would do if it had been asked to do this work*/
        if( !op->flags.bits.completed )
        {
            op->flags.bits.completed = 1;
            op->flags.bits.cancelled = 1;
#if 0
            /* A partial solution to bug 3465. This prevents complete callback from extending and having the false
               illusion that the op can be kept alive for a while. */
            op->flags.bits.denyTimeoutExtend = 1;
#endif
            DOFException_Destroy( op->completeCallbackException );
            op->completeCallbackException = DOFException_Copy( except );
            if (op->notifyOpAcknowledgedCond)
                PCRCondition_Broadcast( op->notifyOpAcknowledgedCond );  /* Signal DOFOp_WaitAcknowledged().*/
            if (op->callbacks.pcb && op->callbacks.pcb->fns && op->callbacks.pcb->fns->Complete)
            {
                DOFOperation_Unlock( op );

                op->callbacks.pcb->fns->Complete( op->callbacks.pcb, op, except );

                DOFOperation_Lock( op );
            }
            op->flags.bits.completedAck = 1;      /* Prevent DOFSystem from calling requestor callback again. */
            op->flags.bits.blockFollowOnIncomingSignalPDUs = 1;                  /* Might be in system queue. */
            if (op->notifyOpCompleteCond)
                PCRCondition_Broadcast( op->notifyOpCompleteCond );  /* Signal DOFOp_WaitComplete(). */
        }
        DOFOperation_Unlock( op );
        DOFOperation_Destroy( op ); /*This cancels op (must because it is toast), and ALSO removes above refcount bump*/
    }
}

static void retryCallback(DOFError   error,
                          void     * context)
{
    DOFOperation op = (DOFOperation) context;

    if(error == DOFERROR_OK)
    {
        DOFSecurityScope secScope = DOFSecurityMembrane_ValidateOutboundPermissions(op->sources->channel->securityMembrane,
                                                                                    op->commandPerm, op->commandPerm);

        if(secScope != NULL)
        {
            DOFSecurityScope_Destroy(op->securityScope);
            DOFOperation_Lock(op);
            op->securityScope = secScope;
            if(OALPermissionScope_GetIdCount(DOFSecurityScope_GetOALPermissionScope(secScope)) > 0)
            {
                DOFRouter_RouteToProvider(op);
            }
            DOFOperation_Unlock(op);
        }
    }
    DOFOperation_Internal_Destroy(op);
}

void DOFSystem_CheckPermissionsRetryAndRoute(DOFOperation op)
{
    /* operation is already locked */

    DOFSecurityScope secScope = DOFSecurityMembrane_ValidateOutboundPermissions(op->sources->channel->securityMembrane,
                                                                                op->commandPerm, op->commandPerm);

    if(secScope != NULL)
    {
        DOFSecurityScope_Destroy(op->securityScope);
        op->securityScope = secScope;
        if(OALPermissionScope_GetIdCount(DOFSecurityScope_GetOALPermissionScope(secScope)) > 0)
        {
            DOFRouter_RouteToProvider(op);
        }
    }
    else
    {
        DOFOperation_Retain(op);
        OALChannel_ExtendOutboundMembranePermissions(op->sources->channel, op->commandPerm, retryCallback, op);
    }
}

/** @internal
    Inform DOFSystem that an operation has timed out.

    @param op The timedOut operation.
**/
void DOFSystem_OpTimedOutCallback( DOFOperation op )
{
    PCR_Assert( op );
    /* Called with operation mutex locked. */
    if (!op->flags.bits.cancelled && !op->flags.bits.completedAck)
    {
        op->flags.bits.timedOut = 1;        /* Cleared by DOFSystem after making a decision. */
        if (op->provider_object)
        {
            /* Send op through DOFSystem queue. DOFSystem will call app callback to see if it will extend timeout. */
            if (OpIsBeginProvide( op ))
            {
                PCR_Assert( DOFObject_GetSystem( op->provider_object ) == (DOFSystem) op->sources->channel );
                if (!DOFSystem_AcceptRequest( op->sources->channel, NULL, op ))           /* NULL: req op. */
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "DOFSystem: Could not notify of provide operation timeout.");
            }
            else
            {
                PCR_Assert( DOFObject_GetSystem( op->provider_object ) == (DOFSystem) op->sources->channel );
                if (!DOFSystem_AcceptResponse( op->sources->channel, op, NULL ))          /* NULL: result. */
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "DOFSystem: Could not notify of request operation timeout.");
            }
        }
        else if (op->operation_type == DOFOPERATIONTYPE_WATCH ||
                 op->operation_type == DOFOPERATIONTYPE_ACTIVATE ||
                 op->operation_type == DOFOPERATIONTYPE_FULLCONNECT ||
                 op->operation_type == DOFOPERATIONTYPE_CONNECT)
        {
            /* Send op through DOFSystem queue. DOFSystem will call app callback to see if it will extend timeout. */
            if (!DOFSystem_AcceptResponse( op->sources->channel, op, NULL ))                      /* NULL: result. */
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "DOFSystem: Could not notify of interest operation timeout.");
        }
        else if (op->operation_type == DOFOPERATIONTYPE_QUERY)
            DOFSystem_AcceptQueryResponse((DOFSystem)op->sources->channel, op, NULL, NULL, QUERY_CALLBACK_COMPLETE);
        else
            DOFOperation_Cancel( op );      /* What circumstances led to this path? */
    }
    else
    {
        /* This is unlikely, but cover it by cancelling the operation. */
        DOFOperation_Cancel( op ); /* router/timeout code could make this check but we are privy to the bits so we do.*/
    }
}


/** @internal
    Inform DOFSystem that an operation has cancelled before it even got through router.

    @param op The cancelled operation.
**/
void DOFSystem_OpEarlyCancelCallback( DOFOperation op )
{
    /* Called with operation mutex locked. */
    PCR_Assert( op );
    PCR_Assert( op->flags.bits.cancelled );
    PCR_Assert( op->flags.bits.denyTimeoutExtend );
    if (!op->flags.bits.completedAck)
    {
        if (op->provider_object)
        {
            /* Send op through DOFSystem queue. DOFSystem will call app callback to see if it will extend timeout. */
            if (OpIsBeginProvide( op ))
            {
                PCR_Assert( DOFObject_GetSystem( op->provider_object ) == (DOFSystem) op->sources->channel );
                if (!DOFSystem_AcceptRequest( op->sources->channel, NULL, op ))           /* NULL: req op. */
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "DOFSystem: Could not notify of provide operation timeout." );
            }
            else
            {
                PCR_Assert( DOFObject_GetSystem( op->provider_object ) == (DOFSystem) op->sources->channel );
                if (!DOFSystem_AcceptResponse( op->sources->channel, op, NULL ))          /* NULL: result. */
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "DOFSystem: Could not notify of request operation timeout." );
            }
        }
        else if (op->operation_type == DOFOPERATIONTYPE_WATCH ||
                 op->operation_type == DOFOPERATIONTYPE_ACTIVATE ||
                 op->operation_type == DOFOPERATIONTYPE_FULLCONNECT ||
                 op->operation_type == DOFOPERATIONTYPE_CONNECT)
        {
            /* Send op through DOFSystem queue. DOFSystem will call app callback to see if it will extend timeout. */
            if (!DOFSystem_AcceptResponse( op->sources->channel, op, NULL ))                      /* NULL: result. */
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "DOFSystem: Could not notify of interest operation timeout." );
        }
        else if (op->operation_type == DOFOPERATIONTYPE_QUERY)
            DOFSystem_AcceptQueryResponse((DOFSystem)op->sources->channel, op, NULL, NULL, QUERY_CALLBACK_COMPLETE);
        else
        {
            DOFOperation_Cancel( op );      /* What circumstances led to this path? */
        }
    }
    else
    {
        /* There is some reason to believe that this code will not be hit, but just in case . . . */
        DOFOperation_Cancel( op ); /* router/timeout code could make this check but we are privy to the bits so we do.*/
    }
}


/** @internal
    Check reqOp perms- is it authorized to be given us from DOF (i.e. for outbound)?

    @param sys
    @param securityConnector
    @param perm
    @param reqOp             MUST not be NULL.
    @param provOp
    @param result            For provOpNotResult FALSE, this MUST not be NULL.
    @param provOpNotResult   For TRUE, provOp used, result ignored. For FALSE, provOp ignored, result used.

    @return TRUE for permission validated right now. FALSE for invalidated, validation in progress, or for any error.
**/
validationStatus DOFSystem_ValidateOperationInboundPermissions( DOFSystem sys, DOFSecurityMembrane securityMembrane,
                                                                DOFPermission explicitPerm, DOFPermission logicalPerm,
                                                                DOFOperation reqOp, DOFOperation provOp, DOFResult result,
                                                                boolean provOpNotResult )
{
    DOFSecurityScope secScope;
    validationStatus valid = VALIDATION_ERROR;

    PCR_Assert(securityMembrane);
    PCR_Assert( reqOp );
    if ((secScope = DOFSecurityMembrane_ValidateInboundPermissions( securityMembrane, explicitPerm, logicalPerm, NULL )) == NULL)
    {
        ExtendAcceptRequestContext context = (ExtendAcceptRequestContext)
                                                DOFMem_Alloc( sizeof( struct ExtendAcceptRequestContext_t ) );

        if (context != NULL)
        {
            boolean ok = TRUE;

            PCRMem_Set( context, 0, sizeof( struct ExtendAcceptRequestContext_t ));
            if (provOpNotResult)
            {
                context->isOp = TRUE;
                context->arg2.provOp = provOp;
                if (context->arg2.provOp != NULL)
                    DOFOperation_Retain( context->arg2.provOp );
            }
            else
            {
                /* context->isOp is FALSE. */
                PCR_Assert( result );
                if ((context->arg2.result = DOFResult_Copy( result )) == NULL)
                {
                    ok = FALSE;
                }
            }

            if (ok && ((context->sys = DOFSystem_Retain( sys )) != NULL))
            {
                context->reqOp = (DOFOperation_Retain( reqOp ), reqOp);
                OALChannel_ExtendInboundMembranePermissions(&(sys->channel), explicitPerm, extendCB_Accept, context);
                valid = VALIDATION_PENDING;
            }
            else
            { 
                /*Set ok FALSE but nothing checks it now. Fall thru to return with valid as VALIDATION_ERROR*/
                DOFMem_Free(context);
            }
        }
    }
    else
    {
        if (OALPermissionScope_GetIdCount(DOFSecurityScope_GetOALPermissionScope( secScope )) != 0)
            valid = VALIDATED;
        else
            valid = INVALIDATED;
        DOFSecurityScope_Destroy( secScope );
    }

    return valid;
}


void DOFSystem_Cleanup( DOFSystem sys )
{
    PCRMutex mutex;

    PCR_Assert( sys );
    mutex = OALMutexPool_Get( sys );
    PCRMutex_Lock( mutex );
    /* Shutdown threads (and deallocate members). */
    if (sys->requestQueue)
    {
        if (OALEnqueue_Enqueue(sys->dof->enqueue, (OALEnqueueItem) sys->requestItem ))
        {
            sys->requestItem = NULL;
            sys->requestQueue = NULL;
        }
    }
    if (sys->responseQueue)
    {
        if (OALEnqueue_Enqueue(sys->dof->enqueue, (OALEnqueueItem) sys->responseItem ))
        {
            sys->responseItem = NULL;
            sys->responseQueue = NULL;
        }
    }
    if (sys->query_queue)
    {
        if (OALEnqueue_Enqueue(sys->dof->enqueue, (OALEnqueueItem) sys->queryItem ))
        {
            sys->queryItem = NULL;
            sys->query_queue = NULL;
        }
    }
    if(sys->threadPool != NULL)
    {
        if (OALEnqueue_Enqueue(sys->dof->enqueue, (OALEnqueueItem) sys->destroyPoolContext ))
        {
            sys->threadPool = NULL;
        }
    }
    PCRMutex_Unlock(mutex);
    OALMutexPool_Release(sys);
}

static OALTimer delayTimer = NULL;

boolean DOFSystem_Init( DOF dof )
{
    delayTimer = OALTimer_Create(dof->threadPool, FALSE);
    if( delayTimer )
        return TRUE;
    return FALSE;
}

OALTimer DOFSystem_GetTimer()
{
    return delayTimer;
}

void DOFSystem_Shutdown(DOF dof)
{
    OALTimer_Destroy(delayTimer);
    DOF_CleanupSystems(dof);
    DOFSystem_QueryCleanup();
}

static void DOFSystem_DestroyThreadPool(OALThreadPoolItem dnode,
                                        boolean           shutdown)
{
    DOFSystemDestroyThreadpoolContext_t * node = (DOFSystemDestroyThreadpoolContext_t*) dnode;

    OALThreadPool_Destroy(node->pool);
    DOFMem_Free(node);
}

DOFSystem DOFSystem_Create(DOF                 dof,
                           DOFSystemConfig     config,
                           DOFSecurityMembrane membrane)
{
    DOFSystem ret = (DOFSystem) DOFMem_Alloc(sizeof(DOFSystem_t));
    STATIC_ASSERT( offsetof( DOFSystem_t, channel ) == 0 );

    PCR_Assert(membrane != NULL);
    if(ret != NULL)
    {
        DOFCredentials neuteredCreds = NULL;

        if(config != NULL)
        {
            DOFCredentials configCreds = DOFSystemConfig_GetCredentials(config);

            neuteredCreds = DOFCredentials_CreateNeutered(configCreds);
            if((neuteredCreds == NULL) && (configCreds != NULL))
            {
                DOFMem_Free(ret);
                return NULL;
            }
        }
        PCRMem_Set( ret, 0, sizeof( DOFSystem_t ) );
        ret->requestItem = (DOFEnqueueContext_t*) DOFMem_Alloc(sizeof(DOFEnqueueContext_t));
        ret->responseItem = (DOFEnqueueContext_t*) DOFMem_Alloc(sizeof(DOFEnqueueContext_t));
        ret->queryItem = (DOFEnqueueContext_t*) DOFMem_Alloc(sizeof(DOFEnqueueContext_t));
        if( (ret->requestItem != NULL) &&
            (ret->responseItem != NULL) &&
            (ret->queryItem != NULL) &&
            DOFSystem_CreateInterestStructure( ret ) )
        {
            uint16 poolSize = 0;

            if(config != NULL)
            {
                poolSize = DOFSystemConfig_GetThreadpoolSize(config);

                if(poolSize > 3)
                {
                    poolSize = 3;
                }
            }
            if(poolSize > 0)
            {
                ret->threadPool = OALThreadPool_Create(poolSize);
                if(ret->threadPool != NULL)
                {
                    ret->destroyPoolContext = (DOFSystemDestroyThreadpoolContext_t*) DOFMem_Alloc(sizeof(DOFSystemDestroyThreadpoolContext_t));
                    if(ret->destroyPoolContext != NULL)
                    {
                        OALEnqueueItem_Init((OALEnqueueItem) ret->destroyPoolContext, DOFSystem_DestroyThreadPool);
                        ret->destroyPoolContext->pool = ret->threadPool;
                    }
                    else
                    {
                        OALThreadPool_Destroy(ret->threadPool);
                        ret->threadPool = NULL;
                    }
                }
            }
            else
            {
                ret->threadPool = dof->threadPool;
            }
            if(ret->threadPool != NULL)
            {
                ret->requestQueue = OALEnqueue_Create(ret->threadPool);
                ret->responseQueue = OALEnqueue_Create(ret->threadPool);
                ret->query_queue = OALEnqueue_Create(ret->threadPool);
                if( (ret->requestQueue != NULL) &&
                    (ret->responseQueue != NULL) &&
                    (ret->query_queue != NULL) )
                {
                    ret->state.credentials = neuteredCreds;
                    ret->state.poolSize = poolSize;
                    ret->state.config = DOFSystemConfig_Copy(config);
                    if((ret->state.config != NULL) || (config == NULL))
                    {
                        if(OALChannel_Ctor(&(ret->channel), FALSE, membrane,
                                        DOFSystem_AcceptRequest_Wrapper,
                                        DOFSystem_AcceptResponse_Wrapper,
                                        NULL, DOFSystem_Dtor, NULL))
                        {
                            DOFAddress sysAddr;

                            dof->remainingSystemCleanupCount++;
                            ret->dof = dof;
                            ret->identity = NULL;                /* Required. */
                            ret->objectList = NULL;
                            OALEnqueueItem_Init((OALEnqueueItem) ret->requestItem, DOF_DestroyEnqueue);
                            ret->requestItem->enqueue = ret->requestQueue;
                            OALEnqueueItem_Init((OALEnqueueItem) ret->responseItem, DOF_DestroyEnqueue);
                            ret->responseItem->enqueue = ret->responseQueue;
                            OALEnqueueItem_Init((OALEnqueueItem) ret->queryItem, DOF_DestroyEnqueue);
                            ret->queryItem->enqueue = ret->query_queue;
                            sysAddr = DOFSystemAddress_Create( ret );
                            if ( sysAddr )
                            {
                                DOFConnectionConfig connConfig = DOFConnectionConfigBuilder_BuildAndDestroy( DOFConnectionConfigBuilder_Create( DOFCONNECTIONTYPE_STREAM, sysAddr ) );

                                ret->channel.sysConnection = DOF_CreateConnection( dof, connConfig ); /* Figure out how to create this connection with a real membrane */
                                DOFConnectionConfig_Destroy( connConfig );
                            }
                            DOFAddress_Destroy( sysAddr );
                            if(ret->channel.sysConnection != NULL)
                            {
                                if(DOF_AddSystem(dof, ret))
                                {
                                    if(poolSize == 0)
                                    {
                                        ret->threadPool = NULL;
                                    }
                                    return ret;
                                }
                                /* DOFConnection_Destroy(ret->channel.sysConnection);  -- cleaned up via OALChannel_Release (I hope) */
                            }
                            OALChannel_Release(&(ret->channel));
                            return NULL;
                        }
                        DOFSystemConfig_Destroy(ret->state.config);
                    }
                }
                if(poolSize > 0)
                {
                    DOFMem_Free(ret->destroyPoolContext);
                    OALThreadPool_Destroy(ret->threadPool);
                }
            }
            OALEnqueue_Destroy(ret->requestQueue);
            OALEnqueue_Destroy(ret->responseQueue);
            OALEnqueue_Destroy(ret->query_queue);
        }
        DOFSystem_DestroyInterestStructure( ret );
        DOFCredentials_Destroy(neuteredCreds);
        DOFMem_Free(ret->requestItem);
        DOFMem_Free(ret->responseItem);
        DOFMem_Free(ret->queryItem);
        DOFMem_Free(ret);
    }
    return NULL;
}

DOFSystemState DOFSystem_CreateState(DOFSystem self)
{
    if(self != NULL)
    {
        DOFSystemState state = DOFMem_Alloc(sizeof(DOFSystemState_t));

        if(state != NULL)
        {
            state->config = DOFSystemConfig_Copy(self->state.config);
            if((state->config != NULL) || (self->state.config == NULL))
            {
                state->credentials = DOFCredentials_Copy(self->state.credentials);
                if((state->credentials != NULL) || (self->state.credentials == NULL))
                {
                    state->systemStateMembrane = DOFSecurityMembrane_Copy(self->channel.securityMembrane);
                    if(state->systemStateMembrane != NULL)
                    {
                        state->poolSize = self->state.poolSize;
                        return state;
                    }
                    DOFCredentials_Destroy(state->credentials);
                }
                DOFSystemConfig_Destroy(state->config);
            }
            DOFMem_Free(state);
        }
    }
    return NULL;
}

/*
**  "Public" Methods
*/


void DOFSystem_Dtor( OALChannel chan )
{
    uint32    remainingSystemCleanupCount;
    PCRMutex  mutex;
    DOFSystem sys = (DOFSystem)chan;

    /* Note that sys refCount is 0 now and has been for a while. */
    /*
     * Remove sys from DOF's systemList. We do this first because it removes sys from the dof->systems list before
     * DOFSystem_Cleanup() is called, and therefore prevents a cleaning-up Requestor side WATCH operation from
     * submitting itself to DOFSystem_AcceptRequest(), as it normally wants to do. If that were allowed, it could happen
     * after DOFSystem_Cleanup(), and there is a crash because sys is gone.
     */
    DOF_RemoveSystem( sys->dof, sys );

    mutex = OALMutexPool_Get( sys );
    PCRMutex_Lock( mutex );

    /* Each object created places a ref count hold on sys, and each object deallocated removes that hold. Our intent is
       that the app have destroyed all sys's objects by the time we get here. If that is true, sys->objectList is empty,
       and *(OALSListNode *)&sys->objectList is NULL. We normally CANNOT get here with anything left in sys->objectList,
       but we do not assert so because the app could get us here by improperly releasing sys too many times and
       releasing its objects not enough times.

       If it happens that sys->objectList is not empty right now, we could make the effort to empty it, using this call:
            OALSList_RemoveAll( (OALSListNode *)&sys->objectList, removeRecord, NULL );
       The only benefit of doing that is to free memory consumed for recording each object in sys->objectList. There is
       no benefit of that, however, compared to the fact that the still allocated objects themselves remain without hope
       of deallocation. So we do not make the call. */
    /* Likewise, each interest listener added places a ref count hold on sys, and each such removed removes that hold.
       Our intent is that the app have removed all sys's interest listeners by the time we get here. If the is true,
       sys->interestCallbacks is empty. We normally CANNOT get here with anything left in sys->interestCallbacks, but we
       do not assert so because the app could get us here by improperly releasing sys too many times and removing not
       enough interest listeners.

       If sys->interestCallbacks is not empty right now, we could make the effort to empty it, using this call:
            OALSList_RemoveAll( (OALSListNode *)&sys->interestCallbacks, removeRecord, NULL );
       The only benefit of doing that is to free memory consumed for recording each interest listener in
       sys->interestCallbacks. There is no benefit of that, however, compared to the fact that the still allocated
       callback information themselves remain without hope of deallocation. So we do not make the call. */
    /* This next call deallocates the 3 sys queues (by now no other code should find sys in dof->systemList). */
    DOFSystem_Cleanup( sys );
    PCRMutex_Unlock( mutex );
    OALMutexPool_Release( sys );

    /* Account for "all systems gone." */
    PCR_Assert( sys->dof );
    DOFCredentials_Destroy(sys->state.credentials);
    DOFSystemConfig_Destroy(sys->state.config);
    DOF_Lock();
    PCR_Assert( sys->dof->remainingSystemCleanupCount );
    remainingSystemCleanupCount = --sys->dof->remainingSystemCleanupCount;
    DOF_Unlock();
    PCR_Assert( sys->dof->systemsGoneEvent );
    if (!remainingSystemCleanupCount)
        PCREvent_Signal( sys->dof->systemsGoneEvent ); /* Inform possibly waiting DOF_Destroy() that all sys are gone.*/

    /* Call parent class destructor. */
    OALChannel_Dtor( (OALChannel_t* )sys );
}


void DOFSystem_Destroy( DOFSystem sys )
{
    DOFSystem_DestroyInterestStructure( sys );
    DOFSystem_Release( sys );
}


DOF DOFSystem_GetDOF( const DOFSystem self )
{
    PCR_Assert( self );
    return( self->dof );
}

DOFObject DOFSystem_CreateObject( DOFSystem sys, const DOFObjectID objectID )
{
    DOFObject retObj = NULL;

    PCR_Assert( sys );
    if (objectID)
    {
        sysObjectNode_t objSearchNode;
        sysObjectNode   pObjListNode = NULL;
        PCRMutex        mutex        = OALMutexPool_Get( sys );

        PCRMutex_Lock( mutex );
        PCR_Assert( sys->channel.securityMembrane );

        /* Try to retrieve an object with this object ID and if one is found just bump the ref count and
           return a copy of that object, otherwise create a new object. */

        /* Since the search node is just used in this function these do not need to be a full copy. */
        objSearchNode.objID = objectID;
        pObjListNode = (sysObjectNode)OALSList_Retrieve( (OALSListNode)sys->objectList,
                                                         InSysObjList, &objSearchNode );
        if (pObjListNode != NULL)
            retObj = DOFObject_Copy( pObjListNode->obj );           /* Protected by mutex to prevent race conditions. */
        else if ((pObjListNode = DOFMem_Alloc( sizeof( sysObjectNode_t ))) != NULL)
        {
            DOFObjectID copyObjectID = DOFObjectID_Copy( objectID );

            if (copyObjectID)
            {
                /* Use testSys as sys cannot become NULL at failure- it needs be valid at OALMutexPool_Release() below*/
                DOFSystem testSys = DOFSystem_Retain( sys );             /* Give new object its own reference to sys. */

                if (testSys)
                {
                    retObj = DOFMem_Alloc( sizeof( DOFObject_t ) );
                    if (retObj)
                    {
                        retObj->refCount = 1;                                      /* No memset- we fill all member. */
                        retObj->oid = copyObjectID;
                        retObj->sys = testSys;
                        /*Add retObject to System object list so retObj will be destroyed when the system is destroyed*/
                        pObjListNode->objID = retObj->oid;
                        pObjListNode->obj = retObj;
                        OALSList_Insert( (OALSListNode *)&sys->objectList, (OALSListNode)pObjListNode );
                        goto newObjSuccessOut;
                    }
                    DOFSystem_Release( testSys );
                }
                DOFObjectID_Destroy( copyObjectID );
            }
            DOFMem_Free( pObjListNode );
        }
    newObjSuccessOut:
        PCRMutex_Unlock( mutex );
        OALMutexPool_Release( sys );
    }

    return retObj;
}

boolean DOFSystem_AddRouteListener( DOFSystem self, DOFSystemRouteListener listener )
{
    PCR_Assert( self );
    if (listener)
    {
        RoutingInfoChange_t* node = DOFMem_Alloc( sizeof( RoutingInfoChange_t ) );
        if (node)
        {
            RoutingInfoChange_t * pRoutingInfoCallbacks = DOFRuntime_GetRoutingInfoCallbacks();

            node->system = self;
            node->callback = listener;
            node->next = NULL;
            OALSList_Insert( ((OALSListNode *)(&pRoutingInfoCallbacks)), (OALSListNode)node );
            return TRUE;
        }
    }

    return FALSE;
}


void DOFSystem_RemoveRouteListener( DOFSystem self, DOFSystemRouteListener listener )
{
    if (self && listener)
    {
        RoutingInfoChange_t node;
        RoutingInfoChange_t* found = NULL;
        RoutingInfoChange_t * pRoutingInfoCallbacks = DOFRuntime_GetRoutingInfoCallbacks();

        node.callback = listener;
        node.next = NULL;
        found = (RoutingInfoChange_t*)OALSList_FindAndRemove( ((OALSListNode *)(&pRoutingInfoCallbacks)), NULL,
                                                               remove_ricb, (void *)&node );
        DOFMem_Free( found );
    }
}


void DOFSystemList_Destroy( DOFSystemList self )
{
    PCR_Assert( self );
    while (self->systemCount--)
        DOFSystem_Release( self->systemArray[self->systemCount] );
    DOFMem_Free(self->systemArray);
    DOFMem_Free( self );
}

/**
@}
**/
