/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/dof.h>
#include <dof/exception.h>
#include <dof/interfaceid.h>
#include <dof/objectid.h>
#include <dof/operation.h>
#include <internal/securityscope.h>
#include <dof/providerinfo.h>
#include <dof/routeinfo.h>
#include <dof/query.h>
#include <dof/system.h>
#include <dof/runtime.h>
#include <internal/exception.h>
#include <internal/filterinternal.h>
#include <internal/interfaceidinternal.h>
#include <internal/objectinternal.h>
#include <internal/operationinternal.h>
#include <internal/providerinfo.h>
#include <internal/router.h>
#include <internal/runtimeinternal.h>
#include <internal/systeminternal.h>


/** @addtogroup DOFSystem
    Represents a view into the DOF network.
@{
**/

/*
**  "Private" Type and Method Declarations
*/

#define LIST_ADD_FACTOR     (8)                                                                /* Must be power of 2. */
#define LIST_ADD_MASK       (LIST_ADD_FACTOR - 1)

typedef struct waitContext_t
{
    PCREvent        evt;
    DOFObject       waitObject;
    boolean         waitComplete;
}waitContext_t;

typedef struct bindingChangedContext_t
{
    DOFObjectID         oid;
    DOFInterfaceID      iid;
    uint8               bindingAdded;
    DOFSecurityScope    scope;
}bindingChangedContext_t;

typedef struct beginQueryListNode_t
{
    struct beginQueryListNode_t *   next;
    DOFOperation                    op;
    DOFSystem                       sys;
    DOFQuery                        filter;
    DOFObjectIDList                 matchingObjIDs;
    DOFSecurityScope                scope;

} beginQueryListNode_t;

/*  The begin query list is used to track all active begin query operations for all systems.  the DOFSystem_BeginQuery
    function adds the query to this list, and then when the operations ends the queryOpClosing function will remove
    the query from the list. */
static PCRMutex                 waitProviderMutex;
static beginQueryListNode_t *   beginQueryList;                                        /* All users must lock router. */


boolean DOFSystem_QueryInit(void);

static boolean findBeginQueryByOp( OALSListNode item, void* op )
{
    if (((beginQueryListNode_t *)item)->op == ((DOFOperation)op))
        return TRUE;

    return FALSE;
}

/* Add a begin query operation to the global list of begin query operations. router must be locked. When this
   succeeds, any caller must put in place an eventual matching call to DOFSystem_RemoveBeginQueryNode(). */
static beginQueryListNode_t * DOFSystem_CreateAndAddQueryNode( DOFOperation op, DOFSystem sys, DOFQuery filter )
{
    /* The list accepts op refCount ownership. On our success, caller shall NOT destroy op. */
    PCR_Assert( sys );
    if ((sys = DOFSystem_Retain( sys )) != NULL)
    {
        beginQueryListNode_t * node = (beginQueryListNode_t *)DOFMem_Alloc( sizeof( beginQueryListNode_t ) );

        if (node)
        {
            PCRMem_Set( node, 0, sizeof( beginQueryListNode_t ) );
            if ((node->matchingObjIDs = (DOFObjectIDList)DOFMem_Alloc( sizeof( struct DOFObjectIDList_t ) )) != NULL)
            {
                node->filter = DOFQuery_DeepCopy( filter );
                if ((!filter || node->filter))
                {
                    DOFSecurityScope scope = OALChannel_GetSecurityScope( DOFSystem_GetChannel( sys ), FALSE );
                    /* Outbound because the scope we care about is when our response comes out from dof. */

                    if (!scope || (node->scope = DOFSecurityScope_Copy( scope )) != NULL)
                    {
                        node->sys = sys;
                        node->matchingObjIDs->count = 0;
                        node->matchingObjIDs->list = NULL;
                        node->op = (DOFOperation_Retain( op ), op);  /* Unbumped in DOFSystem_RemoveBeqinQueryNode(). */
                        OALSList_Insert( (OALSListNode *)(&beginQueryList), (OALSListNode)node );
                        return node;
                    }
                    DOFQuery_Destroy( node->filter );
                }
                DOFMem_Free( node->matchingObjIDs );
            }
            DOFMem_Free( node );
        }
        DOFSystem_Release( sys );
    }

    return NULL;
}

/*  Note: the router mutex must be locked when this function is called.
    This function will remove the begin query operation from the global list of operations. */
static void DOFSystem_RemoveBeginQueryNode( DOFOperation op )
{
    /* Called with router locked, just like all other users of beginQueryList. */
    beginQueryListNode_t * node;

    if ((node = (beginQueryListNode_t *)OALSList_FindAndRemove((OALSListNode *)(&beginQueryList), NULL,
                                                                findBeginQueryByOp, op)) != NULL)
    {
        /* Removed the node. */
        DOFQuery_Destroy( node->filter );
        DOFSystem_Release( node->sys );
        DOFSecurityScope_Destroy( node->scope );
        DOFObjectIDList_Destroy( node->matchingObjIDs );
        /* Note: all callers either cancel op themselves or know the op is already cancelled. */
        DOFOperation_Internal_Destroy( node->op );                  /* Unbump from DOFSystem_CreateAndAddQueryNode(). */
        DOFMem_Free( node );
    }
}

static void WaitObjectInterfaceAddedCbk( DOFSystemQueryCallback self, DOFOperation op, DOFObjectID objectID,
                                         DOFInterfaceID interfaceID )
{
    if (op)
    {
        waitContext_t * waitContextPtr;

        PCRMutex_Lock( waitProviderMutex );
        waitContextPtr = DOFOperation_GetContext( op );
        if (waitContextPtr && !waitContextPtr->waitComplete)
        {
            /* Create and save new object in the operation context, then signal that an object is found, then return. */
            DOFSystem sys = DOFOperationQuery_GetSystem( op );

            PCR_Assert( sys );
            waitContextPtr->waitComplete = TRUE;
            waitContextPtr->waitObject = DOFSystem_CreateObject( sys, objectID );
            if (objectID && !waitContextPtr->waitObject)                    /* Or can we assert objectID? */
            {
                /* Code waiting for our signal can see NULL waitContextPtr->waitObject, but we have no specific way to
                   say that we were given an OID and could not create the obj. Least we can do is to log the travesty.*/
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFSystem: WaitObjectInterfaceAddedCbk() failed to allocate required object." );
            }
            PCR_Assert( waitContextPtr->evt );
            PCREvent_Signal( waitContextPtr->evt );
        }
        PCRMutex_Unlock( waitProviderMutex );
    }
}

static void WaitObjectQueryCompleteCbk (DOFSystemQueryCallback self, DOFOperation op, DOFException except)
{
    if (op != NULL)
    {
        waitContext_t * waitContextPtr;

        PCRMutex_Lock(waitProviderMutex);
        waitContextPtr = DOFOperation_GetContext(op);
        if (waitContextPtr && !waitContextPtr->waitComplete)
        {
            waitContextPtr->waitComplete = TRUE;
            waitContextPtr->waitObject = NULL;
            PCREvent_Signal(waitContextPtr->evt);
        }
        PCRMutex_Unlock(waitProviderMutex);
    }
}

typedef struct query_response_fn_t
{
    OALEnqueueItem_t super;
    DOFOperation op;
    DOFObjectID oid;
    DOFInterfaceID iid;
    uint32 callback_type;
} query_response_fn_t;

void query_response_fn(OALThreadPoolItem dnode, boolean shutdown)
{
    /* Not static only because unit test needs to call it. */
    query_response_fn_t * node = (query_response_fn_t *)dnode;

    PCR_Assert( dnode != NULL );
    if (!shutdown)
    {
        uint32 callback_type = node->callback_type;

        DOFOperation_Lock( node->op );
        if (!node->op->flags.bits.cancelled || callback_type == QUERY_CALLBACK_COMPLETE)
        {
            switch(callback_type)
            {
                case QUERY_CALLBACK_INTERFACE_ADDED:
                    DOFOperation_Unlock( node->op );
                    if (node->op->callbacks.qcb && node->op->callbacks.qcb->fns &&
                        node->op->callbacks.qcb->fns->InterfaceAdded)
                       node->op->callbacks.qcb->fns->InterfaceAdded( node->op->callbacks.qcb, node->op, node->oid,
                                                                     node->iid );
                break;
                case QUERY_CALLBACK_INTERFACE_REMOVED:
                    DOFOperation_Unlock( node->op );
                    if (node->op->callbacks.qcb && node->op->callbacks.qcb->fns &&
                        node->op->callbacks.qcb->fns->InterfaceRemoved)
                       node->op->callbacks.qcb->fns->InterfaceRemoved( node->op->callbacks.qcb, node->op, node->oid,
                                                                       node->iid );
                break;
                case QUERY_CALLBACK_OBJECT_REMOVED:
                    DOFOperation_Unlock( node->op );
                    if (node->op->callbacks.qcb && node->op->callbacks.qcb->fns &&
                        node->op->callbacks.qcb->fns->ProviderRemoved)
                       node->op->callbacks.qcb->fns->ProviderRemoved( node->op->callbacks.qcb, node->op, node->oid );
                break;
                case QUERY_CALLBACK_COMPLETE:
                    if (!node->op->flags.bits.completedAck)
                    {
                        DOFException except = NULL;
                        if (node->op->flags.bits.timedOut)
                            except = DOFEXCEPTION_TIMEOUT;
                        node->op->flags.bits.completed = TRUE;
                        node->op->flags.bits.timedOut = FALSE;
                        if (node->op->callbacks.qcb && node->op->callbacks.qcb->fns &&
                            node->op->callbacks.qcb->fns->Complete)
                        {
                            DOFOperation_Unlock( node->op );

                            node->op->callbacks.qcb->fns->Complete( node->op->callbacks.qcb, node->op, except );

                            DOFOperation_Lock( node->op );
                        }
                        if (node->op->flags.bits.completed)
                        {
                            node->op->flags.bits.completedAck = TRUE;
                            node->op->completeCallbackException = DOFException_Copy( except );
                            if (node->op->notifyOpCompleteCond)
                            {
                                /*No ref_count bump- app that _WaitComplete's does not _destroy until after its return*/
                                PCRCondition_Broadcast( node->op->notifyOpCompleteCond );/*Signal DOFOp_WaitComplete()*/
                            }
                            DOFOperation_Cancel( node->op );/*In some cases we were simply notified- we must do cancel*/
                        }
                        else
                        {
                            /*  Callback successfully extended respOp timeout. The "save" worked. Won't happen if either
                                .cancelled or .denyTimeoutExtend are TRUE. */
                            PCR_Assert( !node->op->flags.bits.cancelled );
                            PCR_Assert( !node->op->flags.bits.denyTimeoutExtend );
                            PCR_Assert( !node->op->flags.bits.completedAck );
                        }
                        if (node->op->notifyOpAcknowledgedCond)
                        {
                            /*No ref_count bump- app that _WaitAcknowledged's does not _destroy until after its return*/
                            PCRCondition_Broadcast( node->op->notifyOpAcknowledgedCond );
                        }
                    }
                    DOFOperation_Unlock( node->op );
                break;
                default:
                    DOFOperation_Unlock( node->op );
                    PCR_Assert( FALSE );
                break;
            }
        }
        else
            DOFOperation_Unlock( node->op );
    }

    DOFOperation_Internal_Destroy( node->op );
    DOFObjectID_Destroy( node->oid );
    DOFInterfaceID_Destroy( node->iid );
    DOFMem_Free( node );
}


/** @internal
    This function is used as a callback with router to iterate through each interface for a given object

    @param arg     A structure with the system, operation, oid, iid, and other information used to process the response.
**/
static void CheckInterfaceFilter(DOFRouter_Traverse_arg_t * arg)
{
    if ((arg->filter == NULL) || DOFQuery_IsRestricted(arg->filter, arg->iid))
        DOFSystem_AcceptQueryResponse( arg->system, arg->op, arg->oid, arg->iid, arg->binding_added );
}


/** @internal
    Called by DOFRouter to determine if the object is included in the DOFQuery  specified

    @param arg     A structure with the system, operation, oid, iid, and other information used to process the response.
**/
static void CheckObjectFilter( DOFRouter_Traverse_arg_t * arg )
{
    DOFObjectIDList alreadyMatching;
    uint32          i;

    PCR_Assert( arg );
    PCR_Assert( arg->context );
    alreadyMatching = arg->context;
    PCR_Assert( alreadyMatching );                                 /* This structure can be empty, but it must exist. */
    /* Check to see if this object already matches */
    for (i = 0; i < alreadyMatching->count; i++)
    {
        if (DOFObjectID_Compare( *(alreadyMatching->list + i), arg->oid ) == 0)
            return; /* Object already in alreadyMatching; nothing else to do.*/
    }

    if (arg->filter == NULL || DOFQuery_QueryMatch( arg->filter, arg->oid, arg->securityScope ))
    {
        /* Object fits the filter, so: 1) add it to alreadyMatching, and
                                       2) if a begin query, look at the interfaces and call the callback if applicable*/
        boolean     success = FALSE;
        DOFObjectID tempOID = DOFObjectID_Copy( arg->oid );

        if (!arg->oid || tempOID)                          /* Or can we assert arg->oid? */
        {
            /* Allocate or reallocate the already matching list and add the new objectID. */
            if (alreadyMatching->count == 0)
            {
                if ((alreadyMatching->list = DOFMem_Alloc( sizeof( DOFObjectID ) * LIST_ADD_FACTOR )) != NULL)
                {
                    success = TRUE;
                    /* Taken together, these next two lines completely initialize alreadyMatching->list. Users shall
                       not require NULL backfilling. */
                    *alreadyMatching->list = tempOID;
                    alreadyMatching->count = 1;
                }
            }
            else
            {
                if ((alreadyMatching->count & LIST_ADD_MASK) == 0)
                {
                    /* Here at ->count of modulo LIST_ADD_FACTOR, so reallocate alreadyMatching list. */
                    DOFObjectID * tempMatching = DOFMem_Realloc( alreadyMatching->list,
                                                 (sizeof( DOFObjectID ) * (alreadyMatching->count + LIST_ADD_FACTOR)) );

                    if (tempMatching)
                    {
                        alreadyMatching->list = tempMatching;
                        success = TRUE;
                    }
                    else
                    { /* We keep the list as it exists, but we do not add to it. */
                    }
                }
                else
                    success = TRUE;      /* Previous allocation or reallocation is sufficient to hold this new entry. */

                if (success)
                {
                    /* Since we use count as an offset, make the assignment before incrementing count. */
                    *(alreadyMatching->list + alreadyMatching->count) = tempOID;
                    alreadyMatching->count++;
                    /* Taken together, the above two lines completely initialize alreadyMatching->list. Users shall not
                       require NULL backfilling. */
                }
            }

            if (!success)
                DOFObjectID_Destroy( tempOID );
        }

        if (!success)
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFSystem: CheckObjectFilter() fails allocation" );

        if (success && arg->op != NULL)
        {
            /* For begin query, check each interface against the restricted list and call the application callback
               function where applicable. */
            DOFRouter_TraverseAllInterfaces( CheckInterfaceFilter, arg );
        }
    }
}


/** @internal
    This function is used as a callback with router to iterate throught each interfaceID for a given object.
    each interface will be added to the DOFInterfaceID list (held in the context pointer).

    @param  arg     A structure with the system, operation, oid, iid, and other information used to process the response
**/
static void BuildInterfaceIDList( DOFRouter_Traverse_arg_t * arg )
{
    DOFInterfaceID tempIID;
    boolean        success = FALSE;

    PCR_Assert( arg );
    PCR_Assert( arg->context );
    tempIID = DOFInterfaceID_Copy( arg->iid );
    if (!arg->iid || tempIID)                      /* Or can we assert arg->iid? */
    {
        DOFInterfaceIDList pIIDList = (DOFInterfaceIDList)arg->context;

        /* Allocate or reallocate the already matching list and add the new IID. */
        if (pIIDList->count == 0)
        {
            if ((pIIDList->list = DOFMem_Alloc( sizeof( DOFInterfaceID ) * LIST_ADD_FACTOR )) != NULL)
            {
                *pIIDList->list = tempIID;
                pIIDList->count = 1;
                success = TRUE;
            }
        }
        else
        {
            if ((pIIDList->count & LIST_ADD_MASK) == 0)
            {
                /* Here at ->count of modulo LIST_ADD_FACTOR, so reallocate alreadyMatching list. */
                DOFInterfaceID * tempList = DOFMem_Realloc( pIIDList->list,
                                                     (sizeof( DOFInterfaceID ) * (pIIDList->count + LIST_ADD_FACTOR)) );

                if (tempList)
                {
                    pIIDList->list = tempList;
                    success = TRUE;
                }
            }
            else
                success = TRUE;          /* Previous allocation or reallocation is sufficient to hold this new entry. */

            if (success)
            {
                /* Since we are using the count as an offset, make the assignment before incrementing count. */
                *(pIIDList->list + pIIDList->count) = tempIID;
                pIIDList->count++;
            }
        }
        if (!success)
            DOFInterfaceID_Destroy( tempIID );
    }
    if (!success)
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFSystem: BuildInterfaceIDList() fails allocation." );
}

static void incrementInterfaceCount(DOFRouter_Traverse_arg_t * arg)
{
    uint32 * countPtr = (uint32 *)arg->context;

    PCR_Assert(countPtr);

    *countPtr += 1;
}

static uint32 getObjectInterfaceCount(DOFRouter_Traverse_arg_t * arg)
{
    uint32 count = 0;

    arg->context = &count;
    DOFRouter_TraverseAllInterfaces(incrementInterfaceCount, arg);
    return count;
}

static boolean traverseBeginQueryList(OALSListNode item, void* arg)
{
    bindingChangedContext_t* ctx = (bindingChangedContext_t*)arg;
    beginQueryListNode_t *      node = (beginQueryListNode_t *)item;
    DOFRouter_Traverse_arg_t    bindingArgs;
    uint32                      j;
    boolean                     alreadyMatching = FALSE;

    /* First, fill bindingArgs with the information given us by DOFSystem_BindingChanged(). */
    bindingArgs.oid = ctx->oid;
    bindingArgs.iid = ctx->iid;
    bindingArgs.securityScope = ctx->scope;
    bindingArgs.binding_added = ctx->bindingAdded;
    /* Second, fill remaining members of bindingArgs. */
    bindingArgs.op = node->op;
    bindingArgs.filter = node->filter;
    bindingArgs.system = node->sys;
    bindingArgs.context = node->matchingObjIDs;

    for (j = 0; j < node->matchingObjIDs->count; ++j)
    {
        if (DOFObjectID_Compare(*(node->matchingObjIDs->list + j), ctx->oid) == 0)
        {
            if (ctx->bindingAdded == QUERY_CALLBACK_INTERFACE_ADDED)
            {
                /*  If the binding was added and the object already matches then just call the interface added
                    callback for this binding. */
                if (DOFQuery_IsRestricted(node->filter, ctx->iid))
                    DOFSystem_AcceptQueryResponse(node->sys, node->op, ctx->oid, ctx->iid, ctx->bindingAdded);
            }
            else if (ctx->bindingAdded == QUERY_CALLBACK_INTERFACE_REMOVED)
            {
                DOFObjectID removedObject = NULL;

                if ((node->filter != NULL) && ((DOFQuery_QueryMatch(node->filter, ctx->oid, node->scope) == FALSE) ||
                                               (!getObjectInterfaceCount(&bindingArgs))))
                {
                    removedObject = *(node->matchingObjIDs->list + j);

                    /* remove the object from the query already matching list */
                    if (node->matchingObjIDs->count == 1)
                    {
                        /* since there is only one in the list free list */
                        DOFMem_Free(node->matchingObjIDs->list);
                        node->matchingObjIDs->count = 0;
                        node->matchingObjIDs->list = NULL;
                    }
                    else
                    {
                        PCRMem_Move(node->matchingObjIDs->list + j, node->matchingObjIDs->list + j +1,
                            ((uint8 *)node->matchingObjIDs->list + node->matchingObjIDs->count*sizeof(DOFObjectID)) -
                            (uint8 *)(node->matchingObjIDs->list + j + 1));
                        node->matchingObjIDs->count--;
                        /* see if we need to realloc (shrink) the list. */
                        if (!(node->matchingObjIDs->count & LIST_ADD_MASK))
                        {
                            DOFObjectID * tempObjIDRef;

                             tempObjIDRef = DOFMem_Realloc(node->matchingObjIDs->list,
                                    sizeof(DOFObjectID) * (node->matchingObjIDs->count));
                             if (tempObjIDRef)
                                node->matchingObjIDs->list = tempObjIDRef;
                        }
                    }

                    /* since this no longer fits the filter remove all bindings including this one */
                    DOFRouter_TraverseAllInterfaces(CheckInterfaceFilter, &bindingArgs);
                }

                /* call the interface removed callback for this binding */
                if (DOFQuery_IsRestricted(node->filter, ctx->iid))
                    DOFSystem_AcceptQueryResponse(node->sys, node->op, ctx->oid, ctx->iid, ctx->bindingAdded);

                if (removedObject)
                {
                    /* call the application callback for object removed */
                    /* Note: after the object complete callback is called the object ID will be freed */
                    DOFSystem_AcceptQueryResponse(node->sys, node->op, removedObject, NULL,
                                                  QUERY_CALLBACK_OBJECT_REMOVED);
                    DOFObjectID_Destroy(removedObject);
                }
            }

            /* save the fact that the objectID was in the matching list, even if it was removed */
            alreadyMatching = TRUE;
            break;
        }
    }
    if ((alreadyMatching == FALSE) && (ctx->bindingAdded == QUERY_CALLBACK_INTERFACE_ADDED))
    {
        /*  If the object was added and does not already match see if it fits the filter now,
            if so then call the interface added callback for each interface. If the object does
            not already match and the interface was removed then there is nothing to do. */
        CheckObjectFilter(&bindingArgs);
    }

    return FALSE;
}

/** @internal

    @param  o               The observable for the operation.
    @param  beginQueryOp    The begin query that is closing.
**/
static void queryOpClosing(OALObserver self, OALObservable o, void* beginQueryOp)
{
    DOFOperation op;

    PCR_Assert( o != NULL );
    PCR_Assert( beginQueryOp != NULL );
    op = (DOFOperation)beginQueryOp;

    DOFOperation_Lock( op );
    DOFRouter_Lock();

    PCR_Assert( op->flags.bits.cancelled );   /* verify that the operation was cancelled */
    DOFSystem_RemoveBeginQueryNode( op );

    /* call the query compete application callback */
    DOFSystem_AcceptQueryResponse( (DOFSystem)op->sources->channel, beginQueryOp, NULL, NULL, QUERY_CALLBACK_COMPLETE );

    DOFRouter_Unlock();
    DOFOperation_Unlock( op );
    DOFOperation_Internal_Destroy( (DOFOperation)beginQueryOp );
}


/*
**  "Public" Methods
*/

void DOFSystem_AcceptQueryResponse( DOFSystem sys, DOFOperation op, DOFObjectID oid,
                                    DOFInterfaceID iid, uint32 binding_added )
{
    if (!oid || (oid = DOFObjectID_Copy( oid )) != NULL)
    {
        if (!iid || (iid = DOFInterfaceID_Copy( iid )) != NULL)
        {
            query_response_fn_t * node = (query_response_fn_t *)DOFMem_Alloc( sizeof( query_response_fn_t ) );

            if (node != NULL)
            {
                OALEnqueueItem_Init( (OALEnqueueItem)node, query_response_fn );
                node->op = (DOFOperation_Retain( op ), op);
                node->oid = oid;
                node->iid = iid;
                node->callback_type = binding_added;
                PCR_Assert( sys->query_queue );
                if (OALEnqueue_Enqueue( sys->query_queue, (OALEnqueueItem)node ))
                    return;                                        /* Success. */
                else
                {
                    DOFOperation_Internal_Destroy( op );
                    DOFMem_Free( node );
                    /* Since our copies of the OID and IID are not passed and we must consume them, destroy them now. */
                }
            }
            DOFInterfaceID_Destroy( iid );
        }
        DOFObjectID_Destroy( oid );
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFSystem: _AcceptQueryResponse() failed allocation." );
}

void DOFSystem_BindingChanged ( DOFSecurityScope securityScope, DOFObjectID oid, DOFInterfaceID iid,
                                uint32 binding_added )
{
    bindingChangedContext_t ctx;

    /*  loop through the query list and for each query see if the changes causes the object to fit
        the filter for any query on any system.  */
    ctx.oid = oid;                                                /* Fill all 4 members of ctx. */
    ctx.iid = iid;
    ctx.bindingAdded = binding_added;
    ctx.scope = securityScope;
    PCR_Assert((binding_added | 1) == 1);
    DOFRouter_Lock();

    /* Traverse the list of begin query operations and for each operation see if the change */
    OALSList_Traverse( (OALSListNode)beginQueryList, traverseBeginQueryList, &ctx );

    DOFRouter_Unlock();
}

DOFObjectIDList DOFSystem_Query( const DOFSystem sys, const DOFQuery  filter )
{
    DOFObjectIDList retVal = (DOFObjectIDList)DOFMem_Alloc( sizeof(struct DOFObjectIDList_t) ); /*8 bytes: uint32, ptr*/

    if (retVal != NULL)
    {
        DOFRouter_Traverse_arg_t    traverseArg;
        DOFSecurityScope scope = OALChannel_GetSecurityScope( DOFSystem_GetChannel( sys ), FALSE );
        /* Outbound because the scope we care about is when our response comes out from dof. */

        PCR_Assert( sys != NULL );
        traverseArg.system = sys;
        traverseArg.filter = filter;
        traverseArg.op = NULL;
        traverseArg.context = retVal;
        traverseArg.securityScope = scope;
        retVal->list = NULL;
        retVal->count = 0;

        /*  Traverse all objects will call the callback for each object in the router's tables.
            CheckObjectFilter will add each object to the list once if it matches the provided filter */
        DOFRouter_TraverseAllObjects(CheckObjectFilter, &traverseArg, DOFROUTERLOCATION_BOTH);
    }

    return retVal;
}


DOFOperation DOFSystem_BeginQuery( const DOFSystem sys, const DOFQuery  filter, uint32 timeout,
                                   const DOFSystemQueryCallback callback, void * context )
{
    DOFOperation             beginQueryOp;
    DOFRouter_Traverse_arg_t traverseArg;

    PCR_Assert( sys != NULL );
    beginQueryOp = DOFOperation_Query_Create( DOFSystem_GetChannel( sys ), timeout, context, filter, callback );
    if (beginQueryOp != NULL)
    {
        boolean success = FALSE;

        /* This next bumps beginQueryOp's ref count. queryOpClosing correspondingly unbumps it. */
        if (DOFOperation_AddOpWillCloseObserver( beginQueryOp, queryOpClosing ))
        {
            beginQueryListNode_t * queryNode;

            DOFOperation_Lock( beginQueryOp );               /* Note: use standard lock order- op first, then router. */
            DOFRouter_Lock();
            /* Add query op to the list, transferring refCount ownership to it. On success, we do NOT destroy op. */
            queryNode = DOFSystem_CreateAndAddQueryNode( beginQueryOp, sys, filter );
            if (queryNode)
            {
                if (OALOperationManager_AddOperation( beginQueryOp, beginQueryOp->opID, FALSE ) == DOFERROR_OK)
                {
                    traverseArg.system = sys;
                    traverseArg.filter = filter;
                    traverseArg.op = beginQueryOp;
                    traverseArg.binding_added = QUERY_CALLBACK_INTERFACE_ADDED;
                    traverseArg.securityScope = queryNode->scope;
                    traverseArg.context = queryNode->matchingObjIDs;

                    /*TraverseAllObjects() calls the callback function for each available binding that fits the filter*/
                    DOFRouter_TraverseAllObjects( CheckObjectFilter, &traverseArg,
                                                  DOFROUTERLOCATION_BOTH );
                    success = TRUE;
                }
                else
                    DOFSystem_RemoveBeginQueryNode( beginQueryOp );                /* And so ownership reverts to us. */
            }
            /*Decisions were made w/ op and router locked, so no race condition about inserting/removing from the list*/
            DOFRouter_Unlock();
            DOFOperation_Unlock( beginQueryOp );
        }
        if (!success)
        {
            DOFOperation_Destroy(beginQueryOp);/*Also cancels op so queryOpClosing gets called to unbump last refCount*/
            return NULL;
        }
    }

    return beginQueryOp;
}


DOFInterfaceIDList DOFSystem_GetProviderInterfaces( const DOFSystem sys, const DOFObjectID objectID )
{
    PCR_Assert( sys != NULL );
    if (objectID != NULL)
    {
        DOFRouter_Traverse_arg_t traverseArg;
        DOFInterfaceIDList       retVal = (DOFInterfaceIDList)DOFMem_Alloc( sizeof( struct DOFInterfaceIDList_t ) );
                                                                                  /* 8 bytes- uint32 and pointer. */

        if (retVal)
        {
            traverseArg.system = sys;
            traverseArg.oid = objectID;
            traverseArg.context = retVal;
            retVal->list = NULL;
            retVal->count = 0;
            traverseArg.securityScope = OALChannel_GetSecurityScope( DOFSystem_GetChannel( sys ), FALSE );
            /* Outbound because the scope we care about is when our response comes out from dof. */

            /*  Traverse all Interfaces will call the callback function for each interface associated with an object.
                BuildInterfaceIDList will simply add each interface to the list */
            DOFRouter_TraverseAllInterfaces(BuildInterfaceIDList, &traverseArg);
            return retVal;
        }
    }
    return NULL;
}


DOFObject DOFSystem_WaitProvider( const DOFSystem sys, const DOFObjectID objectID, uint32 nInterfaces,
                                  const DOFInterfaceID interfaceIDs[], uint32 timeout )
{
    waitContext_t waitContext;

    PCR_Assert( sys != NULL );
    waitContext.waitObject = NULL;
    if (objectID != NULL && nInterfaces > 0 && interfaceIDs != NULL)
    {
        boolean                                         success       = FALSE;
        static const struct DOFSystemQueryCallbackFns_t WaitObjectFns = { WaitObjectQueryCompleteCbk,
                                                                          WaitObjectInterfaceAddedCbk, NULL, NULL };
        static struct DOFSystemQueryCallback_t WaitObjectCallback     = { &WaitObjectFns };

        /* Internally we will create a begin query operation and set up an interface added callback so that we can
           determine when an interface is added and return the object. */
        waitContext.waitComplete = FALSE;
        if ((waitContext.evt = PCREvent_Create( FALSE )) != NULL)
        {
            /*  Create a DOFQuery  using the parameters passed in. */
            DOFQueryBuilder builder = DOFQueryBuilder_Create();

            if (builder)
            {
                if (!DOFObjectID_IsWildcard( objectID ))
                    success = DOFQueryBuilder_AddFilter( builder, 1, &objectID, nInterfaces, interfaceIDs );
                else
                    success = DOFQueryBuilder_AddFilter( builder, 0, NULL, nInterfaces, interfaceIDs );
                /* Add a restriction for the first interfaceID so the interface added callback is only called once */
                success &= DOFQueryBuilder_AddRestriction( builder, 1, interfaceIDs );
                if (success)
                {
                    DOFQuery     filter;
                    DOFOperation queryOp = NULL;

                    DOFQueryBuilder_SetMatchStyle( builder, DOFQUERYMATCHSTYLE_AT_LEAST );
                    if ((filter = DOFQueryBuilder_Build( builder )) != NULL)
                    {
                        queryOp = DOFSystem_BeginQuery( sys, filter, timeout, &WaitObjectCallback, &waitContext );
                        DOFQuery_Destroy( filter );             /* The operation has its copy of the filter now */
                    }
                    if (queryOp)
                    {
                        PCREvent_Wait( waitContext.evt, PCRTIMEOUT_FOREVER );
                        PCRMutex_Lock( waitProviderMutex );
                        /* The query complete callback can still be called after we return, so set context to NULL. */
                        queryOp->context = NULL;
                        PCRMutex_Unlock( waitProviderMutex );
                        DOFOperation_Destroy( queryOp );
                    }
                    else
                        success = FALSE;
                }
                DOFQueryBuilder_Destroy( builder );
            }
            PCREvent_Destroy( waitContext.evt );
        }
        if (!success)
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFSystem: _WaitProvider() fails allocation." );
    }

    return waitContext.waitObject;
}

DOFRouteInfo DOFSystem_CreateProviderRoute( const DOFSystem sys, const DOFObjectID objectID,
                                            const DOFInterfaceID interfaceID )
{
    PCR_Assert( sys );
    if (objectID && interfaceID && !DOFInterfaceID_IsWildcard( interfaceID ) &&
        !DOFObjectID_IsBroadcast( objectID ) && !DOFObjectID_IsMulticast( objectID ))
        return DOFRuntime_CreateProviderRoute( DOF_GetRuntime( DOFSystem_GetDOF( sys ) ), objectID, interfaceID,
                                               OALChannel_GetSecurityScope( DOFSystem_GetChannel( sys ), FALSE ) );

    return NULL;
}

boolean DOFSystem_QueryInit(void)
{
    waitProviderMutex = PCRMutex_Create();
    return (waitProviderMutex != NULL);
}

void DOFSystem_QueryCleanup(void)
{
    PCRMutex_Destroy(waitProviderMutex);
    waitProviderMutex = NULL;
}


/**
@}
**/
