/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/credentials.h>
#include <dof/filter.h>
#include <dof/objectid.h>
#include <dof/system.h>
#include <internal/filter_internal.h>
#include <internal/permissionsetinternal.h>
#include <internal/systeminternal.h>

#define SYSTEM_CONFIG_CREDENTIALS_ERROR     0x01
#define SYSTEM_CONFIG_RECEIVE_FILTER_ERROR  0x02
#define SYSTEM_CONFIG_SEND_FILTER_ERROR     0x04
#define SYSTEM_CONFIG_PERMISSIONS_ERROR     0x08
#define SYSTEM_CONFIG_IDENTITY_ERROR        0x10

typedef struct DOFSystemConfigBuilder_t
{
    DOFCredentials credentials;
    DOFObjectIDAuthentication id;
    DOFOperationFilter receiveFilter;
    DOFOperationFilter sendFilter;
    DOFPermissionSet permissions;
    uint32 refCount;
    uint32 errors;
    uint16 poolSize;
    boolean extendAllowed;
} DOFSystemConfigBuilder_t;

extern PCRMutex ref_count_mutex;

static void doErrorStuff(uint32  * errors,
                         uint32    error,
                         boolean   worked);

DOFSystemConfigBuilder DOFSystemConfigBuilder_Create(void)
{
    DOFSystemConfigBuilder retVal = (DOFSystemConfigBuilder) DOFMem_Alloc(sizeof(DOFSystemConfigBuilder_t));

    if(retVal != NULL)
    {
        retVal->receiveFilter = NULL;
        retVal->sendFilter = NULL;
        retVal->credentials = NULL;
        retVal->permissions = NULL;
        retVal->id = NULL;
        retVal->refCount = 1;
        retVal->poolSize = 0;
        retVal->errors = 0;
        retVal->extendAllowed = TRUE;
    }
    return retVal;
}

DOFSystemConfigBuilder DOFSystemConfigBuilder_Create_Config(DOFSystemConfig self)
{
    if(self != NULL)
    {
        DOFSystemConfigBuilder retVal = (DOFSystemConfigBuilder) DOFMem_Alloc(sizeof(DOFSystemConfigBuilder_t));

        if(retVal != NULL)
        {
            DOFSystemConfigBuilder selfx = (DOFSystemConfigBuilder) self;

            retVal->credentials = DOFCredentials_Copy(selfx->credentials);
            if((retVal->credentials != NULL) || (selfx->credentials == NULL))
            {
                retVal->receiveFilter = DOFOperationFilter_Copy(selfx->receiveFilter);
                if((retVal->receiveFilter != NULL) || (selfx->receiveFilter == NULL))
                {
                    retVal->sendFilter = DOFOperationFilter_Copy(selfx->sendFilter);
                    if((retVal->sendFilter != NULL) || (selfx->sendFilter == NULL))
                    {
                        retVal->permissions = DOFPermissionSet_DeepCopy(selfx->permissions);
                        if((retVal->permissions != NULL) || (selfx->permissions == NULL))
                        {
                            retVal->id = DOFObjectIDAuthentication_Copy(selfx->id);
                            if((retVal->id != NULL) || (selfx->id == NULL))
                            {
                                retVal->refCount = 1;
                                retVal->errors = 0;
                                retVal->poolSize = selfx->poolSize;
                                retVal->extendAllowed = selfx->extendAllowed;
                                return retVal;
                            }
                            DOFPermissionSet_Destroy(retVal->permissions);
                        }
                        DOFOperationFilter_Destroy(retVal->sendFilter);
                    }
                    DOFOperationFilter_Destroy(retVal->receiveFilter);
                }
                DOFCredentials_Destroy(retVal->credentials);
            }
            DOFMem_Free(retVal);
        }
    }
    return NULL;
}

DOFSystemConfigBuilder DOFSystemConfigBuilder_Create_State(DOFSystemState self)
{
    if(self != NULL)
    {
        if(self->config != NULL)
        {
            return DOFSystemConfigBuilder_Create_Config(self->config);
        }
        else
        {
            return DOFSystemConfigBuilder_Create();
        }
    }
    return NULL;
}

boolean DOFSystemConfigBuilder_SetCredentials(DOFSystemConfigBuilder self,
                                              DOFCredentials         credentials)
{
    boolean retVal = FALSE;

    if(self != NULL)
    {
        DOFCredentials newCreds = DOFCredentials_Copy(credentials);

        retVal = ((newCreds != NULL) || (credentials == NULL));
        DOFCredentials_Destroy(self->credentials);
        self->credentials = newCreds;
        doErrorStuff(&(self->errors), SYSTEM_CONFIG_CREDENTIALS_ERROR, retVal);
    }
    return retVal;
}

boolean DOFSystemConfigBuilder_SetReceiveFilter(DOFSystemConfigBuilder self,
                                                DOFOperationFilter     filter)
{
    boolean retVal = FALSE;

    if(self != NULL)
    {
        DOFOperationFilter newFilter = DOFOperationFilter_Copy(filter);

        retVal = ((newFilter != NULL) || (filter == NULL));
        DOFOperationFilter_Destroy(self->receiveFilter);
        self->receiveFilter = newFilter;
        doErrorStuff(&(self->errors), SYSTEM_CONFIG_RECEIVE_FILTER_ERROR, retVal);
    }
    return retVal;
}

boolean DOFSystemConfigBuilder_SetSendFilter(DOFSystemConfigBuilder self,
                                             DOFOperationFilter     filter)
{
    boolean retVal = FALSE;

    if(self != NULL)
    {
        DOFOperationFilter newFilter = DOFOperationFilter_Copy(filter);

        retVal = ((newFilter != NULL) || (filter == NULL));
        DOFOperationFilter_Destroy(self->sendFilter);
        self->sendFilter = newFilter;
        doErrorStuff(&(self->errors), SYSTEM_CONFIG_SEND_FILTER_ERROR, retVal);
    }
    return retVal;
}

boolean DOFSystemConfigBuilder_SetPermissions(DOFSystemConfigBuilder self,
                                              DOFPermissionSet       permissions)
{
    boolean retVal = FALSE;

    if(self != NULL)
    {
        DOFPermissionSet newPerms = DOFPermissionSet_Copy(permissions);

        retVal = ((newPerms != NULL) || (permissions == NULL));
        DOFPermissionSet_Destroy(self->permissions);
        self->permissions = newPerms;
        doErrorStuff(&(self->errors), SYSTEM_CONFIG_PERMISSIONS_ERROR, retVal);
    }
    return retVal;
}

boolean DOFSystemConfigBuilder_SetThreadpoolSize(DOFSystemConfigBuilder self,
                                                 uint16                 maxSize)
{
    if(self != NULL)
    {
        self->poolSize = maxSize;
        return TRUE;
    }
    return FALSE;
}

DOFSystemConfig DOFSystemConfigBuilder_BuildAndDestroy(DOFSystemConfigBuilder self)
{
    if(self)
    {
        if(self->errors == 0)
        {
            if(!self->extendAllowed)
            {
                DOFPermissionSet_SetGrantPerms(self->permissions);
            }
            return (DOFSystemConfig) self;
        }
        DOFSystemConfigBuilder_Destroy(self);
    }
    return NULL;
}

DOFSystemConfig DOFSystemConfigBuilder_Build(DOFSystemConfigBuilder self)
{
    if(self && (self->errors == 0))
    {
        DOFSystemConfigBuilder syscfg = DOFSystemConfigBuilder_Create_Config((DOFSystemConfig) self);
        if(syscfg)
        {
            if(!syscfg->extendAllowed)
            {
                DOFPermissionSet_SetGrantPerms(syscfg->permissions);
            }
            return (DOFSystemConfig) syscfg;
        }
    }
    return NULL;
}

DOFSystemConfigBuilder DOFSystemConfigBuilder_Copy(DOFSystemConfigBuilder self)
{
    return DOFSystemConfigBuilder_Create_Config((DOFSystemConfig) self);
}

void DOFSystemConfigBuilder_Destroy(DOFSystemConfigBuilder self)
{
    if(self != NULL)
    {
        boolean freeOk;

        PCRMutex_Lock(ref_count_mutex);
        freeOk = (--(self->refCount) == 0);
        PCRMutex_Unlock(ref_count_mutex);
        if(freeOk)
        {
            DOFCredentials_Destroy(self->credentials);
            DOFOperationFilter_Destroy(self->receiveFilter);
            DOFOperationFilter_Destroy(self->sendFilter);
            DOFPermissionSet_Destroy(self->permissions);
            DOFMem_Free(self);
        }
    }
}

DOFSystemConfig DOFSystemConfig_Copy(DOFSystemConfig self)
{
    if(self != NULL)
    {
        DOFSystemConfigBuilder selfx = (DOFSystemConfigBuilder) self;

        uint32 oldCount;

        PCRMutex_Lock(ref_count_mutex);
        oldCount = selfx->refCount;
        ++selfx->refCount;
        if(selfx->refCount > oldCount)
        {
            PCRMutex_Unlock(ref_count_mutex);
            return self;
        }
        selfx->refCount = oldCount;
        PCRMutex_Unlock(ref_count_mutex);
    }
    return NULL;
}

void DOFSystemConfig_Destroy(DOFSystemConfig self)
{
    DOFSystemConfigBuilder_Destroy((DOFSystemConfigBuilder) self);
}

DOFCredentials DOFSystemConfig_GetCredentials(DOFSystemConfig self)
{
    if(self != NULL)
    {
        return ((DOFSystemConfigBuilder) self)->credentials;
    }
    return NULL;
}

DOFOperationFilter DOFSystemConfig_GetReceiveFilter(DOFSystemConfig self)
{
    if(self != NULL)
    {
        return ((DOFSystemConfigBuilder) self)->receiveFilter;
    }
    return NULL;
}

DOFOperationFilter DOFSystemConfig_GetSendFilter(DOFSystemConfig self)
{
    if(self != NULL)
    {
        return ((DOFSystemConfigBuilder) self)->sendFilter;
    }
    return NULL;
}

DOFPermissionSet DOFSystemConfig_GetPermissions(DOFSystemConfig self)
{
    if(self != NULL)
    {
        return ((DOFSystemConfigBuilder) self)->permissions;
    }
    return NULL;
}

boolean DOFSystemConfig_IsPermissionsExtendAllowed(DOFSystemConfig self)
{
    if(self != NULL)
    {
        return ((DOFSystemConfigBuilder) self)->extendAllowed;
    }
    return FALSE;
}

uint16 DOFSystemConfig_GetThreadpoolSize(DOFSystemConfig self)
{
    if(self != NULL)
    {
        return ((DOFSystemConfigBuilder) self)->poolSize;
    }
    return 0;
}

static void doErrorStuff(uint32  * errors,
                         uint32    error,
                         boolean   worked)
{
    if(worked)
    {
        *errors &= ~error;
    }
    else
    {
        *errors |= error;
    }
}
