/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/dof.h>
#include <dof/object.h>
#include <dof/system.h>
#include <internal/exception.h>
#include <internal/requestinternal.h>
#include <internal/resultinternal.h>
#include <internal/systeminternal.h>
#include <internal/operationinternal.h>
#include "system_interest.h"


/*
**  "Private" Type and Method Declarations
*/

typedef struct
{
    DOFSystem           sys;
    DOFOperation        reqOp;
    OALSListNode        head;
} NotifyOpInfo;

typedef struct ActivateInterestListenerRemovalState_t
{
    OALEnqueueItem_t                    super;
    DOFSystemActivateInterestListener   listener;
    DOFSystem                           sys;
    DOFException                        exception;
    OALSListNode_t*                     realNode;
} ActivateInterestListenerRemovalState_t;

typedef struct ActivateInterestListenerNode_t
{
    OALSListNode_t                          sListNode;
    ActivateInterestListenerRemovalState_t  listenerInfo;
} ActivateInterestListenerNode_t;

typedef struct InterestNotifyNode_t
{
    OALEnqueueItem_t                super;
    DOFSystem                       sys;
    DOFOperation                    reqOp;
    ActivateInterestListenerNode_t* listenerNode;
} InterestNotifyNode_t;


static void notifyFn(OALThreadPoolItem dnode, boolean shutdown)
{
    InterestNotifyNode_t * node = (InterestNotifyNode_t*) dnode;
    PCR_Assert(dnode != NULL);
    node->listenerNode->listenerInfo.listener->fns->Activate( node->listenerNode->listenerInfo.listener, node->sys, node->reqOp->sources->request, node->reqOp->object_id.object_id, node->reqOp->interface_id );
    DOFSystem_Release( node->sys );
    DOFOperation_Internal_Destroy(node->reqOp);
    DOFMem_Free(node);
}


static boolean NotifyInterestCallbacks_Activate( OALAVLTreeItem_t item, void *arg )
{
    ActivateInterestListenerNode_t* listenerNode = (ActivateInterestListenerNode_t*)item;
    NotifyOpInfo* info = (NotifyOpInfo*)arg;

    PCR_Assert( info );
    if( listenerNode->listenerInfo.listener->fns->Activate )
    {
        if (DOFSystem_Retain( info->sys ) != NULL)
        {
            InterestNotifyNode_t * node = (InterestNotifyNode_t *)DOFMem_Alloc( sizeof( InterestNotifyNode_t ) );
            if (node != NULL)
            {
                OALEnqueueItem_Init( (OALEnqueueItem)node, notifyFn );
                node->sys = info->sys;
                DOFOperation_Retain( info->reqOp );
                node->reqOp = info->reqOp;
                node->listenerNode = listenerNode;
                if( OALEnqueue_Enqueue( info->sys->query_queue, (OALEnqueueItem)node ) )
                {
                    return FALSE; /* Keep Going */
                }
                DOFOperation_Internal_Destroy( info->reqOp );
                DOFMem_Free( node );
            }
            DOFSystem_Release( info->sys );
        }
        OALSList_Insert( &(info->head), listenerNode->listenerInfo.realNode );
    }
    return FALSE; /* Keep Going */
}


static void completeFn( OALThreadPoolItem dnode, boolean shutdown )
{
    InterestNotifyNode_t * node;
    PCR_Assert(dnode != NULL);
    node = (InterestNotifyNode_t*) dnode;
    node->listenerNode->listenerInfo.listener->fns->CancelActivate( node->listenerNode->listenerInfo.listener, node->sys, node->reqOp->sources->request, node->reqOp->object_id.object_id, node->reqOp->interface_id );
    DOFSystem_Release( node->sys );
    DOFOperation_Internal_Destroy(node->reqOp);
    DOFMem_Free(node);
}


static boolean NotifyInterestCallbacks_CancelActivate( OALAVLTreeItem_t item, void *arg )
{
    ActivateInterestListenerNode_t* listenerNode = (ActivateInterestListenerNode_t*)item;
    NotifyOpInfo* info = (NotifyOpInfo*)arg;

    PCR_Assert( info );
    if( listenerNode->listenerInfo.listener->fns->CancelActivate )
    {
        if( DOFSystem_Retain( info->sys ) != NULL )
        {
            InterestNotifyNode_t * node = (InterestNotifyNode_t*) DOFMem_Alloc(sizeof(InterestNotifyNode_t));
            if (node != NULL)
            {
                OALEnqueueItem_Init( (OALEnqueueItem)node, completeFn );
                node->sys = info->sys;
                DOFOperation_Retain( info->reqOp );
                node->reqOp = info->reqOp;
                node->listenerNode = listenerNode;
                if( OALEnqueue_Enqueue( info->sys->query_queue, (OALEnqueueItem)node ) )
                {
                    return FALSE; /* Keep Going */
                }
                DOFOperation_Internal_Destroy( info->reqOp );
                DOFMem_Free( node );
            }
            DOFSystem_Release( info->sys );
        }
        OALSList_Insert( &(info->head), listenerNode->listenerInfo.realNode );
    }
    return FALSE; /* Keep Going */
}

static void callRemovedCallback( OALThreadPoolItem dnode, boolean shutdown)
{
    ActivateInterestListenerRemovalState_t * node = (ActivateInterestListenerRemovalState_t*)dnode;
    
    if( node->listener->fns->Removed )
        node->listener->fns->Removed( node->listener, node->sys, node->exception );

    DOFSystem_Release( node->sys );
    DOFMem_Free( node->realNode );
}

static void NotifyInterestCallbacks_Removed( OALSListNode item, void *arg )
{
    ActivateInterestListenerNode_t* listenerNode = (ActivateInterestListenerNode_t*)item;

    if( OALAVLTree_Delete( listenerNode->listenerInfo.sys->interestListeners, listenerNode->listenerInfo.listener ) )
    {
        listenerNode->listenerInfo.exception = (DOFException)arg;
        OALEnqueueItem_Init( (OALEnqueueItem)&(listenerNode->listenerInfo), callRemovedCallback );
        OALEnqueue_Enqueue( listenerNode->listenerInfo.sys->query_queue, (OALEnqueueItem)&(listenerNode->listenerInfo) );
    }
    else
        PCR_Assert( FALSE ); /* Should be in the tree */
}

void DOFSystem_ProcessInterest( DOFSystem self, DOFOperation reqOp )
{
    NotifyOpInfo info;

    PCR_Assert(self);
    PCR_Assert(reqOp);

    info.sys = self;
    info.reqOp = reqOp;
    info.head = NULL;

    if( ( reqOp->operation_type == DOFOPERATIONTYPE_ACTIVATE ) ||
        ( reqOp->operation_type == DOFOPERATIONTYPE_CONNECT ) ||
        ( reqOp->operation_type == DOFOPERATIONTYPE_FULLCONNECT ) )
    {
        OALChannel_Lock( DOFSystem_GetChannel( self ) );
        if( self->interestListeners )
        {
            OALAVLTree_Traverse( self->interestListeners,
                                 ((reqOp->flags.bits.activation_closing == 0) ?
                                        NotifyInterestCallbacks_Activate :
                                        NotifyInterestCallbacks_CancelActivate),
                                 &info);
            if( info.head )
                OALSList_RemoveAll( &(info.head), NotifyInterestCallbacks_Removed, DOFEXCEPTION_INSUFFICIENT_RESOURCES); /* Remove all that failed */
        }
        OALChannel_Unlock( DOFSystem_GetChannel( self ) );
    }
}

static boolean ActivateInterestListenerNode_GreaterThan( OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context )
{
    DOFSystemActivateInterestListener listener = (DOFSystemActivateInterestListener)key;
    ActivateInterestListenerNode_t* node = (ActivateInterestListenerNode_t*)item;
    return ( listener > node->listenerInfo.listener );
}

static boolean ActivateInterestListenerNode_Equal( OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context )
{
    DOFSystemActivateInterestListener listener = (DOFSystemActivateInterestListener)key;
    ActivateInterestListenerNode_t* node = (ActivateInterestListenerNode_t*)item;
    return ( listener == node->listenerInfo.listener );
}

boolean DOFSystem_CreateInterestStructure( DOFSystem system )
{
    PCR_Assert( system && !system->interestListeners );
    system->interestListeners = OALAVLTree_Create( ActivateInterestListenerNode_GreaterThan, ActivateInterestListenerNode_Equal, NULL );
    if( system->interestListeners )
        return TRUE;
    return FALSE;
}

static void RemoveAllListeners( OALAVLTreeItem_t item, void *arg )
{
    ActivateInterestListenerNode_t* node = (ActivateInterestListenerNode_t*)item;
    OALEnqueueItem_Init( (OALEnqueueItem)&(node->listenerInfo), callRemovedCallback );
    OALEnqueue_Enqueue( node->listenerInfo.sys->query_queue, (OALEnqueueItem)&(node->listenerInfo) );
}

void DOFSystem_DestroyInterestStructure( DOFSystem system )
{
    if( system && system->interestListeners )
    {
        OALAVLTree_Destroy( system->interestListeners, RemoveAllListeners, NULL );
        system->interestListeners = NULL;
    }
}

/*
**  "Internal Public"  Methods.
*/

boolean DOFSystem_AddActivateInterestListener( DOFSystem self, const DOFSystemActivateInterestListener listener )
{
    PCR_Assert( self );
    if( listener && self->interestListeners )
    {
        if( (DOFSystem_Retain( self )) != NULL )
        {
            ActivateInterestListenerNode_t* listenerNode = DOFMem_Alloc( sizeof( ActivateInterestListenerNode_t ) );
            if( listenerNode )
            {
                OALAVLTree_InsertionStatus status;

                listenerNode->listenerInfo.exception = NULL;
                listenerNode->listenerInfo.listener = listener;
                listenerNode->listenerInfo.realNode = &(listenerNode->sListNode);
                listenerNode->listenerInfo.sys = self;

                OALChannel_Lock( DOFSystem_GetChannel( self ) );
                status = OALAVLTree_Insert( self->interestListeners, listener, listenerNode, NULL );
                OALChannel_Unlock( DOFSystem_GetChannel( self ) );

                switch( status )
                {
                case OALAVLTree_AlreadyInserted:
                    DOFSystem_Release( self );
                    DOFMem_Free( listenerNode );
                    /* Fall through */
                case OALAVLTree_InsertSuccess:
                    return TRUE;
                    break;            
                case OALAVLTree_InsertFailure:
                default:
                    DOFMem_Free( listenerNode );
                    break;
                }
            }
            DOFSystem_Release( self );
        }
    }
    return FALSE;
}

void DOFSystem_RemoveActivateInterestListener( DOFSystem self, const DOFSystemActivateInterestListener listener )
{
    PCR_Assert( self );
    if( listener && self->interestListeners )
    {
        ActivateInterestListenerNode_t* ret;

        OALChannel_Lock( DOFSystem_GetChannel( self ) );
        ret = (ActivateInterestListenerNode_t*)OALAVLTree_Retrieve( self->interestListeners, listener );
        if( ret )
        {
            if( OALAVLTree_Delete( self->interestListeners, listener ) )
            { 
                OALChannel_Unlock( DOFSystem_GetChannel( self ) );
                /* Enqueue the removed */
                OALEnqueueItem_Init( (OALEnqueueItem)&(ret->listenerInfo), callRemovedCallback );
                OALEnqueue_Enqueue( ret->listenerInfo.sys->query_queue, (OALEnqueueItem)&(ret->listenerInfo) );
                return;
            }
            else
                PCR_Assert(FALSE); /* We just found it! Why wasn't it removed? */
        }
        OALChannel_Unlock( DOFSystem_GetChannel( self ) );     
    }
}

boolean DOFSystem_IsInterestRequested( const DOFSystem sys, const DOFObjectID objectID, const DOFInterfaceID iID, DOFInterestLevel level )
{
    DOFSecurityScope scope;

    PCR_Assert( sys ); /* System must not be NULL */

    scope = OALChannel_GetSecurityScope( DOFSystem_GetChannel( sys ), FALSE );
    /* Outbound because the scope we care about is when the interest request comes out from dof. */

    if( (objectID == NULL) || (level > DOFINTERESTLEVEL_CONNECT_FULL) )
        return FALSE;

    return DOFRouter_CheckInterest( scope, objectID, iID, level );
}

void DOFSystem_InterestOpClosing(OALObserver self, OALObservable o, void* reqOp )
{
    DOFOperation op = (DOFOperation)reqOp;

    PCR_Assert( op );
    if(op->callbacks.icb != NULL && !op->flags.bits.completedAck)
    {
        /* Send op through queue to call its callback for final time. */
        if (!DOFSystem_AcceptResponse( op->sources->channel, op, NULL )) /* NULL DOFResult: This is cancel or timeout.*/
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "DOFSystem: Failed to notify DOFSystem that an interest op is closing." );
    }
    DOFOperation_Internal_Destroy(op);
}

