/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/exception.h>
#include <dof/interface.h>
#include <dof/object.h>
#include <internal/error.h>
#include <internal/exception.h>
#include <internal/marshal.h>
#include <internal/mutex_pool.h>
#include <internal/requestinternal.h>
#include <internal/resultinternal.h>
#include <internal/systeminternal.h>
#include <internal/value.h>
#include "system_interest.h"


/** @addtogroup DOFSystem
@{
**/


/*
**  "Private" Type and Method Declarations
*/


static boolean op_gt( OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context )
{
    return (DOFOperation)key > DOFRequest_GetReqOp((DOFRequest)item);
}
static boolean op_eq( OALAVLTreeKey_t key, OALAVLTreeItem_t item, void *context )
{
    return (DOFOperation)key == DOFRequest_GetReqOp((DOFRequest)item);
}


/** @internal
   Act on subscribe, register, and session requests only.
**/
static void CleanupAndCallProvideCompleteCallback( DOFOperation provOp, DOFOperation provSideReqOp )
{
    /* Called with both operations locked. When ProvideHandler() creates provOp->requestTree it also has both
       operations locked, likewise for Provider_ProcessRequest() when it adds entries to provOp->requestTree. */
    PCR_Assert( provOp );
    PCR_Assert( provSideReqOp );
    if( provOp->requestTree &&
       (provSideReqOp->operation_type == DOFOPERATIONTYPE_SUBSCRIBE ||
        provSideReqOp->operation_type == DOFOPERATIONTYPE_REGISTER ||
        provSideReqOp->operation_type == DOFOPERATIONTYPE_SESSION ) )
    {
        DOFRequest pCtx;

        /* We have a provider-side closing subscribe, register, or session request op. */
        PCR_Assert( !DOFInterfaceID_Compare( provSideReqOp->interface_id, provOp->interface_id ) );
        /* Get original DOFRequest from tree, call completeCallback OR NOT, then cleanup. */
        if ((pCtx = (DOFRequest)OALAVLTree_Retrieve( provOp->requestTree, (OALAVLTreeKey_t)provSideReqOp )) != NULL)
        {
            boolean success;
            PCR_Assert( pCtx );
            DOFRequest_DisablePend( pCtx );                            /* Denies further use by application. */
            success = OALAVLTree_Delete( provOp->requestTree, (OALAVLTreeKey_t)provSideReqOp );

            /* We could instead assert success TRUE, since we retrieved it and it is deletable. */
            if(!success)
                DOF_LogMessage(DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFSystem: CleanupAndCallProvideCompleteCallback() failed to delete a stored request"); 

            if (provOp->callbacks.pcb && provOp->callbacks.pcb->fns)
            {
                switch (provSideReqOp->operation_type)
                {
                case DOFOPERATIONTYPE_SUBSCRIBE:
                    PCR_Assert( provSideReqOp->item );
                    if (provOp->callbacks.pcb->fns->SubscribeComplete)
                    {
                        DOFOperation_Unlock( provSideReqOp );

                        /* All parameters are not NULL, checked by test above or assert above. */
                        provOp->callbacks.pcb->fns->SubscribeComplete( provOp->callbacks.pcb,         /* <<<< The cb! */
                                                                       provOp, pCtx,
                                                                       (DOFInterfaceProperty)provSideReqOp->item );

                        DOFOperation_Lock( provSideReqOp );
                    }
                break;
                case DOFOPERATIONTYPE_REGISTER:
                    PCR_Assert( provSideReqOp->item );
                    if (provOp->callbacks.pcb->fns->RegisterComplete)
                    {
                        DOFOperation_Unlock( provSideReqOp );

                        /* All parameters are not NULL, checked by test above or assert above. */
                        provOp->callbacks.pcb->fns->RegisterComplete( provOp->callbacks.pcb,         /* <<<<< The cb! */
                                                                      provOp, pCtx,
                                                                      (DOFInterfaceEvent)provSideReqOp->item );

                        DOFOperation_Lock( provSideReqOp );
                    }
                break;
                case DOFOPERATIONTYPE_SESSION:
                    if (provOp->callbacks.pcb->fns->SessionComplete &&
                        provSideReqOp->sessionOID && provSideReqOp->sessionIID)
                    {
                        PCR_Assert( provOp->provider_object );
                        PCR_Assert( provOp->interface_id ); /*Otherwise, DOFObject_BeginProvide will not create provOp*/
                        DOFOperation_Unlock( provSideReqOp );

                        /* All parameters are not NULL, checked by test above or assert above. */
                        provOp->callbacks.pcb->fns->SessionComplete( provOp->callbacks.pcb,         /* <<<<<< The cb! */
                                                                     provOp, pCtx, provOp->provider_object,
                                                                     provOp->interface_id, provSideReqOp->sessionOID,
                                                                     provSideReqOp->sessionIID );

                        DOFOperation_Lock( provSideReqOp );
                    }
                break;
                default: /* Assert above proves we cannot get here. */
                break;
                } /* end switch(). */
            }
            DOFRequest_CleanupRequestTree( pCtx, NULL );
        }
        else
        {
            /* There is a theoretical chance that a request is closing here on the provider side and that we get here
               before that operation was added by Provider_ProcessRequest() to ->requestTree. Therefore we can possibly
               be here for that special case and would have nothing to clean up. Do not call the complete callback
               because we know the original callback was not called (since there is no entry in ->requestTree). */
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "DOFSystem: CleanupAndCallProvideCompleteCallback() did not find the anticipated stored request" );
        }
    }
    else
    {
        /* We have a closing request that is either:
               NOT subscribe, register, or session- there is never anything to clean up on the provider side.
               IS subscribe, register, or session: lacking ->requestTree, we know we have nothing to clean up THIS TIME.
                This can happen two ways:
                    - The ordinary way. The request was cancelled and _DOFOperation_Cancel() already did this cleanup
                      and set ->requestTree to NULL, all before we were called.
                    - The unusual way. There is a theoretical chance that the closing request got us here before that
                      operation was even setup by Provide_Handler() in the first place, and therefore ->requestTree was
                      never filled. For this special case we have nothing to clean up and have no reason to call the
                      complete callback because nothing was setup and the original "opening" callback wasn't called.
           None of these situations is notable for logging. */
    }
}


/** @internal

    @param provOp
**/
static void CloseOutProvideOp( DOFOperation provOp )
{
    /* Called with provOp locked. */
    DOFException completeCallbackException = provOp->completeCallbackException;

    PCR_Assert( provOp );
    if (provOp->flags.bits.timedOut)
        completeCallbackException = DOFEXCEPTION_TIMEOUT;
    provOp->flags.bits.completed = 1;
    provOp->flags.bits.timedOut = 0;          /* So won't be hanging around if app competed callback extends timeout. */
    if (provOp->callbacks.pcb && provOp->callbacks.pcb->fns && provOp->callbacks.pcb->fns->Complete)
    {
        /* We are told of provOp timeout or cancel. So call provide callback "for the last time."
           This means call .complete() and not .get(), .set(), or .invoke(). */
        DOFOperation_Unlock( provOp );
        /* provOp ref_count at least 1 from sysQ bump. provOp destroy can have us be called through the op
           cancel observer callback. By the time that action reaches us, the app's ref_count can be gone. */

        provOp->callbacks.pcb->fns->Complete( provOp->callbacks.pcb, provOp,
                                              completeCallbackException );             /* <<<<<<<<<< APP COMPLETE CB! */

        /* .complete may have destroyed provOp, but provOp ref_count still at least 1 from sysQ bump. */
        /* Do NOT destroy completeCallbackException- they are instead destroyed at exception cleanup. */
        DOFOperation_Lock( provOp );
    }
    if (provOp->flags.bits.completed)
    {
        /* .complete() callback did not extend provOp timeout. Never call provOp's provider callback again. */
        provOp->flags.bits.completedAck = 1;
        completeCallbackException = DOFException_Copy( completeCallbackException );
        DOFException_Destroy( provOp->completeCallbackException );
        provOp->completeCallbackException = completeCallbackException;
        if (provOp->notifyOpCompleteCond)
        {
            PCRCondition_Broadcast( provOp->notifyOpCompleteCond );                   /* Signal DOFOp_WaitComplete(). */
            /* provOp at this moment has at least our enqueue refCount. */
        }
        if (!provOp->flags.bits.cancelled)
        {
            DOFOperation_Cancel(provOp);      /* Timeout case when cb did not cancel, we must do it. Sets cancel bit. */
            /* Causes router to call DOFSystem_AcceptRequest_Wrapper() again (us). This little code chunk will be
               skipped because flags.bits.completedAck is set. That is OK. The important thing was to exercise the
               operation cancel code for its other reasons. */
        }
    }
    else
    {
        /* The .complete() callback successfully extended provOp timeout. The "save" worked. The "save"
           would not be allowed if either .cancelled or .denyTimeoutExtend were TRUE at the time the complete
           callback executed. Note that thread swapping may have allowed enough time since the callback returned
           to us that either or both of those flag bits may have since set. Further, it is useless to call
           DOFOperation_GetTimeRemaining() anywhere near to right now, for the actual timeout extension
           "command" information is likely to be going through a router queue right now. */
        PCR_Assert( !provOp->flags.bits.completedAck );                /* Our thread is only setter so we can assert. */
    }
}


/** @internal

    @param provOp
    @param provSideReqOp

    @return DOFERROR_OK if unmarshal not needed or succeeded; DOFERROR_PARSE_ERROR if unmarshal failed.
**/
static DOFError ProcessOAPUnmarshalLater( DOFOperation provOp, DOFOperation provSideReqOp )
{
    /* Called with provSideReqOp and provOp both locked. */
    DOFError retStatus = DOFERROR_OK;

    DOFOperation sessionOp = DOFRouter_LookupSession(provOp->object_id.object_id);

    if(((provSideReqOp->valueStream != NULL && provSideReqOp->val.values == NULL ) ||
       (provSideReqOp->flags.bits.internally_created && (provSideReqOp->item == NULL))))
    {
        boolean itemRequired = TRUE;

        if( !provSideReqOp->item )
        {
            switch (provSideReqOp->operation_type)       /* Do work right now, if needed and if we can. */
            {
            case DOFOPERATIONTYPE_INVOKE:
                provSideReqOp->item = (DOFItem)DOFInterface_GetMethod( provOp->provider_interface, provSideReqOp->itemId );
            break;
            case DOFOPERATIONTYPE_GET:
            case DOFOPERATIONTYPE_CHANGED:
            case DOFOPERATIONTYPE_SET:
            case DOFOPERATIONTYPE_SUBSCRIBE:
                provSideReqOp->item = (DOFItem)DOFInterface_GetProperty(provOp->provider_interface, provSideReqOp->itemId);
            break;
            case DOFOPERATIONTYPE_REGISTER:
                provSideReqOp->item = (DOFItem)DOFInterface_GetEvent( provOp->provider_interface, provSideReqOp->itemId );
            break;
            default:
                itemRequired = FALSE;
            break;
            }
            DOFItem_RetainDesc( provSideReqOp->item );
        }
        if (provSideReqOp->item)
        {
            /* We can unmarshal, so do it now. */
            switch (provSideReqOp->operation_type)
            {
            case DOFOPERATIONTYPE_INVOKE:
                {
                    OALStream stream = provSideReqOp->valueStream;
                    DOFValue * values;
                    uint32 offset = OALStream_GetReadOffset(provSideReqOp->valueStream);
                    
                    if(sessionOp != NULL)
                    {
                        uint32 bufferSize = OALStream_GetDataSize(stream);
                        uint8 * buffer = (uint8*) OALStream_GetReadPtr(stream);
                        uint8 * newBuffer;

                        /* Unlock in reverse order of ProvideHandler, then relock in the original order */
                        DOFOperation_Unlock(provSideReqOp);
                        DOFOperation_Unlock(provOp);
                        newBuffer = sessionOp->sessionTransformProvider->fns->TransformReceiveData(sessionOp->sessionTransformProvider,
                                                                                                   provSideReqOp->interface_id, &bufferSize, buffer);
                        DOFOperation_Lock(provOp);
                        DOFOperation_Lock(provSideReqOp);
                        if(newBuffer != buffer)
                        {
/*                          OALStream_Release(stream, TRUE); */
                            stream = OALStream_CreateRWWithData(newBuffer, bufferSize);
                            if(stream == NULL)
                            {
                                PCRMem_Free(newBuffer);
                                return DOFERROR_INSUFFICIENT_RESOURCES;
                            }
                        }
                    }
                    values =  DOFInterfaceMethod_CreateInputParamValueListFromStream( (DOFInterfaceMethod)provSideReqOp->item, stream );

                    if ( stream == provSideReqOp->valueStream)
                        OALStream_SetReadOffset(stream, offset);  /* System and OAP both use this stream, so make sure that the read offset is correnct before and after using it here. */

                    if (!values)
                    {
                        retStatus = DOFERROR_PARSE_ERROR;
                        break;
                    }
                    /* ->val.valueStream, ->val.values, ->val.value are equivalent pointers in a union, so release the
                        stream then overwrite it with values. Clearing .unmarshalLater below tells
                        DOFOperation_Internal_Destroy() how to cleanup. */
                    if (stream != provSideReqOp->valueStream)
                        OALStream_Release(stream, TRUE);
/*                  provSideReqOp->valueStream = stream; */
                    PCR_Assert(provSideReqOp->val.values == NULL);
                    provSideReqOp->val.values = values;
                    PCR_Assert(provSideReqOp->valtype.types == NULL);
                    provSideReqOp->valtype.types = DOFInterfaceMethod_GetInputParameterTypeList
                                                                            ( (DOFInterfaceMethod)provSideReqOp->item );
                }
            break;
            case DOFOPERATIONTYPE_CHANGED:
            case DOFOPERATIONTYPE_SET:
                {
                    boolean   failed;
                    OALStream stream = provSideReqOp->valueStream;
                    DOFValue  value;
                    uint32 offset = OALStream_GetReadOffset(provSideReqOp->valueStream);
                    
                    if(sessionOp != NULL)
                    {
                        uint32 bufferSize = OALStream_GetDataSize(stream);
                        uint8 * buffer = (uint8*) OALStream_GetReadPtr(stream);
                        uint8 * newBuffer;

                        DOFOperation_Unlock(provSideReqOp);
                        DOFOperation_Unlock(provOp);
                        newBuffer = sessionOp->sessionTransformProvider->fns->TransformReceiveData(sessionOp->sessionTransformProvider,
                                                                                                   provSideReqOp->interface_id, &bufferSize, buffer);
                        DOFOperation_Lock(provOp);
                        DOFOperation_Lock(provSideReqOp);
                        if(newBuffer != buffer)
                        {
/*                          OALStream_Release(stream, TRUE); */
                            stream = OALStream_CreateRWWithData(newBuffer, bufferSize);
                            if(stream == NULL)
                            {
                                PCRMem_Free(newBuffer);
                                return DOFERROR_INSUFFICIENT_RESOURCES;
                            }
                        }
                    }
                    value = DOFInterfaceProperty_CreateValueFromStream( (DOFInterfaceProperty)provSideReqOp->item, stream, &failed );

                    if ( stream == provSideReqOp->valueStream)
                        OALStream_SetReadOffset(stream, offset);  /* System and OAP both use this stream, so make sure that the read offset is correnct before and after using it here. */

                    if (failed)
                    {
                        retStatus = DOFERROR_PARSE_ERROR;
                        if (stream != provSideReqOp->valueStream)
                            OALStream_Release(stream, TRUE);
                        break;
                    }
                    /* ->val.valueStream and ->val.values and ->val.value are equivalent pointers in a union, so release
                        the stream then overwrite it with value. Setting
                        provSideReqOp->flags.bits.value_type_is_not_list next and clearing .unmarshalLater below tells
                        DOFOperation_Internal_Destroy() how to cleanup. */
                    if (stream != provSideReqOp->valueStream)
                        OALStream_Release(stream, TRUE);
/*                  provSideReqOp->valueStream = NULL; */
                    PCR_Assert(provSideReqOp->val.value == NULL);
                    provSideReqOp->val.value = value;
                    PCR_Assert(provSideReqOp->valtype.type == NULL);
                    provSideReqOp->valtype.type = DOFInterfaceProperty_GetType
                                                                          ( (DOFInterfaceProperty)provSideReqOp->item );
                }
                /* Fall through to set the flag. */
            case DOFOPERATIONTYPE_GET:                                  /* Actual unmarshal not required. */
                /* Note for _GET only: ->val.valueStream and ->val.values and ->val.value are equivalent
                     pointers in a union. Setting provSideReqOp->flags.bits.value_type_is_not_list now and
                     clearing .unmarshalLater below tells DOFOperation_Internal_Destroy() how to cleanup. */
                provSideReqOp->flags.bits.value_type_is_not_list = 1;
            break;
            default:                                                    /* Actual unmarshal not required. */
            break;
            } /* end switch(). */
        }
        else if (itemRequired)
            retStatus = DOFERROR_PARSE_ERROR;
    }

    return retStatus;
}

/** @internal

    @param system
    @param provOp
    @param provSideReqOp
**/
static void CallProviderCallback( DOFSystem system, DOFOperation provOp, DOFOperation provSideReqOp )
{
    /* This is a normal request. So if a million other checks are OK, call the appropriate
       NON-complete provider callback. Only within this function do we call a NON-complete callback. */
    OALChannel  channel;
    DOFObject   provObj;
    DOFObjectID resultOID;
    DOFResult   result    = NULL;
    boolean     reqStored = FALSE;

    PCR_Assert( system );
    PCR_Assert( provOp );
    PCR_Assert( provSideReqOp );
    channel = DOFSystem_GetChannel( system );
    provObj = DOFOperation_GetObject( provOp );
    resultOID = DOFObject_GetObjectID( provObj );

    /* Note about provObj and resultOID: router enqueues this work that we act on here, and supplies our
       provOp and therefore our provObj and resultOID. If the original provSideReqOp was directed, there
       is only one matching provOp. If it was instead a wildcard request, router may call us multiple
       times for any given request, with a different provOp (and provObj/resultOID) each time. */
    PCR_Assert( !provSideReqOp->flags.bits.cancelled );
    PCR_Assert( !provSideReqOp->flags.bits.completedAck );
    /* Only these request types use ->requestTree. */
    if (provSideReqOp->operation_type == DOFOPERATIONTYPE_SESSION ||
        provSideReqOp->operation_type == DOFOPERATIONTYPE_REGISTER ||
        provSideReqOp->operation_type == DOFOPERATIONTYPE_SUBSCRIBE)
    {
        DOFRequest pCtx;

        if (!provOp->requestTree)
        {
            if ((provOp->requestTree = OALAVLTree_Create( op_gt, op_eq, NULL )) == NULL)
            {
                result = DOFResult_Exception_Create( channel, NULL, DOFEXCEPTION_INSUFFICIENT_RESOURCES, resultOID );
                /* let DOFSystem_DeliverResponseBackToRequestor deal with a NULL result. */
                DOFSystem_DeliverResponseBackToRequestor( provSideReqOp, result );                 /* Destroys result */
                return;                                         /* At return, provSideReqOp then provOp are unlocked. */
            }
        }
        PCR_Assert( provOp->requestTree );
        if ((pCtx = (DOFRequest)OALAVLTree_Retrieve( provOp->requestTree, (OALAVLTreeKey_t)provSideReqOp )) != NULL)
        {
            PCR_Assert( pCtx );
            reqStored = TRUE;                           /* provSideReqOp is an incoming retry. */
            /* We have a retry subscribe or retry register or retry session. For subscribe ONLY a
               new DOFRequest will be formed but NOT inserted into the tree. */
        }
        else
        { /* new DOFRequest will be formed and inserted in provOp->requestTree. */
        }
    }
    PCR_Assert( !reqStored ||
                (provSideReqOp->operation_type == DOFOPERATIONTYPE_SESSION ||
                 provSideReqOp->operation_type == DOFOPERATIONTYPE_REGISTER ||
                 provSideReqOp->operation_type == DOFOPERATIONTYPE_SUBSCRIBE) );

    /* DOFRouter filled sysQueue entry provSideReqOp with a request op that is directed to:
       a specific DOFSystem (us), a specific DOFObject, and a specific DOFInterfaceID, as found in
       provSideReqOp->providerObject->sys, provSideReqOp->providerObject, and provSideReqOp->iid.
          But they are also found in provOp->providerObject->sys, provOp->providerObject, and provOp->iid. */
    PCR_Assert( !DOFInterfaceID_Compare(provSideReqOp->interface_id, provOp->interface_id) );
    if (!provObj)
        result = DOFResult_Exception_Create(channel, NULL, DOFEXCEPTION_ILLEGAL_ARGUMENT, NULL);
    else if (provSideReqOp->operation_type != DOFOPERATIONTYPE_SESSION && !provSideReqOp->item)
        result = DOFResult_Exception_Create( channel, NULL, DOFEXCEPTION_ILLEGAL_ARGUMENT, resultOID );
    else
    {
        /* provSideReqOp->flags.bits.oap_cancel is clear for normal requests, and is set when
           the requestor sent over an OAP cancel PDU to give us this provSideReqOp. */
        /* This will call provider cb to fill result (unless an OAP cancel). Only in this path can pend happen. */

        result = Provider_ProcessRequest( provSideReqOp, provOp, reqStored ); /*<< The actual callback calls are here!*/

        /* If the above call called callback, it unlocked both op's for it, then relocked them*/
    }
    /* provSideReqOp still allocated- provider callback has no ref to it and CANNOT destroy it*/

    /* Note: Do not set provSideReqOp->provider_object or provSideReqOp->myProviderOperation, they must
       remain as delivered to us. As we send a response back through router and OAP, PDU send address is
       not determined, as possibly could be done, by examining those fields. They address the response
       PDU instead according to provSideReqOp->sources->channel, which is never changed and is the same
       whether or not the request is wildcard. */
    /* Note: provSideReqOp is locked right now. */
    if (result)
    {
        /* By being here we know that result of non-retry register with no callback can be DOFERROR_OK, or else the
           result status will NOT be DOFERROR_OK (which we cannot check) and will also contain an exception. */
        PCR_Assert( provSideReqOp->operation_type == DOFOPERATIONTYPE_REGISTER ||
                    provSideReqOp->operation_type == DOFOPERATIONTYPE_SESSION ||
                    DOFResult_GetException( result ) != NULL );
        DOFSystem_DeliverResponseBackToRequestor(provSideReqOp, result);                        /* Destroys result. */
    }
    else
    {
        /* Silently drop request here. If we pended the request, it will be acted on when the app returns or throws it.
            If we didn't pend the request, we are here because we couldn't create the DOFResult with an exception to
            tell us what the problem was. */
    }
}


/** @internal

    @note For each request processed, a provider response is formed and passed back to DOFRouter, except when:
        - the request op has been fully completed (and so it is a residual instance to be discarded), or
        - the request op is NULL.

    @param dnode
    @param shutdown
**/
static void ProvideHandler( OALThreadPoolItem dnode, boolean shutdown )
{
    /* Note: This is the enqueued function at the back end of the sys->requestQueue enqueue. */
    DOFOperation      provSideReqOp;
    DOFOperation      provOp;
    SystemContext_t * node = (SystemContext_t*) dnode;

    PCR_Assert( dnode );
    PCR_Assert( node->system );
    provSideReqOp = node->reqOp;;
    provOp = node->organizer.provOp;

    /* Process provOp with companion provSideReqOp. Note: both had their ref count bumped when placed into queue.
       This code is written so ALL paths fall through to have both ref counts unbumped. */
    PCR_Assert( provSideReqOp || provOp );
    /* Meaning of provider-side extractions from the DOFSystem queue:
        ---provSideReqOp---     ---provOp--- */
    /*     valid                   valid      Normal request: get, set, invoke, subscribe, changed, register, session.
                                                    Or: DOF_HandleInterestRequest() called DOFSystem_AcceptRequest(). */
    /*     valid                   NULL       Activate, watch, or connect request. Special, with no provOp.           */
    /*     NULL                    valid      provOp timeout notification from DOFSystem_OpTimedOutCallback(), or
                                              provOp cancel  notification from beginProvideOpClosing().               */
    /*     NULL                    NULL       Invalid.                                                                */
    PCR_Assert( provOp || provSideReqOp->operation_type == DOFOPERATIONTYPE_ACTIVATE ||
                          provSideReqOp->operation_type == DOFOPERATIONTYPE_WATCH ||
                          provSideReqOp->operation_type == DOFOPERATIONTYPE_CONNECT ||
                          provSideReqOp->operation_type == DOFOPERATIONTYPE_FULLCONNECT );
    PCR_Assert( !provOp || !provSideReqOp || (provSideReqOp->operation_type != DOFOPERATIONTYPE_ACTIVATE &&
                                              provSideReqOp->operation_type != DOFOPERATIONTYPE_WATCH &&
                                              provSideReqOp->operation_type != DOFOPERATIONTYPE_CONNECT &&
                                              provSideReqOp->operation_type != DOFOPERATIONTYPE_FULLCONNECT) );
    /* Note about provOp: router enqueues this work that we act on here, and supplies our provOp. If the original
       provSideReqOp was directed, there is only one matching provOp, equalling provSideReqOp->myProviderOperation. If
       it was instead a wildcard request, router may call us multiple times for any given request, with a different
       provOp each time. */

    if (provOp)
        DOFOperation_Lock( provOp );
    if (provSideReqOp && !shutdown)
    {
        DOFOperation_Lock( provSideReqOp );
        /* Note: Nothing in provide handling code sets provSideReqOp->completedAck, because a provider does not own this
           operation. In practical terms it means the completedAck flag NEVER gets set. However, an app that submits
           this request as a local operation (coming from same node) could (maybe) get provSideReqOp->completedAck
           set, and for that reason and that reason only do we check provSideReqOp->completedAck on this next line. */
        if (!provSideReqOp->flags.bits.completedAck && !provSideReqOp->flags.bits.cancelled)
        {
            if (provOp)
            {
                if (ProcessOAPUnmarshalLater( provOp, provSideReqOp ) != DOFERROR_OK)
                {
                    /*
                     * Error at this point means we failed to get an item (bad item ID for the interface)
                     * or we failed to get the value(s) (they do not match the interface).
                     */
                    DOFResult result = DOFResult_Exception_Create( DOFSystem_GetChannel( node->system ), NULL,
                                                                   DOFEXCEPTION_NOT_SUPPORTED,
                                                            DOFObject_GetObjectID( DOFOperation_GetObject( provOp ) ) );

                    DOFSystem_DeliverResponseBackToRequestor( provSideReqOp, result ); /* Destroys result for us. */
                    /* Fall through to unlock provSideReqOp then provOp, then final exit. */
                }
                else if (provSideReqOp->operation_type == DOFOPERATIONTYPE_FINDINTERFACEDEFINITION)
                {
                    DOFResult result = DOFResult_DefineComplete_Create( DOFSystem_GetChannel( node->system ), provOp->provider_interface,
                                                                        provOp->object_id.object_id );
                    DOFSystem_DeliverResponseBackToRequestor( provSideReqOp, result ); /* Destroys result for us. */
                }
                else if (!provOp->flags.bits.completedAck)
                {
                    /* This is a normal request. So if a million other checks are OK, call the appropriate
                       NON-complete provider callback. Only within this call do we call a NON-complete callback. */
                    CallProviderCallback( node->system, provOp, provSideReqOp );
                }
                else if (!provSideReqOp->flags.bits.isFlooded &&
                         !provSideReqOp->flags.bits.provider_oid_is_wildcard)
                {
                    /* provOp->flags.bits.completedAck set, AND this is a directed request, so it is proper that we
                       respond. But the provider doesn't exist (we are fully completed), so send that exception. */
                    DOFResult result = DOFResult_Exception_Create( DOFSystem_GetChannel( node->system ), NULL,
                                                        DOFEXCEPTION_NOT_FOUND,
                                                        DOFObject_GetObjectID( DOFOperation_GetObject( provOp ) ) );

                    DOFSystem_DeliverResponseBackToRequestor( provSideReqOp, result ); /* Destroys result for us. */
                }
            }
            else
            {
                /* NULL provOp, so this is DOFOPERATIONTYPE_ACTIVATE or _WATCH or _CONNECT or _FULLCONNECT request. */
                DOFOperation_Unlock( provSideReqOp );

                /* This calls 0-n InterestAdded callbacks. */
                DOFSystem_ProcessInterest( node->system, provSideReqOp );   /* << calls cb! for interest operations */
                /* By implementation choice, the above call does NOT call the interest callbacks directly, but
                   separately enqueues each callback to (different from our queue) node->system->query_queue. This
                   separates out interest callback from normal requests handling. */

                DOFOperation_Lock( provSideReqOp );
            }
        }
        else
        {
            /* provideSideReqOp is either cancelled or completedAck (in practical terms, nothing sets the latter). */
            if (provOp)
            {
                if (!provOp->flags.bits.completedAck)
                {
                    /* Closing Subscribe, Register, and Session requests require provider side cleanup and calling the
                       matching provider side complete callback, according to original DOFRequest information held in
                       provOp->requestTree. [Furthermore, provOp is still alive.] This next call does that work
                       for those 3 request operation types, and does nothing for all other types. */
                    CleanupAndCallProvideCompleteCallback( provOp, provSideReqOp );
                }
                else
                { /* provOp is completedAck so silently drop this request. */
                }
            }
            else
            {
                /*provOp NULL: It is DOFOP_ACTIVATE or _WATCH or _CONNECT or _FULLCONNECT request that is closing out.*/
                /* It is believable that provSideReqOp could have flags .cancelled and .complete both set, or to have
                   only the .cancelled flag set. For either situation, we choose to call the user interest callback.
                        Which callback to call, InterestAdded or InterestRemoved? DOFSystem_ProcessInterest() next will
                        do one or the other, based on provSideReqOp->flags.bits.activation_closing. */
                /* We react to provSideReqOp->flags.bits AS RECEIVED (and not as originally set when we were enqueued),
                   which presents an unusual case- our behavior here favors provSideReqOp->flags.bits.activation_closing
                   rather than .cancelled. We are willing to call InterestAdded (if .activation_closing is 0 ) even
                   though by being here we know the operation is cancelled. That unusual case can happen this way:
                        1. A new interest request enqueues us in route_interest() with .activation_closing clear, so
                            that we would call the InterestAdded callback. (The request was not cancelled at that time
                            so it was intended that we do the work of calling InterestAdded above in non-cancel code).
                        2. Before we are called at the back end of the enqueue, the request becomes marked as cancelled.
                        3. As we are finally called, the cancelled state of the request gets us HERE rather than above.
                        4. We now call the InterestAdded callback, as intended by route_interest(), even though we know
                            the request op is cancelled, and may wonder why InterestRemoved is not the better to call.
                        5. Follow on behavior is decent: We know that interest_observer() is called as part of operation
                            cancel processing. That enqueues us, this time with flag .activation_closing set, intending
                            therefore that we call the InterestRemoved callback. As we are finally called from that
                            enqueue, we get HERE, as expected, and as also expected, we do call InterestRemoved. */
                DOFOperation_Unlock( provSideReqOp );

                /* This calls 0-n InterestRemoved callbacks (or possibly InterestAdded as explained above). */
                DOFSystem_ProcessInterest( node->system, provSideReqOp );                 /* <<<<<<<<<< The CALLBACK! */
                /* By implementation choice, the above call does NOT call the interest callbacks directly, but
                   separately enqueues each callback to (different from our queue) node->system->query_queue. This
                   separates out interest callback from normal requests handling. */

                DOFOperation_Lock( provSideReqOp );
            }
        }
        /* Any paths that unlocked provideSideReqOp must get it relocked by the time we reach this point. */
        DOFOperation_Unlock( provSideReqOp );
    }
    else
    { /* There is no request to process. */
    }
    /* Whether or not a request was processed, if provOp is going away, close it out and call its complete callback. */
    if (provOp && !provOp->flags.bits.completedAck && (provOp->flags.bits.timedOut || provOp->flags.bits.cancelled ))
    {
        /*We might be here for provSideReqOp NULL (explicit notification of BeginProvide timeout), or because we just
          happened to be servicing a request at the time that timedOut or cancelled flags got themselves set. */
        CloseOutProvideOp( provOp );
    }
    else
    { /* Nothing to do. */
    }
    /* Any paths that unlocked provOp must get it relocked by the time we reach this point.*/
    if (provOp)
        DOFOperation_Unlock( provOp );
    DOFOperation_Internal_Destroy( provSideReqOp );               /* Decr ref counts, had incremented at queue input. */
    DOFOperation_Internal_Destroy( provOp );
    DOFSystem_Release( node->system );
    DOFMem_Free( dnode );
}


static DOFResult FormExceptionStreamResult(OALChannel   channel,
                                           DOFException exception,
                                           DOFObjectID  resultOid)
{
    DOFValue * values = DOFProviderException_GetParameters(exception);
    uint16 count = DOFProviderException_GetParametersCount(exception);
    DOFInterfaceException iException = DOFProviderException_GetInterfaceException(exception);

    uint8 * buffer;
    uint32 bufferSize;
    if(OALMarshal_MarshalValues(values, DOFInterfaceException_GetParameterTypeList(iException), count, &buffer, &bufferSize))
    {
        DOFOperation sessionOp = DOFRouter_LookupSession(resultOid);

        if(sessionOp != NULL)
        {
            uint8 * newBuffer = sessionOp->sessionTransformProvider->fns->TransformSendData(sessionOp->sessionTransformProvider,
                                                                                            DOFInterfaceException_GetInterfaceID(iException),
                                                                                            &bufferSize, buffer);

            if(newBuffer != buffer)
            {
                PCRMem_Free(buffer);
                buffer = newBuffer;
            }
        }
        if(buffer != NULL)
        {
            OALStream stream = OALStream_CreateRWWithData(buffer, bufferSize);

            if(stream != NULL)
            {
                DOFResult result = DOFResult_ExceptionStream_Create(channel, NULL, resultOid, stream, exception);
                OALStream_Release(stream, TRUE);
                return result;
            }
            else
            {
                PCRMem_Free(buffer);
            }
        }
    }
    return NULL;
}


/** @internal
*/
static DOFResult FormResultFromAppProvideGetProcessing( DOFRequest request, DOFValue getValue, DOFInterfaceProperty prp,
                                                        DOFObjectID resultOID, DOFSystem sys, DOFOperation operation )
{
    /* Would be static except a unit test needs this function. */
    DOFResult    retResult = NULL;
    DOFException exception = DOFRequest_GetException( request );
    boolean      exceptOK = DOFOperationControl_GetResponseLevel( DOFOperation_GetControl(operation) )
                            > DOFOPERATIONRESPONSELEVEL_SUCCESS;

    PCR_Assert( prp );
    if (!exception)
    {
        DOFType type = DOFInterfaceProperty_GetType( prp );

        if (!DOF_IsParameterValidation( sys->dof ) ||
            DOFValue_CheckCompatible( getValue, type ))
        {
            uint32 marshalSize;
            uint8 * buffer;

            if(OALMarshal_MarshalValue(getValue, type, &buffer, &marshalSize))
            {
                DOFOperation sessionOp = DOFRouter_LookupSession(resultOID);

                if(sessionOp != NULL)
                {
                    uint8 * newBuffer = sessionOp->sessionTransformProvider->fns->TransformSendData(sessionOp->sessionTransformProvider,
                                                                                                    DOFInterfaceProperty_GetInterfaceID(prp), &marshalSize, buffer);

                    if(newBuffer != buffer)
                    {
                        PCRMem_Free(buffer);
                        buffer = newBuffer;
                    }
                }
                if(buffer != NULL)
                {
                    OALStream stream = OALStream_CreateRWWithData(buffer, marshalSize);

                    if(stream != NULL)
                    {
                        retResult = DOFResult_Stream_Create(DOFSystem_GetChannel(sys), NULL, DOFERROR_OK,
                                                            resultOID, stream, DOFRESULTTYPE_VALUE_STREAM);
                        OALStream_Release(stream, TRUE);
                    }
                    else
                    {
                        PCRMem_Free(buffer);
                    }
                }
            }
            /* If can't make a real retResult, not sure we can make an exception retResult, so don't. */
        }
        else if (exceptOK)
            retResult = DOFResult_Exception_Create( DOFSystem_GetChannel( sys ), 0, DOFEXCEPTION_TYPE_MISMATCH,
                                                    resultOID );
    }
    else if (exceptOK)
    {
        if(DOFException_GetType(exception) == DOFEXCEPTIONTYPE_PROVIDER)
        {
            retResult = FormExceptionStreamResult( DOFSystem_GetChannel( sys ), exception, resultOID );
        }
        else
        {
            retResult = DOFResult_Exception_Create( DOFSystem_GetChannel( sys ), NULL, exception, resultOID );
        }
    }

    return retResult;
}


/** @internal

*/
static DOFResult FormResultFromAppProvideSetProcessing( DOFRequest request, DOFObjectID resultOID, DOFSystem sys, DOFOperation operation )
{
    /* Would be static except a unit test needs this function. */
    DOFResult    retResult = NULL;
    DOFException exception = DOFRequest_GetException( request );
    DOFInterfaceProperty prp = (DOFInterfaceProperty)(operation->item);

    if (!exception)
    {
        retResult = DOFResult_NoValue_Create( DOFSystem_GetChannel( sys ), NULL, DOFERROR_OK, resultOID );
        if ( DOFInterfaceProperty_IsReadable( prp ) )
        {
            /* This shouldn't fail, but there's nothing we can do if it does. */
            (void) DOFRouter_PropertyChanged( resultOID, prp,
                                              OALChannel_GetSecurityScope( DOFSystem_GetChannel( sys ),
                                                                           TRUE ) );        /* inbound because towards dof. */
        }
    }
    else if (DOFOperationControl_GetResponseLevel( DOFOperation_GetControl( operation ) ) >
             DOFOPERATIONRESPONSELEVEL_SUCCESS)
    {
        if(DOFException_GetType(exception) == DOFEXCEPTIONTYPE_PROVIDER)
        {
            retResult = FormExceptionStreamResult( DOFSystem_GetChannel( sys ), exception, resultOID );
        }
        else
        {
            retResult = DOFResult_Exception_Create( DOFSystem_GetChannel( sys ), NULL, exception, resultOID );
        }
    }

    return retResult;
}


/** @internal

*/
static DOFResult FormResultFromAppProvideInvokeProcessing( DOFRequest request, uint16 count,
                                                           const DOFValue invokeValues[], DOFInterfaceMethod mtd,
                                                           DOFObjectID resultOID, DOFSystem sys, DOFOperation operation )
{
    /* Would be static except a unit test needs this function. */
    DOFResult    retResult = NULL;
    DOFException exception = DOFRequest_GetException( request );
    boolean      exceptOK = DOFOperationControl_GetResponseLevel( DOFOperation_GetControl(operation) )
                            > DOFOPERATIONRESPONSELEVEL_SUCCESS;

    PCR_Assert( mtd );
    if (!exception)
    {
        if (count == DOFInterfaceMethod_GetOutputParameterCount( mtd ) && invokeValues)
        {
            DOFType * types = DOFInterfaceMethod_GetOutputParameterTypeList( mtd );

            if (!DOF_IsParameterValidation(sys->dof) ||
                DOFTypeList_IsValueCompatibleWith( count, types,
                                                   invokeValues ))
            {
                uint8 * buffer;
                uint32 marshalSize;

                if(OALMarshal_MarshalValues((DOFValue*) invokeValues, types, count, &buffer, &marshalSize))
                {
                    DOFOperation sessionOp = DOFRouter_LookupSession(resultOID);

                    if(sessionOp != NULL)
                    {
                        uint8 * newBuffer = sessionOp->sessionTransformProvider->fns->TransformSendData(sessionOp->sessionTransformProvider,
                                                                                                        DOFInterfaceMethod_GetInterfaceID(mtd), &marshalSize, buffer);

                        if(newBuffer != buffer)
                        {
                            PCRMem_Free(buffer);
                            buffer = newBuffer;
                        }
                    }
                    if(buffer != NULL)
                    {
                        OALStream stream = OALStream_CreateRWWithData(buffer, marshalSize);

                        if(stream != NULL)
                        {
                            retResult = DOFResult_Stream_Create(DOFSystem_GetChannel(sys), NULL, DOFERROR_OK,
                                                                resultOID, stream, DOFRESULTTYPE_VALUELIST_STREAM);
                            OALStream_Release(stream, TRUE);
                        }
                        else
                        {
                            PCRMem_Free(buffer);
                        }
                    }
                }
            }
            else if (exceptOK)
               retResult = DOFResult_Exception_Create( DOFSystem_GetChannel( sys ), NULL, DOFEXCEPTION_TYPE_MISMATCH,
                                                       resultOID );
        }
        else if (DOFInterfaceMethod_GetOutputParameterCount( mtd ) == 0)
        {
            OALStream stream = OALStream_CreateRO(NULL, 0);

            if(stream != NULL)
            {
                retResult = DOFResult_Stream_Create(DOFSystem_GetChannel(sys), NULL, DOFERROR_OK,
                                                    resultOID, stream, DOFRESULTTYPE_VALUELIST_STREAM);
                OALStream_Release(stream, FALSE);
            }
        }
        else if (exceptOK)
            retResult = DOFResult_Exception_Create( DOFSystem_GetChannel( sys ), NULL, DOFEXCEPTION_APPLICATION_ERROR,
                                                    resultOID );
    }
    else if (exceptOK)
    {
        if(DOFException_GetType(exception) == DOFEXCEPTIONTYPE_PROVIDER)
        {
            retResult = FormExceptionStreamResult( DOFSystem_GetChannel( sys ), exception, resultOID );
        }
        else
        {
            retResult = DOFResult_Exception_Create( DOFSystem_GetChannel( sys ), NULL, exception, resultOID );
        }
    }

    return retResult;
}


/* @internal

*/
static DOFResult FormResultFromAppProvideSessionProcessing( DOFRequest request, DOFObjectID resultOID, DOFSystem sys )
{
    DOFResult       retResult = NULL;
    DOFException    exception = DOFRequest_GetException( request );

    if (exception == NULL)
        retResult = DOFResult_NoValue_Create( DOFSystem_GetChannel( sys ), NULL, DOFERROR_OK, resultOID );
    else if (DOFOperationControl_GetResponseLevel( DOFOperation_GetControl( DOFRequest_GetReqOp( request ) ) ) >
             DOFOPERATIONRESPONSELEVEL_SUCCESS)
    {
        if(DOFException_GetType(exception) == DOFEXCEPTIONTYPE_PROVIDER)
        {
            retResult = FormExceptionStreamResult( DOFSystem_GetChannel( sys ), exception, resultOID );
        }
        else
        {
            retResult = DOFResult_Exception_Create( DOFSystem_GetChannel( sys ), NULL, exception, resultOID );
        }
    }

    return retResult;
}


/* @internal

*/
static DOFResult FormResultFromAppProvideRegisterProcessing( DOFRequest request, DOFObjectID resultOID, DOFSystem sys,
                                                             DOFInterfaceID iid )
{
    DOFResult       retResult = NULL;
    DOFException    exception = DOFRequest_GetException( request );

    if (exception == NULL)
        retResult = DOFResult_NoValue_Create( DOFSystem_GetChannel( sys ), NULL, DOFERROR_OK, resultOID );
    else if (DOFOperationControl_GetResponseLevel( DOFOperation_GetControl( DOFRequest_GetReqOp( request ) ) ) >
             DOFOPERATIONRESPONSELEVEL_SUCCESS)
    {
        if(DOFException_GetType(exception) == DOFEXCEPTIONTYPE_PROVIDER)
        {
            retResult = FormExceptionStreamResult( DOFSystem_GetChannel( sys ), exception, resultOID );
        }
        else
        {
            retResult = DOFResult_Exception_Create( DOFSystem_GetChannel( sys ), NULL, exception, resultOID );
        }
    }

    return retResult;
}


/*
**  "Internal Public"  Methods.
*/

static void DOFSystem_SetUp_AckTimeout( DOFOperation operation )
{
    if(!operation->flags.bits.internally_created)
    {
        if(operation->operationConfig)
        {
            uint32 ackTimeout = DOFOperationControl_GetAckTimeout(operation->operationConfig);

            if( ackTimeout < DOF_TIMEOUT_NEVER )
            {
                if(operation->ackTimerItem == NULL)
                {
                    operation->ackTimerItem = OALTimer_CreateItem(DOFSystem_GetTimer(), DOFSystem_Handle_AckTimeout, 1, (void **)&(operation));
                }
                if ((operation->ackTimerItem != NULL) && !OALTimerItem_IsInserted(operation->ackTimerItem))
                {
                    OALTimerItem_Insert( operation->ackTimerItem, ackTimeout, FALSE, FALSE );
                }
                else
                {
                    /* ack timeout already set, leave it alone. */
                }
            }
        }
    }
}


/* @internal
    process the operation for its type and return a DOFResult

    @param  reqOp;
    @param  provOp;

    @return DOFResult to report status and deliver DOFValue or list of DOFValues. The DOFResult will be error'ed if:
        - provOp's provObj is NULL, or
        - reqOp's item is NULL, or
        - reqOp's value is NULL for _Set request, or
        - prop subscription could not be added for _Subscribe request, or
        - evt registration could not be added for _Register request, or
        - the provider callback returned with an error code.
*/
DOFResult Provider_ProcessRequest( DOFOperation reqOp, DOFOperation provOp, boolean reqStored )
{
    /* Called with both operations locked. */
    DOFObject   provObj;                                /* Not static only because unit test needs to call it. */
    OALChannel  channel;
    DOFObjectID resultOID = NULL;

    PCR_Assert( reqOp );
    PCR_Assert( provOp );
    PCR_Assert( reqOp->operation_type == DOFOPERATIONTYPE_GET ||
                reqOp->operation_type == DOFOPERATIONTYPE_CHANGED ||
                reqOp->operation_type == DOFOPERATIONTYPE_SET ||
                reqOp->operation_type == DOFOPERATIONTYPE_INVOKE ||
                reqOp->operation_type == DOFOPERATIONTYPE_SUBSCRIBE ||
                reqOp->operation_type == DOFOPERATIONTYPE_REGISTER ||
                reqOp->operation_type == DOFOPERATIONTYPE_SESSION );
    /* Note: Both reqOp and provOp have their mutexes locked right now. We do not have an assert to prove it.
             All unlocking of these mutexes here must lock them again before returning. */
    /* Though reqOp->provider_object equals provOp->provider_object for directed provSideReqOp requests coming in, they
             are not necessarily equal when reqOp came in originally as a wildcard request. The reqOp->provider member
             holds whatever router placed there as it submitted reqOp to us. */
    PCR_Assert( provOp->interface_id );                /* Otherwise, DOFObject_BeginProvide() will not create provOp. */
    PCR_Assert( !DOFInterfaceID_Compare( reqOp->interface_id, provOp->interface_id ) );
    /*provOp->reqItem possibly NULL- provOp's factory method did not fill it because provide not restricted to an item*/
    provObj = DOFOperation_GetObject( provOp );
    PCR_Assert( provObj );
    resultOID = DOFObject_GetObjectID( provObj );
    channel = DOFSystem_GetChannel( DOFObject_GetSystem( provObj ) );
    PCR_Assert( reqOp->operation_type != DOFOPERATIONTYPE_CANCEL );       /* DOFRouter blocks this from us. */
    {
        DOFRequest submitRequest;
        DOFItem    item = reqOp->item;

        PCR_Assert( reqOp->sources->request );                        /* OAP stuffs this then router passes it along. */
        PCR_Assert( !DOFRequest_IsPending( reqOp->sources->request ) );
        PCR_Assert( !reqOp->flags.bits.oap_cancel && !reqOp->flags.bits.cancelled ); /*reqOp is locked so we know this*/

        DOFSystem_SetUp_AckTimeout( reqOp );

        switch (reqOp->operation_type)
        {
            /* For all three cases, a new DOFRequest is formed, whether this is the 1st reqOp instance or a retry. */
        case DOFOPERATIONTYPE_SUBSCRIBE:
            /* Behind the complexity here is that we call the subscribe callback once ONLY for any particular reqOp and
               all of its attendant incoming retries. To decide, any incoming reqOp checks its own ->requestTree for a
               matching provOp entry, and calls the subscribe cb only if no entry is found. */
            /*For !reqStored and we DO     have cb, the new DOFRequest IS     inserted in provOp->requestTree.
                We respond only when the app authorizes subscriptions by calling DOFRequestSubscribe_Return(). We react
                in that function call by calling the GET provide cb.

              For !reqStored and we DO NOT have cb, the new DOFRequest IS NOT inserted in provOp->requestTree.
                We respond here by calling only the GET provide cb.

              For reqStored, this subscribe is a retry AND we have a SUBSCRIBE provide cb.
                We had previously called the SUBSCRIBE cb and will not do so again. It is up to the app to use the first
                and only SUBSCRIBE cb to cause a determination be made- authorizing or denying the subscription.
                  We respond here by calling only the GET provide cb, but ONLY if the app has authorized subscriptions.

              Router watches and turns on subscription reporting only when it sees a 1st DOFERROR_OK reqOp response. */
            PCR_Assert( provOp->requestTree );
            /* FALL THROUGH: We may return present value of the subscribed property, according to above rules. */
        case DOFOPERATIONTYPE_CHANGED:
            /* FALL THROUGH: _CHANGED returns the present value of the subscribed property. */
        case DOFOPERATIONTYPE_GET:
            PCR_Assert( item );
            PCR_Assert( !reqStored || reqOp->operation_type == DOFOPERATIONTYPE_SUBSCRIBE );
            /*Check for problems- if found return exception and do not perform actions described by the above comments*/
            if (DOFInterfaceProperty_IsReadable( (DOFInterfaceProperty)item ))
            {
                if ((submitRequest = DOFRequest_Create( (DOFSystem)channel, reqOp,
                                                    DOFRequest_GetConnectionState( reqOp->sources->request ) )) != NULL)
                {
                    DOFRequest_Pend( submitRequest, reqOp, provOp );
                    /* Note: Our need to retain reqOp for all of the request types can be implemented more simply by
                       returning to an original solution where _Pend() and _Unpend() did the reqOp retain and release.*/
                    if (!reqStored && reqOp->operation_type == DOFOPERATIONTYPE_SUBSCRIBE)
                    {
                        /* Store provOp/submitRequest to provOp->requestTree. */
                        OALAVLTree_InsertionStatus status;

                        status = OALAVLTree_Insert( provOp->requestTree, (OALAVLTreeKey_t)reqOp, submitRequest, NULL );
                        PCR_Assert( status != OALAVLTree_AlreadyInserted );     /* If so reqStored would be TRUE. */
                        if (status == OALAVLTree_InsertSuccess)
                        {
                            if (provOp->callbacks.pcb && provOp->callbacks.pcb->fns &&
                                provOp->callbacks.pcb->fns->Subscribe)
                            {
                                DOFOperation_Unlock( reqOp );
                                DOFOperation_Unlock( provOp );
                                /* We pass ownership to app callback. This matches his DOFRequest_Release that takes
                                   place when the app eventualy calls DOFRequestXXX_Return() or DOFRequest_Throw()*/
                                DOFRequest_Retain( submitRequest );

                                /* The first four parameters are not NULL, checked by test above or assert above. */
                                provOp->callbacks.pcb->fns->Subscribe( provOp->callbacks.pcb, provOp, submitRequest,
                                                                       (DOFInterfaceProperty)item, /* << callback!*/
                                                                       reqOp->minPeriod );

                                DOFRequest_Release(submitRequest); /*Release hold- matches above DOFRequest_Create*/
                                DOFOperation_Lock( provOp );
                                DOFOperation_Lock( reqOp );
                                /* Note: subscribe response is filled w/ info from provider get cb, which happens in
                                   DOFSystem_CompleteSubscribeRequest(), calledbyDOFRequestSubscribe_Return(). */
                                return NULL; /* To advance pended request, app calls DOFRequestSubscribe_Return().*/
                            }
                            /* No callback- app stipulates that subscription is to be accepted. */
                            reqOp->flags.bits.subscribeOk = TRUE; /* bit write is OK- reqOp is locked this moment.*/
                            DOFRequestSubscribe_Accept( submitRequest );
                            goto standAloneGet; /* The request of the original subscribe is now in ->requestTree. */
                        }
                        DOFRequest_Unpend( submitRequest );
                        DOFRequest_Release( submitRequest );  /* Release our hold- matches above DOFRequest_Create(). */
                        return DOFResult_Exception_Create( channel, NULL, DOFEXCEPTION_INSUFFICIENT_RESOURCES,
                                                           resultOID );
                    }
                    else
                    {
                        /* This is either _SUBSCRIBE retry with subscribe callback, or any _CHANGED or any _GET. */
                        if ( reqOp->operation_type == DOFOPERATIONTYPE_SUBSCRIBE && !reqOp->flags.bits.subscribeOk )
                        {
                            /* Have to wait for application to accept it 1st! [This is subscribe retry with callback.]*/
                            DOFRequest_Unpend( submitRequest );
                            DOFRequest_Release( submitRequest ); /*Release our hold- matches above DOFRequest_Create()*/
                            return NULL;
                        }
                    }
    standAloneGet:
                    /* Get, Changed, or "Subscribe with no subscribe callback." */
                    if (provOp->callbacks.pcb && provOp->callbacks.pcb->fns && provOp->callbacks.pcb->fns->Get)
                    {
                        DOFOperation_Unlock( reqOp );
                        DOFOperation_Unlock( provOp );
                        /* We pass ownership to app callback. This matches his DOFRequest_Release that takes
                           place when the app eventualy calls DOFRequestXXX_Return() or DOFRequest_Throw()*/
                        DOFRequest_Retain( submitRequest );

                        provOp->callbacks.pcb->fns->Get( provOp->callbacks.pcb, provOp, submitRequest, /* <<< cb! */
                                                         (DOFInterfaceProperty)item );

                        DOFRequest_Release( submitRequest );  /* Release our hold- matches above DOFRequest_Create(). */
                        DOFOperation_Lock( provOp );
                        DOFOperation_Lock( reqOp );
                        return NULL;              /* To advance pended request, app calls DOFRequestGet_Return(). */
                    }
                    else
                    {
                        DOFRequest_Unpend( submitRequest );
                        DOFRequest_Release( submitRequest );  /* Release our hold- matches above DOFRequest_Create(). */
                        return DOFResult_Exception_Create(channel, NULL, DOFEXCEPTION_NOT_SUPPORTED, resultOID);
                    }
                }
                return DOFResult_Exception_Create( channel, NULL, DOFEXCEPTION_INSUFFICIENT_RESOURCES, resultOID );
            }
            return DOFResult_Exception_Create( channel, NULL, DOFEXCEPTION_NOT_READABLE, resultOID );  /* Err out.*/
        break;
        case DOFOPERATIONTYPE_SET:
            if ( provOp->callbacks.pcb && provOp->callbacks.pcb->fns && provOp->callbacks.pcb->fns->Set )
            {
                PCR_Assert( reqOp->item );
                if (DOFInterfaceProperty_IsWritable( (DOFInterfaceProperty)reqOp->item ))
                {
                    /* We can set. Call callback with a new request first time and at each retry. */
                    if ((submitRequest = DOFRequest_Create( (DOFSystem)channel, reqOp,
                                                            DOFRequest_GetConnectionState(reqOp->sources->request) ))
                                                                                                                != NULL)
                    {
                        DOFRequest_Pend( submitRequest, reqOp, provOp );
                        DOFOperation_Unlock( reqOp );
                        DOFOperation_Unlock( provOp );
                        /* We pass ownership to app callback. This matches his DOFRequest_Release that takes
                           place when the app eventualy calls DOFRequestXXX_Return() or DOFRequest_Throw()*/
                        DOFRequest_Retain( submitRequest );

                        provOp->callbacks.pcb->fns->Set( provOp->callbacks.pcb, provOp, submitRequest, /*The callback!*/
                                                         (DOFInterfaceProperty)item, reqOp->val.value );

                        DOFRequest_Release( submitRequest );  /* Release our hold- matches above DOFRequest_Create(). */
                        DOFOperation_Lock( provOp );
                        DOFOperation_Lock( reqOp );
                        return NULL;      /* To advance pended request, app calls DOFRequestSet_Return( DOFRequest ). */
                    }
                    return DOFResult_Exception_Create( channel, NULL, DOFEXCEPTION_INSUFFICIENT_RESOURCES, resultOID );
                }
                return DOFResult_Exception_Create( channel, NULL, DOFEXCEPTION_NOT_WRITABLE, resultOID );
            }
            return DOFResult_Exception_Create( channel, NULL, DOFEXCEPTION_NOT_SUPPORTED, resultOID );
        break;
        case DOFOPERATIONTYPE_INVOKE:
            if (reqOp->val.values && provOp->callbacks.pcb && provOp->callbacks.pcb->fns &&
                provOp->callbacks.pcb->fns->Invoke)
            {
                PCR_Assert( item );
                if ((submitRequest = DOFRequest_Create( (DOFSystem)channel, reqOp,
                                                      DOFRequest_GetConnectionState(reqOp->sources->request) )) != NULL)
                {
                    DOFRequest_Pend( submitRequest, reqOp, provOp );
                    /* We can invoke. Call callback with a new request first time and at each retry. */
                    DOFOperation_Unlock( reqOp );
                    DOFOperation_Unlock( provOp );
                    /* We pass ownership to app callback. This matches his DOFRequest_Release that takes
                       place when the app eventualy calls DOFRequestXXX_Return() or DOFRequest_Throw()*/
                    DOFRequest_Retain( submitRequest );

                    provOp->callbacks.pcb->fns->Invoke( provOp->callbacks.pcb, provOp, submitRequest, /* The callback!*/
                                                        (DOFInterfaceMethod)item,
                                                  DOFInterfaceMethod_GetInputParameterCount( (DOFInterfaceMethod)item ),
                                                         reqOp->val.values );

                    DOFRequest_Release( submitRequest );      /* Release our hold- matches above DOFRequest_Create(). */
                    DOFOperation_Lock( provOp );
                    DOFOperation_Lock( reqOp );
                    return NULL;         /* To advance pended request, app calls DOFRequestInvoke_Return(DOFRequest). */
                }
                return DOFResult_Exception_Create( channel, NULL, DOFEXCEPTION_INSUFFICIENT_RESOURCES, resultOID );
            }
            return DOFResult_Exception_Create( channel, NULL, DOFEXCEPTION_NOT_SUPPORTED, resultOID );
        break;
        case DOFOPERATIONTYPE_REGISTER:
            PCR_Assert( provOp->requestTree );
            PCR_Assert( item );
            if (!reqStored)
            {
                /* provSideReqOp NOT retry and provider wants his register callback called, to perhaps throw exception*/
                if ((submitRequest = DOFRequest_Create( (DOFSystem)channel, reqOp,
                                                      DOFRequest_GetConnectionState(reqOp->sources->request) )) != NULL)
                {
                    /* Store provOp/submitRequest to provOp->requestTree. */
                    OALAVLTree_InsertionStatus status;

                    DOFRequest_Pend( submitRequest, reqOp, provOp );
                    status = OALAVLTree_Insert( provOp->requestTree, (OALAVLTreeKey_t)reqOp, submitRequest, NULL );
                    PCR_Assert( status != OALAVLTree_AlreadyInserted );         /* If so reqStored would be TRUE. */
                    if (status == OALAVLTree_InsertSuccess)
                    {
                        /* Eventual tree cleanup un-does these two holds. That can happen in CleanupRequestTree()
                           when either _DOFOperation_Cancel() or DOFOperation_Internal_Destroy() calls it, or in
                           CleanupAndCallProvideCompleteCallback(), which is called when the request op is cancelled
                           and the app gave it a closing observer, causing ProvideHandler() to be called. */
                        if (provOp->callbacks.pcb && provOp->callbacks.pcb->fns &&
                            provOp->callbacks.pcb->fns->Register)
                        {
                            DOFOperation_Unlock( reqOp );
                            DOFOperation_Unlock( provOp );
                            /* We pass ownership to app callback. This matches his DOFRequest_Release that takes
                               place when the app eventualy calls DOFRequestXXX_Return() or DOFRequest_Throw(). */
                            DOFRequest_Retain( submitRequest );

                            /* All parameters are not NULL, checked by test above or assert above. */
                            provOp->callbacks.pcb->fns->Register( provOp->callbacks.pcb, provOp, /* <The callback!*/
                                                                  submitRequest, (DOFInterfaceEvent)item );

                            DOFRequest_Release(submitRequest); /*Release our hold- matches above DOFRequest_Create*/
                            DOFOperation_Lock( provOp );
                            DOFOperation_Lock( reqOp );
                            return NULL;      /* DOFResult is returned when pended request advances as the app calls
                                                 DOFRequestRegister_Return(DOFRequest,...) or DOFRequest_Throw(). */
                        }
                        reqOp->flags.bits.subscribeOk = TRUE;    /* bit write is OK- reqOp is locked this moment. */
                        DOFRequest_Unpend(submitRequest);
                        goto acceptRegistration;                 /* No Callbacks so accept */
                    }
                    DOFRequest_Unpend(submitRequest);
                    DOFRequest_Release( submitRequest );      /* Release our hold- matches above DOFRequest_Create(). */
                }
                return DOFResult_Exception_Create( channel, NULL, DOFEXCEPTION_INSUFFICIENT_RESOURCES, resultOID );
            }
            else
            { /* retry */
            }
    acceptRegistration:
            /* provSideReqOp is either NOT a retry and we have no callback,
                        (in which case the application stipulates that the registration is to be accepted)
               OR it is a retry
                        (in which case the application stipulated that the registration is accepted only after it calls
                        DOFRequestRegister_Return(). */
            if (reqOp->flags.bits.subscribeOk)
                return DOFResult_NoValue_Create( channel, NULL, DOFERROR_OK, resultOID );                /* accepted. */
            else
                return NULL;                                                        /* Registration not yet accepted. */
        break;
        case DOFOPERATIONTYPE_SESSION:
            {
                /* Call provider supplied Session callback. */
                PCR_Assert( provOp->requestTree );
                if (!reqStored && reqOp->sessionOID && reqOp->sessionIID && reqOp->sources->channel &&
                    provOp->callbacks.pcb && provOp->callbacks.pcb->fns && provOp->callbacks.pcb->fns->Session)
                {
                    if ((submitRequest = DOFRequest_Create( (DOFSystem)channel, reqOp,
                                                      DOFRequest_GetConnectionState(reqOp->sources->request) )) != NULL)
                    {
                        OALAVLTree_InsertionStatus status;

                        DOFRequest_Pend( submitRequest, reqOp, provOp );
                        /* Store provOp/submitRequest to provOp->requestTree. */
                        status = OALAVLTree_Insert( provOp->requestTree, (OALAVLTreeKey_t)reqOp, submitRequest, NULL );
                        PCR_Assert( status != OALAVLTree_AlreadyInserted );/*If so reqStored would be TRUE*/
                        if (status == OALAVLTree_InsertSuccess)
                        {
                            DOFOperation_Unlock( reqOp );                  /* Call the session callback. */
                            /* We pass ownership to app callback. This matches his DOFRequest_Release that takes
                               place when the app eventualy calls DOFRequestXXX_Return() or DOFRequest_Throw()*/
                            DOFOperation_Unlock( provOp );
                            DOFRequest_Retain( submitRequest );

                            /* All parameters are not NULL, checked by test above or assert above. */
                            provOp->callbacks.pcb->fns->Session( provOp->callbacks.pcb, provOp, /*<<<The callback!*/
                                                                 submitRequest, provObj, provOp->interface_id,
                                                                 reqOp->sessionOID, reqOp->sessionIID );

                            DOFRequest_Release(submitRequest); /*Release our hold- matches above DOFRequest_Create*/
                            DOFOperation_Lock( provOp );
                            DOFOperation_Lock( reqOp );
                            return NULL;
                            /* Note: OAP will not send a succesful open-session response PDU if
                               DOFRequestSession_Return() is not called. Nothing in the COAL calls that. If
                               the just called open session callback did not call it, and nothing else in
                               the user app calls it, then the succesful OAP open-session response PDU
                               will not go back to requestor. */
                        }
                        DOFRequest_Unpend(submitRequest);
                        DOFRequest_Release( submitRequest );  /* Release our hold- matches above DOFRequest_Create(). */
                    }
                    return DOFResult_Exception_Create( channel, NULL, DOFEXCEPTION_INSUFFICIENT_RESOURCES, resultOID );
                }
                else if (!reqStored)
                    return DOFResult_Exception_Create( channel, NULL, DOFEXCEPTION_NOT_SUPPORTED, resultOID );
                else
                {
                    /*provSideReqOp is retry of session we implemented already, so return success message to requestor*/
                    return DOFResult_NoValue_Create( channel, NULL, DOFERROR_OK, resultOID );
                }
            }
        break;
        default:
            {
                PCR_Assert( FALSE );
            }
        break;
        } /* end switch( op type). */
    }
    return NULL;
}


void DOFSystem_AcceptRequest_Wrapper(OALChannel chan, DOFOperation reqOp, const DOFOperation provOp, DOFAddress unused)
{
    DOFSystem_AcceptRequest( chan, reqOp, provOp ); /*router calls _Wrapper thru a table and cannot use boolean return*/
}


/** @internal
   Accept DOF request from DOFRouter to provider and queue it for DOF processing that generates a response. Or,
   for interest operations coming from DOFRouter to requestor, queue it to process that interest.

   @note This call does not and shall not block.
   @note Callers must not call with the reqOp mutex locked.

   @param chan   DOFSystem of the request. Though findable in either op param, C OAL form says supply DOFSystem.
              OALChannel param used instead of DOFSystem to generically match the other calls used by OALChannel_Ctor().
   @param reqOp  The new request operation info. It is used for actual provider response processing.
   @param provOp If reqOp coming to provider, is active BeginProvide operation that defines our responses to requests.

   @return TRUE if the OAP request is submitted for further processing, or FALSE if it is discarded.
**/
boolean DOFSystem_AcceptRequest( OALChannel chan, DOFOperation reqOp, const DOFOperation provOp )
{
    boolean   retBoolean;
    boolean   cleanupThree = TRUE;
    DOFSystem sys          = (DOFSystem)chan;

    PCR_Assert( sys );
    PCR_Assert( reqOp || provOp );

    PCRMutex_Lock( ref_count_mutex );
    /* Detect a rare situation, where sys is in the process of being cleaned up just as an interest operation comes to
     the requestor side. This can happen because interest operations do not hold a refCount on the relevant DOFSystem.

     How this can happen (showing why we are sure sys is not freed at this moment EVEN IF sys->refCount is 0):

     1. OALChannel_Release() takes sys->refCount to 0 within held ref_count_mutex.

     2. DOF_HandleInterestRequest() is called within interest.c by either interest_observer() or route_interest(). Both
     calls have a valid operation but neither has knowledge that the relevant DOFSystem is valid, invalid, or perhaps
     just now in the process of being cleaned up.

     3. DOF_HandleInterestRequest finds that sys exists and calls DOFSystem_AcceptRequest, although sys->refCount is 0.
     *********
     Note: DOF_HandleInterestRequest() finds that sys exists by checking dof->systems slist. Fact: dof->systems holds a
     DOFSystem even after its refCount goes to 0 UNTIL DOFSystem_Dtor() calls DOF_RemoveSystem(). For the zero refCount
     sys of this example, that has NOT happened yet and we are assured that sys is not freed within this function.
     *********

     So detect this rare case. If     so: exit, because everything else says sys is GONE, and we know sys is a zombie.
                               If not so: bump sys refCount within the held mutex to prevent a (possibly) imminent set
                                          of DOFSystem_Release() calls from taking sys's refCount to 0. */
    retBoolean = (sys->channel.refCount > 0);               /* Even for refCount 0, above shows sys is known not freed*/
    if (retBoolean)
        retBoolean = (chan = OALChannel_Retain(chan)) != NULL; /*Returns quick because we already hold ref_count_mutex*/
    PCRMutex_Unlock( ref_count_mutex );
    if (!retBoolean)
        return FALSE;

    /* If here, we bumped sys refCount and sys cannot be freed until we release it again. */
    retBoolean = FALSE;
    if((reqOp != NULL) && (reqOp->operation_type != DOFOPERATIONTYPE_CHANGED))
    {
        /*DOF gave us reqOp, hence "outbound." Test its perms against our security connector. Drop reqOp if called for*/
        validationStatus valid = DOFSystem_ValidateOperationInboundPermissions( sys, sys->channel.securityMembrane,
                                                                                (reqOp->responsePerm) ? reqOp->responsePerm : reqOp->commandPerm,
                                                                                reqOp->commandPerm, reqOp, provOp,
                                                                                NULL, TRUE );
        if (valid == VALIDATED)
        { /* OK that we accept the request. Fall through for enqueuing reqOp with provOp to RequestHandler(). */
        }
        else
        {
            /* Not validated at the moment. Validation either in progress or known invalid, or error. We now exit. */
            switch (valid)
            {
            default:
                {
                    PCR_Assert( FALSE );
                }
                /* Fall through. */
            case VALIDATION_ERROR:
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DOFSystem: Error while validating request" );
                /* Fall through. */
            case INVALIDATED:
                /* Break out and drop reqOp. */
            break;
            case VALIDATION_PENDING:
                /* Validation determined by TRP callback, setup in above _ValidateOperationOutboundPermissions call. */
                retBoolean = TRUE;                        /* Return TRUE: reqOp is submitted for further processing. */
            break;
            } /* end switch(). */
            /* Exit but first release our refCount bump from above. For VALIDATION_PENDING, the _Validate operation took
               ownership from us. For other cases we won't be called again so our hold on sys must be removed. */
            OALChannel_Release( chan );
            return retBoolean;
        }
    }
    else
    { /* No reqOp- this is notification of provOp timeout or cancel. */
    }

    /* Meaning of provider-side submittals:
        ---reqOp---             ---provOp--- */
    /*     valid                   valid      Normal request: get, set, invoke, subscribe, changed, register, session.
                                                    Or: DOF_HandleInterestRequest() called DOFSystem_AcceptRequest(). */
    /*     valid                   NULL       Activate, watch, or connect request. Special, with no provOp.           */
    /*     NULL                    valid      provOp timeout notification from DOFSystem_OpTimedOutCallback(), or
                                              provOp cancel  notification from beginProvideOpClosing().               */
    /*     NULL                    NULL       Invalid.                                                                */
    PCR_Assert( provOp || reqOp->operation_type == DOFOPERATIONTYPE_ACTIVATE ||
                          reqOp->operation_type == DOFOPERATIONTYPE_WATCH ||
                          reqOp->operation_type == DOFOPERATIONTYPE_CONNECT ||
                          reqOp->operation_type == DOFOPERATIONTYPE_FULLCONNECT );

    /* If here, we bumped sys refCount and sys cannot be freed until we release it again. */
    if (provOp)
        DOFOperation_Retain( provOp );
    if (reqOp)
        DOFOperation_Retain( reqOp );
    PCR_Assert( !retBoolean );                               /* and cleanupThree is TRUE. */
    {
        PCRMutex mutex = OALMutexPool_Get( sys );

        PCRMutex_Lock( mutex );
        if(sys->requestQueue != NULL)
        {
            SystemContext_t * node = (SystemContext_t *)DOFMem_Alloc(sizeof(SystemContext_t));

            if (node != NULL)
            {
                OALEnqueueItem_Init( (OALEnqueueItem) node, ProvideHandler );
                node->system = sys;
                node->reqOp = reqOp;
                node->organizer.provOp = provOp;
                if ((retBoolean = OALEnqueue_Enqueue( sys->requestQueue, (OALEnqueueItem)node )) == FALSE)
                    DOFMem_Free( node );  /* Fail out. */
                else
                    cleanupThree = FALSE; /*Bumped 3 refCounts above. Cannot be freed until ProvideHandler() releases them*/
            }
        }
        PCRMutex_Unlock(mutex);
        OALMutexPool_Release(sys);
    }
    if (cleanupThree)
    {
        DOFOperation_Internal_Destroy( reqOp );
        DOFOperation_Internal_Destroy( provOp );
        DOFSystem_Release( sys );
        PCRThread_Yield(); /* Ensure the processing thread has a chance to service the queue. */
    }

    return retBoolean;
}

typedef struct MarshallPendingContext_t
{
    OALEnqueueItem_t super;
    DOFSystem system;
    DOFRequest request;
    DOFOperation operation;
    DOFObjectID providerId;
    union {
        DOFValue value;
        struct {
            DOFValue * values;
            uint16 count;
        } values;
    } value;
} * MarshallPendingContext;

static MarshallPendingContext CreateMarshallPendingContext_Base(DOFSystem  system,
                                                                DOFRequest request)
{
    MarshallPendingContext ret = (MarshallPendingContext) DOFMem_Alloc(sizeof(struct MarshallPendingContext_t));

    if(ret != NULL)
    {
        ret->system = DOFSystem_Retain(system);
        if(ret->system != NULL)
        {
            ret->providerId = DOFObjectID_Copy(DOFRequest_GetResultOID(request));
            if(ret->providerId != NULL)
            {
                DOFOperation operation = DOFRequest_GetReqOp(request);

                ret->request = (DOFRequest_Retain(request), request);
                ret->operation = (DOFOperation_Retain(operation), operation);
                return ret;
            }
            DOFSystem_Release(ret->system);
        }
        DOFMem_Free(ret);
    }
    return NULL;
}

static void DestroyMarshallPendingContext_Base(MarshallPendingContext self)
{
    DOFObjectID_Destroy(self->providerId);
    DOFOperation_Internal_Destroy(self->operation);
    DOFRequest_Release(self->request);
    DOFSystem_Release(self->system);
    DOFMem_Free(self);
}

static MarshallPendingContext CreateMarshallPendingContext_Value(DOFSystem   system,
                                                                 DOFRequest  request,
                                                                 DOFValue    value)
{
    MarshallPendingContext ret = CreateMarshallPendingContext_Base(system, request);

    if(ret != NULL)
    {
        ret->value.value = DOFValue_Copy(value);
        if((ret->value.value != NULL) || (value == NULL))
        {
                return ret;
        }
        DestroyMarshallPendingContext_Base(ret);
    }
    return NULL;
}

static MarshallPendingContext CreateMarshallPendingContext_Values(DOFSystem    system,
                                                                  DOFRequest   request,
                                                                  uint16       count,
                                                                  DOFValue   * values)
{
    MarshallPendingContext ret = CreateMarshallPendingContext_Base(system, request);

    if(ret != NULL)
    {
        ret->value.values.count = count;
        ret->value.values.values = DOFMem_Alloc(sizeof(DOFValue) * count);
        if(ret->value.values.values != NULL)
        {
            uint16 i;
            uint16 j;
            boolean failed = FALSE;

            for(i = 0; i < count; ++i)
            {
                ret->value.values.values[i] = DOFValue_Copy(values[i]);
                if((ret->value.values.values[i] == NULL) && (values[i] != NULL))
                {
                    failed = TRUE;
                    break;
                }
            }
            if(failed)
            {
                for(j = 0; j < i; ++j)
                {
                    DOFValue_Destroy(ret->value.values.values[j]);
                }
                DOFMem_Free(ret->value.values.values);
            }
            else
            {
                return ret;
            }
        }
        DestroyMarshallPendingContext_Base(ret);
    }
    return NULL;
}

static void DestroyMarshallPendingContext_Value(MarshallPendingContext self)
{
    DOFValue_Destroy(self->value.value);
    DestroyMarshallPendingContext_Base(self);
}

static void DestroyMarshallPendingContext_Values(MarshallPendingContext self)
{
    uint16 i;

    for(i = 0; i < self->value.values.count; ++i)
    {
        DOFValue_Destroy(self->value.values.values[i]);
    }
    DOFMem_Free(self->value.values.values);
    DestroyMarshallPendingContext_Base(self);
}

static void DoPendingGetMarshal(OALThreadPoolItem self,
                                boolean           shutdown)
{
    MarshallPendingContext realSelf = (MarshallPendingContext) self;

    if(!shutdown)
    {
        DOFResult result;

        /* Note: value can be NULL for a NULLABLE value type; see DOFValue_IsCompatibleWith(). */
        result = FormResultFromAppProvideGetProcessing( realSelf->request, realSelf->value.value, (DOFInterfaceProperty)realSelf->operation->item,
                                                        realSelf->providerId, realSelf->system, realSelf->operation );
        DOFSystem_CheckPermissionsAndRouteResponse( realSelf->system, realSelf->operation, result );
        /* The above call handles the possible but unexpected NULL result and otherwise destroys result for us. */
    }
    DestroyMarshallPendingContext_Value(realSelf);
}

/** @internal
** Called by DOFRequest_CompleteGet when a previously pended Get request completes.
*/
void DOFSystem_CompleteGetRequest( DOFSystem sys, DOFRequest request, DOFValue value )
{
    MarshallPendingContext node;

    PCR_Assert( sys );
    PCR_Assert( request );
    node = CreateMarshallPendingContext_Value(sys, request, value);
    if(node != NULL)
    {
        OALEnqueueItem_Init((OALEnqueueItem) node, DoPendingGetMarshal);
        if(!OALEnqueue_Enqueue(sys->responseQueue, (OALEnqueueItem) node))
        {
            DestroyMarshallPendingContext_Value(node);
        }
    }
}

static void DoPendingSetMarshal(OALThreadPoolItem self,
                                boolean           shutdown)
{
    MarshallPendingContext realSelf = (MarshallPendingContext) self;

    if(!shutdown)
    {
        DOFResult result;

        /* Note: value can be NULL for a NULLABLE value type; see DOFValue_IsCompatibleWith(). */
        result = FormResultFromAppProvideSetProcessing( realSelf->request, realSelf->providerId, realSelf->system, realSelf->operation );
        DOFSystem_CheckPermissionsAndRouteResponse( realSelf->system, realSelf->operation, result );
        /* The above call handles the possible but unexpected NULL result and otherwise destroys result for us. */
    }
    DestroyMarshallPendingContext_Base(realSelf);
}

/* @internal
** Called by DOFRequest_CompleteSet when a previously pended Set request completes.
*/
void DOFSystem_CompleteSetRequest( DOFSystem sys, DOFRequest request )
{
    MarshallPendingContext node;

    PCR_Assert( sys );
    PCR_Assert( request );
    node = CreateMarshallPendingContext_Base(sys, request);
    if(node != NULL)
    {
        OALEnqueueItem_Init((OALEnqueueItem) node, DoPendingSetMarshal);
        if(!OALEnqueue_Enqueue(sys->responseQueue, (OALEnqueueItem) node))
        {
            DestroyMarshallPendingContext_Base(node);
        }
    }
}

static void DoPendingInvokeMarshal(OALThreadPoolItem self,
                                   boolean           shutdown)
{
    MarshallPendingContext realSelf = (MarshallPendingContext) self;

    if(!shutdown)
    {
        DOFResult result;

        /* Note: value can be NULL for a NULLABLE value type; see DOFValue_IsCompatibleWith(). */
        result = FormResultFromAppProvideInvokeProcessing( realSelf->request, realSelf->value.values.count, realSelf->value.values.values,
                                                           (DOFInterfaceMethod)realSelf->operation->item, realSelf->providerId, realSelf->system, realSelf->operation );
        DOFSystem_CheckPermissionsAndRouteResponse( realSelf->system, realSelf->operation, result );
        /* The above call handles the possible but unexpected NULL result and otherwise destroys result for us. */
    }
    DestroyMarshallPendingContext_Values(realSelf);
}

/* @internal
** Called by DOFRequest_CompleteInvoke when a previously pended Invoke request completes.
*/
void DOFSystem_CompleteInvokeRequest( DOFSystem sys, DOFRequest request, uint16 valuesCount, const DOFValue values[] )
{
    MarshallPendingContext node;

    PCR_Assert( sys );
    PCR_Assert( request );
    node = CreateMarshallPendingContext_Values(sys, request, valuesCount, (DOFValue*) values);
    if(node != NULL)
    {
        OALEnqueueItem_Init((OALEnqueueItem) node, DoPendingInvokeMarshal);
        if(!OALEnqueue_Enqueue(sys->responseQueue, (OALEnqueueItem) node))
        {
            DestroyMarshallPendingContext_Values(node);
        }
    }
}


/* @internal
** Called by DOFRequest_CompleteSession when a previously pended Session request completes.
*/
void DOFSystem_CompleteSessionRequest( DOFSystem sys, DOFRequest request )
{
    DOFResult    result;
    DOFOperation provSideReqOp;

    PCR_Assert( sys );
    PCR_Assert( request );
    provSideReqOp = DOFRequest_GetReqOp( request );
    PCR_Assert( provSideReqOp );
    result = FormResultFromAppProvideSessionProcessing( request, DOFRequest_GetResultOID( request ), sys );
    DOFSystem_CheckPermissionsAndRouteResponse( sys, provSideReqOp, result );
    /* The above call handles the possible but unexpected NULL result and otherwise destroys result for us. */
}


/* @internal
** Called by DOFRequestSubscribe_Return() when a previously pended Subscribe request completes.
*/
void DOFSystem_CompleteSubscribeRequest( DOFSystem sys, DOFRequest request )
{
    /* Subscribe OK, call the Get callback to get the property value. */
    DOFOperation provOp, reqOp;

    PCR_Assert( sys );
    PCR_Assert( request );
    provOp = DOFRequest_GetProviderOp( request );
    PCR_Assert( provOp );
    reqOp = DOFRequest_GetReqOp( request );
    PCR_Assert( reqOp );
    if (provOp->callbacks.pcb && provOp->callbacks.pcb->fns && provOp->callbacks.pcb->fns->Get)
    {
        /* We pass ownership to app callback. This matches his DOFRequest_Release that takes
           place when the app eventualy calls DOFRequestXXX_Return() or DOFRequest_Throw(). */
        DOFRequest_Retain( request );

        provOp->callbacks.pcb->fns->Get( provOp->callbacks.pcb, provOp, request,
                                         (DOFInterfaceProperty)reqOp->item );                        /* <<< Callback! */

        DOFRequest_Release( request );                 /* Release our hold- matches our original DOFRequest_Create(). */
    }
    else if (DOFOperationControl_GetResponseLevel( DOFOperation_GetControl( DOFRequest_GetReqOp( request ) ) ) >
             DOFOPERATIONRESPONSELEVEL_SUCCESS)
    {
        DOFResult result = DOFResult_Exception_Create( DOFSystem_GetChannel( sys ), NULL, DOFEXCEPTION_NOT_SUPPORTED,
                                                       DOFObject_GetObjectID( DOFOperation_GetObject( provOp ) ) );

        DOFSystem_CheckPermissionsAndRouteResponse( sys, reqOp, result );                                  /* Err out. */
        /* The above call handles the possible but unexpected NULL result and otherwise destroys result for us. */
    }
    DOFOperation_Lock(reqOp);
    reqOp->flags.bits.subscribeOk = TRUE;
    DOFOperation_Unlock(reqOp);
}


/* @internal
** Called by DOFRequest_CompleteRegister when a previously pended Register request completes.
*/
void DOFSystem_CompleteRegisterRequest( DOFSystem sys, DOFRequest request )
{
    DOFResult    result;
    DOFOperation provSideReqOp;

    PCR_Assert( sys );
    PCR_Assert( request );
    provSideReqOp = DOFRequest_GetReqOp( request );
    PCR_Assert( provSideReqOp );
    result = FormResultFromAppProvideRegisterProcessing( request, DOFRequest_GetResultOID( request ), sys,
                                           DOFInterfaceEvent_GetInterfaceID( (DOFInterfaceEvent)provSideReqOp->item ) );
    DOFSystem_CheckPermissionsAndRouteResponse( sys, provSideReqOp, result );
    /* The above call handles the possible but unexpected NULL result and otherwise destroys result for us. */
    DOFOperation_Lock(provSideReqOp);
    provSideReqOp->flags.bits.subscribeOk = TRUE;
    DOFOperation_Unlock(provSideReqOp);
}

int DOFSystem_Handle_AckTimeout( OALTimerItem item, uint32 argc, void* argv[], boolean shutdown )
{
    DOFOperation op;
    PCR_Assert( argc == 1 );
    op = (DOFOperation) argv[0];
    DOFOperation_Lock(op);
    if ( !op->completeCallbackException )
        op->completeCallbackException = DOFException_Copy( DOFEXCEPTION_ACK_TIMEOUT );
    DOFOperation_Unlock(op);
    DOFOperation_Cancel( op );
    return 0;
}

/**
@}
**/
