/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/bridge.h>
#include <dof/core.h>
#include <dof/exception.h>
#include <dof/interface.h>
#include <dof/object.h>
#include <dof/permissionset.h>
#include <dof/providerinfo.h>
#include <internal/securityscope.h>
#include <internal/bridgeinternal.h>
#include <internal/exception.h>
#include <internal/interfaceidinternal.h>
#include <internal/objectinternal.h>
#include <internal/operationinternal.h>
#include <internal/permissioninternal.h>
#include <internal/resultinternal.h>
#include <internal/systeminternal.h>
#include "system_interest.h"


/** @addtogroup DOFSystem
@{
**/


/*
**  "Private" Type and Method Declarations
*/


typedef struct ExtendPermContext_t
{
    DOFOperation op;
    DOFSecurityMembrane secConn;
} * ExtendPermContext;


static void interestExtendCB( DOFError status, void * context )
{
    ExtendPermContext realContext = (ExtendPermContext)context;

    if (status == DOFERROR_OK)
    {
        PCR_Assert(realContext->secConn);

        DOFSecurityScope_Destroy(realContext->op->securityScope);
        realContext->op->securityScope =
            DOFSecurityMembrane_ValidateOutboundPermissions(realContext->secConn, realContext->op->commandPerm, realContext->op->commandPerm);
        if (realContext->op->securityScope != NULL)
        {
            if (OALPermissionScope_GetIdCount(DOFSecurityScope_GetOALPermissionScope(realContext->op->securityScope)) > 0)
            {
                DOFOperation temp = realContext->op;

                DOFRouterStatusCode_t routeResponseStatus;

                while (temp)
                {
                    /* Be an observer of reqOp's operation_will_close observable. */
                    /* This bumps reqOp's ref count. DOFSystem_ReqOpClosingObserver() correspondingly unbumps it. */
                    DOFOperation_AddOpWillCloseObserver( temp, DOFSystem_InterestOpClosing);
                    /* Send OAP pkt to provider to request the DOFValue corresponding to reqOp's property prp. When that
                       provider sends us the expected OAP response, call requestor callback rcb. */
                    routeResponseStatus = DOFRouter_RouteToProvider( temp );
                    if (routeResponseStatus == DOFROUTERSTATUS_SUCCESS)
                        temp = temp->next_sub_operation; /*Everything is groovy, keep going thru and adding operations*/
                    else
                    {
                        /*Unable to route op properly, so some type of error occured. Cancel the ops and clean them up*/
                        while(realContext->op != NULL)
                        {
                            DOFOperation_Cancel(realContext->op);
                            temp = realContext->op;
                            realContext->op = realContext->op->next_sub_operation;
                            DOFOperation_Destroy(temp);
                        }
                    }
                }
            }
            else
            {
                DOFSystem_CallCompleteWithException( realContext->op, DOFEXCEPTION_ACCESS_DENIED );
            }
        }
        else
        {
            DOFSystem_CallCompleteWithException( realContext->op, DOFEXCEPTION_ACCESS_DENIED );
        }
    }
    else
    {
        DOFSystem_CallCompleteWithException( realContext->op, DOFEXCEPTION_ACCESS_DENIED );
    }
    DOFOperation_Internal_Destroy(realContext->op);
    DOFSecurityMembrane_Destroy(realContext->secConn);
    DOFMem_Free(realContext);
}


/** @internal

   @param dnode
   @param shutdown
**/
static void ResponseHandler( OALThreadPoolItem dnode, boolean shutdown )
{
    /* Note: This is the enqueued function at the back end of the sys->responseQueue enqueue. */
    DOFOperation      respOp;
    SystemContext_t * node = (SystemContext_t *)dnode;

    PCR_Assert( dnode != NULL );
    respOp = node->reqOp;            /* requestor's reqOp is our response, for convenience here call it respOp. */
    PCR_Assert( node->system );
    /* DOFSystem_AcceptResponse() had filled responseQueue with a request operation that is directed to a specific
       original request operation, and also with a corresponding DOFResult. Known here as respOp and respResult. */

    /* Meaning of requestor-side extractions from the DOFSystem queue:
        ---respOp---            ---respResult--- */
    /*     valid                   valid     Normal response: get, set, invoke, subscribe, changed, signal.     */
    /*     valid                   NULL      respOp timeout notification from DOFSystem_OpTimedOutCallback(), or
                                             respOp cancel  notification from DOFSystem_ReqOpClosingObserver(). */
    /*     NULL                      X       Invalid.                                                           */
    PCR_Assert( respOp );
    /* This next locks/unlocks respOp at entry/exit. For any callback it calls it temporarily unlocks respOp. */
    Requestor_ProcessResponse( respOp, node->organizer.result );        /* Call does not destroy this result. */
    DOFOperation_Internal_Destroy( respOp );     /* Reduce ref counts- had incremented at queue's input side. */
    DOFResult_Destroy( node->organizer.result );
    DOFSystem_Release( node->system );
    /*If respOp is deallocated at this moment it is ONLY because that was done within the requestor (complete)callback.
      If respOp not deallocated ****** AT THIS MOMENT ******, ONLY the app will do anything more to destroy it. */
    DOFMem_Free( dnode );
}


/*
**  "Internal Public"  Methods.
*/

static boolean DOFSystem_PreparePermission(DOFOperation operation,
                                           DOFResult    result)
{
    DOFPermission perm = DOFResult_GetPermission(result);

    if(perm == NULL)
    {
        /* Set Required permissions in result. op usually tells us required prm. */
        if ( operation->operation_type == DOFOPERATIONTYPE_CHANGED )
        {
            /* Internal CHANGED operation is handled separately, since it uses the commandPerm and cannot send exceptions. */
            PCR_Assert( operation->commandPerm );
            if ( !DOFResult_GetException( result ) )
                perm = DOFPermission_Copy( operation->commandPerm );
        }
        else if (!(perm = DOFPermission_Copy( operation->responsePerm )))
        {
            DOFObjectID providerID = DOFResult_GetProviderID( result );

            /*
                * The only way a permission is not set for the request operation is if the request is
                * broadcast, in which case the provider ID must be set in the result.
                */
            PCR_Assert( providerID );
            PCR_Assert( operation->interface_id );
            perm = _createNeededPermission( operation->operation_type,
                                            (operation->operation_type != DOFOPERATIONTYPE_CHANGED),
                                            providerID, operation->interface_id );       /* Dictate required prm. */
        }
        if(perm != NULL)
        {
            DOFResult_SetPermission( result, perm );
            DOFPermission_Destroy( perm );
        }
        else
        {
            return FALSE;
        }
    }
    return TRUE;
}

/** @internal

   @note This calls DOFResult_Destroy( result ) but not DOFOperation_Internal_Destroy( provSideReqOp ).
   @note This call does nothing for NULL result.
**/
void DOFSystem_DeliverResponseBackToRequestor( DOFOperation provSideReqOp, DOFResult result )
{
    /* Note: Caller ProvideHandler() has provSideReqOp locked right now, but the four DOFSystem_CompleteXXXRequest()
             callers do not. Ordinarily we require an op be locked if its flags.bits are read. In this case we do not
             need to require it. Because: The worst that can happen at the moment we read is that some other thread may
             be changing the cancelled bit from 0 to 1, and what answer we get depends on exact thread interactions. The
             only effect of "wrongly" getting 0 is that we route our response to the requestor, when we might have been
             just as happy to drop the response. That is no different than wondering whether a particularly slow
             provider response beats our provSideReqOp timeout or not. */
    PCR_Assert( provSideReqOp );
    if (result && /* All callers supposed to give us a result. If NULL it is only by their failure to allocate memory.*/
        !provSideReqOp->flags.bits.cancelled)  /* e.g.: provCallback did too much yield/sleep so NOW op is cancelled. */
    {
        if (DOFSystem_PreparePermission(provSideReqOp, result))
        {
            PCR_Assert(DOFResult_GetPermission(result) != NULL);
            /* Mark result as local to this node. */
            DOFResult_SetLocal( result, TRUE );
            DOFRouter_RouteToRequestor( provSideReqOp, result );
            /* We don't check result. If router call fails we drop provSideReqOp. */
        }
    }
    /* Current decision: drop the response operation if we do not give to router or it does not accept us. */
    DOFResult_Destroy( result );
}

typedef struct RouteResopnseContext_t
{
    DOFSystem system;
    DOFOperation operation;
    DOFResult result;
} RouteResopnseContext_t;

static void routeResponseCallback(DOFError   status,
                                  void     * dcontext)
{
    RouteResopnseContext_t * context = (RouteResopnseContext_t*) dcontext;
    DOFPermission perm = DOFResult_GetPermission(context->result);
    DOFSecurityScope secScope = DOFSecurityMembrane_ValidateOutboundPermissions(context->system->channel.securityMembrane, perm, perm);

    if(secScope != NULL)
    {
        if(OALPermissionScope_GetIdCount(DOFSecurityScope_GetOALPermissionScope(secScope)) > 0)
        {
            DOFSystem_DeliverResponseBackToRequestor(context->operation, context->result);
        }
        DOFSecurityScope_Destroy(secScope);
    }
    DOFSystem_Release(context->system);
    DOFOperation_Internal_Destroy(context->operation);
    DOFResult_Destroy(context->result);
    DOFMem_Free(context);
}

void DOFSystem_CheckPermissionsAndRouteResponse(DOFSystem    system,
                                                DOFOperation operation,
                                                DOFResult    result)
{
    if(result != NULL && DOFSystem_PreparePermission(operation, result))
    {
        DOFPermission perm = DOFResult_GetPermission(result);
        DOFSecurityScope secScope = DOFSecurityMembrane_ValidateOutboundPermissions(system->channel.securityMembrane, perm, perm );

        if(secScope != NULL)
        {
            if(OALPermissionScope_GetIdCount(DOFSecurityScope_GetOALPermissionScope(secScope)) > 0)
            {
                DOFSystem_DeliverResponseBackToRequestor(operation, result);
            }
            DOFSecurityScope_Destroy(secScope);
        }
        else
        {
            RouteResopnseContext_t * context = (RouteResopnseContext_t*) DOFMem_Alloc(sizeof(RouteResopnseContext_t));

            if(context != NULL)
            {
                context->system = DOFSystem_Retain(system);
                context->operation = (DOFOperation_Retain(operation), operation);
                context->result = DOFResult_Copy(result);
                OALChannel_ExtendOutboundMembranePermissions(&(system->channel), perm, routeResponseCallback, context);
            }
        }
    }
}


/** @internal
    Process response from provider by calling the requestor callback.

    @note Though we here checks many things, the primary goal is to call one or both of the requestor callbacks (normal
          or complete callback). The ONLY ways this function will not call at least one of the requestor callbacks are:
            - respOp has already had its requestor callback called for the final time, or
            - the relevant provided callback function pointer is NULL. In that case this code does NOT call the callback
              but does destroy respResult.
            - respResult cannot be copied, as is required for internal operation.
          For the first case, this function does nothing.
          For the second and third cases, this function calls DOFOperation_Cancel().

    @param respOp      The original request operation, called respOp to reflect that we handle it now as a response.
                       MUST not be NULL.
    @param respResult  The DOFResult that contains the response information.
**/
void Requestor_ProcessResponse( DOFOperation respOp, DOFResult respResult )
{
    boolean timedOut;                            /* Would be a static function except that unit tests need access. */
    boolean cancelled;
    boolean runUsAgain = TRUE;

    PCR_Assert( respOp );
    /* Meaning of requestor-side submittals to the DOFSystem queue:
        ---respOp---            ---respResult--- */
    /*     valid                   valid        Normal response: get, set, invoke, subscribe, changed, signal.     */
    /*     valid                   NULL         respOp timeout notification from DOFSystem_OpTimedOutCallback(), or
                                                respOp cancel  notification from DOFSystem_ReqOpClosingObserver(). */
    /*     NULL                      X          Invalid.                                                           */

    while (runUsAgain)                                  /* Run exactly once in almost all cases. */
    {
        runUsAgain = FALSE; /*Can be set TRUE again for DOFEXCEPTION_RESULT_LOST- a rcvd event has sequence # problems*/
        /* Note: respOp does not have its mutex locked right now. We do not have an assert to prove it. */
        DOFOperation_Lock( respOp );                   /* This mutex must be unlocked before returning. */
        timedOut = respOp->flags.bits.timedOut;
        cancelled = respOp->flags.bits.cancelled;
        if (!respOp->flags.bits.completed && !respOp->flags.bits.completedAck)
        {
            /* .timedOut can have been serviced on a previous response and then the provider callback "saved" itself by
               extending its timeout. If belatedly now the official timeout-notifying queue entry gets to us (where
               respResult is NULL), it is expected that .timedOut is 0 again, though a short timeout extension would
               allow otherwise. */
            /* Before calling the callback, study timeout vs. cancel issues. */
            if (respResult || timedOut || cancelled)
            {
                /* We WILL call the callback. respResult can be:
                      not NULL (this is normal response needing service whether or not timeout/cancel happened),
                       or NULL (this is the queue entry to notify us of timeout/cancel). */
                boolean          submitAsComplete;
                boolean          resultCopiedOK             = TRUE;
                DOFException     completeCallbackException  = NULL;                /* Used only in complete callback. */
                DOFOperationType requestorOpType            = respOp->operation_type;

                if (timedOut || cancelled)
                {
                    /* We are told of a timeout or explicit cancel. So call callback "for the last time." [It is
                       legal that respResult be valid right now- that simply means this is 1st normal response to be
                       extracted from the queue after timeout or cancel notification.] */
                    respOp->flags.bits.completed = 1;
                    respOp->flags.bits.timedOut = 0;  /* So won't be hanging around if app complete callback extends. */
                    if (cancelled)
                    {
                        respOp->flags.bits.denyTimeoutExtend = 1;
                        completeCallbackException = respOp->completeCallbackException;
                    }
                    else if (!respOp->flags.bits.responseResultReceived)
                    {
                        completeCallbackException = DOFEXCEPTION_TIMEOUT;   /* Only used in the complete callback. */
                    }
                    else
                    { /* Any result is sufficient to withhold timeout exception being passed to complete callback. */
                    }
                }
                /* If this is a response and the only response we will ever get, close respOp and deny t/o extension. */
                if (respResult && !respOp->flags.bits.provider_oid_is_wildcard &&
                    (DOFResult_GetException( respResult ) != NULL || /* Any directed op getting exception needs close.*/
                     (requestorOpType != DOFOPERATIONTYPE_SUBSCRIBE && requestorOpType != DOFOPERATIONTYPE_REGISTER &&
                      requestorOpType != DOFOPERATIONTYPE_SESSION))) /* These directed 3 ops stay until cancel or t/o.*/
                {
                    respOp->flags.bits.completed = 1;
                    respOp->flags.bits.timedOut = 0;   /* So won't be hanging around if app complete callback extends.*/
                    respOp->flags.bits.denyTimeoutExtend = 1;
                }
                submitAsComplete = respOp->flags.bits.completed;  /* We fill after several paths above had their say. */

                if (respResult)
                {
                    DOFException resultException;

                    /*This is real response (NOT just a timeout/cancel). The response may contain an exception however*/
                    if (!respOp->flags.bits.acknowledged)
                    {
                        PCR_Assert( !respOp->firstResult );
                        if ((respOp->firstResult = DOFResult_Copy( respResult )) != NULL ) /* Released at op destroy. */
                        {
                            respOp->flags.bits.acknowledged = 1; /*DOFOp_IsAck and _WaitAck use this to say ack'ed/not*/
                            /* Nothing else sets .acknowledged- a real result must be received; that happens this path*/
                            /* Acknowledge result before calling cb. Docs do not require before vs. after- our choice.*/
                            if (respOp->notifyOpAcknowledgedCond)
                            {
                                /*No ref_count bump needed- _WaitAcknowledged does not _destroy until after its return*/
                                PCRCondition_Broadcast(respOp->notifyOpAcknowledgedCond); /*Signal Op_WaitAcknowledged*/
                            }
                        }
                        else
                            resultCopiedOK = FALSE;
                    }
                    if (resultCopiedOK)
                    {
                        if ((respResult = DOFResult_Copy( respResult )) != NULL)
                        {
                            /* Submit this normal respResult to app callback. */
                            boolean initialResponse = ( respOp->flags.bits.responseResultReceived == FALSE );

                            /* See the NULL respResult case just below: its newly created respResult cannot be destroyed
                               after we return to the caller, as is the normal case. So we destroy its respResult after
                               calling rcb. Because of that we must retain respResult here in our normal path. */

                            /* Handle automatic ACK lists. */
                            if ( DOFObjectID_IsWildcard( respOp->object_id.object_id ) &&
                                 respOp->operationConfig != NULL &&
                                 DOFOperationControl_IsAutoExclude( respOp->operationConfig ) &&
                                 respOp->operation_type != DOFOPERATIONTYPE_SUBSCRIBE &&
                                 respOp->operation_type != DOFOPERATIONTYPE_REGISTER &&
                                 respOp->operation_type != DOFOPERATIONTYPE_SESSION )
                            {
                                DOFProviderInfo pInfo = DOFResult_GetProviderInfo( respResult );
                                DOFObjectID providerID = DOFProviderInfo_GetProviderID( pInfo );
                                DOFOperationControl_AddProviderExclusion( respOp->operationConfig, providerID );
                            }
                            /* Only in this path do we call any of the non-completion requestor callbacks. This
                               callback at its top can safely check respResult for status. If status OK it can process
                               the response fully, even though it may be that the complete flag is set. */
                            PCR_Assert( respResult );
                            respOp->flags.bits.responseResultReceived = 1;
                            resultException = DOFResult_GetException( respResult );
                            /* resultException either contains a stream or it does not. Containing a stream means the
                               exception item coming in on the wire does not match the request operation's interface
                               definition. */
                            if (resultException && DOFException_IsStream( resultException ))
                            {
                                /*For "stream" type of resultexception call requestor callback with "FAILED" exception*/
                                resultException = DOFEXCEPTION_FAILED;
                            }
                            else
                            { /*For no resultException or "no stream" in it, just call the requestor callback normally*/
                            }
                            /* With all decisions and manipulations made, call the callback. */
                            switch ( respOp->operation_type )
                            {
                                case DOFOPERATIONTYPE_SESSION:
                                    if ((initialResponse || resultException) &&
                                        respOp->callbacks.scb && respOp->callbacks.scb->fns &&
                                        respOp->callbacks.scb->fns->SessionOpen)
                                    {
                                        DOFObject sessionObject = NULL;
                                        DOFProviderInfo providerInfo = DOFResult_GetProviderInfo( respResult );

                                        if (!resultException)
                                        {
                                            sessionObject = DOFSystem_CreateObject(
                                                                         DOFObject_GetSystem( respOp->provider_object ),
                                                                                    respOp->sessionOID );
                                            if (!sessionObject)
                                                resultException = DOFEXCEPTION_INSUFFICIENT_RESOURCES;
                                        }

                                        OALTimerItem_Destroy(respOp->ackTimerItem, FALSE);
                                        respOp->ackTimerItem = NULL;

                                        DOFOperation_Unlock( respOp );
                                        respOp->callbacks.scb->fns->SessionOpen( respOp->callbacks.scb,
                                                                                 respOp,
                                                                                 providerInfo,
                                                                                 sessionObject,
                                                                                 resultException );/*THE APP CALLBACK!*/
                                        DOFOperation_Lock( respOp );

                                        DOFObject_Destroy( sessionObject );
                                    }
                                    break;
                                case DOFOPERATIONTYPE_FINDINTERFACEDEFINITION:
                                    if (respOp->callbacks.rcb_def && respOp->callbacks.rcb_def->fns &&
                                        respOp->callbacks.rcb_def->fns->Define)
                                    {
                                        DOFInterface iface = NULL;

                                        if (!resultException)
                                        {
                                            DOFValue val = DOFResult_GetValue( respResult );

                                            if (val && DOFType_GetTypeID( DOFValue_GetDOFType( val ) ) == DOFTYPE_BLOB)
                                            {
                                                iface = DOFInterface_Create( respOp->interface_id,
                                                                             DOFValueBLOB_GetSize( val ),
                                                                             DOFValueBLOB_Get( val ) );
                                            }
                                        }
                                        if (iface)
                                        {
                                            DOFOperation_Unlock( respOp );
                                            respOp->callbacks.rcb_def->fns->Define( respOp->callbacks.rcb_def,
                                                                                    respOp, iface );/*THE APP CALLBACK*/
                                            DOFOperation_Lock( respOp );
                                        }
                                    }
                                    break;
                                case DOFOPERATIONTYPE_SUBSCRIBE: /* Same signature as Get. */
                                    if( initialResponse )
                                    {
                                        OALTimerItem_Destroy(respOp->ackTimerItem, FALSE);
                                        respOp->ackTimerItem = NULL;
                                    }

                                    if ( !initialResponse && !resultException && !DOFResult_IsNewSequence( respResult ) )
                                        break;
                                case DOFOPERATIONTYPE_CHANGED:   /* Same signature as Get. */
                                case DOFOPERATIONTYPE_GET:
                                    if ( respOp->callbacks.rcb_get &&
                                         respOp->callbacks.rcb_get->fns &&
                                         respOp->callbacks.rcb_get->fns->GetResult )
                                    {
                                        DOFOperation_Unlock( respOp );
                                        respOp->callbacks.rcb_get->fns->GetResult(respOp->callbacks.rcb_get, respOp,
                                                                                  DOFResult_GetProviderInfo(respResult),
                                                                                  DOFResult_GetValue( respResult ),
                                                                                  resultException); /*<<<APP CALLBACK!*/
                                        DOFOperation_Lock( respOp );
                                    }
                                    break;
                                case DOFOPERATIONTYPE_SET:
                                    if ( respOp->callbacks.rcb_set &&
                                         respOp->callbacks.rcb_set->fns &&
                                         respOp->callbacks.rcb_set->fns->SetResult )
                                    {
                                        DOFOperation_Unlock( respOp );
                                        respOp->callbacks.rcb_set->fns->SetResult( respOp->callbacks.rcb_set, respOp,
                                                                                  DOFResult_GetProviderInfo(respResult),
                                                                                   resultException );/*<<APP CALLBACK!*/
                                        DOFOperation_Lock( respOp );
                                    }
                                    break;
                                case DOFOPERATIONTYPE_REGISTER: /* Same signature as Invoke. */
                                    if( initialResponse )
                                    {
                                        OALTimerItem_Destroy(respOp->ackTimerItem, FALSE);
                                        respOp->ackTimerItem = NULL;
                                    }

                                    if (!DOFResult_IsSkippedSequence(respResult))
                                    {
                                        if (respOp->flags.bits.blockFollowOnIncomingSignalPDUs)
                                            break;
                                        if ( !initialResponse && !resultException && !DOFResult_IsNewSequence( respResult ) )
                                            break;
                                    }
                                    else
                                    {
                                        /* Call complete callback w/ exception, instead. */
                                        submitAsComplete = TRUE;
                                        respOp->flags.bits.completed = 1;
                                        completeCallbackException = DOFEXCEPTION_RESULT_LOST;
                                        /*Can overwrite directly DOFEXCEPTION_TIMEOUT in completeCallbackException*/
                                        break;
                                    }
                                    /* Fall through */
                                case DOFOPERATIONTYPE_SIGNAL:   /* Same signature as Invoke. */
                                case DOFOPERATIONTYPE_INVOKE:
                                    if (respOp->callbacks.rcb_inv &&
                                        respOp->callbacks.rcb_inv->fns &&
                                        respOp->callbacks.rcb_inv->fns->InvokeResult)
                                    {
                                        DOFOperation_Unlock( respOp );
                                        respOp->callbacks.rcb_inv->fns->InvokeResult( respOp->callbacks.rcb_inv, respOp,
                                                                                DOFResult_GetProviderInfo( respResult ),
                                                                                DOFResult_GetValueListCount(respResult),
                                                                                DOFResult_GetValueList( respResult ),
                                                                                      resultException );/*APP CALLBACK*/
                                        DOFOperation_Lock( respOp );
                                    }
                                    break;
                                default:
                                    /* Other operation types are not handled here. */
                                    break;
                            } /* end switch(). */
                            /*Don't destroy resultException- is held within result, is cleaned up by DOFResult_Destroy*/
                        }
                        else
                        {
                            resultCopiedOK = FALSE;
                            DOFResult_Destroy( respOp->firstResult ); /* Might be NULL or not-NULL. */
                        }
                    }
                }
                else /* respResult is NULL. */
                {
                    PCR_Assert( submitAsComplete );
                }

                /* Main callback was called above (or it was skipped because there is no supplied app callback. */
                if (resultCopiedOK)
                {
                    /*respOp refCount at least 1 from sysQ bump. rcb or even app loop may now have removed itsrefCount*/
                    if (submitAsComplete)
                    {
                        if (respOp->flags.bits.completed)
                        {
                            PCR_Assert( respOp->firstResult || completeCallbackException || cancelled );
                            if (respOp->flags.bits.denyTimeoutExtend)
                            {
                                DOFException_Destroy( respOp->completeCallbackException );
                                respOp->completeCallbackException = DOFException_Copy( completeCallbackException );
                                /* complete callback will see this in respOp. It is NULL or DOFEXCEPTION_TIMEOUT or
                                   DOFEXCEPTION_RESULT_LOST. */
                            }
                            /* respOp->callbacks is a union of all callback instances. All callback instances have the
                               ->Complete callback function as their second member. We can reach any of the Complete
                               callbacks by selecting any one of them to "call through." We choose
                               respOp->callbacks.rcb_get, but if respOp is instead set or invoke or similar, the correct
                               Complete callback will still be called. */
                            if (respOp->callbacks.rcb_get && respOp->callbacks.rcb_get->fns &&
                                respOp->callbacks.rcb_get->fns->Complete)
                            {
                                /* This callback may or may not allow ->flags.bits.completed to remain set. The only way
                                   it can clear .completed is to call DOFOperation_SetTimeout(), which may not succeed*/
                                DOFOperation_Unlock( respOp );/*With all decisions and manipulations made, call the cb*/

                                respOp->callbacks.rcb_get->fns->Complete( respOp->callbacks.rcb_get, respOp,
                                                                          completeCallbackException );/*<APP CALLBACK!*/

                                /* Do NOT destroy completeCallbackException- would destroy DOFEXCEPTION_TIMEOUT or
                                   DOFEXCEPTION_RESULT_LOST, which are instead destroyed at exception cleanup. */
                                DOFOperation_Lock( respOp );
                            }
                            if (!respOp->flags.bits.denyTimeoutExtend)
                            {
                                /* We did not fill respOp->completeCallbackException above. */
                                PCR_Assert( !respOp->completeCallbackException );
                                if (respOp->flags.bits.completed)                /* Complete callback did not extend. */
                                {
                                    /* respOp will hold the completeCallbackException for its remaining life. */
                                    DOFException_Destroy( respOp->completeCallbackException );
                                    respOp->completeCallbackException = DOFException_Copy( completeCallbackException );
                                    /* Do NOT destroy resp->completeCallbackException- would destroy
                                       DOFEXCEPTION_TIMEOUT or DOFEXCEPTION_RESULT_LOST, which are instead destroyed
                                       at exception cleanup. */
                                }
                                /* resultCopiedOK is still TRUE. */
                                if (completeCallbackException == DOFEXCEPTION_RESULT_LOST)
                                {
                                    /* Found DOFResult skipped sequence. */
                                    if (respOp->flags.bits.completed)
                                    {
                                        /*App did not call DOFOperation_SetTimeout() so cancel respOp, drop respResult*/
                                        DOFOperation_Cancel( respOp );
                                        respOp->flags.bits.blockFollowOnIncomingSignalPDUs = 1;/*Might be in sys queue*/
                                    }
                                    else
                                    {
                                        DOFResult tempResult;

                                        PCR_Assert( respResult );/*DOFEXCEPTION_RESULT_LOST happens only on respResult*/
                                        if ((tempResult = DOFResult_Copy( respResult )) != NULL)
                                        {
                                            /* We copied here to avoid losing respResult for the 2nd "run." */
                                            respResult = tempResult; /* For form. Because of ref counting, not needed.*/
                                            runUsAgain = TRUE;/*App called DOFOp_SetTimeout(respOp,non-0),so run again*/
                                            DOFResult_SetSequenceNumber( respResult,      /* To succeed at 2nd "run." */
                                                                         DOFResult_GetSequenceNumber( respResult ),
                                                                         FALSE,
                                                                         DOFResult_IsNewSequence( respResult ) );
                                        }
                                        else
                                        {
                                            /* ERROR: cannot do 2nd run, so drop the whole thing. Exception still held*/
                                            resultCopiedOK = FALSE;
                                            DOFOperation_Cancel( respOp );
                                            respOp->flags.bits.blockFollowOnIncomingSignalPDUs = 1;/*Might be in sys q*/
                                        }
                                    }
                                }
                            }
                        }
                        /* Callback may not have cleared ->flags.bits.completed, so MUST retest ->flags.bits.completed*/
                        if (resultCopiedOK)
                        {
                            if (respOp->flags.bits.completed)
                            {
                                respOp->flags.bits.completedAck = 1;  /* Never call respOp's requestor callback again.*/
                                /* App must destroy respOp. The above ->Complete callback is often thought a place to do
                                   that. BUT there is the case where DOFOperation_WaitComplete() is used to monitor
                                   progress of the respOp. For that situation, DOFOperation_Cancel() is allowed within
                                   the Complete() callback, but _Destroy() must be called somewhere else in the app. */
                                if (respOp->notifyOpCompleteCond)
                                    PCRCondition_Broadcast( respOp->notifyOpCompleteCond );/*Signal DOFOp_WaitComplete*/
                                DOFOperation_Cancel( respOp ); /*In some cases we were simply notified- we must cancel*/
                            }
                            else
                            {
                                /* The callback successfully extended respOp timeout. The "save" worked. The "save"
                                   would not be allowed if either .cancelled or .denyTimeoutExtend were TRUE at the time
                                   the complete callback executed. Note that thread swapping may have allowed enough
                                   time since the callback returned to us that either/both of those flag bits may have
                                   since set. Further, it is useless to call DOFOperation_GetTimeRemaining() anywhere
                                   near to right now, for the actual timeout extension "command" information is likely
                                   to be going through a router queue right now. */
                                PCR_Assert( !respOp->flags.bits.completedAck );/*Our thread only setter, so can assert*/
                            }
                        }
                    } /* end if (submitAsComplete). */
                    if (resultCopiedOK && respOp->notifyOpAcknowledgedCond && /* App wants acknowledgement. */
                        !respOp->flags.bits.acknowledged &&                   /* We have received no provider result. */
                        respOp->flags.bits.completedAck)                      /* We will never process respOp again. */
                    {
                        /* This is our last chance to signal DOFOperation_WaitAcknowledged- do it now. */
                        /* No ref_count bump needed- _WaitAcknowledged does not _destroy until after its return.*/
                        PCRCondition_Broadcast( respOp->notifyOpAcknowledgedCond ); /* Signal DOFOp_WaitAcknowledged()*/
                    }
                    DOFResult_Destroy( respResult );                                /* Unbump from above. */
                    /* Req callback needs to have extracted all that it needs from DOFResult. As we return from this
                       function, respResult will be destroyed. [We could unlock respOp for this destroy call.] */
                }
                else
                    DOFOperation_Cancel( respOp );  /* resultCopiedOK FALSE, we couldn't call callback, so cancel op. */
            }
            else
            {
                /* respResult is NULL and timedOut and cancelled are both FALSE, so . . . do not call the callback and
                   do not call the complete callback. Reason: There has been no cancel yet, but timedOut notification
                   was handled before by us in a previous queue entry that was a valid response (i.e. having a
                   DOFResult). At that notification, the callback extended the timeout (we can know this because
                   timedOut is now 0). We have now the trailing queue entry to notify us of timeout, but it is already
                   handled, so do nothing with it. */
            }
        }
        else
        { /* respOp is residual after being officially closed out by the app callback. */
        }
        DOFOperation_Unlock( respOp );
    } /* end while (runUsAgain) */
}

static void TransformResult(OALThreadPoolItem dnode,
                            boolean           shutdown)
{
    SystemContext_t * node = (SystemContext_t*) dnode;

    if(DOFResult_TransformStreamResult(node->organizer.result, node->reqOp))
    {
        ResponseHandler( dnode, shutdown );
    }
    else
    {
        DOFOperation_Internal_Destroy(node->reqOp);
        DOFResult_Destroy(node->organizer.result);
        DOFMem_Free(node);
    }
}

void DOFSystem_AcceptResponse_Wrapper( OALChannel chan, DOFOperation respOp, DOFResult respResult )
{
    DOFSystem_AcceptResponse(chan, respOp, respResult);/*router calls _Wrapper thru a table; cannot use boolean return*/
}



/** @internal
   Accept DOF provider response from DOFRouter and queue it for delivery to the requestor callback.

   @note This call does not and shall not block.
   @note Callers must not call with the respOp mutex locked.

   @param chan   DOFSystem of the response. Though findable in either op param, C OAL form says supply DOFSystem.
              OALChannel param used instead of DOFSystem to generically match the other calls used by OALChannel_Ctor().
   @param respOp     The original request operation info, now modified to describe the provider response.
   @param respResult The response DOFResult returned to us.

   @return TRUE if the OAP response is submitted for further processing, or FALSE if it is discarded.
 */
boolean DOFSystem_AcceptResponse( OALChannel chan, DOFOperation respOp, DOFResult respResult )
{
    /* Note: This respOp is never originally created by OAP. In fact, this function is never called where respOp was
             initially created by anything but the requestor. Although we term respOp to be a response operation, it is
             actually the same operation instance that was formed by an original request. */
    boolean   respResultFailure = FALSE;
    boolean   success           = FALSE;
    DOFSystem sys               = (DOFSystem)chan;
    OALThreadPoolFunc callback  = ResponseHandler;

    PCR_Assert( sys );
    PCR_Assert( respOp );
    if ( respOp->operation_type == DOFOPERATIONTYPE_BEGINPROVIDE )
        return TRUE;
    if ((sys = DOFSystem_Retain( sys )) != NULL)
    {
        if (respResult)
        {
            /*DOF gave us respResult, hence outbound. Test perms against security connector. Drop respOp if called for*/
            validationStatus valid = DOFSystem_ValidateOperationInboundPermissions( sys,
                                                                                    sys->channel.securityMembrane,
                                                                                    respOp->commandPerm,
                                                                                    DOFResult_GetPermission( respResult ),
                                                                                    respOp, NULL, respResult, FALSE );

            if (valid == VALIDATED)
            {
                /* If respResult holds a stream it needs to be transformed. */
                if (DOFResult_GetStream( respResult ) != NULL)
                {
                    callback = TransformResult;
                }
                else
                { /* respResult already transformed. */
                }
            }
            else
            {
                /* Not validated at the moment. Validation either in progress or known invalid or error: so do nothing*/
                switch (valid)
                {
                default:
                    {
                        PCR_Assert( FALSE );
                    }
                    /* Fall through. */
                case VALIDATION_ERROR:
                    /* Fall through. */
                case INVALIDATED:
                    /* Break out to drop respOp/respResult. */
                break;
                case VALIDATION_PENDING:
                    /* Validation determined by TRP callback setup in above _ValidateOperationOutboundPermissions call*/
                    DOFSystem_Release(sys);
                    return TRUE; /*Don't proceed with op/result yet, we will be called again w/ same respOp/respResult*/
                break;
                } /* end switch(). */
                respResultFailure = TRUE;
            }
        }
        else
        { /* No respResult- this is notification of respOp timeout or cancel. */
        }

        /* success is still FALSE. Everything must proceed perfectly to get it to be TRUE. */
        if (!respResultFailure)
        {
            /* Meaning of requestor-side submittals to the DOFSystem queue:
                ---respOp---            ---respResult---                                                              */
            /*     valid                   valid    Normal response: get, set, invoke, subscribe, changed, signal.    */
            /*     valid                   NULL     respOp timeout notification from DOFSystem_OpTimedOutCallback(), or
                                                    respOp cancel  notification from DOFSystem_ReqOpClosingObserver().*/
            /*     NULL                      X      Invalid.                                                          */

            if (sys->responseQueue)
            {
                if (!respResult || (respResult = DOFResult_Copy( respResult )) != NULL)
                {
                    /* Successful preparation: ref count bump is matched by unbump at queue's output side. */
                    SystemContext_t * node = (SystemContext_t *)DOFMem_Alloc( sizeof( SystemContext_t ) );

                    if (node != NULL)
                    {
                        /* sys and respResult were copied above. Those and respOp (retained here) are released at queue's output side. */
                        DOFOperation_Retain( respOp );
                        OALEnqueueItem_Init( (OALEnqueueItem)node, callback );
                        node->system = sys;
                        node->reqOp = respOp;
                        node->organizer.result = respResult;
                        /* System mutex could be locked around the if () block we are in, but that is believed unwise.
                           Instead, if problems appear where sys->responseQueue is here found NULL (would be due to
                           shutdown), we may ask router to check shutdown before calling us. */
                        success = OALEnqueue_Enqueue( sys->responseQueue, (OALEnqueueItem) node );
                        if (!success)
                        {
                            DOFOperation_Internal_Destroy( respOp );
                            DOFResult_Destroy( respResult );
                            DOFMem_Free( node );
                        }
                    }
                    else
                    {
                        DOFResult_Destroy( respResult );
                    }
                }
            }
        }
        if (!success)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_TRACE, "DOFSystem: Validation or processing error dropping response to requestor operation." );
            DOFSystem_Release( sys );
            PCRThread_Yield();                        /* So ResponseHandler() thread can extract. */
        }

    }

    return success;                                   /* FALSE: queue full or doesn't exist, or copy or stream error. */
}


/** @internal
 Observer callback to shutdown and/or disable things when a BeginXXX operation closes.

 @param o      Requester observable that BeginXXX Op is observing (originally stored at reqOp->operation_will_close).
 @param reqOp The specific requestor operation that is in the process of shutting down.
 **/
void DOFSystem_ReqOpClosingObserver( OALObserver self, OALObservable o, void* reqOp )
{
    PCR_Assert( o );
    PCR_Assert( reqOp );
    /* Note: _DOFOperation_Cancel() causes us to be called, and does so with reqOp locked. */
    PCR_Assert( ((DOFOperation)reqOp)->flags.bits.cancelled );  /* Cancel already happened or we would not be called. */
    if (!((DOFOperation)reqOp)->flags.bits.completedAck &&      /* This completedAck test is a quick check to avoid. */
                           /* enqueuing. If we enqueue and completedAck happens, the back side handler will drop it. */
        ((DOFOperation)reqOp)->provider_object)
    {
        /* reqOp is one of the DOFObject_BeginXXX calls. Send through queue to call its callback for final time. */
        if (!DOFSystem_AcceptResponse( DOFSystem_GetChannel( ((DOFOperation)reqOp)->provider_object->sys ),
                                       (DOFOperation)reqOp, NULL )) /* NULL DOFResult: This is cancel or timeout.*/
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "DOFSystem: Could not notify of a requestor operation closing." );
    }
    /* We don't remove ourselves as observer of o (which is reqOp->operation_will_close) because DOFOperation_Cancel()
     logic guarantees that reqOp is not cancelled twice, and it takes a _Cancel() to notify its observers. */
    DOFOperation_Internal_Destroy((DOFOperation)reqOp);   /* Unbump ref count bump from setting up observer callback. */
}


/*
**  "Public" Methods
*/


DOFOperation DOFSystem_BeginInterest(const DOFSystem sys,
                                     const DOFObjectID objectID,
                                     const DOFInterfaceID interfaceID,
                                           DOFInterestLevel level,
                                           DOFOperationControl operationConfig,
                                           uint32 timeout,
                                     const DOFSystemInterestCallback icb,
                                           void * ctx)
{
    DOFOperation reqOp;

    if ((sys == NULL) || (objectID == NULL) || (interfaceID == NULL))
        return NULL;

    if( ( level > DOFINTERESTLEVEL_WATCH && DOFObjectID_IsBroadcast( objectID ) ) ||
        ( level > DOFINTERESTLEVEL_ACTIVATE && DOFInterfaceID_IsWildcard( interfaceID ) ) )
        return NULL;

    reqOp = DOFOperation_Interest_Create( operationConfig, DOFSystem_GetChannel( sys ),
                                          objectID, interfaceID, icb, timeout, level,
                                          ctx);
    /* These would now work: DOFOperation_GetSystem( reqOp ), DOFOperation_GetType( reqOp ),
                             DOFOperation_GetObject( reqOp ), DOFOperation_GetProperty( reqOp ),
                             DOFOperation_GetContext( reqOp ), DOFOperation_GetTimeRemaining( reqOp ). */
    if (reqOp == NULL)
        return NULL;

    if (reqOp->securityScope != NULL)
    {
        if (OALPermissionScope_GetIdCount(DOFSecurityScope_GetOALPermissionScope(reqOp->securityScope)) > 0)
        {
            DOFOperation temp = reqOp;

            DOFRouterStatusCode_t routeResponseStatus;

            while (temp)
            {
                /* Be an observer of reqOp's operation_will_close observable. */
                /* This bumps reqOp's ref count. DOFSystem_ReqOpClosingObserver() correspondingly unbumps it. */
                DOFOperation_AddOpWillCloseObserver( temp, DOFSystem_InterestOpClosing);
                /* Send OAP pkt to provider to request the DOFValue corresponding to reqOp's property prp. When that
                   provider sends us the expected OAP response, call requestor callback rcb. */
                PCR_Assert( DOFObjectID_Compare(temp->object_id.object_id, objectID) == 0 );
                routeResponseStatus = DOFRouter_RouteToProvider( temp );
                if (routeResponseStatus == DOFROUTERSTATUS_SUCCESS)
                    temp = temp->next_sub_operation; /* Everything is groovy, keep going through and adding operations*/
                else
                {
                    /* Unable to route op properly, so some type of error occured. Cancel the ops and clean them up. */
                    while(reqOp != NULL)
                    {
                        DOFOperation_Cancel(reqOp);
                        temp = reqOp;
                        reqOp = reqOp->next_sub_operation;
                        DOFOperation_Destroy(temp);
                    }
                    return NULL;                /* return error */
                }
            }
        }
        else
        {
            DOFSystem_CallCompleteWithException( reqOp, DOFEXCEPTION_ACCESS_DENIED );
            /* The complete callback was called, and the operation was provided to the application in the complete callback.
               Since the callback was called we need return an operation here.  But, there is a chance tha the callback destroyed
               the operation, so we cannot access the operation after this point.  The other possibility is that the application
               is not using callbacks, or did not destroy the operation in the callback, in which case the operation will be
               complete, but still valid, and the application can call functions to access the operation that is returned. */
        }
    }
    else
    {
        ExtendPermContext context =
                            (ExtendPermContext)DOFMem_Alloc( sizeof( struct ExtendAcceptRequestContext_t ) );

        if (context != NULL)
        {
            context->secConn = DOFSecurityMembrane_Copy(DOFSystem_GetChannel( sys )->securityMembrane);
            context->op = (DOFOperation_Retain(reqOp), reqOp);
            if (context->secConn && context->op)
            {
                OALChannel_ExtendOutboundMembranePermissions(&(sys->channel), reqOp->commandPerm, interestExtendCB, context);
            }
            else
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DOFSystem: Failed to extend outboundPermissions in DOFSystem_BeginInterest." );
                DOFSecurityMembrane_Destroy(context->secConn);
                DOFOperation_Internal_Destroy(context->op);
                DOFMem_Free(context);
                DOFOperation_Destroy(reqOp);
                reqOp = NULL;
            }
        }
        else
        {
            DOFOperation_Destroy(reqOp);
            reqOp = NULL;
        }
    }

    return reqOp;                       /* if we made it this far, everything went ok */
}


/**
@}
**/
