/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/exception.h>
#include <internal/remotechannel.h>
#include <internal/tep_v1.h>
#include <internal/trp_v1.h>
#include <internal/permissionsetinternal.h>
#include <internal/resultinternal.h>

#include "tep_v1_operation.h"
#include "tep_v1_session.h"
#include "tep_v1_sm.h"

#define TEP_V1_DURATION     (24 * 60) / 8       /* One day */

TEP_V1 TEP_V1_Create(BlackboardData              blackboardData,
                     boolean                     isConnectionLess,
                     TEP_V1_SessionEventCallback sessionEventCallback)
{
    TEP_V1 ret;

    PCR_Assert(blackboardData != NULL);

    ret = (TEP_V1) DOFMem_Alloc(sizeof(TEP_V1_t));

    if(ret != NULL)
    {
        ret->blackboardData       = blackboardData;
        ret->isConnectionLess     = isConnectionLess;
        ret->sessionEventCallback = sessionEventCallback;
    }
    return ret;
}

void TEP_V1_Destroy(TEP_V1 self)
{
    if(self != NULL)
    {
        DOFMem_Free(self);
    }
}

void TEP_V1_ExtendPermissions(OALChannel         channel,
                              DOFAddress         address,
                              DOFPermissionSet   extendPerms,
                              ExtendCompleteCB   callback,
                              void             * context)
{
    TEPV1SessionState state = TEP_V1_SessionLookup(channel);

    if(state != NULL)
    {
        DOFKeyRequest keyRequest;

        PCR_Assert(extendPerms != NULL);
        keyRequest = DOFKeyRequest_Create(channel->remoteChannelData->secureCredentials,
                                          DEFAULT_NONCE_SIZE, FALSE, FALSE, NULL, extendPerms);
        if(keyRequest != NULL)
        {
            OALSessionInitiatorBlock initBlock =
                OALSessionInitiatorBlock_Create(0, channel->remoteChannelData->mode, keyRequest);

            if(initBlock != NULL)
            {
                DOFOperation extendOperation =
                    DOFOperation_Create_ExtendPermission(channel, channel->remoteChannelData->address, 30000,
                                                         initBlock, NULL, callback, context, NULL,
                                                         OALCipher_GetCurrentKeyState(channel->remoteChannelData->cipher),
                                                         TEP_V1_Session_GetTEP(state)->isConnectionLess);

                if(extendOperation != NULL)
                {
                    if(OALOperationManager_AddOperation(extendOperation, extendOperation->opID, FALSE) == DOFERROR_OK)
                    {
                        TEP_V1_SM_ProcessOperation(extendOperation);
                    }
                    DOFOperation_Internal_Destroy(extendOperation);
                }
                OALSessionInitiatorBlock_Release(initBlock);
            }
            DOFKeyRequest_Destroy(keyRequest);
        }
        TEP_V1_Session_Destroy(state);
    }
}

void TEP_V1_EndSession(OALChannel channel)
{
    TEPV1SessionState sessionState;

    PCR_Assert(channel != NULL);
    sessionState = TEP_V1_SessionLookup(channel);
    if(sessionState != NULL)
    {
        DOFOperation endOp = DOFOperation_Create_End(channel, TEP_V1_Session_GetOtherAddress(sessionState));

        if(endOp != NULL)
        {
            channel->routeToProvider(channel, endOp, endOp, TEP_V1_Session_GetOtherAddress(sessionState));
            DOFOperation_Internal_Destroy(endOp);
        }
        TEP_V1_Session_Destroy(sessionState);
    }
}

void TEP_V1_Authenticate(TEP_V1            tep,
                         boolean           isInitiator,
                         void            * sessionEventCallbackContext,
                         OALChannel        associatedChannel,
                         DOFAddress        remoteAddress,
                         DOFPermissionSet  extraPerms)
{
    PCR_Assert(tep != NULL);
    PCR_Assert(associatedChannel != NULL);
    PCR_Assert(associatedChannel->remoteChannelData);
    PCR_Assert(associatedChannel->remoteChannelData->flags.isConnection);
    PCR_Assert(associatedChannel->remoteChannelData->config.connection);
    PCR_Assert(remoteAddress != NULL);

    if(isInitiator)
    {
        DOFKeyRequest keyRequest;

        OALSessionInitiatorBlock initiatorBlock;

        boolean failure = TRUE;

        TEPV1SessionState state;

        PCR_Assert(associatedChannel->remoteChannelData->secureCredentials != NULL);
        PCR_Assert(DOFCredentials_IsResolved(associatedChannel->remoteChannelData->secureCredentials));
        state = TEP_V1_SessionLookup(associatedChannel);
        if(state != NULL)
        {
            /* We already have a session */
            TEP_V1_Session_Destroy(state);
            return;
        }
        if(associatedChannel->remoteChannelData->mode == NULL)
        {
            associatedChannel->remoteChannelData->mode = DOFSecurityModeCCM_Create(DOFCIPHERALGORITHM_AES, DOFCIPHERSTRENGTH_256, 4, TRUE);

            if(associatedChannel->remoteChannelData->mode == NULL)
            {
                tep->sessionEventCallback(sessionEventCallbackContext, NULL, TEP_V1_SessionEvent_NEWSESSION_FAILURE,
                                          NULL, associatedChannel, NULL, NULL, NULL, DOFEXCEPTION_INSUFFICIENT_RESOURCES,
                                          0, remoteAddress, NULL, associatedChannel->remoteChannelData->secureCredentials);
                return;
            }
        }
        keyRequest = DOFKeyRequest_Create( associatedChannel->remoteChannelData->secureCredentials, DEFAULT_NONCE_SIZE, TRUE,
                                           !DOFConnectionConfig_IsPermissionsExtendAllowed(associatedChannel->remoteChannelData->config.connection), NULL,
                                           DOFConnectionConfig_GetPermissions(associatedChannel->remoteChannelData->config.connection));
        if(keyRequest != NULL)
        {
            boolean keepGoing = TRUE;

            if(extraPerms != NULL)
            {
                keepGoing = DOFPermissionSet_AddPermissionsFromSet(keyRequest->permission_set, extraPerms);
            }
            if(keepGoing)
            {
                PCRSecureRandom_Generate(keyRequest->nonceLength, keyRequest->nonce);
                state = TEP_V1_Session_CreateNewState(FALSE, keyRequest->permission_set, associatedChannel,
                                                      NULL, remoteAddress, tep, sessionEventCallbackContext);

                if(state != NULL)
                {
                    initiatorBlock = OALSessionInitiatorBlock_Create(TEP_V1_DURATION,
                                                                     associatedChannel->remoteChannelData->mode,
                                                                     keyRequest);
                    if(initiatorBlock != NULL)
                    {
                        DOFOperation beginSessionOperation =
                            DOFOperation_Create_BeginSession(associatedChannel, remoteAddress,
                                                             TEP_V1_TIMEOUT, initiatorBlock, NULL, tep,
                                                             NULL, tep->isConnectionLess);
                        if(beginSessionOperation != NULL)
                        {
                            if(OALOperationManager_AddOperation(beginSessionOperation, beginSessionOperation->opID, FALSE) == DOFERROR_OK)
                            {
                                TEP_V1_SM_ProcessOperation(beginSessionOperation);
                                failure = FALSE;
                            }
                            DOFOperation_Internal_Destroy(beginSessionOperation);
                        }
                        OALSessionInitiatorBlock_Release(initiatorBlock);
                    }
                    if(failure)
                    {
                        TEP_V1_Session_Remove(state, associatedChannel);
                    }
                    TEP_V1_Session_Destroy(state);
                }
            }
            DOFKeyRequest_Destroy(keyRequest);
        }
        if(failure)
        {
            tep->sessionEventCallback(sessionEventCallbackContext, NULL, TEP_V1_SessionEvent_NEWSESSION_FAILURE,
                                      NULL, associatedChannel, NULL, NULL, NULL, DOFEXCEPTION_FAILED, 0, remoteAddress, NULL,
                                      associatedChannel->remoteChannelData->secureCredentials);
        }
    }
    else    /* responder side of a TEP session */
    {
        TEPV1SessionState state = TEP_V1_Session_CreateNewState(TRUE, NULL, associatedChannel, NULL,
                                                                remoteAddress, tep, sessionEventCallbackContext);

        TEP_V1_Session_Destroy(state);
    }
}

boolean TEP_V1_SendNoOp(OALChannel channel)
{
    DOFOperation noOp;

    PCR_Assert(channel != NULL);
    noOp = (DOFOperation) DOFMem_Alloc(sizeof(DOFOperation_t));
    if(noOp != NULL)
    {
        PCRMem_Set(noOp, 0, sizeof(DOFOperation_t));
        noOp->sources = (DOFOperationSource_t *) DOFMem_Alloc(sizeof(DOFOperationSource_t));
        if(noOp->sources != NULL)
        {
            PCRMem_Set(noOp->sources, 0, sizeof(DOFOperationSource_t));
            noOp->ref_count = 1;
            noOp->flags.bits.cancelled = TRUE;
            noOp->operation_type = DOFOPERATIONTYPE_TEP_NEGOTIATION_NOOP;
            channel->routeToProvider(channel, noOp, noOp, NULL);
            DOFOperation_Internal_Destroy(noOp);
            return TRUE;
        }
        DOFMem_Free(noOp);
    }
    return FALSE;
}
