/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/resultinternal.h>

#include "tep_v1_operation.h"

static DOFOperation DOFOperation_CreateBaseOperation(DOFOperationType opType,
                                                     OALChannel       channel,
                                                     DOFAddress       address);

static void DOFOperation_CleanupBaseOperation(DOFOperation operation);

static void DOFOperation_CleanupTepContext(void * operationContext);

static void DOFOperation_TEPOperationClosing(OALObserver self,
                                             OALObservable   observable,
                                             void          * arg);

DOFOperation DOFOperation_Create_BeginSession(OALChannel               channel,
                                              DOFAddress               address,
                                              uint32                   timeout,
                                              OALSessionInitiatorBlock initiatorBlock,
                                              OALStream                iBlock,
                                              TEP_V1                   tep,
                                              DOFOperationID           operationId,
                                              boolean                  isConnectionless)
{
    TEP_V1_OperationContext tepContext;

    boolean internallyCreated = (operationId != NULL);

    PCR_Assert(channel != NULL);
    PCR_Assert(address != NULL);
    PCR_Assert(tep != NULL);
    PCR_Assert(!iBlock == !operationId);
    tepContext = DOFMem_Alloc(sizeof(TEP_V1_OperationContext_t));
    if(tepContext != NULL)
    {
        DOFOperation ret;

        ret = DOFOperation_CreateBaseOperation(DOFOPERATIONTYPE_TEP_BEGIN, channel, address);
        if(ret != NULL)
        {
            if(operationId == NULL)     /* New security request */
            {
                operationId = DOFOperationID_CreateLocal();
            }
            else
            {
                operationId = DOFOperationID_Copy(operationId);       /* always retain */
            }
            if(operationId != NULL)
            {
                ret->opID = operationId;    /* already retained */
                ret->tepContext = tepContext;
                ret->sources->oal_timeout = timeout;
                PCRMem_Set(tepContext, 0, sizeof(TEP_V1_OperationContext_t));
                tepContext->initiatorBlock = OALSessionInitiatorBlock_Retain(initiatorBlock);
                tepContext->tep = tep;
                if(iBlock != NULL)
                {
                    OALStream_Retain(iBlock);
                    tepContext->iBlock = iBlock;
                }
                ret->securityScope = DOFSecurityScope_Copy(DOF_GetDefaultSecurityScope(DOF_GetDof()));
                ret->flags.bits.internally_created = internallyCreated;
                ret->flags.bits.is_connectionless = isConnectionless;
                DOFOperation_AddOpWillCloseObserver(ret, DOFOperation_TEPOperationClosing);
                return ret;
            }
            DOFOperation_CleanupBaseOperation(ret);
        }
        DOFMem_Free(tepContext);
    }
    return NULL;
}

DOFOperation DOFOperation_Create_ExtendPermission(OALChannel                 channel,
                                                  DOFAddress                 address,
                                                  uint32                     timeout,
                                                  OALSessionInitiatorBlock   initiatorBlock,
                                                  DOFOperationID             operationId,
                                                  ExtendCompleteCB           callback,
                                                  void                     * context,
                                                  OALStream                  iBlock,
                                                  OALCipherKeyState          keyState,
                                                  boolean                    isConnectionless)
{
    TEP_V1_OperationContext tepContext;

    boolean internallyCreated = (operationId != NULL);

    PCR_Assert(channel != NULL);
    tepContext = DOFMem_Alloc(sizeof(TEP_V1_OperationContext_t));
    if(tepContext != NULL)
    {
        DOFOperation ret;

        ret = DOFOperation_CreateBaseOperation(DOFOPERATIONTYPE_TEP_EXTEND, channel, address);
        if(ret != NULL)
        {
            if(operationId == NULL)     /* New security request */
            {
                operationId = DOFOperationID_CreateLocal();
            }
            else
            {
                operationId = DOFOperationID_Copy(operationId);       /* always retain */
            }
            if(operationId != NULL)
            {
                ret->opID = operationId;    /* already retained */
                ret->tepContext = tepContext;
                ret->sources->oal_timeout = timeout;
                PCRMem_Set(tepContext, 0, sizeof(TEP_V1_OperationContext_t));
                tepContext->keyState = keyState;
                tepContext->initiatorBlock = OALSessionInitiatorBlock_Retain(initiatorBlock);
                tepContext->flags.extendingPermission = TRUE;
                DOFOperation_AddOpWillCloseObserver(ret, DOFOperation_TEPOperationClosing);
                ret->securityScope = DOFSecurityScope_Copy(DOF_GetDefaultSecurityScope(DOF_GetDof()));
                ret->flags.bits.is_connectionless = isConnectionless;
                tepContext->extendCompleteCallback = callback;
                tepContext->extendContext = context;
                tepContext->calledCallback = FALSE;
                if(iBlock != NULL)
                {
                    OALStream_Retain(iBlock);
                    tepContext->iBlock = iBlock;
                }
                ret->flags.bits.internally_created = internallyCreated;
                return ret;
            }
            DOFOperation_CleanupBaseOperation(ret);
        }
        DOFMem_Free(tepContext);
    }
    return NULL;
}

DOFOperation DOFOperation_Create_TepRekey(OALChannel       channel,
                                          DOFAddress       address,
                                          DOFPermissionSet permissionSet,
                                          uint8            duration,
                                          TEP_V1           tep,
                                          uint32           timeout)
{
    TEP_V1_OperationContext tepContext;

    tepContext = DOFMem_Alloc(sizeof(TEP_V1_OperationContext_t));
    if(tepContext != NULL)
    {
        DOFKeyRequest keyRequest = DOFKeyRequest_Create(channel->remoteChannelData->secureCredentials,
                                                        DEFAULT_NONCE_SIZE, FALSE,
                                                        !DOFConnectionConfig_IsPermissionsExtendAllowed(channel->remoteChannelData->config.connection),
                                                        NULL, permissionSet);

        if(keyRequest)
        {
            DOFOperation ret = DOFOperation_CreateBaseOperation(DOFOPERATIONTYPE_TEP_BEGIN, channel, address);

            PCRSecureRandom_Generate(keyRequest->nonceLength, keyRequest->nonce);
            if(ret != NULL)
            {
                ret->securityScope = DOFSecurityScope_Copy(DOF_GetDefaultSecurityScope(DOF_GetDof()));
                if(ret->securityScope != NULL)
                {
                    PCRMem_Set(tepContext, 0, sizeof(TEP_V1_OperationContext_t));
                    tepContext->initiatorBlock = OALSessionInitiatorBlock_Create(duration, channel->remoteChannelData->mode, keyRequest);
                    if(tepContext->initiatorBlock != NULL)
                    {
                        OALObserver obs = DOFOperation_AddOpWillCloseObserver(ret, DOFOperation_TEPOperationClosing);

                        DOFKeyRequest_Destroy(keyRequest);
                        if(obs != NULL)
                        {
                            ret->opID = DOFOperationID_CreateLocal();
                            if(ret->opID != NULL)
                            {
                                tepContext->flags.rekeying = TRUE;
                                tepContext->tep = tep;
                                ret->tepContext = tepContext;
                                ret->sources->oal_timeout = timeout;
                                return ret;
                            }
                            DOFOperation_RemoveOpWillCloseObserver(ret, obs);
                        }
                        OALSessionInitiatorBlock_Release(tepContext->initiatorBlock);
                    }
                    DOFSecurityScope_Destroy(ret->securityScope);
                }
                DOFOperation_CleanupBaseOperation(ret);
            }
            DOFKeyRequest_Destroy(keyRequest);
        }
        DOFMem_Free(tepContext);
    }
    return NULL;
}

DOFOperation DOFOperation_Create_TepRekey_Internal(OALChannel               channel,
                                                   DOFAddress               address,
                                                   OALStream                iBlock,
                                                   OALSessionInitiatorBlock initiatorBlock,
                                                   DOFOperationID           operationId,
                                                   TEP_V1                   tep,
                                                   uint32                   timeout)
{
    TEP_V1_OperationContext tepContext;

    tepContext = DOFMem_Alloc(sizeof(TEP_V1_OperationContext_t));
    if(tepContext != NULL)
    {
        DOFOperation ret = DOFOperation_CreateBaseOperation(DOFOPERATIONTYPE_TEP_BEGIN, channel, address);

        if(ret != NULL)
        {
            ret->securityScope = DOFSecurityScope_Copy(DOF_GetDefaultSecurityScope(DOF_GetDof()));
            if(ret->securityScope != NULL)
            {
                PCRMem_Set(tepContext, 0, sizeof(TEP_V1_OperationContext_t));
                tepContext->initiatorBlock = OALSessionInitiatorBlock_Retain(initiatorBlock);
                if(tepContext->initiatorBlock != NULL)
                {
                    tepContext->iBlock = iBlock;
                    OALStream_Retain(iBlock);
                    ret->opID = DOFOperationID_Copy(operationId);
                    if(ret->opID != NULL)
                    {
                        OALObserver obs = DOFOperation_AddOpWillCloseObserver(ret, DOFOperation_TEPOperationClosing);

                        if(obs != NULL)
                        {
                            tepContext->flags.rekeying = TRUE;
                            tepContext->tep = tep;
                            ret->flags.bits.internally_created = TRUE;
                            ret->tepContext = tepContext;
                            ret->sources->oal_timeout = timeout;
                            OALSessionInitiatorBlock_Release(tepContext->initiatorBlock);
                            OALStream_Release(iBlock, TRUE);
                            return ret;
                        }
                        DOFOperationID_Destroy(ret->opID);
                        OALStream_Release(iBlock, TRUE);
                    }
                    OALSessionInitiatorBlock_Release(tepContext->initiatorBlock);
                }
                DOFSecurityScope_Destroy(ret->securityScope);
            }
            DOFOperation_CleanupBaseOperation(ret);
        }
        DOFMem_Free(tepContext);
    }
    return NULL;
}

DOFOperation DOFOperation_Create_Reject(OALChannel     channel,
                                        DOFAddress     address,
                                        DOFOperationID operationId,
                                        boolean        isConnectionless,
                                        uint8          rejectCode)
{
    TEP_V1_OperationContext tepContext;

    /* We don't want to assert here, since this is called based on received data, so just return NULL if we don't have what we need. */
    if ((operationId != NULL) && (channel != NULL) && (address != NULL))
    {
        DOFOperationID_Copy(operationId);       /* always retain */
        tepContext = DOFMem_Alloc(sizeof(TEP_V1_OperationContext_t));
        if(tepContext != NULL)
        {
            DOFOperation ret;

            ret = DOFOperation_CreateBaseOperation(DOFOPERATIONTYPE_TEP_REJECT, channel, address);
            if(ret != NULL)
            {
                ret->opID = operationId;    /* already retained */
                ret->tepContext = tepContext;
                PCRMem_Set(tepContext, 0, sizeof(TEP_V1_OperationContext_t));
                tepContext->rejectCode = rejectCode;
                ret->flags.bits.internally_created = TRUE;
                ret->flags.bits.is_connectionless = isConnectionless;
                ret->securityScope = DOFSecurityScope_Copy(DOF_GetDefaultSecurityScope(DOF_GetDof()));
            }
            else
            {
                DOFMem_Free(tepContext);
            }
            return ret;
        }
        DOFOperationID_Destroy(operationId);        /* something failed to alloc */
    }
    return NULL;
}

DOFOperation DOFOperation_Create_End(OALChannel channel,
                                     DOFAddress address)
{
    DOFOperation op = DOFOperation_CreateBaseOperation(DOFOPERATIONTYPE_TEP_END, channel, address);

    if(op != NULL)
    {
        op->tepContextDestroyFunction = NULL;
    }
    return op;
}

DOFOperation DOFOperation_Create_SessionEnding(OALChannel channel,
                                               DOFAddress address)
{
    DOFOperation op = DOFOperation_CreateBaseOperation(DOFOPERATIONTYPE_TEP_SESSIONENDING, channel, address);

    if(op != NULL)
    {
        op->tepContextDestroyFunction = NULL;
    }
    return op;
}

static DOFOperation DOFOperation_CreateBaseOperation(DOFOperationType opType,
                                                     OALChannel       channel,
                                                     DOFAddress       address)
{
    DOFOperation ret = (DOFOperation) DOFMem_Alloc(sizeof(DOFOperation_t));

    if(ret != NULL)
    {
        OALObservable observer = OALObservable_Create();

        CancelOperationContext_t * cancelContext;

        if(observer == NULL)
        {
            DOFMem_Free(ret);
            return NULL;
        }
        cancelContext = DOFOperationManager_CreateCancelContext(ret);
        if(cancelContext != NULL)
        {
            PCRMem_Set(ret, 0, sizeof(DOFOperation_t));
            ret->sources = (DOFOperationSource)DOFMem_Alloc(sizeof(DOFOperationSource_t));
            if( ret->sources != NULL )
            {
                PCRMem_Set(ret->sources, 0, sizeof(DOFOperationSource_t));
                ret->ref_count = 1;
                ret->operation_will_close = observer;
                ret->operation_type = opType;
                ret->sources->channel = OALChannel_Retain(channel);
                ret->tepContextDestroyFunction = DOFOperation_CleanupTepContext;
                ret->sources->address = DOFAddress_Copy(address);
                ret->cancelContext = cancelContext;
                return ret;
            }
            DOFMem_Free(cancelContext);
        }
        OALObservable_Destroy(observer);
        DOFMem_Free(ret);
        return NULL;
    }

    return ret;
}

static void DOFOperation_CleanupBaseOperation(DOFOperation operation)
{
    PCR_Assert(operation != NULL);

    OALChannel_Release(operation->sources->channel);
    DOFMem_Free(operation->sources);
    DOFMem_Free(operation->cancelContext);
    OALObservable_Destroy(operation->operation_will_close);
    DOFMem_Free(operation);
}

static void DOFOperation_CleanupTepContext(void * operationContext)
{
    TEP_V1_OperationContext tepContext = (TEP_V1_OperationContext) operationContext;

    PCR_Assert(operationContext != NULL);
    if(tepContext->initiatorBlock != NULL)
    {
        OALSessionInitiatorBlock_Release(tepContext->initiatorBlock);
    }
    DOFKeyRequest_Destroy(tepContext->keyRequest);
    DOFObjectIDDomain_Destroy(tepContext->domain);
    OALKeyGrant_Release(tepContext->initGrant);
    OALSecurityMode_Destroy(tepContext->initialState);
    DOFSecurityScope_Destroy(tepContext->connectionSecurityScope);
    OALStream_Release(tepContext->iBlock, TRUE);
    OALStream_Release(tepContext->rBlock, TRUE);
    OALStream_Release(tepContext->aBlock, TRUE);
    OALStream_Release(tepContext->rejectData, TRUE);
    if(tepContext->secureSid != NULL)
    {
        source_release(tepContext->secureSid);
    }

    DOFMem_Free(operationContext);
}

static void DOFOperation_TEPOperationClosing(OALObserver self,
                                             OALObservable   observable,
                                             void          * arg)
{
    DOFOperation op = (DOFOperation) arg;

    TEP_V1_OperationContext tepContext = (TEP_V1_OperationContext) op->tepContext;

    if(tepContext->flags.extendingPermission)
    {
        if(!tepContext->calledCallback && (tepContext->extendCompleteCallback != NULL))
        {
            tepContext->extendCompleteCallback(PCRSTATUS_TIMEOUT, tepContext->extendContext);
            tepContext->calledCallback = TRUE;
        }
    }
    DOFOperation_Internal_Destroy(op);                          /* Observer's retain */
}
