/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/esp_v2_options.h>
#include <internal/marshal.h>
#include <internal/remotechannel.h>
#include <internal/tep_v1.h>
#include <internal/trp_v1.h>

#include "tep_v1_operation.h"
#include "tep_v1_receiveblock.h"
#include "tep_v1_receiveutil.h"
#include "tep_v1_session.h"
#include "tep_v1_sm.h"

typedef void (*TEP_V1_ReceiveFunction)(TEP_V1_ReceiveBlock block);

static void TEP_V1_Receive_NegotiateSession(TEP_V1_ReceiveBlock block);

static void TEP_V1_Receive_NegotiateSession_r(TEP_V1_ReceiveBlock block);

static void TEP_V1_Receive_EndSession(TEP_V1_ReceiveBlock block);

static void TEP_V1_Receive_Reject_r(TEP_V1_ReceiveBlock block);

static void TEP_V1_Receive_SessionEnding(TEP_V1_ReceiveBlock block);

static void TEP_V1_Receive_InvalidOpcode(TEP_V1_ReceiveBlock block);

ProtocolLayerResult TEP_V1_Receive(TEP_V1 self)
{
    /* {response, command} */
    static const TEP_V1_ReceiveFunction functions [7][2] = {
            /* 0 - Reject Session */
        {TEP_V1_Receive_Reject_r,           TEP_V1_Receive_InvalidOpcode},
            /* 1 - Begin Session */
        {TEP_V1_Receive_NegotiateSession_r, TEP_V1_Receive_NegotiateSession},
            /* 2 - Extend Session */
        {TEP_V1_Receive_InvalidOpcode,      TEP_V1_Receive_InvalidOpcode},
            /* 3 - UNUSED */
        {TEP_V1_Receive_InvalidOpcode,      TEP_V1_Receive_InvalidOpcode},
            /* 4 - Confirm Session */
        {TEP_V1_Receive_InvalidOpcode,      TEP_V1_Receive_InvalidOpcode},
            /* 5 - End Session */
        {TEP_V1_Receive_InvalidOpcode,      TEP_V1_Receive_EndSession},
            /* 6 - Session Ending */
        {TEP_V1_Receive_InvalidOpcode,      TEP_V1_Receive_SessionEnding}
    };

    uint16 appId;

    PCR_Assert(self != NULL);
    if(!self->blackboardData->associatedChannel->remoteChannelData->flags.isConnection)
    {
        return PLRESULT_ERROR;
    }
    if(OALMarshal_StreamUncompress2(self->blackboardData->receiveStream, &appId) == DOFERROR_OK)
    {
        if(appId == TEP_V1_APPID)
        {
            uint8 pdu;

            if(OALStream_Read8(self->blackboardData->receiveStream, &pdu))
            {
                if((pdu & 0x0F) < 7)
                {
                    TEP_V1_ReceiveBlock_t receiveBlock;

                    DOFOperation operation = NULL;

                    if(self->blackboardData->opID != NULL)
                    {
                        operation = OALOperationManager_CheckExistingOperation(DOFSecurityScope_GetDomainID(DOF_GetDefaultSecurityScope(DOF_GetDof())), self->blackboardData->opID);
                    }

                    /* Verify that the retry count is higher than the last time this operation was seen.
                       This is really EPP logic, but since it is not included there right now, just do it here. */
                    if (operation)
                    {
                        DOFOperation_Lock(operation);
                        if (self->blackboardData->flags.isCommandNotResponse)
                        {
                            if (operation->sources->retry_count >= self->blackboardData->retry)
                            {
                                DOFOperation_Unlock(operation);
                                DOFOperation_Internal_Destroy(operation);
                                return PLRESULT_OKDONOTHING;
                            }
                            else
                            {
                                operation->sources->retry_count = self->blackboardData->retry;
                                DOFOperation_Unlock(operation);
                            }
                        }
                        else
                        {
                            DOFOperation_Unlock(operation);
                        }
                    }

                    TEP_V1_ReceiveBlock_Init(&receiveBlock, self->blackboardData->receiveStream, self, operation, pdu);
                    functions[(receiveBlock.firstByte) & 0x0F][self->blackboardData->flags.isCommandNotResponse](&receiveBlock);
                    switch(receiveBlock.status)
                    {
                    case TEP_V1_ReceiveStatus_OK:
                        if(receiveBlock.operation != NULL)
                        {
                            TEP_V1_SM_ProcessOperation(receiveBlock.operation);
                            DOFOperation_Internal_Destroy(receiveBlock.operation);
                        }
                        if(receiveBlock.sessionState != NULL)
                        {
                            TEP_V1_Session_Destroy(receiveBlock.sessionState);
                        }
                        return PLRESULT_OKDONOTHING;

                    case TEP_V1_ReceiveStatus_DONOTHING:
                        DOFOperation_Internal_Destroy(receiveBlock.operation);
                        if(receiveBlock.sessionState != NULL)
                        {
                            TEP_V1_Session_Destroy(receiveBlock.sessionState);
                        }
                        return PLRESULT_OKDONOTHING;
                        break;

                    default:
                    case TEP_V1_ReceiveStatus_CLOSE_CONNECTION:
                    case TEP_V1_ReceiveStatus_APP_ERROR:
                    case TEP_V1_ReceiveStatus_MEMORY_ERROR:
                    case TEP_V1_ReceiveStatus_STREAM_ERROR:
                        if((receiveBlock.operation != NULL) && receiveBlock.operation->flags.bits.internally_created)
                        {
                            DOFOperation_Lock(receiveBlock.operation);
                            receiveBlock.operation->operation_type = DOFOPERATIONTYPE_TEP_REJECT;
                            DOFOperation_Unlock(receiveBlock.operation);
                            TEP_V1_SM_ProcessOperation(receiveBlock.operation);
                            DOFOperation_Internal_Destroy(receiveBlock.operation);
                            if(receiveBlock.sessionState != NULL)
                            {
                                TEP_V1_Session_Destroy(receiveBlock.sessionState);
                            }
                            return PLRESULT_OKDONOTHING;        /* Need to send a reject */
                        }
                        if (receiveBlock.operation != NULL)
                        {
                            DOFOperation_Internal_Destroy(receiveBlock.operation);
                        }
                        if (receiveBlock.sessionState != NULL)
                        {
                            TEP_V1_Session_Destroy(receiveBlock.sessionState);
                        }
                        return PLRESULT_ERROR;
                    }
                }
                else
                {
                    return PLRESULT_ERROR;
                }
            }
            else if(self->isConnectionLess && !self->blackboardData->associatedChannel->remoteChannelData->flags.mustFinalizeTep)
            {
                TEPV1SessionState state = TEP_V1_SessionLookup(self->blackboardData->associatedChannel);

                if(state != NULL)
                {
                    void * newContext = TEP_V1_Session_GetContext(state);

                    newContext = self->sessionEventCallback(newContext, NULL,
                                                            TEP_V1_SessionEvent_NEWSESSION_SUCCESS_SENDKEY,
                                                            NULL, self->blackboardData->associatedChannel,
                                                            NULL, NULL, NULL, NULL, 0, self->blackboardData->transportAddr, NULL,
                                                            self->blackboardData->associatedChannel->remoteChannelData->secureCredentials);
                    newContext = self->sessionEventCallback(newContext, NULL,
                                                            TEP_V1_SessionEvent_NEWSESSION_FINALIZE,
                                                            NULL, self->blackboardData->associatedChannel,
                                                            NULL, NULL, NULL, NULL, 0, self->blackboardData->transportAddr, NULL,
                                                            self->blackboardData->associatedChannel->remoteChannelData->secureCredentials);
                    TRPRouter_AddDynamicRoute(self->blackboardData->associatedChannel->remoteChannelData->secureCredentials,
                                            self->blackboardData->associatedChannel, self->blackboardData->transportAddr, FALSE);
                    TEP_V1_Session_SetContext(state, newContext);
                    TEP_V1_Session_IncrementKeyCount(state);
                    TEP_V1_Session_SetupRekeyTimer(state);
                    OALConnection_SetPeerCredentials(self->blackboardData->associatedChannel->remoteChannelData->OALConn,
                                                    TEP_V1_Session_GetOtherCredentials(state));
                    self->blackboardData->associatedChannel->remoteChannelData->flags.mustFinalizeTep = TRUE;
                    TEP_V1_Session_Destroy(state);
                }
            }
            return PLRESULT_OKDONOTHING;
        }
        /* else -- Invalid AppId */
    }
    return PLRESULT_ERROR;
}

static void TEP_V1_Receive_NegotiateSession(TEP_V1_ReceiveBlock block)
{
    PCR_Assert(block != NULL);

    if(block->tep->blackboardData->flags.isMulticast)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_INFO, "TEP V1: Rejecting multicast negotiation.");
        block->status = TEP_V1_ReceiveStatus_CLOSE_CONNECTION;
        return;
    }
    else
    {
        boolean isConfirm  = (block->firstByte & 0x20) >> 5;
        boolean useKey = (block->firstByte & 0x10) >> 4;

        if(useKey && block->tep->blackboardData->flags.receivedUnsecure)
        {           /* New Session */
            static const TEP_V1_ReceiveUtilFunction newSessionFunctions[] = {
                    TEP_V1_Receive_ReadDomain,
                    TEP_V1_Receive_ReadInitiatorBlock,
                    NULL
            };

            if(!TEP_V1_Receive_ProcessFunctions(block, newSessionFunctions))
            {
                uint8 rejectCode;

                switch(block->status)
                {
                case TEP_V1_ReceiveStatus_STREAM_ERROR:
                    rejectCode = TEP_V1_ERROR_PARSEERROR;
                    break;

                case TEP_V1_ReceiveStatus_APP_ERROR:
                    rejectCode = TEP_V1_ERROR_UNKNOWNDOMAIN;
                    break;

                case TEP_V1_ReceiveStatus_MEMORY_ERROR:
                    rejectCode = TEP_V1_ERROR_INSUFFICIENT_RESOURCES;
                    break;

                default:
                    /* Why don't we know a better error? */
                    rejectCode = TEP_V1_ERROR_INSUFFICIENT_RESOURCES;
                    break;
                }
                DOFObjectIDDomain_Destroy(block->domainId);
                if(block->initiatorBlock != NULL)
                {
                    OALSessionInitiatorBlock_Release(block->initiatorBlock);
                }
                block->operation =
                    DOFOperation_Create_Reject(block->tep->blackboardData->associatedChannel,
                                               block->tep->blackboardData->transportAddr,
                                               block->tep->blackboardData->opID,
                                               block->tep->isConnectionLess, rejectCode);
            }
            else
            {
                if(block->operation == NULL)
                {
                    block->operation =
                        DOFOperation_Create_BeginSession(block->tep->blackboardData->associatedChannel,
                                                         block->tep->blackboardData->transportAddr,
                                                         block->tep->blackboardData->delay,
                                                         block->initiatorBlock, block->iBlock,
                                                         block->tep, block->tep->blackboardData->opID,
                                                         block->tep->isConnectionLess);
                    if( block->operation != NULL )
                    {
                        OALSessionInitiatorBlock_Release(block->initiatorBlock); 
                        OALStream_Release(block->iBlock, TRUE);
                        block->context = (TEP_V1_OperationContext) block->operation->tepContext;
                        block->context->domain = block->domainId;
                        block->domainId = NULL;
                    }
                }
                else
                {
                    OALStream oldStream;
                    OALSessionInitiatorBlock oldBlock;

                    DOFOperation_Lock(block->operation);
                    oldStream = block->context->iBlock;
                    oldBlock = block->context->initiatorBlock;
                    block->context->iBlock = block->iBlock;
                    OALStream_Release(oldStream, TRUE);
                    block->context->initiatorBlock = block->initiatorBlock;
                    OALSessionInitiatorBlock_Release(oldBlock);
                    DOFOperation_Unlock(block->operation);
                }
                DOFObjectIDDomain_Destroy(block->domainId);
                if(block->operation == NULL)
                {
                    block->status = TEP_V1_ReceiveStatus_MEMORY_ERROR;
                }
                else
                {
                    if(OALOperationManager_AddOperation(block->operation,
                                                        block->tep->blackboardData->opID,
                                                        FALSE) != DOFERROR_OK)
                    {
                        block->status = TEP_V1_ReceiveStatus_MEMORY_ERROR;
                    }
                }
            }
        }
        else if(!isConfirm)
        {
            static const TEP_V1_ReceiveUtilFunction extendSessionFunctions[] = {
                    TEP_V1_Receive_LookupSession,
                    TEP_V1_Receive_ReadInitiatorBlock,
                    NULL
            };

            if(!TEP_V1_Receive_ProcessFunctions(block, extendSessionFunctions))
            {
                uint8 rejectCode;

                switch(block->status)
                {
                case TEP_V1_ReceiveStatus_STREAM_ERROR:
                    rejectCode = TEP_V1_ERROR_PARSEERROR;
                    break;

                case TEP_V1_ReceiveStatus_MEMORY_ERROR:
                    rejectCode = TEP_V1_ERROR_INSUFFICIENT_RESOURCES;
                    break;

                default:
                    /* Why don't we know a better error? */
                    rejectCode = TEP_V1_ERROR_INSUFFICIENT_RESOURCES;
                    break;
                }
                DOFObjectIDDomain_Destroy(block->domainId);
                if(block->initiatorBlock != NULL)
                {
                    OALSessionInitiatorBlock_Release(block->initiatorBlock);
                }
                if(block->operation != NULL)
                {
                    DOFOperation_Internal_Destroy(block->operation);
                }
                block->operation =
                    DOFOperation_Create_Reject(block->tep->blackboardData->associatedChannel,
                                               block->tep->blackboardData->transportAddr,
                                               block->tep->blackboardData->opID,
                                               block->tep->isConnectionLess, rejectCode);
            }
            else
            {
                if(!block->tep->blackboardData->flags.receivedUnsecure)
                {
                    if(!DOFObjectIDAuthentication_IsEqual(DOFCredentials_GetIdentity(TEP_V1_Session_GetOtherCredentials(block->sessionState)),
                                                          DOFCredentials_GetIdentity(block->initiatorBlock->keyRequest->credentials)))
                    {
                        block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
                        return;
                    }
                }
                if(useKey)
                {
                    if(DOFSecurityMode_IsEqual(block->initiatorBlock->modeOfOperation,
                                               block->tep->blackboardData->associatedChannel->remoteChannelData->mode))
                    {
                        block->operation = DOFOperation_Create_TepRekey_Internal(block->tep->blackboardData->associatedChannel,
                                                                                 block->tep->blackboardData->transportAddr,
                                                                                 block->iBlock, block->initiatorBlock,
                                                                                 block->tep->blackboardData->opID,
                                                                                 block->tep, block->tep->blackboardData->delay);
                        if(OALOperationManager_AddOperation(block->operation, block->tep->blackboardData->opID, FALSE) != DOFERROR_OK)
                        {
                            block->status = TEP_V1_ReceiveStatus_MEMORY_ERROR;
                        }
                        else
                        {
                            block->tep->blackboardData->associatedChannel->remoteChannelData->flags.tepRekeying = TRUE;
                        }
                    }
                    else
                    {
                        TEP_V1_OperationContext newContext;

                        DOFOperation_Internal_Destroy(block->operation);

                        block->operation =
                            DOFOperation_Create_Reject(block->tep->blackboardData->associatedChannel,
                                                       block->tep->blackboardData->transportAddr,
                                                       block->tep->blackboardData->opID,
                                                       block->tep->isConnectionLess, TEP_V1_ERROR_UNACCEPTABLEMODE);
                        newContext = (TEP_V1_OperationContext) block->operation->tepContext;
                        if(newContext != NULL)
                        {
                            newContext->rejectData = OALStream_CreateRW(8);
                            if(newContext->rejectData != NULL)
                            {
                                ESPV2_MarshalCCMReject(newContext->rejectData);
                            }
                        }
                        OALSessionInitiatorBlock_Release(block->initiatorBlock);
                        OALStream_Release(block->iBlock, TRUE);
                    }

                }
                else
                {
                    block->creds = block->tep->blackboardData->associatedChannel->remoteChannelData->secureCredentials;
                    if(block->initiatorBlock->duration != 0)
                    {
                        DOFOperation_Internal_Destroy(block->operation);
                        block->operation =
                            DOFOperation_Create_Reject(block->tep->blackboardData->associatedChannel,
                                                       block->tep->blackboardData->transportAddr,
                                                       block->tep->blackboardData->opID,
                                                       block->tep->isConnectionLess,
                                                       TEP_V1_ERROR_PARSEERROR);
                        block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
                    }
                    else if(!DOFSecurityMode_IsEqual(block->initiatorBlock->modeOfOperation,
                                                     block->tep->blackboardData->associatedChannel->remoteChannelData->mode))
                    {
                        DOFOperation_Internal_Destroy(block->operation);
                        block->operation =
                            DOFOperation_Create_Reject(block->tep->blackboardData->associatedChannel,
                                                       block->tep->blackboardData->transportAddr,
                                                       block->tep->blackboardData->opID,
                                                       block->tep->isConnectionLess,
                                                       TEP_V1_ERROR_PARSEERROR);
                        block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
                    }
                    else
                    {
                        if(block->operation == NULL)
                        {
                            block->operation =
                                DOFOperation_Create_ExtendPermission(block->tep->blackboardData->associatedChannel,
                                                                     block->tep->blackboardData->transportAddr,
                                                                     block->tep->blackboardData->delay,
                                                                     block->initiatorBlock, block->tep->blackboardData->opID,
                                                                     NULL, NULL, block->iBlock, block->tep->blackboardData->keyState,
                                                                     block->tep->isConnectionLess);
                            OALSessionInitiatorBlock_Release(block->initiatorBlock);
                            OALStream_Release(block->iBlock, TRUE);
                            if(block->operation != NULL)
                            {
                                if(OALOperationManager_AddOperation(block->operation,
                                                                    block->tep->blackboardData->opID,
                                                                    FALSE) != DOFERROR_OK)
                                {
                                    block->status = TEP_V1_ReceiveStatus_MEMORY_ERROR;
                                }
                            }
                            else
                            {
                                block->status = TEP_V1_ReceiveStatus_MEMORY_ERROR;
                            }
                        }
                        else
                        {
                            DOFOperation_Lock(block->operation);
                            OALSessionInitiatorBlock_Release(block->context->initiatorBlock);
                            block->context->initiatorBlock = block->initiatorBlock;
                            OALStream_Release(block->context->iBlock, TRUE);
                            block->context->iBlock = block->iBlock;
                            DOFOperation_Unlock(block->operation);
                        }
                    }
                }
            }
        }
        else if(TEP_V1_Receive_VerifyOperation(block))
        {
            static const TEP_V1_ReceiveUtilFunction confirmSessionFunctions[] = {
                    TEP_V1_Receive_LookupSession,
                    TEP_V1_Receive_ReadInitiatorConfirmation,
                    TEP_V1_Receive_VerifyInitiatorConfirmation,
                    NULL
            };

            uint8 index = (block->context->flags.extendingPermission) ? 1 : 0;

            if(TEP_V1_Receive_ProcessFunctions(block, &(confirmSessionFunctions[index])))
            {
                DOFOperation_Lock(block->operation);
                PCRMem_Copy(block->context->initiatorConfirmation, block->initiatorConfirmation, 32);
                block->operation->operation_type = DOFOPERATIONTYPE_TEP_CONFIRM;
                if(!block->context->flags.extendingPermission)
                {
                    TEP_SessionEvent_t eventType;

                    void * newContext;

                    if(block->context->flags.rekeying)
                    {
                        if(block->tep->isConnectionLess)
                        {
                            eventType = TEP_V1_SessionEvent_REKEY_SENDKEY;
                        }
                        else
                        {
                            eventType = TEP_V1_SessionEvent_REKEY_SENDKEYANDDISABLE;
                        }
                    }
                    else
                    {
                        eventType = TEP_V1_SessionEvent_NEWSESSION_SUCCESS_SENDKEY;
                    }
                    DOFOperation_Unlock(block->operation);
                    newContext =
                        block->context->tep->sessionEventCallback(TEP_V1_Session_GetContext(block->sessionState),
                                                                  block->context->sessionkey, eventType,
                                                                  block->context->initiatorBlock->modeOfOperation,
                                                                  block->operation->sources->channel,
                                                                  block->context->connectionSecurityScope,
                                                                  TEPOperation_GetResponderPermissions(block->operation),
                                                                  TEPOperation_GetInitiatorPermissions(block->operation),
                                                                  NULL, block->context->stateIdentifier,
                                                                  block->operation->sources->address, NULL,
                                                                  block->operation->sources->channel->remoteChannelData->secureCredentials);
                    TEP_V1_Session_SetContext(block->sessionState, newContext);
                }
                else
                {
                    OALCipher cipher = block->operation->sources->channel->remoteChannelData->cipher;

                    if(block->tep->isConnectionLess)
                    {
                        if(!DOFPermissionSet_AddPermissionsFromSet(OALCipher_GetLocalPerms( cipher, block->context->keyState ),
                                                                   TEPOperation_GetResponderPermissions(block->operation)))
                        {
                            block->status = TEP_V1_ReceiveStatus_CLOSE_CONNECTION;
                        }
                    }
                    if(!DOFPermissionSet_AddPermissionsFromSet(OALCipher_GetRemotePerms( cipher, block->context->keyState ),
                                                               TEPOperation_GetInitiatorPermissions(block->operation)))
                    {
                        block->status = TEP_V1_ReceiveStatus_CLOSE_CONNECTION;
                    }
                    DOFOperation_Unlock(block->operation);
                }
                if(!block->context->flags.rekeying && !block->context->flags.extendingPermission &&
                   !block->tep->blackboardData->associatedChannel->remoteChannelData->flags.mustFinalizeTep)
                {
                    void * newContext =
                        block->tep->sessionEventCallback(TEP_V1_Session_GetContext(block->sessionState),
                                                         block->context->sessionkey,
                                                         TEP_V1_SessionEvent_NEWSESSION_FINALIZE,
                                                         block->context->initiatorBlock->modeOfOperation,
                                                         block->operation->sources->channel,
                                                         block->context->connectionSecurityScope,
                                                         TEPOperation_GetResponderPermissions(block->operation),
                                                         TEPOperation_GetInitiatorPermissions(block->operation),
                                                         NULL, block->context->stateIdentifier,
                                                         block->operation->sources->address, NULL,
                                                         block->operation->sources->channel->remoteChannelData->secureCredentials);
                    TEP_V1_Session_SetContext(block->sessionState, newContext);
                    if(block->operation->sources->channel->remoteChannelData->secureSid != NULL)
                    {
                        source_release(block->operation->sources->channel->remoteChannelData->secureSid);
                    }
                    block->operation->sources->channel->remoteChannelData->secureSid = block->context->secureSid;
                    source_retain(block->context->secureSid);
                    OALConnection_SetPeerCredentials(block->tep->blackboardData->associatedChannel->remoteChannelData->OALConn, block->context->initiatorBlock->keyRequest->credentials);
                    block->tep->blackboardData->associatedChannel->remoteChannelData->flags.mustFinalizeTep = TRUE;
                }
                block->tep->blackboardData->associatedChannel->remoteChannelData->flags.tepRekeying = FALSE;
            }
            else /* Error case */
            {
                uint8 rejectCode;

                switch(block->status)
                {
                case TEP_V1_ReceiveStatus_STREAM_ERROR:
                    rejectCode = TEP_V1_ERROR_PARSEERROR;
                    break;

                case TEP_V1_ReceiveStatus_APP_ERROR:
                    rejectCode = TEP_V1_ERROR_UNKNOWNDOMAIN;
                    break;

                case TEP_V1_ReceiveStatus_MEMORY_ERROR:
                    rejectCode = TEP_V1_ERROR_INSUFFICIENT_RESOURCES;
                    break;

                default:
                    /* Why don't we know a better error? */
                    rejectCode = TEP_V1_ERROR_INSUFFICIENT_RESOURCES;
                    break;
                }
                DOFObjectIDDomain_Destroy(block->domainId);
                if(block->initiatorBlock != NULL)
                {
                    OALSessionInitiatorBlock_Release(block->initiatorBlock);
                }
                if(block->operation != NULL)
                {
                    DOFOperation_Internal_Destroy(block->operation);
                }
                block->operation =
                    DOFOperation_Create_Reject(block->tep->blackboardData->associatedChannel,
                                               block->tep->blackboardData->transportAddr,
                                               block->tep->blackboardData->opID,
                                               block->tep->isConnectionLess, rejectCode);
            }
        }
    }
}

static void TEP_V1_Receive_NegotiateSession_r(TEP_V1_ReceiveBlock block)
{
    boolean isConfirm;

    PCR_Assert(block != NULL);
    if(block->context == NULL)
    {
        block->status = TEP_V1_ReceiveStatus_MEMORY_ERROR;
    }
    else
    {
        isConfirm  = (block->firstByte & 0x20) >> 5;
        if(isConfirm)
        {
            static const TEP_V1_ReceiveUtilFunction confirmFunctions[] = {
                TEP_V1_Receive_LookupSession,
                TEP_V1_Receive_VerifyOperation,
                NULL
            };

            if(TEP_V1_Receive_ProcessFunctions(block, (block->tep->isConnectionLess) ? &confirmFunctions[0]
                                                                                     : &confirmFunctions[1]))
            {
                DOFOperation_Lock(block->operation);
                block->operation->operation_type = DOFOPERATIONTYPE_TEP_CONFIRMRESPONSE;
                if(block->context->flags.extendingPermission && block->tep->isConnectionLess)
                {
                    OALCipher cipher = block->operation->sources->channel->remoteChannelData->cipher;

                    if(!DOFPermissionSet_AddPermissionsFromSet(OALCipher_GetLocalPerms( cipher, block->context->keyState ),
                                                               TEPOperation_GetInitiatorPermissions(block->operation)))
                    {
                        DOFOperation_Unlock(block->operation);
                        block->status = TEP_V1_ReceiveStatus_CLOSE_CONNECTION;
                        return;
                    }
                    DOFOperation_Unlock(block->operation);
                    if(!block->context->calledCallback && (block->context->extendCompleteCallback != NULL))
                    {
                        block->context->extendCompleteCallback(DOFERROR_OK, block->context->extendContext);
                        block->context->calledCallback = TRUE;
                    }
                }
                else
                {
                    if(block->tep->isConnectionLess)
                    {
                        TEP_SessionEvent_t eventType = TEP_V1_SessionEvent_REKEY_SENDKEY;

                        void * newContext =
                            block->context->tep->sessionEventCallback(TEP_V1_Session_GetContext(block->sessionState),
                                                                      block->context->sessionkey, eventType,
                                                                      block->context->initiatorBlock->modeOfOperation,
                                                                      block->operation->sources->channel,
                                                                      block->context->connectionSecurityScope,
                                                                      TEPOperation_GetInitiatorPermissions(block->operation),
                                                                      TEPOperation_GetResponderPermissions(block->operation),
                                                                      NULL, block->context->stateIdentifier,
                                                                      block->operation->sources->address, NULL,
                                                                      block->operation->sources->channel->remoteChannelData->secureCredentials);
                        TEP_V1_Session_SetContext(block->sessionState, newContext);
                    }
                    block->tep->blackboardData->associatedChannel->remoteChannelData->flags.tepRekeying = FALSE;
                    DOFOperation_Unlock(block->operation);
                }
            }
        }
        else
        {
            static const TEP_V1_ReceiveUtilFunction sessionFunctions[] = {
                    TEP_V1_Receive_VerifyOperation,
                    TEP_V1_Receive_LookupSession,
                    TEP_V1_Receive_ReadInitiatorTicket,
                    TEP_V1_Receive_ReadTicketConfirmation,
                    TEP_V1_Receive_ReadResponderInitialization,
                    TEP_V1_Receive_ReadResponderBlock,
                    TEP_V1_Receive_ReadAuthenticatorInitialization,
                    NULL
            };

            static const TEP_V1_ReceiveUtilFunction extendFunctions[] = {
                    TEP_V1_Receive_VerifyOperation,
                    TEP_V1_Receive_LookupSession,
                    TEP_V1_Receive_ReadInitiatorTicket,
                    TEP_V1_Receive_ReadTicketConfirmation,
                    TEP_V1_Receive_ReadResponderBlock,
                    TEP_V1_Receive_ReadAuthenticatorInitialization,
                    NULL
            };

            if(TEP_V1_Receive_ProcessFunctions(block, (block->context->flags.extendingPermission) ? extendFunctions
                                                                                                  : sessionFunctions))
            {
                if(!block->tep->blackboardData->flags.receivedUnsecure)
                {
                    if(!DOFObjectIDAuthentication_IsEqual(DOFCredentials_GetIdentity(TEP_V1_Session_GetOtherCredentials(block->sessionState)),
                                                          DOFCredentials_GetIdentity(block->responderBlock->credentials)))
                    {
                        /* This sets the fields created in the tep operation context so they can be freed in the standard tep way.
                           Note: These could be freed here to save us the locking of the operation. */
                        DOFOperation_Lock(block->operation);
                        block->context->initGrant = block->initiatorTicket;
                        block->context->initialState = block->initialState;
                        block->context->connectionSecurityScope = block->connectionSecurityScope;
                        block->context->keyRequest = block->responderBlock;
                        block->context->rBlock = block->rBlock;
                        block->context->aBlock = block->aBlock;
                        DOFOperation_Unlock(block->operation);

                        block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
                        return;
                    }
                }
                DOFOperation_Lock(block->operation);
                block->context->grantedDuration = block->grantedDuration;
                block->context->initGrant = block->initiatorTicket;
                PCRMem_Copy(block->context->responderConfirmation, block->ticketConfirmation, 32);
                block->context->initialState = block->initialState;
                block->context->connectionSecurityScope = block->connectionSecurityScope;
                block->context->keyRequest = block->responderBlock;
                block->operation->operation_type = DOFOPERATIONTYPE_TEP_AUTHORIZE;
                block->context->rBlock = block->rBlock;
                block->context->aBlock = block->aBlock;
                DOFOperation_Unlock(block->operation);
                if(!block->context->flags.extendingPermission)
                {
                    if(!block->context->flags.rekeying)
                    {
                        block->context->stateIdentifier = block->stateIdentifier;
                    }
                    TEP_V1_SM_HandleAuthorize(block->operation);
                    block->status = TEP_V1_ReceiveStatus_DONOTHING;
                }
            }
        }
    }
}

static void TEP_V1_Receive_EndSession(TEP_V1_ReceiveBlock block)
{
    static const TEP_V1_ReceiveUtilFunction functions[] = {
            TEP_V1_Receive_LookupSession,
            NULL
    };

    PCR_Assert(block != NULL);
    if(!block->tep->blackboardData->flags.receivedAuthenticated)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_INFO, "TEP V1: Rejecting EndSession. (not authenticated)\n");
        block->status = TEP_V1_ReceiveStatus_CLOSE_CONNECTION;
        return;
    }
    else if(block->tep->blackboardData->flags.isMulticast)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_INFO, "TEP V1: Rejecting EndSession. (multicast not allowed)\n");
        block->status = TEP_V1_ReceiveStatus_CLOSE_CONNECTION;
        return;
    }
    else if(TEP_V1_Receive_ProcessFunctions(block, functions))
    {
        block->tep->blackboardData->associatedChannel->remoteChannelData->flags.mustFinalizeTep = FALSE;
        block->tep->sessionEventCallback(TEP_V1_Session_GetContext(block->sessionState), NULL,
                                         TEP_V1_SessionEvent_ENDING, NULL, block->tep->blackboardData->associatedChannel,
                                           NULL, NULL, NULL, NULL, 0, block->tep->blackboardData->transportAddr,
                                           NULL, block->tep->blackboardData->associatedChannel->remoteChannelData->secureCredentials);
    }
}

static void TEP_V1_Receive_Reject_r(TEP_V1_ReceiveBlock block)
{
    static const TEP_V1_ReceiveUtilFunction functions[] = {
            TEP_V1_Receive_VerifyOperation,
            TEP_V1_Receive_ReadErrorCode,
            NULL
    };

    PCR_Assert(block != NULL);
    if(TEP_V1_Receive_ProcessFunctions(block, functions))
    {
        DOFOperation_Lock(block->operation);
        if((block->operation->operation_type == DOFOPERATIONTYPE_TEP_BEGIN) && (block->rejectErrorCode == TEP_V1_ERROR_UNACCEPTABLEMODE))
        {
            if(block->rejectMode != NULL)
            {
                DOFOperation oldOperation = block->operation;

                block->operation = DOFOperation_Create_BeginSession(block->tep->blackboardData->associatedChannel,
                                                                    block->tep->blackboardData->transportAddr,
                                                                    TEP_V1_TIMEOUT, block->context->initiatorBlock, NULL,
                                                                    block->tep, NULL, block->tep->isConnectionLess);

                if(block->operation != NULL)
                {
                    DOFSecurityMode oldMode;

                    block->context = (TEP_V1_OperationContext) block->operation->tepContext;
                    oldMode = block->context->initiatorBlock->modeOfOperation;
                    block->context->initiatorBlock->modeOfOperation = block->rejectMode;
                    DOFSecurityMode_Destroy(oldMode);
                    if(OALOperationManager_AddOperation(block->operation, block->operation->opID, FALSE) != DOFERROR_OK)
                    {
                        block->status = TEP_V1_ReceiveStatus_DONOTHING;
                    }
                }
                oldOperation->operation_type = DOFOPERATIONTYPE_TEP_REJECT;
                DOFOperation_Unlock(oldOperation);
                DOFOperation_Internal_Destroy(oldOperation);
                return;
            }
            DOFSecurityMode_Destroy(block->rejectMode);
        }
        block->context->rejectCode = block->rejectErrorCode;
        block->operation->operation_type = DOFOPERATIONTYPE_TEP_REJECT;
        DOFOperation_Unlock(block->operation);
    }
}

static void TEP_V1_Receive_SessionEnding(TEP_V1_ReceiveBlock block)
{
    static const TEP_V1_ReceiveUtilFunction functions[] = {
            TEP_V1_Receive_LookupSession,
            NULL
    };

    PCR_Assert(block != NULL);
    if(!block->tep->blackboardData->flags.receivedAuthenticated)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_INFO, "TEP V1: Rejecting EndSession. (not authenticated)\n");
        block->status = TEP_V1_ReceiveStatus_CLOSE_CONNECTION;
        return;
    }
    else if(block->tep->blackboardData->flags.isMulticast)
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_INFO, "TEP V1: Rejecting EndSession. (multicast not allowed)\n");
        block->status = TEP_V1_ReceiveStatus_CLOSE_CONNECTION;
        return;
    }
    else if(TEP_V1_Receive_ProcessFunctions(block, functions))
    {
        TEP_V1_Session_BeginRekey(block->sessionState);
        block->status = TEP_V1_ReceiveStatus_DONOTHING;
    }
}

static void TEP_V1_Receive_InvalidOpcode(TEP_V1_ReceiveBlock block)
{
    PCR_Assert(block != NULL);
    block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
}
