/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/error.h>
#include <internal/esp_v2_options.h>
#include <internal/credentialsinternal.h>
#include <internal/marshal.h>
#include <internal/objectidinternal.h>
#include <internal/remotechannel.h>
#include <internal/serverinternal.h>

#include "tep_v1_receiveutil.h"
#include "tep_v1_sm.h"

boolean TEP_V1_Receive_ProcessFunctions(      TEP_V1_ReceiveBlock          block,
                                        const TEP_V1_ReceiveUtilFunction * functions)
{
    uint8 i = 0;

    PCR_Assert(block != NULL);
    PCR_Assert(functions != NULL);
    while(functions[i] != NULL)
    {
        if(!functions[i](block))
        {
            PCR_Assert(block->status != TEP_V1_ReceiveStatus_OK);   /* Double check return */
            return FALSE;
        }
        PCR_Assert(block->status == TEP_V1_ReceiveStatus_OK);       /* Double check return */
        ++i;
    }
    return TRUE;
}

boolean TEP_V1_Receive_LookupSession(TEP_V1_ReceiveBlock block)
{
    PCR_Assert(block != NULL);

    block->sessionState = TEP_V1_SessionLookup(block->tep->blackboardData->associatedChannel);
    if(block->sessionState == NULL)
    {
        block->status = TEP_V1_ReceiveStatus_MEMORY_ERROR;
        return FALSE;
    }
    return TRUE;
}

boolean TEP_V1_Receive_VerifyOperation(TEP_V1_ReceiveBlock block)
{
    PCR_Assert(block != NULL);

    if((block->operation == NULL) ||
       (block->operation->operation_type < DOFOPERATIONTYPE_TEP_BEGIN) ||
       (block->operation->operation_type > DOFOPERATIONTYPE_TEP_LAST))
    {
        block->status = TEP_V1_ReceiveStatus_MEMORY_ERROR;
        return FALSE;
    }
    return TRUE;
}

boolean TEP_V1_Receive_ReadDomain(TEP_V1_ReceiveBlock block)
{
    DOFObjectID oid;

    PCR_Assert(block != NULL);

    oid = DOFObjectID_Unmarshal(block->stream);
    if(oid == NULL)
    {
        block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
        return FALSE;
    }
    else
    {
        boolean ok = FALSE;

        if(DOFObjectID_IsBroadcast(oid))
        {
            block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
        }
        else
        {
            block->domainId = DOFObjectIDDomain_Create(oid);
            if(block->domainId == NULL)
            {
                block->status = TEP_V1_ReceiveStatus_MEMORY_ERROR;
                ok = FALSE;
            }
            else
            {
                ok = TRUE;
            }
        }
        DOFObjectID_Destroy(oid);
        return ok;
    }
}

boolean TEP_V1_Receive_ReadInitiatorBlock(TEP_V1_ReceiveBlock block)
{
    uint32 streamOffset;

    uint8 * streamPtr;

    PCR_Assert(block != NULL);

    streamPtr = (uint8*) OALStream_GetReadPtr(block->stream);
    streamOffset = OALStream_GetReadOffset(block->stream);
    block->initiatorBlock = OALSessionInitiatorBlock_Unmarshal(block->stream, block->domainId, TRUE, FALSE);
    if(block->initiatorBlock == NULL)
    {
        block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
    }
    else
    {
        /* This will be true if we've already keyed, false otherwise */
        /* boolean isRekey = (block->tep->blackboardData->associatedChannel->remoteChannelData->cipher != NULL); */

        uint32 blockSize = OALStream_GetReadOffset(block->stream) - streamOffset;

        block->iBlock = OALStream_CreateRW(blockSize);
        if(block->iBlock != NULL)
        {
            OALStream_Write(block->iBlock, blockSize, streamPtr);
        }
        else
        {
            block->status = TEP_V1_ReceiveStatus_MEMORY_ERROR;
            return FALSE;
        }
    }
    /* DOFObjectIDDomain_Destroy(block->domainId); - Handled elsewhere */
    return (block->initiatorBlock != NULL);
}

boolean TEP_V1_Receive_ReadInitiatorTicket(TEP_V1_ReceiveBlock block)
{
    PCR_Assert(block != NULL);

    block->initiatorTicket = OALKeyGrant_Unmarshal(block->stream);
    if(block->initiatorTicket == NULL)
    {
        block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
    }
    return (block->initiatorTicket != NULL);
}

boolean TEP_V1_Receive_ReadTicketConfirmation(TEP_V1_ReceiveBlock block)
{
    PCR_Assert(block != NULL);

    if(!OALStream_Read(block->stream, 32, block->ticketConfirmation))
    {
        block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
        return FALSE;
    }
    return TRUE;
}

boolean TEP_V1_Receive_ReadResponderInitialization(TEP_V1_ReceiveBlock block)
{
    PCR_Assert(block != NULL);

    if(block->tep->blackboardData->flags.receivedUnsecure && OALMarshal_StreamUncompress4(block->stream, &(block->stateIdentifier)) != DOFERROR_OK)
    {
        block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
        return FALSE;
    }
    block->initialState = OALSecurityMode_Unmarshal(block->stream);
    if(block->initialState == NULL)
    {
        block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
    }
    return (block->initialState != NULL);
}

boolean TEP_V1_Receive_ReadResponderBlock(TEP_V1_ReceiveBlock block)
{
    uint8 * streamPtr;

    uint32 streamOffset = OALStream_GetReadOffset(block->stream);

    PCR_Assert(block != NULL);

    streamPtr = (uint8*) OALStream_GetReadPtr(block->stream);
    block->responderBlock = (DOFKeyRequest)DOFMem_Alloc( sizeof( DOFKeyRequest_t ) );
    if(block->responderBlock != NULL)
    {
        PCRMem_Set(block->responderBlock, 0, sizeof(DOFKeyRequest_t));
        if(!DOFKeyRequest_Unmarshal(block->stream, block->responderBlock, DOFCredentials_GetDomainID( block->operation->sources->channel->remoteChannelData->secureCredentials ), FALSE, FALSE))
        {
           block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
           return FALSE;
        }
        else
        {
            uint32 blockSize = OALStream_GetReadOffset(block->stream) - streamOffset;

            block->rBlock = OALStream_CreateRW(blockSize);
            if(block->rBlock != NULL)
            {
                OALStream_Write(block->rBlock, blockSize, streamPtr);
            }
            else
            {
                block->status = TEP_V1_ReceiveStatus_MEMORY_ERROR;
                return FALSE;
            }
            return TRUE;
        }
    }
    else
    {
        block->status = TEP_V1_ReceiveStatus_MEMORY_ERROR;
        return FALSE;
    }
}

boolean TEP_V1_Receive_ReadAuthenticatorInitialization(TEP_V1_ReceiveBlock block)
{
    uint32 streamOffset;

    uint8 * streamPtr;

    PCR_Assert(block != NULL);

    streamPtr = (uint8*) OALStream_GetReadPtr(block->stream);
    streamOffset = OALStream_GetReadOffset(block->stream);
    if(OALSessionAuthRespBlock_Unmarshal(&(block->grantedDuration), &(block->connectionSecurityScope),
                                         block->context->initiatorBlock->keyRequest->permission_set,
                                         block->responderBlock->permission_set,
                                         DOFCredentials_GetDomainID(block->operation->sources->channel->remoteChannelData->secureCredentials),
                                         block->stream))
    {
        uint32 blockSize = OALStream_GetReadOffset(block->stream) - streamOffset;

        block->aBlock = OALStream_CreateRW(blockSize);
        if(block->aBlock != NULL)
        {
            OALStream_Write(block->aBlock, blockSize, streamPtr);
        }
        else
        {
            block->status = TEP_V1_ReceiveStatus_MEMORY_ERROR;
            return FALSE;
        }
        return TRUE;
    }
    else
    {
        block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
        return FALSE;
    }
}

boolean TEP_V1_Receive_ReadInitiatorConfirmation(TEP_V1_ReceiveBlock block)
{
    PCR_Assert(block != NULL);

    if(!OALStream_Read(block->stream, 32, block->initiatorConfirmation))
    {
        block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
        return FALSE;
    }
    return TRUE;
}

boolean TEP_V1_Receive_VerifyInitiatorConfirmation(TEP_V1_ReceiveBlock block)
{
    uint8 confirmBlock[32];

    PCR_Assert(block != NULL);
    if(block->context->flags.keyRequested) /* Error to build ticket comfirmation befor we have our keyRequest */
    {
        TEP_V1_SM_BuildTicketConfirmation(block->operation, FALSE, confirmBlock); 
        if(PCRMem_Compare(confirmBlock, block->initiatorConfirmation, 32) != 0)
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR,
                           "TEP_V1: Calculated and sent initiator confirmations do not match");
            block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
            return FALSE;
        }
    }
    else
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "TEP_V1: Tried to build comfirmation without having a keyRequest.");
        block->status = TEP_V1_ReceiveStatus_MEMORY_ERROR;
        return FALSE;
    }
    return TRUE;
}

boolean TEP_V1_Receive_ReadErrorCode(TEP_V1_ReceiveBlock block)
{
    PCR_Assert(block != NULL);

    if(!OALStream_Read8(block->stream, &(block->rejectErrorCode)))
    {
        block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
        return FALSE;
    }
    if(block->rejectErrorCode == TEP_V1_ERROR_UNACCEPTABLEMODE)
    {
        boolean streamError;

        if(ESPV2_CheckRejectOrNakIfNakReturnASecurityMode(block->stream,
                                                          block->tep->blackboardData->associatedChannel->remoteChannelData->config.connection->securityDesire,
                                                          &(block->rejectMode), &streamError))
        /* Negative Acknowledgement and no stream issue */
        {
            PCR_Assert(streamError == FALSE);
        }
        else if(streamError)
        {
            block->status = TEP_V1_ReceiveStatus_STREAM_ERROR;
            return FALSE;
        }
    }
    return TRUE;
}
