/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/marshal.h>
#include <internal/operationinternal.h>
#include <internal/permissionsetinternal.h>
#include <internal/tep_v1.h>
#include <internal/trp_v1.h>

#include "tep_v1_sendblock.h"
#include "tep_v1_sendutil.h"
#include "tep_v1_sm.h"

typedef void (*TEP_V1_SendFunction)(TEP_V1_SendBlock block);

static void TEP_V1_Send_NegotiateSession(TEP_V1_SendBlock block);

static void TEP_V1_Send_NegotiateSession_r(TEP_V1_SendBlock block);

static void TEP_V1_Send_EndSession(TEP_V1_SendBlock block);

static void TEP_V1_Send_Reject_r(TEP_V1_SendBlock block);

static void TEP_V1_Send_SessionEnding(TEP_V1_SendBlock block);

static void TEP_V1_Send_NoOp(TEP_V1_SendBlock block);

static void TEP_V1_Send_NegotiateNoOp(TEP_V1_SendBlock block);

ProtocolLayerResult TEP_V1_Send(TEP_V1       self,
                                DOFOperation operation,
                                boolean      isProviderOperation,
                                DOFResult    result)
{
    TEP_V1_SendFunction functions[DOFOPERATIONTYPE_TEP_LAST - DOFOPERATIONTYPE_TEP_BEGIN] = {
            TEP_V1_Send_NegotiateSession,   /* DOFOPERATIONTYPE_TEP_BEGIN */
            TEP_V1_Send_NegotiateSession_r, /* DOFOPERATIONTYPE_TEP_AUTHORIZE */
            TEP_V1_Send_NegotiateSession,   /* DOFOPERATIONTYPE_TEP_CONFIRM */
            TEP_V1_Send_NegotiateSession_r, /* DOFOPERATIONTYPE_TEP_CONFIRMRESPONSE */
            TEP_V1_Send_EndSession,         /* DOFOPERATIONTYPE_TEP_END */
            TEP_V1_Send_NegotiateSession,   /* DOFOPERATIONTYPE_TEP_EXTEND */
            TEP_V1_Send_NegotiateSession_r, /* DOFOPERATIONTYPE_TEP_EXTENDAUTH */
            TEP_V1_Send_Reject_r,           /* DOFOPERATIONTYPE_TEP_REJECT */
            TEP_V1_Send_SessionEnding,      /* DOFOPERATIONTYPE_TEP_SESSIONENDING */
            TEP_V1_Send_NoOp,               /* DOFOPERATIONTYPE_TEP_NOOP */
            TEP_V1_Send_NegotiateNoOp       /* DOFOPERATIONTYPE_TEP_NEGOTIATION_NOOP */
    };

    TEP_V1_SendBlock_t sendBlock;

    PCR_Assert(self != NULL);
    PCR_Assert(operation != NULL);
    PCR_Assert(operation->operation_type >= DOFOPERATIONTYPE_TEP_BEGIN);
    PCR_Assert(operation->operation_type < DOFOPERATIONTYPE_TEP_LAST);
    TEP_V1_SendBlock_Init(&sendBlock, self, self->blackboardData->sendPacket, operation);
    functions[operation->operation_type - DOFOPERATIONTYPE_TEP_BEGIN](&sendBlock);
    switch(sendBlock.status)
    {
    case TEP_V1_SendStatus_OK:
        self->blackboardData->opID = DOFOperationID_Copy(operation->opID);

    case TEP_V1_SendStatus_OKNOID:
        if(operation->operation_type == DOFOPERATIONTYPE_TEP_CONFIRM)
        {
            self->blackboardData->retry = operation->sources->retry_count + 1;
        }
        else
        {
            self->blackboardData->retry = operation->sources->retry_count;
        }
        if(!self->blackboardData->associatedChannel->remoteChannelData->flags.sessionIsSecure &&
            self->blackboardData->flags.isCommandNotResponse)
        {
            self->blackboardData->seq = 1;
        }
        if(sendBlock.state != NULL)
        {
            TEP_V1_Session_Destroy(sendBlock.state);
        }
        return PLRESULT_OKSENDDOWN;
        break;

    case TEP_V1_SendStatus_DONOTSEND:
        return PLRESULT_OKDONOTHING;
        break;

    default:
    case TEP_V1_SendStatus_CLOSE_CONNECTION:
    case TEP_V1_SendStatus_MARSHAL_ERROR:
    case TEP_V1_SendStatus_MEMORY_ERROR:
        if(sendBlock.state != NULL)
        {
            TEP_V1_Session_Destroy(sendBlock.state);
        }
        return PLRESULT_ERROR;
        break;
    }
}

static void TEP_V1_Send_NegotiateSession(TEP_V1_SendBlock block)
{
    boolean useKey;
    boolean isConfirm;

    PCR_Assert(block != NULL);
    block->pdu = 1;
    block->tep->blackboardData->flags.multicastProtocolNotAllowed = TRUE;
    isConfirm = (block->operation->operation_type == DOFOPERATIONTYPE_TEP_CONFIRM);
    useKey = !block->context->flags.extendingPermission;
    if(isConfirm)
    {
        static const TEP_V1_SendUtilFunction confirmFunctions[] = {
                TEP_V1_Send_LookupSession,
                TEP_V1_Send_WriteFirstBytes,
                TEP_V1_Send_WriteInitiatorConfirmation,
                NULL
        };

        uint8 index = (block->tep->isConnectionLess) ? 1 : 0;

        block->pdu |= 0x20;
        if(TEP_V1_Send_ProcessFunctions(block, &(confirmFunctions[index])))
        {
            block->tep->blackboardData->flags.isCommandNotResponse = TRUE;
            block->tep->blackboardData->delay = block->operation->sources->oal_timeout;
            if(!block->context->flags.extendingPermission)
            {
                if(!block->tep->isConnectionLess)
                {
                    TEP_SessionEvent_t eventType;

                    void * newContext;

                    if(block->context->flags.rekeying)
                    {
                        eventType = TEP_V1_SessionEvent_REKEY_SENDKEYANDDISABLE;
                    }
                    else
                    {
                        block->tep->blackboardData->flags.forceUnsecure = TRUE;
                        eventType = TEP_V1_SessionEvent_NEWSESSION_SUCCESS_SENDKEY;
                    }
                    newContext =
                        block->context->tep->sessionEventCallback(TEP_V1_Session_GetContext(block->state),
                                                                  block->context->sessionkey, eventType,
                                                                  block->context->initiatorBlock->modeOfOperation,
                                                                  block->operation->sources->channel,
                                                                  block->context->connectionSecurityScope,
                                                                  TEPOperation_GetInitiatorPermissions(block->operation),
                                                                  TEPOperation_GetResponderPermissions(block->operation),
                                                                  NULL, block->context->stateIdentifier,
                                                                  block->operation->sources->address, NULL,
                                                                  block->operation->sources->channel->remoteChannelData->secureCredentials);
                    if(!block->context->flags.rekeying && !block->context->tep->blackboardData->associatedChannel->remoteChannelData->flags.mustFinalizeTep)
                    {
                        newContext =
                            block->context->tep->sessionEventCallback(TEP_V1_Session_GetContext(block->state),
                                                                      block->context->sessionkey, TEP_V1_SessionEvent_NEWSESSION_FINALIZE,
                                                                      block->context->initiatorBlock->modeOfOperation,
                                                                      block->operation->sources->channel,
                                                                      block->context->connectionSecurityScope,
                                                                      TEPOperation_GetInitiatorPermissions(block->operation),
                                                                      TEPOperation_GetResponderPermissions(block->operation),
                                                                      NULL, block->context->stateIdentifier,
                                                                      block->operation->sources->address, NULL,
                                                                      block->operation->sources->channel->remoteChannelData->secureCredentials);
                        TRPRouter_AddDynamicRoute(block->context->tep->blackboardData->associatedChannel->remoteChannelData->secureCredentials,
                                                  block->context->tep->blackboardData->associatedChannel, block->context->tep->blackboardData->transportAddr, FALSE);
                        TEP_V1_Session_IncrementKeyCount(block->state);
                        TEP_V1_Session_SetupRekeyTimer(block->state);
                        OALConnection_SetPeerCredentials(block->context->tep->blackboardData->associatedChannel->remoteChannelData->OALConn,
                                                        TEP_V1_Session_GetOtherCredentials(block->state));
                        block->context->tep->blackboardData->associatedChannel->remoteChannelData->flags.mustFinalizeTep = TRUE;
                        block->tep->blackboardData->seq = 1;
                    }
                    TEP_V1_Session_SetContext(block->state, newContext);
                }
            }
            else if(!block->tep->isConnectionLess)
            {
                OALCipher cipher = block->operation->sources->channel->remoteChannelData->cipher;

                if(!DOFPermissionSet_AddPermissionsFromSet(OALCipher_GetLocalPerms( cipher, block->context->keyState ),
                                                           TEPOperation_GetInitiatorPermissions(block->operation)))
                {
                    DOFOperation_Unlock(block->operation);
                    block->status = TEP_V1_SendStatus_CLOSE_CONNECTION;
                    return;
                }

                DOFOperation_Lock(block->operation);
                if(!block->context->calledCallback && (block->context->extendCompleteCallback != NULL))
                {
                    block->context->calledCallback = TRUE;
                    DOFOperation_Unlock(block->operation);
                    block->context->extendCompleteCallback(DOFERROR_OK, block->context->extendContext);
                }
                else
                {
                    DOFOperation_Unlock(block->operation);
                }
            }
        }
        return;
    }
    else if(useKey)
    {
        block->pdu |= 0x10;
    }

    if(block->context->flags.rekeying || block->context->flags.extendingPermission)
    {
        static const TEP_V1_SendUtilFunction rekeyFunctions[] = {
                TEP_V1_Send_WriteFirstBytes,
                TEP_V1_Send_WriteInitiatorBlock,
                NULL
        };
        if(TEP_V1_Send_ProcessFunctions(block, rekeyFunctions))
        {
            if(block->context->flags.rekeying)
            {
                block->tep->blackboardData->associatedChannel->remoteChannelData->flags.tepRekeying = TRUE;
            }
            block->tep->blackboardData->flags.isCommandNotResponse = TRUE;
            block->tep->blackboardData->delay = block->operation->sources->oal_timeout;
        }
    }
    else
    {
        static const TEP_V1_SendUtilFunction newSessionFunctions[] = {
                TEP_V1_Send_WriteFirstBytes,
                TEP_V1_Send_WriteDomain,
                TEP_V1_Send_WriteInitiatorBlock,
                NULL
        };
        if(TEP_V1_Send_ProcessFunctions(block, newSessionFunctions))
        {
            block->tep->blackboardData->flags.isCommandNotResponse = TRUE;
            block->tep->blackboardData->delay = block->operation->sources->oal_timeout;
        }
    }
}

static void TEP_V1_Send_NegotiateSession_r(TEP_V1_SendBlock block)
{
    boolean isConfirm;

    PCR_Assert(block != NULL);
    block->pdu = 1;
    block->tep->blackboardData->flags.isCommandNotResponse = FALSE;
    isConfirm = (block->operation->operation_type == DOFOPERATIONTYPE_TEP_CONFIRMRESPONSE);
    if(isConfirm)
    {
        static const TEP_V1_SendUtilFunction confirmFunctions[] = {
                TEP_V1_Send_LookupSession,
                TEP_V1_Send_WriteFirstBytes,
                NULL
        };

        block->pdu |= 0x20;
        if(TEP_V1_Send_ProcessFunctions(block, confirmFunctions))
        {
            if(!block->context->flags.extendingPermission)
            {
                TEP_V1_SM_ProcessOperation(block->operation);
            }
        }
    }
    else
    {
        static const TEP_V1_SendUtilFunction sessionFunctions[] = {
                TEP_V1_Send_LookupSession,
                TEP_V1_Send_WriteFirstBytes,
                TEP_V1_Send_WriteInitiatorTicket,
                TEP_V1_Send_WriteTicketConfirmation,
                TEP_V1_Send_WriteResponderInitialization,
                TEP_V1_Send_WriteResponderBlock,
                TEP_V1_Send_WriteAuthenticatorInitialization,
                NULL
        };

        static const TEP_V1_SendUtilFunction rekeyFunctions[] = {
                TEP_V1_Send_WriteFirstBytes,
                TEP_V1_Send_WriteInitiatorTicket,
                TEP_V1_Send_WriteTicketConfirmation,
                TEP_V1_Send_WriteResponderBlock,
                TEP_V1_Send_WriteAuthenticatorInitialization,
                NULL
        };

        OALStream_SetReadOffset(block->context->aBlock, 0);
        OALStream_SetReadOffset(block->context->rBlock, 0);
        if(TEP_V1_Send_ProcessFunctions(block, (block->context->flags.extendingPermission) ? rekeyFunctions
                                                                                           : sessionFunctions))
        {
            if(!block->context->flags.extendingPermission)
            {
                if(!block->tep->isConnectionLess)
                {
                    TEP_SessionEvent_t eventType;

                    void * newContext;

                    DOFOperation_Lock(block->operation);
                    if(block->context->flags.rekeying)
                    {
                        eventType = TEP_V1_SessionEvent_REKEY_SENDKEY;
                    }
                    else
                    {
                        block->tep->blackboardData->flags.forceUnsecure = TRUE;
                        eventType = TEP_V1_SessionEvent_NEWSESSION_SUCCESS_SENDKEY;
                    }
                    DOFOperation_Unlock(block->operation);
                    newContext =
                        block->context->tep->sessionEventCallback(TEP_V1_Session_GetContext(block->state),
                                                                  block->context->sessionkey, eventType,
                                                                  block->context->initiatorBlock->modeOfOperation,
                                                                  block->operation->sources->channel,
                                                                  block->context->connectionSecurityScope,
                                                                  TEPOperation_GetResponderPermissions(block->operation),
                                                                  TEPOperation_GetInitiatorPermissions(block->operation),
                                                                  NULL, block->context->stateIdentifier,
                                                                  block->operation->sources->address, NULL,
                                                                  block->operation->sources->channel->remoteChannelData->secureCredentials);
                    TEP_V1_Session_SetContext(block->state, newContext);
                }
            }
            else if(!block->tep->isConnectionLess)
            {
                OALCipher cipher = block->operation->sources->channel->remoteChannelData->cipher;

                if(!DOFPermissionSet_AddPermissionsFromSet(OALCipher_GetLocalPerms( cipher, block->context->keyState ),
                                                           TEPOperation_GetResponderPermissions(block->operation)))
                {
                    block->status = TEP_V1_SendStatus_CLOSE_CONNECTION;
                }
            }
        }
    }
}

static void TEP_V1_Send_EndSession(TEP_V1_SendBlock block)
{
    static const TEP_V1_SendUtilFunction functions[] = {
            TEP_V1_Send_LookupSession,
            TEP_V1_Send_WriteFirstBytes,
            NULL
    };

    PCR_Assert(block != NULL);
    block->pdu = 5;
    block->tep->blackboardData->flags.multicastProtocolNotAllowed = TRUE;
    block->tep->blackboardData->flags.securityRequiredAuthenticate = TRUE;
    if(TEP_V1_Send_ProcessFunctions(block, functions))
    {
        block->tep->blackboardData->flags.isCommandNotResponse = TRUE;
        PCR_Assert(block->status == TEP_V1_SendStatus_OK);
        block->status = TEP_V1_SendStatus_OKNOID;
        block->tep->sessionEventCallback(TEP_V1_Session_GetContext(block->state),
                                         NULL, TEP_V1_SessionEvent_ENDING, NULL,
                                         block->operation->sources->channel,
                                         NULL, NULL, NULL, NULL, 0, block->operation->sources->address, NULL,
                                         block->operation->sources->channel->remoteChannelData->secureCredentials);
    }
}

static void TEP_V1_Send_Reject_r(TEP_V1_SendBlock block)
{
    static const TEP_V1_SendUtilFunction functions[] = {
            TEP_V1_Send_WriteFirstBytes,
            TEP_V1_Send_WriteErrorCode,
            NULL
    };

    PCR_Assert(block != NULL);
    if(block->operation->flags.bits.internally_created)
    {
        block->pdu = 0;
        if(TEP_V1_Send_ProcessFunctions(block, functions))
        {
            block->tep->blackboardData->flags.isCommandNotResponse = FALSE;
        }
    }
    else
    {
        block->status =TEP_V1_SendStatus_DONOTSEND;
    }
}

static void TEP_V1_Send_SessionEnding(TEP_V1_SendBlock block)
{
    static const TEP_V1_SendUtilFunction functions[] = {
            TEP_V1_Send_WriteFirstBytes,
            NULL
    };

    PCR_Assert(block != NULL);
    block->pdu = 6;
    block->tep->blackboardData->flags.multicastProtocolNotAllowed = TRUE;
    block->tep->blackboardData->flags.securityRequiredAuthenticate = TRUE;
    if(TEP_V1_Send_ProcessFunctions(block, functions))
    {
        block->tep->blackboardData->flags.isCommandNotResponse = TRUE;
        block->status = TEP_V1_SendStatus_OKNOID;
    }
}

static void TEP_V1_Send_NoOp(TEP_V1_SendBlock block)
{
    OALStream stream = OALStream_CreateRW(2);

    PCR_Assert(block != NULL);
    if(stream != NULL)
    {
        block->tep->blackboardData->flags.isCommandNotResponse = TRUE;
        OALMarshal_StreamCompress2(stream, TEP_V1_APPID);
        OALPacket_AppendStream(block->tep->blackboardData->sendPacket, stream);
        block->status = TEP_V1_SendStatus_OKNOID;
        DOFOperation_Lock(block->operation);
        block->operation->operation_type = DOFOPERATIONTYPE_TEP_CONFIRMRESPONSE;
        DOFOperation_Unlock(block->operation);
        TEP_V1_SM_ProcessOperation(block->operation);       /* Special case */
    }
    else
    {
        block->status = TEP_V1_SendStatus_MEMORY_ERROR;
    }
}

static void TEP_V1_Send_NegotiateNoOp(TEP_V1_SendBlock block)
{
    OALStream stream = OALStream_CreateRW(2);

    PCR_Assert(block != NULL);
    if(stream != NULL)
    {
        block->tep->blackboardData->flags.isCommandNotResponse = TRUE;
        OALMarshal_StreamCompress2(stream, TEP_V1_APPID);
        OALPacket_AppendStream(block->tep->blackboardData->sendPacket, stream);
        block->status = TEP_V1_SendStatus_OKNOID;
    }
    else
    {
        block->status = TEP_V1_SendStatus_MEMORY_ERROR;
    }
}
