/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/domaininternal.h>
#include <internal/marshal.h>
#include <internal/objectidinternal.h>
#include <internal/security_common.h>
#include <internal/tep_v1.h>

#include "tep_v1_operation.h"
#include "tep_v1_sendutil.h"

boolean TEP_V1_Send_ProcessFunctions(      TEP_V1_SendBlock          block,
                                     const TEP_V1_SendUtilFunction * functions)
{
    uint8 i = 0;

    PCR_Assert(block != NULL);
    PCR_Assert(functions != NULL);
    while(functions[i] != NULL)
    {
        if(!functions[i](block, FALSE))
        {
            return FALSE;
        }
        PCR_Assert(block->status == TEP_V1_SendStatus_OK);      /* Double check return */
        ++i;
    }
    block->stream = OALStream_CreateRW(block->streamSize);
    if(block->stream != NULL)
    {
        i = 0;
        while(functions[i] != NULL)
        {
            if(!functions[i](block, TRUE))
            {
                return FALSE;
            }
            PCR_Assert(block->status == TEP_V1_SendStatus_OK);      /* Double check return */
            ++i;
        }
        if(OALPacket_AppendStream(block->packet, block->stream))
        {
            return TRUE;
        }
        OALStream_Release(block->stream, FALSE);
    }
    return FALSE;
}

boolean TEP_V1_Send_LookupSession(TEP_V1_SendBlock block,
                                  boolean          write)
{
    PCR_Assert(block != NULL);

    if(!write)
    {
        block->state = TEP_V1_SessionLookup(block->tep->blackboardData->associatedChannel);
        if(block->state == NULL)
        {
            block->status = TEP_V1_SendStatus_MEMORY_ERROR;
            return FALSE;
        }
    }
    return TRUE;
}

boolean TEP_V1_Send_WriteFirstBytes(TEP_V1_SendBlock block,
                                    boolean          write)
{
    if(write)
    {
        static uint16 const COMPRESSED_APPID = 0x8080 /* TEP_V1_Send_CompressAppID() */;

        if(OALStream_Write16(block->stream, COMPRESSED_APPID))
        {
            if(OALStream_Write8(block->stream, block->pdu))
            {
                return TRUE;
            }
        }
        block->status = TEP_V1_SendStatus_MARSHAL_ERROR;
        return FALSE;
    }
    else
    {
        static uint8 const APPID_SIZE = 2 /* OALMarshal_GetCompressedValueSize(2, TEP_V1_APPID) */;

        block->streamSize += APPID_SIZE + 1;
        return TRUE;
    }
}

boolean TEP_V1_Send_WriteDomain(TEP_V1_SendBlock block,
                                boolean          write)
{
    DOFObjectID domainId = DOFObjectIDDomain_GetObjectID( DOFCredentials_GetDomainID(block->operation->sources->channel->remoteChannelData->secureCredentials) );

    if(write)
    {
        if(!DOFObjectID_Marshal(domainId, block->stream))
        {
            block->status = TEP_V1_SendStatus_MARSHAL_ERROR;
            return FALSE;
        }
    }
    else
    {
        block->streamSize += (domainId ? DOFObjectID_GetBytesSize( domainId ) : 0);
    }
    return TRUE;
}

boolean TEP_V1_Send_WriteInitiatorBlock(TEP_V1_SendBlock block,
                                        boolean          write)
{
    if(write)
    {
        const uint8 * buffer = OALStream_GetReadPtr(block->context->iBlock);

        uint32 dataSize = OALStream_GetDataSize(block->context->iBlock);

        if(!OALStream_Write(block->stream, dataSize, buffer))
        {
            block->status = TEP_V1_SendStatus_MARSHAL_ERROR;
            return FALSE;
        }
    }
    else
    {
        uint32 blockSize = OALStream_GetDataSize(block->context->iBlock);

        block->streamSize += blockSize;
        OALStream_SetReadOffset(block->context->iBlock, 0);
    }
    return TRUE;
}

boolean TEP_V1_Send_WriteInitiatorTicket(TEP_V1_SendBlock block,
                                         boolean          write)
{
    OALKeyGrant keyGrant = block->context->initGrant;

    if(write)
    {
        if(!OALKeyGrant_Marshal(keyGrant, block->stream))
        {
            block->status = TEP_V1_SendStatus_MARSHAL_ERROR;
            return FALSE;
        }
    }
    else
    {
        block->streamSize += OALKeyGrant_MarshalSize(keyGrant);
    }
    return TRUE;
}

boolean TEP_V1_Send_WriteTicketConfirmation(TEP_V1_SendBlock block,
                                            boolean          write)
{
    if(write)
    {
        if(!OALStream_Write(block->stream, 32, block->context->responderConfirmation))
        {
            block->status = TEP_V1_SendStatus_MARSHAL_ERROR;
            return FALSE;
        }
    }
    else
    {
        block->streamSize += 32;
    }
    return TRUE;
}

boolean TEP_V1_Send_WriteResponderInitialization(TEP_V1_SendBlock block,
                                                 boolean          write)
{
    if(write)
    {
        if(!block->context->flags.rekeying)
        {
            if(!OALMarshal_StreamCompress4(block->stream, block->context->stateIdentifier) == DOFERROR_OK)
            {
                block->status = TEP_V1_SendStatus_MARSHAL_ERROR;
                return FALSE;
            }
        }
        if(OALSecurityMode_Marshal(block->context->initialState, block->stream))
        {
            return TRUE;
        }
        block->status = TEP_V1_SendStatus_MARSHAL_ERROR;
        return FALSE;
    }
    else
    {
        block->streamSize += OALSecurityMode_MarshalSize(block->context->initialState);
        if(!block->context->flags.rekeying)
        {
            block->streamSize += OALMarshal_GetCompressedValueSize(4, block->context->stateIdentifier);
        }
        return TRUE;
    }
}

boolean TEP_V1_Send_WriteResponderBlock(TEP_V1_SendBlock block,
                                        boolean          write)
{
    if(write)
    {
        const uint8 * buffer = OALStream_GetReadPtr(block->context->rBlock);

        uint32 dataSize = OALStream_GetDataSize(block->context->rBlock);

        if(!OALStream_Write(block->stream, dataSize, buffer))
        {
            block->status = TEP_V1_SendStatus_MARSHAL_ERROR;
            return FALSE;
        }
    }
    else
    {
        uint32 blockSize = OALStream_GetDataSize(block->context->rBlock);

        OALStream_SetReadOffset(block->context->rBlock, 0);
        block->streamSize += blockSize;
    }
    return TRUE;
}

boolean TEP_V1_Send_WriteAuthenticatorInitialization(TEP_V1_SendBlock block,
                                                     boolean          write)
{
    if(write)
    {
        if(!OALStream_ReadToStream(block->context->aBlock, OALStream_GetDataSize(block->context->aBlock), block->stream))
        {
            block->status = TEP_V1_SendStatus_MARSHAL_ERROR;
            return FALSE;
        }
    }
    else
    {
        uint32 blockSize = OALStream_GetDataSize(block->context->aBlock);

        block->streamSize += blockSize;
    }
    return TRUE;
}

boolean TEP_V1_Send_WriteInitiatorConfirmation(TEP_V1_SendBlock block,
                                               boolean          write)
{
    if(write)
    {
        if(!OALStream_Write(block->stream, 32, block->context->initiatorConfirmation))
        {
            block->status = TEP_V1_SendStatus_MARSHAL_ERROR;
            return FALSE;
        }
    }
    else
    {
        block->streamSize += 32;
    }
    return TRUE;
}

boolean TEP_V1_Send_WriteErrorCode(TEP_V1_SendBlock block,
                                   boolean          write)
{
    if(write)
    {
        if(!OALStream_Write8(block->stream, block->context->rejectCode))
        {
            block->status = TEP_V1_SendStatus_MARSHAL_ERROR;
            return FALSE;
        }
        else if(block->context->rejectCode == TEP_V1_ERROR_UNACCEPTABLEMODE)
        {
            const uint8 * streamPtr = OALStream_GetReadPtr(block->context->rejectData);

            if(!OALStream_Write(block->stream, OALStream_GetDataSize(block->context->rejectData), streamPtr))
            {
                block->status = TEP_V1_SendStatus_MARSHAL_ERROR;
                return FALSE;
            }
        }
    }
    else
    {
        block->streamSize += 1;
        if(block->context->rejectCode == TEP_V1_ERROR_UNACCEPTABLEMODE)
        {
            PCR_Assert(block->context->rejectData != NULL);
            block->streamSize += OALStream_GetDataSize(block->context->rejectData);
        }
    }
    return TRUE;
}
