/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/address.h>
#include <dof/error.h>
#include <internal/dofinternal.h>
#include <internal/mutex_pool.h>
#include <internal/permissionsetinternal.h>
#include <internal/staticassertinternal.h>
#include <internal/timer.h>

#include "tep_v1_operation.h"
#include "tep_v1_session.h"
#include "tep_v1_sm.h"

#ifdef TEP_V1_FORCE_QUICK_REKEY
#define TEP_V1_REKEYGRACE_INITIATOR     (1500)      /* 1.5 seconds */
#define TEP_V1_REKEYGRACE_RESPONDER     (1200)      /* 1.2 seconds */
#else
#define TEP_V1_REKEYGRACE_INITIATOR     (5 * 60 * 1000)     /* 5 minutes */
#define TEP_V1_REKEYGRACE_RESPONDER     ( TEP_V1_REKEYGRACE_INITIATOR / 2 ) /* Half of initiator */
#endif

#define CONVERT_DURATION_TO_MILLISECONDS(duration)      (duration * 8 * 60 * 1000)

typedef struct TEPV1SessionState_t
{
    OALTimerItem     sessionTimeoutItem;
    OALTimerItem     rekeyTimerItem;
    DOFPermissionSet initialPermissions;
    DOFCredentials   otherCreds;
    DOFAddress       address;
    OALChannel       channel;
    OALObserver      channelObserver;
    TEP_V1           tep;

    void * sessionContext;

    uint32 refCount;
    uint32 keyCount;

    uint8 duration;

    boolean isResponder;
    boolean isComplete;
} TEPV1SessionState_t;

static OALTimer     tepV1Timer;
static OALHashtable tepV1Sessions;

extern PCRMutex     ref_count_mutex;
extern PCRMutex     tepV1Mutex;
extern boolean      isRunning;

static int TEP_V1_SM_ProcessRekey(OALTimerItem   item,
                                  uint32         argc,
                                  void         * argv[],
                                  boolean        shutdown);

static int TEP_V1_SM_SendSessionEnding(OALTimerItem   item,
                                       uint32         argc,
                                       void         * argv[],
                                       boolean        shutdown);

static int TEP_V1_SM_StartEndSession(OALTimerItem   item,
                                     uint32         argc,
                                     void         * argv[],
                                     boolean        shutdown);

static void sessionStateObserverCb(OALObserver     self,
                                   OALObservable   observable,
                                   void          * arg)
{
    OALChannel channel =(OALChannel)arg;

    if (channel)
    {
        TEPV1SessionState state = TEP_V1_SessionLookup(channel);
        if (state)
        {
            PCRMutex_Lock(tepV1Mutex);
            if (state->sessionTimeoutItem)
            {
                /* If the timer was never inserted, the item and state still need to be destroyed. */
                if (!state->isComplete || OALTimerItem_Remove(state->sessionTimeoutItem))
                {
                    OALTimerItem_Destroy(state->sessionTimeoutItem, FALSE);
                    TEP_V1_Session_Destroy(state);  /* Call destroy since it would have been destroyed in the timer callback. */
                    state->sessionTimeoutItem = NULL;
                }
            }
            if (state->rekeyTimerItem && OALTimerItem_Remove(state->rekeyTimerItem))
            {
                OALTimerItem_Destroy(state->rekeyTimerItem, FALSE);
                TEP_V1_Session_Destroy(state);  /* Call destroy since it would have been destroyed in the timer callback. */
                state->rekeyTimerItem = NULL;
            }
            PCRMutex_Unlock(tepV1Mutex);
            TEP_V1_Session_Remove(state, channel);
            TEP_V1_Session_Destroy(state);  /* SessionLookup returned a reference, destroy it here. */
        }
    }
}

boolean TEP_V1_Session_Init(void)
{
    tepV1Timer = OALTimer_Create(DOF_GetDof()->threadPool, TRUE);
    if(tepV1Timer != NULL)
    {
        tepV1Sessions = OALHashtable_Create(1);
        if(tepV1Sessions != NULL)
        {
            return TRUE;
        }
        OALTimer_Destroy(tepV1Timer);
    }
    return FALSE;
}

static boolean TEP_V1_SM_CleanupSession(OALHashItem_t   item,
                                        void          * arg)
{
    TEPV1SessionState state = (TEPV1SessionState) item;

    if(!state->isComplete)
    {
        if (state->sessionTimeoutItem)
        {
            OALTimerItem_Destroy(state->sessionTimeoutItem, FALSE);
            TEP_V1_Session_Destroy(state);
        }
    }
    return FALSE;
}

void TEP_V1_Session_Shutdown(void)
{
    OALHashtable_Destroy(tepV1Sessions, TEP_V1_SM_CleanupSession, NULL);
    OALTimer_Destroy(tepV1Timer);
}

static boolean TEP_V1_SessionAdd(TEPV1SessionState state,
                                 OALChannel        channel)
{
    boolean retVal = FALSE;

    OALHashItem_t keys[1];

    static OALHashKeySize_t keySizes[1] = { sizeof(OALChannel) };

    PCR_Assert(channel != NULL);
    keys[0] = &channel;
    PCRMutex_Lock(tepV1Mutex);
    if(isRunning && OALHashtable_Insert(tepV1Sessions, state, keys, keySizes) == OALHashtable_InsertSuccess)
    {
        retVal = TRUE;
    }
    PCRMutex_Unlock(tepV1Mutex);
    return retVal;
}

OALTimerItem TEP_V1_Session_CreateTimerItem(OALTimer_fn_t   fn,
                                            uint32          argc,
                                            void          * argv[])
{
    return OALTimer_CreateItem(tepV1Timer, fn, argc, argv);
}

TEPV1SessionState TEP_V1_SessionLookup(OALChannel channel)
{
    OALHashItem_t keys[1];

    static OALHashKeySize_t keySizes[1] = { sizeof(OALChannel) };

    TEPV1SessionState retVal;

    PCR_Assert(channel != NULL);
    keys[0] = &channel;
    PCRMutex_Lock(tepV1Mutex);
    if(isRunning)
    {
        retVal = (TEPV1SessionState) OALHashtable_Retrieve(tepV1Sessions, keys, keySizes);
        if(retVal != NULL)
        {
            PCR_Assert(retVal->channel == channel);
            TEP_V1_Session_Retain(retVal);
        }
    }
    else
    {
        retVal = NULL;
    }
    PCRMutex_Unlock(tepV1Mutex);
    return retVal;
}

void TEP_V1_Session_Remove(TEPV1SessionState state,
                           OALChannel        channel)
{
    OALHashItem_t keys[1];

    static OALHashKeySize_t keySizes[1] = { sizeof(OALChannel) };

    PCR_Assert(state != NULL);
    PCR_Assert(channel != NULL);
    keys[0] = &channel;
    PCRMutex_Lock(tepV1Mutex);
    if(isRunning)
    {
        OALHashtable_Remove(tepV1Sessions, state, keys, keySizes);
    }
    PCRMutex_Unlock(tepV1Mutex);
}

TEPV1SessionState TEP_V1_Session_CreateNewState(boolean            isResponder,
                                                DOFPermissionSet   initialPermissions,
                                                OALChannel         channel,
                                                DOFCredentials     otherCredentials,
                                                DOFAddress         address,
                                                TEP_V1             tep,
                                                void             * context)
{
    TEPV1SessionState retVal;

    PCR_Assert(channel != NULL);
    PCR_Assert(address != NULL);
    retVal = DOFMem_Alloc(sizeof(TEPV1SessionState_t));
    if(retVal != NULL)
    {
        retVal->initialPermissions = (initialPermissions != NULL) ? DOFPermissionSet_DeepCopy(initialPermissions) : NULL;
        if((retVal->initialPermissions != NULL) || (initialPermissions == NULL))
        {
            retVal->otherCreds = (otherCredentials != NULL) ? DOFCredentials_Copy(otherCredentials) : NULL;
            if((retVal->otherCreds != NULL) || (otherCredentials == NULL))
            {
                retVal->address = DOFAddress_Copy(address);
                if(retVal->address != NULL)
                {
                    void * argv[1];

                    argv[0] = retVal;
                    retVal->sessionTimeoutItem = OALTimer_CreateItem(tepV1Timer, TEP_V1_SM_StartEndSession, 1, argv);
                    if(retVal->sessionTimeoutItem != NULL)
                    {
                        retVal->channel = OALChannel_Retain(channel);
                        if (retVal->channel)
                        {
                            retVal->rekeyTimerItem = NULL;
                            retVal->sessionContext = context;
                            retVal->duration = 0;
                            retVal->isResponder = isResponder;
                            retVal->refCount = 2;
                            retVal->keyCount = 0;
                            retVal->tep = tep;
                            retVal->isComplete = FALSE;
                            OALChannel_Lock(retVal->channel);
                            retVal->channelObserver = OALObservable_AddObserver(retVal->channel->chObservable, sessionStateObserverCb, retVal->channel);
                            OALChannel_Unlock(retVal->channel);
                            if (retVal->channelObserver)
                            {
                                if(TEP_V1_SessionAdd(retVal, channel))
                                {
                                    return retVal;
                                }
                                OALChannel_Lock(retVal->channel);
                                OALObservable_RemoveObserver(retVal->channel->chObservable, retVal->channelObserver);
                                OALChannel_Unlock(retVal->channel);
                            }
                            OALChannel_Release(retVal->channel);
                        }
                        OALTimerItem_Destroy(retVal->sessionTimeoutItem, FALSE);
                    }
                    DOFAddress_Destroy(retVal->address);
                }
                DOFCredentials_Destroy(retVal->otherCreds);
            }
            DOFPermissionSet_Destroy(retVal->initialPermissions);
        }
        DOFMem_Free(retVal);
    }
    return NULL;
}

TEPV1SessionState TEP_V1_Session_Retain(TEPV1SessionState state)
{
    PCR_Assert(state != NULL);

    PCRMutex_Lock(ref_count_mutex);
    PCR_Assert(state->refCount > 0);
    ++(state->refCount);
    PCR_Assert(state->refCount > 0);        /* Make sure we don't roll over */
    PCRMutex_Unlock(ref_count_mutex);
    return state;
}

void TEP_V1_Session_Destroy(TEPV1SessionState state)
{
    boolean mustFree;
    PCR_Assert(state != NULL);
    PCR_Assert(state->refCount > 0);

    PCRMutex_Lock(ref_count_mutex);
    mustFree = ((--(state->refCount)) == 0);
    PCRMutex_Unlock(ref_count_mutex);
    if(mustFree)
    {
        DOFPermissionSet_Destroy(state->initialPermissions);
        DOFCredentials_Destroy(state->otherCreds);
        DOFAddress_Destroy(state->address);
        OALChannel_Lock(state->channel);
        OALObservable_RemoveObserver(state->channel->chObservable, state->channelObserver);
        OALChannel_Unlock(state->channel);
        OALChannel_Release(state->channel);
        DOFMem_Free(state);
    }
}

void TEP_V1_Session_SetOtherCredentials(TEPV1SessionState state,
                                        DOFCredentials    credentials)
{
    PCR_Assert(state != NULL);
    if (state->otherCreds == NULL)
        state->otherCreds = DOFCredentials_Copy(credentials);
}

DOFCredentials TEP_V1_Session_GetOtherCredentials(TEPV1SessionState state)
{
    PCR_Assert(state != NULL);
    PCR_Assert(state->otherCreds != NULL);

    return state->otherCreds;
}

DOFAddress TEP_V1_Session_GetOtherAddress(TEPV1SessionState state)
{
    PCR_Assert(state != NULL);
    PCR_Assert(state->address != NULL);

    return state->address;
}

DOFPermissionSet TEP_V1_Session_GetInitialPermissions(TEPV1SessionState state)
{
    PCR_Assert(state != NULL);
    return state->initialPermissions;
}

void TEP_V1_Session_SetInitialPermissions(TEPV1SessionState state,
                                          DOFPermissionSet  initialPermissions)
{
    PCR_Assert(state != NULL);
    if(state->initialPermissions == NULL)
    {
        state->initialPermissions = DOFPermissionSet_Copy(initialPermissions);
    }
    else
    {
        DOFLog_Message(DOFLOGLEVEL_WARNING, "TEP Session: Attempting to set initial state permissions multiple times");
    }
}

void * TEP_V1_Session_GetContext(TEPV1SessionState state)
{
    PCR_Assert(state != NULL);
    return state->sessionContext;
}

void TEP_V1_Session_SetContext(TEPV1SessionState   state,
                               void              * newContext)
{
    PCR_Assert(state != NULL);
    state->sessionContext = newContext;
}

void TEP_V1_Session_SetDuration(TEPV1SessionState state,
                                uint8             grantedDuration)
{
    PCR_Assert(state != NULL);

    state->duration = grantedDuration;
}

void TEP_V1_Session_IncrementKeyCount(TEPV1SessionState state)
{
    PCR_Assert(state != NULL);

    /* Overflow is ok here since the actual value doesn't matter.  All that
     * matters is making sure this value changes between key sessions.
     */
    ++(state->keyCount);
}

uint32 TEP_V1_Session_GetKeyCount(TEPV1SessionState state)
{
    return (state->keyCount);
}

TEP_V1 TEP_V1_Session_GetTEP(TEPV1SessionState state)
{
    return state->tep;
}

static int TEP_V1_SM_ProcessRekey(OALTimerItem   item,
                                  uint32         argc,
                                  void         * argv[],
                                  boolean        shutdown)
{
    TEPV1SessionState state;

    PCR_Assert(argc == 2);
    PCR_Assert(argv[0] != NULL);
    state = (TEPV1SessionState) argv[0];
    if(!shutdown)
    {
        /* This typecast is smelly, but it saves us an extra alloc.  It should probably
         * be rewritten at some point, but for now it provides a significant benefit.
         */
        uint32 keyCount = (uint32) argv[1];

        if(keyCount == state->keyCount)
        {
            TEP_V1_Session_BeginRekey(state);
        }
    }
    PCRMutex_Lock(tepV1Mutex);
    OALTimerItem_Destroy(item, FALSE);
    state->rekeyTimerItem = NULL;
    PCRMutex_Unlock(tepV1Mutex);
    TEP_V1_Session_Destroy(state);
    return 0;
}

static int TEP_V1_SM_SendSessionEnding(OALTimerItem   item,
                                       uint32         argc,
                                       void         * argv[],
                                       boolean        shutdown)
{
    TEPV1SessionState state;

    DOFOperation sessionEndingOperation;

    PCR_Assert(argc == 2);
    PCR_Assert(argv[0] != NULL);
    state = (TEPV1SessionState) argv[0];
    if(!shutdown)
    {
        /* This typecast saves us an extra alloc.  It should probably
         * be rewritten at some point, but for now it provides a significant benefit.
         */
        uint32 keyCount = (uint32) argv[1];

        if(keyCount == state->keyCount)
        {
            sessionEndingOperation = DOFOperation_Create_SessionEnding(state->channel, state->address);
            if(sessionEndingOperation != NULL)
            {
                state->channel->routeToProvider(state->channel, sessionEndingOperation, sessionEndingOperation,
                                                TEP_V1_Session_GetOtherAddress(state));
                DOFOperation_Internal_Destroy(sessionEndingOperation);
            }
        }
    }
    PCRMutex_Lock(tepV1Mutex);
    OALTimerItem_Destroy(item, FALSE);
    state->rekeyTimerItem = NULL;
    PCRMutex_Unlock(tepV1Mutex);
    TEP_V1_Session_Destroy(state);
    return 0;
}

/* This assert confirms we aren't trimming the keyCount field used by TEP_V1_SM_ProcessRekey. */
STATIC_ASSERT(sizeof(void *) >= sizeof(uint32));

static int TEP_V1_SM_StartEndSession(OALTimerItem   item,
                                     uint32         argc,
                                     void         * argv[],
                                     boolean        shutdown)
{
    TEPV1SessionState state;

    DOFOperation endOp;

    PCR_Assert(argc == 1);
    PCR_Assert(argv[0] != NULL);
    state = (TEPV1SessionState) argv[0];
    endOp = DOFOperation_Create_End(state->channel, TEP_V1_Session_GetOtherAddress(state));
    if(endOp != NULL)
    {
        state->channel->routeToProvider(state->channel, endOp, endOp, TEP_V1_Session_GetOtherAddress(state));
        DOFOperation_Internal_Destroy(endOp);
    }
    PCRMutex_Lock(tepV1Mutex);
    OALTimerItem_Destroy(item, FALSE);
    state->sessionTimeoutItem = NULL;
    PCRMutex_Unlock(tepV1Mutex);
    TEP_V1_Session_Remove(state, state->channel);
    TEP_V1_Session_Destroy(state);
    return 0;
}

boolean TEP_V1_Session_SetupRekeyTimer(TEPV1SessionState state)
{
    static uint32 times[2] =
    {
        TEP_V1_REKEYGRACE_INITIATOR,
        TEP_V1_REKEYGRACE_RESPONDER
    };
    static OALTimer_fn_t const fns[] = { TEP_V1_SM_ProcessRekey, TEP_V1_SM_SendSessionEnding };
    uint32 embiggenedDuration;
    void * argv[2];

    PCR_Assert(state != NULL);
    embiggenedDuration = CONVERT_DURATION_TO_MILLISECONDS(state->duration);
    argv[0] = (void *) TEP_V1_Session_Retain(state);
    argv[1] = (void *) state->keyCount;

    /* If there is already a timer item for this session, remove it before adding a new one. */
    PCRMutex_Lock(tepV1Mutex);
    if (state->rekeyTimerItem)
    {
        if (OALTimerItem_Remove(state->rekeyTimerItem))
        {
            TEP_V1_Session_Destroy(state);
            OALTimerItem_Destroy(state->rekeyTimerItem, FALSE);
        }
        /* If the timer fired then there is nothing to take care of here. */
    }
    state->rekeyTimerItem = OALTimer_CreateItem(tepV1Timer, fns[state->isResponder], 2, argv);
    if(state->rekeyTimerItem != NULL)
    {
        if(OALTimerItem_Insert(state->rekeyTimerItem,
#ifndef TEP_V1_FORCE_QUICK_REKEY
                               embiggenedDuration -
#endif
                               times[state->isResponder], FALSE, FALSE) == DOFERROR_OK)
        {
            if(OALTimerItem_Insert(state->sessionTimeoutItem, embiggenedDuration, FALSE, FALSE) == DOFERROR_OK)
            {
                state->isComplete = TRUE;
                PCRMutex_Unlock(tepV1Mutex);
                return TRUE;
            }
        }
        OALTimerItem_Destroy(state->rekeyTimerItem, FALSE);
        state->rekeyTimerItem = NULL;
        TEP_V1_Session_Destroy(state);
    }
    PCRMutex_Unlock(tepV1Mutex);
    return FALSE;
}

void TEP_V1_Session_BeginRekey(TEPV1SessionState state)
{
    DOFOperation operation;

    PCR_Assert(state != NULL);
    operation = DOFOperation_Create_TepRekey(state->channel, state->address, state->initialPermissions, state->duration, state->tep, TEP_V1_TIMEOUT);
    if(operation != NULL)
    {
        if(OALOperationManager_AddOperation(operation, operation->opID, FALSE) == DOFERROR_OK)
        {
            TEP_V1_SM_ProcessOperation(operation);
        }
        DOFOperation_Internal_Destroy(operation);
    }
}
