/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/error.h>
#include <dof/oaltypes.h>
#include <internal/channel.h>
#include <internal/credentialsinternal.h>
#include <internal/enqueue.h>
#include <internal/esp_v2_options.h>
#include <internal/exception.h>
#include <internal/hashtable.h>
#include <internal/marshal.h>
#include <internal/operationinternal.h>
#include <internal/permissionsetinternal.h>
#include <internal/remotechannel.h>
#include <internal/serverinternal.h>
#include <internal/trp_v1.h>

#include "tep_v1_session.h"
#include "tep_v1_sm.h"

#define MAX_NONCE_LENGTH        16

#define TEP_V1_ERROR_CANNOT_CONTACT_AS      0x83

static OALEnqueue   tepV1Queue;
       PCRMutex     tepV1Mutex;
       boolean      isRunning;

static const uint8 zeroNonce[MAX_NONCE_LENGTH] = { 0, 0, 0, 0, 0, 0, 0, 0,
                                                   0, 0, 0, 0, 0, 0, 0, 0 };

static void TEP_V1_SM_EnqueueOperation(DOFOperation operation);

static uint32 TEP_V1_SM_GetNextStateIdentifier(void);

static boolean TEP_V1_SM_RequestTRPSession(DOFOperation      operation,
                                           TEPV1SessionState state,
                                           boolean           doRetain);

static void TEP_V1_SM_QueueFunction(OALThreadPoolItem dnode,
                                    boolean           shutdown);

/* Enqueue Stuff */
typedef void (*TEP_V1_SM_Operation)(DOFOperation operation);

static void TEP_V1_SM_BeginSession_Initiator(DOFOperation operation);

static void TEP_V1_SM_BeginSession_Responder(DOFOperation operation);

static void TEP_V1_SM_AuthorizeSession_Initiator(DOFOperation operation);

static void TEP_V1_SM_AuthorizeSession_Responder(DOFOperation operation);

static void TEP_V1_SM_ConfirmSession_Initiator(DOFOperation operation);

static void TEP_V1_SM_ConfirmSession_Responder(DOFOperation operation);

static void TEP_V1_SM_AckSession_Initiator(DOFOperation operation);

static void TEP_V1_SM_AckSession_Responder(DOFOperation operation);

static void TEP_V1_SM_EndSession(DOFOperation operation);

static void TEP_V1_SM_RejectSession_Initiator(DOFOperation operation);

static void TEP_V1_SM_RejectSession_Responder(DOFOperation operation);

static void TEP_V1_SM_ResolveCredsCallback( DOFCredentials self,
                                            DOFException exception,
                                            void* context );

static void TEP_V1_SM_RequestSessionCallback(TRP_V1_Context   trp_context,
                                             void           * context);

boolean TEP_V1_Init(void)
{
    tepV1Queue = OALEnqueue_Create(DOF_GetDof()->threadPool);
    if(tepV1Queue != NULL)
    {
        tepV1Mutex = PCRMutex_Create();
        if(tepV1Mutex != NULL)
        {
            if(TEP_V1_Session_Init())
            {
                isRunning = TRUE;
                return TRUE;
            }
            PCRMutex_Destroy(tepV1Mutex);
        }
        OALEnqueue_Destroy(tepV1Queue);
    }
    return FALSE;
}

void TEP_V1_Shutdown1(void)
{
    PCRMutex_Lock(tepV1Mutex);
    isRunning = FALSE;
    PCRMutex_Unlock(tepV1Mutex);
    OALEnqueue_Empty(tepV1Queue);
    TEP_V1_Session_Shutdown();
}

void TEP_V1_Shutdown2(void)
{
    OALEnqueue_Destroy(tepV1Queue);
    PCRMutex_Destroy(tepV1Mutex);
}

static int TEP_V1_SM_RetryFunction(OALTimerItem   item,
                                   uint32         argc,
                                   void         * argv[],
                                   boolean        shutdown);

boolean TEP_V1_SM_StartRetryTimer(DOFOperation operation)
{
    OALTimerItem timerItem;

    PCR_Assert(operation != NULL);
    PCR_Assert(operation->tepContext != NULL);
    timerItem = TEP_V1_Session_CreateTimerItem(TEP_V1_SM_RetryFunction, 1, (void **) &operation);
    if(timerItem != NULL)
    {
        DOFOperation_Retain(operation);
        if(OALTimerItem_Insert(timerItem, 1000, TRUE, FALSE) != DOFERROR_OK)
        {
            DOFOperation_Internal_Destroy(operation);
            return FALSE;
        }
        return TRUE;
    }
    return FALSE;
}

static int TEP_V1_SM_RetryFunction(OALTimerItem   item,
                                   uint32         argc,
                                   void         * argv[],
                                   boolean        shutdown)
{
    DOFOperation tepOp = (DOFOperation) argv[0];

    PCR_Assert(argc == 1);
    PCR_Assert(item != NULL);
    PCR_Assert(tepOp != NULL);
    if(shutdown)
    {
        OALTimerItem_Destroy(item, FALSE);
        DOFOperation_Internal_Destroy(tepOp);
        return 0;
    }
    else
    {
        switch(tepOp->operation_type)
        {
        case DOFOPERATIONTYPE_TEP_CONFIRM:
        {
            TEPV1SessionState state = TEP_V1_SessionLookup(tepOp->sources->channel);

            boolean doIt = FALSE;

            if(state != NULL)
            {
                doIt = (TEP_V1_Session_GetKeyCount(state) == 0);
                TEP_V1_Session_Destroy(state);
            }
            if(!doIt)
            {
                break;
            }
        }

        case DOFOPERATIONTYPE_TEP_BEGIN:
        case DOFOPERATIONTYPE_TEP_EXTEND:
            DOFOperation_Lock(tepOp);
            ++tepOp->sources->retry_count;
            DOFOperation_Unlock(tepOp);
            tepOp->sources->channel->routeToProvider(tepOp->sources->channel, tepOp, tepOp, tepOp->sources->address);
            return 0;
            break;

        case DOFOPERATIONTYPE_TEP_AUTHORIZE:
        case DOFOPERATIONTYPE_TEP_EXTENDAUTH:
            return 0;

        case DOFOPERATIONTYPE_TEP_CONFIRMRESPONSE:
        case DOFOPERATIONTYPE_TEP_END:
        case DOFOPERATIONTYPE_TEP_REJECT:
        case DOFOPERATIONTYPE_TEP_SESSIONENDING:
        case DOFOPERATIONTYPE_TEP_NOOP:     /* Can this case hit? */
            break;

        default:
            PCR_Assert(FALSE);
            break;
        }
        OALTimerItem_Destroy(item, FALSE);
        DOFOperation_Internal_Destroy(tepOp);
        return 1;
    }
}

void TEP_V1_SM_ProcessOperation(DOFOperation operation)
{
    PCR_Assert(operation != NULL);
    PCRMutex_Lock(tepV1Mutex);
    if(!isRunning)
    {
        PCRMutex_Unlock(tepV1Mutex);            /* avoid deadlock */
        return;
    }
    PCRMutex_Unlock(tepV1Mutex);
    TEP_V1_SM_EnqueueOperation(operation);
}

typedef struct OperationContext_t
{
    OALEnqueueItem_t super;
    DOFOperation operation;
} OperationContext_t;

static void TEP_V1_SM_EnqueueOperation(DOFOperation operation)
{
    OperationContext_t * node;

    PCR_Assert(operation != NULL);
    node = (OperationContext_t*) DOFMem_Alloc(sizeof(OperationContext_t));
    if(node != NULL)
    {
        OALEnqueueItem_Init((OALEnqueueItem) node, TEP_V1_SM_QueueFunction);
        node->operation = operation;
        DOFOperation_Retain(operation);         /* make sure there will be a valid operation in the callback */
        if(isRunning & !OALEnqueue_Enqueue(tepV1Queue, (OALEnqueueItem) node))
        {
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TEP V1: Unable to enqueue an operation.");
            DOFOperation_Internal_Destroy(node->operation);
            DOFMem_Free(node);
        }
    }
}

static uint32 TEP_V1_SM_GetNextStateIdentifier(void)
{
    return 0;
}

void TEP_V1_SM_BuildTicketConfirmation(DOFOperation   operation,
                                       boolean        responder,
                                       uint8        * buffer)
{
    TEP_V1_OperationContext context;

    OALSecureHMACContext confirmationHasher;

    uint16 zeroPaddingLength;

    PCR_Assert(operation != NULL);
    PCR_Assert(buffer != NULL);
    context = (TEP_V1_OperationContext) operation->tepContext;
    OALSecureHMAC_Start(&confirmationHasher, context->sessionkey);

    /* PAD[16](Ni) */
    zeroPaddingLength = MAX_NONCE_LENGTH - context->initiatorBlock->keyRequest->nonceLength;
    if(zeroPaddingLength > 0)
    {
        OALSecureHMAC_Digest(&confirmationHasher, zeroPaddingLength, zeroNonce);
    }
    OALSecureHMAC_Digest(&confirmationHasher, context->initiatorBlock->keyRequest->nonceLength,
                         context->initiatorBlock->keyRequest->nonce);

    /* PAD[16](Nr) */
    zeroPaddingLength = MAX_NONCE_LENGTH - context->keyRequest->nonceLength;
    if(zeroPaddingLength > 0)
    {
        OALSecureHMAC_Digest(&confirmationHasher, zeroPaddingLength, zeroNonce);
    }
    OALSecureHMAC_Digest(&confirmationHasher, context->keyRequest->nonceLength,
                         context->keyRequest->nonce);

    if(responder)
    {
        DOFObjectID identity = DOFObjectIDAuthentication_GetObjectID( DOFCredentials_GetIdentity(context->keyRequest->credentials) );

        /* S */
        if(!context->flags.extendingPermission)
        {
            uint8 stateIdBuffer[4];

            if(!context->flags.rekeying)
            {
                OALMarshal_CompressValue(4, context->stateIdentifier, 4, stateIdBuffer);
                OALSecureHMAC_Digest(&confirmationHasher,
                                     OALMarshal_GetCompressedValueSize(4, context->stateIdentifier),
                                     stateIdBuffer);
                PCR_Assert( context->initialState->length <= 0x7FFF );
            }
            OALMarshal_CompressValue(2, context->initialState->length, 4, stateIdBuffer);
            OALSecureHMAC_Digest(&confirmationHasher,
                                 OALMarshal_GetCompressedValueSize(2, context->initialState->length),
                                 stateIdBuffer);
            OALSecureHMAC_Digest(&confirmationHasher, context->initialState->length,
                                 context->initialState->initialState);
        }

        /* IDr */
        OALSecureHMAC_Digest(&confirmationHasher, (identity ? DOFObjectID_GetBytesSize(identity) : 0),
                             DOFObjectID_GetBytes(identity));
    }
    OALSecureHMAC_Finish(&confirmationHasher, buffer);
}

static boolean TEP_V1_SM_RequestTRPSession(DOFOperation      operation,
                                           TEPV1SessionState state,
                                           boolean           doRetain)
{
    TEP_V1_OperationContext tepContext;

    DOFPermissionSet libPermissions;
    DOFPermissionSet permSet;

    PCR_Assert(operation != NULL);
    PCR_Assert(operation->tepContext != NULL);
    tepContext = (TEP_V1_OperationContext) operation->tepContext;
    DOFOperation_Lock(operation);
    libPermissions = TEP_V1_Session_GetInitialPermissions(state);
    permSet = DOFPermissionSet_CreateComplement(tepContext->initiatorBlock->keyRequest->permission_set);
    if(permSet != NULL)
    {
        if(!tepContext->flags.extendingPermission && (libPermissions != NULL))
        {
            if(!DOFPermissionSet_AddPermissionsFromSet(permSet, libPermissions))
            {
                DOFPermissionSet_Destroy(permSet);
                DOFOperation_Unlock(operation);
                return FALSE;
            }
        }
        tepContext->keyRequest = DOFKeyRequest_Create(operation->sources->channel->remoteChannelData->secureCredentials,
                                                      DEFAULT_NONCE_SIZE, TRUE, FALSE, NULL, permSet);
        if(tepContext->keyRequest != NULL)
        {
            PCRSecureRandom_Generate(DEFAULT_NONCE_SIZE, tepContext->keyRequest->nonce);
            tepContext->rBlock = OALStream_CreateRW(DOFKeyRequest_MarshallByteLength(tepContext->keyRequest, TRUE));
            if(tepContext->rBlock != NULL)
            {
                boolean worked = DOFKeyRequest_Marshal(tepContext->rBlock, tepContext->keyRequest, TRUE);

                PCR_Assert(worked);
            }
            else
            {
                tepContext->rejectCode = TEP_V1_ERROR_INSUFFICIENT_RESOURCES;
                DOFOperation_Unlock(operation);
                DOFPermissionSet_Destroy(permSet);
                operation->sources->channel->routeToProvider(operation->sources->channel,
                                                             operation, operation, NULL);
                return FALSE;
            }
        }
        else
        {
            tepContext->rejectCode = TEP_V1_ERROR_INSUFFICIENT_RESOURCES;
            DOFOperation_Unlock(operation);
            DOFPermissionSet_Destroy(permSet);
            operation->sources->channel->routeToProvider(operation->sources->channel,
                                                         operation, operation, NULL);
            return FALSE;
        }
        DOFPermissionSet_Destroy(permSet);
    }
    else
    {
        tepContext->rejectCode = TEP_V1_ERROR_INSUFFICIENT_RESOURCES;
        DOFOperation_Unlock(operation);
        operation->sources->channel->routeToProvider(operation->sources->channel,
                                                     operation, operation, NULL);
        return FALSE;
    }
    if((tepContext->keyRequest != NULL) && (tepContext->keyRequest->permission_set != NULL))
    {
        tepContext->flags.keyRequested = 1;
        DOFOperation_Unlock(operation);
        if(doRetain)
        {
            DOFOperation_Retain(operation);     /* Make sure op is valid in callback */
        }
        TRP_V1_RequestSession(tepContext->rBlock, tepContext->keyRequest,
                              tepContext->iBlock, tepContext->initiatorBlock,
                              operation->sources->channel, operation->sources->address,
                              TEP_V1_SM_RequestSessionCallback, operation, operation->sources->oal_timeout);
        return TRUE;
    }
    else
    {
        /* TODO: Deal with alloc failure */
        DOFOperation_Unlock(operation);
        return FALSE;
    }
}

static void TEP_V1_SM_QueueFunction(OALThreadPoolItem dnode,
                                    boolean           shutdown)
{
    static const TEP_V1_SM_Operation functions[DOFOPERATIONTYPE_TEP_LAST - DOFOPERATIONTYPE_TEP_BEGIN][2] = {
        /* DOFOPERATIONTYPE_TEP_BEGIN */
            {TEP_V1_SM_BeginSession_Initiator,               TEP_V1_SM_BeginSession_Responder},
        /* DOFOPERATIONTYPE_TEP_AUTHORIZE */
            {TEP_V1_SM_AuthorizeSession_Initiator,           TEP_V1_SM_AuthorizeSession_Responder},
        /* DOFOPERATIONTYPE_TEP_CONFIRM */
            {TEP_V1_SM_ConfirmSession_Initiator,             TEP_V1_SM_ConfirmSession_Responder},
        /* DOFOPERATIONTYPE_TEP_CONFIRMRESPONSE */
            {TEP_V1_SM_AckSession_Initiator,                 TEP_V1_SM_AckSession_Responder},
        /* DOFOPERATIONTYPE_TEP_END */
            {TEP_V1_SM_EndSession,                           TEP_V1_SM_EndSession},
        /* DOFOPERATIONTYPE_TEP_EXTEND */
            {TEP_V1_SM_BeginSession_Initiator,               TEP_V1_SM_BeginSession_Responder},
        /* DOFOPERATIONTYPE_TEP_EXTENDAUTH */
            {TEP_V1_SM_AuthorizeSession_Initiator,           TEP_V1_SM_AuthorizeSession_Responder},
        /* DOFOPERATIONTYPE_TEP_REJECT */
            {TEP_V1_SM_RejectSession_Initiator,              TEP_V1_SM_RejectSession_Responder},
        /* DOFOPERATIONTYPE_TEP_SESSIONENDING */
            {NULL,                                  NULL},  /* Handled in timeritem and receive */
        /* DOFOPERATIONTYPE_TEP_NOOP */
            {NULL,                                  NULL},  /* Should never hit */
        /* DOFOPERATIONTYPE_TEP_NEGOTIATION_NOOP */
            {NULL,                                  NULL}   /* Should never hit */
    };

    OperationContext_t * node = (OperationContext_t*) dnode;

    PCR_Assert(node != NULL);
    DOFOperation_Lock(node->operation);
    PCR_Assert( node->operation->operation_type >= DOFOPERATIONTYPE_TEP_BEGIN && node->operation->operation_type < DOFOPERATIONTYPE_TEP_LAST );
    PCR_Assert(functions[node->operation->operation_type - DOFOPERATIONTYPE_TEP_BEGIN] [node->operation->flags.bits.internally_created] != NULL);
    DOFOperation_Unlock(node->operation);
    if(!shutdown)
    {
        functions[node->operation->operation_type - DOFOPERATIONTYPE_TEP_BEGIN]
                 [node->operation->flags.bits.internally_created](node->operation);
    }
    DOFOperation_Internal_Destroy(node->operation);        /* matches up with TEP_V1_SM_EnqueueOperation */
    DOFMem_Free(node);
}

static void TEP_V1_SM_BeginSession_Initiator(DOFOperation operation)
{
    TEP_V1_OperationContext opContext;

    PCR_Assert(operation != NULL);
    PCR_Assert(operation->tepContext != NULL);
    opContext = (TEP_V1_OperationContext) operation->tepContext;
    DOFOperation_Lock(operation);
    opContext->iBlock = OALStream_CreateRW(OALSessionInitiatorBlock_MarshalSize(opContext->initiatorBlock, TRUE));
    if(opContext->iBlock)
    {
        boolean worked = OALSessionInitiatorBlock_Marshal(opContext->initiatorBlock, opContext->iBlock, TRUE);

        PCR_Assert(worked);     /* Critical failure */
    }
    /* TODO: Handle alloc failure */
    DOFOperation_Unlock(operation);
    if(operation->flags.bits.is_connectionless)
    {
        TEP_V1_SM_StartRetryTimer(operation);
    }
    operation->sources->channel->routeToProvider(operation->sources->channel,
                                                 operation, operation, NULL);
}

static void TEP_V1_SM_GetCredentials(DOFOperation            operation,
                                     TEP_V1_OperationContext context);

static boolean TEP_V1_SM_ValidateSecurityMode(DOFOperation operation);

static void TEP_V1_SM_CreateLibPermsCallback(DOFError           status,
                                             sid_t            * secureSid,
                                             DOFPermissionSet   libPerms,
                                             void             * context);

static void TEP_V1_SM_BeginSession_Responder(DOFOperation operation)
{
    TEP_V1_OperationContext tepContext;

    PCR_Assert(operation != NULL);
    PCR_Assert(operation->tepContext != NULL);
    tepContext = (TEP_V1_OperationContext) operation->tepContext;

    if(!tepContext->flags.keyRequested)
    {
        if(operation->sources->channel->remoteChannelData->secureCredentials == NULL)
        {
            TEP_V1_SM_GetCredentials(operation, tepContext);
            return;
        }
        /* Do we need to resolve the requestor credentials? */
        if(!DOFCredentials_IsResolved(operation->sources->channel->remoteChannelData->secureCredentials))     /* This may hit even after the first time. */
        {
            DOFOperation_Retain(operation);

            DOFCredentials_Internal_BeginResolve( operation->sources->channel->remoteChannelData->secureCredentials,
                                                  DOF_GetDof(), operation->sources->oal_timeout, FALSE,
                                                  operation->sources->channel, operation->sources->address,
                                                  TEP_V1_SM_ResolveCredsCallback, operation );
        }
        else if(!tepContext->flags.rekeying && !tepContext->flags.extendingPermission)
        {
            if(TEP_V1_SM_ValidateSecurityMode(operation))
            {
                DOFOperation_Retain(operation);
                OALRemoteChannel_CreateLibPermissions(TEP_V1_SM_CreateLibPermsCallback, (void *) operation,
                                                      operation->sources->channel->remoteChannelData->secureCredentials,
                                                      operation->sources->channel, operation->sources->address);
            }
        }
        /* Credentials are resolved, so we can continue */
        else
        {
            TEPV1SessionState state = TEP_V1_SessionLookup(operation->sources->channel);

            if(state != NULL)       /* How'd we get here without a state? */
            {
                TEP_V1_SM_RequestTRPSession(operation, state, TRUE);
                TEP_V1_Session_Destroy(state);
            }
        }
    }
}

typedef struct fcs
{
    OALObserver obs;
    DOFOperation operation;
} fcs;

static void getCredsCallback(DOFError  status,
                             void    * context)
{
    fcs * f = (fcs*) context;

    OALChannel_Lock(f->operation->sources->channel);
    OALObservable_RemoveObserver(f->operation->sources->channel->chObservable, f->obs);
    OALChannel_Unlock(f->operation->sources->channel);
    if( status == DOFERROR_OK )
    {
        PCR_Assert(f->operation->sources->channel->remoteChannelData->secureCredentials != NULL);
        TEP_V1_SM_BeginSession_Responder(f->operation);
    }
    else
    {
        TEP_V1_OperationContext context = (TEP_V1_OperationContext) f->operation->tepContext;

        DOFOperation_Lock(f->operation);
        f->operation->operation_type = DOFOPERATIONTYPE_TEP_REJECT;
        context->rejectCode = TEP_V1_ERROR_AUTHENTICATAION_FAILED;
        DOFOperation_Unlock(f->operation);
        f->operation->sources->channel->routeToProvider(f->operation->sources->channel, f->operation,
                                                        NULL, f->operation->sources->address);
    }
    DOFOperation_Internal_Destroy(f->operation);
    DOFMem_Free(f);
}

static void channelDownCallback(OALObserver     self,
                                OALObservable   observable,
                                void          * arg)
{
    fcs * f = (fcs*) arg;

    DOFOperation_Internal_Destroy(f->operation);
    DOFMem_Free(arg);
}

static void TEP_V1_SM_GetCredentials(DOFOperation            operation,
                                     TEP_V1_OperationContext context)
{
    uint32 timeout = operation->sources->oal_timeout;
    fcs * f = (fcs*) DOFMem_Alloc(sizeof(fcs));

    if(f)
    {
        PCR_Assert(operation->sources->channel->remoteChannelData->flags.isConnection);
        OALChannel_Lock(operation->sources->channel);
        f->obs = OALObservable_AddObserver(operation->sources->channel->chObservable, channelDownCallback, f);
        OALChannel_Unlock(operation->sources->channel);
        if(f->obs)
        {
            DOFOperation_Retain(operation);
            f->operation = operation;
            OALRemoteChannel_SetCredentials(operation->sources->channel, context->domain,
                                            timeout, getCredsCallback, f);
        }
        else
        {
            DOFOperation_Lock(operation);
            operation->operation_type = DOFOPERATIONTYPE_TEP_REJECT;
            context->rejectCode = TEP_V1_ERROR_INSUFFICIENT_RESOURCES;
            DOFOperation_Unlock(operation);
            operation->sources->channel->routeToProvider(operation->sources->channel, operation,
                                                         NULL, operation->sources->address);
            DOFMem_Free(f);
        }
    }
}

static boolean TEP_V1_SM_ValidateSecurityMode(DOFOperation operation)
{
    TEP_V1_OperationContext tepContext;

    DOFSecurityDesire desire;

    DOFSecurityMode securityMode;

    PCR_Assert(operation != NULL);
    PCR_Assert(operation->tepContext != NULL);
    tepContext = (TEP_V1_OperationContext) operation->tepContext;
    securityMode = tepContext->initiatorBlock->modeOfOperation;
    desire =
        (operation->sources->channel->remoteChannelData->flags.isConnection ? operation->sources->channel->remoteChannelData->config.connection->securityDesire
                                                                            : operation->sources->channel->remoteChannelData->config.server->securityDesire);
    if(DOFSecurityModeCCM_GetStrength(securityMode) == DOFCIPHERSTRENGTH_256)
    {
        uint8 encrypt = DOFSecurityModeCCM_IsEncrypt(securityMode);

        switch(desire)
        {
        case DOFSECURITYDESIRE_ANY:
        case DOFSECURITYDESIRE_SECURE_ANY:
            return TRUE;
            break;

        case DOFSECURITYDESIRE_SECURE:
            if(encrypt)
            {
                return TRUE;
            }
            else
            {
                if(tepContext->rejectData != NULL)
                {
                    OALStream_Release(tepContext->rejectData, TRUE);
                }
                tepContext->rejectData = OALStream_CreateRW(8);
                if(tepContext->rejectData != NULL)
                {
                    tepContext->rejectCode = TEP_V1_ERROR_UNACCEPTABLEMODE;
                    if(!ESPV2_MarshalCCMNak(tepContext->rejectData, desire))
                    {
                        OALStream_Release(tepContext->rejectData, FALSE);
                        tepContext->rejectData = NULL;
                    }
                }
            }
            break;

        case DOFSECURITYDESIRE_SECURE_AUTHENTICATE_ONLY:
            if(!encrypt)
            {
                return TRUE;
            }
            else
            {
                if(tepContext->rejectData != NULL)
                {
                    OALStream_Release(tepContext->rejectData, TRUE);
                }
                tepContext->rejectData = OALStream_CreateRW(8);
                if(tepContext->rejectData != NULL)
                {
                    tepContext->rejectCode = TEP_V1_ERROR_UNACCEPTABLEMODE;
                    if(!ESPV2_MarshalCCMNak(tepContext->rejectData, desire))
                    {
                        OALStream_Release(tepContext->rejectData, FALSE);
                        tepContext->rejectData = NULL;
                    }
                }
            }
            break;

        default:
            break;
        }
    }
    else
    {
        if(tepContext->rejectData != NULL)
        {
            OALStream_Release(tepContext->rejectData, TRUE);
        }
        tepContext->rejectData = OALStream_CreateRW(8);
        if(tepContext->rejectData != NULL)
        {
            tepContext->rejectCode = TEP_V1_ERROR_UNACCEPTABLEMODE;
            if(!ESPV2_MarshalCCMNak(tepContext->rejectData, desire))
            {
                OALStream_Release(tepContext->rejectData, FALSE);
                tepContext->rejectData = NULL;
            }
        }
    }
    if(tepContext->rejectData != NULL)
    {
        DOFOperation_Lock(operation);
        operation->operation_type = DOFOPERATIONTYPE_TEP_REJECT;
        DOFOperation_Unlock(operation);
        operation->sources->channel->routeToProvider(operation->sources->channel, operation, NULL, operation->sources->address);
    }
    return FALSE;
}

static void TEP_V1_SM_CreateLibPermsCallback(DOFError           status,
                                             sid_t            * secureSid,
                                             DOFPermissionSet   libPerms,
                                             void             * context)
{
    DOFOperation operation;

    TEP_V1_OperationContext tepContext;

    operation = (DOFOperation) context;
    PCR_Assert(operation != NULL);
    PCR_Assert(operation->tepContext != NULL);
    tepContext = (TEP_V1_OperationContext) operation->tepContext;
    if(status == DOFERROR_OK)
    {
        TEPV1SessionState state;

        PCR_Assert(secureSid != NULL);
        PCR_Assert(libPerms != NULL);
        tepContext->secureSid = secureSid;
        source_retain(secureSid);
        state = TEP_V1_SessionLookup(operation->sources->channel);
        if(state == NULL)
        {
            state = TEP_V1_Session_CreateNewState(TRUE, libPerms, operation->sources->channel,
                                                  tepContext->initiatorBlock->keyRequest->credentials,
                                                  operation->sources->address, tepContext->tep, NULL);
        }
        else
        {
            TEP_V1_Session_SetInitialPermissions(state, libPerms);
            TEP_V1_Session_SetOtherCredentials(state, tepContext->initiatorBlock->keyRequest->credentials);
        }
        if(state != NULL)
        {
            TEP_V1_SM_RequestTRPSession(operation, state, TRUE);
            TEP_V1_Session_Destroy(state);
        }
    }
    DOFOperation_Internal_Destroy(operation);
}

void TEP_V1_SM_HandleAuthorize(DOFOperation operation)
{
    TEP_V1_SM_AuthorizeSession_Initiator(operation);
}

static void TEP_V1_SM_AuthorizeSession_Initiator(DOFOperation operation)
{
    TEP_V1_OperationContext opContext;

    uint8 i;

    OALCryptExpandedKey eKey;

    uint8 encMac[32];
    uint8 responderConfirmation[32];

    PCR_Assert(operation != NULL);
    PCR_Assert(operation->tepContext != NULL);
    opContext = (TEP_V1_OperationContext) operation->tepContext;
    PCRMem_Copy(encMac, opContext->initGrant->mac, 32);
    PCR_Validate( DOFCipherRegistry_AES256Cipher );
    DOFCipherRegistry_AES256Cipher->GenerateKeyState( eKey, DOFCredentials_GetKey(operation->sources->channel->remoteChannelData->secureCredentials) );
    DOFCipherRegistry_AES256Cipher->Encrypt( eKey, &encMac[0] );
    DOFCipherRegistry_AES256Cipher->Encrypt( eKey, &encMac[16] );
    for(i = 0; i < 32; ++i)
    {
        opContext->sessionkey[i] = (opContext->initGrant->key[i] ^ encMac[i]);
    }
    /* verify mac */
    TEP_V1_SM_BuildTicketConfirmation(operation, TRUE, responderConfirmation);
    if(PCRMem_Compare(opContext->responderConfirmation, responderConfirmation, 32) != 0)
    {
        /* error, responder confirmation does not match */
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR,
                       "TEP V1: Calculated and sent responder confirmations do not match.");
    }
    else
    {
        uint8 mac[32];

        OALSecureHMACContext hashContext;

        DOFObjectID domain = DOFObjectIDDomain_GetObjectID( DOFCredentials_GetDomainID(operation->sources->channel->remoteChannelData->secureCredentials) );

        /* Do these need to get reset? */
        OALStream_SetReadOffset(opContext->iBlock, 0);
        OALStream_SetReadOffset(opContext->rBlock, 0);
        OALStream_SetReadOffset(opContext->aBlock, 0);

        OALSecureHMAC_Start(&hashContext, (uint8*) DOFCredentials_GetKey(operation->sources->channel->remoteChannelData->secureCredentials));

        OALSecureHMAC_Digest(&hashContext, (domain ? DOFObjectID_GetBytesSize(domain) : 0), DOFObjectID_GetBytes(domain));
        OALSecureHMAC_Digest(&hashContext, OALStream_GetDataSize(opContext->iBlock), OALStream_GetReadPtr(opContext->iBlock));
        OALSecureHMAC_Digest(&hashContext, OALStream_GetDataSize(opContext->rBlock), OALStream_GetReadPtr(opContext->rBlock));
        OALSecureHMAC_Digest(&hashContext, OALStream_GetDataSize(opContext->aBlock), OALStream_GetReadPtr(opContext->aBlock));
        OALSecureHMAC_Digest(&hashContext, 32, opContext->sessionkey);

        OALSecureHMAC_Finish(&hashContext, mac);
        if(PCRMem_Compare(opContext->initGrant->mac, mac, 32) != 0)
        {
            /* error, mac does not match */
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR,
                           "TEP V1: Calculated MAC does not match MAC from AS");
        }
        else
        {
            if(!opContext->flags.extendingPermission)        /* Keying */
            {
                OALStream initialState = opContext->initialState->length ?
                    OALStream_CreateRO(opContext->initialState->initialState, opContext->initialState->length) : NULL;

                TEP_SessionEvent_t eventType;

                void * newContext;

                TEPV1SessionState state = TEP_V1_SessionLookup(operation->sources->channel);

                PCR_Assert(state != NULL);
                if(opContext->flags.rekeying)
                {
                    eventType = TEP_V1_SessionEvent_REKEY_RECVKEY;
                }
                else
                {
                    eventType = TEP_V1_SessionEvent_NEWSESSION_SUCCESS_RECVKEY;
                    TEP_V1_Session_SetOtherCredentials(state, opContext->keyRequest->credentials);
                }
                newContext =
                    opContext->tep->sessionEventCallback(TEP_V1_Session_GetContext(state),
                                                         opContext->sessionkey, eventType,
                                                         opContext->initiatorBlock->modeOfOperation,
                                                         operation->sources->channel,
                                                         opContext->connectionSecurityScope,
                                                         TEPOperation_GetInitiatorPermissions(operation),
                                                         TEPOperation_GetResponderPermissions(operation),
                                                         NULL, opContext->stateIdentifier,
                                                         operation->sources->address, &initialState,
                                                         operation->sources->channel->remoteChannelData->secureCredentials);
                TEP_V1_Session_SetContext(state, newContext);
                TEP_V1_Session_Destroy(state);
                OALStream_Release(initialState, FALSE);
                TEP_V1_Session_SetDuration(state, opContext->grantedDuration);
            }
            else
            {
                DOFPermissionSet_AddPermissionsFromSet(OALCipher_GetRemotePerms(operation->sources->channel->remoteChannelData->cipher, opContext->keyState),
                                                       TEPOperation_GetResponderPermissions(operation));
            }
            DOFOperation_Lock(operation);
            TEP_V1_SM_BuildTicketConfirmation(operation, FALSE, opContext->initiatorConfirmation);
            operation->operation_type = DOFOPERATIONTYPE_TEP_CONFIRM;
            DOFOperation_Unlock(operation);
            operation->sources->channel->routeToProvider(operation->sources->channel,
                                                         operation, operation, NULL);
        }
    }
}

static void TEP_V1_SM_AuthorizeSession_Responder(DOFOperation operation)
{
    TEP_V1_OperationContext context;

    TEP_SessionEvent_t eventType;

    OALStream initialStateStream = NULL;

    PCR_Assert(operation != NULL);
    PCR_Assert(operation->tepContext != NULL);
    context = (TEP_V1_OperationContext) (operation->tepContext);
    if(operation->operation_type == DOFOPERATIONTYPE_TEP_AUTHORIZE)
    {
        void * newContext;

        TEPV1SessionState state = TEP_V1_SessionLookup(operation->sources->channel);

        PCR_Assert(state != NULL);
        if(context->flags.rekeying)
        {
            eventType = TEP_V1_SessionEvent_REKEY_RECVKEY;
        }
        else
        {
            eventType = TEP_V1_SessionEvent_NEWSESSION_SUCCESS_RECVKEY;
            context->stateIdentifier = TEP_V1_SM_GetNextStateIdentifier();
        }
        newContext =
            context->tep->sessionEventCallback(TEP_V1_Session_GetContext(state),
                                               context->sessionkey, eventType,
                                               context->initiatorBlock->modeOfOperation,
                                               operation->sources->channel,
                                               context->connectionSecurityScope,
                                               TEPOperation_GetResponderPermissions(operation),
                                               TEPOperation_GetInitiatorPermissions(operation),
                                               NULL, context->stateIdentifier,
                                               operation->sources->address, &initialStateStream,
                                               operation->sources->channel->remoteChannelData->secureCredentials);
        TEP_V1_Session_SetContext(state, newContext);
        TEP_V1_Session_SetDuration(state, context->grantedDuration);
        TEP_V1_Session_Destroy(state);
    }
    if(context->initialState != NULL)
    {
        OALSecurityMode_Destroy(context->initialState);
    }
    /* initialStateStream NULL means we are streaming, if so this next creates securityMode of length field 0. */
    context->initialState = OALSecurityMode_Create(initialStateStream ? OALStream_GetDataSize(initialStateStream) : 0);
    if (initialStateStream)
    {
        PCRMem_Copy(context->initialState->initialState,
                    OALStream_GetReadPtr(initialStateStream),
                    context->initialState->length);
        OALStream_Release(initialStateStream, FALSE);
    }
    TEP_V1_SM_BuildTicketConfirmation(operation, TRUE, context->responderConfirmation);
    operation->sources->channel->routeToProvider(operation->sources->channel,
                                                 operation, operation, NULL);
}

static void TEP_V1_SM_ConfirmSession_Initiator(DOFOperation operation)
{
}

static void TEP_V1_SM_ConfirmSession_Responder(DOFOperation operation)
{
    TEP_V1_OperationContext tepContext;

    PCR_Assert(operation != NULL);
    tepContext = (TEP_V1_OperationContext) operation->tepContext;
    DOFOperation_Lock(operation);
    if(tepContext->flags.rekeying || tepContext->flags.extendingPermission)
    {
        operation->operation_type = DOFOPERATIONTYPE_TEP_CONFIRMRESPONSE;
    }
    else    /* should only hit on initial session */
    {
        operation->operation_type = DOFOPERATIONTYPE_TEP_NOOP;
    }
    DOFOperation_Unlock(operation);
    operation->sources->channel->routeToProvider(operation->sources->channel,
                                                 operation, operation, NULL);
}

static void TEP_V1_SM_AckSession_Initiator(DOFOperation operation)
{
    TEP_V1_OperationContext opContext;

    PCR_Assert(operation != NULL);
    PCR_Assert(operation->tepContext != NULL);
    opContext = (TEP_V1_OperationContext) operation->tepContext;
    if(!opContext->flags.extendingPermission)
    {
        TEPV1SessionState state = TEP_V1_SessionLookup(operation->sources->channel);

        if(state != NULL)
        {
            TEP_V1_Session_IncrementKeyCount(state);
            if(!TEP_V1_Session_SetupRekeyTimer(state))
            {
                /* failed to setup rekey timers */
            }
            TEP_V1_Session_Destroy(state);
        }
        else { }
    }
    DOFOperation_Cancel(operation);
}

static void TEP_V1_SM_AckSession_Responder(DOFOperation operation)
{
    TEP_V1_OperationContext opContext;

    PCR_Assert(operation != NULL);
    PCR_Assert(operation->tepContext != NULL);
    opContext = (TEP_V1_OperationContext) operation->tepContext;
    if(!opContext->flags.extendingPermission)
    {
        TEPV1SessionState state = TEP_V1_SessionLookup(operation->sources->channel);

        if(state != NULL)
        {
            TEP_V1_Session_IncrementKeyCount(state);
            if(!TEP_V1_Session_SetupRekeyTimer(state))
            {
                /* failed to setup rekey timers */
            }
            TEP_V1_Session_Destroy(state);
        }
        else { }
    }
}

static void TEP_V1_SM_RejectSession_Initiator(DOFOperation operation)
{
    TEP_V1_OperationContext context;

    PCR_Assert(operation != NULL);
    PCR_Assert(operation->tepContext != NULL);
    context = (TEP_V1_OperationContext) operation->tepContext;
    if(!context->flags.extendingPermission)
    {
        TEPV1SessionState sessionState = TEP_V1_SessionLookup(operation->sources->channel);

        if(sessionState != NULL)
        {
            TEP_SessionEvent_t event;

            DOFException exception;

            if(!context->flags.rekeying)
            {
                event = TEP_V1_SessionEvent_NEWSESSION_FAILURE;
            }
            else
            {
                event = TEP_V1_SessionEvent_REKEY_FAILURE;
            }
            if(TRP_V1_TriggerPossibleCacheReset(operation->sources->channel->remoteChannelData->secureCredentials, TRUE, context->rejectCode))
            {
                exception = DOFEXCEPTION_STALE_CREDENTIAL;
            }
            else
            {
                exception = DOFEXCEPTION_FAILED;
            }
            context->tep->sessionEventCallback(TEP_V1_Session_GetContext(sessionState), context->sessionkey,
                                            event, NULL, operation->sources->channel,
                                            context->connectionSecurityScope, NULL, NULL, exception,
                                            context->stateIdentifier, operation->sources->address,
                                            NULL, operation->sources->channel->remoteChannelData->secureCredentials);
            TEP_V1_Session_Destroy(sessionState);
        }
    }
    else
    {
        if(!context->calledCallback && (context->extendCompleteCallback != NULL))
        {
            context->extendCompleteCallback(context->rejectCode, context->extendContext);
            context->calledCallback = TRUE;
        }
    }
}

static void TEP_V1_SM_RejectSession_Responder(DOFOperation operation)
{
    /* can this entire function be optimized away? */
    PCR_Assert(operation != NULL);
    operation->sources->channel->routeToProvider(operation->sources->channel,
                                                 operation, operation, NULL);
}

static void TEP_V1_SM_EndSession(DOFOperation operation)
{
}

static void TEP_V1_SM_ResolveCredsCallback( DOFCredentials self,
                                            DOFException exception,
                                            void* context )
{
    DOFOperation operation;

    TEP_V1_OperationContext opContext;

    PCR_Assert(context != NULL);
    operation = (DOFOperation)context;
    opContext = (TEP_V1_OperationContext) operation->tepContext;
    if( !exception )
    {
        if(!opContext->flags.rekeying && !opContext->flags.extendingPermission)
        {
            if(TEP_V1_SM_ValidateSecurityMode(operation))
            {
                /* The operation was retained by TEP_V1_SM_BeginSession_Responder, so just pass ownership on to the context here. */
                OALRemoteChannel_CreateLibPermissions(TEP_V1_SM_CreateLibPermsCallback, (void *) operation,
                                                      operation->sources->channel->remoteChannelData->secureCredentials,
                                                      operation->sources->channel, operation->sources->address);
            }
        }
        /* Use the existing retain for this function to avoid another */
        else if(!TEP_V1_SM_RequestTRPSession(operation, NULL, FALSE))
        {
            DOFOperation_Internal_Destroy(operation);
        }
    }
    else
    {
        DOFOperation_Lock(operation);
        opContext->rejectCode = DOFErrorException_GetError( exception );
        operation->operation_type = DOFOPERATIONTYPE_TEP_REJECT;
        DOFOperation_Unlock(operation);
        DOFOperation_Internal_Destroy(operation);
    }
}

static void TEP_V1_SM_RequestSessionCallback(TRP_V1_Context   trp_context,
                                             void           * context)
{
    DOFOperation operation;

    TEP_V1_OperationContext opContext;

    PCR_Assert(context != NULL);
    operation = (DOFOperation) context;
    opContext = (TEP_V1_OperationContext) operation->tepContext;
    if(trp_context != NULL)
    {
        if((trp_context->SessionReturn->status == DOFERROR_OK) &&
           (trp_context->SessionReturn->errorCode == 0))
        {
            DOFOperation_Lock(operation);
            PCRMem_Copy(opContext->sessionkey, trp_context->SessionReturn->sessionKey, 32);

            DOFPermissionSet_Copy(trp_context->SessionReturn->initPermValidation);
            DOFPermissionSet_Copy(trp_context->SessionReturn->respPermValidation);
            DOFPermissionSet_Destroy(opContext->keyRequest->permission_set);
            DOFPermissionSet_Destroy(opContext->initiatorBlock->keyRequest->permission_set);
            opContext->initiatorBlock->keyRequest->permission_set = trp_context->SessionReturn->initPermValidation;
            opContext->keyRequest->permission_set = trp_context->SessionReturn->respPermValidation;

            opContext->initGrant = OALKeyGrant_Retain(trp_context->SessionReturn->initGrant);
            OALStream_Retain(trp_context->SessionReturn->A);
            opContext->aBlock = trp_context->SessionReturn->A;
            OALStream_SetReadOffset(trp_context->SessionReturn->A, 0);
            OALStream_SetReadOffset(opContext->rBlock, 0);
            opContext->grantedDuration = trp_context->SessionReturn->grantedDuration;
            opContext->connectionSecurityScope = DOFSecurityScope_Copy(trp_context->SessionReturn->connSecurityScope);

            /* Processing is done, so queue back into the state machine */
            if (operation->operation_type < DOFOPERATIONTYPE_TEP_REJECT)
            {
                ++(operation->operation_type);
            }
            DOFOperation_Unlock(operation);
        }
        else
        {
            uint8 errorCode = TEP_V1_ERROR_NORESPONSE;  /* There's probably a better default, but we'll burn that bridge as we cross it */

            switch(trp_context->ResolveReturn->status)
            {
            case DOFERROR_OK:
                errorCode = trp_context->ResolveReturn->errorCode;
                break;

            case DOFERROR_TIMEOUT:
                errorCode = TEP_V1_ERROR_NORESPONSE;
                break;

            default:
                errorCode = TEP_V1_ERROR_AUTHENTICATAION_FAILED;
                break;
            }
            DOFOperation_Lock(operation);
            opContext->rejectCode = errorCode;
            operation->operation_type = DOFOPERATIONTYPE_TEP_REJECT;
            DOFOperation_Unlock(operation);
        }
    }
    else
    {
        DOFOperation_Lock(operation);
        opContext->rejectCode = TEP_V1_ERROR_CANNOT_CONTACT_AS;
        operation->operation_type = DOFOPERATIONTYPE_TEP_REJECT;
        DOFOperation_Unlock(operation);
    }
    TEP_V1_SM_EnqueueOperation(operation);
    DOFOperation_Internal_Destroy(operation);   /* Matches up with retain in TEP_V1_SM_BeginSession_Responder */
}
