/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <internal/dofinternal.h>
#include <dof/core.h>
#include <internal/threadpool.h>
#include <internal/queue.h>
#include <internal/enqueue.h>
#include <internal/timer.h>

#include "internal_queue.h"
#include "threadpool_internal.h"

extern PCRMutex threadpool_mutex;

struct OALEnqueue_t
{
    OALThreadPool pool;
    InternalQueue queue;
    PCRCondition destroy_cond; /* NULL most of the time */
    struct
    {
        unsigned int accepting    : 1;
        unsigned int enq_enqueued : 1; /* 1 if something is in the thread pool already */
    } flags;
};

OALEnqueue OALEnqueue_Create(OALThreadPool pool)
{
    OALEnqueue enqueue = DOFMem_Alloc(sizeof(struct OALEnqueue_t));

    if (enqueue)
    {
        if ((enqueue->queue = InternalQueue_Create()))
        {
            enqueue->destroy_cond = PCRCondition_Create();
            if(enqueue->destroy_cond != NULL)
            {
                enqueue->flags.accepting = 1;
                enqueue->flags.enq_enqueued = 0;
                enqueue->pool = pool;
                PCRMutex_Lock(threadpool_mutex);
                ++pool->use_count;
                PCRMutex_Unlock(threadpool_mutex);
                return enqueue;
            }
            InternalQueue_Destroy(enqueue->queue);
        }
        DOFMem_Free(enqueue);
    }

    return NULL;
}

void OALEnqueue_Destroy(OALEnqueue enqueue)
{
    if (enqueue)
    {
        PCRMutex_Lock( threadpool_mutex );
        PCR_Assert( enqueue->flags.accepting ); /* don't destroy twice */
        enqueue->flags.accepting = 0;
        if (!enqueue->flags.enq_enqueued)
        {
            --enqueue->pool->use_count;
            PCRMutex_Unlock(threadpool_mutex);
        }
        else
        {
            while (enqueue->flags.enq_enqueued)
                PCRCondition_Wait( enqueue->destroy_cond, threadpool_mutex, PCRTIMEOUT_FOREVER );
            --enqueue->pool->use_count;
            PCRMutex_Unlock( threadpool_mutex );
        }
        /* at this point nobody else should ever use this enqueue */
        PCRCondition_Destroy( enqueue->destroy_cond );
        InternalQueue_Destroy( enqueue->queue );
        DOFMem_Free( enqueue );
    }
}

void OALEnqueue_Empty( OALEnqueue enqueue )
{
    if (enqueue)
    {
        PCRMutex_Lock( threadpool_mutex );
        if (enqueue->flags.enq_enqueued)
        {
            while (enqueue->flags.enq_enqueued)
                PCRCondition_Wait( enqueue->destroy_cond, threadpool_mutex, PCRTIMEOUT_FOREVER );
        }
        PCRMutex_Unlock( threadpool_mutex );
    }
}

static void fn_or_timer(OALEnqueueItem item, boolean shutdown, boolean *destroy)
{
    /* release mutex before calling user function */
    PCRMutex_Unlock( threadpool_mutex );
    item->userFunc( (OALThreadPoolItem) item, shutdown );
    PCRThread_Yield(); /* play nice */
    PCRMutex_Lock( threadpool_mutex );
}

static void enqueue_trampoline( OALThreadPoolItem itemx, boolean shutdown )
{
    OALEnqueueItem item = (OALEnqueueItem)itemx;
    OALEnqueue enqueue;
    boolean    destroy = FALSE;

    PCRMutex_Lock( threadpool_mutex );
    PCR_Assert( item );
    PCR_Assert( !shutdown || !item->enqueue->flags.accepting );
    /* If threadpool is not shutting down it's ok if the enqueue is.
       If it is shutting down, the enqueue must not be accepting anymore. */
    enqueue = item->enqueue;
    PCR_Assert( enqueue->flags.enq_enqueued );

    /* REVIEW: fn_or_timer() does NOT modify destroy right now. Is it important to consider a future use of destroy? */
    fn_or_timer( item, !item->enqueue->flags.accepting, &destroy );
    /* fn_or_timer() can call timer_trampoline(), which unlocks threadpool_mutex while it calls a timer callback. */

    PCR_Assert( enqueue->flags.enq_enqueued );
    /* At this point there is nothing from this enqueue in the thread pool. Check if there is another to add. */
    if (InternalQueue_GetSize( enqueue->queue ))
    {
        if (shutdown)
        {
            boolean destroy2 = FALSE; /* REVIEW: Nothing sets destroy2. */

            /* if shutting down, then just call all functions enqueued with the shutdown flag set */
            do
            {
                OALEnqueueItem item_ = (OALEnqueueItem)InternalQueue_Dequeue(enqueue->queue);

                PCR_Assert( item_ );
                fn_or_timer( item_, TRUE, &destroy2 );

                if (destroy2)
                {
                    PCR_Assert( FALSE );   /* REVIEW: Nothing sets destroy2. */
                    DOFMem_Free( item_ );
                }
            } while (InternalQueue_GetSize( enqueue->queue ));
        }
        else
        {
            OALEnqueueItem item_ = (OALEnqueueItem)InternalQueue_Dequeue(enqueue->queue);

            PCR_Assert(item_);
            OALThreadPool_RunInternal(item_->enqueue->pool, (OALThreadPoolItem)item_);
        }
    }
    else
    {
        enqueue->flags.enq_enqueued = 0;     /* Nothing to add, so return to not-in-use. Only place that clears flag. */
        if (enqueue->destroy_cond)                  /* This was set up by OALEnqueue_Destroy() or OALEnqueue_Empty(). */
            PCRCondition_Signal( enqueue->destroy_cond );
    }

    if (destroy)
    {
        PCR_Assert( FALSE ); /* REVIEW: Nothing sets destroy. */
        DOFMem_Free( item ); /* Free the old item. */
    }
    PCRMutex_Unlock( threadpool_mutex );
}

boolean OALEnqueue_Enqueue( OALEnqueue enqueue, OALEnqueueItem item )
{
    boolean result;

    PCRMutex_Lock( threadpool_mutex );
    result = OALEnqueue_EnqueueInternal( enqueue, item );
    PCRMutex_Unlock( threadpool_mutex );

    return result;
}

boolean OALEnqueue_EnqueueInternal( OALEnqueue enqueue, OALEnqueueItem item )
{
    /* Called with threadpool_mutex already locked, and it must stay locked. */
    PCR_Assert( enqueue );
    PCR_Assert( item != NULL );
    if (enqueue->flags.accepting)
    {
        item->enqueue = enqueue;
        if (!enqueue->flags.enq_enqueued)
        {
            /* nothing is enqueued right now, so just put it in the thread pool */
            OALThreadPool_RunInternal( enqueue->pool, (OALThreadPoolItem)item );
            enqueue->flags.enq_enqueued = 1; /* Only place that sets this flag. */
        }
        else
        {
            /* we already have something in the thread pool, so put it in our queue instead */
            InternalQueue_Enqueue( enqueue->queue, (OALThreadPoolItem)item );
        }
        return TRUE;
    }

    return FALSE;
}

void OALEnqueueItem_Init( OALEnqueueItem item, OALThreadPoolFunc callback )
{
    PCR_Assert( item != NULL );
    PCR_Assert( callback != NULL );

    OALThreadPoolItem_Init( (OALThreadPoolItem)item, enqueue_trampoline );
    item->userFunc = callback;
}

