/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/threadpool.h>
#include <internal/enqueue.h>
#include <internal/timer.h>
#include <internal/dofinternal.h>

#include "threadpool_internal.h"

#define MAX_SHUTDOWN_TIME       5000    /* Maximum time (ms) to wait at shutdown for stray users of the global mutex to cleanup. */

PCRMutex threadpool_mutex = 0;
static uint32 nPools = 0;

void OALThreadPool_Init(void)
{
    PCR_Assert(!threadpool_mutex);
    threadpool_mutex = PCRMutex_Create();
    PCR_Assert(threadpool_mutex); /* this is also ok here */
}

void OALThreadPool_Shutdown(void)
{
    uint32 start;
    PCR_Assert(threadpool_mutex);
    PCRMutex_Lock(threadpool_mutex);
    start = PCRTime_GetMS();
    while ( nPools && ( ( PCRTime_GetMS() - start ) < MAX_SHUTDOWN_TIME ) )
    {
        PCRMutex_Unlock(threadpool_mutex);
        PCRThread_Sleep(500);
        PCRMutex_Lock(threadpool_mutex);
    }
    PCR_Validate( nPools == 0 );  /* This only fails when a threadpool wasn't destroyed prior to shutdown. */
    PCRMutex_Unlock(threadpool_mutex);
    PCRMutex_Destroy(threadpool_mutex);
    threadpool_mutex = NULL;
}

/** @internal @addtogroup OALThreadPool

    The OALThreadPool is a combination of a queue and a fixed-size thread pool.
    All of the worker threads are allocated and started when the threadpool is
    created, thus there is deterministic performance throughout the lifetime of
    the pool.

    As long as free threads exist in the pool, work given to the pool will be
    executed immediately, and concurrently with existing work handled by the
    queue.  It is important to remember that synchronization of the work is the
    responsibility of the user of the thread pool.  It is also important to
    consider the effect of locking and blocking the worker threads, as this
    affects the performance of other work waiting in the queue.  It is
    generally a bad idea to have dependencies between work in the same pool
    (e.g., a thread waiting on other work to be done by the same pool), as this
    can lead to a deadlock (if there is only one thread in the pool and its
    work is waiting on work in the queue, the queue work doesn't have a thread
    and can't be serviced).

    The simplest case of the threadpool is using a single thread.  In this
    case, all work is serviced sequentially by the single thread.  When the
    thread is busy, additional work can still be queued and will be processed
    in the order in which it was enqueued.

@{
**/


/*
**  "Private" Type and Method Declarations
*/

static void threadpool_worker_fn( OALThreadPool tpool );


/*
**  "Internal Public"  Methods.
*/

OALThreadPool OALThreadPool_Create( uint16 nThreads )
{
    OALThreadPool tpool;
    uint32 size = offsetof( struct OALThreadPool_t, thread ) + sizeof(PCRThread) * (nThreads);

    if(!nThreads) return NULL;

    if ( (tpool = (OALThreadPool) DOFMem_Alloc( size )) != NULL )
    {
        PCRMem_Set( tpool, 0, size );
        tpool->flags.running = 1;
        tpool->use_count = 0;

        if ( ( tpool->queue = InternalQueue_Create() ) != NULL )
        {
            if ( ( tpool->cond = PCRCondition_Create() ) != NULL )
            {
                boolean errored = FALSE;

                PCRMutex_Lock(threadpool_mutex);
                tpool->count = nThreads;
                while((nThreads > 0) && !errored)
                {
                    if ( ( tpool->thread[--nThreads] = PCRThread_Create((PCRThreadCall)threadpool_worker_fn, tpool) ) == NULL )
                    {
                        errored = TRUE;
                    }
                }
                if ( !errored )
                {
                    tpool->flags.accepting = 1;
                    ++nPools;
                    PCRMutex_Unlock(threadpool_mutex);
                    return( tpool );
                }

                tpool->flags.running = 0;
                PCRCondition_Broadcast( tpool->cond );
                PCRMutex_Unlock(threadpool_mutex);
                for ( ; nThreads < tpool->count; nThreads++ )
                {
                    PCRThread_Join( tpool->thread[nThreads] );
                    PCRThread_Destroy( tpool->thread[nThreads] );
                }
                PCRCondition_Destroy( tpool->cond );
            }
            InternalQueue_Destroy( tpool->queue );
        }
        DOFMem_Free( tpool );
    }

    return NULL;
}

void OALThreadPool_Destroy( OALThreadPool tpool )
{
    if (!tpool)
        return;

    PCRMutex_Lock(threadpool_mutex);
    PCR_Assert(!tpool->use_count);
    tpool->flags.accepting = 0;
    PCRCondition_Broadcast( tpool->cond );

    while ( tpool->count > 0 )
    {
        PCRThread thread = tpool->thread[--tpool->count];
        PCRMutex_Unlock( threadpool_mutex );
        /* Cannot hold mutex here, since the thread needs it to cleanup and terminate. */
        PCRThread_Join( thread );
        PCRThread_Destroy( thread );
        PCRMutex_Lock( threadpool_mutex );
    }
    --nPools;
    /* All threads are gone; no need for the mutex held anymore (we have sole access to the tpool instance). */
    PCRMutex_Unlock( threadpool_mutex );

    /* Nothing should be left at this point, so make sure of that. */

    PCR_Assert(!InternalQueue_GetSize(tpool->queue));

    PCRCondition_Destroy(tpool->cond);
    InternalQueue_Destroy( tpool->queue );
    DOFMem_Free(tpool);
}

boolean OALThreadPool_Run( OALThreadPool tpool, OALThreadPoolItem item )
{
    boolean result = FALSE;

    PCR_Assert( tpool );
    PCR_Assert(item != NULL);

    PCRMutex_Lock(threadpool_mutex);
    if ( tpool->flags.accepting)
    {
        InternalQueue_Enqueue( tpool->queue, item );
        PCRCondition_Signal( tpool->cond );
        result = TRUE;
    }

    PCRMutex_Unlock(threadpool_mutex);
    return( result );
}

void OALThreadPool_RunInternal(OALThreadPool tpool, OALThreadPoolItem item)
{
    /* called with threadpool mutex already locked, and it must stay locked */
    PCR_Assert(tpool && item);
    PCR_Assert(tpool->flags.accepting);

    InternalQueue_Enqueue(tpool->queue, item);
    PCRCondition_Signal(tpool->cond);
}

/** @internal
    The main thread function for a worker thread in the pool.

    @param tpool  The threadpool to which this worker thread belongs.
**/
static void threadpool_worker_fn( OALThreadPool tpool )
{
    PCRMutex_Lock(threadpool_mutex);
    while ( tpool->flags.running )
    {
        OALThreadPoolItem item;

        boolean shutdown = !tpool->flags.accepting;

        if ( !InternalQueue_GetSize( tpool->queue ) )
        {
            if ( !shutdown )
                PCRCondition_Wait( tpool->cond, threadpool_mutex, PCRTIMEOUT_FOREVER );
            else
                tpool->flags.running = 0;
            continue;
        }
        item = InternalQueue_Dequeue( tpool->queue );
        PCR_Assert( item );
        PCRMutex_Unlock(threadpool_mutex);
        item->callback(item, shutdown);
        PCRThread_Yield();
        PCRMutex_Lock(threadpool_mutex);
    }

    PCRMutex_Unlock(threadpool_mutex);
}

void OALThreadPoolItem_Init(OALThreadPoolItem item,
                            OALThreadPoolFunc fn)
{
    PCR_Assert(item != NULL);
    PCR_Assert(fn != NULL);

    item->next = NULL;
    item->callback = fn;
}

/**
@}
**/