/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/error.h>
#include <internal/avl_multitree.h>
#include <internal/dofinternal.h>
#include <internal/enqueue.h>
#include <internal/timer.h>
#include "threadpool_internal.h"

extern PCRMutex threadpool_mutex;
static PCRThread timer_thread;
static PCRCondition timer_condition;

static void timer_thread_fn(void *unused);

static boolean timer_active;

static OALAVLMultiTree timer_trees[2];

static uint8 tree_index;
uint32 now_;

static boolean timer_gt(void * timeout, OALTimerItem item, void *context);
static boolean timer_eq(void * timeout, OALTimerItem item, void *context);

static void calc_relative_time(uint32 the_time, uint32 *timeout, uint8 *index);

/**
 A timed function queue.

 @internal

 @{
 **/
struct OALTimerItem_t {
    OALEnqueueItem_t super;
    OALTimer_fn_t fn;
    uint32 argc;
    OALDListNode_t list_node;
    OALTimer timer;
    uint32 timeout; /* absolute time */
    uint32 refill; /* for periodic timers, the refill value */
    uint32 offset_refill; /* calculated timer tick */
    PCRCondition destroy_cond;

    struct {
        unsigned int periodic           : 1; /* if true, re-arm after firing */
        unsigned int fired              : 1; /* if true, then item has already fired. periodic items will read 1 if the item has fired at least once */
        unsigned int internal           : 1; /* if true, then destroy the item after it fires */
        unsigned int index              : 1; /* the index into the timer_trees[] array that the item was put in */
        unsigned int refill_index       : 1;
        unsigned int inserted           : 1; /* if true, the item is inserted into the timer's tree and list */
        unsigned int not_removed        : 1; /* if true, the item is still in the timer's list (but has been removed from the tree) */
        unsigned int firing             : 1; /* set 1 when the callback is being called */
        unsigned int pending_destroy    : 1;
    } flags;
    void *argv[1]; /* array of arguments */
};

struct OALTimer_t
{
    OALThreadPool pool;
    OALEnqueue enqueue; /* if NULL, then timer is asynchronous */
    uint32 insert_count;
    uint32 item_count;
    uint32 enqueue_count;
    PCRCondition destroy_cond;
    OALDListNode head;
    struct {
        unsigned int accepting : 1;
    } flags;
};

void OALTimer_Init(void)
{
    timer_active = TRUE;
    timer_condition = PCRCondition_Create();
    PCR_Assert(timer_condition); /* this is also ok here */

    timer_trees[0] = OALAVLMultiTree_Create((OALAVLMultiTree_CompareGreaterThan_fn_t)timer_gt, (OALAVLMultiTree_CompareEqual_fn_t)timer_eq, 0, 0, 0);
    PCR_Assert(timer_trees[0] != NULL);
    timer_trees[1] = OALAVLMultiTree_Create((OALAVLMultiTree_CompareGreaterThan_fn_t)timer_gt, (OALAVLMultiTree_CompareEqual_fn_t)timer_eq, 0, 0, 0);
    PCR_Assert(timer_trees[1] != NULL);

    tree_index = 0;
    timer_thread = PCRThread_Create(timer_thread_fn, NULL);
    PCR_Assert(timer_thread != NULL);
}

void OALTimer_Shutdown(void)
{
    PCRMutex_Lock(threadpool_mutex);
    timer_active = FALSE;
    PCRCondition_Signal(timer_condition);
    PCRMutex_Unlock(threadpool_mutex);

    PCRThread_Join(timer_thread);
    PCRThread_Destroy(timer_thread);
    timer_thread = 0;

    PCRCondition_Destroy(timer_condition);
    timer_condition = 0;

    PCR_Assert(!OALAVLMultiTree_GetNumItems(timer_trees[0])&&!OALAVLMultiTree_GetNumItems(timer_trees[1]));
    OALAVLMultiTree_Destroy(timer_trees[0], 0, 0);
    OALAVLMultiTree_Destroy(timer_trees[1], 0, 0);
    timer_trees[0] = timer_trees[1] = 0;
}

static boolean timer_gt(void * timeout, OALTimerItem item, void *context)
{
    return ((uint32)timeout)>item->timeout;
}

static boolean timer_eq(void * timeout, OALTimerItem item, void *context)
{
    return ((uint32)timeout)==item->timeout;
}

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t find_min(OALAVLMultiTreeItem_t item, void* _item, boolean new_key)
{
    *((OALTimerItem*)_item) = ((OALTimerItem)item);
    return OALAVLMULTITREE_TERMINATE_TRAVERSE; /* found first... so quit! */
}

static OALAVLMultiTree_Traverse_All_Return_Type_fn_t index_switch_traverse(OALAVLMultiTreeItem_t treeItem, void* item, boolean new_key)
{
    *((OALTimerItem*)item) = ((OALTimerItem)treeItem);
    if ((*(OALTimerItem *)item)->timeout < 0x80000000UL)
    {
        /* leave first 24 days in tree. */
        *((OALTimerItem*)item) = NULL;
        return OALAVLMULTITREE_NORMAL;  /* keep going */
    }
    return OALAVLMULTITREE_TERMINATE_TRAVERSE; /* found first that needs to be fired. */
}

static void OALTimerItem_reset_item( OALTimerItem item )
{
    /* called with threadpool_mutex held */
    --item->timer->enqueue_count; /* now have one less timer item floating around */
    item->flags.fired = 1;
    item->flags.firing = 0;
}

/* return TRUE on detecting pending destroy of timer item, after dealing with it appropriately. */
static boolean CheckDestroy( OALTimerItem item, boolean * destroy )
{
    /* called with threadpool_mutex held */
    /* did the user want to destroy the item? */
    if (item->flags.pending_destroy)
    {
        PCR_Assert( !item->flags.inserted && !item->flags.not_removed ); /*From _Remove() done in OALTimerItem_Destroy() that set .pending_destroy*/
        OALTimerItem_reset_item( item );
        if (!item->destroy_cond)
        {
            *destroy = TRUE;
            --item->timer->item_count;
        }
        else
            PCRCondition_Signal( item->destroy_cond ); /* Signal timer ITEM destroy condition, not that of the timer. */
        return TRUE;
    }

    return FALSE;
}

static void timer_trampoline( OALThreadPoolItem ditem, boolean shutdown)
{
    OALTimer     timer;
    OALTimerItem item      = (OALTimerItem)ditem;
    OALTimerItem send_item = item;
    boolean      destroy   = FALSE;

    PCRMutex_Lock( threadpool_mutex );
    PCR_Assert( item );
    timer = item->timer;
    PCR_Assert( !shutdown || !timer->flags.accepting );
    /* If thread pool not shutting down, then it's ok if the timer is. should not be accepting in that case. */
    if (CheckDestroy( item, &destroy ))
    {
        if (timer->destroy_cond)
        {
            PCR_Assert( !item->flags.inserted && !item->flags.not_removed );/*From _Remove() in OALTimerItem_Destroy() that created ->destroy_cond*/
            PCRCondition_Signal( timer->destroy_cond );      /* Signal timer destroy condition, not that of THE ITEM. */
        }
        PCRMutex_Unlock( threadpool_mutex );
        if (destroy)
            DOFMem_Free( item );
        return;
    }

    /* If the item is internal, then don't give it to the timer callback. */
    if (item->flags.internal)
        send_item = NULL;

    {
        int     ret;
        boolean shutdown = !timer->flags.accepting;

        /* The meaning of .firing is we are calling the timer callback (or once called, that we are still in the timer
           callback). After the callback returns, paths below can clear .firing. */
        PCR_Assert( item->flags.firing );  /* Set in timer_thread_fn() only. Cleared only below and in _reset_item(). */
        PCRMutex_Unlock( threadpool_mutex );

        ret = item->fn(send_item, item->argc&0x3fffffffUL, item->argv, shutdown); /* <<<<<<<<<<<< The timer callback! */

        /* When timer_observer() is this fn(), it can enqueue a function that will destroy our timer_item item, but as
           OALTimer_Destroy() is called to do that, it holds to perform the actual destroy until the below
           CheckDestroy() call signals it. Thus item below is guaranteed valid. */
        PCRThread_Yield(); /* play nice */
        PCRMutex_Lock( threadpool_mutex );
        if (ret && !item->flags.inserted)
             item->flags.periodic = 0; /* no longer periodic */
    }

    if (CheckDestroy( item, &destroy ))
    {
        if (timer->destroy_cond)
        {
            PCR_Assert( timer->destroy_cond );
            PCRCondition_Signal( timer->destroy_cond );
        }
        PCRMutex_Unlock( threadpool_mutex );
        if (destroy)
            DOFMem_Free( item );
        return;
    }

    for ( ;; )
    {
        if (timer->flags.accepting && item->flags.periodic)
        {
            OALAVLTree_InsertionStatus sts;

            if ( item->flags.inserted )
            {
                /* Already inserted?  Must've been inserted while it was firing... */
                OALTimerItem_reset_item( item );
                item->flags.fired = 0;
                break;
            }

            item->timeout = item->offset_refill;
            item->flags.index = item->flags.refill_index;

            sts = OALAVLMultiTree_Insert(timer_trees[item->flags.index], (void*)item->timeout, item);
            /* now see if we can insert into the tree. to save memory let's re-use the index variable */
            if ( sts == OALAVLTree_InsertSuccess )
            {
                /* inserted successfully */
                OALTimerItem_reset_item( item );
                OALDList_Insert( &timer->head, &item->list_node );
                item->flags.inserted = 1; /* if the item was waiting to be fired it won't fire now */
                item->flags.not_removed = 1;
                ++timer->insert_count;
                item->flags.fired = 0;
                PCRCondition_Signal( timer_condition );
                break;
            }
            else
            {
                PCR_Assert( sts != OALAVLTree_AlreadyInserted );
                /* no memory? try again */
                PCRMutex_Unlock( threadpool_mutex );
                PCRThread_Sleep( 10 );
                PCRMutex_Lock( threadpool_mutex );
                if (CheckDestroy( item, &destroy ))
                {
                    PCRMutex_Unlock( threadpool_mutex );
                    if (destroy)
                        DOFMem_Free( item );
                    return;
                }
            }
        }
        else if (item->flags.internal)
        {
            /* destroy internal item */
            OALTimerItem_reset_item( item );
            DOFMem_Free( item );
            break;
        }
        else
        {
            OALTimerItem_reset_item( item );
            break;
        }
    }
    PCRMutex_Unlock( threadpool_mutex );
}

static void timer_thread_fn(void *unused)
{
    uint32 time_to_wait;
    PCRMutex_Lock(threadpool_mutex);

    now_ = PCRTime_GetMS();
    time_to_wait = -now_;

    for (;;)
    {
        uint32       prev_;
        OALTimerItem item;
        uint8        cur_index;

        /* If there are no items in the tree, time_to_wait will be set to the next time that PCRTime_GetMS will roll over.
           If there are items, time_to_wait will be the timeout for the next timer item.
           In either case it is ok to wait for less time, since it will just wake up, update prev_and now_, set time_to_wait, and then wait again. */
        if (OALAVLMultiTree_GetNumItems( timer_trees[tree_index] ) == 0 || time_to_wait != -now_)
        {
            if (!timer_active)
                goto timerThreadOut;

            if (time_to_wait > 0x80000000UL)
            {
                /* avoid a situation where the tree index should be switched, but now_ and prev_ are both at zero, or they are both small and prev_ > now_ was missed */
                time_to_wait = 0x80000000UL;
            }

            PCRCondition_Wait(timer_condition, threadpool_mutex, time_to_wait);
        }

        /* Proceed with mutex locked. */
        cur_index = tree_index;
        prev_ = now_;
        now_ = PCRTime_GetMS();
        if (prev_ > now_)
        {
            /* time overflowed, so switch the indexes. Happens every 21 days for 2^^32 millisecond time base. */
            tree_index ^= 1;
        }

        /* handle all items that should fire now (esp. in case of rollover) */
        for (;;)
        {
            item = NULL;

            if (prev_ > now_)
            {
                OALAVLMultiTree_Traverse_All(timer_trees[cur_index], index_switch_traverse, &item);
            }
            else
            {
                OALAVLMultiTree_Traverse_All(timer_trees[cur_index], find_min, &item);
            }
            if (item &&                                                                    /* Timer item soonest to fire. */
                (now_ >= item->timeout ||                                                 /* item's firing time is now. */
                 prev_ > now_))   /* Tells us that time overflowed and item is held in the previous tree- fire it now. */
            {
                /* If the item is firing, we need to wait for it to finish. */
                if ( item->flags.firing )
                {
                    PCRMutex_Unlock(threadpool_mutex);
                    PCRThread_Yield();
                    PCRMutex_Lock(threadpool_mutex);
                    continue;
                }

                /* is the item periodic? */
                if (item->flags.periodic)
                {
                    uint8 index;

                    calc_relative_time(item->refill, &item->offset_refill, &index);
                    item->flags.refill_index = index;
                }

                /*_Traverse_All() above can find matching item->timeout and return item no longer inserted- nothing checks*/
                if (!item->flags.inserted)
                {
                    /* Do not call timer callback! OALAVLMultiTree_Traverse_All() above returned item no longer inserted. */
                    PCR_Assert( OALAVLMultiTree_Retrieve_Item( timer_trees[item->flags.index], (void*)item->timeout, item )
                                                                                                                  == NULL );
                }
                else
                {
                    /* Setup calling of timer callback for this item. */
                    boolean success = TRUE;

                    PCR_Assert( OALAVLMultiTree_Retrieve_Item( timer_trees[item->flags.index], (void*)item->timeout, item )
                                                                                                                  == item );
                    /* Cause timer_trampoline() to execute (in a different thread), which will call item's timer callback.*/
                    if (item->timer->enqueue)
                    {
                        /* OALEnqueue_Enqueue() not needed here because we HAVE threadpool_mutex locked. */
                        success = OALEnqueue_EnqueueInternal( item->timer->enqueue, (OALEnqueueItem)item );
                    }
                    else
                    {
                        /* put item in threadpool */
                        OALThreadPoolItem_Init((OALThreadPoolItem) item, timer_trampoline);
                        PCR_Assert( item->timer->pool );
                        OALThreadPool_RunInternal(item->timer->pool, (OALThreadPoolItem)item);
                    }
                    if (success)
                    {
                        /* Item is to be processed in a new thread that will call timer_trampoline(). We still have the
                           mutex lock, so update the timer status. item cannot go away while mutex is locked. */
                        PCR_Assert( item->timer->insert_count );
                        ++item->timer->enqueue_count; /* Above, we setup timer_trampoline to run- it will decrement this. */
                        item->flags.firing = 1;       /* timer_trampoline() callback will be called by timer_trampoline().*/
                        item->flags.inserted = 0;     /* About to delete item from timer tree- say it is no longer there. */
                        item->flags.not_removed = 0;  /* About to delete item from timer list- say it is no longer there. */
                        --item->timer->insert_count;
                        OALDList_Remove( &item->timer->head, &item->list_node );
                        {
                            boolean r = OALAVLMultiTree_Delete(timer_trees[item->flags.index], (void *)item->timeout, item);

                            PCR_Assert( r );
                            PCR_Assert( OALAVLMultiTree_Retrieve_Item( timer_trees[item->flags.index], (void*)item->timeout,
                                                                       item ) == NULL );
                            /* NOTE: multitrees can store multiple items with same key (key here is a 32 bit time value.
                                     Thus, even though we just deleted timeValue/item, two things are notable:
                                        Understandably, calling OALAVLMultiTree_Retrieve() as a test could return non-NULL,
                                            because other items may remain having same key (our timeValue).
                                        Calling OALAVLMultiTree_Delete() AGAIN as a test can unexpectedly return TRUE,
                                            because the specific implementation has a sneak path when a matching item is not
                                            found. The whole class depends on that behavior, and so we can call this a
                                            feature rather than a bug. */
                        }
                    }
                }
                /* continue inner loop */
            }
            else
            {
                if (prev_ <= now_)
                {
                    /* set time_to_wait to the next item timeout, or when now_ will overfolow. */
                    if (item)
                        time_to_wait = item->timeout - now_;
                    else
                        time_to_wait = -now_;
                }
                else
                {
                    time_to_wait = 0;
                }

                break; /* out of inner loop */
            }
        } /* end for(;;). */
    } /* end for(;;). */

timerThreadOut:
    PCRMutex_Unlock(threadpool_mutex);
}

static void calc_relative_time(uint32 the_time, uint32 *timeout, uint8 *index)
{
    uint32 now = PCRTime_GetMS();

    /* time is relative */
    *timeout = now+the_time;

    if ((0xffffffffUL-now)>=the_time)
        *index = tree_index;
    else
        *index = tree_index^1;                                                          /* Passing time can overflow. */

    /* If PCRTime_GetMS has rolled over, but tree_index has not been changed yet, the index found above will be fixed here.*/
    if (now < now_)
        *index ^= 1;
}

OALTimer OALTimer_Create(OALThreadPool pool, boolean synchronous)
{
    /* there is no reason to lock the threadpool_mutex */
    OALTimer timer = DOFMem_Alloc( sizeof( struct OALTimer_t ) );

    if (timer)
    {
        PCRMutex_Lock(threadpool_mutex);
        timer->pool = pool;
        if (synchronous)
        {
            timer->enqueue = OALEnqueue_Create(pool);
            if(!timer->enqueue)
            {
                DOFMem_Free(timer);
                PCRMutex_Unlock(threadpool_mutex);
                return NULL;
            }
        }
        else
            timer->enqueue = 0;

        timer->insert_count = timer->item_count = 0;
        timer->flags.accepting = 1;

        timer->enqueue_count = 0;
        timer->destroy_cond = NULL;

        timer->head = NULL;
        ++pool->use_count;
        PCRMutex_Unlock(threadpool_mutex);
    }

    return timer;
}

static boolean OALTimer_destroy1(OALDListNode item, void *arg)
{
    PCR_Assert(item);

    {
        /* Called with threadpool_mutex locked. */
        OALTimerItem item_ = node2struct(item, struct OALTimerItem_t, list_node);

        /* any item here should be internal */
        /*PCR_Assert(_item->flags.internal);*/

        /* already inserted, so find it in the trees */
        {
            boolean r = OALAVLMultiTree_Delete(timer_trees[item_->flags.index], (void*)item_->timeout, item_);
            PCR_Assert(r);
        }
        item_->flags.inserted = 0;
        --item_->timer->insert_count;
        item_->flags.fired = 0;
    }

    return FALSE; /* don't exit early */
}

static void OALTimer_destroy2(OALDListNode item, void *arg)
{
    PCR_Assert(item);

    {
        /* Called with threadpool_mutex NOT locked. */
        OALTimerItem item_ = node2struct(item, struct OALTimerItem_t, list_node);

        if(item_->flags.internal)
        {
            item_->fn(NULL, item_->argc&0x3fffffffUL, item_->argv, TRUE);
            DOFMem_Free(item_);
        }
        else
        {
            item_->fn(item_, item_->argc&0x3fffffffUL, item_->argv, TRUE);
            /* expect user to clean up item */
        }
    }
}

void OALTimer_Destroy(OALTimer timer)
{
    if (timer)
    {
        OALDListNode head = NULL;

        PCRMutex_Lock(threadpool_mutex);

        timer->flags.accepting = 0;

        if (timer->insert_count)
        {
            head = timer->head;
            /* Any items left? traverse over all items and remove from trees so timer thread will not see them. */
            timer->head = NULL;
            OALDList_Traverse(head, OALTimer_destroy1, 0);
            /* After the items that are already enqueued run, we'll come back to our head and clean them up. */
        }

        if ( timer->enqueue_count )
        {
            timer->destroy_cond = PCRCondition_Create();
            while ( timer->enqueue_count )
                PCRCondition_Wait(timer->destroy_cond, threadpool_mutex, PCRTIMEOUT_FOREVER);
            PCRCondition_Destroy(timer->destroy_cond);
        }
        PCR_Assert( timer->enqueue_count == 0 );
        --timer->pool->use_count;
        PCRMutex_Unlock(threadpool_mutex); /* REVIEW: Should this move down to protect traverse of _destroy2()? */

        /* at this point our local head pointer begins a doubly linked list of all items removed just above in the timer
           and they need their function called with the shutdown flag true*/
        OALDList_RemoveAll(&head, OALTimer_destroy2, 0); /* Iterate over these and call the function with shutdown flag set.*/
        /* after calling this head is no longer valid */

        OALEnqueue_Destroy( timer->enqueue );
        PCR_Validate( !timer->item_count );     /* at this point all user created items must have been destroyed */
        DOFMem_Free( timer );
    }
}

static OALTimerItem OALTimer_create_item(OALTimer timer, OALTimer_fn_t fn, uint32 argc, void *argv[], boolean internal)
{
    PCR_Assert(timer);

    {
        OALTimerItem item = DOFMem_Alloc(offsetof(struct OALTimerItem_t, argv) + (sizeof(void*)*(argc)));

        if (item)
        {
            OALEnqueueItem_Init((OALEnqueueItem)item, timer_trampoline);
            item->fn = fn;
            item->argc = argc;
            PCRMem_Copy(item->argv, argv, sizeof(void*)*argc);

            item->timer = timer;
            item->timeout = item->refill = 0;
            item->destroy_cond = NULL;

            item->flags.fired = 1; /* indicates the timer item is not in a tree and needs no cleanup work */
            item->flags.internal = internal;
            item->flags.index = 0; /* just clear it for now */
            item->flags.periodic = 0;
            item->flags.inserted = 0;
            item->flags.not_removed = 0;
            item->flags.firing = 0;
            item->flags.pending_destroy = 0;

            item->list_node.next = item->list_node.prev = NULL;

            PCRMutex_Lock(threadpool_mutex);
            if (timer->flags.accepting)
            {
                if (!internal)
                    ++(timer->item_count);
            }
            else
            {
                DOFMem_Free( item );
                item = NULL;
            }
            PCRMutex_Unlock(threadpool_mutex);
        }

        return item;
    }
}

boolean OALTimer_Insert( OALTimer timer, uint32 relative_time, OALTimer_fn_t fn, uint32 argc, void *argv[],
                         boolean periodic )
{
    OALTimerItem item = OALTimer_create_item( timer, fn, argc, argv, TRUE );

    if (item)
    {
        if (OALTimerItem_Insert( item, relative_time, periodic, FALSE ) != DOFERROR_OK)
        {
            DOFMem_Free( item );
            return FALSE;
        }
        return TRUE;
    }

    return FALSE;
}

DOFError OALTimerItem_Insert( OALTimerItem item, uint32 relative_time, boolean periodic, boolean notOnTimerItemFiring )
{
    DOFError retStatus = DOFERROR_REJECTED;   /* Cannot have value < 3, as OALAVLMultiTree_Insert() can return those. */

    PCRMutex_Lock( threadpool_mutex );
    PCR_Assert( item && item->timer && item->timer->flags.accepting && !item->flags.pending_destroy );
    if (!notOnTimerItemFiring || !item->flags.firing)
    {
        uint8                      index;
        OALAVLTree_InsertionStatus treeStatus;

        item->flags.periodic = periodic;
        item->refill = relative_time;

        if (item->flags.inserted)
        {
            /* Delete before re-inserting. */
            boolean r = OALAVLMultiTree_Delete(timer_trees[item->flags.index], (void*)item->timeout, item);

            PCR_Assert( r );
            item->flags.inserted = 0;
            --item->timer->insert_count;
        }
        if ( item->flags.not_removed )
        {
            /* Delete before re-inserting. */
            OALDList_Remove( &item->timer->head, &item->list_node );
            item->flags.not_removed = 0;
        }

        /* figure out new time */
        calc_relative_time( relative_time, &item->timeout, &index );         /* Fills item->timeout, used just below. */
        item->flags.index = index;

        treeStatus = OALAVLMultiTree_Insert( timer_trees[item->flags.index], (void*)item->timeout, item );
        if (treeStatus == OALAVLTree_InsertSuccess)
        {
            /* inserted successfully */
            item->flags.inserted = 1; /* if the item was waiting to be fired it won't fire now */
            item->flags.not_removed = 1;
            ++item->timer->insert_count;
            item->flags.fired = 0;
            OALDList_Insert( &item->timer->head, &item->list_node );
            PCRCondition_Signal( timer_condition );
        }
        PCR_Assert( treeStatus == OALAVLTree_InsertSuccess || treeStatus == OALAVLTree_InsertFailure );
        retStatus = (treeStatus == OALAVLTree_InsertSuccess) ? DOFERROR_OK : DOFERROR_FAILED;
    }
    else
    { /* timer item is firing right now and we are asked not to insert when that is so. */
    }
    PCRMutex_Unlock( threadpool_mutex );

    return retStatus;
}


OALTimerItem OALTimer_CreateItem( OALTimer timer, OALTimer_fn_t fn, uint32 argc, void * argv[] )
{
    return OALTimer_create_item( timer, fn, argc, argv, FALSE );
}

boolean OALTimerItem_Remove( OALTimerItem item )
{
    boolean retBoolean = FALSE;

    PCRMutex_Lock( threadpool_mutex );
    PCR_Assert( item && item->timer && !item->flags.internal );
    if (item->flags.inserted)
    {
        boolean r = OALAVLMultiTree_Delete( timer_trees[item->flags.index], (void*)item->timeout, item );

        PCR_Assert( r );
        PCR_Assert( OALAVLMultiTree_Retrieve_Item( timer_trees[item->flags.index], (void*)item->timeout, item )
                                                                                                      == NULL );
        item->flags.inserted = 0;
        item->flags.fired = 1; /* pretend like it was fired */

        --item->timer->insert_count;
        retBoolean = TRUE;
    }
    if ( item->flags.not_removed )
    {
        if ( item->timer->head )
            OALDList_Remove( &item->timer->head, &item->list_node );
        item->flags.not_removed = 0;
    }

    PCRMutex_Unlock( threadpool_mutex );
    return retBoolean;
}

void OALTimerItem_Destroy( OALTimerItem item, boolean wait )
{
    if (item)
    {
        PCRMutex_Lock( threadpool_mutex );
        /* Note: Do not put log message calls in here. */
        PCR_Assert( item->timer );
        PCR_Assert( !item->flags.internal );
        PCR_Assert( !item->flags.pending_destroy );
        PCR_Assert( !item->destroy_cond );
        OALTimerItem_Remove(item);
        PCR_Assert( !item->flags.inserted ); /* From _Remove() just above. */
        PCR_Assert( !item->flags.not_removed ); /* From _Remove() just above. */
        if (item->flags.firing)
        {
            item->flags.pending_destroy = 1;
            if (!wait)
                goto out;
            item->destroy_cond = PCRCondition_Create();
            while (item->flags.firing)
                PCRCondition_Wait( item->destroy_cond,  threadpool_mutex, PCRTIMEOUT_FOREVER );
            PCRCondition_Destroy( item->destroy_cond );
            PCR_Assert( !item->flags.inserted ); /* From _Remove() above. */
            PCR_Assert( !item->flags.not_removed ); /* From _Remove() above. */
        }
        --(item->timer->item_count);
        DOFMem_Free( item );

    out:
        PCRMutex_Unlock(threadpool_mutex);
    }
}

/**
 Get the time remaining before an item's function is executed.

 @param item The item to check.

 @return The remaining time.
 **/
uint32 OALTimerItem_TimeRemaining(OALTimerItem item)
{
    uint32 temp;
    uint32 curr;
    uint8 curr_index;

    PCRMutex_Lock(threadpool_mutex);

    /* Set curr and curr_index after mutex is locked. */
    curr = PCRTime_GetMS();
    curr_index = tree_index;

    /* If PCRTime_GetMS has rolled over, but tree_index has not been changed yet, use what the index should be set to. */
    if (curr < now_)
        curr_index ^= 1;

    PCR_Assert(item);

    /* If PCRTime_GetMS has rolled over, but tree_index has not been changed yet, the index found above will be fixed here.*/
    if (curr_index == item->flags.index)
    {
        if (item->timeout > curr)
            temp = item->timeout - curr;
        else
            temp = 0;

    }
    else
    {
        if (item->timeout > curr)
            temp = 0;
        else
            temp = item->timeout - curr;
    }

    PCRMutex_Unlock(threadpool_mutex);
    return temp;
}

boolean OALTimerItem_IsInserted(OALTimerItem item)
{
    boolean periodic = FALSE;

    PCRMutex_Lock(threadpool_mutex);
    PCR_Assert(item);
    periodic = item->flags.inserted;
    PCRMutex_Unlock(threadpool_mutex);

    return periodic;
}

boolean OALTimerItem_IsPeriodic(OALTimerItem item)
{
    boolean periodic = FALSE;

    PCRMutex_Lock(threadpool_mutex);
    PCR_Assert(item);
    periodic = item->flags.periodic;
    PCRMutex_Unlock(threadpool_mutex);

    return periodic;
}

void OALTimerItem_SetPeriodic(OALTimerItem item, boolean periodic)
{
    PCRMutex_Lock(threadpool_mutex);
    PCR_Assert(item);
    item->flags.periodic = periodic;
    PCRMutex_Unlock(threadpool_mutex);
}

/**
@}
**/
