/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/timeinternal.h>

/**
    Compare two 32-bit times.

    @param time1 The first time to compare.
    @param time2 The second time to compare.

    @return Zero           if exact match.
    @return Positive value if time1 is after time2.
    @return Negative value if time1 is before time2.
**/
int8 DOFTime_Compare( uint32 time1, uint32 time2 )
{
    /* For the case where submitted times are in milliseconds, this function assumes the two times
                    are known to be within 24 days or so of each other.
       Explanation: PCR times are stored as 4 bytes of milliseconds, which overflow approximately every 49 days.
                    This common overflow denies straightforward comparison. But if the two input times are within 24
                    days of each other, this code truly determines which time is before/after the other. */
    uint32 diff = time1 - time2;

    if (!diff)
        return 0;

    return ((diff & 0x80000000) == 0) ? 1 : -1;
}


/**
    Compare two 16 bit times.

    @param time1 The first time to compare.
    @param time2 The second time to compare.

    @return Zero           if exact match.
    @return Positive value if time1 is after time2.
    @return Negative value if time1 is before time2.
**/
int8 DOFTime_Compare_16Bit( uint16 time1, uint16 time2 )
{
    uint16 diff = time1 - time2;

    if (!diff)
        return 0;

    return ((diff & 0x8000) == 0) ? 1 : -1;
}


/**
    Compare two 15 bit times.

    @param time1 The first time to compare.
    @param time2 The second time to compare.

    @return Zero           if exact match.
    @return Positive value if time1 is after time2.
    @return Negative value if time1 is before time2.
**/
int8 DOFTime_Compare_15Bit( uint16 time1, uint16 time2 )
{
    time1 <<= 1;
    time2 <<= 1;

    return DOFTime_Compare_16Bit( time1, time2 );
}
