/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/address.h>
#include <dof/transport.h>
#include <internal/transportinternal.h>

extern PCRMutex ref_count_mutex;

typedef struct DTLConnectionConfig_t
{
    DTLTransport                  transport;
    DTLConnectionConfig_Destroyer ccdf;
    void*                         config;
    uint32                        size;
    uint32                        refCount;
}DTLConnectionConfig_t;

DTLConnectionConfig DTLConnectionConfig_Create(      uint32                          configSize,
                                               const void                          * config,
                                                     DTLTransport                    transport,
                                                     DTLConnectionConfig_Destroyer   destroyer)
{
    DTLConnectionConfig ret = DOFMem_Alloc(sizeof(DTLConnectionConfig_t) + configSize);

    PCR_Assert(transport);

    if( !ret )
        return NULL;

    ret->refCount = 1;
    ret->ccdf = destroyer;
    ret->size = configSize;
    ret->transport = transport;
    ret->config = (configSize > 0)?(void*)(((uint8*)ret) + sizeof(DTLConnectionConfig_t)) : NULL;
    PCRMem_Copy( ret->config, config, configSize );
    return ret;
}

DTLConnectionConfig DTLConnectionConfig_Copy( const DTLConnectionConfig self )
{
    DTLConnectionConfig ret = NULL;
    if( self )
    {
        PCRMutex_Lock( ref_count_mutex );
        if( self->refCount < 0xFFFFFFFF )
        {
            ++(self->refCount);
            ret = self;
        }
        PCRMutex_Unlock( ref_count_mutex );
    }
    return ret;
}

boolean DTLConnectionConfig_IsEqual(const DTLConnectionConfig self,
                                    const DTLConnectionConfig config)
{
    if ( !self && !config )
        return TRUE;
    if( self && config )
    {
        if( self->transport != config->transport && self->size != config->size )
            return FALSE;
        if( self->transport->fns->IsConnectionConfigEqual )
            return self->transport->fns->IsConnectionConfigEqual( self->transport, self->size, self->config, config->config );
        return (PCRMem_Compare(self->config, config->config, self->size) == 0);
    }
    return FALSE;
}

void DTLConnectionConfig_Destroy( DTLConnectionConfig self )
{
    if( self )
    {
        boolean destroy = FALSE;
        PCRMutex_Lock( ref_count_mutex );
        PCR_Assert( self->refCount > 0 );
        --(self->refCount);
        if( !self->refCount )
            destroy = TRUE;
        PCRMutex_Unlock( ref_count_mutex );
        
        if( destroy )
        {
            if( self->config )
            {
                if(self->ccdf)
                    self->ccdf(self->config);
            }
            DOFMem_Free( self );
        }
    }
}

const void * DTLConnectionConfig_GetConfig( const DTLConnectionConfig self )
{
    if( self )
        return self->config;
    return NULL;
}

uint32 DTLConnectionConfig_GetConfigSize( const DTLConnectionConfig self )
{
    if( self )
        return self->size;
    return 0;
}

DTLTransport DTLConnectionConfig_GetTransport( const DTLConnectionConfig self )
{
    if( self )
        return self->transport;
    return NULL;
}
