/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/address.h>
#include <dof/transport.h>
#include <internal/transportinternal.h>
#include <internal/dofinternal.h>

extern PCRMutex ref_count_mutex;

typedef struct DTLServerConfig_t
{
    DTLTransport                transport;
    DTLServerConfig_Destroyer   scdf;
    void*                       config;
    uint32                      size;
    uint32                      refCount;
}DTLServerConfig_t;

DTLServerConfig DTLServerConfig_Create( uint32 configSize, const void *config, DTLTransport transport, DTLServerConfig_Destroyer destroyer )
{
    DTLServerConfig ret = DOFMem_Alloc(sizeof(DTLServerConfig_t)+ configSize);

    PCR_Assert(transport);

    if( !ret )
        return NULL;
    
    ret->refCount = 1;    
    ret->scdf = destroyer;
    ret->size = configSize;
    ret->transport = transport;
    ret->config = (configSize > 0)?(void*)(((uint8*)ret) + sizeof(DTLServerConfig_t)) : NULL;
    PCRMem_Copy( ret->config, config, configSize );
    
    return ret;
}

DTLServerConfig DTLServerConfig_Copy( const DTLServerConfig self )
{
    DTLServerConfig ret = NULL;
    if( self )
    {
        PCRMutex_Lock( ref_count_mutex );
        if( self->refCount < 0xFFFFFFFF )
        {
            ++(self->refCount);
            ret = self;
        }
        PCRMutex_Unlock( ref_count_mutex );
    }
    return ret;
}

boolean DTLServerConfig_IsEqual( const DTLServerConfig self, const DTLServerConfig config )
{
    if ( !self && !config )
        return TRUE;
    if( self && config )
    {
        if( self->transport != config->transport && self->size != config->size )
            return FALSE;
        if( self->transport->fns->IsServerConfigEqual )
            return self->transport->fns->IsServerConfigEqual( self->transport, self->size, self->config, config->config );
        return (PCRMem_Compare(self->config, config->config, self->size) == 0);
    }
    return FALSE;
}

void DTLServerConfig_Destroy( DTLServerConfig self )
{
    if( self )
    {
        boolean destroy = FALSE;
        PCRMutex_Lock( ref_count_mutex );
        PCR_Assert( self->refCount > 0 );
        --(self->refCount);
        if( !self->refCount )
            destroy = TRUE;
        PCRMutex_Unlock( ref_count_mutex );
        
        if( destroy )
        {
            if( self->config )
            {
                if(self->scdf)
                    self->scdf(self->config);
            }
            DOFMem_Free( self );
        }
    }
}

const void * DTLServerConfig_GetConfig( const DTLServerConfig self )
{
    if( self )
        return self->config;
    return NULL;
}

uint32 DTLServerConfig_GetConfigSize( const DTLServerConfig self )
{
    if( self )
        return self->size;
    return 0;
}

DTLTransport DTLServerConfig_GetTransport( const DTLServerConfig self )
{
    if( self )
        return self->transport;
    return NULL;
}
