/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/address.h>
#include <dof/exception.h>
#include <dof/transport.h>
#include <internal/dofinternal.h>

#include <internal/connectionmanager.h>
#include <internal/epp_negative_one.h>
#include <internal/transportinternal.h>
#include <internal/serverinternal.h>
#include <internal/remotechannel.h>
#include <internal/remotechannelmanager.h>

extern PCRMutex ref_count_mutex;

static uint32       transport_layer_refcount = 0;
static PCRMutex     transport_layer_mutex = NULL;
static OALAVLTree   transport_layer_tree = NULL; /* stores DTLServer and DTLConnection */
static boolean      transport_layer_shutdown = FALSE;

typedef struct dtl_t
{
    union
    {
        struct
        {
            DTLServer           dtlServer;          /* Server-specific function calls. */
            OALServer           server;
        }server;

        struct
        {
            DTLConnection       dtlConnection;
            OALChannel          channel;
        }connection;
    }type;

    uint32  refCount;

    struct
    {
        unsigned int startOpenCalled        : 1; /* start or open called. */
        unsigned int startedOpenedCalled    : 1; /* started or opened called. need to notify transport if stop or closed called */
        unsigned int stopCloseCalled        : 1; /* stop or close called. do not notify channel or server anymore. notify transport if sooCalled*/
        unsigned int stoppedClosedCalled    : 1; /* stopped or closed called if !stopCalled notify terminated happened */
        unsigned int freeable               : 1; /* ref count hit zero but couldn't free because transport had not returned yet */
        unsigned int isConnection           : 1;
        unsigned int initCalled             : 1;
    }bits;
}dtl_t;

static boolean DOFTransportLayer_Opened( DTLTransportHandler handler, DTLConnection connection );
static void DOFTransportLayer_Closed( DTLTransportHandler handler, DTLConnection connection, DOFException except );
static boolean DOFTransportLayer_Started( DTLTransportHandler handler, DTLServer server );
static void DOFTransportLayer_Stopped( DTLTransportHandler handler, DTLServer server, DOFException except );
static boolean DOFTransportLayer_Received_Connection( DTLTransportHandler handler, DTLConnection connection, uint32 size, uint8 * packet, DOFAddress address, DOFAddressType atype);
static boolean DOFTransportLayer_Received_Server( DTLTransportHandler handler, DTLServer server, uint32 size, uint8 * packet, DOFAddress address, DOFAddressType atype );
static int32 DOFTransportLayer_GetRecvSize( DTLTransportHandler handler, const uint8 * buffer, uint32 offset, uint32 length );
static uint32 DOFTransportLayer_GetMaxInitialRecvSize( DTLTransportHandler handler );

static const struct DTLTransportHandlerFns_t DOFTransportLayerFuncs =
{
    DOFTransportLayer_Opened,
    DOFTransportLayer_Closed,
    DOFTransportLayer_Started,
    DOFTransportLayer_Stopped,
    DOFTransportLayer_Received_Connection,
    DOFTransportLayer_Received_Server,
    DOFTransportLayer_GetRecvSize,
    DOFTransportLayer_GetMaxInitialRecvSize
};

static const struct DTLTransportHandler_t DOFTransportLayer = { &DOFTransportLayerFuncs };

static DTLTransportHandler DOFTransportLayer_GetDTLHandler( void )
{
    return ((DTLTransportHandler)&DOFTransportLayer);
}

static boolean DOFTransportLayer_Retain(void);
static void DOFTransportLayer_Release(void);

static void DOFTransportLayer_Lock(void)
{
    PCRMutex_Lock(transport_layer_mutex);
}

static void DOFTransportLayer_Unlock(void)
{
    PCRMutex_Unlock(transport_layer_mutex);
}

/* called with lock */
static boolean DOFTransportLayer_AddDTL( DTL dtl )
{
    if( dtl )
    {
        return (OALAVLTree_Insert( transport_layer_tree, dtl, dtl, NULL) != OALAVLTree_InsertFailure);
    }
    return FALSE;
}

/* check if we can free the transport layer */
static void DOFTransportLayer_Free(void)
{
    if( transport_layer_shutdown && OALAVLTree_GetNumItems( transport_layer_tree ) == 0 )
    {
        OALAVLTree_Destroy(transport_layer_tree, NULL, NULL);
        transport_layer_tree = NULL;
    }
}

/* called with lock */
static boolean DOFTransportLayer_RemoveDTL( DTL dtl )
{
    boolean ret = FALSE;
    if( dtl )
    {
        if( OALAVLTree_Delete( transport_layer_tree, dtl ) )
        {
            DOFTransportLayer_Free();
            ret = TRUE;
        }
    }
    return ret;
}

/* called with lock */
static boolean DOFTransportLayer_IsKnow( DTL dtl )
{
    if( dtl )
    {
        DTL item = (void*)OALAVLTree_Retrieve( transport_layer_tree, dtl );
        if( item )
            return TRUE;
    }
    return FALSE;
}

/* called with lock */
static boolean DOFTransportLayer_IsOpenIO( DTL dtl )
{
    if( dtl )
    {
        if( DOFTransportLayer_IsKnow( dtl ) && dtl->bits.startedOpenedCalled && !(dtl->bits.stopCloseCalled || dtl->bits.stoppedClosedCalled) )
            return TRUE;
    }
    return FALSE;
}

/* called with lock */
static boolean DOFTransportLayer_NotifyUp( DTL dtl )
{
    return ( !dtl->bits.stopCloseCalled );
}

/* called with lock */
static boolean DOFTransportLayer_SendUp( DTL dtl )
{
    return (!( dtl->bits.stopCloseCalled || dtl->bits.stoppedClosedCalled ));
}

static void DTL_Close_DTL( DTL dtl, boolean internal )
{
    if( dtl )
    {
        if( !dtl->bits.stoppedClosedCalled && !dtl->bits.stopCloseCalled && dtl->bits.startedOpenedCalled )
        {
            if( dtl->bits.isConnection )
                dtl->type.connection.dtlConnection->fns->Close( dtl->type.connection.dtlConnection );
            else
                dtl->type.server.dtlServer->fns->Stop( dtl->type.server.dtlServer );
        }
        if( !internal )
            dtl->bits.stopCloseCalled = 1;
    }
}

/* called with lock */
static void DTL_CloseDTL( DTL dtl )
{
    DTL_Close_DTL(dtl, FALSE);
}

DTL DTL_Create( void )
{
    DTL ret = (DTL)DOFMem_Alloc( sizeof( dtl_t ) );

    if( ret )
    {
        PCRMem_Set( ret, 0, sizeof( dtl_t ) );
        DOFTransportLayer_Retain();
        ret->refCount = 1;
        return ret;
    }

    return ret;
}

static void DTL_ErrorDestroy( DTL self )
{
    if( self )
    {
        DOFTransportLayer_Release();
        DOFMem_Free(self);
    }
}

/* called with lock */
static DTL DTL_Create_InternalConnection( DOF dof, DTLConnection dtlconnection )
{
    DTL ret = DTL_Create();

    if( ret )
    {
        if( DOFTransportLayer_AddDTL(ret) )
        {
            ret->bits.initCalled = 1;
            ret->bits.isConnection = 1;
            ret->type.connection.dtlConnection = dtlconnection;
            dtlconnection->fns->SetContext( dtlconnection, ret );
            return ret;
        }
        DTL_ErrorDestroy(ret);
    }

    return ret;
}

boolean DTL_InitConnectionDTL( DTL self, DOF dof, OALChannel channel, DTLTransport transport, DTLConnectionConfig dtlConfig, DOFAddress peer, DOFConnectionType type )
{
    self->bits.isConnection = 1;
    self->type.connection.channel = channel;

    DOFTransportLayer_Lock();
    if ( transport->fns->Open )
    {
        if( DOFTransportLayer_AddDTL(self) )
        {
            if( transport->fns->Open(transport, dtlConfig, peer, DOFTransportLayer_GetDTLHandler(), type, (void*)self ) )
            {
                self->bits.startOpenCalled = 1;
                self->bits.initCalled = 1;
                DOFTransportLayer_Unlock();
                return TRUE;
            }
            else
            {
                DOFTransportLayer_RemoveDTL(self);
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DTLTransportHandler: Failed to allocate transport DTLConnection.");
            }
        }
        else
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DTLTransportHandler: Failed to add dtl to transport layer (connection init).");
    }
    else
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DTLTransportHandler: Transport does not support connections.");

    DOFTransportLayer_Unlock();
    return FALSE;
}

boolean DTL_InitServerDTL( DTL self, DOF dof, OALServer server, DTLTransport transport, DTLServerConfig dtlConfig, DOFAddress listenAddr, DOFServerType type, DTL relatedServer )
{
    self->type.server.server = server;

    DOFTransportLayer_Lock();
    if ( transport->fns->Start )
    {
        if( DOFTransportLayer_AddDTL(self) )
        {
            if( transport->fns->Start( transport, dtlConfig, listenAddr, DOFTransportLayer_GetDTLHandler(), type, (relatedServer)?relatedServer->type.server.dtlServer:NULL, (void*)self ) )
            {
                self->bits.startOpenCalled = 1;
                self->bits.initCalled = 1;
                DOFTransportLayer_Unlock();
                return TRUE;
            }
            else
            {
                DOFTransportLayer_RemoveDTL( self );
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DTLTransportHandler: Failed to allocate transport DTLServer.");
            }
        }
        else
            DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DTLTransportHandler: Failed to add dtl to transport layer (server).");
    }
    else
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "DTLTransportHandler: Transport does not support servers.");

    DOFTransportLayer_Unlock();
    return FALSE;
}

DTL DTL_Retain( DTL dtl )
{
    if( dtl )
    {
        DOFTransportLayer_Lock();
        ++dtl->refCount;
        DOFTransportLayer_Unlock();
        return dtl;
    }
    return NULL;
}

/* called with lock */
static void DTL_Release( DTL dtl )
{
    if( dtl )
    {
        if( dtl->bits.freeable )
        {
            if( dtl->bits.stoppedClosedCalled )
            {
                DOFTransportLayer_RemoveDTL( dtl );
                DOFTransportLayer_Unlock();
                DOFTransportLayer_Release();
                DOFMem_Free( dtl );
            }
            else if ( !dtl->bits.initCalled )
            {
                DOFTransportLayer_Unlock();
                DOFTransportLayer_Release();
                DOFMem_Free( dtl );
            }
            else
            {
                DTL_CloseDTL( dtl );
                DOFTransportLayer_Unlock();
            }
            return;
        }
    }
    DOFTransportLayer_Unlock();
}

static void RemoveContext( DTL dtl )
{
    if( dtl && dtl->type.connection.channel )
    {
        if( dtl->bits.isConnection )
            dtl->type.connection.channel = NULL;
        else
            dtl->type.server.server = NULL;
    }
}

void DTL_Destroy( DTL dtl )
{
    if( dtl )
    {
        DOFTransportLayer_Lock();

        --dtl->refCount;

        if(!dtl->refCount)
        {
            RemoveContext( dtl );
            dtl->bits.freeable = 1;
            DTL_Release( dtl ); /* This call will unlock the mutex */
            return;
        }

        DOFTransportLayer_Unlock();
    }
}

void DTL_StopClose(DTL dtl)
{
    DOFTransportLayer_Lock();
    RemoveContext( dtl );
    DTL_CloseDTL( dtl );
    DOFTransportLayer_Unlock();
}

/* only should be called for connections to our servers */
void DTL_SetChannel( DTL dtl, OALChannel channel )
{
    dtl->type.connection.channel = channel;
}

boolean DTL_IsConnection( DTL dtl )
{
    if( dtl )
        return dtl->bits.isConnection;
    return FALSE;
}

boolean DOFTransportLayer_Init(void)
{
    transport_layer_shutdown = FALSE;

    transport_layer_mutex = PCRMutex_Create();
    if( transport_layer_mutex )
    {
        transport_layer_tree = OALAVLTree_Create(NULL,NULL,NULL);
        if( transport_layer_tree )
        {
            transport_layer_refcount = 1;

            return TRUE;
        }
        PCRMutex_Destroy(transport_layer_mutex);
        transport_layer_mutex = NULL;
    }
    return FALSE;
}

static boolean DOFTransportLayer_Retain(void)
{
    boolean ret = FALSE;
    PCRMutex_Lock(ref_count_mutex);
    if( transport_layer_refcount < 0xFFFFFFFF )
    {
        ++transport_layer_refcount;
        ret = TRUE;
    }
    PCRMutex_Unlock(ref_count_mutex);
    return ret;
}

static void DOFTransportLayer_Release(void)
{
    DOFTransportLayer_Shutdown();
}

void DOFTransportLayer_Shutdown(void)
{
    boolean doFree = FALSE;

    PCRMutex_Lock(ref_count_mutex);
    PCR_Assert(transport_layer_refcount);
    --transport_layer_refcount;
    if( !transport_layer_refcount )
    {
        doFree = TRUE;
    }
    PCRMutex_Unlock(ref_count_mutex);

    if( doFree )
    {
        DOFTransportLayer_Lock();
        transport_layer_shutdown = TRUE;
        DOFTransportLayer_Free();
        DOFTransportLayer_Unlock();
        PCRMutex_Destroy(transport_layer_mutex);
        transport_layer_mutex = NULL;
    }
}

static boolean DOFTransportLayer_Opened( DTLTransportHandler handler, DTLConnection connection )
{
    DTL dtl;
    boolean mustDestroy = FALSE;

    PCR_Assert(handler && connection);

    DOFServer_Lock();
    DOFTransportLayer_Lock();

    dtl = (DTL)connection->fns->GetContext( connection );

    if( !dtl )
    {
        DTLServer dtlSrv = connection->fns->GetServer( connection );
        DTL sDTL = dtlSrv->fns->GetContext( dtlSrv );
        dtl = DTL_Create_InternalConnection( (DOF)DOF_GetDof(), connection );
        if( dtl )
        {
            if(!OALChannelMgr_CreateChannel_StreamServerConnect( dtl, sDTL->type.server.server, connection->fns->GetConnectionType( connection ), connection->fns->GetAddress( connection ), connection->fns->GetConnectionConfig( connection ) ))
            {
                DTL_Destroy(dtl);
                DOFTransportLayer_Unlock();
                DOFServer_Unlock();
                return FALSE;
            }
            /*  If StreamServerConnect is successfull it will retain the dtl, so destroy must be called,
                but just in case it is released by the server before this functin is done using it,
                set mustDestroy and call destroy later. */
            mustDestroy = TRUE;
        }
        else
        {
            DOFTransportLayer_Unlock();
            DOFServer_Unlock();
            return FALSE;
        }
    }

    if( !dtl->type.connection.dtlConnection )
        dtl->type.connection.dtlConnection = connection;

    DOFServer_Unlock();

    if( DOFTransportLayer_NotifyUp(dtl) )
    {
        dtl->type.connection.channel->remoteChannelData->flags.isUp = TRUE;
        dtl->type.connection.channel->remoteChannelData->dtl->bits.isConnection = TRUE;
        OALRemoteChannel_Init( (OALChannel)dtl->type.connection.channel );
    }
    else
    {
        DOFTransportLayer_Unlock();
        if (mustDestroy)
            DTL_Destroy(dtl);

        return FALSE;
    }

    dtl->bits.startedOpenedCalled = 1;

    DOFTransportLayer_Unlock();

    if (mustDestroy)
        DTL_Destroy(dtl);

    return TRUE;
}

static void DOFTransportLayer_Closed( DTLTransportHandler handler, DTLConnection connection, DOFException except )
{
    DTL dtl;

    PCR_Assert(handler && connection);

    DOFTransportLayer_Lock();

    dtl = (DTL)connection->fns->GetContext( connection );

    if( dtl )
    {
        dtl->bits.stoppedClosedCalled = 1;
        if( dtl->bits.startedOpenedCalled && !dtl->bits.stopCloseCalled )
        {
            OALRemoteChannel_NotifyTransportDown(dtl->type.connection.channel, dtl, except);
        }
        DTL_Release(dtl); /* this call will unlock */
        return;
    }

    DOFTransportLayer_Unlock();
}

static boolean DOFTransportLayer_Started( DTLTransportHandler handler, DTLServer server )
{
    DTL dtl;

    PCR_Assert(handler && server);

    DOFTransportLayer_Lock();

    dtl = (DTL)server->fns->GetContext(server);

    if( dtl )
    {
        if(!dtl->type.server.dtlServer)
            dtl->type.server.dtlServer = server;

        dtl->bits.startedOpenedCalled = 1;
        if( DOFTransportLayer_NotifyUp( dtl ) )
        {
            /* This checks transport which is a union of channel and server. */
            if( ((void*)dtl->type.server.server->transport.channel) != ((void*)dtl) )
            {
                dtl->type.server.server->transport.channel->remoteChannelData->flags.isUp = TRUE;
                OALRemoteChannel_Init( dtl->type.server.server->transport.channel );
            }
            else
            {
                DOFTransportLayer_Unlock();
                OALServer_NotifyServerStarted( dtl->type.server.server );
                return TRUE;
            }
        }
    }
    DOFTransportLayer_Unlock();

    return TRUE;
}

static void DOFTransportLayer_Stopped( DTLTransportHandler handler, DTLServer server, DOFException except )
{
    DTL dtl;

    PCR_Assert(handler && server);

    DOFTransportLayer_Lock();

    dtl = (DTL)server->fns->GetContext(server);

    if( dtl )
    {
        dtl->bits.stoppedClosedCalled = 1;
        if( DOFTransportLayer_NotifyUp(dtl) )
        {
            if( dtl->type.server.server )
            {
                if( dtl->bits.startedOpenedCalled && OALServer_IsDatagram( dtl->type.server.server ))
                {
                    OALRemoteChannel_NotifyTransportDown(dtl->type.server.server->transport.channel, dtl, except);
                }
                else
                {
                    DOFTransportLayer_Unlock();
                    OALServer_NotifyServerStopped( dtl->type.server.server, except );
                    DOFTransportLayer_Lock();
                }
            }
        }
        DTL_Release(dtl); /* this call will unlock */
        return;
    }
    DOFTransportLayer_Unlock();
}

static boolean DOFTransportLayer_Received_Connection( DTLTransportHandler handler, DTLConnection connection, uint32 size, uint8 * packet, DOFAddress address, DOFAddressType atype )
{
    OALChannel channel;
    OALStream recvStream;
    boolean ret = FALSE;
    DTL conn;

    PCR_Assert(handler && connection && address);

    DOFTransportLayer_Lock();

    conn = (DTL)connection->fns->GetContext(connection);
    if( DOFTransportLayer_SendUp( conn ) )
    {
        channel = (OALChannel)(conn->type.connection.channel);
        if( channel )
        {
            recvStream = OALStream_CreateRWWithData( packet, size );

            DOF_Inc_InputByteCount( DOF_GetDof(), size );

            if( recvStream )
            {
                switch( connection->fns->GetConnectionType(connection) )
                {
                case DOFCONNECTIONTYPE_STREAM:
                    OALRemoteChannel_ReceiveFromTransport( channel, recvStream, NULL, FALSE, TRUE );
                    ret = TRUE;
                    break;
                case DOFCONNECTIONTYPE_DATAGRAM:
                    /* the ssMulticast paramater for this call should be true for multicast and broadcast. */
                    OALRemoteChannel_ReceiveFromTransport( channel, recvStream, address, atype != DOFADDRESSTYPE_UNICAST, FALSE );
                    ret = TRUE;
                    break;
                case DOFCONNECTIONTYPE_HUB:
                case DOFCONNECTIONTYPE_POINT:
                case DOFCONNECTIONTYPE_GROUP:
                case DOFCONNECTIONTYPE_GROUP_MEMBER_ONLY:
                default:
                    break;
                }
                OALStream_Release( recvStream, TRUE );
                DOFTransportLayer_Unlock();
                return ret;
            }
        }
    }
    DOFTransportLayer_Unlock();

    PCRMem_Free( packet );

    return ret;
}

static boolean DOFTransportLayer_Received_Server( DTLTransportHandler handler, DTLServer server, uint32 size, uint8 * packet, DOFAddress address, DOFAddressType atype )
{
    OALServer srv;
    OALStream recvStream;
    DTL mydtl;

    PCR_Assert(handler && server && address);

    DOFTransportLayer_Lock();

    mydtl = (DTL)server->fns->GetContext(server);
    if( DOFTransportLayer_SendUp( mydtl ) )
    {
        srv = mydtl->type.server.server;
        if( srv )
        {
            recvStream = OALStream_CreateRWWithData( packet, size );

            if( !recvStream )
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DTLTransportHandler: Dropping packet. (out of memory)");
                goto FAIL;
            }

            DOF_Inc_InputByteCount( DOF_GetDof(), size );

            /* the ssMulticast paramater for this call should be true for multicast and broadcast. */
            OALRemoteChannel_ReceiveFromTransport( srv->transport.channel, recvStream, address, atype != DOFADDRESSTYPE_UNICAST, FALSE );

            OALStream_Release( recvStream, TRUE );
            DOFTransportLayer_Unlock();
            return TRUE;
        }
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "DTLTransportHandler: Dropping packet.");
FAIL:
    DOFTransportLayer_Unlock();

    PCRMem_Free(packet);

    return FALSE;
}

static DTL DOFTransportLayer_GetSendServerDTL( DTL dtl )
{
    if( dtl->type.server.server && dtl->type.server.server->transport.channel && dtl->type.server.server->transport.channel->remoteChannelData && dtl->type.server.server->transport.channel->remoteChannelData->baseChannel )
    {
        return dtl->type.server.server->transport.channel->remoteChannelData->baseChannel->remoteChannelData->dtl;
    }
    return dtl;
}

boolean DOFTransportLayer_SendTo(DTL transport, OALPacket data, DOFAddress remoteAddr)
{
    uint8* dataBuf = NULL;
    uint32 dataLen = 0;

    DOFTransportLayer_Lock();

    if( DOFTransportLayer_IsOpenIO( transport ) )
    {
        dataLen = OALPacket_GetTotalStreamSize( data );
        dataBuf = (uint8*)PCRMem_Alloc(dataLen);

        if(dataBuf == NULL)
        {
            DOFTransportLayer_Unlock();
            return FALSE;
        }
        OALPacket_CopyTo( data, dataBuf, 0, dataLen );

        DOF_Inc_OutputByteCount( DOF_GetDof(), dataLen );

        /* send */
        if( transport->bits.isConnection )
        {
            transport->type.connection.dtlConnection->fns->Send( transport->type.connection.dtlConnection,remoteAddr,dataBuf,0,dataLen);
            OALConnection_UpdateOutputByteCount( transport->type.connection.channel->remoteChannelData->OALConn, dataLen );
            DOFTransportLayer_Unlock();
            return TRUE;
        }
        else
        {
            DTL dtl = DOFTransportLayer_GetSendServerDTL( transport );
            DOFAddress addr = remoteAddr;

            if( !dtl )
            {
                dtl = transport;
            }
            else if( !DOFTransportLayer_IsOpenIO( dtl ) )
            {
                DOFMem_Free( dataBuf );
                DOFTransportLayer_Unlock();
                return FALSE;
            }

            if( !addr )
                addr = transport->type.server.server->config->listenAddr; /* listen address of listen server probably multicast addr */

            dtl->type.server.dtlServer->fns->Send(dtl->type.server.dtlServer,addr,dataBuf,0,dataLen);
            OALServer_UpdateOutputByteCount( dtl->type.server.server, dataLen );
            DOFTransportLayer_Unlock();
            return TRUE;
        }
    }
    else
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_DEBUG, "DTLTransportHandler: _ISOpenIO returned FALSE");
    }
    DOFTransportLayer_Unlock();

    return FALSE;
}

static int32 DOFTransportLayer_GetRecvSize( DTLTransportHandler handler, const uint8 * buffer, uint32 offset, uint32 length )
{
    /* ENP logic */
    PCR_Assert( handler );
    if( length == 0 )
        return -2;

    if( !(buffer[offset] & 0x80) ) /* no flags need two */
    {
        if( length >=2 )
            return 2;
        else
            return length - 2;
    }
    else /* flags */
    {
        switch( (buffer[offset] & 0x7F) ) /* version */
        {
        case 1:
            {
                int32 len = 0;
                if( length < 2 )
                    return length - 2;

                len = ((buffer[offset+1] & 0x03));
                if( ((uint32)len+2) <= length )
                {
                    switch( len )
                    {
                    case 0:
                        return 2;
                        break;
                    case 1:
                        return buffer[offset+2];
                        break;
                    case 2:
                        return *((uint16*)(buffer+2));
                        break;
                    case 3:
                        return ((*((uint32*)(buffer+2))) & 0x00FFFFFF);
                        break;
                    default:
                        break;
                    }
                }
                else
                {
                    return length - (len + 2);
                }
            }
            break;
        default:
            break;
        }
    }

    return 0;
}

/**
 * Get maximum initial receive size.  That is, the maximum number of bytes that
 * ever may need to be received before the packet length can be determined.
 */
static uint32 DOFTransportLayer_GetMaxInitialRecvSize( DTLTransportHandler handler )
{
    PCR_Assert( handler );
    return 2;
}

boolean DOFTransportLayer_CreateConnectionConfig( DTL transport, DOFAddress address, DTLConnectionConfig * DTLConfig )
{
    DTLConnectionConfig config;
    boolean retStatus = FALSE;

    *DTLConfig = NULL;

    DOFTransportLayer_Lock();

    if( DOFTransportLayer_IsOpenIO( transport ) )
    {
        if ( transport->bits.isConnection )
            config = transport->type.connection.dtlConnection->fns->GetConnectionConfig( transport->type.connection.dtlConnection );
        else
            config = transport->type.server.dtlServer->fns->GetConnectionConfig( transport->type.server.dtlServer, address );

        if( config )
        {
            *DTLConfig = DTLConnectionConfig_Copy(config);
        }

        if ( *DTLConfig || !config )
        {
            retStatus = TRUE;
        }
    }

    DOFTransportLayer_Unlock();

    return retStatus;
}

