/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/objectid.h>
#include <internal/trp_v1.h>
#include <internal/timer.h>
#include <internal/slist.h>
#include <internal/avl_tree.h>
#include <internal/dofinternal.h>
#include <internal/avl_treelist.h>
#include <internal/domaininternal.h>
#include <internal/connection_internal.h>
#include <internal/operationinternal.h>
#include <internal/permissionsetinternal.h>
#include <internal/exception.h>
#include "trp_shared.h"
#include "trp_v1_utils.h"
#include "trp_v1_sm.h"

typedef struct TRPRouterDomainRoute_Node_t
{
    OALSListNode_t* next; /* pointer to the next item in the list */
    OALChannel      channel;
    DOFAddress      address;
    boolean         configured;
}TRPRouterDomainRoute_Node_t;

typedef struct TRPRouterDomain_Node_t
{
    DOFObjectIDDomain   domain;             /* the domain these routes are for */
    DOFAuthenticator    localAS;            /* Local Domain to handle the AS requests */
    OALSListNode        trustedRoutes;      /* <TRPRouterDomainRoute_Node_t* node> */ /* Trusted Routes Configured by Connections Configs, No Validation Required */
    OALSListNode        currTrustedRoutes;  
    OALSListNode        routes;             /* <TRPRouterDomainRoute_Node_t* node> */ /* Validated Routes either configured on a connection or found by succesful completion on a TRP request. This will hold static and dynamic routes */
    OALSListNode        currRoutes;         
}TRPRouterDomain_Node_t;

typedef struct TRPRouter_t
{
    OALAVLTree      uniqueChannelTree;  /* <OALChannel, OALChannel> */ /* use of tree to enforce single entry of a channel */
    OALAVLTree      domains;            /* <TRPRouterDomain_Node_t* object, DOFObjectID domain> */ /* known AS routes, lookup by the wanted DOFObjectIDDomain */
    uint32          localASCount;       /* holds count of local AS */
}TRPRouter_t;

static boolean trpShutdown = TRUE;
static PCRMutex trpRouter_mtx = NULL;
static TRPRouter_t trpRouter = { NULL, NULL, 0 };

static boolean TRPRouter_Lock(void)
{
    PCRMutex_Lock( trpRouter_mtx );
    if(!trpShutdown)
        return TRUE;
    PCRMutex_Unlock( trpRouter_mtx );
    return FALSE;
}

static void TRPRouter_Unlock(void)
{
    PCRMutex_Unlock( trpRouter_mtx );
}

static void TRPRouter_ChannelShutdowObserver(OALObserver self, OALObservable observable, void* arg);

static boolean TRPRouter_GreaterThanDomainNode( OALAVLTreeKey_t keyx, OALAVLTreeItem_t itemx, void *context )
{
    DOFObjectIDDomain key = (DOFObjectIDDomain)keyx;
    TRPRouterDomain_Node_t* item = (TRPRouterDomain_Node_t*)itemx;

    return ( DOFObjectID_Compare(DOFObjectIDDomain_GetObjectID(key), DOFObjectIDDomain_GetObjectID(item->domain)) > 0 );
}

static boolean TRPRouter_EqualToDomainNode( OALAVLTreeKey_t keyx, OALAVLTreeItem_t itemx, void *context )
{
    DOFObjectIDDomain key = (DOFObjectIDDomain)keyx;
    TRPRouterDomain_Node_t* item = (TRPRouterDomain_Node_t*)itemx;

    return ( DOFObjectID_Compare(DOFObjectIDDomain_GetObjectID(key), DOFObjectIDDomain_GetObjectID(item->domain)) == 0 );
}

boolean TRPRouter_MutexInit(void)
{
    trpRouter_mtx = PCRMutex_Create();
    if(trpRouter_mtx)
        return TRUE;
    return FALSE;
}

void TRPRouter_MutexShutdown(void)
{
    PCRMutex_Destroy(trpRouter_mtx);
    trpRouter_mtx = NULL;
}

boolean TRPRouter_Init(void)
{
    if(trpShutdown)
    {
        trpRouter.uniqueChannelTree = OALAVLTree_Create(NULL, NULL, NULL);
        if( trpRouter.uniqueChannelTree )
        {
            trpRouter.domains = OALAVLTree_Create(TRPRouter_GreaterThanDomainNode, TRPRouter_EqualToDomainNode, NULL);
            if( trpRouter.domains )
            {
                trpShutdown = FALSE;
                return TRUE;
            }
            OALAVLTree_Destroy(trpRouter.uniqueChannelTree, NULL, NULL);
        }
    }
    return FALSE;
}

static void TRPRouterDomainRouteNode_Free(OALSListNode itemx, void *arg)
{
    TRPRouterDomainRoute_Node_t* item = (TRPRouterDomainRoute_Node_t*)itemx;
    DOFAddress_Destroy(item->address);
    DOFMem_Free(item);
}

static void TRPRouter_RemoveAllDomainNodes(OALAVLTreeItem_t item, void *arg)
{
    TRPRouterDomain_Node_t* node = (TRPRouterDomain_Node_t*)item;

    DOFObjectIDDomain_Destroy(node->domain);
    node->localAS = NULL;

    if( node->routes )
    {
        node->currRoutes = NULL;
        OALSList_RemoveAll(&node->routes, TRPRouterDomainRouteNode_Free, NULL);
    }
    if( node->trustedRoutes )
    {
        node->currTrustedRoutes = NULL;
        OALSList_RemoveAll(&node->trustedRoutes, TRPRouterDomainRouteNode_Free, NULL);
    }
    DOFMem_Free(node);
}

void TRPRouter_Shutdown(void)
{
    if(TRPRouter_Lock())
    {
        trpShutdown = TRUE;
        OALAVLTree_Destroy(trpRouter.uniqueChannelTree, NULL, NULL);
        trpRouter.uniqueChannelTree = NULL;
        OALAVLTree_Destroy(trpRouter.domains, TRPRouter_RemoveAllDomainNodes, NULL);
        trpRouter.domains = NULL;
        TRPRouter_Unlock();
    }
}

/* NOTE: Assumes trprouter is locked and will return with it locked */
static boolean TRPRouter_AddUniqueChannel(OALChannel channel)
{
    OALAVLTree_InsertionStatus status = OALAVLTree_Insert(trpRouter.uniqueChannelTree, channel, channel, NULL);
    if( status == OALAVLTree_InsertSuccess )
    {
        TRPRouter_Unlock();
        OALChannel_Lock(channel);
        PCR_Assert( channel->chObservable );
        if( OALObservable_AddObserver(channel->chObservable, TRPRouter_ChannelShutdowObserver, (void*)channel) )
        {
            OALChannel_Unlock(channel);
            TRPRouter_Lock();
            return TRUE;
        }
        OALChannel_Unlock(channel);
        TRPRouter_Lock();
        OALAVLTree_Delete( trpRouter.uniqueChannelTree, channel );
    }
    else if( status == OALAVLTree_AlreadyInserted )
    {
        return TRUE;
    }
    return FALSE;
}

boolean TRPRouter_KnowsAnAS(void)
{
    boolean ret = FALSE;
    if( TRPRouter_Lock() )
    {
        ret = (trpRouter.localASCount > 0 || OALAVLTree_GetNumItems(trpRouter.domains) > 0);
        TRPRouter_Unlock();
    }
    return ret;
}

boolean TRPRouter_HasAnAS(void)
{
    boolean ret = FALSE;
    if( TRPRouter_Lock() )
    {
        ret = trpRouter.localASCount > 0;
        TRPRouter_Unlock();
    }
    return ret;
}

boolean TRPRouter_AddAuthenticator( DOFAuthenticator domain )
{
    boolean retVal = FALSE;
    PCR_Assert(domain);
    if( TRPRouter_Lock() )
    {
        DOFObjectIDDomain domainID = DOFAuthenticator_GetDomainID(domain);
        TRPRouterDomain_Node_t* ret = (TRPRouterDomain_Node_t*)OALAVLTree_Retrieve(trpRouter.domains, domainID);

        if( !ret )
        {
            ret = (TRPRouterDomain_Node_t*)DOFMem_Alloc(sizeof(TRPRouterDomain_Node_t));
            if( !ret )
            {
                TRPRouter_Unlock();
                return FALSE;
            }
            PCRMem_Set(ret,0,sizeof(TRPRouterDomain_Node_t));
            ret->domain = DOFObjectIDDomain_Copy(domainID);
            if( ret->domain )
            {
                OALAVLTree_Insert(trpRouter.domains, domainID, ret, NULL);
            }
            else
            {
                DOFMem_Free( ret );
                TRPRouter_Unlock();
                return FALSE;
            }
        }
        if(!ret->localAS )
        {
            ret->localAS = domain;
            if( !ret->localAS )
            {
                TRPRouter_Unlock();
                return FALSE;
            }
            ++trpRouter.localASCount;
            retVal = TRUE;
        }
        else
        {
            DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_WARNING, "TRPRouter: Two domains were added with the same domain ID. The second will be ignored!");
        }
        TRPRouter_Unlock();
    }
    return retVal;
}

static void TRPRouter_CheckAndRemoveDomainIfEmpty(TRPRouterDomain_Node_t* node)
{
    if ( !node->routes && !node->localAS && !node->trustedRoutes )
    {
        if (OALAVLTree_Delete(trpRouter.domains, node->domain))
        {
            DOFObjectIDDomain_Destroy(node->domain);
            DOFMem_Free(node);
        }
    }
}

void TRPRouter_RemoveAuthenticator( DOFAuthenticator domain )
{
    PCR_Assert(domain);
    if( TRPRouter_Lock() )
    {
        DOFObjectIDDomain domainID = DOFAuthenticator_GetDomainID(domain);
        TRPRouterDomain_Node_t* ret = (TRPRouterDomain_Node_t*)OALAVLTree_Retrieve(trpRouter.domains, domainID);

        if( ret )
        {
            ret->localAS = NULL;
            --trpRouter.localASCount;
            TRPRouter_CheckAndRemoveDomainIfEmpty(ret);
        }
        TRPRouter_Unlock();
    }
}

/* Used for RemoveCredentials and RemoveTrustedDomain */
static boolean RoutesTraversal_fn(OALSListNode item, void *arg)
{
    TRPRouterDomainRoute_Node_t* node = (TRPRouterDomainRoute_Node_t*)item;
    DOFConnection conn = (DOFConnection)arg;
    return (( node->channel == conn->OALConn->channel )&&( node->address == conn->OALConn->channel->remoteChannelData->address ));
}

void TRPRouter_RemoveCredentials( DOFCredentials creds, DOFConnection connection )
{
    if( TRPRouter_Lock() )
    {
        TRPRouterDomain_Node_t* ret = (TRPRouterDomain_Node_t*)OALAVLTree_Retrieve(trpRouter.domains, DOFCredentials_GetDomainID(creds));

        if( ret )
        {
            TRPRouterDomainRoute_Node_t* node = (TRPRouterDomainRoute_Node_t*)OALSList_FindAndRemove( &ret->routes, NULL, RoutesTraversal_fn, connection );
            if( node )
            {
                if( node == (TRPRouterDomainRoute_Node_t*)ret->currRoutes )
                    ret->currRoutes = NULL;
                DOFAddress_Destroy( node->address );
                DOFMem_Free( node );
            }
        }
        TRPRouter_Unlock();
    }
}

boolean TRPRouter_AddCredentials( DOFCredentials creds, DOFConnection connection )
{
    if( TRPRouter_Lock() )
    {
        DOFAddress address = DOFAddress_Copy(connection->OALConn->channel->remoteChannelData->address);
        if( address )
        {
            TRPRouterDomainRoute_Node_t* node;
            DOFObjectIDDomain domainID = DOFCredentials_GetDomainID(creds);
            TRPRouterDomain_Node_t* ret = (TRPRouterDomain_Node_t*)OALAVLTree_Retrieve(trpRouter.domains, domainID);
            OALChannel channel = connection->OALConn->channel;

            if( !ret )
            {
                OALAVLTree_InsertionStatus status;
                ret = (TRPRouterDomain_Node_t*)DOFMem_Alloc(sizeof(TRPRouterDomain_Node_t));
                if( !ret )
                {
                    DOFAddress_Destroy(address);
                    TRPRouter_Unlock();
                    return FALSE;
                }
                PCRMem_Set(ret,0,sizeof(TRPRouterDomain_Node_t));
                ret->domain = DOFObjectIDDomain_Copy(domainID);
                if( !ret->domain )
                {
                    DOFMem_Free( ret );
                    DOFAddress_Destroy( address );
                    TRPRouter_Unlock();
                    return FALSE;
                }

                status = OALAVLTree_Insert(trpRouter.domains, domainID, ret, NULL);
                PCR_Assert( status != OALAVLTree_AlreadyInserted );
                if( status == OALAVLTree_InsertFailure )
                {
                    DOFObjectIDDomain_Destroy( ret->domain );
                    DOFMem_Free( ret );
                    DOFAddress_Destroy( address );
                    TRPRouter_Unlock();
                    return FALSE;
                }
            }

            node = DOFMem_Alloc(sizeof(TRPRouterDomainRoute_Node_t));
            if(node)
            {
                PCRMem_Set(node,0,(sizeof(TRPRouterDomainRoute_Node_t)));

                node->configured = TRUE;
                node->channel = channel;
                node->address = address;

                if( TRPRouter_AddUniqueChannel(channel) )
                {
                    OALSList_Insert(&ret->routes, (void*)node);
                    TRPRouter_Unlock();
                    return TRUE;
                }
                DOFMem_Free( node );
            }
            DOFAddress_Destroy(address);
        }
        TRPRouter_Unlock();
    }
    return FALSE;
}

void TRPRouter_RemoveTrustedDomain( DOFObjectIDDomain domain, DOFConnection connection)
{
    if( TRPRouter_Lock() )
    {
        TRPRouterDomain_Node_t* ret = (TRPRouterDomain_Node_t*)OALAVLTree_Retrieve(trpRouter.domains, domain);

        if( ret )
        {
            TRPRouterDomainRoute_Node_t* node = (TRPRouterDomainRoute_Node_t*)OALSList_FindAndRemove( &ret->trustedRoutes, NULL, RoutesTraversal_fn, connection );
            if( node )
            {
                if( node == (TRPRouterDomainRoute_Node_t*)ret->currTrustedRoutes )
                    ret->currTrustedRoutes = NULL;
                DOFAddress_Destroy( node->address );
                DOFMem_Free( node );
            }
        }
        TRPRouter_Unlock();
    }
}

boolean TRPRouter_AddTrustedDomain( DOFObjectIDDomain domain, OALChannel channel )
{
    if( TRPRouter_Lock() )
    {
        DOFAddress address = DOFAddress_Copy(channel->remoteChannelData->address);
        if( address )
        {
            TRPRouterDomainRoute_Node_t* node;
            TRPRouterDomain_Node_t* ret = (TRPRouterDomain_Node_t*)OALAVLTree_Retrieve(trpRouter.domains, domain);
            boolean created = ( ret == NULL );

            if( !ret )
            {
                OALAVLTree_InsertionStatus status;
                ret = (TRPRouterDomain_Node_t*)DOFMem_Alloc(sizeof(TRPRouterDomain_Node_t));
                if( !ret )
                {
                    DOFAddress_Destroy(address);
                    TRPRouter_Unlock();
                    return FALSE;
                }
                PCRMem_Set(ret,0,sizeof(TRPRouterDomain_Node_t));
                ret->domain = DOFObjectIDDomain_Copy(domain);
                if( !ret->domain )
                {
                    DOFMem_Free(ret);
                    DOFAddress_Destroy(address);
                    TRPRouter_Unlock();
                    return FALSE;
                }

                status = OALAVLTree_Insert( trpRouter.domains, domain, ret, NULL );
                PCR_Assert( status != OALAVLTree_AlreadyInserted ); /* should not already be in the tree, checked above */
                if( status == OALAVLTree_InsertFailure )
                {
                    DOFObjectIDDomain_Destroy( ret->domain );
                    DOFMem_Free(ret);
                    DOFAddress_Destroy(address);
                    TRPRouter_Unlock();
                    return FALSE;
                }
            }

            node = DOFMem_Alloc(sizeof(TRPRouterDomainRoute_Node_t));
            if( node )
            {
                PCRMem_Set(node,0,(sizeof(TRPRouterDomainRoute_Node_t)));

                node->channel = channel;
                node->address = address;

                if( TRPRouter_AddUniqueChannel(channel) )
                {
                    OALSList_Insert(&ret->trustedRoutes, (OALSListNode)node);
                    if( created )
                        ret->currTrustedRoutes = NULL; /* set to head route */
                    TRPRouter_Unlock();
                    return TRUE;
                }
                DOFMem_Free( node );
            }
            DOFAddress_Destroy(address);
        }
        TRPRouter_Unlock();
    }
    return FALSE;
}

void TRPRouter_AddDynamicRoute(DOFCredentials creds, OALChannel channel, DOFAddress address, boolean addAlways)
{
    PCR_Assert(creds);
    PCR_Assert(channel != NULL);
    if( TRPRouter_Lock() )
    {
        DOFObjectIDDomain domain = DOFCredentials_GetDomainID( creds );
        TRPRouterDomainRoute_Node_t* node;
        TRPRouterDomain_Node_t* ret = (TRPRouterDomain_Node_t*)OALAVLTree_Retrieve(trpRouter.domains, domain);

        if( !(ret && !addAlways) )
        {
            if( !ret )
            {
                ret = (TRPRouterDomain_Node_t*)DOFMem_Alloc(sizeof(TRPRouterDomain_Node_t));
                if( !ret )
                {
                    TRPRouter_Unlock();
                    return;
                }
                PCRMem_Set(ret,0,sizeof(TRPRouterDomain_Node_t));
                ret->domain = DOFObjectIDDomain_Copy(domain);
                if( !ret->domain )
                {
                    DOFMem_Free( ret );
                    TRPRouter_Unlock();
                    return;
                }
                OALAVLTree_Insert(trpRouter.domains, domain, ret, NULL);
            }

            node = DOFMem_Alloc( sizeof(TRPRouterDomainRoute_Node_t) );
            if( node )
            {
                PCRMem_Set( node, 0, sizeof(TRPRouterDomainRoute_Node_t) );

                node->address =  DOFAddress_Copy( address );
                if( node->address || !address )
                {
                    if( TRPRouter_AddUniqueChannel( channel ) )
                    {
                        DOFConnectionState state;
                        node->channel = channel;
                        node->address = address;

                        OALSList_Insert( &ret->routes, (void*)node );

                        TRPRouter_Unlock();
                        state = OALChannel_CreateConnectionState( channel, address );
                        if( state )
                        {
                            DOF_NotifyNewDomain( DOF_GetDof(), creds, state );
                            DOFConnectionState_Destroy( state );
                        }
                        return;
                    }
                    DOFAddress_Destroy( node->address );
                }
                DOFMem_Free( node );
            }
        }
        TRPRouter_Unlock();
    }
}

struct channelAndAddress_t
{
    OALChannel  channel;
    DOFAddress  address;
}channelAndAddress_t;

static boolean TRPRouter_FindInList(OALSListNode itemx, void *argx)
{
    TRPRouterDomainRoute_Node_t* item = (TRPRouterDomainRoute_Node_t*)itemx;
    struct channelAndAddress_t* arg = (struct channelAndAddress_t*)argx;

        if( !item->configured && OALChannel_IsEquivalent( item->channel, arg->channel ) && DOFAddress_IsEqual(item->address, arg->address) )
            return TRUE;
    return FALSE;
}

void TRPRouter_NotifyFailure( DOFObjectIDDomain domain, OALChannel channel, DOFAddress address )
{
    PCR_Assert(domain);
    if( TRPRouter_Lock() )
    {
        TRPRouterDomainRoute_Node_t* node;
        TRPRouterDomain_Node_t* ret = (TRPRouterDomain_Node_t*)OALAVLTree_Retrieve(trpRouter.domains, domain);

        if( ret )
        {
            struct channelAndAddress_t ctx;

            ctx.channel = channel;
            ctx.address = address;

            /* dynamic route remove */
            node = (TRPRouterDomainRoute_Node_t*)OALSList_FindAndRemove(&ret->routes, NULL, TRPRouter_FindInList, (void*)&ctx );

            if( node )
            {
                if( node == (TRPRouterDomainRoute_Node_t*)ret->currRoutes )
                    ret->currRoutes = NULL;
                DOFAddress_Destroy(node->address);
                DOFMem_Free(node);
                TRPRouter_CheckAndRemoveDomainIfEmpty(ret);
            }
        }

        TRPRouter_Unlock();
    }
}


/* rotate current as route used */
static void TRPRouter_RotateRoutes( TRPRouterDomain_Node_t* node, boolean validated )
{
    if( validated )
    {
        if( !node->currRoutes || !node->currRoutes->next )
            node->currRoutes = node->routes; /*set to head */
        else
            node->currRoutes = node->currRoutes->next;
    }
    else
    {
        if( !node->currTrustedRoutes || !node->currTrustedRoutes->next )
            node->currTrustedRoutes = node->trustedRoutes; /* set to head */
        else
            node->currTrustedRoutes = node->currTrustedRoutes->next;
    }
}

boolean TRPRouter_GetDomainRoute( DOFObjectIDDomain domain, TRPRouter_Route_t* route )
{
    boolean ret = FALSE;

    if( route )
    {
        if( TRPRouter_Lock() )
        {
            /* look for domain */
            TRPRouterDomain_Node_t* domainNode = (TRPRouterDomain_Node_t*)OALAVLTree_Retrieve(trpRouter.domains, domain);
            route->isLocal = FALSE;

            if( !domainNode )
            {
                /* Wildcard routes */
                domainNode = (TRPRouterDomain_Node_t*)OALAVLTree_Retrieve(trpRouter.domains, DOFOBJECTID_BROADCAST);
            }

            if( domainNode )
            {
                if( domainNode->localAS )
                {
                    /* local AS */
                    route->isLocal = TRUE;
                    route->route.domain = DOFAuthenticator_Copy(domainNode->localAS);
                    if( !route->route.domain )
                    {
                        TRPRouter_Unlock();
                        return FALSE;
                    }
                    ret = TRUE;
                }
                else
                {
                    TRPRouterDomainRoute_Node_t* node = NULL;

                    if( domainNode->routes )
                    {
                        TRPRouter_RotateRoutes( domainNode, TRUE );
                        node = ((TRPRouterDomainRoute_Node_t*)domainNode->currRoutes);
                    }
                    else if( domainNode->trustedRoutes )
                    {
                        TRPRouter_RotateRoutes( domainNode, FALSE );
                        node = ((TRPRouterDomainRoute_Node_t*)domainNode->currTrustedRoutes);
                    }

                    if( node )
                    {
                        route->address = DOFAddress_Copy(node->address);
                        route->route.channel = OALChannel_Retain(node->channel);
                        if( (!route->address && node->address) || !route->route.channel)
                        {
                            TRPRouter_Unlock();
                            DOFAddress_Destroy( route->address );
                            OALChannel_Release( route->route.channel );
                            route->route.channel = NULL;
                            route->address = NULL;
                            return FALSE;
                        }
                        ret = TRUE;
                    }
                }
            }
            TRPRouter_Unlock();
        }
    }

    return ret;
}

struct domTvsRoutes_t
{
    TRPRouter_Route_t*  domainRoutes;
    uint32              count;
    boolean             fail;
}domTvsRoutes_t;

static boolean TRPRouter_TraverseDomains(OALAVLTreeItem_t item, void *arg)
{
    TRPRouterDomain_Node_t* node = (TRPRouterDomain_Node_t*)item;
    struct domTvsRoutes_t* ctx = (struct domTvsRoutes_t*)arg;
    TRPRouter_Route_t* currNode = (TRPRouter_Route_t*)ctx->domainRoutes+ctx->count;

    if( node->localAS )
    {
        currNode->isLocal = TRUE;
        currNode->route.domain = DOFAuthenticator_Copy(node->localAS);
        if( !currNode->route.domain )
        {
            ctx->fail = TRUE;
            return TRUE;
        }
    }
    else
    {
        TRPRouterDomainRoute_Node_t* rnode = NULL;
        currNode->isLocal = FALSE;

        if( node->routes )
        {
            TRPRouter_RotateRoutes( node, TRUE );
            rnode = ((TRPRouterDomainRoute_Node_t*)node->currRoutes);
        }
        else if( node->trustedRoutes )
        {
            TRPRouter_RotateRoutes( node, FALSE );
            rnode = ((TRPRouterDomainRoute_Node_t*)node->currTrustedRoutes);
        }

        if( rnode )
        {
            currNode->address = DOFAddress_Copy(rnode->address);
            currNode->route.channel = OALChannel_Retain(rnode->channel);
            if( !currNode->address || !currNode->route.channel )
            {
                ctx->fail = TRUE;
                DOFAddress_Destroy( currNode->address );
                OALChannel_Release( currNode->route.channel );
                currNode->route.channel = NULL;
                currNode->address = NULL;
                return TRUE;
            }
        }
        else
        {
            PCR_Assert(FALSE); /* shouldn't hit here. Tree is not being cleaned up properly */
            return TRUE;
        }
    }
    ++ctx->count;
    return FALSE;
}

/* Get all AS routes. only for requests that don't know their domain */
TRPRouter_Route_t* TRPRouter_GetKnownRoutes( uint32* count )
{
    struct domTvsRoutes_t ctx = {NULL, 0, FALSE};
    if( TRPRouter_Lock() )
    {
        (*count) = OALAVLTree_GetNumItems(trpRouter.domains);
        if( (*count) > 0 )
        {
            ctx.domainRoutes = (TRPRouter_Route_t*)DOFMem_Alloc(sizeof(TRPRouter_Route_t) * (*count)); /* one route per known domain */
            if( ctx.domainRoutes )
            {
                OALAVLTree_Traverse(trpRouter.domains, TRPRouter_TraverseDomains, (void*)&ctx);

                if( ctx.count != (*count) || ctx.fail )
                {
                    uint32 i = 0;
                    /*error*/
                    for( ; i < ctx.count; ++i )
                    {
                        if(ctx.domainRoutes[i].isLocal)
                        {
                            DOFAuthenticator_Destroy(ctx.domainRoutes[i].route.domain);
                        }
                        else
                        {
                            DOFAddress_Destroy(ctx.domainRoutes[i].address);
                            OALChannel_Release(ctx.domainRoutes[i].route.channel);
                        }
                    }
                    DOFMem_Free(ctx.domainRoutes);
                    ctx.domainRoutes = NULL;
                    (*count) = 0;
                }
            }
            else
            {
                *count = 0;
            }
        }
        TRPRouter_Unlock();
    }
    return ctx.domainRoutes;
}

struct domTvsAS_t
{
    uint32          count;
    DOFAuthenticator*      domains;
    boolean         fail;
}domTvsAS_t;

static boolean TRPRouter_TraverseDomainsGetASs(OALAVLTreeItem_t item, void *arg)
{
    TRPRouterDomain_Node_t* node = (TRPRouterDomain_Node_t*)item;
    struct domTvsAS_t* ctx = (struct domTvsAS_t*)arg;

    if( node->localAS )
    {
        ctx->domains[ctx->count] = DOFAuthenticator_Copy(node->localAS);
        if(!ctx->domains[ctx->count])
        {
            ctx->fail = TRUE;
            return TRUE;
        }
        ++ctx->count;
    }
    return FALSE;
}

/* Get Local ASs, This is only used for Domain Discovery. */
DOFAuthenticator* TRPRouter_GetDomains( uint32* count )
{
    struct domTvsAS_t ctx = {0, NULL, FALSE};
    if( TRPRouter_Lock() )
    {
        (*count) = trpRouter.localASCount;
        if( (*count) > 0 )
        {
            ctx.domains = (DOFAuthenticator*)DOFMem_Alloc(sizeof(DOFAuthenticator) * (*count)); /* one route per known AS */
            if( ctx.domains )
            {
                OALAVLTree_Traverse(trpRouter.domains, TRPRouter_TraverseDomainsGetASs, (void*)&ctx);

                if( (*count) != ctx.count || ctx.fail )
                {
                    uint32 i = 0;

                    for( ; i<ctx.count; ++i )
                    {
                        DOFAuthenticator_Destroy(ctx.domains[i]);
                    }
                    DOFMem_Free(ctx.domains);
                    (*count) = 0;
                    ctx.domains = NULL;
                }
            }
            else
            {
                (*count) = 0;
            }
        }
        TRPRouter_Unlock();
    }
    return ctx.domains;
}

static boolean TRPRouter_FindChannelForRemoval(OALSListNode itemx, void *arg)
{
    TRPRouterDomainRoute_Node_t* item = (TRPRouterDomainRoute_Node_t*)itemx;
    OALChannel channel = (OALChannel)arg;

    if( OALChannel_IsEquivalent( item->channel, channel ) )
        return TRUE;
    return FALSE;
}

static boolean TRPRouter_FindAndRemoveChannels(OALAVLTreeItem_t itemx, void *arg)
{
    TRPRouterDomain_Node_t* item = (TRPRouterDomain_Node_t*)itemx;
    OALChannel channel = (OALChannel)arg;
    TRPRouterDomainRoute_Node_t* rem = NULL;

    /* static */
    do
    {
        rem = (TRPRouterDomainRoute_Node_t*)OALSList_FindAndRemove( (OALSListNode*)&item->routes, NULL, TRPRouter_FindChannelForRemoval, (void*)channel);
        if( rem )
        {
            DOFAddress_Destroy(rem->address);
            DOFMem_Free(rem);
        }
    }while( rem );

    /* trusted */
    do
    {
        rem = (TRPRouterDomainRoute_Node_t*)OALSList_FindAndRemove( (OALSListNode*)&item->trustedRoutes, NULL, TRPRouter_FindChannelForRemoval, (void*)channel);
        if( rem )
        {
            DOFAddress_Destroy(rem->address);
            DOFMem_Free(rem);
        }
    }while( rem );

    /* Just assume we removed the routes we are currently using */
    item->currRoutes = NULL;
    item->currTrustedRoutes = NULL;

    return FALSE;
}

static void TRPRouter_RemoveChannel(OALChannel channel)
{
    OALAVLTree_Traverse(trpRouter.domains, TRPRouter_FindAndRemoveChannels, channel);
    OALAVLTree_Delete(trpRouter.uniqueChannelTree, channel);
}

static void TRPRouter_ChannelShutdowObserver(OALObserver self, OALObservable observable, void* argx)
{
    OALChannel arg = (OALChannel)argx;

    if( TRPRouter_Lock() )
    {
        /* remove all routes associated with this channel */
        TRPRouter_RemoveChannel(arg);
        TRPRouter_Unlock();
    }
}

static boolean routeToAsChannels_internal(OALAVLTreeItem_t item, void *arg)
{
    OALChannel channel = (OALChannel)item;
    DOFOperation op = (DOFOperation)arg;
    TRP_V1_Data data = (TRP_V1_Data)op->context;

    /* if we are pass through, don't send op back to the node that sent it to us. */
    if( !( ( op->operation_type == DOFOPERATIONTYPE_TRP_V1_PASS ) &&
           ( OALChannel_IsEquivalent( data->RequestData.passData->passOnChannel, channel ) ||
           ( ( data->RequestData.passData->address && channel->remoteChannelData->address && DOFAddress_IsEqual( data->RequestData.passData->address, channel->remoteChannelData->address ) ) ) ) ) )
    {
        OALChannel temp = OALChannel_Retain(channel);
        if(temp)
        {
            DOFAddress_Destroy( op->sources->address );
            OALChannel_Release( op->sources->channel );
            op->sources->address = NULL;
            op->sources->channel = temp;
            op->sources->channel->routeToProvider( op->sources->channel, op, NULL, op->sources->address );
        }
    }
    return FALSE;
}

static boolean routeToLocalASs_internal(OALAVLTreeItem_t item, void *arg)
{
    TRPRouterDomain_Node_t* node = (TRPRouterDomain_Node_t*)item;
    DOFOperation op = (DOFOperation)arg;

    if( node->localAS )
    {
        TRP_V1_RouteSendLocalAS( op, node->localAS );
    }
    return FALSE;
}

boolean TRPRouter_RouteToAS(DOFOperation op, OALChannel channel, DOFAddress address, boolean force)
{
    TRP_V1_Data_t* data = op->context;
    boolean ret = FALSE;
    DOFOperation realOp = op;

    DOFOperation_Lock(realOp);

    if(  TRPRouter_Lock() )
    {
        if( !force )
        {
            if( !DOFObjectID_IsBroadcast( DOFObjectIDDomain_GetObjectID( data->secDomain ) ) )
            {
                /* known domain */
                TRPRouter_Route_t route;

                boolean foundRoute = TRPRouter_GetDomainRoute(data->secDomain, &route);

                if( foundRoute )
                {
                    if( !route.isLocal )
                    {
                        DOFOperation_Lock(op);
                        OALChannel_Release( op->sources->channel );

                        op->sources->channel = route.route.channel; /* retained in the route info */
                        op->sources->channel->routeToProvider( op->sources->channel, op, NULL, route.address );
                        DOFOperation_Unlock(op);
                        DOFAddress_Destroy( route.address );
                        ret = TRUE;
                    }
                    else
                    {
                        data->domain = route.route.domain;
                        TRP_V1_RouteSendLocalAS( op, data->domain );
                        ret = TRUE;
                    }

                }
                else
                {
                    if( channel )
                    {
                        OALChannel channel_temp = OALChannel_Retain(channel);
                        DOFAddress address_temp = DOFAddress_Copy(address);
                        if( channel_temp && (!address || address_temp))
                        {
                            DOFOperation_Lock(op);
                            DOFAddress_Destroy( op->sources->address );
                            OALChannel_Release( op->sources->channel );

                            data->flags.bits.passUsed = TRUE;
                            op->sources->address = address_temp;
                            op->sources->channel = channel_temp;
                            op->sources->channel->routeToProvider( op->sources->channel, op, NULL, op->sources->address );
                            DOFOperation_Unlock(op);
                            ret = TRUE;
                        }
                        else
                        {
                            DOFAddress_Destroy( address_temp );
                            OALChannel_Release( channel_temp );
                        }
                    }
                }
            }
            else
            {
                data->flags.bits.isDomainDiscovery = 1;
                /* Remote ASs */
                if( OALAVLTree_GetNumItems(trpRouter.domains) )
                {
                    /* all known channels */
                    OALAVLTree_Traverse(trpRouter.uniqueChannelTree, routeToAsChannels_internal, (void*)op);
                    ret = TRUE;
                }

                /* Local ASs */
                if(trpRouter.localASCount > 0)
                {
                    OALAVLTree_Traverse(trpRouter.domains, routeToLocalASs_internal, op);
                }

                /* Fallback AS */
                if( channel )
                {
                    OALChannel channel_temp = OALChannel_Retain( channel );
                    DOFAddress address_temp = DOFAddress_Copy( address );
                    if( channel_temp && (!address || address_temp))
                    {
                        DOFOperation_Lock(op);
                        DOFAddress_Destroy( op->sources->address );
                        OALChannel_Release( op->sources->channel );

                        data->flags.bits.passUsed = TRUE;
                        op->sources->address = address_temp;
                        op->sources->channel = channel_temp;
                        op->sources->channel->routeToProvider( op->sources->channel, op, NULL, op->sources->address );
                        DOFOperation_Unlock(op);
                        ret = TRUE;
                    }
                    else
                    {
                        DOFAddress_Destroy( address_temp );
                        OALChannel_Release( channel_temp );
                    }
                }
            }
        }
        else if( channel )
        {
            OALChannel channel_temp = OALChannel_Retain( channel );
            DOFAddress address_temp = DOFAddress_Copy( address );
            DOFOperation_Lock(op);
            if( channel_temp && (!address || address_temp))
            {
                DOFAddress_Destroy( op->sources->address );
                OALChannel_Release( op->sources->channel );

                data->flags.bits.passUsed = TRUE;
                op->sources->address = address_temp;
                op->sources->channel = channel_temp;
                op->sources->channel->routeToProvider( op->sources->channel, op, NULL, op->sources->address );
                DOFOperation_Unlock(op);
                ret = TRUE;
            }
            else
            {
                DOFAddress_Destroy( address_temp );
                OALChannel_Release( channel_temp );
            }
        }
        TRPRouter_Unlock();
        DOFOperation_Unlock(realOp);
    }
    return ret;
}

