/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/system.h>
#include <internal/exception.h>
#include <internal/permissionsetinternal.h>
#include <internal/operationinternal.h>
#include <internal/enqueue.h>
#include <internal/blackboarddata.h>
#include <internal/connection_internal.h>
#include <internal/trp_v1.h>
#include <internal/slist.h>
#include <internal/srp.h>
#include <internal/domaininternal.h>
#include "trp_v1_sm.h"
#include "trp_v1_utils.h"

static OALEnqueue TRP_SM_enqueue = NULL;

static boolean TRP_Running = FALSE;
static PCRMutex TRP_Mutex = NULL;

boolean TRP_V1_SM_Create( void )
{
    TRP_SM_enqueue = OALEnqueue_Create(DOF_GetDof()->threadPool);
    if(TRP_SM_enqueue != NULL)
    {
        if(TRP_V1_InitRetryTimer())
        {
            if(TRPRouter_Init())
            {
                TRP_Running = TRUE;
                return TRUE;
            }
            TRP_V1_DestroyRetryTimer();
        }
        OALEnqueue_Destroy(TRP_SM_enqueue);
    }
    return FALSE;
}

boolean TRP_V1_IsRunning(void)
{
    boolean ret = FALSE;
    PCRMutex_Lock(TRP_Mutex);
    ret = TRP_Running;
    PCRMutex_Unlock(TRP_Mutex);
    return ret;
}

void TRP_V1_Shutdown1(void)
{
    PCRMutex_Lock(TRP_Mutex);
    TRP_Running = FALSE;
    PCRMutex_Unlock(TRP_Mutex);
}

boolean TRP_V1_Init(void)
{
    TRP_Mutex = PCRMutex_Create();
    if(TRP_Mutex )
        return TRUE;
    return FALSE;
}

void TRP_V1_Shutdown(void)
{
    PCRMutex_Destroy(TRP_Mutex);
    TRP_Mutex = NULL;
}

void TRP_V1_SM_Destroy()
{
    OALEnqueue_Destroy(TRP_SM_enqueue);
    TRP_SM_enqueue = NULL;
    TRP_V1_DestroyRetryTimer();
    TRPRouter_Shutdown();
}

void TRP_V1_SM_Call(DOFOperation op )
{
    OperationContext_t * node = (OperationContext_t*) DOFMem_Alloc(sizeof(OperationContext_t));

    if(node != NULL)
    {
        OALEnqueueItem_Init((OALEnqueueItem) node, TRP_V1_SM_Receive);
        node->op = op;
        node->domain = NULL;
        node->doFree = TRUE;
        if(!OALEnqueue_Enqueue(TRP_SM_enqueue, (OALEnqueueItem) node))
        {
            DOFMem_Free(node);
        }
    }
}

void TRP_V1_RouteSendLocalAS( DOFOperation op, DOFAuthenticator domain )
{
    OperationContext_t * node = (OperationContext_t*) DOFMem_Alloc(sizeof(OperationContext_t));

    if(node != NULL)
    {
        OALEnqueueItem_Init((OALEnqueueItem) node, TRP_V1_SM_LocalASReceive);
        node->domain = DOFAuthenticator_Copy(domain);
        if( node->domain )
        {
            node->op = (DOFOperation_Retain(op), op);
            node->doFree = TRUE;
            if( OALEnqueue_Enqueue(TRP_SM_enqueue, (OALEnqueueItem) node) )
            {
                return;
            }
            DOFAuthenticator_Destroy(domain);
            DOFOperation_Destroy( op );
        }
        DOFMem_Free(node);
    }
}

void TRP_V1_RouteSendLocal( DOFOperation op )
{
    OperationContext_t * node = (OperationContext_t*) DOFMem_Alloc(sizeof(OperationContext_t));

    if(node != NULL)
    {
        OALEnqueueItem_Init((OALEnqueueItem) node, TRP_V1_SM_Receive);
        node->op = op;
        node->domain = NULL;
        node->doFree = TRUE;
        if(!OALEnqueue_Enqueue(TRP_SM_enqueue, (OALEnqueueItem) node))
        {
            DOFMem_Free(node);
        }
    }
}

void TRP_V1_RouteReceiveLocal( DOFOperation op )
{
    OperationContext_t * node = (OperationContext_t*) DOFMem_Alloc(sizeof(OperationContext_t));

    if(node != NULL)
    {
        OALEnqueueItem_Init((OALEnqueueItem) node, TRP_V1_SM_Receive);
        node->op = op;
        node->domain = NULL;
        node->doFree = TRUE;
        if(!OALEnqueue_Enqueue(TRP_SM_enqueue, (OALEnqueueItem) node))
        {
            DOFMem_Free(node);
        }
    }
}

/* This function will accept a TRP error code and map that error code to a DOFError. */
DOFError TRP_V1_GetDOFErrorFromErrorCode( uint8 errorCode )
{
    DOFError errorMap[] = { DOFERROR_OK,                      /*  TRP_V1_SUCCESS                              0L      */
                            DOFERROR_FAILED,                  /*  TRP_V1_PARSE_ERROR                          1L      */
                            DOFERROR_ACCESS_DENIED,           /*  TRP_V1_ACCESS_DENIED                        2L      */
                            DOFERROR_UNKNOWN_INITIATOR,       /*  TRP_V1_UNKNOWN_INITIATOR                    3L      */
                            DOFERROR_UNKNOWN_RESPONDER,       /*  TRP_V1_UNKNOWN_RESPONDER                    4L      */
                            DOFERROR_UNKNOWN_DOMAIN,          /*  TRP_V1_UNKNOWN_DOMAIN                       5L      */
                            DOFERROR_SERVICE_UNAVAILABLE,     /*  TRP_V1_SERVICE_UNAVAILABLE                  6L      */
                            DOFERROR_BAD_MODE,                /*  TRP_V1_BAD_MODE                             7L      */
                            DOFERROR_FAILED,                  /*  TRP supposedly cannot return this errorCode 8.      */
                            DOFERROR_NOT_IN_GROUP,            /*  TRP_V1_INITIATOR_NOT_GROUP_MEMBER           9L      */
                            DOFERROR_UNSUPPORTED_INITIATOR,   /*  TRP_V1_UNSUPPORTED_IDENTIFICATION_INITIATOR 10L     */
                            DOFERROR_UNSUPPORTED_RESPONDER,   /*  TRP_V1_UNSUPPORTED_IDENTIFICATION_RESPONDER 11L     */
                            DOFERROR_INVALID_INITIATOR,       /*  TRP_V1_INVALID_IDENTIFICATION_INITIATOR     12L     */
                            DOFERROR_INVALID_RESPONDER,       /*  TRP_V1_INVALID_IDENTIFICATION_RESPONDER     13L     */
                            DOFERROR_RESOLUTION_FAILED,       /*  TRP_V1_RESOLUTION_FAILED                    14L     */
                            DOFERROR_VALIDATION_FAILED};      /*  TRP_V1_VALIDATION_FAILED                    15L     */
                                                              /*  TRP_V1_INTERNAL_ERROR                       127L    */

    /* Caller MUST understand that we map interesting param value TRP_V1_INTERNAL_ERROR to DOFERROR_FAILED. */
    if (errorCode > TRP_V1_VALIDATION_FAILED)
        return DOFERROR_FAILED;

    return errorMap[errorCode];
}

TRP_V1 TRP_V1_Create(BlackboardData blackboardData, boolean isConnectionless)
{
    TRP_V1 ret = (TRP_V1) DOFMem_Alloc(sizeof(TRP_V1_t));

    if(ret)
    {
        ret->blackboardData = blackboardData;
        ret->isConnectionless = isConnectionless;
    }
    else
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "TRP V1: Failed to create TRP Operation.");

    return ret;
}

void TRP_V1_Destroy(TRP_V1 self)
{
    if( self )
        DOFMem_Free(self);
}

boolean TRP_V1_TriggerPossibleCacheReset( DOFCredentials creds, boolean initiator, uint8 TRPErrorCode )
{
    if( DOFCredentials_IsCached( creds ) )
    {
        if( ( initiator && TRPErrorCode == TRP_V1_UNKNOWN_INITIATOR ) || 
            ( !initiator && TRPErrorCode == TRP_V1_UNKNOWN_RESPONDER ) )
        {
            DOFCredentials_Reset( creds );
            return TRUE;
        }
    }
    return FALSE;
}
