/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <internal/securityscope.h>
#include <internal/marshal.h>
#include <internal/domaininternal.h>
#include <internal/credentialsinternal.h>
#include <internal/permissionsetinternal.h>
#include "key_hash.h"
#include "trp_v1_utils.h"

static uint32 currTimestamp = 0;
static uint16 sequence[8] = {0,0,0,0,0,0,0,0};

static uint8 TRP_V1_AS_DOFErrorToTRPRejectCode( DOFError status, boolean initiator )
{
    uint8 errorCode;

    if( status >= DOFERROR_ACCESS_DENIED ||
        status <= DOFERROR_VALIDATION_FAILED )
    {
        errorCode = (status & 0x6FFF);
    }
    else
    {
        errorCode = TRP_V1_INTERNAL_ERROR;
    }

    if( !initiator )
    {
        if( errorCode == TRP_V1_UNKNOWN_INITIATOR ||
            errorCode == TRP_V1_UNSUPPORTED_IDENTIFICATION_INITIATOR ||
            errorCode == TRP_V1_INVALID_IDENTIFICATION_INITIATOR )
        {
            ++errorCode;
        }
    }

    return errorCode;
}

static void get_seq(uint32 sid, uint16* seq, uint32* timestamp)
{
    *timestamp = (PCRTime_GetMS()/1000);

    if( *timestamp != currTimestamp )
    {
        currTimestamp = *timestamp;
    }
    else
    {
        /* sets next sequence number and mods it so it won't be larger than 10 bits */
        sequence[sid & 0x7] = (sequence[sid & 0x7]+1) & 0x3FF;
    }

    *seq = sequence[sid & 0x7];
}

static uint8 WriteSIDToBuffer( uint32 sid, uint8 offset, uint8* buffer )
{
    if( sid <= 0x000000FF )
    {
        buffer[offset] = (uint8)sid;
        return 1;
    }
    else if( sid <= 0x0000FFFF )
    {
        buffer[offset+1] = (uint8)(sid & 0xFF);
        buffer[offset] = (uint8)(sid >> 8);
        return 2;
    }
    else if( sid <= 0x00FFFFFF )
    {
        buffer[offset+2] = (uint8)(sid & 0xFF);
        buffer[offset+1] = (uint8)((sid >> 8) & 0xFF);
        buffer[offset] = (uint8)(sid >> 16);
        return 3;
    }
    else
    {
        buffer[offset+3] = (uint8)(sid & 0xFF);
        buffer[offset+2] = (uint8)((sid >> 8) & 0xFF);
        buffer[offset+1] = (uint8)((sid >> 16) & 0xFF);
        buffer[offset] = (uint8)(sid >> 24);
        return 4;
    }
}

static DOFObjectIDSource SID_Create(uint32 sid, uint8 asID)
{
    uint16 seq = 0;
    uint32 time = 0;
    uint8 position = 0;
    uint8 bytes[10]= {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};

    get_seq(sid,&seq,&time);

    position += WriteSIDToBuffer(sid, 0, bytes);

    bytes[position] = asID << 2;
    bytes[position++] |= (seq >> 8) & 0x3;
    bytes[position++] = seq & 0xFF;
    bytes[position++] = time >> 24;
    bytes[position++] = (time >> 16) & 0xFF;
    bytes[position++] = (time >> 8) & 0xFF;
    bytes[position++] = time & 0xFF;

    return (DOFObjectIDSource)DOFObjectID_Create(DOFOBJECTIDCLASS_AS,position,(const uint8*)bytes);
}

static DOFPermissionSet SIDPermissionSet_Create(uint32 sid, uint8 asID)
{
    DOFPermissionSetBuilder builder = DOFPermissionSetBuilder_Create();

    if( builder )
    {
        DOFObjectIDSource theID = SID_Create(sid,asID);
        if( theID )
        {
            DOFPermission perm = DOFPermissionIAm_Create(theID);
            if( perm )
            {
                if( !DOFPermissionSetBuilder_AddPermission(builder, perm) )
                {
                    DOFPermissionSetBuilder_Destroy( builder );
                    builder = NULL;
                }
                DOFPermission_Destroy(perm);
            }
            DOFObjectIDSource_Destroy(theID);
        }
    }
    return DOFPermissionSetBuilder_BuildAndDestroy(builder);
}

static void DOFAuthenticatorNode_Release(DOFAuthenticatorNode node)
{
    if( node )
        node->fns->Release( node );
}

static void shared_key_mac(DOFObjectID domain, DOFCredentials credential, uint32 i_length, uint8* i, uint32 r_length, uint8* r, DOFAuthenticatorNode node, OALSecureHMACContext *context, const uint8 *a, uint32 a_length, const uint8 *session_key, uint8 *mac_output)
{
    OALSecureHMAC_Start(context, (uint8*)DOFCredentials_GetKeyFromStorage( credential,node->fns->GetCredentialsStorage(node)));
    OALSecureHMAC_Digest(context, (domain ? DOFObjectID_GetBytesSize( domain ) : 0), DOFObjectID_GetBytes(domain));
    OALSecureHMAC_Digest(context, i_length, i);
    OALSecureHMAC_Digest(context, r_length, r);
    OALSecureHMAC_Digest(context, a_length, a);
    OALSecureHMAC_Digest(context, 32, session_key);
    OALSecureHMAC_Finish(context, mac_output);
}

static boolean ValidatePermissionsRequestAgainstGranted(DOFPermissionSet permission_set, DOFAuthenticatorNodePermissionSetMap perm_map, uint32 sid, uint8 asID)
{
    DOFPermissionSet sids;

    if(!DOFPermissionSet_StartBuildingValidation(permission_set))
        return FALSE;

    if(!DOFPermissionSet_AddIDToValidation(permission_set, perm_map))
        return FALSE; /* failed to add */

    sids = SIDPermissionSet_Create(sid,asID);
    if( sids )
    {
        if( !DOFPermissionSet_AddALLScopesToValidation( permission_set, sids ) )
        {
            DOFPermissionSet_Destroy(sids);
            return FALSE; /* failed to add */
        }
        DOFPermissionSet_Destroy(sids);
    }

    if( DOFPermissionSet_FinishBuildingValidation( permission_set ) )
        return TRUE;

    return FALSE;
}

static OALStream TRP_V1_AS_CreateReject( uint8 errorCode )
{
    OALStream  ret = OALStream_CreateRW( 2 );

    if( ret && OALStream_Write8( ret, TRP_V1_OP_REJECT ) && OALStream_Write8( ret, errorCode ))
    {
        return ret;
    }
    OALStream_Release( ret, FALSE );
    return NULL;
}

static DOFPermissionSet createGroupPermissionSet( DOFAuthenticatorNodePermissionSetMap perms )
{
    DOFPermissionSet ret = NULL;
    uint32 i = 0;
    uint32 cnt = 0;
    boolean fail = FALSE;

    for( ; i<perms->count; ++i )
    {
        cnt += DOFPermissionSet_GetPermissionCount( perms->list[i].permissionset );
    }

    ret = DOFPermissionSet_CreateInternal(cnt,FALSE);

    if( ret )
    {
        for(i=0; i<perms->count; ++i )
        {
            if( !DOFPermissionSet_AddPermissionsFromSet( ret, perms->list[i].permissionset ) )
            {
                fail = TRUE;
                break;
            }
        }

        if( !fail )
        {
            return ret;
        }

        DOFPermissionSet_Destroy( ret );
    }

    return NULL;
}

static DOFSecurityScope TRP_V1_AS_CreateSecurityScope( DOFObjectIDDomain domain, DOFAuthenticatorNode node )
{
    DOFSecurityScope ret = NULL;
    OALPermissionScopeBuilder psb = OALPermissionScopeBuilder_Create();

    if( psb )
    {
        DOFAuthenticatorNodePermissionSetMap psm = node->fns->GetPermissions( node );
        uint32 i = 0;
        boolean fail = FALSE;

        for(; i<psm->count; ++i)
        {
            if( !OALPermissionScopeBuilder_AddID(psb, psm->list[i].id) )
            {
                fail = TRUE;
                break;
            }
        }

        if( !fail )
        {
            ret = DOFSecurityScope_Create(domain, psb->idCount, psb->ids, 0, NULL);
            OALPermissionScopeBuilder_Destroy(psb);
            return ret;
        }

        OALPermissionScopeBuilder_Destroy( psb );
    }
    return NULL;
}

boolean TRP_V1_AS_RequestKEK( TRP_V1_Data data )
{
    DOFAuthenticatorStorage storage = DOFAuthenticator_GetStorage(data->domain);

    if(!storage || !storage->fns->IsStorageAvailable(storage))
    {
        data->errorCode = TRP_V1_SERVICE_UNAVAILABLE;
    }
    else
    {
        DOFAuthenticatorNode i_node;
        DOFAuthenticatorNode g_node;
        DOFError status = DOFERROR_OK;

        i_node = storage->fns->GetNode( storage, DOFCredentials_GetIdentity(data->RequestData.kekData->initKeyRequest->credentials), DOFCredentials_GetType(data->RequestData.kekData->initKeyRequest->credentials), DOFAUTHENTICATORSECURITYTYPE_MULTI_POINT, 0, NULL, data->RequestData.kekData->initKeyRequest->permission_set, &status );
        if(!i_node)
        {
            data->errorCode = TRP_V1_AS_DOFErrorToTRPRejectCode( status, TRUE );
        }
        else
        {
            if(!i_node->fns->IsMember( i_node, (DOFObjectIDAuthentication)data->RequestData.kekData->groupID ))
            {
                data->errorCode = TRP_V1_INITIATOR_NOT_GROUP_MEMBER;
            }
            else
            {
                g_node = storage->fns->GetGroupNode( storage, (DOFObjectIDAuthentication)data->RequestData.kekData->groupID );
                if(!g_node)
                {
                    data->errorCode = TRP_V1_UNKNOWN_RESPONDER;
                }
                else
                {
                    DOFPermissionSet g_permission_set;
                    DOFSecurityScope i_node_scope = TRP_V1_AS_CreateSecurityScope(DOFAuthenticator_GetDomainID(data->domain), i_node);
                    if( i_node_scope )
                    {
                        DOFSecurityScope g_node_scope = TRP_V1_AS_CreateSecurityScope(DOFAuthenticator_GetDomainID(data->domain), g_node);
                        if( g_node_scope )
                        {
                            DOFSecurityScope compatible_scope = DOFSecurityScope_CreateCompatibleScope( i_node_scope, g_node_scope);
                            if(!compatible_scope)
                            {
                                data->errorCode = TRP_V1_INTERNAL_ERROR;
                            }
                            else
                            {
                                if(!ValidatePermissionsRequestAgainstGranted(data->RequestData.kekData->initKeyRequest->permission_set, i_node->fns->GetPermissions(i_node), i_node->fns->GetSourceIdentifier(i_node), DOFAuthenticator_GetID(data->domain)))
                                {
                                    data->errorCode = TRP_V1_INTERNAL_ERROR;
                                }
                                else
                                {
                                    g_permission_set = createGroupPermissionSet(g_node->fns->GetPermissions(g_node));
                                    if(!g_permission_set)
                                    {
                                        data->errorCode = TRP_V1_INTERNAL_ERROR;
                                    }
                                    else
                                    {
                                        if(!ValidatePermissionsRequestAgainstGranted(g_permission_set, g_node->fns->GetPermissions(g_node), g_node->fns->GetSourceIdentifier(g_node), DOFAuthenticator_GetID(data->domain)))
                                        {
                                            data->errorCode = TRP_V1_INTERNAL_ERROR;
                                        }
                                        else
                                        {
                                            DOFObjectIDSource sid = SID_Create(g_node->fns->GetSourceIdentifier(g_node), DOFAuthenticator_GetID(data->domain));
                                            uint16 max = g_node->fns->GetGroupMaxKEKPeriod(g_node) / 8;
                                            if ( max & 0xff00 )
                                                max = 0xff; /* Maximum KEK period is 1 byte. */
                                            if( !sid )
                                            {
                                                data->errorCode = TRP_V1_INTERNAL_ERROR;
                                            }
                                            else
                                            {
                                                uint32 stream_size = 1+64+1+1+1+2;

                                                stream_size += DOFObjectID_GetBytesSize(DOFObjectIDSource_GetObjectID(sid));
                                                stream_size += DOFSecurityScope_GetMarshalSize(compatible_scope);
                                                stream_size += g_node->fns->GetGroupSecurityModeSize(g_node);
                                                stream_size += OALMarshal_GetCompressedValueSize(4, g_node->fns->GetGroupStateID(g_node));
                                                stream_size += DOFPermissionSet_MarshalRequestSize(g_permission_set);
                                                stream_size += DOFPermissionSet_MarshalValidationSize(g_permission_set);
                                                stream_size += DOFPermissionSet_MarshalValidationSize(data->RequestData.kekData->initKeyRequest->permission_set);

                                                data->A = OALStream_CreateRW(stream_size);
                                                if(!data->A)
                                                {
                                                    data->errorCode = TRP_V1_INTERNAL_ERROR;
                                                }
                                                else
                                                {
                                                    OALSecureHMACContext hmac;
                                                    DOFObjectID domain_id;
                                                    uint8* write_ptr = OALStream_GetWritePtr(data->A);

                                                    /* marshal trp opcode */
                                                    if( OALStream_Write8(data->A, TRP_V1_OP_KEK) )
                                                    {
                                                        OALStream_AdjustWriteOffset(data->A, OALKeyGrant_MarshalSize(NULL));

                                                        /* A */
                                                        if( OALStream_Write8(data->A, g_node->fns->GetGroupHeartbeatPeriod(g_node)) &&
                                                            OALStream_Write8(data->A, g_node->fns->GetGroupMinKEKPeriod(g_node)) &&
                                                            OALStream_Write8(data->A, max) &&
                                                            OALStream_Write16(data->A, g_node->fns->GetGroupEpoch(g_node)) &&
                                                            DOFObjectID_Marshal(DOFObjectIDSource_GetObjectID(sid), data->A ) &&
                                                            DOFSecurityScope_Marshal(compatible_scope, data->A) &&
                                                            OALStream_Write(data->A, g_node->fns->GetGroupSecurityModeSize(g_node),g_node->fns->GetGroupSecurityMode(g_node)) &&
                                                            (OALMarshal_StreamCompress4(data->A, g_node->fns->GetGroupStateID(g_node)) == DOFERROR_OK) &&
                                                            DOFPermissionSet_MarshalRequest(g_permission_set, data->A) &&
                                                            DOFPermissionSet_MarshalValidation(g_permission_set, data->A) &&
                                                            DOFPermissionSet_MarshalValidation(data->RequestData.kekData->initKeyRequest->permission_set, data->A) )
                                                        {
                                                            /* now that everything is marshalled it's time to compute the kek */
                                                            OALSecureHMAC_Start(&hmac, (uint8*)DOFCredentials_GetKeyFromStorage( data->RequestData.kekData->initKeyRequest->credentials, i_node->fns->GetCredentialsStorage(i_node)));
                                                            domain_id = DOFObjectIDDomain_GetObjectID( DOFAuthenticator_GetDomainID(data->domain) );
                                                            OALSecureHMAC_Digest(&hmac, (domain_id ? DOFObjectID_GetBytesSize(domain_id) : 0), DOFObjectID_GetBytes(domain_id));
                                                            OALSecureHMAC_Digest(&hmac, data->I_len, data->I);
                                                            OALSecureHMAC_Digest(&hmac, OALStream_GetWriteOffset(data->A)-(1+64), write_ptr+1+64);
                                                            OALSecureHMAC_Digest(&hmac, g_node->fns->GetCredentialsStorageSize( g_node ), g_node->fns->GetCredentialsStorage( g_node ));
                                                            OALSecureHMAC_Finish(&hmac, write_ptr+1);

                                                            shared_key_key(DOFCredentials_GetKeyFromStorage( data->RequestData.kekData->initKeyRequest->credentials, i_node->fns->GetCredentialsStorage(i_node)), write_ptr+1, g_node->fns->GetCredentialsStorage(g_node), write_ptr+33);
                                                        }
                                                        else
                                                            data->errorCode = TRP_V1_INTERNAL_ERROR;
                                                    }
                                                    else
                                                        data->errorCode = TRP_V1_INTERNAL_ERROR;
                                                }
                                                DOFObjectIDSource_Destroy(sid);
                                            }
                                        }
                                        DOFPermissionSet_Destroy(g_permission_set);
                                    }
                                }
                                DOFSecurityScope_Destroy(compatible_scope);
                            }
                            DOFSecurityScope_Destroy(g_node_scope);
                        }
                        else
                            data->errorCode = TRP_V1_INTERNAL_ERROR;
                        DOFSecurityScope_Destroy(i_node_scope);
                    }
                    else
                        data->errorCode = TRP_V1_INTERNAL_ERROR;
                    DOFAuthenticatorNode_Release(g_node);
                }
            }
            DOFAuthenticatorNode_Release(i_node);
        }
    }
    if( data->errorCode != 0 )
        data->A = TRP_V1_AS_CreateReject( data->errorCode );

    return TRUE;
}

boolean TRP_V1_AS_RequestRandom( TRP_V1_Data data )
{
    DOFAuthenticatorStorage storage = DOFAuthenticator_GetStorage(data->domain);

    if(!storage || !storage->fns->IsStorageAvailable(storage))
    {
        data->errorCode = TRP_V1_SERVICE_UNAVAILABLE;
    }
    else
    {
        DOFError status = DOFERROR_OK;
        DOFAuthenticatorNode i_node = storage->fns->GetNode( storage, DOFCredentials_GetIdentity(data->RequestData.randomData->initKeyRequest->credentials), DOFCredentials_GetType( data->RequestData.randomData->initKeyRequest->credentials ), DOFAUTHENTICATORSECURITYTYPE_UNDEFINED, 0, NULL, data->RequestData.randomData->initKeyRequest->permission_set, &status );

        if(!i_node)
        {
            data->errorCode = TRP_V1_AS_DOFErrorToTRPRejectCode( status, TRUE );
        }
        else
        {
            uint32 len = 1 + OALKeyGrant_MarshalSize(NULL);

            data->A = OALStream_CreateRW( len );
            if(!data->A)
            {
                data->errorCode = TRP_V1_INTERNAL_ERROR;
            }
            else
            {
                uint8 *write_ptr;
                OALSecureHMACContext hmac;
                uint8 random_key[32];
                const uint8 *i_secret;
                DOFObjectID domain = DOFObjectIDDomain_GetObjectID( DOFAuthenticator_GetDomainID( data->domain ) );

                if(!(DOFAuthenticator_IsDebug(data->domain) && DOFAuthenticator_GetDebugKey(data->domain)))
                {
                    PCRSecureRandom_Generate(32, random_key);
                }
                else
                {
                    PCRMem_Copy(random_key, DOFAuthenticator_GetDebugKey(data->domain), 32);
                }

                write_ptr = OALStream_GetWritePtr(data->A);

                write_ptr[0] = TRP_V1_OP_RANDOM;
                i_secret = DOFCredentials_GetKeyFromStorage( data->RequestData.randomData->initKeyRequest->credentials, i_node->fns->GetCredentialsStorage(i_node));

                OALSecureHMAC_Start(&hmac, (uint8*)i_secret);
                OALSecureHMAC_Digest(&hmac, (domain ? DOFObjectID_GetBytesSize( domain ) : 0), DOFObjectID_GetBytes( domain ));
                OALSecureHMAC_Digest(&hmac, data->I_len, data->I);
                OALSecureHMAC_Digest(&hmac, 32, random_key);
                OALSecureHMAC_Finish(&hmac, write_ptr+1); /* write the mac */

                shared_key_key( i_secret, write_ptr+1, random_key, write_ptr+33);
                OALStream_AdjustWriteOffset( data->A, len );
            }
            DOFAuthenticatorNode_Release(i_node);
        }
    }

    if( data->errorCode != 0 )
        data->A = TRP_V1_AS_CreateReject( data->errorCode );

    return TRUE;
}

boolean TRP_V1_AS_RequestSession( TRP_V1_Data data )
{
    DOFAuthenticatorStorage storage = DOFAuthenticator_GetStorage(data->domain);

    if(!storage || ! storage->fns->IsStorageAvailable(storage))
    {
        data->errorCode = TRP_V1_SERVICE_UNAVAILABLE;
    }
    else
    {
        DOFAuthenticatorNode i_node, r_node;
        uint8 session_key[32];
        uint8* mode = data->I+1; /* offset of mode */
        uint16 modeLen = 4 + ((uint8)data->I[1+3]);
        DOFError status = DOFERROR_OK;

        i_node = storage->fns->GetNode( storage, DOFCredentials_GetIdentity(data->RequestData.sessionData->initBlock->keyRequest->credentials), DOFCredentials_GetType( data->RequestData.sessionData->initBlock->keyRequest->credentials ), DOFAUTHENTICATORSECURITYTYPE_TWO_POINT, modeLen, mode, data->RequestData.sessionData->initBlock->keyRequest->permission_set, &status );
        if( !i_node )
        {
            data->errorCode = TRP_V1_AS_DOFErrorToTRPRejectCode( status, TRUE );
        }
        else
        {
            if( !i_node->fns->IsEnabled(i_node) )
            {
                data->errorCode = TRP_V1_ACCESS_DENIED;
            }
            else
            {
                status = DOFERROR_OK;
                r_node = storage->fns->GetNode( storage, DOFCredentials_GetIdentity(data->RequestData.sessionData->respBlock->credentials), DOFCredentials_GetType( data->RequestData.sessionData->respBlock->credentials ), DOFAUTHENTICATORSECURITYTYPE_TWO_POINT, modeLen, mode, data->RequestData.sessionData->respBlock->permission_set, &status );
                if( !r_node )
                {
                    data->errorCode = TRP_V1_AS_DOFErrorToTRPRejectCode( status, FALSE );
                }
                else
                {
                    if( !r_node->fns->IsEnabled(r_node) )
                    {
                        data->errorCode = TRP_V1_ACCESS_DENIED;
                    }
                    else
                    {
                        if( !ValidatePermissionsRequestAgainstGranted( data->RequestData.sessionData->initBlock->keyRequest->permission_set, i_node->fns->GetPermissions(i_node), i_node->fns->GetSourceIdentifier(i_node), DOFAuthenticator_GetID(data->domain)))
                        {
                            data->errorCode = TRP_V1_INTERNAL_ERROR;
                        }
                        else
                        {
                            if( !ValidatePermissionsRequestAgainstGranted( data->RequestData.sessionData->respBlock->permission_set, r_node->fns->GetPermissions(r_node), r_node->fns->GetSourceIdentifier(r_node), DOFAuthenticator_GetID(data->domain)))
                            {
                                data->errorCode = TRP_V1_INTERNAL_ERROR;
                            }
                            else
                            {
                                /* first figure out stream size */
                                uint32 stream_size = 130; /* TRP opcode + granted duration (1 byte) + key grant (128 bytes) */
                                DOFSecurityScope i_node_scope = TRP_V1_AS_CreateSecurityScope(DOFAuthenticator_GetDomainID(data->domain), i_node);
                                if( i_node_scope )
                                {
                                    DOFSecurityScope r_node_scope = TRP_V1_AS_CreateSecurityScope(DOFAuthenticator_GetDomainID(data->domain), r_node);
                                    if( r_node_scope )
                                    {
                                        DOFSecurityScope compatible_scope = DOFSecurityScope_CreateCompatibleScope( i_node_scope, r_node_scope);
                                        if(!compatible_scope)
                                        {
                                            data->errorCode = TRP_V1_INTERNAL_ERROR;
                                        }
                                        else
                                        {
                                            uint8 *stream_ptr;
                                            OALSecureHMACContext hmac;
                                            uint8 max = DOFAuthenticator_GetMaxLifetime( data->domain );
                                            uint8 grantedDuration = (data->RequestData.sessionData->initBlock->duration > max)?max:data->RequestData.sessionData->initBlock->duration;

                                            /* create the session key */
                                            if(!(DOFAuthenticator_IsDebug(data->domain) && DOFAuthenticator_GetDebugKey(data->domain)))
                                            {
                                                PCRSecureRandom_Generate(32, session_key);
                                            }
                                            else
                                            {
                                                PCRMem_Copy(session_key, DOFAuthenticator_GetDebugKey(data->domain), 32);
                                            }

                                            stream_size += DOFPermissionSet_MarshalValidationSize(data->RequestData.sessionData->initBlock->keyRequest->permission_set);
                                            stream_size += DOFPermissionSet_MarshalValidationSize(data->RequestData.sessionData->respBlock->permission_set);
                                            stream_size += DOFSecurityScope_GetMarshalSize(compatible_scope);

                                            data->A = OALStream_CreateRW(stream_size);
                                            if(!data->A)
                                            {
                                                data->errorCode = TRP_V1_INTERNAL_ERROR;
                                            }
                                            else
                                            {

                                                DOFObjectID domain = DOFObjectIDDomain_GetObjectID( DOFAuthenticator_GetDomainID( data->domain ) );
                                                stream_ptr = OALStream_GetWritePtr(data->A);
                                                /* marshal trp opcode */
                                                /* temporarily skip empty grants which we'll fill back in later */
                                                /* start marshal of A block. starts at 129 offset */
                                                if( OALStream_Write8(data->A, TRP_V1_OP_SESSION) )
                                                {
                                                    OALStream_AdjustWriteOffset(data->A, 128);

                                                    if( OALStream_Write8(data->A, grantedDuration ) &&
                                                        DOFSecurityScope_Marshal(compatible_scope, data->A) &&
                                                        DOFPermissionSet_MarshalValidation(data->RequestData.sessionData->initBlock->keyRequest->permission_set, data->A) &&
                                                        DOFPermissionSet_MarshalValidation(data->RequestData.sessionData->respBlock->permission_set, data->A) )
                                                    {
                                                        shared_key_mac(domain, data->RequestData.sessionData->respBlock->credentials, data->I_len, data->I, data->R_len, data->R, r_node, &hmac, stream_ptr+129, OALStream_GetWriteOffset(data->A)-129, session_key, stream_ptr+1 /* r_mac */);
                                                        shared_key_mac(domain, data->RequestData.sessionData->initBlock->keyRequest->credentials, data->I_len, data->I, data->R_len, data->R, i_node, &hmac, stream_ptr+129, OALStream_GetWriteOffset(data->A)-129, session_key, stream_ptr+1+64 /* i_mac */);

                                                        /* now compute the key */
                                                        shared_key_key( DOFCredentials_GetKeyFromStorage( data->RequestData.sessionData->respBlock->credentials,r_node->fns->GetCredentialsStorage(r_node)), stream_ptr+1, session_key, stream_ptr+1+32);
                                                        shared_key_key( DOFCredentials_GetKeyFromStorage( data->RequestData.sessionData->initBlock->keyRequest->credentials, i_node->fns->GetCredentialsStorage(i_node) ), stream_ptr+1+64, session_key, stream_ptr+1+64+32);
                                                        /* we're done! */
                                                    }
                                                    else
                                                        data->errorCode = TRP_V1_INTERNAL_ERROR;
                                                }
                                                else
                                                    data->errorCode = TRP_V1_INTERNAL_ERROR;
                                            }
                                            DOFSecurityScope_Destroy(compatible_scope);
                                        }
                                        DOFSecurityScope_Destroy(r_node_scope);
                                    }
                                    else
                                        data->errorCode = TRP_V1_INTERNAL_ERROR;
                                    DOFSecurityScope_Destroy(i_node_scope);
                                }
                                else
                                    data->errorCode = TRP_V1_INTERNAL_ERROR;
                            }
                        }
                    }
                    DOFAuthenticatorNode_Release(r_node);
                }
            }
            DOFAuthenticatorNode_Release(i_node);
        }
    }

    if( data->errorCode != 0 )
        data->A = TRP_V1_AS_CreateReject( data->errorCode );

    return TRUE;
}

boolean TRP_V1_AS_RequestSecurityScope( TRP_V1_Data data )
{
    DOFAuthenticatorStorage storage = DOFAuthenticator_GetStorage(data->domain);

    if(!storage || !storage->fns->IsStorageAvailable(storage))
    {
        data->errorCode = TRP_V1_SERVICE_UNAVAILABLE;
    }
    else
    {
        DOFError status = DOFERROR_OK;
        DOFAuthenticatorNode i_node = storage->fns->GetNode( storage, DOFCredentials_GetIdentity(data->RequestData.securityScopeData->initKeyRequest->credentials), DOFCredentials_GetType( data->RequestData.securityScopeData->initKeyRequest->credentials ), DOFAUTHENTICATORSECURITYTYPE_UNDEFINED, 0, NULL, data->RequestData.securityScopeData->initKeyRequest->permission_set, &status );
        if(!i_node)
        {
            data->errorCode = TRP_V1_AS_DOFErrorToTRPRejectCode( status, TRUE );
        }
        else
        {
            if( !i_node->fns->IsEnabled(i_node) )
            {
                data->errorCode = TRP_V1_ACCESS_DENIED;
            }
            else
            {
                DOFAuthenticatorNode r_node;
                status = DOFERROR_OK; 
                r_node = storage->fns->GetNode( storage, (DOFObjectIDAuthentication)data->RequestData.securityScopeData->secNodeID, DOFCREDENTIALSTYPE_NONE, DOFAUTHENTICATORSECURITYTYPE_UNDEFINED, 0, NULL, data->RequestData.securityScopeData->initKeyRequest->permission_set, &status );
                if(!r_node)
                {
                    data->errorCode = TRP_V1_AS_DOFErrorToTRPRejectCode( status, FALSE );
                }
                else
                {
                    if( !r_node->fns->IsEnabled(r_node) )
                    {
                        data->errorCode = TRP_V1_ACCESS_DENIED;
                    }
                    else
                    {
                        if(!ValidatePermissionsRequestAgainstGranted(data->RequestData.securityScopeData->initKeyRequest->permission_set, r_node->fns->GetPermissions(r_node), r_node->fns->GetSourceIdentifier(r_node), DOFAuthenticator_GetID(data->domain)))
                        {
                            data->errorCode = TRP_V1_INTERNAL_ERROR;
                        }
                        else
                        {
                            DOFSecurityScope r_scope = TRP_V1_AS_CreateSecurityScope(DOFAuthenticator_GetDomainID(data->domain), r_node);

                            if( !r_scope )
                            {
                                data->errorCode = TRP_V1_INTERNAL_ERROR;
                            }
                            else
                            {
                                uint8 max = DOFAuthenticator_GetMaxLifetime( data->domain );
                                uint8 grantedDuration = (data->RequestData.securityScopeData->grantedDuration > max)?max:data->RequestData.securityScopeData->grantedDuration;

                                uint32 stream_size = 1 /* opcode */ + 1 /* granted duration */ +
                                                     OALKeyGrant_MarshalSize(NULL);
                                stream_size += DOFPermissionSet_MarshalValidationSize(data->RequestData.securityScopeData->initKeyRequest->permission_set);
                                stream_size += DOFSecurityScope_GetMarshalSize(r_scope);

                                data->A = OALStream_CreateRW(stream_size);
                                if(!data->A)
                                {
                                    data->errorCode = TRP_V1_INTERNAL_ERROR;
                                }
                                else
                                {
                                    uint8 *stream_ptr;
                                    OALSecureHMACContext hmac;
                                    uint8 session_key[32];
                                    DOFObjectID domain = DOFObjectIDDomain_GetObjectID( DOFAuthenticator_GetDomainID( data->domain ) );

                                    if(!(DOFAuthenticator_IsDebug(data->domain) && DOFAuthenticator_GetDebugKey(data->domain)))
                                    {
                                        PCRSecureRandom_Generate(32, session_key);
                                    }
                                    else
                                    {
                                        PCRMem_Copy(session_key, DOFAuthenticator_GetDebugKey(data->domain), 32);
                                    }

                                    stream_ptr = OALStream_GetWritePtr(data->A);

                                    OALStream_Write8(data->A, TRP_V1_OP_SECURITY_SCOPE);
                                    OALStream_AdjustWriteOffset(data->A, OALKeyGrant_MarshalSize(NULL));
                                    OALStream_Write8(data->A, grantedDuration);
                                    DOFSecurityScope_Marshal(r_scope, data->A);

                                    if( DOFPermissionSet_MarshalValidation(data->RequestData.securityScopeData->initKeyRequest->permission_set, data->A) )
                                    {
                                        OALSecureHMAC_Start(&hmac, (uint8*)DOFCredentials_GetKeyFromStorage(data->RequestData.securityScopeData->initKeyRequest->credentials, i_node->fns->GetCredentialsStorage(i_node) ));
                                        OALSecureHMAC_Digest(&hmac, (domain ? DOFObjectID_GetBytesSize( domain ) : 0), DOFObjectID_GetBytes(domain));
                                        OALSecureHMAC_Digest(&hmac, data->I_len, data->I);
                                        OALSecureHMAC_Digest(&hmac, OALStream_GetWriteOffset(data->A)-(1+OALKeyGrant_MarshalSize(NULL)), stream_ptr+(1+OALKeyGrant_MarshalSize(NULL)));
                                        OALSecureHMAC_Digest(&hmac, 32, session_key);
                                        OALSecureHMAC_Finish(&hmac, stream_ptr+1);

                                        shared_key_key(DOFCredentials_GetKeyFromStorage(data->RequestData.securityScopeData->initKeyRequest->credentials, i_node->fns->GetCredentialsStorage(i_node) ), stream_ptr+1, session_key, stream_ptr+33);
                                    }
                                    else
                                    {
                                        data->errorCode = TRP_V1_INTERNAL_ERROR;
                                    }
                                }
                                DOFSecurityScope_Destroy(r_scope);
                            }
                        }
                    }
                    DOFAuthenticatorNode_Release(r_node);
                }
            }
        }
        DOFAuthenticatorNode_Release(i_node);
    }

    if( data->errorCode != 0 )
        data->A = TRP_V1_AS_CreateReject( data->errorCode );

    return TRUE;
}

boolean TRP_V1_AS_RequestValidateCredential( TRP_V1_Data data )
{
    DOFAuthenticatorStorage storage = DOFAuthenticator_GetStorage(data->domain);

    if(!storage || !storage->fns->IsStorageAvailable(storage))
    {
         data->errorCode = TRP_V1_SERVICE_UNAVAILABLE;
    }
    else
    {
        DOFError status = DOFERROR_OK;
        DOFAuthenticatorNode i_node = storage->fns->GetNode( storage, DOFCredentials_GetIdentity(data->RequestData.validateData->credentials), DOFCredentials_GetType(data->RequestData.validateData->credentials), DOFAUTHENTICATORSECURITYTYPE_UNDEFINED, 0, NULL, NULL, &status );
        if(!i_node)
        {
            data->errorCode = TRP_V1_AS_DOFErrorToTRPRejectCode( status, TRUE );
        }
        else
        {
            if( !i_node->fns->IsEnabled(i_node) )
            {
                data->errorCode = TRP_V1_INTERNAL_ERROR;
            }
            else
            {
                OALSecureHMACContext hmac;
                uint8* validA = data->RequestData.validateData->validA;
                uint8 validateA[40];
                DOFObjectIDSource sid = SID_Create(i_node->fns->GetSourceIdentifier(i_node),DOFAuthenticator_GetID(data->domain));
                if( sid )
                {
                    uint32 sid_len = DOFObjectID_GetBytesSize(DOFObjectIDSource_GetObjectID(sid));

                    PCRMem_Copy(validateA, validA, 8);

                    /* Validate ValidA */
                    OALSecureHMAC_Start(&hmac, DOFCredentials_GetKeyFromStorage(data->RequestData.validateData->credentials, i_node->fns->GetCredentialsStorage(i_node)));
                    OALSecureHMAC_Digest(&hmac, 8, validateA);
                    OALSecureHMAC_Finish(&hmac,(validateA+8));

                    if( PCRMem_Compare( validateA, validA,40) != 0 )
                    {
                        data->errorCode = TRP_V1_INTERNAL_ERROR;
                    }
                    else
                    {
                        data->A = OALStream_CreateRW( 41 + sid_len);
                        if( !data->A )
                        {
                            data->errorCode = TRP_V1_INTERNAL_ERROR;
                        }
                        else
                        {
                            uint8* writePtr;
                            OALStream_Write8(data->A, TRP_V1_OP_VALIDATE_CREDS);
                            writePtr = OALStream_GetWritePtr( data->A );
                            if( !writePtr )
                            {
                                OALStream_Release(data->A, TRUE);
                                data->A = NULL;
                                data->errorCode = TRP_V1_INTERNAL_ERROR;
                            }
                            else
                            {
                                DOFObjectID domain = DOFObjectIDDomain_GetObjectID( DOFAuthenticator_GetDomainID( data->domain ) );

                                PCRSecureRandom_Generate(8,writePtr);
                                OALStream_AdjustWriteOffset(data->A, 8);
                                DOFObjectID_Marshal(DOFObjectIDSource_GetObjectID(sid), data->A);
                                OALSecureHMAC_Start(&hmac, DOFCredentials_GetKeyFromStorage(data->RequestData.validateData->credentials, i_node->fns->GetCredentialsStorage(i_node)));
                                OALSecureHMAC_Digest(&hmac, 8, validA);
                                OALSecureHMAC_Digest(&hmac, 8, writePtr);
                                OALSecureHMAC_Digest(&hmac, DOFObjectID_GetBytesSize(DOFObjectIDSource_GetObjectID(sid)), DOFObjectID_GetBytes(DOFObjectIDSource_GetObjectID(sid)));
                                OALSecureHMAC_Digest(&hmac, (domain ? DOFObjectID_GetBytesSize(domain) : 0), DOFObjectID_GetBytes(domain));
                                OALSecureHMAC_Finish(&hmac,(writePtr+sid_len+8));
                                OALStream_AdjustWriteOffset(data->A, 32);
                            }
                        }
                    }
                    DOFObjectIDSource_Destroy(sid);
                }
            }
        }
    }

    if( data->errorCode != 0 )
        data->A = TRP_V1_AS_CreateReject( data->errorCode );

    return TRUE;
}

boolean TRP_V1_AS_RequestCredential( TRP_V1_Data data )
{
    DOFAuthenticatorStorage storage = DOFAuthenticator_GetStorage(data->domain);

    if(!storage || !storage->fns->IsStorageAvailable(storage))
    {
         data->errorCode = TRP_V1_SERVICE_UNAVAILABLE;
    }
    else
    {
        DOFError status = storage->fns->Resolve( storage, data->RequestData.credentialData->credentials );
        if( status != DOFERROR_OK )
        {
            uint8 err = TRP_V1_AS_DOFErrorToTRPRejectCode( status, TRUE );
            if( err == TRP_V1_INTERNAL_ERROR )
                data->errorCode = TRP_V1_RESOLUTION_FAILED;
            else
                data->errorCode = err;
        }
        else
        {
            uint32 stream_size = 1 + DOFCredentials_MarshalSize( data->RequestData.credentialData->credentials, FALSE) + ((data->flags.bits.isDomainDiscovery)?DOFObjectID_GetBytesSize(DOFObjectIDDomain_GetObjectID(DOFAuthenticator_GetDomainID(data->domain))):0);

            data->A = OALStream_CreateRW( stream_size );
            if( !data->A )
            {
                data->errorCode = TRP_V1_INTERNAL_ERROR;
            }
            else
            {
                if(!data->flags.bits.isDomainDiscovery)
                {
                    OALStream_Write8(data->A, TRP_V1_OP_CREDENTIAL);
                }
                else
                {
                    OALStream_Write8(data->A, TRP_V1_OP_DOMAINCREDENTIAL);
                    DOFObjectID_Marshal(DOFObjectIDDomain_GetObjectID(DOFAuthenticator_GetDomainID(data->domain)),data->A);
                }
                DOFCredentials_Internal_Marshal( data->RequestData.credentialData->credentials, data->A, FALSE );

                if( DOFCredentials_HasAnotherResolutionStage(data->RequestData.credentialData->credentials) )
                    data->flags.bits.ASDontCancel = 1;
                else
                    data->flags.bits.ASDontCancel = 0;
            }
        }
    }

    if( data->errorCode != 0 )
        data->A = TRP_V1_AS_CreateReject( data->errorCode );

    return TRUE;
}

boolean TRP_V1_AS_RequestLocalDomain( TRP_V1_Data data )
{
    DOFAuthenticatorStorage storage = DOFAuthenticator_GetStorage(data->domain);

    if(!storage || !storage->fns->IsStorageAvailable(storage))
    {
        data->errorCode = TRP_V1_SERVICE_UNAVAILABLE;
    }
    else
    {
        DOFAuthenticatorNode d_node = storage->fns->GetDomainNode( storage, (DOFObjectIDDomain)data->RequestData.domainData->remDomain, DOFCredentials_GetIdentity(data->RequestData.domainData->initRemoteKeyRequest->credentials) );
        if( !d_node )
        {
            data->errorCode = TRP_V1_UNKNOWN_DOMAIN;
        }
        else
        {
            if( !d_node->fns->IsEnabled(d_node) )
            {
                data->errorCode = TRP_V1_ACCESS_DENIED;
            }
            else
            {
                DOFError status = DOFERROR_OK;
                DOFAuthenticatorNode i_node = storage->fns->GetNode( storage, DOFCredentials_GetIdentity(data->RequestData.domainData->initRemoteKeyRequest->credentials), DOFCREDENTIALSTYPE_KEY, DOFAUTHENTICATORSECURITYTYPE_UNDEFINED, 0, NULL, DOFPermissionSet_GetEmptyPermissionSet(), &status );
                if( !i_node )
                {
                    data->errorCode = TRP_V1_AS_DOFErrorToTRPRejectCode( status, TRUE );
                }
                else
                {
                    if( !i_node->fns->IsEnabled(i_node) )
                    {
                        data->errorCode = TRP_V1_ACCESS_DENIED;
                    }
                    else
                    {
                        uint8 mac[32];
                        OALSecureHMACContext hmac;
                        const uint8 *d_secret = d_node->fns->GetCredentialsStorage(d_node);
                        const uint8 *i_secret;
                        uint8 remote_secret[32];

                        i_secret = DOFCredentials_GetKeyFromStorage( data->RequestData.domainData->initRemoteKeyRequest->credentials, i_node->fns->GetCredentialsStorage(i_node));

                        OALKeyGrant_ExtractK(data->RequestData.domainData->initRemoteKeyGrant, d_secret, remote_secret);

                        if(!data->errorCode)
                        {
                            /* find out if macs match */
                            OALSecureHMAC_Start(&hmac, (uint8*)d_secret);
                            OALSecureHMAC_Digest(&hmac, DOFObjectID_GetBytesSize(DOFObjectIDDomain_GetObjectID( data->RequestData.domainData->remDomain )), DOFObjectID_GetBytes(DOFObjectIDDomain_GetObjectID(data->RequestData.domainData->remDomain)));
                            OALSecureHMAC_Digest(&hmac, data->I_len, data->I);
                            OALSecureHMAC_Digest(&hmac, data->RequestData.domainData->responseA_len, data->RequestData.domainData->responseA);
                            OALSecureHMAC_Digest(&hmac, 32, remote_secret);
                            OALSecureHMAC_Finish(&hmac, mac);

                            if(PCRMem_Compare(mac, data->RequestData.domainData->initRemoteKeyGrant->mac, 32) != 0)
                            {
                                data->errorCode = TRP_V1_INTERNAL_ERROR;
                            }
                            else
                            {
                                uint32 stream_size = OALKeyGrant_MarshalSize(NULL)+1; /* key grant and opcode */
                                stream_size += OALMarshal_GetCompressedValueSize(4, d_node->fns->GetRemoteDomainIdentifier(d_node));

                                data->A = OALStream_CreateRW(stream_size);

                                if(!data->A)
                                {
                                    data->errorCode = TRP_V1_INTERNAL_ERROR;
                                }
                                else
                                {
                                    uint8 *write_ptr = OALStream_GetWritePtr(data->A);
                                    DOFObjectID domain = DOFObjectIDDomain_GetObjectID( DOFAuthenticator_GetDomainID( data->domain ) );

                                    OALStream_Write8(data->A, TRP_V1_OP_LOCAL_DOMAIN);

                                    OALStream_AdjustWriteOffset(data->A, OALKeyGrant_MarshalSize(NULL));

                                    OALMarshal_StreamCompress4(data->A, d_node->fns->GetRemoteDomainIdentifier(d_node));

                                    OALSecureHMAC_Start(&hmac, (uint8*)i_secret);
                                    OALSecureHMAC_Digest(&hmac, (domain ? DOFObjectID_GetBytesSize(domain) : 0), DOFObjectID_GetBytes(domain));
                                    OALSecureHMAC_Digest(&hmac, data->I_len, data->I);
                                    OALSecureHMAC_Digest(&hmac, OALStream_GetWriteOffset(data->A)-(1+OALKeyGrant_MarshalSize(NULL)), write_ptr+(1+OALKeyGrant_MarshalSize(NULL)));
                                    OALSecureHMAC_Digest(&hmac, 32, remote_secret);
                                    OALSecureHMAC_Finish(&hmac, write_ptr+1);

                                    shared_key_key(i_secret, write_ptr+1, remote_secret, write_ptr+33);
                                }
                            }
                        }
                    }
                    DOFAuthenticatorNode_Release(i_node);
                }
            }
            DOFAuthenticatorNode_Release(d_node);
        }
    }

    if( data->errorCode != 0 )
        data->A = TRP_V1_AS_CreateReject( data->errorCode );

    return TRUE;
}

boolean TRP_V1_AS_RequestRemoteDomain( TRP_V1_Data data )
{
    DOFAuthenticatorStorage storage = DOFAuthenticator_GetStorage(data->domain);

    if(!storage || !storage->fns->IsStorageAvailable(storage))
    {
        data->errorCode = TRP_V1_SERVICE_UNAVAILABLE;
    }
    else
    {
        DOFAuthenticatorNode d_node = storage->fns->GetDomainNode( storage, data->RequestData.domainData->localDomain, DOFCredentials_GetIdentity(data->RequestData.domainData->initRemoteKeyRequest->credentials) );
        if( !d_node )
        {
            data->errorCode = TRP_V1_UNKNOWN_DOMAIN;
        }
        else
        {
            if( !d_node->fns->IsEnabled(d_node) )
            {
                data->errorCode = TRP_V1_ACCESS_DENIED;
            }
            else
            {
                DOFError status = DOFERROR_OK;
                DOFAuthenticatorNode i_node;
                DOFObjectIDAuthentication authID = d_node->fns->GetLocalID(d_node);
                if ( !authID )
                    authID = DOFCredentials_GetIdentity(data->RequestData.domainData->initRemoteKeyRequest->credentials);

                i_node = storage->fns->GetNode( storage, authID, DOFCREDENTIALSTYPE_KEY, DOFAUTHENTICATORSECURITYTYPE_UNDEFINED, 0, NULL, DOFPermissionSet_GetEmptyPermissionSet(), &status );
                if( !i_node )
                {
                    data->errorCode = TRP_V1_AS_DOFErrorToTRPRejectCode( status, TRUE );
                }
                else
                {
                    if( !i_node->fns->IsEnabled(i_node) )
                    {
                        data->errorCode = TRP_V1_ACCESS_DENIED;
                    }
                    else
                    {
                        uint32 stream_size;
                        DOFObjectID newid = DOFObjectIDAuthentication_GetObjectID( authID );
                        DOFObjectID domain = DOFObjectIDDomain_GetObjectID( DOFAuthenticator_GetDomainID( data->domain ) );
                        DOFObjectIDSize domSize = ((data->flags.bits.isDomainDiscovery) ?
                                                    (domain ? DOFObjectID_GetBytesSize( domain ) : 0) : 0);
                        DOFObjectIDSize idSize;

                        idSize = newid ? DOFObjectID_GetBytesSize( newid ) : 0;
                        stream_size = 1 + OALKeyGrant_MarshalSize(NULL) + idSize + domSize;

                        data->A = OALStream_CreateRW(stream_size);
                        if(!data->A)
                        {
                            data->errorCode = TRP_V1_INTERNAL_ERROR;
                        }
                        else
                        {
                            OALSecureHMACContext hmac;
                            const uint8 *d_secret;
                            const uint8 *i_secret;
                            uint8 *write_ptr;

                            write_ptr = OALStream_GetWritePtr( data->A );

                            if(!data->flags.bits.isDomainDiscovery)
                            {
                                OALStream_Write8(data->A, TRP_V1_OP_REMOTE_DOMAIN);
                            }
                            else
                            {
                                OALStream_Write8(data->A, TRP_V1_OP_DISCOVER_REMOTE);
                                DOFObjectID_Marshal(domain, data->A);
                            }

                            OALStream_AdjustWriteOffset(data->A, OALKeyGrant_MarshalSize(NULL));

                            DOFObjectID_Marshal(newid, data->A);

                            d_secret = d_node->fns->GetCredentialsStorage( d_node ); /* this is the domain seceret */
                            i_secret = i_node->fns->GetCredentialsStorage(i_node); /* This will always be a key in request remote domain */

                            OALSecureHMAC_Start(&hmac, (uint8*)d_secret);
                            OALSecureHMAC_Digest(&hmac, (domain ? DOFObjectID_GetBytesSize(domain) : 0), DOFObjectID_GetBytes(domain));
                            OALSecureHMAC_Digest(&hmac, data->I_len, data->I);
                            OALSecureHMAC_Digest(&hmac, idSize, write_ptr+(1+domSize+OALKeyGrant_MarshalSize(NULL)));
                            OALSecureHMAC_Digest(&hmac, 32, i_secret);
                            OALSecureHMAC_Finish(&hmac, write_ptr+1+domSize);

                            shared_key_key( d_secret, write_ptr+1+domSize, i_secret, write_ptr+33+domSize);
                        }
                    }
                    DOFAuthenticatorNode_Release(i_node);
                }
            }
            DOFAuthenticatorNode_Release(d_node);
        }
    }

    if( data->errorCode != 0 )
       data->A = TRP_V1_AS_CreateReject( data->errorCode );

    return TRUE;
}

