/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/dof.h>
#include <dof/cfg.h>
#include <dof/system.h>
#include <dof/oaltypes.h>
#include <internal/router.h>
#include <internal/stream.h>
#include <internal/trp_v1.h>
#include <internal/srp.h>
#include <internal/objectidinternal.h>
#include <internal/credentialsinternal.h>
#include <internal/blackboarddata.h>
#include <internal/permissionsetinternal.h>
#include <internal/securitymode.h>
#include <internal/security_common.h>
#include <internal/dofinternal.h>
#include <internal/error.h>
#include <internal/resultinternal.h>
#include <internal/marshal.h>

#include "trp_v1_sm.h"
#include "trp_v1_utils.h"
#include "trp_v1_receive.h"
#include "trp_v1_operation.h"
#include "trp_v1_receive_util.h"

#define TRP_V1_STATUS_OK             0
#define TRP_V1_STATUS_ERROR          1
#define TRP_V1_STATUS_FAILED_READ    2

typedef void (*TRP_V1_ReceiveFunction)(TRP_V1_ReceiveData_t data);

static boolean HandlePassThrough(TRP_V1_ReceiveData_t data)
{
    TRP_V1_PassData_t* passData = data.opData->RequestData.passData;

    if( !data.trp->blackboardData->flags.isCommandNotResponse )
    {
        DOFResult ret = NULL;
        OALStream strm = NULL;

        OALStream_AdjustReadOffset(data.stream, -1); /* roll back and get the op code and app ID */
        strm = OALStream_CreateRW(OALStream_GetDataSize(data.stream));
        if(strm && OALStream_ReadToStream(data.stream, OALStream_GetDataSize(data.stream), strm))
        {
            ret = DOFResult_Stream_Create(passData->passOnChannel,passData->address, DOFERROR_OK, DOFObjectIDDomain_GetObjectID(data.opData->secDomain), strm, DOFRESULTTYPE_VALUELIST_STREAM);
            if( ret )
            {
                /* SUCCESS */
                data.opData->trp_status = DOFERROR_OK;
                DOFOperation_Unlock(data.op);
                passData->passOnChannel->routeToRequestor(passData->passOnChannel,data.op,ret);
                DOFOperation_Lock(data.op);
                OALStream_Release(strm, FALSE);
                DOFResult_Destroy(ret);
                return TRUE;
            }
        }
        OALStream_Release(strm, FALSE);
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "TRP V1: Failed to read from stream in HandlePassThrough() response.");
        data.opData->trp_status = DOFERROR_REQUEST_FAILED;
    }
    else
    {
        data.opData->flags.bits.isCommand = data.trp->blackboardData->flags.isCommandNotResponse;

        OALStream_AdjustReadOffset(data.stream, -1); /* roll back and get the op code */
        passData->length = OALStream_GetDataSize(data.stream);
        if(passData->streamData)
            DOFMem_Free(passData->streamData);
        passData->streamData = DOFMem_Alloc(passData->length);
        if( passData->streamData )
        {
            if( OALStream_Read(data.stream, passData->length, passData->streamData ))
            {
                data.opData->trp_status = DOFERROR_OK;
                return FALSE;
            }
        }

        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_DEBUG, "TRP V1: Failed to read from stream in HandlePassThrough() command");
        data.opData->trp_status = DOFERROR_REQUEST_FAILED;
    }
    return FALSE;
}

static void TRP_V1_ExtractK( TRP_V1_ReceiveData_t data, const uint8* key, OALKeyGrant kg )
{
    OALKeyGrant_ExtractK( kg, key, data.opData->k );
}

static boolean TRP_V1_ReceiveAuthorize_Validate(TRP_V1_ReceiveData_t data, const uint8* key, OALKeyGrant kg)
{
    OALSecureHMACDigest digest;

    TRP_V1_ExtractK( data, key, kg );

    /* SecureHMAC has already been started */
    OALSecureHMAC_Digest( &data.opData->ctx, 32, data.opData->k );

    OALSecureHMAC_Finish( &data.opData->ctx, digest );

    if(PCRMem_Compare(digest,kg->mac,32) != 0)
    {
        data.opData->trp_status = DOFERROR_VALIDATION_FAILED;
        PCRMem_Set( data.opData->k, 0, 32);
        return FALSE;
    }

    data.opData->trp_status = DOFERROR_OK;
    return TRUE;
}

static DOFObjectIDDomain TRP_V1_GetSecurityDomainFromStream( OALStream stream )
{
    const uint8* start = OALStream_GetReadPtr(stream);
    DOFObjectIDDomain ret = DOFObjectIDDomain_Unmarshal( stream );
    OALStream_AdjustReadOffset( stream, -(OALStream_GetReadPtr(stream)-start) ); /* roll back */
    return ret;
}

ProtocolLayerResult TRP_V1_Receive(TRP_V1 self)
{
    TRP_V1_ReceiveData_t data;
    uint8 opcode = 12; /* init to out of bounds */
    uint16  appID;
    static const TRP_V1_ReceiveFunction trp_functions[12][2]=
        {
            {TRP_V1_ReceiveReject,                      TRP_V1_ReceiveReject},                  /* TRP_V1_OP_REJECT                 */
            {TRP_V1_ReceiveAuthorizeKEK,                TRP_V1_ReceiveRequestKEK},              /* TRP_V1_OP_KEK                    */
            {TRP_V1_ReceiveAuthorizeRandom,             TRP_V1_ReceiveRequestRandom},           /* TRP_V1_OP_RANDOM                 */
            {TRP_V1_ReceiveAuthorizeSession,            TRP_V1_ReceiveRequestSession},          /* TRP_V1_OP_SESSION                */
            {TRP_V1_ReceiveAuthorizeSecurityScope,      TRP_V1_ReceiveRequestSecurityScope},    /* TRP_V1_OP_SECURITY_SCOPE         */
            {TRP_V1_ReceiveAuthorizeDomainCredential ,  TRP_V1_ReceiveInvalidOpCode},           /* TRP_V1_OP_CREDENTIAL DISCOVERY   */
            {TRP_V1_ReceiveAuthorizeCredential,         TRP_V1_ReceiveRequestCredential},       /* TRP_V1_OP_CREDENTIAL             */
            {TRP_V1_ReceiveAuthorizeLocalDomain,        TRP_V1_ReceiveRequestLocalDomain},      /* TRP_V1_OP_LOCAL_DOMAIN           */
            {TRP_V1_ReceiveAuthorizeRemoteDomain,       TRP_V1_ReceiveRequestRemoteDomain},     /* TRP_V1_OP_REMOTE_DOMAIN          */
            {TRP_V1_ReceiveAuthorizeValidateCreds,      TRP_V1_ReceiveRequestValidateCreds},    /* TRP_V1_OP_VALIDATE_CREDS         */
            {TRP_V1_ReceiveAuthorizeRemoteDiscover,     TRP_V1_ReceiveInvalidOpCode},           /* TRP_V1_OP_DISCOVER_REMOTE        */
            {TRP_V1_ReceiveAuthorizeSecurityScope,      TRP_V1_ReceiveRequestSecurityScope}     /* TRP_V1_OP_SECURITY_SCOPE 0x0B    */
         };

    if( !self->blackboardData->associatedChannel->remoteChannelData->flags.TRP_AcceptsCommands && self->blackboardData->flags.isCommandNotResponse )
    {
        return PLRESULT_OKDONOTHING; /* can't receive it on this channel */
    }

    data.trp = self;
    data.op = NULL;
    data.opData = NULL;

    /* APPID */
    if( self->blackboardData->receiveStream != NULL )
    {
        if( !OALStream_Read16(self->blackboardData->receiveStream, &appID) || appID != 0x8081 )
        {
            PCR_Assert(FALSE); /* ASSERT BECAUSE IF IT GETS HERE WITH ANOTHER APPIP ITS A REALLY BAD SCREW UP */
            return PLRESULT_ERROR;
        }
    }
    else
        return PLRESULT_ERROR;

    if( OALStream_Read8( self->blackboardData->receiveStream, &opcode ) && opcode <= 11 )
    {
        if( self->blackboardData->opID != NULL ) 
        {
            data.op = OALOperationManager_CheckExistingOperation(DOFSecurityScope_GetDomainID(DOF_GetDefaultSecurityScope(DOF_GetDof())),self->blackboardData->opID);

            if( !data.op )
            {
                if( self->blackboardData->flags.isCommandNotResponse )
                {
                    /* we don't know this op so we are the AS or we can pass it on to one we know */
                    DOFObjectIDDomain domain = TRP_V1_GetSecurityDomainFromStream( self->blackboardData->receiveStream );

                    if( domain )
                    {
                        if( !DOFObjectID_IsWildcard( DOFObjectIDDomain_GetObjectID( domain ) ) )
                        {
                            TRPRouter_Route_t route;
                            boolean ret = TRPRouter_GetDomainRoute(domain, &route);

                            if( ret )
                            {
                                if( route.isLocal )
                                {
                                    if( (data.op = TRP_V1_CreateASOperation( route.route.domain, opcode, self->blackboardData->associatedChannel, self->blackboardData->transportAddr, self->blackboardData->opID )) == NULL )
                                    {
                                        DOFObjectIDDomain_Destroy( domain );
                                        DOFAuthenticator_Destroy(route.route.domain);
                                        return PLRESULT_OKDONOTHING;
                                    }
                                    DOFAuthenticator_Destroy(route.route.domain);
                                }
                                else
                                {
                                    if( (data.op = TRP_V1_RelayCreate(domain, self->blackboardData->associatedChannel, self->blackboardData->transportAddr, self->blackboardData->opID, opcode, self->blackboardData->delayWithoutGracePeriod)) == NULL )
                                    {
                                        DOFObjectIDDomain_Destroy( domain );
                                        OALChannel_Release(route.route.channel);
                                        DOFAddress_Destroy(route.address);
                                        return PLRESULT_OKDONOTHING;
                                    }

                                    OALChannel_Release(route.route.channel);
                                    DOFAddress_Destroy(route.address);
                                }
                            }
                            else
                            {
                                OALStream stream = OALStream_CreateRW(4);
                                DOFObjectIDDomain_Destroy( domain );

                                if( stream )
                                {
                                    OALPacket_Release(self->blackboardData->sendPacket);
                                    self->blackboardData->sendPacket = OALPacket_Create();
                                    if( self->blackboardData->sendPacket )
                                    {
                                        if((((self->blackboardData->transportAddr)?DOFAddress_GetAddressType(self->blackboardData->transportAddr):DOFAddress_GetAddressType(self->blackboardData->associatedChannel->remoteChannelData->address)) == DOFADDRESSTYPE_UNICAST))
                                        {
                                            self->blackboardData->flags.isCommandNotResponse=FALSE;
                                            if( OALStream_Write16(stream, 0x8081) &&
                                                OALStream_Write8( stream, TRP_V1_OP_REJECT ) &&
                                                OALStream_Write8( stream, TRP_V1_UNKNOWN_DOMAIN ) &&
                                                OALPacket_AppendStream( self->blackboardData->sendPacket, stream ) )
                                            {
                                                return PLRESULT_OKSENDDOWN;
                                            }
                                        }
                                    }
                                    OALStream_Release( stream, TRUE );
                                }
                                return PLRESULT_OKDONOTHING;
                            }
                        }
                        else
                        {
                            if( (opcode == TRP_V1_OP_CREDENTIAL || opcode == TRP_V1_OP_REMOTE_DOMAIN) && TRPRouter_KnowsAnAS() )
                            {
                                if( (data.op = TRP_V1_RelayCreate(domain, self->blackboardData->associatedChannel, self->blackboardData->transportAddr, self->blackboardData->opID, opcode, self->blackboardData->delayWithoutGracePeriod)) == NULL )
                                {
                                    DOFObjectIDDomain_Destroy( domain );
                                    return PLRESULT_OKDONOTHING;
                                }
                            }
                            else
                            {
                                DOFObjectIDDomain_Destroy( domain );
                                return PLRESULT_OKDONOTHING;
                            }
                        }
                        DOFObjectIDDomain_Destroy( domain );
                        DOFOperation_Lock( data.op );
                        data.opData = (TRP_V1_Data)data.op->context;
                    }
                    else
                    {
                        return PLRESULT_OKDONOTHING;
                    }
                }
                else
                {
                    return PLRESULT_OKDONOTHING; /* Invalid packet */
                }
            }
            else
            {
                DOFOperation_Lock( data.op );
                if( data.op->operation_type >= DOFOPERATIONTYPE_TRP_V1_PASS && data.op->operation_type < DOFOPERATIONTYPE_TRP_LAST ) /* Verify that this is really a TRP op */
                {
                    data.opData = (TRP_V1_Data)data.op->context; 

                    if( data.op->flags.bits.completed || data.op->flags.bits.cancelled || (opcode == TRP_V1_OP_REJECT && (((data.op->sources->address)?DOFAddress_GetAddressType(data.op->sources->address):DOFAddress_GetAddressType(data.op->sources->channel->remoteChannelData->address)) != DOFADDRESSTYPE_UNICAST || data.opData->flags.bits.isDomainDiscovery)))
                    {
                        DOFOperation_Unlock( data.op );
                        DOFOperation_Internal_Destroy( data.op );
                        return PLRESULT_OKDONOTHING;
                    }
                    else if(( data.op->operation_type != DOFOPERATIONTYPE_TRP_V1_PASS && data.op->operation_type != DOFOPERATIONTYPE_TRP_V1_CREDENTIAL && data.op->operation_type != DOFOPERATIONTYPE_TRP_V1_REMOTE_DOMAIN && data.opData->opCode != TRP_V1_OP_DISCOVER_REMOTE) && !(data.op->operation_type == DOFOPERATIONTYPE_TRP_V1_AS && (data.opData->opCode == TRP_V1_OP_CREDENTIAL)))
                    {
                        if( !self->blackboardData->flags.isCommandNotResponse )
                        {
                            data.op->flags.bits.completed = 1;
                            DOFOperation_Internal_Cancel(data.op);
                        }
                        else
                        {
                            DOFOperation_Unlock( data.op );
                            DOFOperation_Internal_Destroy( data.op );
                            return PLRESULT_OKDONOTHING;
                        }
                    }

                    if( self->blackboardData->flags.isCommandNotResponse )
                    {
                        DOFOperation_SetTimeout( data.op, self->blackboardData->delayWithoutGracePeriod );
                    }
                }
                else
                {
                    return PLRESULT_OKDONOTHING;
                }
            }
        }
        else
        {
            return PLRESULT_OKDONOTHING;
        }
    }
    else
    {
        /* ERROR */
        return PLRESULT_ERROR;
    }

    data.stream = self->blackboardData->receiveStream ;
    data.opData->flags.bits.isCommand = self->blackboardData->flags.isCommandNotResponse;
    data.op->sources->retry_count = self->blackboardData->retry;

    if( data.op->operation_type != DOFOPERATIONTYPE_TRP_V1_PASS )
    {
        if( opcode <= 11 ) /* check bounds */
        {
            trp_functions[opcode][data.opData->flags.bits.isCommand](data);
        }
        else
        {
            TRP_V1_ReceiveInvalidOpCode( data );
        }
    }
    else
    {
        if(HandlePassThrough(data))
        {
            DOFOperation_Unlock( data.op );
            DOFOperation_Internal_Destroy(data.op);
            return PLRESULT_OKDONOTHING;
        }
    }

    switch( data.opData->trp_status )
    {
    case DOFERROR_DUPLICATE:
        data.opData->trp_status = DOFERROR_OK;
        DOFOperation_Unlock( data.op );
        DOFOperation_Internal_Destroy(data.op);
        return PLRESULT_OKDONOTHING;
    case DOFERROR_INVALID_REQUEST_STRUCT:
        DOFOperation_Unlock( data.op );
        TRP_V1_CallStateMachine(data.op);
        return PLRESULT_ERROR;
        break;
    default:
        DOFOperation_Unlock( data.op );
        TRP_V1_CallStateMachine(data.op);
        return PLRESULT_OKDONOTHING;
        break;
    }
}

static uint8 TRP_V1_MapErrorCode( uint8 errorCode )
{
    if(  errorCode == 8 ||
         errorCode > 127 ||
       ( errorCode > 15 && errorCode < 127 ) )
    {
        return 127;
    }
    return errorCode;
}

void TRP_V1_ReceiveReject( TRP_V1_ReceiveData_t data )
{
    data.opData->opCode = TRP_V1_OP_REJECT;
    if( TRP_V1_Read1ByteValue(data.stream, data.opData->errorCode) )
    {
        data.opData->errorCode = TRP_V1_MapErrorCode( data.opData->errorCode );
        if( data.opData->errorCode == TRP_V1_BAD_MODE )
        {
            uint32 len = OALStream_GetDataSize(data.stream);
            if( len > 0 )
            {
                const uint8* ptr = OALStream_GetReadPtr(data.stream);
                data.opData->errorData = DOFMem_Alloc(len);
                if( !data.opData->errorData )
                {
                    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unable to allocate memory for Reject data.");
                    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
                    return;
                }
                PCRMem_Copy(data.opData->errorData, ptr, len);
            }
        }

        data.opData->trp_status = DOFERROR_OK;
        return;
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unmarshal Failed.");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_ReceiveAuthorizeKEK( TRP_V1_ReceiveData_t data )
{
    TRP_V1_KEKData_t* kekData = data.opData->RequestData.kekData;
    uint8* start = NULL;

    if( TRP_V1_Unmarshal_KeyGrant( data.stream, &kekData->initKeyGrant ) )
    {
        start = (uint8*)OALStream_GetReadPtr( data.stream );

        if( TRP_V1_Unmarshal_KEKInitBlock( data.stream, kekData ) )
        {
            TRP_V1_HASH( data, start );
            TRP_V1_ReceiveAuthorize_Validate( data, DOFCredentials_GetKey( kekData->initKeyRequest->credentials ), kekData->initKeyGrant );
            return;
        }
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unmarshal Failed.");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_ReceiveRequestKEK( TRP_V1_ReceiveData_t data )
{
    TRP_V1_KEKData_t* kekData = data.opData->RequestData.kekData;

    data.opData->opCode = TRP_V1_OP_KEK;
    if( TRP_V1_Unmarshal_SecurityDomainID( data.stream, &data.opData->secDomain ) )
    {
        data.opData->I = (uint8*)OALStream_GetReadPtr( data.stream );

        if( TRP_V1_Unmarshal_KEKReqBlock( data.stream, kekData, data.opData->secDomain ) )
        {
            uint8* temp = (uint8*)OALStream_GetReadPtr( data.stream );
            data.opData->I_len = (temp - data.opData->I);
            OALStream_Retain( data.stream ); /* we retain till we send the response */
            data.opData->stream = data.stream;
            data.opData->flags.bits.retainedStream = 1;
            return;
        }
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unmarshal Failed.");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_ReceiveAuthorizeRandom( TRP_V1_ReceiveData_t data )
{
    TRP_V1_RandomData_t* randData = data.opData->RequestData.randomData;

    if( TRP_V1_Unmarshal_KeyGrant( data.stream, &randData->initKeyGrant ) )
    {
        TRP_V1_ReceiveAuthorize_Validate( data, DOFCredentials_GetKey( randData->initKeyRequest->credentials ), randData->initKeyGrant );
        return;
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unmarshal Failed.");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_ReceiveRequestRandom( TRP_V1_ReceiveData_t data )
{
    TRP_V1_RandomData_t* randData = data.opData->RequestData.randomData;

    data.opData->opCode = TRP_V1_OP_RANDOM;
    if( TRP_V1_Unmarshal_SecurityDomainID( data.stream, &data.opData->secDomain ) )
    {
        data.opData->I = (uint8*)OALStream_GetReadPtr( data.stream );

        if( TRP_V1_Unmarshal_KeyRequest( data.stream, &randData->initKeyRequest, data.opData->secDomain ) )
        {
            uint8* temp = (uint8*)OALStream_GetReadPtr( data.stream );
            data.opData->I_len = (temp - data.opData->I);
            OALStream_Retain( data.stream ); /* we retain till we send the response */
            data.opData->stream = data.stream;
            data.opData->flags.bits.retainedStream = 1;
            return;
        }
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unmarshal Failed.");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_ReceiveAuthorizeSession( TRP_V1_ReceiveData_t data )
{
    TRP_V1_SessionData_t* sessData = data.opData->RequestData.sessionData;
    uint8* start = NULL;

    if( TRP_V1_Unmarshal_KeyGrant( data.stream, &sessData->respKeyGrant ) &&
        TRP_V1_Unmarshal_KeyGrant( data.stream, &sessData->initKeyGrant ) )
    {
        data.opData->A = OALStream_CreateRW(OALStream_GetDataSize(data.stream));
        start = (uint8*)OALStream_GetReadPtr(data.stream);
        if( TRP_V1_Unmarshal_SessAuthBlock( data.stream, sessData ) )
        {
            OALStream_Write( data.opData->A, (uint32)(OALStream_GetReadPtr(data.stream)-start), start);
            TRP_V1_HASH( data, start );
            TRP_V1_ReceiveAuthorize_Validate( data, DOFCredentials_GetKey( sessData->respBlock->credentials ), sessData->respKeyGrant );
            return;
        }
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unmarshal Failed.");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_ReceiveRequestSession( TRP_V1_ReceiveData_t data )
{
    TRP_V1_SessionData_t* sessData = data.opData->RequestData.sessionData;

    data.opData->opCode = TRP_V1_OP_SESSION;
    if( TRP_V1_Unmarshal_SecurityDomainID( data.stream, &data.opData->secDomain ) )
    {
        data.opData->R = (uint8*)OALStream_GetReadPtr( data.stream );

        if( TRP_V1_Unmarshal_SessRespBlock( data.stream, sessData, data.opData->secDomain ) )
        {
            data.opData->I = (uint8*)OALStream_GetReadPtr( data.stream );
            data.opData->R_len = (data.opData->I - data.opData->R);
            if( TRP_V1_Unmarshal_SessInitBlock( data.stream, sessData, data.opData->secDomain ) )
            {
                uint8* temp = (uint8*)OALStream_GetReadPtr( data.stream );
                data.opData->I_len = (temp - data.opData->I);
                OALStream_Retain( data.stream ); /* we retain till we send the response */
                data.opData->stream = data.stream;
                data.opData->flags.bits.retainedStream = 1;
                return;
            }
        }
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unmarshal Failed.");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_ReceiveAuthorizeSecurityScope( TRP_V1_ReceiveData_t data )
{
    TRP_V1_SecurityScopeData_t* ssData = data.opData->RequestData.securityScopeData;
    uint8* start = NULL;

    if( TRP_V1_Unmarshal_KeyGrant( data.stream, &ssData->initKeyGrant ) )
    {
        start = (uint8*)OALStream_GetReadPtr(data.stream);
        if( OALStream_Read8( data.stream, &ssData->grantedDuration ) &&
            TRP_V1_Unmarshal_SecurityScopeInit( data.stream, ssData ))
        {
            TRP_V1_HASH( data, start );
            TRP_V1_ReceiveAuthorize_Validate( data, DOFCredentials_GetKey( ssData->initKeyRequest->credentials ), ssData->initKeyGrant );
            return;
        }
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unmarshal Failed.");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_ReceiveRequestSecurityScope( TRP_V1_ReceiveData_t data )
{
    TRP_V1_SecurityScopeData_t* ssData = data.opData->RequestData.securityScopeData;

    data.opData->opCode = TRP_V1_OP_SECURITY_SCOPE;
    if( TRP_V1_Unmarshal_SecurityDomainID( data.stream, &data.opData->secDomain ) )
    {
        data.opData->I = (uint8*)OALStream_GetReadPtr( data.stream );
        if( OALStream_Read8( data.stream, &ssData->grantedDuration ) &&
            TRP_V1_Unmarshal_KeyRequest( data.stream, &ssData->initKeyRequest, data.opData->secDomain ) &&
            TRP_V1_Unmarshal_SecurityNodeID( data.stream, &ssData->secNodeID ) )
        {
            uint8* temp;
            if(data.opData->opCode == TRP_V1_OP_SECURITY_SCOPE_EXT &&
               OALStream_Read8(data.stream, ((uint8*)&ssData->secType)))
            {
                ssData->modeLen = OALStream_GetDataSize(data.stream);
                ssData->mode = DOFMem_Alloc(ssData->modeLen);
                if(!ssData->mode)
                {
                    goto FAIL;
                }
                if(!OALStream_Read(data.stream, ssData->modeLen, ssData->mode) ||
                   !((*((uint16*)(ssData->mode+1)) >= 0x6000) && (*((uint16*)(ssData->mode+1)) < 0x7000)) ||
                   !((*((uint8*)(ssData->mode+3))) == ssData->modeLen - 4) )
                {
                    goto FAIL;
                }
            }

            temp = (uint8*)OALStream_GetReadPtr( data.stream );
            data.opData->I_len = (temp - data.opData->I);
            OALStream_Retain( data.stream ); /* we retain till we send the response */
            data.opData->stream = data.stream;
            data.opData->flags.bits.retainedStream = 1;
            return;
        }
    }
FAIL:
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unmarshal Failed.");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_ReceiveAuthorizeCredential( TRP_V1_ReceiveData_t data )
{
    TRP_V1_CredentialData_t* credData = data.opData->RequestData.credentialData;
    CredentialsUnmarshalReturn ret = DOFCredentials_Unmarshal( &credData->credentials, data.opData->secDomain, data.stream, FALSE );

    if( ret == CREDENTIALS_SUCCESSFUL_UNMARSHAL )
    {
        return;
    }
    else if( ret == CREDENTIALS_WRONGSTAGE_UNMARSHAL )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Received wrong credential stage in packet.");
        data.opData->trp_status = DOFERROR_DUPLICATE;
        return;
    }

    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unmarshal Failed.");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}


void TRP_V1_ReceiveRequestCredential( TRP_V1_ReceiveData_t data )
{
    TRP_V1_CredentialData_t* credData = data.opData->RequestData.credentialData;
    CredentialsUnmarshalReturn ret = CREDENTIALS_FAILED_UNMARSHAL;
    DOFObjectIDDomain temp = NULL;

    data.opData->opCode = TRP_V1_OP_CREDENTIAL;
    if( (TRP_V1_Unmarshal_SecurityDomainID( data.stream, &temp)) &&
        (ret = DOFCredentials_Unmarshal( &credData->credentials, temp, data.stream, TRUE )) == CREDENTIALS_SUCCESSFUL_UNMARSHAL )
    {
        DOFObjectIDDomain_Destroy(temp);
        return;
    }

    DOFObjectIDDomain_Destroy(temp);

    if( ret == CREDENTIALS_FAILED_UNMARSHAL )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unmarshal Failed.");
        data.opData->trp_status = DOFERROR_REQUEST_FAILED;
    }
    else
    {
        data.opData->trp_status = DOFERROR_DUPLICATE;
    }
}

void TRP_V1_ReceiveAuthorizeDomainCredential( TRP_V1_ReceiveData_t data )
{
    TRP_V1_CredentialData_t* credData = data.opData->RequestData.credentialData;
    DOFObjectIDDomain temp;

    if( TRP_V1_Unmarshal_SecurityDomainID( data.stream, &temp) &&
        (DOFCredentials_Unmarshal( &credData->credentials, temp, data.stream, FALSE ) == CREDENTIALS_SUCCESSFUL_UNMARSHAL) )
    {
        TRP_V1_ResolveDomainCredentials( data.op, temp, credData->credentials, data.trp->blackboardData->associatedChannel, data.trp->blackboardData->transportAddr );
    }
    DOFObjectIDDomain_Destroy( temp );
    data.opData->trp_status = DOFERROR_DUPLICATE;
}

void TRP_V1_ReceiveAuthorizeLocalDomain( TRP_V1_ReceiveData_t data )
{
    TRP_V1_DomainData_t* domData = data.opData->RequestData.domainData;
    uint8* start = NULL;

    if( TRP_V1_Unmarshal_KeyGrant( data.stream, &domData->initLocalKeyGrant ) )
    {
        start = (uint8*)OALStream_GetReadPtr(data.stream);
        if( DOFERROR_OK == OALMarshal_StreamUncompress4( data.stream, &domData->localDomainRespBlock) )
        {
            TRP_V1_HASH( data, start );
            TRP_V1_ReceiveAuthorize_Validate( data, DOFCredentials_GetKey( domData->initRemoteKeyRequest->credentials ), domData->initLocalKeyGrant );
            return;
        }
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unmarshal Failed.");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_ReceiveRequestLocalDomain( TRP_V1_ReceiveData_t data )
{
    TRP_V1_DomainData_t* domData = data.opData->RequestData.domainData;

    data.opData->opCode = TRP_V1_OP_LOCAL_DOMAIN;
    if( TRP_V1_Unmarshal_SecurityDomainID( data.stream, &domData->localDomain ) &&
        TRP_V1_Unmarshal_SecurityDomainID( data.stream, &domData->remDomain ) )
    {
        data.opData->secDomain = DOFObjectIDDomain_Copy(domData->localDomain);
        if(data.opData->secDomain)
        {
            data.opData->I = (uint8*)OALStream_GetReadPtr(data.stream);
            if(TRP_V1_Unmarshal_KeyRequest( data.stream, &domData->initRemoteKeyRequest, domData->localDomain ))
            {
                uint8* temp = (uint8*)OALStream_GetReadPtr( data.stream );
                data.opData->I_len = (temp - data.opData->I);
                OALStream_Retain( data.stream ); /* we retain till we send the response */
                if( (domData->initRemoteKeyGrant = OALKeyGrant_Unmarshal( data.stream )) )
                {
                    domData->responseA = (uint8*)OALStream_GetReadPtr(data.stream);
                    if( TRP_V1_Unmarshal_SecurityNodeID( data.stream, &domData->remNodeID ) )
                    {
                        temp = (uint8*)OALStream_GetReadPtr( data.stream );
                        domData->responseA_len = (temp - domData->responseA);
                        data.opData->stream = data.stream;
                        data.opData->flags.bits.retainedStream = 1;
                        return;
                    }
                }
                OALStream_Release( data.stream, TRUE );
            }
        }
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unmarshal Failed.");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_ReceiveAuthorizeRemoteDomain( TRP_V1_ReceiveData_t data )
{
    TRP_V1_DomainData_t* domData = data.opData->RequestData.domainData;

    if( TRP_V1_Unmarshal_KeyGrant_ToStream( data.stream, domData->rd ) &&
        TRP_V1_Unmarshal_SecurityNodeID( data.stream, &domData->remNodeID ) )
    {
        return;
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unmarshal Failed.");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_ReceiveRequestRemoteDomain( TRP_V1_ReceiveData_t data )
{
    TRP_V1_DomainData_t* domData = data.opData->RequestData.domainData;

    data.opData->opCode = TRP_V1_OP_REMOTE_DOMAIN;
    if( TRP_V1_Unmarshal_SecurityDomainID( data.stream, &domData->remDomain ) &&
        TRP_V1_Unmarshal_SecurityDomainID( data.stream, &domData->localDomain ) )
    {
        data.opData->secDomain = DOFObjectIDDomain_Copy(domData->remDomain);
        if(data.opData->secDomain)
        {
            data.opData->I = (uint8*)OALStream_GetReadPtr(data.stream);
            if( TRP_V1_Unmarshal_KeyRequest( data.stream, &domData->initRemoteKeyRequest, domData->localDomain ))
            {
                uint8* temp = (uint8*)OALStream_GetReadPtr( data.stream );
                data.opData->I_len = (temp - data.opData->I);
                OALStream_Retain( data.stream ); /* we retain till we send the response */
                data.opData->stream = data.stream;
                data.opData->flags.bits.retainedStream = 1;
                return;
            }
        }
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unmarshal Failed.");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_ReceiveAuthorizeValidateCreds( TRP_V1_ReceiveData_t data )
{
    TRP_V1_ValidateCredsData_t* vcData = data.opData->RequestData.validateData;

    vcData->bLen = OALStream_GetDataSize(data.stream);
    vcData->validB = PCRMem_Alloc(vcData->bLen);
    if( OALStream_Read(data.stream, vcData->bLen, vcData->validB) )
    {
        return;
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unmarshal Failed.");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_ReceiveRequestValidateCreds( TRP_V1_ReceiveData_t data )
{
    TRP_V1_ValidateCredsData_t* vcData = data.opData->RequestData.validateData;

    data.opData->opCode = TRP_V1_OP_VALIDATE_CREDS;
    if( TRP_V1_Unmarshal_SecurityDomainID( data.stream, &data.opData->secDomain ) &&
        (DOFCredentials_Unmarshal( &vcData->credentials, data.opData->secDomain, data.stream, TRUE ) != CREDENTIALS_FAILED_UNMARSHAL) &&
        OALStream_Read( data.stream, 40, vcData->validA) )
    {
        return;
    }
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Unmarshal Failed.");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_ReceiveAuthorizeRemoteDiscover( TRP_V1_ReceiveData_t data )
{
    TRP_V1_DomainData_t* domData = data.opData->RequestData.domainData;
    DOFObjectIDDomain temp = NULL;

    if( TRP_V1_Unmarshal_SecurityDomainID( data.stream, &temp) &&
        TRP_V1_Unmarshal_KeyGrant_ToStream( data.stream, domData->rd ) &&
        TRP_V1_Unmarshal_SecurityNodeID( data.stream, &domData->remNodeID ) )
    {
        DOFKeyRequest kr = NULL;
        DOFCredentials tempc = DOFCredentials_Copy(domData->initRemoteKeyRequest->credentials);

        kr = DOFKeyRequest_Create(tempc, DEFAULT_NONCE_SIZE, TRUE, FALSE, NULL, domData->initRemoteKeyRequest->permission_set);

        TRP_V1_RequestRemoteDomainDiscovery(temp,data.op,kr,data.trp->blackboardData->associatedChannel,data.trp->blackboardData->transportAddr, DOFOperation_GetTimeRemaining(data.op));

        DOFObjectIDDomain_Destroy(temp);
        DOFCredentials_Destroy(tempc);
        DOFKeyRequest_Destroy(kr);
        data.opData->trp_status = DOFERROR_DUPLICATE;
        return;
    }
    DOFObjectIDDomain_Destroy(temp);
}

void TRP_V1_ReceiveInvalidOpCode( TRP_V1_ReceiveData_t data )
{
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Received invalid op code.");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_CallStateMachine( DOFOperation op )
{
    TRP_V1_SM_Call(op);
}

