/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/dof.h>
#include <dof/cfg.h>
#include <dof/system.h>
#include <dof/oaltypes.h>
#include <dof/credentials.h>
#include <internal/router.h>
#include <internal/stream.h>
#include <internal/trp_v1.h>
#include <internal/srp.h>
#include <internal/objectidinternal.h>
#include <internal/credentialsinternal.h>
#include <internal/blackboarddata.h>
#include <internal/permissionsetinternal.h>
#include <internal/securitymode.h>
#include <internal/security_common.h>
#include "trp_v1_receive_util.h"
#include "trp_v1_receive.h"
#include "trp_v1_utils.h"

boolean TRP_V1_PermissionSet_Read(OALStream stream, DOFPermissionSet* permSet)
{
    return ((*permSet = DOFPermissionSet_UnmarshalRequest( stream, FALSE )) != NULL);
}

boolean TRP_V1_ReadStateID( OALStream stream, OALStateIdentifier* stateID )
{
    return (OALStateIdentifier_Read(stream, *stateID) == DOFERROR_OK);
}

void TRP_V1_HASH(TRP_V1_ReceiveData_t opData, uint8* start)
{
    uint8* end = (uint8*)OALStream_GetReadPtr(opData.stream);
    OALSecureHMAC_Digest( &opData.opData->ctx, (uint32)(end-start), start );
}

boolean TRP_V1_Unmarshal_KeyRequest( OALStream stream, DOFKeyRequest* keyRequest, DOFObjectIDDomain domain )
{
    if( *keyRequest == NULL )
    {
        *keyRequest = DOFMem_Alloc( sizeof( DOFKeyRequest_t ) );
        if( *keyRequest == NULL )
            return FALSE;
        PCRMem_Set( *keyRequest, 0, sizeof( DOFKeyRequest_t ));
    }
    return DOFKeyRequest_Unmarshal( stream, *keyRequest, domain, TRUE, TRUE );
}

boolean TRP_V1_Unmarshal_KeyGrant( OALStream stream, OALKeyGrant* keyGrant )
{
    (*keyGrant) = OALKeyGrant_Unmarshal( stream );
    return ((*keyGrant) != NULL );
}

boolean TRP_V1_Unmarshal_KeyGrant_ToStream( OALStream from,  OALStream to )
{
    uint8* grantPos = OALStream_GetWritePtr(to);

    if( OALStream_GetDataSize(from) >= 64 &&
        OALStream_GetFreeSize(to) >= 64 &&
        OALStream_Read(from, 64, grantPos))
    {
        OALStream_AdjustWriteOffset(to,64);

        return TRUE;
    }
    return FALSE;
}

boolean TRP_V1_Unmarshal_SourceNodeID( OALStream stream, DOFObjectIDSource* secNode )
{
    (*secNode) = (DOFObjectIDSource)DOFObjectID_Unmarshal( stream );
    if( !DOFObjectID_IsSourceID(DOFObjectIDSource_GetObjectID((*secNode))))
    {
        DOFObjectIDSource_Destroy((*secNode));
        (*secNode) = NULL;
    }
    return ((*secNode) != NULL );
}

boolean TRP_V1_Unmarshal_SecurityNodeID( OALStream stream, DOFObjectIDAuthentication* secNode )
{
    (*secNode) = DOFObjectIDAuthentication_Unmarshal( stream );
    return ((*secNode) != NULL );
}

boolean TRP_V1_Unmarshal_SecurityDomainID( OALStream stream, DOFObjectIDDomain* secDomain )
{
    (*secDomain) = DOFObjectIDDomain_Unmarshal( stream );
    return ((*secDomain) != NULL );
}

boolean TRP_V1_Unmarshal_KEKReqBlock( OALStream stream, TRP_V1_KEKData_t* kekData, DOFObjectIDDomain domain )
{
    if( TRP_V1_Unmarshal_KeyRequest( stream, &kekData->initKeyRequest, domain ) &&
        TRP_V1_Unmarshal_SecurityNodeID( stream, &kekData->groupID ) )
        return TRUE;
    return FALSE;
}



boolean TRP_V1_Unmarshal_KEKInitBlock( OALStream stream, TRP_V1_KEKData_t* kekData )
{
    if( TRP_V1_Read1ByteValue( stream, kekData->thb ) &&
        TRP_V1_Read1ByteValue( stream, kekData->tmin ) &&
        TRP_V1_Read1ByteValue( stream, kekData->tmax ) &&
        TRP_V1_Read2ByteValue( stream, kekData->epoch )  &&
        TRP_V1_Unmarshal_SourceNodeID( stream, &kekData->sid) &&
        (DOFObjectID_GetIDClass(DOFObjectIDSource_GetObjectID(kekData->sid)) == DOFOBJECTIDCLASS_AS) &&
        ((kekData->secScope = DOFSecurityScope_Unmarshal( stream, DOFCredentials_GetDomainID( kekData->initKeyRequest->credentials ))) != NULL )&&
        ((kekData->securityMode = DOFSecurityMode_Create_UnmarshalFromStream(stream)) != NULL ) &&
        (OALMarshal_StreamUncompress4( stream, &kekData->stateID ) == DOFERROR_OK) &&
        TRP_V1_PermissionSet_Read( stream, &kekData->groupsPermissions ) &&
        DOFPermissionSet_UnmarshalValidation( kekData->groupsPermissions, stream, DOFSecurityScope_GetOALPermissionScope(kekData->secScope) ) &&
        DOFPermissionSet_UnmarshalValidation( kekData->initKeyRequest->permission_set, stream, DOFSecurityScope_GetOALPermissionScope(kekData->secScope) ))
    {
        DOFPermissionSet_SetGrantPerms(kekData->groupsPermissions);
        return TRUE;
    }
    return FALSE;
}

boolean TRP_V1_Unmarshal_SessInitBlock( OALStream stream, TRP_V1_SessionData_t* sessData, DOFObjectIDDomain domain )
{
    if( (sessData->initBlock = OALSessionInitiatorBlock_Unmarshal(stream, domain, TRUE, TRUE)) != NULL )
        return TRUE;
    return FALSE;
}

boolean TRP_V1_Unmarshal_SessRespBlock( OALStream stream, TRP_V1_SessionData_t* sessData, DOFObjectIDDomain domain )
{
    if( !sessData->respBlock )
        sessData->respBlock = DOFKeyRequest_Create( NULL, 0, FALSE, FALSE, NULL, NULL );

    return DOFKeyRequest_Unmarshal( stream, sessData->respBlock, domain, FALSE, TRUE );
}

boolean TRP_V1_Unmarshal_SessAuthBlock( OALStream stream, TRP_V1_SessionData_t* sessData )
{
    return OALSessionAuthRespBlock_Unmarshal( &sessData->grantedDuration, &(sessData->connSecurityScope), sessData->initBlock->keyRequest->permission_set, sessData->respBlock->permission_set, DOFCredentials_GetDomainID( sessData->respBlock->credentials ), stream );
}

boolean TRP_V1_Unmarshal_SecurityScopeInit( OALStream stream, TRP_V1_SecurityScopeData_t* ssData )
{
    if( (((ssData->securityScope = DOFSecurityScope_Unmarshal( stream,  DOFCredentials_GetDomainID( ssData->initKeyRequest->credentials )))) != NULL) &&
         DOFPermissionSet_UnmarshalValidation( ssData->initKeyRequest->permission_set, stream,  DOFSecurityScope_GetOALPermissionScope(ssData->securityScope) ) )
        return TRUE;
    return FALSE;
}

