/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/dof.h>
#include <dof/system.h>
#include <dof/credentials.h>
#include <internal/bridgeinternal.h>
#include <internal/credentialsinternal.h>
#include <internal/securitymode.h>
#include <internal/security_common.h>
#include <internal/stream.h>
#include <internal/trp_v1.h>
#include <internal/dofinternal.h>
#include <internal/error.h>
#include <internal/operationinternal.h>
#include <internal/resultinternal.h>

#include "trp_v1_utils.h"
#include "trp_v1_send.h"
#include "trp_v1_receive.h"

typedef void (*TRP_V1_SendFunction)(TRP_V1_SendData_t data);

#define TRP_V1_Write1ByteValue(stream, valueHolder) OALStream_Write8(stream, valueHolder)
#define TRP_V1_Write2ByteValue(stream, valueHolder) OALStream_Write16(stream, valueHolder)
#define TRP_V1_WriteSignature(stream, signature )   OALStream_Write(stream, 32, signature)

static void AddPassThroughToperationpacket(TRP_V1_SendData_t data)
{
    TRP_V1_PassData_t* passData = data.opData->RequestData.passData;
    if( !data.result )
    {
        OALStream stream = OALStream_CreateRW(passData->length);
        if( stream )
        {
            /* fill stream */
            if( OALStream_Write(stream, passData->length, passData->streamData) )
            {
                /* add to packet */
                if(OALPacket_AppendStream(data.packet, stream))
                {
                    data.opData->trp_status = DOFERROR_OK;
                    return;
                }
            }
            OALStream_Release(stream, TRUE);
        }
        data.opData->trp_status = DOFERROR_REQUEST_FAILED;
    }
    else
    {
        if( data.trp->blackboardData->seq < 255 )
        {
            OALStream strm = DOFResult_GetStream(data.result);
            if( strm )
            {
                if( OALPacket_AppendStream(data.packet, strm) ) 
                {
                    OALStream_Retain(strm);
                    DOFAddress_Destroy(data.trp->blackboardData->transportAddr);
                    data.trp->blackboardData->transportAddr = DOFAddress_Copy( DOFResult_GetAddress(data.result) );
                    data.trp->blackboardData->flags.isCommandNotResponse = 0;
                    data.trp->blackboardData->seq++;
                    data.opData->trp_status = DOFERROR_OK;
                    return;
                }
            }
        }
    }
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

static void TRP_V1_HandleHash(TRP_V1_SendData_t opdata, const uint8* data, uint32 dataLen)
{
    OALSecureHMAC_Digest( &opdata.opData->ctx, dataLen, data );
}

ProtocolLayerResult TRP_V1_Send(TRP_V1 self, DOFOperation operation, boolean isProviderOperation, DOFResult result)
{
    TRP_V1_SendData_t data;
    OALStream stream;
    uint16 appid = 0x8081;
    boolean forwarded = FALSE;

    static const TRP_V1_SendFunction trp_functions[12][2]=
        {
            {TRP_V1_SendReject,    TRP_V1_SendReject},                  /* TRP_V1_OP_REJECT                */
            {TRP_V1_SendAuthorize, TRP_V1_SendRequestKEK},              /* TRP_V1_OP_KEK                   */
            {TRP_V1_SendAuthorize, TRP_V1_SendRequestRandom},           /* TRP_V1_OP_RANDOM                */
            {TRP_V1_SendAuthorize, TRP_V1_SendRequestSession},          /* TRP_V1_OP_SESSION               */
            {TRP_V1_SendAuthorize, TRP_V1_SendRequestSecurityScope},    /* TRP_V1_OP_SECURITY_SCOPE        */
            {TRP_V1_SendAuthorize, TRP_V1_SendInvalidOpCode},           /* TRP_V1_OP_CREDENTIAL DISCOVERY  */
            {TRP_V1_SendAuthorize, TRP_V1_SendRequestCredential},       /* TRP_V1_OP_CREDENTIAL            */
            {TRP_V1_SendAuthorize, TRP_V1_SendRequestLocalDomain},      /* TRP_V1_OP_LOCAL_DOMAIN          */
            {TRP_V1_SendAuthorize, TRP_V1_SendRequestRemoteDomain},     /* TRP_V1_OP_REMOTE_DOMAIN         */
            {TRP_V1_SendAuthorize, TRP_V1_SendRequestValidateCreds},    /* TRP_V1_OP_VALIDATE_CREDS        */
            {TRP_V1_SendAuthorize, TRP_V1_SendInvalidOpCode},           /* TRP_V1_OP_DISCOVER_REMOTE       */
            {TRP_V1_SendAuthorize, TRP_V1_SendRequestSecurityScope}     /* TRP_V1_OP_SECURITY_SCOPE 0x0B   */
        };

    data.trp = self;
    data.packet =  self->blackboardData->sendPacket;
    data.result = result;

    data.op = operation;

    DOFOperation_Lock( data.op );
    data.opData = (TRP_V1_Data)data.op->context;
    data.opData->trp_status = DOFERROR_OK;

    if( data.opData->fromDomain && !DOFObjectID_IsEqual( DOFObjectIDDomain_GetObjectID( data.opData->fromDomain ), DOFObjectIDDomain_GetObjectID( DOFSecurityMembrane_GetDomainID( self->blackboardData->associatedChannel->securityMembrane ) ) ) )
    {
        forwarded = TRUE;
        if(!OALOperationManager_SetupForwardingOperation(data.op))
        {
            DOFOperation_Unlock( data.op );
            return PLRESULT_OKDONOTHING; /* can't create forward op */
        }
    }

    if( !result && ( data.opData->flags.bits.isCommand && ( !self->blackboardData->associatedChannel->remoteChannelData->flags.TRP_SendCommands || data.op->operation_type == DOFOPERATIONTYPE_TRP_V1_AS ) ) )
    {
        DOFOperation_Unlock( data.op );
        return PLRESULT_OKDONOTHING; /* can't send it on this channel */
    }

    /* get trp context data */
    self->blackboardData->flags.isCommandNotResponse = data.opData->flags.bits.isCommand;
    self->blackboardData->retry = data.op->sources->retry_count;
    self->blackboardData->seq = 1;
    self->blackboardData->delay = DOFOperation_GetTimeRemaining(data.op);

    if(!forwarded)
    {
        self->blackboardData->opID = DOFOperationID_Copy(data.op->opID);
    }
    else
    {
        PCR_Assert(data.op->forwardingId);
        self->blackboardData->opID = DOFOperationID_Copy(data.op->forwardingId);
    }

    if( !self->blackboardData->opID )
    {
        DOFOperation_Unlock(data.op);
        return PLRESULT_ERROR;
    }


    /* this is only done for non pass through packets because the channels are switched for pass through channels in AddPassThroughToperationpacket */
    if( !data.opData->flags.bits.isCommand && (data.opData->secure != self->blackboardData->associatedChannel->remoteChannelData->flags.sessionIsSecure))
    {
        DOF_LogMessage( DOF_GetDof(),DOFLOGLEVEL_WARNING, "TRP V1: Dropping packet because security session has changed.");
        DOFOperation_Unlock(data.op);
        return PLRESULT_OKDONOTHING;
    }

    /* write APPID */
    stream = OALStream_CreateRW(2);
    if( stream )
    {
        OALStream_Write16(stream, appid);
        if(!OALPacket_AppendStream(data.packet, stream))
        {
            DOFOperation_Unlock( data.op );
            return PLRESULT_ERROR;
        }
    }
    else
    {
        DOFOperation_Unlock( data.op );
        return PLRESULT_ERROR;
    }

    if( data.op->operation_type != DOFOPERATIONTYPE_TRP_V1_PASS )
    {
        if( data.opData->opCode <= 11 ) /* check bounds */
            trp_functions[data.opData->opCode][data.opData->flags.bits.isCommand](data);
        else
            TRP_V1_SendInvalidOpCode( data );
    }
    else
    {
        AddPassThroughToperationpacket( data );
    }

    switch( data.opData->trp_status )
    {
    case DOFERROR_INVALID_REQUEST_STRUCT:
        DOFOperation_Unlock( data.op );
        return PLRESULT_ERROR;
        break;
    case DOFERROR_DUPLICATE:
        DOFOperation_Unlock( data.op );
        return PLRESULT_OKDONOTHING;
        break;
    case DOFERROR_REQUEST_FAILED:
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "TRP V1: Error occured while trp was trying to send.");
        return PLRESULT_OKDONOTHING;
    case DOFERROR_OK:
    default:
        DOFOperation_Unlock( data.op );
        return PLRESULT_OKSENDDOWN;
        break;
    }
}

void TRP_V1_SendAuthorize( TRP_V1_SendData_t data )
{
    if( data.opData->A )
    {
        if( OALPacket_AppendStream(data.packet, data.opData->A) )
        {
            data.opData->A = NULL;
            if(!data.opData->flags.bits.ASDontCancel )
                DOFOperation_Internal_Cancel(data.op);
            return;
        }
        data.opData->trp_status = DOFERROR_REQUEST_FAILED;
    }
    data.opData->trp_status = DOFERROR_DUPLICATE;
}

void TRP_V1_SendReject( TRP_V1_SendData_t data )
{
    OALStream stream = OALStream_CreateRW(2);

    data.trp->blackboardData->seq = 1;

    if( stream )
    {
        if( TRP_V1_Write1ByteValue( stream, TRP_V1_OP_REJECT ) &&
            TRP_V1_Write1ByteValue( stream, data.opData->errorCode ) &&
            OALPacket_AppendStream( data.packet, stream ) )
        {
            return;
        }
        OALStream_Release( stream, TRUE );
    }
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_SendRequestKEK(TRP_V1_SendData_t data)
{
    TRP_V1_KEKData_t* KEKData = data.opData->RequestData.kekData;
    uint32 hashLength = DOFKeyRequest_MarshallByteLength( KEKData->initKeyRequest, TRUE ) +
                        DOFObjectID_GetBytesSize( DOFObjectIDAuthentication_GetObjectID( KEKData->groupID ));
    uint32 streamLen = 1 + DOFObjectID_GetBytesSize( DOFObjectIDDomain_GetObjectID( data.opData->secDomain )) + hashLength;

    OALStream stream = OALStream_CreateRW(streamLen);

    data.trp->blackboardData->seq = 1;

    if( stream )
    {
        if( TRP_V1_Write1ByteValue(stream, TRP_V1_OP_KEK) &&
            DOFObjectID_Marshal( DOFObjectIDDomain_GetObjectID( data.opData->secDomain ), stream ) )
        {
            const uint8* writePtr = OALStream_GetWritePtr(stream);
            if( DOFKeyRequest_Marshal( stream, KEKData->initKeyRequest, TRUE ) &&
                DOFObjectID_Marshal( DOFObjectIDAuthentication_GetObjectID( KEKData->groupID ), stream ) )
            {
                PCR_Assert( OALStream_GetFreeSize(stream) == 0 );
                TRP_V1_HandleHash( data, writePtr, hashLength );
                if(OALPacket_AppendStream(data.packet, stream))
                    return;
            }
        }
        OALStream_Release( stream, TRUE );
    }
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_SendRequestRandom(TRP_V1_SendData_t data)
{
    TRP_V1_RandomData_t* RandData = data.opData->RequestData.randomData;
    uint32 hashLength = DOFKeyRequest_MarshallByteLength( RandData->initKeyRequest, TRUE ) ;
    DOFObjectID domain = DOFObjectIDDomain_GetObjectID( DOFCredentials_GetDomainID( RandData->initKeyRequest->credentials ) );
    uint32 streamLen = 1 + (domain ? DOFObjectID_GetBytesSize( domain ) : 0) + hashLength;

    OALStream stream = OALStream_CreateRW(streamLen);

    data.trp->blackboardData->seq = 1;

    if( stream )
    {
        if( TRP_V1_Write1ByteValue( stream, TRP_V1_OP_RANDOM ) &&
            DOFObjectID_Marshal( domain, stream ) )
        {
            const uint8* writePtr = OALStream_GetWritePtr( stream );
            if( DOFKeyRequest_Marshal( stream, RandData->initKeyRequest, TRUE ) )
            {
                TRP_V1_HandleHash( data, writePtr, hashLength );
                if(OALPacket_AppendStream(data.packet, stream))
                    return;
            }
        }
        OALStream_Release( stream, TRUE );
    }
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_SendRequestSession(TRP_V1_SendData_t data)
{
    TRP_V1_SessionData_t* sessData = data.opData->RequestData.sessionData;
    uint32 hashLength = OALStream_GetDataSize(sessData->R) + OALStream_GetDataSize(sessData->I);
    DOFObjectID domain = DOFObjectIDDomain_GetObjectID( DOFCredentials_GetDomainID( sessData->respBlock->credentials ) );
    uint32 streamLen = 1 + (domain ? DOFObjectID_GetBytesSize( domain ) : 0) + hashLength;

    OALStream stream = OALStream_CreateRW(streamLen);

    data.trp->blackboardData->seq = 1;

    if( stream )
    {
        uint32 rRead = OALStream_GetReadOffset(sessData->R);
        uint32 iRead = OALStream_GetReadOffset(sessData->I);

        if( TRP_V1_Write1ByteValue(stream, TRP_V1_OP_SESSION) &&
            DOFObjectID_Marshal( domain, stream ) &&
            OALStream_ReadToStream(sessData->R,OALStream_GetDataSize(sessData->R), stream) &&
            OALStream_ReadToStream(sessData->I,OALStream_GetDataSize(sessData->I), stream))
        {
            OALStream_SetReadOffset(sessData->R, rRead);
            OALStream_SetReadOffset(sessData->I, iRead);

            TRP_V1_HandleHash( data, OALStream_GetReadPtr(sessData->I), OALStream_GetDataSize(sessData->I));
            TRP_V1_HandleHash( data, OALStream_GetReadPtr(sessData->R), OALStream_GetDataSize(sessData->R));

            if(OALPacket_AppendStream(data.packet, stream))
                return;
        }
        OALStream_Release( stream, TRUE );
    }
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_SendRequestSecurityScope( TRP_V1_SendData_t data )
{
    TRP_V1_SecurityScopeData_t* scopeData = data.opData->RequestData.securityScopeData;
    uint32 hashLength =  1 + DOFKeyRequest_MarshallByteLength( scopeData->initKeyRequest, TRUE ) + DOFObjectID_GetBytesSize( DOFObjectIDAuthentication_GetObjectID( scopeData->secNodeID )) +
                        ((data.opData->opCode == TRP_V1_OP_SECURITY_SCOPE_EXT)?scopeData->modeLen+1:0);
    DOFObjectID domain = DOFObjectIDDomain_GetObjectID( DOFCredentials_GetDomainID( scopeData->initKeyRequest->credentials ) );
    uint32 streamLen = 1 + (domain ? DOFObjectID_GetBytesSize( domain ) : 0) + hashLength;

    OALStream stream = OALStream_CreateRW(streamLen);

    data.trp->blackboardData->seq = 1;

    if( stream )
    {
        if( TRP_V1_Write1ByteValue( stream, data.opData->opCode) &&
            DOFObjectID_Marshal( domain, stream ) )
        {
            const uint8* writePtr = OALStream_GetWritePtr( stream );
            if( OALStream_Write8( stream, scopeData->grantedDuration ) &&
                DOFKeyRequest_Marshal( stream, scopeData->initKeyRequest, TRUE ) &&
                DOFObjectID_Marshal( DOFObjectIDAuthentication_GetObjectID(scopeData->secNodeID), stream ))
            {
                if( data.opData->opCode == TRP_V1_OP_SECURITY_SCOPE ||
                   (data.opData->opCode == TRP_V1_OP_SECURITY_SCOPE_EXT &&
                    OALStream_Write8(stream, scopeData->secType) &&
                    OALStream_Write(stream, scopeData->modeLen, scopeData->mode)))
                {
                    TRP_V1_HandleHash( data, writePtr, hashLength );
                    if(OALPacket_AppendStream(data.packet, stream))
                        return;
                }
            }
        }
        OALStream_Release( stream, TRUE );
    }
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_SendRequestCredential( TRP_V1_SendData_t data )
{
    TRP_V1_CredentialData_t* credData = data.opData->RequestData.credentialData;
    uint32 streamLen = 1 + DOFObjectID_GetBytesSize(  DOFObjectIDDomain_GetObjectID(DOFCredentials_GetDomainID(credData->credentials)) ) + DOFCredentials_MarshalSize( credData->credentials, TRUE );

    OALStream stream = OALStream_CreateRW(streamLen);

    data.trp->blackboardData->seq = 1;

    if( stream )
    {
        if( TRP_V1_Write1ByteValue( stream, TRP_V1_OP_CREDENTIAL ) &&
            DOFObjectID_Marshal( DOFObjectIDDomain_GetObjectID( DOFCredentials_GetDomainID(credData->credentials) ), stream ) &&
            DOFCredentials_Internal_Marshal( credData->credentials, stream, TRUE ) &&
            OALPacket_AppendStream(data.packet, stream) )
        {
            return;
        }
        OALStream_Release( stream, TRUE );
    }
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_SendRequestLocalDomain( TRP_V1_SendData_t data )
{
    TRP_V1_DomainData_t* ldData = data.opData->RequestData.domainData;
    uint32 hashLength =  OALStream_GetDataSize(ldData->rd)-64;
    uint32 streamLen = 1 + DOFObjectID_GetBytesSize( DOFObjectIDDomain_GetObjectID( ldData->localDomain ) ) +
                       DOFObjectID_GetBytesSize( DOFObjectIDDomain_GetObjectID( ldData->remDomain )) +
                       hashLength +
                       OALKeyGrant_MarshalSize( ldData->initRemoteKeyGrant ) +
                       DOFObjectID_GetBytesSize( DOFObjectIDAuthentication_GetObjectID( ldData->remNodeID ) );
    OALStream stream = OALStream_CreateRW(streamLen);

    data.trp->blackboardData->seq = 1;

    if( stream )
    {
        if( TRP_V1_Write1ByteValue( stream, TRP_V1_OP_LOCAL_DOMAIN ) &&
            DOFObjectID_Marshal( DOFObjectIDDomain_GetObjectID( ldData->localDomain ), stream ) &&
            DOFObjectID_Marshal( DOFObjectIDDomain_GetObjectID( ldData->remDomain ), stream ))
        {
            OALStream_AdjustReadOffset( ldData->rd, -OALStream_GetReadOffset( ldData->rd ));
            if( OALStream_Write(stream, OALStream_GetDataSize(ldData->rd), OALStream_GetReadPtr(ldData->rd)) )
            {
                TRP_V1_HandleHash( data, OALStream_GetReadPtr(ldData->rd), hashLength );
                if( DOFObjectID_Marshal( DOFObjectIDAuthentication_GetObjectID( ldData->remNodeID ), stream ) &&
                    OALPacket_AppendStream(data.packet, stream) )
                    return;
            }
        }
        OALStream_Release( stream, TRUE );
    }
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_SendRequestRemoteDomain( TRP_V1_SendData_t data )
{
    TRP_V1_DomainData_t* rdData = data.opData->RequestData.domainData;
    uint32 streamLen = 1 +
                       DOFObjectID_GetBytesSize( DOFObjectIDDomain_GetObjectID( rdData->remDomain ) ) +
                       DOFObjectID_GetBytesSize( DOFObjectIDDomain_GetObjectID( rdData->localDomain ) );
    OALStream stream = OALStream_CreateRW(streamLen);
    if( stream )
    {
        data.opData->I_len = DOFKeyRequest_MarshallByteLength(rdData->initRemoteKeyRequest, TRUE); 
        rdData->rd = OALStream_CreateRW(data.opData->I_len + 64);

        data.trp->blackboardData->seq = 1;
        if( rdData->rd )
        {
            data.opData->I = (uint8*)OALStream_GetReadPtr(rdData->rd);
            if( TRP_V1_Write1ByteValue( stream, TRP_V1_OP_REMOTE_DOMAIN ) &&
                DOFObjectID_Marshal( DOFObjectIDDomain_GetObjectID( rdData->remDomain ), stream ) &&
                DOFObjectID_Marshal( DOFObjectIDDomain_GetObjectID( rdData->localDomain ), stream ) &&
                DOFKeyRequest_Marshal( rdData->rd, rdData->initRemoteKeyRequest, TRUE ) )
            {
                if( OALPacket_AppendStream(data.packet, stream) )
                {
                    if( OALPacket_AppendStream(data.packet, rdData->rd) )
                    {   
                        /* Retained because we keep a copy in TRP op data and the reference we had was given to OALPacket_AppendStream */
                        OALStream_Retain(rdData->rd); 
                        return;
                    }
                    /* Set to NULL so we don't free bellow because OALPacket_AppendStream succeded and took ownership */
                    stream = NULL; 
                }
            }
            OALStream_Release( rdData->rd, TRUE );
            rdData->rd = NULL;
        }
        OALStream_Release( stream, TRUE );
    }
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_SendRequestValidateCreds( TRP_V1_SendData_t data )
{
    TRP_V1_ValidateCredsData_t* vcData = data.opData->RequestData.validateData;
    uint32 streamLen = 1+
                       DOFObjectID_GetBytesSize( DOFObjectIDDomain_GetObjectID( data.opData->secDomain ) ) +
                       DOFCredentials_MarshalSize( vcData->credentials, TRUE )+
                       40;
    OALStream stream = OALStream_CreateRW( streamLen );

    data.trp->blackboardData->seq = 1;

    if( stream )
    {
        if( TRP_V1_Write1ByteValue( stream, TRP_V1_OP_VALIDATE_CREDS ) &&
            DOFObjectID_Marshal( DOFObjectIDDomain_GetObjectID( data.opData->secDomain ), stream ) &&
            DOFCredentials_Internal_Marshal( vcData->credentials, stream, TRUE ) &&
            OALStream_Write( stream, 40, vcData->validA ) &&
            OALPacket_AppendStream(data.packet, stream))
        {
            return;
        }
        OALStream_Release(stream, FALSE);
    }
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}

void TRP_V1_SendInvalidOpCode( TRP_V1_SendData_t data )
{
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_WARNING, "TRP V1: Received invalid opcode");
    data.opData->trp_status = DOFERROR_REQUEST_FAILED;
}
