/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/cfg.h>
#include <dof/permissionset.h>
#include <dof/credentials.h>
#include <dof/securitymode.h>
#include <internal/bridgeinternal.h>
#include <internal/epp_negative_one.h>
#include <internal/objectinternal.h>
#include <internal/credentialsinternal.h>
#include <internal/permissionsetinternal.h>
#include <internal/resultinternal.h>
#include <internal/domaininternal.h>
#include <internal/router.h>
#include <internal/srp.h>
#include <internal/error.h>
#include <internal/trp_v1.h>
#include "trp_v1_utils.h"
#include "trp_v1_receive_util.h"
#include "trp_v1_sm.h"
#include "trp_v1_operation.h"
#include <internal/dofinternal.h>

static OALTimer TRPRetryTimer = NULL;

boolean TRP_V1_InitRetryTimer( void )
{
    if ((TRPRetryTimer = OALTimer_Create(DOF_GetDof()->threadPool, FALSE)) == NULL)
        return FALSE;
    return TRUE;
}

void TRP_V1_DestroyRetryTimer( void )
{
    if (TRPRetryTimer)
    {
        OALTimer_Destroy(TRPRetryTimer);
        TRPRetryTimer = NULL;
    }
}

static void TRP_V1_TimeOut( OALObserver self, OALObservable observable, void* ctx )
{
    /* Handle Time Out */
    DOFOperation op = (DOFOperation)ctx;
    if( op )
    {
        TRP_V1_Data data = (TRP_V1_Data)op->context;
        if(!op->flags.bits.completed && op->operation_type != DOFOPERATIONTYPE_TRP_V1_PASS )
        {
            OperationContext_t node;

            OALEnqueueItem_Init((OALEnqueueItem) &node, TRP_V1_SM_Receive);
            node.op = op;
            node.domain = NULL;
            node.doFree = FALSE;
            op->flags.bits.completed = 1;
            data->trp_status = DOFERROR_TIMEOUT;
            TRP_V1_SM_Receive( (OALThreadPoolItem) &node, FALSE );
        }
        else
        {
            op->flags.bits.completed = 1;
            DOFOperation_Internal_Destroy(op);
        }
    }
}

static void session_context_destroy(void* context)
{
    TRP_V1_Data data = (TRP_V1_Data)context;
    OALStream_Release( data->RequestData.sessionData->I, TRUE );
    OALStream_Release( data->RequestData.sessionData->R, TRUE );
    if( data->A )
        OALStream_Release( data->A, TRUE );
    DOFObjectIDDomain_Destroy( data->secDomain );
    DOFSecurityScope_Destroy( data->RequestData.sessionData->connSecurityScope );
    OALSessionInitiatorBlock_Release( data->RequestData.sessionData->initBlock );
    OALKeyGrant_Release( data->RequestData.sessionData->initKeyGrant );
    DOFKeyRequest_Destroy( data->RequestData.sessionData->respBlock );
    OALKeyGrant_Release( data->RequestData.sessionData->respKeyGrant );
    if(data->flags.bits.retainedStream)
        OALStream_Release(data->stream, TRUE);
    DOFOperation_Destroy(data->parent);
    DOFAuthenticator_Destroy( data->domain );
    DOFObjectIDDomain_Destroy(data->fromDomain);
    DOFMem_Free(data->errorData);
    DOFMem_Free(data);
}

static void domain_context_destroy(void* context)
{
    TRP_V1_Data data = (TRP_V1_Data)context;
    if( data->RequestData.domainData )
    {
        if( data->RequestData.domainData->rd )
            OALStream_Release(data->RequestData.domainData->rd, TRUE);
        DOFObjectIDAuthentication_Destroy( data->RequestData.domainData->remNodeID );
        DOFObjectIDDomain_Destroy( data->RequestData.domainData->remDomain );
        DOFObjectIDDomain_Destroy( data->RequestData.domainData->localDomain );
        DOFKeyRequest_Destroy( data->RequestData.domainData->initRemoteKeyRequest );
        OALKeyGrant_Release( data->RequestData.domainData->initRemoteKeyGrant );
        OALKeyGrant_Release( data->RequestData.domainData->initLocalKeyGrant );
        DOFMem_Free(data->RequestData.domainData);
    }
    if(data->flags.bits.retainedStream)
        OALStream_Release(data->stream, TRUE);
    DOFObjectIDDomain_Destroy( data->secDomain );
    DOFOperation_Destroy(data->parent);
    DOFAuthenticator_Destroy( data->domain );
    OALStream_Release(data->A, TRUE);
    DOFObjectIDDomain_Destroy(data->fromDomain);
    DOFMem_Free(data->errorData);
    DOFMem_Free(data);
}

static void as_domain_context_destroy(void* context)
{
    TRP_V1_Data data = (TRP_V1_Data)context;
    if( data->RequestData.domainData->rd )
        OALStream_Release(data->RequestData.domainData->rd, TRUE);
    DOFObjectIDDomain_Destroy( data->RequestData.domainData->remDomain );
    DOFObjectIDDomain_Destroy( data->RequestData.domainData->localDomain );
    DOFKeyRequest_Destroy( data->RequestData.domainData->initRemoteKeyRequest );
    OALKeyGrant_Release( data->RequestData.domainData->initRemoteKeyGrant );
    OALKeyGrant_Release( data->RequestData.domainData->initLocalKeyGrant );
    DOFObjectIDAuthentication_Destroy( data->RequestData.domainData->remNodeID );
    if(data->flags.bits.retainedStream)
        OALStream_Release(data->stream, TRUE);
    DOFObjectIDDomain_Destroy( data->secDomain );
    DOFOperation_Destroy(data->parent);
    DOFAuthenticator_Destroy( data->domain );
    DOFObjectIDDomain_Destroy(data->fromDomain);
    DOFMem_Free(data->errorData);
    OALStream_Release(data->A, TRUE);
    DOFMem_Free(data);
}

static void securityScope_context_destroy(void* context)
{
    TRP_V1_Data data = (TRP_V1_Data)context;
    DOFObjectIDDomain_Destroy( data->secDomain );
    DOFObjectIDAuthentication_Destroy( data->RequestData.securityScopeData->secNodeID );
    OALKeyGrant_Release( data->RequestData.securityScopeData->initKeyGrant );
    DOFKeyRequest_Destroy( data->RequestData.securityScopeData->initKeyRequest );
    DOFSecurityScope_Destroy( data->RequestData.securityScopeData->securityScope );
    DOFMem_Free(data->RequestData.securityScopeData->mode);
    if(data->flags.bits.retainedStream)
        OALStream_Release(data->stream, TRUE);
    DOFOperation_Destroy(data->parent);
    DOFAuthenticator_Destroy( data->domain );
    DOFMem_Free(data->errorData);
    DOFObjectIDDomain_Destroy(data->fromDomain);
    OALStream_Release(data->A, TRUE);
    DOFMem_Free(data);
}

static void kek_context_destroy(void* context)
{
    TRP_V1_Data data = (TRP_V1_Data)context;
    DOFObjectIDDomain_Destroy( data->secDomain );
    DOFPermissionSet_Destroy( data->RequestData.kekData->groupsPermissions );
    DOFKeyRequest_Destroy( data->RequestData.kekData->initKeyRequest );
    DOFSecurityMode_Destroy( data->RequestData.kekData->securityMode );
    DOFSecurityScope_Destroy( data->RequestData.kekData->secScope );
    DOFObjectIDSource_Destroy( data->RequestData.kekData->sid );
    OALKeyGrant_Release( data->RequestData.kekData->initKeyGrant );
    DOFObjectIDAuthentication_Destroy( data->RequestData.kekData->groupID );
    if(data->flags.bits.retainedStream)
        OALStream_Release(data->stream, TRUE);
    DOFOperation_Destroy(data->parent);
    DOFAuthenticator_Destroy( data->domain );
    DOFMem_Free(data->errorData);
    DOFObjectIDDomain_Destroy(data->fromDomain);
    OALStream_Release(data->A, TRUE);
    DOFMem_Free(data);
}

static void random_context_destroy(void* context)
{
    TRP_V1_Data data = (TRP_V1_Data)context;
    DOFObjectIDDomain_Destroy( data->secDomain );
    OALKeyGrant_Release( data->RequestData.randomData->initKeyGrant );
    DOFKeyRequest_Destroy( data->RequestData.randomData->initKeyRequest );
    DOFOperation_Destroy(data->parent);
    DOFAuthenticator_Destroy( data->domain );
    if(data->flags.bits.retainedStream)
        OALStream_Release(data->stream, TRUE);
    DOFMem_Free(data->errorData);
    DOFObjectIDDomain_Destroy(data->fromDomain);
    OALStream_Release(data->A, TRUE);
    DOFMem_Free(data);
}

static void resolve_context_destroy(void* context)
{
    TRP_V1_Data data = (TRP_V1_Data)context;
    DOFObjectIDDomain_Destroy( data->secDomain );
    DOFCredentials_Destroy( data->RequestData.credentialData->credentials );
    DOFOperation_Destroy(data->parent);
    DOFAuthenticator_Destroy( data->domain );
    DOFMem_Free(data->errorData);
    if(data->flags.bits.retainedStream)
        OALStream_Release(data->stream, TRUE);
    OALStream_Release(data->A, TRUE);
    DOFObjectIDDomain_Destroy(data->fromDomain);
    DOFMem_Free(data);
}

static void validate_context_destroy(void* context)
{
    TRP_V1_Data data = (TRP_V1_Data)context;
    PCRMem_Free(data->RequestData.validateData->validB);
    DOFObjectIDDomain_Destroy( data->secDomain );
    DOFCredentials_Destroy( data->RequestData.validateData->credentials);
    DOFOperation_Destroy(data->parent);
    DOFAuthenticator_Destroy( data->domain );
    if(data->flags.bits.retainedStream)
        OALStream_Release(data->stream, TRUE);
    DOFMem_Free(data->errorData);
    DOFObjectIDDomain_Destroy(data->fromDomain);
    OALStream_Release(data->A, TRUE);
    DOFMem_Free(data);
}

static void pass_relay_destroy(void* context)
{
    TRP_V1_Data data = (TRP_V1_Data)context;
    OALChannel_Release(data->RequestData.passData->passOnChannel);
    DOFAddress_Destroy(data->RequestData.passData->address);
    DOFMem_Free(data->RequestData.passData->streamData);
    DOFObjectIDDomain_Destroy( data->secDomain );
    DOFOperation_Destroy(data->parent);
    DOFAuthenticator_Destroy( data->domain );
    if(data->flags.bits.retainedStream)
        OALStream_Release(data->stream, TRUE);
    DOFMem_Free(data->errorData);
    DOFObjectIDDomain_Destroy(data->fromDomain);
    OALStream_Release(data->A, TRUE);
    DOFMem_Free(data);
}

void TRP_V1_FreeLocalASRequestDestroy( TRP_V1_Data data )
{
    if( data )
    {
        switch( data->opCode )
        {
        case TRP_V1_OP_KEK:
            kek_context_destroy(data);
            break;
        case TRP_V1_OP_RANDOM:
            random_context_destroy(data);
            break;
        case TRP_V1_OP_SESSION:
            session_context_destroy(data);
            break;
        case TRP_V1_OP_SECURITY_SCOPE:
        case TRP_V1_OP_SECURITY_SCOPE_EXT:
            securityScope_context_destroy(data);
            break;
        case TRP_V1_OP_CREDENTIAL:
            resolve_context_destroy(data);
            break;
        case TRP_V1_OP_LOCAL_DOMAIN:
        case TRP_V1_OP_REMOTE_DOMAIN:
            domain_context_destroy(data);
            break;
        case TRP_V1_OP_VALIDATE_CREDS:
            validate_context_destroy(data);
            break;
        default:
            break;
        }
    }
}

static void TRP_ParentOpClosing(OALObserver self, OALObservable observable, void* ctx)
{
    DOFOperation op = (DOFOperation)ctx;
    if( op )
    {
        DOFOperation temp;
        DOFOperation_Lock(op);
        temp = ((TRP_V1_Data_t*)op->context)->parent;
        op->flags.bits.completed = 1;
        DOFOperation_Unlock(op);
        DOFOperation_Internal_Destroy(temp);
        DOFOperation_Internal_Destroy(op);
    }
}

static DOFOperation create_trp_operation( DOFOperationType type, TRP_V1_Data myData, OALChannel channel, DOFAddress address, uint32 timeout, DOFOperation_DestroyContext dest_fn, boolean force )
{
    /* create operation */
    DOFOperation op = DOFOperation_TRP_V1_New( type, myData, timeout );
    if( op )
    {
        /* add the observer for timeout*/
        OALObserver temp_obsrv = NULL;
        op->operation_will_close = OALObservable_Create();
        if( op->operation_will_close )
        {
            OALObserver obs = NULL;
            if( OALOperationManager_AddOperation( op, op->opID, TRUE) == DOFERROR_OK )
            {
                DOFOperation_Lock(op);
                /* The op could be cancelled by the add operation. There is no need to route if op was cancelled. */
                if( !op->flags.bits.cancelled ) 
                {
                    if( TRPRouter_RouteToAS(op, channel, address, force) )
                    {
                        op->contextDestroyFunction = dest_fn;
                        temp_obsrv = DOFOperation_AddOpWillCloseObserver(op, TRP_V1_TimeOut);
                        if (temp_obsrv)
                        {
                            if( myData->parent )
                            {
                                DOFOperation_Retain(op);
                                obs = DOFOperation_AddOpWillCloseObserverOtherArg(myData->parent, TRP_ParentOpClosing, op);
                                if( obs )
                                {
                                    DOFOperation_Unlock(op);
                                    return op;
                                }
                                DOFOperation_Internal_Destroy(op);
                                /* Fall through to error handling */
                            }
                            else
                            {
                                DOFOperation_Unlock(op);
                                return op;
                            }

                            /* The op is already locked */
                            op->flags.bits.completed = 1;
                            op->contextDestroyFunction = NULL;
                            DOFOperation_RemoveOpWillCloseObserver( op, temp_obsrv );
                        }
                    }
                    op->flags.bits.completed = 1;
                    op->contextDestroyFunction = NULL;
                }
                OALOperationManager_RemoveOperationAndUnlock(op); 
            }
        }
        DOFOperation_Internal_Destroy(op);
    }
    return NULL;
}

static boolean fill_in_data_std(TRP_V1_Data data, trp_callback_fn trp_fn, void* context, boolean isCommand, DOFObjectIDDomain secDomain, uint8 trp_op)
{
    data->RequestData.kekData = (TRP_V1_KEKData_t*)(((uint8*)data) + sizeof(TRP_V1_Data_t));
    data->secDomain = DOFObjectIDDomain_Copy( secDomain );
    if( data->secDomain )
    {
        data->callback = trp_fn;
        data->trp_status = DOFERROR_OK;
        data->userContext = context;
        data->opCode = trp_op;
        data->flags.bits.isCommand = isCommand;
        return TRUE;
    }
    return FALSE;
}

static void clear_data_std( TRP_V1_Data data )
{
    DOFObjectIDDomain_Destroy( data->secDomain );
}

void TRP_V1_StartHash( TRP_V1_Data data, DOFCredentials creds )
{
    DOFObjectID domain = DOFObjectIDDomain_GetObjectID( DOFCredentials_GetDomainID( creds ) );
    PCR_Assert( domain );
    OALSecureHMAC_Start( &data->ctx, (uint8*)DOFCredentials_GetKey( creds ) );
    OALSecureHMAC_Digest( &data->ctx, DOFObjectID_GetBytesSize( domain ), DOFObjectID_GetBytes( domain ));
}

void TRP_V1_RequestKEK( DOFKeyRequest keyReq, DOFObjectIDAuthentication groupID, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout)
{
    TRP_V1_Data data;
    DOFOperation op;

    if( !DOFCredentials_IsResolved( keyReq->credentials ) )
        goto FAIL;

    data = DOFMem_Alloc(sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_KEKData_t));

    if( !data )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "TRP V1: Failed to create request KEK.");
        goto FAIL;
    }

    PCRMem_Set(data,0,sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_KEKData_t));

    if( fill_in_data_std( data, trp_fn, context, TRUE, DOFCredentials_GetDomainID( keyReq->credentials ), TRP_V1_OP_KEK ) )
    {
        data->RequestData.kekData->initKeyRequest = DOFKeyRequest_Retain( keyReq );
        if( data->RequestData.kekData->initKeyRequest )
        {
            data->RequestData.kekData->groupID = DOFObjectIDAuthentication_Copy( groupID );
            if( data->RequestData.kekData->groupID )
            {
                TRP_V1_StartHash( data, keyReq->credentials );

                if((op = create_trp_operation( DOFOPERATIONTYPE_TRP_V1_KEK, data, channel, address, timeout, kek_context_destroy, FALSE)) != NULL )
                {
                    DOFOperation_Internal_Destroy(op);
                    return; /* success */
                }
                DOFObjectIDAuthentication_Destroy( groupID );
            }
            DOFKeyRequest_Destroy( data->RequestData.kekData->initKeyRequest );
        }
        clear_data_std( data );
    }
    DOFMem_Free( data );
FAIL:
    trp_fn(NULL, context);
}

void TRP_V1_RequestRandom( DOFKeyRequest keyReq, trp_callback_fn trp_fn, void* context, uint32 timeout )
{
    TRP_V1_Data data;
    DOFOperation op;

    if( !DOFCredentials_IsResolved( keyReq->credentials ) )
        goto FAIL;

    data = DOFMem_Alloc(sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_RandomData_t));

    if( !data )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "TRP V1: Failed to create request Random.");
        goto FAIL;
    }

    PCRMem_Set(data,0,sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_RandomData_t));

    if( fill_in_data_std( data, trp_fn,context, TRUE, DOFCredentials_GetDomainID( keyReq->credentials ), TRP_V1_OP_RANDOM ) )
    {
        data->RequestData.randomData->initKeyRequest = DOFKeyRequest_Retain( keyReq );
        if( data->RequestData.randomData->initKeyRequest )
        {
            TRP_V1_StartHash( data, keyReq->credentials );
            if((op = create_trp_operation(DOFOPERATIONTYPE_TRP_V1_RANDOM, data, NULL, NULL, timeout, random_context_destroy, FALSE)) != NULL)
            {
                DOFOperation_Internal_Destroy(op);
                return; /* success */
            }
            DOFKeyRequest_Destroy( data->RequestData.randomData->initKeyRequest );
        }
        clear_data_std( data );
    }
    DOFMem_Free( data );
FAIL:
    trp_fn(NULL, context);
}

void TRP_V1_RequestSession( OALStream rBlock, DOFKeyRequest respKeyReq, OALStream iBlock, OALSessionInitiatorBlock sessInitBlock, OALChannel channel, DOFAddress address, trp_callback_fn trp_fn, void* context, uint32 timeout )
{
    TRP_V1_Data data;
    DOFOperation op;

    PCR_Assert((rBlock != NULL) && (iBlock  != NULL));

    if( !DOFCredentials_IsResolved( respKeyReq->credentials ) )
        goto FAIL;

    data = DOFMem_Alloc(sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_SessionData_t));

    if( !data )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "TVP V1: Failed to create request Session.");
        goto FAIL;
    }

    PCRMem_Set(data,0,sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_SessionData_t));
    if( fill_in_data_std( data, trp_fn, context, TRUE, DOFCredentials_GetDomainID( respKeyReq->credentials ), TRP_V1_OP_SESSION ) )
    {
        data->RequestData.sessionData->respBlock = DOFKeyRequest_Retain(respKeyReq);
        if( data->RequestData.sessionData->respBlock )
        {
            data->RequestData.sessionData->initBlock = OALSessionInitiatorBlock_Retain( sessInitBlock );
            if( data->RequestData.sessionData->initBlock )
            {
                OALStream_Retain(iBlock);
                data->RequestData.sessionData->I = iBlock;
                OALStream_Retain(rBlock);
                data->RequestData.sessionData->R = rBlock;

                TRP_V1_StartHash( data, respKeyReq->credentials );

                if((op = create_trp_operation(DOFOPERATIONTYPE_TRP_V1_SESSION, data, channel, address, timeout, session_context_destroy, FALSE)) != NULL )
                {
                    DOFOperation_Internal_Destroy(op);
                    return; /* success */
                }

                OALStream_Release( iBlock, TRUE );
                OALStream_Release( rBlock, TRUE );
                OALSessionInitiatorBlock_Release( data->RequestData.sessionData->initBlock );
            }
            DOFKeyRequest_Destroy( data->RequestData.sessionData->respBlock );
        }
        clear_data_std( data );
    }
    DOFMem_Free( data );
FAIL:
    trp_fn(NULL, context);
}

static void TRP_V1_Internal_RequestSecurityScope( uint8 opCode, uint8 duration, DOFKeyRequest initKeyReq, DOFObjectIDAuthentication secNodeID, DOFAuthenticatorSecurityType secType, uint32 modeLen, uint8* mode, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout)
{
    TRP_V1_Data data;
    DOFOperation op;

    if( !DOFCredentials_IsResolved( initKeyReq->credentials ) )
        goto FAIL;

    data = DOFMem_Alloc(sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_SecurityScopeData_t));

    if( !data )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "TRP V1: Failed to create request Security Scope.");
        goto FAIL;
    }

    PCRMem_Set(data,0,sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_SecurityScopeData_t));

    if( fill_in_data_std( data, trp_fn, context, TRUE, DOFCredentials_GetDomainID( initKeyReq->credentials ), opCode ) )
    {
        if( opCode == TRP_V1_OP_SECURITY_SCOPE_EXT )
        {
            data->RequestData.securityScopeData->secType = secType;
            data->RequestData.securityScopeData->modeLen = modeLen;
            data->RequestData.securityScopeData->mode = DOFMem_Alloc(modeLen);
            if( !data->RequestData.securityScopeData->mode )
            {
                DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "TRP V1: Failed to create request Security Scope.");
                goto FAIL_DESTROY;
            }
            PCRMem_Copy(data->RequestData.securityScopeData->mode, mode, modeLen);
        }

        data->RequestData.securityScopeData->initKeyRequest = DOFKeyRequest_Retain( initKeyReq );
        if( data->RequestData.securityScopeData->initKeyRequest )
        {
            data->RequestData.securityScopeData->secNodeID = DOFObjectIDAuthentication_Copy( secNodeID );
            if( data->RequestData.securityScopeData->secNodeID )
            {
                data->RequestData.securityScopeData->grantedDuration = duration;
                TRP_V1_StartHash( data, initKeyReq->credentials );

                if((op = create_trp_operation(DOFOPERATIONTYPE_TRP_V1_SECURITY_SCOPE, data, channel, address, timeout, securityScope_context_destroy, FALSE)) != NULL )
                {
                    DOFOperation_Internal_Destroy(op);
                    return; /* success */
                }
                DOFObjectIDAuthentication_Destroy( data->RequestData.securityScopeData->secNodeID );
            }
            DOFKeyRequest_Destroy( data->RequestData.securityScopeData->initKeyRequest );
        }
FAIL_DESTROY:
        clear_data_std( data );
    }
    DOFMem_Free( data );
FAIL:
    trp_fn(NULL, context);
}

void TRP_V1_RequestSecurityScope( uint8 duration, DOFKeyRequest initKeyReq, DOFObjectIDAuthentication secNodeID, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout)
{
    TRP_V1_Internal_RequestSecurityScope(TRP_V1_OP_SECURITY_SCOPE, duration, initKeyReq, secNodeID, DOFAUTHENTICATORSECURITYTYPE_UNDEFINED, 0, NULL, trp_fn, context, channel, address, timeout);
}

void TRP_V1_RequestSecurityScopeAndMode(uint8 duration, DOFKeyRequest initKeyReq, DOFObjectIDAuthentication secNodeID, DOFAuthenticatorSecurityType secType, uint32 modeLen, uint8* mode, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout)
{
    TRP_V1_Internal_RequestSecurityScope(TRP_V1_OP_SECURITY_SCOPE_EXT, duration, initKeyReq, secNodeID, secType, modeLen, mode, trp_fn, context, channel, address, timeout);
}

void TRP_V1_Internal_RequestResolveCredentials( DOFObjectIDDomain domain, DOFOperation parent, DOFCredentials creds, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout, boolean force )
{
    TRP_V1_Data data = DOFMem_Alloc(sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_CredentialData_t));
    DOFOperation op;

    if( !data )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "TRP V1: Failed to create request Resolve Credentials.");
        trp_fn(NULL, context);
        return;
    }

    PCRMem_Set(data,0,sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_CredentialData_t));

    if( fill_in_data_std( data, trp_fn, context, TRUE, domain, TRP_V1_OP_CREDENTIAL ) )
    {
        data->RequestData.credentialData->credentials = DOFCredentials_Copy( creds );
        if( data->RequestData.credentialData->credentials )
        {
            if( parent )
            {
                DOFOperation_Retain( parent );
                data->parent = parent;
            }

            if( (op = create_trp_operation( DOFOPERATIONTYPE_TRP_V1_CREDENTIAL, data, channel, address, timeout, resolve_context_destroy, force)) != NULL)
            {
                DOFOperation_Internal_Destroy(op);
                return;
            }

            if( parent )
            {
                data->parent = NULL;
                DOFOperation_Internal_Destroy( parent );
            }

            DOFCredentials_Destroy( data->RequestData.credentialData->credentials );
        }
        clear_data_std( data );
    }
    DOFMem_Free( data );

    trp_fn(NULL, context);
}

void TRP_V1_RequestResolveCredentials( DOFCredentials creds, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout )
{
    if( !DOFCredentials_IsResolved(creds) )
    {
        TRP_V1_Internal_RequestResolveCredentials( DOFCredentials_GetDomainID( creds ), NULL, creds, trp_fn, context, channel, address, timeout, FALSE );
    }
    else
    {
        trp_v1_context trp_contex;
        Resolve_return_t ret;

        ret.errorCode = 0;
        ret.creds = creds;
        ret.status = DOFERROR_OK;

        trp_contex.ResolveReturn = &ret;

        trp_fn(&trp_contex,context);
    }
}

void TRP_V1_Internal_RequestValidateCredentials(DOFCredentials creds, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout, boolean force)
{
    TRP_V1_Data data;
    DOFOperation op;
    OALSecureHMACContext hmac;

    if( !DOFCredentials_IsResolved( creds ) )
        goto FAIL;

    data = DOFMem_Alloc(sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_ValidateCredsData_t));

    if( !data )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "TRP V1: Failed to create request Random");
        goto FAIL;
    }

    PCRMem_Set(data,0,sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_ValidateCredsData_t));

    if( fill_in_data_std( data, trp_fn,context, TRUE, DOFCredentials_GetDomainID( creds ), TRP_V1_OP_VALIDATE_CREDS ) )
    {
        data->RequestData.validateData->credentials = DOFCredentials_Copy(creds);
        if( data->RequestData.validateData->credentials )
        {
            PCRSecureRandom_Generate(8, data->RequestData.validateData->validA);

            OALSecureHMAC_Start(&hmac, (uint8*)DOFCredentials_GetKey(data->RequestData.validateData->credentials));
            OALSecureHMAC_Digest(&hmac, 8, data->RequestData.validateData->validA);
            OALSecureHMAC_Finish(&hmac, data->RequestData.validateData->validA+8);

            if((op = create_trp_operation(DOFOPERATIONTYPE_TRP_V1_VALIDATE_CREDENTIALS, data, channel, address, timeout, validate_context_destroy, force)) != NULL)
            {
                DOFOperation_Internal_Destroy(op);
                return; /* success */
            }

            DOFCredentials_Destroy( data->RequestData.validateData->credentials );
        }
        clear_data_std( data );
    }
    DOFMem_Free( data );
FAIL:
    trp_fn(NULL, context);
}

void TRP_V1_RequestValidateCredentials(DOFCredentials creds, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout)
{
    TRP_V1_Internal_RequestValidateCredentials(creds, trp_fn, context, channel, address, timeout, FALSE);
}

static void TRP_V1_Internal_RequestRemoteDomain( DOFObjectIDDomain domain, DOFOperation parent, DOFKeyRequest keyReq, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout, boolean force)
{
    TRP_V1_Data data;
    DOFOperation op;

    if( !DOFCredentials_IsResolved( keyReq->credentials ) )
        goto FAIL;

    data = DOFMem_Alloc(sizeof(TRP_V1_Data_t));

    if( !data )
    {
        goto FAIL;
    }

    PCRMem_Set(data,0,sizeof(TRP_V1_Data_t));

    if( fill_in_data_std( data, trp_fn, context, TRUE, domain, TRP_V1_OP_REMOTE_DOMAIN ) )
    {
        data->RequestData.domainData = (TRP_V1_DomainData_t*)DOFMem_Alloc( sizeof(TRP_V1_DomainData_t) );
        if( data->RequestData.domainData )
        {
            PCRMem_Set(data->RequestData.domainData,0,sizeof(TRP_V1_DomainData_t));
            data->RequestData.domainData->remDomain = DOFObjectIDDomain_Copy( domain );
            if( data->RequestData.domainData->remDomain )
            {
                data->RequestData.domainData->localDomain =  DOFObjectIDDomain_Copy(DOFCredentials_GetDomainID( keyReq->credentials ));
                if( data->RequestData.domainData->localDomain )
                {
                    data->RequestData.domainData->initRemoteKeyRequest = DOFKeyRequest_Retain( keyReq );
                    if( data->RequestData.domainData->initRemoteKeyRequest )
                    {
                        if( parent )
                        {
                            DOFOperation_Retain(parent);
                            data->parent = parent;
                        }

                        if((op = create_trp_operation(DOFOPERATIONTYPE_TRP_V1_REMOTE_DOMAIN, data, channel, address, timeout, domain_context_destroy, force)) != NULL)
                        {
                            DOFOperation_Internal_Destroy(op);
                            return; /* success */
                        }

                        if( parent )
                        {
                            data->parent = NULL;
                            DOFOperation_Internal_Destroy( parent );
                        }
                        DOFKeyRequest_Destroy( data->RequestData.domainData->initRemoteKeyRequest );
                    }
                    DOFObjectIDDomain_Destroy( data->RequestData.domainData->localDomain );
                }
                DOFObjectIDDomain_Destroy( data->RequestData.domainData->remDomain );
            }
            DOFMem_Free( data->RequestData.domainData );
        }
        clear_data_std( data );
    }
    DOFMem_Free( data );
FAIL:
    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "TRP V1: Failed to create request Remote Domain");
    trp_fn(NULL, context);
}

static void remoteDomainDiscoveryRet(TRP_V1_Context trp_context, DOFOperation oldOp)
{
    if( trp_context )
    {
        if( trp_context->RemoteDomainReturn->status == DOFERROR_OK && trp_context->RemoteDomainReturn->errorCode == 0 && DOFCredentials_IsResolved( trp_context->RemoteDomainReturn->creds ) && !oldOp->flags.bits.completed )
        {
            TRP_V1_Data data = (TRP_V1_Data)oldOp->context;
            oldOp->flags.bits.completed = 1;
            data->callback(trp_context, data->userContext);
        }
    }
}

void TRP_V1_RequestRemoteDomainDiscovery( DOFObjectIDDomain domain, DOFOperation parent, DOFKeyRequest keyReq, OALChannel channel, DOFAddress address, uint32 timeout)
{
    TRP_V1_Internal_RequestRemoteDomain( domain, parent, keyReq, (trp_callback_fn)remoteDomainDiscoveryRet, parent, channel, address, timeout, FALSE );
}

void TRP_V1_RequestRemoteDomain( DOFObjectIDDomain domain, DOFKeyRequest keyReq, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout)
{
    TRP_V1_Internal_RequestRemoteDomain( domain, NULL, keyReq, trp_fn, context, channel, address, timeout, FALSE );
}

void TRP_V1_RequestRemoteDomainForcable( DOFObjectIDDomain domain, DOFKeyRequest keyReq, trp_callback_fn trp_fn, void* context, OALChannel channel, DOFAddress address, uint32 timeout, boolean force )
{
    TRP_V1_Internal_RequestRemoteDomain( domain, NULL, keyReq, trp_fn, context, channel, address, timeout, force );
}

static void domainDiscoverRet( TRP_V1_Context trp_context, DOFOperation oldOp )
{
    if( trp_context )
    {
        if( trp_context->ResolveReturn->status == DOFERROR_OK && trp_context->ResolveReturn->errorCode == 0 && DOFCredentials_IsResolved( trp_context->ResolveReturn->creds ) && !oldOp->flags.bits.completed )
        {

            TRP_V1_Data data = (TRP_V1_Data)oldOp->context;
            oldOp->flags.bits.completed = 1;
            data->callback(trp_context, data->userContext);
            return;
        }
    }
}

void TRP_V1_ResolveDomainCredentials( DOFOperation oldOp, DOFObjectIDDomain domain, DOFCredentials creds, OALChannel channel, DOFAddress address )
{
    DOFCredentials temp = DOFCredentials_CreateCopyWithDomain(creds,domain);
    /* resolve new credentials */
    TRP_V1_Internal_RequestResolveCredentials( domain, oldOp, temp, (trp_callback_fn)domainDiscoverRet, (void*)oldOp, channel, address, DOFOperation_GetTimeRemaining(oldOp), FALSE);
    DOFCredentials_Destroy(temp);
}

void TRP_V1_RequestLocalDomain( DOFOperation op )
{
    DOFOperation nop;
    TRP_V1_Data data = DOFMem_Alloc(sizeof(TRP_V1_Data_t));
    TRP_V1_Data oldOpData = (TRP_V1_Data)op->context;

    if( !data )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "TRP V1: Failed to create request Local Domain");
        goto FAIL;
    }

    PCRMem_Set(data,0,sizeof(TRP_V1_Data_t));

    if( fill_in_data_std( data, oldOpData->callback, oldOpData->userContext, TRUE, oldOpData->RequestData.domainData->localDomain, TRP_V1_OP_LOCAL_DOMAIN ) )
    {
        data->RequestData.domainData = oldOpData->RequestData.domainData;
        oldOpData->RequestData.domainData = NULL;

        if( oldOpData->I )
        {
            data->I = oldOpData->I;
            data->I_len = oldOpData->I_len;
            oldOpData->I = NULL;
            oldOpData->I_len = 0;

            data->flags.flags = oldOpData->flags.flags;
            data->flags.bits.isMarshalled = FALSE;
            data->flags.bits.isCommand = TRUE;
        }

        DOFOperation_Retain(op);
        data->parent = op;

        TRP_V1_StartHash( data, data->RequestData.domainData->initRemoteKeyRequest->credentials );
        if((nop = create_trp_operation(DOFOPERATIONTYPE_TRP_V1_LOCAL_DOMAIN, data, op->sources->channel, (op->sources->channel)?op->sources->channel->remoteChannelData->address:NULL, DOFOperation_GetTimeRemaining(op), domain_context_destroy, FALSE)) != NULL)
        {
            DOFOperation_Internal_Destroy(nop);
            return; /* success */
        }

        DOFOperation_Internal_Destroy( data->parent );
        clear_data_std( data );
    }
    DOFMem_Free( data );
FAIL:
    oldOpData->callback(NULL, oldOpData->userContext);
}

DOFOperation TRP_V1_CreateASOperation( DOFAuthenticator domain, uint8 opCode, OALChannel channel, DOFAddress addr, DOFOperationID opID )
{
    DOFOperation op;
    TRP_V1_Data_t* data = NULL;
    DOFOperation_DestroyContext dtor;

    switch( opCode )
    {
    case TRP_V1_OP_KEK:
        dtor = kek_context_destroy;
        data = (TRP_V1_Data_t*)DOFMem_Alloc(sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_KEKData_t));
        if( data )PCRMem_Set(data,0,sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_KEKData_t));
        break;
    case TRP_V1_OP_RANDOM:
        dtor = random_context_destroy;
        data = (TRP_V1_Data_t*)DOFMem_Alloc(sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_RandomData_t));
        if( data )PCRMem_Set(data,0,sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_RandomData_t));
        break;
    case TRP_V1_OP_SESSION:
        dtor = session_context_destroy;
        data = (TRP_V1_Data_t*)DOFMem_Alloc(sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_SessionData_t));
        if( data )PCRMem_Set(data,0,sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_SessionData_t));
        break;
    case TRP_V1_OP_SECURITY_SCOPE:
    case TRP_V1_OP_SECURITY_SCOPE_EXT:
        dtor = securityScope_context_destroy;
        data = (TRP_V1_Data_t*)DOFMem_Alloc(sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_SecurityScopeData_t));
        if( data )PCRMem_Set(data,0,sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_SecurityScopeData_t));
        break;
    case TRP_V1_OP_CREDENTIAL:
        dtor = resolve_context_destroy;
        data = (TRP_V1_Data_t*)DOFMem_Alloc(sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_CredentialData_t));
        if( data )PCRMem_Set(data,0,sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_CredentialData_t));
        break;
    case TRP_V1_OP_LOCAL_DOMAIN:
    case TRP_V1_OP_REMOTE_DOMAIN:
        dtor= as_domain_context_destroy;
        data = (TRP_V1_Data_t*)DOFMem_Alloc(sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_DomainData_t));
        if( data )PCRMem_Set(data,0,sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_DomainData_t));
        break;
    case TRP_V1_OP_VALIDATE_CREDS:
        dtor = validate_context_destroy;
        data = (TRP_V1_Data_t*)DOFMem_Alloc(sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_ValidateCredsData_t));
        if( data )PCRMem_Set(data,0,sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_ValidateCredsData_t));
        break;
    default:
        goto FAIL_0;
        break;
    }

    if( !data )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "TRP V1: Failed to create TRP AS operation.");
        goto FAIL_0;
    }

    /* this will work for all types */
    data->RequestData.kekData = (TRP_V1_KEKData_t*)(((uint8*)data) + sizeof(TRP_V1_Data_t));

    data->secure = channel->remoteChannelData->flags.sessionIsSecure;

    data->trp_status = DOFERROR_OK;
    data->domain = DOFAuthenticator_Copy(domain);
    data->opCode = opCode;
    DOFOperationID_Copy(opID);

    op = TRP_V1_CreateOperation(DOFOPERATIONTYPE_TRP_V1_AS, data, opID, 15000 );

    if( !op )
    {
        DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "TRP V1: Failed to create TRP AS operation.");
        goto FAIL_1;
    }

    op->sources->channel = OALChannel_Retain( channel );
    op->sources->address = DOFAddress_Copy( addr );
    if(op->sources->channel && op->sources->address)
    {
        op->operation_will_close = OALObservable_Create();
        if( op->operation_will_close )
        {
            if(DOFOperation_AddOpWillCloseObserver(op, TRP_V1_TimeOut))
            {
                op->contextDestroyFunction = dtor;
                if(OALOperationManager_AddOperation( op, op->opID, TRUE)== DOFERROR_OK)
                {
                    return op;
                }
            }
        }
    }
    else
    {
        goto FAIL_1;
    }

    OALObservable_Destroy(op->operation_will_close);
    op->operation_will_close = NULL;
    DOFOperation_Internal_Destroy(op);
    return NULL;
FAIL_1:
    DOFAddress_Destroy(addr);
    OALChannel_Release(channel);
    DOFOperationID_Destroy(opID);
    DOFMem_Free( data );
FAIL_0:
    return NULL;
}

DOFOperation TRP_V1_RelayCreate( DOFObjectIDDomain domain, OALChannel channel, DOFAddress addr, DOFOperationID opID, uint8 opCode, uint32 timeout)
{
    DOFOperation op;
    TRP_V1_Data_t* data = (TRP_V1_Data_t*)DOFMem_Alloc(sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_PassData_t));
    DOFOperationID opid = NULL;

    if( data )
    {
        PCRMem_Set(data,0,sizeof(TRP_V1_Data_t) + sizeof(TRP_V1_PassData_t));
        data->RequestData.passData = (TRP_V1_PassData_t*)(((uint8*)data) + sizeof(TRP_V1_Data_t));

        data->secDomain = DOFObjectIDDomain_Copy(domain);
        if( data->secDomain )
        {
            data->RequestData.passData->address =  DOFAddress_Copy(addr);
            if( data->RequestData.passData->address )
            {
                data->fromDomain = DOFObjectIDDomain_Copy( DOFSecurityMembrane_GetDomainID( channel->securityMembrane ) );
                if( data->fromDomain )
                {
                    data->RequestData.passData->passOnChannel = OALChannel_Retain(channel);
                    if( data->RequestData.passData->passOnChannel )
                    {
                        data->trp_status = DOFERROR_OK;
                        data->flags.bits.isCommand = TRUE;
                        data->RequestData.passData->length = 0;
                        data->RequestData.passData->streamData = NULL;
                        data->opCode = opCode;

                        data->secure = channel->remoteChannelData->flags.sessionIsSecure;

                        opid = DOFOperationID_Copy(opID);
                        if( opid )
                        {
                            op = TRP_V1_CreateOperation(DOFOPERATIONTYPE_TRP_V1_PASS, data, opID, timeout );
                            if( op )
                            {
                                op->operation_will_close = OALObservable_Create();
                                if( op->operation_will_close )
                                {
                                    if(DOFOperation_AddOpWillCloseObserver(op, TRP_V1_TimeOut))
                                    {
                                        op->contextDestroyFunction = pass_relay_destroy;
                                        OALOperationManager_AddOperation( op, op->opID, TRUE);
                                        return op;
                                    }
                                    OALObservable_Destroy(op->operation_will_close);
                                    op->operation_will_close = NULL;
                                }
                                DOFOperation_Internal_Destroy(op);
                            }
                            DOFOperationID_Destroy(opID);
                        }
                        OALChannel_Release( data->RequestData.passData->passOnChannel );
                    }
                    DOFObjectIDDomain_Destroy( data->fromDomain );
                }
                DOFAddress_Destroy( data->RequestData.passData->address );
                OALChannel_Release( data->RequestData.passData->passOnChannel );
            }
            DOFObjectIDDomain_Destroy( data->secDomain );
        }
        DOFMem_Free( data );
    }

    DOF_LogMessage( DOF_GetDof(), DOFLOGLEVEL_ERROR, "TRP V1: Failed to create TRP forwarded op.");

    return NULL;
}

void TRP_V1_SM_Receive( OALThreadPoolItem dnode, boolean shutdown )
{
    /* should be operation */
    OperationContext_t * node = (OperationContext_t*) dnode;

    if(!shutdown)
    {
        /* process */
        switch( node->op->operation_type )
        {
        case DOFOPERATIONTYPE_TRP_V1_PASS:
            TRP_V1_SM_PassThrough(node->op);
            break;
        case DOFOPERATIONTYPE_TRP_V1_AS:
            TRP_V1_SM_AS(node->op);
            break;
        case DOFOPERATIONTYPE_TRP_V1_KEK:
            TRP_V1_SM_KEK(node->op);
            break;
        case DOFOPERATIONTYPE_TRP_V1_RANDOM:
            TRP_V1_SM_Random(node->op);
            break;
        case DOFOPERATIONTYPE_TRP_V1_SESSION:
            TRP_V1_SM_Session(node->op);
            break;
        case DOFOPERATIONTYPE_TRP_V1_SECURITY_SCOPE:
            TRP_V1_SM_SecurityScope(node->op);
            break;
        case DOFOPERATIONTYPE_TRP_V1_CREDENTIAL:
            TRP_V1_SM_Credential(node->op);
            break;
        case DOFOPERATIONTYPE_TRP_V1_LOCAL_DOMAIN:
            TRP_V1_SM_LocalDomain(node->op);
            break;
        case DOFOPERATIONTYPE_TRP_V1_REMOTE_DOMAIN:
            TRP_V1_SM_RemoteDomain(node->op);
            break;
        case DOFOPERATIONTYPE_TRP_V1_VALIDATE_CREDENTIALS:
            TRP_V1_SM_ValidateCredentials(node->op);
            break;
        default:
            break;
        }
    }

    PCR_Assert(dnode != NULL);

    DOFOperation_Internal_Destroy(node->op);
    if(node->doFree)
    {
        DOFMem_Free(node);
    }
}

void TRP_V1_SM_PassThrough(DOFOperation myOperation)
{
    TRP_V1_Data data = (TRP_V1_Data)myOperation->context;

    if( data->trp_status == DOFERROR_OK && data->flags.bits.isCommand)
    {
        TRPRouter_RouteToAS(myOperation,NULL,NULL,FALSE);
    }
}

/* Handles Local traffic and forwarded traffic to local AS */
void TRP_V1_SM_LocalASReceive( OALThreadPoolItem dnode, boolean shutdown )
{
    OperationContext_t * node = (OperationContext_t*) dnode;
    DOFOperation op = NULL;
    TRP_V1_Data data = NULL;
    TRP_V1_Data localData = NULL;
    DOFAuthenticator domain = NULL;

    PCR_Assert( dnode != NULL );
    op = node->op;
    domain = node->domain;

    if( shutdown )
    {
        DOFOperation_Internal_Destroy(op);
        if(node->doFree)
        {
            DOFMem_Free(node);
        }
        return;
    }

    DOFOperation_Lock(op);

    data = (TRP_V1_Data)op->context;

    if( op->operation_type != DOFOPERATIONTYPE_TRP_V1_PASS )
    {
        /* Local AS request */
        /* if broadcast copy opdata */
        if( data->flags.bits.isDomainDiscovery )
        {
            if( !(data->opCode == TRP_V1_OP_CREDENTIAL || data->opCode == TRP_V1_OP_REMOTE_DOMAIN) || ((localData = TRP_V1_LocalASDiscoveryRequest(data, domain)) == NULL))
            {
                goto FAIL;
            }
        }
        else
        {
            if( !data->flags.bits.isMarshalled )
            {
                /* Marshal it */
                if( !TRP_V1_MarshalLocalASRequest(data) )
                {
                    goto FAIL;
                }
                if( data->opCode != TRP_V1_OP_CREDENTIAL )
                {
                    data->flags.bits.isMarshalled = 1;
                }
            }
            localData = data;
        }
    }
    else
    {
        /* Pass Through Local AS Request */
        /* Unmarshal it */
        if( (localData = TRP_V1_UnmarshalForwardedLocalASRequest(data, domain)) == NULL )
        {
            goto FAIL;
        }
    }

    /* call as */
    switch( localData->opCode )
    {
    case TRP_V1_OP_KEK:
        TRP_V1_AS_RequestKEK( localData );
        break;
    case TRP_V1_OP_RANDOM:
        TRP_V1_AS_RequestRandom( localData );
        break;
    case TRP_V1_OP_SESSION:
        TRP_V1_AS_RequestSession( localData );
        break;
    case TRP_V1_OP_SECURITY_SCOPE:
        TRP_V1_AS_RequestSecurityScope( localData );
        break;
    case TRP_V1_OP_CREDENTIAL:
        TRP_V1_AS_RequestCredential( localData );
        break;
    case TRP_V1_OP_LOCAL_DOMAIN:
        TRP_V1_AS_RequestLocalDomain( localData );
        break;
    case TRP_V1_OP_REMOTE_DOMAIN:
        TRP_V1_AS_RequestRemoteDomain( localData );
        break;
    case TRP_V1_OP_VALIDATE_CREDS:
        TRP_V1_AS_RequestValidateCredential( localData );
        break;
    default:
        goto FAIL;
        break;
    }

    /* errors get dropped for domain discovery requests and forwarding requests */
    if( localData->errorCode != 0 && (data->flags.bits.isDomainDiscovery || op->operation_type == DOFOPERATIONTYPE_TRP_V1_PASS) )
    {
        goto FAIL;
    }

    /* Send Responses */
    if(  op->operation_type != DOFOPERATIONTYPE_TRP_V1_PASS )
    {
        if( data->flags.bits.isDomainDiscovery )
        {
            /* local domain discovery request */
            if( op->operation_type == DOFOPERATIONTYPE_TRP_V1_CREDENTIAL )
            {
                TRP_V1_ResolveDomainCredentials(op, DOFAuthenticator_GetDomainID(domain), data->RequestData.credentialData->credentials, NULL, NULL);
            }
            else if( op->operation_type == DOFOPERATIONTYPE_TRP_V1_REMOTE_DOMAIN )
            {
                DOFKeyRequest kr = NULL;
                DOFCredentials temp = DOFCredentials_CreateCopy(data->RequestData.domainData->initRemoteKeyRequest->credentials);
                if( temp )
                {
                    kr = DOFKeyRequest_Create(temp, DEFAULT_NONCE_SIZE, data->RequestData.domainData->initRemoteKeyRequest->newLogin, data->RequestData.domainData->initRemoteKeyRequest->fixedPermissions, NULL, data->RequestData.domainData->initRemoteKeyRequest->permission_set);
                    if( kr )
                    {
                        TRP_V1_RequestRemoteDomainDiscovery(DOFAuthenticator_GetDomainID(domain), op, kr, NULL, NULL, DOFOperation_GetTimeRemaining(op));
                        DOFKeyRequest_Destroy(kr);
                    }
                    DOFCredentials_Destroy(temp);
                }
            }
            else
            {
                PCR_Assert(FALSE); /* shouldn't hit this, if we do there is a bug */
            }
        }
        else
        {
            if((!TRP_V1_UnmarshalLocalASRequest(localData) || !TRP_V1_ValidateLocalASRequest(localData,DOFAuthenticator_GetDomainID(domain))))
            {
                localData = NULL;
                goto FAIL;
            }

            /* directed local as request */
            if(!op->flags.bits.completed)
            {
                OperationContext_t node;

                OALEnqueueItem_Init((OALEnqueueItem) &node, TRP_V1_SM_Receive);
                node.op = op;
                node.domain = NULL;
                node.doFree = FALSE;
                DOFOperation_Retain(op); /* bumps ref count because we will send it back throught the state machine */
                if(data->opCode != TRP_V1_OP_CREDENTIAL && data->opCode != TRP_V1_OP_REMOTE_DOMAIN )
                    op->flags.bits.completed = 1;
                data->trp_status = DOFERROR_OK;
                DOFOperation_Unlock( op );
                TRP_V1_SM_Receive( (OALThreadPoolItem) &node, FALSE );
                DOFOperation_Lock( op );
            }
            localData = NULL;
        }
    }
    else
    {
        /* route response to sender */
        TRP_V1_PassData_t* passData = data->RequestData.passData;

        if(!op->flags.bits.completed)
        {
            DOFResult res = DOFResult_Stream_Create(passData->passOnChannel, passData->address, DOFERROR_OK,DOFObjectIDDomain_GetObjectID(localData->secDomain), localData->A, DOFRESULTTYPE_VALUELIST_STREAM);
            OALStream_Release(data->A, TRUE);
            if( res )
            {
                OALChannel channel_tmp = OALChannel_Retain(passData->passOnChannel);
                DOFAddress address_tmp = DOFAddress_Copy(passData->address);

                if( channel_tmp && (!passData->address || address_tmp ))
                {
                    data->flags.bits.isCommand = 0;
                    data->A = localData->A;
                    localData->A = NULL;

                    DOFAddress_Destroy( op->sources->address );
                    OALChannel_Release( op->sources->channel );

                    op->sources->address = address_tmp;
                    op->sources->channel = channel_tmp;
                    DOFOperation_Unlock( op );
                    op->sources->channel->routeToRequestor( op->sources->channel, op, res);
                    DOFOperation_Lock( op );
                }
                else
                {
                    DOFAddress_Destroy( address_tmp );
                    OALChannel_Release( channel_tmp );
                }
                DOFResult_Destroy(res);
            }
        }
    }

FAIL:
    TRP_V1_FreeLocalASRequestDestroy(localData);
    DOFOperation_Unlock(op);
    DOFAuthenticator_Destroy(domain);
    DOFOperation_Internal_Destroy(op);
    if(node->doFree)
    {
        DOFMem_Free(node);
    }
}

void TRP_V1_SM_AS(DOFOperation myOperation)
{
    TRP_V1_Data data = (TRP_V1_Data)myOperation->context;
    boolean ret = FALSE;

    DOFOperation_Lock(myOperation);
    if( data->trp_status == DOFERROR_OK )
    {
        DOFOperation_IncrementRetry(myOperation);
        switch( data->opCode )
        {
        case TRP_V1_OP_REJECT:
            break;
        case TRP_V1_OP_KEK:
            ret = TRP_V1_AS_RequestKEK( data );
            break;
        case TRP_V1_OP_RANDOM:
            ret = TRP_V1_AS_RequestRandom( data );
            break;
        case TRP_V1_OP_SESSION:
            ret = TRP_V1_AS_RequestSession( data );
            break;
        case TRP_V1_OP_SECURITY_SCOPE:
            ret = TRP_V1_AS_RequestSecurityScope( data );
            break;
        case TRP_V1_OP_DOMAINCREDENTIAL:
            /* should not be received */
            break;
        case TRP_V1_OP_CREDENTIAL:
            ret = TRP_V1_AS_RequestCredential( data );
            break;
        case TRP_V1_OP_LOCAL_DOMAIN:
            ret = TRP_V1_AS_RequestLocalDomain( data );
            break;
        case TRP_V1_OP_REMOTE_DOMAIN:
            ret = TRP_V1_AS_RequestRemoteDomain( data );
            break;
        case TRP_V1_OP_VALIDATE_CREDS:
            ret = TRP_V1_AS_RequestValidateCredential( data );
            break;
        case TRP_V1_OP_DISCOVER_REMOTE:
            /* should not be received */
            break;
        default:
            break;
        }

        if( ret )
        {
            data->flags.bits.isCommand = FALSE;
            myOperation->sources->channel->routeToRequestor( myOperation->sources->channel, myOperation, NULL );
        }
    }
    DOFOperation_Unlock(myOperation);
}

void TRP_V1_SM_KEK(DOFOperation myOperation)
{
    TRP_V1_Data data = (TRP_V1_Data)myOperation->context;
    trp_v1_context trc;
    KEK_return_t kekRet;
    trc.KEKReturn = &kekRet;

    DOFOperation_Lock(myOperation);
    PCRMem_Set( &kekRet, 0, sizeof(KEK_return_t));
    kekRet.status = data->trp_status;

    if( kekRet.status == DOFERROR_OK )
    {
        if( data->opCode == TRP_V1_OP_REJECT )
        {
            kekRet.errorCode = data->errorCode;
            kekRet.errorData = data->errorData;
        }
        else
        {
            if( data->flags.bits.passUsed )
                TRPRouter_AddDynamicRoute(data->RequestData.kekData->initKeyRequest->credentials, myOperation->sources->channel, myOperation->sources->address, TRUE);
            kekRet.KEK = (uint8*)data->k;
            kekRet.epoch = data->RequestData.kekData->epoch;
            kekRet.sid = data->RequestData.kekData->sid;
            kekRet.groupsPermissions = data->RequestData.kekData->groupsPermissions;
            kekRet.secScope = data->RequestData.kekData->secScope;
            kekRet.thb = data->RequestData.kekData->thb;
            kekRet.tmax = data->RequestData.kekData->tmax;
            kekRet.tmin = data->RequestData.kekData->tmin;
            kekRet.securityMode = data->RequestData.kekData->securityMode;
            kekRet.stateID = data->RequestData.kekData->stateID;
            kekRet.initPermissions = data->RequestData.kekData->initKeyRequest->permission_set;
        }
    }

    DOFOperation_Unlock(myOperation);
    data->callback((TRP_V1_Context)&trc, data->userContext);
}

void TRP_V1_SM_Random(DOFOperation myOperation)
{
    TRP_V1_Data data = (TRP_V1_Data)myOperation->context;
    trp_v1_context trc;
    Random_return_t ranRet;
    trc.RandomReturn = &ranRet;

    DOFOperation_Lock(myOperation);
    PCRMem_Set( &ranRet, 0, sizeof(Random_return_t));
    ranRet.status = data->trp_status;

    if( data->trp_status == DOFERROR_OK )
    {
        if( data->opCode == TRP_V1_OP_REJECT )
        {
            ranRet.errorCode = data->errorCode;
            ranRet.errorData = data->errorData;
        }
        else
        {
            if( data->flags.bits.passUsed )
                TRPRouter_AddDynamicRoute(data->RequestData.randomData->initKeyRequest->credentials, myOperation->sources->channel, myOperation->sources->address, TRUE);
            ranRet.Random = (uint8*)data->k;
        }
    }
    DOFOperation_Unlock(myOperation);
    data->callback((TRP_V1_Context)&trc, data->userContext);
}

void TRP_V1_SM_Session(DOFOperation myOperation)
{
    TRP_V1_Data data = (TRP_V1_Data)myOperation->context;
    trp_v1_context trc;
    Session_return_t sesRet;
    trc.SessionReturn = &sesRet;

    DOFOperation_Lock(myOperation);
    PCRMem_Set( &sesRet, 0, sizeof(Session_return_t));
    sesRet.status = data->trp_status;

    if( data->trp_status == DOFERROR_OK )
    {
        if( data->opCode == TRP_V1_OP_REJECT )
        {
            sesRet.errorCode = data->errorCode;
            sesRet.errorData = data->errorData;
        }
        else
        {
            if( data->flags.bits.passUsed )
                TRPRouter_AddDynamicRoute(data->RequestData.sessionData->respBlock->credentials, myOperation->sources->channel, myOperation->sources->address, TRUE);
            sesRet.grantedDuration = data->RequestData.sessionData->grantedDuration;
            sesRet.connSecurityScope = data->RequestData.sessionData->connSecurityScope;
            sesRet.initPermValidation = data->RequestData.sessionData->initBlock->keyRequest->permission_set;
            sesRet.respPermValidation = data->RequestData.sessionData->respBlock->permission_set;
            sesRet.initGrant = data->RequestData.sessionData->initKeyGrant;
            sesRet.sessionKey = (uint8*)data->k;
            sesRet.A = data->A;
        }
    }
    DOFOperation_Unlock(myOperation);
    data->callback((TRP_V1_Context)&trc, data->userContext);
}

void TRP_V1_SM_SecurityScope(DOFOperation myOperation)
{
    TRP_V1_Data data = (TRP_V1_Data)myOperation->context;
    trp_v1_context trc;
    Scope_return_t ssRet;
    trc.ScopeReturn = &ssRet;

    DOFOperation_Lock(myOperation);
    PCRMem_Set( &ssRet, 0, sizeof(Scope_return_t));
    ssRet.status = data->trp_status;

    if( data->trp_status == DOFERROR_OK )
    {
        if( data->opCode == TRP_V1_OP_REJECT )
        {
            ssRet.errorCode = data->errorCode;
            ssRet.errorData = data->errorData;
        }
        else
        {
            if( data->flags.bits.passUsed )
                TRPRouter_AddDynamicRoute(data->RequestData.securityScopeData->initKeyRequest->credentials, myOperation->sources->channel, myOperation->sources->address, TRUE);
            ssRet.grantedDuration = data->RequestData.securityScopeData->grantedDuration;
            ssRet.securityScope = data->RequestData.securityScopeData->securityScope;
            ssRet.permValidation = data->RequestData.securityScopeData->initKeyRequest->permission_set;
        }
    }
    DOFOperation_Unlock(myOperation);
    data->callback((TRP_V1_Context)&trc, data->userContext);
}

void TRP_V1_SM_Credential(DOFOperation myOperation)
{
    TRP_V1_Data data = (TRP_V1_Data)myOperation->context;

    trp_v1_context trc;
    Resolve_return_t resRet;
    trc.ResolveReturn = &resRet;

    DOFOperation_Lock(myOperation);
    PCRMem_Set( &resRet, 0, sizeof(Resolve_return_t));
    resRet.status = data->trp_status;
    resRet.creds = data->RequestData.credentialData->credentials;

    if( data->trp_status == DOFERROR_OK )
    {
        if( data->opCode == TRP_V1_OP_REJECT )
        {
            resRet.errorCode = data->errorCode;
            resRet.errorData = data->errorData;
        }
        else
        {
            if( !DOFCredentials_IsResolved( data->RequestData.credentialData->credentials ) )
            {
                data->flags.bits.isCommand = TRUE;
                DOFOperation_IncrementRetry(myOperation);
                TRPRouter_RouteToAS( myOperation, myOperation->sources->channel, myOperation->sources->address, FALSE);
                DOFOperation_Unlock(myOperation);
                return;
            }

            if( data->flags.bits.passUsed )
                TRPRouter_AddDynamicRoute(data->RequestData.credentialData->credentials, myOperation->sources->channel, myOperation->sources->address, TRUE);
        }
    }
    else if( data->trp_status == DOFERROR_DUPLICATE )
    {
        data->trp_status = DOFERROR_OK;
        return;
    }

    if( data->callback )
    {
        DOFOperation_Unlock(myOperation);
        data->callback((TRP_V1_Context)&trc, data->userContext);
    }

    myOperation->flags.bits.completed = 1;
    DOFOperation_Internal_Cancel(myOperation);
}

void TRP_V1_SM_LocalDomain(DOFOperation myOperation)
{
    TRP_V1_Data data = (TRP_V1_Data)myOperation->context;
    DOFOperation op = data->parent;
    TRP_V1_Data oldData = (TRP_V1_Data)op->context;
    trp_v1_context trc;
    RemoteDomain_return_t rdRet;
    trc.RemoteDomainReturn = &rdRet;

    DOFOperation_Lock(op);
    if( !op->flags.bits.timedOut && !op->flags.bits.cancelled && !op->flags.bits.completed )
    {
        DOFOperation_Lock(myOperation);
        PCRMem_Set( &rdRet, 0, sizeof(RemoteDomain_return_t));
        rdRet.status = data->trp_status;

        if( data->trp_status == DOFERROR_OK )
        {
            if( data->opCode == TRP_V1_OP_REJECT )
            {
                if( !oldData->flags.bits.isDomainDiscovery )
                {
                    rdRet.errorCode = data->errorCode;
                    rdRet.errorData = data->errorData;
                }
                else
                {
                    myOperation->flags.bits.completed = 1;
                    DOFOperation_Unlock(myOperation);
                    DOFOperation_Unlock(op);
                    return;
                }
            }
            else
            {
                if( data->flags.bits.passUsed )
                    TRPRouter_AddDynamicRoute(data->RequestData.domainData->initRemoteKeyRequest->credentials, myOperation->sources->channel, myOperation->sources->address, TRUE);
                rdRet.remoteDomainIdentifier = data->RequestData.domainData->localDomainRespBlock;
                rdRet.creds = DOFCredentialsKey_Create( data->RequestData.domainData->remDomain, data->RequestData.domainData->remNodeID, data->k );
            }


            op->flags.bits.completed = 1;
            myOperation->flags.bits.completed = 1;
            DOFOperation_Unlock(myOperation);
            DOFOperation_Unlock(op);
            data->callback((TRP_V1_Context)&trc, data->userContext);
            DOFCredentials_Destroy( rdRet.creds );
            DOFOperation_Internal_Cancel(op);
        }
    }
    else
    {
        DOFOperation_Lock(myOperation);
        myOperation->flags.bits.completed = 1;
        DOFOperation_Unlock(myOperation);
        DOFOperation_Unlock(op);
    }
}

void TRP_V1_SM_RemoteDomain(DOFOperation myOperation)
{
    TRP_V1_Data data = (TRP_V1_Data)myOperation->context;
    trp_v1_context trc;
    RemoteDomain_return_t rdRet;

    DOFOperation_Lock(myOperation);
    trc.RemoteDomainReturn = &rdRet;
    rdRet.errorCode = data->errorCode;
    rdRet.status = data->trp_status;

    if( !myOperation->flags.bits.completed )
    {
        if( data->trp_status == DOFERROR_OK && data->opCode != TRP_V1_OP_REJECT)
        {
            TRP_V1_RequestLocalDomain( myOperation );
        }
        else
        {
            myOperation->flags.bits.completed = 1;

            DOFOperation_Unlock(myOperation);
            data->callback((TRP_V1_Context)&trc, data->userContext);
            return;
        }
    }
    DOFOperation_Unlock(myOperation);
}

void TRP_V1_SM_ValidateCredentials(DOFOperation myOperation)
{
    TRP_V1_Data data = (TRP_V1_Data)myOperation->context;

    trp_v1_context trc;
    Validate_return_t resRet;
    trc.ValidateReturn = &resRet;

    DOFOperation_Lock(myOperation);
    PCRMem_Set( &resRet, 0, sizeof(Validate_return_t));
    resRet.status = data->trp_status;
    resRet.creds = data->RequestData.validateData->credentials;

    if( data->trp_status == DOFERROR_OK )
    {
        if( data->opCode == TRP_V1_OP_REJECT )
        {
            resRet.errorCode = data->errorCode;
            resRet.errorData = data->errorData;
        }
        else
        {
            OALSecureHMACContext hmac;
            uint32 len = data->RequestData.validateData->bLen;
            uint8* validA = data->RequestData.validateData->validA;
            uint8* validB = data->RequestData.validateData->validB;
            uint8* validateB = NULL;
            uint32 sidLen = len - 40;
            DOFObjectID sid = DOFObjectID_Create_Unmarshal(&sidLen, validB+8 );
            resRet.sid = DOFObjectIDSource_Create(sid);
            DOFObjectID_Destroy(sid);

            if( resRet.sid )
            {
                len = 40 + sidLen;
                validateB = PCRMem_Alloc(len);
                if( validateB )
                {
                    PCRMem_Copy(validateB,validB,8);
                    PCRMem_Copy(validateB+8,DOFObjectID_GetBytes(DOFObjectIDSource_GetObjectID(resRet.sid)),sidLen);

                    OALSecureHMAC_Start(&hmac, (uint8*)DOFCredentials_GetKey(data->RequestData.validateData->credentials));
                    OALSecureHMAC_Digest(&hmac, 8, validA);
                    OALSecureHMAC_Digest(&hmac, 8, validB);
                    OALSecureHMAC_Digest(&hmac, sidLen, DOFObjectID_GetBytes(DOFObjectIDSource_GetObjectID(resRet.sid)));
                    OALSecureHMAC_Digest(&hmac, DOFObjectID_GetBytesSize( DOFObjectIDDomain_GetObjectID( data->secDomain )),DOFObjectID_GetBytes(DOFObjectIDDomain_GetObjectID( data->secDomain )));
                    OALSecureHMAC_Finish(&hmac, validateB+8+sidLen);

                    if( PCRMem_Compare( validB, validateB, len ) != 0 )
                    {
                        resRet.errorCode = DOFERROR_REJECTED;
                        DOFObjectIDSource_Destroy(resRet.sid);
                        resRet.sid = NULL;
                    }
                    else
                    {
                        if( data->flags.bits.passUsed )
                            TRPRouter_AddDynamicRoute(data->RequestData.validateData->credentials, myOperation->sources->channel, myOperation->sources->address, TRUE);
                        DOFCredentials_UpdateSID( data->RequestData.validateData->credentials, resRet.sid );
                        DOFObjectIDSource_Destroy(resRet.sid);
                        resRet.sid = DOFCredentials_GetSID( data->RequestData.validateData->credentials ); /* Valid while credentials exist (owned by caller), so don't bother copying here. */
                    }
                    PCRMem_Free(validateB);
                }
                else
                {
                    DOFObjectIDSource_Destroy(resRet.sid);
                    resRet.sid = NULL;
                    resRet.status = DOFERROR_INTERNAL_ERROR;
                }
            }
            else
            {
                resRet.status = DOFERROR_INTERNAL_ERROR;
            }
        }
    }
    DOFOperation_Unlock(myOperation);
    data->callback((TRP_V1_Context)&trc, data->userContext);
}
