/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef TRP_V1_UTILS_H
#define TRP_V1_UTILS_H

#include <internal/trp_v1.h>
#include <internal/security_common.h>
#include <internal/stream.h>

#ifdef  __cplusplus
extern "C" {
#endif

typedef struct TRP_V1_Data_t* TRP_V1_Data;
typedef boolean (*TRP_V1_Helper)(TRP_V1_Data data);

typedef enum TRP_V1_Status
{
    TRP_V1_Status_OK = 0,
    TRP_V1_Status_ERROR,
    TRP_V1_Status_DONOTHING,
    TRP_V1_Status_SEND_DOWN
} TRP_V1_Status;

/* Used for all routes */
/* The uncommon case: LocalAS a pointer is wasted */
typedef struct TRPRouter_Route_t
{
    boolean                 isLocal;
    DOFAddress              address;
    union
    {
        OALChannel          channel;
        DOFAuthenticator    domain;
    }route;
}TRPRouter_Route_t;

typedef struct TRP_V1_KEKData_t
{
    DOFKeyRequest               initKeyRequest;
    DOFObjectIDAuthentication   groupID;

    OALKeyGrant                 initKeyGrant;
    uint8                       thb;
    uint8                       tmin;
    uint8                       tmax;
    uint16                      epoch;
    DOFSecurityScope            secScope;
    DOFSecurityMode             securityMode;
    uint32                      stateID;
    DOFObjectIDSource           sid;
    DOFPermissionSet            groupsPermissions;
}TRP_V1_KEKData_t;

typedef struct TRP_V1_RandomData_t
{
    DOFKeyRequest           initKeyRequest;

    OALKeyGrant             initKeyGrant;
}TRP_V1_RandomData_t;

typedef struct TRP_V1_SessionData_t
{
    OALStream                   I;
    OALStream                   R;

    DOFKeyRequest               respBlock;
    OALSessionInitiatorBlock    initBlock;

    OALKeyGrant                 respKeyGrant;
    OALKeyGrant                 initKeyGrant;
    uint8                       grantedDuration;
    DOFSecurityScope            connSecurityScope;
}TRP_V1_SessionData_t;

typedef struct TRP_V1_SecurityScopeData_t
{
    DOFKeyRequest               initKeyRequest;
    DOFObjectIDAuthentication   secNodeID;
    DOFAuthenticatorSecurityType       secType;

    uint8                       grantedDuration;
    OALKeyGrant                 initKeyGrant;
    DOFSecurityScope            securityScope;
    uint32                      modeLen;
    uint8*                      mode;
}TRP_V1_SecurityScopeData_t;

typedef struct TRP_V1_CredentialData_t
{
    DOFCredentials          credentials;
    DOFCredentials          localcredentials;
}TRP_V1_CredentialData_t;

typedef struct TRP_V1_ValidateCredsData_t
{
    DOFCredentials          credentials;
    uint8                   validA[40];
    uint8*                  validB;
    uint32                  bLen;
}TRP_V1_ValidateCredsData_t;

typedef struct TRP_V1_DomainData_t
{
    DOFObjectIDDomain           remDomain;
    DOFObjectIDDomain           localDomain;
    DOFKeyRequest               initRemoteKeyRequest;
    OALKeyGrant                 initRemoteKeyGrant;
    DOFObjectIDAuthentication   remNodeID;
    OALKeyGrant                 initLocalKeyGrant;
    uint32                      localDomainRespBlock;
    OALStream                   rd;
    uint32                      I_len;
    uint32                      responseA_len;
    uint8*                      responseA;
}TRP_V1_DomainData_t;

typedef struct TRP_V1_PassData_t
{
    OALChannel  passOnChannel;
    DOFAddress  address;
    uint32      length;
    uint8*      streamData;
}TRP_V1_PassData_t;

typedef struct TRP_V1_Data_t
{
    DOFError                    trp_status;

    uint8*                      errorData;

    trp_callback_fn             callback;
    void*                       userContext;

    DOFObjectIDDomain           secDomain;

    DOFAuthenticator            domain; 

    DOFAddress                  address;
    OALChannel                  channel;
    DOFOperation                parent;

    DOFObjectIDDomain           fromDomain; 

    union RequestData
    {
        TRP_V1_KEKData_t*             kekData;
        TRP_V1_RandomData_t*          randomData;
        TRP_V1_SessionData_t*         sessionData;
        TRP_V1_SecurityScopeData_t*   securityScopeData;
        TRP_V1_CredentialData_t*      credentialData;
        TRP_V1_DomainData_t*          domainData;
        TRP_V1_PassData_t*            passData;
        TRP_V1_ValidateCredsData_t*   validateData;
    }RequestData;                               /* data associated with operation not needed for Random || Reject */

    uint8                       k[32];

    OALSecureHMACContext        ctx;           /* hmac context fill during request */

    OALStream                   stream;        /* for storing Request Stream */
    uint8*                      I;             /* For AS */
    uint32                      I_len;         /* For AS */
    uint8*                      R;             /* For AS */
    uint32                      R_len;         /* For AS */
    OALStream                   A;             /* For AS */
    uint8*                      A_pointer;     /* For Local AS */
    uint32                      A_len;

    boolean                     secure;        /* it needs to go out the same way it came in secure/non-secure or be dropped */

    uint8                       errorCode;      /* If AS returned an error this will hold the status */

    uint8                       opCode;         /* opCode */

    union
    {
        uint8 flags;
        struct
        {
            unsigned int passUsed           : 1;
            unsigned int recReply           : 1;
            unsigned int isCommand          : 1;
            unsigned int isLocal            : 1;
            unsigned int isMarshalled       : 1;
            unsigned int isDomainDiscovery  : 1;
            unsigned int retainedStream     : 1;
            unsigned int ASDontCancel       : 1;
        } bits;
    } flags;

}TRP_V1_Data_t;

extern void TRP_V1_SendRejectOp(OALChannel channel, DOFAddress addr, DOFOperationID opID);
extern boolean TRP_V1_AS_RequestKEK( TRP_V1_Data data );
extern boolean TRP_V1_AS_RequestRandom( TRP_V1_Data data );
extern boolean TRP_V1_AS_RequestSession( TRP_V1_Data data );
extern boolean TRP_V1_AS_RequestSecurityScope( TRP_V1_Data data );
extern boolean TRP_V1_AS_RequestCredential( TRP_V1_Data data );
extern boolean TRP_V1_AS_RequestLocalDomain( TRP_V1_Data data );
extern boolean TRP_V1_AS_RequestRemoteDomain( TRP_V1_Data data );
extern boolean TRP_V1_AS_RequestValidateCredential( TRP_V1_Data data );

extern boolean TRPRouter_Init(void);
extern void TRPRouter_Shutdown(void);
extern boolean TRPRouter_GetDomainRoute( DOFObjectIDDomain domain, TRPRouter_Route_t* route );
extern TRPRouter_Route_t* TRPRouter_GetKnownRoutes( uint32* count ); /* all routes */
extern DOFAuthenticator* TRPRouter_GetDomains( uint32* count ); /* local ASs */
extern void TRPRouter_NotifyFailure(DOFObjectIDDomain domain, OALChannel channel, DOFAddress address);

extern void TRP_V1_CreateLocalASOperation( DOFOperation operation, DOFAuthenticator domain );

extern boolean TRP_V1_MarshalLocalASRequest( TRP_V1_Data data );
extern boolean TRP_V1_UnmarshalLocalASRequest( TRP_V1_Data data );
extern boolean TRP_V1_ValidateLocalASRequest( TRP_V1_Data data, DOFObjectIDDomain domain );
extern TRP_V1_Data TRP_V1_UnmarshalForwardedLocalASRequest( TRP_V1_Data data, DOFAuthenticator domain );
extern TRP_V1_Data TRP_V1_LocalASDiscoveryRequest( TRP_V1_Data data, DOFAuthenticator domain );
extern void TRP_V1_FreeLocalASRequestDestroy( TRP_V1_Data data );

#ifdef  __cplusplus
}
#endif

#endif

