/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/type.h>
#include <dof/value.h>
#include <internal/type.h>
#include <internal/marshal.h>
#include <internal/dofinternal.h>

#include <dof/interface.h>
#include <dof/interfaceid.h>
#include <internal/interfaceidinternal.h>
#include <dof/objectid.h>
#include <internal/objectidinternal.h>
#include <internal/iteminternal.h>
#include "../value/valueprivate.h"

/** @addtogroup DOFType
@{
**/


/*
**  "Private" Type and Method Declarations
*/

static DOFType DOFTypeContext_GetType( DOFTypeContext context, uint16 index, boolean must_define );
static uint16 DOFTypeContext_GetIndex( DOFTypeContext context, DOFType type );

/** @internal
    DOFType Structure.
**/
struct DOFType_t 
{
    DOFType  nextRef;           /**< Used for cleanup. */
    uint32   testRefCount;      /**< Used for cleanup. */
    uint32   externalRefCount;  /**< Reference Count. */
    uint32   internalRefCount;  /**< Reference Count. */
    uint32   minLength;         /**< Minimum Element Count (for Array types), minimum size (for BLOB type), or MIBenum for encoding (for a String type). **/
    uint32   maxLength;         /**< Maximum Element Count (for Array types), maximum size (for BLOB type), Maximum String Length (for String types), or Total Element Count (for Structure type). **/
    DOFTypeID type;             /**< DOF Type Specifier. **/
    uint16   typeSize;          /**< Size of the marshalled type, excluding sub-elements.  Type is predefined (not allocated) if typeSize == 0x8000. **/
    const struct DOFType_t *
             subtype[1];        /**< Subtype (for Array, Nullable, and Reference types), List of subtypes (for Structure type).  Space is allocated appropriately starting at this location for the data that needs to be contained.  **/
};

/*
 *  DOFType is declared as a pointer to a const structure (because some are).
 *  However, in some cases we know it is not const (e.g., it needs to be
 *  freed), and need a pointer to non-const data.  We use this union to bypass
 *  the compiler warning.
 */
union ConstCast
{
    DOFType            ctype;
    struct DOFType_t * type;
};

struct DOFTypeContextElement_t {
    DOFType type;
    uint32 offset;
    uint16 size;
    boolean processing;
};

struct DOFTypeContext_t {
    struct DOFTypeContextElement_t * elements;
    const uint8 * buffer;
    uint16 count;
    uint16 allocCount;
};

static const struct DOFType_t doftype_int8 =       { NULL, 0, 0, 0, 0,        0, DOFTYPE_INT8,         0x8000, { NULL } };
static const struct DOFType_t doftype_int16 =      { NULL, 0, 0, 0, 0,        0, DOFTYPE_INT16,        0x8000, { NULL } };
static const struct DOFType_t doftype_int32 =      { NULL, 0, 0, 0, 0,        0, DOFTYPE_INT32,        0x8000, { NULL } };
static const struct DOFType_t doftype_int64 =      { NULL, 0, 0, 0, 0,        0, DOFTYPE_INT64,        0x8000, { NULL } };
static const struct DOFType_t doftype_uint8 =      { NULL, 0, 0, 0, 0,        0, DOFTYPE_UINT8,        0x8000, { NULL } };
static const struct DOFType_t doftype_uint16 =     { NULL, 0, 0, 0, 0,        0, DOFTYPE_UINT16,       0x8000, { NULL } };
static const struct DOFType_t doftype_uint32 =     { NULL, 0, 0, 0, 0,        0, DOFTYPE_UINT32,       0x8000, { NULL } };
static const struct DOFType_t doftype_uint64 =     { NULL, 0, 0, 0, 0,        0, DOFTYPE_UINT64,       0x8000, { NULL } };
static const struct DOFType_t doftype_float32 =    { NULL, 0, 0, 0, 0,        0, DOFTYPE_FLOAT32,      0x8000, { NULL } };
static const struct DOFType_t doftype_float64 =    { NULL, 0, 0, 0, 0,        0, DOFTYPE_FLOAT64,      0x8000, { NULL } };
static const struct DOFType_t doftype_datetime =   { NULL, 0, 0, 0, 0,        0, DOFTYPE_DATETIME,     0x8000, { NULL } };
static const struct DOFType_t doftype_boolean =    { NULL, 0, 0, 0, 0,        0, DOFTYPE_BOOLEAN,      0x8000, { NULL } };
static const struct DOFType_t doftype_iid =        { NULL, 0, 0, 0, 0,        0, DOFTYPE_INTERFACEID,  0x8000, { NULL } };
static const struct DOFType_t doftype_oid =        { NULL, 0, 0, 0, 0,        0, DOFTYPE_OBJECTID,     0x8000, { NULL } };
static const struct DOFType_t doftype_guid =       { NULL, 0, 0, 0, 0,        0, DOFTYPE_GUID,         0x8000, { NULL } };
static const struct DOFType_t doftype_blob =       { NULL, 0, 0, 0, 0, 0x3FFFFF, DOFTYPE_BLOB,         0x8000, { NULL } };
static const struct DOFType_t doftype_void =       { NULL, 0, 0, 0, 0,        0, DOFTYPE_VOID,         0x8000, { NULL } };

DOFType DOFVALUEINT8_TYPE = &doftype_int8;
DOFType DOFVALUEINT16_TYPE = &doftype_int16;
DOFType DOFVALUEINT32_TYPE = &doftype_int32;
DOFType DOFVALUEINT64_TYPE = &doftype_int64;
DOFType DOFVALUEUINT8_TYPE = &doftype_uint8;
DOFType DOFVALUEUINT16_TYPE = &doftype_uint16;
DOFType DOFVALUEUINT32_TYPE = &doftype_uint32;
DOFType DOFVALUEUINT64_TYPE = &doftype_uint64;
DOFType DOFVALUEFLOAT32_TYPE = &doftype_float32;
DOFType DOFVALUEFLOAT64_TYPE = &doftype_float64;
DOFType DOFVALUEDATETIME_TYPE = &doftype_datetime;
DOFType DOFVALUEBOOLEAN_TYPE = &doftype_boolean;
DOFType DOFVALUEINTERFACEID_TYPE = &doftype_iid;
DOFType DOFVALUEOBJECTID_TYPE = &doftype_oid;
DOFType DOFVALUEGUID_TYPE = &doftype_guid;
DOFType DOFVALUEBLOB_TYPE = &doftype_blob;
DOFType DOFTYPEVOID_TYPE = &doftype_void;

static PCRMutex typeMutex = NULL;
static OALSListNode typeGCList = NULL;

static DOFType DOFType_Internal_Copy( DOFType self );
static void DOFType_Internal_Destroy( DOFType type );

#define DOFTYPE_TRY_NOT_CHECKED 0xFFFFFFFF

typedef struct DOFTypeGCNode_t {
    struct OALSListNode_t *next;
    DOFType                type;
} DOFTypeGCNode_t;

static void DOFType_Lock( void )
{
    if( typeMutex )
    {
        PCRMutex_Lock(typeMutex);
    }
}

static void DOFType_Unlock( void )
{
    if( typeMutex )
    {
        PCRMutex_Unlock(typeMutex);
    }
}

boolean DOFType_Init( void )
{
    typeMutex = PCRMutex_Create();
    if( typeMutex )
        return TRUE;
    return FALSE;
}

void DOFType_Shutdown( void )
{
    PCRMutex_Destroy( typeMutex );
    typeMutex = NULL;
}

static boolean DOFType_IsComposite( DOFType self )
{
    if( ( self->type == DOFTYPE_STRUCTURE ) || 
        ( self->type == DOFTYPE_ARRAY ) ||
        ( self->type == DOFTYPE_NULLABLE ) ||
        ( self->type == DOFTYPE_REFERENCE ) )
        return TRUE;
    return FALSE;
}

static void DOFType_ResetTryData( DOFType self )
{
    DOFType curr = self;
    while( curr != NULL )
    {
        DOFType next = curr->nextRef;
        union ConstCast temp;
        temp.ctype = curr;

        temp.type->testRefCount = DOFTYPE_TRY_NOT_CHECKED;
        temp.type->nextRef = NULL;
        curr = next;
    }
}

static boolean DOFType_TryTraverse( DOFType headx, DOFType currx, boolean noDec )
{
    boolean ret = TRUE;
    if( currx && currx->typeSize != 0x8000 )
    {
        if( currx->externalRefCount == 0) 
        {
            union ConstCast curr;
            curr.ctype = currx;

            /* check if we have visited before */
            if( currx->testRefCount == DOFTYPE_TRY_NOT_CHECKED )
            {
                union ConstCast headRef;
                headRef.ctype = headx;

                PCR_Assert( currx->testRefCount == DOFTYPE_TRY_NOT_CHECKED );

                curr.type->testRefCount = (noDec)?curr.type->internalRefCount : curr.type->internalRefCount-1;

                PCR_Assert( curr.type->testRefCount != DOFTYPE_TRY_NOT_CHECKED );
                PCR_Assert( curr.type->nextRef == NULL );

                /* add after headx */
                if( headx != currx )
                {
                    curr.type->nextRef = headRef.type->nextRef;
                    headRef.type->nextRef = currx;
                }

                /* handle children */
                switch ( currx->type )
                {
                    case DOFTYPE_STRUCTURE:
                        /* Structures have a list of fields (subtypes), so free them all. */
                        if ( currx->maxLength )
                        {
                            uint32 count = 0;
                            while ( count < currx->maxLength && ret)
                            {
                                ret = DOFType_TryTraverse( headx, currx->subtype[count], FALSE );
                                ++count;
                            }
                        }
                        break;
                    case DOFTYPE_ARRAY:
                    case DOFTYPE_NULLABLE:
                    case DOFTYPE_REFERENCE:
                        /* Arrays, Reference and Nullable contain an element type, so free it. */
                        ret = DOFType_TryTraverse( headx, currx->subtype[0], FALSE );
                        /* Fall through */

                    default:
                        break;
                }
            }
            else
            {
                --curr.type->testRefCount;
                PCR_Assert( curr.type->testRefCount != DOFTYPE_TRY_NOT_CHECKED );
                return TRUE;
            }

            if( ret && 
                curr.type->testRefCount != 0 && 
                DOFType_IsComposite( currx ))
            {
                ret = FALSE;
            }
        }
        else
        {
            ret = FALSE;
        }        
    }
    return ret;
}

static boolean DOFType_TryDelete( DOFType self )
{
    if( !DOFType_TryTraverse( self, self, TRUE ) )
    {
        DOFType_ResetTryData( self );
        return FALSE;
    }
    return TRUE;
}

static boolean DOFType_GC_Add( DOFType type )
{
    DOFTypeGCNode_t* node = (DOFTypeGCNode_t*)DOFMem_Alloc(sizeof(DOFTypeGCNode_t));
    if( node )
    {
        node->type = type;
        OALSList_Insert( &typeGCList, (OALSListNode)node );
        return TRUE;
    }
    return FALSE;
}

static boolean DOFType_GC_FindRemove(OALSListNode item, void *arg)
{
    DOFTypeGCNode_t* node = (DOFTypeGCNode_t*)item;
    DOFType type = (DOFType)arg;

    if( node->type == type )
        return TRUE;
    return FALSE;
}

static boolean DOFType_GC_Remove( DOFType type )
{
    DOFTypeGCNode_t* node;
    DOFType_Lock();
    node = (DOFTypeGCNode_t*)OALSList_FindAndRemove( &typeGCList, NULL, DOFType_GC_FindRemove, (void*)type);
    if( node )
    {
        DOFMem_Free(node);
        DOFType_Unlock();
        return TRUE;
    }
    DOFType_Unlock();
    return FALSE;
}

static boolean DOFType_GC_FindDelete(OALSListNode item, void *arg)
{
    DOFTypeGCNode_t* node = (DOFTypeGCNode_t*)item;
    DOFType* my = (DOFType*)arg;
    (*my) = node->type;

    return DOFType_TryDelete( node->type );
}

static void DOFType_Delete( DOFType self )
{
    DOFType curr = self;
    while( curr != NULL )
    {
        DOFType next = curr->nextRef;
        union ConstCast temp;
        temp.ctype = curr;

        if( curr->type == DOFTYPE_REFERENCE )
            DOFType_GC_Remove( curr );

        if(temp.type->testRefCount == 0)
        {
            DOFMem_Free( temp.type );
        }
        else if( !DOFType_IsComposite( self ) )
        {
            temp.type->internalRefCount = temp.type->testRefCount;
            temp.type->testRefCount = DOFTYPE_TRY_NOT_CHECKED;
            temp.type->nextRef = NULL;
        }
        curr = next;
    }
}

static void DOFType_GC_Run( void )
{
    DOFTypeGCNode_t* node = NULL;
    DOFType_Lock();
    
    do
    {
        DOFType temp = NULL;
        node = (DOFTypeGCNode_t*)OALSList_FindAndRemove( &typeGCList, NULL, DOFType_GC_FindDelete, &temp );
        if( node )
        {
            DOFType_Delete( temp );
            DOFMem_Free( node );
        }
    }
    while( node != NULL);

    DOFType_Unlock();
}
/** @internal
    Create a new DOFType.

    @param  type        DOF Type Specifier to be constructed.
    @param  typeSize    Size of the marshalled type specifier, excluding sub-elements.
    @param  subtypeCount  Count of sub-elements.  This must be the number of fields for a Structure type, 1 for an Array type, and 0 for everything else.
    @param  subtypes    Array of sub-elements.  The number of sub-elements must match the @a subtypeCount parameter.
    @param  minLength   Minimum Element Count (for Array and BLOB types).
    @param  maxLength   Maximum Element Count (for Array and BLOB types), Maximum String Length (for String type), or Total Element Count (for Structure type).
    @param  encoding    Encoding MIBenum (only for String type).

    @return A new DOFType, created as specified, or NULL if there are
    insufficient resources to allocate one.
**/
static struct DOFType_t * DOFType_Create( DOFTypeID type, uint32 typeSize, uint16 subtypeCount, const DOFType * subtypes, uint32 minLength, uint32 maxLength, uint16 encoding )
{
    struct DOFType_t * rval;

    if ( ( encoding && ( subtypeCount || minLength ) ) || ( type == DOFTYPE_STRING && encoding <= DOFVALUESTRING_UNKNOWN ) || ( minLength > maxLength ) || ( typeSize >= 0x8000 )|| ( subtypeCount >= 0x8000 ) )
        return NULL;

    rval = DOFMem_Alloc( offsetof( struct DOFType_t, subtype ) + subtypeCount * sizeof( DOFType ) );
    if ( !rval )
        return NULL;

    rval->nextRef = NULL;
    rval->testRefCount = DOFTYPE_TRY_NOT_CHECKED;
    rval->internalRefCount = 0;
    rval->externalRefCount = 1;
    rval->minLength = minLength | encoding;
    rval->maxLength = maxLength;
    rval->type = type;
    rval->typeSize = (uint16) typeSize;
    PCR_Assert( ( rval->typeSize == 0 ) || ( type & 0x1 ) );
    {
        PCRMem_Set( &rval->subtype[0], 0, subtypeCount * sizeof( DOFType ) );
        if ( subtypes )
            while ( subtypeCount-- > 0 )
            {
                rval->subtype[subtypeCount] = DOFType_Internal_Copy( subtypes[subtypeCount] );
                if ( !rval->subtype[subtypeCount] )
                {
                    DOFType_Internal_Destroy( rval );
                    return NULL;
                }
            }
    }
    return rval;
}


static DOFType DOFType_GetEffectiveType( DOFType self )
{
    if ( self && self->type == DOFTYPE_REFERENCE )
    {
        PCR_Assert( !self->subtype[0] || self->subtype[0]->type != DOFTYPE_REFERENCE );
        return( self->subtype[0] );
    }
    return( self );
}


boolean DOFType_IsDefined( DOFType self )
{
    if ( !self )
        return( FALSE );
    if ( self->type == DOFTYPE_REFERENCE )
        return( self->subtype[0] != NULL );
    return( TRUE );
}


/*
**  "Public" Methods
*/


DOFType DOFValueInt8Type_Create( void )
{
    return( DOFType_Create( DOFTYPE_INT8, 0, 0, NULL, 0, 0, 0 ) );
}


DOFType DOFValueInt16Type_Create( void )
{
    return( DOFType_Create( DOFTYPE_INT16, 0, 0, NULL, 0, 0, 0 ) );
}


DOFType DOFValueInt32Type_Create( void )
{
    return( DOFType_Create( DOFTYPE_INT32, 0, 0, NULL, 0, 0, 0 ) );
}


DOFType DOFValueInt64Type_Create( void )
{
    return( DOFType_Create( DOFTYPE_INT64, 0, 0, NULL, 0, 0, 0 ) );
}


DOFType DOFValueUInt8Type_Create( void )
{
    return( DOFType_Create( DOFTYPE_UINT8, 0, 0, NULL, 0, 0, 0 ) );
}


DOFType DOFValueUInt16Type_Create( void )
{
    return( DOFType_Create( DOFTYPE_UINT16, 0, 0, NULL, 0, 0, 0 ) );
}


DOFType DOFValueUInt32Type_Create( void )
{
    return( DOFType_Create( DOFTYPE_UINT32, 0, 0, NULL, 0, 0, 0 ) );
}


DOFType DOFValueUInt64Type_Create( void )
{
    return( DOFType_Create( DOFTYPE_UINT64, 0, 0, NULL, 0, 0, 0 ) );
}


DOFType DOFValueFloat32Type_Create( void )
{
    return( DOFType_Create( DOFTYPE_FLOAT32, 0, 0, NULL, 0, 0, 0 ) );
}


DOFType DOFValueFloat64Type_Create( void )
{
    return( DOFType_Create( DOFTYPE_FLOAT64, 0, 0, NULL, 0, 0, 0 ) );
}


DOFType DOFValueDateTimeType_Create( void )
{
    return( DOFType_Create( DOFTYPE_DATETIME, 0, 0, NULL, 0, 0, 0 ) );
}


DOFType DOFValueBooleanType_Create( void )
{
    return( DOFType_Create( DOFTYPE_BOOLEAN, 0, 0, NULL, 0, 0, 0 ) );
}


DOFType DOFValueGUIDType_Create( void )
{
    return( DOFType_Create( DOFTYPE_GUID, 0, 0, NULL, 0, 0, 0 ) );
}


DOFType DOFValueInterfaceIDType_Create( void )
{
    return( DOFType_Create( DOFTYPE_INTERFACEID, 0, 0, NULL, 0, 0, 0 ) );
}


DOFType DOFValueObjectIDType_Create( void )
{
    return( DOFType_Create( DOFTYPE_OBJECTID, 0, 0, NULL, 0, 0, 0 ) );
}


DOFType DOFValueStringType_Create( uint16 encoding, uint32 length )
{
    if ( encoding > DOFVALUESTRING_UNKNOWN )
    {
        return ( DOFType_Create( DOFTYPE_STRING, OALMarshal_GetCompressedValueSize( 3, length ) + OALMarshal_GetCompressedValueSize( 2, encoding ), 0, NULL, 0, length, encoding ) );
    }
    return NULL;
}


DOFType DOFValueBLOBType_Create( uint32 minSize, uint32 maxSize )
{
    return ( DOFType_Create( DOFTYPE_BLOB, ( ( minSize == maxSize ) ? 0 : OALMarshal_GetCompressedValueSize( 3, minSize ) ) + OALMarshal_GetCompressedValueSize( 3, maxSize ), 0, NULL, minSize, maxSize, 0 ) );
}


DOFType DOFValueStructureType_Create( uint16 count, const DOFType fields[] )
{
    uint16 i;
    for ( i = 0; i < count; i++ )
    {
        if ( !DOFType_IsDefined( fields[i] ) )
            return NULL;
    }
    return ( DOFType_Create( DOFTYPE_STRUCTURE, OALMarshal_GetCompressedValueSize( 2, count ), count, fields, count, count, 0 ) );
}


/** @internal
    Create a new Structure Type based on an array of @ref DOFValue.

    @param count   The number of fields in the Structure.
    @param fields  An array of DOFValues, whose types represent the field types for the Structure.

    @return The new DOFType, or NULL if it could not be created.
**/
DOFType DOFValueStructureType_Create_Values( uint16 count, const DOFValue fields[] )
{
    struct DOFType_t * structure = DOFType_Create( DOFTYPE_STRUCTURE, OALMarshal_GetCompressedValueSize( 2, count ), count, NULL, count, count, 0 );
    while ( count-- > 0 )
    {
        DOFType field = DOFValue_GetDOFType( fields[count] );
        structure->subtype[count] = DOFType_Internal_Copy( field );
        if ( !DOFType_IsDefined( structure->subtype[count] ) )
        {
            DOFType_Internal_Destroy( structure );
            return( NULL );
        }
    }
    return ( structure );
}


DOFType DOFValueArrayType_Create( const DOFType elementType, uint32 minLength, uint32 maxLength )
{
    if ( !elementType || ( minLength > 0 && !DOFType_IsDefined( elementType ) ) )
        return NULL;
        return ( DOFType_Create( DOFTYPE_ARRAY, ( ( minLength == maxLength ) ? 0 : OALMarshal_GetCompressedValueSize( 2, minLength ) ) + OALMarshal_GetCompressedValueSize( 2, maxLength ), 1, &elementType, minLength, maxLength, 0 ) );
}


DOFType DOFTypeTypedef_Create( const DOFType xrefType )
{
    DOFType refType = DOFType_GetEffectiveType( xrefType );
    if ( xrefType )
    {
        if ( refType && DOFType_IsDefined( refType ) && DOFType_GetTypeID( refType ) != DOFTYPE_NULLABLE )
        {
            DOFType type = DOFType_Create( DOFTYPE_REFERENCE, 0, 1, &refType, 0, 0, 0 );
            if( type )
            {
                DOFType_Lock();
                if( DOFType_IsComposite( refType ) )
                {
                    if( DOFType_GC_Add( type ) )
                    {
                        DOFType_Unlock();
                        return type;
                    }
                }
                else
                {
                    DOFType_Unlock();
                    return type;
                }
                DOFType_Unlock();
                DOFType_Destroy(type);
            }
        }
    }
    else
        return ( DOFType_Create( DOFTYPE_REFERENCE, 0, 1, NULL, 0, 0, 0 ) );
    return NULL;
}


boolean DOFTypeTypedef_DefineAs( DOFType type, const DOFType xrefType )
{
    DOFType refType = DOFType_GetEffectiveType( xrefType );
    if ( type && !DOFType_IsDefined( type ) && refType && DOFType_IsDefined( refType ) && DOFType_GetTypeID( refType ) != DOFTYPE_NULLABLE )
    {
        union ConstCast temp;
        temp.ctype = type;

        DOFType_Lock();

        temp.type->subtype[0] = DOFType_Internal_Copy( refType );
        if( temp.type->subtype[0] )
        {
            if( DOFType_IsComposite( refType ) )
            {
                if( DOFType_GC_Add( type ) )
                {
                    DOFType_Unlock();
                    return TRUE;
                }
            }
            else
            {
                DOFType_Unlock();
                return TRUE;
            }
            DOFType_Internal_Destroy( refType );
            temp.type->subtype[0] = NULL;
        }
        DOFType_Unlock();
    }
    return FALSE;
}


DOFType DOFTypeNullable_Create( const DOFType subType )
{
    if ( !subType || subType->type == DOFTYPE_NULLABLE )
        return NULL;
    return ( DOFType_Create( DOFTYPE_NULLABLE, 0, 1, &subType, 0, 0, 0 ) );
}

static DOFType DOFType_Internal_Copy( DOFType self )
{
    if ( !self )
        return NULL;
    if ( self->typeSize == 0x8000 )
    {
        /* One of the predefined types. */
        return self;
    }
    DOFType_Lock();
    if( self->internalRefCount < 0xFFFFFFFF )
    {
        union ConstCast temp;
        temp.ctype = self;
        temp.type->internalRefCount++;

        DOFType_Unlock();
        return self;
    }
    DOFType_Unlock();
    return NULL;
}

DOFType DOFType_Copy( DOFType self )
{
    if ( !self )
        return NULL;
    if ( self->typeSize == 0x8000 )
    {
        /* One of the predefined types. */
        return self;
    }
    DOFType_Lock();
    if( self->externalRefCount < 0xFFFFFFFF )
    {
        union ConstCast temp;
        temp.ctype = self;
        temp.type->externalRefCount++;

        DOFType_Unlock();
        return self;
    }
    DOFType_Unlock();
    return NULL;
}

static void DOFType_Internal_Destroy( DOFType type )
{
    if ( type )
    {
        if ( type->typeSize != 0x8000 )
        {
            boolean destroy = FALSE;
            union ConstCast temp;

            temp.ctype = type;
            DOFType_Lock();
            PCR_Assert( type->internalRefCount );
            if ( --temp.type->internalRefCount == 0 &&
                 temp.type->externalRefCount == 0 )
            {
                temp.type->externalRefCount++; /* bump and run through the normal delete */
                destroy = TRUE;
            }

            if ( destroy )
            {
                DOFType_Destroy(type);
            }
            DOFType_Unlock();
        }
    }
}

void DOFType_Destroy( DOFType type )
{
    if ( type )
    {
        if ( type->typeSize != 0x8000 )
        {
            boolean destroy = FALSE;
            union ConstCast temp;

            temp.ctype = type;
            DOFType_Lock();
            PCR_Assert( type->externalRefCount );
            if ( --temp.type->externalRefCount == 0 )
                destroy = TRUE;

            if ( destroy )
            {
                if( temp.type->internalRefCount == 0 )
                {
                    switch ( type->type )
                    {
                        case DOFTYPE_STRUCTURE:
                            /* Structures have a list of fields (subtypes), so free them all. */
                            if ( type->maxLength )
                            {
                                uint32 count = 0;
                                while ( count < type->maxLength)
                                {
                                    DOFType_Internal_Destroy( type->subtype[count] );
                                    ++count;
                                }
                            }
                            break;
                        case DOFTYPE_ARRAY:
                        case DOFTYPE_NULLABLE:
                        case DOFTYPE_REFERENCE:
                            /* Arrays, Reference and Nullable contain an element type, so free it. */
                            DOFType_Internal_Destroy( type->subtype[0] );
                            /* Fall through */

                            if( type->type == DOFTYPE_REFERENCE)
                            {
                                DOFType_GC_Remove(type);
                            }

                        default:
                            break;
                    }
                    
                    DOFMem_Free( temp.type );
                }
                else if( DOFType_IsComposite( type ) )
                {
                    DOFType_GC_Run();
                }
            }
            DOFType_Unlock();
        }
    }
}


/** @internal
    Determine the size of buffer required to marshal a @ref DOFType.

    @param  type    The DOFType to be marshalled.
    @param  context The DOFTypeContext within which the type is defined.

    @return The marshalled size of the type (the amount of buffer space needed
    to store the marshalled type data).
**/
uint32 DOFType_GetMarshalSize( const DOFType type, DOFTypeContext context )
{
    uint32 size;
    if ( !type )
        return 0;
    if ( ( type->type & 0x1 ) == 0)
    {
        return ( OALMarshal_GetCompressedValueSize( 2, type->type ) );
    }
    size = OALMarshal_GetCompressedValueSize( 2, type->type ) + OALMarshal_GetCompressedValueSize( 2, type->typeSize ) + type->typeSize;
    if ( size >= 0x8000 )
        return 0;
    switch ( type->type )
    {
        case DOFTYPE_STRUCTURE:
            {
                uint16 i = 0;
                for ( ; i < ( type->maxLength & 0x7FFF ); i++ )
                {
                    uint16 typeref = DOFTypeContext_GetIndex( context, type->subtype[i] );
                    if ( typeref >= 0x8000 )
                        return 0;
                    size += OALMarshal_GetCompressedValueSize( 2, typeref );
                    if ( size >= 0x8000 )
                        return 0;
                }
            }
            break;

        case DOFTYPE_ARRAY:
        case DOFTYPE_NULLABLE:
        case DOFTYPE_REFERENCE:
            {
                uint16 typeref = DOFTypeContext_GetIndex( context, type->subtype[0] );
                if ( typeref >= 0x8000 )
                    return 0;
                size += OALMarshal_GetCompressedValueSize( 2, typeref );
                if ( size >= 0x8000 )
                    return 0;
            }
            break;

        default:
            break;
    }
    return ( size );
}


/** @internal
    Marshal a @ref DOFType into a buffer.

    @param  type  The DOFType to be marshalled.
    @param  length  The length of the buffer in which to marshal the DOFType.
    @param  buffer  The buffer in which to marshal the DOFType, or NULL to only return the marshalled size.
    @param  context The DOFTypeContext within which the type (and its subtypes) is defined.

    @return The marshalled size of the type (the amount of buffer space used to
    store the marshalled type data).
**/
uint32 DOFType_Marshal( const DOFType type, uint32 length, uint8 *buffer, DOFTypeContext context )
{
    uint32 used = 0;
    uint32 mSize = DOFType_GetMarshalSize( type, context );

    if ( length == 0 || !buffer || !type )
        return mSize;

    PCR_Assert( mSize );
    if ( length < mSize )
        return 0;

    PCR_Assert( type->type < 0x8000 );
    used += OALMarshal_CompressValue( 2, type->type, length - used, buffer + used );
    if ( type->type & 0x1 )
    {
        PCR_Assert( mSize < 0x8000 );
        used += OALMarshal_CompressValue( 2, mSize, length - used, buffer + used );
        switch ( type->type )
        {
            case DOFTYPE_STRING:
                PCR_Assert( type->maxLength < 0x400000 );
                used += OALMarshal_CompressValue( 3, type->maxLength, length - used, buffer + used );
                used += OALMarshal_CompressValue( 2, type->minLength, length - used, buffer + used );
                break;

            case DOFTYPE_STRUCTURE:
                PCR_Assert( type->maxLength < 0x8000 );
                used += OALMarshal_CompressValue( 2, type->maxLength, length - used, buffer + used );
                {
                    uint16 i;
                    for ( i = 0; i < ( type->maxLength & 0x7FFF ); i++ )
                        used += OALMarshal_CompressValue( 2, DOFTypeContext_GetIndex( context, type->subtype[i] ), length - used, buffer + used );
                }
                break;

            case DOFTYPE_BLOB:
                PCR_Assert( type->maxLength < 0x400000 );
                PCR_Assert( type->minLength < 0x400000 );
                used += OALMarshal_CompressValue( 3, type->maxLength, length - used, buffer + used );
                if ( type->minLength != type->maxLength )
                    used += OALMarshal_CompressValue( 3, type->minLength, length - used, buffer + used );
                break;

            case DOFTYPE_NULLABLE:
                used += OALMarshal_CompressValue( 2, DOFTypeContext_GetIndex( context, type->subtype[0] ), length - used, buffer + used );
                break;

            case DOFTYPE_ARRAY:
                PCR_Assert( type->maxLength < 0x8000 );
                PCR_Assert( type->minLength < 0x8000 );
                used += OALMarshal_CompressValue( 2, DOFTypeContext_GetIndex( context, type->subtype[0] ), length - used, buffer + used );
                used += OALMarshal_CompressValue( 2, type->maxLength, length - used, buffer + used );
                if ( type->minLength != type->maxLength )
                    used += OALMarshal_CompressValue( 2, type->minLength, length - used, buffer + used );
                break;

            case DOFTYPE_REFERENCE:
                PCR_Assert( type->maxLength < 0x8000 );
                used += OALMarshal_CompressValue( 2, DOFTypeContext_GetIndex( context, type->subtype[0] ), length - used, buffer + used );
                break;

            default:
                return( 0 );
        }
    }
    PCR_Assert( used == mSize );
    return( used );
}


/** @internal
    Unmarshal a @ref DOFType from a buffer.

    @param  length  [IN/OUT] The length of the buffer from which to unmarshal the DOFType (IN); the length of the data unmarshalled (OUT).
    @param  buffer  The buffer from which to unmarshal the DOFType.
    @param  context The DOFTypeContext into which the type (and its subtypes) will be added.
    @param  must_define Indicate whether the type must be fully defined or not after unmarshal.

    @return A new DOFType created from the data within the buffer, or NULL if
    either the buffer contained an invalid type or if there are insufficient
    resources to allocate the described type.
**/
DOFType DOFType_Create_Unmarshal( uint32 * length, const uint8 * buffer, DOFTypeContext context, boolean must_define)
{
    uint32 used = 0;
    DOFType rval = NULL;
    struct DOFType_t * type;
    uint32 maxLength;
    uint32 minLength;
    uint16 typeLength = 0;

    if ( !length )
        return( NULL );
    if ( !buffer || *length < 1 )
        goto bail;

    /* Get the type specifier (store in maxLength temporarily for convenience). */
    used = *length;
    maxLength = OALMarshal_UncompressValue( 2, &used, buffer );
    if ( maxLength == OALMARSHAL_INVALID_VALUE )
        goto bail;
    if ( maxLength & 0x1 )
    {
        uint32 tlen = *length - used;

        /* Get the length of the whole type. */
        minLength = OALMarshal_UncompressValue( 2, &tlen, buffer + used );
        if ( minLength == OALMARSHAL_INVALID_VALUE )
            goto bail;
        used += tlen;
        typeLength = (uint16) minLength;
    }
    if ( *length - used < typeLength )
        goto bail;
    *length = used + typeLength;

    switch ( maxLength )
    {
        case DOFTYPE_INT8:
            rval = DOFValueInt8Type_Create();
            break;

        case DOFTYPE_INT16:
            rval = DOFValueInt16Type_Create();
            break;

        case DOFTYPE_INT32:
            rval = DOFValueInt32Type_Create();
            break;

        case DOFTYPE_INT64:
            rval = DOFValueInt64Type_Create();
            break;

        case DOFTYPE_UINT8:
            rval = DOFValueUInt8Type_Create();
            break;

        case DOFTYPE_UINT16:
            rval = DOFValueUInt16Type_Create();
            break;

        case DOFTYPE_UINT32:
            rval = DOFValueUInt32Type_Create();
            break;

        case DOFTYPE_UINT64:
            rval = DOFValueUInt64Type_Create();
            break;

        case DOFTYPE_FLOAT32:
            rval = DOFValueFloat32Type_Create();
            break;

        case DOFTYPE_FLOAT64:
            rval = DOFValueFloat64Type_Create();
            break;

        case DOFTYPE_DATETIME:
            rval = DOFValueDateTimeType_Create();
            break;

        case DOFTYPE_BOOLEAN:
            rval = DOFValueBooleanType_Create();
            break;

        case DOFTYPE_INTERFACEID:
            rval = DOFValueInterfaceIDType_Create();
            break;

        case DOFTYPE_OBJECTID:
            rval = DOFValueObjectIDType_Create();
            break;

        case DOFTYPE_GUID:
            rval = DOFValueGUIDType_Create();
            break;

        case DOFTYPE_STRING:
            {
                uint32 tlen = *length - used;
                uint32 tlen2;
                maxLength = OALMarshal_UncompressValue( 3, &tlen, buffer + used );
                if ( maxLength == OALMARSHAL_INVALID_VALUE )
                    goto bail;
                used += tlen;
                tlen2 = *length - used;
                minLength = OALMarshal_UncompressValue( 2, &tlen2, buffer + used );
                if ( minLength == OALMARSHAL_INVALID_VALUE )
                    goto bail;
                used += tlen2;
                rval = DOFType_Create( DOFTYPE_STRING, tlen + tlen2, 0, NULL, 0, maxLength, (uint16) minLength );
            }
            break;

        case DOFTYPE_BLOB:
            {
                uint32 tlen = *length - used;
                maxLength = OALMarshal_UncompressValue( 3, &tlen, buffer + used );
                if ( maxLength == OALMARSHAL_INVALID_VALUE )
                    goto bail;
                used += tlen;
                tlen = *length - used;
                if ( tlen )
                {
                    minLength = OALMarshal_UncompressValue( 3, &tlen, buffer + used );
                    if ( minLength == OALMARSHAL_INVALID_VALUE )
                        goto bail;
                    used += tlen;
                }
                else
                    minLength = maxLength;
            }
            rval = DOFValueBLOBType_Create( minLength, maxLength );
            break;

        case DOFTYPE_STRUCTURE:
            {
                uint32 tlen = *length - used;
                maxLength = OALMarshal_UncompressValue( 2, &tlen, buffer + used );
                if ( maxLength == OALMARSHAL_INVALID_VALUE )
                    goto bail;
                used += tlen;
                rval = type = DOFType_Create( DOFTYPE_STRUCTURE, tlen, (uint16) maxLength, NULL, maxLength, maxLength, 0 );
            }
            if ( type )
            {
                for ( minLength = 0; minLength < maxLength; minLength++ )
                {
                    uint32 sublen = *length - used;
                    uint32 index = OALMarshal_UncompressValue( 2, &sublen, &buffer[used] );
                    if ( index == OALMARSHAL_INVALID_VALUE )
                        goto bail;
                    type->subtype[minLength] = DOFType_Internal_Copy( DOFTypeContext_GetType( context, index, TRUE ) );
                    if ( !type->subtype[minLength] )
                        goto bail;
                    used += sublen;
                }
            }
            else
                goto bail;
            break;

        case DOFTYPE_NULLABLE:
            rval = type = DOFType_Create( DOFTYPE_NULLABLE, 0, 1, NULL, 0, 0, 0 );
            if ( rval )
            {
                uint32 sublen = *length - used;
                uint32 index = OALMarshal_UncompressValue( 2, &sublen, &buffer[used] );
                if ( index == OALMARSHAL_INVALID_VALUE )
                    goto bail;
                type->subtype[0] = DOFType_Internal_Copy( DOFTypeContext_GetType( context, index, FALSE ) );
                if ( !type->subtype[0] || type->subtype[0]->type == DOFTYPE_NULLABLE )
                    goto bail;
                used += sublen;
            }
            else
                goto bail;
            break;

        case DOFTYPE_ARRAY:
            {
                uint32 tlen1 = *length - used;
                uint32 tlen2;
                uint32 index = OALMarshal_UncompressValue( 2, &tlen1, buffer + used );
                if ( index == OALMARSHAL_INVALID_VALUE )
                    goto bail;
                used += tlen1;
                tlen1 = *length - used;
                maxLength = OALMarshal_UncompressValue( 2, &tlen1, buffer + used );
                if ( maxLength == OALMARSHAL_INVALID_VALUE )
                    goto bail;
                used += tlen1;
                tlen2 = *length - used;
                if ( tlen2 )
                {
                    minLength = OALMarshal_UncompressValue( 2, &tlen2, buffer + used );
                    if ( minLength == OALMARSHAL_INVALID_VALUE )
                        goto bail;
                    used += tlen2;
                }
                else
                    minLength = maxLength;
                rval = type = DOFType_Create( DOFTYPE_ARRAY, tlen1+tlen2, 1, NULL, minLength, maxLength, 0 );
                if ( rval )
                {
                    type->subtype[0] = DOFType_Internal_Copy( DOFTypeContext_GetType( context, index, minLength != 0 ) );
                    if ( !type->subtype[0] )
                        goto bail;
                }
                else
                    goto bail;
            }
            break;

        case DOFTYPE_REFERENCE:
            {
                uint32 tlen = *length - used;
                maxLength = OALMarshal_UncompressValue( 2, &tlen, buffer + used );
                if ( maxLength == OALMARSHAL_INVALID_VALUE )
                    goto bail;
                used += tlen;
                rval = DOFType_Create( DOFTYPE_REFERENCE, 0, 1, NULL, maxLength, maxLength, 0 );
                if ( must_define && !DOFTypeTypedef_DefineAs( rval, DOFTypeContext_GetType( context, maxLength, must_define ) ) )
                    goto bail;
            }
            if ( !rval )
                goto bail;
            break;

        default:
            goto bail;
    }

    if ( !rval || *length != used )
        goto bail;

    /* *length = used; */
    return( rval );

  bail:
    *length = 0;
    DOFType_Destroy( rval );
    return( NULL );
}



/** @internal
    Determine the size and validity of a marshalled @ref DOFType in a buffer.

    @param  length  The length of the buffer in which the DOFType was marshalled.
    @param  buffer  The buffer in which the DOFType was marshalled.

    @return The length of the data of the DOFType that could be unmarshalled,
    or 0 if the data does not appear to be a valid DOFType.
**/
uint32 DOFType_GetUnmarshalSize( uint32 length, const uint8 *buffer )
{
    uint32 used = 0;
    uint32 tlen;
    DOFTypeID type;

    if ( !buffer || length < 1 )
        return( 0 );

    tlen = length;
    if ( ( used = OALMarshal_UncompressValue( 2, &tlen, buffer ) ) == OALMARSHAL_INVALID_VALUE )
        goto bail;
    type = (DOFTypeID) used;
    used = tlen;

    if ( type & 0x1 )
    {
        uint32 var;
        tlen = length - used;
        if ( ( var = OALMarshal_UncompressValue( 2, &tlen, buffer + used ) ) == OALMARSHAL_INVALID_VALUE )
            goto bail;
        used += tlen + var;
    }
    return( used );

  bail:
    return( 0 );
}



/** @internal
    Unmarshal a @ref DOFTypeContext from a buffer.

    @param  length  [IN/OUT] The length of the buffer from which to unmarshal the DOFTypeContext (IN); the length of the data unmarshalled (OUT).
    @param  buffer  The buffer from which to unmarshal the DOFType.
    @param  context The DOFTypeContext into which the type (and its subtypes) will be added.

    @return A new DOFType created from the data within the buffer, or NULL if
    either the buffer contained an invalid type or if there are insufficient
    resources to allocate the described type.
**/
DOFTypeContext DOFTypeContext_Create_Unmarshal( uint32 * length, const uint8 * buffer )
{
    DOFTypeContext context = NULL;
    uint32 used = 0;
    uint32 i;

    if ( !length )
        return( NULL );
    if ( !buffer || *length < 1 )
        goto bail;

    context = DOFMem_Alloc( sizeof(struct DOFTypeContext_t) );
    if ( !context )
        goto bail;

    /* Get the count of types in the context. */
    used = *length;
    i = OALMarshal_UncompressValue( 2, &used, buffer );
    if ( i == OALMARSHAL_INVALID_VALUE )
        goto bail;

    context->buffer = buffer;
    context->allocCount = context->count = (uint16) i;
    context->elements = DOFMem_Alloc( sizeof(struct DOFTypeContextElement_t) * context->allocCount );
    if ( !context->elements )
        goto bail;
    PCRMem_Set( context->elements, 0, sizeof(struct DOFTypeContextElement_t) * context->allocCount );

    for ( i = 0; i < context->count; i++ )
    {
        context->elements[i].offset = used;
        context->elements[i].size = (uint16) DOFType_GetUnmarshalSize( *length - used, buffer + used );
        if ( context->elements[i].size == 0 || context->elements[i].size + used > *length )
            goto bail;
        used += context->elements[i].size;
    }

    for ( i = 0; i < context->count; i++ )
    {
        if ( !DOFTypeContext_GetType( context, i, FALSE ) ) /* Automatically stores it in the context. */
            goto bail;
    }

    for ( i = 0; i < context->count; i++ )
    {
        if ( !DOFTypeContext_GetType( context, i, TRUE ) ) /* Automatically stores it in the context. */
            goto bail;
    }

    *length = used;
    return( context );

  bail:
    *length = 0;
    DOFTypeContext_Destroy( context );
    return( NULL );
}


void DOFTypeContext_Destroy( DOFTypeContext context )
{
    if ( context )
    {
        uint16 i;
        for ( i = 0; i < context->count; i++ )
            DOFType_Destroy( context->elements[i].type );
        DOFMem_Free( context->elements );
        DOFMem_Free( context );
    }
}


static DOFType DOFTypeContext_GetType( DOFTypeContext context, uint16 index, boolean must_define )
{
    if ( index < context->count )
    {
        if ( !context->elements[index].processing )
        {
            context->elements[index].processing = TRUE;
            if ( !context->elements[index].type )
            {
                uint32 tlen = context->elements[index].size;
                context->elements[index].type = DOFType_Create_Unmarshal( &tlen, context->buffer + context->elements[index].offset, context, must_define );
                if ( !context->elements[index].type || tlen != context->elements[index].size )
                {
                    context->elements[index].processing = FALSE;
                    return NULL;
                }
            }
            if ( must_define && !DOFType_IsDefined( context->elements[index].type ) &&
                !DOFTypeTypedef_DefineAs( context->elements[index].type, DOFTypeContext_GetType( context, context->elements[index].type->maxLength, must_define ) ) )
            {
                context->elements[index].processing = FALSE;
                return NULL;
            }
            context->elements[index].processing = FALSE;
            return context->elements[index].type;
        }
    }
    return NULL;
}


DOFType DOFTypeContext_UnmarshalTypeRef( DOFTypeContext context, uint32 * length, const uint8 * buffer )
{
    uint32 used = 0;
    uint32 ref;
    DOFType rval = NULL;

    if ( !length )
        return( NULL );
    if ( !buffer || *length < 1 )
        goto bail;

    /* Get the type reference. */
    used = *length;
    ref = OALMarshal_UncompressValue( 2, &used, buffer );
    if ( ref == OALMARSHAL_INVALID_VALUE )
        goto bail;
    if ( ref < context->count )
        rval = context->elements[ref].type;

    *length = used;
    return( rval );

  bail:
    *length = 0;
    return( NULL );
}


static uint16 DOFTypeContext_GetIndex( DOFTypeContext context, DOFType type )
{
    uint16 i;
    for ( i = 0; i < context->count; i++ )
    {
        if ( context->elements[i].type == type )
            return i;
    }
    
    /* Note: When Interface Builder is implemented, this function should append new type of the end of the array */
    return 0xFFFF;
}


DOFTypeID DOFType_GetTypeID( const DOFType ttype )
{
    DOFType type = DOFType_GetEffectiveType( ttype );
    if ( !type )
        return( DOFTYPE_VOID );
    return( type->type );
}


/**
    Report whether or not the type has a meaningful, defined character encoding format (e.g., a string).
    Types that have an encoding can report useful data via @ref DOFType_GetEncoding.

    @param  type  The DOFType.  This must not be NULL.

    @return TRUE if the type has an encoding.
    @return FALSE if the type does not have an encoding.
**/
static boolean DOFType_HasEncoding( const DOFType ttype )
{
    DOFType type = DOFType_GetEffectiveType( ttype );
    if ( !type )
        return FALSE;
    return( type->type == DOFTYPE_STRING );
}


/**
    Report whether or not the type has a meaningful length.
    Types that have length can report useful data via @ref DOFType_GetMaxLength and @ref DOFType_GetMinLength.

    @param  type  The DOFType.  This must not be NULL.

    @return TRUE if the type has a length.
    @return FALSE if the type does not have a length.
**/
static boolean DOFType_HasLength( const DOFType ttype )
{
    DOFType type = DOFType_GetEffectiveType( ttype );
    if ( !type )
        return FALSE;
    return( ( type->type == DOFTYPE_BLOB ) || ( type->type == DOFTYPE_ARRAY ) || ( type->type == DOFTYPE_STRING ) );
}


/**
    Report whether or not the type has meaningful subtypes.
    Subtypes can be array elements, nullable targets, or structure fields, etc.
    Types that have subtypes can report useful data via @ref DOFType_GetSubtype and @ref DOFType_GetSubtypes.

    @param  type  The DOFType.  This must not be NULL.

    @return TRUE if the type has subtypes.
    @return FALSE if the type does not have subtypes.
**/
static boolean DOFType_HasSubtypes( const DOFType ttype )
{
    DOFType type = DOFType_GetEffectiveType( ttype );
    if ( !type )
        return FALSE;
    return( ( type->type == DOFTYPE_NULLABLE ) || ( type->type == DOFTYPE_STRUCTURE ) || ( type->type == DOFTYPE_ARRAY ) );
}


uint32 DOFType_GetMaxLength( const DOFType ttype )
{
    DOFType type = DOFType_GetEffectiveType( ttype );
    if ( !type || !DOFType_HasLength( type ) )
        return 0;
    return( type->maxLength );
}


uint32 DOFType_GetMinLength( const DOFType ttype )
{
    DOFType type = DOFType_GetEffectiveType( ttype );
    if ( !type || !DOFType_HasLength( type ) || DOFType_HasEncoding( type ) )
        return 0;
    return( type->minLength );
}


uint16 DOFType_GetEncoding( const DOFType ttype )
{
    DOFType type = DOFType_GetEffectiveType( ttype );
    if ( type && DOFType_HasEncoding( type ) )
        return( (uint16) type->minLength );
    return DOFVALUESTRING_UNKNOWN;
}


uint32 DOFType_GetSubtypesCount( const DOFType ttype )
{
    DOFType type = DOFType_GetEffectiveType( ttype );
    if ( !type || !DOFType_HasSubtypes( type ) )
        return 0;
    if ( type->type == DOFTYPE_STRUCTURE )
        return( type->maxLength );
    return 1;
}


DOFType DOFType_GetSubtype( const DOFType ttype, uint32 index )
{
    DOFType type = DOFType_GetEffectiveType( ttype );
    if ( index < DOFType_GetSubtypesCount( type ) )
        return( type->subtype[ index ] );
    return (const DOFType) NULL;
}


const DOFType * DOFType_GetSubtypes( const DOFType ttype )
{
    DOFType type = DOFType_GetEffectiveType( ttype );
    if ( DOFType_HasSubtypes( type ) )
        return( &type->subtype[0] );
    return NULL;
}


/** Value unmarshalling (marshaling done in DOFValue) **/

/** @internal
    Unmarshal a @ref DOFValue from a buffer, based on the given type.

    @param  type   The DOFType of the DOFValue to be unmarshalled.
    @param  length [IN] The total length of the buffer; [OUT] the length of the data unmarshalled.
    @param  buffer The buffer in which the DOFValue is marshalled.

    @return The new @ref DOFValue created.  A Nullable type may cause a NULL to be unmarshalled (correctly), in which case the @a length will be non-zero (since data was read).

    @exception NULL is returned and @a length is zero if insufficient resources
               are available to create the value or there were insufficient
               data bytes in the buffer to correctly unmarshal the specified
               @ref DOFType.
**/
DOFValue DOFType_CreateValue_Unmarshal( const DOFType const_type, uint32 * length, const uint8 *buffer )
{
    uint32 used = 0;
    uint32 size;
    DOFValue rval = NULL;
    DOFType type = DOFType_GetEffectiveType( const_type );
    DOFTypeID typeSpec = DOFType_GetTypeID( type );

    switch ( typeSpec )
    {
        case DOFTYPE_INT8:
        case DOFTYPE_UINT8:
        case DOFTYPE_BOOLEAN:
            size = 1;
            goto FIXED;

        case DOFTYPE_INT16:
        case DOFTYPE_UINT16:
            size = 2;
            goto FIXED;

        case DOFTYPE_INT32:
        case DOFTYPE_UINT32:
        case DOFTYPE_FLOAT32:
            size = 4;
            goto FIXED;

        case DOFTYPE_INT64:
        case DOFTYPE_UINT64:
        case DOFTYPE_FLOAT64:
        case DOFTYPE_DATETIME:
            size = 8;
            goto FIXED;

        case DOFTYPE_INTERFACEID:
            {
                DOFInterfaceID iid;
                size = *length - used;
                iid = DOFInterfaceID_Create_Unmarshal(&size, &buffer[used]);
                if( iid )
                    rval = DOFValue_Create(DOFType_Copy(type), sizeof(iid), (uint8 *)&iid, FALSE, 0); 
                used += size;
                if ( !rval )
                {
                    DOFInterfaceID_Destroy(iid);
                    goto bail;
                }
                break;
            }

        case DOFTYPE_OBJECTID:
            {
                DOFObjectID oid;
                size = *length - used;
                oid = DOFObjectID_Create_Unmarshal(&size, &buffer[used]);
                if( oid )
                    rval = DOFValue_Create(DOFType_Copy(type), sizeof(oid), (uint8 *)&oid, FALSE, 0); 
                used += size;
                if ( !rval )
                {
                    DOFObjectID_Destroy(oid);
                    goto bail;
                }
                break;
            }

        case DOFTYPE_GUID:
            size = 16;
        FIXED:
            if ( *length < used + size )
                goto bail;
            rval = DOFValue_Create( DOFType_Copy( type ), size, &buffer[used], TRUE, 0 );
            used += size;
            if ( !rval )
                goto bail;
            break;

        case DOFTYPE_NULLABLE:
            size = 1;
            if ( *length < used + 1 )
                goto bail;
            if ( buffer[used++] )
            {
                uint32 subLength = *length - used;
                rval = DOFType_CreateValue_Unmarshal( type->subtype[0], &subLength, &buffer[used] );
                if ( !rval ) /* Do not permit NULLABLE NULLABLEs. */
                    goto bail;
                used += subLength;
            }
            else
                rval = NULL;
            break;

        case DOFTYPE_STRING:
        case DOFTYPE_BLOB:
            used = *length;
            size = OALMarshal_UncompressValue( 3, &used, &buffer[0] );
            if ( size == 0xFFFFFFFF || *length < used + size )
                goto bail;
            if ( ( size < type->minLength && typeSpec != DOFTYPE_STRING ) || size > type->maxLength )
                goto bail;
            rval = DOFValue_Create( DOFType_Copy( type ), size, &buffer[used], TRUE, (typeSpec == DOFTYPE_STRING)? DOFType_GetEncoding( type ) : 0 );
            used += size;
            if ( !rval )
                goto bail;
            break;

        case DOFTYPE_STRUCTURE:
            size = type->maxLength;
            rval = DOFValue_Create( DOFType_Copy( type ), size, NULL, FALSE, 0 );
            if ( rval )
            {
                uint32 i = 0;

                for ( ; i < size; i++ )
                {
                    uint32 subLength = *length - used;
                    rval->value[i] = DOFType_CreateValue_Unmarshal( type->subtype[i], &subLength, &buffer[used] );
                    if ( subLength == 0 )
                        goto bail;
                    used += subLength;
                }
            }
            else
                goto bail;
            break;

        case DOFTYPE_ARRAY:
            used = *length;
            size = OALMarshal_UncompressValue( 2, &used, &buffer[0] );
            if ( size == 0xFFFFFFFF )
                goto bail;
            if ( size < type->minLength || size > type->maxLength )
                goto bail;
            rval = DOFValue_Create( DOFType_Copy( type ), size, NULL, FALSE, 0 );
            if ( rval )
            {
                uint16 i = 0;

                for ( ; i < size; i++ )
                {
                    uint32 subLength = *length - used;
                    rval->value[i] = DOFType_CreateValue_Unmarshal( type->subtype[0], &subLength, &buffer[used] );
                    if ( subLength == 0 )
                        goto bail;
                    used += subLength;
                }
            }
            else
                goto bail;
            break;

        default:
            goto bail;
    }

    *length = used;
    return( rval ); /* May be NULL if the type is NULLABLE. */

  bail:
    *length = 0;
    DOFValue_Destroy( rval );
    return( NULL );
}


/** @internal
    Unmarshal a @ref DOFValue from an @ref OALStream.

    @param  type    The DOFType of the DOFValue to be unmarshalled.
    @param  stream  The OALStream in which the DOFValue is marshalled.
    @param  failed  [OUT] Set to zero (FALSE) when value was successfully unmarshaled (including a NULL value), or non-zero (TRUE) if the unmarshal failed.

    @return The new @ref DOFValue created, or NULL if insufficient resources
    are available to create it or there were insufficient data bytes in the
    stream to correctly unmarshal the specified @ref DOFType.
**/
DOFValue DOFType_CreateValue_UnmarshalFromStream( const DOFType self, OALStream stream, boolean * failed )
{
    uint32 marshalSize = OALStream_GetDataSize( stream );
    DOFValue val = DOFType_CreateValue_Unmarshal( self, &marshalSize, OALStream_GetReadPtr( stream ) );
    if ( marshalSize )
        OALStream_AdjustReadOffset( stream, (int32) marshalSize & 0x7FFFFFFFUL );
    *failed = ( marshalSize == 0 );
    return( val );
}


/**
@}
**/
