/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/type.h>
#include <dof/error.h>
#include <internal/type.h>
#include <internal/value.h>


/** @addtogroup DOFType
@{
**/


/*
**  "Private" Type and Method Declarations
*/

extern PCRMutex ref_count_mutex;

/*
**  "Public" Methods
*/

void DOFTypeList_Destroy( uint16 valueCount, DOFType typeList[] )
{
    uint16 i;

    for ( i = 0; i < valueCount; i++ )
    {
        DOFType_Destroy( typeList[i] );
        typeList[i] = NULL;
    }
}


DOFType * DOFTypeListAlloc_Create( uint16 typeCount )
{
    DOFType * rval = DOFMem_Alloc( sizeof(DOFType) * ( (typeCount)? typeCount : 1 ) );

    if ( rval )
    {
        PCRMem_Set( rval, 0, sizeof(DOFType) * typeCount );
    }
    return rval;
}


void DOFTypeListAlloc_Destroy( uint16 typeCount, DOFType * typeList )
{
    if ( typeList )
    {
        DOFTypeList_Destroy( typeCount, typeList );
        DOFMem_Free( typeList );
    }
}


DOFType * DOFTypeListAlloc_Copy( uint16 typeCount, const DOFType typeList[] )
{
    uint16 i;
    DOFType * rval = DOFTypeListAlloc_Create( typeCount );

    if ( rval )
    {
        for ( i = 0; i < typeCount; i++ )
        {
            rval[i] = DOFType_Copy( typeList[i] );
            if ( !rval[i] )
            {
                DOFTypeList_Destroy( typeCount, rval );
                DOFMem_Free( rval );
                return NULL;
            }
        }
    }
    return rval;
}


boolean DOFTypeList_IsValueCompatibleWith( uint16 count, const DOFType types[], const DOFValue values[] )
{
    uint16    i;

    if (!values || !types)
        return (count > 0) ? FALSE : TRUE;
    /* Proceed now knowing that types and values are not NULL. */
    for (i = 0; i < count; i++)
    {
        if (!DOFValue_CheckCompatible( values[i], types[i] ))
            return FALSE;
    }

    return TRUE;
}


/* This class legally accepts its "self" parameter as NULL. */
uint32 DOFTypeList_GetMarshalValueSize( uint16 count, const DOFType types[], const DOFValue values[] )
{
    uint16 i;
    uint32 size = 0;

    if (!values || !types)
        return 0;
    /* Proceed now knowing that types and values are not NULL. */
    for (i = 0; i < count; i++)
        size += DOFValue_GetMarshalSize( values[i], types[i] );

    return size;
}


/* This class legally accepts its "self" parameter as NULL. */
boolean DOFTypeList_MarshalValueToStream( uint16 count, const DOFType types[], const DOFValue values[], OALStream stream )
{
    uint16 i;

    if (!values || !types)
        return 0;
    /* Proceed now knowing that types and values are not NULL. */
    for (i = 0; i < count; i++)
        if ( !DOFValue_MarshalToStream( values[i], types[i], stream ) )
            return FALSE;
    return TRUE;
}


/**
@}
**/
