/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/value.h>
#include <dof/type.h>
#include "valueprivate.h"


/** @addtogroup DOFValueArray
@{
**/


/*
**  "Private" Type and Method Declarations
*/

/*
**  "Public" Methods
*/


DOFValue DOFValueArray_Create( uint16 elementCount, const DOFValue elements[] )
{
    DOFType elementType = DOFTYPEVOID_TYPE;
    uint16 i;

    if ( elementCount && ( !elements || ( elementCount & 0x8000 ) ) )
        return NULL;
    if ( elementCount > 0 )
    {
        DOFValue elementTypeValue = elements[0];
        elementType = DOFValue_GetDOFType( elementTypeValue );

        for ( i = 1; i < elementCount; i++ )
        {
            if ( !DOFValue_CheckCompatible( elements[i], elementType ) )
            {
                elementType = DOFValue_GetDOFType( elements[i] );
                if ( !DOFValue_CheckCompatible( elementTypeValue, elementType ) )
                {
                    return NULL;
                }
                elementTypeValue = elements[i];
            }
        }
    }
    return( DOFValue_Create( DOFValueArrayType_Create( elementType, elementCount, elementCount ), elementCount, (const uint8 *)elements, FALSE, 0 ) );
}


DOFValue DOFValueArray_Create_Type( const DOFType type, uint16 elementCount, const DOFValue elements[] )
{
    uint16 i;

    if ( !type || ( elementCount && ( !elements || ( elementCount & 0x8000 ) ) ) || ( DOFType_GetTypeID( type ) != DOFTYPE_ARRAY ) || ( elementCount > DOFType_GetMaxLength( type ) ) || ( elementCount < DOFType_GetMinLength( type ) ) )
        return NULL;
    if ( elementCount > 0 )
    {
        DOFType elementType = DOFType_GetSubtype( type, 0 );

        for ( i = 0; i < elementCount; i++ )
        {
            if ( !DOFValue_CheckCompatible( elements[i], elementType ) )
            {
                return NULL;
            }
        }
    }
    return( DOFValue_Create( DOFType_Copy( type ), elementCount, (const uint8 *)elements, FALSE, 0 ) );
}


uint16 DOFValueArray_GetElementsCount( const DOFValue self )
{
    if( !self || DOFType_GetTypeID(self->type) != DOFTYPE_ARRAY )
        return 0;
    return(uint16)( self->length );
}


DOFValue DOFValueArray_GetElement( const DOFValue self, uint16 element )
{
    if( !self || DOFType_GetTypeID(self->type) != DOFTYPE_ARRAY )
        return NULL;

    if ( element < self->length )
        return( self->value[element] );
    return( NULL );
}


const DOFValue * DOFValueArray_GetElements( const DOFValue self )
{
    if( !self || DOFType_GetTypeID(self->type) != DOFTYPE_ARRAY )
        return NULL;

    return( self->value );
}


/** @internal
    Create a new native uint8 array of from the contents of the specified @ref
    DOFValue.

    This is a helper function that can be used to extract string/blob array types from a DOFValueArray.

    @note This allocates resources that must later be deallocated by
    calling PCRMem_Free().

    @param  value           The DOFValue to be extracted.
    @param  expectedType    The expected element type of the array (for type checking).
    @param  elementOffset   Offset of the first array element to extract (0 starts at the beginning of the array).
    @param  elementCount    Maximum number of array elements to extract (less may be returned if the end of the array is reached).
    @param  destroyDOFValueArray Destroy the @a value before return (if TRUE).

    @return The new native uint8 array from @a value, or NULL if it could not be created.

    @see    DOFValueArray_GetElementsCount
**/
uint8 ** DOFValueArray_ToComboMalloc( const DOFValue value, DOFTypeID expectedType, uint16 elementOffset, uint16 elementCount, boolean destroyDOFValueArray )
{
    uint8 ** rval = NULL;
    uint8 * dataPtr;
    uint32 allocSize = sizeof(uint8 *) * elementCount;
    uint16 i;
    DOFType valueType = DOFValue_GetDOFType( value );

    /* The request is invalid if there is no value, it is not the correct type, or the elementOffset and/or elementCount specify results that are outside the array bounds. */
    if( value &&
        DOFType_GetMaxLength( valueType ) &&
        DOFType_GetSubtypesCount( valueType ) &&
        DOFType_GetTypeID( DOFType_GetSubtype( valueType, 0 ) ) == expectedType &&
        elementOffset < value->length &&
        elementCount > 0 &&
        ( (uint32) elementOffset + elementCount ) <= value->length )
    {
        /* Strings are stored with a NULL terminator, just in case.  It is *not* stored in length. */
        boolean isString = (expectedType == DOFTYPE_STRING);
        /* Determine the size of the data we will be allocating. */
        for ( i = 0; i < elementCount; i++ )
            allocSize += value->value[i + elementOffset]->length + isString;

        /* Allocate the block -- the first part holds the pointers into the second part where each value's data is stored end-to-end. */
        rval = (uint8 **) PCRMem_Alloc( allocSize );
        if ( rval )
        {
            dataPtr = ( (uint8 *) rval ) + sizeof(uint8 *) * elementCount;

            for ( i = 0; i < elementCount; i++ )
            {
                uint32 currLength = value->value[i + elementOffset]->length + isString;

                rval[i] = dataPtr;
                PCRMem_Copy( dataPtr, value->value[i + elementOffset]->data, currLength );
                dataPtr += currLength;
            }
        }
    }

    if ( destroyDOFValueArray )
        DOFValue_Destroy( value );

    return( rval );
}


/**
@}
**/
