/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <dof/core.h>
#include <dof/interfaceid.h>
#include <internal/interfaceidinternal.h>
#include <internal/marshal.h>
#include <dof/type.h>
#include <internal/type.h>
#include <dof/objectid.h>
#include <internal/objectidinternal.h>
#include <dof/value.h>
#include "valueprivate.h"

/** @addtogroup DOFValue
@{
**/

extern PCRMutex ref_count_mutex;

/*
**  "Private" Type and Method Declarations
*/


/*
**  "Public" Methods
*/

/** @internal
    Create a new DOFValue.

    @note This allocates resources that must later be deallocated by calling @ref DOFValue_Destroy.

    Functions returning a @ref DOFValue parameter always transfer ownership
    of the @ref DOFValue, so there is no need to free them.  Conversely,
    functions implemented by the application (such as Provider callbacks) that
    take a @ref DOFValue parameter always transfer ownership as well; these
    must therefore be freed by the application.

    @param  type  The type of value to be created.  This becomes owned by the DOFValue.  ALWAYS copy or create a new type for this parameter.
    @param  size  The size of the data field to be created.
    @param  data  The value data, if any.
    @param  isMarshalled  A flag indicating if the data specified is in a marshalled format.
    @param  encoding    Encoding MIBenum (only for String type).

    @return The new @ref DOFValue.

    @see    DOFValue_Destroy
**/
DOFValue DOFValue_Create( DOFType type, uint32 size, const uint8 *data, boolean isMarshalled, uint16 encoding )
{
    DOFValue rval;
    uint32 allocSize;
    uint32 dataSize;
    DOFTypeID typespec;

    if ( !type )
        return NULL;

    typespec = DOFType_GetTypeID( type );
    switch ( typespec )
    {
        case DOFTYPE_STRUCTURE:
        case DOFTYPE_ARRAY:
            dataSize = size * sizeof(DOFValue);
            allocSize = offsetof(struct DOFValue_t, value) + dataSize;
            break;
        case DOFTYPE_STRING:
            ++size; /* NULL Termination */
            /* Fall through */
        default:
            dataSize = size;
            allocSize = offsetof(struct DOFValue_t, data) + dataSize;
            break;
    }
    rval = (DOFValue) DOFMem_Alloc( allocSize );
    if ( rval )
    {
        PCRMem_Set( rval, 0, allocSize );
        rval->type = type;
        if ( rval->type )
        {
            rval->refCount = 1;
            rval->length = size;
            rval->encoding = encoding;
            if ( data )
            {
                switch ( typespec )
                {
                    case DOFTYPE_STRING:
                        rval->data[--size] = 0;
                        --(rval->length);
                        /* Fall through */
                    default:
                        PCRMem_Copy( &rval->data[0], data, size );
                        break;
                    case DOFTYPE_STRUCTURE:
                    case DOFTYPE_ARRAY:
                    {
                        uint32 i;

                        PCRMem_Set( &rval->value[0], 0, dataSize );
                        for ( i = 0; i < size; i++ )
                        {
                            DOFValue v = ((DOFValue *)data)[i];
                            if ( v )
                            {
                                rval->value[i] = DOFValue_Copy( v );
                                if ( !rval->value[i] )
                                {
                                    DOFValue_Destroy( rval );
                                    DOFType_Destroy( type );
                                    return NULL;
                                }
                            }
                        }
                    }
                }
            }
            return( rval );
        }
        DOFMem_Free( rval );
    }
    DOFType_Destroy( type );
    return NULL;
}


DOFValue DOFValue_Copy(const DOFValue self)
{
    if(!self)
        return NULL;

    PCRMutex_Lock(ref_count_mutex);
    if( self->refCount < 0xFFFFFFFF )
    {
        self->refCount++;
        PCRMutex_Unlock(ref_count_mutex);
        return self;
    }
    else
    {
        PCRMutex_Unlock(ref_count_mutex);
        return NULL;
    }
}

void DOFValue_Destroy( DOFValue value )
{
    if (value != NULL)
    {
        boolean gone = FALSE;
        PCRMutex_Lock(ref_count_mutex);
        PCR_Assert( value->refCount );
        (value->refCount)--;
        if( !(value->refCount) )
            gone = TRUE;
        PCRMutex_Unlock(ref_count_mutex);

        if ( gone )
        {
            /* Free Structure, Array, and special object types! */
            switch ( DOFType_GetTypeID( value->type ) )
            {
                case DOFTYPE_STRUCTURE:
                case DOFTYPE_ARRAY:
                    {
                        uint16 i = 0;
                        PCR_Assert( value->length < 0x8000 );
                        for ( ; i < value->length; i++ )
                            DOFValue_Destroy( value->value[i] );
                    }
                    break;

                case DOFTYPE_INTERFACEID:
                    {
                        DOFInterfaceID iid;
                        PCRMem_Copy( &iid, &value->data[0], sizeof(iid) );
                        DOFInterfaceID_Destroy( iid );
                    }
                    break;

                case DOFTYPE_OBJECTID:
                    {
                        DOFObjectID oid;
                        PCRMem_Copy( &oid, &value->data[0], sizeof(oid) );
                        DOFObjectID_Destroy( oid );
                    }
                    break;

                default:
                    break;
            }
            DOFType_Destroy( value->type );
            DOFMem_Free( value );
        }
    }
}

DOFType DOFValue_GetDOFType( const DOFValue value )
{
    if( !value )
        return NULL;
    return( value->type );
}


boolean DOFValue_CheckCompatible( DOFValue value, DOFType type )
{
    /* Handle NULL / NULLABLE compatibility. */
    if ( DOFType_GetTypeID( type ) == DOFTYPE_NULLABLE )
        return( !value || DOFValue_IsCompatibleWith( value, DOFType_GetSubtype( type,  0 ) ) );
    return( DOFValue_IsCompatibleWith( value, type ) );
}


boolean DOFValue_IsCompatibleWith( DOFValue value, DOFType type )
{
    /* Only a NULL value is compatible with the NULLABLE type. */
    DOFType valueType = DOFValue_GetDOFType( value );
    DOFTypeID typeSpec, valueTypeSpec;

    typeSpec = DOFType_GetTypeID( type );

    if ( !type )
        return( FALSE );
    if ( valueType == type )
        return( TRUE );

    if ( DOFType_IsNullable( valueType ) )
    {
        valueType = DOFType_GetSubtype( valueType, 0 );
        if ( valueType == type )
            return( TRUE );
    }
    PCR_Assert( !DOFType_IsNullable( valueType ) );
    valueTypeSpec = DOFType_GetTypeID( valueType );

    if ( value && !valueType )
        return( FALSE );

    PCR_Assert( value || valueTypeSpec == DOFTYPE_VOID );

    switch ( typeSpec )
    {
        case DOFTYPE_VOID:
            /* Everything is technically compatible, but the type should never be VOID. */
            return( FALSE );

        case DOFTYPE_STRUCTURE:
            if ( valueTypeSpec != typeSpec )
                return( FALSE );
            if ( value->length != DOFType_GetSubtypesCount( type ) )
                return( FALSE );
            {
                const DOFValue * fields = DOFValueStructure_GetFields( value );
                uint32 i = value->length;
                while ( i-- > 0 )
                {
                    if ( !DOFValue_CheckCompatible( fields[i], DOFType_GetSubtype( type, i ) ) )
                        return( FALSE );
                }
            }
            break;

        case DOFTYPE_ARRAY:
            /* Permit null for 0-length array types. */
            if ( !value )
                return( DOFType_GetMinLength( type ) == 0 );
            if ( valueTypeSpec != typeSpec )
                return( FALSE );
            if ( ( value->length > DOFType_GetMaxLength( type ) ) ||
                 ( value->length < DOFType_GetMinLength( type ) ) )
                return( FALSE );
            {
                const DOFValue * elements = DOFValueArray_GetElements( value );
                uint32 i = value->length;
                DOFType elementType = DOFType_GetSubtype( type, 0 );
                while ( i-- > 0 )
                    if ( !DOFValue_CheckCompatible( elements[i], elementType ) )
                        return( FALSE );
            }
            break;

        case DOFTYPE_NULLABLE:
            /* Permit null for NULLABLE type. */
            if ( !value )
                return( TRUE );
            /* A value cannot be NULLABLE; it is either NULL or it is a value. */
            return( FALSE );
            break;

        case DOFTYPE_STRING:
            if ( ( valueTypeSpec != typeSpec ) ||
                 ( DOFType_GetMaxLength( valueType ) > DOFType_GetMaxLength( type ) ) ||
                 ( DOFType_GetEncoding( valueType ) != DOFType_GetEncoding( type ) ) )
                return( FALSE );
            break;

        case DOFTYPE_BLOB:
            if ( ( valueTypeSpec != typeSpec ) ||
                 ( value->length > DOFType_GetMaxLength( type ) ) ||
                 ( value->length < DOFType_GetMinLength( type ) ) )
                return( FALSE );
            break;

        default:
            if ( valueTypeSpec != typeSpec )
                return( FALSE );
            /* Nothing else to compare... */
            break;
    }
    return( TRUE );
}


/** Value marshalling (unmarshaling done in DOFType) **/

/** @internal
    Report the number of bytes required in a buffer to marshal the given @ref DOFValue.

    @param  value The @ref DOFValue.  This may be NULL to marshal a null value.
    @param  type  The @ref DOFType indicating how to marshal the @a value.  This must not be NULL.

    @return The number of bytes required.
**/
uint32 DOFValue_GetMarshalSize( const DOFValue value, const DOFType type )
{
    DOFType valueType = DOFValue_GetDOFType( value );
    DOFTypeID valueTypeSpec = DOFType_GetTypeID( valueType );
    DOFTypeID typeSpec = DOFType_GetTypeID( type );
    uint8 lenbytes = 0;

    switch ( typeSpec )
    {
        case DOFTYPE_VOID:
            /* Type should never be VOID. */
            return 0;

        case DOFTYPE_INT8:
        case DOFTYPE_INT16:
        case DOFTYPE_INT32:
        case DOFTYPE_INT64:
        case DOFTYPE_UINT8:
        case DOFTYPE_UINT16:
        case DOFTYPE_UINT32:
        case DOFTYPE_UINT64:
        case DOFTYPE_FLOAT32:
        case DOFTYPE_FLOAT64:
        case DOFTYPE_DATETIME:
        case DOFTYPE_BOOLEAN:
        case DOFTYPE_GUID:
            if ( !value || typeSpec != valueTypeSpec )
                return 0;
            return( value->length );

        case DOFTYPE_INTERFACEID:
            if ( !value || typeSpec != valueTypeSpec )
                return 0;
            {
                DOFInterfaceID iid;
                /* Copy in case data field is not aligned correctly. */
                PCRMem_Copy( &iid, &value->data[0], sizeof(iid) );
                return( DOFInterfaceID_GetBytesSize( iid ) );
            }

        case DOFTYPE_OBJECTID:
            if ( !value || typeSpec != valueTypeSpec )
                return 0;
            {
                DOFObjectID oid;
                /* Copy in case data field is not aligned correctly. */
                PCRMem_Copy( &oid, &value->data[0], sizeof(oid) );
                return oid ? DOFObjectID_GetBytesSize( oid ) : 0;
            }

        case DOFTYPE_STRING:
        case DOFTYPE_BLOB:
            if ( !value || !DOFValue_CheckCompatible( value, type ) )
                return 0;
            lenbytes = OALMarshal_GetCompressedValueSize( 3, value->length );
            return( value->length + lenbytes );

        case DOFTYPE_ARRAY:
            if ( !value || typeSpec != valueTypeSpec )
                return 0;
            PCR_Assert( value->length < 0x8000 );
            lenbytes = OALMarshal_GetCompressedValueSize( 2, value->length );
            {
                uint16 i;
                uint32 size = lenbytes;
                DOFType elementType = DOFType_GetSubtype( type, 0 );
                for ( i = 0; i < value->length; i++ )
                    size += DOFValue_GetMarshalSize( value->value[i], elementType );
                return( size );
            }

        case DOFTYPE_STRUCTURE:
            if ( !value || typeSpec != valueTypeSpec )
                return 0;
            PCR_Assert( value->length < 0x8000 );
            {
                uint16 i;
                uint32 size = lenbytes;
                for ( i = 0; i < value->length; i++ )
                    size += DOFValue_GetMarshalSize( value->value[i], DOFType_GetSubtype( type, i ) );
                return( size );
            }

        case DOFTYPE_NULLABLE:
            if ( value )
            {
                uint32 size = DOFValue_GetMarshalSize( value, DOFType_GetSubtype( type, 0 ) );
                if ( !size )
                    return 0;
                return size + 1;
            }
            return( 1 );

        default:
            break;
    }
    return 0;
}


/** @internal
    Marshal a @ref DOFValue into a buffer.

    @param  value  The @ref DOFValue to be marshalled.  This may be NULL to marshal a null value.
    @param  type   The @ref DOFType indicating how to marshal the @a value.  This must not be NULL.
    @param  length The length of the buffer in which to marshal the DOFValue.
    @param  buffer The buffer in which to marshal the DOFValue, or NULL to only return the marshalled size.

    @return The length of the marshalled DOFValue, or 0 if it does not fit in the buffer.
**/
uint32 DOFValue_Marshal( const DOFValue value, const DOFType type, uint32 length, uint8 * buffer )
{
    DOFType valueType = DOFValue_GetDOFType( value );
    DOFTypeID valueTypeSpec = DOFType_GetTypeID( valueType );
    DOFTypeID typeSpec = DOFType_GetTypeID( type );
    uint32 mSize = DOFValue_GetMarshalSize( value, type );
    uint32 used = 0;

    if ( !buffer )
        return mSize;

    if ( length < mSize )
        return 0;

    switch ( typeSpec )
    {
        case DOFTYPE_VOID:
            /* Type should never be VOID. */
            return 0;

        case DOFTYPE_INT8:
        case DOFTYPE_INT16:
        case DOFTYPE_INT32:
        case DOFTYPE_INT64:
        case DOFTYPE_UINT8:
        case DOFTYPE_UINT16:
        case DOFTYPE_UINT32:
        case DOFTYPE_UINT64:
        case DOFTYPE_FLOAT32:
        case DOFTYPE_FLOAT64:
        case DOFTYPE_DATETIME:
        case DOFTYPE_BOOLEAN:
        case DOFTYPE_GUID:
            if ( !value || typeSpec != valueTypeSpec )
                return 0;
            PCRMem_Copy( &buffer[used], &value->data[0], value->length );
            used += value->length;
            break;

        case DOFTYPE_INTERFACEID:
            if ( !value || typeSpec != valueTypeSpec )
                return 0;
            {
                DOFInterfaceID iid;
                uint16 len;
                PCR_Assert( value && DOFValue_CheckCompatible( value, type ) );
                PCRMem_Copy( &iid, &value->data[0], sizeof(iid) );
                PCRMem_Copy( &buffer[used], DOFInterfaceID_GetBytes( iid ), len = DOFInterfaceID_GetBytesSize( iid ) );
                used += len;
                break;
            }

        case DOFTYPE_OBJECTID:
            if ( !value || typeSpec != valueTypeSpec )
                return 0;
            {
                DOFObjectID oid;
                uint16 len;
                PCR_Assert( value && DOFValue_CheckCompatible( value, type ) );
                PCRMem_Copy( &oid, &value->data[0], sizeof(oid) );
                PCRMem_Copy( &buffer[used], DOFObjectID_GetBytes( oid ), len = DOFObjectID_GetBytesSize( oid ) );
                used += len;
                break;
            }

        case DOFTYPE_STRING:
        case DOFTYPE_BLOB:
            if ( !value || !DOFValue_CheckCompatible( value, type ) )
                return 0;
            PCR_Assert( value->length < 0x400000 );
            used += OALMarshal_CompressValue( 3, value->length, length - used, buffer + used );
            PCRMem_Copy( &buffer[used], &value->data[0], value->length );
            used += value->length;
            break;

        case DOFTYPE_ARRAY:
            if ( !value || typeSpec != valueTypeSpec )
                return 0;
            PCR_Assert( value->length < 0x8000 );
            used += OALMarshal_CompressValue( 2, value->length, length - used, buffer + used );
            {
                uint16 i = 0;
                DOFType elementType = DOFType_GetSubtype( type, 0 );
                for ( ; i < value->length; i++ )
                    used += DOFValue_Marshal( value->value[i], elementType, length - used, buffer + used );
            }
            break;

        case DOFTYPE_STRUCTURE:
            if ( !value || typeSpec != valueTypeSpec )
                return 0;
            PCR_Assert( value->length < 0x8000 );
            {
                uint16 i = 0;
                for ( ; i < value->length; i++ )
                    used += DOFValue_Marshal( value->value[i], DOFType_GetSubtype( type, i ), length - used, buffer + used );
            }
            break;

        case DOFTYPE_NULLABLE:
            if ( value )
            {
                uint32 size;
                buffer[used++] = 1; /* Not NULL */
                size = DOFValue_Marshal( value, DOFType_GetSubtype( type, 0 ), length - used, buffer + used );
                if ( !size )
                    return 0;
                used += size;
            }
            else
            {
                buffer[used++] = 0; /* NULL */
            }
            break;

        default:
            return 0;
    }
    PCR_Assert( used == mSize );
    return( used );
}


/** @internal
    Marshal a @ref DOFValue into an @ref OALStream.

    @param  type    The @ref DOFType indicating how to marshal the @a value.
    @param  value   The DOFValue to be marshalled.
    @param  stream  The OALStream in which to marshal the DOFValue.

    @return TRUE if the marshal was successful; FALSE otherwise.
**/
boolean DOFValue_MarshalToStream( const DOFValue value, const DOFType type, OALStream stream )
{
    uint32 marshalSize = DOFValue_Marshal( value, type, OALStream_GetFreeSize( stream ), OALStream_GetWritePtr( stream ) );
    if ( marshalSize )
        OALStream_AdjustWriteOffset( stream, (int32) marshalSize & 0x7FFFFFFFUL );
    return( marshalSize > 0 );
}


/**
@}
**/
