/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef DOF_PCR_H
#define DOF_PCR_H

#include <dof/pcrtypes.h>

#ifdef PCR_DEV_DOCS_ONLY

/** @mainpage The Portable C Runtime (PCR) API
    @section intro Introduction

        The Portable C Runtime (PCR) provides a common interface for projects
        to use given that different operating systems and platforms have different
        abilities or APIs that can be used. The PCR defines common system-level
        definitions, types, and functions. The dof-pcr-dev component provides the
        interface, however the implementation of these functions will be in platform
        specific implementations of the defined interface.

    @section introUsage Using This Reference

        This reference defines the APIs for use by a C-based PCR implementation.

    @section bugs How to Report a Bug

        See the README file associated with this project.

    @section addinfo Additional Information

        See the COPYING and README files associated with this project.
**/

#else

/** @defgroup PCR Portable C Runtime (PCR)
    Common system-level definitions, types, and functions.

    Different operating systems and platforms have different abilities or APIs
    that can be used.  Here, the functionality required for the DOF C OAL
    is defined and encapsulated in this portable abstraction layer.  Thus, the
    C OAL can easily be ported to new platforms and operating systems simply by
    porting this layer to the new platform.

@{
**/

#endif

/** @file
    Header file for the Portable C Runtime functions, defines and data types.
    @ingroup PCRCondition
    @ingroup PCREvent
    @ingroup PCRMem
    @ingroup PCRMutex
    @ingroup PCRRandom
    @ingroup PCRSemaphore
    @ingroup PCRString
    @ingroup PCRSwap
    @ingroup PCRThread
    @ingroup PCRTime
    @ingroup PCRTypes
**/

#ifdef __cplusplus
extern "C" {
#endif


/** Infinite timeout value. **/
#define PCRTIMEOUT_FOREVER 0xFFFFFFFF

/** Common status/error codes for system functions. **/
typedef enum PCRStatus {
    PCRSTATUS_OK            = 0x0000,  /**< Success (no errors). **/
    PCRSTATUS_ERROR         = 0x0001,  /**< General Failure.  Function call failed to execute or complete normally. **/
    PCRSTATUS_TIMEOUT       = 0x0002,  /**< Timeout reached. **/
    PCRSTATUS_DESTROYED     = 0x0003   /**< Operation cancelled due to resource being destroyed. **/
} PCRStatus;


#ifdef NDEBUG

/* No debugging: disable assertions. */
# define PCR_Assert(expr)

#else

/* Debugging: enable assertions. */

/**
 * Assert that an expression is always true.
 * This is a developer tool to verify code invariants; failure of an assertion must only be due to a coding or logic error.
 * If the expression evaluates to zero (FALSE), code will not proceed past this check.
 * @note If @c NDEBUG is defined, this check evaluates to an empty statement.
 * @param expr Expression to check.
 */
# define PCR_Assert(expr) ((expr) ? ((void)0) : PCR_AssertFail(#expr, __FILE__, __LINE__))

/**
 * Assertion failure handler.
 *
 * This function should never be called directly.
 * This handles the failure of a @ref PCR_Assert.
 *
 * @note This function should end by calling @ref PCR_Die with a code of 0.
 *
 * @param expression The expression given to @ref PCR_Assert that failed to evaluate to non-zero (TRUE), given as an
 *                   ASCII C string.
 * @param filename   The filename in which the failing @ref PCR_Assert is located, given as an ASCII C string.
 * @param line       The line number on which the failing @ref PCR_Assert is located.
 */
extern DOF_PCR_DECLSPEC void PCR_AssertFail( const char * expression, const char *filename, uint32 line );

#endif

/**
 * Validate that an expression is true.
 * This is a tool to check for fatal runtime errors where recovery is not possible. If the expression evaluates to zero
 * (FALSE), code will not proceed past this check.
 * @param expr Expression to check.
 */
#define PCR_Validate(expr) ((expr) ? ((void)0) : PCR_ValidateFail(#expr, __FILE__, __LINE__))

/**
 * Validation failure handler.
 *
 * This function should never be called directly.
 * This handles the failure of a @ref PCR_Validate.
 *
 * @note This function should end by calling @ref PCR_Die with a code of 1.
 *
 * @param expression The expression given to @ref PCR_Validate that failed to evaluate to non-zero (TRUE), given as an
 *                   ASCII C string.
 * @param filename   The filename in which the failing @ref PCR_Validate is located, given as an ASCII C string.
 * @param line       The line number on which the failing @ref PCR_Validate is located.
 */
extern DOF_PCR_DECLSPEC void PCR_ValidateFail( const char * expression, const char * filename, uint32 line );

/**
    Initialize the Portable C Runtime.
**/
extern DOF_PCR_DECLSPEC void PCR_Init( void );

/**
    Terminate use of the library and free internal resources.
**/
extern DOF_PCR_DECLSPEC void PCR_Shutdown( void );

/**
 * Platform-specific abnormal termination / catastrophic failure handler.
 * This is called to indicate a catastrophic failure has occurred and no further processing can occur (there is no recovery possible from this condition).
 * The platform may choose to halt, abort, reboot, or any other acceptable action that prevents further execution.
 * @note <B>This function must not return.</B>
 * @param code A code for internal use only.
 */
extern DOF_PCR_DECLSPEC void PCR_Die( uint32 code );


/** @addtogroup PCRThread PCRThread
    Common system-level thread operations.
@{
**/

/** Represents an allocated thread. **/
typedef struct PCRThread_t *PCRThread;

/** A runnable thread function. **/
typedef void (*PCRThreadCall)(void *);

/**
    Create and start a new running thread.

    @mustfree{PCRThread_Destroy}

    @param threadFunc   The function to execute as the new thread. This must not be NULL.
    @param threadParam  A parameter to the thread function. May be NULL.

    @return A representation of the newly created thread, or NULL if a thread
            could not be created.

    @see    PCRThread_Join
    @see    PCRThread_Destroy
**/
extern DOF_PCR_DECLSPEC PCRThread PCRThread_Create( PCRThreadCall threadFunc, void * threadParam );

/**
    Wait for a thread to finish running and cleanup.

    This function should always be called prior to @ref PCRThread_Destroy.

    @param  thread      The thread for which to wait.

    @return Returns @ref PCRSTATUS_OK if successful, or @ref PCRSTATUS_ERROR if there was an error.

    @exception Returns @ref PCRSTATUS_ERROR if there was an error.

    @see    PCRThread_Destroy
    @see    PCRThread_Create
**/
extern DOF_PCR_DECLSPEC PCRStatus PCRThread_Join( PCRThread thread );

/**
    Yield the calling thread to the OS.

    @return Returns @ref PCRSTATUS_OK if successful, or @ref PCRSTATUS_ERROR if not successful.

    @see    PCRThread_Destroy
    @see    PCRThread_Create
**/
extern DOF_PCR_DECLSPEC PCRStatus PCRThread_Yield( void );

/**
    Sleep the calling thread for a specified amount of time.

    @param mseconds Duration of the sleep, in milliseconds.
**/
extern DOF_PCR_DECLSPEC void PCRThread_Sleep( uint32 mseconds );

/**
    Free the resources associated with a thread.

    @ref PCRThread_Join should always be called prior to destroying the thread.

    @param  thread      The thread to destroy. May be NULL.

    @see    PCRThread_Join
    @see    PCRThread_Create
**/
extern DOF_PCR_DECLSPEC void PCRThread_Destroy( PCRThread thread );

/**
@}
**/


/** @addtogroup PCREvent PCREvent
    Synchronization event operations.
@{
**/

/** Represents an allocated event. **/
typedef struct PCREvent_t *PCREvent;


/**
    Create a thread synchronization event.

    @mustfree{PCREvent_Destroy}

    @param manualReset  Flag indicating whether the event must be manually reset.

    @return A representation of the newly created event, or NULL if an event
            could not be created.
**/
extern DOF_PCR_DECLSPEC PCREvent PCREvent_Create( boolean manualReset );

/**
    Signal an event.

    Depending on the style of event this will either wake up a single thread
    waiting on the event (if the manualReset flag was FALSE when the event was
    created), or it will wake up all threads waiting on the event (if the
    manualReset flag was TRUE when the event was created).

    The event must be a valid @ref PCREvent.

    @param  evt     The event to signal. This may not be NULL.
**/
extern DOF_PCR_DECLSPEC void PCREvent_Signal( PCREvent evt );

/**
    Wait for an event to occur.

    If the event is already signaled, this operation will return without
    blocking.

    The event must be a valid @ref PCREvent.  The event must not be destroyed
    while it is being waited upon.

    @param  evt  The event for which to wait. This may not be NULL.
    @param  time Maximum time to wait, in milliseconds, or @ref PCRTIMEOUT_FOREVER to wait indefinitely.

    @return Returns a status code:
            - @ref PCRSTATUS_OK if the event was signaled before the timeout expired.
            - @ref PCRSTATUS_TIMEOUT if the event was not signaled prior to the timeout.
**/
extern DOF_PCR_DECLSPEC PCRStatus PCREvent_Wait( PCREvent evt, uint32 time );

/**
    Manually reset an event.

    This has no effect on an event that was created with the manualReset flag
    set to FALSE.

    The event must be a valid @ref PCREvent.

    @param  evt     The event to reset. This may not be NULL.
**/
extern DOF_PCR_DECLSPEC void PCREvent_Reset( PCREvent evt );

/**
    Destroy an event.

    The event must not be currently waited upon.

    @param  evt     The event to destroy. This may be NULL.
**/
extern DOF_PCR_DECLSPEC void PCREvent_Destroy( PCREvent evt );

/**
@}
**/


/** @addtogroup PCRMutex PCRMutex
    Mutual exclusion operations.
@{
**/

/** Represents an allocated mutex. **/
typedef struct PCRMutex_t *PCRMutex;

/**
    Create a mutex.

    @mustfree{PCRMutex_Destroy}

    @return A representation of the newly created mutex, or NULL if a mutex
            could not be created.
**/
extern DOF_PCR_DECLSPEC PCRMutex PCRMutex_Create ( void );

/**
    Destroy a mutex.

    @param  mtx     Mutex to destroy. May be NULL.
**/
extern DOF_PCR_DECLSPEC void PCRMutex_Destroy( PCRMutex mtx );

/**
    Wait for a mutex to become available and lock it.

    If the current thread already has the mutex locked, this function will not
    block, but will keep track of the number of times it is locked (the mutex
    must be unlocked the same number of times in order to be unlocked for
    another thread to use).

    The mutex must be a valid PCRMutex.  The mutex must not be destroyed while
    it is locked.  Except for a serious programming error, this function does
    not return without the mutex locked.

    @param  mtx     Mutex to lock. This may not be NULL.
**/
extern DOF_PCR_DECLSPEC void PCRMutex_Lock   ( PCRMutex mtx );

/**
    Unlock a mutex.

    If the current thread locked the mutex multiple times, this function will reduce the lock count by 1.  The mutex is only unlocked when @b all of its recursive locks have been released.

    The mutex must be a valid PCRMutex.

    @param  mtx     Mutex to unlock. This may not be NULL.
**/
extern DOF_PCR_DECLSPEC void PCRMutex_Unlock ( PCRMutex mtx );

/**
@}
**/


/** @addtogroup PCRCondition PCRCondition
    Condition variable operations.
@{
**/

/** Represents an allocated condition variable. **/
typedef struct PCRCondition_t *PCRCondition;

/**
    Create a condition variable.

    @mustfree{PCRCondition_Destroy}

    @return A representation of the newly created condition variable, or NULL if it
            could not be created.
**/
extern DOF_PCR_DECLSPEC PCRCondition PCRCondition_Create( void );

/**
    Destroy a condition variable.

    The condition variable must not be currently in use (nothing can be waiting
    on or signaling the condition).

    @param  cond    Condition variable to destroy. May be NULL.
**/
extern DOF_PCR_DECLSPEC void PCRCondition_Destroy( PCRCondition cond );

/**
    Wait for a condition to be signaled.

    The condition variable and mutex must be valid, and the mutex must already
    be held (locked) by the calling thread.  Furthermore, the condition
    variable must not be used concurrently with different mutexes.

    @note On some platforms, a spurious wakeup may occur, causing this
    function to return as if the condition was signaled (when it actually
    was not).  It is strongly encouraged for the caller to test for a separate
    flag to know for sure whether the signal is meaningful.

    @param  cond    Condition for which to wait. Must not be NULL.
    @param  mtx     Locked mutex associated with condition. Must not be NULL.
    @param  timeout Maximum time to wait, in milliseconds.

    @return Return value is one of:
            - PCRSTATUS_OK: Condition was signaled.
            - PCRSTATUS_TIMEOUT: Timeout reached; condition was not signaled.
**/
extern DOF_PCR_DECLSPEC PCRStatus PCRCondition_Wait( PCRCondition cond, PCRMutex mtx, uint32 timeout );

/**
    Signal a condition variable.

    This will wake up at most one thread waiting on the condition variable.
    The signal is lost if no threads are waiting on the condition variable.

    The condition variable must be a valid @ref PCRCondition.

    @param  cond    Condition variable to signal. Must not be NULL.
**/
extern DOF_PCR_DECLSPEC void PCRCondition_Signal( PCRCondition cond );

/**
    Signal a condition variable.

    This will wake up all threads waiting on the condition variable.
    The signal is lost if no threads are waiting on the condition variable.

    The condition variable must be a valid PCRCondition.

    @param  cond    Condition variable to signal. Must not be NULL.
**/
extern DOF_PCR_DECLSPEC void PCRCondition_Broadcast( PCRCondition cond );

/**
@}
**/


/** @addtogroup PCRSemaphore PCRSemaphore
    Semaphore operations.
@{
**/

/** Represents an allocated semaphore. **/
typedef struct PCRSemaphore_t *PCRSemaphore;

/**
    Create a semaphore.

    @mustfree{PCRSemaphore_Destroy}

    @param count Initial semaphore signal count. The semaphore will behave as if
                 @a count @ref PCRSemaphore_Signal calls have already occurred.

    @return A representation of the newly created semaphore, or NULL if it could not be created.
**/
extern DOF_PCR_DECLSPEC PCRSemaphore PCRSemaphore_Create( uint32 count );

/**
    Destroy a semaphore.

    The semaphore must not be currently in use (nothing can be waiting on or signaling the semaphore).
    The semaphore signal count must also be equal to or greater than the starting count. This means
    that PCRSemaphore_Signal must be called for each call made to PCRSemaphore_Wiat. It is ok if
    there are extra calls to PCRSemaphore_Signal.

    @param  sem Semaphore to destroy. May be NULL.
**/
extern DOF_PCR_DECLSPEC void PCRSemaphore_Destroy( PCRSemaphore sem );

/**
    Wait indefinitely for a semaphore to be signaled.

    If the semaphore is already signaled at least once, this call returns immediately.

    The semaphore must be a valid @ref PCRSemaphore.

    @param  sem Semaphore for which to wait. Must not be NULL.
**/
extern DOF_PCR_DECLSPEC void PCRSemaphore_Wait( PCRSemaphore sem );

/**
    Signal a semaphore.

    This will wake up at most one thread waiting on the semaphore.
    If no threads are waiting, the signal is retained until the next call to @ref PCRSemaphore_Wait.

    The semaphore must be a valid @ref PCRSemaphore.

    @param  sem Semaphore to signal. Must not be NULL.

    @retval PCRSTATUS_OK if the semaphore was successfully signaled.
    @retval PCRSTATUS_ERROR if the semaphore could not be signaled (e.g., it would cause an internal overflow of the
            signal count).
**/
extern DOF_PCR_DECLSPEC PCRStatus PCRSemaphore_Signal( PCRSemaphore sem );

/**
@}
**/


/** @addtogroup PCRMem PCRMem
    Common system-level memory operations including allocation, deallocation, and copy.
@{
**/

/**
    Free a block of memory previously allocated by @ref PCRMem_Alloc.

    @param  mem Pointer to the memory to be freed. May be NULL. If memory block
                is already freed, or is invalid, then behavior is undefined.

    @see    PCRMem_Alloc
    @see    PCRMem_Realloc
**/
extern DOF_PCR_DECLSPEC void PCRMem_Free( void * mem );

/**
    Allocate a block of memory.

    Storage is dynamically allocated from the available system memory.  The
    block is @b not initialized and may contain arbitrary data.

    @mustfree{PCRMem_Free}

    @note If @ref PCRMem_Realloc is used on a value returned from this function,
                    then the value returned from @ref PCRMem_Realloc is the only
                    one that needs to be freed.

    @param  size    The total amount of memory to allocate.  This must be greater than 0.

    @return The pointer to the newly allocated memory, or NULL if the
            allocation failed.

    @see    PCRMem_Free
**/
extern DOF_PCR_DECLSPEC void * PCRMem_Alloc( uint32 size );

/**
    Change the allocated size of a block of memory.

    @param  mem     Pointer to the memory to be freed. Must not be NULL.
    @param  size    The new size of the allocated memory block.  It may be larger or smaller than the original allocated size.  This must be greater than 0.

    @mustfree{PCRMem_Free}

    @return The pointer to the newly allocated memory, or NULL if the
            reallocation failed.

    @see    PCRMem_Alloc
    @see    PCRMem_Free
**/
extern DOF_PCR_DECLSPEC void * PCRMem_Realloc( void * mem, uint32 size );

/**
    Compare the contents of two memory locations for equality.

    @param buf1 Pointer to first input buffer. May not be NULL unless @a n == 0.
    @param buf2 Pointer to second input buffer. May not be NULL unless @a n == 0.
    @param n    Byte count to compare. Must be >0 for compare to take place.

    @return The return value is a standard integer, and its exact value is platform dependent.  It may not always be correctly represented in 8 bits.  It is intended to be used only in comparison, as follows:
            - Zero ( == 0 ) if buf1 == buf2
            - Positive ( > 0 ) if buf1 > buf2
            - Negative ( < 0 ) if buf1 < buf2
**/
extern DOF_PCR_DECLSPEC int PCRMem_Compare( const void * buf1, const void * buf2, uint32 n );

/**
    Fills the contents of a memory location with a constant value.

    @param  ptr     Memory location to fill. May not be NULL unless @a n == 0.
    @param  c       Constant byte value with which to fill.
    @param  n       Number of bytes to fill starting at @a ptr. Must be >0 or nothing happens.

    @return None.
**/
extern DOF_PCR_DECLSPEC void PCRMem_Set( void *ptr, uint8 c, uint32 n );

/**
    Copy the contents of one memory location to another.
    If the memory areas overlap, behavior is undefined.  For this reason, it is recommended to use @ref PCRMem_Move instead.  However, if it can be guaranteed that the two regions do not overlap, this function may be more optimal.

    @param  dest    Destination memory location. May not be NULL.
    @param  src     Source memory location. May not be NULL unless @a n == 0.
    @param  n       Number of bytes to copy from @a src to @a dest. Must be >0 or nothing happens.

    @return None.

    @see    PCRMem_Move
**/
extern DOF_PCR_DECLSPEC void PCRMem_Copy( void *dest, const void *src, uint32 n );

/**
    Copy the contents of one memory location to another.
    The copy is correct even if the memory areas overlap.  For this reason, it is recommended to use this function in place of @ref PCRMem_Copy.

    @param  dest    Destination memory location. May not be NULL unless @a n == 0.
    @param  src     Source memory location. May not be NULL unless @a n == 0.
    @param  n       Number of bytes to copy from @a src to @a dest. Must be >0 or nothing happens.

    @return None.

    @see    PCRMem_Copy
**/
extern DOF_PCR_DECLSPEC void PCRMem_Move( void *dest, const void *src, uint32 n );


/**
@}
**/


/** @addtogroup PCRString PCRString
    Common system-level string operations.
@{
**/


/**
    Function to determine the character length of a string.
    Equivalent to the standard C library function strlen().

    @param  s String of which to calculate the length. Must not be NULL.

    @return Length of the string, in characters, excluding the terminating null.
**/
extern DOF_PCR_DECLSPEC uint32 PCRString_Length( const char *s );

/**
    Appends up to the first n characters of pStr2 to the end of pStr1.
    Equivalent to the standard C library function strncat().

    @param pStr1 - Pointer to the string that pStr2 will be added to. Should not be NULL.
    @param pStr2 - Pointer to the string to be added at the end of pStr1. Should not be NULL.
    @param n     The maximum number. Can be @c 0, in which case no action is taken.
**/
extern DOF_PCR_DECLSPEC void PCRString_Concat( char * pStr1, const char * pStr2, uint32 n );

/**
    Copy a string from src to dest, up to the given length.
    Equivalent to the standard C library function strncpy().

    @param dest  Pointer to a buffer to receive the copied string. Must not be NULL.
    @param src   Pointer to a string to be copied. Must not be NULL.
    @param n     Maximum number of characters to copy. Can be @c 0, in which case no action is taken.
**/
extern DOF_PCR_DECLSPEC void PCRString_Copy( char * dest, const char * src, uint32 n );


/**
@}
**/


/** @addtogroup PCRTime PCRTime
    Common system-level time operations.
@{
**/


/**
    Report the current value of the system clock, in milliseconds.

    This is not an absolute time, but may be used to track relative time with millisecond resolution.  The granularity and accuracy is platform-dependent.

    @return Returns the current value of the millisecond clock.
**/
extern DOF_PCR_DECLSPEC uint32 PCRTime_GetMS      ( void );


/**
@}
**/


/** @addtogroup PCRRandom PCRRandom
    A Pseudo-Random Number Generator implementation.

    @note   The quality of the randomness generated is platform-dependent, and
            may not be suitable for cryptography or security applications.
@{
**/


/**
    Generate a sequence of pseudo-random data, filling the given buffer.

    @note   The quality of the random data is platform-dependent, and may not
            be suitable for cryptography or security applications.

    @param  len Length of buffer to be filled.
    @param  buf Buffer to be filled with random data. Must not be NULL if len is greater than zero.
**/
extern DOF_PCR_DECLSPEC void PCRRandom_Generate( uint32 len, uint8 * buf );

/**
    Set the seed for the Pseudo-Random Number Generator.

    @param  seed New seed to use for the generator.
**/
extern DOF_PCR_DECLSPEC void PCRRandom_Seed( uint32 seed );

/**
    Retrieve a pseudo-random number.

    @note   The quality of the random number is platform-dependent, and may not
            be suitable for cryptography or security applications.

    @return A 32-bit pseudo-random number.
**/
extern DOF_PCR_DECLSPEC uint32 PCRRandom_Get( void );




/**
@}
**/


/** @addtogroup PCRSecureRandom PCRSecureRandom
    A Cryptographically Secure Random Number Generator implementation.

    @note   The quality of the randomness should be suitable for cryptography and security applications.
@{
**/


/**
    Generate a sequence of random data, filling the given buffer.

    @note   The quality of the random data is platform-dependent, and may not
            be truly random (depending on the amount of entropy available).

    @param  len Length of buffer to be filled.
    @param  buf Buffer to be filled with random data. Must not be NULL if len is greater than zero.
**/
extern DOF_PCR_DECLSPEC void PCRSecureRandom_Generate( uint32 len, uint8 * buf );

/**
    Retrieve a random number.

    @note   The quality of the random number is platform-dependent, and may not
            be truly random (depending on the amount of entropy available).

    @return A 32-bit random number.
**/
extern DOF_PCR_DECLSPEC uint32 PCRSecureRandom_Get( void );

/**
@}
**/


/** @addtogroup PCRSwap PCRSwap
    Network byte-order conversion, based on machine endianness.
@{
**/


/**
 * Byte-swap an 8-bit value to/from Network Byte Order.  Single-byte values never need to be converted; this macro exists only for completeness.
 * @param value The value to convert.
 * @return The value, in Network Byte Order.
 */
# define PCRSwap_Net8(value)     ( (uint8)  ( (value) ) )


#ifdef WORDS_BIGENDIAN

# define PCRSwap_Net16(value)    ( (uint16) ( (value) ) )
# define PCRSwap_Net32(value)    ( (uint32) ( (value) ) )
# define PCRSwap_Net64(value)    ( (uint64) ( (value) ) )

#else

/**
 * Byte-swap a 16-bit value to/from Network Byte Order.  On platforms where the native and network byte order are identical, this routine performs no conversion.
 * @param value The value to convert.
 * @return The value, in Network Byte Order.
 */
extern DOF_PCR_DECLSPEC uint16 PCRSwap_Net16( uint16 value );

/**
 * Byte-swap a 32-bit value to/from Network Byte Order.  On platforms where the native and network byte order are identical, this routine performs no conversion.
 * @param value The value to convert.
 * @return The value, in Network Byte Order.
 */
extern DOF_PCR_DECLSPEC uint32 PCRSwap_Net32( uint32 value );

/**
 * Byte-swap a 64-bit value to/from Network Byte Order.  On platforms where the native and network byte order are identical, this routine performs no conversion.
 * @param value The value to convert.
 * @return The value, in Network Byte Order.
 */
extern DOF_PCR_DECLSPEC uint64 PCRSwap_Net64( uint64 value );

#endif


/**
@}
**/




#ifdef __cplusplus
}
#endif  /* extern "C" */


/**
@}
**/

#endif /* DOF_PCR_H */
