/*
**  Copyright (c) 2014-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include "config.h"
#include <dof/pcr.h>
#include "pcrsync.h"

/* Standard headers */
#include <pthread.h>
#include <time.h>
#include <errno.h>


/* Internal Event representation. **/
struct PCRCondition_t
{
    pthread_cond_t cond;
};


PCRCondition PCRCondition_Create(void)
{
    PCRCondition cond;

    cond = (PCRCondition)PCRMem_Alloc(sizeof(struct PCRCondition_t));
    if(cond != NULL)
    {
        PCRMem_Set( cond, 0, sizeof(struct PCRCondition_t) );
        if ( pthread_cond_init(&cond->cond, NULL) != 0 )
        {
            PCRMem_Free(cond);
            cond = NULL;
        }
    }
    return(cond);
}


void PCRCondition_Destroy(PCRCondition cond)
{
    if(cond != NULL)
    {
        int result = pthread_cond_destroy(&cond->cond);
        /*
         * The only way this can fail is if cond is not valid (e.g., memory
         * corruption), or it is currently in use (programming error).
         */
        PCR_Assert(result == 0);
        PCRMem_Free(cond);
    }
}


PCRStatus PCRCondition_Wait(PCRCondition cond, PCRMutex mtx, uint32 timeout)
{
    int result = 0;
    PCR_Assert(cond);
    PCR_Assert(mtx);

    if ( timeout == PCRTIMEOUT_FOREVER )
        result = pthread_cond_wait(&cond->cond, (pthread_mutex_t*) mtx);
    else
    {
        struct timespec ts;

        PCRSync_GetAbsoluteTimeFromNow( timeout, &ts );
        while( (result = pthread_cond_timedwait(&cond->cond, (pthread_mutex_t*) mtx, &ts)) == EINTR )
            ;
        if ( result == ETIMEDOUT )
            return(PCRSTATUS_TIMEOUT);
    }
    /*
     * At this point, the only errors possible are coding errors (misuse of the API):
     *   The mutex was not owned (locked) by the current thread, or
     *   A different mutex was supplied for concurrent waits on the same condition variable.
     */
    PCR_Assert( result == 0 );
    return(PCRSTATUS_OK);
}


void PCRCondition_Signal(PCRCondition cond)
{
    int result;
    PCR_Assert(cond);
    result = pthread_cond_signal(&cond->cond);
    /* The only way this can fail is if cond is not valid (e.g., memory corruption). */
    PCR_Assert( result == 0 );
}


void PCRCondition_Broadcast(PCRCondition cond)
{
    int result;
    PCR_Assert(cond);
    result = pthread_cond_broadcast(&cond->cond);
    /* The only way this can fail is if cond is not valid (e.g., memory corruption). */
    PCR_Assert( result == 0 );
}

