/*
**  Copyright (c) 2014-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include "config.h"
#include <dof/pcr.h>
#include "pcrsync.h"

/* Standard headers */
#include <pthread.h>
#include <time.h>
#include <errno.h>

/* Internal Event representation. **/
struct PCREvent_t
{
    pthread_cond_t cond;
    pthread_mutex_t mutex;
    uint32 waitCount;
    unsigned int manualReset : 1;
    unsigned int signaled    : 1;
};


PCREvent PCREvent_Create(boolean manualReset)
{
    PCREvent evt;

    evt = (PCREvent)PCRMem_Alloc(sizeof(struct PCREvent_t));
    if(evt != NULL)
    {
        PCRMem_Set( evt, 0, sizeof(struct PCREvent_t) );
        if ( pthread_mutex_init(&evt->mutex, NULL) == 0 &&
             pthread_cond_init(&evt->cond, NULL) == 0 )
        {
            /* Since evt->manualReset is a bitfield, we need to ensure the parameter (which may not be 0 or 1) will fit into a single bit.  Hence the double-not. */
            evt->manualReset = !!manualReset;
            evt->signaled = FALSE;
            evt->waitCount = 0;
        }
        else
        {
            PCRMem_Free( evt );
            evt = NULL;
        }
    }
    return(evt);
}


void PCREvent_Signal(PCREvent evt)
{
    int result;

    PCR_Assert( evt );
    result = pthread_mutex_lock(&evt->mutex);
    /* The mutex is owned by the event, so this must always be safe. */
    PCR_Assert( result == 0 );
    if((evt->manualReset) || (evt->waitCount == 0))
        evt->signaled = TRUE;
    if(evt->waitCount != 0) {
        if(evt->manualReset)
            result = pthread_cond_broadcast(&evt->cond);
        else
            result = pthread_cond_signal(&evt->cond);
        /* The condition is owned by the event, so this must always be safe. */
        PCR_Assert( result == 0 );
    }
    result = pthread_mutex_unlock(&evt->mutex);
    PCR_Assert( result == 0 );
}


PCRStatus PCREvent_Wait(PCREvent evt, uint32 timeout)
{
    int result;
    PCRStatus myresult = PCRSTATUS_OK;

    /* The waitCount handles the case where an event is signaled before the wait. */
    PCR_Assert(evt);
    result = pthread_mutex_lock(&evt->mutex);
    /* The mutex is owned by the event, so this must always be safe. */
    PCR_Assert( result == 0 );
    evt->waitCount++;
    if(evt->signaled) {
        if(! evt->manualReset) {
            evt->signaled = FALSE;
        }
    } else {
        if(timeout == PCRTIMEOUT_FOREVER) {
            result = pthread_cond_wait(&evt->cond, &evt->mutex);
            PCR_Assert( result == 0 );
        } else {
            struct timespec ts;
            PCRSync_GetAbsoluteTimeFromNow( timeout, &ts );
            while( (result = pthread_cond_timedwait(&evt->cond, &evt->mutex, &ts)) == EINTR ){
                ;
            }
            PCR_Assert( result == 0 || result == ETIMEDOUT );
        }
    }
    evt->waitCount--;

    if (result == ETIMEDOUT)
        myresult = PCRSTATUS_TIMEOUT;

    result = pthread_mutex_unlock(&evt->mutex);
    PCR_Assert( result == 0);

    return(myresult);
}


void PCREvent_Reset(PCREvent evt)
{
    int result;

    PCR_Assert( evt );
    result = pthread_mutex_lock(&evt->mutex);
    /* The mutex is owned by the event, so this must always be safe. */
    PCR_Assert( result == 0 );
    if ( evt->manualReset )
        evt->signaled = FALSE;
    result = pthread_mutex_unlock(&evt->mutex);
    PCR_Assert( result == 0 );
}


void PCREvent_Destroy(PCREvent evt)
{
    int result;
    /* There must not be anything waiting on the event! */
    if(evt != NULL) {
        result = pthread_mutex_lock(&evt->mutex);
        /* The mutex is owned by the event, so this must always be safe. */
        PCR_Assert( result == 0 );
        evt->manualReset = TRUE;
        evt->signaled = TRUE;
        
        /* There must not be anything waiting on the event! */
        PCR_Assert( evt->waitCount == 0 );

        result = pthread_mutex_unlock(&evt->mutex);
        PCR_Assert( result == 0 );

        /* Clean up handles */
        result = pthread_cond_destroy(&evt->cond);
        PCR_Assert( result == 0 );
        result = pthread_mutex_destroy(&evt->mutex);
        PCR_Assert( result == 0 );
        PCRMem_Free(evt);
    }
}

