/*
**  Copyright (c) 2014-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include "config.h"
#include <dof/pcr.h>
#include <internal/mutex.h>
#include "pcrsync.h"

/* Standard headers */
#include <pthread.h>
#include <time.h>

static pthread_mutexattr_t our_attr;
static boolean attr_inited = FALSE;

/* Internal Event representation. **/
struct PCRMutex_t
{
    pthread_mutex_t mutex;
};


/*
    Initialize a global mutex attribute used to create mutexes.
*/
void PCRMutex_Init(void)
{
    if(!attr_inited)
    {
        PCRMem_Set(&our_attr, 0, sizeof(our_attr));
        pthread_mutexattr_init(&our_attr);
        pthread_mutexattr_settype(&our_attr, PTHREAD_MUTEX_RECURSIVE);
        attr_inited = TRUE;
    }
}


PCRMutex PCRMutex_Create(void)
{
    PCRMutex mtx = NULL;

    mtx = (PCRMutex)PCRMem_Alloc(sizeof(struct PCRMutex_t));
    if(mtx != NULL)
    {
        PCRMem_Set( mtx, 0, sizeof(*mtx) );
        if(!pthread_mutex_init(&mtx->mutex, &our_attr)) {
            return mtx;
        }
        PCRMem_Free(mtx);
    }
    return NULL;
}


void PCRMutex_Destroy(PCRMutex mtx)
{
    if(mtx != NULL)
    {
        int result = pthread_mutex_destroy(&mtx->mutex);
        /*
         * The only way this can fail is if the mutex is not valid (e.g., memory
         * corruption), or it is currently locked (programming error).
         */
        PCR_Assert( result == 0 );
        PCRMem_Free(mtx);
    }
}


void PCRMutex_Lock(PCRMutex mtx)
{
    int result;
    PCR_Assert( mtx );
#ifdef PCRMUTEX_DEADLOCK_TIMER
# ifdef HAVE_PTHREAD_MUTEX_TIMEDLOCK
    {
        struct timespec ts;
        PCRSync_GetAbsoluteTimeFromNow( PCRMUTEX_DEADLOCK_TIMER, &ts );
        result = pthread_mutex_timedlock( &mtx->mutex, &ts );
    }
# else
#  error "The pthread_mutex_timedlock function is not supported."
# endif
#else
    result = pthread_mutex_lock( &mtx->mutex );
#endif
    /*
     * The only way this fails is if the mutex is invalid (memory corruption),
     * if the recursive reference count is exceeded (runaway recursion), or if
     * a deadlock is detected (programming error).  If DEADLOCK_TIMER is
     * defined, then the timeout was reached, indicating a probable deadlock.
     */
    PCR_Assert( result == 0 );
}


void PCRMutex_Unlock(PCRMutex mtx)
{
    int result;

    PCR_Assert( mtx );
    result = pthread_mutex_unlock(&mtx->mutex);
    PCR_Assert( result == 0 );
}

