/*
**  Copyright (c) 2014-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include "config.h"
#include <dof/pcr.h>
#include "pcrsync.h"

/* Standard headers */
#ifndef APPLE_SEM
#include <semaphore.h>
#include <errno.h>
#else
#include <dispatch/dispatch.h>
#endif


/* Internal Semaphore representation. */
struct PCRSemaphore_t
{
#ifndef APPLE_SEM
    sem_t sem;
    uint32 count;
#else
    dispatch_semaphore_t sem;
#endif
};


PCRSemaphore PCRSemaphore_Create(uint32 count)
{
    PCRSemaphore sem;

    sem = (PCRSemaphore)PCRMem_Alloc(sizeof(struct PCRSemaphore_t));
    if(sem != NULL)
    {
        PCRMem_Set( sem, 0, sizeof(struct PCRSemaphore_t) );
#ifndef APPLE_SEM
        sem->count = count;
        if( sem_init(&sem->sem, 0, count) != 0 )
#else
        if( (sem->sem = dispatch_semaphore_create(count)) == NULL )
#endif
        {
            PCRMem_Free(sem);
            sem = NULL;
        }
    }
    return(sem);
}


void PCRSemaphore_Destroy(PCRSemaphore sem)
{
    if(sem != NULL)
    {
#ifndef APPLE_SEM
        int result = sem_destroy(&sem->sem);
        int scount;
        /*
         * The only way this can fail is if sem is not valid (e.g., memory corruption).
         * Undefined behavior will result if it is still being waited on in other threads.
         */
        PCR_Assert(result == 0);
        result = sem_getvalue(&sem->sem, &scount);
        PCR_Validate(scount >= sem->count);
#else
        dispatch_release(sem->sem);
#endif

        PCRMem_Free(sem);
    }
}


void PCRSemaphore_Wait(PCRSemaphore sem)
{
    int result;
    PCR_Assert(sem);

#ifndef APPLE_SEM
    while( (result = sem_wait(&sem->sem)) == EINTR )
        ;
    /*
     * At this point, the only error possible is if sem is not valid (e.g., memory corruption).
     */
#else
    result = dispatch_semaphore_wait(sem->sem, DISPATCH_TIME_FOREVER);
#endif
    PCR_Assert( result == 0 );
}


PCRStatus PCRSemaphore_Signal(PCRSemaphore sem)
{
#ifndef APPLE_SEM
    int result;
#endif
    PCR_Assert(sem);
    
#ifndef APPLE_SEM
    result = sem_post(&sem->sem);
    /*
     * The only way this can fail is if sem is not valid (e.g., memory corruption), or the signal count has overflowed.
     */
    if ( result == 0 )
        return PCRSTATUS_OK;
    return PCRSTATUS_ERROR;
#else
    dispatch_semaphore_signal(sem->sem);
    return PCRSTATUS_OK;
#endif
}

